--
--  Copyright (C) 2014-2023, AdaCore
--  SPDX-License-Identifier: Apache-2.0
--

with Ada.Containers;                  use Ada.Containers;
with Ada.Containers.Hashed_Maps;
with Ada.Containers.Vectors;
with Ada.Directories;
with Ada.Exceptions;
with Ada.Finalization;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Ada.Strings.Wide_Wide_Unbounded; use Ada.Strings.Wide_Wide_Unbounded;
pragma Warnings (Off, "internal");
with Ada.Strings.Wide_Wide_Unbounded.Aux;
use Ada.Strings.Wide_Wide_Unbounded.Aux;
pragma Warnings (On, "internal");
with Ada.Text_IO;                     use Ada.Text_IO;
with Ada.Unchecked_Conversion;
with Ada.Unchecked_Deallocation;
with System;
with GNAT.Task_Lock;
with GNAT.Traceback.Symbolic;
with GNATCOLL.Traces;
with Langkit_Support.Adalog.Debug;
with Langkit_Support.Hashes; use Langkit_Support.Hashes;
with Langkit_Support.Images; use Langkit_Support.Images;
with Langkit_Support.Relative_Get;
with Libadalang.Private_Converters;
use Libadalang.Private_Converters;
with Libadalang.Introspection_Implementation;
pragma Warnings (Off, "referenced");
          with Ada.Containers.Hashed_Sets;
          with Libadalang.Env_Hooks;
            use Libadalang.Env_Hooks;
          with Libadalang.Implementation.Extensions;
            use Libadalang.Implementation.Extensions;
          with Libadalang.Internal_Default_Provider;
          with Libadalang.Sources;
pragma Warnings (On, "referenced");
package body Libadalang.Implementation is
   use Precomputed_Symbols;
   pragma Warnings (Off, "has no effect");
   use Solver;
   pragma Warnings (On, "has no effect");
   package Context_Vectors is new Ada.Containers.Vectors
     (Index_Type   => Positive,
      Element_Type => Internal_Context);
   type Contexts_Destructor is limited
      new Ada.Finalization.Limited_Controlled with null record;
   overriding procedure Finalize (CD : in out Contexts_Destructor);
   --  Helper to destroy all contexts when terminating the process
   package Context_Pool is
      procedure Acquire (Context : out Internal_Context)
         with Post => Context /= null;
      --  If a context is free for reuse, increment its serial number and
      --  return it. Otherwise, allocate a new one. In any case, this does not
      --  initialize it, except for the Serial_Number field.
      procedure Release (Context : in out Internal_Context)
         with Pre  => Context /= null,
              Post => Context = null;
      --  Tag Context as free for reuse and set it to null
      procedure Free;
      --  Free all contexts in this pool. Intended to be called only when the
      --  process is terminating, to avoid reported memory leaks.
   private
      Available : Context_Vectors.Vector;
      --  List of allocated contexts that can be re-used right now
      CD : Contexts_Destructor with Unreferenced;
      --  Singleton whose only purpose is to free all contexts in Available
      --  when finalized.
   end Context_Pool;
   procedure Register_Destroyable_Helper
     (Unit    : Internal_Unit;
      Object  : System.Address;
      Destroy : Destroy_Procedure);
   --  Common underlying implementation for Register_Destroyable_Gen
   pragma Warnings (Off, "referenced");
   function Construct_Entity_Array
     (V : AST_Envs.Entity_Vectors.Vector) return Internal_Entity_Array_Access;
   pragma Warnings (On, "referenced");
   procedure Destroy (Env : in out Lexical_Env_Access);
   function Snaps_At_Start (Self : Bare_Ada_Node) return Boolean;
   function Snaps_At_End (Self : Bare_Ada_Node) return Boolean;
   --  Those maps are used to give unique ids to lexical envs while pretty
   --  printing them.
   package Address_To_Id_Maps is new Ada.Containers.Hashed_Maps
     (Lexical_Env, Integer, Hash, "=");
   type Dump_Lexical_Env_State is record
      Env_Ids : Address_To_Id_Maps.Map;
      --  Mapping: Lexical_Env -> Integer, used to remember which unique Ids we
      --  assigned to the lexical environments we found.
      Next_Id : Positive := 1;
      --  Id to assign to the next unknown lexical environment
      Root_Env : Lexical_Env;
      --  Lexical environment we consider a root (this is the Root_Scope from
      --  the current analysis context), or null if unknown.
   end record;
   --  Holder for the state of lexical environment dumpers
   function Get_Env_Id
     (E : Lexical_Env; State : in out Dump_Lexical_Env_State) return String;
   --  If E is known, return its unique Id from State. Otherwise, assign it a
   --  new unique Id and return it.
   ------------------------
   -- Precomputed_Symbol --
   ------------------------
   pragma Warnings (Off, "referenced");
   function Precomputed_Symbol
     (Index : Precomputed_Symbol_Index) return Text_Type is
   pragma Warnings (On, "referenced");
   begin
         declare
            Raw_Text : constant Text_Type := (case Index is
               when Precomputed_Sym => """&""",
               when Precomputed_Sym_1 => """&""""+""",
               when Precomputed_Sym_10 => """=""",
               when Precomputed_Sym_11 => """>""",
               when Precomputed_Sym_12 => """>=""",
               when Precomputed_Sym_2 => """*""",
               when Precomputed_Sym_20 => "<<>>",
               when Precomputed_Sym_3 => """**""",
               when Precomputed_Sym_4 => """+""",
               when Precomputed_Sym_5 => """-""",
               when Precomputed_Sym_6 => """/""",
               when Precomputed_Sym_7 => """/=""",
               when Precomputed_Sym_8 => """<""",
               when Precomputed_Sym_9 => """<=""",
               when Precomputed_Sym_Abort => "abort",
               when Precomputed_Sym_Abort_Signal => "Abort_Signal",
               when Precomputed_Sym_Abort_Signal_256 => "abort_signal_",
               when Precomputed_Sym_Abs => """abs""",
               when Precomputed_Sym_Abs_257 => "abs",
               when Precomputed_Sym_Abstract => "abstract",
               when Precomputed_Sym_Abstract_State => "Abstract_State",
               when Precomputed_Sym_Accept => "accept",
               when Precomputed_Sym_Access => "Access",
               when Precomputed_Sym_Access_260 => "access",
               when Precomputed_Sym_Ada => "Ada",
               when Precomputed_Sym_Ada_261 => "ada",
               when Precomputed_Sym_Address => "Address",
               when Precomputed_Sym_Address_Size => "Address_Size",
               when Precomputed_Sym_Adjacent => "Adjacent",
               when Precomputed_Sym_Aft => "Aft",
               when Precomputed_Sym_Aliased => "aliased",
               when Precomputed_Sym_Alignment => "Alignment",
               when Precomputed_Sym_All => "all",
               when Precomputed_Sym_And => """and""",
               when Precomputed_Sym_And_264 => "and",
               when Precomputed_Sym_Array => "array",
               when Precomputed_Sym_Asm_Input => "Asm_Input",
               when Precomputed_Sym_Asm_Input_Operand => "Asm_Input_Operand",
               when Precomputed_Sym_Asm_Output => "Asm_Output",
               when Precomputed_Sym_Asm_Output_Operand => "Asm_Output_Operand",
               when Precomputed_Sym_Assert => "Assert",
               when Precomputed_Sym_Assert_And_Cut => "Assert_And_Cut",
               when Precomputed_Sym_Assume => "Assume",
               when Precomputed_Sym_Asynchronous => "Asynchronous",
               when Precomputed_Sym_At => "at",
               when Precomputed_Sym_Atomic => "Atomic",
               when Precomputed_Sym_Atomic_Components => "Atomic_Components",
               when Precomputed_Sym_Attach_Handler => "Attach_Handler",
               when Precomputed_Sym_Aux_Dec => "Aux_DEC",
               when Precomputed_Sym_Base => "Base",
               when Precomputed_Sym_Begin => "begin",
               when Precomputed_Sym_Bit_Order => "Bit_Order",
               when Precomputed_Sym_Body => "body",
               when Precomputed_Sym_Boolean => "Boolean",
               when Precomputed_Sym_Callable => "Callable",
               when Precomputed_Sym_Caller => "Caller",
               when Precomputed_Sym_Case => "case",
               when Precomputed_Sym_Ceiling => "Ceiling",
               when Precomputed_Sym_Character => "Character",
               when Precomputed_Sym_Class => "Class",
               when Precomputed_Sym_Compile_Time_Error => "Compile_Time_Error",
               when Precomputed_Sym_Compile_Time_Warning => "Compile_Time_Warning",
               when Precomputed_Sym_Component_Size => "Component_Size",
               when Precomputed_Sym_Compose => "Compose",
               when Precomputed_Sym_Constant => "constant",
               when Precomputed_Sym_Constant_Indexing => "Constant_Indexing",
               when Precomputed_Sym_Constrained => "Constrained",
               when Precomputed_Sym_Contract_Cases => "Contract_Cases",
               when Precomputed_Sym_Convention => "Convention",
               when Precomputed_Sym_Copy_Sign => "Copy_Sign",
               when Precomputed_Sym_Count => "Count",
               when Precomputed_Sym_Cursor => "Cursor",
               when Precomputed_Sym_Debug => "Debug",
               when Precomputed_Sym_Declare => "declare",
               when Precomputed_Sym_Default_Bit_Order => "Default_Bit_Order",
               when Precomputed_Sym_Default_Initial_Condition => "Default_Initial_Condition",
               when Precomputed_Sym_Definite => "Definite",
               when Precomputed_Sym_Delay => "delay",
               when Precomputed_Sym_Delta => "Delta",
               when Precomputed_Sym_Delta_273 => "delta",
               when Precomputed_Sym_Depends => "Depends",
               when Precomputed_Sym_Descriptor_Size => "Descriptor_Size",
               when Precomputed_Sym_Digits => "Digits",
               when Precomputed_Sym_Digits_274 => "digits",
               when Precomputed_Sym_Discard_Names => "Discard_Names",
               when Precomputed_Sym_Do => "do",
               when Precomputed_Sym_Duration => "Duration",
               when Precomputed_Sym_Dynamic_Predicate => "Dynamic_Predicate",
               when Precomputed_Sym_Elaborate_Body => "Elaborate_Body",
               when Precomputed_Sym_Element => "Element",
               when Precomputed_Sym_Else => "else",
               when Precomputed_Sym_Elsif => "elsif",
               when Precomputed_Sym_End => "end",
               when Precomputed_Sym_Ensures => "Ensures",
               when Precomputed_Sym_Entry => "entry",
               when Precomputed_Sym_Enum_Rep => "Enum_Rep",
               when Precomputed_Sym_Enum_Val => "Enum_Val",
               when Precomputed_Sym_Epsilon => "Epsilon",
               when Precomputed_Sym_Exception => "exception",
               when Precomputed_Sym_Exception_Id => "Exception_Id",
               when Precomputed_Sym_Exception_Occurrence => "Exception_Occurrence",
               when Precomputed_Sym_Exceptions => "Exceptions",
               when Precomputed_Sym_Exit => "exit",
               when Precomputed_Sym_Exponent => "Exponent",
               when Precomputed_Sym_Export => "Export",
               when Precomputed_Sym_External_Tag => "External_Tag",
               when Precomputed_Sym_Finalization_Size => "Finalization_Size",
               when Precomputed_Sym_First => "First",
               when Precomputed_Sym_First_Bit => "First_Bit",
               when Precomputed_Sym_Fixed_Value => "Fixed_Value",
               when Precomputed_Sym_Floor => "Floor",
               when Precomputed_Sym_For => "for",
               when Precomputed_Sym_Fore => "Fore",
               when Precomputed_Sym_Fraction => "Fraction",
               when Precomputed_Sym_Function => "function",
               when Precomputed_Sym_Generic => "generic",
               when Precomputed_Sym_Ghost => "Ghost",
               when Precomputed_Sym_Global => "Global",
               when Precomputed_Sym_Goto => "goto",
               when Precomputed_Sym_Has_Access_Values => "Has_Access_Values",
               when Precomputed_Sym_Has_Discriminants => "Has_Discriminants",
               when Precomputed_Sym_Has_Tagged_Values => "Has_Tagged_Values",
               when Precomputed_Sym_Identity => "Identity",
               when Precomputed_Sym_If => "if",
               when Precomputed_Sym_Image => "Image",
               when Precomputed_Sym_Img => "Img",
               when Precomputed_Sym_Implicit_Dereference => "Implicit_Dereference",
               when Precomputed_Sym_Import => "Import",
               when Precomputed_Sym_Import_Function => "Import_Function",
               when Precomputed_Sym_Import_Procedure => "Import_Procedure",
               when Precomputed_Sym_In => "in",
               when Precomputed_Sym_Independent => "Independent",
               when Precomputed_Sym_Independent_Components => "Independent_Components",
               when Precomputed_Sym_Index => "Index",
               when Precomputed_Sym_Initialized => "Initialized",
               when Precomputed_Sym_Inline => "Inline",
               when Precomputed_Sym_Input => "Input",
               when Precomputed_Sym_Integer => "Integer",
               when Precomputed_Sym_Integer_Literal => "Integer_Literal",
               when Precomputed_Sym_Integer_Value => "Integer_Value",
               when Precomputed_Sym_Interface => "Interface",
               when Precomputed_Sym_Interface_288 => "interface",
               when Precomputed_Sym_Interrupt_Handler => "Interrupt_Handler",
               when Precomputed_Sym_Invalid_Value => "Invalid_Value",
               when Precomputed_Sym_Invariant => "Invariant",
               when Precomputed_Sym_Invariant_Class => "Invariant'Class",
               when Precomputed_Sym_Is => "is",
               when Precomputed_Sym_Iterable => "Iterable",
               when Precomputed_Sym_Iterator_Element => "Iterator_Element",
               when Precomputed_Sym_Iterator_Interfaces => "Iterator_Interfaces",
               when Precomputed_Sym_Large => "Large",
               when Precomputed_Sym_Last => "Last",
               when Precomputed_Sym_Last_Bit => "Last_Bit",
               when Precomputed_Sym_Left => "left",
               when Precomputed_Sym_Length => "Length",
               when Precomputed_Sym_Limited => "limited",
               when Precomputed_Sym_Loop => "loop",
               when Precomputed_Sym_Loop_Entry => "Loop_Entry",
               when Precomputed_Sym_Loop_Invariant => "Loop_Invariant",
               when Precomputed_Sym_Machine => "Machine",
               when Precomputed_Sym_Machine_Code => "Machine_Code",
               when Precomputed_Sym_Machine_Emax => "Machine_Emax",
               when Precomputed_Sym_Machine_Emin => "Machine_Emin",
               when Precomputed_Sym_Machine_Mantissa => "Machine_Mantissa",
               when Precomputed_Sym_Machine_Overflows => "Machine_Overflows",
               when Precomputed_Sym_Machine_Radix => "Machine_Radix",
               when Precomputed_Sym_Machine_Rounding => "Machine_Rounding",
               when Precomputed_Sym_Machine_Rounds => "Machine_Rounds",
               when Precomputed_Sym_Mantissa => "Mantissa",
               when Precomputed_Sym_Max => "Max",
               when Precomputed_Sym_Max_Integer_Size => "Max_Integer_Size",
               when Precomputed_Sym_Max_Size_In_Storage_Elements => "Max_Size_In_Storage_Elements",
               when Precomputed_Sym_Maximum_Alignment => "Maximum_Alignment",
               when Precomputed_Sym_Min => "Min",
               when Precomputed_Sym_Mod => """mod""",
               when Precomputed_Sym_Mod_141 => "Mod",
               when Precomputed_Sym_Mod_293 => "mod",
               when Precomputed_Sym_Model_Epsilon => "Model_Epsilon",
               when Precomputed_Sym_Model_Mantissa => "Model_Mantissa",
               when Precomputed_Sym_Model_Small => "Model_Small",
               when Precomputed_Sym_Modulus => "Modulus",
               when Precomputed_Sym_New => "new",
               when Precomputed_Sym_Nextpart => "__nextpart",
               when Precomputed_Sym_No_Elaboration_Code => "No_Elaboration_Code",
               when Precomputed_Sym_No_Return => "No_Return",
               when Precomputed_Sym_Not => """not""",
               when Precomputed_Sym_Not_295 => "not",
               when Precomputed_Sym_Null => "null",
               when Precomputed_Sym_Object_Size => "Object_Size",
               when Precomputed_Sym_Of => "of",
               when Precomputed_Sym_Old => "Old",
               when Precomputed_Sym_Or => """or""",
               when Precomputed_Sym_Or_298 => "or",
               when Precomputed_Sym_Others => "others",
               when Precomputed_Sym_Out => "out",
               when Precomputed_Sym_Output => "Output",
               when Precomputed_Sym_Overriding => "overriding",
               when Precomputed_Sym_Pack => "Pack",
               when Precomputed_Sym_Package => "package",
               when Precomputed_Sym_Pos => "Pos",
               when Precomputed_Sym_Position => "Position",
               when Precomputed_Sym_Post => "Post",
               when Precomputed_Sym_Post_Class => "Post'Class",
               when Precomputed_Sym_Postcondition => "Postcondition",
               when Precomputed_Sym_Postcondition_Class => "Postcondition'Class",
               when Precomputed_Sym_Pragma => "pragma",
               when Precomputed_Sym_Pre => "Pre",
               when Precomputed_Sym_Pre_Class => "Pre'Class",
               when Precomputed_Sym_Precondition => "Precondition",
               when Precomputed_Sym_Precondition_Class => "Precondition'Class",
               when Precomputed_Sym_Pred => "Pred",
               when Precomputed_Sym_Predicate => "Predicate",
               when Precomputed_Sym_Predicate_Failure => "Predicate_Failure",
               when Precomputed_Sym_Preelaborate => "Preelaborate",
               when Precomputed_Sym_Private => "private",
               when Precomputed_Sym_Privatepart => "__privatepart",
               when Precomputed_Sym_Procedure => "procedure",
               when Precomputed_Sym_Protected => "protected",
               when Precomputed_Sym_Pure => "Pure",
               when Precomputed_Sym_Put_Image => "Put_Image",
               when Precomputed_Sym_Put_Image_307 => "put_image",
               when Precomputed_Sym_Raise => "raise",
               when Precomputed_Sym_Range => "Range",
               when Precomputed_Sym_Range_309 => "range",
               when Precomputed_Sym_Range_Length => "Range_Length",
               when Precomputed_Sym_Read => "Read",
               when Precomputed_Sym_Real_Literal => "Real_Literal",
               when Precomputed_Sym_Record => "record",
               when Precomputed_Sym_Reduce => "Reduce",
               when Precomputed_Sym_Refined_Depends => "Refined_Depends",
               when Precomputed_Sym_Refined_Global => "Refined_Global",
               when Precomputed_Sym_Refined_Post => "Refined_Post",
               when Precomputed_Sym_Refined_State => "Refined_State",
               when Precomputed_Sym_Rem => """rem""",
               when Precomputed_Sym_Rem_311 => "rem",
               when Precomputed_Sym_Remainder => "Remainder",
               when Precomputed_Sym_Remote_Call_Interface => "Remote_Call_Interface",
               when Precomputed_Sym_Remote_Types => "Remote_Types",
               when Precomputed_Sym_Renames => "renames",
               when Precomputed_Sym_Requeue => "requeue",
               when Precomputed_Sym_Requires => "Requires",
               when Precomputed_Sym_Restrictions => "Restrictions",
               when Precomputed_Sym_Result => "Result",
               when Precomputed_Sym_Return => "return",
               when Precomputed_Sym_Reverse => "reverse",
               when Precomputed_Sym_Right => "right",
               when Precomputed_Sym_Root_Buffer_Type => "Root_Buffer_Type",
               when Precomputed_Sym_Root_Integer => "root_integer",
               when Precomputed_Sym_Root_Real => "root_real",
               when Precomputed_Sym_Root_Storage_Pool => "Root_Storage_Pool",
               when Precomputed_Sym_Root_Stream_Type => "Root_Stream_Type",
               when Precomputed_Sym_Root_Types => "root_types_",
               when Precomputed_Sym_Round => "Round",
               when Precomputed_Sym_Rounding => "Rounding",
               when Precomputed_Sym_Safe_Large => "Safe_Large",
               when Precomputed_Sym_Safe_Small => "Safe_Small",
               when Precomputed_Sym_Scalar_Storage_Order => "Scalar_Storage_Order",
               when Precomputed_Sym_Scaling => "Scaling",
               when Precomputed_Sym_Select => "select",
               when Precomputed_Sym_Separate => "separate",
               when Precomputed_Sym_Shared_Passive => "Shared_Passive",
               when Precomputed_Sym_Size => "Size",
               when Precomputed_Sym_Small => "Small",
               when Precomputed_Sym_Small_Denominator => "Small_Denominator",
               when Precomputed_Sym_Small_Numerator => "Small_Numerator",
               when Precomputed_Sym_Some => "some",
               when Precomputed_Sym_Stable_Properties => "Stable_Properties",
               when Precomputed_Sym_Standard => "Standard",
               when Precomputed_Sym_Standard_323 => "standard",
               when Precomputed_Sym_Static_Predicate => "Static_Predicate",
               when Precomputed_Sym_Storage_Elements => "Storage_Elements",
               when Precomputed_Sym_Storage_Pool => "Storage_Pool",
               when Precomputed_Sym_Storage_Pools => "Storage_Pools",
               when Precomputed_Sym_Storage_Size => "Storage_Size",
               when Precomputed_Sym_Storage_Unit => "Storage_Unit",
               when Precomputed_Sym_Streams => "Streams",
               when Precomputed_Sym_String => "String",
               when Precomputed_Sym_String_Literal => "String_Literal",
               when Precomputed_Sym_Strings => "Strings",
               when Precomputed_Sym_Style_Checks => "Style_Checks",
               when Precomputed_Sym_Subtype => "subtype",
               when Precomputed_Sym_Succ => "Succ",
               when Precomputed_Sym_Synchronized => "synchronized",
               when Precomputed_Sym_System => "System",
               when Precomputed_Sym_System_Allocator_Alignment => "System_Allocator_Alignment",
               when Precomputed_Sym_Tag => "Tag",
               when Precomputed_Sym_Tagged => "tagged",
               when Precomputed_Sym_Tags => "Tags",
               when Precomputed_Sym_Target_Name => "Target_Name",
               when Precomputed_Sym_Task => "task",
               when Precomputed_Sym_Task_Id => "Task_Id",
               when Precomputed_Sym_Task_Identification => "Task_Identification",
               when Precomputed_Sym_Terminate => "terminate",
               when Precomputed_Sym_Terminated => "Terminated",
               when Precomputed_Sym_Test_Case => "Test_Case",
               when Precomputed_Sym_Text_Buffers => "Text_Buffers",
               when Precomputed_Sym_Text_Io => "text_io",
               when Precomputed_Sym_Then => "then",
               when Precomputed_Sym_To_Address => "To_Address",
               when Precomputed_Sym_Truncation => "Truncation",
               when Precomputed_Sym_Type => "type",
               when Precomputed_Sym_Type_Class => "Type_Class",
               when Precomputed_Sym_Type_Invariant => "Type_Invariant",
               when Precomputed_Sym_Type_Invariant_Class => "Type_Invariant'Class",
               when Precomputed_Sym_Unchecked_Access => "Unchecked_Access",
               when Precomputed_Sym_Unchecked_Union => "Unchecked_Union",
               when Precomputed_Sym_Universal_Int_Type => "Universal_Int_Type_",
               when Precomputed_Sym_Universal_Real_Type => "Universal_Real_Type_",
               when Precomputed_Sym_Unreferenced => "Unreferenced",
               when Precomputed_Sym_Unrestricted_Access => "Unrestricted_Access",
               when Precomputed_Sym_Until => "until",
               when Precomputed_Sym_Update => "Update",
               when Precomputed_Sym_Use => "use",
               when Precomputed_Sym_Vads_Size => "VADS_Size",
               when Precomputed_Sym_Val => "Val",
               when Precomputed_Sym_Valid => "Valid",
               when Precomputed_Sym_Value => "Value",
               when Precomputed_Sym_Value_Size => "Value_Size",
               when Precomputed_Sym_Variable_Indexing => "Variable_Indexing",
               when Precomputed_Sym_Volatile => "Volatile",
               when Precomputed_Sym_Volatile_Components => "Volatile_Components",
               when Precomputed_Sym_Warnings => "Warnings",
               when Precomputed_Sym_When => "when",
               when Precomputed_Sym_While => "while",
               when Precomputed_Sym_Wide_Character => "Wide_Character",
               when Precomputed_Sym_Wide_Image => "Wide_Image",
               when Precomputed_Sym_Wide_String => "Wide_String",
               when Precomputed_Sym_Wide_Text_Io => "wide_text_io",
               when Precomputed_Sym_Wide_Value => "Wide_Value",
               when Precomputed_Sym_Wide_Wide_Character => "Wide_Wide_Character",
               when Precomputed_Sym_Wide_Wide_Image => "Wide_Wide_Image",
               when Precomputed_Sym_Wide_Wide_String => "Wide_Wide_String",
               when Precomputed_Sym_Wide_Wide_Text_Io => "wide_wide_text_io",
               when Precomputed_Sym_Wide_Wide_Value => "Wide_Wide_Value",
               when Precomputed_Sym_Width => "Width",
               when Precomputed_Sym_With => "with",
               when Precomputed_Sym_Word_Size => "Word_Size",
               when Precomputed_Sym_Write => "Write",
               when Precomputed_Sym_Xor => """xor""",
               when Precomputed_Sym_Xor_339 => "xor"
            );
            Symbol : constant Symbolization_Result :=
                  Libadalang.Sources.Canonicalize (Raw_Text)
            ;
         begin
            if Symbol.Success then
               return Symbol.Symbol;
            else
               raise Program_Error with
                 "Cannot canonicalize symbol literal: " & Image (Raw_Text);
            end if;
         end;
   end Precomputed_Symbol;
   ----------------------------
   -- Construct_Entity_Array --
   ----------------------------
   function Construct_Entity_Array
     (V : AST_Envs.Entity_Vectors.Vector) return Internal_Entity_Array_Access
   is
      Ret : Internal_Entity_Array_Access :=
        Create_Internal_Entity_Array (V.Length);
   begin
      for J in V.First_Index .. V.Last_Index loop
         Ret.Items (J) := V.Get (J);
      end loop;
      declare
         Tmp : AST_Envs.Entity_Vectors.Vector := V;
      begin
         Tmp.Destroy;
      end;
      return Ret;
   end Construct_Entity_Array;
   ----------------
   -- Enter_Call --
   ----------------
   procedure Enter_Call
     (Context : Internal_Context; Call_Depth : access Natural)
   is
      Max             : Natural renames Context.Max_Call_Depth;
      Current         : Natural renames Context.Current_Call_Depth;
      High_Water_Mark : Natural renames Context.Call_Depth_High_Water_Mark;
   begin
      Current := Current + 1;
      High_Water_Mark := Natural'Max (High_Water_Mark, Current);
      Call_Depth.all := Current;
      if Current > Max then
         raise Property_Error with "stack overflow";
      end if;
   end Enter_Call;
   ---------------
   -- Exit_Call --
   ---------------
   procedure Exit_Call (Context : Internal_Context; Call_Depth : Natural) is
      Current : Natural renames Context.Current_Call_Depth;
   begin
      if Call_Depth /= Current then
         raise Unexpected_Call_Depth with
            "Langkit code generation bug for call depth handling detected";
      end if;
      Current := Current - 1;
   end Exit_Call;
   -----------
   -- Image --
   -----------
   function Image (Self : Symbol_Type) return String_Type is
   begin
      return Create_String (Image (Self));
   end Image;
   ------------------
   -- Context_Pool --
   ------------------
   package body Context_Pool is
      -------------
      -- Acquire --
      -------------
      procedure Acquire (Context : out Internal_Context) is
      begin
         GNAT.Task_Lock.Lock;
         if Available.Is_Empty then
            Context := new Analysis_Context_Type;
            Context.Serial_Number := 1;
         else
            Context := Available.Last_Element;
            Available.Delete_Last;
         end if;
         GNAT.Task_Lock.Unlock;
      exception
         when others =>
            GNAT.Task_Lock.Unlock;
            raise;
      end Acquire;
      -------------
      -- Release --
      -------------
      procedure Release (Context : in out Internal_Context) is
      begin
         GNAT.Task_Lock.Lock;
         Available.Append (Context);
         Context.Serial_Number := Context.Serial_Number + 1;
         Context := null;
         GNAT.Task_Lock.Unlock;
      exception
         when others =>
            GNAT.Task_Lock.Unlock;
            raise;
      end Release;
      ----------
      -- Free --
      ----------
      procedure Free is
      begin
         GNAT.Task_Lock.Lock;
         for C of Available loop
            Free (C);
         end loop;
         GNAT.Task_Lock.Unlock;
      exception
         when others =>
            GNAT.Task_Lock.Unlock;
            raise;
      end Free;
   end Context_Pool;
   --------------
   -- Finalize --
   --------------
   overriding procedure Finalize (CD : in out Contexts_Destructor) is
      pragma Unreferenced (CD);
   begin
      Context_Pool.Free;
   end Finalize;
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (File_Reader : in out Internal_File_Reader_Access) is
      procedure Destroy is new Ada.Unchecked_Deallocation
        (Internal_File_Reader'Class, Internal_File_Reader_Access);
   begin
      if File_Reader /= null and then File_Reader.all.Dec_Ref then
         Destroy (File_Reader);
      end if;
   end Dec_Ref;
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (Self : in out Internal_Event_Handler_Access) is
      procedure Destroy is new Ada.Unchecked_Deallocation
        (Internal_Event_Handler'Class, Internal_Event_Handler_Access);
   begin
      if Self /= null and then Self.all.Dec_Ref then
         Destroy (Self);
      end if;
   end Dec_Ref;
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (Provider : in out Internal_Unit_Provider_Access) is
      procedure Destroy is new Ada.Unchecked_Deallocation
        (Internal_Unit_Provider'Class, Internal_Unit_Provider_Access);
   begin
      if Provider /= null and then Provider.all.Dec_Ref then
         Destroy (Provider);
      end if;
   end Dec_Ref;
   ----------------
   -- Get_Env_Id --
   ----------------
   function Get_Env_Id
     (E : Lexical_Env; State : in out Dump_Lexical_Env_State) return String
   is
      C        : Address_To_Id_Maps.Cursor;
      Inserted : Boolean;
   begin
      if E = Null_Lexical_Env then
         return "$null";
      elsif E = State.Root_Env then
         --  Insert root env with a special Id so that we only print it once
         State.Env_Ids.Insert (E, -1, C, Inserted);
         return "$root";
      end if;
      State.Env_Ids.Insert (E, State.Next_Id, C, Inserted);
      if Inserted then
         State.Next_Id := State.Next_Id + 1;
      end if;
      return '@' & Stripped_Image (Address_To_Id_Maps.Element (C));
   end Get_Env_Id;
   pragma Warnings (Off, "referenced");
   function To_Lookup_Kind_Type (K : Lookup_Kind) return Lookup_Kind_Type
   is
     (Lookup_Kind_Type'Val (Lookup_Kind'Pos (K)));
   pragma Warnings (On, "referenced");
   --------------------
   -- Create_Context --
   --------------------
   function Create_Context
     (Charset        : String;
      File_Reader    : Internal_File_Reader_Access;
      Unit_Provider  : Internal_Unit_Provider_Access;
      Event_Handler  : Internal_Event_Handler_Access;
      With_Trivia    : Boolean;
      Tab_Stop       : Positive;
      Max_Call_Depth : Natural := 1000)
      return Internal_Context
   is
      Actual_Charset : constant String :=
        (if Charset = "" then Default_Charset else Charset);
      Symbols        : constant Precomputed_Symbol_Table
        := Create_Symbol_Table;
      Context        : Internal_Context;
   begin
      Context_Pool.Acquire (Context);
      Context.Ref_Count := 1;
      Context.Symbols := Symbol_Table (Symbols);
      Context.Charset := To_Unbounded_String (Actual_Charset);
      Context.Tab_Stop := Tab_Stop;
      Context.With_Trivia := With_Trivia;
      Context.Root_Scope := Create_Static_Lexical_Env
        (Parent => Null_Lexical_Env,
         Node   => null);
      --  Create a new ownership share for Event_Handler so that it lives at
      --  least as long as this analysis context.
      Context.Event_Handler := Event_Handler;
      if Context.Event_Handler /= null then
         Context.Event_Handler.Inc_Ref;
      end if;
      --  Create a new ownership share for File_Reader so that it lives at
      --  least as long as this analysis context.
      Context.File_Reader := File_Reader;
      if Context.File_Reader /= null then
         Context.File_Reader.Inc_Ref;
      end if;
      --  Create a new ownership share for Unit_Provider so that it lives at
      --  least as long as this analysis context.
      Context.Unit_Provider := Unit_Provider;
      if Context.Unit_Provider /= null then
         Context.Unit_Provider.Inc_Ref;
      end if;
         if Context.Unit_Provider = null then
            Context.Unit_Provider := Libadalang.Internal_Default_Provider.Create;
         end if;
      Initialize (Context.Parser);
      Context.Discard_Errors_In_Populate_Lexical_Env := True;
      Context.Logic_Resolution_Timeout :=
        Langkit_Support.Adalog.Default_Timeout_Ticks_Number;
      Context.In_Populate_Lexical_Env := False;
      Context.Cache_Version := 0;
      Context.Reparse_Cache_Version := 0;
      Context.Rewriting_Handle := No_Rewriting_Handle_Pointer;
      Context.Templates_Unit := No_Analysis_Unit;
      Context.Max_Call_Depth := Max_Call_Depth;
      Context.Available_Rebindings := Env_Rebindings_Vectors.Empty_Vector;
Fetch_Standard (Context);
      return Context;
   end Create_Context;
   -----------------
   -- Create_Unit --
   -----------------
   function Create_Unit
     (Context             : Internal_Context;
      Normalized_Filename : Virtual_File;
      Charset             : String;
      Rule                : Grammar_Rule) return Internal_Unit
   is
      use Units_Maps;
      Unit : Internal_Unit;
   begin
      Unit := Create_Special_Unit
        (Context, Normalized_Filename, Charset, Rule);
      Context.Units.Insert (Normalized_Filename, Unit);
      return Unit;
   end Create_Unit;
   --------------
   -- Get_Unit --
   --------------
   function Get_Unit
     (Context           : Internal_Context;
      Filename, Charset : String;
      Reparse           : Boolean;
      Input             : Internal_Lexer_Input;
      Rule              : Grammar_Rule;
      Is_Internal       : Boolean := False) return Internal_Unit
   is
      use Units_Maps;
      Normalized_Filename : constant GNATCOLL.VFS.Virtual_File :=
         Normalized_Unit_Filename (Context, Filename);
      Cur     : constant Cursor :=
         Context.Units.Find (Normalized_Filename);
      Created : constant Boolean := Cur = No_Element;
      Unit    : Internal_Unit;
      Actual_Charset : Unbounded_String;
      Refined_Input  : Internal_Lexer_Input := Input;
   begin
      --  Determine which encoding to use. Use the Charset parameter (if
      --  provided), otherwise use the context-wide default.
      Actual_Charset := (if Charset'Length /= 0
                         then To_Unbounded_String (Charset)
                         else Context.Charset);
      if Refined_Input.Kind = File then
         Refined_Input.Filename := Normalized_Filename;
      end if;
      if Refined_Input.Kind in File | Bytes_Buffer then
         Refined_Input.Charset := Actual_Charset;
         --  Unless the caller requested a specific charset for this unit,
         --  allow the lexer to automatically discover the source file encoding
         --  before defaulting to the context-specific one. We do this trying
         --  to match a byte order mark.
         Refined_Input.Read_BOM := Charset'Length = 0;
      end if;
      --  Create the Internal_Unit if needed
      Unit :=
        (if Created
         then Create_Unit (Context, Normalized_Filename,
                           To_String (Actual_Charset), Rule)
         else Element (Cur));
      --  If an internal unit is requested, set the corresponding flag.
      --  Otherwise, make sure that the unit we return isn't internal.
      if Is_Internal then
         Unit.Is_Internal := True;
      end if;
      --  (Re)parse it if needed
      if Created or else Reparse then
         --  It is illegal to reparse an internal unit for public API users.
         --  Since public APIs do not allow to pass True to Is_Internal, we can
         --  check here that only the implementation can ask to reparse an
         --  internal unit.
         if Unit.Is_Internal and then not Is_Internal then
            raise Precondition_Failure with "cannot reparse an internal unit";
         end if;
         declare
            Reparsed : Reparsed_Unit;
         begin
            Do_Parsing (Unit, Refined_Input, Reparsed);
            Update_After_Reparse (Unit, Reparsed);
         end;
         --  Now that we have removed reparsed the unit, update its current
         --  charset.
         Unit.Charset := Actual_Charset;
      end if;
      if Context.Event_Handler /= null then
         Context.Event_Handler.Unit_Parsed_Callback (Context, Unit, Reparse);
      end if;
      return Unit;
   end Get_Unit;
   --------------
   -- Has_Unit --
   --------------
   function Has_Unit
     (Context : Internal_Context; Unit_Filename : String) return Boolean is
   begin
      return Context.Units.Contains
        (Normalized_Unit_Filename (Context, Unit_Filename));
   end Has_Unit;
   -------------------
   -- Get_From_File --
   -------------------
   function Get_From_File
     (Context  : Internal_Context;
      Filename : String;
      Charset  : String;
      Reparse  : Boolean;
      Rule     : Grammar_Rule) return Internal_Unit
   is
      Input : constant Internal_Lexer_Input :=
        (Kind     => File,
         Charset  => <>,
         Read_BOM => False,
         Filename => <>);
   begin
      if Reparse and then Has_Rewriting_Handle (Context) then
         raise Precondition_Failure with
            "cannot reparse during tree rewriting";
      end if;
      return Get_Unit (Context, Filename, Charset, Reparse, Input, Rule);
   end Get_From_File;
   ---------------------
   -- Get_From_Buffer --
   ---------------------
   function Get_From_Buffer
     (Context  : Internal_Context;
      Filename : String;
      Charset  : String;
      Buffer   : String;
      Rule     : Grammar_Rule) return Internal_Unit
   is
      Input : constant Internal_Lexer_Input :=
        (Kind        => Bytes_Buffer,
         Charset     => <>,
         Read_BOM    => False,
         Bytes       => Buffer'Address,
         Bytes_Count => Buffer'Length);
   begin
      if Has_Rewriting_Handle (Context) then
         raise Precondition_Failure with
            "cannot parse from buffer during tree rewriting";
      elsif Context.File_Reader /= null then
         raise Precondition_Failure with
            "cannot parse from buffer with a file reader";
      end if;
      return Get_Unit (Context, Filename, Charset, True, Input, Rule);
   end Get_From_Buffer;
   --------------------
   -- Get_With_Error --
   --------------------
   function Get_With_Error
     (Context  : Internal_Context;
      Filename : String;
      Error    : Text_Type;
      Charset  : String;
      Rule     : Grammar_Rule) return Internal_Unit
   is
      use Units_Maps;
      Normalized_Filename : constant Virtual_File :=
         Normalized_Unit_Filename (Context, Filename);
      Cur                 : constant Cursor :=
         Context.Units.Find (Normalized_Filename);
   begin
      if Cur = No_Element then
         declare
            Unit : constant Internal_Unit := Create_Unit
              (Context, Normalized_Filename, Charset, Rule);
         begin
            Append (Unit.Diagnostics, No_Source_Location_Range, Error);
            return Unit;
         end;
      else
         return Element (Cur);
      end if;
   end Get_With_Error;
   -----------------------
   -- Get_From_Provider --
   -----------------------
   function Get_From_Provider
     (Context : Internal_Context;
      Name    : Text_Type;
      Kind    : Analysis_Unit_Kind;
      Charset : String;
      Reparse : Boolean) return Internal_Unit is
   begin
      if Reparse and then Has_Rewriting_Handle (Context) then
         raise Precondition_Failure with
            "cannot reparse during tree rewriting";
      end if;
      begin
         return Context.Unit_Provider.Get_Unit
           (Context, Name, Kind, Charset, Reparse);
      exception
         when Precondition_Failure | Property_Error =>
            raise Invalid_Unit_Name_Error with
               "Invalid unit name: " & Image (Name, With_Quotes => True)
               & " (" & Analysis_Unit_Kind'Image (Kind) & ")";
      end;
   end Get_From_Provider;
   -------------------
   -- Unit_Provider --
   -------------------
   function Unit_Provider
     (Context : Internal_Context) return Internal_Unit_Provider_Access
   is (Context.Unit_Provider);
   ----------
   -- Hash --
   ----------
   function Hash (Context : Internal_Context) return Hash_Type is
      function H is new Hash_Access (Analysis_Context_Type, Internal_Context);
   begin
      return H (Context);
   end Hash;
   ---------------------
   -- Has_With_Trivia --
   ---------------------
   function Has_With_Trivia (Context : Internal_Context) return Boolean is
   begin
      return Context.With_Trivia;
   end Has_With_Trivia;
   --------------------------------------------
   -- Discard_Errors_In_Populate_Lexical_Env --
   --------------------------------------------
   procedure Discard_Errors_In_Populate_Lexical_Env
     (Context : Internal_Context; Discard : Boolean) is
   begin
      Context.Discard_Errors_In_Populate_Lexical_Env := Discard;
   end Discard_Errors_In_Populate_Lexical_Env;
   ----------------------------------
   -- Set_Logic_Resolution_Timeout --
   ----------------------------------
   procedure Set_Logic_Resolution_Timeout
     (Context : Internal_Context; Timeout : Natural) is
   begin
      Context.Logic_Resolution_Timeout := Timeout;
   end Set_Logic_Resolution_Timeout;
   --------------------------
   -- Has_Rewriting_Handle --
   --------------------------
   function Has_Rewriting_Handle (Context : Internal_Context) return Boolean is
   begin
      return Context.Rewriting_Handle /= No_Rewriting_Handle_Pointer;
   end Has_Rewriting_Handle;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (Context : Internal_Context) is
   begin
      if Context /= null then
         Context.Ref_Count := Context.Ref_Count + 1;
      end if;
   end Inc_Ref;
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (Context : in out Internal_Context) is
   begin
      if Context /= null then
         Context.Ref_Count := Context.Ref_Count - 1;
         if Context.Ref_Count = 0 then
            Destroy (Context);
         end if;
      end if;
   end Dec_Ref;
   -------------
   -- Destroy --
   -------------
   procedure Destroy (Context : in out Internal_Context) is
   begin
      --  Destroy all named environment data structures
      for Desc of Context.Named_Envs loop
         for V of Desc.Foreign_Nodes loop
            V.Destroy;
         end loop;
         Destroy (Desc);
      end loop;
      Context.Named_Envs.Clear;
      --  If we are asked to free this context, it means that no one else have
      --  references to its analysis units, so it's safe to destroy these.
      for Unit of Context.Units loop
         Destroy (Unit);
      end loop;
      Context.Units := Units_Maps.Empty_Map;
      Context.Filenames := Virtual_File_Maps.Empty_Map;
      declare
         procedure Destroy is new Ada.Unchecked_Deallocation
           (Env_Rebindings_Type, Env_Rebindings);
         AR : Env_Rebindings_Vectors.Vector renames
            Context.Available_Rebindings;
         R  : Env_Rebindings;
      begin
         for I in AR.First_Index .. AR.Last_Index loop
            R := AR.Get (I);
            Destroy (R);
         end loop;
         AR.Destroy;
      end;
      Destroy (Context.Templates_Unit);
      AST_Envs.Destroy (Context.Root_Scope);
      Destroy (Context.Symbols);
      Destroy (Context.Parser);
      Dec_Ref (Context.File_Reader);
      Dec_Ref (Context.Unit_Provider);
      Dec_Ref (Context.Event_Handler);
      Context_Pool.Release (Context);
   end Destroy;
   -------------
   -- Context --
   -------------
   function Context (Unit : Internal_Unit) return Internal_Context is
   begin
      return Unit.Context;
   end Context;
   ----------
   -- Hash --
   ----------
   function Hash (Unit : Internal_Unit) return Hash_Type is
      function H is new Hash_Access (Analysis_Unit_Type, Internal_Unit);
   begin
      return H (Unit);
   end Hash;
   -------------
   -- Reparse --
   -------------
   procedure Reparse (Unit : Internal_Unit; Charset : String) is
      Dummy : constant Internal_Unit := Get_From_File
        (Unit.Context, +Unit.Filename.Full_Name, Charset,
         Reparse => True,
         Rule    => Unit.Rule);
   begin
      null;
   end Reparse;
   -------------
   -- Reparse --
   -------------
   procedure Reparse (Unit : Internal_Unit; Charset : String; Buffer : String)
   is
      Dummy : constant Internal_Unit := Get_From_Buffer
        (Unit.Context, +Unit.Filename.Full_Name, Charset, Buffer, Unit.Rule);
   begin
      null;
   end Reparse;
   --------------------------
   -- Populate_Lexical_Env --
   --------------------------
   procedure Populate_Lexical_Env (Unit : Internal_Unit) is
      Context : constant Internal_Context := Unit.Context;
      Has_Errors : Boolean := False;
      --  Whether at least one Property_Error occurred during this PLE pass
      Saved_In_Populate_Lexical_Env : constant Boolean :=
         Unit.Context.In_Populate_Lexical_Env;
      procedure Reset_Envs_Caches (Unit : Internal_Unit) is
         procedure Internal (Node : Bare_Ada_Node) is
         begin
            if Node = null then
               return;
            end if;
            Reset_Caches (Node.Self_Env);
            for I in 1 .. Children_Count (Node) loop
               Internal (Child (Node, I));
            end loop;
         end Internal;
      begin
         Internal (Unit.Ast_Root);
      end Reset_Envs_Caches;
   begin
      --  TODO??? Handle env invalidation when reparsing a unit and when a
      --  previous call raised a Property_Error.
      if Unit.Is_Env_Populated then
         return;
      end if;
      Unit.Is_Env_Populated := True;
      if Unit.Ast_Root = null then
         return;
      end if;
      GNATCOLL.Traces.Trace (Main_Trace, "Populating lexical envs for unit: "
                                         & Basename (Unit));
      if GNATCOLL.Traces.Active (Main_Trace) then
         GNATCOLL.Traces.Increase_Indent (Main_Trace);
      end if;
      Context.In_Populate_Lexical_Env := True;
      Has_Errors := Populate_Lexical_Env (Unit.Ast_Root);
      Context.In_Populate_Lexical_Env := Saved_In_Populate_Lexical_Env;
      if GNATCOLL.Traces.Active (Main_Trace) then
         GNATCOLL.Traces.Decrease_Indent (Main_Trace);
         GNATCOLL.Traces.Trace
           (Main_Trace,
            "Finished populating lexical envs for unit: " & Basename (Unit));
      end if;
      Reset_Envs_Caches (Unit);
      if Has_Errors and then not Context.Discard_Errors_In_Populate_Lexical_Env
      then
         raise Property_Error with
            "errors occurred in Populate_Lexical_Env";
      end if;
   end Populate_Lexical_Env;
   ------------------
   -- Get_Filename --
   ------------------
   function Get_Filename (Unit : Internal_Unit) return String is
     (+Unit.Filename.Full_Name);
   -----------------
   -- Get_Charset --
   -----------------
   function Get_Charset (Unit : Internal_Unit) return String is
   begin
      return To_String (Unit.Charset);
   end Get_Charset;
   ---------------------
   -- Has_Diagnostics --
   ---------------------
   function Has_Diagnostics (Unit : Internal_Unit) return Boolean is
   begin
      return not Unit.Diagnostics.Is_Empty;
   end Has_Diagnostics;
   -----------------
   -- Diagnostics --
   -----------------
   function Diagnostics (Unit : Internal_Unit) return Diagnostics_Array is
      Result : Diagnostics_Array (1 .. Natural (Unit.Diagnostics.Length));
      I      : Natural := 1;
   begin
      for D of Unit.Diagnostics loop
         Result (I) := D;
         I := I + 1;
      end loop;
      return Result;
   end Diagnostics;
   ---------------------------
   -- Format_GNU_Diagnostic --
   ---------------------------
   function Format_GNU_Diagnostic
     (Unit : Internal_Unit; D : Diagnostic) return String
   is
      Filename : constant String := Basename (Unit);
      Sloc     : constant Source_Location := Start_Sloc (D.Sloc_Range);
      Msg      : constant String :=
         Image
           (Ada.Strings.Wide_Wide_Unbounded.To_Wide_Wide_String (D.Message));
   begin
      return (Filename
              & (if Sloc = No_Source_Location then "" else ":" & Image (Sloc))
              & ": " & Msg);
   end Format_GNU_Diagnostic;
   ----------
   -- Root --
   ----------
   function Root (Unit : Internal_Unit) return Bare_Ada_Node is
     (Unit.Ast_Root);
   -----------------
   -- First_Token --
   -----------------
   function First_Token (Unit : Internal_Unit) return Token_Reference is
     (Wrap_Token_Reference (Unit.Context,
                            Unit.TDH'Access,
                            First_Token_Or_Trivia (Unit.TDH)));
   ----------------
   -- Last_Token --
   ----------------
   function Last_Token (Unit : Internal_Unit) return Token_Reference is
     (Wrap_Token_Reference (Unit.Context,
                            Unit.TDH'Access,
                            Last_Token_Or_Trivia (Unit.TDH)));
   -----------------
   -- Token_Count --
   -----------------
   function Token_Count (Unit : Internal_Unit) return Natural is
     (Unit.TDH.Tokens.Length);
   ------------------
   -- Trivia_Count --
   ------------------
   function Trivia_Count (Unit : Internal_Unit) return Natural is
     (Unit.TDH.Trivias.Length);
   ----------
   -- Text --
   ----------
   function Text (Unit : Internal_Unit) return Text_Type is
   begin
      return Text (First_Token (Unit), Last_Token (Unit));
   end Text;
   ------------------
   -- Lookup_Token --
   ------------------
   function Lookup_Token
     (Unit : Internal_Unit; Sloc : Source_Location) return Token_Reference
   is
      Result : constant Token_Or_Trivia_Index := Lookup_Token (Unit.TDH, Sloc);
   begin
      return Wrap_Token_Reference (Unit.Context, Unit.TDH'Access, Result);
   end Lookup_Token;
   ----------------------
   -- Dump_Lexical_Env --
   ----------------------
   procedure Dump_Lexical_Env (Unit : Internal_Unit) is
      Node     : constant Bare_Ada_Node := Unit.Ast_Root;
      Root_Env : constant Lexical_Env := Unit.Context.Root_Scope;
      State    : Dump_Lexical_Env_State := (Root_Env => Root_Env, others => <>);
      function Get_Parent (Env : Lexical_Env) return Lexical_Env
      is (Unwrap (Env).Parent);
      --------------------------
      -- Explore_Parent_Chain --
      --------------------------
      procedure Explore_Parent_Chain (Env : Lexical_Env) is
         P : Lexical_Env;
      begin
         if Env /= Null_Lexical_Env then
            P := Get_Parent (Env);
            Dump_One_Lexical_Env
              (Env, Get_Env_Id (Env, State), Get_Env_Id (P, State));
            Explore_Parent_Chain (P);
         end if;
      end Explore_Parent_Chain;
      --------------
      -- Internal --
      --------------
      procedure Internal (Current : Bare_Ada_Node) is
         Explore_Parent : Boolean := False;
         Env, Parent    : Lexical_Env;
      begin
         if Current = null then
            return;
         end if;
         --  We only dump environments that we haven't dumped before. This way
         --  we'll only dump environments at the site of their creation, and
         --  not in any subsequent link. We use the Env_Ids map to check which
         --  envs we have already seen or not.
         if not State.Env_Ids.Contains (Current.Self_Env) then
            Env := Current.Self_Env;
            Parent := Get_Parent (Env);
            Explore_Parent := not State.Env_Ids.Contains (Parent);
            Dump_One_Lexical_Env
              (Env, Get_Env_Id (Env, State), Get_Env_Id (Parent, State));
            if Explore_Parent then
               Explore_Parent_Chain (Parent);
            end if;
         end if;
         for Child of Internal_Bare_Ada_Node_Array'(Children (Current))
         loop
            Internal (Child);
         end loop;
      end Internal;
      --  This procedure implements the main recursive logic of dumping the
      --  environments.
   begin
      Internal (Bare_Ada_Node (Node));
   end Dump_Lexical_Env;
   --------------
   -- Get_Line --
   --------------
   function Get_Line
     (Unit : Internal_Unit; Line_Number : Positive) return Text_Type
   is
   begin
      return Get_Line (Unit.TDH, Line_Number);
   end Get_Line;
   -----------
   -- Print --
   -----------
   procedure Print (Unit : Internal_Unit; Show_Slocs : Boolean) is
   begin
      if Unit.Ast_Root = null then
         Put_Line ("<empty analysis unit>");
      else
         Print (Unit.Ast_Root, Show_Slocs);
      end if;
   end Print;
   ---------------
   -- PP_Trivia --
   ---------------
   procedure PP_Trivia (Unit : Internal_Unit) is
      procedure Process (Trivia : Token_Index) is
         Data : constant Stored_Token_Data :=
            Unit.TDH.Trivias.Get (Natural (Trivia)).T;
      begin
         Put_Line (Image (Text (Unit.TDH, Data)));
      end Process;
      Last_Token : constant Token_Index :=
         Token_Index (Token_Vectors.Last_Index (Unit.TDH.Tokens) - 1);
      --  Index for the last token in Unit excluding the Termination token
      --  (hence the -1).
   begin
      for Tok of Get_Leading_Trivias (Unit.TDH) loop
         Process (Tok);
      end loop;
      PP_Trivia (Unit.Ast_Root);
      for Tok of Get_Trivias (Unit.TDH, Last_Token) loop
         Process (Tok);
      end loop;
   end PP_Trivia;
   -------------
   -- Destroy --
   -------------
   procedure Destroy (Unit : in out Internal_Unit) is
   begin
      if Unit = No_Analysis_Unit then
         return;
      end if;
      Unit.Exiled_Entries.Destroy;
      Unit.Foreign_Nodes.Destroy;
      Unit.Exiled_Entries_In_NED.Destroy;
      Unit.Exiled_Envs.Destroy;
      Unit.Named_Envs.Destroy;
      Analysis_Unit_Sets.Destroy (Unit.Referenced_Units);
         Destroy (Unit.Memoization_Map);
      Destroy_Rebindings (Unit.Rebindings'Access);
      Unit.Rebindings.Destroy;
      if Unit.Ast_Root /= null then
         Destroy (Unit.Ast_Root);
      end if;
      Free (Unit.TDH);
      Free (Unit.Ast_Mem_Pool);
      Destroy_Unit_Destroyables (Unit);
      Destroyable_Vectors.Destroy (Unit.Destroyables);
      Free (Unit);
   end Destroy;
   -------------------
   -- Is_Token_Node --
   -------------------
   function Is_Token_Node (Node : Bare_Ada_Node) return Boolean is
   begin
      return Is_Token_Node (Node.Kind);
   end Is_Token_Node;
   ------------------
   -- Is_Synthetic --
   ------------------
   function Is_Synthetic (Node : Bare_Ada_Node) return Boolean is
   begin
      return Node.Kind in Synthetic_Nodes;
   end Is_Synthetic;
   ------------------------------
   -- Register_Destroyable_Gen --
   ------------------------------
   procedure Register_Destroyable_Gen
     (Unit : Internal_Unit; Object : T_Access)
   is
      function Convert is new Ada.Unchecked_Conversion
        (System.Address, Destroy_Procedure);
      procedure Destroy_Procedure (Object : in out T_Access) renames Destroy;
   begin
      Register_Destroyable_Helper
        (Unit,
         Object.all'Address,
         Convert (Destroy_Procedure'Address));
   end Register_Destroyable_Gen;
   ---------
   -- Get --
   ---------
   function Get
     (T       : Bare_Ada_Node_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Bare_Ada_Node
   is
      function Absolute_Get
        (T : Bare_Ada_Node_Array_Access; Index : Integer)
         return Bare_Ada_Node
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Bare_Ada_Node,
         Sequence_Type => Bare_Ada_Node_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Bare_Ada_Node;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Bare_Ada_Node;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Bare_Ada_Node_Array_Access) return Bare_Ada_Node_Array_Access is
      Ret : Bare_Ada_Node_Array_Access := Create_Bare_Ada_Node_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Bare_Ada_Node_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Bare_Ada_Node_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Bare_Ada_Node_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Bare_Ada_Node_Array (Items_Count : Natural) return Bare_Ada_Node_Array_Access
   is (if Items_Count = 0
       then No_Bare_Ada_Node_Array_Type
       else new Bare_Ada_Node_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Bare_Ada_Node_Array
     (Items : Internal_Bare_Ada_Node_Array) return Bare_Ada_Node_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Bare_Ada_Node_Array_Type;
      end if;
      return new Bare_Ada_Node_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Bare_Ada_Node_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Bare_Ada_Node_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Bare_Anonymous_Expr_Decl_Vectors is new Langkit_Support.Vectors (Bare_Anonymous_Expr_Decl);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Bare_Anonymous_Expr_Decl_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Bare_Anonymous_Expr_Decl
   is
      function Absolute_Get
        (T : Bare_Anonymous_Expr_Decl_Array_Access; Index : Integer)
         return Bare_Anonymous_Expr_Decl
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Bare_Anonymous_Expr_Decl,
         Sequence_Type => Bare_Anonymous_Expr_Decl_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Bare_Anonymous_Expr_Decl;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Bare_Ada_Node;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Bare_Anonymous_Expr_Decl_Array_Access) return Bare_Anonymous_Expr_Decl_Array_Access is
      Ret : Bare_Anonymous_Expr_Decl_Array_Access := Create_Bare_Anonymous_Expr_Decl_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Bare_Anonymous_Expr_Decl_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Bare_Anonymous_Expr_Decl_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Bare_Anonymous_Expr_Decl_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Bare_Anonymous_Expr_Decl_Array (Items_Count : Natural) return Bare_Anonymous_Expr_Decl_Array_Access
   is (if Items_Count = 0
       then No_Bare_Anonymous_Expr_Decl_Array_Type
       else new Bare_Anonymous_Expr_Decl_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Bare_Anonymous_Expr_Decl_Array
     (Items : Internal_Bare_Anonymous_Expr_Decl_Array) return Bare_Anonymous_Expr_Decl_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Bare_Anonymous_Expr_Decl_Array_Type;
      end if;
      return new Bare_Anonymous_Expr_Decl_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Bare_Anonymous_Expr_Decl_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Bare_Anonymous_Expr_Decl_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Bare_Compilation_Unit_Vectors is new Langkit_Support.Vectors (Bare_Compilation_Unit);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Bare_Compilation_Unit_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Bare_Compilation_Unit
   is
      function Absolute_Get
        (T : Bare_Compilation_Unit_Array_Access; Index : Integer)
         return Bare_Compilation_Unit
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Bare_Compilation_Unit,
         Sequence_Type => Bare_Compilation_Unit_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Bare_Compilation_Unit;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Bare_Ada_Node;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Bare_Compilation_Unit_Array_Access) return Bare_Compilation_Unit_Array_Access is
      Ret : Bare_Compilation_Unit_Array_Access := Create_Bare_Compilation_Unit_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Bare_Compilation_Unit_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Bare_Compilation_Unit_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Bare_Compilation_Unit_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Bare_Compilation_Unit_Array (Items_Count : Natural) return Bare_Compilation_Unit_Array_Access
   is (if Items_Count = 0
       then No_Bare_Compilation_Unit_Array_Type
       else new Bare_Compilation_Unit_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Bare_Compilation_Unit_Array
     (Items : Internal_Bare_Compilation_Unit_Array) return Bare_Compilation_Unit_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Bare_Compilation_Unit_Array_Type;
      end if;
      return new Bare_Compilation_Unit_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Bare_Compilation_Unit_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Bare_Compilation_Unit_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Bare_Expr_Vectors is new Langkit_Support.Vectors (Bare_Expr);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Bare_Expr_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Bare_Expr
   is
      function Absolute_Get
        (T : Bare_Expr_Array_Access; Index : Integer)
         return Bare_Expr
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Bare_Expr,
         Sequence_Type => Bare_Expr_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Bare_Expr;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Bare_Ada_Node;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Bare_Expr_Array_Access) return Bare_Expr_Array_Access is
      Ret : Bare_Expr_Array_Access := Create_Bare_Expr_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Bare_Expr_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Bare_Expr_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Bare_Expr_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Bare_Expr_Array (Items_Count : Natural) return Bare_Expr_Array_Access
   is (if Items_Count = 0
       then No_Bare_Expr_Array_Type
       else new Bare_Expr_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Bare_Expr_Array
     (Items : Internal_Bare_Expr_Array) return Bare_Expr_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Bare_Expr_Array_Type;
      end if;
      return new Bare_Expr_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Bare_Expr_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Bare_Expr_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Bare_Name_Vectors is new Langkit_Support.Vectors (Bare_Name);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Bare_Name_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Bare_Name
   is
      function Absolute_Get
        (T : Bare_Name_Array_Access; Index : Integer)
         return Bare_Name
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Bare_Name,
         Sequence_Type => Bare_Name_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Bare_Name;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Bare_Ada_Node;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Bare_Name_Array_Access) return Bare_Name_Array_Access is
      Ret : Bare_Name_Array_Access := Create_Bare_Name_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Bare_Name_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Bare_Name_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Bare_Name_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Bare_Name_Array (Items_Count : Natural) return Bare_Name_Array_Access
   is (if Items_Count = 0
       then No_Bare_Name_Array_Type
       else new Bare_Name_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Bare_Name_Array
     (Items : Internal_Bare_Name_Array) return Bare_Name_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Bare_Name_Array_Type;
      end if;
      return new Bare_Name_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Bare_Name_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Bare_Name_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Bare_Pragma_Node_Vectors is new Langkit_Support.Vectors (Bare_Pragma_Node);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Bare_Pragma_Node_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Bare_Pragma_Node
   is
      function Absolute_Get
        (T : Bare_Pragma_Node_Array_Access; Index : Integer)
         return Bare_Pragma_Node
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Bare_Pragma_Node,
         Sequence_Type => Bare_Pragma_Node_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Bare_Pragma_Node;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Bare_Ada_Node;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Bare_Pragma_Node_Array_Access) return Bare_Pragma_Node_Array_Access is
      Ret : Bare_Pragma_Node_Array_Access := Create_Bare_Pragma_Node_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Bare_Pragma_Node_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Bare_Pragma_Node_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Bare_Pragma_Node_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Bare_Pragma_Node_Array (Items_Count : Natural) return Bare_Pragma_Node_Array_Access
   is (if Items_Count = 0
       then No_Bare_Pragma_Node_Array_Type
       else new Bare_Pragma_Node_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Bare_Pragma_Node_Array
     (Items : Internal_Bare_Pragma_Node_Array) return Bare_Pragma_Node_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Bare_Pragma_Node_Array_Type;
      end if;
      return new Bare_Pragma_Node_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Bare_Pragma_Node_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Bare_Pragma_Node_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
   ---------
   -- Get --
   ---------
   function Get
     (T       : Bare_Single_Tok_Node_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Bare_Single_Tok_Node
   is
      function Absolute_Get
        (T : Bare_Single_Tok_Node_Array_Access; Index : Integer)
         return Bare_Single_Tok_Node
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Bare_Single_Tok_Node,
         Sequence_Type => Bare_Single_Tok_Node_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Bare_Single_Tok_Node;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Bare_Ada_Node;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Bare_Single_Tok_Node_Array_Access) return Bare_Single_Tok_Node_Array_Access is
      Ret : Bare_Single_Tok_Node_Array_Access := Create_Bare_Single_Tok_Node_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Bare_Single_Tok_Node_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Bare_Single_Tok_Node_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Bare_Single_Tok_Node_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Bare_Single_Tok_Node_Array (Items_Count : Natural) return Bare_Single_Tok_Node_Array_Access
   is (if Items_Count = 0
       then No_Bare_Single_Tok_Node_Array_Type
       else new Bare_Single_Tok_Node_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Bare_Single_Tok_Node_Array
     (Items : Internal_Bare_Single_Tok_Node_Array) return Bare_Single_Tok_Node_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Bare_Single_Tok_Node_Array_Type;
      end if;
      return new Bare_Single_Tok_Node_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Bare_Single_Tok_Node_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Bare_Single_Tok_Node_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Bare_Use_Clause_Vectors is new Langkit_Support.Vectors (Bare_Use_Clause);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Bare_Use_Clause_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Bare_Use_Clause
   is
      function Absolute_Get
        (T : Bare_Use_Clause_Array_Access; Index : Integer)
         return Bare_Use_Clause
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Bare_Use_Clause,
         Sequence_Type => Bare_Use_Clause_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Bare_Use_Clause;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Bare_Ada_Node;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Bare_Use_Clause_Array_Access) return Bare_Use_Clause_Array_Access is
      Ret : Bare_Use_Clause_Array_Access := Create_Bare_Use_Clause_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Bare_Use_Clause_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Bare_Use_Clause_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Bare_Use_Clause_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Bare_Use_Clause_Array (Items_Count : Natural) return Bare_Use_Clause_Array_Access
   is (if Items_Count = 0
       then No_Bare_Use_Clause_Array_Type
       else new Bare_Use_Clause_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Bare_Use_Clause_Array
     (Items : Internal_Bare_Use_Clause_Array) return Bare_Use_Clause_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Bare_Use_Clause_Array_Type;
      end if;
      return new Bare_Use_Clause_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Bare_Use_Clause_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Bare_Use_Clause_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Aspect_Vectors is new Langkit_Support.Vectors (Internal_Aspect);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Aspect_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Aspect
   is
      function Absolute_Get
        (T : Internal_Aspect_Array_Access; Index : Integer)
         return Internal_Aspect
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Aspect,
         Sequence_Type => Internal_Aspect_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Aspect;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Aspect;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Aspect_Array_Access) return Internal_Aspect_Array_Access is
      Ret : Internal_Aspect_Array_Access := Create_Internal_Aspect_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Aspect_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Aspect_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Aspect_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Aspect_Array (Items_Count : Natural) return Internal_Aspect_Array_Access
   is (if Items_Count = 0
       then No_Internal_Aspect_Array_Type
       else new Internal_Aspect_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Aspect_Array
     (Items : Internal_Internal_Aspect_Array) return Internal_Aspect_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Aspect_Array_Type;
      end if;
      return new Internal_Aspect_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Aspect_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Aspect_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Completion_Item_Vectors is new Langkit_Support.Vectors (Internal_Completion_Item);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Completion_Item_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Completion_Item
   is
      function Absolute_Get
        (T : Internal_Completion_Item_Array_Access; Index : Integer)
         return Internal_Completion_Item
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Completion_Item,
         Sequence_Type => Internal_Completion_Item_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Completion_Item;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Completion_Item;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Completion_Item_Array_Access) return Internal_Completion_Item_Array_Access is
      Ret : Internal_Completion_Item_Array_Access := Create_Internal_Completion_Item_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Completion_Item_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Completion_Item_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Completion_Item_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Completion_Item_Array (Items_Count : Natural) return Internal_Completion_Item_Array_Access
   is (if Items_Count = 0
       then No_Internal_Completion_Item_Array_Type
       else new Internal_Completion_Item_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Completion_Item_Array
     (Items : Internal_Internal_Completion_Item_Array) return Internal_Completion_Item_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Completion_Item_Array_Type;
      end if;
      return new Internal_Completion_Item_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Completion_Item_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- To_Iterator --
      -----------------
      function To_Iterator
        (Self    : Internal_Completion_Item_Array_Access;
         Context : Internal_Context) return Internal_Completion_Item_Iterator_Access
      is
      begin
         Inc_Ref (Self);
         return new Internal_Internal_Completion_Item_Iterator'
           (Ref_Count  => 1,
            Safety_Net => Create_Safety_Net (Context),
            Elements   => Self,
            Index      => 1);
      end To_Iterator;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Completion_Item_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Discriminant_Values_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Discriminant_Values
   is
      function Absolute_Get
        (T : Internal_Discriminant_Values_Array_Access; Index : Integer)
         return Internal_Discriminant_Values
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Discriminant_Values,
         Sequence_Type => Internal_Discriminant_Values_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Discriminant_Values;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Discriminant_Values;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Discriminant_Values_Array_Access) return Internal_Discriminant_Values_Array_Access is
      Ret : Internal_Discriminant_Values_Array_Access := Create_Internal_Discriminant_Values_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Discriminant_Values_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Discriminant_Values_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Discriminant_Values_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Discriminant_Values_Array (Items_Count : Natural) return Internal_Discriminant_Values_Array_Access
   is (if Items_Count = 0
       then No_Internal_Discriminant_Values_Array_Type
       else new Internal_Discriminant_Values_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Discriminant_Values_Array
     (Items : Internal_Internal_Discriminant_Values_Array) return Internal_Discriminant_Values_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Discriminant_Values_Array_Type;
      end if;
      return new Internal_Discriminant_Values_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Discriminant_Values_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Discriminant_Values_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Doc_Annotation_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Doc_Annotation
   is
      function Absolute_Get
        (T : Internal_Doc_Annotation_Array_Access; Index : Integer)
         return Internal_Doc_Annotation
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Doc_Annotation,
         Sequence_Type => Internal_Doc_Annotation_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Doc_Annotation;
   begin
      if Relative_Get (T, Index, Result) then
            Inc_Ref (Result);
         return Result;
      elsif Or_Null then
         return No_Doc_Annotation;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Doc_Annotation_Array_Access) return Internal_Doc_Annotation_Array_Access is
      Ret : Internal_Doc_Annotation_Array_Access := Create_Internal_Doc_Annotation_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
         for Item of Ret.Items loop
            Inc_Ref (Item);
         end loop;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Doc_Annotation_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Doc_Annotation_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Doc_Annotation_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
            for Item of T.Items loop
               Dec_Ref (Item);
            end loop;
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Doc_Annotation_Array (Items_Count : Natural) return Internal_Doc_Annotation_Array_Access
   is (if Items_Count = 0
       then No_Internal_Doc_Annotation_Array_Type
       else new Internal_Doc_Annotation_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Doc_Annotation_Array
     (Items : Internal_Internal_Doc_Annotation_Array) return Internal_Doc_Annotation_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Doc_Annotation_Array_Type;
      end if;
         for El of Items loop
            Inc_Ref (El);
         end loop;
      return new Internal_Doc_Annotation_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Doc_Annotation_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               not Equivalent (L.Items (I), R.Items (I))
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Doc_Annotation_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Entity_Accept_Stmt_Vectors is new Langkit_Support.Vectors (Internal_Entity_Accept_Stmt);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Entity_Accept_Stmt_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Entity_Accept_Stmt
   is
      function Absolute_Get
        (T : Internal_Entity_Accept_Stmt_Array_Access; Index : Integer)
         return Internal_Entity_Accept_Stmt
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Entity_Accept_Stmt,
         Sequence_Type => Internal_Entity_Accept_Stmt_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Entity_Accept_Stmt;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Entity_Accept_Stmt;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Entity_Accept_Stmt_Array_Access) return Internal_Entity_Accept_Stmt_Array_Access is
      Ret : Internal_Entity_Accept_Stmt_Array_Access := Create_Internal_Entity_Accept_Stmt_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Accept_Stmt_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Entity_Accept_Stmt_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Accept_Stmt_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Entity_Accept_Stmt_Array (Items_Count : Natural) return Internal_Entity_Accept_Stmt_Array_Access
   is (if Items_Count = 0
       then No_Internal_Entity_Accept_Stmt_Array_Type
       else new Internal_Entity_Accept_Stmt_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Entity_Accept_Stmt_Array
     (Items : Internal_Internal_Entity_Accept_Stmt_Array) return Internal_Entity_Accept_Stmt_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Entity_Accept_Stmt_Array_Type;
      end if;
      return new Internal_Entity_Accept_Stmt_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Entity_Accept_Stmt_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Accept_Stmt_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Entity_Vectors is new Langkit_Support.Vectors (Internal_Entity);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Entity_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Entity
   is
      function Absolute_Get
        (T : Internal_Entity_Array_Access; Index : Integer)
         return Internal_Entity
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Entity,
         Sequence_Type => Internal_Entity_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Entity;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Entity;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Entity_Array_Access) return Internal_Entity_Array_Access is
      Ret : Internal_Entity_Array_Access := Create_Internal_Entity_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Entity_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Entity_Array (Items_Count : Natural) return Internal_Entity_Array_Access
   is (if Items_Count = 0
       then No_Internal_Entity_Array_Type
       else new Internal_Entity_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Entity_Array
     (Items : AST_Envs.Entity_Array) return Internal_Entity_Array_Access
   is (if Items'Length = 0
       then No_Internal_Entity_Array_Type
       else new Internal_Entity_Array_Record'
         (N         => Items'Length,
          Items     => Implementation.Internal_Internal_Entity_Array (Items),
          Ref_Count => 1));
   function Create_Internal_Entity_Array
     (Items : Internal_Internal_Entity_Array) return Internal_Entity_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Entity_Array_Type;
      end if;
      return new Internal_Entity_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Entity_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Entity_Base_Assoc_Vectors is new Langkit_Support.Vectors (Internal_Entity_Base_Assoc);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Entity_Base_Assoc_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Entity_Base_Assoc
   is
      function Absolute_Get
        (T : Internal_Entity_Base_Assoc_Array_Access; Index : Integer)
         return Internal_Entity_Base_Assoc
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Entity_Base_Assoc,
         Sequence_Type => Internal_Entity_Base_Assoc_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Entity_Base_Assoc;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Entity_Base_Assoc;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Entity_Base_Assoc_Array_Access) return Internal_Entity_Base_Assoc_Array_Access is
      Ret : Internal_Entity_Base_Assoc_Array_Access := Create_Internal_Entity_Base_Assoc_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Base_Assoc_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Entity_Base_Assoc_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Base_Assoc_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Entity_Base_Assoc_Array (Items_Count : Natural) return Internal_Entity_Base_Assoc_Array_Access
   is (if Items_Count = 0
       then No_Internal_Entity_Base_Assoc_Array_Type
       else new Internal_Entity_Base_Assoc_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Entity_Base_Assoc_Array
     (Items : Internal_Internal_Entity_Base_Assoc_Array) return Internal_Entity_Base_Assoc_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Entity_Base_Assoc_Array_Type;
      end if;
      return new Internal_Entity_Base_Assoc_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Entity_Base_Assoc_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Base_Assoc_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Entity_Base_Formal_Param_Decl_Vectors is new Langkit_Support.Vectors (Internal_Entity_Base_Formal_Param_Decl);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Entity_Base_Formal_Param_Decl
   is
      function Absolute_Get
        (T : Internal_Entity_Base_Formal_Param_Decl_Array_Access; Index : Integer)
         return Internal_Entity_Base_Formal_Param_Decl
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Entity_Base_Formal_Param_Decl,
         Sequence_Type => Internal_Entity_Base_Formal_Param_Decl_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Entity_Base_Formal_Param_Decl;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Entity_Base_Formal_Param_Decl;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Entity_Base_Formal_Param_Decl_Array_Access) return Internal_Entity_Base_Formal_Param_Decl_Array_Access is
      Ret : Internal_Entity_Base_Formal_Param_Decl_Array_Access := Create_Internal_Entity_Base_Formal_Param_Decl_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Base_Formal_Param_Decl_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Entity_Base_Formal_Param_Decl_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Base_Formal_Param_Decl_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Entity_Base_Formal_Param_Decl_Array (Items_Count : Natural) return Internal_Entity_Base_Formal_Param_Decl_Array_Access
   is (if Items_Count = 0
       then No_Internal_Entity_Base_Formal_Param_Decl_Array_Type
       else new Internal_Entity_Base_Formal_Param_Decl_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Entity_Base_Formal_Param_Decl_Array
     (Items : Internal_Internal_Entity_Base_Formal_Param_Decl_Array) return Internal_Entity_Base_Formal_Param_Decl_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Entity_Base_Formal_Param_Decl_Array_Type;
      end if;
      return new Internal_Entity_Base_Formal_Param_Decl_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Entity_Base_Formal_Param_Decl_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Base_Formal_Param_Decl_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Entity_Base_Type_Decl_Vectors is new Langkit_Support.Vectors (Internal_Entity_Base_Type_Decl);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Entity_Base_Type_Decl_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Entity_Base_Type_Decl
   is
      function Absolute_Get
        (T : Internal_Entity_Base_Type_Decl_Array_Access; Index : Integer)
         return Internal_Entity_Base_Type_Decl
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Entity_Base_Type_Decl,
         Sequence_Type => Internal_Entity_Base_Type_Decl_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Entity_Base_Type_Decl;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Entity_Base_Type_Decl;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Entity_Base_Type_Decl_Array_Access) return Internal_Entity_Base_Type_Decl_Array_Access is
      Ret : Internal_Entity_Base_Type_Decl_Array_Access := Create_Internal_Entity_Base_Type_Decl_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Base_Type_Decl_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Entity_Base_Type_Decl_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Base_Type_Decl_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Entity_Base_Type_Decl_Array (Items_Count : Natural) return Internal_Entity_Base_Type_Decl_Array_Access
   is (if Items_Count = 0
       then No_Internal_Entity_Base_Type_Decl_Array_Type
       else new Internal_Entity_Base_Type_Decl_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Entity_Base_Type_Decl_Array
     (Items : Internal_Internal_Entity_Base_Type_Decl_Array) return Internal_Entity_Base_Type_Decl_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Entity_Base_Type_Decl_Array_Type;
      end if;
      return new Internal_Entity_Base_Type_Decl_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Entity_Base_Type_Decl_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Base_Type_Decl_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      function Make_Unique (A : Internal_Entity_Base_Type_Decl_Array_Access) return Internal_Entity_Base_Type_Decl_Array_Access is
         --  To keep the order of the result deterministic, this function works
         --  in two steps. First, determine for each item in A whether it
         --  duplicate a previous item. Only then, go through all items in A
         --  and add only non-duplicated ones to the result.
         Duplicates : array (A.Items'Range) of Boolean := (others => False);
         --  For each item in A, indicate whether the value is a duplicate of a
         --  previous one in the array.
         Unique_Elements : Natural;
         --  Number of unique elements in A
      begin
         --  Step 1: compute Duplicates
         declare
            package Sets is new Ada.Containers.Hashed_Sets
              (Element_Type        => Internal_Entity_Base_Type_Decl,
               "="                 => "=",
               Equivalent_Elements => "=",
               Hash                => Hash);
            Set : Sets.Set;
         begin
            for I in A.Items'Range loop
               declare
                  Item : Internal_Entity_Base_Type_Decl renames A.Items (I);
               begin
                  if Set.Contains (Item) then
                     Duplicates (I) := True;
                  else
                     Set.Insert (Item);
                  end if;
               end;
            end loop;
            Unique_Elements := Natural (Set.Length);
         end;
         --  Step 2: create the result
         declare
            Next   : Positive := 1;
            Result : constant Internal_Entity_Base_Type_Decl_Array_Access :=
               Create_Internal_Entity_Base_Type_Decl_Array (Unique_Elements);
         begin
            for I in A.Items'Range loop
               declare
                  Item : Internal_Entity_Base_Type_Decl renames A.Items (I);
               begin
                  if not Duplicates (I) then
                     Result.Items (Next) := Item;
                     Next := Next + 1;
                  end if;
               end;
            end loop;
            pragma Assert (Next = Result.N + 1);
            return Result;
         end;
      end Make_Unique;
      ----------
      -- Hash --
      ----------
      function Hash (R : Internal_Entity_Base_Type_Decl_Array_Access) return Hash_Type is
         Result : Hash_Type := Initial_Hash;
      begin
         for I in R.Items'Range loop
            Result := Combine (Result, Hash (R.Items (I)));
         end loop;
         return Result;
      end Hash;
      package Internal_Entity_Basic_Assoc_Vectors is new Langkit_Support.Vectors (Internal_Entity_Basic_Assoc);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Entity_Basic_Assoc_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Entity_Basic_Assoc
   is
      function Absolute_Get
        (T : Internal_Entity_Basic_Assoc_Array_Access; Index : Integer)
         return Internal_Entity_Basic_Assoc
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Entity_Basic_Assoc,
         Sequence_Type => Internal_Entity_Basic_Assoc_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Entity_Basic_Assoc;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Entity_Basic_Assoc;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Entity_Basic_Assoc_Array_Access) return Internal_Entity_Basic_Assoc_Array_Access is
      Ret : Internal_Entity_Basic_Assoc_Array_Access := Create_Internal_Entity_Basic_Assoc_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Basic_Assoc_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Entity_Basic_Assoc_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Basic_Assoc_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Entity_Basic_Assoc_Array (Items_Count : Natural) return Internal_Entity_Basic_Assoc_Array_Access
   is (if Items_Count = 0
       then No_Internal_Entity_Basic_Assoc_Array_Type
       else new Internal_Entity_Basic_Assoc_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Entity_Basic_Assoc_Array
     (Items : Internal_Internal_Entity_Basic_Assoc_Array) return Internal_Entity_Basic_Assoc_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Entity_Basic_Assoc_Array_Type;
      end if;
      return new Internal_Entity_Basic_Assoc_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Entity_Basic_Assoc_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Basic_Assoc_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Entity_Basic_Decl_Vectors is new Langkit_Support.Vectors (Internal_Entity_Basic_Decl);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Entity_Basic_Decl_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Entity_Basic_Decl
   is
      function Absolute_Get
        (T : Internal_Entity_Basic_Decl_Array_Access; Index : Integer)
         return Internal_Entity_Basic_Decl
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Entity_Basic_Decl,
         Sequence_Type => Internal_Entity_Basic_Decl_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Entity_Basic_Decl;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Entity_Basic_Decl;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Entity_Basic_Decl_Array_Access) return Internal_Entity_Basic_Decl_Array_Access is
      Ret : Internal_Entity_Basic_Decl_Array_Access := Create_Internal_Entity_Basic_Decl_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Basic_Decl_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Entity_Basic_Decl_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Basic_Decl_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Entity_Basic_Decl_Array (Items_Count : Natural) return Internal_Entity_Basic_Decl_Array_Access
   is (if Items_Count = 0
       then No_Internal_Entity_Basic_Decl_Array_Type
       else new Internal_Entity_Basic_Decl_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Entity_Basic_Decl_Array
     (Items : Internal_Internal_Entity_Basic_Decl_Array) return Internal_Entity_Basic_Decl_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Entity_Basic_Decl_Array_Type;
      end if;
      return new Internal_Entity_Basic_Decl_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Entity_Basic_Decl_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Basic_Decl_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      function Make_Unique (A : Internal_Entity_Basic_Decl_Array_Access) return Internal_Entity_Basic_Decl_Array_Access is
         --  To keep the order of the result deterministic, this function works
         --  in two steps. First, determine for each item in A whether it
         --  duplicate a previous item. Only then, go through all items in A
         --  and add only non-duplicated ones to the result.
         Duplicates : array (A.Items'Range) of Boolean := (others => False);
         --  For each item in A, indicate whether the value is a duplicate of a
         --  previous one in the array.
         Unique_Elements : Natural;
         --  Number of unique elements in A
      begin
         --  Step 1: compute Duplicates
         declare
            package Sets is new Ada.Containers.Hashed_Sets
              (Element_Type        => Internal_Entity_Basic_Decl,
               "="                 => "=",
               Equivalent_Elements => "=",
               Hash                => Hash);
            Set : Sets.Set;
         begin
            for I in A.Items'Range loop
               declare
                  Item : Internal_Entity_Basic_Decl renames A.Items (I);
               begin
                  if Set.Contains (Item) then
                     Duplicates (I) := True;
                  else
                     Set.Insert (Item);
                  end if;
               end;
            end loop;
            Unique_Elements := Natural (Set.Length);
         end;
         --  Step 2: create the result
         declare
            Next   : Positive := 1;
            Result : constant Internal_Entity_Basic_Decl_Array_Access :=
               Create_Internal_Entity_Basic_Decl_Array (Unique_Elements);
         begin
            for I in A.Items'Range loop
               declare
                  Item : Internal_Entity_Basic_Decl renames A.Items (I);
               begin
                  if not Duplicates (I) then
                     Result.Items (Next) := Item;
                     Next := Next + 1;
                  end if;
               end;
            end loop;
            pragma Assert (Next = Result.N + 1);
            return Result;
         end;
      end Make_Unique;
      package Internal_Entity_Compilation_Unit_Vectors is new Langkit_Support.Vectors (Internal_Entity_Compilation_Unit);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Entity_Compilation_Unit_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Entity_Compilation_Unit
   is
      function Absolute_Get
        (T : Internal_Entity_Compilation_Unit_Array_Access; Index : Integer)
         return Internal_Entity_Compilation_Unit
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Entity_Compilation_Unit,
         Sequence_Type => Internal_Entity_Compilation_Unit_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Entity_Compilation_Unit;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Entity_Compilation_Unit;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Entity_Compilation_Unit_Array_Access) return Internal_Entity_Compilation_Unit_Array_Access is
      Ret : Internal_Entity_Compilation_Unit_Array_Access := Create_Internal_Entity_Compilation_Unit_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Compilation_Unit_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Entity_Compilation_Unit_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Compilation_Unit_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Entity_Compilation_Unit_Array (Items_Count : Natural) return Internal_Entity_Compilation_Unit_Array_Access
   is (if Items_Count = 0
       then No_Internal_Entity_Compilation_Unit_Array_Type
       else new Internal_Entity_Compilation_Unit_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Entity_Compilation_Unit_Array
     (Items : Internal_Internal_Entity_Compilation_Unit_Array) return Internal_Entity_Compilation_Unit_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Entity_Compilation_Unit_Array_Type;
      end if;
      return new Internal_Entity_Compilation_Unit_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Entity_Compilation_Unit_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Compilation_Unit_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      function Make_Unique (A : Internal_Entity_Compilation_Unit_Array_Access) return Internal_Entity_Compilation_Unit_Array_Access is
         --  To keep the order of the result deterministic, this function works
         --  in two steps. First, determine for each item in A whether it
         --  duplicate a previous item. Only then, go through all items in A
         --  and add only non-duplicated ones to the result.
         Duplicates : array (A.Items'Range) of Boolean := (others => False);
         --  For each item in A, indicate whether the value is a duplicate of a
         --  previous one in the array.
         Unique_Elements : Natural;
         --  Number of unique elements in A
      begin
         --  Step 1: compute Duplicates
         declare
            package Sets is new Ada.Containers.Hashed_Sets
              (Element_Type        => Internal_Entity_Compilation_Unit,
               "="                 => "=",
               Equivalent_Elements => "=",
               Hash                => Hash);
            Set : Sets.Set;
         begin
            for I in A.Items'Range loop
               declare
                  Item : Internal_Entity_Compilation_Unit renames A.Items (I);
               begin
                  if Set.Contains (Item) then
                     Duplicates (I) := True;
                  else
                     Set.Insert (Item);
                  end if;
               end;
            end loop;
            Unique_Elements := Natural (Set.Length);
         end;
         --  Step 2: create the result
         declare
            Next   : Positive := 1;
            Result : constant Internal_Entity_Compilation_Unit_Array_Access :=
               Create_Internal_Entity_Compilation_Unit_Array (Unique_Elements);
         begin
            for I in A.Items'Range loop
               declare
                  Item : Internal_Entity_Compilation_Unit renames A.Items (I);
               begin
                  if not Duplicates (I) then
                     Result.Items (Next) := Item;
                     Next := Next + 1;
                  end if;
               end;
            end loop;
            pragma Assert (Next = Result.N + 1);
            return Result;
         end;
      end Make_Unique;
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Entity_Declarative_Part_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Entity_Declarative_Part
   is
      function Absolute_Get
        (T : Internal_Entity_Declarative_Part_Array_Access; Index : Integer)
         return Internal_Entity_Declarative_Part
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Entity_Declarative_Part,
         Sequence_Type => Internal_Entity_Declarative_Part_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Entity_Declarative_Part;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Entity_Declarative_Part;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Entity_Declarative_Part_Array_Access) return Internal_Entity_Declarative_Part_Array_Access is
      Ret : Internal_Entity_Declarative_Part_Array_Access := Create_Internal_Entity_Declarative_Part_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Declarative_Part_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Entity_Declarative_Part_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Declarative_Part_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Entity_Declarative_Part_Array (Items_Count : Natural) return Internal_Entity_Declarative_Part_Array_Access
   is (if Items_Count = 0
       then No_Internal_Entity_Declarative_Part_Array_Type
       else new Internal_Entity_Declarative_Part_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Entity_Declarative_Part_Array
     (Items : Internal_Internal_Entity_Declarative_Part_Array) return Internal_Entity_Declarative_Part_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Entity_Declarative_Part_Array_Type;
      end if;
      return new Internal_Entity_Declarative_Part_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Entity_Declarative_Part_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Declarative_Part_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Entity_Defining_Name_Vectors is new Langkit_Support.Vectors (Internal_Entity_Defining_Name);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Entity_Defining_Name_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Entity_Defining_Name
   is
      function Absolute_Get
        (T : Internal_Entity_Defining_Name_Array_Access; Index : Integer)
         return Internal_Entity_Defining_Name
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Entity_Defining_Name,
         Sequence_Type => Internal_Entity_Defining_Name_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Entity_Defining_Name;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Entity_Defining_Name;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Entity_Defining_Name_Array_Access) return Internal_Entity_Defining_Name_Array_Access is
      Ret : Internal_Entity_Defining_Name_Array_Access := Create_Internal_Entity_Defining_Name_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Defining_Name_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Entity_Defining_Name_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Defining_Name_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Entity_Defining_Name_Array (Items_Count : Natural) return Internal_Entity_Defining_Name_Array_Access
   is (if Items_Count = 0
       then No_Internal_Entity_Defining_Name_Array_Type
       else new Internal_Entity_Defining_Name_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Entity_Defining_Name_Array
     (Items : Internal_Internal_Entity_Defining_Name_Array) return Internal_Entity_Defining_Name_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Entity_Defining_Name_Array_Type;
      end if;
      return new Internal_Entity_Defining_Name_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Entity_Defining_Name_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Defining_Name_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Entity_Expr_Vectors is new Langkit_Support.Vectors (Internal_Entity_Expr);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Entity_Expr_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Entity_Expr
   is
      function Absolute_Get
        (T : Internal_Entity_Expr_Array_Access; Index : Integer)
         return Internal_Entity_Expr
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Entity_Expr,
         Sequence_Type => Internal_Entity_Expr_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Entity_Expr;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Entity_Expr;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Entity_Expr_Array_Access) return Internal_Entity_Expr_Array_Access is
      Ret : Internal_Entity_Expr_Array_Access := Create_Internal_Entity_Expr_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Expr_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Entity_Expr_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Expr_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Entity_Expr_Array (Items_Count : Natural) return Internal_Entity_Expr_Array_Access
   is (if Items_Count = 0
       then No_Internal_Entity_Expr_Array_Type
       else new Internal_Entity_Expr_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Entity_Expr_Array
     (Items : Internal_Internal_Entity_Expr_Array) return Internal_Entity_Expr_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Entity_Expr_Array_Type;
      end if;
      return new Internal_Entity_Expr_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Entity_Expr_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Expr_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Entity_Generic_Instantiation_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Entity_Generic_Instantiation
   is
      function Absolute_Get
        (T : Internal_Entity_Generic_Instantiation_Array_Access; Index : Integer)
         return Internal_Entity_Generic_Instantiation
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Entity_Generic_Instantiation,
         Sequence_Type => Internal_Entity_Generic_Instantiation_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Entity_Generic_Instantiation;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Entity_Generic_Instantiation;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Entity_Generic_Instantiation_Array_Access) return Internal_Entity_Generic_Instantiation_Array_Access is
      Ret : Internal_Entity_Generic_Instantiation_Array_Access := Create_Internal_Entity_Generic_Instantiation_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Generic_Instantiation_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Entity_Generic_Instantiation_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Generic_Instantiation_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Entity_Generic_Instantiation_Array (Items_Count : Natural) return Internal_Entity_Generic_Instantiation_Array_Access
   is (if Items_Count = 0
       then No_Internal_Entity_Generic_Instantiation_Array_Type
       else new Internal_Entity_Generic_Instantiation_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Entity_Generic_Instantiation_Array
     (Items : Internal_Internal_Entity_Generic_Instantiation_Array) return Internal_Entity_Generic_Instantiation_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Entity_Generic_Instantiation_Array_Type;
      end if;
      return new Internal_Entity_Generic_Instantiation_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Entity_Generic_Instantiation_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Generic_Instantiation_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Entity_Identifier_Vectors is new Langkit_Support.Vectors (Internal_Entity_Identifier);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Entity_Identifier_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Entity_Identifier
   is
      function Absolute_Get
        (T : Internal_Entity_Identifier_Array_Access; Index : Integer)
         return Internal_Entity_Identifier
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Entity_Identifier,
         Sequence_Type => Internal_Entity_Identifier_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Entity_Identifier;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Entity_Identifier;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Entity_Identifier_Array_Access) return Internal_Entity_Identifier_Array_Access is
      Ret : Internal_Entity_Identifier_Array_Access := Create_Internal_Entity_Identifier_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Identifier_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Entity_Identifier_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Identifier_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Entity_Identifier_Array (Items_Count : Natural) return Internal_Entity_Identifier_Array_Access
   is (if Items_Count = 0
       then No_Internal_Entity_Identifier_Array_Type
       else new Internal_Entity_Identifier_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Entity_Identifier_Array
     (Items : Internal_Internal_Entity_Identifier_Array) return Internal_Entity_Identifier_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Entity_Identifier_Array_Type;
      end if;
      return new Internal_Entity_Identifier_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Entity_Identifier_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Identifier_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Entity_Mode_Vectors is new Langkit_Support.Vectors (Internal_Entity_Mode);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Entity_Mode_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Entity_Mode
   is
      function Absolute_Get
        (T : Internal_Entity_Mode_Array_Access; Index : Integer)
         return Internal_Entity_Mode
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Entity_Mode,
         Sequence_Type => Internal_Entity_Mode_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Entity_Mode;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Entity_Mode;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Entity_Mode_Array_Access) return Internal_Entity_Mode_Array_Access is
      Ret : Internal_Entity_Mode_Array_Access := Create_Internal_Entity_Mode_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Mode_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Entity_Mode_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Mode_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Entity_Mode_Array (Items_Count : Natural) return Internal_Entity_Mode_Array_Access
   is (if Items_Count = 0
       then No_Internal_Entity_Mode_Array_Type
       else new Internal_Entity_Mode_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Entity_Mode_Array
     (Items : Internal_Internal_Entity_Mode_Array) return Internal_Entity_Mode_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Entity_Mode_Array_Type;
      end if;
      return new Internal_Entity_Mode_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Entity_Mode_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Mode_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Entity_Name_Vectors is new Langkit_Support.Vectors (Internal_Entity_Name);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Entity_Name_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Entity_Name
   is
      function Absolute_Get
        (T : Internal_Entity_Name_Array_Access; Index : Integer)
         return Internal_Entity_Name
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Entity_Name,
         Sequence_Type => Internal_Entity_Name_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Entity_Name;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Entity_Name;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Entity_Name_Array_Access) return Internal_Entity_Name_Array_Access is
      Ret : Internal_Entity_Name_Array_Access := Create_Internal_Entity_Name_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Name_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Entity_Name_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Name_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Entity_Name_Array (Items_Count : Natural) return Internal_Entity_Name_Array_Access
   is (if Items_Count = 0
       then No_Internal_Entity_Name_Array_Type
       else new Internal_Entity_Name_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Entity_Name_Array
     (Items : Internal_Internal_Entity_Name_Array) return Internal_Entity_Name_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Entity_Name_Array_Type;
      end if;
      return new Internal_Entity_Name_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Entity_Name_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Name_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Entity_Param_Spec_Vectors is new Langkit_Support.Vectors (Internal_Entity_Param_Spec);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Entity_Param_Spec_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Entity_Param_Spec
   is
      function Absolute_Get
        (T : Internal_Entity_Param_Spec_Array_Access; Index : Integer)
         return Internal_Entity_Param_Spec
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Entity_Param_Spec,
         Sequence_Type => Internal_Entity_Param_Spec_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Entity_Param_Spec;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Entity_Param_Spec;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Entity_Param_Spec_Array_Access) return Internal_Entity_Param_Spec_Array_Access is
      Ret : Internal_Entity_Param_Spec_Array_Access := Create_Internal_Entity_Param_Spec_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Param_Spec_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Entity_Param_Spec_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Param_Spec_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Entity_Param_Spec_Array (Items_Count : Natural) return Internal_Entity_Param_Spec_Array_Access
   is (if Items_Count = 0
       then No_Internal_Entity_Param_Spec_Array_Type
       else new Internal_Entity_Param_Spec_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Entity_Param_Spec_Array
     (Items : Internal_Internal_Entity_Param_Spec_Array) return Internal_Entity_Param_Spec_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Entity_Param_Spec_Array_Type;
      end if;
      return new Internal_Entity_Param_Spec_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Entity_Param_Spec_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Param_Spec_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Entity_Pragma_Node_Vectors is new Langkit_Support.Vectors (Internal_Entity_Pragma_Node);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Entity_Pragma_Node_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Entity_Pragma_Node
   is
      function Absolute_Get
        (T : Internal_Entity_Pragma_Node_Array_Access; Index : Integer)
         return Internal_Entity_Pragma_Node
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Entity_Pragma_Node,
         Sequence_Type => Internal_Entity_Pragma_Node_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Entity_Pragma_Node;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Entity_Pragma_Node;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Entity_Pragma_Node_Array_Access) return Internal_Entity_Pragma_Node_Array_Access is
      Ret : Internal_Entity_Pragma_Node_Array_Access := Create_Internal_Entity_Pragma_Node_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Pragma_Node_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Entity_Pragma_Node_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Pragma_Node_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Entity_Pragma_Node_Array (Items_Count : Natural) return Internal_Entity_Pragma_Node_Array_Access
   is (if Items_Count = 0
       then No_Internal_Entity_Pragma_Node_Array_Type
       else new Internal_Entity_Pragma_Node_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Entity_Pragma_Node_Array
     (Items : Internal_Internal_Entity_Pragma_Node_Array) return Internal_Entity_Pragma_Node_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Entity_Pragma_Node_Array_Type;
      end if;
      return new Internal_Entity_Pragma_Node_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Entity_Pragma_Node_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Pragma_Node_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Entity_Type_Decl_Vectors is new Langkit_Support.Vectors (Internal_Entity_Type_Decl);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Entity_Type_Decl_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Entity_Type_Decl
   is
      function Absolute_Get
        (T : Internal_Entity_Type_Decl_Array_Access; Index : Integer)
         return Internal_Entity_Type_Decl
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Entity_Type_Decl,
         Sequence_Type => Internal_Entity_Type_Decl_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Entity_Type_Decl;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Entity_Type_Decl;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Entity_Type_Decl_Array_Access) return Internal_Entity_Type_Decl_Array_Access is
      Ret : Internal_Entity_Type_Decl_Array_Access := Create_Internal_Entity_Type_Decl_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Type_Decl_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Entity_Type_Decl_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Type_Decl_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Entity_Type_Decl_Array (Items_Count : Natural) return Internal_Entity_Type_Decl_Array_Access
   is (if Items_Count = 0
       then No_Internal_Entity_Type_Decl_Array_Type
       else new Internal_Entity_Type_Decl_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Entity_Type_Decl_Array
     (Items : Internal_Internal_Entity_Type_Decl_Array) return Internal_Entity_Type_Decl_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Entity_Type_Decl_Array_Type;
      end if;
      return new Internal_Entity_Type_Decl_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Entity_Type_Decl_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Type_Decl_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Entity_Type_Expr_Vectors is new Langkit_Support.Vectors (Internal_Entity_Type_Expr);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Entity_Type_Expr_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Entity_Type_Expr
   is
      function Absolute_Get
        (T : Internal_Entity_Type_Expr_Array_Access; Index : Integer)
         return Internal_Entity_Type_Expr
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Entity_Type_Expr,
         Sequence_Type => Internal_Entity_Type_Expr_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Entity_Type_Expr;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Entity_Type_Expr;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Entity_Type_Expr_Array_Access) return Internal_Entity_Type_Expr_Array_Access is
      Ret : Internal_Entity_Type_Expr_Array_Access := Create_Internal_Entity_Type_Expr_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Type_Expr_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Entity_Type_Expr_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Type_Expr_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Entity_Type_Expr_Array (Items_Count : Natural) return Internal_Entity_Type_Expr_Array_Access
   is (if Items_Count = 0
       then No_Internal_Entity_Type_Expr_Array_Type
       else new Internal_Entity_Type_Expr_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Entity_Type_Expr_Array
     (Items : Internal_Internal_Entity_Type_Expr_Array) return Internal_Entity_Type_Expr_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Entity_Type_Expr_Array_Type;
      end if;
      return new Internal_Entity_Type_Expr_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Entity_Type_Expr_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Type_Expr_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Env_Assoc_Vectors is new Langkit_Support.Vectors (Internal_Env_Assoc);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Env_Assoc_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Env_Assoc
   is
      function Absolute_Get
        (T : Internal_Env_Assoc_Array_Access; Index : Integer)
         return Internal_Env_Assoc
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Env_Assoc,
         Sequence_Type => Internal_Env_Assoc_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Env_Assoc;
   begin
      if Relative_Get (T, Index, Result) then
            Inc_Ref (Result);
         return Result;
      elsif Or_Null then
         return No_Env_Assoc;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Env_Assoc_Array_Access) return Internal_Env_Assoc_Array_Access is
      Ret : Internal_Env_Assoc_Array_Access := Create_Internal_Env_Assoc_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
         for Item of Ret.Items loop
            Inc_Ref (Item);
         end loop;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Env_Assoc_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Env_Assoc_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Env_Assoc_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
            for Item of T.Items loop
               Dec_Ref (Item);
            end loop;
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Env_Assoc_Array (Items_Count : Natural) return Internal_Env_Assoc_Array_Access
   is (if Items_Count = 0
       then No_Internal_Env_Assoc_Array_Type
       else new Internal_Env_Assoc_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Env_Assoc_Array
     (Items : Internal_Internal_Env_Assoc_Array) return Internal_Env_Assoc_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Env_Assoc_Array_Type;
      end if;
         for El of Items loop
            Inc_Ref (El);
         end loop;
      return new Internal_Env_Assoc_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Env_Assoc_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               not Equivalent (L.Items (I), R.Items (I))
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Env_Assoc_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Expected_Type_For_Expr_Vectors is new Langkit_Support.Vectors (Internal_Expected_Type_For_Expr);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Expected_Type_For_Expr_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Expected_Type_For_Expr
   is
      function Absolute_Get
        (T : Internal_Expected_Type_For_Expr_Array_Access; Index : Integer)
         return Internal_Expected_Type_For_Expr
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Expected_Type_For_Expr,
         Sequence_Type => Internal_Expected_Type_For_Expr_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Expected_Type_For_Expr;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Expected_Type_For_Expr;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Expected_Type_For_Expr_Array_Access) return Internal_Expected_Type_For_Expr_Array_Access is
      Ret : Internal_Expected_Type_For_Expr_Array_Access := Create_Internal_Expected_Type_For_Expr_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Expected_Type_For_Expr_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Expected_Type_For_Expr_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Expected_Type_For_Expr_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Expected_Type_For_Expr_Array (Items_Count : Natural) return Internal_Expected_Type_For_Expr_Array_Access
   is (if Items_Count = 0
       then No_Internal_Expected_Type_For_Expr_Array_Type
       else new Internal_Expected_Type_For_Expr_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Expected_Type_For_Expr_Array
     (Items : Internal_Internal_Expected_Type_For_Expr_Array) return Internal_Expected_Type_For_Expr_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Expected_Type_For_Expr_Array_Type;
      end if;
      return new Internal_Expected_Type_For_Expr_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Expected_Type_For_Expr_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Expected_Type_For_Expr_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Inner_Env_Assoc_Vectors is new Langkit_Support.Vectors (Internal_Inner_Env_Assoc);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Inner_Env_Assoc_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Inner_Env_Assoc
   is
      function Absolute_Get
        (T : Internal_Inner_Env_Assoc_Array_Access; Index : Integer)
         return Internal_Inner_Env_Assoc
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Inner_Env_Assoc,
         Sequence_Type => Internal_Inner_Env_Assoc_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Inner_Env_Assoc;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Inner_Env_Assoc;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Inner_Env_Assoc_Array_Access) return Internal_Inner_Env_Assoc_Array_Access is
      Ret : Internal_Inner_Env_Assoc_Array_Access := Create_Internal_Inner_Env_Assoc_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Inner_Env_Assoc_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Inner_Env_Assoc_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Inner_Env_Assoc_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Inner_Env_Assoc_Array (Items_Count : Natural) return Internal_Inner_Env_Assoc_Array_Access
   is (if Items_Count = 0
       then No_Internal_Inner_Env_Assoc_Array_Type
       else new Internal_Inner_Env_Assoc_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Inner_Env_Assoc_Array
     (Items : Internal_Internal_Inner_Env_Assoc_Array) return Internal_Inner_Env_Assoc_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Inner_Env_Assoc_Array_Type;
      end if;
      return new Internal_Inner_Env_Assoc_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Inner_Env_Assoc_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Inner_Env_Assoc_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Param_Actual_Vectors is new Langkit_Support.Vectors (Internal_Param_Actual);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Param_Actual_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Param_Actual
   is
      function Absolute_Get
        (T : Internal_Param_Actual_Array_Access; Index : Integer)
         return Internal_Param_Actual
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Param_Actual,
         Sequence_Type => Internal_Param_Actual_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Param_Actual;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Param_Actual;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Param_Actual_Array_Access) return Internal_Param_Actual_Array_Access is
      Ret : Internal_Param_Actual_Array_Access := Create_Internal_Param_Actual_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Param_Actual_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Param_Actual_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Param_Actual_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Param_Actual_Array (Items_Count : Natural) return Internal_Param_Actual_Array_Access
   is (if Items_Count = 0
       then No_Internal_Param_Actual_Array_Type
       else new Internal_Param_Actual_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Param_Actual_Array
     (Items : Internal_Internal_Param_Actual_Array) return Internal_Param_Actual_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Param_Actual_Array_Type;
      end if;
      return new Internal_Param_Actual_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Param_Actual_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Param_Actual_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Param_Match_Vectors is new Langkit_Support.Vectors (Internal_Param_Match);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Param_Match_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Param_Match
   is
      function Absolute_Get
        (T : Internal_Param_Match_Array_Access; Index : Integer)
         return Internal_Param_Match
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Param_Match,
         Sequence_Type => Internal_Param_Match_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Param_Match;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Param_Match;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Param_Match_Array_Access) return Internal_Param_Match_Array_Access is
      Ret : Internal_Param_Match_Array_Access := Create_Internal_Param_Match_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Param_Match_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Param_Match_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Param_Match_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Param_Match_Array (Items_Count : Natural) return Internal_Param_Match_Array_Access
   is (if Items_Count = 0
       then No_Internal_Param_Match_Array_Type
       else new Internal_Param_Match_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Param_Match_Array
     (Items : Internal_Internal_Param_Match_Array) return Internal_Param_Match_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Param_Match_Array_Type;
      end if;
      return new Internal_Param_Match_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Param_Match_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Param_Match_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Ref_Result_Vectors is new Langkit_Support.Vectors (Internal_Ref_Result);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Ref_Result_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Ref_Result
   is
      function Absolute_Get
        (T : Internal_Ref_Result_Array_Access; Index : Integer)
         return Internal_Ref_Result
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Ref_Result,
         Sequence_Type => Internal_Ref_Result_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Ref_Result;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Ref_Result;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Ref_Result_Array_Access) return Internal_Ref_Result_Array_Access is
      Ret : Internal_Ref_Result_Array_Access := Create_Internal_Ref_Result_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Ref_Result_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Ref_Result_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Ref_Result_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Ref_Result_Array (Items_Count : Natural) return Internal_Ref_Result_Array_Access
   is (if Items_Count = 0
       then No_Internal_Ref_Result_Array_Type
       else new Internal_Ref_Result_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Ref_Result_Array
     (Items : Internal_Internal_Ref_Result_Array) return Internal_Ref_Result_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Ref_Result_Array_Type;
      end if;
      return new Internal_Ref_Result_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Ref_Result_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Ref_Result_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Shape_Vectors is new Langkit_Support.Vectors (Internal_Shape);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Shape_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Shape
   is
      function Absolute_Get
        (T : Internal_Shape_Array_Access; Index : Integer)
         return Internal_Shape
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Shape,
         Sequence_Type => Internal_Shape_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Shape;
   begin
      if Relative_Get (T, Index, Result) then
            Inc_Ref (Result);
         return Result;
      elsif Or_Null then
         return No_Shape;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Shape_Array_Access) return Internal_Shape_Array_Access is
      Ret : Internal_Shape_Array_Access := Create_Internal_Shape_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
         for Item of Ret.Items loop
            Inc_Ref (Item);
         end loop;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Shape_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Shape_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Shape_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
            for Item of T.Items loop
               Dec_Ref (Item);
            end loop;
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Shape_Array (Items_Count : Natural) return Internal_Shape_Array_Access
   is (if Items_Count = 0
       then No_Internal_Shape_Array_Type
       else new Internal_Shape_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Shape_Array
     (Items : Internal_Internal_Shape_Array) return Internal_Shape_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Shape_Array_Type;
      end if;
         for El of Items loop
            Inc_Ref (El);
         end loop;
      return new Internal_Shape_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Shape_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               not Equivalent (L.Items (I), R.Items (I))
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Shape_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Single_Actual_Vectors is new Langkit_Support.Vectors (Internal_Single_Actual);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Single_Actual_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Single_Actual
   is
      function Absolute_Get
        (T : Internal_Single_Actual_Array_Access; Index : Integer)
         return Internal_Single_Actual
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Single_Actual,
         Sequence_Type => Internal_Single_Actual_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Single_Actual;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return No_Single_Actual;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Single_Actual_Array_Access) return Internal_Single_Actual_Array_Access is
      Ret : Internal_Single_Actual_Array_Access := Create_Internal_Single_Actual_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Single_Actual_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Single_Actual_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Single_Actual_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Single_Actual_Array (Items_Count : Natural) return Internal_Single_Actual_Array_Access
   is (if Items_Count = 0
       then No_Internal_Single_Actual_Array_Type
       else new Internal_Single_Actual_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Single_Actual_Array
     (Items : Internal_Internal_Single_Actual_Array) return Internal_Single_Actual_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Single_Actual_Array_Type;
      end if;
      return new Internal_Single_Actual_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Single_Actual_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Single_Actual_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Substitution_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Substitution
   is
      function Absolute_Get
        (T : Internal_Substitution_Array_Access; Index : Integer)
         return Internal_Substitution
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Substitution,
         Sequence_Type => Internal_Substitution_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Substitution;
   begin
      if Relative_Get (T, Index, Result) then
            Inc_Ref (Result);
         return Result;
      elsif Or_Null then
         return No_Substitution;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Substitution_Array_Access) return Internal_Substitution_Array_Access is
      Ret : Internal_Substitution_Array_Access := Create_Internal_Substitution_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
         for Item of Ret.Items loop
            Inc_Ref (Item);
         end loop;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Substitution_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Substitution_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Substitution_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
            for Item of T.Items loop
               Dec_Ref (Item);
            end loop;
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Substitution_Array (Items_Count : Natural) return Internal_Substitution_Array_Access
   is (if Items_Count = 0
       then No_Internal_Substitution_Array_Type
       else new Internal_Substitution_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Substitution_Array
     (Items : Internal_Internal_Substitution_Array) return Internal_Substitution_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Substitution_Array_Type;
      end if;
         for El of Items loop
            Inc_Ref (El);
         end loop;
      return new Internal_Substitution_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Substitution_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               not Equivalent (L.Items (I), R.Items (I))
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Substitution_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Internal_Unit_Vectors is new Langkit_Support.Vectors (Internal_Unit);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Internal_Unit_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Internal_Unit
   is
      function Absolute_Get
        (T : Internal_Unit_Array_Access; Index : Integer)
         return Internal_Unit
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Internal_Unit,
         Sequence_Type => Internal_Unit_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Internal_Unit;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return null;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Internal_Unit_Array_Access) return Internal_Unit_Array_Access is
      Ret : Internal_Unit_Array_Access := Create_Internal_Unit_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Unit_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Internal_Unit_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Unit_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Internal_Unit_Array (Items_Count : Natural) return Internal_Unit_Array_Access
   is (if Items_Count = 0
       then No_Internal_Unit_Array_Type
       else new Internal_Unit_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Internal_Unit_Array
     (Items : Internal_Internal_Unit_Array) return Internal_Unit_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Internal_Unit_Array_Type;
      end if;
      return new Internal_Unit_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Internal_Unit_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Unit_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      function Make_Unique (A : Internal_Unit_Array_Access) return Internal_Unit_Array_Access is
         --  To keep the order of the result deterministic, this function works
         --  in two steps. First, determine for each item in A whether it
         --  duplicate a previous item. Only then, go through all items in A
         --  and add only non-duplicated ones to the result.
         Duplicates : array (A.Items'Range) of Boolean := (others => False);
         --  For each item in A, indicate whether the value is a duplicate of a
         --  previous one in the array.
         Unique_Elements : Natural;
         --  Number of unique elements in A
      begin
         --  Step 1: compute Duplicates
         declare
            package Sets is new Ada.Containers.Hashed_Sets
              (Element_Type        => Internal_Unit,
               "="                 => "=",
               Equivalent_Elements => "=",
               Hash                => Hash);
            Set : Sets.Set;
         begin
            for I in A.Items'Range loop
               declare
                  Item : Internal_Unit renames A.Items (I);
               begin
                  if Set.Contains (Item) then
                     Duplicates (I) := True;
                  else
                     Set.Insert (Item);
                  end if;
               end;
            end loop;
            Unique_Elements := Natural (Set.Length);
         end;
         --  Step 2: create the result
         declare
            Next   : Positive := 1;
            Result : constant Internal_Unit_Array_Access :=
               Create_Internal_Unit_Array (Unique_Elements);
         begin
            for I in A.Items'Range loop
               declare
                  Item : Internal_Unit renames A.Items (I);
               begin
                  if not Duplicates (I) then
                     Result.Items (Next) := Item;
                     Next := Next + 1;
                  end if;
               end;
            end loop;
            pragma Assert (Next = Result.N + 1);
            return Result;
         end;
      end Make_Unique;
      package Lexical_Env_Vectors is new Langkit_Support.Vectors (Lexical_Env);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Lexical_Env_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Lexical_Env
   is
      function Absolute_Get
        (T : Lexical_Env_Array_Access; Index : Integer)
         return Lexical_Env
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Lexical_Env,
         Sequence_Type => Lexical_Env_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Lexical_Env;
   begin
      if Relative_Get (T, Index, Result) then
            Inc_Ref (Result);
         return Result;
      elsif Or_Null then
         return Empty_Env;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Lexical_Env_Array_Access) return Lexical_Env_Array_Access is
      Ret : Lexical_Env_Array_Access := Create_Lexical_Env_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
         for Item of Ret.Items loop
            Inc_Ref (Item);
         end loop;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Lexical_Env_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Lexical_Env_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Lexical_Env_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
            for Item of T.Items loop
               Dec_Ref (Item);
            end loop;
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Lexical_Env_Array (Items_Count : Natural) return Lexical_Env_Array_Access
   is (if Items_Count = 0
       then No_Lexical_Env_Array_Type
       else new Lexical_Env_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Lexical_Env_Array
     (Items : Internal_Lexical_Env_Array) return Lexical_Env_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Lexical_Env_Array_Type;
      end if;
         for El of Items loop
            Inc_Ref (El);
         end loop;
      return new Lexical_Env_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Lexical_Env_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               not Equivalent (L.Items (I), R.Items (I))
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Lexical_Env_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Logic_Equation_Vectors is new Langkit_Support.Vectors (Logic_Equation);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Logic_Equation_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Logic_Equation
   is
      function Absolute_Get
        (T : Logic_Equation_Array_Access; Index : Integer)
         return Logic_Equation
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Logic_Equation,
         Sequence_Type => Logic_Equation_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Logic_Equation;
   begin
      if Relative_Get (T, Index, Result) then
            Inc_Ref (Result);
         return Result;
      elsif Or_Null then
         return Null_Logic_Equation;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Logic_Equation_Array_Access) return Logic_Equation_Array_Access is
      Ret : Logic_Equation_Array_Access := Create_Logic_Equation_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
         for Item of Ret.Items loop
            Inc_Ref (Item);
         end loop;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Logic_Equation_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Logic_Equation_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Logic_Equation_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
            for Item of T.Items loop
               Dec_Ref (Item);
            end loop;
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Logic_Equation_Array (Items_Count : Natural) return Logic_Equation_Array_Access
   is (if Items_Count = 0
       then No_Logic_Equation_Array_Type
       else new Logic_Equation_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Logic_Equation_Array
     (Items : Internal_Logic_Equation_Array) return Logic_Equation_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Logic_Equation_Array_Type;
      end if;
         for El of Items loop
            Inc_Ref (El);
         end loop;
      return new Logic_Equation_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Logic_Equation_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Logic_Equation_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package String_Type_Vectors is new Langkit_Support.Vectors (String_Type);
   ---------
   -- Get --
   ---------
   function Get
     (T       : String_Type_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return String_Type
   is
      function Absolute_Get
        (T : String_Type_Array_Access; Index : Integer)
         return String_Type
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => String_Type,
         Sequence_Type => String_Type_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : String_Type;
   begin
      if Relative_Get (T, Index, Result) then
            Inc_Ref (Result);
         return Result;
      elsif Or_Null then
         return Empty_String;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : String_Type_Array_Access) return String_Type_Array_Access is
      Ret : String_Type_Array_Access := Create_String_Type_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
         for Item of Ret.Items loop
            Inc_Ref (Item);
         end loop;
      return Ret;
   end Concat;
      ------------------
      -- Join_Strings --
      ------------------
      function Join_Strings
        (Separator : String_Type;
         Strings   : String_Type_Array_Access) return String_Type
      is
         Separator_Length : constant Natural := Separator.Length;
         Length           : Natural := 0;
         First            : Boolean;
      begin
         --  First, compute the length of the result: the sum of all string
         --  lengths in Strings.
         First := True;
         for S of Strings.Items loop
            if First then
               First := False;
            else
               Length := Length + Separator_Length;
            end if;
            Length := Length + S.Length;
         end loop;
         --  Create the result string with the correct length. Do not use our
         --  constructor and initialize the content in-place, to avoid extra
         --  copies.
         return Result : constant String_Type :=
            new String_Record (Length)
         do
            Result.Ref_Count := 1;
            --  Now copy the content of all strings into the result
            declare
               Last : Natural := 0;
            begin
               First := True;
               for S of Strings.Items loop
                  if First then
                     First := False;
                  else
                     Result.Content (Last + 1 .. Last + Separator_Length) :=
                        Separator.Content;
                     Last := Last + Separator_Length;
                  end if;
                  Result.Content (Last + 1 .. Last + S.Length) := S.Content;
                  Last := Last + S.Length;
               end loop;
            end;
         end return;
      end Join_Strings;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : String_Type_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : String_Type_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out String_Type_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
            for Item of T.Items loop
               Dec_Ref (Item);
            end loop;
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_String_Type_Array (Items_Count : Natural) return String_Type_Array_Access
   is (if Items_Count = 0
       then No_String_Type_Array_Type
       else new String_Type_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_String_Type_Array
     (Items : Internal_String_Type_Array) return String_Type_Array_Access is
   begin
      if Items'Length = 0 then
         return No_String_Type_Array_Type;
      end if;
         for El of Items loop
            Inc_Ref (El);
         end loop;
      return new String_Type_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : String_Type_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               not Equivalent (L.Items (I), R.Items (I))
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : String_Type_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
      package Symbol_Type_Vectors is new Langkit_Support.Vectors (Symbol_Type);
   ---------
   -- Get --
   ---------
   function Get
     (T       : Symbol_Type_Array_Access;
      Index   : Integer;
      Or_Null : Boolean := False) return Symbol_Type
   is
      function Absolute_Get
        (T : Symbol_Type_Array_Access; Index : Integer)
         return Symbol_Type
      is
        (T.Items (Index + 1)); --  T.Items is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Symbol_Type,
         Sequence_Type => Symbol_Type_Array_Access,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Symbol_Type;
   begin
      if Relative_Get (T, Index, Result) then
         return Result;
      elsif Or_Null then
         return null;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   ------------
   -- Concat --
   ------------
   function Concat (L, R : Symbol_Type_Array_Access) return Symbol_Type_Array_Access is
      Ret : Symbol_Type_Array_Access := Create_Symbol_Type_Array (Length (L) + Length (R));
   begin
      Ret.Items := L.Items & R.Items;
      return Ret;
   end Concat;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Symbol_Type_Array_Access) is
   begin
      if T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   ------------
   -- Length --
   ------------
   function Length (T : Symbol_Type_Array_Access) return Natural is (T.N);
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Symbol_Type_Array_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
   function Create_Symbol_Type_Array (Items_Count : Natural) return Symbol_Type_Array_Access
   is (if Items_Count = 0
       then No_Symbol_Type_Array_Type
       else new Symbol_Type_Array_Record'(N => Items_Count, Ref_Count => 1, Items => <>));
   function Create_Symbol_Type_Array
     (Items : Internal_Symbol_Type_Array) return Symbol_Type_Array_Access is
   begin
      if Items'Length = 0 then
         return No_Symbol_Type_Array_Type;
      end if;
      return new Symbol_Type_Array_Record'
        (N => Items'Length, Ref_Count => 1, Items => Items);
   end;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (L, R : Symbol_Type_Array_Access) return Boolean is
   begin
      if L.N /= R.N then
         return False;
      end if;
      for I in L.Items'Range loop
         if
               L.Items (I) /= R.Items (I)
         then
            return False;
         end if;
      end loop;
      return True;
   end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Symbol_Type_Array_Access) return String is
         Result : Unbounded_String;
      begin
         Append (Result, "[");
         for I in A.Items'Range loop
            if I > A.Items'First then
               Append (Result, ", ");
            end if;
            Append (Result, Trace_Image (A.Items (I)));
         end loop;
         Append (Result, "]");
         return To_String (Result);
      end Trace_Image;
   ----------
   -- Next --
   ----------
   function Next
     (T       : Bare_Ada_Node_Iterator_Access;
      Element : out Bare_Ada_Node) return Boolean is
   begin
      if T = null then
         raise Property_Error with "null access dereference";
      end if;
      Check_Safety_Net (T.Safety_Net);
      if T.Index > T.Elements.Items'Last then
         return False;
      else
         Element := T.Elements.Items (T.Index);
         T.Index := T.Index + 1;
         return True;
      end if;
   end Next;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Bare_Ada_Node_Iterator_Access) is
   begin
      if T /= null and then T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Bare_Ada_Node_Iterator_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Dec_Ref (T.Elements);
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Bare_Ada_Node_Iterator_Access) return String is
      begin
         return "<Iterator of AdaNode, index="
                & A.Index'Image & ">";
      end Trace_Image;
   ----------
   -- Next --
   ----------
   function Next
     (T       : Internal_Completion_Item_Iterator_Access;
      Element : out Internal_Completion_Item) return Boolean is
   begin
      if T = null then
         raise Property_Error with "null access dereference";
      end if;
      Check_Safety_Net (T.Safety_Net);
      if T.Index > T.Elements.Items'Last then
         return False;
      else
         Element := T.Elements.Items (T.Index);
         T.Index := T.Index + 1;
         return True;
      end if;
   end Next;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Completion_Item_Iterator_Access) is
   begin
      if T /= null and then T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Completion_Item_Iterator_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Dec_Ref (T.Elements);
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Completion_Item_Iterator_Access) return String is
      begin
         return "<Iterator of CompletionItem, index="
                & A.Index'Image & ">";
      end Trace_Image;
   ----------
   -- Next --
   ----------
   function Next
     (T       : Internal_Entity_Iterator_Access;
      Element : out Internal_Entity) return Boolean is
   begin
      if T = null then
         raise Property_Error with "null access dereference";
      end if;
      Check_Safety_Net (T.Safety_Net);
      if T.Index > T.Elements.Items'Last then
         return False;
      else
         Element := T.Elements.Items (T.Index);
         T.Index := T.Index + 1;
         return True;
      end if;
   end Next;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Entity_Iterator_Access) is
   begin
      if T /= null and then T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Entity_Iterator_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Dec_Ref (T.Elements);
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Entity_Iterator_Access) return String is
      begin
         return "<Iterator of AdaNode.entity, index="
                & A.Index'Image & ">";
      end Trace_Image;
   ----------
   -- Next --
   ----------
   function Next
     (T       : Internal_Inner_Env_Assoc_Iterator_Access;
      Element : out Internal_Inner_Env_Assoc) return Boolean is
   begin
      if T = null then
         raise Property_Error with "null access dereference";
      end if;
      Check_Safety_Net (T.Safety_Net);
      if T.Index > T.Elements.Items'Last then
         return False;
      else
         Element := T.Elements.Items (T.Index);
         T.Index := T.Index + 1;
         return True;
      end if;
   end Next;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (T : Internal_Inner_Env_Assoc_Iterator_Access) is
   begin
      if T /= null and then T.Ref_Count >= 0 then
         T.Ref_Count := T.Ref_Count + 1;
      end if;
   end Inc_Ref;
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (T : in out Internal_Inner_Env_Assoc_Iterator_Access) is
   begin
      if T = null or else T.Ref_Count < 0 then
         return;
      end if;
      if T.Ref_Count = 1 then
         Dec_Ref (T.Elements);
         Free (T);
      else
         T.Ref_Count := T.Ref_Count - 1;
         T := null;
      end if;
   end Dec_Ref;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (A : Internal_Inner_Env_Assoc_Iterator_Access) return String is
      begin
         return "<Iterator of InnerEnvAssoc, index="
                & A.Index'Image & ">";
      end Trace_Image;
   ---------
   -- "<" --
   ---------
   function "<" (Left, Right : Internal_Unit) return Boolean is
   begin
      return Left.Filename < Right.Filename;
   end "<";
function Hash (Key : Mmz_Key_Item) return Hash_Type;
function Equivalent (L, R : Mmz_Key_Item) return Boolean;
procedure Destroy (Key : in out Mmz_Key_Array_Access);
----------------
-- Equivalent --
----------------
function Equivalent (L, R : Mmz_Key_Item) return Boolean is
begin
   if L.Kind /= R.Kind then
      return False;
   end if;
   case L.Kind is
         when Mmz_Bare_Access_Def =>
               return L.As_Bare_Access_Def = R.As_Bare_Access_Def;
         when Mmz_Bare_Ada_Node =>
               return L.As_Bare_Ada_Node = R.As_Bare_Ada_Node;
         when Mmz_Bare_Anonymous_Expr_Decl =>
               return L.As_Bare_Anonymous_Expr_Decl = R.As_Bare_Anonymous_Expr_Decl;
         when Mmz_Bare_Array_Type_Def =>
               return L.As_Bare_Array_Type_Def = R.As_Bare_Array_Type_Def;
         when Mmz_Bare_Assoc_List =>
               return L.As_Bare_Assoc_List = R.As_Bare_Assoc_List;
         when Mmz_Bare_Base_Aggregate =>
               return L.As_Bare_Base_Aggregate = R.As_Bare_Base_Aggregate;
         when Mmz_Bare_Base_Id =>
               return L.As_Bare_Base_Id = R.As_Bare_Base_Id;
         when Mmz_Bare_Base_Subp_Spec =>
               return L.As_Bare_Base_Subp_Spec = R.As_Bare_Base_Subp_Spec;
         when Mmz_Bare_Base_Type_Decl =>
               return L.As_Bare_Base_Type_Decl = R.As_Bare_Base_Type_Decl;
         when Mmz_Bare_Basic_Decl =>
               return L.As_Bare_Basic_Decl = R.As_Bare_Basic_Decl;
         when Mmz_Bare_Body_Node =>
               return L.As_Bare_Body_Node = R.As_Bare_Body_Node;
         when Mmz_Bare_Classwide_Type_Decl =>
               return L.As_Bare_Classwide_Type_Decl = R.As_Bare_Classwide_Type_Decl;
         when Mmz_Bare_Compilation_Unit =>
               return L.As_Bare_Compilation_Unit = R.As_Bare_Compilation_Unit;
         when Mmz_Bare_Decimal_Fixed_Point_Def =>
               return L.As_Bare_Decimal_Fixed_Point_Def = R.As_Bare_Decimal_Fixed_Point_Def;
         when Mmz_Bare_Defining_Name =>
               return L.As_Bare_Defining_Name = R.As_Bare_Defining_Name;
         when Mmz_Bare_End_Name =>
               return L.As_Bare_End_Name = R.As_Bare_End_Name;
         when Mmz_Bare_Enum_Literal_Decl =>
               return L.As_Bare_Enum_Literal_Decl = R.As_Bare_Enum_Literal_Decl;
         when Mmz_Bare_Enum_Type_Def =>
               return L.As_Bare_Enum_Type_Def = R.As_Bare_Enum_Type_Def;
         when Mmz_Bare_For_Loop_Spec =>
               return L.As_Bare_For_Loop_Spec = R.As_Bare_For_Loop_Spec;
         when Mmz_Bare_For_Loop_Var_Decl =>
               return L.As_Bare_For_Loop_Var_Decl = R.As_Bare_For_Loop_Var_Decl;
         when Mmz_Bare_Formal_Discrete_Type_Def =>
               return L.As_Bare_Formal_Discrete_Type_Def = R.As_Bare_Formal_Discrete_Type_Def;
         when Mmz_Bare_Generic_Instantiation =>
               return L.As_Bare_Generic_Instantiation = R.As_Bare_Generic_Instantiation;
         when Mmz_Bare_Mod_Int_Type_Def =>
               return L.As_Bare_Mod_Int_Type_Def = R.As_Bare_Mod_Int_Type_Def;
         when Mmz_Bare_Name =>
               return L.As_Bare_Name = R.As_Bare_Name;
         when Mmz_Bare_Ordinary_Fixed_Point_Def =>
               return L.As_Bare_Ordinary_Fixed_Point_Def = R.As_Bare_Ordinary_Fixed_Point_Def;
         when Mmz_Bare_Private_Type_Def =>
               return L.As_Bare_Private_Type_Def = R.As_Bare_Private_Type_Def;
         when Mmz_Bare_Real_Type_Def =>
               return L.As_Bare_Real_Type_Def = R.As_Bare_Real_Type_Def;
         when Mmz_Bare_Record_Type_Def =>
               return L.As_Bare_Record_Type_Def = R.As_Bare_Record_Type_Def;
         when Mmz_Bare_Signed_Int_Type_Def =>
               return L.As_Bare_Signed_Int_Type_Def = R.As_Bare_Signed_Int_Type_Def;
         when Mmz_Bare_Type_Decl =>
               return L.As_Bare_Type_Decl = R.As_Bare_Type_Decl;
         when Mmz_Bare_Type_Expr =>
               return L.As_Bare_Type_Expr = R.As_Bare_Type_Expr;
         when Mmz_Boolean =>
               return L.As_Boolean = R.As_Boolean;
         when Mmz_Env_Rebindings =>
               return L.As_Env_Rebindings = R.As_Env_Rebindings;
         when Mmz_Integer =>
               return L.As_Integer = R.As_Integer;
         when Mmz_Internal_Entity_Base_Type_Decl_Array_Access =>
               return Equivalent (L.As_Internal_Entity_Base_Type_Decl_Array_Access, R.As_Internal_Entity_Base_Type_Decl_Array_Access);
         when Mmz_Internal_Entity_Info =>
               return L.As_Internal_Entity_Info = R.As_Internal_Entity_Info;
         when Mmz_Internal_Metadata =>
               return L.As_Internal_Metadata = R.As_Internal_Metadata;
         when Mmz_Lexical_Env =>
               return Equivalent (L.As_Lexical_Env, R.As_Lexical_Env);
         when Mmz_Symbol_Type =>
               return L.As_Symbol_Type = R.As_Symbol_Type;
   end case;
end Equivalent;
----------
-- Hash --
----------
function Hash (Key : Mmz_Key_Item) return Hash_Type is
begin
   case Key.Kind is
         when Mmz_Bare_Access_Def =>
            return Hash (Key.As_Bare_Access_Def);
         when Mmz_Bare_Ada_Node =>
            return Hash (Key.As_Bare_Ada_Node);
         when Mmz_Bare_Anonymous_Expr_Decl =>
            return Hash (Key.As_Bare_Anonymous_Expr_Decl);
         when Mmz_Bare_Array_Type_Def =>
            return Hash (Key.As_Bare_Array_Type_Def);
         when Mmz_Bare_Assoc_List =>
            return Hash (Key.As_Bare_Assoc_List);
         when Mmz_Bare_Base_Aggregate =>
            return Hash (Key.As_Bare_Base_Aggregate);
         when Mmz_Bare_Base_Id =>
            return Hash (Key.As_Bare_Base_Id);
         when Mmz_Bare_Base_Subp_Spec =>
            return Hash (Key.As_Bare_Base_Subp_Spec);
         when Mmz_Bare_Base_Type_Decl =>
            return Hash (Key.As_Bare_Base_Type_Decl);
         when Mmz_Bare_Basic_Decl =>
            return Hash (Key.As_Bare_Basic_Decl);
         when Mmz_Bare_Body_Node =>
            return Hash (Key.As_Bare_Body_Node);
         when Mmz_Bare_Classwide_Type_Decl =>
            return Hash (Key.As_Bare_Classwide_Type_Decl);
         when Mmz_Bare_Compilation_Unit =>
            return Hash (Key.As_Bare_Compilation_Unit);
         when Mmz_Bare_Decimal_Fixed_Point_Def =>
            return Hash (Key.As_Bare_Decimal_Fixed_Point_Def);
         when Mmz_Bare_Defining_Name =>
            return Hash (Key.As_Bare_Defining_Name);
         when Mmz_Bare_End_Name =>
            return Hash (Key.As_Bare_End_Name);
         when Mmz_Bare_Enum_Literal_Decl =>
            return Hash (Key.As_Bare_Enum_Literal_Decl);
         when Mmz_Bare_Enum_Type_Def =>
            return Hash (Key.As_Bare_Enum_Type_Def);
         when Mmz_Bare_For_Loop_Spec =>
            return Hash (Key.As_Bare_For_Loop_Spec);
         when Mmz_Bare_For_Loop_Var_Decl =>
            return Hash (Key.As_Bare_For_Loop_Var_Decl);
         when Mmz_Bare_Formal_Discrete_Type_Def =>
            return Hash (Key.As_Bare_Formal_Discrete_Type_Def);
         when Mmz_Bare_Generic_Instantiation =>
            return Hash (Key.As_Bare_Generic_Instantiation);
         when Mmz_Bare_Mod_Int_Type_Def =>
            return Hash (Key.As_Bare_Mod_Int_Type_Def);
         when Mmz_Bare_Name =>
            return Hash (Key.As_Bare_Name);
         when Mmz_Bare_Ordinary_Fixed_Point_Def =>
            return Hash (Key.As_Bare_Ordinary_Fixed_Point_Def);
         when Mmz_Bare_Private_Type_Def =>
            return Hash (Key.As_Bare_Private_Type_Def);
         when Mmz_Bare_Real_Type_Def =>
            return Hash (Key.As_Bare_Real_Type_Def);
         when Mmz_Bare_Record_Type_Def =>
            return Hash (Key.As_Bare_Record_Type_Def);
         when Mmz_Bare_Signed_Int_Type_Def =>
            return Hash (Key.As_Bare_Signed_Int_Type_Def);
         when Mmz_Bare_Type_Decl =>
            return Hash (Key.As_Bare_Type_Decl);
         when Mmz_Bare_Type_Expr =>
            return Hash (Key.As_Bare_Type_Expr);
         when Mmz_Boolean =>
            return Hash (Key.As_Boolean);
         when Mmz_Env_Rebindings =>
            return Hash (Key.As_Env_Rebindings);
         when Mmz_Integer =>
            return Hash (Key.As_Integer);
         when Mmz_Internal_Entity_Base_Type_Decl_Array_Access =>
            return Hash (Key.As_Internal_Entity_Base_Type_Decl_Array_Access);
         when Mmz_Internal_Entity_Info =>
            return Hash (Key.As_Internal_Entity_Info);
         when Mmz_Internal_Metadata =>
            return Hash (Key.As_Internal_Metadata);
         when Mmz_Lexical_Env =>
            return Hash (Key.As_Lexical_Env);
         when Mmz_Symbol_Type =>
            return Hash (Key.As_Symbol_Type);
   end case;
end Hash;
----------
-- Hash --
----------
function Hash (Key : Mmz_Key) return Hash_Type is
   Result : Hash_Type := Mmz_Property'Pos (Key.Property);
begin
   for K of Key.Items.all loop
      Result := Combine (Result, Hash (K));
   end loop;
   return Result;
end Hash;
----------------
-- Equivalent --
----------------
function Equivalent (L, R : Mmz_Key) return Boolean is
   L_Items : Mmz_Key_Array renames L.Items.all;
   R_Items : Mmz_Key_Array renames R.Items.all;
begin
   if L.Property /= R.Property or else L_Items'Length /= R_Items'Length then
      return False;
   end if;
   for I in L_Items'Range loop
      if not Equivalent (L_Items (I), R_Items (I)) then
         return False;
      end if;
   end loop;
   return True;
end Equivalent;
-------------
-- Destroy --
-------------
procedure Destroy (Map : in out Memoization_Maps.Map) is
   use Memoization_Maps;
   --  We need keys and values to be valid when clearing the memoization map,
   --  but on the other hand we need to free keys and values as well. To
   --  achieve both goals, we first copy key and values into arrays, then we
   --  clear the map, and then we free keys/values in the arrays. Allocate both
   --  arrays on the heap to avoid stack overflow, as they can be quite big.
   Length : constant Natural := Natural (Map.Length);
   type Key_Array is array (1 .. Length) of Mmz_Key_Array_Access;
   type Key_Array_Access is access Key_Array;
   procedure Free is new Ada.Unchecked_Deallocation
     (Key_Array, Key_Array_Access);
   type Value_Array is array (1 .. Length) of Mmz_Value;
   type Value_Array_Access is access Value_Array;
   procedure Free is new Ada.Unchecked_Deallocation
     (Value_Array, Value_Array_Access);
   Keys   : Key_Array_Access := new Key_Array;
   Values : Value_Array_Access := new Value_Array;
   I      : Positive := 1;
begin
   for Cur in Map.Iterate loop
      Keys (I) := Key (Cur).Items;
      Values (I) := Element (Cur);
      I := I + 1;
   end loop;
   Map.Clear;
   for K_Array of Keys.all loop
      Destroy (K_Array);
   end loop;
      for V of Values.all loop
         case V.Kind is
               when Mmz_Internal_Entity_Array_Access =>
                  Dec_Ref (V.As_Internal_Entity_Array_Access);
               when Mmz_Internal_Entity_Base_Formal_Param_Decl_Array_Access =>
                  Dec_Ref (V.As_Internal_Entity_Base_Formal_Param_Decl_Array_Access);
               when Mmz_Internal_Entity_Base_Type_Decl_Array_Access =>
                  Dec_Ref (V.As_Internal_Entity_Base_Type_Decl_Array_Access);
               when Mmz_Internal_Entity_Basic_Decl_Array_Access =>
                  Dec_Ref (V.As_Internal_Entity_Basic_Decl_Array_Access);
               when Mmz_Internal_Entity_Compilation_Unit_Array_Access =>
                  Dec_Ref (V.As_Internal_Entity_Compilation_Unit_Array_Access);
               when Mmz_Internal_Env_Assoc =>
                  Dec_Ref (V.As_Internal_Env_Assoc);
               when Mmz_Internal_Env_Assoc_Array_Access =>
                  Dec_Ref (V.As_Internal_Env_Assoc_Array_Access);
               when Mmz_Internal_Inner_Env_Assoc_Array_Access =>
                  Dec_Ref (V.As_Internal_Inner_Env_Assoc_Array_Access);
               when Mmz_Internal_Param_Match_Array_Access =>
                  Dec_Ref (V.As_Internal_Param_Match_Array_Access);
               when Mmz_Internal_Single_Actual_Array_Access =>
                  Dec_Ref (V.As_Internal_Single_Actual_Array_Access);
               when Mmz_Lexical_Env =>
                  Dec_Ref (V.As_Lexical_Env);
            when others => null;
         end case;
      end loop;
   Free (Keys);
   Free (Values);
end Destroy;
-------------
-- Destroy --
-------------
procedure Destroy (Key : in out Mmz_Key_Array_Access) is
   procedure Free is new Ada.Unchecked_Deallocation
     (Mmz_Key_Array, Mmz_Key_Array_Access);
begin
      for K of Key.all loop
         case K.Kind is
               when Mmz_Internal_Entity_Base_Type_Decl_Array_Access =>
                  Dec_Ref (K.As_Internal_Entity_Base_Type_Decl_Array_Access);
               when Mmz_Lexical_Env =>
                  Dec_Ref (K.As_Lexical_Env);
            when others => null;
         end case;
      end loop;
   Free (Key);
end Destroy;
-------------------------
-- Find_Memoized_Value --
-------------------------
function Find_Memoized_Value
  (Unit       : Internal_Unit;
   Handle     : out Memoization_Handle;
   Value      : out Mmz_Value;
   Create_Key : access function return Mmz_Key) return Boolean
is
   Inserted : Boolean;
begin
   --  Make sure that we don't lookup stale caches
   Reset_Caches (Unit);
   --  Initialize handle: create the key and create a cursor pointing to an
   --  existing entry.
   Handle.Key := Create_Key.all;
   Handle.Cache_Version := Unit.Cache_Version;
   Value := (Kind => Mmz_Evaluating);
   Unit.Memoization_Map.Insert (Handle.Key, Value, Handle.Cur, Inserted);
   --  No existing entry yet? The above just created one. Otherwise, destroy
   --  our key and reuse the existing entry's.
   if not Inserted then
      Destroy (Handle.Key.Items);
      Handle.Key := Memoization_Maps.Key (Handle.Cur);
      Value := Memoization_Maps.Element (Handle.Cur);
   end if;
   return not Inserted;
end Find_Memoized_Value;
------------------------
-- Add_Memoized_Value --
------------------------
procedure Add_Memoized_Value
  (Unit   : Internal_Unit;
   Handle : in out Memoization_Handle;
   Value  : Mmz_Value;
   Stored : out Boolean) is
begin
   --  If Handle was created using a memoization map that has been since then
   --  reset, do nothing: the result can be partly stale due to the event that
   --  triggered the memoization tables reset.
   Stored := Unit.Cache_Version <= Handle.Cache_Version;
   if Stored then
      Unit.Memoization_Map.Replace_Element (Handle.Cur, Value);
   end if;
end Add_Memoized_Value;
   -------------------
   -- Solve_Wrapper --
   -------------------
   function Solve_Wrapper
     (R            : Solver.Relation;
      Context_Node : Bare_Ada_Node) return Boolean is
   begin
      if Context_Node /= null and then Langkit_Support.Adalog.Debug.Debug then
         Assign_Names_To_Logic_Vars (Context_Node);
      end if;
      begin
         return Solver.Solve_First
           (R, Timeout => Context_Node.Unit.Context.Logic_Resolution_Timeout);
      exception
         when Langkit_Support.Adalog.Early_Binding_Error =>
            raise Property_Error with "invalid equation for logic resolution";
         when Langkit_Support.Adalog.Timeout_Error =>
            raise Property_Error with "logic resolution timed out";
      end;
   end Solve_Wrapper;
   -------------
   -- Destroy --
   -------------
   procedure Destroy (Env : in out Lexical_Env_Access) is
      Mutable_Env : Lexical_Env :=
        (Wrap (Env), 0, Env.Kind, No_Generic_Unit, 0);
   begin
      Destroy (Mutable_Env);
      Env := null;
   end Destroy;
   ----------------
   -- Initialize --
   ----------------
   procedure Initialize
     (Self              : Bare_Ada_Node;
      Kind              : Ada_Node_Kind_Type;
      Unit              : Internal_Unit;
      Token_Start_Index : Token_Index;
      Token_End_Index   : Token_Index;
      Parent            : Bare_Ada_Node := null;
      Self_Env          : Lexical_Env := AST_Envs.Empty_Env) is
   begin
      pragma Unreferenced (Kind);
      Self.Parent := Parent;
      Self.Unit := Unit;
      Self.Token_Start_Index := Token_Start_Index;
      Self.Token_End_Index := Token_End_Index;
      Self.Self_Env := Self_Env;
      Self.Last_Attempted_Child := -1;
   end Initialize;
   --------------------
   -- Use_Direct_Env --
   --------------------
   procedure Use_Direct_Env (State : in out PLE_Node_State; Env : Lexical_Env)
   is
   begin
      State.Current_Env := Env;
      State.Current_NED := null;
   end Use_Direct_Env;
   -------------------
   -- Use_Named_Env --
   -------------------
   procedure Use_Named_Env
     (State   : in out PLE_Node_State;
      Context : Internal_Context;
      Name    : Symbol_Type) is
   begin
      State.Current_NED := Get_Named_Env_Descriptor (Context, Name);
      State.Current_Env := State.Current_NED.Env_With_Precedence;
   end Use_Named_Env;
   ---------------------
   -- Set_Initial_Env --
   ---------------------
   procedure Set_Initial_Env
     (Self         : Bare_Ada_Node;
      State        : in out PLE_Node_State;
      Env          : Internal_Designated_Env;
      DSL_Location : String) is
   begin
      case Env.Kind is
         when None =>
            Use_Direct_Env (State, Empty_Env);
         when Current_Env =>
            null;
         when Named_Env =>
            Use_Named_Env (State, Self.Unit.Context, Env.Env_Name);
         when Direct_Env =>
            --  Sanitize this environment value: make sure it's a non-foreign
            --  and primary environment.
            if Env.Direct_Env.Kind /= Static_Primary then
               raise Property_Error with
                  "Cannot set an env that is not static-primary as the"
                  & " initial env";
            elsif Is_Foreign_Strict (Env.Direct_Env, Self) then
               raise Property_Error with
                  "unsound foreign environment in SetInitialEnv ("
                  & DSL_Location & ")";
            end if;
            Use_Direct_Env (State, Env.Direct_Env);
      end case;
   end Set_Initial_Env;
   ----------------
   -- Add_To_Env --
   ----------------
   procedure Add_To_Env
     (Self         : Bare_Ada_Node;
      State        : PLE_Node_State;
      Key          : Symbol_Type;
      Value        : Bare_Ada_Node;
      Md           : Internal_Metadata;
      Resolver     : Entity_Resolver;
      Dest_Env     : Internal_Designated_Env;
      DSL_Location : String)
   is
      Context    : constant Internal_Context := Self.Unit.Context;
      Root_Scope : Lexical_Env renames Context.Root_Scope;
      --  Shortcuts
      Actual_Dest_Env : Lexical_Env;
      Dest_NED        : Named_Env_Descriptor_Access;
      --  Description for the destination environment
   begin
      --  Skip the env addition if explicitly requested
      if Key = null
         or else Value = null
         or else (case Dest_Env.Kind is
                  when None        => True,
                  when Current_Env => False,
                  when Named_Env   => Dest_Env.Env_Name = null,
                  when Direct_Env  => Dest_Env.Direct_Env = Empty_Env)
      then
         return;
      end if;
      if Value.Unit /= Self.Unit then
         raise Property_Error with "Cannot add_to_env an AST node that comes"
                                   & " from another analysis unit";
      end if;
         --  Make sure metadata does not contain any foreign node
         if (Md.Primitive /= null and then Md.Primitive.Unit /= Self.Unit) or else (Md.Primitive_Real_Type /= null and then Md.Primitive_Real_Type.Unit /= Self.Unit)
         then
            raise Property_Error
               with "Cannot add metadata that contains foreign nodes";
         end if;
      --  Then determine the destination environment
      case Dest_Env.Kind is
         when None =>
            raise Program_Error with "unreachable code";
         when Current_Env =>
            --  Just use the current environment
            Dest_NED := State.Current_NED;
            Actual_Dest_Env := State.Current_Env;
         when Named_Env =>
            --  There is an environment name: just lookup the corresponding
            --  NED/env.
            Dest_NED := Get_Named_Env_Descriptor (Context, Dest_Env.Env_Name);
            Actual_Dest_Env := Dest_NED.Env_With_Precedence;
         when Direct_Env =>
            --  There is an explicit destination environment
            Dest_NED := null;
            Actual_Dest_Env := Dest_Env.Direct_Env;
      end case;
      --  Sanitize it
      if Actual_Dest_Env.Kind /= Static_Primary then
         raise Property_Error with
            "Cannot add elements to a lexical env that is not static-primary";
      elsif
         --  Since lexical envs need to sort the foreign nodes they contain,
         --  and that the total order on nodes is not defined for synthetic
         --  nodes, it is not possible to add a synthetic node to a foreign
         --  lexical environment.
         --
         --  This reasoning applies to environments that belong to foreign
         --  units, but also to the root environment.
         Is_Foreign (Actual_Dest_Env, Self) and then Is_Synthetic (Value)
      then
         raise Property_Error with
            "Cannot add a synthetic node to a lexical env from another"
            & " analysis unit";
      elsif
         --  Reject direct references to foreign destination environments.
         --
         --  This is an attempt at identifying uses of the unsound relocation
         --  mechanism (as opposed to named environments), so this applies to
         --  all foreign environments (root scope included).
         DSL_Location'Length > 0
         and then Dest_Env.Kind = Direct_Env
         and then Is_Foreign_Strict (Actual_Dest_Env, Self)
      then
         raise Property_Error with
            "unsound foreign environment in AddToEnv (" & DSL_Location & ")";
      end if;
      --  Now that everything is sanitized, we can proceed with the actual
      --  key/value pair addition. Note that this does nothing if
      --  Actual_Dest_Env ended up empty.
      Add (Actual_Dest_Env, Key, Value, Md, Resolver);
      --  If we're adding the element to an environment by env name, we must
      --  register this association in two places: in the target named env
      --  entry, and in Value's unit.
      if Dest_NED /= null then
         declare
            use NED_Assoc_Maps;
            FN    : Map renames Dest_NED.Foreign_Nodes;
            Dummy : Boolean;
            Cur   : Cursor;
         begin
            FN.Insert (Key      => Key,
                       New_Item => Internal_Map_Node_Vectors.Empty_Vector,
                       Position => Cur,
                       Inserted => Dummy);
            declare
               V : Internal_Map_Node_Vectors.Vector renames
                  FN.Reference (Cur);
            begin
               V.Append ((Value, Md, Resolver));
            end;
         end;
         Value.Unit.Exiled_Entries_In_NED.Append ((Dest_NED, Key, Value));
      --  Otherwise, if we're adding the element to an environment that belongs
      --  to a different unit, or to the root scope, then...
      elsif Is_Foreign_Not_Empty (Actual_Dest_Env, Self) then
         --  Add the Key/Value association to the list of entries contained in
         --  other units, so we can remove them when reparsing Value's unit.
         Value.Unit.Exiled_Entries.Append ((Actual_Dest_Env, Key, Value));
         if Actual_Dest_Env /= Root_Scope then
            --  Add Val to the list of foreign nodes that Actual_Dest_Env's
            --  unit contains, so that when that unit is reparsed, we can call
            --  Add_To_Env again on those nodes.
            Convert_Unit (Actual_Dest_Env.Owner).Foreign_Nodes.Append
              ((Value, Self.Unit));
         end if;
      end if;
   end Add_To_Env;
   -------------
   -- Ref_Env --
   -------------
   procedure Ref_Env
     (Self                : Bare_Ada_Node;
      Dest_Env            : Lexical_Env;
      Ref_Env_Nodes       : in out Bare_Ada_Node_Array_Access;
      Resolver            : Lexical_Env_Resolver;
      Kind                : Ref_Kind;
      Cats                : Ref_Categories;
      Shed_Rebindings     : Boolean) is
   begin
      for N of Ref_Env_Nodes.Items loop
         if N /= null then
            if N.Unit /= Self.Unit then
               raise Property_Error with
                  "attempt to add a referenced environment to a foreign unit";
            end if;
            Reference (Dest_Env, N, Resolver, Kind, Cats, Shed_Rebindings);
         end if;
      end loop;
      Dec_Ref (Ref_Env_Nodes);
   end Ref_Env;
   -------------
   -- Add_Env --
   -------------
   procedure Add_Env
     (Self              : Bare_Ada_Node;
      State             : in out PLE_Node_State;
      No_Parent         : Boolean;
      Transitive_Parent : Boolean;
      Names             : in out Symbol_Type_Array_Access)
   is
      Parent_From_Name : constant Boolean := State.Current_NED /= null;
      --  Does the parent environment comes from a named environment lookup?
      --  Determine the parent of this new environment:
      --
      --  (1) no parent if requested;
      --  (2) the current environment as the static parent if it comes from a
      --      named env lookup or if it is not foreign (or is the empty/root
      --      environment).
      Parent : constant Lexical_Env :=
        (if No_Parent
         then Null_Lexical_Env
         else State.Current_Env);
   begin
      --  Create the environment itself
      Self.Self_Env := Create_Static_Lexical_Env
        (Parent            => Parent,
         Node              => Self,
         Transitive_Parent => Transitive_Parent);
      --  If the parent of this new environment comes from a named environment
      --  lookup, register this new environment so that its parent is updated
      --  when the precence for this named environment changes.
      if Parent_From_Name then
         declare
            NED : constant Named_Env_Descriptor_Access := State.Current_NED;
         begin
            Self.Unit.Exiled_Envs.Append ((NED, Self.Self_Env));
            NED.Foreign_Envs.Insert (Self, Self.Self_Env);
         end;
      end if;
      --  From now on, the current environment is Self.Self_Env, with a direct
      --  access to it. It does not go through the env naming scheme, since
      --  only this node and its children (i.e. non-foreign nodes) will access
      --  it as a "current" environment during PLE.
      Use_Direct_Env (State, Self.Self_Env);
      --  Register the environment we just created on all the requested names
      if Names /= null then
         declare
            Context   : constant Internal_Context := Self.Unit.Context;
            Env       : constant Lexical_Env := Self.Self_Env;
            NENU      : NED_Maps.Map renames
               State.Unit_State.Named_Envs_Needing_Update;
         begin
            for N of Names.Items loop
               Register_Named_Env (Context, N, Env, NENU);
            end loop;
            Dec_Ref (Names);
         end;
      end if;
   end Add_Env;
   ---------------------
   -- Pre_Env_Actions --
   ---------------------
   procedure Pre_Env_Actions
     (Self            : Bare_Ada_Node;
      State           : in out PLE_Node_State;
      Add_To_Env_Only : Boolean := False) is
   begin
   case Self.Kind is
            when Ada_With_Clause =>
            With_Clause_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Use_Type_Clause =>
            Use_Type_Clause_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Use_Package_Clause =>
            Use_Package_Clause_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Access_To_Subp_Def =>
            Access_To_Subp_Def_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Named_Stmt =>
            Named_Stmt_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Extended_Return_Stmt =>
            Extended_Return_Stmt_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Begin_Block .. Ada_Decl_Block =>
            Block_Stmt_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_For_Loop_Stmt =>
            For_Loop_Stmt_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Accept_Stmt .. Ada_Accept_Stmt_With_Stmts =>
            Accept_Stmt_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Pragma_Node =>
            Pragma_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Decl_Expr =>
            Decl_Expr_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Private_Part =>
            Private_Part_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Single_Task_Decl =>
            Single_Task_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Single_Protected_Decl =>
            Single_Protected_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Package_Renaming_Decl =>
            Package_Renaming_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Object_Decl .. Ada_No_Type_Object_Renaming_Decl =>
            Object_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Number_Decl =>
            Number_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Label_Decl =>
            Label_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Generic_Subp_Renaming_Decl =>
            Generic_Subp_Renaming_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Generic_Package_Renaming_Decl =>
            Generic_Package_Renaming_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Generic_Subp_Instantiation =>
            Generic_Subp_Instantiation_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Generic_Package_Instantiation =>
            Generic_Package_Instantiation_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Generic_Subp_Decl =>
            Generic_Subp_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Generic_Package_Decl =>
            Generic_Package_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_For_Loop_Var_Decl =>
            For_Loop_Var_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Exception_Handler =>
            Exception_Handler_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Exception_Decl =>
            Exception_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Entry_Index_Spec =>
            Entry_Index_Spec_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Task_Body =>
            Task_Body_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Protected_Body =>
            Protected_Body_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Package_Body =>
            Package_Body_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Entry_Body =>
            Entry_Body_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Task_Body_Stub =>
            Task_Body_Stub_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Subp_Body_Stub =>
            Subp_Body_Stub_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Protected_Body_Stub =>
            Protected_Body_Stub_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Package_Body_Stub =>
            Package_Body_Stub_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Expr_Function .. Ada_Subp_Renaming_Decl =>
            Base_Subp_Body_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Generic_Subp_Internal =>
            Generic_Subp_Internal_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Enum_Literal_Decl .. Ada_Synthetic_Char_Enum_Lit =>
            Enum_Literal_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Entry_Decl =>
            Entry_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Abstract_Subp_Decl .. Ada_Subp_Decl | Ada_Synthetic_Subp_Decl =>
            Basic_Subp_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Anonymous_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
            null;
            when Ada_Type_Decl =>
            Type_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Single_Task_Type_Decl =>
            Single_Task_Type_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Task_Type_Decl =>
            Task_Type_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Protected_Type_Decl =>
            Protected_Type_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Incomplete_Type_Decl .. Ada_Incomplete_Tagged_Type_Decl =>
            Incomplete_Type_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Classwide_Type_Decl =>
            null;
            when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
            Base_Type_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Package_Decl =>
            Package_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Generic_Package_Internal =>
            Generic_Package_Internal_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Param_Spec =>
            Param_Spec_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Discriminant_Spec =>
            Discriminant_Spec_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Component_Decl =>
            Component_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
            when Ada_Abstract_State_Decl =>
            Abstract_State_Decl_Pre_Env_Actions (Self, State, Add_To_Env_Only);
      when others =>  null; 
   end case;
   end Pre_Env_Actions;
   ----------------------
   -- Post_Env_Actions --
   ----------------------
   procedure Post_Env_Actions
     (Self : Bare_Ada_Node; State : in out PLE_Node_State) is
   begin
   case Self.Kind is
            when Ada_With_Clause =>
            null;
            when Ada_Use_Type_Clause =>
            Use_Type_Clause_Post_Env_Actions (Self, State);
            when Ada_Use_Package_Clause =>
            Use_Package_Clause_Post_Env_Actions (Self, State);
            when Ada_Access_To_Subp_Def =>
            null;
            when Ada_Named_Stmt =>
            null;
            when Ada_Extended_Return_Stmt =>
            null;
            when Ada_Begin_Block .. Ada_Decl_Block =>
            null;
            when Ada_For_Loop_Stmt =>
            null;
            when Ada_Accept_Stmt .. Ada_Accept_Stmt_With_Stmts =>
            null;
            when Ada_Pragma_Node =>
            null;
            when Ada_Decl_Expr =>
            null;
            when Ada_Private_Part =>
            null;
            when Ada_Single_Task_Decl =>
            null;
            when Ada_Single_Protected_Decl =>
            null;
            when Ada_Package_Renaming_Decl =>
            null;
            when Ada_Object_Decl .. Ada_No_Type_Object_Renaming_Decl =>
            null;
            when Ada_Number_Decl =>
            null;
            when Ada_Label_Decl =>
            null;
            when Ada_Generic_Subp_Renaming_Decl =>
            null;
            when Ada_Generic_Package_Renaming_Decl =>
            null;
            when Ada_Generic_Subp_Instantiation =>
            null;
            when Ada_Generic_Package_Instantiation =>
            null;
            when Ada_Generic_Subp_Decl =>
            Generic_Subp_Decl_Post_Env_Actions (Self, State);
            when Ada_Generic_Package_Decl =>
            null;
            when Ada_For_Loop_Var_Decl =>
            null;
            when Ada_Exception_Handler =>
            null;
            when Ada_Exception_Decl =>
            null;
            when Ada_Entry_Index_Spec =>
            null;
            when Ada_Task_Body =>
            null;
            when Ada_Protected_Body =>
            null;
            when Ada_Package_Body =>
            null;
            when Ada_Entry_Body =>
            null;
            when Ada_Task_Body_Stub =>
            null;
            when Ada_Subp_Body_Stub =>
            null;
            when Ada_Protected_Body_Stub =>
            null;
            when Ada_Package_Body_Stub =>
            null;
            when Ada_Expr_Function .. Ada_Subp_Renaming_Decl =>
            null;
            when Ada_Generic_Subp_Internal =>
            null;
            when Ada_Enum_Literal_Decl .. Ada_Synthetic_Char_Enum_Lit =>
            null;
            when Ada_Entry_Decl =>
            null;
            when Ada_Abstract_Subp_Decl .. Ada_Subp_Decl | Ada_Synthetic_Subp_Decl =>
            Basic_Subp_Decl_Post_Env_Actions (Self, State);
            when Ada_Anonymous_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
            null;
            when Ada_Type_Decl =>
            Type_Decl_Post_Env_Actions (Self, State);
            when Ada_Single_Task_Type_Decl =>
            null;
            when Ada_Task_Type_Decl =>
            null;
            when Ada_Protected_Type_Decl =>
            null;
            when Ada_Incomplete_Type_Decl .. Ada_Incomplete_Tagged_Type_Decl =>
            null;
            when Ada_Classwide_Type_Decl =>
            null;
            when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
            null;
            when Ada_Package_Decl =>
            null;
            when Ada_Generic_Package_Internal =>
            null;
            when Ada_Param_Spec =>
            null;
            when Ada_Discriminant_Spec =>
            null;
            when Ada_Component_Decl =>
            null;
            when Ada_Abstract_State_Decl =>
            null;
      when others =>  null; 
   end case;
   end Post_Env_Actions;
   ----------------
   -- Get_Symbol --
   ----------------
   function Get_Symbol
     (Node : Bare_Ada_Node) return Symbol_Type is
   begin
      if Node = null then
         raise Property_Error with "cannot get the symbol of a null node";
      end if;
      return Get_Symbol (Token (Node, Node.Token_Start_Index));
   end Get_Symbol;
   ----------
   -- Text --
   ----------
   function Text
     (Node : Bare_Ada_Node) return Text_Type
   is
   begin
      if Node = null then
         raise Property_Error with "cannot get the text of a null node";
      end if;
      declare
         Start_T : constant Token_Reference :=
            Token (Node, Node.Token_Start_Index);
         End_T   : constant Token_Reference :=
            Token (Node, Node.Token_End_Index);
      begin
         --  No text is associated to synthetic and ghost nodes
         if Is_Synthetic (Node) then
            return "";
         end if;
         if Is_Ghost (Node) then
            return "";
         end if;
         return Text (Start_T, End_T);
      end;
   end Text;
   ---------------------
   -- Is_Visible_From --
   ---------------------
   function Is_Visible_From
     (Referenced_Env, Base_Env : Lexical_Env) return Boolean
   is
      Referenced_Unit : constant Internal_Unit :=
         Convert_Unit (Referenced_Env.Owner);
      Base_Unit       : constant Internal_Unit :=
         Convert_Unit (Base_Env.Owner);
   begin
      if Referenced_Unit = null then
         raise Property_Error with
            "referenced environment does not belong to any analysis unit";
      elsif Base_Unit = null then
         raise Property_Error with
            "base environment does not belong to any analysis unit";
      end if;
      return Is_Referenced_From (Referenced_Unit, Base_Unit);
   end Is_Visible_From;
   ----------
   -- Unit --
   ----------
   function Unit (Node : Bare_Ada_Node) return Internal_Unit is
   begin
      return Node.Unit;
   end Unit;
   function Lookup_Internal
     (Node : Bare_Ada_Node;
      Sloc : Source_Location) return Bare_Ada_Node;
   procedure Lookup_Relative
     (Node       : Bare_Ada_Node;
      Sloc       : Source_Location;
      Position   : out Relative_Position;
      Node_Found : out Bare_Ada_Node);
   --  Implementation helpers for the looking up process
   -----------------
   -- Set_Parents --
   -----------------
   procedure Set_Parents
     (Node, Parent : Bare_Ada_Node)
   is
   begin
      if Node = null then
         return;
      end if;
      Node.Parent := Bare_Ada_Node (Parent);
      for I in 1 .. Children_Count (Node) loop
         Set_Parents (Child (Node, I), Node);
      end loop;
   end Set_Parents;
   -------------
   -- Destroy --
   -------------
   procedure Destroy (Node : Bare_Ada_Node) is
   begin
      if Node = null then
         return;
      end if;
      Free_User_Fields (Node);
      for I in 1 .. Children_Count (Node) loop
         Destroy (Child (Node, I));
      end loop;
   end Destroy;
   -----------
   -- Child --
   -----------
   function Child (Node  : Bare_Ada_Node;
                   Index : Positive) return Bare_Ada_Node
   is
      Result          : Bare_Ada_Node;
      Index_In_Bounds : Boolean;
   begin
      Get_Child (Node, Index, Index_In_Bounds, Result);
      return Result;
   end Child;
   --------------
   -- Traverse --
   --------------
   function Traverse
     (Node  : Bare_Ada_Node;
      Visit : access function (Node : Bare_Ada_Node)
              return Visit_Status)
     return Visit_Status
   is
      Status : Visit_Status := Into;
   begin
      if Node /= null then
         Status := Visit (Node);
         --  Skip processing the child nodes if the returned status is Over
         --  or Stop. In the former case the previous call to Visit has taken
         --  care of processing the needed childs, and in the latter case we
         --  must immediately stop processing the tree.
         if Status = Into then
            for I in 1 .. Children_Count (Node) loop
               declare
                  Cur_Child : constant Bare_Ada_Node :=
                     Child (Node, I);
               begin
                  if Cur_Child /= null then
                     Status := Traverse (Cur_Child, Visit);
                     exit when Status /= Into;
                  end if;
               end;
            end loop;
         end if;
      end if;
      if Status = Stop then
         return Stop;
      --  At this stage the Over status has no sense and we just continue
      --  processing the tree.
      else
         return Into;
      end if;
   end Traverse;
   --------------
   -- Traverse --
   --------------
   procedure Traverse
     (Node  : Bare_Ada_Node;
      Visit : access function (Node : Bare_Ada_Node)
                               return Visit_Status)
   is
      Result_Status : Visit_Status;
      pragma Unreferenced (Result_Status);
   begin
      Result_Status := Traverse (Node, Visit);
   end Traverse;
   ------------------------
   -- Traverse_With_Data --
   ------------------------
   function Traverse_With_Data
     (Node  : Bare_Ada_Node;
      Visit : access function (Node : Bare_Ada_Node;
                               Data : in out Data_Type)
                               return Visit_Status;
      Data  : in out Data_Type)
      return Visit_Status
   is
      function Helper (Node : Bare_Ada_Node) return Visit_Status;
      ------------
      -- Helper --
      ------------
      function Helper (Node : Bare_Ada_Node) return Visit_Status is
      begin
         return Visit (Node, Data);
      end Helper;
      Saved_Data : Data_Type;
      Result     : Visit_Status;
   begin
      if Reset_After_Traversal then
         Saved_Data := Data;
      end if;
      Result := Traverse (Node, Helper'Access);
      if Reset_After_Traversal then
         Data := Saved_Data;
      end if;
      return Result;
   end Traverse_With_Data;
   ----------------
   -- Sloc_Range --
   ----------------
   function Sloc_Range
     (Node : Bare_Ada_Node) return Source_Location_Range
   is
      type Token_Anchor is (T_Start, T_End);
      type Token_Pos is record
         Pos    : Token_Index;
         Anchor : Token_Anchor;
      end record;
      TDH                    : Token_Data_Handler renames Node.Unit.TDH;
      Token_Start, Token_End : Token_Pos;
      function Get (Index : Token_Index) return Stored_Token_Data is
        (Get_Token (TDH, Index));
      function Sloc (T : Token_Pos) return Source_Location is
        (if T.Anchor = T_Start
         then Sloc_Start (TDH, Get (T.Pos))
         else Sloc_End (TDH, Get (T.Pos)));
   begin
      if Is_Synthetic (Node) then
         return Sloc_Range (Node.Parent);
      end if;
      if Is_Ghost (Node) then
         Token_Start := (if Node.Token_Start_Index = 1
                         then (1, T_Start)
                         else (Node.Token_Start_Index - 1, T_End));
         Token_End := Token_Start;
      else
         Token_Start := (Node.Token_Start_Index, T_Start);
         Token_End := (Node.Token_End_Index, T_End);
      end if;
      if Snaps_At_Start (Node)
         and then not Is_Ghost (Node)
         and then Token_Start.Pos /= 1
      then
         Token_Start := (Token_Start.Pos - 1, T_End);
      end if;
      if Snaps_At_End (Node) and then Token_End.Pos /= Last_Token (TDH) then
         Token_End := (Token_End.Pos + 1, T_Start);
      end if;
      return Make_Range (Sloc (Token_Start), Sloc (Token_End));
   end Sloc_Range;
   ------------
   -- Lookup --
   ------------
   function Lookup
     (Node : Bare_Ada_Node;
      Sloc : Source_Location) return Bare_Ada_Node
   is
      Position : Relative_Position;
      Result   : Bare_Ada_Node;
   begin
      if Sloc = No_Source_Location then
         return null;
      end if;
      Lookup_Relative
        (Bare_Ada_Node (Node), Sloc, Position, Result);
      return Result;
   end Lookup;
   ---------------------
   -- Lookup_Internal --
   ---------------------
   function Lookup_Internal
     (Node : Bare_Ada_Node;
      Sloc : Source_Location) return Bare_Ada_Node
   is
      --  For this implementation helper (i.e. internal primitive), we can
      --  assume that all lookups fall into this node's sloc range.
      pragma Assert (Compare (Sloc_Range (Node), Sloc) = Inside);
      Children : constant Internal_Bare_Ada_Node_Array :=
         Implementation.Children (Node);
      Pos      : Relative_Position;
      Result   : Bare_Ada_Node;
   begin
      --  Look for a child node that contains Sloc (i.e. return the most
      --  precise result).
      for Child of Children loop
         --  Note that we assume here that child nodes are ordered so that the
         --  first one has a sloc range that is before the sloc range of the
         --  second child node, etc.
         if Child /= null then
            Lookup_Relative (Child, Sloc, Pos, Result);
            case Pos is
               when Before =>
                   --  If this is the first node, Sloc is before it, so we can
                   --  stop here.  Otherwise, Sloc is between the previous
                   --  child node and the next one...  so we can stop here,
                   --  too.
                   return Node;
               when Inside =>
                   return Result;
               when After =>
                   --  Sloc is after the current child node, so see with the
                   --  next one.
                   null;
            end case;
         end if;
      end loop;
      --  If we reach this point, we found no children that covers Sloc, but
      --  Node still covers it (see the assertion).
      return Node;
   end Lookup_Internal;
   -------------
   -- Compare --
   -------------
   function Compare
     (Node : Bare_Ada_Node;
      Sloc : Source_Location) return Relative_Position is
   begin
      return Compare (Sloc_Range (Node), Sloc);
   end Compare;
   ---------------------
   -- Lookup_Relative --
   ---------------------
   procedure Lookup_Relative
     (Node       : Bare_Ada_Node;
      Sloc       : Source_Location;
      Position   : out Relative_Position;
      Node_Found : out Bare_Ada_Node)
   is
      Result : constant Relative_Position :=
        Compare (Node, Sloc);
   begin
      Position := Result;
      Node_Found := (if Result = Inside
                     then Lookup_Internal (Node, Sloc)
                     else null);
   end Lookup_Relative;
   -------------
   -- Compare --
   -------------
   function Compare
     (Left, Right : Bare_Ada_Node;
      Relation    : Comparison_Relation) return Boolean
   is
      LS, RS : Source_Location;
   begin
      if Left = null or else Right = null or else Left.Unit /= Right.Unit then
         raise Property_Error with "invalid node comparison";
      end if;
      LS := Start_Sloc (Sloc_Range (Left));
      RS := Start_Sloc (Sloc_Range (Right));
      return (case Relation is
              when Langkit_Support.Types.Less_Than        => LS < RS,
              when Langkit_Support.Types.Less_Or_Equal    => LS <= RS,
              when Langkit_Support.Types.Greater_Than     => LS > RS,
              when Langkit_Support.Types.Greater_Or_Equal => LS >= RS);
   end Compare;
   --------------
   -- Children --
   --------------
   function Children
     (Node : Bare_Ada_Node) return Internal_Bare_Ada_Node_Array
   is
      First : constant Integer := Bare_Ada_Node_Vectors.Index_Type'First;
      Last  : constant Integer := First + Children_Count (Node) - 1;
   begin
      return A : Internal_Bare_Ada_Node_Array (First .. Last)
      do
         for I in First .. Last loop
            A (I) := Child (Node, I);
         end loop;
      end return;
   end Children;
   function Children
     (Node : Bare_Ada_Node) return Bare_Ada_Node_Array_Access
   is
      C : Internal_Bare_Ada_Node_Array := Children (Node);
   begin
      return Ret : Bare_Ada_Node_Array_Access :=
         Create_Bare_Ada_Node_Array (C'Length)
      do
         Ret.Items := C;
      end return;
   end Children;
   ---------
   -- Get --
   ---------
   function Get
     (Node    : Bare_Ada_List;
      Index   : Integer;
      Or_Null : Boolean := False) return Bare_Ada_Node
   is
      function Length (Node : Bare_Ada_List) return Natural
      is (Node.Count);
      --  Wrapper around the Length primitive to get the compiler happy for the
      --  the package instantiation below.
      function Absolute_Get
        (L     : Bare_Ada_List;
         Index : Integer) return Bare_Ada_Node
      is (L.Nodes.all (Index + 1));
      --  L.Nodes is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Bare_Ada_Node,
         Sequence_Type => Bare_Ada_List,
         Length        => Length,
         Get           => Absolute_Get);
      Result : Bare_Ada_Node;
   begin
      if Node = null and then Or_Null then
         return null;
      elsif Relative_Get (Node, Index, Result) then
         return Result;
      elsif Or_Null then
         return null;
      else
         raise Property_Error with "out-of-bounds AST list access";
      end if;
   end Get;
   ---------------
   -- PP_Trivia --
   ---------------
   procedure PP_Trivia
     (Node        : Bare_Ada_Node;
      Line_Prefix : String := "")
   is
      Children_Prefix : constant String := Line_Prefix & "|  ";
   begin
      if Node = null then
         Put_Line (Line_Prefix & "None");
         return;
      end if;
      Put_Line (Line_Prefix & Kind_Name (Node));
      for C of Children_And_Trivia (Node) loop
         case C.Kind is
            when Trivia =>
               Put_Line (Children_Prefix & Image (Text (C.Trivia)));
            when Child =>
               PP_Trivia (C.Node, Children_Prefix);
         end case;
      end loop;
   end PP_Trivia;
   --------------------------
   -- Populate_Lexical_Env --
   --------------------------
   function Populate_Lexical_Env (Node : Bare_Ada_Node) return Boolean is
      Context    : constant Internal_Context := Node.Unit.Context;
      Unit_State : aliased PLE_Unit_State := (Named_Envs_Needing_Update => <>);
      Root_State : constant PLE_Node_State :=
        (Unit_State  => Unit_State'Unchecked_Access,
         Current_Env => Context.Root_Scope,
         Current_NED => null);
      function Populate_Internal
        (Node         : Bare_Ada_Node;
         Parent_State : PLE_Node_State) return Boolean;
      --  Do the lexical env population on Node and recurse on its children
      procedure Register_Foreign_Env
        (Node : Bare_Ada_Node; State : PLE_Node_State);
      --  Given a node and its PLE state, register Node.Self_Env as being
      --  initialized through the named environment mechanism, if that's indeed
      --  the case. Do nothing otherwise.
      -----------------------
      -- Populate_Internal --
      -----------------------
      function Populate_Internal
        (Node         : Bare_Ada_Node;
         Parent_State : PLE_Node_State) return Boolean
      is
         Result : Boolean := False;
         State  : PLE_Node_State := Parent_State;
      begin
         if Node = null then
            return Result;
         end if;
         --  By default (i.e. unless env actions add a new env), the
         --  environment we store in Node is the current one.
         Node.Self_Env := State.Current_Env;
         --  Run pre/post actions, and run PLE on children in between. Make
         --  sure we register the potential foreign Node.Self_Env environment
         --  at the end, even when an exception interrupts PLE to keep the
         --  state consistent.
         begin
            Pre_Env_Actions (Node, State);
            if State.Current_Env /= Null_Lexical_Env then
               Node.Self_Env := State.Current_Env;
               Register_Foreign_Env (Node, State);
            end if;
            --  Call recursively on children
            for I in First_Child_Index (Node) .. Last_Child_Index (Node) loop
               Result := Populate_Internal
                 (Child (Node, I), State) or else Result;
            end loop;
            Post_Env_Actions (Node, State);
         exception
            when Exc : Precondition_Failure | Property_Error =>
               if PLE_Errors_Trace.Is_Active then
                   GNATCOLL.Traces.Trace
                     (PLE_Errors_Trace,
                      "Exception raised during PLE "
                      & Ada.Exceptions.Exception_Name (Exc) & " : "
                      & Ada.Exceptions.Exception_Message (Exc));
                   GNATCOLL.Traces.Trace
                     (PLE_Errors_Trace,
                      GNAT.Traceback.Symbolic.Symbolic_Traceback (Exc));
               end if;
               Register_Foreign_Env (Node, State);
               return True;
         end;
         return Result;
      end Populate_Internal;
      --------------------------
      -- Register_Foreign_Env --
      --------------------------
      procedure Register_Foreign_Env
        (Node : Bare_Ada_Node; State : PLE_Node_State) is
      begin
         if State.Current_NED /= null then
            State.Current_NED.Nodes_With_Foreign_Env.Insert (Node);
            Node.Unit.Nodes_With_Foreign_Env.Insert (Node, State.Current_NED);
         end if;
      end Register_Foreign_Env;
   begin
      --  This function is meant to be called during an existing PLE pass. If
      --  if is called outside of this context, run the PLE pass on Node's
      --  analysis unit. Likewise, if PLE has not run on the unit that owns
      --  this PLE unit yet, do a full run, which will in the end trigger the
      --  PLE on this PLE unit.
      --
      --  We do this so that as soon as PLE is required on a PLE unit: the
      --  whole unit end up with its lexical environments populated.
      if not Context.In_Populate_Lexical_Env then
         begin
            Populate_Lexical_Env (Node.Unit);
            return False;
         exception
            when Precondition_Failure | Property_Error =>
               return True;
         end;
      end if;
      --  This is intended to be called on the root node only
      if Node.Parent /= null then
         raise Program_Error;
      end if;
      return Result : constant Boolean :=
         Populate_Internal (Node, Root_State)
      do
         Update_Named_Envs (Context, Unit_State.Named_Envs_Needing_Update);
      end return;
   end Populate_Lexical_Env;
   ------------------------------
   -- AST_Envs_Node_Text_Image --
   ------------------------------
   function AST_Envs_Node_Text_Image
     (Node  : Bare_Ada_Node;
      Short : Boolean := True) return Text_Type is
   begin
      if Short then
         if Node = null then
            return "null";
         end if;
         return To_Text (Basename (Node.Unit))
           & ":" & To_Text (Image (Start_Sloc (Sloc_Range (Node))));
      else
         return Short_Text_Image (Node);
      end if;
   end AST_Envs_Node_Text_Image;
   -------------------
   -- Is_Rebindable --
   -------------------
   function Is_Rebindable (Node : Bare_Ada_Node) return Boolean is
   begin
         return Node.Kind in Ada_Generic_Package_Decl .. Ada_Generic_Subp_Decl;
   end Is_Rebindable;
   -----------------------
   -- Acquire_Rebinding --
   -----------------------
   function Acquire_Rebinding
     (Node             : Bare_Ada_Node;
      Parent           : Env_Rebindings;
      Old_Env, New_Env : Lexical_Env) return Env_Rebindings
   is
      Result    : Env_Rebindings;
      Available : Env_Rebindings_Vectors.Vector renames
         Node.Unit.Context.Available_Rebindings;
   begin
      --  Use an existing and available Env_Rebindings_Type record for Node's
      --  Context, otherwise allocate a new rebinding.
      Result := (if Available.Is_Empty
                 then new Env_Rebindings_Type'(Version => 0, others => <>)
                 else Available.Pop);
      Result.Parent := Parent;
      Result.Old_Env := Old_Env;
      Result.New_Env := New_Env;
      Result.Children := Env_Rebindings_Vectors.Empty_Vector;
      return Result;
   end Acquire_Rebinding;
   -----------------------
   -- Release_Rebinding --
   -----------------------
   procedure Release_Rebinding (Self : in out Env_Rebindings) is
      Available : Env_Rebindings_Vectors.Vector renames
         Unwrap (Self.Old_Env).Node.Unit.Context.Available_Rebindings;
   begin
      --  Bumping the version number, to invalidate existing references to
      --  Self.
      Self.Version := Self.Version + 1;
      Self.Children.Destroy;
      Available.Append (Self);
      Self := null;
   end Release_Rebinding;
   ------------------------
   -- Register_Rebinding --
   ------------------------
   procedure Register_Rebinding
     (Node : Bare_Ada_Node; Rebinding : Env_Rebindings) is
   begin
      Node.Unit.Rebindings.Append (Rebinding);
   end Register_Rebinding;
   --------------------
   -- Element_Parent --
   --------------------
   function Element_Parent
     (Node : Bare_Ada_Node) return Bare_Ada_Node
   is (Node.Parent);
   ---------------
   -- Node_Unit --
   ---------------
   function Node_Unit (Node : Bare_Ada_Node) return Generic_Unit_Ptr is
   begin
      return Convert_Unit (Node.Unit);
   end Node_Unit;
   ----------
   -- Hash --
   ----------
   function Hash (Node : Bare_Ada_Node) return Hash_Type
   is
      function H is new Hash_Access
        (Root_Node_Record, Bare_Ada_Node);
   begin
      return H (Node);
   end Hash;
      function Hash (B : Boolean) return Hash_Type is (Boolean'Pos (B));
      function Hash (I : Integer) return Hash_Type is (Hash_Type'Mod (I));
   ------------------------
   -- Named environments --
   ------------------------
   ---------
   -- Add --
   ---------
   procedure Add
     (Self : in out NED_Assoc_Maps.Map;
      Key  : Symbol_Type;
      Node : AST_Envs.Internal_Map_Node)
   is
      use NED_Assoc_Maps;
      Pos   : Cursor;
      Dummy : Boolean;
   begin
      --  Make sure there is a vector entry for Key
      Self.Insert (Key, Internal_Map_Node_Vectors.Empty_Vector, Pos, Dummy);
      --  Append Node to that vector
      declare
         V : Internal_Map_Node_Vectors.Vector renames Self.Reference (Pos);
      begin
         V.Append (Node);
      end;
   end Add;
   ------------
   -- Remove --
   ------------
   procedure Remove
     (Self : in out NED_Assoc_Maps.Map;
      Key  : Symbol_Type;
      Node : Bare_Ada_Node)
   is
      use NED_Assoc_Maps;
      V : Internal_Map_Node_Vectors.Vector renames Self.Reference (Key);
   begin
      --  Remove the (assumed unique) entry in V whose node is Node. The order
      --  of items in V is not significant, so we can use Pop for efficient
      --  removal. Do the traversal in reverse order for correctness.
      for I in reverse 1 .. V.Length loop
         if V.Get_Access (I).Node = Node then
            V.Pop (I);
            exit;
         end if;
      end loop;
   end Remove;
   ------------------------------
   -- Get_Named_Env_Descriptor --
   ------------------------------
   function Get_Named_Env_Descriptor
     (Context : Internal_Context;
      Name    : Symbol_Type) return Named_Env_Descriptor_Access
   is
      use NED_Maps;
      --  Look for an existing entry for Name
      Pos : constant Cursor := Context.Named_Envs.Find (Name);
   begin
      if Has_Element (Pos) then
         return Element (Pos);
      end if;
      --  There is no such entry: create one
      return Result : constant Named_Env_Descriptor_Access :=
         new Named_Env_Descriptor'
           (Name                   => Name,
            Envs                   => <>,
            Env_With_Precedence    => Empty_Env,
            Foreign_Nodes          => <>,
            Foreign_Envs           => <>,
            Nodes_With_Foreign_Env => <>)
      do
         Context.Named_Envs.Insert (Name, Result);
      end return;
   end Get_Named_Env_Descriptor;
   ------------------------
   -- Register_Named_Env --
   ------------------------
   procedure Register_Named_Env
     (Context                   : Internal_Context;
      Name                      : Symbol_Type;
      Env                       : Lexical_Env;
      Named_Envs_Needing_Update : in out NED_Maps.Map)
   is
      NED_Access : constant Named_Env_Descriptor_Access :=
         Get_Named_Env_Descriptor (Context, Name);
      NED        : Named_Env_Descriptor renames NED_Access.all;
      Node       : constant Bare_Ada_Node := Env_Node (Env);
   begin
      NED.Envs.Insert (Node, Env);
      Node.Unit.Named_Envs.Append ((Name, Env));
      --  If that insertion must change the env that has precedence, signal
      --  that NED requires an update.
      if NED.Envs.First_Element /= NED.Env_With_Precedence then
         Named_Envs_Needing_Update.Include (Name, NED_Access);
      end if;
   end Register_Named_Env;
   ----------------------
   -- Update_Named_Env --
   ----------------------
   procedure Update_Named_Envs
     (Context : Internal_Context; Named_Envs : NED_Maps.Map)
   is
      Require_Cache_Reset : Boolean := False;
   begin
      for Cur in Named_Envs.Iterate loop
         declare
            NE      : Named_Env_Descriptor renames NED_Maps.Element (Cur).all;
            New_Env : constant Lexical_Env :=
              (if NE.Envs.Is_Empty
               then Empty_Env
               else NE.Envs.First_Element);
         begin
            --  If there was an environment with precedence, remove its foreign
            --  nodes.
            if NE.Env_With_Precedence /= Empty_Env then
               for Cur in NE.Foreign_Nodes.Iterate loop
                  declare
                     Key   : constant Symbol_Type :=
                        NED_Assoc_Maps.Key (Cur);
                     Nodes : Internal_Map_Node_Vectors.Vector renames
                        NE.Foreign_Nodes.Reference (Cur);
                  begin
                     for N of Nodes loop
                        Remove (NE.Env_With_Precedence, Key, N.Node);
                     end loop;
                  end;
               end loop;
            end if;
            --  Now, set the new environment that has precedence
            NE.Env_With_Precedence := New_Env;
            --  Add the foreign nodes to the new environment with precedence,
            --  if any.
            for Cur in NE.Foreign_Nodes.Iterate loop
               declare
                  Key   : constant Symbol_Type :=
                     NED_Assoc_Maps.Key (Cur);
                  Nodes : Internal_Map_Node_Vectors.Vector renames
                     NE.Foreign_Nodes.Reference (Cur);
               begin
                  for N of Nodes loop
                     Add (New_Env, Key, N.Node, N.Md, N.Resolver);
                  end loop;
               end;
            end loop;
            --  Set the parent environment of all foreign environments
            for Cur in NE.Foreign_Envs.Iterate loop
               declare
                  Env : Lexical_Env_Record renames
                     Unwrap (Sorted_Env_Maps.Element (Cur)).all;
               begin
                  Env.Parent := New_Env;
                  --  We have updated the lexical env hierarchy (e.g. an env
                  --  which had no parent may have one now), so the cached
                  --  entries for queries that traveresed the old env hierarchy
                  --  need to be invalidated.
                  Require_Cache_Reset := True;
               end;
            end loop;
            --  Update nodes whose environment was the old env with precedence
            for N of NE.Nodes_With_Foreign_Env loop
               N.Self_Env := New_Env;
            end loop;
         end;
      end loop;
      if Require_Cache_Reset then
         Invalidate_Caches (Context, Invalidate_Envs => True);
      end if;
   end Update_Named_Envs;
   --------------------------
   -- Big integers wrapper --
   --------------------------
   ------------------------
   -- Create_Big_Integer --
   ------------------------
   function Create_Big_Integer
     (Image : String; Base : Integer := 10) return Big_Integer_Type
   is
      use GNATCOLL.GMP;
      use GNATCOLL.GMP.Integers;
   begin
      return new Big_Integer_Record'(Value     => Make (Image, Int (Base)),
                                     Ref_Count => 1);
   end Create_Big_Integer;
   ------------------------
   -- Create_Big_Integer --
   ------------------------
   function Create_Big_Integer
     (Big_Int : GNATCOLL.GMP.Integers.Big_Integer) return Big_Integer_Type
   is
      Result : constant Big_Integer_Type :=
         new Big_Integer_Record'(Value     => <>,
                                 Ref_Count => 1);
   begin
      Result.Value.Set (Big_Int);
      return Result;
   end Create_Big_Integer;
   ------------------------
   -- Create_Big_Integer --
   ------------------------
   function Create_Big_Integer (Int : Integer) return Big_Integer_Type is
      Result : constant Big_Integer_Type :=
         new Big_Integer_Record'(Value     => <>,
                                 Ref_Count => 1);
   begin
      Result.Value.Set (GNATCOLL.GMP.Long (Int));
      return Result;
   end Create_Big_Integer;
   -------------------------------
   -- Create_Public_Big_Integer --
   -------------------------------
   function Create_Public_Big_Integer
     (Big_Int : Big_Integer_Type) return GNATCOLL.GMP.Integers.Big_Integer is
   begin
      return Result : GNATCOLL.GMP.Integers.Big_Integer do
         Result.Set (Big_Int.Value);
      end return;
   end Create_Public_Big_Integer;
   -----------------
   -- Trace_Image --
   -----------------
   function Trace_Image (I : Big_Integer_Type) return String is
   begin
      return GNATCOLL.GMP.Integers.Image (I.Value);
   end Trace_Image;
   ----------------
   -- To_Integer --
   ----------------
   function To_Integer (Big_Int : Big_Integer_Type) return Integer is
      Image : constant String := Big_Int.Value.Image;
   begin
      return Integer'Value (Image);
   exception
      when Constraint_Error =>
         raise Property_Error with "out of range big integer";
   end To_Integer;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (Big_Int : Big_Integer_Type) is
   begin
      if Big_Int.Ref_Count /= -1 then
         Big_Int.Ref_Count := Big_Int.Ref_Count + 1;
      end if;
   end Inc_Ref;
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (Big_Int : in out Big_Integer_Type) is
      procedure Destroy is new Ada.Unchecked_Deallocation
        (Big_Integer_Record, Big_Integer_Type);
   begin
      if Big_Int = null or else Big_Int.Ref_Count = -1 then
         return;
      end if;
      Big_Int.Ref_Count := Big_Int.Ref_Count - 1;
      if Big_Int.Ref_Count = 0 then
         Destroy (Big_Int);
      end if;
   end Dec_Ref;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (Left, Right : Big_Integer_Type) return Boolean is
      use type GNATCOLL.GMP.Integers.Big_Integer;
   begin
      return Left.Value = Right.Value;
   end Equivalent;
   ---------
   -- "<" --
   ---------
   function "<" (Left, Right : Big_Integer_Type) return Boolean is
      use type GNATCOLL.GMP.Integers.Big_Integer;
   begin
      return Left.Value < Right.Value;
   end "<";
   ----------
   -- "<=" --
   ----------
   function "<=" (Left, Right : Big_Integer_Type) return Boolean is
      use type GNATCOLL.GMP.Integers.Big_Integer;
   begin
      return Left.Value <= Right.Value;
   end "<=";
   ---------
   -- ">" --
   ---------
   function ">" (Left, Right : Big_Integer_Type) return Boolean is
      use type GNATCOLL.GMP.Integers.Big_Integer;
   begin
      return Left.Value > Right.Value;
   end ">";
   ----------
   -- ">=" --
   ----------
   function ">=" (Left, Right : Big_Integer_Type) return Boolean is
      use type GNATCOLL.GMP.Integers.Big_Integer;
   begin
      return Left.Value >= Right.Value;
   end ">=";
   ---------
   -- "+" --
   ---------
   function "+" (Left, Right : Big_Integer_Type) return Big_Integer_Type is
      use type GNATCOLL.GMP.Integers.Big_Integer;
   begin
      return Create_Big_Integer (Left.Value + Right.Value);
   end "+";
   ---------
   -- "-" --
   ---------
   function "-" (Left, Right : Big_Integer_Type) return Big_Integer_Type is
      use type GNATCOLL.GMP.Integers.Big_Integer;
   begin
      return Create_Big_Integer (Left.Value - Right.Value);
   end "-";
   ------------------
   -- Unit_Version --
   ------------------
   function Unit_Version (Unit : Generic_Unit_Ptr) return Version_Number is
   begin
      return Convert_Unit (Unit).Unit_Version;
   end Unit_Version;
   -------------------------
   -- Get_Context_Version --
   -------------------------
   function Get_Context_Version
     (Node : Bare_Ada_Node) return Version_Number is
   begin
      return Node.Unit.Context.Cache_Version;
   end Get_Context_Version;
   --------------------------
   -- Properties_May_Raise --
   --------------------------
   function Properties_May_Raise
     (Exc : Ada.Exceptions.Exception_Occurrence) return Boolean is
   begin
      return Ada.Exceptions.Exception_Identity (Exc) in
            Precondition_Failure'Identity
               |
            Property_Error'Identity
      ;
   end Properties_May_Raise;
   ----------------------
   -- Short_Text_Image --
   ----------------------
   function Short_Text_Image (Self : Bare_Ada_Node) return Text_Type
   is
   begin
      if Self = null then
         return "None";
      end if;
   case Self.Kind is
            when Ada_Char_Literal .. Ada_String_Literal =>
         return Base_Id_Short_Image (Self);
            when Ada_Abstract_State_Decl .. Ada_Single_Task_Decl =>
         return Basic_Decl_Short_Image (Self);
      when others => 
         return "<" & To_Text (Kind_Name (Self))
                & " "
                & To_Text
                  (Ada.Directories.Simple_Name
                     (Get_Filename (Unit (Self))))
                & ":" & To_Text (Image (Sloc_Range (Self))) & ">";
   end case;
   end Short_Text_Image;
   --------------------
   -- Snaps_At_Start --
   --------------------
   function Snaps_At_Start (Self : Bare_Ada_Node) return Boolean is
   begin
   case Self.Kind is
            when Ada_Handled_Stmts =>
         return True;
            when Ada_Declarative_Part .. Ada_Public_Part =>
         return True;
      when others => 
         return False;
   end case;
   end Snaps_At_Start;
   ------------------
   -- Snaps_At_End --
   ------------------
   function Snaps_At_End (Self : Bare_Ada_Node) return Boolean is
   begin
   case Self.Kind is
            when Ada_Handled_Stmts =>
         return True;
            when Ada_Declarative_Part .. Ada_Public_Part =>
         return True;
      when others => 
         return Is_Incomplete (Self);
   end case;
   end Snaps_At_End;
   -------------
   -- Parents --
   -------------
   function Parents
     (Node      : Bare_Ada_Node;
      With_Self : Boolean := True)
      return Bare_Ada_Node_Array_Access
   is
      Count : Natural := 0;
      Start : Bare_Ada_Node :=
        (if With_Self then Node else Node.Parent);
      Cur   : Bare_Ada_Node := Start;
   begin
      while Cur /= null loop
         Count := Count + 1;
         Cur := Cur.Parent;
      end loop;
      declare
         Result : constant Bare_Ada_Node_Array_Access :=
            Create_Bare_Ada_Node_Array (Count);
      begin
         Cur := Start;
         for I in Result.Items'Range loop
            Result.Items (I) := Cur;
            Cur := Cur.Parent;
         end loop;
         return Result;
      end;
   end Parents;
   -----------------------
   -- First_Child_Index --
   -----------------------
   function First_Child_Index (Node : Bare_Ada_Node) return Natural
   is (1);
   ----------------------
   -- Last_Child_Index --
   ----------------------
   function Last_Child_Index (Node : Bare_Ada_Node) return Natural
   is (Children_Count (Node));
   ---------------
   -- Get_Child --
   ---------------
   procedure Get_Child
     (Node            : Bare_Ada_Node;
      Index           : Positive;
      Index_In_Bounds : out Boolean;
      Result          : out Bare_Ada_Node)
   is
      K : constant Ada_Node_Kind_Type := Node.Kind;
   begin
      Index_In_Bounds := True;
      Result := null;
      case Ada_Ada_Node (K) is
when Ada_Ada_List =>
declare
N_Bare_Ada_List : constant Bare_Ada_List := Node;
begin
                    if Index > N_Bare_Ada_List.Count then
                        Index_In_Bounds := False;
                    else
                        Result := N_Bare_Ada_List.Nodes (Index);
                    end if;
                    return;
end;
when Ada_Constrained_Array_Indices_Range =>
declare
N_Bare_Constrained_Array_Indices : constant Bare_Constrained_Array_Indices := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Constrained_Array_Indices.Constrained_Array_Indices_F_List;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Unconstrained_Array_Indices_Range =>
declare
N_Bare_Unconstrained_Array_Indices : constant Bare_Unconstrained_Array_Indices := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Unconstrained_Array_Indices.Unconstrained_Array_Indices_F_Types;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Aspect_Assoc_Range =>
declare
N_Bare_Aspect_Assoc : constant Bare_Aspect_Assoc := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Aspect_Assoc.Aspect_Assoc_F_Id;
                            return;
                        when 2 =>
                            Result := N_Bare_Aspect_Assoc.Aspect_Assoc_F_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_At_Clause_Range =>
declare
N_Bare_At_Clause : constant Bare_At_Clause := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_At_Clause.At_Clause_F_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_At_Clause.At_Clause_F_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Attribute_Def_Clause_Range =>
declare
N_Bare_Attribute_Def_Clause : constant Bare_Attribute_Def_Clause := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Attribute_Def_Clause.Attribute_Def_Clause_F_Attribute_Expr;
                            return;
                        when 2 =>
                            Result := N_Bare_Attribute_Def_Clause.Attribute_Def_Clause_F_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Enum_Rep_Clause_Range =>
declare
N_Bare_Enum_Rep_Clause : constant Bare_Enum_Rep_Clause := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Enum_Rep_Clause.Enum_Rep_Clause_F_Type_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Enum_Rep_Clause.Enum_Rep_Clause_F_Aggregate;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Record_Rep_Clause_Range =>
declare
N_Bare_Record_Rep_Clause : constant Bare_Record_Rep_Clause := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Record_Rep_Clause.Record_Rep_Clause_F_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Record_Rep_Clause.Record_Rep_Clause_F_At_Expr;
                            return;
                        when 3 =>
                            Result := N_Bare_Record_Rep_Clause.Record_Rep_Clause_F_Components;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Aspect_Spec_Range =>
declare
N_Bare_Aspect_Spec : constant Bare_Aspect_Spec := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Aspect_Spec.Aspect_Spec_F_Aspect_Assocs;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Contract_Case_Assoc_Range =>
declare
N_Bare_Contract_Case_Assoc : constant Bare_Contract_Case_Assoc := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Contract_Case_Assoc.Contract_Case_Assoc_F_Guard;
                            return;
                        when 2 =>
                            Result := N_Bare_Contract_Case_Assoc.Contract_Case_Assoc_F_Consequence;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Pragma_Argument_Assoc_Range =>
declare
N_Bare_Pragma_Argument_Assoc : constant Bare_Pragma_Argument_Assoc := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Pragma_Argument_Assoc.Pragma_Argument_Assoc_F_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Pragma_Argument_Assoc.Pragma_Argument_Assoc_F_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Entry_Spec_Range =>
declare
N_Bare_Entry_Spec : constant Bare_Entry_Spec := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Entry_Spec.Entry_Spec_F_Entry_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Entry_Spec.Entry_Spec_F_Family_Type;
                            return;
                        when 3 =>
                            Result := N_Bare_Entry_Spec.Entry_Spec_F_Entry_Params;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Predefined_Bin_Op_Spec_Range =>
declare
N_Bare_Predefined_Bin_Op_Spec : constant Bare_Predefined_Bin_Op_Spec := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Predefined_Bin_Op_Spec.Predefined_Bin_Op_Spec_F_Left_Param;
                            return;
                        when 2 =>
                            Result := N_Bare_Predefined_Bin_Op_Spec.Predefined_Bin_Op_Spec_F_Right_Param;
                            return;
                        when 3 =>
                            Result := N_Bare_Predefined_Bin_Op_Spec.Predefined_Bin_Op_Spec_F_Return_Type_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Predefined_Un_Op_Spec_Range =>
declare
N_Bare_Predefined_Un_Op_Spec : constant Bare_Predefined_Un_Op_Spec := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Predefined_Un_Op_Spec.Predefined_Un_Op_Spec_F_Right_Param;
                            return;
                        when 2 =>
                            Result := N_Bare_Predefined_Un_Op_Spec.Predefined_Un_Op_Spec_F_Return_Type_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Subp_Spec_Range =>
declare
N_Bare_Subp_Spec : constant Bare_Subp_Spec := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Subp_Spec.Subp_Spec_F_Subp_Kind;
                            return;
                        when 2 =>
                            Result := N_Bare_Subp_Spec.Subp_Spec_F_Subp_Name;
                            return;
                        when 3 =>
                            Result := N_Bare_Subp_Spec.Subp_Spec_F_Subp_Params;
                            return;
                        when 4 =>
                            Result := N_Bare_Subp_Spec.Subp_Spec_F_Subp_Returns;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Component_List_Range =>
declare
N_Bare_Component_List : constant Bare_Component_List := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Component_List.Component_List_F_Components;
                            return;
                        when 2 =>
                            Result := N_Bare_Component_List.Component_List_F_Variant_Part;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Known_Discriminant_Part_Range =>
declare
N_Bare_Known_Discriminant_Part : constant Bare_Known_Discriminant_Part := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Known_Discriminant_Part.Known_Discriminant_Part_F_Discr_Specs;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Entry_Completion_Formal_Params_Range =>
declare
N_Bare_Entry_Completion_Formal_Params : constant Bare_Entry_Completion_Formal_Params := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Entry_Completion_Formal_Params.Entry_Completion_Formal_Params_F_Params;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Generic_Formal_Part_Range =>
declare
N_Bare_Generic_Formal_Part : constant Bare_Generic_Formal_Part := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Generic_Formal_Part.Generic_Formal_Part_F_Decls;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Base_Record_Def =>
declare
N_Bare_Base_Record_Def : constant Bare_Base_Record_Def := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Base_Record_Def.Base_Record_Def_F_Components;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Aggregate_Assoc_Range =>
declare
N_Bare_Aggregate_Assoc : constant Bare_Aggregate_Assoc := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Aggregate_Assoc.Aggregate_Assoc_F_Designators;
                            return;
                        when 2 =>
                            Result := N_Bare_Aggregate_Assoc.Aggregate_Assoc_F_R_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Composite_Constraint_Assoc_Range =>
declare
N_Bare_Composite_Constraint_Assoc : constant Bare_Composite_Constraint_Assoc := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Composite_Constraint_Assoc.Composite_Constraint_Assoc_F_Ids;
                            return;
                        when 2 =>
                            Result := N_Bare_Composite_Constraint_Assoc.Composite_Constraint_Assoc_F_Constraint_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Iterated_Assoc_Range =>
declare
N_Bare_Iterated_Assoc : constant Bare_Iterated_Assoc := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Iterated_Assoc.Iterated_Assoc_F_Spec;
                            return;
                        when 2 =>
                            Result := N_Bare_Iterated_Assoc.Iterated_Assoc_F_R_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Param_Assoc_Range =>
declare
N_Bare_Param_Assoc : constant Bare_Param_Assoc := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Param_Assoc.Param_Assoc_F_Designator;
                            return;
                        when 2 =>
                            Result := N_Bare_Param_Assoc.Param_Assoc_F_R_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Abstract_State_Decl_Range =>
declare
N_Bare_Abstract_State_Decl : constant Bare_Abstract_State_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Abstract_State_Decl.Abstract_State_Decl_F_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Abstract_State_Decl.Abstract_State_Decl_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Anonymous_Expr_Decl_Range =>
declare
N_Bare_Anonymous_Expr_Decl : constant Bare_Anonymous_Expr_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Anonymous_Expr_Decl.Anonymous_Expr_Decl_F_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Component_Decl_Range =>
declare
N_Bare_Component_Decl : constant Bare_Component_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Component_Decl.Component_Decl_F_Ids;
                            return;
                        when 2 =>
                            Result := N_Bare_Component_Decl.Component_Decl_F_Component_Def;
                            return;
                        when 3 =>
                            Result := N_Bare_Component_Decl.Component_Decl_F_Default_Expr;
                            return;
                        when 4 =>
                            Result := N_Bare_Component_Decl.Component_Decl_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Discriminant_Spec_Range =>
declare
N_Bare_Discriminant_Spec : constant Bare_Discriminant_Spec := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Discriminant_Spec.Discriminant_Spec_F_Ids;
                            return;
                        when 2 =>
                            Result := N_Bare_Discriminant_Spec.Discriminant_Spec_F_Type_Expr;
                            return;
                        when 3 =>
                            Result := N_Bare_Discriminant_Spec.Discriminant_Spec_F_Default_Expr;
                            return;
                        when 4 =>
                            Result := N_Bare_Discriminant_Spec.Discriminant_Spec_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Generic_Formal =>
declare
N_Bare_Generic_Formal : constant Bare_Generic_Formal := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Generic_Formal.Generic_Formal_F_Decl;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Param_Spec_Range =>
declare
N_Bare_Param_Spec : constant Bare_Param_Spec := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Param_Spec.Param_Spec_F_Ids;
                            return;
                        when 2 =>
                            Result := N_Bare_Param_Spec.Param_Spec_F_Has_Aliased;
                            return;
                        when 3 =>
                            Result := N_Bare_Param_Spec.Param_Spec_F_Mode;
                            return;
                        when 4 =>
                            Result := N_Bare_Param_Spec.Param_Spec_F_Type_Expr;
                            return;
                        when 5 =>
                            Result := N_Bare_Param_Spec.Param_Spec_F_Default_Expr;
                            return;
                        when 6 =>
                            Result := N_Bare_Param_Spec.Param_Spec_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Synthetic_Formal_Param_Decl_Range =>
declare
N_Bare_Synthetic_Formal_Param_Decl : constant Bare_Synthetic_Formal_Param_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Synthetic_Formal_Param_Decl.Synthetic_Formal_Param_Decl_F_Param_Type;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Base_Package_Decl =>
declare
N_Bare_Base_Package_Decl : constant Bare_Base_Package_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Base_Package_Decl.Base_Package_Decl_F_Package_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Base_Package_Decl.Base_Package_Decl_F_Aspects;
                            return;
                        when 3 =>
                            Result := N_Bare_Base_Package_Decl.Base_Package_Decl_F_Public_Part;
                            return;
                        when 4 =>
                            Result := N_Bare_Base_Package_Decl.Base_Package_Decl_F_Private_Part;
                            return;
                        when 5 =>
                            Result := N_Bare_Base_Package_Decl.Base_Package_Decl_F_End_Name;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Base_Type_Decl =>
declare
N_Bare_Base_Type_Decl : constant Bare_Base_Type_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Base_Type_Decl.Base_Type_Decl_F_Name;
                            return;
                        when others => null;
                    end case;
case Ada_Base_Type_Decl (K) is
when Ada_Subtype_Decl_Range =>
declare
N_Bare_Subtype_Decl : constant Bare_Subtype_Decl := N_Bare_Base_Type_Decl;
begin
case Index is
                        when 2 =>
                            Result := N_Bare_Subtype_Decl.Subtype_Decl_F_Subtype;
                            return;
                        when 3 =>
                            Result := N_Bare_Subtype_Decl.Subtype_Decl_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Incomplete_Type_Decl_Range =>
declare
N_Bare_Incomplete_Type_Decl : constant Bare_Incomplete_Type_Decl := N_Bare_Base_Type_Decl;
begin
case Index is
                        when 2 =>
                            Result := N_Bare_Incomplete_Type_Decl.Incomplete_Type_Decl_F_Discriminants;
                            return;
                        when others => null;
                    end case;
case Ada_Incomplete_Type_Decl_Range (K) is
when Ada_Incomplete_Tagged_Type_Decl_Range =>
declare
N_Bare_Incomplete_Tagged_Type_Decl : constant Bare_Incomplete_Tagged_Type_Decl := N_Bare_Incomplete_Type_Decl;
begin
case Index is
                        when 3 =>
                            Result := N_Bare_Incomplete_Tagged_Type_Decl.Incomplete_Tagged_Type_Decl_F_Has_Abstract;
                            return;
                        when others => null;
                    end case;
end;
when others => null;
end case;
end;
when Ada_Protected_Type_Decl_Range =>
declare
N_Bare_Protected_Type_Decl : constant Bare_Protected_Type_Decl := N_Bare_Base_Type_Decl;
begin
case Index is
                        when 2 =>
                            Result := N_Bare_Protected_Type_Decl.Protected_Type_Decl_F_Discriminants;
                            return;
                        when 3 =>
                            Result := N_Bare_Protected_Type_Decl.Protected_Type_Decl_F_Aspects;
                            return;
                        when 4 =>
                            Result := N_Bare_Protected_Type_Decl.Protected_Type_Decl_F_Interfaces;
                            return;
                        when 5 =>
                            Result := N_Bare_Protected_Type_Decl.Protected_Type_Decl_F_Definition;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Task_Type_Decl_Range =>
declare
N_Bare_Task_Type_Decl : constant Bare_Task_Type_Decl := N_Bare_Base_Type_Decl;
begin
case Index is
                        when 2 =>
                            Result := N_Bare_Task_Type_Decl.Task_Type_Decl_F_Discriminants;
                            return;
                        when 3 =>
                            Result := N_Bare_Task_Type_Decl.Task_Type_Decl_F_Aspects;
                            return;
                        when 4 =>
                            Result := N_Bare_Task_Type_Decl.Task_Type_Decl_F_Definition;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Type_Decl_Range =>
declare
N_Bare_Type_Decl : constant Bare_Type_Decl := N_Bare_Base_Type_Decl;
begin
case Index is
                        when 2 =>
                            Result := N_Bare_Type_Decl.Type_Decl_F_Discriminants;
                            return;
                        when 3 =>
                            Result := N_Bare_Type_Decl.Type_Decl_F_Type_Def;
                            return;
                        when 4 =>
                            Result := N_Bare_Type_Decl.Type_Decl_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when others => null;
end case;
end;
when Ada_Classic_Subp_Decl =>
declare
N_Bare_Classic_Subp_Decl : constant Bare_Classic_Subp_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Classic_Subp_Decl.Classic_Subp_Decl_F_Overriding;
                            return;
                        when 2 =>
                            Result := N_Bare_Classic_Subp_Decl.Classic_Subp_Decl_F_Subp_Spec;
                            return;
                        when others => null;
                    end case;
case Ada_Classic_Subp_Decl (K) is
when Ada_Abstract_Subp_Decl_Range =>
declare
N_Bare_Abstract_Subp_Decl : constant Bare_Abstract_Subp_Decl := N_Bare_Classic_Subp_Decl;
begin
case Index is
                        when 3 =>
                            Result := N_Bare_Abstract_Subp_Decl.Abstract_Subp_Decl_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Formal_Subp_Decl =>
declare
N_Bare_Formal_Subp_Decl : constant Bare_Formal_Subp_Decl := N_Bare_Classic_Subp_Decl;
begin
case Index is
                        when 3 =>
                            Result := N_Bare_Formal_Subp_Decl.Formal_Subp_Decl_F_Default_Expr;
                            return;
                        when 4 =>
                            Result := N_Bare_Formal_Subp_Decl.Formal_Subp_Decl_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Subp_Decl_Range =>
declare
N_Bare_Subp_Decl : constant Bare_Subp_Decl := N_Bare_Classic_Subp_Decl;
begin
case Index is
                        when 3 =>
                            Result := N_Bare_Subp_Decl.Subp_Decl_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when others => null;
end case;
end;
when Ada_Entry_Decl_Range =>
declare
N_Bare_Entry_Decl : constant Bare_Entry_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Entry_Decl.Entry_Decl_F_Overriding;
                            return;
                        when 2 =>
                            Result := N_Bare_Entry_Decl.Entry_Decl_F_Spec;
                            return;
                        when 3 =>
                            Result := N_Bare_Entry_Decl.Entry_Decl_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Enum_Literal_Decl_Range =>
declare
N_Bare_Enum_Literal_Decl : constant Bare_Enum_Literal_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Enum_Literal_Decl.Enum_Literal_Decl_F_Name;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Generic_Subp_Internal_Range =>
declare
N_Bare_Generic_Subp_Internal : constant Bare_Generic_Subp_Internal := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Generic_Subp_Internal.Generic_Subp_Internal_F_Subp_Spec;
                            return;
                        when 2 =>
                            Result := N_Bare_Generic_Subp_Internal.Generic_Subp_Internal_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Synthetic_Subp_Decl_Range =>
declare
N_Bare_Synthetic_Subp_Decl : constant Bare_Synthetic_Subp_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Synthetic_Subp_Decl.Synthetic_Subp_Decl_F_Spec;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Base_Subp_Body =>
declare
N_Bare_Base_Subp_Body : constant Bare_Base_Subp_Body := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Base_Subp_Body.Base_Subp_Body_F_Overriding;
                            return;
                        when 2 =>
                            Result := N_Bare_Base_Subp_Body.Base_Subp_Body_F_Subp_Spec;
                            return;
                        when others => null;
                    end case;
case Ada_Base_Subp_Body (K) is
when Ada_Expr_Function_Range =>
declare
N_Bare_Expr_Function : constant Bare_Expr_Function := N_Bare_Base_Subp_Body;
begin
case Index is
                        when 3 =>
                            Result := N_Bare_Expr_Function.Expr_Function_F_Expr;
                            return;
                        when 4 =>
                            Result := N_Bare_Expr_Function.Expr_Function_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Null_Subp_Decl_Range =>
declare
N_Bare_Null_Subp_Decl : constant Bare_Null_Subp_Decl := N_Bare_Base_Subp_Body;
begin
case Index is
                        when 3 =>
                            Result := N_Bare_Null_Subp_Decl.Null_Subp_Decl_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Subp_Body_Range =>
declare
N_Bare_Subp_Body : constant Bare_Subp_Body := N_Bare_Base_Subp_Body;
begin
case Index is
                        when 3 =>
                            Result := N_Bare_Subp_Body.Subp_Body_F_Aspects;
                            return;
                        when 4 =>
                            Result := N_Bare_Subp_Body.Subp_Body_F_Decls;
                            return;
                        when 5 =>
                            Result := N_Bare_Subp_Body.Subp_Body_F_Stmts;
                            return;
                        when 6 =>
                            Result := N_Bare_Subp_Body.Subp_Body_F_End_Name;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Subp_Renaming_Decl_Range =>
declare
N_Bare_Subp_Renaming_Decl : constant Bare_Subp_Renaming_Decl := N_Bare_Base_Subp_Body;
begin
case Index is
                        when 3 =>
                            Result := N_Bare_Subp_Renaming_Decl.Subp_Renaming_Decl_F_Renames;
                            return;
                        when 4 =>
                            Result := N_Bare_Subp_Renaming_Decl.Subp_Renaming_Decl_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when others => null;
end case;
end;
when Ada_Package_Body_Stub_Range =>
declare
N_Bare_Package_Body_Stub : constant Bare_Package_Body_Stub := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Package_Body_Stub.Package_Body_Stub_F_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Package_Body_Stub.Package_Body_Stub_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Protected_Body_Stub_Range =>
declare
N_Bare_Protected_Body_Stub : constant Bare_Protected_Body_Stub := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Protected_Body_Stub.Protected_Body_Stub_F_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Protected_Body_Stub.Protected_Body_Stub_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Subp_Body_Stub_Range =>
declare
N_Bare_Subp_Body_Stub : constant Bare_Subp_Body_Stub := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Subp_Body_Stub.Subp_Body_Stub_F_Overriding;
                            return;
                        when 2 =>
                            Result := N_Bare_Subp_Body_Stub.Subp_Body_Stub_F_Subp_Spec;
                            return;
                        when 3 =>
                            Result := N_Bare_Subp_Body_Stub.Subp_Body_Stub_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Task_Body_Stub_Range =>
declare
N_Bare_Task_Body_Stub : constant Bare_Task_Body_Stub := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Task_Body_Stub.Task_Body_Stub_F_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Task_Body_Stub.Task_Body_Stub_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Entry_Body_Range =>
declare
N_Bare_Entry_Body : constant Bare_Entry_Body := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Entry_Body.Entry_Body_F_Entry_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Entry_Body.Entry_Body_F_Index_Spec;
                            return;
                        when 3 =>
                            Result := N_Bare_Entry_Body.Entry_Body_F_Params;
                            return;
                        when 4 =>
                            Result := N_Bare_Entry_Body.Entry_Body_F_Aspects;
                            return;
                        when 5 =>
                            Result := N_Bare_Entry_Body.Entry_Body_F_Barrier;
                            return;
                        when 6 =>
                            Result := N_Bare_Entry_Body.Entry_Body_F_Decls;
                            return;
                        when 7 =>
                            Result := N_Bare_Entry_Body.Entry_Body_F_Stmts;
                            return;
                        when 8 =>
                            Result := N_Bare_Entry_Body.Entry_Body_F_End_Name;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Package_Body_Range =>
declare
N_Bare_Package_Body : constant Bare_Package_Body := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Package_Body.Package_Body_F_Package_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Package_Body.Package_Body_F_Aspects;
                            return;
                        when 3 =>
                            Result := N_Bare_Package_Body.Package_Body_F_Decls;
                            return;
                        when 4 =>
                            Result := N_Bare_Package_Body.Package_Body_F_Stmts;
                            return;
                        when 5 =>
                            Result := N_Bare_Package_Body.Package_Body_F_End_Name;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Protected_Body_Range =>
declare
N_Bare_Protected_Body : constant Bare_Protected_Body := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Protected_Body.Protected_Body_F_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Protected_Body.Protected_Body_F_Aspects;
                            return;
                        when 3 =>
                            Result := N_Bare_Protected_Body.Protected_Body_F_Decls;
                            return;
                        when 4 =>
                            Result := N_Bare_Protected_Body.Protected_Body_F_End_Name;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Task_Body_Range =>
declare
N_Bare_Task_Body : constant Bare_Task_Body := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Task_Body.Task_Body_F_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Task_Body.Task_Body_F_Aspects;
                            return;
                        when 3 =>
                            Result := N_Bare_Task_Body.Task_Body_F_Decls;
                            return;
                        when 4 =>
                            Result := N_Bare_Task_Body.Task_Body_F_Stmts;
                            return;
                        when 5 =>
                            Result := N_Bare_Task_Body.Task_Body_F_End_Name;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Entry_Index_Spec_Range =>
declare
N_Bare_Entry_Index_Spec : constant Bare_Entry_Index_Spec := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Entry_Index_Spec.Entry_Index_Spec_F_Id;
                            return;
                        when 2 =>
                            Result := N_Bare_Entry_Index_Spec.Entry_Index_Spec_F_Subtype;
                            return;
                        when 3 =>
                            Result := N_Bare_Entry_Index_Spec.Entry_Index_Spec_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Exception_Decl_Range =>
declare
N_Bare_Exception_Decl : constant Bare_Exception_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Exception_Decl.Exception_Decl_F_Ids;
                            return;
                        when 2 =>
                            Result := N_Bare_Exception_Decl.Exception_Decl_F_Renames;
                            return;
                        when 3 =>
                            Result := N_Bare_Exception_Decl.Exception_Decl_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Exception_Handler_Range =>
declare
N_Bare_Exception_Handler : constant Bare_Exception_Handler := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Exception_Handler.Exception_Handler_F_Exception_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Exception_Handler.Exception_Handler_F_Handled_Exceptions;
                            return;
                        when 3 =>
                            Result := N_Bare_Exception_Handler.Exception_Handler_F_Stmts;
                            return;
                        when others => null;
                    end case;
end;
when Ada_For_Loop_Var_Decl_Range =>
declare
N_Bare_For_Loop_Var_Decl : constant Bare_For_Loop_Var_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_For_Loop_Var_Decl.For_Loop_Var_Decl_F_Id;
                            return;
                        when 2 =>
                            Result := N_Bare_For_Loop_Var_Decl.For_Loop_Var_Decl_F_Id_Type;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Generic_Decl =>
declare
N_Bare_Generic_Decl : constant Bare_Generic_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Generic_Decl.Generic_Decl_F_Formal_Part;
                            return;
                        when others => null;
                    end case;
case Ada_Generic_Decl (K) is
when Ada_Generic_Package_Decl_Range =>
declare
N_Bare_Generic_Package_Decl : constant Bare_Generic_Package_Decl := N_Bare_Generic_Decl;
begin
case Index is
                        when 2 =>
                            Result := N_Bare_Generic_Package_Decl.Generic_Package_Decl_F_Package_Decl;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Generic_Subp_Decl_Range =>
declare
N_Bare_Generic_Subp_Decl : constant Bare_Generic_Subp_Decl := N_Bare_Generic_Decl;
begin
case Index is
                        when 2 =>
                            Result := N_Bare_Generic_Subp_Decl.Generic_Subp_Decl_F_Subp_Decl;
                            return;
                        when others => null;
                    end case;
end;
when others => null;
end case;
end;
when Ada_Generic_Package_Instantiation_Range =>
declare
N_Bare_Generic_Package_Instantiation : constant Bare_Generic_Package_Instantiation := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Generic_Package_Instantiation.Generic_Package_Instantiation_F_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Generic_Package_Instantiation.Generic_Package_Instantiation_F_Generic_Pkg_Name;
                            return;
                        when 3 =>
                            Result := N_Bare_Generic_Package_Instantiation.Generic_Package_Instantiation_F_Params;
                            return;
                        when 4 =>
                            Result := N_Bare_Generic_Package_Instantiation.Generic_Package_Instantiation_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Generic_Subp_Instantiation_Range =>
declare
N_Bare_Generic_Subp_Instantiation : constant Bare_Generic_Subp_Instantiation := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Generic_Subp_Instantiation.Generic_Subp_Instantiation_F_Overriding;
                            return;
                        when 2 =>
                            Result := N_Bare_Generic_Subp_Instantiation.Generic_Subp_Instantiation_F_Kind;
                            return;
                        when 3 =>
                            Result := N_Bare_Generic_Subp_Instantiation.Generic_Subp_Instantiation_F_Subp_Name;
                            return;
                        when 4 =>
                            Result := N_Bare_Generic_Subp_Instantiation.Generic_Subp_Instantiation_F_Generic_Subp_Name;
                            return;
                        when 5 =>
                            Result := N_Bare_Generic_Subp_Instantiation.Generic_Subp_Instantiation_F_Params;
                            return;
                        when 6 =>
                            Result := N_Bare_Generic_Subp_Instantiation.Generic_Subp_Instantiation_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Generic_Package_Renaming_Decl_Range =>
declare
N_Bare_Generic_Package_Renaming_Decl : constant Bare_Generic_Package_Renaming_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Generic_Package_Renaming_Decl.Generic_Package_Renaming_Decl_F_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Generic_Package_Renaming_Decl.Generic_Package_Renaming_Decl_F_Renames;
                            return;
                        when 3 =>
                            Result := N_Bare_Generic_Package_Renaming_Decl.Generic_Package_Renaming_Decl_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Generic_Subp_Renaming_Decl_Range =>
declare
N_Bare_Generic_Subp_Renaming_Decl : constant Bare_Generic_Subp_Renaming_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Generic_Subp_Renaming_Decl.Generic_Subp_Renaming_Decl_F_Kind;
                            return;
                        when 2 =>
                            Result := N_Bare_Generic_Subp_Renaming_Decl.Generic_Subp_Renaming_Decl_F_Name;
                            return;
                        when 3 =>
                            Result := N_Bare_Generic_Subp_Renaming_Decl.Generic_Subp_Renaming_Decl_F_Renames;
                            return;
                        when 4 =>
                            Result := N_Bare_Generic_Subp_Renaming_Decl.Generic_Subp_Renaming_Decl_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Label_Decl_Range =>
declare
N_Bare_Label_Decl : constant Bare_Label_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Label_Decl.Label_Decl_F_Name;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Named_Stmt_Decl_Range =>
declare
N_Bare_Named_Stmt_Decl : constant Bare_Named_Stmt_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Named_Stmt_Decl.Named_Stmt_Decl_F_Name;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Number_Decl_Range =>
declare
N_Bare_Number_Decl : constant Bare_Number_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Number_Decl.Number_Decl_F_Ids;
                            return;
                        when 2 =>
                            Result := N_Bare_Number_Decl.Number_Decl_F_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Object_Decl_Range =>
declare
N_Bare_Object_Decl : constant Bare_Object_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Object_Decl.Object_Decl_F_Ids;
                            return;
                        when 2 =>
                            Result := N_Bare_Object_Decl.Object_Decl_F_Has_Aliased;
                            return;
                        when 3 =>
                            Result := N_Bare_Object_Decl.Object_Decl_F_Has_Constant;
                            return;
                        when 4 =>
                            Result := N_Bare_Object_Decl.Object_Decl_F_Mode;
                            return;
                        when 5 =>
                            Result := N_Bare_Object_Decl.Object_Decl_F_Type_Expr;
                            return;
                        when 6 =>
                            Result := N_Bare_Object_Decl.Object_Decl_F_Default_Expr;
                            return;
                        when 7 =>
                            Result := N_Bare_Object_Decl.Object_Decl_F_Renaming_Clause;
                            return;
                        when 8 =>
                            Result := N_Bare_Object_Decl.Object_Decl_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Package_Renaming_Decl_Range =>
declare
N_Bare_Package_Renaming_Decl : constant Bare_Package_Renaming_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Package_Renaming_Decl.Package_Renaming_Decl_F_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Package_Renaming_Decl.Package_Renaming_Decl_F_Renames;
                            return;
                        when 3 =>
                            Result := N_Bare_Package_Renaming_Decl.Package_Renaming_Decl_F_Aspects;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Single_Protected_Decl_Range =>
declare
N_Bare_Single_Protected_Decl : constant Bare_Single_Protected_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Single_Protected_Decl.Single_Protected_Decl_F_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Single_Protected_Decl.Single_Protected_Decl_F_Aspects;
                            return;
                        when 3 =>
                            Result := N_Bare_Single_Protected_Decl.Single_Protected_Decl_F_Interfaces;
                            return;
                        when 4 =>
                            Result := N_Bare_Single_Protected_Decl.Single_Protected_Decl_F_Definition;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Single_Task_Decl_Range =>
declare
N_Bare_Single_Task_Decl : constant Bare_Single_Task_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Single_Task_Decl.Single_Task_Decl_F_Task_Type;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Case_Stmt_Alternative_Range =>
declare
N_Bare_Case_Stmt_Alternative : constant Bare_Case_Stmt_Alternative := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Case_Stmt_Alternative.Case_Stmt_Alternative_F_Choices;
                            return;
                        when 2 =>
                            Result := N_Bare_Case_Stmt_Alternative.Case_Stmt_Alternative_F_Stmts;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Compilation_Unit_Range =>
declare
N_Bare_Compilation_Unit : constant Bare_Compilation_Unit := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Compilation_Unit.Compilation_Unit_F_Prelude;
                            return;
                        when 2 =>
                            Result := N_Bare_Compilation_Unit.Compilation_Unit_F_Body;
                            return;
                        when 3 =>
                            Result := N_Bare_Compilation_Unit.Compilation_Unit_F_Pragmas;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Component_Clause_Range =>
declare
N_Bare_Component_Clause : constant Bare_Component_Clause := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Component_Clause.Component_Clause_F_Id;
                            return;
                        when 2 =>
                            Result := N_Bare_Component_Clause.Component_Clause_F_Position;
                            return;
                        when 3 =>
                            Result := N_Bare_Component_Clause.Component_Clause_F_Range;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Component_Def_Range =>
declare
N_Bare_Component_Def : constant Bare_Component_Def := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Component_Def.Component_Def_F_Has_Aliased;
                            return;
                        when 2 =>
                            Result := N_Bare_Component_Def.Component_Def_F_Has_Constant;
                            return;
                        when 3 =>
                            Result := N_Bare_Component_Def.Component_Def_F_Type_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Composite_Constraint_Range =>
declare
N_Bare_Composite_Constraint : constant Bare_Composite_Constraint := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Composite_Constraint.Composite_Constraint_F_Constraints;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Delta_Constraint_Range =>
declare
N_Bare_Delta_Constraint : constant Bare_Delta_Constraint := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Delta_Constraint.Delta_Constraint_F_Digits;
                            return;
                        when 2 =>
                            Result := N_Bare_Delta_Constraint.Delta_Constraint_F_Range;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Digits_Constraint_Range =>
declare
N_Bare_Digits_Constraint : constant Bare_Digits_Constraint := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Digits_Constraint.Digits_Constraint_F_Digits;
                            return;
                        when 2 =>
                            Result := N_Bare_Digits_Constraint.Digits_Constraint_F_Range;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Range_Constraint_Range =>
declare
N_Bare_Range_Constraint : constant Bare_Range_Constraint := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Range_Constraint.Range_Constraint_F_Range;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Declarative_Part_Range =>
declare
N_Bare_Declarative_Part : constant Bare_Declarative_Part := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Declarative_Part.Declarative_Part_F_Decls;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Elsif_Expr_Part_Range =>
declare
N_Bare_Elsif_Expr_Part : constant Bare_Elsif_Expr_Part := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Elsif_Expr_Part.Elsif_Expr_Part_F_Cond_Expr;
                            return;
                        when 2 =>
                            Result := N_Bare_Elsif_Expr_Part.Elsif_Expr_Part_F_Then_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Elsif_Stmt_Part_Range =>
declare
N_Bare_Elsif_Stmt_Part : constant Bare_Elsif_Stmt_Part := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Elsif_Stmt_Part.Elsif_Stmt_Part_F_Cond_Expr;
                            return;
                        when 2 =>
                            Result := N_Bare_Elsif_Stmt_Part.Elsif_Stmt_Part_F_Stmts;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Abstract_State_Decl_Expr_Range =>
declare
N_Bare_Abstract_State_Decl_Expr : constant Bare_Abstract_State_Decl_Expr := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Abstract_State_Decl_Expr.Abstract_State_Decl_Expr_F_State_Decl;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Allocator_Range =>
declare
N_Bare_Allocator : constant Bare_Allocator := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Allocator.Allocator_F_Subpool;
                            return;
                        when 2 =>
                            Result := N_Bare_Allocator.Allocator_F_Type_Or_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Base_Aggregate =>
declare
N_Bare_Base_Aggregate : constant Bare_Base_Aggregate := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Base_Aggregate.Base_Aggregate_F_Ancestor_Expr;
                            return;
                        when 2 =>
                            Result := N_Bare_Base_Aggregate.Base_Aggregate_F_Assocs;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Bin_Op_Range =>
declare
N_Bare_Bin_Op : constant Bare_Bin_Op := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Bin_Op.Bin_Op_F_Left;
                            return;
                        when 2 =>
                            Result := N_Bare_Bin_Op.Bin_Op_F_Op;
                            return;
                        when 3 =>
                            Result := N_Bare_Bin_Op.Bin_Op_F_Right;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Case_Expr_Alternative_Range =>
declare
N_Bare_Case_Expr_Alternative : constant Bare_Case_Expr_Alternative := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Case_Expr_Alternative.Case_Expr_Alternative_F_Choices;
                            return;
                        when 2 =>
                            Result := N_Bare_Case_Expr_Alternative.Case_Expr_Alternative_F_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Case_Expr_Range =>
declare
N_Bare_Case_Expr : constant Bare_Case_Expr := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Case_Expr.Case_Expr_F_Expr;
                            return;
                        when 2 =>
                            Result := N_Bare_Case_Expr.Case_Expr_F_Cases;
                            return;
                        when others => null;
                    end case;
end;
when Ada_If_Expr_Range =>
declare
N_Bare_If_Expr : constant Bare_If_Expr := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_If_Expr.If_Expr_F_Cond_Expr;
                            return;
                        when 2 =>
                            Result := N_Bare_If_Expr.If_Expr_F_Then_Expr;
                            return;
                        when 3 =>
                            Result := N_Bare_If_Expr.If_Expr_F_Alternatives;
                            return;
                        when 4 =>
                            Result := N_Bare_If_Expr.If_Expr_F_Else_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Contract_Cases_Range =>
declare
N_Bare_Contract_Cases : constant Bare_Contract_Cases := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Contract_Cases.Contract_Cases_F_Contract_Cases;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Decl_Expr_Range =>
declare
N_Bare_Decl_Expr : constant Bare_Decl_Expr := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Decl_Expr.Decl_Expr_F_Decls;
                            return;
                        when 2 =>
                            Result := N_Bare_Decl_Expr.Decl_Expr_F_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Membership_Expr_Range =>
declare
N_Bare_Membership_Expr : constant Bare_Membership_Expr := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Membership_Expr.Membership_Expr_F_Expr;
                            return;
                        when 2 =>
                            Result := N_Bare_Membership_Expr.Membership_Expr_F_Op;
                            return;
                        when 3 =>
                            Result := N_Bare_Membership_Expr.Membership_Expr_F_Membership_Exprs;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Attribute_Ref_Range =>
declare
N_Bare_Attribute_Ref : constant Bare_Attribute_Ref := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Attribute_Ref.Attribute_Ref_F_Prefix;
                            return;
                        when 2 =>
                            Result := N_Bare_Attribute_Ref.Attribute_Ref_F_Attribute;
                            return;
                        when 3 =>
                            Result := N_Bare_Attribute_Ref.Attribute_Ref_F_Args;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Call_Expr_Range =>
declare
N_Bare_Call_Expr : constant Bare_Call_Expr := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Call_Expr.Call_Expr_F_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Call_Expr.Call_Expr_F_Suffix;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Defining_Name_Range =>
declare
N_Bare_Defining_Name : constant Bare_Defining_Name := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Defining_Name.Defining_Name_F_Name;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Discrete_Subtype_Name_Range =>
declare
N_Bare_Discrete_Subtype_Name : constant Bare_Discrete_Subtype_Name := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Discrete_Subtype_Name.Discrete_Subtype_Name_F_Subtype;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Dotted_Name_Range =>
declare
N_Bare_Dotted_Name : constant Bare_Dotted_Name := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Dotted_Name.Dotted_Name_F_Prefix;
                            return;
                        when 2 =>
                            Result := N_Bare_Dotted_Name.Dotted_Name_F_Suffix;
                            return;
                        when others => null;
                    end case;
end;
when Ada_End_Name_Range =>
declare
N_Bare_End_Name : constant Bare_End_Name := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_End_Name.End_Name_F_Name;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Explicit_Deref_Range =>
declare
N_Bare_Explicit_Deref : constant Bare_Explicit_Deref := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Explicit_Deref.Explicit_Deref_F_Prefix;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Qual_Expr_Range =>
declare
N_Bare_Qual_Expr : constant Bare_Qual_Expr := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Qual_Expr.Qual_Expr_F_Prefix;
                            return;
                        when 2 =>
                            Result := N_Bare_Qual_Expr.Qual_Expr_F_Suffix;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Reduce_Attribute_Ref_Range =>
declare
N_Bare_Reduce_Attribute_Ref : constant Bare_Reduce_Attribute_Ref := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Reduce_Attribute_Ref.Reduce_Attribute_Ref_F_Prefix;
                            return;
                        when 2 =>
                            Result := N_Bare_Reduce_Attribute_Ref.Reduce_Attribute_Ref_F_Attribute;
                            return;
                        when 3 =>
                            Result := N_Bare_Reduce_Attribute_Ref.Reduce_Attribute_Ref_F_Args;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Paren_Expr_Range =>
declare
N_Bare_Paren_Expr : constant Bare_Paren_Expr := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Paren_Expr.Paren_Expr_F_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Quantified_Expr_Range =>
declare
N_Bare_Quantified_Expr : constant Bare_Quantified_Expr := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Quantified_Expr.Quantified_Expr_F_Quantifier;
                            return;
                        when 2 =>
                            Result := N_Bare_Quantified_Expr.Quantified_Expr_F_Loop_Spec;
                            return;
                        when 3 =>
                            Result := N_Bare_Quantified_Expr.Quantified_Expr_F_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Raise_Expr_Range =>
declare
N_Bare_Raise_Expr : constant Bare_Raise_Expr := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Raise_Expr.Raise_Expr_F_Exception_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Raise_Expr.Raise_Expr_F_Error_Message;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Un_Op_Range =>
declare
N_Bare_Un_Op : constant Bare_Un_Op := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Un_Op.Un_Op_F_Op;
                            return;
                        when 2 =>
                            Result := N_Bare_Un_Op.Un_Op_F_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Handled_Stmts_Range =>
declare
N_Bare_Handled_Stmts : constant Bare_Handled_Stmts := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Handled_Stmts.Handled_Stmts_F_Stmts;
                            return;
                        when 2 =>
                            Result := N_Bare_Handled_Stmts.Handled_Stmts_F_Exceptions;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Library_Item_Range =>
declare
N_Bare_Library_Item : constant Bare_Library_Item := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Library_Item.Library_Item_F_Has_Private;
                            return;
                        when 2 =>
                            Result := N_Bare_Library_Item.Library_Item_F_Item;
                            return;
                        when others => null;
                    end case;
end;
when Ada_For_Loop_Spec_Range =>
declare
N_Bare_For_Loop_Spec : constant Bare_For_Loop_Spec := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_For_Loop_Spec.For_Loop_Spec_F_Var_Decl;
                            return;
                        when 2 =>
                            Result := N_Bare_For_Loop_Spec.For_Loop_Spec_F_Loop_Type;
                            return;
                        when 3 =>
                            Result := N_Bare_For_Loop_Spec.For_Loop_Spec_F_Has_Reverse;
                            return;
                        when 4 =>
                            Result := N_Bare_For_Loop_Spec.For_Loop_Spec_F_Iter_Expr;
                            return;
                        when 5 =>
                            Result := N_Bare_For_Loop_Spec.For_Loop_Spec_F_Iter_Filter;
                            return;
                        when others => null;
                    end case;
end;
when Ada_While_Loop_Spec_Range =>
declare
N_Bare_While_Loop_Spec : constant Bare_While_Loop_Spec := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_While_Loop_Spec.While_Loop_Spec_F_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Multi_Abstract_State_Decl_Range =>
declare
N_Bare_Multi_Abstract_State_Decl : constant Bare_Multi_Abstract_State_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Multi_Abstract_State_Decl.Multi_Abstract_State_Decl_F_Decls;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Params_Range =>
declare
N_Bare_Params : constant Bare_Params := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Params.Params_F_Params;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Paren_Abstract_State_Decl_Range =>
declare
N_Bare_Paren_Abstract_State_Decl : constant Bare_Paren_Abstract_State_Decl := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Paren_Abstract_State_Decl.Paren_Abstract_State_Decl_F_Decl;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Pp_Elsif_Directive_Range =>
declare
N_Bare_Pp_Elsif_Directive : constant Bare_Pp_Elsif_Directive := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Pp_Elsif_Directive.Pp_Elsif_Directive_F_Expr;
                            return;
                        when 2 =>
                            Result := N_Bare_Pp_Elsif_Directive.Pp_Elsif_Directive_F_Then_Kw;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Pp_If_Directive_Range =>
declare
N_Bare_Pp_If_Directive : constant Bare_Pp_If_Directive := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Pp_If_Directive.Pp_If_Directive_F_Expr;
                            return;
                        when 2 =>
                            Result := N_Bare_Pp_If_Directive.Pp_If_Directive_F_Then_Kw;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Pragma_Node_Range =>
declare
N_Bare_Pragma_Node : constant Bare_Pragma_Node := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Pragma_Node.Pragma_Node_F_Id;
                            return;
                        when 2 =>
                            Result := N_Bare_Pragma_Node.Pragma_Node_F_Args;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Protected_Def_Range =>
declare
N_Bare_Protected_Def : constant Bare_Protected_Def := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Protected_Def.Protected_Def_F_Public_Part;
                            return;
                        when 2 =>
                            Result := N_Bare_Protected_Def.Protected_Def_F_Private_Part;
                            return;
                        when 3 =>
                            Result := N_Bare_Protected_Def.Protected_Def_F_End_Name;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Range_Spec_Range =>
declare
N_Bare_Range_Spec : constant Bare_Range_Spec := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Range_Spec.Range_Spec_F_Range;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Renaming_Clause_Range =>
declare
N_Bare_Renaming_Clause : constant Bare_Renaming_Clause := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Renaming_Clause.Renaming_Clause_F_Renamed_Object;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Select_When_Part_Range =>
declare
N_Bare_Select_When_Part : constant Bare_Select_When_Part := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Select_When_Part.Select_When_Part_F_Cond_Expr;
                            return;
                        when 2 =>
                            Result := N_Bare_Select_When_Part.Select_When_Part_F_Stmts;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Accept_Stmt_Range =>
declare
N_Bare_Accept_Stmt : constant Bare_Accept_Stmt := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Accept_Stmt.Accept_Stmt_F_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Accept_Stmt.Accept_Stmt_F_Entry_Index_Expr;
                            return;
                        when 3 =>
                            Result := N_Bare_Accept_Stmt.Accept_Stmt_F_Params;
                            return;
                        when others => null;
                    end case;
case Ada_Accept_Stmt_Range (K) is
when Ada_Accept_Stmt_With_Stmts_Range =>
declare
N_Bare_Accept_Stmt_With_Stmts : constant Bare_Accept_Stmt_With_Stmts := N_Bare_Accept_Stmt;
begin
case Index is
                        when 4 =>
                            Result := N_Bare_Accept_Stmt_With_Stmts.Accept_Stmt_With_Stmts_F_Stmts;
                            return;
                        when 5 =>
                            Result := N_Bare_Accept_Stmt_With_Stmts.Accept_Stmt_With_Stmts_F_End_Name;
                            return;
                        when others => null;
                    end case;
end;
when others => null;
end case;
end;
when Ada_Base_Loop_Stmt =>
declare
N_Bare_Base_Loop_Stmt : constant Bare_Base_Loop_Stmt := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Base_Loop_Stmt.Base_Loop_Stmt_F_Spec;
                            return;
                        when 2 =>
                            Result := N_Bare_Base_Loop_Stmt.Base_Loop_Stmt_F_Stmts;
                            return;
                        when 3 =>
                            Result := N_Bare_Base_Loop_Stmt.Base_Loop_Stmt_F_End_Name;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Begin_Block_Range =>
declare
N_Bare_Begin_Block : constant Bare_Begin_Block := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Begin_Block.Begin_Block_F_Stmts;
                            return;
                        when 2 =>
                            Result := N_Bare_Begin_Block.Begin_Block_F_End_Name;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Decl_Block_Range =>
declare
N_Bare_Decl_Block : constant Bare_Decl_Block := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Decl_Block.Decl_Block_F_Decls;
                            return;
                        when 2 =>
                            Result := N_Bare_Decl_Block.Decl_Block_F_Stmts;
                            return;
                        when 3 =>
                            Result := N_Bare_Decl_Block.Decl_Block_F_End_Name;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Case_Stmt_Range =>
declare
N_Bare_Case_Stmt : constant Bare_Case_Stmt := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Case_Stmt.Case_Stmt_F_Expr;
                            return;
                        when 2 =>
                            Result := N_Bare_Case_Stmt.Case_Stmt_F_Pragmas;
                            return;
                        when 3 =>
                            Result := N_Bare_Case_Stmt.Case_Stmt_F_Alternatives;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Extended_Return_Stmt_Range =>
declare
N_Bare_Extended_Return_Stmt : constant Bare_Extended_Return_Stmt := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Extended_Return_Stmt.Extended_Return_Stmt_F_Decl;
                            return;
                        when 2 =>
                            Result := N_Bare_Extended_Return_Stmt.Extended_Return_Stmt_F_Stmts;
                            return;
                        when others => null;
                    end case;
end;
when Ada_If_Stmt_Range =>
declare
N_Bare_If_Stmt : constant Bare_If_Stmt := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_If_Stmt.If_Stmt_F_Cond_Expr;
                            return;
                        when 2 =>
                            Result := N_Bare_If_Stmt.If_Stmt_F_Then_Stmts;
                            return;
                        when 3 =>
                            Result := N_Bare_If_Stmt.If_Stmt_F_Alternatives;
                            return;
                        when 4 =>
                            Result := N_Bare_If_Stmt.If_Stmt_F_Else_Stmts;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Named_Stmt_Range =>
declare
N_Bare_Named_Stmt : constant Bare_Named_Stmt := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Named_Stmt.Named_Stmt_F_Decl;
                            return;
                        when 2 =>
                            Result := N_Bare_Named_Stmt.Named_Stmt_F_Stmt;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Select_Stmt_Range =>
declare
N_Bare_Select_Stmt : constant Bare_Select_Stmt := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Select_Stmt.Select_Stmt_F_Guards;
                            return;
                        when 2 =>
                            Result := N_Bare_Select_Stmt.Select_Stmt_F_Else_Stmts;
                            return;
                        when 3 =>
                            Result := N_Bare_Select_Stmt.Select_Stmt_F_Abort_Stmts;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Abort_Stmt_Range =>
declare
N_Bare_Abort_Stmt : constant Bare_Abort_Stmt := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Abort_Stmt.Abort_Stmt_F_Names;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Assign_Stmt_Range =>
declare
N_Bare_Assign_Stmt : constant Bare_Assign_Stmt := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Assign_Stmt.Assign_Stmt_F_Dest;
                            return;
                        when 2 =>
                            Result := N_Bare_Assign_Stmt.Assign_Stmt_F_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Call_Stmt_Range =>
declare
N_Bare_Call_Stmt : constant Bare_Call_Stmt := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Call_Stmt.Call_Stmt_F_Call;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Delay_Stmt_Range =>
declare
N_Bare_Delay_Stmt : constant Bare_Delay_Stmt := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Delay_Stmt.Delay_Stmt_F_Has_Until;
                            return;
                        when 2 =>
                            Result := N_Bare_Delay_Stmt.Delay_Stmt_F_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Exit_Stmt_Range =>
declare
N_Bare_Exit_Stmt : constant Bare_Exit_Stmt := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Exit_Stmt.Exit_Stmt_F_Loop_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Exit_Stmt.Exit_Stmt_F_Cond_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Goto_Stmt_Range =>
declare
N_Bare_Goto_Stmt : constant Bare_Goto_Stmt := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Goto_Stmt.Goto_Stmt_F_Label_Name;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Label_Range =>
declare
N_Bare_Label : constant Bare_Label := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Label.Label_F_Decl;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Raise_Stmt_Range =>
declare
N_Bare_Raise_Stmt : constant Bare_Raise_Stmt := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Raise_Stmt.Raise_Stmt_F_Exception_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Raise_Stmt.Raise_Stmt_F_Error_Message;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Requeue_Stmt_Range =>
declare
N_Bare_Requeue_Stmt : constant Bare_Requeue_Stmt := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Requeue_Stmt.Requeue_Stmt_F_Call_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Requeue_Stmt.Requeue_Stmt_F_Has_Abort;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Return_Stmt_Range =>
declare
N_Bare_Return_Stmt : constant Bare_Return_Stmt := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Return_Stmt.Return_Stmt_F_Return_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Subunit_Range =>
declare
N_Bare_Subunit : constant Bare_Subunit := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Subunit.Subunit_F_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Subunit.Subunit_F_Body;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Task_Def_Range =>
declare
N_Bare_Task_Def : constant Bare_Task_Def := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Task_Def.Task_Def_F_Interfaces;
                            return;
                        when 2 =>
                            Result := N_Bare_Task_Def.Task_Def_F_Public_Part;
                            return;
                        when 3 =>
                            Result := N_Bare_Task_Def.Task_Def_F_Private_Part;
                            return;
                        when 4 =>
                            Result := N_Bare_Task_Def.Task_Def_F_End_Name;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Access_Def =>
declare
N_Bare_Access_Def : constant Bare_Access_Def := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Access_Def.Access_Def_F_Has_Not_Null;
                            return;
                        when others => null;
                    end case;
case Ada_Access_Def (K) is
when Ada_Access_To_Subp_Def_Range =>
declare
N_Bare_Access_To_Subp_Def : constant Bare_Access_To_Subp_Def := N_Bare_Access_Def;
begin
case Index is
                        when 2 =>
                            Result := N_Bare_Access_To_Subp_Def.Access_To_Subp_Def_F_Has_Protected;
                            return;
                        when 3 =>
                            Result := N_Bare_Access_To_Subp_Def.Access_To_Subp_Def_F_Subp_Spec;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Anonymous_Type_Access_Def_Range =>
declare
N_Bare_Anonymous_Type_Access_Def : constant Bare_Anonymous_Type_Access_Def := N_Bare_Access_Def;
begin
case Index is
                        when 2 =>
                            Result := N_Bare_Anonymous_Type_Access_Def.Anonymous_Type_Access_Def_F_Type_Decl;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Type_Access_Def_Range =>
declare
N_Bare_Type_Access_Def : constant Bare_Type_Access_Def := N_Bare_Access_Def;
begin
case Index is
                        when 2 =>
                            Result := N_Bare_Type_Access_Def.Type_Access_Def_F_Has_All;
                            return;
                        when 3 =>
                            Result := N_Bare_Type_Access_Def.Type_Access_Def_F_Has_Constant;
                            return;
                        when 4 =>
                            Result := N_Bare_Type_Access_Def.Type_Access_Def_F_Subtype_Indication;
                            return;
                        when others => null;
                    end case;
end;
when others => null;
end case;
end;
when Ada_Array_Type_Def_Range =>
declare
N_Bare_Array_Type_Def : constant Bare_Array_Type_Def := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Array_Type_Def.Array_Type_Def_F_Indices;
                            return;
                        when 2 =>
                            Result := N_Bare_Array_Type_Def.Array_Type_Def_F_Component_Type;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Derived_Type_Def_Range =>
declare
N_Bare_Derived_Type_Def : constant Bare_Derived_Type_Def := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Derived_Type_Def.Derived_Type_Def_F_Has_Abstract;
                            return;
                        when 2 =>
                            Result := N_Bare_Derived_Type_Def.Derived_Type_Def_F_Has_Limited;
                            return;
                        when 3 =>
                            Result := N_Bare_Derived_Type_Def.Derived_Type_Def_F_Has_Synchronized;
                            return;
                        when 4 =>
                            Result := N_Bare_Derived_Type_Def.Derived_Type_Def_F_Subtype_Indication;
                            return;
                        when 5 =>
                            Result := N_Bare_Derived_Type_Def.Derived_Type_Def_F_Interfaces;
                            return;
                        when 6 =>
                            Result := N_Bare_Derived_Type_Def.Derived_Type_Def_F_Record_Extension;
                            return;
                        when 7 =>
                            Result := N_Bare_Derived_Type_Def.Derived_Type_Def_F_Has_With_Private;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Enum_Type_Def_Range =>
declare
N_Bare_Enum_Type_Def : constant Bare_Enum_Type_Def := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Enum_Type_Def.Enum_Type_Def_F_Enum_Literals;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Interface_Type_Def_Range =>
declare
N_Bare_Interface_Type_Def : constant Bare_Interface_Type_Def := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Interface_Type_Def.Interface_Type_Def_F_Interface_Kind;
                            return;
                        when 2 =>
                            Result := N_Bare_Interface_Type_Def.Interface_Type_Def_F_Interfaces;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Mod_Int_Type_Def_Range =>
declare
N_Bare_Mod_Int_Type_Def : constant Bare_Mod_Int_Type_Def := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Mod_Int_Type_Def.Mod_Int_Type_Def_F_Expr;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Private_Type_Def_Range =>
declare
N_Bare_Private_Type_Def : constant Bare_Private_Type_Def := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Private_Type_Def.Private_Type_Def_F_Has_Abstract;
                            return;
                        when 2 =>
                            Result := N_Bare_Private_Type_Def.Private_Type_Def_F_Has_Tagged;
                            return;
                        when 3 =>
                            Result := N_Bare_Private_Type_Def.Private_Type_Def_F_Has_Limited;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Decimal_Fixed_Point_Def_Range =>
declare
N_Bare_Decimal_Fixed_Point_Def : constant Bare_Decimal_Fixed_Point_Def := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Decimal_Fixed_Point_Def.Decimal_Fixed_Point_Def_F_Delta;
                            return;
                        when 2 =>
                            Result := N_Bare_Decimal_Fixed_Point_Def.Decimal_Fixed_Point_Def_F_Digits;
                            return;
                        when 3 =>
                            Result := N_Bare_Decimal_Fixed_Point_Def.Decimal_Fixed_Point_Def_F_Range;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Floating_Point_Def_Range =>
declare
N_Bare_Floating_Point_Def : constant Bare_Floating_Point_Def := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Floating_Point_Def.Floating_Point_Def_F_Num_Digits;
                            return;
                        when 2 =>
                            Result := N_Bare_Floating_Point_Def.Floating_Point_Def_F_Range;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Ordinary_Fixed_Point_Def_Range =>
declare
N_Bare_Ordinary_Fixed_Point_Def : constant Bare_Ordinary_Fixed_Point_Def := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Ordinary_Fixed_Point_Def.Ordinary_Fixed_Point_Def_F_Delta;
                            return;
                        when 2 =>
                            Result := N_Bare_Ordinary_Fixed_Point_Def.Ordinary_Fixed_Point_Def_F_Range;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Record_Type_Def_Range =>
declare
N_Bare_Record_Type_Def : constant Bare_Record_Type_Def := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Record_Type_Def.Record_Type_Def_F_Has_Abstract;
                            return;
                        when 2 =>
                            Result := N_Bare_Record_Type_Def.Record_Type_Def_F_Has_Tagged;
                            return;
                        when 3 =>
                            Result := N_Bare_Record_Type_Def.Record_Type_Def_F_Has_Limited;
                            return;
                        when 4 =>
                            Result := N_Bare_Record_Type_Def.Record_Type_Def_F_Record_Def;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Signed_Int_Type_Def_Range =>
declare
N_Bare_Signed_Int_Type_Def : constant Bare_Signed_Int_Type_Def := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Signed_Int_Type_Def.Signed_Int_Type_Def_F_Range;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Anonymous_Type_Range =>
declare
N_Bare_Anonymous_Type : constant Bare_Anonymous_Type := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Anonymous_Type.Anonymous_Type_F_Type_Decl;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Subtype_Indication_Range =>
declare
N_Bare_Subtype_Indication : constant Bare_Subtype_Indication := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Subtype_Indication.Subtype_Indication_F_Has_Not_Null;
                            return;
                        when 2 =>
                            Result := N_Bare_Subtype_Indication.Subtype_Indication_F_Name;
                            return;
                        when 3 =>
                            Result := N_Bare_Subtype_Indication.Subtype_Indication_F_Constraint;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Synthetic_Type_Expr_Range =>
declare
N_Bare_Synthetic_Type_Expr : constant Bare_Synthetic_Type_Expr := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Synthetic_Type_Expr.Synthetic_Type_Expr_F_Target_Type;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Unconstrained_Array_Index_Range =>
declare
N_Bare_Unconstrained_Array_Index : constant Bare_Unconstrained_Array_Index := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Unconstrained_Array_Index.Unconstrained_Array_Index_F_Subtype_Indication;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Use_Package_Clause_Range =>
declare
N_Bare_Use_Package_Clause : constant Bare_Use_Package_Clause := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Use_Package_Clause.Use_Package_Clause_F_Packages;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Use_Type_Clause_Range =>
declare
N_Bare_Use_Type_Clause : constant Bare_Use_Type_Clause := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Use_Type_Clause.Use_Type_Clause_F_Has_All;
                            return;
                        when 2 =>
                            Result := N_Bare_Use_Type_Clause.Use_Type_Clause_F_Types;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Value_Sequence_Range =>
declare
N_Bare_Value_Sequence : constant Bare_Value_Sequence := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Value_Sequence.Value_Sequence_F_Iter_Assoc;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Variant_Range =>
declare
N_Bare_Variant : constant Bare_Variant := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Variant.Variant_F_Choices;
                            return;
                        when 2 =>
                            Result := N_Bare_Variant.Variant_F_Components;
                            return;
                        when others => null;
                    end case;
end;
when Ada_Variant_Part_Range =>
declare
N_Bare_Variant_Part : constant Bare_Variant_Part := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_Variant_Part.Variant_Part_F_Discr_Name;
                            return;
                        when 2 =>
                            Result := N_Bare_Variant_Part.Variant_Part_F_Variant;
                            return;
                        when others => null;
                    end case;
end;
when Ada_With_Clause_Range =>
declare
N_Bare_With_Clause : constant Bare_With_Clause := Node;
begin
case Index is
                        when 1 =>
                            Result := N_Bare_With_Clause.With_Clause_F_Has_Limited;
                            return;
                        when 2 =>
                            Result := N_Bare_With_Clause.With_Clause_F_Has_Private;
                            return;
                        when 3 =>
                            Result := N_Bare_With_Clause.With_Clause_F_Packages;
                            return;
                        when others => null;
                    end case;
end;
when others => null;
end case;
      --  Execution should reach this point iff nothing matched this index, so
      --  we must be out of bounds.
      Index_In_Bounds := False;
   end Get_Child;
   -----------
   -- Print --
   -----------
   procedure Print
     (Node        : Bare_Ada_Node;
      Show_Slocs  : Boolean;
      Line_Prefix : String := "")
   is
      K : Ada_Node_Kind_Type;
   begin
      if Node = null then
         Put_Line ("None");
         return;
      end if;
      K := Node.Kind;
      Put (Line_Prefix & Kind_Name (Node));
      if Show_Slocs then
         Put ("[" & Image (Sloc_Range (Node)) & "]");
      end if;
      if Is_Incomplete (Node) then
         Put (" <<INCOMPLETE>>");
      end if;
      if Is_Token_Node (Node.Kind) then
         Put_Line (": " & Image (Text (Node)));
      elsif Node.Kind not in Ada_Ada_List then
         New_Line;
      end if;
         --  List nodes are displayed in a special way (they have no field)
         if K in Ada_Ada_List then
            if Node.Count = 0 then
               Put_Line (": <empty list>");
               return;
            end if;
            New_Line;
            for Child of Node.Nodes (1 .. Node.Count) loop
               if Child /= null then
                  Print (Child, Show_Slocs, Line_Prefix & "|  ");
               end if;
            end loop;
            return;
         end if;
         --  This is for regular nodes: display each field
         declare
            use Libadalang.Introspection_Implementation;
            Attr_Prefix     : constant String := Line_Prefix & "|";
            Children_Prefix : constant String := Line_Prefix & "|  ";
            Field_List      : constant Syntax_Field_Reference_Array :=
              Syntax_Fields (K);
         begin
            for I in Field_List'Range loop
               declare
                  Child : constant Bare_Ada_Node :=
                     Implementation.Child (Node, I);
               begin
                  Put
                    (Attr_Prefix
                     & Image (Syntax_Field_Name (Field_List (I)))
                     & ":");
                  if Child /= null then
                     New_Line;
                     Print (Child, Show_Slocs, Children_Prefix);
                  else
                     Put_Line (" <null>");
                  end if;
               end;
            end loop;
         end;
   end Print;
   ------------
   -- Parent --
   ------------
   function Parent (Node : Bare_Ada_Node) return Bare_Ada_Node is
   begin
      return Node.Parent;
   end Parent;
   ------------------
   -- Stored_Token --
   ------------------
   function Stored_Token
     (Node  : Bare_Ada_Node;
      Token : Token_Reference) return Token_Index
   is
      Index : constant Token_Or_Trivia_Index := Get_Token_Index (Token);
   begin
      if Node.Unit.TDH'Access /= Get_Token_TDH (Token) then
         raise Property_Error with
           ("Cannot associate a token and a node from different analysis"
            & " units");
      elsif Index.Trivia /= No_Token_Index then
         raise Property_Error with
           ("A node cannot hold trivia");
      end if;
      return Index.Token;
   end Stored_Token;
   -------------------------
   -- Children_And_Trivia --
   -------------------------
   function Children_And_Trivia
     (Node : Bare_Ada_Node) return Bare_Children_Array
   is
      package Children_Vectors is new Ada.Containers.Vectors
        (Positive, Bare_Child_Record);
      use Children_Vectors;
      Ret_Vec : Vector;
      Ctx     : Internal_Context renames Node.Unit.Context;
      TDH     : Token_Data_Handler renames Node.Unit.TDH;
      procedure Append_Trivias (First, Last : Token_Index);
      --  Append all the trivias of tokens between indices First and Last to
      --  the returned vector.
      function Filter_Children
        (Parent : Bare_Ada_Node)
         return Internal_Bare_Ada_Node_Array;
      --  Return an array for all children in Parent that are not null
      --------------------
      -- Append_Trivias --
      --------------------
      procedure Append_Trivias (First, Last : Token_Index) is
      begin
         for I in First .. Last loop
            for D of Get_Trivias (TDH, I) loop
               Ret_Vec.Append
                 (Bare_Child_Record'
                    (Kind   => Trivia,
                     Trivia => Wrap_Token_Reference
                                 (Ctx, TDH'Access, (I, D))));
            end loop;
         end loop;
      end Append_Trivias;
      ---------------------
      -- Filter_Children --
      ---------------------
      function Filter_Children
        (Parent : Bare_Ada_Node)
         return Internal_Bare_Ada_Node_Array
      is
         Children : constant Internal_Bare_Ada_Node_Array :=
            Implementation.Children (Parent);
         Result   : Internal_Bare_Ada_Node_Array (Children'Range);
         Next     : Integer := Result'First;
      begin
         for I in Children'Range loop
            if Children (I) /= null then
               Result (Next) := Children (I);
               Next := Next + 1;
            end if;
         end loop;
         return Result (Result'First .. Next - 1);
      end Filter_Children;
      First_Child : constant Positive := 1;
      N_Children  : constant Internal_Bare_Ada_Node_Array :=
         Filter_Children (Node);
   begin
      if N_Children'Length > 0
        and then (Node.Token_Start_Index
                    /= N_Children (First_Child).Token_Start_Index)
      then
         Append_Trivias (Node.Token_Start_Index,
                         N_Children (First_Child).Token_Start_Index - 1);
      end if;
      --  Append each node to Ret_Vec, and append trivia that follow after each
      --  non-ghost nodes.
      for I in N_Children'Range loop
         Ret_Vec.Append (Bare_Child_Record'(Child, N_Children (I)));
         if not Is_Ghost (N_Children (I)) then
            Append_Trivias (N_Children (I).Token_End_Index,
                            (if I = N_Children'Last
                             then Node.Token_End_Index - 1
                             else N_Children (I + 1).Token_Start_Index - 1));
         end if;
      end loop;
      declare
         A : Bare_Children_Array (1 .. Natural (Ret_Vec.Length));
      begin
         for I in A'Range loop
            A (I) := Ret_Vec.Element (I);
         end loop;
         return A;
      end;
   end Children_And_Trivia;
   --------------
   -- Is_Ghost --
   --------------
   function Is_Ghost (Node : Bare_Ada_Node) return Boolean
   is (Node.Token_End_Index = No_Token_Index);
   -------------------
   -- Is_Incomplete --
   -------------------
   function Is_Incomplete (Node : Bare_Ada_Node) return Boolean
   is
      LGC : Bare_Ada_Node;
   begin
     if Is_List_Node (Node.Kind) then
        LGC := (if Last_Child_Index (Node) /= 0
                then Child (Node, Last_Child_Index (Node))
                else null);
        return LGC /= null and then Is_Incomplete (LGC);
      else
         return Node.Last_Attempted_Child > -1;
      end if;
   end;
   -----------------
   -- Token_Start --
   -----------------
   function Token_Start (Node : Bare_Ada_Node) return Token_Reference
   is (Token (Node, Node.Token_Start_Index));
   ---------------
   -- Token_End --
   ---------------
   function Token_End (Node : Bare_Ada_Node) return Token_Reference
   is
     (if Node.Token_End_Index = No_Token_Index
      then Token_Start (Node)
      else Token (Node, Node.Token_End_Index));
   -----------
   -- Token --
   -----------
   function Token
     (Node  : Bare_Ada_Node;
      Index : Token_Index) return Token_Reference
   is
      Unit    : constant Internal_Unit := Node.Unit;
      Context : constant Internal_Context := Unit.Context;
   begin
      return Wrap_Token_Reference
        (Context, Token_Data (Unit), (Index, No_Token_Index));
   end Token;
   ---------
   -- "<" --
   ---------
   function "<" (Left, Right : Bare_Ada_Node) return Boolean is
   begin
      --  Reject invalid inputs
      if Left /= null and Is_Synthetic (Left) then
         raise Property_Error with "left node is synthetic";
      elsif Right /= null and Is_Synthetic (Right) then
         raise Property_Error with "right node is synthetic";
      end if;
      --  Null nodes come first
      if Left = null then
         return Right /= null;
      elsif Right = null then
         return False;
      end if;
      --  So we have two non-null nodes. Sort by unit filename
      if Left.Unit < Right.Unit then
         return True;
      elsif Left.Unit /= Right.Unit then
         return False;
      end if;
      --  Both nodes come from the same unit: compare their token indexes
      if Left.Token_Start_Index < Right.Token_Start_Index then
         return True;
      elsif Left.Token_Start_Index > Right.Token_Start_Index then
         return False;
      else
         return Left.Token_End_Index < Right.Token_End_Index;
      end if;
   end "<";
   -------------
   -- Is_Null --
   -------------
   function Is_Null (Node : Bare_Ada_Node) return Boolean
   is (Node = null);
   ----------
   -- Kind --
   ----------
   function Kind (Node : Bare_Ada_Node) return Ada_Node_Kind_Type
   is (Node.Kind);
   -----------------
   -- Child_Index --
   -----------------
   function Child_Index (Node : Bare_Ada_Node) return Integer
   is
      N : Bare_Ada_Node := null;
   begin
      if Node.Parent = null then
         raise Property_Error with
            "Trying to get the child index of a root node";
      end if;
      for I in First_Child_Index (Node.Parent)
            .. Last_Child_Index (Node.Parent)
      loop
         N := Child (Node.Parent, I);
         if N = Node then
            return I - 1;
         end if;
      end loop;
      --  If we reach this point, then Node isn't a Child of Node.Parent. This
      --  is not supposed to happen.
      raise Program_Error;
   end Child_Index;
   -------------------
   -- Fetch_Sibling --
   -------------------
   function Fetch_Sibling
     (Node   : Bare_Ada_Node;
      Offset : Integer) return Bare_Ada_Node is
   begin
      --  Root nodes have no sibling: handle them now to avoid invalid requests
      --  in the code below.
      if Node.Parent = null then
         return null;
      end if;
      declare
         Node_Index : constant Positive := Child_Index (Node) + 1;
         --  Child_Index is 0-based, but the Child primitive expects a 1-based
         --  index.
         Sibling_Index : constant Integer := Node_Index + Offset;
      begin
         --  Child returns null for out-of-bound indexes
         return (if Sibling_Index >= 1
                 then Child (Node.Parent, Sibling_Index)
                 else null);
      end;
   end Fetch_Sibling;
   -------------------
   -- Fetch_Sibling --
   -------------------
   function Fetch_Sibling
     (Node   : Bare_Ada_Node;
      E_Info : Internal_Entity_Info;
      Offset : Integer) return Internal_Entity
   is
      Sibling : constant Bare_Ada_Node := Fetch_Sibling (Node, Offset);
   begin
      --  Don't forget to clear entity info if the result is nul
      return (if Sibling = null
              then No_Entity
              else (Sibling, E_Info));
   end Fetch_Sibling;
   ----------------------
   -- Previous_Sibling --
   ----------------------
   function Previous_Sibling
     (Node   : Bare_Ada_Node;
      E_Info : Internal_Entity_Info := No_Entity_Info)
      return Internal_Entity is
   begin
      return Fetch_Sibling (Node, E_Info, -1);
   end Previous_Sibling;
   ------------------
   -- Next_Sibling --
   ------------------
   function Next_Sibling
     (Node   : Bare_Ada_Node;
      E_Info : Internal_Entity_Info := No_Entity_Info)
      return Internal_Entity is
   begin
      return Fetch_Sibling (Node, E_Info, 1);
   end Next_Sibling;
   -------------
   -- Combine --
   -------------
   function Combine
     (L, R : Internal_Metadata) return Internal_Metadata
   is
      Ret : Internal_Metadata := No_Metadata;
   begin
         Ret.Dottable_Subp := L.Dottable_Subp or R.Dottable_Subp;
         if L.Primitive = No_Bare_Ada_Node then
            Ret.Primitive := R.Primitive;
         elsif R.Primitive = No_Bare_Ada_Node then
            Ret.Primitive := L.Primitive;
         else
            raise Property_Error with "Wrong combine for metadata field";
         end if;
         if L.Primitive_Real_Type = No_Bare_Ada_Node then
            Ret.Primitive_Real_Type := R.Primitive_Real_Type;
         elsif R.Primitive_Real_Type = No_Bare_Ada_Node then
            Ret.Primitive_Real_Type := L.Primitive_Real_Type;
         else
            raise Property_Error with "Wrong combine for metadata field";
         end if;
      return Ret;
   end Combine;
   -------------------------------
   -- Create_Static_Lexical_Env --
   -------------------------------
   function Create_Static_Lexical_Env
     (Parent            : Lexical_Env;
      Node              : Bare_Ada_Node;
      Transitive_Parent : Boolean := False) return Lexical_Env
   is
      Unit : constant Internal_Unit :=
        (if Node = null then null else Node.Unit);
   begin
      return Result : Lexical_Env := Create_Lexical_Env
        (Parent, Node, Transitive_Parent, Convert_Unit (Unit))
      do
         if Unit /= null then
            Register_Destroyable (Unit, Unwrap (Result.Env));
         end if;
      end return;
   end Create_Static_Lexical_Env;
   ---------
   -- Get --
   ---------
   function Get
     (A     : AST_Envs.Entity_Array;
      Index : Integer) return Internal_Entity
   is
      function Length (A : AST_Envs.Entity_Array) return Natural
      is (A'Length);
      function Get
        (A     : AST_Envs.Entity_Array;
         Index : Integer) return Internal_Entity
      is (A (Index + 1)); --  A is 1-based but Index is 0-based
      function Relative_Get is new Langkit_Support.Relative_Get
        (Item_Type     => Entity,
         Sequence_Type => AST_Envs.Entity_Array,
         Length        => Length,
         Get           => Get);
      Result : Internal_Entity;
   begin
      if Relative_Get (A, Index, Result) then
         return Result;
      else
         raise Property_Error with "out-of-bounds array access";
      end if;
   end Get;
   -----------
   -- Group --
   -----------
   function Group
     (Envs   : Lexical_Env_Array_Access;
      Env_Md : Internal_Metadata := No_Metadata) return Lexical_Env
   is (Group (Lexical_Env_Array (Envs.Items), Env_Md));
      --
      --  Untyped wrappers for Bare_Ada_Node
      --
function Ada_Node_P_Resolve_Generic_Actual
  (E : Entity
  ) return Internal_Entity;
function Ada_Node_P_Use_Clauses_In_Spec_Of_Subp_Body
  (E : Entity
  ) return Lexical_Env;
function Ada_Node_P_Use_Clauses_In_Generic_Formal_Part
  (E : Entity
  ) return Lexical_Env;
function Ada_Node_P_Nested_Generic_Formal_Part
  (E : Entity
  ) return Lexical_Env;
      --
      --  Untyped wrappers for Bare_Base_Type_Decl
      --
function Base_Type_Decl_P_Dottable_Subps
  (E : Entity
  ) return Internal_Inner_Env_Assoc_Array_Access;
      --
      --  Untyped wrappers for Bare_Type_Decl
      --
function Type_Decl_F_Direct_Primitive_Subps
  (E : Entity
  ) return Internal_Inner_Env_Assoc_Array_Access;
function Type_Decl_P_Refined_Parent_Primitives_Env
  (E : Entity
  ) return Lexical_Env;
      --
      --  Untyped wrappers for Bare_Body_Node
      --
function Body_Node_P_Subunit_Decl_Env
  (E : Entity
  ) return Lexical_Env;
function Body_Node_P_Body_Decl_Scope
  (E : Entity
  ) return Lexical_Env;
      --
      --  Untyped wrappers for Bare_Package_Body
      --
function Package_Body_P_Package_Decl_Uses_Clauses_Envs
  (E : Entity
  ) return Lexical_Env;
      --
      --  Untyped wrappers for Bare_Task_Body
      --
function Task_Body_P_Task_Type_Decl_Scope
  (E : Entity
  ) return Lexical_Env;
      --
      --  Untyped wrappers for Bare_Generic_Instantiation
      --
function Generic_Instantiation_P_Instantiation_Bindings
  (E : Entity
  ) return Internal_Inner_Env_Assoc_Array_Access;
      --
      --  Untyped wrappers for Bare_Generic_Subp_Instantiation
      --
function Generic_Subp_Instantiation_P_Designated_Subp
  (E : Entity
  ) return Internal_Entity;
      --
      --  Untyped wrappers for Bare_Name
      --
function Name_P_Name_Designated_Type_Env
  (E : Entity
  ) return Lexical_Env;
function Name_P_Use_Package_Name_Designated_Env
  (E : Entity
  ) return Lexical_Env;
   ------------------
   -- Children_Env --
   ------------------
   function Children_Env
     (Node   : Bare_Ada_Node;
      E_Info : Internal_Entity_Info := No_Entity_Info)
      return Lexical_Env
   is (Rebind_Env (Node.Self_Env, E_Info));
   --------------
   -- Node_Env --
   --------------
   function Node_Env
     (Node   : Bare_Ada_Node;
      E_Info : Internal_Entity_Info := No_Entity_Info)
      return Lexical_Env
   is
      function Get_Base_Env return Lexical_Env;
      --  Return the environment that we need to rebind before returning
      ------------------
      -- Get_Base_Env --
      ------------------
      function Get_Base_Env return Lexical_Env is
         pragma Warnings (Off, "referenced");
         function Get_Parent_Env return Lexical_Env;
         pragma Warnings (On, "referenced");
         --------------------
         -- Get_Parent_Env --
         --------------------
         function Get_Parent_Env return Lexical_Env is
            Parent : constant Lexical_Env := AST_Envs.Parent (Node.Self_Env);
         begin
            --  If Node is the root scope or the empty environment, Parent can
            --  be a wrapper around the null node. Turn this into the
            --  Empty_Env, as null envs are erroneous values in properties.
            return (if Unwrap (Parent) = null
                    then Empty_Env
                    else Parent);
         end Get_Parent_Env;
      begin
         return
           (if Node.Kind in Ada_Generic_Package_Internal | Ada_Package_Decl | Ada_Incomplete_Type_Decl | Ada_Incomplete_Tagged_Type_Decl | Ada_Protected_Type_Decl | Ada_Task_Type_Decl | Ada_Single_Task_Type_Decl | Ada_Type_Decl | Ada_Abstract_Subp_Decl | Ada_Abstract_Formal_Subp_Decl | Ada_Concrete_Formal_Subp_Decl | Ada_Subp_Decl | Ada_Entry_Decl | Ada_Enum_Literal_Decl | Ada_Synthetic_Char_Enum_Lit | Ada_Generic_Subp_Internal | Ada_Synthetic_Subp_Decl | Ada_Expr_Function | Ada_Null_Subp_Decl | Ada_Subp_Body | Ada_Subp_Renaming_Decl | Ada_Package_Body_Stub | Ada_Protected_Body_Stub | Ada_Subp_Body_Stub | Ada_Task_Body_Stub | Ada_Entry_Body | Ada_Package_Body | Ada_Protected_Body | Ada_Task_Body | Ada_Exception_Handler | Ada_Generic_Package_Decl | Ada_Generic_Subp_Decl | Ada_Generic_Package_Instantiation | Ada_Generic_Subp_Instantiation | Ada_Generic_Package_Renaming_Decl | Ada_Generic_Subp_Renaming_Decl | Ada_Package_Renaming_Decl | Ada_Single_Protected_Decl | Ada_Single_Task_Decl | Ada_Private_Part | Ada_Decl_Expr | Ada_Accept_Stmt | Ada_Accept_Stmt_With_Stmts | Ada_For_Loop_Stmt | Ada_Begin_Block | Ada_Decl_Block | Ada_Extended_Return_Stmt | Ada_Named_Stmt | Ada_Access_To_Subp_Def
            then Get_Parent_Env
            else Node.Self_Env);
      end Get_Base_Env;
      Base_Env : Lexical_Env := Get_Base_Env;
      Result   : constant Lexical_Env := Rebind_Env (Base_Env, E_Info);
   begin
      Dec_Ref (Base_Env);
      return Result;
   end Node_Env;
   ------------
   -- Parent --
   ------------
   function Parent
     (Node   : Bare_Ada_Node;
      E_Info : Internal_Entity_Info := No_Entity_Info)
      return Internal_Entity is
   begin
      --  TODO: shed entity information as appropriate
      return (Node.Parent, E_Info);
   end Parent;
   -------------
   -- Parents --
   -------------
   function Parents
     (Node      : Bare_Ada_Node;
      With_Self : Boolean := True;
      E_Info    : Internal_Entity_Info := No_Entity_Info)
      return Internal_Entity_Array_Access
   is
      Bare_Parents : Bare_Ada_Node_Array_Access := Parents (Node, With_Self);
      Result       : Internal_Entity_Array_Access :=
         Create_Internal_Entity_Array (Bare_Parents.N);
   begin
      --  TODO: shed entity information as appropriate
      for I in Bare_Parents.Items'Range loop
         Result.Items (I) := (Bare_Parents.Items (I), E_Info);
      end loop;
      Dec_Ref (Bare_Parents);
      return Result;
   end Parents;
   --------------
   -- Children --
   --------------
   function Children
     (Node   : Bare_Ada_Node;
      E_Info : Internal_Entity_Info := No_Entity_Info)
      return Internal_Entity_Array_Access
   is
      Bare_Children : Bare_Ada_Node_Array_Access := Children (Node);
      Result        : Internal_Entity_Array_Access :=
         Create_Internal_Entity_Array (Bare_Children.N);
   begin
      --  TODO: shed entity information as appropriate
      for I in Bare_Children.Items'Range loop
         Result.Items (I) := (Bare_Children.Items (I), E_Info);
      end loop;
      Dec_Ref (Bare_Children);
      return Result;
   end Children;
   ---------------------
   -- New_Unit_String --
   ---------------------
   function New_Unit_String
     (Unit : Internal_Unit; Str : String) return String_Access
   is
      procedure Register_Destroyable_String is new Register_Destroyable_Gen
        (String, String_Access, Free);
   begin
      return Ret : String_Access := new String'(Str) do
         Register_Destroyable_String (Unit, Ret);
      end return;
   end New_Unit_String;
   --------------------------------
   -- Assign_Names_To_Logic_Vars --
   --------------------------------
   procedure Assign_Names_To_Logic_Vars (Node : Bare_Ada_Node) is
      pragma Warnings (Off, "referenced");
      procedure Assign
        (Node  : Bare_Ada_Node;
         LV    : in out Logic_Var_Record;
         Field : String);
      --  Assign a name to the LV logic variable. Node must be the node that
      --  owns LV, and Field must be the name of the field in Node that holds
      --  LV.
      ------------
      -- Assign --
      ------------
      procedure Assign
        (Node  : Bare_Ada_Node;
         LV    : in out Logic_Var_Record;
         Field : String) is
      begin
         LV.Dbg_Name :=
           new String'(Image (Short_Text_Image (Node)) & "." & Field);
      end Assign;
      K : constant Ada_Node_Kind_Type := Node.Kind;
      pragma Warnings (On, "referenced");
   begin
      case Ada_Ada_Node (K) is
when Ada_Attribute_Ref_Range =>
declare
N_Bare_Attribute_Ref : constant Bare_Attribute_Ref := Node;
begin
Assign (N_Bare_Attribute_Ref,        N_Bare_Attribute_Ref.Attribute_Ref_R_Ref_Var,        "r_ref_var");
end;
when Ada_Call_Expr_Range =>
declare
N_Bare_Call_Expr : constant Bare_Call_Expr := Node;
begin
Assign (N_Bare_Call_Expr,        N_Bare_Call_Expr.Call_Expr_R_Called_Spec,        "r_called_spec");
end;
when Ada_Explicit_Deref_Range =>
declare
N_Bare_Explicit_Deref : constant Bare_Explicit_Deref := Node;
begin
Assign (N_Bare_Explicit_Deref,        N_Bare_Explicit_Deref.Explicit_Deref_R_Called_Spec,        "r_called_spec");
end;
when Ada_Reduce_Attribute_Ref_Range =>
declare
N_Bare_Reduce_Attribute_Ref : constant Bare_Reduce_Attribute_Ref := Node;
begin
Assign (N_Bare_Reduce_Attribute_Ref,        N_Bare_Reduce_Attribute_Ref.Reduce_Attribute_Ref_R_Ref_Var,        "r_ref_var");
end;
when Ada_Target_Name_Range =>
declare
N_Bare_Target_Name : constant Bare_Target_Name := Node;
begin
Assign (N_Bare_Target_Name,        N_Bare_Target_Name.Target_Name_R_Ref_Var,        "r_ref_var");
end;
when others => null;
end case;
      for Child of Internal_Bare_Ada_Node_Array'(Children (Node)) loop
         if Child /= null then
            Assign_Names_To_Logic_Vars (Child);
         end if;
      end loop;
   end Assign_Names_To_Logic_Vars;
   ----------------
   -- Text_Image --
   ----------------
   function Text_Image (Ent : Internal_Entity) return Text_Type is
   begin
      if Ent.Node /= null then
         declare
            Node_Image : constant Text_Type := Short_Text_Image (Ent.Node);
         begin
            return
            (if Ent.Info.Rebindings /= null
             then "<| "
             & Node_Image (Node_Image'First + 1 .. Node_Image'Last - 1) & " "
             & AST_Envs.Text_Image (Ent.Info.Rebindings) & " |>"
             else Node_Image);
         end;
      else
         return "None";
      end if;
   end Text_Image;
   ---------------------
   -- Full_Sloc_Image --
   ---------------------
   function Full_Sloc_Image (Node : Bare_Ada_Node) return String_Type
   is
      Res      : constant Text_Type :=
        To_Text
          (Ada.Directories.Simple_Name
             (Get_Filename (Unit (Node))))
           & ":" & To_Text (Image (Start_Sloc (Sloc_Range (Node)))) & ": ";
   begin
      return Create_String (Res);
   end Full_Sloc_Image;
   -----------
   -- Image --
   -----------
   function Image (Ent : Internal_Entity) return String is
      Result : constant Text_Type := Text_Image (Ent);
   begin
      return Image (Result);
   end Image;
   ---------------
   -- Can_Reach --
   ---------------
   function Can_Reach (El, From : Bare_Ada_Node) return Boolean is
   begin
      --  Since this function is only used to implement sequential semantics in
      --  envs, we consider that elements coming from different units are
      --  always visible for each other, and let the user implement language
      --  specific visibility rules in the DSL.
      if El = null or else From = null or else El.Unit /= From.Unit then
         return True;
      end if;
      return El.Token_Start_Index < From.Token_Start_Index;
   end Can_Reach;
   -----------------
   -- Hash_Entity --
   -----------------
   function Hash_Entity (Self : Internal_Entity) return Hash_Type is
   begin
      return Combine (Hash (Self.Node), Hash (Self.Info.Rebindings));
   end Hash_Entity;
   --------------------
   -- Compare_Entity --
   --------------------
   function Compare_Entity (Left, Right : Internal_Entity) return Boolean
   is
   begin
      return (Left.Node = Right.Node
              and then Left.Info.Rebindings = Right.Info.Rebindings);
   end Compare_Entity;
   --------------------------------
   -- Create_Dynamic_Lexical_Env --
   --------------------------------
   function Create_Dynamic_Lexical_Env
     (Self              : Bare_Ada_Node;
      Assocs_Getter     : Inner_Env_Assocs_Resolver;
      Assoc_Resolver    : Entity_Resolver;
      Transitive_Parent : Boolean) return Lexical_Env
   is
      Unit : constant Internal_Unit := Self.Unit;
   begin
      --  This restriction is necessary to avoid relocation issues when
      --  Self.Self_Env is terminated.
      if Is_Foreign_Strict (Self.Self_Env, Self) then
         raise Property_Error with
           ("cannot create a dynamic lexical env when Self.Self_Env is"
            & " foreign");
      end if;
      return Result : constant Lexical_Env := Create_Dynamic_Lexical_Env
        (Parent            => Null_Lexical_Env,
         Node              => Self,
         Transitive_Parent => Transitive_Parent,
         Owner             => Convert_Unit (Unit),
         Assocs_Getter     => Assocs_Getter,
         Assoc_Resolver    => Assoc_Resolver)
      do
         --  Since dynamic lexical environments can only be created in lazy
         --  field initializers, it is fine to tie Result's lifetime to the
         --  its owning unit's lifetime.
         Register_Destroyable (Unit, Unwrap (Result));
      end return;
   end Create_Dynamic_Lexical_Env;
   procedure Destroy_Synthetic_Node (Node : in out Bare_Ada_Node);
   --  Helper for the Register_Destroyable above
   ------------
   -- Length --
   ------------
   function Length (Node : Bare_Ada_List) return Natural
   is (if Node = null then 0 else Children_Count (Node));
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (B : Boolean) return String is
      begin
         return (if B then "True" else "False");
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (I : Integer) return String is
      begin
         return Integer'Image (I);
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (S : Symbol_Type) return String is
      begin
         return (if S = null
                 then "None"
                 else Image (S.all, With_Quotes => True));
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (C : Character_Type) return String is
         C_Str : constant Text_Type := (1 => C);
      begin
         return "'" & Image (C_Str) & "'";
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (S : String_Type) return String is
      begin
         return Image (S.Content, With_Quotes => True);
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (Env : Lexical_Env) return String is
      begin
         case Env.Kind is
         when Static_Primary =>
            return "<LexicalEnv static-primary for "
                   & Trace_Image (Env_Node (Env)) & ">";
         when others =>
            return "<LexicalEnv synthetic>";
         end case;
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (R : Env_Rebindings) return String is
      begin
         return Image (Text_Image (R));
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (Unit : Internal_Unit) return String is
      begin
         return "Internal_Unit (""" & Basename (Unit) & """)";
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (Eq : Logic_Equation) return String is
         pragma Unreferenced (Eq);
      begin
         return "<LogicEquation>";
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (Var : Logic_Var) return String is
         pragma Unreferenced (Var);
      begin
         return "<LogicVariable>";
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (K : Analysis_Unit_Kind) return String is
      begin
         return Analysis_Unit_Kind'Image (K);
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image (Self : Ref_Categories) return String is
         Result : Unbounded_String;
         First  : Boolean := True;
      begin
         Append (Result, "RefCategories(");
         for C in Ref_Category loop
            if Self (C) then
               if First then
                  First := False;
               else
                  Append (Result, ", ");
               end if;
               Append (Result, C'Image);
            end if;
         end loop;
         Append (Result, ")");
         return To_String (Result);
      end Trace_Image;
      ----------
      -- Hash --
      ----------
      pragma Warnings (Off, "referenced");
      function Hash (R : Internal_Metadata) return Hash_Type is
         pragma Warnings (On, "referenced");
      begin
            return Combine ((Hash (R.Dottable_Subp), Hash (R.Primitive), Hash (R.Primitive_Real_Type)));
      end Hash;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Metadata) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Dottable_Subp => " & Trace_Image (R.Dottable_Subp)
                        & ", "
                     & "Primitive => " & Trace_Image (R.Primitive)
                        & ", "
                     & "Primitive_Real_Type => " & Trace_Image (R.Primitive_Real_Type)
               & ")");
      end Trace_Image;
      ----------
      -- Hash --
      ----------
      pragma Warnings (Off, "referenced");
      function Hash (R : Internal_Entity_Info) return Hash_Type is
         pragma Warnings (On, "referenced");
      begin
            return Combine ((Hash (R.Md), Hash (R.Rebindings), Hash (R.From_Rebound)));
      end Hash;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Info) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Md => " & Trace_Image (R.Md)
                        & ", "
                     & "Rebindings => " & Trace_Image (R.Rebindings)
                        & ", "
                     & "From_Rebound => " & Trace_Image (R.From_Rebound)
               & ")");
      end Trace_Image;
      function Create_Internal_Entity
        (Node : Bare_Ada_Node; Info : Internal_Entity_Info)
         return Internal_Entity is
      begin
         if Node = null then
            return No_Entity;
         end if;
         return (Node => Node, Info => Info);
      end;
      ----------
      -- Hash --
      ----------
      pragma Warnings (Off, "referenced");
      function Hash (R : Internal_Entity) return Hash_Type is
         pragma Warnings (On, "referenced");
      begin
            return Combine
              (Hash (R.Node), Hash (R.Info));
      end Hash;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Expr
        (Node : Bare_Expr; Info : Internal_Entity_Info)
         return Internal_Entity_Expr is
      begin
         if Node = null then
            return No_Entity_Expr;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Expr) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Aspect) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Exists => " & Trace_Image (R.Exists)
                        & ", "
                     & "Node => " & Trace_Image (R.Node)
                        & ", "
                     & "Value => " & Trace_Image (R.Value)
               & ")");
      end Trace_Image;
      function Create_Internal_Entity_Basic_Decl
        (Node : Bare_Basic_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Basic_Decl is
      begin
         if Node = null then
            return No_Entity_Basic_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      ----------
      -- Hash --
      ----------
      pragma Warnings (Off, "referenced");
      function Hash (R : Internal_Entity_Basic_Decl) return Hash_Type is
         pragma Warnings (On, "referenced");
      begin
            return Combine
              (Hash (R.Node), Hash (R.Info));
      end Hash;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Basic_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Completion_Item) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Decl => " & Trace_Image (R.Decl)
                        & ", "
                     & "Is_Dot_Call => " & Trace_Image (R.Is_Dot_Call)
                        & ", "
                     & "Is_Visible => " & Trace_Image (R.Is_Visible)
                        & ", "
                     & "Weight => " & Trace_Image (R.Weight)
               & ")");
      end Trace_Image;
      -------------
      -- Inc_Ref --
      -------------
      procedure Inc_Ref (R : Internal_Designated_Env) is
      begin
               Inc_Ref (R.Direct_Env);
      end Inc_Ref;
      -------------
      -- Dec_Ref --
      -------------
      procedure Dec_Ref (R : in out Internal_Designated_Env) is
      begin
               Dec_Ref (R.Direct_Env);
      end Dec_Ref;
      ----------------
      -- Equivalent --
      ----------------
      function Equivalent (L, R : Internal_Designated_Env) return Boolean is
      begin
         return L.Kind = R.Kind and then L.Env_Name = R.Env_Name and then Equivalent (L.Direct_Env, R.Direct_Env);
      end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Designated_Env) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Kind => " & Trace_Image (R.Kind)
                        & ", "
                     & "Env_Name => " & Trace_Image (R.Env_Name)
                        & ", "
                     & "Direct_Env => " & Trace_Image (R.Direct_Env)
               & ")");
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Discrete_Range) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Low_Bound => " & Trace_Image (R.Low_Bound)
                        & ", "
                     & "High_Bound => " & Trace_Image (R.High_Bound)
               & ")");
      end Trace_Image;
      function Create_Internal_Entity_Ada_List
        (Node : Bare_Ada_List; Info : Internal_Entity_Info)
         return Internal_Entity_Ada_List is
      begin
         if Node = null then
            return No_Entity_Ada_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Ada_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Ada_Node_List
        (Node : Bare_Ada_Node_List; Info : Internal_Entity_Info)
         return Internal_Entity_Ada_Node_List is
      begin
         if Node = null then
            return No_Entity_Ada_Node_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Ada_Node_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Alternatives_List
        (Node : Bare_Alternatives_List; Info : Internal_Entity_Info)
         return Internal_Entity_Alternatives_List is
      begin
         if Node = null then
            return No_Entity_Alternatives_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Alternatives_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Name
        (Node : Bare_Name; Info : Internal_Entity_Info)
         return Internal_Entity_Name is
      begin
         if Node = null then
            return No_Entity_Name;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Name) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Single_Tok_Node
        (Node : Bare_Single_Tok_Node; Info : Internal_Entity_Info)
         return Internal_Entity_Single_Tok_Node is
      begin
         if Node = null then
            return No_Entity_Single_Tok_Node;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Single_Tok_Node) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Base_Id
        (Node : Bare_Base_Id; Info : Internal_Entity_Info)
         return Internal_Entity_Base_Id is
      begin
         if Node = null then
            return No_Entity_Base_Id;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Base_Id) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Identifier
        (Node : Bare_Identifier; Info : Internal_Entity_Info)
         return Internal_Entity_Identifier is
      begin
         if Node = null then
            return No_Entity_Identifier;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Identifier) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Discriminant_Values) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Discriminant => " & Trace_Image (R.Discriminant)
                        & ", "
                     & "Values => " & Trace_Image (R.Values)
               & ")");
      end Trace_Image;
      -------------
      -- Inc_Ref --
      -------------
      procedure Inc_Ref (R : Internal_Doc_Annotation) is
      begin
               Inc_Ref (R.Key);
               Inc_Ref (R.Value);
      end Inc_Ref;
      -------------
      -- Dec_Ref --
      -------------
      procedure Dec_Ref (R : in out Internal_Doc_Annotation) is
      begin
               Dec_Ref (R.Key);
               Dec_Ref (R.Value);
      end Dec_Ref;
      ----------------
      -- Equivalent --
      ----------------
      function Equivalent (L, R : Internal_Doc_Annotation) return Boolean is
      begin
         return Equivalent (L.Key, R.Key) and then Equivalent (L.Value, R.Value);
      end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Doc_Annotation) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Key => " & Trace_Image (R.Key)
                        & ", "
                     & "Value => " & Trace_Image (R.Value)
               & ")");
      end Trace_Image;
      function Create_Internal_Entity_Abort_Node
        (Node : Bare_Abort_Node; Info : Internal_Entity_Info)
         return Internal_Entity_Abort_Node is
      begin
         if Node = null then
            return No_Entity_Abort_Node;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Abort_Node) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Abort_Absent
        (Node : Bare_Abort_Absent; Info : Internal_Entity_Info)
         return Internal_Entity_Abort_Absent is
      begin
         if Node = null then
            return No_Entity_Abort_Absent;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Abort_Absent) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Abort_Present
        (Node : Bare_Abort_Present; Info : Internal_Entity_Info)
         return Internal_Entity_Abort_Present is
      begin
         if Node = null then
            return No_Entity_Abort_Present;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Abort_Present) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Stmt
        (Node : Bare_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Stmt is
      begin
         if Node = null then
            return No_Entity_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Simple_Stmt
        (Node : Bare_Simple_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Simple_Stmt is
      begin
         if Node = null then
            return No_Entity_Simple_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Simple_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Abort_Stmt
        (Node : Bare_Abort_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Abort_Stmt is
      begin
         if Node = null then
            return No_Entity_Abort_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Abort_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Abstract_Node
        (Node : Bare_Abstract_Node; Info : Internal_Entity_Info)
         return Internal_Entity_Abstract_Node is
      begin
         if Node = null then
            return No_Entity_Abstract_Node;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Abstract_Node) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Abstract_Absent
        (Node : Bare_Abstract_Absent; Info : Internal_Entity_Info)
         return Internal_Entity_Abstract_Absent is
      begin
         if Node = null then
            return No_Entity_Abstract_Absent;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Abstract_Absent) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Basic_Subp_Decl
        (Node : Bare_Basic_Subp_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Basic_Subp_Decl is
      begin
         if Node = null then
            return No_Entity_Basic_Subp_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Basic_Subp_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Classic_Subp_Decl
        (Node : Bare_Classic_Subp_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Classic_Subp_Decl is
      begin
         if Node = null then
            return No_Entity_Classic_Subp_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Classic_Subp_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Formal_Subp_Decl
        (Node : Bare_Formal_Subp_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Formal_Subp_Decl is
      begin
         if Node = null then
            return No_Entity_Formal_Subp_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Formal_Subp_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Abstract_Formal_Subp_Decl
        (Node : Bare_Abstract_Formal_Subp_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Abstract_Formal_Subp_Decl is
      begin
         if Node = null then
            return No_Entity_Abstract_Formal_Subp_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Abstract_Formal_Subp_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Abstract_Present
        (Node : Bare_Abstract_Present; Info : Internal_Entity_Info)
         return Internal_Entity_Abstract_Present is
      begin
         if Node = null then
            return No_Entity_Abstract_Present;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Abstract_Present) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Abstract_State_Decl
        (Node : Bare_Abstract_State_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Abstract_State_Decl is
      begin
         if Node = null then
            return No_Entity_Abstract_State_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Abstract_State_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Abstract_State_Decl_Expr
        (Node : Bare_Abstract_State_Decl_Expr; Info : Internal_Entity_Info)
         return Internal_Entity_Abstract_State_Decl_Expr is
      begin
         if Node = null then
            return No_Entity_Abstract_State_Decl_Expr;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Abstract_State_Decl_Expr) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Abstract_State_Decl_List
        (Node : Bare_Abstract_State_Decl_List; Info : Internal_Entity_Info)
         return Internal_Entity_Abstract_State_Decl_List is
      begin
         if Node = null then
            return No_Entity_Abstract_State_Decl_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Abstract_State_Decl_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Abstract_Subp_Decl
        (Node : Bare_Abstract_Subp_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Abstract_Subp_Decl is
      begin
         if Node = null then
            return No_Entity_Abstract_Subp_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Abstract_Subp_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Composite_Stmt
        (Node : Bare_Composite_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Composite_Stmt is
      begin
         if Node = null then
            return No_Entity_Composite_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Composite_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Accept_Stmt
        (Node : Bare_Accept_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Accept_Stmt is
      begin
         if Node = null then
            return No_Entity_Accept_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Accept_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Accept_Stmt_With_Stmts
        (Node : Bare_Accept_Stmt_With_Stmts; Info : Internal_Entity_Info)
         return Internal_Entity_Accept_Stmt_With_Stmts is
      begin
         if Node = null then
            return No_Entity_Accept_Stmt_With_Stmts;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Accept_Stmt_With_Stmts) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Type_Def
        (Node : Bare_Type_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Type_Def is
      begin
         if Node = null then
            return No_Entity_Type_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Type_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Access_Def
        (Node : Bare_Access_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Access_Def is
      begin
         if Node = null then
            return No_Entity_Access_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Access_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Access_To_Subp_Def
        (Node : Bare_Access_To_Subp_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Access_To_Subp_Def is
      begin
         if Node = null then
            return No_Entity_Access_To_Subp_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Access_To_Subp_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Base_Aggregate
        (Node : Bare_Base_Aggregate; Info : Internal_Entity_Info)
         return Internal_Entity_Base_Aggregate is
      begin
         if Node = null then
            return No_Entity_Base_Aggregate;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Base_Aggregate) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Aggregate
        (Node : Bare_Aggregate; Info : Internal_Entity_Info)
         return Internal_Entity_Aggregate is
      begin
         if Node = null then
            return No_Entity_Aggregate;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Aggregate) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Basic_Assoc
        (Node : Bare_Basic_Assoc; Info : Internal_Entity_Info)
         return Internal_Entity_Basic_Assoc is
      begin
         if Node = null then
            return No_Entity_Basic_Assoc;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Basic_Assoc) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Aggregate_Assoc
        (Node : Bare_Aggregate_Assoc; Info : Internal_Entity_Info)
         return Internal_Entity_Aggregate_Assoc is
      begin
         if Node = null then
            return No_Entity_Aggregate_Assoc;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Aggregate_Assoc) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Aliased_Node
        (Node : Bare_Aliased_Node; Info : Internal_Entity_Info)
         return Internal_Entity_Aliased_Node is
      begin
         if Node = null then
            return No_Entity_Aliased_Node;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Aliased_Node) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Aliased_Absent
        (Node : Bare_Aliased_Absent; Info : Internal_Entity_Info)
         return Internal_Entity_Aliased_Absent is
      begin
         if Node = null then
            return No_Entity_Aliased_Absent;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Aliased_Absent) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Aliased_Present
        (Node : Bare_Aliased_Present; Info : Internal_Entity_Info)
         return Internal_Entity_Aliased_Present is
      begin
         if Node = null then
            return No_Entity_Aliased_Present;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Aliased_Present) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_All_Node
        (Node : Bare_All_Node; Info : Internal_Entity_Info)
         return Internal_Entity_All_Node is
      begin
         if Node = null then
            return No_Entity_All_Node;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_All_Node) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_All_Absent
        (Node : Bare_All_Absent; Info : Internal_Entity_Info)
         return Internal_Entity_All_Absent is
      begin
         if Node = null then
            return No_Entity_All_Absent;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_All_Absent) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_All_Present
        (Node : Bare_All_Present; Info : Internal_Entity_Info)
         return Internal_Entity_All_Present is
      begin
         if Node = null then
            return No_Entity_All_Present;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_All_Present) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Allocator
        (Node : Bare_Allocator; Info : Internal_Entity_Info)
         return Internal_Entity_Allocator is
      begin
         if Node = null then
            return No_Entity_Allocator;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Allocator) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Anonymous_Expr_Decl
        (Node : Bare_Anonymous_Expr_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Anonymous_Expr_Decl is
      begin
         if Node = null then
            return No_Entity_Anonymous_Expr_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Anonymous_Expr_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Type_Expr
        (Node : Bare_Type_Expr; Info : Internal_Entity_Info)
         return Internal_Entity_Type_Expr is
      begin
         if Node = null then
            return No_Entity_Type_Expr;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Type_Expr) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Anonymous_Type
        (Node : Bare_Anonymous_Type; Info : Internal_Entity_Info)
         return Internal_Entity_Anonymous_Type is
      begin
         if Node = null then
            return No_Entity_Anonymous_Type;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Anonymous_Type) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Base_Type_Access_Def
        (Node : Bare_Base_Type_Access_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Base_Type_Access_Def is
      begin
         if Node = null then
            return No_Entity_Base_Type_Access_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Base_Type_Access_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Anonymous_Type_Access_Def
        (Node : Bare_Anonymous_Type_Access_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Anonymous_Type_Access_Def is
      begin
         if Node = null then
            return No_Entity_Anonymous_Type_Access_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Anonymous_Type_Access_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Base_Type_Decl
        (Node : Bare_Base_Type_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Base_Type_Decl is
      begin
         if Node = null then
            return No_Entity_Base_Type_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      ----------
      -- Hash --
      ----------
      pragma Warnings (Off, "referenced");
      function Hash (R : Internal_Entity_Base_Type_Decl) return Hash_Type is
         pragma Warnings (On, "referenced");
      begin
            return Combine
              (Hash (R.Node), Hash (R.Info));
      end Hash;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Base_Type_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Type_Decl
        (Node : Bare_Type_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Type_Decl is
      begin
         if Node = null then
            return No_Entity_Type_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Type_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Anonymous_Type_Decl
        (Node : Bare_Anonymous_Type_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Anonymous_Type_Decl is
      begin
         if Node = null then
            return No_Entity_Anonymous_Type_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Anonymous_Type_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Array_Indices
        (Node : Bare_Array_Indices; Info : Internal_Entity_Info)
         return Internal_Entity_Array_Indices is
      begin
         if Node = null then
            return No_Entity_Array_Indices;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Array_Indices) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Array_Type_Def
        (Node : Bare_Array_Type_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Array_Type_Def is
      begin
         if Node = null then
            return No_Entity_Array_Type_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Array_Type_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Aspect_Assoc
        (Node : Bare_Aspect_Assoc; Info : Internal_Entity_Info)
         return Internal_Entity_Aspect_Assoc is
      begin
         if Node = null then
            return No_Entity_Aspect_Assoc;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Aspect_Assoc) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Aspect_Assoc_List
        (Node : Bare_Aspect_Assoc_List; Info : Internal_Entity_Info)
         return Internal_Entity_Aspect_Assoc_List is
      begin
         if Node = null then
            return No_Entity_Aspect_Assoc_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Aspect_Assoc_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Aspect_Clause
        (Node : Bare_Aspect_Clause; Info : Internal_Entity_Info)
         return Internal_Entity_Aspect_Clause is
      begin
         if Node = null then
            return No_Entity_Aspect_Clause;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Aspect_Clause) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Aspect_Spec
        (Node : Bare_Aspect_Spec; Info : Internal_Entity_Info)
         return Internal_Entity_Aspect_Spec is
      begin
         if Node = null then
            return No_Entity_Aspect_Spec;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Aspect_Spec) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Assign_Stmt
        (Node : Bare_Assign_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Assign_Stmt is
      begin
         if Node = null then
            return No_Entity_Assign_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Assign_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Basic_Assoc_List
        (Node : Bare_Basic_Assoc_List; Info : Internal_Entity_Info)
         return Internal_Entity_Basic_Assoc_List is
      begin
         if Node = null then
            return No_Entity_Basic_Assoc_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Basic_Assoc_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Assoc_List
        (Node : Bare_Assoc_List; Info : Internal_Entity_Info)
         return Internal_Entity_Assoc_List is
      begin
         if Node = null then
            return No_Entity_Assoc_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Assoc_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_At_Clause
        (Node : Bare_At_Clause; Info : Internal_Entity_Info)
         return Internal_Entity_At_Clause is
      begin
         if Node = null then
            return No_Entity_At_Clause;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_At_Clause) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Attribute_Def_Clause
        (Node : Bare_Attribute_Def_Clause; Info : Internal_Entity_Info)
         return Internal_Entity_Attribute_Def_Clause is
      begin
         if Node = null then
            return No_Entity_Attribute_Def_Clause;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Attribute_Def_Clause) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Attribute_Ref
        (Node : Bare_Attribute_Ref; Info : Internal_Entity_Info)
         return Internal_Entity_Attribute_Ref is
      begin
         if Node = null then
            return No_Entity_Attribute_Ref;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Attribute_Ref) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Base_Assoc
        (Node : Bare_Base_Assoc; Info : Internal_Entity_Info)
         return Internal_Entity_Base_Assoc is
      begin
         if Node = null then
            return No_Entity_Base_Assoc;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Base_Assoc) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Base_Assoc_List
        (Node : Bare_Base_Assoc_List; Info : Internal_Entity_Info)
         return Internal_Entity_Base_Assoc_List is
      begin
         if Node = null then
            return No_Entity_Base_Assoc_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Base_Assoc_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Base_Formal_Param_Decl
        (Node : Bare_Base_Formal_Param_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Base_Formal_Param_Decl is
      begin
         if Node = null then
            return No_Entity_Base_Formal_Param_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Base_Formal_Param_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Base_Formal_Param_Holder
        (Node : Bare_Base_Formal_Param_Holder; Info : Internal_Entity_Info)
         return Internal_Entity_Base_Formal_Param_Holder is
      begin
         if Node = null then
            return No_Entity_Base_Formal_Param_Holder;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Base_Formal_Param_Holder) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Base_Loop_Stmt
        (Node : Bare_Base_Loop_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Base_Loop_Stmt is
      begin
         if Node = null then
            return No_Entity_Base_Loop_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Base_Loop_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Base_Package_Decl
        (Node : Bare_Base_Package_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Base_Package_Decl is
      begin
         if Node = null then
            return No_Entity_Base_Package_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Base_Package_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Base_Record_Def
        (Node : Bare_Base_Record_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Base_Record_Def is
      begin
         if Node = null then
            return No_Entity_Base_Record_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Base_Record_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Body_Node
        (Node : Bare_Body_Node; Info : Internal_Entity_Info)
         return Internal_Entity_Body_Node is
      begin
         if Node = null then
            return No_Entity_Body_Node;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Body_Node) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Base_Subp_Body
        (Node : Bare_Base_Subp_Body; Info : Internal_Entity_Info)
         return Internal_Entity_Base_Subp_Body is
      begin
         if Node = null then
            return No_Entity_Base_Subp_Body;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Base_Subp_Body) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Base_Subp_Spec
        (Node : Bare_Base_Subp_Spec; Info : Internal_Entity_Info)
         return Internal_Entity_Base_Subp_Spec is
      begin
         if Node = null then
            return No_Entity_Base_Subp_Spec;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Base_Subp_Spec) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Base_Subtype_Decl
        (Node : Bare_Base_Subtype_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Base_Subtype_Decl is
      begin
         if Node = null then
            return No_Entity_Base_Subtype_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Base_Subtype_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Basic_Decl_List
        (Node : Bare_Basic_Decl_List; Info : Internal_Entity_Info)
         return Internal_Entity_Basic_Decl_List is
      begin
         if Node = null then
            return No_Entity_Basic_Decl_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Basic_Decl_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Block_Stmt
        (Node : Bare_Block_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Block_Stmt is
      begin
         if Node = null then
            return No_Entity_Block_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Block_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Begin_Block
        (Node : Bare_Begin_Block; Info : Internal_Entity_Info)
         return Internal_Entity_Begin_Block is
      begin
         if Node = null then
            return No_Entity_Begin_Block;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Begin_Block) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Bin_Op
        (Node : Bare_Bin_Op; Info : Internal_Entity_Info)
         return Internal_Entity_Bin_Op is
      begin
         if Node = null then
            return No_Entity_Bin_Op;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Bin_Op) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Body_Stub
        (Node : Bare_Body_Stub; Info : Internal_Entity_Info)
         return Internal_Entity_Body_Stub is
      begin
         if Node = null then
            return No_Entity_Body_Stub;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Body_Stub) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Box_Expr
        (Node : Bare_Box_Expr; Info : Internal_Entity_Info)
         return Internal_Entity_Box_Expr is
      begin
         if Node = null then
            return No_Entity_Box_Expr;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Box_Expr) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Bracket_Aggregate
        (Node : Bare_Bracket_Aggregate; Info : Internal_Entity_Info)
         return Internal_Entity_Bracket_Aggregate is
      begin
         if Node = null then
            return No_Entity_Bracket_Aggregate;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Bracket_Aggregate) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Delta_Aggregate
        (Node : Bare_Delta_Aggregate; Info : Internal_Entity_Info)
         return Internal_Entity_Delta_Aggregate is
      begin
         if Node = null then
            return No_Entity_Delta_Aggregate;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Delta_Aggregate) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Bracket_Delta_Aggregate
        (Node : Bare_Bracket_Delta_Aggregate; Info : Internal_Entity_Info)
         return Internal_Entity_Bracket_Delta_Aggregate is
      begin
         if Node = null then
            return No_Entity_Bracket_Delta_Aggregate;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Bracket_Delta_Aggregate) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Call_Expr
        (Node : Bare_Call_Expr; Info : Internal_Entity_Info)
         return Internal_Entity_Call_Expr is
      begin
         if Node = null then
            return No_Entity_Call_Expr;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Call_Expr) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Call_Stmt
        (Node : Bare_Call_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Call_Stmt is
      begin
         if Node = null then
            return No_Entity_Call_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Call_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Cond_Expr
        (Node : Bare_Cond_Expr; Info : Internal_Entity_Info)
         return Internal_Entity_Cond_Expr is
      begin
         if Node = null then
            return No_Entity_Cond_Expr;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Cond_Expr) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Case_Expr
        (Node : Bare_Case_Expr; Info : Internal_Entity_Info)
         return Internal_Entity_Case_Expr is
      begin
         if Node = null then
            return No_Entity_Case_Expr;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Case_Expr) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Case_Expr_Alternative
        (Node : Bare_Case_Expr_Alternative; Info : Internal_Entity_Info)
         return Internal_Entity_Case_Expr_Alternative is
      begin
         if Node = null then
            return No_Entity_Case_Expr_Alternative;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Case_Expr_Alternative) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Case_Expr_Alternative_List
        (Node : Bare_Case_Expr_Alternative_List; Info : Internal_Entity_Info)
         return Internal_Entity_Case_Expr_Alternative_List is
      begin
         if Node = null then
            return No_Entity_Case_Expr_Alternative_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Case_Expr_Alternative_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Case_Stmt
        (Node : Bare_Case_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Case_Stmt is
      begin
         if Node = null then
            return No_Entity_Case_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Case_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Case_Stmt_Alternative
        (Node : Bare_Case_Stmt_Alternative; Info : Internal_Entity_Info)
         return Internal_Entity_Case_Stmt_Alternative is
      begin
         if Node = null then
            return No_Entity_Case_Stmt_Alternative;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Case_Stmt_Alternative) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Case_Stmt_Alternative_List
        (Node : Bare_Case_Stmt_Alternative_List; Info : Internal_Entity_Info)
         return Internal_Entity_Case_Stmt_Alternative_List is
      begin
         if Node = null then
            return No_Entity_Case_Stmt_Alternative_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Case_Stmt_Alternative_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Char_Literal
        (Node : Bare_Char_Literal; Info : Internal_Entity_Info)
         return Internal_Entity_Char_Literal is
      begin
         if Node = null then
            return No_Entity_Char_Literal;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Char_Literal) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Classwide_Type_Decl
        (Node : Bare_Classwide_Type_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Classwide_Type_Decl is
      begin
         if Node = null then
            return No_Entity_Classwide_Type_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Classwide_Type_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Compilation_Unit
        (Node : Bare_Compilation_Unit; Info : Internal_Entity_Info)
         return Internal_Entity_Compilation_Unit is
      begin
         if Node = null then
            return No_Entity_Compilation_Unit;
         end if;
         return (Node => Node, Info => Info);
      end;
      ----------
      -- Hash --
      ----------
      pragma Warnings (Off, "referenced");
      function Hash (R : Internal_Entity_Compilation_Unit) return Hash_Type is
         pragma Warnings (On, "referenced");
      begin
            return Combine
              (Hash (R.Node), Hash (R.Info));
      end Hash;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Compilation_Unit) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Compilation_Unit_List
        (Node : Bare_Compilation_Unit_List; Info : Internal_Entity_Info)
         return Internal_Entity_Compilation_Unit_List is
      begin
         if Node = null then
            return No_Entity_Compilation_Unit_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Compilation_Unit_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Component_Clause
        (Node : Bare_Component_Clause; Info : Internal_Entity_Info)
         return Internal_Entity_Component_Clause is
      begin
         if Node = null then
            return No_Entity_Component_Clause;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Component_Clause) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Component_Decl
        (Node : Bare_Component_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Component_Decl is
      begin
         if Node = null then
            return No_Entity_Component_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Component_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Component_Def
        (Node : Bare_Component_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Component_Def is
      begin
         if Node = null then
            return No_Entity_Component_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Component_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Component_List
        (Node : Bare_Component_List; Info : Internal_Entity_Info)
         return Internal_Entity_Component_List is
      begin
         if Node = null then
            return No_Entity_Component_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Component_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Constraint
        (Node : Bare_Constraint; Info : Internal_Entity_Info)
         return Internal_Entity_Constraint is
      begin
         if Node = null then
            return No_Entity_Constraint;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Constraint) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Composite_Constraint
        (Node : Bare_Composite_Constraint; Info : Internal_Entity_Info)
         return Internal_Entity_Composite_Constraint is
      begin
         if Node = null then
            return No_Entity_Composite_Constraint;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Composite_Constraint) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Composite_Constraint_Assoc
        (Node : Bare_Composite_Constraint_Assoc; Info : Internal_Entity_Info)
         return Internal_Entity_Composite_Constraint_Assoc is
      begin
         if Node = null then
            return No_Entity_Composite_Constraint_Assoc;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Composite_Constraint_Assoc) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Concrete_Formal_Subp_Decl
        (Node : Bare_Concrete_Formal_Subp_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Concrete_Formal_Subp_Decl is
      begin
         if Node = null then
            return No_Entity_Concrete_Formal_Subp_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Concrete_Formal_Subp_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Constant_Node
        (Node : Bare_Constant_Node; Info : Internal_Entity_Info)
         return Internal_Entity_Constant_Node is
      begin
         if Node = null then
            return No_Entity_Constant_Node;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Constant_Node) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Constant_Absent
        (Node : Bare_Constant_Absent; Info : Internal_Entity_Info)
         return Internal_Entity_Constant_Absent is
      begin
         if Node = null then
            return No_Entity_Constant_Absent;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Constant_Absent) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Constant_Present
        (Node : Bare_Constant_Present; Info : Internal_Entity_Info)
         return Internal_Entity_Constant_Present is
      begin
         if Node = null then
            return No_Entity_Constant_Present;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Constant_Present) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Constrained_Array_Indices
        (Node : Bare_Constrained_Array_Indices; Info : Internal_Entity_Info)
         return Internal_Entity_Constrained_Array_Indices is
      begin
         if Node = null then
            return No_Entity_Constrained_Array_Indices;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Constrained_Array_Indices) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Subtype_Indication
        (Node : Bare_Subtype_Indication; Info : Internal_Entity_Info)
         return Internal_Entity_Subtype_Indication is
      begin
         if Node = null then
            return No_Entity_Subtype_Indication;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Subtype_Indication) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Constrained_Subtype_Indication
        (Node : Bare_Constrained_Subtype_Indication; Info : Internal_Entity_Info)
         return Internal_Entity_Constrained_Subtype_Indication is
      begin
         if Node = null then
            return No_Entity_Constrained_Subtype_Indication;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Constrained_Subtype_Indication) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Constraint_List
        (Node : Bare_Constraint_List; Info : Internal_Entity_Info)
         return Internal_Entity_Constraint_List is
      begin
         if Node = null then
            return No_Entity_Constraint_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Constraint_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Contract_Case_Assoc
        (Node : Bare_Contract_Case_Assoc; Info : Internal_Entity_Info)
         return Internal_Entity_Contract_Case_Assoc is
      begin
         if Node = null then
            return No_Entity_Contract_Case_Assoc;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Contract_Case_Assoc) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Contract_Case_Assoc_List
        (Node : Bare_Contract_Case_Assoc_List; Info : Internal_Entity_Info)
         return Internal_Entity_Contract_Case_Assoc_List is
      begin
         if Node = null then
            return No_Entity_Contract_Case_Assoc_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Contract_Case_Assoc_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Contract_Cases
        (Node : Bare_Contract_Cases; Info : Internal_Entity_Info)
         return Internal_Entity_Contract_Cases is
      begin
         if Node = null then
            return No_Entity_Contract_Cases;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Contract_Cases) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Real_Type_Def
        (Node : Bare_Real_Type_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Real_Type_Def is
      begin
         if Node = null then
            return No_Entity_Real_Type_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Real_Type_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Decimal_Fixed_Point_Def
        (Node : Bare_Decimal_Fixed_Point_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Decimal_Fixed_Point_Def is
      begin
         if Node = null then
            return No_Entity_Decimal_Fixed_Point_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Decimal_Fixed_Point_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Decl_Block
        (Node : Bare_Decl_Block; Info : Internal_Entity_Info)
         return Internal_Entity_Decl_Block is
      begin
         if Node = null then
            return No_Entity_Decl_Block;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Decl_Block) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Decl_Expr
        (Node : Bare_Decl_Expr; Info : Internal_Entity_Info)
         return Internal_Entity_Decl_Expr is
      begin
         if Node = null then
            return No_Entity_Decl_Expr;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Decl_Expr) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Decl_List
        (Node : Bare_Decl_List; Info : Internal_Entity_Info)
         return Internal_Entity_Decl_List is
      begin
         if Node = null then
            return No_Entity_Decl_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Decl_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Declarative_Part
        (Node : Bare_Declarative_Part; Info : Internal_Entity_Info)
         return Internal_Entity_Declarative_Part is
      begin
         if Node = null then
            return No_Entity_Declarative_Part;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Declarative_Part) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Defining_Name
        (Node : Bare_Defining_Name; Info : Internal_Entity_Info)
         return Internal_Entity_Defining_Name is
      begin
         if Node = null then
            return No_Entity_Defining_Name;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Defining_Name) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Defining_Name_List
        (Node : Bare_Defining_Name_List; Info : Internal_Entity_Info)
         return Internal_Entity_Defining_Name_List is
      begin
         if Node = null then
            return No_Entity_Defining_Name_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Defining_Name_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Delay_Stmt
        (Node : Bare_Delay_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Delay_Stmt is
      begin
         if Node = null then
            return No_Entity_Delay_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Delay_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Delta_Constraint
        (Node : Bare_Delta_Constraint; Info : Internal_Entity_Info)
         return Internal_Entity_Delta_Constraint is
      begin
         if Node = null then
            return No_Entity_Delta_Constraint;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Delta_Constraint) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Derived_Type_Def
        (Node : Bare_Derived_Type_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Derived_Type_Def is
      begin
         if Node = null then
            return No_Entity_Derived_Type_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Derived_Type_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Digits_Constraint
        (Node : Bare_Digits_Constraint; Info : Internal_Entity_Info)
         return Internal_Entity_Digits_Constraint is
      begin
         if Node = null then
            return No_Entity_Digits_Constraint;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Digits_Constraint) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Discrete_Base_Subtype_Decl
        (Node : Bare_Discrete_Base_Subtype_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Discrete_Base_Subtype_Decl is
      begin
         if Node = null then
            return No_Entity_Discrete_Base_Subtype_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Discrete_Base_Subtype_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Discrete_Subtype_Indication
        (Node : Bare_Discrete_Subtype_Indication; Info : Internal_Entity_Info)
         return Internal_Entity_Discrete_Subtype_Indication is
      begin
         if Node = null then
            return No_Entity_Discrete_Subtype_Indication;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Discrete_Subtype_Indication) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Discrete_Subtype_Name
        (Node : Bare_Discrete_Subtype_Name; Info : Internal_Entity_Info)
         return Internal_Entity_Discrete_Subtype_Name is
      begin
         if Node = null then
            return No_Entity_Discrete_Subtype_Name;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Discrete_Subtype_Name) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Identifier_List
        (Node : Bare_Identifier_List; Info : Internal_Entity_Info)
         return Internal_Entity_Identifier_List is
      begin
         if Node = null then
            return No_Entity_Identifier_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Identifier_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Discriminant_Choice_List
        (Node : Bare_Discriminant_Choice_List; Info : Internal_Entity_Info)
         return Internal_Entity_Discriminant_Choice_List is
      begin
         if Node = null then
            return No_Entity_Discriminant_Choice_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Discriminant_Choice_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Discriminant_Part
        (Node : Bare_Discriminant_Part; Info : Internal_Entity_Info)
         return Internal_Entity_Discriminant_Part is
      begin
         if Node = null then
            return No_Entity_Discriminant_Part;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Discriminant_Part) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Discriminant_Spec
        (Node : Bare_Discriminant_Spec; Info : Internal_Entity_Info)
         return Internal_Entity_Discriminant_Spec is
      begin
         if Node = null then
            return No_Entity_Discriminant_Spec;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Discriminant_Spec) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Discriminant_Spec_List
        (Node : Bare_Discriminant_Spec_List; Info : Internal_Entity_Info)
         return Internal_Entity_Discriminant_Spec_List is
      begin
         if Node = null then
            return No_Entity_Discriminant_Spec_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Discriminant_Spec_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Dotted_Name
        (Node : Bare_Dotted_Name; Info : Internal_Entity_Info)
         return Internal_Entity_Dotted_Name is
      begin
         if Node = null then
            return No_Entity_Dotted_Name;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Dotted_Name) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Elsif_Expr_Part
        (Node : Bare_Elsif_Expr_Part; Info : Internal_Entity_Info)
         return Internal_Entity_Elsif_Expr_Part is
      begin
         if Node = null then
            return No_Entity_Elsif_Expr_Part;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Elsif_Expr_Part) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Elsif_Expr_Part_List
        (Node : Bare_Elsif_Expr_Part_List; Info : Internal_Entity_Info)
         return Internal_Entity_Elsif_Expr_Part_List is
      begin
         if Node = null then
            return No_Entity_Elsif_Expr_Part_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Elsif_Expr_Part_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Elsif_Stmt_Part
        (Node : Bare_Elsif_Stmt_Part; Info : Internal_Entity_Info)
         return Internal_Entity_Elsif_Stmt_Part is
      begin
         if Node = null then
            return No_Entity_Elsif_Stmt_Part;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Elsif_Stmt_Part) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Elsif_Stmt_Part_List
        (Node : Bare_Elsif_Stmt_Part_List; Info : Internal_Entity_Info)
         return Internal_Entity_Elsif_Stmt_Part_List is
      begin
         if Node = null then
            return No_Entity_Elsif_Stmt_Part_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Elsif_Stmt_Part_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_End_Name
        (Node : Bare_End_Name; Info : Internal_Entity_Info)
         return Internal_Entity_End_Name is
      begin
         if Node = null then
            return No_Entity_End_Name;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_End_Name) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Entry_Body
        (Node : Bare_Entry_Body; Info : Internal_Entity_Info)
         return Internal_Entity_Entry_Body is
      begin
         if Node = null then
            return No_Entity_Entry_Body;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Entry_Body) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Entry_Completion_Formal_Params
        (Node : Bare_Entry_Completion_Formal_Params; Info : Internal_Entity_Info)
         return Internal_Entity_Entry_Completion_Formal_Params is
      begin
         if Node = null then
            return No_Entity_Entry_Completion_Formal_Params;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Entry_Completion_Formal_Params) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Entry_Decl
        (Node : Bare_Entry_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Entry_Decl is
      begin
         if Node = null then
            return No_Entity_Entry_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Entry_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Entry_Index_Spec
        (Node : Bare_Entry_Index_Spec; Info : Internal_Entity_Info)
         return Internal_Entity_Entry_Index_Spec is
      begin
         if Node = null then
            return No_Entity_Entry_Index_Spec;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Entry_Index_Spec) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Entry_Spec
        (Node : Bare_Entry_Spec; Info : Internal_Entity_Info)
         return Internal_Entity_Entry_Spec is
      begin
         if Node = null then
            return No_Entity_Entry_Spec;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Entry_Spec) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Enum_Lit_Synth_Type_Expr
        (Node : Bare_Enum_Lit_Synth_Type_Expr; Info : Internal_Entity_Info)
         return Internal_Entity_Enum_Lit_Synth_Type_Expr is
      begin
         if Node = null then
            return No_Entity_Enum_Lit_Synth_Type_Expr;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Enum_Lit_Synth_Type_Expr) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Enum_Literal_Decl
        (Node : Bare_Enum_Literal_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Enum_Literal_Decl is
      begin
         if Node = null then
            return No_Entity_Enum_Literal_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Enum_Literal_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Enum_Literal_Decl_List
        (Node : Bare_Enum_Literal_Decl_List; Info : Internal_Entity_Info)
         return Internal_Entity_Enum_Literal_Decl_List is
      begin
         if Node = null then
            return No_Entity_Enum_Literal_Decl_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Enum_Literal_Decl_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Enum_Rep_Clause
        (Node : Bare_Enum_Rep_Clause; Info : Internal_Entity_Info)
         return Internal_Entity_Enum_Rep_Clause is
      begin
         if Node = null then
            return No_Entity_Enum_Rep_Clause;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Enum_Rep_Clause) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Enum_Subp_Spec
        (Node : Bare_Enum_Subp_Spec; Info : Internal_Entity_Info)
         return Internal_Entity_Enum_Subp_Spec is
      begin
         if Node = null then
            return No_Entity_Enum_Subp_Spec;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Enum_Subp_Spec) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Enum_Type_Def
        (Node : Bare_Enum_Type_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Enum_Type_Def is
      begin
         if Node = null then
            return No_Entity_Enum_Type_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Enum_Type_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Error_Decl
        (Node : Bare_Error_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Error_Decl is
      begin
         if Node = null then
            return No_Entity_Error_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Error_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Error_Stmt
        (Node : Bare_Error_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Error_Stmt is
      begin
         if Node = null then
            return No_Entity_Error_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Error_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Exception_Decl
        (Node : Bare_Exception_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Exception_Decl is
      begin
         if Node = null then
            return No_Entity_Exception_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Exception_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Exception_Handler
        (Node : Bare_Exception_Handler; Info : Internal_Entity_Info)
         return Internal_Entity_Exception_Handler is
      begin
         if Node = null then
            return No_Entity_Exception_Handler;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Exception_Handler) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Exit_Stmt
        (Node : Bare_Exit_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Exit_Stmt is
      begin
         if Node = null then
            return No_Entity_Exit_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Exit_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Explicit_Deref
        (Node : Bare_Explicit_Deref; Info : Internal_Entity_Info)
         return Internal_Entity_Explicit_Deref is
      begin
         if Node = null then
            return No_Entity_Explicit_Deref;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Explicit_Deref) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Expr_List
        (Node : Bare_Expr_List; Info : Internal_Entity_Info)
         return Internal_Entity_Expr_List is
      begin
         if Node = null then
            return No_Entity_Expr_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Expr_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Expr_Alternatives_List
        (Node : Bare_Expr_Alternatives_List; Info : Internal_Entity_Info)
         return Internal_Entity_Expr_Alternatives_List is
      begin
         if Node = null then
            return No_Entity_Expr_Alternatives_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Expr_Alternatives_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Expr_Function
        (Node : Bare_Expr_Function; Info : Internal_Entity_Info)
         return Internal_Entity_Expr_Function is
      begin
         if Node = null then
            return No_Entity_Expr_Function;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Expr_Function) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Extended_Return_Stmt
        (Node : Bare_Extended_Return_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Extended_Return_Stmt is
      begin
         if Node = null then
            return No_Entity_Extended_Return_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Extended_Return_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Object_Decl
        (Node : Bare_Object_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Object_Decl is
      begin
         if Node = null then
            return No_Entity_Object_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Object_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Extended_Return_Stmt_Object_Decl
        (Node : Bare_Extended_Return_Stmt_Object_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Extended_Return_Stmt_Object_Decl is
      begin
         if Node = null then
            return No_Entity_Extended_Return_Stmt_Object_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Extended_Return_Stmt_Object_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Floating_Point_Def
        (Node : Bare_Floating_Point_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Floating_Point_Def is
      begin
         if Node = null then
            return No_Entity_Floating_Point_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Floating_Point_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Loop_Spec
        (Node : Bare_Loop_Spec; Info : Internal_Entity_Info)
         return Internal_Entity_Loop_Spec is
      begin
         if Node = null then
            return No_Entity_Loop_Spec;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Loop_Spec) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_For_Loop_Spec
        (Node : Bare_For_Loop_Spec; Info : Internal_Entity_Info)
         return Internal_Entity_For_Loop_Spec is
      begin
         if Node = null then
            return No_Entity_For_Loop_Spec;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_For_Loop_Spec) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_For_Loop_Stmt
        (Node : Bare_For_Loop_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_For_Loop_Stmt is
      begin
         if Node = null then
            return No_Entity_For_Loop_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_For_Loop_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_For_Loop_Var_Decl
        (Node : Bare_For_Loop_Var_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_For_Loop_Var_Decl is
      begin
         if Node = null then
            return No_Entity_For_Loop_Var_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_For_Loop_Var_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Formal_Discrete_Type_Def
        (Node : Bare_Formal_Discrete_Type_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Formal_Discrete_Type_Def is
      begin
         if Node = null then
            return No_Entity_Formal_Discrete_Type_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Formal_Discrete_Type_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Generic_Decl
        (Node : Bare_Generic_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Generic_Decl is
      begin
         if Node = null then
            return No_Entity_Generic_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Generic_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Generic_Formal
        (Node : Bare_Generic_Formal; Info : Internal_Entity_Info)
         return Internal_Entity_Generic_Formal is
      begin
         if Node = null then
            return No_Entity_Generic_Formal;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Generic_Formal) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Generic_Formal_Obj_Decl
        (Node : Bare_Generic_Formal_Obj_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Generic_Formal_Obj_Decl is
      begin
         if Node = null then
            return No_Entity_Generic_Formal_Obj_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Generic_Formal_Obj_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Generic_Formal_Package
        (Node : Bare_Generic_Formal_Package; Info : Internal_Entity_Info)
         return Internal_Entity_Generic_Formal_Package is
      begin
         if Node = null then
            return No_Entity_Generic_Formal_Package;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Generic_Formal_Package) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Generic_Formal_Part
        (Node : Bare_Generic_Formal_Part; Info : Internal_Entity_Info)
         return Internal_Entity_Generic_Formal_Part is
      begin
         if Node = null then
            return No_Entity_Generic_Formal_Part;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Generic_Formal_Part) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Generic_Formal_Subp_Decl
        (Node : Bare_Generic_Formal_Subp_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Generic_Formal_Subp_Decl is
      begin
         if Node = null then
            return No_Entity_Generic_Formal_Subp_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Generic_Formal_Subp_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Generic_Formal_Type_Decl
        (Node : Bare_Generic_Formal_Type_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Generic_Formal_Type_Decl is
      begin
         if Node = null then
            return No_Entity_Generic_Formal_Type_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Generic_Formal_Type_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Generic_Instantiation
        (Node : Bare_Generic_Instantiation; Info : Internal_Entity_Info)
         return Internal_Entity_Generic_Instantiation is
      begin
         if Node = null then
            return No_Entity_Generic_Instantiation;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Generic_Instantiation) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Generic_Package_Decl
        (Node : Bare_Generic_Package_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Generic_Package_Decl is
      begin
         if Node = null then
            return No_Entity_Generic_Package_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Generic_Package_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Generic_Package_Instantiation
        (Node : Bare_Generic_Package_Instantiation; Info : Internal_Entity_Info)
         return Internal_Entity_Generic_Package_Instantiation is
      begin
         if Node = null then
            return No_Entity_Generic_Package_Instantiation;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Generic_Package_Instantiation) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Generic_Package_Internal
        (Node : Bare_Generic_Package_Internal; Info : Internal_Entity_Info)
         return Internal_Entity_Generic_Package_Internal is
      begin
         if Node = null then
            return No_Entity_Generic_Package_Internal;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Generic_Package_Internal) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Generic_Renaming_Decl
        (Node : Bare_Generic_Renaming_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Generic_Renaming_Decl is
      begin
         if Node = null then
            return No_Entity_Generic_Renaming_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Generic_Renaming_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Generic_Package_Renaming_Decl
        (Node : Bare_Generic_Package_Renaming_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Generic_Package_Renaming_Decl is
      begin
         if Node = null then
            return No_Entity_Generic_Package_Renaming_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Generic_Package_Renaming_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Generic_Subp_Decl
        (Node : Bare_Generic_Subp_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Generic_Subp_Decl is
      begin
         if Node = null then
            return No_Entity_Generic_Subp_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Generic_Subp_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Generic_Subp_Instantiation
        (Node : Bare_Generic_Subp_Instantiation; Info : Internal_Entity_Info)
         return Internal_Entity_Generic_Subp_Instantiation is
      begin
         if Node = null then
            return No_Entity_Generic_Subp_Instantiation;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Generic_Subp_Instantiation) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Generic_Subp_Internal
        (Node : Bare_Generic_Subp_Internal; Info : Internal_Entity_Info)
         return Internal_Entity_Generic_Subp_Internal is
      begin
         if Node = null then
            return No_Entity_Generic_Subp_Internal;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Generic_Subp_Internal) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Generic_Subp_Renaming_Decl
        (Node : Bare_Generic_Subp_Renaming_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Generic_Subp_Renaming_Decl is
      begin
         if Node = null then
            return No_Entity_Generic_Subp_Renaming_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Generic_Subp_Renaming_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Goto_Stmt
        (Node : Bare_Goto_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Goto_Stmt is
      begin
         if Node = null then
            return No_Entity_Goto_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Goto_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Handled_Stmts
        (Node : Bare_Handled_Stmts; Info : Internal_Entity_Info)
         return Internal_Entity_Handled_Stmts is
      begin
         if Node = null then
            return No_Entity_Handled_Stmts;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Handled_Stmts) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_If_Expr
        (Node : Bare_If_Expr; Info : Internal_Entity_Info)
         return Internal_Entity_If_Expr is
      begin
         if Node = null then
            return No_Entity_If_Expr;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_If_Expr) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_If_Stmt
        (Node : Bare_If_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_If_Stmt is
      begin
         if Node = null then
            return No_Entity_If_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_If_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Incomplete_Type_Decl
        (Node : Bare_Incomplete_Type_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Incomplete_Type_Decl is
      begin
         if Node = null then
            return No_Entity_Incomplete_Type_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Incomplete_Type_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Incomplete_Tagged_Type_Decl
        (Node : Bare_Incomplete_Tagged_Type_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Incomplete_Tagged_Type_Decl is
      begin
         if Node = null then
            return No_Entity_Incomplete_Tagged_Type_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Incomplete_Tagged_Type_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Num_Literal
        (Node : Bare_Num_Literal; Info : Internal_Entity_Info)
         return Internal_Entity_Num_Literal is
      begin
         if Node = null then
            return No_Entity_Num_Literal;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Num_Literal) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Int_Literal
        (Node : Bare_Int_Literal; Info : Internal_Entity_Info)
         return Internal_Entity_Int_Literal is
      begin
         if Node = null then
            return No_Entity_Int_Literal;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Int_Literal) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Interface_Kind
        (Node : Bare_Interface_Kind; Info : Internal_Entity_Info)
         return Internal_Entity_Interface_Kind is
      begin
         if Node = null then
            return No_Entity_Interface_Kind;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Interface_Kind) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Interface_Kind_Limited
        (Node : Bare_Interface_Kind_Limited; Info : Internal_Entity_Info)
         return Internal_Entity_Interface_Kind_Limited is
      begin
         if Node = null then
            return No_Entity_Interface_Kind_Limited;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Interface_Kind_Limited) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Interface_Kind_Protected
        (Node : Bare_Interface_Kind_Protected; Info : Internal_Entity_Info)
         return Internal_Entity_Interface_Kind_Protected is
      begin
         if Node = null then
            return No_Entity_Interface_Kind_Protected;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Interface_Kind_Protected) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Interface_Kind_Synchronized
        (Node : Bare_Interface_Kind_Synchronized; Info : Internal_Entity_Info)
         return Internal_Entity_Interface_Kind_Synchronized is
      begin
         if Node = null then
            return No_Entity_Interface_Kind_Synchronized;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Interface_Kind_Synchronized) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Interface_Kind_Task
        (Node : Bare_Interface_Kind_Task; Info : Internal_Entity_Info)
         return Internal_Entity_Interface_Kind_Task is
      begin
         if Node = null then
            return No_Entity_Interface_Kind_Task;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Interface_Kind_Task) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Interface_Type_Def
        (Node : Bare_Interface_Type_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Interface_Type_Def is
      begin
         if Node = null then
            return No_Entity_Interface_Type_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Interface_Type_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Iter_Type
        (Node : Bare_Iter_Type; Info : Internal_Entity_Info)
         return Internal_Entity_Iter_Type is
      begin
         if Node = null then
            return No_Entity_Iter_Type;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Iter_Type) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Iter_Type_In
        (Node : Bare_Iter_Type_In; Info : Internal_Entity_Info)
         return Internal_Entity_Iter_Type_In is
      begin
         if Node = null then
            return No_Entity_Iter_Type_In;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Iter_Type_In) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Iter_Type_Of
        (Node : Bare_Iter_Type_Of; Info : Internal_Entity_Info)
         return Internal_Entity_Iter_Type_Of is
      begin
         if Node = null then
            return No_Entity_Iter_Type_Of;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Iter_Type_Of) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Iterated_Assoc
        (Node : Bare_Iterated_Assoc; Info : Internal_Entity_Info)
         return Internal_Entity_Iterated_Assoc is
      begin
         if Node = null then
            return No_Entity_Iterated_Assoc;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Iterated_Assoc) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Known_Discriminant_Part
        (Node : Bare_Known_Discriminant_Part; Info : Internal_Entity_Info)
         return Internal_Entity_Known_Discriminant_Part is
      begin
         if Node = null then
            return No_Entity_Known_Discriminant_Part;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Known_Discriminant_Part) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Label
        (Node : Bare_Label; Info : Internal_Entity_Info)
         return Internal_Entity_Label is
      begin
         if Node = null then
            return No_Entity_Label;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Label) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Label_Decl
        (Node : Bare_Label_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Label_Decl is
      begin
         if Node = null then
            return No_Entity_Label_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Label_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Library_Item
        (Node : Bare_Library_Item; Info : Internal_Entity_Info)
         return Internal_Entity_Library_Item is
      begin
         if Node = null then
            return No_Entity_Library_Item;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Library_Item) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Limited_Node
        (Node : Bare_Limited_Node; Info : Internal_Entity_Info)
         return Internal_Entity_Limited_Node is
      begin
         if Node = null then
            return No_Entity_Limited_Node;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Limited_Node) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Limited_Absent
        (Node : Bare_Limited_Absent; Info : Internal_Entity_Info)
         return Internal_Entity_Limited_Absent is
      begin
         if Node = null then
            return No_Entity_Limited_Absent;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Limited_Absent) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Limited_Present
        (Node : Bare_Limited_Present; Info : Internal_Entity_Info)
         return Internal_Entity_Limited_Present is
      begin
         if Node = null then
            return No_Entity_Limited_Present;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Limited_Present) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Loop_Stmt
        (Node : Bare_Loop_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Loop_Stmt is
      begin
         if Node = null then
            return No_Entity_Loop_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Loop_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Membership_Expr
        (Node : Bare_Membership_Expr; Info : Internal_Entity_Info)
         return Internal_Entity_Membership_Expr is
      begin
         if Node = null then
            return No_Entity_Membership_Expr;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Membership_Expr) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Mod_Int_Type_Def
        (Node : Bare_Mod_Int_Type_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Mod_Int_Type_Def is
      begin
         if Node = null then
            return No_Entity_Mod_Int_Type_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Mod_Int_Type_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Mode
        (Node : Bare_Mode; Info : Internal_Entity_Info)
         return Internal_Entity_Mode is
      begin
         if Node = null then
            return No_Entity_Mode;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Mode) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Mode_Default
        (Node : Bare_Mode_Default; Info : Internal_Entity_Info)
         return Internal_Entity_Mode_Default is
      begin
         if Node = null then
            return No_Entity_Mode_Default;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Mode_Default) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Mode_In
        (Node : Bare_Mode_In; Info : Internal_Entity_Info)
         return Internal_Entity_Mode_In is
      begin
         if Node = null then
            return No_Entity_Mode_In;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Mode_In) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Mode_In_Out
        (Node : Bare_Mode_In_Out; Info : Internal_Entity_Info)
         return Internal_Entity_Mode_In_Out is
      begin
         if Node = null then
            return No_Entity_Mode_In_Out;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Mode_In_Out) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Mode_Out
        (Node : Bare_Mode_Out; Info : Internal_Entity_Info)
         return Internal_Entity_Mode_Out is
      begin
         if Node = null then
            return No_Entity_Mode_Out;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Mode_Out) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Multi_Abstract_State_Decl
        (Node : Bare_Multi_Abstract_State_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Multi_Abstract_State_Decl is
      begin
         if Node = null then
            return No_Entity_Multi_Abstract_State_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Multi_Abstract_State_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Multi_Dim_Array_Assoc
        (Node : Bare_Multi_Dim_Array_Assoc; Info : Internal_Entity_Info)
         return Internal_Entity_Multi_Dim_Array_Assoc is
      begin
         if Node = null then
            return No_Entity_Multi_Dim_Array_Assoc;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Multi_Dim_Array_Assoc) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Name_List
        (Node : Bare_Name_List; Info : Internal_Entity_Info)
         return Internal_Entity_Name_List is
      begin
         if Node = null then
            return No_Entity_Name_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Name_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Named_Stmt
        (Node : Bare_Named_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Named_Stmt is
      begin
         if Node = null then
            return No_Entity_Named_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Named_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Named_Stmt_Decl
        (Node : Bare_Named_Stmt_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Named_Stmt_Decl is
      begin
         if Node = null then
            return No_Entity_Named_Stmt_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Named_Stmt_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_No_Type_Object_Renaming_Decl
        (Node : Bare_No_Type_Object_Renaming_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_No_Type_Object_Renaming_Decl is
      begin
         if Node = null then
            return No_Entity_No_Type_Object_Renaming_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_No_Type_Object_Renaming_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Not_Null
        (Node : Bare_Not_Null; Info : Internal_Entity_Info)
         return Internal_Entity_Not_Null is
      begin
         if Node = null then
            return No_Entity_Not_Null;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Not_Null) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Not_Null_Absent
        (Node : Bare_Not_Null_Absent; Info : Internal_Entity_Info)
         return Internal_Entity_Not_Null_Absent is
      begin
         if Node = null then
            return No_Entity_Not_Null_Absent;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Not_Null_Absent) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Not_Null_Present
        (Node : Bare_Not_Null_Present; Info : Internal_Entity_Info)
         return Internal_Entity_Not_Null_Present is
      begin
         if Node = null then
            return No_Entity_Not_Null_Present;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Not_Null_Present) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Null_Component_Decl
        (Node : Bare_Null_Component_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Null_Component_Decl is
      begin
         if Node = null then
            return No_Entity_Null_Component_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Null_Component_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Null_Literal
        (Node : Bare_Null_Literal; Info : Internal_Entity_Info)
         return Internal_Entity_Null_Literal is
      begin
         if Node = null then
            return No_Entity_Null_Literal;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Null_Literal) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Null_Record_Aggregate
        (Node : Bare_Null_Record_Aggregate; Info : Internal_Entity_Info)
         return Internal_Entity_Null_Record_Aggregate is
      begin
         if Node = null then
            return No_Entity_Null_Record_Aggregate;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Null_Record_Aggregate) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Null_Record_Def
        (Node : Bare_Null_Record_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Null_Record_Def is
      begin
         if Node = null then
            return No_Entity_Null_Record_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Null_Record_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Null_Stmt
        (Node : Bare_Null_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Null_Stmt is
      begin
         if Node = null then
            return No_Entity_Null_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Null_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Null_Subp_Decl
        (Node : Bare_Null_Subp_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Null_Subp_Decl is
      begin
         if Node = null then
            return No_Entity_Null_Subp_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Null_Subp_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Number_Decl
        (Node : Bare_Number_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Number_Decl is
      begin
         if Node = null then
            return No_Entity_Number_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Number_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op
        (Node : Bare_Op; Info : Internal_Entity_Info)
         return Internal_Entity_Op is
      begin
         if Node = null then
            return No_Entity_Op;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Abs
        (Node : Bare_Op_Abs; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Abs is
      begin
         if Node = null then
            return No_Entity_Op_Abs;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Abs) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_And
        (Node : Bare_Op_And; Info : Internal_Entity_Info)
         return Internal_Entity_Op_And is
      begin
         if Node = null then
            return No_Entity_Op_And;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_And) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_And_Then
        (Node : Bare_Op_And_Then; Info : Internal_Entity_Info)
         return Internal_Entity_Op_And_Then is
      begin
         if Node = null then
            return No_Entity_Op_And_Then;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_And_Then) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Concat
        (Node : Bare_Op_Concat; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Concat is
      begin
         if Node = null then
            return No_Entity_Op_Concat;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Concat) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Div
        (Node : Bare_Op_Div; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Div is
      begin
         if Node = null then
            return No_Entity_Op_Div;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Div) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Double_Dot
        (Node : Bare_Op_Double_Dot; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Double_Dot is
      begin
         if Node = null then
            return No_Entity_Op_Double_Dot;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Double_Dot) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Eq
        (Node : Bare_Op_Eq; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Eq is
      begin
         if Node = null then
            return No_Entity_Op_Eq;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Eq) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Gt
        (Node : Bare_Op_Gt; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Gt is
      begin
         if Node = null then
            return No_Entity_Op_Gt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Gt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Gte
        (Node : Bare_Op_Gte; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Gte is
      begin
         if Node = null then
            return No_Entity_Op_Gte;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Gte) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_In
        (Node : Bare_Op_In; Info : Internal_Entity_Info)
         return Internal_Entity_Op_In is
      begin
         if Node = null then
            return No_Entity_Op_In;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_In) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Lt
        (Node : Bare_Op_Lt; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Lt is
      begin
         if Node = null then
            return No_Entity_Op_Lt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Lt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Lte
        (Node : Bare_Op_Lte; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Lte is
      begin
         if Node = null then
            return No_Entity_Op_Lte;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Lte) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Minus
        (Node : Bare_Op_Minus; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Minus is
      begin
         if Node = null then
            return No_Entity_Op_Minus;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Minus) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Mod
        (Node : Bare_Op_Mod; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Mod is
      begin
         if Node = null then
            return No_Entity_Op_Mod;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Mod) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Mult
        (Node : Bare_Op_Mult; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Mult is
      begin
         if Node = null then
            return No_Entity_Op_Mult;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Mult) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Neq
        (Node : Bare_Op_Neq; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Neq is
      begin
         if Node = null then
            return No_Entity_Op_Neq;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Neq) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Not
        (Node : Bare_Op_Not; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Not is
      begin
         if Node = null then
            return No_Entity_Op_Not;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Not) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Not_In
        (Node : Bare_Op_Not_In; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Not_In is
      begin
         if Node = null then
            return No_Entity_Op_Not_In;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Not_In) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Or
        (Node : Bare_Op_Or; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Or is
      begin
         if Node = null then
            return No_Entity_Op_Or;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Or) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Or_Else
        (Node : Bare_Op_Or_Else; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Or_Else is
      begin
         if Node = null then
            return No_Entity_Op_Or_Else;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Or_Else) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Plus
        (Node : Bare_Op_Plus; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Plus is
      begin
         if Node = null then
            return No_Entity_Op_Plus;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Plus) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Pow
        (Node : Bare_Op_Pow; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Pow is
      begin
         if Node = null then
            return No_Entity_Op_Pow;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Pow) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Rem
        (Node : Bare_Op_Rem; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Rem is
      begin
         if Node = null then
            return No_Entity_Op_Rem;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Rem) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Op_Xor
        (Node : Bare_Op_Xor; Info : Internal_Entity_Info)
         return Internal_Entity_Op_Xor is
      begin
         if Node = null then
            return No_Entity_Op_Xor;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Op_Xor) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Ordinary_Fixed_Point_Def
        (Node : Bare_Ordinary_Fixed_Point_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Ordinary_Fixed_Point_Def is
      begin
         if Node = null then
            return No_Entity_Ordinary_Fixed_Point_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Ordinary_Fixed_Point_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Others_Designator
        (Node : Bare_Others_Designator; Info : Internal_Entity_Info)
         return Internal_Entity_Others_Designator is
      begin
         if Node = null then
            return No_Entity_Others_Designator;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Others_Designator) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Overriding_Node
        (Node : Bare_Overriding_Node; Info : Internal_Entity_Info)
         return Internal_Entity_Overriding_Node is
      begin
         if Node = null then
            return No_Entity_Overriding_Node;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Overriding_Node) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Overriding_Not_Overriding
        (Node : Bare_Overriding_Not_Overriding; Info : Internal_Entity_Info)
         return Internal_Entity_Overriding_Not_Overriding is
      begin
         if Node = null then
            return No_Entity_Overriding_Not_Overriding;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Overriding_Not_Overriding) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Overriding_Overriding
        (Node : Bare_Overriding_Overriding; Info : Internal_Entity_Info)
         return Internal_Entity_Overriding_Overriding is
      begin
         if Node = null then
            return No_Entity_Overriding_Overriding;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Overriding_Overriding) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Overriding_Unspecified
        (Node : Bare_Overriding_Unspecified; Info : Internal_Entity_Info)
         return Internal_Entity_Overriding_Unspecified is
      begin
         if Node = null then
            return No_Entity_Overriding_Unspecified;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Overriding_Unspecified) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Package_Body
        (Node : Bare_Package_Body; Info : Internal_Entity_Info)
         return Internal_Entity_Package_Body is
      begin
         if Node = null then
            return No_Entity_Package_Body;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Package_Body) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Package_Body_Stub
        (Node : Bare_Package_Body_Stub; Info : Internal_Entity_Info)
         return Internal_Entity_Package_Body_Stub is
      begin
         if Node = null then
            return No_Entity_Package_Body_Stub;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Package_Body_Stub) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Package_Decl
        (Node : Bare_Package_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Package_Decl is
      begin
         if Node = null then
            return No_Entity_Package_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Package_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Package_Renaming_Decl
        (Node : Bare_Package_Renaming_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Package_Renaming_Decl is
      begin
         if Node = null then
            return No_Entity_Package_Renaming_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Package_Renaming_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Param_Assoc
        (Node : Bare_Param_Assoc; Info : Internal_Entity_Info)
         return Internal_Entity_Param_Assoc is
      begin
         if Node = null then
            return No_Entity_Param_Assoc;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Param_Assoc) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Param_Spec
        (Node : Bare_Param_Spec; Info : Internal_Entity_Info)
         return Internal_Entity_Param_Spec is
      begin
         if Node = null then
            return No_Entity_Param_Spec;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Param_Spec) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Param_Spec_List
        (Node : Bare_Param_Spec_List; Info : Internal_Entity_Info)
         return Internal_Entity_Param_Spec_List is
      begin
         if Node = null then
            return No_Entity_Param_Spec_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Param_Spec_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Params
        (Node : Bare_Params; Info : Internal_Entity_Info)
         return Internal_Entity_Params is
      begin
         if Node = null then
            return No_Entity_Params;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Params) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Paren_Abstract_State_Decl
        (Node : Bare_Paren_Abstract_State_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Paren_Abstract_State_Decl is
      begin
         if Node = null then
            return No_Entity_Paren_Abstract_State_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Paren_Abstract_State_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Paren_Expr
        (Node : Bare_Paren_Expr; Info : Internal_Entity_Info)
         return Internal_Entity_Paren_Expr is
      begin
         if Node = null then
            return No_Entity_Paren_Expr;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Paren_Expr) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Parent_List
        (Node : Bare_Parent_List; Info : Internal_Entity_Info)
         return Internal_Entity_Parent_List is
      begin
         if Node = null then
            return No_Entity_Parent_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Parent_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Pp_Directive
        (Node : Bare_Pp_Directive; Info : Internal_Entity_Info)
         return Internal_Entity_Pp_Directive is
      begin
         if Node = null then
            return No_Entity_Pp_Directive;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Pp_Directive) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Pp_Else_Directive
        (Node : Bare_Pp_Else_Directive; Info : Internal_Entity_Info)
         return Internal_Entity_Pp_Else_Directive is
      begin
         if Node = null then
            return No_Entity_Pp_Else_Directive;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Pp_Else_Directive) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Pp_Elsif_Directive
        (Node : Bare_Pp_Elsif_Directive; Info : Internal_Entity_Info)
         return Internal_Entity_Pp_Elsif_Directive is
      begin
         if Node = null then
            return No_Entity_Pp_Elsif_Directive;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Pp_Elsif_Directive) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Pp_End_If_Directive
        (Node : Bare_Pp_End_If_Directive; Info : Internal_Entity_Info)
         return Internal_Entity_Pp_End_If_Directive is
      begin
         if Node = null then
            return No_Entity_Pp_End_If_Directive;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Pp_End_If_Directive) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Pp_If_Directive
        (Node : Bare_Pp_If_Directive; Info : Internal_Entity_Info)
         return Internal_Entity_Pp_If_Directive is
      begin
         if Node = null then
            return No_Entity_Pp_If_Directive;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Pp_If_Directive) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Pp_Then_Kw
        (Node : Bare_Pp_Then_Kw; Info : Internal_Entity_Info)
         return Internal_Entity_Pp_Then_Kw is
      begin
         if Node = null then
            return No_Entity_Pp_Then_Kw;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Pp_Then_Kw) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Pragma_Argument_Assoc
        (Node : Bare_Pragma_Argument_Assoc; Info : Internal_Entity_Info)
         return Internal_Entity_Pragma_Argument_Assoc is
      begin
         if Node = null then
            return No_Entity_Pragma_Argument_Assoc;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Pragma_Argument_Assoc) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Pragma_Node
        (Node : Bare_Pragma_Node; Info : Internal_Entity_Info)
         return Internal_Entity_Pragma_Node is
      begin
         if Node = null then
            return No_Entity_Pragma_Node;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Pragma_Node) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Pragma_Node_List
        (Node : Bare_Pragma_Node_List; Info : Internal_Entity_Info)
         return Internal_Entity_Pragma_Node_List is
      begin
         if Node = null then
            return No_Entity_Pragma_Node_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Pragma_Node_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Predefined_Bin_Op_Spec
        (Node : Bare_Predefined_Bin_Op_Spec; Info : Internal_Entity_Info)
         return Internal_Entity_Predefined_Bin_Op_Spec is
      begin
         if Node = null then
            return No_Entity_Predefined_Bin_Op_Spec;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Predefined_Bin_Op_Spec) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Predefined_Un_Op_Spec
        (Node : Bare_Predefined_Un_Op_Spec; Info : Internal_Entity_Info)
         return Internal_Entity_Predefined_Un_Op_Spec is
      begin
         if Node = null then
            return No_Entity_Predefined_Un_Op_Spec;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Predefined_Un_Op_Spec) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Private_Node
        (Node : Bare_Private_Node; Info : Internal_Entity_Info)
         return Internal_Entity_Private_Node is
      begin
         if Node = null then
            return No_Entity_Private_Node;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Private_Node) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Private_Absent
        (Node : Bare_Private_Absent; Info : Internal_Entity_Info)
         return Internal_Entity_Private_Absent is
      begin
         if Node = null then
            return No_Entity_Private_Absent;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Private_Absent) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Private_Part
        (Node : Bare_Private_Part; Info : Internal_Entity_Info)
         return Internal_Entity_Private_Part is
      begin
         if Node = null then
            return No_Entity_Private_Part;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Private_Part) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Private_Present
        (Node : Bare_Private_Present; Info : Internal_Entity_Info)
         return Internal_Entity_Private_Present is
      begin
         if Node = null then
            return No_Entity_Private_Present;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Private_Present) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Private_Type_Def
        (Node : Bare_Private_Type_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Private_Type_Def is
      begin
         if Node = null then
            return No_Entity_Private_Type_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Private_Type_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Protected_Node
        (Node : Bare_Protected_Node; Info : Internal_Entity_Info)
         return Internal_Entity_Protected_Node is
      begin
         if Node = null then
            return No_Entity_Protected_Node;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Protected_Node) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Protected_Absent
        (Node : Bare_Protected_Absent; Info : Internal_Entity_Info)
         return Internal_Entity_Protected_Absent is
      begin
         if Node = null then
            return No_Entity_Protected_Absent;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Protected_Absent) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Protected_Body
        (Node : Bare_Protected_Body; Info : Internal_Entity_Info)
         return Internal_Entity_Protected_Body is
      begin
         if Node = null then
            return No_Entity_Protected_Body;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Protected_Body) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Protected_Body_Stub
        (Node : Bare_Protected_Body_Stub; Info : Internal_Entity_Info)
         return Internal_Entity_Protected_Body_Stub is
      begin
         if Node = null then
            return No_Entity_Protected_Body_Stub;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Protected_Body_Stub) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Protected_Def
        (Node : Bare_Protected_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Protected_Def is
      begin
         if Node = null then
            return No_Entity_Protected_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Protected_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Protected_Present
        (Node : Bare_Protected_Present; Info : Internal_Entity_Info)
         return Internal_Entity_Protected_Present is
      begin
         if Node = null then
            return No_Entity_Protected_Present;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Protected_Present) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Protected_Type_Decl
        (Node : Bare_Protected_Type_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Protected_Type_Decl is
      begin
         if Node = null then
            return No_Entity_Protected_Type_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Protected_Type_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Public_Part
        (Node : Bare_Public_Part; Info : Internal_Entity_Info)
         return Internal_Entity_Public_Part is
      begin
         if Node = null then
            return No_Entity_Public_Part;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Public_Part) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Qual_Expr
        (Node : Bare_Qual_Expr; Info : Internal_Entity_Info)
         return Internal_Entity_Qual_Expr is
      begin
         if Node = null then
            return No_Entity_Qual_Expr;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Qual_Expr) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Quantified_Expr
        (Node : Bare_Quantified_Expr; Info : Internal_Entity_Info)
         return Internal_Entity_Quantified_Expr is
      begin
         if Node = null then
            return No_Entity_Quantified_Expr;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Quantified_Expr) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Quantifier
        (Node : Bare_Quantifier; Info : Internal_Entity_Info)
         return Internal_Entity_Quantifier is
      begin
         if Node = null then
            return No_Entity_Quantifier;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Quantifier) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Quantifier_All
        (Node : Bare_Quantifier_All; Info : Internal_Entity_Info)
         return Internal_Entity_Quantifier_All is
      begin
         if Node = null then
            return No_Entity_Quantifier_All;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Quantifier_All) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Quantifier_Some
        (Node : Bare_Quantifier_Some; Info : Internal_Entity_Info)
         return Internal_Entity_Quantifier_Some is
      begin
         if Node = null then
            return No_Entity_Quantifier_Some;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Quantifier_Some) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Raise_Expr
        (Node : Bare_Raise_Expr; Info : Internal_Entity_Info)
         return Internal_Entity_Raise_Expr is
      begin
         if Node = null then
            return No_Entity_Raise_Expr;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Raise_Expr) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Raise_Stmt
        (Node : Bare_Raise_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Raise_Stmt is
      begin
         if Node = null then
            return No_Entity_Raise_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Raise_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Range_Constraint
        (Node : Bare_Range_Constraint; Info : Internal_Entity_Info)
         return Internal_Entity_Range_Constraint is
      begin
         if Node = null then
            return No_Entity_Range_Constraint;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Range_Constraint) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Range_Spec
        (Node : Bare_Range_Spec; Info : Internal_Entity_Info)
         return Internal_Entity_Range_Spec is
      begin
         if Node = null then
            return No_Entity_Range_Spec;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Range_Spec) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Real_Literal
        (Node : Bare_Real_Literal; Info : Internal_Entity_Info)
         return Internal_Entity_Real_Literal is
      begin
         if Node = null then
            return No_Entity_Real_Literal;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Real_Literal) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Record_Def
        (Node : Bare_Record_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Record_Def is
      begin
         if Node = null then
            return No_Entity_Record_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Record_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Record_Rep_Clause
        (Node : Bare_Record_Rep_Clause; Info : Internal_Entity_Info)
         return Internal_Entity_Record_Rep_Clause is
      begin
         if Node = null then
            return No_Entity_Record_Rep_Clause;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Record_Rep_Clause) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Record_Type_Def
        (Node : Bare_Record_Type_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Record_Type_Def is
      begin
         if Node = null then
            return No_Entity_Record_Type_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Record_Type_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Reduce_Attribute_Ref
        (Node : Bare_Reduce_Attribute_Ref; Info : Internal_Entity_Info)
         return Internal_Entity_Reduce_Attribute_Ref is
      begin
         if Node = null then
            return No_Entity_Reduce_Attribute_Ref;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Reduce_Attribute_Ref) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Relation_Op
        (Node : Bare_Relation_Op; Info : Internal_Entity_Info)
         return Internal_Entity_Relation_Op is
      begin
         if Node = null then
            return No_Entity_Relation_Op;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Relation_Op) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Renaming_Clause
        (Node : Bare_Renaming_Clause; Info : Internal_Entity_Info)
         return Internal_Entity_Renaming_Clause is
      begin
         if Node = null then
            return No_Entity_Renaming_Clause;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Renaming_Clause) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Requeue_Stmt
        (Node : Bare_Requeue_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Requeue_Stmt is
      begin
         if Node = null then
            return No_Entity_Requeue_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Requeue_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Return_Stmt
        (Node : Bare_Return_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Return_Stmt is
      begin
         if Node = null then
            return No_Entity_Return_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Return_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Reverse_Node
        (Node : Bare_Reverse_Node; Info : Internal_Entity_Info)
         return Internal_Entity_Reverse_Node is
      begin
         if Node = null then
            return No_Entity_Reverse_Node;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Reverse_Node) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Reverse_Absent
        (Node : Bare_Reverse_Absent; Info : Internal_Entity_Info)
         return Internal_Entity_Reverse_Absent is
      begin
         if Node = null then
            return No_Entity_Reverse_Absent;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Reverse_Absent) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Reverse_Present
        (Node : Bare_Reverse_Present; Info : Internal_Entity_Info)
         return Internal_Entity_Reverse_Present is
      begin
         if Node = null then
            return No_Entity_Reverse_Present;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Reverse_Present) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Select_Stmt
        (Node : Bare_Select_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_Select_Stmt is
      begin
         if Node = null then
            return No_Entity_Select_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Select_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Select_When_Part
        (Node : Bare_Select_When_Part; Info : Internal_Entity_Info)
         return Internal_Entity_Select_When_Part is
      begin
         if Node = null then
            return No_Entity_Select_When_Part;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Select_When_Part) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Select_When_Part_List
        (Node : Bare_Select_When_Part_List; Info : Internal_Entity_Info)
         return Internal_Entity_Select_When_Part_List is
      begin
         if Node = null then
            return No_Entity_Select_When_Part_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Select_When_Part_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Signed_Int_Type_Def
        (Node : Bare_Signed_Int_Type_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Signed_Int_Type_Def is
      begin
         if Node = null then
            return No_Entity_Signed_Int_Type_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Signed_Int_Type_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Single_Protected_Decl
        (Node : Bare_Single_Protected_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Single_Protected_Decl is
      begin
         if Node = null then
            return No_Entity_Single_Protected_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Single_Protected_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Single_Task_Decl
        (Node : Bare_Single_Task_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Single_Task_Decl is
      begin
         if Node = null then
            return No_Entity_Single_Task_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Single_Task_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Task_Type_Decl
        (Node : Bare_Task_Type_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Task_Type_Decl is
      begin
         if Node = null then
            return No_Entity_Task_Type_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Task_Type_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Single_Task_Type_Decl
        (Node : Bare_Single_Task_Type_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Single_Task_Type_Decl is
      begin
         if Node = null then
            return No_Entity_Single_Task_Type_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Single_Task_Type_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Stmt_List
        (Node : Bare_Stmt_List; Info : Internal_Entity_Info)
         return Internal_Entity_Stmt_List is
      begin
         if Node = null then
            return No_Entity_Stmt_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Stmt_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_String_Literal
        (Node : Bare_String_Literal; Info : Internal_Entity_Info)
         return Internal_Entity_String_Literal is
      begin
         if Node = null then
            return No_Entity_String_Literal;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_String_Literal) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Subp_Body
        (Node : Bare_Subp_Body; Info : Internal_Entity_Info)
         return Internal_Entity_Subp_Body is
      begin
         if Node = null then
            return No_Entity_Subp_Body;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Subp_Body) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Subp_Body_Stub
        (Node : Bare_Subp_Body_Stub; Info : Internal_Entity_Info)
         return Internal_Entity_Subp_Body_Stub is
      begin
         if Node = null then
            return No_Entity_Subp_Body_Stub;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Subp_Body_Stub) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Subp_Decl
        (Node : Bare_Subp_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Subp_Decl is
      begin
         if Node = null then
            return No_Entity_Subp_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Subp_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Subp_Kind
        (Node : Bare_Subp_Kind; Info : Internal_Entity_Info)
         return Internal_Entity_Subp_Kind is
      begin
         if Node = null then
            return No_Entity_Subp_Kind;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Subp_Kind) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Subp_Kind_Function
        (Node : Bare_Subp_Kind_Function; Info : Internal_Entity_Info)
         return Internal_Entity_Subp_Kind_Function is
      begin
         if Node = null then
            return No_Entity_Subp_Kind_Function;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Subp_Kind_Function) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Subp_Kind_Procedure
        (Node : Bare_Subp_Kind_Procedure; Info : Internal_Entity_Info)
         return Internal_Entity_Subp_Kind_Procedure is
      begin
         if Node = null then
            return No_Entity_Subp_Kind_Procedure;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Subp_Kind_Procedure) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Subp_Renaming_Decl
        (Node : Bare_Subp_Renaming_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Subp_Renaming_Decl is
      begin
         if Node = null then
            return No_Entity_Subp_Renaming_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Subp_Renaming_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Subp_Spec
        (Node : Bare_Subp_Spec; Info : Internal_Entity_Info)
         return Internal_Entity_Subp_Spec is
      begin
         if Node = null then
            return No_Entity_Subp_Spec;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Subp_Spec) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Subtype_Decl
        (Node : Bare_Subtype_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Subtype_Decl is
      begin
         if Node = null then
            return No_Entity_Subtype_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Subtype_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Subunit
        (Node : Bare_Subunit; Info : Internal_Entity_Info)
         return Internal_Entity_Subunit is
      begin
         if Node = null then
            return No_Entity_Subunit;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Subunit) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Synchronized_Node
        (Node : Bare_Synchronized_Node; Info : Internal_Entity_Info)
         return Internal_Entity_Synchronized_Node is
      begin
         if Node = null then
            return No_Entity_Synchronized_Node;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Synchronized_Node) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Synchronized_Absent
        (Node : Bare_Synchronized_Absent; Info : Internal_Entity_Info)
         return Internal_Entity_Synchronized_Absent is
      begin
         if Node = null then
            return No_Entity_Synchronized_Absent;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Synchronized_Absent) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Synchronized_Present
        (Node : Bare_Synchronized_Present; Info : Internal_Entity_Info)
         return Internal_Entity_Synchronized_Present is
      begin
         if Node = null then
            return No_Entity_Synchronized_Present;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Synchronized_Present) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Synth_Anonymous_Type_Decl
        (Node : Bare_Synth_Anonymous_Type_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Synth_Anonymous_Type_Decl is
      begin
         if Node = null then
            return No_Entity_Synth_Anonymous_Type_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Synth_Anonymous_Type_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Synthetic_Char_Enum_Lit
        (Node : Bare_Synthetic_Char_Enum_Lit; Info : Internal_Entity_Info)
         return Internal_Entity_Synthetic_Char_Enum_Lit is
      begin
         if Node = null then
            return No_Entity_Synthetic_Char_Enum_Lit;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Synthetic_Char_Enum_Lit) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Synthetic_Defining_Name
        (Node : Bare_Synthetic_Defining_Name; Info : Internal_Entity_Info)
         return Internal_Entity_Synthetic_Defining_Name is
      begin
         if Node = null then
            return No_Entity_Synthetic_Defining_Name;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Synthetic_Defining_Name) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Synthetic_Formal_Param_Decl
        (Node : Bare_Synthetic_Formal_Param_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Synthetic_Formal_Param_Decl is
      begin
         if Node = null then
            return No_Entity_Synthetic_Formal_Param_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Synthetic_Formal_Param_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Synthetic_Identifier
        (Node : Bare_Synthetic_Identifier; Info : Internal_Entity_Info)
         return Internal_Entity_Synthetic_Identifier is
      begin
         if Node = null then
            return No_Entity_Synthetic_Identifier;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Synthetic_Identifier) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Synthetic_Renaming_Clause
        (Node : Bare_Synthetic_Renaming_Clause; Info : Internal_Entity_Info)
         return Internal_Entity_Synthetic_Renaming_Clause is
      begin
         if Node = null then
            return No_Entity_Synthetic_Renaming_Clause;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Synthetic_Renaming_Clause) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Synthetic_Subp_Decl
        (Node : Bare_Synthetic_Subp_Decl; Info : Internal_Entity_Info)
         return Internal_Entity_Synthetic_Subp_Decl is
      begin
         if Node = null then
            return No_Entity_Synthetic_Subp_Decl;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Synthetic_Subp_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Synthetic_Type_Expr
        (Node : Bare_Synthetic_Type_Expr; Info : Internal_Entity_Info)
         return Internal_Entity_Synthetic_Type_Expr is
      begin
         if Node = null then
            return No_Entity_Synthetic_Type_Expr;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Synthetic_Type_Expr) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Tagged_Node
        (Node : Bare_Tagged_Node; Info : Internal_Entity_Info)
         return Internal_Entity_Tagged_Node is
      begin
         if Node = null then
            return No_Entity_Tagged_Node;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Tagged_Node) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Tagged_Absent
        (Node : Bare_Tagged_Absent; Info : Internal_Entity_Info)
         return Internal_Entity_Tagged_Absent is
      begin
         if Node = null then
            return No_Entity_Tagged_Absent;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Tagged_Absent) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Tagged_Present
        (Node : Bare_Tagged_Present; Info : Internal_Entity_Info)
         return Internal_Entity_Tagged_Present is
      begin
         if Node = null then
            return No_Entity_Tagged_Present;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Tagged_Present) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Target_Name
        (Node : Bare_Target_Name; Info : Internal_Entity_Info)
         return Internal_Entity_Target_Name is
      begin
         if Node = null then
            return No_Entity_Target_Name;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Target_Name) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Task_Body
        (Node : Bare_Task_Body; Info : Internal_Entity_Info)
         return Internal_Entity_Task_Body is
      begin
         if Node = null then
            return No_Entity_Task_Body;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Task_Body) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Task_Body_Stub
        (Node : Bare_Task_Body_Stub; Info : Internal_Entity_Info)
         return Internal_Entity_Task_Body_Stub is
      begin
         if Node = null then
            return No_Entity_Task_Body_Stub;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Task_Body_Stub) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Task_Def
        (Node : Bare_Task_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Task_Def is
      begin
         if Node = null then
            return No_Entity_Task_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Task_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Terminate_Alternative
        (Node : Bare_Terminate_Alternative; Info : Internal_Entity_Info)
         return Internal_Entity_Terminate_Alternative is
      begin
         if Node = null then
            return No_Entity_Terminate_Alternative;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Terminate_Alternative) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Type_Access_Def
        (Node : Bare_Type_Access_Def; Info : Internal_Entity_Info)
         return Internal_Entity_Type_Access_Def is
      begin
         if Node = null then
            return No_Entity_Type_Access_Def;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Type_Access_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Un_Op
        (Node : Bare_Un_Op; Info : Internal_Entity_Info)
         return Internal_Entity_Un_Op is
      begin
         if Node = null then
            return No_Entity_Un_Op;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Un_Op) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Unconstrained_Array_Index
        (Node : Bare_Unconstrained_Array_Index; Info : Internal_Entity_Info)
         return Internal_Entity_Unconstrained_Array_Index is
      begin
         if Node = null then
            return No_Entity_Unconstrained_Array_Index;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Unconstrained_Array_Index) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Unconstrained_Array_Index_List
        (Node : Bare_Unconstrained_Array_Index_List; Info : Internal_Entity_Info)
         return Internal_Entity_Unconstrained_Array_Index_List is
      begin
         if Node = null then
            return No_Entity_Unconstrained_Array_Index_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Unconstrained_Array_Index_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Unconstrained_Array_Indices
        (Node : Bare_Unconstrained_Array_Indices; Info : Internal_Entity_Info)
         return Internal_Entity_Unconstrained_Array_Indices is
      begin
         if Node = null then
            return No_Entity_Unconstrained_Array_Indices;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Unconstrained_Array_Indices) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Unknown_Discriminant_Part
        (Node : Bare_Unknown_Discriminant_Part; Info : Internal_Entity_Info)
         return Internal_Entity_Unknown_Discriminant_Part is
      begin
         if Node = null then
            return No_Entity_Unknown_Discriminant_Part;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Unknown_Discriminant_Part) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Until_Node
        (Node : Bare_Until_Node; Info : Internal_Entity_Info)
         return Internal_Entity_Until_Node is
      begin
         if Node = null then
            return No_Entity_Until_Node;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Until_Node) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Until_Absent
        (Node : Bare_Until_Absent; Info : Internal_Entity_Info)
         return Internal_Entity_Until_Absent is
      begin
         if Node = null then
            return No_Entity_Until_Absent;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Until_Absent) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Until_Present
        (Node : Bare_Until_Present; Info : Internal_Entity_Info)
         return Internal_Entity_Until_Present is
      begin
         if Node = null then
            return No_Entity_Until_Present;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Until_Present) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Update_Attribute_Ref
        (Node : Bare_Update_Attribute_Ref; Info : Internal_Entity_Info)
         return Internal_Entity_Update_Attribute_Ref is
      begin
         if Node = null then
            return No_Entity_Update_Attribute_Ref;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Update_Attribute_Ref) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Use_Clause
        (Node : Bare_Use_Clause; Info : Internal_Entity_Info)
         return Internal_Entity_Use_Clause is
      begin
         if Node = null then
            return No_Entity_Use_Clause;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Use_Clause) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Use_Package_Clause
        (Node : Bare_Use_Package_Clause; Info : Internal_Entity_Info)
         return Internal_Entity_Use_Package_Clause is
      begin
         if Node = null then
            return No_Entity_Use_Package_Clause;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Use_Package_Clause) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Use_Type_Clause
        (Node : Bare_Use_Type_Clause; Info : Internal_Entity_Info)
         return Internal_Entity_Use_Type_Clause is
      begin
         if Node = null then
            return No_Entity_Use_Type_Clause;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Use_Type_Clause) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Value_Sequence
        (Node : Bare_Value_Sequence; Info : Internal_Entity_Info)
         return Internal_Entity_Value_Sequence is
      begin
         if Node = null then
            return No_Entity_Value_Sequence;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Value_Sequence) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Variant
        (Node : Bare_Variant; Info : Internal_Entity_Info)
         return Internal_Entity_Variant is
      begin
         if Node = null then
            return No_Entity_Variant;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Variant) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Variant_List
        (Node : Bare_Variant_List; Info : Internal_Entity_Info)
         return Internal_Entity_Variant_List is
      begin
         if Node = null then
            return No_Entity_Variant_List;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Variant_List) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_Variant_Part
        (Node : Bare_Variant_Part; Info : Internal_Entity_Info)
         return Internal_Entity_Variant_Part is
      begin
         if Node = null then
            return No_Entity_Variant_Part;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_Variant_Part) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_While_Loop_Spec
        (Node : Bare_While_Loop_Spec; Info : Internal_Entity_Info)
         return Internal_Entity_While_Loop_Spec is
      begin
         if Node = null then
            return No_Entity_While_Loop_Spec;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_While_Loop_Spec) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_While_Loop_Stmt
        (Node : Bare_While_Loop_Stmt; Info : Internal_Entity_Info)
         return Internal_Entity_While_Loop_Stmt is
      begin
         if Node = null then
            return No_Entity_While_Loop_Stmt;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_While_Loop_Stmt) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_With_Clause
        (Node : Bare_With_Clause; Info : Internal_Entity_Info)
         return Internal_Entity_With_Clause is
      begin
         if Node = null then
            return No_Entity_With_Clause;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_With_Clause) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_With_Private
        (Node : Bare_With_Private; Info : Internal_Entity_Info)
         return Internal_Entity_With_Private is
      begin
         if Node = null then
            return No_Entity_With_Private;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_With_Private) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_With_Private_Absent
        (Node : Bare_With_Private_Absent; Info : Internal_Entity_Info)
         return Internal_Entity_With_Private_Absent is
      begin
         if Node = null then
            return No_Entity_With_Private_Absent;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_With_Private_Absent) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      function Create_Internal_Entity_With_Private_Present
        (Node : Bare_With_Private_Present; Info : Internal_Entity_Info)
         return Internal_Entity_With_Private_Present is
      begin
         if Node = null then
            return No_Entity_With_Private_Present;
         end if;
         return (Node => Node, Info => Info);
      end;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Entity_With_Private_Present) return String is
         pragma Warnings (On, "referenced");
      begin
            return Image (Entity'(Node => R.Node, Info => R.Info));
      end Trace_Image;
      -------------
      -- Inc_Ref --
      -------------
      procedure Inc_Ref (R : Internal_Env_Assoc) is
      begin
               Inc_Ref (R.Dest_Env);
      end Inc_Ref;
      -------------
      -- Dec_Ref --
      -------------
      procedure Dec_Ref (R : in out Internal_Env_Assoc) is
      begin
               Dec_Ref (R.Dest_Env);
      end Dec_Ref;
      ----------------
      -- Equivalent --
      ----------------
      function Equivalent (L, R : Internal_Env_Assoc) return Boolean is
      begin
         return L.Key = R.Key and then L.Value = R.Value and then Equivalent (L.Dest_Env, R.Dest_Env) and then L.Metadata = R.Metadata;
      end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Env_Assoc) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Key => " & Trace_Image (R.Key)
                        & ", "
                     & "Value => " & Trace_Image (R.Value)
                        & ", "
                     & "Dest_Env => " & Trace_Image (R.Dest_Env)
                        & ", "
                     & "Metadata => " & Trace_Image (R.Metadata)
               & ")");
      end Trace_Image;
      -------------
      -- Inc_Ref --
      -------------
      procedure Inc_Ref (R : Internal_Eval_Discrete_Range) is
      begin
               Inc_Ref (R.Low_Bound);
               Inc_Ref (R.High_Bound);
      end Inc_Ref;
      -------------
      -- Dec_Ref --
      -------------
      procedure Dec_Ref (R : in out Internal_Eval_Discrete_Range) is
      begin
               Dec_Ref (R.Low_Bound);
               Dec_Ref (R.High_Bound);
      end Dec_Ref;
      ----------------
      -- Equivalent --
      ----------------
      function Equivalent (L, R : Internal_Eval_Discrete_Range) return Boolean is
      begin
         return Equivalent (L.Low_Bound, R.Low_Bound) and then Equivalent (L.High_Bound, R.High_Bound);
      end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Eval_Discrete_Range) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Low_Bound => " & Trace_Image (R.Low_Bound)
                        & ", "
                     & "High_Bound => " & Trace_Image (R.High_Bound)
               & ")");
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Expected_Type_For_Expr) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Expected_Type => " & Trace_Image (R.Expected_Type)
                        & ", "
                     & "Expr => " & Trace_Image (R.Expr)
               & ")");
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Inner_Env_Assoc) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Key => " & Trace_Image (R.Key)
                        & ", "
                     & "Value => " & Trace_Image (R.Value)
                        & ", "
                     & "Metadata => " & Trace_Image (R.Metadata)
               & ")");
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Logic_Val_Result) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Success => " & Trace_Image (R.Success)
                        & ", "
                     & "Value => " & Trace_Image (R.Value)
               & ")");
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Multidim_Aggregate_Info) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Agg => " & Trace_Image (R.Agg)
                        & ", "
                     & "Typ => " & Trace_Image (R.Typ)
                        & ", "
                     & "Rank => " & Trace_Image (R.Rank)
               & ")");
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Param_Actual) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Param => " & Trace_Image (R.Param)
                        & ", "
                     & "Actual => " & Trace_Image (R.Actual)
               & ")");
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Single_Actual) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Name => " & Trace_Image (R.Name)
                        & ", "
                     & "Assoc => " & Trace_Image (R.Assoc)
               & ")");
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Param_Match) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Has_Matched => " & Trace_Image (R.Has_Matched)
                        & ", "
                     & "Actual => " & Trace_Image (R.Actual)
                        & ", "
                     & "Formal => " & Trace_Image (R.Formal)
               & ")");
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Ref_Result) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Ref => " & Trace_Image (R.Ref)
                        & ", "
                     & "Kind => " & Trace_Image (R.Kind)
               & ")");
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Refd_Decl) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Decl => " & Trace_Image (R.Decl)
                        & ", "
                     & "Kind => " & Trace_Image (R.Kind)
               & ")");
      end Trace_Image;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Refd_Def) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Def_Name => " & Trace_Image (R.Def_Name)
                        & ", "
                     & "Kind => " & Trace_Image (R.Kind)
               & ")");
      end Trace_Image;
      -------------
      -- Inc_Ref --
      -------------
      procedure Inc_Ref (R : Internal_Shape) is
      begin
               Inc_Ref (R.Components);
               Inc_Ref (R.Discriminants_Values);
      end Inc_Ref;
      -------------
      -- Dec_Ref --
      -------------
      procedure Dec_Ref (R : in out Internal_Shape) is
      begin
               Dec_Ref (R.Components);
               Dec_Ref (R.Discriminants_Values);
      end Dec_Ref;
      ----------------
      -- Equivalent --
      ----------------
      function Equivalent (L, R : Internal_Shape) return Boolean is
      begin
         return Equivalent (L.Components, R.Components) and then Equivalent (L.Discriminants_Values, R.Discriminants_Values);
      end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Shape) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "Components => " & Trace_Image (R.Components)
                        & ", "
                     & "Discriminants_Values => " & Trace_Image (R.Discriminants_Values)
               & ")");
      end Trace_Image;
      -------------
      -- Inc_Ref --
      -------------
      procedure Inc_Ref (R : Internal_Substitution) is
      begin
               Inc_Ref (R.To_Value);
      end Inc_Ref;
      -------------
      -- Dec_Ref --
      -------------
      procedure Dec_Ref (R : in out Internal_Substitution) is
      begin
               Dec_Ref (R.To_Value);
      end Dec_Ref;
      ----------------
      -- Equivalent --
      ----------------
      function Equivalent (L, R : Internal_Substitution) return Boolean is
      begin
         return L.From_Decl = R.From_Decl and then Equivalent (L.To_Value, R.To_Value) and then L.Value_Type = R.Value_Type;
      end Equivalent;
      -----------------
      -- Trace_Image --
      -----------------
      pragma Warnings (Off, "referenced");
      function Trace_Image (R : Internal_Substitution) return String is
         pragma Warnings (On, "referenced");
      begin
            return
              ("("
                     & "From_Decl => " & Trace_Image (R.From_Decl)
                        & ", "
                     & "To_Value => " & Trace_Image (R.To_Value)
                        & ", "
                     & "Value_Type => " & Trace_Image (R.Value_Type)
               & ")");
      end Trace_Image;
   type Ada_Node_P_Is_Not_Null_0_Predicate is
   new Solver_Ifc.Predicate_Type
   with record
         null;
   end record;
      overriding function Call
        (Self : Ada_Node_P_Is_Not_Null_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   ;
   overriding function Image (Self : Ada_Node_P_Is_Not_Null_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Ada_Node_P_Is_Not_Null_0_Predicate);
   function Create_Ada_Node_P_Is_Not_Null_0_Predicate
      return Ada_Node_P_Is_Not_Null_0_Predicate is
   begin
      return Ada_Node_P_Is_Not_Null_0_Predicate'(Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Ada_Node_P_Is_Not_Null_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   is
         pragma Unreferenced (Self);
      Node : constant Bare_Ada_Node := Entity.Node;
   begin
      return Ada_Node_P_Is_Not_Null (Node, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Ada_Node_P_Is_Not_Null_0_Predicate) return String is
   begin
      return "AdaNode.is_not_null";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Ada_Node_P_Is_Not_Null_0_Predicate) is
      begin
         null;
      end Destroy;
   type Basic_Decl_P_Subp_Decl_Match_Signature_0_Predicate is
   new Solver_Ifc.Predicate_Type
   with record
         Field_0 : Internal_Entity_Basic_Decl;
   end record;
      overriding function Call
        (Self : Basic_Decl_P_Subp_Decl_Match_Signature_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   ;
   overriding function Image (Self : Basic_Decl_P_Subp_Decl_Match_Signature_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Basic_Decl_P_Subp_Decl_Match_Signature_0_Predicate);
   function Create_Basic_Decl_P_Subp_Decl_Match_Signature_0_Predicate
   (
         Field_0 : Internal_Entity_Basic_Decl
   )
      return Basic_Decl_P_Subp_Decl_Match_Signature_0_Predicate is
   begin
      return Basic_Decl_P_Subp_Decl_Match_Signature_0_Predicate'(Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1, Field_0 => Field_0);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Basic_Decl_P_Subp_Decl_Match_Signature_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   is
      Node : constant Bare_Basic_Decl := Entity.Node;
   begin
      return Basic_Decl_P_Subp_Decl_Match_Signature (Node, Self.Field_0, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Basic_Decl_P_Subp_Decl_Match_Signature_0_Predicate) return String is
   begin
      return "BasicDecl.subp_decl_match_signature";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Basic_Decl_P_Subp_Decl_Match_Signature_0_Predicate) is
      begin
         null;
      end Destroy;
   type Basic_Decl_P_Is_Subprogram_0_Predicate is
   new Solver_Ifc.Predicate_Type
   with record
         null;
   end record;
      overriding function Call
        (Self : Basic_Decl_P_Is_Subprogram_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   ;
   overriding function Image (Self : Basic_Decl_P_Is_Subprogram_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Basic_Decl_P_Is_Subprogram_0_Predicate);
   function Create_Basic_Decl_P_Is_Subprogram_0_Predicate
      return Basic_Decl_P_Is_Subprogram_0_Predicate is
   begin
      return Basic_Decl_P_Is_Subprogram_0_Predicate'(Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Basic_Decl_P_Is_Subprogram_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   is
         pragma Unreferenced (Self);
      Node : constant Bare_Basic_Decl := Entity.Node;
   begin
      return Basic_Decl_P_Is_Subprogram (Node);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Basic_Decl_P_Is_Subprogram_0_Predicate) return String is
   begin
      return "BasicDecl.is_subprogram";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Basic_Decl_P_Is_Subprogram_0_Predicate) is
      begin
         null;
      end Destroy;
   type Basic_Decl_P_Is_Stream_Subprogram_For_Type_0_Predicate is
   new Solver_Ifc.Predicate_Type
   with record
         Field_0 : Internal_Entity_Base_Type_Decl;
         Field_1 : Boolean;
   end record;
      overriding function Call
        (Self : Basic_Decl_P_Is_Stream_Subprogram_For_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   ;
   overriding function Image (Self : Basic_Decl_P_Is_Stream_Subprogram_For_Type_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Basic_Decl_P_Is_Stream_Subprogram_For_Type_0_Predicate);
   function Create_Basic_Decl_P_Is_Stream_Subprogram_For_Type_0_Predicate
   (
         Field_0 : Internal_Entity_Base_Type_Decl;
         Field_1 : Boolean
   )
      return Basic_Decl_P_Is_Stream_Subprogram_For_Type_0_Predicate is
   begin
      return Basic_Decl_P_Is_Stream_Subprogram_For_Type_0_Predicate'(Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1, Field_0 => Field_0, Field_1 => Field_1);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Basic_Decl_P_Is_Stream_Subprogram_For_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   is
      Node : constant Bare_Basic_Decl := Entity.Node;
   begin
      return Basic_Decl_P_Is_Stream_Subprogram_For_Type (Node, Self.Field_0, Self.Field_1, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Basic_Decl_P_Is_Stream_Subprogram_For_Type_0_Predicate) return String is
   begin
      return "BasicDecl.is_stream_subprogram_for_type";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Basic_Decl_P_Is_Stream_Subprogram_For_Type_0_Predicate) is
      begin
         null;
      end Destroy;
   type Basic_Decl_P_Is_Put_Image_Subprogram_For_Type_0_Predicate is
   new Solver_Ifc.Predicate_Type
   with record
         Field_0 : Internal_Entity_Base_Type_Decl;
   end record;
      overriding function Call
        (Self : Basic_Decl_P_Is_Put_Image_Subprogram_For_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   ;
   overriding function Image (Self : Basic_Decl_P_Is_Put_Image_Subprogram_For_Type_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Basic_Decl_P_Is_Put_Image_Subprogram_For_Type_0_Predicate);
   function Create_Basic_Decl_P_Is_Put_Image_Subprogram_For_Type_0_Predicate
   (
         Field_0 : Internal_Entity_Base_Type_Decl
   )
      return Basic_Decl_P_Is_Put_Image_Subprogram_For_Type_0_Predicate is
   begin
      return Basic_Decl_P_Is_Put_Image_Subprogram_For_Type_0_Predicate'(Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1, Field_0 => Field_0);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Basic_Decl_P_Is_Put_Image_Subprogram_For_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   is
      Node : constant Bare_Basic_Decl := Entity.Node;
   begin
      return Basic_Decl_P_Is_Put_Image_Subprogram_For_Type (Node, Self.Field_0, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Basic_Decl_P_Is_Put_Image_Subprogram_For_Type_0_Predicate) return String is
   begin
      return "BasicDecl.is_put_image_subprogram_for_type";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Basic_Decl_P_Is_Put_Image_Subprogram_For_Type_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Is_Array_Or_Rec_0_Predicate is
   new Solver_Ifc.Predicate_Type
   with record
         Field_0 : Bare_Ada_Node;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Array_Or_Rec_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Is_Array_Or_Rec_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Array_Or_Rec_0_Predicate);
   function Create_Base_Type_Decl_P_Is_Array_Or_Rec_0_Predicate
   (
         Field_0 : Bare_Ada_Node
   )
      return Base_Type_Decl_P_Is_Array_Or_Rec_0_Predicate is
   begin
      return Base_Type_Decl_P_Is_Array_Or_Rec_0_Predicate'(Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1, Field_0 => Field_0);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Array_Or_Rec_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   is
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Base_Type_Decl_P_Is_Array_Or_Rec (Node, Self.Field_0, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Is_Array_Or_Rec_0_Predicate) return String is
   begin
      return "BaseTypeDecl.is_array_or_rec";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Array_Or_Rec_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Is_Task_Type_0_Predicate is
   new Solver_Ifc.Predicate_Type
   with record
         null;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Task_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Is_Task_Type_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Task_Type_0_Predicate);
   function Create_Base_Type_Decl_P_Is_Task_Type_0_Predicate
      return Base_Type_Decl_P_Is_Task_Type_0_Predicate is
   begin
      return Base_Type_Decl_P_Is_Task_Type_0_Predicate'(Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Task_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   is
         pragma Unreferenced (Self);
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Dispatcher_Base_Type_Decl_P_Is_Task_Type (Node);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Is_Task_Type_0_Predicate) return String is
   begin
      return "BaseTypeDecl.is_task_type";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Task_Type_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_One_Non_Universal_0_Predicate is
   new Solver_Ifc.N_Predicate_Type
   with record
         null;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_One_Non_Universal_0_Predicate;
            Entities : Entity_Vars.Value_Array
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_One_Non_Universal_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_One_Non_Universal_0_Predicate);
   function Create_Base_Type_Decl_P_One_Non_Universal_0_Predicate
      return Base_Type_Decl_P_One_Non_Universal_0_Predicate is
   begin
      return Base_Type_Decl_P_One_Non_Universal_0_Predicate'(N => 2, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_One_Non_Universal_0_Predicate;
            Entities : Entity_Vars.Value_Array
        ) return Boolean
   is
         pragma Unreferenced (Self);
         Entity : Internal_Entity := Entities (1);
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Base_Type_Decl_P_One_Non_Universal (Node, (Node => Entities (2).Node, Info => Entities (2).Info), E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_One_Non_Universal_0_Predicate) return String is
   begin
      return "BaseTypeDecl.one_non_universal";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_One_Non_Universal_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Is_Non_Null_Char_Type_0_Predicate is
   new Solver_Ifc.Predicate_Type
   with record
         Field_0 : Bare_Ada_Node;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Non_Null_Char_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Is_Non_Null_Char_Type_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Non_Null_Char_Type_0_Predicate);
   function Create_Base_Type_Decl_P_Is_Non_Null_Char_Type_0_Predicate
   (
         Field_0 : Bare_Ada_Node
   )
      return Base_Type_Decl_P_Is_Non_Null_Char_Type_0_Predicate is
   begin
      return Base_Type_Decl_P_Is_Non_Null_Char_Type_0_Predicate'(Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1, Field_0 => Field_0);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Non_Null_Char_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   is
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Base_Type_Decl_P_Is_Non_Null_Char_Type (Node, Self.Field_0, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Is_Non_Null_Char_Type_0_Predicate) return String is
   begin
      return "BaseTypeDecl.is_non_null_char_type";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Non_Null_Char_Type_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Is_Universal_Type_0_Predicate is
   new Solver_Ifc.Predicate_Type
   with record
         null;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Universal_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Is_Universal_Type_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Universal_Type_0_Predicate);
   function Create_Base_Type_Decl_P_Is_Universal_Type_0_Predicate
      return Base_Type_Decl_P_Is_Universal_Type_0_Predicate is
   begin
      return Base_Type_Decl_P_Is_Universal_Type_0_Predicate'(Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Universal_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   is
         pragma Unreferenced (Self);
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Base_Type_Decl_P_Is_Universal_Type (Node, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Is_Universal_Type_0_Predicate) return String is
   begin
      return "BaseTypeDecl.is_universal_type";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Universal_Type_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Is_Not_Universal_Type_0_Predicate is
   new Solver_Ifc.Predicate_Type
   with record
         null;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Not_Universal_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Is_Not_Universal_Type_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Not_Universal_Type_0_Predicate);
   function Create_Base_Type_Decl_P_Is_Not_Universal_Type_0_Predicate
      return Base_Type_Decl_P_Is_Not_Universal_Type_0_Predicate is
   begin
      return Base_Type_Decl_P_Is_Not_Universal_Type_0_Predicate'(Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Not_Universal_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   is
         pragma Unreferenced (Self);
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Base_Type_Decl_P_Is_Not_Universal_Type (Node, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Is_Not_Universal_Type_0_Predicate) return String is
   begin
      return "BaseTypeDecl.is_not_universal_type";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Not_Universal_Type_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Is_Access_Type_Predicate_0_Predicate is
   new Solver_Ifc.Predicate_Type
   with record
         Field_0 : Bare_Ada_Node;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Access_Type_Predicate_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Is_Access_Type_Predicate_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Access_Type_Predicate_0_Predicate);
   function Create_Base_Type_Decl_P_Is_Access_Type_Predicate_0_Predicate
   (
         Field_0 : Bare_Ada_Node
   )
      return Base_Type_Decl_P_Is_Access_Type_Predicate_0_Predicate is
   begin
      return Base_Type_Decl_P_Is_Access_Type_Predicate_0_Predicate'(Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1, Field_0 => Field_0);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Access_Type_Predicate_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   is
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Base_Type_Decl_P_Is_Access_Type_Predicate (Node, Self.Field_0, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Is_Access_Type_Predicate_0_Predicate) return String is
   begin
      return "BaseTypeDecl.is_access_type_predicate";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Access_Type_Predicate_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Is_Iterator_Type_0_Predicate is
   new Solver_Ifc.Predicate_Type
   with record
         Field_0 : Bare_Ada_Node;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Iterator_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Is_Iterator_Type_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Iterator_Type_0_Predicate);
   function Create_Base_Type_Decl_P_Is_Iterator_Type_0_Predicate
   (
         Field_0 : Bare_Ada_Node
   )
      return Base_Type_Decl_P_Is_Iterator_Type_0_Predicate is
   begin
      return Base_Type_Decl_P_Is_Iterator_Type_0_Predicate'(Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1, Field_0 => Field_0);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Iterator_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   is
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Base_Type_Decl_P_Is_Iterator_Type (Node, Self.Field_0, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Is_Iterator_Type_0_Predicate) return String is
   begin
      return "BaseTypeDecl.is_iterator_type";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Iterator_Type_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Is_Discrete_Type_0_Predicate is
   new Solver_Ifc.Predicate_Type
   with record
         Field_0 : Bare_Ada_Node;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Discrete_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Is_Discrete_Type_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Discrete_Type_0_Predicate);
   function Create_Base_Type_Decl_P_Is_Discrete_Type_0_Predicate
   (
         Field_0 : Bare_Ada_Node
   )
      return Base_Type_Decl_P_Is_Discrete_Type_0_Predicate is
   begin
      return Base_Type_Decl_P_Is_Discrete_Type_0_Predicate'(Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1, Field_0 => Field_0);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Discrete_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   is
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Dispatcher_Base_Type_Decl_P_Is_Discrete_Type (Node, Self.Field_0, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Is_Discrete_Type_0_Predicate) return String is
   begin
      return "BaseTypeDecl.is_discrete_type";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Discrete_Type_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Is_Not_Root_Int_Type_0_Predicate is
   new Solver_Ifc.Predicate_Type
   with record
         null;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Not_Root_Int_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Is_Not_Root_Int_Type_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Not_Root_Int_Type_0_Predicate);
   function Create_Base_Type_Decl_P_Is_Not_Root_Int_Type_0_Predicate
      return Base_Type_Decl_P_Is_Not_Root_Int_Type_0_Predicate is
   begin
      return Base_Type_Decl_P_Is_Not_Root_Int_Type_0_Predicate'(Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Not_Root_Int_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   is
         pragma Unreferenced (Self);
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Base_Type_Decl_P_Is_Not_Root_Int_Type (Node, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Is_Not_Root_Int_Type_0_Predicate) return String is
   begin
      return "BaseTypeDecl.is_not_root_int_type";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Not_Root_Int_Type_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Allows_String_Literal_0_Predicate is
   new Solver_Ifc.Predicate_Type
   with record
         Field_0 : Bare_Ada_Node;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Allows_String_Literal_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Allows_String_Literal_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Allows_String_Literal_0_Predicate);
   function Create_Base_Type_Decl_P_Allows_String_Literal_0_Predicate
   (
         Field_0 : Bare_Ada_Node
   )
      return Base_Type_Decl_P_Allows_String_Literal_0_Predicate is
   begin
      return Base_Type_Decl_P_Allows_String_Literal_0_Predicate'(Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1, Field_0 => Field_0);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Allows_String_Literal_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   is
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Base_Type_Decl_P_Allows_String_Literal (Node, Self.Field_0, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Allows_String_Literal_0_Predicate) return String is
   begin
      return "BaseTypeDecl.allows_string_literal";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Allows_String_Literal_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Is_Subp_Access_Of_0_Predicate is
   new Solver_Ifc.N_Predicate_Type
   with record
         Field_0 : Bare_Ada_Node;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Subp_Access_Of_0_Predicate;
            Entities : Entity_Vars.Value_Array
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Is_Subp_Access_Of_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Subp_Access_Of_0_Predicate);
   function Create_Base_Type_Decl_P_Is_Subp_Access_Of_0_Predicate
   (
         Field_0 : Bare_Ada_Node
   )
      return Base_Type_Decl_P_Is_Subp_Access_Of_0_Predicate is
   begin
      return Base_Type_Decl_P_Is_Subp_Access_Of_0_Predicate'(N => 2, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1, Field_0 => Field_0);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Subp_Access_Of_0_Predicate;
            Entities : Entity_Vars.Value_Array
        ) return Boolean
   is
         Entity : Internal_Entity := Entities (1);
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Base_Type_Decl_P_Is_Subp_Access_Of (Node, (Node => Entities (2).Node, Info => Entities (2).Info), Self.Field_0, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Is_Subp_Access_Of_0_Predicate) return String is
   begin
      return "BaseTypeDecl.is_subp_access_of";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Subp_Access_Of_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Is_Tagged_Type_With_Deref_0_Predicate is
   new Solver_Ifc.Predicate_Type
   with record
         Field_0 : Bare_Ada_Node;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Tagged_Type_With_Deref_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Is_Tagged_Type_With_Deref_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Tagged_Type_With_Deref_0_Predicate);
   function Create_Base_Type_Decl_P_Is_Tagged_Type_With_Deref_0_Predicate
   (
         Field_0 : Bare_Ada_Node
   )
      return Base_Type_Decl_P_Is_Tagged_Type_With_Deref_0_Predicate is
   begin
      return Base_Type_Decl_P_Is_Tagged_Type_With_Deref_0_Predicate'(Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1, Field_0 => Field_0);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Tagged_Type_With_Deref_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   is
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Base_Type_Decl_P_Is_Tagged_Type_With_Deref (Node, Self.Field_0, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Is_Tagged_Type_With_Deref_0_Predicate) return String is
   begin
      return "BaseTypeDecl.is_tagged_type_with_deref";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Tagged_Type_With_Deref_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Is_Array_Def_With_Deref_0_Predicate is
   new Solver_Ifc.Predicate_Type
   with record
         Field_0 : Bare_Ada_Node;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Array_Def_With_Deref_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Is_Array_Def_With_Deref_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Array_Def_With_Deref_0_Predicate);
   function Create_Base_Type_Decl_P_Is_Array_Def_With_Deref_0_Predicate
   (
         Field_0 : Bare_Ada_Node
   )
      return Base_Type_Decl_P_Is_Array_Def_With_Deref_0_Predicate is
   begin
      return Base_Type_Decl_P_Is_Array_Def_With_Deref_0_Predicate'(Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1, Field_0 => Field_0);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Array_Def_With_Deref_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   is
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Base_Type_Decl_P_Is_Array_Def_With_Deref (Node, Self.Field_0, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Is_Array_Def_With_Deref_0_Predicate) return String is
   begin
      return "BaseTypeDecl.is_array_def_with_deref";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Array_Def_With_Deref_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Is_Array_Def_With_Deref_Or_Null_0_Predicate is
   new Solver_Ifc.Predicate_Type
   with record
         Field_0 : Bare_Ada_Node;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Array_Def_With_Deref_Or_Null_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Is_Array_Def_With_Deref_Or_Null_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Array_Def_With_Deref_Or_Null_0_Predicate);
   function Create_Base_Type_Decl_P_Is_Array_Def_With_Deref_Or_Null_0_Predicate
   (
         Field_0 : Bare_Ada_Node
   )
      return Base_Type_Decl_P_Is_Array_Def_With_Deref_Or_Null_0_Predicate is
   begin
      return Base_Type_Decl_P_Is_Array_Def_With_Deref_Or_Null_0_Predicate'(Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1, Field_0 => Field_0);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Is_Array_Def_With_Deref_Or_Null_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   is
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Base_Type_Decl_P_Is_Array_Def_With_Deref_Or_Null (Node, Self.Field_0, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Is_Array_Def_With_Deref_Or_Null_0_Predicate) return String is
   begin
      return "BaseTypeDecl.is_array_def_with_deref_or_null";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Is_Array_Def_With_Deref_Or_Null_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Matching_Prefix_Type_0_Predicate is
   new Solver_Ifc.N_Predicate_Type
   with record
         Field_0 : Bare_Ada_Node;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Matching_Prefix_Type_0_Predicate;
            Entities : Entity_Vars.Value_Array
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Matching_Prefix_Type_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Matching_Prefix_Type_0_Predicate);
   function Create_Base_Type_Decl_P_Matching_Prefix_Type_0_Predicate
   (
         Field_0 : Bare_Ada_Node
   )
      return Base_Type_Decl_P_Matching_Prefix_Type_0_Predicate is
   begin
      return Base_Type_Decl_P_Matching_Prefix_Type_0_Predicate'(N => 2, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1, Field_0 => Field_0);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Matching_Prefix_Type_0_Predicate;
            Entities : Entity_Vars.Value_Array
        ) return Boolean
   is
         Entity : Internal_Entity := Entities (1);
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Base_Type_Decl_P_Matching_Prefix_Type (Node, (Node => Entities (2).Node, Info => Entities (2).Info), Self.Field_0, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Matching_Prefix_Type_0_Predicate) return String is
   begin
      return "BaseTypeDecl.matching_prefix_type";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Matching_Prefix_Type_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Matching_Formal_Prim_Type_0_Predicate is
   new Solver_Ifc.N_Predicate_Type
   with record
         Field_0 : Bare_Ada_Node;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Matching_Formal_Prim_Type_0_Predicate;
            Entities : Entity_Vars.Value_Array
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Matching_Formal_Prim_Type_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Matching_Formal_Prim_Type_0_Predicate);
   function Create_Base_Type_Decl_P_Matching_Formal_Prim_Type_0_Predicate
   (
         Field_0 : Bare_Ada_Node
   )
      return Base_Type_Decl_P_Matching_Formal_Prim_Type_0_Predicate is
   begin
      return Base_Type_Decl_P_Matching_Formal_Prim_Type_0_Predicate'(N => 2, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1, Field_0 => Field_0);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Matching_Formal_Prim_Type_0_Predicate;
            Entities : Entity_Vars.Value_Array
        ) return Boolean
   is
         Entity : Internal_Entity := Entities (1);
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Base_Type_Decl_P_Matching_Formal_Prim_Type (Node, (Node => Entities (2).Node, Info => Entities (2).Info), Self.Field_0, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Matching_Formal_Prim_Type_0_Predicate) return String is
   begin
      return "BaseTypeDecl.matching_formal_prim_type";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Matching_Formal_Prim_Type_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Matching_Formal_Type_0_Predicate is
   new Solver_Ifc.N_Predicate_Type
   with record
         Field_0 : Bare_Ada_Node;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Matching_Formal_Type_0_Predicate;
            Entities : Entity_Vars.Value_Array
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Matching_Formal_Type_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Matching_Formal_Type_0_Predicate);
   function Create_Base_Type_Decl_P_Matching_Formal_Type_0_Predicate
   (
         Field_0 : Bare_Ada_Node
   )
      return Base_Type_Decl_P_Matching_Formal_Type_0_Predicate is
   begin
      return Base_Type_Decl_P_Matching_Formal_Type_0_Predicate'(N => 2, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1, Field_0 => Field_0);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Matching_Formal_Type_0_Predicate;
            Entities : Entity_Vars.Value_Array
        ) return Boolean
   is
         Entity : Internal_Entity := Entities (1);
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Base_Type_Decl_P_Matching_Formal_Type (Node, (Node => Entities (2).Node, Info => Entities (2).Info), Self.Field_0, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Matching_Formal_Type_0_Predicate) return String is
   begin
      return "BaseTypeDecl.matching_formal_type";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Matching_Formal_Type_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Matching_Assign_Type_0_Predicate is
   new Solver_Ifc.N_Predicate_Type
   with record
         Field_0 : Bare_Ada_Node;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Matching_Assign_Type_0_Predicate;
            Entities : Entity_Vars.Value_Array
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Matching_Assign_Type_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Matching_Assign_Type_0_Predicate);
   function Create_Base_Type_Decl_P_Matching_Assign_Type_0_Predicate
   (
         Field_0 : Bare_Ada_Node
   )
      return Base_Type_Decl_P_Matching_Assign_Type_0_Predicate is
   begin
      return Base_Type_Decl_P_Matching_Assign_Type_0_Predicate'(N => 2, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1, Field_0 => Field_0);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Matching_Assign_Type_0_Predicate;
            Entities : Entity_Vars.Value_Array
        ) return Boolean
   is
         Entity : Internal_Entity := Entities (1);
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Base_Type_Decl_P_Matching_Assign_Type (Node, (Node => Entities (2).Node, Info => Entities (2).Info), Self.Field_0, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Matching_Assign_Type_0_Predicate) return String is
   begin
      return "BaseTypeDecl.matching_assign_type";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Matching_Assign_Type_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Matching_Type_0_Predicate is
   new Solver_Ifc.N_Predicate_Type
   with record
         Field_0 : Bare_Ada_Node;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Matching_Type_0_Predicate;
            Entities : Entity_Vars.Value_Array
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Matching_Type_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Matching_Type_0_Predicate);
   function Create_Base_Type_Decl_P_Matching_Type_0_Predicate
   (
         Field_0 : Bare_Ada_Node
   )
      return Base_Type_Decl_P_Matching_Type_0_Predicate is
   begin
      return Base_Type_Decl_P_Matching_Type_0_Predicate'(N => 2, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1, Field_0 => Field_0);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Matching_Type_0_Predicate;
            Entities : Entity_Vars.Value_Array
        ) return Boolean
   is
         Entity : Internal_Entity := Entities (1);
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Base_Type_Decl_P_Matching_Type (Node, (Node => Entities (2).Node, Info => Entities (2).Info), Self.Field_0, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Matching_Type_0_Predicate) return String is
   begin
      return "BaseTypeDecl.matching_type";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Matching_Type_0_Predicate) is
      begin
         null;
      end Destroy;
   type Base_Type_Decl_P_Matching_Allocator_Type_0_Predicate is
   new Solver_Ifc.Predicate_Type
   with record
         Field_0 : Internal_Entity_Base_Type_Decl;
         Field_1 : Bare_Ada_Node;
   end record;
      overriding function Call
        (Self : Base_Type_Decl_P_Matching_Allocator_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   ;
   overriding function Image (Self : Base_Type_Decl_P_Matching_Allocator_Type_0_Predicate) return String;
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Matching_Allocator_Type_0_Predicate);
   function Create_Base_Type_Decl_P_Matching_Allocator_Type_0_Predicate
   (
         Field_0 : Internal_Entity_Base_Type_Decl;
         Field_1 : Bare_Ada_Node
   )
      return Base_Type_Decl_P_Matching_Allocator_Type_0_Predicate is
   begin
      return Base_Type_Decl_P_Matching_Allocator_Type_0_Predicate'(Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Ref_Count => 1, Field_0 => Field_0, Field_1 => Field_1);
   end;
   ----------
   -- Call --
   ----------
      overriding function Call
        (Self : Base_Type_Decl_P_Matching_Allocator_Type_0_Predicate;
            Entity : Internal_Entity
        ) return Boolean
   is
      Node : constant Bare_Base_Type_Decl := Entity.Node;
   begin
      return Base_Type_Decl_P_Matching_Allocator_Type (Node, Self.Field_0, Self.Field_1, E_Info => Entity.Info);
   end Call;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Base_Type_Decl_P_Matching_Allocator_Type_0_Predicate) return String is
   begin
      return "BaseTypeDecl.matching_allocator_type";
   end Image;
      -------------
      -- Destroy --
      -------------
      overriding procedure Destroy (Self : in out Base_Type_Decl_P_Matching_Allocator_Type_0_Predicate) is
      begin
         null;
      end Destroy;
   type Logic_Functor_604 is new Solver_Ifc.Converter_Type with
      record
            Origin : Bare_Ada_Node;
      end record;
   overriding function Convert (Self : Logic_Functor_604; From : Internal_Entity) return Internal_Entity with Inline;
   overriding function Image (Self : Logic_Functor_604) return String;
   overriding function Convert (Self : Logic_Functor_604; From : Internal_Entity) return Internal_Entity is
      Ret : Internal_Entity_Base_Type_Decl;
   begin
            if From.Node /= null
               and then From.Node.Kind not in Ada_Base_Type_Decl
            then
               raise Property_Error
                 with "mismatching node type for conv_prop";
            end if;
      Ret := Base_Type_Decl_P_Accessed_Type_No_Call (Node => From.Node, Origin => Self.Origin, E_Info => From.Info);
      return (Node => Ret.Node, Info => Ret.Info);
   end Convert;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Logic_Functor_604) return String is
   begin
      return ("BaseTypeDecl.accessed_type_no_call");
   end Image;
   type Logic_Functor_554 is new Solver_Ifc.Converter_Type with
      null record;
   overriding function Convert (Self : Logic_Functor_554; From : Internal_Entity) return Internal_Entity with Inline;
   overriding function Image (Self : Logic_Functor_554) return String;
   overriding function Convert (Self : Logic_Functor_554; From : Internal_Entity) return Internal_Entity is
         pragma Unreferenced (Self);
      Ret : Internal_Entity_Base_Type_Decl;
   begin
            if From.Node /= null
               and then From.Node.Kind not in Ada_Base_Type_Decl
            then
               raise Property_Error
                 with "mismatching node type for conv_prop";
            end if;
      Ret := Base_Type_Decl_P_Anonymous_Access_Type_Or_Null (Node => From.Node, E_Info => From.Info);
      return (Node => Ret.Node, Info => Ret.Info);
   end Convert;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Logic_Functor_554) return String is
   begin
      return ("BaseTypeDecl.anonymous_access_type_or_null");
   end Image;
   type Logic_Functor_581 is new Solver_Ifc.Combiner_Type with
      record
            Origin : Bare_Ada_Node;
      end record;
   overriding function Combine (Self : Logic_Functor_581; Vals : Entity_Vars.Value_Array) return Internal_Entity with Inline;
   overriding function Image (Self : Logic_Functor_581) return String;
   overriding function Combine (Self : Logic_Functor_581; Vals : Entity_Vars.Value_Array) return Internal_Entity is
         From : constant Internal_Entity := Vals (1);
      Ret : Internal_Entity_Base_Type_Decl;
   begin
            if From.Node /= null
               and then From.Node.Kind not in Ada_Base_Type_Decl
            then
               raise Property_Error
                 with "mismatching node type for comb_prop";
            end if;
            if Vals (2).Node /= null
               and then Vals (2).Node.Kind not in Ada_Base_Type_Decl
            then
               raise Property_Error
                 with "mismatching node type for comb_prop";
            end if;
      Ret := Base_Type_Decl_P_Array_Concat_Result_Type (Node => From.Node, Other => (Node => Vals (2).Node,  Info => Vals (2).Info), Origin => Self.Origin, E_Info => From.Info);
      return (Node => Ret.Node, Info => Ret.Info);
   end Combine;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Logic_Functor_581) return String is
   begin
      return ("BaseTypeDecl.array_concat_result_type");
   end Image;
   type Logic_Functor_552 is new Solver_Ifc.Converter_Type with
      record
            Origin : Bare_Ada_Node;
      end record;
   overriding function Convert (Self : Logic_Functor_552; From : Internal_Entity) return Internal_Entity with Inline;
   overriding function Image (Self : Logic_Functor_552) return String;
   overriding function Convert (Self : Logic_Functor_552; From : Internal_Entity) return Internal_Entity is
      Ret : Internal_Entity_Base_Type_Decl;
   begin
            if From.Node /= null
               and then From.Node.Kind not in Ada_Base_Type_Decl
            then
               raise Property_Error
                 with "mismatching node type for conv_prop";
            end if;
      Ret := Base_Type_Decl_P_Base_Subtype (Node => From.Node, Origin => Self.Origin, E_Info => From.Info);
      return (Node => Ret.Node, Info => Ret.Info);
   end Convert;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Logic_Functor_552) return String is
   begin
      return ("BaseTypeDecl.base_subtype");
   end Image;
   type Logic_Functor_620 is new Solver_Ifc.Converter_Type with
      record
            Origin : Bare_Ada_Node;
      end record;
   overriding function Convert (Self : Logic_Functor_620; From : Internal_Entity) return Internal_Entity with Inline;
   overriding function Image (Self : Logic_Functor_620) return String;
   overriding function Convert (Self : Logic_Functor_620; From : Internal_Entity) return Internal_Entity is
      Ret : Internal_Entity_Base_Type_Decl;
   begin
            if From.Node /= null
               and then From.Node.Kind not in Ada_Base_Type_Decl
            then
               raise Property_Error
                 with "mismatching node type for conv_prop";
            end if;
      Ret := Base_Type_Decl_P_Comp_Type (Node => From.Node, Origin => Self.Origin, E_Info => From.Info);
      return (Node => Ret.Node, Info => Ret.Info);
   end Convert;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Logic_Functor_620) return String is
   begin
      return ("BaseTypeDecl.comp_type");
   end Image;
   type Logic_Functor_645 is new Solver_Ifc.Converter_Type with
      record
            Origin : Bare_Ada_Node;
      end record;
   overriding function Convert (Self : Logic_Functor_645; From : Internal_Entity) return Internal_Entity with Inline;
   overriding function Image (Self : Logic_Functor_645) return String;
   overriding function Convert (Self : Logic_Functor_645; From : Internal_Entity) return Internal_Entity is
      Ret : Internal_Entity_Basic_Decl;
   begin
            if From.Node /= null
               and then From.Node.Kind not in Ada_Base_Type_Decl
            then
               raise Property_Error
                 with "mismatching node type for conv_prop";
            end if;
      Ret := Base_Type_Decl_P_Corresponding_Char_Literal (Node => From.Node, Origin => Self.Origin, E_Info => From.Info);
      return (Node => Ret.Node, Info => Ret.Info);
   end Convert;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Logic_Functor_645) return String is
   begin
      return ("BaseTypeDecl.corresponding_char_literal");
   end Image;
   type Logic_Functor_578 is new Solver_Ifc.Converter_Type with
      record
            Origin : Bare_Ada_Node;
      end record;
   overriding function Convert (Self : Logic_Functor_578; From : Internal_Entity) return Internal_Entity with Inline;
   overriding function Image (Self : Logic_Functor_578) return String;
   overriding function Convert (Self : Logic_Functor_578; From : Internal_Entity) return Internal_Entity is
      Ret : Internal_Entity_Base_Type_Decl;
   begin
            if From.Node /= null
               and then From.Node.Kind not in Ada_Base_Type_Decl
            then
               raise Property_Error
                 with "mismatching node type for conv_prop";
            end if;
      Ret := Base_Type_Decl_P_Derefed_Base_Subtype (Node => From.Node, Origin => Self.Origin, E_Info => From.Info);
      return (Node => Ret.Node, Info => Ret.Info);
   end Convert;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Logic_Functor_578) return String is
   begin
      return ("BaseTypeDecl.derefed_base_subtype");
   end Image;
   type Logic_Functor_577 is new Solver_Ifc.Converter_Type with
      record
            Origin : Bare_Ada_Node;
      end record;
   overriding function Convert (Self : Logic_Functor_577; From : Internal_Entity) return Internal_Entity with Inline;
   overriding function Image (Self : Logic_Functor_577) return String;
   overriding function Convert (Self : Logic_Functor_577; From : Internal_Entity) return Internal_Entity is
      Ret : Internal_Entity_Base_Type_Decl;
   begin
            if From.Node /= null
               and then From.Node.Kind not in Ada_Base_Type_Decl
            then
               raise Property_Error
                 with "mismatching node type for conv_prop";
            end if;
      Ret := Base_Type_Decl_P_Derefed_Type (Node => From.Node, Origin => Self.Origin, E_Info => From.Info);
      return (Node => Ret.Node, Info => Ret.Info);
   end Convert;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Logic_Functor_577) return String is
   begin
      return ("BaseTypeDecl.derefed_type");
   end Image;
   type Logic_Functor_582 is new Solver_Ifc.Combiner_Type with
      record
            Origin : Bare_Ada_Node;
      end record;
   overriding function Combine (Self : Logic_Functor_582; Vals : Entity_Vars.Value_Array) return Internal_Entity with Inline;
   overriding function Image (Self : Logic_Functor_582) return String;
   overriding function Combine (Self : Logic_Functor_582; Vals : Entity_Vars.Value_Array) return Internal_Entity is
         From : constant Internal_Entity := Vals (1);
      Ret : Internal_Entity_Base_Type_Decl;
   begin
            if From.Node /= null
               and then From.Node.Kind not in Ada_Base_Type_Decl
            then
               raise Property_Error
                 with "mismatching node type for comb_prop";
            end if;
            if Vals (2).Node /= null
               and then Vals (2).Node.Kind not in Ada_Base_Type_Decl
            then
               raise Property_Error
                 with "mismatching node type for comb_prop";
            end if;
      Ret := Base_Type_Decl_P_Expected_Array_Concat_Operand_Type (Node => From.Node, Operand_Type => (Node => Vals (2).Node,  Info => Vals (2).Info), Origin => Self.Origin, E_Info => From.Info);
      return (Node => Ret.Node, Info => Ret.Info);
   end Combine;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Logic_Functor_582) return String is
   begin
      return ("BaseTypeDecl.expected_array_concat_operand_type");
   end Image;
   type Logic_Functor_580 is new Solver_Ifc.Combiner_Type with
      record
            Origin : Bare_Ada_Node;
      end record;
   overriding function Combine (Self : Logic_Functor_580; Vals : Entity_Vars.Value_Array) return Internal_Entity with Inline;
   overriding function Image (Self : Logic_Functor_580) return String;
   overriding function Combine (Self : Logic_Functor_580; Vals : Entity_Vars.Value_Array) return Internal_Entity is
         From : constant Internal_Entity := Vals (1);
      Ret : Internal_Entity_Base_Type_Decl;
   begin
            if From.Node /= null
               and then From.Node.Kind not in Ada_Base_Type_Decl
            then
               raise Property_Error
                 with "mismatching node type for comb_prop";
            end if;
            if Vals (2).Node /= null
               and then Vals (2).Node.Kind not in Ada_Base_Type_Decl
            then
               raise Property_Error
                 with "mismatching node type for comb_prop";
            end if;
      Ret := Base_Type_Decl_P_Non_Universal_Base_Subtype (Node => From.Node, Second => (Node => Vals (2).Node,  Info => Vals (2).Info), Origin => Self.Origin, E_Info => From.Info);
      return (Node => Ret.Node, Info => Ret.Info);
   end Combine;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Logic_Functor_580) return String is
   begin
      return ("BaseTypeDecl.non_universal_base_subtype");
   end Image;
   type Logic_Functor_587 is new Solver_Ifc.Converter_Type with
      null record;
   overriding function Convert (Self : Logic_Functor_587; From : Internal_Entity) return Internal_Entity with Inline;
   overriding function Image (Self : Logic_Functor_587) return String;
   overriding function Convert (Self : Logic_Functor_587; From : Internal_Entity) return Internal_Entity is
         pragma Unreferenced (Self);
      Ret : Internal_Entity_Discrete_Base_Subtype_Decl;
   begin
            if From.Node /= null
               and then From.Node.Kind not in Ada_Base_Type_Decl
            then
               raise Property_Error
                 with "mismatching node type for conv_prop";
            end if;
      Ret := Base_Type_Decl_P_Scalar_Base_Type (Node => From.Node, E_Info => From.Info);
      return (Node => Ret.Node, Info => Ret.Info);
   end Convert;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Logic_Functor_587) return String is
   begin
      return ("BaseTypeDecl.scalar_base_type");
   end Image;
   type Logic_Functor_204 is new Solver_Ifc.Converter_Type with
      record
            Origin : Bare_Ada_Node;
      end record;
   overriding function Convert (Self : Logic_Functor_204; From : Internal_Entity) return Internal_Entity with Inline;
   overriding function Image (Self : Logic_Functor_204) return String;
   overriding function Convert (Self : Logic_Functor_204; From : Internal_Entity) return Internal_Entity is
      Ret : Internal_Entity_Base_Type_Decl;
   begin
            if From.Node /= null
               and then From.Node.Kind not in Ada_Basic_Decl
            then
               raise Property_Error
                 with "mismatching node type for conv_prop";
            end if;
      Ret := Basic_Decl_P_Identity_Type (Node => From.Node, Origin => Self.Origin, E_Info => From.Info);
      return (Node => Ret.Node, Info => Ret.Info);
   end Convert;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Logic_Functor_204) return String is
   begin
      return ("BasicDecl.identity_type");
   end Image;
   type Logic_Functor_209 is new Solver_Ifc.Converter_Type with
      null record;
   overriding function Convert (Self : Logic_Functor_209; From : Internal_Entity) return Internal_Entity with Inline;
   overriding function Image (Self : Logic_Functor_209) return String;
   overriding function Convert (Self : Logic_Functor_209; From : Internal_Entity) return Internal_Entity is
         pragma Unreferenced (Self);
      Ret : Internal_Entity_Base_Subp_Spec;
   begin
            if From.Node /= null
               and then From.Node.Kind not in Ada_Basic_Decl
            then
               raise Property_Error
                 with "mismatching node type for conv_prop";
            end if;
      Ret := Basic_Decl_P_Subp_Spec_Or_Null (Node => From.Node, E_Info => From.Info);
      return (Node => Ret.Node, Info => Ret.Info);
   end Convert;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Logic_Functor_209) return String is
   begin
      return ("BasicDecl.subp_spec_or_null");
   end Image;
   type Logic_Functor_586 is new Solver_Ifc.Converter_Type with
      null record;
   overriding function Convert (Self : Logic_Functor_586; From : Internal_Entity) return Internal_Entity with Inline;
   overriding function Image (Self : Logic_Functor_586) return String;
   overriding function Convert (Self : Logic_Functor_586; From : Internal_Entity) return Internal_Entity is
         pragma Unreferenced (Self);
      Ret : Internal_Entity_Classwide_Type_Decl;
   begin
            if From.Node /= null
               and then From.Node.Kind not in Ada_Base_Type_Decl
            then
               raise Property_Error
                 with "mismatching node type for conv_prop";
            end if;
      Ret := Dispatcher_Base_Type_Decl_P_Classwide_Type (Node => From.Node, E_Info => From.Info);
      return (Node => Ret.Node, Info => Ret.Info);
   end Convert;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Logic_Functor_586) return String is
   begin
      return ("BaseTypeDecl.classwide_type");
   end Image;
   type Logic_Functor_207 is new Solver_Ifc.Converter_Type with
      record
            Origin : Bare_Ada_Node;
      end record;
   overriding function Convert (Self : Logic_Functor_207; From : Internal_Entity) return Internal_Entity with Inline;
   overriding function Image (Self : Logic_Functor_207) return String;
   overriding function Convert (Self : Logic_Functor_207; From : Internal_Entity) return Internal_Entity is
      Ret : Internal_Entity_Base_Type_Decl;
   begin
            if From.Node /= null
               and then From.Node.Kind not in Ada_Basic_Decl
            then
               raise Property_Error
                 with "mismatching node type for conv_prop";
            end if;
      Ret := Dispatcher_Basic_Decl_P_Expr_Type (Node => From.Node, Origin => Self.Origin, E_Info => From.Info);
      return (Node => Ret.Node, Info => Ret.Info);
   end Convert;
   -----------
   -- Image --
   -----------
   overriding function Image (Self : Logic_Functor_207) return String is
   begin
      return ("BasicDecl.expr_type");
   end Image;
   --
   --  Primitives for Bare_Ada_Node
   --
--# property-start AdaNode.declarative_scope /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:110
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Declarative_Scope
  (Node : Bare_Ada_Node
  )
   return Bare_Declarative_Part
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Declarative_Part;
            procedure Finalizer_Scope_2 with Inline_Always;
      Node_Parents : Bare_Ada_Node_Array_Access;
Is_A : Boolean;
Find_Result : Bare_Ada_Node;
Cast_Result : Bare_Declarative_Part;
            procedure Finalizer_Scope_2 is
            begin
                     Dec_Ref (Node_Parents);
            end Finalizer_Scope_2;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4 '<Cast to DeclarativePart at ast.py:111>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:111
--# expr-start 3 '<Find at ast.py:111>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:111
   --# expr-start 1 '<FieldAccess for parents at ast.py:111>' Node_Parents /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:111
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents := Parents (Self, True); 
--# expr-done 1
      Find_Result := No_Bare_Ada_Node;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents;
      begin
         for Item_0 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_0
      --# expr-start 2 '<IsA DeclarativePart at ast.py:112>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:112
Is_A := Item_0 /= null 
and then Item_0.Kind in Ada_Declarative_Part_Range; 
--# expr-done 2
      if Is_A then
         Find_Result := Item_0;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 3
   if Find_Result = null
      or else Find_Result.Kind in Ada_Declarative_Part_Range
   then
      Cast_Result := Find_Result;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 4
      Property_Result := Cast_Result;
   --# end
      Finalizer_Scope_2;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_2;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Declarative_Scope;
--# end
--# property-start AdaNode.owning_unit_kind /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:120
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Owning_Unit_Kind
  (Node : Bare_Ada_Node
  )
   return Analysis_Unit_Kind
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Analysis_Unit_Kind;
      Fld : Internal_Unit;
Fld_1 : Bare_Ada_Node;
Cast_Result : Bare_Compilation_Unit;
Fld_2 : Analysis_Unit_Kind;
Let_Result : Analysis_Unit_Kind;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8 '<FieldAccess for unit_kind at ast.py:124>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:124
--# expr-start 7 '<Cast to CompilationUnit at ast.py:124>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:124
--# expr-start 6 '<FieldAccess for root at ast.py:124>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:124
--# expr-start 5 '<FieldAccess for unit at ast.py:124>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:124
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Unit (Node => Self);
--# expr-done 5
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Ast_Root;
--# expr-done 6
   if Fld_1 = null
      or else Fld_1.Kind in Ada_Compilation_Unit_Range
   then
      Cast_Result := Fld_1;
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 7
      if Cast_Result = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.unit_kind
Fld_2 := Libadalang.Implementation.Compilation_Unit_P_Unit_Kind (Node => Cast_Result);
--# end
--# expr-done 8
Let_Result := Fld_2; 
--# end
--# expr-done 9
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Owning_Unit_Kind;
--# end
--# property-start AdaNode.withed_unit_helper /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:127
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Withed_Unit_Helper
  (Node : Bare_Ada_Node
      ; Unit_Name : Bare_Name
  )
   return Bare_Compilation_Unit
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind unit_name Unit_Name
   Property_Result : Bare_Compilation_Unit;
            procedure Finalizer_Scope_4 with Inline_Always;
      Block_Var : Symbol_Type_Array_Access;
Block_Var_1 : Bare_Compilation_Unit;
Block_Var_2 : Bare_Compilation_Unit;
Block_Var_3 : Bare_Compilation_Unit;
Var_Expr : Bare_Compilation_Unit;
Var_Expr_1 : Bare_Compilation_Unit;
Fld : Symbol_Type_Array_Access;
Fld_1 : Bare_Compilation_Unit;
Fld_2 : Bare_Compilation_Unit;
Is_Null : Boolean;
If_Result : Bare_Compilation_Unit;
Fld_3 : Bare_Compilation_Unit;
Fld_4 : Bare_Basic_Decl;
Scope_Result : Bare_Basic_Decl;
Result_Var : Bare_Basic_Decl;
Is_A : Boolean;
Is_Null_1 : Boolean;
If_Result_1 : Boolean;
Is_Null_2 : Boolean;
Is_Null_3 : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Boolean;
If_Result_4 : Bare_Compilation_Unit;
Scope_Result_1 : Bare_Compilation_Unit;
Result_Var_1 : Bare_Compilation_Unit;
Let_Result : Bare_Compilation_Unit;
            procedure Finalizer_Scope_4 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Block_Var);
            end Finalizer_Scope_4;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 22 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10 '<FieldAccess for as_symbol_array at ast.py:138>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:138
      if Unit_Name = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.as_symbol_array'
Fld := Libadalang.Implementation.Dispatcher_Name_P_As_Symbol_Array (Node => Unit_Name);
--# end
--# expr-done 10
Block_Var := Fld; Inc_Ref (Block_Var);
--# bind unit_name_array Block_Var
--# expr-start 11 '<FieldAccess for designated_compilation_unit at ast.py:139>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:139
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.designated_compilation_unit
Fld_1 := Libadalang.Implementation.Ada_Node_P_Designated_Compilation_Unit (Node => Self, Name => Block_Var, Kind => Unit_Specification, Not_Found_Is_Error => False);
--# end
--# expr-done 11
Block_Var_1 := Fld_1; 
--# bind spec Block_Var_1
--# expr-start 14 '<If at ast.py:144>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:144
--# expr-start 12 '<IsNull at ast.py:145>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:145
Is_Null := Block_Var_1 = null; 
--# expr-done 12
if Is_Null then
   --# expr-start 13 '<FieldAccess for designated_compilation_unit at ast.py:146>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:146
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.designated_compilation_unit
Fld_2 := Libadalang.Implementation.Ada_Node_P_Designated_Compilation_Unit (Node => Self, Name => Block_Var, Kind => Unit_Body, Not_Found_Is_Error => False);
--# end
--# expr-done 13
   If_Result := Fld_2;
else
   If_Result := No_Bare_Ada_Node;
end if;
--# expr-done 14
Block_Var_2 := If_Result; 
--# bind body Block_Var_2
--# expr-start 21 '<If at ast.py:158>' If_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:158
--# expr-start 16 '<IsA PackageBody at ast.py:159>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:159
Var_Expr := Block_Var_2;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 15 '<FieldAccess for decl at ast.py:159>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:159
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.decl
Fld_4 := Libadalang.Implementation.Compilation_Unit_P_Decl (Node => Var_Expr);
--# end
--# expr-done 15
Scope_Result := Fld_4;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node;
end if;
Is_A := Result_Var /= null 
and then Result_Var.Kind in Ada_Package_Body_Range; 
--# expr-done 16
if Is_A then
   --# expr-start 17 '<IsNull at ast.py:159>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:159
Is_Null_1 := Block_Var_1 = null; 
--# expr-done 17
   If_Result_1 := Is_Null_1;
else
   If_Result_1 := False;
end if;
if If_Result_1 then
   If_Result_3 := True;
else
--# expr-start 18 '<IsNull at ast.py:160>' Is_Null_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:160
Is_Null_2 := Block_Var_1 = null; 
--# expr-done 18
if Is_Null_2 then
   --# expr-start 19 '<IsNull at ast.py:160>' Is_Null_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:160
Is_Null_3 := Block_Var_2 = null; 
--# expr-done 19
   If_Result_2 := Is_Null_3;
else
   If_Result_2 := False;
end if;
   If_Result_3 := If_Result_2;
end if;
if If_Result_3 then
   --# expr-start 20 '<FieldAccess for designated_compilation_unit at ast.py:162>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:162
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.designated_compilation_unit
Fld_3 := Libadalang.Implementation.Ada_Node_P_Designated_Compilation_Unit (Node => Self, Name => Block_Var, Kind => Unit_Specification, Not_Found_Is_Error => True);
--# end
--# expr-done 20
   If_Result_4 := Fld_3;
else
   If_Result_4 := No_Bare_Ada_Node;
end if;
--# expr-done 21
Block_Var_3 := If_Result_4; 
Var_Expr_1 := Block_Var_1;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind e Var_Expr_1
Scope_Result_1 := Var_Expr_1;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := Block_Var_2;
end if;
Let_Result := Result_Var_1; 
--# end
--# expr-done 22
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_4;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_4;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Withed_Unit_Helper;
--# end
--# property-start '[dispatcher]AdaNode.custom_id_text' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Ada_Node_P_Custom_Id_Text
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return String_Type
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : String_Type;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Ada_Node (Self.Kind) is
               when Ada_Abort_Absent .. Ada_Param_Assoc | Ada_Case_Stmt_Alternative .. Ada_Signed_Int_Type_Def | Ada_Synthetic_Type_Expr .. Ada_With_Private_Present =>
                  --# property-call-start AdaNode.custom_id_text
                  Property_Result := Ada_Node_P_Custom_Id_Text
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Abstract_State_Decl .. Ada_Single_Task_Decl =>
                  --# property-call-start BasicDecl.custom_id_text
                  Property_Result := Basic_Decl_P_Custom_Id_Text
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Anonymous_Type =>
                  --# property-call-start AnonymousType.custom_id_text
                  Property_Result := Anonymous_Type_P_Custom_Id_Text
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Enum_Lit_Synth_Type_Expr =>
                  --# property-call-start EnumLitSynthTypeExpr.custom_id_text
                  Property_Result := Enum_Lit_Synth_Type_Expr_P_Custom_Id_Text
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Subtype_Indication .. Ada_Discrete_Subtype_Indication =>
                  --# property-call-start SubtypeIndication.custom_id_text
                  Property_Result := Subtype_Indication_P_Custom_Id_Text
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Ada_Node_P_Custom_Id_Text;
--# end
--# property-start AdaNode.is_contract_aspect /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:183
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Is_Contract_Aspect
  (Node : Bare_Ada_Node
      ; Name : Symbol_Type
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind name Name
   Property_Result : Boolean;
      Sym : Symbol_Type;
Is_Equal : Boolean;
Sym_1 : Symbol_Type;
Is_Equal_1 : Boolean;
If_Result : Boolean;
Sym_2 : Symbol_Type;
Is_Equal_2 : Boolean;
If_Result_1 : Boolean;
Sym_3 : Symbol_Type;
Is_Equal_3 : Boolean;
If_Result_2 : Boolean;
Sym_4 : Symbol_Type;
Is_Equal_4 : Boolean;
If_Result_3 : Boolean;
Sym_5 : Symbol_Type;
Is_Equal_5 : Boolean;
If_Result_4 : Boolean;
Sym_6 : Symbol_Type;
Is_Equal_6 : Boolean;
If_Result_5 : Boolean;
Sym_7 : Symbol_Type;
Is_Equal_7 : Boolean;
If_Result_6 : Boolean;
Sym_8 : Symbol_Type;
Is_Equal_8 : Boolean;
If_Result_7 : Boolean;
Sym_9 : Symbol_Type;
Is_Equal_9 : Boolean;
If_Result_8 : Boolean;
Sym_10 : Symbol_Type;
Is_Equal_10 : Boolean;
If_Result_9 : Boolean;
Sym_11 : Symbol_Type;
Is_Equal_11 : Boolean;
If_Result_10 : Boolean;
Sym_12 : Symbol_Type;
Is_Equal_12 : Boolean;
If_Result_11 : Boolean;
Sym_13 : Symbol_Type;
Is_Equal_13 : Boolean;
If_Result_12 : Boolean;
Sym_14 : Symbol_Type;
Is_Equal_14 : Boolean;
If_Result_13 : Boolean;
Sym_15 : Symbol_Type;
Is_Equal_15 : Boolean;
If_Result_14 : Boolean;
Sym_16 : Symbol_Type;
Is_Equal_16 : Boolean;
If_Result_15 : Boolean;
Sym_17 : Symbol_Type;
Is_Equal_17 : Boolean;
If_Result_16 : Boolean;
Sym_18 : Symbol_Type;
Is_Equal_18 : Boolean;
If_Result_17 : Boolean;
Sym_19 : Symbol_Type;
Is_Equal_19 : Boolean;
If_Result_18 : Boolean;
Sym_20 : Symbol_Type;
Is_Equal_20 : Boolean;
If_Result_19 : Boolean;
Sym_21 : Symbol_Type;
Is_Equal_21 : Boolean;
If_Result_20 : Boolean;
Sym_22 : Symbol_Type;
Is_Equal_22 : Boolean;
If_Result_21 : Boolean;
Sym_23 : Symbol_Type;
Is_Equal_23 : Boolean;
If_Result_22 : Boolean;
Sym_24 : Symbol_Type;
Is_Equal_24 : Boolean;
If_Result_23 : Boolean;
Sym_25 : Symbol_Type;
Is_Equal_25 : Boolean;
If_Result_24 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 75 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 24 '<Eq at ast.py:184>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 23 '<Symbol Pre at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Pre); 
--# expr-done 23
Is_Equal := Name = Sym; 
--# expr-done 24
if Is_Equal then
   If_Result := True;
else
   --# expr-start 26 '<Eq at ast.py:184>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 25 '<Symbol Pre'"'"'Class at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Pre_Class); 
--# expr-done 25
Is_Equal_1 := Name = Sym_1; 
--# expr-done 26
   If_Result := Is_Equal_1;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 28 '<Eq at ast.py:184>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 27 '<Symbol Post at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Post); 
--# expr-done 27
Is_Equal_2 := Name = Sym_2; 
--# expr-done 28
   If_Result_1 := Is_Equal_2;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
   --# expr-start 30 '<Eq at ast.py:184>' Is_Equal_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 29 '<Symbol Post'"'"'Class at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Post_Class); 
--# expr-done 29
Is_Equal_3 := Name = Sym_3; 
--# expr-done 30
   If_Result_2 := Is_Equal_3;
end if;
if If_Result_2 then
   If_Result_3 := True;
else
   --# expr-start 32 '<Eq at ast.py:184>' Is_Equal_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 31 '<Symbol Refined_Post at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Refined_Post); 
--# expr-done 31
Is_Equal_4 := Name = Sym_4; 
--# expr-done 32
   If_Result_3 := Is_Equal_4;
end if;
if If_Result_3 then
   If_Result_4 := True;
else
   --# expr-start 34 '<Eq at ast.py:184>' Is_Equal_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 33 '<Symbol Precondition at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Precondition); 
--# expr-done 33
Is_Equal_5 := Name = Sym_5; 
--# expr-done 34
   If_Result_4 := Is_Equal_5;
end if;
if If_Result_4 then
   If_Result_5 := True;
else
   --# expr-start 36 '<Eq at ast.py:184>' Is_Equal_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 35 '<Symbol Postcondition at ???>' Sym_6 None
Sym_6 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Postcondition); 
--# expr-done 35
Is_Equal_6 := Name = Sym_6; 
--# expr-done 36
   If_Result_5 := Is_Equal_6;
end if;
if If_Result_5 then
   If_Result_6 := True;
else
   --# expr-start 38 '<Eq at ast.py:184>' Is_Equal_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 37 '<Symbol Precondition'"'"'Class at ???>' Sym_7 None
Sym_7 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Precondition_Class); 
--# expr-done 37
Is_Equal_7 := Name = Sym_7; 
--# expr-done 38
   If_Result_6 := Is_Equal_7;
end if;
if If_Result_6 then
   If_Result_7 := True;
else
   --# expr-start 40 '<Eq at ast.py:184>' Is_Equal_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 39 '<Symbol Postcondition'"'"'Class at ???>' Sym_8 None
Sym_8 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Postcondition_Class); 
--# expr-done 39
Is_Equal_8 := Name = Sym_8; 
--# expr-done 40
   If_Result_7 := Is_Equal_8;
end if;
if If_Result_7 then
   If_Result_8 := True;
else
   --# expr-start 42 '<Eq at ast.py:184>' Is_Equal_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 41 '<Symbol Invariant at ???>' Sym_9 None
Sym_9 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Invariant); 
--# expr-done 41
Is_Equal_9 := Name = Sym_9; 
--# expr-done 42
   If_Result_8 := Is_Equal_9;
end if;
if If_Result_8 then
   If_Result_9 := True;
else
   --# expr-start 44 '<Eq at ast.py:184>' Is_Equal_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 43 '<Symbol Invariant'"'"'Class at ???>' Sym_10 None
Sym_10 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Invariant_Class); 
--# expr-done 43
Is_Equal_10 := Name = Sym_10; 
--# expr-done 44
   If_Result_9 := Is_Equal_10;
end if;
if If_Result_9 then
   If_Result_10 := True;
else
   --# expr-start 46 '<Eq at ast.py:184>' Is_Equal_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 45 '<Symbol Type_Invariant at ???>' Sym_11 None
Sym_11 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Type_Invariant); 
--# expr-done 45
Is_Equal_11 := Name = Sym_11; 
--# expr-done 46
   If_Result_10 := Is_Equal_11;
end if;
if If_Result_10 then
   If_Result_11 := True;
else
   --# expr-start 48 '<Eq at ast.py:184>' Is_Equal_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 47 '<Symbol Type_Invariant'"'"'Class at ???>' Sym_12 None
Sym_12 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Type_Invariant_Class); 
--# expr-done 47
Is_Equal_12 := Name = Sym_12; 
--# expr-done 48
   If_Result_11 := Is_Equal_12;
end if;
if If_Result_11 then
   If_Result_12 := True;
else
   --# expr-start 50 '<Eq at ast.py:184>' Is_Equal_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 49 '<Symbol Predicate at ???>' Sym_13 None
Sym_13 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Predicate); 
--# expr-done 49
Is_Equal_13 := Name = Sym_13; 
--# expr-done 50
   If_Result_12 := Is_Equal_13;
end if;
if If_Result_12 then
   If_Result_13 := True;
else
   --# expr-start 52 '<Eq at ast.py:184>' Is_Equal_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 51 '<Symbol Static_Predicate at ???>' Sym_14 None
Sym_14 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Static_Predicate); 
--# expr-done 51
Is_Equal_14 := Name = Sym_14; 
--# expr-done 52
   If_Result_13 := Is_Equal_14;
end if;
if If_Result_13 then
   If_Result_14 := True;
else
   --# expr-start 54 '<Eq at ast.py:184>' Is_Equal_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 53 '<Symbol Dynamic_Predicate at ???>' Sym_15 None
Sym_15 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Dynamic_Predicate); 
--# expr-done 53
Is_Equal_15 := Name = Sym_15; 
--# expr-done 54
   If_Result_14 := Is_Equal_15;
end if;
if If_Result_14 then
   If_Result_15 := True;
else
   --# expr-start 56 '<Eq at ast.py:184>' Is_Equal_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 55 '<Symbol Default_Initial_Condition at ???>' Sym_16 None
Sym_16 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Default_Initial_Condition); 
--# expr-done 55
Is_Equal_16 := Name = Sym_16; 
--# expr-done 56
   If_Result_15 := Is_Equal_16;
end if;
if If_Result_15 then
   If_Result_16 := True;
else
   --# expr-start 58 '<Eq at ast.py:184>' Is_Equal_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 57 '<Symbol Contract_Cases at ???>' Sym_17 None
Sym_17 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Contract_Cases); 
--# expr-done 57
Is_Equal_17 := Name = Sym_17; 
--# expr-done 58
   If_Result_16 := Is_Equal_17;
end if;
if If_Result_16 then
   If_Result_17 := True;
else
   --# expr-start 60 '<Eq at ast.py:184>' Is_Equal_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 59 '<Symbol Test_Case at ???>' Sym_18 None
Sym_18 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Test_Case); 
--# expr-done 59
Is_Equal_18 := Name = Sym_18; 
--# expr-done 60
   If_Result_17 := Is_Equal_18;
end if;
if If_Result_17 then
   If_Result_18 := True;
else
   --# expr-start 62 '<Eq at ast.py:184>' Is_Equal_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 61 '<Symbol Global at ???>' Sym_19 None
Sym_19 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Global); 
--# expr-done 61
Is_Equal_19 := Name = Sym_19; 
--# expr-done 62
   If_Result_18 := Is_Equal_19;
end if;
if If_Result_18 then
   If_Result_19 := True;
else
   --# expr-start 64 '<Eq at ast.py:184>' Is_Equal_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 63 '<Symbol Refined_Global at ???>' Sym_20 None
Sym_20 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Refined_Global); 
--# expr-done 63
Is_Equal_20 := Name = Sym_20; 
--# expr-done 64
   If_Result_19 := Is_Equal_20;
end if;
if If_Result_19 then
   If_Result_20 := True;
else
   --# expr-start 66 '<Eq at ast.py:184>' Is_Equal_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 65 '<Symbol Refined_State at ???>' Sym_21 None
Sym_21 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Refined_State); 
--# expr-done 65
Is_Equal_21 := Name = Sym_21; 
--# expr-done 66
   If_Result_20 := Is_Equal_21;
end if;
if If_Result_20 then
   If_Result_21 := True;
else
   --# expr-start 68 '<Eq at ast.py:184>' Is_Equal_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 67 '<Symbol Stable_Properties at ???>' Sym_22 None
Sym_22 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Stable_Properties); 
--# expr-done 67
Is_Equal_22 := Name = Sym_22; 
--# expr-done 68
   If_Result_21 := Is_Equal_22;
end if;
if If_Result_21 then
   If_Result_22 := True;
else
   --# expr-start 70 '<Eq at ast.py:184>' Is_Equal_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 69 '<Symbol Depends at ???>' Sym_23 None
Sym_23 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Depends); 
--# expr-done 69
Is_Equal_23 := Name = Sym_23; 
--# expr-done 70
   If_Result_22 := Is_Equal_23;
end if;
if If_Result_22 then
   If_Result_23 := True;
else
   --# expr-start 72 '<Eq at ast.py:184>' Is_Equal_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 71 '<Symbol Refined_Depends at ???>' Sym_24 None
Sym_24 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Refined_Depends); 
--# expr-done 71
Is_Equal_24 := Name = Sym_24; 
--# expr-done 72
   If_Result_23 := Is_Equal_24;
end if;
if If_Result_23 then
   If_Result_24 := True;
else
   --# expr-start 74 '<Eq at ast.py:184>' Is_Equal_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:184
--# expr-start 73 '<Symbol Predicate_Failure at ???>' Sym_25 None
Sym_25 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Predicate_Failure); 
--# expr-done 73
Is_Equal_25 := Name = Sym_25; 
--# expr-done 74
   If_Result_24 := Is_Equal_25;
end if;
Let_Result := If_Result_24; 
--# end
--# expr-done 75
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Is_Contract_Aspect;
--# end
--# property-start AdaNode.in_contract /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:198
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_In_Contract
  (Node : Bare_Ada_Node
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
            procedure Finalizer_Scope_7 with Inline_Always;
      Var_Expr : Boolean;
Var_Expr_1 : Bare_Aspect_Assoc;
Var_Expr_2 : Bare_Pragma_Node;
Node_Parents : Bare_Ada_Node_Array_Access;
Cast_Expr : Bare_Ada_Node;
Cast_Result : Bare_Aspect_Assoc;
Fld : Bare_Name;
As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld_1 : Symbol_Type;
Fld_2 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Scope_Result_1 : Boolean;
Cast_Expr_1 : Bare_Ada_Node;
Cast_Result_1 : Bare_Pragma_Node;
Fld_3 : Bare_Identifier;
As_Entity_1 : Internal_Entity_Identifier;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Identifier;
Fld_4 : Symbol_Type;
Fld_5 : Boolean;
Scope_Result_2 : Boolean;
Result_Var_1 : Boolean;
Result_Var_2 : Boolean;
Find_Result : Bare_Ada_Node;
Is_Null_2 : Boolean;
Not_Val : Boolean;
            procedure Finalizer_Scope_7 is
            begin
                     Dec_Ref (Node_Parents);
            end Finalizer_Scope_7;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 89 '<Not at ast.py:198>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:198
--# expr-start 88 '<IsNull at ast.py:198>' Is_Null_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:198
--# expr-start 87 '<Find at ast.py:198>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:198
   --# expr-start 76 '<FieldAccess for parents at ast.py:198>' Node_Parents /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:198
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents := Parents (Self, True); 
--# expr-done 76
      Find_Result := No_Bare_Ada_Node;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents;
      begin
         for Item_1 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_1
--# expr-start 77 '<Cast to AspectAssoc at ast.py:199>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:199
Cast_Expr := Item_1; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_Aspect_Assoc_Range
   then
      Cast_Result := Cast_Expr;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 77
Var_Expr_1 := Cast_Result;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind a Var_Expr_1
--# expr-start 81 '<FieldAccess for is_contract_aspect at ast.py:200>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:200
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 80 '<FieldAccess for name_symbol at ast.py:200>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:200
--# expr-start 79 '<as_bare_entity at ast.py:200>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:200
--# expr-start 78 '<FieldAccess for id at ast.py:200>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:200
      if Var_Expr_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Var_Expr_1.Aspect_Assoc_F_Id;
--# expr-done 78
Is_Null := Fld = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
As_Entity := (Info => No_Entity_Info, Node => Fld); 
   If_Result := As_Entity;
end if;
--# expr-done 79
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => If_Result.Node);
--# end
--# expr-done 80
--# property-call-start AdaNode.is_contract_aspect
Fld_2 := Libadalang.Implementation.Ada_Node_P_Is_Contract_Aspect (Node => Self, Name => Fld_1);
--# end
--# expr-done 81
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Var_Expr := Result_Var;
if Var_Expr /= False then
   --# scope-start
   --# bind e Var_Expr
Scope_Result_1 := Var_Expr;
   --# end
   Result_Var_2 := Scope_Result_1;
else
--# expr-start 82 '<Cast to Pragma at ast.py:201>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:201
Cast_Expr_1 := Item_1; 
   if Cast_Expr_1 = null
      or else Cast_Expr_1.Kind in Ada_Pragma_Node_Range
   then
      Cast_Result_1 := Cast_Expr_1;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 82
Var_Expr_2 := Cast_Result_1;
if Var_Expr_2 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind p Var_Expr_2
--# expr-start 86 '<FieldAccess for is_contract_aspect at ast.py:202>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:202
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 85 '<FieldAccess for name_symbol at ast.py:202>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:202
--# expr-start 84 '<as_bare_entity at ast.py:202>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:202
--# expr-start 83 '<FieldAccess for id at ast.py:202>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:202
      if Var_Expr_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Var_Expr_2.Pragma_Node_F_Id;
--# expr-done 83
Is_Null_1 := Fld_3 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Identifier;
else
As_Entity_1 := (Info => No_Entity_Info, Node => Fld_3); 
   If_Result_1 := As_Entity_1;
end if;
--# expr-done 84
      if If_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_4 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => If_Result_1.Node);
--# end
--# expr-done 85
--# property-call-start AdaNode.is_contract_aspect
Fld_5 := Libadalang.Implementation.Ada_Node_P_Is_Contract_Aspect (Node => Self, Name => Fld_4);
--# end
--# expr-done 86
Scope_Result_2 := Fld_5;
   --# end
   Result_Var_1 := Scope_Result_2;
else
   Result_Var_1 := False;
end if;
   Result_Var_2 := Result_Var_1;
end if;
      if Result_Var_2 then
         Find_Result := Item_1;
      end if;
   --# end
               exit when Result_Var_2;
         end loop;
      end;
--# expr-done 87
Is_Null_2 := Find_Result = null; 
--# expr-done 88
Not_Val := not (Is_Null_2); 
--# expr-done 89
      Property_Result := Not_Val;
   --# end
      Finalizer_Scope_7;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_7;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_In_Contract;
--# end
--# property-start AdaNode.in_aspect /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:207
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_In_Aspect
  (Node : Bare_Ada_Node
      ; Name : Symbol_Type
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind name Name
   Property_Result : Boolean;
            procedure Finalizer_Scope_8 with Inline_Always;
      Var_Expr : Bare_Aspect_Assoc;
Node_Parents : Bare_Ada_Node_Array_Access;
Cast_Expr : Bare_Ada_Node;
Cast_Result : Bare_Aspect_Assoc;
Fld : Bare_Name;
Fld_1 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Quantifier_Result : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_8 is
            begin
                     Dec_Ref (Node_Parents);
            end Finalizer_Scope_8;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 95 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 94 '<AnyQuantifier at ast.py:211>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:211
   --# expr-start 90 '<FieldAccess for parents at ast.py:211>' Node_Parents /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:211
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents := Parents (Self, True); 
--# expr-done 90
      Quantifier_Result := False;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents;
      begin
         for Item_2 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_2
--# expr-start 91 '<Cast to AspectAssoc at ast.py:212>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:212
Cast_Expr := Item_2; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_Aspect_Assoc_Range
   then
      Cast_Result := Cast_Expr;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 91
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind a Var_Expr
--# expr-start 93 '<FieldAccess for name_is at ast.py:212>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:212
--# expr-start 92 '<FieldAccess for id at ast.py:212>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:212
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Var_Expr.Aspect_Assoc_F_Id;
--# expr-done 92
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_is
Fld_1 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld, Sym => Name);
--# end
--# expr-done 93
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
      Quantifier_Result := Result_Var;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
--# expr-done 94
Let_Result := Quantifier_Result; 
--# end
--# expr-done 95
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_8;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_8;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_In_Aspect;
--# end
--# property-start AdaNode.empty_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:215
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Empty_Env
  (Node : Bare_Ada_Node
  )
   return Lexical_Env
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_9 with Inline_Always;
      Node_Parents : Bare_Ada_Node_Array_Access;
Is_A : Boolean;
Find_Result : Bare_Ada_Node;
Cast_Result : Bare_Compilation_Unit;
Fld : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_9 is
            begin
                     Dec_Ref (Node_Parents);
                     Dec_Ref (Fld);
            end Finalizer_Scope_9;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 100 '<FieldAccess for get_empty_env at ast.py:216>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:216
--# expr-start 99 '<Cast to CompilationUnit at ast.py:216>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:216
--# expr-start 98 '<Find at ast.py:216>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:216
   --# expr-start 96 '<FieldAccess for parents at ast.py:216>' Node_Parents /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:216
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents := Parents (Self, True); 
--# expr-done 96
      Find_Result := No_Bare_Ada_Node;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents;
      begin
         for Item_3 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_3
      --# expr-start 97 '<IsA CompilationUnit at ast.py:216>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:216
Is_A := Item_3 /= null 
and then Item_3.Kind in Ada_Compilation_Unit_Range; 
--# expr-done 97
      if Is_A then
         Find_Result := Item_3;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 98
   if Find_Result = null
      or else Find_Result.Kind in Ada_Compilation_Unit_Range
   then
      Cast_Result := Find_Result;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 99
      if Cast_Result = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Compilation_Unit_P_Get_Empty_Env (Node => Cast_Result);
--# expr-done 100
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_9;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_9;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Empty_Env;
--# end
--# property-start AdaNode.is_not_null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:221
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Is_Not_Null
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity :=
     Internal_Entity'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Is_Null : Boolean;
Not_Val : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 103 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 102 '<Not at ast.py:229>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:229
--# expr-start 101 '<IsNull at ast.py:229>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:229
Is_Null := Ent.Node = null; 
--# expr-done 101
Not_Val := not (Is_Null); 
--# expr-done 102
Let_Result := Not_Val; 
--# end
--# expr-done 103
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Is_Not_Null;
--# end
--# property-start AdaNode.eval_discrete_range /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:232
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Eval_Discrete_Range
  (Node : Bare_Ada_Node
      ; Dr : Internal_Discrete_Range
  )
   return Internal_Eval_Discrete_Range
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind dr Dr
   Property_Result : Internal_Eval_Discrete_Range;
            procedure Finalizer_Scope_11 with Inline_Always;
            procedure Finalizer_Scope_1691 with Inline_Always;
      Var_Expr : Internal_Entity_Expr;
Fld : Internal_Entity_Expr;
Fld_1 : Big_Integer_Type;
Scope_Result : Big_Integer_Type;
Big_Int : Big_Integer_Type;
Result_Var : Big_Integer_Type;
Fld_2 : Internal_Entity_Expr;
Fld_3 : Big_Integer_Type;
New_Struct : Internal_Eval_Discrete_Range := No_Eval_Discrete_Range;
Is_Equal : Boolean;
If_Result : Internal_Eval_Discrete_Range := No_Eval_Discrete_Range;
Let_Result : Internal_Eval_Discrete_Range := No_Eval_Discrete_Range;
            procedure Finalizer_Scope_11 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Big_Int);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Fld_3);
                     Dec_Ref (New_Struct);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_11;
            procedure Finalizer_Scope_1691 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1691;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 112 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 111 '<If at ast.py:236>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:236
--# expr-start 104 '<Eq at ast.py:237>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:237
Is_Equal := Dr = No_Discrete_Range; 
--# expr-done 104
if Is_Equal then
   If_Result := raise Property_Error with "Attempting to evaluate a null discrete range";
else
   --# expr-start 110 '<New EvalDiscreteRange at ast.py:242>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:242
--# expr-start 106 '<FieldAccess for eval_as_int at ast.py:247>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:247
--# expr-start 105 '<FieldAccess for high_bound at ast.py:247>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:247
Fld_2 := Dr.High_Bound;
--# expr-done 105
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.eval_as_int
Fld_3 := Libadalang.Implementation.Expr_P_Eval_As_Int (Node => Fld_2.Node, E_Info => Fld_2.Info);
--# end
--# expr-done 106
--# expr-start 107 '<FieldAccess for low_bound at ast.py:243>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:243
Fld := Dr.Low_Bound;
--# expr-done 107
Var_Expr := Fld;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind lb Var_Expr
--# expr-start 108 '<FieldAccess for eval_as_int at ast.py:244>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:244
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.eval_as_int
Fld_1 := Libadalang.Implementation.Expr_P_Eval_As_Int (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 108
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1691;
   Result_Var := Scope_Result;
else
   --# expr-start 109 '<BigInteger 0 at ast.py:245>' Big_Int /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:245
Big_Int := Create_Big_Integer ("0"); 
--# expr-done 109
   Result_Var := Big_Int;
end if;
      Inc_Ref (Result_Var);
Inc_Ref (Fld_3);
Inc_Ref (Result_Var);
New_Struct := (High_Bound => Fld_3, Low_Bound => Result_Var); 
--# expr-done 110
   If_Result := New_Struct;
end if;
      Inc_Ref (If_Result);
--# expr-done 111
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 112
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_11;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_11;
               Finalizer_Scope_1691;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Eval_Discrete_Range;
--# end
--# property-start AdaNode.sym_join /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:252
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Sym_Join
  (Node : Bare_Ada_Node
      ; Syms : Symbol_Type_Array_Access
      ; Sep : String_Type
  )
   return String_Type
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind syms Syms
   --# bind sep Sep
   Property_Result : String_Type;
            procedure Finalizer_Scope_12 with Inline_Always;
            procedure Finalizer_Scope_1692 with Inline_Always;
      Fld : String_Type;
Map_Result : String_Type_Array_Access;
Join_Result : String_Type;
Let_Result : String_Type;
            procedure Finalizer_Scope_12 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Join_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_12;
            procedure Finalizer_Scope_1692 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1692;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 116 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 115 '<Join at ast.py:256>' Join_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:256
--# expr-start 114 '<Map at ast.py:256>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:256
      declare
         Map_Result_Vec : String_Type_Vectors.Vector;
      begin
      declare
         Collection : constant Symbol_Type_Array_Access := Syms;
      begin
         for Item_4 of
               Collection.Items
         loop
   --# scope-start
               --# bind s Item_4
      --# expr-start 113 '<FieldAccess for image at ast.py:256>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:256
      if Item_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Image (Self => Item_4);
--# expr-done 113
         declare
            Item_To_Append : constant String_Type := Fld;
         begin
               Inc_Ref (Item_To_Append);
            String_Type_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_1692;
         end loop;
      end;
         Map_Result := Create_String_Type_Array
           (Items_Count => Natural (String_Type_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := String_Type_Vectors.Get
              (Map_Result_Vec,
               I + String_Type_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         String_Type_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 114
Join_Result := Join_Strings (Sep, Map_Result); 
--# expr-done 115
Let_Result := Join_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 116
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_12;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_12;
               Finalizer_Scope_1692;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Sym_Join;
--# end
--# property-start AdaNode.enclosing_compilation_unit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:260
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Enclosing_Compilation_Unit
  (Node : Bare_Ada_Node
  )
   return Bare_Compilation_Unit
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Compilation_Unit;
            procedure Finalizer_Scope_13 with Inline_Always;
      Node_Parents : Bare_Ada_Node_Array_Access;
Is_A : Boolean;
Find_Result : Bare_Ada_Node;
Cast_Result : Bare_Compilation_Unit;
Let_Result : Bare_Compilation_Unit;
            procedure Finalizer_Scope_13 is
            begin
                     Dec_Ref (Node_Parents);
            end Finalizer_Scope_13;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 121 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 120 '<Cast to CompilationUnit at ast.py:268>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:268
--# expr-start 119 '<Find at ast.py:268>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:268
   --# expr-start 117 '<FieldAccess for parents at ast.py:268>' Node_Parents /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:268
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents := Parents (Self, True); 
--# expr-done 117
      Find_Result := No_Bare_Ada_Node;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents;
      begin
         for Item_5 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_5
      --# expr-start 118 '<IsA CompilationUnit at ast.py:269>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:269
Is_A := Item_5 /= null 
and then Item_5.Kind in Ada_Compilation_Unit_Range; 
--# expr-done 118
      if Is_A then
         Find_Result := Item_5;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 119
   if Find_Result = null
      or else Find_Result.Kind in Ada_Compilation_Unit_Range
   then
      Cast_Result := Find_Result;
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 120
Let_Result := Cast_Result; 
--# end
--# expr-done 121
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_13;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_13;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Enclosing_Compilation_Unit;
--# end
--# property-start AdaNode.is_children_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:273
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Is_Children_Env
  (Node : Bare_Ada_Node
      ; Parent : Lexical_Env
      ; Current_Env : Lexical_Env
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind parent Parent
   --# bind current_env Current_Env
   Property_Result : Boolean;
            procedure Finalizer_Scope_14 with Inline_Always;
      Is_Equal : Boolean;
Is_Equal_1 : Boolean;
Is_Equal_2 : Boolean;
Env_Parent : Lexical_Env := Empty_Env;
Fld : Boolean;
If_Result : Boolean;
If_Result_1 : Boolean;
If_Result_2 : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_14 is
            begin
                     Dec_Ref (Env_Parent);
            end Finalizer_Scope_14;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 128 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 127 '<Cond at ast.py:278>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:278
--# expr-start 122 '<Eq at ast.py:279>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:279
Is_Equal := Equivalent (Parent, Empty_Env); 
--# expr-done 122
if Is_Equal then
   If_Result_2 := False;
else
--# expr-start 123 '<Eq at ast.py:280>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:280
Is_Equal_1 := Equivalent (Current_Env, Parent); 
--# expr-done 123
if Is_Equal_1 then
   If_Result_1 := True;
else
--# expr-start 124 '<IsNull at ast.py:281>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:281
Is_Equal_2 := Equivalent (Current_Env, Empty_Env); 
--# expr-done 124
if Is_Equal_2 then
   If_Result := False;
else
   --# expr-start 126 '<FieldAccess for is_children_env at ast.py:282>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:282
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 125 '<env_parent at ast.py:282>' Env_Parent /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:282
Env_Parent := AST_Envs.Parent (Current_Env); 
--# expr-done 125
--# property-call-start AdaNode.is_children_env
Fld := Libadalang.Implementation.Ada_Node_P_Is_Children_Env (Node => Self, Parent => Parent, Current_Env => Env_Parent);
--# end
--# expr-done 126
   If_Result := Fld;
end if;
   If_Result_1 := If_Result;
end if;
   If_Result_2 := If_Result_1;
end if;
--# expr-done 127
Let_Result := If_Result_2; 
--# end
--# expr-done 128
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_14;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_14;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Is_Children_Env;
--# end
--# property-start AdaNode.without_md /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:286
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Without_Md
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity :=
     Internal_Entity'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity;
      Fld : Bare_Ada_Node;
Fld_1 : Internal_Entity_Info;
Fld_2 : Env_Rebindings;
Fld_3 : Internal_Entity_Info;
Fld_4 : Boolean;
New_Struct : Internal_Entity_Info;
New_Struct_1 : Internal_Entity;
Let_Result : Internal_Entity;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 136 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 135 '<New AdaNode.entity at ???>' New_Struct_1 None
--# expr-start 133 '<New EntityInfo at ast.py:291>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:291
--# expr-start 130 '<FieldAccess for from_rebound at ast.py:294>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:294
--# expr-start 129 '<FieldAccess for info at ast.py:294>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:294
Fld_3 := Ent.Info;
--# expr-done 129
Fld_4 := Fld_3.From_Rebound;
--# expr-done 130
--# expr-start 132 '<FieldAccess for rebindings at ast.py:292>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:292
--# expr-start 131 '<FieldAccess for info at ast.py:292>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:292
Fld_1 := Ent.Info;
--# expr-done 131
Fld_2 := Fld_1.Rebindings;
--# expr-done 132
New_Struct := (From_Rebound => Fld_4, Md => No_Metadata, Rebindings => Fld_2); 
--# expr-done 133
--# expr-start 134 '<FieldAccess for node at ast.py:291>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:291
Fld := Ent.Node;
--# expr-done 134
New_Struct_1 := (Info => New_Struct, Node => Fld); 
--# expr-done 135
Let_Result := New_Struct_1; 
--# end
--# expr-done 136
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Without_Md;
--# end
--# property-start AdaNode.get_uninstantiated_node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:299
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Get_Uninstantiated_Node
  (Node : Bare_Ada_Node
  )
   return Internal_Entity
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity;
      As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Let_Result : Internal_Entity;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 138 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 137 '<as_bare_entity at ast.py:305>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:305
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 137
Let_Result := If_Result; 
--# end
--# expr-done 138
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Get_Uninstantiated_Node;
--# end
--# property-start AdaNode.complete /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:309
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Complete
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Completion_Item_Iterator_Access
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity :=
     Internal_Entity'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Completion_Item_Iterator_Access;
            procedure Finalizer_Scope_17 with Inline_Always;
      Fld : Internal_Completion_Item_Array_Access;
Fld_1 : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity_Synthetic_Subp_Decl;
Is_Null : Boolean;
Map_Result : Internal_Completion_Item_Array_Access;
Iter : Internal_Completion_Item_Iterator_Access;
Let_Result : Internal_Completion_Item_Iterator_Access;
            procedure Finalizer_Scope_17 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Iter);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_17;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 145 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 144 '<FieldAccess for to_iterator at ast.py:313>' Iter /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:313
--# expr-start 143 '<Filter at ast.py:313>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:313
   --# expr-start 139 '<FieldAccess for complete_items at ast.py:313>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:313
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.complete_items'
Fld := Libadalang.Implementation.Dispatcher_Ada_Node_P_Complete_Items (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 139
      declare
         Map_Result_Vec : Internal_Completion_Item_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Completion_Item_Array_Access := Fld;
      begin
         for Item_6 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_6
         --# expr-start 142 '<IsNull at ast.py:317>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:317
--# expr-start 141 '<Cast to SyntheticSubpDecl at ast.py:317>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:317
--# expr-start 140 '<FieldAccess for decl at ast.py:317>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:317
Fld_1 := Item_6.Decl;
--# expr-done 140
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Synthetic_Subp_Decl_Range
   then
      Cast_Result := Create_Internal_Entity_Synthetic_Subp_Decl
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         Cast_Result := No_Entity_Synthetic_Subp_Decl;
   end if;
--# expr-done 141
Is_Null := Cast_Result.Node = null; 
--# expr-done 142
         if Is_Null then
         declare
            Item_To_Append : constant Internal_Completion_Item := Item_6;
         begin
            Internal_Completion_Item_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Completion_Item_Array
           (Items_Count => Natural (Internal_Completion_Item_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Completion_Item_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Completion_Item_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Completion_Item_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 143
Iter := To_Iterator (Map_Result, Self.Unit.Context); 
--# expr-done 144
Let_Result := Iter; Inc_Ref (Let_Result);
--# end
--# expr-done 145
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_17;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_17;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Complete;
--# end
--# property-start '[dispatcher]AdaNode.complete_items' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Ada_Node_P_Complete_Items
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Completion_Item_Array_Access
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Completion_Item_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Ada_Node (Self.Kind) is
               when Ada_Abort_Absent .. Ada_Abstract_State_Decl_List | Ada_Constraint_List .. Ada_Discrete_Subtype_Name | Ada_End_Name .. Ada_Char_Literal | Ada_Op_Abs .. Ada_Enum_Lit_Synth_Type_Expr | Ada_Synthetic_Type_Expr .. Ada_With_Private_Present =>
                  --# property-call-start AdaNode.complete_items
                  Property_Result := Ada_Node_P_Complete_Items
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Alternatives_List =>
                  --# property-call-start AlternativesList.complete_items
                  Property_Result := Alternatives_List_P_Complete_Items
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Dotted_Name =>
                  --# property-call-start DottedName.complete_items
                  Property_Result := Dotted_Name_P_Complete_Items
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Identifier =>
                  --# property-call-start Identifier.complete_items
                  Property_Result := Identifier_P_Complete_Items
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Subtype_Indication .. Ada_Discrete_Subtype_Indication =>
                  --# property-call-start SubtypeIndication.complete_items
                  Property_Result := Subtype_Indication_P_Complete_Items
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Ada_Node_P_Complete_Items;
--# end
--# property-start AdaNode.valid_keywords /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:339
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Valid_Keywords
  (Node : Bare_Ada_Node
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_21 with Inline_Always;
      Sym : Symbol_Type;
Sym_1 : Symbol_Type;
Sym_2 : Symbol_Type;
Sym_3 : Symbol_Type;
Sym_4 : Symbol_Type;
Sym_5 : Symbol_Type;
Sym_6 : Symbol_Type;
Sym_7 : Symbol_Type;
Sym_8 : Symbol_Type;
Sym_9 : Symbol_Type;
Sym_10 : Symbol_Type;
Sym_11 : Symbol_Type;
Sym_12 : Symbol_Type;
Sym_13 : Symbol_Type;
Sym_14 : Symbol_Type;
Sym_15 : Symbol_Type;
Sym_16 : Symbol_Type;
Sym_17 : Symbol_Type;
Sym_18 : Symbol_Type;
Sym_19 : Symbol_Type;
Sym_20 : Symbol_Type;
Sym_21 : Symbol_Type;
Sym_22 : Symbol_Type;
Sym_23 : Symbol_Type;
Sym_24 : Symbol_Type;
Sym_25 : Symbol_Type;
Sym_26 : Symbol_Type;
Sym_27 : Symbol_Type;
Sym_28 : Symbol_Type;
Sym_29 : Symbol_Type;
Sym_30 : Symbol_Type;
Sym_31 : Symbol_Type;
Sym_32 : Symbol_Type;
Sym_33 : Symbol_Type;
Sym_34 : Symbol_Type;
Sym_35 : Symbol_Type;
Sym_36 : Symbol_Type;
Sym_37 : Symbol_Type;
Sym_38 : Symbol_Type;
Sym_39 : Symbol_Type;
Sym_40 : Symbol_Type;
Sym_41 : Symbol_Type;
Sym_42 : Symbol_Type;
Sym_43 : Symbol_Type;
Sym_44 : Symbol_Type;
Sym_45 : Symbol_Type;
Sym_46 : Symbol_Type;
Sym_47 : Symbol_Type;
Sym_48 : Symbol_Type;
Sym_49 : Symbol_Type;
Sym_50 : Symbol_Type;
Sym_51 : Symbol_Type;
Sym_52 : Symbol_Type;
Sym_53 : Symbol_Type;
Sym_54 : Symbol_Type;
Sym_55 : Symbol_Type;
Sym_56 : Symbol_Type;
Sym_57 : Symbol_Type;
Sym_58 : Symbol_Type;
Sym_59 : Symbol_Type;
Sym_60 : Symbol_Type;
Sym_61 : Symbol_Type;
Sym_62 : Symbol_Type;
Sym_63 : Symbol_Type;
Sym_64 : Symbol_Type;
Sym_65 : Symbol_Type;
Sym_66 : Symbol_Type;
Sym_67 : Symbol_Type;
Sym_68 : Symbol_Type;
Sym_69 : Symbol_Type;
Sym_70 : Symbol_Type;
Sym_71 : Symbol_Type;
Sym_72 : Symbol_Type;
Array_Lit : Symbol_Type_Array_Access;
Let_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_21 is
            begin
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_21;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 220 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 219 '<ArrayLiteral at ast.py:347>' Array_Lit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:347
--# expr-start 146 '<Symbol abort at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Abort); 
--# expr-done 146
--# expr-start 147 '<Symbol abs at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Abs_257); 
--# expr-done 147
--# expr-start 148 '<Symbol abstract at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Abstract); 
--# expr-done 148
--# expr-start 149 '<Symbol accept at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Accept); 
--# expr-done 149
--# expr-start 150 '<Symbol access at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Access_260); 
--# expr-done 150
--# expr-start 151 '<Symbol aliased at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Aliased); 
--# expr-done 151
--# expr-start 152 '<Symbol all at ???>' Sym_6 None
Sym_6 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_All); 
--# expr-done 152
--# expr-start 153 '<Symbol and at ???>' Sym_7 None
Sym_7 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_And_264); 
--# expr-done 153
--# expr-start 154 '<Symbol array at ???>' Sym_8 None
Sym_8 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Array); 
--# expr-done 154
--# expr-start 155 '<Symbol at at ???>' Sym_9 None
Sym_9 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_At); 
--# expr-done 155
--# expr-start 156 '<Symbol begin at ???>' Sym_10 None
Sym_10 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Begin); 
--# expr-done 156
--# expr-start 157 '<Symbol body at ???>' Sym_11 None
Sym_11 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Body); 
--# expr-done 157
--# expr-start 158 '<Symbol case at ???>' Sym_12 None
Sym_12 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Case); 
--# expr-done 158
--# expr-start 159 '<Symbol constant at ???>' Sym_13 None
Sym_13 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Constant); 
--# expr-done 159
--# expr-start 160 '<Symbol declare at ???>' Sym_14 None
Sym_14 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Declare); 
--# expr-done 160
--# expr-start 161 '<Symbol delay at ???>' Sym_15 None
Sym_15 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Delay); 
--# expr-done 161
--# expr-start 162 '<Symbol delta at ???>' Sym_16 None
Sym_16 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Delta_273); 
--# expr-done 162
--# expr-start 163 '<Symbol digits at ???>' Sym_17 None
Sym_17 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Digits_274); 
--# expr-done 163
--# expr-start 164 '<Symbol do at ???>' Sym_18 None
Sym_18 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Do); 
--# expr-done 164
--# expr-start 165 '<Symbol else at ???>' Sym_19 None
Sym_19 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Else); 
--# expr-done 165
--# expr-start 166 '<Symbol elsif at ???>' Sym_20 None
Sym_20 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Elsif); 
--# expr-done 166
--# expr-start 167 '<Symbol end at ???>' Sym_21 None
Sym_21 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_End); 
--# expr-done 167
--# expr-start 168 '<Symbol entry at ???>' Sym_22 None
Sym_22 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Entry); 
--# expr-done 168
--# expr-start 169 '<Symbol exception at ???>' Sym_23 None
Sym_23 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Exception); 
--# expr-done 169
--# expr-start 170 '<Symbol exit at ???>' Sym_24 None
Sym_24 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Exit); 
--# expr-done 170
--# expr-start 171 '<Symbol for at ???>' Sym_25 None
Sym_25 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_For); 
--# expr-done 171
--# expr-start 172 '<Symbol function at ???>' Sym_26 None
Sym_26 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Function); 
--# expr-done 172
--# expr-start 173 '<Symbol generic at ???>' Sym_27 None
Sym_27 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Generic); 
--# expr-done 173
--# expr-start 174 '<Symbol goto at ???>' Sym_28 None
Sym_28 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Goto); 
--# expr-done 174
--# expr-start 175 '<Symbol if at ???>' Sym_29 None
Sym_29 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_If); 
--# expr-done 175
--# expr-start 176 '<Symbol in at ???>' Sym_30 None
Sym_30 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_In); 
--# expr-done 176
--# expr-start 177 '<Symbol interface at ???>' Sym_31 None
Sym_31 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Interface_288); 
--# expr-done 177
--# expr-start 178 '<Symbol is at ???>' Sym_32 None
Sym_32 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Is); 
--# expr-done 178
--# expr-start 179 '<Symbol limited at ???>' Sym_33 None
Sym_33 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Limited); 
--# expr-done 179
--# expr-start 180 '<Symbol loop at ???>' Sym_34 None
Sym_34 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Loop); 
--# expr-done 180
--# expr-start 181 '<Symbol mod at ???>' Sym_35 None
Sym_35 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Mod_293); 
--# expr-done 181
--# expr-start 182 '<Symbol new at ???>' Sym_36 None
Sym_36 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_New); 
--# expr-done 182
--# expr-start 183 '<Symbol not at ???>' Sym_37 None
Sym_37 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Not_295); 
--# expr-done 183
--# expr-start 184 '<Symbol null at ???>' Sym_38 None
Sym_38 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Null); 
--# expr-done 184
--# expr-start 185 '<Symbol others at ???>' Sym_39 None
Sym_39 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Others); 
--# expr-done 185
--# expr-start 186 '<Symbol out at ???>' Sym_40 None
Sym_40 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Out); 
--# expr-done 186
--# expr-start 187 '<Symbol of at ???>' Sym_41 None
Sym_41 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Of); 
--# expr-done 187
--# expr-start 188 '<Symbol or at ???>' Sym_42 None
Sym_42 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Or_298); 
--# expr-done 188
--# expr-start 189 '<Symbol overriding at ???>' Sym_43 None
Sym_43 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Overriding); 
--# expr-done 189
--# expr-start 190 '<Symbol package at ???>' Sym_44 None
Sym_44 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Package); 
--# expr-done 190
--# expr-start 191 '<Symbol pragma at ???>' Sym_45 None
Sym_45 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Pragma); 
--# expr-done 191
--# expr-start 192 '<Symbol private at ???>' Sym_46 None
Sym_46 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Private); 
--# expr-done 192
--# expr-start 193 '<Symbol procedure at ???>' Sym_47 None
Sym_47 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Procedure); 
--# expr-done 193
--# expr-start 194 '<Symbol protected at ???>' Sym_48 None
Sym_48 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Protected); 
--# expr-done 194
--# expr-start 195 '<Symbol raise at ???>' Sym_49 None
Sym_49 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Raise); 
--# expr-done 195
--# expr-start 196 '<Symbol range at ???>' Sym_50 None
Sym_50 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Range_309); 
--# expr-done 196
--# expr-start 197 '<Symbol record at ???>' Sym_51 None
Sym_51 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Record); 
--# expr-done 197
--# expr-start 198 '<Symbol rem at ???>' Sym_52 None
Sym_52 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Rem_311); 
--# expr-done 198
--# expr-start 199 '<Symbol renames at ???>' Sym_53 None
Sym_53 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Renames); 
--# expr-done 199
--# expr-start 200 '<Symbol requeue at ???>' Sym_54 None
Sym_54 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Requeue); 
--# expr-done 200
--# expr-start 201 '<Symbol return at ???>' Sym_55 None
Sym_55 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Return); 
--# expr-done 201
--# expr-start 202 '<Symbol reverse at ???>' Sym_56 None
Sym_56 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Reverse); 
--# expr-done 202
--# expr-start 203 '<Symbol select at ???>' Sym_57 None
Sym_57 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Select); 
--# expr-done 203
--# expr-start 204 '<Symbol separate at ???>' Sym_58 None
Sym_58 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Separate); 
--# expr-done 204
--# expr-start 205 '<Symbol some at ???>' Sym_59 None
Sym_59 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Some); 
--# expr-done 205
--# expr-start 206 '<Symbol subtype at ???>' Sym_60 None
Sym_60 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Subtype); 
--# expr-done 206
--# expr-start 207 '<Symbol synchronized at ???>' Sym_61 None
Sym_61 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Synchronized); 
--# expr-done 207
--# expr-start 208 '<Symbol tagged at ???>' Sym_62 None
Sym_62 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Tagged); 
--# expr-done 208
--# expr-start 209 '<Symbol task at ???>' Sym_63 None
Sym_63 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Task); 
--# expr-done 209
--# expr-start 210 '<Symbol terminate at ???>' Sym_64 None
Sym_64 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Terminate); 
--# expr-done 210
--# expr-start 211 '<Symbol then at ???>' Sym_65 None
Sym_65 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Then); 
--# expr-done 211
--# expr-start 212 '<Symbol type at ???>' Sym_66 None
Sym_66 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Type); 
--# expr-done 212
--# expr-start 213 '<Symbol until at ???>' Sym_67 None
Sym_67 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Until); 
--# expr-done 213
--# expr-start 214 '<Symbol use at ???>' Sym_68 None
Sym_68 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Use); 
--# expr-done 214
--# expr-start 215 '<Symbol when at ???>' Sym_69 None
Sym_69 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_When); 
--# expr-done 215
--# expr-start 216 '<Symbol while at ???>' Sym_70 None
Sym_70 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_While); 
--# expr-done 216
--# expr-start 217 '<Symbol with at ???>' Sym_71 None
Sym_71 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_With); 
--# expr-done 217
--# expr-start 218 '<Symbol xor at ???>' Sym_72 None
Sym_72 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Xor_339); 
--# expr-done 218
Array_Lit := Create_Symbol_Type_Array (Internal_Symbol_Type_Array'(1 => Sym, 2 => Sym_1, 3 => Sym_2, 4 => Sym_3, 5 => Sym_4, 6 => Sym_5, 7 => Sym_6, 8 => Sym_7, 9 => Sym_8, 10 => Sym_9, 11 => Sym_10, 12 => Sym_11, 13 => Sym_12, 14 => Sym_13, 15 => Sym_14, 16 => Sym_15, 17 => Sym_16, 18 => Sym_17, 19 => Sym_18, 20 => Sym_19, 21 => Sym_20, 22 => Sym_21, 23 => Sym_22, 24 => Sym_23, 25 => Sym_24, 26 => Sym_25, 27 => Sym_26, 28 => Sym_27, 29 => Sym_28, 30 => Sym_29, 31 => Sym_30, 32 => Sym_31, 33 => Sym_32, 34 => Sym_33, 35 => Sym_34, 36 => Sym_35, 37 => Sym_36, 38 => Sym_37, 39 => Sym_38, 40 => Sym_39, 41 => Sym_40, 42 => Sym_41, 43 => Sym_42, 44 => Sym_43, 45 => Sym_44, 46 => Sym_45, 47 => Sym_46, 48 => Sym_47, 49 => Sym_48, 50 => Sym_49, 51 => Sym_50, 52 => Sym_51, 53 => Sym_52, 54 => Sym_53, 55 => Sym_54, 56 => Sym_55, 57 => Sym_56, 58 => Sym_57, 59 => Sym_58, 60 => Sym_59, 61 => Sym_60, 62 => Sym_61, 63 => Sym_62, 64 => Sym_63, 65 => Sym_64, 66 => Sym_65, 67 => Sym_66, 68 => Sym_67, 69 => Sym_68, 70 => Sym_69, 71 => Sym_70, 72 => Sym_71, 73 => Sym_72)); 
--# expr-done 219
Let_Result := Array_Lit; Inc_Ref (Let_Result);
--# end
--# expr-done 220
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_21;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_21;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Valid_Keywords;
--# end
--# property-start AdaNode.generic_instantiations /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:362
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Generic_Instantiations
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Generic_Instantiation_Array_Access
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity :=
     Internal_Entity'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Generic_Instantiation_Array_Access;
            procedure Finalizer_Scope_22 with Inline_Always;
      Fld : Internal_Entity_Info;
Fld_1 : Env_Rebindings;
Fld_2 : Internal_Entity_Generic_Instantiation_Array_Access;
Let_Result : Internal_Entity_Generic_Instantiation_Array_Access;
            procedure Finalizer_Scope_22 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_22;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 224 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 223 '<FieldAccess for generic_instantiations_internal at ast.py:368>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:368
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 222 '<FieldAccess for rebindings at ast.py:368>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:368
--# expr-start 221 '<FieldAccess for info at ast.py:368>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:368
Fld := Ent.Info;
--# expr-done 221
Fld_1 := Fld.Rebindings;
--# expr-done 222
--# property-call-start AdaNode.generic_instantiations_internal
Fld_2 := Libadalang.Implementation.Ada_Node_P_Generic_Instantiations_Internal (Node => Self, R => Fld_1);
--# end
--# expr-done 223
Let_Result := Fld_2; Inc_Ref (Let_Result);
--# end
--# expr-done 224
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_22;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_22;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Generic_Instantiations;
--# end
--# property-start AdaNode.generic_instantiations_internal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:371
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Generic_Instantiations_Internal
  (Node : Bare_Ada_Node
      ; R : Env_Rebindings
  )
   return Internal_Entity_Generic_Instantiation_Array_Access
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind r R
   Property_Result : Internal_Entity_Generic_Instantiation_Array_Access;
            procedure Finalizer_Scope_23 with Inline_Always;
      Head : Internal_Entity_Generic_Instantiation;
Tail : Internal_Entity_Generic_Instantiation_Array_Access;
Fld : Lexical_Env := Empty_Env;
Env_Node : Bare_Ada_Node;
Cast_Result : Bare_Generic_Instantiation;
As_Entity : Internal_Entity_Generic_Instantiation;
Is_Null : Boolean;
If_Result : Internal_Entity_Generic_Instantiation;
Fld_1 : Env_Rebindings;
Fld_2 : Internal_Entity_Generic_Instantiation_Array_Access;
Singleton : Internal_Entity_Generic_Instantiation_Array_Access;
Concat_Result : Internal_Entity_Generic_Instantiation_Array_Access;
Let_Result : Internal_Entity_Generic_Instantiation_Array_Access;
Is_Equal : Boolean;
If_Result_1 : Internal_Entity_Generic_Instantiation_Array_Access;
Let_Result_1 : Internal_Entity_Generic_Instantiation_Array_Access;
            procedure Finalizer_Scope_23 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Tail);
                     Dec_Ref (Singleton);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Let_Result);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Let_Result_1);
            end Finalizer_Scope_23;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 235 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 234 '<If at ast.py:372>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:372
--# expr-start 225 '<Eq at ast.py:373>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:373
Is_Equal := R = null; 
--# expr-done 225
if Is_Equal then
   If_Result_1 := No_Internal_Entity_Generic_Instantiation_Array_Type;
else
   --# expr-start 233 '<Let at ast.py:376>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:376
--# scope-start
--# expr-start 229 '<as_bare_entity at ast.py:377>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:377
--# expr-start 228 '<Cast to GenericInstantiation at ast.py:377>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:377
--# expr-start 227 '<env_node at ast.py:377>' Env_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:377
--# expr-start 226 '<FieldAccess for new_env at ast.py:377>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:377
      if R = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := R.New_Env;
Inc_Ref (Fld);
--# expr-done 226
Env_Node := AST_Envs.Env_Node (Fld); 
--# expr-done 227
   if Env_Node = null
      or else Env_Node.Kind in Ada_Generic_Instantiation
   then
      Cast_Result := Env_Node;
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 228
Is_Null := Cast_Result = null; 
if Is_Null then
   If_Result := No_Entity_Generic_Instantiation;
else
As_Entity := (Info => No_Entity_Info, Node => Cast_Result); 
   If_Result := As_Entity;
end if;
--# expr-done 229
Head := If_Result; 
--# bind head Head
--# expr-start 231 '<FieldAccess for generic_instantiations_internal at ast.py:379>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:379
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 230 '<FieldAccess for get_parent at ast.py:379>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:379
      if R = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := R.Parent;
--# expr-done 230
--# property-call-start AdaNode.generic_instantiations_internal
Fld_2 := Libadalang.Implementation.Ada_Node_P_Generic_Instantiations_Internal (Node => Self, R => Fld_1);
--# end
--# expr-done 231
Tail := Fld_2; Inc_Ref (Tail);
--# bind tail Tail
--# expr-start 232 '<Concat at ast.py:380>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:380
                Singleton := Create_Internal_Entity_Generic_Instantiation_Array (Items_Count => 1);
                Singleton.Items (1) := Head;
Concat_Result := Concat (Singleton, Tail); 
--# expr-done 232
Let_Result := Concat_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 233
   If_Result_1 := Let_Result;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 234
Let_Result_1 := If_Result_1; Inc_Ref (Let_Result_1);
--# end
--# expr-done 235
      Property_Result := Let_Result_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_23;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_23;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Generic_Instantiations_Internal;
--# end
--# property-start AdaNode.remove_rebindings /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:384
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Remove_Rebindings
  (Node : Bare_Ada_Node
      ; Base : Env_Rebindings
      ; Suffix : Env_Rebindings
  )
   return Env_Rebindings
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind base Base
   --# bind suffix Suffix
   Property_Result : Env_Rebindings;
            procedure Finalizer_Scope_24 with Inline_Always;
      Is_Equal : Boolean;
Is_Equal_1 : Boolean;
If_Result : Boolean;
Fld : Lexical_Env := Empty_Env;
Fld_1 : Lexical_Env := Empty_Env;
Is_Equal_2 : Boolean;
Fld_2 : Lexical_Env := Empty_Env;
Fld_3 : Lexical_Env := Empty_Env;
Is_Equal_3 : Boolean;
If_Result_1 : Boolean;
Fld_4 : Env_Rebindings;
Fld_5 : Env_Rebindings;
Fld_6 : Env_Rebindings;
If_Result_2 : Env_Rebindings;
If_Result_3 : Env_Rebindings;
Let_Result : Env_Rebindings;
            procedure Finalizer_Scope_24 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_24;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 248 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 247 '<Cond at ast.py:389>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:389
--# expr-start 236 '<IsNull at ast.py:390>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:390
Is_Equal := Base = null; 
--# expr-done 236
if Is_Equal then
   If_Result := True;
else
   --# expr-start 237 '<IsNull at ast.py:390>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:390
Is_Equal_1 := Suffix = null; 
--# expr-done 237
   If_Result := Is_Equal_1;
end if;
if If_Result then
   If_Result_3 := Base;
else
--# expr-start 240 '<Eq at ast.py:393>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:393
--# expr-start 238 '<FieldAccess for old_env at ast.py:393>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:393
      if Base = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Base.Old_Env;
Inc_Ref (Fld);
--# expr-done 238
--# expr-start 239 '<FieldAccess for old_env at ast.py:393>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:393
      if Suffix = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Suffix.Old_Env;
Inc_Ref (Fld_1);
--# expr-done 239
Is_Equal_2 := Equivalent (Fld, Fld_1); 
--# expr-done 240
if Is_Equal_2 then
   --# expr-start 243 '<Eq at ast.py:394>' Is_Equal_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:394
--# expr-start 241 '<FieldAccess for new_env at ast.py:394>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:394
      if Base = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Base.New_Env;
Inc_Ref (Fld_2);
--# expr-done 241
--# expr-start 242 '<FieldAccess for new_env at ast.py:394>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:394
      if Suffix = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Suffix.New_Env;
Inc_Ref (Fld_3);
--# expr-done 242
Is_Equal_3 := Equivalent (Fld_2, Fld_3); 
--# expr-done 243
   If_Result_1 := Is_Equal_3;
else
   If_Result_1 := False;
end if;
if If_Result_1 then
   --# expr-start 246 '<FieldAccess for remove_rebindings at ast.py:395>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:395
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 244 '<FieldAccess for get_parent at ast.py:395>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:395
      if Base = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Base.Parent;
--# expr-done 244
--# expr-start 245 '<FieldAccess for get_parent at ast.py:395>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:395
      if Suffix = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Suffix.Parent;
--# expr-done 245
--# property-call-start AdaNode.remove_rebindings
Fld_6 := Libadalang.Implementation.Ada_Node_P_Remove_Rebindings (Node => Self, Base => Fld_4, Suffix => Fld_5);
--# end
--# expr-done 246
   If_Result_2 := Fld_6;
else
   If_Result_2 := Base;
end if;
   If_Result_3 := If_Result_2;
end if;
--# expr-done 247
Let_Result := If_Result_3; 
--# end
--# expr-done 248
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_24;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_24;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Remove_Rebindings;
--# end
--# property-start AdaNode.logic_val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:405
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Logic_Val
  (Node : Bare_Ada_Node
      ; From_Node : Internal_Entity
      ; Lvar : Logic_Var
  )
   return Internal_Logic_Val_Result
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind from_node From_Node
   --# bind lvar Lvar
   Property_Result : Internal_Logic_Val_Result;
      Block_Var : Boolean;
Fld : Boolean;
Is_Logic_Var_Defined : Boolean;
Eq_Solution : Internal_Entity;
If_Result : Internal_Entity;
If_Result_1 : Internal_Entity;
New_Struct : Internal_Logic_Val_Result;
Let_Result : Internal_Logic_Val_Result;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 253 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 249 '<FieldAccess for resolve_names_from_closest_entry_point at ast.py:406>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:406
      if From_Node.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.resolve_names_from_closest_entry_point
Fld := Libadalang.Implementation.Ada_Node_P_Resolve_Names_From_Closest_Entry_Point (Node => From_Node.Node, E_Info => From_Node.Info);
--# end
--# expr-done 249
Block_Var := Fld; 
--# bind success Block_Var
--# expr-start 252 '<New LogicValResult at ast.py:408>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:408
--# expr-start 251 '<If at ast.py:408>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:408
if Block_Var then
   --# expr-start 250 '<get_value at ast.py:409>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:409
Is_Logic_Var_Defined := Entity_Vars.Is_Defined (Lvar); 
if Is_Logic_Var_Defined then
Eq_Solution := Entity_Vars.Get_Value (Lvar); 
   If_Result := Eq_Solution;
else
   If_Result := No_Entity;
end if;
--# expr-done 250
   If_Result_1 := If_Result;
else
   If_Result_1 := No_Entity;
end if;
--# expr-done 251
New_Struct := (Success => Block_Var, Value => If_Result_1); 
--# expr-done 252
Let_Result := New_Struct; 
--# end
--# expr-done 253
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Logic_Val;
--# end
--# property-start AdaNode.semantic_parent_helper /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:413
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Semantic_Parent_Helper
  (Node : Bare_Ada_Node
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity :=
     Internal_Entity'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   Property_Result : Internal_Entity;
            procedure Finalizer_Scope_1704 with Inline_Always;
      Var_Expr : Lexical_Env := Empty_Env;
Var_Expr_1 : Internal_Entity;
Env_Node : Bare_Ada_Node;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Scope_Result : Internal_Entity;
Env_Parent : Lexical_Env := Empty_Env;
Fld : Internal_Entity;
Result_Var : Internal_Entity;
Scope_Result_1 : Internal_Entity;
Result_Var_1 : Internal_Entity;
Let_Result : Internal_Entity;
            procedure Finalizer_Scope_1704 is
            begin
                     Dec_Ref (Var_Expr);
                     Dec_Ref (Env_Parent);
            end Finalizer_Scope_1704;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 258 '<Block at ???>' Let_Result None
--# scope-start
Var_Expr := Env;
      Inc_Ref (Var_Expr);
if Var_Expr /= Empty_Env then
   --# scope-start
   --# bind env Var_Expr
--# expr-start 255 '<as_entity at ast.py:414>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:414
--# expr-start 254 '<env_node at ast.py:414>' Env_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:414
Env_Node := AST_Envs.Env_Node (Var_Expr); 
--# expr-done 254
Is_Null := Env_Node = null; 
if Is_Null then
   If_Result := No_Entity;
else
As_Entity := (Info => E_Info, Node => Env_Node); 
   If_Result := As_Entity;
end if;
--# expr-done 255
Var_Expr_1 := If_Result;
if Var_Expr_1 /= No_Entity then
   --# scope-start
   --# bind e Var_Expr_1
Scope_Result := Var_Expr_1;
   --# end
   Result_Var := Scope_Result;
else
   --# expr-start 257 '<FieldAccess for semantic_parent_helper at ast.py:415>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:415
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 256 '<env_parent at ast.py:415>' Env_Parent /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:415
Env_Parent := AST_Envs.Parent (Var_Expr); 
--# expr-done 256
--# property-call-start AdaNode.semantic_parent_helper
Fld := Libadalang.Implementation.Ada_Node_P_Semantic_Parent_Helper (Node => Ent.Node, Env => Env_Parent, E_Info => Ent.Info);
--# end
--# expr-done 257
   Result_Var := Fld;
end if;
Scope_Result_1 := Result_Var;
   --# end
      Finalizer_Scope_1704;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity;
end if;
Let_Result := Result_Var_1; 
--# end
--# expr-done 258
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1704;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Semantic_Parent_Helper;
--# end
--# property-start AdaNode.semantic_parent /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:419
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Semantic_Parent
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity :=
     Internal_Entity'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity;
            procedure Finalizer_Scope_27 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Fld_1 : Internal_Entity;
Let_Result : Internal_Entity;
            procedure Finalizer_Scope_27 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_27;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 261 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 260 '<FieldAccess for semantic_parent_helper at ast.py:428>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:428
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 259 '<FieldAccess for node_env at ast.py:428>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:428
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Node_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 259
--# property-call-start AdaNode.semantic_parent_helper
Fld_1 := Libadalang.Implementation.Ada_Node_P_Semantic_Parent_Helper (Node => Ent.Node, Env => Fld, E_Info => Ent.Info);
--# end
--# expr-done 260
Let_Result := Fld_1; 
--# end
--# expr-done 261
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_27;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_27;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Semantic_Parent;
--# end
--# property-start AdaNode.is_in_top_level_public_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:431
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Is_In_Top_Level_Public_Part
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity :=
     Internal_Entity'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Block_Var : Internal_Entity;
Var_Expr : Internal_Entity_Base_Package_Decl;
Fld : Internal_Entity;
Is_Null : Boolean;
Is_A : Boolean;
Fld_1 : Internal_Entity;
Cast_Result : Internal_Entity_Base_Package_Decl;
Fld_2 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result : Boolean;
Is_A_1 : Boolean;
Cast_Expr : Internal_Entity;
Cast_Result_1 : Internal_Entity_Package_Body;
Fld_3 : Boolean;
If_Result_1 : Boolean;
Is_A_2 : Boolean;
Cast_Expr_1 : Internal_Entity;
Cast_Result_2 : Internal_Entity_Base_Package_Decl;
Fld_4 : Boolean;
If_Result_2 : Boolean;
Fld_5 : Boolean;
If_Result_3 : Boolean;
If_Result_4 : Boolean;
If_Result_5 : Boolean;
If_Result_6 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 276 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 262 '<FieldAccess for semantic_parent at ast.py:436>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:436
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.semantic_parent
Fld := Libadalang.Implementation.Ada_Node_P_Semantic_Parent (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 262
Block_Var := Fld; 
--# bind node Block_Var
--# expr-start 275 '<Cond at ast.py:437>' If_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:437
--# expr-start 263 '<IsNull at ast.py:438>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:438
Is_Null := Block_Var.Node = null; 
--# expr-done 263
if Is_Null then
   If_Result_6 := False;
else
--# expr-start 264 '<IsA PrivatePart at ast.py:441>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:441
Is_A := Block_Var.Node /= null 
and then Block_Var.Node.Kind in Ada_Private_Part_Range; 
--# expr-done 264
if Is_A then
--# expr-start 266 '<Cast to BasePackageDecl at ast.py:442>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:442
--# expr-start 265 '<parent at ast.py:442>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:442
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Parent (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# expr-done 265
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Base_Package_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Package_Decl
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         Cast_Result := No_Entity_Base_Package_Decl;
   end if;
--# expr-done 266
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Base_Package_Decl then
   --# scope-start
--# expr-start 267 '<FieldAccess for is_compilation_unit_root at ast.py:442>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:442
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_compilation_unit_root
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Is_Compilation_Unit_Root (Node => Var_Expr.Node);
--# end
--# expr-done 267
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
   If_Result := Result_Var;
else
   If_Result := False;
end if;
if If_Result then
   If_Result_5 := False;
else
--# expr-start 268 '<IsA PackageBody at ast.py:445>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:445
Is_A_1 := Block_Var.Node /= null 
and then Block_Var.Node.Kind in Ada_Package_Body_Range; 
--# expr-done 268
if Is_A_1 then
   --# expr-start 270 '<FieldAccess for is_compilation_unit_root at ast.py:446>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:446
--# expr-start 269 '<Cast to PackageBody at ast.py:446>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:446
Cast_Expr := Block_Var; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Package_Body_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Package_Body
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result_1 := No_Entity_Package_Body;
   end if;
--# expr-done 269
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_compilation_unit_root
Fld_3 := Libadalang.Implementation.Basic_Decl_P_Is_Compilation_Unit_Root (Node => Cast_Result_1.Node);
--# end
--# expr-done 270
   If_Result_1 := Fld_3;
else
   If_Result_1 := False;
end if;
if If_Result_1 then
   If_Result_4 := False;
else
--# expr-start 271 '<IsA BasePackageDecl at ast.py:449>' Is_A_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:449
Is_A_2 := Block_Var.Node /= null 
and then Block_Var.Node.Kind in Ada_Base_Package_Decl; 
--# expr-done 271
if Is_A_2 then
   --# expr-start 273 '<FieldAccess for is_compilation_unit_root at ast.py:450>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:450
--# expr-start 272 '<Cast to BasePackageDecl at ast.py:450>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:450
Cast_Expr_1 := Block_Var; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Base_Package_Decl
   then
      Cast_Result_2 := Create_Internal_Entity_Base_Package_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_2 := No_Entity_Base_Package_Decl;
   end if;
--# expr-done 272
      if Cast_Result_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_compilation_unit_root
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Is_Compilation_Unit_Root (Node => Cast_Result_2.Node);
--# end
--# expr-done 273
   If_Result_2 := Fld_4;
else
   If_Result_2 := False;
end if;
if If_Result_2 then
   If_Result_3 := True;
else
   --# expr-start 274 '<FieldAccess for is_in_top_level_public_part at ast.py:453>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:453
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.is_in_top_level_public_part
Fld_5 := Libadalang.Implementation.Ada_Node_P_Is_In_Top_Level_Public_Part (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 274
   If_Result_3 := Fld_5;
end if;
   If_Result_4 := If_Result_3;
end if;
   If_Result_5 := If_Result_4;
end if;
   If_Result_6 := If_Result_5;
end if;
--# expr-done 275
Let_Result := If_Result_6; 
--# end
--# expr-done 276
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Is_In_Top_Level_Public_Part;
--# end
--# property-start AdaNode.parent_basic_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:457
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Parent_Basic_Decl
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity :=
     Internal_Entity'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Basic_Decl;
      Var_Expr : Internal_Entity;
Var_Expr_1 : Internal_Entity_Basic_Decl;
Fld : Internal_Entity;
Fld_1 : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Fld_3 : Internal_Entity_Generic_Instantiation;
Cast_Expr_1 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Basic_Decl;
Cast_Result_2 : Internal_Entity_Basic_Decl;
Is_A : Boolean;
If_Result : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Fld_4 : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Scope_Result_1 : Internal_Entity_Basic_Decl;
Result_Var_1 : Internal_Entity_Basic_Decl;
Is_A_1 : Boolean;
If_Result_1 : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 288 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 287 '<If at ast.py:472>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:472
--# expr-start 277 '<IsA ClasswideTypeDecl, DiscreteBaseSubtypeDecl, SynthAnonymousTypeDecl at ast.py:477>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:477
Is_A_1 := Ent.Node /= null 
and then Ent.Node.Kind in Ada_Classwide_Type_Decl_Range | Ada_Discrete_Base_Subtype_Decl_Range | Ada_Synth_Anonymous_Type_Decl_Range; 
--# expr-done 277
if Is_A_1 then
   --# expr-start 279 '<FieldAccess for parent_basic_decl at ast.py:479>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:479
--# expr-start 278 '<FieldAccess for semantic_parent at ast.py:479>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:479
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.semantic_parent
Fld := Libadalang.Implementation.Ada_Node_P_Semantic_Parent (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 278
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.parent_basic_decl
Fld_1 := Libadalang.Implementation.Ada_Node_P_Parent_Basic_Decl (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 279
   If_Result_1 := Fld_1;
else
--# expr-start 280 '<FieldAccess for semantic_parent at ast.py:480>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:480
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.semantic_parent
Fld_2 := Libadalang.Implementation.Ada_Node_P_Semantic_Parent (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 280
Var_Expr := Fld_2;
if Var_Expr /= No_Entity then
   --# scope-start
   --# bind sp Var_Expr
--# expr-start 285 '<If at ast.py:481>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:481
--# expr-start 281 '<IsA GenericSubpInternal, GenericPackageInternal at ast.py:482>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:482
Is_A := Var_Expr.Node /= null 
and then Var_Expr.Node.Kind in Ada_Generic_Subp_Internal_Range | Ada_Generic_Package_Internal_Range; 
--# expr-done 281
if Is_A then
--# expr-start 283 '<FieldAccess for get_instantiation at ast.py:483>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:483
--# expr-start 282 '<Cast to BasicDecl at ast.py:483>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:483
Cast_Expr := Var_Expr; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 282
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.get_instantiation
Fld_3 := Libadalang.Implementation.Basic_Decl_P_Get_Instantiation (Node => Cast_Result.Node, E_Info => Cast_Result.Info);
--# end
--# expr-done 283
      Cast_Result_2 := Create_Internal_Entity_Basic_Decl
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
   If_Result := Cast_Result_2;
else
   --# expr-start 284 '<Cast to BasicDecl at ast.py:484>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:484
Cast_Expr_1 := Var_Expr; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_1 := No_Entity_Basic_Decl;
   end if;
--# expr-done 284
   If_Result := Cast_Result_1;
end if;
--# expr-done 285
Var_Expr_1 := If_Result;
if Var_Expr_1 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind e Var_Expr_1
Scope_Result := Var_Expr_1;
   --# end
   Result_Var := Scope_Result;
else
   --# expr-start 286 '<FieldAccess for parent_basic_decl at ast.py:485>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:485
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.parent_basic_decl
Fld_4 := Libadalang.Implementation.Ada_Node_P_Parent_Basic_Decl (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 286
   Result_Var := Fld_4;
end if;
Scope_Result_1 := Result_Var;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Basic_Decl;
end if;
   If_Result_1 := Result_Var_1;
end if;
--# expr-done 287
Let_Result := If_Result_1; 
--# end
--# expr-done 288
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Parent_Basic_Decl;
--# end
--# property-start '[dispatcher]AdaNode.immediate_declarative_region' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Ada_Node_P_Immediate_Declarative_Region
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Lexical_Env;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Ada_Node (Self.Kind) is
               when Ada_Abort_Absent .. Ada_Param_Assoc | Ada_Case_Stmt_Alternative .. Ada_Declarative_Part | Ada_Public_Part .. Ada_Begin_Block | Ada_Case_Stmt .. Ada_With_Private_Present =>
                  --# property-call-start AdaNode.immediate_declarative_region
                  Property_Result := Ada_Node_P_Immediate_Declarative_Region
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Abstract_State_Decl .. Ada_Single_Task_Decl =>
                  --# property-call-start BasicDecl.immediate_declarative_region
                  Property_Result := Basic_Decl_P_Immediate_Declarative_Region
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Private_Part =>
                  --# property-call-start PrivatePart.immediate_declarative_region
                  Property_Result := Private_Part_P_Immediate_Declarative_Region
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Decl_Block =>
                  --# property-call-start DeclBlock.immediate_declarative_region
                  Property_Result := Decl_Block_P_Immediate_Declarative_Region
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Ada_Node_P_Immediate_Declarative_Region;
--# end
--# property-start AdaNode.designated_compilation_unit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:533
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Designated_Compilation_Unit
  (Node : Bare_Ada_Node
      ; Name : Symbol_Type_Array_Access
      ; Kind : Analysis_Unit_Kind
      ; Load_If_Needed : Boolean
         := True
      ; Not_Found_Is_Error : Boolean
         := True
      ; Process_Parents : Boolean
         := True
  )
   return Bare_Compilation_Unit
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind name Name
   --# bind kind Kind
   --# bind load_if_needed Load_If_Needed
   --# bind not_found_is_error Not_Found_Is_Error
   --# bind process_parents Process_Parents
   Property_Result : Bare_Compilation_Unit;
      Block_Var : Internal_Unit;
Fld : Internal_Unit;
Fld_1 : Bare_Compilation_Unit;
Let_Result : Bare_Compilation_Unit;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 291 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 289 '<FieldAccess for get_unit at ast.py:543>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:543
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Ada_Node_P_Get_Unit (Node => Self, Name => Name, Kind => Kind, Load_If_Needed => Load_If_Needed, Not_Found_Is_Error => Not_Found_Is_Error, Process_Parents => Process_Parents);
--# expr-done 289
Block_Var := Fld; 
--# bind designated_analysis_unit Block_Var
--# expr-start 290 '<FieldAccess for compilation_unit_with_name at ast.py:547>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:547
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.compilation_unit_with_name
Fld_1 := Libadalang.Implementation.Ada_Node_P_Compilation_Unit_With_Name (Node => Self, Unit => Block_Var, Name => Name);
--# end
--# expr-done 290
Let_Result := Fld_1; 
--# end
--# expr-done 291
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Designated_Compilation_Unit;
--# end
--# property-start AdaNode.compilation_unit_with_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:550
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Compilation_Unit_With_Name
  (Node : Bare_Ada_Node
      ; Unit : Internal_Unit
      ; Name : Symbol_Type_Array_Access
  )
   return Bare_Compilation_Unit
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind unit Unit
   --# bind name Name
   Property_Result : Bare_Compilation_Unit;
            procedure Finalizer_Scope_1718 with Inline_Always;
      Var_Expr : Bare_Ada_Node;
Match_0 : Bare_Compilation_Unit;
Match_1 : Bare_Compilation_Unit_List;
Match_2 : Bare_Ada_Node;
Fld : Bare_Ada_Node;
Fld_1 : Symbol_Type_Array_Access;
Is_Equal : Boolean;
Item_8 : Bare_Compilation_Unit;
Find_Result : Bare_Compilation_Unit;
Match_Prefix : Bare_Ada_Node;
Cast_Result : Bare_Compilation_Unit;
Let_Result : Bare_Compilation_Unit;
Scope_Result : Bare_Compilation_Unit;
Cast_Result_1 : Bare_Compilation_Unit_List;
Let_Result_1 : Bare_Compilation_Unit;
Scope_Result_1 : Bare_Compilation_Unit;
Cast_Result_2 : Bare_Ada_Node;
Let_Result_2 : Bare_Compilation_Unit;
Scope_Result_2 : Bare_Compilation_Unit;
Match_Result : Bare_Compilation_Unit;
Scope_Result_3 : Bare_Compilation_Unit;
Result_Var : Bare_Compilation_Unit;
Let_Result_3 : Bare_Compilation_Unit;
            procedure Finalizer_Scope_1718 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1718;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 297 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 292 '<FieldAccess for root at ast.py:556>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:556
      if Unit = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Unit.Ast_Root;
--# expr-done 292
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 296 '<Match at ast.py:556>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:556
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Var_Expr; 
case Ada_Ada_Node (Match_Prefix.Kind) is
         when Ada_Compilation_Unit =>
   --# scope-start
--# scope-start
      Cast_Result := Match_Prefix;
Match_0 := Cast_Result; 
--# bind single Match_0
Let_Result := Match_0; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Compilation_Unit_List =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Match_Prefix;
Match_1 := Cast_Result_1; 
--# bind multi Match_1
--# expr-start 295 '<Find at ast.py:563>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:563
      Find_Result := No_Bare_Ada_Node;
      if Match_1 = null then
      Find_Result := No_Bare_Ada_Node;
      else
      declare
         Collection : constant Bare_Compilation_Unit_List := Match_1;
      begin
         for Untyped_Item_8 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_8 := Untyped_Item_8; 
   --# scope-start
               --# bind c Item_8
      --# expr-start 294 '<Eq at ast.py:564>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:564
--# expr-start 293 '<FieldAccess for syntactic_fully_qualified_name at ast.py:564>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:564
      if Item_8 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.syntactic_fully_qualified_name
Fld_1 := Libadalang.Implementation.Compilation_Unit_P_Syntactic_Fully_Qualified_Name (Node => Item_8);
--# end
--# expr-done 293
Is_Equal := Equivalent (Fld_1, Name); 
--# expr-done 294
      if Is_Equal then
         Find_Result := Item_8;
      end if;
   --# end
      Finalizer_Scope_1718;
               exit when Is_Equal;
         end loop;
      end;
      end if;
--# expr-done 295
Let_Result_1 := Find_Result; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abort_Absent .. Ada_Case_Stmt_Alternative_List | Ada_Contract_Case_Assoc_List .. Ada_Case_Stmt_Alternative | Ada_Component_Clause .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Match_Prefix;
Match_2 := Cast_Result_2; 
--# bind _ Match_2
Let_Result_2 := raise Property_Error with "Unexpected analysis unit root"; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 296
Scope_Result_3 := Match_Result;
   --# end
   Result_Var := Scope_Result_3;
else
   Result_Var := No_Bare_Ada_Node;
end if;
Let_Result_3 := Result_Var; 
--# end
--# expr-done 297
      Property_Result := Let_Result_3;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1718;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Compilation_Unit_With_Name;
--# end
--# property-start AdaNode.get_unit_root_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:573
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Get_Unit_Root_Decl
  (Node : Bare_Ada_Node
      ; Name : Symbol_Type_Array_Access
      ; Kind : Analysis_Unit_Kind
      ; Load_If_Needed : Boolean
         := True
      ; Not_Found_Is_Error : Boolean
         := True
      ; Process_Parents : Boolean
         := True
  )
   return Bare_Basic_Decl
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind name Name
   --# bind kind Kind
   --# bind load_if_needed Load_If_Needed
   --# bind not_found_is_error Not_Found_Is_Error
   --# bind process_parents Process_Parents
   Property_Result : Bare_Basic_Decl;
      Block_Var : Bare_Compilation_Unit;
Var_Expr : Bare_Compilation_Unit;
Fld : Bare_Compilation_Unit;
Fld_1 : Bare_Basic_Decl;
Scope_Result : Bare_Basic_Decl;
Result_Var : Bare_Basic_Decl;
Let_Result : Bare_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 300 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 298 '<FieldAccess for designated_compilation_unit at ast.py:582>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:582
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.designated_compilation_unit
Fld := Libadalang.Implementation.Ada_Node_P_Designated_Compilation_Unit (Node => Self, Name => Name, Kind => Kind, Load_If_Needed => Load_If_Needed, Not_Found_Is_Error => Not_Found_Is_Error, Process_Parents => Process_Parents);
--# end
--# expr-done 298
Block_Var := Fld; 
--# bind cu Block_Var
Var_Expr := Block_Var;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 299 '<FieldAccess for decl at ast.py:585>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:585
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.decl
Fld_1 := Libadalang.Implementation.Compilation_Unit_P_Decl (Node => Var_Expr);
--# end
--# expr-done 299
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 300
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Get_Unit_Root_Decl;
--# end
--# property-start '[dispatcher]AdaNode.xref_equation' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Ada_Node_P_Xref_Equation
  (Node : Bare_Ada_Node
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Ada_Node (Self.Kind) is
               when Ada_Abort_Absent .. Ada_All_Present | Ada_Aspect_Spec .. Ada_Pragma_Argument_Assoc | Ada_Component_List .. Ada_Record_Def | Ada_Composite_Constraint_Assoc | Ada_Abstract_State_Decl .. Ada_Anonymous_Expr_Decl | Ada_Synthetic_Formal_Param_Decl .. Ada_Discrete_Base_Subtype_Decl | Ada_Classwide_Type_Decl .. Ada_Incomplete_Tagged_Type_Decl | Ada_Abstract_Subp_Decl | Ada_Subp_Decl .. Ada_Synthetic_Subp_Decl | Ada_Null_Subp_Decl .. Ada_Subp_Body | Ada_Package_Body_Stub .. Ada_Task_Body_Stub | Ada_Package_Body .. Ada_Task_Body | Ada_Error_Decl | Ada_Generic_Package_Decl .. Ada_Generic_Subp_Decl | Ada_Label_Decl .. Ada_Named_Stmt_Decl | Ada_Single_Task_Decl | Ada_Compilation_Unit | Ada_Constant_Absent .. Ada_Constant_Present | Ada_Declarative_Part .. Ada_Elsif_Expr_Part | Ada_Case_Expr_Alternative | Ada_Contract_Cases | Ada_Discrete_Subtype_Name | Ada_Synthetic_Identifier | Ada_Handled_Stmts .. Ada_Limited_Present | Ada_Mode_Default .. Ada_Null_Component_Decl | Ada_Overriding_Not_Overriding .. Ada_Pp_Then_Kw | Ada_Private_Absent .. Ada_Quantifier_Some | Ada_Renaming_Clause .. Ada_Reverse_Present | Ada_Error_Stmt | Ada_Subp_Kind_Function .. Ada_Subp_Kind_Procedure | Ada_Synchronized_Absent .. Ada_Tagged_Present | Ada_Anonymous_Type_Access_Def | Ada_Enum_Lit_Synth_Type_Expr | Ada_Synthetic_Type_Expr .. Ada_Until_Present | Ada_Variant | Ada_With_Private_Absent .. Ada_With_Private_Present =>
                  --# property-call-start AdaNode.xref_equation
                  Property_Result := Ada_Node_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Constrained_Array_Indices =>
                  --# property-call-start ConstrainedArrayIndices.xref_equation
                  Property_Result := Constrained_Array_Indices_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Unconstrained_Array_Indices =>
                  --# property-call-start UnconstrainedArrayIndices.xref_equation
                  Property_Result := Unconstrained_Array_Indices_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Aspect_Assoc =>
                  --# property-call-start AspectAssoc.xref_equation
                  Property_Result := Aspect_Assoc_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_At_Clause =>
                  --# property-call-start AtClause.xref_equation
                  Property_Result := At_Clause_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Attribute_Def_Clause =>
                  --# property-call-start AttributeDefClause.xref_equation
                  Property_Result := Attribute_Def_Clause_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Enum_Rep_Clause =>
                  --# property-call-start EnumRepClause.xref_equation
                  Property_Result := Enum_Rep_Clause_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Record_Rep_Clause =>
                  --# property-call-start RecordRepClause.xref_equation
                  Property_Result := Record_Rep_Clause_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Enum_Subp_Spec .. Ada_Subp_Spec =>
                  --# property-call-start BaseSubpSpec.xref_equation
                  Property_Result := Base_Subp_Spec_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Entry_Spec =>
                  --# property-call-start EntrySpec.xref_equation
                  Property_Result := Entry_Spec_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Aggregate_Assoc .. Ada_Multi_Dim_Array_Assoc =>
                  --# property-call-start AggregateAssoc.xref_equation
                  Property_Result := Aggregate_Assoc_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Iterated_Assoc =>
                  --# property-call-start IteratedAssoc.xref_equation
                  Property_Result := Iterated_Assoc_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Param_Assoc =>
                  --# property-call-start ParamAssoc.xref_equation
                  Property_Result := Param_Assoc_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Component_Decl =>
                  --# property-call-start ComponentDecl.xref_equation
                  Property_Result := Component_Decl_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discriminant_Spec =>
                  --# property-call-start DiscriminantSpec.xref_equation
                  Property_Result := Discriminant_Spec_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Formal_Obj_Decl .. Ada_Generic_Formal_Type_Decl =>
                  --# property-call-start GenericFormal.xref_equation
                  Property_Result := Generic_Formal_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Param_Spec =>
                  --# property-call-start ParamSpec.xref_equation
                  Property_Result := Param_Spec_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Subtype_Decl =>
                  --# property-call-start SubtypeDecl.xref_equation
                  Property_Result := Subtype_Decl_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Protected_Type_Decl =>
                  --# property-call-start ProtectedTypeDecl.xref_equation
                  Property_Result := Protected_Type_Decl_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Task_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start TaskTypeDecl.xref_equation
                  Property_Result := Task_Type_Decl_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.xref_equation
                  Property_Result := Type_Decl_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Abstract_Formal_Subp_Decl .. Ada_Concrete_Formal_Subp_Decl =>
                  --# property-call-start FormalSubpDecl.xref_equation
                  Property_Result := Formal_Subp_Decl_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Expr_Function =>
                  --# property-call-start ExprFunction.xref_equation
                  Property_Result := Expr_Function_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Subp_Renaming_Decl =>
                  --# property-call-start SubpRenamingDecl.xref_equation
                  Property_Result := Subp_Renaming_Decl_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Entry_Body =>
                  --# property-call-start EntryBody.xref_equation
                  Property_Result := Entry_Body_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Entry_Index_Spec =>
                  --# property-call-start EntryIndexSpec.xref_equation
                  Property_Result := Entry_Index_Spec_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Exception_Decl =>
                  --# property-call-start ExceptionDecl.xref_equation
                  Property_Result := Exception_Decl_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Exception_Handler =>
                  --# property-call-start ExceptionHandler.xref_equation
                  Property_Result := Exception_Handler_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_For_Loop_Var_Decl =>
                  --# property-call-start ForLoopVarDecl.xref_equation
                  Property_Result := For_Loop_Var_Decl_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Package_Instantiation .. Ada_Generic_Subp_Instantiation =>
                  --# property-call-start GenericInstantiation.xref_equation
                  Property_Result := Generic_Instantiation_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Package_Renaming_Decl .. Ada_Generic_Subp_Renaming_Decl =>
                  --# property-call-start GenericRenamingDecl.xref_equation
                  Property_Result := Generic_Renaming_Decl_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Number_Decl =>
                  --# property-call-start NumberDecl.xref_equation
                  Property_Result := Number_Decl_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Object_Decl .. Ada_Extended_Return_Stmt_Object_Decl =>
                  --# property-call-start ObjectDecl.xref_equation
                  Property_Result := Object_Decl_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_No_Type_Object_Renaming_Decl =>
                  --# property-call-start NoTypeObjectRenamingDecl.xref_equation
                  Property_Result := No_Type_Object_Renaming_Decl_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Package_Renaming_Decl =>
                  --# property-call-start PackageRenamingDecl.xref_equation
                  Property_Result := Package_Renaming_Decl_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Single_Protected_Decl =>
                  --# property-call-start SingleProtectedDecl.xref_equation
                  Property_Result := Single_Protected_Decl_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Case_Stmt_Alternative =>
                  --# property-call-start CaseStmtAlternative.xref_equation
                  Property_Result := Case_Stmt_Alternative_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Component_Clause =>
                  --# property-call-start ComponentClause.xref_equation
                  Property_Result := Component_Clause_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Component_Def =>
                  --# property-call-start ComponentDef.xref_equation
                  Property_Result := Component_Def_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Composite_Constraint =>
                  --# property-call-start CompositeConstraint.xref_equation
                  Property_Result := Composite_Constraint_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Delta_Constraint =>
                  --# property-call-start DeltaConstraint.xref_equation
                  Property_Result := Delta_Constraint_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Digits_Constraint =>
                  --# property-call-start DigitsConstraint.xref_equation
                  Property_Result := Digits_Constraint_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Range_Constraint =>
                  --# property-call-start RangeConstraint.xref_equation
                  Property_Result := Range_Constraint_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Elsif_Stmt_Part =>
                  --# property-call-start ElsifStmtPart.xref_equation
                  Property_Result := Elsif_Stmt_Part_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Abstract_State_Decl_Expr =>
                  --# property-call-start AbstractStateDeclExpr.xref_equation
                  Property_Result := Abstract_State_Decl_Expr_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Allocator =>
                  --# property-call-start Allocator.xref_equation
                  Property_Result := Allocator_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Aggregate .. Ada_Null_Record_Aggregate =>
                  --# property-call-start BaseAggregate.xref_equation
                  Property_Result := Base_Aggregate_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Bin_Op .. Ada_Relation_Op =>
                  --# property-call-start BinOp.xref_equation
                  Property_Result := Bin_Op_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Box_Expr =>
                  --# property-call-start BoxExpr.xref_equation
                  Property_Result := Box_Expr_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Case_Expr =>
                  --# property-call-start CaseExpr.xref_equation
                  Property_Result := Case_Expr_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_If_Expr =>
                  --# property-call-start IfExpr.xref_equation
                  Property_Result := If_Expr_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Decl_Expr =>
                  --# property-call-start DeclExpr.xref_equation
                  Property_Result := Decl_Expr_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Membership_Expr =>
                  --# property-call-start MembershipExpr.xref_equation
                  Property_Result := Membership_Expr_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Attribute_Ref =>
                  --# property-call-start AttributeRef.xref_equation
                  Property_Result := Attribute_Ref_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Update_Attribute_Ref =>
                  --# property-call-start UpdateAttributeRef.xref_equation
                  Property_Result := Update_Attribute_Ref_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Call_Expr =>
                  --# property-call-start CallExpr.xref_equation
                  Property_Result := Call_Expr_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Defining_Name .. Ada_Synthetic_Defining_Name =>
                  --# property-call-start DefiningName.xref_equation
                  Property_Result := Defining_Name_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Dotted_Name =>
                  --# property-call-start DottedName.xref_equation
                  Property_Result := Dotted_Name_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_End_Name =>
                  --# property-call-start EndName.xref_equation
                  Property_Result := End_Name_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Explicit_Deref =>
                  --# property-call-start ExplicitDeref.xref_equation
                  Property_Result := Explicit_Deref_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Qual_Expr =>
                  --# property-call-start QualExpr.xref_equation
                  Property_Result := Qual_Expr_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Reduce_Attribute_Ref =>
                  --# property-call-start ReduceAttributeRef.xref_equation
                  Property_Result := Reduce_Attribute_Ref_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Identifier =>
                  --# property-call-start BaseId.xref_equation
                  Property_Result := Base_Id_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Char_Literal =>
                  --# property-call-start CharLiteral.xref_equation
                  Property_Result := Char_Literal_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Op_Abs .. Ada_Op_Xor =>
                  --# property-call-start Op.xref_equation
                  Property_Result := Op_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_String_Literal =>
                  --# property-call-start StringLiteral.xref_equation
                  Property_Result := String_Literal_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Null_Literal =>
                  --# property-call-start NullLiteral.xref_equation
                  Property_Result := Null_Literal_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Int_Literal =>
                  --# property-call-start IntLiteral.xref_equation
                  Property_Result := Int_Literal_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Real_Literal =>
                  --# property-call-start RealLiteral.xref_equation
                  Property_Result := Real_Literal_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Target_Name =>
                  --# property-call-start TargetName.xref_equation
                  Property_Result := Target_Name_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Paren_Expr =>
                  --# property-call-start ParenExpr.xref_equation
                  Property_Result := Paren_Expr_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Quantified_Expr =>
                  --# property-call-start QuantifiedExpr.xref_equation
                  Property_Result := Quantified_Expr_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Raise_Expr =>
                  --# property-call-start RaiseExpr.xref_equation
                  Property_Result := Raise_Expr_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Un_Op =>
                  --# property-call-start UnOp.xref_equation
                  Property_Result := Un_Op_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_For_Loop_Spec =>
                  --# property-call-start ForLoopSpec.xref_equation
                  Property_Result := For_Loop_Spec_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_While_Loop_Spec =>
                  --# property-call-start WhileLoopSpec.xref_equation
                  Property_Result := While_Loop_Spec_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Others_Designator =>
                  --# property-call-start OthersDesignator.xref_equation
                  Property_Result := Others_Designator_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Pragma_Node =>
                  --# property-call-start Pragma.xref_equation
                  Property_Result := Pragma_Node_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Range_Spec =>
                  --# property-call-start RangeSpec.xref_equation
                  Property_Result := Range_Spec_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Select_When_Part =>
                  --# property-call-start SelectWhenPart.xref_equation
                  Property_Result := Select_When_Part_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Accept_Stmt .. Ada_Accept_Stmt_With_Stmts =>
                  --# property-call-start AcceptStmt.xref_equation
                  Property_Result := Accept_Stmt_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_For_Loop_Stmt .. Ada_While_Loop_Stmt =>
                  --# property-call-start BaseLoopStmt.xref_equation
                  Property_Result := Base_Loop_Stmt_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Begin_Block .. Ada_Decl_Block =>
                  --# property-call-start BlockStmt.xref_equation
                  Property_Result := Block_Stmt_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Case_Stmt =>
                  --# property-call-start CaseStmt.xref_equation
                  Property_Result := Case_Stmt_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Extended_Return_Stmt =>
                  --# property-call-start ExtendedReturnStmt.xref_equation
                  Property_Result := Extended_Return_Stmt_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_If_Stmt =>
                  --# property-call-start IfStmt.xref_equation
                  Property_Result := If_Stmt_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Named_Stmt =>
                  --# property-call-start NamedStmt.xref_equation
                  Property_Result := Named_Stmt_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Select_Stmt =>
                  --# property-call-start SelectStmt.xref_equation
                  Property_Result := Select_Stmt_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Abort_Stmt =>
                  --# property-call-start AbortStmt.xref_equation
                  Property_Result := Abort_Stmt_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Assign_Stmt =>
                  --# property-call-start AssignStmt.xref_equation
                  Property_Result := Assign_Stmt_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Call_Stmt =>
                  --# property-call-start CallStmt.xref_equation
                  Property_Result := Call_Stmt_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Delay_Stmt =>
                  --# property-call-start DelayStmt.xref_equation
                  Property_Result := Delay_Stmt_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Exit_Stmt =>
                  --# property-call-start ExitStmt.xref_equation
                  Property_Result := Exit_Stmt_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Goto_Stmt =>
                  --# property-call-start GotoStmt.xref_equation
                  Property_Result := Goto_Stmt_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Label =>
                  --# property-call-start Label.xref_equation
                  Property_Result := Label_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Null_Stmt =>
                  --# property-call-start NullStmt.xref_equation
                  Property_Result := Null_Stmt_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Raise_Stmt =>
                  --# property-call-start RaiseStmt.xref_equation
                  Property_Result := Raise_Stmt_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Requeue_Stmt =>
                  --# property-call-start RequeueStmt.xref_equation
                  Property_Result := Requeue_Stmt_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Return_Stmt =>
                  --# property-call-start ReturnStmt.xref_equation
                  Property_Result := Return_Stmt_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Terminate_Alternative =>
                  --# property-call-start TerminateAlternative.xref_equation
                  Property_Result := Terminate_Alternative_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Subunit =>
                  --# property-call-start Subunit.xref_equation
                  Property_Result := Subunit_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Task_Def =>
                  --# property-call-start TaskDef.xref_equation
                  Property_Result := Task_Def_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Access_To_Subp_Def =>
                  --# property-call-start AccessToSubpDef.xref_equation
                  Property_Result := Access_To_Subp_Def_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Access_Def =>
                  --# property-call-start TypeAccessDef.xref_equation
                  Property_Result := Type_Access_Def_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Array_Type_Def =>
                  --# property-call-start ArrayTypeDef.xref_equation
                  Property_Result := Array_Type_Def_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Derived_Type_Def =>
                  --# property-call-start DerivedTypeDef.xref_equation
                  Property_Result := Derived_Type_Def_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Enum_Type_Def =>
                  --# property-call-start EnumTypeDef.xref_equation
                  Property_Result := Enum_Type_Def_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Formal_Discrete_Type_Def =>
                  --# property-call-start FormalDiscreteTypeDef.xref_equation
                  Property_Result := Formal_Discrete_Type_Def_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Interface_Type_Def =>
                  --# property-call-start InterfaceTypeDef.xref_equation
                  Property_Result := Interface_Type_Def_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Mod_Int_Type_Def =>
                  --# property-call-start ModIntTypeDef.xref_equation
                  Property_Result := Mod_Int_Type_Def_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Private_Type_Def =>
                  --# property-call-start PrivateTypeDef.xref_equation
                  Property_Result := Private_Type_Def_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Decimal_Fixed_Point_Def =>
                  --# property-call-start DecimalFixedPointDef.xref_equation
                  Property_Result := Decimal_Fixed_Point_Def_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Floating_Point_Def =>
                  --# property-call-start FloatingPointDef.xref_equation
                  Property_Result := Floating_Point_Def_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Ordinary_Fixed_Point_Def =>
                  --# property-call-start OrdinaryFixedPointDef.xref_equation
                  Property_Result := Ordinary_Fixed_Point_Def_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Record_Type_Def =>
                  --# property-call-start RecordTypeDef.xref_equation
                  Property_Result := Record_Type_Def_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Signed_Int_Type_Def =>
                  --# property-call-start SignedIntTypeDef.xref_equation
                  Property_Result := Signed_Int_Type_Def_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Anonymous_Type =>
                  --# property-call-start AnonymousType.xref_equation
                  Property_Result := Anonymous_Type_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Subtype_Indication .. Ada_Discrete_Subtype_Indication =>
                  --# property-call-start SubtypeIndication.xref_equation
                  Property_Result := Subtype_Indication_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Use_Package_Clause =>
                  --# property-call-start UsePackageClause.xref_equation
                  Property_Result := Use_Package_Clause_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Use_Type_Clause =>
                  --# property-call-start UseTypeClause.xref_equation
                  Property_Result := Use_Type_Clause_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Value_Sequence =>
                  --# property-call-start ValueSequence.xref_equation
                  Property_Result := Value_Sequence_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Variant_Part =>
                  --# property-call-start VariantPart.xref_equation
                  Property_Result := Variant_Part_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_With_Clause =>
                  --# property-call-start WithClause.xref_equation
                  Property_Result := With_Clause_P_Xref_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Ada_Node_P_Xref_Equation;
--# end
--# property-start '[dispatcher]AdaNode.xref_stop_resolution' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Ada_Node_P_Xref_Stop_Resolution
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Ada_Node (Self.Kind) is
               when Ada_Abort_Absent .. Ada_Record_Def | Ada_Composite_Constraint_Assoc | Ada_Param_Assoc .. Ada_Elsif_Stmt_Part | Ada_Handled_Stmts .. Ada_Quantifier_Some | Ada_Renaming_Clause .. Ada_With_Private_Present =>
                  --# property-call-start AdaNode.xref_stop_resolution
                  Property_Result := Ada_Node_P_Xref_Stop_Resolution
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Aggregate_Assoc .. Ada_Multi_Dim_Array_Assoc =>
                  --# property-call-start AggregateAssoc.xref_stop_resolution
                  Property_Result := Aggregate_Assoc_P_Xref_Stop_Resolution
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Iterated_Assoc =>
                  --# property-call-start IteratedAssoc.xref_stop_resolution
                  Property_Result := Iterated_Assoc_P_Xref_Stop_Resolution
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Abstract_State_Decl_Expr .. Ada_Update_Attribute_Ref | Ada_Defining_Name .. Ada_Un_Op =>
                  --# property-call-start Expr.xref_stop_resolution
                  Property_Result := Expr_P_Xref_Stop_Resolution
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Call_Expr =>
                  --# property-call-start CallExpr.xref_stop_resolution
                  Property_Result := Call_Expr_P_Xref_Stop_Resolution
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Range_Spec =>
                  --# property-call-start RangeSpec.xref_stop_resolution
                  Property_Result := Range_Spec_P_Xref_Stop_Resolution
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Ada_Node_P_Xref_Stop_Resolution;
--# end
--# property-start '[dispatcher]AdaNode.stop_resolution_equation' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Ada_Node_P_Stop_Resolution_Equation
  (Node : Bare_Ada_Node
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Ada_Node (Self.Kind) is
               when Ada_Abort_Absent .. Ada_Update_Attribute_Ref | Ada_Defining_Name .. Ada_With_Private_Present =>
                  --# property-call-start AdaNode.stop_resolution_equation
                  Property_Result := Ada_Node_P_Stop_Resolution_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Call_Expr =>
                  --# property-call-start CallExpr.stop_resolution_equation
                  Property_Result := Call_Expr_P_Stop_Resolution_Equation
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Ada_Node_P_Stop_Resolution_Equation;
--# end
--# property-start '[dispatcher]AdaNode.xref_initial_env' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Ada_Node_P_Xref_Initial_Env
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity :=
     Internal_Entity'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Ada_Node (Self.Kind) is
               when Ada_Abort_Absent .. Ada_Pp_Then_Kw | Ada_Private_Absent .. Ada_With_Private_Present =>
                  --# property-call-start AdaNode.xref_initial_env
                  Property_Result := Ada_Node_P_Xref_Initial_Env
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Pragma_Node =>
                  --# property-call-start Pragma.xref_initial_env
                  Property_Result := Pragma_Node_P_Xref_Initial_Env
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Ada_Node_P_Xref_Initial_Env;
--# end
--# property-start AdaNode.sub_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:630
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Sub_Equation
  (Node : Bare_Ada_Node
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity :=
     Internal_Entity'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_40 with Inline_Always;
      Fld : Logic_Equation;
Fld_1 : Logic_Equation;
Fld_2 : Boolean;
If_Result : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_40 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_1);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_40;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 305 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 304 '<If at ast.py:638>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:638
--# expr-start 301 '<FieldAccess for xref_stop_resolution at ast.py:638>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:638
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_stop_resolution'
Fld_2 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Stop_Resolution (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 301
if Fld_2 then
   --# expr-start 302 '<FieldAccess for stop_resolution_equation at ast.py:639>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:639
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.stop_resolution_equation'
Fld := Libadalang.Implementation.Dispatcher_Ada_Node_P_Stop_Resolution_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 302
   If_Result := Fld;
else
   --# expr-start 303 '<FieldAccess for xref_equation at ast.py:640>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:640
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_1 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 303
   If_Result := Fld_1;
end if;
      Inc_Ref (If_Result);
--# expr-done 304
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 305
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_40;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_40;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Sub_Equation;
--# end
--# property-start AdaNode.resolve_children_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:653
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Resolve_Children_Names
  (Node : Bare_Ada_Node
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Boolean;
            procedure Finalizer_Scope_42 with Inline_Always;
      Var_Expr : Bare_Ada_Node;
Node_Children : Bare_Ada_Node_Array_Access;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Fld : Boolean;
As_Entity_1 : Internal_Entity;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity;
Fld_1 : Boolean;
If_Result_2 : Boolean;
As_Entity_2 : Internal_Entity;
Is_Null_2 : Boolean;
If_Result_3 : Internal_Entity;
Fld_2 : Boolean;
If_Result_4 : Boolean;
Fld_3 : Boolean;
If_Result_5 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Quantifier_Result : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_42 is
            begin
                     Dec_Ref (Node_Children);
            end Finalizer_Scope_42;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 317 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 316 '<AllQuantifier at ast.py:658>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:658
   --# expr-start 306 '<children at ast.py:658>' Node_Children /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:658
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Children := Children (Self); 
--# expr-done 306
      Quantifier_Result := True;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Children;
      begin
         for Item_9 of
               Collection.Items
         loop
   --# scope-start
               --# bind c Item_9
Var_Expr := Item_9;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind c Var_Expr
--# expr-start 315 '<If at ast.py:661>' If_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:661
--# expr-start 307 '<FieldAccess for xref_entry_point at ast.py:662>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:662
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_entry_point'
Fld_3 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Entry_Point (Node => Var_Expr);
--# end
--# expr-done 307
if Fld_3 then
   If_Result_5 := True;
else
--# expr-start 312 '<If at ast.py:664>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:664
--# expr-start 309 '<FieldAccess for xref_stop_resolution at ast.py:665>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:665
--# expr-start 308 '<as_entity at ast.py:665>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:665
Is_Null_1 := Var_Expr = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity;
else
As_Entity_1 := (Info => E_Info, Node => Var_Expr); 
   If_Result_1 := As_Entity_1;
end if;
--# expr-done 308
      if If_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_stop_resolution'
Fld_1 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Stop_Resolution (Node => If_Result_1.Node, E_Info => If_Result_1.Info);
--# end
--# expr-done 309
if Fld_1 then
   --# expr-start 311 '<FieldAccess for resolve_own_names at ast.py:666>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:666
--# expr-start 310 '<as_entity at ast.py:666>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:666
Is_Null := Var_Expr = null; 
if Is_Null then
   If_Result := No_Entity;
else
As_Entity := (Info => E_Info, Node => Var_Expr); 
   If_Result := As_Entity;
end if;
--# expr-done 310
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Ada_Node_P_Resolve_Own_Names (Node => If_Result.Node, Env => Env, Origin => Origin, E_Info => If_Result.Info);
--# expr-done 311
   If_Result_2 := Fld;
else
   If_Result_2 := True;
end if;
--# expr-done 312
if If_Result_2 then
   --# expr-start 314 '<FieldAccess for resolve_children_names at ast.py:668>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:668
--# expr-start 313 '<as_entity at ast.py:668>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:668
Is_Null_2 := Var_Expr = null; 
if Is_Null_2 then
   If_Result_3 := No_Entity;
else
As_Entity_2 := (Info => E_Info, Node => Var_Expr); 
   If_Result_3 := As_Entity_2;
end if;
--# expr-done 313
      if If_Result_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.resolve_children_names
Fld_2 := Libadalang.Implementation.Ada_Node_P_Resolve_Children_Names (Node => If_Result_3.Node, Env => Env, Origin => Origin, E_Info => If_Result_3.Info);
--# end
--# expr-done 314
   If_Result_4 := Fld_2;
else
   If_Result_4 := False;
end if;
   If_Result_5 := If_Result_4;
end if;
--# expr-done 315
Scope_Result := If_Result_5;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := True;
end if;
      Quantifier_Result := Result_Var;
   --# end
               exit when not Quantifier_Result;
         end loop;
      end;
--# expr-done 316
Let_Result := Quantifier_Result; 
--# end
--# expr-done 317
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_42;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_42;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Resolve_Children_Names;
--# end
--# property-start AdaNode.resolve_names_internal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:674
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Resolve_Names_Internal
  (Node : Bare_Ada_Node
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity :=
     Internal_Entity'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Boolean;
Fld_1 : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 320 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 318 '<FieldAccess for resolve_own_names at ast.py:678>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:678
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Ada_Node_P_Resolve_Own_Names (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# expr-done 318
if Fld then
   --# expr-start 319 '<FieldAccess for resolve_children_names at ast.py:678>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:678
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.resolve_children_names
Fld_1 := Libadalang.Implementation.Ada_Node_P_Resolve_Children_Names (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 319
   If_Result := Fld_1;
else
   If_Result := False;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 320
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Resolve_Names_Internal;
--# end
--# property-start AdaNode.resolve_names_internal_with_eq /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:681
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Resolve_Names_Internal_With_Eq
  (Node : Bare_Ada_Node
      ; Additional_Equation : Logic_Equation
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity :=
     Internal_Entity'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind additional_equation Additional_Equation
   --# bind env Env
   --# bind origin Origin
   Property_Result : Boolean;
            procedure Finalizer_Scope_44 with Inline_Always;
      Fld : Logic_Equation;
And_Pred : Logic_Equation;
Solve_Success : Boolean;
Fld_1 : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_44 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (And_Pred);
            end Finalizer_Scope_44;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 325 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 323 '<solve at ast.py:686>' Solve_Success /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:686
--# expr-start 322 '<And at ast.py:686>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:686
--# expr-start 321 '<FieldAccess for xref_equation at ast.py:686>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:686
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 321
And_Pred := Create_And (Fld, Additional_Equation, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:686") else null)); 
--# expr-done 322
Solve_Success := Solve_Wrapper (And_Pred, Self); 
--# expr-done 323
if Solve_Success then
   --# expr-start 324 '<FieldAccess for resolve_children_names at ast.py:687>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:687
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.resolve_children_names
Fld_1 := Libadalang.Implementation.Ada_Node_P_Resolve_Children_Names (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 324
   If_Result := Fld_1;
else
   If_Result := False;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 325
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_44;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_44;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Resolve_Names_Internal_With_Eq;
--# end
--# property-start '[dispatcher]AdaNode.xref_entry_point' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Ada_Node_P_Xref_Entry_Point
  (Node : Bare_Ada_Node
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Ada_Node (Self.Kind) is
               when Ada_Abort_Absent .. Ada_Unconstrained_Array_Indices | Ada_Aspect_Spec .. Ada_Pragma_Argument_Assoc | Ada_Component_List .. Ada_Iterated_Assoc | Ada_Abstract_State_Decl .. Ada_Anonymous_Expr_Decl | Ada_Synthetic_Formal_Param_Decl .. Ada_Discrete_Base_Subtype_Decl | Ada_Classwide_Type_Decl .. Ada_Incomplete_Tagged_Type_Decl | Ada_Abstract_Subp_Decl .. Ada_Synthetic_Subp_Decl | Ada_Null_Subp_Decl .. Ada_Subp_Body | Ada_Package_Body_Stub .. Ada_Task_Body_Stub | Ada_Package_Body .. Ada_Task_Body | Ada_Error_Decl | Ada_For_Loop_Var_Decl .. Ada_Generic_Subp_Decl | Ada_Label_Decl .. Ada_Named_Stmt_Decl | Ada_Single_Task_Decl | Ada_Compilation_Unit | Ada_Component_Def .. Ada_Elsif_Expr_Part | Ada_Abstract_State_Decl_Expr .. Ada_Call_Expr | Ada_Discrete_Subtype_Name .. Ada_Dotted_Name | Ada_Explicit_Deref .. Ada_Limited_Present | Ada_While_Loop_Spec .. Ada_Pp_Then_Kw | Ada_Private_Absent .. Ada_Select_When_Part | Ada_Subp_Kind_Function .. Ada_Subp_Kind_Procedure | Ada_Synchronized_Absent .. Ada_Type_Access_Def | Ada_Derived_Type_Def .. Ada_Until_Present | Ada_Value_Sequence .. Ada_Variant | Ada_With_Private_Absent .. Ada_With_Private_Present =>
                  --# property-call-start AdaNode.xref_entry_point
                  Property_Result := Ada_Node_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Aspect_Assoc =>
                  --# property-call-start AspectAssoc.xref_entry_point
                  Property_Result := Aspect_Assoc_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_At_Clause .. Ada_Record_Rep_Clause =>
                  --# property-call-start AspectClause.xref_entry_point
                  Property_Result := Aspect_Clause_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Entry_Spec .. Ada_Subp_Spec =>
                  --# property-call-start BaseSubpSpec.xref_entry_point
                  Property_Result := Base_Subp_Spec_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Param_Assoc =>
                  --# property-call-start ParamAssoc.xref_entry_point
                  Property_Result := Param_Assoc_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Component_Decl =>
                  --# property-call-start ComponentDecl.xref_entry_point
                  Property_Result := Component_Decl_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Discriminant_Spec =>
                  --# property-call-start DiscriminantSpec.xref_entry_point
                  Property_Result := Discriminant_Spec_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Generic_Formal_Obj_Decl .. Ada_Generic_Formal_Type_Decl =>
                  --# property-call-start GenericFormal.xref_entry_point
                  Property_Result := Generic_Formal_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Param_Spec =>
                  --# property-call-start ParamSpec.xref_entry_point
                  Property_Result := Param_Spec_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Subtype_Decl =>
                  --# property-call-start SubtypeDecl.xref_entry_point
                  Property_Result := Subtype_Decl_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Protected_Type_Decl =>
                  --# property-call-start ProtectedTypeDecl.xref_entry_point
                  Property_Result := Protected_Type_Decl_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Task_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start TaskTypeDecl.xref_entry_point
                  Property_Result := Task_Type_Decl_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Type_Decl =>
                  --# property-call-start TypeDecl.xref_entry_point
                  Property_Result := Type_Decl_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Anonymous_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start AnonymousTypeDecl.xref_entry_point
                  Property_Result := Anonymous_Type_Decl_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Expr_Function =>
                  --# property-call-start ExprFunction.xref_entry_point
                  Property_Result := Expr_Function_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Subp_Renaming_Decl =>
                  --# property-call-start SubpRenamingDecl.xref_entry_point
                  Property_Result := Subp_Renaming_Decl_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Entry_Body =>
                  --# property-call-start EntryBody.xref_entry_point
                  Property_Result := Entry_Body_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Entry_Index_Spec =>
                  --# property-call-start EntryIndexSpec.xref_entry_point
                  Property_Result := Entry_Index_Spec_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Exception_Decl =>
                  --# property-call-start ExceptionDecl.xref_entry_point
                  Property_Result := Exception_Decl_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Exception_Handler =>
                  --# property-call-start ExceptionHandler.xref_entry_point
                  Property_Result := Exception_Handler_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Generic_Package_Instantiation .. Ada_Generic_Subp_Instantiation =>
                  --# property-call-start GenericInstantiation.xref_entry_point
                  Property_Result := Generic_Instantiation_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Generic_Package_Renaming_Decl .. Ada_Generic_Subp_Renaming_Decl =>
                  --# property-call-start GenericRenamingDecl.xref_entry_point
                  Property_Result := Generic_Renaming_Decl_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Number_Decl =>
                  --# property-call-start NumberDecl.xref_entry_point
                  Property_Result := Number_Decl_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Object_Decl .. Ada_No_Type_Object_Renaming_Decl =>
                  --# property-call-start ObjectDecl.xref_entry_point
                  Property_Result := Object_Decl_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Package_Renaming_Decl =>
                  --# property-call-start PackageRenamingDecl.xref_entry_point
                  Property_Result := Package_Renaming_Decl_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Single_Protected_Decl =>
                  --# property-call-start SingleProtectedDecl.xref_entry_point
                  Property_Result := Single_Protected_Decl_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Case_Stmt_Alternative =>
                  --# property-call-start CaseStmtAlternative.xref_entry_point
                  Property_Result := Case_Stmt_Alternative_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Component_Clause =>
                  --# property-call-start ComponentClause.xref_entry_point
                  Property_Result := Component_Clause_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Elsif_Stmt_Part =>
                  --# property-call-start ElsifStmtPart.xref_entry_point
                  Property_Result := Elsif_Stmt_Part_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Defining_Name .. Ada_Synthetic_Defining_Name =>
                  --# property-call-start DefiningName.xref_entry_point
                  Property_Result := Defining_Name_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_End_Name =>
                  --# property-call-start EndName.xref_entry_point
                  Property_Result := End_Name_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_For_Loop_Spec =>
                  --# property-call-start ForLoopSpec.xref_entry_point
                  Property_Result := For_Loop_Spec_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Pragma_Node =>
                  --# property-call-start Pragma.xref_entry_point
                  Property_Result := Pragma_Node_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Accept_Stmt .. Ada_Terminate_Alternative =>
                  --# property-call-start Stmt.xref_entry_point
                  Property_Result := Stmt_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Subunit =>
                  --# property-call-start Subunit.xref_entry_point
                  Property_Result := Subunit_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Array_Type_Def =>
                  --# property-call-start ArrayTypeDef.xref_entry_point
                  Property_Result := Array_Type_Def_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Use_Package_Clause .. Ada_Use_Type_Clause =>
                  --# property-call-start UseClause.xref_entry_point
                  Property_Result := Use_Clause_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_Variant_Part =>
                  --# property-call-start VariantPart.xref_entry_point
                  Property_Result := Variant_Part_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
               when Ada_With_Clause =>
                  --# property-call-start WithClause.xref_entry_point
                  Property_Result := With_Clause_P_Xref_Entry_Point
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Ada_Node_P_Xref_Entry_Point;
--# end
--# property-start AdaNode.resolve_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:706
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Resolve_Names
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity :=
     Internal_Entity'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
            procedure Finalizer_Scope_46 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_1 : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_2 : Boolean;
Dyn_Var_Bind_Result : Boolean;
Dyn_Var_Bind_Result_1 : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_46 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Bound_Env);
            end Finalizer_Scope_46;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Ada_Node_Ada_Node_P_Resolve_Names,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Ada_Node,
                                As_Bare_Ada_Node => Self);
               Mmz_K.Items (2) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Boolean;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 331 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 330 '<bind at ast.py:712>' Dyn_Var_Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:712
--# expr-start 326 '<FieldAccess for xref_initial_env at ast.py:713>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:713
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_initial_env'
Fld := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Initial_Env (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 326
Bound_Env := Fld; Inc_Ref (Bound_Env);
--# expr-start 329 '<bind at ast.py:714>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:714
--# expr-start 327 '<FieldAccess for origin_node at ast.py:714>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:714
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_1 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 327
Bound_Origin := Fld_1; 
--# expr-start 328 '<FieldAccess for resolve_names_internal at ast.py:714>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:714
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.resolve_names_internal
Fld_2 := Libadalang.Implementation.Ada_Node_P_Resolve_Names_Internal (Node => Ent.Node, Env => Bound_Env, Origin => Bound_Origin, E_Info => Ent.Info);
--# end
--# expr-done 328
Dyn_Var_Bind_Result := Fld_2; 
--# expr-done 329
Dyn_Var_Bind_Result_1 := Dyn_Var_Bind_Result; 
--# expr-done 330
Let_Result := Dyn_Var_Bind_Result_1; 
--# end
--# expr-done 331
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_46;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Boolean,
                        As_Boolean => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_46;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Resolve_Names;
--# end
--# property-start AdaNode.resolve_names_from_closest_entry_point_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:718
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Resolve_Names_From_Closest_Entry_Point_Impl
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity :=
     Internal_Entity'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_47 with Inline_Always;
            procedure Finalizer_Scope_1727 with Inline_Always;
      Res : Lexical_Env := Empty_Env;
Var_Expr : Internal_Entity;
Fld : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_1 : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_2 : Boolean;
If_Result : Lexical_Env := Empty_Env;
Dyn_Var_Bind_Result : Lexical_Env := Empty_Env;
Dyn_Var_Bind_Result_1 : Lexical_Env := Empty_Env;
Fld_3 : Internal_Entity;
Fld_4 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Is_Equal : Boolean;
Fld_5 : Boolean;
Bound_Env_1 : Lexical_Env := Empty_Env;
Fld_6 : Bare_Ada_Node;
Bound_Origin_1 : Bare_Ada_Node;
Fld_7 : Boolean;
If_Result_1 : Lexical_Env := Empty_Env;
Dyn_Var_Bind_Result_2 : Lexical_Env := Empty_Env;
Dyn_Var_Bind_Result_3 : Lexical_Env := Empty_Env;
If_Result_2 : Lexical_Env := Empty_Env;
If_Result_3 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
Fld_8 : Boolean;
If_Result_4 : Lexical_Env := Empty_Env;
Let_Result_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_47 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (If_Result);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Dyn_Var_Bind_Result_1);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Res);
                     Dec_Ref (Bound_Env_1);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Dyn_Var_Bind_Result_2);
                     Dec_Ref (Dyn_Var_Bind_Result_3);
                     Dec_Ref (If_Result_2);
                     Dec_Ref (If_Result_3);
                     Dec_Ref (Let_Result);
                     Dec_Ref (If_Result_4);
                     Dec_Ref (Let_Result_1);
            end Finalizer_Scope_47;
            procedure Finalizer_Scope_1727 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_1727;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 351 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 350 '<If at ast.py:728>' If_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:728
--# expr-start 332 '<FieldAccess for xref_entry_point at ast.py:732>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:732
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_entry_point'
Fld_8 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Entry_Point (Node => Ent.Node);
--# end
--# expr-done 332
if Fld_8 then
   --# expr-start 338 '<bind at ast.py:733>' Dyn_Var_Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:733
--# expr-start 333 '<FieldAccess for xref_initial_env at ast.py:734>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:734
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_initial_env'
Fld := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Initial_Env (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 333
Bound_Env := Fld; Inc_Ref (Bound_Env);
--# expr-start 337 '<bind at ast.py:735>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:735
--# expr-start 334 '<FieldAccess for origin_node at ast.py:736>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:736
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_1 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 334
Bound_Origin := Fld_1; 
--# expr-start 336 '<If at ast.py:737>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:737
--# expr-start 335 '<FieldAccess for resolve_own_names at ast.py:737>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:737
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Extensions.Ada_Node_P_Resolve_Own_Names (Node => Ent.Node, Env => Bound_Env, Origin => Bound_Origin, E_Info => Ent.Info);
--# expr-done 335
if Fld_2 then
   If_Result := Bound_Env;
else
   If_Result := Empty_Env;
end if;
      Inc_Ref (If_Result);
--# expr-done 336
Dyn_Var_Bind_Result := If_Result; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 337
Dyn_Var_Bind_Result_1 := Dyn_Var_Bind_Result; Inc_Ref (Dyn_Var_Bind_Result_1);
--# expr-done 338
   If_Result_4 := Dyn_Var_Bind_Result_1;
else
   --# expr-start 349 '<Let at ast.py:741>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:741
--# scope-start
--# expr-start 339 '<parent at ast.py:746>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:746
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 339
Var_Expr := Fld_3;
if Var_Expr /= No_Entity then
   --# scope-start
--# expr-start 340 '<FieldAccess for resolve_names_from_closest_entry_point_impl at ast.py:746>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:746
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.resolve_names_from_closest_entry_point_impl
Fld_4 := Libadalang.Implementation.Ada_Node_P_Resolve_Names_From_Closest_Entry_Point_Impl (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 340
Scope_Result := Fld_4;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1727;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
Res := Result_Var; Inc_Ref (Res);
--# bind res Res
--# expr-start 348 '<Cond at ast.py:749>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:749
--# expr-start 341 '<Eq at ast.py:751>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:751
Is_Equal := Equivalent (Res, Empty_Env); 
--# expr-done 341
if Is_Equal then
   If_Result_3 := Res;
else
--# expr-start 342 '<FieldAccess for xref_stop_resolution at ast.py:757>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:757
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_stop_resolution'
Fld_5 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Stop_Resolution (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 342
if Fld_5 then
   --# expr-start 347 '<bind at ast.py:758>' Dyn_Var_Bind_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:758
Bound_Env_1 := Res; Inc_Ref (Bound_Env_1);
--# expr-start 346 '<bind at ast.py:760>' Dyn_Var_Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:760
--# expr-start 343 '<FieldAccess for origin_node at ast.py:761>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:761
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_6 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 343
Bound_Origin_1 := Fld_6; 
--# expr-start 345 '<If at ast.py:762>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:762
--# expr-start 344 '<FieldAccess for resolve_own_names at ast.py:762>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:762
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Extensions.Ada_Node_P_Resolve_Own_Names (Node => Ent.Node, Env => Bound_Env_1, Origin => Bound_Origin_1, E_Info => Ent.Info);
--# expr-done 344
if Fld_7 then
   If_Result_1 := Res;
else
   If_Result_1 := Empty_Env;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 345
Dyn_Var_Bind_Result_2 := If_Result_1; Inc_Ref (Dyn_Var_Bind_Result_2);
--# expr-done 346
Dyn_Var_Bind_Result_3 := Dyn_Var_Bind_Result_2; Inc_Ref (Dyn_Var_Bind_Result_3);
--# expr-done 347
   If_Result_2 := Dyn_Var_Bind_Result_3;
else
   If_Result_2 := Res;
end if;
      Inc_Ref (If_Result_2);
   If_Result_3 := If_Result_2;
end if;
      Inc_Ref (If_Result_3);
--# expr-done 348
Let_Result := If_Result_3; Inc_Ref (Let_Result);
--# end
--# expr-done 349
   If_Result_4 := Let_Result;
end if;
      Inc_Ref (If_Result_4);
--# expr-done 350
Let_Result_1 := If_Result_4; Inc_Ref (Let_Result_1);
--# end
--# expr-done 351
      Property_Result := Let_Result_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_47;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_47;
               Finalizer_Scope_1727;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Resolve_Names_From_Closest_Entry_Point_Impl;
--# end
--# property-start AdaNode.resolve_names_from_closest_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:775
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Resolve_Names_From_Closest_Entry_Point
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity :=
     Internal_Entity'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
            procedure Finalizer_Scope_48 with Inline_Always;
      Block_Var : Lexical_Env := Empty_Env;
Fld : Lexical_Env := Empty_Env;
Is_Equal : Boolean;
Not_Val : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_48 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Block_Var);
            end Finalizer_Scope_48;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 355 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 352 '<FieldAccess for resolve_names_from_closest_entry_point_impl at ast.py:797>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:797
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.resolve_names_from_closest_entry_point_impl
Fld := Libadalang.Implementation.Ada_Node_P_Resolve_Names_From_Closest_Entry_Point_Impl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 352
Block_Var := Fld; Inc_Ref (Block_Var);
--# bind result Block_Var
--# expr-start 354 '<Not at ast.py:798>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:798
--# expr-start 353 '<Eq at ast.py:798>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:798
Is_Equal := Equivalent (Block_Var, Empty_Env); 
--# expr-done 353
Not_Val := not (Is_Equal); 
--# expr-done 354
Let_Result := Not_Val; 
--# end
--# expr-done 355
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_48;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_48;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Resolve_Names_From_Closest_Entry_Point;
--# end
--# property-start AdaNode.parent_unit_env_helper /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:801
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Parent_Unit_Env_Helper
  (Node : Bare_Ada_Node
      ; Unit : Internal_Unit
      ; Env : Lexical_Env
  )
   return Lexical_Env
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind unit Unit
   --# bind env Env
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_49 with Inline_Always;
            procedure Finalizer_Scope_1728 with Inline_Always;
            procedure Finalizer_Scope_1729 with Inline_Always;
      Var_Expr : Lexical_Env := Empty_Env;
Var_Expr_1 : Bare_Ada_Node;
Env_Parent : Lexical_Env := Empty_Env;
Env_Node : Bare_Ada_Node;
Fld : Lexical_Env := Empty_Env;
Fld_1 : Internal_Unit;
Is_Equal : Boolean;
If_Result : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Scope_Result_1 : Lexical_Env := Empty_Env;
Result_Var_1 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_49 is
            begin
                     Dec_Ref (Env_Parent);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_49;
            procedure Finalizer_Scope_1728 is
            begin
                     Dec_Ref (Var_Expr);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1728;
            procedure Finalizer_Scope_1729 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (If_Result);
            end Finalizer_Scope_1729;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 362 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 356 '<env_parent at ast.py:802>' Env_Parent /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:802
Env_Parent := AST_Envs.Parent (Env); 
--# expr-done 356
Var_Expr := Env_Parent;
      Inc_Ref (Var_Expr);
if Var_Expr /= Empty_Env then
   --# scope-start
   --# bind parent_env Var_Expr
--# expr-start 357 '<env_node at ast.py:802>' Env_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:802
Env_Node := AST_Envs.Env_Node (Var_Expr); 
--# expr-done 357
Var_Expr_1 := Env_Node;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind parent_node Var_Expr_1
--# expr-start 361 '<If at ast.py:803>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:803
--# expr-start 359 '<Eq at ast.py:804>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:804
--# expr-start 358 '<FieldAccess for unit at ast.py:804>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:804
      if Var_Expr_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Unit (Node => Var_Expr_1);
--# expr-done 358
Is_Equal := Fld_1 = Unit; 
--# expr-done 359
if Is_Equal then
   --# expr-start 360 '<FieldAccess for parent_unit_env_helper at ast.py:805>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:805
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.parent_unit_env_helper
Fld := Libadalang.Implementation.Ada_Node_P_Parent_Unit_Env_Helper (Node => Self, Unit => Unit, Env => Var_Expr);
--# end
--# expr-done 360
   If_Result := Fld;
else
   If_Result := Var_Expr;
end if;
      Inc_Ref (If_Result);
--# expr-done 361
Scope_Result := If_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1729;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
Scope_Result_1 := Result_Var;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_1728;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := Empty_Env;
end if;
      Inc_Ref (Result_Var_1);
Let_Result := Result_Var_1; Inc_Ref (Let_Result);
--# end
--# expr-done 362
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_49;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_49;
               Finalizer_Scope_1728;
               Finalizer_Scope_1729;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Parent_Unit_Env_Helper;
--# end
--# property-start AdaNode.parent_unit_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:811
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Parent_Unit_Env
  (Node : Bare_Ada_Node
      ; Env : Lexical_Env
  )
   return Lexical_Env
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_50 with Inline_Always;
            procedure Finalizer_Scope_1730 with Inline_Always;
      Var_Expr : Lexical_Env := Empty_Env;
Env_Node : Bare_Ada_Node;
Fld : Internal_Unit;
Fld_1 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_50 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_50;
            procedure Finalizer_Scope_1730 is
            begin
                     Dec_Ref (Var_Expr);
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1730;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 366 '<Block at ???>' Let_Result None
--# scope-start
Var_Expr := Env;
      Inc_Ref (Var_Expr);
if Var_Expr /= Empty_Env then
   --# scope-start
   --# bind env Var_Expr
--# expr-start 365 '<FieldAccess for parent_unit_env_helper at ast.py:817>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:817
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 364 '<FieldAccess for unit at ast.py:817>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:817
--# expr-start 363 '<env_node at ast.py:817>' Env_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:817
Env_Node := AST_Envs.Env_Node (Var_Expr); 
--# expr-done 363
      if Env_Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Unit (Node => Env_Node);
--# expr-done 364
--# property-call-start AdaNode.parent_unit_env_helper
Fld_1 := Libadalang.Implementation.Ada_Node_P_Parent_Unit_Env_Helper (Node => Self, Unit => Fld, Env => Var_Expr);
--# end
--# expr-done 365
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1730;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 366
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_50;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_50;
               Finalizer_Scope_1730;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Parent_Unit_Env;
--# end
--# property-start AdaNode.std /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:829
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Std
  (Node : Bare_Ada_Node
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Internal_Unit;
Fld_1 : Bare_Ada_Node;
Cast_Result : Bare_Compilation_Unit;
Fld_2 : Bare_Ada_Node;
Cast_Result_1 : Bare_Library_Item;
Fld_3 : Bare_Basic_Decl;
As_Entity : Internal_Entity_Basic_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 373 '<as_bare_entity at ast.py:830>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:830
--# expr-start 372 '<FieldAccess for item at ast.py:830>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:830
--# expr-start 371 '<Cast to LibraryItem at ast.py:830>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:830
--# expr-start 370 '<FieldAccess for body at ast.py:830>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:830
--# expr-start 369 '<Cast to CompilationUnit at ast.py:830>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:830
--# expr-start 368 '<FieldAccess for root at ast.py:830>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:830
--# expr-start 367 '<FieldAccess for standard_unit at ast.py:830>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:830
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Ada_Node_P_Standard_Unit (Node => Self);
--# expr-done 367
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Ast_Root;
--# expr-done 368
   if Fld_1 = null
      or else Fld_1.Kind in Ada_Compilation_Unit_Range
   then
      Cast_Result := Fld_1;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 369
      if Cast_Result = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Cast_Result.Compilation_Unit_F_Body;
--# expr-done 370
   if Fld_2 = null
      or else Fld_2.Kind in Ada_Library_Item_Range
   then
      Cast_Result_1 := Fld_2;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 371
      if Cast_Result_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Cast_Result_1.Library_Item_F_Item;
--# expr-done 372
Is_Null := Fld_3 = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Decl;
else
As_Entity := (Info => No_Entity_Info, Node => Fld_3); 
   If_Result := As_Entity;
end if;
--# expr-done 373
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Std;
--# end
--# property-start AdaNode.std_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:838
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Std_Env
  (Node : Bare_Ada_Node
  )
   return Lexical_Env
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_53 with Inline_Always;
      Fld : Internal_Entity_Basic_Decl;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_53 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_53;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 375 '<FieldAccess for children_env at ast.py:839>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:839
--# expr-start 374 '<FieldAccess for std at ast.py:839>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:839
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.std
Fld := Libadalang.Implementation.Ada_Node_P_Std (Node => Self);
--# end
--# expr-done 374
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Children_Env (Node => Fld.Node, E_Info => Fld.Info);
--# expr-done 375
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_53;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_53;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Std_Env;
--# end
--# property-start AdaNode.std_entity /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:843
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Std_Entity
  (Node : Bare_Ada_Node
      ; Sym : Symbol_Type
  )
   return Internal_Entity
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind sym Sym
   Property_Result : Internal_Entity;
      Fld : Internal_Unit;
Fld_1 : Bare_Ada_Node;
Fld_2 : Internal_Entity;
Let_Result : Internal_Entity;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 379 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 378 '<FieldAccess for std_entity_implem at ast.py:844>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:844
--# expr-start 377 '<FieldAccess for root at ast.py:844>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:844
--# expr-start 376 '<FieldAccess for unit at ast.py:844>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:844
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Unit (Node => Self);
--# expr-done 376
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Ast_Root;
--# expr-done 377
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.std_entity_implem
Fld_2 := Libadalang.Implementation.Ada_Node_P_Std_Entity_Implem (Node => Fld_1, Sym => Sym);
--# end
--# expr-done 378
Let_Result := Fld_2; 
--# end
--# expr-done 379
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Std_Entity;
--# end
--# property-start AdaNode.std_entity_implem /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:850
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Std_Entity_Implem
  (Node : Bare_Ada_Node
      ; Sym : Symbol_Type
  )
   return Internal_Entity
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind sym Sym
   Property_Result : Internal_Entity;
            procedure Finalizer_Scope_55 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Env_Get_Result : Internal_Entity;
Let_Result : Internal_Entity;
            procedure Finalizer_Scope_55 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_55;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Ada_Node_Ada_Node_P_Std_Entity_Implem,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Ada_Node,
                                As_Bare_Ada_Node => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Symbol_Type,
                                      As_Symbol_Type => Sym);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 382 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 381 '<EnvGet at ast.py:851>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:851
--# expr-start 380 '<FieldAccess for std_env at ast.py:851>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:851
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.std_env
Fld := Libadalang.Implementation.Ada_Node_P_Std_Env (Node => Self);
--# end
--# expr-done 380
Env_Get_Result := AST_Envs.Get_First (Self => Fld, Key => Sym, Lookup_Kind => To_Lookup_Kind_Type (Recursive), Categories => (Inherited_Primitives => False, Nocat => True)); 
--# expr-done 381
Let_Result := Env_Get_Result; 
--# end
--# expr-done 382
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_55;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity,
                        As_Internal_Entity => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_55;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Std_Entity_Implem;
--# end
--# property-start AdaNode.bool_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:855
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Bool_Type
  (Node : Bare_Ada_Node
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Type_Decl;
      Sym : Symbol_Type;
Fld : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 385 '<Cast to BaseTypeDecl at ast.py:856>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:856
--# expr-start 384 '<FieldAccess for std_entity at ast.py:856>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:856
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 383 '<Symbol Boolean at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Boolean); 
--# expr-done 383
--# property-call-start AdaNode.std_entity
Fld := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym);
--# end
--# expr-done 384
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 385
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Bool_Type;
--# end
--# property-start AdaNode.int_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:860
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Int_Type
  (Node : Bare_Ada_Node
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Type_Decl;
      Sym : Symbol_Type;
Fld : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 388 '<Cast to BaseTypeDecl at ast.py:861>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:861
--# expr-start 387 '<FieldAccess for std_entity at ast.py:861>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:861
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 386 '<Symbol Integer at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Integer); 
--# expr-done 386
--# property-call-start AdaNode.std_entity
Fld := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym);
--# end
--# expr-done 387
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 388
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Int_Type;
--# end
--# property-start AdaNode.universal_int_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:865
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Universal_Int_Type
  (Node : Bare_Ada_Node
  )
   return Internal_Entity
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity;
      Sym : Symbol_Type;
Fld : Internal_Entity;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 390 '<FieldAccess for std_entity at ast.py:866>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:866
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 389 '<Symbol Universal_Int_Type_ at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Universal_Int_Type); 
--# expr-done 389
--# property-call-start AdaNode.std_entity
Fld := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym);
--# end
--# expr-done 390
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Universal_Int_Type;
--# end
--# property-start AdaNode.universal_real_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:870
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Universal_Real_Type
  (Node : Bare_Ada_Node
  )
   return Internal_Entity
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity;
      Sym : Symbol_Type;
Fld : Internal_Entity;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 392 '<FieldAccess for std_entity at ast.py:871>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:871
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 391 '<Symbol Universal_Real_Type_ at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Universal_Real_Type); 
--# expr-done 391
--# property-call-start AdaNode.std_entity
Fld := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym);
--# end
--# expr-done 392
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Universal_Real_Type;
--# end
--# property-start AdaNode.std_char_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:875
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Std_Char_Type
  (Node : Bare_Ada_Node
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Type_Decl;
      Sym : Symbol_Type;
Fld : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 395 '<Cast to BaseTypeDecl at ast.py:876>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:876
--# expr-start 394 '<FieldAccess for std_entity at ast.py:876>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:876
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 393 '<Symbol Character at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Character); 
--# expr-done 393
--# property-call-start AdaNode.std_entity
Fld := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym);
--# end
--# expr-done 394
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 395
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Std_Char_Type;
--# end
--# property-start AdaNode.std_wide_char_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:880
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Std_Wide_Char_Type
  (Node : Bare_Ada_Node
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Type_Decl;
      Sym : Symbol_Type;
Fld : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 398 '<Cast to BaseTypeDecl at ast.py:881>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:881
--# expr-start 397 '<FieldAccess for std_entity at ast.py:881>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:881
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 396 '<Symbol Wide_Character at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Wide_Character); 
--# expr-done 396
--# property-call-start AdaNode.std_entity
Fld := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym);
--# end
--# expr-done 397
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 398
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Std_Wide_Char_Type;
--# end
--# property-start AdaNode.std_wide_wide_char_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:886
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Std_Wide_Wide_Char_Type
  (Node : Bare_Ada_Node
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Type_Decl;
      Sym : Symbol_Type;
Fld : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 401 '<Cast to BaseTypeDecl at ast.py:887>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:887
--# expr-start 400 '<FieldAccess for std_entity at ast.py:887>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:887
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 399 '<Symbol Wide_Wide_Character at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Wide_Wide_Character); 
--# expr-done 399
--# property-call-start AdaNode.std_entity
Fld := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym);
--# end
--# expr-done 400
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 401
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Std_Wide_Wide_Char_Type;
--# end
--# property-start AdaNode.std_root_types /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:893
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Std_Root_Types
  (Node : Bare_Ada_Node
  )
   return Lexical_Env
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_63 with Inline_Always;
            procedure Finalizer_Scope_1731 with Inline_Always;
      Var_Expr : Internal_Entity_Package_Decl;
Sym : Symbol_Type;
Fld : Internal_Entity;
Cast_Result : Internal_Entity_Package_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_63 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_63;
            procedure Finalizer_Scope_1731 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1731;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
--# expr-start 404 '<Cast to PackageDecl at ast.py:894>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:894
--# expr-start 403 '<FieldAccess for std_entity at ast.py:894>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:894
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 402 '<Symbol root_types_ at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Root_Types); 
--# expr-done 402
--# property-call-start AdaNode.std_entity
Fld := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym);
--# end
--# expr-done 403
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Package_Decl_Range
   then
      Cast_Result := Create_Internal_Entity_Package_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Package_Decl;
   end if;
--# expr-done 404
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Package_Decl then
   --# scope-start
--# expr-start 405 '<FieldAccess for children_env at ast.py:894>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:894
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Children_Env (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# expr-done 405
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1731;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_63;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_63;
               Finalizer_Scope_1731;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Std_Root_Types;
--# end
--# property-start AdaNode.root_int_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:900
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Root_Int_Type
  (Node : Bare_Ada_Node
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_64 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Sym : Symbol_Type;
Env_Get_Result : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_64 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_64;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 409 '<Cast to BaseTypeDecl at ast.py:901>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:901
--# expr-start 408 '<EnvGet at ast.py:901>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:901
--# expr-start 406 '<FieldAccess for std_root_types at ast.py:901>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:901
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.std_root_types
Fld := Libadalang.Implementation.Ada_Node_P_Std_Root_Types (Node => Self);
--# end
--# expr-done 406
--# expr-start 407 '<Symbol root_integer at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Root_Integer); 
--# expr-done 407
Env_Get_Result := AST_Envs.Get_First (Self => Fld, Key => Sym, Lookup_Kind => To_Lookup_Kind_Type (Minimal), Categories => (Inherited_Primitives => False, Nocat => True)); 
--# expr-done 408
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 409
      Property_Result := Cast_Result;
   --# end
      Finalizer_Scope_64;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_64;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Root_Int_Type;
--# end
--# property-start AdaNode.root_real_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:906
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Root_Real_Type
  (Node : Bare_Ada_Node
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_65 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Sym : Symbol_Type;
Env_Get_Result : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_65 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_65;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 413 '<Cast to BaseTypeDecl at ast.py:907>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:907
--# expr-start 412 '<EnvGet at ast.py:907>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:907
--# expr-start 410 '<FieldAccess for std_root_types at ast.py:907>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:907
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.std_root_types
Fld := Libadalang.Implementation.Ada_Node_P_Std_Root_Types (Node => Self);
--# end
--# expr-done 410
--# expr-start 411 '<Symbol root_real at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Root_Real); 
--# expr-done 411
Env_Get_Result := AST_Envs.Get_First (Self => Fld, Key => Sym, Lookup_Kind => To_Lookup_Kind_Type (Minimal), Categories => (Inherited_Primitives => False, Nocat => True)); 
--# expr-done 412
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 413
      Property_Result := Cast_Result;
   --# end
      Finalizer_Scope_65;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_65;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Root_Real_Type;
--# end
--# property-start AdaNode.root_type_ops_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:915
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Root_Type_Ops_Impl
  (Node : Bare_Ada_Node
      ; Sym : Symbol_Type
  )
   return Internal_Entity_Basic_Decl_Array_Access
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind sym Sym
   Property_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_66 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Env_Get_Result : Internal_Entity_Array_Access;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Cast_Expr_1 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Basic_Decl;
Fld_1 : Boolean;
Map_Result : Internal_Entity_Basic_Decl_Array_Access;
Let_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_66 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Env_Get_Result);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_66;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Ada_Node_Ada_Node_P_Root_Type_Ops_Impl,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Ada_Node,
                                As_Bare_Ada_Node => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Symbol_Type,
                                      As_Symbol_Type => Sym);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Basic_Decl_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
      end if;
   --# scope-start
      --# expr-start 420 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 419 '<FilterMap at ast.py:919>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:919
   --# expr-start 415 '<EnvGet at ast.py:919>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:919
--# expr-start 414 '<FieldAccess for std_root_types at ast.py:919>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:919
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.std_root_types
Fld := Libadalang.Implementation.Ada_Node_P_Std_Root_Types (Node => Self);
--# end
--# expr-done 414
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Fld, Key => Sym, Lookup_Kind => To_Lookup_Kind_Type (Minimal), Categories => (Inherited_Primitives => False, Nocat => True))); 
--# expr-done 415
      declare
         Map_Result_Vec : Internal_Entity_Basic_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Env_Get_Result;
      begin
         for Item_10 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_10
         --# expr-start 417 '<FieldAccess for is_subprogram at ast.py:923>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:923
--# expr-start 416 '<Cast to BasicDecl at ast.py:923>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:923
Cast_Expr_1 := Item_10; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 416
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_subprogram
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Is_Subprogram (Node => Cast_Result_1.Node);
--# end
--# expr-done 417
         if Fld_1 then
      --# expr-start 418 '<Cast to BasicDecl at ast.py:922>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:922
Cast_Expr := Item_10; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 418
         declare
            Item_To_Append : constant Internal_Entity_Basic_Decl := Cast_Result;
         begin
            Internal_Entity_Basic_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Basic_Decl_Array
           (Items_Count => Natural (Internal_Entity_Basic_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Basic_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Basic_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Basic_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 419
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 420
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_66;
      if Self /= null then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Basic_Decl_Array_Access,
                        As_Internal_Entity_Basic_Decl_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_66;
         if Self /= null then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Root_Type_Ops_Impl;
--# end
--# property-start AdaNode.root_type_ops /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:927
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Root_Type_Ops
  (Node : Bare_Ada_Node
      ; Sym : Symbol_Type
  )
   return Internal_Entity_Basic_Decl_Array_Access
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind sym Sym
   Property_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_67 with Inline_Always;
      Fld : Internal_Unit;
Fld_1 : Bare_Ada_Node;
Fld_2 : Internal_Entity_Basic_Decl_Array_Access;
Let_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_67 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_67;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 424 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 423 '<FieldAccess for root_type_ops_impl at ast.py:933>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:933
--# expr-start 422 '<FieldAccess for root at ast.py:933>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:933
--# expr-start 421 '<FieldAccess for unit at ast.py:933>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:933
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Unit (Node => Self);
--# expr-done 421
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Ast_Root;
--# expr-done 422
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.root_type_ops_impl
Fld_2 := Libadalang.Implementation.Ada_Node_P_Root_Type_Ops_Impl (Node => Fld_1, Sym => Sym);
--# end
--# expr-done 423
Let_Result := Fld_2; Inc_Ref (Let_Result);
--# end
--# expr-done 424
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_67;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_67;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Root_Type_Ops;
--# end
--# property-start AdaNode.exc_id_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:935
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Exc_Id_Type
  (Node : Bare_Ada_Node
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_68 with Inline_Always;
            procedure Finalizer_Scope_1733 with Inline_Always;
      Var_Expr : Bare_Basic_Decl;
Sym : Symbol_Type;
Sym_1 : Symbol_Type;
Array_Lit : Symbol_Type_Array_Access;
Fld : Bare_Basic_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Sym_2 : Symbol_Type;
Env_Get_Result : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_68 is
            begin
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_68;
            procedure Finalizer_Scope_1733 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1733;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 432 '<Cast to BaseTypeDecl at ast.py:936>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:936
--# expr-start 431 '<EnvGet at ast.py:936>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:936
--# expr-start 428 '<FieldAccess for get_unit_root_decl at ast.py:936>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:936
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 427 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 425 '<Symbol Ada at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Ada); 
--# expr-done 425
--# expr-start 426 '<Symbol Exceptions at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Exceptions); 
--# expr-done 426
Array_Lit := Create_Symbol_Type_Array (Internal_Symbol_Type_Array'(1 => Sym, 2 => Sym_1)); 
--# expr-done 427
--# property-call-start AdaNode.get_unit_root_decl
Fld := Libadalang.Implementation.Ada_Node_P_Get_Unit_Root_Decl (Node => Self, Name => Array_Lit, Kind => Unit_Specification);
--# end
--# expr-done 428
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 429 '<FieldAccess for children_env at ast.py:936>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:936
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Children_Env (Node => Var_Expr);
--# expr-done 429
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1733;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
--# expr-start 430 '<Symbol Exception_Id at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Exception_Id); 
--# expr-done 430
Env_Get_Result := AST_Envs.Get_First (Self => Result_Var, Key => Sym_2, Lookup_Kind => To_Lookup_Kind_Type (Flat), Categories => (Inherited_Primitives => True, Nocat => True)); 
--# expr-done 431
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 432
      Property_Result := Cast_Result;
   --# end
      Finalizer_Scope_68;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_68;
               Finalizer_Scope_1733;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Exc_Id_Type;
--# end
--# property-start AdaNode.task_id_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:945
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Task_Id_Type
  (Node : Bare_Ada_Node
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_69 with Inline_Always;
            procedure Finalizer_Scope_1734 with Inline_Always;
      Var_Expr : Bare_Basic_Decl;
Sym : Symbol_Type;
Sym_1 : Symbol_Type;
Array_Lit : Symbol_Type_Array_Access;
Fld : Bare_Basic_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Sym_2 : Symbol_Type;
Env_Get_Result : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_69 is
            begin
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_69;
            procedure Finalizer_Scope_1734 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1734;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 440 '<Cast to BaseTypeDecl at ast.py:946>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:946
--# expr-start 439 '<EnvGet at ast.py:946>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:946
--# expr-start 436 '<FieldAccess for get_unit_root_decl at ast.py:946>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:946
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 435 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 433 '<Symbol Ada at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Ada); 
--# expr-done 433
--# expr-start 434 '<Symbol Task_Identification at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Task_Identification); 
--# expr-done 434
Array_Lit := Create_Symbol_Type_Array (Internal_Symbol_Type_Array'(1 => Sym, 2 => Sym_1)); 
--# expr-done 435
--# property-call-start AdaNode.get_unit_root_decl
Fld := Libadalang.Implementation.Ada_Node_P_Get_Unit_Root_Decl (Node => Self, Name => Array_Lit, Kind => Unit_Specification);
--# end
--# expr-done 436
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 437 '<FieldAccess for children_env at ast.py:946>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:946
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Children_Env (Node => Var_Expr);
--# expr-done 437
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1734;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
--# expr-start 438 '<Symbol Task_Id at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Task_Id); 
--# expr-done 438
Env_Get_Result := AST_Envs.Get_First (Self => Result_Var, Key => Sym_2, Lookup_Kind => To_Lookup_Kind_Type (Flat), Categories => (Inherited_Primitives => True, Nocat => True)); 
--# expr-done 439
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 440
      Property_Result := Cast_Result;
   --# end
      Finalizer_Scope_69;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_69;
               Finalizer_Scope_1734;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Task_Id_Type;
--# end
--# property-start AdaNode.root_buffer_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:954
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Root_Buffer_Type
  (Node : Bare_Ada_Node
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_70 with Inline_Always;
            procedure Finalizer_Scope_1735 with Inline_Always;
      Var_Expr : Bare_Basic_Decl;
Sym : Symbol_Type;
Sym_1 : Symbol_Type;
Sym_2 : Symbol_Type;
Array_Lit : Symbol_Type_Array_Access;
Fld : Bare_Basic_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Sym_3 : Symbol_Type;
Env_Get_Result : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_70 is
            begin
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_70;
            procedure Finalizer_Scope_1735 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1735;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 449 '<Cast to BaseTypeDecl at ast.py:955>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:955
--# expr-start 448 '<EnvGet at ast.py:955>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:955
--# expr-start 445 '<FieldAccess for get_unit_root_decl at ast.py:955>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:955
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 444 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 441 '<Symbol Ada at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Ada); 
--# expr-done 441
--# expr-start 442 '<Symbol Strings at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Strings); 
--# expr-done 442
--# expr-start 443 '<Symbol Text_Buffers at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Text_Buffers); 
--# expr-done 443
Array_Lit := Create_Symbol_Type_Array (Internal_Symbol_Type_Array'(1 => Sym, 2 => Sym_1, 3 => Sym_2)); 
--# expr-done 444
--# property-call-start AdaNode.get_unit_root_decl
Fld := Libadalang.Implementation.Ada_Node_P_Get_Unit_Root_Decl (Node => Self, Name => Array_Lit, Kind => Unit_Specification);
--# end
--# expr-done 445
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 446 '<FieldAccess for children_env at ast.py:955>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:955
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Children_Env (Node => Var_Expr);
--# expr-done 446
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1735;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
--# expr-start 447 '<Symbol Root_Buffer_Type at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Root_Buffer_Type); 
--# expr-done 447
Env_Get_Result := AST_Envs.Get_First (Self => Result_Var, Key => Sym_3, Lookup_Kind => To_Lookup_Kind_Type (Flat), Categories => (Inherited_Primitives => True, Nocat => True)); 
--# expr-done 448
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 449
      Property_Result := Cast_Result;
   --# end
      Finalizer_Scope_70;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_70;
               Finalizer_Scope_1735;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Root_Buffer_Type;
--# end
--# property-start AdaNode.has_with_visibility /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:964
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Has_With_Visibility
  (Node : Bare_Ada_Node
      ; Refd_Unit : Internal_Unit
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind refd_unit Refd_Unit
   Property_Result : Boolean;
            procedure Finalizer_Scope_71 with Inline_Always;
      Var_Expr : Bare_Ada_Node;
Var_Expr_1 : Internal_Entity_Base_Subp_Body;
Var_Expr_2 : Internal_Unit;
Var_Expr_3 : Bare_Ada_Node;
Var_Expr_4 : Bare_Compilation_Unit;
Fld : Internal_Unit;
Fld_1 : Boolean;
Fld_2 : Internal_Unit;
Fld_3 : Bare_Basic_Decl;
Fld_4 : Lexical_Env := Empty_Env;
Fld_5 : Lexical_Env := Empty_Env;
Env_Node : Bare_Ada_Node;
Fld_6 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result : Boolean;
Fld_7 : Internal_Unit;
Fld_8 : Bare_Basic_Decl;
As_Entity : Internal_Entity_Basic_Decl;
Is_Null : Boolean;
If_Result_1 : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity_Base_Subp_Body;
Fld_9 : Internal_Entity_Defining_Name;
Fld_10 : Internal_Unit;
Fld_11 : Bare_Ada_Node;
Fld_12 : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
Fld_13 : Boolean;
If_Result_2 : Boolean;
Scope_Result_3 : Boolean;
Result_Var_3 : Boolean;
If_Result_3 : Boolean;
Fld_14 : Bare_Ada_Node;
Cast_Result_1 : Bare_Compilation_Unit;
Fld_15 : Boolean;
Scope_Result_4 : Boolean;
Result_Var_4 : Boolean;
If_Result_4 : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_71 is
            begin
                     Dec_Ref (Fld_4);
                     Dec_Ref (Fld_5);
            end Finalizer_Scope_71;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 471 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 451 '<FieldAccess for is_referenced_from at ast.py:972>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:972
      if Refd_Unit = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 450 '<FieldAccess for unit at ast.py:972>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:972
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Unit (Node => Self);
--# expr-done 450
Fld_1 := Libadalang.Implementation.Is_Referenced_From (Self => Refd_Unit, Unit => Fld);
--# expr-done 451
if Fld_1 then
   If_Result := True;
else
--# expr-start 456 '<env_node at ast.py:973>' Env_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:973
--# expr-start 455 '<FieldAccess for parent_unit_env at ast.py:973>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:973
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 454 '<FieldAccess for children_env at ast.py:978>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:978
--# expr-start 453 '<FieldAccess for top_level_decl at ast.py:978>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:978
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 452 '<FieldAccess for unit at ast.py:978>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:978
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Unit (Node => Self);
--# expr-done 452
--# property-call-start AdaNode.top_level_decl
Fld_3 := Libadalang.Implementation.Ada_Node_P_Top_Level_Decl (Node => Self, Unit => Fld_2);
--# end
--# expr-done 453
      if Fld_3 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Children_Env (Node => Fld_3);
--# expr-done 454
--# property-call-start AdaNode.parent_unit_env
Fld_5 := Libadalang.Implementation.Ada_Node_P_Parent_Unit_Env (Node => Self, Env => Fld_4);
--# end
--# expr-done 455
Env_Node := AST_Envs.Env_Node (Fld_5); 
--# expr-done 456
Var_Expr := Env_Node;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 457 '<FieldAccess for has_with_visibility at ast.py:973>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:973
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.has_with_visibility
Fld_6 := Libadalang.Implementation.Ada_Node_P_Has_With_Visibility (Node => Var_Expr, Refd_Unit => Refd_Unit);
--# end
--# expr-done 457
Scope_Result := Fld_6;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
   If_Result := Result_Var;
end if;
if If_Result then
   If_Result_3 := True;
else
--# expr-start 461 '<Cast to BaseSubpBody at ast.py:986>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:986
--# expr-start 460 '<as_bare_entity at ast.py:986>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:986
--# expr-start 459 '<FieldAccess for top_level_decl at ast.py:986>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:986
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 458 '<FieldAccess for unit at ast.py:987>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:987
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Unit (Node => Self);
--# expr-done 458
--# property-call-start AdaNode.top_level_decl
Fld_8 := Libadalang.Implementation.Ada_Node_P_Top_Level_Decl (Node => Self, Unit => Fld_7);
--# end
--# expr-done 459
Is_Null := Fld_8 = null; 
if Is_Null then
   If_Result_1 := No_Entity_Basic_Decl;
else
As_Entity := (Info => No_Entity_Info, Node => Fld_8); 
   If_Result_1 := As_Entity;
end if;
--# expr-done 460
   if If_Result_1.Node = null
      or else If_Result_1.Node.Kind in Ada_Base_Subp_Body
   then
      Cast_Result := Create_Internal_Entity_Base_Subp_Body
        (Node => If_Result_1.Node,
         Info => If_Result_1.Info);
   else
         Cast_Result := No_Entity_Base_Subp_Body;
   end if;
--# expr-done 461
Var_Expr_1 := Cast_Result;
if Var_Expr_1 /= No_Entity_Base_Subp_Body then
   --# scope-start
   --# bind b Var_Expr_1
--# expr-start 467 '<If at ast.py:989>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:989
--# expr-start 462 '<FieldAccess for is_library_item at ast.py:990>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:990
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_library_item'
Fld_13 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Library_Item (Node => Var_Expr_1.Node);
--# end
--# expr-done 462
if Fld_13 then
--# expr-start 464 '<FieldAccess for referenced_unit at ast.py:991>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:991
--# expr-start 463 '<FieldAccess for defining_name at ast.py:991>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:991
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_9 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 463
      if Fld_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_unit
Fld_10 := Libadalang.Implementation.Name_P_Referenced_Unit (Node => Fld_9.Node, Kind => Unit_Specification, Not_Found_Is_Error => False);
--# end
--# expr-done 464
Var_Expr_2 := Fld_10;
if Var_Expr_2 /= null then
   --# scope-start
   --# bind u Var_Expr_2
--# expr-start 465 '<FieldAccess for root at ast.py:994>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:994
      if Var_Expr_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Var_Expr_2.Ast_Root;
--# expr-done 465
Var_Expr_3 := Fld_11;
if Var_Expr_3 /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 466 '<FieldAccess for has_with_visibility at ast.py:994>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:994
      if Var_Expr_3 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.has_with_visibility
Fld_12 := Libadalang.Implementation.Ada_Node_P_Has_With_Visibility (Node => Var_Expr_3, Refd_Unit => Refd_Unit);
--# end
--# expr-done 466
Scope_Result_1 := Fld_12;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
Scope_Result_2 := Result_Var_1;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := False;
end if;
   If_Result_2 := Result_Var_2;
else
   If_Result_2 := False;
end if;
--# expr-done 467
Scope_Result_3 := If_Result_2;
   --# end
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := False;
end if;
   If_Result_3 := Result_Var_3;
end if;
if If_Result_3 then
   If_Result_4 := True;
else
--# expr-start 469 '<Cast to CompilationUnit at ast.py:1007>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1007
--# expr-start 468 '<FieldAccess for root at ast.py:1007>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1007
      if Refd_Unit = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Refd_Unit.Ast_Root;
--# expr-done 468
   if Fld_14 = null
      or else Fld_14.Kind in Ada_Compilation_Unit_Range
   then
      Cast_Result_1 := Fld_14;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 469
Var_Expr_4 := Cast_Result_1;
if Var_Expr_4 /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 470 '<FieldAccess for is_text_io_child at ast.py:1007>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1007
      if Var_Expr_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.is_text_io_child
Fld_15 := Libadalang.Implementation.Compilation_Unit_P_Is_Text_Io_Child (Node => Var_Expr_4);
--# end
--# expr-done 470
Scope_Result_4 := Fld_15;
   --# end
   Result_Var_4 := Scope_Result_4;
else
   Result_Var_4 := False;
end if;
   If_Result_4 := Result_Var_4;
end if;
Let_Result := If_Result_4; 
--# end
--# expr-done 471
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_71;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_71;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Has_With_Visibility;
--# end
--# property-start AdaNode.has_visibility /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1011
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Has_Visibility
  (Node : Bare_Ada_Node
      ; Other_Entity : Internal_Entity
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind other_entity Other_Entity
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Generic_Package_Instantiation;
Var_Expr_1 : Internal_Entity_Package_Renaming_Decl;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Generic_Package_Instantiation;
Fld : Internal_Entity_Info;
Fld_1 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Cast_Expr_1 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Package_Renaming_Decl;
Fld_2 : Internal_Entity_Info;
Fld_3 : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
If_Result : Boolean;
Cast_Expr_2 : Internal_Entity;
Cast_Result_2 : Internal_Entity_Basic_Decl;
Fld_4 : Boolean;
Not_Val : Boolean;
If_Result_1 : Boolean;
Fld_5 : Bare_Ada_Node;
Fld_6 : Internal_Unit;
Fld_7 : Boolean;
If_Result_2 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 484 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 472 '<Cast to GenericPackageInstantiation at ast.py:1015>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1015
Cast_Expr := Other_Entity; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Generic_Package_Instantiation_Range
   then
      Cast_Result := Create_Internal_Entity_Generic_Package_Instantiation
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Generic_Package_Instantiation;
   end if;
--# expr-done 472
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Generic_Package_Instantiation then
   --# scope-start
--# expr-start 474 '<FieldAccess for from_rebound at ast.py:1015>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1015
--# expr-start 473 '<FieldAccess for info at ast.py:1015>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1015
Fld := Var_Expr.Info;
--# expr-done 473
Fld_1 := Fld.From_Rebound;
--# expr-done 474
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
if Result_Var then
   If_Result := True;
else
--# expr-start 475 '<Cast to PackageRenamingDecl at ast.py:1017>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1017
Cast_Expr_1 := Other_Entity; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Package_Renaming_Decl_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Package_Renaming_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_1 := No_Entity_Package_Renaming_Decl;
   end if;
--# expr-done 475
Var_Expr_1 := Cast_Result_1;
if Var_Expr_1 /= No_Entity_Package_Renaming_Decl then
   --# scope-start
--# expr-start 477 '<FieldAccess for from_rebound at ast.py:1017>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1017
--# expr-start 476 '<FieldAccess for info at ast.py:1017>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1017
Fld_2 := Var_Expr_1.Info;
--# expr-done 476
Fld_3 := Fld_2.From_Rebound;
--# expr-done 477
Scope_Result_1 := Fld_3;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
   If_Result := Result_Var_1;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 480 '<Not at ast.py:1020>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1020
--# expr-start 479 '<FieldAccess for is_compilation_unit_root at ast.py:1020>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1020
--# expr-start 478 '<Cast to BasicDecl at ast.py:1020>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1020
Cast_Expr_2 := Other_Entity; 
   if Cast_Expr_2.Node = null
      or else Cast_Expr_2.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_2 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
   else
         Cast_Result_2 := No_Entity_Basic_Decl;
   end if;
--# expr-done 478
      if Cast_Result_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_compilation_unit_root
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Is_Compilation_Unit_Root (Node => Cast_Result_2.Node);
--# end
--# expr-done 479
Not_Val := not (Fld_4); 
--# expr-done 480
   If_Result_1 := Not_Val;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
   --# expr-start 483 '<FieldAccess for has_with_visibility at ast.py:1023>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1023
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 482 '<FieldAccess for unit at ast.py:1023>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1023
--# expr-start 481 '<FieldAccess for node at ast.py:1023>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1023
Fld_5 := Other_Entity.Node;
--# expr-done 481
      if Fld_5 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Unit (Node => Fld_5);
--# expr-done 482
--# property-call-start AdaNode.has_with_visibility
Fld_7 := Libadalang.Implementation.Ada_Node_P_Has_With_Visibility (Node => Self, Refd_Unit => Fld_6);
--# end
--# expr-done 483
   If_Result_2 := Fld_7;
end if;
Let_Result := If_Result_2; 
--# end
--# expr-done 484
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Has_Visibility;
--# end
--# property-start AdaNode.resolve_generic_actual /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1027
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Resolve_Generic_Actual
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity :=
     Internal_Entity'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity;
            procedure Finalizer_Scope_1746 with Inline_Always;
      Match_0 : Internal_Entity_Anonymous_Expr_Decl;
Match_1 : Internal_Entity_Name;
Match_2 : Internal_Entity;
Var_Expr : Internal_Entity;
Fld : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Scope_Result : Internal_Entity;
Fld_1 : Internal_Entity_Array_Access;
Get_Result : Internal_Entity;
Result_Var : Internal_Entity;
Match_Prefix : Internal_Entity;
Cast_Result_1 : Internal_Entity_Anonymous_Expr_Decl;
Cast_Expr : Internal_Entity_Anonymous_Expr_Decl;
Cast_Result_2 : Internal_Entity;
Let_Result : Internal_Entity;
Scope_Result_1 : Internal_Entity;
Cast_Result_3 : Internal_Entity_Name;
Let_Result_1 : Internal_Entity;
Scope_Result_2 : Internal_Entity;
Cast_Result_4 : Internal_Entity;
Let_Result_2 : Internal_Entity;
Scope_Result_3 : Internal_Entity;
Match_Result : Internal_Entity;
Let_Result_3 : Internal_Entity;
            procedure Finalizer_Scope_1746 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1746;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 490 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 489 '<Match at ast.py:1031>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1031
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Anonymous_Expr_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Anonymous_Expr_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_1; 
--# bind aod Match_0
Cast_Expr := Match_0; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Let_Result := Cast_Result_2; 
--# end
Scope_Result_1 := Let_Result;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Attribute_Ref .. Ada_Target_Name =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_3; 
--# bind n Match_1
--# expr-start 486 '<Cast to AdaNode.entity at ast.py:1037>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1037
--# expr-start 485 '<FieldAccess for name_designated_type at ast.py:1037>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1037
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Match_1.Node, E_Info => Match_1.Info);
--# end
--# expr-done 485
      Cast_Result := Create_Internal_Entity
        (Node => Fld.Node,
         Info => Fld.Info);
--# expr-done 486
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity then
   --# scope-start
   --# bind e Var_Expr
Scope_Result := Var_Expr;
   --# end
   Result_Var := Scope_Result;
else
   --# expr-start 488 '<at at ast.py:1039>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1039
--# expr-start 487 '<FieldAccess for all_env_elements at ast.py:1039>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1039
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.all_env_elements
Fld_1 := Libadalang.Implementation.Name_P_All_Env_Elements (Node => Match_1.Node, E_Info => Match_1.Info);
--# end
--# expr-done 487
Get_Result := Get (Fld_1, 0, True); 
--# expr-done 488
   Result_Var := Get_Result;
end if;
Let_Result_1 := Result_Var; 
--# end
Scope_Result_2 := Let_Result_1;
   --# end
      Finalizer_Scope_1746;
            Match_Result := Scope_Result_2; 
         when Ada_Abort_Absent .. Ada_Abstract_State_Decl | Ada_Component_Decl .. Ada_Membership_Expr | Ada_Paren_Expr .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_4; 
--# bind _ Match_2
Let_Result_2 := No_Entity; 
--# end
Scope_Result_3 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_3; 
end case;
--# expr-done 489
Let_Result_3 := Match_Result; 
--# end
--# expr-done 490
      Property_Result := Let_Result_3;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1746;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Resolve_Generic_Actual;
--# end
--# property-start AdaNode.top_level_use_package_clauses /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1046
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Top_Level_Use_Package_Clauses
  (Node : Bare_Ada_Node
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_74 with Inline_Always;
            procedure Finalizer_Scope_1750 with Inline_Always;
      Fld : Bare_Ada_Node;
Fld_1 : Bare_Ada_Node;
Cast_Result : Bare_Compilation_Unit;
Fld_2 : Bare_Ada_Node_List;
Item_12 : Bare_Ada_Node;
Is_A : Boolean;
Map_Result : Bare_Ada_Node_Array_Access;
Cast_Expr : Bare_Ada_Node;
Cast_Result_1 : Bare_Use_Package_Clause;
Fld_3 : Bare_Name_List;
Cast_Expr_1 : Bare_Name;
Cast_Result_2 : Bare_Ada_Node;
Item_13 : Bare_Name;
Map_Result_1 : Bare_Ada_Node_Array_Access;
Map_Result_2 : Bare_Ada_Node_Array_Access;
Let_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_74 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Map_Result_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_74;
            procedure Finalizer_Scope_1750 is
            begin
                     Dec_Ref (Map_Result_1);
            end Finalizer_Scope_1750;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 502 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 501 '<Mapcat at ast.py:1053>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1053
   --# expr-start 496 '<Filter at ast.py:1053>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1053
   --# expr-start 494 '<FieldAccess for prelude at ast.py:1053>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1053
--# expr-start 493 '<Cast to CompilationUnit at ast.py:1053>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1053
--# expr-start 492 '<parent at ast.py:1053>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1053
--# expr-start 491 '<parent at ast.py:1053>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1053
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 491
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Parent; 
--# expr-done 492
   if Fld_1 = null
      or else Fld_1.Kind in Ada_Compilation_Unit_Range
   then
      Cast_Result := Fld_1;
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 493
      if Cast_Result = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Cast_Result.Compilation_Unit_F_Prelude;
--# expr-done 494
      declare
         Map_Result_Vec : Bare_Ada_Node_Vectors.Vector;
      begin
      if Fld_2 = null then
      Map_Result := Create_Bare_Ada_Node_Array (0);
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_2;
      begin
         for Untyped_Item_12 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_12 := Untyped_Item_12; 
   --# scope-start
               --# bind p Item_12
         --# expr-start 495 '<IsA UsePackageClause at ast.py:1055>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1055
Is_A := Item_12 /= null 
and then Item_12.Kind in Ada_Use_Package_Clause_Range; 
--# expr-done 495
         if Is_A then
         declare
            Item_To_Append : constant Bare_Ada_Node := Item_12;
         begin
            Bare_Ada_Node_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Bare_Ada_Node_Array
           (Items_Count => Natural (Bare_Ada_Node_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Bare_Ada_Node_Vectors.Get
              (Map_Result_Vec,
               I + Bare_Ada_Node_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Bare_Ada_Node_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 496
      declare
         Map_Result_2_Vec : Bare_Ada_Node_Vectors.Vector;
      begin
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Map_Result;
      begin
         for Item_11 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_11
      --# expr-start 500 '<Map at ast.py:1057>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1057
   --# expr-start 498 '<FieldAccess for packages at ast.py:1057>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1057
--# expr-start 497 '<Cast to UsePackageClause at ast.py:1057>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1057
Cast_Expr := Item_11; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_Use_Package_Clause_Range
   then
      Cast_Result_1 := Cast_Expr;
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 497
      if Cast_Result_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Cast_Result_1.Use_Package_Clause_F_Packages;
--# expr-done 498
      declare
         Map_Result_1_Vec : Bare_Ada_Node_Vectors.Vector;
      begin
      if Fld_3 = null then
      Map_Result_1 := Create_Bare_Ada_Node_Array (0);
      else
      declare
         Collection : constant Bare_Name_List := Fld_3;
      begin
         for Untyped_Item_13 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_13 := Untyped_Item_13; 
   --# scope-start
               --# bind n Item_13
      --# expr-start 499 '<Cast to AdaNode at ast.py:1058>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1058
Cast_Expr_1 := Item_13; 
      Cast_Result_2 := Cast_Expr_1;
--# expr-done 499
         declare
            Item_To_Append : constant Bare_Ada_Node := Cast_Result_2;
         begin
            Bare_Ada_Node_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result_1 := Create_Bare_Ada_Node_Array
           (Items_Count => Natural (Bare_Ada_Node_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Bare_Ada_Node_Vectors.Get
              (Map_Result_1_Vec,
               I + Bare_Ada_Node_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Bare_Ada_Node_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 500
         for Item_To_Append of
               Map_Result_1.Items
         loop
            Bare_Ada_Node_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_1750;
         end loop;
      end;
         Map_Result_2 := Create_Bare_Ada_Node_Array
           (Items_Count => Natural (Bare_Ada_Node_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Bare_Ada_Node_Vectors.Get
              (Map_Result_2_Vec,
               I + Bare_Ada_Node_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Bare_Ada_Node_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 501
Let_Result := Map_Result_2; Inc_Ref (Let_Result);
--# end
--# expr-done 502
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_74;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_74;
               Finalizer_Scope_1750;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Top_Level_Use_Package_Clauses;
--# end
--# property-start AdaNode.top_level_use_type_clauses /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1064
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Top_Level_Use_Type_Clauses
  (Node : Bare_Ada_Node
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_75 with Inline_Always;
            procedure Finalizer_Scope_1753 with Inline_Always;
      Fld : Bare_Ada_Node;
Fld_1 : Bare_Ada_Node;
Cast_Result : Bare_Compilation_Unit;
Fld_2 : Bare_Ada_Node_List;
Item_15 : Bare_Ada_Node;
Is_A : Boolean;
Map_Result : Bare_Ada_Node_Array_Access;
Cast_Expr : Bare_Ada_Node;
Cast_Result_1 : Bare_Use_Type_Clause;
Fld_3 : Bare_Name_List;
Cast_Expr_1 : Bare_Name;
Cast_Result_2 : Bare_Ada_Node;
Item_16 : Bare_Name;
Map_Result_1 : Bare_Ada_Node_Array_Access;
Map_Result_2 : Bare_Ada_Node_Array_Access;
Let_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_75 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Map_Result_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_75;
            procedure Finalizer_Scope_1753 is
            begin
                     Dec_Ref (Map_Result_1);
            end Finalizer_Scope_1753;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 514 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 513 '<Mapcat at ast.py:1071>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1071
   --# expr-start 508 '<Filter at ast.py:1071>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1071
   --# expr-start 506 '<FieldAccess for prelude at ast.py:1071>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1071
--# expr-start 505 '<Cast to CompilationUnit at ast.py:1071>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1071
--# expr-start 504 '<parent at ast.py:1071>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1071
--# expr-start 503 '<parent at ast.py:1071>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1071
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 503
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Parent; 
--# expr-done 504
   if Fld_1 = null
      or else Fld_1.Kind in Ada_Compilation_Unit_Range
   then
      Cast_Result := Fld_1;
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 505
      if Cast_Result = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Cast_Result.Compilation_Unit_F_Prelude;
--# expr-done 506
      declare
         Map_Result_Vec : Bare_Ada_Node_Vectors.Vector;
      begin
      if Fld_2 = null then
      Map_Result := Create_Bare_Ada_Node_Array (0);
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_2;
      begin
         for Untyped_Item_15 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_15 := Untyped_Item_15; 
   --# scope-start
               --# bind p Item_15
         --# expr-start 507 '<IsA UseTypeClause at ast.py:1073>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1073
Is_A := Item_15 /= null 
and then Item_15.Kind in Ada_Use_Type_Clause_Range; 
--# expr-done 507
         if Is_A then
         declare
            Item_To_Append : constant Bare_Ada_Node := Item_15;
         begin
            Bare_Ada_Node_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Bare_Ada_Node_Array
           (Items_Count => Natural (Bare_Ada_Node_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Bare_Ada_Node_Vectors.Get
              (Map_Result_Vec,
               I + Bare_Ada_Node_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Bare_Ada_Node_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 508
      declare
         Map_Result_2_Vec : Bare_Ada_Node_Vectors.Vector;
      begin
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Map_Result;
      begin
         for Item_14 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_14
      --# expr-start 512 '<Map at ast.py:1075>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1075
   --# expr-start 510 '<FieldAccess for types at ast.py:1075>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1075
--# expr-start 509 '<Cast to UseTypeClause at ast.py:1075>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1075
Cast_Expr := Item_14; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_Use_Type_Clause_Range
   then
      Cast_Result_1 := Cast_Expr;
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 509
      if Cast_Result_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Cast_Result_1.Use_Type_Clause_F_Types;
--# expr-done 510
      declare
         Map_Result_1_Vec : Bare_Ada_Node_Vectors.Vector;
      begin
      if Fld_3 = null then
      Map_Result_1 := Create_Bare_Ada_Node_Array (0);
      else
      declare
         Collection : constant Bare_Name_List := Fld_3;
      begin
         for Untyped_Item_16 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_16 := Untyped_Item_16; 
   --# scope-start
               --# bind n Item_16
      --# expr-start 511 '<Cast to AdaNode at ast.py:1076>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1076
Cast_Expr_1 := Item_16; 
      Cast_Result_2 := Cast_Expr_1;
--# expr-done 511
         declare
            Item_To_Append : constant Bare_Ada_Node := Cast_Result_2;
         begin
            Bare_Ada_Node_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result_1 := Create_Bare_Ada_Node_Array
           (Items_Count => Natural (Bare_Ada_Node_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Bare_Ada_Node_Vectors.Get
              (Map_Result_1_Vec,
               I + Bare_Ada_Node_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Bare_Ada_Node_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 512
         for Item_To_Append of
               Map_Result_1.Items
         loop
            Bare_Ada_Node_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_1753;
         end loop;
      end;
         Map_Result_2 := Create_Bare_Ada_Node_Array
           (Items_Count => Natural (Bare_Ada_Node_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Bare_Ada_Node_Vectors.Get
              (Map_Result_2_Vec,
               I + Bare_Ada_Node_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Bare_Ada_Node_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 513
Let_Result := Map_Result_2; Inc_Ref (Let_Result);
--# end
--# expr-done 514
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_75;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_75;
               Finalizer_Scope_1753;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Top_Level_Use_Type_Clauses;
--# end
--# property-start AdaNode.top_level_use_clauses /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1082
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Top_Level_Use_Clauses
  (Node : Bare_Ada_Node
  )
   return Bare_Use_Clause_Array_Access
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Use_Clause_Array_Access;
            procedure Finalizer_Scope_76 with Inline_Always;
      Block_Var : Bare_Compilation_Unit;
Fld : Bare_Ada_Node;
Fld_1 : Bare_Ada_Node;
Cast_Result : Bare_Compilation_Unit;
Fld_2 : Bare_Ada_Node_List;
Cast_Expr : Bare_Ada_Node;
Cast_Result_1 : Bare_Use_Clause;
Item_17 : Bare_Ada_Node;
Is_A : Boolean;
Map_Result : Bare_Use_Clause_Array_Access;
Let_Result : Bare_Use_Clause_Array_Access;
            procedure Finalizer_Scope_76 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_76;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 522 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 517 '<Cast to CompilationUnit at ast.py:1087>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1087
--# expr-start 516 '<parent at ast.py:1087>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1087
--# expr-start 515 '<parent at ast.py:1087>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1087
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 515
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Parent; 
--# expr-done 516
   if Fld_1 = null
      or else Fld_1.Kind in Ada_Compilation_Unit_Range
   then
      Cast_Result := Fld_1;
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 517
Block_Var := Cast_Result; 
--# bind cu Block_Var
--# expr-start 521 '<FilterMap at ast.py:1088>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1088
   --# expr-start 518 '<FieldAccess for prelude at ast.py:1088>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1088
      if Block_Var = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Block_Var.Compilation_Unit_F_Prelude;
--# expr-done 518
      declare
         Map_Result_Vec : Bare_Use_Clause_Vectors.Vector;
      begin
      if Fld_2 = null then
      Map_Result := Create_Bare_Use_Clause_Array (0);
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_2;
      begin
         for Untyped_Item_17 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_17 := Untyped_Item_17; 
   --# scope-start
               --# bind p Item_17
         --# expr-start 519 '<IsA UseClause at ast.py:1090>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1090
Is_A := Item_17 /= null 
and then Item_17.Kind in Ada_Use_Clause; 
--# expr-done 519
         if Is_A then
      --# expr-start 520 '<Cast to UseClause at ast.py:1089>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1089
Cast_Expr := Item_17; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_Use_Clause
   then
      Cast_Result_1 := Cast_Expr;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 520
         declare
            Item_To_Append : constant Bare_Use_Clause := Cast_Result_1;
         begin
            Bare_Use_Clause_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Bare_Use_Clause_Array
           (Items_Count => Natural (Bare_Use_Clause_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Bare_Use_Clause_Vectors.Get
              (Map_Result_Vec,
               I + Bare_Use_Clause_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Bare_Use_Clause_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 521
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 522
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_76;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_76;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Top_Level_Use_Clauses;
--# end
--# property-start AdaNode.top_level_with_package_clauses /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1094
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Top_Level_With_Package_Clauses
  (Node : Bare_Ada_Node
  )
   return Bare_Name_Array_Access
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Name_Array_Access;
            procedure Finalizer_Scope_77 with Inline_Always;
            procedure Finalizer_Scope_1756 with Inline_Always;
      Var_Expr : Bare_With_Clause;
Fld : Bare_Compilation_Unit;
Fld_1 : Bare_Ada_Node_List;
Cast_Expr : Bare_Ada_Node;
Cast_Result : Bare_With_Clause;
Fld_2 : Bare_Name_List;
Scope_Result : Bare_Name_List;
Result_Var : Bare_Name_List;
Item_356 : Bare_Name;
Map_Result : Bare_Name_Array_Access;
Item_18 : Bare_Ada_Node;
Map_Result_1 : Bare_Name_Array_Access;
Let_Result : Bare_Name_Array_Access;
            procedure Finalizer_Scope_77 is
            begin
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_77;
            procedure Finalizer_Scope_1756 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_1756;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 529 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 528 '<Mapcat at ast.py:1100>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1100
   --# expr-start 524 '<FieldAccess for prelude at ast.py:1100>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1100
--# expr-start 523 '<FieldAccess for enclosing_compilation_unit at ast.py:1100>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1100
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.enclosing_compilation_unit
Fld := Libadalang.Implementation.Ada_Node_P_Enclosing_Compilation_Unit (Node => Self);
--# end
--# expr-done 523
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Compilation_Unit_F_Prelude;
--# expr-done 524
      declare
         Map_Result_1_Vec : Bare_Name_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result_1 := Create_Bare_Name_Array (0);
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_1;
      begin
         for Untyped_Item_18 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_18 := Untyped_Item_18; 
   --# scope-start
               --# bind p Item_18
      --# expr-start 527 '<TakeWhile at ???>' Map_Result None
--# expr-start 525 '<Cast to WithClause at ast.py:1102>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1102
Cast_Expr := Item_18; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_With_Clause_Range
   then
      Cast_Result := Cast_Expr;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 525
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 526 '<FieldAccess for packages at ast.py:1102>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1102
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Var_Expr.With_Clause_F_Packages;
--# expr-done 526
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node;
end if;
      declare
         Map_Result_Vec : Bare_Name_Vectors.Vector;
      begin
      if Result_Var = null then
      Map_Result := Create_Bare_Name_Array (0);
      else
      declare
         Collection : constant Bare_Name_List := Result_Var;
      begin
         for Untyped_Item_356 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_356 := Untyped_Item_356; 
   --# scope-start
         declare
            Item_To_Append : constant Bare_Name := Item_356;
         begin
            Bare_Name_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Bare_Name_Array
           (Items_Count => Natural (Bare_Name_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Bare_Name_Vectors.Get
              (Map_Result_Vec,
               I + Bare_Name_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Bare_Name_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 527
         for Item_To_Append of
               Map_Result.Items
         loop
            Bare_Name_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_1756;
         end loop;
      end;
      end if;
         Map_Result_1 := Create_Bare_Name_Array
           (Items_Count => Natural (Bare_Name_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Bare_Name_Vectors.Get
              (Map_Result_1_Vec,
               I + Bare_Name_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Bare_Name_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 528
Let_Result := Map_Result_1; Inc_Ref (Let_Result);
--# end
--# expr-done 529
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_77;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_77;
               Finalizer_Scope_1756;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Top_Level_With_Package_Clauses;
--# end
--# property-start AdaNode.use_clauses_in_spec_of_subp_body /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1106
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Use_Clauses_In_Spec_Of_Subp_Body
  (Node : Bare_Ada_Node
  )
   return Lexical_Env
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_78 with Inline_Always;
            procedure Finalizer_Scope_1760 with Inline_Always;
            procedure Finalizer_Scope_1761 with Inline_Always;
      Block_Var : Symbol_Type_Array_Access;
Block_Var_1 : Bare_Compilation_Unit;
Var_Expr : Bare_Basic_Decl;
Var_Expr_1 : Bare_Compilation_Unit;
Fld : Bare_Compilation_Unit;
Fld_1 : Symbol_Type_Array_Access;
Fld_2 : Bare_Compilation_Unit;
Fld_3 : Bare_Basic_Decl;
Scope_Result : Bare_Basic_Decl;
Result_Var : Bare_Basic_Decl;
Fld_4 : Bare_Use_Clause_Array_Access;
Scope_Result_1 : Bare_Use_Clause_Array_Access;
Result_Var_1 : Bare_Use_Clause_Array_Access;
As_Entity : Internal_Entity_Use_Clause;
Is_Null : Boolean;
If_Result : Internal_Entity_Use_Clause;
Fld_5 : Lexical_Env := Empty_Env;
Map_Result : Lexical_Env_Array_Access;
Group_Env : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_78 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Group_Env);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_78;
            procedure Finalizer_Scope_1760 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_1760;
            procedure Finalizer_Scope_1761 is
            begin
                     Dec_Ref (Fld_5);
            end Finalizer_Scope_1761;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 539 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 531 '<FieldAccess for syntactic_fully_qualified_name at ast.py:1112>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1112
--# expr-start 530 '<FieldAccess for enclosing_compilation_unit at ast.py:1112>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1112
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.enclosing_compilation_unit
Fld := Libadalang.Implementation.Ada_Node_P_Enclosing_Compilation_Unit (Node => Self);
--# end
--# expr-done 530
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.syntactic_fully_qualified_name
Fld_1 := Libadalang.Implementation.Compilation_Unit_P_Syntactic_Fully_Qualified_Name (Node => Fld);
--# end
--# expr-done 531
Block_Var := Fld_1; Inc_Ref (Block_Var);
--# bind fqn Block_Var
--# expr-start 532 '<FieldAccess for designated_compilation_unit at ast.py:1114>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1114
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.designated_compilation_unit
Fld_2 := Libadalang.Implementation.Ada_Node_P_Designated_Compilation_Unit (Node => Self, Name => Block_Var, Kind => Unit_Specification, Not_Found_Is_Error => False);
--# end
--# expr-done 532
Block_Var_1 := Fld_2; 
--# bind spec Block_Var_1
--# expr-start 538 '<env_group at ast.py:1119>' Group_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1119
--# expr-start 537 '<Map at ast.py:1119>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1119
Var_Expr_1 := Block_Var_1;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 533 '<FieldAccess for decl at ast.py:1119>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1119
      if Var_Expr_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.decl
Fld_3 := Libadalang.Implementation.Compilation_Unit_P_Decl (Node => Var_Expr_1);
--# end
--# expr-done 533
Scope_Result := Fld_3;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node;
end if;
Var_Expr := Result_Var;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 534 '<FieldAccess for top_level_use_clauses at ast.py:1119>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1119
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_clauses
Fld_4 := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Clauses (Node => Var_Expr);
--# end
--# expr-done 534
Scope_Result_1 := Fld_4;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_1760;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Bare_Use_Clause_Array_Type;
end if;
      Inc_Ref (Result_Var_1);
      declare
         Map_Result_Vec : Lexical_Env_Vectors.Vector;
      begin
      declare
         Collection : constant Bare_Use_Clause_Array_Access := Result_Var_1;
      begin
         for Item_19 of
               Collection.Items
         loop
   --# scope-start
               --# bind clause Item_19
      --# expr-start 536 '<FieldAccess for used_envs at ast.py:1120>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1120
--# expr-start 535 '<as_bare_entity at ast.py:1120>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1120
Is_Null := Item_19 = null; 
if Is_Null then
   If_Result := No_Entity_Use_Clause;
else
As_Entity := (Info => No_Entity_Info, Node => Item_19); 
   If_Result := As_Entity;
end if;
--# expr-done 535
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start UseClause.used_envs
Fld_5 := Libadalang.Implementation.Use_Clause_P_Used_Envs (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 536
         declare
            Item_To_Append : constant Lexical_Env := Fld_5;
         begin
               Inc_Ref (Item_To_Append);
            Lexical_Env_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_1761;
         end loop;
      end;
         Map_Result := Create_Lexical_Env_Array
           (Items_Count => Natural (Lexical_Env_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Lexical_Env_Vectors.Get
              (Map_Result_Vec,
               I + Lexical_Env_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Lexical_Env_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 537
Group_Env := Group (Map_Result, No_Metadata); 
--# expr-done 538
Let_Result := Group_Env; Inc_Ref (Let_Result);
--# end
--# expr-done 539
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_78;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_78;
               Finalizer_Scope_1760;
               Finalizer_Scope_1761;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Use_Clauses_In_Spec_Of_Subp_Body;
--# end
--# property-start AdaNode.use_clauses_in_generic_formal_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1124
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Use_Clauses_In_Generic_Formal_Part
  (Node : Bare_Ada_Node
  )
   return Lexical_Env
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_79 with Inline_Always;
            procedure Finalizer_Scope_1763 with Inline_Always;
      Block_Var : Internal_Entity_Generic_Decl;
Var_Expr : Internal_Entity_Body_Node;
Var_Expr_1 : Internal_Entity_Generic_Decl;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Cast_Result : Internal_Entity_Body_Node;
Fld : Internal_Entity_Generic_Decl;
Scope_Result : Internal_Entity_Generic_Decl;
Result_Var : Internal_Entity_Generic_Decl;
Fld_1 : Internal_Entity_Generic_Formal_Part;
Fld_2 : Lexical_Env := Empty_Env;
Scope_Result_1 : Lexical_Env := Empty_Env;
Fld_3 : Lexical_Env := Empty_Env;
Result_Var_1 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_79 is
            begin
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_79;
            procedure Finalizer_Scope_1763 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_1763;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 546 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 541 '<Cast to Body at ast.py:1132>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1132
--# expr-start 540 '<as_bare_entity at ast.py:1132>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1132
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 540
   if If_Result.Node = null
      or else If_Result.Node.Kind in Ada_Body_Node
   then
      Cast_Result := Create_Internal_Entity_Body_Node
        (Node => If_Result.Node,
         Info => If_Result.Info);
   else
         Cast_Result := No_Entity_Body_Node;
   end if;
--# expr-done 541
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Body_Node then
   --# scope-start
--# expr-start 542 '<FieldAccess for safe_generic_decl_part at ast.py:1132>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1132
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.safe_generic_decl_part
Fld := Libadalang.Implementation.Body_Node_P_Safe_Generic_Decl_Part (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 542
Scope_Result := Fld;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Generic_Decl;
end if;
Block_Var := Result_Var; 
--# bind gen_decl Block_Var
Var_Expr_1 := Block_Var;
if Var_Expr_1 /= No_Entity_Generic_Decl then
   --# scope-start
   --# bind gd Var_Expr_1
--# expr-start 544 '<FieldAccess for use_clauses_envs at ast.py:1134>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1134
--# expr-start 543 '<FieldAccess for formal_part at ast.py:1134>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1134
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Generic_Formal_Part (Node => Var_Expr_1.Node.Generic_Decl_F_Formal_Part, Info => Var_Expr_1.Info);
--# expr-done 543
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericFormalPart.use_clauses_envs
Fld_2 := Libadalang.Implementation.Generic_Formal_Part_P_Use_Clauses_Envs (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 544
Scope_Result_1 := Fld_2;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_1763;
   Result_Var_1 := Scope_Result_1;
else
   --# expr-start 545 '<FieldAccess for empty_env at ast.py:1135>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1135
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.empty_env
Fld_3 := Libadalang.Implementation.Ada_Node_P_Empty_Env (Node => Self);
--# end
--# expr-done 545
   Result_Var_1 := Fld_3;
end if;
      Inc_Ref (Result_Var_1);
Let_Result := Result_Var_1; Inc_Ref (Let_Result);
--# end
--# expr-done 546
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_79;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_79;
               Finalizer_Scope_1763;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Use_Clauses_In_Generic_Formal_Part;
--# end
--# property-start AdaNode.nested_generic_formal_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1139
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Nested_Generic_Formal_Part
  (Node : Bare_Ada_Node
  )
   return Lexical_Env
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_80 with Inline_Always;
            procedure Finalizer_Scope_1765 with Inline_Always;
      Block_Var : Internal_Entity_Generic_Decl;
Var_Expr : Internal_Entity_Body_Node;
Var_Expr_1 : Internal_Entity_Generic_Decl;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Cast_Result : Internal_Entity_Body_Node;
Fld : Internal_Entity_Generic_Decl;
Scope_Result : Internal_Entity_Generic_Decl;
Result_Var : Internal_Entity_Generic_Decl;
Fld_1 : Bare_Generic_Decl;
Fld_2 : Lexical_Env := Empty_Env;
Scope_Result_1 : Lexical_Env := Empty_Env;
Fld_3 : Lexical_Env := Empty_Env;
Result_Var_1 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_80 is
            begin
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_80;
            procedure Finalizer_Scope_1765 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_1765;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 553 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 548 '<Cast to Body at ast.py:1153>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1153
--# expr-start 547 '<as_bare_entity at ast.py:1153>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1153
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 547
   if If_Result.Node = null
      or else If_Result.Node.Kind in Ada_Body_Node
   then
      Cast_Result := Create_Internal_Entity_Body_Node
        (Node => If_Result.Node,
         Info => If_Result.Info);
   else
         Cast_Result := No_Entity_Body_Node;
   end if;
--# expr-done 548
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Body_Node then
   --# scope-start
--# expr-start 549 '<FieldAccess for safe_generic_decl_part at ast.py:1153>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1153
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.safe_generic_decl_part
Fld := Libadalang.Implementation.Body_Node_P_Safe_Generic_Decl_Part (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 549
Scope_Result := Fld;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Generic_Decl;
end if;
Block_Var := Result_Var; 
--# bind gen_decl Block_Var
Var_Expr_1 := Block_Var;
if Var_Expr_1 /= No_Entity_Generic_Decl then
   --# scope-start
   --# bind gd Var_Expr_1
--# expr-start 551 '<FieldAccess for children_env at ast.py:1155>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1155
--# expr-start 550 '<FieldAccess for node at ast.py:1155>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1155
Fld_1 := Var_Expr_1.Node;
--# expr-done 550
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Children_Env (Node => Fld_1);
--# expr-done 551
Scope_Result_1 := Fld_2;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_1765;
   Result_Var_1 := Scope_Result_1;
else
   --# expr-start 552 '<FieldAccess for empty_env at ast.py:1156>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1156
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.empty_env
Fld_3 := Libadalang.Implementation.Ada_Node_P_Empty_Env (Node => Self);
--# end
--# expr-done 552
   Result_Var_1 := Fld_3;
end if;
      Inc_Ref (Result_Var_1);
Let_Result := Result_Var_1; Inc_Ref (Let_Result);
--# end
--# expr-done 553
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_80;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_80;
               Finalizer_Scope_1765;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Nested_Generic_Formal_Part;
--# end
--# property-start AdaNode.is_package /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1160
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Is_Package
  (Node : Bare_Ada_Node
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Is_A : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 555 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 554 '<IsA PackageDecl, PackageBody, GenericPackageInstantiation, PackageRenamingDecl, GenericPackageDecl at ast.py:1164>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1164
Is_A := Self /= null 
and then Self.Kind in Ada_Package_Decl_Range | Ada_Package_Body_Range | Ada_Generic_Package_Instantiation_Range | Ada_Package_Renaming_Decl_Range | Ada_Generic_Package_Decl_Range; 
--# expr-done 554
Let_Result := Is_A; 
--# end
--# expr-done 555
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Is_Package;
--# end
--# property-start AdaNode.default_initial_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1168
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Default_Initial_Env
  (Node : Bare_Ada_Node
  )
   return Lexical_Env
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_82 with Inline_Always;
            procedure Finalizer_Scope_1766 with Inline_Always;
      Var_Expr : Bare_Ada_Node;
Fld : Bare_Ada_Node;
Fld_1 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Fld_2 : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_82 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_82;
            procedure Finalizer_Scope_1766 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1766;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 559 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 556 '<parent at ast.py:1173>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1173
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 556
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind p Var_Expr
--# expr-start 557 '<FieldAccess for children_env at ast.py:1173>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1173
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Children_Env (Node => Var_Expr);
--# expr-done 557
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1766;
   Result_Var := Scope_Result;
else
   --# expr-start 558 '<FieldAccess for children_env at ast.py:1174>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1174
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Children_Env (Node => Self);
--# expr-done 558
   Result_Var := Fld_2;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 559
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_82;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_82;
               Finalizer_Scope_1766;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Default_Initial_Env;
--# end
--# property-start AdaNode.top_level_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1177
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Top_Level_Decl
  (Node : Bare_Ada_Node
      ; Unit : Internal_Unit
  )
   return Bare_Basic_Decl
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind unit Unit
   Property_Result : Bare_Basic_Decl;
      Var_Expr : Bare_Ada_Node;
Var_Expr_1 : Internal_Unit;
Fld : Bare_Ada_Node;
Scope_Result : Bare_Ada_Node;
Result_Var : Bare_Ada_Node;
Cast_Expr : Bare_Ada_Node;
Cast_Result : Bare_Compilation_Unit;
Fld_1 : Bare_Basic_Decl;
Scope_Result_1 : Bare_Basic_Decl;
Result_Var_1 : Bare_Basic_Decl;
Let_Result : Bare_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 563 '<Block at ???>' Let_Result None
--# scope-start
Var_Expr_1 := Unit;
if Var_Expr_1 /= null then
   --# scope-start
--# expr-start 560 '<FieldAccess for root at ast.py:1182>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1182
      if Var_Expr_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Var_Expr_1.Ast_Root;
--# expr-done 560
Scope_Result := Fld;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node;
end if;
Var_Expr := Result_Var;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 562 '<FieldAccess for decl at ast.py:1182>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1182
--# expr-start 561 '<Cast to CompilationUnit at ast.py:1182>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1182
Cast_Expr := Var_Expr; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_Compilation_Unit_Range
   then
      Cast_Result := Cast_Expr;
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 561
      if Cast_Result = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.decl
Fld_1 := Libadalang.Implementation.Compilation_Unit_P_Decl (Node => Cast_Result);
--# end
--# expr-done 562
Scope_Result_1 := Fld_1;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Bare_Ada_Node;
end if;
Let_Result := Result_Var_1; 
--# end
--# expr-done 563
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Top_Level_Decl;
--# end
--# property-start AdaNode.unpack_formals /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1185
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Unpack_Formals
  (Node : Bare_Ada_Node
      ; Formal_Params : Internal_Entity_Base_Formal_Param_Decl_Array_Access
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind formal_params Formal_Params
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_84 with Inline_Always;
      Fld : Internal_Unit;
Fld_1 : Bare_Ada_Node;
Fld_2 : Internal_Entity_Defining_Name_Array_Access;
Let_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_84 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_84;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 567 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 566 '<FieldAccess for unpack_formals_impl at ast.py:1189>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1189
--# expr-start 565 '<FieldAccess for root at ast.py:1189>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1189
--# expr-start 564 '<FieldAccess for unit at ast.py:1189>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1189
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Unit (Node => Self);
--# expr-done 564
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Ast_Root;
--# expr-done 565
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.unpack_formals_impl
Fld_2 := Libadalang.Implementation.Ada_Node_P_Unpack_Formals_Impl (Node => Fld_1, Formal_Params => Formal_Params);
--# end
--# expr-done 566
Let_Result := Fld_2; Inc_Ref (Let_Result);
--# end
--# expr-done 567
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_84;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_84;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Unpack_Formals;
--# end
--# property-start AdaNode.unpack_formals_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1192
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Unpack_Formals_Impl
  (Node : Bare_Ada_Node
      ; Formal_Params : Internal_Entity_Base_Formal_Param_Decl_Array_Access
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind formal_params Formal_Params
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_85 with Inline_Always;
            procedure Finalizer_Scope_1767 with Inline_Always;
      Fld : Internal_Entity_Defining_Name_Array_Access;
Map_Result : Internal_Entity_Defining_Name_Array_Access;
Let_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_85 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_85;
            procedure Finalizer_Scope_1767 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1767;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 570 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 569 '<Mapcat at ast.py:1193>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1193
      declare
         Map_Result_Vec : Internal_Entity_Defining_Name_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Base_Formal_Param_Decl_Array_Access := Formal_Params;
      begin
         for Item_20 of
               Collection.Items
         loop
   --# scope-start
               --# bind spec Item_20
      --# expr-start 568 '<FieldAccess for defining_names at ast.py:1193>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1193
      if Item_20.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_names'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Names (Node => Item_20.Node, E_Info => Item_20.Info);
--# end
--# expr-done 568
         for Item_To_Append of
               Fld.Items
         loop
            Internal_Entity_Defining_Name_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_1767;
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Defining_Name_Array
           (Items_Count => Natural (Internal_Entity_Defining_Name_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Defining_Name_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Defining_Name_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Defining_Name_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 569
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 570
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_85;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_85;
               Finalizer_Scope_1767;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Unpack_Formals_Impl;
--# end
--# property-start AdaNode.match_formals /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1196
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Match_Formals
  (Node : Bare_Ada_Node
      ; Formal_Params : Internal_Entity_Base_Formal_Param_Decl_Array_Access
      ; Params : Internal_Entity_Assoc_List
      ; Is_Dottable_Subp : Boolean
  )
   return Internal_Param_Match_Array_Access
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind formal_params Formal_Params
   --# bind params Params
   --# bind is_dottable_subp Is_Dottable_Subp
   Property_Result : Internal_Param_Match_Array_Access;
            procedure Finalizer_Scope_86 with Inline_Always;
            procedure Finalizer_Scope_1768 with Inline_Always;
      Block_Var : Internal_Entity_Defining_Name_Array_Access;
Var_Expr : Internal_Entity_Assoc_List;
Idx : Integer;
Var_Expr_1 : Internal_Entity_Defining_Name;
Var_Expr_2 : Bare_Base_Id;
Var_Expr_3 : Internal_Entity_Defining_Name;
Fld : Internal_Entity_Defining_Name_Array_Access;
Fld_1 : Internal_Single_Actual_Array_Access;
Arith_Result : Integer;
If_Result : Integer;
Get_Result : Internal_Entity_Defining_Name;
New_Struct : Internal_Param_Match;
Scope_Result : Internal_Param_Match;
Result_Var : Internal_Param_Match;
Let_Result : Internal_Param_Match;
Fld_2 : Bare_Base_Id;
Fld_3 : Internal_Entity_Name;
Cast_Expr : Bare_Base_Id;
Cast_Result : Bare_Name;
Fld_4 : Boolean;
Find_Result : Internal_Entity_Defining_Name;
New_Struct_1 : Internal_Param_Match;
Scope_Result_1 : Internal_Param_Match;
Result_Var_1 : Internal_Param_Match;
Scope_Result_2 : Internal_Param_Match;
Result_Var_2 : Internal_Param_Match;
Fld_5 : Bare_Base_Id;
Is_Null : Boolean;
If_Result_1 : Internal_Param_Match;
Index_22 : Integer;
Map_Result : Internal_Param_Match_Array_Access;
Scope_Result_3 : Internal_Param_Match_Array_Access;
Result_Var_3 : Internal_Param_Match_Array_Access;
Let_Result_1 : Internal_Param_Match_Array_Access;
            procedure Finalizer_Scope_86 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Result_Var_3);
                     Dec_Ref (Let_Result_1);
            end Finalizer_Scope_86;
            procedure Finalizer_Scope_1768 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_1768;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 587 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 571 '<FieldAccess for unpack_formals at ast.py:1204>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1204
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.unpack_formals
Fld := Libadalang.Implementation.Ada_Node_P_Unpack_Formals (Node => Self, Formal_Params => Formal_Params);
--# end
--# expr-done 571
Block_Var := Fld; Inc_Ref (Block_Var);
--# bind unpacked_formals Block_Var
Var_Expr := Params;
if Var_Expr /= No_Entity_Assoc_List then
   --# scope-start
   --# bind p Var_Expr
--# expr-start 586 '<Map at ast.py:1206>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1206
   --# expr-start 572 '<FieldAccess for unpacked_params at ast.py:1206>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1206
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AssocList.unpacked_params
Fld_1 := Libadalang.Implementation.Assoc_List_P_Unpacked_Params (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 572
      declare
         Map_Result_Vec : Internal_Param_Match_Vectors.Vector;
      begin
         Index_22 := 0;
      declare
         Collection : constant Internal_Single_Actual_Array_Access := Fld_1;
      begin
         for Item_21 of
               Collection.Items
         loop
   --# scope-start
               --# bind a Item_21
               --# bind i Index_22
      --# expr-start 585 '<If at ast.py:1206>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1206
--# expr-start 574 '<IsNull at ast.py:1207>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1207
--# expr-start 573 '<FieldAccess for name at ast.py:1207>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1207
Fld_5 := Item_21.Name;
--# expr-done 573
Is_Null := Fld_5 = null; 
--# expr-done 574
if Is_Null then
   --# expr-start 579 '<Let at ast.py:1209>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1209
--# scope-start
--# expr-start 576 '<If at ast.py:1210>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1210
if Is_Dottable_Subp then
   --# expr-start 575 '<Arithmetic + at ast.py:1210>' Arith_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1210
Arith_Result := (Index_22 + 1); 
--# expr-done 575
   If_Result := Arith_Result;
else
   If_Result := Index_22;
end if;
--# expr-done 576
Idx := If_Result; 
--# bind idx Idx
--# expr-start 577 '<at at ast.py:1213>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1213
Get_Result := Get (Block_Var, Idx, True); 
--# expr-done 577
Var_Expr_1 := Get_Result;
if Var_Expr_1 /= No_Entity_Defining_Name then
   --# scope-start
   --# bind sp Var_Expr_1
--# expr-start 578 '<New ParamMatch at ast.py:1214>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1214
New_Struct := (Actual => Item_21, Formal => Var_Expr_1, Has_Matched => True); 
--# expr-done 578
Scope_Result := New_Struct;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Param_Match;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 579
   If_Result_1 := Let_Result;
else
--# expr-start 580 '<FieldAccess for name at ast.py:1224>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1224
Fld_2 := Item_21.Name;
--# expr-done 580
Var_Expr_2 := Fld_2;
if Var_Expr_2 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind id Var_Expr_2
--# expr-start 583 '<Find at ast.py:1225>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1225
      Find_Result := No_Entity_Defining_Name;
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Block_Var;
      begin
         for Item_23 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_23
      --# expr-start 582 '<FieldAccess for matches at ast.py:1225>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1225
--# expr-start 581 '<FieldAccess for name at ast.py:1225>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1225
      if Item_23.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Name (Node => Item_23.Node.Defining_Name_F_Name, Info => Item_23.Info);
--# expr-done 581
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Cast_Expr := Var_Expr_2; 
      Cast_Result := Cast_Expr;
--# property-call-start Name.matches
Fld_4 := Libadalang.Implementation.Name_P_Matches (Node => Fld_3.Node, N => Cast_Result);
--# end
--# expr-done 582
      if Fld_4 then
         Find_Result := Item_23;
      end if;
   --# end
               exit when Fld_4;
         end loop;
      end;
--# expr-done 583
Var_Expr_3 := Find_Result;
if Var_Expr_3 /= No_Entity_Defining_Name then
   --# scope-start
   --# bind sp Var_Expr_3
--# expr-start 584 '<New ParamMatch at ast.py:1226>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1226
New_Struct_1 := (Actual => Item_21, Formal => Var_Expr_3, Has_Matched => True); 
--# expr-done 584
Scope_Result_1 := New_Struct_1;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Param_Match;
end if;
Scope_Result_2 := Result_Var_1;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Param_Match;
end if;
   If_Result_1 := Result_Var_2;
end if;
--# expr-done 585
         declare
            Item_To_Append : constant Internal_Param_Match := If_Result_1;
         begin
            Internal_Param_Match_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
               Index_22 := Index_22 + 1;
         end loop;
      end;
         Map_Result := Create_Internal_Param_Match_Array
           (Items_Count => Natural (Internal_Param_Match_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Param_Match_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Param_Match_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Param_Match_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 586
Scope_Result_3 := Map_Result;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_1768;
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := No_Internal_Param_Match_Array_Type;
end if;
      Inc_Ref (Result_Var_3);
Let_Result_1 := Result_Var_3; Inc_Ref (Let_Result_1);
--# end
--# expr-done 587
      Property_Result := Let_Result_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_86;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_86;
               Finalizer_Scope_1768;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Match_Formals;
--# end
--# property-start AdaNode.choice_match /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1235
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Choice_Match
  (Node : Bare_Ada_Node
      ; Value : Big_Integer_Type
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity :=
     Internal_Entity'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind value Value
   Property_Result : Boolean;
            procedure Finalizer_Scope_1777 with Inline_Always;
            procedure Finalizer_Scope_1778 with Inline_Always;
            procedure Finalizer_Scope_1780 with Inline_Always;
            procedure Finalizer_Scope_1783 with Inline_Always;
            procedure Finalizer_Scope_1784 with Inline_Always;
      Match_0 : Internal_Entity_Bin_Op;
Match_1 : Internal_Entity_Name;
Match_2 : Internal_Entity_Subtype_Indication;
Match_3 : Internal_Entity_Expr;
Match_4 : Internal_Entity_Others_Designator;
Match_5 : Internal_Entity;
Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Discrete_Range;
Edr : Internal_Eval_Discrete_Range := No_Eval_Discrete_Range;
Var_Expr_2 : Internal_Discrete_Range;
Edr_1 : Internal_Eval_Discrete_Range := No_Eval_Discrete_Range;
Fld : Internal_Entity_Expr;
Fld_1 : Big_Integer_Type;
Comp_Result : Boolean;
Fld_2 : Internal_Entity_Expr;
Fld_3 : Big_Integer_Type;
Comp_Result_1 : Boolean;
If_Result : Boolean;
Fld_4 : Big_Integer_Type;
Is_Equal : Boolean;
Fld_5 : Internal_Entity_Op;
Is_A : Boolean;
If_Result_1 : Boolean;
Fld_6 : Internal_Entity_Base_Type_Decl;
Fld_7 : Internal_Discrete_Range;
Fld_8 : Internal_Eval_Discrete_Range := No_Eval_Discrete_Range;
Fld_9 : Big_Integer_Type;
Comp_Result_2 : Boolean;
Fld_10 : Big_Integer_Type;
Comp_Result_3 : Boolean;
If_Result_2 : Boolean;
Let_Result : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Bound_Origin : Bare_Ada_Node;
Bound_Imprecise_Fallback : Boolean;
Fld_11 : Boolean;
Dyn_Var_Bind_Result : Boolean;
Dyn_Var_Bind_Result_1 : Boolean;
If_Result_3 : Boolean;
Scope_Result_1 : Boolean;
Fld_12 : Big_Integer_Type;
Is_Equal_1 : Boolean;
Result_Var_1 : Boolean;
Fld_13 : Internal_Discrete_Range;
Fld_14 : Internal_Eval_Discrete_Range := No_Eval_Discrete_Range;
Fld_15 : Big_Integer_Type;
Comp_Result_4 : Boolean;
Fld_16 : Big_Integer_Type;
Comp_Result_5 : Boolean;
If_Result_4 : Boolean;
Let_Result_1 : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
Bound_Origin_1 : Bare_Ada_Node;
Bound_Imprecise_Fallback_1 : Boolean;
Fld_17 : Internal_Entity_Base_Type_Decl;
Fld_18 : Boolean;
Dyn_Var_Bind_Result_2 : Boolean;
Dyn_Var_Bind_Result_3 : Boolean;
If_Result_5 : Boolean;
Fld_19 : Big_Integer_Type;
Is_Equal_2 : Boolean;
Match_Prefix : Internal_Entity;
Cast_Result : Internal_Entity_Bin_Op;
Let_Result_2 : Boolean;
Scope_Result_3 : Boolean;
Cast_Result_1 : Internal_Entity_Name;
Let_Result_3 : Boolean;
Scope_Result_4 : Boolean;
Cast_Result_2 : Internal_Entity_Subtype_Indication;
Let_Result_4 : Boolean;
Scope_Result_5 : Boolean;
Cast_Result_3 : Internal_Entity_Expr;
Let_Result_5 : Boolean;
Scope_Result_6 : Boolean;
Cast_Result_4 : Internal_Entity_Others_Designator;
Let_Result_6 : Boolean;
Scope_Result_7 : Boolean;
Cast_Result_5 : Internal_Entity;
Let_Result_7 : Boolean;
Scope_Result_8 : Boolean;
Match_Result : Boolean;
Let_Result_8 : Boolean;
            procedure Finalizer_Scope_1777 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_1777;
            procedure Finalizer_Scope_1778 is
            begin
                     Dec_Ref (Fld_12);
            end Finalizer_Scope_1778;
            procedure Finalizer_Scope_1780 is
            begin
                     Dec_Ref (Fld_8);
                     Dec_Ref (Edr);
                     Dec_Ref (Fld_9);
                     Dec_Ref (Fld_10);
            end Finalizer_Scope_1780;
            procedure Finalizer_Scope_1783 is
            begin
                     Dec_Ref (Fld_14);
                     Dec_Ref (Edr_1);
                     Dec_Ref (Fld_15);
                     Dec_Ref (Fld_16);
            end Finalizer_Scope_1783;
            procedure Finalizer_Scope_1784 is
            begin
                     Dec_Ref (Fld_19);
            end Finalizer_Scope_1784;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 620 '<Block at ???>' Let_Result_8 None
--# scope-start
--# expr-start 619 '<Match at ast.py:1246>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1246
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Bin_Op .. Ada_Relation_Op =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Bin_Op
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind bo Match_0
--# expr-start 596 '<If at ast.py:1250>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1250
--# expr-start 589 '<IsA Op.DoubleDot at ast.py:1252>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1252
--# expr-start 588 '<FieldAccess for op at ast.py:1252>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1252
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Op (Node => Match_0.Node.Bin_Op_F_Op, Info => Match_0.Info);
--# expr-done 588
Is_A := Fld_5.Node /= null 
and then Fld_5.Node.Kind in Ada_Op_Double_Dot_Range; 
--# expr-done 589
if Is_A then
--# expr-start 591 '<FieldAccess for eval_as_int at ast.py:1254>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1254
--# expr-start 590 '<FieldAccess for left at ast.py:1254>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1254
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Match_0.Node.Bin_Op_F_Left, Info => Match_0.Info);
--# expr-done 590
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.eval_as_int
Fld_1 := Libadalang.Implementation.Expr_P_Eval_As_Int (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 591
Comp_Result := Value >= Fld_1; 
if Comp_Result then
--# expr-start 593 '<FieldAccess for eval_as_int at ast.py:1255>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1255
--# expr-start 592 '<FieldAccess for right at ast.py:1255>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1255
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Expr (Node => Match_0.Node.Bin_Op_F_Right, Info => Match_0.Info);
--# expr-done 592
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.eval_as_int
Fld_3 := Libadalang.Implementation.Expr_P_Eval_As_Int (Node => Fld_2.Node, E_Info => Fld_2.Info);
--# end
--# expr-done 593
Comp_Result_1 := Value <= Fld_3; 
   If_Result := Comp_Result_1;
else
   If_Result := False;
end if;
   If_Result_1 := If_Result;
else
   --# expr-start 595 '<Eq at ast.py:1257>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1257
--# expr-start 594 '<FieldAccess for eval_as_int at ast.py:1257>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1257
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.eval_as_int
Fld_4 := Libadalang.Implementation.Expr_P_Eval_As_Int (Node => Match_0.Node, E_Info => Match_0.Info);
--# end
--# expr-done 594
Is_Equal := Equivalent (Value, Fld_4); 
--# expr-done 595
   If_Result_1 := Is_Equal;
end if;
--# expr-done 596
Let_Result_2 := If_Result_1; 
--# end
Scope_Result_3 := Let_Result_2;
   --# end
      Finalizer_Scope_1777;
            Match_Result := Scope_Result_3; 
         when Ada_Attribute_Ref .. Ada_Target_Name =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind n Match_1
--# expr-start 597 '<FieldAccess for name_designated_type at ast.py:1262>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1262
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_6 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Match_1.Node, E_Info => Match_1.Info);
--# end
--# expr-done 597
Var_Expr := Fld_6;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind dt Var_Expr
--# expr-start 598 '<FieldAccess for discrete_range at ast.py:1263>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1263
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.discrete_range'
Fld_7 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Discrete_Range (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 598
Var_Expr_1 := Fld_7;
if Var_Expr_1 /= No_Discrete_Range then
   --# scope-start
   --# bind dr Var_Expr_1
--# expr-start 602 '<Let at ast.py:1264>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1264
--# scope-start
--# expr-start 599 '<FieldAccess for eval_discrete_range at ast.py:1265>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1265
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.eval_discrete_range
Fld_8 := Libadalang.Implementation.Ada_Node_P_Eval_Discrete_Range (Node => Self, Dr => Var_Expr_1);
--# end
--# expr-done 599
Edr := Fld_8; Inc_Ref (Edr);
--# bind edr Edr
--# expr-start 600 '<FieldAccess for low_bound at ast.py:1266>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1266
Fld_9 := Edr.Low_Bound;
Inc_Ref (Fld_9);
--# expr-done 600
Comp_Result_2 := Value >= Fld_9; 
if Comp_Result_2 then
--# expr-start 601 '<FieldAccess for high_bound at ast.py:1266>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1266
Fld_10 := Edr.High_Bound;
Inc_Ref (Fld_10);
--# expr-done 601
Comp_Result_3 := Value <= Fld_10; 
   If_Result_2 := Comp_Result_3;
else
   If_Result_2 := False;
end if;
Let_Result := If_Result_2; 
--# end
--# expr-done 602
Scope_Result := Let_Result;
   --# end
      Finalizer_Scope_1780;
   Result_Var := Scope_Result;
else
   Result_Var := True;
end if;
if Result_Var then
   --# expr-start 605 '<bind at ast.py:1270>' Dyn_Var_Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1270
Bound_Origin := Self; 
--# expr-start 604 '<bind at ast.py:1271>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1271
Bound_Imprecise_Fallback := False; 
--# expr-start 603 '<FieldAccess for satisfies_type_predicates at ast.py:1272>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1272
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.satisfies_type_predicates
Fld_11 := Libadalang.Implementation.Base_Type_Decl_P_Satisfies_Type_Predicates (Node => Var_Expr.Node, Value => Value, Imprecise_Fallback => Bound_Imprecise_Fallback, Origin => Bound_Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 603
Dyn_Var_Bind_Result := Fld_11; 
--# expr-done 604
Dyn_Var_Bind_Result_1 := Dyn_Var_Bind_Result; 
--# expr-done 605
   If_Result_3 := Dyn_Var_Bind_Result_1;
else
   If_Result_3 := False;
end if;
Scope_Result_1 := If_Result_3;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   --# expr-start 607 '<Eq at ast.py:1275>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1275
--# expr-start 606 '<FieldAccess for eval_as_int at ast.py:1275>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1275
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.eval_as_int
Fld_12 := Libadalang.Implementation.Expr_P_Eval_As_Int (Node => Match_1.Node, E_Info => Match_1.Info);
--# end
--# expr-done 606
Is_Equal_1 := Equivalent (Value, Fld_12); 
--# expr-done 607
   Result_Var_1 := Is_Equal_1;
end if;
Let_Result_3 := Result_Var_1; 
--# end
Scope_Result_4 := Let_Result_3;
   --# end
      Finalizer_Scope_1778;
            Match_Result := Scope_Result_4; 
         when Ada_Subtype_Indication .. Ada_Discrete_Subtype_Indication =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Subtype_Indication
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind st Match_2
--# expr-start 608 '<FieldAccess for discrete_range at ast.py:1279>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1279
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SubtypeIndication.discrete_range
Fld_13 := Libadalang.Implementation.Subtype_Indication_P_Discrete_Range (Node => Match_2.Node, E_Info => Match_2.Info);
--# end
--# expr-done 608
Var_Expr_2 := Fld_13;
if Var_Expr_2 /= No_Discrete_Range then
   --# scope-start
   --# bind dr Var_Expr_2
--# expr-start 612 '<Let at ast.py:1280>' Let_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1280
--# scope-start
--# expr-start 609 '<FieldAccess for eval_discrete_range at ast.py:1281>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1281
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.eval_discrete_range
Fld_14 := Libadalang.Implementation.Ada_Node_P_Eval_Discrete_Range (Node => Self, Dr => Var_Expr_2);
--# end
--# expr-done 609
Edr_1 := Fld_14; Inc_Ref (Edr_1);
--# bind edr Edr_1
--# expr-start 610 '<FieldAccess for low_bound at ast.py:1282>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1282
Fld_15 := Edr_1.Low_Bound;
Inc_Ref (Fld_15);
--# expr-done 610
Comp_Result_4 := Value >= Fld_15; 
if Comp_Result_4 then
--# expr-start 611 '<FieldAccess for high_bound at ast.py:1282>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1282
Fld_16 := Edr_1.High_Bound;
Inc_Ref (Fld_16);
--# expr-done 611
Comp_Result_5 := Value <= Fld_16; 
   If_Result_4 := Comp_Result_5;
else
   If_Result_4 := False;
end if;
Let_Result_1 := If_Result_4; 
--# end
--# expr-done 612
Scope_Result_2 := Let_Result_1;
   --# end
      Finalizer_Scope_1783;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := True;
end if;
if Result_Var_2 then
   --# expr-start 616 '<bind at ast.py:1286>' Dyn_Var_Bind_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1286
Bound_Origin_1 := Self; 
--# expr-start 615 '<bind at ast.py:1287>' Dyn_Var_Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1287
Bound_Imprecise_Fallback_1 := False; 
--# expr-start 614 '<FieldAccess for satisfies_type_predicates at ast.py:1288>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1288
--# expr-start 613 '<FieldAccess for designated_type at ast.py:1288>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1288
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_17 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Match_2.Node, Origin => Bound_Origin_1, E_Info => Match_2.Info);
--# end
--# expr-done 613
      if Fld_17.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.satisfies_type_predicates
Fld_18 := Libadalang.Implementation.Base_Type_Decl_P_Satisfies_Type_Predicates (Node => Fld_17.Node, Value => Value, Imprecise_Fallback => Bound_Imprecise_Fallback_1, Origin => Bound_Origin_1, E_Info => Fld_17.Info);
--# end
--# expr-done 614
Dyn_Var_Bind_Result_2 := Fld_18; 
--# expr-done 615
Dyn_Var_Bind_Result_3 := Dyn_Var_Bind_Result_2; 
--# expr-done 616
   If_Result_5 := Dyn_Var_Bind_Result_3;
else
   If_Result_5 := False;
end if;
Let_Result_4 := If_Result_5; 
--# end
Scope_Result_5 := Let_Result_4;
   --# end
            Match_Result := Scope_Result_5; 
         when Ada_Abstract_State_Decl_Expr .. Ada_Null_Record_Aggregate | Ada_Box_Expr .. Ada_Membership_Expr | Ada_Paren_Expr .. Ada_Un_Op =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_3; 
--# bind e Match_3
--# expr-start 618 '<Eq at ast.py:1293>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1293
--# expr-start 617 '<FieldAccess for eval_as_int at ast.py:1293>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1293
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.eval_as_int
Fld_19 := Libadalang.Implementation.Expr_P_Eval_As_Int (Node => Match_3.Node, E_Info => Match_3.Info);
--# end
--# expr-done 617
Is_Equal_2 := Equivalent (Value, Fld_19); 
--# expr-done 618
Let_Result_5 := Is_Equal_2; 
--# end
Scope_Result_6 := Let_Result_5;
   --# end
      Finalizer_Scope_1784;
            Match_Result := Scope_Result_6; 
         when Ada_Others_Designator =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Others_Designator
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_4 := Cast_Result_4; 
--# bind _ Match_4
Let_Result_6 := True; 
--# end
Scope_Result_7 := Let_Result_6;
   --# end
            Match_Result := Scope_Result_7; 
         when Ada_Abort_Absent .. Ada_Elsif_Stmt_Part | Ada_Handled_Stmts .. Ada_Null_Component_Decl | Ada_Overriding_Not_Overriding .. Ada_Enum_Lit_Synth_Type_Expr | Ada_Synthetic_Type_Expr .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_5 := Cast_Result_5; 
--# bind _ Match_5
Let_Result_7 := False; 
--# end
Scope_Result_8 := Let_Result_7;
   --# end
            Match_Result := Scope_Result_8; 
end case;
--# expr-done 619
Let_Result_8 := Match_Result; 
--# end
--# expr-done 620
      Property_Result := Let_Result_8;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1777;
               Finalizer_Scope_1778;
               Finalizer_Scope_1780;
               Finalizer_Scope_1783;
               Finalizer_Scope_1784;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Choice_Match;
--# end
--# property-start AdaNode.gnat_xref /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1302
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Gnat_Xref
  (Node : Bare_Ada_Node
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity :=
     Internal_Entity'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_1800 with Inline_Always;
      Block_Var : Internal_Entity_Basic_Decl;
Var_Expr : Internal_Entity_Defining_Name;
Var_Expr_1 : Internal_Entity_Basic_Decl;
Var_Expr_2 : Internal_Entity_Basic_Decl;
Var_Expr_3 : Internal_Entity_Basic_Decl;
Var_Expr_4 : Internal_Entity_Basic_Decl;
Var_Expr_5 : Internal_Entity_Basic_Decl;
Var_Expr_6 : Internal_Entity_Base_Type_Decl;
Var_Expr_7 : Boolean;
Var_Expr_8 : Internal_Entity_Defining_Name;
Var_Expr_9 : Internal_Entity_Base_Type_Decl;
Var_Expr_10 : Internal_Entity_Defining_Name;
Var_Expr_11 : Internal_Entity_Name;
Dbd : Internal_Entity_Basic_Decl;
Var_Expr_12 : Internal_Entity_Generic_Instantiation;
Var_Expr_13 : Internal_Entity_Defining_Name;
Var_Expr_14 : Internal_Entity_Basic_Decl;
Var_Expr_15 : Internal_Entity_Basic_Decl;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Name;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Bound_Origin : Bare_Ada_Node;
Is_A : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
Fld_2 : Internal_Entity;
Is_A_1 : Boolean;
If_Result : Boolean;
Fld_3 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Basic_Decl;
Fld_4 : Internal_Entity_Defining_Name;
Is_A_2 : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
Fld_5 : Internal_Entity;
Cast_Result_2 : Internal_Entity_Basic_Decl;
Fld_6 : Internal_Entity_Defining_Name;
Is_A_3 : Boolean;
Scope_Result_3 : Boolean;
Result_Var_3 : Boolean;
Fld_7 : Internal_Entity;
Is_A_4 : Boolean;
If_Result_1 : Boolean;
Fld_8 : Internal_Entity;
Cast_Result_3 : Internal_Entity_Basic_Decl;
Fld_9 : Internal_Entity_Defining_Name;
Is_A_5 : Boolean;
Scope_Result_4 : Boolean;
Result_Var_4 : Boolean;
Cast_Expr_1 : Internal_Entity_Basic_Decl;
Cast_Result_4 : Internal_Entity_Abstract_Subp_Decl;
Fld_10 : Internal_Entity_Base_Subp_Spec;
Fld_11 : Internal_Entity_Base_Type_Decl;
Fld_12 : Internal_Entity_Defining_Name;
Is_A_6 : Boolean;
Scope_Result_5 : Boolean;
Result_Var_5 : Boolean;
Cast_Expr_2 : Internal_Entity_Basic_Decl;
Cast_Result_5 : Internal_Entity_Basic_Subp_Decl;
Fld_13 : Internal_Entity_Base_Subp_Spec;
Fld_14 : Internal_Entity_Base_Type_Decl;
Fld_15 : Boolean;
Fld_16 : Internal_Entity_Base_Type_Decl;
Fld_17 : Internal_Entity_Defining_Name;
Scope_Result_6 : Internal_Entity_Defining_Name;
Result_Var_6 : Internal_Entity_Defining_Name;
Scope_Result_7 : Internal_Entity_Defining_Name;
Fld_18 : Internal_Entity_Defining_Name;
Result_Var_7 : Internal_Entity_Defining_Name;
Scope_Result_8 : Internal_Entity_Defining_Name;
Result_Var_8 : Internal_Entity_Defining_Name;
Scope_Result_9 : Internal_Entity_Defining_Name;
Result_Var_9 : Internal_Entity_Defining_Name;
Cast_Expr_3 : Internal_Entity;
Cast_Result_6 : Internal_Entity_Name;
Fld_19 : Internal_Entity_Defining_Name;
Scope_Result_10 : Internal_Entity_Defining_Name;
Result_Var_10 : Internal_Entity_Defining_Name;
Fld_20 : Internal_Entity_Basic_Decl;
Is_A_7 : Boolean;
Cast_Expr_4 : Internal_Entity_Basic_Decl;
Cast_Result_7 : Internal_Entity_Param_Spec;
Fld_21 : Internal_Entity_Defining_Name;
Is_A_8 : Boolean;
Fld_22 : Internal_Entity_Generic_Instantiation_Array_Access;
Get_Result : Internal_Entity_Generic_Instantiation;
Cast_Expr_5 : Internal_Entity_Generic_Instantiation;
Cast_Result_8 : Internal_Entity_Basic_Decl;
Fld_23 : Internal_Entity_Defining_Name;
Scope_Result_11 : Internal_Entity_Defining_Name;
Result_Var_11 : Internal_Entity_Defining_Name;
Is_A_9 : Boolean;
Cast_Expr_6 : Internal_Entity_Basic_Decl;
Cast_Result_9 : Internal_Entity_Object_Decl;
Fld_24 : Internal_Entity_Basic_Decl;
Fld_25 : Internal_Entity_Defining_Name;
Scope_Result_12 : Internal_Entity_Defining_Name;
Result_Var_12 : Internal_Entity_Defining_Name;
Scope_Result_13 : Internal_Entity_Defining_Name;
Result_Var_13 : Internal_Entity_Defining_Name;
Is_A_10 : Boolean;
Cast_Expr_7 : Internal_Entity_Basic_Decl;
Cast_Result_10 : Internal_Entity_Base_Subp_Body;
Fld_26 : Internal_Entity_Basic_Decl;
Scope_Result_14 : Internal_Entity_Basic_Decl;
Result_Var_14 : Internal_Entity_Basic_Decl;
Fld_27 : Internal_Entity_Defining_Name;
If_Result_2 : Internal_Entity_Defining_Name;
If_Result_3 : Internal_Entity_Defining_Name;
If_Result_4 : Internal_Entity_Defining_Name;
If_Result_5 : Internal_Entity_Defining_Name;
Let_Result : Internal_Entity_Defining_Name;
Scope_Result_15 : Internal_Entity_Defining_Name;
Result_Var_15 : Internal_Entity_Defining_Name;
If_Result_6 : Internal_Entity_Defining_Name;
If_Result_7 : Internal_Entity_Defining_Name;
If_Result_8 : Internal_Entity_Defining_Name;
If_Result_9 : Internal_Entity_Defining_Name;
If_Result_10 : Internal_Entity_Defining_Name;
Dyn_Var_Bind_Result : Internal_Entity_Defining_Name;
Let_Result_1 : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_1800 is
            begin
                     Dec_Ref (Fld_22);
            end Finalizer_Scope_1800;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 676 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 622 '<FieldAccess for enclosing_defining_name at ast.py:1308>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1308
--# expr-start 621 '<Cast to Name at ast.py:1308>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1308
Cast_Expr := Ent; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Name
   then
      Cast_Result := Create_Internal_Entity_Name
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Name;
   end if;
--# expr-done 621
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.enclosing_defining_name
Fld := Libadalang.Implementation.Name_P_Enclosing_Defining_Name (Node => Cast_Result.Node, E_Info => Cast_Result.Info);
--# end
--# expr-done 622
Var_Expr := Fld;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
   --# bind dn Var_Expr
--# expr-start 623 '<FieldAccess for basic_decl at ast.py:1309>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1309
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld_1 := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 623
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Basic_Decl;
end if;
Block_Var := Result_Var; 
--# bind bd Block_Var
--# expr-start 675 '<bind at ast.py:1311>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1311
Bound_Origin := Self; 
--# expr-start 674 '<Cond at ast.py:1311>' If_Result_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1311
Var_Expr_1 := Block_Var;
if Var_Expr_1 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind bd Var_Expr_1
--# expr-start 624 '<IsA ParamSpec at ast.py:1312>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1312
Is_A := Var_Expr_1.Node /= null 
and then Var_Expr_1.Node.Kind in Ada_Param_Spec_Range; 
--# expr-done 624
Scope_Result_1 := Is_A;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
if Result_Var_1 then
   --# expr-start 626 '<IsA SubpDecl, ExprFunction, GenericSubpInternal, BaseTypeDecl at ast.py:1313>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1313
--# expr-start 625 '<FieldAccess for semantic_parent at ast.py:1313>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1313
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.semantic_parent
Fld_2 := Libadalang.Implementation.Ada_Node_P_Semantic_Parent (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 625
Is_A_1 := Fld_2.Node /= null 
and then Fld_2.Node.Kind in Ada_Subp_Decl_Range | Ada_Expr_Function_Range | Ada_Generic_Subp_Internal_Range | Ada_Base_Type_Decl; 
--# expr-done 626
   If_Result := Is_A_1;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 629 '<FieldAccess for defining_name at ast.py:1316>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1316
--# expr-start 628 '<Cast to BasicDecl at ast.py:1316>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1316
--# expr-start 627 '<FieldAccess for semantic_parent at ast.py:1316>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1316
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.semantic_parent
Fld_3 := Libadalang.Implementation.Ada_Node_P_Semantic_Parent (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 627
   if Fld_3.Node = null
      or else Fld_3.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Basic_Decl
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
   else
         Cast_Result_1 := No_Entity_Basic_Decl;
   end if;
--# expr-done 628
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Cast_Result_1.Node, E_Info => Cast_Result_1.Info);
--# end
--# expr-done 629
   If_Result_10 := Fld_4;
else
Var_Expr_2 := Block_Var;
if Var_Expr_2 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind bd Var_Expr_2
--# expr-start 630 '<IsA DiscriminantSpec at ast.py:1318>' Is_A_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1318
Is_A_2 := Var_Expr_2.Node /= null 
and then Var_Expr_2.Node.Kind in Ada_Discriminant_Spec_Range; 
--# expr-done 630
Scope_Result_2 := Is_A_2;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := False;
end if;
if Result_Var_2 then
   --# expr-start 633 '<FieldAccess for defining_name at ast.py:1319>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1319
--# expr-start 632 '<Cast to BasicDecl at ast.py:1319>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1319
--# expr-start 631 '<FieldAccess for semantic_parent at ast.py:1319>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1319
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.semantic_parent
Fld_5 := Libadalang.Implementation.Ada_Node_P_Semantic_Parent (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 631
   if Fld_5.Node = null
      or else Fld_5.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_2 := Create_Internal_Entity_Basic_Decl
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
   else
         Cast_Result_2 := No_Entity_Basic_Decl;
   end if;
--# expr-done 632
      if Cast_Result_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_6 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Cast_Result_2.Node, E_Info => Cast_Result_2.Info);
--# end
--# expr-done 633
   If_Result_9 := Fld_6;
else
Var_Expr_3 := Block_Var;
if Var_Expr_3 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind bd Var_Expr_3
--# expr-start 634 '<IsA ParamSpec at ast.py:1321>' Is_A_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1321
Is_A_3 := Var_Expr_3.Node /= null 
and then Var_Expr_3.Node.Kind in Ada_Param_Spec_Range; 
--# expr-done 634
Scope_Result_3 := Is_A_3;
   --# end
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := False;
end if;
if Result_Var_3 then
   --# expr-start 636 '<IsA AbstractSubpDecl, FormalSubpDecl, NullSubpDecl at ast.py:1322>' Is_A_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1322
--# expr-start 635 '<FieldAccess for semantic_parent at ast.py:1322>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1322
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.semantic_parent
Fld_7 := Libadalang.Implementation.Ada_Node_P_Semantic_Parent (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 635
Is_A_4 := Fld_7.Node /= null 
and then Fld_7.Node.Kind in Ada_Abstract_Subp_Decl_Range | Ada_Formal_Subp_Decl | Ada_Null_Subp_Decl_Range; 
--# expr-done 636
   If_Result_1 := Is_A_4;
else
   If_Result_1 := False;
end if;
if If_Result_1 then
   --# expr-start 639 '<FieldAccess for defining_name at ast.py:1324>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1324
--# expr-start 638 '<Cast to BasicDecl at ast.py:1324>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1324
--# expr-start 637 '<FieldAccess for semantic_parent at ast.py:1324>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1324
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.semantic_parent
Fld_8 := Libadalang.Implementation.Ada_Node_P_Semantic_Parent (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 637
   if Fld_8.Node = null
      or else Fld_8.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_3 := Create_Internal_Entity_Basic_Decl
        (Node => Fld_8.Node,
         Info => Fld_8.Info);
   else
         Cast_Result_3 := No_Entity_Basic_Decl;
   end if;
--# expr-done 638
      if Cast_Result_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_9 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Cast_Result_3.Node, E_Info => Cast_Result_3.Info);
--# end
--# expr-done 639
   If_Result_8 := Fld_9;
else
Var_Expr_4 := Block_Var;
if Var_Expr_4 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind bd Var_Expr_4
--# expr-start 640 '<IsA AbstractSubpDecl at ast.py:1326>' Is_A_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1326
Is_A_5 := Var_Expr_4.Node /= null 
and then Var_Expr_4.Node.Kind in Ada_Abstract_Subp_Decl_Range; 
--# expr-done 640
Scope_Result_4 := Is_A_5;
   --# end
   Result_Var_4 := Scope_Result_4;
else
   Result_Var_4 := False;
end if;
if Result_Var_4 then
   --# expr-start 644 '<FieldAccess for defining_name at ast.py:1327>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1327
--# expr-start 643 '<FieldAccess for primitive_subp_first_type at ast.py:1327>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1327
--# expr-start 642 '<FieldAccess for subp_decl_spec at ast.py:1327>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1327
--# expr-start 641 '<Cast to AbstractSubpDecl at ast.py:1327>' Cast_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1327
Cast_Expr_1 := Block_Var; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Abstract_Subp_Decl_Range
   then
      Cast_Result_4 := Create_Internal_Entity_Abstract_Subp_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_4 := No_Entity_Abstract_Subp_Decl;
   end if;
--# expr-done 641
      if Cast_Result_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicSubpDecl.subp_decl_spec'
Fld_10 := Libadalang.Implementation.Dispatcher_Basic_Subp_Decl_P_Subp_Decl_Spec (Node => Cast_Result_4.Node, E_Info => Cast_Result_4.Info);
--# end
--# expr-done 642
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.primitive_subp_first_type
Fld_11 := Libadalang.Implementation.Base_Subp_Spec_P_Primitive_Subp_First_Type (Node => Fld_10.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_10.Info);
--# end
--# expr-done 643
      if Fld_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_12 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Fld_11.Node, E_Info => Fld_11.Info);
--# end
--# expr-done 644
   If_Result_7 := Fld_12;
else
Var_Expr_5 := Block_Var;
if Var_Expr_5 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind bd Var_Expr_5
--# expr-start 645 '<IsA BasicSubpDecl at ast.py:1330>' Is_A_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1330
Is_A_6 := Var_Expr_5.Node /= null 
and then Var_Expr_5.Node.Kind in Ada_Basic_Subp_Decl; 
--# expr-done 645
Scope_Result_5 := Is_A_6;
   --# end
   Result_Var_5 := Scope_Result_5;
else
   Result_Var_5 := False;
end if;
if Result_Var_5 then
--# expr-start 648 '<FieldAccess for primitive_subp_first_type at ast.py:1331>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1331
--# expr-start 647 '<FieldAccess for subp_decl_spec at ast.py:1331>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1331
--# expr-start 646 '<Cast to BasicSubpDecl at ast.py:1331>' Cast_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1331
Cast_Expr_2 := Block_Var; 
   if Cast_Expr_2.Node = null
      or else Cast_Expr_2.Node.Kind in Ada_Basic_Subp_Decl
   then
      Cast_Result_5 := Create_Internal_Entity_Basic_Subp_Decl
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
   else
         Cast_Result_5 := No_Entity_Basic_Subp_Decl;
   end if;
--# expr-done 646
      if Cast_Result_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicSubpDecl.subp_decl_spec'
Fld_13 := Libadalang.Implementation.Dispatcher_Basic_Subp_Decl_P_Subp_Decl_Spec (Node => Cast_Result_5.Node, E_Info => Cast_Result_5.Info);
--# end
--# expr-done 647
      if Fld_13.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.primitive_subp_first_type
Fld_14 := Libadalang.Implementation.Base_Subp_Spec_P_Primitive_Subp_First_Type (Node => Fld_13.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_13.Info);
--# end
--# expr-done 648
Var_Expr_6 := Fld_14;
if Var_Expr_6 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind prim_typ Var_Expr_6
--# expr-start 649 '<FieldAccess for is_tagged_type at ast.py:1334>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1334
      if Var_Expr_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_tagged_type'
Fld_15 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Tagged_Type (Node => Var_Expr_6.Node, Origin => Bound_Origin, E_Info => Var_Expr_6.Info);
--# end
--# expr-done 649
Var_Expr_7 := Fld_15;
if Var_Expr_7 /= False then
   --# scope-start
   --# bind _ Var_Expr_7
--# expr-start 650 '<FieldAccess for private_completion at ast.py:1335>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1335
      if Var_Expr_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.private_completion
Fld_16 := Libadalang.Implementation.Base_Type_Decl_P_Private_Completion (Node => Var_Expr_6.Node, E_Info => Var_Expr_6.Info);
--# end
--# expr-done 650
Var_Expr_9 := Fld_16;
if Var_Expr_9 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind pc Var_Expr_9
--# expr-start 651 '<FieldAccess for defining_name at ast.py:1336>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1336
      if Var_Expr_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_17 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Var_Expr_9.Node, E_Info => Var_Expr_9.Info);
--# end
--# expr-done 651
Scope_Result_6 := Fld_17;
   --# end
   Result_Var_6 := Scope_Result_6;
else
   Result_Var_6 := No_Entity_Defining_Name;
end if;
Var_Expr_8 := Result_Var_6;
if Var_Expr_8 /= No_Entity_Defining_Name then
   --# scope-start
   --# bind e Var_Expr_8
Scope_Result_7 := Var_Expr_8;
   --# end
   Result_Var_7 := Scope_Result_7;
else
   --# expr-start 652 '<FieldAccess for defining_name at ast.py:1337>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1337
      if Var_Expr_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_18 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Var_Expr_6.Node, E_Info => Var_Expr_6.Info);
--# end
--# expr-done 652
   Result_Var_7 := Fld_18;
end if;
Scope_Result_8 := Result_Var_7;
   --# end
   Result_Var_8 := Scope_Result_8;
else
   Result_Var_8 := No_Entity_Defining_Name;
end if;
Scope_Result_9 := Result_Var_8;
   --# end
   Result_Var_9 := Scope_Result_9;
else
   Result_Var_9 := No_Entity_Defining_Name;
end if;
   If_Result_6 := Result_Var_9;
else
--# expr-start 653 '<Cast to Name at ast.py:1341>' Cast_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1341
Cast_Expr_3 := Ent; 
   if Cast_Expr_3.Node = null
      or else Cast_Expr_3.Node.Kind in Ada_Name
   then
      Cast_Result_6 := Create_Internal_Entity_Name
        (Node => Cast_Expr_3.Node,
         Info => Cast_Expr_3.Info);
   else
         Cast_Result_6 := No_Entity_Name;
   end if;
--# expr-done 653
Var_Expr_11 := Cast_Result_6;
if Var_Expr_11 /= No_Entity_Name then
   --# scope-start
--# expr-start 654 '<FieldAccess for gnat_xref_decl at ast.py:1341>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1341
      if Var_Expr_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.gnat_xref_decl
Fld_19 := Libadalang.Implementation.Name_P_Gnat_Xref_Decl (Node => Var_Expr_11.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr_11.Info);
--# end
--# expr-done 654
Scope_Result_10 := Fld_19;
   --# end
   Result_Var_10 := Scope_Result_10;
else
   Result_Var_10 := No_Entity_Defining_Name;
end if;
Var_Expr_10 := Result_Var_10;
if Var_Expr_10 /= No_Entity_Defining_Name then
   --# scope-start
   --# bind ret Var_Expr_10
--# expr-start 673 '<Let at ast.py:1343>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1343
--# scope-start
--# expr-start 655 '<FieldAccess for basic_decl at ast.py:1343>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1343
      if Var_Expr_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld_20 := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Var_Expr_10.Node, E_Info => Var_Expr_10.Info);
--# end
--# expr-done 655
Dbd := Fld_20; 
--# bind dbd Dbd
--# expr-start 672 '<Cond at ast.py:1343>' If_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1343
--# expr-start 656 '<IsA ParamSpec at ast.py:1344>' Is_A_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1344
Is_A_7 := Dbd.Node /= null 
and then Dbd.Node.Kind in Ada_Param_Spec_Range; 
--# expr-done 656
if Is_A_7 then
   --# expr-start 658 '<FieldAccess for decl_param at ast.py:1345>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1345
--# expr-start 657 '<Cast to ParamSpec at ast.py:1345>' Cast_Result_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1345
Cast_Expr_4 := Dbd; 
   if Cast_Expr_4.Node = null
      or else Cast_Expr_4.Node.Kind in Ada_Param_Spec_Range
   then
      Cast_Result_7 := Create_Internal_Entity_Param_Spec
        (Node => Cast_Expr_4.Node,
         Info => Cast_Expr_4.Info);
   else
         Cast_Result_7 := No_Entity_Param_Spec;
   end if;
--# expr-done 657
      if Cast_Result_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamDecl.decl_param
Fld_21 := Libadalang.Implementation.Base_Formal_Param_Decl_P_Decl_Param (Node => Cast_Result_7.Node, Param => Var_Expr_10, Imprecise_Fallback => Imprecise_Fallback, E_Info => Cast_Result_7.Info);
--# end
--# expr-done 658
   If_Result_5 := Fld_21;
else
--# expr-start 659 '<IsA GenericSubpInternal, GenericPackageInternal at ast.py:1347>' Is_A_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1347
Is_A_8 := Dbd.Node /= null 
and then Dbd.Node.Kind in Ada_Generic_Subp_Internal_Range | Ada_Generic_Package_Internal_Range; 
--# expr-done 659
if Is_A_8 then
--# expr-start 661 '<at at ast.py:1348>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1348
--# expr-start 660 '<FieldAccess for generic_instantiations at ast.py:1348>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1348
      if Dbd.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.generic_instantiations
Fld_22 := Libadalang.Implementation.Ada_Node_P_Generic_Instantiations (Node => Dbd.Node, E_Info => Dbd.Info);
--# end
--# expr-done 660
Get_Result := Get (Fld_22, 0, True); 
--# expr-done 661
Var_Expr_12 := Get_Result;
if Var_Expr_12 /= No_Entity_Generic_Instantiation then
   --# scope-start
   --# bind gi Var_Expr_12
--# expr-start 663 '<FieldAccess for defining_name at ast.py:1349>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1349
--# expr-start 662 '<Cast to BasicDecl at ast.py:1349>' Cast_Result_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1349
Cast_Expr_5 := Var_Expr_12; 
      Cast_Result_8 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_5.Node,
         Info => Cast_Expr_5.Info);
--# expr-done 662
      if Cast_Result_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_23 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Cast_Result_8.Node, E_Info => Cast_Result_8.Info);
--# end
--# expr-done 663
Scope_Result_11 := Fld_23;
   --# end
   Result_Var_11 := Scope_Result_11;
else
   Result_Var_11 := Var_Expr_10;
end if;
   If_Result_4 := Result_Var_11;
else
--# expr-start 664 '<IsA ObjectDecl at ast.py:1353>' Is_A_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1353
Is_A_9 := Dbd.Node /= null 
and then Dbd.Node.Kind in Ada_Object_Decl_Range; 
--# expr-done 664
if Is_A_9 then
--# expr-start 666 '<FieldAccess for public_part_decl at ast.py:1354>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1354
--# expr-start 665 '<Cast to ObjectDecl at ast.py:1354>' Cast_Result_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1354
Cast_Expr_6 := Dbd; 
   if Cast_Expr_6.Node = null
      or else Cast_Expr_6.Node.Kind in Ada_Object_Decl_Range
   then
      Cast_Result_9 := Create_Internal_Entity_Object_Decl
        (Node => Cast_Expr_6.Node,
         Info => Cast_Expr_6.Info);
   else
         Cast_Result_9 := No_Entity_Object_Decl;
   end if;
--# expr-done 665
      if Cast_Result_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ObjectDecl.public_part_decl
Fld_24 := Libadalang.Implementation.Object_Decl_P_Public_Part_Decl (Node => Cast_Result_9.Node, E_Info => Cast_Result_9.Info);
--# end
--# expr-done 666
Var_Expr_14 := Fld_24;
if Var_Expr_14 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind ppd Var_Expr_14
--# expr-start 667 '<FieldAccess for defining_name at ast.py:1355>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1355
      if Var_Expr_14.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_25 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Var_Expr_14.Node, E_Info => Var_Expr_14.Info);
--# end
--# expr-done 667
Scope_Result_12 := Fld_25;
   --# end
   Result_Var_12 := Scope_Result_12;
else
   Result_Var_12 := No_Entity_Defining_Name;
end if;
Var_Expr_13 := Result_Var_12;
if Var_Expr_13 /= No_Entity_Defining_Name then
   --# scope-start
   --# bind e Var_Expr_13
Scope_Result_13 := Var_Expr_13;
   --# end
   Result_Var_13 := Scope_Result_13;
else
   Result_Var_13 := Var_Expr_10;
end if;
   If_Result_3 := Result_Var_13;
else
--# expr-start 668 '<IsA BaseSubpBody at ast.py:1359>' Is_A_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1359
Is_A_10 := Dbd.Node /= null 
and then Dbd.Node.Kind in Ada_Base_Subp_Body; 
--# expr-done 668
if Is_A_10 then
   --# expr-start 671 '<FieldAccess for defining_name at ast.py:1360>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1360
--# expr-start 670 '<FieldAccess for decl_part at ast.py:1360>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1360
--# expr-start 669 '<Cast to BaseSubpBody at ast.py:1360>' Cast_Result_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1360
Cast_Expr_7 := Dbd; 
   if Cast_Expr_7.Node = null
      or else Cast_Expr_7.Node.Kind in Ada_Base_Subp_Body
   then
      Cast_Result_10 := Create_Internal_Entity_Base_Subp_Body
        (Node => Cast_Expr_7.Node,
         Info => Cast_Expr_7.Info);
   else
         Cast_Result_10 := No_Entity_Base_Subp_Body;
   end if;
--# expr-done 669
      if Cast_Result_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.decl_part
Fld_26 := Libadalang.Implementation.Body_Node_P_Decl_Part (Node => Cast_Result_10.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Cast_Result_10.Info);
--# end
--# expr-done 670
Var_Expr_15 := Fld_26;
if Var_Expr_15 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind e Var_Expr_15
Scope_Result_14 := Var_Expr_15;
   --# end
   Result_Var_14 := Scope_Result_14;
else
   Result_Var_14 := Dbd;
end if;
      if Result_Var_14.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_27 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Result_Var_14.Node, E_Info => Result_Var_14.Info);
--# end
--# expr-done 671
   If_Result_2 := Fld_27;
else
   If_Result_2 := Var_Expr_10;
end if;
   If_Result_3 := If_Result_2;
end if;
   If_Result_4 := If_Result_3;
end if;
   If_Result_5 := If_Result_4;
end if;
--# expr-done 672
Let_Result := If_Result_5; 
--# end
--# expr-done 673
Scope_Result_15 := Let_Result;
   --# end
      Finalizer_Scope_1800;
   Result_Var_15 := Scope_Result_15;
else
   Result_Var_15 := No_Entity_Defining_Name;
end if;
   If_Result_6 := Result_Var_15;
end if;
   If_Result_7 := If_Result_6;
end if;
   If_Result_8 := If_Result_7;
end if;
   If_Result_9 := If_Result_8;
end if;
   If_Result_10 := If_Result_9;
end if;
--# expr-done 674
Dyn_Var_Bind_Result := If_Result_10; 
--# expr-done 675
Let_Result_1 := Dyn_Var_Bind_Result; 
--# end
--# expr-done 676
      Property_Result := Let_Result_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1800;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Gnat_Xref;
--# end
--# property-start AdaNode.env_get_real_from_node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1369
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Env_Get_Real_From_Node
  (Node : Bare_Ada_Node
      ; From_Node : Bare_Ada_Node
  )
   return Bare_Ada_Node
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind from_node From_Node
   Property_Result : Bare_Ada_Node;
            procedure Finalizer_Scope_89 with Inline_Always;
      C : Bare_Ada_Node;
Node_Parents : Bare_Ada_Node_Array_Access;
Is_A : Boolean;
Find_Result : Bare_Ada_Node;
Is_Null : Boolean;
Is_A_1 : Boolean;
Cast_Expr : Bare_Ada_Node;
Cast_Result : Bare_Base_Subp_Spec;
As_Entity : Internal_Entity_Base_Subp_Spec;
Is_Null_1 : Boolean;
If_Result : Internal_Entity_Base_Subp_Spec;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Bare_Defining_Name;
Cast_Result_1 : Bare_Ada_Node;
If_Result_1 : Bare_Ada_Node;
If_Result_2 : Bare_Ada_Node;
Let_Result : Bare_Ada_Node;
Is_Null_2 : Boolean;
If_Result_3 : Bare_Ada_Node;
Let_Result_1 : Bare_Ada_Node;
            procedure Finalizer_Scope_89 is
            begin
                     Dec_Ref (Node_Parents);
            end Finalizer_Scope_89;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 690 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 689 '<If at ast.py:1376>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1376
--# expr-start 677 '<IsNull at ast.py:1376>' Is_Null_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1376
Is_Null_2 := From_Node = null; 
--# expr-done 677
if Is_Null_2 then
   If_Result_3 := From_Node;
else
   --# expr-start 688 '<Let at ast.py:1376>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1376
--# scope-start
--# expr-start 680 '<Find at ast.py:1377>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1377
   --# expr-start 678 '<FieldAccess for parents at ast.py:1377>' Node_Parents /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1377
      if From_Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents := Parents (From_Node, True); 
--# expr-done 678
      Find_Result := No_Bare_Ada_Node;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents;
      begin
         for Item_24 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_24
      --# expr-start 679 '<IsA GenericInstantiation, BaseSubpSpec at ast.py:1378>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1378
Is_A := Item_24 /= null 
and then Item_24.Kind in Ada_Generic_Instantiation | Ada_Base_Subp_Spec; 
--# expr-done 679
      if Is_A then
         Find_Result := Item_24;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 680
C := Find_Result; 
--# bind c C
--# expr-start 687 '<Cond at ast.py:1379>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1379
--# expr-start 681 '<IsNull at ast.py:1380>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1380
Is_Null := C = null; 
--# expr-done 681
if Is_Null then
   If_Result_2 := From_Node;
else
--# expr-start 682 '<IsA BaseSubpSpec at ast.py:1383>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1383
Is_A_1 := C /= null 
and then C.Kind in Ada_Base_Subp_Spec; 
--# expr-done 682
if Is_A_1 then
--# expr-start 686 '<FieldAccess for node at ast.py:1384>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1384
--# expr-start 685 '<FieldAccess for name at ast.py:1384>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1384
--# expr-start 684 '<as_bare_entity at ast.py:1384>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1384
--# expr-start 683 '<Cast to BaseSubpSpec at ast.py:1384>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1384
Cast_Expr := C; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_Base_Subp_Spec
   then
      Cast_Result := Cast_Expr;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 683
Is_Null_1 := Cast_Result = null; 
if Is_Null_1 then
   If_Result := No_Entity_Base_Subp_Spec;
else
As_Entity := (Info => No_Entity_Info, Node => Cast_Result); 
   If_Result := As_Entity;
end if;
--# expr-done 684
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.name'
Fld := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Name (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 685
Fld_1 := Fld.Node;
--# expr-done 686
      Cast_Result_1 := Fld_1;
   If_Result_1 := Cast_Result_1;
else
   If_Result_1 := C;
end if;
   If_Result_2 := If_Result_1;
end if;
--# expr-done 687
Let_Result := If_Result_2; 
--# end
--# expr-done 688
   If_Result_3 := Let_Result;
end if;
--# expr-done 689
Let_Result_1 := If_Result_3; 
--# end
--# expr-done 690
      Property_Result := Let_Result_1;
   --# end
      Finalizer_Scope_89;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_89;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Env_Get_Real_From_Node;
--# end
--# property-start AdaNode.entity_no_md /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1391
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Entity_No_Md
  (Node : Bare_Ada_Node
      ; N : Bare_Ada_Node
      ; Rebindings : Env_Rebindings
      ; From_Rebound : Boolean
  )
   return Internal_Entity
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind n N
   --# bind rebindings Rebindings
   --# bind from_rebound From_Rebound
   Property_Result : Internal_Entity;
      New_Struct : Internal_Entity_Info;
Is_Null : Boolean;
If_Result : Internal_Entity_Info;
New_Struct_1 : Internal_Entity;
Let_Result : Internal_Entity;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 695 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 694 '<New AdaNode.entity at ast.py:1397>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1397
--# expr-start 693 '<If at ast.py:1399>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1399
--# expr-start 691 '<IsNull at ast.py:1399>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1399
Is_Null := N = null; 
--# expr-done 691
if Is_Null then
   If_Result := No_Entity_Info;
else
   --# expr-start 692 '<New EntityInfo at ast.py:1399>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1399
New_Struct := (From_Rebound => From_Rebound, Md => No_Metadata, Rebindings => Rebindings); 
--# expr-done 692
   If_Result := New_Struct;
end if;
--# expr-done 693
New_Struct_1 := (Info => If_Result, Node => N); 
--# expr-done 694
Let_Result := New_Struct_1; 
--# end
--# expr-done 695
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Entity_No_Md;
--# end
--# property-start AdaNode.env_mappings /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1407
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Env_Mappings
  (Node : Bare_Ada_Node
      ; Defining_Names : Bare_Defining_Name_List
      ; Value : Bare_Ada_Node
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind defining_names Defining_Names
   --# bind value Value
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_91 with Inline_Always;
            procedure Finalizer_Scope_1809 with Inline_Always;
      Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
Item_25 : Bare_Defining_Name;
Map_Result : Internal_Env_Assoc_Array_Access;
Let_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_91 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_91;
            procedure Finalizer_Scope_1809 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1809;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 700 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 699 '<Map at ast.py:1412>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1412
      declare
         Map_Result_Vec : Internal_Env_Assoc_Vectors.Vector;
      begin
      if Defining_Names = null then
      Map_Result := Create_Internal_Env_Assoc_Array (0);
      else
      declare
         Collection : constant Bare_Defining_Name_List := Defining_Names;
      begin
         for Untyped_Item_25 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_25 := Untyped_Item_25; 
   --# scope-start
               --# bind n Item_25
      --# expr-start 698 '<New EnvAssoc at ast.py:1413>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1413
--# expr-start 696 '<New DesignatedEnv at ast.py:1413>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1413
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 696
--# expr-start 697 '<FieldAccess for name_symbol at ast.py:1413>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1413
      if Item_25 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Item_25);
--# end
--# expr-done 697
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Value); 
--# expr-done 698
         declare
            Item_To_Append : constant Internal_Env_Assoc := New_Struct_1;
         begin
               Inc_Ref (Item_To_Append);
            Internal_Env_Assoc_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_1809;
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Env_Assoc_Array
           (Items_Count => Natural (Internal_Env_Assoc_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Env_Assoc_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Env_Assoc_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Env_Assoc_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 699
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 700
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_91;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_91;
               Finalizer_Scope_1809;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Env_Mappings;
--# end
--# property-start AdaNode.comp_bind /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1417
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Comp_Bind
  (Node : Bare_Ada_Node
      ; Left : Logic_Var
      ; Right : Logic_Var
      ; Origin : Bare_Ada_Node
  )
   return Logic_Equation
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind left Left
   --# bind right Right
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_92 with Inline_Always;
      Bind_Result : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_92 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_92;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 702 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 701 '<Bind at ast.py:1418>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1418
Left.Value := No_Entity;
Entity_Vars.Reset (Left);
Right.Value := No_Entity;
Entity_Vars.Reset (Right);
Bind_Result := Solver.Create_Propagate (Left, Right, Logic_Functor_620'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Origin), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:1418") else null)); 
--# expr-done 701
Let_Result := Bind_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 702
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_92;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_92;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Comp_Bind;
--# end
--# property-start AdaNode.universal_int_bind /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1421
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Universal_Int_Bind
  (Node : Bare_Ada_Node
      ; Type_Var : Logic_Var
      ; Origin : Bare_Ada_Node
  )
   return Logic_Equation
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind type_var Type_Var
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_93 with Inline_Always;
      Fld : Internal_Entity;
Bind_Result : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_93 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_93;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 705 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 704 '<Bind at ast.py:1427>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1427
Type_Var.Value := No_Entity;
Entity_Vars.Reset (Type_Var);
--# expr-start 703 '<FieldAccess for universal_int_type at ast.py:1427>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1427
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.universal_int_type
Fld := Libadalang.Implementation.Ada_Node_P_Universal_Int_Type (Node => Self);
--# end
--# expr-done 703
Bind_Result := Solver.Create_Assign (Type_Var, Fld, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:1427") else null)); 
--# expr-done 704
Let_Result := Bind_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 705
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_93;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_93;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Universal_Int_Bind;
--# end
--# property-start AdaNode.universal_real_bind /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1430
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Universal_Real_Bind
  (Node : Bare_Ada_Node
      ; Type_Var : Logic_Var
      ; Origin : Bare_Ada_Node
  )
   return Logic_Equation
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind type_var Type_Var
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_94 with Inline_Always;
      Fld : Internal_Entity;
Bind_Result : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_94 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_94;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 708 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 707 '<Bind at ast.py:1437>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1437
Type_Var.Value := No_Entity;
Entity_Vars.Reset (Type_Var);
--# expr-start 706 '<FieldAccess for universal_real_type at ast.py:1437>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1437
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.universal_real_type
Fld := Libadalang.Implementation.Ada_Node_P_Universal_Real_Type (Node => Self);
--# end
--# expr-done 706
Bind_Result := Solver.Create_Assign (Type_Var, Fld, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:1437") else null)); 
--# expr-done 707
Let_Result := Bind_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 708
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_94;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_94;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Universal_Real_Bind;
--# end
--# property-start AdaNode.origin_node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1440
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Origin_Node
  (Node : Bare_Ada_Node
  )
   return Bare_Ada_Node
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node;
      Fld : Boolean;
If_Result : Bare_Ada_Node;
Let_Result : Bare_Ada_Node;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 711 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 710 '<If at ast.py:1445>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1445
--# expr-start 709 '<FieldAccess for in_contract at ast.py:1445>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1445
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.in_contract
Fld := Libadalang.Implementation.Ada_Node_P_In_Contract (Node => Self);
--# end
--# expr-done 709
if Fld then
   If_Result := No_Bare_Ada_Node;
else
   If_Result := Self;
end if;
--# expr-done 710
Let_Result := If_Result; 
--# end
--# expr-done 711
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Origin_Node;
--# end
--# property-start AdaNode.env_hook /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1448
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Env_Hook
  (Node : Bare_Ada_Node
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Match_0 : Bare_Library_Item;
Match_1 : Bare_Subunit;
Match_2 : Bare_Ada_Node;
Match_0_1 : Bare_Body_Node;
Match_1_1 : Bare_Basic_Decl;
Match_2_1 : Bare_Ada_Node;
Fld : Bare_Ada_Node;
Fld_1 : Boolean;
Fld_2 : Boolean;
Match_Prefix : Bare_Ada_Node;
Cast_Result : Bare_Body_Node;
Let_Result : Boolean;
Scope_Result : Boolean;
Cast_Result_1 : Bare_Basic_Decl;
Let_Result_1 : Boolean;
Scope_Result_1 : Boolean;
Cast_Result_2 : Bare_Ada_Node;
Let_Result_2 : Boolean;
Scope_Result_2 : Boolean;
Match_Result : Boolean;
Fld_3 : Boolean;
Match_Prefix_1 : Bare_Ada_Node;
Cast_Result_3 : Bare_Library_Item;
Let_Result_3 : Boolean;
Scope_Result_3 : Boolean;
Cast_Result_4 : Bare_Subunit;
Let_Result_4 : Boolean;
Scope_Result_4 : Boolean;
Cast_Result_5 : Bare_Ada_Node;
Let_Result_5 : Boolean;
Scope_Result_5 : Boolean;
Match_Result_1 : Boolean;
Let_Result_6 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 718 '<Block at ???>' Let_Result_6 None
--# scope-start
--# expr-start 717 '<Match at ast.py:1455>' Match_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1455
--# expr-start 712 '<parent at ast.py:1455>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1455
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 712
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix_1 := Fld; 
case Ada_Ada_Node (Match_Prefix_1.Kind) is
         when Ada_Library_Item =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Match_Prefix_1;
Match_0 := Cast_Result_3; 
--# bind _ Match_0
--# expr-start 715 '<Match at ast.py:1456>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1456
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Self; 
case Ada_Ada_Node (Match_Prefix.Kind) is
         when Ada_Expr_Function .. Ada_Task_Body =>
   --# scope-start
--# scope-start
      Cast_Result := Match_Prefix;
Match_0_1 := Cast_Result; 
--# bind b Match_0_1
--# expr-start 713 '<FieldAccess for env_hook_body at ast.py:1457>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1457
      if Match_0_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.env_hook_body
Fld_1 := Libadalang.Implementation.Body_Node_P_Env_Hook_Body (Node => Match_0_1);
--# end
--# expr-done 713
Let_Result := Fld_1; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Abstract_State_Decl .. Ada_Synthetic_Subp_Decl | Ada_Entry_Index_Spec .. Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Match_Prefix;
Match_1_1 := Cast_Result_1; 
--# bind bd Match_1_1
--# expr-start 714 '<FieldAccess for env_hook_basic_decl at ast.py:1458>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1458
      if Match_1_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.env_hook_basic_decl
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Env_Hook_Basic_Decl (Node => Match_1_1);
--# end
--# expr-done 714
Let_Result_1 := Fld_2; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abort_Absent .. Ada_Param_Assoc | Ada_Case_Stmt_Alternative .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Match_Prefix;
Match_2_1 := Cast_Result_2; 
--# bind _ Match_2_1
Let_Result_2 := False; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 715
Let_Result_3 := Match_Result; 
--# end
Scope_Result_3 := Let_Result_3;
   --# end
            Match_Result_1 := Scope_Result_3; 
         when Ada_Subunit =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Match_Prefix_1;
Match_1 := Cast_Result_4; 
--# bind su Match_1
--# expr-start 716 '<FieldAccess for env_hook_subunit at ast.py:1461>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1461
      if Match_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Subunit.env_hook_subunit
Fld_3 := Libadalang.Implementation.Subunit_P_Env_Hook_Subunit (Node => Match_1);
--# end
--# expr-done 716
Let_Result_4 := Fld_3; 
--# end
Scope_Result_4 := Let_Result_4;
   --# end
            Match_Result_1 := Scope_Result_4; 
         when Ada_Abort_Absent .. Ada_Iter_Type_Of | Ada_Limited_Absent .. Ada_Subp_Kind_Procedure | Ada_Synchronized_Absent .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Match_Prefix_1;
Match_2 := Cast_Result_5; 
--# bind _ Match_2
Let_Result_5 := False; 
--# end
Scope_Result_5 := Let_Result_5;
   --# end
            Match_Result_1 := Scope_Result_5; 
end case;
--# expr-done 717
Let_Result_6 := Match_Result_1; 
--# end
--# expr-done 718
      Property_Result := Let_Result_6;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Env_Hook;
--# end
--# property-start AdaNode.env_get /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1466
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Env_Get
  (Node : Bare_Ada_Node
      ; Env : Lexical_Env
      ; Symbol : Symbol_Type
      ; Lookup : Lookup_Kind
         := Recursive
      ; From_Node : Bare_Ada_Node
         := No_Bare_Ada_Node
      ; Categories : Ref_Categories
         := (Inherited_Primitives => True, Nocat => True)
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind symbol Symbol
   --# bind lookup Lookup
   --# bind from_node From_Node
   --# bind categories Categories
   Property_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_97 with Inline_Always;
      Block_Var : Bare_Ada_Node;
Block_Var_1 : Internal_Entity_Array_Access;
Block_Var_2 : Bare_Basic_Decl;
Var_Expr : Bare_Basic_Decl;
Var_Expr_1 : Bare_Defining_Name;
Var_Expr_2 : Internal_Entity_Defining_Name;
Var_Expr_3 : Internal_Entity_Basic_Decl;
Fld : Bare_Ada_Node;
Env_Get_Result : Internal_Entity_Array_Access;
Cast_Expr : Bare_Ada_Node;
Cast_Result : Bare_Basic_Decl;
Scope_Result : Bare_Basic_Decl;
Cast_Expr_1 : Bare_Ada_Node;
Cast_Result_1 : Bare_Defining_Name;
Fld_1 : Bare_Ada_Node;
Fld_2 : Bare_Ada_Node;
Cast_Result_2 : Bare_Basic_Decl;
Scope_Result_1 : Bare_Basic_Decl;
Result_Var : Bare_Basic_Decl;
Result_Var_1 : Bare_Basic_Decl;
Fld_3 : Bare_Ada_Node;
Cast_Expr_2 : Bare_Basic_Decl;
Cast_Result_3 : Bare_Ada_Node;
Is_Equal : Boolean;
Not_Val : Boolean;
Map_Result : Internal_Entity_Array_Access;
As_Entity : Internal_Entity_Basic_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Decl;
Fld_4 : Internal_Entity_Defining_Name;
Scope_Result_2 : Internal_Entity_Defining_Name;
Result_Var_2 : Internal_Entity_Defining_Name;
Fld_5 : Boolean;
Scope_Result_3 : Boolean;
Result_Var_3 : Boolean;
If_Result_1 : Internal_Entity_Array_Access;
Let_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_97 is
            begin
                     Dec_Ref (Env_Get_Result);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Map_Result);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_97;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 734 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 719 '<FieldAccess for env_get_real_from_node at ast.py:1489>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1489
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.env_get_real_from_node
Fld := Libadalang.Implementation.Ada_Node_P_Env_Get_Real_From_Node (Node => Self, From_Node => From_Node);
--# end
--# expr-done 719
Block_Var := Fld; 
--# bind real_from_node Block_Var
--# expr-start 720 '<EnvGet at ast.py:1490>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1490
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Env, Key => Symbol, Lookup_Kind => To_Lookup_Kind_Type (Lookup), Categories => Categories, From => Block_Var)); 
--# expr-done 720
Block_Var_1 := Env_Get_Result; Inc_Ref (Block_Var_1);
--# bind results Block_Var_1
--# expr-start 721 '<Cast to BasicDecl at ast.py:1494>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1494
Cast_Expr := Block_Var; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_Basic_Decl
   then
      Cast_Result := Cast_Expr;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 721
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind e Var_Expr
Scope_Result := Var_Expr;
   --# end
   Result_Var_1 := Scope_Result;
else
--# expr-start 722 '<Cast to DefiningName at ast.py:1495>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1495
Cast_Expr_1 := Block_Var; 
   if Cast_Expr_1 = null
      or else Cast_Expr_1.Kind in Ada_Defining_Name_Range
   then
      Cast_Result_1 := Cast_Expr_1;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 722
Var_Expr_1 := Cast_Result_1;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind name Var_Expr_1
--# expr-start 725 '<Cast to BasicDecl at ast.py:1496>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1496
--# expr-start 724 '<parent at ast.py:1496>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1496
--# expr-start 723 '<parent at ast.py:1496>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1496
      if Var_Expr_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Var_Expr_1.Parent; 
--# expr-done 723
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Fld_1.Parent; 
--# expr-done 724
   if Fld_2 = null
      or else Fld_2.Kind in Ada_Basic_Decl
   then
      Cast_Result_2 := Fld_2;
   else
         Cast_Result_2 := No_Bare_Ada_Node;
   end if;
--# expr-done 725
Scope_Result_1 := Cast_Result_2;
   --# end
   Result_Var := Scope_Result_1;
else
   Result_Var := No_Bare_Ada_Node;
end if;
   Result_Var_1 := Result_Var;
end if;
Block_Var_2 := Result_Var_1; 
--# bind enclosing_bd Block_Var_2
--# expr-start 733 '<If at ast.py:1500>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1500
--# expr-start 726 '<as_bare_entity at ast.py:1504>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1504
Is_Null := Block_Var_2 = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Decl;
else
As_Entity := (Info => No_Entity_Info, Node => Block_Var_2); 
   If_Result := As_Entity;
end if;
--# expr-done 726
Var_Expr_3 := If_Result;
if Var_Expr_3 /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 727 '<FieldAccess for defining_name at ast.py:1504>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1504
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Var_Expr_3.Node, E_Info => Var_Expr_3.Info);
--# end
--# expr-done 727
Scope_Result_2 := Fld_4;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Entity_Defining_Name;
end if;
Var_Expr_2 := Result_Var_2;
if Var_Expr_2 /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 728 '<FieldAccess for name_is at ast.py:1504>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1504
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_is
Fld_5 := Libadalang.Implementation.Name_P_Name_Is (Node => Var_Expr_2.Node, Sym => Symbol);
--# end
--# expr-done 728
Scope_Result_3 := Fld_5;
   --# end
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := False;
end if;
if Result_Var_3 then
   --# expr-start 732 '<Filter at ast.py:1505>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1505
      declare
         Map_Result_Vec : Internal_Entity_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Block_Var_1;
      begin
         for Item_26 of
               Collection.Items
         loop
   --# scope-start
               --# bind r Item_26
         --# expr-start 731 '<Not at ast.py:1505>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1505
--# expr-start 730 '<Eq at ast.py:1505>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1505
--# expr-start 729 '<FieldAccess for node at ast.py:1505>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1505
Fld_3 := Item_26.Node;
--# expr-done 729
Cast_Expr_2 := Block_Var_2; 
      Cast_Result_3 := Cast_Expr_2;
Is_Equal := Fld_3 = Cast_Result_3; 
--# expr-done 730
Not_Val := not (Is_Equal); 
--# expr-done 731
         if Not_Val then
         declare
            Item_To_Append : constant Internal_Entity := Item_26;
         begin
            Internal_Entity_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 732
   If_Result_1 := Map_Result;
else
   If_Result_1 := Block_Var_1;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 733
Let_Result := If_Result_1; Inc_Ref (Let_Result);
--# end
--# expr-done 734
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_97;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_97;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Env_Get;
--# end
--# property-start AdaNode.env_get_public /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1510
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Env_Get_Public
  (Node : Bare_Ada_Node
      ; Env : Lexical_Env
      ; Symbol : Symbol_Type
      ; Lookup : Lookup_Kind
         := Recursive
      ; From_Node : Bare_Ada_Node
         := No_Bare_Ada_Node
      ; Categories : Ref_Categories
         := (Inherited_Primitives => True, Nocat => True)
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind symbol Symbol
   --# bind lookup Lookup
   --# bind from_node From_Node
   --# bind categories Categories
   Property_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_98 with Inline_Always;
      Var_Expr : Internal_Entity_Package_Decl;
Fld : Internal_Entity_Array_Access;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Package_Decl;
Fld_1 : Symbol_Type;
Sym : Symbol_Type;
Is_Equal : Boolean;
Not_Val : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Map_Result : Internal_Entity_Array_Access;
Let_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_98 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_98;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 742 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 741 '<Filter at ast.py:1522>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1522
   --# expr-start 735 '<FieldAccess for env_get at ast.py:1522>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1522
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.env_get
Fld := Libadalang.Implementation.Ada_Node_P_Env_Get (Node => Self, Env => Env, Symbol => Symbol, Lookup => Lookup, From_Node => From_Node, Categories => Categories);
--# end
--# expr-done 735
      declare
         Map_Result_Vec : Internal_Entity_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Fld;
      begin
         for Item_27 of
               Collection.Items
         loop
   --# scope-start
               --# bind x Item_27
--# expr-start 736 '<Cast to PackageDecl at ast.py:1523>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1523
Cast_Expr := Item_27; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Package_Decl_Range
   then
      Cast_Result := Create_Internal_Entity_Package_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Package_Decl;
   end if;
--# expr-done 736
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Package_Decl then
   --# scope-start
   --# bind pkg Var_Expr
--# expr-start 740 '<Not at ast.py:1524>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1524
--# expr-start 739 '<Eq at ast.py:1524>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1524
--# expr-start 737 '<FieldAccess for name_symbol at ast.py:1524>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1524
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Var_Expr.Node);
--# end
--# expr-done 737
--# expr-start 738 '<Symbol root_types_ at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Root_Types); 
--# expr-done 738
Is_Equal := Fld_1 = Sym; 
--# expr-done 739
Not_Val := not (Is_Equal); 
--# expr-done 740
Scope_Result := Not_Val;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := True;
end if;
         if Result_Var then
         declare
            Item_To_Append : constant Internal_Entity := Item_27;
         begin
            Internal_Entity_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 741
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 742
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_98;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_98;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Env_Get_Public;
--# end
--# property-start AdaNode.synthesize_defining_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1531
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Synthesize_Defining_Name
  (Node : Bare_Ada_Node
      ; Sym : Symbol_Type
  )
   return Bare_Defining_Name
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind sym Sym
   Property_Result : Bare_Defining_Name;
      New_Node : Bare_Synthetic_Identifier;
Cast_Result : Bare_Name;
New_Node_1 : Bare_Synthetic_Defining_Name;
Let_Result : Bare_Synthetic_Defining_Name;
Cast_Result_1 : Bare_Defining_Name;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Ada_Node_Ada_Node_P_Synthesize_Defining_Name,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Ada_Node,
                                As_Bare_Ada_Node => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Symbol_Type,
                                      As_Symbol_Type => Sym);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Bare_Defining_Name;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
--# expr-start 745 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 744 '<New SyntheticDefiningName at ast.py:1536>' New_Node_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1536
--# expr-start 743 '<New SyntheticIdentifier at ast.py:1537>' New_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1537
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node := new Root_Node_Record
  (Ada_Synthetic_Identifier);
Initialize
  (Self => New_Node,
   Kind => Ada_Synthetic_Identifier,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node);
      New_Node.Expr_Logic_Vars :=
         System.Null_Address;
      New_Node.Synthetic_Identifier_Sym :=
         Sym;
--# expr-done 743
      Cast_Result := New_Node;
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node_1 := new Root_Node_Record
  (Ada_Synthetic_Defining_Name);
Initialize
  (Self => New_Node_1,
   Kind => Ada_Synthetic_Defining_Name,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node_1);
   Initialize_Fields_For_Synthetic_Defining_Name
     (Self => New_Node_1,
      Defining_Name_F_Name => Cast_Result);
      New_Node_1.Expr_Logic_Vars :=
         System.Null_Address;
--# expr-done 744
Let_Result := New_Node_1; 
--# end
--# expr-done 745
      Cast_Result_1 := Let_Result;
      Property_Result := Cast_Result_1;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Bare_Defining_Name,
                        As_Bare_Defining_Name => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Synthesize_Defining_Name;
--# end
--# property-start AdaNode.create_unop_assoc /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1545
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Create_Unop_Assoc
  (Node : Bare_Ada_Node
      ; Op : Symbol_Type
      ; Rhs : Bare_Base_Type_Decl
      ; Ret : Bare_Base_Type_Decl
  )
   return Internal_Env_Assoc
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind op Op
   --# bind rhs Rhs
   --# bind ret Ret
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_100 with Inline_Always;
      Sym : Symbol_Type;
New_Node : Bare_Synthetic_Type_Expr;
Cast_Result : Bare_Type_Expr;
New_Node_1 : Bare_Synthetic_Formal_Param_Decl;
New_Node_2 : Bare_Synthetic_Type_Expr;
New_Node_3 : Bare_Predefined_Un_Op_Spec;
Cast_Result_1 : Bare_Base_Subp_Spec;
New_Node_4 : Bare_Synthetic_Subp_Decl;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Result_2 : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
Let_Result : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_100 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_100;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Ada_Node_Ada_Node_P_Create_Unop_Assoc,
            Items    => new Mmz_Key_Array (1 ..  4))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Ada_Node,
                                As_Bare_Ada_Node => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Symbol_Type,
                                      As_Symbol_Type => Op);
               Mmz_K.Items (3) := (Kind => Mmz_Bare_Base_Type_Decl,
                                      As_Bare_Base_Type_Decl => Rhs);
               Mmz_K.Items (4) := (Kind => Mmz_Bare_Base_Type_Decl,
                                      As_Bare_Base_Type_Decl => Ret);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Env_Assoc;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 754 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 753 '<New EnvAssoc at ast.py:1551>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1551
--# expr-start 746 '<New DesignatedEnv at ast.py:1551>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1551
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 746
--# expr-start 752 '<New SyntheticSubpDecl at ast.py:1553>' New_Node_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1553
--# expr-start 751 '<New PredefinedUnOpSpec at ast.py:1553>' New_Node_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1553
--# expr-start 747 '<New SyntheticTypeExpr at ast.py:1559>' New_Node_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1559
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node_2 := new Root_Node_Record
  (Ada_Synthetic_Type_Expr);
Initialize
  (Self => New_Node_2,
   Kind => Ada_Synthetic_Type_Expr,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node_2);
   Initialize_Fields_For_Synthetic_Type_Expr
     (Self => New_Node_2,
      Synthetic_Type_Expr_F_Target_Type => Ret);
--# expr-done 747
--# expr-start 750 '<New SyntheticFormalParamDecl at ast.py:1555>' New_Node_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1555
--# expr-start 748 '<New SyntheticTypeExpr at ast.py:1557>' New_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1557
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node := new Root_Node_Record
  (Ada_Synthetic_Type_Expr);
Initialize
  (Self => New_Node,
   Kind => Ada_Synthetic_Type_Expr,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node);
   Initialize_Fields_For_Synthetic_Type_Expr
     (Self => New_Node,
      Synthetic_Type_Expr_F_Target_Type => Rhs);
--# expr-done 748
      Cast_Result := New_Node;
--# expr-start 749 '<Symbol right at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Right); 
--# expr-done 749
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node_1 := new Root_Node_Record
  (Ada_Synthetic_Formal_Param_Decl);
Initialize
  (Self => New_Node_1,
   Kind => Ada_Synthetic_Formal_Param_Decl,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node_1);
   Initialize_Fields_For_Synthetic_Formal_Param_Decl
     (Self => New_Node_1,
      Synthetic_Formal_Param_Decl_F_Param_Type => Cast_Result);
      New_Node_1.Synthetic_Formal_Param_Decl_Param_Name :=
         Sym;
--# expr-done 750
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node_3 := new Root_Node_Record
  (Ada_Predefined_Un_Op_Spec);
Initialize
  (Self => New_Node_3,
   Kind => Ada_Predefined_Un_Op_Spec,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node_3);
   Initialize_Fields_For_Predefined_Un_Op_Spec
     (Self => New_Node_3,
      Predefined_Un_Op_Spec_F_Return_Type_Expr => New_Node_2, Predefined_Un_Op_Spec_F_Right_Param => New_Node_1);
      New_Node_3.Predefined_Un_Op_Spec_Op_Symbol :=
         Op;
--# expr-done 751
      Cast_Result_1 := New_Node_3;
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node_4 := new Root_Node_Record
  (Ada_Synthetic_Subp_Decl);
Initialize
  (Self => New_Node_4,
   Kind => Ada_Synthetic_Subp_Decl,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node_4);
   Initialize_Fields_For_Synthetic_Subp_Decl
     (Self => New_Node_4,
      Synthetic_Subp_Decl_F_Spec => Cast_Result_1);
--# expr-done 752
      Cast_Result_2 := New_Node_4;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Op, Metadata => No_Metadata, Value => Cast_Result_2); 
--# expr-done 753
Let_Result := New_Struct_1; Inc_Ref (Let_Result);
--# end
--# expr-done 754
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_100;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Env_Assoc,
                        As_Internal_Env_Assoc => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_100;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Create_Unop_Assoc;
--# end
--# property-start AdaNode.create_binop_assoc_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1564
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Create_Binop_Assoc_Impl
  (Node : Bare_Ada_Node
      ; Op : Symbol_Type
      ; Lhs : Bare_Type_Expr
      ; Rhs : Bare_Type_Expr
      ; Ret : Bare_Type_Expr
  )
   return Internal_Env_Assoc
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind op Op
   --# bind lhs Lhs
   --# bind rhs Rhs
   --# bind ret Ret
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_101 with Inline_Always;
      Sym : Symbol_Type;
New_Node : Bare_Synthetic_Formal_Param_Decl;
Sym_1 : Symbol_Type;
New_Node_1 : Bare_Synthetic_Formal_Param_Decl;
New_Node_2 : Bare_Predefined_Bin_Op_Spec;
Cast_Result : Bare_Base_Subp_Spec;
New_Node_3 : Bare_Synthetic_Subp_Decl;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Result_1 : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
Let_Result : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_101 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_101;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Ada_Node_Ada_Node_P_Create_Binop_Assoc_Impl,
            Items    => new Mmz_Key_Array (1 ..  5))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Ada_Node,
                                As_Bare_Ada_Node => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Symbol_Type,
                                      As_Symbol_Type => Op);
               Mmz_K.Items (3) := (Kind => Mmz_Bare_Type_Expr,
                                      As_Bare_Type_Expr => Lhs);
               Mmz_K.Items (4) := (Kind => Mmz_Bare_Type_Expr,
                                      As_Bare_Type_Expr => Rhs);
               Mmz_K.Items (5) := (Kind => Mmz_Bare_Type_Expr,
                                      As_Bare_Type_Expr => Ret);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Env_Assoc;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 763 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 762 '<New EnvAssoc at ast.py:1571>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1571
--# expr-start 755 '<New DesignatedEnv at ast.py:1571>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1571
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 755
--# expr-start 761 '<New SyntheticSubpDecl at ast.py:1573>' New_Node_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1573
--# expr-start 760 '<New PredefinedBinOpSpec at ast.py:1573>' New_Node_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1573
--# expr-start 757 '<New SyntheticFormalParamDecl at ast.py:1575>' New_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1575
--# expr-start 756 '<Symbol left at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Left); 
--# expr-done 756
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node := new Root_Node_Record
  (Ada_Synthetic_Formal_Param_Decl);
Initialize
  (Self => New_Node,
   Kind => Ada_Synthetic_Formal_Param_Decl,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node);
   Initialize_Fields_For_Synthetic_Formal_Param_Decl
     (Self => New_Node,
      Synthetic_Formal_Param_Decl_F_Param_Type => Lhs);
      New_Node.Synthetic_Formal_Param_Decl_Param_Name :=
         Sym;
--# expr-done 757
--# expr-start 759 '<New SyntheticFormalParamDecl at ast.py:1579>' New_Node_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1579
--# expr-start 758 '<Symbol right at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Right); 
--# expr-done 758
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node_1 := new Root_Node_Record
  (Ada_Synthetic_Formal_Param_Decl);
Initialize
  (Self => New_Node_1,
   Kind => Ada_Synthetic_Formal_Param_Decl,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node_1);
   Initialize_Fields_For_Synthetic_Formal_Param_Decl
     (Self => New_Node_1,
      Synthetic_Formal_Param_Decl_F_Param_Type => Rhs);
      New_Node_1.Synthetic_Formal_Param_Decl_Param_Name :=
         Sym_1;
--# expr-done 759
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node_2 := new Root_Node_Record
  (Ada_Predefined_Bin_Op_Spec);
Initialize
  (Self => New_Node_2,
   Kind => Ada_Predefined_Bin_Op_Spec,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node_2);
   Initialize_Fields_For_Predefined_Bin_Op_Spec
     (Self => New_Node_2,
      Predefined_Bin_Op_Spec_F_Left_Param => New_Node, Predefined_Bin_Op_Spec_F_Return_Type_Expr => Ret, Predefined_Bin_Op_Spec_F_Right_Param => New_Node_1);
      New_Node_2.Predefined_Bin_Op_Spec_Op_Symbol :=
         Op;
--# expr-done 760
      Cast_Result := New_Node_2;
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node_3 := new Root_Node_Record
  (Ada_Synthetic_Subp_Decl);
Initialize
  (Self => New_Node_3,
   Kind => Ada_Synthetic_Subp_Decl,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node_3);
   Initialize_Fields_For_Synthetic_Subp_Decl
     (Self => New_Node_3,
      Synthetic_Subp_Decl_F_Spec => Cast_Result);
--# expr-done 761
      Cast_Result_1 := New_Node_3;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Op, Metadata => No_Metadata, Value => Cast_Result_1); 
--# expr-done 762
Let_Result := New_Struct_1; Inc_Ref (Let_Result);
--# end
--# expr-done 763
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_101;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Env_Assoc,
                        As_Internal_Env_Assoc => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_101;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Create_Binop_Assoc_Impl;
--# end
--# property-start AdaNode.create_binop_assoc /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1588
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Create_Binop_Assoc
  (Node : Bare_Ada_Node
      ; Op : Symbol_Type
      ; Lhs : Bare_Base_Type_Decl
      ; Rhs : Bare_Base_Type_Decl
      ; Ret : Bare_Base_Type_Decl
  )
   return Internal_Env_Assoc
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind op Op
   --# bind lhs Lhs
   --# bind rhs Rhs
   --# bind ret Ret
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_102 with Inline_Always;
      New_Node : Bare_Synthetic_Type_Expr;
Cast_Result : Bare_Type_Expr;
New_Node_1 : Bare_Synthetic_Type_Expr;
Cast_Result_1 : Bare_Type_Expr;
New_Node_2 : Bare_Synthetic_Type_Expr;
Cast_Result_2 : Bare_Type_Expr;
Fld : Internal_Env_Assoc := No_Env_Assoc;
Let_Result : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_102 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_102;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Ada_Node_Ada_Node_P_Create_Binop_Assoc,
            Items    => new Mmz_Key_Array (1 ..  5))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Ada_Node,
                                As_Bare_Ada_Node => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Symbol_Type,
                                      As_Symbol_Type => Op);
               Mmz_K.Items (3) := (Kind => Mmz_Bare_Base_Type_Decl,
                                      As_Bare_Base_Type_Decl => Lhs);
               Mmz_K.Items (4) := (Kind => Mmz_Bare_Base_Type_Decl,
                                      As_Bare_Base_Type_Decl => Rhs);
               Mmz_K.Items (5) := (Kind => Mmz_Bare_Base_Type_Decl,
                                      As_Bare_Base_Type_Decl => Ret);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Env_Assoc;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 768 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 767 '<FieldAccess for create_binop_assoc_impl at ast.py:1596>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1596
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 764 '<New SyntheticTypeExpr at ast.py:1598>' New_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1598
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node := new Root_Node_Record
  (Ada_Synthetic_Type_Expr);
Initialize
  (Self => New_Node,
   Kind => Ada_Synthetic_Type_Expr,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node);
   Initialize_Fields_For_Synthetic_Type_Expr
     (Self => New_Node,
      Synthetic_Type_Expr_F_Target_Type => Lhs);
--# expr-done 764
      Cast_Result := New_Node;
--# expr-start 765 '<New SyntheticTypeExpr at ast.py:1599>' New_Node_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1599
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node_1 := new Root_Node_Record
  (Ada_Synthetic_Type_Expr);
Initialize
  (Self => New_Node_1,
   Kind => Ada_Synthetic_Type_Expr,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node_1);
   Initialize_Fields_For_Synthetic_Type_Expr
     (Self => New_Node_1,
      Synthetic_Type_Expr_F_Target_Type => Rhs);
--# expr-done 765
      Cast_Result_1 := New_Node_1;
--# expr-start 766 '<New SyntheticTypeExpr at ast.py:1600>' New_Node_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1600
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node_2 := new Root_Node_Record
  (Ada_Synthetic_Type_Expr);
Initialize
  (Self => New_Node_2,
   Kind => Ada_Synthetic_Type_Expr,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node_2);
   Initialize_Fields_For_Synthetic_Type_Expr
     (Self => New_Node_2,
      Synthetic_Type_Expr_F_Target_Type => Ret);
--# expr-done 766
      Cast_Result_2 := New_Node_2;
--# property-call-start AdaNode.create_binop_assoc_impl
Fld := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc_Impl (Node => Self, Op => Op, Lhs => Cast_Result, Rhs => Cast_Result_1, Ret => Cast_Result_2);
--# end
--# expr-done 767
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 768
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_102;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Env_Assoc,
                        As_Internal_Env_Assoc => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_102;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Create_Binop_Assoc;
--# end
--# property-start AdaNode.create_binop_assoc_l_expr /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1604
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Create_Binop_Assoc_L_Expr
  (Node : Bare_Ada_Node
      ; Op : Symbol_Type
      ; Lhs : Bare_Type_Expr
      ; Rhs : Bare_Base_Type_Decl
      ; Ret : Bare_Base_Type_Decl
  )
   return Internal_Env_Assoc
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind op Op
   --# bind lhs Lhs
   --# bind rhs Rhs
   --# bind ret Ret
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_103 with Inline_Always;
      New_Node : Bare_Synthetic_Type_Expr;
Cast_Result : Bare_Type_Expr;
New_Node_1 : Bare_Synthetic_Type_Expr;
Cast_Result_1 : Bare_Type_Expr;
Fld : Internal_Env_Assoc := No_Env_Assoc;
Let_Result : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_103 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_103;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Ada_Node_Ada_Node_P_Create_Binop_Assoc_L_Expr,
            Items    => new Mmz_Key_Array (1 ..  5))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Ada_Node,
                                As_Bare_Ada_Node => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Symbol_Type,
                                      As_Symbol_Type => Op);
               Mmz_K.Items (3) := (Kind => Mmz_Bare_Type_Expr,
                                      As_Bare_Type_Expr => Lhs);
               Mmz_K.Items (4) := (Kind => Mmz_Bare_Base_Type_Decl,
                                      As_Bare_Base_Type_Decl => Rhs);
               Mmz_K.Items (5) := (Kind => Mmz_Bare_Base_Type_Decl,
                                      As_Bare_Base_Type_Decl => Ret);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Env_Assoc;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 772 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 771 '<FieldAccess for create_binop_assoc_impl at ast.py:1611>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1611
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 769 '<New SyntheticTypeExpr at ast.py:1614>' New_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1614
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node := new Root_Node_Record
  (Ada_Synthetic_Type_Expr);
Initialize
  (Self => New_Node,
   Kind => Ada_Synthetic_Type_Expr,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node);
   Initialize_Fields_For_Synthetic_Type_Expr
     (Self => New_Node,
      Synthetic_Type_Expr_F_Target_Type => Rhs);
--# expr-done 769
      Cast_Result := New_Node;
--# expr-start 770 '<New SyntheticTypeExpr at ast.py:1615>' New_Node_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1615
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node_1 := new Root_Node_Record
  (Ada_Synthetic_Type_Expr);
Initialize
  (Self => New_Node_1,
   Kind => Ada_Synthetic_Type_Expr,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node_1);
   Initialize_Fields_For_Synthetic_Type_Expr
     (Self => New_Node_1,
      Synthetic_Type_Expr_F_Target_Type => Ret);
--# expr-done 770
      Cast_Result_1 := New_Node_1;
--# property-call-start AdaNode.create_binop_assoc_impl
Fld := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc_Impl (Node => Self, Op => Op, Lhs => Lhs, Rhs => Cast_Result, Ret => Cast_Result_1);
--# end
--# expr-done 771
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 772
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_103;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Env_Assoc,
                        As_Internal_Env_Assoc => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_103;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Create_Binop_Assoc_L_Expr;
--# end
--# property-start AdaNode.create_binop_assoc_r_expr /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1619
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Create_Binop_Assoc_R_Expr
  (Node : Bare_Ada_Node
      ; Op : Symbol_Type
      ; Lhs : Bare_Base_Type_Decl
      ; Rhs : Bare_Type_Expr
      ; Ret : Bare_Base_Type_Decl
  )
   return Internal_Env_Assoc
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind op Op
   --# bind lhs Lhs
   --# bind rhs Rhs
   --# bind ret Ret
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_104 with Inline_Always;
      New_Node : Bare_Synthetic_Type_Expr;
Cast_Result : Bare_Type_Expr;
New_Node_1 : Bare_Synthetic_Type_Expr;
Cast_Result_1 : Bare_Type_Expr;
Fld : Internal_Env_Assoc := No_Env_Assoc;
Let_Result : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_104 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_104;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Ada_Node_Ada_Node_P_Create_Binop_Assoc_R_Expr,
            Items    => new Mmz_Key_Array (1 ..  5))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Ada_Node,
                                As_Bare_Ada_Node => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Symbol_Type,
                                      As_Symbol_Type => Op);
               Mmz_K.Items (3) := (Kind => Mmz_Bare_Base_Type_Decl,
                                      As_Bare_Base_Type_Decl => Lhs);
               Mmz_K.Items (4) := (Kind => Mmz_Bare_Type_Expr,
                                      As_Bare_Type_Expr => Rhs);
               Mmz_K.Items (5) := (Kind => Mmz_Bare_Base_Type_Decl,
                                      As_Bare_Base_Type_Decl => Ret);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Env_Assoc;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 776 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 775 '<FieldAccess for create_binop_assoc_impl at ast.py:1626>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1626
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 773 '<New SyntheticTypeExpr at ast.py:1628>' New_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1628
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node := new Root_Node_Record
  (Ada_Synthetic_Type_Expr);
Initialize
  (Self => New_Node,
   Kind => Ada_Synthetic_Type_Expr,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node);
   Initialize_Fields_For_Synthetic_Type_Expr
     (Self => New_Node,
      Synthetic_Type_Expr_F_Target_Type => Lhs);
--# expr-done 773
      Cast_Result := New_Node;
--# expr-start 774 '<New SyntheticTypeExpr at ast.py:1630>' New_Node_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1630
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node_1 := new Root_Node_Record
  (Ada_Synthetic_Type_Expr);
Initialize
  (Self => New_Node_1,
   Kind => Ada_Synthetic_Type_Expr,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node_1);
   Initialize_Fields_For_Synthetic_Type_Expr
     (Self => New_Node_1,
      Synthetic_Type_Expr_F_Target_Type => Ret);
--# expr-done 774
      Cast_Result_1 := New_Node_1;
--# property-call-start AdaNode.create_binop_assoc_impl
Fld := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc_Impl (Node => Self, Op => Op, Lhs => Cast_Result, Rhs => Rhs, Ret => Cast_Result_1);
--# end
--# expr-done 775
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 776
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_104;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Env_Assoc,
                        As_Internal_Env_Assoc => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_104;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Create_Binop_Assoc_R_Expr;
--# end
--# property-start AdaNode.create_binop_assoc_l_r_expr /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1634
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Create_Binop_Assoc_L_R_Expr
  (Node : Bare_Ada_Node
      ; Op : Symbol_Type
      ; Lhs : Bare_Type_Expr
      ; Rhs : Bare_Type_Expr
      ; Ret : Bare_Base_Type_Decl
  )
   return Internal_Env_Assoc
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind op Op
   --# bind lhs Lhs
   --# bind rhs Rhs
   --# bind ret Ret
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_105 with Inline_Always;
      New_Node : Bare_Synthetic_Type_Expr;
Cast_Result : Bare_Type_Expr;
Fld : Internal_Env_Assoc := No_Env_Assoc;
Let_Result : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_105 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_105;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Ada_Node_Ada_Node_P_Create_Binop_Assoc_L_R_Expr,
            Items    => new Mmz_Key_Array (1 ..  5))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Ada_Node,
                                As_Bare_Ada_Node => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Symbol_Type,
                                      As_Symbol_Type => Op);
               Mmz_K.Items (3) := (Kind => Mmz_Bare_Type_Expr,
                                      As_Bare_Type_Expr => Lhs);
               Mmz_K.Items (4) := (Kind => Mmz_Bare_Type_Expr,
                                      As_Bare_Type_Expr => Rhs);
               Mmz_K.Items (5) := (Kind => Mmz_Bare_Base_Type_Decl,
                                      As_Bare_Base_Type_Decl => Ret);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Env_Assoc;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 779 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 778 '<FieldAccess for create_binop_assoc_impl at ast.py:1641>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1641
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 777 '<New SyntheticTypeExpr at ast.py:1642>' New_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1642
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node := new Root_Node_Record
  (Ada_Synthetic_Type_Expr);
Initialize
  (Self => New_Node,
   Kind => Ada_Synthetic_Type_Expr,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node);
   Initialize_Fields_For_Synthetic_Type_Expr
     (Self => New_Node,
      Synthetic_Type_Expr_F_Target_Type => Ret);
--# expr-done 777
      Cast_Result := New_Node;
--# property-call-start AdaNode.create_binop_assoc_impl
Fld := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc_Impl (Node => Self, Op => Op, Lhs => Lhs, Rhs => Rhs, Ret => Cast_Result);
--# end
--# expr-done 778
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 779
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_105;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Env_Assoc,
                        As_Internal_Env_Assoc => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_105;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Create_Binop_Assoc_L_R_Expr;
--# end
--# property-start AdaNode.custom_id_text /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:174
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Custom_Id_Text
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return String_Type
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : String_Type;
            procedure Finalizer_Scope_5 with Inline_Always;
      Str : String_Type;
Let_Result : String_Type;
            procedure Finalizer_Scope_5 is
            begin
                     Dec_Ref (Str);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_5;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 781 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 780 '<String at ast.py:180>' Str /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:180
Str := Create_String (""); 
--# expr-done 780
Let_Result := Str; Inc_Ref (Let_Result);
--# end
--# expr-done 781
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_5;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_5;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Custom_Id_Text;
--# end
--# property-start AdaNode.complete_items /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:321
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Complete_Items
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Completion_Item_Array_Access
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Completion_Item_Array_Access;
            procedure Finalizer_Scope_19 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Fld_1 : Internal_Entity_Array_Access;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity_Info;
Fld_3 : Internal_Metadata;
Fld_4 : Boolean;
Fld_5 : Internal_Unit;
Fld_6 : Boolean;
New_Struct : Internal_Completion_Item;
Map_Result : Internal_Completion_Item_Array_Access;
Let_Result : Internal_Completion_Item_Array_Access;
            procedure Finalizer_Scope_19 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_19;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 792 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 791 '<Map at ast.py:327>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:327
   --# expr-start 783 '<FieldAccess for env_get_public at ast.py:327>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:327
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 782 '<FieldAccess for children_env at ast.py:328>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:328
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children_Env (Node => Self);
--# expr-done 782
--# property-call-start AdaNode.env_get_public
Fld_1 := Libadalang.Implementation.Ada_Node_P_Env_Get_Public (Node => Self, Env => Fld, Symbol => null);
--# end
--# expr-done 783
      declare
         Map_Result_Vec : Internal_Completion_Item_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Fld_1;
      begin
         for Item_7 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_7
      --# expr-start 790 '<New CompletionItem at ast.py:331>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:331
--# expr-start 784 '<Cast to BasicDecl at ast.py:332>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:332
Cast_Expr := Item_7; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 784
--# expr-start 787 '<FieldAccess for dottable_subp at ast.py:333>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:333
--# expr-start 786 '<FieldAccess for md at ast.py:333>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:333
--# expr-start 785 '<FieldAccess for info at ast.py:333>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:333
Fld_2 := Item_7.Info;
--# expr-done 785
Fld_3 := Fld_2.Md;
--# expr-done 786
Fld_4 := Fld_3.Dottable_Subp;
--# expr-done 787
--# expr-start 789 '<FieldAccess for has_with_visibility at ast.py:334>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:334
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 788 '<FieldAccess for unit at ast.py:334>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:334
      if Item_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Unit (Node => Item_7.Node);
--# expr-done 788
--# property-call-start AdaNode.has_with_visibility
Fld_6 := Libadalang.Implementation.Ada_Node_P_Has_With_Visibility (Node => Self, Refd_Unit => Fld_5);
--# end
--# expr-done 789
New_Struct := (Decl => Cast_Result, Is_Dot_Call => Fld_4, Is_Visible => Fld_6, Weight => 0); 
--# expr-done 790
         declare
            Item_To_Append : constant Internal_Completion_Item := New_Struct;
         begin
            Internal_Completion_Item_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Completion_Item_Array
           (Items_Count => Natural (Internal_Completion_Item_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Completion_Item_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Completion_Item_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Completion_Item_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 791
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 792
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_19;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_19;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Complete_Items;
--# end
--# property-start AdaNode.immediate_declarative_region /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:490
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Immediate_Declarative_Region
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_30 with Inline_Always;
      Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_30 is
            begin
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_30;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 793 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := Empty_Env; Inc_Ref (Let_Result);
--# end
--# expr-done 793
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_30;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_30;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Immediate_Declarative_Region;
--# end
 function Ada_Node_P_Xref_Equation
  (Node : Bare_Ada_Node
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is (raise Property_Error
    with "Property AdaNode.xref_equation not implemented on type "
    & Kind_Name (Node));
--# property-start AdaNode.xref_stop_resolution /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:614
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Xref_Stop_Resolution
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Xref_Stop_Resolution;
--# end
--# property-start AdaNode.stop_resolution_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:615
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Stop_Resolution_Equation
  (Node : Bare_Ada_Node
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_38 with Inline_Always;
      True_Rel : Logic_Equation;
            procedure Finalizer_Scope_38 is
            begin
                     Dec_Ref (True_Rel);
            end Finalizer_Scope_38;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:616") else null)); 
      Property_Result := True_Rel;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_38;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_38;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Stop_Resolution_Equation;
--# end
--# property-start AdaNode.xref_initial_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:621
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Xref_Initial_Env
  (Node : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity :=
     Internal_Entity'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_39 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_39 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_39;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 795 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 794 '<FieldAccess for children_env at ast.py:627>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:627
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 794
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 795
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_39;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_39;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Xref_Initial_Env;
--# end
--# property-start AdaNode.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:689
pragma Warnings (Off, "is not referenced");
 function Ada_Node_P_Xref_Entry_Point
  (Node : Bare_Ada_Node
  )
   return Boolean
is
   Self : Bare_Ada_Node := Bare_Ada_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ada_Node_P_Xref_Entry_Point;
--# end
function Ada_Node_P_Resolve_Generic_Actual
  (E : Entity
  ) return Internal_Entity
is
      E_Info : Internal_Entity_Info :=
         Shed_Rebindings (E.Info, Node_Env (E.Node));
   Result : constant Internal_Entity :=
      Ada_Node_P_Resolve_Generic_Actual (E.Node, E_Info);
begin
      return (Node => Result.Node, Info => Result.Info);
end;
function Ada_Node_P_Use_Clauses_In_Spec_Of_Subp_Body
  (E : Entity
  ) return Lexical_Env
is
   Result : constant Lexical_Env :=
      Ada_Node_P_Use_Clauses_In_Spec_Of_Subp_Body (E.Node);
begin
      return Result;
end;
function Ada_Node_P_Use_Clauses_In_Generic_Formal_Part
  (E : Entity
  ) return Lexical_Env
is
   Result : constant Lexical_Env :=
      Ada_Node_P_Use_Clauses_In_Generic_Formal_Part (E.Node);
begin
      return Result;
end;
function Ada_Node_P_Nested_Generic_Formal_Part
  (E : Entity
  ) return Lexical_Env
is
   Result : constant Lexical_Env :=
      Ada_Node_P_Nested_Generic_Formal_Part (E.Node);
begin
      return Result;
end;
   --
   --  Primitives for Bare_Abort_Node
   --
--# property-start '[dispatcher]Abort.as_bool' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Abort_Node_P_As_Bool
  (Node : Bare_Abort_Node
  )
   return Boolean
is
   Self : Bare_Abort_Node := Bare_Abort_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Abort_Node (Self.Kind) is
               when Ada_Abort_Absent =>
                  --# property-call-start Abort.Absent.as_bool
                  Property_Result := Abort_Absent_P_As_Bool
                    (Self
                    );
                  --# end
               when Ada_Abort_Present =>
                  --# property-call-start Abort.Present.as_bool
                  Property_Result := Abort_Present_P_As_Bool
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Abort_Node_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Abort_Absent
   --
--# property-start Abort.Absent.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4922
pragma Warnings (Off, "is not referenced");
 function Abort_Absent_P_As_Bool
  (Node : Bare_Abort_Absent
  )
   return Boolean
is
   Self : Bare_Abort_Absent := Bare_Abort_Absent (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Abort_Absent_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Abort_Present
   --
--# property-start Abort.Present.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4922
pragma Warnings (Off, "is not referenced");
 function Abort_Present_P_As_Bool
  (Node : Bare_Abort_Present
  )
   return Boolean
is
   Self : Bare_Abort_Present := Bare_Abort_Present (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Abort_Present_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Abstract_Node
   --
--# property-start '[dispatcher]Abstract.as_bool' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Abstract_Node_P_As_Bool
  (Node : Bare_Abstract_Node
  )
   return Boolean
is
   Self : Bare_Abstract_Node := Bare_Abstract_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Abstract_Node (Self.Kind) is
               when Ada_Abstract_Absent =>
                  --# property-call-start Abstract.Absent.as_bool
                  Property_Result := Abstract_Absent_P_As_Bool
                    (Self
                    );
                  --# end
               when Ada_Abstract_Present =>
                  --# property-call-start Abstract.Present.as_bool
                  Property_Result := Abstract_Present_P_As_Bool
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Abstract_Node_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Abstract_Absent
   --
--# property-start Abstract.Absent.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4866
pragma Warnings (Off, "is not referenced");
 function Abstract_Absent_P_As_Bool
  (Node : Bare_Abstract_Absent
  )
   return Boolean
is
   Self : Bare_Abstract_Absent := Bare_Abstract_Absent (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Abstract_Absent_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Abstract_Present
   --
--# property-start Abstract.Present.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4866
pragma Warnings (Off, "is not referenced");
 function Abstract_Present_P_As_Bool
  (Node : Bare_Abstract_Present
  )
   return Boolean
is
   Self : Bare_Abstract_Present := Bare_Abstract_Present (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Abstract_Present_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Ada_List
   --
   --
   --  Primitives for Bare_Ada_Node_List
   --
   --
   --  Primitives for Bare_Abstract_State_Decl_List
   --
   --
   --  Primitives for Bare_Alternatives_List
   --
--# property-start AlternativesList.enum_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20247
pragma Warnings (Off, "is not referenced");
 function Alternatives_List_P_Enum_Type
  (Node : Bare_Alternatives_List
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Alternatives_List := Bare_Alternatives_List (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Alternatives_List :=
     Internal_Entity_Alternatives_List'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Case_Stmt;
Fld : Internal_Entity;
Fld_1 : Internal_Entity;
Fld_2 : Internal_Entity;
Cast_Result : Internal_Entity_Case_Stmt;
Fld_3 : Internal_Entity_Expr;
Fld_4 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 802 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 799 '<Cast to CaseStmt at ast.py:20252>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20252
--# expr-start 798 '<parent at ast.py:20252>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20252
--# expr-start 797 '<parent at ast.py:20252>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20252
--# expr-start 796 '<parent at ast.py:20252>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20252
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 796
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Parent (Node => Fld.Node, E_Info => Fld.Info);
--# expr-done 797
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Parent (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# expr-done 798
   if Fld_2.Node = null
      or else Fld_2.Node.Kind in Ada_Case_Stmt_Range
   then
      Cast_Result := Create_Internal_Entity_Case_Stmt
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
   else
         Cast_Result := No_Entity_Case_Stmt;
   end if;
--# expr-done 799
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Case_Stmt then
   --# scope-start
   --# bind cs Var_Expr
--# expr-start 801 '<FieldAccess for expression_type at ast.py:20253>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20253
--# expr-start 800 '<FieldAccess for expr at ast.py:20253>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20253
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Expr (Node => Var_Expr.Node.Case_Stmt_F_Expr, Info => Var_Expr.Info);
--# expr-done 800
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.expression_type
Fld_4 := Libadalang.Implementation.Expr_P_Expression_Type (Node => Fld_3.Node, E_Info => Fld_3.Info);
--# end
--# expr-done 801
Scope_Result := Fld_4;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 802
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Alternatives_List_P_Enum_Type;
--# end
--# property-start AlternativesList.complete_items /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20257
pragma Warnings (Off, "is not referenced");
 function Alternatives_List_P_Complete_Items
  (Node : Bare_Alternatives_List
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Completion_Item_Array_Access
is
   Self : Bare_Alternatives_List := Bare_Alternatives_List (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Alternatives_List :=
     Internal_Entity_Alternatives_List'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Completion_Item_Array_Access;
            procedure Finalizer_Scope_1411 with Inline_Always;
      Match_0 : Internal_Entity_Enum_Literal_Decl;
Match_1 : Internal_Entity;
Fld : Lexical_Env := Empty_Env;
Env_Get_Result : Internal_Entity_Array_Access;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Info;
Fld_2 : Internal_Metadata;
Fld_3 : Boolean;
Fld_4 : Internal_Unit;
Fld_5 : Boolean;
Fld_6 : Internal_Entity_Base_Type_Decl;
Fld_7 : Internal_Entity_Type_Decl;
Cast_Result_1 : Internal_Entity;
Cast_Result_2 : Internal_Entity;
Is_Equiv : Boolean;
If_Result : Integer;
Match_Prefix : Internal_Entity;
Cast_Result_3 : Internal_Entity_Enum_Literal_Decl;
Let_Result : Integer;
Scope_Result : Integer;
Cast_Result_4 : Internal_Entity;
Let_Result_1 : Integer;
Scope_Result_1 : Integer;
Match_Result : Integer;
New_Struct : Internal_Completion_Item;
Map_Result : Internal_Completion_Item_Array_Access;
Let_Result_2 : Internal_Completion_Item_Array_Access;
            procedure Finalizer_Scope_1411 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Env_Get_Result);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result_2);
            end Finalizer_Scope_1411;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 818 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 817 '<Map at ast.py:20261>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20261
   --# expr-start 804 '<EnvGet at ast.py:20261>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20261
--# expr-start 803 '<FieldAccess for children_env at ast.py:20261>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20261
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children_Env (Node => Self);
--# expr-done 803
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Fld, Key => null, Lookup_Kind => To_Lookup_Kind_Type (Recursive), Categories => (Inherited_Primitives => True, Nocat => True))); 
--# expr-done 804
      declare
         Map_Result_Vec : Internal_Completion_Item_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Env_Get_Result;
      begin
         for Item_28 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_28
      --# expr-start 816 '<New CompletionItem at ast.py:20262>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20262
--# expr-start 805 '<Cast to BasicDecl at ast.py:20263>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20263
Cast_Expr := Item_28; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 805
--# expr-start 808 '<FieldAccess for dottable_subp at ast.py:20264>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20264
--# expr-start 807 '<FieldAccess for md at ast.py:20264>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20264
--# expr-start 806 '<FieldAccess for info at ast.py:20264>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20264
Fld_1 := Item_28.Info;
--# expr-done 806
Fld_2 := Fld_1.Md;
--# expr-done 807
Fld_3 := Fld_2.Dottable_Subp;
--# expr-done 808
--# expr-start 810 '<FieldAccess for has_with_visibility at ast.py:20265>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20265
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 809 '<FieldAccess for unit at ast.py:20265>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20265
      if Item_28.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Unit (Node => Item_28.Node);
--# expr-done 809
--# property-call-start AdaNode.has_with_visibility
Fld_5 := Libadalang.Implementation.Ada_Node_P_Has_With_Visibility (Node => Self, Refd_Unit => Fld_4);
--# end
--# expr-done 810
--# expr-start 815 '<Match at ast.py:20266>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20266
      if Item_28.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Item_28; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Enum_Literal_Decl .. Ada_Synthetic_Char_Enum_Lit =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Enum_Literal_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_3; 
--# bind eld Match_0
--# expr-start 814 '<If at ast.py:20267>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20267
--# expr-start 813 '<Eq at ast.py:20268>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20268
--# expr-start 811 '<FieldAccess for enum_type at ast.py:20268>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20268
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AlternativesList.enum_type
Fld_6 := Libadalang.Implementation.Alternatives_List_P_Enum_Type (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 811
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_6.Node,
         Info => Fld_6.Info);
--# expr-start 812 '<FieldAccess for enum_type at ast.py:20268>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20268
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]EnumLiteralDecl.enum_type'
Fld_7 := Libadalang.Implementation.Dispatcher_Enum_Literal_Decl_P_Enum_Type (Node => Match_0.Node, E_Info => Match_0.Info);
--# end
--# expr-done 812
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_7.Node,
         Info => Fld_7.Info);
Is_Equiv := Equivalent (Cast_Result_1, Cast_Result_2); 
--# expr-done 813
if Is_Equiv then
   If_Result := 100;
else
   If_Result := 0;
end if;
--# expr-done 814
Let_Result := If_Result; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Abort_Absent .. Ada_Entry_Decl | Ada_Generic_Subp_Internal .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_4; 
--# bind _ Match_1
Let_Result_1 := 0; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
end case;
--# expr-done 815
New_Struct := (Decl => Cast_Result, Is_Dot_Call => Fld_3, Is_Visible => Fld_5, Weight => Match_Result); 
--# expr-done 816
         declare
            Item_To_Append : constant Internal_Completion_Item := New_Struct;
         begin
            Internal_Completion_Item_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Completion_Item_Array
           (Items_Count => Natural (Internal_Completion_Item_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Completion_Item_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Completion_Item_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Completion_Item_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 817
Let_Result_2 := Map_Result; Inc_Ref (Let_Result_2);
--# end
--# expr-done 818
      Property_Result := Let_Result_2;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1411;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1411;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Alternatives_List_P_Complete_Items;
--# end
   --
   --  Primitives for Bare_Constraint_List
   --
   --
   --  Primitives for Bare_Decl_List
   --
   --
   --  Primitives for Bare_Stmt_List
   --
   --
   --  Primitives for Bare_Aspect_Assoc_List
   --
   --
   --  Primitives for Bare_Base_Assoc_List
   --
   --
   --  Primitives for Bare_Basic_Assoc_List
   --
   --
   --  Primitives for Bare_Assoc_List
   --
--# property-start AssocList.actual_for_param_at /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14326
pragma Warnings (Off, "is not referenced");
 function Assoc_List_P_Actual_For_Param_At
  (Node : Bare_Assoc_List
      ; Param : Internal_Entity_Defining_Name
      ; Pos : Integer
      ; Default_Expr : Internal_Entity_Expr
         := No_Entity_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Expr
is
   Self : Bare_Assoc_List := Bare_Assoc_List (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Assoc_List :=
     Internal_Entity_Assoc_List'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind param Param
   --# bind pos Pos
   --# bind default_expr Default_Expr
   Property_Result : Internal_Entity_Expr;
            procedure Finalizer_Scope_1035 with Inline_Always;
      Block_Var : Internal_Single_Actual_Array_Access;
Var_Expr : Internal_Single_Actual;
Var_Expr_1 : Bare_Base_Id;
Fld : Internal_Single_Actual_Array_Access;
Fld_1 : Bare_Base_Id;
Fld_2 : Internal_Entity_Name;
Fld_3 : Bare_Name;
Fld_4 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Find_Result : Internal_Single_Actual;
Fld_5 : Internal_Entity_Basic_Assoc;
Fld_6 : Internal_Entity_Expr;
Scope_Result_1 : Internal_Entity_Expr;
Get_Result : Internal_Single_Actual;
Fld_7 : Internal_Entity_Basic_Assoc;
Fld_8 : Internal_Entity_Expr;
Get_Result_1 : Internal_Single_Actual;
Is_Equal : Boolean;
Get_Result_2 : Internal_Single_Actual;
Fld_9 : Internal_Entity_Basic_Assoc;
Fld_10 : Bare_Ada_Node_Array_Access;
Is_Equal_1 : Boolean;
Not_Val : Boolean;
If_Result : Boolean;
If_Result_1 : Internal_Entity_Expr;
Result_Var_1 : Internal_Entity_Expr;
Let_Result : Internal_Entity_Expr;
            procedure Finalizer_Scope_1035 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Fld_10);
            end Finalizer_Scope_1035;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 838 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 819 '<FieldAccess for unpacked_params at ast.py:14333>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14333
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AssocList.unpacked_params
Fld := Libadalang.Implementation.Assoc_List_P_Unpacked_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 819
Block_Var := Fld; Inc_Ref (Block_Var);
--# bind up Block_Var
--# expr-start 824 '<Find at ast.py:14335>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14335
      Find_Result := No_Single_Actual;
      declare
         Collection : constant Internal_Single_Actual_Array_Access := Block_Var;
      begin
         for Item_29 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_29
--# expr-start 820 '<FieldAccess for name at ast.py:14337>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14337
Fld_1 := Item_29.Name;
--# expr-done 820
Var_Expr_1 := Fld_1;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 823 '<FieldAccess for matches at ast.py:14337>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14337
      if Var_Expr_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 822 '<FieldAccess for node at ast.py:14337>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14337
--# expr-start 821 '<FieldAccess for name at ast.py:14337>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14337
      if Param.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Param.Node.Defining_Name_F_Name, Info => Param.Info);
--# expr-done 821
Fld_3 := Fld_2.Node;
--# expr-done 822
--# property-call-start Name.matches
Fld_4 := Libadalang.Implementation.Name_P_Matches (Node => Var_Expr_1, N => Fld_3);
--# end
--# expr-done 823
Scope_Result := Fld_4;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
      if Result_Var then
         Find_Result := Item_29;
      end if;
   --# end
               exit when Result_Var;
         end loop;
      end;
--# expr-done 824
Var_Expr := Find_Result;
if Var_Expr /= No_Single_Actual then
   --# scope-start
   --# bind a Var_Expr
--# expr-start 826 '<FieldAccess for expr at ast.py:14339>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14339
--# expr-start 825 '<FieldAccess for assoc at ast.py:14339>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14339
Fld_5 := Var_Expr.Assoc;
--# expr-done 825
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_6 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_5.Node, E_Info => Fld_5.Info);
--# end
--# expr-done 826
Scope_Result_1 := Fld_6;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   --# expr-start 837 '<If at ast.py:14341>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14341
--# expr-start 828 '<IsNull at ast.py:14343>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14343
--# expr-start 827 '<at at ast.py:14343>' Get_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14343
Get_Result_1 := Get (Block_Var, Pos, True); 
--# expr-done 827
Is_Equal := Get_Result_1 = No_Single_Actual; 
--# expr-done 828
if Is_Equal then
   If_Result := True;
else
   --# expr-start 833 '<Not at ast.py:14344>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14344
--# expr-start 832 '<IsNull at ast.py:14344>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14344
--# expr-start 831 '<FieldAccess for names at ast.py:14344>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14344
--# expr-start 830 '<FieldAccess for assoc at ast.py:14344>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14344
--# expr-start 829 '<at at ast.py:14344>' Get_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14344
Get_Result_2 := Get (Block_Var, Pos, True); 
--# expr-done 829
Fld_9 := Get_Result_2.Assoc;
--# expr-done 830
      if Fld_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.names'
Fld_10 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Names (Node => Fld_9.Node);
--# end
--# expr-done 831
Is_Equal_1 := Equivalent (Fld_10, No_Bare_Ada_Node_Array_Type); 
--# expr-done 832
Not_Val := not (Is_Equal_1); 
--# expr-done 833
   If_Result := Not_Val;
end if;
if If_Result then
   If_Result_1 := Default_Expr;
else
   --# expr-start 836 '<FieldAccess for expr at ast.py:14350>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14350
--# expr-start 835 '<FieldAccess for assoc at ast.py:14350>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14350
--# expr-start 834 '<at at ast.py:14350>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14350
Get_Result := Get (Block_Var, Pos, True); 
--# expr-done 834
Fld_7 := Get_Result.Assoc;
--# expr-done 835
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_8 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_7.Node, E_Info => Fld_7.Info);
--# end
--# expr-done 836
   If_Result_1 := Fld_8;
end if;
--# expr-done 837
   Result_Var_1 := If_Result_1;
end if;
Let_Result := Result_Var_1; 
--# end
--# expr-done 838
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1035;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1035;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Assoc_List_P_Actual_For_Param_At;
--# end
--# property-start AssocList.unpacked_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14355
pragma Warnings (Off, "is not referenced");
 function Assoc_List_P_Unpacked_Params
  (Node : Bare_Assoc_List
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Single_Actual_Array_Access
is
   Self : Bare_Assoc_List := Bare_Assoc_List (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Assoc_List :=
     Internal_Entity_Assoc_List'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Single_Actual_Array_Access;
            procedure Finalizer_Scope_1036 with Inline_Always;
            procedure Finalizer_Scope_1769 with Inline_Always;
      Names : Bare_Ada_Node_Array_Access;
Saved : Internal_Entity_Assoc_List;
Fld : Bare_Assoc_List;
Fld_1 : Internal_Entity_Info;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_2 : Bare_Ada_Node_Array_Access;
Cast_Expr : Bare_Identifier;
Cast_Result : Bare_Base_Id;
New_Struct : Internal_Single_Actual;
Singleton : Internal_Single_Actual_Array_Access;
Cast_Expr_1 : Bare_Ada_Node;
Cast_Result_1 : Bare_Base_Id;
New_Struct_1 : Internal_Single_Actual;
Is_A : Boolean;
Map_Result : Internal_Single_Actual_Array_Access;
Len : Integer;
Is_Equal : Boolean;
If_Result_1 : Internal_Single_Actual_Array_Access;
Let_Result : Internal_Single_Actual_Array_Access;
Item_30 : Internal_Entity_Basic_Assoc;
Bare_Item_30 : Bare_Basic_Assoc;
Map_Result_1 : Internal_Single_Actual_Array_Access;
Let_Result_1 : Internal_Single_Actual_Array_Access;
            procedure Finalizer_Scope_1036 is
            begin
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Let_Result_1);
            end Finalizer_Scope_1036;
            procedure Finalizer_Scope_1769 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Names);
                     Dec_Ref (Singleton);
                     Dec_Ref (Map_Result);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1769;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Assoc_List_Assoc_List_P_Unpacked_Params,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Assoc_List,
                                As_Bare_Assoc_List => Self);
               Mmz_K.Items (2) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Single_Actual_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 850 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 849 '<Mapcat at ast.py:14361>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14361
Saved := Ent; 
Fld := Saved.Node;
      declare
         Map_Result_1_Vec : Internal_Single_Actual_Vectors.Vector;
      begin
      if Fld = null then
      Map_Result_1 := Create_Internal_Single_Actual_Array (0);
      else
      declare
         Collection : constant Bare_Assoc_List := Fld;
      begin
         for Untyped_Item_30 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_30 := Untyped_Item_30; 
Is_Null := Bare_Item_30 = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_1 := Saved.Info;
As_Entity := (Info => Fld_1, Node => Bare_Item_30); 
   If_Result := As_Entity;
end if;
                  Item_30 := If_Result; 
   --# scope-start
               --# bind pa Item_30
      --# expr-start 848 '<Let at ast.py:14361>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14361
--# scope-start
--# expr-start 839 '<FieldAccess for names at ast.py:14361>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14361
      if Item_30.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.names'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Names (Node => Item_30.Node);
--# end
--# expr-done 839
Names := Fld_2; Inc_Ref (Names);
--# bind names Names
--# expr-start 847 '<If at ast.py:14361>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14361
--# expr-start 841 '<Eq at ast.py:14362>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14362
--# expr-start 840 '<length at ast.py:14362>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14362
Len := Length (Names); 
--# expr-done 840
Is_Equal := Len = 0; 
--# expr-done 841
if Is_Equal then
   --# expr-start 842 '<New SingleActual at ast.py:14363>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14363
Cast_Expr := No_Bare_Ada_Node; 
      Cast_Result := Cast_Expr;
New_Struct := (Assoc => Item_30, Name => Cast_Result); 
--# expr-done 842
                Singleton := Create_Internal_Single_Actual_Array (Items_Count => 1);
                Singleton.Items (1) := New_Struct;
   If_Result_1 := Singleton;
else
   --# expr-start 846 '<FilterMap at ast.py:14364>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14364
      declare
         Map_Result_Vec : Internal_Single_Actual_Vectors.Vector;
      begin
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Names;
      begin
         for Item_31 of
               Collection.Items
         loop
   --# scope-start
               --# bind i Item_31
         --# expr-start 843 '<IsA BaseId at ast.py:14366>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14366
Is_A := Item_31 /= null 
and then Item_31.Kind in Ada_Base_Id; 
--# expr-done 843
         if Is_A then
      --# expr-start 845 '<New SingleActual at ast.py:14365>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14365
--# expr-start 844 '<Cast to BaseId at ast.py:14365>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14365
Cast_Expr_1 := Item_31; 
   if Cast_Expr_1 = null
      or else Cast_Expr_1.Kind in Ada_Base_Id
   then
      Cast_Result_1 := Cast_Expr_1;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 844
New_Struct_1 := (Assoc => Item_30, Name => Cast_Result_1); 
--# expr-done 845
         declare
            Item_To_Append : constant Internal_Single_Actual := New_Struct_1;
         begin
            Internal_Single_Actual_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Single_Actual_Array
           (Items_Count => Natural (Internal_Single_Actual_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Single_Actual_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Single_Actual_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Single_Actual_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 846
   If_Result_1 := Map_Result;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 847
Let_Result := If_Result_1; Inc_Ref (Let_Result);
--# end
--# expr-done 848
         for Item_To_Append of
               Let_Result.Items
         loop
            Internal_Single_Actual_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_1769;
         end loop;
      end;
      end if;
         Map_Result_1 := Create_Internal_Single_Actual_Array
           (Items_Count => Natural (Internal_Single_Actual_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Single_Actual_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Single_Actual_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Single_Actual_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 849
Let_Result_1 := Map_Result_1; Inc_Ref (Let_Result_1);
--# end
--# expr-done 850
      Property_Result := Let_Result_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1036;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Single_Actual_Array_Access,
                        As_Internal_Single_Actual_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1036;
               Finalizer_Scope_1769;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Assoc_List_P_Unpacked_Params;
--# end
--# property-start AssocList.zip_with_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14372
pragma Warnings (Off, "is not referenced");
 function Assoc_List_P_Zip_With_Params
  (Node : Bare_Assoc_List
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Param_Actual_Array_Access
is
   Self : Bare_Assoc_List := Bare_Assoc_List (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Assoc_List :=
     Internal_Entity_Assoc_List'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Param_Actual_Array_Access;
            procedure Finalizer_Scope_1037 with Inline_Always;
            procedure Finalizer_Scope_1828 with Inline_Always;
            procedure Finalizer_Scope_1829 with Inline_Always;
            procedure Finalizer_Scope_1830 with Inline_Always;
            procedure Finalizer_Scope_1831 with Inline_Always;
            procedure Finalizer_Scope_1832 with Inline_Always;
            procedure Finalizer_Scope_1833 with Inline_Always;
            procedure Finalizer_Scope_1834 with Inline_Always;
            procedure Finalizer_Scope_1835 with Inline_Always;
            procedure Finalizer_Scope_1836 with Inline_Always;
            procedure Finalizer_Scope_1838 with Inline_Always;
            procedure Finalizer_Scope_1840 with Inline_Always;
            procedure Finalizer_Scope_1842 with Inline_Always;
            procedure Finalizer_Scope_1848 with Inline_Always;
      Block_Var : Boolean;
Block_Var_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Block_Var_2 : Internal_Entity_Basic_Assoc;
Block_Var_3 : Internal_Param_Actual_Array_Access;
Block_Var_4 : Internal_Param_Actual_Array_Access;
Block_Var_5 : Internal_Param_Actual_Array_Access;
Block_Var_6 : Internal_Param_Actual_Array_Access;
Var_Expr : Internal_Entity_Name;
Var_Expr_1 : Internal_Entity;
Match_0 : Internal_Entity_Call_Expr;
Match_1 : Internal_Entity_Generic_Instantiation;
Match_2 : Internal_Entity_Composite_Constraint;
Match_3 : Internal_Entity_Base_Aggregate;
Match_4 : Internal_Entity;
Var_Expr_2 : Internal_Entity_Base_Formal_Param_Holder;
Var_Expr_3 : Internal_Entity_Generic_Decl;
Var_Expr_4 : Internal_Entity_Base_Type_Decl;
Var_Expr_5 : Internal_Entity_Base_Record_Def;
Var_Expr_6 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Var_Expr_7 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Decl : Internal_Entity_Generic_Formal_Subp_Decl;
Subp : Internal_Entity_Formal_Subp_Decl;
Var_Expr_8 : Internal_Entity_Basic_Decl;
Var_Expr_9 : Internal_Entity_Generic_Formal_Subp_Decl;
Var_Expr_10 : Internal_Entity_Formal_Subp_Decl;
Var_Expr_11 : Internal_Entity_Basic_Assoc;
Bound_Imprecise_Fallback : Boolean;
Fld : Internal_Entity;
Cast_Result : Internal_Entity_Name;
Fld_1 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Dyn_Var_Bind_Result : Boolean;
Fld_2 : Internal_Entity;
Fld_3 : Internal_Entity_Base_Formal_Param_Holder;
Fld_4 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Scope_Result_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Result_Var_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_5 : Internal_Entity_Name;
Fld_6 : Internal_Entity_Basic_Decl;
Cast_Result_1 : Internal_Entity_Generic_Decl;
Fld_7 : Internal_Entity_Generic_Formal_Part;
Fld_8 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Scope_Result_2 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Result_Var_2 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_9 : Internal_Entity_Base_Type_Decl;
Fld_10 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Scope_Result_3 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Result_Var_3 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Cast_Expr : Bare_Assoc_List;
Cast_Result_2 : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_11 : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_12 : Internal_Entity_Base_Type_Decl;
Fld_13 : Internal_Entity_Base_Record_Def;
Fld_14 : Internal_Entity_Component_List;
Fld_15 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Scope_Result_4 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Result_Var_4 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Dyn_Var_Bind_Result_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Dyn_Var_Bind_Result_2 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Match_Prefix : Internal_Entity;
Cast_Result_3 : Internal_Entity_Call_Expr;
Let_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Scope_Result_5 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Cast_Result_4 : Internal_Entity_Generic_Instantiation;
Let_Result_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Scope_Result_6 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Cast_Result_5 : Internal_Entity_Composite_Constraint;
Let_Result_2 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Scope_Result_7 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Cast_Result_6 : Internal_Entity_Base_Aggregate;
Let_Result_3 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Scope_Result_8 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Cast_Result_7 : Internal_Entity;
Let_Result_4 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Scope_Result_9 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Match_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Scope_Result_10 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Result_Var_5 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Saved : Internal_Entity_Assoc_List;
Fld_16 : Bare_Assoc_List;
Fld_17 : Internal_Entity_Info;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_18 : Bare_Ada_Node_Array_Access;
Is_A : Boolean;
Quantifier_Result : Boolean;
Item_32 : Internal_Entity_Basic_Assoc;
Bare_Item_32 : Bare_Basic_Assoc;
Find_Result : Internal_Entity_Basic_Assoc;
Fld_19 : Internal_Param_Match_Array_Access;
Fld_20 : Internal_Entity_Defining_Name;
Fld_21 : Internal_Single_Actual;
Fld_22 : Internal_Entity_Basic_Assoc;
Fld_23 : Internal_Entity_Expr;
New_Struct : Internal_Param_Actual;
Map_Result : Internal_Param_Actual_Array_Access;
Scope_Result_11 : Internal_Param_Actual_Array_Access;
Result_Var_6 : Internal_Param_Actual_Array_Access;
Cast_Expr_1 : Internal_Entity_Base_Formal_Param_Decl;
Cast_Result_8 : Internal_Entity_Generic_Formal_Subp_Decl;
Cast_Expr_2 : Internal_Entity_Base_Formal_Param_Decl;
Cast_Result_9 : Internal_Entity_Generic_Formal_Subp_Decl;
Fld_24 : Internal_Entity_Basic_Decl;
Cast_Result_10 : Internal_Entity_Formal_Subp_Decl;
Fld_25 : Internal_Entity_Defining_Name;
Fld_26 : Internal_Entity_Expr;
Fld_27 : Internal_Entity;
Cast_Result_11 : Internal_Entity_Generic_Instantiation;
Fld_28 : Internal_Entity_Basic_Decl;
Fld_29 : Internal_Entity_Defining_Name;
Scope_Result_12 : Internal_Entity_Defining_Name;
Result_Var_7 : Internal_Entity_Defining_Name;
Cast_Result_12 : Internal_Entity_Expr;
Fld_30 : Internal_Entity_Expr;
Is_A_1 : Boolean;
If_Result_1 : Internal_Entity_Expr;
New_Struct_1 : Internal_Param_Actual;
Let_Result_5 : Internal_Param_Actual;
Cast_Expr_3 : Internal_Entity_Base_Formal_Param_Decl;
Cast_Result_13 : Internal_Entity_Generic_Formal_Subp_Decl;
Fld_31 : Internal_Entity_Basic_Decl;
Cast_Result_14 : Internal_Entity_Formal_Subp_Decl;
Fld_32 : Internal_Entity_Expr;
Is_A_2 : Boolean;
Fld_33 : Internal_Entity_Defining_Name;
Fld_34 : Internal_Entity_Defining_Name;
Cast_Result_15 : Internal_Entity;
Cast_Result_16 : Internal_Entity;
Is_Equiv : Boolean;
Quantifier_Result_1 : Boolean;
Not_Val : Boolean;
If_Result_2 : Boolean;
Scope_Result_13 : Boolean;
Result_Var_8 : Boolean;
Scope_Result_14 : Boolean;
Result_Var_9 : Boolean;
Map_Result_1 : Internal_Param_Actual_Array_Access;
Scope_Result_15 : Internal_Param_Actual_Array_Access;
Result_Var_10 : Internal_Param_Actual_Array_Access;
Concat_Result : Internal_Param_Actual_Array_Access;
Fld_35 : Internal_Entity_Defining_Name_Array_Access;
Fld_36 : Internal_Entity_Expr;
New_Struct_2 : Internal_Param_Actual;
Fld_37 : Internal_Entity_Defining_Name;
Cast_Result_17 : Internal_Entity;
Cast_Expr_4 : Internal_Entity_Defining_Name;
Cast_Result_18 : Internal_Entity;
Is_Equiv_1 : Boolean;
Quantifier_Result_2 : Boolean;
Not_Val_1 : Boolean;
Map_Result_2 : Internal_Param_Actual_Array_Access;
Scope_Result_16 : Internal_Param_Actual_Array_Access;
Result_Var_11 : Internal_Param_Actual_Array_Access;
Concat_Result_1 : Internal_Param_Actual_Array_Access;
Let_Result_6 : Internal_Param_Actual_Array_Access;
            procedure Finalizer_Scope_1037 is
            begin
                     Dec_Ref (Scope_Result_10);
                     Dec_Ref (Result_Var_5);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Scope_Result_11);
                     Dec_Ref (Result_Var_6);
                     Dec_Ref (Block_Var_3);
                     Dec_Ref (Scope_Result_15);
                     Dec_Ref (Result_Var_10);
                     Dec_Ref (Block_Var_4);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Block_Var_5);
                     Dec_Ref (Scope_Result_16);
                     Dec_Ref (Result_Var_11);
                     Dec_Ref (Block_Var_6);
                     Dec_Ref (Concat_Result_1);
                     Dec_Ref (Let_Result_6);
            end Finalizer_Scope_1037;
            procedure Finalizer_Scope_1828 is
            begin
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result_5);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_6);
                     Dec_Ref (Let_Result_2);
                     Dec_Ref (Scope_Result_7);
                     Dec_Ref (Let_Result_3);
                     Dec_Ref (Scope_Result_8);
                     Dec_Ref (Let_Result_4);
                     Dec_Ref (Scope_Result_9);
                     Dec_Ref (Match_Result);
            end Finalizer_Scope_1828;
            procedure Finalizer_Scope_1829 is
            begin
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
            end Finalizer_Scope_1829;
            procedure Finalizer_Scope_1830 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_1830;
            procedure Finalizer_Scope_1831 is
            begin
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Result_Var_2);
            end Finalizer_Scope_1831;
            procedure Finalizer_Scope_1832 is
            begin
                     Dec_Ref (Fld_8);
            end Finalizer_Scope_1832;
            procedure Finalizer_Scope_1833 is
            begin
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Result_Var_3);
            end Finalizer_Scope_1833;
            procedure Finalizer_Scope_1834 is
            begin
                     Dec_Ref (Fld_10);
            end Finalizer_Scope_1834;
            procedure Finalizer_Scope_1835 is
            begin
                     Dec_Ref (Fld_11);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Scope_Result_4);
                     Dec_Ref (Result_Var_4);
                     Dec_Ref (Dyn_Var_Bind_Result_1);
                     Dec_Ref (Dyn_Var_Bind_Result_2);
            end Finalizer_Scope_1835;
            procedure Finalizer_Scope_1836 is
            begin
                     Dec_Ref (Fld_15);
            end Finalizer_Scope_1836;
            procedure Finalizer_Scope_1838 is
            begin
                     Dec_Ref (Fld_18);
            end Finalizer_Scope_1838;
            procedure Finalizer_Scope_1840 is
            begin
                     Dec_Ref (Var_Expr_6);
                     Dec_Ref (Fld_19);
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_1840;
            procedure Finalizer_Scope_1842 is
            begin
                     Dec_Ref (Var_Expr_7);
                     Dec_Ref (Map_Result_1);
            end Finalizer_Scope_1842;
            procedure Finalizer_Scope_1848 is
            begin
                     Dec_Ref (Fld_35);
                     Dec_Ref (Map_Result_2);
            end Finalizer_Scope_1848;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 920 '<Block at ???>' Let_Result_6 None
--# scope-start
--# expr-start 854 '<bind at ast.py:14380>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14380
Bound_Imprecise_Fallback := False; 
--# expr-start 852 '<Cast to Name at ast.py:14381>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14381
--# expr-start 851 '<parent at ast.py:14381>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14381
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 851
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Name
   then
      Cast_Result := Create_Internal_Entity_Name
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Name;
   end if;
--# expr-done 852
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Name then
   --# scope-start
   --# bind e Var_Expr
--# expr-start 853 '<FieldAccess for is_dot_call at ast.py:14381>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14381
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_dot_call
Fld_1 := Libadalang.Implementation.Name_P_Is_Dot_Call (Node => Var_Expr.Node, Imprecise_Fallback => Bound_Imprecise_Fallback, E_Info => Var_Expr.Info);
--# end
--# expr-done 853
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Dyn_Var_Bind_Result := Result_Var; 
--# expr-done 854
Block_Var := Dyn_Var_Bind_Result; 
--# bind is_dottable_subp Block_Var
--# expr-start 855 '<parent at ast.py:14384>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14384
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 855
Var_Expr_1 := Fld_2;
if Var_Expr_1 /= No_Entity then
   --# scope-start
--# expr-start 872 '<Match at ast.py:14384>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14384
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Var_Expr_1; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Call_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Call_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_3; 
--# bind e Match_0
--# expr-start 856 '<FieldAccess for called_subp_spec at ast.py:14385>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14385
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.called_subp_spec
Fld_3 := Libadalang.Implementation.Name_P_Called_Subp_Spec (Node => Match_0.Node, E_Info => Match_0.Info);
--# end
--# expr-done 856
Var_Expr_2 := Fld_3;
if Var_Expr_2 /= No_Entity_Base_Formal_Param_Holder then
   --# scope-start
--# expr-start 857 '<FieldAccess for abstract_formal_params at ast.py:14385>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14385
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseFormalParamHolder.abstract_formal_params'
Fld_4 := Libadalang.Implementation.Dispatcher_Base_Formal_Param_Holder_P_Abstract_Formal_Params (Node => Var_Expr_2.Node, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 857
Scope_Result_1 := Fld_4;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_1830;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Internal_Entity_Base_Formal_Param_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var_1);
Let_Result := Result_Var_1; Inc_Ref (Let_Result);
--# end
Scope_Result_5 := Let_Result;
   Inc_Ref (Scope_Result_5);
   --# end
      Finalizer_Scope_1829;
            Match_Result := Scope_Result_5; Inc_Ref (Match_Result);
         when Ada_Generic_Package_Instantiation .. Ada_Generic_Subp_Instantiation =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Generic_Instantiation
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_4; 
--# bind i Match_1
--# expr-start 860 '<Cast to GenericDecl at ast.py:14388>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14388
--# expr-start 859 '<FieldAccess for referenced_decl at ast.py:14388>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14388
--# expr-start 858 '<FieldAccess for generic_entity_name at ast.py:14388>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14388
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericInstantiation.generic_entity_name'
Fld_5 := Libadalang.Implementation.Dispatcher_Generic_Instantiation_P_Generic_Entity_Name (Node => Match_1.Node, E_Info => Match_1.Info);
--# end
--# expr-done 858
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld_6 := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Fld_5.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_5.Info);
--# end
--# expr-done 859
   if Fld_6.Node = null
      or else Fld_6.Node.Kind in Ada_Generic_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Generic_Decl
        (Node => Fld_6.Node,
         Info => Fld_6.Info);
   else
         Cast_Result_1 := No_Entity_Generic_Decl;
   end if;
--# expr-done 860
Var_Expr_3 := Cast_Result_1;
if Var_Expr_3 /= No_Entity_Generic_Decl then
   --# scope-start
--# expr-start 862 '<FieldAccess for abstract_formal_params at ast.py:14388>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14388
--# expr-start 861 '<FieldAccess for formal_part at ast.py:14388>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14388
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Generic_Formal_Part (Node => Var_Expr_3.Node.Generic_Decl_F_Formal_Part, Info => Var_Expr_3.Info);
--# expr-done 861
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseFormalParamHolder.abstract_formal_params'
Fld_8 := Libadalang.Implementation.Dispatcher_Base_Formal_Param_Holder_P_Abstract_Formal_Params (Node => Fld_7.Node, E_Info => Fld_7.Info);
--# end
--# expr-done 862
Scope_Result_2 := Fld_8;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_1832;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Internal_Entity_Base_Formal_Param_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var_2);
Let_Result_1 := Result_Var_2; Inc_Ref (Let_Result_1);
--# end
Scope_Result_6 := Let_Result_1;
   Inc_Ref (Scope_Result_6);
   --# end
      Finalizer_Scope_1831;
            Match_Result := Scope_Result_6; Inc_Ref (Match_Result);
         when Ada_Composite_Constraint =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Create_Internal_Entity_Composite_Constraint
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_5; 
--# bind c Match_2
--# expr-start 863 '<FieldAccess for subtype at ast.py:14392>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14392
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Constraint.subtype
Fld_9 := Libadalang.Implementation.Constraint_P_Subtype (Node => Match_2.Node, E_Info => Match_2.Info);
--# end
--# expr-done 863
Var_Expr_4 := Fld_9;
if Var_Expr_4 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 864 '<FieldAccess for discriminants_list at ast.py:14392>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14392
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.discriminants_list'
Fld_10 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Discriminants_List (Node => Var_Expr_4.Node, Origin => No_Bare_Ada_Node, E_Info => Var_Expr_4.Info);
--# end
--# expr-done 864
Scope_Result_3 := Fld_10;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_1834;
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := No_Internal_Entity_Base_Formal_Param_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var_3);
Let_Result_2 := Result_Var_3; Inc_Ref (Let_Result_2);
--# end
Scope_Result_7 := Let_Result_2;
   Inc_Ref (Scope_Result_7);
   --# end
      Finalizer_Scope_1833;
            Match_Result := Scope_Result_7; Inc_Ref (Match_Result);
         when Ada_Aggregate .. Ada_Null_Record_Aggregate =>
   --# scope-start
--# scope-start
      Cast_Result_6 := Create_Internal_Entity_Base_Aggregate
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_6; 
--# bind a Match_3
--# expr-start 871 '<bind at ast.py:14394>' Dyn_Var_Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14394
Cast_Expr := Self; 
      Cast_Result_2 := Cast_Expr;
Bound_Origin := Cast_Result_2; 
--# expr-start 870 '<bind at ast.py:14394>' Dyn_Var_Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14394
--# expr-start 865 '<FieldAccess for node_env at ast.py:14395>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14395
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Libadalang.Implementation.Node_Env (Node => Self);
--# expr-done 865
Bound_Env := Fld_11; Inc_Ref (Bound_Env);
--# expr-start 867 '<FieldAccess for record_def at ast.py:14396>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14396
--# expr-start 866 '<FieldAccess for expression_type at ast.py:14396>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14396
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.expression_type
Fld_12 := Libadalang.Implementation.Expr_P_Expression_Type (Node => Match_3.Node, E_Info => Match_3.Info);
--# end
--# expr-done 866
      if Fld_12.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.record_def'
Fld_13 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Record_Def (Node => Fld_12.Node, Origin => Bound_Origin, E_Info => Fld_12.Info);
--# end
--# expr-done 867
Var_Expr_5 := Fld_13;
if Var_Expr_5 /= No_Entity_Base_Record_Def then
   --# scope-start
--# expr-start 869 '<FieldAccess for abstract_formal_params_for_assocs at ast.py:14396>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14396
--# expr-start 868 '<FieldAccess for components at ast.py:14396>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14396
      if Var_Expr_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Create_Internal_Entity_Component_List (Node => Var_Expr_5.Node.Base_Record_Def_F_Components, Info => Var_Expr_5.Info);
--# expr-done 868
      if Fld_14.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.abstract_formal_params_for_assocs
Fld_15 := Libadalang.Implementation.Component_List_P_Abstract_Formal_Params_For_Assocs (Node => Fld_14.Node, Assocs => Ent, Env => Bound_Env, Origin => Bound_Origin, E_Info => Fld_14.Info);
--# end
--# expr-done 869
Scope_Result_4 := Fld_15;
   Inc_Ref (Scope_Result_4);
   --# end
      Finalizer_Scope_1836;
   Result_Var_4 := Scope_Result_4;
else
   Result_Var_4 := No_Internal_Entity_Base_Formal_Param_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var_4);
Dyn_Var_Bind_Result_1 := Result_Var_4; Inc_Ref (Dyn_Var_Bind_Result_1);
--# expr-done 870
Dyn_Var_Bind_Result_2 := Dyn_Var_Bind_Result_1; Inc_Ref (Dyn_Var_Bind_Result_2);
--# expr-done 871
Let_Result_3 := Dyn_Var_Bind_Result_2; Inc_Ref (Let_Result_3);
--# end
Scope_Result_8 := Let_Result_3;
   Inc_Ref (Scope_Result_8);
   --# end
      Finalizer_Scope_1835;
            Match_Result := Scope_Result_8; Inc_Ref (Match_Result);
         when Ada_Abort_Absent .. Ada_Generic_Subp_Decl | Ada_Generic_Package_Renaming_Decl .. Ada_Constant_Present | Ada_Delta_Constraint .. Ada_Allocator | Ada_Bin_Op .. Ada_Update_Attribute_Ref | Ada_Defining_Name .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_7 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_4 := Cast_Result_7; 
--# bind _ Match_4
Let_Result_4 := No_Internal_Entity_Base_Formal_Param_Decl_Array_Type; Inc_Ref (Let_Result_4);
--# end
Scope_Result_9 := Let_Result_4;
   Inc_Ref (Scope_Result_9);
   --# end
            Match_Result := Scope_Result_9; Inc_Ref (Match_Result);
end case;
--# expr-done 872
Scope_Result_10 := Match_Result;
   Inc_Ref (Scope_Result_10);
   --# end
      Finalizer_Scope_1828;
   Result_Var_5 := Scope_Result_10;
else
   Result_Var_5 := No_Internal_Entity_Base_Formal_Param_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var_5);
Block_Var_1 := Result_Var_5; Inc_Ref (Block_Var_1);
--# bind params Block_Var_1
--# expr-start 876 '<Find at ast.py:14403>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14403
Saved := Ent; 
Fld_16 := Saved.Node;
      Find_Result := No_Entity_Basic_Assoc;
      if Fld_16 = null then
      Find_Result := No_Entity_Basic_Assoc;
      else
      declare
         Collection : constant Bare_Assoc_List := Fld_16;
      begin
         for Untyped_Item_32 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_32 := Untyped_Item_32; 
Is_Null := Bare_Item_32 = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_17 := Saved.Info;
As_Entity := (Info => Fld_17, Node => Bare_Item_32); 
   If_Result := As_Entity;
end if;
                  Item_32 := If_Result; 
   --# scope-start
               --# bind assoc Item_32
      --# expr-start 875 '<AnyQuantifier at ast.py:14404>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14404
   --# expr-start 873 '<FieldAccess for names at ast.py:14404>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14404
      if Item_32.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.names'
Fld_18 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Names (Node => Item_32.Node);
--# end
--# expr-done 873
      Quantifier_Result := False;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Fld_18;
      begin
         for Item_33 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_33
      --# expr-start 874 '<IsA OthersDesignator at ast.py:14405>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14405
Is_A := Item_33 /= null 
and then Item_33.Kind in Ada_Others_Designator_Range; 
--# expr-done 874
      Quantifier_Result := Is_A;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
--# expr-done 875
      if Quantifier_Result then
         Find_Result := Item_32;
      end if;
   --# end
      Finalizer_Scope_1838;
               exit when Quantifier_Result;
         end loop;
      end;
      end if;
--# expr-done 876
Block_Var_2 := Find_Result; 
--# bind others_assoc Block_Var_2
Var_Expr_6 := Block_Var_1;
      Inc_Ref (Var_Expr_6);
if Var_Expr_6 /= No_Internal_Entity_Base_Formal_Param_Decl_Array_Type then
   --# scope-start
   --# bind _ Var_Expr_6
--# expr-start 883 '<Map at ast.py:14410>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14410
   --# expr-start 877 '<FieldAccess for match_formals at ast.py:14410>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14410
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.match_formals
Fld_19 := Libadalang.Implementation.Ada_Node_P_Match_Formals (Node => Self, Formal_Params => Block_Var_1, Params => Ent, Is_Dottable_Subp => Block_Var);
--# end
--# expr-done 877
      declare
         Map_Result_Vec : Internal_Param_Actual_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Param_Match_Array_Access := Fld_19;
      begin
         for Item_34 of
               Collection.Items
         loop
   --# scope-start
               --# bind m Item_34
      --# expr-start 882 '<New ParamActual at ast.py:14411>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14411
--# expr-start 880 '<FieldAccess for expr at ast.py:14413>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14413
--# expr-start 879 '<FieldAccess for assoc at ast.py:14413>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14413
--# expr-start 878 '<FieldAccess for actual at ast.py:14413>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14413
Fld_21 := Item_34.Actual;
--# expr-done 878
Fld_22 := Fld_21.Assoc;
--# expr-done 879
      if Fld_22.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_23 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_22.Node, E_Info => Fld_22.Info);
--# end
--# expr-done 880
--# expr-start 881 '<FieldAccess for formal at ast.py:14412>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14412
Fld_20 := Item_34.Formal;
--# expr-done 881
New_Struct := (Actual => Fld_23, Param => Fld_20); 
--# expr-done 882
         declare
            Item_To_Append : constant Internal_Param_Actual := New_Struct;
         begin
            Internal_Param_Actual_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Param_Actual_Array
           (Items_Count => Natural (Internal_Param_Actual_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Param_Actual_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Param_Actual_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Param_Actual_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 883
Scope_Result_11 := Map_Result;
   Inc_Ref (Scope_Result_11);
   --# end
      Finalizer_Scope_1840;
   Result_Var_6 := Scope_Result_11;
else
   Result_Var_6 := No_Internal_Param_Actual_Array_Type;
end if;
      Inc_Ref (Result_Var_6);
Block_Var_3 := Result_Var_6; Inc_Ref (Block_Var_3);
--# bind explicit_matches Block_Var_3
Var_Expr_7 := Block_Var_1;
      Inc_Ref (Var_Expr_7);
if Var_Expr_7 /= No_Internal_Entity_Base_Formal_Param_Decl_Array_Type then
   --# scope-start
   --# bind _ Var_Expr_7
--# expr-start 909 '<FilterMap at ast.py:14418>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14418
      declare
         Map_Result_1_Vec : Internal_Param_Actual_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Base_Formal_Param_Decl_Array_Access := Block_Var_1;
      begin
         for Item_35 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_35
--# expr-start 884 '<Cast to GenericFormalSubpDecl at ast.py:14437>' Cast_Result_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14437
Cast_Expr_3 := Item_35; 
   if Cast_Expr_3.Node = null
      or else Cast_Expr_3.Node.Kind in Ada_Generic_Formal_Subp_Decl_Range
   then
      Cast_Result_13 := Create_Internal_Entity_Generic_Formal_Subp_Decl
        (Node => Cast_Expr_3.Node,
         Info => Cast_Expr_3.Info);
   else
         Cast_Result_13 := No_Entity_Generic_Formal_Subp_Decl;
   end if;
--# expr-done 884
Var_Expr_9 := Cast_Result_13;
if Var_Expr_9 /= No_Entity_Generic_Formal_Subp_Decl then
   --# scope-start
   --# bind fd Var_Expr_9
--# expr-start 886 '<Cast to FormalSubpDecl at ast.py:14438>' Cast_Result_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14438
--# expr-start 885 '<FieldAccess for decl at ast.py:14438>' Fld_31 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14438
      if Var_Expr_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_31 := Create_Internal_Entity_Basic_Decl (Node => Var_Expr_9.Node.Generic_Formal_F_Decl, Info => Var_Expr_9.Info);
--# expr-done 885
   if Fld_31.Node = null
      or else Fld_31.Node.Kind in Ada_Formal_Subp_Decl
   then
      Cast_Result_14 := Create_Internal_Entity_Formal_Subp_Decl
        (Node => Fld_31.Node,
         Info => Fld_31.Info);
   else
         Cast_Result_14 := No_Entity_Formal_Subp_Decl;
   end if;
--# expr-done 886
Var_Expr_10 := Cast_Result_14;
if Var_Expr_10 /= No_Entity_Formal_Subp_Decl then
   --# scope-start
   --# bind subp Var_Expr_10
--# expr-start 888 '<IsA BoxExpr, Name at ast.py:14442>' Is_A_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14442
--# expr-start 887 '<FieldAccess for default_expr at ast.py:14442>' Fld_32 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14442
      if Var_Expr_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_32 := Create_Internal_Entity_Expr (Node => Var_Expr_10.Node.Formal_Subp_Decl_F_Default_Expr, Info => Var_Expr_10.Info);
--# expr-done 887
Is_A_2 := Fld_32.Node /= null 
and then Fld_32.Node.Kind in Ada_Box_Expr_Range | Ada_Name; 
--# expr-done 888
if Is_A_2 then
   --# expr-start 893 '<Not at ast.py:14445>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14445
--# expr-start 892 '<AnyQuantifier at ast.py:14445>' Quantifier_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14445
      Quantifier_Result_1 := False;
      declare
         Collection : constant Internal_Param_Actual_Array_Access := Block_Var_3;
      begin
         for Item_36 of
               Collection.Items
         loop
   --# scope-start
               --# bind m Item_36
      --# expr-start 891 '<Eq at ast.py:14446>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14446
--# expr-start 889 '<FieldAccess for param at ast.py:14446>' Fld_33 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14446
Fld_33 := Item_36.Param;
--# expr-done 889
      Cast_Result_15 := Create_Internal_Entity
        (Node => Fld_33.Node,
         Info => Fld_33.Info);
--# expr-start 890 '<FieldAccess for defining_name at ast.py:14446>' Fld_34 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14446
      if Var_Expr_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_34 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Var_Expr_10.Node, E_Info => Var_Expr_10.Info);
--# end
--# expr-done 890
      Cast_Result_16 := Create_Internal_Entity
        (Node => Fld_34.Node,
         Info => Fld_34.Info);
Is_Equiv := Equivalent (Cast_Result_15, Cast_Result_16); 
--# expr-done 891
      Quantifier_Result_1 := Is_Equiv;
   --# end
               exit when Quantifier_Result_1;
         end loop;
      end;
--# expr-done 892
Not_Val := not (Quantifier_Result_1); 
--# expr-done 893
   If_Result_2 := Not_Val;
else
   If_Result_2 := False;
end if;
Scope_Result_13 := If_Result_2;
   --# end
   Result_Var_8 := Scope_Result_13;
else
   Result_Var_8 := False;
end if;
Scope_Result_14 := Result_Var_8;
   --# end
   Result_Var_9 := Scope_Result_14;
else
   Result_Var_9 := False;
end if;
         if Result_Var_9 then
      --# expr-start 908 '<Let at ast.py:14421>' Let_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14421
--# scope-start
--# expr-start 894 '<Cast to GenericFormalSubpDecl at ast.py:14423>' Cast_Result_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14423
Cast_Expr_1 := Item_35; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Generic_Formal_Subp_Decl_Range
   then
      Cast_Result_8 := Create_Internal_Entity_Generic_Formal_Subp_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_8 := No_Entity_Generic_Formal_Subp_Decl;
   end if;
--# expr-done 894
Decl := Cast_Result_8; 
--# bind decl Decl
--# expr-start 897 '<Cast to FormalSubpDecl at ast.py:14424>' Cast_Result_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14424
--# expr-start 896 '<FieldAccess for decl at ast.py:14424>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14424
--# expr-start 895 '<Cast to GenericFormalSubpDecl at ast.py:14424>' Cast_Result_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14424
Cast_Expr_2 := Item_35; 
   if Cast_Expr_2.Node = null
      or else Cast_Expr_2.Node.Kind in Ada_Generic_Formal_Subp_Decl_Range
   then
      Cast_Result_9 := Create_Internal_Entity_Generic_Formal_Subp_Decl
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
   else
         Cast_Result_9 := No_Entity_Generic_Formal_Subp_Decl;
   end if;
--# expr-done 895
      if Cast_Result_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_24 := Create_Internal_Entity_Basic_Decl (Node => Cast_Result_9.Node.Generic_Formal_F_Decl, Info => Cast_Result_9.Info);
--# expr-done 896
   if Fld_24.Node = null
      or else Fld_24.Node.Kind in Ada_Formal_Subp_Decl
   then
      Cast_Result_10 := Create_Internal_Entity_Formal_Subp_Decl
        (Node => Fld_24.Node,
         Info => Fld_24.Info);
   else
         Cast_Result_10 := No_Entity_Formal_Subp_Decl;
   end if;
--# expr-done 897
Subp := Cast_Result_10; 
--# bind subp Subp
--# expr-start 907 '<New ParamActual at ast.py:14426>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14426
--# expr-start 905 '<If at ast.py:14428>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14428
--# expr-start 899 '<IsA Name at ast.py:14429>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14429
--# expr-start 898 '<FieldAccess for default_expr at ast.py:14429>' Fld_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14429
      if Subp.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_30 := Create_Internal_Entity_Expr (Node => Subp.Node.Formal_Subp_Decl_F_Default_Expr, Info => Subp.Info);
--# expr-done 898
Is_A_1 := Fld_30.Node /= null 
and then Fld_30.Node.Kind in Ada_Name; 
--# expr-done 899
if Is_A_1 then
   --# expr-start 900 '<FieldAccess for default_expr at ast.py:14430>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14430
      if Subp.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_26 := Create_Internal_Entity_Expr (Node => Subp.Node.Formal_Subp_Decl_F_Default_Expr, Info => Subp.Info);
--# expr-done 900
   If_Result_1 := Fld_26;
else
--# expr-start 903 '<FieldAccess for designated_subprogram_from at ast.py:14431>' Fld_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14431
      if Subp.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 902 '<Cast to GenericInstantiation at ast.py:14432>' Cast_Result_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14432
--# expr-start 901 '<parent at ast.py:14432>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14432
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_27 := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 901
   if Fld_27.Node = null
      or else Fld_27.Node.Kind in Ada_Generic_Instantiation
   then
      Cast_Result_11 := Create_Internal_Entity_Generic_Instantiation
        (Node => Fld_27.Node,
         Info => Fld_27.Info);
   else
         Cast_Result_11 := No_Entity_Generic_Instantiation;
   end if;
--# expr-done 902
--# property-call-start FormalSubpDecl.designated_subprogram_from
Fld_28 := Libadalang.Implementation.Formal_Subp_Decl_P_Designated_Subprogram_From (Node => Subp.Node, Inst => Cast_Result_11, E_Info => Subp.Info);
--# end
--# expr-done 903
Var_Expr_8 := Fld_28;
if Var_Expr_8 /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 904 '<FieldAccess for defining_name at ast.py:14431>' Fld_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14431
      if Var_Expr_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_29 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Var_Expr_8.Node, E_Info => Var_Expr_8.Info);
--# end
--# expr-done 904
Scope_Result_12 := Fld_29;
   --# end
   Result_Var_7 := Scope_Result_12;
else
   Result_Var_7 := No_Entity_Defining_Name;
end if;
      Cast_Result_12 := Create_Internal_Entity_Expr
        (Node => Result_Var_7.Node,
         Info => Result_Var_7.Info);
   If_Result_1 := Cast_Result_12;
end if;
--# expr-done 905
--# expr-start 906 '<FieldAccess for defining_name at ast.py:14427>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14427
      if Decl.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_25 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Decl.Node, E_Info => Decl.Info);
--# end
--# expr-done 906
New_Struct_1 := (Actual => If_Result_1, Param => Fld_25); 
--# expr-done 907
Let_Result_5 := New_Struct_1; 
--# end
--# expr-done 908
         declare
            Item_To_Append : constant Internal_Param_Actual := Let_Result_5;
         begin
            Internal_Param_Actual_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result_1 := Create_Internal_Param_Actual_Array
           (Items_Count => Natural (Internal_Param_Actual_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Param_Actual_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Param_Actual_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Param_Actual_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 909
Scope_Result_15 := Map_Result_1;
   Inc_Ref (Scope_Result_15);
   --# end
      Finalizer_Scope_1842;
   Result_Var_10 := Scope_Result_15;
else
   Result_Var_10 := No_Internal_Param_Actual_Array_Type;
end if;
      Inc_Ref (Result_Var_10);
Block_Var_4 := Result_Var_10; Inc_Ref (Block_Var_4);
--# bind default_subp_matches Block_Var_4
--# expr-start 910 '<Concat at ast.py:14453>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14453
Concat_Result := Concat (Block_Var_3, Block_Var_4); 
--# expr-done 910
Block_Var_5 := Concat_Result; Inc_Ref (Block_Var_5);
--# bind given_matches Block_Var_5
Var_Expr_11 := Block_Var_2;
if Var_Expr_11 /= No_Entity_Basic_Assoc then
   --# scope-start
   --# bind oa Var_Expr_11
--# expr-start 918 '<FilterMap at ast.py:14456>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14456
   --# expr-start 911 '<FieldAccess for unpack_formals at ast.py:14456>' Fld_35 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14456
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.unpack_formals
Fld_35 := Libadalang.Implementation.Ada_Node_P_Unpack_Formals (Node => Self, Formal_Params => Block_Var_1);
--# end
--# expr-done 911
      declare
         Map_Result_2_Vec : Internal_Param_Actual_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Fld_35;
      begin
         for Item_37 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_37
         --# expr-start 915 '<Not at ast.py:14461>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14461
--# expr-start 914 '<AnyQuantifier at ast.py:14461>' Quantifier_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14461
      Quantifier_Result_2 := False;
      declare
         Collection : constant Internal_Param_Actual_Array_Access := Block_Var_5;
      begin
         for Item_38 of
               Collection.Items
         loop
   --# scope-start
               --# bind m Item_38
      --# expr-start 913 '<Eq at ast.py:14461>' Is_Equiv_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14461
--# expr-start 912 '<FieldAccess for param at ast.py:14461>' Fld_37 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14461
Fld_37 := Item_38.Param;
--# expr-done 912
      Cast_Result_17 := Create_Internal_Entity
        (Node => Fld_37.Node,
         Info => Fld_37.Info);
Cast_Expr_4 := Item_37; 
      Cast_Result_18 := Create_Internal_Entity
        (Node => Cast_Expr_4.Node,
         Info => Cast_Expr_4.Info);
Is_Equiv_1 := Equivalent (Cast_Result_17, Cast_Result_18); 
--# expr-done 913
      Quantifier_Result_2 := Is_Equiv_1;
   --# end
               exit when Quantifier_Result_2;
         end loop;
      end;
--# expr-done 914
Not_Val_1 := not (Quantifier_Result_2); 
--# expr-done 915
         if Not_Val_1 then
      --# expr-start 917 '<New ParamActual at ast.py:14457>' New_Struct_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14457
--# expr-start 916 '<FieldAccess for expr at ast.py:14459>' Fld_36 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14459
      if Var_Expr_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_36 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Var_Expr_11.Node, E_Info => Var_Expr_11.Info);
--# end
--# expr-done 916
New_Struct_2 := (Actual => Fld_36, Param => Item_37); 
--# expr-done 917
         declare
            Item_To_Append : constant Internal_Param_Actual := New_Struct_2;
         begin
            Internal_Param_Actual_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result_2 := Create_Internal_Param_Actual_Array
           (Items_Count => Natural (Internal_Param_Actual_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Internal_Param_Actual_Vectors.Get
              (Map_Result_2_Vec,
               I + Internal_Param_Actual_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Internal_Param_Actual_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 918
Scope_Result_16 := Map_Result_2;
   Inc_Ref (Scope_Result_16);
   --# end
      Finalizer_Scope_1848;
   Result_Var_11 := Scope_Result_16;
else
   Result_Var_11 := No_Internal_Param_Actual_Array_Type;
end if;
      Inc_Ref (Result_Var_11);
Block_Var_6 := Result_Var_11; Inc_Ref (Block_Var_6);
--# bind others_matches Block_Var_6
--# expr-start 919 '<Concat at ast.py:14465>' Concat_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14465
Concat_Result_1 := Concat (Block_Var_5, Block_Var_6); 
--# expr-done 919
Let_Result_6 := Concat_Result_1; Inc_Ref (Let_Result_6);
--# end
--# expr-done 920
      Property_Result := Let_Result_6;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1037;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1037;
               Finalizer_Scope_1828;
               Finalizer_Scope_1829;
               Finalizer_Scope_1830;
               Finalizer_Scope_1831;
               Finalizer_Scope_1832;
               Finalizer_Scope_1833;
               Finalizer_Scope_1834;
               Finalizer_Scope_1835;
               Finalizer_Scope_1836;
               Finalizer_Scope_1838;
               Finalizer_Scope_1840;
               Finalizer_Scope_1842;
               Finalizer_Scope_1848;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Assoc_List_P_Zip_With_Params;
--# end
   --
   --  Primitives for Bare_Basic_Decl_List
   --
   --
   --  Primitives for Bare_Case_Expr_Alternative_List
   --
   --
   --  Primitives for Bare_Case_Stmt_Alternative_List
   --
   --
   --  Primitives for Bare_Compilation_Unit_List
   --
   --
   --  Primitives for Bare_Contract_Case_Assoc_List
   --
   --
   --  Primitives for Bare_Defining_Name_List
   --
   --
   --  Primitives for Bare_Discriminant_Spec_List
   --
   --
   --  Primitives for Bare_Elsif_Expr_Part_List
   --
   --
   --  Primitives for Bare_Elsif_Stmt_Part_List
   --
   --
   --  Primitives for Bare_Enum_Literal_Decl_List
   --
   --
   --  Primitives for Bare_Expr_List
   --
   --
   --  Primitives for Bare_Expr_Alternatives_List
   --
   --
   --  Primitives for Bare_Identifier_List
   --
   --
   --  Primitives for Bare_Discriminant_Choice_List
   --
   --
   --  Primitives for Bare_Name_List
   --
   --
   --  Primitives for Bare_Parent_List
   --
   --
   --  Primitives for Bare_Param_Spec_List
   --
   --
   --  Primitives for Bare_Pragma_Node_List
   --
   --
   --  Primitives for Bare_Select_When_Part_List
   --
   --
   --  Primitives for Bare_Unconstrained_Array_Index_List
   --
   --
   --  Primitives for Bare_Variant_List
   --
   --
   --  Primitives for Bare_Aliased_Node
   --
--# property-start '[dispatcher]Aliased.as_bool' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Aliased_Node_P_As_Bool
  (Node : Bare_Aliased_Node
  )
   return Boolean
is
   Self : Bare_Aliased_Node := Bare_Aliased_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Aliased_Node (Self.Kind) is
               when Ada_Aliased_Absent =>
                  --# property-call-start Aliased.Absent.as_bool
                  Property_Result := Aliased_Absent_P_As_Bool
                    (Self
                    );
                  --# end
               when Ada_Aliased_Present =>
                  --# property-call-start Aliased.Present.as_bool
                  Property_Result := Aliased_Present_P_As_Bool
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Aliased_Node_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Aliased_Absent
   --
--# property-start Aliased.Absent.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4890
pragma Warnings (Off, "is not referenced");
 function Aliased_Absent_P_As_Bool
  (Node : Bare_Aliased_Absent
  )
   return Boolean
is
   Self : Bare_Aliased_Absent := Bare_Aliased_Absent (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aliased_Absent_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Aliased_Present
   --
--# property-start Aliased.Present.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4890
pragma Warnings (Off, "is not referenced");
 function Aliased_Present_P_As_Bool
  (Node : Bare_Aliased_Present
  )
   return Boolean
is
   Self : Bare_Aliased_Present := Bare_Aliased_Present (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aliased_Present_P_As_Bool;
--# end
   --
   --  Primitives for Bare_All_Node
   --
--# property-start '[dispatcher]All.as_bool' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_All_Node_P_As_Bool
  (Node : Bare_All_Node
  )
   return Boolean
is
   Self : Bare_All_Node := Bare_All_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_All_Node (Self.Kind) is
               when Ada_All_Absent =>
                  --# property-call-start All.Absent.as_bool
                  Property_Result := All_Absent_P_As_Bool
                    (Self
                    );
                  --# end
               when Ada_All_Present =>
                  --# property-call-start All.Present.as_bool
                  Property_Result := All_Present_P_As_Bool
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_All_Node_P_As_Bool;
--# end
   --
   --  Primitives for Bare_All_Absent
   --
--# property-start All.Absent.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4914
pragma Warnings (Off, "is not referenced");
 function All_Absent_P_As_Bool
  (Node : Bare_All_Absent
  )
   return Boolean
is
   Self : Bare_All_Absent := Bare_All_Absent (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end All_Absent_P_As_Bool;
--# end
   --
   --  Primitives for Bare_All_Present
   --
--# property-start All.Present.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4914
pragma Warnings (Off, "is not referenced");
 function All_Present_P_As_Bool
  (Node : Bare_All_Present
  )
   return Boolean
is
   Self : Bare_All_Present := Bare_All_Present (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end All_Present_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Array_Indices
   --
--# property-start '[dispatcher]ArrayIndices.ndims' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Array_Indices_P_Ndims
  (Node : Bare_Array_Indices
  )
   return Integer
is
   Self : Bare_Array_Indices := Bare_Array_Indices (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Integer;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Array_Indices (Self.Kind) is
               when Ada_Constrained_Array_Indices =>
                  --# property-call-start ConstrainedArrayIndices.ndims
                  Property_Result := Constrained_Array_Indices_P_Ndims
                    (Self
                    );
                  --# end
               when Ada_Unconstrained_Array_Indices =>
                  --# property-call-start UnconstrainedArrayIndices.ndims
                  Property_Result := Unconstrained_Array_Indices_P_Ndims
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Array_Indices_P_Ndims;
--# end
--# property-start '[dispatcher]ArrayIndices.constrain_index_expr' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Array_Indices_P_Constrain_Index_Expr
  (Node : Bare_Array_Indices
      ; Index_Expr : Internal_Entity_Expr
      ; Dim : Integer
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Array_Indices := Bare_Array_Indices (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind index_expr Index_Expr
   --# bind dim Dim
   --# bind origin Origin
   Property_Result : Logic_Equation;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Array_Indices (Self.Kind) is
               when Ada_Constrained_Array_Indices =>
                  --# property-call-start ConstrainedArrayIndices.constrain_index_expr
                  Property_Result := Constrained_Array_Indices_P_Constrain_Index_Expr
                    (Self
                        , Index_Expr
                        , Dim
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Unconstrained_Array_Indices =>
                  --# property-call-start UnconstrainedArrayIndices.constrain_index_expr
                  Property_Result := Unconstrained_Array_Indices_P_Constrain_Index_Expr
                    (Self
                        , Index_Expr
                        , Dim
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Array_Indices_P_Constrain_Index_Expr;
--# end
--# property-start '[dispatcher]ArrayIndices.index_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Array_Indices_P_Index_Type
  (Node : Bare_Array_Indices
      ; Dim : Integer
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Array_Indices := Bare_Array_Indices (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind dim Dim
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Array_Indices (Self.Kind) is
               when Ada_Constrained_Array_Indices =>
                  --# property-call-start ConstrainedArrayIndices.index_type
                  Property_Result := Constrained_Array_Indices_P_Index_Type
                    (Self
                        , Dim
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Unconstrained_Array_Indices =>
                  --# property-call-start UnconstrainedArrayIndices.index_type
                  Property_Result := Unconstrained_Array_Indices_P_Index_Type
                    (Self
                        , Dim
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Array_Indices_P_Index_Type;
--# end
--# property-start '[dispatcher]ArrayIndices.is_static' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Array_Indices_P_Is_Static
  (Node : Bare_Array_Indices
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Array_Indices := Bare_Array_Indices (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Array_Indices (Self.Kind) is
               when Ada_Constrained_Array_Indices =>
                  --# property-call-start ConstrainedArrayIndices.is_static
                  Property_Result := Constrained_Array_Indices_P_Is_Static
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Unconstrained_Array_Indices =>
                  --# property-call-start UnconstrainedArrayIndices.is_static
                  Property_Result := Unconstrained_Array_Indices_P_Is_Static
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Array_Indices_P_Is_Static;
--# end
   --
   --  Primitives for Bare_Constrained_Array_Indices
   --
      procedure Initialize_Fields_For_Constrained_Array_Indices
        (Self : Bare_Constrained_Array_Indices
         ; Constrained_Array_Indices_F_List : Bare_Constraint_List
        ) is
      begin
            Self.Constrained_Array_Indices_F_List := Constrained_Array_Indices_F_List;
      end Initialize_Fields_For_Constrained_Array_Indices;
   function Constrained_Array_Indices_F_List
     (Node : Bare_Constrained_Array_Indices) return Bare_Constraint_List
   is
   begin
         return Node.Constrained_Array_Indices_F_List;
   end;
--# property-start ConstrainedArrayIndices.ndims /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7752
pragma Warnings (Off, "is not referenced");
 function Constrained_Array_Indices_P_Ndims
  (Node : Bare_Constrained_Array_Indices
  )
   return Integer
is
   Self : Bare_Constrained_Array_Indices := Bare_Constrained_Array_Indices (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Integer;
      Fld : Bare_Constraint_List;
Len : Integer;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 922 '<length at ast.py:7752>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7752
--# expr-start 921 '<FieldAccess for list at ast.py:7752>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7752
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Constrained_Array_Indices_F_List;
--# expr-done 921
Len := Length (Fld); 
--# expr-done 922
      Property_Result := Len;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Constrained_Array_Indices_P_Ndims;
--# end
--# property-start ConstrainedArrayIndices.constrain_index_expr /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7755
pragma Warnings (Off, "is not referenced");
 function Constrained_Array_Indices_P_Constrain_Index_Expr
  (Node : Bare_Constrained_Array_Indices
      ; Index_Expr : Internal_Entity_Expr
      ; Dim : Integer
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Constrained_Array_Indices := Bare_Constrained_Array_Indices (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Constrained_Array_Indices :=
     Internal_Entity_Constrained_Array_Indices'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind index_expr Index_Expr
   --# bind dim Dim
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_594 with Inline_Always;
      Fld : Logic_Var;
Fld_1 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_2 : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_594 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_2);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_594;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 928 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 927 '<And at ast.py:7756>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7756
--# expr-start 925 '<Bind at ast.py:7757>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7757
--# expr-start 923 '<FieldAccess for expected_type_var at ast.py:7757>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7757
      if Index_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Index_Expr.Node);
--# expr-done 923
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
--# expr-start 924 '<FieldAccess for index_type at ast.py:7757>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7757
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]ArrayIndices.index_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Array_Indices_P_Index_Type (Node => Ent.Node, Dim => Dim, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 924
      Cast_Result := Create_Internal_Entity
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
Bind_Result := Solver.Create_Assign (Fld, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7757") else null)); 
--# expr-done 925
--# expr-start 926 '<FieldAccess for matches_expected_type at ast.py:7758>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7758
      if Index_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_2 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Index_Expr.Node, Origin => Origin);
--# end
--# expr-done 926
And_Pred := Create_And (Bind_Result, Fld_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7756") else null)); 
--# expr-done 927
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 928
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_594;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_594;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Constrained_Array_Indices_P_Constrain_Index_Expr;
--# end
--# property-start ConstrainedArrayIndices.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7762
pragma Warnings (Off, "is not referenced");
 function Constrained_Array_Indices_P_Xref_Equation
  (Node : Bare_Constrained_Array_Indices
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Constrained_Array_Indices := Bare_Constrained_Array_Indices (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Constrained_Array_Indices :=
     Internal_Entity_Constrained_Array_Indices'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_595 with Inline_Always;
            procedure Finalizer_Scope_1852 with Inline_Always;
            procedure Finalizer_Scope_1853 with Inline_Always;
      Var_Expr : Internal_Entity_Expr;
Fld : Internal_Entity_Constraint_List;
Fld_1 : Bare_Constraint_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Fld_3 : Logic_Equation;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Expr;
Fld_4 : Logic_Var;
Pred : Logic_Equation;
Fld_5 : Logic_Var;
Pred_1 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_6 : Logic_Var;
Pred_2 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred_2 : Logic_Equation;
Item_39 : Internal_Entity;
Bare_Item_39 : Bare_Ada_Node;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_595 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_595;
            procedure Finalizer_Scope_1852 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred_2);
            end Finalizer_Scope_1852;
            procedure Finalizer_Scope_1853 is
            begin
                     Dec_Ref (Pred);
                     Dec_Ref (Pred_1);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Pred_2);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_1853;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 943 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 942 '<LogicAll at ast.py:7763>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7763
--# expr-start 941 '<Map at ast.py:7763>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7763
   --# expr-start 929 '<FieldAccess for list at ast.py:7763>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7763
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Constraint_List (Node => Ent.Node.Constrained_Array_Indices_F_List, Info => Ent.Info);
--# expr-done 929
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Constraint_List := Fld_1;
      begin
         for Untyped_Item_39 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_39 := Untyped_Item_39; 
Is_Null := Bare_Item_39 = null; 
if Is_Null then
   If_Result := No_Entity;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_39); 
   If_Result := As_Entity;
end if;
                  Item_39 := If_Result; 
   --# scope-start
               --# bind index Item_39
      --# expr-start 940 '<And at ast.py:7765>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7765
--# expr-start 930 '<FieldAccess for sub_equation at ast.py:7765>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7765
      if Item_39.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Item_39.Node, Env => Env, Origin => Origin, E_Info => Item_39.Info);
--# end
--# expr-done 930
--# expr-start 931 '<Cast to Expr at ast.py:7766>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7766
Cast_Expr := Item_39; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Expr
   then
      Cast_Result := Create_Internal_Entity_Expr
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Expr;
   end if;
--# expr-done 931
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind expr Var_Expr
--# expr-start 939 '<And at ast.py:7768>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7768
--# expr-start 936 '<And at ast.py:7768>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7768
--# expr-start 933 '<Predicate on AdaNode.is_not_null at ast.py:7768>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7768
--# expr-start 932 '<FieldAccess for type_var at ast.py:7768>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7768
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Var_Expr.Node);
--# expr-done 932
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
Pred := Solver.Create_Predicate (Fld_4, Create_Ada_Node_P_Is_Not_Null_0_Predicate); 
--# expr-done 933
--# expr-start 935 '<Predicate on BaseTypeDecl.is_discrete_type at ast.py:7769>' Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7769
--# expr-start 934 '<FieldAccess for type_var at ast.py:7769>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7769
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Var_Expr.Node);
--# expr-done 934
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
Pred_1 := Solver.Create_Predicate (Fld_5, Create_Base_Type_Decl_P_Is_Discrete_Type_0_Predicate (Origin)); 
--# expr-done 935
And_Pred := Create_And (Pred, Pred_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7768") else null)); 
--# expr-done 936
--# expr-start 938 '<Predicate on BaseTypeDecl.is_not_root_int_type at ast.py:7770>' Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7770
--# expr-start 937 '<FieldAccess for type_var at ast.py:7770>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7770
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Var_Expr.Node);
--# expr-done 937
Fld_6.Value := No_Entity;
Entity_Vars.Reset (Fld_6);
Pred_2 := Solver.Create_Predicate (Fld_6, Create_Base_Type_Decl_P_Is_Not_Root_Int_Type_0_Predicate); 
--# expr-done 938
And_Pred_1 := Create_And (And_Pred, Pred_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7768") else null)); 
--# expr-done 939
Scope_Result := And_Pred_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1853;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7771") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred_2 := Create_And (Fld_3, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7765") else null)); 
--# expr-done 940
         declare
            Item_To_Append : constant Logic_Equation := And_Pred_2;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_1852;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 941
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7763") else null)); 
--# expr-done 942
Let_Result := Logic_Boolean_Op; Inc_Ref (Let_Result);
--# end
--# expr-done 943
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_595;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_595;
               Finalizer_Scope_1852;
               Finalizer_Scope_1853;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Constrained_Array_Indices_P_Xref_Equation;
--# end
--# property-start ConstrainedArrayIndices.index_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7776
pragma Warnings (Off, "is not referenced");
 function Constrained_Array_Indices_P_Index_Type
  (Node : Bare_Constrained_Array_Indices
      ; Dim : Integer
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Constrained_Array_Indices := Bare_Constrained_Array_Indices (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Constrained_Array_Indices :=
     Internal_Entity_Constrained_Array_Indices'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind dim Dim
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_596 with Inline_Always;
      Block_Var : Boolean;
Var_Expr : Internal_Entity;
Match_0 : Internal_Entity_Subtype_Indication;
Match_1 : Internal_Entity_Expr;
Match_2 : Internal_Entity;
Node_Parents : Bare_Ada_Node_Array_Access;
Fld : Boolean;
Find_Result : Bare_Ada_Node;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Fld_1 : Boolean;
Fld_2 : Internal_Entity_Constraint_List;
Fld_3 : Bare_Constraint_List;
Fld_4 : Internal_Entity_Info;
Get_Result : Bare_Ada_Node;
As_Entity_1 : Internal_Entity;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity;
Fld_5 : Internal_Entity_Base_Type_Decl;
Fld_6 : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
Match_Prefix : Internal_Entity;
Cast_Result_1 : Internal_Entity_Subtype_Indication;
Let_Result : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity_Expr;
Let_Result_1 : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity;
Let_Result_2 : Internal_Entity_Base_Type_Decl;
Scope_Result_2 : Internal_Entity_Base_Type_Decl;
Match_Result : Internal_Entity_Base_Type_Decl;
Scope_Result_3 : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Let_Result_3 : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_596 is
            begin
                     Dec_Ref (Node_Parents);
            end Finalizer_Scope_596;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 954 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 948 '<FieldAccess for resolve_names at ast.py:7778>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7778
--# expr-start 947 '<as_entity at ast.py:7778>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7778
--# expr-start 946 '<Find at ast.py:7778>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7778
   --# expr-start 944 '<FieldAccess for parents at ast.py:7778>' Node_Parents /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7778
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents := Parents (Self, True); 
--# expr-done 944
      Find_Result := No_Bare_Ada_Node;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents;
      begin
         for Item_40 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_40
      --# expr-start 945 '<FieldAccess for xref_entry_point at ast.py:7779>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7779
      if Item_40 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_entry_point'
Fld := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Entry_Point (Node => Item_40);
--# end
--# expr-done 945
      if Fld then
         Find_Result := Item_40;
      end if;
   --# end
               exit when Fld;
         end loop;
      end;
--# expr-done 946
Is_Null := Find_Result = null; 
if Is_Null then
   If_Result := No_Entity;
else
As_Entity := (Info => E_Info, Node => Find_Result); 
   If_Result := As_Entity;
end if;
--# expr-done 947
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.resolve_names
Fld_1 := Libadalang.Implementation.Ada_Node_P_Resolve_Names (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 948
Block_Var := Fld_1; 
--# expr-start 949 '<FieldAccess for list at ast.py:7782>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7782
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Constraint_List (Node => Ent.Node.Constrained_Array_Indices_F_List, Info => Ent.Info);
--# expr-done 949
Fld_3 := Fld_2.Node;
Get_Result := Get (Fld_3, Dim, True); 
Is_Null_1 := Get_Result = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity;
else
Fld_4 := Fld_2.Info;
As_Entity_1 := (Info => Fld_4, Node => Get_Result); 
   If_Result_1 := As_Entity_1;
end if;
Var_Expr := If_Result_1;
if Var_Expr /= No_Entity then
   --# scope-start
--# expr-start 953 '<Match at ast.py:7782>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7782
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Var_Expr; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Subtype_Indication .. Ada_Discrete_Subtype_Indication =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Subtype_Indication
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_1; 
--# bind st Match_0
--# expr-start 950 '<FieldAccess for designated_type at ast.py:7783>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7783
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_5 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Match_0.Node, Origin => Origin, E_Info => Match_0.Info);
--# end
--# expr-done 950
Let_Result := Fld_5; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Abstract_State_Decl_Expr .. Ada_Un_Op =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_2; 
--# bind e Match_1
--# expr-start 952 '<Cast to BaseTypeDecl.entity at ast.py:7784>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7784
--# expr-start 951 '<FieldAccess for type_val at ast.py:7784>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7784
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.type_val
Fld_6 := Libadalang.Implementation.Expr_P_Type_Val (Node => Match_1.Node);
--# end
--# expr-done 951
   if Fld_6.Node = null
      or else Fld_6.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld_6.Node,
         Info => Fld_6.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 952
Let_Result_1 := Cast_Result; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abort_Absent .. Ada_Elsif_Stmt_Part | Ada_Handled_Stmts .. Ada_Enum_Lit_Synth_Type_Expr | Ada_Synthetic_Type_Expr .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_3; 
--# bind _ Match_2
Let_Result_2 := No_Entity_Base_Type_Decl; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 953
Scope_Result_3 := Match_Result;
   --# end
   Result_Var := Scope_Result_3;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Let_Result_3 := Result_Var; 
--# end
--# expr-done 954
      Property_Result := Let_Result_3;
   --# end
      Finalizer_Scope_596;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_596;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Constrained_Array_Indices_P_Index_Type;
--# end
--# property-start ConstrainedArrayIndices.is_static /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7788
pragma Warnings (Off, "is not referenced");
 function Constrained_Array_Indices_P_Is_Static
  (Node : Bare_Constrained_Array_Indices
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Constrained_Array_Indices := Bare_Constrained_Array_Indices (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Constrained_Array_Indices :=
     Internal_Entity_Constrained_Array_Indices'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Match_0 : Internal_Entity_Subtype_Indication;
Match_1 : Internal_Entity_Bin_Op;
Match_2 : Internal_Entity;
Fld : Internal_Entity_Constraint_List;
Fld_1 : Bare_Constraint_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Fld_3 : Boolean;
Fld_4 : Internal_Entity_Expr;
Fld_5 : Boolean;
Fld_6 : Internal_Entity_Expr;
Fld_7 : Boolean;
If_Result_1 : Boolean;
Match_Prefix : Internal_Entity;
Cast_Result : Internal_Entity_Subtype_Indication;
Let_Result : Boolean;
Scope_Result : Boolean;
Cast_Result_1 : Internal_Entity_Bin_Op;
Let_Result_1 : Boolean;
Scope_Result_1 : Boolean;
Cast_Result_2 : Internal_Entity;
Let_Result_2 : Boolean;
Scope_Result_2 : Boolean;
Match_Result : Boolean;
Item_41 : Internal_Entity;
Bare_Item_41 : Bare_Ada_Node;
Quantifier_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 962 '<AllQuantifier at ast.py:7788>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7788
   --# expr-start 955 '<FieldAccess for list at ast.py:7788>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7788
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Constraint_List (Node => Ent.Node.Constrained_Array_Indices_F_List, Info => Ent.Info);
--# expr-done 955
Fld_1 := Fld.Node;
      Quantifier_Result := True;
      if Fld_1 = null then
      null;
      else
      declare
         Collection : constant Bare_Constraint_List := Fld_1;
      begin
         for Untyped_Item_41 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_41 := Untyped_Item_41; 
Is_Null := Bare_Item_41 = null; 
if Is_Null then
   If_Result := No_Entity;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_41); 
   If_Result := As_Entity;
end if;
                  Item_41 := If_Result; 
   --# scope-start
               --# bind t Item_41
      --# expr-start 961 '<Match at ast.py:7789>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7789
      if Item_41.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Item_41; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Subtype_Indication .. Ada_Discrete_Subtype_Indication =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Subtype_Indication
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind st Match_0
--# expr-start 956 '<FieldAccess for is_static_subtype at ast.py:7790>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7790
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SubtypeIndication.is_static_subtype
Fld_3 := Libadalang.Implementation.Subtype_Indication_P_Is_Static_Subtype (Node => Match_0.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Match_0.Info);
--# end
--# expr-done 956
Let_Result := Fld_3; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Bin_Op .. Ada_Relation_Op =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Bin_Op
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind e Match_1
--# expr-start 958 '<FieldAccess for is_static_expr at ast.py:7791>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7791
--# expr-start 957 '<FieldAccess for left at ast.py:7791>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7791
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Expr (Node => Match_1.Node.Bin_Op_F_Left, Info => Match_1.Info);
--# expr-done 957
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_5 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_4.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_4.Info);
--# end
--# expr-done 958
if Fld_5 then
   --# expr-start 960 '<FieldAccess for is_static_expr at ast.py:7791>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7791
--# expr-start 959 '<FieldAccess for right at ast.py:7791>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7791
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Expr (Node => Match_1.Node.Bin_Op_F_Right, Info => Match_1.Info);
--# expr-done 959
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_7 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_6.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_6.Info);
--# end
--# expr-done 960
   If_Result_1 := Fld_7;
else
   If_Result_1 := False;
end if;
Let_Result_1 := If_Result_1; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abort_Absent .. Ada_Null_Record_Aggregate | Ada_Box_Expr .. Ada_Enum_Lit_Synth_Type_Expr | Ada_Synthetic_Type_Expr .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind _ Match_2
Let_Result_2 := False; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 961
      Quantifier_Result := Match_Result;
   --# end
               exit when not Quantifier_Result;
         end loop;
      end;
      end if;
--# expr-done 962
      Property_Result := Quantifier_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Constrained_Array_Indices_P_Is_Static;
--# end
   --
   --  Primitives for Bare_Unconstrained_Array_Indices
   --
      procedure Initialize_Fields_For_Unconstrained_Array_Indices
        (Self : Bare_Unconstrained_Array_Indices
         ; Unconstrained_Array_Indices_F_Types : Bare_Unconstrained_Array_Index_List
        ) is
      begin
            Self.Unconstrained_Array_Indices_F_Types := Unconstrained_Array_Indices_F_Types;
      end Initialize_Fields_For_Unconstrained_Array_Indices;
   function Unconstrained_Array_Indices_F_Types
     (Node : Bare_Unconstrained_Array_Indices) return Bare_Unconstrained_Array_Index_List
   is
   begin
         return Node.Unconstrained_Array_Indices_F_Types;
   end;
--# property-start UnconstrainedArrayIndices.ndims /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7722
pragma Warnings (Off, "is not referenced");
 function Unconstrained_Array_Indices_P_Ndims
  (Node : Bare_Unconstrained_Array_Indices
  )
   return Integer
is
   Self : Bare_Unconstrained_Array_Indices := Bare_Unconstrained_Array_Indices (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Integer;
      Fld : Bare_Unconstrained_Array_Index_List;
Len : Integer;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 964 '<length at ast.py:7722>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7722
--# expr-start 963 '<FieldAccess for types at ast.py:7722>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7722
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Unconstrained_Array_Indices_F_Types;
--# expr-done 963
Len := Length (Fld); 
--# expr-done 964
      Property_Result := Len;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Unconstrained_Array_Indices_P_Ndims;
--# end
--# property-start UnconstrainedArrayIndices.constrain_index_expr /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7725
pragma Warnings (Off, "is not referenced");
 function Unconstrained_Array_Indices_P_Constrain_Index_Expr
  (Node : Bare_Unconstrained_Array_Indices
      ; Index_Expr : Internal_Entity_Expr
      ; Dim : Integer
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Unconstrained_Array_Indices := Bare_Unconstrained_Array_Indices (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Unconstrained_Array_Indices :=
     Internal_Entity_Unconstrained_Array_Indices'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind index_expr Index_Expr
   --# bind dim Dim
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_589 with Inline_Always;
      Fld : Logic_Var;
Fld_1 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_2 : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_589 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_2);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_589;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 970 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 969 '<And at ast.py:7726>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7726
--# expr-start 967 '<Bind at ast.py:7727>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7727
--# expr-start 965 '<FieldAccess for expected_type_var at ast.py:7727>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7727
      if Index_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Index_Expr.Node);
--# expr-done 965
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
--# expr-start 966 '<FieldAccess for index_type at ast.py:7727>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7727
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]ArrayIndices.index_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Array_Indices_P_Index_Type (Node => Ent.Node, Dim => Dim, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 966
      Cast_Result := Create_Internal_Entity
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
Bind_Result := Solver.Create_Assign (Fld, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7727") else null)); 
--# expr-done 967
--# expr-start 968 '<FieldAccess for matches_expected_type at ast.py:7728>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7728
      if Index_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_2 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Index_Expr.Node, Origin => Origin);
--# end
--# expr-done 968
And_Pred := Create_And (Bind_Result, Fld_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7726") else null)); 
--# expr-done 969
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 970
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_589;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_589;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Unconstrained_Array_Indices_P_Constrain_Index_Expr;
--# end
--# property-start UnconstrainedArrayIndices.index_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7732
pragma Warnings (Off, "is not referenced");
 function Unconstrained_Array_Indices_P_Index_Type
  (Node : Bare_Unconstrained_Array_Indices
      ; Dim : Integer
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Unconstrained_Array_Indices := Bare_Unconstrained_Array_Indices (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Unconstrained_Array_Indices :=
     Internal_Entity_Unconstrained_Array_Indices'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind dim Dim
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Unconstrained_Array_Index;
Fld : Internal_Entity_Unconstrained_Array_Index_List;
Fld_1 : Bare_Unconstrained_Array_Index_List;
Fld_2 : Internal_Entity_Info;
Get_Result : Bare_Unconstrained_Array_Index;
As_Entity : Internal_Entity_Unconstrained_Array_Index;
Is_Null : Boolean;
If_Result : Internal_Entity_Unconstrained_Array_Index;
Fld_3 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 973 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 971 '<FieldAccess for types at ast.py:7733>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7733
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Unconstrained_Array_Index_List (Node => Ent.Node.Unconstrained_Array_Indices_F_Types, Info => Ent.Info);
--# expr-done 971
Fld_1 := Fld.Node;
Get_Result := Get (Fld_1, Dim, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Unconstrained_Array_Index;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Get_Result); 
   If_Result := As_Entity;
end if;
Var_Expr := If_Result;
if Var_Expr /= No_Entity_Unconstrained_Array_Index then
   --# scope-start
--# expr-start 972 '<FieldAccess for designated_type at ast.py:7733>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7733
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start UnconstrainedArrayIndex.designated_type
Fld_3 := Libadalang.Implementation.Unconstrained_Array_Index_P_Designated_Type (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 972
Scope_Result := Fld_3;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 973
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Unconstrained_Array_Indices_P_Index_Type;
--# end
--# property-start UnconstrainedArrayIndices.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7736
pragma Warnings (Off, "is not referenced");
 function Unconstrained_Array_Indices_P_Xref_Equation
  (Node : Bare_Unconstrained_Array_Indices
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Unconstrained_Array_Indices := Bare_Unconstrained_Array_Indices (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Unconstrained_Array_Indices :=
     Internal_Entity_Unconstrained_Array_Indices'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_591 with Inline_Always;
            procedure Finalizer_Scope_1864 with Inline_Always;
      Fld : Internal_Entity_Unconstrained_Array_Index_List;
Fld_1 : Bare_Unconstrained_Array_Index_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity_Unconstrained_Array_Index;
Is_Null : Boolean;
If_Result : Internal_Entity_Unconstrained_Array_Index;
Fld_3 : Internal_Entity_Subtype_Indication;
Fld_4 : Logic_Equation;
Item_42 : Internal_Entity_Unconstrained_Array_Index;
Bare_Item_42 : Bare_Unconstrained_Array_Index;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_591 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_591;
            procedure Finalizer_Scope_1864 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_1864;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 979 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 978 '<LogicAll at ast.py:7737>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7737
--# expr-start 977 '<Map at ast.py:7737>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7737
   --# expr-start 974 '<FieldAccess for types at ast.py:7737>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7737
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Unconstrained_Array_Index_List (Node => Ent.Node.Unconstrained_Array_Indices_F_Types, Info => Ent.Info);
--# expr-done 974
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Unconstrained_Array_Index_List := Fld_1;
      begin
         for Untyped_Item_42 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_42 := Untyped_Item_42; 
Is_Null := Bare_Item_42 = null; 
if Is_Null then
   If_Result := No_Entity_Unconstrained_Array_Index;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_42); 
   If_Result := As_Entity;
end if;
                  Item_42 := If_Result; 
   --# scope-start
               --# bind typ Item_42
      --# expr-start 976 '<FieldAccess for sub_equation at ast.py:7738>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7738
--# expr-start 975 '<FieldAccess for subtype_indication at ast.py:7738>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7738
      if Item_42.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Subtype_Indication (Node => Item_42.Node.Unconstrained_Array_Index_F_Subtype_Indication, Info => Item_42.Info);
--# expr-done 975
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_4 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_3.Node, Env => Env, Origin => Origin, E_Info => Fld_3.Info);
--# end
--# expr-done 976
         declare
            Item_To_Append : constant Logic_Equation := Fld_4;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_1864;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 977
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7737") else null)); 
--# expr-done 978
Let_Result := Logic_Boolean_Op; Inc_Ref (Let_Result);
--# end
--# expr-done 979
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_591;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_591;
               Finalizer_Scope_1864;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Unconstrained_Array_Indices_P_Xref_Equation;
--# end
--# property-start UnconstrainedArrayIndices.is_static /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7741
pragma Warnings (Off, "is not referenced");
 function Unconstrained_Array_Indices_P_Is_Static
  (Node : Bare_Unconstrained_Array_Indices
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Unconstrained_Array_Indices := Bare_Unconstrained_Array_Indices (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Unconstrained_Array_Indices :=
     Internal_Entity_Unconstrained_Array_Indices'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Fld : Internal_Entity_Unconstrained_Array_Index_List;
Fld_1 : Bare_Unconstrained_Array_Index_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity_Unconstrained_Array_Index;
Is_Null : Boolean;
If_Result : Internal_Entity_Unconstrained_Array_Index;
Fld_3 : Internal_Entity_Subtype_Indication;
Fld_4 : Boolean;
Item_43 : Internal_Entity_Unconstrained_Array_Index;
Bare_Item_43 : Bare_Unconstrained_Array_Index;
Quantifier_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 983 '<AllQuantifier at ast.py:7741>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7741
   --# expr-start 980 '<FieldAccess for types at ast.py:7741>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7741
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Unconstrained_Array_Index_List (Node => Ent.Node.Unconstrained_Array_Indices_F_Types, Info => Ent.Info);
--# expr-done 980
Fld_1 := Fld.Node;
      Quantifier_Result := True;
      if Fld_1 = null then
      null;
      else
      declare
         Collection : constant Bare_Unconstrained_Array_Index_List := Fld_1;
      begin
         for Untyped_Item_43 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_43 := Untyped_Item_43; 
Is_Null := Bare_Item_43 = null; 
if Is_Null then
   If_Result := No_Entity_Unconstrained_Array_Index;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_43); 
   If_Result := As_Entity;
end if;
                  Item_43 := If_Result; 
   --# scope-start
               --# bind t Item_43
      --# expr-start 982 '<FieldAccess for is_static_subtype at ast.py:7742>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7742
--# expr-start 981 '<FieldAccess for subtype_indication at ast.py:7742>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7742
      if Item_43.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Subtype_Indication (Node => Item_43.Node.Unconstrained_Array_Index_F_Subtype_Indication, Info => Item_43.Info);
--# expr-done 981
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SubtypeIndication.is_static_subtype
Fld_4 := Libadalang.Implementation.Subtype_Indication_P_Is_Static_Subtype (Node => Fld_3.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_3.Info);
--# end
--# expr-done 982
      Quantifier_Result := Fld_4;
   --# end
               exit when not Quantifier_Result;
         end loop;
      end;
      end if;
--# expr-done 983
      Property_Result := Quantifier_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Unconstrained_Array_Indices_P_Is_Static;
--# end
   --
   --  Primitives for Bare_Aspect_Assoc
   --
      procedure Initialize_Fields_For_Aspect_Assoc
        (Self : Bare_Aspect_Assoc
         ; Aspect_Assoc_F_Id : Bare_Name
         ; Aspect_Assoc_F_Expr : Bare_Expr
        ) is
      begin
            Self.Aspect_Assoc_F_Id := Aspect_Assoc_F_Id;
            Self.Aspect_Assoc_F_Expr := Aspect_Assoc_F_Expr;
      end Initialize_Fields_For_Aspect_Assoc;
   function Aspect_Assoc_F_Id
     (Node : Bare_Aspect_Assoc) return Bare_Name
   is
   begin
         return Node.Aspect_Assoc_F_Id;
   end;
   function Aspect_Assoc_F_Expr
     (Node : Bare_Aspect_Assoc) return Bare_Expr
   is
   begin
         return Node.Aspect_Assoc_F_Expr;
   end;
--# property-start AspectAssoc.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9433
pragma Warnings (Off, "is not referenced");
 function Aspect_Assoc_P_Xref_Entry_Point
  (Node : Bare_Aspect_Assoc
  )
   return Boolean
is
   Self : Bare_Aspect_Assoc := Bare_Aspect_Assoc (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aspect_Assoc_P_Xref_Entry_Point;
--# end
--# property-start AspectAssoc.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9436
pragma Warnings (Off, "is not referenced");
 function Aspect_Assoc_P_Xref_Equation
  (Node : Bare_Aspect_Assoc
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Aspect_Assoc := Bare_Aspect_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Aspect_Assoc :=
     Internal_Entity_Aspect_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_755 with Inline_Always;
            procedure Finalizer_Scope_1866 with Inline_Always;
            procedure Finalizer_Scope_1868 with Inline_Always;
      Block_Var : Bare_Ada_Node;
Expr_Equation_Env : Lexical_Env := Empty_Env;
Var_Expr : Bare_Type_Decl;
Var_Expr_1 : Internal_Entity_Expr;
Fld : Bare_Ada_Node;
Fld_1 : Bare_Ada_Node;
Fld_2 : Bare_Ada_Node;
Fld_3 : Internal_Entity_Name;
Sym : Symbol_Type;
Fld_4 : Boolean;
Fld_5 : Internal_Entity_Expr;
Cast_Result : Internal_Entity_Aggregate;
Fld_6 : Internal_Entity_Assoc_List;
Fld_7 : Internal_Single_Actual_Array_Access;
Fld_8 : Internal_Entity_Basic_Assoc;
Fld_9 : Internal_Entity_Expr;
Cast_Result_1 : Internal_Entity_Name;
Fld_10 : Logic_Equation;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Fld_11 : Internal_Entity_Name;
Fld_12 : Symbol_Type;
Sym_1 : Symbol_Type;
Is_Equal : Boolean;
Fld_13 : Internal_Entity_Name;
Fld_14 : Symbol_Type;
Sym_2 : Symbol_Type;
Is_Equal_1 : Boolean;
If_Result : Boolean;
Fld_15 : Internal_Entity_Name;
Fld_16 : Symbol_Type;
Sym_3 : Symbol_Type;
Is_Equal_2 : Boolean;
If_Result_1 : Boolean;
Fld_17 : Internal_Entity_Name;
Fld_18 : Symbol_Type;
Sym_4 : Symbol_Type;
Is_Equal_3 : Boolean;
If_Result_2 : Boolean;
Fld_19 : Internal_Entity_Name;
Fld_20 : Symbol_Type;
Sym_5 : Symbol_Type;
Is_Equal_4 : Boolean;
If_Result_3 : Boolean;
Fld_21 : Internal_Entity_Name;
Fld_22 : Symbol_Type;
Sym_6 : Symbol_Type;
Is_Equal_5 : Boolean;
If_Result_4 : Boolean;
Fld_23 : Internal_Entity_Name;
Fld_24 : Symbol_Type;
Sym_7 : Symbol_Type;
Is_Equal_6 : Boolean;
If_Result_5 : Boolean;
Fld_25 : Internal_Entity_Name;
Fld_26 : Symbol_Type;
Sym_8 : Symbol_Type;
Is_Equal_7 : Boolean;
If_Result_6 : Boolean;
Cast_Expr : Bare_Ada_Node;
Cast_Result_2 : Bare_Type_Decl;
Fld_27 : Bare_Type_Def;
Fld_28 : Lexical_Env := Empty_Env;
Fld_29 : Lexical_Env := Empty_Env;
Fld_30 : Internal_Entity_Name;
Fld_31 : Symbol_Type;
Sym_9 : Symbol_Type;
Is_Equal_8 : Boolean;
Fld_32 : Internal_Entity_Name;
Fld_33 : Symbol_Type;
Sym_10 : Symbol_Type;
Is_Equal_9 : Boolean;
If_Result_7 : Boolean;
Cast_Expr_1 : Bare_Ada_Node;
Cast_Result_3 : Bare_Type_Decl;
Fld_34 : Bare_Type_Def;
Scope_Result : Bare_Type_Def;
Result_Var : Bare_Type_Def;
Is_A : Boolean;
If_Result_8 : Boolean;
If_Result_9 : Lexical_Env := Empty_Env;
Fld_35 : Bare_Expr;
Fld_36 : Logic_Var;
Fld_37 : Internal_Entity_Base_Type_Decl;
Cast_Result_4 : Internal_Entity;
Bind_Result : Logic_Equation;
Bound_Env : Lexical_Env := Empty_Env;
Fld_38 : Internal_Entity_Expr;
Fld_39 : Logic_Equation;
Dyn_Var_Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_40 : Bare_Expr;
Fld_41 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
Fld_42 : Internal_Entity_Name;
Sym_11 : Symbol_Type;
Fld_43 : Boolean;
Fld_44 : Internal_Entity_Expr;
Fld_45 : Logic_Equation;
Fld_46 : Internal_Entity_Name;
Sym_12 : Symbol_Type;
Fld_47 : Boolean;
Fld_48 : Internal_Entity_Expr;
Cast_Result_5 : Internal_Entity_Name;
Fld_49 : Logic_Equation;
Fld_50 : Internal_Entity_Expr;
Cast_Result_6 : Internal_Entity_Name;
Fld_51 : Logic_Var;
Cast_Expr_2 : Bare_Ada_Node;
Cast_Result_7 : Bare_Base_Type_Decl;
As_Entity : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
If_Result_10 : Internal_Entity_Base_Type_Decl;
Pred : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_52 : Internal_Entity_Name;
Sym_13 : Symbol_Type;
Fld_53 : Boolean;
Fld_54 : Internal_Entity_Name;
Sym_14 : Symbol_Type;
Fld_55 : Boolean;
If_Result_11 : Boolean;
True_Rel : Logic_Equation;
Fld_56 : Internal_Entity_Expr;
Fld_57 : Logic_Equation;
Fld_58 : Internal_Entity_Expr;
Is_A_1 : Boolean;
If_Result_12 : Logic_Equation;
Fld_59 : Internal_Entity_Name;
Sym_15 : Symbol_Type;
Fld_60 : Boolean;
True_Rel_1 : Logic_Equation;
Fld_61 : Internal_Entity_Name;
Sym_16 : Symbol_Type;
Fld_62 : Boolean;
Fld_63 : Logic_Equation;
Fld_64 : Internal_Entity_Expr;
Fld_65 : Logic_Equation;
Scope_Result_1 : Logic_Equation;
True_Rel_2 : Logic_Equation;
Result_Var_1 : Logic_Equation;
True_Rel_3 : Logic_Equation;
Or_Pred : Logic_Equation;
If_Result_13 : Logic_Equation;
If_Result_14 : Logic_Equation;
If_Result_15 : Logic_Equation;
If_Result_16 : Logic_Equation;
If_Result_17 : Logic_Equation;
If_Result_18 : Logic_Equation;
If_Result_19 : Logic_Equation;
Let_Result_1 : Logic_Equation;
            procedure Finalizer_Scope_755 is
            begin
                     Dec_Ref (Fld_7);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Fld_28);
                     Dec_Ref (Fld_29);
                     Dec_Ref (If_Result_9);
                     Dec_Ref (Expr_Equation_Env);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_39);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_41);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
                     Dec_Ref (Fld_45);
                     Dec_Ref (Fld_49);
                     Dec_Ref (Pred);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Fld_57);
                     Dec_Ref (If_Result_12);
                     Dec_Ref (True_Rel_1);
                     Dec_Ref (Fld_63);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (True_Rel_2);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (True_Rel_3);
                     Dec_Ref (Or_Pred);
                     Dec_Ref (If_Result_13);
                     Dec_Ref (If_Result_14);
                     Dec_Ref (If_Result_15);
                     Dec_Ref (If_Result_16);
                     Dec_Ref (If_Result_17);
                     Dec_Ref (If_Result_18);
                     Dec_Ref (If_Result_19);
                     Dec_Ref (Let_Result_1);
            end Finalizer_Scope_755;
            procedure Finalizer_Scope_1866 is
            begin
                     Dec_Ref (Fld_10);
            end Finalizer_Scope_1866;
            procedure Finalizer_Scope_1868 is
            begin
                     Dec_Ref (Fld_65);
            end Finalizer_Scope_1868;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1100 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 986 '<parent at ast.py:9437>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9437
--# expr-start 985 '<parent at ast.py:9437>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9437
--# expr-start 984 '<parent at ast.py:9437>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9437
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 984
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Parent; 
--# expr-done 985
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Fld_1.Parent; 
--# expr-done 986
Block_Var := Fld_2; 
--# bind target Block_Var
--# expr-start 1099 '<Cond at ast.py:9438>' If_Result_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9438
--# expr-start 989 '<FieldAccess for name_is at ast.py:9440>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9440
--# expr-start 987 '<FieldAccess for id at ast.py:9440>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9440
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 987
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 988 '<Symbol Iterable at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Iterable); 
--# expr-done 988
--# property-call-start Name.name_is
Fld_4 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_3.Node, Sym => Sym);
--# end
--# expr-done 989
if Fld_4 then
   --# expr-start 999 '<LogicAll at ast.py:9441>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9441
--# expr-start 998 '<Map at ast.py:9441>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9441
   --# expr-start 993 '<FieldAccess for unpacked_params at ast.py:9441>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9441
--# expr-start 992 '<FieldAccess for assocs at ast.py:9441>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9441
--# expr-start 991 '<Cast to Aggregate at ast.py:9441>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9441
--# expr-start 990 '<FieldAccess for expr at ast.py:9441>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9441
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Expr (Node => Ent.Node.Aspect_Assoc_F_Expr, Info => Ent.Info);
--# expr-done 990
   if Fld_5.Node = null
      or else Fld_5.Node.Kind in Ada_Aggregate_Range
   then
      Cast_Result := Create_Internal_Entity_Aggregate
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
   else
         Cast_Result := No_Entity_Aggregate;
   end if;
--# expr-done 991
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Assoc_List (Node => Cast_Result.Node.Base_Aggregate_F_Assocs, Info => Cast_Result.Info);
--# expr-done 992
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AssocList.unpacked_params
Fld_7 := Libadalang.Implementation.Assoc_List_P_Unpacked_Params (Node => Fld_6.Node, E_Info => Fld_6.Info);
--# end
--# expr-done 993
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Single_Actual_Array_Access := Fld_7;
      begin
         for Item_44 of
               Collection.Items
         loop
   --# scope-start
               --# bind sa Item_44
      --# expr-start 997 '<FieldAccess for xref_no_overloading at ast.py:9443>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9443
--# expr-start 996 '<Cast to Name at ast.py:9443>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9443
--# expr-start 995 '<FieldAccess for expr at ast.py:9443>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9443
--# expr-start 994 '<FieldAccess for assoc at ast.py:9443>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9443
Fld_8 := Item_44.Assoc;
--# expr-done 994
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_9 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_8.Node, E_Info => Fld_8.Info);
--# end
--# expr-done 995
   if Fld_9.Node = null
      or else Fld_9.Node.Kind in Ada_Name
   then
      Cast_Result_1 := Create_Internal_Entity_Name
        (Node => Fld_9.Node,
         Info => Fld_9.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 996
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_10 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Cast_Result_1.Node, Sequential => False, Env => Env, Origin => Origin, E_Info => Cast_Result_1.Info);
--# end
--# expr-done 997
         declare
            Item_To_Append : constant Logic_Equation := Fld_10;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_1866;
         end loop;
      end;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 998
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9441") else null)); 
--# expr-done 999
   If_Result_19 := Logic_Boolean_Op;
else
--# expr-start 1003 '<Eq at ast.py:9448>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
--# expr-start 1001 '<FieldAccess for name_symbol at ast.py:9448>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
--# expr-start 1000 '<FieldAccess for id at ast.py:9448>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 1000
      if Fld_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_12 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_11.Node);
--# end
--# expr-done 1001
--# expr-start 1002 '<Symbol Pre at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Pre); 
--# expr-done 1002
Is_Equal := Fld_12 = Sym_1; 
--# expr-done 1003
if Is_Equal then
   If_Result := True;
else
   --# expr-start 1007 '<Eq at ast.py:9448>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
--# expr-start 1005 '<FieldAccess for name_symbol at ast.py:9448>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
--# expr-start 1004 '<FieldAccess for id at ast.py:9448>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 1004
      if Fld_13.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_14 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_13.Node);
--# end
--# expr-done 1005
--# expr-start 1006 '<Symbol Post at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Post); 
--# expr-done 1006
Is_Equal_1 := Fld_14 = Sym_2; 
--# expr-done 1007
   If_Result := Is_Equal_1;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 1011 '<Eq at ast.py:9448>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
--# expr-start 1009 '<FieldAccess for name_symbol at ast.py:9448>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
--# expr-start 1008 '<FieldAccess for id at ast.py:9448>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 1008
      if Fld_15.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_16 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_15.Node);
--# end
--# expr-done 1009
--# expr-start 1010 '<Symbol Refined_Post at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Refined_Post); 
--# expr-done 1010
Is_Equal_2 := Fld_16 = Sym_3; 
--# expr-done 1011
   If_Result_1 := Is_Equal_2;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
   --# expr-start 1015 '<Eq at ast.py:9448>' Is_Equal_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
--# expr-start 1013 '<FieldAccess for name_symbol at ast.py:9448>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
--# expr-start 1012 '<FieldAccess for id at ast.py:9448>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_17 := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 1012
      if Fld_17.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_18 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_17.Node);
--# end
--# expr-done 1013
--# expr-start 1014 '<Symbol Type_Invariant at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Type_Invariant); 
--# expr-done 1014
Is_Equal_3 := Fld_18 = Sym_4; 
--# expr-done 1015
   If_Result_2 := Is_Equal_3;
end if;
if If_Result_2 then
   If_Result_3 := True;
else
   --# expr-start 1019 '<Eq at ast.py:9448>' Is_Equal_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
--# expr-start 1017 '<FieldAccess for name_symbol at ast.py:9448>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
--# expr-start 1016 '<FieldAccess for id at ast.py:9448>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 1016
      if Fld_19.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_20 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_19.Node);
--# end
--# expr-done 1017
--# expr-start 1018 '<Symbol Invariant at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Invariant); 
--# expr-done 1018
Is_Equal_4 := Fld_20 = Sym_5; 
--# expr-done 1019
   If_Result_3 := Is_Equal_4;
end if;
if If_Result_3 then
   If_Result_4 := True;
else
   --# expr-start 1023 '<Eq at ast.py:9448>' Is_Equal_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
--# expr-start 1021 '<FieldAccess for name_symbol at ast.py:9448>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
--# expr-start 1020 '<FieldAccess for id at ast.py:9448>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_21 := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 1020
      if Fld_21.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_22 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_21.Node);
--# end
--# expr-done 1021
--# expr-start 1022 '<Symbol Predicate at ???>' Sym_6 None
Sym_6 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Predicate); 
--# expr-done 1022
Is_Equal_5 := Fld_22 = Sym_6; 
--# expr-done 1023
   If_Result_4 := Is_Equal_5;
end if;
if If_Result_4 then
   If_Result_5 := True;
else
   --# expr-start 1027 '<Eq at ast.py:9448>' Is_Equal_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
--# expr-start 1025 '<FieldAccess for name_symbol at ast.py:9448>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
--# expr-start 1024 '<FieldAccess for id at ast.py:9448>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_23 := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 1024
      if Fld_23.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_24 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_23.Node);
--# end
--# expr-done 1025
--# expr-start 1026 '<Symbol Static_Predicate at ???>' Sym_7 None
Sym_7 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Static_Predicate); 
--# expr-done 1026
Is_Equal_6 := Fld_24 = Sym_7; 
--# expr-done 1027
   If_Result_5 := Is_Equal_6;
end if;
if If_Result_5 then
   If_Result_6 := True;
else
   --# expr-start 1031 '<Eq at ast.py:9448>' Is_Equal_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
--# expr-start 1029 '<FieldAccess for name_symbol at ast.py:9448>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
--# expr-start 1028 '<FieldAccess for id at ast.py:9448>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9448
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_25 := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 1028
      if Fld_25.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_26 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_25.Node);
--# end
--# expr-done 1029
--# expr-start 1030 '<Symbol Dynamic_Predicate at ???>' Sym_8 None
Sym_8 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Dynamic_Predicate); 
--# expr-done 1030
Is_Equal_7 := Fld_26 = Sym_8; 
--# expr-done 1031
   If_Result_6 := Is_Equal_7;
end if;
if If_Result_6 then
   --# expr-start 1059 '<Let at ast.py:9452>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9452
--# scope-start
--# expr-start 1047 '<If at ast.py:9453>' If_Result_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9453
--# expr-start 1035 '<Eq at ast.py:9458>' Is_Equal_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9458
--# expr-start 1033 '<FieldAccess for name_symbol at ast.py:9458>' Fld_31 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9458
--# expr-start 1032 '<FieldAccess for id at ast.py:9458>' Fld_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9458
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_30 := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 1032
      if Fld_30.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_31 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_30.Node);
--# end
--# expr-done 1033
--# expr-start 1034 '<Symbol Pre at ???>' Sym_9 None
Sym_9 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Pre); 
--# expr-done 1034
Is_Equal_8 := Fld_31 = Sym_9; 
--# expr-done 1035
if Is_Equal_8 then
   If_Result_7 := True;
else
   --# expr-start 1039 '<Eq at ast.py:9458>' Is_Equal_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9458
--# expr-start 1037 '<FieldAccess for name_symbol at ast.py:9458>' Fld_33 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9458
--# expr-start 1036 '<FieldAccess for id at ast.py:9458>' Fld_32 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9458
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_32 := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 1036
      if Fld_32.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_33 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_32.Node);
--# end
--# expr-done 1037
--# expr-start 1038 '<Symbol Post at ???>' Sym_10 None
Sym_10 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Post); 
--# expr-done 1038
Is_Equal_9 := Fld_33 = Sym_10; 
--# expr-done 1039
   If_Result_7 := Is_Equal_9;
end if;
if If_Result_7 then
   --# expr-start 1042 '<IsA AccessToSubpDef at ast.py:9459>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9459
--# expr-start 1040 '<Cast to TypeDecl at ast.py:9459>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9459
Cast_Expr_1 := Block_Var; 
   if Cast_Expr_1 = null
      or else Cast_Expr_1.Kind in Ada_Type_Decl_Range
   then
      Cast_Result_3 := Cast_Expr_1;
   else
         Cast_Result_3 := No_Bare_Ada_Node;
   end if;
--# expr-done 1040
Var_Expr := Cast_Result_3;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 1041 '<FieldAccess for type_def at ast.py:9459>' Fld_34 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9459
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_34 := Var_Expr.Type_Decl_F_Type_Def;
--# expr-done 1041
Scope_Result := Fld_34;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node;
end if;
Is_A := Result_Var /= null 
and then Result_Var.Kind in Ada_Access_To_Subp_Def_Range; 
--# expr-done 1042
   If_Result_8 := Is_A;
else
   If_Result_8 := False;
end if;
if If_Result_8 then
   --# expr-start 1045 '<FieldAccess for children_env at ast.py:9460>' Fld_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9460
--# expr-start 1044 '<FieldAccess for type_def at ast.py:9460>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9460
--# expr-start 1043 '<Cast to TypeDecl at ast.py:9460>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9460
Cast_Expr := Block_Var; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_Type_Decl_Range
   then
      Cast_Result_2 := Cast_Expr;
   else
         Cast_Result_2 := No_Bare_Ada_Node;
   end if;
--# expr-done 1043
      if Cast_Result_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_27 := Cast_Result_2.Type_Decl_F_Type_Def;
--# expr-done 1044
      if Fld_27 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_28 := Libadalang.Implementation.Children_Env (Node => Fld_27);
--# expr-done 1045
   If_Result_9 := Fld_28;
else
   --# expr-start 1046 '<FieldAccess for node_env at ast.py:9461>' Fld_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9461
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_29 := Libadalang.Implementation.Node_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 1046
   If_Result_9 := Fld_29;
end if;
      Inc_Ref (If_Result_9);
--# expr-done 1047
Expr_Equation_Env := If_Result_9; Inc_Ref (Expr_Equation_Env);
--# bind expr_equation_env Expr_Equation_Env
--# expr-start 1058 '<And at ast.py:9463>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9463
--# expr-start 1055 '<And at ast.py:9463>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9463
--# expr-start 1051 '<Bind at ast.py:9463>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9463
--# expr-start 1049 '<FieldAccess for expected_type_var at ast.py:9463>' Fld_36 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9463
--# expr-start 1048 '<FieldAccess for expr at ast.py:9463>' Fld_35 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9463
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_35 := Self.Aspect_Assoc_F_Expr;
--# expr-done 1048
      if Fld_35 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_36 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_35);
--# expr-done 1049
Fld_36.Value := No_Entity;
Entity_Vars.Reset (Fld_36);
--# expr-start 1050 '<FieldAccess for bool_type at ast.py:9463>' Fld_37 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9463
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_37 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 1050
      Cast_Result_4 := Create_Internal_Entity
        (Node => Fld_37.Node,
         Info => Fld_37.Info);
Bind_Result := Solver.Create_Assign (Fld_36, Cast_Result_4, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9463") else null)); 
--# expr-done 1051
--# expr-start 1054 '<bind at ast.py:9464>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9464
Bound_Env := Expr_Equation_Env; Inc_Ref (Bound_Env);
--# expr-start 1053 '<FieldAccess for sub_equation at ast.py:9465>' Fld_39 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9465
--# expr-start 1052 '<FieldAccess for expr at ast.py:9465>' Fld_38 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9465
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_38 := Create_Internal_Entity_Expr (Node => Ent.Node.Aspect_Assoc_F_Expr, Info => Ent.Info);
--# expr-done 1052
      if Fld_38.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_39 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_38.Node, Env => Bound_Env, Origin => Origin, E_Info => Fld_38.Info);
--# end
--# expr-done 1053
Dyn_Var_Bind_Result := Fld_39; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 1054
And_Pred := Create_And (Bind_Result, Dyn_Var_Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9463") else null)); 
--# expr-done 1055
--# expr-start 1057 '<FieldAccess for matches_expected_formal_prim_type at ast.py:9466>' Fld_41 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9466
--# expr-start 1056 '<FieldAccess for expr at ast.py:9466>' Fld_40 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9466
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_40 := Self.Aspect_Assoc_F_Expr;
--# expr-done 1056
      if Fld_40 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_prim_type
Fld_41 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Prim_Type (Node => Fld_40, Origin => Origin);
--# end
--# expr-done 1057
And_Pred_1 := Create_And (And_Pred, Fld_41, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9463") else null)); 
--# expr-done 1058
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 1059
   If_Result_18 := Let_Result;
else
--# expr-start 1062 '<FieldAccess for name_is at ast.py:9469>' Fld_43 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9469
--# expr-start 1060 '<FieldAccess for id at ast.py:9469>' Fld_42 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9469
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_42 := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 1060
      if Fld_42.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1061 '<Symbol Contract_Cases at ???>' Sym_11 None
Sym_11 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Contract_Cases); 
--# expr-done 1061
--# property-call-start Name.name_is
Fld_43 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_42.Node, Sym => Sym_11);
--# end
--# expr-done 1062
if Fld_43 then
   --# expr-start 1064 '<FieldAccess for sub_equation at ast.py:9470>' Fld_45 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9470
--# expr-start 1063 '<FieldAccess for expr at ast.py:9470>' Fld_44 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9470
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_44 := Create_Internal_Entity_Expr (Node => Ent.Node.Aspect_Assoc_F_Expr, Info => Ent.Info);
--# expr-done 1063
      if Fld_44.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_45 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_44.Node, Env => Env, Origin => Origin, E_Info => Fld_44.Info);
--# end
--# expr-done 1064
   If_Result_17 := Fld_45;
else
--# expr-start 1067 '<FieldAccess for name_is at ast.py:9473>' Fld_47 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9473
--# expr-start 1065 '<FieldAccess for id at ast.py:9473>' Fld_46 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9473
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_46 := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 1065
      if Fld_46.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1066 '<Symbol Put_Image at ???>' Sym_12 None
Sym_12 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Put_Image); 
--# expr-done 1066
--# property-call-start Name.name_is
Fld_47 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_46.Node, Sym => Sym_12);
--# end
--# expr-done 1067
if Fld_47 then
   --# expr-start 1077 '<And at ast.py:9474>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9474
--# expr-start 1070 '<FieldAccess for xref_no_overloading at ast.py:9474>' Fld_49 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9474
--# expr-start 1069 '<Cast to Name at ast.py:9474>' Cast_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9474
--# expr-start 1068 '<FieldAccess for expr at ast.py:9474>' Fld_48 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9474
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_48 := Create_Internal_Entity_Expr (Node => Ent.Node.Aspect_Assoc_F_Expr, Info => Ent.Info);
--# expr-done 1068
   if Fld_48.Node = null
      or else Fld_48.Node.Kind in Ada_Name
   then
      Cast_Result_5 := Create_Internal_Entity_Name
        (Node => Fld_48.Node,
         Info => Fld_48.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 1069
      if Cast_Result_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_49 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Cast_Result_5.Node, Sequential => False, All_Els => True, Env => Env, Origin => Origin, E_Info => Cast_Result_5.Info);
--# end
--# expr-done 1070
--# expr-start 1076 '<Predicate on BasicDecl.is_put_image_subprogram_for_type at ast.py:9476>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9476
--# expr-start 1073 '<FieldAccess for ref_var at ast.py:9478>' Fld_51 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9478
--# expr-start 1072 '<Cast to Name at ast.py:9478>' Cast_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9478
--# expr-start 1071 '<FieldAccess for expr at ast.py:9478>' Fld_50 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9478
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_50 := Create_Internal_Entity_Expr (Node => Ent.Node.Aspect_Assoc_F_Expr, Info => Ent.Info);
--# expr-done 1071
   if Fld_50.Node = null
      or else Fld_50.Node.Kind in Ada_Name
   then
      Cast_Result_6 := Create_Internal_Entity_Name
        (Node => Fld_50.Node,
         Info => Fld_50.Info);
   else
         Cast_Result_6 := No_Entity_Name;
   end if;
--# expr-done 1072
      if Cast_Result_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_51 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Cast_Result_6.Node);
--# end
--# expr-done 1073
Fld_51.Value := No_Entity;
Entity_Vars.Reset (Fld_51);
--# expr-start 1075 '<as_entity at ast.py:9479>' If_Result_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9479
--# expr-start 1074 '<Cast to BaseTypeDecl at ast.py:9479>' Cast_Result_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9479
Cast_Expr_2 := Block_Var; 
   if Cast_Expr_2 = null
      or else Cast_Expr_2.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_7 := Cast_Expr_2;
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 1074
Is_Null := Cast_Result_7 = null; 
if Is_Null then
   If_Result_10 := No_Entity_Base_Type_Decl;
else
As_Entity := (Info => E_Info, Node => Cast_Result_7); 
   If_Result_10 := As_Entity;
end if;
--# expr-done 1075
Pred := Solver.Create_Predicate (Fld_51, Create_Basic_Decl_P_Is_Put_Image_Subprogram_For_Type_0_Predicate (If_Result_10)); 
--# expr-done 1076
And_Pred_2 := Create_And (Fld_49, Pred, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9474") else null)); 
--# expr-done 1077
   If_Result_16 := And_Pred_2;
else
--# expr-start 1080 '<FieldAccess for name_is at ast.py:9484>' Fld_53 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9484
--# expr-start 1078 '<FieldAccess for id at ast.py:9484>' Fld_52 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9484
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_52 := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 1078
      if Fld_52.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1079 '<Symbol Global at ???>' Sym_13 None
Sym_13 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Global); 
--# expr-done 1079
--# property-call-start Name.name_is
Fld_53 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_52.Node, Sym => Sym_13);
--# end
--# expr-done 1080
if Fld_53 then
   If_Result_11 := True;
else
   --# expr-start 1083 '<FieldAccess for name_is at ast.py:9484>' Fld_55 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9484
--# expr-start 1081 '<FieldAccess for id at ast.py:9484>' Fld_54 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9484
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_54 := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 1081
      if Fld_54.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1082 '<Symbol Refined_Global at ???>' Sym_14 None
Sym_14 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Refined_Global); 
--# expr-done 1082
--# property-call-start Name.name_is
Fld_55 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_54.Node, Sym => Sym_14);
--# end
--# expr-done 1083
   If_Result_11 := Fld_55;
end if;
if If_Result_11 then
   --# expr-start 1088 '<If at ast.py:9485>' If_Result_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9485
--# expr-start 1085 '<IsA NullLiteral at ast.py:9486>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9486
--# expr-start 1084 '<FieldAccess for expr at ast.py:9486>' Fld_58 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9486
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_58 := Create_Internal_Entity_Expr (Node => Ent.Node.Aspect_Assoc_F_Expr, Info => Ent.Info);
--# expr-done 1084
Is_A_1 := Fld_58.Node /= null 
and then Fld_58.Node.Kind in Ada_Null_Literal_Range; 
--# expr-done 1085
if Is_A_1 then
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9487") else null)); 
   If_Result_12 := True_Rel;
else
   --# expr-start 1087 '<FieldAccess for sub_equation at ast.py:9488>' Fld_57 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9488
--# expr-start 1086 '<FieldAccess for expr at ast.py:9488>' Fld_56 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9488
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_56 := Create_Internal_Entity_Expr (Node => Ent.Node.Aspect_Assoc_F_Expr, Info => Ent.Info);
--# expr-done 1086
      if Fld_56.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_57 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_56.Node, Env => Env, Origin => Origin, E_Info => Fld_56.Info);
--# end
--# expr-done 1087
   If_Result_12 := Fld_57;
end if;
      Inc_Ref (If_Result_12);
--# expr-done 1088
   If_Result_15 := If_Result_12;
else
--# expr-start 1091 '<FieldAccess for name_is at ast.py:9494>' Fld_60 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9494
--# expr-start 1089 '<FieldAccess for id at ast.py:9494>' Fld_59 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9494
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_59 := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 1089
      if Fld_59.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1090 '<Symbol Convention at ???>' Sym_15 None
Sym_15 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Convention); 
--# expr-done 1090
--# property-call-start Name.name_is
Fld_60 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_59.Node, Sym => Sym_15);
--# end
--# expr-done 1091
if Fld_60 then
   True_Rel_1 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9494") else null)); 
   If_Result_14 := True_Rel_1;
else
--# expr-start 1094 '<FieldAccess for name_is at ast.py:9496>' Fld_62 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9496
--# expr-start 1092 '<FieldAccess for id at ast.py:9496>' Fld_61 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9496
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_61 := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 1092
      if Fld_61.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1093 '<Symbol Stable_Properties at ???>' Sym_16 None
Sym_16 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Stable_Properties); 
--# expr-done 1093
--# property-call-start Name.name_is
Fld_62 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_61.Node, Sym => Sym_16);
--# end
--# expr-done 1094
if Fld_62 then
   --# expr-start 1095 '<FieldAccess for stable_properties_assoc_equation at ast.py:9497>' Fld_63 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9497
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AspectAssoc.stable_properties_assoc_equation
Fld_63 := Libadalang.Implementation.Aspect_Assoc_P_Stable_Properties_Assoc_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 1095
   If_Result_13 := Fld_63;
else
   --# expr-start 1098 '<Or at ast.py:9502>' Or_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9502
--# expr-start 1096 '<FieldAccess for expr at ast.py:9502>' Fld_64 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9502
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_64 := Create_Internal_Entity_Expr (Node => Ent.Node.Aspect_Assoc_F_Expr, Info => Ent.Info);
--# expr-done 1096
Var_Expr_1 := Fld_64;
if Var_Expr_1 /= No_Entity_Expr then
   --# scope-start
   --# bind e Var_Expr_1
--# expr-start 1097 '<FieldAccess for sub_equation at ast.py:9502>' Fld_65 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9502
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_65 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr_1.Node, Env => Env, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 1097
Scope_Result_1 := Fld_65;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_1868;
   Result_Var_1 := Scope_Result_1;
else
   True_Rel_2 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9502") else null)); 
   Result_Var_1 := True_Rel_2;
end if;
      Inc_Ref (Result_Var_1);
True_Rel_3 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9503") else null)); 
Or_Pred := Create_Or (Result_Var_1, True_Rel_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9502") else null)); 
--# expr-done 1098
   If_Result_13 := Or_Pred;
end if;
      Inc_Ref (If_Result_13);
   If_Result_14 := If_Result_13;
end if;
      Inc_Ref (If_Result_14);
   If_Result_15 := If_Result_14;
end if;
      Inc_Ref (If_Result_15);
   If_Result_16 := If_Result_15;
end if;
      Inc_Ref (If_Result_16);
   If_Result_17 := If_Result_16;
end if;
      Inc_Ref (If_Result_17);
   If_Result_18 := If_Result_17;
end if;
      Inc_Ref (If_Result_18);
   If_Result_19 := If_Result_18;
end if;
      Inc_Ref (If_Result_19);
--# expr-done 1099
Let_Result_1 := If_Result_19; Inc_Ref (Let_Result_1);
--# end
--# expr-done 1100
      Property_Result := Let_Result_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_755;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_755;
               Finalizer_Scope_1866;
               Finalizer_Scope_1868;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aspect_Assoc_P_Xref_Equation;
--# end
--# property-start AspectAssoc.stable_properties_assoc_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9507
pragma Warnings (Off, "is not referenced");
 function Aspect_Assoc_P_Stable_Properties_Assoc_Equation
  (Node : Bare_Aspect_Assoc
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Aspect_Assoc := Bare_Aspect_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Aspect_Assoc :=
     Internal_Entity_Aspect_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_756 with Inline_Always;
            procedure Finalizer_Scope_1869 with Inline_Always;
            procedure Finalizer_Scope_1871 with Inline_Always;
            procedure Finalizer_Scope_1872 with Inline_Always;
            procedure Finalizer_Scope_1877 with Inline_Always;
            procedure Finalizer_Scope_1879 with Inline_Always;
      Block_Var : Internal_Entity_Identifier_Array_Access;
Match_0 : Internal_Entity_Aggregate;
Match_1 : Internal_Entity_Paren_Expr;
Match_2 : Internal_Entity_Expr;
Var_Expr : Internal_Entity_Un_Op;
Fld : Internal_Entity_Expr;
Fld_1 : Internal_Entity_Assoc_List;
Fld_2 : Bare_Assoc_List;
Fld_3 : Internal_Entity_Info;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_4 : Internal_Entity_Expr;
Item_47 : Internal_Entity_Basic_Assoc;
Bare_Item_47 : Bare_Basic_Assoc;
Map_Result : Internal_Entity_Expr_Array_Access;
Fld_5 : Internal_Entity_Expr;
Singleton : Internal_Entity_Expr_Array_Access;
Singleton_1 : Internal_Entity_Expr_Array_Access;
Match_Prefix : Internal_Entity_Expr;
Cast_Result : Internal_Entity_Aggregate;
Let_Result : Internal_Entity_Expr_Array_Access;
Scope_Result : Internal_Entity_Expr_Array_Access;
Cast_Result_1 : Internal_Entity_Paren_Expr;
Let_Result_1 : Internal_Entity_Expr_Array_Access;
Scope_Result_1 : Internal_Entity_Expr_Array_Access;
Cast_Result_2 : Internal_Entity_Expr;
Let_Result_2 : Internal_Entity_Expr_Array_Access;
Scope_Result_2 : Internal_Entity_Expr_Array_Access;
Match_Result : Internal_Entity_Expr_Array_Access;
Cast_Expr : Internal_Entity_Expr;
Cast_Result_3 : Internal_Entity_Un_Op;
Fld_6 : Internal_Entity_Expr;
Fld_7 : Internal_Entity_Op;
Is_A : Boolean;
If_Result_1 : Internal_Entity_Expr;
Scope_Result_3 : Internal_Entity_Expr;
Result_Var : Internal_Entity_Expr;
Map_Result_1 : Internal_Entity_Expr_Array_Access;
Cast_Expr_1 : Internal_Entity_Expr;
Cast_Result_4 : Internal_Entity_Identifier;
Map_Result_2 : Internal_Entity_Identifier_Array_Access;
Fld_8 : Symbol_Type;
Fld_9 : Bare_Ada_Node;
Fld_10 : Internal_Entity_Array_Access;
Is_A_1 : Boolean;
Cast_Expr_2 : Internal_Entity;
Cast_Result_5 : Internal_Entity_Basic_Decl;
Fld_11 : Internal_Entity_Base_Subp_Spec;
Fld_12 : Boolean;
If_Result_2 : Boolean;
Map_Result_3 : Internal_Entity_Array_Access;
Fld_13 : Logic_Var;
Bind_Result : Logic_Equation;
Map_Result_4 : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Map_Result_5 : Logic_Equation_Array_Access;
Logic_Boolean_Op_1 : Logic_Equation;
Let_Result_3 : Logic_Equation;
            procedure Finalizer_Scope_756 is
            begin
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Let_Result_2);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Match_Result);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Map_Result_2);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Map_Result_5);
                     Dec_Ref (Logic_Boolean_Op_1);
                     Dec_Ref (Let_Result_3);
            end Finalizer_Scope_756;
            procedure Finalizer_Scope_1869 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_1869;
            procedure Finalizer_Scope_1871 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1871;
            procedure Finalizer_Scope_1872 is
            begin
                     Dec_Ref (Singleton_1);
            end Finalizer_Scope_1872;
            procedure Finalizer_Scope_1877 is
            begin
                     Dec_Ref (Fld_10);
                     Dec_Ref (Map_Result_3);
                     Dec_Ref (Map_Result_4);
                     Dec_Ref (Logic_Boolean_Op);
            end Finalizer_Scope_1877;
            procedure Finalizer_Scope_1879 is
            begin
                     Dec_Ref (Bind_Result);
            end Finalizer_Scope_1879;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1129 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 1114 '<Map at ast.py:9514>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9514
   --# expr-start 1112 '<Map at ast.py:9514>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9514
   --# expr-start 1106 '<Match at ast.py:9514>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9514
--# expr-start 1101 '<FieldAccess for expr at ast.py:9514>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9514
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Aspect_Assoc_F_Expr, Info => Ent.Info);
--# expr-done 1101
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld; 
case Ada_Expr (Match_Prefix.Node.Kind) is
         when Ada_Aggregate .. Ada_Bracket_Aggregate =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Aggregate
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind a Match_0
--# expr-start 1104 '<Map at ast.py:9516>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9516
   --# expr-start 1102 '<FieldAccess for assocs at ast.py:9516>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9516
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Assoc_List (Node => Match_0.Node.Base_Aggregate_F_Assocs, Info => Match_0.Info);
--# expr-done 1102
Fld_2 := Fld_1.Node;
      declare
         Map_Result_Vec : Internal_Entity_Expr_Vectors.Vector;
      begin
      if Fld_2 = null then
      Map_Result := Create_Internal_Entity_Expr_Array (0);
      else
      declare
         Collection : constant Bare_Assoc_List := Fld_2;
      begin
         for Untyped_Item_47 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_47 := Untyped_Item_47; 
Is_Null := Bare_Item_47 = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_3 := Fld_1.Info;
As_Entity := (Info => Fld_3, Node => Bare_Item_47); 
   If_Result := As_Entity;
end if;
                  Item_47 := If_Result; 
   --# scope-start
               --# bind i Item_47
      --# expr-start 1103 '<FieldAccess for expr at ast.py:9516>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9516
      if Item_47.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_4 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Item_47.Node, E_Info => Item_47.Info);
--# end
--# expr-done 1103
         declare
            Item_To_Append : constant Internal_Entity_Expr := Fld_4;
         begin
            Internal_Entity_Expr_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Expr_Array
           (Items_Count => Natural (Internal_Entity_Expr_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Expr_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Expr_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Expr_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1104
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
Scope_Result := Let_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1869;
            Match_Result := Scope_Result; Inc_Ref (Match_Result);
         when Ada_Paren_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Paren_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind pe Match_1
--# expr-start 1105 '<FieldAccess for expr at ast.py:9518>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9518
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Expr (Node => Match_1.Node.Paren_Expr_F_Expr, Info => Match_1.Info);
--# expr-done 1105
                Singleton := Create_Internal_Entity_Expr_Array (Items_Count => 1);
                Singleton.Items (1) := Fld_5;
Let_Result_1 := Singleton; Inc_Ref (Let_Result_1);
--# end
Scope_Result_1 := Let_Result_1;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_1871;
            Match_Result := Scope_Result_1; Inc_Ref (Match_Result);
         when Ada_Abstract_State_Decl_Expr .. Ada_Allocator | Ada_Delta_Aggregate .. Ada_Target_Name | Ada_Quantified_Expr .. Ada_Un_Op =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind _ Match_2
                Singleton_1 := Create_Internal_Entity_Expr_Array (Items_Count => 1);
                Singleton_1.Items (1) := No_Entity_Expr;
Let_Result_2 := Singleton_1; Inc_Ref (Let_Result_2);
--# end
Scope_Result_2 := Let_Result_2;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_1872;
            Match_Result := Scope_Result_2; Inc_Ref (Match_Result);
end case;
--# expr-done 1106
      declare
         Map_Result_1_Vec : Internal_Entity_Expr_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Expr_Array_Access := Match_Result;
      begin
         for Item_46 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_46
--# expr-start 1107 '<Cast to UnOp at ast.py:9523>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9523
Cast_Expr := Item_46; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Un_Op_Range
   then
      Cast_Result_3 := Create_Internal_Entity_Un_Op
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result_3 := No_Entity_Un_Op;
   end if;
--# expr-done 1107
Var_Expr := Cast_Result_3;
if Var_Expr /= No_Entity_Un_Op then
   --# scope-start
   --# bind uo Var_Expr
--# expr-start 1111 '<If at ast.py:9524>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9524
--# expr-start 1109 '<IsA Op.Not at ast.py:9524>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9524
--# expr-start 1108 '<FieldAccess for op at ast.py:9524>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9524
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Op (Node => Var_Expr.Node.Un_Op_F_Op, Info => Var_Expr.Info);
--# expr-done 1108
Is_A := Fld_7.Node /= null 
and then Fld_7.Node.Kind in Ada_Op_Not_Range; 
--# expr-done 1109
if Is_A then
   --# expr-start 1110 '<FieldAccess for expr at ast.py:9524>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9524
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Expr (Node => Var_Expr.Node.Un_Op_F_Expr, Info => Var_Expr.Info);
--# expr-done 1110
   If_Result_1 := Fld_6;
else
   If_Result_1 := Item_46;
end if;
--# expr-done 1111
Scope_Result_3 := If_Result_1;
   --# end
   Result_Var := Scope_Result_3;
else
   Result_Var := Item_46;
end if;
         declare
            Item_To_Append : constant Internal_Entity_Expr := Result_Var;
         begin
            Internal_Entity_Expr_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result_1 := Create_Internal_Entity_Expr_Array
           (Items_Count => Natural (Internal_Entity_Expr_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Entity_Expr_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Entity_Expr_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Entity_Expr_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 1112
      declare
         Map_Result_2_Vec : Internal_Entity_Identifier_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Expr_Array_Access := Map_Result_1;
      begin
         for Item_45 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_45
      --# expr-start 1113 '<Cast to Identifier at ast.py:9529>' Cast_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9529
Cast_Expr_1 := Item_45; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Identifier_Range
   then
      Cast_Result_4 := Create_Internal_Entity_Identifier
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 1113
         declare
            Item_To_Append : constant Internal_Entity_Identifier := Cast_Result_4;
         begin
            Internal_Entity_Identifier_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result_2 := Create_Internal_Entity_Identifier_Array
           (Items_Count => Natural (Internal_Entity_Identifier_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Internal_Entity_Identifier_Vectors.Get
              (Map_Result_2_Vec,
               I + Internal_Entity_Identifier_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Internal_Entity_Identifier_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 1114
Block_Var := Map_Result_2; Inc_Ref (Block_Var);
--# bind identifiers Block_Var
--# expr-start 1128 '<LogicAll at ast.py:9533>' Logic_Boolean_Op_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9533
--# expr-start 1127 '<Map at ast.py:9533>' Map_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9533
      declare
         Map_Result_5_Vec : Logic_Equation_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Identifier_Array_Access := Block_Var;
      begin
         for Item_48 of
               Collection.Items
         loop
   --# scope-start
               --# bind i Item_48
      --# expr-start 1126 '<LogicAny at ast.py:9534>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9534
--# expr-start 1125 '<Map at ast.py:9534>' Map_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9534
   --# expr-start 1122 '<Filter at ast.py:9534>' Map_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9534
   --# expr-start 1117 '<FieldAccess for env_get at ast.py:9534>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9534
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1115 '<FieldAccess for sym at ast.py:9536>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9536
      if Item_48.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTokNode.sym
Fld_8 := Libadalang.Implementation.Single_Tok_Node_P_Sym (Node => Item_48.Node);
--# end
--# expr-done 1115
--# expr-start 1116 '<FieldAccess for origin_node at ast.py:9538>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9538
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_9 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 1116
--# property-call-start AdaNode.env_get
Fld_10 := Libadalang.Implementation.Ada_Node_P_Env_Get (Node => Self, Env => Env, Symbol => Fld_8, Lookup => Recursive, From_Node => Fld_9, Categories => (Inherited_Primitives => True, Nocat => True));
--# end
--# expr-done 1117
      declare
         Map_Result_3_Vec : Internal_Entity_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Fld_10;
      begin
         for Item_50 of
               Collection.Items
         loop
   --# scope-start
               --# bind f Item_50
--# expr-start 1118 '<IsA SubpDecl, ExprFunction at ast.py:9542>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9542
Is_A_1 := Item_50.Node /= null 
and then Item_50.Node.Kind in Ada_Subp_Decl_Range | Ada_Expr_Function_Range; 
--# expr-done 1118
if Is_A_1 then
   --# expr-start 1121 '<FieldAccess for denotes_the_property_function at ast.py:9543>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9543
      if Item_48.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1120 '<FieldAccess for subp_spec_or_null at ast.py:9544>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9544
--# expr-start 1119 '<Cast to BasicDecl at ast.py:9544>' Cast_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9544
Cast_Expr_2 := Item_50; 
   if Cast_Expr_2.Node = null
      or else Cast_Expr_2.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_5 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
   else
         Cast_Result_5 := No_Entity_Basic_Decl;
   end if;
--# expr-done 1119
      if Cast_Result_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_11 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Cast_Result_5.Node, E_Info => Cast_Result_5.Info);
--# end
--# expr-done 1120
--# property-call-start BaseId.denotes_the_property_function
Fld_12 := Libadalang.Implementation.Base_Id_P_Denotes_The_Property_Function (Node => Item_48.Node, Subp_Spec => Fld_11, E_Info => Item_48.Info);
--# end
--# expr-done 1121
   If_Result_2 := Fld_12;
else
   If_Result_2 := False;
end if;
         if If_Result_2 then
         declare
            Item_To_Append : constant Internal_Entity := Item_50;
         begin
            Internal_Entity_Vectors.Append (Map_Result_3_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result_3 := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_3_Vec)));
         for I in Map_Result_3.Items'Range loop
            Map_Result_3.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_3_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result_3.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_3_Vec);
      end;
--# expr-done 1122
      declare
         Map_Result_4_Vec : Logic_Equation_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Map_Result_3;
      begin
         for Item_49 of
               Collection.Items
         loop
   --# scope-start
               --# bind f Item_49
      --# expr-start 1124 '<Bind at ast.py:9547>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9547
--# expr-start 1123 '<FieldAccess for ref_var at ast.py:9547>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9547
      if Item_48.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_13 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Item_48.Node);
--# end
--# expr-done 1123
Fld_13.Value := No_Entity;
Entity_Vars.Reset (Fld_13);
Bind_Result := Solver.Create_Assign (Fld_13, Item_49, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9547") else null)); 
--# expr-done 1124
         declare
            Item_To_Append : constant Logic_Equation := Bind_Result;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_4_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_1879;
         end loop;
      end;
         Map_Result_4 := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_4_Vec)));
         for I in Map_Result_4.Items'Range loop
            Map_Result_4.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_4_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result_4.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_4_Vec);
      end;
--# expr-done 1125
Logic_Boolean_Op := Solver.Create_Any (Relation_Array (Map_Result_4.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9534") else null)); 
--# expr-done 1126
         declare
            Item_To_Append : constant Logic_Equation := Logic_Boolean_Op;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_5_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_1877;
         end loop;
      end;
         Map_Result_5 := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_5_Vec)));
         for I in Map_Result_5.Items'Range loop
            Map_Result_5.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_5_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result_5.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_5_Vec);
      end;
--# expr-done 1127
Logic_Boolean_Op_1 := Solver.Create_All (Relation_Array (Map_Result_5.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9533") else null)); 
--# expr-done 1128
Let_Result_3 := Logic_Boolean_Op_1; Inc_Ref (Let_Result_3);
--# end
--# expr-done 1129
      Property_Result := Let_Result_3;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_756;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_756;
               Finalizer_Scope_1869;
               Finalizer_Scope_1871;
               Finalizer_Scope_1872;
               Finalizer_Scope_1877;
               Finalizer_Scope_1879;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aspect_Assoc_P_Stable_Properties_Assoc_Equation;
--# end
--# property-start AspectAssoc.aspect_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9552
pragma Warnings (Off, "is not referenced");
 function Aspect_Assoc_P_Aspect_Name
  (Node : Bare_Aspect_Assoc
      ; N : Internal_Entity_Name
  )
   return String_Type
is
   Self : Bare_Aspect_Assoc := Bare_Aspect_Assoc (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind n N
   Property_Result : String_Type;
            procedure Finalizer_Scope_757 with Inline_Always;
            procedure Finalizer_Scope_1880 with Inline_Always;
            procedure Finalizer_Scope_1881 with Inline_Always;
      Match_0 : Internal_Entity_Base_Id;
Match_1 : Internal_Entity_Attribute_Ref;
Match_2 : Internal_Entity_Name;
Fld : Symbol_Type;
Fld_1 : String_Type;
Fld_2 : Internal_Entity_Name;
Fld_3 : String_Type;
Str : String_Type;
Concat_Result : String_Type;
Fld_4 : Internal_Entity_Identifier;
Fld_5 : Symbol_Type;
Fld_6 : String_Type;
Concat_Result_1 : String_Type;
Match_Prefix : Internal_Entity_Name;
Cast_Result : Internal_Entity_Base_Id;
Let_Result : String_Type;
Scope_Result : String_Type;
Cast_Result_1 : Internal_Entity_Attribute_Ref;
Let_Result_1 : String_Type;
Scope_Result_1 : String_Type;
Cast_Result_2 : Internal_Entity_Name;
Let_Result_2 : String_Type;
Scope_Result_2 : String_Type;
Match_Result : String_Type;
Let_Result_3 : String_Type;
            procedure Finalizer_Scope_757 is
            begin
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Let_Result_2);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Match_Result);
                     Dec_Ref (Let_Result_3);
            end Finalizer_Scope_757;
            procedure Finalizer_Scope_1880 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1880;
            procedure Finalizer_Scope_1881 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Str);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Fld_6);
                     Dec_Ref (Concat_Result_1);
            end Finalizer_Scope_1881;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1141 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 1140 '<Match at ast.py:9559>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9559
      if N.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := N; 
case Ada_Name (Match_Prefix.Node.Kind) is
         when Ada_Char_Literal .. Ada_String_Literal =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Base_Id
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind bid Match_0
--# expr-start 1131 '<FieldAccess for image at ast.py:9560>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9560
--# expr-start 1130 '<FieldAccess for sym at ast.py:9560>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9560
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTokNode.sym
Fld := Libadalang.Implementation.Single_Tok_Node_P_Sym (Node => Match_0.Node);
--# end
--# expr-done 1130
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Image (Self => Fld);
--# expr-done 1131
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
Scope_Result := Let_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1880;
            Match_Result := Scope_Result; Inc_Ref (Match_Result);
         when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Attribute_Ref
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind ar Match_1
--# expr-start 1139 '<Concat at ast.py:9561>' Concat_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9561
--# expr-start 1135 '<Concat at ast.py:9561>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9561
--# expr-start 1133 '<FieldAccess for aspect_name at ast.py:9561>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9561
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1132 '<FieldAccess for prefix at ast.py:9561>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9561
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Match_1.Node.Attribute_Ref_F_Prefix, Info => Match_1.Info);
--# expr-done 1132
--# property-call-start AspectAssoc.aspect_name
Fld_3 := Libadalang.Implementation.Aspect_Assoc_P_Aspect_Name (Node => Self, N => Fld_2);
--# end
--# expr-done 1133
--# expr-start 1134 '<String at ast.py:9562>' Str /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9562
Str := Create_String ("'"); 
--# expr-done 1134
Concat_Result := Concat_String (Fld_3, Str); 
--# expr-done 1135
--# expr-start 1138 '<FieldAccess for image at ast.py:9563>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9563
--# expr-start 1137 '<FieldAccess for sym at ast.py:9563>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9563
--# expr-start 1136 '<FieldAccess for attribute at ast.py:9563>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9563
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Identifier (Node => Match_1.Node.Attribute_Ref_F_Attribute, Info => Match_1.Info);
--# expr-done 1136
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTokNode.sym
Fld_5 := Libadalang.Implementation.Single_Tok_Node_P_Sym (Node => Fld_4.Node);
--# end
--# expr-done 1137
      if Fld_5 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Image (Self => Fld_5);
--# expr-done 1138
Concat_Result_1 := Concat_String (Concat_Result, Fld_6); 
--# expr-done 1139
Let_Result_1 := Concat_Result_1; Inc_Ref (Let_Result_1);
--# end
Scope_Result_1 := Let_Result_1;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_1881;
            Match_Result := Scope_Result_1; Inc_Ref (Match_Result);
         when Ada_Call_Expr .. Ada_Reduce_Attribute_Ref | Ada_Null_Literal .. Ada_Target_Name =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind _ Match_2
Let_Result_2 := raise Property_Error with "aspect_name called on an invalid aspect name"; Inc_Ref (Let_Result_2);
--# end
Scope_Result_2 := Let_Result_2;
   Inc_Ref (Scope_Result_2);
   --# end
            Match_Result := Scope_Result_2; Inc_Ref (Match_Result);
end case;
--# expr-done 1140
Let_Result_3 := Match_Result; Inc_Ref (Let_Result_3);
--# end
--# expr-done 1141
      Property_Result := Let_Result_3;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_757;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_757;
               Finalizer_Scope_1880;
               Finalizer_Scope_1881;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aspect_Assoc_P_Aspect_Name;
--# end
--# property-start AspectAssoc.is_ghost_code /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9571
pragma Warnings (Off, "is not referenced");
 function Aspect_Assoc_P_Is_Ghost_Code
  (Node : Bare_Aspect_Assoc
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Aspect_Assoc := Bare_Aspect_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Aspect_Assoc :=
     Internal_Entity_Aspect_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Internal_Entity_Name;
Fld_1 : Symbol_Type;
Sym : Symbol_Type;
Is_Equal : Boolean;
Fld_2 : Internal_Entity_Name;
Fld_3 : Symbol_Type;
Sym_1 : Symbol_Type;
Is_Equal_1 : Boolean;
If_Result : Boolean;
Fld_4 : Internal_Entity_Name;
Fld_5 : Symbol_Type;
Sym_2 : Symbol_Type;
Is_Equal_2 : Boolean;
If_Result_1 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1154 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1145 '<Eq at ast.py:9575>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9575
--# expr-start 1143 '<FieldAccess for name_symbol at ast.py:9575>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9575
--# expr-start 1142 '<FieldAccess for id at ast.py:9575>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9575
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 1142
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld.Node);
--# end
--# expr-done 1143
--# expr-start 1144 '<Symbol Pre at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Pre); 
--# expr-done 1144
Is_Equal := Fld_1 = Sym; 
--# expr-done 1145
if Is_Equal then
   If_Result := True;
else
   --# expr-start 1149 '<Eq at ast.py:9575>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9575
--# expr-start 1147 '<FieldAccess for name_symbol at ast.py:9575>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9575
--# expr-start 1146 '<FieldAccess for id at ast.py:9575>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9575
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 1146
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_2.Node);
--# end
--# expr-done 1147
--# expr-start 1148 '<Symbol Post at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Post); 
--# expr-done 1148
Is_Equal_1 := Fld_3 = Sym_1; 
--# expr-done 1149
   If_Result := Is_Equal_1;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 1153 '<Eq at ast.py:9575>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9575
--# expr-start 1151 '<FieldAccess for name_symbol at ast.py:9575>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9575
--# expr-start 1150 '<FieldAccess for id at ast.py:9575>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9575
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Name (Node => Ent.Node.Aspect_Assoc_F_Id, Info => Ent.Info);
--# expr-done 1150
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_5 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_4.Node);
--# end
--# expr-done 1151
--# expr-start 1152 '<Symbol Contract_Cases at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Contract_Cases); 
--# expr-done 1152
Is_Equal_2 := Fld_5 = Sym_2; 
--# expr-done 1153
   If_Result_1 := Is_Equal_2;
end if;
Let_Result := If_Result_1; 
--# end
--# expr-done 1154
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aspect_Assoc_P_Is_Ghost_Code;
--# end
   --
   --  Primitives for Bare_Aspect_Clause
   --
--# property-start AspectClause.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9227
pragma Warnings (Off, "is not referenced");
 function Aspect_Clause_P_Xref_Entry_Point
  (Node : Bare_Aspect_Clause
  )
   return Boolean
is
   Self : Bare_Aspect_Clause := Bare_Aspect_Clause (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aspect_Clause_P_Xref_Entry_Point;
--# end
   --
   --  Primitives for Bare_At_Clause
   --
      procedure Initialize_Fields_For_At_Clause
        (Self : Bare_At_Clause
         ; At_Clause_F_Name : Bare_Base_Id
         ; At_Clause_F_Expr : Bare_Expr
        ) is
      begin
            Self.At_Clause_F_Name := At_Clause_F_Name;
            Self.At_Clause_F_Expr := At_Clause_F_Expr;
      end Initialize_Fields_For_At_Clause;
   function At_Clause_F_Name
     (Node : Bare_At_Clause) return Bare_Base_Id
   is
   begin
         return Node.At_Clause_F_Name;
   end;
   function At_Clause_F_Expr
     (Node : Bare_At_Clause) return Bare_Expr
   is
   begin
         return Node.At_Clause_F_Expr;
   end;
--# property-start AtClause.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9360
pragma Warnings (Off, "is not referenced");
 function At_Clause_P_Xref_Equation
  (Node : Bare_At_Clause
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_At_Clause := Bare_At_Clause (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_At_Clause :=
     Internal_Entity_At_Clause'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_743 with Inline_Always;
      Fld : Internal_Entity_Base_Id;
Fld_1 : Logic_Equation;
Fld_2 : Internal_Entity_Expr;
Fld_3 : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_743 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Fld_3);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_743;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1160 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1159 '<And at ast.py:9361>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9361
--# expr-start 1156 '<FieldAccess for sub_equation at ast.py:9361>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9361
--# expr-start 1155 '<FieldAccess for name at ast.py:9361>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9361
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Base_Id (Node => Ent.Node.At_Clause_F_Name, Info => Ent.Info);
--# expr-done 1155
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 1156
--# expr-start 1158 '<FieldAccess for sub_equation at ast.py:9361>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9361
--# expr-start 1157 '<FieldAccess for expr at ast.py:9361>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9361
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Expr (Node => Ent.Node.At_Clause_F_Expr, Info => Ent.Info);
--# expr-done 1157
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_2.Node, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 1158
And_Pred := Create_And (Fld_1, Fld_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9361") else null)); 
--# expr-done 1159
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 1160
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_743;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_743;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end At_Clause_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Attribute_Def_Clause
   --
      procedure Initialize_Fields_For_Attribute_Def_Clause
        (Self : Bare_Attribute_Def_Clause
         ; Attribute_Def_Clause_F_Attribute_Expr : Bare_Name
         ; Attribute_Def_Clause_F_Expr : Bare_Expr
        ) is
      begin
            Self.Attribute_Def_Clause_F_Attribute_Expr := Attribute_Def_Clause_F_Attribute_Expr;
            Self.Attribute_Def_Clause_F_Expr := Attribute_Def_Clause_F_Expr;
      end Initialize_Fields_For_Attribute_Def_Clause;
   function Attribute_Def_Clause_F_Attribute_Expr
     (Node : Bare_Attribute_Def_Clause) return Bare_Name
   is
   begin
         return Node.Attribute_Def_Clause_F_Attribute_Expr;
   end;
   function Attribute_Def_Clause_F_Expr
     (Node : Bare_Attribute_Def_Clause) return Bare_Expr
   is
   begin
         return Node.Attribute_Def_Clause_F_Expr;
   end;
--# property-start AttributeDefClause.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9272
pragma Warnings (Off, "is not referenced");
 function Attribute_Def_Clause_P_Xref_Equation
  (Node : Bare_Attribute_Def_Clause
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Def_Clause := Bare_Attribute_Def_Clause (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Def_Clause :=
     Internal_Entity_Attribute_Def_Clause'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_739 with Inline_Always;
            procedure Finalizer_Scope_1883 with Inline_Always;
      Block_Var : Internal_Entity_Attribute_Ref;
Block_Var_1 : Symbol_Type;
Var_Expr : Internal_Entity_Attribute_Ref;
Fld : Internal_Entity_Name;
Cast_Result : Internal_Entity_Attribute_Ref;
Fld_1 : Internal_Entity_Identifier;
Fld_2 : Symbol_Type;
Sym : Symbol_Type;
Is_Equal : Boolean;
Sym_1 : Symbol_Type;
Is_Equal_1 : Boolean;
If_Result : Boolean;
Sym_2 : Symbol_Type;
Is_Equal_2 : Boolean;
If_Result_1 : Boolean;
Sym_3 : Symbol_Type;
Is_Equal_3 : Boolean;
If_Result_2 : Boolean;
Fld_3 : Internal_Entity_Expr;
Cast_Result_1 : Internal_Entity_Name;
Fld_4 : Logic_Equation;
Fld_5 : Internal_Entity_Expr;
Cast_Result_2 : Internal_Entity_Name;
Fld_6 : Logic_Var;
Fld_7 : Internal_Entity_Name;
Fld_8 : Internal_Entity_Base_Type_Decl;
Sym_4 : Symbol_Type;
Is_Equal_4 : Boolean;
Pred : Logic_Equation;
And_Pred : Logic_Equation;
Sym_5 : Symbol_Type;
Is_Equal_5 : Boolean;
Fld_9 : Internal_Entity_Expr;
Cast_Result_3 : Internal_Entity_Name;
Fld_10 : Logic_Equation;
Fld_11 : Internal_Entity_Expr;
Cast_Result_4 : Internal_Entity_Name;
Fld_12 : Logic_Var;
Fld_13 : Internal_Entity_Name;
Fld_14 : Internal_Entity_Base_Type_Decl;
Pred_1 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_15 : Internal_Entity_Expr;
Fld_16 : Logic_Equation;
Sym_6 : Symbol_Type;
Is_Equal_6 : Boolean;
Fld_17 : Bare_Expr;
Fld_18 : Logic_Var;
Sym_7 : Symbol_Type;
Fld_19 : Internal_Entity;
Bind_Result : Logic_Equation;
True_Rel : Logic_Equation;
If_Result_3 : Logic_Equation;
And_Pred_2 : Logic_Equation;
If_Result_4 : Logic_Equation;
If_Result_5 : Logic_Equation;
Fld_20 : Internal_Entity_Name;
Fld_21 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel_1 : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred_3 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_739 is
            begin
                     Dec_Ref (Fld_4);
                     Dec_Ref (Pred);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_10);
                     Dec_Ref (Pred_1);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_16);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (If_Result_3);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (If_Result_4);
                     Dec_Ref (If_Result_5);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel_1);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_739;
            procedure Finalizer_Scope_1883 is
            begin
                     Dec_Ref (Fld_21);
            end Finalizer_Scope_1883;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1212 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1162 '<Cast to AttributeRef at ast.py:9273>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9273
--# expr-start 1161 '<FieldAccess for attribute_expr at ast.py:9273>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9273
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Def_Clause_F_Attribute_Expr, Info => Ent.Info);
--# expr-done 1161
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Attribute_Ref_Range
   then
      Cast_Result := Create_Internal_Entity_Attribute_Ref
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 1162
Block_Var := Cast_Result; 
--# bind attr Block_Var
--# expr-start 1164 '<FieldAccess for name_symbol at ast.py:9274>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9274
--# expr-start 1163 '<FieldAccess for attribute at ast.py:9274>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9274
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Identifier (Node => Block_Var.Node.Attribute_Ref_F_Attribute, Info => Block_Var.Info);
--# expr-done 1163
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_1.Node);
--# end
--# expr-done 1164
Block_Var_1 := Fld_2; 
--# bind rel_name Block_Var_1
--# expr-start 1211 '<And at ast.py:9276>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9276
--# expr-start 1208 '<Cond at ast.py:9276>' If_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9276
--# expr-start 1166 '<Eq at ast.py:9277>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9277
--# expr-start 1165 '<Symbol Read at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Read); 
--# expr-done 1165
Is_Equal := Block_Var_1 = Sym; 
--# expr-done 1166
if Is_Equal then
   If_Result := True;
else
   --# expr-start 1168 '<Eq at ast.py:9277>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9277
--# expr-start 1167 '<Symbol Write at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Write); 
--# expr-done 1167
Is_Equal_1 := Block_Var_1 = Sym_1; 
--# expr-done 1168
   If_Result := Is_Equal_1;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 1170 '<Eq at ast.py:9277>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9277
--# expr-start 1169 '<Symbol Input at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Input); 
--# expr-done 1169
Is_Equal_2 := Block_Var_1 = Sym_2; 
--# expr-done 1170
   If_Result_1 := Is_Equal_2;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
   --# expr-start 1172 '<Eq at ast.py:9277>' Is_Equal_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9277
--# expr-start 1171 '<Symbol Output at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Output); 
--# expr-done 1171
Is_Equal_3 := Block_Var_1 = Sym_3; 
--# expr-done 1172
   If_Result_2 := Is_Equal_3;
end if;
if If_Result_2 then
   --# expr-start 1184 '<And at ast.py:9278>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9278
--# expr-start 1175 '<FieldAccess for xref_no_overloading at ast.py:9278>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9278
--# expr-start 1174 '<Cast to Name at ast.py:9278>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9278
--# expr-start 1173 '<FieldAccess for expr at ast.py:9278>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9278
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Expr (Node => Ent.Node.Attribute_Def_Clause_F_Expr, Info => Ent.Info);
--# expr-done 1173
   if Fld_3.Node = null
      or else Fld_3.Node.Kind in Ada_Name
   then
      Cast_Result_1 := Create_Internal_Entity_Name
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 1174
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_4 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Cast_Result_1.Node, Env => Env, Origin => Origin, E_Info => Cast_Result_1.Info);
--# end
--# expr-done 1175
--# expr-start 1183 '<Predicate on BasicDecl.is_stream_subprogram_for_type at ast.py:9279>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9279
--# expr-start 1178 '<FieldAccess for ref_var at ast.py:9281>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9281
--# expr-start 1177 '<Cast to Name at ast.py:9281>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9281
--# expr-start 1176 '<FieldAccess for expr at ast.py:9281>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9281
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Expr (Node => Ent.Node.Attribute_Def_Clause_F_Expr, Info => Ent.Info);
--# expr-done 1176
   if Fld_5.Node = null
      or else Fld_5.Node.Kind in Ada_Name
   then
      Cast_Result_2 := Create_Internal_Entity_Name
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
   else
         Cast_Result_2 := No_Entity_Name;
   end if;
--# expr-done 1177
      if Cast_Result_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_6 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Cast_Result_2.Node);
--# end
--# expr-done 1178
Fld_6.Value := No_Entity;
Entity_Vars.Reset (Fld_6);
--# expr-start 1180 '<FieldAccess for name_designated_type at ast.py:9282>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9282
--# expr-start 1179 '<FieldAccess for prefix at ast.py:9282>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9282
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Name (Node => Block_Var.Node.Attribute_Ref_F_Prefix, Info => Block_Var.Info);
--# expr-done 1179
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_8 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld_7.Node, E_Info => Fld_7.Info);
--# end
--# expr-done 1180
--# expr-start 1182 '<Eq at ast.py:9283>' Is_Equal_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9283
--# expr-start 1181 '<Symbol Input at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Input); 
--# expr-done 1181
Is_Equal_4 := Block_Var_1 = Sym_4; 
--# expr-done 1182
Pred := Solver.Create_Predicate (Fld_6, Create_Basic_Decl_P_Is_Stream_Subprogram_For_Type_0_Predicate (Fld_8, Is_Equal_4)); 
--# expr-done 1183
And_Pred := Create_And (Fld_4, Pred, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9278") else null)); 
--# expr-done 1184
   If_Result_5 := And_Pred;
else
--# expr-start 1186 '<Eq at ast.py:9286>' Is_Equal_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9286
--# expr-start 1185 '<Symbol Put_Image at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Put_Image); 
--# expr-done 1185
Is_Equal_5 := Block_Var_1 = Sym_5; 
--# expr-done 1186
if Is_Equal_5 then
   --# expr-start 1196 '<And at ast.py:9287>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9287
--# expr-start 1189 '<FieldAccess for xref_no_overloading at ast.py:9287>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9287
--# expr-start 1188 '<Cast to Name at ast.py:9287>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9287
--# expr-start 1187 '<FieldAccess for expr at ast.py:9287>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9287
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Expr (Node => Ent.Node.Attribute_Def_Clause_F_Expr, Info => Ent.Info);
--# expr-done 1187
   if Fld_9.Node = null
      or else Fld_9.Node.Kind in Ada_Name
   then
      Cast_Result_3 := Create_Internal_Entity_Name
        (Node => Fld_9.Node,
         Info => Fld_9.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 1188
      if Cast_Result_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_10 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Cast_Result_3.Node, All_Els => True, Env => Env, Origin => Origin, E_Info => Cast_Result_3.Info);
--# end
--# expr-done 1189
--# expr-start 1195 '<Predicate on BasicDecl.is_put_image_subprogram_for_type at ast.py:9288>' Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9288
--# expr-start 1192 '<FieldAccess for ref_var at ast.py:9290>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9290
--# expr-start 1191 '<Cast to Name at ast.py:9290>' Cast_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9290
--# expr-start 1190 '<FieldAccess for expr at ast.py:9290>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9290
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Create_Internal_Entity_Expr (Node => Ent.Node.Attribute_Def_Clause_F_Expr, Info => Ent.Info);
--# expr-done 1190
   if Fld_11.Node = null
      or else Fld_11.Node.Kind in Ada_Name
   then
      Cast_Result_4 := Create_Internal_Entity_Name
        (Node => Fld_11.Node,
         Info => Fld_11.Info);
   else
         Cast_Result_4 := No_Entity_Name;
   end if;
--# expr-done 1191
      if Cast_Result_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_12 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Cast_Result_4.Node);
--# end
--# expr-done 1192
Fld_12.Value := No_Entity;
Entity_Vars.Reset (Fld_12);
--# expr-start 1194 '<FieldAccess for name_designated_type at ast.py:9291>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9291
--# expr-start 1193 '<FieldAccess for prefix at ast.py:9291>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9291
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Create_Internal_Entity_Name (Node => Block_Var.Node.Attribute_Ref_F_Prefix, Info => Block_Var.Info);
--# expr-done 1193
      if Fld_13.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_14 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld_13.Node, E_Info => Fld_13.Info);
--# end
--# expr-done 1194
Pred_1 := Solver.Create_Predicate (Fld_12, Create_Basic_Decl_P_Is_Put_Image_Subprogram_For_Type_0_Predicate (Fld_14)); 
--# expr-done 1195
And_Pred_1 := Create_And (Fld_10, Pred_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9287") else null)); 
--# expr-done 1196
   If_Result_4 := And_Pred_1;
else
   --# expr-start 1207 '<And at ast.py:9294>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9294
--# expr-start 1198 '<FieldAccess for sub_equation at ast.py:9294>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9294
--# expr-start 1197 '<FieldAccess for expr at ast.py:9294>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9294
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Create_Internal_Entity_Expr (Node => Ent.Node.Attribute_Def_Clause_F_Expr, Info => Ent.Info);
--# expr-done 1197
      if Fld_15.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_16 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_15.Node, Env => Env, Origin => Origin, E_Info => Fld_15.Info);
--# end
--# expr-done 1198
--# expr-start 1206 '<Cond at ast.py:9294>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9294
--# expr-start 1200 '<Eq at ast.py:9295>' Is_Equal_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9295
--# expr-start 1199 '<Symbol External_Tag at ???>' Sym_6 None
Sym_6 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_External_Tag); 
--# expr-done 1199
Is_Equal_6 := Block_Var_1 = Sym_6; 
--# expr-done 1200
if Is_Equal_6 then
   --# expr-start 1205 '<Bind at ast.py:9296>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9296
--# expr-start 1202 '<FieldAccess for expected_type_var at ast.py:9296>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9296
--# expr-start 1201 '<FieldAccess for expr at ast.py:9296>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9296
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_17 := Self.Attribute_Def_Clause_F_Expr;
--# expr-done 1201
      if Fld_17 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_17);
--# expr-done 1202
Fld_18.Value := No_Entity;
Entity_Vars.Reset (Fld_18);
--# expr-start 1204 '<FieldAccess for std_entity at ast.py:9296>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9296
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1203 '<Symbol String at ???>' Sym_7 None
Sym_7 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_String); 
--# expr-done 1203
--# property-call-start AdaNode.std_entity
Fld_19 := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym_7);
--# end
--# expr-done 1204
Bind_Result := Solver.Create_Assign (Fld_18, Fld_19, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9296") else null)); 
--# expr-done 1205
   If_Result_3 := Bind_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9298") else null)); 
   If_Result_3 := True_Rel;
end if;
      Inc_Ref (If_Result_3);
--# expr-done 1206
And_Pred_2 := Create_And (Fld_16, If_Result_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9294") else null)); 
--# expr-done 1207
   If_Result_4 := And_Pred_2;
end if;
      Inc_Ref (If_Result_4);
   If_Result_5 := If_Result_4;
end if;
      Inc_Ref (If_Result_5);
--# expr-done 1208
Var_Expr := Block_Var;
if Var_Expr /= No_Entity_Attribute_Ref then
   --# scope-start
   --# bind ar Var_Expr
--# expr-start 1210 '<FieldAccess for sub_equation at ast.py:9301>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9301
--# expr-start 1209 '<FieldAccess for prefix at ast.py:9301>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9301
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_20 := Create_Internal_Entity_Name (Node => Var_Expr.Node.Attribute_Ref_F_Prefix, Info => Var_Expr.Info);
--# expr-done 1209
      if Fld_20.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_21 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_20.Node, Env => Env, Origin => Origin, E_Info => Fld_20.Info);
--# end
--# expr-done 1210
Scope_Result := Fld_21;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1883;
   Result_Var := Scope_Result;
else
   True_Rel_1 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9301") else null)); 
   Result_Var := True_Rel_1;
end if;
      Inc_Ref (Result_Var);
And_Pred_3 := Create_And (If_Result_5, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9276") else null)); 
--# expr-done 1211
Let_Result := And_Pred_3; Inc_Ref (Let_Result);
--# end
--# expr-done 1212
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_739;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_739;
               Finalizer_Scope_1883;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Def_Clause_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Enum_Rep_Clause
   --
      procedure Initialize_Fields_For_Enum_Rep_Clause
        (Self : Bare_Enum_Rep_Clause
         ; Enum_Rep_Clause_F_Type_Name : Bare_Name
         ; Enum_Rep_Clause_F_Aggregate : Bare_Base_Aggregate
        ) is
      begin
            Self.Enum_Rep_Clause_F_Type_Name := Enum_Rep_Clause_F_Type_Name;
            Self.Enum_Rep_Clause_F_Aggregate := Enum_Rep_Clause_F_Aggregate;
      end Initialize_Fields_For_Enum_Rep_Clause;
   function Enum_Rep_Clause_F_Type_Name
     (Node : Bare_Enum_Rep_Clause) return Bare_Name
   is
   begin
         return Node.Enum_Rep_Clause_F_Type_Name;
   end;
   function Enum_Rep_Clause_F_Aggregate
     (Node : Bare_Enum_Rep_Clause) return Bare_Base_Aggregate
   is
   begin
         return Node.Enum_Rep_Clause_F_Aggregate;
   end;
--# property-start EnumRepClause.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9238
pragma Warnings (Off, "is not referenced");
 function Enum_Rep_Clause_P_Xref_Equation
  (Node : Bare_Enum_Rep_Clause
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Enum_Rep_Clause := Bare_Enum_Rep_Clause (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Enum_Rep_Clause :=
     Internal_Entity_Enum_Rep_Clause'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_737 with Inline_Always;
      Fld : Internal_Entity_Name;
Fld_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_737 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_737;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1215 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1214 '<FieldAccess for xref_no_overloading at ast.py:9240>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9240
--# expr-start 1213 '<FieldAccess for type_name at ast.py:9240>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9240
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Enum_Rep_Clause_F_Type_Name, Info => Ent.Info);
--# expr-done 1213
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_1 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 1214
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 1215
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_737;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_737;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Rep_Clause_P_Xref_Equation;
--# end
--# property-start EnumRepClause.params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9243
pragma Warnings (Off, "is not referenced");
 function Enum_Rep_Clause_P_Params
  (Node : Bare_Enum_Rep_Clause
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Param_Actual_Array_Access
is
   Self : Bare_Enum_Rep_Clause := Bare_Enum_Rep_Clause (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Enum_Rep_Clause :=
     Internal_Entity_Enum_Rep_Clause'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Param_Actual_Array_Access;
            procedure Finalizer_Scope_738 with Inline_Always;
      Block_Var : Internal_Entity_Enum_Literal_Decl_List;
Block_Var_1 : Internal_Entity_Assoc_List;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity_Type_Decl;
Fld_3 : Internal_Entity_Type_Def;
Cast_Result_2 : Internal_Entity_Enum_Type_Def;
Fld_4 : Internal_Entity_Enum_Literal_Decl_List;
Fld_5 : Internal_Entity_Base_Aggregate;
Fld_6 : Internal_Entity_Assoc_List;
Saved : Internal_Entity_Enum_Literal_Decl_List;
Fld_7 : Bare_Enum_Literal_Decl_List;
Fld_8 : Internal_Entity_Info;
As_Entity : Internal_Entity_Enum_Literal_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Enum_Literal_Decl;
Fld_9 : Internal_Entity_Defining_Name;
Fld_10 : Internal_Entity_Defining_Name;
Fld_11 : Internal_Entity_Expr;
New_Struct : Internal_Param_Actual;
Item_51 : Internal_Entity_Enum_Literal_Decl;
Bare_Item_51 : Bare_Enum_Literal_Decl;
Index_52 : Integer;
Map_Result : Internal_Param_Actual_Array_Access;
Let_Result : Internal_Param_Actual_Array_Access;
            procedure Finalizer_Scope_738 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_738;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1231 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1223 '<FieldAccess for enum_literals at ast.py:9249>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9249
--# expr-start 1222 '<Cast to EnumTypeDef at ast.py:9249>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9249
--# expr-start 1221 '<FieldAccess for type_def at ast.py:9249>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9249
--# expr-start 1220 '<Cast to TypeDecl at ast.py:9249>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9249
--# expr-start 1219 '<FieldAccess for root_type at ast.py:9249>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9249
--# expr-start 1218 '<Cast to BaseTypeDecl at ast.py:9249>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9249
--# expr-start 1217 '<FieldAccess for referenced_decl at ast.py:9249>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9249
--# expr-start 1216 '<FieldAccess for type_name at ast.py:9249>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9249
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Enum_Rep_Clause_F_Type_Name, Info => Ent.Info);
--# expr-done 1216
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld_1 := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Fld.Node, Imprecise_Fallback => False, E_Info => Fld.Info);
--# end
--# expr-done 1217
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 1218
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.root_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Root_Type (Node => Cast_Result.Node, Origin => No_Bare_Ada_Node, E_Info => Cast_Result.Info);
--# end
--# expr-done 1219
   if Fld_2.Node = null
      or else Fld_2.Node.Kind in Ada_Type_Decl_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Type_Decl
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
   else
         Cast_Result_1 := No_Entity_Type_Decl;
   end if;
--# expr-done 1220
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Type_Def (Node => Cast_Result_1.Node.Type_Decl_F_Type_Def, Info => Cast_Result_1.Info);
--# expr-done 1221
   if Fld_3.Node = null
      or else Fld_3.Node.Kind in Ada_Enum_Type_Def_Range
   then
      Cast_Result_2 := Create_Internal_Entity_Enum_Type_Def
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
   else
         Cast_Result_2 := No_Entity_Enum_Type_Def;
   end if;
--# expr-done 1222
      if Cast_Result_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Enum_Literal_Decl_List (Node => Cast_Result_2.Node.Enum_Type_Def_F_Enum_Literals, Info => Cast_Result_2.Info);
--# expr-done 1223
Block_Var := Fld_4; 
--# bind el Block_Var
--# expr-start 1225 '<FieldAccess for assocs at ast.py:9253>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9253
--# expr-start 1224 '<FieldAccess for aggregate at ast.py:9253>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9253
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Base_Aggregate (Node => Ent.Node.Enum_Rep_Clause_F_Aggregate, Info => Ent.Info);
--# expr-done 1224
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Assoc_List (Node => Fld_5.Node.Base_Aggregate_F_Assocs, Info => Fld_5.Info);
--# expr-done 1225
Block_Var_1 := Fld_6; 
--# bind ra Block_Var_1
--# expr-start 1230 '<Map at ast.py:9255>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9255
Saved := Block_Var; 
Fld_7 := Saved.Node;
      declare
         Map_Result_Vec : Internal_Param_Actual_Vectors.Vector;
      begin
      if Fld_7 = null then
      Map_Result := Create_Internal_Param_Actual_Array (0);
      else
         Index_52 := 0;
      declare
         Collection : constant Bare_Enum_Literal_Decl_List := Fld_7;
      begin
         for Untyped_Item_51 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_51 := Untyped_Item_51; 
Is_Null := Bare_Item_51 = null; 
if Is_Null then
   If_Result := No_Entity_Enum_Literal_Decl;
else
Fld_8 := Saved.Info;
As_Entity := (Info => Fld_8, Node => Bare_Item_51); 
   If_Result := As_Entity;
end if;
                  Item_51 := If_Result; 
   --# scope-start
               --# bind l Item_51
               --# bind i Index_52
      --# expr-start 1229 '<New ParamActual at ast.py:9256>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9256
--# expr-start 1227 '<FieldAccess for actual_for_param_at at ast.py:9258>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9258
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1226 '<FieldAccess for name at ast.py:9258>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9258
      if Item_51.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Create_Internal_Entity_Defining_Name (Node => Item_51.Node.Enum_Literal_Decl_F_Name, Info => Item_51.Info);
--# expr-done 1226
--# property-call-start AssocList.actual_for_param_at
Fld_11 := Libadalang.Implementation.Assoc_List_P_Actual_For_Param_At (Node => Block_Var_1.Node, Param => Fld_10, Pos => Index_52, E_Info => Block_Var_1.Info);
--# end
--# expr-done 1227
--# expr-start 1228 '<FieldAccess for name at ast.py:9257>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9257
      if Item_51.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Defining_Name (Node => Item_51.Node.Enum_Literal_Decl_F_Name, Info => Item_51.Info);
--# expr-done 1228
New_Struct := (Actual => Fld_11, Param => Fld_9); 
--# expr-done 1229
         declare
            Item_To_Append : constant Internal_Param_Actual := New_Struct;
         begin
            Internal_Param_Actual_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
               Index_52 := Index_52 + 1;
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Param_Actual_Array
           (Items_Count => Natural (Internal_Param_Actual_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Param_Actual_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Param_Actual_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Param_Actual_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1230
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 1231
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_738;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_738;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Rep_Clause_P_Params;
--# end
   --
   --  Primitives for Bare_Record_Rep_Clause
   --
      procedure Initialize_Fields_For_Record_Rep_Clause
        (Self : Bare_Record_Rep_Clause
         ; Record_Rep_Clause_F_Name : Bare_Name
         ; Record_Rep_Clause_F_At_Expr : Bare_Expr
         ; Record_Rep_Clause_F_Components : Bare_Ada_Node_List
        ) is
      begin
            Self.Record_Rep_Clause_F_Name := Record_Rep_Clause_F_Name;
            Self.Record_Rep_Clause_F_At_Expr := Record_Rep_Clause_F_At_Expr;
            Self.Record_Rep_Clause_F_Components := Record_Rep_Clause_F_Components;
      end Initialize_Fields_For_Record_Rep_Clause;
   function Record_Rep_Clause_F_Name
     (Node : Bare_Record_Rep_Clause) return Bare_Name
   is
   begin
         return Node.Record_Rep_Clause_F_Name;
   end;
   function Record_Rep_Clause_F_At_Expr
     (Node : Bare_Record_Rep_Clause) return Bare_Expr
   is
   begin
         return Node.Record_Rep_Clause_F_At_Expr;
   end;
   function Record_Rep_Clause_F_Components
     (Node : Bare_Record_Rep_Clause) return Bare_Ada_Node_List
   is
   begin
         return Node.Record_Rep_Clause_F_Components;
   end;
--# property-start RecordRepClause.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9342
pragma Warnings (Off, "is not referenced");
 function Record_Rep_Clause_P_Xref_Equation
  (Node : Bare_Record_Rep_Clause
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Record_Rep_Clause := Bare_Record_Rep_Clause (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Record_Rep_Clause :=
     Internal_Entity_Record_Rep_Clause'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_742 with Inline_Always;
            procedure Finalizer_Scope_1885 with Inline_Always;
      Var_Expr : Internal_Entity_Expr;
Fld : Internal_Entity_Name;
Fld_1 : Logic_Equation;
Fld_2 : Internal_Entity_Expr;
Fld_3 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_742 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_742;
            procedure Finalizer_Scope_1885 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_1885;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1237 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1236 '<And at ast.py:9343>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9343
--# expr-start 1233 '<FieldAccess for xref_no_overloading at ast.py:9344>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9344
--# expr-start 1232 '<FieldAccess for name at ast.py:9344>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9344
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Record_Rep_Clause_F_Name, Info => Ent.Info);
--# expr-done 1232
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_1 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 1233
--# expr-start 1234 '<FieldAccess for at_expr at ast.py:9345>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9345
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Expr (Node => Ent.Node.Record_Rep_Clause_F_At_Expr, Info => Ent.Info);
--# expr-done 1234
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind e Var_Expr
--# expr-start 1235 '<FieldAccess for sub_equation at ast.py:9346>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9346
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 1235
Scope_Result := Fld_3;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1885;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9347") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred := Create_And (Fld_1, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9343") else null)); 
--# expr-done 1236
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 1237
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_742;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_742;
               Finalizer_Scope_1885;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Record_Rep_Clause_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Aspect_Spec
   --
      procedure Initialize_Fields_For_Aspect_Spec
        (Self : Bare_Aspect_Spec
         ; Aspect_Spec_F_Aspect_Assocs : Bare_Aspect_Assoc_List
        ) is
      begin
            Self.Aspect_Spec_F_Aspect_Assocs := Aspect_Spec_F_Aspect_Assocs;
      end Initialize_Fields_For_Aspect_Spec;
   function Aspect_Spec_F_Aspect_Assocs
     (Node : Bare_Aspect_Spec) return Bare_Aspect_Assoc_List
   is
   begin
         return Node.Aspect_Spec_F_Aspect_Assocs;
   end;
   --
   --  Primitives for Bare_Base_Assoc
   --
--# property-start '[dispatcher]BaseAssoc.assoc_expr' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Assoc_P_Assoc_Expr
  (Node : Bare_Base_Assoc
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Expr
is
   Self : Bare_Base_Assoc := Bare_Base_Assoc (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Assoc (Self.Kind) is
               when Ada_Contract_Case_Assoc =>
                  --# property-call-start ContractCaseAssoc.assoc_expr
                  Property_Result := Contract_Case_Assoc_P_Assoc_Expr
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Pragma_Argument_Assoc =>
                  --# property-call-start PragmaArgumentAssoc.assoc_expr
                  Property_Result := Pragma_Argument_Assoc_P_Assoc_Expr
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Assoc_P_Assoc_Expr;
--# end
   --
   --  Primitives for Bare_Contract_Case_Assoc
   --
      procedure Initialize_Fields_For_Contract_Case_Assoc
        (Self : Bare_Contract_Case_Assoc
         ; Contract_Case_Assoc_F_Guard : Bare_Ada_Node
         ; Contract_Case_Assoc_F_Consequence : Bare_Expr
        ) is
      begin
            Self.Contract_Case_Assoc_F_Guard := Contract_Case_Assoc_F_Guard;
            Self.Contract_Case_Assoc_F_Consequence := Contract_Case_Assoc_F_Consequence;
      end Initialize_Fields_For_Contract_Case_Assoc;
   function Contract_Case_Assoc_F_Guard
     (Node : Bare_Contract_Case_Assoc) return Bare_Ada_Node
   is
   begin
         return Node.Contract_Case_Assoc_F_Guard;
   end;
   function Contract_Case_Assoc_F_Consequence
     (Node : Bare_Contract_Case_Assoc) return Bare_Expr
   is
   begin
         return Node.Contract_Case_Assoc_F_Consequence;
   end;
--# property-start ContractCaseAssoc.assoc_expr /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11490
pragma Warnings (Off, "is not referenced");
 function Contract_Case_Assoc_P_Assoc_Expr
  (Node : Bare_Contract_Case_Assoc
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Expr
is
   Self : Bare_Contract_Case_Assoc := Bare_Contract_Case_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Contract_Case_Assoc :=
     Internal_Entity_Contract_Case_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Expr;
      Fld : Internal_Entity_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1238 '<FieldAccess for consequence at ast.py:11490>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11490
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Contract_Case_Assoc_F_Consequence, Info => Ent.Info);
--# expr-done 1238
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Contract_Case_Assoc_P_Assoc_Expr;
--# end
   --
   --  Primitives for Bare_Pragma_Argument_Assoc
   --
      procedure Initialize_Fields_For_Pragma_Argument_Assoc
        (Self : Bare_Pragma_Argument_Assoc
         ; Pragma_Argument_Assoc_F_Name : Bare_Name
         ; Pragma_Argument_Assoc_F_Expr : Bare_Expr
        ) is
      begin
            Self.Pragma_Argument_Assoc_F_Name := Pragma_Argument_Assoc_F_Name;
            Self.Pragma_Argument_Assoc_F_Expr := Pragma_Argument_Assoc_F_Expr;
      end Initialize_Fields_For_Pragma_Argument_Assoc;
   function Pragma_Argument_Assoc_F_Name
     (Node : Bare_Pragma_Argument_Assoc) return Bare_Name
   is
   begin
         return Node.Pragma_Argument_Assoc_F_Name;
   end;
   function Pragma_Argument_Assoc_F_Expr
     (Node : Bare_Pragma_Argument_Assoc) return Bare_Expr
   is
   begin
         return Node.Pragma_Argument_Assoc_F_Expr;
   end;
--# property-start PragmaArgumentAssoc.assoc_expr /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9219
pragma Warnings (Off, "is not referenced");
 function Pragma_Argument_Assoc_P_Assoc_Expr
  (Node : Bare_Pragma_Argument_Assoc
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Expr
is
   Self : Bare_Pragma_Argument_Assoc := Bare_Pragma_Argument_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Pragma_Argument_Assoc :=
     Internal_Entity_Pragma_Argument_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Expr;
      Fld : Internal_Entity_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1239 '<FieldAccess for expr at ast.py:9219>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9219
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Pragma_Argument_Assoc_F_Expr, Info => Ent.Info);
--# expr-done 1239
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Pragma_Argument_Assoc_P_Assoc_Expr;
--# end
   --
   --  Primitives for Bare_Base_Formal_Param_Holder
   --
--# property-start '[dispatcher]BaseFormalParamHolder.abstract_formal_params' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Formal_Param_Holder_P_Abstract_Formal_Params
  (Node : Bare_Base_Formal_Param_Holder
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Base_Formal_Param_Holder := Bare_Base_Formal_Param_Holder (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Formal_Param_Holder (Self.Kind) is
               when Ada_Entry_Spec .. Ada_Enum_Subp_Spec | Ada_Subp_Spec =>
                  --# property-call-start BaseSubpSpec.abstract_formal_params
                  Property_Result := Base_Subp_Spec_P_Abstract_Formal_Params
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Predefined_Bin_Op_Spec =>
                  --# property-call-start PredefinedBinOpSpec.abstract_formal_params
                  Property_Result := Predefined_Bin_Op_Spec_P_Abstract_Formal_Params
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Predefined_Un_Op_Spec =>
                  --# property-call-start PredefinedUnOpSpec.abstract_formal_params
                  Property_Result := Predefined_Un_Op_Spec_P_Abstract_Formal_Params
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Component_List =>
                  --# property-call-start ComponentList.abstract_formal_params
                  Property_Result := Component_List_P_Abstract_Formal_Params
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Unknown_Discriminant_Part =>
                  --# property-call-start DiscriminantPart.abstract_formal_params
                  Property_Result := Discriminant_Part_P_Abstract_Formal_Params
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Known_Discriminant_Part =>
                  --# property-call-start KnownDiscriminantPart.abstract_formal_params
                  Property_Result := Known_Discriminant_Part_P_Abstract_Formal_Params
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Entry_Completion_Formal_Params =>
                  --# property-call-start EntryCompletionFormalParams.abstract_formal_params
                  Property_Result := Entry_Completion_Formal_Params_P_Abstract_Formal_Params
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Formal_Part =>
                  --# property-call-start GenericFormalPart.abstract_formal_params
                  Property_Result := Generic_Formal_Part_P_Abstract_Formal_Params
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Formal_Param_Holder_P_Abstract_Formal_Params;
--# end
--# property-start BaseFormalParamHolder.unpacked_formal_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4023
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Holder_P_Unpacked_Formal_Params
  (Node : Bare_Base_Formal_Param_Holder
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Base_Formal_Param_Holder := Bare_Base_Formal_Param_Holder (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Formal_Param_Holder :=
     Internal_Entity_Base_Formal_Param_Holder'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_255 with Inline_Always;
      Fld : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_1 : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_255 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_255;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1241 '<FieldAccess for unpack_formals at ast.py:4024>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4024
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1240 '<FieldAccess for abstract_formal_params at ast.py:4024>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4024
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseFormalParamHolder.abstract_formal_params'
Fld := Libadalang.Implementation.Dispatcher_Base_Formal_Param_Holder_P_Abstract_Formal_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1240
--# property-call-start AdaNode.unpack_formals
Fld_1 := Libadalang.Implementation.Ada_Node_P_Unpack_Formals (Node => Self, Formal_Params => Fld);
--# end
--# expr-done 1241
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_255;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_255;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Holder_P_Unpacked_Formal_Params;
--# end
--# property-start BaseFormalParamHolder.formal_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4031
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Holder_P_Formal_Params
  (Node : Bare_Base_Formal_Param_Holder
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Base_Formal_Param_Holder := Bare_Base_Formal_Param_Holder (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Formal_Param_Holder :=
     Internal_Entity_Base_Formal_Param_Holder'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_256 with Inline_Always;
            procedure Finalizer_Scope_1886 with Inline_Always;
            procedure Finalizer_Scope_1888 with Inline_Always;
      Match_0 : Internal_Entity_Component_List;
Match_1 : Internal_Entity_Base_Formal_Param_Holder;
Fld : Internal_Entity_Ada_Node_List;
Fld_1 : Bare_Ada_Node_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Base_Formal_Param_Decl;
Item_53 : Internal_Entity;
Bare_Item_53 : Bare_Ada_Node;
Is_A : Boolean;
Map_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_3 : Internal_Entity_Defining_Name_Array_Access;
Fld_4 : Internal_Entity_Defining_Name_Array_Access;
Match_Prefix : Internal_Entity_Base_Formal_Param_Holder;
Cast_Result_1 : Internal_Entity_Component_List;
Let_Result : Internal_Entity_Defining_Name_Array_Access;
Scope_Result : Internal_Entity_Defining_Name_Array_Access;
Cast_Result_2 : Internal_Entity_Base_Formal_Param_Holder;
Let_Result_1 : Internal_Entity_Defining_Name_Array_Access;
Scope_Result_1 : Internal_Entity_Defining_Name_Array_Access;
Match_Result : Internal_Entity_Defining_Name_Array_Access;
Let_Result_2 : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_256 is
            begin
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Match_Result);
                     Dec_Ref (Let_Result_2);
            end Finalizer_Scope_256;
            procedure Finalizer_Scope_1886 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_1886;
            procedure Finalizer_Scope_1888 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_1888;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1249 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 1248 '<Match at ast.py:4037>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4037
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Base_Formal_Param_Holder (Match_Prefix.Node.Kind) is
         when Ada_Component_List =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Component_List
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_1; 
--# bind r Match_0
--# expr-start 1246 '<FieldAccess for unpack_formals at ast.py:4039>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4039
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1245 '<FilterMap at ast.py:4039>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4039
   --# expr-start 1242 '<FieldAccess for components at ast.py:4039>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4039
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Ada_Node_List (Node => Match_0.Node.Component_List_F_Components, Info => Match_0.Info);
--# expr-done 1242
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Internal_Entity_Base_Formal_Param_Decl_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Internal_Entity_Base_Formal_Param_Decl_Array (0);
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_1;
      begin
         for Untyped_Item_53 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_53 := Untyped_Item_53; 
Is_Null := Bare_Item_53 = null; 
if Is_Null then
   If_Result := No_Entity;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_53); 
   If_Result := As_Entity;
end if;
                  Item_53 := If_Result; 
   --# scope-start
               --# bind e Item_53
         --# expr-start 1243 '<IsA BaseFormalParamDecl at ???>' Is_A None
Is_A := Item_53.Node /= null 
and then Item_53.Node.Kind in Ada_Base_Formal_Param_Decl; 
--# expr-done 1243
         if Is_A then
      --# expr-start 1244 '<Cast to BaseFormalParamDecl at ???>' Cast_Result None
Cast_Expr := Item_53; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Base_Formal_Param_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Formal_Param_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Base_Formal_Param_Decl;
   end if;
--# expr-done 1244
         declare
            Item_To_Append : constant Internal_Entity_Base_Formal_Param_Decl := Cast_Result;
         begin
            Internal_Entity_Base_Formal_Param_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Base_Formal_Param_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Formal_Param_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Base_Formal_Param_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Base_Formal_Param_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Base_Formal_Param_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1245
--# property-call-start AdaNode.unpack_formals
Fld_3 := Libadalang.Implementation.Ada_Node_P_Unpack_Formals (Node => Self, Formal_Params => Map_Result);
--# end
--# expr-done 1246
Let_Result := Fld_3; Inc_Ref (Let_Result);
--# end
Scope_Result := Let_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1886;
            Match_Result := Scope_Result; Inc_Ref (Match_Result);
         when Ada_Entry_Spec .. Ada_Subp_Spec | Ada_Known_Discriminant_Part .. Ada_Generic_Formal_Part =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Base_Formal_Param_Holder
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_2; 
--# bind _ Match_1
--# expr-start 1247 '<FieldAccess for unpacked_formal_params at ast.py:4040>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4040
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld_4 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1247
Let_Result_1 := Fld_4; Inc_Ref (Let_Result_1);
--# end
Scope_Result_1 := Let_Result_1;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_1888;
            Match_Result := Scope_Result_1; Inc_Ref (Match_Result);
end case;
--# expr-done 1248
Let_Result_2 := Match_Result; Inc_Ref (Let_Result_2);
--# end
--# expr-done 1249
      Property_Result := Let_Result_2;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_256;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_256;
               Finalizer_Scope_1886;
               Finalizer_Scope_1888;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Holder_P_Formal_Params;
--# end
--# property-start BaseFormalParamHolder.match_param_list /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4044
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Holder_P_Match_Param_List
  (Node : Bare_Base_Formal_Param_Holder
      ; Params : Internal_Entity_Assoc_List
      ; Is_Dottable_Subp : Boolean
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Param_Match_Array_Access
is
   Self : Bare_Base_Formal_Param_Holder := Bare_Base_Formal_Param_Holder (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Formal_Param_Holder :=
     Internal_Entity_Base_Formal_Param_Holder'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind params Params
   --# bind is_dottable_subp Is_Dottable_Subp
   --# bind env Env
   Property_Result : Internal_Param_Match_Array_Access;
            procedure Finalizer_Scope_257 with Inline_Always;
      Fld : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_1 : Internal_Param_Match_Array_Access;
Let_Result : Internal_Param_Match_Array_Access;
            procedure Finalizer_Scope_257 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_257;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1252 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1251 '<FieldAccess for match_formals at ast.py:4046>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4046
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1250 '<FieldAccess for abstract_formal_params at ast.py:4047>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4047
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseFormalParamHolder.abstract_formal_params'
Fld := Libadalang.Implementation.Dispatcher_Base_Formal_Param_Holder_P_Abstract_Formal_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1250
--# property-call-start AdaNode.match_formals
Fld_1 := Libadalang.Implementation.Ada_Node_P_Match_Formals (Node => Self, Formal_Params => Fld, Params => Params, Is_Dottable_Subp => Is_Dottable_Subp);
--# end
--# expr-done 1251
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 1252
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_257;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_257;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Holder_P_Match_Param_List;
--# end
--# property-start BaseFormalParamHolder.nb_min_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4050
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Holder_P_Nb_Min_Params
  (Node : Bare_Base_Formal_Param_Holder
  )
   return Integer
is
   Self : Bare_Base_Formal_Param_Holder := Bare_Base_Formal_Param_Holder (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Integer;
            procedure Finalizer_Scope_258 with Inline_Always;
      As_Entity : Internal_Entity_Base_Formal_Param_Holder;
Is_Null : Boolean;
If_Result : Internal_Entity_Base_Formal_Param_Holder;
Fld : Internal_Entity_Defining_Name_Array_Access;
Fld_1 : Internal_Entity_Base_Formal_Param_Decl;
Fld_2 : Boolean;
Map_Result : Internal_Entity_Defining_Name_Array_Access;
Len : Integer;
            procedure Finalizer_Scope_258 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_258;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1258 '<length at ast.py:4051>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4051
--# expr-start 1257 '<Filter at ast.py:4051>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4051
   --# expr-start 1254 '<FieldAccess for unpacked_formal_params at ast.py:4051>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4051
--# expr-start 1253 '<as_bare_entity at ast.py:4051>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4051
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity_Base_Formal_Param_Holder;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 1253
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 1254
      declare
         Map_Result_Vec : Internal_Entity_Defining_Name_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Fld;
      begin
         for Item_54 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_54
         --# expr-start 1256 '<FieldAccess for is_mandatory at ast.py:4052>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4052
--# expr-start 1255 '<FieldAccess for formal_decl at ast.py:4052>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4052
      if Item_54.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_1 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Item_54.Node, E_Info => Item_54.Info);
--# end
--# expr-done 1255
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseFormalParamDecl.is_mandatory'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Formal_Param_Decl_P_Is_Mandatory (Node => Fld_1.Node);
--# end
--# expr-done 1256
         if Fld_2 then
         declare
            Item_To_Append : constant Internal_Entity_Defining_Name := Item_54;
         begin
            Internal_Entity_Defining_Name_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Defining_Name_Array
           (Items_Count => Natural (Internal_Entity_Defining_Name_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Defining_Name_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Defining_Name_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Defining_Name_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1257
Len := Length (Map_Result); 
--# expr-done 1258
      Property_Result := Len;
   --# end
      Finalizer_Scope_258;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_258;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Holder_P_Nb_Min_Params;
--# end
--# property-start BaseFormalParamHolder.nb_max_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4060
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Holder_P_Nb_Max_Params
  (Node : Bare_Base_Formal_Param_Holder
  )
   return Integer
is
   Self : Bare_Base_Formal_Param_Holder := Bare_Base_Formal_Param_Holder (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Integer;
            procedure Finalizer_Scope_259 with Inline_Always;
      As_Entity : Internal_Entity_Base_Formal_Param_Holder;
Is_Null : Boolean;
If_Result : Internal_Entity_Base_Formal_Param_Holder;
Fld : Internal_Entity_Defining_Name_Array_Access;
Len : Integer;
            procedure Finalizer_Scope_259 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_259;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1261 '<length at ast.py:4061>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4061
--# expr-start 1260 '<FieldAccess for unpacked_formal_params at ast.py:4061>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4061
--# expr-start 1259 '<as_bare_entity at ast.py:4061>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4061
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity_Base_Formal_Param_Holder;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 1259
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 1260
Len := Length (Fld); 
--# expr-done 1261
      Property_Result := Len;
   --# end
      Finalizer_Scope_259;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_259;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Holder_P_Nb_Max_Params;
--# end
--# property-start BaseFormalParamHolder.paramless /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4069
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Holder_P_Paramless
  (Node : Bare_Base_Formal_Param_Holder
      ; Dottable_Subp : Boolean
      ; Can_Be : Boolean
         := True
  )
   return Boolean
is
   Self : Bare_Base_Formal_Param_Holder := Bare_Base_Formal_Param_Holder (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind dottable_subp Dottable_Subp
   --# bind can_be Can_Be
   Property_Result : Boolean;
      Block_Var : Integer;
Fld : Integer;
Fld_1 : Integer;
If_Result : Integer;
Is_Equal : Boolean;
If_Result_1 : Boolean;
Is_Equal_1 : Boolean;
If_Result_2 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1267 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1264 '<If at ast.py:4075>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4075
if Can_Be then
   --# expr-start 1262 '<FieldAccess for nb_min_params at ast.py:4075>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4075
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.nb_min_params
Fld := Libadalang.Implementation.Base_Formal_Param_Holder_P_Nb_Min_Params (Node => Self);
--# end
--# expr-done 1262
   If_Result := Fld;
else
   --# expr-start 1263 '<FieldAccess for nb_max_params at ast.py:4075>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4075
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.nb_max_params
Fld_1 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Nb_Max_Params (Node => Self);
--# end
--# expr-done 1263
   If_Result := Fld_1;
end if;
--# expr-done 1264
Block_Var := If_Result; 
--# bind nb_params Block_Var
if Dottable_Subp then
   --# expr-start 1265 '<Eq at ast.py:4077>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4077
Is_Equal := Block_Var = 1; 
--# expr-done 1265
   If_Result_1 := Is_Equal;
else
   If_Result_1 := False;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
   --# expr-start 1266 '<Eq at ast.py:4078>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4078
Is_Equal_1 := Block_Var = 0; 
--# expr-done 1266
   If_Result_2 := Is_Equal_1;
end if;
Let_Result := If_Result_2; 
--# end
--# expr-done 1267
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Holder_P_Paramless;
--# end
--# property-start BaseFormalParamHolder.is_matching_param_list /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4082
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Holder_P_Is_Matching_Param_List
  (Node : Bare_Base_Formal_Param_Holder
      ; Params : Internal_Entity_Assoc_List
      ; Is_Dottable_Subp : Boolean
      ; Env : Lexical_Env
  )
   return Boolean
is
   Self : Bare_Base_Formal_Param_Holder := Bare_Base_Formal_Param_Holder (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind params Params
   --# bind is_dottable_subp Is_Dottable_Subp
   --# bind env Env
   Property_Result : Boolean;
            procedure Finalizer_Scope_261 with Inline_Always;
      Block_Var : Internal_Entity_Base_Formal_Param_Holder;
Block_Var_1 : Internal_Param_Match_Array_Access;
Block_Var_2 : Integer;
Block_Var_3 : Integer;
As_Entity : Internal_Entity_Base_Formal_Param_Holder;
Is_Null : Boolean;
If_Result : Internal_Entity_Base_Formal_Param_Holder;
Fld : Internal_Param_Match_Array_Access;
Fld_1 : Integer;
Arith_Result : Integer;
Fld_2 : Integer;
If_Result_1 : Integer;
Fld_3 : Integer;
Arith_Result_1 : Integer;
Fld_4 : Integer;
If_Result_2 : Integer;
Fld_5 : Bare_Assoc_List;
Len : Integer;
Comp_Result : Boolean;
Fld_6 : Boolean;
Quantifier_Result : Boolean;
If_Result_3 : Boolean;
Fld_7 : Internal_Entity_Defining_Name;
Fld_8 : Internal_Entity_Base_Formal_Param_Decl;
Fld_9 : Boolean;
Map_Result : Internal_Param_Match_Array_Access;
Len_1 : Integer;
Is_Equal : Boolean;
If_Result_4 : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_261 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_261;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1287 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1268 '<as_bare_entity at ast.py:4088>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4088
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity_Base_Formal_Param_Holder;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 1268
Block_Var := If_Result; 
--# bind bare Block_Var
--# expr-start 1269 '<FieldAccess for match_param_list at ast.py:4089>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4089
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.match_param_list
Fld := Libadalang.Implementation.Base_Formal_Param_Holder_P_Match_Param_List (Node => Block_Var.Node, Params => Params, Is_Dottable_Subp => Is_Dottable_Subp, Env => Env, E_Info => Block_Var.Info);
--# end
--# expr-done 1269
Block_Var_1 := Fld; Inc_Ref (Block_Var_1);
--# bind match_list Block_Var_1
--# expr-start 1273 '<If at ast.py:4091>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4091
if Is_Dottable_Subp then
   --# expr-start 1271 '<Arithmetic - at ast.py:4091>' Arith_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4091
--# expr-start 1270 '<FieldAccess for nb_max_params at ast.py:4091>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4091
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.nb_max_params
Fld_1 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Nb_Max_Params (Node => Block_Var.Node);
--# end
--# expr-done 1270
Arith_Result := (Fld_1 - 1); 
--# expr-done 1271
   If_Result_1 := Arith_Result;
else
   --# expr-start 1272 '<FieldAccess for nb_max_params at ast.py:4091>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4091
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.nb_max_params
Fld_2 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Nb_Max_Params (Node => Block_Var.Node);
--# end
--# expr-done 1272
   If_Result_1 := Fld_2;
end if;
--# expr-done 1273
Block_Var_2 := If_Result_1; 
--# bind nb_max_params Block_Var_2
--# expr-start 1277 '<If at ast.py:4094>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4094
if Is_Dottable_Subp then
   --# expr-start 1275 '<Arithmetic - at ast.py:4094>' Arith_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4094
--# expr-start 1274 '<FieldAccess for nb_min_params at ast.py:4094>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4094
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.nb_min_params
Fld_3 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Nb_Min_Params (Node => Block_Var.Node);
--# end
--# expr-done 1274
Arith_Result_1 := (Fld_3 - 1); 
--# expr-done 1275
   If_Result_2 := Arith_Result_1;
else
   --# expr-start 1276 '<FieldAccess for nb_min_params at ast.py:4094>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4094
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.nb_min_params
Fld_4 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Nb_Min_Params (Node => Block_Var.Node);
--# end
--# expr-done 1276
   If_Result_2 := Fld_4;
end if;
--# expr-done 1277
Block_Var_3 := If_Result_2; 
--# bind nb_min_params Block_Var_3
--# expr-start 1278 '<length at ast.py:4098>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4098
Fld_5 := Params.Node; 
Len := Length (Fld_5); 
--# expr-done 1278
Comp_Result := Len <= Block_Var_2; 
if Comp_Result then
   --# expr-start 1280 '<AllQuantifier at ast.py:4099>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4099
      Quantifier_Result := True;
      declare
         Collection : constant Internal_Param_Match_Array_Access := Block_Var_1;
      begin
         for Item_55 of
               Collection.Items
         loop
   --# scope-start
               --# bind m Item_55
      --# expr-start 1279 '<FieldAccess for has_matched at ast.py:4099>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4099
Fld_6 := Item_55.Has_Matched;
--# expr-done 1279
      Quantifier_Result := Fld_6;
   --# end
               exit when not Quantifier_Result;
         end loop;
      end;
--# expr-done 1280
   If_Result_3 := Quantifier_Result;
else
   If_Result_3 := False;
end if;
if If_Result_3 then
   --# expr-start 1286 '<Eq at ast.py:4100>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4100
--# expr-start 1285 '<length at ast.py:4100>' Len_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4100
--# expr-start 1284 '<Filter at ast.py:4100>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4100
      declare
         Map_Result_Vec : Internal_Param_Match_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Param_Match_Array_Access := Block_Var_1;
      begin
         for Item_56 of
               Collection.Items
         loop
   --# scope-start
               --# bind m Item_56
         --# expr-start 1283 '<FieldAccess for is_mandatory at ast.py:4101>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4101
--# expr-start 1282 '<FieldAccess for formal_decl at ast.py:4101>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4101
--# expr-start 1281 '<FieldAccess for formal at ast.py:4101>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4101
Fld_7 := Item_56.Formal;
--# expr-done 1281
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_8 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Fld_7.Node, E_Info => Fld_7.Info);
--# end
--# expr-done 1282
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseFormalParamDecl.is_mandatory'
Fld_9 := Libadalang.Implementation.Dispatcher_Base_Formal_Param_Decl_P_Is_Mandatory (Node => Fld_8.Node);
--# end
--# expr-done 1283
         if Fld_9 then
         declare
            Item_To_Append : constant Internal_Param_Match := Item_56;
         begin
            Internal_Param_Match_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Param_Match_Array
           (Items_Count => Natural (Internal_Param_Match_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Param_Match_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Param_Match_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Param_Match_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1284
Len_1 := Length (Map_Result); 
--# expr-done 1285
Is_Equal := Len_1 = Block_Var_3; 
--# expr-done 1286
   If_Result_4 := Is_Equal;
else
   If_Result_4 := False;
end if;
Let_Result := If_Result_4; 
--# end
--# expr-done 1287
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_261;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_261;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Holder_P_Is_Matching_Param_List;
--# end
--# property-start BaseFormalParamHolder.param_types /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4107
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Holder_P_Param_Types
  (Node : Bare_Base_Formal_Param_Holder
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl_Array_Access
is
   Self : Bare_Base_Formal_Param_Holder := Bare_Base_Formal_Param_Holder (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Formal_Param_Holder :=
     Internal_Entity_Base_Formal_Param_Holder'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_262 with Inline_Always;
      Fld : Internal_Entity_Defining_Name_Array_Access;
Fld_1 : Internal_Entity_Base_Formal_Param_Decl;
Fld_2 : Internal_Entity_Type_Expr;
Fld_3 : Internal_Entity_Base_Type_Decl;
Map_Result : Internal_Entity_Base_Type_Decl_Array_Access;
Let_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_262 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_262;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1293 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1292 '<Map at ast.py:4111>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4111
   --# expr-start 1288 '<FieldAccess for unpacked_formal_params at ast.py:4111>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4111
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1288
      declare
         Map_Result_Vec : Internal_Entity_Base_Type_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Fld;
      begin
         for Item_57 of
               Collection.Items
         loop
   --# scope-start
               --# bind fp Item_57
      --# expr-start 1291 '<FieldAccess for real_designated_type at ast.py:4113>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4113
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1290 '<FieldAccess for type_expression at ast.py:4113>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4113
--# expr-start 1289 '<FieldAccess for formal_decl at ast.py:4113>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4113
      if Item_57.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_1 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Item_57.Node, E_Info => Item_57.Info);
--# end
--# expr-done 1289
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.type_expression'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Type_Expression (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 1290
--# property-call-start BaseFormalParamHolder.real_designated_type
Fld_3 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Real_Designated_Type (Node => Ent.Node, Typ => Fld_2, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 1291
         declare
            Item_To_Append : constant Internal_Entity_Base_Type_Decl := Fld_3;
         begin
            Internal_Entity_Base_Type_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Base_Type_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Type_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Base_Type_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Base_Type_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Base_Type_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1292
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 1293
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_262;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_262;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Holder_P_Param_Types;
--# end
--# property-start BaseFormalParamHolder.param_modes /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4117
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Holder_P_Param_Modes
  (Node : Bare_Base_Formal_Param_Holder
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Mode_Array_Access
is
   Self : Bare_Base_Formal_Param_Holder := Bare_Base_Formal_Param_Holder (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Formal_Param_Holder :=
     Internal_Entity_Base_Formal_Param_Holder'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Mode_Array_Access;
            procedure Finalizer_Scope_263 with Inline_Always;
      Fld : Internal_Entity_Defining_Name_Array_Access;
Fld_1 : Internal_Entity_Base_Formal_Param_Decl;
Cast_Result : Internal_Entity_Param_Spec;
Fld_2 : Internal_Entity_Mode;
Map_Result : Internal_Entity_Mode_Array_Access;
Let_Result : Internal_Entity_Mode_Array_Access;
            procedure Finalizer_Scope_263 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_263;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1299 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1298 '<Map at ast.py:4121>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4121
   --# expr-start 1294 '<FieldAccess for unpacked_formal_params at ast.py:4121>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4121
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1294
      declare
         Map_Result_Vec : Internal_Entity_Mode_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Fld;
      begin
         for Item_58 of
               Collection.Items
         loop
   --# scope-start
               --# bind fp Item_58
      --# expr-start 1297 '<FieldAccess for mode at ast.py:4122>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4122
--# expr-start 1296 '<Cast to ParamSpec at ast.py:4122>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4122
--# expr-start 1295 '<FieldAccess for formal_decl at ast.py:4122>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4122
      if Item_58.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_1 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Item_58.Node, E_Info => Item_58.Info);
--# end
--# expr-done 1295
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Param_Spec_Range
   then
      Cast_Result := Create_Internal_Entity_Param_Spec
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         Cast_Result := No_Entity_Param_Spec;
   end if;
--# expr-done 1296
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Mode (Node => Cast_Result.Node.Param_Spec_F_Mode, Info => Cast_Result.Info);
--# expr-done 1297
         declare
            Item_To_Append : constant Internal_Entity_Mode := Fld_2;
         begin
            Internal_Entity_Mode_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Mode_Array
           (Items_Count => Natural (Internal_Entity_Mode_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Mode_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Mode_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Mode_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1298
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 1299
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_263;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_263;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Holder_P_Param_Modes;
--# end
--# property-start BaseFormalParamHolder.real_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4127
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Holder_P_Real_Type
  (Node : Bare_Base_Formal_Param_Holder
      ; Typ : Internal_Entity_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Formal_Param_Holder := Bare_Base_Formal_Param_Holder (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Formal_Param_Holder :=
     Internal_Entity_Base_Formal_Param_Holder'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind typ Typ
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Block_Var : Internal_Entity_Base_Type_Decl;
Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Entity_Base_Type_Decl;
Var_Expr_2 : Bare_Ada_Node;
Var_Expr_3 : Internal_Entity_Anonymous_Type_Decl;
Var_Expr_4 : Internal_Entity_Base_Type_Decl;
Var_Expr_5 : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Info;
Fld_1 : Internal_Metadata;
Fld_2 : Bare_Ada_Node;
Fld_3 : Internal_Entity_Info;
Fld_4 : Env_Rebindings;
Fld_5 : Internal_Entity_Info;
Fld_6 : Boolean;
Fld_7 : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
Fld_8 : Bare_Base_Type_Decl;
Is_Null : Boolean;
Fld_9 : Internal_Entity_Base_Type_Decl;
Fld_10 : Bare_Base_Type_Decl;
Scope_Result : Bare_Base_Type_Decl;
Result_Var : Bare_Base_Type_Decl;
Fld_11 : Internal_Entity_Base_Type_Decl;
Fld_12 : Bare_Base_Type_Decl;
Scope_Result_1 : Bare_Base_Type_Decl;
Result_Var_1 : Bare_Base_Type_Decl;
Cast_Result_1 : Bare_Base_Type_Decl;
Is_Equal : Boolean;
Fld_13 : Internal_Entity_Info;
Fld_14 : Internal_Metadata;
Fld_15 : Bare_Ada_Node;
Scope_Result_2 : Bare_Ada_Node;
Fld_16 : Bare_Base_Type_Decl;
Cast_Result_2 : Bare_Ada_Node;
Result_Var_2 : Bare_Ada_Node;
Fld_17 : Internal_Entity_Info;
Fld_18 : Env_Rebindings;
Fld_19 : Internal_Entity_Info;
Fld_20 : Boolean;
Fld_21 : Internal_Entity;
Cast_Result_3 : Internal_Entity_Base_Type_Decl;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_4 : Internal_Entity_Anonymous_Type_Decl;
Fld_22 : Internal_Entity_Type_Def;
Is_A : Boolean;
Not_Val : Boolean;
Scope_Result_3 : Boolean;
Result_Var_3 : Boolean;
Fld_23 : Internal_Entity_Base_Type_Decl;
Fld_24 : Internal_Entity_Base_Type_Decl;
Fld_25 : Internal_Entity_Base_Type_Decl;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_5 : Internal_Entity;
Cast_Expr_2 : Internal_Entity_Base_Type_Decl;
Cast_Result_6 : Internal_Entity;
Is_Equiv : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
Scope_Result_4 : Internal_Entity_Base_Type_Decl;
Result_Var_4 : Internal_Entity_Base_Type_Decl;
Scope_Result_5 : Internal_Entity_Base_Type_Decl;
Result_Var_5 : Internal_Entity_Base_Type_Decl;
If_Result_1 : Internal_Entity_Base_Type_Decl;
If_Result_2 : Internal_Entity_Base_Type_Decl;
If_Result_3 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1336 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1308 '<Cast to BaseTypeDecl at ast.py:4134>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4134
--# expr-start 1307 '<FieldAccess for entity_no_md at ast.py:4134>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4134
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1302 '<FieldAccess for primitive at ast.py:4135>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4135
--# expr-start 1301 '<FieldAccess for md at ast.py:4135>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4135
--# expr-start 1300 '<FieldAccess for info at ast.py:4135>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4135
Fld := Ent.Info;
--# expr-done 1300
Fld_1 := Fld.Md;
--# expr-done 1301
Fld_2 := Fld_1.Primitive;
--# expr-done 1302
--# expr-start 1304 '<FieldAccess for rebindings at ast.py:4136>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4136
--# expr-start 1303 '<FieldAccess for info at ast.py:4136>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4136
Fld_3 := Ent.Info;
--# expr-done 1303
Fld_4 := Fld_3.Rebindings;
--# expr-done 1304
--# expr-start 1306 '<FieldAccess for from_rebound at ast.py:4137>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4137
--# expr-start 1305 '<FieldAccess for info at ast.py:4137>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4137
Fld_5 := Ent.Info;
--# expr-done 1305
Fld_6 := Fld_5.From_Rebound;
--# expr-done 1306
--# property-call-start AdaNode.entity_no_md
Fld_7 := Libadalang.Implementation.Ada_Node_P_Entity_No_Md (Node => Ent.Node, N => Fld_2, Rebindings => Fld_4, From_Rebound => Fld_6);
--# end
--# expr-done 1307
   if Fld_7.Node = null
      or else Fld_7.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld_7.Node,
         Info => Fld_7.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 1308
Block_Var := Cast_Result; 
--# bind prim_type Block_Var
--# expr-start 1335 '<Cond at ast.py:4140>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4140
--# expr-start 1310 '<IsNull at ast.py:4141>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4141
--# expr-start 1309 '<FieldAccess for node at ast.py:4141>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4141
Fld_8 := Block_Var.Node;
--# expr-done 1309
Is_Null := Fld_8 = null; 
--# expr-done 1310
if Is_Null then
   If_Result_3 := Typ;
else
--# expr-start 1315 '<Eq at ast.py:4144>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4144
Var_Expr := Block_Var;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 1312 '<FieldAccess for node at ast.py:4144>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4144
--# expr-start 1311 '<FieldAccess for canonical_type at ast.py:4144>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4144
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.canonical_type'
Fld_9 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Canonical_Type (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 1311
Fld_10 := Fld_9.Node;
--# expr-done 1312
Scope_Result := Fld_10;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node;
end if;
      Cast_Result_1 := Result_Var;
Var_Expr_1 := Typ;
if Var_Expr_1 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 1314 '<FieldAccess for node at ast.py:4144>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4144
--# expr-start 1313 '<FieldAccess for canonical_type at ast.py:4144>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4144
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.canonical_type'
Fld_11 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Canonical_Type (Node => Var_Expr_1.Node, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 1313
Fld_12 := Fld_11.Node;
--# expr-done 1314
Scope_Result_1 := Fld_12;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Bare_Ada_Node;
end if;
Is_Equal := Cast_Result_1 = Result_Var_1; 
--# expr-done 1315
if Is_Equal then
   --# expr-start 1325 '<Cast to BaseTypeDecl at ast.py:4145>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4145
--# expr-start 1324 '<FieldAccess for entity_no_md at ast.py:4145>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4145
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1318 '<FieldAccess for primitive_real_type at ast.py:4146>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4146
--# expr-start 1317 '<FieldAccess for md at ast.py:4146>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4146
--# expr-start 1316 '<FieldAccess for info at ast.py:4146>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4146
Fld_13 := Ent.Info;
--# expr-done 1316
Fld_14 := Fld_13.Md;
--# expr-done 1317
Fld_15 := Fld_14.Primitive_Real_Type;
--# expr-done 1318
Var_Expr_2 := Fld_15;
if Var_Expr_2 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind e Var_Expr_2
Scope_Result_2 := Var_Expr_2;
   --# end
   Result_Var_2 := Scope_Result_2;
else
--# expr-start 1319 '<FieldAccess for node at ast.py:4146>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4146
Fld_16 := Typ.Node;
--# expr-done 1319
      Cast_Result_2 := Fld_16;
   Result_Var_2 := Cast_Result_2;
end if;
--# expr-start 1321 '<FieldAccess for rebindings at ast.py:4147>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4147
--# expr-start 1320 '<FieldAccess for info at ast.py:4147>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4147
Fld_17 := Ent.Info;
--# expr-done 1320
Fld_18 := Fld_17.Rebindings;
--# expr-done 1321
--# expr-start 1323 '<FieldAccess for from_rebound at ast.py:4148>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4148
--# expr-start 1322 '<FieldAccess for info at ast.py:4148>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4148
Fld_19 := Ent.Info;
--# expr-done 1322
Fld_20 := Fld_19.From_Rebound;
--# expr-done 1323
--# property-call-start AdaNode.entity_no_md
Fld_21 := Libadalang.Implementation.Ada_Node_P_Entity_No_Md (Node => Ent.Node, N => Result_Var_2, Rebindings => Fld_18, From_Rebound => Fld_20);
--# end
--# expr-done 1324
   if Fld_21.Node = null
      or else Fld_21.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_3 := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld_21.Node,
         Info => Fld_21.Info);
   else
         Cast_Result_3 := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 1325
   If_Result_2 := Cast_Result_3;
else
--# expr-start 1326 '<Cast to AnonymousTypeDecl at ast.py:4154>' Cast_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4154
Cast_Expr := Typ; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Anonymous_Type_Decl_Range
   then
      Cast_Result_4 := Create_Internal_Entity_Anonymous_Type_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result_4 := No_Entity_Anonymous_Type_Decl;
   end if;
--# expr-done 1326
Var_Expr_3 := Cast_Result_4;
if Var_Expr_3 /= No_Entity_Anonymous_Type_Decl then
   --# scope-start
   --# bind td Var_Expr_3
--# expr-start 1329 '<Not at ast.py:4155>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4155
--# expr-start 1328 '<IsA AccessToSubpDef at ast.py:4155>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4155
--# expr-start 1327 '<FieldAccess for type_def at ast.py:4155>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4155
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_22 := Create_Internal_Entity_Type_Def (Node => Var_Expr_3.Node.Type_Decl_F_Type_Def, Info => Var_Expr_3.Info);
--# expr-done 1327
Is_A := Fld_22.Node /= null 
and then Fld_22.Node.Kind in Ada_Access_To_Subp_Def_Range; 
--# expr-done 1328
Not_Val := not (Is_A); 
--# expr-done 1329
Scope_Result_3 := Not_Val;
   --# end
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := False;
end if;
if Result_Var_3 then
--# expr-start 1330 '<FieldAccess for accessed_type at ast.py:4157>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4157
      if Typ.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_23 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Typ.Node, Origin => Origin, E_Info => Typ.Info);
--# end
--# expr-done 1330
Var_Expr_4 := Fld_23;
if Var_Expr_4 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind at Var_Expr_4
--# expr-start 1331 '<FieldAccess for real_type at ast.py:4158>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4158
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.real_type
Fld_24 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Real_Type (Node => Ent.Node, Typ => Var_Expr_4, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 1331
Var_Expr_5 := Fld_24;
if Var_Expr_5 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind rat Var_Expr_5
--# expr-start 1334 '<If at ast.py:4159>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4159
--# expr-start 1332 '<Eq at ast.py:4160>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4160
Cast_Expr_1 := Var_Expr_4; 
      Cast_Result_5 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Cast_Expr_2 := Var_Expr_5; 
      Cast_Result_6 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
Is_Equiv := Equivalent (Cast_Result_5, Cast_Result_6); 
--# expr-done 1332
if Is_Equiv then
   If_Result := Typ;
else
   --# expr-start 1333 '<FieldAccess for anonymous_access_type at ast.py:4162>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4162
      if Var_Expr_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.anonymous_access_type
Fld_25 := Libadalang.Implementation.Base_Type_Decl_P_Anonymous_Access_Type (Node => Var_Expr_5.Node, E_Info => Var_Expr_5.Info);
--# end
--# expr-done 1333
   If_Result := Fld_25;
end if;
--# expr-done 1334
Scope_Result_4 := If_Result;
   --# end
   Result_Var_4 := Scope_Result_4;
else
   Result_Var_4 := No_Entity_Base_Type_Decl;
end if;
Scope_Result_5 := Result_Var_4;
   --# end
   Result_Var_5 := Scope_Result_5;
else
   Result_Var_5 := No_Entity_Base_Type_Decl;
end if;
   If_Result_1 := Result_Var_5;
else
   If_Result_1 := Typ;
end if;
   If_Result_2 := If_Result_1;
end if;
   If_Result_3 := If_Result_2;
end if;
--# expr-done 1335
Let_Result := If_Result_3; 
--# end
--# expr-done 1336
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Holder_P_Real_Type;
--# end
--# property-start BaseFormalParamHolder.real_designated_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4172
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Holder_P_Real_Designated_Type
  (Node : Bare_Base_Formal_Param_Holder
      ; Typ : Internal_Entity_Type_Expr
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Formal_Param_Holder := Bare_Base_Formal_Param_Holder (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Formal_Param_Holder :=
     Internal_Entity_Base_Formal_Param_Holder'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind typ Typ
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Block_Var : Env_Rebindings;
Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Entity_Base_Type_Decl;
Var_Expr_2 : Internal_Entity_Base_Type_Decl;
Rt : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Info;
Fld_1 : Internal_Metadata;
Fld_2 : Bare_Ada_Node;
Cast_Result : Bare_Base_Type_Decl;
As_Entity : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
Fld_3 : Internal_Entity_Info;
Fld_4 : Internal_Metadata;
Fld_5 : Bare_Ada_Node;
Cast_Result_1 : Bare_Base_Type_Decl;
Fld_6 : Env_Rebindings;
Scope_Result : Env_Rebindings;
Result_Var : Env_Rebindings;
Fld_7 : Internal_Entity_Base_Type_Decl;
Fld_8 : Internal_Entity;
Scope_Result_1 : Internal_Entity;
Result_Var_1 : Internal_Entity;
Cast_Result_2 : Internal_Entity_Base_Type_Decl;
Fld_9 : Internal_Entity_Base_Type_Decl;
Fld_10 : Bare_Type_Expr;
Fld_11 : Internal_Entity_Info;
Fld_12 : Internal_Metadata;
Fld_13 : Internal_Entity_Info;
Fld_14 : Env_Rebindings;
Rebinding : Env_Rebindings;
Fld_15 : Internal_Entity_Info;
Fld_16 : Boolean;
New_Struct : Internal_Entity_Info;
New_Struct_1 : Internal_Entity_Type_Expr;
Fld_17 : Internal_Entity_Base_Type_Decl;
Is_Equal : Boolean;
If_Result_1 : Internal_Entity_Base_Type_Decl;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_4 : Internal_Entity;
Is_Equiv : Boolean;
If_Result_2 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
Scope_Result_2 : Internal_Entity_Base_Type_Decl;
Result_Var_2 : Internal_Entity_Base_Type_Decl;
Let_Result_1 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1367 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 1341 '<as_bare_entity at ast.py:4234>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4234
--# expr-start 1340 '<Cast to BaseTypeDecl at ast.py:4234>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4234
--# expr-start 1339 '<FieldAccess for primitive_real_type at ast.py:4234>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4234
--# expr-start 1338 '<FieldAccess for md at ast.py:4234>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4234
--# expr-start 1337 '<FieldAccess for info at ast.py:4234>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4234
Fld := Ent.Info;
--# expr-done 1337
Fld_1 := Fld.Md;
--# expr-done 1338
Fld_2 := Fld_1.Primitive_Real_Type;
--# expr-done 1339
   if Fld_2 = null
      or else Fld_2.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Fld_2;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 1340
Is_Null := Cast_Result = null; 
if Is_Null then
   If_Result := No_Entity_Base_Type_Decl;
else
As_Entity := (Info => No_Entity_Info, Node => Cast_Result); 
   If_Result := As_Entity;
end if;
--# expr-done 1341
Var_Expr := If_Result;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 1346 '<FieldAccess for find_base_type_rebindings at ast.py:4234>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4234
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1345 '<Cast to BaseTypeDecl at ast.py:4236>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4236
--# expr-start 1344 '<FieldAccess for primitive at ast.py:4236>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4236
--# expr-start 1343 '<FieldAccess for md at ast.py:4236>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4236
--# expr-start 1342 '<FieldAccess for info at ast.py:4236>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4236
Fld_3 := Ent.Info;
--# expr-done 1342
Fld_4 := Fld_3.Md;
--# expr-done 1343
Fld_5 := Fld_4.Primitive;
--# expr-done 1344
   if Fld_5 = null
      or else Fld_5.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_1 := Fld_5;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 1345
--# property-call-start BaseTypeDecl.find_base_type_rebindings
Fld_6 := Libadalang.Implementation.Base_Type_Decl_P_Find_Base_Type_Rebindings (Node => Var_Expr.Node, Target => Cast_Result_1, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 1346
Scope_Result := Fld_6;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := null;
end if;
Block_Var := Result_Var; 
--# bind base_rebindings Block_Var
--# expr-start 1349 '<Cast to BaseTypeDecl at ast.py:4240>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4240
--# expr-start 1347 '<FieldAccess for designated_type at ast.py:4240>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4240
      if Typ.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_7 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Typ.Node, Origin => Origin, E_Info => Typ.Info);
--# end
--# expr-done 1347
Var_Expr_2 := Fld_7;
if Var_Expr_2 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 1348 '<FieldAccess for without_md at ast.py:4240>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4240
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.without_md
Fld_8 := Libadalang.Implementation.Ada_Node_P_Without_Md (Node => Var_Expr_2.Node, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 1348
Scope_Result_1 := Fld_8;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity;
end if;
   if Result_Var_1.Node = null
      or else Result_Var_1.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_2 := Create_Internal_Entity_Base_Type_Decl
        (Node => Result_Var_1.Node,
         Info => Result_Var_1.Info);
   else
         Cast_Result_2 := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 1349
Var_Expr_1 := Cast_Result_2;
if Var_Expr_1 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind t Var_Expr_1
--# expr-start 1366 '<Let at ast.py:4241>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4241
--# scope-start
--# expr-start 1350 '<FieldAccess for real_type at ast.py:4242>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4242
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.real_type
Fld_9 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Real_Type (Node => Ent.Node, Typ => Var_Expr_1, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 1350
Rt := Fld_9; 
--# bind rt Rt
--# expr-start 1365 '<If at ast.py:4243>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4243
--# expr-start 1351 '<Eq at ast.py:4244>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4244
Cast_Expr := Var_Expr_1; 
      Cast_Result_3 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Cast_Expr_1 := Rt; 
      Cast_Result_4 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Is_Equiv := Equivalent (Cast_Result_3, Cast_Result_4); 
--# expr-done 1351
if Is_Equiv then
   --# expr-start 1364 '<If at ast.py:4245>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4245
--# expr-start 1352 '<IsNull at ast.py:4246>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4246
Is_Equal := Block_Var = null; 
--# expr-done 1352
if Is_Equal then
   If_Result_1 := Var_Expr_1;
else
   --# expr-start 1363 '<FieldAccess for designated_type at ???>' Fld_17 None
--# expr-start 1362 '<New TypeExpr.entity at ???>' New_Struct_1 None
--# expr-start 1360 '<New EntityInfo at ast.py:4250>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4250
--# expr-start 1354 '<FieldAccess for from_rebound at ast.py:4254>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4254
--# expr-start 1353 '<FieldAccess for info at ast.py:4254>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4254
Fld_15 := Typ.Info;
--# expr-done 1353
Fld_16 := Fld_15.From_Rebound;
--# expr-done 1354
--# expr-start 1356 '<FieldAccess for md at ast.py:4251>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4251
--# expr-start 1355 '<FieldAccess for info at ast.py:4251>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4251
Fld_11 := Typ.Info;
--# expr-done 1355
Fld_12 := Fld_11.Md;
--# expr-done 1356
--# expr-start 1359 '<concat_rebindings at ast.py:4252>' Rebinding /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4252
--# expr-start 1358 '<FieldAccess for rebindings at ast.py:4252>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4252
--# expr-start 1357 '<FieldAccess for info at ast.py:4252>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4252
Fld_13 := Typ.Info;
--# expr-done 1357
Fld_14 := Fld_13.Rebindings;
--# expr-done 1358
Rebinding := AST_Envs.Combine (Fld_14, Block_Var); 
--# expr-done 1359
New_Struct := (From_Rebound => Fld_16, Md => Fld_12, Rebindings => Rebinding); 
--# expr-done 1360
--# expr-start 1361 '<FieldAccess for node at ast.py:4249>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4249
Fld_10 := Typ.Node;
--# expr-done 1361
New_Struct_1 := (Info => New_Struct, Node => Fld_10); 
--# expr-done 1362
      if New_Struct_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_17 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => New_Struct_1.Node, Origin => Origin, E_Info => New_Struct_1.Info);
--# end
--# expr-done 1363
   If_Result_1 := Fld_17;
end if;
--# expr-done 1364
   If_Result_2 := If_Result_1;
else
   If_Result_2 := Rt;
end if;
--# expr-done 1365
Let_Result := If_Result_2; 
--# end
--# expr-done 1366
Scope_Result_2 := Let_Result;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Entity_Base_Type_Decl;
end if;
Let_Result_1 := Result_Var_2; 
--# end
--# expr-done 1367
      Property_Result := Let_Result_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Holder_P_Real_Designated_Type;
--# end
--# property-start BaseFormalParamHolder.call_argument_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4264
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Holder_P_Call_Argument_Equation
  (Node : Bare_Base_Formal_Param_Holder
      ; Param : Internal_Entity_Base_Formal_Param_Decl
      ; Arg : Internal_Entity_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Base_Formal_Param_Holder := Bare_Base_Formal_Param_Holder (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Formal_Param_Holder :=
     Internal_Entity_Base_Formal_Param_Holder'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind param Param
   --# bind arg Arg
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_266 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Type_Expr;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_3 : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_266 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_3);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_266;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1374 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1369 '<FieldAccess for real_designated_type at ast.py:4272>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4272
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1368 '<FieldAccess for type_expression at ast.py:4272>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4272
      if Param.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.type_expression'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Type_Expression (Node => Param.Node, E_Info => Param.Info);
--# end
--# expr-done 1368
--# property-call-start BaseFormalParamHolder.real_designated_type
Fld_1 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Real_Designated_Type (Node => Ent.Node, Typ => Fld, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 1369
Block_Var := Fld_1; 
--# bind param_type Block_Var
--# expr-start 1373 '<And at ast.py:4273>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4273
--# expr-start 1371 '<Bind at ast.py:4274>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4274
--# expr-start 1370 '<FieldAccess for expected_type_var at ast.py:4274>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4274
      if Arg.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Arg.Node);
--# expr-done 1370
Fld_2.Value := No_Entity;
Entity_Vars.Reset (Fld_2);
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_2, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:4274") else null)); 
--# expr-done 1371
--# expr-start 1372 '<FieldAccess for matches_expected_formal_type at ast.py:4275>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4275
      if Arg.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_3 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Arg.Node, Origin => Origin);
--# end
--# expr-done 1372
And_Pred := Create_And (Bind_Result, Fld_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:4273") else null)); 
--# expr-done 1373
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 1374
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_266;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_266;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Holder_P_Call_Argument_Equation;
--# end
--# property-start BaseFormalParamHolder.match_formal_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4279
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Holder_P_Match_Formal_Params
  (Node : Bare_Base_Formal_Param_Holder
      ; Other : Internal_Entity_Base_Formal_Param_Holder
      ; Match_Names : Boolean
         := True
      ; Ignore_First_Param : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Formal_Param_Holder := Bare_Base_Formal_Param_Holder (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Formal_Param_Holder :=
     Internal_Entity_Base_Formal_Param_Holder'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind other Other
   --# bind match_names Match_Names
   --# bind ignore_first_param Ignore_First_Param
   Property_Result : Boolean;
            procedure Finalizer_Scope_267 with Inline_Always;
            procedure Finalizer_Scope_1903 with Inline_Always;
      Block_Var : Internal_Entity_Defining_Name_Array_Access;
Block_Var_1 : Internal_Entity_Defining_Name_Array_Access;
Block_Var_2 : Internal_Entity_Base_Type_Decl_Array_Access;
Block_Var_3 : Internal_Entity_Base_Type_Decl_Array_Access;
Var_Expr : Internal_Entity_Defining_Name_Array_Access;
Var_Expr_1 : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Defining_Name_Array_Access;
Index_60 : Integer;
Is_Equal : Boolean;
Not_Val : Boolean;
Map_Result : Internal_Entity_Defining_Name_Array_Access;
If_Result : Internal_Entity_Defining_Name_Array_Access;
Scope_Result : Internal_Entity_Defining_Name_Array_Access;
Result_Var : Internal_Entity_Defining_Name_Array_Access;
Fld_1 : Internal_Entity_Defining_Name_Array_Access;
Fld_2 : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_3 : Internal_Entity_Base_Type_Decl_Array_Access;
Dyn_Var_Bind_Result : Internal_Entity_Base_Type_Decl_Array_Access;
Fld_4 : Bare_Base_Formal_Param_Holder;
Fld_5 : Bare_Ada_Node;
Bound_Origin_1 : Bare_Ada_Node;
Fld_6 : Internal_Entity_Base_Type_Decl_Array_Access;
Dyn_Var_Bind_Result_1 : Internal_Entity_Base_Type_Decl_Array_Access;
Len : Integer;
Len_1 : Integer;
Is_Equal_1 : Boolean;
Fld_7 : Bare_Ada_Node;
Bound_Origin_2 : Bare_Ada_Node;
Not_Val_1 : Boolean;
Fld_8 : Internal_Entity_Name;
Get_Result : Internal_Entity_Defining_Name;
Fld_9 : Internal_Entity_Name;
Fld_10 : Bare_Name;
Fld_11 : Boolean;
If_Result_1 : Boolean;
Get_Result_1 : Internal_Entity_Base_Type_Decl;
Get_Result_2 : Internal_Entity_Base_Type_Decl;
Fld_12 : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
If_Result_2 : Boolean;
Index_62 : Integer;
Quantifier_Result : Boolean;
Dyn_Var_Bind_Result_2 : Boolean;
If_Result_3 : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_267 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Block_Var_2);
                     Dec_Ref (Fld_6);
                     Dec_Ref (Dyn_Var_Bind_Result_1);
                     Dec_Ref (Block_Var_3);
            end Finalizer_Scope_267;
            procedure Finalizer_Scope_1903 is
            begin
                     Dec_Ref (Var_Expr);
                     Dec_Ref (Map_Result);
                     Dec_Ref (If_Result);
            end Finalizer_Scope_1903;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1403 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1375 '<FieldAccess for unpacked_formal_params at ast.py:4287>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4287
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1375
Var_Expr := Fld;
      Inc_Ref (Var_Expr);
if Var_Expr /= No_Internal_Entity_Defining_Name_Array_Type then
   --# scope-start
   --# bind params Var_Expr
--# expr-start 1379 '<If at ast.py:4288>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4288
if Ignore_First_Param then
   --# expr-start 1378 '<Filter at ast.py:4290>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4290
      declare
         Map_Result_Vec : Internal_Entity_Defining_Name_Vectors.Vector;
      begin
         Index_60 := 0;
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Var_Expr;
      begin
         for Item_59 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_59
               --# bind i Index_60
         --# expr-start 1377 '<Not at ast.py:4290>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4290
--# expr-start 1376 '<Eq at ast.py:4290>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4290
Is_Equal := Index_60 = 0; 
--# expr-done 1376
Not_Val := not (Is_Equal); 
--# expr-done 1377
         if Not_Val then
         declare
            Item_To_Append : constant Internal_Entity_Defining_Name := Item_59;
         begin
            Internal_Entity_Defining_Name_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
               Index_60 := Index_60 + 1;
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Defining_Name_Array
           (Items_Count => Natural (Internal_Entity_Defining_Name_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Defining_Name_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Defining_Name_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Defining_Name_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1378
   If_Result := Map_Result;
else
   If_Result := Var_Expr;
end if;
      Inc_Ref (If_Result);
--# expr-done 1379
Scope_Result := If_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1903;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Defining_Name_Array_Type;
end if;
      Inc_Ref (Result_Var);
Block_Var := Result_Var; Inc_Ref (Block_Var);
--# bind self_params Block_Var
--# expr-start 1380 '<FieldAccess for unpacked_formal_params at ast.py:4294>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4294
      if Other.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld_1 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => Other.Node, E_Info => Other.Info);
--# end
--# expr-done 1380
Block_Var_1 := Fld_1; Inc_Ref (Block_Var_1);
--# bind other_params Block_Var_1
--# expr-start 1383 '<bind at ast.py:4296>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4296
--# expr-start 1381 '<FieldAccess for origin_node at ast.py:4296>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4296
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_2 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 1381
Bound_Origin := Fld_2; 
--# expr-start 1382 '<FieldAccess for param_types at ast.py:4296>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4296
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.param_types
Fld_3 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Param_Types (Node => Ent.Node, Origin => Bound_Origin, E_Info => Ent.Info);
--# end
--# expr-done 1382
Dyn_Var_Bind_Result := Fld_3; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 1383
Block_Var_2 := Dyn_Var_Bind_Result; Inc_Ref (Block_Var_2);
--# bind self_types Block_Var_2
--# expr-start 1387 '<bind at ast.py:4297>' Dyn_Var_Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4297
--# expr-start 1385 '<FieldAccess for origin_node at ast.py:4297>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4297
--# expr-start 1384 '<FieldAccess for node at ast.py:4297>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4297
Fld_4 := Other.Node;
--# expr-done 1384
      if Fld_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_5 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Fld_4);
--# end
--# expr-done 1385
Bound_Origin_1 := Fld_5; 
--# expr-start 1386 '<FieldAccess for param_types at ast.py:4298>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4298
      if Other.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.param_types
Fld_6 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Param_Types (Node => Other.Node, Origin => Bound_Origin_1, E_Info => Other.Info);
--# end
--# expr-done 1386
Dyn_Var_Bind_Result_1 := Fld_6; Inc_Ref (Dyn_Var_Bind_Result_1);
--# expr-done 1387
Block_Var_3 := Dyn_Var_Bind_Result_1; Inc_Ref (Block_Var_3);
--# bind other_types Block_Var_3
--# expr-start 1390 '<Eq at ast.py:4300>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4300
--# expr-start 1388 '<length at ast.py:4300>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4300
Len := Length (Block_Var); 
--# expr-done 1388
--# expr-start 1389 '<length at ast.py:4300>' Len_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4300
Len_1 := Length (Block_Var_1); 
--# expr-done 1389
Is_Equal_1 := Len = Len_1; 
--# expr-done 1390
if Is_Equal_1 then
   --# expr-start 1402 '<bind at ast.py:4301>' Dyn_Var_Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4301
--# expr-start 1391 '<FieldAccess for origin_node at ast.py:4301>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4301
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_7 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 1391
Bound_Origin_2 := Fld_7; 
--# expr-start 1401 '<AllQuantifier at ast.py:4301>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4301
      Quantifier_Result := True;
         Index_62 := 0;
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Block_Var;
      begin
         for Item_61 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_61
               --# bind i Index_62
--# expr-start 1392 '<Not at ast.py:4303>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4303
Not_Val_1 := not (Match_Names); 
--# expr-done 1392
if Not_Val_1 then
   If_Result_1 := True;
else
   --# expr-start 1397 '<FieldAccess for matches at ast.py:4304>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4304
--# expr-start 1393 '<FieldAccess for name at ast.py:4304>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4304
      if Item_61.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Name (Node => Item_61.Node.Defining_Name_F_Name, Info => Item_61.Info);
--# expr-done 1393
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1396 '<FieldAccess for node at ast.py:4304>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4304
--# expr-start 1395 '<FieldAccess for name at ast.py:4304>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4304
--# expr-start 1394 '<at at ast.py:4304>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4304
Get_Result := Get (Block_Var_1, Index_62, True); 
--# expr-done 1394
      if Get_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Name (Node => Get_Result.Node.Defining_Name_F_Name, Info => Get_Result.Info);
--# expr-done 1395
Fld_10 := Fld_9.Node;
--# expr-done 1396
--# property-call-start Name.matches
Fld_11 := Libadalang.Implementation.Name_P_Matches (Node => Fld_8.Node, N => Fld_10);
--# end
--# expr-done 1397
   If_Result_1 := Fld_11;
end if;
if If_Result_1 then
--# expr-start 1398 '<at at ast.py:4305>' Get_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4305
Get_Result_1 := Get (Block_Var_2, Index_62, True); 
--# expr-done 1398
Var_Expr_1 := Get_Result_1;
if Var_Expr_1 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 1400 '<FieldAccess for matching_type at ast.py:4305>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4305
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1399 '<at at ast.py:4305>' Get_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4305
Get_Result_2 := Get (Block_Var_3, Index_62, True); 
--# expr-done 1399
--# property-call-start BaseTypeDecl.matching_type
Fld_12 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Type (Node => Var_Expr_1.Node, Expected_Type => Get_Result_2, Origin => Bound_Origin_2, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 1400
Scope_Result_1 := Fld_12;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
   If_Result_2 := Result_Var_1;
else
   If_Result_2 := False;
end if;
      Quantifier_Result := If_Result_2;
   --# end
               exit when not Quantifier_Result;
               Index_62 := Index_62 + 1;
         end loop;
      end;
--# expr-done 1401
Dyn_Var_Bind_Result_2 := Quantifier_Result; 
--# expr-done 1402
   If_Result_3 := Dyn_Var_Bind_Result_2;
else
   If_Result_3 := False;
end if;
Let_Result := If_Result_3; 
--# end
--# expr-done 1403
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_267;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_267;
               Finalizer_Scope_1903;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Holder_P_Match_Formal_Params;
--# end
--# property-start '[dispatcher]BaseFormalParamHolder.match_other' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Formal_Param_Holder_P_Match_Other
  (Node : Bare_Base_Formal_Param_Holder
      ; Other : Internal_Entity_Base_Formal_Param_Holder
      ; Match_Names : Boolean
         := True
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Formal_Param_Holder := Bare_Base_Formal_Param_Holder (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Formal_Param_Holder :=
     Internal_Entity_Base_Formal_Param_Holder'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind other Other
   --# bind match_names Match_Names
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Formal_Param_Holder (Self.Kind) is
               when Ada_Component_List .. Ada_Generic_Formal_Part =>
                  --# property-call-start BaseFormalParamHolder.match_other
                  Property_Result := Base_Formal_Param_Holder_P_Match_Other
                    (Self
                        , Other
                        , Match_Names
                        , E_Info
                    );
                  --# end
               when Ada_Entry_Spec .. Ada_Subp_Spec =>
                  --# property-call-start BaseSubpSpec.match_other
                  Property_Result := Base_Subp_Spec_P_Match_Other
                    (Self
                        , Other
                        , Match_Names
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Formal_Param_Holder_P_Match_Other;
--# end
--# property-start BaseFormalParamHolder.match_other /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4310
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Holder_P_Match_Other
  (Node : Bare_Base_Formal_Param_Holder
      ; Other : Internal_Entity_Base_Formal_Param_Holder
      ; Match_Names : Boolean
         := True
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Formal_Param_Holder := Bare_Base_Formal_Param_Holder (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Formal_Param_Holder :=
     Internal_Entity_Base_Formal_Param_Holder'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind other Other
   --# bind match_names Match_Names
   Property_Result : Boolean;
      Fld : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1405 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1404 '<FieldAccess for match_formal_params at ast.py:4318>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4318
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.match_formal_params
Fld := Libadalang.Implementation.Base_Formal_Param_Holder_P_Match_Formal_Params (Node => Ent.Node, Other => Other, Match_Names => Match_Names, E_Info => Ent.Info);
--# end
--# expr-done 1404
Let_Result := Fld; 
--# end
--# expr-done 1405
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Holder_P_Match_Other;
--# end
   --
   --  Primitives for Bare_Base_Subp_Spec
   --
--# property-start '[dispatcher]BaseSubpSpec.name' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Subp_Spec_P_Name
  (Node : Bare_Base_Subp_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Defining_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Subp_Spec (Self.Kind) is
               when Ada_Entry_Spec =>
                  --# property-call-start EntrySpec.name
                  Property_Result := Entry_Spec_P_Name
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Enum_Subp_Spec =>
                  --# property-call-start EnumSubpSpec.name
                  Property_Result := Enum_Subp_Spec_P_Name
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Predefined_Bin_Op_Spec =>
                  --# property-call-start PredefinedBinOpSpec.name
                  Property_Result := Predefined_Bin_Op_Spec_P_Name
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Predefined_Un_Op_Spec =>
                  --# property-call-start PredefinedUnOpSpec.name
                  Property_Result := Predefined_Un_Op_Spec_P_Name
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Subp_Spec =>
                  --# property-call-start SubpSpec.name
                  Property_Result := Subp_Spec_P_Name
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Subp_Spec_P_Name;
--# end
--# property-start '[dispatcher]BaseSubpSpec.returns' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Subp_Spec_P_Returns
  (Node : Bare_Base_Subp_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Expr
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Type_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Subp_Spec (Self.Kind) is
               when Ada_Entry_Spec =>
                  --# property-call-start EntrySpec.returns
                  Property_Result := Entry_Spec_P_Returns
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Enum_Subp_Spec =>
                  --# property-call-start EnumSubpSpec.returns
                  Property_Result := Enum_Subp_Spec_P_Returns
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Predefined_Bin_Op_Spec =>
                  --# property-call-start PredefinedBinOpSpec.returns
                  Property_Result := Predefined_Bin_Op_Spec_P_Returns
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Predefined_Un_Op_Spec =>
                  --# property-call-start PredefinedUnOpSpec.returns
                  Property_Result := Predefined_Un_Op_Spec_P_Returns
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Subp_Spec =>
                  --# property-call-start SubpSpec.returns
                  Property_Result := Subp_Spec_P_Returns
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Subp_Spec_P_Returns;
--# end
--# property-start '[dispatcher]BaseSubpSpec.params' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Subp_Spec_P_Params
  (Node : Bare_Base_Subp_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Param_Spec_Array_Access
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Param_Spec_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Subp_Spec (Self.Kind) is
               when Ada_Predefined_Bin_Op_Spec .. Ada_Predefined_Un_Op_Spec =>
                  --# property-call-start BaseSubpSpec.params
                  Property_Result := Base_Subp_Spec_P_Params
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Entry_Spec =>
                  --# property-call-start EntrySpec.params
                  Property_Result := Entry_Spec_P_Params
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Enum_Subp_Spec =>
                  --# property-call-start EnumSubpSpec.params
                  Property_Result := Enum_Subp_Spec_P_Params
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Subp_Spec =>
                  --# property-call-start SubpSpec.params
                  Property_Result := Subp_Spec_P_Params
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Subp_Spec_P_Params;
--# end
--# property-start BaseSubpSpec.abstract_formal_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16291
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Abstract_Formal_Params
  (Node : Bare_Base_Subp_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Spec :=
     Internal_Entity_Base_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_1161 with Inline_Always;
      Fld : Internal_Entity_Param_Spec_Array_Access;
Cast_Expr : Internal_Entity_Param_Spec;
Cast_Result : Internal_Entity_Base_Formal_Param_Decl;
Map_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_1161 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_1161;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1408 '<Map at ast.py:16292>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16292
   --# expr-start 1406 '<FieldAccess for params at ast.py:16292>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16292
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.params'
Fld := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1406
      declare
         Map_Result_Vec : Internal_Entity_Base_Formal_Param_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Param_Spec_Array_Access := Fld;
      begin
         for Item_63 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_63
      --# expr-start 1407 '<Cast to BaseFormalParamDecl at ast.py:16292>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16292
Cast_Expr := Item_63; 
      Cast_Result := Create_Internal_Entity_Base_Formal_Param_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
--# expr-done 1407
         declare
            Item_To_Append : constant Internal_Entity_Base_Formal_Param_Decl := Cast_Result;
         begin
            Internal_Entity_Base_Formal_Param_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Base_Formal_Param_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Formal_Param_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Base_Formal_Param_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Base_Formal_Param_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Base_Formal_Param_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1408
      Property_Result := Map_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1161;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1161;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Abstract_Formal_Params;
--# end
--# property-start BaseSubpSpec.match_return_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16296
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Match_Return_Type
  (Node : Bare_Base_Subp_Spec
      ; Other : Internal_Entity_Base_Subp_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Spec :=
     Internal_Entity_Base_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind other Other
   Property_Result : Boolean;
      Block_Var : Internal_Entity_Base_Type_Decl;
Block_Var_1 : Internal_Entity_Base_Type_Decl;
Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_1 : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Base_Type_Decl;
Fld_2 : Bare_Base_Subp_Spec;
Fld_3 : Bare_Ada_Node;
Bound_Origin_1 : Bare_Ada_Node;
Fld_4 : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result_1 : Internal_Entity_Base_Type_Decl;
Fld_5 : Internal_Entity_Type_Expr;
Is_Null : Boolean;
Fld_6 : Internal_Entity_Type_Expr;
Is_Null_1 : Boolean;
If_Result : Boolean;
Fld_7 : Internal_Entity_Type_Expr;
Is_Null_2 : Boolean;
Not_Val : Boolean;
Fld_8 : Internal_Entity_Type_Expr;
Is_Null_3 : Boolean;
Not_Val_1 : Boolean;
If_Result_1 : Boolean;
Fld_9 : Bare_Ada_Node;
Bound_Origin_2 : Bare_Ada_Node;
Fld_10 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Dyn_Var_Bind_Result_2 : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1429 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1411 '<bind at ast.py:16302>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16302
--# expr-start 1409 '<FieldAccess for origin_node at ast.py:16302>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16302
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 1409
Bound_Origin := Fld; 
--# expr-start 1410 '<FieldAccess for return_type at ast.py:16302>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16302
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.return_type
Fld_1 := Libadalang.Implementation.Base_Subp_Spec_P_Return_Type (Node => Ent.Node, Origin => Bound_Origin, E_Info => Ent.Info);
--# end
--# expr-done 1410
Dyn_Var_Bind_Result := Fld_1; 
--# expr-done 1411
Block_Var := Dyn_Var_Bind_Result; 
--# bind self_ret Block_Var
--# expr-start 1415 '<bind at ast.py:16303>' Dyn_Var_Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16303
--# expr-start 1413 '<FieldAccess for origin_node at ast.py:16303>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16303
--# expr-start 1412 '<FieldAccess for node at ast.py:16303>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16303
Fld_2 := Other.Node;
--# expr-done 1412
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_3 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Fld_2);
--# end
--# expr-done 1413
Bound_Origin_1 := Fld_3; 
--# expr-start 1414 '<FieldAccess for return_type at ast.py:16303>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16303
      if Other.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.return_type
Fld_4 := Libadalang.Implementation.Base_Subp_Spec_P_Return_Type (Node => Other.Node, Origin => Bound_Origin_1, E_Info => Other.Info);
--# end
--# expr-done 1414
Dyn_Var_Bind_Result_1 := Fld_4; 
--# expr-done 1415
Block_Var_1 := Dyn_Var_Bind_Result_1; 
--# bind other_ret Block_Var_1
--# expr-start 1417 '<IsNull at ast.py:16305>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16305
--# expr-start 1416 '<FieldAccess for returns at ast.py:16305>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16305
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.returns'
Fld_5 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Returns (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1416
Is_Null := Fld_5.Node = null; 
--# expr-done 1417
if Is_Null then
   --# expr-start 1419 '<IsNull at ast.py:16305>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16305
--# expr-start 1418 '<FieldAccess for returns at ast.py:16305>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16305
      if Other.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.returns'
Fld_6 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Returns (Node => Other.Node, E_Info => Other.Info);
--# end
--# expr-done 1418
Is_Null_1 := Fld_6.Node = null; 
--# expr-done 1419
   If_Result := Is_Null_1;
else
   If_Result := False;
end if;
if If_Result then
   If_Result_3 := True;
else
--# expr-start 1422 '<Not at ast.py:16307>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16307
--# expr-start 1421 '<IsNull at ast.py:16307>' Is_Null_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16307
--# expr-start 1420 '<FieldAccess for returns at ast.py:16307>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16307
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.returns'
Fld_7 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Returns (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1420
Is_Null_2 := Fld_7.Node = null; 
--# expr-done 1421
Not_Val := not (Is_Null_2); 
--# expr-done 1422
if Not_Val then
   --# expr-start 1425 '<Not at ast.py:16307>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16307
--# expr-start 1424 '<IsNull at ast.py:16307>' Is_Null_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16307
--# expr-start 1423 '<FieldAccess for returns at ast.py:16307>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16307
      if Other.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.returns'
Fld_8 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Returns (Node => Other.Node, E_Info => Other.Info);
--# end
--# expr-done 1423
Is_Null_3 := Fld_8.Node = null; 
--# expr-done 1424
Not_Val_1 := not (Is_Null_3); 
--# expr-done 1425
   If_Result_1 := Not_Val_1;
else
   If_Result_1 := False;
end if;
if If_Result_1 then
   --# expr-start 1428 '<bind at ast.py:16308>' Dyn_Var_Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16308
--# expr-start 1426 '<FieldAccess for origin_node at ast.py:16308>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16308
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_9 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 1426
Bound_Origin_2 := Fld_9; 
Var_Expr := Block_Var;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 1427 '<FieldAccess for matching_type at ast.py:16309>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16309
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_type
Fld_10 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Type (Node => Var_Expr.Node, Expected_Type => Block_Var_1, Origin => Bound_Origin_2, E_Info => Var_Expr.Info);
--# end
--# expr-done 1427
Scope_Result := Fld_10;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Dyn_Var_Bind_Result_2 := Result_Var; 
--# expr-done 1428
   If_Result_2 := Dyn_Var_Bind_Result_2;
else
   If_Result_2 := False;
end if;
   If_Result_3 := If_Result_2;
end if;
Let_Result := If_Result_3; 
--# end
--# expr-done 1429
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Match_Return_Type;
--# end
--# property-start BaseSubpSpec.match_signature /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16314
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Match_Signature
  (Node : Bare_Base_Subp_Spec
      ; Other : Internal_Entity_Base_Subp_Spec
      ; Match_Name : Boolean
      ; Use_Entity_Info : Boolean
         := True
      ; Ignore_First_Param : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Spec :=
     Internal_Entity_Base_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind other Other
   --# bind match_name Match_Name
   --# bind use_entity_info Use_Entity_Info
   --# bind ignore_first_param Ignore_First_Param
   Property_Result : Boolean;
      Block_Var : Internal_Entity_Base_Subp_Spec;
As_Entity : Internal_Entity_Base_Subp_Spec;
Is_Null : Boolean;
If_Result : Internal_Entity_Base_Subp_Spec;
If_Result_1 : Internal_Entity_Base_Subp_Spec;
Not_Val : Boolean;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Bare_Defining_Name;
Fld_2 : Internal_Entity_Defining_Name;
Fld_3 : Bare_Defining_Name;
Cast_Result : Bare_Name;
Fld_4 : Boolean;
If_Result_2 : Boolean;
Fld_5 : Boolean;
If_Result_3 : Boolean;
Cast_Expr : Internal_Entity_Base_Subp_Spec;
Cast_Result_1 : Internal_Entity_Base_Formal_Param_Holder;
Fld_6 : Boolean;
If_Result_4 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1440 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1431 '<If at ast.py:16337>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16337
if Use_Entity_Info then
   If_Result_1 := Ent;
else
   --# expr-start 1430 '<as_bare_entity at ast.py:16337>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16337
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity_Base_Subp_Spec;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 1430
   If_Result_1 := If_Result;
end if;
--# expr-done 1431
Block_Var := If_Result_1; 
--# bind ent Block_Var
--# expr-start 1432 '<Not at ast.py:16340>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16340
Not_Val := not (Match_Name); 
--# expr-done 1432
if Not_Val then
   If_Result_2 := True;
else
   --# expr-start 1437 '<FieldAccess for matches at ast.py:16340>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16340
--# expr-start 1434 '<FieldAccess for node at ast.py:16340>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16340
--# expr-start 1433 '<FieldAccess for name at ast.py:16340>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16340
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.name'
Fld := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Name (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 1433
Fld_1 := Fld.Node;
--# expr-done 1434
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1436 '<FieldAccess for node at ast.py:16340>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16340
--# expr-start 1435 '<FieldAccess for name at ast.py:16340>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16340
      if Other.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.name'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Name (Node => Other.Node, E_Info => Other.Info);
--# end
--# expr-done 1435
Fld_3 := Fld_2.Node;
--# expr-done 1436
      Cast_Result := Fld_3;
--# property-call-start Name.matches
Fld_4 := Libadalang.Implementation.Name_P_Matches (Node => Fld_1, N => Cast_Result);
--# end
--# expr-done 1437
   If_Result_2 := Fld_4;
end if;
if If_Result_2 then
   --# expr-start 1438 '<FieldAccess for match_return_type at ast.py:16341>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16341
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.match_return_type
Fld_5 := Libadalang.Implementation.Base_Subp_Spec_P_Match_Return_Type (Node => Block_Var.Node, Other => Other, E_Info => Block_Var.Info);
--# end
--# expr-done 1438
   If_Result_3 := Fld_5;
else
   If_Result_3 := False;
end if;
if If_Result_3 then
   --# expr-start 1439 '<FieldAccess for match_formal_params at ast.py:16342>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16342
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Cast_Expr := Other; 
      Cast_Result_1 := Create_Internal_Entity_Base_Formal_Param_Holder
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
--# property-call-start BaseFormalParamHolder.match_formal_params
Fld_6 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Match_Formal_Params (Node => Block_Var.Node, Other => Cast_Result_1, Match_Names => Match_Name, Ignore_First_Param => Ignore_First_Param, E_Info => Block_Var.Info);
--# end
--# expr-done 1439
   If_Result_4 := Fld_6;
else
   If_Result_4 := False;
end if;
Let_Result := If_Result_4; 
--# end
--# expr-done 1440
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Match_Signature;
--# end
--# property-start BaseSubpSpec.match_other /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16346
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Match_Other
  (Node : Bare_Base_Subp_Spec
      ; Other : Internal_Entity_Base_Formal_Param_Holder
      ; Match_Names : Boolean
         := True
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Spec :=
     Internal_Entity_Base_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind other Other
   --# bind match_names Match_Names
   Property_Result : Boolean;
      Cast_Expr : Internal_Entity_Base_Formal_Param_Holder;
Cast_Result : Internal_Entity_Base_Subp_Spec;
Fld : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1443 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1442 '<FieldAccess for match_signature at ast.py:16348>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16348
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1441 '<Cast to BaseSubpSpec at ast.py:16349>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16349
Cast_Expr := Other; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Base_Subp_Spec
   then
      Cast_Result := Create_Internal_Entity_Base_Subp_Spec
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 1441
--# property-call-start BaseSubpSpec.match_signature
Fld := Libadalang.Implementation.Base_Subp_Spec_P_Match_Signature (Node => Ent.Node, Other => Cast_Result, Match_Name => Match_Names, E_Info => Ent.Info);
--# end
--# expr-done 1442
Let_Result := Fld; 
--# end
--# expr-done 1443
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Match_Other;
--# end
--# property-start BaseSubpSpec.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16354
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Defining_Env
  (Node : Bare_Base_Subp_Spec
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Spec :=
     Internal_Entity_Base_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1165 with Inline_Always;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Fld_2 : Internal_Entity_Type_Expr;
Is_Null : Boolean;
If_Result : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1165 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1165;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1449 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1448 '<If at ast.py:16358>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16358
--# expr-start 1445 '<IsNull at ast.py:16358>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16358
--# expr-start 1444 '<FieldAccess for returns at ast.py:16358>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16358
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.returns'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Returns (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1444
Is_Null := Fld_2.Node = null; 
--# expr-done 1445
if Is_Null then
   If_Result := Empty_Env;
else
   --# expr-start 1447 '<FieldAccess for defining_env at ast.py:16359>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16359
--# expr-start 1446 '<FieldAccess for return_type at ast.py:16359>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16359
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.return_type
Fld := Libadalang.Implementation.Base_Subp_Spec_P_Return_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 1446
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Fld.Node, Origin => Origin, Include_Ud_Indexing => False, E_Info => Fld.Info);
--# end
--# expr-done 1447
   If_Result := Fld_1;
end if;
      Inc_Ref (If_Result);
--# expr-done 1448
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 1449
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1165;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1165;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Defining_Env;
--# end
--# property-start BaseSubpSpec.potential_dottable_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16362
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Potential_Dottable_Type
  (Node : Bare_Base_Subp_Spec
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Spec :=
     Internal_Entity_Base_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1166 with Inline_Always;
      Var_Expr : Internal_Entity_Base_Formal_Param_Decl;
Var_Expr_1 : Internal_Entity_Type_Expr;
Fld : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Get_Result : Internal_Entity_Base_Formal_Param_Decl;
Fld_1 : Internal_Entity_Type_Expr;
Fld_2 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Result_Var_1 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1166 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1166;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1454 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1451 '<at at ast.py:16367>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16367
--# expr-start 1450 '<FieldAccess for abstract_formal_params at ast.py:16367>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16367
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseFormalParamHolder.abstract_formal_params'
Fld := Libadalang.Implementation.Dispatcher_Base_Formal_Param_Holder_P_Abstract_Formal_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1450
Get_Result := Get (Fld, 0, True); 
--# expr-done 1451
Var_Expr := Get_Result;
if Var_Expr /= No_Entity_Base_Formal_Param_Decl then
   --# scope-start
   --# bind p Var_Expr
--# expr-start 1452 '<FieldAccess for type_expression at ast.py:16368>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16368
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.type_expression'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Type_Expression (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 1452
Var_Expr_1 := Fld_1;
if Var_Expr_1 /= No_Entity_Type_Expr then
   --# scope-start
--# expr-start 1453 '<FieldAccess for element_type at ast.py:16368>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16368
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeExpr.element_type
Fld_2 := Libadalang.Implementation.Type_Expr_P_Element_Type (Node => Var_Expr_1.Node, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 1453
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Scope_Result_1 := Result_Var;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Base_Type_Decl;
end if;
Let_Result := Result_Var_1; 
--# end
--# expr-done 1454
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1166;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1166;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Potential_Dottable_Type;
--# end
--# property-start BaseSubpSpec.get_candidate_type_for_primitive /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16372
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Get_Candidate_Type_For_Primitive
  (Node : Bare_Base_Subp_Spec
      ; Type_Expr : Internal_Entity_Type_Expr
      ; Canonicalize : Boolean
         := True
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind type_expr Type_Expr
   --# bind canonicalize Canonicalize
   Property_Result : Internal_Entity_Base_Type_Decl;
      Block_Var : Bare_Declarative_Part;
Block_Var_1 : Internal_Entity_Base_Type_Decl;
Block_Var_2 : Internal_Entity_Base_Type_Decl;
Block_Var_3 : Internal_Entity_Base_Type_Decl;
Block_Var_4 : Bare_Declarative_Part;
Match_0 : Internal_Entity_Anonymous_Type;
Match_1 : Internal_Entity_Type_Expr;
Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Boolean;
Var_Expr_2 : Internal_Entity_Base_Type_Decl;
Var_Expr_3 : Internal_Entity_Incomplete_Type_Decl;
Var_Expr_4 : Internal_Entity_Base_Type_Decl;
Var_Expr_5 : Bare_Declarative_Part;
Var_Expr_6 : Bare_Declarative_Part;
Var_Expr_7 : Internal_Entity_Type_Decl;
Var_Expr_8 : Internal_Entity_Base_Type_Decl;
Var_Expr_9 : Bare_Declarative_Part;
Var_Expr_10 : Bare_Base_Package_Decl;
Var_Expr_11 : Bare_Declarative_Part;
Fld : Bare_Ada_Node;
Fld_1 : Bare_Ada_Node;
Fld_2 : Bare_Ada_Node;
Cast_Result : Bare_Declarative_Part;
Fld_3 : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_4 : Internal_Entity_Base_Type_Decl;
Fld_5 : Boolean;
Not_Val : Boolean;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Result_Var_1 : Internal_Entity_Base_Type_Decl;
Fld_6 : Internal_Entity_Base_Type_Decl;
Match_Prefix : Internal_Entity_Type_Expr;
Cast_Result_1 : Internal_Entity_Anonymous_Type;
Let_Result : Internal_Entity_Base_Type_Decl;
Scope_Result_2 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity_Type_Expr;
Let_Result_1 : Internal_Entity_Base_Type_Decl;
Scope_Result_3 : Internal_Entity_Base_Type_Decl;
Match_Result : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Base_Type_Decl;
Fld_7 : Bare_Ada_Node;
Bound_Origin_1 : Bare_Ada_Node;
Fld_8 : Internal_Entity_Base_Type_Decl;
Scope_Result_4 : Internal_Entity_Base_Type_Decl;
Result_Var_2 : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result_1 : Internal_Entity_Base_Type_Decl;
If_Result : Internal_Entity_Base_Type_Decl;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity_Incomplete_Type_Decl;
Fld_9 : Internal_Entity_Base_Type_Decl;
Scope_Result_5 : Internal_Entity_Base_Type_Decl;
Result_Var_3 : Internal_Entity_Base_Type_Decl;
Fld_10 : Bare_Base_Type_Decl;
Fld_11 : Bare_Ada_Node;
Fld_12 : Bare_Ada_Node;
Cast_Result_4 : Bare_Declarative_Part;
Scope_Result_6 : Bare_Declarative_Part;
Result_Var_4 : Bare_Declarative_Part;
Fld_13 : Bare_Ada_Node;
Scope_Result_7 : Bare_Ada_Node;
Result_Var_5 : Bare_Ada_Node;
Is_A : Boolean;
Fld_14 : Bare_Ada_Node;
Scope_Result_8 : Bare_Ada_Node;
Result_Var_6 : Bare_Ada_Node;
Is_A_1 : Boolean;
If_Result_1 : Boolean;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_5 : Internal_Entity_Type_Decl;
Fld_15 : Boolean;
Scope_Result_9 : Boolean;
Result_Var_7 : Boolean;
If_Result_2 : Boolean;
Fld_16 : Boolean;
Scope_Result_10 : Boolean;
Result_Var_8 : Boolean;
Not_Val_1 : Boolean;
If_Result_3 : Boolean;
Cast_Expr_2 : Bare_Declarative_Part;
Cast_Result_6 : Bare_Declarative_Part;
Is_Equal : Boolean;
Fld_17 : Bare_Ada_Node;
Scope_Result_11 : Bare_Ada_Node;
Result_Var_9 : Bare_Ada_Node;
Cast_Result_7 : Bare_Base_Package_Decl;
Fld_18 : Bare_Public_Part;
Scope_Result_12 : Bare_Public_Part;
Result_Var_10 : Bare_Public_Part;
Cast_Result_8 : Bare_Declarative_Part;
Is_Equal_1 : Boolean;
If_Result_4 : Boolean;
Scope_Result_13 : Boolean;
Result_Var_11 : Boolean;
If_Result_5 : Boolean;
If_Result_6 : Internal_Entity_Base_Type_Decl;
Let_Result_2 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1490 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 1458 '<Cast to DeclarativePart at ast.py:16383>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16383
--# expr-start 1457 '<parent at ast.py:16383>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16383
--# expr-start 1456 '<parent at ast.py:16383>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16383
--# expr-start 1455 '<parent at ast.py:16383>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16383
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 1455
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Parent; 
--# expr-done 1456
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Fld_1.Parent; 
--# expr-done 1457
   if Fld_2 = null
      or else Fld_2.Kind in Ada_Declarative_Part_Range
   then
      Cast_Result := Fld_2;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 1458
Block_Var := Cast_Result; 
--# bind decl_scope Block_Var
--# expr-start 1465 '<bind at ast.py:16385>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16385
--# expr-start 1459 '<FieldAccess for origin_node at ast.py:16385>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16385
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_3 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 1459
Bound_Origin := Fld_3; 
--# expr-start 1464 '<Match at ast.py:16385>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16385
      if Type_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Type_Expr; 
case Ada_Type_Expr (Match_Prefix.Node.Kind) is
         when Ada_Anonymous_Type =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Anonymous_Type
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_1; 
--# bind at Match_0
--# expr-start 1460 '<FieldAccess for element_type at ast.py:16386>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16386
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeExpr.element_type
Fld_4 := Libadalang.Implementation.Type_Expr_P_Element_Type (Node => Match_0.Node, Origin => Bound_Origin, E_Info => Match_0.Info);
--# end
--# expr-done 1460
Var_Expr := Fld_4;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind et Var_Expr
--# expr-start 1462 '<Not at ast.py:16389>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16389
--# expr-start 1461 '<FieldAccess for is_classwide at ast.py:16389>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16389
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_classwide'
Fld_5 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Classwide (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 1461
Not_Val := not (Fld_5); 
--# expr-done 1462
Var_Expr_1 := Not_Val;
if Var_Expr_1 /= False then
   --# scope-start
   --# bind _ Var_Expr_1
Scope_Result := Var_Expr;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Scope_Result_1 := Result_Var;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Base_Type_Decl;
end if;
Let_Result := Result_Var_1; 
--# end
Scope_Result_2 := Let_Result;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Enum_Lit_Synth_Type_Expr .. Ada_Synthetic_Type_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Type_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_2; 
--# bind other Match_1
--# expr-start 1463 '<FieldAccess for designated_type at ast.py:16393>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16393
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_6 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Match_1.Node, Origin => Bound_Origin, E_Info => Match_1.Info);
--# end
--# expr-done 1463
Let_Result_1 := Fld_6; 
--# end
Scope_Result_3 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_3; 
end case;
--# expr-done 1464
Dyn_Var_Bind_Result := Match_Result; 
--# expr-done 1465
Block_Var_1 := Dyn_Var_Bind_Result; 
--# bind typ Block_Var_1
--# expr-start 1469 '<If at ast.py:16397>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16397
if Canonicalize then
   --# expr-start 1468 '<bind at ast.py:16399>' Dyn_Var_Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16399
--# expr-start 1466 '<FieldAccess for origin_node at ast.py:16399>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16399
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_7 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 1466
Bound_Origin_1 := Fld_7; 
Var_Expr_2 := Block_Var_1;
if Var_Expr_2 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 1467 '<FieldAccess for canonical_type at ast.py:16399>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16399
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.canonical_type'
Fld_8 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Canonical_Type (Node => Var_Expr_2.Node, Origin => Bound_Origin_1, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 1467
Scope_Result_4 := Fld_8;
   --# end
   Result_Var_2 := Scope_Result_4;
else
   Result_Var_2 := No_Entity_Base_Type_Decl;
end if;
Dyn_Var_Bind_Result_1 := Result_Var_2; 
--# expr-done 1468
   If_Result := Dyn_Var_Bind_Result_1;
else
   If_Result := Block_Var_1;
end if;
--# expr-done 1469
Block_Var_2 := If_Result; 
--# bind canon_type Block_Var_2
--# expr-start 1470 '<Cast to IncompleteTypeDecl at ast.py:16403>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16403
Cast_Expr := Block_Var_2; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Incomplete_Type_Decl_Range
   then
      Cast_Result_3 := Create_Internal_Entity_Incomplete_Type_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result_3 := No_Entity_Incomplete_Type_Decl;
   end if;
--# expr-done 1470
Var_Expr_3 := Cast_Result_3;
if Var_Expr_3 /= No_Entity_Incomplete_Type_Decl then
   --# scope-start
   --# bind i Var_Expr_3
--# expr-start 1471 '<FieldAccess for next_part at ast.py:16404>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16404
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.next_part
Fld_9 := Libadalang.Implementation.Base_Type_Decl_P_Next_Part (Node => Var_Expr_3.Node, E_Info => Var_Expr_3.Info);
--# end
--# expr-done 1471
Scope_Result_5 := Fld_9;
   --# end
   Result_Var_3 := Scope_Result_5;
else
   Result_Var_3 := Block_Var_2;
end if;
Block_Var_3 := Result_Var_3; 
--# bind final_type Block_Var_3
Var_Expr_4 := Block_Var_3;
if Var_Expr_4 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind typ Var_Expr_4
--# expr-start 1475 '<Cast to DeclarativePart at ast.py:16409>' Cast_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16409
--# expr-start 1474 '<parent at ast.py:16409>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16409
--# expr-start 1473 '<parent at ast.py:16409>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16409
--# expr-start 1472 '<FieldAccess for node at ast.py:16409>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16409
Fld_10 := Var_Expr_4.Node;
--# expr-done 1472
      if Fld_10 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Fld_10.Parent; 
--# expr-done 1473
      if Fld_11 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Fld_11.Parent; 
--# expr-done 1474
   if Fld_12 = null
      or else Fld_12.Kind in Ada_Declarative_Part_Range
   then
      Cast_Result_4 := Fld_12;
   else
         Cast_Result_4 := No_Bare_Ada_Node;
   end if;
--# expr-done 1475
Scope_Result_6 := Cast_Result_4;
   --# end
   Result_Var_4 := Scope_Result_6;
else
   Result_Var_4 := No_Bare_Ada_Node;
end if;
Block_Var_4 := Result_Var_4; 
--# bind type_scope Block_Var_4
--# expr-start 1489 '<If at ast.py:16412>' If_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16412
--# expr-start 1477 '<IsA BasePackageDecl at ast.py:16418>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16418
Var_Expr_5 := Block_Var_4;
if Var_Expr_5 /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 1476 '<parent at ast.py:16418>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16418
      if Var_Expr_5 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Var_Expr_5.Parent; 
--# expr-done 1476
Scope_Result_7 := Fld_13;
   --# end
   Result_Var_5 := Scope_Result_7;
else
   Result_Var_5 := No_Bare_Ada_Node;
end if;
Is_A := Result_Var_5 /= null 
and then Result_Var_5.Kind in Ada_Base_Package_Decl; 
--# expr-done 1477
if Is_A then
   --# expr-start 1479 '<IsA BasePackageDecl at ast.py:16419>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16419
Var_Expr_6 := Block_Var;
if Var_Expr_6 /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 1478 '<parent at ast.py:16419>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16419
      if Var_Expr_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Var_Expr_6.Parent; 
--# expr-done 1478
Scope_Result_8 := Fld_14;
   --# end
   Result_Var_6 := Scope_Result_8;
else
   Result_Var_6 := No_Bare_Ada_Node;
end if;
Is_A_1 := Result_Var_6 /= null 
and then Result_Var_6.Kind in Ada_Base_Package_Decl; 
--# expr-done 1479
   If_Result_1 := Is_A_1;
else
   If_Result_1 := False;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
--# expr-start 1480 '<Cast to TypeDecl at ast.py:16434>' Cast_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16434
Cast_Expr_1 := Block_Var_3; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Type_Decl_Range
   then
      Cast_Result_5 := Create_Internal_Entity_Type_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_5 := No_Entity_Type_Decl;
   end if;
--# expr-done 1480
Var_Expr_7 := Cast_Result_5;
if Var_Expr_7 /= No_Entity_Type_Decl then
   --# scope-start
--# expr-start 1481 '<FieldAccess for is_derived_tagged_type at ast.py:16434>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16434
      if Var_Expr_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDecl.is_derived_tagged_type
Fld_15 := Libadalang.Implementation.Type_Decl_P_Is_Derived_Tagged_Type (Node => Var_Expr_7.Node, E_Info => Var_Expr_7.Info);
--# end
--# expr-done 1481
Scope_Result_9 := Fld_15;
   --# end
   Result_Var_7 := Scope_Result_9;
else
   Result_Var_7 := False;
end if;
   If_Result_2 := Result_Var_7;
end if;
if If_Result_2 then
   --# expr-start 1483 '<Not at ast.py:16438>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16438
Var_Expr_8 := Block_Var_3;
if Var_Expr_8 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 1482 '<FieldAccess for is_classwide at ast.py:16438>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16438
      if Var_Expr_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_classwide'
Fld_16 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Classwide (Node => Var_Expr_8.Node, E_Info => Var_Expr_8.Info);
--# end
--# expr-done 1482
Scope_Result_10 := Fld_16;
   --# end
   Result_Var_8 := Scope_Result_10;
else
   Result_Var_8 := False;
end if;
Not_Val_1 := not (Result_Var_8); 
--# expr-done 1483
   If_Result_3 := Not_Val_1;
else
   If_Result_3 := False;
end if;
if If_Result_3 then
Var_Expr_9 := Block_Var_4;
if Var_Expr_9 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind ds Var_Expr_9
--# expr-start 1484 '<Eq at ast.py:16443>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16443
Cast_Expr_2 := Block_Var; 
      Cast_Result_6 := Cast_Expr_2;
Is_Equal := Cast_Result_6 = Var_Expr_9; 
--# expr-done 1484
if Is_Equal then
   If_Result_4 := True;
else
   --# expr-start 1488 '<Eq at ast.py:16443>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16443
--# expr-start 1486 '<Cast to BasePackageDecl at ast.py:16445>' Cast_Result_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16445
Var_Expr_11 := Block_Var;
if Var_Expr_11 /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 1485 '<parent at ast.py:16445>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16445
      if Var_Expr_11 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_17 := Var_Expr_11.Parent; 
--# expr-done 1485
Scope_Result_11 := Fld_17;
   --# end
   Result_Var_9 := Scope_Result_11;
else
   Result_Var_9 := No_Bare_Ada_Node;
end if;
   if Result_Var_9 = null
      or else Result_Var_9.Kind in Ada_Base_Package_Decl
   then
      Cast_Result_7 := Result_Var_9;
   else
         Cast_Result_7 := No_Bare_Ada_Node;
   end if;
--# expr-done 1486
Var_Expr_10 := Cast_Result_7;
if Var_Expr_10 /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 1487 '<FieldAccess for public_part at ast.py:16445>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16445
      if Var_Expr_10 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Var_Expr_10.Base_Package_Decl_F_Public_Part;
--# expr-done 1487
Scope_Result_12 := Fld_18;
   --# end
   Result_Var_10 := Scope_Result_12;
else
   Result_Var_10 := No_Bare_Ada_Node;
end if;
      Cast_Result_8 := Result_Var_10;
Is_Equal_1 := Var_Expr_9 = Cast_Result_8; 
--# expr-done 1488
   If_Result_4 := Is_Equal_1;
end if;
Scope_Result_13 := If_Result_4;
   --# end
   Result_Var_11 := Scope_Result_13;
else
   Result_Var_11 := False;
end if;
   If_Result_5 := Result_Var_11;
else
   If_Result_5 := False;
end if;
if If_Result_5 then
   If_Result_6 := Block_Var_3;
else
   If_Result_6 := No_Entity_Base_Type_Decl;
end if;
--# expr-done 1489
Let_Result_2 := If_Result_6; 
--# end
--# expr-done 1490
      Property_Result := Let_Result_2;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Get_Candidate_Type_For_Primitive;
--# end
--# property-start BaseSubpSpec.candidate_primitive_subp_types /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16453
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Candidate_Primitive_Subp_Types
  (Node : Bare_Base_Subp_Spec
      ; Canonicalize : Boolean
         := True
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl_Array_Access
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Spec :=
     Internal_Entity_Base_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind canonicalize Canonicalize
   Property_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_1168 with Inline_Always;
            procedure Finalizer_Scope_1927 with Inline_Always;
      Block_Var : Internal_Entity_Defining_Name_Array_Access;
Block_Var_1 : Internal_Entity_Type_Expr_Array_Access;
Var_Expr : Internal_Entity_Type_Expr;
Var_Expr_1 : Internal_Entity_Incomplete_Type_Decl;
Fld : Internal_Entity_Defining_Name_Array_Access;
Fld_1 : Internal_Entity_Base_Formal_Param_Decl;
Fld_2 : Internal_Entity_Type_Expr;
Map_Result : Internal_Entity_Type_Expr_Array_Access;
Fld_3 : Internal_Entity_Type_Expr;
Singleton : Internal_Entity_Type_Expr_Array_Access;
Scope_Result : Internal_Entity_Type_Expr_Array_Access;
Result_Var : Internal_Entity_Type_Expr_Array_Access;
Concat_Result : Internal_Entity_Type_Expr_Array_Access;
Fld_4 : Internal_Entity_Base_Type_Decl;
Map_Result_1 : Internal_Entity_Base_Type_Decl_Array_Access;
Is_Null : Boolean;
Not_Val : Boolean;
Map_Result_2 : Internal_Entity_Base_Type_Decl_Array_Access;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity_Incomplete_Type_Decl;
Fld_5 : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Result_Var_1 : Internal_Entity_Base_Type_Decl;
Map_Result_3 : Internal_Entity_Base_Type_Decl_Array_Access;
Unique_Array : Internal_Entity_Base_Type_Decl_Array_Access;
Let_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_1168 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Map_Result_2);
                     Dec_Ref (Map_Result_3);
                     Dec_Ref (Unique_Array);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1168;
            procedure Finalizer_Scope_1927 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1927;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Subp_Spec_Base_Subp_Spec_P_Candidate_Primitive_Subp_Types,
            Items    => new Mmz_Key_Array (1 ..  3))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Subp_Spec,
                                As_Bare_Base_Subp_Spec => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Boolean,
                                      As_Boolean => Canonicalize);
               Mmz_K.Items (3) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Base_Type_Decl_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 1506 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1491 '<FieldAccess for unpacked_formal_params at ast.py:16462>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16462
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1491
Block_Var := Fld; Inc_Ref (Block_Var);
--# bind params Block_Var
--# expr-start 1496 '<Concat at ast.py:16463>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16463
--# expr-start 1494 '<Map at ast.py:16463>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16463
      declare
         Map_Result_Vec : Internal_Entity_Type_Expr_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Block_Var;
      begin
         for Item_64 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_64
      --# expr-start 1493 '<FieldAccess for type_expression at ast.py:16463>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16463
--# expr-start 1492 '<FieldAccess for formal_decl at ast.py:16463>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16463
      if Item_64.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_1 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Item_64.Node, E_Info => Item_64.Info);
--# end
--# expr-done 1492
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.type_expression'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Type_Expression (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 1493
         declare
            Item_To_Append : constant Internal_Entity_Type_Expr := Fld_2;
         begin
            Internal_Entity_Type_Expr_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Type_Expr_Array
           (Items_Count => Natural (Internal_Entity_Type_Expr_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Type_Expr_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Type_Expr_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Type_Expr_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1494
--# expr-start 1495 '<FieldAccess for returns at ast.py:16464>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16464
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.returns'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Returns (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1495
Var_Expr := Fld_3;
if Var_Expr /= No_Entity_Type_Expr then
   --# scope-start
                Singleton := Create_Internal_Entity_Type_Expr_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1927;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Type_Expr_Array_Type;
end if;
      Inc_Ref (Result_Var);
Concat_Result := Concat (Map_Result, Result_Var); 
--# expr-done 1496
Block_Var_1 := Concat_Result; Inc_Ref (Block_Var_1);
--# bind types Block_Var_1
--# expr-start 1505 '<unique at ast.py:16467>' Unique_Array /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16467
--# expr-start 1504 '<Map at ast.py:16467>' Map_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16467
   --# expr-start 1501 '<Filter at ast.py:16467>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16467
   --# expr-start 1498 '<Map at ast.py:16467>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16467
      declare
         Map_Result_1_Vec : Internal_Entity_Base_Type_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Type_Expr_Array_Access := Block_Var_1;
      begin
         for Item_67 of
               Collection.Items
         loop
   --# scope-start
               --# bind t Item_67
      --# expr-start 1497 '<FieldAccess for get_candidate_type_for_primitive at ast.py:16468>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16468
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.get_candidate_type_for_primitive
Fld_4 := Libadalang.Implementation.Base_Subp_Spec_P_Get_Candidate_Type_For_Primitive (Node => Ent.Node, Type_Expr => Item_67, Canonicalize => Canonicalize);
--# end
--# expr-done 1497
         declare
            Item_To_Append : constant Internal_Entity_Base_Type_Decl := Fld_4;
         begin
            Internal_Entity_Base_Type_Decl_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result_1 := Create_Internal_Entity_Base_Type_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Type_Decl_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Entity_Base_Type_Decl_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Entity_Base_Type_Decl_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Entity_Base_Type_Decl_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 1498
      declare
         Map_Result_2_Vec : Internal_Entity_Base_Type_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Base_Type_Decl_Array_Access := Map_Result_1;
      begin
         for Item_66 of
               Collection.Items
         loop
   --# scope-start
               --# bind t Item_66
         --# expr-start 1500 '<Not at ast.py:16472>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16472
--# expr-start 1499 '<IsNull at ast.py:16472>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16472
Is_Null := Item_66.Node = null; 
--# expr-done 1499
Not_Val := not (Is_Null); 
--# expr-done 1500
         if Not_Val then
         declare
            Item_To_Append : constant Internal_Entity_Base_Type_Decl := Item_66;
         begin
            Internal_Entity_Base_Type_Decl_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result_2 := Create_Internal_Entity_Base_Type_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Type_Decl_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Internal_Entity_Base_Type_Decl_Vectors.Get
              (Map_Result_2_Vec,
               I + Internal_Entity_Base_Type_Decl_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Internal_Entity_Base_Type_Decl_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 1501
      declare
         Map_Result_3_Vec : Internal_Entity_Base_Type_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Base_Type_Decl_Array_Access := Map_Result_2;
      begin
         for Item_65 of
               Collection.Items
         loop
   --# scope-start
               --# bind t Item_65
--# expr-start 1502 '<Cast to IncompleteTypeDecl at ast.py:16474>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16474
Cast_Expr := Item_65; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Incomplete_Type_Decl_Range
   then
      Cast_Result := Create_Internal_Entity_Incomplete_Type_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Incomplete_Type_Decl;
   end if;
--# expr-done 1502
Var_Expr_1 := Cast_Result;
if Var_Expr_1 /= No_Entity_Incomplete_Type_Decl then
   --# scope-start
   --# bind i Var_Expr_1
--# expr-start 1503 '<FieldAccess for next_part at ast.py:16475>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16475
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.next_part
Fld_5 := Libadalang.Implementation.Base_Type_Decl_P_Next_Part (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 1503
Scope_Result_1 := Fld_5;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := Item_65;
end if;
         declare
            Item_To_Append : constant Internal_Entity_Base_Type_Decl := Result_Var_1;
         begin
            Internal_Entity_Base_Type_Decl_Vectors.Append (Map_Result_3_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result_3 := Create_Internal_Entity_Base_Type_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Type_Decl_Vectors.Length (Map_Result_3_Vec)));
         for I in Map_Result_3.Items'Range loop
            Map_Result_3.Items (I) := Internal_Entity_Base_Type_Decl_Vectors.Get
              (Map_Result_3_Vec,
               I + Internal_Entity_Base_Type_Decl_Vectors.Index_Type'First - Map_Result_3.Items'First);
         end loop;
         Internal_Entity_Base_Type_Decl_Vectors.Destroy (Map_Result_3_Vec);
      end;
--# expr-done 1504
Unique_Array := Make_Unique (Map_Result_3); 
--# expr-done 1505
Let_Result := Unique_Array; Inc_Ref (Let_Result);
--# end
--# expr-done 1506
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1168;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Base_Type_Decl_Array_Access,
                        As_Internal_Entity_Base_Type_Decl_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1168;
               Finalizer_Scope_1927;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Candidate_Primitive_Subp_Types;
--# end
--# property-start BaseSubpSpec.candidate_primitive_subp_first_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16481
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Candidate_Primitive_Subp_First_Type
  (Node : Bare_Base_Subp_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Spec :=
     Internal_Entity_Base_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1169 with Inline_Always;
            procedure Finalizer_Scope_1932 with Inline_Always;
      Var_Expr : Internal_Entity_Base_Type_Decl_Array_Access;
Fld : Internal_Entity_Base_Type_Decl_Array_Access;
Get_Result : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1169 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1169;
            procedure Finalizer_Scope_1932 is
            begin
                     Dec_Ref (Var_Expr);
            end Finalizer_Scope_1932;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1509 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1507 '<FieldAccess for candidate_primitive_subp_types at ast.py:16486>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16486
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.candidate_primitive_subp_types
Fld := Libadalang.Implementation.Base_Subp_Spec_P_Candidate_Primitive_Subp_Types (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1507
Var_Expr := Fld;
      Inc_Ref (Var_Expr);
if Var_Expr /= No_Internal_Entity_Base_Type_Decl_Array_Type then
   --# scope-start
   --# bind p Var_Expr
--# expr-start 1508 '<at at ast.py:16486>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16486
Get_Result := Get (Var_Expr, 0, True); 
--# expr-done 1508
Scope_Result := Get_Result;
   --# end
      Finalizer_Scope_1932;
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 1509
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1169;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1169;
               Finalizer_Scope_1932;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Candidate_Primitive_Subp_First_Type;
--# end
--# property-start BaseSubpSpec.candidate_primitive_subp_tagged_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16489
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Candidate_Primitive_Subp_Tagged_Type
  (Node : Bare_Base_Subp_Spec
      ; Canonicalize : Boolean
         := True
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Spec :=
     Internal_Entity_Base_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind canonicalize Canonicalize
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1170 with Inline_Always;
      Cast_Expr : Bare_Base_Subp_Spec;
Cast_Result : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld : Internal_Entity_Base_Type_Decl_Array_Access;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Boolean;
Find_Result : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1170 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1170;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Subp_Spec_Base_Subp_Spec_P_Candidate_Primitive_Subp_Tagged_Type,
            Items    => new Mmz_Key_Array (1 ..  3))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Subp_Spec,
                                As_Bare_Base_Subp_Spec => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Boolean,
                                      As_Boolean => Canonicalize);
               Mmz_K.Items (3) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Base_Type_Decl;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 1515 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1514 '<bind at ast.py:16495>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16495
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Bound_Origin := Cast_Result; 
--# expr-start 1513 '<Find at ast.py:16496>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16496
   --# expr-start 1510 '<FieldAccess for candidate_primitive_subp_types at ast.py:16496>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16496
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.candidate_primitive_subp_types
Fld := Libadalang.Implementation.Base_Subp_Spec_P_Candidate_Primitive_Subp_Types (Node => Ent.Node, Canonicalize => Canonicalize, E_Info => Ent.Info);
--# end
--# expr-done 1510
      Find_Result := No_Entity_Base_Type_Decl;
      declare
         Collection : constant Internal_Entity_Base_Type_Decl_Array_Access := Fld;
      begin
         for Item_68 of
               Collection.Items
         loop
   --# scope-start
               --# bind t Item_68
      --# expr-start 1512 '<FieldAccess for is_tagged_type at ast.py:16497>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16497
--# expr-start 1511 '<FieldAccess for full_view at ast.py:16497>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16497
      if Item_68.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.full_view
Fld_1 := Libadalang.Implementation.Base_Type_Decl_P_Full_View (Node => Item_68.Node, E_Info => Item_68.Info);
--# end
--# expr-done 1511
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_tagged_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Tagged_Type (Node => Fld_1.Node, Origin => Bound_Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 1512
      if Fld_2 then
         Find_Result := Item_68;
      end if;
   --# end
               exit when Fld_2;
         end loop;
      end;
--# expr-done 1513
Dyn_Var_Bind_Result := Find_Result; 
--# expr-done 1514
Let_Result := Dyn_Var_Bind_Result; 
--# end
--# expr-done 1515
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1170;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Base_Type_Decl,
                        As_Internal_Entity_Base_Type_Decl => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1170;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Candidate_Primitive_Subp_Tagged_Type;
--# end
--# property-start BaseSubpSpec.decl_spec /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16503
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Decl_Spec
  (Node : Bare_Base_Subp_Spec
      ; Follow_Generic : Boolean
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Subp_Spec
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Spec :=
     Internal_Entity_Base_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind follow_generic Follow_Generic
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Base_Subp_Spec;
      Block_Var : Internal_Entity_Basic_Decl;
Var_Expr : Internal_Entity_Defining_Name;
Var_Expr_1 : Internal_Entity_Basic_Decl;
Var_Expr_2 : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity_Basic_Decl;
Scope_Result_1 : Internal_Entity_Basic_Decl;
Result_Var_1 : Internal_Entity_Basic_Decl;
Fld_3 : Internal_Entity_Base_Subp_Spec;
Not_Val : Boolean;
Is_A : Boolean;
If_Result : Boolean;
If_Result_1 : Internal_Entity_Base_Subp_Spec;
Scope_Result_2 : Internal_Entity_Base_Subp_Spec;
Result_Var_2 : Internal_Entity_Base_Subp_Spec;
Let_Result : Internal_Entity_Base_Subp_Spec;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1523 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1516 '<FieldAccess for name at ast.py:16515>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16515
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.name'
Fld := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Name (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1516
Var_Expr := Fld;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 1517 '<FieldAccess for basic_decl at ast.py:16515>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16515
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld_1 := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 1517
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Basic_Decl;
end if;
Block_Var := Result_Var; 
--# bind bd Block_Var
Var_Expr_2 := Block_Var;
if Var_Expr_2 /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 1518 '<FieldAccess for canonical_part at ast.py:16516>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16516
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.canonical_part
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Canonical_Part (Node => Var_Expr_2.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 1518
Scope_Result_1 := Fld_2;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Basic_Decl;
end if;
Var_Expr_1 := Result_Var_1;
if Var_Expr_1 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind dp Var_Expr_1
--# expr-start 1522 '<If at ast.py:16517>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16517
--# expr-start 1519 '<Not at ast.py:16518>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16518
Not_Val := not (Follow_Generic); 
--# expr-done 1519
if Not_Val then
   --# expr-start 1520 '<IsA GenericSubpInternal at ast.py:16518>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16518
Is_A := Var_Expr_1.Node /= null 
and then Var_Expr_1.Node.Kind in Ada_Generic_Subp_Internal_Range; 
--# expr-done 1520
   If_Result := Is_A;
else
   If_Result := False;
end if;
if If_Result then
   If_Result_1 := No_Entity_Base_Subp_Spec;
else
   --# expr-start 1521 '<FieldAccess for subp_spec_or_null at ast.py:16520>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16520
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_3 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Var_Expr_1.Node, Follow_Generic => Follow_Generic, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 1521
   If_Result_1 := Fld_3;
end if;
--# expr-done 1522
Scope_Result_2 := If_Result_1;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := Ent;
end if;
Let_Result := Result_Var_2; 
--# end
--# expr-done 1523
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Decl_Spec;
--# end
--# property-start BaseSubpSpec.primitive_decl_spec /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16527
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Primitive_Decl_Spec
  (Node : Bare_Base_Subp_Spec
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Subp_Spec
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Spec :=
     Internal_Entity_Base_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Base_Subp_Spec;
      Fld : Internal_Entity_Base_Subp_Spec;
Let_Result : Internal_Entity_Base_Subp_Spec;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1525 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1524 '<FieldAccess for decl_spec at ast.py:16533>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16533
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.decl_spec
Fld := Libadalang.Implementation.Base_Subp_Spec_P_Decl_Spec (Node => Ent.Node, Follow_Generic => False, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 1524
Let_Result := Fld; 
--# end
--# expr-done 1525
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Primitive_Decl_Spec;
--# end
--# property-start BaseSubpSpec.as_primitive_subp_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16536
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_As_Primitive_Subp_Type
  (Node : Bare_Base_Subp_Spec
      ; Typ : Internal_Entity_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Spec :=
     Internal_Entity_Base_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind typ Typ
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1937 with Inline_Always;
            procedure Finalizer_Scope_1938 with Inline_Always;
      Var_Expr : Internal_Entity_Defining_Name;
Var_Expr_1 : Boolean;
Var_Expr_2 : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Fld_2 : Symbol_Type;
Env_Get_Result : Internal_Entity_Array_Access;
Fld_3 : Bare_Ada_Node;
Fld_4 : Internal_Entity_Basic_Decl;
Fld_5 : Bare_Basic_Decl;
Cast_Result : Bare_Ada_Node;
Is_Equal : Boolean;
Quantifier_Result : Boolean;
Fld_6 : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Result_Var_1 : Internal_Entity_Base_Type_Decl;
Scope_Result_2 : Internal_Entity_Base_Type_Decl;
Result_Var_2 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1937 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Env_Get_Result);
            end Finalizer_Scope_1937;
            procedure Finalizer_Scope_1938 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1938;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1536 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1526 '<FieldAccess for name at ast.py:16545>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16545
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.name'
Fld := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Name (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1526
Var_Expr := Fld;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
   --# bind name Var_Expr
--# expr-start 1534 '<AnyQuantifier at ast.py:16549>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16549
   --# expr-start 1529 '<EnvGet at ast.py:16549>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16549
Var_Expr_2 := Typ;
if Var_Expr_2 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 1527 '<FieldAccess for primitives_env at ast.py:16549>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16549
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.primitives_env'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Primitives_Env (Node => Var_Expr_2.Node, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 1527
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1938;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
--# expr-start 1528 '<FieldAccess for name_symbol at ast.py:16549>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16549
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Var_Expr.Node);
--# end
--# expr-done 1528
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Result_Var, Key => Fld_2, Lookup_Kind => To_Lookup_Kind_Type (Recursive), Categories => (Inherited_Primitives => True, Nocat => True))); 
--# expr-done 1529
      Quantifier_Result := False;
      declare
         Collection : constant Internal_Entity_Array_Access := Env_Get_Result;
      begin
         for Item_69 of
               Collection.Items
         loop
   --# scope-start
               --# bind b Item_69
      --# expr-start 1533 '<Eq at ast.py:16550>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16550
--# expr-start 1530 '<FieldAccess for node at ast.py:16550>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16550
Fld_3 := Item_69.Node;
--# expr-done 1530
--# expr-start 1532 '<FieldAccess for node at ast.py:16550>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16550
--# expr-start 1531 '<FieldAccess for basic_decl at ast.py:16550>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16550
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld_4 := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 1531
Fld_5 := Fld_4.Node;
--# expr-done 1532
      Cast_Result := Fld_5;
Is_Equal := Fld_3 = Cast_Result; 
--# expr-done 1533
      Quantifier_Result := Is_Equal;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
--# expr-done 1534
Var_Expr_1 := Quantifier_Result;
if Var_Expr_1 /= False then
   --# scope-start
   --# bind _ Var_Expr_1
--# expr-start 1535 '<FieldAccess for real_type at ast.py:16555>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16555
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.real_type
Fld_6 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Real_Type (Node => Ent.Node, Typ => Typ, Origin => No_Bare_Ada_Node, E_Info => Ent.Info);
--# end
--# expr-done 1535
Scope_Result_1 := Fld_6;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Base_Type_Decl;
end if;
Scope_Result_2 := Result_Var_1;
   --# end
      Finalizer_Scope_1937;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Entity_Base_Type_Decl;
end if;
Let_Result := Result_Var_2; 
--# end
--# expr-done 1536
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1937;
               Finalizer_Scope_1938;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_As_Primitive_Subp_Type;
--# end
--# property-start BaseSubpSpec.primitive_subp_types /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16561
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Primitive_Subp_Types
  (Node : Bare_Base_Subp_Spec
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl_Array_Access
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Spec :=
     Internal_Entity_Base_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_1174 with Inline_Always;
            procedure Finalizer_Scope_1941 with Inline_Always;
      Var_Expr : Internal_Entity_Base_Subp_Spec;
Fld : Internal_Entity_Base_Subp_Spec;
Fld_1 : Internal_Entity_Base_Type_Decl_Array_Access;
Fld_2 : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
Not_Val : Boolean;
Map_Result : Internal_Entity_Base_Type_Decl_Array_Access;
Scope_Result : Internal_Entity_Base_Type_Decl_Array_Access;
Result_Var : Internal_Entity_Base_Type_Decl_Array_Access;
Let_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_1174 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1174;
            procedure Finalizer_Scope_1941 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_1941;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1543 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1537 '<FieldAccess for primitive_decl_spec at ast.py:16565>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16565
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.primitive_decl_spec
Fld := Libadalang.Implementation.Base_Subp_Spec_P_Primitive_Decl_Spec (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 1537
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Subp_Spec then
   --# scope-start
   --# bind spec Var_Expr
--# expr-start 1542 '<Filter at ast.py:16566>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16566
   --# expr-start 1538 '<FieldAccess for candidate_primitive_subp_types at ast.py:16566>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16566
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.candidate_primitive_subp_types
Fld_1 := Libadalang.Implementation.Base_Subp_Spec_P_Candidate_Primitive_Subp_Types (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 1538
      declare
         Map_Result_Vec : Internal_Entity_Base_Type_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Base_Type_Decl_Array_Access := Fld_1;
      begin
         for Item_70 of
               Collection.Items
         loop
   --# scope-start
               --# bind c Item_70
         --# expr-start 1541 '<Not at ast.py:16567>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16567
--# expr-start 1540 '<IsNull at ast.py:16567>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16567
--# expr-start 1539 '<FieldAccess for as_primitive_subp_type at ast.py:16567>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16567
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.as_primitive_subp_type
Fld_2 := Libadalang.Implementation.Base_Subp_Spec_P_As_Primitive_Subp_Type (Node => Var_Expr.Node, Typ => Item_70, E_Info => Var_Expr.Info);
--# end
--# expr-done 1539
Is_Null := Fld_2.Node = null; 
--# expr-done 1540
Not_Val := not (Is_Null); 
--# expr-done 1541
         if Not_Val then
         declare
            Item_To_Append : constant Internal_Entity_Base_Type_Decl := Item_70;
         begin
            Internal_Entity_Base_Type_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Base_Type_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Type_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Base_Type_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Base_Type_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Base_Type_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1542
Scope_Result := Map_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1941;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Base_Type_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 1543
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1174;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1174;
               Finalizer_Scope_1941;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Primitive_Subp_Types;
--# end
--# property-start BaseSubpSpec.primitive_subp_first_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16573
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Primitive_Subp_First_Type
  (Node : Bare_Base_Subp_Spec
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Spec :=
     Internal_Entity_Base_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Base_Subp_Spec;
Fld : Internal_Entity_Base_Subp_Spec;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1547 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1544 '<FieldAccess for primitive_decl_spec at ast.py:16577>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16577
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.primitive_decl_spec
Fld := Libadalang.Implementation.Base_Subp_Spec_P_Primitive_Decl_Spec (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 1544
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Subp_Spec then
   --# scope-start
   --# bind spec Var_Expr
--# expr-start 1546 '<FieldAccess for as_primitive_subp_type at ast.py:16578>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16578
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1545 '<FieldAccess for candidate_primitive_subp_first_type at ast.py:16579>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16579
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.candidate_primitive_subp_first_type
Fld_1 := Libadalang.Implementation.Base_Subp_Spec_P_Candidate_Primitive_Subp_First_Type (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 1545
--# property-call-start BaseSubpSpec.as_primitive_subp_type
Fld_2 := Libadalang.Implementation.Base_Subp_Spec_P_As_Primitive_Subp_Type (Node => Var_Expr.Node, Typ => Fld_1, E_Info => Var_Expr.Info);
--# end
--# expr-done 1546
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 1547
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Primitive_Subp_First_Type;
--# end
--# property-start BaseSubpSpec.primitive_subp_tagged_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16585
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Primitive_Subp_Tagged_Type
  (Node : Bare_Base_Subp_Spec
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Spec :=
     Internal_Entity_Base_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Base_Subp_Spec;
Fld : Internal_Entity_Base_Subp_Spec;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1551 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1548 '<FieldAccess for primitive_decl_spec at ast.py:16590>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16590
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.primitive_decl_spec
Fld := Libadalang.Implementation.Base_Subp_Spec_P_Primitive_Decl_Spec (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 1548
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Subp_Spec then
   --# scope-start
   --# bind spec Var_Expr
--# expr-start 1550 '<FieldAccess for as_primitive_subp_type at ast.py:16591>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16591
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1549 '<FieldAccess for candidate_primitive_subp_tagged_type at ast.py:16592>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16592
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.candidate_primitive_subp_tagged_type
Fld_1 := Libadalang.Implementation.Base_Subp_Spec_P_Candidate_Primitive_Subp_Tagged_Type (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 1549
--# property-call-start BaseSubpSpec.as_primitive_subp_type
Fld_2 := Libadalang.Implementation.Base_Subp_Spec_P_As_Primitive_Subp_Type (Node => Var_Expr.Node, Typ => Fld_1, E_Info => Var_Expr.Info);
--# end
--# expr-done 1550
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 1551
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Primitive_Subp_Tagged_Type;
--# end
--# property-start BaseSubpSpec.has_controlling_result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16598
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Has_Controlling_Result
  (Node : Bare_Base_Subp_Spec
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Spec :=
     Internal_Entity_Base_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Block_Var : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
Not_Val : Boolean;
Fld_1 : Internal_Entity_Base_Type_Decl;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Cast_Result_1 : Internal_Entity;
Is_Equiv : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1557 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1552 '<FieldAccess for candidate_primitive_subp_tagged_type at ast.py:16604>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16604
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.candidate_primitive_subp_tagged_type
Fld := Libadalang.Implementation.Base_Subp_Spec_P_Candidate_Primitive_Subp_Tagged_Type (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1552
Block_Var := Fld; 
--# bind typ Block_Var
--# expr-start 1554 '<Not at ast.py:16605>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16605
--# expr-start 1553 '<IsNull at ast.py:16605>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16605
Is_Null := Block_Var.Node = null; 
--# expr-done 1553
Not_Val := not (Is_Null); 
--# expr-done 1554
if Not_Val then
   --# expr-start 1556 '<Eq at ast.py:16605>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16605
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
--# expr-start 1555 '<FieldAccess for return_type at ast.py:16605>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16605
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.return_type
Fld_1 := Libadalang.Implementation.Base_Subp_Spec_P_Return_Type (Node => Ent.Node, Origin => No_Bare_Ada_Node, E_Info => Ent.Info);
--# end
--# expr-done 1555
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
Is_Equiv := Equivalent (Cast_Result, Cast_Result_1); 
--# expr-done 1556
   If_Result := Is_Equiv;
else
   If_Result := False;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 1557
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Has_Controlling_Result;
--# end
--# property-start BaseSubpSpec.dottable_subp_of /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16608
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Dottable_Subp_Of
  (Node : Bare_Base_Subp_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Spec :=
     Internal_Entity_Base_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_2 : Internal_Entity_Base_Type_Decl;
Is_A : Boolean;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity_Classwide_Type_Decl;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Internal_Entity_Base_Type_Decl;
Fld_5 : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
If_Result_1 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Fld_6 : Integer;
Comp_Result : Boolean;
If_Result_2 : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Subp_Spec_Base_Subp_Spec_P_Dottable_Subp_Of,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Subp_Spec,
                                As_Bare_Base_Subp_Spec => Self);
               Mmz_K.Items (2) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Base_Type_Decl;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 1570 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1569 '<bind at ast.py:16613>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16613
--# expr-start 1559 '<FieldAccess for origin_node at ast.py:16613>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16613
--# expr-start 1558 '<FieldAccess for name at ast.py:16613>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16613
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.name'
Fld := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Name (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1558
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_1 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Fld.Node);
--# end
--# expr-done 1559
Bound_Origin := Fld_1; 
--# expr-start 1568 '<If at ast.py:16613>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16613
--# expr-start 1560 '<FieldAccess for nb_max_params at ast.py:16614>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16614
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.nb_max_params
Fld_6 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Nb_Max_Params (Node => Ent.Node);
--# end
--# expr-done 1560
Comp_Result := Fld_6 > 0; 
if Comp_Result then
--# expr-start 1561 '<FieldAccess for potential_dottable_type at ast.py:16615>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16615
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.potential_dottable_type
Fld_2 := Libadalang.Implementation.Base_Subp_Spec_P_Potential_Dottable_Type (Node => Ent.Node, Origin => Bound_Origin, E_Info => Ent.Info);
--# end
--# expr-done 1561
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind t Var_Expr
--# expr-start 1567 '<Cond at ast.py:16615>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16615
--# expr-start 1562 '<IsA ClasswideTypeDecl at ast.py:16616>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16616
Is_A := Var_Expr.Node /= null 
and then Var_Expr.Node.Kind in Ada_Classwide_Type_Decl_Range; 
--# expr-done 1562
if Is_A then
   --# expr-start 1564 '<FieldAccess for typedecl at ast.py:16617>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16617
--# expr-start 1563 '<Cast to ClasswideTypeDecl at ast.py:16617>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16617
Cast_Expr := Var_Expr; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Classwide_Type_Decl_Range
   then
      Cast_Result := Create_Internal_Entity_Classwide_Type_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Classwide_Type_Decl;
   end if;
--# expr-done 1563
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ClasswideTypeDecl.typedecl
Fld_3 := Libadalang.Implementation.Classwide_Type_Decl_P_Typedecl (Node => Cast_Result.Node, E_Info => Cast_Result.Info);
--# end
--# expr-done 1564
   If_Result_1 := Fld_3;
else
--# expr-start 1566 '<FieldAccess for is_tagged_type at ast.py:16624>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16624
--# expr-start 1565 '<FieldAccess for full_view at ast.py:16624>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16624
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.full_view
Fld_4 := Libadalang.Implementation.Base_Type_Decl_P_Full_View (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 1565
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_tagged_type'
Fld_5 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Tagged_Type (Node => Fld_4.Node, Origin => Bound_Origin, E_Info => Fld_4.Info);
--# end
--# expr-done 1566
if Fld_5 then
   If_Result := Var_Expr;
else
   If_Result := No_Entity_Base_Type_Decl;
end if;
   If_Result_1 := If_Result;
end if;
--# expr-done 1567
Scope_Result := If_Result_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
   If_Result_2 := Result_Var;
else
   If_Result_2 := No_Entity_Base_Type_Decl;
end if;
--# expr-done 1568
Dyn_Var_Bind_Result := If_Result_2; 
--# expr-done 1569
Let_Result := Dyn_Var_Bind_Result; 
--# end
--# expr-done 1570
      Property_Result := Let_Result;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Base_Type_Decl,
                        As_Internal_Entity_Base_Type_Decl => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Dottable_Subp_Of;
--# end
--# property-start BaseSubpSpec.return_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16634
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Return_Type
  (Node : Bare_Base_Subp_Spec
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Spec :=
     Internal_Entity_Base_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Type_Expr;
Fld : Internal_Entity_Type_Expr;
Fld_1 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1573 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1571 '<FieldAccess for returns at ast.py:16639>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16639
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.returns'
Fld := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Returns (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1571
Var_Expr := Fld;
if Var_Expr /= No_Entity_Type_Expr then
   --# scope-start
   --# bind rt Var_Expr
--# expr-start 1572 '<FieldAccess for real_designated_type at ast.py:16640>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16640
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.real_designated_type
Fld_1 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Real_Designated_Type (Node => Ent.Node, Typ => Var_Expr, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 1572
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 1573
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Return_Type;
--# end
--# property-start BaseSubpSpec.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16643
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Xref_Entry_Point
  (Node : Bare_Base_Subp_Spec
  )
   return Boolean
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Xref_Entry_Point;
--# end
--# property-start BaseSubpSpec.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16644
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Spec_P_Xref_Equation
  (Node : Bare_Base_Subp_Spec
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Base_Subp_Spec := Bare_Base_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Spec :=
     Internal_Entity_Base_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1181 with Inline_Always;
            procedure Finalizer_Scope_1947 with Inline_Always;
      Var_Expr : Internal_Entity_Type_Expr;
Fld : Internal_Entity_Type_Expr;
Fld_1 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
            procedure Finalizer_Scope_1181 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1181;
            procedure Finalizer_Scope_1947 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1947;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
--# expr-start 1574 '<FieldAccess for returns at ast.py:16644>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16644
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.returns'
Fld := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Returns (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1574
Var_Expr := Fld;
if Var_Expr /= No_Entity_Type_Expr then
   --# scope-start
   --# bind r Var_Expr
--# expr-start 1575 '<FieldAccess for sub_equation at ast.py:16644>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16644
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 1575
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1947;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:16645") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1181;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1181;
               Finalizer_Scope_1947;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Spec_P_Xref_Equation;
--# end
 function Base_Subp_Spec_P_Params
  (Node : Bare_Base_Subp_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Param_Spec_Array_Access
is (raise Property_Error
    with "Property BaseSubpSpec.params not implemented on type "
    & Kind_Name (Node));
   --
   --  Primitives for Bare_Entry_Spec
   --
      procedure Initialize_Fields_For_Entry_Spec
        (Self : Bare_Entry_Spec
         ; Entry_Spec_F_Entry_Name : Bare_Defining_Name
         ; Entry_Spec_F_Family_Type : Bare_Ada_Node
         ; Entry_Spec_F_Entry_Params : Bare_Params
        ) is
      begin
            Self.Entry_Spec_F_Entry_Name := Entry_Spec_F_Entry_Name;
            Self.Entry_Spec_F_Family_Type := Entry_Spec_F_Family_Type;
            Self.Entry_Spec_F_Entry_Params := Entry_Spec_F_Entry_Params;
      end Initialize_Fields_For_Entry_Spec;
   function Entry_Spec_F_Entry_Name
     (Node : Bare_Entry_Spec) return Bare_Defining_Name
   is
   begin
         return Node.Entry_Spec_F_Entry_Name;
   end;
   function Entry_Spec_F_Family_Type
     (Node : Bare_Entry_Spec) return Bare_Ada_Node
   is
   begin
         return Node.Entry_Spec_F_Family_Type;
   end;
   function Entry_Spec_F_Entry_Params
     (Node : Bare_Entry_Spec) return Bare_Params
   is
   begin
         return Node.Entry_Spec_F_Entry_Params;
   end;
--# property-start EntrySpec.name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16911
pragma Warnings (Off, "is not referenced");
 function Entry_Spec_P_Name
  (Node : Bare_Entry_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Entry_Spec := Bare_Entry_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Spec :=
     Internal_Entity_Entry_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name;
      Fld : Internal_Entity_Defining_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1576 '<FieldAccess for entry_name at ast.py:16911>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16911
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Entry_Spec_F_Entry_Name, Info => Ent.Info);
--# expr-done 1576
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Spec_P_Name;
--# end
--# property-start EntrySpec.params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16912
pragma Warnings (Off, "is not referenced");
 function Entry_Spec_P_Params
  (Node : Bare_Entry_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Param_Spec_Array_Access
is
   Self : Bare_Entry_Spec := Bare_Entry_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Spec :=
     Internal_Entity_Entry_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Param_Spec_Array_Access;
            procedure Finalizer_Scope_1212 with Inline_Always;
            procedure Finalizer_Scope_1948 with Inline_Always;
      Var_Expr : Internal_Entity_Params;
Fld : Internal_Entity_Params;
Fld_1 : Internal_Entity_Param_Spec_List;
Fld_2 : Bare_Param_Spec_List;
Fld_3 : Internal_Entity_Info;
As_Entity : Internal_Entity_Param_Spec;
Is_Null : Boolean;
If_Result : Internal_Entity_Param_Spec;
Item_71 : Internal_Entity_Param_Spec;
Bare_Item_71 : Bare_Param_Spec;
Map_Result : Internal_Entity_Param_Spec_Array_Access;
Scope_Result : Internal_Entity_Param_Spec_Array_Access;
Result_Var : Internal_Entity_Param_Spec_Array_Access;
            procedure Finalizer_Scope_1212 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1212;
            procedure Finalizer_Scope_1948 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_1948;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 1577 '<FieldAccess for entry_params at ast.py:16913>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16913
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Params (Node => Ent.Node.Entry_Spec_F_Entry_Params, Info => Ent.Info);
--# expr-done 1577
Var_Expr := Fld;
if Var_Expr /= No_Entity_Params then
   --# scope-start
   --# bind p Var_Expr
--# expr-start 1579 '<Map at ast.py:16914>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16914
   --# expr-start 1578 '<FieldAccess for params at ast.py:16914>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16914
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Param_Spec_List (Node => Var_Expr.Node.Params_F_Params, Info => Var_Expr.Info);
--# expr-done 1578
Fld_2 := Fld_1.Node;
      declare
         Map_Result_Vec : Internal_Entity_Param_Spec_Vectors.Vector;
      begin
      if Fld_2 = null then
      Map_Result := Create_Internal_Entity_Param_Spec_Array (0);
      else
      declare
         Collection : constant Bare_Param_Spec_List := Fld_2;
      begin
         for Untyped_Item_71 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_71 := Untyped_Item_71; 
Is_Null := Bare_Item_71 = null; 
if Is_Null then
   If_Result := No_Entity_Param_Spec;
else
Fld_3 := Fld_1.Info;
As_Entity := (Info => Fld_3, Node => Bare_Item_71); 
   If_Result := As_Entity;
end if;
                  Item_71 := If_Result; 
   --# scope-start
               --# bind p Item_71
         declare
            Item_To_Append : constant Internal_Entity_Param_Spec := Item_71;
         begin
            Internal_Entity_Param_Spec_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Param_Spec_Array
           (Items_Count => Natural (Internal_Entity_Param_Spec_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Param_Spec_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Param_Spec_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Param_Spec_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1579
Scope_Result := Map_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1948;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Param_Spec_Array_Type;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1212;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1212;
               Finalizer_Scope_1948;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Spec_P_Params;
--# end
--# property-start EntrySpec.returns /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16918
pragma Warnings (Off, "is not referenced");
 function Entry_Spec_P_Returns
  (Node : Bare_Entry_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Expr
is
   Self : Bare_Entry_Spec := Bare_Entry_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Type_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := No_Entity_Type_Expr;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Spec_P_Returns;
--# end
--# property-start EntrySpec.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16920
pragma Warnings (Off, "is not referenced");
 function Entry_Spec_P_Xref_Equation
  (Node : Bare_Entry_Spec
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Entry_Spec := Bare_Entry_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Spec :=
     Internal_Entity_Entry_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1214 with Inline_Always;
            procedure Finalizer_Scope_1950 with Inline_Always;
      Var_Expr : Internal_Entity;
Fld : Internal_Entity;
Fld_1 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
            procedure Finalizer_Scope_1214 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1214;
            procedure Finalizer_Scope_1950 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1950;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
--# expr-start 1580 '<FieldAccess for family_type at ast.py:16920>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16920
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity (Node => Ent.Node.Entry_Spec_F_Family_Type, Info => Ent.Info);
--# expr-done 1580
Var_Expr := Fld;
if Var_Expr /= No_Entity then
   --# scope-start
   --# bind r Var_Expr
--# expr-start 1581 '<FieldAccess for sub_equation at ast.py:16921>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16921
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 1581
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1950;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:16922") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1214;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1214;
               Finalizer_Scope_1950;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Spec_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Enum_Subp_Spec
   --
--# property-start EnumSubpSpec.enum_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16656
pragma Warnings (Off, "is not referenced");
 function Enum_Subp_Spec_P_Enum_Decl
  (Node : Bare_Enum_Subp_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Enum_Literal_Decl
is
   Self : Bare_Enum_Subp_Spec := Bare_Enum_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Enum_Literal_Decl;
      Fld : Bare_Ada_Node;
Cast_Result : Bare_Enum_Literal_Decl;
As_Entity : Internal_Entity_Enum_Literal_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Enum_Literal_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1584 '<as_entity at ast.py:16656>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16656
--# expr-start 1583 '<Cast to EnumLiteralDecl at ast.py:16656>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16656
--# expr-start 1582 '<parent at ast.py:16656>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16656
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 1582
   if Fld = null
      or else Fld.Kind in Ada_Enum_Literal_Decl_Range
   then
      Cast_Result := Fld;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 1583
Is_Null := Cast_Result = null; 
if Is_Null then
   If_Result := No_Entity_Enum_Literal_Decl;
else
As_Entity := (Info => E_Info, Node => Cast_Result); 
   If_Result := As_Entity;
end if;
--# expr-done 1584
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Subp_Spec_P_Enum_Decl;
--# end
--# property-start EnumSubpSpec.name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16658
pragma Warnings (Off, "is not referenced");
 function Enum_Subp_Spec_P_Name
  (Node : Bare_Enum_Subp_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Enum_Subp_Spec := Bare_Enum_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Enum_Subp_Spec :=
     Internal_Entity_Enum_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name;
      Fld : Internal_Entity_Enum_Literal_Decl;
Fld_1 : Internal_Entity_Defining_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1586 '<FieldAccess for enum_decl_name at ast.py:16658>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16658
--# expr-start 1585 '<FieldAccess for enum_decl at ast.py:16658>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16658
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start EnumSubpSpec.enum_decl
Fld := Libadalang.Implementation.Enum_Subp_Spec_P_Enum_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1585
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]EnumLiteralDecl.enum_decl_name'
Fld_1 := Libadalang.Implementation.Dispatcher_Enum_Literal_Decl_P_Enum_Decl_Name (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 1586
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Subp_Spec_P_Name;
--# end
--# property-start EnumSubpSpec.returns /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16659
pragma Warnings (Off, "is not referenced");
 function Enum_Subp_Spec_P_Returns
  (Node : Bare_Enum_Subp_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Expr
is
   Self : Bare_Enum_Subp_Spec := Bare_Enum_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Enum_Subp_Spec :=
     Internal_Entity_Enum_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Type_Expr;
      Fld : Internal_Entity_Enum_Literal_Decl;
Fld_1 : Internal_Entity_Enum_Lit_Synth_Type_Expr;
Cast_Result : Internal_Entity_Type_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
--# expr-start 1588 '<FieldAccess for synth_type_expr at ast.py:16659>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16659
--# expr-start 1587 '<FieldAccess for enum_decl at ast.py:16659>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16659
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start EnumSubpSpec.enum_decl
Fld := Libadalang.Implementation.Enum_Subp_Spec_P_Enum_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1587
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start EnumLiteralDecl.synth_type_expr
Fld_1 := Libadalang.Implementation.Enum_Literal_Decl_P_Synth_Type_Expr (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 1588
      Cast_Result := Create_Internal_Entity_Type_Expr
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Subp_Spec_P_Returns;
--# end
--# property-start EnumSubpSpec.params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16660
pragma Warnings (Off, "is not referenced");
 function Enum_Subp_Spec_P_Params
  (Node : Bare_Enum_Subp_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Param_Spec_Array_Access
is
   Self : Bare_Enum_Subp_Spec := Bare_Enum_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Param_Spec_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := No_Internal_Entity_Param_Spec_Array_Type;
         Inc_Ref (Property_Result);
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Subp_Spec_P_Params;
--# end
   --
   --  Primitives for Bare_Predefined_Bin_Op_Spec
   --
      procedure Initialize_Fields_For_Predefined_Bin_Op_Spec
        (Self : Bare_Predefined_Bin_Op_Spec
         ; Predefined_Bin_Op_Spec_F_Left_Param : Bare_Synthetic_Formal_Param_Decl
         ; Predefined_Bin_Op_Spec_F_Right_Param : Bare_Synthetic_Formal_Param_Decl
         ; Predefined_Bin_Op_Spec_F_Return_Type_Expr : Bare_Type_Expr
        ) is
      begin
            Self.Predefined_Bin_Op_Spec_F_Left_Param := Predefined_Bin_Op_Spec_F_Left_Param;
            Self.Predefined_Bin_Op_Spec_F_Right_Param := Predefined_Bin_Op_Spec_F_Right_Param;
            Self.Predefined_Bin_Op_Spec_F_Return_Type_Expr := Predefined_Bin_Op_Spec_F_Return_Type_Expr;
      Self.Predefined_Bin_Op_Spec_Op_Symbol := null;
      end Initialize_Fields_For_Predefined_Bin_Op_Spec;
   function Predefined_Bin_Op_Spec_F_Left_Param
     (Node : Bare_Predefined_Bin_Op_Spec) return Bare_Synthetic_Formal_Param_Decl
   is
   begin
         return Node.Predefined_Bin_Op_Spec_F_Left_Param;
   end;
   function Predefined_Bin_Op_Spec_F_Right_Param
     (Node : Bare_Predefined_Bin_Op_Spec) return Bare_Synthetic_Formal_Param_Decl
   is
   begin
         return Node.Predefined_Bin_Op_Spec_F_Right_Param;
   end;
   function Predefined_Bin_Op_Spec_F_Return_Type_Expr
     (Node : Bare_Predefined_Bin_Op_Spec) return Bare_Type_Expr
   is
   begin
         return Node.Predefined_Bin_Op_Spec_F_Return_Type_Expr;
   end;
--# property-start PredefinedBinOpSpec.name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16772
pragma Warnings (Off, "is not referenced");
 function Predefined_Bin_Op_Spec_P_Name
  (Node : Bare_Predefined_Bin_Op_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Predefined_Bin_Op_Spec := Bare_Predefined_Bin_Op_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Defining_Name;
      Fld : Symbol_Type;
Fld_1 : Bare_Defining_Name;
As_Entity : Internal_Entity_Defining_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Defining_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1591 '<as_entity at ast.py:16772>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16772
--# expr-start 1590 '<FieldAccess for synthesize_defining_name at ast.py:16772>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16772
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1589 '<FieldAccess for op_symbol at ast.py:16772>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16772
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Predefined_Bin_Op_Spec_Op_Symbol;
--# expr-done 1589
--# property-call-start AdaNode.synthesize_defining_name
Fld_1 := Libadalang.Implementation.Ada_Node_P_Synthesize_Defining_Name (Node => Self, Sym => Fld);
--# end
--# expr-done 1590
Is_Null := Fld_1 = null; 
if Is_Null then
   If_Result := No_Entity_Defining_Name;
else
As_Entity := (Info => E_Info, Node => Fld_1); 
   If_Result := As_Entity;
end if;
--# expr-done 1591
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Predefined_Bin_Op_Spec_P_Name;
--# end
--# property-start PredefinedBinOpSpec.returns /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16773
pragma Warnings (Off, "is not referenced");
 function Predefined_Bin_Op_Spec_P_Returns
  (Node : Bare_Predefined_Bin_Op_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Expr
is
   Self : Bare_Predefined_Bin_Op_Spec := Bare_Predefined_Bin_Op_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Predefined_Bin_Op_Spec :=
     Internal_Entity_Predefined_Bin_Op_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Type_Expr;
      Fld : Internal_Entity_Type_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1592 '<FieldAccess for return_type_expr at ast.py:16773>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16773
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Expr (Node => Ent.Node.Predefined_Bin_Op_Spec_F_Return_Type_Expr, Info => Ent.Info);
--# expr-done 1592
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Predefined_Bin_Op_Spec_P_Returns;
--# end
--# property-start PredefinedBinOpSpec.abstract_formal_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16776
pragma Warnings (Off, "is not referenced");
 function Predefined_Bin_Op_Spec_P_Abstract_Formal_Params
  (Node : Bare_Predefined_Bin_Op_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Predefined_Bin_Op_Spec := Bare_Predefined_Bin_Op_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Predefined_Bin_Op_Spec :=
     Internal_Entity_Predefined_Bin_Op_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_1197 with Inline_Always;
      Fld : Internal_Entity_Synthetic_Formal_Param_Decl;
Cast_Result : Internal_Entity_Base_Formal_Param_Decl;
Fld_1 : Internal_Entity_Synthetic_Formal_Param_Decl;
Cast_Result_1 : Internal_Entity_Base_Formal_Param_Decl;
Array_Lit : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Let_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_1197 is
            begin
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1197;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1598 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1597 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 1594 '<Cast to BaseFormalParamDecl at ast.py:16778>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16778
--# expr-start 1593 '<FieldAccess for left_param at ast.py:16778>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16778
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Synthetic_Formal_Param_Decl (Node => Ent.Node.Predefined_Bin_Op_Spec_F_Left_Param, Info => Ent.Info);
--# expr-done 1593
      Cast_Result := Create_Internal_Entity_Base_Formal_Param_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
--# expr-done 1594
--# expr-start 1596 '<Cast to BaseFormalParamDecl at ast.py:16779>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16779
--# expr-start 1595 '<FieldAccess for right_param at ast.py:16779>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16779
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Synthetic_Formal_Param_Decl (Node => Ent.Node.Predefined_Bin_Op_Spec_F_Right_Param, Info => Ent.Info);
--# expr-done 1595
      Cast_Result_1 := Create_Internal_Entity_Base_Formal_Param_Decl
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
--# expr-done 1596
Array_Lit := Create_Internal_Entity_Base_Formal_Param_Decl_Array (Internal_Internal_Entity_Base_Formal_Param_Decl_Array'(1 => Cast_Result, 2 => Cast_Result_1)); 
--# expr-done 1597
Let_Result := Array_Lit; Inc_Ref (Let_Result);
--# end
--# expr-done 1598
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1197;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1197;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Predefined_Bin_Op_Spec_P_Abstract_Formal_Params;
--# end
   --
   --  Primitives for Bare_Predefined_Un_Op_Spec
   --
      procedure Initialize_Fields_For_Predefined_Un_Op_Spec
        (Self : Bare_Predefined_Un_Op_Spec
         ; Predefined_Un_Op_Spec_F_Right_Param : Bare_Synthetic_Formal_Param_Decl
         ; Predefined_Un_Op_Spec_F_Return_Type_Expr : Bare_Synthetic_Type_Expr
        ) is
      begin
            Self.Predefined_Un_Op_Spec_F_Right_Param := Predefined_Un_Op_Spec_F_Right_Param;
            Self.Predefined_Un_Op_Spec_F_Return_Type_Expr := Predefined_Un_Op_Spec_F_Return_Type_Expr;
      Self.Predefined_Un_Op_Spec_Op_Symbol := null;
      end Initialize_Fields_For_Predefined_Un_Op_Spec;
   function Predefined_Un_Op_Spec_F_Right_Param
     (Node : Bare_Predefined_Un_Op_Spec) return Bare_Synthetic_Formal_Param_Decl
   is
   begin
         return Node.Predefined_Un_Op_Spec_F_Right_Param;
   end;
   function Predefined_Un_Op_Spec_F_Return_Type_Expr
     (Node : Bare_Predefined_Un_Op_Spec) return Bare_Synthetic_Type_Expr
   is
   begin
         return Node.Predefined_Un_Op_Spec_F_Return_Type_Expr;
   end;
--# property-start PredefinedUnOpSpec.name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16754
pragma Warnings (Off, "is not referenced");
 function Predefined_Un_Op_Spec_P_Name
  (Node : Bare_Predefined_Un_Op_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Predefined_Un_Op_Spec := Bare_Predefined_Un_Op_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Defining_Name;
      Fld : Symbol_Type;
Fld_1 : Bare_Defining_Name;
As_Entity : Internal_Entity_Defining_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Defining_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1601 '<as_entity at ast.py:16754>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16754
--# expr-start 1600 '<FieldAccess for synthesize_defining_name at ast.py:16754>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16754
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1599 '<FieldAccess for op_symbol at ast.py:16754>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16754
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Predefined_Un_Op_Spec_Op_Symbol;
--# expr-done 1599
--# property-call-start AdaNode.synthesize_defining_name
Fld_1 := Libadalang.Implementation.Ada_Node_P_Synthesize_Defining_Name (Node => Self, Sym => Fld);
--# end
--# expr-done 1600
Is_Null := Fld_1 = null; 
if Is_Null then
   If_Result := No_Entity_Defining_Name;
else
As_Entity := (Info => E_Info, Node => Fld_1); 
   If_Result := As_Entity;
end if;
--# expr-done 1601
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Predefined_Un_Op_Spec_P_Name;
--# end
--# property-start PredefinedUnOpSpec.returns /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16755
pragma Warnings (Off, "is not referenced");
 function Predefined_Un_Op_Spec_P_Returns
  (Node : Bare_Predefined_Un_Op_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Expr
is
   Self : Bare_Predefined_Un_Op_Spec := Bare_Predefined_Un_Op_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Predefined_Un_Op_Spec :=
     Internal_Entity_Predefined_Un_Op_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Type_Expr;
      Fld : Internal_Entity_Synthetic_Type_Expr;
Cast_Result : Internal_Entity_Type_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 1602 '<FieldAccess for return_type_expr at ast.py:16755>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16755
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Synthetic_Type_Expr (Node => Ent.Node.Predefined_Un_Op_Spec_F_Return_Type_Expr, Info => Ent.Info);
--# expr-done 1602
      Cast_Result := Create_Internal_Entity_Type_Expr
        (Node => Fld.Node,
         Info => Fld.Info);
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Predefined_Un_Op_Spec_P_Returns;
--# end
--# property-start PredefinedUnOpSpec.abstract_formal_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16758
pragma Warnings (Off, "is not referenced");
 function Predefined_Un_Op_Spec_P_Abstract_Formal_Params
  (Node : Bare_Predefined_Un_Op_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Predefined_Un_Op_Spec := Bare_Predefined_Un_Op_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Predefined_Un_Op_Spec :=
     Internal_Entity_Predefined_Un_Op_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_1194 with Inline_Always;
      Fld : Internal_Entity_Synthetic_Formal_Param_Decl;
Cast_Result : Internal_Entity_Base_Formal_Param_Decl;
Array_Lit : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Let_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_1194 is
            begin
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1194;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1606 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1605 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 1604 '<Cast to BaseFormalParamDecl at ast.py:16759>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16759
--# expr-start 1603 '<FieldAccess for right_param at ast.py:16759>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16759
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Synthetic_Formal_Param_Decl (Node => Ent.Node.Predefined_Un_Op_Spec_F_Right_Param, Info => Ent.Info);
--# expr-done 1603
      Cast_Result := Create_Internal_Entity_Base_Formal_Param_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
--# expr-done 1604
Array_Lit := Create_Internal_Entity_Base_Formal_Param_Decl_Array (Internal_Internal_Entity_Base_Formal_Param_Decl_Array'(1 => Cast_Result)); 
--# expr-done 1605
Let_Result := Array_Lit; Inc_Ref (Let_Result);
--# end
--# expr-done 1606
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1194;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1194;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Predefined_Un_Op_Spec_P_Abstract_Formal_Params;
--# end
   --
   --  Primitives for Bare_Subp_Spec
   --
      procedure Initialize_Fields_For_Subp_Spec
        (Self : Bare_Subp_Spec
         ; Subp_Spec_F_Subp_Kind : Bare_Subp_Kind
         ; Subp_Spec_F_Subp_Name : Bare_Defining_Name
         ; Subp_Spec_F_Subp_Params : Bare_Params
         ; Subp_Spec_F_Subp_Returns : Bare_Type_Expr
        ) is
      begin
            Self.Subp_Spec_F_Subp_Kind := Subp_Spec_F_Subp_Kind;
            Self.Subp_Spec_F_Subp_Name := Subp_Spec_F_Subp_Name;
            Self.Subp_Spec_F_Subp_Params := Subp_Spec_F_Subp_Params;
            Self.Subp_Spec_F_Subp_Returns := Subp_Spec_F_Subp_Returns;
      end Initialize_Fields_For_Subp_Spec;
   function Subp_Spec_F_Subp_Kind
     (Node : Bare_Subp_Spec) return Bare_Subp_Kind
   is
   begin
         return Node.Subp_Spec_F_Subp_Kind;
   end;
   function Subp_Spec_F_Subp_Name
     (Node : Bare_Subp_Spec) return Bare_Defining_Name
   is
   begin
         return Node.Subp_Spec_F_Subp_Name;
   end;
   function Subp_Spec_F_Subp_Params
     (Node : Bare_Subp_Spec) return Bare_Params
   is
   begin
         return Node.Subp_Spec_F_Subp_Params;
   end;
   function Subp_Spec_F_Subp_Returns
     (Node : Bare_Subp_Spec) return Bare_Type_Expr
   is
   begin
         return Node.Subp_Spec_F_Subp_Returns;
   end;
--# property-start SubpSpec.name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16814
pragma Warnings (Off, "is not referenced");
 function Subp_Spec_P_Name
  (Node : Bare_Subp_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Subp_Spec := Bare_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Subp_Spec :=
     Internal_Entity_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name;
      Fld : Internal_Entity_Defining_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1607 '<FieldAccess for subp_name at ast.py:16814>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16814
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Subp_Spec_F_Subp_Name, Info => Ent.Info);
--# expr-done 1607
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subp_Spec_P_Name;
--# end
--# property-start SubpSpec.params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16815
pragma Warnings (Off, "is not referenced");
 function Subp_Spec_P_Params
  (Node : Bare_Subp_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Param_Spec_Array_Access
is
   Self : Bare_Subp_Spec := Bare_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Subp_Spec :=
     Internal_Entity_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Param_Spec_Array_Access;
            procedure Finalizer_Scope_1202 with Inline_Always;
      Var_Expr : Internal_Entity_Params;
Fld : Internal_Entity_Params;
Fld_1 : Internal_Entity_Param_Spec_List;
Scope_Result : Internal_Entity_Param_Spec_List;
Result_Var : Internal_Entity_Param_Spec_List;
Fld_2 : Bare_Param_Spec_List;
Fld_3 : Internal_Entity_Info;
As_Entity : Internal_Entity_Param_Spec;
Is_Null : Boolean;
If_Result : Internal_Entity_Param_Spec;
Item_72 : Internal_Entity_Param_Spec;
Bare_Item_72 : Bare_Param_Spec;
Map_Result : Internal_Entity_Param_Spec_Array_Access;
            procedure Finalizer_Scope_1202 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_1202;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1610 '<Map at ast.py:16815>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16815
--# expr-start 1608 '<FieldAccess for subp_params at ast.py:16815>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16815
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Params (Node => Ent.Node.Subp_Spec_F_Subp_Params, Info => Ent.Info);
--# expr-done 1608
Var_Expr := Fld;
if Var_Expr /= No_Entity_Params then
   --# scope-start
--# expr-start 1609 '<FieldAccess for params at ast.py:16815>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16815
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Param_Spec_List (Node => Var_Expr.Node.Params_F_Params, Info => Var_Expr.Info);
--# expr-done 1609
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Param_Spec_List;
end if;
Fld_2 := Result_Var.Node;
      declare
         Map_Result_Vec : Internal_Entity_Param_Spec_Vectors.Vector;
      begin
      if Fld_2 = null then
      Map_Result := Create_Internal_Entity_Param_Spec_Array (0);
      else
      declare
         Collection : constant Bare_Param_Spec_List := Fld_2;
      begin
         for Untyped_Item_72 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_72 := Untyped_Item_72; 
Is_Null := Bare_Item_72 = null; 
if Is_Null then
   If_Result := No_Entity_Param_Spec;
else
Fld_3 := Result_Var.Info;
As_Entity := (Info => Fld_3, Node => Bare_Item_72); 
   If_Result := As_Entity;
end if;
                  Item_72 := If_Result; 
   --# scope-start
               --# bind p Item_72
         declare
            Item_To_Append : constant Internal_Entity_Param_Spec := Item_72;
         begin
            Internal_Entity_Param_Spec_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Param_Spec_Array
           (Items_Count => Natural (Internal_Entity_Param_Spec_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Param_Spec_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Param_Spec_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Param_Spec_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1610
      Property_Result := Map_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1202;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1202;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subp_Spec_P_Params;
--# end
--# property-start SubpSpec.returns /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16817
pragma Warnings (Off, "is not referenced");
 function Subp_Spec_P_Returns
  (Node : Bare_Subp_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Expr
is
   Self : Bare_Subp_Spec := Bare_Subp_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Subp_Spec :=
     Internal_Entity_Subp_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Type_Expr;
      Fld : Internal_Entity_Type_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1611 '<FieldAccess for subp_returns at ast.py:16817>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16817
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Expr (Node => Ent.Node.Subp_Spec_F_Subp_Returns, Info => Ent.Info);
--# expr-done 1611
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subp_Spec_P_Returns;
--# end
   --
   --  Primitives for Bare_Component_List
   --
      procedure Initialize_Fields_For_Component_List
        (Self : Bare_Component_List
         ; Component_List_F_Components : Bare_Ada_Node_List
         ; Component_List_F_Variant_Part : Bare_Variant_Part
        ) is
      begin
            Self.Component_List_F_Components := Component_List_F_Components;
            Self.Component_List_F_Variant_Part := Component_List_F_Variant_Part;
      end Initialize_Fields_For_Component_List;
   function Component_List_F_Components
     (Node : Bare_Component_List) return Bare_Ada_Node_List
   is
   begin
         return Node.Component_List_F_Components;
   end;
   function Component_List_F_Variant_Part
     (Node : Bare_Component_List) return Bare_Variant_Part
   is
   begin
         return Node.Component_List_F_Variant_Part;
   end;
--# property-start ComponentList.type_def /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4712
pragma Warnings (Off, "is not referenced");
 function Component_List_P_Type_Def
  (Node : Bare_Component_List
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Def
is
   Self : Bare_Component_List := Bare_Component_List (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Type_Def;
      Fld : Bare_Ada_Node;
Fld_1 : Bare_Ada_Node;
Cast_Result : Bare_Type_Def;
As_Entity : Internal_Entity_Type_Def;
Is_Null : Boolean;
If_Result : Internal_Entity_Type_Def;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1615 '<as_entity at ast.py:4712>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4712
--# expr-start 1614 '<Cast to TypeDef at ast.py:4712>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4712
--# expr-start 1613 '<parent at ast.py:4712>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4712
--# expr-start 1612 '<parent at ast.py:4712>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4712
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 1612
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Parent; 
--# expr-done 1613
   if Fld_1 = null
      or else Fld_1.Kind in Ada_Type_Def
   then
      Cast_Result := Fld_1;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 1614
Is_Null := Cast_Result = null; 
if Is_Null then
   If_Result := No_Entity_Type_Def;
else
As_Entity := (Info => E_Info, Node => Cast_Result); 
   If_Result := As_Entity;
end if;
--# expr-done 1615
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_List_P_Type_Def;
--# end
--# property-start ComponentList.type_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4713
pragma Warnings (Off, "is not referenced");
 function Component_List_P_Type_Decl
  (Node : Bare_Component_List
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Decl
is
   Self : Bare_Component_List := Bare_Component_List (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Component_List :=
     Internal_Entity_Component_List'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Type_Decl;
      Fld : Internal_Entity_Type_Def;
Fld_1 : Internal_Entity;
Cast_Result : Internal_Entity_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1618 '<Cast to TypeDecl at ast.py:4713>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4713
--# expr-start 1617 '<parent at ast.py:4713>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4713
--# expr-start 1616 '<FieldAccess for type_def at ast.py:4713>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4713
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.type_def
Fld := Libadalang.Implementation.Component_List_P_Type_Def (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1616
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Parent (Node => Fld.Node, E_Info => Fld.Info);
--# expr-done 1617
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Type_Decl_Range
   then
      Cast_Result := Create_Internal_Entity_Type_Decl
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         Cast_Result := No_Entity_Type_Decl;
   end if;
--# expr-done 1618
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_List_P_Type_Decl;
--# end
--# property-start ComponentList.parent_component_list /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4715
pragma Warnings (Off, "is not referenced");
 function Component_List_P_Parent_Component_List
  (Node : Bare_Component_List
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Component_List
is
   Self : Bare_Component_List := Bare_Component_List (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Component_List :=
     Internal_Entity_Component_List'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Component_List;
      Var_Expr : Internal_Entity_Base_Record_Def;
Var_Expr_1 : Internal_Entity_Derived_Type_Def;
Cast_Expr : Bare_Component_List;
Cast_Result : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld : Internal_Entity_Type_Def;
Cast_Result_1 : Internal_Entity_Derived_Type_Def;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Base_Record_Def;
Scope_Result : Internal_Entity_Base_Record_Def;
Result_Var : Internal_Entity_Base_Record_Def;
Fld_3 : Internal_Entity_Component_List;
Scope_Result_1 : Internal_Entity_Component_List;
Result_Var_1 : Internal_Entity_Component_List;
Dyn_Var_Bind_Result : Internal_Entity_Component_List;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1624 '<bind at ast.py:4715>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4715
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Bound_Origin := Cast_Result; 
--# expr-start 1620 '<Cast to DerivedTypeDef at ast.py:4717>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4717
--# expr-start 1619 '<FieldAccess for type_def at ast.py:4717>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4717
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.type_def
Fld := Libadalang.Implementation.Component_List_P_Type_Def (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1619
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Derived_Type_Def_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Derived_Type_Def
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result_1 := No_Entity_Derived_Type_Def;
   end if;
--# expr-done 1620
Var_Expr_1 := Cast_Result_1;
if Var_Expr_1 /= No_Entity_Derived_Type_Def then
   --# scope-start
--# expr-start 1622 '<FieldAccess for record_def at ast.py:4717>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4717
--# expr-start 1621 '<FieldAccess for base_type at ast.py:4717>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4717
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => Var_Expr_1.Node, Origin => Bound_Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 1621
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.record_def'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Record_Def (Node => Fld_1.Node, Origin => Bound_Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 1622
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Record_Def;
end if;
Var_Expr := Result_Var;
if Var_Expr /= No_Entity_Base_Record_Def then
   --# scope-start
--# expr-start 1623 '<FieldAccess for comps at ast.py:4717>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4717
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseRecordDef.comps
Fld_3 := Libadalang.Implementation.Base_Record_Def_P_Comps (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 1623
Scope_Result_1 := Fld_3;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Component_List;
end if;
Dyn_Var_Bind_Result := Result_Var_1; 
--# expr-done 1624
      Property_Result := Dyn_Var_Bind_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_List_P_Parent_Component_List;
--# end
--# property-start ComponentList.abstract_formal_params_for_assocs /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4722
pragma Warnings (Off, "is not referenced");
 function Component_List_P_Abstract_Formal_Params_For_Assocs
  (Node : Bare_Component_List
      ; Assocs : Internal_Entity_Assoc_List
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Component_List := Bare_Component_List (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Component_List :=
     Internal_Entity_Component_List'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind assocs Assocs
   --# bind env Env
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_311 with Inline_Always;
      Block_Var : Internal_Entity_Type_Decl;
Block_Var_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Block_Var_2 : Internal_Param_Match_Array_Access;
Fld : Internal_Entity_Type_Decl;
Fld_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_2 : Internal_Param_Match_Array_Access;
Fld_3 : Internal_Entity_Defining_Name;
Is_Null : Boolean;
Not_Val : Boolean;
Fld_4 : Internal_Entity_Defining_Name;
Fld_5 : Internal_Entity_Base_Formal_Param_Decl;
Cast_Expr : Internal_Entity_Base_Formal_Param_Decl;
Cast_Result : Internal_Entity;
Cast_Result_1 : Internal_Entity;
Is_Equiv : Boolean;
Find_Result : Internal_Entity_Base_Formal_Param_Decl;
Is_Null_1 : Boolean;
Not_Val_1 : Boolean;
If_Result : Boolean;
Map_Result : Internal_Param_Match_Array_Access;
Fld_6 : Internal_Entity_Base_Record_Def;
Fld_7 : Internal_Entity_Component_List;
Fld_8 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Let_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_311 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Block_Var_2);
                     Dec_Ref (Fld_8);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_311;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1641 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1625 '<FieldAccess for type_decl at ast.py:4724>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4724
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.type_decl
Fld := Libadalang.Implementation.Component_List_P_Type_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1625
Block_Var := Fld; 
--# bind td Block_Var
--# expr-start 1626 '<FieldAccess for discriminants_list at ast.py:4725>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4725
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.discriminants_list'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Discriminants_List (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 1626
Block_Var_1 := Fld_1; Inc_Ref (Block_Var_1);
--# bind discriminants Block_Var_1
--# expr-start 1637 '<Filter at ast.py:4728>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4728
   --# expr-start 1627 '<FieldAccess for match_formals at ast.py:4728>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4728
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.match_formals
Fld_2 := Libadalang.Implementation.Ada_Node_P_Match_Formals (Node => Self, Formal_Params => Block_Var_1, Params => Assocs, Is_Dottable_Subp => False);
--# end
--# expr-done 1627
      declare
         Map_Result_Vec : Internal_Param_Match_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Param_Match_Array_Access := Fld_2;
      begin
         for Item_73 of
               Collection.Items
         loop
   --# scope-start
               --# bind pm Item_73
--# expr-start 1630 '<Not at ast.py:4732>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4732
--# expr-start 1629 '<IsNull at ast.py:4732>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4732
--# expr-start 1628 '<FieldAccess for formal at ast.py:4732>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4732
Fld_3 := Item_73.Formal;
--# expr-done 1628
Is_Null := Fld_3.Node = null; 
--# expr-done 1629
Not_Val := not (Is_Null); 
--# expr-done 1630
if Not_Val then
   --# expr-start 1636 '<Not at ast.py:4733>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4733
--# expr-start 1635 '<IsNull at ast.py:4733>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4733
--# expr-start 1634 '<Find at ast.py:4733>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4733
      Find_Result := No_Entity_Base_Formal_Param_Decl;
      declare
         Collection : constant Internal_Entity_Base_Formal_Param_Decl_Array_Access := Block_Var_1;
      begin
         for Item_74 of
               Collection.Items
         loop
   --# scope-start
               --# bind d Item_74
      --# expr-start 1633 '<Eq at ast.py:4734>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4734
Cast_Expr := Item_74; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
--# expr-start 1632 '<FieldAccess for formal_decl at ast.py:4734>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4734
--# expr-start 1631 '<FieldAccess for formal at ast.py:4734>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4734
Fld_4 := Item_73.Formal;
--# expr-done 1631
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_5 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Fld_4.Node, E_Info => Fld_4.Info);
--# end
--# expr-done 1632
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
Is_Equiv := Equivalent (Cast_Result, Cast_Result_1); 
--# expr-done 1633
      if Is_Equiv then
         Find_Result := Item_74;
      end if;
   --# end
               exit when Is_Equiv;
         end loop;
      end;
--# expr-done 1634
Is_Null_1 := Find_Result.Node = null; 
--# expr-done 1635
Not_Val_1 := not (Is_Null_1); 
--# expr-done 1636
   If_Result := Not_Val_1;
else
   If_Result := False;
end if;
         if If_Result then
         declare
            Item_To_Append : constant Internal_Param_Match := Item_73;
         begin
            Internal_Param_Match_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Param_Match_Array
           (Items_Count => Natural (Internal_Param_Match_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Param_Match_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Param_Match_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Param_Match_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1637
Block_Var_2 := Map_Result; Inc_Ref (Block_Var_2);
--# bind discriminants_matches Block_Var_2
--# expr-start 1640 '<FieldAccess for abstract_formal_params_impl at ast.py:4742>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4742
--# expr-start 1639 '<FieldAccess for comps at ast.py:4742>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4742
--# expr-start 1638 '<FieldAccess for record_def at ast.py:4742>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4742
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.record_def'
Fld_6 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Record_Def (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 1638
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseRecordDef.comps
Fld_7 := Libadalang.Implementation.Base_Record_Def_P_Comps (Node => Fld_6.Node, E_Info => Fld_6.Info);
--# end
--# expr-done 1639
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.abstract_formal_params_impl
Fld_8 := Libadalang.Implementation.Component_List_P_Abstract_Formal_Params_Impl (Node => Fld_7.Node, Discriminants => Block_Var_2, E_Info => Fld_7.Info);
--# end
--# expr-done 1640
Let_Result := Fld_8; Inc_Ref (Let_Result);
--# end
--# expr-done 1641
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_311;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_311;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_List_P_Abstract_Formal_Params_For_Assocs;
--# end
--# property-start ComponentList.abstract_formal_params_for_delta_assocs /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4747
pragma Warnings (Off, "is not referenced");
 function Component_List_P_Abstract_Formal_Params_For_Delta_Assocs
  (Node : Bare_Component_List
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Component_List := Bare_Component_List (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Component_List :=
     Internal_Entity_Component_List'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_312 with Inline_Always;
            procedure Finalizer_Scope_1958 with Inline_Always;
            procedure Finalizer_Scope_1960 with Inline_Always;
      Block_Var : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Block_Var_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Block_Var_2 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Var_Expr : Internal_Entity_Variant_Part;
Var_Expr_1 : Internal_Entity_Component_List;
Fld : Internal_Entity_Variant_Part;
Fld_1 : Internal_Entity_Variant_List;
Scope_Result : Internal_Entity_Variant_List;
Result_Var : Internal_Entity_Variant_List;
Fld_2 : Bare_Variant_List;
Fld_3 : Internal_Entity_Info;
As_Entity : Internal_Entity_Variant;
Is_Null : Boolean;
If_Result : Internal_Entity_Variant;
Fld_4 : Internal_Entity_Component_List;
Fld_5 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Item_75 : Internal_Entity_Variant;
Bare_Item_75 : Bare_Variant;
Map_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_6 : Internal_Entity_Ada_Node_List;
Fld_7 : Bare_Ada_Node_List;
Fld_8 : Internal_Entity_Info;
As_Entity_1 : Internal_Entity;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Base_Formal_Param_Decl;
Item_76 : Internal_Entity;
Bare_Item_76 : Bare_Ada_Node;
Is_A : Boolean;
Map_Result_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_9 : Internal_Entity_Component_List;
Fld_10 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Scope_Result_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Result_Var_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Concat_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Concat_Result_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Let_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_312 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Block_Var_2);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Concat_Result_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_312;
            procedure Finalizer_Scope_1958 is
            begin
                     Dec_Ref (Fld_5);
            end Finalizer_Scope_1958;
            procedure Finalizer_Scope_1960 is
            begin
                     Dec_Ref (Fld_10);
            end Finalizer_Scope_1960;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1655 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1646 '<Mapcat at ast.py:4754>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4754
--# expr-start 1642 '<FieldAccess for variant_part at ast.py:4754>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4754
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Variant_Part (Node => Ent.Node.Component_List_F_Variant_Part, Info => Ent.Info);
--# expr-done 1642
Var_Expr := Fld;
if Var_Expr /= No_Entity_Variant_Part then
   --# scope-start
--# expr-start 1643 '<FieldAccess for variant at ast.py:4754>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4754
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Variant_List (Node => Var_Expr.Node.Variant_Part_F_Variant, Info => Var_Expr.Info);
--# expr-done 1643
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Variant_List;
end if;
Fld_2 := Result_Var.Node;
      declare
         Map_Result_Vec : Internal_Entity_Base_Formal_Param_Decl_Vectors.Vector;
      begin
      if Fld_2 = null then
      Map_Result := Create_Internal_Entity_Base_Formal_Param_Decl_Array (0);
      else
      declare
         Collection : constant Bare_Variant_List := Fld_2;
      begin
         for Untyped_Item_75 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_75 := Untyped_Item_75; 
Is_Null := Bare_Item_75 = null; 
if Is_Null then
   If_Result := No_Entity_Variant;
else
Fld_3 := Result_Var.Info;
As_Entity := (Info => Fld_3, Node => Bare_Item_75); 
   If_Result := As_Entity;
end if;
                  Item_75 := If_Result; 
   --# scope-start
               --# bind v Item_75
      --# expr-start 1645 '<FieldAccess for abstract_formal_params_for_delta_assocs at ast.py:4755>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4755
--# expr-start 1644 '<FieldAccess for components at ast.py:4755>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4755
      if Item_75.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Component_List (Node => Item_75.Node.Variant_F_Components, Info => Item_75.Info);
--# expr-done 1644
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.abstract_formal_params_for_delta_assocs
Fld_5 := Libadalang.Implementation.Component_List_P_Abstract_Formal_Params_For_Delta_Assocs (Node => Fld_4.Node, E_Info => Fld_4.Info);
--# end
--# expr-done 1645
         for Item_To_Append of
               Fld_5.Items
         loop
            Internal_Entity_Base_Formal_Param_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_1958;
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Base_Formal_Param_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Formal_Param_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Base_Formal_Param_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Base_Formal_Param_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Base_Formal_Param_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1646
Block_Var := Map_Result; Inc_Ref (Block_Var);
--# bind variant_part_components Block_Var
--# expr-start 1650 '<FilterMap at ast.py:4757>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4757
   --# expr-start 1647 '<FieldAccess for components at ast.py:4757>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4757
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Ada_Node_List (Node => Ent.Node.Component_List_F_Components, Info => Ent.Info);
--# expr-done 1647
Fld_7 := Fld_6.Node;
      declare
         Map_Result_1_Vec : Internal_Entity_Base_Formal_Param_Decl_Vectors.Vector;
      begin
      if Fld_7 = null then
      Map_Result_1 := Create_Internal_Entity_Base_Formal_Param_Decl_Array (0);
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_7;
      begin
         for Untyped_Item_76 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_76 := Untyped_Item_76; 
Is_Null_1 := Bare_Item_76 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity;
else
Fld_8 := Fld_6.Info;
As_Entity_1 := (Info => Fld_8, Node => Bare_Item_76); 
   If_Result_1 := As_Entity_1;
end if;
                  Item_76 := If_Result_1; 
   --# scope-start
               --# bind e Item_76
         --# expr-start 1648 '<IsA BaseFormalParamDecl at ???>' Is_A None
Is_A := Item_76.Node /= null 
and then Item_76.Node.Kind in Ada_Base_Formal_Param_Decl; 
--# expr-done 1648
         if Is_A then
      --# expr-start 1649 '<Cast to BaseFormalParamDecl at ???>' Cast_Result None
Cast_Expr := Item_76; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Base_Formal_Param_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Formal_Param_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Base_Formal_Param_Decl;
   end if;
--# expr-done 1649
         declare
            Item_To_Append : constant Internal_Entity_Base_Formal_Param_Decl := Cast_Result;
         begin
            Internal_Entity_Base_Formal_Param_Decl_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
      end if;
         Map_Result_1 := Create_Internal_Entity_Base_Formal_Param_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Formal_Param_Decl_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Entity_Base_Formal_Param_Decl_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Entity_Base_Formal_Param_Decl_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Entity_Base_Formal_Param_Decl_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 1650
Block_Var_1 := Map_Result_1; Inc_Ref (Block_Var_1);
--# bind self_components Block_Var_1
--# expr-start 1651 '<FieldAccess for parent_component_list at ast.py:4758>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4758
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.parent_component_list
Fld_9 := Libadalang.Implementation.Component_List_P_Parent_Component_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1651
Var_Expr_1 := Fld_9;
if Var_Expr_1 /= No_Entity_Component_List then
   --# scope-start
--# expr-start 1652 '<FieldAccess for abstract_formal_params_for_delta_assocs at ast.py:4758>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4758
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.abstract_formal_params_for_delta_assocs
Fld_10 := Libadalang.Implementation.Component_List_P_Abstract_Formal_Params_For_Delta_Assocs (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 1652
Scope_Result_1 := Fld_10;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_1960;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Internal_Entity_Base_Formal_Param_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var_1);
Block_Var_2 := Result_Var_1; Inc_Ref (Block_Var_2);
--# bind parent_components Block_Var_2
--# expr-start 1654 '<Concat at ast.py:4760>' Concat_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4760
--# expr-start 1653 '<Concat at ast.py:4760>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4760
Concat_Result := Concat (Block_Var, Block_Var_1); 
--# expr-done 1653
Concat_Result_1 := Concat (Concat_Result, Block_Var_2); 
--# expr-done 1654
Let_Result := Concat_Result_1; Inc_Ref (Let_Result);
--# end
--# expr-done 1655
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_312;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_312;
               Finalizer_Scope_1958;
               Finalizer_Scope_1960;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_List_P_Abstract_Formal_Params_For_Delta_Assocs;
--# end
--# property-start ComponentList.abstract_formal_params_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4764
pragma Warnings (Off, "is not referenced");
 function Component_List_P_Abstract_Formal_Params_Impl
  (Node : Bare_Component_List
      ; Discriminants : Internal_Param_Match_Array_Access
      ; Include_Discriminants : Boolean
         := True
      ; Recurse : Boolean
         := True
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Component_List := Bare_Component_List (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Component_List :=
     Internal_Entity_Component_List'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind discriminants Discriminants
   --# bind include_discriminants Include_Discriminants
   --# bind recurse Recurse
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_313 with Inline_Always;
            procedure Finalizer_Scope_1962 with Inline_Always;
            procedure Finalizer_Scope_1963 with Inline_Always;
            procedure Finalizer_Scope_1965 with Inline_Always;
      Block_Var : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Block_Var_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Var_Expr : Internal_Entity_Variant_Part;
Var_Expr_1 : Internal_Entity_Component_List;
Var_Expr_2 : Internal_Entity_Composite_Constraint;
Var_Expr_3 : Internal_Entity_Type_Decl;
Fld : Internal_Entity_Ada_Node_List;
Fld_1 : Bare_Ada_Node_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Base_Formal_Param_Decl;
Item_77 : Internal_Entity;
Bare_Item_77 : Bare_Ada_Node;
Is_A : Boolean;
Map_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_3 : Internal_Entity_Variant_Part;
Fld_4 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Scope_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Result_Var : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Concat_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_5 : Internal_Entity_Component_List;
Fld_6 : Internal_Entity_Type_Decl;
Fld_7 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_8 : Internal_Entity_Type_Def;
Cast_Result_1 : Internal_Entity_Derived_Type_Def;
Fld_9 : Internal_Entity_Subtype_Indication;
Fld_10 : Internal_Entity_Constraint;
Cast_Result_2 : Internal_Entity_Composite_Constraint;
Fld_11 : Internal_Entity_Assoc_List;
Scope_Result_1 : Internal_Entity_Assoc_List;
Result_Var_1 : Internal_Entity_Assoc_List;
Fld_12 : Internal_Param_Match_Array_Access;
Fld_13 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Concat_Result_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Scope_Result_2 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Result_Var_2 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
If_Result_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_14 : Internal_Entity_Type_Decl;
Fld_15 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Scope_Result_3 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Result_Var_3 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Concat_Result_2 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
If_Result_2 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Let_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_313 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Result_Var_2);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Result_Var_3);
                     Dec_Ref (Concat_Result_2);
                     Dec_Ref (If_Result_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_313;
            procedure Finalizer_Scope_1962 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_1962;
            procedure Finalizer_Scope_1963 is
            begin
                     Dec_Ref (Fld_7);
                     Dec_Ref (Fld_12);
                     Dec_Ref (Fld_13);
                     Dec_Ref (Concat_Result_1);
            end Finalizer_Scope_1963;
            procedure Finalizer_Scope_1965 is
            begin
                     Dec_Ref (Fld_15);
            end Finalizer_Scope_1965;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1680 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1662 '<Concat at ast.py:4772>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4772
--# expr-start 1659 '<FilterMap at ast.py:4772>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4772
   --# expr-start 1656 '<FieldAccess for components at ast.py:4772>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4772
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Ada_Node_List (Node => Ent.Node.Component_List_F_Components, Info => Ent.Info);
--# expr-done 1656
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Internal_Entity_Base_Formal_Param_Decl_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Internal_Entity_Base_Formal_Param_Decl_Array (0);
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_1;
      begin
         for Untyped_Item_77 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_77 := Untyped_Item_77; 
Is_Null := Bare_Item_77 = null; 
if Is_Null then
   If_Result := No_Entity;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_77); 
   If_Result := As_Entity;
end if;
                  Item_77 := If_Result; 
   --# scope-start
               --# bind e Item_77
         --# expr-start 1657 '<IsA BaseFormalParamDecl at ???>' Is_A None
Is_A := Item_77.Node /= null 
and then Item_77.Node.Kind in Ada_Base_Formal_Param_Decl; 
--# expr-done 1657
         if Is_A then
      --# expr-start 1658 '<Cast to BaseFormalParamDecl at ???>' Cast_Result None
Cast_Expr := Item_77; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Base_Formal_Param_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Formal_Param_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Base_Formal_Param_Decl;
   end if;
--# expr-done 1658
         declare
            Item_To_Append : constant Internal_Entity_Base_Formal_Param_Decl := Cast_Result;
         begin
            Internal_Entity_Base_Formal_Param_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Base_Formal_Param_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Formal_Param_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Base_Formal_Param_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Base_Formal_Param_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Base_Formal_Param_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1659
--# expr-start 1660 '<FieldAccess for variant_part at ast.py:4773>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4773
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Variant_Part (Node => Ent.Node.Component_List_F_Variant_Part, Info => Ent.Info);
--# expr-done 1660
Var_Expr := Fld_3;
if Var_Expr /= No_Entity_Variant_Part then
   --# scope-start
--# expr-start 1661 '<FieldAccess for get_components at ast.py:4773>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4773
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start VariantPart.get_components
Fld_4 := Libadalang.Implementation.Variant_Part_P_Get_Components (Node => Var_Expr.Node, Discriminants => Discriminants, E_Info => Var_Expr.Info);
--# end
--# expr-done 1661
Scope_Result := Fld_4;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1962;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Base_Formal_Param_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var);
Concat_Result := Concat (Map_Result, Result_Var); 
--# expr-done 1662
Block_Var := Concat_Result; Inc_Ref (Block_Var);
--# bind self_comps Block_Var
--# expr-start 1675 '<If at ast.py:4780>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4780
if Recurse then
--# expr-start 1663 '<FieldAccess for parent_component_list at ast.py:4782>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4782
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.parent_component_list
Fld_5 := Libadalang.Implementation.Component_List_P_Parent_Component_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1663
Var_Expr_1 := Fld_5;
if Var_Expr_1 /= No_Entity_Component_List then
   --# scope-start
   --# bind pcl Var_Expr_1
--# expr-start 1674 '<Concat at ast.py:4783>' Concat_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4783
--# expr-start 1673 '<FieldAccess for abstract_formal_params_impl at ast.py:4783>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4783
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1672 '<FieldAccess for match_formals at ast.py:4784>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4784
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1665 '<FieldAccess for discriminants_list at ast.py:4785>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4785
--# expr-start 1664 '<FieldAccess for type_decl at ast.py:4785>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4785
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.type_decl
Fld_6 := Libadalang.Implementation.Component_List_P_Type_Decl (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 1664
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.discriminants_list'
Fld_7 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Discriminants_List (Node => Fld_6.Node, Origin => No_Bare_Ada_Node, E_Info => Fld_6.Info);
--# end
--# expr-done 1665
--# expr-start 1670 '<Cast to CompositeConstraint at ast.py:4786>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4786
--# expr-start 1669 '<FieldAccess for constraint at ast.py:4786>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4786
--# expr-start 1668 '<FieldAccess for subtype_indication at ast.py:4786>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4786
--# expr-start 1667 '<Cast to DerivedTypeDef at ast.py:4786>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4786
--# expr-start 1666 '<FieldAccess for type_def at ast.py:4786>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4786
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.type_def
Fld_8 := Libadalang.Implementation.Component_List_P_Type_Def (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1666
   if Fld_8.Node = null
      or else Fld_8.Node.Kind in Ada_Derived_Type_Def_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Derived_Type_Def
        (Node => Fld_8.Node,
         Info => Fld_8.Info);
   else
         Cast_Result_1 := No_Entity_Derived_Type_Def;
   end if;
--# expr-done 1667
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Subtype_Indication (Node => Cast_Result_1.Node.Derived_Type_Def_F_Subtype_Indication, Info => Cast_Result_1.Info);
--# expr-done 1668
      if Fld_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Create_Internal_Entity_Constraint (Node => Fld_9.Node.Subtype_Indication_F_Constraint, Info => Fld_9.Info);
--# expr-done 1669
   if Fld_10.Node = null
      or else Fld_10.Node.Kind in Ada_Composite_Constraint_Range
   then
      Cast_Result_2 := Create_Internal_Entity_Composite_Constraint
        (Node => Fld_10.Node,
         Info => Fld_10.Info);
   else
         Cast_Result_2 := No_Entity_Composite_Constraint;
   end if;
--# expr-done 1670
Var_Expr_2 := Cast_Result_2;
if Var_Expr_2 /= No_Entity_Composite_Constraint then
   --# scope-start
--# expr-start 1671 '<FieldAccess for constraints at ast.py:4786>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4786
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Create_Internal_Entity_Assoc_List (Node => Var_Expr_2.Node.Composite_Constraint_F_Constraints, Info => Var_Expr_2.Info);
--# expr-done 1671
Scope_Result_1 := Fld_11;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Assoc_List;
end if;
--# property-call-start AdaNode.match_formals
Fld_12 := Libadalang.Implementation.Ada_Node_P_Match_Formals (Node => Var_Expr_1.Node, Formal_Params => Fld_7, Params => Result_Var_1, Is_Dottable_Subp => False);
--# end
--# expr-done 1672
--# property-call-start ComponentList.abstract_formal_params_impl
Fld_13 := Libadalang.Implementation.Component_List_P_Abstract_Formal_Params_Impl (Node => Var_Expr_1.Node, Discriminants => Fld_12, Include_Discriminants => False, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 1673
Concat_Result_1 := Concat (Fld_13, Block_Var); 
--# expr-done 1674
Scope_Result_2 := Concat_Result_1;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_1963;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := Block_Var;
end if;
      Inc_Ref (Result_Var_2);
   If_Result_1 := Result_Var_2;
else
   If_Result_1 := Block_Var;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 1675
Block_Var_1 := If_Result_1; Inc_Ref (Block_Var_1);
--# bind ret Block_Var_1
--# expr-start 1679 '<If at ast.py:4798>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4798
if Include_Discriminants then
   --# expr-start 1678 '<Concat at ast.py:4800>' Concat_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4800
--# expr-start 1676 '<FieldAccess for type_decl at ast.py:4800>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4800
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.type_decl
Fld_14 := Libadalang.Implementation.Component_List_P_Type_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1676
Var_Expr_3 := Fld_14;
if Var_Expr_3 /= No_Entity_Type_Decl then
   --# scope-start
--# expr-start 1677 '<FieldAccess for discriminants_list at ast.py:4800>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4800
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.discriminants_list'
Fld_15 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Discriminants_List (Node => Var_Expr_3.Node, Origin => No_Bare_Ada_Node, E_Info => Var_Expr_3.Info);
--# end
--# expr-done 1677
Scope_Result_3 := Fld_15;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_1965;
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := No_Internal_Entity_Base_Formal_Param_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var_3);
Concat_Result_2 := Concat (Result_Var_3, Block_Var_1); 
--# expr-done 1678
   If_Result_2 := Concat_Result_2;
else
   If_Result_2 := Block_Var_1;
end if;
      Inc_Ref (If_Result_2);
--# expr-done 1679
Let_Result := If_Result_2; Inc_Ref (Let_Result);
--# end
--# expr-done 1680
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_313;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_313;
               Finalizer_Scope_1962;
               Finalizer_Scope_1963;
               Finalizer_Scope_1965;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_List_P_Abstract_Formal_Params_Impl;
--# end
--# property-start ComponentList.abstract_formal_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4805
pragma Warnings (Off, "is not referenced");
 function Component_List_P_Abstract_Formal_Params
  (Node : Bare_Component_List
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Component_List := Bare_Component_List (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Component_List :=
     Internal_Entity_Component_List'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_314 with Inline_Always;
      Fld : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Let_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_314 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_314;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1682 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1681 '<FieldAccess for abstract_formal_params_impl at ast.py:4806>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4806
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.abstract_formal_params_impl
Fld := Libadalang.Implementation.Component_List_P_Abstract_Formal_Params_Impl (Node => Ent.Node, Discriminants => No_Internal_Param_Match_Array_Type, E_Info => Ent.Info);
--# end
--# expr-done 1681
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 1682
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_314;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_314;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_List_P_Abstract_Formal_Params;
--# end
--# property-start ComponentList.shapes /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4809
pragma Warnings (Off, "is not referenced");
 function Component_List_P_Shapes
  (Node : Bare_Component_List
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Shape_Array_Access
is
   Self : Bare_Component_List := Bare_Component_List (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Component_List :=
     Internal_Entity_Component_List'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Shape_Array_Access;
            procedure Finalizer_Scope_315 with Inline_Always;
            procedure Finalizer_Scope_1967 with Inline_Always;
            procedure Finalizer_Scope_1968 with Inline_Always;
            procedure Finalizer_Scope_1969 with Inline_Always;
      Block_Var : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Var_Expr : Internal_Entity_Variant_Part;
Fld : Internal_Entity_Ada_Node_List;
Fld_1 : Bare_Ada_Node_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Base_Formal_Param_Decl;
Item_78 : Internal_Entity;
Bare_Item_78 : Bare_Ada_Node;
Is_A : Boolean;
Map_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_3 : Internal_Entity_Variant_Part;
Fld_4 : Internal_Entity_Variant_List;
Fld_5 : Bare_Variant_List;
Fld_6 : Internal_Entity_Info;
As_Entity_1 : Internal_Entity_Variant;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Variant;
Fld_7 : Internal_Entity_Component_List;
Fld_8 : Internal_Shape_Array_Access;
Fld_9 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Concat_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_10 : Internal_Entity_Identifier;
Fld_11 : Internal_Entity_Alternatives_List;
New_Struct : Internal_Discriminant_Values;
Singleton : Internal_Discriminant_Values_Array_Access;
Fld_12 : Internal_Discriminant_Values_Array_Access;
Concat_Result_1 : Internal_Discriminant_Values_Array_Access;
New_Struct_1 : Internal_Shape := No_Shape;
Map_Result_1 : Internal_Shape_Array_Access;
Item_79 : Internal_Entity_Variant;
Bare_Item_79 : Bare_Variant;
Map_Result_2 : Internal_Shape_Array_Access;
Scope_Result : Internal_Shape_Array_Access;
New_Struct_2 : Internal_Shape := No_Shape;
Singleton_1 : Internal_Shape_Array_Access;
Result_Var : Internal_Shape_Array_Access;
Let_Result : Internal_Shape_Array_Access;
            procedure Finalizer_Scope_315 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (New_Struct_2);
                     Dec_Ref (Singleton_1);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_315;
            procedure Finalizer_Scope_1967 is
            begin
                     Dec_Ref (Map_Result_2);
            end Finalizer_Scope_1967;
            procedure Finalizer_Scope_1968 is
            begin
                     Dec_Ref (Fld_8);
                     Dec_Ref (Map_Result_1);
            end Finalizer_Scope_1968;
            procedure Finalizer_Scope_1969 is
            begin
                     Dec_Ref (Fld_9);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Singleton);
                     Dec_Ref (Fld_12);
                     Dec_Ref (Concat_Result_1);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1969;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1702 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1686 '<FilterMap at ast.py:4813>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4813
   --# expr-start 1683 '<FieldAccess for components at ast.py:4813>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4813
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Ada_Node_List (Node => Ent.Node.Component_List_F_Components, Info => Ent.Info);
--# expr-done 1683
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Internal_Entity_Base_Formal_Param_Decl_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Internal_Entity_Base_Formal_Param_Decl_Array (0);
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_1;
      begin
         for Untyped_Item_78 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_78 := Untyped_Item_78; 
Is_Null := Bare_Item_78 = null; 
if Is_Null then
   If_Result := No_Entity;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_78); 
   If_Result := As_Entity;
end if;
                  Item_78 := If_Result; 
   --# scope-start
               --# bind e Item_78
         --# expr-start 1684 '<IsA BaseFormalParamDecl at ???>' Is_A None
Is_A := Item_78.Node /= null 
and then Item_78.Node.Kind in Ada_Base_Formal_Param_Decl; 
--# expr-done 1684
         if Is_A then
      --# expr-start 1685 '<Cast to BaseFormalParamDecl at ???>' Cast_Result None
Cast_Expr := Item_78; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Base_Formal_Param_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Formal_Param_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Base_Formal_Param_Decl;
   end if;
--# expr-done 1685
         declare
            Item_To_Append : constant Internal_Entity_Base_Formal_Param_Decl := Cast_Result;
         begin
            Internal_Entity_Base_Formal_Param_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Base_Formal_Param_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Formal_Param_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Base_Formal_Param_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Base_Formal_Param_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Base_Formal_Param_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1686
Block_Var := Map_Result; Inc_Ref (Block_Var);
--# bind self_comps Block_Var
--# expr-start 1687 '<FieldAccess for variant_part at ast.py:4814>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4814
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Variant_Part (Node => Ent.Node.Component_List_F_Variant_Part, Info => Ent.Info);
--# expr-done 1687
Var_Expr := Fld_3;
if Var_Expr /= No_Entity_Variant_Part then
   --# scope-start
   --# bind vpart Var_Expr
--# expr-start 1700 '<Mapcat at ast.py:4815>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4815
   --# expr-start 1688 '<FieldAccess for variant at ast.py:4815>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4815
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Variant_List (Node => Var_Expr.Node.Variant_Part_F_Variant, Info => Var_Expr.Info);
--# expr-done 1688
Fld_5 := Fld_4.Node;
      declare
         Map_Result_2_Vec : Internal_Shape_Vectors.Vector;
      begin
      if Fld_5 = null then
      Map_Result_2 := Create_Internal_Shape_Array (0);
      else
      declare
         Collection : constant Bare_Variant_List := Fld_5;
      begin
         for Untyped_Item_79 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_79 := Untyped_Item_79; 
Is_Null_1 := Bare_Item_79 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Variant;
else
Fld_6 := Fld_4.Info;
As_Entity_1 := (Info => Fld_6, Node => Bare_Item_79); 
   If_Result_1 := As_Entity_1;
end if;
                  Item_79 := If_Result_1; 
   --# scope-start
               --# bind v Item_79
      --# expr-start 1699 '<Map at ast.py:4816>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4816
   --# expr-start 1690 '<FieldAccess for shapes at ast.py:4816>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4816
--# expr-start 1689 '<FieldAccess for components at ast.py:4816>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4816
      if Item_79.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Component_List (Node => Item_79.Node.Variant_F_Components, Info => Item_79.Info);
--# expr-done 1689
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.shapes
Fld_8 := Libadalang.Implementation.Component_List_P_Shapes (Node => Fld_7.Node, E_Info => Fld_7.Info);
--# end
--# expr-done 1690
      declare
         Map_Result_1_Vec : Internal_Shape_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Shape_Array_Access := Fld_8;
      begin
         for Item_80 of
               Collection.Items
         loop
   --# scope-start
               --# bind s Item_80
      --# expr-start 1698 '<New Shape at ast.py:4817>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4817
--# expr-start 1692 '<Concat at ast.py:4818>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4818
--# expr-start 1691 '<FieldAccess for components at ast.py:4818>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4818
Fld_9 := Item_80.Components;
Inc_Ref (Fld_9);
--# expr-done 1691
Concat_Result := Concat (Block_Var, Fld_9); 
--# expr-done 1692
--# expr-start 1697 '<Concat at ast.py:4819>' Concat_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4819
--# expr-start 1695 '<New DiscriminantValues at ast.py:4819>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4819
--# expr-start 1693 '<FieldAccess for discr_name at ast.py:4820>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4820
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Create_Internal_Entity_Identifier (Node => Var_Expr.Node.Variant_Part_F_Discr_Name, Info => Var_Expr.Info);
--# expr-done 1693
--# expr-start 1694 '<FieldAccess for choices at ast.py:4821>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4821
      if Item_79.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Create_Internal_Entity_Alternatives_List (Node => Item_79.Node.Variant_F_Choices, Info => Item_79.Info);
--# expr-done 1694
New_Struct := (Discriminant => Fld_10, Values => Fld_11); 
--# expr-done 1695
                Singleton := Create_Internal_Discriminant_Values_Array (Items_Count => 1);
                Singleton.Items (1) := New_Struct;
--# expr-start 1696 '<FieldAccess for discriminants_values at ast.py:4822>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4822
Fld_12 := Item_80.Discriminants_Values;
Inc_Ref (Fld_12);
--# expr-done 1696
Concat_Result_1 := Concat (Singleton, Fld_12); 
--# expr-done 1697
Inc_Ref (Concat_Result);
Inc_Ref (Concat_Result_1);
New_Struct_1 := (Components => Concat_Result, Discriminants_Values => Concat_Result_1); 
--# expr-done 1698
         declare
            Item_To_Append : constant Internal_Shape := New_Struct_1;
         begin
               Inc_Ref (Item_To_Append);
            Internal_Shape_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_1969;
         end loop;
      end;
         Map_Result_1 := Create_Internal_Shape_Array
           (Items_Count => Natural (Internal_Shape_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Shape_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Shape_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Shape_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 1699
         for Item_To_Append of
               Map_Result_1.Items
         loop
               Inc_Ref (Item_To_Append);
            Internal_Shape_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_1968;
         end loop;
      end;
      end if;
         Map_Result_2 := Create_Internal_Shape_Array
           (Items_Count => Natural (Internal_Shape_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Internal_Shape_Vectors.Get
              (Map_Result_2_Vec,
               I + Internal_Shape_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Internal_Shape_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 1700
Scope_Result := Map_Result_2;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1967;
   Result_Var := Scope_Result;
else
   --# expr-start 1701 '<New Shape at ast.py:4826>' New_Struct_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4826
Inc_Ref (Block_Var);
Inc_Ref (No_Internal_Discriminant_Values_Array_Type);
New_Struct_2 := (Components => Block_Var, Discriminants_Values => No_Internal_Discriminant_Values_Array_Type); 
--# expr-done 1701
                Singleton_1 := Create_Internal_Shape_Array (Items_Count => 1);
                Singleton_1.Items (1) := New_Struct_2;
                Inc_Ref (Singleton_1.Items (1));
   Result_Var := Singleton_1;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 1702
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_315;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_315;
               Finalizer_Scope_1967;
               Finalizer_Scope_1968;
               Finalizer_Scope_1969;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_List_P_Shapes;
--# end
   --
   --  Primitives for Bare_Discriminant_Part
   --
--# property-start DiscriminantPart.abstract_formal_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4328
pragma Warnings (Off, "is not referenced");
 function Discriminant_Part_P_Abstract_Formal_Params
  (Node : Bare_Discriminant_Part
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Discriminant_Part := Bare_Discriminant_Part (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_269 with Inline_Always;
      Let_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_269 is
            begin
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_269;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1703 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := No_Internal_Entity_Base_Formal_Param_Decl_Array_Type; Inc_Ref (Let_Result);
--# end
--# expr-done 1703
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_269;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_269;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Discriminant_Part_P_Abstract_Formal_Params;
--# end
   --
   --  Primitives for Bare_Known_Discriminant_Part
   --
      procedure Initialize_Fields_For_Known_Discriminant_Part
        (Self : Bare_Known_Discriminant_Part
         ; Known_Discriminant_Part_F_Discr_Specs : Bare_Discriminant_Spec_List
        ) is
      begin
            Self.Known_Discriminant_Part_F_Discr_Specs := Known_Discriminant_Part_F_Discr_Specs;
      end Initialize_Fields_For_Known_Discriminant_Part;
   function Known_Discriminant_Part_F_Discr_Specs
     (Node : Bare_Known_Discriminant_Part) return Bare_Discriminant_Spec_List
   is
   begin
         return Node.Known_Discriminant_Part_F_Discr_Specs;
   end;
--# property-start KnownDiscriminantPart.abstract_formal_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4340
pragma Warnings (Off, "is not referenced");
 function Known_Discriminant_Part_P_Abstract_Formal_Params
  (Node : Bare_Known_Discriminant_Part
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Known_Discriminant_Part := Bare_Known_Discriminant_Part (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_270 with Inline_Always;
      Fld : Bare_Discriminant_Spec_List;
Cast_Expr : Bare_Discriminant_Spec;
Cast_Result : Bare_Base_Formal_Param_Decl;
As_Entity : Internal_Entity_Base_Formal_Param_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Base_Formal_Param_Decl;
Item_81 : Bare_Discriminant_Spec;
Map_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Let_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_270 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_270;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1708 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1707 '<Map at ast.py:4341>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4341
   --# expr-start 1704 '<FieldAccess for discr_specs at ast.py:4341>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4341
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Known_Discriminant_Part_F_Discr_Specs;
--# expr-done 1704
      declare
         Map_Result_Vec : Internal_Entity_Base_Formal_Param_Decl_Vectors.Vector;
      begin
      if Fld = null then
      Map_Result := Create_Internal_Entity_Base_Formal_Param_Decl_Array (0);
      else
      declare
         Collection : constant Bare_Discriminant_Spec_List := Fld;
      begin
         for Untyped_Item_81 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_81 := Untyped_Item_81; 
   --# scope-start
               --# bind e Item_81
      --# expr-start 1706 '<as_entity at ast.py:4342>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4342
--# expr-start 1705 '<Cast to BaseFormalParamDecl at ast.py:4342>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4342
Cast_Expr := Item_81; 
      Cast_Result := Cast_Expr;
--# expr-done 1705
Is_Null := Cast_Result = null; 
if Is_Null then
   If_Result := No_Entity_Base_Formal_Param_Decl;
else
As_Entity := (Info => E_Info, Node => Cast_Result); 
   If_Result := As_Entity;
end if;
--# expr-done 1706
         declare
            Item_To_Append : constant Internal_Entity_Base_Formal_Param_Decl := If_Result;
         begin
            Internal_Entity_Base_Formal_Param_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Base_Formal_Param_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Formal_Param_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Base_Formal_Param_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Base_Formal_Param_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Base_Formal_Param_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1707
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 1708
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_270;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_270;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Known_Discriminant_Part_P_Abstract_Formal_Params;
--# end
   --
   --  Primitives for Bare_Unknown_Discriminant_Part
   --
   --
   --  Primitives for Bare_Entry_Completion_Formal_Params
   --
      procedure Initialize_Fields_For_Entry_Completion_Formal_Params
        (Self : Bare_Entry_Completion_Formal_Params
         ; Entry_Completion_Formal_Params_F_Params : Bare_Params
        ) is
      begin
            Self.Entry_Completion_Formal_Params_F_Params := Entry_Completion_Formal_Params_F_Params;
      end Initialize_Fields_For_Entry_Completion_Formal_Params;
   function Entry_Completion_Formal_Params_F_Params
     (Node : Bare_Entry_Completion_Formal_Params) return Bare_Params
   is
   begin
         return Node.Entry_Completion_Formal_Params_F_Params;
   end;
--# property-start EntryCompletionFormalParams.abstract_formal_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19597
pragma Warnings (Off, "is not referenced");
 function Entry_Completion_Formal_Params_P_Abstract_Formal_Params
  (Node : Bare_Entry_Completion_Formal_Params
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Entry_Completion_Formal_Params := Bare_Entry_Completion_Formal_Params (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Completion_Formal_Params :=
     Internal_Entity_Entry_Completion_Formal_Params'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_1368 with Inline_Always;
      Var_Expr : Internal_Entity_Params;
Fld : Internal_Entity_Params;
Fld_1 : Internal_Entity_Param_Spec_List;
Scope_Result : Internal_Entity_Param_Spec_List;
Result_Var : Internal_Entity_Param_Spec_List;
Fld_2 : Bare_Param_Spec_List;
Fld_3 : Internal_Entity_Info;
As_Entity : Internal_Entity_Param_Spec;
Is_Null : Boolean;
If_Result : Internal_Entity_Param_Spec;
Cast_Expr : Internal_Entity_Param_Spec;
Cast_Result : Internal_Entity_Base_Formal_Param_Decl;
Item_82 : Internal_Entity_Param_Spec;
Bare_Item_82 : Bare_Param_Spec;
Map_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_1368 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_1368;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1712 '<Map at ast.py:19598>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19598
--# expr-start 1709 '<FieldAccess for params at ast.py:19598>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19598
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Params (Node => Ent.Node.Entry_Completion_Formal_Params_F_Params, Info => Ent.Info);
--# expr-done 1709
Var_Expr := Fld;
if Var_Expr /= No_Entity_Params then
   --# scope-start
--# expr-start 1710 '<FieldAccess for params at ast.py:19598>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19598
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Param_Spec_List (Node => Var_Expr.Node.Params_F_Params, Info => Var_Expr.Info);
--# expr-done 1710
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Param_Spec_List;
end if;
Fld_2 := Result_Var.Node;
      declare
         Map_Result_Vec : Internal_Entity_Base_Formal_Param_Decl_Vectors.Vector;
      begin
      if Fld_2 = null then
      Map_Result := Create_Internal_Entity_Base_Formal_Param_Decl_Array (0);
      else
      declare
         Collection : constant Bare_Param_Spec_List := Fld_2;
      begin
         for Untyped_Item_82 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_82 := Untyped_Item_82; 
Is_Null := Bare_Item_82 = null; 
if Is_Null then
   If_Result := No_Entity_Param_Spec;
else
Fld_3 := Result_Var.Info;
As_Entity := (Info => Fld_3, Node => Bare_Item_82); 
   If_Result := As_Entity;
end if;
                  Item_82 := If_Result; 
   --# scope-start
               --# bind p Item_82
      --# expr-start 1711 '<Cast to BaseFormalParamDecl at ast.py:19598>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19598
Cast_Expr := Item_82; 
      Cast_Result := Create_Internal_Entity_Base_Formal_Param_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
--# expr-done 1711
         declare
            Item_To_Append : constant Internal_Entity_Base_Formal_Param_Decl := Cast_Result;
         begin
            Internal_Entity_Base_Formal_Param_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Base_Formal_Param_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Formal_Param_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Base_Formal_Param_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Base_Formal_Param_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Base_Formal_Param_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1712
      Property_Result := Map_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1368;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1368;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Completion_Formal_Params_P_Abstract_Formal_Params;
--# end
   --
   --  Primitives for Bare_Generic_Formal_Part
   --
      procedure Initialize_Fields_For_Generic_Formal_Part
        (Self : Bare_Generic_Formal_Part
         ; Generic_Formal_Part_F_Decls : Bare_Ada_Node_List
        ) is
      begin
            Self.Generic_Formal_Part_F_Decls := Generic_Formal_Part_F_Decls;
      end Initialize_Fields_For_Generic_Formal_Part;
   function Generic_Formal_Part_F_Decls
     (Node : Bare_Generic_Formal_Part) return Bare_Ada_Node_List
   is
   begin
         return Node.Generic_Formal_Part_F_Decls;
   end;
--# property-start GenericFormalPart.abstract_formal_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10712
pragma Warnings (Off, "is not referenced");
 function Generic_Formal_Part_P_Abstract_Formal_Params
  (Node : Bare_Generic_Formal_Part
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Generic_Formal_Part := Bare_Generic_Formal_Part (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Formal_Part :=
     Internal_Entity_Generic_Formal_Part'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_840 with Inline_Always;
      Fld : Internal_Entity_Ada_Node_List;
Fld_1 : Bare_Ada_Node_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Base_Formal_Param_Decl;
Item_83 : Internal_Entity;
Bare_Item_83 : Bare_Ada_Node;
Is_A : Boolean;
Map_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_840 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_840;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1716 '<FilterMap at ast.py:10712>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10712
   --# expr-start 1713 '<FieldAccess for decls at ast.py:10712>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10712
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Ada_Node_List (Node => Ent.Node.Generic_Formal_Part_F_Decls, Info => Ent.Info);
--# expr-done 1713
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Internal_Entity_Base_Formal_Param_Decl_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Internal_Entity_Base_Formal_Param_Decl_Array (0);
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_1;
      begin
         for Untyped_Item_83 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_83 := Untyped_Item_83; 
Is_Null := Bare_Item_83 = null; 
if Is_Null then
   If_Result := No_Entity;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_83); 
   If_Result := As_Entity;
end if;
                  Item_83 := If_Result; 
   --# scope-start
               --# bind e Item_83
         --# expr-start 1714 '<IsA BaseFormalParamDecl at ???>' Is_A None
Is_A := Item_83.Node /= null 
and then Item_83.Node.Kind in Ada_Base_Formal_Param_Decl; 
--# expr-done 1714
         if Is_A then
      --# expr-start 1715 '<Cast to BaseFormalParamDecl at ???>' Cast_Result None
Cast_Expr := Item_83; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Base_Formal_Param_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Formal_Param_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Base_Formal_Param_Decl;
   end if;
--# expr-done 1715
         declare
            Item_To_Append : constant Internal_Entity_Base_Formal_Param_Decl := Cast_Result;
         begin
            Internal_Entity_Base_Formal_Param_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Base_Formal_Param_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Formal_Param_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Base_Formal_Param_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Base_Formal_Param_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Base_Formal_Param_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1716
      Property_Result := Map_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_840;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_840;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Formal_Part_P_Abstract_Formal_Params;
--# end
--# property-start GenericFormalPart.use_clauses_envs /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10715
pragma Warnings (Off, "is not referenced");
 function Generic_Formal_Part_P_Use_Clauses_Envs
  (Node : Bare_Generic_Formal_Part
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Generic_Formal_Part := Bare_Generic_Formal_Part (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Formal_Part :=
     Internal_Entity_Generic_Formal_Part'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_841 with Inline_Always;
            procedure Finalizer_Scope_1974 with Inline_Always;
      Fld : Internal_Entity_Ada_Node_List;
Fld_1 : Bare_Ada_Node_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Use_Clause;
Fld_3 : Lexical_Env := Empty_Env;
Item_84 : Internal_Entity;
Bare_Item_84 : Bare_Ada_Node;
Is_A : Boolean;
Map_Result : Lexical_Env_Array_Access;
Group_Env : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_841 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Group_Env);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_841;
            procedure Finalizer_Scope_1974 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_1974;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1723 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1722 '<env_group at ast.py:10720>' Group_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10720
--# expr-start 1721 '<FilterMap at ast.py:10720>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10720
   --# expr-start 1717 '<FieldAccess for decls at ast.py:10720>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10720
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Ada_Node_List (Node => Ent.Node.Generic_Formal_Part_F_Decls, Info => Ent.Info);
--# expr-done 1717
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Lexical_Env_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Lexical_Env_Array (0);
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_1;
      begin
         for Untyped_Item_84 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_84 := Untyped_Item_84; 
Is_Null := Bare_Item_84 = null; 
if Is_Null then
   If_Result := No_Entity;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_84); 
   If_Result := As_Entity;
end if;
                  Item_84 := If_Result; 
   --# scope-start
               --# bind u Item_84
         --# expr-start 1718 '<IsA UseClause at ast.py:10722>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10722
Is_A := Item_84.Node /= null 
and then Item_84.Node.Kind in Ada_Use_Clause; 
--# expr-done 1718
         if Is_A then
      --# expr-start 1720 '<FieldAccess for used_envs at ast.py:10721>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10721
--# expr-start 1719 '<Cast to UseClause at ast.py:10721>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10721
Cast_Expr := Item_84; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Use_Clause
   then
      Cast_Result := Create_Internal_Entity_Use_Clause
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Use_Clause;
   end if;
--# expr-done 1719
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start UseClause.used_envs
Fld_3 := Libadalang.Implementation.Use_Clause_P_Used_Envs (Node => Cast_Result.Node, E_Info => Cast_Result.Info);
--# end
--# expr-done 1720
         declare
            Item_To_Append : constant Lexical_Env := Fld_3;
         begin
               Inc_Ref (Item_To_Append);
            Lexical_Env_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
      Finalizer_Scope_1974;
         end loop;
      end;
      end if;
         Map_Result := Create_Lexical_Env_Array
           (Items_Count => Natural (Lexical_Env_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Lexical_Env_Vectors.Get
              (Map_Result_Vec,
               I + Lexical_Env_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Lexical_Env_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1721
Group_Env := Group (Map_Result, No_Metadata); 
--# expr-done 1722
Let_Result := Group_Env; Inc_Ref (Let_Result);
--# end
--# expr-done 1723
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_841;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_841;
               Finalizer_Scope_1974;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Formal_Part_P_Use_Clauses_Envs;
--# end
   --
   --  Primitives for Bare_Base_Record_Def
   --
      procedure Initialize_Fields_For_Base_Record_Def
        (Self : Bare_Base_Record_Def
         ; Base_Record_Def_F_Components : Bare_Component_List
        ) is
      begin
            Self.Base_Record_Def_F_Components := Base_Record_Def_F_Components;
      end Initialize_Fields_For_Base_Record_Def;
   function Base_Record_Def_F_Components
     (Node : Bare_Base_Record_Def) return Bare_Component_List
   is
   begin
         return Node.Base_Record_Def_F_Components;
   end;
--# property-start BaseRecordDef.comps /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4841
pragma Warnings (Off, "is not referenced");
 function Base_Record_Def_P_Comps
  (Node : Bare_Base_Record_Def
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Component_List
is
   Self : Bare_Base_Record_Def := Bare_Base_Record_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Record_Def :=
     Internal_Entity_Base_Record_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Component_List;
      Fld : Internal_Entity_Component_List;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1724 '<FieldAccess for components at ast.py:4841>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4841
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Component_List (Node => Ent.Node.Base_Record_Def_F_Components, Info => Ent.Info);
--# expr-done 1724
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Record_Def_P_Comps;
--# end
   --
   --  Primitives for Bare_Null_Record_Def
   --
      procedure Initialize_Fields_For_Null_Record_Def
        (Self : Bare_Null_Record_Def
         ; Base_Record_Def_F_Components : Bare_Component_List
        ) is
      begin
            Initialize_Fields_For_Base_Record_Def
              (Self, Base_Record_Def_F_Components);
      end Initialize_Fields_For_Null_Record_Def;
   --
   --  Primitives for Bare_Record_Def
   --
      procedure Initialize_Fields_For_Record_Def
        (Self : Bare_Record_Def
         ; Base_Record_Def_F_Components : Bare_Component_List
        ) is
      begin
            Initialize_Fields_For_Base_Record_Def
              (Self, Base_Record_Def_F_Components);
      end Initialize_Fields_For_Record_Def;
   --
   --  Primitives for Bare_Basic_Assoc
   --
--# property-start '[dispatcher]BasicAssoc.expr' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Assoc_P_Expr
  (Node : Bare_Basic_Assoc
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Expr
is
   Self : Bare_Basic_Assoc := Bare_Basic_Assoc (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Assoc (Self.Kind) is
               when Ada_Aggregate_Assoc .. Ada_Multi_Dim_Array_Assoc =>
                  --# property-call-start AggregateAssoc.expr
                  Property_Result := Aggregate_Assoc_P_Expr
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Composite_Constraint_Assoc =>
                  --# property-call-start CompositeConstraintAssoc.expr
                  Property_Result := Composite_Constraint_Assoc_P_Expr
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Iterated_Assoc =>
                  --# property-call-start IteratedAssoc.expr
                  Property_Result := Iterated_Assoc_P_Expr
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Param_Assoc =>
                  --# property-call-start ParamAssoc.expr
                  Property_Result := Param_Assoc_P_Expr
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Assoc_P_Expr;
--# end
--# property-start '[dispatcher]BasicAssoc.names' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Assoc_P_Names
  (Node : Bare_Basic_Assoc
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Basic_Assoc := Bare_Basic_Assoc (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Assoc (Self.Kind) is
               when Ada_Aggregate_Assoc .. Ada_Multi_Dim_Array_Assoc =>
                  --# property-call-start AggregateAssoc.names
                  Property_Result := Aggregate_Assoc_P_Names
                    (Self
                    );
                  --# end
               when Ada_Composite_Constraint_Assoc =>
                  --# property-call-start CompositeConstraintAssoc.names
                  Property_Result := Composite_Constraint_Assoc_P_Names
                    (Self
                    );
                  --# end
               when Ada_Iterated_Assoc =>
                  --# property-call-start IteratedAssoc.names
                  Property_Result := Iterated_Assoc_P_Names
                    (Self
                    );
                  --# end
               when Ada_Param_Assoc =>
                  --# property-call-start ParamAssoc.names
                  Property_Result := Param_Assoc_P_Names
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Assoc_P_Names;
--# end
--# property-start BasicAssoc.get_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7484
pragma Warnings (Off, "is not referenced");
 function Basic_Assoc_P_Get_Params
  (Node : Bare_Basic_Assoc
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Basic_Assoc := Bare_Basic_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Assoc :=
     Internal_Entity_Basic_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_554 with Inline_Always;
      Fld : Internal_Entity;
Cast_Result : Internal_Entity_Assoc_List;
Fld_1 : Internal_Param_Actual_Array_Access;
Fld_2 : Internal_Entity_Defining_Name;
Fld_3 : Internal_Entity_Expr;
Fld_4 : Internal_Entity_Expr;
Cast_Result_1 : Internal_Entity;
Cast_Result_2 : Internal_Entity;
Is_Equiv : Boolean;
Map_Result : Internal_Entity_Defining_Name_Array_Access;
Let_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_554 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_554;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1733 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1732 '<FilterMap at ast.py:7489>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7489
   --# expr-start 1727 '<FieldAccess for zip_with_params at ast.py:7489>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7489
--# expr-start 1726 '<Cast to AssocList at ast.py:7489>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7489
--# expr-start 1725 '<parent at ast.py:7489>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7489
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 1725
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Assoc_List_Range
   then
      Cast_Result := Create_Internal_Entity_Assoc_List
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 1726
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AssocList.zip_with_params
Fld_1 := Libadalang.Implementation.Assoc_List_P_Zip_With_Params (Node => Cast_Result.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Cast_Result.Info);
--# end
--# expr-done 1727
      declare
         Map_Result_Vec : Internal_Entity_Defining_Name_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Param_Actual_Array_Access := Fld_1;
      begin
         for Item_85 of
               Collection.Items
         loop
   --# scope-start
               --# bind m Item_85
         --# expr-start 1730 '<Eq at ast.py:7491>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7491
--# expr-start 1728 '<FieldAccess for actual at ast.py:7491>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7491
Fld_3 := Item_85.Actual;
--# expr-done 1728
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
--# expr-start 1729 '<FieldAccess for expr at ast.py:7491>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7491
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_4 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1729
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_4.Node,
         Info => Fld_4.Info);
Is_Equiv := Equivalent (Cast_Result_1, Cast_Result_2); 
--# expr-done 1730
         if Is_Equiv then
      --# expr-start 1731 '<FieldAccess for param at ast.py:7490>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7490
Fld_2 := Item_85.Param;
--# expr-done 1731
         declare
            Item_To_Append : constant Internal_Entity_Defining_Name := Fld_2;
         begin
            Internal_Entity_Defining_Name_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Defining_Name_Array
           (Items_Count => Natural (Internal_Entity_Defining_Name_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Defining_Name_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Defining_Name_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Defining_Name_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1732
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 1733
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_554;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_554;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Assoc_P_Get_Params;
--# end
   --
   --  Primitives for Bare_Aggregate_Assoc
   --
      procedure Initialize_Fields_For_Aggregate_Assoc
        (Self : Bare_Aggregate_Assoc
         ; Aggregate_Assoc_F_Designators : Bare_Alternatives_List
         ; Aggregate_Assoc_F_R_Expr : Bare_Expr
        ) is
      begin
            Self.Aggregate_Assoc_F_Designators := Aggregate_Assoc_F_Designators;
            Self.Aggregate_Assoc_F_R_Expr := Aggregate_Assoc_F_R_Expr;
      end Initialize_Fields_For_Aggregate_Assoc;
   function Aggregate_Assoc_F_Designators
     (Node : Bare_Aggregate_Assoc) return Bare_Alternatives_List
   is
   begin
         return Node.Aggregate_Assoc_F_Designators;
   end;
   function Aggregate_Assoc_F_R_Expr
     (Node : Bare_Aggregate_Assoc) return Bare_Expr
   is
   begin
         return Node.Aggregate_Assoc_F_R_Expr;
   end;
--# property-start AggregateAssoc.expr /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13965
pragma Warnings (Off, "is not referenced");
 function Aggregate_Assoc_P_Expr
  (Node : Bare_Aggregate_Assoc
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Expr
is
   Self : Bare_Aggregate_Assoc := Bare_Aggregate_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Aggregate_Assoc :=
     Internal_Entity_Aggregate_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Expr;
      Fld : Internal_Entity_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1734 '<FieldAccess for r_expr at ast.py:13965>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13965
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Aggregate_Assoc_F_R_Expr, Info => Ent.Info);
--# expr-done 1734
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aggregate_Assoc_P_Expr;
--# end
--# property-start AggregateAssoc.names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13966
pragma Warnings (Off, "is not referenced");
 function Aggregate_Assoc_P_Names
  (Node : Bare_Aggregate_Assoc
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Aggregate_Assoc := Bare_Aggregate_Assoc (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1018 with Inline_Always;
      Fld : Bare_Alternatives_List;
Item_86 : Bare_Ada_Node;
Map_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1018 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_1018;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1736 '<Map at ast.py:13966>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13966
   --# expr-start 1735 '<FieldAccess for designators at ast.py:13966>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13966
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Aggregate_Assoc_F_Designators;
--# expr-done 1735
      declare
         Map_Result_Vec : Bare_Ada_Node_Vectors.Vector;
      begin
      if Fld = null then
      Map_Result := Create_Bare_Ada_Node_Array (0);
      else
      declare
         Collection : constant Bare_Alternatives_List := Fld;
      begin
         for Untyped_Item_86 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_86 := Untyped_Item_86; 
   --# scope-start
               --# bind d Item_86
         declare
            Item_To_Append : constant Bare_Ada_Node := Item_86;
         begin
            Bare_Ada_Node_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Bare_Ada_Node_Array
           (Items_Count => Natural (Bare_Ada_Node_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Bare_Ada_Node_Vectors.Get
              (Map_Result_Vec,
               I + Bare_Ada_Node_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Bare_Ada_Node_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1736
      Property_Result := Map_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1018;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1018;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aggregate_Assoc_P_Names;
--# end
--# property-start AggregateAssoc.xref_stop_resolution /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13968
pragma Warnings (Off, "is not referenced");
 function Aggregate_Assoc_P_Xref_Stop_Resolution
  (Node : Bare_Aggregate_Assoc
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Aggregate_Assoc := Bare_Aggregate_Assoc (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aggregate_Assoc_P_Xref_Stop_Resolution;
--# end
--# property-start AggregateAssoc.base_aggregate /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13970
pragma Warnings (Off, "is not referenced");
 function Aggregate_Assoc_P_Base_Aggregate
  (Node : Bare_Aggregate_Assoc
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Aggregate
is
   Self : Bare_Aggregate_Assoc := Bare_Aggregate_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Aggregate_Assoc :=
     Internal_Entity_Aggregate_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Aggregate;
      Fld : Internal_Entity;
Fld_1 : Internal_Entity;
Cast_Result : Internal_Entity_Base_Aggregate;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1739 '<Cast to BaseAggregate at ast.py:13971>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13971
--# expr-start 1738 '<parent at ast.py:13971>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13971
--# expr-start 1737 '<parent at ast.py:13971>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13971
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 1737
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Parent (Node => Fld.Node, E_Info => Fld.Info);
--# expr-done 1738
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Base_Aggregate
   then
      Cast_Result := Create_Internal_Entity_Base_Aggregate
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 1739
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aggregate_Assoc_P_Base_Aggregate;
--# end
--# property-start AggregateAssoc.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13975
pragma Warnings (Off, "is not referenced");
 function Aggregate_Assoc_P_Xref_Equation
  (Node : Bare_Aggregate_Assoc
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Aggregate_Assoc := Bare_Aggregate_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Aggregate_Assoc :=
     Internal_Entity_Aggregate_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1021 with Inline_Always;
      Block_Var : Internal_Entity_Base_Aggregate;
Block_Var_1 : Internal_Multidim_Aggregate_Info;
Block_Var_2 : Internal_Entity_Base_Type_Decl;
Block_Var_3 : Internal_Entity_Array_Type_Def;
Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Base_Aggregate;
Fld_1 : Internal_Multidim_Aggregate_Info;
Fld_2 : Internal_Entity_Base_Type_Decl;
Fld_3 : Internal_Entity_Base_Aggregate;
Fld_4 : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
Is_Equal : Boolean;
Not_Val : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
Fld_5 : Internal_Entity_Array_Type_Def;
Scope_Result : Internal_Entity_Array_Type_Def;
Result_Var : Internal_Entity_Array_Type_Def;
Sym : Symbol_Type;
Fld_6 : Boolean;
Sym_1 : Symbol_Type;
Fld_7 : Boolean;
If_Result_1 : Boolean;
Fld_8 : Logic_Equation;
Sym_2 : Symbol_Type;
Fld_9 : Boolean;
Sym_3 : Symbol_Type;
Fld_10 : Boolean;
If_Result_2 : Boolean;
Fld_11 : Logic_Equation;
Sym_4 : Symbol_Type;
Fld_12 : Boolean;
Fld_13 : Logic_Equation;
Sym_5 : Symbol_Type;
Fld_14 : Boolean;
Fld_15 : Logic_Equation;
Sym_6 : Symbol_Type;
Fld_16 : Boolean;
Fld_17 : Logic_Equation;
Fld_18 : Internal_Entity;
Is_A : Boolean;
True_Rel : Logic_Equation;
Is_Null : Boolean;
Fld_19 : Logic_Equation;
Fld_20 : Logic_Equation;
If_Result_3 : Logic_Equation;
If_Result_4 : Logic_Equation;
If_Result_5 : Logic_Equation;
If_Result_6 : Logic_Equation;
If_Result_7 : Logic_Equation;
If_Result_8 : Logic_Equation;
If_Result_9 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1021 is
            begin
                     Dec_Ref (Fld_8);
                     Dec_Ref (Fld_11);
                     Dec_Ref (Fld_13);
                     Dec_Ref (Fld_15);
                     Dec_Ref (Fld_17);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Fld_19);
                     Dec_Ref (Fld_20);
                     Dec_Ref (If_Result_3);
                     Dec_Ref (If_Result_4);
                     Dec_Ref (If_Result_5);
                     Dec_Ref (If_Result_6);
                     Dec_Ref (If_Result_7);
                     Dec_Ref (If_Result_8);
                     Dec_Ref (If_Result_9);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1021;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1775 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1740 '<FieldAccess for base_aggregate at ast.py:13976>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13976
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AggregateAssoc.base_aggregate
Fld := Libadalang.Implementation.Aggregate_Assoc_P_Base_Aggregate (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1740
Block_Var := Fld; 
--# bind agg Block_Var
--# expr-start 1741 '<FieldAccess for multidim_root_aggregate at ast.py:13978>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13978
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseAggregate.multidim_root_aggregate
Fld_1 := Libadalang.Implementation.Base_Aggregate_P_Multidim_Root_Aggregate (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 1741
Block_Var_1 := Fld_1; 
--# bind mra Block_Var_1
--# expr-start 1748 '<If at ast.py:13983>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13983
--# expr-start 1743 '<Not at ast.py:13984>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13984
--# expr-start 1742 '<IsNull at ast.py:13984>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13984
Is_Equal := Block_Var_1 = No_Multidim_Aggregate_Info; 
--# expr-done 1742
Not_Val := not (Is_Equal); 
--# expr-done 1743
if Not_Val then
   --# expr-start 1744 '<FieldAccess for typ at ast.py:13985>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13985
Fld_2 := Block_Var_1.Typ;
--# expr-done 1744
   If_Result := Fld_2;
else
   --# expr-start 1747 '<Cast to BaseTypeDecl at ast.py:13986>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13986
--# expr-start 1746 '<FieldAccess for type_val at ast.py:13986>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13986
--# expr-start 1745 '<FieldAccess for base_aggregate at ast.py:13986>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13986
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AggregateAssoc.base_aggregate
Fld_3 := Libadalang.Implementation.Aggregate_Assoc_P_Base_Aggregate (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1745
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.type_val
Fld_4 := Libadalang.Implementation.Expr_P_Type_Val (Node => Fld_3.Node);
--# end
--# expr-done 1746
   if Fld_4.Node = null
      or else Fld_4.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld_4.Node,
         Info => Fld_4.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 1747
   If_Result := Cast_Result;
end if;
--# expr-done 1748
Block_Var_2 := If_Result; 
--# bind td Block_Var_2
Var_Expr := Block_Var_2;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 1749 '<FieldAccess for array_def at ast.py:13989>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13989
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.array_def'
Fld_5 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Array_Def (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 1749
Scope_Result := Fld_5;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Array_Type_Def;
end if;
Block_Var_3 := Result_Var; 
--# bind atd Block_Var_3
--# expr-start 1774 '<Cond at ast.py:13991>' If_Result_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13991
--# expr-start 1751 '<FieldAccess for in_aspect at ast.py:13992>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13992
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1750 '<Symbol Global at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Global); 
--# expr-done 1750
--# property-call-start AdaNode.in_aspect
Fld_6 := Libadalang.Implementation.Ada_Node_P_In_Aspect (Node => Block_Var.Node, Name => Sym);
--# end
--# expr-done 1751
if Fld_6 then
   If_Result_1 := True;
else
   --# expr-start 1753 '<FieldAccess for in_aspect at ast.py:13992>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13992
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1752 '<Symbol Refined_Global at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Refined_Global); 
--# expr-done 1752
--# property-call-start AdaNode.in_aspect
Fld_7 := Libadalang.Implementation.Ada_Node_P_In_Aspect (Node => Block_Var.Node, Name => Sym_1);
--# end
--# expr-done 1753
   If_Result_1 := Fld_7;
end if;
if If_Result_1 then
   --# expr-start 1754 '<FieldAccess for globals_assoc_equation at ast.py:13993>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13993
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AggregateAssoc.globals_assoc_equation
Fld_8 := Libadalang.Implementation.Aggregate_Assoc_P_Globals_Assoc_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 1754
   If_Result_9 := Fld_8;
else
--# expr-start 1756 '<FieldAccess for in_aspect at ast.py:13995>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13995
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1755 '<Symbol Depends at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Depends); 
--# expr-done 1755
--# property-call-start AdaNode.in_aspect
Fld_9 := Libadalang.Implementation.Ada_Node_P_In_Aspect (Node => Block_Var.Node, Name => Sym_2);
--# end
--# expr-done 1756
if Fld_9 then
   If_Result_2 := True;
else
   --# expr-start 1758 '<FieldAccess for in_aspect at ast.py:13995>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13995
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1757 '<Symbol Refined_Depends at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Refined_Depends); 
--# expr-done 1757
--# property-call-start AdaNode.in_aspect
Fld_10 := Libadalang.Implementation.Ada_Node_P_In_Aspect (Node => Block_Var.Node, Name => Sym_3);
--# end
--# expr-done 1758
   If_Result_2 := Fld_10;
end if;
if If_Result_2 then
   --# expr-start 1759 '<FieldAccess for depends_assoc_equation at ast.py:13996>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13996
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AggregateAssoc.depends_assoc_equation
Fld_11 := Libadalang.Implementation.Aggregate_Assoc_P_Depends_Assoc_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 1759
   If_Result_8 := Fld_11;
else
--# expr-start 1761 '<FieldAccess for in_aspect at ast.py:13998>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13998
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1760 '<Symbol Test_Case at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Test_Case); 
--# expr-done 1760
--# property-call-start AdaNode.in_aspect
Fld_12 := Libadalang.Implementation.Ada_Node_P_In_Aspect (Node => Block_Var.Node, Name => Sym_4);
--# end
--# expr-done 1761
if Fld_12 then
   --# expr-start 1762 '<FieldAccess for test_case_assoc_equation at ast.py:13999>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13999
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AggregateAssoc.test_case_assoc_equation
Fld_13 := Libadalang.Implementation.Aggregate_Assoc_P_Test_Case_Assoc_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 1762
   If_Result_7 := Fld_13;
else
--# expr-start 1764 '<FieldAccess for in_aspect at ast.py:14001>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14001
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1763 '<Symbol Refined_State at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Refined_State); 
--# expr-done 1763
--# property-call-start AdaNode.in_aspect
Fld_14 := Libadalang.Implementation.Ada_Node_P_In_Aspect (Node => Block_Var.Node, Name => Sym_5);
--# end
--# expr-done 1764
if Fld_14 then
   --# expr-start 1765 '<FieldAccess for exprs_assoc_equation at ast.py:14004>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14004
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AggregateAssoc.exprs_assoc_equation
Fld_15 := Libadalang.Implementation.Aggregate_Assoc_P_Exprs_Assoc_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 1765
   If_Result_6 := Fld_15;
else
--# expr-start 1767 '<FieldAccess for in_aspect at ast.py:14006>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14006
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1766 '<Symbol Contract_Cases at ???>' Sym_6 None
Sym_6 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Contract_Cases); 
--# expr-done 1766
--# property-call-start AdaNode.in_aspect
Fld_16 := Libadalang.Implementation.Ada_Node_P_In_Aspect (Node => Block_Var.Node, Name => Sym_6);
--# end
--# expr-done 1767
if Fld_16 then
   --# expr-start 1768 '<FieldAccess for contract_cases_assoc_equation at ast.py:14007>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14007
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AggregateAssoc.contract_cases_assoc_equation
Fld_17 := Libadalang.Implementation.Aggregate_Assoc_P_Contract_Cases_Assoc_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 1768
   If_Result_5 := Fld_17;
else
--# expr-start 1770 '<IsA AspectClause, AspectAssoc, PragmaArgumentAssoc at ast.py:14009>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14009
--# expr-start 1769 '<parent at ast.py:14009>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14009
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Libadalang.Implementation.Parent (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# expr-done 1769
Is_A := Fld_18.Node /= null 
and then Fld_18.Node.Kind in Ada_Aspect_Clause | Ada_Aspect_Assoc_Range | Ada_Pragma_Argument_Assoc_Range; 
--# expr-done 1770
if Is_A then
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14010") else null)); 
   If_Result_4 := True_Rel;
else
--# expr-start 1771 '<IsNull at ast.py:14012>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14012
Is_Null := Block_Var_3.Node = null; 
--# expr-done 1771
if Is_Null then
   --# expr-start 1772 '<FieldAccess for record_assoc_equation at ast.py:14013>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14013
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AggregateAssoc.record_assoc_equation
Fld_19 := Libadalang.Implementation.Aggregate_Assoc_P_Record_Assoc_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 1772
   If_Result_3 := Fld_19;
else
   --# expr-start 1773 '<FieldAccess for array_assoc_equation at ast.py:14014>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14014
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AggregateAssoc.array_assoc_equation
Fld_20 := Libadalang.Implementation.Aggregate_Assoc_P_Array_Assoc_Equation (Node => Ent.Node, Atd => Block_Var_3, Mra => Block_Var_1, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 1773
   If_Result_3 := Fld_20;
end if;
      Inc_Ref (If_Result_3);
   If_Result_4 := If_Result_3;
end if;
      Inc_Ref (If_Result_4);
   If_Result_5 := If_Result_4;
end if;
      Inc_Ref (If_Result_5);
   If_Result_6 := If_Result_5;
end if;
      Inc_Ref (If_Result_6);
   If_Result_7 := If_Result_6;
end if;
      Inc_Ref (If_Result_7);
   If_Result_8 := If_Result_7;
end if;
      Inc_Ref (If_Result_8);
   If_Result_9 := If_Result_8;
end if;
      Inc_Ref (If_Result_9);
--# expr-done 1774
Let_Result := If_Result_9; Inc_Ref (Let_Result);
--# end
--# expr-done 1775
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1021;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1021;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aggregate_Assoc_P_Xref_Equation;
--# end
--# property-start AggregateAssoc.record_assoc_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14018
pragma Warnings (Off, "is not referenced");
 function Aggregate_Assoc_P_Record_Assoc_Equation
  (Node : Bare_Aggregate_Assoc
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Aggregate_Assoc := Bare_Aggregate_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Aggregate_Assoc :=
     Internal_Entity_Aggregate_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1022 with Inline_Always;
            procedure Finalizer_Scope_1981 with Inline_Always;
            procedure Finalizer_Scope_1982 with Inline_Always;
            procedure Finalizer_Scope_1983 with Inline_Always;
      Block_Var : Internal_Entity_Base_Aggregate;
Block_Var_1 : Internal_Param_Match_Array_Access;
Block_Var_2 : Internal_Param_Match_Array_Access;
Block_Var_3 : Boolean;
Var_Expr : Bare_Base_Id;
Var_Expr_1 : Internal_Entity_Defining_Name;
Fld : Internal_Entity_Base_Aggregate;
Fld_1 : Internal_Param_Match_Array_Access;
Fld_2 : Internal_Single_Actual;
Fld_3 : Internal_Entity_Basic_Assoc;
Cast_Result : Internal_Entity;
Cast_Expr : Internal_Entity_Aggregate_Assoc;
Cast_Result_1 : Internal_Entity;
Is_Equiv : Boolean;
Map_Result : Internal_Param_Match_Array_Access;
Fld_4 : Internal_Param_Match_Array_Access;
Fld_5 : Internal_Single_Actual;
Fld_6 : Internal_Entity_Basic_Assoc;
Cast_Result_2 : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Aggregate_Assoc;
Cast_Result_3 : Internal_Entity;
Is_Equiv_1 : Boolean;
Map_Result_1 : Internal_Param_Match_Array_Access;
Is_Equal : Boolean;
Not_Val : Boolean;
If_Result : Internal_Param_Match_Array_Access;
Fld_7 : Bare_Ada_Node_Array_Access;
Is_A : Boolean;
Quantifier_Result : Boolean;
Fld_8 : Internal_Single_Actual;
Fld_9 : Internal_Entity_Basic_Assoc;
Fld_10 : Internal_Entity_Expr;
Fld_11 : Logic_Var;
Fld_12 : Internal_Entity_Defining_Name;
Fld_13 : Internal_Entity_Base_Formal_Param_Decl;
Fld_14 : Internal_Entity_Type_Expr;
Fld_15 : Internal_Entity_Base_Type_Decl;
Cast_Result_4 : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_16 : Internal_Single_Actual;
Fld_17 : Internal_Entity_Basic_Assoc;
Fld_18 : Internal_Entity_Expr;
Fld_19 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_20 : Internal_Single_Actual;
Fld_21 : Internal_Entity_Basic_Assoc;
Fld_22 : Internal_Entity_Expr;
Fld_23 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_24 : Internal_Single_Actual;
Fld_25 : Bare_Base_Id;
Fld_26 : Logic_Var;
Fld_27 : Internal_Entity_Defining_Name;
Fld_28 : Internal_Entity_Base_Formal_Param_Decl;
Cast_Result_5 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred_2 : Logic_Equation;
Map_Result_2 : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Fld_29 : Internal_Entity_Defining_Name;
Fld_30 : Internal_Entity_Expr;
Fld_31 : Logic_Var;
Fld_32 : Internal_Entity_Base_Formal_Param_Decl;
Fld_33 : Internal_Entity_Type_Expr;
Fld_34 : Internal_Entity_Base_Type_Decl;
Cast_Result_6 : Internal_Entity;
Bind_Result_2 : Logic_Equation;
Fld_35 : Internal_Entity_Expr;
Fld_36 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_37 : Internal_Entity_Expr;
Fld_38 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Scope_Result_1 : Logic_Equation;
True_Rel_1 : Logic_Equation;
Result_Var_1 : Logic_Equation;
Not_Val_1 : Boolean;
If_Result_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1022 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (If_Result);
                     Dec_Ref (Block_Var_2);
                     Dec_Ref (Fld_7);
                     Dec_Ref (Map_Result_2);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (True_Rel_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1022;
            procedure Finalizer_Scope_1981 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_19);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_23);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred_2);
            end Finalizer_Scope_1981;
            procedure Finalizer_Scope_1982 is
            begin
                     Dec_Ref (Bind_Result_1);
            end Finalizer_Scope_1982;
            procedure Finalizer_Scope_1983 is
            begin
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (Fld_36);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Fld_38);
                     Dec_Ref (And_Pred_4);
            end Finalizer_Scope_1983;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1836 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1776 '<FieldAccess for base_aggregate at ast.py:14023>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14023
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AggregateAssoc.base_aggregate
Fld := Libadalang.Implementation.Aggregate_Assoc_P_Base_Aggregate (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1776
Block_Var := Fld; 
--# bind agg Block_Var
--# expr-start 1781 '<Filter at ast.py:14026>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14026
   --# expr-start 1777 '<FieldAccess for matched_discriminants at ast.py:14026>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14026
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseAggregate.matched_discriminants
Fld_1 := Libadalang.Implementation.Base_Aggregate_P_Matched_Discriminants (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 1777
      declare
         Map_Result_Vec : Internal_Param_Match_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Param_Match_Array_Access := Fld_1;
      begin
         for Item_87 of
               Collection.Items
         loop
   --# scope-start
               --# bind pm Item_87
         --# expr-start 1780 '<Eq at ast.py:14027>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14027
--# expr-start 1779 '<FieldAccess for assoc at ast.py:14027>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14027
--# expr-start 1778 '<FieldAccess for actual at ast.py:14027>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14027
Fld_2 := Item_87.Actual;
--# expr-done 1778
Fld_3 := Fld_2.Assoc;
--# expr-done 1779
      Cast_Result := Create_Internal_Entity
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
Cast_Expr := Ent; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Is_Equiv := Equivalent (Cast_Result, Cast_Result_1); 
--# expr-done 1780
         if Is_Equiv then
         declare
            Item_To_Append : constant Internal_Param_Match := Item_87;
         begin
            Internal_Param_Match_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Param_Match_Array
           (Items_Count => Natural (Internal_Param_Match_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Param_Match_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Param_Match_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Param_Match_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1781
Block_Var_1 := Map_Result; Inc_Ref (Block_Var_1);
--# bind discr_matches Block_Var_1
--# expr-start 1789 '<If at ast.py:14039>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14039
--# expr-start 1783 '<Not at ast.py:14040>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14040
--# expr-start 1782 '<IsNull at ast.py:14040>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14040
Is_Equal := Equivalent (Block_Var_1, No_Internal_Param_Match_Array_Type); 
--# expr-done 1782
Not_Val := not (Is_Equal); 
--# expr-done 1783
if Not_Val then
   If_Result := Block_Var_1;
else
   --# expr-start 1788 '<Filter at ast.py:14044>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14044
   --# expr-start 1784 '<FieldAccess for matched_components at ast.py:14044>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14044
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseAggregate.matched_components
Fld_4 := Libadalang.Implementation.Base_Aggregate_P_Matched_Components (Node => Block_Var.Node, Origin => Origin, Env => Env, E_Info => Block_Var.Info);
--# end
--# expr-done 1784
      declare
         Map_Result_1_Vec : Internal_Param_Match_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Param_Match_Array_Access := Fld_4;
      begin
         for Item_88 of
               Collection.Items
         loop
   --# scope-start
               --# bind pm Item_88
         --# expr-start 1787 '<Eq at ast.py:14045>' Is_Equiv_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14045
--# expr-start 1786 '<FieldAccess for assoc at ast.py:14045>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14045
--# expr-start 1785 '<FieldAccess for actual at ast.py:14045>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14045
Fld_5 := Item_88.Actual;
--# expr-done 1785
Fld_6 := Fld_5.Assoc;
--# expr-done 1786
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_6.Node,
         Info => Fld_6.Info);
Cast_Expr_1 := Ent; 
      Cast_Result_3 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Is_Equiv_1 := Equivalent (Cast_Result_2, Cast_Result_3); 
--# expr-done 1787
         if Is_Equiv_1 then
         declare
            Item_To_Append : constant Internal_Param_Match := Item_88;
         begin
            Internal_Param_Match_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result_1 := Create_Internal_Param_Match_Array
           (Items_Count => Natural (Internal_Param_Match_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Param_Match_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Param_Match_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Param_Match_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 1788
   If_Result := Map_Result_1;
end if;
      Inc_Ref (If_Result);
--# expr-done 1789
Block_Var_2 := If_Result; Inc_Ref (Block_Var_2);
--# bind matches Block_Var_2
--# expr-start 1792 '<AnyQuantifier at ast.py:14050>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14050
   --# expr-start 1790 '<FieldAccess for names at ast.py:14050>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14050
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.names'
Fld_7 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Names (Node => Ent.Node);
--# end
--# expr-done 1790
      Quantifier_Result := False;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Fld_7;
      begin
         for Item_89 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_89
      --# expr-start 1791 '<IsA OthersDesignator at ast.py:14051>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14051
Is_A := Item_89 /= null 
and then Item_89.Kind in Ada_Others_Designator_Range; 
--# expr-done 1791
      Quantifier_Result := Is_A;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
--# expr-done 1792
Block_Var_3 := Quantifier_Result; 
--# bind is_others_assoc Block_Var_3
--# expr-start 1835 '<If at ast.py:14054>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14054
--# expr-start 1793 '<Not at ast.py:14055>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14055
Not_Val_1 := not (Block_Var_3); 
--# expr-done 1793
if Not_Val_1 then
   --# expr-start 1821 '<LogicAll at ast.py:14057>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14057
--# expr-start 1820 '<Map at ast.py:14057>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14057
      declare
         Map_Result_2_Vec : Logic_Equation_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Param_Match_Array_Access := Block_Var_2;
      begin
         for Item_90 of
               Collection.Items
         loop
   --# scope-start
               --# bind match Item_90
      --# expr-start 1819 '<And at ast.py:14057>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14057
--# expr-start 1812 '<And at ast.py:14057>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14057
--# expr-start 1807 '<And at ast.py:14057>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14057
--# expr-start 1802 '<Bind at ast.py:14058>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14058
--# expr-start 1797 '<FieldAccess for expected_type_var at ast.py:14059>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14059
--# expr-start 1796 '<FieldAccess for expr at ast.py:14059>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14059
--# expr-start 1795 '<FieldAccess for assoc at ast.py:14059>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14059
--# expr-start 1794 '<FieldAccess for actual at ast.py:14059>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14059
Fld_8 := Item_90.Actual;
--# expr-done 1794
Fld_9 := Fld_8.Assoc;
--# expr-done 1795
      if Fld_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_10 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_9.Node, E_Info => Fld_9.Info);
--# end
--# expr-done 1796
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_10.Node);
--# expr-done 1797
Fld_11.Value := No_Entity;
Entity_Vars.Reset (Fld_11);
--# expr-start 1801 '<FieldAccess for designated_type at ast.py:14060>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14060
--# expr-start 1800 '<FieldAccess for type_expression at ast.py:14060>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14060
--# expr-start 1799 '<FieldAccess for formal_decl at ast.py:14060>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14060
--# expr-start 1798 '<FieldAccess for formal at ast.py:14060>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14060
Fld_12 := Item_90.Formal;
--# expr-done 1798
      if Fld_12.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_13 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Fld_12.Node, E_Info => Fld_12.Info);
--# end
--# expr-done 1799
      if Fld_13.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.type_expression'
Fld_14 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Type_Expression (Node => Fld_13.Node, E_Info => Fld_13.Info);
--# end
--# expr-done 1800
      if Fld_14.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_15 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Fld_14.Node, Origin => Origin, E_Info => Fld_14.Info);
--# end
--# expr-done 1801
      Cast_Result_4 := Create_Internal_Entity
        (Node => Fld_15.Node,
         Info => Fld_15.Info);
Bind_Result := Solver.Create_Assign (Fld_11, Cast_Result_4, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14058") else null)); 
--# expr-done 1802
--# expr-start 1806 '<FieldAccess for sub_equation at ast.py:14062>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14062
--# expr-start 1805 '<FieldAccess for expr at ast.py:14062>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14062
--# expr-start 1804 '<FieldAccess for assoc at ast.py:14062>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14062
--# expr-start 1803 '<FieldAccess for actual at ast.py:14062>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14062
Fld_16 := Item_90.Actual;
--# expr-done 1803
Fld_17 := Fld_16.Assoc;
--# expr-done 1804
      if Fld_17.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_18 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_17.Node, E_Info => Fld_17.Info);
--# end
--# expr-done 1805
      if Fld_18.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_19 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_18.Node, Env => Env, Origin => Origin, E_Info => Fld_18.Info);
--# end
--# expr-done 1806
And_Pred := Create_And (Bind_Result, Fld_19, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14057") else null)); 
--# expr-done 1807
--# expr-start 1811 '<FieldAccess for matches_expected_type at ast.py:14063>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14063
--# expr-start 1810 '<FieldAccess for expr at ast.py:14063>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14063
--# expr-start 1809 '<FieldAccess for assoc at ast.py:14063>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14063
--# expr-start 1808 '<FieldAccess for actual at ast.py:14063>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14063
Fld_20 := Item_90.Actual;
--# expr-done 1808
Fld_21 := Fld_20.Assoc;
--# expr-done 1809
      if Fld_21.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_22 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_21.Node, E_Info => Fld_21.Info);
--# end
--# expr-done 1810
      if Fld_22.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_23 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_22.Node, Origin => Origin);
--# end
--# expr-done 1811
And_Pred_1 := Create_And (And_Pred, Fld_23, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14057") else null)); 
--# expr-done 1812
--# expr-start 1814 '<FieldAccess for name at ast.py:14064>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14064
--# expr-start 1813 '<FieldAccess for actual at ast.py:14064>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14064
Fld_24 := Item_90.Actual;
--# expr-done 1813
Fld_25 := Fld_24.Name;
--# expr-done 1814
Var_Expr := Fld_25;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind n Var_Expr
--# expr-start 1818 '<Bind at ast.py:14065>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14065
--# expr-start 1815 '<FieldAccess for ref_var at ast.py:14065>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14065
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_26 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Var_Expr);
--# end
--# expr-done 1815
Fld_26.Value := No_Entity;
Entity_Vars.Reset (Fld_26);
--# expr-start 1817 '<FieldAccess for formal_decl at ast.py:14065>' Fld_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14065
--# expr-start 1816 '<FieldAccess for formal at ast.py:14065>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14065
Fld_27 := Item_90.Formal;
--# expr-done 1816
      if Fld_27.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_28 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Fld_27.Node, E_Info => Fld_27.Info);
--# end
--# expr-done 1817
      Cast_Result_5 := Create_Internal_Entity
        (Node => Fld_28.Node,
         Info => Fld_28.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_26, Cast_Result_5, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14065") else null)); 
--# expr-done 1818
Scope_Result := Bind_Result_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1982;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14066") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred_2 := Create_And (And_Pred_1, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14057") else null)); 
--# expr-done 1819
         declare
            Item_To_Append : constant Logic_Equation := And_Pred_2;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_1981;
         end loop;
      end;
         Map_Result_2 := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_2_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 1820
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result_2.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14057") else null)); 
--# expr-done 1821
   If_Result_1 := Logic_Boolean_Op;
else
--# expr-start 1822 '<FieldAccess for first_unmatched_formal at ast.py:14074>' Fld_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14074
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseAggregate.first_unmatched_formal
Fld_29 := Libadalang.Implementation.Base_Aggregate_P_First_Unmatched_Formal (Node => Block_Var.Node, Origin => Origin, Env => Env, E_Info => Block_Var.Info);
--# end
--# expr-done 1822
Var_Expr_1 := Fld_29;
if Var_Expr_1 /= No_Entity_Defining_Name then
   --# scope-start
   --# bind unmatched_formal Var_Expr_1
--# expr-start 1834 '<And at ast.py:14076>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14076
--# expr-start 1831 '<And at ast.py:14076>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14076
--# expr-start 1828 '<Bind at ast.py:14076>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14076
--# expr-start 1824 '<FieldAccess for expected_type_var at ast.py:14077>' Fld_31 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14077
--# expr-start 1823 '<FieldAccess for expr at ast.py:14077>' Fld_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14077
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_30 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1823
      if Fld_30.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_31 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_30.Node);
--# expr-done 1824
Fld_31.Value := No_Entity;
Entity_Vars.Reset (Fld_31);
--# expr-start 1827 '<FieldAccess for designated_type at ast.py:14078>' Fld_34 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14078
--# expr-start 1826 '<FieldAccess for type_expression at ast.py:14078>' Fld_33 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14078
--# expr-start 1825 '<FieldAccess for formal_decl at ast.py:14078>' Fld_32 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14078
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_32 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 1825
      if Fld_32.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.type_expression'
Fld_33 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Type_Expression (Node => Fld_32.Node, E_Info => Fld_32.Info);
--# end
--# expr-done 1826
      if Fld_33.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_34 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Fld_33.Node, Origin => Origin, E_Info => Fld_33.Info);
--# end
--# expr-done 1827
      Cast_Result_6 := Create_Internal_Entity
        (Node => Fld_34.Node,
         Info => Fld_34.Info);
Bind_Result_2 := Solver.Create_Assign (Fld_31, Cast_Result_6, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14076") else null)); 
--# expr-done 1828
--# expr-start 1830 '<FieldAccess for sub_equation at ast.py:14081>' Fld_36 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14081
--# expr-start 1829 '<FieldAccess for expr at ast.py:14081>' Fld_35 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14081
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_35 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1829
      if Fld_35.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_36 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_35.Node, Env => Env, Origin => Origin, E_Info => Fld_35.Info);
--# end
--# expr-done 1830
And_Pred_3 := Create_And (Bind_Result_2, Fld_36, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14076") else null)); 
--# expr-done 1831
--# expr-start 1833 '<FieldAccess for matches_expected_type at ast.py:14082>' Fld_38 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14082
--# expr-start 1832 '<FieldAccess for expr at ast.py:14082>' Fld_37 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14082
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_37 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1832
      if Fld_37.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_38 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_37.Node, Origin => Origin);
--# end
--# expr-done 1833
And_Pred_4 := Create_And (And_Pred_3, Fld_38, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14076") else null)); 
--# expr-done 1834
Scope_Result_1 := And_Pred_4;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_1983;
   Result_Var_1 := Scope_Result_1;
else
   True_Rel_1 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14083") else null)); 
   Result_Var_1 := True_Rel_1;
end if;
      Inc_Ref (Result_Var_1);
   If_Result_1 := Result_Var_1;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 1835
Let_Result := If_Result_1; Inc_Ref (Let_Result);
--# end
--# expr-done 1836
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1022;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1022;
               Finalizer_Scope_1981;
               Finalizer_Scope_1982;
               Finalizer_Scope_1983;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aggregate_Assoc_P_Record_Assoc_Equation;
--# end
--# property-start AggregateAssoc.array_assoc_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14088
pragma Warnings (Off, "is not referenced");
 function Aggregate_Assoc_P_Array_Assoc_Equation
  (Node : Bare_Aggregate_Assoc
      ; Atd : Internal_Entity_Array_Type_Def
      ; Mra : Internal_Multidim_Aggregate_Info
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Aggregate_Assoc := Bare_Aggregate_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Aggregate_Assoc :=
     Internal_Entity_Aggregate_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind atd Atd
   --# bind mra Mra
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1023 with Inline_Always;
            procedure Finalizer_Scope_1984 with Inline_Always;
            procedure Finalizer_Scope_1985 with Inline_Always;
      Var_Expr : Internal_Entity_Name;
Var_Expr_1 : Internal_Entity_Expr;
Fld : Internal_Entity_Expr;
Fld_1 : Logic_Var;
Fld_2 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_3 : Internal_Entity_Expr;
Fld_4 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_5 : Internal_Entity_Expr;
Fld_6 : Logic_Equation;
And_Pred_1 : Logic_Equation;
True_Rel : Logic_Equation;
Is_Equal : Boolean;
Fld_7 : Integer;
Fld_8 : Integer;
Arith_Result : Integer;
Is_Equal_1 : Boolean;
If_Result : Boolean;
If_Result_1 : Logic_Equation;
Fld_9 : Internal_Entity_Alternatives_List;
Fld_10 : Bare_Alternatives_List;
Fld_11 : Internal_Entity_Info;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result_2 : Internal_Entity;
Fld_12 : Logic_Equation;
Cast_Expr : Internal_Entity;
Cast_Result_1 : Internal_Entity_Name;
Fld_13 : Logic_Equation;
Cast_Expr_1 : Internal_Entity;
Cast_Result_2 : Internal_Entity_Expr;
Fld_14 : Logic_Var;
Fld_15 : Integer;
Fld_16 : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
Fld_17 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel_1 : Logic_Equation;
Result_Var : Logic_Equation;
Cast_Expr_2 : Internal_Entity;
Cast_Result_4 : Internal_Entity_Name;
Fld_18 : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Result_Var_1 : Internal_Entity_Base_Type_Decl;
Is_Null_1 : Boolean;
Not_Val : Boolean;
If_Result_3 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Item_91 : Internal_Entity;
Bare_Item_91 : Bare_Ada_Node;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
And_Pred_4 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1023 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_4);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (True_Rel);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1023;
            procedure Finalizer_Scope_1984 is
            begin
                     Dec_Ref (Fld_12);
                     Dec_Ref (Fld_13);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel_1);
                     Dec_Ref (Result_Var);
                     Dec_Ref (If_Result_3);
                     Dec_Ref (And_Pred_3);
            end Finalizer_Scope_1984;
            procedure Finalizer_Scope_1985 is
            begin
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Fld_17);
                     Dec_Ref (And_Pred_2);
            end Finalizer_Scope_1985;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1873 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1872 '<And at ast.py:14094>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14094
--# expr-start 1852 '<If at ast.py:14095>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14095
--# expr-start 1837 '<IsNull at ast.py:14098>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14098
Is_Equal := Mra = No_Multidim_Aggregate_Info; 
--# expr-done 1837
if Is_Equal then
   If_Result := True;
else
   --# expr-start 1841 '<Eq at ast.py:14098>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14098
--# expr-start 1838 '<FieldAccess for rank at ast.py:14098>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14098
Fld_7 := Mra.Rank;
--# expr-done 1838
--# expr-start 1840 '<Arithmetic - at ast.py:14098>' Arith_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14098
--# expr-start 1839 '<FieldAccess for array_ndims at ast.py:14098>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14098
      if Atd.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.array_ndims'
Fld_8 := Libadalang.Implementation.Dispatcher_Type_Def_P_Array_Ndims (Node => Atd.Node, Origin => Origin, E_Info => Atd.Info);
--# end
--# expr-done 1839
Arith_Result := (Fld_8 - 1); 
--# expr-done 1840
Is_Equal_1 := Fld_7 = Arith_Result; 
--# expr-done 1841
   If_Result := Is_Equal_1;
end if;
if If_Result then
   --# expr-start 1851 '<And at ast.py:14101>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14101
--# expr-start 1848 '<And at ast.py:14101>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14101
--# expr-start 1845 '<Bind at ast.py:14101>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14101
--# expr-start 1843 '<FieldAccess for expected_type_var at ast.py:14101>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14101
--# expr-start 1842 '<FieldAccess for expr at ast.py:14101>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14101
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1842
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld.Node);
--# expr-done 1843
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 1844 '<FieldAccess for comp_type at ast.py:14101>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14101
      if Atd.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ArrayTypeDef.comp_type
Fld_2 := Libadalang.Implementation.Array_Type_Def_P_Comp_Type (Node => Atd.Node, Origin => Origin, E_Info => Atd.Info);
--# end
--# expr-done 1844
      Cast_Result := Create_Internal_Entity
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Bind_Result := Solver.Create_Assign (Fld_1, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14101") else null)); 
--# expr-done 1845
--# expr-start 1847 '<FieldAccess for sub_equation at ast.py:14102>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14102
--# expr-start 1846 '<FieldAccess for expr at ast.py:14102>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14102
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_3 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1846
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_4 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_3.Node, Env => Env, Origin => Origin, E_Info => Fld_3.Info);
--# end
--# expr-done 1847
And_Pred := Create_And (Bind_Result, Fld_4, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14101") else null)); 
--# expr-done 1848
--# expr-start 1850 '<FieldAccess for matches_expected_type at ast.py:14103>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14103
--# expr-start 1849 '<FieldAccess for expr at ast.py:14103>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14103
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_5 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1849
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_6 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_5.Node, Origin => Origin);
--# end
--# expr-done 1850
And_Pred_1 := Create_And (And_Pred, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14101") else null)); 
--# expr-done 1851
   If_Result_1 := And_Pred_1;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14107") else null)); 
   If_Result_1 := True_Rel;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 1852
--# expr-start 1871 '<LogicAll at ast.py:14110>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14110
--# expr-start 1870 '<Map at ast.py:14110>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14110
   --# expr-start 1853 '<FieldAccess for designators at ast.py:14110>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14110
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Alternatives_List (Node => Ent.Node.Aggregate_Assoc_F_Designators, Info => Ent.Info);
--# expr-done 1853
Fld_10 := Fld_9.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_10 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Alternatives_List := Fld_10;
      begin
         for Untyped_Item_91 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_91 := Untyped_Item_91; 
Is_Null := Bare_Item_91 = null; 
if Is_Null then
   If_Result_2 := No_Entity;
else
Fld_11 := Fld_9.Info;
As_Entity := (Info => Fld_11, Node => Bare_Item_91); 
   If_Result_2 := As_Entity;
end if;
                  Item_91 := If_Result_2; 
   --# scope-start
               --# bind n Item_91
      --# expr-start 1869 '<And at ast.py:14111>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14111
--# expr-start 1854 '<FieldAccess for sub_equation at ast.py:14111>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14111
      if Item_91.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_12 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Item_91.Node, Env => Env, Origin => Origin, E_Info => Item_91.Info);
--# end
--# expr-done 1854
--# expr-start 1868 '<If at ast.py:14111>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14111
--# expr-start 1858 '<Not at ast.py:14112>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14112
--# expr-start 1857 '<IsNull at ast.py:14112>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14112
--# expr-start 1855 '<Cast to Name at ast.py:14112>' Cast_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14112
Cast_Expr_2 := Item_91; 
   if Cast_Expr_2.Node = null
      or else Cast_Expr_2.Node.Kind in Ada_Name
   then
      Cast_Result_4 := Create_Internal_Entity_Name
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
   else
         Cast_Result_4 := No_Entity_Name;
   end if;
--# expr-done 1855
Var_Expr := Cast_Result_4;
if Var_Expr /= No_Entity_Name then
   --# scope-start
--# expr-start 1856 '<FieldAccess for name_designated_type at ast.py:14112>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14112
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_18 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 1856
Scope_Result_1 := Fld_18;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Base_Type_Decl;
end if;
Is_Null_1 := Result_Var_1.Node = null; 
--# expr-done 1857
Not_Val := not (Is_Null_1); 
--# expr-done 1858
if Not_Val then
   --# expr-start 1860 '<FieldAccess for xref_no_overloading at ast.py:14114>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14114
--# expr-start 1859 '<Cast to Name at ast.py:14114>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14114
Cast_Expr := Item_91; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Name
   then
      Cast_Result_1 := Create_Internal_Entity_Name
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result_1 := No_Entity_Name;
   end if;
--# expr-done 1859
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_13 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Cast_Result_1.Node, Env => Env, Origin => Origin, E_Info => Cast_Result_1.Info);
--# end
--# expr-done 1860
   If_Result_3 := Fld_13;
else
--# expr-start 1861 '<Cast to Expr at ast.py:14116>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14116
Cast_Expr_1 := Item_91; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Expr
   then
      Cast_Result_2 := Create_Internal_Entity_Expr
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_2 := No_Entity_Expr;
   end if;
--# expr-done 1861
Var_Expr_1 := Cast_Result_2;
if Var_Expr_1 /= No_Entity_Expr then
   --# scope-start
   --# bind n Var_Expr_1
--# expr-start 1867 '<And at ast.py:14118>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14118
--# expr-start 1865 '<Bind at ast.py:14118>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14118
--# expr-start 1862 '<FieldAccess for expected_type_var at ast.py:14118>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14118
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Var_Expr_1.Node);
--# expr-done 1862
Fld_14.Value := No_Entity;
Entity_Vars.Reset (Fld_14);
--# expr-start 1864 '<FieldAccess for index_type at ast.py:14118>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14118
      if Atd.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1863 '<FieldAccess for rank at ast.py:14118>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14118
Fld_15 := Mra.Rank;
--# expr-done 1863
--# property-call-start ArrayTypeDef.index_type
Fld_16 := Libadalang.Implementation.Array_Type_Def_P_Index_Type (Node => Atd.Node, Dim => Fld_15, Origin => Origin, E_Info => Atd.Info);
--# end
--# expr-done 1864
      Cast_Result_3 := Create_Internal_Entity
        (Node => Fld_16.Node,
         Info => Fld_16.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_14, Cast_Result_3, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14118") else null)); 
--# expr-done 1865
--# expr-start 1866 '<FieldAccess for matches_expected_type at ast.py:14119>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14119
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_17 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Var_Expr_1.Node, Origin => Origin);
--# end
--# expr-done 1866
And_Pred_2 := Create_And (Bind_Result_1, Fld_17, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14118") else null)); 
--# expr-done 1867
Scope_Result := And_Pred_2;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1985;
   Result_Var := Scope_Result;
else
   True_Rel_1 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14120") else null)); 
   Result_Var := True_Rel_1;
end if;
      Inc_Ref (Result_Var);
   If_Result_3 := Result_Var;
end if;
      Inc_Ref (If_Result_3);
--# expr-done 1868
And_Pred_3 := Create_And (Fld_12, If_Result_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14111") else null)); 
--# expr-done 1869
         declare
            Item_To_Append : constant Logic_Equation := And_Pred_3;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_1984;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1870
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14110") else null)); 
--# expr-done 1871
And_Pred_4 := Create_And (If_Result_1, Logic_Boolean_Op, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14094") else null)); 
--# expr-done 1872
Let_Result := And_Pred_4; Inc_Ref (Let_Result);
--# end
--# expr-done 1873
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1023;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1023;
               Finalizer_Scope_1984;
               Finalizer_Scope_1985;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aggregate_Assoc_P_Array_Assoc_Equation;
--# end
--# property-start AggregateAssoc.globals_assoc_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14127
pragma Warnings (Off, "is not referenced");
 function Aggregate_Assoc_P_Globals_Assoc_Equation
  (Node : Bare_Aggregate_Assoc
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Aggregate_Assoc := Bare_Aggregate_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Aggregate_Assoc :=
     Internal_Entity_Aggregate_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1024 with Inline_Always;
      Fld : Internal_Entity_Expr;
Fld_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1024 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1024;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1876 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1875 '<FieldAccess for sub_equation at ast.py:14135>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14135
--# expr-start 1874 '<FieldAccess for expr at ast.py:14135>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14135
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1874
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 1875
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 1876
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1024;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1024;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aggregate_Assoc_P_Globals_Assoc_Equation;
--# end
--# property-start AggregateAssoc.depends_assoc_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14138
pragma Warnings (Off, "is not referenced");
 function Aggregate_Assoc_P_Depends_Assoc_Equation
  (Node : Bare_Aggregate_Assoc
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Aggregate_Assoc := Bare_Aggregate_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Aggregate_Assoc :=
     Internal_Entity_Aggregate_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1025 with Inline_Always;
            procedure Finalizer_Scope_1987 with Inline_Always;
      Var_Expr : Internal_Entity_Identifier;
Fld : Internal_Entity_Expr;
Fld_1 : Logic_Equation;
Fld_2 : Bare_Ada_Node_Array_Access;
Get_Result : Bare_Ada_Node;
Cast_Result : Bare_Identifier;
As_Entity : Internal_Entity_Identifier;
Is_Null : Boolean;
If_Result : Internal_Entity_Identifier;
Fld_3 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1025 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1025;
            procedure Finalizer_Scope_1987 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_1987;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1885 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1884 '<And at ast.py:14143>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14143
--# expr-start 1878 '<FieldAccess for sub_equation at ast.py:14148>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14148
--# expr-start 1877 '<FieldAccess for expr at ast.py:14148>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14148
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1877
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 1878
--# expr-start 1882 '<as_entity at ast.py:14156>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14156
--# expr-start 1881 '<Cast to Identifier at ast.py:14156>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14156
--# expr-start 1880 '<at at ast.py:14156>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14156
--# expr-start 1879 '<FieldAccess for names at ast.py:14156>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14156
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.names'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Names (Node => Ent.Node);
--# end
--# expr-done 1879
Get_Result := Get (Fld_2, 0, True); 
--# expr-done 1880
   if Get_Result = null
      or else Get_Result.Kind in Ada_Identifier_Range
   then
      Cast_Result := Get_Result;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 1881
Is_Null := Cast_Result = null; 
if Is_Null then
   If_Result := No_Entity_Identifier;
else
As_Entity := (Info => E_Info, Node => Cast_Result); 
   If_Result := As_Entity;
end if;
--# expr-done 1882
Var_Expr := If_Result;
if Var_Expr /= No_Entity_Identifier then
   --# scope-start
   --# bind n Var_Expr
--# expr-start 1883 '<FieldAccess for sub_equation at ast.py:14157>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14157
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 1883
Scope_Result := Fld_3;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1987;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14157") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred := Create_And (Fld_1, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14143") else null)); 
--# expr-done 1884
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 1885
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1025;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1025;
               Finalizer_Scope_1987;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aggregate_Assoc_P_Depends_Assoc_Equation;
--# end
--# property-start AggregateAssoc.test_case_assoc_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14162
pragma Warnings (Off, "is not referenced");
 function Aggregate_Assoc_P_Test_Case_Assoc_Equation
  (Node : Bare_Aggregate_Assoc
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Aggregate_Assoc := Bare_Aggregate_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Aggregate_Assoc :=
     Internal_Entity_Aggregate_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1026 with Inline_Always;
      Var_Expr : Bare_Base_Id;
Fld : Internal_Entity_Expr;
Fld_1 : Logic_Equation;
True_Rel : Logic_Equation;
Fld_2 : Bare_Ada_Node_Array_Access;
Get_Result : Bare_Ada_Node;
Cast_Result : Bare_Base_Id;
Fld_3 : Symbol_Type;
Sym : Symbol_Type;
Is_Equal : Boolean;
Fld_4 : Symbol_Type;
Sym_1 : Symbol_Type;
Is_Equal_1 : Boolean;
If_Result : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1026 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Fld_2);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1026;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1898 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1897 '<If at ast.py:14167>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14167
--# expr-start 1888 '<Cast to BaseId at ast.py:14170>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14170
--# expr-start 1887 '<at at ast.py:14170>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14170
--# expr-start 1886 '<FieldAccess for names at ast.py:14170>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14170
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.names'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Names (Node => Ent.Node);
--# end
--# expr-done 1886
Get_Result := Get (Fld_2, 0, True); 
--# expr-done 1887
   if Get_Result = null
      or else Get_Result.Kind in Ada_Base_Id
   then
      Cast_Result := Get_Result;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 1888
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind name Var_Expr
--# expr-start 1891 '<Eq at ast.py:14171>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14171
--# expr-start 1889 '<FieldAccess for name_symbol at ast.py:14171>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14171
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Var_Expr);
--# end
--# expr-done 1889
--# expr-start 1890 '<Symbol Requires at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Requires); 
--# expr-done 1890
Is_Equal := Fld_3 = Sym; 
--# expr-done 1891
if Is_Equal then
   If_Result := True;
else
   --# expr-start 1894 '<Eq at ast.py:14171>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14171
--# expr-start 1892 '<FieldAccess for name_symbol at ast.py:14171>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14171
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_4 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Var_Expr);
--# end
--# expr-done 1892
--# expr-start 1893 '<Symbol Ensures at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Ensures); 
--# expr-done 1893
Is_Equal_1 := Fld_4 = Sym_1; 
--# expr-done 1894
   If_Result := Is_Equal_1;
end if;
Scope_Result := If_Result;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
if Result_Var then
   --# expr-start 1896 '<FieldAccess for sub_equation at ast.py:14175>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14175
--# expr-start 1895 '<FieldAccess for expr at ast.py:14175>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14175
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1895
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 1896
   If_Result_1 := Fld_1;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14176") else null)); 
   If_Result_1 := True_Rel;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 1897
Let_Result := If_Result_1; Inc_Ref (Let_Result);
--# end
--# expr-done 1898
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1026;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1026;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aggregate_Assoc_P_Test_Case_Assoc_Equation;
--# end
--# property-start AggregateAssoc.contract_cases_assoc_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14180
pragma Warnings (Off, "is not referenced");
 function Aggregate_Assoc_P_Contract_Cases_Assoc_Equation
  (Node : Bare_Aggregate_Assoc
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Aggregate_Assoc := Bare_Aggregate_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Aggregate_Assoc :=
     Internal_Entity_Aggregate_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1027 with Inline_Always;
            procedure Finalizer_Scope_1989 with Inline_Always;
            procedure Finalizer_Scope_1990 with Inline_Always;
      Var_Expr : Internal_Entity_Expr;
Fld : Internal_Entity_Alternatives_List;
Fld_1 : Bare_Alternatives_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Expr;
Fld_3 : Logic_Var;
Fld_4 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_5 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_6 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
Item_92 : Internal_Entity;
Bare_Item_92 : Bare_Ada_Node;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Fld_7 : Internal_Entity_Expr;
Fld_8 : Logic_Var;
Fld_9 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_10 : Internal_Entity_Expr;
Fld_11 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_12 : Internal_Entity_Expr;
Fld_13 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1027 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Fld_11);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Fld_13);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1027;
            procedure Finalizer_Scope_1989 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1989;
            procedure Finalizer_Scope_1990 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_5);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_1990;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1921 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1920 '<And at ast.py:14186>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14186
--# expr-start 1917 '<And at ast.py:14186>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14186
--# expr-start 1914 '<And at ast.py:14186>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14186
--# expr-start 1909 '<LogicAll at ast.py:14187>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14187
--# expr-start 1908 '<Map at ast.py:14187>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14187
   --# expr-start 1899 '<FieldAccess for designators at ast.py:14187>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14187
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Alternatives_List (Node => Ent.Node.Aggregate_Assoc_F_Designators, Info => Ent.Info);
--# expr-done 1899
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Alternatives_List := Fld_1;
      begin
         for Untyped_Item_92 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_92 := Untyped_Item_92; 
Is_Null := Bare_Item_92 = null; 
if Is_Null then
   If_Result := No_Entity;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_92); 
   If_Result := As_Entity;
end if;
                  Item_92 := If_Result; 
   --# scope-start
               --# bind d Item_92
--# expr-start 1900 '<Cast to Expr at ast.py:14188>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14188
Cast_Expr := Item_92; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Expr
   then
      Cast_Result := Create_Internal_Entity_Expr
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Expr;
   end if;
--# expr-done 1900
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind e Var_Expr
--# expr-start 1907 '<And at ast.py:14190>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14190
--# expr-start 1905 '<And at ast.py:14190>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14190
--# expr-start 1903 '<Bind at ast.py:14190>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14190
--# expr-start 1901 '<FieldAccess for expected_type_var at ast.py:14190>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14190
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Var_Expr.Node);
--# expr-done 1901
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
--# expr-start 1902 '<FieldAccess for bool_type at ast.py:14190>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14190
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_4 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 1902
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_4.Node,
         Info => Fld_4.Info);
Bind_Result := Solver.Create_Assign (Fld_3, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14190") else null)); 
--# expr-done 1903
--# expr-start 1904 '<FieldAccess for sub_equation at ast.py:14191>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14191
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_5 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 1904
And_Pred := Create_And (Bind_Result, Fld_5, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14190") else null)); 
--# expr-done 1905
--# expr-start 1906 '<FieldAccess for matches_expected_formal_prim_type at ast.py:14192>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14192
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_prim_type
Fld_6 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Prim_Type (Node => Var_Expr.Node, Origin => Origin);
--# end
--# expr-done 1906
And_Pred_1 := Create_And (And_Pred, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14190") else null)); 
--# expr-done 1907
Scope_Result := And_Pred_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1990;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14195") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
         declare
            Item_To_Append : constant Logic_Equation := Result_Var;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_1989;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1908
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14187") else null)); 
--# expr-done 1909
--# expr-start 1913 '<Bind at ast.py:14198>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14198
--# expr-start 1911 '<FieldAccess for expected_type_var at ast.py:14198>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14198
--# expr-start 1910 '<FieldAccess for expr at ast.py:14198>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14198
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_7 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1910
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_7.Node);
--# expr-done 1911
Fld_8.Value := No_Entity;
Entity_Vars.Reset (Fld_8);
--# expr-start 1912 '<FieldAccess for bool_type at ast.py:14198>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14198
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_9 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 1912
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_9.Node,
         Info => Fld_9.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_8, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14198") else null)); 
--# expr-done 1913
And_Pred_2 := Create_And (Logic_Boolean_Op, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14186") else null)); 
--# expr-done 1914
--# expr-start 1916 '<FieldAccess for sub_equation at ast.py:14199>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14199
--# expr-start 1915 '<FieldAccess for expr at ast.py:14199>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14199
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_10 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1915
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_11 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_10.Node, Env => Env, Origin => Origin, E_Info => Fld_10.Info);
--# end
--# expr-done 1916
And_Pred_3 := Create_And (And_Pred_2, Fld_11, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14186") else null)); 
--# expr-done 1917
--# expr-start 1919 '<FieldAccess for matches_expected_formal_prim_type at ast.py:14200>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14200
--# expr-start 1918 '<FieldAccess for expr at ast.py:14200>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14200
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_12 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1918
      if Fld_12.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_prim_type
Fld_13 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Prim_Type (Node => Fld_12.Node, Origin => Origin);
--# end
--# expr-done 1919
And_Pred_4 := Create_And (And_Pred_3, Fld_13, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14186") else null)); 
--# expr-done 1920
Let_Result := And_Pred_4; Inc_Ref (Let_Result);
--# end
--# expr-done 1921
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1027;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1027;
               Finalizer_Scope_1989;
               Finalizer_Scope_1990;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aggregate_Assoc_P_Contract_Cases_Assoc_Equation;
--# end
--# property-start AggregateAssoc.exprs_assoc_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14204
pragma Warnings (Off, "is not referenced");
 function Aggregate_Assoc_P_Exprs_Assoc_Equation
  (Node : Bare_Aggregate_Assoc
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Aggregate_Assoc := Bare_Aggregate_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Aggregate_Assoc :=
     Internal_Entity_Aggregate_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1028 with Inline_Always;
            procedure Finalizer_Scope_1991 with Inline_Always;
            procedure Finalizer_Scope_1992 with Inline_Always;
      Var_Expr : Internal_Entity_Name;
Fld : Internal_Entity_Alternatives_List;
Fld_1 : Bare_Alternatives_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Fld_3 : Logic_Equation;
Item_93 : Internal_Entity;
Bare_Item_93 : Bare_Ada_Node;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Fld_4 : Internal_Entity_Expr;
Cast_Result : Internal_Entity_Name;
Fld_5 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1028 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1028;
            procedure Finalizer_Scope_1991 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_1991;
            procedure Finalizer_Scope_1992 is
            begin
                     Dec_Ref (Fld_5);
            end Finalizer_Scope_1992;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1930 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1929 '<And at ast.py:14210>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14210
--# expr-start 1925 '<LogicAll at ast.py:14211>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14211
--# expr-start 1924 '<Map at ast.py:14211>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14211
   --# expr-start 1922 '<FieldAccess for designators at ast.py:14211>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14211
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Alternatives_List (Node => Ent.Node.Aggregate_Assoc_F_Designators, Info => Ent.Info);
--# expr-done 1922
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Alternatives_List := Fld_1;
      begin
         for Untyped_Item_93 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_93 := Untyped_Item_93; 
Is_Null := Bare_Item_93 = null; 
if Is_Null then
   If_Result := No_Entity;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_93); 
   If_Result := As_Entity;
end if;
                  Item_93 := If_Result; 
   --# scope-start
               --# bind d Item_93
      --# expr-start 1923 '<FieldAccess for sub_equation at ast.py:14211>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14211
      if Item_93.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Item_93.Node, Env => Env, Origin => Origin, E_Info => Item_93.Info);
--# end
--# expr-done 1923
         declare
            Item_To_Append : constant Logic_Equation := Fld_3;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_1991;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1924
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14211") else null)); 
--# expr-done 1925
--# expr-start 1927 '<Cast to Name at ast.py:14212>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14212
--# expr-start 1926 '<FieldAccess for expr at ast.py:14212>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14212
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_4 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1926
   if Fld_4.Node = null
      or else Fld_4.Node.Kind in Ada_Name
   then
      Cast_Result := Create_Internal_Entity_Name
        (Node => Fld_4.Node,
         Info => Fld_4.Info);
   else
         Cast_Result := No_Entity_Name;
   end if;
--# expr-done 1927
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Name then
   --# scope-start
   --# bind n Var_Expr
--# expr-start 1928 '<FieldAccess for sub_equation at ast.py:14213>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14213
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_5 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 1928
Scope_Result := Fld_5;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1992;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14214") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred := Create_And (Logic_Boolean_Op, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14210") else null)); 
--# expr-done 1929
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 1930
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1028;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1028;
               Finalizer_Scope_1991;
               Finalizer_Scope_1992;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Aggregate_Assoc_P_Exprs_Assoc_Equation;
--# end
   --
   --  Primitives for Bare_Multi_Dim_Array_Assoc
   --
      procedure Initialize_Fields_For_Multi_Dim_Array_Assoc
        (Self : Bare_Multi_Dim_Array_Assoc
         ; Aggregate_Assoc_F_Designators : Bare_Alternatives_List
         ; Aggregate_Assoc_F_R_Expr : Bare_Expr
        ) is
      begin
            Initialize_Fields_For_Aggregate_Assoc
              (Self, Aggregate_Assoc_F_Designators, Aggregate_Assoc_F_R_Expr);
      end Initialize_Fields_For_Multi_Dim_Array_Assoc;
   --
   --  Primitives for Bare_Composite_Constraint_Assoc
   --
      procedure Initialize_Fields_For_Composite_Constraint_Assoc
        (Self : Bare_Composite_Constraint_Assoc
         ; Composite_Constraint_Assoc_F_Ids : Bare_Discriminant_Choice_List
         ; Composite_Constraint_Assoc_F_Constraint_Expr : Bare_Ada_Node
        ) is
      begin
            Self.Composite_Constraint_Assoc_F_Ids := Composite_Constraint_Assoc_F_Ids;
            Self.Composite_Constraint_Assoc_F_Constraint_Expr := Composite_Constraint_Assoc_F_Constraint_Expr;
      end Initialize_Fields_For_Composite_Constraint_Assoc;
   function Composite_Constraint_Assoc_F_Ids
     (Node : Bare_Composite_Constraint_Assoc) return Bare_Discriminant_Choice_List
   is
   begin
         return Node.Composite_Constraint_Assoc_F_Ids;
   end;
   function Composite_Constraint_Assoc_F_Constraint_Expr
     (Node : Bare_Composite_Constraint_Assoc) return Bare_Ada_Node
   is
   begin
         return Node.Composite_Constraint_Assoc_F_Constraint_Expr;
   end;
--# property-start CompositeConstraintAssoc.expr /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7503
pragma Warnings (Off, "is not referenced");
 function Composite_Constraint_Assoc_P_Expr
  (Node : Bare_Composite_Constraint_Assoc
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Expr
is
   Self : Bare_Composite_Constraint_Assoc := Bare_Composite_Constraint_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Composite_Constraint_Assoc :=
     Internal_Entity_Composite_Constraint_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Expr;
      Fld : Internal_Entity;
Cast_Result : Internal_Entity_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1932 '<Cast to Expr at ast.py:7503>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7503
--# expr-start 1931 '<FieldAccess for constraint_expr at ast.py:7503>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7503
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity (Node => Ent.Node.Composite_Constraint_Assoc_F_Constraint_Expr, Info => Ent.Info);
--# expr-done 1931
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Expr
   then
      Cast_Result := Create_Internal_Entity_Expr
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 1932
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Composite_Constraint_Assoc_P_Expr;
--# end
--# property-start CompositeConstraintAssoc.names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7504
pragma Warnings (Off, "is not referenced");
 function Composite_Constraint_Assoc_P_Names
  (Node : Bare_Composite_Constraint_Assoc
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Composite_Constraint_Assoc := Bare_Composite_Constraint_Assoc (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_556 with Inline_Always;
      Fld : Bare_Discriminant_Choice_List;
Cast_Expr : Bare_Identifier;
Cast_Result : Bare_Ada_Node;
Item_94 : Bare_Identifier;
Map_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_556 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_556;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1935 '<Map at ast.py:7504>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7504
   --# expr-start 1933 '<FieldAccess for ids at ast.py:7504>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7504
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Composite_Constraint_Assoc_F_Ids;
--# expr-done 1933
      declare
         Map_Result_Vec : Bare_Ada_Node_Vectors.Vector;
      begin
      if Fld = null then
      Map_Result := Create_Bare_Ada_Node_Array (0);
      else
      declare
         Collection : constant Bare_Discriminant_Choice_List := Fld;
      begin
         for Untyped_Item_94 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_94 := Untyped_Item_94; 
   --# scope-start
               --# bind i Item_94
      --# expr-start 1934 '<Cast to AdaNode at ast.py:7504>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7504
Cast_Expr := Item_94; 
      Cast_Result := Cast_Expr;
--# expr-done 1934
         declare
            Item_To_Append : constant Bare_Ada_Node := Cast_Result;
         begin
            Bare_Ada_Node_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Bare_Ada_Node_Array
           (Items_Count => Natural (Bare_Ada_Node_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Bare_Ada_Node_Vectors.Get
              (Map_Result_Vec,
               I + Bare_Ada_Node_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Bare_Ada_Node_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 1935
      Property_Result := Map_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_556;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_556;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Composite_Constraint_Assoc_P_Names;
--# end
   --
   --  Primitives for Bare_Iterated_Assoc
   --
      procedure Initialize_Fields_For_Iterated_Assoc
        (Self : Bare_Iterated_Assoc
         ; Iterated_Assoc_F_Spec : Bare_For_Loop_Spec
         ; Iterated_Assoc_F_R_Expr : Bare_Expr
        ) is
      begin
            Self.Iterated_Assoc_F_Spec := Iterated_Assoc_F_Spec;
            Self.Iterated_Assoc_F_R_Expr := Iterated_Assoc_F_R_Expr;
      end Initialize_Fields_For_Iterated_Assoc;
   function Iterated_Assoc_F_Spec
     (Node : Bare_Iterated_Assoc) return Bare_For_Loop_Spec
   is
   begin
         return Node.Iterated_Assoc_F_Spec;
   end;
   function Iterated_Assoc_F_R_Expr
     (Node : Bare_Iterated_Assoc) return Bare_Expr
   is
   begin
         return Node.Iterated_Assoc_F_R_Expr;
   end;
--# property-start IteratedAssoc.expr /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14226
pragma Warnings (Off, "is not referenced");
 function Iterated_Assoc_P_Expr
  (Node : Bare_Iterated_Assoc
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Expr
is
   Self : Bare_Iterated_Assoc := Bare_Iterated_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Iterated_Assoc :=
     Internal_Entity_Iterated_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Expr;
      Fld : Internal_Entity_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1936 '<FieldAccess for r_expr at ast.py:14226>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14226
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Iterated_Assoc_F_R_Expr, Info => Ent.Info);
--# expr-done 1936
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Iterated_Assoc_P_Expr;
--# end
--# property-start IteratedAssoc.names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14227
pragma Warnings (Off, "is not referenced");
 function Iterated_Assoc_P_Names
  (Node : Bare_Iterated_Assoc
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Iterated_Assoc := Bare_Iterated_Assoc (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := No_Bare_Ada_Node_Array_Type;
         Inc_Ref (Property_Result);
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Iterated_Assoc_P_Names;
--# end
--# property-start IteratedAssoc.base_aggregate /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14229
pragma Warnings (Off, "is not referenced");
 function Iterated_Assoc_P_Base_Aggregate
  (Node : Bare_Iterated_Assoc
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Aggregate
is
   Self : Bare_Iterated_Assoc := Bare_Iterated_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Iterated_Assoc :=
     Internal_Entity_Iterated_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Aggregate;
      Fld : Internal_Entity;
Fld_1 : Internal_Entity;
Cast_Result : Internal_Entity_Base_Aggregate;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1939 '<Cast to BaseAggregate at ast.py:14230>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14230
--# expr-start 1938 '<parent at ast.py:14230>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14230
--# expr-start 1937 '<parent at ast.py:14230>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14230
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 1937
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Parent (Node => Fld.Node, E_Info => Fld.Info);
--# expr-done 1938
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Base_Aggregate
   then
      Cast_Result := Create_Internal_Entity_Base_Aggregate
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 1939
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Iterated_Assoc_P_Base_Aggregate;
--# end
--# property-start IteratedAssoc.xref_stop_resolution /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14233
pragma Warnings (Off, "is not referenced");
 function Iterated_Assoc_P_Xref_Stop_Resolution
  (Node : Bare_Iterated_Assoc
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Iterated_Assoc := Bare_Iterated_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Iterated_Assoc :=
     Internal_Entity_Iterated_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Internal_Entity;
Fld_1 : Internal_Entity;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 1942 '<IsA BaseAggregate at ast.py:14233>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14233
--# expr-start 1941 '<parent at ast.py:14233>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14233
--# expr-start 1940 '<parent at ast.py:14233>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14233
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 1940
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Parent (Node => Fld.Node, E_Info => Fld.Info);
--# expr-done 1941
Is_A := Fld_1.Node /= null 
and then Fld_1.Node.Kind in Ada_Base_Aggregate; 
--# expr-done 1942
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Iterated_Assoc_P_Xref_Stop_Resolution;
--# end
--# property-start IteratedAssoc.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14236
pragma Warnings (Off, "is not referenced");
 function Iterated_Assoc_P_Xref_Equation
  (Node : Bare_Iterated_Assoc
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Iterated_Assoc := Bare_Iterated_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Iterated_Assoc :=
     Internal_Entity_Iterated_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1033 with Inline_Always;
      Block_Var : Internal_Entity_Base_Aggregate;
Block_Var_1 : Internal_Multidim_Aggregate_Info;
Block_Var_2 : Internal_Entity_Base_Type_Decl;
Block_Var_3 : Internal_Entity_Array_Type_Def;
Block_Var_4 : Boolean;
Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Base_Aggregate;
Fld_1 : Internal_Multidim_Aggregate_Info;
Fld_2 : Internal_Entity_Base_Type_Decl;
Fld_3 : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
Is_Equal : Boolean;
Not_Val : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
Fld_4 : Internal_Entity_Array_Type_Def;
Scope_Result : Internal_Entity_Array_Type_Def;
Result_Var : Internal_Entity_Array_Type_Def;
Fld_5 : Internal_Entity_For_Loop_Spec;
Fld_6 : Internal_Entity_For_Loop_Spec;
Fld_7 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Expr;
Fld_8 : Logic_Var;
Fld_9 : Integer;
Fld_10 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_11 : Internal_Entity_For_Loop_Spec;
Fld_12 : Internal_Entity;
Cast_Result_3 : Internal_Entity_Expr;
Fld_13 : Logic_Equation;
And_Pred : Logic_Equation;
True_Rel : Logic_Equation;
Fld_14 : Bare_For_Loop_Spec;
Fld_15 : Bare_Iter_Type;
Is_A : Boolean;
If_Result_1 : Logic_Equation;
Fld_16 : Boolean;
Fld_17 : Internal_Entity_Expr;
Fld_18 : Logic_Equation;
Fld_19 : Internal_Entity_Expr;
Fld_20 : Logic_Var;
Fld_21 : Internal_Entity_Base_Type_Decl;
Cast_Result_4 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_22 : Internal_Entity_Expr;
Fld_23 : Logic_Equation;
And_Pred_2 : Logic_Equation;
True_Rel_1 : Logic_Equation;
Is_Equal_1 : Boolean;
Fld_24 : Integer;
Fld_25 : Integer;
Arith_Result : Integer;
Is_Equal_2 : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Logic_Equation;
False_Rel : Logic_Equation;
If_Result_4 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1033 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_13);
                     Dec_Ref (And_Pred);
                     Dec_Ref (True_Rel);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Fld_18);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_23);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (True_Rel_1);
                     Dec_Ref (If_Result_3);
                     Dec_Ref (False_Rel);
                     Dec_Ref (If_Result_4);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1033;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 1987 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1943 '<FieldAccess for base_aggregate at ast.py:14237>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14237
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start IteratedAssoc.base_aggregate
Fld := Libadalang.Implementation.Iterated_Assoc_P_Base_Aggregate (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1943
Block_Var := Fld; 
--# bind aggregate Block_Var
--# expr-start 1944 '<FieldAccess for multidim_root_aggregate at ast.py:14239>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14239
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseAggregate.multidim_root_aggregate
Fld_1 := Libadalang.Implementation.Base_Aggregate_P_Multidim_Root_Aggregate (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 1944
Block_Var_1 := Fld_1; 
--# bind root_agg Block_Var_1
--# expr-start 1950 '<If at ast.py:14244>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14244
--# expr-start 1946 '<Not at ast.py:14245>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14245
--# expr-start 1945 '<IsNull at ast.py:14245>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14245
Is_Equal := Block_Var_1 = No_Multidim_Aggregate_Info; 
--# expr-done 1945
Not_Val := not (Is_Equal); 
--# expr-done 1946
if Not_Val then
   --# expr-start 1947 '<FieldAccess for typ at ast.py:14246>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14246
Fld_2 := Block_Var_1.Typ;
--# expr-done 1947
   If_Result := Fld_2;
else
   --# expr-start 1949 '<Cast to BaseTypeDecl at ast.py:14247>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14247
--# expr-start 1948 '<FieldAccess for type_val at ast.py:14247>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14247
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.type_val
Fld_3 := Libadalang.Implementation.Expr_P_Type_Val (Node => Block_Var.Node);
--# end
--# expr-done 1948
   if Fld_3.Node = null
      or else Fld_3.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 1949
   If_Result := Cast_Result;
end if;
--# expr-done 1950
Block_Var_2 := If_Result; 
--# bind type_decl Block_Var_2
Var_Expr := Block_Var_2;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 1951 '<FieldAccess for array_def at ast.py:14250>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14250
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.array_def'
Fld_4 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Array_Def (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 1951
Scope_Result := Fld_4;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Array_Type_Def;
end if;
Block_Var_3 := Result_Var; 
--# bind array_type_def Block_Var_3
--# expr-start 1969 '<FieldAccess for resolve_names_internal_with_eq at ast.py:14258>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14258
--# expr-start 1952 '<FieldAccess for spec at ast.py:14258>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14258
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_For_Loop_Spec (Node => Ent.Node.Iterated_Assoc_F_Spec, Info => Ent.Info);
--# expr-done 1952
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1968 '<If at ast.py:14259>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14259
--# expr-start 1955 '<IsA IterType.In at ast.py:14259>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14259
--# expr-start 1954 '<FieldAccess for loop_type at ast.py:14259>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14259
--# expr-start 1953 '<FieldAccess for spec at ast.py:14259>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14259
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Self.Iterated_Assoc_F_Spec;
--# expr-done 1953
      if Fld_14 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Fld_14.For_Loop_Spec_F_Loop_Type;
--# expr-done 1954
Is_A := Fld_15 /= null 
and then Fld_15.Kind in Ada_Iter_Type_In_Range; 
--# expr-done 1955
if Is_A then
   --# expr-start 1967 '<And at ast.py:14260>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14260
--# expr-start 1962 '<Bind at ast.py:14260>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14260
--# expr-start 1959 '<FieldAccess for expected_type_var at ast.py:14260>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14260
--# expr-start 1958 '<Cast to Expr at ast.py:14260>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14260
--# expr-start 1957 '<FieldAccess for iter_expr at ast.py:14260>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14260
--# expr-start 1956 '<FieldAccess for spec at ast.py:14260>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14260
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_For_Loop_Spec (Node => Ent.Node.Iterated_Assoc_F_Spec, Info => Ent.Info);
--# expr-done 1956
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity (Node => Fld_6.Node.For_Loop_Spec_F_Iter_Expr, Info => Fld_6.Info);
--# expr-done 1957
   if Fld_7.Node = null
      or else Fld_7.Node.Kind in Ada_Expr
   then
      Cast_Result_1 := Create_Internal_Entity_Expr
        (Node => Fld_7.Node,
         Info => Fld_7.Info);
   else
         Cast_Result_1 := No_Entity_Expr;
   end if;
--# expr-done 1958
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Cast_Result_1.Node);
--# expr-done 1959
Fld_8.Value := No_Entity;
Entity_Vars.Reset (Fld_8);
--# expr-start 1961 '<FieldAccess for index_type at ast.py:14261>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14261
      if Block_Var_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 1960 '<FieldAccess for rank at ast.py:14261>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14261
Fld_9 := Block_Var_1.Rank;
--# expr-done 1960
--# property-call-start ArrayTypeDef.index_type
Fld_10 := Libadalang.Implementation.Array_Type_Def_P_Index_Type (Node => Block_Var_3.Node, Dim => Fld_9, Origin => Origin, E_Info => Block_Var_3.Info);
--# end
--# expr-done 1961
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_10.Node,
         Info => Fld_10.Info);
Bind_Result := Solver.Create_Assign (Fld_8, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14260") else null)); 
--# expr-done 1962
--# expr-start 1966 '<FieldAccess for matches_expected_type at ast.py:14262>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14262
--# expr-start 1965 '<Cast to Expr at ast.py:14262>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14262
--# expr-start 1964 '<FieldAccess for iter_expr at ast.py:14262>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14262
--# expr-start 1963 '<FieldAccess for spec at ast.py:14262>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14262
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Create_Internal_Entity_For_Loop_Spec (Node => Ent.Node.Iterated_Assoc_F_Spec, Info => Ent.Info);
--# expr-done 1963
      if Fld_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Create_Internal_Entity (Node => Fld_11.Node.For_Loop_Spec_F_Iter_Expr, Info => Fld_11.Info);
--# expr-done 1964
   if Fld_12.Node = null
      or else Fld_12.Node.Kind in Ada_Expr
   then
      Cast_Result_3 := Create_Internal_Entity_Expr
        (Node => Fld_12.Node,
         Info => Fld_12.Info);
   else
         Cast_Result_3 := No_Entity_Expr;
   end if;
--# expr-done 1965
      if Cast_Result_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_13 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Cast_Result_3.Node, Origin => Origin);
--# end
--# expr-done 1966
And_Pred := Create_And (Bind_Result, Fld_13, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14260") else null)); 
--# expr-done 1967
   If_Result_1 := And_Pred;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14263") else null)); 
   If_Result_1 := True_Rel;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 1968
--# property-call-start AdaNode.resolve_names_internal_with_eq
Fld_16 := Libadalang.Implementation.Ada_Node_P_Resolve_Names_Internal_With_Eq (Node => Fld_5.Node, Additional_Equation => If_Result_1, Env => Env, Origin => Origin, E_Info => Fld_5.Info);
--# end
--# expr-done 1969
Block_Var_4 := Fld_16; 
--# bind spec_success Block_Var_4
--# expr-start 1986 '<If at ast.py:14266>' If_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14266
if Block_Var_4 then
   --# expr-start 1985 '<If at ast.py:14268>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14268
--# expr-start 1970 '<IsNull at ast.py:14271>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14271
Is_Equal_1 := Block_Var_1 = No_Multidim_Aggregate_Info; 
--# expr-done 1970
if Is_Equal_1 then
   If_Result_2 := True;
else
   --# expr-start 1974 '<Eq at ast.py:14272>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14272
--# expr-start 1971 '<FieldAccess for rank at ast.py:14272>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14272
Fld_24 := Block_Var_1.Rank;
--# expr-done 1971
--# expr-start 1973 '<Arithmetic - at ast.py:14272>' Arith_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14272
--# expr-start 1972 '<FieldAccess for array_ndims at ast.py:14272>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14272
      if Block_Var_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.array_ndims'
Fld_25 := Libadalang.Implementation.Dispatcher_Type_Def_P_Array_Ndims (Node => Block_Var_3.Node, Origin => Origin, E_Info => Block_Var_3.Info);
--# end
--# expr-done 1972
Arith_Result := (Fld_25 - 1); 
--# expr-done 1973
Is_Equal_2 := Fld_24 = Arith_Result; 
--# expr-done 1974
   If_Result_2 := Is_Equal_2;
end if;
if If_Result_2 then
   --# expr-start 1984 '<And at ast.py:14275>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14275
--# expr-start 1981 '<And at ast.py:14275>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14275
--# expr-start 1976 '<FieldAccess for sub_equation at ast.py:14275>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14275
--# expr-start 1975 '<FieldAccess for expr at ast.py:14275>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14275
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_17 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1975
      if Fld_17.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_18 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_17.Node, Env => Env, Origin => Origin, E_Info => Fld_17.Info);
--# end
--# expr-done 1976
--# expr-start 1980 '<Bind at ast.py:14276>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14276
--# expr-start 1978 '<FieldAccess for expected_type_var at ast.py:14276>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14276
--# expr-start 1977 '<FieldAccess for expr at ast.py:14276>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14276
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_19 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1977
      if Fld_19.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_20 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_19.Node);
--# expr-done 1978
Fld_20.Value := No_Entity;
Entity_Vars.Reset (Fld_20);
--# expr-start 1979 '<FieldAccess for comp_type at ast.py:14276>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14276
      if Block_Var_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ArrayTypeDef.comp_type
Fld_21 := Libadalang.Implementation.Array_Type_Def_P_Comp_Type (Node => Block_Var_3.Node, Origin => Origin, E_Info => Block_Var_3.Info);
--# end
--# expr-done 1979
      Cast_Result_4 := Create_Internal_Entity
        (Node => Fld_21.Node,
         Info => Fld_21.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_20, Cast_Result_4, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14276") else null)); 
--# expr-done 1980
And_Pred_1 := Create_And (Fld_18, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14275") else null)); 
--# expr-done 1981
--# expr-start 1983 '<FieldAccess for matches_expected_type at ast.py:14277>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14277
--# expr-start 1982 '<FieldAccess for expr at ast.py:14277>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14277
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_22 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 1982
      if Fld_22.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_23 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_22.Node, Origin => Origin);
--# end
--# expr-done 1983
And_Pred_2 := Create_And (And_Pred_1, Fld_23, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14275") else null)); 
--# expr-done 1984
   If_Result_3 := And_Pred_2;
else
   True_Rel_1 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14281") else null)); 
   If_Result_3 := True_Rel_1;
end if;
      Inc_Ref (If_Result_3);
--# expr-done 1985
   If_Result_4 := If_Result_3;
else
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14283") else null)); 
   If_Result_4 := False_Rel;
end if;
      Inc_Ref (If_Result_4);
--# expr-done 1986
Let_Result := If_Result_4; Inc_Ref (Let_Result);
--# end
--# expr-done 1987
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1033;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1033;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Iterated_Assoc_P_Xref_Equation;
--# end
--# property-start IteratedAssoc.xref_equation_for_reduce /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14287
pragma Warnings (Off, "is not referenced");
 function Iterated_Assoc_P_Xref_Equation_For_Reduce
  (Node : Bare_Iterated_Assoc
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Iterated_Assoc := Bare_Iterated_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Iterated_Assoc :=
     Internal_Entity_Iterated_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1034 with Inline_Always;
      Block_Var : Logic_Var;
Block_Var_1 : Boolean;
Fld : Internal_Entity;
Cast_Result : Internal_Entity_Value_Sequence;
Fld_1 : Internal_Entity_Iterated_Assoc;
Fld_2 : Internal_Entity_Expr;
Fld_3 : Logic_Var;
Fld_4 : Internal_Entity_For_Loop_Spec;
Fld_5 : Internal_Entity_For_Loop_Spec;
Fld_6 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Expr;
Fld_7 : Logic_Var;
Bind_Result : Logic_Equation;
Fld_8 : Internal_Entity_For_Loop_Spec;
Fld_9 : Internal_Entity;
Cast_Result_2 : Internal_Entity_Expr;
Fld_10 : Logic_Equation;
And_Pred : Logic_Equation;
True_Rel : Logic_Equation;
Fld_11 : Bare_For_Loop_Spec;
Fld_12 : Bare_Iter_Type;
Is_A : Boolean;
If_Result : Logic_Equation;
Fld_13 : Boolean;
Fld_14 : Internal_Entity_Expr;
Fld_15 : Logic_Equation;
Fld_16 : Internal_Entity_Expr;
Fld_17 : Logic_Var;
Bind_Result_1 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_18 : Internal_Entity_Expr;
Fld_19 : Logic_Equation;
And_Pred_2 : Logic_Equation;
False_Rel : Logic_Equation;
If_Result_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1034 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_10);
                     Dec_Ref (And_Pred);
                     Dec_Ref (True_Rel);
                     Dec_Ref (If_Result);
                     Dec_Ref (Fld_15);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_19);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (False_Rel);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1034;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2019 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 1992 '<FieldAccess for expected_type_var at ast.py:14294>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14294
--# expr-start 1991 '<FieldAccess for expr at ast.py:14294>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14294
--# expr-start 1990 '<FieldAccess for iter_assoc at ast.py:14294>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14294
--# expr-start 1989 '<Cast to ValueSequence at ast.py:14294>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14294
--# expr-start 1988 '<parent at ast.py:14294>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14294
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 1988
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Value_Sequence_Range
   then
      Cast_Result := Create_Internal_Entity_Value_Sequence
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Value_Sequence;
   end if;
--# expr-done 1989
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Iterated_Assoc (Node => Cast_Result.Node.Value_Sequence_F_Iter_Assoc, Info => Cast_Result.Info);
--# expr-done 1990
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 1991
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_2.Node);
--# expr-done 1992
Block_Var := Fld_3; 
--# bind expected_typ Block_Var
--# expr-start 2008 '<FieldAccess for resolve_names_internal_with_eq at ast.py:14298>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14298
--# expr-start 1993 '<FieldAccess for spec at ast.py:14298>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14298
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_For_Loop_Spec (Node => Ent.Node.Iterated_Assoc_F_Spec, Info => Ent.Info);
--# expr-done 1993
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2007 '<If at ast.py:14299>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14299
--# expr-start 1996 '<IsA IterType.In at ast.py:14299>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14299
--# expr-start 1995 '<FieldAccess for loop_type at ast.py:14299>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14299
--# expr-start 1994 '<FieldAccess for spec at ast.py:14299>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14299
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Self.Iterated_Assoc_F_Spec;
--# expr-done 1994
      if Fld_11 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Fld_11.For_Loop_Spec_F_Loop_Type;
--# expr-done 1995
Is_A := Fld_12 /= null 
and then Fld_12.Kind in Ada_Iter_Type_In_Range; 
--# expr-done 1996
if Is_A then
   --# expr-start 2006 '<And at ast.py:14300>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14300
--# expr-start 2001 '<Bind at ast.py:14300>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14300
--# expr-start 2000 '<FieldAccess for expected_type_var at ast.py:14300>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14300
--# expr-start 1999 '<Cast to Expr at ast.py:14300>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14300
--# expr-start 1998 '<FieldAccess for iter_expr at ast.py:14300>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14300
--# expr-start 1997 '<FieldAccess for spec at ast.py:14300>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14300
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_For_Loop_Spec (Node => Ent.Node.Iterated_Assoc_F_Spec, Info => Ent.Info);
--# expr-done 1997
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity (Node => Fld_5.Node.For_Loop_Spec_F_Iter_Expr, Info => Fld_5.Info);
--# expr-done 1998
   if Fld_6.Node = null
      or else Fld_6.Node.Kind in Ada_Expr
   then
      Cast_Result_1 := Create_Internal_Entity_Expr
        (Node => Fld_6.Node,
         Info => Fld_6.Info);
   else
         Cast_Result_1 := No_Entity_Expr;
   end if;
--# expr-done 1999
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Cast_Result_1.Node);
--# expr-done 2000
Fld_7.Value := No_Entity;
Entity_Vars.Reset (Fld_7);
Block_Var.Value := No_Entity;
Entity_Vars.Reset (Block_Var);
Bind_Result := Solver.Create_Unify (Fld_7, Block_Var, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14300") else null)); 
--# expr-done 2001
--# expr-start 2005 '<FieldAccess for matches_expected_type at ast.py:14302>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14302
--# expr-start 2004 '<Cast to Expr at ast.py:14302>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14302
--# expr-start 2003 '<FieldAccess for iter_expr at ast.py:14302>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14302
--# expr-start 2002 '<FieldAccess for spec at ast.py:14302>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14302
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_For_Loop_Spec (Node => Ent.Node.Iterated_Assoc_F_Spec, Info => Ent.Info);
--# expr-done 2002
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity (Node => Fld_8.Node.For_Loop_Spec_F_Iter_Expr, Info => Fld_8.Info);
--# expr-done 2003
   if Fld_9.Node = null
      or else Fld_9.Node.Kind in Ada_Expr
   then
      Cast_Result_2 := Create_Internal_Entity_Expr
        (Node => Fld_9.Node,
         Info => Fld_9.Info);
   else
         Cast_Result_2 := No_Entity_Expr;
   end if;
--# expr-done 2004
      if Cast_Result_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_10 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Cast_Result_2.Node, Origin => Origin);
--# end
--# expr-done 2005
And_Pred := Create_And (Bind_Result, Fld_10, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14300") else null)); 
--# expr-done 2006
   If_Result := And_Pred;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14303") else null)); 
   If_Result := True_Rel;
end if;
      Inc_Ref (If_Result);
--# expr-done 2007
--# property-call-start AdaNode.resolve_names_internal_with_eq
Fld_13 := Libadalang.Implementation.Ada_Node_P_Resolve_Names_Internal_With_Eq (Node => Fld_4.Node, Additional_Equation => If_Result, Env => Env, Origin => Origin, E_Info => Fld_4.Info);
--# end
--# expr-done 2008
Block_Var_1 := Fld_13; 
--# bind spec_success Block_Var_1
--# expr-start 2018 '<If at ast.py:14306>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14306
if Block_Var_1 then
   --# expr-start 2017 '<And at ast.py:14307>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14307
--# expr-start 2014 '<And at ast.py:14307>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14307
--# expr-start 2010 '<FieldAccess for sub_equation at ast.py:14307>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14307
--# expr-start 2009 '<FieldAccess for expr at ast.py:14307>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14307
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_14 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2009
      if Fld_14.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_15 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_14.Node, Env => Env, Origin => Origin, E_Info => Fld_14.Info);
--# end
--# expr-done 2010
--# expr-start 2013 '<Bind at ast.py:14308>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14308
--# expr-start 2012 '<FieldAccess for expected_type_var at ast.py:14308>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14308
--# expr-start 2011 '<FieldAccess for expr at ast.py:14308>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14308
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_16 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2011
      if Fld_16.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_17 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_16.Node);
--# expr-done 2012
Fld_17.Value := No_Entity;
Entity_Vars.Reset (Fld_17);
Block_Var.Value := No_Entity;
Entity_Vars.Reset (Block_Var);
Bind_Result_1 := Solver.Create_Unify (Fld_17, Block_Var, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14308") else null)); 
--# expr-done 2013
And_Pred_1 := Create_And (Fld_15, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14307") else null)); 
--# expr-done 2014
--# expr-start 2016 '<FieldAccess for matches_expected_type at ast.py:14309>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14309
--# expr-start 2015 '<FieldAccess for expr at ast.py:14309>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14309
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_18 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2015
      if Fld_18.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_19 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_18.Node, Origin => Origin);
--# end
--# expr-done 2016
And_Pred_2 := Create_And (And_Pred_1, Fld_19, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14307") else null)); 
--# expr-done 2017
   If_Result_1 := And_Pred_2;
else
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14310") else null)); 
   If_Result_1 := False_Rel;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 2018
Let_Result := If_Result_1; Inc_Ref (Let_Result);
--# end
--# expr-done 2019
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1034;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1034;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Iterated_Assoc_P_Xref_Equation_For_Reduce;
--# end
   --
   --  Primitives for Bare_Param_Assoc
   --
      procedure Initialize_Fields_For_Param_Assoc
        (Self : Bare_Param_Assoc
         ; Param_Assoc_F_Designator : Bare_Ada_Node
         ; Param_Assoc_F_R_Expr : Bare_Expr
        ) is
      begin
            Self.Param_Assoc_F_Designator := Param_Assoc_F_Designator;
            Self.Param_Assoc_F_R_Expr := Param_Assoc_F_R_Expr;
      end Initialize_Fields_For_Param_Assoc;
   function Param_Assoc_F_Designator
     (Node : Bare_Param_Assoc) return Bare_Ada_Node
   is
   begin
         return Node.Param_Assoc_F_Designator;
   end;
   function Param_Assoc_F_R_Expr
     (Node : Bare_Param_Assoc) return Bare_Expr
   is
   begin
         return Node.Param_Assoc_F_R_Expr;
   end;
--# property-start ParamAssoc.expr /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13938
pragma Warnings (Off, "is not referenced");
 function Param_Assoc_P_Expr
  (Node : Bare_Param_Assoc
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Expr
is
   Self : Bare_Param_Assoc := Bare_Param_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Param_Assoc :=
     Internal_Entity_Param_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Expr;
      Fld : Internal_Entity_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2020 '<FieldAccess for r_expr at ast.py:13938>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13938
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Param_Assoc_F_R_Expr, Info => Ent.Info);
--# expr-done 2020
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Param_Assoc_P_Expr;
--# end
--# property-start ParamAssoc.names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13939
pragma Warnings (Off, "is not referenced");
 function Param_Assoc_P_Names
  (Node : Bare_Param_Assoc
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Param_Assoc := Bare_Param_Assoc (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1013 with Inline_Always;
      Fld : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
Fld_1 : Bare_Ada_Node;
Is_Null : Boolean;
If_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1013 is
            begin
                     Dec_Ref (Singleton);
                     Dec_Ref (If_Result);
            end Finalizer_Scope_1013;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2024 '<If at ast.py:13939>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13939
--# expr-start 2022 '<IsNull at ast.py:13939>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13939
--# expr-start 2021 '<FieldAccess for designator at ast.py:13939>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13939
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Self.Param_Assoc_F_Designator;
--# expr-done 2021
Is_Null := Fld_1 = null; 
--# expr-done 2022
if Is_Null then
   If_Result := No_Bare_Ada_Node_Array_Type;
else
   --# expr-start 2023 '<FieldAccess for designator at ast.py:13940>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13940
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Param_Assoc_F_Designator;
--# expr-done 2023
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
   If_Result := Singleton;
end if;
      Inc_Ref (If_Result);
--# expr-done 2024
      Property_Result := If_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1013;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1013;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Param_Assoc_P_Names;
--# end
--# property-start ParamAssoc.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13942
pragma Warnings (Off, "is not referenced");
 function Param_Assoc_P_Xref_Entry_Point
  (Node : Bare_Param_Assoc
  )
   return Boolean
is
   Self : Bare_Param_Assoc := Bare_Param_Assoc (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2025 '<FieldAccess for is_static_attribute_assoc at ast.py:13942>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13942
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ParamAssoc.is_static_attribute_assoc
Fld := Libadalang.Implementation.Param_Assoc_P_Is_Static_Attribute_Assoc (Node => Self);
--# end
--# expr-done 2025
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Param_Assoc_P_Xref_Entry_Point;
--# end
--# property-start ParamAssoc.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13943
pragma Warnings (Off, "is not referenced");
 function Param_Assoc_P_Xref_Equation
  (Node : Bare_Param_Assoc
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Param_Assoc := Bare_Param_Assoc (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Param_Assoc :=
     Internal_Entity_Param_Assoc'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1015 with Inline_Always;
      Fld : Internal_Entity_Expr;
Fld_1 : Logic_Equation;
False_Rel : Logic_Equation;
Fld_2 : Boolean;
If_Result : Logic_Equation;
            procedure Finalizer_Scope_1015 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (False_Rel);
                     Dec_Ref (If_Result);
            end Finalizer_Scope_1015;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2029 '<If at ast.py:13943>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13943
--# expr-start 2026 '<FieldAccess for xref_entry_point at ast.py:13944>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13944
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_entry_point'
Fld_2 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Entry_Point (Node => Self);
--# end
--# expr-done 2026
if Fld_2 then
   --# expr-start 2028 '<FieldAccess for sub_equation at ast.py:13945>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13945
--# expr-start 2027 '<FieldAccess for expr at ast.py:13945>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13945
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2027
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 2028
   If_Result := Fld_1;
else
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13946") else null)); 
   If_Result := False_Rel;
end if;
      Inc_Ref (If_Result);
--# expr-done 2029
      Property_Result := If_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1015;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1015;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Param_Assoc_P_Xref_Equation;
--# end
--# property-start ParamAssoc.is_static_attribute_assoc /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13950
pragma Warnings (Off, "is not referenced");
 function Param_Assoc_P_Is_Static_Attribute_Assoc
  (Node : Bare_Param_Assoc
  )
   return Boolean
is
   Self : Bare_Param_Assoc := Bare_Param_Assoc (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Var_Expr : Bare_Attribute_Ref;
Fld : Bare_Ada_Node;
Fld_1 : Bare_Ada_Node;
Cast_Result : Bare_Attribute_Ref;
Fld_2 : Bare_Identifier;
Fld_3 : Symbol_Type;
Sym : Symbol_Type;
Is_Equal : Boolean;
Fld_4 : Bare_Identifier;
Fld_5 : Symbol_Type;
Sym_1 : Symbol_Type;
Is_Equal_1 : Boolean;
If_Result : Boolean;
Fld_6 : Bare_Identifier;
Fld_7 : Symbol_Type;
Sym_2 : Symbol_Type;
Is_Equal_2 : Boolean;
If_Result_1 : Boolean;
Fld_8 : Bare_Identifier;
Fld_9 : Symbol_Type;
Sym_3 : Symbol_Type;
Is_Equal_3 : Boolean;
If_Result_2 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2049 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2032 '<Cast to AttributeRef at ast.py:13951>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13951
--# expr-start 2031 '<parent at ast.py:13951>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13951
--# expr-start 2030 '<parent at ast.py:13951>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13951
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 2030
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Parent; 
--# expr-done 2031
   if Fld_1 = null
      or else Fld_1.Kind in Ada_Attribute_Ref_Range
   then
      Cast_Result := Fld_1;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 2032
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind ar Var_Expr
--# expr-start 2036 '<Eq at ast.py:13952>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13952
--# expr-start 2034 '<FieldAccess for name_symbol at ast.py:13952>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13952
--# expr-start 2033 '<FieldAccess for attribute at ast.py:13952>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13952
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Var_Expr.Attribute_Ref_F_Attribute;
--# expr-done 2033
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_2);
--# end
--# expr-done 2034
--# expr-start 2035 '<Symbol First at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_First); 
--# expr-done 2035
Is_Equal := Fld_3 = Sym; 
--# expr-done 2036
if Is_Equal then
   If_Result := True;
else
   --# expr-start 2040 '<Eq at ast.py:13952>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13952
--# expr-start 2038 '<FieldAccess for name_symbol at ast.py:13952>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13952
--# expr-start 2037 '<FieldAccess for attribute at ast.py:13952>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13952
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Var_Expr.Attribute_Ref_F_Attribute;
--# expr-done 2037
      if Fld_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_5 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_4);
--# end
--# expr-done 2038
--# expr-start 2039 '<Symbol Last at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Last); 
--# expr-done 2039
Is_Equal_1 := Fld_5 = Sym_1; 
--# expr-done 2040
   If_Result := Is_Equal_1;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 2044 '<Eq at ast.py:13952>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13952
--# expr-start 2042 '<FieldAccess for name_symbol at ast.py:13952>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13952
--# expr-start 2041 '<FieldAccess for attribute at ast.py:13952>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13952
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Var_Expr.Attribute_Ref_F_Attribute;
--# expr-done 2041
      if Fld_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_7 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_6);
--# end
--# expr-done 2042
--# expr-start 2043 '<Symbol Range at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Range); 
--# expr-done 2043
Is_Equal_2 := Fld_7 = Sym_2; 
--# expr-done 2044
   If_Result_1 := Is_Equal_2;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
   --# expr-start 2048 '<Eq at ast.py:13952>' Is_Equal_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13952
--# expr-start 2046 '<FieldAccess for name_symbol at ast.py:13952>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13952
--# expr-start 2045 '<FieldAccess for attribute at ast.py:13952>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13952
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Var_Expr.Attribute_Ref_F_Attribute;
--# expr-done 2045
      if Fld_8 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_9 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_8);
--# end
--# expr-done 2046
--# expr-start 2047 '<Symbol Length at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Length); 
--# expr-done 2047
Is_Equal_3 := Fld_9 = Sym_3; 
--# expr-done 2048
   If_Result_2 := Is_Equal_3;
end if;
Scope_Result := If_Result_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 2049
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Param_Assoc_P_Is_Static_Attribute_Assoc;
--# end
   --
   --  Primitives for Bare_Basic_Decl
   --
   function Basic_Decl_F_Aspects
     (Node : Bare_Basic_Decl) return Bare_Aspect_Spec
   is
         Kind : constant Ada_Basic_Decl := Node.Kind;
   begin
         case Kind is
               when Ada_Error_Decl =>
                     return No_Bare_Ada_Node;
               when Ada_Discriminant_Spec =>
         return Node.Discriminant_Spec_F_Aspects;
               when Ada_Component_Decl =>
         return Node.Component_Decl_F_Aspects;
               when Ada_Classwide_Type_Decl =>
                     return No_Bare_Ada_Node;
               when Ada_Type_Decl | Ada_Anonymous_Type_Decl | Ada_Synth_Anonymous_Type_Decl =>
         return Node.Type_Decl_F_Aspects;
               when Ada_Subtype_Decl =>
         return Node.Subtype_Decl_F_Aspects;
               when Ada_Discrete_Base_Subtype_Decl =>
                     return No_Bare_Ada_Node;
               when Ada_Task_Type_Decl | Ada_Single_Task_Type_Decl =>
         return Node.Task_Type_Decl_F_Aspects;
               when Ada_Protected_Type_Decl =>
         return Node.Protected_Type_Decl_F_Aspects;
               when Ada_Param_Spec =>
         return Node.Param_Spec_F_Aspects;
               when Ada_Subp_Decl =>
         return Node.Subp_Decl_F_Aspects;
               when Ada_Abstract_Subp_Decl =>
         return Node.Abstract_Subp_Decl_F_Aspects;
               when Ada_Single_Task_Decl =>
                     return No_Bare_Ada_Node;
               when Ada_Single_Protected_Decl =>
         return Node.Single_Protected_Decl_F_Aspects;
               when Ada_Number_Decl =>
                     return No_Bare_Ada_Node;
               when Ada_Object_Decl | Ada_Extended_Return_Stmt_Object_Decl | Ada_No_Type_Object_Renaming_Decl =>
         return Node.Object_Decl_F_Aspects;
               when Ada_Anonymous_Expr_Decl =>
                     return No_Bare_Ada_Node;
               when Ada_Generic_Package_Internal | Ada_Package_Decl =>
         return Node.Base_Package_Decl_F_Aspects;
               when Ada_Exception_Decl =>
         return Node.Exception_Decl_F_Aspects;
               when Ada_Generic_Subp_Instantiation =>
         return Node.Generic_Subp_Instantiation_F_Aspects;
               when Ada_Generic_Package_Instantiation =>
         return Node.Generic_Package_Instantiation_F_Aspects;
               when Ada_Package_Renaming_Decl =>
         return Node.Package_Renaming_Decl_F_Aspects;
               when Ada_Generic_Package_Renaming_Decl =>
         return Node.Generic_Package_Renaming_Decl_F_Aspects;
               when Ada_Generic_Subp_Renaming_Decl =>
         return Node.Generic_Subp_Renaming_Decl_F_Aspects;
               when Ada_Abstract_Formal_Subp_Decl | Ada_Concrete_Formal_Subp_Decl =>
         return Node.Formal_Subp_Decl_F_Aspects;
               when Ada_Generic_Formal_Obj_Decl | Ada_Generic_Formal_Package | Ada_Generic_Formal_Subp_Decl | Ada_Generic_Formal_Type_Decl =>
                     return No_Bare_Ada_Node;
               when Ada_Generic_Subp_Internal =>
         return Node.Generic_Subp_Internal_F_Aspects;
               when Ada_Generic_Subp_Decl =>
                     return No_Bare_Ada_Node;
               when Ada_Generic_Package_Decl =>
                     return No_Bare_Ada_Node;
               when Ada_Enum_Literal_Decl | Ada_Synthetic_Char_Enum_Lit =>
                     return No_Bare_Ada_Node;
               when Ada_Synthetic_Formal_Param_Decl =>
                     return No_Bare_Ada_Node;
               when Ada_Synthetic_Subp_Decl =>
                     return No_Bare_Ada_Node;
               when Ada_Entry_Decl =>
         return Node.Entry_Decl_F_Aspects;
               when Ada_For_Loop_Var_Decl =>
                     return No_Bare_Ada_Node;
               when Ada_Expr_Function =>
         return Node.Expr_Function_F_Aspects;
               when Ada_Null_Subp_Decl =>
         return Node.Null_Subp_Decl_F_Aspects;
               when Ada_Subp_Renaming_Decl =>
         return Node.Subp_Renaming_Decl_F_Aspects;
               when Ada_Subp_Body =>
         return Node.Subp_Body_F_Aspects;
               when Ada_Exception_Handler =>
                     return No_Bare_Ada_Node;
               when Ada_Label_Decl =>
                     return No_Bare_Ada_Node;
               when Ada_Named_Stmt_Decl =>
                     return No_Bare_Ada_Node;
               when Ada_Package_Body =>
         return Node.Package_Body_F_Aspects;
               when Ada_Task_Body =>
         return Node.Task_Body_F_Aspects;
               when Ada_Protected_Body =>
         return Node.Protected_Body_F_Aspects;
               when Ada_Entry_Body =>
         return Node.Entry_Body_F_Aspects;
               when Ada_Entry_Index_Spec =>
         return Node.Entry_Index_Spec_F_Aspects;
               when Ada_Protected_Body_Stub =>
         return Node.Protected_Body_Stub_F_Aspects;
               when Ada_Subp_Body_Stub =>
         return Node.Subp_Body_Stub_F_Aspects;
               when Ada_Package_Body_Stub =>
         return Node.Package_Body_Stub_F_Aspects;
               when Ada_Task_Body_Stub =>
         return Node.Task_Body_Stub_F_Aspects;
               when Ada_Incomplete_Type_Decl | Ada_Incomplete_Tagged_Type_Decl =>
                     return No_Bare_Ada_Node;
               when Ada_Abstract_State_Decl =>
         return Node.Abstract_State_Decl_F_Aspects;
         end case;
   end;
--# property-start BasicDecl.env_hook_basic_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1673
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Env_Hook_Basic_Decl
  (Node : Bare_Basic_Decl
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
            procedure Finalizer_Scope_1813 with Inline_Always;
      Var_Expr : Internal_Entity_Dotted_Name;
Var_Expr_1 : Internal_Unit;
As_Entity : Internal_Entity_Basic_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Internal_Entity_Name;
Cast_Result : Internal_Entity_Dotted_Name;
Fld_2 : Internal_Entity_Name;
Fld_3 : Symbol_Type_Array_Access;
Is_A : Boolean;
Not_Val : Boolean;
Fld_4 : Internal_Unit;
Scope_Result : Boolean;
Result_Var : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
Is_A_1 : Boolean;
If_Result_1 : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_1813 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_1813;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2061 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2060 '<If at ast.py:1677>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1677
--# expr-start 2050 '<IsA PackageDecl, BasicSubpDecl, PackageRenamingDecl, GenericPackageDecl, GenericPackageInstantiation, GenericSubpInstantiation, GenericSubpDecl, SubpBody at ast.py:1680>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1680
Is_A_1 := Self /= null 
and then Self.Kind in Ada_Package_Decl_Range | Ada_Basic_Subp_Decl | Ada_Package_Renaming_Decl_Range | Ada_Generic_Package_Decl_Range | Ada_Generic_Package_Instantiation_Range | Ada_Generic_Subp_Instantiation_Range | Ada_Generic_Subp_Decl_Range | Ada_Subp_Body_Range; 
--# expr-done 2050
if Is_A_1 then
--# expr-start 2054 '<Cast to DottedName at ast.py:1685>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1685
--# expr-start 2053 '<FieldAccess for name at ast.py:1685>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1685
--# expr-start 2052 '<FieldAccess for defining_name at ast.py:1685>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1685
--# expr-start 2051 '<as_bare_entity at ast.py:1685>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1685
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Decl;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 2051
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 2052
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Fld.Node.Defining_Name_F_Name, Info => Fld.Info);
--# expr-done 2053
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Dotted_Name_Range
   then
      Cast_Result := Create_Internal_Entity_Dotted_Name
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         Cast_Result := No_Entity_Dotted_Name;
   end if;
--# expr-done 2054
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Dotted_Name then
   --# scope-start
   --# bind dn Var_Expr
--# expr-start 2059 '<FieldAccess for get_unit at ast.py:1687>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1687
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2056 '<FieldAccess for as_symbol_array at ast.py:1687>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1687
--# expr-start 2055 '<FieldAccess for prefix at ast.py:1687>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1687
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Var_Expr.Node.Dotted_Name_F_Prefix, Info => Var_Expr.Info);
--# expr-done 2055
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.as_symbol_array'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_As_Symbol_Array (Node => Fld_2.Node);
--# end
--# expr-done 2056
--# expr-start 2058 '<Not at ast.py:1690>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1690
--# expr-start 2057 '<IsA SubpBody at ast.py:1690>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1690
Is_A := Self /= null 
and then Self.Kind in Ada_Subp_Body_Range; 
--# expr-done 2057
Not_Val := not (Is_A); 
--# expr-done 2058
Fld_4 := Libadalang.Implementation.Extensions.Ada_Node_P_Get_Unit (Node => Self, Name => Fld_3, Kind => Unit_Specification, Load_If_Needed => True, Not_Found_Is_Error => Not_Val);
--# expr-done 2059
Var_Expr_1 := Fld_4;
if Var_Expr_1 /= null then
   --# scope-start
   --# bind _ Var_Expr_1
Scope_Result := False;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Scope_Result_1 := Result_Var;
   --# end
      Finalizer_Scope_1813;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
   If_Result_1 := Result_Var_1;
else
   If_Result_1 := False;
end if;
--# expr-done 2060
Let_Result := If_Result_1; 
--# end
--# expr-done 2061
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1813;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Env_Hook_Basic_Decl;
--# end
--# property-start BasicDecl.populate_body_unit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1697
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Populate_Body_Unit
  (Node : Bare_Basic_Decl
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
            procedure Finalizer_Scope_122 with Inline_Always;
      Var_Expr : Internal_Unit;
As_Entity : Internal_Entity_Basic_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Internal_Entity_Name;
Fld_2 : Symbol_Type_Array_Access;
Fld_3 : Internal_Unit;
Scope_Result : Boolean;
Result_Var : Boolean;
Fld_4 : Boolean;
If_Result_1 : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_122 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_122;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2069 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2068 '<If at ast.py:1703>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1703
--# expr-start 2062 '<FieldAccess for is_library_item at ast.py:1704>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1704
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_library_item'
Fld_4 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Library_Item (Node => Self);
--# end
--# expr-done 2062
if Fld_4 then
--# expr-start 2067 '<FieldAccess for get_unit at ast.py:1705>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1705
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2066 '<FieldAccess for as_symbol_array at ast.py:1706>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1706
--# expr-start 2065 '<FieldAccess for name at ast.py:1706>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1706
--# expr-start 2064 '<FieldAccess for defining_name at ast.py:1706>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1706
--# expr-start 2063 '<as_bare_entity at ast.py:1706>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1706
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Decl;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 2063
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 2064
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Fld.Node.Defining_Name_F_Name, Info => Fld.Info);
--# expr-done 2065
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.as_symbol_array'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_As_Symbol_Array (Node => Fld_1.Node);
--# end
--# expr-done 2066
Fld_3 := Libadalang.Implementation.Extensions.Ada_Node_P_Get_Unit (Node => Self, Name => Fld_2, Kind => Unit_Body, Load_If_Needed => True, Not_Found_Is_Error => False);
--# expr-done 2067
Var_Expr := Fld_3;
if Var_Expr /= null then
   --# scope-start
   --# bind _ Var_Expr
Scope_Result := False;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
   If_Result_1 := Result_Var;
else
   If_Result_1 := False;
end if;
--# expr-done 2068
Let_Result := If_Result_1; 
--# end
--# expr-done 2069
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_122;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_122;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Populate_Body_Unit;
--# end
--# property-start BasicDecl.has_top_level_env_name_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1715
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Has_Top_Level_Env_Name_Impl
  (Node : Bare_Basic_Decl
      ; Allow_Bodies : Boolean
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind allow_bodies Allow_Bodies
   Property_Result : Boolean;
            procedure Finalizer_Scope_123 with Inline_Always;
      Block_Var : Boolean;
Block_Var_1 : Boolean;
Var_Expr : Bare_Ada_Node;
Var_Expr_1 : Bare_Basic_Decl;
Var_Expr_2 : Bare_Basic_Decl;
Is_A : Boolean;
Is_A_1 : Boolean;
Fld : Boolean;
If_Result : Boolean;
If_Result_1 : Boolean;
Fld_1 : Lexical_Env := Empty_Env;
Env_Node : Bare_Ada_Node;
Cast_Expr : Bare_Ada_Node;
Cast_Result : Bare_Basic_Decl;
Is_A_2 : Boolean;
Not_Val : Boolean;
If_Result_2 : Boolean;
Is_A_3 : Boolean;
Not_Val_1 : Boolean;
If_Result_3 : Boolean;
Fld_2 : Boolean;
Cast_Expr_1 : Bare_Basic_Decl;
Cast_Result_1 : Bare_Basic_Decl;
Is_Equal : Boolean;
If_Result_4 : Boolean;
Scope_Result : Boolean;
Is_A_4 : Boolean;
Fld_3 : Bare_Ada_Node;
Cast_Result_2 : Bare_Basic_Decl;
Fld_4 : Boolean;
Scope_Result_1 : Boolean;
Result_Var : Boolean;
If_Result_5 : Boolean;
Result_Var_1 : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
If_Result_6 : Boolean;
If_Result_7 : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_123 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_123;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2087 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2070 '<IsA BasePackageDecl, BasicSubpDecl, GenericDecl, TaskTypeDecl, ProtectedTypeDecl, SingleTaskDecl, SingleProtectedDecl at ast.py:1720>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1720
Is_A := Self /= null 
and then Self.Kind in Ada_Base_Package_Decl | Ada_Basic_Subp_Decl | Ada_Generic_Decl | Ada_Task_Type_Decl_Range | Ada_Protected_Type_Decl_Range | Ada_Single_Task_Decl_Range | Ada_Single_Protected_Decl_Range; 
--# expr-done 2070
Block_Var := Is_A; 
--# bind is_decl Block_Var
--# expr-start 2071 '<IsA Body at ast.py:1726>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1726
Is_A_1 := Self /= null 
and then Self.Kind in Ada_Body_Node; 
--# expr-done 2071
Block_Var_1 := Is_A_1; 
--# bind is_body Block_Var_1
--# expr-start 2072 '<FieldAccess for is_compilation_unit_root at ast.py:1728>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1728
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_compilation_unit_root
Fld := Libadalang.Implementation.Basic_Decl_P_Is_Compilation_Unit_Root (Node => Self);
--# end
--# expr-done 2072
if Fld then
   If_Result_7 := True;
else
if Block_Var then
   If_Result_1 := True;
else
if Allow_Bodies then
   If_Result := Block_Var_1;
else
   If_Result := False;
end if;
   If_Result_1 := If_Result;
end if;
if If_Result_1 then
--# expr-start 2074 '<env_node at ast.py:1730>' Env_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1730
--# expr-start 2073 '<FieldAccess for node_env at ast.py:1730>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1730
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Node_Env (Node => Self);
--# expr-done 2073
Env_Node := AST_Envs.Env_Node (Fld_1); 
--# expr-done 2074
Var_Expr := Env_Node;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind node Var_Expr
--# expr-start 2075 '<Cast to BasicDecl at ast.py:1731>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1731
Cast_Expr := Var_Expr; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_Basic_Decl
   then
      Cast_Result := Cast_Expr;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 2075
Var_Expr_1 := Cast_Result;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind p Var_Expr_1
--# expr-start 2082 '<If at ast.py:1732>' If_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1732
--# expr-start 2076 '<Eq at ast.py:1733>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1733
Cast_Expr_1 := Self; 
      Cast_Result_1 := Cast_Expr_1;
Is_Equal := Cast_Result_1 = Var_Expr_1; 
--# expr-done 2076
if Is_Equal then
   If_Result_4 := True;
else
   --# expr-start 2081 '<FieldAccess for has_top_level_env_name_impl at ast.py:1735>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1735
      if Var_Expr_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
if Allow_Bodies then
   --# expr-start 2078 '<Not at ast.py:1738>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1738
--# expr-start 2077 '<IsA BaseSubpBody at ast.py:1738>' Is_A_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1738
Is_A_2 := Self /= null 
and then Self.Kind in Ada_Base_Subp_Body; 
--# expr-done 2077
Not_Val := not (Is_A_2); 
--# expr-done 2078
   If_Result_2 := Not_Val;
else
   If_Result_2 := False;
end if;
if If_Result_2 then
   --# expr-start 2080 '<Not at ast.py:1739>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1739
--# expr-start 2079 '<IsA BaseSubpBody at ast.py:1739>' Is_A_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1739
Is_A_3 := Var_Expr_1 /= null 
and then Var_Expr_1.Kind in Ada_Base_Subp_Body; 
--# expr-done 2079
Not_Val_1 := not (Is_A_3); 
--# expr-done 2080
   If_Result_3 := Not_Val_1;
else
   If_Result_3 := False;
end if;
--# property-call-start BasicDecl.has_top_level_env_name_impl
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Has_Top_Level_Env_Name_Impl (Node => Var_Expr_1, Allow_Bodies => If_Result_3);
--# end
--# expr-done 2081
   If_Result_4 := Fld_2;
end if;
--# expr-done 2082
Scope_Result := If_Result_4;
   --# end
   Result_Var_1 := Scope_Result;
else
--# expr-start 2083 '<IsA PrivatePart at ast.py:1744>' Is_A_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1744
Is_A_4 := Var_Expr /= null 
and then Var_Expr.Kind in Ada_Private_Part_Range; 
--# expr-done 2083
if Is_A_4 then
--# expr-start 2085 '<Cast to BasicDecl at ast.py:1745>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1745
--# expr-start 2084 '<parent at ast.py:1745>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1745
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Var_Expr.Parent; 
--# expr-done 2084
   if Fld_3 = null
      or else Fld_3.Kind in Ada_Basic_Decl
   then
      Cast_Result_2 := Fld_3;
   else
         Cast_Result_2 := No_Bare_Ada_Node;
   end if;
--# expr-done 2085
Var_Expr_2 := Cast_Result_2;
if Var_Expr_2 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind bd Var_Expr_2
--# expr-start 2086 '<FieldAccess for has_top_level_env_name_impl at ast.py:1747>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1747
      if Var_Expr_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.has_top_level_env_name_impl
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Has_Top_Level_Env_Name_Impl (Node => Var_Expr_2, Allow_Bodies => Allow_Bodies);
--# end
--# expr-done 2086
Scope_Result_1 := Fld_4;
   --# end
   Result_Var := Scope_Result_1;
else
   Result_Var := False;
end if;
   If_Result_5 := Result_Var;
else
   If_Result_5 := False;
end if;
   Result_Var_1 := If_Result_5;
end if;
Scope_Result_2 := Result_Var_1;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := True;
end if;
   If_Result_6 := Result_Var_2;
else
   If_Result_6 := False;
end if;
   If_Result_7 := If_Result_6;
end if;
Let_Result := If_Result_7; 
--# end
--# expr-done 2087
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_123;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_123;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Has_Top_Level_Env_Name_Impl;
--# end
--# property-start BasicDecl.has_top_level_env_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1756
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Has_Top_Level_Env_Name
  (Node : Bare_Basic_Decl
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
            procedure Finalizer_Scope_124 with Inline_Always;
      Var_Expr : Bare_Ada_Node;
Var_Expr_1 : Bare_Basic_Decl;
Var_Expr_2 : Bare_Basic_Decl;
Fld : Lexical_Env := Empty_Env;
Env_Node : Bare_Ada_Node;
Cast_Expr : Bare_Ada_Node;
Cast_Result : Bare_Basic_Decl;
Fld_1 : Boolean;
Scope_Result : Boolean;
Is_A : Boolean;
Fld_2 : Bare_Ada_Node;
Cast_Result_1 : Bare_Basic_Decl;
Fld_3 : Boolean;
Scope_Result_1 : Boolean;
Result_Var : Boolean;
If_Result : Boolean;
Result_Var_1 : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_124 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_124;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2096 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2089 '<env_node at ast.py:1824>' Env_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1824
--# expr-start 2088 '<FieldAccess for children_env at ast.py:1824>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1824
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children_Env (Node => Self);
--# expr-done 2088
Env_Node := AST_Envs.Env_Node (Fld); 
--# expr-done 2089
Var_Expr := Env_Node;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind node Var_Expr
--# expr-start 2090 '<Cast to BasicDecl at ast.py:1825>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1825
Cast_Expr := Var_Expr; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_Basic_Decl
   then
      Cast_Result := Cast_Expr;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 2090
Var_Expr_1 := Cast_Result;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind bd Var_Expr_1
--# expr-start 2091 '<FieldAccess for has_top_level_env_name_impl at ast.py:1826>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1826
      if Var_Expr_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.has_top_level_env_name_impl
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Has_Top_Level_Env_Name_Impl (Node => Var_Expr_1, Allow_Bodies => True);
--# end
--# expr-done 2091
Scope_Result := Fld_1;
   --# end
   Result_Var_1 := Scope_Result;
else
--# expr-start 2092 '<IsA PrivatePart at ast.py:1830>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1830
Is_A := Var_Expr /= null 
and then Var_Expr.Kind in Ada_Private_Part_Range; 
--# expr-done 2092
if Is_A then
--# expr-start 2094 '<Cast to BasicDecl at ast.py:1831>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1831
--# expr-start 2093 '<parent at ast.py:1831>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1831
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Var_Expr.Parent; 
--# expr-done 2093
   if Fld_2 = null
      or else Fld_2.Kind in Ada_Basic_Decl
   then
      Cast_Result_1 := Fld_2;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 2094
Var_Expr_2 := Cast_Result_1;
if Var_Expr_2 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind bd Var_Expr_2
--# expr-start 2095 '<FieldAccess for has_top_level_env_name_impl at ast.py:1832>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1832
      if Var_Expr_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.has_top_level_env_name_impl
Fld_3 := Libadalang.Implementation.Basic_Decl_P_Has_Top_Level_Env_Name_Impl (Node => Var_Expr_2, Allow_Bodies => True);
--# end
--# expr-done 2095
Scope_Result_1 := Fld_3;
   --# end
   Result_Var := Scope_Result_1;
else
   Result_Var := False;
end if;
   If_Result := Result_Var;
else
   If_Result := False;
end if;
   Result_Var_1 := If_Result;
end if;
Scope_Result_2 := Result_Var_1;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := True;
end if;
Let_Result := Result_Var_2; 
--# end
--# expr-done 2096
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_124;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_124;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Has_Top_Level_Env_Name;
--# end
--# property-start BasicDecl.env_spec_fully_qualified_name_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1842
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Env_Spec_Fully_Qualified_Name_Impl
  (Node : Bare_Basic_Decl
      ; Self_Env : Lexical_Env
  )
   return String_Type
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind self_env Self_Env
   Property_Result : String_Type;
            procedure Finalizer_Scope_125 with Inline_Always;
            procedure Finalizer_Scope_2005 with Inline_Always;
      Var_Expr : Bare_Basic_Decl;
Var_Expr_1 : Bare_Basic_Decl;
Var_Expr_2 : Bare_Private_Part;
Fld : Boolean;
Fld_1 : Bare_Compilation_Unit;
Fld_2 : Symbol_Type_Array_Access;
Str : String_Type;
Fld_3 : String_Type;
Is_A : Boolean;
Fld_4 : Bare_Ada_Node;
Cast_Result : Bare_Basic_Decl;
Fld_5 : Bare_Ada_Node;
Fld_6 : Lexical_Env := Empty_Env;
Fld_7 : String_Type;
Env_Node : Bare_Ada_Node;
Cast_Result_1 : Bare_Basic_Decl;
Scope_Result : Bare_Basic_Decl;
Env_Node_1 : Bare_Ada_Node;
Cast_Result_2 : Bare_Private_Part;
Fld_8 : Bare_Ada_Node;
Scope_Result_1 : Bare_Ada_Node;
Result_Var : Bare_Ada_Node;
Cast_Result_3 : Bare_Basic_Decl;
Result_Var_1 : Bare_Basic_Decl;
Fld_9 : Lexical_Env := Empty_Env;
Fld_10 : String_Type;
Str_1 : String_Type;
Concat_Result : String_Type;
Fld_11 : Symbol_Type;
Fld_12 : String_Type;
Concat_Result_1 : String_Type;
Scope_Result_2 : String_Type;
Result_Var_2 : String_Type;
If_Result : String_Type;
If_Result_1 : String_Type;
Let_Result : String_Type;
            procedure Finalizer_Scope_125 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Str);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Fld_6);
                     Dec_Ref (Fld_7);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Result_Var_2);
                     Dec_Ref (If_Result);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_125;
            procedure Finalizer_Scope_2005 is
            begin
                     Dec_Ref (Fld_9);
                     Dec_Ref (Fld_10);
                     Dec_Ref (Str_1);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Fld_12);
                     Dec_Ref (Concat_Result_1);
            end Finalizer_Scope_2005;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2122 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2121 '<Cond at ast.py:1846>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1846
--# expr-start 2097 '<FieldAccess for is_compilation_unit_root at ast.py:1849>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1849
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_compilation_unit_root
Fld := Libadalang.Implementation.Basic_Decl_P_Is_Compilation_Unit_Root (Node => Self);
--# end
--# expr-done 2097
if Fld then
   --# expr-start 2101 '<FieldAccess for sym_join at ast.py:1850>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1850
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2099 '<FieldAccess for syntactic_fully_qualified_name at ast.py:1851>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1851
--# expr-start 2098 '<FieldAccess for enclosing_compilation_unit at ast.py:1851>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1851
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.enclosing_compilation_unit
Fld_1 := Libadalang.Implementation.Ada_Node_P_Enclosing_Compilation_Unit (Node => Self);
--# end
--# expr-done 2098
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.syntactic_fully_qualified_name
Fld_2 := Libadalang.Implementation.Compilation_Unit_P_Syntactic_Fully_Qualified_Name (Node => Fld_1);
--# end
--# expr-done 2099
--# expr-start 2100 '<String at ast.py:1852>' Str /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1852
Str := Create_String ("."); 
--# expr-done 2100
--# property-call-start AdaNode.sym_join
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sym_Join (Node => Self, Syms => Fld_2, Sep => Str);
--# end
--# expr-done 2101
   If_Result_1 := Fld_3;
else
--# expr-start 2102 '<IsA GenericPackageInternal, GenericSubpInternal at ast.py:1857>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1857
Is_A := Self /= null 
and then Self.Kind in Ada_Generic_Package_Internal_Range | Ada_Generic_Subp_Internal_Range; 
--# expr-done 2102
if Is_A then
   --# expr-start 2107 '<FieldAccess for env_spec_fully_qualified_name_impl at ast.py:1858>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1858
--# expr-start 2104 '<Cast to BasicDecl at ast.py:1858>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1858
--# expr-start 2103 '<parent at ast.py:1858>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1858
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Self.Parent; 
--# expr-done 2103
   if Fld_4 = null
      or else Fld_4.Kind in Ada_Basic_Decl
   then
      Cast_Result := Fld_4;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 2104
      if Cast_Result = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2106 '<FieldAccess for node_env at ast.py:1859>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1859
--# expr-start 2105 '<parent at ast.py:1859>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1859
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Self.Parent; 
--# expr-done 2105
      if Fld_5 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Node_Env (Node => Fld_5);
--# expr-done 2106
--# property-call-start BasicDecl.env_spec_fully_qualified_name_impl
Fld_7 := Libadalang.Implementation.Basic_Decl_P_Env_Spec_Fully_Qualified_Name_Impl (Node => Cast_Result, Self_Env => Fld_6);
--# end
--# expr-done 2107
   If_Result := Fld_7;
else
--# expr-start 2109 '<Cast to BasicDecl at ast.py:1863>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1863
--# expr-start 2108 '<env_node at ast.py:1863>' Env_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1863
Env_Node := AST_Envs.Env_Node (Self_Env); 
--# expr-done 2108
   if Env_Node = null
      or else Env_Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_1 := Env_Node;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 2109
Var_Expr_1 := Cast_Result_1;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind e Var_Expr_1
Scope_Result := Var_Expr_1;
   --# end
   Result_Var_1 := Scope_Result;
else
   --# expr-start 2113 '<Cast to BasicDecl at ast.py:1864>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1864
--# expr-start 2111 '<Cast to PrivatePart at ast.py:1864>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1864
--# expr-start 2110 '<env_node at ast.py:1864>' Env_Node_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1864
Env_Node_1 := AST_Envs.Env_Node (Self_Env); 
--# expr-done 2110
   if Env_Node_1 = null
      or else Env_Node_1.Kind in Ada_Private_Part_Range
   then
      Cast_Result_2 := Env_Node_1;
   else
         Cast_Result_2 := No_Bare_Ada_Node;
   end if;
--# expr-done 2111
Var_Expr_2 := Cast_Result_2;
if Var_Expr_2 /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 2112 '<parent at ast.py:1864>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1864
      if Var_Expr_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Var_Expr_2.Parent; 
--# expr-done 2112
Scope_Result_1 := Fld_8;
   --# end
   Result_Var := Scope_Result_1;
else
   Result_Var := No_Bare_Ada_Node;
end if;
   if Result_Var = null
      or else Result_Var.Kind in Ada_Basic_Decl
   then
      Cast_Result_3 := Result_Var;
   else
         Cast_Result_3 := No_Bare_Ada_Node;
   end if;
--# expr-done 2113
   Result_Var_1 := Cast_Result_3;
end if;
Var_Expr := Result_Var_1;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind bd Var_Expr
--# expr-start 2120 '<Concat at ast.py:1868>' Concat_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1868
--# expr-start 2117 '<Concat at ast.py:1868>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1868
--# expr-start 2115 '<FieldAccess for env_spec_fully_qualified_name_impl at ast.py:1868>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1868
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2114 '<FieldAccess for node_env at ast.py:1868>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1868
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Libadalang.Implementation.Node_Env (Node => Var_Expr);
--# expr-done 2114
--# property-call-start BasicDecl.env_spec_fully_qualified_name_impl
Fld_10 := Libadalang.Implementation.Basic_Decl_P_Env_Spec_Fully_Qualified_Name_Impl (Node => Var_Expr, Self_Env => Fld_9);
--# end
--# expr-done 2115
--# expr-start 2116 '<String at ast.py:1869>' Str_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1869
Str_1 := Create_String ("."); 
--# expr-done 2116
Concat_Result := Concat_String (Fld_10, Str_1); 
--# expr-done 2117
--# expr-start 2119 '<FieldAccess for image at ast.py:1870>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1870
--# expr-start 2118 '<FieldAccess for name_symbol at ast.py:1870>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1870
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld_11 := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Self);
--# end
--# expr-done 2118
      if Fld_11 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Libadalang.Implementation.Image (Self => Fld_11);
--# expr-done 2119
Concat_Result_1 := Concat_String (Concat_Result, Fld_12); 
--# expr-done 2120
Scope_Result_2 := Concat_Result_1;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2005;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := Empty_String;
end if;
      Inc_Ref (Result_Var_2);
   If_Result := Result_Var_2;
end if;
      Inc_Ref (If_Result);
   If_Result_1 := If_Result;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 2121
Let_Result := If_Result_1; Inc_Ref (Let_Result);
--# end
--# expr-done 2122
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_125;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_125;
               Finalizer_Scope_2005;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Env_Spec_Fully_Qualified_Name_Impl;
--# end
--# property-start BasicDecl.env_spec_fully_qualified_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1875
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Env_Spec_Fully_Qualified_Name
  (Node : Bare_Basic_Decl
  )
   return String_Type
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : String_Type;
            procedure Finalizer_Scope_126 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Fld_1 : String_Type;
Let_Result : String_Type;
            procedure Finalizer_Scope_126 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_126;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2125 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2124 '<FieldAccess for env_spec_fully_qualified_name_impl at ast.py:1884>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1884
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2123 '<FieldAccess for children_env at ast.py:1884>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1884
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children_Env (Node => Self);
--# expr-done 2123
--# property-call-start BasicDecl.env_spec_fully_qualified_name_impl
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Env_Spec_Fully_Qualified_Name_Impl (Node => Self, Self_Env => Fld);
--# end
--# expr-done 2124
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 2125
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_126;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_126;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Env_Spec_Fully_Qualified_Name;
--# end
--# property-start BasicDecl.top_level_env_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1887
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Top_Level_Env_Name
  (Node : Bare_Basic_Decl
  )
   return String_Type
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : String_Type;
            procedure Finalizer_Scope_127 with Inline_Always;
      Fld : String_Type;
Fld_1 : Boolean;
If_Result : String_Type;
Let_Result : String_Type;
            procedure Finalizer_Scope_127 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_127;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2129 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2128 '<If at ast.py:1893>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1893
--# expr-start 2126 '<FieldAccess for has_top_level_env_name at ast.py:1894>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1894
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.has_top_level_env_name
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Has_Top_Level_Env_Name (Node => Self);
--# end
--# expr-done 2126
if Fld_1 then
   --# expr-start 2127 '<FieldAccess for env_spec_fully_qualified_name at ast.py:1895>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1895
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.env_spec_fully_qualified_name
Fld := Libadalang.Implementation.Basic_Decl_P_Env_Spec_Fully_Qualified_Name (Node => Self);
--# end
--# expr-done 2127
   If_Result := Fld;
else
   If_Result := Empty_String;
end if;
      Inc_Ref (If_Result);
--# expr-done 2128
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 2129
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_127;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_127;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Top_Level_Env_Name;
--# end
--# property-start '[dispatcher]BasicDecl.child_decl_initial_env_name' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Decl_P_Child_Decl_Initial_Env_Name
  (Node : Bare_Basic_Decl
      ; Private_Part : Boolean
         := False
  )
   return Symbol_Type
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind private_part Private_Part
   Property_Result : Symbol_Type;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Decl (Self.Kind) is
               when Ada_Abstract_State_Decl .. Ada_Entry_Index_Spec | Ada_Exception_Decl .. Ada_Single_Task_Decl =>
                  --# property-call-start BasicDecl.child_decl_initial_env_name
                  Property_Result := Basic_Decl_P_Child_Decl_Initial_Env_Name
                    (Self
                        , Private_Part
                    );
                  --# end
               when Ada_Error_Decl =>
                  --# property-call-start ErrorDecl.child_decl_initial_env_name
                  Property_Result := Error_Decl_P_Child_Decl_Initial_Env_Name
                    (Self
                        , Private_Part
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Decl_P_Child_Decl_Initial_Env_Name;
--# end
--# property-start BasicDecl.child_decl_initial_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1934
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Child_Decl_Initial_Env
  (Node : Bare_Basic_Decl
      ; Private_Part : Boolean
         := False
  )
   return Internal_Designated_Env
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind private_part Private_Part
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_129 with Inline_Always;
            procedure Finalizer_Scope_2007 with Inline_Always;
      Var_Expr : Symbol_Type;
Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Scope_Result : Internal_Designated_Env := No_Designated_Env;
Fld_1 : Lexical_Env := Empty_Env;
New_Struct_1 : Internal_Designated_Env := No_Designated_Env;
Result_Var : Internal_Designated_Env := No_Designated_Env;
Let_Result : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_129 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Fld_1);
                     Dec_Ref (New_Struct_1);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_129;
            procedure Finalizer_Scope_2007 is
            begin
                     Dec_Ref (New_Struct);
            end Finalizer_Scope_2007;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2134 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2130 '<FieldAccess for child_decl_initial_env_name at ast.py:1943>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1943
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.child_decl_initial_env_name'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Child_Decl_Initial_Env_Name (Node => Self, Private_Part => Private_Part);
--# end
--# expr-done 2130
Var_Expr := Fld;
if Var_Expr /= null then
   --# scope-start
   --# bind name Var_Expr
--# expr-start 2131 '<New DesignatedEnv at ast.py:1944>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1944
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => Var_Expr, Kind => Named_Env); 
--# expr-done 2131
Scope_Result := New_Struct;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2007;
   Result_Var := Scope_Result;
else
   --# expr-start 2133 '<New DesignatedEnv at ast.py:1945>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1945
--# expr-start 2132 '<FieldAccess for default_initial_env at ast.py:1945>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1945
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.default_initial_env
Fld_1 := Libadalang.Implementation.Ada_Node_P_Default_Initial_Env (Node => Self);
--# end
--# expr-done 2132
Inc_Ref (Fld_1);
New_Struct_1 := (Direct_Env => Fld_1, Env_Name => null, Kind => Direct_Env); 
--# expr-done 2133
   Result_Var := New_Struct_1;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 2134
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_129;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_129;
               Finalizer_Scope_2007;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Child_Decl_Initial_Env;
--# end
--# property-start BasicDecl.basic_decl_env_assocs /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1949
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Basic_Decl_Env_Assocs
  (Node : Bare_Basic_Decl
      ; Dest_Env : Internal_Designated_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind dest_env Dest_Env
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_131 with Inline_Always;
      Block_Var : Symbol_Type;
Block_Var_1 : Internal_Env_Assoc_Array_Access;
Block_Var_2 : Internal_Env_Assoc_Array_Access;
Fld : Symbol_Type;
Fld_1 : Symbol_Type;
Cast_Expr : Bare_Basic_Decl;
Cast_Result : Bare_Ada_Node;
New_Struct : Internal_Env_Assoc := No_Env_Assoc;
Singleton : Internal_Env_Assoc_Array_Access;
Sym : Symbol_Type;
Cast_Expr_1 : Bare_Basic_Decl;
Cast_Result_1 : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
Singleton_1 : Internal_Env_Assoc_Array_Access;
Sym_1 : Symbol_Type;
Is_Equal : Boolean;
If_Result : Internal_Env_Assoc_Array_Access;
Concat_Result : Internal_Env_Assoc_Array_Access;
Let_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_131 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (Singleton);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (New_Struct_1);
                     Dec_Ref (Singleton_1);
                     Dec_Ref (If_Result);
                     Dec_Ref (Block_Var_2);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_131;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2144 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2135 '<FieldAccess for name_symbol at ast.py:1958>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1958
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 2135
Block_Var := Fld; 
--# bind name Block_Var
--# expr-start 2137 '<New EnvAssoc at ast.py:1961>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1961
--# expr-start 2136 '<FieldAccess for name_symbol at ast.py:1962>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1962
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 2136
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (Dest_Env);
New_Struct := (Dest_Env => Dest_Env, Key => Fld_1, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 2137
                Singleton := Create_Internal_Env_Assoc_Array (Items_Count => 1);
                Singleton.Items (1) := New_Struct;
                Inc_Ref (Singleton.Items (1));
Block_Var_1 := Singleton; Inc_Ref (Block_Var_1);
--# bind base_assoc Block_Var_1
--# expr-start 2142 '<If at ast.py:1968>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1968
--# expr-start 2139 '<Eq at ast.py:1969>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1969
--# expr-start 2138 '<Symbol "=" at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_10); 
--# expr-done 2138
Is_Equal := Block_Var = Sym_1; 
--# expr-done 2139
if Is_Equal then
   --# expr-start 2141 '<New EnvAssoc at ast.py:1970>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1970
--# expr-start 2140 '<Symbol "/=" at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_7); 
--# expr-done 2140
Cast_Expr_1 := Self; 
      Cast_Result_1 := Cast_Expr_1;
Inc_Ref (Dest_Env);
New_Struct_1 := (Dest_Env => Dest_Env, Key => Sym, Metadata => No_Metadata, Value => Cast_Result_1); 
--# expr-done 2141
                Singleton_1 := Create_Internal_Env_Assoc_Array (Items_Count => 1);
                Singleton_1.Items (1) := New_Struct_1;
                Inc_Ref (Singleton_1.Items (1));
   If_Result := Singleton_1;
else
   If_Result := No_Internal_Env_Assoc_Array_Type;
end if;
      Inc_Ref (If_Result);
--# expr-done 2142
Block_Var_2 := If_Result; Inc_Ref (Block_Var_2);
--# bind implicit_neq_assoc Block_Var_2
--# expr-start 2143 '<Concat at ast.py:1978>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1978
Concat_Result := Concat (Block_Var_1, Block_Var_2); 
--# expr-done 2143
Let_Result := Concat_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 2144
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_131;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_131;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Basic_Decl_Env_Assocs;
--# end
--# property-start BasicDecl.child_decl_env_assocs /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1981
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Child_Decl_Env_Assocs
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_132 with Inline_Always;
            procedure Finalizer_Scope_2008 with Inline_Always;
      Block_Var : Internal_Designated_Env := No_Designated_Env;
Var_Expr : Symbol_Type;
Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Scope_Result : Internal_Designated_Env := No_Designated_Env;
New_Struct_1 : Internal_Designated_Env := No_Designated_Env;
Result_Var : Internal_Designated_Env := No_Designated_Env;
Fld_1 : Internal_Env_Assoc_Array_Access;
Let_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_132 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (New_Struct_1);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_132;
            procedure Finalizer_Scope_2008 is
            begin
                     Dec_Ref (New_Struct);
            end Finalizer_Scope_2008;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2149 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2145 '<FieldAccess for child_decl_initial_env_name at ast.py:1992>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1992
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.child_decl_initial_env_name'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Child_Decl_Initial_Env_Name (Node => Self, Private_Part => False);
--# end
--# expr-done 2145
Var_Expr := Fld;
if Var_Expr /= null then
   --# scope-start
   --# bind non_null_name Var_Expr
--# expr-start 2146 '<New DesignatedEnv at ???>' New_Struct None
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => Var_Expr, Kind => Named_Env); 
--# expr-done 2146
Scope_Result := New_Struct;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2008;
   Result_Var := Scope_Result;
else
   --# expr-start 2147 '<New DesignatedEnv at ast.py:1991>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1991
Inc_Ref (Empty_Env);
New_Struct_1 := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 2147
   Result_Var := New_Struct_1;
end if;
      Inc_Ref (Result_Var);
Block_Var := Result_Var; Inc_Ref (Block_Var);
--# bind dest_env Block_Var
--# expr-start 2148 '<FieldAccess for basic_decl_env_assocs at ast.py:1995>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1995
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.basic_decl_env_assocs
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Basic_Decl_Env_Assocs (Node => Ent.Node, Dest_Env => Block_Var, E_Info => Ent.Info);
--# end
--# expr-done 2148
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 2149
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_132;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_132;
               Finalizer_Scope_2008;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Child_Decl_Env_Assocs;
--# end
--# property-start BasicDecl.is_formal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1997
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Is_Formal
  (Node : Bare_Basic_Decl
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2151 '<IsA GenericFormal at ast.py:1998>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1998
--# expr-start 2150 '<parent at ast.py:1998>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1998
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 2150
Is_A := Fld /= null 
and then Fld.Kind in Ada_Generic_Formal; 
--# expr-done 2151
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Is_Formal;
--# end
--# property-start '[dispatcher]BasicDecl.previous_part_for_decl' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Decl_P_Previous_Part_For_Decl
  (Node : Bare_Basic_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Decl (Self.Kind) is
               when Ada_Abstract_State_Decl .. Ada_Anonymous_Expr_Decl | Ada_Generic_Package_Internal .. Ada_Number_Decl | Ada_Package_Renaming_Decl .. Ada_Single_Task_Decl =>
                  --# property-call-start BasicDecl.previous_part_for_decl
                  Property_Result := Basic_Decl_P_Previous_Part_For_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Component_Decl .. Ada_Synthetic_Formal_Param_Decl =>
                  --# property-call-start BaseFormalParamDecl.previous_part_for_decl
                  Property_Result := Base_Formal_Param_Decl_P_Previous_Part_For_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Object_Decl .. Ada_No_Type_Object_Renaming_Decl =>
                  --# property-call-start ObjectDecl.previous_part_for_decl
                  Property_Result := Object_Decl_P_Previous_Part_For_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Decl_P_Previous_Part_For_Decl;
--# end
--# property-start BasicDecl.canonical_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2056
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Canonical_Part
  (Node : Bare_Basic_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Var_Expr : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Basic_Decl_Basic_Decl_P_Canonical_Part,
            Items    => new Mmz_Key_Array (1 ..  3))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Basic_Decl,
                                As_Bare_Basic_Decl => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Boolean,
                                      As_Boolean => Imprecise_Fallback);
               Mmz_K.Items (3) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Basic_Decl;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 2154 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2152 '<FieldAccess for previous_part_for_decl at ast.py:2061>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2061
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.previous_part_for_decl'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Previous_Part_For_Decl (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2152
Var_Expr := Fld;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind pp Var_Expr
--# expr-start 2153 '<FieldAccess for canonical_part at ast.py:2062>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2062
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.canonical_part
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Canonical_Part (Node => Var_Expr.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr.Info);
--# end
--# expr-done 2153
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := Ent;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 2154
      Property_Result := Let_Result;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Basic_Decl,
                        As_Internal_Entity_Basic_Decl => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Canonical_Part;
--# end
--# property-start BasicDecl.all_previous_parts /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2067
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_All_Previous_Parts
  (Node : Bare_Basic_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl_Array_Access
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_138 with Inline_Always;
            procedure Finalizer_Scope_2013 with Inline_Always;
      Var_Expr : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Basic_Decl_Array_Access;
Singleton : Internal_Entity_Basic_Decl_Array_Access;
Concat_Result : Internal_Entity_Basic_Decl_Array_Access;
Cast_Expr : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Basic_Decl;
Cast_Result_1 : Internal_Entity;
Is_Equiv : Boolean;
If_Result : Internal_Entity_Basic_Decl_Array_Access;
Scope_Result : Internal_Entity_Basic_Decl_Array_Access;
Result_Var : Internal_Entity_Basic_Decl_Array_Access;
Let_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_138 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_138;
            procedure Finalizer_Scope_2013 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Singleton);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (If_Result);
            end Finalizer_Scope_2013;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2160 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2155 '<FieldAccess for previous_part_for_decl at ast.py:2072>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2072
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.previous_part_for_decl'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Previous_Part_For_Decl (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2155
Var_Expr := Fld;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind pp Var_Expr
--# expr-start 2159 '<If at ast.py:2073>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2073
--# expr-start 2156 '<Eq at ast.py:2074>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2074
Cast_Expr := Ent; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Cast_Expr_1 := Var_Expr; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Is_Equiv := Equivalent (Cast_Result, Cast_Result_1); 
--# expr-done 2156
if Is_Equiv then
   If_Result := No_Internal_Entity_Basic_Decl_Array_Type;
else
   --# expr-start 2158 '<Concat at ast.py:2076>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2076
--# expr-start 2157 '<FieldAccess for all_previous_parts at ast.py:2076>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2076
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.all_previous_parts
Fld_1 := Libadalang.Implementation.Basic_Decl_P_All_Previous_Parts (Node => Var_Expr.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr.Info);
--# end
--# expr-done 2157
                Singleton := Create_Internal_Entity_Basic_Decl_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
Concat_Result := Concat (Fld_1, Singleton); 
--# expr-done 2158
   If_Result := Concat_Result;
end if;
      Inc_Ref (If_Result);
--# expr-done 2159
Scope_Result := If_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2013;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Basic_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 2160
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_138;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_138;
               Finalizer_Scope_2013;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_All_Previous_Parts;
--# end
--# property-start BasicDecl.all_next_parts /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2082
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_All_Next_Parts
  (Node : Bare_Basic_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl_Array_Access
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_139 with Inline_Always;
            procedure Finalizer_Scope_2018 with Inline_Always;
      Var_Expr : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Basic_Decl;
Singleton : Internal_Entity_Basic_Decl_Array_Access;
Fld_1 : Internal_Entity_Basic_Decl_Array_Access;
Concat_Result : Internal_Entity_Basic_Decl_Array_Access;
Cast_Expr : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Basic_Decl;
Cast_Result_1 : Internal_Entity;
Is_Equiv : Boolean;
If_Result : Internal_Entity_Basic_Decl_Array_Access;
Scope_Result : Internal_Entity_Basic_Decl_Array_Access;
Result_Var : Internal_Entity_Basic_Decl_Array_Access;
Let_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_139 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_139;
            procedure Finalizer_Scope_2018 is
            begin
                     Dec_Ref (Singleton);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (If_Result);
            end Finalizer_Scope_2018;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2166 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2161 '<FieldAccess for next_part_for_decl at ast.py:2087>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2087
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.next_part_for_decl'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Next_Part_For_Decl (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2161
Var_Expr := Fld;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind np Var_Expr
--# expr-start 2165 '<If at ast.py:2088>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2088
--# expr-start 2162 '<Eq at ast.py:2089>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2089
Cast_Expr := Ent; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Cast_Expr_1 := Var_Expr; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Is_Equiv := Equivalent (Cast_Result, Cast_Result_1); 
--# expr-done 2162
if Is_Equiv then
   If_Result := No_Internal_Entity_Basic_Decl_Array_Type;
else
   --# expr-start 2164 '<Concat at ast.py:2091>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2091
                Singleton := Create_Internal_Entity_Basic_Decl_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
--# expr-start 2163 '<FieldAccess for all_next_parts at ast.py:2091>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2091
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.all_next_parts
Fld_1 := Libadalang.Implementation.Basic_Decl_P_All_Next_Parts (Node => Var_Expr.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr.Info);
--# end
--# expr-done 2163
Concat_Result := Concat (Singleton, Fld_1); 
--# expr-done 2164
   If_Result := Concat_Result;
end if;
      Inc_Ref (If_Result);
--# expr-done 2165
Scope_Result := If_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2018;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Basic_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 2166
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_139;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_139;
               Finalizer_Scope_2018;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_All_Next_Parts;
--# end
--# property-start BasicDecl.all_parts /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2097
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_All_Parts
  (Node : Bare_Basic_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl_Array_Access
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_140 with Inline_Always;
      Block_Var : Internal_Entity_Basic_Decl_Array_Access;
Block_Var_1 : Internal_Entity_Basic_Decl_Array_Access;
Fld : Internal_Entity_Basic_Decl_Array_Access;
Fld_1 : Internal_Entity_Basic_Decl_Array_Access;
Singleton : Internal_Entity_Basic_Decl_Array_Access;
Concat_Result : Internal_Entity_Basic_Decl_Array_Access;
Concat_Result_1 : Internal_Entity_Basic_Decl_Array_Access;
Let_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_140 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Singleton);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Concat_Result_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_140;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2171 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2167 '<FieldAccess for all_previous_parts at ast.py:2102>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2102
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.all_previous_parts
Fld := Libadalang.Implementation.Basic_Decl_P_All_Previous_Parts (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2167
Block_Var := Fld; Inc_Ref (Block_Var);
--# bind prevs Block_Var
--# expr-start 2168 '<FieldAccess for all_next_parts at ast.py:2103>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2103
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.all_next_parts
Fld_1 := Libadalang.Implementation.Basic_Decl_P_All_Next_Parts (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2168
Block_Var_1 := Fld_1; Inc_Ref (Block_Var_1);
--# bind nexts Block_Var_1
--# expr-start 2170 '<Concat at ast.py:2104>' Concat_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2104
--# expr-start 2169 '<Concat at ast.py:2104>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2104
                Singleton := Create_Internal_Entity_Basic_Decl_Array (Items_Count => 1);
                Singleton.Items (1) := Ent;
Concat_Result := Concat (Block_Var, Singleton); 
--# expr-done 2169
Concat_Result_1 := Concat (Concat_Result, Block_Var_1); 
--# expr-done 2170
Let_Result := Concat_Result_1; Inc_Ref (Let_Result);
--# end
--# expr-done 2171
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_140;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_140;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_All_Parts;
--# end
--# property-start '[dispatcher]BasicDecl.is_static_decl' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Decl_P_Is_Static_Decl
  (Node : Bare_Basic_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Decl (Self.Kind) is
               when Ada_Abstract_State_Decl | Ada_Component_Decl .. Ada_Package_Decl | Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl | Ada_Abstract_Subp_Decl .. Ada_Entry_Decl | Ada_Generic_Subp_Internal | Ada_Expr_Function .. Ada_Named_Stmt_Decl | Ada_Package_Renaming_Decl .. Ada_Single_Task_Decl =>
                  --# property-call-start BasicDecl.is_static_decl
                  Property_Result := Basic_Decl_P_Is_Static_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Anonymous_Expr_Decl =>
                  --# property-call-start AnonymousExprDecl.is_static_decl
                  Property_Result := Anonymous_Expr_Decl_P_Is_Static_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl =>
                  --# property-call-start DiscreteBaseSubtypeDecl.is_static_decl
                  Property_Result := Discrete_Base_Subtype_Decl_P_Is_Static_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Subtype_Decl =>
                  --# property-call-start SubtypeDecl.is_static_decl
                  Property_Result := Subtype_Decl_P_Is_Static_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.is_static_decl
                  Property_Result := Type_Decl_P_Is_Static_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Enum_Literal_Decl .. Ada_Synthetic_Char_Enum_Lit =>
                  --# property-call-start EnumLiteralDecl.is_static_decl
                  Property_Result := Enum_Literal_Decl_P_Is_Static_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Synthetic_Subp_Decl =>
                  --# property-call-start SyntheticSubpDecl.is_static_decl
                  Property_Result := Synthetic_Subp_Decl_P_Is_Static_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Number_Decl =>
                  --# property-call-start NumberDecl.is_static_decl
                  Property_Result := Number_Decl_P_Is_Static_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Object_Decl .. Ada_No_Type_Object_Renaming_Decl =>
                  --# property-call-start ObjectDecl.is_static_decl
                  Property_Result := Object_Decl_P_Is_Static_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Decl_P_Is_Static_Decl;
--# end
--# property-start BasicDecl.unshed_rebindings /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2114
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Unshed_Rebindings
  (Node : Bare_Basic_Decl
      ; Rebindings : Env_Rebindings
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind rebindings Rebindings
   Property_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_142 with Inline_Always;
      Is_Equal : Boolean;
Fld : Lexical_Env := Empty_Env;
Env_Node : Bare_Ada_Node;
Cast_Expr : Bare_Basic_Decl;
Cast_Result : Bare_Ada_Node;
Is_Equal_1 : Boolean;
Fld_1 : Internal_Entity_Info;
Fld_2 : Internal_Metadata;
Fld_3 : Internal_Entity_Info;
Fld_4 : Boolean;
New_Struct : Internal_Entity_Info;
New_Struct_1 : Internal_Entity_Basic_Decl;
Fld_5 : Internal_Entity_Info;
Fld_6 : Env_Rebindings;
Is_Equal_2 : Boolean;
Fld_7 : Env_Rebindings;
Fld_8 : Internal_Entity_Info;
Fld_9 : Env_Rebindings;
Is_Equal_3 : Boolean;
If_Result : Boolean;
If_Result_1 : Internal_Entity_Basic_Decl;
Fld_10 : Env_Rebindings;
Fld_11 : Internal_Entity_Basic_Decl;
If_Result_2 : Internal_Entity_Basic_Decl;
If_Result_3 : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_142 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_142;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2193 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2192 '<Cond at ast.py:2121>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2121
--# expr-start 2172 '<Eq at ast.py:2122>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2122
Is_Equal := Rebindings = null; 
--# expr-done 2172
if Is_Equal then
   If_Result_3 := Ent;
else
--# expr-start 2175 '<Eq at ast.py:2125>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2125
--# expr-start 2174 '<env_node at ast.py:2125>' Env_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2125
--# expr-start 2173 '<FieldAccess for old_env at ast.py:2125>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2125
      if Rebindings = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Rebindings.Old_Env;
Inc_Ref (Fld);
--# expr-done 2173
Env_Node := AST_Envs.Env_Node (Fld); 
--# expr-done 2174
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Is_Equal_1 := Env_Node = Cast_Result; 
--# expr-done 2175
if Is_Equal_1 then
   --# expr-start 2189 '<If at ast.py:2127>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2127
--# expr-start 2178 '<Eq at ast.py:2129>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2129
--# expr-start 2177 '<FieldAccess for rebindings at ast.py:2129>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2129
--# expr-start 2176 '<FieldAccess for info at ast.py:2129>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2129
Fld_5 := Ent.Info;
--# expr-done 2176
Fld_6 := Fld_5.Rebindings;
--# expr-done 2177
Is_Equal_2 := Fld_6 = null; 
--# expr-done 2178
if Is_Equal_2 then
   If_Result := True;
else
   --# expr-start 2182 '<Eq at ast.py:2130>' Is_Equal_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2130
--# expr-start 2179 '<FieldAccess for get_parent at ast.py:2130>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2130
      if Rebindings = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Rebindings.Parent;
--# expr-done 2179
--# expr-start 2181 '<FieldAccess for rebindings at ast.py:2130>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2130
--# expr-start 2180 '<FieldAccess for info at ast.py:2130>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2130
Fld_8 := Ent.Info;
--# expr-done 2180
Fld_9 := Fld_8.Rebindings;
--# expr-done 2181
Is_Equal_3 := Fld_7 = Fld_9; 
--# expr-done 2182
   If_Result := Is_Equal_3;
end if;
if If_Result then
   --# expr-start 2188 '<New BasicDecl.entity at ???>' New_Struct_1 None
--# expr-start 2187 '<New EntityInfo at ast.py:2135>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2135
--# expr-start 2184 '<FieldAccess for from_rebound at ast.py:2138>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2138
--# expr-start 2183 '<FieldAccess for info at ast.py:2138>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2138
Fld_3 := Ent.Info;
--# expr-done 2183
Fld_4 := Fld_3.From_Rebound;
--# expr-done 2184
--# expr-start 2186 '<FieldAccess for md at ast.py:2137>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2137
--# expr-start 2185 '<FieldAccess for info at ast.py:2137>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2137
Fld_1 := Ent.Info;
--# expr-done 2185
Fld_2 := Fld_1.Md;
--# expr-done 2186
New_Struct := (From_Rebound => Fld_4, Md => Fld_2, Rebindings => Rebindings); 
--# expr-done 2187
New_Struct_1 := (Info => New_Struct, Node => Self); 
--# expr-done 2188
   If_Result_1 := New_Struct_1;
else
   If_Result_1 := raise Property_Error with "Incorrect rebindings";
end if;
--# expr-done 2189
   If_Result_2 := If_Result_1;
else
   --# expr-start 2191 '<FieldAccess for unshed_rebindings at ast.py:2144>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2144
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2190 '<FieldAccess for get_parent at ast.py:2144>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2144
      if Rebindings = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Rebindings.Parent;
--# expr-done 2190
--# property-call-start BasicDecl.unshed_rebindings
Fld_11 := Libadalang.Implementation.Basic_Decl_P_Unshed_Rebindings (Node => Ent.Node, Rebindings => Fld_10, E_Info => Ent.Info);
--# end
--# expr-done 2191
   If_Result_2 := Fld_11;
end if;
   If_Result_3 := If_Result_2;
end if;
--# expr-done 2192
Let_Result := If_Result_3; 
--# end
--# expr-done 2193
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_142;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_142;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Unshed_Rebindings;
--# end
--# property-start '[dispatcher]BasicDecl.is_library_item' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Decl_P_Is_Library_Item
  (Node : Bare_Basic_Decl
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Decl (Self.Kind) is
               when Ada_Abstract_State_Decl .. Ada_Synthetic_Formal_Param_Decl | Ada_Package_Decl .. Ada_Single_Task_Decl =>
                  --# property-call-start BasicDecl.is_library_item
                  Property_Result := Basic_Decl_P_Is_Library_Item
                    (Self
                    );
                  --# end
               when Ada_Generic_Package_Internal =>
                  --# property-call-start GenericPackageInternal.is_library_item
                  Property_Result := Generic_Package_Internal_P_Is_Library_Item
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Decl_P_Is_Library_Item;
--# end
--# property-start BasicDecl.decl_private_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2154
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Decl_Private_Part
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Private_Part
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Private_Part;
      Match_0 : Internal_Entity_Base_Package_Decl;
Match_1 : Internal_Entity_Task_Type_Decl;
Match_2 : Internal_Entity_Single_Task_Decl;
Match_3 : Internal_Entity_Protected_Type_Decl;
Match_4 : Internal_Entity_Single_Protected_Decl;
Match_5 : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Private_Part;
Fld_1 : Internal_Entity_Task_Def;
Fld_2 : Internal_Entity_Private_Part;
Fld_3 : Internal_Entity_Single_Task_Type_Decl;
Fld_4 : Internal_Entity_Task_Def;
Fld_5 : Internal_Entity_Private_Part;
Fld_6 : Internal_Entity_Protected_Def;
Fld_7 : Internal_Entity_Private_Part;
Fld_8 : Internal_Entity_Protected_Def;
Fld_9 : Internal_Entity_Private_Part;
Match_Prefix : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity_Base_Package_Decl;
Let_Result : Internal_Entity_Private_Part;
Scope_Result : Internal_Entity_Private_Part;
Cast_Result_1 : Internal_Entity_Task_Type_Decl;
Let_Result_1 : Internal_Entity_Private_Part;
Scope_Result_1 : Internal_Entity_Private_Part;
Cast_Result_2 : Internal_Entity_Single_Task_Decl;
Let_Result_2 : Internal_Entity_Private_Part;
Scope_Result_2 : Internal_Entity_Private_Part;
Cast_Result_3 : Internal_Entity_Protected_Type_Decl;
Let_Result_3 : Internal_Entity_Private_Part;
Scope_Result_3 : Internal_Entity_Private_Part;
Cast_Result_4 : Internal_Entity_Single_Protected_Decl;
Let_Result_4 : Internal_Entity_Private_Part;
Scope_Result_4 : Internal_Entity_Private_Part;
Cast_Result_5 : Internal_Entity_Basic_Decl;
Let_Result_5 : Internal_Entity_Private_Part;
Scope_Result_5 : Internal_Entity_Private_Part;
Match_Result : Internal_Entity_Private_Part;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2204 '<Match at ast.py:2154>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2154
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Basic_Decl (Match_Prefix.Node.Kind) is
         when Ada_Generic_Package_Internal .. Ada_Package_Decl =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Base_Package_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind bpd Match_0
--# expr-start 2194 '<FieldAccess for private_part at ast.py:2155>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2155
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Private_Part (Node => Match_0.Node.Base_Package_Decl_F_Private_Part, Info => Match_0.Info);
--# expr-done 2194
Let_Result := Fld; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Task_Type_Decl .. Ada_Single_Task_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Task_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind ttd Match_1
--# expr-start 2196 '<FieldAccess for private_part at ast.py:2156>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2156
--# expr-start 2195 '<FieldAccess for definition at ast.py:2156>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2156
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Task_Def (Node => Match_1.Node.Task_Type_Decl_F_Definition, Info => Match_1.Info);
--# expr-done 2195
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Private_Part (Node => Fld_1.Node.Task_Def_F_Private_Part, Info => Fld_1.Info);
--# expr-done 2196
Let_Result_1 := Fld_2; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Single_Task_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind td Match_2
--# expr-start 2199 '<FieldAccess for private_part at ast.py:2157>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2157
--# expr-start 2198 '<FieldAccess for definition at ast.py:2157>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2157
--# expr-start 2197 '<FieldAccess for task_type at ast.py:2157>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2157
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Single_Task_Type_Decl (Node => Match_2.Node.Single_Task_Decl_F_Task_Type, Info => Match_2.Info);
--# expr-done 2197
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Task_Def (Node => Fld_3.Node.Task_Type_Decl_F_Definition, Info => Fld_3.Info);
--# expr-done 2198
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Private_Part (Node => Fld_4.Node.Task_Def_F_Private_Part, Info => Fld_4.Info);
--# expr-done 2199
Let_Result_2 := Fld_5; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Protected_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Protected_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_3; 
--# bind ptd Match_3
--# expr-start 2201 '<FieldAccess for private_part at ast.py:2158>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2158
--# expr-start 2200 '<FieldAccess for definition at ast.py:2158>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2158
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Protected_Def (Node => Match_3.Node.Protected_Type_Decl_F_Definition, Info => Match_3.Info);
--# expr-done 2200
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Private_Part (Node => Fld_6.Node.Protected_Def_F_Private_Part, Info => Fld_6.Info);
--# expr-done 2201
Let_Result_3 := Fld_7; 
--# end
Scope_Result_3 := Let_Result_3;
   --# end
            Match_Result := Scope_Result_3; 
         when Ada_Single_Protected_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Single_Protected_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_4 := Cast_Result_4; 
--# bind spd Match_4
--# expr-start 2203 '<FieldAccess for private_part at ast.py:2159>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2159
--# expr-start 2202 '<FieldAccess for definition at ast.py:2159>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2159
      if Match_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Protected_Def (Node => Match_4.Node.Single_Protected_Decl_F_Definition, Info => Match_4.Info);
--# expr-done 2202
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Private_Part (Node => Fld_8.Node.Protected_Def_F_Private_Part, Info => Fld_8.Info);
--# expr-done 2203
Let_Result_4 := Fld_9; 
--# end
Scope_Result_4 := Let_Result_4;
   --# end
            Match_Result := Scope_Result_4; 
         when Ada_Abstract_State_Decl .. Ada_Synthetic_Formal_Param_Decl | Ada_Discrete_Base_Subtype_Decl .. Ada_Incomplete_Tagged_Type_Decl | Ada_Type_Decl .. Ada_Package_Renaming_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Create_Internal_Entity_Basic_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_5 := Cast_Result_5; 
--# bind _ Match_5
Let_Result_5 := No_Entity_Private_Part; 
--# end
Scope_Result_5 := Let_Result_5;
   --# end
            Match_Result := Scope_Result_5; 
end case;
--# expr-done 2204
      Property_Result := Match_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Decl_Private_Part;
--# end
--# property-start '[dispatcher]BasicDecl.declarative_parts' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Decl_P_Declarative_Parts
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Declarative_Part_Array_Access
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Declarative_Part_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Decl (Self.Kind) is
               when Ada_Abstract_State_Decl .. Ada_Synthetic_Formal_Param_Decl | Ada_Discrete_Base_Subtype_Decl .. Ada_Incomplete_Tagged_Type_Decl | Ada_Type_Decl .. Ada_Null_Subp_Decl | Ada_Subp_Renaming_Decl .. Ada_Task_Body_Stub | Ada_Entry_Index_Spec .. Ada_Package_Renaming_Decl | Ada_Single_Task_Decl =>
                  --# property-call-start BasicDecl.declarative_parts
                  Property_Result := Basic_Decl_P_Declarative_Parts
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Package_Internal .. Ada_Package_Decl =>
                  --# property-call-start BasePackageDecl.declarative_parts
                  Property_Result := Base_Package_Decl_P_Declarative_Parts
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Protected_Type_Decl =>
                  --# property-call-start ProtectedTypeDecl.declarative_parts
                  Property_Result := Protected_Type_Decl_P_Declarative_Parts
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Task_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start TaskTypeDecl.declarative_parts
                  Property_Result := Task_Type_Decl_P_Declarative_Parts
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Subp_Body =>
                  --# property-call-start SubpBody.declarative_parts
                  Property_Result := Subp_Body_P_Declarative_Parts
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Entry_Body =>
                  --# property-call-start EntryBody.declarative_parts
                  Property_Result := Entry_Body_P_Declarative_Parts
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Package_Body =>
                  --# property-call-start PackageBody.declarative_parts
                  Property_Result := Package_Body_P_Declarative_Parts
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Protected_Body =>
                  --# property-call-start ProtectedBody.declarative_parts
                  Property_Result := Protected_Body_P_Declarative_Parts
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Task_Body =>
                  --# property-call-start TaskBody.declarative_parts
                  Property_Result := Task_Body_P_Declarative_Parts
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Single_Protected_Decl =>
                  --# property-call-start SingleProtectedDecl.declarative_parts
                  Property_Result := Single_Protected_Decl_P_Declarative_Parts
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Decl_P_Declarative_Parts;
--# end
--# property-start BasicDecl.immediate_declarative_region /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2172
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Immediate_Declarative_Region
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_146 with Inline_Always;
            procedure Finalizer_Scope_2025 with Inline_Always;
            procedure Finalizer_Scope_2026 with Inline_Always;
      Fld : Internal_Entity_Basic_Decl_Array_Access;
Singleton : Internal_Entity_Basic_Decl_Array_Access;
Concat_Result : Internal_Entity_Basic_Decl_Array_Access;
Fld_1 : Internal_Entity_Declarative_Part_Array_Access;
Fld_2 : Lexical_Env := Empty_Env;
Map_Result : Lexical_Env_Array_Access;
Map_Result_1 : Lexical_Env_Array_Access;
Group_Env : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_146 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Singleton);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Group_Env);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_146;
            procedure Finalizer_Scope_2025 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_2025;
            procedure Finalizer_Scope_2026 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_2026;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Basic_Decl_Basic_Decl_P_Immediate_Declarative_Region,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Basic_Decl,
                                As_Bare_Basic_Decl => Self);
               Mmz_K.Items (2) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Lexical_Env;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 2212 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2211 '<env_group at ast.py:2173>' Group_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2173
--# expr-start 2210 '<Mapcat at ast.py:2173>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2173
   --# expr-start 2206 '<Concat at ast.py:2173>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2173
--# expr-start 2205 '<FieldAccess for all_previous_parts at ast.py:2173>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2173
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.all_previous_parts
Fld := Libadalang.Implementation.Basic_Decl_P_All_Previous_Parts (Node => Ent.Node, Imprecise_Fallback => False, E_Info => Ent.Info);
--# end
--# expr-done 2205
                Singleton := Create_Internal_Entity_Basic_Decl_Array (Items_Count => 1);
                Singleton.Items (1) := Ent;
Concat_Result := Concat (Fld, Singleton); 
--# expr-done 2206
      declare
         Map_Result_1_Vec : Lexical_Env_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Basic_Decl_Array_Access := Concat_Result;
      begin
         for Item_95 of
               Collection.Items
         loop
   --# scope-start
               --# bind part Item_95
      --# expr-start 2209 '<Map at ast.py:2174>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2174
   --# expr-start 2207 '<FieldAccess for declarative_parts at ast.py:2174>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2174
      if Item_95.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.declarative_parts'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Declarative_Parts (Node => Item_95.Node, E_Info => Item_95.Info);
--# end
--# expr-done 2207
      declare
         Map_Result_Vec : Lexical_Env_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Declarative_Part_Array_Access := Fld_1;
      begin
         for Item_96 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_96
      --# expr-start 2208 '<FieldAccess for children_env at ast.py:2175>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2175
      if Item_96.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Children_Env (Node => Item_96.Node, E_Info => Item_96.Info);
--# expr-done 2208
         declare
            Item_To_Append : constant Lexical_Env := Fld_2;
         begin
               Inc_Ref (Item_To_Append);
            Lexical_Env_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2026;
         end loop;
      end;
         Map_Result := Create_Lexical_Env_Array
           (Items_Count => Natural (Lexical_Env_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Lexical_Env_Vectors.Get
              (Map_Result_Vec,
               I + Lexical_Env_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Lexical_Env_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 2209
         for Item_To_Append of
               Map_Result.Items
         loop
               Inc_Ref (Item_To_Append);
            Lexical_Env_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_2025;
         end loop;
      end;
         Map_Result_1 := Create_Lexical_Env_Array
           (Items_Count => Natural (Lexical_Env_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Lexical_Env_Vectors.Get
              (Map_Result_1_Vec,
               I + Lexical_Env_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Lexical_Env_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 2210
Group_Env := Group (Map_Result_1, No_Metadata); 
--# expr-done 2211
Let_Result := Group_Env; Inc_Ref (Let_Result);
--# end
--# expr-done 2212
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_146;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Lexical_Env,
                        As_Lexical_Env => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_146;
               Finalizer_Scope_2025;
               Finalizer_Scope_2026;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Immediate_Declarative_Region;
--# end
--# property-start '[dispatcher]BasicDecl.get_aspect_assoc' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Decl_P_Get_Aspect_Assoc
  (Node : Bare_Basic_Decl
      ; Name : Symbol_Type
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Aspect_Assoc
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind name Name
   Property_Result : Internal_Entity_Aspect_Assoc;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Decl (Self.Kind) is
               when Ada_Abstract_State_Decl .. Ada_Subtype_Decl | Ada_Incomplete_Type_Decl .. Ada_For_Loop_Var_Decl | Ada_Generic_Package_Instantiation .. Ada_Single_Task_Decl =>
                  --# property-call-start BasicDecl.get_aspect_assoc
                  Property_Result := Basic_Decl_P_Get_Aspect_Assoc
                    (Self
                        , Name
                        , E_Info
                    );
                  --# end
               when Ada_Classwide_Type_Decl =>
                  --# property-call-start ClasswideTypeDecl.get_aspect_assoc
                  Property_Result := Classwide_Type_Decl_P_Get_Aspect_Assoc
                    (Self
                        , Name
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Package_Decl .. Ada_Generic_Subp_Decl =>
                  --# property-call-start GenericDecl.get_aspect_assoc
                  Property_Result := Generic_Decl_P_Get_Aspect_Assoc
                    (Self
                        , Name
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Decl_P_Get_Aspect_Assoc;
--# end
--# property-start BasicDecl.get_aspect_spec_expr /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2193
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Get_Aspect_Spec_Expr
  (Node : Bare_Basic_Decl
      ; Name : Symbol_Type
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Expr
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind name Name
   Property_Result : Internal_Entity_Expr;
      Var_Expr : Internal_Entity_Aspect_Assoc;
Fld : Internal_Entity_Aspect_Assoc;
Fld_1 : Internal_Entity_Expr;
Scope_Result : Internal_Entity_Expr;
Result_Var : Internal_Entity_Expr;
Let_Result : Internal_Entity_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2215 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2213 '<FieldAccess for get_aspect_assoc at ast.py:2198>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2198
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.get_aspect_assoc'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Get_Aspect_Assoc (Node => Ent.Node, Name => Name, E_Info => Ent.Info);
--# end
--# expr-done 2213
Var_Expr := Fld;
if Var_Expr /= No_Entity_Aspect_Assoc then
   --# scope-start
--# expr-start 2214 '<FieldAccess for expr at ast.py:2198>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2198
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Expr (Node => Var_Expr.Node.Aspect_Assoc_F_Expr, Info => Var_Expr.Info);
--# expr-done 2214
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Expr;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 2215
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Get_Aspect_Spec_Expr;
--# end
--# property-start BasicDecl.library_item_pragmas /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2201
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Library_Item_Pragmas
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Pragma_Node_List
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Pragma_Node_List;
      Fld : Internal_Entity;
Is_A : Boolean;
Fld_1 : Internal_Entity;
Fld_2 : Internal_Entity;
Cast_Result : Internal_Entity_Compilation_Unit;
Fld_3 : Internal_Entity_Pragma_Node_List;
Fld_4 : Internal_Entity;
Fld_5 : Internal_Entity;
Is_A_1 : Boolean;
Fld_6 : Internal_Entity;
Fld_7 : Internal_Entity;
Fld_8 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Compilation_Unit;
Fld_9 : Internal_Entity_Pragma_Node_List;
If_Result : Internal_Entity_Pragma_Node_List;
If_Result_1 : Internal_Entity_Pragma_Node_List;
Let_Result : Internal_Entity_Pragma_Node_List;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2231 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2230 '<Cond at ast.py:2205>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2205
--# expr-start 2217 '<IsA LibraryItem at ast.py:2206>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2206
--# expr-start 2216 '<parent at ast.py:2206>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2206
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 2216
Is_A := Fld.Node /= null 
and then Fld.Node.Kind in Ada_Library_Item_Range; 
--# expr-done 2217
if Is_A then
   --# expr-start 2221 '<FieldAccess for pragmas at ast.py:2207>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2207
--# expr-start 2220 '<Cast to CompilationUnit at ast.py:2207>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2207
--# expr-start 2219 '<parent at ast.py:2207>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2207
--# expr-start 2218 '<parent at ast.py:2207>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2207
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 2218
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Parent (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# expr-done 2219
   if Fld_2.Node = null
      or else Fld_2.Node.Kind in Ada_Compilation_Unit_Range
   then
      Cast_Result := Create_Internal_Entity_Compilation_Unit
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
   else
         Cast_Result := No_Entity_Compilation_Unit;
   end if;
--# expr-done 2220
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Pragma_Node_List (Node => Cast_Result.Node.Compilation_Unit_F_Pragmas, Info => Cast_Result.Info);
--# expr-done 2221
   If_Result_1 := Fld_3;
else
--# expr-start 2224 '<IsA LibraryItem at ast.py:2209>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2209
--# expr-start 2223 '<parent at ast.py:2209>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2209
--# expr-start 2222 '<parent at ast.py:2209>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2209
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 2222
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Parent (Node => Fld_4.Node, E_Info => Fld_4.Info);
--# expr-done 2223
Is_A_1 := Fld_5.Node /= null 
and then Fld_5.Node.Kind in Ada_Library_Item_Range; 
--# expr-done 2224
if Is_A_1 then
   --# expr-start 2229 '<FieldAccess for pragmas at ast.py:2210>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2210
--# expr-start 2228 '<Cast to CompilationUnit at ast.py:2210>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2210
--# expr-start 2227 '<parent at ast.py:2210>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2210
--# expr-start 2226 '<parent at ast.py:2210>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2210
--# expr-start 2225 '<parent at ast.py:2210>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2210
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 2225
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Parent (Node => Fld_6.Node, E_Info => Fld_6.Info);
--# expr-done 2226
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Parent (Node => Fld_7.Node, E_Info => Fld_7.Info);
--# expr-done 2227
   if Fld_8.Node = null
      or else Fld_8.Node.Kind in Ada_Compilation_Unit_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Compilation_Unit
        (Node => Fld_8.Node,
         Info => Fld_8.Info);
   else
         Cast_Result_1 := No_Entity_Compilation_Unit;
   end if;
--# expr-done 2228
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Pragma_Node_List (Node => Cast_Result_1.Node.Compilation_Unit_F_Pragmas, Info => Cast_Result_1.Info);
--# expr-done 2229
   If_Result := Fld_9;
else
   If_Result := No_Entity_Pragma_Node_List;
end if;
   If_Result_1 := If_Result;
end if;
--# expr-done 2230
Let_Result := If_Result_1; 
--# end
--# expr-done 2231
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Library_Item_Pragmas;
--# end
--# property-start BasicDecl.get_aspect /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2217
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Get_Aspect
  (Node : Bare_Basic_Decl
      ; Name : Symbol_Type
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Aspect
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind name Name
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Aspect;
      Var_Expr : Internal_Entity_Defining_Name;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Internal_Aspect;
Scope_Result : Internal_Aspect;
Result_Var : Internal_Aspect;
Let_Result : Internal_Aspect;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2234 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2232 '<FieldAccess for defining_name_or_raise at ast.py:2230>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2230
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name_or_raise
Fld := Libadalang.Implementation.Basic_Decl_P_Defining_Name_Or_Raise (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2232
Var_Expr := Fld;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 2233 '<FieldAccess for get_aspect at ast.py:2230>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2230
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.get_aspect
Fld_1 := Libadalang.Implementation.Defining_Name_P_Get_Aspect (Node => Var_Expr.Node, Name => Name, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr.Info);
--# end
--# expr-done 2233
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Aspect;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 2234
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Get_Aspect;
--# end
--# property-start BasicDecl.has_aspect /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2234
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Has_Aspect
  (Node : Bare_Basic_Decl
      ; Name : Symbol_Type
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind name Name
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Defining_Name;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2237 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2235 '<FieldAccess for defining_name_or_raise at ast.py:2241>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2241
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name_or_raise
Fld := Libadalang.Implementation.Basic_Decl_P_Defining_Name_Or_Raise (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2235
Var_Expr := Fld;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 2236 '<FieldAccess for has_aspect at ast.py:2241>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2241
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.has_aspect
Fld_1 := Libadalang.Implementation.Defining_Name_P_Has_Aspect (Node => Var_Expr.Node, Name => Name, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr.Info);
--# end
--# expr-done 2236
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 2237
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Has_Aspect;
--# end
--# property-start BasicDecl.get_pragma /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2244
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Get_Pragma
  (Node : Bare_Basic_Decl
      ; Name : Symbol_Type
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Pragma_Node
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind name Name
   Property_Result : Internal_Entity_Pragma_Node;
      Var_Expr : Internal_Entity_Defining_Name;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Internal_Entity_Pragma_Node;
Scope_Result : Internal_Entity_Pragma_Node;
Result_Var : Internal_Entity_Pragma_Node;
Let_Result : Internal_Entity_Pragma_Node;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2240 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2238 '<FieldAccess for defining_name_or_raise at ast.py:2252>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2252
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name_or_raise
Fld := Libadalang.Implementation.Basic_Decl_P_Defining_Name_Or_Raise (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2238
Var_Expr := Fld;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 2239 '<FieldAccess for get_pragma at ast.py:2252>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2252
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.get_pragma
Fld_1 := Libadalang.Implementation.Defining_Name_P_Get_Pragma (Node => Var_Expr.Node, Name => Name, E_Info => Var_Expr.Info);
--# end
--# expr-done 2239
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Pragma_Node;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 2240
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Get_Pragma;
--# end
--# property-start BasicDecl.get_representation_clause /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2256
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Get_Representation_Clause
  (Node : Bare_Basic_Decl
      ; Name : Symbol_Type
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Attribute_Def_Clause
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind name Name
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Attribute_Def_Clause;
      Var_Expr : Internal_Entity_Defining_Name;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Internal_Entity_Attribute_Def_Clause;
Scope_Result : Internal_Entity_Attribute_Def_Clause;
Result_Var : Internal_Entity_Attribute_Def_Clause;
Let_Result : Internal_Entity_Attribute_Def_Clause;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2243 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2241 '<FieldAccess for defining_name_or_raise at ast.py:2261>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2261
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name_or_raise
Fld := Libadalang.Implementation.Basic_Decl_P_Defining_Name_Or_Raise (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2241
Var_Expr := Fld;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 2242 '<FieldAccess for get_representation_clause at ast.py:2261>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2261
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.get_representation_clause
Fld_1 := Libadalang.Implementation.Defining_Name_P_Get_Representation_Clause (Node => Var_Expr.Node, Name => Name, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr.Info);
--# end
--# expr-done 2242
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Attribute_Def_Clause;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 2243
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Get_Representation_Clause;
--# end
--# property-start BasicDecl.get_at_clause /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2265
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Get_At_Clause
  (Node : Bare_Basic_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_At_Clause
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_At_Clause;
      Var_Expr : Internal_Entity_Defining_Name;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Internal_Entity_At_Clause;
Scope_Result : Internal_Entity_At_Clause;
Result_Var : Internal_Entity_At_Clause;
Let_Result : Internal_Entity_At_Clause;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2246 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2244 '<FieldAccess for defining_name_or_raise at ast.py:2269>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2269
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name_or_raise
Fld := Libadalang.Implementation.Basic_Decl_P_Defining_Name_Or_Raise (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2244
Var_Expr := Fld;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 2245 '<FieldAccess for get_at_clause at ast.py:2269>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2269
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.get_at_clause
Fld_1 := Libadalang.Implementation.Defining_Name_P_Get_At_Clause (Node => Var_Expr.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr.Info);
--# end
--# expr-done 2245
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_At_Clause;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 2246
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Get_At_Clause;
--# end
--# property-start '[dispatcher]BasicDecl.is_imported' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Decl_P_Is_Imported
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Decl (Self.Kind) is
               when Ada_Abstract_State_Decl .. Ada_Generic_Package_Decl | Ada_Generic_Package_Instantiation .. Ada_Single_Task_Decl =>
                  --# property-call-start BasicDecl.is_imported
                  Property_Result := Basic_Decl_P_Is_Imported
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Subp_Decl =>
                  --# property-call-start GenericSubpDecl.is_imported
                  Property_Result := Generic_Subp_Decl_P_Is_Imported
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Decl_P_Is_Imported;
--# end
--# property-start BasicDecl.is_ghost_code /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2279
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Is_Ghost_Code
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Defining_Name;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2249 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2247 '<FieldAccess for defining_name_or_raise at ast.py:2283>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2283
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name_or_raise
Fld := Libadalang.Implementation.Basic_Decl_P_Defining_Name_Or_Raise (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2247
Var_Expr := Fld;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 2248 '<FieldAccess for is_ghost_code at ast.py:2283>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2283
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.is_ghost_code
Fld_1 := Libadalang.Implementation.Defining_Name_P_Is_Ghost_Code (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 2248
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 2249
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Is_Ghost_Code;
--# end
--# property-start BasicDecl.get_instantiation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2286
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Get_Instantiation
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Generic_Instantiation
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Generic_Instantiation;
            procedure Finalizer_Scope_2037 with Inline_Always;
      Block_Var : Bare_Generic_Instantiation;
Block_Var_1 : Internal_Entity_Basic_Decl;
Var_Expr : Env_Rebindings;
Var_Expr_1 : Internal_Entity_Generic_Instantiation;
Fld : Internal_Entity_Info;
Fld_1 : Env_Rebindings;
Fld_2 : Lexical_Env := Empty_Env;
Env_Node : Bare_Ada_Node;
Cast_Result : Bare_Generic_Instantiation;
Scope_Result : Bare_Generic_Instantiation;
Result_Var : Bare_Generic_Instantiation;
As_Entity : Internal_Entity_Generic_Instantiation;
Is_Null : Boolean;
If_Result : Internal_Entity_Generic_Instantiation;
Fld_3 : Internal_Entity_Basic_Decl;
Scope_Result_1 : Internal_Entity_Basic_Decl;
Result_Var_1 : Internal_Entity_Basic_Decl;
Fld_4 : Internal_Entity_Info;
Fld_5 : Env_Rebindings;
Fld_6 : Internal_Entity_Info;
Fld_7 : Env_Rebindings;
Fld_8 : Env_Rebindings;
Fld_9 : Internal_Entity_Info;
Fld_10 : Boolean;
New_Struct : Internal_Metadata;
New_Struct_1 : Internal_Entity_Info;
New_Struct_2 : Internal_Entity_Generic_Instantiation;
Fld_11 : Bare_Basic_Decl;
Fld_12 : Bare_Ada_Node;
Cast_Result_1 : Bare_Ada_Node;
Is_Equal : Boolean;
If_Result_1 : Internal_Entity_Generic_Instantiation;
Let_Result : Internal_Entity_Generic_Instantiation;
            procedure Finalizer_Scope_2037 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_2037;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2271 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2251 '<FieldAccess for rebindings at ast.py:2298>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2298
--# expr-start 2250 '<FieldAccess for info at ast.py:2298>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2298
Fld := Ent.Info;
--# expr-done 2250
Fld_1 := Fld.Rebindings;
--# expr-done 2251
Var_Expr := Fld_1;
if Var_Expr /= null then
   --# scope-start
   --# bind r Var_Expr
--# expr-start 2254 '<Cast to GenericInstantiation at ast.py:2299>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2299
--# expr-start 2253 '<env_node at ast.py:2299>' Env_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2299
--# expr-start 2252 '<FieldAccess for new_env at ast.py:2299>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2299
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Var_Expr.New_Env;
Inc_Ref (Fld_2);
--# expr-done 2252
Env_Node := AST_Envs.Env_Node (Fld_2); 
--# expr-done 2253
   if Env_Node = null
      or else Env_Node.Kind in Ada_Generic_Instantiation
   then
      Cast_Result := Env_Node;
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 2254
Scope_Result := Cast_Result;
   --# end
      Finalizer_Scope_2037;
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node;
end if;
Block_Var := Result_Var; 
--# bind inst_node Block_Var
--# expr-start 2255 '<as_bare_entity at ast.py:2302>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2302
Is_Null := Block_Var = null; 
if Is_Null then
   If_Result := No_Entity_Generic_Instantiation;
else
As_Entity := (Info => No_Entity_Info, Node => Block_Var); 
   If_Result := As_Entity;
end if;
--# expr-done 2255
Var_Expr_1 := If_Result;
if Var_Expr_1 /= No_Entity_Generic_Instantiation then
   --# scope-start
--# expr-start 2256 '<FieldAccess for designated_generic_decl at ast.py:2302>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2302
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericInstantiation.designated_generic_decl'
Fld_3 := Libadalang.Implementation.Dispatcher_Generic_Instantiation_P_Designated_Generic_Decl (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 2256
Scope_Result_1 := Fld_3;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Basic_Decl;
end if;
Block_Var_1 := Result_Var_1; 
--# bind designated_decl Block_Var_1
--# expr-start 2270 '<If at ast.py:2304>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2304
--# expr-start 2259 '<Eq at ast.py:2305>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2305
--# expr-start 2257 '<FieldAccess for node at ast.py:2305>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2305
Fld_11 := Block_Var_1.Node;
--# expr-done 2257
      Cast_Result_1 := Fld_11;
--# expr-start 2258 '<parent at ast.py:2305>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2305
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Self.Parent; 
--# expr-done 2258
Is_Equal := Cast_Result_1 = Fld_12; 
--# expr-done 2259
if Is_Equal then
   --# expr-start 2269 '<New GenericInstantiation.entity at ???>' New_Struct_2 None
--# expr-start 2268 '<New EntityInfo at ast.py:2308>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2308
--# expr-start 2261 '<FieldAccess for from_rebound at ast.py:2315>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2315
--# expr-start 2260 '<FieldAccess for info at ast.py:2315>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2315
Fld_9 := Ent.Info;
--# expr-done 2260
Fld_10 := Fld_9.From_Rebound;
--# expr-done 2261
--# expr-start 2262 '<New Metadata at ast.py:2316>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2316
New_Struct := (Dottable_Subp => False, Primitive => No_Bare_Ada_Node, Primitive_Real_Type => No_Bare_Ada_Node); 
--# expr-done 2262
--# expr-start 2267 '<FieldAccess for remove_rebindings at ast.py:2311>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2311
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2264 '<FieldAccess for rebindings at ast.py:2312>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2312
--# expr-start 2263 '<FieldAccess for info at ast.py:2312>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2312
Fld_4 := Ent.Info;
--# expr-done 2263
Fld_5 := Fld_4.Rebindings;
--# expr-done 2264
--# expr-start 2266 '<FieldAccess for rebindings at ast.py:2313>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2313
--# expr-start 2265 '<FieldAccess for info at ast.py:2313>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2313
Fld_6 := Block_Var_1.Info;
--# expr-done 2265
Fld_7 := Fld_6.Rebindings;
--# expr-done 2266
--# property-call-start AdaNode.remove_rebindings
Fld_8 := Libadalang.Implementation.Ada_Node_P_Remove_Rebindings (Node => Self, Base => Fld_5, Suffix => Fld_7);
--# end
--# expr-done 2267
New_Struct_1 := (From_Rebound => Fld_10, Md => New_Struct, Rebindings => Fld_8); 
--# expr-done 2268
New_Struct_2 := (Info => New_Struct_1, Node => Block_Var); 
--# expr-done 2269
   If_Result_1 := New_Struct_2;
else
   If_Result_1 := No_Entity_Generic_Instantiation;
end if;
--# expr-done 2270
Let_Result := If_Result_1; 
--# end
--# expr-done 2271
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_2037;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Get_Instantiation;
--# end
--# property-start BasicDecl.is_compilation_unit_root /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2323
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Is_Compilation_Unit_Root
  (Node : Bare_Basic_Decl
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Var_Expr : Bare_Ada_Node;
Match_0 : Bare_Library_Item;
Match_1 : Bare_Generic_Package_Decl;
Match_2 : Bare_Subunit;
Match_3 : Bare_Ada_Node;
Var_Expr_1 : Bare_Ada_Node;
Fld : Bare_Ada_Node;
Fld_1 : Bare_Ada_Node;
Is_A : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Match_Prefix : Bare_Ada_Node;
Cast_Result : Bare_Library_Item;
Let_Result : Boolean;
Scope_Result_1 : Boolean;
Cast_Result_1 : Bare_Generic_Package_Decl;
Let_Result_1 : Boolean;
Scope_Result_2 : Boolean;
Cast_Result_2 : Bare_Subunit;
Let_Result_2 : Boolean;
Scope_Result_3 : Boolean;
Cast_Result_3 : Bare_Ada_Node;
Let_Result_3 : Boolean;
Scope_Result_4 : Boolean;
Match_Result : Boolean;
Scope_Result_5 : Boolean;
Result_Var_1 : Boolean;
Let_Result_4 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2276 '<Block at ???>' Let_Result_4 None
--# scope-start
--# expr-start 2272 '<parent at ast.py:2327>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2327
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 2272
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind p Var_Expr
--# expr-start 2275 '<Match at ast.py:2327>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2327
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Var_Expr; 
case Ada_Ada_Node (Match_Prefix.Kind) is
         when Ada_Library_Item =>
   --# scope-start
--# scope-start
      Cast_Result := Match_Prefix;
Match_0 := Cast_Result; 
--# bind _ Match_0
Let_Result := True; 
--# end
Scope_Result_1 := Let_Result;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Generic_Package_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Match_Prefix;
Match_1 := Cast_Result_1; 
--# bind gen_pkg_decl Match_1
--# expr-start 2273 '<parent at ast.py:2330>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2330
      if Match_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Match_1.Parent; 
--# expr-done 2273
Var_Expr_1 := Fld_1;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind p Var_Expr_1
--# expr-start 2274 '<IsA LibraryItem at ast.py:2330>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2330
Is_A := Var_Expr_1 /= null 
and then Var_Expr_1.Kind in Ada_Library_Item_Range; 
--# expr-done 2274
Scope_Result := Is_A;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Let_Result_1 := Result_Var; 
--# end
Scope_Result_2 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Subunit =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Match_Prefix;
Match_2 := Cast_Result_2; 
--# bind _ Match_2
Let_Result_2 := True; 
--# end
Scope_Result_3 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_3; 
         when Ada_Abort_Absent .. Ada_For_Loop_Var_Decl | Ada_Generic_Subp_Decl .. Ada_Iter_Type_Of | Ada_Limited_Absent .. Ada_Subp_Kind_Procedure | Ada_Synchronized_Absent .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Match_Prefix;
Match_3 := Cast_Result_3; 
--# bind _ Match_3
Let_Result_3 := False; 
--# end
Scope_Result_4 := Let_Result_3;
   --# end
            Match_Result := Scope_Result_4; 
end case;
--# expr-done 2275
Scope_Result_5 := Match_Result;
   --# end
   Result_Var_1 := Scope_Result_5;
else
   Result_Var_1 := False;
end if;
Let_Result_4 := Result_Var_1; 
--# end
--# expr-done 2276
      Property_Result := Let_Result_4;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Is_Compilation_Unit_Root;
--# end
--# property-start BasicDecl.populate_dependent_units /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2336
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Populate_Dependent_Units
  (Node : Bare_Basic_Decl
  )
   return Bare_Compilation_Unit_Array_Access
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_159 with Inline_Always;
      Fld : Bare_Name_Array_Access;
Fld_1 : Bare_Compilation_Unit;
Map_Result : Bare_Compilation_Unit_Array_Access;
Fld_2 : Boolean;
If_Result : Bare_Compilation_Unit_Array_Access;
Let_Result : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_159 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Map_Result);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_159;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2282 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2281 '<If at ast.py:2337>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2337
--# expr-start 2277 '<FieldAccess for is_compilation_unit_root at ast.py:2338>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2338
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_compilation_unit_root
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Is_Compilation_Unit_Root (Node => Self);
--# end
--# expr-done 2277
if Fld_2 then
   --# expr-start 2280 '<Map at ast.py:2339>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2339
   --# expr-start 2278 '<FieldAccess for top_level_with_package_clauses at ast.py:2339>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2339
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_with_package_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_With_Package_Clauses (Node => Self);
--# end
--# expr-done 2278
      declare
         Map_Result_Vec : Bare_Compilation_Unit_Vectors.Vector;
      begin
      declare
         Collection : constant Bare_Name_Array_Access := Fld;
      begin
         for Item_98 of
               Collection.Items
         loop
   --# scope-start
               --# bind package_name Item_98
      --# expr-start 2279 '<FieldAccess for withed_unit_helper at ast.py:2341>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2341
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.withed_unit_helper
Fld_1 := Libadalang.Implementation.Ada_Node_P_Withed_Unit_Helper (Node => Self, Unit_Name => Item_98);
--# end
--# expr-done 2279
         declare
            Item_To_Append : constant Bare_Compilation_Unit := Fld_1;
         begin
            Bare_Compilation_Unit_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Bare_Compilation_Unit_Array
           (Items_Count => Natural (Bare_Compilation_Unit_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Bare_Compilation_Unit_Vectors.Get
              (Map_Result_Vec,
               I + Bare_Compilation_Unit_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Bare_Compilation_Unit_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 2280
   If_Result := Map_Result;
else
   If_Result := No_Bare_Compilation_Unit_Array_Type;
end if;
      Inc_Ref (If_Result);
--# expr-done 2281
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 2282
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_159;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_159;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Populate_Dependent_Units;
--# end
--# property-start BasicDecl.should_ref_generic_formals /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2347
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Should_Ref_Generic_Formals
  (Node : Bare_Basic_Decl
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
Not_Val : Boolean;
Fld_1 : Bare_Ada_Node;
Is_A : Boolean;
If_Result : Boolean;
Is_A_1 : Boolean;
If_Result_1 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2288 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2284 '<Not at ast.py:2357>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2357
--# expr-start 2283 '<FieldAccess for is_compilation_unit_root at ast.py:2357>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2357
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_compilation_unit_root
Fld := Libadalang.Implementation.Basic_Decl_P_Is_Compilation_Unit_Root (Node => Self);
--# end
--# expr-done 2283
Not_Val := not (Fld); 
--# expr-done 2284
if Not_Val then
   If_Result := True;
else
   --# expr-start 2286 '<IsA Subunit at ast.py:2362>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2362
--# expr-start 2285 '<parent at ast.py:2362>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2362
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Self.Parent; 
--# expr-done 2285
Is_A := Fld_1 /= null 
and then Fld_1.Kind in Ada_Subunit_Range; 
--# expr-done 2286
   If_Result := Is_A;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 2287 '<IsA BaseSubpBody at ast.py:2368>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2368
Is_A_1 := Self /= null 
and then Self.Kind in Ada_Base_Subp_Body; 
--# expr-done 2287
   If_Result_1 := Is_A_1;
end if;
Let_Result := If_Result_1; 
--# end
--# expr-done 2288
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Should_Ref_Generic_Formals;
--# end
--# property-start BasicDecl.is_in_public_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2371
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Is_In_Public_Part
  (Node : Bare_Basic_Decl
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Fld_1 : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2291 '<IsA PublicPart at ast.py:2371>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2371
--# expr-start 2290 '<parent at ast.py:2371>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2371
--# expr-start 2289 '<parent at ast.py:2371>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2371
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 2289
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Parent; 
--# expr-done 2290
Is_A := Fld_1 /= null 
and then Fld_1.Kind in Ada_Public_Part_Range; 
--# expr-done 2291
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Is_In_Public_Part;
--# end
--# property-start '[dispatcher]BasicDecl.is_in_private_part' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Decl_P_Is_In_Private_Part
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Decl (Self.Kind) is
               when Ada_Abstract_State_Decl .. Ada_Subtype_Decl | Ada_Incomplete_Type_Decl .. Ada_Single_Task_Decl =>
                  --# property-call-start BasicDecl.is_in_private_part
                  Property_Result := Basic_Decl_P_Is_In_Private_Part
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Classwide_Type_Decl =>
                  --# property-call-start ClasswideTypeDecl.is_in_private_part
                  Property_Result := Classwide_Type_Decl_P_Is_In_Private_Part
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Decl_P_Is_In_Private_Part;
--# end
--# property-start BasicDecl.is_visible /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2375
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Is_Visible
  (Node : Bare_Basic_Decl
      ; From_Node : Internal_Entity
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind from_node From_Node
   Property_Result : Boolean;
            procedure Finalizer_Scope_163 with Inline_Always;
      Var_Expr : Internal_Entity_Body_Node;
Fld : Boolean;
Fld_1 : Lexical_Env := Empty_Env;
Fld_2 : Symbol_Type;
Env_Get_Result : Internal_Entity_Array_Access;
Cast_Expr : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity;
Is_Equiv : Boolean;
Quantifier_Result : Boolean;
Fld_3 : Internal_Entity;
Fld_4 : Internal_Entity;
Fld_5 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Package_Decl;
Fld_6 : Boolean;
Fld_7 : Boolean;
If_Result : Boolean;
Not_Val : Boolean;
If_Result_1 : Boolean;
Fld_8 : Boolean;
Fld_9 : Internal_Entity_Basic_Decl;
Fld_10 : Boolean;
Fld_11 : Boolean;
Cast_Expr_1 : Internal_Entity_Basic_Decl;
Cast_Result_2 : Internal_Entity_Body_Node;
Fld_12 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Not_Val_1 : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Boolean;
If_Result_4 : Boolean;
If_Result_5 : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_163 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Env_Get_Result);
            end Finalizer_Scope_163;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2313 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2312 '<Cond at ast.py:2384>' If_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2384
--# expr-start 2292 '<FieldAccess for is_in_private_part at ast.py:2387>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2387
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_in_private_part'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_In_Private_Part (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2292
if Fld then
--# expr-start 2297 '<Contains at ast.py:2388>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2388
   --# expr-start 2295 '<EnvGet at ast.py:2388>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2388
--# expr-start 2293 '<FieldAccess for node_env at ast.py:2388>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2388
      if From_Node.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Node_Env (Node => From_Node.Node, E_Info => From_Node.Info);
--# expr-done 2293
--# expr-start 2294 '<FieldAccess for name_symbol at ast.py:2388>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2388
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 2294
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Fld_1, Key => Fld_2, Lookup_Kind => To_Lookup_Kind_Type (Recursive), Categories => (Inherited_Primitives => True, Nocat => True))); 
--# expr-done 2295
      Quantifier_Result := False;
      declare
         Collection : constant Internal_Entity_Array_Access := Env_Get_Result;
      begin
         for Item_99 of
               Collection.Items
         loop
   --# scope-start
               --# bind item Item_99
      --# expr-start 2296 '<Eq at ???>' Is_Equiv None
Cast_Expr := Ent; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Is_Equiv := Equivalent (Item_99, Cast_Result); 
--# expr-done 2296
      Quantifier_Result := Is_Equiv;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
--# expr-done 2297
if Quantifier_Result then
   --# expr-start 2304 '<Not at ast.py:2395>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2395
--# expr-start 2302 '<FieldAccess for is_compilation_unit_root at ast.py:2396>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2396
--# expr-start 2301 '<Cast to PackageDecl at ast.py:2396>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2396
--# expr-start 2300 '<parent at ast.py:2396>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2396
--# expr-start 2299 '<parent at ast.py:2396>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2396
--# expr-start 2298 '<parent at ast.py:2396>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2396
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 2298
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Parent (Node => Fld_3.Node, E_Info => Fld_3.Info);
--# expr-done 2299
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Parent (Node => Fld_4.Node, E_Info => Fld_4.Info);
--# expr-done 2300
   if Fld_5.Node = null
      or else Fld_5.Node.Kind in Ada_Package_Decl_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Package_Decl
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
   else
         Cast_Result_1 := No_Entity_Package_Decl;
   end if;
--# expr-done 2301
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_compilation_unit_root
Fld_6 := Libadalang.Implementation.Basic_Decl_P_Is_Compilation_Unit_Root (Node => Cast_Result_1.Node);
--# end
--# expr-done 2302
if Fld_6 then
   --# expr-start 2303 '<FieldAccess for is_in_top_level_public_part at ast.py:2398>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2398
      if From_Node.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.is_in_top_level_public_part
Fld_7 := Libadalang.Implementation.Ada_Node_P_Is_In_Top_Level_Public_Part (Node => From_Node.Node, E_Info => From_Node.Info);
--# end
--# expr-done 2303
   If_Result := Fld_7;
else
   If_Result := False;
end if;
Not_Val := not (If_Result); 
--# expr-done 2304
   If_Result_1 := Not_Val;
else
   If_Result_1 := False;
end if;
   If_Result_5 := If_Result_1;
else
--# expr-start 2305 '<FieldAccess for is_in_public_part at ast.py:2404>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2404
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_in_public_part
Fld_8 := Libadalang.Implementation.Basic_Decl_P_Is_In_Public_Part (Node => Ent.Node);
--# end
--# expr-done 2305
if Fld_8 then
   --# expr-start 2307 '<FieldAccess for is_visible at ast.py:2405>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2405
--# expr-start 2306 '<FieldAccess for parent_basic_decl at ast.py:2405>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2405
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.parent_basic_decl
Fld_9 := Libadalang.Implementation.Ada_Node_P_Parent_Basic_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2306
      if Fld_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_visible
Fld_10 := Libadalang.Implementation.Basic_Decl_P_Is_Visible (Node => Fld_9.Node, From_Node => From_Node, E_Info => Fld_9.Info);
--# end
--# expr-done 2307
   If_Result_4 := Fld_10;
else
--# expr-start 2308 '<FieldAccess for is_compilation_unit_root at ast.py:2410>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2410
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_compilation_unit_root
Fld_11 := Libadalang.Implementation.Basic_Decl_P_Is_Compilation_Unit_Root (Node => Ent.Node);
--# end
--# expr-done 2308
if Fld_11 then
   --# expr-start 2311 '<Not at ast.py:2411>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2411
--# expr-start 2309 '<Cast to Body at ast.py:2411>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2411
Cast_Expr_1 := Ent; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Body_Node
   then
      Cast_Result_2 := Create_Internal_Entity_Body_Node
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_2 := No_Entity_Body_Node;
   end if;
--# expr-done 2309
Var_Expr := Cast_Result_2;
if Var_Expr /= No_Entity_Body_Node then
   --# scope-start
--# expr-start 2310 '<FieldAccess for is_subunit at ast.py:2411>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2411
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.is_subunit
Fld_12 := Libadalang.Implementation.Body_Node_P_Is_Subunit (Node => Var_Expr.Node);
--# end
--# expr-done 2310
Scope_Result := Fld_12;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Not_Val_1 := not (Result_Var); 
--# expr-done 2311
   If_Result_2 := Not_Val_1;
else
   If_Result_2 := False;
end if;
if If_Result_2 then
   If_Result_3 := True;
else
   If_Result_3 := raise Property_Error with "Only package-level declaration support visibility checks for now.";
end if;
   If_Result_4 := If_Result_3;
end if;
   If_Result_5 := If_Result_4;
end if;
--# expr-done 2312
Let_Result := If_Result_5; 
--# end
--# expr-done 2313
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_163;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_163;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Is_Visible;
--# end
--# property-start BasicDecl.subp_decl_match_signature /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2421
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Subp_Decl_Match_Signature
  (Node : Bare_Basic_Decl
      ; Other : Internal_Entity_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind other Other
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Subp_Spec;
Fld_1 : Internal_Entity_Base_Subp_Spec;
Cast_Result : Internal_Entity_Subp_Spec;
Cast_Result_1 : Internal_Entity_Base_Subp_Spec;
Fld_2 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2318 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2317 '<FieldAccess for match_signature at ast.py:2423>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2423
--# expr-start 2314 '<FieldAccess for subp_spec_or_null at ast.py:2423>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2423
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2314
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2316 '<Cast to SubpSpec at ast.py:2424>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2424
--# expr-start 2315 '<FieldAccess for subp_spec_or_null at ast.py:2424>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2424
      if Other.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Other.Node, E_Info => Other.Info);
--# end
--# expr-done 2315
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Subp_Spec_Range
   then
      Cast_Result := Create_Internal_Entity_Subp_Spec
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 2316
      Cast_Result_1 := Create_Internal_Entity_Base_Subp_Spec
        (Node => Cast_Result.Node,
         Info => Cast_Result.Info);
--# property-call-start BaseSubpSpec.match_signature
Fld_2 := Libadalang.Implementation.Base_Subp_Spec_P_Match_Signature (Node => Fld.Node, Other => Cast_Result_1, Match_Name => False, E_Info => Fld.Info);
--# end
--# expr-done 2317
Let_Result := Fld_2; 
--# end
--# expr-done 2318
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Subp_Decl_Match_Signature;
--# end
--# property-start BasicDecl.base_subp_declarations /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2431
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Base_Subp_Declarations
  (Node : Bare_Basic_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl_Array_Access
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_165 with Inline_Always;
            procedure Finalizer_Scope_2043 with Inline_Always;
            procedure Finalizer_Scope_2044 with Inline_Always;
      Block_Var : Boolean;
Block_Var_1 : Internal_Entity_Basic_Decl;
Block_Var_2 : Boolean;
Block_Var_3 : Internal_Entity_Base_Subp_Spec;
Block_Var_4 : Lexical_Env := Empty_Env;
Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Entity_Base_Subp_Spec;
Var_Expr_2 : Internal_Entity_Basic_Decl;
Fld : Boolean;
Fld_1 : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity_Basic_Decl;
Is_A : Boolean;
Fld_3 : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Fld_4 : Internal_Entity_Base_Subp_Spec;
If_Result : Boolean;
Cast_Expr : Internal_Entity_Basic_Decl;
Cast_Result_1 : Internal_Entity_Base_Type_Decl;
Fld_5 : Internal_Entity_Base_Type_Decl_Array_Access;
Fld_6 : Internal_Entity_Base_Type_Decl;
Fld_7 : Lexical_Env := Empty_Env;
Map_Result : Lexical_Env_Array_Access;
Fld_8 : Lexical_Env := Empty_Env;
Singleton : Lexical_Env_Array_Access;
Concat_Result : Lexical_Env_Array_Access;
Group_Env : Lexical_Env := Empty_Env;
Fld_9 : Internal_Entity_Base_Type_Decl;
Fld_10 : Internal_Entity_Base_Type_Decl;
Fld_11 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
If_Result_1 : Lexical_Env := Empty_Env;
If_Result_2 : Lexical_Env := Empty_Env;
Fld_12 : Symbol_Type;
Env_Get_Result : Internal_Entity_Array_Access;
Cast_Expr_1 : Internal_Entity;
Cast_Result_2 : Internal_Entity_Basic_Decl;
Fld_13 : Internal_Entity_Basic_Decl;
Cast_Expr_2 : Internal_Entity;
Cast_Result_3 : Internal_Entity_Basic_Decl;
Fld_14 : Internal_Entity_Base_Subp_Spec;
Scope_Result_1 : Internal_Entity_Base_Subp_Spec;
Result_Var_1 : Internal_Entity_Base_Subp_Spec;
Fld_15 : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
Map_Result_1 : Internal_Entity_Basic_Decl_Array_Access;
Unique_Array : Internal_Entity_Basic_Decl_Array_Access;
Let_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_165 is
            begin
                     Dec_Ref (Fld_5);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Fld_8);
                     Dec_Ref (Singleton);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Group_Env);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (If_Result_2);
                     Dec_Ref (Block_Var_4);
                     Dec_Ref (Env_Get_Result);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Unique_Array);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_165;
            procedure Finalizer_Scope_2043 is
            begin
                     Dec_Ref (Fld_7);
            end Finalizer_Scope_2043;
            procedure Finalizer_Scope_2044 is
            begin
                     Dec_Ref (Fld_11);
            end Finalizer_Scope_2044;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Basic_Decl_Basic_Decl_P_Base_Subp_Declarations,
            Items    => new Mmz_Key_Array (1 ..  3))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Basic_Decl,
                                As_Bare_Basic_Decl => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Boolean,
                                      As_Boolean => Imprecise_Fallback);
               Mmz_K.Items (3) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Basic_Decl_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 2347 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2319 '<FieldAccess for is_subprogram at ast.py:2440>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2440
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_subprogram
Fld := Libadalang.Implementation.Basic_Decl_P_Is_Subprogram (Node => Ent.Node);
--# end
--# expr-done 2319
Block_Var := Fld; 
--# bind is_subp Block_Var
--# expr-start 2321 '<FieldAccess for parent_basic_decl at ast.py:2441>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2441
--# expr-start 2320 '<FieldAccess for canonical_part at ast.py:2441>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2441
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.canonical_part
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Canonical_Part (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2320
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.parent_basic_decl
Fld_2 := Libadalang.Implementation.Ada_Node_P_Parent_Basic_Decl (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 2321
Block_Var_1 := Fld_2; 
--# bind parent Block_Var_1
--# expr-start 2322 '<IsA ProtectedTypeDecl, TaskTypeDecl at ast.py:2442>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2442
Is_A := Block_Var_1.Node /= null 
and then Block_Var_1.Node.Kind in Ada_Protected_Type_Decl_Range | Ada_Task_Type_Decl_Range; 
--# expr-done 2322
Block_Var_2 := Is_A; 
--# bind task_or_protected Block_Var_2
--# expr-start 2325 '<FieldAccess for subp_spec_or_null at ast.py:2447>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2447
--# expr-start 2324 '<Cast to BasicDecl at ast.py:2447>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2447
--# expr-start 2323 '<FieldAccess for without_md at ast.py:2447>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2447
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.without_md
Fld_3 := Libadalang.Implementation.Ada_Node_P_Without_Md (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2323
   if Fld_3.Node = null
      or else Fld_3.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 2324
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Cast_Result.Node, E_Info => Cast_Result.Info);
--# end
--# expr-done 2325
Block_Var_3 := Fld_4; 
--# bind spec Block_Var_3
--# expr-start 2337 '<Cond at ast.py:2451>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2451
if Block_Var then
   If_Result := Block_Var_2;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 2333 '<env_group at ast.py:2456>' Group_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2456
--# expr-start 2332 '<Concat at ast.py:2456>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2456
--# expr-start 2330 '<Map at ast.py:2456>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2456
   --# expr-start 2327 '<FieldAccess for base_types at ast.py:2456>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2456
--# expr-start 2326 '<Cast to BaseTypeDecl at ast.py:2456>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2456
Cast_Expr := Block_Var_1; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Base_Type_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result_1 := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 2326
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.base_types'
Fld_5 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Base_Types (Node => Cast_Result_1.Node, Origin => No_Bare_Ada_Node, E_Info => Cast_Result_1.Info);
--# end
--# expr-done 2327
      declare
         Map_Result_Vec : Lexical_Env_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Base_Type_Decl_Array_Access := Fld_5;
      begin
         for Item_100 of
               Collection.Items
         loop
   --# scope-start
               --# bind bt Item_100
      --# expr-start 2329 '<FieldAccess for primitives_env at ast.py:2457>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2457
--# expr-start 2328 '<FieldAccess for full_view at ast.py:2457>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2457
      if Item_100.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.full_view
Fld_6 := Libadalang.Implementation.Base_Type_Decl_P_Full_View (Node => Item_100.Node, E_Info => Item_100.Info);
--# end
--# expr-done 2328
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.primitives_env'
Fld_7 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Primitives_Env (Node => Fld_6.Node, E_Info => Fld_6.Info);
--# end
--# expr-done 2329
         declare
            Item_To_Append : constant Lexical_Env := Fld_7;
         begin
               Inc_Ref (Item_To_Append);
            Lexical_Env_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2043;
         end loop;
      end;
         Map_Result := Create_Lexical_Env_Array
           (Items_Count => Natural (Lexical_Env_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Lexical_Env_Vectors.Get
              (Map_Result_Vec,
               I + Lexical_Env_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Lexical_Env_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 2330
--# expr-start 2331 '<FieldAccess for children_env at ast.py:2458>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2458
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Children_Env (Node => Block_Var_1.Node, E_Info => Block_Var_1.Info);
--# expr-done 2331
                Singleton := Create_Lexical_Env_Array (Items_Count => 1);
                Singleton.Items (1) := Fld_8;
                Inc_Ref (Singleton.Items (1));
Concat_Result := Concat (Map_Result, Singleton); 
--# expr-done 2332
Group_Env := Group (Concat_Result, No_Metadata); 
--# expr-done 2333
   If_Result_2 := Group_Env;
else
if Block_Var then
--# expr-start 2334 '<FieldAccess for candidate_primitive_subp_tagged_type at ast.py:2462>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2462
      if Block_Var_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.candidate_primitive_subp_tagged_type
Fld_9 := Libadalang.Implementation.Base_Subp_Spec_P_Candidate_Primitive_Subp_Tagged_Type (Node => Block_Var_3.Node, Canonicalize => False, E_Info => Block_Var_3.Info);
--# end
--# expr-done 2334
Var_Expr := Fld_9;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind t Var_Expr
--# expr-start 2336 '<FieldAccess for primitives_env at ast.py:2463>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2463
--# expr-start 2335 '<FieldAccess for full_view at ast.py:2463>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2463
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.full_view
Fld_10 := Libadalang.Implementation.Base_Type_Decl_P_Full_View (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 2335
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.primitives_env'
Fld_11 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Primitives_Env (Node => Fld_10.Node, E_Info => Fld_10.Info);
--# end
--# expr-done 2336
Scope_Result := Fld_11;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2044;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
   If_Result_1 := Result_Var;
else
   If_Result_1 := Empty_Env;
end if;
      Inc_Ref (If_Result_1);
   If_Result_2 := If_Result_1;
end if;
      Inc_Ref (If_Result_2);
--# expr-done 2337
Block_Var_4 := If_Result_2; Inc_Ref (Block_Var_4);
--# bind prims_env Block_Var_4
--# expr-start 2346 '<unique at ast.py:2481>' Unique_Array /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2481
--# expr-start 2345 '<FilterMap at ast.py:2481>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2481
   --# expr-start 2339 '<EnvGet at ast.py:2481>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2481
--# expr-start 2338 '<FieldAccess for name_symbol at ast.py:2481>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2481
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld_12 := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 2338
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Block_Var_4, Key => Fld_12, Lookup_Kind => To_Lookup_Kind_Type (Minimal), Categories => (Inherited_Primitives => True, Nocat => True))); 
--# expr-done 2339
      declare
         Map_Result_1_Vec : Internal_Entity_Basic_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Env_Get_Result;
      begin
         for Item_101 of
               Collection.Items
         loop
   --# scope-start
               --# bind bd Item_101
--# expr-start 2340 '<Cast to BasicDecl at ast.py:2483>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2483
Cast_Expr_2 := Item_101; 
   if Cast_Expr_2.Node = null
      or else Cast_Expr_2.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_3 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
   else
         Cast_Result_3 := No_Entity_Basic_Decl;
   end if;
--# expr-done 2340
Var_Expr_2 := Cast_Result_3;
if Var_Expr_2 /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 2341 '<FieldAccess for subp_spec_or_null at ast.py:2483>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2483
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_14 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Var_Expr_2.Node, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 2341
Scope_Result_1 := Fld_14;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Base_Subp_Spec;
end if;
Var_Expr_1 := Result_Var_1;
if Var_Expr_1 /= No_Entity_Base_Subp_Spec then
   --# scope-start
   --# bind s Var_Expr_1
--# expr-start 2342 '<FieldAccess for match_signature at ast.py:2490>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2490
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.match_signature
Fld_15 := Libadalang.Implementation.Base_Subp_Spec_P_Match_Signature (Node => Var_Expr_1.Node, Other => Block_Var_3, Match_Name => False, Use_Entity_Info => True, Ignore_First_Param => Block_Var_2, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 2342
Scope_Result_2 := Fld_15;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := False;
end if;
         if Result_Var_2 then
      --# expr-start 2344 '<FieldAccess for canonical_part at ast.py:2482>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2482
--# expr-start 2343 '<Cast to BasicDecl at ast.py:2482>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2482
Cast_Expr_1 := Item_101; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_2 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_2 := No_Entity_Basic_Decl;
   end if;
--# expr-done 2343
      if Cast_Result_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.canonical_part
Fld_13 := Libadalang.Implementation.Basic_Decl_P_Canonical_Part (Node => Cast_Result_2.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Cast_Result_2.Info);
--# end
--# expr-done 2344
         declare
            Item_To_Append : constant Internal_Entity_Basic_Decl := Fld_13;
         begin
            Internal_Entity_Basic_Decl_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result_1 := Create_Internal_Entity_Basic_Decl_Array
           (Items_Count => Natural (Internal_Entity_Basic_Decl_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Entity_Basic_Decl_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Entity_Basic_Decl_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Entity_Basic_Decl_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 2345
Unique_Array := Make_Unique (Map_Result_1); 
--# expr-done 2346
Let_Result := Unique_Array; Inc_Ref (Let_Result);
--# end
--# expr-done 2347
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_165;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Basic_Decl_Array_Access,
                        As_Internal_Entity_Basic_Decl_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_165;
               Finalizer_Scope_2043;
               Finalizer_Scope_2044;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Base_Subp_Declarations;
--# end
--# property-start BasicDecl.root_subp_declarations /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2502
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Root_Subp_Declarations
  (Node : Bare_Basic_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl_Array_Access
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_166 with Inline_Always;
      Block_Var : Internal_Entity_Basic_Decl_Array_Access;
Block_Var_1 : Internal_Entity_Basic_Decl;
Block_Var_2 : Boolean;
Block_Var_3 : Internal_Entity_Basic_Decl_Array_Access;
Block_Var_4 : Internal_Entity_Base_Type_Decl_Array_Access;
Fld : Internal_Entity_Basic_Decl_Array_Access;
Fld_1 : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity_Basic_Decl;
Is_A : Boolean;
Fld_3 : Bare_Basic_Decl;
Cast_Result : Bare_Basic_Decl;
Is_Equal : Boolean;
Not_Val : Boolean;
Map_Result : Internal_Entity_Basic_Decl_Array_Access;
Len : Integer;
Comp_Result : Boolean;
If_Result : Boolean;
If_Result_1 : Internal_Entity_Basic_Decl_Array_Access;
Fld_4 : Internal_Entity_Info;
Fld_5 : Internal_Metadata;
Fld_6 : Bare_Ada_Node;
Cast_Result_1 : Bare_Base_Type_Decl;
As_Entity : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
If_Result_2 : Internal_Entity_Base_Type_Decl;
Map_Result_1 : Internal_Entity_Base_Type_Decl_Array_Access;
Unique_Array : Internal_Entity_Base_Type_Decl_Array_Access;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity;
Is_Equiv : Boolean;
Not_Val_1 : Boolean;
Fld_7 : Boolean;
If_Result_3 : Boolean;
Quantifier_Result : Boolean;
Not_Val_2 : Boolean;
Map_Result_2 : Internal_Entity_Base_Type_Decl_Array_Access;
Fld_8 : Internal_Entity_Info;
Fld_9 : Internal_Metadata;
Fld_10 : Bare_Ada_Node;
Fld_11 : Bare_Base_Type_Decl;
Cast_Result_4 : Bare_Ada_Node;
Is_Equal_1 : Boolean;
Find_Result : Internal_Entity_Basic_Decl;
Map_Result_3 : Internal_Entity_Basic_Decl_Array_Access;
Let_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_166 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Map_Result);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Block_Var_3);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Unique_Array);
                     Dec_Ref (Block_Var_4);
                     Dec_Ref (Map_Result_2);
                     Dec_Ref (Map_Result_3);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_166;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2378 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2348 '<FieldAccess for base_subp_declarations at ast.py:2520>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2520
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.base_subp_declarations
Fld := Libadalang.Implementation.Basic_Decl_P_Base_Subp_Declarations (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2348
Block_Var := Fld; Inc_Ref (Block_Var);
--# bind raw_base_decls Block_Var
--# expr-start 2350 '<FieldAccess for parent_basic_decl at ast.py:2529>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2529
--# expr-start 2349 '<FieldAccess for canonical_part at ast.py:2529>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2529
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.canonical_part
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Canonical_Part (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2349
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.parent_basic_decl
Fld_2 := Libadalang.Implementation.Ada_Node_P_Parent_Basic_Decl (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 2350
Block_Var_1 := Fld_2; 
--# bind parent Block_Var_1
--# expr-start 2351 '<IsA ProtectedTypeDecl, TaskTypeDecl at ast.py:2530>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2530
Is_A := Block_Var_1.Node /= null 
and then Block_Var_1.Node.Kind in Ada_Protected_Type_Decl_Range | Ada_Task_Type_Decl_Range; 
--# expr-done 2351
Block_Var_2 := Is_A; 
--# bind task_or_protected Block_Var_2
--# expr-start 2357 '<If at ast.py:2531>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2531
if Block_Var_2 then
   --# expr-start 2352 '<length at ast.py:2532>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2532
Len := Length (Block_Var); 
--# expr-done 2352
Comp_Result := Len > 1; 
   If_Result := Comp_Result;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 2356 '<Filter at ast.py:2533>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2533
      declare
         Map_Result_Vec : Internal_Entity_Basic_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Basic_Decl_Array_Access := Block_Var;
      begin
         for Item_102 of
               Collection.Items
         loop
   --# scope-start
               --# bind bd Item_102
         --# expr-start 2355 '<Not at ast.py:2533>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2533
--# expr-start 2354 '<Eq at ast.py:2533>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2533
--# expr-start 2353 '<FieldAccess for node at ast.py:2533>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2533
Fld_3 := Item_102.Node;
--# expr-done 2353
      Cast_Result := Fld_3;
Is_Equal := Cast_Result = Self; 
--# expr-done 2354
Not_Val := not (Is_Equal); 
--# expr-done 2355
         if Not_Val then
         declare
            Item_To_Append : constant Internal_Entity_Basic_Decl := Item_102;
         begin
            Internal_Entity_Basic_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Basic_Decl_Array
           (Items_Count => Natural (Internal_Entity_Basic_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Basic_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Basic_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Basic_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 2356
   If_Result_1 := Map_Result;
else
   If_Result_1 := Block_Var;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 2357
Block_Var_3 := If_Result_1; Inc_Ref (Block_Var_3);
--# bind base_decls Block_Var_3
--# expr-start 2364 '<unique at ast.py:2538>' Unique_Array /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2538
--# expr-start 2363 '<Map at ast.py:2538>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2538
      declare
         Map_Result_1_Vec : Internal_Entity_Base_Type_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Basic_Decl_Array_Access := Block_Var_3;
      begin
         for Item_103 of
               Collection.Items
         loop
   --# scope-start
               --# bind d Item_103
      --# expr-start 2362 '<as_bare_entity at ast.py:2539>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2539
--# expr-start 2361 '<Cast to BaseTypeDecl at ast.py:2539>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2539
--# expr-start 2360 '<FieldAccess for primitive at ast.py:2539>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2539
--# expr-start 2359 '<FieldAccess for md at ast.py:2539>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2539
--# expr-start 2358 '<FieldAccess for info at ast.py:2539>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2539
Fld_4 := Item_103.Info;
--# expr-done 2358
Fld_5 := Fld_4.Md;
--# expr-done 2359
Fld_6 := Fld_5.Primitive;
--# expr-done 2360
   if Fld_6 = null
      or else Fld_6.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_1 := Fld_6;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 2361
Is_Null := Cast_Result_1 = null; 
if Is_Null then
   If_Result_2 := No_Entity_Base_Type_Decl;
else
As_Entity := (Info => No_Entity_Info, Node => Cast_Result_1); 
   If_Result_2 := As_Entity;
end if;
--# expr-done 2362
         declare
            Item_To_Append : constant Internal_Entity_Base_Type_Decl := If_Result_2;
         begin
            Internal_Entity_Base_Type_Decl_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result_1 := Create_Internal_Entity_Base_Type_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Type_Decl_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Entity_Base_Type_Decl_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Entity_Base_Type_Decl_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Entity_Base_Type_Decl_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 2363
Unique_Array := Make_Unique (Map_Result_1); 
--# expr-done 2364
Block_Var_4 := Unique_Array; Inc_Ref (Block_Var_4);
--# bind base_types Block_Var_4
--# expr-start 2377 '<Map at ast.py:2545>' Map_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2545
   --# expr-start 2370 '<Filter at ast.py:2545>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2545
      declare
         Map_Result_2_Vec : Internal_Entity_Base_Type_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Base_Type_Decl_Array_Access := Block_Var_4;
      begin
         for Item_105 of
               Collection.Items
         loop
   --# scope-start
               --# bind t Item_105
         --# expr-start 2369 '<Not at ast.py:2546>' Not_Val_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2546
--# expr-start 2368 '<AnyQuantifier at ast.py:2546>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2546
      Quantifier_Result := False;
      declare
         Collection : constant Internal_Entity_Base_Type_Decl_Array_Access := Block_Var_4;
      begin
         for Item_106 of
               Collection.Items
         loop
   --# scope-start
               --# bind u Item_106
--# expr-start 2366 '<Not at ast.py:2548>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2548
--# expr-start 2365 '<Eq at ast.py:2548>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2548
Cast_Expr := Item_105; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Cast_Expr_1 := Item_106; 
      Cast_Result_3 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Is_Equiv := Equivalent (Cast_Result_2, Cast_Result_3); 
--# expr-done 2365
Not_Val_1 := not (Is_Equiv); 
--# expr-done 2366
if Not_Val_1 then
   --# expr-start 2367 '<FieldAccess for is_derived_type at ast.py:2549>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2549
      if Item_105.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_derived_type
Fld_7 := Libadalang.Implementation.Base_Type_Decl_P_Is_Derived_Type (Node => Item_105.Node, Other_Type => Item_106, Origin => Origin, E_Info => Item_105.Info);
--# end
--# expr-done 2367
   If_Result_3 := Fld_7;
else
   If_Result_3 := False;
end if;
      Quantifier_Result := If_Result_3;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
--# expr-done 2368
Not_Val_2 := not (Quantifier_Result); 
--# expr-done 2369
         if Not_Val_2 then
         declare
            Item_To_Append : constant Internal_Entity_Base_Type_Decl := Item_105;
         begin
            Internal_Entity_Base_Type_Decl_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result_2 := Create_Internal_Entity_Base_Type_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Type_Decl_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Internal_Entity_Base_Type_Decl_Vectors.Get
              (Map_Result_2_Vec,
               I + Internal_Entity_Base_Type_Decl_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Internal_Entity_Base_Type_Decl_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 2370
      declare
         Map_Result_3_Vec : Internal_Entity_Basic_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Base_Type_Decl_Array_Access := Map_Result_2;
      begin
         for Item_104 of
               Collection.Items
         loop
   --# scope-start
               --# bind root_type Item_104
      --# expr-start 2376 '<Find at ast.py:2554>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2554
      Find_Result := No_Entity_Basic_Decl;
      declare
         Collection : constant Internal_Entity_Basic_Decl_Array_Access := Block_Var_3;
      begin
         for Item_107 of
               Collection.Items
         loop
   --# scope-start
               --# bind d Item_107
      --# expr-start 2375 '<Eq at ast.py:2555>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2555
--# expr-start 2373 '<FieldAccess for primitive at ast.py:2555>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2555
--# expr-start 2372 '<FieldAccess for md at ast.py:2555>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2555
--# expr-start 2371 '<FieldAccess for info at ast.py:2555>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2555
Fld_8 := Item_107.Info;
--# expr-done 2371
Fld_9 := Fld_8.Md;
--# expr-done 2372
Fld_10 := Fld_9.Primitive;
--# expr-done 2373
--# expr-start 2374 '<FieldAccess for node at ast.py:2555>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2555
Fld_11 := Item_104.Node;
--# expr-done 2374
      Cast_Result_4 := Fld_11;
Is_Equal_1 := Fld_10 = Cast_Result_4; 
--# expr-done 2375
      if Is_Equal_1 then
         Find_Result := Item_107;
      end if;
   --# end
               exit when Is_Equal_1;
         end loop;
      end;
--# expr-done 2376
         declare
            Item_To_Append : constant Internal_Entity_Basic_Decl := Find_Result;
         begin
            Internal_Entity_Basic_Decl_Vectors.Append (Map_Result_3_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result_3 := Create_Internal_Entity_Basic_Decl_Array
           (Items_Count => Natural (Internal_Entity_Basic_Decl_Vectors.Length (Map_Result_3_Vec)));
         for I in Map_Result_3.Items'Range loop
            Map_Result_3.Items (I) := Internal_Entity_Basic_Decl_Vectors.Get
              (Map_Result_3_Vec,
               I + Internal_Entity_Basic_Decl_Vectors.Index_Type'First - Map_Result_3.Items'First);
         end loop;
         Internal_Entity_Basic_Decl_Vectors.Destroy (Map_Result_3_Vec);
      end;
--# expr-done 2377
Let_Result := Map_Result_3; Inc_Ref (Let_Result);
--# end
--# expr-done 2378
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_166;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_166;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Root_Subp_Declarations;
--# end
--# property-start BasicDecl.find_all_overrides /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2561
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Find_All_Overrides
  (Node : Bare_Basic_Decl
      ; Units : Internal_Unit_Array_Access
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl_Array_Access
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind units Units
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_168 with Inline_Always;
            procedure Finalizer_Scope_2055 with Inline_Always;
            procedure Finalizer_Scope_2056 with Inline_Always;
      Block_Var : Internal_Entity_Base_Subp_Spec;
Block_Var_1 : Internal_Entity_Base_Type_Decl;
Block_Var_2 : Internal_Entity_Type_Decl_Array_Access;
Var_Expr : Internal_Entity_Base_Subp_Spec;
Var_Expr_1 : Internal_Entity_Base_Type_Decl;
Prims : Internal_Entity_Basic_Decl_Array_Access;
Base_P : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Base_Subp_Spec;
Fld_1 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Type_Decl_Array_Access;
Scope_Result_1 : Internal_Entity_Type_Decl_Array_Access;
Result_Var_1 : Internal_Entity_Type_Decl_Array_Access;
Fld_3 : Lexical_Env := Empty_Env;
Fld_4 : Symbol_Type;
Env_Get_Result : Internal_Entity_Array_Access;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Map_Result : Internal_Entity_Basic_Decl_Array_Access;
Fld_5 : Bare_Basic_Decl;
Cast_Result_1 : Bare_Basic_Decl;
Is_Equal : Boolean;
Find_Result : Internal_Entity_Basic_Decl;
Fld_6 : Internal_Entity_Info;
Fld_7 : Internal_Metadata;
Fld_8 : Bare_Ada_Node;
Fld_9 : Bare_Type_Decl;
Cast_Result_2 : Bare_Ada_Node;
Is_Equal_1 : Boolean;
Fld_10 : Internal_Entity_Base_Subp_Spec;
Fld_11 : Internal_Entity_Base_Subp_Spec;
Fld_12 : Boolean;
If_Result : Boolean;
Map_Result_1 : Internal_Entity_Basic_Decl_Array_Access;
Fld_13 : Internal_Entity_Basic_Decl;
Fld_14 : Internal_Entity;
Cast_Result_3 : Internal_Entity_Basic_Decl;
Map_Result_2 : Internal_Entity_Basic_Decl_Array_Access;
Let_Result : Internal_Entity_Basic_Decl_Array_Access;
Let_Result_1 : Internal_Entity_Basic_Decl_Array_Access;
Map_Result_3 : Internal_Entity_Basic_Decl_Array_Access;
Unique_Array : Internal_Entity_Basic_Decl_Array_Access;
Let_Result_2 : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_168 is
            begin
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Block_Var_2);
                     Dec_Ref (Map_Result_3);
                     Dec_Ref (Unique_Array);
                     Dec_Ref (Let_Result_2);
            end Finalizer_Scope_168;
            procedure Finalizer_Scope_2055 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_2055;
            procedure Finalizer_Scope_2056 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Env_Get_Result);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Prims);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Map_Result_2);
                     Dec_Ref (Let_Result);
                     Dec_Ref (Let_Result_1);
            end Finalizer_Scope_2056;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2407 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 2379 '<FieldAccess for subp_spec_or_null at ast.py:2567>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2567
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2379
Block_Var := Fld; 
--# bind spec Block_Var
Var_Expr := Block_Var;
if Var_Expr /= No_Entity_Base_Subp_Spec then
   --# scope-start
--# expr-start 2380 '<FieldAccess for candidate_primitive_subp_tagged_type at ast.py:2571>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2571
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.candidate_primitive_subp_tagged_type
Fld_1 := Libadalang.Implementation.Base_Subp_Spec_P_Candidate_Primitive_Subp_Tagged_Type (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 2380
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Block_Var_1 := Result_Var; 
--# bind prim_type Block_Var_1
Var_Expr_1 := Block_Var_1;
if Var_Expr_1 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 2381 '<FieldAccess for find_all_derived_types at ast.py:2573>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2573
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.find_all_derived_types
Fld_2 := Libadalang.Implementation.Base_Type_Decl_P_Find_All_Derived_Types (Node => Var_Expr_1.Node, Units => Units, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 2381
Scope_Result_1 := Fld_2;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2055;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Internal_Entity_Type_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var_1);
Block_Var_2 := Result_Var_1; Inc_Ref (Block_Var_2);
--# bind derivations Block_Var_2
--# expr-start 2406 '<unique at ast.py:2575>' Unique_Array /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2575
--# expr-start 2405 '<Mapcat at ast.py:2575>' Map_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2575
      declare
         Map_Result_3_Vec : Internal_Entity_Basic_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Type_Decl_Array_Access := Block_Var_2;
      begin
         for Item_108 of
               Collection.Items
         loop
   --# scope-start
               --# bind t Item_108
      --# expr-start 2404 '<Let at ast.py:2576>' Let_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2576
--# scope-start
--# expr-start 2386 '<Map at ast.py:2578>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2578
   --# expr-start 2384 '<EnvGet at ast.py:2578>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2578
--# expr-start 2382 '<FieldAccess for primitives_env at ast.py:2578>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2578
      if Item_108.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.primitives_env'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Primitives_Env (Node => Item_108.Node, E_Info => Item_108.Info);
--# end
--# expr-done 2382
--# expr-start 2383 '<FieldAccess for name_symbol at ast.py:2578>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2578
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 2383
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Fld_3, Key => Fld_4, Lookup_Kind => To_Lookup_Kind_Type (Recursive), Categories => (Inherited_Primitives => True, Nocat => True))); 
--# expr-done 2384
      declare
         Map_Result_Vec : Internal_Entity_Basic_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Env_Get_Result;
      begin
         for Item_109 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_109
      --# expr-start 2385 '<Cast to BasicDecl at ast.py:2579>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2579
Cast_Expr := Item_109; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 2385
         declare
            Item_To_Append : constant Internal_Entity_Basic_Decl := Cast_Result;
         begin
            Internal_Entity_Basic_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Basic_Decl_Array
           (Items_Count => Natural (Internal_Entity_Basic_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Basic_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Basic_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Basic_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 2386
Prims := Map_Result; Inc_Ref (Prims);
--# bind prims Prims
--# expr-start 2403 '<Let at ast.py:2580>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2580
--# scope-start
--# expr-start 2389 '<Find at ast.py:2583>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2583
      Find_Result := No_Entity_Basic_Decl;
      declare
         Collection : constant Internal_Entity_Basic_Decl_Array_Access := Prims;
      begin
         for Item_110 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_110
      --# expr-start 2388 '<Eq at ast.py:2583>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2583
--# expr-start 2387 '<FieldAccess for node at ast.py:2583>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2583
Fld_5 := Item_110.Node;
--# expr-done 2387
      Cast_Result_1 := Fld_5;
Is_Equal := Cast_Result_1 = Self; 
--# expr-done 2388
      if Is_Equal then
         Find_Result := Item_110;
      end if;
   --# end
               exit when Is_Equal;
         end loop;
      end;
--# expr-done 2389
Base_P := Find_Result; 
--# bind base_p Base_P
--# expr-start 2402 '<Map at ast.py:2585>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2585
   --# expr-start 2398 '<Filter at ast.py:2585>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2585
      declare
         Map_Result_1_Vec : Internal_Entity_Basic_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Basic_Decl_Array_Access := Prims;
      begin
         for Item_112 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_112
--# expr-start 2394 '<Eq at ast.py:2591>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2591
--# expr-start 2392 '<FieldAccess for primitive at ast.py:2591>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2591
--# expr-start 2391 '<FieldAccess for md at ast.py:2591>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2591
--# expr-start 2390 '<FieldAccess for info at ast.py:2591>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2591
Fld_6 := Item_112.Info;
--# expr-done 2390
Fld_7 := Fld_6.Md;
--# expr-done 2391
Fld_8 := Fld_7.Primitive;
--# expr-done 2392
--# expr-start 2393 '<FieldAccess for node at ast.py:2591>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2591
Fld_9 := Item_108.Node;
--# expr-done 2393
      Cast_Result_2 := Fld_9;
Is_Equal_1 := Fld_8 = Cast_Result_2; 
--# expr-done 2394
if Is_Equal_1 then
   --# expr-start 2397 '<FieldAccess for match_signature at ast.py:2594>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2594
--# expr-start 2395 '<FieldAccess for subp_spec_or_null at ast.py:2594>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2594
      if Base_P.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_10 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Base_P.Node, E_Info => Base_P.Info);
--# end
--# expr-done 2395
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2396 '<FieldAccess for subp_spec_or_null at ast.py:2595>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2595
      if Item_112.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_11 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Item_112.Node, E_Info => Item_112.Info);
--# end
--# expr-done 2396
--# property-call-start BaseSubpSpec.match_signature
Fld_12 := Libadalang.Implementation.Base_Subp_Spec_P_Match_Signature (Node => Fld_10.Node, Other => Fld_11, Match_Name => False, Use_Entity_Info => True, E_Info => Fld_10.Info);
--# end
--# expr-done 2397
   If_Result := Fld_12;
else
   If_Result := False;
end if;
         if If_Result then
         declare
            Item_To_Append : constant Internal_Entity_Basic_Decl := Item_112;
         begin
            Internal_Entity_Basic_Decl_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result_1 := Create_Internal_Entity_Basic_Decl_Array
           (Items_Count => Natural (Internal_Entity_Basic_Decl_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Entity_Basic_Decl_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Entity_Basic_Decl_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Entity_Basic_Decl_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 2398
      declare
         Map_Result_2_Vec : Internal_Entity_Basic_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Basic_Decl_Array_Access := Map_Result_1;
      begin
         for Item_111 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_111
      --# expr-start 2401 '<Cast to BasicDecl at ast.py:2601>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2601
--# expr-start 2400 '<FieldAccess for without_md at ast.py:2601>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2601
--# expr-start 2399 '<FieldAccess for canonical_part at ast.py:2601>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2601
      if Item_111.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.canonical_part
Fld_13 := Libadalang.Implementation.Basic_Decl_P_Canonical_Part (Node => Item_111.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Item_111.Info);
--# end
--# expr-done 2399
      if Fld_13.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.without_md
Fld_14 := Libadalang.Implementation.Ada_Node_P_Without_Md (Node => Fld_13.Node, E_Info => Fld_13.Info);
--# end
--# expr-done 2400
   if Fld_14.Node = null
      or else Fld_14.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_3 := Create_Internal_Entity_Basic_Decl
        (Node => Fld_14.Node,
         Info => Fld_14.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 2401
         declare
            Item_To_Append : constant Internal_Entity_Basic_Decl := Cast_Result_3;
         begin
            Internal_Entity_Basic_Decl_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result_2 := Create_Internal_Entity_Basic_Decl_Array
           (Items_Count => Natural (Internal_Entity_Basic_Decl_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Internal_Entity_Basic_Decl_Vectors.Get
              (Map_Result_2_Vec,
               I + Internal_Entity_Basic_Decl_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Internal_Entity_Basic_Decl_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 2402
Let_Result := Map_Result_2; Inc_Ref (Let_Result);
--# end
--# expr-done 2403
Let_Result_1 := Let_Result; Inc_Ref (Let_Result_1);
--# end
--# expr-done 2404
         for Item_To_Append of
               Let_Result_1.Items
         loop
            Internal_Entity_Basic_Decl_Vectors.Append (Map_Result_3_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_2056;
         end loop;
      end;
         Map_Result_3 := Create_Internal_Entity_Basic_Decl_Array
           (Items_Count => Natural (Internal_Entity_Basic_Decl_Vectors.Length (Map_Result_3_Vec)));
         for I in Map_Result_3.Items'Range loop
            Map_Result_3.Items (I) := Internal_Entity_Basic_Decl_Vectors.Get
              (Map_Result_3_Vec,
               I + Internal_Entity_Basic_Decl_Vectors.Index_Type'First - Map_Result_3.Items'First);
         end loop;
         Internal_Entity_Basic_Decl_Vectors.Destroy (Map_Result_3_Vec);
      end;
--# expr-done 2405
Unique_Array := Make_Unique (Map_Result_3); 
--# expr-done 2406
Let_Result_2 := Unique_Array; Inc_Ref (Let_Result_2);
--# end
--# expr-done 2407
      Property_Result := Let_Result_2;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_168;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_168;
               Finalizer_Scope_2055;
               Finalizer_Scope_2056;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Find_All_Overrides;
--# end
--# property-start '[dispatcher]BasicDecl.defining_names' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Decl_P_Defining_Names
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Decl (Self.Kind) is
               when Ada_Abstract_State_Decl =>
                  --# property-call-start AbstractStateDecl.defining_names
                  Property_Result := Abstract_State_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Anonymous_Expr_Decl =>
                  --# property-call-start AnonymousExprDecl.defining_names
                  Property_Result := Anonymous_Expr_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Component_Decl =>
                  --# property-call-start ComponentDecl.defining_names
                  Property_Result := Component_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Discriminant_Spec =>
                  --# property-call-start DiscriminantSpec.defining_names
                  Property_Result := Discriminant_Spec_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Formal_Obj_Decl .. Ada_Generic_Formal_Type_Decl =>
                  --# property-call-start GenericFormal.defining_names
                  Property_Result := Generic_Formal_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Param_Spec =>
                  --# property-call-start ParamSpec.defining_names
                  Property_Result := Param_Spec_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Synthetic_Formal_Param_Decl =>
                  --# property-call-start SyntheticFormalParamDecl.defining_names
                  Property_Result := Synthetic_Formal_Param_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Package_Internal .. Ada_Package_Decl =>
                  --# property-call-start BasePackageDecl.defining_names
                  Property_Result := Base_Package_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start BaseTypeDecl.defining_names
                  Property_Result := Base_Type_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Abstract_Subp_Decl | Ada_Subp_Decl | Ada_Generic_Subp_Internal .. Ada_Synthetic_Subp_Decl =>
                  --# property-call-start BasicSubpDecl.defining_names
                  Property_Result := Basic_Subp_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Abstract_Formal_Subp_Decl .. Ada_Concrete_Formal_Subp_Decl =>
                  --# property-call-start FormalSubpDecl.defining_names
                  Property_Result := Formal_Subp_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Entry_Decl =>
                  --# property-call-start EntryDecl.defining_names
                  Property_Result := Entry_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Enum_Literal_Decl =>
                  --# property-call-start EnumLiteralDecl.defining_names
                  Property_Result := Enum_Literal_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Synthetic_Char_Enum_Lit =>
                  --# property-call-start SyntheticCharEnumLit.defining_names
                  Property_Result := Synthetic_Char_Enum_Lit_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Expr_Function .. Ada_Subp_Renaming_Decl =>
                  --# property-call-start BaseSubpBody.defining_names
                  Property_Result := Base_Subp_Body_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Package_Body_Stub =>
                  --# property-call-start PackageBodyStub.defining_names
                  Property_Result := Package_Body_Stub_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Protected_Body_Stub =>
                  --# property-call-start ProtectedBodyStub.defining_names
                  Property_Result := Protected_Body_Stub_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Subp_Body_Stub =>
                  --# property-call-start SubpBodyStub.defining_names
                  Property_Result := Subp_Body_Stub_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Task_Body_Stub =>
                  --# property-call-start TaskBodyStub.defining_names
                  Property_Result := Task_Body_Stub_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Entry_Body =>
                  --# property-call-start EntryBody.defining_names
                  Property_Result := Entry_Body_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Package_Body =>
                  --# property-call-start PackageBody.defining_names
                  Property_Result := Package_Body_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Protected_Body =>
                  --# property-call-start ProtectedBody.defining_names
                  Property_Result := Protected_Body_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Task_Body =>
                  --# property-call-start TaskBody.defining_names
                  Property_Result := Task_Body_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Entry_Index_Spec =>
                  --# property-call-start EntryIndexSpec.defining_names
                  Property_Result := Entry_Index_Spec_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Error_Decl =>
                  --# property-call-start ErrorDecl.defining_names
                  Property_Result := Error_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Exception_Decl =>
                  --# property-call-start ExceptionDecl.defining_names
                  Property_Result := Exception_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Exception_Handler =>
                  --# property-call-start ExceptionHandler.defining_names
                  Property_Result := Exception_Handler_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_For_Loop_Var_Decl =>
                  --# property-call-start ForLoopVarDecl.defining_names
                  Property_Result := For_Loop_Var_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Package_Decl =>
                  --# property-call-start GenericPackageDecl.defining_names
                  Property_Result := Generic_Package_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Subp_Decl =>
                  --# property-call-start GenericSubpDecl.defining_names
                  Property_Result := Generic_Subp_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Package_Instantiation =>
                  --# property-call-start GenericPackageInstantiation.defining_names
                  Property_Result := Generic_Package_Instantiation_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Subp_Instantiation =>
                  --# property-call-start GenericSubpInstantiation.defining_names
                  Property_Result := Generic_Subp_Instantiation_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Package_Renaming_Decl =>
                  --# property-call-start GenericPackageRenamingDecl.defining_names
                  Property_Result := Generic_Package_Renaming_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Subp_Renaming_Decl =>
                  --# property-call-start GenericSubpRenamingDecl.defining_names
                  Property_Result := Generic_Subp_Renaming_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Label_Decl =>
                  --# property-call-start LabelDecl.defining_names
                  Property_Result := Label_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Named_Stmt_Decl =>
                  --# property-call-start NamedStmtDecl.defining_names
                  Property_Result := Named_Stmt_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Number_Decl =>
                  --# property-call-start NumberDecl.defining_names
                  Property_Result := Number_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Object_Decl .. Ada_No_Type_Object_Renaming_Decl =>
                  --# property-call-start ObjectDecl.defining_names
                  Property_Result := Object_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Package_Renaming_Decl =>
                  --# property-call-start PackageRenamingDecl.defining_names
                  Property_Result := Package_Renaming_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Single_Protected_Decl =>
                  --# property-call-start SingleProtectedDecl.defining_names
                  Property_Result := Single_Protected_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Single_Task_Decl =>
                  --# property-call-start SingleTaskDecl.defining_names
                  Property_Result := Single_Task_Decl_P_Defining_Names
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Decl_P_Defining_Names;
--# end
--# property-start BasicDecl.defining_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2616
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Defining_Name
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_170 with Inline_Always;
      Fld : Internal_Entity_Defining_Name_Array_Access;
Get_Result : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_170 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_170;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2409 '<at at ast.py:2617>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2617
--# expr-start 2408 '<FieldAccess for defining_names at ast.py:2617>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2617
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_names'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Names (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2408
Get_Result := Get (Fld, 0, True); 
--# expr-done 2409
      Property_Result := Get_Result;
   --# end
      Finalizer_Scope_170;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_170;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Defining_Name;
--# end
--# property-start '[dispatcher]BasicDecl.defining_env' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Decl_P_Defining_Env
  (Node : Bare_Basic_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Decl (Self.Kind) is
               when Ada_Abstract_State_Decl | Ada_Generic_Formal_Obj_Decl .. Ada_Generic_Formal_Type_Decl | Ada_Synthetic_Formal_Param_Decl | Ada_Package_Body_Stub .. Ada_Entry_Body | Ada_Protected_Body .. Ada_Task_Body | Ada_Error_Decl .. Ada_Exception_Handler | Ada_Generic_Subp_Decl | Ada_Generic_Subp_Instantiation | Ada_Generic_Subp_Renaming_Decl .. Ada_Label_Decl | Ada_Number_Decl =>
                  --# property-call-start BasicDecl.defining_env
                  Property_Result := Basic_Decl_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Anonymous_Expr_Decl =>
                  --# property-call-start AnonymousExprDecl.defining_env
                  Property_Result := Anonymous_Expr_Decl_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Component_Decl =>
                  --# property-call-start ComponentDecl.defining_env
                  Property_Result := Component_Decl_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Discriminant_Spec =>
                  --# property-call-start DiscriminantSpec.defining_env
                  Property_Result := Discriminant_Spec_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Param_Spec =>
                  --# property-call-start ParamSpec.defining_env
                  Property_Result := Param_Spec_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Package_Internal .. Ada_Package_Decl =>
                  --# property-call-start BasePackageDecl.defining_env
                  Property_Result := Base_Package_Decl_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.defining_env
                  Property_Result := Base_Subtype_Decl_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Classwide_Type_Decl =>
                  --# property-call-start ClasswideTypeDecl.defining_env
                  Property_Result := Classwide_Type_Decl_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Incomplete_Type_Decl .. Ada_Incomplete_Tagged_Type_Decl =>
                  --# property-call-start IncompleteTypeDecl.defining_env
                  Property_Result := Incomplete_Type_Decl_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Protected_Type_Decl =>
                  --# property-call-start ProtectedTypeDecl.defining_env
                  Property_Result := Protected_Type_Decl_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Task_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start TaskTypeDecl.defining_env
                  Property_Result := Task_Type_Decl_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.defining_env
                  Property_Result := Type_Decl_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Abstract_Subp_Decl .. Ada_Synthetic_Subp_Decl =>
                  --# property-call-start BasicSubpDecl.defining_env
                  Property_Result := Basic_Subp_Decl_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Expr_Function .. Ada_Subp_Renaming_Decl =>
                  --# property-call-start BaseSubpBody.defining_env
                  Property_Result := Base_Subp_Body_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Package_Body =>
                  --# property-call-start PackageBody.defining_env
                  Property_Result := Package_Body_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Entry_Index_Spec =>
                  --# property-call-start EntryIndexSpec.defining_env
                  Property_Result := Entry_Index_Spec_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_For_Loop_Var_Decl =>
                  --# property-call-start ForLoopVarDecl.defining_env
                  Property_Result := For_Loop_Var_Decl_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Package_Decl =>
                  --# property-call-start GenericPackageDecl.defining_env
                  Property_Result := Generic_Package_Decl_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Package_Instantiation =>
                  --# property-call-start GenericPackageInstantiation.defining_env
                  Property_Result := Generic_Package_Instantiation_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Package_Renaming_Decl =>
                  --# property-call-start GenericPackageRenamingDecl.defining_env
                  Property_Result := Generic_Package_Renaming_Decl_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Named_Stmt_Decl =>
                  --# property-call-start NamedStmtDecl.defining_env
                  Property_Result := Named_Stmt_Decl_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Object_Decl .. Ada_No_Type_Object_Renaming_Decl =>
                  --# property-call-start ObjectDecl.defining_env
                  Property_Result := Object_Decl_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Package_Renaming_Decl =>
                  --# property-call-start PackageRenamingDecl.defining_env
                  Property_Result := Package_Renaming_Decl_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Single_Protected_Decl =>
                  --# property-call-start SingleProtectedDecl.defining_env
                  Property_Result := Single_Protected_Decl_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
               when Ada_Single_Task_Decl =>
                  --# property-call-start SingleTaskDecl.defining_env
                  Property_Result := Single_Task_Decl_P_Defining_Env
                    (Self
                        , Origin
                        , Include_Ud_Indexing
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Decl_P_Defining_Env;
--# end
--# property-start BasicDecl.defining_name_or_raise /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2634
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Defining_Name_Or_Raise
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_172 with Inline_Always;
      Block_Var : Internal_Entity_Defining_Name_Array_Access;
Fld : Internal_Entity_Defining_Name_Array_Access;
Get_Result : Internal_Entity_Defining_Name;
Len : Integer;
Comp_Result : Boolean;
If_Result : Internal_Entity_Defining_Name;
Let_Result : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_172 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Block_Var);
            end Finalizer_Scope_172;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2414 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2410 '<FieldAccess for defining_names at ast.py:2639>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2639
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_names'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Names (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2410
Block_Var := Fld; Inc_Ref (Block_Var);
--# bind dns Block_Var
--# expr-start 2413 '<If at ast.py:2640>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2640
--# expr-start 2411 '<length at ast.py:2641>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2641
Len := Length (Block_Var); 
--# expr-done 2411
Comp_Result := Len > 1; 
if Comp_Result then
   If_Result := raise Property_Error with "BasicDecl with multiple defining names";
else
   --# expr-start 2412 '<at at ast.py:2644>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2644
Get_Result := Get (Block_Var, 0, True); 
--# expr-done 2412
   If_Result := Get_Result;
end if;
--# expr-done 2413
Let_Result := If_Result; 
--# end
--# expr-done 2414
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_172;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_172;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Defining_Name_Or_Raise;
--# end
--# property-start BasicDecl.identity_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2648
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Identity_Type
  (Node : Bare_Basic_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Match_0 : Internal_Entity_Exception_Decl;
Match_1 : Internal_Entity_Single_Task_Decl;
Match_2 : Internal_Entity_Object_Decl;
Match_3 : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Base_Type_Decl;
Match_Prefix : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity_Exception_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity_Single_Task_Decl;
Let_Result_1 : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity_Object_Decl;
Let_Result_2 : Internal_Entity_Base_Type_Decl;
Scope_Result_2 : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity_Basic_Decl;
Let_Result_3 : Internal_Entity_Base_Type_Decl;
Scope_Result_3 : Internal_Entity_Base_Type_Decl;
Match_Result : Internal_Entity_Base_Type_Decl;
Let_Result_4 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2419 '<Block at ???>' Let_Result_4 None
--# scope-start
--# expr-start 2418 '<Match at ast.py:2649>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2649
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Basic_Decl (Match_Prefix.Node.Kind) is
         when Ada_Exception_Decl =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Exception_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind _ Match_0
--# expr-start 2415 '<FieldAccess for exc_id_type at ast.py:2650>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2650
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.exc_id_type
Fld := Libadalang.Implementation.Ada_Node_P_Exc_Id_Type (Node => Self);
--# end
--# expr-done 2415
Let_Result := Fld; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Single_Task_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind _ Match_1
--# expr-start 2416 '<FieldAccess for task_id_type at ast.py:2651>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2651
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.task_id_type
Fld_1 := Libadalang.Implementation.Ada_Node_P_Task_Id_Type (Node => Self);
--# end
--# expr-done 2416
Let_Result_1 := Fld_1; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Object_Decl .. Ada_No_Type_Object_Renaming_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Object_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind _ Match_2
--# expr-start 2417 '<FieldAccess for task_id_type at ast.py:2655>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2655
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.task_id_type
Fld_2 := Libadalang.Implementation.Ada_Node_P_Task_Id_Type (Node => Self);
--# end
--# expr-done 2417
Let_Result_2 := Fld_2; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Abstract_State_Decl .. Ada_Error_Decl | Ada_Exception_Handler .. Ada_Number_Decl | Ada_Package_Renaming_Decl .. Ada_Single_Protected_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Basic_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_3; 
--# bind _ Match_3
Let_Result_3 := No_Entity_Base_Type_Decl; 
--# end
Scope_Result_3 := Let_Result_3;
   --# end
            Match_Result := Scope_Result_3; 
end case;
--# expr-done 2418
Let_Result_4 := Match_Result; 
--# end
--# expr-done 2419
      Property_Result := Let_Result_4;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Identity_Type;
--# end
--# property-start '[dispatcher]BasicDecl.array_ndims' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Decl_P_Array_Ndims
  (Node : Bare_Basic_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Integer
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Integer;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Decl (Self.Kind) is
               when Ada_Abstract_State_Decl .. Ada_Anonymous_Expr_Decl | Ada_Discriminant_Spec .. Ada_Package_Decl | Ada_Abstract_Subp_Decl .. Ada_Single_Task_Decl =>
                  --# property-call-start BasicDecl.array_ndims
                  Property_Result := Basic_Decl_P_Array_Ndims
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Component_Decl =>
                  --# property-call-start ComponentDecl.array_ndims
                  Property_Result := Component_Decl_P_Array_Ndims
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.array_ndims
                  Property_Result := Base_Type_Decl_P_Array_Ndims
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.array_ndims
                  Property_Result := Base_Subtype_Decl_P_Array_Ndims
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.array_ndims
                  Property_Result := Type_Decl_P_Array_Ndims
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Decl_P_Array_Ndims;
--# end
--# property-start BasicDecl.is_array /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2663
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Is_Array
  (Node : Bare_Basic_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Integer;
Comp_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2420 '<FieldAccess for array_ndims at ast.py:2663>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2663
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.array_ndims'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Array_Ndims (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 2420
Comp_Result := Fld > 0; 
      Property_Result := Comp_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Is_Array;
--# end
--# property-start '[dispatcher]BasicDecl.expr_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Decl_P_Expr_Type
  (Node : Bare_Basic_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Decl (Self.Kind) is
               when Ada_Abstract_State_Decl .. Ada_Package_Decl | Ada_Package_Body_Stub .. Ada_Task_Body | Ada_Error_Decl .. Ada_Exception_Decl | Ada_Generic_Package_Decl .. Ada_Named_Stmt_Decl | Ada_Object_Decl .. Ada_Single_Protected_Decl =>
                  --# property-call-start BasicDecl.expr_type
                  Property_Result := Basic_Decl_P_Expr_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start BaseTypeDecl.expr_type
                  Property_Result := Base_Type_Decl_P_Expr_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Abstract_Subp_Decl .. Ada_Synthetic_Subp_Decl =>
                  --# property-call-start BasicSubpDecl.expr_type
                  Property_Result := Basic_Subp_Decl_P_Expr_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Expr_Function .. Ada_Subp_Renaming_Decl =>
                  --# property-call-start BaseSubpBody.expr_type
                  Property_Result := Base_Subp_Body_P_Expr_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Entry_Index_Spec =>
                  --# property-call-start EntryIndexSpec.expr_type
                  Property_Result := Entry_Index_Spec_P_Expr_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Exception_Handler =>
                  --# property-call-start ExceptionHandler.expr_type
                  Property_Result := Exception_Handler_P_Expr_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_For_Loop_Var_Decl =>
                  --# property-call-start ForLoopVarDecl.expr_type
                  Property_Result := For_Loop_Var_Decl_P_Expr_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Number_Decl =>
                  --# property-call-start NumberDecl.expr_type
                  Property_Result := Number_Decl_P_Expr_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Single_Task_Decl =>
                  --# property-call-start SingleTaskDecl.expr_type
                  Property_Result := Single_Task_Decl_P_Expr_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Decl_P_Expr_Type;
--# end
--# property-start '[dispatcher]BasicDecl.type_expression' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Decl_P_Type_Expression
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Expr
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Type_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Decl (Self.Kind) is
               when Ada_Generic_Package_Internal .. Ada_Synth_Anonymous_Type_Decl | Ada_Package_Body_Stub .. Ada_Protected_Body_Stub | Ada_Task_Body_Stub .. Ada_Number_Decl | Ada_Package_Renaming_Decl .. Ada_Single_Task_Decl =>
                  --# property-call-start BasicDecl.type_expression
                  Property_Result := Basic_Decl_P_Type_Expression
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Abstract_State_Decl =>
                  --# property-call-start AbstractStateDecl.type_expression
                  Property_Result := Abstract_State_Decl_P_Type_Expression
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Anonymous_Expr_Decl =>
                  --# property-call-start AnonymousExprDecl.type_expression
                  Property_Result := Anonymous_Expr_Decl_P_Type_Expression
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Component_Decl =>
                  --# property-call-start ComponentDecl.type_expression
                  Property_Result := Component_Decl_P_Type_Expression
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Discriminant_Spec =>
                  --# property-call-start DiscriminantSpec.type_expression
                  Property_Result := Discriminant_Spec_P_Type_Expression
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Formal_Obj_Decl .. Ada_Generic_Formal_Type_Decl =>
                  --# property-call-start GenericFormal.type_expression
                  Property_Result := Generic_Formal_P_Type_Expression
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Param_Spec =>
                  --# property-call-start ParamSpec.type_expression
                  Property_Result := Param_Spec_P_Type_Expression
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Synthetic_Formal_Param_Decl =>
                  --# property-call-start SyntheticFormalParamDecl.type_expression
                  Property_Result := Synthetic_Formal_Param_Decl_P_Type_Expression
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Abstract_Subp_Decl .. Ada_Synthetic_Subp_Decl =>
                  --# property-call-start BasicSubpDecl.type_expression
                  Property_Result := Basic_Subp_Decl_P_Type_Expression
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Expr_Function .. Ada_Subp_Renaming_Decl =>
                  --# property-call-start BaseSubpBody.type_expression
                  Property_Result := Base_Subp_Body_P_Type_Expression
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Subp_Body_Stub =>
                  --# property-call-start SubpBodyStub.type_expression
                  Property_Result := Subp_Body_Stub_P_Type_Expression
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Object_Decl .. Ada_No_Type_Object_Renaming_Decl =>
                  --# property-call-start ObjectDecl.type_expression
                  Property_Result := Object_Decl_P_Type_Expression
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Decl_P_Type_Expression;
--# end
--# property-start BasicDecl.subp_spec_or_null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2700
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Subp_Spec_Or_Null
  (Node : Bare_Basic_Decl
      ; Follow_Generic : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Subp_Spec
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind follow_generic Follow_Generic
   Property_Result : Internal_Entity_Base_Subp_Spec;
      Match_0 : Internal_Entity_Basic_Subp_Decl;
Match_1 : Internal_Entity_Base_Subp_Body;
Match_2 : Internal_Entity_Subp_Body_Stub;
Match_3 : Internal_Entity_Generic_Subp_Decl;
Match_4 : Internal_Entity_Generic_Subp_Instantiation;
Match_5 : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Base_Subp_Spec;
Fld_1 : Internal_Entity_Subp_Spec;
Fld_2 : Internal_Entity_Subp_Spec;
Fld_3 : Internal_Entity_Generic_Subp_Internal;
Fld_4 : Internal_Entity_Subp_Spec;
If_Result : Internal_Entity_Subp_Spec;
Fld_5 : Internal_Entity_Basic_Decl;
Fld_6 : Internal_Entity_Base_Subp_Spec;
Cast_Expr : Internal_Entity_Subp_Spec;
Cast_Result : Internal_Entity_Base_Subp_Spec;
If_Result_1 : Internal_Entity_Base_Subp_Spec;
Match_Prefix : Internal_Entity_Basic_Decl;
Cast_Result_1 : Internal_Entity_Basic_Subp_Decl;
Let_Result : Internal_Entity_Base_Subp_Spec;
Scope_Result : Internal_Entity_Base_Subp_Spec;
Cast_Result_2 : Internal_Entity_Base_Subp_Body;
Cast_Result_3 : Internal_Entity_Base_Subp_Spec;
Let_Result_1 : Internal_Entity_Base_Subp_Spec;
Scope_Result_1 : Internal_Entity_Base_Subp_Spec;
Cast_Result_4 : Internal_Entity_Subp_Body_Stub;
Cast_Result_5 : Internal_Entity_Base_Subp_Spec;
Let_Result_2 : Internal_Entity_Base_Subp_Spec;
Scope_Result_2 : Internal_Entity_Base_Subp_Spec;
Cast_Result_6 : Internal_Entity_Generic_Subp_Decl;
Cast_Result_7 : Internal_Entity_Base_Subp_Spec;
Let_Result_3 : Internal_Entity_Base_Subp_Spec;
Scope_Result_3 : Internal_Entity_Base_Subp_Spec;
Cast_Result_8 : Internal_Entity_Generic_Subp_Instantiation;
Let_Result_4 : Internal_Entity_Base_Subp_Spec;
Scope_Result_4 : Internal_Entity_Base_Subp_Spec;
Cast_Result_9 : Internal_Entity_Basic_Decl;
Cast_Expr_1 : Internal_Entity_Subp_Spec;
Cast_Result_10 : Internal_Entity_Base_Subp_Spec;
Let_Result_5 : Internal_Entity_Base_Subp_Spec;
Scope_Result_5 : Internal_Entity_Base_Subp_Spec;
Match_Result : Internal_Entity_Base_Subp_Spec;
Let_Result_6 : Internal_Entity_Base_Subp_Spec;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2431 '<Block at ???>' Let_Result_6 None
--# scope-start
--# expr-start 2430 '<Match at ast.py:2707>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2707
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Basic_Decl (Match_Prefix.Node.Kind) is
         when Ada_Abstract_Subp_Decl .. Ada_Synthetic_Subp_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Basic_Subp_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_1; 
--# bind subp Match_0
--# expr-start 2421 '<FieldAccess for subp_decl_spec at ast.py:2708>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2708
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicSubpDecl.subp_decl_spec'
Fld := Libadalang.Implementation.Dispatcher_Basic_Subp_Decl_P_Subp_Decl_Spec (Node => Match_0.Node, E_Info => Match_0.Info);
--# end
--# expr-done 2421
Let_Result := Fld; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Expr_Function .. Ada_Subp_Renaming_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Base_Subp_Body
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_2; 
--# bind subp Match_1
--# expr-start 2422 '<FieldAccess for subp_spec at ast.py:2709>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2709
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Subp_Spec (Node => Match_1.Node.Base_Subp_Body_F_Subp_Spec, Info => Match_1.Info);
--# expr-done 2422
      Cast_Result_3 := Create_Internal_Entity_Base_Subp_Spec
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
Let_Result_1 := Cast_Result_3; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Subp_Body_Stub =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Subp_Body_Stub
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_4; 
--# bind subp Match_2
--# expr-start 2423 '<FieldAccess for subp_spec at ast.py:2710>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2710
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Subp_Spec (Node => Match_2.Node.Subp_Body_Stub_F_Subp_Spec, Info => Match_2.Info);
--# expr-done 2423
      Cast_Result_5 := Create_Internal_Entity_Base_Subp_Spec
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Let_Result_2 := Cast_Result_5; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Generic_Subp_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_6 := Create_Internal_Entity_Generic_Subp_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_6; 
--# bind gsp Match_3
--# expr-start 2426 '<If at ast.py:2712>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2712
if Follow_Generic then
   --# expr-start 2425 '<FieldAccess for subp_spec at ast.py:2712>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2712
--# expr-start 2424 '<FieldAccess for subp_decl at ast.py:2712>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2712
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Generic_Subp_Internal (Node => Match_3.Node.Generic_Subp_Decl_F_Subp_Decl, Info => Match_3.Info);
--# expr-done 2424
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Subp_Spec (Node => Fld_3.Node.Generic_Subp_Internal_F_Subp_Spec, Info => Fld_3.Info);
--# expr-done 2425
   If_Result := Fld_4;
else
   If_Result := No_Entity_Subp_Spec;
end if;
--# expr-done 2426
      Cast_Result_7 := Create_Internal_Entity_Base_Subp_Spec
        (Node => If_Result.Node,
         Info => If_Result.Info);
Let_Result_3 := Cast_Result_7; 
--# end
Scope_Result_3 := Let_Result_3;
   --# end
            Match_Result := Scope_Result_3; 
         when Ada_Generic_Subp_Instantiation =>
   --# scope-start
--# scope-start
      Cast_Result_8 := Create_Internal_Entity_Generic_Subp_Instantiation
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_4 := Cast_Result_8; 
--# bind gsi Match_4
--# expr-start 2429 '<If at ast.py:2714>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2714
if Follow_Generic then
   --# expr-start 2428 '<FieldAccess for subp_spec_or_null at ast.py:2715>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2715
--# expr-start 2427 '<FieldAccess for designated_generic_decl at ast.py:2715>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2715
      if Match_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericInstantiation.designated_generic_decl'
Fld_5 := Libadalang.Implementation.Dispatcher_Generic_Instantiation_P_Designated_Generic_Decl (Node => Match_4.Node, E_Info => Match_4.Info);
--# end
--# expr-done 2427
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_6 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Fld_5.Node, Follow_Generic => True, E_Info => Fld_5.Info);
--# end
--# expr-done 2428
   If_Result_1 := Fld_6;
else
Cast_Expr := No_Entity_Subp_Spec; 
      Cast_Result := Create_Internal_Entity_Base_Subp_Spec
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   If_Result_1 := Cast_Result;
end if;
--# expr-done 2429
Let_Result_4 := If_Result_1; 
--# end
Scope_Result_4 := Let_Result_4;
   --# end
            Match_Result := Scope_Result_4; 
         when Ada_Abstract_State_Decl .. Ada_Synth_Anonymous_Type_Decl | Ada_Package_Body_Stub .. Ada_Protected_Body_Stub | Ada_Task_Body_Stub .. Ada_Generic_Package_Decl | Ada_Generic_Package_Instantiation | Ada_Generic_Package_Renaming_Decl .. Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_9 := Create_Internal_Entity_Basic_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_5 := Cast_Result_9; 
--# bind _ Match_5
Cast_Expr_1 := No_Entity_Subp_Spec; 
      Cast_Result_10 := Create_Internal_Entity_Base_Subp_Spec
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Let_Result_5 := Cast_Result_10; 
--# end
Scope_Result_5 := Let_Result_5;
   --# end
            Match_Result := Scope_Result_5; 
end case;
--# expr-done 2430
Let_Result_6 := Match_Result; 
--# end
--# expr-done 2431
      Property_Result := Let_Result_6;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Subp_Spec_Or_Null;
--# end
--# property-start BasicDecl.formal_param_holder_or_null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2721
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Formal_Param_Holder_Or_Null
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Holder
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Formal_Param_Holder;
      Match_0 : Internal_Entity_Type_Decl;
Match_1 : Internal_Entity_Entry_Body;
Match_2 : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Discriminant_Part;
Fld_1 : Internal_Entity_Entry_Completion_Formal_Params;
Fld_2 : Internal_Entity_Base_Subp_Spec;
Match_Prefix : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity_Type_Decl;
Cast_Result_1 : Internal_Entity_Base_Formal_Param_Holder;
Let_Result : Internal_Entity_Base_Formal_Param_Holder;
Scope_Result : Internal_Entity_Base_Formal_Param_Holder;
Cast_Result_2 : Internal_Entity_Entry_Body;
Cast_Result_3 : Internal_Entity_Base_Formal_Param_Holder;
Let_Result_1 : Internal_Entity_Base_Formal_Param_Holder;
Scope_Result_1 : Internal_Entity_Base_Formal_Param_Holder;
Cast_Result_4 : Internal_Entity_Basic_Decl;
Cast_Result_5 : Internal_Entity_Base_Formal_Param_Holder;
Let_Result_2 : Internal_Entity_Base_Formal_Param_Holder;
Scope_Result_2 : Internal_Entity_Base_Formal_Param_Holder;
Match_Result : Internal_Entity_Base_Formal_Param_Holder;
Let_Result_3 : Internal_Entity_Base_Formal_Param_Holder;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2436 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 2435 '<Match at ast.py:2722>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2722
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Basic_Decl (Match_Prefix.Node.Kind) is
         when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind t Match_0
--# expr-start 2432 '<FieldAccess for discriminants at ast.py:2723>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2723
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Discriminant_Part (Node => Match_0.Node.Type_Decl_F_Discriminants, Info => Match_0.Info);
--# expr-done 2432
      Cast_Result_1 := Create_Internal_Entity_Base_Formal_Param_Holder
        (Node => Fld.Node,
         Info => Fld.Info);
Let_Result := Cast_Result_1; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Entry_Body =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Entry_Body
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_2; 
--# bind e Match_1
--# expr-start 2433 '<FieldAccess for params at ast.py:2724>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2724
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Entry_Completion_Formal_Params (Node => Match_1.Node.Entry_Body_F_Params, Info => Match_1.Info);
--# expr-done 2433
      Cast_Result_3 := Create_Internal_Entity_Base_Formal_Param_Holder
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
Let_Result_1 := Cast_Result_3; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abstract_State_Decl .. Ada_Single_Task_Type_Decl | Ada_Abstract_Subp_Decl .. Ada_Task_Body_Stub | Ada_Package_Body .. Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Basic_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_4; 
--# bind _ Match_2
--# expr-start 2434 '<FieldAccess for subp_spec_or_null at ast.py:2725>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2725
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Ent.Node, Follow_Generic => True, E_Info => Ent.Info);
--# end
--# expr-done 2434
      Cast_Result_5 := Create_Internal_Entity_Base_Formal_Param_Holder
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Let_Result_2 := Cast_Result_5; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 2435
Let_Result_3 := Match_Result; 
--# end
--# expr-done 2436
      Property_Result := Let_Result_3;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Formal_Param_Holder_Or_Null;
--# end
--# property-start BasicDecl.is_subprogram /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2729
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Is_Subprogram
  (Node : Bare_Basic_Decl
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Is_A : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2438 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2437 '<IsA BasicSubpDecl, BaseSubpBody, SubpBodyStub, EntryDecl at ast.py:2739>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2739
Is_A := Self /= null 
and then Self.Kind in Ada_Basic_Subp_Decl | Ada_Base_Subp_Body | Ada_Subp_Body_Stub_Range | Ada_Entry_Decl_Range; 
--# expr-done 2437
Let_Result := Is_A; 
--# end
--# expr-done 2438
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Is_Subprogram;
--# end
--# property-start BasicDecl.is_valid_reducer_candidate /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2742
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Is_Valid_Reducer_Candidate
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
            procedure Finalizer_Scope_2076 with Inline_Always;
      Var_Expr : Boolean;
Var_Expr_1 : Internal_Entity_Base_Subp_Spec;
Param_Types : Internal_Entity_Base_Type_Decl_Array_Access;
Return_Type : Internal_Entity_Base_Type_Decl;
Param_Modes : Internal_Entity_Mode_Array_Access;
Fld : Boolean;
Fld_1 : Internal_Entity_Base_Subp_Spec;
Fld_2 : Internal_Entity_Base_Type_Decl_Array_Access;
Fld_3 : Internal_Entity_Base_Type_Decl;
Len : Integer;
Is_Equal : Boolean;
Fld_4 : Internal_Entity_Mode_Array_Access;
Get_Result : Internal_Entity_Mode;
Is_A : Boolean;
Get_Result_1 : Internal_Entity_Mode;
Is_A_1 : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
Get_Result_2 : Internal_Entity_Base_Type_Decl;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Cast_Result_1 : Internal_Entity;
Is_Equiv : Boolean;
Is_Null : Boolean;
If_Result_1 : Boolean;
If_Result_2 : Boolean;
Let_Result_1 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
Let_Result_2 : Boolean;
            procedure Finalizer_Scope_2076 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Param_Types);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Param_Modes);
            end Finalizer_Scope_2076;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2456 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 2439 '<FieldAccess for is_subprogram at ast.py:2749>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2749
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_subprogram
Fld := Libadalang.Implementation.Basic_Decl_P_Is_Subprogram (Node => Ent.Node);
--# end
--# expr-done 2439
Var_Expr := Fld;
if Var_Expr /= False then
   --# scope-start
   --# bind _ Var_Expr
--# expr-start 2440 '<FieldAccess for subp_spec_or_null at ast.py:2750>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2750
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2440
Var_Expr_1 := Fld_1;
if Var_Expr_1 /= No_Entity_Base_Subp_Spec then
   --# scope-start
   --# bind spec Var_Expr_1
--# expr-start 2455 '<Let at ast.py:2751>' Let_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2751
--# scope-start
--# expr-start 2441 '<FieldAccess for param_types at ast.py:2752>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2752
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.param_types
Fld_2 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Param_Types (Node => Var_Expr_1.Node, Origin => No_Bare_Ada_Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 2441
Param_Types := Fld_2; Inc_Ref (Param_Types);
--# bind param_types Param_Types
--# expr-start 2442 '<FieldAccess for return_type at ast.py:2753>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2753
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.return_type
Fld_3 := Libadalang.Implementation.Base_Subp_Spec_P_Return_Type (Node => Var_Expr_1.Node, Origin => No_Bare_Ada_Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 2442
Return_Type := Fld_3; 
--# bind return_type Return_Type
--# expr-start 2444 '<Eq at ast.py:2757>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2757
--# expr-start 2443 '<length at ast.py:2757>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2757
Len := Length (Param_Types); 
--# expr-done 2443
Is_Equal := Len = 2; 
--# expr-done 2444
if Is_Equal then
   --# expr-start 2454 '<If at ast.py:2759>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2759
--# expr-start 2445 '<IsNull at ast.py:2760>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2760
Is_Null := Return_Type.Node = null; 
--# expr-done 2445
if Is_Null then
   --# expr-start 2451 '<Let at ast.py:2763>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2763
--# scope-start
--# expr-start 2446 '<FieldAccess for param_modes at ast.py:2764>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2764
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.param_modes
Fld_4 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Param_Modes (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 2446
Param_Modes := Fld_4; Inc_Ref (Param_Modes);
--# bind param_modes Param_Modes
--# expr-start 2448 '<IsA Mode.InOut at ast.py:2765>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2765
--# expr-start 2447 '<at at ast.py:2765>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2765
Get_Result := Get (Param_Modes, 0, True); 
--# expr-done 2447
Is_A := Get_Result.Node /= null 
and then Get_Result.Node.Kind in Ada_Mode_In_Out_Range; 
--# expr-done 2448
if Is_A then
   --# expr-start 2450 '<IsA Mode.In, Mode.Default at ast.py:2766>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2766
--# expr-start 2449 '<at at ast.py:2766>' Get_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2766
Get_Result_1 := Get (Param_Modes, 1, True); 
--# expr-done 2449
Is_A_1 := Get_Result_1.Node /= null 
and then Get_Result_1.Node.Kind in Ada_Mode_In_Range | Ada_Mode_Default_Range; 
--# expr-done 2450
   If_Result := Is_A_1;
else
   If_Result := False;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 2451
   If_Result_1 := Let_Result;
else
   --# expr-start 2453 '<Eq at ast.py:2771>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2771
Cast_Expr := Return_Type; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
--# expr-start 2452 '<at at ast.py:2771>' Get_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2771
Get_Result_2 := Get (Param_Types, 0, True); 
--# expr-done 2452
      Cast_Result_1 := Create_Internal_Entity
        (Node => Get_Result_2.Node,
         Info => Get_Result_2.Info);
Is_Equiv := Equivalent (Cast_Result, Cast_Result_1); 
--# expr-done 2453
   If_Result_1 := Is_Equiv;
end if;
--# expr-done 2454
   If_Result_2 := If_Result_1;
else
   If_Result_2 := False;
end if;
Let_Result_1 := If_Result_2; 
--# end
--# expr-done 2455
Scope_Result := Let_Result_1;
   --# end
      Finalizer_Scope_2076;
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Scope_Result_1 := Result_Var;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
Let_Result_2 := Result_Var_1; 
--# end
--# expr-done 2456
      Property_Result := Let_Result_2;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_2076;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Is_Valid_Reducer_Candidate;
--# end
--# property-start BasicDecl.is_stream_subprogram_for_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2779
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Is_Stream_Subprogram_For_Type
  (Node : Bare_Basic_Decl
      ; Typ : Internal_Entity_Base_Type_Decl
      ; Return_Obj : Boolean
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind typ Typ
   --# bind return_obj Return_Obj
   Property_Result : Boolean;
            procedure Finalizer_Scope_182 with Inline_Always;
            procedure Finalizer_Scope_2077 with Inline_Always;
            procedure Finalizer_Scope_2078 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Block_Var_1 : Internal_Entity_Defining_Name_Array_Access;
Var_Expr : Bare_Basic_Decl;
Var_Expr_1 : Internal_Entity_Base_Subp_Spec;
Sym : Symbol_Type;
Sym_1 : Symbol_Type;
Array_Lit : Symbol_Type_Array_Access;
Fld : Bare_Basic_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Sym_2 : Symbol_Type;
Env_Get_Result : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Classwide_Type_Decl;
Cast_Result_1 : Internal_Entity_Base_Type_Decl;
Fld_3 : Internal_Entity_Base_Subp_Spec;
Fld_4 : Internal_Entity_Defining_Name_Array_Access;
Scope_Result_1 : Internal_Entity_Defining_Name_Array_Access;
Result_Var_1 : Internal_Entity_Defining_Name_Array_Access;
Fld_5 : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_6 : Boolean;
Get_Result : Internal_Entity_Defining_Name;
Fld_7 : Internal_Entity_Base_Formal_Param_Decl;
Fld_8 : Internal_Entity_Base_Type_Decl;
Fld_9 : Boolean;
If_Result : Boolean;
Fld_10 : Internal_Entity_Base_Subp_Spec;
Fld_11 : Internal_Entity_Base_Type_Decl;
Fld_12 : Boolean;
Get_Result_1 : Internal_Entity_Defining_Name;
Fld_13 : Internal_Entity_Base_Formal_Param_Decl;
Fld_14 : Internal_Entity_Base_Type_Decl;
Fld_15 : Boolean;
If_Result_1 : Boolean;
If_Result_2 : Boolean;
Dyn_Var_Bind_Result : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_182 is
            begin
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Block_Var_1);
            end Finalizer_Scope_182;
            procedure Finalizer_Scope_2077 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2077;
            procedure Finalizer_Scope_2078 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_2078;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2484 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2466 '<Cast to BaseTypeDecl at ast.py:2782>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2782
--# expr-start 2465 '<FieldAccess for classwide_type at ast.py:2782>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2782
--# expr-start 2464 '<Cast to BaseTypeDecl at ast.py:2782>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2782
--# expr-start 2463 '<EnvGet at ast.py:2782>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2782
--# expr-start 2460 '<FieldAccess for get_unit_root_decl at ast.py:2782>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2782
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2459 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 2457 '<Symbol Ada at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Ada); 
--# expr-done 2457
--# expr-start 2458 '<Symbol Streams at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Streams); 
--# expr-done 2458
Array_Lit := Create_Symbol_Type_Array (Internal_Symbol_Type_Array'(1 => Sym, 2 => Sym_1)); 
--# expr-done 2459
--# property-call-start AdaNode.get_unit_root_decl
Fld := Libadalang.Implementation.Ada_Node_P_Get_Unit_Root_Decl (Node => Ent.Node, Name => Array_Lit, Kind => Unit_Specification);
--# end
--# expr-done 2460
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 2461 '<FieldAccess for children_env at ast.py:2782>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2782
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Children_Env (Node => Var_Expr);
--# expr-done 2461
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2077;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
--# expr-start 2462 '<Symbol Root_Stream_Type at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Root_Stream_Type); 
--# expr-done 2462
Env_Get_Result := AST_Envs.Get_First (Self => Result_Var, Key => Sym_2, Lookup_Kind => To_Lookup_Kind_Type (Flat), Categories => (Inherited_Primitives => True, Nocat => True)); 
--# expr-done 2463
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 2464
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.classwide_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Classwide_Type (Node => Cast_Result.Node, E_Info => Cast_Result.Info);
--# end
--# expr-done 2465
      Cast_Result_1 := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
--# expr-done 2466
Block_Var := Cast_Result_1; 
--# bind root_stream_type Block_Var
--# expr-start 2467 '<FieldAccess for subp_spec_or_null at ast.py:2788>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2788
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_3 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2467
Var_Expr_1 := Fld_3;
if Var_Expr_1 /= No_Entity_Base_Subp_Spec then
   --# scope-start
--# expr-start 2468 '<FieldAccess for unpacked_formal_params at ast.py:2788>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2788
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld_4 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 2468
Scope_Result_1 := Fld_4;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2078;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Internal_Entity_Defining_Name_Array_Type;
end if;
      Inc_Ref (Result_Var_1);
Block_Var_1 := Result_Var_1; Inc_Ref (Block_Var_1);
--# bind params Block_Var_1
--# expr-start 2483 '<bind at ast.py:2790>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2790
--# expr-start 2469 '<FieldAccess for origin_node at ast.py:2791>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2791
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_5 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 2469
Bound_Origin := Fld_5; 
--# expr-start 2470 '<FieldAccess for is_subprogram at ast.py:2792>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2792
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_subprogram
Fld_6 := Libadalang.Implementation.Basic_Decl_P_Is_Subprogram (Node => Self);
--# end
--# expr-done 2470
if Fld_6 then
   --# expr-start 2474 '<FieldAccess for is_access_to at ast.py:2793>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2793
--# expr-start 2473 '<FieldAccess for formal_type at ast.py:2793>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2793
--# expr-start 2472 '<FieldAccess for formal_decl at ast.py:2793>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2793
--# expr-start 2471 '<at at ast.py:2793>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2793
Get_Result := Get (Block_Var_1, 0, True); 
--# expr-done 2471
      if Get_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_7 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Get_Result.Node, E_Info => Get_Result.Info);
--# end
--# expr-done 2472
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamDecl.formal_type
Fld_8 := Libadalang.Implementation.Base_Formal_Param_Decl_P_Formal_Type (Node => Fld_7.Node, Origin => Bound_Origin, E_Info => Fld_7.Info);
--# end
--# expr-done 2473
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_access_to
Fld_9 := Libadalang.Implementation.Base_Type_Decl_P_Is_Access_To (Node => Fld_8.Node, Typ => Block_Var, Origin => Bound_Origin, E_Info => Fld_8.Info);
--# end
--# expr-done 2474
   If_Result := Fld_9;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 2482 '<If at ast.py:2795>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2795
if Return_Obj then
   --# expr-start 2477 '<FieldAccess for matching_formal_type at ast.py:2797>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2797
--# expr-start 2476 '<FieldAccess for return_type at ast.py:2797>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2797
--# expr-start 2475 '<FieldAccess for subp_spec_or_null at ast.py:2797>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2797
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_10 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2475
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.return_type
Fld_11 := Libadalang.Implementation.Base_Subp_Spec_P_Return_Type (Node => Fld_10.Node, Origin => Bound_Origin, E_Info => Fld_10.Info);
--# end
--# expr-done 2476
      if Fld_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_formal_type
Fld_12 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Type (Node => Fld_11.Node, Formal_Type => Typ, Origin => Bound_Origin, E_Info => Fld_11.Info);
--# end
--# expr-done 2477
   If_Result_1 := Fld_12;
else
   --# expr-start 2481 '<FieldAccess for matching_formal_type at ast.py:2798>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2798
--# expr-start 2480 '<FieldAccess for formal_type at ast.py:2798>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2798
--# expr-start 2479 '<FieldAccess for formal_decl at ast.py:2798>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2798
--# expr-start 2478 '<at at ast.py:2798>' Get_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2798
Get_Result_1 := Get (Block_Var_1, 1, True); 
--# expr-done 2478
      if Get_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_13 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Get_Result_1.Node, E_Info => Get_Result_1.Info);
--# end
--# expr-done 2479
      if Fld_13.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamDecl.formal_type
Fld_14 := Libadalang.Implementation.Base_Formal_Param_Decl_P_Formal_Type (Node => Fld_13.Node, Origin => Bound_Origin, E_Info => Fld_13.Info);
--# end
--# expr-done 2480
      if Fld_14.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_formal_type
Fld_15 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Type (Node => Fld_14.Node, Formal_Type => Typ, Origin => Bound_Origin, E_Info => Fld_14.Info);
--# end
--# expr-done 2481
   If_Result_1 := Fld_15;
end if;
--# expr-done 2482
   If_Result_2 := If_Result_1;
else
   If_Result_2 := False;
end if;
Dyn_Var_Bind_Result := If_Result_2; 
--# expr-done 2483
Let_Result := Dyn_Var_Bind_Result; 
--# end
--# expr-done 2484
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_182;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_182;
               Finalizer_Scope_2077;
               Finalizer_Scope_2078;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Is_Stream_Subprogram_For_Type;
--# end
--# property-start BasicDecl.is_put_image_subprogram_for_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2803
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Is_Put_Image_Subprogram_For_Type
  (Node : Bare_Basic_Decl
      ; Typ : Internal_Entity_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind typ Typ
   Property_Result : Boolean;
            procedure Finalizer_Scope_183 with Inline_Always;
            procedure Finalizer_Scope_2079 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Block_Var_1 : Internal_Entity_Defining_Name_Array_Access;
Var_Expr : Internal_Entity_Base_Subp_Spec;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Classwide_Type_Decl;
Cast_Result : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Base_Subp_Spec;
Fld_3 : Internal_Entity_Defining_Name_Array_Access;
Scope_Result : Internal_Entity_Defining_Name_Array_Access;
Result_Var : Internal_Entity_Defining_Name_Array_Access;
Fld_4 : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_5 : Boolean;
Get_Result : Internal_Entity_Defining_Name;
Fld_6 : Internal_Entity_Base_Formal_Param_Decl;
Fld_7 : Internal_Entity_Base_Type_Decl;
Fld_8 : Boolean;
If_Result : Boolean;
Get_Result_1 : Internal_Entity_Defining_Name;
Fld_9 : Internal_Entity_Base_Formal_Param_Decl;
Fld_10 : Internal_Entity_Base_Type_Decl;
Fld_11 : Boolean;
If_Result_1 : Boolean;
Dyn_Var_Bind_Result : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_183 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Block_Var_1);
            end Finalizer_Scope_183;
            procedure Finalizer_Scope_2079 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_2079;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2501 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2487 '<Cast to BaseTypeDecl at ast.py:2809>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2809
--# expr-start 2486 '<FieldAccess for classwide_type at ast.py:2809>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2809
--# expr-start 2485 '<FieldAccess for root_buffer_type at ast.py:2809>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2809
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.root_buffer_type
Fld := Libadalang.Implementation.Ada_Node_P_Root_Buffer_Type (Node => Self);
--# end
--# expr-done 2485
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.classwide_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Classwide_Type (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 2486
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
--# expr-done 2487
Block_Var := Cast_Result; 
--# bind root_buffer_type Block_Var
--# expr-start 2488 '<FieldAccess for subp_spec_or_null at ast.py:2811>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2811
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2488
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Base_Subp_Spec then
   --# scope-start
--# expr-start 2489 '<FieldAccess for unpacked_formal_params at ast.py:2811>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2811
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld_3 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 2489
Scope_Result := Fld_3;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2079;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Defining_Name_Array_Type;
end if;
      Inc_Ref (Result_Var);
Block_Var_1 := Result_Var; Inc_Ref (Block_Var_1);
--# bind params Block_Var_1
--# expr-start 2500 '<bind at ast.py:2813>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2813
--# expr-start 2490 '<FieldAccess for origin_node at ast.py:2814>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2814
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_4 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 2490
Bound_Origin := Fld_4; 
--# expr-start 2491 '<FieldAccess for is_subprogram at ast.py:2815>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2815
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_subprogram
Fld_5 := Libadalang.Implementation.Basic_Decl_P_Is_Subprogram (Node => Self);
--# end
--# expr-done 2491
if Fld_5 then
   --# expr-start 2495 '<FieldAccess for matching_formal_type at ast.py:2817>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2817
--# expr-start 2494 '<FieldAccess for formal_type at ast.py:2817>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2817
--# expr-start 2493 '<FieldAccess for formal_decl at ast.py:2817>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2817
--# expr-start 2492 '<at at ast.py:2817>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2817
Get_Result := Get (Block_Var_1, 0, True); 
--# expr-done 2492
      if Get_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_6 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Get_Result.Node, E_Info => Get_Result.Info);
--# end
--# expr-done 2493
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamDecl.formal_type
Fld_7 := Libadalang.Implementation.Base_Formal_Param_Decl_P_Formal_Type (Node => Fld_6.Node, Origin => Bound_Origin, E_Info => Fld_6.Info);
--# end
--# expr-done 2494
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_formal_type
Fld_8 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Type (Node => Fld_7.Node, Formal_Type => Block_Var, Origin => Bound_Origin, E_Info => Fld_7.Info);
--# end
--# expr-done 2495
   If_Result := Fld_8;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 2499 '<FieldAccess for matching_formal_type at ast.py:2820>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2820
--# expr-start 2498 '<FieldAccess for formal_type at ast.py:2820>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2820
--# expr-start 2497 '<FieldAccess for formal_decl at ast.py:2820>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2820
--# expr-start 2496 '<at at ast.py:2820>' Get_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2820
Get_Result_1 := Get (Block_Var_1, 1, True); 
--# expr-done 2496
      if Get_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_9 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Get_Result_1.Node, E_Info => Get_Result_1.Info);
--# end
--# expr-done 2497
      if Fld_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamDecl.formal_type
Fld_10 := Libadalang.Implementation.Base_Formal_Param_Decl_P_Formal_Type (Node => Fld_9.Node, Origin => Bound_Origin, E_Info => Fld_9.Info);
--# end
--# expr-done 2498
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_formal_type
Fld_11 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Type (Node => Fld_10.Node, Formal_Type => Typ, Origin => Bound_Origin, E_Info => Fld_10.Info);
--# end
--# expr-done 2499
   If_Result_1 := Fld_11;
else
   If_Result_1 := False;
end if;
Dyn_Var_Bind_Result := If_Result_1; 
--# expr-done 2500
Let_Result := Dyn_Var_Bind_Result; 
--# end
--# expr-done 2501
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_183;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_183;
               Finalizer_Scope_2079;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Is_Put_Image_Subprogram_For_Type;
--# end
--# property-start BasicDecl.can_be_paramless /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2825
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Can_Be_Paramless
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Base_Subp_Spec;
Fld : Internal_Entity_Base_Subp_Spec;
Fld_1 : Internal_Entity_Info;
Fld_2 : Internal_Metadata;
Fld_3 : Boolean;
Fld_4 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2507 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2502 '<FieldAccess for subp_spec_or_null at ast.py:2830>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2830
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2502
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Subp_Spec then
   --# scope-start
   --# bind ss Var_Expr
--# expr-start 2506 '<FieldAccess for paramless at ast.py:2831>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2831
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2505 '<FieldAccess for dottable_subp at ast.py:2832>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2832
--# expr-start 2504 '<FieldAccess for md at ast.py:2832>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2832
--# expr-start 2503 '<FieldAccess for info at ast.py:2832>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2832
Fld_1 := Ent.Info;
--# expr-done 2503
Fld_2 := Fld_1.Md;
--# expr-done 2504
Fld_3 := Fld_2.Dottable_Subp;
--# expr-done 2505
--# property-call-start BaseFormalParamHolder.paramless
Fld_4 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Paramless (Node => Var_Expr.Node, Dottable_Subp => Fld_3, Can_Be => True);
--# end
--# expr-done 2506
Scope_Result := Fld_4;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := True;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 2507
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Can_Be_Paramless;
--# end
--# property-start BasicDecl.is_paramless /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2838
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Is_Paramless
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Base_Subp_Spec;
Fld : Internal_Entity_Base_Subp_Spec;
Fld_1 : Internal_Entity_Info;
Fld_2 : Internal_Metadata;
Fld_3 : Boolean;
Fld_4 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2513 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2508 '<FieldAccess for subp_spec_or_null at ast.py:2843>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2843
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2508
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Subp_Spec then
   --# scope-start
   --# bind ss Var_Expr
--# expr-start 2512 '<FieldAccess for paramless at ast.py:2844>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2844
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2511 '<FieldAccess for dottable_subp at ast.py:2845>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2845
--# expr-start 2510 '<FieldAccess for md at ast.py:2845>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2845
--# expr-start 2509 '<FieldAccess for info at ast.py:2845>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2845
Fld_1 := Ent.Info;
--# expr-done 2509
Fld_2 := Fld_1.Md;
--# expr-done 2510
Fld_3 := Fld_2.Dottable_Subp;
--# expr-done 2511
--# property-call-start BaseFormalParamHolder.paramless
Fld_4 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Paramless (Node => Var_Expr.Node, Dottable_Subp => Fld_3, Can_Be => False);
--# end
--# expr-done 2512
Scope_Result := Fld_4;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := True;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 2513
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Is_Paramless;
--# end
--# property-start '[dispatcher]BasicDecl.constrain_prefix' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Decl_P_Constrain_Prefix
  (Node : Bare_Basic_Decl
      ; Prefix : Bare_Expr
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind prefix Prefix
   --# bind origin Origin
   Property_Result : Logic_Equation;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Decl (Self.Kind) is
               when Ada_Abstract_State_Decl .. Ada_Anonymous_Expr_Decl | Ada_Discriminant_Spec .. Ada_Generic_Formal_Type_Decl | Ada_Synthetic_Formal_Param_Decl .. Ada_Synth_Anonymous_Type_Decl | Ada_Expr_Function .. Ada_Single_Task_Decl =>
                  --# property-call-start BasicDecl.constrain_prefix
                  Property_Result := Basic_Decl_P_Constrain_Prefix
                    (Self
                        , Prefix
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Component_Decl =>
                  --# property-call-start ComponentDecl.constrain_prefix
                  Property_Result := Component_Decl_P_Constrain_Prefix
                    (Self
                        , Prefix
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Param_Spec =>
                  --# property-call-start ParamSpec.constrain_prefix
                  Property_Result := Param_Spec_P_Constrain_Prefix
                    (Self
                        , Prefix
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Abstract_Subp_Decl .. Ada_Synthetic_Subp_Decl =>
                  --# property-call-start BasicSubpDecl.constrain_prefix
                  Property_Result := Basic_Subp_Decl_P_Constrain_Prefix
                    (Self
                        , Prefix
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Decl_P_Constrain_Prefix;
--# end
--# property-start BasicDecl.relative_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2877
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Relative_Name
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Single_Tok_Node
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Single_Tok_Node;
      Var_Expr : Internal_Entity_Defining_Name;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Internal_Entity_Single_Tok_Node;
Scope_Result : Internal_Entity_Single_Tok_Node;
Result_Var : Internal_Entity_Single_Tok_Node;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
--# expr-start 2514 '<FieldAccess for defining_name at ast.py:2878>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2878
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2514
Var_Expr := Fld;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 2515 '<FieldAccess for relative_name at ast.py:2878>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2878
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.relative_name'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Relative_Name (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 2515
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Single_Tok_Node;
end if;
      Property_Result := Result_Var;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Relative_Name;
--# end
--# property-start BasicDecl.relative_name_text /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2884
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Relative_Name_Text
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Symbol_Type
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Symbol_Type;
      Var_Expr : Internal_Entity_Single_Tok_Node;
Fld : Internal_Entity_Single_Tok_Node;
Fld_1 : Bare_Single_Tok_Node;
Sym : Symbol_Type;
Scope_Result : Symbol_Type;
Result_Var : Symbol_Type;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
--# expr-start 2516 '<FieldAccess for relative_name at ast.py:2885>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2885
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.relative_name
Fld := Libadalang.Implementation.Basic_Decl_P_Relative_Name (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2516
Var_Expr := Fld;
if Var_Expr /= No_Entity_Single_Tok_Node then
   --# scope-start
--# expr-start 2517 '<GetSymbol at ast.py:2885>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2885
Fld_1 := Var_Expr.Node; 
Sym := Get_Symbol (Fld_1); 
--# expr-done 2517
Scope_Result := Sym;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := null;
end if;
      Property_Result := Result_Var;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Relative_Name_Text;
--# end
--# property-start BasicDecl.name_symbol /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2890
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Name_Symbol
  (Node : Bare_Basic_Decl
  )
   return Symbol_Type
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type;
      As_Entity : Internal_Entity_Basic_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Single_Tok_Node;
Fld_1 : Bare_Single_Tok_Node;
Sym : Symbol_Type;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2520 '<GetSymbol at ast.py:2890>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2890
--# expr-start 2519 '<FieldAccess for relative_name at ast.py:2890>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2890
--# expr-start 2518 '<as_bare_entity at ast.py:2890>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2890
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Decl;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 2518
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.relative_name
Fld := Libadalang.Implementation.Basic_Decl_P_Relative_Name (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 2519
Fld_1 := Fld.Node; 
Sym := Get_Symbol (Fld_1); 
--# expr-done 2520
      Property_Result := Sym;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Name_Symbol;
--# end
--# property-start BasicDecl.basic_decl_next_part_for_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2893
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Basic_Decl_Next_Part_For_Decl
  (Node : Bare_Basic_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_190 with Inline_Always;
      Block_Var : Internal_Unit;
Match_0 : Bare_Body_Node;
Match_1 : Bare_Basic_Decl;
Var_Expr : Internal_Entity_Defining_Name;
Var_Expr_1 : Internal_Entity_Basic_Decl;
Fld : Bare_Compilation_Unit;
Fld_1 : Bare_Basic_Decl;
As_Entity : Internal_Entity_Basic_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity_Defining_Name;
Scope_Result : Internal_Entity_Defining_Name;
Result_Var : Internal_Entity_Defining_Name;
Is_A : Boolean;
Is_A_1 : Boolean;
As_Entity_1 : Internal_Entity_Basic_Decl;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Basic_Decl;
Sym : Symbol_Type;
Fld_3 : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Boolean;
Not_Val : Boolean;
Fld_4 : Internal_Unit;
Scope_Result_1 : Internal_Unit;
Result_Var_1 : Internal_Unit;
Match_Prefix : Bare_Basic_Decl;
Cast_Result : Bare_Body_Node;
Let_Result : Internal_Unit;
Scope_Result_2 : Internal_Unit;
Cast_Result_1 : Bare_Basic_Decl;
Let_Result_1 : Internal_Unit;
Scope_Result_3 : Internal_Unit;
Match_Result : Internal_Unit;
Fld_5 : Lexical_Env := Empty_Env;
Sym_1 : Symbol_Type;
Env_Get_Result : Internal_Entity;
Cast_Result_2 : Internal_Entity_Basic_Decl;
Let_Result_2 : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_190 is
            begin
                     Dec_Ref (Fld_5);
            end Finalizer_Scope_190;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2537 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 2532 '<Match at ast.py:2901>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2901
--# expr-start 2522 '<FieldAccess for decl at ast.py:2901>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2901
--# expr-start 2521 '<FieldAccess for enclosing_compilation_unit at ast.py:2901>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2901
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.enclosing_compilation_unit
Fld := Libadalang.Implementation.Ada_Node_P_Enclosing_Compilation_Unit (Node => Self);
--# end
--# expr-done 2521
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.decl
Fld_1 := Libadalang.Implementation.Compilation_Unit_P_Decl (Node => Fld);
--# end
--# expr-done 2522
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld_1; 
case Ada_Basic_Decl (Match_Prefix.Kind) is
         when Ada_Expr_Function .. Ada_Task_Body =>
   --# scope-start
--# scope-start
      Cast_Result := Match_Prefix;
Match_0 := Cast_Result; 
--# bind _ Match_0
Let_Result := null; 
--# end
Scope_Result_2 := Let_Result;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Abstract_State_Decl .. Ada_Synthetic_Subp_Decl | Ada_Entry_Index_Spec .. Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Match_Prefix;
Match_1 := Cast_Result_1; 
--# bind b Match_1
--# expr-start 2523 '<as_bare_entity at ast.py:2904>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2904
Is_Null := Match_1 = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Decl;
else
As_Entity := (Info => No_Entity_Info, Node => Match_1); 
   If_Result := As_Entity;
end if;
--# expr-done 2523
Var_Expr_1 := If_Result;
if Var_Expr_1 /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 2524 '<FieldAccess for defining_name at ast.py:2904>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2904
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 2524
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Defining_Name;
end if;
Var_Expr := Result_Var;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 2531 '<FieldAccess for referenced_unit at ast.py:2904>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2904
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2530 '<Not at ast.py:2906>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2906
--# expr-start 2525 '<IsA BasePackageDecl, GenericPackageDecl, GenericPackageInstantiation, PackageRenamingDecl, SubpRenamingDecl, GenericRenamingDecl at ast.py:2910>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2910
Is_A := Match_1 /= null 
and then Match_1.Kind in Ada_Base_Package_Decl | Ada_Generic_Package_Decl_Range | Ada_Generic_Package_Instantiation_Range | Ada_Package_Renaming_Decl_Range | Ada_Subp_Renaming_Decl_Range | Ada_Generic_Renaming_Decl; 
--# expr-done 2525
if Is_A then
   If_Result_3 := True;
else
--# expr-start 2526 '<IsA BasicSubpDecl, GenericSubpDecl at ast.py:2921>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2921
Is_A_1 := Match_1 /= null 
and then Match_1.Kind in Ada_Basic_Subp_Decl | Ada_Generic_Subp_Decl_Range; 
--# expr-done 2526
if Is_A_1 then
   --# expr-start 2529 '<FieldAccess for has_aspect at ast.py:2922>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2922
--# expr-start 2527 '<as_entity at ast.py:2922>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2922
Is_Null_1 := Match_1 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Basic_Decl;
else
As_Entity_1 := (Info => E_Info, Node => Match_1); 
   If_Result_1 := As_Entity_1;
end if;
--# expr-done 2527
      if If_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2528 '<Symbol Import at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Import); 
--# expr-done 2528
--# property-call-start BasicDecl.has_aspect
Fld_3 := Libadalang.Implementation.Basic_Decl_P_Has_Aspect (Node => If_Result_1.Node, Name => Sym, Imprecise_Fallback => Imprecise_Fallback, E_Info => If_Result_1.Info);
--# end
--# expr-done 2529
   If_Result_2 := Fld_3;
else
   If_Result_2 := False;
end if;
   If_Result_3 := If_Result_2;
end if;
Not_Val := not (If_Result_3); 
--# expr-done 2530
--# property-call-start Name.referenced_unit
Fld_4 := Libadalang.Implementation.Name_P_Referenced_Unit (Node => Var_Expr.Node, Kind => Unit_Body, Not_Found_Is_Error => Not_Val);
--# end
--# expr-done 2531
Scope_Result_1 := Fld_4;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := null;
end if;
Let_Result_1 := Result_Var_1; 
--# end
Scope_Result_3 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_3; 
end case;
--# expr-done 2532
Block_Var := Match_Result; 
--# expr-start 2536 '<Cast to BasicDecl at ast.py:2928>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2928
--# expr-start 2535 '<EnvGet at ast.py:2928>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2928
--# expr-start 2533 '<FieldAccess for children_env at ast.py:2928>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2928
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Children_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 2533
--# expr-start 2534 '<Symbol __nextpart at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Nextpart); 
--# expr-done 2534
Env_Get_Result := AST_Envs.Get_First (Self => Fld_5, Key => Sym_1, Lookup_Kind => To_Lookup_Kind_Type (Minimal), Categories => (Inherited_Primitives => False, Nocat => True)); 
--# expr-done 2535
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_2 := Create_Internal_Entity_Basic_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         Cast_Result_2 := No_Entity_Basic_Decl;
   end if;
--# expr-done 2536
Let_Result_2 := Cast_Result_2; 
--# end
--# expr-done 2537
      Property_Result := Let_Result_2;
   --# end
      Finalizer_Scope_190;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_190;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Basic_Decl_Next_Part_For_Decl;
--# end
--# property-start '[dispatcher]BasicDecl.next_part_for_decl' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Decl_P_Next_Part_For_Decl
  (Node : Bare_Basic_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Decl (Self.Kind) is
               when Ada_Abstract_State_Decl .. Ada_Anonymous_Expr_Decl | Ada_Generic_Package_Internal .. Ada_Package_Decl | Ada_Entry_Index_Spec .. Ada_Error_Decl | Ada_Exception_Handler .. Ada_For_Loop_Var_Decl | Ada_Generic_Package_Instantiation .. Ada_Number_Decl | Ada_Package_Renaming_Decl .. Ada_Single_Task_Decl =>
                  --# property-call-start BasicDecl.next_part_for_decl
                  Property_Result := Basic_Decl_P_Next_Part_For_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Component_Decl .. Ada_Synthetic_Formal_Param_Decl =>
                  --# property-call-start BaseFormalParamDecl.next_part_for_decl
                  Property_Result := Base_Formal_Param_Decl_P_Next_Part_For_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Incomplete_Tagged_Type_Decl | Ada_Task_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start BaseTypeDecl.next_part_for_decl
                  Property_Result := Base_Type_Decl_P_Next_Part_For_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Protected_Type_Decl =>
                  --# property-call-start ProtectedTypeDecl.next_part_for_decl
                  Property_Result := Protected_Type_Decl_P_Next_Part_For_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Abstract_Subp_Decl .. Ada_Generic_Subp_Internal =>
                  --# property-call-start BasicSubpDecl.next_part_for_decl
                  Property_Result := Basic_Subp_Decl_P_Next_Part_For_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Synthetic_Subp_Decl =>
                  --# property-call-start SyntheticSubpDecl.next_part_for_decl
                  Property_Result := Synthetic_Subp_Decl_P_Next_Part_For_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Expr_Function .. Ada_Subp_Renaming_Decl | Ada_Entry_Body .. Ada_Task_Body =>
                  --# property-call-start Body.next_part_for_decl
                  Property_Result := Body_Node_P_Next_Part_For_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Package_Body_Stub .. Ada_Task_Body_Stub =>
                  --# property-call-start BodyStub.next_part_for_decl
                  Property_Result := Body_Stub_P_Next_Part_For_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Exception_Decl =>
                  --# property-call-start ExceptionDecl.next_part_for_decl
                  Property_Result := Exception_Decl_P_Next_Part_For_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Package_Decl =>
                  --# property-call-start GenericDecl.next_part_for_decl
                  Property_Result := Generic_Decl_P_Next_Part_For_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Subp_Decl =>
                  --# property-call-start GenericSubpDecl.next_part_for_decl
                  Property_Result := Generic_Subp_Decl_P_Next_Part_For_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Object_Decl .. Ada_No_Type_Object_Renaming_Decl =>
                  --# property-call-start ObjectDecl.next_part_for_decl
                  Property_Result := Object_Decl_P_Next_Part_For_Decl
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Decl_P_Next_Part_For_Decl;
--# end
--# property-start BasicDecl.body_part_for_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2947
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Body_Part_For_Decl
  (Node : Bare_Basic_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Body_Node
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Body_Node;
      Var_Expr : Internal_Entity_Basic_Decl;
Match_0 : Internal_Entity_Body_Stub;
Match_1 : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Basic_Decl;
Match_Prefix : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity_Body_Stub;
Let_Result : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Cast_Result_1 : Internal_Entity_Basic_Decl;
Let_Result_1 : Internal_Entity_Basic_Decl;
Scope_Result_1 : Internal_Entity_Basic_Decl;
Match_Result : Internal_Entity_Basic_Decl;
Scope_Result_2 : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Cast_Result_2 : Internal_Entity_Body_Node;
Let_Result_2 : Internal_Entity_Body_Node;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2542 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 2541 '<Cast to Body at ast.py:2956>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2956
--# expr-start 2538 '<FieldAccess for next_part_for_decl at ast.py:2956>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2956
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.next_part_for_decl'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Next_Part_For_Decl (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2538
Var_Expr := Fld;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind np Var_Expr
--# expr-start 2540 '<Match at ast.py:2957>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2957
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Var_Expr; 
case Ada_Basic_Decl (Match_Prefix.Node.Kind) is
         when Ada_Package_Body_Stub .. Ada_Task_Body_Stub =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Body_Stub
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind stub Match_0
--# expr-start 2539 '<FieldAccess for next_part_for_decl at ast.py:2958>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2958
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.next_part_for_decl'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Next_Part_For_Decl (Node => Match_0.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Match_0.Info);
--# end
--# expr-done 2539
Let_Result := Fld_1; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Abstract_State_Decl .. Ada_Subp_Renaming_Decl | Ada_Entry_Body .. Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Basic_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind other Match_1
Let_Result_1 := Match_1; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
end case;
--# expr-done 2540
Scope_Result_2 := Match_Result;
   --# end
   Result_Var := Scope_Result_2;
else
   Result_Var := No_Entity_Basic_Decl;
end if;
   if Result_Var.Node = null
      or else Result_Var.Node.Kind in Ada_Body_Node
   then
      Cast_Result_2 := Create_Internal_Entity_Body_Node
        (Node => Result_Var.Node,
         Info => Result_Var.Info);
   else
         Cast_Result_2 := No_Entity_Body_Node;
   end if;
--# expr-done 2541
Let_Result_2 := Cast_Result_2; 
--# end
--# expr-done 2542
      Property_Result := Let_Result_2;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Body_Part_For_Decl;
--# end
--# property-start '[dispatcher]BasicDecl.next_part_for_name' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Decl_P_Next_Part_For_Name
  (Node : Bare_Basic_Decl
      ; Sym : Symbol_Type
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind sym Sym
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Decl (Self.Kind) is
               when Ada_Abstract_State_Decl .. Ada_Anonymous_Expr_Decl | Ada_Generic_Package_Internal .. Ada_Number_Decl | Ada_Package_Renaming_Decl .. Ada_Single_Task_Decl =>
                  --# property-call-start BasicDecl.next_part_for_name
                  Property_Result := Basic_Decl_P_Next_Part_For_Name
                    (Self
                        , Sym
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Component_Decl .. Ada_Synthetic_Formal_Param_Decl =>
                  --# property-call-start BaseFormalParamDecl.next_part_for_name
                  Property_Result := Base_Formal_Param_Decl_P_Next_Part_For_Name
                    (Self
                        , Sym
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Object_Decl .. Ada_No_Type_Object_Renaming_Decl =>
                  --# property-call-start ObjectDecl.next_part_for_name
                  Property_Result := Object_Decl_P_Next_Part_For_Name
                    (Self
                        , Sym
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Decl_P_Next_Part_For_Name;
--# end
--# property-start '[dispatcher]BasicDecl.previous_part_for_name' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Decl_P_Previous_Part_For_Name
  (Node : Bare_Basic_Decl
      ; Sym : Symbol_Type
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind sym Sym
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Decl (Self.Kind) is
               when Ada_Abstract_State_Decl .. Ada_Anonymous_Expr_Decl | Ada_Generic_Package_Internal .. Ada_Number_Decl | Ada_Package_Renaming_Decl .. Ada_Single_Task_Decl =>
                  --# property-call-start BasicDecl.previous_part_for_name
                  Property_Result := Basic_Decl_P_Previous_Part_For_Name
                    (Self
                        , Sym
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Component_Decl .. Ada_Synthetic_Formal_Param_Decl =>
                  --# property-call-start BaseFormalParamDecl.previous_part_for_name
                  Property_Result := Base_Formal_Param_Decl_P_Previous_Part_For_Name
                    (Self
                        , Sym
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Object_Decl .. Ada_No_Type_Object_Renaming_Decl =>
                  --# property-call-start ObjectDecl.previous_part_for_name
                  Property_Result := Object_Decl_P_Previous_Part_For_Name
                    (Self
                        , Sym
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Decl_P_Previous_Part_For_Name;
--# end
--# property-start BasicDecl.canonical_part_for_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3013
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Canonical_Part_For_Name
  (Node : Bare_Basic_Decl
      ; Sym : Symbol_Type
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind sym Sym
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Var_Expr : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2545 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2543 '<FieldAccess for previous_part_for_name at ast.py:3018>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3018
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.previous_part_for_name'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Previous_Part_For_Name (Node => Ent.Node, Sym => Sym, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2543
Var_Expr := Fld;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind pp Var_Expr
--# expr-start 2544 '<FieldAccess for canonical_part_for_name at ast.py:3019>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3019
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.canonical_part_for_name
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Canonical_Part_For_Name (Node => Var_Expr.Node, Sym => Sym, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr.Info);
--# end
--# expr-done 2544
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := Ent;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 2545
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Canonical_Part_For_Name;
--# end
--# property-start BasicDecl.most_visible_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3024
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Most_Visible_Part
  (Node : Bare_Basic_Decl
      ; Origin : Bare_Ada_Node
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Var_Expr : Internal_Entity_Defining_Name;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Internal_Entity_Defining_Name;
Fld_2 : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2549 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2547 '<FieldAccess for most_visible_part at ast.py:3030>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3030
--# expr-start 2546 '<FieldAccess for defining_name_or_raise at ast.py:3030>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3030
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name_or_raise
Fld := Libadalang.Implementation.Basic_Decl_P_Defining_Name_Or_Raise (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2546
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.most_visible_part
Fld_1 := Libadalang.Implementation.Defining_Name_P_Most_Visible_Part (Node => Fld.Node, Origin => Origin, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld.Info);
--# end
--# expr-done 2547
Var_Expr := Fld_1;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 2548 '<FieldAccess for basic_decl at ast.py:3030>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3030
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld_2 := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 2548
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Basic_Decl;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 2549
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Most_Visible_Part;
--# end
--# property-start BasicDecl.most_visible_part_for_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3034
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Most_Visible_Part_For_Name
  (Node : Bare_Basic_Decl
      ; Sym : Symbol_Type
      ; Origin : Bare_Ada_Node
      ; Imprecise_Fallback : Boolean
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind sym Sym
   --# bind origin Origin
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_197 with Inline_Always;
      Block_Var : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Basic_Decl;
Is_Null : Boolean;
Is_Null_1 : Boolean;
Fld_1 : Internal_Entity_Basic_Decl;
Fld_2 : Boolean;
Fld_3 : Boolean;
If_Result : Boolean;
Not_Val : Boolean;
Fld_4 : Internal_Entity_Basic_Decl;
Fld_5 : Lexical_Env := Empty_Env;
Env_Get_Result : Internal_Entity_Array_Access;
Cast_Expr : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity;
Is_Equiv : Boolean;
Quantifier_Result : Boolean;
If_Result_1 : Internal_Entity_Basic_Decl;
As_Entity : Internal_Entity;
Is_Null_2 : Boolean;
If_Result_2 : Internal_Entity;
Fld_6 : Boolean;
Fld_7 : Internal_Entity_Basic_Decl;
If_Result_3 : Internal_Entity_Basic_Decl;
If_Result_4 : Internal_Entity_Basic_Decl;
If_Result_5 : Internal_Entity_Basic_Decl;
If_Result_6 : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_197 is
            begin
                     Dec_Ref (Fld_5);
                     Dec_Ref (Env_Get_Result);
            end Finalizer_Scope_197;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2567 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2550 '<FieldAccess for next_part_for_name at ast.py:3039>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3039
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.next_part_for_name'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Next_Part_For_Name (Node => Ent.Node, Sym => Sym, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2550
Block_Var := Fld; 
--# bind np Block_Var
--# expr-start 2566 '<Cond at ast.py:3040>' If_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3040
--# expr-start 2551 '<IsNull at ast.py:3042>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3042
Is_Null := Block_Var.Node = null; 
--# expr-done 2551
if Is_Null then
   If_Result_6 := Ent;
else
--# expr-start 2552 '<IsNull at ast.py:3046>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3046
Is_Null_1 := Origin = null; 
--# expr-done 2552
if Is_Null_1 then
   --# expr-start 2553 '<FieldAccess for most_visible_part_for_name at ast.py:3047>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3047
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.most_visible_part_for_name
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Most_Visible_Part_For_Name (Node => Block_Var.Node, Sym => Sym, Origin => Origin, Imprecise_Fallback => Imprecise_Fallback, E_Info => Block_Var.Info);
--# end
--# expr-done 2553
   If_Result_5 := Fld_1;
else
--# expr-start 2556 '<Not at ast.py:3051>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3051
--# expr-start 2554 '<FieldAccess for is_in_private_part at ast.py:3051>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3051
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_in_private_part'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_In_Private_Part (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 2554
if Fld_2 then
   If_Result := True;
else
   --# expr-start 2555 '<FieldAccess for is_in_public_part at ast.py:3051>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3051
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_in_public_part
Fld_3 := Libadalang.Implementation.Basic_Decl_P_Is_In_Public_Part (Node => Block_Var.Node);
--# end
--# expr-done 2555
   If_Result := Fld_3;
end if;
Not_Val := not (If_Result); 
--# expr-done 2556
if Not_Val then
   --# expr-start 2562 '<If at ast.py:3052>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3052
--# expr-start 2560 '<Contains at ast.py:3053>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3053
   --# expr-start 2558 '<EnvGet at ast.py:3053>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3053
--# expr-start 2557 '<FieldAccess for node_env at ast.py:3053>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3053
      if Origin = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Node_Env (Node => Origin);
--# expr-done 2557
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Fld_5, Key => Sym, Lookup_Kind => To_Lookup_Kind_Type (Recursive), Categories => (Inherited_Primitives => True, Nocat => True))); 
--# expr-done 2558
      Quantifier_Result := False;
      declare
         Collection : constant Internal_Entity_Array_Access := Env_Get_Result;
      begin
         for Item_113 of
               Collection.Items
         loop
   --# scope-start
               --# bind item Item_113
      --# expr-start 2559 '<Eq at ???>' Is_Equiv None
Cast_Expr := Ent; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Is_Equiv := Equivalent (Item_113, Cast_Result); 
--# expr-done 2559
      Quantifier_Result := Is_Equiv;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
--# expr-done 2560
if Quantifier_Result then
   --# expr-start 2561 '<FieldAccess for most_visible_part_for_name at ast.py:3054>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3054
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.most_visible_part_for_name
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Most_Visible_Part_For_Name (Node => Block_Var.Node, Sym => Sym, Origin => Origin, Imprecise_Fallback => Imprecise_Fallback, E_Info => Block_Var.Info);
--# end
--# expr-done 2561
   If_Result_1 := Fld_4;
else
   If_Result_1 := Ent;
end if;
--# expr-done 2562
   If_Result_4 := If_Result_1;
else
--# expr-start 2564 '<FieldAccess for is_visible at ast.py:3060>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3060
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2563 '<as_bare_entity at ast.py:3060>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3060
Is_Null_2 := Origin = null; 
if Is_Null_2 then
   If_Result_2 := No_Entity;
else
As_Entity := (Info => No_Entity_Info, Node => Origin); 
   If_Result_2 := As_Entity;
end if;
--# expr-done 2563
--# property-call-start BasicDecl.is_visible
Fld_6 := Libadalang.Implementation.Basic_Decl_P_Is_Visible (Node => Block_Var.Node, From_Node => If_Result_2, E_Info => Block_Var.Info);
--# end
--# expr-done 2564
if Fld_6 then
   --# expr-start 2565 '<FieldAccess for most_visible_part_for_name at ast.py:3061>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3061
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.most_visible_part_for_name
Fld_7 := Libadalang.Implementation.Basic_Decl_P_Most_Visible_Part_For_Name (Node => Block_Var.Node, Sym => Sym, Origin => Origin, Imprecise_Fallback => Imprecise_Fallback, E_Info => Block_Var.Info);
--# end
--# expr-done 2565
   If_Result_3 := Fld_7;
else
   If_Result_3 := Ent;
end if;
   If_Result_4 := If_Result_3;
end if;
   If_Result_5 := If_Result_4;
end if;
   If_Result_6 := If_Result_5;
end if;
--# expr-done 2566
Let_Result := If_Result_6; 
--# end
--# expr-done 2567
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_197;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_197;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Most_Visible_Part_For_Name;
--# end
--# property-start BasicDecl.fully_qualified_name_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3068
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Fully_Qualified_Name_Impl
  (Node : Bare_Basic_Decl
      ; Include_Profile : Boolean
         := False
      ; Dn : Internal_Entity_Defining_Name
         := No_Entity_Defining_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return String_Type_Array_Access
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind include_profile Include_Profile
   --# bind dn Dn
   Property_Result : String_Type_Array_Access;
            procedure Finalizer_Scope_199 with Inline_Always;
            procedure Finalizer_Scope_2089 with Inline_Always;
            procedure Finalizer_Scope_2090 with Inline_Always;
            procedure Finalizer_Scope_2091 with Inline_Always;
            procedure Finalizer_Scope_2092 with Inline_Always;
            procedure Finalizer_Scope_2093 with Inline_Always;
            procedure Finalizer_Scope_2094 with Inline_Always;
            procedure Finalizer_Scope_2095 with Inline_Always;
            procedure Finalizer_Scope_2096 with Inline_Always;
            procedure Finalizer_Scope_2097 with Inline_Always;
      Block_Var : Boolean;
Block_Var_1 : String_Type_Array_Access;
Block_Var_2 : String_Type_Array_Access;
Block_Var_3 : String_Type_Array_Access;
Match_0 : Internal_Entity_Synthetic_Defining_Name;
Match_1 : Internal_Entity_Defining_Name;
Var_Expr : String_Type_Array_Access;
Var_Expr_1 : Internal_Entity_Basic_Decl;
Var_Expr_2 : String_Type_Array_Access;
Var_Expr_3 : Bare_Subunit;
Is_Null : Boolean;
Fld : Internal_Entity_Defining_Name_Array_Access;
Len : Integer;
Comp_Result : Boolean;
If_Result : Boolean;
If_Result_1 : Boolean;
Fld_1 : Internal_Entity_Defining_Name;
Is_Null_1 : Boolean;
If_Result_2 : Internal_Entity_Defining_Name;
Fld_2 : Symbol_Type;
Fld_3 : String_Type;
Array_Lit : String_Type_Array_Access;
Fld_4 : Bare_Single_Tok_Node_Array_Access;
Fld_5 : String_Type;
Map_Result : String_Type_Array_Access;
Match_Prefix : Internal_Entity_Defining_Name;
Cast_Result : Internal_Entity_Synthetic_Defining_Name;
Let_Result : String_Type_Array_Access;
Scope_Result : String_Type_Array_Access;
Cast_Result_1 : Internal_Entity_Defining_Name;
Let_Result_1 : String_Type_Array_Access;
Scope_Result_1 : String_Type_Array_Access;
Match_Result : String_Type_Array_Access;
Fld_6 : String_Type;
Str : String_Type;
If_Result_3 : String_Type;
Concat_Result : String_Type;
Is_A : Boolean;
Str_1 : String_Type;
Is_A_1 : Boolean;
Str_2 : String_Type;
Is_A_2 : Boolean;
Str_3 : String_Type;
Str_4 : String_Type;
If_Result_4 : String_Type;
If_Result_5 : String_Type;
If_Result_6 : String_Type;
Str_5 : String_Type;
Len_1 : Integer;
Arith_Result : Integer;
Is_Equal : Boolean;
If_Result_7 : String_Type;
Concat_Result_1 : String_Type;
Index_116 : Integer;
Map_Result_1 : String_Type_Array_Access;
Fld_7 : Internal_Entity_Basic_Decl;
Fld_8 : String_Type_Array_Access;
Scope_Result_2 : String_Type_Array_Access;
Result_Var : String_Type_Array_Access;
Concat_Result_2 : String_Type_Array_Access;
Is_A_3 : Boolean;
If_Result_8 : String_Type_Array_Access;
Scope_Result_3 : String_Type_Array_Access;
Result_Var_1 : String_Type_Array_Access;
Fld_9 : Boolean;
If_Result_9 : String_Type_Array_Access;
Fld_10 : Bare_Ada_Node;
Cast_Result_2 : Bare_Subunit;
Fld_11 : Bare_Name;
Fld_12 : Bare_Single_Tok_Node_Array_Access;
Scope_Result_4 : Bare_Single_Tok_Node_Array_Access;
Result_Var_2 : Bare_Single_Tok_Node_Array_Access;
Fld_13 : String_Type;
Map_Result_2 : String_Type_Array_Access;
Concat_Result_3 : String_Type_Array_Access;
Scope_Result_5 : String_Type_Array_Access;
Result_Var_3 : String_Type_Array_Access;
Let_Result_2 : String_Type_Array_Access;
            procedure Finalizer_Scope_199 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Match_Result);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Block_Var_2);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (If_Result_9);
                     Dec_Ref (Block_Var_3);
                     Dec_Ref (Scope_Result_4);
                     Dec_Ref (Result_Var_2);
                     Dec_Ref (Map_Result_2);
                     Dec_Ref (Concat_Result_3);
                     Dec_Ref (Scope_Result_5);
                     Dec_Ref (Result_Var_3);
                     Dec_Ref (Let_Result_2);
            end Finalizer_Scope_199;
            procedure Finalizer_Scope_2089 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Array_Lit);
            end Finalizer_Scope_2089;
            procedure Finalizer_Scope_2090 is
            begin
                     Dec_Ref (Fld_4);
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_2090;
            procedure Finalizer_Scope_2091 is
            begin
                     Dec_Ref (Fld_5);
            end Finalizer_Scope_2091;
            procedure Finalizer_Scope_2092 is
            begin
                     Dec_Ref (Fld_6);
                     Dec_Ref (Str);
                     Dec_Ref (If_Result_3);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Str_1);
                     Dec_Ref (Str_2);
                     Dec_Ref (Str_3);
                     Dec_Ref (Str_4);
                     Dec_Ref (If_Result_4);
                     Dec_Ref (If_Result_5);
                     Dec_Ref (If_Result_6);
                     Dec_Ref (Str_5);
                     Dec_Ref (If_Result_7);
                     Dec_Ref (Concat_Result_1);
            end Finalizer_Scope_2092;
            procedure Finalizer_Scope_2093 is
            begin
                     Dec_Ref (Fld_8);
            end Finalizer_Scope_2093;
            procedure Finalizer_Scope_2094 is
            begin
                     Dec_Ref (Var_Expr);
                     Dec_Ref (Concat_Result_2);
                     Dec_Ref (If_Result_8);
            end Finalizer_Scope_2094;
            procedure Finalizer_Scope_2095 is
            begin
                     Dec_Ref (Fld_12);
            end Finalizer_Scope_2095;
            procedure Finalizer_Scope_2096 is
            begin
                     Dec_Ref (Fld_13);
            end Finalizer_Scope_2096;
            procedure Finalizer_Scope_2097 is
            begin
                     Dec_Ref (Var_Expr_2);
            end Finalizer_Scope_2097;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2615 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 2571 '<If at ast.py:3084>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3084
--# expr-start 2568 '<IsNull at ast.py:3085>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3085
Is_Null := Dn.Node = null; 
--# expr-done 2568
if Is_Null then
   --# expr-start 2570 '<length at ast.py:3085>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3085
--# expr-start 2569 '<FieldAccess for defining_names at ast.py:3085>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3085
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_names'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Names (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2569
Len := Length (Fld); 
--# expr-done 2570
Comp_Result := Len > 1; 
   If_Result := Comp_Result;
else
   If_Result := False;
end if;
if If_Result then
   If_Result_1 := raise Property_Error with "Can't call on a declaration with several names";
else
   If_Result_1 := False;
end if;
--# expr-done 2571
Block_Var := If_Result_1; 
--# expr-start 2581 '<Match at ast.py:3091>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3091
--# expr-start 2574 '<If at ast.py:3091>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3091
--# expr-start 2572 '<IsNull at ast.py:3092>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3092
Is_Null_1 := Dn.Node = null; 
--# expr-done 2572
if Is_Null_1 then
   --# expr-start 2573 '<FieldAccess for defining_name at ast.py:3092>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3092
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2573
   If_Result_2 := Fld_1;
else
   If_Result_2 := Dn;
end if;
--# expr-done 2574
      if If_Result_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := If_Result_2; 
case Ada_Defining_Name_Range (Match_Prefix.Node.Kind) is
         when Ada_Synthetic_Defining_Name =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Synthetic_Defining_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind scel Match_0
--# expr-start 2577 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 2576 '<FieldAccess for image at ast.py:3094>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3094
--# expr-start 2575 '<FieldAccess for name_symbol at ast.py:3094>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3094
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Match_0.Node);
--# end
--# expr-done 2575
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Image (Self => Fld_2);
--# expr-done 2576
Array_Lit := Create_String_Type_Array (Internal_String_Type_Array'(1 => Fld_3)); 
--# expr-done 2577
Let_Result := Array_Lit; Inc_Ref (Let_Result);
--# end
Scope_Result := Let_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2089;
            Match_Result := Scope_Result; Inc_Ref (Match_Result);
         when Ada_Defining_Name =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Defining_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind n Match_1
--# expr-start 2580 '<Map at ast.py:3095>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3095
   --# expr-start 2578 '<FieldAccess for as_single_tok_node_array at ast.py:3095>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3095
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.as_single_tok_node_array
Fld_4 := Libadalang.Implementation.Name_P_As_Single_Tok_Node_Array (Node => Match_1.Node);
--# end
--# expr-done 2578
      declare
         Map_Result_Vec : String_Type_Vectors.Vector;
      begin
      declare
         Collection : constant Bare_Single_Tok_Node_Array_Access := Fld_4;
      begin
         for Item_114 of
               Collection.Items
         loop
   --# scope-start
               --# bind t Item_114
      --# expr-start 2579 '<FieldAccess for text at ast.py:3095>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3095
      if Item_114 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Text (Node => Item_114);
--# expr-done 2579
         declare
            Item_To_Append : constant String_Type := Fld_5;
         begin
               Inc_Ref (Item_To_Append);
            String_Type_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2091;
         end loop;
      end;
         Map_Result := Create_String_Type_Array
           (Items_Count => Natural (String_Type_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := String_Type_Vectors.Get
              (Map_Result_Vec,
               I + String_Type_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         String_Type_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 2580
Let_Result_1 := Map_Result; Inc_Ref (Let_Result_1);
--# end
Scope_Result_1 := Let_Result_1;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2090;
            Match_Result := Scope_Result_1; Inc_Ref (Match_Result);
end case;
--# expr-done 2581
Block_Var_1 := Match_Result; Inc_Ref (Block_Var_1);
--# bind def_name Block_Var_1
--# expr-start 2600 '<Map at ast.py:3098>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3098
      declare
         Map_Result_1_Vec : String_Type_Vectors.Vector;
      begin
         Index_116 := 0;
      declare
         Collection : constant String_Type_Array_Access := Block_Var_1;
      begin
         for Item_115 of
               Collection.Items
         loop
   --# scope-start
               --# bind t Item_115
               --# bind i Index_116
      --# expr-start 2599 '<Concat at ast.py:3099>' Concat_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3099
--# expr-start 2585 '<Concat at ast.py:3099>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3099
--# expr-start 2584 '<If at ast.py:3100>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3100
if Include_Profile then
   --# expr-start 2582 '<FieldAccess for custom_id_text at ast.py:3100>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3100
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.custom_id_text'
Fld_6 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Custom_Id_Text (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2582
   If_Result_3 := Fld_6;
else
   --# expr-start 2583 '<String at ast.py:3100>' Str /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3100
Str := Create_String (""); 
--# expr-done 2583
   If_Result_3 := Str;
end if;
      Inc_Ref (If_Result_3);
--# expr-done 2584
Concat_Result := Concat_String (Item_115, If_Result_3); 
--# expr-done 2585
--# expr-start 2598 '<If at ast.py:3101>' If_Result_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3101
--# expr-start 2588 '<Eq at ast.py:3102>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3102
--# expr-start 2587 '<Arithmetic - at ast.py:3102>' Arith_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3102
--# expr-start 2586 '<length at ast.py:3102>' Len_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3102
Len_1 := Length (Block_Var_1); 
--# expr-done 2586
Arith_Result := (Len_1 - 1); 
--# expr-done 2587
Is_Equal := Index_116 = Arith_Result; 
--# expr-done 2588
if Is_Equal then
   --# expr-start 2596 '<Cond at ast.py:3103>' If_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3103
--# expr-start 2589 '<IsA ClasswideTypeDecl at ast.py:3103>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3103
Is_A := Ent.Node /= null 
and then Ent.Node.Kind in Ada_Classwide_Type_Decl_Range; 
--# expr-done 2589
if Is_A then
   --# expr-start 2590 '<String at ast.py:3103>' Str_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3103
Str_1 := Create_String ("'Class"); 
--# expr-done 2590
   If_Result_6 := Str_1;
else
--# expr-start 2591 '<IsA DiscreteBaseSubtypeDecl at ast.py:3104>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3104
Is_A_1 := Ent.Node /= null 
and then Ent.Node.Kind in Ada_Discrete_Base_Subtype_Decl_Range; 
--# expr-done 2591
if Is_A_1 then
   --# expr-start 2592 '<String at ast.py:3104>' Str_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3104
Str_2 := Create_String ("'Base"); 
--# expr-done 2592
   If_Result_5 := Str_2;
else
--# expr-start 2593 '<IsA SynthAnonymousTypeDecl at ast.py:3112>' Is_A_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3112
Is_A_2 := Ent.Node /= null 
and then Ent.Node.Kind in Ada_Synth_Anonymous_Type_Decl_Range; 
--# expr-done 2593
if Is_A_2 then
   --# expr-start 2594 '<String at ast.py:3113>' Str_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3113
Str_3 := Create_String ("'Anonymous_Access"); 
--# expr-done 2594
   If_Result_4 := Str_3;
else
   --# expr-start 2595 '<String at ast.py:3115>' Str_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3115
Str_4 := Create_String (""); 
--# expr-done 2595
   If_Result_4 := Str_4;
end if;
      Inc_Ref (If_Result_4);
   If_Result_5 := If_Result_4;
end if;
      Inc_Ref (If_Result_5);
   If_Result_6 := If_Result_5;
end if;
      Inc_Ref (If_Result_6);
--# expr-done 2596
   If_Result_7 := If_Result_6;
else
   --# expr-start 2597 '<String at ast.py:3116>' Str_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3116
Str_5 := Create_String (""); 
--# expr-done 2597
   If_Result_7 := Str_5;
end if;
      Inc_Ref (If_Result_7);
--# expr-done 2598
Concat_Result_1 := Concat_String (Concat_Result, If_Result_7); 
--# expr-done 2599
         declare
            Item_To_Append : constant String_Type := Concat_Result_1;
         begin
               Inc_Ref (Item_To_Append);
            String_Type_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2092;
               Index_116 := Index_116 + 1;
         end loop;
      end;
         Map_Result_1 := Create_String_Type_Array
           (Items_Count => Natural (String_Type_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := String_Type_Vectors.Get
              (Map_Result_1_Vec,
               I + String_Type_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         String_Type_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 2600
Block_Var_2 := Map_Result_1; Inc_Ref (Block_Var_2);
--# bind self_name Block_Var_2
--# expr-start 2607 '<If at ast.py:3120>' If_Result_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3120
--# expr-start 2601 '<FieldAccess for is_compilation_unit_root at ast.py:3121>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3121
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_compilation_unit_root
Fld_9 := Libadalang.Implementation.Basic_Decl_P_Is_Compilation_Unit_Root (Node => Ent.Node);
--# end
--# expr-done 2601
if Fld_9 then
   If_Result_9 := Block_Var_2;
else
--# expr-start 2602 '<FieldAccess for parent_basic_decl at ast.py:3124>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3124
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.parent_basic_decl
Fld_7 := Libadalang.Implementation.Ada_Node_P_Parent_Basic_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2602
Var_Expr_1 := Fld_7;
if Var_Expr_1 /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 2603 '<FieldAccess for fully_qualified_name_impl at ast.py:3124>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3124
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.fully_qualified_name_impl
Fld_8 := Libadalang.Implementation.Basic_Decl_P_Fully_Qualified_Name_Impl (Node => Var_Expr_1.Node, Include_Profile => Include_Profile, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 2603
Scope_Result_2 := Fld_8;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2093;
   Result_Var := Scope_Result_2;
else
   Result_Var := No_String_Type_Array_Type;
end if;
      Inc_Ref (Result_Var);
Var_Expr := Result_Var;
      Inc_Ref (Var_Expr);
if Var_Expr /= No_String_Type_Array_Type then
   --# scope-start
   --# bind fqn Var_Expr
--# expr-start 2606 '<If at ast.py:3126>' If_Result_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3126
--# expr-start 2604 '<IsA GenericPackageInternal, GenericSubpInternal at ast.py:3127>' Is_A_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3127
Is_A_3 := Self /= null 
and then Self.Kind in Ada_Generic_Package_Internal_Range | Ada_Generic_Subp_Internal_Range; 
--# expr-done 2604
if Is_A_3 then
   If_Result_8 := Var_Expr;
else
   --# expr-start 2605 '<Concat at ast.py:3129>' Concat_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3129
Concat_Result_2 := Concat (Var_Expr, Block_Var_2); 
--# expr-done 2605
   If_Result_8 := Concat_Result_2;
end if;
      Inc_Ref (If_Result_8);
--# expr-done 2606
Scope_Result_3 := If_Result_8;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_2094;
   Result_Var_1 := Scope_Result_3;
else
   Result_Var_1 := No_String_Type_Array_Type;
end if;
      Inc_Ref (Result_Var_1);
   If_Result_9 := Result_Var_1;
end if;
      Inc_Ref (If_Result_9);
--# expr-done 2607
Block_Var_3 := If_Result_9; Inc_Ref (Block_Var_3);
--# bind fqn Block_Var_3
--# expr-start 2614 '<Concat at ast.py:3133>' Concat_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3133
--# expr-start 2613 '<Map at ast.py:3133>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3133
--# expr-start 2609 '<Cast to Subunit at ast.py:3133>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3133
--# expr-start 2608 '<parent at ast.py:3133>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3133
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Self.Parent; 
--# expr-done 2608
   if Fld_10 = null
      or else Fld_10.Kind in Ada_Subunit_Range
   then
      Cast_Result_2 := Fld_10;
   else
         Cast_Result_2 := No_Bare_Ada_Node;
   end if;
--# expr-done 2609
Var_Expr_3 := Cast_Result_2;
if Var_Expr_3 /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 2611 '<FieldAccess for as_single_tok_node_array at ast.py:3133>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3133
--# expr-start 2610 '<FieldAccess for name at ast.py:3133>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3133
      if Var_Expr_3 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Var_Expr_3.Subunit_F_Name;
--# expr-done 2610
      if Fld_11 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.as_single_tok_node_array
Fld_12 := Libadalang.Implementation.Name_P_As_Single_Tok_Node_Array (Node => Fld_11);
--# end
--# expr-done 2611
Scope_Result_4 := Fld_12;
   Inc_Ref (Scope_Result_4);
   --# end
      Finalizer_Scope_2095;
   Result_Var_2 := Scope_Result_4;
else
   Result_Var_2 := No_Bare_Single_Tok_Node_Array_Type;
end if;
      Inc_Ref (Result_Var_2);
      declare
         Map_Result_2_Vec : String_Type_Vectors.Vector;
      begin
      declare
         Collection : constant Bare_Single_Tok_Node_Array_Access := Result_Var_2;
      begin
         for Item_117 of
               Collection.Items
         loop
   --# scope-start
               --# bind t Item_117
      --# expr-start 2612 '<FieldAccess for text at ast.py:3135>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3135
      if Item_117 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Libadalang.Implementation.Text (Node => Item_117);
--# expr-done 2612
         declare
            Item_To_Append : constant String_Type := Fld_13;
         begin
               Inc_Ref (Item_To_Append);
            String_Type_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2096;
         end loop;
      end;
         Map_Result_2 := Create_String_Type_Array
           (Items_Count => Natural (String_Type_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := String_Type_Vectors.Get
              (Map_Result_2_Vec,
               I + String_Type_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         String_Type_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 2613
Concat_Result_3 := Concat (Map_Result_2, Block_Var_3); 
--# expr-done 2614
Var_Expr_2 := Concat_Result_3;
      Inc_Ref (Var_Expr_2);
if Var_Expr_2 /= No_String_Type_Array_Type then
   --# scope-start
   --# bind e Var_Expr_2
Scope_Result_5 := Var_Expr_2;
   Inc_Ref (Scope_Result_5);
   --# end
      Finalizer_Scope_2097;
   Result_Var_3 := Scope_Result_5;
else
   Result_Var_3 := Block_Var_3;
end if;
      Inc_Ref (Result_Var_3);
Let_Result_2 := Result_Var_3; Inc_Ref (Let_Result_2);
--# end
--# expr-done 2615
      Property_Result := Let_Result_2;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_199;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_199;
               Finalizer_Scope_2089;
               Finalizer_Scope_2090;
               Finalizer_Scope_2091;
               Finalizer_Scope_2092;
               Finalizer_Scope_2093;
               Finalizer_Scope_2094;
               Finalizer_Scope_2095;
               Finalizer_Scope_2096;
               Finalizer_Scope_2097;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Fully_Qualified_Name_Impl;
--# end
--# property-start BasicDecl.fully_qualified_name_string_array /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3138
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Fully_Qualified_Name_String_Array
  (Node : Bare_Basic_Decl
      ; Include_Profile : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return String_Type_Array_Access
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind include_profile Include_Profile
   Property_Result : String_Type_Array_Access;
            procedure Finalizer_Scope_200 with Inline_Always;
      Fld : String_Type_Array_Access;
Let_Result : String_Type_Array_Access;
            procedure Finalizer_Scope_200 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_200;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2617 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2616 '<FieldAccess for fully_qualified_name_impl at ast.py:3143>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3143
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.fully_qualified_name_impl
Fld := Libadalang.Implementation.Basic_Decl_P_Fully_Qualified_Name_Impl (Node => Ent.Node, Include_Profile => Include_Profile, E_Info => Ent.Info);
--# end
--# expr-done 2616
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 2617
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_200;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_200;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Fully_Qualified_Name_String_Array;
--# end
--# property-start BasicDecl.fully_qualified_name_array /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3148
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Fully_Qualified_Name_Array
  (Node : Bare_Basic_Decl
      ; Include_Profile : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind include_profile Include_Profile
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_201 with Inline_Always;
      Fld : String_Type_Array_Access;
Sym : Symbol_Type;
Map_Result : Symbol_Type_Array_Access;
Let_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_201 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_201;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2621 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2620 '<Map at ast.py:3153>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3153
   --# expr-start 2618 '<FieldAccess for fully_qualified_name_string_array at ast.py:3153>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3153
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.fully_qualified_name_string_array
Fld := Libadalang.Implementation.Basic_Decl_P_Fully_Qualified_Name_String_Array (Node => Ent.Node, Include_Profile => Include_Profile, E_Info => Ent.Info);
--# end
--# expr-done 2618
      declare
         Map_Result_Vec : Symbol_Type_Vectors.Vector;
      begin
      declare
         Collection : constant String_Type_Array_Access := Fld;
      begin
         for Item_118 of
               Collection.Items
         loop
   --# scope-start
               --# bind t Item_118
      --# expr-start 2619 '<to_symbol at ast.py:3155>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3155
Sym := String_To_Symbol (Self.Unit.Context, Item_118); 
--# expr-done 2619
         declare
            Item_To_Append : constant Symbol_Type := Sym;
         begin
            Symbol_Type_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Symbol_Type_Array
           (Items_Count => Natural (Symbol_Type_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Symbol_Type_Vectors.Get
              (Map_Result_Vec,
               I + Symbol_Type_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Symbol_Type_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 2620
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 2621
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_201;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_201;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Fully_Qualified_Name_Array;
--# end
--# property-start BasicDecl.fully_qualified_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3158
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Fully_Qualified_Name
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return String_Type
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : String_Type;
            procedure Finalizer_Scope_202 with Inline_Always;
      Str : String_Type;
Fld : String_Type_Array_Access;
Join_Result : String_Type;
Let_Result : String_Type;
            procedure Finalizer_Scope_202 is
            begin
                     Dec_Ref (Str);
                     Dec_Ref (Fld);
                     Dec_Ref (Join_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_202;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2625 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2624 '<Join at ast.py:3162>' Join_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3162
--# expr-start 2622 '<String at ast.py:3162>' Str /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3162
Str := Create_String ("."); 
--# expr-done 2622
--# expr-start 2623 '<FieldAccess for fully_qualified_name_impl at ast.py:3162>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3162
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.fully_qualified_name_impl
Fld := Libadalang.Implementation.Basic_Decl_P_Fully_Qualified_Name_Impl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2623
Join_Result := Join_Strings (Str, Fld); 
--# expr-done 2624
Let_Result := Join_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 2625
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_202;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_202;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Fully_Qualified_Name;
--# end
--# property-start BasicDecl.canonical_fully_qualified_name_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3165
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Canonical_Fully_Qualified_Name_Impl
  (Node : Bare_Basic_Decl
      ; Include_Profile : Boolean
         := False
      ; Dn : Internal_Entity_Defining_Name
         := No_Entity_Defining_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return String_Type
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind include_profile Include_Profile
   --# bind dn Dn
   Property_Result : String_Type;
            procedure Finalizer_Scope_203 with Inline_Always;
            procedure Finalizer_Scope_2100 with Inline_Always;
      Str : String_Type;
Fld : String_Type_Array_Access;
Sym : Symbol_Type;
Map_Result : Symbol_Type_Array_Access;
Fld_1 : String_Type;
Map_Result_1 : String_Type_Array_Access;
Join_Result : String_Type;
Let_Result : String_Type;
            procedure Finalizer_Scope_203 is
            begin
                     Dec_Ref (Str);
                     Dec_Ref (Fld);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Join_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_203;
            procedure Finalizer_Scope_2100 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2100;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2633 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2632 '<Join at ast.py:3172>' Join_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3172
--# expr-start 2626 '<String at ast.py:3172>' Str /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3172
Str := Create_String ("."); 
--# expr-done 2626
--# expr-start 2631 '<Map at ast.py:3173>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3173
   --# expr-start 2629 '<Map at ast.py:3173>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3173
   --# expr-start 2627 '<FieldAccess for fully_qualified_name_impl at ast.py:3173>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3173
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.fully_qualified_name_impl
Fld := Libadalang.Implementation.Basic_Decl_P_Fully_Qualified_Name_Impl (Node => Ent.Node, Include_Profile => Include_Profile, Dn => Dn, E_Info => Ent.Info);
--# end
--# expr-done 2627
      declare
         Map_Result_Vec : Symbol_Type_Vectors.Vector;
      begin
      declare
         Collection : constant String_Type_Array_Access := Fld;
      begin
         for Item_120 of
               Collection.Items
         loop
   --# scope-start
               --# bind t Item_120
      --# expr-start 2628 '<to_symbol at ast.py:3177>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3177
Sym := String_To_Symbol (Self.Unit.Context, Item_120); 
--# expr-done 2628
         declare
            Item_To_Append : constant Symbol_Type := Sym;
         begin
            Symbol_Type_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Symbol_Type_Array
           (Items_Count => Natural (Symbol_Type_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Symbol_Type_Vectors.Get
              (Map_Result_Vec,
               I + Symbol_Type_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Symbol_Type_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 2629
      declare
         Map_Result_1_Vec : String_Type_Vectors.Vector;
      begin
      declare
         Collection : constant Symbol_Type_Array_Access := Map_Result;
      begin
         for Item_119 of
               Collection.Items
         loop
   --# scope-start
               --# bind t Item_119
      --# expr-start 2630 '<FieldAccess for image at ast.py:3177>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3177
      if Item_119 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Image (Self => Item_119);
--# expr-done 2630
         declare
            Item_To_Append : constant String_Type := Fld_1;
         begin
               Inc_Ref (Item_To_Append);
            String_Type_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2100;
         end loop;
      end;
         Map_Result_1 := Create_String_Type_Array
           (Items_Count => Natural (String_Type_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := String_Type_Vectors.Get
              (Map_Result_1_Vec,
               I + String_Type_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         String_Type_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 2631
Join_Result := Join_Strings (Str, Map_Result_1); 
--# expr-done 2632
Let_Result := Join_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 2633
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_203;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_203;
               Finalizer_Scope_2100;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Canonical_Fully_Qualified_Name_Impl;
--# end
--# property-start BasicDecl.canonical_fully_qualified_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3181
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Canonical_Fully_Qualified_Name
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return String_Type
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : String_Type;
            procedure Finalizer_Scope_204 with Inline_Always;
      Fld : String_Type;
Let_Result : String_Type;
            procedure Finalizer_Scope_204 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_204;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2635 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2634 '<FieldAccess for canonical_fully_qualified_name_impl at ast.py:3186>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3186
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.canonical_fully_qualified_name_impl
Fld := Libadalang.Implementation.Basic_Decl_P_Canonical_Fully_Qualified_Name_Impl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2634
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 2635
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_204;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_204;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Canonical_Fully_Qualified_Name;
--# end
--# property-start BasicDecl.unique_identifying_name_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3189
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Unique_Identifying_Name_Impl
  (Node : Bare_Basic_Decl
      ; Dn : Internal_Entity_Defining_Name
         := No_Entity_Defining_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return String_Type
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind dn Dn
   Property_Result : String_Type;
            procedure Finalizer_Scope_205 with Inline_Always;
            procedure Finalizer_Scope_2101 with Inline_Always;
            procedure Finalizer_Scope_2102 with Inline_Always;
      Match_0 : Internal_Entity_Anonymous_Type_Decl;
Match_1 : Internal_Entity_Basic_Decl;
Fld : String_Type;
Fld_1 : String_Type;
Match_Prefix : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity_Anonymous_Type_Decl;
Let_Result : String_Type;
Scope_Result : String_Type;
Cast_Result_1 : Internal_Entity_Basic_Decl;
Let_Result_1 : String_Type;
Scope_Result_1 : String_Type;
Match_Result : String_Type;
Let_Result_2 : String_Type;
            procedure Finalizer_Scope_205 is
            begin
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Match_Result);
                     Dec_Ref (Let_Result_2);
            end Finalizer_Scope_205;
            procedure Finalizer_Scope_2101 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_2101;
            procedure Finalizer_Scope_2102 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2102;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2639 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 2638 '<Match at ast.py:3195>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3195
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Basic_Decl (Match_Prefix.Node.Kind) is
         when Ada_Anonymous_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Anonymous_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind _ Match_0
--# expr-start 2636 '<FieldAccess for custom_id_text at ast.py:3196>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3196
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.custom_id_text'
Fld := Libadalang.Implementation.Dispatcher_Ada_Node_P_Custom_Id_Text (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2636
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
Scope_Result := Let_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2101;
            Match_Result := Scope_Result; Inc_Ref (Match_Result);
         when Ada_Abstract_State_Decl .. Ada_Type_Decl | Ada_Abstract_Subp_Decl .. Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Basic_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind _ Match_1
--# expr-start 2637 '<FieldAccess for canonical_fully_qualified_name_impl at ast.py:3197>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3197
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.canonical_fully_qualified_name_impl
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Canonical_Fully_Qualified_Name_Impl (Node => Ent.Node, Include_Profile => True, Dn => Dn, E_Info => Ent.Info);
--# end
--# expr-done 2637
Let_Result_1 := Fld_1; Inc_Ref (Let_Result_1);
--# end
Scope_Result_1 := Let_Result_1;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2102;
            Match_Result := Scope_Result_1; Inc_Ref (Match_Result);
end case;
--# expr-done 2638
Let_Result_2 := Match_Result; Inc_Ref (Let_Result_2);
--# end
--# expr-done 2639
      Property_Result := Let_Result_2;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_205;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_205;
               Finalizer_Scope_2101;
               Finalizer_Scope_2102;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Unique_Identifying_Name_Impl;
--# end
--# property-start BasicDecl.unique_identifying_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3203
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Unique_Identifying_Name
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return String_Type
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : String_Type;
            procedure Finalizer_Scope_206 with Inline_Always;
      Fld : String_Type;
Let_Result : String_Type;
            procedure Finalizer_Scope_206 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_206;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2641 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2640 '<FieldAccess for unique_identifying_name_impl at ast.py:3214>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3214
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.unique_identifying_name_impl
Fld := Libadalang.Implementation.Basic_Decl_P_Unique_Identifying_Name_Impl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2640
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 2641
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_206;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_206;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Unique_Identifying_Name;
--# end
--# property-start BasicDecl.custom_id_text /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3217
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Custom_Id_Text
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return String_Type
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : String_Type;
            procedure Finalizer_Scope_207 with Inline_Always;
            procedure Finalizer_Scope_2103 with Inline_Always;
            procedure Finalizer_Scope_2104 with Inline_Always;
            procedure Finalizer_Scope_2105 with Inline_Always;
            procedure Finalizer_Scope_2106 with Inline_Always;
            procedure Finalizer_Scope_2107 with Inline_Always;
            procedure Finalizer_Scope_2108 with Inline_Always;
      Var_Expr : Internal_Entity_Base_Subp_Spec;
Var_Expr_1 : Internal_Entity_Type_Expr;
Var_Expr_2 : Internal_Entity_Defining_Name_Array_Access;
Var_Expr_3 : Internal_Entity_Type_Expr;
Var_Expr_4 : Internal_Entity_Type_Expr;
Fld : Internal_Entity_Base_Subp_Spec;
Str : String_Type;
Fld_1 : Internal_Entity_Type_Expr;
Str_1 : String_Type;
Scope_Result : String_Type;
Result_Var : String_Type;
Fld_2 : Internal_Entity_Defining_Name_Array_Access;
Str_2 : String_Type;
Fld_3 : Internal_Entity_Base_Formal_Param_Decl;
Fld_4 : Internal_Entity_Type_Expr;
Fld_5 : String_Type;
Map_Result : String_Type_Array_Access;
Join_Result : String_Type;
Scope_Result_1 : String_Type;
Result_Var_1 : String_Type;
Fld_6 : Internal_Entity_Type_Expr;
Str_3 : String_Type;
Scope_Result_2 : String_Type;
Result_Var_2 : String_Type;
Concat_Result : String_Type;
Concat_Result_1 : String_Type;
Concat_Result_2 : String_Type;
Fld_7 : Internal_Entity_Type_Expr;
Str_4 : String_Type;
Fld_8 : String_Type;
Concat_Result_3 : String_Type;
Scope_Result_3 : String_Type;
Result_Var_3 : String_Type;
Concat_Result_4 : String_Type;
Str_5 : String_Type;
Concat_Result_5 : String_Type;
Scope_Result_4 : String_Type;
Str_6 : String_Type;
Result_Var_4 : String_Type;
Let_Result : String_Type;
            procedure Finalizer_Scope_207 is
            begin
                     Dec_Ref (Scope_Result_4);
                     Dec_Ref (Str_6);
                     Dec_Ref (Result_Var_4);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_207;
            procedure Finalizer_Scope_2103 is
            begin
                     Dec_Ref (Str);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Result_Var_2);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Concat_Result_1);
                     Dec_Ref (Concat_Result_2);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Result_Var_3);
                     Dec_Ref (Concat_Result_4);
                     Dec_Ref (Str_5);
                     Dec_Ref (Concat_Result_5);
            end Finalizer_Scope_2103;
            procedure Finalizer_Scope_2104 is
            begin
                     Dec_Ref (Str_1);
            end Finalizer_Scope_2104;
            procedure Finalizer_Scope_2105 is
            begin
                     Dec_Ref (Var_Expr_2);
                     Dec_Ref (Str_2);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Join_Result);
            end Finalizer_Scope_2105;
            procedure Finalizer_Scope_2106 is
            begin
                     Dec_Ref (Fld_5);
            end Finalizer_Scope_2106;
            procedure Finalizer_Scope_2107 is
            begin
                     Dec_Ref (Str_3);
            end Finalizer_Scope_2107;
            procedure Finalizer_Scope_2108 is
            begin
                     Dec_Ref (Str_4);
                     Dec_Ref (Fld_8);
                     Dec_Ref (Concat_Result_3);
            end Finalizer_Scope_2108;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2666 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2642 '<FieldAccess for subp_spec_or_null at ast.py:3218>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3218
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2642
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Subp_Spec then
   --# scope-start
   --# bind ss Var_Expr
--# expr-start 2664 '<Concat at ast.py:3221>' Concat_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3221
--# expr-start 2662 '<Concat at ast.py:3221>' Concat_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3221
--# expr-start 2657 '<Concat at ast.py:3221>' Concat_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3221
--# expr-start 2643 '<String at ast.py:3221>' Str /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3221
Str := Create_String ("("); 
--# expr-done 2643
--# expr-start 2656 '<Concat at ast.py:3222>' Concat_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3222
--# expr-start 2644 '<FieldAccess for returns at ast.py:3222>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3222
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.returns'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Returns (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 2644
Var_Expr_1 := Fld_1;
if Var_Expr_1 /= No_Entity_Type_Expr then
   --# scope-start
   --# bind _ Var_Expr_1
--# expr-start 2645 '<String at ast.py:3222>' Str_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3222
Str_1 := Create_String ("("); 
--# expr-done 2645
Scope_Result := Str_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2104;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_String;
end if;
      Inc_Ref (Result_Var);
--# expr-start 2655 '<Concat at ast.py:3223>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3223
--# expr-start 2646 '<FieldAccess for unpacked_formal_params at ast.py:3223>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3223
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld_2 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 2646
Var_Expr_2 := Fld_2;
      Inc_Ref (Var_Expr_2);
if Var_Expr_2 /= No_Internal_Entity_Defining_Name_Array_Type then
   --# scope-start
   --# bind ufp Var_Expr_2
--# expr-start 2652 '<Join at ast.py:3224>' Join_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3224
--# expr-start 2647 '<String at ast.py:3224>' Str_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3224
Str_2 := Create_String (", "); 
--# expr-done 2647
--# expr-start 2651 '<Map at ast.py:3225>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3225
      declare
         Map_Result_Vec : String_Type_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Var_Expr_2;
      begin
         for Item_121 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_121
      --# expr-start 2650 '<FieldAccess for custom_id_text at ast.py:3227>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3227
--# expr-start 2649 '<FieldAccess for type_expression at ast.py:3227>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3227
--# expr-start 2648 '<FieldAccess for formal_decl at ast.py:3227>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3227
      if Item_121.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_3 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Item_121.Node, E_Info => Item_121.Info);
--# end
--# expr-done 2648
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.type_expression'
Fld_4 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Type_Expression (Node => Fld_3.Node, E_Info => Fld_3.Info);
--# end
--# expr-done 2649
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.custom_id_text'
Fld_5 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Custom_Id_Text (Node => Fld_4.Node, E_Info => Fld_4.Info);
--# end
--# expr-done 2650
         declare
            Item_To_Append : constant String_Type := Fld_5;
         begin
               Inc_Ref (Item_To_Append);
            String_Type_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2106;
         end loop;
      end;
         Map_Result := Create_String_Type_Array
           (Items_Count => Natural (String_Type_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := String_Type_Vectors.Get
              (Map_Result_Vec,
               I + String_Type_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         String_Type_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 2651
Join_Result := Join_Strings (Str_2, Map_Result); 
--# expr-done 2652
Scope_Result_1 := Join_Result;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2105;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := Empty_String;
end if;
      Inc_Ref (Result_Var_1);
--# expr-start 2653 '<FieldAccess for returns at ast.py:3230>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3230
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.returns'
Fld_6 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Returns (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 2653
Var_Expr_3 := Fld_6;
if Var_Expr_3 /= No_Entity_Type_Expr then
   --# scope-start
   --# bind _ Var_Expr_3
--# expr-start 2654 '<String at ast.py:3230>' Str_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3230
Str_3 := Create_String (")"); 
--# expr-done 2654
Scope_Result_2 := Str_3;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2107;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := Empty_String;
end if;
      Inc_Ref (Result_Var_2);
Concat_Result := Concat_String (Result_Var_1, Result_Var_2); 
--# expr-done 2655
Concat_Result_1 := Concat_String (Result_Var, Concat_Result); 
--# expr-done 2656
Concat_Result_2 := Concat_String (Str, Concat_Result_1); 
--# expr-done 2657
--# expr-start 2658 '<FieldAccess for returns at ast.py:3232>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3232
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.returns'
Fld_7 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Returns (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 2658
Var_Expr_4 := Fld_7;
if Var_Expr_4 /= No_Entity_Type_Expr then
   --# scope-start
   --# bind r Var_Expr_4
--# expr-start 2661 '<Concat at ast.py:3233>' Concat_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3233
--# expr-start 2659 '<String at ast.py:3233>' Str_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3233
Str_4 := Create_String (" -> "); 
--# expr-done 2659
--# expr-start 2660 '<FieldAccess for custom_id_text at ast.py:3233>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3233
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.custom_id_text'
Fld_8 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Custom_Id_Text (Node => Var_Expr_4.Node, E_Info => Var_Expr_4.Info);
--# end
--# expr-done 2660
Concat_Result_3 := Concat_String (Str_4, Fld_8); 
--# expr-done 2661
Scope_Result_3 := Concat_Result_3;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_2108;
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := Empty_String;
end if;
      Inc_Ref (Result_Var_3);
Concat_Result_4 := Concat_String (Concat_Result_2, Result_Var_3); 
--# expr-done 2662
--# expr-start 2663 '<String at ast.py:3234>' Str_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3234
Str_5 := Create_String (")"); 
--# expr-done 2663
Concat_Result_5 := Concat_String (Concat_Result_4, Str_5); 
--# expr-done 2664
Scope_Result_4 := Concat_Result_5;
   Inc_Ref (Scope_Result_4);
   --# end
      Finalizer_Scope_2103;
   Result_Var_4 := Scope_Result_4;
else
   --# expr-start 2665 '<String at ast.py:3235>' Str_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3235
Str_6 := Create_String (""); 
--# expr-done 2665
   Result_Var_4 := Str_6;
end if;
      Inc_Ref (Result_Var_4);
Let_Result := Result_Var_4; Inc_Ref (Let_Result);
--# end
--# expr-done 2666
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_207;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_207;
               Finalizer_Scope_2103;
               Finalizer_Scope_2104;
               Finalizer_Scope_2105;
               Finalizer_Scope_2106;
               Finalizer_Scope_2107;
               Finalizer_Scope_2108;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Custom_Id_Text;
--# end
--# property-start BasicDecl.does_aspects_make_preelaborable /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3239
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Does_Aspects_Make_Preelaborable
  (Node : Bare_Basic_Decl
      ; From_Body : Boolean
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind from_body From_Body
   Property_Result : Boolean;
      Bound_Imprecise_Fallback : Boolean;
Sym : Symbol_Type;
Fld : Boolean;
Sym_1 : Symbol_Type;
Fld_1 : Boolean;
If_Result : Boolean;
Sym_2 : Symbol_Type;
Fld_2 : Boolean;
If_Result_1 : Boolean;
Not_Val : Boolean;
Sym_3 : Symbol_Type;
Fld_3 : Boolean;
Sym_4 : Symbol_Type;
Fld_4 : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Boolean;
If_Result_4 : Boolean;
Dyn_Var_Bind_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2679 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2678 '<bind at ast.py:3248>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3248
Bound_Imprecise_Fallback := False; 
--# expr-start 2668 '<FieldAccess for has_aspect at ast.py:3250>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3250
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2667 '<Symbol Pure at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Pure); 
--# expr-done 2667
--# property-call-start BasicDecl.has_aspect
Fld := Libadalang.Implementation.Basic_Decl_P_Has_Aspect (Node => Ent.Node, Name => Sym, Imprecise_Fallback => Bound_Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2668
if Fld then
   If_Result := True;
else
   --# expr-start 2670 '<FieldAccess for has_aspect at ast.py:3251>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3251
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2669 '<Symbol Preelaborate at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Preelaborate); 
--# expr-done 2669
--# property-call-start BasicDecl.has_aspect
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Has_Aspect (Node => Ent.Node, Name => Sym_1, Imprecise_Fallback => Bound_Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2670
   If_Result := Fld_1;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 2672 '<FieldAccess for has_aspect at ast.py:3252>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3252
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2671 '<Symbol Shared_Passive at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Shared_Passive); 
--# expr-done 2671
--# property-call-start BasicDecl.has_aspect
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Has_Aspect (Node => Ent.Node, Name => Sym_2, Imprecise_Fallback => Bound_Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2672
   If_Result_1 := Fld_2;
end if;
if If_Result_1 then
   If_Result_4 := True;
else
--# expr-start 2673 '<Not at ast.py:3253>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3253
Not_Val := not (From_Body); 
--# expr-done 2673
if Not_Val then
--# expr-start 2675 '<FieldAccess for has_aspect at ast.py:3255>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3255
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2674 '<Symbol Remote_Types at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Remote_Types); 
--# expr-done 2674
--# property-call-start BasicDecl.has_aspect
Fld_3 := Libadalang.Implementation.Basic_Decl_P_Has_Aspect (Node => Ent.Node, Name => Sym_3, Imprecise_Fallback => Bound_Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2675
if Fld_3 then
   If_Result_2 := True;
else
   --# expr-start 2677 '<FieldAccess for has_aspect at ast.py:3256>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3256
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2676 '<Symbol Remote_Call_Interface at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Remote_Call_Interface); 
--# expr-done 2676
--# property-call-start BasicDecl.has_aspect
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Has_Aspect (Node => Ent.Node, Name => Sym_4, Imprecise_Fallback => Bound_Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2677
   If_Result_2 := Fld_4;
end if;
   If_Result_3 := If_Result_2;
else
   If_Result_3 := False;
end if;
   If_Result_4 := If_Result_3;
end if;
Dyn_Var_Bind_Result := If_Result_4; 
--# expr-done 2678
Let_Result := Dyn_Var_Bind_Result; 
--# end
--# expr-done 2679
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Does_Aspects_Make_Preelaborable;
--# end
--# property-start BasicDecl.wrap_public_reference /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3261
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Wrap_Public_Reference
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Basic_Decl;
      Var_Expr : Internal_Entity_Generic_Instantiation;
Fld : Internal_Entity_Generic_Instantiation;
Scope_Result : Internal_Entity_Generic_Instantiation;
Cast_Result : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Is_A : Boolean;
If_Result : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2683 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2682 '<If at ast.py:3283>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3283
--# expr-start 2680 '<IsA GenericSubpInternal, GenericPackageInternal at ast.py:3284>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3284
Is_A := Ent.Node /= null 
and then Ent.Node.Kind in Ada_Generic_Subp_Internal_Range | Ada_Generic_Package_Internal_Range; 
--# expr-done 2680
if Is_A then
--# expr-start 2681 '<FieldAccess for get_instantiation at ast.py:3285>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3285
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.get_instantiation
Fld := Libadalang.Implementation.Basic_Decl_P_Get_Instantiation (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2681
Var_Expr := Fld;
if Var_Expr /= No_Entity_Generic_Instantiation then
   --# scope-start
   --# bind e Var_Expr
Scope_Result := Var_Expr;
   --# end
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Scope_Result.Node,
         Info => Scope_Result.Info);
   Result_Var := Cast_Result;
else
   Result_Var := Ent;
end if;
   If_Result := Result_Var;
else
   If_Result := Ent;
end if;
--# expr-done 2682
Let_Result := If_Result; 
--# end
--# expr-done 2683
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Wrap_Public_Reference;
--# end
--# property-start '[dispatcher]BasicDecl.is_constant_object' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Decl_P_Is_Constant_Object
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Decl (Self.Kind) is
               when Ada_Abstract_State_Decl .. Ada_Anonymous_Expr_Decl | Ada_Generic_Formal_Obj_Decl .. Ada_Generic_Formal_Type_Decl | Ada_Synthetic_Formal_Param_Decl .. Ada_Concrete_Formal_Subp_Decl | Ada_Entry_Decl | Ada_Generic_Subp_Internal .. Ada_Synthetic_Subp_Decl | Ada_Package_Body_Stub .. Ada_Exception_Decl | Ada_Generic_Package_Decl .. Ada_Named_Stmt_Decl | Ada_Package_Renaming_Decl .. Ada_Single_Task_Decl =>
                  --# property-call-start BasicDecl.is_constant_object
                  Property_Result := Basic_Decl_P_Is_Constant_Object
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Component_Decl =>
                  --# property-call-start ComponentDecl.is_constant_object
                  Property_Result := Component_Decl_P_Is_Constant_Object
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Discriminant_Spec =>
                  --# property-call-start DiscriminantSpec.is_constant_object
                  Property_Result := Discriminant_Spec_P_Is_Constant_Object
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Param_Spec =>
                  --# property-call-start ParamSpec.is_constant_object
                  Property_Result := Param_Spec_P_Is_Constant_Object
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Subp_Decl =>
                  --# property-call-start SubpDecl.is_constant_object
                  Property_Result := Subp_Decl_P_Is_Constant_Object
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Enum_Literal_Decl .. Ada_Synthetic_Char_Enum_Lit =>
                  --# property-call-start EnumLiteralDecl.is_constant_object
                  Property_Result := Enum_Literal_Decl_P_Is_Constant_Object
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Expr_Function .. Ada_Subp_Renaming_Decl =>
                  --# property-call-start BaseSubpBody.is_constant_object
                  Property_Result := Base_Subp_Body_P_Is_Constant_Object
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Exception_Handler =>
                  --# property-call-start ExceptionHandler.is_constant_object
                  Property_Result := Exception_Handler_P_Is_Constant_Object
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_For_Loop_Var_Decl =>
                  --# property-call-start ForLoopVarDecl.is_constant_object
                  Property_Result := For_Loop_Var_Decl_P_Is_Constant_Object
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Number_Decl =>
                  --# property-call-start NumberDecl.is_constant_object
                  Property_Result := Number_Decl_P_Is_Constant_Object
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Object_Decl .. Ada_No_Type_Object_Renaming_Decl =>
                  --# property-call-start ObjectDecl.is_constant_object
                  Property_Result := Object_Decl_P_Is_Constant_Object
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Decl_P_Is_Constant_Object;
--# end
--# property-start BasicDecl.child_decl_initial_env_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1900
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Child_Decl_Initial_Env_Name
  (Node : Bare_Basic_Decl
      ; Private_Part : Boolean
         := False
  )
   return Symbol_Type
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind private_part Private_Part
   Property_Result : Symbol_Type;
            procedure Finalizer_Scope_128 with Inline_Always;
            procedure Finalizer_Scope_2006 with Inline_Always;
      Block_Var : Internal_Entity_Defining_Name;
Block_Var_1 : Internal_Entity_Dotted_Name;
Var_Expr : Internal_Entity_Dotted_Name;
As_Entity : Internal_Entity_Basic_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Internal_Entity_Name;
Cast_Result : Internal_Entity_Dotted_Name;
Fld_2 : String_Type;
Sym : Symbol_Type;
Sym_1 : Symbol_Type;
Is_Equal : Boolean;
Fld_3 : Boolean;
Fld_4 : Internal_Entity_Name;
Fld_5 : String_Type;
Str : String_Type;
Concat_Result : String_Type;
Sym_2 : Symbol_Type;
Fld_6 : Internal_Entity_Name;
Fld_7 : String_Type;
Sym_3 : Symbol_Type;
If_Result_1 : Symbol_Type;
Scope_Result : Symbol_Type;
Sym_4 : Symbol_Type;
Result_Var : Symbol_Type;
If_Result_2 : Symbol_Type;
If_Result_3 : Symbol_Type;
Let_Result : Symbol_Type;
            procedure Finalizer_Scope_128 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_128;
            procedure Finalizer_Scope_2006 is
            begin
                     Dec_Ref (Fld_5);
                     Dec_Ref (Str);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Fld_7);
            end Finalizer_Scope_2006;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2704 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2685 '<FieldAccess for defining_name at ast.py:1908>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1908
--# expr-start 2684 '<as_bare_entity at ast.py:1908>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1908
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Decl;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 2684
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 2685
Block_Var := Fld; 
--# bind defining_name Block_Var
--# expr-start 2687 '<Cast to DottedName at ast.py:1909>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1909
--# expr-start 2686 '<FieldAccess for name at ast.py:1909>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1909
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Block_Var.Node.Defining_Name_F_Name, Info => Block_Var.Info);
--# expr-done 2686
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Dotted_Name_Range
   then
      Cast_Result := Create_Internal_Entity_Dotted_Name
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         Cast_Result := No_Entity_Dotted_Name;
   end if;
--# expr-done 2687
Block_Var_1 := Cast_Result; 
--# bind child_name Block_Var_1
--# expr-start 2703 '<Cond at ast.py:1910>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1910
--# expr-start 2691 '<Eq at ast.py:1913>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1913
--# expr-start 2689 '<to_symbol at ast.py:1913>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1913
--# expr-start 2688 '<FieldAccess for text at ast.py:1913>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1913
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Text (Node => Block_Var.Node);
--# expr-done 2688
Sym := String_To_Symbol (Self.Unit.Context, Fld_2); 
--# expr-done 2689
--# expr-start 2690 '<Symbol standard at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Standard_323); 
--# expr-done 2690
Is_Equal := Sym = Sym_1; 
--# expr-done 2691
if Is_Equal then
   If_Result_3 := null;
else
--# expr-start 2692 '<FieldAccess for is_library_item at ast.py:1916>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1916
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_library_item'
Fld_3 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Library_Item (Node => Self);
--# end
--# expr-done 2692
if Fld_3 then
Var_Expr := Block_Var_1;
if Var_Expr /= No_Entity_Dotted_Name then
   --# scope-start
   --# bind n Var_Expr
--# expr-start 2701 '<If at ast.py:1920>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1920
if Private_Part then
   --# expr-start 2697 '<to_symbol at ast.py:1922>' Sym_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1922
--# expr-start 2696 '<Concat at ast.py:1922>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1922
--# expr-start 2694 '<FieldAccess for text at ast.py:1922>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1922
--# expr-start 2693 '<FieldAccess for prefix at ast.py:1922>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1922
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Name (Node => Var_Expr.Node.Dotted_Name_F_Prefix, Info => Var_Expr.Info);
--# expr-done 2693
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Text (Node => Fld_4.Node);
--# expr-done 2694
--# expr-start 2695 '<String at ast.py:1922>' Str /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1922
Str := Create_String (".__privatepart"); 
--# expr-done 2695
Concat_Result := Concat_String (Fld_5, Str); 
--# expr-done 2696
Sym_2 := String_To_Symbol (Self.Unit.Context, Concat_Result); 
--# expr-done 2697
   If_Result_1 := Sym_2;
else
   --# expr-start 2700 '<to_symbol at ast.py:1923>' Sym_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1923
--# expr-start 2699 '<FieldAccess for text at ast.py:1923>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1923
--# expr-start 2698 '<FieldAccess for prefix at ast.py:1923>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:1923
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Name (Node => Var_Expr.Node.Dotted_Name_F_Prefix, Info => Var_Expr.Info);
--# expr-done 2698
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Text (Node => Fld_6.Node);
--# expr-done 2699
Sym_3 := String_To_Symbol (Self.Unit.Context, Fld_7); 
--# expr-done 2700
   If_Result_1 := Sym_3;
end if;
--# expr-done 2701
Scope_Result := If_Result_1;
   --# end
      Finalizer_Scope_2006;
   Result_Var := Scope_Result;
else
   --# expr-start 2702 '<Symbol standard at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Standard_323); 
--# expr-done 2702
   Result_Var := Sym_4;
end if;
   If_Result_2 := Result_Var;
else
   If_Result_2 := null;
end if;
   If_Result_3 := If_Result_2;
end if;
--# expr-done 2703
Let_Result := If_Result_3; 
--# end
--# expr-done 2704
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_128;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_128;
               Finalizer_Scope_2006;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Child_Decl_Initial_Env_Name;
--# end
--# property-start BasicDecl.previous_part_for_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2037
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Previous_Part_For_Decl
  (Node : Bare_Basic_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Match_0 : Internal_Entity_Base_Type_Decl;
Match_1 : Internal_Entity_Body_Node;
Match_2 : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Basic_Decl;
Match_Prefix : Internal_Entity_Basic_Decl;
Cast_Result_1 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Cast_Result_2 : Internal_Entity_Body_Node;
Let_Result_1 : Internal_Entity_Basic_Decl;
Scope_Result_1 : Internal_Entity_Basic_Decl;
Cast_Result_3 : Internal_Entity_Basic_Decl;
Let_Result_2 : Internal_Entity_Basic_Decl;
Scope_Result_2 : Internal_Entity_Basic_Decl;
Match_Result : Internal_Entity_Basic_Decl;
Let_Result_3 : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2709 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 2708 '<Match at ast.py:2046>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2046
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Basic_Decl (Match_Prefix.Node.Kind) is
         when Ada_Discrete_Base_Subtype_Decl .. Ada_Synth_Anonymous_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Base_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_1; 
--# bind btd Match_0
--# expr-start 2706 '<Cast to BasicDecl at ast.py:2048>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2048
--# expr-start 2705 '<FieldAccess for previous_part at ast.py:2048>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2048
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.previous_part'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Previous_Part (Node => Match_0.Node, Go_To_Incomplete => True, E_Info => Match_0.Info);
--# end
--# expr-done 2705
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
--# expr-done 2706
Let_Result := Cast_Result; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Expr_Function .. Ada_Task_Body =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Body_Node
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_2; 
--# bind bd Match_1
--# expr-start 2707 '<FieldAccess for previous_part_internal at ast.py:2049>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2049
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.previous_part_internal
Fld_1 := Libadalang.Implementation.Body_Node_P_Previous_Part_Internal (Node => Match_1.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Match_1.Info);
--# end
--# expr-done 2707
Let_Result_1 := Fld_1; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abstract_State_Decl .. Ada_Package_Decl | Ada_Abstract_Subp_Decl .. Ada_Synthetic_Subp_Decl | Ada_Entry_Index_Spec .. Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Basic_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_3; 
--# bind _ Match_2
Let_Result_2 := No_Entity_Basic_Decl; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 2708
Let_Result_3 := Match_Result; 
--# end
--# expr-done 2709
      Property_Result := Let_Result_3;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Previous_Part_For_Decl;
--# end
--# property-start BasicDecl.is_static_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2107
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Is_Static_Decl
  (Node : Bare_Basic_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2710 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := False; 
--# end
--# expr-done 2710
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Is_Static_Decl;
--# end
--# property-start BasicDecl.is_library_item /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2148
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Is_Library_Item
  (Node : Bare_Basic_Decl
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2713 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2712 '<IsA LibraryItem at ast.py:2152>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2152
--# expr-start 2711 '<parent at ast.py:2152>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2152
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 2711
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range; 
--# expr-done 2712
Let_Result := Is_A; 
--# end
--# expr-done 2713
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Is_Library_Item;
--# end
--# property-start BasicDecl.declarative_parts /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2164
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Declarative_Parts
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Declarative_Part_Array_Access
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_145 with Inline_Always;
      Let_Result : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_145 is
            begin
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_145;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2714 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := No_Internal_Entity_Declarative_Part_Array_Type; Inc_Ref (Let_Result);
--# end
--# expr-done 2714
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_145;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_145;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Declarative_Parts;
--# end
--# property-start BasicDecl.get_aspect_assoc /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2184
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Get_Aspect_Assoc
  (Node : Bare_Basic_Decl
      ; Name : Symbol_Type
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Aspect_Assoc
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind name Name
   Property_Result : Internal_Entity_Aspect_Assoc;
            procedure Finalizer_Scope_2028 with Inline_Always;
      Var_Expr : Internal_Entity_Aspect_Spec;
Fld : Internal_Entity_Aspect_Spec;
Fld_1 : Internal_Entity_Aspect_Assoc_List;
Scope_Result : Internal_Entity_Aspect_Assoc_List;
Result_Var : Internal_Entity_Aspect_Assoc_List;
Fld_2 : Bare_Aspect_Assoc_List;
Fld_3 : Internal_Entity_Info;
As_Entity : Internal_Entity_Aspect_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Aspect_Assoc;
Fld_4 : Internal_Entity_Name;
Fld_5 : String_Type;
Fld_6 : String_Type;
Is_Equal : Boolean;
Item_97 : Internal_Entity_Aspect_Assoc;
Bare_Item_97 : Bare_Aspect_Assoc;
Find_Result : Internal_Entity_Aspect_Assoc;
Let_Result : Internal_Entity_Aspect_Assoc;
            procedure Finalizer_Scope_2028 is
            begin
                     Dec_Ref (Fld_5);
                     Dec_Ref (Fld_6);
            end Finalizer_Scope_2028;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2722 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2721 '<Find at ast.py:2188>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2188
--# expr-start 2715 '<FieldAccess for aspects at ast.py:2188>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2188
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Aspect_Spec (Node => Implementation.Basic_Decl_F_Aspects (Ent.Node), Info => Ent.Info);
--# expr-done 2715
Var_Expr := Fld;
if Var_Expr /= No_Entity_Aspect_Spec then
   --# scope-start
--# expr-start 2716 '<FieldAccess for aspect_assocs at ast.py:2188>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2188
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Aspect_Assoc_List (Node => Var_Expr.Node.Aspect_Spec_F_Aspect_Assocs, Info => Var_Expr.Info);
--# expr-done 2716
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Aspect_Assoc_List;
end if;
Fld_2 := Result_Var.Node;
      Find_Result := No_Entity_Aspect_Assoc;
      if Fld_2 = null then
      Find_Result := No_Entity_Aspect_Assoc;
      else
      declare
         Collection : constant Bare_Aspect_Assoc_List := Fld_2;
      begin
         for Untyped_Item_97 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_97 := Untyped_Item_97; 
Is_Null := Bare_Item_97 = null; 
if Is_Null then
   If_Result := No_Entity_Aspect_Assoc;
else
Fld_3 := Result_Var.Info;
As_Entity := (Info => Fld_3, Node => Bare_Item_97); 
   If_Result := As_Entity;
end if;
                  Item_97 := If_Result; 
   --# scope-start
               --# bind asp Item_97
      --# expr-start 2720 '<Eq at ast.py:2189>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2189
--# expr-start 2718 '<FieldAccess for aspect_name at ast.py:2189>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2189
      if Item_97.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2717 '<FieldAccess for id at ast.py:2189>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2189
      if Item_97.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Name (Node => Item_97.Node.Aspect_Assoc_F_Id, Info => Item_97.Info);
--# expr-done 2717
--# property-call-start AspectAssoc.aspect_name
Fld_5 := Libadalang.Implementation.Aspect_Assoc_P_Aspect_Name (Node => Item_97.Node, N => Fld_4);
--# end
--# expr-done 2718
--# expr-start 2719 '<FieldAccess for image at ast.py:2189>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2189
      if Name = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Image (Self => Name);
--# expr-done 2719
Is_Equal := Equivalent (Fld_5, Fld_6); 
--# expr-done 2720
      if Is_Equal then
         Find_Result := Item_97;
      end if;
   --# end
      Finalizer_Scope_2028;
               exit when Is_Equal;
         end loop;
      end;
      end if;
--# expr-done 2721
Let_Result := Find_Result; 
--# end
--# expr-done 2722
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_2028;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Get_Aspect_Assoc;
--# end
--# property-start BasicDecl.is_imported /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2272
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Is_Imported
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Defining_Name;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2725 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2723 '<FieldAccess for defining_name_or_raise at ast.py:2276>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2276
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name_or_raise
Fld := Libadalang.Implementation.Basic_Decl_P_Defining_Name_Or_Raise (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2723
Var_Expr := Fld;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 2724 '<FieldAccess for is_imported at ast.py:2276>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2276
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.is_imported
Fld_1 := Libadalang.Implementation.Defining_Name_P_Is_Imported (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 2724
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 2725
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Is_Imported;
--# end
--# property-start BasicDecl.is_in_private_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2372
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Is_In_Private_Part
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Fld_1 : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2728 '<IsA PrivatePart at ast.py:2372>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2372
--# expr-start 2727 '<parent at ast.py:2372>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2372
--# expr-start 2726 '<parent at ast.py:2372>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2372
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 2726
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Parent; 
--# expr-done 2727
Is_A := Fld_1 /= null 
and then Fld_1.Kind in Ada_Private_Part_Range; 
--# expr-done 2728
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Is_In_Private_Part;
--# end
--# property-start BasicDecl.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2624
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Defining_Env
  (Node : Bare_Basic_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := Empty_Env;
         Inc_Ref (Property_Result);
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Defining_Env;
--# end
--# property-start BasicDecl.array_ndims /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2660
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Array_Ndims
  (Node : Bare_Basic_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Integer
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Integer;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Integer;
Let_Result : Integer;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2731 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2730 '<FieldAccess for array_ndims at ast.py:2661>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2661
--# expr-start 2729 '<FieldAccess for expr_type at ast.py:2661>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2661
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.expr_type'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Expr_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 2729
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.array_ndims'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Array_Ndims (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 2730
Let_Result := Fld_1; 
--# end
--# expr-done 2731
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Array_Ndims;
--# end
--# property-start BasicDecl.expr_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2668
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Expr_Type
  (Node : Bare_Basic_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Type_Expr;
Fld : Internal_Entity_Type_Expr;
Fld_1 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2734 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2732 '<FieldAccess for type_expression at ast.py:2687>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2687
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.type_expression'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Type_Expression (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2732
Var_Expr := Fld;
if Var_Expr /= No_Entity_Type_Expr then
   --# scope-start
   --# bind te Var_Expr
--# expr-start 2733 '<FieldAccess for designated_type at ast.py:2687>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2687
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 2733
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 2734
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Expr_Type;
--# end
--# property-start BasicDecl.type_expression /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2689
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Type_Expression
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Expr
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Type_Expr;
      Node_For_Entity : Bare_Type_Expr;
As_Entity : Internal_Entity_Type_Expr;
Is_Null : Boolean;
If_Result : Internal_Entity_Type_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2736 '<as_entity at ast.py:2690>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2690
--# expr-start 2735 '<No TypeExpr at ast.py:2690>' Node_For_Entity /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2690
Node_For_Entity := No_Bare_Ada_Node;
--# expr-done 2735
Is_Null := Node_For_Entity = null; 
if Is_Null then
   If_Result := No_Entity_Type_Expr;
else
As_Entity := (Info => E_Info, Node => Node_For_Entity); 
   If_Result := As_Entity;
end if;
--# expr-done 2736
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Type_Expression;
--# end
--# property-start BasicDecl.constrain_prefix /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2851
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Constrain_Prefix
  (Node : Bare_Basic_Decl
      ; Prefix : Bare_Expr
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind prefix Prefix
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_186 with Inline_Always;
      True_Rel : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_186 is
            begin
                     Dec_Ref (True_Rel);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_186;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2737 '<Block at ???>' Let_Result None
--# scope-start
True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:2875") else null)); 
Let_Result := True_Rel; Inc_Ref (Let_Result);
--# end
--# expr-done 2737
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_186;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_186;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Constrain_Prefix;
--# end
--# property-start BasicDecl.next_part_for_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2935
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Next_Part_For_Decl
  (Node : Bare_Basic_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2739 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2738 '<FieldAccess for basic_decl_next_part_for_decl at ast.py:2944>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2944
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.basic_decl_next_part_for_decl
Fld := Libadalang.Implementation.Basic_Decl_P_Basic_Decl_Next_Part_For_Decl (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2738
Let_Result := Fld; 
--# end
--# expr-done 2739
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Next_Part_For_Decl;
--# end
--# property-start BasicDecl.next_part_for_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2965
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Next_Part_For_Name
  (Node : Bare_Basic_Decl
      ; Sym : Symbol_Type
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind sym Sym
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2741 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2740 '<FieldAccess for next_part_for_decl at ast.py:2985>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2985
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.next_part_for_decl'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Next_Part_For_Decl (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2740
Let_Result := Fld; 
--# end
--# expr-done 2741
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Next_Part_For_Name;
--# end
--# property-start BasicDecl.previous_part_for_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:2989
pragma Warnings (Off, "is not referenced");
 function Basic_Decl_P_Previous_Part_For_Name
  (Node : Bare_Basic_Decl
      ; Sym : Symbol_Type
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Basic_Decl := Bare_Basic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Decl :=
     Internal_Entity_Basic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind sym Sym
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2743 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2742 '<FieldAccess for previous_part_for_decl at ast.py:3009>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3009
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.previous_part_for_decl'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Previous_Part_For_Decl (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2742
Let_Result := Fld; 
--# end
--# expr-done 2743
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Decl_P_Previous_Part_For_Name;
--# end
 function Basic_Decl_P_Is_Constant_Object
  (Node : Bare_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is (raise Property_Error
    with "Property BasicDecl.is_constant_object not implemented on type "
    & Kind_Name (Node));
   --
   --  Primitives for Bare_Abstract_State_Decl
   --
      procedure Abstract_State_Decl_Pre_Env_Actions
        (Self            : Bare_Abstract_State_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_233 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:20325");
         Dec_Ref (Mapping.Dest_Env);
      end;
      end;
      procedure Initialize_Fields_For_Abstract_State_Decl
        (Self : Bare_Abstract_State_Decl
         ; Abstract_State_Decl_F_Name : Bare_Defining_Name
         ; Abstract_State_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Self.Abstract_State_Decl_F_Name := Abstract_State_Decl_F_Name;
            Self.Abstract_State_Decl_F_Aspects := Abstract_State_Decl_F_Aspects;
      end Initialize_Fields_For_Abstract_State_Decl;
   function Abstract_State_Decl_F_Name
     (Node : Bare_Abstract_State_Decl) return Bare_Defining_Name
   is
   begin
         return Node.Abstract_State_Decl_F_Name;
   end;
--# property-start AbstractStateDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20321
pragma Warnings (Off, "is not referenced");
 function Abstract_State_Decl_P_Defining_Names
  (Node : Bare_Abstract_State_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Abstract_State_Decl := Bare_Abstract_State_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Abstract_State_Decl :=
     Internal_Entity_Abstract_State_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1413 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1413 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1413;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2744 '<FieldAccess for name at ast.py:20321>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20321
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Abstract_State_Decl_F_Name, Info => Ent.Info);
--# expr-done 2744
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1413;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1413;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Abstract_State_Decl_P_Defining_Names;
--# end
--# property-start AbstractStateDecl.type_expression /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20322
pragma Warnings (Off, "is not referenced");
 function Abstract_State_Decl_P_Type_Expression
  (Node : Bare_Abstract_State_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Expr
is
   Self : Bare_Abstract_State_Decl := Bare_Abstract_State_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Type_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := No_Entity_Type_Expr;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Abstract_State_Decl_P_Type_Expression;
--# end
--# property-start AbstractStateDecl._env_mappings_233 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20325
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_233
  (Node : Bare_Abstract_State_Decl
  )
   return Internal_Env_Assoc
is
   Self : Bare_Abstract_State_Decl := Bare_Abstract_State_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1650 with Inline_Always;
      Fld : Bare_Defining_Name;
Fld_1 : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Abstract_State_Decl;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1650 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1650;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2748 '<New EnvAssoc at ast.py:20325>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20325
--# expr-start 2745 '<New DesignatedEnv at ast.py:20325>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20325
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 2745
--# expr-start 2747 '<FieldAccess for name_symbol at ast.py:20325>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20325
--# expr-start 2746 '<FieldAccess for name at ast.py:20325>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20325
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Abstract_State_Decl_F_Name;
--# expr-done 2746
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld);
--# end
--# expr-done 2747
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld_1, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 2748
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1650;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1650;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_233;
--# end
   --
   --  Primitives for Bare_Anonymous_Expr_Decl
   --
      procedure Initialize_Fields_For_Anonymous_Expr_Decl
        (Self : Bare_Anonymous_Expr_Decl
         ; Anonymous_Expr_Decl_F_Expr : Bare_Expr
        ) is
      begin
            Self.Anonymous_Expr_Decl_F_Expr := Anonymous_Expr_Decl_F_Expr;
      end Initialize_Fields_For_Anonymous_Expr_Decl;
   function Anonymous_Expr_Decl_F_Expr
     (Node : Bare_Anonymous_Expr_Decl) return Bare_Expr
   is
   begin
         return Node.Anonymous_Expr_Decl_F_Expr;
   end;
--# property-start AnonymousExprDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9760
pragma Warnings (Off, "is not referenced");
 function Anonymous_Expr_Decl_P_Defining_Names
  (Node : Bare_Anonymous_Expr_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Anonymous_Expr_Decl := Bare_Anonymous_Expr_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := No_Internal_Entity_Defining_Name_Array_Type;
         Inc_Ref (Property_Result);
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Anonymous_Expr_Decl_P_Defining_Names;
--# end
--# property-start AnonymousExprDecl.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9761
pragma Warnings (Off, "is not referenced");
 function Anonymous_Expr_Decl_P_Defining_Env
  (Node : Bare_Anonymous_Expr_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Anonymous_Expr_Decl := Bare_Anonymous_Expr_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Anonymous_Expr_Decl :=
     Internal_Entity_Anonymous_Expr_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_779 with Inline_Always;
      Fld : Internal_Entity_Type_Expr;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_779 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_779;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2750 '<FieldAccess for defining_env at ast.py:9761>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9761
--# expr-start 2749 '<FieldAccess for type_expression at ast.py:9761>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9761
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.type_expression'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Type_Expression (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2749
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeExpr.defining_env
Fld_1 := Libadalang.Implementation.Type_Expr_P_Defining_Env (Node => Fld.Node, Origin => Origin, Include_Ud_Indexing => Include_Ud_Indexing, E_Info => Fld.Info);
--# end
--# expr-done 2750
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_779;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_779;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Anonymous_Expr_Decl_P_Defining_Env;
--# end
--# property-start AnonymousExprDecl.get_formal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9766
pragma Warnings (Off, "is not referenced");
 function Anonymous_Expr_Decl_P_Get_Formal
  (Node : Bare_Anonymous_Expr_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Anonymous_Expr_Decl := Bare_Anonymous_Expr_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Anonymous_Expr_Decl :=
     Internal_Entity_Anonymous_Expr_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_780 with Inline_Always;
      Block_Var : Internal_Entity_Basic_Assoc;
Fld : Internal_Entity_Expr;
Fld_1 : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Assoc;
Fld_2 : Internal_Entity_Defining_Name_Array_Access;
Get_Result : Internal_Entity_Defining_Name;
Let_Result : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_780 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_780;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Anonymous_Expr_Decl_Anonymous_Expr_Decl_P_Get_Formal,
            Items    => new Mmz_Key_Array (1 ..  3))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Anonymous_Expr_Decl,
                                As_Bare_Anonymous_Expr_Decl => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Boolean,
                                      As_Boolean => Imprecise_Fallback);
               Mmz_K.Items (3) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Defining_Name;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 2756 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2753 '<Cast to BasicAssoc at ast.py:9771>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9771
--# expr-start 2752 '<parent at ast.py:9771>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9771
--# expr-start 2751 '<FieldAccess for expr at ast.py:9771>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9771
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Anonymous_Expr_Decl_F_Expr, Info => Ent.Info);
--# expr-done 2751
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Parent (Node => Fld.Node, E_Info => Fld.Info);
--# expr-done 2752
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Basic_Assoc
   then
      Cast_Result := Create_Internal_Entity_Basic_Assoc
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         Cast_Result := No_Entity_Basic_Assoc;
   end if;
--# expr-done 2753
Block_Var := Cast_Result; 
--# bind assoc Block_Var
--# expr-start 2755 '<at at ast.py:9772>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9772
--# expr-start 2754 '<FieldAccess for get_params at ast.py:9772>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9772
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicAssoc.get_params
Fld_2 := Libadalang.Implementation.Basic_Assoc_P_Get_Params (Node => Block_Var.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Block_Var.Info);
--# end
--# expr-done 2754
Get_Result := Get (Fld_2, 0, True); 
--# expr-done 2755
Let_Result := Get_Result; 
--# end
--# expr-done 2756
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_780;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Defining_Name,
                        As_Internal_Entity_Defining_Name => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_780;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Anonymous_Expr_Decl_P_Get_Formal;
--# end
--# property-start AnonymousExprDecl.type_expression /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9775
pragma Warnings (Off, "is not referenced");
 function Anonymous_Expr_Decl_P_Type_Expression
  (Node : Bare_Anonymous_Expr_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Expr
is
   Self : Bare_Anonymous_Expr_Decl := Bare_Anonymous_Expr_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Anonymous_Expr_Decl :=
     Internal_Entity_Anonymous_Expr_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Type_Expr;
      Fld : Internal_Entity_Defining_Name;
Fld_1 : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity_Type_Expr;
Let_Result : Internal_Entity_Type_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2760 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2759 '<FieldAccess for type_expression at ast.py:9781>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9781
--# expr-start 2758 '<FieldAccess for basic_decl at ast.py:9781>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9781
--# expr-start 2757 '<FieldAccess for get_formal at ast.py:9781>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9781
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AnonymousExprDecl.get_formal
Fld := Libadalang.Implementation.Anonymous_Expr_Decl_P_Get_Formal (Node => Ent.Node, Imprecise_Fallback => False, E_Info => Ent.Info);
--# end
--# expr-done 2757
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld_1 := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 2758
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.type_expression'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Type_Expression (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 2759
Let_Result := Fld_2; 
--# end
--# expr-done 2760
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Anonymous_Expr_Decl_P_Type_Expression;
--# end
--# property-start AnonymousExprDecl.is_static_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9784
pragma Warnings (Off, "is not referenced");
 function Anonymous_Expr_Decl_P_Is_Static_Decl
  (Node : Bare_Anonymous_Expr_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Anonymous_Expr_Decl := Bare_Anonymous_Expr_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Anonymous_Expr_Decl :=
     Internal_Entity_Anonymous_Expr_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Fld : Internal_Entity_Expr;
Fld_1 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2763 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2762 '<FieldAccess for is_static_expr at ast.py:9785>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9785
--# expr-start 2761 '<FieldAccess for expr at ast.py:9785>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9785
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Anonymous_Expr_Decl_F_Expr, Info => Ent.Info);
--# expr-done 2761
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_1 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld.Info);
--# end
--# expr-done 2762
Let_Result := Fld_1; 
--# end
--# expr-done 2763
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Anonymous_Expr_Decl_P_Is_Static_Decl;
--# end
   --
   --  Primitives for Bare_Base_Formal_Param_Decl
   --
--# property-start '[dispatcher]BaseFormalParamDecl.is_mandatory' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Formal_Param_Decl_P_Is_Mandatory
  (Node : Bare_Base_Formal_Param_Decl
  )
   return Boolean
is
   Self : Bare_Base_Formal_Param_Decl := Bare_Base_Formal_Param_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Formal_Param_Decl (Self.Kind) is
               when Ada_Component_Decl .. Ada_Generic_Formal_Type_Decl | Ada_Synthetic_Formal_Param_Decl =>
                  --# property-call-start BaseFormalParamDecl.is_mandatory
                  Property_Result := Base_Formal_Param_Decl_P_Is_Mandatory
                    (Self
                    );
                  --# end
               when Ada_Param_Spec =>
                  --# property-call-start ParamSpec.is_mandatory
                  Property_Result := Param_Spec_P_Is_Mandatory
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Formal_Param_Decl_P_Is_Mandatory;
--# end
--# property-start BaseFormalParamDecl.formal_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3913
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Decl_P_Formal_Type
  (Node : Bare_Base_Formal_Param_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Formal_Param_Decl := Bare_Base_Formal_Param_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Formal_Param_Decl :=
     Internal_Entity_Base_Formal_Param_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Type_Expr;
Fld : Internal_Entity_Type_Expr;
Fld_1 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2766 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2764 '<FieldAccess for type_expression at ast.py:3917>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3917
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.type_expression'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Type_Expression (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2764
Var_Expr := Fld;
if Var_Expr /= No_Entity_Type_Expr then
   --# scope-start
--# expr-start 2765 '<FieldAccess for designated_type at ast.py:3917>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3917
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 2765
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 2766
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Decl_P_Formal_Type;
--# end
--# property-start BaseFormalParamDecl.parent_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3920
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Decl_P_Parent_Decl
  (Node : Bare_Base_Formal_Param_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Base_Formal_Param_Decl := Bare_Base_Formal_Param_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Formal_Param_Decl :=
     Internal_Entity_Base_Formal_Param_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2769 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2768 '<Cast to BasicDecl at ast.py:3921>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3921
--# expr-start 2767 '<FieldAccess for semantic_parent at ast.py:3921>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3921
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.semantic_parent
Fld := Libadalang.Implementation.Ada_Node_P_Semantic_Parent (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2767
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 2768
Let_Result := Cast_Result; 
--# end
--# expr-done 2769
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Decl_P_Parent_Decl;
--# end
--# property-start BaseFormalParamDecl.get_param /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3924
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Decl_P_Get_Param
  (Node : Bare_Base_Formal_Param_Decl
      ; Part : Internal_Entity_Basic_Decl
      ; Param : Symbol_Type
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Base_Formal_Param_Decl := Bare_Base_Formal_Param_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind part Part
   --# bind param Param
   Property_Result : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_2111 with Inline_Always;
            procedure Finalizer_Scope_2112 with Inline_Always;
      Var_Expr : Internal_Entity_Basic_Decl;
Var_Expr_1 : Internal_Entity_Base_Formal_Param_Holder;
Fld : Internal_Entity_Base_Formal_Param_Holder;
Fld_1 : Internal_Entity_Defining_Name_Array_Access;
Scope_Result : Internal_Entity_Defining_Name_Array_Access;
Result_Var : Internal_Entity_Defining_Name_Array_Access;
Fld_2 : Boolean;
Find_Result : Internal_Entity_Defining_Name;
Scope_Result_1 : Internal_Entity_Defining_Name;
Result_Var_1 : Internal_Entity_Defining_Name;
Let_Result : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_2111 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_2111;
            procedure Finalizer_Scope_2112 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2112;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2774 '<Block at ???>' Let_Result None
--# scope-start
Var_Expr := Part;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind d Var_Expr
--# expr-start 2773 '<Find at ast.py:3927>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3927
--# expr-start 2770 '<FieldAccess for formal_param_holder_or_null at ast.py:3927>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3927
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.formal_param_holder_or_null
Fld := Libadalang.Implementation.Basic_Decl_P_Formal_Param_Holder_Or_Null (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 2770
Var_Expr_1 := Fld;
if Var_Expr_1 /= No_Entity_Base_Formal_Param_Holder then
   --# scope-start
--# expr-start 2771 '<FieldAccess for unpacked_formal_params at ast.py:3927>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3927
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld_1 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 2771
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2112;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Defining_Name_Array_Type;
end if;
      Inc_Ref (Result_Var);
      Find_Result := No_Entity_Defining_Name;
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Result_Var;
      begin
         for Item_122 of
               Collection.Items
         loop
   --# scope-start
               --# bind sf Item_122
      --# expr-start 2772 '<FieldAccess for name_is at ast.py:3928>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3928
      if Item_122.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_is
Fld_2 := Libadalang.Implementation.Name_P_Name_Is (Node => Item_122.Node, Sym => Param);
--# end
--# expr-done 2772
      if Fld_2 then
         Find_Result := Item_122;
      end if;
   --# end
               exit when Fld_2;
         end loop;
      end;
--# expr-done 2773
Scope_Result_1 := Find_Result;
   --# end
      Finalizer_Scope_2111;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Defining_Name;
end if;
Let_Result := Result_Var_1; 
--# end
--# expr-done 2774
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_2111;
               Finalizer_Scope_2112;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Decl_P_Get_Param;
--# end
--# property-start BaseFormalParamDecl.decl_param /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3933
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Decl_P_Decl_Param
  (Node : Bare_Base_Formal_Param_Decl
      ; Param : Internal_Entity_Defining_Name
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Base_Formal_Param_Decl := Bare_Base_Formal_Param_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Formal_Param_Decl :=
     Internal_Entity_Base_Formal_Param_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind param Param
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Defining_Name;
      Var_Expr : Internal_Entity_Defining_Name;
Var_Expr_1 : Internal_Entity_Base_Subp_Body;
Fld : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity_Base_Subp_Body;
Fld_1 : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Fld_2 : Symbol_Type;
Fld_3 : Internal_Entity_Defining_Name;
Scope_Result_1 : Internal_Entity_Defining_Name;
Result_Var_1 : Internal_Entity_Defining_Name;
Let_Result : Internal_Entity_Defining_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2780 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2779 '<FieldAccess for get_param at ast.py:3938>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3938
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2776 '<Cast to BaseSubpBody at ast.py:3939>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3939
--# expr-start 2775 '<FieldAccess for parent_decl at ast.py:3939>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3939
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamDecl.parent_decl
Fld := Libadalang.Implementation.Base_Formal_Param_Decl_P_Parent_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2775
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Base_Subp_Body
   then
      Cast_Result := Create_Internal_Entity_Base_Subp_Body
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Base_Subp_Body;
   end if;
--# expr-done 2776
Var_Expr_1 := Cast_Result;
if Var_Expr_1 /= No_Entity_Base_Subp_Body then
   --# scope-start
--# expr-start 2777 '<FieldAccess for decl_part at ast.py:3939>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3939
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.decl_part
Fld_1 := Libadalang.Implementation.Body_Node_P_Decl_Part (Node => Var_Expr_1.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 2777
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Basic_Decl;
end if;
--# expr-start 2778 '<FieldAccess for name_symbol at ast.py:3940>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3940
      if Param.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Param.Node);
--# end
--# expr-done 2778
--# property-call-start BaseFormalParamDecl.get_param
Fld_3 := Libadalang.Implementation.Base_Formal_Param_Decl_P_Get_Param (Node => Ent.Node, Part => Result_Var, Param => Fld_2);
--# end
--# expr-done 2779
Var_Expr := Fld_3;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
   --# bind e Var_Expr
Scope_Result_1 := Var_Expr;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := Param;
end if;
Let_Result := Result_Var_1; 
--# end
--# expr-done 2780
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Decl_P_Decl_Param;
--# end
--# property-start BaseFormalParamDecl.next_part_for_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3944
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Decl_P_Next_Part_For_Name
  (Node : Bare_Base_Formal_Param_Decl
      ; Sym : Symbol_Type
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Base_Formal_Param_Decl := Bare_Base_Formal_Param_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Formal_Param_Decl :=
     Internal_Entity_Base_Formal_Param_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind sym Sym
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Var_Expr : Internal_Entity_Defining_Name;
Var_Expr_1 : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity_Defining_Name;
Fld_3 : Internal_Entity_Basic_Decl;
Scope_Result_1 : Internal_Entity_Basic_Decl;
Result_Var_1 : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2785 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2783 '<FieldAccess for get_param at ast.py:3945>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3945
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2781 '<FieldAccess for parent_decl at ast.py:3946>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3946
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamDecl.parent_decl
Fld := Libadalang.Implementation.Base_Formal_Param_Decl_P_Parent_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2781
Var_Expr_1 := Fld;
if Var_Expr_1 /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 2782 '<FieldAccess for next_part_for_decl at ast.py:3946>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3946
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.next_part_for_decl'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Next_Part_For_Decl (Node => Var_Expr_1.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 2782
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Basic_Decl;
end if;
--# property-call-start BaseFormalParamDecl.get_param
Fld_2 := Libadalang.Implementation.Base_Formal_Param_Decl_P_Get_Param (Node => Ent.Node, Part => Result_Var, Param => Sym);
--# end
--# expr-done 2783
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 2784 '<FieldAccess for basic_decl at ast.py:3945>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3945
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld_3 := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 2784
Scope_Result_1 := Fld_3;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Basic_Decl;
end if;
Let_Result := Result_Var_1; 
--# end
--# expr-done 2785
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Decl_P_Next_Part_For_Name;
--# end
--# property-start BaseFormalParamDecl.previous_part_for_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3951
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Decl_P_Previous_Part_For_Name
  (Node : Bare_Base_Formal_Param_Decl
      ; Sym : Symbol_Type
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Base_Formal_Param_Decl := Bare_Base_Formal_Param_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Formal_Param_Decl :=
     Internal_Entity_Base_Formal_Param_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind sym Sym
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Var_Expr : Internal_Entity_Defining_Name;
Var_Expr_1 : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity_Defining_Name;
Fld_3 : Internal_Entity_Basic_Decl;
Scope_Result_1 : Internal_Entity_Basic_Decl;
Result_Var_1 : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2790 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2788 '<FieldAccess for get_param at ast.py:3952>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3952
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2786 '<FieldAccess for parent_decl at ast.py:3953>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3953
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamDecl.parent_decl
Fld := Libadalang.Implementation.Base_Formal_Param_Decl_P_Parent_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2786
Var_Expr_1 := Fld;
if Var_Expr_1 /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 2787 '<FieldAccess for previous_part_for_decl at ast.py:3953>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3953
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.previous_part_for_decl'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Previous_Part_For_Decl (Node => Var_Expr_1.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 2787
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Basic_Decl;
end if;
--# property-call-start BaseFormalParamDecl.get_param
Fld_2 := Libadalang.Implementation.Base_Formal_Param_Decl_P_Get_Param (Node => Ent.Node, Part => Result_Var, Param => Sym);
--# end
--# expr-done 2788
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 2789 '<FieldAccess for basic_decl at ast.py:3952>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3952
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld_3 := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 2789
Scope_Result_1 := Fld_3;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Basic_Decl;
end if;
Let_Result := Result_Var_1; 
--# end
--# expr-done 2790
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Decl_P_Previous_Part_For_Name;
--# end
--# property-start BaseFormalParamDecl.next_part_for_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3958
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Decl_P_Next_Part_For_Decl
  (Node : Bare_Base_Formal_Param_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Base_Formal_Param_Decl := Bare_Base_Formal_Param_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Formal_Param_Decl :=
     Internal_Entity_Base_Formal_Param_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Symbol_Type;
Fld_1 : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2793 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2792 '<FieldAccess for next_part_for_name at ast.py:3959>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3959
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2791 '<FieldAccess for name_symbol at ast.py:3959>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3959
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 2791
--# property-call-start '[dispatcher]BasicDecl.next_part_for_name'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Next_Part_For_Name (Node => Ent.Node, Sym => Fld, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2792
Let_Result := Fld_1; 
--# end
--# expr-done 2793
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Decl_P_Next_Part_For_Decl;
--# end
--# property-start BaseFormalParamDecl.previous_part_for_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3962
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Decl_P_Previous_Part_For_Decl
  (Node : Bare_Base_Formal_Param_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Base_Formal_Param_Decl := Bare_Base_Formal_Param_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Formal_Param_Decl :=
     Internal_Entity_Base_Formal_Param_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Symbol_Type;
Fld_1 : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2796 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2795 '<FieldAccess for previous_part_for_name at ast.py:3963>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3963
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2794 '<FieldAccess for name_symbol at ast.py:3963>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3963
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 2794
--# property-call-start '[dispatcher]BasicDecl.previous_part_for_name'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Previous_Part_For_Name (Node => Ent.Node, Sym => Fld, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2795
Let_Result := Fld_1; 
--# end
--# expr-done 2796
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Decl_P_Previous_Part_For_Decl;
--# end
--# property-start BaseFormalParamDecl.is_mandatory /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3909
pragma Warnings (Off, "is not referenced");
 function Base_Formal_Param_Decl_P_Is_Mandatory
  (Node : Bare_Base_Formal_Param_Decl
  )
   return Boolean
is
   Self : Bare_Base_Formal_Param_Decl := Bare_Base_Formal_Param_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Formal_Param_Decl_P_Is_Mandatory;
--# end
   --
   --  Primitives for Bare_Component_Decl
   --
      procedure Component_Decl_Pre_Env_Actions
        (Self            : Bare_Component_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mappings : Internal_Env_Assoc_Array_Access :=
            Libadalang.Implementation.Env_Mappings_1 (Node => 
Self);
      begin
         for Mapping of Mappings.Items loop
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:4629");
         end loop;
         Dec_Ref (Mappings);
      end;
      end;
      procedure Initialize_Fields_For_Component_Decl
        (Self : Bare_Component_Decl
         ; Component_Decl_F_Ids : Bare_Defining_Name_List
         ; Component_Decl_F_Component_Def : Bare_Component_Def
         ; Component_Decl_F_Default_Expr : Bare_Expr
         ; Component_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Self.Component_Decl_F_Ids := Component_Decl_F_Ids;
            Self.Component_Decl_F_Component_Def := Component_Decl_F_Component_Def;
            Self.Component_Decl_F_Default_Expr := Component_Decl_F_Default_Expr;
            Self.Component_Decl_F_Aspects := Component_Decl_F_Aspects;
      end Initialize_Fields_For_Component_Decl;
   function Component_Decl_F_Ids
     (Node : Bare_Component_Decl) return Bare_Defining_Name_List
   is
   begin
         return Node.Component_Decl_F_Ids;
   end;
   function Component_Decl_F_Component_Def
     (Node : Bare_Component_Decl) return Bare_Component_Def
   is
   begin
         return Node.Component_Decl_F_Component_Def;
   end;
   function Component_Decl_F_Default_Expr
     (Node : Bare_Component_Decl) return Bare_Expr
   is
   begin
         return Node.Component_Decl_F_Default_Expr;
   end;
--# property-start ComponentDecl.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4631
pragma Warnings (Off, "is not referenced");
 function Component_Decl_P_Defining_Env
  (Node : Bare_Component_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Component_Decl := Bare_Component_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Component_Decl :=
     Internal_Entity_Component_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_298 with Inline_Always;
      Fld : Internal_Entity_Component_Def;
Fld_1 : Internal_Entity_Type_Expr;
Fld_2 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_298 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_298;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2799 '<FieldAccess for defining_env at ast.py:4632>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4632
--# expr-start 2798 '<FieldAccess for type_expr at ast.py:4632>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4632
--# expr-start 2797 '<FieldAccess for component_def at ast.py:4632>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4632
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Component_Def (Node => Ent.Node.Component_Decl_F_Component_Def, Info => Ent.Info);
--# expr-done 2797
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Type_Expr (Node => Fld.Node.Component_Def_F_Type_Expr, Info => Fld.Info);
--# expr-done 2798
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeExpr.defining_env
Fld_2 := Libadalang.Implementation.Type_Expr_P_Defining_Env (Node => Fld_1.Node, Origin => Origin, Include_Ud_Indexing => Include_Ud_Indexing, E_Info => Fld_1.Info);
--# end
--# expr-done 2799
      Property_Result := Fld_2;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_298;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_298;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_Decl_P_Defining_Env;
--# end
--# property-start ComponentDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4636
pragma Warnings (Off, "is not referenced");
 function Component_Decl_P_Defining_Names
  (Node : Bare_Component_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Component_Decl := Bare_Component_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_299 with Inline_Always;
      Fld : Bare_Defining_Name_List;
As_Entity : Internal_Entity_Defining_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Defining_Name;
Item_123 : Bare_Defining_Name;
Map_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_299 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_299;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2802 '<Map at ast.py:4636>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4636
   --# expr-start 2800 '<FieldAccess for ids at ast.py:4636>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4636
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Component_Decl_F_Ids;
--# expr-done 2800
      declare
         Map_Result_Vec : Internal_Entity_Defining_Name_Vectors.Vector;
      begin
      if Fld = null then
      Map_Result := Create_Internal_Entity_Defining_Name_Array (0);
      else
      declare
         Collection : constant Bare_Defining_Name_List := Fld;
      begin
         for Untyped_Item_123 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_123 := Untyped_Item_123; 
   --# scope-start
               --# bind i Item_123
      --# expr-start 2801 '<as_entity at ast.py:4636>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4636
Is_Null := Item_123 = null; 
if Is_Null then
   If_Result := No_Entity_Defining_Name;
else
As_Entity := (Info => E_Info, Node => Item_123); 
   If_Result := As_Entity;
end if;
--# expr-done 2801
         declare
            Item_To_Append : constant Internal_Entity_Defining_Name := If_Result;
         begin
            Internal_Entity_Defining_Name_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Defining_Name_Array
           (Items_Count => Natural (Internal_Entity_Defining_Name_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Defining_Name_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Defining_Name_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Defining_Name_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 2802
      Property_Result := Map_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_299;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_299;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_Decl_P_Defining_Names;
--# end
--# property-start ComponentDecl.array_ndims /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4638
pragma Warnings (Off, "is not referenced");
 function Component_Decl_P_Array_Ndims
  (Node : Bare_Component_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Integer
is
   Self : Bare_Component_Decl := Bare_Component_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Component_Decl :=
     Internal_Entity_Component_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Integer;
      Fld : Internal_Entity_Component_Def;
Fld_1 : Internal_Entity_Type_Expr;
Fld_2 : Integer;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2805 '<FieldAccess for array_ndims at ast.py:4638>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4638
--# expr-start 2804 '<FieldAccess for type_expr at ast.py:4638>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4638
--# expr-start 2803 '<FieldAccess for component_def at ast.py:4638>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4638
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Component_Def (Node => Ent.Node.Component_Decl_F_Component_Def, Info => Ent.Info);
--# expr-done 2803
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Type_Expr (Node => Fld.Node.Component_Def_F_Type_Expr, Info => Fld.Info);
--# expr-done 2804
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeExpr.array_ndims
Fld_2 := Libadalang.Implementation.Type_Expr_P_Array_Ndims (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 2805
      Property_Result := Fld_2;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_Decl_P_Array_Ndims;
--# end
--# property-start ComponentDecl.type_expression /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4640
pragma Warnings (Off, "is not referenced");
 function Component_Decl_P_Type_Expression
  (Node : Bare_Component_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Expr
is
   Self : Bare_Component_Decl := Bare_Component_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Type_Expr;
      Fld : Bare_Component_Def;
Fld_1 : Bare_Type_Expr;
As_Entity : Internal_Entity_Type_Expr;
Is_Null : Boolean;
If_Result : Internal_Entity_Type_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2808 '<as_entity at ast.py:4640>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4640
--# expr-start 2807 '<FieldAccess for type_expr at ast.py:4640>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4640
--# expr-start 2806 '<FieldAccess for component_def at ast.py:4640>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4640
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Component_Decl_F_Component_Def;
--# expr-done 2806
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Component_Def_F_Type_Expr;
--# expr-done 2807
Is_Null := Fld_1 = null; 
if Is_Null then
   If_Result := No_Entity_Type_Expr;
else
As_Entity := (Info => E_Info, Node => Fld_1); 
   If_Result := As_Entity;
end if;
--# expr-done 2808
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_Decl_P_Type_Expression;
--# end
--# property-start ComponentDecl.is_constant_object /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4642
pragma Warnings (Off, "is not referenced");
 function Component_Decl_P_Is_Constant_Object
  (Node : Bare_Component_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Component_Decl := Bare_Component_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_Decl_P_Is_Constant_Object;
--# end
--# property-start ComponentDecl.constrain_prefix /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4645
pragma Warnings (Off, "is not referenced");
 function Component_Decl_P_Constrain_Prefix
  (Node : Bare_Component_Decl
      ; Prefix : Bare_Expr
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Component_Decl := Bare_Component_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Component_Decl :=
     Internal_Entity_Component_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind prefix Prefix
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_303 with Inline_Always;
      Var_Expr : Bare_Attribute_Ref;
True_Rel : Logic_Equation;
Fld : Logic_Var;
Fld_1 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_2 : Logic_Equation;
And_Pred : Logic_Equation;
Cast_Expr : Bare_Expr;
Cast_Result_1 : Bare_Attribute_Ref;
Fld_3 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_303 is
            begin
                     Dec_Ref (True_Rel);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_2);
                     Dec_Ref (And_Pred);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_303;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2817 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2816 '<If at ast.py:4646>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4646
--# expr-start 2809 '<Cast to AttributeRef at ast.py:4650>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4650
Cast_Expr := Prefix; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_Attribute_Ref_Range
   then
      Cast_Result_1 := Cast_Expr;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 2809
Var_Expr := Cast_Result_1;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 2810 '<FieldAccess for is_access_attr at ast.py:4650>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4650
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.is_access_attr
Fld_3 := Libadalang.Implementation.Attribute_Ref_P_Is_Access_Attr (Node => Var_Expr);
--# end
--# expr-done 2810
Scope_Result := Fld_3;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
if Result_Var then
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:4652") else null)); 
   If_Result := True_Rel;
else
   --# expr-start 2815 '<And at ast.py:4656>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4656
--# expr-start 2813 '<Bind at ast.py:4657>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4657
--# expr-start 2811 '<FieldAccess for expected_type_var at ast.py:4657>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4657
      if Prefix = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Prefix);
--# expr-done 2811
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
--# expr-start 2812 '<FieldAccess for container_type at ast.py:4657>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4657
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentDecl.container_type
Fld_1 := Libadalang.Implementation.Component_Decl_P_Container_Type (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2812
      Cast_Result := Create_Internal_Entity
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
Bind_Result := Solver.Create_Assign (Fld, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:4657") else null)); 
--# expr-done 2813
--# expr-start 2814 '<FieldAccess for matches_expected_prefix_type at ast.py:4658>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4658
      if Prefix = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_prefix_type
Fld_2 := Libadalang.Implementation.Expr_P_Matches_Expected_Prefix_Type (Node => Prefix, Origin => Origin);
--# end
--# expr-done 2814
And_Pred := Create_And (Bind_Result, Fld_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:4656") else null)); 
--# expr-done 2815
   If_Result := And_Pred;
end if;
      Inc_Ref (If_Result);
--# expr-done 2816
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 2817
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_303;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_303;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_Decl_P_Constrain_Prefix;
--# end
--# property-start ComponentDecl.container_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4663
pragma Warnings (Off, "is not referenced");
 function Component_Decl_P_Container_Type
  (Node : Bare_Component_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Component_Decl := Bare_Component_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_304 with Inline_Always;
      Node_Parents : Bare_Ada_Node_Array_Access;
Is_A : Boolean;
Find_Result : Bare_Ada_Node;
Cast_Result : Bare_Base_Type_Decl;
As_Entity : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_304 is
            begin
                     Dec_Ref (Node_Parents);
            end Finalizer_Scope_304;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2823 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2822 '<as_entity at ast.py:4667>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4667
--# expr-start 2821 '<Cast to BaseTypeDecl at ast.py:4667>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4667
--# expr-start 2820 '<Find at ast.py:4667>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4667
   --# expr-start 2818 '<FieldAccess for parents at ast.py:4667>' Node_Parents /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4667
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents := Parents (Self, True); 
--# expr-done 2818
      Find_Result := No_Bare_Ada_Node;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents;
      begin
         for Item_124 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_124
      --# expr-start 2819 '<IsA BaseTypeDecl at ast.py:4668>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4668
Is_A := Item_124 /= null 
and then Item_124.Kind in Ada_Base_Type_Decl; 
--# expr-done 2819
      if Is_A then
         Find_Result := Item_124;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 2820
   if Find_Result = null
      or else Find_Result.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Find_Result;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 2821
Is_Null := Cast_Result = null; 
if Is_Null then
   If_Result := No_Entity_Base_Type_Decl;
else
As_Entity := (Info => E_Info, Node => Cast_Result); 
   If_Result := As_Entity;
end if;
--# expr-done 2822
Let_Result := If_Result; 
--# end
--# expr-done 2823
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_304;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_304;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_Decl_P_Container_Type;
--# end
--# property-start ComponentDecl.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4672
pragma Warnings (Off, "is not referenced");
 function Component_Decl_P_Xref_Equation
  (Node : Bare_Component_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Component_Decl := Bare_Component_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Component_Decl :=
     Internal_Entity_Component_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_305 with Inline_Always;
            procedure Finalizer_Scope_2123 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Var_Expr : Internal_Entity_Expr;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Component_Def;
Fld_2 : Internal_Entity_Type_Expr;
Fld_3 : Logic_Equation;
Fld_4 : Internal_Entity_Expr;
Fld_5 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_6 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_7 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred_2 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_305 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_305;
            procedure Finalizer_Scope_2123 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_7);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_2123;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2836 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2824 '<FieldAccess for expr_type at ast.py:4673>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4673
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.expr_type'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Expr_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 2824
Block_Var := Fld; 
--# bind typ Block_Var
--# expr-start 2835 '<And at ast.py:4675>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4675
--# expr-start 2827 '<FieldAccess for sub_equation at ast.py:4675>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4675
--# expr-start 2826 '<FieldAccess for type_expr at ast.py:4675>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4675
--# expr-start 2825 '<FieldAccess for component_def at ast.py:4675>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4675
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Component_Def (Node => Ent.Node.Component_Decl_F_Component_Def, Info => Ent.Info);
--# expr-done 2825
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Type_Expr (Node => Fld_1.Node.Component_Def_F_Type_Expr, Info => Fld_1.Info);
--# expr-done 2826
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_2.Node, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 2827
--# expr-start 2828 '<FieldAccess for default_expr at ast.py:4676>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4676
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Expr (Node => Ent.Node.Component_Decl_F_Default_Expr, Info => Ent.Info);
--# expr-done 2828
Var_Expr := Fld_4;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind de Var_Expr
--# expr-start 2834 '<And at ast.py:4678>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4678
--# expr-start 2832 '<And at ast.py:4678>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4678
--# expr-start 2830 '<Bind at ast.py:4678>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4678
--# expr-start 2829 '<FieldAccess for expected_type_var at ast.py:4678>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4678
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Var_Expr.Node);
--# expr-done 2829
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_5, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:4678") else null)); 
--# expr-done 2830
--# expr-start 2831 '<FieldAccess for sub_equation at ast.py:4679>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4679
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_6 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 2831
And_Pred := Create_And (Bind_Result, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:4678") else null)); 
--# expr-done 2832
--# expr-start 2833 '<FieldAccess for matches_expected_assign_type at ast.py:4680>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4680
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_assign_type
Fld_7 := Libadalang.Implementation.Expr_P_Matches_Expected_Assign_Type (Node => Var_Expr.Node, Origin => Origin);
--# end
--# expr-done 2833
And_Pred_1 := Create_And (And_Pred, Fld_7, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:4678") else null)); 
--# expr-done 2834
Scope_Result := And_Pred_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2123;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:4681") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred_2 := Create_And (Fld_3, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:4675") else null)); 
--# expr-done 2835
Let_Result := And_Pred_2; Inc_Ref (Let_Result);
--# end
--# expr-done 2836
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_305;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_305;
               Finalizer_Scope_2123;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_Decl_P_Xref_Equation;
--# end
--# property-start ComponentDecl.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4685
pragma Warnings (Off, "is not referenced");
 function Component_Decl_P_Xref_Entry_Point
  (Node : Bare_Component_Decl
  )
   return Boolean
is
   Self : Bare_Component_Decl := Bare_Component_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_Decl_P_Xref_Entry_Point;
--# end
--# property-start ComponentDecl._env_mappings_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4629
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_1
  (Node : Bare_Component_Decl
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Component_Decl := Bare_Component_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1417 with Inline_Always;
      Fld : Bare_Defining_Name_List;
Cast_Expr : Bare_Component_Decl;
Cast_Result : Bare_Ada_Node;
Fld_1 : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1417 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1417;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2838 '<FieldAccess for env_mappings at ast.py:4629>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4629
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2837 '<FieldAccess for ids at ast.py:4629>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4629
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Component_Decl_F_Ids;
--# expr-done 2837
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# property-call-start AdaNode.env_mappings
Fld_1 := Libadalang.Implementation.Ada_Node_P_Env_Mappings (Node => Self, Defining_Names => Fld, Value => Cast_Result);
--# end
--# expr-done 2838
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1417;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1417;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_1;
--# end
   --
   --  Primitives for Bare_Discriminant_Spec
   --
      procedure Discriminant_Spec_Pre_Env_Actions
        (Self            : Bare_Discriminant_Spec;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mappings : Internal_Env_Assoc_Array_Access :=
            Libadalang.Implementation.Env_Mappings_0 (Node => 
Self);
      begin
         for Mapping of Mappings.Items loop
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:3975");
         end loop;
         Dec_Ref (Mappings);
      end;
      end;
      procedure Initialize_Fields_For_Discriminant_Spec
        (Self : Bare_Discriminant_Spec
         ; Discriminant_Spec_F_Ids : Bare_Defining_Name_List
         ; Discriminant_Spec_F_Type_Expr : Bare_Type_Expr
         ; Discriminant_Spec_F_Default_Expr : Bare_Expr
         ; Discriminant_Spec_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Self.Discriminant_Spec_F_Ids := Discriminant_Spec_F_Ids;
            Self.Discriminant_Spec_F_Type_Expr := Discriminant_Spec_F_Type_Expr;
            Self.Discriminant_Spec_F_Default_Expr := Discriminant_Spec_F_Default_Expr;
            Self.Discriminant_Spec_F_Aspects := Discriminant_Spec_F_Aspects;
      end Initialize_Fields_For_Discriminant_Spec;
   function Discriminant_Spec_F_Ids
     (Node : Bare_Discriminant_Spec) return Bare_Defining_Name_List
   is
   begin
         return Node.Discriminant_Spec_F_Ids;
   end;
   function Discriminant_Spec_F_Type_Expr
     (Node : Bare_Discriminant_Spec) return Bare_Type_Expr
   is
   begin
         return Node.Discriminant_Spec_F_Type_Expr;
   end;
   function Discriminant_Spec_F_Default_Expr
     (Node : Bare_Discriminant_Spec) return Bare_Expr
   is
   begin
         return Node.Discriminant_Spec_F_Default_Expr;
   end;
--# property-start DiscriminantSpec.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3977
pragma Warnings (Off, "is not referenced");
 function Discriminant_Spec_P_Defining_Names
  (Node : Bare_Discriminant_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Discriminant_Spec := Bare_Discriminant_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_248 with Inline_Always;
      Fld : Bare_Defining_Name_List;
As_Entity : Internal_Entity_Defining_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Defining_Name;
Item_125 : Bare_Defining_Name;
Map_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_248 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_248;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2841 '<Map at ast.py:3977>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3977
   --# expr-start 2839 '<FieldAccess for ids at ast.py:3977>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3977
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Discriminant_Spec_F_Ids;
--# expr-done 2839
      declare
         Map_Result_Vec : Internal_Entity_Defining_Name_Vectors.Vector;
      begin
      if Fld = null then
      Map_Result := Create_Internal_Entity_Defining_Name_Array (0);
      else
      declare
         Collection : constant Bare_Defining_Name_List := Fld;
      begin
         for Untyped_Item_125 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_125 := Untyped_Item_125; 
   --# scope-start
               --# bind id Item_125
      --# expr-start 2840 '<as_entity at ast.py:3977>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3977
Is_Null := Item_125 = null; 
if Is_Null then
   If_Result := No_Entity_Defining_Name;
else
As_Entity := (Info => E_Info, Node => Item_125); 
   If_Result := As_Entity;
end if;
--# expr-done 2840
         declare
            Item_To_Append : constant Internal_Entity_Defining_Name := If_Result;
         begin
            Internal_Entity_Defining_Name_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Defining_Name_Array
           (Items_Count => Natural (Internal_Entity_Defining_Name_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Defining_Name_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Defining_Name_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Defining_Name_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 2841
      Property_Result := Map_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_248;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_248;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Discriminant_Spec_P_Defining_Names;
--# end
--# property-start DiscriminantSpec.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3978
pragma Warnings (Off, "is not referenced");
 function Discriminant_Spec_P_Defining_Env
  (Node : Bare_Discriminant_Spec
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Discriminant_Spec := Bare_Discriminant_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Discriminant_Spec :=
     Internal_Entity_Discriminant_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_249 with Inline_Always;
      Fld : Internal_Entity_Type_Expr;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_249 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_249;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2843 '<FieldAccess for defining_env at ast.py:3978>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3978
--# expr-start 2842 '<FieldAccess for type_expr at ast.py:3978>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3978
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Expr (Node => Ent.Node.Discriminant_Spec_F_Type_Expr, Info => Ent.Info);
--# expr-done 2842
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeExpr.defining_env
Fld_1 := Libadalang.Implementation.Type_Expr_P_Defining_Env (Node => Fld.Node, Origin => Origin, Include_Ud_Indexing => Include_Ud_Indexing, E_Info => Fld.Info);
--# end
--# expr-done 2843
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_249;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_249;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Discriminant_Spec_P_Defining_Env;
--# end
--# property-start DiscriminantSpec.type_expression /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3980
pragma Warnings (Off, "is not referenced");
 function Discriminant_Spec_P_Type_Expression
  (Node : Bare_Discriminant_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Expr
is
   Self : Bare_Discriminant_Spec := Bare_Discriminant_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Discriminant_Spec :=
     Internal_Entity_Discriminant_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Type_Expr;
      Fld : Internal_Entity_Type_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2844 '<FieldAccess for type_expr at ast.py:3980>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3980
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Expr (Node => Ent.Node.Discriminant_Spec_F_Type_Expr, Info => Ent.Info);
--# expr-done 2844
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Discriminant_Spec_P_Type_Expression;
--# end
--# property-start DiscriminantSpec.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3982
pragma Warnings (Off, "is not referenced");
 function Discriminant_Spec_P_Xref_Entry_Point
  (Node : Bare_Discriminant_Spec
  )
   return Boolean
is
   Self : Bare_Discriminant_Spec := Bare_Discriminant_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Discriminant_Spec_P_Xref_Entry_Point;
--# end
--# property-start DiscriminantSpec.is_constant_object /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3984
pragma Warnings (Off, "is not referenced");
 function Discriminant_Spec_P_Is_Constant_Object
  (Node : Bare_Discriminant_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Discriminant_Spec := Bare_Discriminant_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Discriminant_Spec_P_Is_Constant_Object;
--# end
--# property-start DiscriminantSpec.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3987
pragma Warnings (Off, "is not referenced");
 function Discriminant_Spec_P_Xref_Equation
  (Node : Bare_Discriminant_Spec
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Discriminant_Spec := Bare_Discriminant_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Discriminant_Spec :=
     Internal_Entity_Discriminant_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_253 with Inline_Always;
            procedure Finalizer_Scope_2125 with Inline_Always;
      Var_Expr : Internal_Entity_Expr;
Fld : Internal_Entity_Type_Expr;
Fld_1 : Logic_Equation;
Fld_2 : Internal_Entity_Expr;
Fld_3 : Logic_Var;
Fld_4 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_5 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_6 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred_2 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_253 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_253;
            procedure Finalizer_Scope_2125 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_5);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_2125;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2856 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2855 '<And at ast.py:3988>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3988
--# expr-start 2846 '<FieldAccess for sub_equation at ast.py:3989>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3989
--# expr-start 2845 '<FieldAccess for type_expr at ast.py:3989>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3989
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Expr (Node => Ent.Node.Discriminant_Spec_F_Type_Expr, Info => Ent.Info);
--# expr-done 2845
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 2846
--# expr-start 2847 '<FieldAccess for default_expr at ast.py:3990>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3990
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Expr (Node => Ent.Node.Discriminant_Spec_F_Default_Expr, Info => Ent.Info);
--# expr-done 2847
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind de Var_Expr
--# expr-start 2854 '<And at ast.py:3992>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3992
--# expr-start 2852 '<And at ast.py:3992>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3992
--# expr-start 2850 '<Bind at ast.py:3992>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3992
--# expr-start 2848 '<FieldAccess for expected_type_var at ast.py:3992>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3992
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Var_Expr.Node);
--# expr-done 2848
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
--# expr-start 2849 '<FieldAccess for expr_type at ast.py:3992>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3992
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.expr_type'
Fld_4 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Expr_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 2849
      Cast_Result := Create_Internal_Entity
        (Node => Fld_4.Node,
         Info => Fld_4.Info);
Bind_Result := Solver.Create_Assign (Fld_3, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:3992") else null)); 
--# expr-done 2850
--# expr-start 2851 '<FieldAccess for sub_equation at ast.py:3993>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3993
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_5 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 2851
And_Pred := Create_And (Bind_Result, Fld_5, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:3992") else null)); 
--# expr-done 2852
--# expr-start 2853 '<FieldAccess for matches_expected_assign_type at ast.py:3994>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3994
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_assign_type
Fld_6 := Libadalang.Implementation.Expr_P_Matches_Expected_Assign_Type (Node => Var_Expr.Node, Origin => Origin);
--# end
--# expr-done 2853
And_Pred_1 := Create_And (And_Pred, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:3992") else null)); 
--# expr-done 2854
Scope_Result := And_Pred_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2125;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:3995") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred_2 := Create_And (Fld_1, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:3988") else null)); 
--# expr-done 2855
Let_Result := And_Pred_2; Inc_Ref (Let_Result);
--# end
--# expr-done 2856
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_253;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_253;
               Finalizer_Scope_2125;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Discriminant_Spec_P_Xref_Equation;
--# end
--# property-start DiscriminantSpec._env_mappings_0 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3975
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_0
  (Node : Bare_Discriminant_Spec
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Discriminant_Spec := Bare_Discriminant_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1416 with Inline_Always;
      Fld : Bare_Defining_Name_List;
Cast_Expr : Bare_Discriminant_Spec;
Cast_Result : Bare_Ada_Node;
Fld_1 : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1416 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1416;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2858 '<FieldAccess for env_mappings at ast.py:3975>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3975
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2857 '<FieldAccess for ids at ast.py:3975>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3975
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Discriminant_Spec_F_Ids;
--# expr-done 2857
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# property-call-start AdaNode.env_mappings
Fld_1 := Libadalang.Implementation.Ada_Node_P_Env_Mappings (Node => Self, Defining_Names => Fld, Value => Cast_Result);
--# end
--# expr-done 2858
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1416;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1416;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_0;
--# end
   --
   --  Primitives for Bare_Generic_Formal
   --
      procedure Initialize_Fields_For_Generic_Formal
        (Self : Bare_Generic_Formal
         ; Generic_Formal_F_Decl : Bare_Basic_Decl
        ) is
      begin
            Self.Generic_Formal_F_Decl := Generic_Formal_F_Decl;
      end Initialize_Fields_For_Generic_Formal;
   function Generic_Formal_F_Decl
     (Node : Bare_Generic_Formal) return Bare_Basic_Decl
   is
   begin
         return Node.Generic_Formal_F_Decl;
   end;
--# property-start GenericFormal.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10735
pragma Warnings (Off, "is not referenced");
 function Generic_Formal_P_Defining_Names
  (Node : Bare_Generic_Formal
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Generic_Formal := Bare_Generic_Formal (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Formal :=
     Internal_Entity_Generic_Formal'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_842 with Inline_Always;
      Fld : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_842 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_842;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2860 '<FieldAccess for defining_names at ast.py:10735>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10735
--# expr-start 2859 '<FieldAccess for decl at ast.py:10735>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10735
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Basic_Decl (Node => Ent.Node.Generic_Formal_F_Decl, Info => Ent.Info);
--# expr-done 2859
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_names'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Names (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 2860
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_842;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_842;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Formal_P_Defining_Names;
--# end
--# property-start GenericFormal.type_expression /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10736
pragma Warnings (Off, "is not referenced");
 function Generic_Formal_P_Type_Expression
  (Node : Bare_Generic_Formal
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Expr
is
   Self : Bare_Generic_Formal := Bare_Generic_Formal (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Formal :=
     Internal_Entity_Generic_Formal'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Type_Expr;
      Fld : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Type_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2862 '<FieldAccess for type_expression at ast.py:10736>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10736
--# expr-start 2861 '<FieldAccess for decl at ast.py:10736>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10736
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Basic_Decl (Node => Ent.Node.Generic_Formal_F_Decl, Info => Ent.Info);
--# expr-done 2861
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.type_expression'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Type_Expression (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 2862
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Formal_P_Type_Expression;
--# end
--# property-start GenericFormal.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10738
pragma Warnings (Off, "is not referenced");
 function Generic_Formal_P_Xref_Entry_Point
  (Node : Bare_Generic_Formal
  )
   return Boolean
is
   Self : Bare_Generic_Formal := Bare_Generic_Formal (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Formal_P_Xref_Entry_Point;
--# end
--# property-start GenericFormal.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10739
pragma Warnings (Off, "is not referenced");
 function Generic_Formal_P_Xref_Equation
  (Node : Bare_Generic_Formal
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Generic_Formal := Bare_Generic_Formal (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Formal :=
     Internal_Entity_Generic_Formal'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_845 with Inline_Always;
      Fld : Internal_Entity_Basic_Decl;
Fld_1 : Logic_Equation;
            procedure Finalizer_Scope_845 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_845;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2864 '<FieldAccess for xref_equation at ast.py:10739>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10739
--# expr-start 2863 '<FieldAccess for decl at ast.py:10739>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10739
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Basic_Decl (Node => Ent.Node.Generic_Formal_F_Decl, Info => Ent.Info);
--# expr-done 2863
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_1 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 2864
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_845;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_845;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Formal_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Generic_Formal_Obj_Decl
   --
      procedure Initialize_Fields_For_Generic_Formal_Obj_Decl
        (Self : Bare_Generic_Formal_Obj_Decl
         ; Generic_Formal_F_Decl : Bare_Basic_Decl
        ) is
      begin
            Initialize_Fields_For_Generic_Formal
              (Self, Generic_Formal_F_Decl);
      end Initialize_Fields_For_Generic_Formal_Obj_Decl;
   --
   --  Primitives for Bare_Generic_Formal_Package
   --
      procedure Initialize_Fields_For_Generic_Formal_Package
        (Self : Bare_Generic_Formal_Package
         ; Generic_Formal_F_Decl : Bare_Basic_Decl
        ) is
      begin
            Initialize_Fields_For_Generic_Formal
              (Self, Generic_Formal_F_Decl);
      end Initialize_Fields_For_Generic_Formal_Package;
   --
   --  Primitives for Bare_Generic_Formal_Subp_Decl
   --
      procedure Initialize_Fields_For_Generic_Formal_Subp_Decl
        (Self : Bare_Generic_Formal_Subp_Decl
         ; Generic_Formal_F_Decl : Bare_Basic_Decl
        ) is
      begin
            Initialize_Fields_For_Generic_Formal
              (Self, Generic_Formal_F_Decl);
      end Initialize_Fields_For_Generic_Formal_Subp_Decl;
   --
   --  Primitives for Bare_Generic_Formal_Type_Decl
   --
      procedure Initialize_Fields_For_Generic_Formal_Type_Decl
        (Self : Bare_Generic_Formal_Type_Decl
         ; Generic_Formal_F_Decl : Bare_Basic_Decl
        ) is
      begin
            Initialize_Fields_For_Generic_Formal
              (Self, Generic_Formal_F_Decl);
      end Initialize_Fields_For_Generic_Formal_Type_Decl;
   --
   --  Primitives for Bare_Param_Spec
   --
      procedure Param_Spec_Pre_Env_Actions
        (Self            : Bare_Param_Spec;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mappings : Internal_Env_Assoc_Array_Access :=
            Libadalang.Implementation.Env_Mappings_24 (Node => 
Self);
      begin
         for Mapping of Mappings.Items loop
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:8675");
         end loop;
         Dec_Ref (Mappings);
      end;
      end;
      procedure Initialize_Fields_For_Param_Spec
        (Self : Bare_Param_Spec
         ; Param_Spec_F_Ids : Bare_Defining_Name_List
         ; Param_Spec_F_Has_Aliased : Bare_Aliased_Node
         ; Param_Spec_F_Mode : Bare_Mode
         ; Param_Spec_F_Type_Expr : Bare_Type_Expr
         ; Param_Spec_F_Default_Expr : Bare_Expr
         ; Param_Spec_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Self.Param_Spec_F_Ids := Param_Spec_F_Ids;
            Self.Param_Spec_F_Has_Aliased := Param_Spec_F_Has_Aliased;
            Self.Param_Spec_F_Mode := Param_Spec_F_Mode;
            Self.Param_Spec_F_Type_Expr := Param_Spec_F_Type_Expr;
            Self.Param_Spec_F_Default_Expr := Param_Spec_F_Default_Expr;
            Self.Param_Spec_F_Aspects := Param_Spec_F_Aspects;
      end Initialize_Fields_For_Param_Spec;
   function Param_Spec_F_Ids
     (Node : Bare_Param_Spec) return Bare_Defining_Name_List
   is
   begin
         return Node.Param_Spec_F_Ids;
   end;
   function Param_Spec_F_Has_Aliased
     (Node : Bare_Param_Spec) return Bare_Aliased_Node
   is
   begin
         return Node.Param_Spec_F_Has_Aliased;
   end;
   function Param_Spec_F_Mode
     (Node : Bare_Param_Spec) return Bare_Mode
   is
   begin
         return Node.Param_Spec_F_Mode;
   end;
   function Param_Spec_F_Type_Expr
     (Node : Bare_Param_Spec) return Bare_Type_Expr
   is
   begin
         return Node.Param_Spec_F_Type_Expr;
   end;
   function Param_Spec_F_Default_Expr
     (Node : Bare_Param_Spec) return Bare_Expr
   is
   begin
         return Node.Param_Spec_F_Default_Expr;
   end;
--# property-start ParamSpec.is_mandatory /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8671
pragma Warnings (Off, "is not referenced");
 function Param_Spec_P_Is_Mandatory
  (Node : Bare_Param_Spec
  )
   return Boolean
is
   Self : Bare_Param_Spec := Bare_Param_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Expr;
Is_Null : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2866 '<IsNull at ast.py:8671>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8671
--# expr-start 2865 '<FieldAccess for default_expr at ast.py:8671>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8671
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Param_Spec_F_Default_Expr;
--# expr-done 2865
Is_Null := Fld = null; 
--# expr-done 2866
      Property_Result := Is_Null;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Param_Spec_P_Is_Mandatory;
--# end
--# property-start ParamSpec.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8672
pragma Warnings (Off, "is not referenced");
 function Param_Spec_P_Defining_Names
  (Node : Bare_Param_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Param_Spec := Bare_Param_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_708 with Inline_Always;
      Fld : Bare_Defining_Name_List;
As_Entity : Internal_Entity_Defining_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Defining_Name;
Item_126 : Bare_Defining_Name;
Map_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_708 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_708;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2869 '<Map at ast.py:8672>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8672
   --# expr-start 2867 '<FieldAccess for ids at ast.py:8672>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8672
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Param_Spec_F_Ids;
--# expr-done 2867
      declare
         Map_Result_Vec : Internal_Entity_Defining_Name_Vectors.Vector;
      begin
      if Fld = null then
      Map_Result := Create_Internal_Entity_Defining_Name_Array (0);
      else
      declare
         Collection : constant Bare_Defining_Name_List := Fld;
      begin
         for Untyped_Item_126 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_126 := Untyped_Item_126; 
   --# scope-start
               --# bind id Item_126
      --# expr-start 2868 '<as_entity at ast.py:8672>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8672
Is_Null := Item_126 = null; 
if Is_Null then
   If_Result := No_Entity_Defining_Name;
else
As_Entity := (Info => E_Info, Node => Item_126); 
   If_Result := As_Entity;
end if;
--# expr-done 2868
         declare
            Item_To_Append : constant Internal_Entity_Defining_Name := If_Result;
         begin
            Internal_Entity_Defining_Name_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Defining_Name_Array
           (Items_Count => Natural (Internal_Entity_Defining_Name_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Defining_Name_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Defining_Name_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Defining_Name_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 2869
      Property_Result := Map_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_708;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_708;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Param_Spec_P_Defining_Names;
--# end
--# property-start ParamSpec.type_expression /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8678
pragma Warnings (Off, "is not referenced");
 function Param_Spec_P_Type_Expression
  (Node : Bare_Param_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Expr
is
   Self : Bare_Param_Spec := Bare_Param_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Param_Spec :=
     Internal_Entity_Param_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Type_Expr;
      Fld : Internal_Entity_Type_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2870 '<FieldAccess for type_expr at ast.py:8678>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8678
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Expr (Node => Ent.Node.Param_Spec_F_Type_Expr, Info => Ent.Info);
--# expr-done 2870
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Param_Spec_P_Type_Expression;
--# end
--# property-start ParamSpec.is_constant_object /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8681
pragma Warnings (Off, "is not referenced");
 function Param_Spec_P_Is_Constant_Object
  (Node : Bare_Param_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Param_Spec := Bare_Param_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Mode;
Is_A : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2873 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2872 '<IsA Mode.In, Mode.Default at ast.py:8682>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8682
--# expr-start 2871 '<FieldAccess for mode at ast.py:8682>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8682
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Param_Spec_F_Mode;
--# expr-done 2871
Is_A := Fld /= null 
and then Fld.Kind in Ada_Mode_In_Range | Ada_Mode_Default_Range; 
--# expr-done 2872
Let_Result := Is_A; 
--# end
--# expr-done 2873
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Param_Spec_P_Is_Constant_Object;
--# end
--# property-start ParamSpec.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8685
pragma Warnings (Off, "is not referenced");
 function Param_Spec_P_Defining_Env
  (Node : Bare_Param_Spec
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Param_Spec := Bare_Param_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Param_Spec :=
     Internal_Entity_Param_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_711 with Inline_Always;
      Fld : Internal_Entity_Type_Expr;
Fld_1 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_711 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_711;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2876 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2875 '<FieldAccess for defining_env at ast.py:8686>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8686
--# expr-start 2874 '<FieldAccess for type_expr at ast.py:8686>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8686
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Expr (Node => Ent.Node.Param_Spec_F_Type_Expr, Info => Ent.Info);
--# expr-done 2874
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeExpr.defining_env
Fld_1 := Libadalang.Implementation.Type_Expr_P_Defining_Env (Node => Fld.Node, Origin => Origin, Include_Ud_Indexing => Include_Ud_Indexing, E_Info => Fld.Info);
--# end
--# expr-done 2875
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 2876
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_711;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_711;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Param_Spec_P_Defining_Env;
--# end
--# property-start ParamSpec.constrain_prefix /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8689
pragma Warnings (Off, "is not referenced");
 function Param_Spec_P_Constrain_Prefix
  (Node : Bare_Param_Spec
      ; Prefix : Bare_Expr
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Param_Spec := Bare_Param_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind prefix Prefix
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_712 with Inline_Always;
            procedure Finalizer_Scope_2127 with Inline_Always;
      Var_Expr : Bare_Name;
Fld : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Cast_Expr_1 : Bare_Expr;
Cast_Result_1 : Bare_Name;
Fld_1 : Logic_Var;
Cast_Expr_2 : Internal_Entity_Base_Formal_Param_Holder;
Cast_Result_2 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
Scope_Result : Logic_Equation;
False_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_712 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (False_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_712;
            procedure Finalizer_Scope_2127 is
            begin
                     Dec_Ref (Bind_Result_1);
            end Finalizer_Scope_2127;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2883 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2882 '<And at ast.py:8695>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8695
--# expr-start 2878 '<Bind at ast.py:8696>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8696
--# expr-start 2877 '<FieldAccess for type_var at ast.py:8696>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8696
      if Prefix = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Prefix);
--# expr-done 2877
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
Cast_Expr := No_Entity_Base_Type_Decl; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8696") else null)); 
--# expr-done 2878
--# expr-start 2879 '<Cast to Name at ast.py:8697>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8697
Cast_Expr_1 := Prefix; 
   if Cast_Expr_1 = null
      or else Cast_Expr_1.Kind in Ada_Name
   then
      Cast_Result_1 := Cast_Expr_1;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 2879
Var_Expr := Cast_Result_1;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind name Var_Expr
--# expr-start 2881 '<Bind at ast.py:8698>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8698
--# expr-start 2880 '<FieldAccess for subp_spec_var at ast.py:8698>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8698
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.subp_spec_var'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Subp_Spec_Var (Node => Var_Expr);
--# end
--# expr-done 2880
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
Cast_Expr_2 := No_Entity_Base_Formal_Param_Holder; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_1, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8698") else null)); 
--# expr-done 2881
Scope_Result := Bind_Result_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2127;
   Result_Var := Scope_Result;
else
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8700") else null)); 
   Result_Var := False_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred := Create_And (Bind_Result, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8695") else null)); 
--# expr-done 2882
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 2883
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_712;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_712;
               Finalizer_Scope_2127;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Param_Spec_P_Constrain_Prefix;
--# end
--# property-start ParamSpec.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8705
pragma Warnings (Off, "is not referenced");
 function Param_Spec_P_Xref_Equation
  (Node : Bare_Param_Spec
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Param_Spec := Bare_Param_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Param_Spec :=
     Internal_Entity_Param_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_713 with Inline_Always;
            procedure Finalizer_Scope_2128 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Var_Expr : Internal_Entity_Expr;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Type_Expr;
Fld_2 : Logic_Equation;
Fld_3 : Internal_Entity_Expr;
Fld_4 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_5 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_6 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred_2 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_713 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_713;
            procedure Finalizer_Scope_2128 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_5);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_2128;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2895 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2884 '<FieldAccess for expr_type at ast.py:8706>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8706
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.expr_type'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Expr_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 2884
Block_Var := Fld; 
--# bind typ Block_Var
--# expr-start 2894 '<And at ast.py:8708>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8708
--# expr-start 2886 '<FieldAccess for sub_equation at ast.py:8708>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8708
--# expr-start 2885 '<FieldAccess for type_expr at ast.py:8708>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8708
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Type_Expr (Node => Ent.Node.Param_Spec_F_Type_Expr, Info => Ent.Info);
--# expr-done 2885
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_2 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_1.Node, Env => Env, Origin => Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 2886
--# expr-start 2887 '<FieldAccess for default_expr at ast.py:8710>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8710
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Expr (Node => Ent.Node.Param_Spec_F_Default_Expr, Info => Ent.Info);
--# expr-done 2887
Var_Expr := Fld_3;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind de Var_Expr
--# expr-start 2893 '<And at ast.py:8712>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8712
--# expr-start 2891 '<And at ast.py:8712>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8712
--# expr-start 2889 '<Bind at ast.py:8712>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8712
--# expr-start 2888 '<FieldAccess for expected_type_var at ast.py:8712>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8712
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Var_Expr.Node);
--# expr-done 2888
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_4, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8712") else null)); 
--# expr-done 2889
--# expr-start 2890 '<FieldAccess for sub_equation at ast.py:8713>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8713
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_5 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 2890
And_Pred := Create_And (Bind_Result, Fld_5, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8712") else null)); 
--# expr-done 2891
--# expr-start 2892 '<FieldAccess for matches_expected_assign_type at ast.py:8714>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8714
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_assign_type
Fld_6 := Libadalang.Implementation.Expr_P_Matches_Expected_Assign_Type (Node => Var_Expr.Node, Origin => Origin);
--# end
--# expr-done 2892
And_Pred_1 := Create_And (And_Pred, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8712") else null)); 
--# expr-done 2893
Scope_Result := And_Pred_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2128;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8715") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred_2 := Create_And (Fld_2, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8708") else null)); 
--# expr-done 2894
Let_Result := And_Pred_2; Inc_Ref (Let_Result);
--# end
--# expr-done 2895
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_713;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_713;
               Finalizer_Scope_2128;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Param_Spec_P_Xref_Equation;
--# end
--# property-start ParamSpec.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8719
pragma Warnings (Off, "is not referenced");
 function Param_Spec_P_Xref_Entry_Point
  (Node : Bare_Param_Spec
  )
   return Boolean
is
   Self : Bare_Param_Spec := Bare_Param_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Param_Spec_P_Xref_Entry_Point;
--# end
--# property-start ParamSpec._env_mappings_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8675
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_24
  (Node : Bare_Param_Spec
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Param_Spec := Bare_Param_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1441 with Inline_Always;
      Fld : Bare_Defining_Name_List;
Cast_Expr : Bare_Param_Spec;
Cast_Result : Bare_Ada_Node;
Fld_1 : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1441 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1441;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2897 '<FieldAccess for env_mappings at ast.py:8675>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8675
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2896 '<FieldAccess for ids at ast.py:8675>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8675
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Param_Spec_F_Ids;
--# expr-done 2896
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# property-call-start AdaNode.env_mappings
Fld_1 := Libadalang.Implementation.Ada_Node_P_Env_Mappings (Node => Self, Defining_Names => Fld, Value => Cast_Result);
--# end
--# expr-done 2897
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1441;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1441;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_24;
--# end
   --
   --  Primitives for Bare_Synthetic_Formal_Param_Decl
   --
      procedure Initialize_Fields_For_Synthetic_Formal_Param_Decl
        (Self : Bare_Synthetic_Formal_Param_Decl
         ; Synthetic_Formal_Param_Decl_F_Param_Type : Bare_Type_Expr
        ) is
      begin
            Self.Synthetic_Formal_Param_Decl_F_Param_Type := Synthetic_Formal_Param_Decl_F_Param_Type;
      Self.Synthetic_Formal_Param_Decl_Param_Name := null;
      end Initialize_Fields_For_Synthetic_Formal_Param_Decl;
   function Synthetic_Formal_Param_Decl_F_Param_Type
     (Node : Bare_Synthetic_Formal_Param_Decl) return Bare_Type_Expr
   is
   begin
         return Node.Synthetic_Formal_Param_Decl_F_Param_Type;
   end;
--# property-start SyntheticFormalParamDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16738
pragma Warnings (Off, "is not referenced");
 function Synthetic_Formal_Param_Decl_P_Defining_Names
  (Node : Bare_Synthetic_Formal_Param_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Synthetic_Formal_Param_Decl := Bare_Synthetic_Formal_Param_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1190 with Inline_Always;
      Fld : Symbol_Type;
Fld_1 : Bare_Defining_Name;
As_Entity : Internal_Entity_Defining_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Defining_Name;
Array_Lit : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1190 is
            begin
                     Dec_Ref (Array_Lit);
            end Finalizer_Scope_1190;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2901 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 2900 '<as_entity at ast.py:16739>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16739
--# expr-start 2899 '<FieldAccess for synthesize_defining_name at ast.py:16739>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16739
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2898 '<FieldAccess for param_name at ast.py:16739>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16739
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Synthetic_Formal_Param_Decl_Param_Name;
--# expr-done 2898
--# property-call-start AdaNode.synthesize_defining_name
Fld_1 := Libadalang.Implementation.Ada_Node_P_Synthesize_Defining_Name (Node => Self, Sym => Fld);
--# end
--# expr-done 2899
Is_Null := Fld_1 = null; 
if Is_Null then
   If_Result := No_Entity_Defining_Name;
else
As_Entity := (Info => E_Info, Node => Fld_1); 
   If_Result := As_Entity;
end if;
--# expr-done 2900
Array_Lit := Create_Internal_Entity_Defining_Name_Array (Internal_Internal_Entity_Defining_Name_Array'(1 => If_Result)); 
--# expr-done 2901
      Property_Result := Array_Lit;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1190;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1190;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Synthetic_Formal_Param_Decl_P_Defining_Names;
--# end
--# property-start SyntheticFormalParamDecl.type_expression /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16742
pragma Warnings (Off, "is not referenced");
 function Synthetic_Formal_Param_Decl_P_Type_Expression
  (Node : Bare_Synthetic_Formal_Param_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Expr
is
   Self : Bare_Synthetic_Formal_Param_Decl := Bare_Synthetic_Formal_Param_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Synthetic_Formal_Param_Decl :=
     Internal_Entity_Synthetic_Formal_Param_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Type_Expr;
      Fld : Internal_Entity_Type_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2902 '<FieldAccess for param_type at ast.py:16742>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16742
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Expr (Node => Ent.Node.Synthetic_Formal_Param_Decl_F_Param_Type, Info => Ent.Info);
--# expr-done 2902
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Synthetic_Formal_Param_Decl_P_Type_Expression;
--# end
   --
   --  Primitives for Bare_Base_Package_Decl
   --
      procedure Initialize_Fields_For_Base_Package_Decl
        (Self : Bare_Base_Package_Decl
         ; Base_Package_Decl_F_Package_Name : Bare_Defining_Name
         ; Base_Package_Decl_F_Aspects : Bare_Aspect_Spec
         ; Base_Package_Decl_F_Public_Part : Bare_Public_Part
         ; Base_Package_Decl_F_Private_Part : Bare_Private_Part
         ; Base_Package_Decl_F_End_Name : Bare_End_Name
        ) is
      begin
            Self.Base_Package_Decl_F_Package_Name := Base_Package_Decl_F_Package_Name;
            Self.Base_Package_Decl_F_Aspects := Base_Package_Decl_F_Aspects;
            Self.Base_Package_Decl_F_Public_Part := Base_Package_Decl_F_Public_Part;
            Self.Base_Package_Decl_F_Private_Part := Base_Package_Decl_F_Private_Part;
            Self.Base_Package_Decl_F_End_Name := Base_Package_Decl_F_End_Name;
      end Initialize_Fields_For_Base_Package_Decl;
   function Base_Package_Decl_F_Package_Name
     (Node : Bare_Base_Package_Decl) return Bare_Defining_Name
   is
   begin
         return Node.Base_Package_Decl_F_Package_Name;
   end;
   function Base_Package_Decl_F_Public_Part
     (Node : Bare_Base_Package_Decl) return Bare_Public_Part
   is
   begin
         return Node.Base_Package_Decl_F_Public_Part;
   end;
   function Base_Package_Decl_F_Private_Part
     (Node : Bare_Base_Package_Decl) return Bare_Private_Part
   is
   begin
         return Node.Base_Package_Decl_F_Private_Part;
   end;
   function Base_Package_Decl_F_End_Name
     (Node : Bare_Base_Package_Decl) return Bare_End_Name
   is
   begin
         return Node.Base_Package_Decl_F_End_Name;
   end;
--# property-start BasePackageDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9877
pragma Warnings (Off, "is not referenced");
 function Base_Package_Decl_P_Defining_Names
  (Node : Bare_Base_Package_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Base_Package_Decl := Bare_Base_Package_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Package_Decl :=
     Internal_Entity_Base_Package_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_787 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_787 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_787;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2903 '<FieldAccess for package_name at ast.py:9877>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9877
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Base_Package_Decl_F_Package_Name, Info => Ent.Info);
--# expr-done 2903
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_787;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_787;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Package_Decl_P_Defining_Names;
--# end
--# property-start BasePackageDecl.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9878
pragma Warnings (Off, "is not referenced");
 function Base_Package_Decl_P_Defining_Env
  (Node : Bare_Base_Package_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Base_Package_Decl := Bare_Base_Package_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Package_Decl :=
     Internal_Entity_Base_Package_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_788 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_788 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_788;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2904 '<FieldAccess for children_env at ast.py:9878>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9878
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 2904
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_788;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_788;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Package_Decl_P_Defining_Env;
--# end
--# property-start BasePackageDecl.body_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9881
pragma Warnings (Off, "is not referenced");
 function Base_Package_Decl_P_Body_Part
  (Node : Bare_Base_Package_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Package_Body
is
   Self : Bare_Base_Package_Decl := Bare_Base_Package_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Package_Decl :=
     Internal_Entity_Base_Package_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Package_Body;
      Bound_Imprecise_Fallback : Boolean;
Fld : Internal_Entity_Body_Node;
Cast_Result : Internal_Entity_Package_Body;
Dyn_Var_Bind_Result : Internal_Entity_Package_Body;
Let_Result : Internal_Entity_Package_Body;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2908 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2907 '<bind at ast.py:9885>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9885
Bound_Imprecise_Fallback := False; 
--# expr-start 2906 '<Cast to PackageBody at ast.py:9887>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9887
--# expr-start 2905 '<FieldAccess for body_part_for_decl at ast.py:9887>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9887
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.body_part_for_decl
Fld := Libadalang.Implementation.Basic_Decl_P_Body_Part_For_Decl (Node => Ent.Node, Imprecise_Fallback => Bound_Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 2905
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Package_Body_Range
   then
      Cast_Result := Create_Internal_Entity_Package_Body
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Package_Body;
   end if;
--# expr-done 2906
Dyn_Var_Bind_Result := Cast_Result; 
--# expr-done 2907
Let_Result := Dyn_Var_Bind_Result; 
--# end
--# expr-done 2908
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Package_Decl_P_Body_Part;
--# end
--# property-start BasePackageDecl.declarative_parts /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9892
pragma Warnings (Off, "is not referenced");
 function Base_Package_Decl_P_Declarative_Parts
  (Node : Bare_Base_Package_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Declarative_Part_Array_Access
is
   Self : Bare_Base_Package_Decl := Bare_Base_Package_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Package_Decl :=
     Internal_Entity_Base_Package_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_790 with Inline_Always;
            procedure Finalizer_Scope_2129 with Inline_Always;
      Var_Expr : Internal_Entity_Declarative_Part;
Fld : Internal_Entity_Public_Part;
Cast_Result : Internal_Entity_Declarative_Part;
Singleton : Internal_Entity_Declarative_Part_Array_Access;
Fld_1 : Internal_Entity_Private_Part;
Cast_Result_1 : Internal_Entity_Declarative_Part;
Singleton_1 : Internal_Entity_Declarative_Part_Array_Access;
Scope_Result : Internal_Entity_Declarative_Part_Array_Access;
Result_Var : Internal_Entity_Declarative_Part_Array_Access;
Concat_Result : Internal_Entity_Declarative_Part_Array_Access;
Let_Result : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_790 is
            begin
                     Dec_Ref (Singleton);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_790;
            procedure Finalizer_Scope_2129 is
            begin
                     Dec_Ref (Singleton_1);
            end Finalizer_Scope_2129;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2914 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2913 '<Concat at ast.py:9893>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9893
--# expr-start 2910 '<Cast to DeclarativePart at ast.py:9893>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9893
--# expr-start 2909 '<FieldAccess for public_part at ast.py:9893>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9893
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Public_Part (Node => Ent.Node.Base_Package_Decl_F_Public_Part, Info => Ent.Info);
--# expr-done 2909
      Cast_Result := Create_Internal_Entity_Declarative_Part
        (Node => Fld.Node,
         Info => Fld.Info);
--# expr-done 2910
                Singleton := Create_Internal_Entity_Declarative_Part_Array (Items_Count => 1);
                Singleton.Items (1) := Cast_Result;
--# expr-start 2912 '<Cast to DeclarativePart at ast.py:9894>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9894
--# expr-start 2911 '<FieldAccess for private_part at ast.py:9894>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9894
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Private_Part (Node => Ent.Node.Base_Package_Decl_F_Private_Part, Info => Ent.Info);
--# expr-done 2911
      Cast_Result_1 := Create_Internal_Entity_Declarative_Part
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
--# expr-done 2912
Var_Expr := Cast_Result_1;
if Var_Expr /= No_Entity_Declarative_Part then
   --# scope-start
                Singleton_1 := Create_Internal_Entity_Declarative_Part_Array (Items_Count => 1);
                Singleton_1.Items (1) := Var_Expr;
Scope_Result := Singleton_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2129;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Declarative_Part_Array_Type;
end if;
      Inc_Ref (Result_Var);
Concat_Result := Concat (Singleton, Result_Var); 
--# expr-done 2913
Let_Result := Concat_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 2914
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_790;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_790;
               Finalizer_Scope_2129;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Package_Decl_P_Declarative_Parts;
--# end
--# property-start BasePackageDecl.env_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9898
pragma Warnings (Off, "is not referenced");
 function Base_Package_Decl_P_Env_Names
  (Node : Bare_Base_Package_Decl
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Base_Package_Decl := Bare_Base_Package_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_791 with Inline_Always;
            procedure Finalizer_Scope_2130 with Inline_Always;
      Block_Var : String_Type;
Var_Expr : String_Type;
Fld : String_Type;
Sym : Symbol_Type;
Array_Lit : Symbol_Type_Array_Access;
Sym_1 : Symbol_Type;
Str : String_Type;
Concat_Result : String_Type;
Sym_2 : Symbol_Type;
Array_Lit_1 : Symbol_Type_Array_Access;
Fld_1 : Bare_Private_Part;
Is_Null : Boolean;
Not_Val : Boolean;
If_Result : Symbol_Type_Array_Access;
Scope_Result : Symbol_Type_Array_Access;
Result_Var : Symbol_Type_Array_Access;
Let_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_791 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_791;
            procedure Finalizer_Scope_2130 is
            begin
                     Dec_Ref (Var_Expr);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Str);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Array_Lit_1);
                     Dec_Ref (If_Result);
            end Finalizer_Scope_2130;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2927 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2915 '<FieldAccess for top_level_env_name at ast.py:9905>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9905
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.top_level_env_name
Fld := Libadalang.Implementation.Basic_Decl_P_Top_Level_Env_Name (Node => Self);
--# end
--# expr-done 2915
Block_Var := Fld; Inc_Ref (Block_Var);
--# bind fqn Block_Var
Var_Expr := Block_Var;
      Inc_Ref (Var_Expr);
if Var_Expr /= Empty_String then
   --# scope-start
   --# bind fqn Var_Expr
--# expr-start 2926 '<If at ast.py:9907>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9907
--# expr-start 2918 '<Not at ast.py:9908>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9908
--# expr-start 2917 '<IsNull at ast.py:9908>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9908
--# expr-start 2916 '<FieldAccess for private_part at ast.py:9908>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9908
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Self.Base_Package_Decl_F_Private_Part;
--# expr-done 2916
Is_Null := Fld_1 = null; 
--# expr-done 2917
Not_Val := not (Is_Null); 
--# expr-done 2918
if Not_Val then
   --# expr-start 2920 '<ArrayLiteral at ast.py:9909>' Array_Lit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9909
--# expr-start 2919 '<to_symbol at ast.py:9910>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9910
Sym := String_To_Symbol (Self.Unit.Context, Var_Expr); 
--# expr-done 2919
Array_Lit := Create_Symbol_Type_Array (Internal_Symbol_Type_Array'(1 => Sym)); 
--# expr-done 2920
   If_Result := Array_Lit;
else
   --# expr-start 2925 '<ArrayLiteral at ast.py:9912>' Array_Lit_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9912
--# expr-start 2921 '<to_symbol at ast.py:9913>' Sym_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9913
Sym_1 := String_To_Symbol (Self.Unit.Context, Var_Expr); 
--# expr-done 2921
--# expr-start 2924 '<to_symbol at ast.py:9914>' Sym_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9914
--# expr-start 2923 '<Concat at ast.py:9914>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9914
--# expr-start 2922 '<String at ast.py:9914>' Str /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9914
Str := Create_String (".__privatepart"); 
--# expr-done 2922
Concat_Result := Concat_String (Var_Expr, Str); 
--# expr-done 2923
Sym_2 := String_To_Symbol (Self.Unit.Context, Concat_Result); 
--# expr-done 2924
Array_Lit_1 := Create_Symbol_Type_Array (Internal_Symbol_Type_Array'(1 => Sym_1, 2 => Sym_2)); 
--# expr-done 2925
   If_Result := Array_Lit_1;
end if;
      Inc_Ref (If_Result);
--# expr-done 2926
Scope_Result := If_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2130;
   Result_Var := Scope_Result;
else
   Result_Var := No_Symbol_Type_Array_Type;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 2927
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_791;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_791;
               Finalizer_Scope_2130;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Package_Decl_P_Env_Names;
--# end
   --
   --  Primitives for Bare_Generic_Package_Internal
   --
      procedure Generic_Package_Internal_Pre_Env_Actions
        (Self            : Bare_Generic_Package_Internal;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_117 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            Libadalang.Implementation.Env_Names_118 (Node => 
Self);
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Generic_Package_Internal
        (Self : Bare_Generic_Package_Internal
         ; Base_Package_Decl_F_Package_Name : Bare_Defining_Name
         ; Base_Package_Decl_F_Aspects : Bare_Aspect_Spec
         ; Base_Package_Decl_F_Public_Part : Bare_Public_Part
         ; Base_Package_Decl_F_Private_Part : Bare_Private_Part
         ; Base_Package_Decl_F_End_Name : Bare_End_Name
        ) is
      begin
            Initialize_Fields_For_Base_Package_Decl
              (Self, Base_Package_Decl_F_Package_Name, Base_Package_Decl_F_Aspects, Base_Package_Decl_F_Public_Part, Base_Package_Decl_F_Private_Part, Base_Package_Decl_F_End_Name);
      end Initialize_Fields_For_Generic_Package_Internal;
--# property-start GenericPackageInternal.is_library_item /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10874
pragma Warnings (Off, "is not referenced");
 function Generic_Package_Internal_P_Is_Library_Item
  (Node : Bare_Generic_Package_Internal
  )
   return Boolean
is
   Self : Bare_Generic_Package_Internal := Bare_Generic_Package_Internal (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Fld_1 : Bare_Ada_Node;
Is_A : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2931 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2930 '<IsA LibraryItem at ast.py:10878>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10878
--# expr-start 2929 '<parent at ast.py:10878>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10878
--# expr-start 2928 '<parent at ast.py:10878>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10878
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 2928
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Parent; 
--# expr-done 2929
Is_A := Fld_1 /= null 
and then Fld_1.Kind in Ada_Library_Item_Range; 
--# expr-done 2930
Let_Result := Is_A; 
--# end
--# expr-done 2931
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Package_Internal_P_Is_Library_Item;
--# end
--# property-start GenericPackageInternal._env_trans_parent_117 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10880
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_117
  (Node : Bare_Generic_Package_Internal
  )
   return Boolean
is
   Self : Bare_Generic_Package_Internal := Bare_Generic_Package_Internal (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_117;
--# end
--# property-start GenericPackageInternal._env_names_118 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10880
pragma Warnings (Off, "is not referenced");
 function Env_Names_118
  (Node : Bare_Generic_Package_Internal
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Generic_Package_Internal := Bare_Generic_Package_Internal (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1535 with Inline_Always;
      Fld : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1535 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1535;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2932 '<FieldAccess for env_names at ast.py:10880>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10880
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasePackageDecl.env_names
Fld := Libadalang.Implementation.Base_Package_Decl_P_Env_Names (Node => Self);
--# end
--# expr-done 2932
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1535;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1535;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Names_118;
--# end
   --
   --  Primitives for Bare_Package_Decl
   --
      procedure Package_Decl_Pre_Env_Actions
        (Self            : Bare_Package_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Dummy : Boolean;
      begin
         Dummy := Libadalang.Implementation.Env_Do_48 (Node => 
Self);
      end;
      declare
         Env : constant Internal_Designated_Env :=
           Libadalang.Implementation.Initial_Env_49 (Node => 
Self);
      begin
         Set_Initial_Env
           (Self,
            State,
            Env,
            DSL_Location => "ast.py:9927");
      end;
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mappings : Internal_Env_Assoc_Array_Access :=
            Libadalang.Implementation.Env_Mappings_50 (Node => 
Self);
      begin
         for Mapping of Mappings.Items loop
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:9934");
         end loop;
         Dec_Ref (Mappings);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_51 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            Libadalang.Implementation.Env_Names_52 (Node => 
Self);
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      declare
         Dummy : Bare_Compilation_Unit_Array_Access;
      begin
         Dummy := Libadalang.Implementation.Env_Do_53 (Node => 
Self);
         Dec_Ref (Dummy);
      end;
      if Libadalang.Implementation.Ref_Cond_55 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_54 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Use_Package_Name_Designated_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_57 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_56 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Name_Designated_Type_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      end;
      procedure Initialize_Fields_For_Package_Decl
        (Self : Bare_Package_Decl
         ; Base_Package_Decl_F_Package_Name : Bare_Defining_Name
         ; Base_Package_Decl_F_Aspects : Bare_Aspect_Spec
         ; Base_Package_Decl_F_Public_Part : Bare_Public_Part
         ; Base_Package_Decl_F_Private_Part : Bare_Private_Part
         ; Base_Package_Decl_F_End_Name : Bare_End_Name
        ) is
      begin
            Initialize_Fields_For_Base_Package_Decl
              (Self, Base_Package_Decl_F_Package_Name, Base_Package_Decl_F_Aspects, Base_Package_Decl_F_Public_Part, Base_Package_Decl_F_Private_Part, Base_Package_Decl_F_End_Name);
      end Initialize_Fields_For_Package_Decl;
--# property-start PackageDecl._env_do_48 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9925
pragma Warnings (Off, "is not referenced");
 function Env_Do_48
  (Node : Bare_Package_Decl
  )
   return Boolean
is
   Self : Bare_Package_Decl := Bare_Package_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2933 '<FieldAccess for env_hook at ast.py:9925>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9925
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.env_hook
Fld := Libadalang.Implementation.Ada_Node_P_Env_Hook (Node => Self);
--# end
--# expr-done 2933
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_48;
--# end
--# property-start PackageDecl._initial_env_49 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9931
pragma Warnings (Off, "is not referenced");
 function Initial_Env_49
  (Node : Bare_Package_Decl
  )
   return Internal_Designated_Env
is
   Self : Bare_Package_Decl := Bare_Package_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_1466 with Inline_Always;
      Fld : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_1466 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1466;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2934 '<FieldAccess for child_decl_initial_env at ast.py:9931>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9931
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.child_decl_initial_env
Fld := Libadalang.Implementation.Basic_Decl_P_Child_Decl_Initial_Env (Node => Self, Private_Part => True);
--# end
--# expr-done 2934
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1466;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1466;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Initial_Env_49;
--# end
--# property-start PackageDecl._env_mappings_50 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9934
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_50
  (Node : Bare_Package_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Package_Decl := Bare_Package_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Package_Decl :=
     Internal_Entity_Package_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1467 with Inline_Always;
      Fld : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1467 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1467;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2935 '<FieldAccess for child_decl_env_assocs at ast.py:9934>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9934
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.child_decl_env_assocs
Fld := Libadalang.Implementation.Basic_Decl_P_Child_Decl_Env_Assocs (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 2935
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1467;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1467;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_50;
--# end
--# property-start PackageDecl._env_trans_parent_51 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9924
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_51
  (Node : Bare_Package_Decl
  )
   return Boolean
is
   Self : Bare_Package_Decl := Bare_Package_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_51;
--# end
--# property-start PackageDecl._env_names_52 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9936
pragma Warnings (Off, "is not referenced");
 function Env_Names_52
  (Node : Bare_Package_Decl
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Package_Decl := Bare_Package_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1469 with Inline_Always;
      Fld : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1469 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1469;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2936 '<FieldAccess for env_names at ast.py:9936>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9936
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasePackageDecl.env_names
Fld := Libadalang.Implementation.Base_Package_Decl_P_Env_Names (Node => Self);
--# end
--# expr-done 2936
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1469;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1469;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Names_52;
--# end
--# property-start PackageDecl._env_do_53 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9938
pragma Warnings (Off, "is not referenced");
 function Env_Do_53
  (Node : Bare_Package_Decl
  )
   return Bare_Compilation_Unit_Array_Access
is
   Self : Bare_Package_Decl := Bare_Package_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1470 with Inline_Always;
      Fld : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1470 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1470;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2937 '<FieldAccess for populate_dependent_units at ast.py:9938>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9938
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.populate_dependent_units
Fld := Libadalang.Implementation.Basic_Decl_P_Populate_Dependent_Units (Node => Self);
--# end
--# expr-done 2937
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1470;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1470;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_53;
--# end
--# property-start PackageDecl._ref_env_nodes_54 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9941
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_54
  (Node : Bare_Package_Decl
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Package_Decl := Bare_Package_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1471 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1471 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1471;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2938 '<FieldAccess for top_level_use_package_clauses at ast.py:9941>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9941
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_package_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Package_Clauses (Node => Self);
--# end
--# expr-done 2938
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1471;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1471;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_54;
--# end
--# property-start PackageDecl._ref_cond_55 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9943
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_55
  (Node : Bare_Package_Decl
  )
   return Boolean
is
   Self : Bare_Package_Decl := Bare_Package_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2940 '<IsA LibraryItem, Subunit at ast.py:9943>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9943
--# expr-start 2939 '<parent at ast.py:9943>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9943
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 2939
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 2940
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_55;
--# end
--# property-start PackageDecl._ref_env_nodes_56 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9947
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_56
  (Node : Bare_Package_Decl
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Package_Decl := Bare_Package_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1473 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1473 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1473;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2941 '<FieldAccess for top_level_use_type_clauses at ast.py:9947>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9947
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_type_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Type_Clauses (Node => Self);
--# end
--# expr-done 2941
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1473;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1473;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_56;
--# end
--# property-start PackageDecl._ref_cond_57 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9949
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_57
  (Node : Bare_Package_Decl
  )
   return Boolean
is
   Self : Bare_Package_Decl := Bare_Package_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2943 '<IsA LibraryItem, Subunit at ast.py:9949>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9949
--# expr-start 2942 '<parent at ast.py:9949>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9949
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 2942
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 2943
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_57;
--# end
   --
   --  Primitives for Bare_Base_Type_Decl
   --
      procedure Base_Type_Decl_Pre_Env_Actions
        (Self            : Bare_Base_Type_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_2 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:5083");
         Dec_Ref (Mapping.Dest_Env);
      end;
      end;
      procedure Initialize_Fields_For_Base_Type_Decl
        (Self : Bare_Base_Type_Decl
         ; Base_Type_Decl_F_Name : Bare_Defining_Name
        ) is
      begin
            Self.Base_Type_Decl_F_Name := Base_Type_Decl_F_Name;
      Self.Lf_Present_Dottable_Subps_Env := False;
      Self.Lf_Stg_Dottable_Subps_Env := Empty_Env;
      end Initialize_Fields_For_Base_Type_Decl;
   function Base_Type_Decl_F_Name
     (Node : Bare_Base_Type_Decl) return Bare_Defining_Name
   is
   begin
         return Node.Base_Type_Decl_F_Name;
   end;
--# property-start BaseTypeDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5085
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Defining_Names
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_364 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_364 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_364;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2944 '<FieldAccess for name at ast.py:5085>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5085
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Base_Type_Decl_F_Name, Info => Ent.Info);
--# expr-done 2944
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_364;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_364;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Defining_Names;
--# end
--# property-start BaseTypeDecl.base_subtype /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5090
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Base_Subtype
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Match_0 : Internal_Entity_Discrete_Base_Subtype_Decl;
Match_1 : Internal_Entity_Base_Subtype_Decl;
Match_2 : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Match_Prefix : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity_Discrete_Base_Subtype_Decl;
Cast_Expr : Internal_Entity_Discrete_Base_Subtype_Decl;
Cast_Result_1 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity_Base_Subtype_Decl;
Let_Result_1 : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity_Base_Type_Decl;
Let_Result_2 : Internal_Entity_Base_Type_Decl;
Scope_Result_2 : Internal_Entity_Base_Type_Decl;
Match_Result : Internal_Entity_Base_Type_Decl;
Let_Result_3 : Internal_Entity_Base_Type_Decl;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Type_Decl_Base_Type_Decl_P_Base_Subtype,
            Items    => new Mmz_Key_Array (1 ..  3))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Type_Decl,
                                As_Bare_Base_Type_Decl => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Bare_Ada_Node,
                                      As_Bare_Ada_Node => Origin);
               Mmz_K.Items (3) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Base_Type_Decl;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 2948 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 2947 '<Match at ast.py:5095>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5095
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Base_Type_Decl (Match_Prefix.Node.Kind) is
         when Ada_Discrete_Base_Subtype_Decl =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Discrete_Base_Subtype_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind db Match_0
Cast_Expr := Match_0; 
      Cast_Result_1 := Create_Internal_Entity_Base_Type_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Let_Result := Cast_Result_1; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Subtype_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Base_Subtype_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_2; 
--# bind st Match_1
--# expr-start 2946 '<FieldAccess for base_subtype at ast.py:5097>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5097
--# expr-start 2945 '<FieldAccess for get_type at ast.py:5097>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5097
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Match_1.Node, Origin => Origin, E_Info => Match_1.Info);
--# end
--# expr-done 2945
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.base_subtype
Fld_1 := Libadalang.Implementation.Base_Type_Decl_P_Base_Subtype (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 2946
Let_Result_1 := Fld_1; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Classwide_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Base_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_3; 
--# bind _ Match_2
Let_Result_2 := Ent; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 2947
Let_Result_3 := Match_Result; 
--# end
--# expr-done 2948
      Property_Result := Let_Result_3;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Base_Type_Decl,
                        As_Internal_Entity_Base_Type_Decl => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Base_Subtype;
--# end
--# property-start BaseTypeDecl.anonymous_access_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5102
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Anonymous_Access_Type
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Bare_Defining_Name;
New_Node : Bare_Not_Null_Absent;
Cast_Result : Bare_Not_Null;
New_Node_1 : Bare_Anonymous_Type_Access_Def;
Cast_Result_1 : Bare_Type_Def;
New_Node_2 : Bare_Synth_Anonymous_Type_Decl;
Cast_Result_2 : Bare_Base_Type_Decl;
As_Entity : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Type_Decl_Base_Type_Decl_P_Anonymous_Access_Type,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Type_Decl,
                                As_Bare_Base_Type_Decl => Self);
               Mmz_K.Items (2) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Base_Type_Decl;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 2955 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2954 '<as_entity at ???>' If_Result None
--# expr-start 2953 '<Cast to BaseTypeDecl at ???>' Cast_Result_2 None
--# expr-start 2952 '<New SynthAnonymousTypeDecl at ???>' New_Node_2 None
--# expr-start 2949 '<FieldAccess for name at ast.py:5104>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5104
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Base_Type_Decl_F_Name;
--# expr-done 2949
--# expr-start 2951 '<New AnonymousTypeAccessDef at ???>' New_Node_1 None
--# expr-start 2950 '<New NotNull.Absent at ???>' New_Node None
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node := new Root_Node_Record
  (Ada_Not_Null_Absent);
Initialize
  (Self => New_Node,
   Kind => Ada_Not_Null_Absent,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node);
--# expr-done 2950
      Cast_Result := New_Node;
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node_1 := new Root_Node_Record
  (Ada_Anonymous_Type_Access_Def);
Initialize
  (Self => New_Node_1,
   Kind => Ada_Anonymous_Type_Access_Def,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node_1);
   Initialize_Fields_For_Anonymous_Type_Access_Def
     (Self => New_Node_1,
      Access_Def_F_Has_Not_Null => Cast_Result, Anonymous_Type_Access_Def_F_Type_Decl => Self);
--# expr-done 2951
      Cast_Result_1 := New_Node_1;
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node_2 := new Root_Node_Record
  (Ada_Synth_Anonymous_Type_Decl);
Initialize
  (Self => New_Node_2,
   Kind => Ada_Synth_Anonymous_Type_Decl,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node_2);
   Initialize_Fields_For_Synth_Anonymous_Type_Decl
     (Self => New_Node_2,
      Base_Type_Decl_F_Name => Fld, Type_Decl_F_Aspects => No_Bare_Ada_Node, Type_Decl_F_Discriminants => No_Bare_Ada_Node, Type_Decl_F_Type_Def => Cast_Result_1);
--# expr-done 2952
      Cast_Result_2 := New_Node_2;
--# expr-done 2953
Is_Null := Cast_Result_2 = null; 
if Is_Null then
   If_Result := No_Entity_Base_Type_Decl;
else
As_Entity := (Info => E_Info, Node => Cast_Result_2); 
   If_Result := As_Entity;
end if;
--# expr-done 2954
Let_Result := If_Result; 
--# end
--# expr-done 2955
      Property_Result := Let_Result;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Base_Type_Decl,
                        As_Internal_Entity_Base_Type_Decl => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Anonymous_Access_Type;
--# end
--# property-start BaseTypeDecl.anonymous_access_type_or_null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5114
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Anonymous_Access_Type_Or_Null
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2957 '<Block at ???>' Let_Result None
--# scope-start
Var_Expr := Ent;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 2956 '<FieldAccess for anonymous_access_type at ast.py:5115>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5115
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.anonymous_access_type
Fld := Libadalang.Implementation.Base_Type_Decl_P_Anonymous_Access_Type (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 2956
Scope_Result := Fld;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 2957
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Anonymous_Access_Type_Or_Null;
--# end
--# property-start BaseTypeDecl.private_completion /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5120
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Private_Completion
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Bare_Ada_Node_List;
Var_Expr_1 : Bare_Private_Part;
Var_Expr_2 : Bare_Base_Package_Decl;
Var_Expr_3 : Bare_Public_Part;
Var_Expr_4 : Bare_Base_Type_Decl;
Fld : Bare_Declarative_Part;
Cast_Result : Bare_Public_Part;
Fld_1 : Bare_Ada_Node;
Scope_Result : Bare_Ada_Node;
Result_Var : Bare_Ada_Node;
Cast_Result_1 : Bare_Base_Package_Decl;
Fld_2 : Bare_Private_Part;
Scope_Result_1 : Bare_Private_Part;
Result_Var_1 : Bare_Private_Part;
Fld_3 : Bare_Ada_Node_List;
Scope_Result_2 : Bare_Ada_Node_List;
Result_Var_2 : Bare_Ada_Node_List;
Cast_Expr : Bare_Ada_Node;
Cast_Result_2 : Bare_Base_Type_Decl;
Fld_4 : Symbol_Type;
Fld_5 : Symbol_Type;
Is_Equal : Boolean;
Scope_Result_3 : Boolean;
Result_Var_3 : Boolean;
Item_127 : Bare_Ada_Node;
Find_Result : Bare_Ada_Node;
Scope_Result_4 : Bare_Ada_Node;
Result_Var_4 : Bare_Ada_Node;
Cast_Result_3 : Bare_Base_Type_Decl;
As_Entity : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Type_Decl_Base_Type_Decl_P_Private_Completion,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Type_Decl,
                                As_Bare_Base_Type_Decl => Self);
               Mmz_K.Items (2) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Base_Type_Decl;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 2971 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2970 '<as_entity at ast.py:5125>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5125
--# expr-start 2969 '<Cast to BaseTypeDecl at ast.py:5125>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5125
--# expr-start 2961 '<Cast to BasePackageDecl at ast.py:5125>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5125
--# expr-start 2959 '<Cast to PublicPart at ast.py:5125>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5125
--# expr-start 2958 '<FieldAccess for declarative_scope at ast.py:5125>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5125
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.declarative_scope
Fld := Libadalang.Implementation.Ada_Node_P_Declarative_Scope (Node => Ent.Node);
--# end
--# expr-done 2958
   if Fld = null
      or else Fld.Kind in Ada_Public_Part_Range
   then
      Cast_Result := Fld;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 2959
Var_Expr_3 := Cast_Result;
if Var_Expr_3 /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 2960 '<parent at ast.py:5125>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5125
      if Var_Expr_3 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Var_Expr_3.Parent; 
--# expr-done 2960
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node;
end if;
   if Result_Var = null
      or else Result_Var.Kind in Ada_Base_Package_Decl
   then
      Cast_Result_1 := Result_Var;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 2961
Var_Expr_2 := Cast_Result_1;
if Var_Expr_2 /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 2962 '<FieldAccess for private_part at ast.py:5125>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5125
      if Var_Expr_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Var_Expr_2.Base_Package_Decl_F_Private_Part;
--# expr-done 2962
Scope_Result_1 := Fld_2;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Bare_Ada_Node;
end if;
Var_Expr_1 := Result_Var_1;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 2963 '<FieldAccess for decls at ast.py:5125>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5125
      if Var_Expr_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Var_Expr_1.Declarative_Part_F_Decls;
--# expr-done 2963
Scope_Result_2 := Fld_3;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Bare_Ada_Node;
end if;
Var_Expr := Result_Var_2;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 2968 '<Find at ast.py:5125>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5125
      Find_Result := No_Bare_Ada_Node;
      if Var_Expr = null then
      Find_Result := No_Bare_Ada_Node;
      else
      declare
         Collection : constant Bare_Ada_Node_List := Var_Expr;
      begin
         for Untyped_Item_127 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_127 := Untyped_Item_127; 
   --# scope-start
               --# bind d Item_127
--# expr-start 2964 '<Cast to BaseTypeDecl at ast.py:5128>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5128
Cast_Expr := Item_127; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_2 := Cast_Expr;
   else
         Cast_Result_2 := No_Bare_Ada_Node;
   end if;
--# expr-done 2964
Var_Expr_4 := Cast_Result_2;
if Var_Expr_4 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind pp Var_Expr_4
--# expr-start 2967 '<Eq at ast.py:5130>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5130
--# expr-start 2965 '<FieldAccess for name_symbol at ast.py:5130>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5130
      if Var_Expr_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Var_Expr_4);
--# end
--# expr-done 2965
--# expr-start 2966 '<FieldAccess for name_symbol at ast.py:5130>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5130
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld_5 := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 2966
Is_Equal := Fld_4 = Fld_5; 
--# expr-done 2967
Scope_Result_3 := Is_Equal;
   --# end
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := False;
end if;
      if Result_Var_3 then
         Find_Result := Item_127;
      end if;
   --# end
               exit when Result_Var_3;
         end loop;
      end;
      end if;
--# expr-done 2968
Scope_Result_4 := Find_Result;
   --# end
   Result_Var_4 := Scope_Result_4;
else
   Result_Var_4 := No_Bare_Ada_Node;
end if;
   if Result_Var_4 = null
      or else Result_Var_4.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_3 := Result_Var_4;
   else
         Cast_Result_3 := No_Bare_Ada_Node;
   end if;
--# expr-done 2969
Is_Null := Cast_Result_3 = null; 
if Is_Null then
   If_Result := No_Entity_Base_Type_Decl;
else
As_Entity := (Info => E_Info, Node => Cast_Result_3); 
   If_Result := As_Entity;
end if;
--# expr-done 2970
Let_Result := If_Result; 
--# end
--# expr-done 2971
      Property_Result := Let_Result;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Base_Type_Decl,
                        As_Internal_Entity_Base_Type_Decl => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Private_Completion;
--# end
--# property-start BaseTypeDecl.is_view_of_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5136
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_View_Of_Type
  (Node : Bare_Base_Type_Decl
      ; Comp_View : Internal_Entity_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind comp_view Comp_View
   Property_Result : Boolean;
      Block_Var : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Is_Equiv : Boolean;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
If_Result : Boolean;
If_Result_1 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2977 '<Block at ???>' Let_Result None
--# scope-start
Block_Var := Ent; 
--# bind typ Block_Var
--# expr-start 2976 '<Cond at ast.py:5142>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5142
--# expr-start 2972 '<IsNull at ast.py:5143>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5143
Is_Null := Comp_View.Node = null; 
--# expr-done 2972
if Is_Null then
   If_Result_1 := False;
else
--# expr-start 2973 '<Eq at ast.py:5144>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5144
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Cast_Expr_1 := Comp_View; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Is_Equiv := Equivalent (Cast_Result, Cast_Result_1); 
--# expr-done 2973
if Is_Equiv then
   If_Result := True;
else
   --# expr-start 2975 '<FieldAccess for is_view_of_type at ast.py:5145>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5145
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 2974 '<FieldAccess for previous_part at ast.py:5145>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5145
      if Comp_View.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.previous_part'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Previous_Part (Node => Comp_View.Node, Go_To_Incomplete => True, E_Info => Comp_View.Info);
--# end
--# expr-done 2974
--# property-call-start BaseTypeDecl.is_view_of_type
Fld_1 := Libadalang.Implementation.Base_Type_Decl_P_Is_View_Of_Type (Node => Block_Var.Node, Comp_View => Fld, E_Info => Block_Var.Info);
--# end
--# expr-done 2975
   If_Result := Fld_1;
end if;
   If_Result_1 := If_Result;
end if;
--# expr-done 2976
Let_Result := If_Result_1; 
--# end
--# expr-done 2977
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_View_Of_Type;
--# end
--# property-start BaseTypeDecl.is_array_or_rec /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5149
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Array_Or_Rec
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Boolean;
Fld_1 : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2980 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2978 '<FieldAccess for is_array at ast.py:5150>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5150
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_array
Fld := Libadalang.Implementation.Basic_Decl_P_Is_Array (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 2978
if Fld then
   If_Result := True;
else
   --# expr-start 2979 '<FieldAccess for is_record_type at ast.py:5150>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5150
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_record_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Record_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 2979
   If_Result := Fld_1;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 2980
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Array_Or_Rec;
--# end
--# property-start BaseTypeDecl.is_inherited_primitive /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5153
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Inherited_Primitive
  (Node : Bare_Base_Type_Decl
      ; P : Internal_Entity_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind p P
   Property_Result : Boolean;
      Fld : Bare_Base_Type_Decl;
Fld_1 : Internal_Entity_Info;
Fld_2 : Internal_Metadata;
Fld_3 : Bare_Ada_Node;
Cast_Result : Bare_Ada_Node;
Is_Equal : Boolean;
Not_Val : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 2987 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2986 '<Not at ast.py:5158>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5158
--# expr-start 2985 '<Eq at ast.py:5158>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5158
--# expr-start 2981 '<FieldAccess for node at ast.py:5158>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5158
Fld := Ent.Node;
--# expr-done 2981
      Cast_Result := Fld;
--# expr-start 2984 '<FieldAccess for primitive at ast.py:5158>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5158
--# expr-start 2983 '<FieldAccess for md at ast.py:5158>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5158
--# expr-start 2982 '<FieldAccess for info at ast.py:5158>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5158
Fld_1 := P.Info;
--# expr-done 2982
Fld_2 := Fld_1.Md;
--# expr-done 2983
Fld_3 := Fld_2.Primitive;
--# expr-done 2984
Is_Equal := Cast_Result = Fld_3; 
--# expr-done 2985
Not_Val := not (Is_Equal); 
--# expr-done 2986
Let_Result := Not_Val; 
--# end
--# expr-done 2987
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Inherited_Primitive;
--# end
--# property-start BaseTypeDecl.get_record_representation_clause /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5162
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Get_Record_Representation_Clause
  (Node : Bare_Base_Type_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Record_Rep_Clause
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Record_Rep_Clause;
      Var_Expr : Bare_Declarative_Part;
Var_Expr_1 : Internal_Entity_Record_Rep_Clause;
Fld : Bare_Declarative_Part;
Fld_1 : Bare_Ada_Node_List;
Scope_Result : Bare_Ada_Node_List;
Result_Var : Bare_Ada_Node_List;
As_Entity : Internal_Entity_Ada_Node_List;
Is_Null : Boolean;
If_Result : Internal_Entity_Ada_Node_List;
Fld_2 : Bare_Ada_Node_List;
Fld_3 : Internal_Entity_Info;
As_Entity_1 : Internal_Entity;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Record_Rep_Clause;
Fld_4 : Internal_Entity_Name;
Fld_5 : Internal_Entity_Basic_Decl;
Cast_Result_1 : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Is_Equiv : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
Item_128 : Internal_Entity;
Bare_Item_128 : Bare_Ada_Node;
Find_Result : Internal_Entity;
Cast_Result_3 : Internal_Entity_Record_Rep_Clause;
Let_Result : Internal_Entity_Record_Rep_Clause;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 2997 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 2996 '<Cast to RecordRepClause.entity at ast.py:5167>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5167
--# expr-start 2995 '<Find at ast.py:5167>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5167
   --# expr-start 2990 '<as_entity at ast.py:5167>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5167
--# expr-start 2988 '<FieldAccess for declarative_scope at ast.py:5167>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5167
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.declarative_scope
Fld := Libadalang.Implementation.Ada_Node_P_Declarative_Scope (Node => Ent.Node);
--# end
--# expr-done 2988
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 2989 '<FieldAccess for decls at ast.py:5167>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5167
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Var_Expr.Declarative_Part_F_Decls;
--# expr-done 2989
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node;
end if;
Is_Null := Result_Var = null; 
if Is_Null then
   If_Result := No_Entity_Ada_Node_List;
else
As_Entity := (Info => E_Info, Node => Result_Var); 
   If_Result := As_Entity;
end if;
--# expr-done 2990
Fld_2 := If_Result.Node;
      Find_Result := No_Entity;
      if Fld_2 = null then
      Find_Result := No_Entity;
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_2;
      begin
         for Untyped_Item_128 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_128 := Untyped_Item_128; 
Is_Null_1 := Bare_Item_128 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity;
else
Fld_3 := If_Result.Info;
As_Entity_1 := (Info => Fld_3, Node => Bare_Item_128); 
   If_Result_1 := As_Entity_1;
end if;
                  Item_128 := If_Result_1; 
   --# scope-start
               --# bind d Item_128
--# expr-start 2991 '<Cast to RecordRepClause at ast.py:5168>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5168
Cast_Expr := Item_128; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Record_Rep_Clause_Range
   then
      Cast_Result := Create_Internal_Entity_Record_Rep_Clause
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Record_Rep_Clause;
   end if;
--# expr-done 2991
Var_Expr_1 := Cast_Result;
if Var_Expr_1 /= No_Entity_Record_Rep_Clause then
   --# scope-start
   --# bind p Var_Expr_1
--# expr-start 2994 '<Eq at ast.py:5169>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5169
--# expr-start 2993 '<FieldAccess for referenced_decl at ast.py:5169>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5169
--# expr-start 2992 '<FieldAccess for name at ast.py:5169>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5169
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Name (Node => Var_Expr_1.Node.Record_Rep_Clause_F_Name, Info => Var_Expr_1.Info);
--# expr-done 2992
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld_5 := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Fld_4.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_4.Info);
--# end
--# expr-done 2993
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
Cast_Expr_1 := Ent; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Is_Equiv := Equivalent (Cast_Result_1, Cast_Result_2); 
--# expr-done 2994
Scope_Result_1 := Is_Equiv;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
      if Result_Var_1 then
         Find_Result := Item_128;
      end if;
   --# end
               exit when Result_Var_1;
         end loop;
      end;
      end if;
--# expr-done 2995
   if Find_Result.Node = null
      or else Find_Result.Node.Kind in Ada_Record_Rep_Clause_Range
   then
      Cast_Result_3 := Create_Internal_Entity_Record_Rep_Clause
        (Node => Find_Result.Node,
         Info => Find_Result.Info);
   else
         Cast_Result_3 := No_Entity_Record_Rep_Clause;
   end if;
--# expr-done 2996
Let_Result := Cast_Result_3; 
--# end
--# expr-done 2997
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Get_Record_Representation_Clause;
--# end
--# property-start BaseTypeDecl.get_enum_representation_clause /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5175
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Get_Enum_Representation_Clause
  (Node : Bare_Base_Type_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Enum_Rep_Clause
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Enum_Rep_Clause;
      Var_Expr : Bare_Declarative_Part;
Var_Expr_1 : Internal_Entity_Enum_Rep_Clause;
Fld : Bare_Declarative_Part;
Fld_1 : Bare_Ada_Node_List;
Scope_Result : Bare_Ada_Node_List;
Result_Var : Bare_Ada_Node_List;
As_Entity : Internal_Entity_Ada_Node_List;
Is_Null : Boolean;
If_Result : Internal_Entity_Ada_Node_List;
Fld_2 : Bare_Ada_Node_List;
Fld_3 : Internal_Entity_Info;
As_Entity_1 : Internal_Entity;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Enum_Rep_Clause;
Fld_4 : Internal_Entity_Name;
Fld_5 : Internal_Entity_Basic_Decl;
Cast_Result_1 : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Is_Equiv : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
Item_129 : Internal_Entity;
Bare_Item_129 : Bare_Ada_Node;
Find_Result : Internal_Entity;
Cast_Result_3 : Internal_Entity_Enum_Rep_Clause;
Let_Result : Internal_Entity_Enum_Rep_Clause;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3007 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3006 '<Cast to EnumRepClause.entity at ast.py:5180>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5180
--# expr-start 3005 '<Find at ast.py:5180>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5180
   --# expr-start 3000 '<as_entity at ast.py:5180>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5180
--# expr-start 2998 '<FieldAccess for declarative_scope at ast.py:5180>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5180
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.declarative_scope
Fld := Libadalang.Implementation.Ada_Node_P_Declarative_Scope (Node => Ent.Node);
--# end
--# expr-done 2998
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 2999 '<FieldAccess for decls at ast.py:5180>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5180
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Var_Expr.Declarative_Part_F_Decls;
--# expr-done 2999
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node;
end if;
Is_Null := Result_Var = null; 
if Is_Null then
   If_Result := No_Entity_Ada_Node_List;
else
As_Entity := (Info => E_Info, Node => Result_Var); 
   If_Result := As_Entity;
end if;
--# expr-done 3000
Fld_2 := If_Result.Node;
      Find_Result := No_Entity;
      if Fld_2 = null then
      Find_Result := No_Entity;
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_2;
      begin
         for Untyped_Item_129 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_129 := Untyped_Item_129; 
Is_Null_1 := Bare_Item_129 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity;
else
Fld_3 := If_Result.Info;
As_Entity_1 := (Info => Fld_3, Node => Bare_Item_129); 
   If_Result_1 := As_Entity_1;
end if;
                  Item_129 := If_Result_1; 
   --# scope-start
               --# bind d Item_129
--# expr-start 3001 '<Cast to EnumRepClause at ast.py:5181>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5181
Cast_Expr := Item_129; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Enum_Rep_Clause_Range
   then
      Cast_Result := Create_Internal_Entity_Enum_Rep_Clause
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Enum_Rep_Clause;
   end if;
--# expr-done 3001
Var_Expr_1 := Cast_Result;
if Var_Expr_1 /= No_Entity_Enum_Rep_Clause then
   --# scope-start
   --# bind p Var_Expr_1
--# expr-start 3004 '<Eq at ast.py:5182>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5182
--# expr-start 3003 '<FieldAccess for referenced_decl at ast.py:5182>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5182
--# expr-start 3002 '<FieldAccess for type_name at ast.py:5182>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5182
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Name (Node => Var_Expr_1.Node.Enum_Rep_Clause_F_Type_Name, Info => Var_Expr_1.Info);
--# expr-done 3002
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld_5 := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Fld_4.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_4.Info);
--# end
--# expr-done 3003
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
Cast_Expr_1 := Ent; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Is_Equiv := Equivalent (Cast_Result_1, Cast_Result_2); 
--# expr-done 3004
Scope_Result_1 := Is_Equiv;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
      if Result_Var_1 then
         Find_Result := Item_129;
      end if;
   --# end
               exit when Result_Var_1;
         end loop;
      end;
      end if;
--# expr-done 3005
   if Find_Result.Node = null
      or else Find_Result.Node.Kind in Ada_Enum_Rep_Clause_Range
   then
      Cast_Result_3 := Create_Internal_Entity_Enum_Rep_Clause
        (Node => Find_Result.Node,
         Info => Find_Result.Info);
   else
         Cast_Result_3 := No_Entity_Enum_Rep_Clause;
   end if;
--# expr-done 3006
Let_Result := Cast_Result_3; 
--# end
--# expr-done 3007
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Get_Enum_Representation_Clause;
--# end
--# property-start '[dispatcher]BaseTypeDecl.primitives_env' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Primitives_Env
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Lexical_Env;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.primitives_env
                  Property_Result := Base_Type_Decl_P_Primitives_Env
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.primitives_env
                  Property_Result := Base_Subtype_Decl_P_Primitives_Env
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.primitives_env
                  Property_Result := Type_Decl_P_Primitives_Env
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Primitives_Env;
--# end
--# property-start '[dispatcher]BaseTypeDecl.is_record_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Is_Record_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Classwide_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start BaseTypeDecl.is_record_type
                  Property_Result := Base_Type_Decl_P_Is_Record_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.is_record_type
                  Property_Result := Base_Subtype_Decl_P_Is_Record_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Is_Record_Type;
--# end
--# property-start BaseTypeDecl.is_array_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5201
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Array_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3009 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3008 '<FieldAccess for is_array at ast.py:5205>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5205
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_array
Fld := Libadalang.Implementation.Basic_Decl_P_Is_Array (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3008
Let_Result := Fld; 
--# end
--# expr-done 3009
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Array_Type;
--# end
--# property-start BaseTypeDecl.find_derived_types /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5209
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Find_Derived_Types
  (Node : Bare_Base_Type_Decl
      ; Root : Internal_Entity
      ; Origin : Bare_Ada_Node
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Decl_Array_Access
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind root Root
   --# bind origin Origin
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Type_Decl_Array_Access;
            procedure Finalizer_Scope_377 with Inline_Always;
            procedure Finalizer_Scope_2147 with Inline_Always;
            procedure Finalizer_Scope_2149 with Inline_Always;
            procedure Finalizer_Scope_2150 with Inline_Always;
            procedure Finalizer_Scope_2151 with Inline_Always;
      Var_Expr : Internal_Entity_Array_Access;
Var_Expr_1 : Internal_Entity_Type_Decl;
Var_Expr_2 : Boolean;
Fld : Internal_Entity_Array_Access;
Is_Null : Boolean;
Fld_1 : Bare_Ada_Node;
Cast_Result : Bare_Ada_Node;
Is_Equal : Boolean;
If_Result : Boolean;
Not_Val : Boolean;
Map_Result : Internal_Entity_Array_Access;
Fld_2 : Internal_Entity_Type_Decl_Array_Access;
Map_Result_1 : Internal_Entity_Type_Decl_Array_Access;
Scope_Result : Internal_Entity_Type_Decl_Array_Access;
Result_Var : Internal_Entity_Type_Decl_Array_Access;
Cast_Expr : Internal_Entity;
Cast_Result_1 : Internal_Entity_Type_Decl;
Fld_3 : Boolean;
Singleton : Internal_Entity_Type_Decl_Array_Access;
Scope_Result_1 : Internal_Entity_Type_Decl_Array_Access;
Result_Var_1 : Internal_Entity_Type_Decl_Array_Access;
Scope_Result_2 : Internal_Entity_Type_Decl_Array_Access;
Result_Var_2 : Internal_Entity_Type_Decl_Array_Access;
Concat_Result : Internal_Entity_Type_Decl_Array_Access;
Let_Result : Internal_Entity_Type_Decl_Array_Access;
            procedure Finalizer_Scope_377 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Result_Var_2);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_377;
            procedure Finalizer_Scope_2147 is
            begin
                     Dec_Ref (Var_Expr);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Map_Result_1);
            end Finalizer_Scope_2147;
            procedure Finalizer_Scope_2149 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_2149;
            procedure Finalizer_Scope_2150 is
            begin
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
            end Finalizer_Scope_2150;
            procedure Finalizer_Scope_2151 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2151;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3021 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3020 '<Concat at ast.py:5214>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5214
--# expr-start 3010 '<children at ast.py:5214>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5214
      if Root.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children (Node => Root.Node, E_Info => Root.Info);
--# expr-done 3010
Var_Expr := Fld;
      Inc_Ref (Var_Expr);
if Var_Expr /= No_Internal_Entity_Array_Type then
   --# scope-start
   --# bind c Var_Expr
--# expr-start 3017 '<Mapcat at ast.py:5215>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5215
   --# expr-start 3015 '<Filter at ast.py:5215>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5215
      declare
         Map_Result_Vec : Internal_Entity_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Var_Expr;
      begin
         for Item_131 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_131
         --# expr-start 3014 '<Not at ast.py:5215>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5215
--# expr-start 3011 '<IsNull at ast.py:5215>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5215
Is_Null := Item_131.Node = null; 
--# expr-done 3011
if Is_Null then
   If_Result := True;
else
   --# expr-start 3013 '<Eq at ast.py:5215>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5215
--# expr-start 3012 '<FieldAccess for node at ast.py:5215>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5215
Fld_1 := Item_131.Node;
--# expr-done 3012
      Cast_Result := Fld_1;
Is_Equal := Cast_Result = Origin; 
--# expr-done 3013
   If_Result := Is_Equal;
end if;
Not_Val := not (If_Result); 
--# expr-done 3014
         if Not_Val then
         declare
            Item_To_Append : constant Internal_Entity := Item_131;
         begin
            Internal_Entity_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 3015
      declare
         Map_Result_1_Vec : Internal_Entity_Type_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Map_Result;
      begin
         for Item_130 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_130
      --# expr-start 3016 '<FieldAccess for find_derived_types at ast.py:5216>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5216
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.find_derived_types
Fld_2 := Libadalang.Implementation.Base_Type_Decl_P_Find_Derived_Types (Node => Ent.Node, Root => Item_130, Origin => Origin, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 3016
         for Item_To_Append of
               Fld_2.Items
         loop
            Internal_Entity_Type_Decl_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_2149;
         end loop;
      end;
         Map_Result_1 := Create_Internal_Entity_Type_Decl_Array
           (Items_Count => Natural (Internal_Entity_Type_Decl_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Entity_Type_Decl_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Entity_Type_Decl_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Entity_Type_Decl_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 3017
Scope_Result := Map_Result_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2147;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Type_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var);
--# expr-start 3018 '<Cast to TypeDecl at ast.py:5217>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5217
Cast_Expr := Root; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Type_Decl_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Type_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result_1 := No_Entity_Type_Decl;
   end if;
--# expr-done 3018
Var_Expr_1 := Cast_Result_1;
if Var_Expr_1 /= No_Entity_Type_Decl then
   --# scope-start
   --# bind type_decl Var_Expr_1
--# expr-start 3019 '<FieldAccess for is_derived_type at ast.py:5218>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5218
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_derived_type
Fld_3 := Libadalang.Implementation.Base_Type_Decl_P_Is_Derived_Type (Node => Var_Expr_1.Node, Other_Type => Ent, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 3019
Var_Expr_2 := Fld_3;
if Var_Expr_2 /= False then
   --# scope-start
   --# bind _ Var_Expr_2
                Singleton := Create_Internal_Entity_Type_Decl_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr_1;
Scope_Result_1 := Singleton;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2151;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Internal_Entity_Type_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var_1);
Scope_Result_2 := Result_Var_1;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2150;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Internal_Entity_Type_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var_2);
Concat_Result := Concat (Result_Var, Result_Var_2); 
--# expr-done 3020
Let_Result := Concat_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 3021
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_377;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_377;
               Finalizer_Scope_2147;
               Finalizer_Scope_2149;
               Finalizer_Scope_2150;
               Finalizer_Scope_2151;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Find_Derived_Types;
--# end
--# property-start '[dispatcher]BaseTypeDecl.is_task_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Is_Task_Type
  (Node : Bare_Base_Type_Decl
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Protected_Type_Decl | Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start BaseTypeDecl.is_task_type
                  Property_Result := Base_Type_Decl_P_Is_Task_Type
                    (Self
                    );
                  --# end
               when Ada_Task_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start TaskTypeDecl.is_task_type
                  Property_Result := Task_Type_Decl_P_Is_Task_Type
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Is_Task_Type;
--# end
--# property-start '[dispatcher]BaseTypeDecl.is_real_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Is_Real_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.is_real_type
                  Property_Result := Base_Type_Decl_P_Is_Real_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.is_real_type
                  Property_Result := Base_Subtype_Decl_P_Is_Real_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.is_real_type
                  Property_Result := Type_Decl_P_Is_Real_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Is_Real_Type;
--# end
--# property-start '[dispatcher]BaseTypeDecl.is_float_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Is_Float_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.is_float_type
                  Property_Result := Base_Type_Decl_P_Is_Float_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.is_float_type
                  Property_Result := Base_Subtype_Decl_P_Is_Float_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.is_float_type
                  Property_Result := Type_Decl_P_Is_Float_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Is_Float_Type;
--# end
--# property-start '[dispatcher]BaseTypeDecl.is_fixed_point' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Is_Fixed_Point
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.is_fixed_point
                  Property_Result := Base_Type_Decl_P_Is_Fixed_Point
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.is_fixed_point
                  Property_Result := Base_Subtype_Decl_P_Is_Fixed_Point
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.is_fixed_point
                  Property_Result := Type_Decl_P_Is_Fixed_Point
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Is_Fixed_Point;
--# end
--# property-start '[dispatcher]BaseTypeDecl.is_enum_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Is_Enum_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.is_enum_type
                  Property_Result := Base_Type_Decl_P_Is_Enum_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.is_enum_type
                  Property_Result := Base_Subtype_Decl_P_Is_Enum_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.is_enum_type
                  Property_Result := Type_Decl_P_Is_Enum_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Is_Enum_Type;
--# end
--# property-start '[dispatcher]BaseTypeDecl.is_classwide' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Is_Classwide
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Incomplete_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start BaseTypeDecl.is_classwide
                  Property_Result := Base_Type_Decl_P_Is_Classwide
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.is_classwide
                  Property_Result := Base_Subtype_Decl_P_Is_Classwide
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Classwide_Type_Decl =>
                  --# property-call-start ClasswideTypeDecl.is_classwide
                  Property_Result := Classwide_Type_Decl_P_Is_Classwide
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Is_Classwide;
--# end
--# property-start '[dispatcher]BaseTypeDecl.is_access_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Is_Access_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.is_access_type
                  Property_Result := Base_Type_Decl_P_Is_Access_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.is_access_type
                  Property_Result := Base_Subtype_Decl_P_Is_Access_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.is_access_type
                  Property_Result := Type_Decl_P_Is_Access_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Is_Access_Type;
--# end
--# property-start BaseTypeDecl.is_implicit_deref /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5251
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Implicit_Deref
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Boolean;
Fld_1 : Internal_Entity_Expr;
Is_Null : Boolean;
Not_Val : Boolean;
If_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
--# expr-start 3022 '<FieldAccess for is_access_type at ast.py:5252>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5252
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_access_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Access_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3022
if Fld then
   If_Result := True;
else
   --# expr-start 3025 '<Not at ast.py:5252>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5252
--# expr-start 3024 '<IsNull at ast.py:5252>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5252
--# expr-start 3023 '<FieldAccess for get_imp_deref at ast.py:5252>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5252
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.get_imp_deref'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Get_Imp_Deref (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3023
Is_Null := Fld_1.Node = null; 
--# expr-done 3024
Not_Val := not (Is_Null); 
--# expr-done 3025
   If_Result := Not_Val;
end if;
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Implicit_Deref;
--# end
--# property-start '[dispatcher]BaseTypeDecl.has_ud_indexing' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Has_Ud_Indexing
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.has_ud_indexing
                  Property_Result := Base_Type_Decl_P_Has_Ud_Indexing
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.has_ud_indexing
                  Property_Result := Base_Subtype_Decl_P_Has_Ud_Indexing
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.has_ud_indexing
                  Property_Result := Type_Decl_P_Has_Ud_Indexing
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Has_Ud_Indexing;
--# end
--# property-start '[dispatcher]BaseTypeDecl.constant_indexing_fns' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Constant_Indexing_Fns
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl_Array_Access
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Basic_Decl_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.constant_indexing_fns
                  Property_Result := Base_Type_Decl_P_Constant_Indexing_Fns
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.constant_indexing_fns
                  Property_Result := Base_Subtype_Decl_P_Constant_Indexing_Fns
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.constant_indexing_fns
                  Property_Result := Type_Decl_P_Constant_Indexing_Fns
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Constant_Indexing_Fns;
--# end
--# property-start '[dispatcher]BaseTypeDecl.variable_indexing_fns' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Variable_Indexing_Fns
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl_Array_Access
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Basic_Decl_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.variable_indexing_fns
                  Property_Result := Base_Type_Decl_P_Variable_Indexing_Fns
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.variable_indexing_fns
                  Property_Result := Base_Subtype_Decl_P_Variable_Indexing_Fns
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.variable_indexing_fns
                  Property_Result := Type_Decl_P_Variable_Indexing_Fns
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Variable_Indexing_Fns;
--# end
--# property-start '[dispatcher]BaseTypeDecl.get_imp_deref' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Get_Imp_Deref
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Expr
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.get_imp_deref
                  Property_Result := Base_Type_Decl_P_Get_Imp_Deref
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.get_imp_deref
                  Property_Result := Type_Decl_P_Get_Imp_Deref
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Get_Imp_Deref;
--# end
--# property-start BaseTypeDecl.derefed_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5283
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Derefed_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
Fld_1 : Internal_Entity_Expr;
Is_Null_1 : Boolean;
If_Result : Boolean;
If_Result_1 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3031 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3030 '<If at ast.py:5288>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5288
--# expr-start 3026 '<IsNull at ast.py:5289>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5289
Is_Null := Ent.Node = null; 
--# expr-done 3026
if Is_Null then
   If_Result := True;
else
   --# expr-start 3028 '<IsNull at ast.py:5289>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5289
--# expr-start 3027 '<FieldAccess for get_imp_deref at ast.py:5289>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5289
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.get_imp_deref'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Get_Imp_Deref (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3027
Is_Null_1 := Fld_1.Node = null; 
--# expr-done 3028
   If_Result := Is_Null_1;
end if;
if If_Result then
   If_Result_1 := Ent;
else
   --# expr-start 3029 '<FieldAccess for accessed_type at ast.py:5291>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5291
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3029
   If_Result_1 := Fld;
end if;
--# expr-done 3030
Let_Result := If_Result_1; 
--# end
--# expr-done 3031
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Derefed_Type;
--# end
--# property-start BaseTypeDecl.derefed_base_subtype /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5296
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Derefed_Base_Subtype
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Expr;
Is_Null : Boolean;
Not_Val : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Base_Type_Decl;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3040 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3039 '<If at ast.py:5302>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5302
--# expr-start 3032 '<IsNull at ast.py:5303>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5303
Is_Null_1 := Ent.Node = null; 
--# expr-done 3032
if Is_Null_1 then
   If_Result_1 := Ent;
else
   --# expr-start 3038 '<FieldAccess for base_subtype at ast.py:5305>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5305
--# expr-start 3037 '<If at ast.py:5305>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5305
--# expr-start 3035 '<Not at ast.py:5306>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5306
--# expr-start 3034 '<IsNull at ast.py:5306>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5306
--# expr-start 3033 '<FieldAccess for get_imp_deref at ast.py:5306>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5306
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.get_imp_deref'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Get_Imp_Deref (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3033
Is_Null := Fld_1.Node = null; 
--# expr-done 3034
Not_Val := not (Is_Null); 
--# expr-done 3035
if Not_Val then
   --# expr-start 3036 '<FieldAccess for accessed_type at ast.py:5307>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5307
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3036
   If_Result := Fld;
else
   If_Result := Ent;
end if;
--# expr-done 3037
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.base_subtype
Fld_2 := Libadalang.Implementation.Base_Type_Decl_P_Base_Subtype (Node => If_Result.Node, Origin => Origin, E_Info => If_Result.Info);
--# end
--# expr-done 3038
   If_Result_1 := Fld_2;
end if;
--# expr-done 3039
Let_Result := If_Result_1; 
--# end
--# expr-done 3040
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Derefed_Base_Subtype;
--# end
--# property-start BaseTypeDecl.one_non_universal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5313
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_One_Non_Universal
  (Node : Bare_Base_Type_Decl
      ; Second : Internal_Entity_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind second Second
   Property_Result : Boolean;
      Is_Null : Boolean;
Not_Val : Boolean;
Fld : Boolean;
If_Result : Boolean;
Is_Null_1 : Boolean;
Not_Val_1 : Boolean;
Fld_1 : Boolean;
If_Result_1 : Boolean;
If_Result_2 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3047 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3042 '<Not at ast.py:5315>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5315
--# expr-start 3041 '<IsNull at ast.py:5315>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5315
Is_Null := Ent.Node = null; 
--# expr-done 3041
Not_Val := not (Is_Null); 
--# expr-done 3042
if Not_Val then
   --# expr-start 3043 '<FieldAccess for is_not_universal_type at ast.py:5315>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5315
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_not_universal_type
Fld := Libadalang.Implementation.Base_Type_Decl_P_Is_Not_Universal_Type (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3043
   If_Result := Fld;
else
   If_Result := False;
end if;
if If_Result then
   If_Result_2 := True;
else
--# expr-start 3045 '<Not at ast.py:5316>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5316
--# expr-start 3044 '<IsNull at ast.py:5316>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5316
Is_Null_1 := Second.Node = null; 
--# expr-done 3044
Not_Val_1 := not (Is_Null_1); 
--# expr-done 3045
if Not_Val_1 then
   --# expr-start 3046 '<FieldAccess for is_not_universal_type at ast.py:5316>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5316
      if Second.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_not_universal_type
Fld_1 := Libadalang.Implementation.Base_Type_Decl_P_Is_Not_Universal_Type (Node => Second.Node, E_Info => Second.Info);
--# end
--# expr-done 3046
   If_Result_1 := Fld_1;
else
   If_Result_1 := False;
end if;
   If_Result_2 := If_Result_1;
end if;
Let_Result := If_Result_2; 
--# end
--# expr-done 3047
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_One_Non_Universal;
--# end
--# property-start BaseTypeDecl.non_universal_base_subtype /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5321
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Non_Universal_Base_Subtype
  (Node : Bare_Base_Type_Decl
      ; Second : Internal_Entity_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind second Second
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Block_Var : Boolean;
Block_Var_1 : Boolean;
First_St : Internal_Entity_Base_Type_Decl;
Second_St : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
Not_Val : Boolean;
Fld : Boolean;
If_Result : Boolean;
Is_Null_1 : Boolean;
Not_Val_1 : Boolean;
Fld_1 : Boolean;
If_Result_1 : Boolean;
If_Result_2 : Boolean;
Fld_2 : Internal_Entity_Base_Type_Decl;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Boolean;
If_Result_3 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
Fld_5 : Internal_Entity_Base_Type_Decl;
Fld_6 : Internal_Entity_Base_Type_Decl;
If_Result_4 : Internal_Entity_Base_Type_Decl;
If_Result_5 : Internal_Entity_Base_Type_Decl;
If_Result_6 : Internal_Entity_Base_Type_Decl;
Let_Result_1 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3062 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 3049 '<Not at ast.py:5322>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5322
--# expr-start 3048 '<IsNull at ast.py:5322>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5322
Is_Null := Ent.Node = null; 
--# expr-done 3048
Not_Val := not (Is_Null); 
--# expr-done 3049
if Not_Val then
   --# expr-start 3050 '<FieldAccess for is_not_universal_type at ast.py:5322>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5322
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_not_universal_type
Fld := Libadalang.Implementation.Base_Type_Decl_P_Is_Not_Universal_Type (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3050
   If_Result := Fld;
else
   If_Result := False;
end if;
Block_Var := If_Result; 
--# bind first_ok Block_Var
--# expr-start 3052 '<Not at ast.py:5323>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5323
--# expr-start 3051 '<IsNull at ast.py:5323>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5323
Is_Null_1 := Second.Node = null; 
--# expr-done 3051
Not_Val_1 := not (Is_Null_1); 
--# expr-done 3052
if Not_Val_1 then
   --# expr-start 3053 '<FieldAccess for is_not_universal_type at ast.py:5323>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5323
      if Second.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_not_universal_type
Fld_1 := Libadalang.Implementation.Base_Type_Decl_P_Is_Not_Universal_Type (Node => Second.Node, E_Info => Second.Info);
--# end
--# expr-done 3053
   If_Result_1 := Fld_1;
else
   If_Result_1 := False;
end if;
Block_Var_1 := If_Result_1; 
--# bind second_ok Block_Var_1
--# expr-start 3061 '<Cond at ast.py:5324>' If_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5324
if Block_Var then
   If_Result_2 := Block_Var_1;
else
   If_Result_2 := False;
end if;
if If_Result_2 then
   --# expr-start 3058 '<Let at ast.py:5326>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5326
--# scope-start
--# expr-start 3054 '<FieldAccess for derefed_base_subtype at ast.py:5328>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5328
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.derefed_base_subtype
Fld_2 := Libadalang.Implementation.Base_Type_Decl_P_Derefed_Base_Subtype (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3054
First_St := Fld_2; 
--# bind first_st First_St
--# expr-start 3055 '<FieldAccess for derefed_base_subtype at ast.py:5329>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5329
      if Second.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.derefed_base_subtype
Fld_3 := Libadalang.Implementation.Base_Type_Decl_P_Derefed_Base_Subtype (Node => Second.Node, Origin => Origin, E_Info => Second.Info);
--# end
--# expr-done 3055
Second_St := Fld_3; 
--# bind second_st Second_St
--# expr-start 3057 '<If at ast.py:5330>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5330
--# expr-start 3056 '<FieldAccess for matching_formal_type at ast.py:5331>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5331
      if Second_St.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_formal_type
Fld_4 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Type (Node => Second_St.Node, Formal_Type => First_St, Origin => Origin, E_Info => Second_St.Info);
--# end
--# expr-done 3056
if Fld_4 then
   If_Result_3 := First_St;
else
   If_Result_3 := Second_St;
end if;
--# expr-done 3057
Let_Result := If_Result_3; 
--# end
--# expr-done 3058
   If_Result_6 := Let_Result;
else
if Block_Var then
   --# expr-start 3059 '<FieldAccess for derefed_base_subtype at ast.py:5338>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5338
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.derefed_base_subtype
Fld_5 := Libadalang.Implementation.Base_Type_Decl_P_Derefed_Base_Subtype (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3059
   If_Result_5 := Fld_5;
else
if Block_Var_1 then
   --# expr-start 3060 '<FieldAccess for derefed_base_subtype at ast.py:5341>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5341
      if Second.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.derefed_base_subtype
Fld_6 := Libadalang.Implementation.Base_Type_Decl_P_Derefed_Base_Subtype (Node => Second.Node, Origin => Origin, E_Info => Second.Info);
--# end
--# expr-done 3060
   If_Result_4 := Fld_6;
else
   If_Result_4 := No_Entity_Base_Type_Decl;
end if;
   If_Result_5 := If_Result_4;
end if;
   If_Result_6 := If_Result_5;
end if;
--# expr-done 3061
Let_Result_1 := If_Result_6; 
--# end
--# expr-done 3062
      Property_Result := Let_Result_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Non_Universal_Base_Subtype;
--# end
--# property-start BaseTypeDecl.array_concat_result_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5348
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Array_Concat_Result_Type
  (Node : Bare_Base_Type_Decl
      ; Other : Internal_Entity_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind other Other
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Is_Null : Boolean;
Is_Null_1 : Boolean;
If_Result : Boolean;
Fld : Boolean;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Boolean;
If_Result_1 : Boolean;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Boolean;
Fld_5 : Internal_Entity_Base_Type_Decl;
Fld_6 : Boolean;
If_Result_2 : Boolean;
Fld_7 : Internal_Entity_Base_Type_Decl;
If_Result_3 : Internal_Entity_Base_Type_Decl;
If_Result_4 : Internal_Entity_Base_Type_Decl;
If_Result_5 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3074 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3073 '<Cond at ast.py:5354>' If_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5354
--# expr-start 3063 '<IsNull at ast.py:5355>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5355
Is_Null := Ent.Node = null; 
--# expr-done 3063
if Is_Null then
   If_Result := True;
else
   --# expr-start 3064 '<IsNull at ast.py:5355>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5355
Is_Null_1 := Other.Node = null; 
--# expr-done 3064
   If_Result := Is_Null_1;
end if;
if If_Result then
   If_Result_5 := No_Entity_Base_Type_Decl;
else
--# expr-start 3065 '<FieldAccess for matching_formal_type at ast.py:5358>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5358
      if Other.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_formal_type
Fld := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Type (Node => Other.Node, Formal_Type => Ent, Origin => Origin, E_Info => Other.Info);
--# end
--# expr-done 3065
if Fld then
   If_Result_1 := True;
else
   --# expr-start 3067 '<FieldAccess for matching_formal_type at ast.py:5359>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5359
      if Other.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3066 '<FieldAccess for comp_type at ast.py:5359>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5359
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.comp_type
Fld_1 := Libadalang.Implementation.Base_Type_Decl_P_Comp_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3066
--# property-call-start BaseTypeDecl.matching_formal_type
Fld_2 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Type (Node => Other.Node, Formal_Type => Fld_1, Origin => Origin, E_Info => Other.Info);
--# end
--# expr-done 3067
   If_Result_1 := Fld_2;
end if;
if If_Result_1 then
   --# expr-start 3068 '<FieldAccess for derefed_base_subtype at ast.py:5360>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5360
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.derefed_base_subtype
Fld_3 := Libadalang.Implementation.Base_Type_Decl_P_Derefed_Base_Subtype (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3068
   If_Result_4 := Fld_3;
else
--# expr-start 3069 '<FieldAccess for matching_formal_type at ast.py:5362>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5362
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_formal_type
Fld_4 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Type (Node => Ent.Node, Formal_Type => Other, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3069
if Fld_4 then
   If_Result_2 := True;
else
   --# expr-start 3071 '<FieldAccess for matching_formal_type at ast.py:5363>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5363
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3070 '<FieldAccess for comp_type at ast.py:5363>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5363
      if Other.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.comp_type
Fld_5 := Libadalang.Implementation.Base_Type_Decl_P_Comp_Type (Node => Other.Node, Origin => Origin, E_Info => Other.Info);
--# end
--# expr-done 3070
--# property-call-start BaseTypeDecl.matching_formal_type
Fld_6 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Type (Node => Ent.Node, Formal_Type => Fld_5, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3071
   If_Result_2 := Fld_6;
end if;
if If_Result_2 then
   --# expr-start 3072 '<FieldAccess for derefed_base_subtype at ast.py:5364>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5364
      if Other.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.derefed_base_subtype
Fld_7 := Libadalang.Implementation.Base_Type_Decl_P_Derefed_Base_Subtype (Node => Other.Node, Origin => Origin, E_Info => Other.Info);
--# end
--# expr-done 3072
   If_Result_3 := Fld_7;
else
   If_Result_3 := No_Entity_Base_Type_Decl;
end if;
   If_Result_4 := If_Result_3;
end if;
   If_Result_5 := If_Result_4;
end if;
--# expr-done 3073
Let_Result := If_Result_5; 
--# end
--# expr-done 3074
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Array_Concat_Result_Type;
--# end
--# property-start BaseTypeDecl.expected_array_concat_operand_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5371
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Expected_Array_Concat_Operand_Type
  (Node : Bare_Base_Type_Decl
      ; Operand_Type : Internal_Entity_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind operand_type Operand_Type
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Is_Null : Boolean;
Is_Null_1 : Boolean;
If_Result : Boolean;
Fld : Boolean;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Boolean;
If_Result_1 : Boolean;
Fld_3 : Internal_Entity_Base_Type_Decl;
If_Result_2 : Internal_Entity_Base_Type_Decl;
If_Result_3 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3082 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3081 '<Cond at ast.py:5380>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5380
--# expr-start 3075 '<IsNull at ast.py:5381>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5381
Is_Null := Ent.Node = null; 
--# expr-done 3075
if Is_Null then
   If_Result := True;
else
   --# expr-start 3076 '<IsNull at ast.py:5381>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5381
Is_Null_1 := Operand_Type.Node = null; 
--# expr-done 3076
   If_Result := Is_Null_1;
end if;
if If_Result then
   If_Result_3 := No_Entity_Base_Type_Decl;
else
--# expr-start 3077 '<FieldAccess for matching_formal_type at ast.py:5384>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5384
      if Operand_Type.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_formal_type
Fld := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Type (Node => Operand_Type.Node, Formal_Type => Ent, Origin => Origin, E_Info => Operand_Type.Info);
--# end
--# expr-done 3077
if Fld then
   If_Result_1 := True;
else
   --# expr-start 3079 '<FieldAccess for matching_formal_type at ast.py:5385>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5385
      if Operand_Type.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3078 '<FieldAccess for comp_type at ast.py:5385>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5385
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.comp_type
Fld_1 := Libadalang.Implementation.Base_Type_Decl_P_Comp_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3078
--# property-call-start BaseTypeDecl.matching_formal_type
Fld_2 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Type (Node => Operand_Type.Node, Formal_Type => Fld_1, Origin => Origin, E_Info => Operand_Type.Info);
--# end
--# expr-done 3079
   If_Result_1 := Fld_2;
end if;
if If_Result_1 then
   --# expr-start 3080 '<FieldAccess for derefed_base_subtype at ast.py:5386>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5386
      if Operand_Type.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.derefed_base_subtype
Fld_3 := Libadalang.Implementation.Base_Type_Decl_P_Derefed_Base_Subtype (Node => Operand_Type.Node, Origin => Origin, E_Info => Operand_Type.Info);
--# end
--# expr-done 3080
   If_Result_2 := Fld_3;
else
   If_Result_2 := No_Entity_Base_Type_Decl;
end if;
   If_Result_3 := If_Result_2;
end if;
--# expr-done 3081
Let_Result := If_Result_3; 
--# end
--# expr-done 3082
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Expected_Array_Concat_Operand_Type;
--# end
--# property-start '[dispatcher]BaseTypeDecl.access_def' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Access_Def
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Access_Def
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Access_Def;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.access_def
                  Property_Result := Base_Type_Decl_P_Access_Def
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.access_def
                  Property_Result := Base_Subtype_Decl_P_Access_Def
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.access_def
                  Property_Result := Type_Decl_P_Access_Def
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Access_Def;
--# end
--# property-start '[dispatcher]BaseTypeDecl.is_char_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Is_Char_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.is_char_type
                  Property_Result := Base_Type_Decl_P_Is_Char_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.is_char_type
                  Property_Result := Base_Subtype_Decl_P_Is_Char_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.is_char_type
                  Property_Result := Type_Decl_P_Is_Char_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Is_Char_Type;
--# end
--# property-start BaseTypeDecl.is_non_null_char_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5400
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Non_Null_Char_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Is_Null : Boolean;
Not_Val : Boolean;
Fld : Boolean;
If_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
--# expr-start 3084 '<Not at ast.py:5401>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5401
--# expr-start 3083 '<IsNull at ast.py:5401>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5401
Is_Null := Self = null; 
--# expr-done 3083
Not_Val := not (Is_Null); 
--# expr-done 3084
if Not_Val then
   --# expr-start 3085 '<FieldAccess for is_char_type at ast.py:5401>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5401
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_char_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Char_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3085
   If_Result := Fld;
else
   If_Result := False;
end if;
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Non_Null_Char_Type;
--# end
--# property-start '[dispatcher]BaseTypeDecl.classwide_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Classwide_Type
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Classwide_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Classwide_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl | Ada_Incomplete_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start BaseTypeDecl.classwide_type
                  Property_Result := Base_Type_Decl_P_Classwide_Type
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Classwide_Type_Decl =>
                  --# property-call-start ClasswideTypeDecl.classwide_type
                  Property_Result := Classwide_Type_Decl_P_Classwide_Type
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Classwide_Type;
--# end
--# property-start BaseTypeDecl.scalar_base_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5412
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Scalar_Base_Type
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Discrete_Base_Subtype_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Discrete_Base_Subtype_Decl;
      Fld : Bare_Discrete_Base_Subtype_Decl;
As_Entity : Internal_Entity_Discrete_Base_Subtype_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Discrete_Base_Subtype_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3087 '<as_entity at ast.py:5412>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5412
--# expr-start 3086 '<FieldAccess for scalar_base_subtype_node at ast.py:5412>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5412
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.scalar_base_subtype_node
Fld := Libadalang.Implementation.Base_Type_Decl_P_Scalar_Base_Subtype_Node (Node => Self);
--# end
--# expr-done 3086
Is_Null := Fld = null; 
if Is_Null then
   If_Result := No_Entity_Discrete_Base_Subtype_Decl;
else
As_Entity := (Info => E_Info, Node => Fld); 
   If_Result := As_Entity;
end if;
--# expr-done 3087
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Scalar_Base_Type;
--# end
--# property-start BaseTypeDecl.is_universal_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5415
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Universal_Type
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Is_Null : Boolean;
Not_Val : Boolean;
Fld : Internal_Entity;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Is_Equiv : Boolean;
Fld_1 : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Is_Equiv_1 : Boolean;
If_Result : Boolean;
If_Result_1 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3094 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3089 '<Not at ast.py:5423>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5423
--# expr-start 3088 '<IsNull at ast.py:5423>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5423
Is_Null := Ent.Node = null; 
--# expr-done 3088
Not_Val := not (Is_Null); 
--# expr-done 3089
if Not_Val then
--# expr-start 3091 '<Eq at ast.py:5424>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5424
Cast_Expr := Ent; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
--# expr-start 3090 '<FieldAccess for universal_int_type at ast.py:5424>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5424
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.universal_int_type
Fld := Libadalang.Implementation.Ada_Node_P_Universal_Int_Type (Node => Self);
--# end
--# expr-done 3090
Is_Equiv := Equivalent (Cast_Result, Fld); 
--# expr-done 3091
if Is_Equiv then
   If_Result := True;
else
   --# expr-start 3093 '<Eq at ast.py:5425>' Is_Equiv_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5425
Cast_Expr_1 := Ent; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
--# expr-start 3092 '<FieldAccess for universal_real_type at ast.py:5425>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5425
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.universal_real_type
Fld_1 := Libadalang.Implementation.Ada_Node_P_Universal_Real_Type (Node => Self);
--# end
--# expr-done 3092
Is_Equiv_1 := Equivalent (Cast_Result_1, Fld_1); 
--# expr-done 3093
   If_Result := Is_Equiv_1;
end if;
   If_Result_1 := If_Result;
else
   If_Result_1 := False;
end if;
Let_Result := If_Result_1; 
--# end
--# expr-done 3094
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Universal_Type;
--# end
--# property-start BaseTypeDecl.is_not_universal_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5429
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Not_Universal_Type
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Is_Null : Boolean;
Not_Val : Boolean;
Fld : Boolean;
Not_Val_1 : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3099 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3096 '<Not at ast.py:5437>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5437
--# expr-start 3095 '<IsNull at ast.py:5437>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5437
Is_Null := Ent.Node = null; 
--# expr-done 3095
Not_Val := not (Is_Null); 
--# expr-done 3096
if Not_Val then
   --# expr-start 3098 '<Not at ast.py:5437>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5437
--# expr-start 3097 '<FieldAccess for is_universal_type at ast.py:5437>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5437
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_universal_type
Fld := Libadalang.Implementation.Base_Type_Decl_P_Is_Universal_Type (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3097
Not_Val_1 := not (Fld); 
--# expr-done 3098
   If_Result := Not_Val_1;
else
   If_Result := False;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 3099
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Not_Universal_Type;
--# end
--# property-start BaseTypeDecl.is_access_type_predicate /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5440
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Access_Type_Predicate
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3101 '<Block at ???>' Let_Result None
--# scope-start
Var_Expr := Ent;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 3100 '<FieldAccess for is_access_type at ast.py:5446>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5446
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_access_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Access_Type (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 3100
Scope_Result := Fld;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 3101
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Access_Type_Predicate;
--# end
--# property-start BaseTypeDecl.array_ndims /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5449
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Array_Ndims
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Integer
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Integer;
      Let_Result : Integer;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 3102 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := 0; 
--# end
--# expr-done 3102
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Array_Ndims;
--# end
--# property-start '[dispatcher]BaseTypeDecl.discrete_range' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Discrete_Range
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Discrete_Range
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Discrete_Range;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Discrete_Base_Subtype_Decl | Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.discrete_range
                  Property_Result := Base_Type_Decl_P_Discrete_Range
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Subtype_Decl =>
                  --# property-call-start SubtypeDecl.discrete_range
                  Property_Result := Subtype_Decl_P_Discrete_Range
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.discrete_range
                  Property_Result := Type_Decl_P_Discrete_Range
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Discrete_Range;
--# end
--# property-start BaseTypeDecl.static_predicate /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5461
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Static_Predicate
  (Node : Bare_Base_Type_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Expr
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Expr;
      Var_Expr : Internal_Entity_Expr;
Sym : Symbol_Type;
Fld : Internal_Aspect;
Fld_1 : Internal_Entity_Expr;
Scope_Result : Internal_Entity_Expr;
Sym_1 : Symbol_Type;
Fld_2 : Internal_Aspect;
Fld_3 : Internal_Entity_Expr;
Result_Var : Internal_Entity_Expr;
Let_Result : Internal_Entity_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3109 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3105 '<FieldAccess for value at ast.py:5466>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5466
--# expr-start 3104 '<FieldAccess for get_aspect at ast.py:5466>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5466
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3103 '<Symbol Static_Predicate at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Static_Predicate); 
--# expr-done 3103
--# property-call-start BasicDecl.get_aspect
Fld := Libadalang.Implementation.Basic_Decl_P_Get_Aspect (Node => Ent.Node, Name => Sym, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 3104
Fld_1 := Fld.Value;
--# expr-done 3105
Var_Expr := Fld_1;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind e Var_Expr
Scope_Result := Var_Expr;
   --# end
   Result_Var := Scope_Result;
else
   --# expr-start 3108 '<FieldAccess for value at ast.py:5467>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5467
--# expr-start 3107 '<FieldAccess for get_aspect at ast.py:5467>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5467
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3106 '<Symbol Predicate at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Predicate); 
--# expr-done 3106
--# property-call-start BasicDecl.get_aspect
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Get_Aspect (Node => Ent.Node, Name => Sym_1, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 3107
Fld_3 := Fld_2.Value;
--# expr-done 3108
   Result_Var := Fld_3;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 3109
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Static_Predicate;
--# end
--# property-start BaseTypeDecl.satisfies_type_predicates /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5473
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Satisfies_Type_Predicates
  (Node : Bare_Base_Type_Decl
      ; Value : Big_Integer_Type
      ; Imprecise_Fallback : Boolean
         := False
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind value Value
   --# bind imprecise_fallback Imprecise_Fallback
   --# bind origin Origin
   Property_Result : Boolean;
            procedure Finalizer_Scope_407 with Inline_Always;
            procedure Finalizer_Scope_2154 with Inline_Always;
      Block_Var : Big_Integer_Type;
Block_Var_1 : Boolean;
Block_Var_2 : Internal_Entity_Base_Type_Decl;
Block_Var_3 : Internal_Entity_Base_Type_Decl;
Var_Expr : Internal_Entity_Expr;
Var_Expr_1 : Internal_Entity_Base_Subtype_Decl;
Big_Int : Big_Integer_Type;
Fld : Internal_Entity_Expr;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity_Basic_Decl;
New_Struct : Internal_Substitution := No_Substitution;
Singleton : Internal_Substitution_Array_Access;
Fld_1 : Big_Integer_Type;
Is_Equal : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity_Base_Subtype_Decl;
Fld_2 : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Result_Var_1 : Internal_Entity_Base_Type_Decl;
Fld_3 : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
Not_Val : Boolean;
Fld_4 : Boolean;
Is_Null_1 : Boolean;
Not_Val_1 : Boolean;
Fld_5 : Boolean;
If_Result : Boolean;
If_Result_1 : Boolean;
If_Result_2 : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_407 is
            begin
                     Dec_Ref (Big_Int);
                     Dec_Ref (Block_Var);
            end Finalizer_Scope_407;
            procedure Finalizer_Scope_2154 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (Singleton);
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2154;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3125 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3110 '<BigInteger 1 at ast.py:5480>' Big_Int /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5480
Big_Int := Create_Big_Integer ("1"); 
--# expr-done 3110
Block_Var := Big_Int; Inc_Ref (Block_Var);
--# bind true_val Block_Var
--# expr-start 3111 '<FieldAccess for static_predicate at ast.py:5483>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5483
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.static_predicate
Fld := Libadalang.Implementation.Base_Type_Decl_P_Static_Predicate (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 3111
Var_Expr := Fld;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind pred Var_Expr
--# expr-start 3114 '<Eq at ast.py:5484>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5484
--# expr-start 3113 '<FieldAccess for eval_as_int_in_env at ast.py:5484>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5484
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3112 '<New Substitution at ast.py:5485>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5485
Cast_Expr := Ent; 
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Inc_Ref (Value);
New_Struct := (From_Decl => Cast_Result, To_Value => Value, Value_Type => Ent); 
--# expr-done 3112
                Singleton := Create_Internal_Substitution_Array (Items_Count => 1);
                Singleton.Items (1) := New_Struct;
                Inc_Ref (Singleton.Items (1));
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Eval_As_Int_In_Env (Node => Var_Expr.Node, Env => Singleton, E_Info => Var_Expr.Info);
--# expr-done 3113
Is_Equal := Equivalent (Block_Var, Fld_1); 
--# expr-done 3114
Scope_Result := Is_Equal;
   --# end
      Finalizer_Scope_2154;
   Result_Var := Scope_Result;
else
   Result_Var := True;
end if;
Block_Var_1 := Result_Var; 
--# bind satisfies_own_predicate Block_Var_1
--# expr-start 3115 '<Cast to BaseSubtypeDecl at ast.py:5495>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5495
Cast_Expr_1 := Ent; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Base_Subtype_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Base_Subtype_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_1 := No_Entity_Base_Subtype_Decl;
   end if;
--# expr-done 3115
Var_Expr_1 := Cast_Result_1;
if Var_Expr_1 /= No_Entity_Base_Subtype_Decl then
   --# scope-start
--# expr-start 3116 '<FieldAccess for get_type at ast.py:5495>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5495
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Var_Expr_1.Node, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 3116
Scope_Result_1 := Fld_2;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Base_Type_Decl;
end if;
Block_Var_2 := Result_Var_1; 
--# bind from_type Block_Var_2
--# expr-start 3117 '<FieldAccess for base_type at ast.py:5496>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5496
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.base_type'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Base_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3117
Block_Var_3 := Fld_3; 
--# bind base_type Block_Var_3
if Block_Var_1 then
   --# expr-start 3124 '<Cond at ast.py:5498>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5498
--# expr-start 3119 '<Not at ast.py:5499>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5499
--# expr-start 3118 '<IsNull at ast.py:5499>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5499
Is_Null := Block_Var_2.Node = null; 
--# expr-done 3118
Not_Val := not (Is_Null); 
--# expr-done 3119
if Not_Val then
   --# expr-start 3120 '<FieldAccess for satisfies_type_predicates at ast.py:5500>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5500
      if Block_Var_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.satisfies_type_predicates
Fld_4 := Libadalang.Implementation.Base_Type_Decl_P_Satisfies_Type_Predicates (Node => Block_Var_2.Node, Value => Value, Imprecise_Fallback => Imprecise_Fallback, Origin => Origin, E_Info => Block_Var_2.Info);
--# end
--# expr-done 3120
   If_Result_1 := Fld_4;
else
--# expr-start 3122 '<Not at ast.py:5502>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5502
--# expr-start 3121 '<IsNull at ast.py:5502>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5502
Is_Null_1 := Block_Var_3.Node = null; 
--# expr-done 3121
Not_Val_1 := not (Is_Null_1); 
--# expr-done 3122
if Not_Val_1 then
   --# expr-start 3123 '<FieldAccess for satisfies_type_predicates at ast.py:5503>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5503
      if Block_Var_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.satisfies_type_predicates
Fld_5 := Libadalang.Implementation.Base_Type_Decl_P_Satisfies_Type_Predicates (Node => Block_Var_3.Node, Value => Value, Imprecise_Fallback => Imprecise_Fallback, Origin => Origin, E_Info => Block_Var_3.Info);
--# end
--# expr-done 3123
   If_Result := Fld_5;
else
   If_Result := True;
end if;
   If_Result_1 := If_Result;
end if;
--# expr-done 3124
   If_Result_2 := If_Result_1;
else
   If_Result_2 := False;
end if;
Let_Result := If_Result_2; 
--# end
--# expr-done 3125
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_407;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_407;
               Finalizer_Scope_2154;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Satisfies_Type_Predicates;
--# end
--# property-start BaseTypeDecl.is_iterator_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5509
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Iterator_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
            procedure Finalizer_Scope_408 with Inline_Always;
      Block_Var : Bare_Basic_Decl;
Block_Var_1 : Internal_Entity_Base_Type_Decl;
Var_Expr : Internal_Entity_Classwide_Type_Decl;
Sym : Symbol_Type;
Sym_1 : Symbol_Type;
Array_Lit : Symbol_Type_Array_Access;
Fld : Bare_Basic_Decl;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity_Classwide_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity;
Fld_3 : Internal_Entity;
Fld_4 : Bare_Ada_Node;
Cast_Expr_1 : Bare_Basic_Decl;
Cast_Result_1 : Bare_Ada_Node;
Is_Equal : Boolean;
Fld_5 : Internal_Entity_Basic_Decl;
Sym_2 : Symbol_Type;
Fld_6 : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_408 is
            begin
                     Dec_Ref (Array_Lit);
            end Finalizer_Scope_408;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Type_Decl_Base_Type_Decl_P_Is_Iterator_Type,
            Items    => new Mmz_Key_Array (1 ..  3))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Type_Decl,
                                As_Bare_Base_Type_Decl => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Bare_Ada_Node,
                                      As_Bare_Ada_Node => Origin);
               Mmz_K.Items (3) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Boolean;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 3139 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3129 '<FieldAccess for get_unit_root_decl at ast.py:5510>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5510
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3128 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 3126 '<Symbol Ada at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Ada); 
--# expr-done 3126
--# expr-start 3127 '<Symbol Iterator_Interfaces at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Iterator_Interfaces); 
--# expr-done 3127
Array_Lit := Create_Symbol_Type_Array (Internal_Symbol_Type_Array'(1 => Sym, 2 => Sym_1)); 
--# expr-done 3128
--# property-call-start AdaNode.get_unit_root_decl
Fld := Libadalang.Implementation.Ada_Node_P_Get_Unit_Root_Decl (Node => Ent.Node, Name => Array_Lit, Kind => Unit_Specification);
--# end
--# expr-done 3129
Block_Var := Fld; 
--# bind iifcs Block_Var
--# expr-start 3130 '<Cast to ClasswideTypeDecl at ast.py:5513>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5513
Cast_Expr := Ent; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Classwide_Type_Decl_Range
   then
      Cast_Result := Create_Internal_Entity_Classwide_Type_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Classwide_Type_Decl;
   end if;
--# expr-done 3130
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Classwide_Type_Decl then
   --# scope-start
   --# bind cw Var_Expr
--# expr-start 3131 '<FieldAccess for typedecl at ast.py:5514>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5514
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ClasswideTypeDecl.typedecl
Fld_1 := Libadalang.Implementation.Classwide_Type_Decl_P_Typedecl (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 3131
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := Ent;
end if;
Block_Var_1 := Result_Var; 
--# bind typ Block_Var_1
--# expr-start 3135 '<Eq at ast.py:5517>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5517
--# expr-start 3134 '<FieldAccess for node at ast.py:5517>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5517
--# expr-start 3133 '<FieldAccess for semantic_parent at ast.py:5517>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5517
--# expr-start 3132 '<FieldAccess for semantic_parent at ast.py:5517>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5517
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.semantic_parent
Fld_2 := Libadalang.Implementation.Ada_Node_P_Semantic_Parent (Node => Block_Var_1.Node, E_Info => Block_Var_1.Info);
--# end
--# expr-done 3132
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.semantic_parent
Fld_3 := Libadalang.Implementation.Ada_Node_P_Semantic_Parent (Node => Fld_2.Node, E_Info => Fld_2.Info);
--# end
--# expr-done 3133
Fld_4 := Fld_3.Node;
--# expr-done 3134
Cast_Expr_1 := Block_Var; 
      Cast_Result_1 := Cast_Expr_1;
Is_Equal := Fld_4 = Cast_Result_1; 
--# expr-done 3135
if Is_Equal then
   If_Result := True;
else
   --# expr-start 3138 '<FieldAccess for has_aspect at ast.py:5518>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5518
--# expr-start 3136 '<FieldAccess for canonical_part at ast.py:5518>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5518
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.canonical_part
Fld_5 := Libadalang.Implementation.Basic_Decl_P_Canonical_Part (Node => Ent.Node, Imprecise_Fallback => False, E_Info => Ent.Info);
--# end
--# expr-done 3136
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3137 '<Symbol Iterable at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Iterable); 
--# expr-done 3137
--# property-call-start BasicDecl.has_aspect
Fld_6 := Libadalang.Implementation.Basic_Decl_P_Has_Aspect (Node => Fld_5.Node, Name => Sym_2, Imprecise_Fallback => False, E_Info => Fld_5.Info);
--# end
--# expr-done 3138
   If_Result := Fld_6;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 3139
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_408;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Boolean,
                        As_Boolean => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_408;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Iterator_Type;
--# end
--# property-start '[dispatcher]BaseTypeDecl.is_discrete_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Is_Discrete_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.is_discrete_type
                  Property_Result := Base_Type_Decl_P_Is_Discrete_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.is_discrete_type
                  Property_Result := Base_Subtype_Decl_P_Is_Discrete_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.is_discrete_type
                  Property_Result := Type_Decl_P_Is_Discrete_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Is_Discrete_Type;
--# end
--# property-start BaseTypeDecl.is_not_root_int_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5529
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Not_Root_Int_Type
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Is_Null : Boolean;
Not_Val : Boolean;
Fld : Internal_Entity_Base_Type_Decl;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Cast_Result_1 : Internal_Entity;
Is_Equiv : Boolean;
Not_Val_1 : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3145 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3141 '<Not at ast.py:5530>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5530
--# expr-start 3140 '<IsNull at ast.py:5530>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5530
Is_Null := Self = null; 
--# expr-done 3140
Not_Val := not (Is_Null); 
--# expr-done 3141
if Not_Val then
   --# expr-start 3144 '<Not at ast.py:5530>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5530
--# expr-start 3143 '<Eq at ast.py:5530>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5530
Cast_Expr := Ent; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
--# expr-start 3142 '<FieldAccess for root_int_type at ast.py:5530>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5530
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.root_int_type
Fld := Libadalang.Implementation.Ada_Node_P_Root_Int_Type (Node => Self);
--# end
--# expr-done 3142
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld.Node,
         Info => Fld.Info);
Is_Equiv := Equivalent (Cast_Result, Cast_Result_1); 
--# expr-done 3143
Not_Val_1 := not (Is_Equiv); 
--# expr-done 3144
   If_Result := Not_Val_1;
else
   If_Result := False;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 3145
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Not_Root_Int_Type;
--# end
--# property-start '[dispatcher]BaseTypeDecl.is_int_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Is_Int_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.is_int_type
                  Property_Result := Base_Type_Decl_P_Is_Int_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.is_int_type
                  Property_Result := Base_Subtype_Decl_P_Is_Int_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.is_int_type
                  Property_Result := Type_Decl_P_Is_Int_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Is_Int_Type;
--# end
--# property-start BaseTypeDecl.allows_universal_int /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5538
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Allows_Universal_Int
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Boolean;
Fld_1 : Internal_Entity_Base_Type_Decl;
Sym : Symbol_Type;
Fld_2 : Internal_Entity_Expr;
Is_Null : Boolean;
Not_Val : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3152 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3146 '<FieldAccess for is_int_type at ast.py:5544>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5544
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_int_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Int_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3146
if Fld then
   If_Result := True;
else
--# expr-start 3147 '<FieldAccess for base_subtype at ast.py:5544>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5544
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.base_subtype
Fld_1 := Libadalang.Implementation.Base_Type_Decl_P_Base_Subtype (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3147
Var_Expr := Fld_1;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind bt Var_Expr
--# expr-start 3151 '<Not at ast.py:5547>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5547
--# expr-start 3150 '<IsNull at ast.py:5547>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5547
--# expr-start 3149 '<FieldAccess for get_aspect_spec_expr at ast.py:5547>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5547
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3148 '<Symbol Integer_Literal at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Integer_Literal); 
--# expr-done 3148
--# property-call-start BasicDecl.get_aspect_spec_expr
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Get_Aspect_Spec_Expr (Node => Var_Expr.Node, Name => Sym, E_Info => Var_Expr.Info);
--# end
--# expr-done 3149
Is_Null := Fld_2.Node = null; 
--# expr-done 3150
Not_Val := not (Is_Null); 
--# expr-done 3151
Scope_Result := Not_Val;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
   If_Result := Result_Var;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 3152
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Allows_Universal_Int;
--# end
--# property-start BaseTypeDecl.allows_universal_real /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5552
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Allows_Universal_Real
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Boolean;
Fld_1 : Internal_Entity_Base_Type_Decl;
Sym : Symbol_Type;
Fld_2 : Internal_Entity_Expr;
Is_Null : Boolean;
Not_Val : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3159 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3153 '<FieldAccess for is_real_type at ast.py:5558>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5558
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_real_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Real_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3153
if Fld then
   If_Result := True;
else
--# expr-start 3154 '<FieldAccess for base_subtype at ast.py:5558>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5558
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.base_subtype
Fld_1 := Libadalang.Implementation.Base_Type_Decl_P_Base_Subtype (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3154
Var_Expr := Fld_1;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind bt Var_Expr
--# expr-start 3158 '<Not at ast.py:5561>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5561
--# expr-start 3157 '<IsNull at ast.py:5561>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5561
--# expr-start 3156 '<FieldAccess for get_aspect_spec_expr at ast.py:5561>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5561
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3155 '<Symbol Real_Literal at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Real_Literal); 
--# expr-done 3155
--# property-call-start BasicDecl.get_aspect_spec_expr
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Get_Aspect_Spec_Expr (Node => Var_Expr.Node, Name => Sym, E_Info => Var_Expr.Info);
--# end
--# expr-done 3156
Is_Null := Fld_2.Node = null; 
--# expr-done 3157
Not_Val := not (Is_Null); 
--# expr-done 3158
Scope_Result := Not_Val;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
   If_Result := Result_Var;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 3159
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Allows_Universal_Real;
--# end
--# property-start BaseTypeDecl.allows_string_literal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5566
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Allows_String_Literal
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
Not_Val : Boolean;
Fld : Boolean;
Fld_1 : Internal_Entity_Base_Type_Decl;
Sym : Symbol_Type;
Fld_2 : Internal_Entity_Expr;
Is_Null_1 : Boolean;
Not_Val_1 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result : Boolean;
If_Result_1 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3168 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3161 '<Not at ast.py:5577>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5577
--# expr-start 3160 '<IsNull at ast.py:5577>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5577
Is_Null := Self = null; 
--# expr-done 3160
Not_Val := not (Is_Null); 
--# expr-done 3161
if Not_Val then
--# expr-start 3162 '<FieldAccess for is_str_type at ast.py:5578>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5578
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_str_type
Fld := Libadalang.Implementation.Base_Type_Decl_P_Is_Str_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3162
if Fld then
   If_Result := True;
else
--# expr-start 3163 '<FieldAccess for base_subtype at ast.py:5578>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5578
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.base_subtype
Fld_1 := Libadalang.Implementation.Base_Type_Decl_P_Base_Subtype (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3163
Var_Expr := Fld_1;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind bt Var_Expr
--# expr-start 3167 '<Not at ast.py:5582>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5582
--# expr-start 3166 '<IsNull at ast.py:5582>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5582
--# expr-start 3165 '<FieldAccess for get_aspect_spec_expr at ast.py:5582>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5582
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3164 '<Symbol String_Literal at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_String_Literal); 
--# expr-done 3164
--# property-call-start BasicDecl.get_aspect_spec_expr
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Get_Aspect_Spec_Expr (Node => Var_Expr.Node, Name => Sym, E_Info => Var_Expr.Info);
--# end
--# expr-done 3165
Is_Null_1 := Fld_2.Node = null; 
--# expr-done 3166
Not_Val_1 := not (Is_Null_1); 
--# expr-done 3167
Scope_Result := Not_Val_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
   If_Result := Result_Var;
end if;
   If_Result_1 := If_Result;
else
   If_Result_1 := False;
end if;
Let_Result := If_Result_1; 
--# end
--# expr-done 3168
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Allows_String_Literal;
--# end
--# property-start BaseTypeDecl.is_str_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5588
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Str_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Integer;
Is_Equal : Boolean;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3173 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3170 '<Eq at ast.py:5592>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5592
--# expr-start 3169 '<FieldAccess for array_ndims at ast.py:5592>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5592
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.array_ndims'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Array_Ndims (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3169
Is_Equal := Fld = 1; 
--# expr-done 3170
if Is_Equal then
--# expr-start 3171 '<FieldAccess for comp_type at ast.py:5592>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5592
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.comp_type
Fld_1 := Libadalang.Implementation.Base_Type_Decl_P_Comp_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3171
Var_Expr := Fld_1;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 3172 '<FieldAccess for is_char_type at ast.py:5592>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5592
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_char_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Char_Type (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 3172
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
   If_Result := Result_Var;
else
   If_Result := False;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 3173
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Str_Type;
--# end
--# property-start '[dispatcher]BaseTypeDecl.accessed_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Accessed_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.accessed_type
                  Property_Result := Base_Type_Decl_P_Accessed_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.accessed_type
                  Property_Result := Base_Subtype_Decl_P_Accessed_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.accessed_type
                  Property_Result := Type_Decl_P_Accessed_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Accessed_Type;
--# end
--# property-start BaseTypeDecl.accessed_type_no_call /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5603
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Accessed_Type_No_Call
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Access_Def;
Is_A : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3178 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3177 '<If at ast.py:5608>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5608
--# expr-start 3175 '<IsA AccessToSubpDef at ast.py:5609>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5609
--# expr-start 3174 '<FieldAccess for access_def at ast.py:5609>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5609
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.access_def'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Access_Def (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3174
Is_A := Fld_1.Node /= null 
and then Fld_1.Node.Kind in Ada_Access_To_Subp_Def_Range; 
--# expr-done 3175
if Is_A then
   If_Result := No_Entity_Base_Type_Decl;
else
   --# expr-start 3176 '<FieldAccess for accessed_type at ast.py:5611>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5611
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3176
   If_Result := Fld;
end if;
--# expr-done 3177
Let_Result := If_Result; 
--# end
--# expr-done 3178
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Accessed_Type_No_Call;
--# end
--# property-start BaseTypeDecl.final_accessed_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5615
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Final_Accessed_Type
  (Node : Bare_Base_Type_Decl
      ; First_Call : Boolean
         := True
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind first_call First_Call
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
If_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3182 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3179 '<FieldAccess for accessed_type at ast.py:5626>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5626
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3179
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind at Var_Expr
--# expr-start 3180 '<FieldAccess for final_accessed_type at ast.py:5627>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5627
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.final_accessed_type
Fld_1 := Libadalang.Implementation.Base_Type_Decl_P_Final_Accessed_Type (Node => Var_Expr.Node, First_Call => False, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 3180
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   --# expr-start 3181 '<If at ast.py:5628>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5628
if First_Call then
   If_Result := No_Entity_Base_Type_Decl;
else
   If_Result := Ent;
end if;
--# expr-done 3181
   Result_Var := If_Result;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 3182
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Final_Accessed_Type;
--# end
--# property-start BaseTypeDecl.is_access_to /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5632
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Access_To
  (Node : Bare_Base_Type_Decl
      ; Typ : Internal_Entity_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind typ Typ
   --# bind origin Origin
   Property_Result : Boolean;
      Block_Var : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3185 '<Block at ???>' Let_Result None
--# scope-start
Block_Var := Ent; 
--# bind access_type Block_Var
--# expr-start 3184 '<FieldAccess for matching_formal_type at ast.py:5634>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5634
--# expr-start 3183 '<FieldAccess for accessed_type at ast.py:5634>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5634
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 3183
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_formal_type
Fld_1 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Type (Node => Fld.Node, Formal_Type => Typ, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3184
Let_Result := Fld_1; 
--# end
--# expr-done 3185
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Access_To;
--# end
--# property-start BaseTypeDecl.is_subp_access_of /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5637
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Subp_Access_Of
  (Node : Bare_Base_Type_Decl
      ; Entity : Internal_Entity_Basic_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind entity Entity
   --# bind origin Origin
   Property_Result : Boolean;
      Block_Var : Internal_Entity_Base_Type_Decl;
Var_Expr : Internal_Entity_Access_To_Subp_Def;
Var_Expr_1 : Internal_Entity_Base_Subp_Spec;
Fld : Internal_Entity_Access_Def;
Cast_Result : Internal_Entity_Access_To_Subp_Def;
Fld_1 : Internal_Entity_Base_Subp_Spec;
Fld_2 : Internal_Entity_Subp_Spec;
Fld_3 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3191 '<Block at ???>' Let_Result None
--# scope-start
Block_Var := Ent; 
--# bind access_type Block_Var
--# expr-start 3187 '<Cast to AccessToSubpDef at ast.py:5643>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5643
--# expr-start 3186 '<FieldAccess for access_def at ast.py:5643>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5643
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.access_def'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Access_Def (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 3186
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Access_To_Subp_Def_Range
   then
      Cast_Result := Create_Internal_Entity_Access_To_Subp_Def
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Access_To_Subp_Def;
   end if;
--# expr-done 3187
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Access_To_Subp_Def then
   --# scope-start
   --# bind sa Var_Expr
--# expr-start 3188 '<FieldAccess for subp_spec_or_null at ast.py:5644>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5644
      if Entity.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Entity.Node, E_Info => Entity.Info);
--# end
--# expr-done 3188
Var_Expr_1 := Fld_1;
if Var_Expr_1 /= No_Entity_Base_Subp_Spec then
   --# scope-start
   --# bind se Var_Expr_1
--# expr-start 3190 '<FieldAccess for match_signature at ast.py:5645>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5645
--# expr-start 3189 '<FieldAccess for subp_spec at ast.py:5645>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5645
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Subp_Spec (Node => Var_Expr.Node.Access_To_Subp_Def_F_Subp_Spec, Info => Var_Expr.Info);
--# expr-done 3189
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.match_signature
Fld_3 := Libadalang.Implementation.Base_Subp_Spec_P_Match_Signature (Node => Fld_2.Node, Other => Var_Expr_1, Match_Name => False, E_Info => Fld_2.Info);
--# end
--# expr-done 3190
Scope_Result := Fld_3;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Scope_Result_1 := Result_Var;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
Let_Result := Result_Var_1; 
--# end
--# expr-done 3191
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Subp_Access_Of;
--# end
--# property-start BaseTypeDecl.is_generic_formal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5650
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Generic_Formal
  (Node : Bare_Base_Type_Decl
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Var_Expr : Bare_Base_Type_Decl;
Fld : Bare_Ada_Node;
Is_A : Boolean;
Fld_1 : Bare_Ada_Node;
Cast_Result : Bare_Base_Type_Decl;
Fld_2 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 3197 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3193 '<IsA GenericFormalTypeDecl at ast.py:5655>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5655
--# expr-start 3192 '<parent at ast.py:5655>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5655
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 3192
Is_A := Fld /= null 
and then Fld.Kind in Ada_Generic_Formal_Type_Decl_Range; 
--# expr-done 3193
if Is_A then
   If_Result := True;
else
--# expr-start 3195 '<Cast to BaseTypeDecl at ast.py:5656>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5656
--# expr-start 3194 '<parent at ast.py:5656>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5656
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Self.Parent; 
--# expr-done 3194
   if Fld_1 = null
      or else Fld_1.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Fld_1;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 3195
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 3196 '<FieldAccess for is_generic_formal at ast.py:5656>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5656
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_generic_formal
Fld_2 := Libadalang.Implementation.Base_Type_Decl_P_Is_Generic_Formal (Node => Var_Expr);
--# end
--# expr-done 3196
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
   If_Result := Result_Var;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 3197
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Generic_Formal;
--# end
--# property-start '[dispatcher]BaseTypeDecl.is_tagged_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Is_Tagged_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Incomplete_Type_Decl | Ada_Protected_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.is_tagged_type
                  Property_Result := Base_Type_Decl_P_Is_Tagged_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.is_tagged_type
                  Property_Result := Base_Subtype_Decl_P_Is_Tagged_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Classwide_Type_Decl =>
                  --# property-call-start ClasswideTypeDecl.is_tagged_type
                  Property_Result := Classwide_Type_Decl_P_Is_Tagged_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Incomplete_Tagged_Type_Decl =>
                  --# property-call-start IncompleteTaggedTypeDecl.is_tagged_type
                  Property_Result := Incomplete_Tagged_Type_Decl_P_Is_Tagged_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.is_tagged_type
                  Property_Result := Type_Decl_P_Is_Tagged_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Is_Tagged_Type;
--# end
--# property-start BaseTypeDecl.is_tagged_type_with_deref /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5665
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Tagged_Type_With_Deref
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Expr;
Is_Null : Boolean;
Not_Val : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
Fld_2 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3204 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3203 '<FieldAccess for is_tagged_type at ast.py:5670>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5670
--# expr-start 3202 '<If at ast.py:5670>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5670
--# expr-start 3200 '<Not at ast.py:5671>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5671
--# expr-start 3199 '<IsNull at ast.py:5671>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5671
--# expr-start 3198 '<FieldAccess for get_imp_deref at ast.py:5671>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5671
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.get_imp_deref'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Get_Imp_Deref (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3198
Is_Null := Fld_1.Node = null; 
--# expr-done 3199
Not_Val := not (Is_Null); 
--# expr-done 3200
if Not_Val then
   --# expr-start 3201 '<FieldAccess for accessed_type at ast.py:5672>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5672
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3201
   If_Result := Fld;
else
   If_Result := Ent;
end if;
--# expr-done 3202
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_tagged_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Tagged_Type (Node => If_Result.Node, Origin => Origin, E_Info => If_Result.Info);
--# end
--# expr-done 3203
Let_Result := Fld_2; 
--# end
--# expr-done 3204
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Tagged_Type_With_Deref;
--# end
--# property-start '[dispatcher]BaseTypeDecl.base_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Base_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Incomplete_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.base_type
                  Property_Result := Base_Type_Decl_P_Base_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.base_type
                  Property_Result := Base_Subtype_Decl_P_Base_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Classwide_Type_Decl =>
                  --# property-call-start ClasswideTypeDecl.base_type
                  Property_Result := Classwide_Type_Decl_P_Base_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.base_type
                  Property_Result := Type_Decl_P_Base_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Base_Type;
--# end
--# property-start '[dispatcher]BaseTypeDecl.base_types' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Base_Types
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl_Array_Access
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Classwide_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start BaseTypeDecl.base_types
                  Property_Result := Base_Type_Decl_P_Base_Types
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.base_types
                  Property_Result := Base_Subtype_Decl_P_Base_Types
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Base_Types;
--# end
--# property-start '[dispatcher]BaseTypeDecl.base_interfaces' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Base_Interfaces
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl_Array_Access
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Incomplete_Type_Decl .. Ada_Incomplete_Tagged_Type_Decl =>
                  --# property-call-start BaseTypeDecl.base_interfaces
                  Property_Result := Base_Type_Decl_P_Base_Interfaces
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.base_interfaces
                  Property_Result := Base_Subtype_Decl_P_Base_Interfaces
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Classwide_Type_Decl =>
                  --# property-call-start ClasswideTypeDecl.base_interfaces
                  Property_Result := Classwide_Type_Decl_P_Base_Interfaces
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Protected_Type_Decl =>
                  --# property-call-start ProtectedTypeDecl.base_interfaces
                  Property_Result := Protected_Type_Decl_P_Base_Interfaces
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Task_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start TaskTypeDecl.base_interfaces
                  Property_Result := Task_Type_Decl_P_Base_Interfaces
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.base_interfaces
                  Property_Result := Type_Decl_P_Base_Interfaces
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Base_Interfaces;
--# end
--# property-start BaseTypeDecl.find_all_derived_types /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5696
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Find_All_Derived_Types
  (Node : Bare_Base_Type_Decl
      ; Units : Internal_Unit_Array_Access
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Decl_Array_Access
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind units Units
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Type_Decl_Array_Access;
            procedure Finalizer_Scope_427 with Inline_Always;
            procedure Finalizer_Scope_2165 with Inline_Always;
            procedure Finalizer_Scope_2166 with Inline_Always;
      Var_Expr : Bare_Ada_Node;
Cast_Expr : Bare_Base_Type_Decl;
Cast_Result : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Unit_Array_Access;
Fld_2 : Bare_Ada_Node;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Fld_3 : Internal_Entity_Type_Decl_Array_Access;
Scope_Result : Internal_Entity_Type_Decl_Array_Access;
Result_Var : Internal_Entity_Type_Decl_Array_Access;
Map_Result : Internal_Entity_Type_Decl_Array_Access;
Dyn_Var_Bind_Result : Internal_Entity_Type_Decl_Array_Access;
Let_Result : Internal_Entity_Type_Decl_Array_Access;
            procedure Finalizer_Scope_427 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_427;
            procedure Finalizer_Scope_2165 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_2165;
            procedure Finalizer_Scope_2166 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_2166;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3212 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3211 '<bind at ast.py:5701>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5701
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Bound_Origin := Cast_Result; 
--# expr-start 3210 '<Mapcat at ast.py:5703>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5703
   --# expr-start 3206 '<FieldAccess for filter_is_imported_by at ast.py:5703>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5703
--# expr-start 3205 '<FieldAccess for canonical_type at ast.py:5703>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5703
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.canonical_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Canonical_Type (Node => Ent.Node, Origin => Bound_Origin, E_Info => Ent.Info);
--# end
--# expr-done 3205
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Ada_Node_P_Filter_Is_Imported_By (Node => Fld.Node, Units => Units, Transitive => True);
--# expr-done 3206
      declare
         Map_Result_Vec : Internal_Entity_Type_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Unit_Array_Access := Fld_1;
      begin
         for Item_132 of
               Collection.Items
         loop
   --# scope-start
               --# bind u Item_132
--# expr-start 3207 '<FieldAccess for root at ast.py:5704>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5704
      if Item_132 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Item_132.Ast_Root;
--# expr-done 3207
Var_Expr := Fld_2;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind r Var_Expr
--# expr-start 3209 '<FieldAccess for find_derived_types at ast.py:5705>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5705
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3208 '<as_bare_entity at ast.py:5705>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5705
Is_Null := Var_Expr = null; 
if Is_Null then
   If_Result := No_Entity;
else
As_Entity := (Info => No_Entity_Info, Node => Var_Expr); 
   If_Result := As_Entity;
end if;
--# expr-done 3208
--# property-call-start BaseTypeDecl.find_derived_types
Fld_3 := Libadalang.Implementation.Base_Type_Decl_P_Find_Derived_Types (Node => Ent.Node, Root => If_Result, Origin => Bound_Origin, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 3209
Scope_Result := Fld_3;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2166;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Type_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var);
         for Item_To_Append of
               Result_Var.Items
         loop
            Internal_Entity_Type_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_2165;
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Type_Decl_Array
           (Items_Count => Natural (Internal_Entity_Type_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Type_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Type_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Type_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 3210
Dyn_Var_Bind_Result := Map_Result; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 3211
Let_Result := Dyn_Var_Bind_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 3212
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_427;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_427;
               Finalizer_Scope_2165;
               Finalizer_Scope_2166;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Find_All_Derived_Types;
--# end
--# property-start '[dispatcher]BaseTypeDecl.record_def' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Record_Def
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Record_Def
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Record_Def;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Incomplete_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.record_def
                  Property_Result := Base_Type_Decl_P_Record_Def
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.record_def
                  Property_Result := Base_Subtype_Decl_P_Record_Def
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Classwide_Type_Decl =>
                  --# property-call-start ClasswideTypeDecl.record_def
                  Property_Result := Classwide_Type_Decl_P_Record_Def
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.record_def
                  Property_Result := Type_Decl_P_Record_Def
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Record_Def;
--# end
--# property-start '[dispatcher]BaseTypeDecl.array_def' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Array_Def
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Type_Def
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Array_Type_Def;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.array_def
                  Property_Result := Base_Type_Decl_P_Array_Def
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.array_def
                  Property_Result := Base_Subtype_Decl_P_Array_Def
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.array_def
                  Property_Result := Type_Decl_P_Array_Def
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Array_Def;
--# end
--# property-start BaseTypeDecl.array_def_with_deref /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5713
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Array_Def_With_Deref
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Type_Def
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Array_Type_Def;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Boolean;
Fld_1 : Internal_Entity_Array_Type_Def;
Fld_2 : Boolean;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Internal_Entity_Array_Type_Def;
Scope_Result : Internal_Entity_Array_Type_Def;
Result_Var : Internal_Entity_Array_Type_Def;
If_Result : Internal_Entity_Array_Type_Def;
If_Result_1 : Internal_Entity_Array_Type_Def;
Let_Result : Internal_Entity_Array_Type_Def;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3219 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3218 '<Cond at ast.py:5719>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5719
--# expr-start 3213 '<FieldAccess for is_array at ast.py:5720>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5720
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_array
Fld := Libadalang.Implementation.Basic_Decl_P_Is_Array (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3213
if Fld then
   --# expr-start 3214 '<FieldAccess for array_def at ast.py:5720>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5720
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.array_def'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Array_Def (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3214
   If_Result_1 := Fld_1;
else
--# expr-start 3215 '<FieldAccess for is_implicit_deref at ast.py:5722>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5722
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_implicit_deref
Fld_2 := Libadalang.Implementation.Base_Type_Decl_P_Is_Implicit_Deref (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3215
if Fld_2 then
--# expr-start 3216 '<FieldAccess for accessed_type at ast.py:5723>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5723
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3216
Var_Expr := Fld_3;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind c Var_Expr
--# expr-start 3217 '<FieldAccess for array_def at ast.py:5723>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5723
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.array_def'
Fld_4 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Array_Def (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 3217
Scope_Result := Fld_4;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Array_Type_Def;
end if;
   If_Result := Result_Var;
else
   If_Result := No_Entity_Array_Type_Def;
end if;
   If_Result_1 := If_Result;
end if;
--# expr-done 3218
Let_Result := If_Result_1; 
--# end
--# expr-done 3219
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Array_Def_With_Deref;
--# end
--# property-start BaseTypeDecl.is_array_def_with_deref /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5728
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Array_Def_With_Deref
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Is_Null : Boolean;
Not_Val : Boolean;
Fld : Internal_Entity_Array_Type_Def;
Is_Null_1 : Boolean;
Not_Val_1 : Boolean;
If_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
--# expr-start 3221 '<Not at ast.py:5729>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5729
--# expr-start 3220 '<IsNull at ast.py:5729>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5729
Is_Null := Self = null; 
--# expr-done 3220
Not_Val := not (Is_Null); 
--# expr-done 3221
if Not_Val then
   --# expr-start 3224 '<Not at ast.py:5729>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5729
--# expr-start 3223 '<IsNull at ast.py:5729>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5729
--# expr-start 3222 '<FieldAccess for array_def_with_deref at ast.py:5729>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5729
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.array_def_with_deref
Fld := Libadalang.Implementation.Base_Type_Decl_P_Array_Def_With_Deref (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3222
Is_Null_1 := Fld.Node = null; 
--# expr-done 3223
Not_Val_1 := not (Is_Null_1); 
--# expr-done 3224
   If_Result := Not_Val_1;
else
   If_Result := False;
end if;
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Array_Def_With_Deref;
--# end
--# property-start BaseTypeDecl.is_array_def_with_deref_or_null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5733
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Array_Def_With_Deref_Or_Null
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Is_Null : Boolean;
Fld : Internal_Entity_Array_Type_Def;
Is_Null_1 : Boolean;
Not_Val : Boolean;
If_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
--# expr-start 3225 '<IsNull at ast.py:5734>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5734
Is_Null := Self = null; 
--# expr-done 3225
if Is_Null then
   If_Result := True;
else
   --# expr-start 3228 '<Not at ast.py:5734>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5734
--# expr-start 3227 '<IsNull at ast.py:5734>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5734
--# expr-start 3226 '<FieldAccess for array_def_with_deref at ast.py:5734>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5734
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.array_def_with_deref
Fld := Libadalang.Implementation.Base_Type_Decl_P_Array_Def_With_Deref (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3226
Is_Null_1 := Fld.Node = null; 
--# expr-done 3227
Not_Val := not (Is_Null_1); 
--# expr-done 3228
   If_Result := Not_Val;
end if;
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Array_Def_With_Deref_Or_Null;
--# end
--# property-start BaseTypeDecl.comp_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5740
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Comp_Type
  (Node : Bare_Base_Type_Decl
      ; Is_Subscript : Boolean
         := False
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind is_subscript Is_Subscript
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Ad : Internal_Entity_Array_Type_Def;
Var_Expr_1 : Internal_Entity_Base_Type_Decl;
Var_Expr_2 : Internal_Entity_Array_Type_Def;
Var_Expr_3 : Internal_Entity_Access_Def;
Match_0 : Internal_Entity_Access_To_Subp_Def;
Match_1 : Internal_Entity_Base_Type_Access_Def;
Fld : Internal_Entity_Array_Type_Def;
Fld_1 : Internal_Entity_Array_Type_Def;
If_Result : Internal_Entity_Array_Type_Def;
Fld_2 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Fld_3 : Internal_Entity_Access_Def;
Fld_4 : Internal_Entity_Subp_Spec;
Fld_5 : Internal_Entity_Base_Type_Decl;
Fld_6 : Internal_Entity_Base_Type_Decl;
Match_Prefix : Internal_Entity_Access_Def;
Cast_Result : Internal_Entity_Access_To_Subp_Def;
Let_Result : Internal_Entity_Base_Type_Decl;
Scope_Result_2 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity_Base_Type_Access_Def;
Let_Result_1 : Internal_Entity_Base_Type_Decl;
Scope_Result_3 : Internal_Entity_Base_Type_Decl;
Match_Result : Internal_Entity_Base_Type_Decl;
Scope_Result_4 : Internal_Entity_Base_Type_Decl;
Result_Var_1 : Internal_Entity_Base_Type_Decl;
Result_Var_2 : Internal_Entity_Base_Type_Decl;
Let_Result_2 : Internal_Entity_Base_Type_Decl;
Scope_Result_5 : Internal_Entity_Base_Type_Decl;
Result_Var_3 : Internal_Entity_Base_Type_Decl;
Let_Result_3 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3239 '<Block at ???>' Let_Result_3 None
--# scope-start
Var_Expr := Ent;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind e Var_Expr
--# expr-start 3238 '<Let at ast.py:5750>' Let_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5750
--# scope-start
--# expr-start 3231 '<If at ast.py:5751>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5751
if Is_Subscript then
   --# expr-start 3229 '<FieldAccess for array_def_with_deref at ast.py:5752>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5752
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.array_def_with_deref
Fld := Libadalang.Implementation.Base_Type_Decl_P_Array_Def_With_Deref (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3229
   If_Result := Fld;
else
   --# expr-start 3230 '<FieldAccess for array_def at ast.py:5753>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5753
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.array_def'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Array_Def (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3230
   If_Result := Fld_1;
end if;
--# expr-done 3231
Ad := If_Result; 
--# bind ad Ad
Var_Expr_2 := Ad;
if Var_Expr_2 /= No_Entity_Array_Type_Def then
   --# scope-start
   --# bind ad Var_Expr_2
--# expr-start 3232 '<FieldAccess for comp_type at ast.py:5754>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5754
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ArrayTypeDef.comp_type
Fld_2 := Libadalang.Implementation.Array_Type_Def_P_Comp_Type (Node => Var_Expr_2.Node, Origin => Origin, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 3232
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Var_Expr_1 := Result_Var;
if Var_Expr_1 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind e Var_Expr_1
Scope_Result_1 := Var_Expr_1;
   --# end
   Result_Var_2 := Scope_Result_1;
else
--# expr-start 3233 '<FieldAccess for access_def at ast.py:5755>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5755
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.access_def'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Access_Def (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 3233
Var_Expr_3 := Fld_3;
if Var_Expr_3 /= No_Entity_Access_Def then
   --# scope-start
--# expr-start 3237 '<Match at ast.py:5755>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5755
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Var_Expr_3; 
case Ada_Access_Def (Match_Prefix.Node.Kind) is
         when Ada_Access_To_Subp_Def =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Access_To_Subp_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind asd Match_0
--# expr-start 3235 '<FieldAccess for return_type at ast.py:5757>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5757
--# expr-start 3234 '<FieldAccess for subp_spec at ast.py:5757>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5757
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Subp_Spec (Node => Match_0.Node.Access_To_Subp_Def_F_Subp_Spec, Info => Match_0.Info);
--# expr-done 3234
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.return_type
Fld_5 := Libadalang.Implementation.Base_Subp_Spec_P_Return_Type (Node => Fld_4.Node, Origin => Origin, E_Info => Fld_4.Info);
--# end
--# expr-done 3235
Let_Result := Fld_5; 
--# end
Scope_Result_2 := Let_Result;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Anonymous_Type_Access_Def .. Ada_Type_Access_Def =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Base_Type_Access_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind tad Match_1
--# expr-start 3236 '<FieldAccess for accessed_type at ast.py:5758>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5758
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.accessed_type'
Fld_6 := Libadalang.Implementation.Dispatcher_Type_Def_P_Accessed_Type (Node => Match_1.Node, Origin => Origin, E_Info => Match_1.Info);
--# end
--# expr-done 3236
Let_Result_1 := Fld_6; 
--# end
Scope_Result_3 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_3; 
end case;
--# expr-done 3237
Scope_Result_4 := Match_Result;
   --# end
   Result_Var_1 := Scope_Result_4;
else
   Result_Var_1 := No_Entity_Base_Type_Decl;
end if;
   Result_Var_2 := Result_Var_1;
end if;
Let_Result_2 := Result_Var_2; 
--# end
--# expr-done 3238
Scope_Result_5 := Let_Result_2;
   --# end
   Result_Var_3 := Scope_Result_5;
else
   Result_Var_3 := No_Entity_Base_Type_Decl;
end if;
Let_Result_3 := Result_Var_3; 
--# end
--# expr-done 3239
      Property_Result := Let_Result_3;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Comp_Type;
--# end
--# property-start BaseTypeDecl.index_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5765
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Index_Type
  (Node : Bare_Base_Type_Decl
      ; Dim : Integer
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind dim Dim
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Array_Type_Def;
Fld : Internal_Entity_Array_Type_Def;
Fld_1 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3242 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3240 '<FieldAccess for array_def_with_deref at ast.py:5773>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5773
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.array_def_with_deref
Fld := Libadalang.Implementation.Base_Type_Decl_P_Array_Def_With_Deref (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3240
Var_Expr := Fld;
if Var_Expr /= No_Entity_Array_Type_Def then
   --# scope-start
   --# bind ad Var_Expr
--# expr-start 3241 '<FieldAccess for index_type at ast.py:5773>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5773
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ArrayTypeDef.index_type
Fld_1 := Libadalang.Implementation.Array_Type_Def_P_Index_Type (Node => Var_Expr.Node, Dim => Dim, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 3241
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 3242
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Index_Type;
--# end
--# property-start BaseTypeDecl.expr_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5777
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Expr_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := Ent;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Expr_Type;
--# end
--# property-start BaseTypeDecl.is_derived_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5781
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Derived_Type
  (Node : Bare_Base_Type_Decl
      ; Other_Type : Internal_Entity_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind other_type Other_Type
   --# bind origin Origin
   Property_Result : Boolean;
            procedure Finalizer_Scope_436 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Block_Var_1 : Internal_Entity_Base_Type_Decl;
Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Is_Equiv : Boolean;
Fld_2 : Internal_Entity_Classwide_Type_Decl;
Is_Null : Boolean;
Not_Val : Boolean;
Fld_3 : Internal_Entity_Classwide_Type_Decl;
Fld_4 : Internal_Entity_Classwide_Type_Decl;
Cast_Result_2 : Internal_Entity;
Cast_Result_3 : Internal_Entity;
Is_Equiv_1 : Boolean;
If_Result : Boolean;
If_Result_1 : Boolean;
Fld_5 : Internal_Entity_Base_Type_Decl_Array_Access;
Fld_6 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Quantifier_Result : Boolean;
If_Result_2 : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_436 is
            begin
                     Dec_Ref (Fld_5);
            end Finalizer_Scope_436;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3255 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3243 '<FieldAccess for canonical_type at ast.py:5785>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5785
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.canonical_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Canonical_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3243
Block_Var := Fld; 
--# bind entity_can Block_Var
--# expr-start 3244 '<FieldAccess for canonical_type at ast.py:5786>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5786
      if Other_Type.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.canonical_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Canonical_Type (Node => Other_Type.Node, Origin => Origin, E_Info => Other_Type.Info);
--# end
--# expr-done 3244
Block_Var_1 := Fld_1; 
--# bind other_can Block_Var_1
--# expr-start 3245 '<Eq at ast.py:5788>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5788
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Cast_Expr_1 := Block_Var_1; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Is_Equiv := Equivalent (Cast_Result, Cast_Result_1); 
--# expr-done 3245
if Is_Equiv then
   If_Result_1 := True;
else
--# expr-start 3248 '<Not at ast.py:5789>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5789
--# expr-start 3247 '<IsNull at ast.py:5789>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5789
--# expr-start 3246 '<FieldAccess for classwide_type at ast.py:5789>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5789
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.classwide_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Classwide_Type (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 3246
Is_Null := Fld_2.Node = null; 
--# expr-done 3247
Not_Val := not (Is_Null); 
--# expr-done 3248
if Not_Val then
   --# expr-start 3251 '<Eq at ast.py:5790>' Is_Equiv_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5790
--# expr-start 3249 '<FieldAccess for classwide_type at ast.py:5790>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5790
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.classwide_type'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Classwide_Type (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 3249
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
--# expr-start 3250 '<FieldAccess for classwide_type at ast.py:5790>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5790
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.classwide_type'
Fld_4 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Classwide_Type (Node => Block_Var_1.Node, E_Info => Block_Var_1.Info);
--# end
--# expr-done 3250
      Cast_Result_3 := Create_Internal_Entity
        (Node => Fld_4.Node,
         Info => Fld_4.Info);
Is_Equiv_1 := Equivalent (Cast_Result_2, Cast_Result_3); 
--# expr-done 3251
   If_Result := Is_Equiv_1;
else
   If_Result := False;
end if;
   If_Result_1 := If_Result;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
   --# expr-start 3254 '<AnyQuantifier at ast.py:5791>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5791
   --# expr-start 3252 '<FieldAccess for base_types at ast.py:5791>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5791
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.base_types'
Fld_5 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Base_Types (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3252
      Quantifier_Result := False;
      declare
         Collection : constant Internal_Entity_Base_Type_Decl_Array_Access := Fld_5;
      begin
         for Item_133 of
               Collection.Items
         loop
   --# scope-start
               --# bind bt Item_133
Var_Expr := Item_133;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 3253 '<FieldAccess for is_derived_type at ast.py:5791>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5791
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_derived_type
Fld_6 := Libadalang.Implementation.Base_Type_Decl_P_Is_Derived_Type (Node => Var_Expr.Node, Other_Type => Other_Type, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 3253
Scope_Result := Fld_6;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
      Quantifier_Result := Result_Var;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
--# expr-done 3254
   If_Result_2 := Quantifier_Result;
end if;
Let_Result := If_Result_2; 
--# end
--# expr-done 3255
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_436;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_436;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Derived_Type;
--# end
--# property-start '[dispatcher]BaseTypeDecl.is_iterable_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Is_Iterable_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Incomplete_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.is_iterable_type
                  Property_Result := Base_Type_Decl_P_Is_Iterable_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.is_iterable_type
                  Property_Result := Base_Subtype_Decl_P_Is_Iterable_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Classwide_Type_Decl =>
                  --# property-call-start ClasswideTypeDecl.is_iterable_type
                  Property_Result := Classwide_Type_Decl_P_Is_Iterable_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.is_iterable_type
                  Property_Result := Type_Decl_P_Is_Iterable_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Is_Iterable_Type;
--# end
--# property-start '[dispatcher]BaseTypeDecl.is_interface_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Is_Interface_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl | Ada_Incomplete_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start BaseTypeDecl.is_interface_type
                  Property_Result := Base_Type_Decl_P_Is_Interface_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Classwide_Type_Decl =>
                  --# property-call-start ClasswideTypeDecl.is_interface_type
                  Property_Result := Classwide_Type_Decl_P_Is_Interface_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Is_Interface_Type;
--# end
--# property-start '[dispatcher]BaseTypeDecl.iterable_comp_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Iterable_Comp_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Incomplete_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.iterable_comp_type
                  Property_Result := Base_Type_Decl_P_Iterable_Comp_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.iterable_comp_type
                  Property_Result := Base_Subtype_Decl_P_Iterable_Comp_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Classwide_Type_Decl =>
                  --# property-call-start ClasswideTypeDecl.iterable_comp_type
                  Property_Result := Classwide_Type_Decl_P_Iterable_Comp_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.iterable_comp_type
                  Property_Result := Type_Decl_P_Iterable_Comp_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Iterable_Comp_Type;
--# end
--# property-start BaseTypeDecl.matching_prefix_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5819
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Matching_Prefix_Type
  (Node : Bare_Base_Type_Decl
      ; Container_Type : Internal_Entity_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind container_type Container_Type
   --# bind origin Origin
   Property_Result : Boolean;
      Block_Var : Internal_Entity_Base_Type_Decl;
Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
Not_Val : Boolean;
Is_Null_1 : Boolean;
Not_Val_1 : Boolean;
If_Result : Boolean;
Fld : Boolean;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result_1 : Boolean;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3265 '<Block at ???>' Let_Result None
--# scope-start
Block_Var := Container_Type; 
--# bind cont_type Block_Var
--# expr-start 3257 '<Not at ast.py:5826>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5826
--# expr-start 3256 '<IsNull at ast.py:5826>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5826
Is_Null := Self = null; 
--# expr-done 3256
Not_Val := not (Is_Null); 
--# expr-done 3257
if Not_Val then
   --# expr-start 3259 '<Not at ast.py:5826>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5826
--# expr-start 3258 '<IsNull at ast.py:5826>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5826
Is_Null_1 := Container_Type.Node = null; 
--# expr-done 3258
Not_Val_1 := not (Is_Null_1); 
--# expr-done 3259
   If_Result := Not_Val_1;
else
   If_Result := False;
end if;
if If_Result then
--# expr-start 3260 '<FieldAccess for matching_formal_prim_type at ast.py:5828>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5828
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_formal_prim_type
Fld := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Prim_Type (Node => Ent.Node, Formal_Type => Block_Var, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3260
if Fld then
   If_Result_1 := True;
else
--# expr-start 3261 '<FieldAccess for final_accessed_type at ast.py:5831>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5831
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.final_accessed_type
Fld_1 := Libadalang.Implementation.Base_Type_Decl_P_Final_Accessed_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3261
Var_Expr := Fld_1;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 3262 '<FieldAccess for matching_formal_prim_type at ast.py:5831>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5831
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_formal_prim_type
Fld_2 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Prim_Type (Node => Var_Expr.Node, Formal_Type => Block_Var, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 3262
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
   If_Result_1 := Result_Var;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
--# expr-start 3263 '<FieldAccess for accessed_type at ast.py:5835>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5835
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 3263
Var_Expr_1 := Fld_3;
if Var_Expr_1 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind at Var_Expr_1
--# expr-start 3264 '<FieldAccess for matching_formal_prim_type at ast.py:5836>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5836
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_formal_prim_type
Fld_4 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Prim_Type (Node => Ent.Node, Formal_Type => Var_Expr_1, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3264
Scope_Result_1 := Fld_4;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
   If_Result_2 := Result_Var_1;
end if;
   If_Result_3 := If_Result_2;
else
   If_Result_3 := False;
end if;
Let_Result := If_Result_3; 
--# end
--# expr-done 3265
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Matching_Prefix_Type;
--# end
--# property-start BaseTypeDecl.matching_access_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5841
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Matching_Access_Type
  (Node : Bare_Base_Type_Decl
      ; Expected_Type : Internal_Entity_Base_Type_Decl
      ; For_Assignment : Boolean
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind expected_type Expected_Type
   --# bind for_assignment For_Assignment
   --# bind origin Origin
   Property_Result : Boolean;
      Block_Var : Internal_Entity_Base_Type_Decl;
Match_0 : Internal_Entity_Anonymous_Type_Decl;
Match_1 : Internal_Entity_Base_Type_Decl;
Match_0_1 : Internal_Entity_Anonymous_Type_Decl;
Match_1_1 : Internal_Entity_Base_Type_Decl;
Fld : Boolean;
Fld_1 : Boolean;
Match_Prefix : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity_Anonymous_Type_Decl;
Let_Result : Boolean;
Scope_Result : Boolean;
Cast_Result_1 : Internal_Entity_Base_Type_Decl;
Let_Result_1 : Boolean;
Scope_Result_1 : Boolean;
Match_Result : Boolean;
Match_Prefix_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity_Anonymous_Type_Decl;
Let_Result_2 : Boolean;
Scope_Result_2 : Boolean;
Cast_Result_3 : Internal_Entity_Base_Type_Decl;
Let_Result_3 : Boolean;
Scope_Result_3 : Boolean;
Match_Result_1 : Boolean;
Let_Result_4 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3270 '<Block at ???>' Let_Result_4 None
--# scope-start
Block_Var := Ent; 
--# bind actual_type Block_Var
--# expr-start 3269 '<Match at ast.py:5847>' Match_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5847
      if Expected_Type.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix_1 := Expected_Type; 
case Ada_Base_Type_Decl (Match_Prefix_1.Node.Kind) is
         when Ada_Anonymous_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Anonymous_Type_Decl
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_0 := Cast_Result_2; 
--# bind atd Match_0
--# expr-start 3266 '<FieldAccess for access_def_matches at ast.py:5849>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5849
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AnonymousTypeDecl.access_def_matches
Fld := Libadalang.Implementation.Anonymous_Type_Decl_P_Access_Def_Matches (Node => Match_0.Node, Other => Block_Var, For_Assignment => For_Assignment, Origin => Origin, E_Info => Match_0.Info);
--# end
--# expr-done 3266
Let_Result_2 := Fld; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result_1 := Scope_Result_2; 
         when Ada_Discrete_Base_Subtype_Decl .. Ada_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Base_Type_Decl
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_1 := Cast_Result_3; 
--# bind _ Match_1
--# expr-start 3268 '<Match at ast.py:5850>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5850
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Block_Var; 
case Ada_Base_Type_Decl (Match_Prefix.Node.Kind) is
         when Ada_Anonymous_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Anonymous_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0_1 := Cast_Result; 
--# bind atd2 Match_0_1
--# expr-start 3267 '<FieldAccess for access_def_matches at ast.py:5852>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5852
      if Match_0_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AnonymousTypeDecl.access_def_matches
Fld_1 := Libadalang.Implementation.Anonymous_Type_Decl_P_Access_Def_Matches (Node => Match_0_1.Node, Other => Expected_Type, For_Assignment => For_Assignment, Origin => Origin, E_Info => Match_0_1.Info);
--# end
--# expr-done 3267
Let_Result := Fld_1; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Discrete_Base_Subtype_Decl .. Ada_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Base_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1_1 := Cast_Result_1; 
--# bind _ Match_1_1
Let_Result_1 := False; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
end case;
--# expr-done 3268
Let_Result_3 := Match_Result; 
--# end
Scope_Result_3 := Let_Result_3;
   --# end
            Match_Result_1 := Scope_Result_3; 
end case;
--# expr-done 3269
Let_Result_4 := Match_Result_1; 
--# end
--# expr-done 3270
      Property_Result := Let_Result_4;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Matching_Access_Type;
--# end
--# property-start BaseTypeDecl.matching_formal_prim_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5858
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Matching_Formal_Prim_Type
  (Node : Bare_Base_Type_Decl
      ; Formal_Type : Internal_Entity_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind formal_type Formal_Type
   --# bind origin Origin
   Property_Result : Boolean;
      Is_Null : Boolean;
Not_Val : Boolean;
Is_Null_1 : Boolean;
Not_Val_1 : Boolean;
If_Result : Boolean;
Fld : Boolean;
If_Result_1 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3276 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3272 '<Not at ast.py:5860>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5860
--# expr-start 3271 '<IsNull at ast.py:5860>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5860
Is_Null := Formal_Type.Node = null; 
--# expr-done 3271
Not_Val := not (Is_Null); 
--# expr-done 3272
if Not_Val then
   --# expr-start 3274 '<Not at ast.py:5861>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5861
--# expr-start 3273 '<IsNull at ast.py:5861>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5861
Is_Null_1 := Self = null; 
--# expr-done 3273
Not_Val_1 := not (Is_Null_1); 
--# expr-done 3274
   If_Result := Not_Val_1;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 3275 '<FieldAccess for matching_formal_type_impl at ast.py:5862>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5862
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_formal_type_impl
Fld := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Type_Impl (Node => Ent.Node, Formal_Type => Formal_Type, Accept_Derived => True, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3275
   If_Result_1 := Fld;
else
   If_Result_1 := False;
end if;
Let_Result := If_Result_1; 
--# end
--# expr-done 3276
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Matching_Formal_Prim_Type;
--# end
--# property-start BaseTypeDecl.matching_formal_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5866
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Matching_Formal_Type
  (Node : Bare_Base_Type_Decl
      ; Formal_Type : Internal_Entity_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind formal_type Formal_Type
   --# bind origin Origin
   Property_Result : Boolean;
      Is_Null : Boolean;
Not_Val : Boolean;
Is_Null_1 : Boolean;
Not_Val_1 : Boolean;
If_Result : Boolean;
Fld : Boolean;
If_Result_1 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3282 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3278 '<Not at ast.py:5868>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5868
--# expr-start 3277 '<IsNull at ast.py:5868>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5868
Is_Null := Formal_Type.Node = null; 
--# expr-done 3277
Not_Val := not (Is_Null); 
--# expr-done 3278
if Not_Val then
   --# expr-start 3280 '<Not at ast.py:5869>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5869
--# expr-start 3279 '<IsNull at ast.py:5869>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5869
Is_Null_1 := Self = null; 
--# expr-done 3279
Not_Val_1 := not (Is_Null_1); 
--# expr-done 3280
   If_Result := Not_Val_1;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 3281 '<FieldAccess for matching_formal_type_impl at ast.py:5870>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5870
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_formal_type_impl
Fld := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Type_Impl (Node => Ent.Node, Formal_Type => Formal_Type, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3281
   If_Result_1 := Fld;
else
   If_Result_1 := False;
end if;
Let_Result := If_Result_1; 
--# end
--# expr-done 3282
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Matching_Formal_Type;
--# end
--# property-start BaseTypeDecl.matching_formal_type_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5874
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Matching_Formal_Type_Impl
  (Node : Bare_Base_Type_Decl
      ; Formal_Type : Internal_Entity_Base_Type_Decl
      ; Accept_Derived : Boolean
         := False
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind formal_type Formal_Type
   --# bind accept_derived Accept_Derived
   --# bind origin Origin
   Property_Result : Boolean;
      Block_Var : Internal_Entity_Base_Type_Decl;
Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Entity_Base_Type_Decl;
Fld : Boolean;
If_Result : Boolean;
Fld_1 : Boolean;
If_Result_1 : Boolean;
Fld_2 : Boolean;
Fld_3 : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Boolean;
Fld_4 : Internal_Entity_Base_Type_Decl;
Fld_5 : Internal_Entity_Base_Type_Decl;
Fld_6 : Boolean;
If_Result_4 : Boolean;
Fld_7 : Boolean;
If_Result_5 : Boolean;
Fld_8 : Boolean;
Fld_9 : Boolean;
If_Result_6 : Boolean;
If_Result_7 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
If_Result_8 : Boolean;
Fld_10 : Internal_Entity_Expr;
Is_Null : Boolean;
Not_Val : Boolean;
Fld_11 : Internal_Entity_Base_Type_Decl;
Fld_12 : Boolean;
If_Result_9 : Boolean;
If_Result_10 : Boolean;
Fld_13 : Boolean;
If_Result_11 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3299 '<Block at ???>' Let_Result None
--# scope-start
Block_Var := Ent; 
--# bind actual_type Block_Var
--# expr-start 3283 '<FieldAccess for is_classwide at ast.py:5878>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5878
      if Formal_Type.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_classwide'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Classwide (Node => Formal_Type.Node, E_Info => Formal_Type.Info);
--# end
--# expr-done 3283
if Fld then
   If_Result := True;
else
   If_Result := Accept_Derived;
end if;
if If_Result then
   --# expr-start 3284 '<FieldAccess for is_derived_type at ast.py:5879>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5879
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_derived_type
Fld_1 := Libadalang.Implementation.Base_Type_Decl_P_Is_Derived_Type (Node => Block_Var.Node, Other_Type => Formal_Type, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 3284
   If_Result_1 := Fld_1;
else
   If_Result_1 := False;
end if;
if If_Result_1 then
   If_Result_3 := True;
else
--# expr-start 3285 '<FieldAccess for is_classwide at ast.py:5881>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5881
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_classwide'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Classwide (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 3285
if Fld_2 then
   --# expr-start 3286 '<FieldAccess for is_derived_type at ast.py:5882>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5882
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_derived_type
Fld_3 := Libadalang.Implementation.Base_Type_Decl_P_Is_Derived_Type (Node => Block_Var.Node, Other_Type => Formal_Type, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 3286
   If_Result_2 := Fld_3;
else
   If_Result_2 := False;
end if;
   If_Result_3 := If_Result_2;
end if;
if If_Result_3 then
   If_Result_8 := True;
else
--# expr-start 3287 '<FieldAccess for accessed_type at ast.py:5885>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5885
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_4 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 3287
Var_Expr := Fld_4;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind actual_access Var_Expr
--# expr-start 3288 '<FieldAccess for accessed_type at ast.py:5887>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5887
      if Formal_Type.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_5 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Formal_Type.Node, Origin => Origin, E_Info => Formal_Type.Info);
--# end
--# expr-done 3288
Var_Expr_1 := Fld_5;
if Var_Expr_1 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind formal_access Var_Expr_1
--# expr-start 3289 '<FieldAccess for is_classwide at ast.py:5889>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5889
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_classwide'
Fld_6 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Classwide (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 3289
if Fld_6 then
   If_Result_4 := True;
else
   If_Result_4 := Accept_Derived;
end if;
if If_Result_4 then
   --# expr-start 3290 '<FieldAccess for is_derived_type at ast.py:5890>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5890
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_derived_type
Fld_7 := Libadalang.Implementation.Base_Type_Decl_P_Is_Derived_Type (Node => Var_Expr.Node, Other_Type => Var_Expr_1, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 3290
   If_Result_5 := Fld_7;
else
   If_Result_5 := False;
end if;
if If_Result_5 then
   If_Result_7 := True;
else
--# expr-start 3291 '<FieldAccess for is_classwide at ast.py:5891>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5891
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_classwide'
Fld_8 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Classwide (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 3291
if Fld_8 then
   --# expr-start 3292 '<FieldAccess for is_derived_type at ast.py:5892>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5892
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_derived_type
Fld_9 := Libadalang.Implementation.Base_Type_Decl_P_Is_Derived_Type (Node => Var_Expr.Node, Other_Type => Var_Expr_1, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 3292
   If_Result_6 := Fld_9;
else
   If_Result_6 := False;
end if;
   If_Result_7 := If_Result_6;
end if;
Scope_Result := If_Result_7;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Scope_Result_1 := Result_Var;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
   If_Result_8 := Result_Var_1;
end if;
if If_Result_8 then
   If_Result_10 := True;
else
--# expr-start 3295 '<Not at ast.py:5897>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5897
--# expr-start 3294 '<IsNull at ast.py:5897>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5897
--# expr-start 3293 '<FieldAccess for get_imp_deref at ast.py:5897>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5897
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.get_imp_deref'
Fld_10 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Get_Imp_Deref (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 3293
Is_Null := Fld_10.Node = null; 
--# expr-done 3294
Not_Val := not (Is_Null); 
--# expr-done 3295
if Not_Val then
   --# expr-start 3297 '<FieldAccess for matching_formal_type at ast.py:5898>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5898
--# expr-start 3296 '<FieldAccess for accessed_type at ast.py:5898>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5898
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_11 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 3296
      if Fld_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_formal_type
Fld_12 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Type (Node => Fld_11.Node, Formal_Type => Formal_Type, Origin => Origin, E_Info => Fld_11.Info);
--# end
--# expr-done 3297
   If_Result_9 := Fld_12;
else
   If_Result_9 := False;
end if;
   If_Result_10 := If_Result_9;
end if;
if If_Result_10 then
   If_Result_11 := True;
else
   --# expr-start 3298 '<FieldAccess for matching_type at ast.py:5901>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5901
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_type
Fld_13 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Type (Node => Block_Var.Node, Expected_Type => Formal_Type, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 3298
   If_Result_11 := Fld_13;
end if;
Let_Result := If_Result_11; 
--# end
--# expr-done 3299
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Matching_Formal_Type_Impl;
--# end
--# property-start BaseTypeDecl.matching_assign_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5905
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Matching_Assign_Type
  (Node : Bare_Base_Type_Decl
      ; Expected_Type : Internal_Entity_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind expected_type Expected_Type
   --# bind origin Origin
   Property_Result : Boolean;
      Block_Var : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
Not_Val : Boolean;
Is_Null_1 : Boolean;
Not_Val_1 : Boolean;
If_Result : Boolean;
Fld : Boolean;
Fld_1 : Boolean;
Fld_2 : Boolean;
If_Result_1 : Boolean;
If_Result_2 : Boolean;
Fld_3 : Internal_Entity_Expr;
Is_Null_2 : Boolean;
Not_Val_2 : Boolean;
Fld_4 : Internal_Entity_Base_Type_Decl;
Fld_5 : Boolean;
If_Result_3 : Boolean;
If_Result_4 : Boolean;
Fld_6 : Internal_Entity_Expr;
Is_Null_3 : Boolean;
Not_Val_3 : Boolean;
Fld_7 : Internal_Entity_Base_Type_Decl;
Fld_8 : Boolean;
If_Result_5 : Boolean;
If_Result_6 : Boolean;
Fld_9 : Boolean;
If_Result_7 : Boolean;
If_Result_8 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3318 '<Block at ???>' Let_Result None
--# scope-start
Block_Var := Ent; 
--# bind actual_type Block_Var
--# expr-start 3301 '<Not at ast.py:5907>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5907
--# expr-start 3300 '<IsNull at ast.py:5907>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5907
Is_Null := Self = null; 
--# expr-done 3300
Not_Val := not (Is_Null); 
--# expr-done 3301
if Not_Val then
   --# expr-start 3303 '<Not at ast.py:5907>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5907
--# expr-start 3302 '<IsNull at ast.py:5907>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5907
Is_Null_1 := Expected_Type.Node = null; 
--# expr-done 3302
Not_Val_1 := not (Is_Null_1); 
--# expr-done 3303
   If_Result := Not_Val_1;
else
   If_Result := False;
end if;
if If_Result then
--# expr-start 3304 '<FieldAccess for matching_type at ast.py:5908>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5908
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_type
Fld := Libadalang.Implementation.Base_Type_Decl_P_Matching_Type (Node => Ent.Node, Expected_Type => Expected_Type, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3304
if Fld then
   If_Result_2 := True;
else
--# expr-start 3305 '<FieldAccess for is_classwide at ast.py:5911>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5911
      if Expected_Type.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_classwide'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Classwide (Node => Expected_Type.Node, E_Info => Expected_Type.Info);
--# end
--# expr-done 3305
if Fld_1 then
   --# expr-start 3306 '<FieldAccess for matching_formal_prim_type at ast.py:5912>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5912
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_formal_prim_type
Fld_2 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Prim_Type (Node => Block_Var.Node, Formal_Type => Expected_Type, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 3306
   If_Result_1 := Fld_2;
else
   If_Result_1 := False;
end if;
   If_Result_2 := If_Result_1;
end if;
if If_Result_2 then
   If_Result_4 := True;
else
--# expr-start 3309 '<Not at ast.py:5916>' Not_Val_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5916
--# expr-start 3308 '<IsNull at ast.py:5916>' Is_Null_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5916
--# expr-start 3307 '<FieldAccess for get_imp_deref at ast.py:5916>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5916
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.get_imp_deref'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Get_Imp_Deref (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 3307
Is_Null_2 := Fld_3.Node = null; 
--# expr-done 3308
Not_Val_2 := not (Is_Null_2); 
--# expr-done 3309
if Not_Val_2 then
   --# expr-start 3311 '<FieldAccess for matching_assign_type at ast.py:5917>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5917
--# expr-start 3310 '<FieldAccess for accessed_type at ast.py:5917>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5917
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_4 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 3310
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_assign_type
Fld_5 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Assign_Type (Node => Fld_4.Node, Expected_Type => Expected_Type, Origin => Origin, E_Info => Fld_4.Info);
--# end
--# expr-done 3311
   If_Result_3 := Fld_5;
else
   If_Result_3 := False;
end if;
   If_Result_4 := If_Result_3;
end if;
if If_Result_4 then
   If_Result_6 := True;
else
--# expr-start 3314 '<Not at ast.py:5922>' Not_Val_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5922
--# expr-start 3313 '<IsNull at ast.py:5922>' Is_Null_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5922
--# expr-start 3312 '<FieldAccess for get_imp_deref at ast.py:5922>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5922
      if Expected_Type.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.get_imp_deref'
Fld_6 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Get_Imp_Deref (Node => Expected_Type.Node, E_Info => Expected_Type.Info);
--# end
--# expr-done 3312
Is_Null_3 := Fld_6.Node = null; 
--# expr-done 3313
Not_Val_3 := not (Is_Null_3); 
--# expr-done 3314
if Not_Val_3 then
   --# expr-start 3316 '<FieldAccess for matching_assign_type at ast.py:5923>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5923
--# expr-start 3315 '<FieldAccess for accessed_type at ast.py:5923>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5923
      if Expected_Type.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_7 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Expected_Type.Node, Origin => Origin, E_Info => Expected_Type.Info);
--# end
--# expr-done 3315
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_assign_type
Fld_8 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Assign_Type (Node => Fld_7.Node, Expected_Type => Block_Var, Origin => Origin, E_Info => Fld_7.Info);
--# end
--# expr-done 3316
   If_Result_5 := Fld_8;
else
   If_Result_5 := False;
end if;
   If_Result_6 := If_Result_5;
end if;
if If_Result_6 then
   If_Result_7 := True;
else
   --# expr-start 3317 '<FieldAccess for matching_access_type at ast.py:5927>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5927
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_access_type
Fld_9 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Access_Type (Node => Ent.Node, Expected_Type => Expected_Type, For_Assignment => True, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3317
   If_Result_7 := Fld_9;
end if;
   If_Result_8 := If_Result_7;
else
   If_Result_8 := False;
end if;
Let_Result := If_Result_8; 
--# end
--# expr-done 3318
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Matching_Assign_Type;
--# end
--# property-start BaseTypeDecl.matching_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5932
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Matching_Type
  (Node : Bare_Base_Type_Decl
      ; Expected_Type : Internal_Entity_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind expected_type Expected_Type
   --# bind origin Origin
   Property_Result : Boolean;
      Block_Var : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
Not_Val : Boolean;
Is_Null_1 : Boolean;
Not_Val_1 : Boolean;
If_Result : Boolean;
Fld : Internal_Entity;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Is_Equiv : Boolean;
Fld_1 : Boolean;
If_Result_1 : Boolean;
Fld_2 : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Is_Equiv_1 : Boolean;
Fld_3 : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Boolean;
Fld_4 : Internal_Entity;
Cast_Expr_2 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Is_Equiv_2 : Boolean;
Fld_5 : Boolean;
If_Result_4 : Boolean;
If_Result_5 : Boolean;
Fld_6 : Internal_Entity;
Cast_Expr_3 : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity;
Is_Equiv_3 : Boolean;
Fld_7 : Boolean;
If_Result_6 : Boolean;
If_Result_7 : Boolean;
Fld_8 : Internal_Entity_Base_Type_Decl;
Fld_9 : Internal_Entity_Base_Type_Decl;
Cast_Result_4 : Internal_Entity;
Cast_Result_5 : Internal_Entity;
Is_Equiv_4 : Boolean;
If_Result_8 : Boolean;
Fld_10 : Boolean;
If_Result_9 : Boolean;
If_Result_10 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3339 '<Block at ???>' Let_Result None
--# scope-start
Block_Var := Ent; 
--# bind actual_type Block_Var
--# expr-start 3320 '<Not at ast.py:5938>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5938
--# expr-start 3319 '<IsNull at ast.py:5938>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5938
Is_Null := Expected_Type.Node = null; 
--# expr-done 3319
Not_Val := not (Is_Null); 
--# expr-done 3320
if Not_Val then
   --# expr-start 3322 '<Not at ast.py:5939>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5939
--# expr-start 3321 '<IsNull at ast.py:5939>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5939
Is_Null_1 := Block_Var.Node = null; 
--# expr-done 3321
Not_Val_1 := not (Is_Null_1); 
--# expr-done 3322
   If_Result := Not_Val_1;
else
   If_Result := False;
end if;
if If_Result then
--# expr-start 3324 '<Eq at ast.py:5941>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5941
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
--# expr-start 3323 '<FieldAccess for universal_int_type at ast.py:5941>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5941
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.universal_int_type
Fld := Libadalang.Implementation.Ada_Node_P_Universal_Int_Type (Node => Self);
--# end
--# expr-done 3323
Is_Equiv := Equivalent (Cast_Result, Fld); 
--# expr-done 3324
if Is_Equiv then
   --# expr-start 3325 '<FieldAccess for allows_universal_int at ast.py:5942>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5942
      if Expected_Type.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.allows_universal_int
Fld_1 := Libadalang.Implementation.Base_Type_Decl_P_Allows_Universal_Int (Node => Expected_Type.Node, Origin => Origin, E_Info => Expected_Type.Info);
--# end
--# expr-done 3325
   If_Result_1 := Fld_1;
else
   If_Result_1 := False;
end if;
if If_Result_1 then
   If_Result_3 := True;
else
--# expr-start 3327 '<Eq at ast.py:5944>' Is_Equiv_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5944
Cast_Expr_1 := Expected_Type; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
--# expr-start 3326 '<FieldAccess for universal_int_type at ast.py:5944>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5944
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.universal_int_type
Fld_2 := Libadalang.Implementation.Ada_Node_P_Universal_Int_Type (Node => Self);
--# end
--# expr-done 3326
Is_Equiv_1 := Equivalent (Cast_Result_1, Fld_2); 
--# expr-done 3327
if Is_Equiv_1 then
   --# expr-start 3328 '<FieldAccess for is_int_type at ast.py:5945>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5945
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_int_type'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Int_Type (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 3328
   If_Result_2 := Fld_3;
else
   If_Result_2 := False;
end if;
   If_Result_3 := If_Result_2;
end if;
if If_Result_3 then
   If_Result_5 := True;
else
--# expr-start 3330 '<Eq at ast.py:5947>' Is_Equiv_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5947
Cast_Expr_2 := Block_Var; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
--# expr-start 3329 '<FieldAccess for universal_real_type at ast.py:5947>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5947
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.universal_real_type
Fld_4 := Libadalang.Implementation.Ada_Node_P_Universal_Real_Type (Node => Self);
--# end
--# expr-done 3329
Is_Equiv_2 := Equivalent (Cast_Result_2, Fld_4); 
--# expr-done 3330
if Is_Equiv_2 then
   --# expr-start 3331 '<FieldAccess for allows_universal_real at ast.py:5948>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5948
      if Expected_Type.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.allows_universal_real
Fld_5 := Libadalang.Implementation.Base_Type_Decl_P_Allows_Universal_Real (Node => Expected_Type.Node, Origin => Origin, E_Info => Expected_Type.Info);
--# end
--# expr-done 3331
   If_Result_4 := Fld_5;
else
   If_Result_4 := False;
end if;
   If_Result_5 := If_Result_4;
end if;
if If_Result_5 then
   If_Result_7 := True;
else
--# expr-start 3333 '<Eq at ast.py:5950>' Is_Equiv_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5950
Cast_Expr_3 := Expected_Type; 
      Cast_Result_3 := Create_Internal_Entity
        (Node => Cast_Expr_3.Node,
         Info => Cast_Expr_3.Info);
--# expr-start 3332 '<FieldAccess for universal_real_type at ast.py:5950>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5950
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.universal_real_type
Fld_6 := Libadalang.Implementation.Ada_Node_P_Universal_Real_Type (Node => Self);
--# end
--# expr-done 3332
Is_Equiv_3 := Equivalent (Cast_Result_3, Fld_6); 
--# expr-done 3333
if Is_Equiv_3 then
   --# expr-start 3334 '<FieldAccess for is_real_type at ast.py:5951>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5951
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_real_type'
Fld_7 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Real_Type (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 3334
   If_Result_6 := Fld_7;
else
   If_Result_6 := False;
end if;
   If_Result_7 := If_Result_6;
end if;
if If_Result_7 then
   If_Result_8 := True;
else
   --# expr-start 3337 '<Eq at ast.py:5953>' Is_Equiv_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5953
--# expr-start 3335 '<FieldAccess for canonical_type at ast.py:5953>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5953
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.canonical_type'
Fld_8 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Canonical_Type (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 3335
      Cast_Result_4 := Create_Internal_Entity
        (Node => Fld_8.Node,
         Info => Fld_8.Info);
--# expr-start 3336 '<FieldAccess for canonical_type at ast.py:5953>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5953
      if Expected_Type.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.canonical_type'
Fld_9 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Canonical_Type (Node => Expected_Type.Node, Origin => Origin, E_Info => Expected_Type.Info);
--# end
--# expr-done 3336
      Cast_Result_5 := Create_Internal_Entity
        (Node => Fld_9.Node,
         Info => Fld_9.Info);
Is_Equiv_4 := Equivalent (Cast_Result_4, Cast_Result_5); 
--# expr-done 3337
   If_Result_8 := Is_Equiv_4;
end if;
if If_Result_8 then
   If_Result_9 := True;
else
   --# expr-start 3338 '<FieldAccess for matching_access_type at ast.py:5955>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5955
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_access_type
Fld_10 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Access_Type (Node => Block_Var.Node, Expected_Type => Expected_Type, For_Assignment => False, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 3338
   If_Result_9 := Fld_10;
end if;
   If_Result_10 := If_Result_9;
else
   If_Result_10 := False;
end if;
Let_Result := If_Result_10; 
--# end
--# expr-done 3339
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Matching_Type;
--# end
--# property-start BaseTypeDecl.matching_allocator_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5960
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Matching_Allocator_Type
  (Node : Bare_Base_Type_Decl
      ; Allocated_Type : Internal_Entity_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind allocated_type Allocated_Type
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Boolean;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3343 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3340 '<FieldAccess for is_access_type at ast.py:5962>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5962
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_access_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Access_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3340
if Fld then
   --# expr-start 3342 '<FieldAccess for matching_formal_type at ast.py:5963>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5963
      if Allocated_Type.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3341 '<FieldAccess for accessed_type at ast.py:5963>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5963
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3341
--# property-call-start BaseTypeDecl.matching_formal_type
Fld_2 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Type (Node => Allocated_Type.Node, Formal_Type => Fld_1, Origin => Origin, E_Info => Allocated_Type.Info);
--# end
--# expr-done 3342
   If_Result := Fld_2;
else
   If_Result := False;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 3343
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Matching_Allocator_Type;
--# end
--# property-start '[dispatcher]BaseTypeDecl.canonical_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Canonical_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Incomplete_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start BaseTypeDecl.canonical_type
                  Property_Result := Base_Type_Decl_P_Canonical_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.canonical_type
                  Property_Result := Base_Subtype_Decl_P_Canonical_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Classwide_Type_Decl =>
                  --# property-call-start ClasswideTypeDecl.canonical_type
                  Property_Result := Classwide_Type_Decl_P_Canonical_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Canonical_Type;
--# end
--# property-start BaseTypeDecl.classwide_type_node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5979
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Classwide_Type_Node
  (Node : Bare_Base_Type_Decl
  )
   return Bare_Classwide_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Classwide_Type_Decl;
      Fld : Bare_Defining_Name;
New_Node : Bare_Classwide_Type_Decl;
Let_Result : Bare_Classwide_Type_Decl;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Type_Decl_Base_Type_Decl_P_Classwide_Type_Node,
            Items    => new Mmz_Key_Array (1 ..  1))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Type_Decl,
                                As_Bare_Base_Type_Decl => Self);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Bare_Classwide_Type_Decl;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
      end if;
   --# scope-start
      --# expr-start 3346 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3345 '<New ClasswideTypeDecl at ???>' New_Node None
--# expr-start 3344 '<FieldAccess for name at ast.py:5980>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5980
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Base_Type_Decl_F_Name;
--# expr-done 3344
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node := new Root_Node_Record
  (Ada_Classwide_Type_Decl);
Initialize
  (Self => New_Node,
   Kind => Ada_Classwide_Type_Decl,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node);
   Initialize_Fields_For_Classwide_Type_Decl
     (Self => New_Node,
      Base_Type_Decl_F_Name => Fld);
--# expr-done 3345
Let_Result := New_Node; 
--# end
--# expr-done 3346
      Property_Result := Let_Result;
   --# end
      if Self /= null then
            Mmz_Val := (Kind => Mmz_Bare_Classwide_Type_Decl,
                        As_Bare_Classwide_Type_Decl => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Classwide_Type_Node;
--# end
--# property-start BaseTypeDecl.scalar_base_subtype_node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5985
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Scalar_Base_Subtype_Node
  (Node : Bare_Base_Type_Decl
  )
   return Bare_Discrete_Base_Subtype_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Discrete_Base_Subtype_Decl;
      Fld : Bare_Defining_Name;
New_Node : Bare_Discrete_Base_Subtype_Decl;
Let_Result : Bare_Discrete_Base_Subtype_Decl;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Type_Decl_Base_Type_Decl_P_Scalar_Base_Subtype_Node,
            Items    => new Mmz_Key_Array (1 ..  1))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Type_Decl,
                                As_Bare_Base_Type_Decl => Self);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Bare_Discrete_Base_Subtype_Decl;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
      end if;
   --# scope-start
      --# expr-start 3349 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3348 '<New DiscreteBaseSubtypeDecl at ast.py:5990>' New_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5990
--# expr-start 3347 '<FieldAccess for name at ast.py:5990>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5990
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Base_Type_Decl_F_Name;
--# expr-done 3347
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node := new Root_Node_Record
  (Ada_Discrete_Base_Subtype_Decl);
Initialize
  (Self => New_Node,
   Kind => Ada_Discrete_Base_Subtype_Decl,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node);
   Initialize_Fields_For_Discrete_Base_Subtype_Decl
     (Self => New_Node,
      Base_Type_Decl_F_Name => Fld);
--# expr-done 3348
Let_Result := New_Node; 
--# end
--# expr-done 3349
      Property_Result := Let_Result;
   --# end
      if Self /= null then
            Mmz_Val := (Kind => Mmz_Bare_Discrete_Base_Subtype_Decl,
                        As_Bare_Discrete_Base_Subtype_Decl => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Scalar_Base_Subtype_Node;
--# end
--# property-start BaseTypeDecl.scalar_base_subtype /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5993
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Scalar_Base_Subtype
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Discrete_Base_Subtype_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Discrete_Base_Subtype_Decl;
      Fld : Bare_Discrete_Base_Subtype_Decl;
As_Entity : Internal_Entity_Discrete_Base_Subtype_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Discrete_Base_Subtype_Decl;
Let_Result : Internal_Entity_Discrete_Base_Subtype_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3352 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3351 '<as_entity at ast.py:5998>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5998
--# expr-start 3350 '<FieldAccess for scalar_base_subtype_node at ast.py:5998>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5998
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.scalar_base_subtype_node
Fld := Libadalang.Implementation.Base_Type_Decl_P_Scalar_Base_Subtype_Node (Node => Self);
--# end
--# expr-done 3350
Is_Null := Fld = null; 
if Is_Null then
   If_Result := No_Entity_Discrete_Base_Subtype_Decl;
else
As_Entity := (Info => E_Info, Node => Fld); 
   If_Result := As_Entity;
end if;
--# expr-done 3351
Let_Result := If_Result; 
--# end
--# expr-done 3352
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Scalar_Base_Subtype;
--# end
--# property-start '[dispatcher]BaseTypeDecl.previous_part' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Previous_Part
  (Node : Bare_Base_Type_Decl
      ; Go_To_Incomplete : Boolean
         := True
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind go_to_incomplete Go_To_Incomplete
   Property_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl | Ada_Incomplete_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start BaseTypeDecl.previous_part
                  Property_Result := Base_Type_Decl_P_Previous_Part
                    (Self
                        , Go_To_Incomplete
                        , E_Info
                    );
                  --# end
               when Ada_Classwide_Type_Decl =>
                  --# property-call-start ClasswideTypeDecl.previous_part
                  Property_Result := Classwide_Type_Decl_P_Previous_Part
                    (Self
                        , Go_To_Incomplete
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Previous_Part;
--# end
--# property-start BaseTypeDecl.next_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6033
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Next_Part
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Type_Decl;
      Match_0 : Internal_Entity_Incomplete_Type_Decl;
Match_1 : Internal_Entity_Base_Type_Decl;
Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Bare_Declarative_Part;
Var_Expr_2 : Boolean;
Var_Expr_3 : Internal_Entity_Package_Body;
Fld : Bare_Declarative_Part;
As_Entity : Internal_Entity_Declarative_Part;
Is_Null : Boolean;
If_Result : Internal_Entity_Declarative_Part;
Fld_1 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Boolean;
Fld_3 : Bare_Declarative_Part;
Fld_4 : Bare_Ada_Node;
Cast_Result : Bare_Base_Package_Decl;
As_Entity_1 : Internal_Entity_Base_Package_Decl;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Base_Package_Decl;
Fld_5 : Internal_Entity_Package_Body;
Fld_6 : Internal_Entity_Declarative_Part;
Fld_7 : Internal_Entity_Base_Type_Decl;
Scope_Result_2 : Internal_Entity_Base_Type_Decl;
Result_Var_1 : Internal_Entity_Base_Type_Decl;
Scope_Result_3 : Internal_Entity_Base_Type_Decl;
Result_Var_2 : Internal_Entity_Base_Type_Decl;
Result_Var_3 : Internal_Entity_Base_Type_Decl;
Fld_8 : Internal_Entity_Base_Type_Decl;
Fld_9 : Boolean;
Fld_10 : Boolean;
Not_Val : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Internal_Entity_Base_Type_Decl;
Match_Prefix : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity_Incomplete_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
Scope_Result_4 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity_Base_Type_Decl;
Let_Result_1 : Internal_Entity_Base_Type_Decl;
Scope_Result_5 : Internal_Entity_Base_Type_Decl;
Match_Result : Internal_Entity_Base_Type_Decl;
Let_Result_2 : Internal_Entity_Base_Type_Decl;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Type_Decl_Base_Type_Decl_P_Next_Part,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Type_Decl,
                                As_Bare_Base_Type_Decl => Self);
               Mmz_K.Items (2) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Base_Type_Decl;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 3370 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 3369 '<Match at ast.py:6043>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6043
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Base_Type_Decl (Match_Prefix.Node.Kind) is
         when Ada_Incomplete_Type_Decl .. Ada_Incomplete_Tagged_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Incomplete_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_1; 
--# bind itd Match_0
--# expr-start 3353 '<FieldAccess for declarative_scope at ast.py:6047>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6047
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.declarative_scope
Fld := Libadalang.Implementation.Ada_Node_P_Declarative_Scope (Node => Match_0.Node);
--# end
--# expr-done 3353
Var_Expr_1 := Fld;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind s Var_Expr_1
--# expr-start 3355 '<FieldAccess for find_next_part_in at ast.py:6048>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6048
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3354 '<as_entity at ast.py:6048>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6048
Is_Null := Var_Expr_1 = null; 
if Is_Null then
   If_Result := No_Entity_Declarative_Part;
else
As_Entity := (Info => E_Info, Node => Var_Expr_1); 
   If_Result := As_Entity;
end if;
--# expr-done 3354
--# property-call-start IncompleteTypeDecl.find_next_part_in
Fld_1 := Libadalang.Implementation.Incomplete_Type_Decl_P_Find_Next_Part_In (Node => Match_0.Node, Decl_Part => If_Result, E_Info => Match_0.Info);
--# end
--# expr-done 3355
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Var_Expr := Result_Var;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind e Var_Expr
Scope_Result_1 := Var_Expr;
   --# end
   Result_Var_3 := Scope_Result_1;
else
--# expr-start 3356 '<FieldAccess for is_in_private_part at ast.py:6054>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6054
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_in_private_part'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_In_Private_Part (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3356
Var_Expr_2 := Fld_2;
if Var_Expr_2 /= False then
   --# scope-start
   --# bind _ Var_Expr_2
--# expr-start 3361 '<FieldAccess for body_part at ast.py:6056>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6056
--# expr-start 3360 '<as_entity at ast.py:6056>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6056
--# expr-start 3359 '<Cast to BasePackageDecl at ast.py:6056>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6056
--# expr-start 3358 '<parent at ast.py:6056>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6056
--# expr-start 3357 '<FieldAccess for declarative_scope at ast.py:6056>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6056
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.declarative_scope
Fld_3 := Libadalang.Implementation.Ada_Node_P_Declarative_Scope (Node => Ent.Node);
--# end
--# expr-done 3357
      if Fld_3 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Fld_3.Parent; 
--# expr-done 3358
   if Fld_4 = null
      or else Fld_4.Kind in Ada_Base_Package_Decl
   then
      Cast_Result := Fld_4;
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 3359
Is_Null_1 := Cast_Result = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Base_Package_Decl;
else
As_Entity_1 := (Info => E_Info, Node => Cast_Result); 
   If_Result_1 := As_Entity_1;
end if;
--# expr-done 3360
      if If_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasePackageDecl.body_part
Fld_5 := Libadalang.Implementation.Base_Package_Decl_P_Body_Part (Node => If_Result_1.Node, E_Info => If_Result_1.Info);
--# end
--# expr-done 3361
Var_Expr_3 := Fld_5;
if Var_Expr_3 /= No_Entity_Package_Body then
   --# scope-start
   --# bind p Var_Expr_3
--# expr-start 3363 '<FieldAccess for find_next_part_in at ast.py:6058>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6058
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3362 '<FieldAccess for decls at ast.py:6058>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6058
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Declarative_Part (Node => Var_Expr_3.Node.Package_Body_F_Decls, Info => Var_Expr_3.Info);
--# expr-done 3362
--# property-call-start IncompleteTypeDecl.find_next_part_in
Fld_7 := Libadalang.Implementation.Incomplete_Type_Decl_P_Find_Next_Part_In (Node => Match_0.Node, Decl_Part => Fld_6, E_Info => Match_0.Info);
--# end
--# expr-done 3363
Scope_Result_2 := Fld_7;
   --# end
   Result_Var_1 := Scope_Result_2;
else
   Result_Var_1 := No_Entity_Base_Type_Decl;
end if;
Scope_Result_3 := Result_Var_1;
   --# end
   Result_Var_2 := Scope_Result_3;
else
   Result_Var_2 := No_Entity_Base_Type_Decl;
end if;
   Result_Var_3 := Result_Var_2;
end if;
Let_Result := Result_Var_3; 
--# end
Scope_Result_4 := Let_Result;
   --# end
            Match_Result := Scope_Result_4; 
         when Ada_Discrete_Base_Subtype_Decl .. Ada_Classwide_Type_Decl | Ada_Protected_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Base_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_2; 
--# bind _ Match_1
--# expr-start 3368 '<If at ast.py:6062>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6062
--# expr-start 3364 '<FieldAccess for is_private at ast.py:6063>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6063
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_private'
Fld_9 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Private (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3364
if Fld_9 then
   --# expr-start 3366 '<Not at ast.py:6064>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6064
--# expr-start 3365 '<FieldAccess for is_generic_formal at ast.py:6064>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6064
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_generic_formal
Fld_10 := Libadalang.Implementation.Base_Type_Decl_P_Is_Generic_Formal (Node => Ent.Node);
--# end
--# expr-done 3365
Not_Val := not (Fld_10); 
--# expr-done 3366
   If_Result_2 := Not_Val;
else
   If_Result_2 := False;
end if;
if If_Result_2 then
   --# expr-start 3367 '<FieldAccess for private_completion at ast.py:6065>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6065
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.private_completion
Fld_8 := Libadalang.Implementation.Base_Type_Decl_P_Private_Completion (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3367
   If_Result_3 := Fld_8;
else
   If_Result_3 := No_Entity_Base_Type_Decl;
end if;
--# expr-done 3368
Let_Result_1 := If_Result_3; 
--# end
Scope_Result_5 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_5; 
end case;
--# expr-done 3369
Let_Result_2 := Match_Result; 
--# end
--# expr-done 3370
      Property_Result := Let_Result_2;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Base_Type_Decl,
                        As_Internal_Entity_Base_Type_Decl => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Next_Part;
--# end
--# property-start BaseTypeDecl.full_view /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6072
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Full_View
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3373 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3371 '<FieldAccess for next_part at ast.py:6076>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6076
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.next_part
Fld := Libadalang.Implementation.Base_Type_Decl_P_Next_Part (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3371
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind np Var_Expr
--# expr-start 3372 '<FieldAccess for full_view at ast.py:6077>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6077
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.full_view
Fld_1 := Libadalang.Implementation.Base_Type_Decl_P_Full_View (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 3372
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := Ent;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 3373
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Full_View;
--# end
--# property-start BaseTypeDecl.is_definite_subtype /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6083
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Definite_Subtype
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Match_0 : Internal_Entity_Incomplete_Type_Decl;
Match_1 : Internal_Entity_Type_Decl;
Match_2 : Internal_Entity_Subtype_Decl;
Match_3 : Internal_Entity_Classwide_Type_Decl;
Match_4 : Internal_Entity_Task_Type_Decl;
Match_5 : Internal_Entity_Protected_Type_Decl;
Match_6 : Internal_Entity_Base_Type_Decl;
Match_0_1 : Internal_Entity_Derived_Type_Def;
Match_1_1 : Internal_Entity_Array_Type_Def;
Match_2_1 : Internal_Entity_Type_Def;
Fld : Internal_Entity_Discriminant_Part;
Is_Null : Boolean;
Fld_1 : Internal_Entity_Type_Def;
Fld_2 : Internal_Entity_Subtype_Indication;
Fld_3 : Internal_Entity_Constraint;
Is_Null_1 : Boolean;
Not_Val : Boolean;
Fld_4 : Internal_Entity_Base_Type_Decl;
Fld_5 : Boolean;
If_Result : Boolean;
Fld_6 : Internal_Entity_Array_Indices;
Is_A : Boolean;
Match_Prefix : Internal_Entity_Type_Def;
Cast_Result : Internal_Entity_Derived_Type_Def;
Let_Result : Boolean;
Scope_Result : Boolean;
Cast_Result_1 : Internal_Entity_Array_Type_Def;
Let_Result_1 : Boolean;
Scope_Result_1 : Boolean;
Cast_Result_2 : Internal_Entity_Type_Def;
Let_Result_2 : Boolean;
Scope_Result_2 : Boolean;
Match_Result : Boolean;
If_Result_1 : Boolean;
Fld_7 : Internal_Entity_Subtype_Indication;
Fld_8 : Internal_Entity_Constraint;
Is_Null_2 : Boolean;
Not_Val_1 : Boolean;
Fld_9 : Internal_Entity_Base_Type_Decl;
Fld_10 : Boolean;
If_Result_2 : Boolean;
Fld_11 : Internal_Entity_Discriminant_Part;
Is_Null_3 : Boolean;
Fld_12 : Internal_Entity_Discriminant_Part;
Is_Null_4 : Boolean;
Match_Prefix_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity_Incomplete_Type_Decl;
Let_Result_3 : Boolean;
Scope_Result_3 : Boolean;
Cast_Result_4 : Internal_Entity_Type_Decl;
Let_Result_4 : Boolean;
Scope_Result_4 : Boolean;
Cast_Result_5 : Internal_Entity_Subtype_Decl;
Let_Result_5 : Boolean;
Scope_Result_5 : Boolean;
Cast_Result_6 : Internal_Entity_Classwide_Type_Decl;
Let_Result_6 : Boolean;
Scope_Result_6 : Boolean;
Cast_Result_7 : Internal_Entity_Task_Type_Decl;
Let_Result_7 : Boolean;
Scope_Result_7 : Boolean;
Cast_Result_8 : Internal_Entity_Protected_Type_Decl;
Let_Result_8 : Boolean;
Scope_Result_8 : Boolean;
Cast_Result_9 : Internal_Entity_Base_Type_Decl;
Let_Result_9 : Boolean;
Scope_Result_9 : Boolean;
Match_Result_1 : Boolean;
Let_Result_10 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3397 '<Block at ???>' Let_Result_10 None
--# scope-start
--# expr-start 3396 '<Match at ast.py:6090>' Match_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6090
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix_1 := Ent; 
case Ada_Base_Type_Decl (Match_Prefix_1.Node.Kind) is
         when Ada_Incomplete_Type_Decl .. Ada_Incomplete_Tagged_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Incomplete_Type_Decl
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_0 := Cast_Result_3; 
--# bind _ Match_0
Let_Result_3 := False; 
--# end
Scope_Result_3 := Let_Result_3;
   --# end
            Match_Result_1 := Scope_Result_3; 
         when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Type_Decl
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_1 := Cast_Result_4; 
--# bind td Match_1
--# expr-start 3375 '<IsNull at ast.py:6092>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6092
--# expr-start 3374 '<FieldAccess for discriminants at ast.py:6092>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6092
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Discriminant_Part (Node => Match_1.Node.Type_Decl_F_Discriminants, Info => Match_1.Info);
--# expr-done 3374
Is_Null := Fld.Node = null; 
--# expr-done 3375
if Is_Null then
   --# expr-start 3385 '<Match at ast.py:6092>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6092
--# expr-start 3376 '<FieldAccess for type_def at ast.py:6092>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6092
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Type_Def (Node => Match_1.Node.Type_Decl_F_Type_Def, Info => Match_1.Info);
--# expr-done 3376
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld_1; 
case Ada_Type_Def (Match_Prefix.Node.Kind) is
         when Ada_Derived_Type_Def =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Derived_Type_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0_1 := Cast_Result; 
--# bind dtd Match_0_1
--# expr-start 3380 '<Not at ast.py:6094>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6094
--# expr-start 3379 '<IsNull at ast.py:6094>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6094
--# expr-start 3378 '<FieldAccess for constraint at ast.py:6094>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6094
--# expr-start 3377 '<FieldAccess for subtype_indication at ast.py:6094>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6094
      if Match_0_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Subtype_Indication (Node => Match_0_1.Node.Derived_Type_Def_F_Subtype_Indication, Info => Match_0_1.Info);
--# expr-done 3377
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Constraint (Node => Fld_2.Node.Subtype_Indication_F_Constraint, Info => Fld_2.Info);
--# expr-done 3378
Is_Null_1 := Fld_3.Node = null; 
--# expr-done 3379
Not_Val := not (Is_Null_1); 
--# expr-done 3380
if Not_Val then
   If_Result := True;
else
   --# expr-start 3382 '<FieldAccess for is_definite_subtype at ast.py:6095>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6095
--# expr-start 3381 '<FieldAccess for base_type at ast.py:6095>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6095
      if Match_0_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld_4 := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => Match_0_1.Node, Origin => Origin, E_Info => Match_0_1.Info);
--# end
--# expr-done 3381
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_definite_subtype
Fld_5 := Libadalang.Implementation.Base_Type_Decl_P_Is_Definite_Subtype (Node => Fld_4.Node, Origin => Origin, E_Info => Fld_4.Info);
--# end
--# expr-done 3382
   If_Result := Fld_5;
end if;
Let_Result := If_Result; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Array_Type_Def =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Array_Type_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1_1 := Cast_Result_1; 
--# bind atd Match_1_1
--# expr-start 3384 '<IsA ConstrainedArrayIndices at ast.py:6098>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6098
--# expr-start 3383 '<FieldAccess for indices at ast.py:6098>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6098
      if Match_1_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Array_Indices (Node => Match_1_1.Node.Array_Type_Def_F_Indices, Info => Match_1_1.Info);
--# expr-done 3383
Is_A := Fld_6.Node /= null 
and then Fld_6.Node.Kind in Ada_Constrained_Array_Indices_Range; 
--# expr-done 3384
Let_Result_1 := Is_A; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Access_To_Subp_Def .. Ada_Type_Access_Def | Ada_Enum_Type_Def .. Ada_Signed_Int_Type_Def =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Type_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2_1 := Cast_Result_2; 
--# bind _ Match_2_1
Let_Result_2 := True; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 3385
   If_Result_1 := Match_Result;
else
   If_Result_1 := False;
end if;
Let_Result_4 := If_Result_1; 
--# end
Scope_Result_4 := Let_Result_4;
   --# end
            Match_Result_1 := Scope_Result_4; 
         when Ada_Subtype_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Create_Internal_Entity_Subtype_Decl
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_2 := Cast_Result_5; 
--# bind st Match_2
--# expr-start 3389 '<Not at ast.py:6102>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6102
--# expr-start 3388 '<IsNull at ast.py:6102>' Is_Null_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6102
--# expr-start 3387 '<FieldAccess for constraint at ast.py:6102>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6102
--# expr-start 3386 '<FieldAccess for subtype at ast.py:6102>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6102
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Subtype_Indication (Node => Match_2.Node.Subtype_Decl_F_Subtype, Info => Match_2.Info);
--# expr-done 3386
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Constraint (Node => Fld_7.Node.Subtype_Indication_F_Constraint, Info => Fld_7.Info);
--# expr-done 3387
Is_Null_2 := Fld_8.Node = null; 
--# expr-done 3388
Not_Val_1 := not (Is_Null_2); 
--# expr-done 3389
if Not_Val_1 then
   If_Result_2 := True;
else
   --# expr-start 3391 '<FieldAccess for is_definite_subtype at ast.py:6103>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6103
--# expr-start 3390 '<FieldAccess for get_type at ast.py:6103>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6103
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld_9 := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Match_2.Node, Origin => Origin, E_Info => Match_2.Info);
--# end
--# expr-done 3390
      if Fld_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_definite_subtype
Fld_10 := Libadalang.Implementation.Base_Type_Decl_P_Is_Definite_Subtype (Node => Fld_9.Node, Origin => Origin, E_Info => Fld_9.Info);
--# end
--# expr-done 3391
   If_Result_2 := Fld_10;
end if;
Let_Result_5 := If_Result_2; 
--# end
Scope_Result_5 := Let_Result_5;
   --# end
            Match_Result_1 := Scope_Result_5; 
         when Ada_Classwide_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_6 := Create_Internal_Entity_Classwide_Type_Decl
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_3 := Cast_Result_6; 
--# bind _ Match_3
Let_Result_6 := False; 
--# end
Scope_Result_6 := Let_Result_6;
   --# end
            Match_Result_1 := Scope_Result_6; 
         when Ada_Task_Type_Decl .. Ada_Single_Task_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_7 := Create_Internal_Entity_Task_Type_Decl
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_4 := Cast_Result_7; 
--# bind ttd Match_4
--# expr-start 3393 '<IsNull at ast.py:6105>' Is_Null_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6105
--# expr-start 3392 '<FieldAccess for discriminants at ast.py:6105>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6105
      if Match_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Create_Internal_Entity_Discriminant_Part (Node => Match_4.Node.Task_Type_Decl_F_Discriminants, Info => Match_4.Info);
--# expr-done 3392
Is_Null_3 := Fld_11.Node = null; 
--# expr-done 3393
Let_Result_7 := Is_Null_3; 
--# end
Scope_Result_7 := Let_Result_7;
   --# end
            Match_Result_1 := Scope_Result_7; 
         when Ada_Protected_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_8 := Create_Internal_Entity_Protected_Type_Decl
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_5 := Cast_Result_8; 
--# bind ptd Match_5
--# expr-start 3395 '<IsNull at ast.py:6106>' Is_Null_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6106
--# expr-start 3394 '<FieldAccess for discriminants at ast.py:6106>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6106
      if Match_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Create_Internal_Entity_Discriminant_Part (Node => Match_5.Node.Protected_Type_Decl_F_Discriminants, Info => Match_5.Info);
--# expr-done 3394
Is_Null_4 := Fld_12.Node = null; 
--# expr-done 3395
Let_Result_8 := Is_Null_4; 
--# end
Scope_Result_8 := Let_Result_8;
   --# end
            Match_Result_1 := Scope_Result_8; 
         when Ada_Discrete_Base_Subtype_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_9 := Create_Internal_Entity_Base_Type_Decl
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_6 := Cast_Result_9; 
--# bind _ Match_6
Let_Result_9 := True; 
--# end
Scope_Result_9 := Let_Result_9;
   --# end
            Match_Result_1 := Scope_Result_9; 
end case;
--# expr-done 3396
Let_Result_10 := Match_Result_1; 
--# end
--# expr-done 3397
      Property_Result := Let_Result_10;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Definite_Subtype;
--# end
--# property-start '[dispatcher]BaseTypeDecl.is_private' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Is_Private
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Incomplete_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.is_private
                  Property_Result := Base_Type_Decl_P_Is_Private
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.is_private
                  Property_Result := Base_Subtype_Decl_P_Is_Private
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Classwide_Type_Decl =>
                  --# property-call-start ClasswideTypeDecl.is_private
                  Property_Result := Classwide_Type_Decl_P_Is_Private
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.is_private
                  Property_Result := Type_Decl_P_Is_Private
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Is_Private;
--# end
--# property-start '[dispatcher]BaseTypeDecl.discriminants_list' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Discriminants_List
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.discriminants_list
                  Property_Result := Base_Subtype_Decl_P_Discriminants_List
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Classwide_Type_Decl =>
                  --# property-call-start ClasswideTypeDecl.discriminants_list
                  Property_Result := Classwide_Type_Decl_P_Discriminants_List
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Incomplete_Type_Decl .. Ada_Incomplete_Tagged_Type_Decl =>
                  --# property-call-start IncompleteTypeDecl.discriminants_list
                  Property_Result := Incomplete_Type_Decl_P_Discriminants_List
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Protected_Type_Decl =>
                  --# property-call-start ProtectedTypeDecl.discriminants_list
                  Property_Result := Protected_Type_Decl_P_Discriminants_List
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Task_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start TaskTypeDecl.discriminants_list
                  Property_Result := Task_Type_Decl_P_Discriminants_List
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.discriminants_list
                  Property_Result := Type_Decl_P_Discriminants_List
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Discriminants_List;
--# end
--# property-start BaseTypeDecl.corresponding_char_literal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6130
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Corresponding_Char_Literal
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Basic_Decl;
      Block_Var : Internal_Entity_Base_Type_Decl;
Block_Var_1 : Internal_Entity_Enum_Type_Def;
Block_Var_2 : Symbol_Type;
Block_Var_3 : Internal_Entity_Enum_Literal_Decl;
Var_Expr : Internal_Entity_Type_Decl;
Var_Expr_1 : Bare_Char_Literal;
Var_Expr_2 : Internal_Entity_Enum_Type_Def;
Fld : Internal_Entity_Base_Type_Decl;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity_Type_Decl;
Fld_1 : Internal_Entity_Type_Def;
Scope_Result : Internal_Entity_Type_Def;
Result_Var : Internal_Entity_Type_Def;
Cast_Result_1 : Internal_Entity_Enum_Type_Def;
Cast_Expr_1 : Bare_Ada_Node;
Cast_Result_2 : Bare_Char_Literal;
Sym : Symbol_Type;
Scope_Result_1 : Symbol_Type;
Result_Var_1 : Symbol_Type;
Fld_2 : Internal_Entity_Enum_Literal_Decl_List;
Scope_Result_2 : Internal_Entity_Enum_Literal_Decl_List;
Result_Var_2 : Internal_Entity_Enum_Literal_Decl_List;
Fld_3 : Bare_Enum_Literal_Decl_List;
Fld_4 : Internal_Entity_Info;
As_Entity : Internal_Entity_Enum_Literal_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Enum_Literal_Decl;
Fld_5 : Internal_Entity_Defining_Name;
Fld_6 : Boolean;
Item_135 : Internal_Entity_Enum_Literal_Decl;
Bare_Item_135 : Bare_Enum_Literal_Decl;
Find_Result : Internal_Entity_Enum_Literal_Decl;
Fld_7 : Internal_Entity;
Cast_Result_3 : Internal_Entity_Type_Decl;
New_Node : Bare_Synthetic_Char_Enum_Lit;
As_Entity_1 : Internal_Entity_Synthetic_Char_Enum_Lit;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Synthetic_Char_Enum_Lit;
Cast_Result_4 : Internal_Entity_Enum_Literal_Decl;
Is_Null_2 : Boolean;
Fld_8 : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Internal_Entity_Enum_Literal_Decl;
Let_Result : Internal_Entity_Enum_Literal_Decl;
Cast_Result_5 : Internal_Entity_Basic_Decl;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Type_Decl_Base_Type_Decl_P_Corresponding_Char_Literal,
            Items    => new Mmz_Key_Array (1 ..  3))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Type_Decl,
                                As_Bare_Base_Type_Decl => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Bare_Ada_Node,
                                      As_Bare_Ada_Node => Origin);
               Mmz_K.Items (3) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Basic_Decl;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
--# expr-start 3415 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3398 '<FieldAccess for root_type at ast.py:6140>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6140
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.root_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Root_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3398
Block_Var := Fld; 
--# bind root Block_Var
--# expr-start 3401 '<Cast to EnumTypeDef at ast.py:6141>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6141
--# expr-start 3399 '<Cast to TypeDecl at ast.py:6141>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6141
Cast_Expr := Block_Var; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Type_Decl_Range
   then
      Cast_Result := Create_Internal_Entity_Type_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Type_Decl;
   end if;
--# expr-done 3399
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Type_Decl then
   --# scope-start
--# expr-start 3400 '<FieldAccess for type_def at ast.py:6141>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6141
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Type_Def (Node => Var_Expr.Node.Type_Decl_F_Type_Def, Info => Var_Expr.Info);
--# expr-done 3400
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Type_Def;
end if;
   if Result_Var.Node = null
      or else Result_Var.Node.Kind in Ada_Enum_Type_Def_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Enum_Type_Def
        (Node => Result_Var.Node,
         Info => Result_Var.Info);
   else
         Cast_Result_1 := No_Entity_Enum_Type_Def;
   end if;
--# expr-done 3401
Block_Var_1 := Cast_Result_1; 
--# bind enum_type Block_Var_1
--# expr-start 3402 '<Cast to CharLiteral at ast.py:6142>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6142
Cast_Expr_1 := Origin; 
   if Cast_Expr_1 = null
      or else Cast_Expr_1.Kind in Ada_Char_Literal_Range
   then
      Cast_Result_2 := Cast_Expr_1;
   else
         Cast_Result_2 := No_Bare_Ada_Node;
   end if;
--# expr-done 3402
Var_Expr_1 := Cast_Result_2;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind l Var_Expr_1
--# expr-start 3403 '<GetSymbol at ast.py:6142>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6142
Sym := Get_Symbol (Var_Expr_1); 
--# expr-done 3403
Scope_Result_1 := Sym;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := null;
end if;
Block_Var_2 := Result_Var_1; 
--# bind sym Block_Var_2
--# expr-start 3407 '<Find at ast.py:6144>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6144
Var_Expr_2 := Block_Var_1;
if Var_Expr_2 /= No_Entity_Enum_Type_Def then
   --# scope-start
--# expr-start 3404 '<FieldAccess for enum_literals at ast.py:6144>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6144
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Enum_Literal_Decl_List (Node => Var_Expr_2.Node.Enum_Type_Def_F_Enum_Literals, Info => Var_Expr_2.Info);
--# expr-done 3404
Scope_Result_2 := Fld_2;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Entity_Enum_Literal_Decl_List;
end if;
Fld_3 := Result_Var_2.Node;
      Find_Result := No_Entity_Enum_Literal_Decl;
      if Fld_3 = null then
      Find_Result := No_Entity_Enum_Literal_Decl;
      else
      declare
         Collection : constant Bare_Enum_Literal_Decl_List := Fld_3;
      begin
         for Untyped_Item_135 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_135 := Untyped_Item_135; 
Is_Null := Bare_Item_135 = null; 
if Is_Null then
   If_Result := No_Entity_Enum_Literal_Decl;
else
Fld_4 := Result_Var_2.Info;
As_Entity := (Info => Fld_4, Node => Bare_Item_135); 
   If_Result := As_Entity;
end if;
                  Item_135 := If_Result; 
   --# scope-start
               --# bind lit_decl Item_135
      --# expr-start 3406 '<FieldAccess for name_is at ast.py:6145>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6145
--# expr-start 3405 '<FieldAccess for name at ast.py:6145>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6145
      if Item_135.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Defining_Name (Node => Item_135.Node.Enum_Literal_Decl_F_Name, Info => Item_135.Info);
--# expr-done 3405
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_is
Fld_6 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_5.Node, Sym => Block_Var_2);
--# end
--# expr-done 3406
      if Fld_6 then
         Find_Result := Item_135;
      end if;
   --# end
               exit when Fld_6;
         end loop;
      end;
      end if;
--# expr-done 3407
Block_Var_3 := Find_Result; 
--# bind char_lit Block_Var_3
--# expr-start 3414 '<If at ast.py:6148>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6148
--# expr-start 3408 '<IsNull at ast.py:6152>' Is_Null_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6152
Is_Null_2 := Block_Var_3.Node = null; 
--# expr-done 3408
if Is_Null_2 then
   --# expr-start 3409 '<FieldAccess for is_std_char_type at ast.py:6152>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6152
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start EnumTypeDef.is_std_char_type
Fld_8 := Libadalang.Implementation.Enum_Type_Def_P_Is_Std_Char_Type (Node => Block_Var_1.Node);
--# end
--# expr-done 3409
   If_Result_2 := Fld_8;
else
   If_Result_2 := False;
end if;
if If_Result_2 then
--# expr-start 3413 '<as_entity at ast.py:6154>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6154
--# expr-start 3412 '<New SyntheticCharEnumLit at ast.py:6154>' New_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6154
--# expr-start 3411 '<Cast to TypeDecl at ast.py:6164>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6164
--# expr-start 3410 '<parent at ast.py:6164>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6164
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Parent (Node => Block_Var_1.Node, E_Info => Block_Var_1.Info);
--# expr-done 3410
   if Fld_7.Node = null
      or else Fld_7.Node.Kind in Ada_Type_Decl_Range
   then
      Cast_Result_3 := Create_Internal_Entity_Type_Decl
        (Node => Fld_7.Node,
         Info => Fld_7.Info);
   else
         Cast_Result_3 := No_Entity_Type_Decl;
   end if;
--# expr-done 3411
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node := new Root_Node_Record
  (Ada_Synthetic_Char_Enum_Lit);
Initialize
  (Self => New_Node,
   Kind => Ada_Synthetic_Char_Enum_Lit,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node);
   Initialize_Fields_For_Synthetic_Char_Enum_Lit
     (Self => New_Node,
      Enum_Literal_Decl_F_Name => No_Bare_Ada_Node);
      New_Node.Synthetic_Char_Enum_Lit_Char_Symbol :=
         Block_Var_2;
      New_Node.Synthetic_Char_Enum_Lit_Enum_Type_Decl :=
         Cast_Result_3;
--# expr-done 3412
Is_Null_1 := New_Node = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Synthetic_Char_Enum_Lit;
else
As_Entity_1 := (Info => E_Info, Node => New_Node); 
   If_Result_1 := As_Entity_1;
end if;
--# expr-done 3413
      Cast_Result_4 := Create_Internal_Entity_Enum_Literal_Decl
        (Node => If_Result_1.Node,
         Info => If_Result_1.Info);
   If_Result_3 := Cast_Result_4;
else
   If_Result_3 := Block_Var_3;
end if;
--# expr-done 3414
Let_Result := If_Result_3; 
--# end
--# expr-done 3415
      Cast_Result_5 := Create_Internal_Entity_Basic_Decl
        (Node => Let_Result.Node,
         Info => Let_Result.Info);
      Property_Result := Cast_Result_5;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Basic_Decl,
                        As_Internal_Entity_Basic_Decl => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Corresponding_Char_Literal;
--# end
--# property-start '[dispatcher]BaseTypeDecl.root_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Type_Decl_P_Root_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Type_Decl (Self.Kind) is
               when Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
                  --# property-call-start BaseTypeDecl.root_type
                  Property_Result := Base_Type_Decl_P_Root_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
                  --# property-call-start BaseSubtypeDecl.root_type
                  Property_Result := Base_Subtype_Decl_P_Root_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
                  --# property-call-start TypeDecl.root_type
                  Property_Result := Type_Decl_P_Root_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Type_Decl_P_Root_Type;
--# end
--# property-start BaseTypeDecl.next_part_for_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6181
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Next_Part_For_Decl
  (Node : Bare_Base_Type_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Match_0 : Internal_Entity_Single_Task_Type_Decl;
Match_1 : Internal_Entity_Task_Type_Decl;
Match_2 : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity_Basic_Decl;
Fld_3 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity_Basic_Decl;
Match_Prefix : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity_Single_Task_Type_Decl;
Let_Result : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Cast_Result_2 : Internal_Entity_Task_Type_Decl;
Let_Result_1 : Internal_Entity_Basic_Decl;
Scope_Result_1 : Internal_Entity_Basic_Decl;
Cast_Result_3 : Internal_Entity_Base_Type_Decl;
Let_Result_2 : Internal_Entity_Basic_Decl;
Scope_Result_2 : Internal_Entity_Basic_Decl;
Match_Result : Internal_Entity_Basic_Decl;
Let_Result_3 : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3422 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 3421 '<Match at ast.py:6182>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6182
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Base_Type_Decl (Match_Prefix.Node.Kind) is
         when Ada_Single_Task_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Single_Task_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_1; 
--# bind sttd Match_0
--# expr-start 3417 '<FieldAccess for basic_decl_next_part_for_decl at ast.py:6186>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6186
--# expr-start 3416 '<FieldAccess for parent_basic_decl at ast.py:6186>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6186
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.parent_basic_decl
Fld := Libadalang.Implementation.Ada_Node_P_Parent_Basic_Decl (Node => Match_0.Node, E_Info => Match_0.Info);
--# end
--# expr-done 3416
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.basic_decl_next_part_for_decl
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Basic_Decl_Next_Part_For_Decl (Node => Fld.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld.Info);
--# end
--# expr-done 3417
Let_Result := Fld_1; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Task_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Task_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_2; 
--# bind ttd Match_1
--# expr-start 3418 '<FieldAccess for basic_decl_next_part_for_decl at ast.py:6187>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6187
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.basic_decl_next_part_for_decl
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Basic_Decl_Next_Part_For_Decl (Node => Match_1.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Match_1.Info);
--# end
--# expr-done 3418
Let_Result_1 := Fld_2; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Discrete_Base_Subtype_Decl .. Ada_Protected_Type_Decl | Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Base_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_3; 
--# bind _ Match_2
--# expr-start 3420 '<Cast to BasicDecl.entity at ast.py:6188>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6188
--# expr-start 3419 '<FieldAccess for next_part at ast.py:6188>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6188
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.next_part
Fld_3 := Libadalang.Implementation.Base_Type_Decl_P_Next_Part (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3419
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
--# expr-done 3420
Let_Result_2 := Cast_Result; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 3421
Let_Result_3 := Match_Result; 
--# end
--# expr-done 3422
      Property_Result := Let_Result_3;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Next_Part_For_Decl;
--# end
--# property-start BaseTypeDecl.shapes /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6193
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Shapes
  (Node : Bare_Base_Type_Decl
      ; Include_Discriminants : Boolean
         := True
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Shape_Array_Access
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind include_discriminants Include_Discriminants
   --# bind origin Origin
   Property_Result : Internal_Shape_Array_Access;
            procedure Finalizer_Scope_461 with Inline_Always;
            procedure Finalizer_Scope_2219 with Inline_Always;
            procedure Finalizer_Scope_2220 with Inline_Always;
            procedure Finalizer_Scope_2221 with Inline_Always;
            procedure Finalizer_Scope_2222 with Inline_Always;
      Block_Var : Internal_Entity_Base_Record_Def;
Block_Var_1 : Internal_Entity_Component_List;
Block_Var_2 : Internal_Entity_Base_Type_Decl;
Block_Var_3 : Internal_Entity_Base_Record_Def;
Block_Var_4 : Internal_Shape_Array_Access;
Block_Var_5 : Internal_Shape_Array_Access;
Block_Var_6 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Var_Expr : Internal_Entity_Derived_Type_Def;
Var_Expr_1 : Internal_Entity_Base_Type_Decl;
Var_Expr_2 : Internal_Entity_Base_Record_Def;
Fld : Internal_Entity_Base_Record_Def;
Fld_1 : Internal_Entity_Component_List;
Fld_2 : Internal_Entity_Type_Def;
Cast_Result : Internal_Entity_Derived_Type_Def;
Fld_3 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Fld_4 : Internal_Entity_Base_Record_Def;
Scope_Result_1 : Internal_Entity_Base_Record_Def;
Result_Var_1 : Internal_Entity_Base_Record_Def;
Fld_5 : Internal_Shape_Array_Access;
Fld_6 : Internal_Shape_Array_Access;
Fld_7 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_8 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Concat_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_9 : Internal_Discriminant_Values_Array_Access;
Fld_10 : Internal_Discriminant_Values_Array_Access;
Concat_Result_1 : Internal_Discriminant_Values_Array_Access;
New_Struct : Internal_Shape := No_Shape;
Map_Result : Internal_Shape_Array_Access;
Map_Result_1 : Internal_Shape_Array_Access;
Scope_Result_2 : Internal_Shape_Array_Access;
Result_Var_2 : Internal_Shape_Array_Access;
Fld_11 : Internal_Entity_Type_Decl;
Fld_12 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_13 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Concat_Result_2 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_14 : Internal_Discriminant_Values_Array_Access;
New_Struct_1 : Internal_Shape := No_Shape;
Map_Result_2 : Internal_Shape_Array_Access;
If_Result : Internal_Shape_Array_Access;
Let_Result : Internal_Shape_Array_Access;
            procedure Finalizer_Scope_461 is
            begin
                     Dec_Ref (Fld_5);
                     Dec_Ref (Block_Var_4);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Result_Var_2);
                     Dec_Ref (Block_Var_5);
                     Dec_Ref (Fld_12);
                     Dec_Ref (Block_Var_6);
                     Dec_Ref (Map_Result_2);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_461;
            procedure Finalizer_Scope_2219 is
            begin
                     Dec_Ref (Fld_6);
                     Dec_Ref (Map_Result_1);
            end Finalizer_Scope_2219;
            procedure Finalizer_Scope_2220 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_2220;
            procedure Finalizer_Scope_2221 is
            begin
                     Dec_Ref (Fld_7);
                     Dec_Ref (Fld_8);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Fld_9);
                     Dec_Ref (Fld_10);
                     Dec_Ref (Concat_Result_1);
                     Dec_Ref (New_Struct);
            end Finalizer_Scope_2221;
            procedure Finalizer_Scope_2222 is
            begin
                     Dec_Ref (Fld_13);
                     Dec_Ref (Concat_Result_2);
                     Dec_Ref (Fld_14);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_2222;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3448 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3423 '<FieldAccess for record_def at ast.py:6247>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6247
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.record_def'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Record_Def (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3423
Block_Var := Fld; 
--# bind rdef Block_Var
--# expr-start 3424 '<FieldAccess for components at ast.py:6248>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6248
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Component_List (Node => Block_Var.Node.Base_Record_Def_F_Components, Info => Block_Var.Info);
--# expr-done 3424
Block_Var_1 := Fld_1; 
--# bind comps Block_Var_1
--# expr-start 3426 '<Cast to DerivedTypeDef at ast.py:6250>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6250
--# expr-start 3425 '<FieldAccess for type_def at ast.py:6250>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6250
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.type_def
Fld_2 := Libadalang.Implementation.Component_List_P_Type_Def (Node => Block_Var_1.Node, E_Info => Block_Var_1.Info);
--# end
--# expr-done 3425
   if Fld_2.Node = null
      or else Fld_2.Node.Kind in Ada_Derived_Type_Def_Range
   then
      Cast_Result := Create_Internal_Entity_Derived_Type_Def
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
   else
         Cast_Result := No_Entity_Derived_Type_Def;
   end if;
--# expr-done 3426
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Derived_Type_Def then
   --# scope-start
--# expr-start 3427 '<FieldAccess for base_type at ast.py:6250>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6250
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld_3 := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 3427
Scope_Result := Fld_3;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Block_Var_2 := Result_Var; 
--# bind parent_type Block_Var_2
Var_Expr_1 := Block_Var_2;
if Var_Expr_1 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 3428 '<FieldAccess for record_def at ast.py:6252>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6252
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.record_def'
Fld_4 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Record_Def (Node => Var_Expr_1.Node, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 3428
Scope_Result_1 := Fld_4;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Base_Record_Def;
end if;
Block_Var_3 := Result_Var_1; 
--# bind parent_record Block_Var_3
--# expr-start 3429 '<FieldAccess for shapes at ast.py:6254>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6254
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.shapes
Fld_5 := Libadalang.Implementation.Component_List_P_Shapes (Node => Block_Var_1.Node, E_Info => Block_Var_1.Info);
--# end
--# expr-done 3429
Block_Var_4 := Fld_5; Inc_Ref (Block_Var_4);
--# bind own_shapes Block_Var_4
Var_Expr_2 := Block_Var_3;
if Var_Expr_2 /= No_Entity_Base_Record_Def then
   --# scope-start
   --# bind _ Var_Expr_2
--# expr-start 3439 '<Mapcat at ast.py:6259>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6259
   --# expr-start 3430 '<FieldAccess for shapes at ast.py:6259>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6259
      if Block_Var_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.shapes
Fld_6 := Libadalang.Implementation.Base_Type_Decl_P_Shapes (Node => Block_Var_2.Node, Include_Discriminants => False, Origin => Origin, E_Info => Block_Var_2.Info);
--# end
--# expr-done 3430
      declare
         Map_Result_1_Vec : Internal_Shape_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Shape_Array_Access := Fld_6;
      begin
         for Item_136 of
               Collection.Items
         loop
   --# scope-start
               --# bind parent_shape Item_136
      --# expr-start 3438 '<Map at ast.py:6260>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6260
      declare
         Map_Result_Vec : Internal_Shape_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Shape_Array_Access := Block_Var_4;
      begin
         for Item_137 of
               Collection.Items
         loop
   --# scope-start
               --# bind own_shape Item_137
      --# expr-start 3437 '<New Shape at ast.py:6261>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6261
--# expr-start 3433 '<Concat at ast.py:6262>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6262
--# expr-start 3431 '<FieldAccess for components at ast.py:6262>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6262
Fld_7 := Item_136.Components;
Inc_Ref (Fld_7);
--# expr-done 3431
--# expr-start 3432 '<FieldAccess for components at ast.py:6263>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6263
Fld_8 := Item_137.Components;
Inc_Ref (Fld_8);
--# expr-done 3432
Concat_Result := Concat (Fld_7, Fld_8); 
--# expr-done 3433
--# expr-start 3436 '<Concat at ast.py:6265>' Concat_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6265
--# expr-start 3434 '<FieldAccess for discriminants_values at ast.py:6265>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6265
Fld_9 := Item_136.Discriminants_Values;
Inc_Ref (Fld_9);
--# expr-done 3434
--# expr-start 3435 '<FieldAccess for discriminants_values at ast.py:6267>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6267
Fld_10 := Item_137.Discriminants_Values;
Inc_Ref (Fld_10);
--# expr-done 3435
Concat_Result_1 := Concat (Fld_9, Fld_10); 
--# expr-done 3436
Inc_Ref (Concat_Result);
Inc_Ref (Concat_Result_1);
New_Struct := (Components => Concat_Result, Discriminants_Values => Concat_Result_1); 
--# expr-done 3437
         declare
            Item_To_Append : constant Internal_Shape := New_Struct;
         begin
               Inc_Ref (Item_To_Append);
            Internal_Shape_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2221;
         end loop;
      end;
         Map_Result := Create_Internal_Shape_Array
           (Items_Count => Natural (Internal_Shape_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Shape_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Shape_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Shape_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 3438
         for Item_To_Append of
               Map_Result.Items
         loop
               Inc_Ref (Item_To_Append);
            Internal_Shape_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_2220;
         end loop;
      end;
         Map_Result_1 := Create_Internal_Shape_Array
           (Items_Count => Natural (Internal_Shape_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Shape_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Shape_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Shape_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 3439
Scope_Result_2 := Map_Result_1;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2219;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := Block_Var_4;
end if;
      Inc_Ref (Result_Var_2);
Block_Var_5 := Result_Var_2; Inc_Ref (Block_Var_5);
--# bind all_shapes Block_Var_5
--# expr-start 3441 '<FieldAccess for discriminants_list at ast.py:6275>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6275
--# expr-start 3440 '<FieldAccess for type_decl at ast.py:6275>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6275
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.type_decl
Fld_11 := Libadalang.Implementation.Component_List_P_Type_Decl (Node => Block_Var_1.Node, E_Info => Block_Var_1.Info);
--# end
--# expr-done 3440
      if Fld_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.discriminants_list'
Fld_12 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Discriminants_List (Node => Fld_11.Node, Origin => Origin, E_Info => Fld_11.Info);
--# end
--# expr-done 3441
Block_Var_6 := Fld_12; Inc_Ref (Block_Var_6);
--# bind discrs Block_Var_6
--# expr-start 3447 '<If at ast.py:6276>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6276
if Include_Discriminants then
   --# expr-start 3446 '<Map at ast.py:6278>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6278
      declare
         Map_Result_2_Vec : Internal_Shape_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Shape_Array_Access := Block_Var_5;
      begin
         for Item_138 of
               Collection.Items
         loop
   --# scope-start
               --# bind s Item_138
      --# expr-start 3445 '<New Shape at ast.py:6279>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6279
--# expr-start 3443 '<Concat at ast.py:6280>' Concat_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6280
--# expr-start 3442 '<FieldAccess for components at ast.py:6280>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6280
Fld_13 := Item_138.Components;
Inc_Ref (Fld_13);
--# expr-done 3442
Concat_Result_2 := Concat (Block_Var_6, Fld_13); 
--# expr-done 3443
--# expr-start 3444 '<FieldAccess for discriminants_values at ast.py:6281>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6281
Fld_14 := Item_138.Discriminants_Values;
Inc_Ref (Fld_14);
--# expr-done 3444
Inc_Ref (Concat_Result_2);
Inc_Ref (Fld_14);
New_Struct_1 := (Components => Concat_Result_2, Discriminants_Values => Fld_14); 
--# expr-done 3445
         declare
            Item_To_Append : constant Internal_Shape := New_Struct_1;
         begin
               Inc_Ref (Item_To_Append);
            Internal_Shape_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2222;
         end loop;
      end;
         Map_Result_2 := Create_Internal_Shape_Array
           (Items_Count => Natural (Internal_Shape_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Internal_Shape_Vectors.Get
              (Map_Result_2_Vec,
               I + Internal_Shape_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Internal_Shape_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 3446
   If_Result := Map_Result_2;
else
   If_Result := Block_Var_5;
end if;
      Inc_Ref (If_Result);
--# expr-done 3447
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 3448
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_461;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_461;
               Finalizer_Scope_2219;
               Finalizer_Scope_2220;
               Finalizer_Scope_2221;
               Finalizer_Scope_2222;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Shapes;
--# end
--# property-start BaseTypeDecl.dottable_subps_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6288
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_F_Dottable_Subps_Env
  (Node : Bare_Base_Type_Decl
  )
   return Lexical_Env
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_462 with Inline_Always;
      Dyn_Env : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_462 is
            begin
                     Dec_Ref (Dyn_Env);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_462;
begin
   --# property-body-start
      if Self.Lf_Present_Dottable_Subps_Env then
         Property_Result := Self.Lf_Stg_Dottable_Subps_Env;
            Inc_Ref (Property_Result);
         return Property_Result;
      end if;
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3450 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3449 '<DynamicLexicalEnv at ast.py:6293>' Dyn_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6293
Dyn_Env := Create_Dynamic_Lexical_Env (Self, Base_Type_Decl_P_Dottable_Subps'Access, null, False); 
--# expr-done 3449
Let_Result := Dyn_Env; Inc_Ref (Let_Result);
--# end
--# expr-done 3450
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_462;
      Self.Lf_Present_Dottable_Subps_Env := True;
      Self.Lf_Stg_Dottable_Subps_Env := Property_Result;
         Inc_Ref (Property_Result);
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_462;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_F_Dottable_Subps_Env;
--# end
--# property-start BaseTypeDecl.dottable_subps /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6299
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Dottable_Subps
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Inner_Env_Assoc_Array_Access
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Inner_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_464 with Inline_Always;
            procedure Finalizer_Scope_2227 with Inline_Always;
      Block_Var : Bare_Declarative_Part;
Block_Var_1 : Internal_Entity_Base_Package_Decl;
Var_Expr : Bare_Package_Body;
Var_Expr_1 : Bare_Declarative_Part;
Var_Expr_2 : Bare_Declarative_Part;
Var_Expr_3 : Internal_Entity_Package_Body;
Var_Expr_4 : Internal_Entity_Declarative_Part;
Bd : Internal_Entity_Basic_Decl;
Var_Expr_5 : Internal_Entity_Base_Subp_Spec;
Var_Expr_6 : Internal_Entity_Basic_Decl;
Fld : Bare_Declarative_Part;
Fld_1 : Bare_Ada_Node;
Scope_Result : Bare_Ada_Node;
Result_Var : Bare_Ada_Node;
Cast_Result : Bare_Package_Body;
Bound_Imprecise_Fallback : Boolean;
As_Entity : Internal_Entity_Package_Body;
Is_Null : Boolean;
If_Result : Internal_Entity_Package_Body;
Fld_2 : Internal_Entity_Basic_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Basic_Decl;
Cast_Result_1 : Internal_Entity_Base_Package_Decl;
Scope_Result_1 : Internal_Entity_Base_Package_Decl;
Fld_3 : Bare_Ada_Node;
Scope_Result_2 : Bare_Ada_Node;
Result_Var_1 : Bare_Ada_Node;
Cast_Result_2 : Bare_Base_Package_Decl;
As_Entity_1 : Internal_Entity_Base_Package_Decl;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Base_Package_Decl;
Result_Var_2 : Internal_Entity_Base_Package_Decl;
Fld_4 : Boolean;
Not_Val : Boolean;
Is_Null_2 : Boolean;
Fld_5 : Internal_Entity_Public_Part;
Cast_Result_3 : Internal_Entity_Declarative_Part;
Fld_6 : Internal_Entity_Private_Part;
Cast_Result_4 : Internal_Entity_Declarative_Part;
Fld_7 : Internal_Entity_Package_Body;
Fld_8 : Internal_Entity_Declarative_Part;
Scope_Result_3 : Internal_Entity_Declarative_Part;
Result_Var_3 : Internal_Entity_Declarative_Part;
Array_Lit : Internal_Entity_Declarative_Part_Array_Access;
Fld_9 : Internal_Entity_Ada_Node_List;
Scope_Result_4 : Internal_Entity_Ada_Node_List;
Result_Var_4 : Internal_Entity_Ada_Node_List;
Fld_10 : Bare_Ada_Node_List;
Fld_11 : Internal_Entity_Info;
As_Entity_2 : Internal_Entity;
Is_Null_3 : Boolean;
If_Result_2 : Internal_Entity;
Item_357 : Internal_Entity;
Bare_Item_357 : Bare_Ada_Node;
Map_Result : Internal_Entity_Array_Access;
Map_Result_1 : Internal_Entity_Array_Access;
Cast_Expr : Internal_Entity;
Cast_Result_5 : Internal_Entity_Basic_Decl;
Fld_12 : Internal_Entity_Defining_Name;
Fld_13 : Symbol_Type;
Fld_14 : Bare_Basic_Decl;
New_Struct : Internal_Metadata;
Cast_Result_6 : Bare_Ada_Node;
New_Struct_1 : Internal_Inner_Env_Assoc;
Let_Result : Internal_Inner_Env_Assoc;
Cast_Expr_1 : Internal_Entity;
Cast_Result_7 : Internal_Entity_Basic_Decl;
Fld_15 : Internal_Entity_Base_Subp_Spec;
Scope_Result_5 : Internal_Entity_Base_Subp_Spec;
Result_Var_5 : Internal_Entity_Base_Subp_Spec;
Fld_16 : Internal_Entity_Base_Type_Decl;
Scope_Result_6 : Internal_Entity_Base_Type_Decl;
Result_Var_6 : Internal_Entity_Base_Type_Decl;
Cast_Result_8 : Internal_Entity;
Cast_Expr_2 : Internal_Entity_Base_Type_Decl;
Cast_Result_9 : Internal_Entity;
Is_Equiv : Boolean;
Map_Result_2 : Internal_Inner_Env_Assoc_Array_Access;
If_Result_3 : Internal_Inner_Env_Assoc_Array_Access;
If_Result_4 : Internal_Inner_Env_Assoc_Array_Access;
Let_Result_1 : Internal_Inner_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_464 is
            begin
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Map_Result_2);
                     Dec_Ref (If_Result_3);
                     Dec_Ref (If_Result_4);
                     Dec_Ref (Let_Result_1);
            end Finalizer_Scope_464;
            procedure Finalizer_Scope_2227 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_2227;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Type_Decl_Base_Type_Decl_P_Dottable_Subps,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Type_Decl,
                                As_Bare_Base_Type_Decl => Self);
               Mmz_K.Items (2) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Inner_Env_Assoc_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 3487 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 3451 '<FieldAccess for declarative_scope at ast.py:6305>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6305
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.declarative_scope
Fld := Libadalang.Implementation.Ada_Node_P_Declarative_Scope (Node => Ent.Node);
--# end
--# expr-done 3451
Block_Var := Fld; 
--# bind scope Block_Var
--# expr-start 3453 '<Cast to PackageBody at ast.py:6307>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6307
Var_Expr_1 := Block_Var;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 3452 '<parent at ast.py:6307>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6307
      if Var_Expr_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Var_Expr_1.Parent; 
--# expr-done 3452
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node;
end if;
   if Result_Var = null
      or else Result_Var.Kind in Ada_Package_Body_Range
   then
      Cast_Result := Result_Var;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 3453
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind body Var_Expr
--# expr-start 3457 '<Cast to BasePackageDecl at ast.py:6308>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6308
--# expr-start 3456 '<bind at ast.py:6308>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6308
Bound_Imprecise_Fallback := False; 
--# expr-start 3455 '<FieldAccess for previous_part at ast.py:6310>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6310
--# expr-start 3454 '<as_entity at ast.py:6310>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6310
Is_Null := Var_Expr = null; 
if Is_Null then
   If_Result := No_Entity_Package_Body;
else
As_Entity := (Info => E_Info, Node => Var_Expr); 
   If_Result := As_Entity;
end if;
--# expr-done 3454
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.previous_part
Fld_2 := Libadalang.Implementation.Body_Node_P_Previous_Part (Node => If_Result.Node, Imprecise_Fallback => Bound_Imprecise_Fallback, E_Info => If_Result.Info);
--# end
--# expr-done 3455
Dyn_Var_Bind_Result := Fld_2; 
--# expr-done 3456
   if Dyn_Var_Bind_Result.Node = null
      or else Dyn_Var_Bind_Result.Node.Kind in Ada_Base_Package_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Base_Package_Decl
        (Node => Dyn_Var_Bind_Result.Node,
         Info => Dyn_Var_Bind_Result.Info);
   else
         Cast_Result_1 := No_Entity_Base_Package_Decl;
   end if;
--# expr-done 3457
Scope_Result_1 := Cast_Result_1;
   --# end
   Result_Var_2 := Scope_Result_1;
else
   --# expr-start 3460 '<as_entity at ast.py:6312>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6312
--# expr-start 3459 '<Cast to BasePackageDecl at ast.py:6312>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6312
Var_Expr_2 := Block_Var;
if Var_Expr_2 /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 3458 '<parent at ast.py:6312>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6312
      if Var_Expr_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Var_Expr_2.Parent; 
--# expr-done 3458
Scope_Result_2 := Fld_3;
   --# end
   Result_Var_1 := Scope_Result_2;
else
   Result_Var_1 := No_Bare_Ada_Node;
end if;
   if Result_Var_1 = null
      or else Result_Var_1.Kind in Ada_Base_Package_Decl
   then
      Cast_Result_2 := Result_Var_1;
   else
         Cast_Result_2 := No_Bare_Ada_Node;
   end if;
--# expr-done 3459
Is_Null_1 := Cast_Result_2 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Base_Package_Decl;
else
As_Entity_1 := (Info => E_Info, Node => Cast_Result_2); 
   If_Result_1 := As_Entity_1;
end if;
--# expr-done 3460
   Result_Var_2 := If_Result_1;
end if;
Block_Var_1 := Result_Var_2; 
--# bind pkg Block_Var_1
--# expr-start 3486 '<Cond at ast.py:6315>' If_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6315
--# expr-start 3462 '<Not at ast.py:6316>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6316
--# expr-start 3461 '<FieldAccess for is_tagged_type at ast.py:6316>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6316
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_tagged_type'
Fld_4 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Tagged_Type (Node => Ent.Node, Origin => No_Bare_Ada_Node, E_Info => Ent.Info);
--# end
--# expr-done 3461
Not_Val := not (Fld_4); 
--# expr-done 3462
if Not_Val then
   If_Result_4 := No_Internal_Inner_Env_Assoc_Array_Type;
else
--# expr-start 3463 '<IsNull at ast.py:6319>' Is_Null_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6319
Is_Null_2 := Block_Var_1.Node = null; 
--# expr-done 3463
if Is_Null_2 then
   If_Result_3 := No_Internal_Inner_Env_Assoc_Array_Type;
else
   --# expr-start 3485 '<FilterMap at ast.py:6322>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6322
   --# expr-start 3473 '<Mapcat at ast.py:6322>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6322
   --# expr-start 3470 '<ArrayLiteral at ast.py:6322>' Array_Lit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6322
--# expr-start 3465 '<Cast to DeclarativePart at ast.py:6323>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6323
--# expr-start 3464 '<FieldAccess for public_part at ast.py:6323>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6323
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Public_Part (Node => Block_Var_1.Node.Base_Package_Decl_F_Public_Part, Info => Block_Var_1.Info);
--# expr-done 3464
      Cast_Result_3 := Create_Internal_Entity_Declarative_Part
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
--# expr-done 3465
--# expr-start 3467 '<Cast to DeclarativePart at ast.py:6324>' Cast_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6324
--# expr-start 3466 '<FieldAccess for private_part at ast.py:6324>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6324
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Private_Part (Node => Block_Var_1.Node.Base_Package_Decl_F_Private_Part, Info => Block_Var_1.Info);
--# expr-done 3466
      Cast_Result_4 := Create_Internal_Entity_Declarative_Part
        (Node => Fld_6.Node,
         Info => Fld_6.Info);
--# expr-done 3467
--# expr-start 3468 '<FieldAccess for body_part at ast.py:6325>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6325
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasePackageDecl.body_part
Fld_7 := Libadalang.Implementation.Base_Package_Decl_P_Body_Part (Node => Block_Var_1.Node, E_Info => Block_Var_1.Info);
--# end
--# expr-done 3468
Var_Expr_3 := Fld_7;
if Var_Expr_3 /= No_Entity_Package_Body then
   --# scope-start
--# expr-start 3469 '<FieldAccess for decls at ast.py:6325>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6325
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Declarative_Part (Node => Var_Expr_3.Node.Package_Body_F_Decls, Info => Var_Expr_3.Info);
--# expr-done 3469
Scope_Result_3 := Fld_8;
   --# end
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := No_Entity_Declarative_Part;
end if;
Array_Lit := Create_Internal_Entity_Declarative_Part_Array (Internal_Internal_Entity_Declarative_Part_Array'(1 => Cast_Result_3, 2 => Cast_Result_4, 3 => Result_Var_3)); 
--# expr-done 3470
      declare
         Map_Result_1_Vec : Internal_Entity_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Declarative_Part_Array_Access := Array_Lit;
      begin
         for Item_140 of
               Collection.Items
         loop
   --# scope-start
               --# bind dp Item_140
      --# expr-start 3472 '<TakeWhile at ???>' Map_Result None
Var_Expr_4 := Item_140;
if Var_Expr_4 /= No_Entity_Declarative_Part then
   --# scope-start
--# expr-start 3471 '<FieldAccess for decls at ast.py:6327>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6327
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Ada_Node_List (Node => Var_Expr_4.Node.Declarative_Part_F_Decls, Info => Var_Expr_4.Info);
--# expr-done 3471
Scope_Result_4 := Fld_9;
   --# end
   Result_Var_4 := Scope_Result_4;
else
   Result_Var_4 := No_Entity_Ada_Node_List;
end if;
Fld_10 := Result_Var_4.Node;
      declare
         Map_Result_Vec : Internal_Entity_Vectors.Vector;
      begin
      if Fld_10 = null then
      Map_Result := Create_Internal_Entity_Array (0);
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_10;
      begin
         for Untyped_Item_357 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_357 := Untyped_Item_357; 
Is_Null_3 := Bare_Item_357 = null; 
if Is_Null_3 then
   If_Result_2 := No_Entity;
else
Fld_11 := Result_Var_4.Info;
As_Entity_2 := (Info => Fld_11, Node => Bare_Item_357); 
   If_Result_2 := As_Entity_2;
end if;
                  Item_357 := If_Result_2; 
   --# scope-start
         declare
            Item_To_Append : constant Internal_Entity := Item_357;
         begin
            Internal_Entity_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 3472
         for Item_To_Append of
               Map_Result.Items
         loop
            Internal_Entity_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_2227;
         end loop;
      end;
         Map_Result_1 := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 3473
      declare
         Map_Result_2_Vec : Internal_Inner_Env_Assoc_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Map_Result_1;
      begin
         for Item_139 of
               Collection.Items
         loop
   --# scope-start
               --# bind decl Item_139
         --# expr-start 3477 '<Eq at ast.py:6338>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6338
--# expr-start 3474 '<Cast to BasicDecl at ast.py:6338>' Cast_Result_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6338
Cast_Expr_1 := Item_139; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_7 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_7 := No_Entity_Basic_Decl;
   end if;
--# expr-done 3474
Var_Expr_6 := Cast_Result_7;
if Var_Expr_6 /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 3475 '<FieldAccess for subp_spec_or_null at ast.py:6338>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6338
      if Var_Expr_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_15 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Var_Expr_6.Node, E_Info => Var_Expr_6.Info);
--# end
--# expr-done 3475
Scope_Result_5 := Fld_15;
   --# end
   Result_Var_5 := Scope_Result_5;
else
   Result_Var_5 := No_Entity_Base_Subp_Spec;
end if;
Var_Expr_5 := Result_Var_5;
if Var_Expr_5 /= No_Entity_Base_Subp_Spec then
   --# scope-start
--# expr-start 3476 '<FieldAccess for dottable_subp_of at ast.py:6338>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6338
      if Var_Expr_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.dottable_subp_of
Fld_16 := Libadalang.Implementation.Base_Subp_Spec_P_Dottable_Subp_Of (Node => Var_Expr_5.Node, E_Info => Var_Expr_5.Info);
--# end
--# expr-done 3476
Scope_Result_6 := Fld_16;
   --# end
   Result_Var_6 := Scope_Result_6;
else
   Result_Var_6 := No_Entity_Base_Type_Decl;
end if;
      Cast_Result_8 := Create_Internal_Entity
        (Node => Result_Var_6.Node,
         Info => Result_Var_6.Info);
Cast_Expr_2 := Ent; 
      Cast_Result_9 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
Is_Equiv := Equivalent (Cast_Result_8, Cast_Result_9); 
--# expr-done 3477
         if Is_Equiv then
      --# expr-start 3484 '<Let at ast.py:6329>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6329
--# scope-start
--# expr-start 3478 '<Cast to BasicDecl at ast.py:6330>' Cast_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6330
Cast_Expr := Item_139; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_5 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result_5 := No_Entity_Basic_Decl;
   end if;
--# expr-done 3478
Bd := Cast_Result_5; 
--# bind bd Bd
--# expr-start 3483 '<New InnerEnvAssoc at ast.py:6330>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6330
--# expr-start 3480 '<FieldAccess for name_symbol at ast.py:6331>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6331
--# expr-start 3479 '<FieldAccess for defining_name at ast.py:6331>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6331
      if Bd.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_12 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Bd.Node, E_Info => Bd.Info);
--# end
--# expr-done 3479
      if Fld_12.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_13 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_12.Node);
--# end
--# expr-done 3480
--# expr-start 3481 '<New Metadata at ast.py:6333>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6333
New_Struct := (Dottable_Subp => True, Primitive => No_Bare_Ada_Node, Primitive_Real_Type => No_Bare_Ada_Node); 
--# expr-done 3481
--# expr-start 3482 '<FieldAccess for node at ast.py:6332>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6332
Fld_14 := Bd.Node;
--# expr-done 3482
      Cast_Result_6 := Fld_14;
New_Struct_1 := (Key => Fld_13, Metadata => New_Struct, Value => Cast_Result_6); 
--# expr-done 3483
Let_Result := New_Struct_1; 
--# end
--# expr-done 3484
         declare
            Item_To_Append : constant Internal_Inner_Env_Assoc := Let_Result;
         begin
            Internal_Inner_Env_Assoc_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result_2 := Create_Internal_Inner_Env_Assoc_Array
           (Items_Count => Natural (Internal_Inner_Env_Assoc_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Internal_Inner_Env_Assoc_Vectors.Get
              (Map_Result_2_Vec,
               I + Internal_Inner_Env_Assoc_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Internal_Inner_Env_Assoc_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 3485
   If_Result_3 := Map_Result_2;
end if;
      Inc_Ref (If_Result_3);
   If_Result_4 := If_Result_3;
end if;
      Inc_Ref (If_Result_4);
--# expr-done 3486
Let_Result_1 := If_Result_4; Inc_Ref (Let_Result_1);
--# end
--# expr-done 3487
      Property_Result := Let_Result_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_464;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Inner_Env_Assoc_Array_Access,
                        As_Internal_Inner_Env_Assoc_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_464;
               Finalizer_Scope_2227;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Dottable_Subps;
--# end
--# property-start BaseTypeDecl.find_base_type_rebindings /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6345
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Find_Base_Type_Rebindings
  (Node : Bare_Base_Type_Decl
      ; Target : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Env_Rebindings
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind target Target
   --# bind origin Origin
   Property_Result : Env_Rebindings;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Info;
Fld_1 : Env_Rebindings;
Fld_2 : Internal_Entity_Base_Type_Decl;
Fld_3 : Env_Rebindings;
Scope_Result : Env_Rebindings;
Result_Var : Env_Rebindings;
Cast_Expr : Bare_Base_Type_Decl;
Cast_Result : Bare_Base_Type_Decl;
Is_Equal : Boolean;
If_Result : Env_Rebindings;
Let_Result : Env_Rebindings;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Type_Decl_Base_Type_Decl_P_Find_Base_Type_Rebindings,
            Items    => new Mmz_Key_Array (1 ..  4))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Type_Decl,
                                As_Bare_Base_Type_Decl => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Bare_Base_Type_Decl,
                                      As_Bare_Base_Type_Decl => Target);
               Mmz_K.Items (3) := (Kind => Mmz_Bare_Ada_Node,
                                      As_Bare_Ada_Node => Origin);
               Mmz_K.Items (4) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Env_Rebindings;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 3494 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3493 '<If at ast.py:6362>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6362
--# expr-start 3488 '<Eq at ast.py:6363>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6363
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Is_Equal := Cast_Result = Target; 
--# expr-done 3488
if Is_Equal then
   --# expr-start 3490 '<FieldAccess for rebindings at ast.py:6364>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6364
--# expr-start 3489 '<FieldAccess for info at ast.py:6364>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6364
Fld := Ent.Info;
--# expr-done 3489
Fld_1 := Fld.Rebindings;
--# expr-done 3490
   If_Result := Fld_1;
else
--# expr-start 3491 '<FieldAccess for base_type at ast.py:6365>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6365
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.base_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Base_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3491
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind bt Var_Expr
--# expr-start 3492 '<FieldAccess for find_base_type_rebindings at ast.py:6366>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6366
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.find_base_type_rebindings
Fld_3 := Libadalang.Implementation.Base_Type_Decl_P_Find_Base_Type_Rebindings (Node => Var_Expr.Node, Target => Target, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 3492
Scope_Result := Fld_3;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := null;
end if;
   If_Result := Result_Var;
end if;
--# expr-done 3493
Let_Result := If_Result; 
--# end
--# expr-done 3494
      Property_Result := Let_Result;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Env_Rebindings,
                        As_Env_Rebindings => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Find_Base_Type_Rebindings;
--# end
--# property-start BaseTypeDecl.has_base_type_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6371
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Has_Base_Type_Impl
  (Node : Bare_Base_Type_Decl
      ; Target : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind target Target
   Property_Result : Boolean;
            procedure Finalizer_Scope_466 with Inline_Always;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Bare_Base_Type_Decl;
Cast_Result : Bare_Base_Type_Decl;
Is_Equal : Boolean;
Fld_2 : Internal_Entity_Base_Type_Decl_Array_Access;
Fld_3 : Boolean;
Quantifier_Result : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_466 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_466;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3501 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3497 '<Eq at ast.py:6377>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6377
--# expr-start 3496 '<FieldAccess for node at ast.py:6377>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6377
--# expr-start 3495 '<FieldAccess for canonical_type at ast.py:6377>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6377
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.canonical_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Canonical_Type (Node => Ent.Node, Origin => No_Bare_Ada_Node, E_Info => Ent.Info);
--# end
--# expr-done 3495
Fld_1 := Fld.Node;
--# expr-done 3496
      Cast_Result := Fld_1;
Is_Equal := Cast_Result = Target; 
--# expr-done 3497
if Is_Equal then
   If_Result := True;
else
   --# expr-start 3500 '<AnyQuantifier at ast.py:6378>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6378
   --# expr-start 3498 '<FieldAccess for base_types at ast.py:6378>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6378
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.base_types'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Base_Types (Node => Ent.Node, Origin => No_Bare_Ada_Node, E_Info => Ent.Info);
--# end
--# expr-done 3498
      Quantifier_Result := False;
      declare
         Collection : constant Internal_Entity_Base_Type_Decl_Array_Access := Fld_2;
      begin
         for Item_141 of
               Collection.Items
         loop
   --# scope-start
               --# bind bt Item_141
      --# expr-start 3499 '<FieldAccess for has_base_type_impl at ast.py:6379>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6379
      if Item_141.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.has_base_type_impl
Fld_3 := Libadalang.Implementation.Base_Type_Decl_P_Has_Base_Type_Impl (Node => Item_141.Node, Target => Target, E_Info => Item_141.Info);
--# end
--# expr-done 3499
      Quantifier_Result := Fld_3;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
--# expr-done 3500
   If_Result := Quantifier_Result;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 3501
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_466;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_466;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Has_Base_Type_Impl;
--# end
--# property-start BaseTypeDecl.has_base_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6384
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Has_Base_Type
  (Node : Bare_Base_Type_Decl
      ; Target : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind target Target
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
As_Entity : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Bare_Base_Type_Decl;
Fld_3 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3507 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3506 '<FieldAccess for has_base_type_impl at ast.py:6393>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6393
--# expr-start 3502 '<FieldAccess for full_view at ast.py:6393>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6393
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.full_view
Fld := Libadalang.Implementation.Base_Type_Decl_P_Full_View (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3502
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3505 '<FieldAccess for node at ast.py:6394>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6394
--# expr-start 3504 '<FieldAccess for canonical_type at ast.py:6394>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6394
--# expr-start 3503 '<as_bare_entity at ast.py:6394>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6394
Is_Null := Target = null; 
if Is_Null then
   If_Result := No_Entity_Base_Type_Decl;
else
As_Entity := (Info => No_Entity_Info, Node => Target); 
   If_Result := As_Entity;
end if;
--# expr-done 3503
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.canonical_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Canonical_Type (Node => If_Result.Node, Origin => No_Bare_Ada_Node, E_Info => If_Result.Info);
--# end
--# expr-done 3504
Fld_2 := Fld_1.Node;
--# expr-done 3505
--# property-call-start BaseTypeDecl.has_base_type_impl
Fld_3 := Libadalang.Implementation.Base_Type_Decl_P_Has_Base_Type_Impl (Node => Fld.Node, Target => Fld_2, E_Info => Fld.Info);
--# end
--# expr-done 3506
Let_Result := Fld_3; 
--# end
--# expr-done 3507
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Has_Base_Type;
--# end
--# property-start BaseTypeDecl._env_mappings_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5083
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_2
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1418 with Inline_Always;
      Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Base_Type_Decl;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1418 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1418;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3510 '<New EnvAssoc at ast.py:5083>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5083
--# expr-start 3508 '<New DesignatedEnv at ast.py:5083>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5083
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 3508
--# expr-start 3509 '<FieldAccess for name_symbol at ast.py:5083>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5083
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 3509
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 3510
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1418;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1418;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_2;
--# end
--# property-start BaseTypeDecl.primitives_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5187
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Primitives_Env
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_374 with Inline_Always;
      Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_374 is
            begin
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_374;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 3511 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := Empty_Env; Inc_Ref (Let_Result);
--# end
--# expr-done 3511
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_374;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_374;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Primitives_Env;
--# end
--# property-start BaseTypeDecl.is_record_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5191
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Record_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Record_Def;
Is_Null : Boolean;
Not_Val : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3515 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3514 '<Not at ast.py:5198>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5198
--# expr-start 3513 '<IsNull at ast.py:5198>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5198
--# expr-start 3512 '<FieldAccess for record_def at ast.py:5198>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5198
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.record_def'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Record_Def (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3512
Is_Null := Fld.Node = null; 
--# expr-done 3513
Not_Val := not (Is_Null); 
--# expr-done 3514
Let_Result := Not_Val; 
--# end
--# expr-done 3515
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Record_Type;
--# end
--# property-start BaseTypeDecl.is_task_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5223
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Task_Type
  (Node : Bare_Base_Type_Decl
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Task_Type;
--# end
--# property-start BaseTypeDecl.is_real_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5225
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Real_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Real_Type;
--# end
--# property-start BaseTypeDecl.is_float_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5229
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Float_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Float_Type;
--# end
--# property-start BaseTypeDecl.is_fixed_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5233
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Fixed_Point
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Fixed_Point;
--# end
--# property-start BaseTypeDecl.is_enum_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5238
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Enum_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Enum_Type;
--# end
--# property-start BaseTypeDecl.is_classwide /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5243
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Classwide
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Classwide;
--# end
--# property-start BaseTypeDecl.is_access_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5245
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Access_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Access_Type;
--# end
--# property-start BaseTypeDecl.has_ud_indexing /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5257
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Has_Ud_Indexing
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Has_Ud_Indexing;
--# end
--# property-start BaseTypeDecl.constant_indexing_fns /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5261
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Constant_Indexing_Fns
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl_Array_Access
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Basic_Decl_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := No_Internal_Entity_Basic_Decl_Array_Type;
         Inc_Ref (Property_Result);
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Constant_Indexing_Fns;
--# end
--# property-start BaseTypeDecl.variable_indexing_fns /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5268
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Variable_Indexing_Fns
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl_Array_Access
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Basic_Decl_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := No_Internal_Entity_Basic_Decl_Array_Type;
         Inc_Ref (Property_Result);
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Variable_Indexing_Fns;
--# end
--# property-start BaseTypeDecl.get_imp_deref /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5276
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Get_Imp_Deref
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Expr
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := No_Entity_Expr;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Get_Imp_Deref;
--# end
--# property-start BaseTypeDecl.access_def /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5391
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Access_Def
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Access_Def
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Access_Def;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := No_Entity_Access_Def;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Access_Def;
--# end
--# property-start BaseTypeDecl.is_char_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5393
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Char_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Char_Type;
--# end
--# property-start BaseTypeDecl.classwide_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5406
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Classwide_Type
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Classwide_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Classwide_Type_Decl;
      Cast_Expr : Bare_Base_Type_Decl;
Cast_Result : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld : Bare_Classwide_Type_Decl;
As_Entity : Internal_Entity_Classwide_Type_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Classwide_Type_Decl;
Fld_1 : Boolean;
If_Result_1 : Internal_Entity_Classwide_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Classwide_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3520 '<bind at ast.py:5406>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5406
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Bound_Origin := Cast_Result; 
--# expr-start 3519 '<If at ast.py:5406>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5406
--# expr-start 3516 '<FieldAccess for is_tagged_type at ast.py:5407>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5407
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_tagged_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Tagged_Type (Node => Ent.Node, Origin => Bound_Origin, E_Info => Ent.Info);
--# end
--# expr-done 3516
if Fld_1 then
   --# expr-start 3518 '<as_entity at ast.py:5408>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5408
--# expr-start 3517 '<FieldAccess for classwide_type_node at ast.py:5408>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5408
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.classwide_type_node
Fld := Libadalang.Implementation.Base_Type_Decl_P_Classwide_Type_Node (Node => Self);
--# end
--# expr-done 3517
Is_Null := Fld = null; 
if Is_Null then
   If_Result := No_Entity_Classwide_Type_Decl;
else
As_Entity := (Info => E_Info, Node => Fld); 
   If_Result := As_Entity;
end if;
--# expr-done 3518
   If_Result_1 := If_Result;
else
   If_Result_1 := No_Entity_Classwide_Type_Decl;
end if;
--# expr-done 3519
Dyn_Var_Bind_Result := If_Result_1; 
--# expr-done 3520
      Property_Result := Dyn_Var_Bind_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Classwide_Type;
--# end
--# property-start BaseTypeDecl.discrete_range /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5453
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Discrete_Range
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Discrete_Range
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Discrete_Range;
      Let_Result : Internal_Discrete_Range;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 3521 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := No_Discrete_Range; 
--# end
--# expr-done 3521
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Discrete_Range;
--# end
--# property-start BaseTypeDecl.is_discrete_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5522
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Discrete_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Boolean;
Fld_1 : Boolean;
If_Result : Boolean;
Fld_2 : Boolean;
If_Result_1 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3525 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3522 '<FieldAccess for is_int_type at ast.py:5526>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5526
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_int_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Int_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3522
if Fld then
   If_Result := True;
else
   --# expr-start 3523 '<FieldAccess for is_enum_type at ast.py:5526>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5526
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_enum_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Enum_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3523
   If_Result := Fld_1;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 3524 '<FieldAccess for is_char_type at ast.py:5526>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5526
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_char_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Char_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3524
   If_Result_1 := Fld_2;
end if;
Let_Result := If_Result_1; 
--# end
--# expr-done 3525
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Discrete_Type;
--# end
--# property-start BaseTypeDecl.is_int_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5533
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Int_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
      Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 3526 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := False; 
--# end
--# expr-done 3526
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Int_Type;
--# end
--# property-start BaseTypeDecl.accessed_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5595
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Accessed_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 3527 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := No_Entity_Base_Type_Decl; 
--# end
--# expr-done 3527
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Accessed_Type;
--# end
--# property-start BaseTypeDecl.is_tagged_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5659
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Tagged_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Tagged_Type;
--# end
--# property-start BaseTypeDecl.base_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5676
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Base_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := No_Entity_Base_Type_Decl;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Base_Type;
--# end
--# property-start BaseTypeDecl.base_types /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5682
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Base_Types
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl_Array_Access
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_425 with Inline_Always;
            procedure Finalizer_Scope_2042 with Inline_Always;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Singleton : Internal_Entity_Base_Type_Decl_Array_Access;
Scope_Result : Internal_Entity_Base_Type_Decl_Array_Access;
Result_Var : Internal_Entity_Base_Type_Decl_Array_Access;
Fld_1 : Internal_Entity_Base_Type_Decl_Array_Access;
Concat_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_425 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Concat_Result);
            end Finalizer_Scope_425;
            procedure Finalizer_Scope_2042 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2042;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3530 '<Concat at ast.py:5683>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5683
--# expr-start 3528 '<FieldAccess for base_type at ast.py:5683>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5683
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.base_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Base_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3528
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind bt Var_Expr
                Singleton := Create_Internal_Entity_Base_Type_Decl_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2042;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Base_Type_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var);
--# expr-start 3529 '<FieldAccess for base_interfaces at ast.py:5684>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5684
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.base_interfaces'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Base_Interfaces (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3529
Concat_Result := Concat (Result_Var, Fld_1); 
--# expr-done 3530
      Property_Result := Concat_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_425;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_425;
               Finalizer_Scope_2042;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Base_Types;
--# end
--# property-start BaseTypeDecl.base_interfaces /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5690
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Base_Interfaces
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl_Array_Access
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := No_Internal_Entity_Base_Type_Decl_Array_Type;
         Inc_Ref (Property_Result);
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Base_Interfaces;
--# end
--# property-start BaseTypeDecl.record_def /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5709
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Record_Def
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Record_Def
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Record_Def;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := No_Entity_Base_Record_Def;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Record_Def;
--# end
--# property-start BaseTypeDecl.array_def /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5710
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Array_Def
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Type_Def
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Array_Type_Def;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := No_Entity_Array_Type_Def;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Array_Def;
--# end
--# property-start BaseTypeDecl.is_iterable_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5794
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Iterable_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Iterable_Type;
--# end
--# property-start BaseTypeDecl.is_interface_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5804
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Interface_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Match_0 : Internal_Entity_Type_Decl;
Match_1 : Internal_Entity_Subtype_Decl;
Match_2 : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Type_Def;
Is_A : Boolean;
Fld_2 : Internal_Entity_Base_Type_Decl;
Fld_3 : Boolean;
Match_Prefix : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity_Type_Decl;
Let_Result : Boolean;
Scope_Result : Boolean;
Cast_Result_1 : Internal_Entity_Subtype_Decl;
Let_Result_1 : Boolean;
Scope_Result_1 : Boolean;
Cast_Result_2 : Internal_Entity_Base_Type_Decl;
Let_Result_2 : Boolean;
Scope_Result_2 : Boolean;
Match_Result : Boolean;
Scope_Result_3 : Boolean;
Result_Var : Boolean;
Let_Result_3 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3537 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 3531 '<FieldAccess for full_view at ast.py:5808>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5808
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.full_view
Fld := Libadalang.Implementation.Base_Type_Decl_P_Full_View (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3531
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 3536 '<Match at ast.py:5808>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5808
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Var_Expr; 
case Ada_Base_Type_Decl (Match_Prefix.Node.Kind) is
         when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind td Match_0
--# expr-start 3533 '<IsA InterfaceTypeDef at ast.py:5809>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5809
--# expr-start 3532 '<FieldAccess for type_def at ast.py:5809>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5809
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Type_Def (Node => Match_0.Node.Type_Decl_F_Type_Def, Info => Match_0.Info);
--# expr-done 3532
Is_A := Fld_1.Node /= null 
and then Fld_1.Node.Kind in Ada_Interface_Type_Def_Range; 
--# expr-done 3533
Let_Result := Is_A; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Subtype_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Subtype_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind sb Match_1
--# expr-start 3535 '<FieldAccess for is_interface_type at ast.py:5810>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5810
--# expr-start 3534 '<FieldAccess for get_type at ast.py:5810>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5810
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Match_1.Node, Origin => Origin, E_Info => Match_1.Info);
--# end
--# expr-done 3534
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_interface_type'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Interface_Type (Node => Fld_2.Node, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 3535
Let_Result_1 := Fld_3; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Discrete_Base_Subtype_Decl | Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Base_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind _ Match_2
Let_Result_2 := False; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 3536
Scope_Result_3 := Match_Result;
   --# end
   Result_Var := Scope_Result_3;
else
   Result_Var := False;
end if;
Let_Result_3 := Result_Var; 
--# end
--# expr-done 3537
      Property_Result := Let_Result_3;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Interface_Type;
--# end
--# property-start BaseTypeDecl.iterable_comp_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5815
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Iterable_Comp_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 3538 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := No_Entity_Base_Type_Decl; 
--# end
--# expr-done 3538
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Iterable_Comp_Type;
--# end
--# property-start BaseTypeDecl.canonical_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5968
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Canonical_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Bound_Imprecise_Fallback : Boolean;
Fld : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3542 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3541 '<bind at ast.py:5973>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5973
Bound_Imprecise_Fallback := False; 
--# expr-start 3540 '<Cast to BaseTypeDecl at ast.py:5974>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5974
--# expr-start 3539 '<FieldAccess for canonical_part at ast.py:5974>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5974
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.canonical_part
Fld := Libadalang.Implementation.Basic_Decl_P_Canonical_Part (Node => Ent.Node, Imprecise_Fallback => Bound_Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 3539
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 3540
Dyn_Var_Bind_Result := Cast_Result; 
--# expr-done 3541
Let_Result := Dyn_Var_Bind_Result; 
--# end
--# expr-done 3542
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Canonical_Type;
--# end
--# property-start BaseTypeDecl.previous_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6003
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Previous_Part
  (Node : Bare_Base_Type_Decl
      ; Go_To_Incomplete : Boolean
         := True
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind go_to_incomplete Go_To_Incomplete
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_2189 with Inline_Always;
            procedure Finalizer_Scope_2190 with Inline_Always;
      Var_Expr : Bare_Defining_Name;
Var_Expr_1 : Internal_Entity_Array_Access;
Var_Expr_2 : Internal_Entity_Base_Type_Decl;
Fld : Bare_Defining_Name;
Fld_1 : Internal_Entity;
Fld_2 : Lexical_Env := Empty_Env;
Fld_3 : Symbol_Type;
Cast_Expr : Bare_Base_Type_Decl;
Cast_Result : Bare_Ada_Node;
Env_Get_Result : Internal_Entity_Array_Access;
Fld_4 : Boolean;
Cast_Expr_1 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Base_Type_Decl;
Fld_5 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result : Boolean;
Is_A : Boolean;
If_Result_1 : Boolean;
If_Result_2 : Boolean;
Find_Result : Internal_Entity;
Scope_Result_1 : Internal_Entity;
Result_Var_1 : Internal_Entity;
Cast_Result_2 : Internal_Entity_Base_Type_Decl;
Scope_Result_2 : Internal_Entity_Base_Type_Decl;
Result_Var_2 : Internal_Entity_Base_Type_Decl;
Fld_6 : Boolean;
If_Result_3 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_2189 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Env_Get_Result);
            end Finalizer_Scope_2189;
            procedure Finalizer_Scope_2190 is
            begin
                     Dec_Ref (Var_Expr_1);
            end Finalizer_Scope_2190;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3556 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3555 '<If at ast.py:6007>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6007
--# expr-start 3543 '<FieldAccess for is_generic_formal at ast.py:6008>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6008
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_generic_formal
Fld_6 := Libadalang.Implementation.Base_Type_Decl_P_Is_Generic_Formal (Node => Self);
--# end
--# expr-done 3543
if Fld_6 then
   If_Result_3 := No_Entity_Base_Type_Decl;
else
--# expr-start 3544 '<FieldAccess for name at ast.py:6015>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6015
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Base_Type_Decl_F_Name;
--# expr-done 3544
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind type_name Var_Expr
--# expr-start 3554 '<Cast to BaseTypeDecl at ast.py:6018>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6018
--# expr-start 3548 '<EnvGet at ast.py:6018>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6018
--# expr-start 3546 '<FieldAccess for immediate_declarative_region at ast.py:6018>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6018
--# expr-start 3545 '<FieldAccess for semantic_parent at ast.py:6018>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6018
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.semantic_parent
Fld_1 := Libadalang.Implementation.Ada_Node_P_Semantic_Parent (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3545
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.immediate_declarative_region'
Fld_2 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Immediate_Declarative_Region (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 3546
--# expr-start 3547 '<FieldAccess for name_symbol at ast.py:6019>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6019
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Var_Expr);
--# end
--# expr-done 3547
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Fld_2, Key => Fld_3, Lookup_Kind => To_Lookup_Kind_Type (Minimal), Categories => (Inherited_Primitives => True, Nocat => True), From => Cast_Result)); 
--# expr-done 3548
Var_Expr_1 := Env_Get_Result;
      Inc_Ref (Var_Expr_1);
if Var_Expr_1 /= No_Internal_Entity_Array_Type then
   --# scope-start
   --# bind pp Var_Expr_1
--# expr-start 3553 '<Find at ast.py:6022>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6022
      Find_Result := No_Entity;
      declare
         Collection : constant Internal_Entity_Array_Access := Var_Expr_1;
      begin
         for Item_134 of
               Collection.Items
         loop
   --# scope-start
               --# bind pp Item_134
--# expr-start 3549 '<FieldAccess for is_in_private_part at ast.py:6023>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6023
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_in_private_part'
Fld_4 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_In_Private_Part (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3549
if Fld_4 then
--# expr-start 3550 '<Cast to BaseTypeDecl at ast.py:6024>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6024
Cast_Expr_1 := Item_134; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Base_Type_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_1 := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 3550
Var_Expr_2 := Cast_Result_1;
if Var_Expr_2 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 3551 '<FieldAccess for is_private at ast.py:6024>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6024
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_private'
Fld_5 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Private (Node => Var_Expr_2.Node, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 3551
Scope_Result := Fld_5;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
   If_Result := Result_Var;
else
   If_Result := False;
end if;
if If_Result then
   If_Result_2 := True;
else
if Go_To_Incomplete then
   --# expr-start 3552 '<IsA IncompleteTypeDecl at ast.py:6026>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6026
Is_A := Item_134.Node /= null 
and then Item_134.Node.Kind in Ada_Incomplete_Type_Decl_Range; 
--# expr-done 3552
   If_Result_1 := Is_A;
else
   If_Result_1 := False;
end if;
   If_Result_2 := If_Result_1;
end if;
      if If_Result_2 then
         Find_Result := Item_134;
      end if;
   --# end
               exit when If_Result_2;
         end loop;
      end;
--# expr-done 3553
Scope_Result_1 := Find_Result;
   --# end
      Finalizer_Scope_2190;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity;
end if;
   if Result_Var_1.Node = null
      or else Result_Var_1.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_2 := Create_Internal_Entity_Base_Type_Decl
        (Node => Result_Var_1.Node,
         Info => Result_Var_1.Info);
   else
         Cast_Result_2 := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 3554
Scope_Result_2 := Cast_Result_2;
   --# end
      Finalizer_Scope_2189;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Entity_Base_Type_Decl;
end if;
   If_Result_3 := Result_Var_2;
end if;
--# expr-done 3555
Let_Result := If_Result_3; 
--# end
--# expr-done 3556
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_2189;
               Finalizer_Scope_2190;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Previous_Part;
--# end
--# property-start BaseTypeDecl.is_private /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6110
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Is_Private
  (Node : Bare_Base_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Is_Private;
--# end
--# property-start BaseTypeDecl.root_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6170
pragma Warnings (Off, "is not referenced");
 function Base_Type_Decl_P_Root_Type
  (Node : Bare_Base_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Type_Decl := Bare_Base_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Type_Decl :=
     Internal_Entity_Base_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := Ent;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Type_Decl_P_Root_Type;
--# end
function Base_Type_Decl_P_Dottable_Subps
  (E : Entity
  ) return Internal_Inner_Env_Assoc_Array_Access
is
      E_Info : Internal_Entity_Info :=
         Shed_Rebindings (E.Info, Node_Env (E.Node));
   Result : constant Internal_Inner_Env_Assoc_Array_Access :=
      Base_Type_Decl_P_Dottable_Subps (E.Node, E_Info);
begin
      return Result;
end;
   --
   --  Primitives for Bare_Base_Subtype_Decl
   --
      procedure Initialize_Fields_For_Base_Subtype_Decl
        (Self : Bare_Base_Subtype_Decl
         ; Base_Type_Decl_F_Name : Bare_Defining_Name
        ) is
      begin
            Initialize_Fields_For_Base_Type_Decl
              (Self, Base_Type_Decl_F_Name);
      end Initialize_Fields_For_Base_Subtype_Decl;
--# property-start BaseSubtypeDecl.from_type_bound /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7914
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_From_Type_Bound
  (Node : Bare_Base_Subtype_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_1 : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3560 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3559 '<bind at ast.py:7919>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7919
--# expr-start 3557 '<FieldAccess for origin_node at ast.py:7919>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7919
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 3557
Bound_Origin := Fld; 
--# expr-start 3558 '<FieldAccess for get_type at ast.py:7919>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7919
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Bound_Origin, E_Info => Ent.Info);
--# end
--# expr-done 3558
Dyn_Var_Bind_Result := Fld_1; 
--# expr-done 3559
Let_Result := Dyn_Var_Bind_Result; 
--# end
--# expr-done 3560
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_From_Type_Bound;
--# end
--# property-start '[dispatcher]BaseSubtypeDecl.get_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Base_Subtype_Decl_P_Get_Type
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Base_Subtype_Decl (Self.Kind) is
               when Ada_Discrete_Base_Subtype_Decl =>
                  --# property-call-start DiscreteBaseSubtypeDecl.get_type
                  Property_Result := Discrete_Base_Subtype_Decl_P_Get_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Subtype_Decl =>
                  --# property-call-start SubtypeDecl.get_type
                  Property_Result := Subtype_Decl_P_Get_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Base_Subtype_Decl_P_Get_Type;
--# end
--# property-start BaseSubtypeDecl.primitives_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7931
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Primitives_Env
  (Node : Bare_Base_Subtype_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_611 with Inline_Always;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_611 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_611;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3562 '<FieldAccess for primitives_env at ast.py:7931>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7931
--# expr-start 3561 '<FieldAccess for from_type_bound at ast.py:7931>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7931
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubtypeDecl.from_type_bound
Fld := Libadalang.Implementation.Base_Subtype_Decl_P_From_Type_Bound (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3561
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.primitives_env'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Primitives_Env (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 3562
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_611;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_611;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Primitives_Env;
--# end
--# property-start BaseSubtypeDecl.array_ndims /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7933
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Array_Ndims
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Integer
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Integer;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Integer;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3564 '<FieldAccess for array_ndims at ast.py:7933>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7933
--# expr-start 3563 '<FieldAccess for get_type at ast.py:7933>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7933
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3563
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.array_ndims'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Array_Ndims (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3564
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Array_Ndims;
--# end
--# property-start BaseSubtypeDecl.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7934
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Defining_Env
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_613 with Inline_Always;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_613 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_613;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3566 '<FieldAccess for defining_env at ast.py:7934>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7934
--# expr-start 3565 '<FieldAccess for get_type at ast.py:7934>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7934
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3565
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Fld.Node, Origin => Origin, Include_Ud_Indexing => Include_Ud_Indexing, E_Info => Fld.Info);
--# end
--# expr-done 3566
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_613;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_613;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Defining_Env;
--# end
--# property-start BaseSubtypeDecl.canonical_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7936
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Canonical_Type
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3568 '<FieldAccess for canonical_type at ast.py:7936>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7936
--# expr-start 3567 '<FieldAccess for get_type at ast.py:7936>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7936
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3567
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.canonical_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Canonical_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3568
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Canonical_Type;
--# end
--# property-start BaseSubtypeDecl.record_def /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7937
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Record_Def
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Record_Def
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Record_Def;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Record_Def;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3570 '<FieldAccess for record_def at ast.py:7937>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7937
--# expr-start 3569 '<FieldAccess for get_type at ast.py:7937>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7937
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3569
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.record_def'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Record_Def (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3570
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Record_Def;
--# end
--# property-start BaseSubtypeDecl.accessed_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7938
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Accessed_Type
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3572 '<FieldAccess for accessed_type at ast.py:7938>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7938
--# expr-start 3571 '<FieldAccess for get_type at ast.py:7938>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7938
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3571
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3572
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Accessed_Type;
--# end
--# property-start BaseSubtypeDecl.is_int_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7939
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Is_Int_Type
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3574 '<FieldAccess for is_int_type at ast.py:7939>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7939
--# expr-start 3573 '<FieldAccess for get_type at ast.py:7939>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7939
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3573
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_int_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Int_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3574
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Is_Int_Type;
--# end
--# property-start BaseSubtypeDecl.is_discrete_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7940
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Is_Discrete_Type
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3576 '<FieldAccess for is_discrete_type at ast.py:7940>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7940
--# expr-start 3575 '<FieldAccess for get_type at ast.py:7940>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7940
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3575
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_discrete_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Discrete_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3576
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Is_Discrete_Type;
--# end
--# property-start BaseSubtypeDecl.is_real_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7942
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Is_Real_Type
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3578 '<FieldAccess for is_real_type at ast.py:7942>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7942
--# expr-start 3577 '<FieldAccess for get_type at ast.py:7942>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7942
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3577
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_real_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Real_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3578
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Is_Real_Type;
--# end
--# property-start BaseSubtypeDecl.is_float_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7943
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Is_Float_Type
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3580 '<FieldAccess for is_float_type at ast.py:7943>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7943
--# expr-start 3579 '<FieldAccess for get_type at ast.py:7943>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7943
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3579
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_float_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Float_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3580
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Is_Float_Type;
--# end
--# property-start BaseSubtypeDecl.is_fixed_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7944
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Is_Fixed_Point
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3582 '<FieldAccess for is_fixed_point at ast.py:7944>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7944
--# expr-start 3581 '<FieldAccess for get_type at ast.py:7944>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7944
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3581
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_fixed_point'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Fixed_Point (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3582
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Is_Fixed_Point;
--# end
--# property-start BaseSubtypeDecl.is_enum_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7945
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Is_Enum_Type
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3584 '<FieldAccess for is_enum_type at ast.py:7945>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7945
--# expr-start 3583 '<FieldAccess for get_type at ast.py:7945>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7945
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3583
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_enum_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Enum_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3584
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Is_Enum_Type;
--# end
--# property-start BaseSubtypeDecl.is_access_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7946
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Is_Access_Type
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3586 '<FieldAccess for is_access_type at ast.py:7946>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7946
--# expr-start 3585 '<FieldAccess for get_type at ast.py:7946>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7946
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3585
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_access_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Access_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3586
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Is_Access_Type;
--# end
--# property-start BaseSubtypeDecl.access_def /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7947
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Access_Def
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Access_Def
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Access_Def;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Access_Def;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3588 '<FieldAccess for access_def at ast.py:7947>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7947
--# expr-start 3587 '<FieldAccess for get_type at ast.py:7947>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7947
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3587
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.access_def'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Access_Def (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3588
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Access_Def;
--# end
--# property-start BaseSubtypeDecl.is_char_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7948
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Is_Char_Type
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3590 '<FieldAccess for is_char_type at ast.py:7948>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7948
--# expr-start 3589 '<FieldAccess for get_type at ast.py:7948>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7948
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3589
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_char_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Char_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3590
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Is_Char_Type;
--# end
--# property-start BaseSubtypeDecl.is_tagged_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7949
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Is_Tagged_Type
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3592 '<FieldAccess for is_tagged_type at ast.py:7949>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7949
--# expr-start 3591 '<FieldAccess for get_type at ast.py:7949>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7949
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3591
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_tagged_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Tagged_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3592
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Is_Tagged_Type;
--# end
--# property-start BaseSubtypeDecl.base_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7950
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Base_Type
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3594 '<FieldAccess for base_type at ast.py:7950>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7950
--# expr-start 3593 '<FieldAccess for get_type at ast.py:7950>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7950
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3593
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.base_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Base_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3594
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Base_Type;
--# end
--# property-start BaseSubtypeDecl.base_interfaces /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7951
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Base_Interfaces
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl_Array_Access
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_628 with Inline_Always;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_628 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_628;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3596 '<FieldAccess for base_interfaces at ast.py:7951>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7951
--# expr-start 3595 '<FieldAccess for get_type at ast.py:7951>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7951
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3595
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.base_interfaces'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Base_Interfaces (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3596
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_628;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_628;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Base_Interfaces;
--# end
--# property-start BaseSubtypeDecl.base_types /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7952
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Base_Types
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl_Array_Access
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_629 with Inline_Always;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_629 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_629;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3598 '<FieldAccess for base_types at ast.py:7952>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7952
--# expr-start 3597 '<FieldAccess for get_type at ast.py:7952>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7952
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3597
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.base_types'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Base_Types (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3598
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_629;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_629;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Base_Types;
--# end
--# property-start BaseSubtypeDecl.array_def /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7953
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Array_Def
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Type_Def
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Array_Type_Def;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Array_Type_Def;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3600 '<FieldAccess for array_def at ast.py:7953>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7953
--# expr-start 3599 '<FieldAccess for get_type at ast.py:7953>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7953
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3599
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.array_def'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Array_Def (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3600
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Array_Def;
--# end
--# property-start BaseSubtypeDecl.is_classwide /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7954
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Is_Classwide
  (Node : Bare_Base_Subtype_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3602 '<FieldAccess for is_classwide at ast.py:7954>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7954
--# expr-start 3601 '<FieldAccess for from_type_bound at ast.py:7954>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7954
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubtypeDecl.from_type_bound
Fld := Libadalang.Implementation.Base_Subtype_Decl_P_From_Type_Bound (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3601
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_classwide'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Classwide (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 3602
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Is_Classwide;
--# end
--# property-start BaseSubtypeDecl.discriminants_list /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7955
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Discriminants_List
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_632 with Inline_Always;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_632 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_632;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3604 '<FieldAccess for discriminants_list at ast.py:7955>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7955
--# expr-start 3603 '<FieldAccess for get_type at ast.py:7955>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7955
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3603
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.discriminants_list'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Discriminants_List (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3604
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_632;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_632;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Discriminants_List;
--# end
--# property-start BaseSubtypeDecl.is_iterable_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7956
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Is_Iterable_Type
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3606 '<FieldAccess for is_iterable_type at ast.py:7956>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7956
--# expr-start 3605 '<FieldAccess for get_type at ast.py:7956>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7956
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3605
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_iterable_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Iterable_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3606
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Is_Iterable_Type;
--# end
--# property-start BaseSubtypeDecl.iterable_comp_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7957
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Iterable_Comp_Type
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3608 '<FieldAccess for iterable_comp_type at ast.py:7957>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7957
--# expr-start 3607 '<FieldAccess for get_type at ast.py:7957>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7957
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3607
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.iterable_comp_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Iterable_Comp_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3608
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Iterable_Comp_Type;
--# end
--# property-start BaseSubtypeDecl.is_record_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7958
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Is_Record_Type
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3610 '<FieldAccess for is_record_type at ast.py:7958>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7958
--# expr-start 3609 '<FieldAccess for get_type at ast.py:7958>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7958
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3609
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_record_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Record_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3610
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Is_Record_Type;
--# end
--# property-start BaseSubtypeDecl.is_private /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7959
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Is_Private
  (Node : Bare_Base_Subtype_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3612 '<FieldAccess for is_private at ast.py:7959>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7959
--# expr-start 3611 '<FieldAccess for from_type_bound at ast.py:7959>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7959
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubtypeDecl.from_type_bound
Fld := Libadalang.Implementation.Base_Subtype_Decl_P_From_Type_Bound (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3611
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_private'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Private (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 3612
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Is_Private;
--# end
--# property-start BaseSubtypeDecl.root_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7960
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Root_Type
  (Node : Bare_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3614 '<FieldAccess for root_type at ast.py:7960>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7960
--# expr-start 3613 '<FieldAccess for get_type at ast.py:7960>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7960
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3613
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.root_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Root_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3614
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Root_Type;
--# end
--# property-start BaseSubtypeDecl.has_ud_indexing /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7962
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Has_Ud_Indexing
  (Node : Bare_Base_Subtype_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3616 '<FieldAccess for has_ud_indexing at ast.py:7963>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7963
--# expr-start 3615 '<FieldAccess for from_type_bound at ast.py:7963>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7963
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubtypeDecl.from_type_bound
Fld := Libadalang.Implementation.Base_Subtype_Decl_P_From_Type_Bound (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3615
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.has_ud_indexing'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Has_Ud_Indexing (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 3616
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Has_Ud_Indexing;
--# end
--# property-start BaseSubtypeDecl.constant_indexing_fns /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7965
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Constant_Indexing_Fns
  (Node : Bare_Base_Subtype_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl_Array_Access
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_639 with Inline_Always;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_639 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_639;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3618 '<FieldAccess for constant_indexing_fns at ast.py:7966>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7966
--# expr-start 3617 '<FieldAccess for from_type_bound at ast.py:7966>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7966
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubtypeDecl.from_type_bound
Fld := Libadalang.Implementation.Base_Subtype_Decl_P_From_Type_Bound (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3617
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.constant_indexing_fns'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Constant_Indexing_Fns (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 3618
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_639;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_639;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Constant_Indexing_Fns;
--# end
--# property-start BaseSubtypeDecl.variable_indexing_fns /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7968
pragma Warnings (Off, "is not referenced");
 function Base_Subtype_Decl_P_Variable_Indexing_Fns
  (Node : Bare_Base_Subtype_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl_Array_Access
is
   Self : Bare_Base_Subtype_Decl := Bare_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subtype_Decl :=
     Internal_Entity_Base_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_640 with Inline_Always;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_640 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_640;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3620 '<FieldAccess for variable_indexing_fns at ast.py:7969>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7969
--# expr-start 3619 '<FieldAccess for from_type_bound at ast.py:7969>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7969
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubtypeDecl.from_type_bound
Fld := Libadalang.Implementation.Base_Subtype_Decl_P_From_Type_Bound (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3619
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.variable_indexing_fns'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Variable_Indexing_Fns (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 3620
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_640;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_640;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subtype_Decl_P_Variable_Indexing_Fns;
--# end
   --
   --  Primitives for Bare_Discrete_Base_Subtype_Decl
   --
      procedure Initialize_Fields_For_Discrete_Base_Subtype_Decl
        (Self : Bare_Discrete_Base_Subtype_Decl
         ; Base_Type_Decl_F_Name : Bare_Defining_Name
        ) is
      begin
            Initialize_Fields_For_Base_Subtype_Decl
              (Self, Base_Type_Decl_F_Name);
      end Initialize_Fields_For_Discrete_Base_Subtype_Decl;
--# property-start DiscreteBaseSubtypeDecl.is_static_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8007
pragma Warnings (Off, "is not referenced");
 function Discrete_Base_Subtype_Decl_P_Is_Static_Decl
  (Node : Bare_Discrete_Base_Subtype_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Discrete_Base_Subtype_Decl := Bare_Discrete_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Discrete_Base_Subtype_Decl_P_Is_Static_Decl;
--# end
--# property-start DiscreteBaseSubtypeDecl.get_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8011
pragma Warnings (Off, "is not referenced");
 function Discrete_Base_Subtype_Decl_P_Get_Type
  (Node : Bare_Discrete_Base_Subtype_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Discrete_Base_Subtype_Decl := Bare_Discrete_Base_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Bare_Ada_Node;
Cast_Result : Bare_Base_Type_Decl;
As_Entity : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 3623 '<as_entity at ast.py:8012>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8012
--# expr-start 3622 '<Cast to BaseTypeDecl at ast.py:8012>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8012
--# expr-start 3621 '<parent at ast.py:8012>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8012
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 3621
   if Fld = null
      or else Fld.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Fld;
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 3622
Is_Null := Cast_Result = null; 
if Is_Null then
   If_Result := No_Entity_Base_Type_Decl;
else
As_Entity := (Info => E_Info, Node => Cast_Result); 
   If_Result := As_Entity;
end if;
--# expr-done 3623
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Discrete_Base_Subtype_Decl_P_Get_Type;
--# end
   --
   --  Primitives for Bare_Subtype_Decl
   --
      procedure Initialize_Fields_For_Subtype_Decl
        (Self : Bare_Subtype_Decl
         ; Base_Type_Decl_F_Name : Bare_Defining_Name
         ; Subtype_Decl_F_Subtype : Bare_Subtype_Indication
         ; Subtype_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Initialize_Fields_For_Base_Subtype_Decl
              (Self, Base_Type_Decl_F_Name);
            Self.Subtype_Decl_F_Subtype := Subtype_Decl_F_Subtype;
            Self.Subtype_Decl_F_Aspects := Subtype_Decl_F_Aspects;
      end Initialize_Fields_For_Subtype_Decl;
   function Subtype_Decl_F_Subtype
     (Node : Bare_Subtype_Decl) return Bare_Subtype_Indication
   is
   begin
         return Node.Subtype_Decl_F_Subtype;
   end;
--# property-start SubtypeDecl.get_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7981
pragma Warnings (Off, "is not referenced");
 function Subtype_Decl_P_Get_Type
  (Node : Bare_Subtype_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Subtype_Decl := Bare_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Subtype_Decl :=
     Internal_Entity_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Match_0 : Internal_Entity_Subtype_Decl;
Match_1 : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Subtype_Indication;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Base_Type_Decl;
Match_Prefix : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity_Subtype_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity_Base_Type_Decl;
Let_Result_1 : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Match_Result : Internal_Entity_Base_Type_Decl;
Let_Result_2 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3628 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 3627 '<Match at ast.py:7982>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7982
--# expr-start 3625 '<FieldAccess for designated_type at ast.py:7982>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7982
--# expr-start 3624 '<FieldAccess for subtype at ast.py:7982>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7982
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Subtype_Indication (Node => Ent.Node.Subtype_Decl_F_Subtype, Info => Ent.Info);
--# expr-done 3624
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3625
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld_1; 
case Ada_Base_Type_Decl (Match_Prefix.Node.Kind) is
         when Ada_Subtype_Decl =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Subtype_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind st Match_0
--# expr-start 3626 '<FieldAccess for get_type at ast.py:7983>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7983
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Match_0.Node, Origin => Origin, E_Info => Match_0.Info);
--# end
--# expr-done 3626
Let_Result := Fld_2; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Discrete_Base_Subtype_Decl | Ada_Classwide_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Base_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind t Match_1
Let_Result_1 := Match_1; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
end case;
--# expr-done 3627
Let_Result_2 := Match_Result; 
--# end
--# expr-done 3628
      Property_Result := Let_Result_2;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subtype_Decl_P_Get_Type;
--# end
--# property-start SubtypeDecl.discrete_range /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7988
pragma Warnings (Off, "is not referenced");
 function Subtype_Decl_P_Discrete_Range
  (Node : Bare_Subtype_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Discrete_Range
is
   Self : Bare_Subtype_Decl := Bare_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Subtype_Decl :=
     Internal_Entity_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Discrete_Range;
      Fld : Internal_Entity_Subtype_Indication;
Fld_1 : Internal_Discrete_Range;
Let_Result : Internal_Discrete_Range;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3631 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3630 '<FieldAccess for discrete_range at ast.py:7989>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7989
--# expr-start 3629 '<FieldAccess for subtype at ast.py:7989>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7989
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Subtype_Indication (Node => Ent.Node.Subtype_Decl_F_Subtype, Info => Ent.Info);
--# expr-done 3629
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SubtypeIndication.discrete_range
Fld_1 := Libadalang.Implementation.Subtype_Indication_P_Discrete_Range (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 3630
Let_Result := Fld_1; 
--# end
--# expr-done 3631
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subtype_Decl_P_Discrete_Range;
--# end
--# property-start SubtypeDecl.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7992
pragma Warnings (Off, "is not referenced");
 function Subtype_Decl_P_Xref_Equation
  (Node : Bare_Subtype_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Subtype_Decl := Bare_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Subtype_Decl :=
     Internal_Entity_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_643 with Inline_Always;
      Fld : Internal_Entity_Subtype_Indication;
Fld_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_643 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_643;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3634 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3633 '<FieldAccess for sub_equation at ast.py:7993>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7993
--# expr-start 3632 '<FieldAccess for subtype at ast.py:7993>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7993
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Subtype_Indication (Node => Ent.Node.Subtype_Decl_F_Subtype, Info => Ent.Info);
--# expr-done 3632
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3633
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 3634
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_643;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_643;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subtype_Decl_P_Xref_Equation;
--# end
--# property-start SubtypeDecl.is_static_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7995
pragma Warnings (Off, "is not referenced");
 function Subtype_Decl_P_Is_Static_Decl
  (Node : Bare_Subtype_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Subtype_Decl := Bare_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Subtype_Decl :=
     Internal_Entity_Subtype_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Fld : Internal_Entity_Subtype_Indication;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3636 '<FieldAccess for is_static_subtype at ast.py:7995>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7995
--# expr-start 3635 '<FieldAccess for subtype at ast.py:7995>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7995
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Subtype_Indication (Node => Ent.Node.Subtype_Decl_F_Subtype, Info => Ent.Info);
--# expr-done 3635
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SubtypeIndication.is_static_subtype
Fld_1 := Libadalang.Implementation.Subtype_Indication_P_Is_Static_Subtype (Node => Fld.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld.Info);
--# end
--# expr-done 3636
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subtype_Decl_P_Is_Static_Decl;
--# end
--# property-start SubtypeDecl.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7996
pragma Warnings (Off, "is not referenced");
 function Subtype_Decl_P_Xref_Entry_Point
  (Node : Bare_Subtype_Decl
  )
   return Boolean
is
   Self : Bare_Subtype_Decl := Bare_Subtype_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subtype_Decl_P_Xref_Entry_Point;
--# end
   --
   --  Primitives for Bare_Classwide_Type_Decl
   --
      procedure Initialize_Fields_For_Classwide_Type_Decl
        (Self : Bare_Classwide_Type_Decl
         ; Base_Type_Decl_F_Name : Bare_Defining_Name
        ) is
      begin
            Initialize_Fields_For_Base_Type_Decl
              (Self, Base_Type_Decl_F_Name);
      end Initialize_Fields_For_Classwide_Type_Decl;
--# property-start ClasswideTypeDecl.typedecl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6409
pragma Warnings (Off, "is not referenced");
 function Classwide_Type_Decl_P_Typedecl
  (Node : Bare_Classwide_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Classwide_Type_Decl := Bare_Classwide_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Bare_Ada_Node;
Cast_Result : Bare_Base_Type_Decl;
As_Entity : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 3639 '<as_entity at ast.py:6409>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6409
--# expr-start 3638 '<Cast to BaseTypeDecl at ast.py:6409>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6409
--# expr-start 3637 '<parent at ast.py:6409>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6409
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 3637
   if Fld = null
      or else Fld.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Fld;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 3638
Is_Null := Cast_Result = null; 
if Is_Null then
   If_Result := No_Entity_Base_Type_Decl;
else
As_Entity := (Info => E_Info, Node => Cast_Result); 
   If_Result := As_Entity;
end if;
--# expr-done 3639
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Classwide_Type_Decl_P_Typedecl;
--# end
--# property-start ClasswideTypeDecl.is_classwide /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6411
pragma Warnings (Off, "is not referenced");
 function Classwide_Type_Decl_P_Is_Classwide
  (Node : Bare_Classwide_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Classwide_Type_Decl := Bare_Classwide_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Classwide_Type_Decl_P_Is_Classwide;
--# end
--# property-start ClasswideTypeDecl.is_tagged_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6413
pragma Warnings (Off, "is not referenced");
 function Classwide_Type_Decl_P_Is_Tagged_Type
  (Node : Bare_Classwide_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Classwide_Type_Decl := Bare_Classwide_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Classwide_Type_Decl_P_Is_Tagged_Type;
--# end
--# property-start ClasswideTypeDecl.base_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6414
pragma Warnings (Off, "is not referenced");
 function Classwide_Type_Decl_P_Base_Type
  (Node : Bare_Classwide_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Classwide_Type_Decl := Bare_Classwide_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Classwide_Type_Decl :=
     Internal_Entity_Classwide_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3641 '<FieldAccess for base_type at ast.py:6414>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6414
--# expr-start 3640 '<FieldAccess for typedecl at ast.py:6414>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6414
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ClasswideTypeDecl.typedecl
Fld := Libadalang.Implementation.Classwide_Type_Decl_P_Typedecl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3640
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.base_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Base_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3641
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Classwide_Type_Decl_P_Base_Type;
--# end
--# property-start ClasswideTypeDecl.base_interfaces /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6415
pragma Warnings (Off, "is not referenced");
 function Classwide_Type_Decl_P_Base_Interfaces
  (Node : Bare_Classwide_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl_Array_Access
is
   Self : Bare_Classwide_Type_Decl := Bare_Classwide_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Classwide_Type_Decl :=
     Internal_Entity_Classwide_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_472 with Inline_Always;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_472 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_472;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3643 '<FieldAccess for base_interfaces at ast.py:6415>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6415
--# expr-start 3642 '<FieldAccess for typedecl at ast.py:6415>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6415
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ClasswideTypeDecl.typedecl
Fld := Libadalang.Implementation.Classwide_Type_Decl_P_Typedecl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3642
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.base_interfaces'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Base_Interfaces (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3643
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_472;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_472;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Classwide_Type_Decl_P_Base_Interfaces;
--# end
--# property-start ClasswideTypeDecl.record_def /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6416
pragma Warnings (Off, "is not referenced");
 function Classwide_Type_Decl_P_Record_Def
  (Node : Bare_Classwide_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Record_Def
is
   Self : Bare_Classwide_Type_Decl := Bare_Classwide_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Classwide_Type_Decl :=
     Internal_Entity_Classwide_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Record_Def;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Record_Def;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3645 '<FieldAccess for record_def at ast.py:6416>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6416
--# expr-start 3644 '<FieldAccess for typedecl at ast.py:6416>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6416
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ClasswideTypeDecl.typedecl
Fld := Libadalang.Implementation.Classwide_Type_Decl_P_Typedecl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3644
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.record_def'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Record_Def (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3645
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Classwide_Type_Decl_P_Record_Def;
--# end
--# property-start ClasswideTypeDecl.classwide_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6417
pragma Warnings (Off, "is not referenced");
 function Classwide_Type_Decl_P_Classwide_Type
  (Node : Bare_Classwide_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Classwide_Type_Decl
is
   Self : Bare_Classwide_Type_Decl := Bare_Classwide_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Classwide_Type_Decl :=
     Internal_Entity_Classwide_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Classwide_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := Ent;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Classwide_Type_Decl_P_Classwide_Type;
--# end
--# property-start ClasswideTypeDecl.is_iterable_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6418
pragma Warnings (Off, "is not referenced");
 function Classwide_Type_Decl_P_Is_Iterable_Type
  (Node : Bare_Classwide_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Classwide_Type_Decl := Bare_Classwide_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Classwide_Type_Decl :=
     Internal_Entity_Classwide_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3647 '<FieldAccess for is_iterable_type at ast.py:6418>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6418
--# expr-start 3646 '<FieldAccess for typedecl at ast.py:6418>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6418
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ClasswideTypeDecl.typedecl
Fld := Libadalang.Implementation.Classwide_Type_Decl_P_Typedecl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3646
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_iterable_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Iterable_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3647
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Classwide_Type_Decl_P_Is_Iterable_Type;
--# end
--# property-start ClasswideTypeDecl.iterable_comp_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6419
pragma Warnings (Off, "is not referenced");
 function Classwide_Type_Decl_P_Iterable_Comp_Type
  (Node : Bare_Classwide_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Classwide_Type_Decl := Bare_Classwide_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Classwide_Type_Decl :=
     Internal_Entity_Classwide_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3649 '<FieldAccess for iterable_comp_type at ast.py:6419>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6419
--# expr-start 3648 '<FieldAccess for typedecl at ast.py:6419>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6419
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ClasswideTypeDecl.typedecl
Fld := Libadalang.Implementation.Classwide_Type_Decl_P_Typedecl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3648
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.iterable_comp_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Iterable_Comp_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3649
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Classwide_Type_Decl_P_Iterable_Comp_Type;
--# end
--# property-start ClasswideTypeDecl.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6420
pragma Warnings (Off, "is not referenced");
 function Classwide_Type_Decl_P_Defining_Env
  (Node : Bare_Classwide_Type_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Classwide_Type_Decl := Bare_Classwide_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Classwide_Type_Decl :=
     Internal_Entity_Classwide_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_477 with Inline_Always;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_477 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_477;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3651 '<FieldAccess for defining_env at ast.py:6420>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6420
--# expr-start 3650 '<FieldAccess for typedecl at ast.py:6420>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6420
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ClasswideTypeDecl.typedecl
Fld := Libadalang.Implementation.Classwide_Type_Decl_P_Typedecl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3650
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Fld.Node, Origin => Origin, Include_Ud_Indexing => Include_Ud_Indexing, E_Info => Fld.Info);
--# end
--# expr-done 3651
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_477;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_477;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Classwide_Type_Decl_P_Defining_Env;
--# end
--# property-start ClasswideTypeDecl.is_private /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6421
pragma Warnings (Off, "is not referenced");
 function Classwide_Type_Decl_P_Is_Private
  (Node : Bare_Classwide_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Classwide_Type_Decl := Bare_Classwide_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Classwide_Type_Decl :=
     Internal_Entity_Classwide_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3653 '<FieldAccess for is_private at ast.py:6421>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6421
--# expr-start 3652 '<FieldAccess for typedecl at ast.py:6421>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6421
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ClasswideTypeDecl.typedecl
Fld := Libadalang.Implementation.Classwide_Type_Decl_P_Typedecl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3652
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_private'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Private (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 3653
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Classwide_Type_Decl_P_Is_Private;
--# end
--# property-start ClasswideTypeDecl.is_in_private_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6422
pragma Warnings (Off, "is not referenced");
 function Classwide_Type_Decl_P_Is_In_Private_Part
  (Node : Bare_Classwide_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Classwide_Type_Decl := Bare_Classwide_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Classwide_Type_Decl :=
     Internal_Entity_Classwide_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 3655 '<FieldAccess for is_in_private_part at ast.py:6422>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6422
--# expr-start 3654 '<FieldAccess for typedecl at ast.py:6422>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6422
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ClasswideTypeDecl.typedecl
Fld := Libadalang.Implementation.Classwide_Type_Decl_P_Typedecl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3654
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_in_private_part'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_In_Private_Part (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 3655
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Classwide_Type_Decl_P_Is_In_Private_Part;
--# end
--# property-start ClasswideTypeDecl.get_aspect_assoc /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6425
pragma Warnings (Off, "is not referenced");
 function Classwide_Type_Decl_P_Get_Aspect_Assoc
  (Node : Bare_Classwide_Type_Decl
      ; Name : Symbol_Type
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Aspect_Assoc
is
   Self : Bare_Classwide_Type_Decl := Bare_Classwide_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Classwide_Type_Decl :=
     Internal_Entity_Classwide_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind name Name
   Property_Result : Internal_Entity_Aspect_Assoc;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Aspect_Assoc;
Let_Result : Internal_Entity_Aspect_Assoc;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3658 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3657 '<FieldAccess for get_aspect_assoc at ast.py:6426>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6426
--# expr-start 3656 '<FieldAccess for typedecl at ast.py:6426>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6426
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ClasswideTypeDecl.typedecl
Fld := Libadalang.Implementation.Classwide_Type_Decl_P_Typedecl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3656
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.get_aspect_assoc'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Get_Aspect_Assoc (Node => Fld.Node, Name => Name, E_Info => Fld.Info);
--# end
--# expr-done 3657
Let_Result := Fld_1; 
--# end
--# expr-done 3658
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Classwide_Type_Decl_P_Get_Aspect_Assoc;
--# end
--# property-start ClasswideTypeDecl.is_interface_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6428
pragma Warnings (Off, "is not referenced");
 function Classwide_Type_Decl_P_Is_Interface_Type
  (Node : Bare_Classwide_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Classwide_Type_Decl := Bare_Classwide_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Classwide_Type_Decl :=
     Internal_Entity_Classwide_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3660 '<FieldAccess for is_interface_type at ast.py:6428>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6428
--# expr-start 3659 '<FieldAccess for typedecl at ast.py:6428>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6428
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ClasswideTypeDecl.typedecl
Fld := Libadalang.Implementation.Classwide_Type_Decl_P_Typedecl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3659
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_interface_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Interface_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3660
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Classwide_Type_Decl_P_Is_Interface_Type;
--# end
--# property-start ClasswideTypeDecl.discriminants_list /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6430
pragma Warnings (Off, "is not referenced");
 function Classwide_Type_Decl_P_Discriminants_List
  (Node : Bare_Classwide_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Classwide_Type_Decl := Bare_Classwide_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Classwide_Type_Decl :=
     Internal_Entity_Classwide_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_482 with Inline_Always;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_482 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_482;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3662 '<FieldAccess for discriminants_list at ast.py:6430>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6430
--# expr-start 3661 '<FieldAccess for typedecl at ast.py:6430>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6430
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ClasswideTypeDecl.typedecl
Fld := Libadalang.Implementation.Classwide_Type_Decl_P_Typedecl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3661
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.discriminants_list'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Discriminants_List (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3662
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_482;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_482;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Classwide_Type_Decl_P_Discriminants_List;
--# end
--# property-start ClasswideTypeDecl.previous_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6434
pragma Warnings (Off, "is not referenced");
 function Classwide_Type_Decl_P_Previous_Part
  (Node : Bare_Classwide_Type_Decl
      ; Go_To_Incomplete : Boolean
         := True
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Classwide_Type_Decl := Bare_Classwide_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Classwide_Type_Decl :=
     Internal_Entity_Classwide_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind go_to_incomplete Go_To_Incomplete
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Classwide_Type_Decl;
Scope_Result : Internal_Entity_Classwide_Type_Decl;
Result_Var : Internal_Entity_Classwide_Type_Decl;
Let_Result : Internal_Entity_Classwide_Type_Decl;
Cast_Result : Internal_Entity_Base_Type_Decl;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Classwide_Type_Decl_Classwide_Type_Decl_P_Previous_Part,
            Items    => new Mmz_Key_Array (1 ..  3))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Classwide_Type_Decl,
                                As_Bare_Classwide_Type_Decl => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Boolean,
                                      As_Boolean => Go_To_Incomplete);
               Mmz_K.Items (3) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Base_Type_Decl;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
--# expr-start 3666 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3664 '<FieldAccess for previous_part at ast.py:6435>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6435
--# expr-start 3663 '<FieldAccess for typedecl at ast.py:6435>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6435
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ClasswideTypeDecl.typedecl
Fld := Libadalang.Implementation.Classwide_Type_Decl_P_Typedecl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3663
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.previous_part'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Previous_Part (Node => Fld.Node, Go_To_Incomplete => Go_To_Incomplete, E_Info => Fld.Info);
--# end
--# expr-done 3664
Var_Expr := Fld_1;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind pp Var_Expr
--# expr-start 3665 '<FieldAccess for classwide_type at ast.py:6436>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6436
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.classwide_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Classwide_Type (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 3665
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Classwide_Type_Decl;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 3666
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Let_Result.Node,
         Info => Let_Result.Info);
      Property_Result := Cast_Result;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Base_Type_Decl,
                        As_Internal_Entity_Base_Type_Decl => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Classwide_Type_Decl_P_Previous_Part;
--# end
--# property-start ClasswideTypeDecl.canonical_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6439
pragma Warnings (Off, "is not referenced");
 function Classwide_Type_Decl_P_Canonical_Type
  (Node : Bare_Classwide_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Classwide_Type_Decl := Bare_Classwide_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Classwide_Type_Decl :=
     Internal_Entity_Classwide_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Entity_Classwide_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Classwide_Type_Decl;
Scope_Result : Internal_Entity_Classwide_Type_Decl;
Cast_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Result_Var_1 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
--# expr-start 3668 '<FieldAccess for canonical_type at ast.py:6439>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6439
--# expr-start 3667 '<FieldAccess for typedecl at ast.py:6439>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6439
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ClasswideTypeDecl.typedecl
Fld := Libadalang.Implementation.Classwide_Type_Decl_P_Typedecl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3667
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.canonical_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Canonical_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3668
Var_Expr := Fld_1;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind t Var_Expr
--# expr-start 3669 '<FieldAccess for classwide_type at ast.py:6443>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6443
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.classwide_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Classwide_Type (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 3669
Var_Expr_1 := Fld_2;
if Var_Expr_1 /= No_Entity_Classwide_Type_Decl then
   --# scope-start
   --# bind e Var_Expr_1
Scope_Result := Var_Expr_1;
   --# end
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Scope_Result.Node,
         Info => Scope_Result.Info);
   Result_Var := Cast_Result;
else
   Result_Var := Var_Expr;
end if;
Scope_Result_1 := Result_Var;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Base_Type_Decl;
end if;
      Property_Result := Result_Var_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Classwide_Type_Decl_P_Canonical_Type;
--# end
   --
   --  Primitives for Bare_Incomplete_Type_Decl
   --
      procedure Incomplete_Type_Decl_Pre_Env_Actions
        (Self            : Bare_Incomplete_Type_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_231 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:20195");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_232 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Incomplete_Type_Decl
        (Self : Bare_Incomplete_Type_Decl
         ; Base_Type_Decl_F_Name : Bare_Defining_Name
         ; Incomplete_Type_Decl_F_Discriminants : Bare_Discriminant_Part
        ) is
      begin
            Initialize_Fields_For_Base_Type_Decl
              (Self, Base_Type_Decl_F_Name);
            Self.Incomplete_Type_Decl_F_Discriminants := Incomplete_Type_Decl_F_Discriminants;
      end Initialize_Fields_For_Incomplete_Type_Decl;
   function Incomplete_Type_Decl_F_Discriminants
     (Node : Bare_Incomplete_Type_Decl) return Bare_Discriminant_Part
   is
   begin
         return Node.Incomplete_Type_Decl_F_Discriminants;
   end;
--# property-start IncompleteTypeDecl.find_next_part_in /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20181
pragma Warnings (Off, "is not referenced");
 function Incomplete_Type_Decl_P_Find_Next_Part_In
  (Node : Bare_Incomplete_Type_Decl
      ; Decl_Part : Internal_Entity_Declarative_Part
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Incomplete_Type_Decl := Bare_Incomplete_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Incomplete_Type_Decl :=
     Internal_Entity_Incomplete_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind decl_part Decl_Part
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Ada_Node_List;
Fld_1 : Bare_Ada_Node_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
Fld_3 : Internal_Entity_Defining_Name;
Fld_4 : Symbol_Type;
Fld_5 : Boolean;
Cast_Expr_1 : Internal_Entity_Incomplete_Type_Decl;
Cast_Result_1 : Internal_Entity;
Cast_Expr_2 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Is_Equiv : Boolean;
Not_Val : Boolean;
If_Result_1 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Item_142 : Internal_Entity;
Bare_Item_142 : Bare_Ada_Node;
Find_Result : Internal_Entity;
Cast_Result_3 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3679 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3678 '<Cast to BaseTypeDecl at ast.py:20187>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20187
--# expr-start 3677 '<Find at ast.py:20187>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20187
   --# expr-start 3670 '<FieldAccess for decls at ast.py:20187>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20187
      if Decl_Part.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Ada_Node_List (Node => Decl_Part.Node.Declarative_Part_F_Decls, Info => Decl_Part.Info);
--# expr-done 3670
Fld_1 := Fld.Node;
      Find_Result := No_Entity;
      if Fld_1 = null then
      Find_Result := No_Entity;
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_1;
      begin
         for Untyped_Item_142 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_142 := Untyped_Item_142; 
Is_Null := Bare_Item_142 = null; 
if Is_Null then
   If_Result := No_Entity;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_142); 
   If_Result := As_Entity;
end if;
                  Item_142 := If_Result; 
   --# scope-start
               --# bind t Item_142
--# expr-start 3671 '<Cast to BaseTypeDecl at ast.py:20188>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20188
Cast_Expr := Item_142; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 3671
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind btd Var_Expr
--# expr-start 3674 '<FieldAccess for name_is at ast.py:20190>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20190
--# expr-start 3672 '<FieldAccess for name at ast.py:20190>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20190
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Defining_Name (Node => Var_Expr.Node.Base_Type_Decl_F_Name, Info => Var_Expr.Info);
--# expr-done 3672
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3673 '<FieldAccess for name_symbol at ast.py:20190>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20190
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Self);
--# end
--# expr-done 3673
--# property-call-start Name.name_is
Fld_5 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_3.Node, Sym => Fld_4);
--# end
--# expr-done 3674
if Fld_5 then
   --# expr-start 3676 '<Not at ast.py:20190>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20190
--# expr-start 3675 '<Eq at ast.py:20190>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20190
Cast_Expr_1 := Ent; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Cast_Expr_2 := Var_Expr; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
Is_Equiv := Equivalent (Cast_Result_1, Cast_Result_2); 
--# expr-done 3675
Not_Val := not (Is_Equiv); 
--# expr-done 3676
   If_Result_1 := Not_Val;
else
   If_Result_1 := False;
end if;
Scope_Result := If_Result_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
      if Result_Var then
         Find_Result := Item_142;
      end if;
   --# end
               exit when Result_Var;
         end loop;
      end;
      end if;
--# expr-done 3677
   if Find_Result.Node = null
      or else Find_Result.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_3 := Create_Internal_Entity_Base_Type_Decl
        (Node => Find_Result.Node,
         Info => Find_Result.Info);
   else
         Cast_Result_3 := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 3678
Let_Result := Cast_Result_3; 
--# end
--# expr-done 3679
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Incomplete_Type_Decl_P_Find_Next_Part_In;
--# end
--# property-start IncompleteTypeDecl.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20199
pragma Warnings (Off, "is not referenced");
 function Incomplete_Type_Decl_P_Defining_Env
  (Node : Bare_Incomplete_Type_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Incomplete_Type_Decl := Bare_Incomplete_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1407 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Fld_1 : Lexical_Env := Empty_Env;
Array_Lit : Lexical_Env_Array_Access;
Group_Env : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1407 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Group_Env);
            end Finalizer_Scope_1407;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3683 '<env_group at ast.py:20200>' Group_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20200
--# expr-start 3682 '<ArrayLiteral at ast.py:20200>' Array_Lit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20200
--# expr-start 3680 '<FieldAccess for children_env at ast.py:20200>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20200
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children_Env (Node => Self);
--# expr-done 3680
--# expr-start 3681 '<FieldAccess for dottable_subps_env at ast.py:20200>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20200
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.dottable_subps_env
Fld_1 := Libadalang.Implementation.Base_Type_Decl_F_Dottable_Subps_Env (Node => Self);
--# end
--# expr-done 3681
Array_Lit := Create_Lexical_Env_Array (Internal_Lexical_Env_Array'(1 => Fld, 2 => Fld_1)); 
--# expr-done 3682
Group_Env := Group (Array_Lit, No_Metadata); 
--# expr-done 3683
      Property_Result := Group_Env;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1407;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1407;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Incomplete_Type_Decl_P_Defining_Env;
--# end
--# property-start IncompleteTypeDecl.discriminants_list /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20204
pragma Warnings (Off, "is not referenced");
 function Incomplete_Type_Decl_P_Discriminants_List
  (Node : Bare_Incomplete_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Incomplete_Type_Decl := Bare_Incomplete_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Incomplete_Type_Decl :=
     Internal_Entity_Incomplete_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_1408 with Inline_Always;
      Fld : Internal_Entity_Discriminant_Part;
Fld_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_1408 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1408;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 3685 '<FieldAccess for abstract_formal_params at ast.py:20204>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20204
--# expr-start 3684 '<FieldAccess for discriminants at ast.py:20204>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20204
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Discriminant_Part (Node => Ent.Node.Incomplete_Type_Decl_F_Discriminants, Info => Ent.Info);
--# expr-done 3684
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseFormalParamHolder.abstract_formal_params'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Formal_Param_Holder_P_Abstract_Formal_Params (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 3685
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1408;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1408;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Incomplete_Type_Decl_P_Discriminants_List;
--# end
--# property-start IncompleteTypeDecl._env_mappings_231 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20195
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_231
  (Node : Bare_Incomplete_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Incomplete_Type_Decl := Bare_Incomplete_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Incomplete_Type_Decl :=
     Internal_Entity_Incomplete_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1648 with Inline_Always;
      Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Incomplete_Type_Decl;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1648 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1648;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3688 '<New EnvAssoc at ast.py:20195>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20195
--# expr-start 3686 '<New DesignatedEnv at ast.py:20195>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20195
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 3686
--# expr-start 3687 '<FieldAccess for name_symbol at ast.py:20195>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20195
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 3687
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 3688
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1648;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1648;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_231;
--# end
--# property-start IncompleteTypeDecl._env_trans_parent_232 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20194
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_232
  (Node : Bare_Incomplete_Type_Decl
  )
   return Boolean
is
   Self : Bare_Incomplete_Type_Decl := Bare_Incomplete_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_232;
--# end
   --
   --  Primitives for Bare_Incomplete_Tagged_Type_Decl
   --
      procedure Initialize_Fields_For_Incomplete_Tagged_Type_Decl
        (Self : Bare_Incomplete_Tagged_Type_Decl
         ; Base_Type_Decl_F_Name : Bare_Defining_Name
         ; Incomplete_Type_Decl_F_Discriminants : Bare_Discriminant_Part
         ; Incomplete_Tagged_Type_Decl_F_Has_Abstract : Bare_Abstract_Node
        ) is
      begin
            Initialize_Fields_For_Incomplete_Type_Decl
              (Self, Base_Type_Decl_F_Name, Incomplete_Type_Decl_F_Discriminants);
            Self.Incomplete_Tagged_Type_Decl_F_Has_Abstract := Incomplete_Tagged_Type_Decl_F_Has_Abstract;
      end Initialize_Fields_For_Incomplete_Tagged_Type_Decl;
   function Incomplete_Tagged_Type_Decl_F_Has_Abstract
     (Node : Bare_Incomplete_Tagged_Type_Decl) return Bare_Abstract_Node
   is
   begin
         return Node.Incomplete_Tagged_Type_Decl_F_Has_Abstract;
   end;
--# property-start IncompleteTaggedTypeDecl.is_tagged_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20214
pragma Warnings (Off, "is not referenced");
 function Incomplete_Tagged_Type_Decl_P_Is_Tagged_Type
  (Node : Bare_Incomplete_Tagged_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Incomplete_Tagged_Type_Decl := Bare_Incomplete_Tagged_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Incomplete_Tagged_Type_Decl_P_Is_Tagged_Type;
--# end
   --
   --  Primitives for Bare_Protected_Type_Decl
   --
      procedure Protected_Type_Decl_Pre_Env_Actions
        (Self            : Bare_Protected_Type_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_13 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:8139");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_14 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            Libadalang.Implementation.Env_Names_15 (Node => 
Self);
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Protected_Type_Decl
        (Self : Bare_Protected_Type_Decl
         ; Base_Type_Decl_F_Name : Bare_Defining_Name
         ; Protected_Type_Decl_F_Discriminants : Bare_Discriminant_Part
         ; Protected_Type_Decl_F_Aspects : Bare_Aspect_Spec
         ; Protected_Type_Decl_F_Interfaces : Bare_Parent_List
         ; Protected_Type_Decl_F_Definition : Bare_Protected_Def
        ) is
      begin
            Initialize_Fields_For_Base_Type_Decl
              (Self, Base_Type_Decl_F_Name);
            Self.Protected_Type_Decl_F_Discriminants := Protected_Type_Decl_F_Discriminants;
            Self.Protected_Type_Decl_F_Aspects := Protected_Type_Decl_F_Aspects;
            Self.Protected_Type_Decl_F_Interfaces := Protected_Type_Decl_F_Interfaces;
            Self.Protected_Type_Decl_F_Definition := Protected_Type_Decl_F_Definition;
      end Initialize_Fields_For_Protected_Type_Decl;
   function Protected_Type_Decl_F_Discriminants
     (Node : Bare_Protected_Type_Decl) return Bare_Discriminant_Part
   is
   begin
         return Node.Protected_Type_Decl_F_Discriminants;
   end;
   function Protected_Type_Decl_F_Interfaces
     (Node : Bare_Protected_Type_Decl) return Bare_Parent_List
   is
   begin
         return Node.Protected_Type_Decl_F_Interfaces;
   end;
   function Protected_Type_Decl_F_Definition
     (Node : Bare_Protected_Type_Decl) return Bare_Protected_Def
   is
   begin
         return Node.Protected_Type_Decl_F_Definition;
   end;
--# property-start ProtectedTypeDecl.discriminants_list /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8107
pragma Warnings (Off, "is not referenced");
 function Protected_Type_Decl_P_Discriminants_List
  (Node : Bare_Protected_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Protected_Type_Decl := Bare_Protected_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Protected_Type_Decl :=
     Internal_Entity_Protected_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_657 with Inline_Always;
      Fld : Internal_Entity_Discriminant_Part;
Fld_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_657 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_657;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 3690 '<FieldAccess for abstract_formal_params at ast.py:8107>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8107
--# expr-start 3689 '<FieldAccess for discriminants at ast.py:8107>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Discriminant_Part (Node => Ent.Node.Protected_Type_Decl_F_Discriminants, Info => Ent.Info);
--# expr-done 3689
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseFormalParamHolder.abstract_formal_params'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Formal_Param_Holder_P_Abstract_Formal_Params (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 3690
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_657;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_657;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Protected_Type_Decl_P_Discriminants_List;
--# end
--# property-start ProtectedTypeDecl.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8109
pragma Warnings (Off, "is not referenced");
 function Protected_Type_Decl_P_Defining_Env
  (Node : Bare_Protected_Type_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Protected_Type_Decl := Bare_Protected_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Protected_Type_Decl :=
     Internal_Entity_Protected_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_658 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_658 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_658;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3691 '<FieldAccess for children_env at ast.py:8109>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8109
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 3691
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_658;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_658;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Protected_Type_Decl_P_Defining_Env;
--# end
--# property-start ProtectedTypeDecl.base_interfaces /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8111
pragma Warnings (Off, "is not referenced");
 function Protected_Type_Decl_P_Base_Interfaces
  (Node : Bare_Protected_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl_Array_Access
is
   Self : Bare_Protected_Type_Decl := Bare_Protected_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Protected_Type_Decl :=
     Internal_Entity_Protected_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_659 with Inline_Always;
      Fld : Internal_Entity_Parent_List;
Fld_1 : Bare_Parent_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld_3 : Internal_Entity_Base_Type_Decl;
Item_143 : Internal_Entity_Name;
Bare_Item_143 : Bare_Name;
Map_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_659 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_659;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3694 '<Map at ast.py:8112>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8112
   --# expr-start 3692 '<FieldAccess for interfaces at ast.py:8112>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8112
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Parent_List (Node => Ent.Node.Protected_Type_Decl_F_Interfaces, Info => Ent.Info);
--# expr-done 3692
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Internal_Entity_Base_Type_Decl_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Internal_Entity_Base_Type_Decl_Array (0);
      else
      declare
         Collection : constant Bare_Parent_List := Fld_1;
      begin
         for Untyped_Item_143 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_143 := Untyped_Item_143; 
Is_Null := Bare_Item_143 = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_143); 
   If_Result := As_Entity;
end if;
                  Item_143 := If_Result; 
   --# scope-start
               --# bind i Item_143
      --# expr-start 3693 '<FieldAccess for name_designated_type at ast.py:8112>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8112
      if Item_143.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_3 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Item_143.Node, E_Info => Item_143.Info);
--# end
--# expr-done 3693
         declare
            Item_To_Append : constant Internal_Entity_Base_Type_Decl := Fld_3;
         begin
            Internal_Entity_Base_Type_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Base_Type_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Type_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Base_Type_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Base_Type_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Base_Type_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 3694
      Property_Result := Map_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_659;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_659;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Protected_Type_Decl_P_Base_Interfaces;
--# end
--# property-start ProtectedTypeDecl.next_part_for_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8116
pragma Warnings (Off, "is not referenced");
 function Protected_Type_Decl_P_Next_Part_For_Decl
  (Node : Bare_Protected_Type_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Protected_Type_Decl := Bare_Protected_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Protected_Type_Decl :=
     Internal_Entity_Protected_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3696 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3695 '<FieldAccess for basic_decl_next_part_for_decl at ast.py:8117>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8117
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.basic_decl_next_part_for_decl
Fld := Libadalang.Implementation.Basic_Decl_P_Basic_Decl_Next_Part_For_Decl (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 3695
Let_Result := Fld; 
--# end
--# expr-done 3696
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Protected_Type_Decl_P_Next_Part_For_Decl;
--# end
--# property-start ProtectedTypeDecl.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8119
pragma Warnings (Off, "is not referenced");
 function Protected_Type_Decl_P_Xref_Entry_Point
  (Node : Bare_Protected_Type_Decl
  )
   return Boolean
is
   Self : Bare_Protected_Type_Decl := Bare_Protected_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Protected_Type_Decl_P_Xref_Entry_Point;
--# end
--# property-start ProtectedTypeDecl.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8122
pragma Warnings (Off, "is not referenced");
 function Protected_Type_Decl_P_Xref_Equation
  (Node : Bare_Protected_Type_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Protected_Type_Decl := Bare_Protected_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Protected_Type_Decl :=
     Internal_Entity_Protected_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_662 with Inline_Always;
            procedure Finalizer_Scope_2243 with Inline_Always;
      Fld : Internal_Entity_Parent_List;
Fld_1 : Bare_Parent_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld_3 : Logic_Equation;
Item_144 : Internal_Entity_Name;
Bare_Item_144 : Bare_Name;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_662 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_662;
            procedure Finalizer_Scope_2243 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_2243;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3701 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3700 '<LogicAll at ast.py:8123>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8123
--# expr-start 3699 '<Map at ast.py:8123>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8123
   --# expr-start 3697 '<FieldAccess for interfaces at ast.py:8123>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8123
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Parent_List (Node => Ent.Node.Protected_Type_Decl_F_Interfaces, Info => Ent.Info);
--# expr-done 3697
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Parent_List := Fld_1;
      begin
         for Untyped_Item_144 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_144 := Untyped_Item_144; 
Is_Null := Bare_Item_144 = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_144); 
   If_Result := As_Entity;
end if;
                  Item_144 := If_Result; 
   --# scope-start
               --# bind ifc Item_144
      --# expr-start 3698 '<FieldAccess for xref_equation at ast.py:8123>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8123
      if Item_144.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_3 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Item_144.Node, Env => Env, Origin => Origin, E_Info => Item_144.Info);
--# end
--# expr-done 3698
         declare
            Item_To_Append : constant Logic_Equation := Fld_3;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2243;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 3699
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8123") else null)); 
--# expr-done 3700
Let_Result := Logic_Boolean_Op; Inc_Ref (Let_Result);
--# end
--# expr-done 3701
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_662;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_662;
               Finalizer_Scope_2243;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Protected_Type_Decl_P_Xref_Equation;
--# end
--# property-start ProtectedTypeDecl.declarative_parts /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8126
pragma Warnings (Off, "is not referenced");
 function Protected_Type_Decl_P_Declarative_Parts
  (Node : Bare_Protected_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Declarative_Part_Array_Access
is
   Self : Bare_Protected_Type_Decl := Bare_Protected_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Protected_Type_Decl :=
     Internal_Entity_Protected_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_663 with Inline_Always;
            procedure Finalizer_Scope_2244 with Inline_Always;
      Block_Var : Internal_Entity_Protected_Def;
Var_Expr : Internal_Entity_Declarative_Part;
Fld : Internal_Entity_Protected_Def;
Fld_1 : Internal_Entity_Public_Part;
Cast_Result : Internal_Entity_Declarative_Part;
Singleton : Internal_Entity_Declarative_Part_Array_Access;
Fld_2 : Internal_Entity_Private_Part;
Cast_Result_1 : Internal_Entity_Declarative_Part;
Singleton_1 : Internal_Entity_Declarative_Part_Array_Access;
Scope_Result : Internal_Entity_Declarative_Part_Array_Access;
Result_Var : Internal_Entity_Declarative_Part_Array_Access;
Concat_Result : Internal_Entity_Declarative_Part_Array_Access;
Let_Result : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_663 is
            begin
                     Dec_Ref (Singleton);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_663;
            procedure Finalizer_Scope_2244 is
            begin
                     Dec_Ref (Singleton_1);
            end Finalizer_Scope_2244;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 3708 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3702 '<FieldAccess for definition at ast.py:8127>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8127
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Protected_Def (Node => Ent.Node.Protected_Type_Decl_F_Definition, Info => Ent.Info);
--# expr-done 3702
Block_Var := Fld; 
--# bind pdef Block_Var
--# expr-start 3707 '<Concat at ast.py:8128>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8128
--# expr-start 3704 '<Cast to DeclarativePart at ast.py:8128>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8128
--# expr-start 3703 '<FieldAccess for public_part at ast.py:8128>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8128
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Public_Part (Node => Block_Var.Node.Protected_Def_F_Public_Part, Info => Block_Var.Info);
--# expr-done 3703
      Cast_Result := Create_Internal_Entity_Declarative_Part
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
--# expr-done 3704
                Singleton := Create_Internal_Entity_Declarative_Part_Array (Items_Count => 1);
                Singleton.Items (1) := Cast_Result;
--# expr-start 3706 '<Cast to DeclarativePart at ast.py:8129>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8129
--# expr-start 3705 '<FieldAccess for private_part at ast.py:8129>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8129
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Private_Part (Node => Block_Var.Node.Protected_Def_F_Private_Part, Info => Block_Var.Info);
--# expr-done 3705
      Cast_Result_1 := Create_Internal_Entity_Declarative_Part
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
--# expr-done 3706
Var_Expr := Cast_Result_1;
if Var_Expr /= No_Entity_Declarative_Part then
   --# scope-start
                Singleton_1 := Create_Internal_Entity_Declarative_Part_Array (Items_Count => 1);
                Singleton_1.Items (1) := Var_Expr;
Scope_Result := Singleton_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2244;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Declarative_Part_Array_Type;
end if;
      Inc_Ref (Result_Var);
Concat_Result := Concat (Singleton, Result_Var); 
--# expr-done 3707
Let_Result := Concat_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 3708
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_663;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_663;
               Finalizer_Scope_2244;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Protected_Type_Decl_P_Declarative_Parts;
--# end
--# property-start ProtectedTypeDecl.env_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8133
pragma Warnings (Off, "is not referenced");
 function Protected_Type_Decl_P_Env_Names
  (Node : Bare_Protected_Type_Decl
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Protected_Type_Decl := Bare_Protected_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_664 with Inline_Always;
            procedure Finalizer_Scope_2245 with Inline_Always;
      Var_Expr : String_Type;
Fld : String_Type;
Sym : Symbol_Type;
Singleton : Symbol_Type_Array_Access;
Scope_Result : Symbol_Type_Array_Access;
Result_Var : Symbol_Type_Array_Access;
Let_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_664 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_664;
            procedure Finalizer_Scope_2245 is
            begin
                     Dec_Ref (Var_Expr);
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2245;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3711 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3709 '<FieldAccess for top_level_env_name at ast.py:8134>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8134
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.top_level_env_name
Fld := Libadalang.Implementation.Basic_Decl_P_Top_Level_Env_Name (Node => Self);
--# end
--# expr-done 3709
Var_Expr := Fld;
      Inc_Ref (Var_Expr);
if Var_Expr /= Empty_String then
   --# scope-start
   --# bind fqn Var_Expr
--# expr-start 3710 '<to_symbol at ast.py:8135>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8135
Sym := String_To_Symbol (Self.Unit.Context, Var_Expr); 
--# expr-done 3710
                Singleton := Create_Symbol_Type_Array (Items_Count => 1);
                Singleton.Items (1) := Sym;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2245;
   Result_Var := Scope_Result;
else
   Result_Var := No_Symbol_Type_Array_Type;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 3711
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_664;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_664;
               Finalizer_Scope_2245;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Protected_Type_Decl_P_Env_Names;
--# end
--# property-start ProtectedTypeDecl._env_mappings_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8139
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_13
  (Node : Bare_Protected_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Protected_Type_Decl := Bare_Protected_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Protected_Type_Decl :=
     Internal_Entity_Protected_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1430 with Inline_Always;
      Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Protected_Type_Decl;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1430 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1430;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3714 '<New EnvAssoc at ast.py:8139>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8139
--# expr-start 3712 '<New DesignatedEnv at ast.py:8139>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8139
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 3712
--# expr-start 3713 '<FieldAccess for name_symbol at ast.py:8139>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8139
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 3713
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 3714
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1430;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1430;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_13;
--# end
--# property-start ProtectedTypeDecl._env_trans_parent_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8138
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_14
  (Node : Bare_Protected_Type_Decl
  )
   return Boolean
is
   Self : Bare_Protected_Type_Decl := Bare_Protected_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_14;
--# end
--# property-start ProtectedTypeDecl._env_names_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8140
pragma Warnings (Off, "is not referenced");
 function Env_Names_15
  (Node : Bare_Protected_Type_Decl
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Protected_Type_Decl := Bare_Protected_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1432 with Inline_Always;
      Fld : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1432 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1432;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3715 '<FieldAccess for env_names at ast.py:8140>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8140
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ProtectedTypeDecl.env_names
Fld := Libadalang.Implementation.Protected_Type_Decl_P_Env_Names (Node => Self);
--# end
--# expr-done 3715
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1432;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1432;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Names_15;
--# end
   --
   --  Primitives for Bare_Task_Type_Decl
   --
      procedure Task_Type_Decl_Pre_Env_Actions
        (Self            : Bare_Task_Type_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_9 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:8059");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_10 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            Libadalang.Implementation.Env_Names_11 (Node => 
Self);
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Task_Type_Decl
        (Self : Bare_Task_Type_Decl
         ; Base_Type_Decl_F_Name : Bare_Defining_Name
         ; Task_Type_Decl_F_Discriminants : Bare_Discriminant_Part
         ; Task_Type_Decl_F_Aspects : Bare_Aspect_Spec
         ; Task_Type_Decl_F_Definition : Bare_Task_Def
        ) is
      begin
            Initialize_Fields_For_Base_Type_Decl
              (Self, Base_Type_Decl_F_Name);
            Self.Task_Type_Decl_F_Discriminants := Task_Type_Decl_F_Discriminants;
            Self.Task_Type_Decl_F_Aspects := Task_Type_Decl_F_Aspects;
            Self.Task_Type_Decl_F_Definition := Task_Type_Decl_F_Definition;
      end Initialize_Fields_For_Task_Type_Decl;
   function Task_Type_Decl_F_Discriminants
     (Node : Bare_Task_Type_Decl) return Bare_Discriminant_Part
   is
   begin
         return Node.Task_Type_Decl_F_Discriminants;
   end;
   function Task_Type_Decl_F_Definition
     (Node : Bare_Task_Type_Decl) return Bare_Task_Def
   is
   begin
         return Node.Task_Type_Decl_F_Definition;
   end;
--# property-start TaskTypeDecl.is_task_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8046
pragma Warnings (Off, "is not referenced");
 function Task_Type_Decl_P_Is_Task_Type
  (Node : Bare_Task_Type_Decl
  )
   return Boolean
is
   Self : Bare_Task_Type_Decl := Bare_Task_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Task_Type_Decl_P_Is_Task_Type;
--# end
--# property-start TaskTypeDecl.base_interfaces /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8048
pragma Warnings (Off, "is not referenced");
 function Task_Type_Decl_P_Base_Interfaces
  (Node : Bare_Task_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl_Array_Access
is
   Self : Bare_Task_Type_Decl := Bare_Task_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Task_Type_Decl :=
     Internal_Entity_Task_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_650 with Inline_Always;
      Fld : Internal_Entity_Task_Def;
Fld_1 : Internal_Entity_Parent_List;
Fld_2 : Bare_Parent_List;
Fld_3 : Internal_Entity_Info;
As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld_4 : Internal_Entity_Base_Type_Decl;
Item_145 : Internal_Entity_Name;
Bare_Item_145 : Bare_Name;
Map_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_650 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_650;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3719 '<Map at ast.py:8049>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8049
   --# expr-start 3717 '<FieldAccess for interfaces at ast.py:8049>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8049
--# expr-start 3716 '<FieldAccess for definition at ast.py:8049>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8049
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Task_Def (Node => Ent.Node.Task_Type_Decl_F_Definition, Info => Ent.Info);
--# expr-done 3716
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Parent_List (Node => Fld.Node.Task_Def_F_Interfaces, Info => Fld.Info);
--# expr-done 3717
Fld_2 := Fld_1.Node;
      declare
         Map_Result_Vec : Internal_Entity_Base_Type_Decl_Vectors.Vector;
      begin
      if Fld_2 = null then
      Map_Result := Create_Internal_Entity_Base_Type_Decl_Array (0);
      else
      declare
         Collection : constant Bare_Parent_List := Fld_2;
      begin
         for Untyped_Item_145 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_145 := Untyped_Item_145; 
Is_Null := Bare_Item_145 = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
Fld_3 := Fld_1.Info;
As_Entity := (Info => Fld_3, Node => Bare_Item_145); 
   If_Result := As_Entity;
end if;
                  Item_145 := If_Result; 
   --# scope-start
               --# bind i Item_145
      --# expr-start 3718 '<FieldAccess for name_designated_type at ast.py:8049>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8049
      if Item_145.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_4 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Item_145.Node, E_Info => Item_145.Info);
--# end
--# expr-done 3718
         declare
            Item_To_Append : constant Internal_Entity_Base_Type_Decl := Fld_4;
         begin
            Internal_Entity_Base_Type_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Base_Type_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Type_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Base_Type_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Base_Type_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Base_Type_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 3719
      Property_Result := Map_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_650;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_650;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Task_Type_Decl_P_Base_Interfaces;
--# end
--# property-start TaskTypeDecl.env_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8053
pragma Warnings (Off, "is not referenced");
 function Task_Type_Decl_P_Env_Names
  (Node : Bare_Task_Type_Decl
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Task_Type_Decl := Bare_Task_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_651 with Inline_Always;
            procedure Finalizer_Scope_2247 with Inline_Always;
      Var_Expr : String_Type;
Fld : String_Type;
Sym : Symbol_Type;
Singleton : Symbol_Type_Array_Access;
Scope_Result : Symbol_Type_Array_Access;
Result_Var : Symbol_Type_Array_Access;
Let_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_651 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_651;
            procedure Finalizer_Scope_2247 is
            begin
                     Dec_Ref (Var_Expr);
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2247;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3722 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3720 '<FieldAccess for top_level_env_name at ast.py:8054>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8054
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.top_level_env_name
Fld := Libadalang.Implementation.Basic_Decl_P_Top_Level_Env_Name (Node => Self);
--# end
--# expr-done 3720
Var_Expr := Fld;
      Inc_Ref (Var_Expr);
if Var_Expr /= Empty_String then
   --# scope-start
   --# bind fqn Var_Expr
--# expr-start 3721 '<to_symbol at ast.py:8055>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8055
Sym := String_To_Symbol (Self.Unit.Context, Var_Expr); 
--# expr-done 3721
                Singleton := Create_Symbol_Type_Array (Items_Count => 1);
                Singleton.Items (1) := Sym;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2247;
   Result_Var := Scope_Result;
else
   Result_Var := No_Symbol_Type_Array_Type;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 3722
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_651;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_651;
               Finalizer_Scope_2247;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Task_Type_Decl_P_Env_Names;
--# end
--# property-start TaskTypeDecl.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8063
pragma Warnings (Off, "is not referenced");
 function Task_Type_Decl_P_Defining_Env
  (Node : Bare_Task_Type_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Task_Type_Decl := Bare_Task_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Task_Type_Decl :=
     Internal_Entity_Task_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_652 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_652 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_652;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3723 '<FieldAccess for children_env at ast.py:8063>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8063
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 3723
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_652;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_652;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Task_Type_Decl_P_Defining_Env;
--# end
--# property-start TaskTypeDecl.discriminants_list /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8065
pragma Warnings (Off, "is not referenced");
 function Task_Type_Decl_P_Discriminants_List
  (Node : Bare_Task_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Task_Type_Decl := Bare_Task_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Task_Type_Decl :=
     Internal_Entity_Task_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_653 with Inline_Always;
      Fld : Internal_Entity_Discriminant_Part;
Fld_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_653 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_653;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 3725 '<FieldAccess for abstract_formal_params at ast.py:8065>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8065
--# expr-start 3724 '<FieldAccess for discriminants at ast.py:8065>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8065
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Discriminant_Part (Node => Ent.Node.Task_Type_Decl_F_Discriminants, Info => Ent.Info);
--# expr-done 3724
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseFormalParamHolder.abstract_formal_params'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Formal_Param_Holder_P_Abstract_Formal_Params (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 3725
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_653;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_653;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Task_Type_Decl_P_Discriminants_List;
--# end
--# property-start TaskTypeDecl.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8067
pragma Warnings (Off, "is not referenced");
 function Task_Type_Decl_P_Xref_Entry_Point
  (Node : Bare_Task_Type_Decl
  )
   return Boolean
is
   Self : Bare_Task_Type_Decl := Bare_Task_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Task_Type_Decl_P_Xref_Entry_Point;
--# end
--# property-start TaskTypeDecl.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8070
pragma Warnings (Off, "is not referenced");
 function Task_Type_Decl_P_Xref_Equation
  (Node : Bare_Task_Type_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Task_Type_Decl := Bare_Task_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Task_Type_Decl :=
     Internal_Entity_Task_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_655 with Inline_Always;
            procedure Finalizer_Scope_2248 with Inline_Always;
      Var_Expr : Internal_Entity_Task_Def;
Fld : Internal_Entity_Task_Def;
Fld_1 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_655 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_655;
            procedure Finalizer_Scope_2248 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2248;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3728 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3726 '<FieldAccess for definition at ast.py:8071>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8071
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Task_Def (Node => Ent.Node.Task_Type_Decl_F_Definition, Info => Ent.Info);
--# expr-done 3726
Var_Expr := Fld;
if Var_Expr /= No_Entity_Task_Def then
   --# scope-start
   --# bind d Var_Expr
--# expr-start 3727 '<FieldAccess for xref_equation at ast.py:8072>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8072
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_1 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 3727
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2248;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8073") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 3728
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_655;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_655;
               Finalizer_Scope_2248;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Task_Type_Decl_P_Xref_Equation;
--# end
--# property-start TaskTypeDecl.declarative_parts /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8077
pragma Warnings (Off, "is not referenced");
 function Task_Type_Decl_P_Declarative_Parts
  (Node : Bare_Task_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Declarative_Part_Array_Access
is
   Self : Bare_Task_Type_Decl := Bare_Task_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Task_Type_Decl :=
     Internal_Entity_Task_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_656 with Inline_Always;
            procedure Finalizer_Scope_2249 with Inline_Always;
            procedure Finalizer_Scope_2250 with Inline_Always;
      Var_Expr : Internal_Entity_Task_Def;
Var_Expr_1 : Internal_Entity_Declarative_Part;
Fld : Internal_Entity_Task_Def;
Fld_1 : Internal_Entity_Public_Part;
Cast_Result : Internal_Entity_Declarative_Part;
Singleton : Internal_Entity_Declarative_Part_Array_Access;
Fld_2 : Internal_Entity_Private_Part;
Cast_Result_1 : Internal_Entity_Declarative_Part;
Singleton_1 : Internal_Entity_Declarative_Part_Array_Access;
Scope_Result : Internal_Entity_Declarative_Part_Array_Access;
Result_Var : Internal_Entity_Declarative_Part_Array_Access;
Concat_Result : Internal_Entity_Declarative_Part_Array_Access;
Scope_Result_1 : Internal_Entity_Declarative_Part_Array_Access;
Result_Var_1 : Internal_Entity_Declarative_Part_Array_Access;
Let_Result : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_656 is
            begin
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_656;
            procedure Finalizer_Scope_2249 is
            begin
                     Dec_Ref (Singleton);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Concat_Result);
            end Finalizer_Scope_2249;
            procedure Finalizer_Scope_2250 is
            begin
                     Dec_Ref (Singleton_1);
            end Finalizer_Scope_2250;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 3735 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3729 '<FieldAccess for definition at ast.py:8078>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8078
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Task_Def (Node => Ent.Node.Task_Type_Decl_F_Definition, Info => Ent.Info);
--# expr-done 3729
Var_Expr := Fld;
if Var_Expr /= No_Entity_Task_Def then
   --# scope-start
   --# bind tdef Var_Expr
--# expr-start 3734 '<Concat at ast.py:8080>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8080
--# expr-start 3731 '<Cast to DeclarativePart at ast.py:8080>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8080
--# expr-start 3730 '<FieldAccess for public_part at ast.py:8080>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8080
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Public_Part (Node => Var_Expr.Node.Task_Def_F_Public_Part, Info => Var_Expr.Info);
--# expr-done 3730
      Cast_Result := Create_Internal_Entity_Declarative_Part
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
--# expr-done 3731
                Singleton := Create_Internal_Entity_Declarative_Part_Array (Items_Count => 1);
                Singleton.Items (1) := Cast_Result;
--# expr-start 3733 '<Cast to DeclarativePart at ast.py:8081>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8081
--# expr-start 3732 '<FieldAccess for private_part at ast.py:8081>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8081
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Private_Part (Node => Var_Expr.Node.Task_Def_F_Private_Part, Info => Var_Expr.Info);
--# expr-done 3732
      Cast_Result_1 := Create_Internal_Entity_Declarative_Part
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
--# expr-done 3733
Var_Expr_1 := Cast_Result_1;
if Var_Expr_1 /= No_Entity_Declarative_Part then
   --# scope-start
                Singleton_1 := Create_Internal_Entity_Declarative_Part_Array (Items_Count => 1);
                Singleton_1.Items (1) := Var_Expr_1;
Scope_Result := Singleton_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2250;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Declarative_Part_Array_Type;
end if;
      Inc_Ref (Result_Var);
Concat_Result := Concat (Singleton, Result_Var); 
--# expr-done 3734
Scope_Result_1 := Concat_Result;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2249;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Internal_Entity_Declarative_Part_Array_Type;
end if;
      Inc_Ref (Result_Var_1);
Let_Result := Result_Var_1; Inc_Ref (Let_Result);
--# end
--# expr-done 3735
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_656;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_656;
               Finalizer_Scope_2249;
               Finalizer_Scope_2250;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Task_Type_Decl_P_Declarative_Parts;
--# end
--# property-start TaskTypeDecl._env_mappings_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8059
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_9
  (Node : Bare_Task_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Task_Type_Decl := Bare_Task_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Task_Type_Decl :=
     Internal_Entity_Task_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1426 with Inline_Always;
      Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Task_Type_Decl;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1426 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1426;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3738 '<New EnvAssoc at ast.py:8059>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8059
--# expr-start 3736 '<New DesignatedEnv at ast.py:8059>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8059
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 3736
--# expr-start 3737 '<FieldAccess for name_symbol at ast.py:8059>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8059
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 3737
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 3738
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1426;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1426;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_9;
--# end
--# property-start TaskTypeDecl._env_trans_parent_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8058
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_10
  (Node : Bare_Task_Type_Decl
  )
   return Boolean
is
   Self : Bare_Task_Type_Decl := Bare_Task_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_10;
--# end
--# property-start TaskTypeDecl._env_names_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8060
pragma Warnings (Off, "is not referenced");
 function Env_Names_11
  (Node : Bare_Task_Type_Decl
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Task_Type_Decl := Bare_Task_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1428 with Inline_Always;
      Fld : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1428 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1428;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3739 '<FieldAccess for env_names at ast.py:8060>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8060
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TaskTypeDecl.env_names
Fld := Libadalang.Implementation.Task_Type_Decl_P_Env_Names (Node => Self);
--# end
--# expr-done 3739
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1428;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1428;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Names_11;
--# end
   --
   --  Primitives for Bare_Single_Task_Type_Decl
   --
      procedure Single_Task_Type_Decl_Pre_Env_Actions
        (Self            : Bare_Single_Task_Type_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_12 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Single_Task_Type_Decl
        (Self : Bare_Single_Task_Type_Decl
         ; Base_Type_Decl_F_Name : Bare_Defining_Name
         ; Task_Type_Decl_F_Discriminants : Bare_Discriminant_Part
         ; Task_Type_Decl_F_Aspects : Bare_Aspect_Spec
         ; Task_Type_Decl_F_Definition : Bare_Task_Def
        ) is
      begin
            Initialize_Fields_For_Task_Type_Decl
              (Self, Base_Type_Decl_F_Name, Task_Type_Decl_F_Discriminants, Task_Type_Decl_F_Aspects, Task_Type_Decl_F_Definition);
      end Initialize_Fields_For_Single_Task_Type_Decl;
--# property-start SingleTaskTypeDecl._env_trans_parent_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8090
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_12
  (Node : Bare_Single_Task_Type_Decl
  )
   return Boolean
is
   Self : Bare_Single_Task_Type_Decl := Bare_Single_Task_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_12;
--# end
   --
   --  Primitives for Bare_Type_Decl
   --
      procedure Type_Decl_Pre_Env_Actions
        (Self            : Bare_Type_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_3 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:6732");
         Dec_Ref (Mapping.Dest_Env);
      end;
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mappings : Internal_Env_Assoc_Array_Access :=
            Libadalang.Implementation.Env_Mappings_4 (Node => 
Self);
      begin
         for Mapping of Mappings.Items loop
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:6733");
         end loop;
         Dec_Ref (Mappings);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_5 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Type_Decl_Post_Env_Actions
        (Self : Bare_Type_Decl; State : in out PLE_Node_State) is
      begin
      if Libadalang.Implementation.Ref_Cond_8 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_6 (Node => 
Self);
            Env : Lexical_Env :=
              Libadalang.Implementation.Env_Dest_7 (Node => 
Self);
         begin
               if Is_Foreign_Strict (Env, Self) then
                  raise Property_Error with
                     "unsound foreign environment in RefEnvs ("
                     & "ast.py:6736)";
               end if;
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Type_Decl_P_Refined_Parent_Primitives_Env'Access,
               Transitive,
               (Inherited_Primitives => True, others => False),
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      end;
      procedure Initialize_Fields_For_Type_Decl
        (Self : Bare_Type_Decl
         ; Base_Type_Decl_F_Name : Bare_Defining_Name
         ; Type_Decl_F_Discriminants : Bare_Discriminant_Part
         ; Type_Decl_F_Type_Def : Bare_Type_Def
         ; Type_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Initialize_Fields_For_Base_Type_Decl
              (Self, Base_Type_Decl_F_Name);
            Self.Type_Decl_F_Discriminants := Type_Decl_F_Discriminants;
            Self.Type_Decl_F_Type_Def := Type_Decl_F_Type_Def;
            Self.Type_Decl_F_Aspects := Type_Decl_F_Aspects;
      Self.Lf_Present_Direct_Primitives_Env := False;
      Self.Lf_Stg_Direct_Primitives_Env := Empty_Env;
      Self.Lf_Present_Direct_Primitive_Subps := False;
      Self.Lf_Stg_Direct_Primitive_Subps := No_Internal_Inner_Env_Assoc_Array_Type;
      end Initialize_Fields_For_Type_Decl;
   function Type_Decl_F_Discriminants
     (Node : Bare_Type_Decl) return Bare_Discriminant_Part
   is
   begin
         return Node.Type_Decl_F_Discriminants;
   end;
   function Type_Decl_F_Type_Def
     (Node : Bare_Type_Decl) return Bare_Type_Def
   is
   begin
         return Node.Type_Decl_F_Type_Def;
   end;
--# property-start TypeDecl.is_iterable_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6456
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Is_Iterable_Type
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Match_0 : Internal_Entity_Derived_Type_Def;
Match_1 : Internal_Entity_Type_Def;
Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Entity_Base_Type_Decl;
Fld : Boolean;
Sym : Symbol_Type;
Fld_1 : Internal_Entity_Expr;
Is_Null : Boolean;
Not_Val : Boolean;
If_Result : Boolean;
Sym_1 : Symbol_Type;
Fld_2 : Internal_Entity_Expr;
Is_Null_1 : Boolean;
Not_Val_1 : Boolean;
If_Result_1 : Boolean;
Fld_3 : Internal_Entity_Type_Def;
Fld_4 : Internal_Entity_Base_Type_Decl;
Fld_5 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Match_Prefix : Internal_Entity_Type_Def;
Cast_Result : Internal_Entity_Derived_Type_Def;
Let_Result : Boolean;
Scope_Result_1 : Boolean;
Cast_Result_1 : Internal_Entity_Type_Def;
Let_Result_1 : Boolean;
Scope_Result_2 : Boolean;
Match_Result : Boolean;
If_Result_2 : Boolean;
Fld_6 : Internal_Entity_Base_Type_Decl;
Fld_7 : Boolean;
Scope_Result_3 : Boolean;
Result_Var_1 : Boolean;
If_Result_3 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
--# expr-start 3740 '<FieldAccess for is_array at ast.py:6461>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6461
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_array
Fld := Libadalang.Implementation.Basic_Decl_P_Is_Array (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3740
if Fld then
   If_Result := True;
else
   --# expr-start 3744 '<Not at ast.py:6462>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6462
--# expr-start 3743 '<IsNull at ast.py:6462>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6462
--# expr-start 3742 '<FieldAccess for get_aspect_spec_expr at ast.py:6462>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6462
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3741 '<Symbol Iterator_Element at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Iterator_Element); 
--# expr-done 3741
--# property-call-start BasicDecl.get_aspect_spec_expr
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Get_Aspect_Spec_Expr (Node => Ent.Node, Name => Sym, E_Info => Ent.Info);
--# end
--# expr-done 3742
Is_Null := Fld_1.Node = null; 
--# expr-done 3743
Not_Val := not (Is_Null); 
--# expr-done 3744
   If_Result := Not_Val;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 3748 '<Not at ast.py:6463>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6463
--# expr-start 3747 '<IsNull at ast.py:6463>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6463
--# expr-start 3746 '<FieldAccess for get_aspect_spec_expr at ast.py:6463>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6463
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3745 '<Symbol Iterable at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Iterable); 
--# expr-done 3745
--# property-call-start BasicDecl.get_aspect_spec_expr
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Get_Aspect_Spec_Expr (Node => Ent.Node, Name => Sym_1, E_Info => Ent.Info);
--# end
--# expr-done 3746
Is_Null_1 := Fld_2.Node = null; 
--# expr-done 3747
Not_Val_1 := not (Is_Null_1); 
--# expr-done 3748
   If_Result_1 := Not_Val_1;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
   --# expr-start 3752 '<Match at ast.py:6464>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6464
--# expr-start 3749 '<FieldAccess for type_def at ast.py:6464>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6464
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3749
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld_3; 
case Ada_Type_Def (Match_Prefix.Node.Kind) is
         when Ada_Derived_Type_Def =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Derived_Type_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind dtd Match_0
--# expr-start 3750 '<FieldAccess for base_type at ast.py:6466>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6466
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld_4 := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => Match_0.Node, Origin => Origin, E_Info => Match_0.Info);
--# end
--# expr-done 3750
Var_Expr := Fld_4;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind bt Var_Expr
--# expr-start 3751 '<FieldAccess for is_iterable_type at ast.py:6466>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6466
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_iterable_type'
Fld_5 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Iterable_Type (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 3751
Scope_Result := Fld_5;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Let_Result := Result_Var; 
--# end
Scope_Result_1 := Let_Result;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Access_To_Subp_Def .. Ada_Array_Type_Def | Ada_Enum_Type_Def .. Ada_Signed_Int_Type_Def =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Type_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind _ Match_1
Let_Result_1 := False; 
--# end
Scope_Result_2 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 3752
   If_Result_2 := Match_Result;
end if;
if If_Result_2 then
   If_Result_3 := True;
else
--# expr-start 3753 '<FieldAccess for previous_part at ast.py:6469>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6469
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.previous_part'
Fld_6 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Previous_Part (Node => Ent.Node, Go_To_Incomplete => False, E_Info => Ent.Info);
--# end
--# expr-done 3753
Var_Expr_1 := Fld_6;
if Var_Expr_1 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind pp Var_Expr_1
--# expr-start 3754 '<FieldAccess for is_iterable_type at ast.py:6469>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6469
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_iterable_type'
Fld_7 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Iterable_Type (Node => Var_Expr_1.Node, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 3754
Scope_Result_3 := Fld_7;
   --# end
   Result_Var_1 := Scope_Result_3;
else
   Result_Var_1 := False;
end if;
   If_Result_3 := Result_Var_1;
end if;
      Property_Result := If_Result_3;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Is_Iterable_Type;
--# end
--# property-start TypeDecl.iterable_comp_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6478
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Iterable_Comp_Type
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_486 with Inline_Always;
            procedure Finalizer_Scope_2255 with Inline_Always;
      Block_Var : Internal_Entity_Expr;
Block_Var_1 : Internal_Entity_Expr;
Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Entity_Name;
Match_0 : Internal_Entity_Derived_Type_Def;
Match_1 : Internal_Entity_Type_Def;
Var_Expr_2 : Internal_Entity_Base_Type_Decl;
Var_Expr_3 : Internal_Entity_Base_Type_Decl;
Sym : Symbol_Type;
Fld : Internal_Entity_Expr;
Sym_1 : Symbol_Type;
Fld_1 : Internal_Entity_Expr;
Bound_Imprecise_Fallback : Boolean;
Fld_2 : Boolean;
Fld_3 : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
Not_Val : Boolean;
Cast_Expr : Internal_Entity_Expr;
Cast_Result : Internal_Entity_Name;
Fld_4 : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_5 : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Is_Null_1 : Boolean;
Not_Val_1 : Boolean;
Cast_Expr_1 : Internal_Entity_Expr;
Cast_Result_1 : Internal_Entity_Aggregate;
Fld_6 : Internal_Entity_Assoc_List;
Fld_7 : Internal_Single_Actual_Array_Access;
Fld_8 : Bare_Base_Id;
Sym_2 : Symbol_Type;
Fld_9 : Boolean;
Find_Result : Internal_Single_Actual;
Fld_10 : Internal_Entity_Basic_Assoc;
Fld_11 : Internal_Entity_Expr;
Cast_Result_2 : Internal_Entity_Name;
Fld_12 : Internal_Entity_Basic_Decl;
Fld_13 : Internal_Entity_Base_Type_Decl;
Fld_14 : Internal_Entity_Type_Def;
Fld_15 : Internal_Entity_Base_Type_Decl;
Fld_16 : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Result_Var_1 : Internal_Entity_Base_Type_Decl;
Match_Prefix : Internal_Entity_Type_Def;
Cast_Result_3 : Internal_Entity_Derived_Type_Def;
Let_Result : Internal_Entity_Base_Type_Decl;
Scope_Result_2 : Internal_Entity_Base_Type_Decl;
Cast_Result_4 : Internal_Entity_Type_Def;
Let_Result_1 : Internal_Entity_Base_Type_Decl;
Scope_Result_3 : Internal_Entity_Base_Type_Decl;
Match_Result : Internal_Entity_Base_Type_Decl;
If_Result : Internal_Entity_Base_Type_Decl;
If_Result_1 : Internal_Entity_Base_Type_Decl;
If_Result_2 : Internal_Entity_Base_Type_Decl;
Scope_Result_4 : Internal_Entity_Base_Type_Decl;
Fld_17 : Internal_Entity_Base_Type_Decl;
Fld_18 : Internal_Entity_Base_Type_Decl;
Scope_Result_5 : Internal_Entity_Base_Type_Decl;
Result_Var_2 : Internal_Entity_Base_Type_Decl;
Result_Var_3 : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result_1 : Internal_Entity_Base_Type_Decl;
Let_Result_2 : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_486 is
            begin
                     Dec_Ref (Fld_7);
            end Finalizer_Scope_486;
            procedure Finalizer_Scope_2255 is
            begin
                     Dec_Ref (Fld_4);
                     Dec_Ref (Bound_Env);
            end Finalizer_Scope_2255;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3789 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 3756 '<FieldAccess for get_aspect_spec_expr at ast.py:6479>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6479
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3755 '<Symbol Iterator_Element at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Iterator_Element); 
--# expr-done 3755
--# property-call-start BasicDecl.get_aspect_spec_expr
Fld := Libadalang.Implementation.Basic_Decl_P_Get_Aspect_Spec_Expr (Node => Ent.Node, Name => Sym, E_Info => Ent.Info);
--# end
--# expr-done 3756
Block_Var := Fld; 
--# bind ie Block_Var
--# expr-start 3758 '<FieldAccess for get_aspect_spec_expr at ast.py:6480>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6480
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3757 '<Symbol Iterable at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Iterable); 
--# expr-done 3757
--# property-call-start BasicDecl.get_aspect_spec_expr
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Get_Aspect_Spec_Expr (Node => Ent.Node, Name => Sym_1, E_Info => Ent.Info);
--# end
--# expr-done 3758
Block_Var_1 := Fld_1; 
--# bind it Block_Var_1
--# expr-start 3788 '<bind at ast.py:6482>' Dyn_Var_Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6482
Bound_Imprecise_Fallback := False; 
--# expr-start 3785 '<Cond at ast.py:6482>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6482
--# expr-start 3759 '<FieldAccess for is_array at ast.py:6483>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6483
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_array
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Is_Array (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3759
if Fld_2 then
   --# expr-start 3760 '<FieldAccess for comp_type at ast.py:6483>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6483
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.comp_type
Fld_3 := Libadalang.Implementation.Base_Type_Decl_P_Comp_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3760
   If_Result_2 := Fld_3;
else
--# expr-start 3762 '<Not at ast.py:6485>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6485
--# expr-start 3761 '<IsNull at ast.py:6485>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6485
Is_Null := Block_Var.Node = null; 
--# expr-done 3761
Not_Val := not (Is_Null); 
--# expr-done 3762
if Not_Val then
--# expr-start 3763 '<Cast to Name at ast.py:6485>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6485
Cast_Expr := Block_Var; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Name
   then
      Cast_Result := Create_Internal_Entity_Name
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Name;
   end if;
--# expr-done 3763
Var_Expr_1 := Cast_Result;
if Var_Expr_1 /= No_Entity_Name then
   --# scope-start
   --# bind name Var_Expr_1
--# expr-start 3766 '<bind at ast.py:6486>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6486
--# expr-start 3764 '<FieldAccess for node_env at ast.py:6486>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6486
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Node_Env (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# expr-done 3764
Bound_Env := Fld_4; Inc_Ref (Bound_Env);
--# expr-start 3765 '<FieldAccess for designated_type_impl at ast.py:6486>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6486
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.designated_type_impl'
Fld_5 := Libadalang.Implementation.Dispatcher_Name_P_Designated_Type_Impl (Node => Var_Expr_1.Node, Env => Bound_Env, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 3765
Dyn_Var_Bind_Result := Fld_5; 
--# expr-done 3766
Scope_Result := Dyn_Var_Bind_Result;
   --# end
      Finalizer_Scope_2255;
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
   If_Result_1 := Result_Var;
else
--# expr-start 3768 '<Not at ast.py:6489>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6489
--# expr-start 3767 '<IsNull at ast.py:6489>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6489
Is_Null_1 := Block_Var_1.Node = null; 
--# expr-done 3767
Not_Val_1 := not (Is_Null_1); 
--# expr-done 3768
if Not_Val_1 then
   --# expr-start 3780 '<FieldAccess for expr_type at ast.py:6490>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6490
--# expr-start 3779 '<FieldAccess for referenced_decl at ast.py:6490>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6490
--# expr-start 3778 '<Cast to Name at ast.py:6490>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6490
--# expr-start 3777 '<FieldAccess for expr at ast.py:6490>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6490
--# expr-start 3776 '<FieldAccess for assoc at ast.py:6490>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6490
--# expr-start 3775 '<Find at ast.py:6490>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6490
   --# expr-start 3771 '<FieldAccess for unpacked_params at ast.py:6490>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6490
--# expr-start 3770 '<FieldAccess for assocs at ast.py:6490>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6490
--# expr-start 3769 '<Cast to Aggregate at ast.py:6490>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6490
Cast_Expr_1 := Block_Var_1; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Aggregate_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Aggregate
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_1 := No_Entity_Aggregate;
   end if;
--# expr-done 3769
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Assoc_List (Node => Cast_Result_1.Node.Base_Aggregate_F_Assocs, Info => Cast_Result_1.Info);
--# expr-done 3770
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AssocList.unpacked_params
Fld_7 := Libadalang.Implementation.Assoc_List_P_Unpacked_Params (Node => Fld_6.Node, E_Info => Fld_6.Info);
--# end
--# expr-done 3771
      Find_Result := No_Single_Actual;
      declare
         Collection : constant Internal_Single_Actual_Array_Access := Fld_7;
      begin
         for Item_146 of
               Collection.Items
         loop
   --# scope-start
               --# bind sa Item_146
      --# expr-start 3774 '<FieldAccess for name_is at ast.py:6491>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6491
--# expr-start 3772 '<FieldAccess for name at ast.py:6491>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6491
Fld_8 := Item_146.Name;
--# expr-done 3772
      if Fld_8 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3773 '<Symbol Element at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Element); 
--# expr-done 3773
--# property-call-start Name.name_is
Fld_9 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_8, Sym => Sym_2);
--# end
--# expr-done 3774
      if Fld_9 then
         Find_Result := Item_146;
      end if;
   --# end
               exit when Fld_9;
         end loop;
      end;
--# expr-done 3775
Fld_10 := Find_Result.Assoc;
--# expr-done 3776
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_11 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_10.Node, E_Info => Fld_10.Info);
--# end
--# expr-done 3777
   if Fld_11.Node = null
      or else Fld_11.Node.Kind in Ada_Name
   then
      Cast_Result_2 := Create_Internal_Entity_Name
        (Node => Fld_11.Node,
         Info => Fld_11.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 3778
      if Cast_Result_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld_12 := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Cast_Result_2.Node, Imprecise_Fallback => Bound_Imprecise_Fallback, E_Info => Cast_Result_2.Info);
--# end
--# expr-done 3779
      if Fld_12.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.expr_type'
Fld_13 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Expr_Type (Node => Fld_12.Node, Origin => Origin, E_Info => Fld_12.Info);
--# end
--# expr-done 3780
   If_Result := Fld_13;
else
   --# expr-start 3784 '<Match at ast.py:6495>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6495
--# expr-start 3781 '<FieldAccess for type_def at ast.py:6495>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6495
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3781
      if Fld_14.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld_14; 
case Ada_Type_Def (Match_Prefix.Node.Kind) is
         when Ada_Derived_Type_Def =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Derived_Type_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_3; 
--# bind dtd Match_0
--# expr-start 3782 '<FieldAccess for base_type at ast.py:6497>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6497
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld_15 := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => Match_0.Node, Origin => Origin, E_Info => Match_0.Info);
--# end
--# expr-done 3782
Var_Expr_2 := Fld_15;
if Var_Expr_2 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind bt Var_Expr_2
--# expr-start 3783 '<FieldAccess for iterable_comp_type at ast.py:6497>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6497
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.iterable_comp_type'
Fld_16 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Iterable_Comp_Type (Node => Var_Expr_2.Node, Origin => Origin, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 3783
Scope_Result_1 := Fld_16;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Base_Type_Decl;
end if;
Let_Result := Result_Var_1; 
--# end
Scope_Result_2 := Let_Result;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Access_To_Subp_Def .. Ada_Array_Type_Def | Ada_Enum_Type_Def .. Ada_Signed_Int_Type_Def =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Type_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_4; 
--# bind _ Match_1
Let_Result_1 := No_Entity_Base_Type_Decl; 
--# end
Scope_Result_3 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_3; 
end case;
--# expr-done 3784
   If_Result := Match_Result;
end if;
   If_Result_1 := If_Result;
end if;
   If_Result_2 := If_Result_1;
end if;
--# expr-done 3785
Var_Expr := If_Result_2;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind e Var_Expr
Scope_Result_4 := Var_Expr;
   --# end
   Result_Var_3 := Scope_Result_4;
else
--# expr-start 3786 '<FieldAccess for previous_part at ast.py:6500>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6500
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.previous_part'
Fld_17 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Previous_Part (Node => Ent.Node, Go_To_Incomplete => False, E_Info => Ent.Info);
--# end
--# expr-done 3786
Var_Expr_3 := Fld_17;
if Var_Expr_3 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind pp Var_Expr_3
--# expr-start 3787 '<FieldAccess for iterable_comp_type at ast.py:6501>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6501
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.iterable_comp_type'
Fld_18 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Iterable_Comp_Type (Node => Var_Expr_3.Node, Origin => Origin, E_Info => Var_Expr_3.Info);
--# end
--# expr-done 3787
Scope_Result_5 := Fld_18;
   --# end
   Result_Var_2 := Scope_Result_5;
else
   Result_Var_2 := No_Entity_Base_Type_Decl;
end if;
   Result_Var_3 := Result_Var_2;
end if;
Dyn_Var_Bind_Result_1 := Result_Var_3; 
--# expr-done 3788
Let_Result_2 := Dyn_Var_Bind_Result_1; 
--# end
--# expr-done 3789
      Property_Result := Let_Result_2;
   --# end
      Finalizer_Scope_486;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_486;
               Finalizer_Scope_2255;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Iterable_Comp_Type;
--# end
--# property-start TypeDecl.discrete_range /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6504
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Discrete_Range
  (Node : Bare_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Discrete_Range
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Discrete_Range;
      Fld : Internal_Entity_Type_Def;
Fld_1 : Internal_Discrete_Range;
Let_Result : Internal_Discrete_Range;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3792 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3791 '<FieldAccess for discrete_range at ast.py:6505>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6505
--# expr-start 3790 '<FieldAccess for type_def at ast.py:6505>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6505
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3790
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.discrete_range'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Discrete_Range (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 3791
Let_Result := Fld_1; 
--# end
--# expr-done 3792
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Discrete_Range;
--# end
--# property-start TypeDecl.discriminants_list /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6508
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Discriminants_List
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_488 with Inline_Always;
            procedure Finalizer_Scope_2262 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Block_Var_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Var_Expr : Internal_Entity_Discriminant_Part;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Discriminant_Part;
Fld_2 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Scope_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Result_Var : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_3 : Boolean;
Fld_4 : Internal_Entity_Base_Type_Decl;
Fld_5 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Len : Integer;
Comp_Result : Boolean;
Is_Null : Boolean;
Not_Val : Boolean;
Fld_6 : Internal_Entity_Base_Type_Decl;
Fld_7 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
If_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
If_Result_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
If_Result_2 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Let_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_488 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Fld_5);
                     Dec_Ref (Fld_7);
                     Dec_Ref (If_Result);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (If_Result_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_488;
            procedure Finalizer_Scope_2262 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_2262;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3805 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3793 '<FieldAccess for base_type at ast.py:6510>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6510
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.base_type'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Base_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3793
Block_Var := Fld; 
--# bind base_type Block_Var
--# expr-start 3794 '<FieldAccess for discriminants at ast.py:6511>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6511
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Discriminant_Part (Node => Ent.Node.Type_Decl_F_Discriminants, Info => Ent.Info);
--# expr-done 3794
Var_Expr := Fld_1;
if Var_Expr /= No_Entity_Discriminant_Part then
   --# scope-start
   --# bind d Var_Expr
--# expr-start 3795 '<FieldAccess for abstract_formal_params at ast.py:6512>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6512
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseFormalParamHolder.abstract_formal_params'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Formal_Param_Holder_P_Abstract_Formal_Params (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 3795
Scope_Result := Fld_2;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2262;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Base_Formal_Param_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var);
Block_Var_1 := Result_Var; Inc_Ref (Block_Var_1);
--# bind self_discs Block_Var_1
--# expr-start 3804 '<Cond at ast.py:6515>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6515
--# expr-start 3796 '<FieldAccess for is_access_type at ast.py:6516>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6516
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_access_type'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Access_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3796
if Fld_3 then
   --# expr-start 3798 '<FieldAccess for discriminants_list at ast.py:6517>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6517
--# expr-start 3797 '<FieldAccess for accessed_type at ast.py:6517>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6517
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_4 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3797
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.discriminants_list'
Fld_5 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Discriminants_List (Node => Fld_4.Node, Origin => Origin, E_Info => Fld_4.Info);
--# end
--# expr-done 3798
   If_Result_2 := Fld_5;
else
--# expr-start 3799 '<length at ast.py:6519>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6519
Len := Length (Block_Var_1); 
--# expr-done 3799
Comp_Result := Len > 0; 
if Comp_Result then
   If_Result_1 := Block_Var_1;
else
--# expr-start 3801 '<Not at ast.py:6520>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6520
--# expr-start 3800 '<IsNull at ast.py:6520>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6520
Is_Null := Block_Var.Node = null; 
--# expr-done 3800
Not_Val := not (Is_Null); 
--# expr-done 3801
if Not_Val then
   --# expr-start 3803 '<FieldAccess for discriminants_list at ast.py:6520>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6520
--# expr-start 3802 '<FieldAccess for base_type at ast.py:6520>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6520
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.base_type'
Fld_6 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Base_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3802
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.discriminants_list'
Fld_7 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Discriminants_List (Node => Fld_6.Node, Origin => Origin, E_Info => Fld_6.Info);
--# end
--# expr-done 3803
   If_Result := Fld_7;
else
   If_Result := No_Internal_Entity_Base_Formal_Param_Decl_Array_Type;
end if;
      Inc_Ref (If_Result);
   If_Result_1 := If_Result;
end if;
      Inc_Ref (If_Result_1);
   If_Result_2 := If_Result_1;
end if;
      Inc_Ref (If_Result_2);
--# expr-done 3804
Let_Result := If_Result_2; Inc_Ref (Let_Result);
--# end
--# expr-done 3805
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_488;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_488;
               Finalizer_Scope_2262;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Discriminants_List;
--# end
--# property-start TypeDecl.direct_primitives_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6525
pragma Warnings (Off, "is not referenced");
 function Type_Decl_F_Direct_Primitives_Env
  (Node : Bare_Type_Decl
  )
   return Lexical_Env
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_489 with Inline_Always;
      Dyn_Env : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_489 is
            begin
                     Dec_Ref (Dyn_Env);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_489;
begin
   --# property-body-start
      if Self.Lf_Present_Direct_Primitives_Env then
         Property_Result := Self.Lf_Stg_Direct_Primitives_Env;
            Inc_Ref (Property_Result);
         return Property_Result;
      end if;
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3807 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3806 '<DynamicLexicalEnv at ast.py:6530>' Dyn_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6530
Dyn_Env := Create_Dynamic_Lexical_Env (Self, Type_Decl_F_Direct_Primitive_Subps'Access, null, False); 
--# expr-done 3806
Let_Result := Dyn_Env; Inc_Ref (Let_Result);
--# end
--# expr-done 3807
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_489;
      Self.Lf_Present_Direct_Primitives_Env := True;
      Self.Lf_Stg_Direct_Primitives_Env := Property_Result;
         Inc_Ref (Property_Result);
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_489;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_F_Direct_Primitives_Env;
--# end
--# property-start TypeDecl.direct_primitive_subps /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6536
pragma Warnings (Off, "is not referenced");
 function Type_Decl_F_Direct_Primitive_Subps
  (Node : Bare_Type_Decl
  )
   return Internal_Inner_Env_Assoc_Array_Access
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Inner_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_490 with Inline_Always;
            procedure Finalizer_Scope_2264 with Inline_Always;
            procedure Finalizer_Scope_2265 with Inline_Always;
            procedure Finalizer_Scope_2269 with Inline_Always;
            procedure Finalizer_Scope_2270 with Inline_Always;
            procedure Finalizer_Scope_2272 with Inline_Always;
            procedure Finalizer_Scope_2276 with Inline_Always;
            procedure Finalizer_Scope_2279 with Inline_Always;
      Block_Var : Bare_Declarative_Part;
Block_Var_1 : Boolean;
Block_Var_2 : Internal_Entity_Array_Access;
Block_Var_3 : Internal_Inner_Env_Assoc_Array_Access;
Block_Var_4 : Internal_Inner_Env_Assoc_Array_Access;
Block_Var_5 : Internal_Inner_Env_Assoc_Array_Access;
Var_Expr : Internal_Entity;
Var_Expr_1 : Bare_Declarative_Part;
Match_0 : Internal_Entity_Base_Package_Decl;
Match_1 : Internal_Entity;
Var_Expr_2 : Internal_Entity_Private_Part;
Var_Expr_3 : Bare_Declarative_Part;
Var_Expr_4 : Bare_Enum_Type_Def;
Var_Expr_5 : Internal_Entity_Base_Subp_Spec;
Var_Expr_6 : Internal_Entity_Basic_Decl;
Bd : Internal_Entity_Basic_Decl;
Fld : Bare_Declarative_Part;
As_Entity : Internal_Entity_Type_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Type_Decl;
Fld_1 : Boolean;
Fld_2 : Bare_Ada_Node;
Scope_Result : Bare_Ada_Node;
Result_Var : Bare_Ada_Node;
As_Entity_1 : Internal_Entity;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity;
Fld_3 : Internal_Entity_Public_Part;
Fld_4 : Internal_Entity_Ada_Node_List;
Fld_5 : Bare_Ada_Node_List;
Fld_6 : Internal_Entity_Info;
As_Entity_2 : Internal_Entity;
Is_Null_2 : Boolean;
If_Result_2 : Internal_Entity;
Item_358 : Internal_Entity;
Bare_Item_358 : Bare_Ada_Node;
Map_Result : Internal_Entity_Array_Access;
Fld_7 : Internal_Entity_Private_Part;
Fld_8 : Internal_Entity_Ada_Node_List;
Scope_Result_1 : Internal_Entity_Ada_Node_List;
Result_Var_1 : Internal_Entity_Ada_Node_List;
Fld_9 : Bare_Ada_Node_List;
Fld_10 : Internal_Entity_Info;
As_Entity_3 : Internal_Entity;
Is_Null_3 : Boolean;
If_Result_3 : Internal_Entity;
Item_359 : Internal_Entity;
Bare_Item_359 : Bare_Ada_Node;
Map_Result_1 : Internal_Entity_Array_Access;
Concat_Result : Internal_Entity_Array_Access;
Fld_11 : Bare_Ada_Node;
Fld_12 : Bare_Ada_Node;
Cast_Result : Bare_Declarative_Part;
As_Entity_4 : Internal_Entity_Declarative_Part;
Is_Null_4 : Boolean;
If_Result_4 : Internal_Entity_Declarative_Part;
Fld_13 : Internal_Entity_Ada_Node_List;
Fld_14 : Bare_Ada_Node_List;
Fld_15 : Internal_Entity_Info;
As_Entity_5 : Internal_Entity;
Is_Null_5 : Boolean;
If_Result_5 : Internal_Entity;
Item_360 : Internal_Entity;
Bare_Item_360 : Bare_Ada_Node;
Map_Result_2 : Internal_Entity_Array_Access;
Scope_Result_2 : Internal_Entity_Array_Access;
Result_Var_2 : Internal_Entity_Array_Access;
If_Result_6 : Internal_Entity_Array_Access;
Match_Prefix : Internal_Entity;
Cast_Result_1 : Internal_Entity_Base_Package_Decl;
Let_Result : Internal_Entity_Array_Access;
Scope_Result_3 : Internal_Entity_Array_Access;
Cast_Result_2 : Internal_Entity;
Let_Result_1 : Internal_Entity_Array_Access;
Scope_Result_4 : Internal_Entity_Array_Access;
Match_Result : Internal_Entity_Array_Access;
Scope_Result_5 : Internal_Entity_Array_Access;
Result_Var_3 : Internal_Entity_Array_Access;
Fld_16 : Bare_Type_Def;
Cast_Result_3 : Bare_Enum_Type_Def;
Fld_17 : Bare_Enum_Literal_Decl_List;
Fld_18 : Bare_Defining_Name;
Fld_19 : Symbol_Type;
Cast_Expr : Bare_Type_Decl;
Cast_Result_4 : Bare_Ada_Node;
New_Struct : Internal_Metadata;
Cast_Expr_1 : Bare_Enum_Literal_Decl;
Cast_Result_5 : Bare_Ada_Node;
New_Struct_1 : Internal_Inner_Env_Assoc;
Item_147 : Bare_Enum_Literal_Decl;
Map_Result_3 : Internal_Inner_Env_Assoc_Array_Access;
Scope_Result_6 : Internal_Inner_Env_Assoc_Array_Access;
Result_Var_4 : Internal_Inner_Env_Assoc_Array_Access;
Cast_Expr_2 : Internal_Entity;
Cast_Result_6 : Internal_Entity_Basic_Decl;
Fld_20 : Internal_Entity_Base_Subp_Spec;
Scope_Result_7 : Internal_Entity_Base_Subp_Spec;
Result_Var_5 : Internal_Entity_Base_Subp_Spec;
Fld_21 : Internal_Entity_Base_Type_Decl_Array_Access;
As_Entity_6 : Internal_Entity_Type_Decl;
Is_Null_6 : Boolean;
If_Result_7 : Internal_Entity_Type_Decl;
Cast_Expr_3 : Internal_Entity_Base_Type_Decl;
Cast_Result_7 : Internal_Entity;
Cast_Result_8 : Internal_Entity;
Is_Equiv : Boolean;
Quantifier_Result : Boolean;
As_Entity_7 : Internal_Entity_Type_Decl;
Is_Null_7 : Boolean;
If_Result_8 : Internal_Entity_Type_Decl;
Fld_22 : Lexical_Env := Empty_Env;
Fld_23 : Internal_Entity_Defining_Name;
Fld_24 : Symbol_Type;
Env_Get_Result : Internal_Entity_Array_Access;
Cast_Expr_4 : Internal_Entity;
Cast_Result_9 : Internal_Entity_Basic_Decl;
Fld_25 : Internal_Entity_Base_Subp_Spec;
Fld_26 : Boolean;
Quantifier_Result_1 : Boolean;
Fld_27 : Bare_Ada_Node;
Is_A : Boolean;
Not_Val : Boolean;
If_Result_9 : Boolean;
If_Result_10 : Boolean;
Scope_Result_8 : Boolean;
Result_Var_6 : Boolean;
Map_Result_4 : Internal_Entity_Array_Access;
Cast_Expr_5 : Internal_Entity;
Cast_Result_10 : Internal_Entity_Basic_Decl;
Fld_28 : Internal_Entity_Defining_Name;
Fld_29 : Symbol_Type;
Fld_30 : Bare_Basic_Decl;
Cast_Expr_6 : Bare_Type_Decl;
Cast_Result_11 : Bare_Ada_Node;
New_Struct_2 : Internal_Metadata;
Cast_Result_12 : Bare_Ada_Node;
New_Struct_3 : Internal_Inner_Env_Assoc;
Singleton : Internal_Inner_Env_Assoc_Array_Access;
Sym : Symbol_Type;
Fld_31 : Bare_Basic_Decl;
Cast_Expr_7 : Bare_Type_Decl;
Cast_Result_13 : Bare_Ada_Node;
New_Struct_4 : Internal_Metadata;
Cast_Result_14 : Bare_Ada_Node;
New_Struct_5 : Internal_Inner_Env_Assoc;
Singleton_1 : Internal_Inner_Env_Assoc_Array_Access;
Fld_32 : Internal_Entity_Defining_Name;
Sym_1 : Symbol_Type;
Fld_33 : Boolean;
If_Result_11 : Internal_Inner_Env_Assoc_Array_Access;
Concat_Result_1 : Internal_Inner_Env_Assoc_Array_Access;
Let_Result_2 : Internal_Inner_Env_Assoc_Array_Access;
Map_Result_5 : Internal_Inner_Env_Assoc_Array_Access;
Fld_34 : Internal_Env_Assoc_Array_Access;
Fld_35 : Symbol_Type;
Fld_36 : Bare_Ada_Node;
Cast_Expr_8 : Bare_Type_Decl;
Cast_Result_15 : Bare_Ada_Node;
New_Struct_6 : Internal_Metadata;
New_Struct_7 : Internal_Inner_Env_Assoc;
Map_Result_6 : Internal_Inner_Env_Assoc_Array_Access;
Concat_Result_2 : Internal_Inner_Env_Assoc_Array_Access;
Concat_Result_3 : Internal_Inner_Env_Assoc_Array_Access;
Let_Result_3 : Internal_Inner_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_490 is
            begin
                     Dec_Ref (Scope_Result_5);
                     Dec_Ref (Result_Var_3);
                     Dec_Ref (Block_Var_2);
                     Dec_Ref (Scope_Result_6);
                     Dec_Ref (Result_Var_4);
                     Dec_Ref (Block_Var_3);
                     Dec_Ref (Map_Result_4);
                     Dec_Ref (Map_Result_5);
                     Dec_Ref (Block_Var_4);
                     Dec_Ref (Fld_34);
                     Dec_Ref (Map_Result_6);
                     Dec_Ref (Block_Var_5);
                     Dec_Ref (Concat_Result_2);
                     Dec_Ref (Concat_Result_3);
                     Dec_Ref (Let_Result_3);
            end Finalizer_Scope_490;
            procedure Finalizer_Scope_2264 is
            begin
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_4);
                     Dec_Ref (Match_Result);
            end Finalizer_Scope_2264;
            procedure Finalizer_Scope_2265 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Concat_Result);
            end Finalizer_Scope_2265;
            procedure Finalizer_Scope_2269 is
            begin
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Result_Var_2);
                     Dec_Ref (If_Result_6);
            end Finalizer_Scope_2269;
            procedure Finalizer_Scope_2270 is
            begin
                     Dec_Ref (Map_Result_2);
            end Finalizer_Scope_2270;
            procedure Finalizer_Scope_2272 is
            begin
                     Dec_Ref (Map_Result_3);
            end Finalizer_Scope_2272;
            procedure Finalizer_Scope_2276 is
            begin
                     Dec_Ref (Fld_21);
                     Dec_Ref (Fld_22);
                     Dec_Ref (Env_Get_Result);
            end Finalizer_Scope_2276;
            procedure Finalizer_Scope_2279 is
            begin
                     Dec_Ref (Singleton);
                     Dec_Ref (Singleton_1);
                     Dec_Ref (If_Result_11);
                     Dec_Ref (Concat_Result_1);
                     Dec_Ref (Let_Result_2);
            end Finalizer_Scope_2279;
begin
   --# property-body-start
      if Self.Lf_Present_Direct_Primitive_Subps then
         Property_Result := Self.Lf_Stg_Direct_Primitive_Subps;
            Inc_Ref (Property_Result);
         return Property_Result;
      end if;
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3881 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 3808 '<FieldAccess for declarative_scope at ast.py:6542>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6542
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.declarative_scope
Fld := Libadalang.Implementation.Ada_Node_P_Declarative_Scope (Node => Self);
--# end
--# expr-done 3808
Block_Var := Fld; 
--# bind scope Block_Var
--# expr-start 3810 '<FieldAccess for is_derived_tagged_type at ast.py:6543>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6543
--# expr-start 3809 '<as_bare_entity at ast.py:6543>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6543
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity_Type_Decl;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 3809
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDecl.is_derived_tagged_type
Fld_1 := Libadalang.Implementation.Type_Decl_P_Is_Derived_Tagged_Type (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 3810
Block_Var_1 := Fld_1; 
--# bind is_derived_tagged Block_Var_1
--# expr-start 3812 '<as_bare_entity at ast.py:6544>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6544
Var_Expr_1 := Block_Var;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 3811 '<parent at ast.py:6544>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6544
      if Var_Expr_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Var_Expr_1.Parent; 
--# expr-done 3811
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node;
end if;
Is_Null_1 := Result_Var = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity;
else
As_Entity_1 := (Info => No_Entity_Info, Node => Result_Var); 
   If_Result_1 := As_Entity_1;
end if;
--# expr-done 3812
Var_Expr := If_Result_1;
if Var_Expr /= No_Entity then
   --# scope-start
--# expr-start 3827 '<Match at ast.py:6544>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6544
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Var_Expr; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Generic_Package_Internal .. Ada_Package_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Base_Package_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_1; 
--# bind pkg_decl Match_0
--# expr-start 3819 '<Concat at ast.py:6548>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6548
--# expr-start 3815 '<TakeWhile at ???>' Map_Result None
   --# expr-start 3814 '<FieldAccess for decls at ast.py:6548>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6548
--# expr-start 3813 '<FieldAccess for public_part at ast.py:6548>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6548
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Public_Part (Node => Match_0.Node.Base_Package_Decl_F_Public_Part, Info => Match_0.Info);
--# expr-done 3813
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Ada_Node_List (Node => Fld_3.Node.Declarative_Part_F_Decls, Info => Fld_3.Info);
--# expr-done 3814
Fld_5 := Fld_4.Node;
      declare
         Map_Result_Vec : Internal_Entity_Vectors.Vector;
      begin
      if Fld_5 = null then
      Map_Result := Create_Internal_Entity_Array (0);
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_5;
      begin
         for Untyped_Item_358 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_358 := Untyped_Item_358; 
Is_Null_2 := Bare_Item_358 = null; 
if Is_Null_2 then
   If_Result_2 := No_Entity;
else
Fld_6 := Fld_4.Info;
As_Entity_2 := (Info => Fld_6, Node => Bare_Item_358); 
   If_Result_2 := As_Entity_2;
end if;
                  Item_358 := If_Result_2; 
   --# scope-start
         declare
            Item_To_Append : constant Internal_Entity := Item_358;
         begin
            Internal_Entity_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 3815
--# expr-start 3818 '<TakeWhile at ???>' Map_Result_1 None
--# expr-start 3816 '<FieldAccess for private_part at ast.py:6549>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6549
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Private_Part (Node => Match_0.Node.Base_Package_Decl_F_Private_Part, Info => Match_0.Info);
--# expr-done 3816
Var_Expr_2 := Fld_7;
if Var_Expr_2 /= No_Entity_Private_Part then
   --# scope-start
--# expr-start 3817 '<FieldAccess for decls at ast.py:6549>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6549
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Ada_Node_List (Node => Var_Expr_2.Node.Declarative_Part_F_Decls, Info => Var_Expr_2.Info);
--# expr-done 3817
Scope_Result_1 := Fld_8;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Ada_Node_List;
end if;
Fld_9 := Result_Var_1.Node;
      declare
         Map_Result_1_Vec : Internal_Entity_Vectors.Vector;
      begin
      if Fld_9 = null then
      Map_Result_1 := Create_Internal_Entity_Array (0);
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_9;
      begin
         for Untyped_Item_359 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_359 := Untyped_Item_359; 
Is_Null_3 := Bare_Item_359 = null; 
if Is_Null_3 then
   If_Result_3 := No_Entity;
else
Fld_10 := Result_Var_1.Info;
As_Entity_3 := (Info => Fld_10, Node => Bare_Item_359); 
   If_Result_3 := As_Entity_3;
end if;
                  Item_359 := If_Result_3; 
   --# scope-start
         declare
            Item_To_Append : constant Internal_Entity := Item_359;
         begin
            Internal_Entity_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result_1 := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 3818
Concat_Result := Concat (Map_Result, Map_Result_1); 
--# expr-done 3819
Let_Result := Concat_Result; Inc_Ref (Let_Result);
--# end
Scope_Result_3 := Let_Result;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_2265;
            Match_Result := Scope_Result_3; Inc_Ref (Match_Result);
         when Ada_Abort_Absent .. Ada_Synthetic_Formal_Param_Decl | Ada_Discrete_Base_Subtype_Decl .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_2; 
--# bind _ Match_1
--# expr-start 3826 '<If at ast.py:6555>' If_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6555
if Block_Var_1 then
--# expr-start 3822 '<Cast to DeclarativePart at ast.py:6557>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6557
--# expr-start 3821 '<parent at ast.py:6557>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6557
--# expr-start 3820 '<parent at ast.py:6557>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6557
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Self.Parent; 
--# expr-done 3820
      if Fld_11 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Fld_11.Parent; 
--# expr-done 3821
   if Fld_12 = null
      or else Fld_12.Kind in Ada_Declarative_Part_Range
   then
      Cast_Result := Fld_12;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 3822
Var_Expr_3 := Cast_Result;
if Var_Expr_3 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind dp Var_Expr_3
--# expr-start 3825 '<TakeWhile at ???>' Map_Result_2 None
   --# expr-start 3824 '<FieldAccess for decls at ast.py:6558>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6558
--# expr-start 3823 '<as_bare_entity at ast.py:6558>' If_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6558
Is_Null_4 := Var_Expr_3 = null; 
if Is_Null_4 then
   If_Result_4 := No_Entity_Declarative_Part;
else
As_Entity_4 := (Info => No_Entity_Info, Node => Var_Expr_3); 
   If_Result_4 := As_Entity_4;
end if;
--# expr-done 3823
      if If_Result_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Create_Internal_Entity_Ada_Node_List (Node => If_Result_4.Node.Declarative_Part_F_Decls, Info => If_Result_4.Info);
--# expr-done 3824
Fld_14 := Fld_13.Node;
      declare
         Map_Result_2_Vec : Internal_Entity_Vectors.Vector;
      begin
      if Fld_14 = null then
      Map_Result_2 := Create_Internal_Entity_Array (0);
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_14;
      begin
         for Untyped_Item_360 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_360 := Untyped_Item_360; 
Is_Null_5 := Bare_Item_360 = null; 
if Is_Null_5 then
   If_Result_5 := No_Entity;
else
Fld_15 := Fld_13.Info;
As_Entity_5 := (Info => Fld_15, Node => Bare_Item_360); 
   If_Result_5 := As_Entity_5;
end if;
                  Item_360 := If_Result_5; 
   --# scope-start
         declare
            Item_To_Append : constant Internal_Entity := Item_360;
         begin
            Internal_Entity_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result_2 := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_2_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 3825
Scope_Result_2 := Map_Result_2;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2270;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Internal_Entity_Array_Type;
end if;
      Inc_Ref (Result_Var_2);
   If_Result_6 := Result_Var_2;
else
   If_Result_6 := No_Internal_Entity_Array_Type;
end if;
      Inc_Ref (If_Result_6);
--# expr-done 3826
Let_Result_1 := If_Result_6; Inc_Ref (Let_Result_1);
--# end
Scope_Result_4 := Let_Result_1;
   Inc_Ref (Scope_Result_4);
   --# end
      Finalizer_Scope_2269;
            Match_Result := Scope_Result_4; Inc_Ref (Match_Result);
end case;
--# expr-done 3827
Scope_Result_5 := Match_Result;
   Inc_Ref (Scope_Result_5);
   --# end
      Finalizer_Scope_2264;
   Result_Var_3 := Scope_Result_5;
else
   Result_Var_3 := No_Internal_Entity_Array_Type;
end if;
      Inc_Ref (Result_Var_3);
Block_Var_2 := Result_Var_3; Inc_Ref (Block_Var_2);
--# bind decl_parts Block_Var_2
--# expr-start 3829 '<Cast to EnumTypeDef at ast.py:6563>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6563
--# expr-start 3828 '<FieldAccess for type_def at ast.py:6563>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6563
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_16 := Self.Type_Decl_F_Type_Def;
--# expr-done 3828
   if Fld_16 = null
      or else Fld_16.Kind in Ada_Enum_Type_Def_Range
   then
      Cast_Result_3 := Fld_16;
   else
         Cast_Result_3 := No_Bare_Ada_Node;
   end if;
--# expr-done 3829
Var_Expr_4 := Cast_Result_3;
if Var_Expr_4 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind etf Var_Expr_4
--# expr-start 3835 '<Map at ast.py:6564>' Map_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6564
   --# expr-start 3830 '<FieldAccess for enum_literals at ast.py:6564>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6564
      if Var_Expr_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_17 := Var_Expr_4.Enum_Type_Def_F_Enum_Literals;
--# expr-done 3830
      declare
         Map_Result_3_Vec : Internal_Inner_Env_Assoc_Vectors.Vector;
      begin
      if Fld_17 = null then
      Map_Result_3 := Create_Internal_Inner_Env_Assoc_Array (0);
      else
      declare
         Collection : constant Bare_Enum_Literal_Decl_List := Fld_17;
      begin
         for Untyped_Item_147 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_147 := Untyped_Item_147; 
   --# scope-start
               --# bind lit Item_147
      --# expr-start 3834 '<New InnerEnvAssoc at ast.py:6565>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6565
--# expr-start 3832 '<FieldAccess for name_symbol at ast.py:6566>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6566
--# expr-start 3831 '<FieldAccess for name at ast.py:6566>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6566
      if Item_147 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Item_147.Enum_Literal_Decl_F_Name;
--# expr-done 3831
      if Fld_18 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_19 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_18);
--# end
--# expr-done 3832
--# expr-start 3833 '<New Metadata at ast.py:6568>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6568
Cast_Expr := Self; 
      Cast_Result_4 := Cast_Expr;
New_Struct := (Dottable_Subp => False, Primitive => Cast_Result_4, Primitive_Real_Type => No_Bare_Ada_Node); 
--# expr-done 3833
Cast_Expr_1 := Item_147; 
      Cast_Result_5 := Cast_Expr_1;
New_Struct_1 := (Key => Fld_19, Metadata => New_Struct, Value => Cast_Result_5); 
--# expr-done 3834
         declare
            Item_To_Append : constant Internal_Inner_Env_Assoc := New_Struct_1;
         begin
            Internal_Inner_Env_Assoc_Vectors.Append (Map_Result_3_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result_3 := Create_Internal_Inner_Env_Assoc_Array
           (Items_Count => Natural (Internal_Inner_Env_Assoc_Vectors.Length (Map_Result_3_Vec)));
         for I in Map_Result_3.Items'Range loop
            Map_Result_3.Items (I) := Internal_Inner_Env_Assoc_Vectors.Get
              (Map_Result_3_Vec,
               I + Internal_Inner_Env_Assoc_Vectors.Index_Type'First - Map_Result_3.Items'First);
         end loop;
         Internal_Inner_Env_Assoc_Vectors.Destroy (Map_Result_3_Vec);
      end;
--# expr-done 3835
Scope_Result_6 := Map_Result_3;
   Inc_Ref (Scope_Result_6);
   --# end
      Finalizer_Scope_2272;
   Result_Var_4 := Scope_Result_6;
else
   Result_Var_4 := No_Internal_Inner_Env_Assoc_Array_Type;
end if;
      Inc_Ref (Result_Var_4);
Block_Var_3 := Result_Var_4; Inc_Ref (Block_Var_3);
--# bind enum_lits Block_Var_3
--# expr-start 3872 '<Mapcat at ast.py:6572>' Map_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6572
   --# expr-start 3855 '<Filter at ast.py:6572>' Map_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6572
      declare
         Map_Result_4_Vec : Internal_Entity_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Block_Var_2;
      begin
         for Item_149 of
               Collection.Items
         loop
   --# scope-start
               --# bind decl Item_149
--# expr-start 3836 '<Cast to BasicDecl at ast.py:6573>' Cast_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6573
Cast_Expr_2 := Item_149; 
   if Cast_Expr_2.Node = null
      or else Cast_Expr_2.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_6 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
   else
         Cast_Result_6 := No_Entity_Basic_Decl;
   end if;
--# expr-done 3836
Var_Expr_6 := Cast_Result_6;
if Var_Expr_6 /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 3837 '<FieldAccess for subp_spec_or_null at ast.py:6573>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6573
      if Var_Expr_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_20 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Var_Expr_6.Node, E_Info => Var_Expr_6.Info);
--# end
--# expr-done 3837
Scope_Result_7 := Fld_20;
   --# end
   Result_Var_5 := Scope_Result_7;
else
   Result_Var_5 := No_Entity_Base_Subp_Spec;
end if;
Var_Expr_5 := Result_Var_5;
if Var_Expr_5 /= No_Entity_Base_Subp_Spec then
   --# scope-start
   --# bind spec Var_Expr_5
--# expr-start 3841 '<Contains at ast.py:6574>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6574
   --# expr-start 3838 '<FieldAccess for candidate_primitive_subp_types at ast.py:6574>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6574
      if Var_Expr_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.candidate_primitive_subp_types
Fld_21 := Libadalang.Implementation.Base_Subp_Spec_P_Candidate_Primitive_Subp_Types (Node => Var_Expr_5.Node, E_Info => Var_Expr_5.Info);
--# end
--# expr-done 3838
      Quantifier_Result := False;
      declare
         Collection : constant Internal_Entity_Base_Type_Decl_Array_Access := Fld_21;
      begin
         for Item_150 of
               Collection.Items
         loop
   --# scope-start
               --# bind item Item_150
      --# expr-start 3840 '<Eq at ???>' Is_Equiv None
Cast_Expr_3 := Item_150; 
      Cast_Result_7 := Create_Internal_Entity
        (Node => Cast_Expr_3.Node,
         Info => Cast_Expr_3.Info);
--# expr-start 3839 '<as_bare_entity at ast.py:6575>' If_Result_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6575
Is_Null_6 := Self = null; 
if Is_Null_6 then
   If_Result_7 := No_Entity_Type_Decl;
else
As_Entity_6 := (Info => No_Entity_Info, Node => Self); 
   If_Result_7 := As_Entity_6;
end if;
--# expr-done 3839
      Cast_Result_8 := Create_Internal_Entity
        (Node => If_Result_7.Node,
         Info => If_Result_7.Info);
Is_Equiv := Equivalent (Cast_Result_7, Cast_Result_8); 
--# expr-done 3840
      Quantifier_Result := Is_Equiv;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
--# expr-done 3841
if Quantifier_Result then
   --# expr-start 3854 '<If at ast.py:6576>' If_Result_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6576
--# expr-start 3844 '<Not at ast.py:6583>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6583
--# expr-start 3843 '<IsA BasePackageDecl at ast.py:6583>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6583
--# expr-start 3842 '<parent at ast.py:6583>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6583
      if Block_Var = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_27 := Block_Var.Parent; 
--# expr-done 3842
Is_A := Fld_27 /= null 
and then Fld_27.Kind in Ada_Base_Package_Decl; 
--# expr-done 3843
Not_Val := not (Is_A); 
--# expr-done 3844
if Not_Val then
   --# expr-start 3853 '<AnyQuantifier at ast.py:6584>' Quantifier_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6584
   --# expr-start 3849 '<EnvGet at ast.py:6584>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6584
--# expr-start 3846 '<FieldAccess for parent_primitives_env at ast.py:6584>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6584
--# expr-start 3845 '<as_bare_entity at ast.py:6584>' If_Result_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6584
Is_Null_7 := Self = null; 
if Is_Null_7 then
   If_Result_8 := No_Entity_Type_Decl;
else
As_Entity_7 := (Info => No_Entity_Info, Node => Self); 
   If_Result_8 := As_Entity_7;
end if;
--# expr-done 3845
      if If_Result_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDecl.parent_primitives_env
Fld_22 := Libadalang.Implementation.Type_Decl_P_Parent_Primitives_Env (Node => If_Result_8.Node, E_Info => If_Result_8.Info);
--# end
--# expr-done 3846
--# expr-start 3848 '<FieldAccess for name_symbol at ast.py:6585>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6585
--# expr-start 3847 '<FieldAccess for name at ast.py:6585>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6585
      if Var_Expr_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.name'
Fld_23 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Name (Node => Var_Expr_5.Node, E_Info => Var_Expr_5.Info);
--# end
--# expr-done 3847
      if Fld_23.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_24 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_23.Node);
--# end
--# expr-done 3848
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Fld_22, Key => Fld_24, Lookup_Kind => To_Lookup_Kind_Type (Recursive), Categories => (Inherited_Primitives => True, Nocat => True))); 
--# expr-done 3849
      Quantifier_Result_1 := False;
      declare
         Collection : constant Internal_Entity_Array_Access := Env_Get_Result;
      begin
         for Item_151 of
               Collection.Items
         loop
   --# scope-start
               --# bind x Item_151
      --# expr-start 3852 '<FieldAccess for match_signature at ast.py:6588>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6588
--# expr-start 3851 '<FieldAccess for subp_spec_or_null at ast.py:6588>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6588
--# expr-start 3850 '<Cast to BasicDecl at ast.py:6588>' Cast_Result_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6588
Cast_Expr_4 := Item_151; 
   if Cast_Expr_4.Node = null
      or else Cast_Expr_4.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_9 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_4.Node,
         Info => Cast_Expr_4.Info);
   else
         Cast_Result_9 := No_Entity_Basic_Decl;
   end if;
--# expr-done 3850
      if Cast_Result_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_25 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Cast_Result_9.Node, E_Info => Cast_Result_9.Info);
--# end
--# expr-done 3851
      if Fld_25.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.match_signature
Fld_26 := Libadalang.Implementation.Base_Subp_Spec_P_Match_Signature (Node => Fld_25.Node, Other => Var_Expr_5, Match_Name => False, Use_Entity_Info => True, E_Info => Fld_25.Info);
--# end
--# expr-done 3852
      Quantifier_Result_1 := Fld_26;
   --# end
               exit when Quantifier_Result_1;
         end loop;
      end;
--# expr-done 3853
   If_Result_9 := Quantifier_Result_1;
else
   If_Result_9 := True;
end if;
--# expr-done 3854
   If_Result_10 := If_Result_9;
else
   If_Result_10 := False;
end if;
Scope_Result_8 := If_Result_10;
   --# end
      Finalizer_Scope_2276;
   Result_Var_6 := Scope_Result_8;
else
   Result_Var_6 := False;
end if;
         if Result_Var_6 then
         declare
            Item_To_Append : constant Internal_Entity := Item_149;
         begin
            Internal_Entity_Vectors.Append (Map_Result_4_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result_4 := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_4_Vec)));
         for I in Map_Result_4.Items'Range loop
            Map_Result_4.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_4_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result_4.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_4_Vec);
      end;
--# expr-done 3855
      declare
         Map_Result_5_Vec : Internal_Inner_Env_Assoc_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Map_Result_4;
      begin
         for Item_148 of
               Collection.Items
         loop
   --# scope-start
               --# bind decl Item_148
      --# expr-start 3871 '<Let at ast.py:6596>' Let_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6596
--# scope-start
--# expr-start 3856 '<Cast to BasicDecl at ast.py:6597>' Cast_Result_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6597
Cast_Expr_5 := Item_148; 
   if Cast_Expr_5.Node = null
      or else Cast_Expr_5.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_10 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_5.Node,
         Info => Cast_Expr_5.Info);
   else
         Cast_Result_10 := No_Entity_Basic_Decl;
   end if;
--# expr-done 3856
Bd := Cast_Result_10; 
--# bind bd Bd
--# expr-start 3870 '<Concat at ast.py:6597>' Concat_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6597
--# expr-start 3861 '<New InnerEnvAssoc at ast.py:6597>' New_Struct_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6597
--# expr-start 3858 '<FieldAccess for name_symbol at ast.py:6598>' Fld_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6598
--# expr-start 3857 '<FieldAccess for defining_name at ast.py:6598>' Fld_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6598
      if Bd.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_28 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Bd.Node, E_Info => Bd.Info);
--# end
--# expr-done 3857
      if Fld_28.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_29 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_28.Node);
--# end
--# expr-done 3858
--# expr-start 3859 '<New Metadata at ast.py:6600>' New_Struct_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6600
Cast_Expr_6 := Self; 
      Cast_Result_11 := Cast_Expr_6;
New_Struct_2 := (Dottable_Subp => False, Primitive => Cast_Result_11, Primitive_Real_Type => No_Bare_Ada_Node); 
--# expr-done 3859
--# expr-start 3860 '<FieldAccess for node at ast.py:6599>' Fld_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6599
Fld_30 := Bd.Node;
--# expr-done 3860
      Cast_Result_12 := Fld_30;
New_Struct_3 := (Key => Fld_29, Metadata => New_Struct_2, Value => Cast_Result_12); 
--# expr-done 3861
                Singleton := Create_Internal_Inner_Env_Assoc_Array (Items_Count => 1);
                Singleton.Items (1) := New_Struct_3;
--# expr-start 3869 '<If at ast.py:6601>' If_Result_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6601
--# expr-start 3864 '<FieldAccess for name_is at ast.py:6602>' Fld_33 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6602
--# expr-start 3862 '<FieldAccess for defining_name at ast.py:6602>' Fld_32 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6602
      if Bd.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_32 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Bd.Node, E_Info => Bd.Info);
--# end
--# expr-done 3862
      if Fld_32.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 3863 '<Symbol "=" at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_10); 
--# expr-done 3863
--# property-call-start Name.name_is
Fld_33 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_32.Node, Sym => Sym_1);
--# end
--# expr-done 3864
if Fld_33 then
   --# expr-start 3868 '<New InnerEnvAssoc at ast.py:6603>' New_Struct_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6603
--# expr-start 3865 '<Symbol "/=" at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_7); 
--# expr-done 3865
--# expr-start 3866 '<New Metadata at ast.py:6606>' New_Struct_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6606
Cast_Expr_7 := Self; 
      Cast_Result_13 := Cast_Expr_7;
New_Struct_4 := (Dottable_Subp => False, Primitive => Cast_Result_13, Primitive_Real_Type => No_Bare_Ada_Node); 
--# expr-done 3866
--# expr-start 3867 '<FieldAccess for node at ast.py:6605>' Fld_31 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6605
Fld_31 := Bd.Node;
--# expr-done 3867
      Cast_Result_14 := Fld_31;
New_Struct_5 := (Key => Sym, Metadata => New_Struct_4, Value => Cast_Result_14); 
--# expr-done 3868
                Singleton_1 := Create_Internal_Inner_Env_Assoc_Array (Items_Count => 1);
                Singleton_1.Items (1) := New_Struct_5;
   If_Result_11 := Singleton_1;
else
   If_Result_11 := No_Internal_Inner_Env_Assoc_Array_Type;
end if;
      Inc_Ref (If_Result_11);
--# expr-done 3869
Concat_Result_1 := Concat (Singleton, If_Result_11); 
--# expr-done 3870
Let_Result_2 := Concat_Result_1; Inc_Ref (Let_Result_2);
--# end
--# expr-done 3871
         for Item_To_Append of
               Let_Result_2.Items
         loop
            Internal_Inner_Env_Assoc_Vectors.Append (Map_Result_5_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_2279;
         end loop;
      end;
         Map_Result_5 := Create_Internal_Inner_Env_Assoc_Array
           (Items_Count => Natural (Internal_Inner_Env_Assoc_Vectors.Length (Map_Result_5_Vec)));
         for I in Map_Result_5.Items'Range loop
            Map_Result_5.Items (I) := Internal_Inner_Env_Assoc_Vectors.Get
              (Map_Result_5_Vec,
               I + Internal_Inner_Env_Assoc_Vectors.Index_Type'First - Map_Result_5.Items'First);
         end loop;
         Internal_Inner_Env_Assoc_Vectors.Destroy (Map_Result_5_Vec);
      end;
--# expr-done 3872
Block_Var_4 := Map_Result_5; Inc_Ref (Block_Var_4);
--# bind prim_subps Block_Var_4
--# expr-start 3878 '<Map at ast.py:6614>' Map_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6614
   --# expr-start 3873 '<FieldAccess for predefined_operators at ast.py:6614>' Fld_34 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6614
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDecl.predefined_operators
Fld_34 := Libadalang.Implementation.Type_Decl_P_Predefined_Operators (Node => Self);
--# end
--# expr-done 3873
      declare
         Map_Result_6_Vec : Internal_Inner_Env_Assoc_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Env_Assoc_Array_Access := Fld_34;
      begin
         for Item_152 of
               Collection.Items
         loop
   --# scope-start
               --# bind assoc Item_152
      --# expr-start 3877 '<New InnerEnvAssoc at ast.py:6615>' New_Struct_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6615
--# expr-start 3874 '<FieldAccess for key at ast.py:6616>' Fld_35 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6616
Fld_35 := Item_152.Key;
--# expr-done 3874
--# expr-start 3875 '<New Metadata at ast.py:6618>' New_Struct_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6618
Cast_Expr_8 := Self; 
      Cast_Result_15 := Cast_Expr_8;
New_Struct_6 := (Dottable_Subp => False, Primitive => Cast_Result_15, Primitive_Real_Type => No_Bare_Ada_Node); 
--# expr-done 3875
--# expr-start 3876 '<FieldAccess for value at ast.py:6617>' Fld_36 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6617
Fld_36 := Item_152.Value;
--# expr-done 3876
New_Struct_7 := (Key => Fld_35, Metadata => New_Struct_6, Value => Fld_36); 
--# expr-done 3877
         declare
            Item_To_Append : constant Internal_Inner_Env_Assoc := New_Struct_7;
         begin
            Internal_Inner_Env_Assoc_Vectors.Append (Map_Result_6_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result_6 := Create_Internal_Inner_Env_Assoc_Array
           (Items_Count => Natural (Internal_Inner_Env_Assoc_Vectors.Length (Map_Result_6_Vec)));
         for I in Map_Result_6.Items'Range loop
            Map_Result_6.Items (I) := Internal_Inner_Env_Assoc_Vectors.Get
              (Map_Result_6_Vec,
               I + Internal_Inner_Env_Assoc_Vectors.Index_Type'First - Map_Result_6.Items'First);
         end loop;
         Internal_Inner_Env_Assoc_Vectors.Destroy (Map_Result_6_Vec);
      end;
--# expr-done 3878
Block_Var_5 := Map_Result_6; Inc_Ref (Block_Var_5);
--# bind predefined_ops Block_Var_5
--# expr-start 3880 '<Concat at ast.py:6622>' Concat_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6622
--# expr-start 3879 '<Concat at ast.py:6622>' Concat_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6622
Concat_Result_2 := Concat (Block_Var_3, Block_Var_4); 
--# expr-done 3879
Concat_Result_3 := Concat (Concat_Result_2, Block_Var_5); 
--# expr-done 3880
Let_Result_3 := Concat_Result_3; Inc_Ref (Let_Result_3);
--# end
--# expr-done 3881
      Property_Result := Let_Result_3;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_490;
      Self.Lf_Present_Direct_Primitive_Subps := True;
      Self.Lf_Stg_Direct_Primitive_Subps := Property_Result;
         Inc_Ref (Property_Result);
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_490;
               Finalizer_Scope_2264;
               Finalizer_Scope_2265;
               Finalizer_Scope_2269;
               Finalizer_Scope_2270;
               Finalizer_Scope_2272;
               Finalizer_Scope_2276;
               Finalizer_Scope_2279;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_F_Direct_Primitive_Subps;
--# end
--# property-start TypeDecl.array_ndims /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6624
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Array_Ndims
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Integer
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Integer;
      Fld : Internal_Entity_Type_Def;
Fld_1 : Integer;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3883 '<FieldAccess for array_ndims at ast.py:6624>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6624
--# expr-start 3882 '<FieldAccess for type_def at ast.py:6624>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6624
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3882
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.array_ndims'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Array_Ndims (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3883
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Array_Ndims;
--# end
--# property-start TypeDecl.is_real_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6626
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Is_Real_Type
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Type_Def;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3885 '<FieldAccess for is_real_type at ast.py:6626>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6626
--# expr-start 3884 '<FieldAccess for type_def at ast.py:6626>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6626
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3884
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDef.is_real_type
Fld_1 := Libadalang.Implementation.Type_Def_P_Is_Real_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3885
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Is_Real_Type;
--# end
--# property-start TypeDecl.is_float_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6627
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Is_Float_Type
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Type_Def;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3887 '<FieldAccess for is_float_type at ast.py:6627>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6627
--# expr-start 3886 '<FieldAccess for type_def at ast.py:6627>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6627
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3886
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.is_float_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Is_Float_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3887
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Is_Float_Type;
--# end
--# property-start TypeDecl.is_fixed_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6628
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Is_Fixed_Point
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Type_Def;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3889 '<FieldAccess for is_fixed_point at ast.py:6628>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6628
--# expr-start 3888 '<FieldAccess for type_def at ast.py:6628>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6628
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3888
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.is_fixed_point'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Is_Fixed_Point (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3889
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Is_Fixed_Point;
--# end
--# property-start TypeDecl.is_int_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6629
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Is_Int_Type
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Type_Def;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3891 '<FieldAccess for is_int_type at ast.py:6629>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6629
--# expr-start 3890 '<FieldAccess for type_def at ast.py:6629>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6629
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3890
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.is_int_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Is_Int_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3891
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Is_Int_Type;
--# end
--# property-start TypeDecl.is_access_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6630
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Is_Access_Type
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
      As_Entity : Internal_Entity_Type_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Type_Decl;
Fld : Internal_Entity_Type_Def;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3894 '<FieldAccess for is_access_type at ast.py:6630>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6630
--# expr-start 3893 '<FieldAccess for type_def at ast.py:6630>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6630
--# expr-start 3892 '<as_bare_entity at ast.py:6630>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6630
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity_Type_Decl;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 3892
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => If_Result.Node.Type_Decl_F_Type_Def, Info => If_Result.Info);
--# expr-done 3893
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.is_access_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Is_Access_Type (Node => Fld.Node, Origin => Origin);
--# end
--# expr-done 3894
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Is_Access_Type;
--# end
--# property-start TypeDecl.is_static_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6631
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Is_Static_Decl
  (Node : Bare_Type_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      As_Entity : Internal_Entity_Type_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Type_Decl;
Fld : Internal_Entity_Type_Def;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3897 '<FieldAccess for is_static at ast.py:6631>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6631
--# expr-start 3896 '<FieldAccess for type_def at ast.py:6631>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6631
--# expr-start 3895 '<as_bare_entity at ast.py:6631>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6631
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity_Type_Decl;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 3895
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => If_Result.Node.Type_Decl_F_Type_Def, Info => If_Result.Info);
--# expr-done 3896
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.is_static'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Is_Static (Node => Fld.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld.Info);
--# end
--# expr-done 3897
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Is_Static_Decl;
--# end
--# property-start TypeDecl.accessed_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6634
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Accessed_Type
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_498 with Inline_Always;
      Block_Var : Internal_Entity_Expr;
Fld : Internal_Entity_Expr;
Fld_1 : Internal_Entity_Type_Def;
Fld_2 : Internal_Entity_Base_Type_Decl;
Bound_Include_Ud_Indexing : Boolean;
Fld_3 : Internal_Entity_Type_Def;
Fld_4 : Lexical_Env := Empty_Env;
Cast_Expr : Internal_Entity_Expr;
Cast_Result : Internal_Entity_Name;
Fld_5 : Symbol_Type;
Env_Get_Result : Internal_Entity;
Dyn_Var_Bind_Result : Internal_Entity;
Cast_Result_1 : Internal_Entity_Base_Formal_Param_Decl;
Fld_6 : Internal_Entity_Base_Type_Decl;
Fld_7 : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_498 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_498;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3912 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3898 '<FieldAccess for get_imp_deref at ast.py:6635>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6635
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.get_imp_deref'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Get_Imp_Deref (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3898
Block_Var := Fld; 
--# bind imp_deref Block_Var
--# expr-start 3911 '<If at ast.py:6637>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6637
--# expr-start 3899 '<IsNull at ast.py:6638>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6638
Is_Null := Block_Var.Node = null; 
--# expr-done 3899
if Is_Null then
   --# expr-start 3901 '<FieldAccess for accessed_type at ast.py:6639>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6639
--# expr-start 3900 '<FieldAccess for type_def at ast.py:6639>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6639
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3900
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.accessed_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Type_Def_P_Accessed_Type (Node => Fld_1.Node, Origin => Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 3901
   If_Result := Fld_2;
else
   --# expr-start 3910 '<FieldAccess for accessed_type at ast.py:6643>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6643
--# expr-start 3909 '<FieldAccess for formal_type at ast.py:6643>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6643
--# expr-start 3908 '<Cast to BaseFormalParamDecl at ast.py:6643>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6643
--# expr-start 3907 '<bind at ast.py:6643>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6643
Bound_Include_Ud_Indexing := False; 
--# expr-start 3906 '<EnvGet at ast.py:6645>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6645
--# expr-start 3903 '<FieldAccess for defining_env at ast.py:6645>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6645
--# expr-start 3902 '<FieldAccess for type_def at ast.py:6645>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6645
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3902
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDef.defining_env
Fld_4 := Libadalang.Implementation.Type_Def_P_Defining_Env (Node => Fld_3.Node, Origin => Origin, Include_Ud_Indexing => Bound_Include_Ud_Indexing, E_Info => Fld_3.Info);
--# end
--# expr-done 3903
--# expr-start 3905 '<FieldAccess for name_symbol at ast.py:6646>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6646
--# expr-start 3904 '<Cast to Name at ast.py:6646>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6646
Cast_Expr := Block_Var; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Name
   then
      Cast_Result := Create_Internal_Entity_Name
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Name;
   end if;
--# expr-done 3904
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_5 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Cast_Result.Node);
--# end
--# expr-done 3905
Env_Get_Result := AST_Envs.Get_First (Self => Fld_4, Key => Fld_5, Lookup_Kind => To_Lookup_Kind_Type (Recursive), Categories => (Inherited_Primitives => False, Nocat => True)); 
--# expr-done 3906
Dyn_Var_Bind_Result := Env_Get_Result; 
--# expr-done 3907
   if Dyn_Var_Bind_Result.Node = null
      or else Dyn_Var_Bind_Result.Node.Kind in Ada_Base_Formal_Param_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Base_Formal_Param_Decl
        (Node => Dyn_Var_Bind_Result.Node,
         Info => Dyn_Var_Bind_Result.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 3908
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamDecl.formal_type
Fld_6 := Libadalang.Implementation.Base_Formal_Param_Decl_P_Formal_Type (Node => Cast_Result_1.Node, Origin => Origin, E_Info => Cast_Result_1.Info);
--# end
--# expr-done 3909
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_7 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Fld_6.Node, Origin => Origin, E_Info => Fld_6.Info);
--# end
--# expr-done 3910
   If_Result := Fld_7;
end if;
--# expr-done 3911
Let_Result := If_Result; 
--# end
--# expr-done 3912
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_498;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_498;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Accessed_Type;
--# end
--# property-start TypeDecl.access_def /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6656
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Access_Def
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Access_Def
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Access_Def;
      Match_0 : Internal_Entity_Access_Def;
Match_1 : Internal_Entity_Derived_Type_Def;
Match_2 : Internal_Entity_Type_Def;
Fld : Internal_Entity_Type_Def;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Access_Def;
Match_Prefix : Internal_Entity_Type_Def;
Cast_Result : Internal_Entity_Access_Def;
Let_Result : Internal_Entity_Access_Def;
Scope_Result : Internal_Entity_Access_Def;
Cast_Result_1 : Internal_Entity_Derived_Type_Def;
Let_Result_1 : Internal_Entity_Access_Def;
Scope_Result_1 : Internal_Entity_Access_Def;
Cast_Result_2 : Internal_Entity_Type_Def;
Let_Result_2 : Internal_Entity_Access_Def;
Scope_Result_2 : Internal_Entity_Access_Def;
Match_Result : Internal_Entity_Access_Def;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3916 '<Match at ast.py:6656>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6656
--# expr-start 3913 '<FieldAccess for type_def at ast.py:6656>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6656
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3913
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld; 
case Ada_Type_Def (Match_Prefix.Node.Kind) is
         when Ada_Access_To_Subp_Def .. Ada_Type_Access_Def =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Access_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind ad Match_0
Let_Result := Match_0; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Derived_Type_Def =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Derived_Type_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind dtd Match_1
--# expr-start 3915 '<FieldAccess for access_def at ast.py:6658>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6658
--# expr-start 3914 '<FieldAccess for base_type at ast.py:6658>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6658
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => Match_1.Node, Origin => Origin, E_Info => Match_1.Info);
--# end
--# expr-done 3914
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.access_def'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Access_Def (Node => Fld_1.Node, Origin => Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 3915
Let_Result_1 := Fld_2; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Array_Type_Def | Ada_Enum_Type_Def .. Ada_Signed_Int_Type_Def =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Type_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind _ Match_2
Let_Result_2 := No_Entity_Access_Def; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 3916
      Property_Result := Match_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Access_Def;
--# end
--# property-start TypeDecl.is_tagged_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6662
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Is_Tagged_Type
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Type_Def;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 3918 '<FieldAccess for is_tagged_type at ast.py:6662>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6662
--# expr-start 3917 '<FieldAccess for type_def at ast.py:6662>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6662
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3917
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.is_tagged_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Is_Tagged_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3918
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Is_Tagged_Type;
--# end
--# property-start TypeDecl.base_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6663
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Base_Type
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Type_Def;
Fld_1 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3920 '<FieldAccess for base_type at ast.py:6663>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6663
--# expr-start 3919 '<FieldAccess for type_def at ast.py:6663>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6663
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3919
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3920
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Base_Type;
--# end
--# property-start TypeDecl.base_interfaces /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6664
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Base_Interfaces
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl_Array_Access
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_502 with Inline_Always;
      Fld : Internal_Entity_Type_Def;
Fld_1 : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_502 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_502;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3922 '<FieldAccess for base_interfaces at ast.py:6664>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6664
--# expr-start 3921 '<FieldAccess for type_def at ast.py:6664>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6664
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3921
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_interfaces'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Interfaces (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 3922
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_502;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_502;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Base_Interfaces;
--# end
--# property-start TypeDecl.is_char_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6665
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Is_Char_Type
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Type_Def;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3924 '<FieldAccess for is_char_type at ast.py:6665>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6665
--# expr-start 3923 '<FieldAccess for type_def at ast.py:6665>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6665
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3923
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.is_char_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Is_Char_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3924
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Is_Char_Type;
--# end
--# property-start TypeDecl.is_enum_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6666
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Is_Enum_Type
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Type_Def;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3926 '<FieldAccess for is_enum_type at ast.py:6666>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6666
--# expr-start 3925 '<FieldAccess for type_def at ast.py:6666>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6666
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3925
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.is_enum_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Is_Enum_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3926
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Is_Enum_Type;
--# end
--# property-start TypeDecl.is_private /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6667
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Is_Private
  (Node : Bare_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Var_Expr : Bare_Derived_Type_Def;
Fld : Bare_Type_Def;
Is_A : Boolean;
Fld_1 : Bare_Type_Def;
Cast_Result : Bare_Derived_Type_Def;
Fld_2 : Bare_With_Private;
Fld_3 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 3928 '<IsA PrivateTypeDef at ast.py:6668>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6668
--# expr-start 3927 '<FieldAccess for type_def at ast.py:6668>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6668
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Type_Decl_F_Type_Def;
--# expr-done 3927
Is_A := Fld /= null 
and then Fld.Kind in Ada_Private_Type_Def_Range; 
--# expr-done 3928
if Is_A then
   If_Result := True;
else
--# expr-start 3930 '<Cast to DerivedTypeDef at ast.py:6669>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6669
--# expr-start 3929 '<FieldAccess for type_def at ast.py:6669>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6669
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Self.Type_Decl_F_Type_Def;
--# expr-done 3929
   if Fld_1 = null
      or else Fld_1.Kind in Ada_Derived_Type_Def_Range
   then
      Cast_Result := Fld_1;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 3930
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind dtd Var_Expr
--# expr-start 3932 '<FieldAccess for as_bool at ast.py:6670>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6670
--# expr-start 3931 '<FieldAccess for has_with_private at ast.py:6670>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6670
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Var_Expr.Derived_Type_Def_F_Has_With_Private;
--# expr-done 3931
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]WithPrivate.as_bool'
Fld_3 := Libadalang.Implementation.Dispatcher_With_Private_P_As_Bool (Node => Fld_2);
--# end
--# expr-done 3932
Scope_Result := Fld_3;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
   If_Result := Result_Var;
end if;
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Is_Private;
--# end
--# property-start TypeDecl.is_derived_tagged_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6674
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Is_Derived_Tagged_Type
  (Node : Bare_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Internal_Entity_Type_Def;
Fld_1 : Boolean;
Fld_2 : Internal_Entity_Type_Def;
Is_A : Boolean;
If_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 3934 '<FieldAccess for is_tagged_type at ast.py:6675>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6675
--# expr-start 3933 '<FieldAccess for type_def at ast.py:6675>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6675
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3933
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.is_tagged_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Is_Tagged_Type (Node => Fld.Node, Origin => No_Bare_Ada_Node, E_Info => Fld.Info);
--# end
--# expr-done 3934
if Fld_1 then
   --# expr-start 3936 '<IsA DerivedTypeDef, InterfaceTypeDef at ast.py:6676>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6676
--# expr-start 3935 '<FieldAccess for type_def at ast.py:6676>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6676
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3935
Is_A := Fld_2.Node /= null 
and then Fld_2.Node.Kind in Ada_Derived_Type_Def_Range | Ada_Interface_Type_Def_Range; 
--# expr-done 3936
   If_Result := Is_A;
else
   If_Result := False;
end if;
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Is_Derived_Tagged_Type;
--# end
--# property-start TypeDecl.array_def /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6679
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Array_Def
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Type_Def
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Array_Type_Def;
      Match_0 : Internal_Entity_Array_Type_Def;
Match_1 : Internal_Entity_Derived_Type_Def;
Match_2 : Internal_Entity_Type_Def;
Fld : Internal_Entity_Type_Def;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Array_Type_Def;
Match_Prefix : Internal_Entity_Type_Def;
Cast_Result : Internal_Entity_Array_Type_Def;
Let_Result : Internal_Entity_Array_Type_Def;
Scope_Result : Internal_Entity_Array_Type_Def;
Cast_Result_1 : Internal_Entity_Derived_Type_Def;
Let_Result_1 : Internal_Entity_Array_Type_Def;
Scope_Result_1 : Internal_Entity_Array_Type_Def;
Cast_Result_2 : Internal_Entity_Type_Def;
Let_Result_2 : Internal_Entity_Array_Type_Def;
Scope_Result_2 : Internal_Entity_Array_Type_Def;
Match_Result : Internal_Entity_Array_Type_Def;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3940 '<Match at ast.py:6679>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6679
--# expr-start 3937 '<FieldAccess for type_def at ast.py:6679>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6679
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3937
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld; 
case Ada_Type_Def (Match_Prefix.Node.Kind) is
         when Ada_Array_Type_Def =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Array_Type_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind atd Match_0
Let_Result := Match_0; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Derived_Type_Def =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Derived_Type_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind dtd Match_1
--# expr-start 3939 '<FieldAccess for array_def at ast.py:6681>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6681
--# expr-start 3938 '<FieldAccess for base_type at ast.py:6681>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6681
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => Match_1.Node, Origin => Origin, E_Info => Match_1.Info);
--# end
--# expr-done 3938
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.array_def'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Array_Def (Node => Fld_1.Node, Origin => Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 3939
Let_Result_1 := Fld_2; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Access_To_Subp_Def .. Ada_Type_Access_Def | Ada_Enum_Type_Def .. Ada_Signed_Int_Type_Def =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Type_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind _ Match_2
Let_Result_2 := No_Entity_Array_Type_Def; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 3940
      Property_Result := Match_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Array_Def;
--# end
--# property-start TypeDecl.root_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6685
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Root_Type
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Match_0 : Internal_Entity_Derived_Type_Def;
Match_1 : Internal_Entity_Type_Def;
Fld : Internal_Entity_Type_Def;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Base_Type_Decl;
Match_Prefix : Internal_Entity_Type_Def;
Cast_Result : Internal_Entity_Derived_Type_Def;
Let_Result : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity_Type_Def;
Cast_Expr : Internal_Entity_Type_Decl;
Cast_Result_2 : Internal_Entity_Base_Type_Decl;
Let_Result_1 : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Match_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3944 '<Match at ast.py:6685>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6685
--# expr-start 3941 '<FieldAccess for type_def at ast.py:6685>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6685
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3941
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld; 
case Ada_Type_Def (Match_Prefix.Node.Kind) is
         when Ada_Derived_Type_Def =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Derived_Type_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind dtd Match_0
--# expr-start 3943 '<FieldAccess for root_type at ast.py:6686>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6686
--# expr-start 3942 '<FieldAccess for base_type at ast.py:6686>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6686
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => Match_0.Node, Origin => Origin, E_Info => Match_0.Info);
--# end
--# expr-done 3942
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.root_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Root_Type (Node => Fld_1.Node, Origin => Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 3943
Let_Result := Fld_2; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Access_To_Subp_Def .. Ada_Array_Type_Def | Ada_Enum_Type_Def .. Ada_Signed_Int_Type_Def =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Type_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind _ Match_1
Cast_Expr := Ent; 
      Cast_Result_2 := Create_Internal_Entity_Base_Type_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Let_Result_1 := Cast_Result_2; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
end case;
--# expr-done 3944
      Property_Result := Match_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Root_Type;
--# end
--# property-start TypeDecl.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6691
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Defining_Env
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_509 with Inline_Always;
            procedure Finalizer_Scope_2298 with Inline_Always;
      Block_Var : Internal_Entity_Expr;
Block_Var_1 : Lexical_Env := Empty_Env;
Fld : Internal_Entity_Expr;
Fld_1 : Internal_Entity_Type_Def;
Fld_2 : Lexical_Env := Empty_Env;
Is_Null : Boolean;
Not_Val : Boolean;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Lexical_Env := Empty_Env;
Array_Lit : Lexical_Env_Array_Access;
Group_Env : Lexical_Env := Empty_Env;
Fld_5 : Boolean;
If_Result : Boolean;
Bound_Include_Ud_Indexing : Boolean;
Fld_6 : Internal_Entity_Basic_Decl_Array_Access;
Fld_7 : Internal_Entity_Basic_Decl_Array_Access;
Concat_Result : Internal_Entity_Basic_Decl_Array_Access;
Fld_8 : Lexical_Env := Empty_Env;
Map_Result : Lexical_Env_Array_Access;
Array_Lit_1 : Lexical_Env_Array_Access;
Concat_Result_1 : Lexical_Env_Array_Access;
Group_Env_1 : Lexical_Env := Empty_Env;
Dyn_Var_Bind_Result : Lexical_Env := Empty_Env;
If_Result_1 : Lexical_Env := Empty_Env;
If_Result_2 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_509 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Group_Env);
                     Dec_Ref (Fld_6);
                     Dec_Ref (Fld_7);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Array_Lit_1);
                     Dec_Ref (Concat_Result_1);
                     Dec_Ref (Group_Env_1);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (If_Result_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_509;
            procedure Finalizer_Scope_2298 is
            begin
                     Dec_Ref (Fld_8);
            end Finalizer_Scope_2298;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3965 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3945 '<FieldAccess for get_imp_deref at ast.py:6692>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6692
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.get_imp_deref'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Get_Imp_Deref (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3945
Block_Var := Fld; 
--# bind imp_deref Block_Var
--# expr-start 3947 '<FieldAccess for defining_env at ast.py:6696>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6696
--# expr-start 3946 '<FieldAccess for type_def at ast.py:6696>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6696
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3946
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDef.defining_env
Fld_2 := Libadalang.Implementation.Type_Def_P_Defining_Env (Node => Fld_1.Node, Origin => Origin, Include_Ud_Indexing => Include_Ud_Indexing, E_Info => Fld_1.Info);
--# end
--# expr-done 3947
Block_Var_1 := Fld_2; Inc_Ref (Block_Var_1);
--# bind self_env Block_Var_1
--# expr-start 3964 '<Cond at ast.py:6698>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6698
--# expr-start 3949 '<Not at ast.py:6699>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6699
--# expr-start 3948 '<IsNull at ast.py:6699>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6699
Is_Null := Block_Var.Node = null; 
--# expr-done 3948
Not_Val := not (Is_Null); 
--# expr-done 3949
if Not_Val then
   --# expr-start 3953 '<env_group at ast.py:6700>' Group_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6700
--# expr-start 3952 '<ArrayLiteral at ast.py:6700>' Array_Lit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6700
--# expr-start 3951 '<FieldAccess for defining_env at ast.py:6700>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6700
--# expr-start 3950 '<FieldAccess for accessed_type at ast.py:6700>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6700
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3950
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_4 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Fld_3.Node, Origin => Origin, Include_Ud_Indexing => Include_Ud_Indexing, E_Info => Fld_3.Info);
--# end
--# expr-done 3951
Array_Lit := Create_Lexical_Env_Array (Internal_Lexical_Env_Array'(1 => Block_Var_1, 2 => Fld_4)); 
--# expr-done 3952
Group_Env := Group (Array_Lit, No_Metadata); 
--# expr-done 3953
   If_Result_2 := Group_Env;
else
if Include_Ud_Indexing then
   --# expr-start 3954 '<FieldAccess for has_ud_indexing at ast.py:6702>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6702
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.has_ud_indexing'
Fld_5 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Has_Ud_Indexing (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3954
   If_Result := Fld_5;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 3963 '<bind at ast.py:6703>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6703
Bound_Include_Ud_Indexing := False; 
--# expr-start 3962 '<env_group at ast.py:6705>' Group_Env_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6705
--# expr-start 3961 '<Concat at ast.py:6705>' Concat_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6705
--# expr-start 3959 '<Map at ast.py:6705>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6705
   --# expr-start 3957 '<Concat at ast.py:6705>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6705
--# expr-start 3955 '<FieldAccess for constant_indexing_fns at ast.py:6705>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6705
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.constant_indexing_fns'
Fld_6 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Constant_Indexing_Fns (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3955
--# expr-start 3956 '<FieldAccess for variable_indexing_fns at ast.py:6706>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6706
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.variable_indexing_fns'
Fld_7 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Variable_Indexing_Fns (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 3956
Concat_Result := Concat (Fld_6, Fld_7); 
--# expr-done 3957
      declare
         Map_Result_Vec : Lexical_Env_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Basic_Decl_Array_Access := Concat_Result;
      begin
         for Item_153 of
               Collection.Items
         loop
   --# scope-start
               --# bind fn Item_153
      --# expr-start 3958 '<FieldAccess for defining_env at ast.py:6707>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6707
      if Item_153.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_8 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Item_153.Node, Origin => Origin, Include_Ud_Indexing => Bound_Include_Ud_Indexing, E_Info => Item_153.Info);
--# end
--# expr-done 3958
         declare
            Item_To_Append : constant Lexical_Env := Fld_8;
         begin
               Inc_Ref (Item_To_Append);
            Lexical_Env_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2298;
         end loop;
      end;
         Map_Result := Create_Lexical_Env_Array
           (Items_Count => Natural (Lexical_Env_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Lexical_Env_Vectors.Get
              (Map_Result_Vec,
               I + Lexical_Env_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Lexical_Env_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 3959
--# expr-start 3960 '<ArrayLiteral at ???>' Array_Lit_1 None
Array_Lit_1 := Create_Lexical_Env_Array (Internal_Lexical_Env_Array'(1 => Block_Var_1)); 
--# expr-done 3960
Concat_Result_1 := Concat (Map_Result, Array_Lit_1); 
--# expr-done 3961
Group_Env_1 := Group (Concat_Result_1, No_Metadata); 
--# expr-done 3962
Dyn_Var_Bind_Result := Group_Env_1; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 3963
   If_Result_1 := Dyn_Var_Bind_Result;
else
   If_Result_1 := Block_Var_1;
end if;
      Inc_Ref (If_Result_1);
   If_Result_2 := If_Result_1;
end if;
      Inc_Ref (If_Result_2);
--# expr-done 3964
Let_Result := If_Result_2; Inc_Ref (Let_Result);
--# end
--# expr-done 3965
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_509;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_509;
               Finalizer_Scope_2298;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Defining_Env;
--# end
--# property-start TypeDecl.predefined_operators /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6715
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Predefined_Operators
  (Node : Bare_Type_Decl
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_510 with Inline_Always;
      Fld : Bare_Type_Def;
Fld_1 : Internal_Env_Assoc_Array_Access;
Fld_2 : Internal_Entity;
Fld_3 : Bare_Ada_Node;
Cast_Expr : Bare_Type_Decl;
Cast_Result : Bare_Ada_Node;
Is_Equal : Boolean;
Fld_4 : Internal_Entity;
Fld_5 : Bare_Ada_Node;
Cast_Expr_1 : Bare_Type_Decl;
Cast_Result_1 : Bare_Ada_Node;
Is_Equal_1 : Boolean;
If_Result : Boolean;
If_Result_1 : Internal_Env_Assoc_Array_Access;
Let_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_510 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_510;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Type_Decl_Type_Decl_P_Predefined_Operators,
            Items    => new Mmz_Key_Array (1 ..  1))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Type_Decl,
                                As_Bare_Type_Decl => Self);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Env_Assoc_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 3975 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3974 '<If at ast.py:6724>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6724
--# expr-start 3968 '<Eq at ast.py:6725>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6725
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-start 3967 '<FieldAccess for node at ast.py:6725>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6725
--# expr-start 3966 '<FieldAccess for universal_int_type at ast.py:6725>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6725
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.universal_int_type
Fld_2 := Libadalang.Implementation.Ada_Node_P_Universal_Int_Type (Node => Self);
--# end
--# expr-done 3966
Fld_3 := Fld_2.Node;
--# expr-done 3967
Is_Equal := Cast_Result = Fld_3; 
--# expr-done 3968
if Is_Equal then
   If_Result := True;
else
   --# expr-start 3971 '<Eq at ast.py:6725>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6725
Cast_Expr_1 := Self; 
      Cast_Result_1 := Cast_Expr_1;
--# expr-start 3970 '<FieldAccess for node at ast.py:6726>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6726
--# expr-start 3969 '<FieldAccess for universal_real_type at ast.py:6726>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6726
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.universal_real_type
Fld_4 := Libadalang.Implementation.Ada_Node_P_Universal_Real_Type (Node => Self);
--# end
--# expr-done 3969
Fld_5 := Fld_4.Node;
--# expr-done 3970
Is_Equal_1 := Cast_Result_1 = Fld_5; 
--# expr-done 3971
   If_Result := Is_Equal_1;
end if;
if If_Result then
   If_Result_1 := No_Internal_Env_Assoc_Array_Type;
else
   --# expr-start 3973 '<FieldAccess for predefined_operators at ast.py:6728>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6728
--# expr-start 3972 '<FieldAccess for type_def at ast.py:6728>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6728
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Type_Decl_F_Type_Def;
--# expr-done 3972
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.predefined_operators'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Predefined_Operators (Node => Fld);
--# end
--# expr-done 3973
   If_Result_1 := Fld_1;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 3974
Let_Result := If_Result_1; Inc_Ref (Let_Result);
--# end
--# expr-done 3975
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_510;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Env_Assoc_Array_Access,
                        As_Internal_Env_Assoc_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_510;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Predefined_Operators;
--# end
--# property-start TypeDecl.record_def /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6746
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Record_Def
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Record_Def
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Record_Def;
      Match_0 : Internal_Entity_Record_Type_Def;
Match_1 : Internal_Entity_Derived_Type_Def;
Match_2 : Internal_Entity_Type_Def;
Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Type_Def;
Fld_1 : Internal_Entity_Base_Record_Def;
Fld_2 : Internal_Entity_Base_Record_Def;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Internal_Entity_Base_Record_Def;
Scope_Result : Internal_Entity_Base_Record_Def;
Result_Var : Internal_Entity_Base_Record_Def;
Fld_5 : Boolean;
If_Result : Internal_Entity_Base_Record_Def;
Match_Prefix : Internal_Entity_Type_Def;
Cast_Result : Internal_Entity_Record_Type_Def;
Let_Result : Internal_Entity_Base_Record_Def;
Scope_Result_1 : Internal_Entity_Base_Record_Def;
Cast_Result_1 : Internal_Entity_Derived_Type_Def;
Let_Result_1 : Internal_Entity_Base_Record_Def;
Scope_Result_2 : Internal_Entity_Base_Record_Def;
Cast_Result_2 : Internal_Entity_Type_Def;
Let_Result_2 : Internal_Entity_Base_Record_Def;
Scope_Result_3 : Internal_Entity_Base_Record_Def;
Match_Result : Internal_Entity_Base_Record_Def;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3983 '<Match at ast.py:6747>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6747
--# expr-start 3976 '<FieldAccess for type_def at ast.py:6747>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6747
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3976
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld; 
case Ada_Type_Def (Match_Prefix.Node.Kind) is
         when Ada_Record_Type_Def =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Record_Type_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind r Match_0
--# expr-start 3977 '<FieldAccess for record_def at ast.py:6748>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6748
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Base_Record_Def (Node => Match_0.Node.Record_Type_Def_F_Record_Def, Info => Match_0.Info);
--# expr-done 3977
Let_Result := Fld_1; 
--# end
Scope_Result_1 := Let_Result;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Derived_Type_Def =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Derived_Type_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind d Match_1
--# expr-start 3982 '<If at ast.py:6751>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6751
--# expr-start 3978 '<FieldAccess for is_tagged_type at ast.py:6752>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6752
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_tagged_type'
Fld_5 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Tagged_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 3978
if Fld_5 then
   --# expr-start 3979 '<FieldAccess for record_extension at ast.py:6753>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6753
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Base_Record_Def (Node => Match_1.Node.Derived_Type_Def_F_Record_Extension, Info => Match_1.Info);
--# expr-done 3979
   If_Result := Fld_2;
else
--# expr-start 3980 '<FieldAccess for base_type at ast.py:6754>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6754
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld_3 := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => Match_1.Node, Origin => Origin, E_Info => Match_1.Info);
--# end
--# expr-done 3980
Var_Expr := Fld_3;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 3981 '<FieldAccess for record_def at ast.py:6754>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6754
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.record_def'
Fld_4 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Record_Def (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 3981
Scope_Result := Fld_4;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Record_Def;
end if;
   If_Result := Result_Var;
end if;
--# expr-done 3982
Let_Result_1 := If_Result; 
--# end
Scope_Result_2 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Access_To_Subp_Def .. Ada_Array_Type_Def | Ada_Enum_Type_Def .. Ada_Ordinary_Fixed_Point_Def | Ada_Signed_Int_Type_Def =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Type_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind _ Match_2
Let_Result_2 := No_Entity_Base_Record_Def; 
--# end
Scope_Result_3 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_3; 
end case;
--# expr-done 3983
      Property_Result := Match_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Record_Def;
--# end
--# property-start TypeDecl.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6760
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Xref_Entry_Point
  (Node : Bare_Type_Decl
  )
   return Boolean
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Xref_Entry_Point;
--# end
--# property-start TypeDecl.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6763
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Xref_Equation
  (Node : Bare_Type_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_513 with Inline_Always;
      Fld : Internal_Entity_Type_Def;
Fld_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_513 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_513;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3986 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3985 '<FieldAccess for sub_equation at ast.py:6765>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6765
--# expr-start 3984 '<FieldAccess for type_def at ast.py:6765>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6765
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3984
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3985
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 3986
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_513;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_513;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Xref_Equation;
--# end
--# property-start TypeDecl.is_discrete_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6767
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Is_Discrete_Type
  (Node : Bare_Type_Decl
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Type_Def;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3988 '<FieldAccess for is_discrete_type at ast.py:6767>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6767
--# expr-start 3987 '<FieldAccess for type_def at ast.py:6767>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6767
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 3987
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.is_discrete_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Is_Discrete_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 3988
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Is_Discrete_Type;
--# end
--# property-start TypeDecl.own_primitives_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6770
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Own_Primitives_Env
  (Node : Bare_Type_Decl
      ; With_Rebindings : Env_Rebindings
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind with_rebindings With_Rebindings
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_515 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Rebound_Env : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_515 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Rebound_Env);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_515;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3991 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3990 '<rebind_env at ast.py:6775>' Rebound_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6775
--# expr-start 3989 '<FieldAccess for direct_primitives_env at ast.py:6775>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6775
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDecl.direct_primitives_env
Fld := Libadalang.Implementation.Type_Decl_F_Direct_Primitives_Env (Node => Ent.Node);
--# end
--# expr-done 3989
Rebound_Env := Rebind_Env (Fld, With_Rebindings); 
--# expr-done 3990
Let_Result := Rebound_Env; Inc_Ref (Let_Result);
--# end
--# expr-done 3991
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_515;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_515;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Own_Primitives_Env;
--# end
--# property-start TypeDecl.own_primitives_envs /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6778
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Own_Primitives_Envs
  (Node : Bare_Type_Decl
      ; With_Rebindings : Env_Rebindings
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env_Array_Access
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind with_rebindings With_Rebindings
   Property_Result : Lexical_Env_Array_Access;
            procedure Finalizer_Scope_516 with Inline_Always;
            procedure Finalizer_Scope_2304 with Inline_Always;
      Var_Expr : Internal_Entity_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity_Type_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Fld_2 : Lexical_Env := Empty_Env;
Array_Lit : Lexical_Env_Array_Access;
Scope_Result : Lexical_Env_Array_Access;
Fld_3 : Lexical_Env := Empty_Env;
Singleton : Lexical_Env_Array_Access;
Result_Var : Lexical_Env_Array_Access;
Let_Result : Lexical_Env_Array_Access;
            procedure Finalizer_Scope_516 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Singleton);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_516;
            procedure Finalizer_Scope_2304 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Array_Lit);
            end Finalizer_Scope_2304;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 3998 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 3993 '<Cast to TypeDecl at ast.py:6785>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6785
--# expr-start 3992 '<FieldAccess for previous_part at ast.py:6785>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6785
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.previous_part'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Previous_Part (Node => Ent.Node, Go_To_Incomplete => False, E_Info => Ent.Info);
--# end
--# expr-done 3992
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Type_Decl_Range
   then
      Cast_Result := Create_Internal_Entity_Type_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Type_Decl;
   end if;
--# expr-done 3993
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Type_Decl then
   --# scope-start
   --# bind pp Var_Expr
--# expr-start 3996 '<ArrayLiteral at ast.py:6786>' Array_Lit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6786
--# expr-start 3994 '<FieldAccess for own_primitives_env at ast.py:6787>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6787
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDecl.own_primitives_env
Fld_1 := Libadalang.Implementation.Type_Decl_P_Own_Primitives_Env (Node => Ent.Node, With_Rebindings => With_Rebindings, E_Info => Ent.Info);
--# end
--# expr-done 3994
--# expr-start 3995 '<FieldAccess for own_primitives_env at ast.py:6788>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6788
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDecl.own_primitives_env
Fld_2 := Libadalang.Implementation.Type_Decl_P_Own_Primitives_Env (Node => Var_Expr.Node, With_Rebindings => With_Rebindings, E_Info => Var_Expr.Info);
--# end
--# expr-done 3995
Array_Lit := Create_Lexical_Env_Array (Internal_Lexical_Env_Array'(1 => Fld_1, 2 => Fld_2)); 
--# expr-done 3996
Scope_Result := Array_Lit;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2304;
   Result_Var := Scope_Result;
else
   --# expr-start 3997 '<FieldAccess for own_primitives_env at ast.py:6790>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6790
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDecl.own_primitives_env
Fld_3 := Libadalang.Implementation.Type_Decl_P_Own_Primitives_Env (Node => Ent.Node, With_Rebindings => With_Rebindings, E_Info => Ent.Info);
--# end
--# expr-done 3997
                Singleton := Create_Lexical_Env_Array (Items_Count => 1);
                Singleton.Items (1) := Fld_3;
                Inc_Ref (Singleton.Items (1));
   Result_Var := Singleton;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 3998
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_516;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_516;
               Finalizer_Scope_2304;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Own_Primitives_Envs;
--# end
--# property-start TypeDecl.primitives_envs /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6794
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Primitives_Envs
  (Node : Bare_Type_Decl
      ; With_Rebindings : Env_Rebindings
      ; Stop_At : Internal_Entity_Base_Type_Decl_Array_Access
      ; Include_Self : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env_Array_Access
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind with_rebindings With_Rebindings
   --# bind stop_at Stop_At
   --# bind include_self Include_Self
   Property_Result : Lexical_Env_Array_Access;
            procedure Finalizer_Scope_517 with Inline_Always;
            procedure Finalizer_Scope_2305 with Inline_Always;
            procedure Finalizer_Scope_2309 with Inline_Always;
      Var_Expr : Internal_Entity_Type_Decl;
Match_0 : Internal_Entity_Type_Decl;
Match_1 : Internal_Entity_Base_Subtype_Decl;
Match_2 : Internal_Entity_Base_Type_Decl;
Cast_Expr : Bare_Type_Decl;
Cast_Result : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld : Internal_Entity_Base_Type_Decl_Array_Access;
Fld_1 : Bare_Base_Subtype_Decl;
Fld_2 : Bare_Ada_Node;
Bound_Origin_1 : Bare_Ada_Node;
Fld_3 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Type_Decl;
Match_Prefix : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity_Type_Decl;
Let_Result : Internal_Entity_Type_Decl;
Scope_Result : Internal_Entity_Type_Decl;
Cast_Result_3 : Internal_Entity_Base_Subtype_Decl;
Let_Result_1 : Internal_Entity_Type_Decl;
Scope_Result_1 : Internal_Entity_Type_Decl;
Cast_Result_4 : Internal_Entity_Base_Type_Decl;
Let_Result_2 : Internal_Entity_Type_Decl;
Scope_Result_2 : Internal_Entity_Type_Decl;
Match_Result : Internal_Entity_Type_Decl;
Fld_4 : Lexical_Env_Array_Access;
Fld_5 : Lexical_Env_Array_Access;
Concat_Result : Lexical_Env_Array_Access;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_5 : Internal_Entity;
Cast_Expr_2 : Internal_Entity_Type_Decl;
Cast_Result_6 : Internal_Entity;
Is_Equiv : Boolean;
Quantifier_Result : Boolean;
If_Result : Lexical_Env_Array_Access;
Scope_Result_3 : Lexical_Env_Array_Access;
Result_Var : Lexical_Env_Array_Access;
Map_Result : Lexical_Env_Array_Access;
Fld_6 : Lexical_Env_Array_Access;
If_Result_1 : Lexical_Env_Array_Access;
Concat_Result_1 : Lexical_Env_Array_Access;
Dyn_Var_Bind_Result_1 : Lexical_Env_Array_Access;
Let_Result_3 : Lexical_Env_Array_Access;
            procedure Finalizer_Scope_517 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Fld_6);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Concat_Result_1);
                     Dec_Ref (Dyn_Var_Bind_Result_1);
                     Dec_Ref (Let_Result_3);
            end Finalizer_Scope_517;
            procedure Finalizer_Scope_2305 is
            begin
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_2305;
            procedure Finalizer_Scope_2309 is
            begin
                     Dec_Ref (Fld_4);
                     Dec_Ref (Fld_5);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (If_Result);
            end Finalizer_Scope_2309;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4017 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 4016 '<bind at ast.py:6806>' Dyn_Var_Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6806
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Bound_Origin := Cast_Result; 
--# expr-start 4015 '<Concat at ast.py:6806>' Concat_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6806
--# expr-start 4012 '<Mapcat at ast.py:6806>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6806
   --# expr-start 3999 '<FieldAccess for base_types at ast.py:6806>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6806
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.base_types'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Base_Types (Node => Ent.Node, Origin => Bound_Origin, E_Info => Ent.Info);
--# end
--# expr-done 3999
      declare
         Map_Result_Vec : Lexical_Env_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Base_Type_Decl_Array_Access := Fld;
      begin
         for Item_154 of
               Collection.Items
         loop
   --# scope-start
               --# bind t Item_154
--# expr-start 4005 '<Match at ast.py:6806>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6806
      if Item_154.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Item_154; 
case Ada_Base_Type_Decl (Match_Prefix.Node.Kind) is
         when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_2; 
--# bind td Match_0
Let_Result := Match_0; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Discrete_Base_Subtype_Decl .. Ada_Subtype_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Base_Subtype_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_3; 
--# bind std Match_1
--# expr-start 4004 '<bind at ast.py:6808>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6808
--# expr-start 4001 '<FieldAccess for origin_node at ast.py:6809>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6809
--# expr-start 4000 '<FieldAccess for node at ast.py:6809>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6809
Fld_1 := Match_1.Node;
--# expr-done 4000
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_2 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Fld_1);
--# end
--# expr-done 4001
Bound_Origin_1 := Fld_2; 
--# expr-start 4003 '<Cast to TypeDecl at ast.py:6809>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6809
--# expr-start 4002 '<FieldAccess for get_type at ast.py:6809>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6809
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubtypeDecl.get_type'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Subtype_Decl_P_Get_Type (Node => Match_1.Node, Origin => Bound_Origin_1, E_Info => Match_1.Info);
--# end
--# expr-done 4002
   if Fld_3.Node = null
      or else Fld_3.Node.Kind in Ada_Type_Decl_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Type_Decl
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
   else
         Cast_Result_1 := No_Entity_Type_Decl;
   end if;
--# expr-done 4003
Dyn_Var_Bind_Result := Cast_Result_1; 
--# expr-done 4004
Let_Result_1 := Dyn_Var_Bind_Result; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Classwide_Type_Decl .. Ada_Single_Task_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Base_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_4; 
--# bind _ Match_2
Let_Result_2 := No_Entity_Type_Decl; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 4005
Var_Expr := Match_Result;
if Var_Expr /= No_Entity_Type_Decl then
   --# scope-start
   --# bind bt Var_Expr
--# expr-start 4011 '<If at ast.py:6813>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6813
--# expr-start 4007 '<Contains at ast.py:6814>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6814
      Quantifier_Result := False;
      declare
         Collection : constant Internal_Entity_Base_Type_Decl_Array_Access := Stop_At;
      begin
         for Item_155 of
               Collection.Items
         loop
   --# scope-start
               --# bind item Item_155
      --# expr-start 4006 '<Eq at ???>' Is_Equiv None
Cast_Expr_1 := Item_155; 
      Cast_Result_5 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Cast_Expr_2 := Var_Expr; 
      Cast_Result_6 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
Is_Equiv := Equivalent (Cast_Result_5, Cast_Result_6); 
--# expr-done 4006
      Quantifier_Result := Is_Equiv;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
--# expr-done 4007
if Quantifier_Result then
   If_Result := No_Lexical_Env_Array_Type;
else
   --# expr-start 4010 '<Concat at ast.py:6816>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6816
--# expr-start 4008 '<FieldAccess for own_primitives_envs at ast.py:6816>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6816
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDecl.own_primitives_envs
Fld_4 := Libadalang.Implementation.Type_Decl_P_Own_Primitives_Envs (Node => Var_Expr.Node, With_Rebindings => With_Rebindings, E_Info => Var_Expr.Info);
--# end
--# expr-done 4008
--# expr-start 4009 '<FieldAccess for primitives_envs at ast.py:6817>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6817
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDecl.primitives_envs
Fld_5 := Libadalang.Implementation.Type_Decl_P_Primitives_Envs (Node => Var_Expr.Node, With_Rebindings => With_Rebindings, Stop_At => Stop_At, Include_Self => False, E_Info => Var_Expr.Info);
--# end
--# expr-done 4009
Concat_Result := Concat (Fld_4, Fld_5); 
--# expr-done 4010
   If_Result := Concat_Result;
end if;
      Inc_Ref (If_Result);
--# expr-done 4011
Scope_Result_3 := If_Result;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_2309;
   Result_Var := Scope_Result_3;
else
   Result_Var := No_Lexical_Env_Array_Type;
end if;
      Inc_Ref (Result_Var);
         for Item_To_Append of
               Result_Var.Items
         loop
               Inc_Ref (Item_To_Append);
            Lexical_Env_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_2305;
         end loop;
      end;
         Map_Result := Create_Lexical_Env_Array
           (Items_Count => Natural (Lexical_Env_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Lexical_Env_Vectors.Get
              (Map_Result_Vec,
               I + Lexical_Env_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Lexical_Env_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 4012
--# expr-start 4014 '<If at ast.py:6820>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6820
if Include_Self then
   --# expr-start 4013 '<FieldAccess for own_primitives_envs at ast.py:6821>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6821
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDecl.own_primitives_envs
Fld_6 := Libadalang.Implementation.Type_Decl_P_Own_Primitives_Envs (Node => Ent.Node, With_Rebindings => With_Rebindings, E_Info => Ent.Info);
--# end
--# expr-done 4013
   If_Result_1 := Fld_6;
else
   If_Result_1 := No_Lexical_Env_Array_Type;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 4014
Concat_Result_1 := Concat (Map_Result, If_Result_1); 
--# expr-done 4015
Dyn_Var_Bind_Result_1 := Concat_Result_1; Inc_Ref (Dyn_Var_Bind_Result_1);
--# expr-done 4016
Let_Result_3 := Dyn_Var_Bind_Result_1; Inc_Ref (Let_Result_3);
--# end
--# expr-done 4017
      Property_Result := Let_Result_3;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_517;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_517;
               Finalizer_Scope_2305;
               Finalizer_Scope_2309;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Primitives_Envs;
--# end
--# property-start TypeDecl.compute_primitives_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6826
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Compute_Primitives_Env
  (Node : Bare_Type_Decl
      ; Include_Self : Boolean
         := True
      ; Stop_At : Internal_Entity_Base_Type_Decl_Array_Access
         := No_Internal_Entity_Base_Type_Decl_Array_Type
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind include_self Include_Self
   --# bind stop_at Stop_At
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_518 with Inline_Always;
      Fld : Internal_Entity_Info;
Fld_1 : Env_Rebindings;
Fld_2 : Lexical_Env_Array_Access;
Cast_Expr : Bare_Type_Decl;
Cast_Result : Bare_Ada_Node;
New_Struct : Internal_Metadata;
Group_Env : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_518 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Group_Env);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_518;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Type_Decl_Type_Decl_P_Compute_Primitives_Env,
            Items    => new Mmz_Key_Array (1 ..  4))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Type_Decl,
                                As_Bare_Type_Decl => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Boolean,
                                      As_Boolean => Include_Self);
               Mmz_K.Items (3) := (Kind => Mmz_Internal_Entity_Base_Type_Decl_Array_Access,
                                      As_Internal_Entity_Base_Type_Decl_Array_Access => Stop_At);
                  Inc_Ref (Mmz_K.Items (3).As_Internal_Entity_Base_Type_Decl_Array_Access);
               Mmz_K.Items (4) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Lexical_Env;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 4023 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4022 '<env_group at ast.py:6834>' Group_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6834
--# expr-start 4020 '<FieldAccess for primitives_envs at ast.py:6834>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6834
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4019 '<FieldAccess for rebindings at ast.py:6835>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6835
--# expr-start 4018 '<FieldAccess for info at ast.py:6835>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6835
Fld := Ent.Info;
--# expr-done 4018
Fld_1 := Fld.Rebindings;
--# expr-done 4019
--# property-call-start TypeDecl.primitives_envs
Fld_2 := Libadalang.Implementation.Type_Decl_P_Primitives_Envs (Node => Ent.Node, With_Rebindings => Fld_1, Stop_At => Stop_At, Include_Self => Include_Self, E_Info => Ent.Info);
--# end
--# expr-done 4020
--# expr-start 4021 '<New Metadata at ast.py:6839>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6839
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
New_Struct := (Dottable_Subp => False, Primitive => No_Bare_Ada_Node, Primitive_Real_Type => Cast_Result); 
--# expr-done 4021
Group_Env := Group (Fld_2, New_Struct); 
--# expr-done 4022
Let_Result := Group_Env; Inc_Ref (Let_Result);
--# end
--# expr-done 4023
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_518;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Lexical_Env,
                        As_Lexical_Env => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_518;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Compute_Primitives_Env;
--# end
--# property-start TypeDecl.parent_primitives_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6843
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Parent_Primitives_Env
  (Node : Bare_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_519 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Fld_1 : Lexical_Env := Empty_Env;
Fld_2 : Bare_Type_Def;
Is_A : Boolean;
If_Result : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_519 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_1);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_519;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4029 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4028 '<If at ast.py:6844>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6844
--# expr-start 4025 '<IsA DerivedTypeDef, InterfaceTypeDef at ast.py:6845>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6845
--# expr-start 4024 '<FieldAccess for type_def at ast.py:6845>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6845
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Type_Decl_F_Type_Def;
--# expr-done 4024
Is_A := Fld_2 /= null 
and then Fld_2.Kind in Ada_Derived_Type_Def_Range | Ada_Interface_Type_Def_Range; 
--# expr-done 4025
if Is_A then
   --# expr-start 4026 '<FieldAccess for compute_primitives_env at ast.py:6846>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6846
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDecl.compute_primitives_env
Fld := Libadalang.Implementation.Type_Decl_P_Compute_Primitives_Env (Node => Ent.Node, Include_Self => False, E_Info => Ent.Info);
--# end
--# expr-done 4026
   If_Result := Fld;
else
   --# expr-start 4027 '<FieldAccess for empty_env at ast.py:6847>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6847
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.empty_env
Fld_1 := Libadalang.Implementation.Ada_Node_P_Empty_Env (Node => Self);
--# end
--# expr-done 4027
   If_Result := Fld_1;
end if;
      Inc_Ref (If_Result);
--# expr-done 4028
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 4029
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_519;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_519;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Parent_Primitives_Env;
--# end
--# property-start TypeDecl.refined_parent_primitives_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6851
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Refined_Parent_Primitives_Env
  (Node : Bare_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_520 with Inline_Always;
            procedure Finalizer_Scope_2311 with Inline_Always;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl_Array_Access;
Scope_Result : Internal_Entity_Base_Type_Decl_Array_Access;
Result_Var : Internal_Entity_Base_Type_Decl_Array_Access;
Fld_2 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_520 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_520;
            procedure Finalizer_Scope_2311 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2311;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4033 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4032 '<FieldAccess for compute_primitives_env at ast.py:6872>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6872
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4030 '<FieldAccess for previous_part at ast.py:6874>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6874
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.previous_part'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Previous_Part (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4030
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 4031 '<FieldAccess for base_types at ast.py:6874>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6874
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.base_types'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Base_Types (Node => Var_Expr.Node, Origin => No_Bare_Ada_Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 4031
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2311;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Base_Type_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var);
--# property-call-start TypeDecl.compute_primitives_env
Fld_2 := Libadalang.Implementation.Type_Decl_P_Compute_Primitives_Env (Node => Ent.Node, Include_Self => False, Stop_At => Result_Var, E_Info => Ent.Info);
--# end
--# expr-done 4032
Let_Result := Fld_2; Inc_Ref (Let_Result);
--# end
--# expr-done 4033
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_520;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_520;
               Finalizer_Scope_2311;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Refined_Parent_Primitives_Env;
--# end
--# property-start TypeDecl.primitives_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6878
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Primitives_Env
  (Node : Bare_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_521 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_521 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_521;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4035 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4034 '<FieldAccess for compute_primitives_env at ast.py:6879>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6879
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDecl.compute_primitives_env
Fld := Libadalang.Implementation.Type_Decl_P_Compute_Primitives_Env (Node => Ent.Node, Include_Self => True, E_Info => Ent.Info);
--# end
--# expr-done 4034
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 4035
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_521;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_521;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Primitives_Env;
--# end
--# property-start TypeDecl.get_primitives /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6882
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Get_Primitives
  (Node : Bare_Type_Decl
      ; Only_Inherited : Boolean
         := False
      ; Include_Predefined_Operators : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl_Array_Access
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind only_inherited Only_Inherited
   --# bind include_predefined_operators Include_Predefined_Operators
   Property_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_522 with Inline_Always;
      Block_Var : Lexical_Env := Empty_Env;
Block_Var_1 : Internal_Entity_Basic_Decl_Array_Access;
Block_Var_2 : Internal_Entity_Basic_Decl_Array_Access;
A_Spec : Internal_Entity_Base_Subp_Spec;
A_Prim : Internal_Entity_Base_Type_Decl;
B_Prim : Bare_Base_Type_Decl;
B_Prim_Ent : Internal_Entity_Base_Type_Decl;
Fld : Lexical_Env := Empty_Env;
Fld_1 : Lexical_Env := Empty_Env;
If_Result : Lexical_Env := Empty_Env;
Env_Get_Result : Internal_Entity_Array_Access;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Map_Result : Internal_Entity_Basic_Decl_Array_Access;
Is_A : Boolean;
Not_Val : Boolean;
Map_Result_1 : Internal_Entity_Basic_Decl_Array_Access;
If_Result_1 : Internal_Entity_Basic_Decl_Array_Access;
Index_159 : Integer;
Fld_2 : Internal_Entity_Base_Subp_Spec;
Fld_3 : Internal_Entity_Info;
Fld_4 : Internal_Metadata;
Fld_5 : Bare_Ada_Node;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result_2 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Base_Type_Decl;
Fld_6 : Internal_Entity_Info;
Fld_7 : Internal_Metadata;
Fld_8 : Bare_Ada_Node;
Cast_Result_2 : Bare_Base_Type_Decl;
Fld_9 : Internal_Entity_Defining_Name;
Fld_10 : Bare_Defining_Name;
Fld_11 : Internal_Entity_Defining_Name;
Fld_12 : Bare_Defining_Name;
Cast_Result_3 : Bare_Name;
Fld_13 : Boolean;
Not_Val_1 : Boolean;
Fld_14 : Internal_Entity_Base_Subp_Spec;
Fld_15 : Boolean;
Not_Val_2 : Boolean;
If_Result_3 : Boolean;
As_Entity_1 : Internal_Entity_Base_Type_Decl;
Is_Null_1 : Boolean;
If_Result_4 : Internal_Entity_Base_Type_Decl;
Comp_Result : Boolean;
Fld_16 : Bare_Base_Type_Decl;
Fld_17 : Boolean;
Not_Val_3 : Boolean;
If_Result_5 : Boolean;
Comp_Result_1 : Boolean;
Fld_18 : Bare_Base_Type_Decl;
Fld_19 : Boolean;
Not_Val_4 : Boolean;
If_Result_6 : Boolean;
Fld_20 : Bare_Base_Type_Decl;
Fld_21 : Boolean;
If_Result_7 : Boolean;
Let_Result : Boolean;
If_Result_8 : Boolean;
Let_Result_1 : Boolean;
Index_161 : Integer;
Quantifier_Result : Boolean;
Let_Result_2 : Boolean;
Map_Result_2 : Internal_Entity_Basic_Decl_Array_Access;
Let_Result_3 : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_522 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_1);
                     Dec_Ref (If_Result);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Env_Get_Result);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Block_Var_2);
                     Dec_Ref (Map_Result_2);
                     Dec_Ref (Let_Result_3);
            end Finalizer_Scope_522;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4080 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 4038 '<If at ast.py:6891>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6891
if Only_Inherited then
   --# expr-start 4036 '<FieldAccess for parent_primitives_env at ast.py:6893>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6893
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDecl.parent_primitives_env
Fld := Libadalang.Implementation.Type_Decl_P_Parent_Primitives_Env (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4036
   If_Result := Fld;
else
   --# expr-start 4037 '<FieldAccess for primitives_env at ast.py:6894>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6894
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.primitives_env'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Primitives_Env (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4037
   If_Result := Fld_1;
end if;
      Inc_Ref (If_Result);
--# expr-done 4038
Block_Var := If_Result; Inc_Ref (Block_Var);
--# bind prim_env Block_Var
--# expr-start 4041 '<Map at ast.py:6897>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6897
   --# expr-start 4039 '<EnvGet at ast.py:6897>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6897
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Block_Var, Key => null, Lookup_Kind => To_Lookup_Kind_Type (Recursive), Categories => (Inherited_Primitives => True, Nocat => True))); 
--# expr-done 4039
      declare
         Map_Result_Vec : Internal_Entity_Basic_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Env_Get_Result;
      begin
         for Item_156 of
               Collection.Items
         loop
   --# scope-start
               --# bind t Item_156
      --# expr-start 4040 '<Cast to BasicDecl at ast.py:6898>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6898
Cast_Expr := Item_156; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 4040
         declare
            Item_To_Append : constant Internal_Entity_Basic_Decl := Cast_Result;
         begin
            Internal_Entity_Basic_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Basic_Decl_Array
           (Items_Count => Natural (Internal_Entity_Basic_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Basic_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Basic_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Basic_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 4041
Block_Var_1 := Map_Result; Inc_Ref (Block_Var_1);
--# bind all_prims Block_Var_1
--# expr-start 4045 '<If at ast.py:6901>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6901
if Include_Predefined_Operators then
   If_Result_1 := Block_Var_1;
else
   --# expr-start 4044 '<Filter at ast.py:6904>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6904
      declare
         Map_Result_1_Vec : Internal_Entity_Basic_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Basic_Decl_Array_Access := Block_Var_1;
      begin
         for Item_157 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_157
         --# expr-start 4043 '<Not at ast.py:6904>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6904
--# expr-start 4042 '<IsA SyntheticSubpDecl at ast.py:6904>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6904
Is_A := Item_157.Node /= null 
and then Item_157.Node.Kind in Ada_Synthetic_Subp_Decl_Range; 
--# expr-done 4042
Not_Val := not (Is_A); 
--# expr-done 4043
         if Not_Val then
         declare
            Item_To_Append : constant Internal_Entity_Basic_Decl := Item_157;
         begin
            Internal_Entity_Basic_Decl_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result_1 := Create_Internal_Entity_Basic_Decl_Array
           (Items_Count => Natural (Internal_Entity_Basic_Decl_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Entity_Basic_Decl_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Entity_Basic_Decl_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Entity_Basic_Decl_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 4044
   If_Result_1 := Map_Result_1;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 4045
Block_Var_2 := If_Result_1; Inc_Ref (Block_Var_2);
--# bind bds Block_Var_2
--# expr-start 4079 '<Filter at ast.py:6909>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6909
      declare
         Map_Result_2_Vec : Internal_Entity_Basic_Decl_Vectors.Vector;
      begin
         Index_159 := 0;
      declare
         Collection : constant Internal_Entity_Basic_Decl_Array_Access := Block_Var_2;
      begin
         for Item_158 of
               Collection.Items
         loop
   --# scope-start
               --# bind a Item_158
               --# bind i Index_159
         --# expr-start 4078 '<Let at ast.py:6909>' Let_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6909
--# scope-start
--# expr-start 4046 '<FieldAccess for subp_spec_or_null at ast.py:6911>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6911
      if Item_158.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Item_158.Node, E_Info => Item_158.Info);
--# end
--# expr-done 4046
A_Spec := Fld_2; 
--# bind a_spec A_Spec
--# expr-start 4051 '<Cast to BaseTypeDecl at ast.py:6912>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6912
--# expr-start 4050 '<as_bare_entity at ast.py:6912>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6912
--# expr-start 4049 '<FieldAccess for primitive at ast.py:6912>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6912
--# expr-start 4048 '<FieldAccess for md at ast.py:6912>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6912
--# expr-start 4047 '<FieldAccess for info at ast.py:6912>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6912
Fld_3 := Item_158.Info;
--# expr-done 4047
Fld_4 := Fld_3.Md;
--# expr-done 4048
Fld_5 := Fld_4.Primitive;
--# expr-done 4049
Is_Null := Fld_5 = null; 
if Is_Null then
   If_Result_2 := No_Entity;
else
As_Entity := (Info => No_Entity_Info, Node => Fld_5); 
   If_Result_2 := As_Entity;
end if;
--# expr-done 4050
   if If_Result_2.Node = null
      or else If_Result_2.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Base_Type_Decl
        (Node => If_Result_2.Node,
         Info => If_Result_2.Info);
   else
         Cast_Result_1 := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 4051
A_Prim := Cast_Result_1; 
--# bind a_prim A_Prim
--# expr-start 4077 '<AllQuantifier at ast.py:6914>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6914
      Quantifier_Result := True;
         Index_161 := 0;
      declare
         Collection : constant Internal_Entity_Basic_Decl_Array_Access := Block_Var_2;
      begin
         for Item_160 of
               Collection.Items
         loop
   --# scope-start
               --# bind b Item_160
               --# bind j Index_161
      --# expr-start 4076 '<Let at ast.py:6914>' Let_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6914
--# scope-start
--# expr-start 4055 '<Cast to BaseTypeDecl at ast.py:6915>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6915
--# expr-start 4054 '<FieldAccess for primitive at ast.py:6915>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6915
--# expr-start 4053 '<FieldAccess for md at ast.py:6915>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6915
--# expr-start 4052 '<FieldAccess for info at ast.py:6915>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6915
Fld_6 := Item_160.Info;
--# expr-done 4052
Fld_7 := Fld_6.Md;
--# expr-done 4053
Fld_8 := Fld_7.Primitive;
--# expr-done 4054
   if Fld_8 = null
      or else Fld_8.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_2 := Fld_8;
   else
         Cast_Result_2 := No_Bare_Ada_Node;
   end if;
--# expr-done 4055
B_Prim := Cast_Result_2; 
--# bind b_prim B_Prim
--# expr-start 4061 '<Not at ast.py:6922>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6922
--# expr-start 4060 '<FieldAccess for matches at ast.py:6922>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6922
--# expr-start 4057 '<FieldAccess for node at ast.py:6922>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6922
--# expr-start 4056 '<FieldAccess for defining_name at ast.py:6922>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6922
      if Item_158.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_9 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Item_158.Node, E_Info => Item_158.Info);
--# end
--# expr-done 4056
Fld_10 := Fld_9.Node;
--# expr-done 4057
      if Fld_10 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4059 '<FieldAccess for node at ast.py:6922>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6922
--# expr-start 4058 '<FieldAccess for defining_name at ast.py:6922>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6922
      if Item_160.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_11 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Item_160.Node, E_Info => Item_160.Info);
--# end
--# expr-done 4058
Fld_12 := Fld_11.Node;
--# expr-done 4059
      Cast_Result_3 := Fld_12;
--# property-call-start Name.matches
Fld_13 := Libadalang.Implementation.Name_P_Matches (Node => Fld_10, N => Cast_Result_3);
--# end
--# expr-done 4060
Not_Val_1 := not (Fld_13); 
--# expr-done 4061
if Not_Val_1 then
   If_Result_3 := True;
else
   --# expr-start 4064 '<Not at ast.py:6925>' Not_Val_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6925
--# expr-start 4063 '<FieldAccess for match_signature at ast.py:6925>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6925
      if A_Spec.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4062 '<FieldAccess for subp_spec_or_null at ast.py:6926>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6926
      if Item_160.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_14 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Item_160.Node, E_Info => Item_160.Info);
--# end
--# expr-done 4062
--# property-call-start BaseSubpSpec.match_signature
Fld_15 := Libadalang.Implementation.Base_Subp_Spec_P_Match_Signature (Node => A_Spec.Node, Other => Fld_14, Match_Name => False, Use_Entity_Info => True, E_Info => A_Spec.Info);
--# end
--# expr-done 4063
Not_Val_2 := not (Fld_15); 
--# expr-done 4064
   If_Result_3 := Not_Val_2;
end if;
if If_Result_3 then
   If_Result_8 := True;
else
   --# expr-start 4075 '<Let at ast.py:6930>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6930
--# scope-start
--# expr-start 4065 '<as_bare_entity at ast.py:6930>' If_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6930
Is_Null_1 := B_Prim = null; 
if Is_Null_1 then
   If_Result_4 := No_Entity_Base_Type_Decl;
else
As_Entity_1 := (Info => No_Entity_Info, Node => B_Prim); 
   If_Result_4 := As_Entity_1;
end if;
--# expr-done 4065
B_Prim_Ent := If_Result_4; 
--# bind b_prim_ent B_Prim_Ent
--# expr-start 4074 '<If at ast.py:6930>' If_Result_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6930
--# expr-start 4067 '<FieldAccess for has_base_type at ast.py:6933>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6933
      if A_Prim.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4066 '<FieldAccess for node at ast.py:6933>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6933
Fld_20 := B_Prim_Ent.Node;
--# expr-done 4066
--# property-call-start BaseTypeDecl.has_base_type
Fld_21 := Libadalang.Implementation.Base_Type_Decl_P_Has_Base_Type (Node => A_Prim.Node, Target => Fld_20, E_Info => A_Prim.Info);
--# end
--# expr-done 4067
if Fld_21 then
Comp_Result := Index_159 >= Index_161; 
if Comp_Result then
   If_Result_5 := True;
else
   --# expr-start 4070 '<Not at ast.py:6945>' Not_Val_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6945
--# expr-start 4069 '<FieldAccess for has_base_type at ast.py:6945>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6945
      if B_Prim_Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4068 '<FieldAccess for node at ast.py:6945>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6945
Fld_16 := A_Prim.Node;
--# expr-done 4068
--# property-call-start BaseTypeDecl.has_base_type
Fld_17 := Libadalang.Implementation.Base_Type_Decl_P_Has_Base_Type (Node => B_Prim_Ent.Node, Target => Fld_16, E_Info => B_Prim_Ent.Info);
--# end
--# expr-done 4069
Not_Val_3 := not (Fld_17); 
--# expr-done 4070
   If_Result_5 := Not_Val_3;
end if;
   If_Result_7 := If_Result_5;
else
Comp_Result_1 := Index_159 <= Index_161; 
if Comp_Result_1 then
   --# expr-start 4073 '<Not at ast.py:6957>' Not_Val_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6957
--# expr-start 4072 '<FieldAccess for has_base_type at ast.py:6957>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6957
      if B_Prim_Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4071 '<FieldAccess for node at ast.py:6957>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6957
Fld_18 := A_Prim.Node;
--# expr-done 4071
--# property-call-start BaseTypeDecl.has_base_type
Fld_19 := Libadalang.Implementation.Base_Type_Decl_P_Has_Base_Type (Node => B_Prim_Ent.Node, Target => Fld_18, E_Info => B_Prim_Ent.Info);
--# end
--# expr-done 4072
Not_Val_4 := not (Fld_19); 
--# expr-done 4073
   If_Result_6 := Not_Val_4;
else
   If_Result_6 := False;
end if;
   If_Result_7 := If_Result_6;
end if;
--# expr-done 4074
Let_Result := If_Result_7; 
--# end
--# expr-done 4075
   If_Result_8 := Let_Result;
end if;
Let_Result_1 := If_Result_8; 
--# end
--# expr-done 4076
      Quantifier_Result := Let_Result_1;
   --# end
               exit when not Quantifier_Result;
               Index_161 := Index_161 + 1;
         end loop;
      end;
--# expr-done 4077
Let_Result_2 := Quantifier_Result; 
--# end
--# expr-done 4078
         if Let_Result_2 then
         declare
            Item_To_Append : constant Internal_Entity_Basic_Decl := Item_158;
         begin
            Internal_Entity_Basic_Decl_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
         end if;
   --# end
               Index_159 := Index_159 + 1;
         end loop;
      end;
         Map_Result_2 := Create_Internal_Entity_Basic_Decl_Array
           (Items_Count => Natural (Internal_Entity_Basic_Decl_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Internal_Entity_Basic_Decl_Vectors.Get
              (Map_Result_2_Vec,
               I + Internal_Entity_Basic_Decl_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Internal_Entity_Basic_Decl_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 4079
Let_Result_3 := Map_Result_2; Inc_Ref (Let_Result_3);
--# end
--# expr-done 4080
      Property_Result := Let_Result_3;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_522;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_522;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Get_Primitives;
--# end
--# property-start TypeDecl.get_imp_deref /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6963
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Get_Imp_Deref
  (Node : Bare_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Expr
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Expr;
      Sym : Symbol_Type;
Fld : Internal_Entity_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4082 '<FieldAccess for get_aspect_spec_expr at ast.py:6964>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6964
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4081 '<Symbol Implicit_Dereference at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Implicit_Dereference); 
--# expr-done 4081
--# property-call-start BasicDecl.get_aspect_spec_expr
Fld := Libadalang.Implementation.Basic_Decl_P_Get_Aspect_Spec_Expr (Node => Ent.Node, Name => Sym, E_Info => Ent.Info);
--# end
--# expr-done 4082
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Get_Imp_Deref;
--# end
--# property-start TypeDecl.has_ud_indexing /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6967
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Has_Ud_Indexing
  (Node : Bare_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Sym : Symbol_Type;
Fld : Internal_Entity_Expr;
Is_Null : Boolean;
Not_Val : Boolean;
Sym_1 : Symbol_Type;
Fld_1 : Internal_Entity_Expr;
Is_Null_1 : Boolean;
Not_Val_1 : Boolean;
If_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
--# expr-start 4086 '<Not at ast.py:6968>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6968
--# expr-start 4085 '<IsNull at ast.py:6968>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6968
--# expr-start 4084 '<FieldAccess for get_aspect_spec_expr at ast.py:6968>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6968
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4083 '<Symbol Constant_Indexing at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Constant_Indexing); 
--# expr-done 4083
--# property-call-start BasicDecl.get_aspect_spec_expr
Fld := Libadalang.Implementation.Basic_Decl_P_Get_Aspect_Spec_Expr (Node => Ent.Node, Name => Sym, E_Info => Ent.Info);
--# end
--# expr-done 4084
Is_Null := Fld.Node = null; 
--# expr-done 4085
Not_Val := not (Is_Null); 
--# expr-done 4086
if Not_Val then
   If_Result := True;
else
   --# expr-start 4090 '<Not at ast.py:6969>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6969
--# expr-start 4089 '<IsNull at ast.py:6969>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6969
--# expr-start 4088 '<FieldAccess for get_aspect_spec_expr at ast.py:6969>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6969
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4087 '<Symbol Variable_Indexing at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Variable_Indexing); 
--# expr-done 4087
--# property-call-start BasicDecl.get_aspect_spec_expr
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Get_Aspect_Spec_Expr (Node => Ent.Node, Name => Sym_1, E_Info => Ent.Info);
--# end
--# expr-done 4088
Is_Null_1 := Fld_1.Node = null; 
--# expr-done 4089
Not_Val_1 := not (Is_Null_1); 
--# expr-done 4090
   If_Result := Not_Val_1;
end if;
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Has_Ud_Indexing;
--# end
--# property-start TypeDecl.constant_indexing_fns /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6973
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Constant_Indexing_Fns
  (Node : Bare_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl_Array_Access
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_525 with Inline_Always;
            procedure Finalizer_Scope_2316 with Inline_Always;
            procedure Finalizer_Scope_2318 with Inline_Always;
      Var_Expr : Internal_Entity_Expr;
Var_Expr_1 : Internal_Entity_Base_Subp_Spec;
Var_Expr_2 : Internal_Entity_Defining_Name;
Sym : Symbol_Type;
Fld : Internal_Entity_Expr;
Cast_Expr : Internal_Entity_Expr;
Cast_Result : Internal_Entity_Name;
Fld_1 : Internal_Entity_Array_Access;
Scope_Result : Internal_Entity_Array_Access;
Result_Var : Internal_Entity_Array_Access;
Cast_Expr_1 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Basic_Decl;
Cast_Expr_2 : Internal_Entity;
Cast_Result_2 : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity_Base_Subp_Spec;
Fld_3 : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_4 : Internal_Entity_Defining_Name_Array_Access;
Get_Result : Internal_Entity_Defining_Name;
Fld_5 : Internal_Entity_Base_Formal_Param_Decl;
Fld_6 : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Result_Var_1 : Internal_Entity_Base_Type_Decl;
Cast_Expr_3 : Internal_Entity_Type_Decl;
Cast_Result_3 : Internal_Entity_Base_Type_Decl;
Fld_7 : Boolean;
Dyn_Var_Bind_Result : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
Map_Result : Internal_Entity_Basic_Decl_Array_Access;
Let_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_525 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_525;
            procedure Finalizer_Scope_2316 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2316;
            procedure Finalizer_Scope_2318 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_2318;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4106 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4105 '<FilterMap at ast.py:6975>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6975
--# expr-start 4092 '<FieldAccess for get_aspect_spec_expr at ast.py:6975>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6975
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4091 '<Symbol Constant_Indexing at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Constant_Indexing); 
--# expr-done 4091
--# property-call-start BasicDecl.get_aspect_spec_expr
Fld := Libadalang.Implementation.Basic_Decl_P_Get_Aspect_Spec_Expr (Node => Ent.Node, Name => Sym, E_Info => Ent.Info);
--# end
--# expr-done 4092
Var_Expr := Fld;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
--# expr-start 4094 '<FieldAccess for all_env_elements_internal at ast.py:6975>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6975
--# expr-start 4093 '<Cast to Name at ast.py:6975>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6975
Cast_Expr := Var_Expr; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Name
   then
      Cast_Result := Create_Internal_Entity_Name
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 4093
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.all_env_elements_internal
Fld_1 := Libadalang.Implementation.Name_P_All_Env_Elements_Internal (Node => Cast_Result.Node, Seq => False, E_Info => Cast_Result.Info);
--# end
--# expr-done 4094
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2316;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Array_Type;
end if;
      Inc_Ref (Result_Var);
      declare
         Map_Result_Vec : Internal_Entity_Basic_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Result_Var;
      begin
         for Item_162 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_162
--# expr-start 4096 '<FieldAccess for subp_spec_or_null at ast.py:6980>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6980
--# expr-start 4095 '<Cast to BasicDecl at ast.py:6980>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6980
Cast_Expr_2 := Item_162; 
   if Cast_Expr_2.Node = null
      or else Cast_Expr_2.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_2 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 4095
      if Cast_Result_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Cast_Result_2.Node, E_Info => Cast_Result_2.Info);
--# end
--# expr-done 4096
Var_Expr_1 := Fld_2;
if Var_Expr_1 /= No_Entity_Base_Subp_Spec then
   --# scope-start
   --# bind ss Var_Expr_1
--# expr-start 4103 '<bind at ast.py:6982>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6982
--# expr-start 4097 '<FieldAccess for origin_node at ast.py:6983>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6983
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_3 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 4097
Bound_Origin := Fld_3; 
--# expr-start 4102 '<FieldAccess for matching_formal_type at ast.py:6984>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6984
--# expr-start 4099 '<at at ast.py:6984>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6984
--# expr-start 4098 '<FieldAccess for unpacked_formal_params at ast.py:6984>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6984
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld_4 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 4098
Get_Result := Get (Fld_4, 0, True); 
--# expr-done 4099
Var_Expr_2 := Get_Result;
if Var_Expr_2 /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 4101 '<FieldAccess for formal_type at ast.py:6984>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6984
--# expr-start 4100 '<FieldAccess for formal_decl at ast.py:6984>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6984
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_5 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Var_Expr_2.Node, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 4100
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamDecl.formal_type
Fld_6 := Libadalang.Implementation.Base_Formal_Param_Decl_P_Formal_Type (Node => Fld_5.Node, Origin => Bound_Origin, E_Info => Fld_5.Info);
--# end
--# expr-done 4101
Scope_Result_1 := Fld_6;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Base_Type_Decl;
end if;
      if Result_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Cast_Expr_3 := Ent; 
      Cast_Result_3 := Create_Internal_Entity_Base_Type_Decl
        (Node => Cast_Expr_3.Node,
         Info => Cast_Expr_3.Info);
--# property-call-start BaseTypeDecl.matching_formal_type
Fld_7 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Type (Node => Result_Var_1.Node, Formal_Type => Cast_Result_3, Origin => Bound_Origin, E_Info => Result_Var_1.Info);
--# end
--# expr-done 4102
Dyn_Var_Bind_Result := Fld_7; 
--# expr-done 4103
Scope_Result_2 := Dyn_Var_Bind_Result;
   --# end
      Finalizer_Scope_2318;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := False;
end if;
         if Result_Var_2 then
      --# expr-start 4104 '<Cast to BasicDecl at ast.py:6978>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6978
Cast_Expr_1 := Item_162; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_1 := No_Entity_Basic_Decl;
   end if;
--# expr-done 4104
         declare
            Item_To_Append : constant Internal_Entity_Basic_Decl := Cast_Result_1;
         begin
            Internal_Entity_Basic_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Basic_Decl_Array
           (Items_Count => Natural (Internal_Entity_Basic_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Basic_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Basic_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Basic_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 4105
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 4106
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_525;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_525;
               Finalizer_Scope_2316;
               Finalizer_Scope_2318;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Constant_Indexing_Fns;
--# end
--# property-start TypeDecl.variable_indexing_fns /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6992
pragma Warnings (Off, "is not referenced");
 function Type_Decl_P_Variable_Indexing_Fns
  (Node : Bare_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl_Array_Access
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_526 with Inline_Always;
            procedure Finalizer_Scope_2320 with Inline_Always;
            procedure Finalizer_Scope_2322 with Inline_Always;
      Var_Expr : Internal_Entity_Expr;
Var_Expr_1 : Internal_Entity_Base_Subp_Spec;
Var_Expr_2 : Internal_Entity_Defining_Name;
Fld : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Sym : Symbol_Type;
Fld_1 : Internal_Entity_Expr;
Cast_Expr : Internal_Entity_Expr;
Cast_Result : Internal_Entity_Name;
Fld_2 : Internal_Entity_Array_Access;
Cast_Expr_1 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Basic_Decl;
Cast_Expr_2 : Internal_Entity;
Cast_Result_2 : Internal_Entity_Basic_Decl;
Fld_3 : Internal_Entity_Base_Subp_Spec;
Fld_4 : Internal_Entity_Defining_Name_Array_Access;
Get_Result : Internal_Entity_Defining_Name;
Fld_5 : Internal_Entity_Base_Formal_Param_Decl;
Fld_6 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Cast_Expr_3 : Internal_Entity_Type_Decl;
Cast_Result_3 : Internal_Entity_Base_Type_Decl;
Fld_7 : Boolean;
Fld_8 : Internal_Entity_Base_Type_Decl;
Fld_9 : Boolean;
If_Result : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
Map_Result : Internal_Entity_Basic_Decl_Array_Access;
Scope_Result_2 : Internal_Entity_Basic_Decl_Array_Access;
Result_Var_2 : Internal_Entity_Basic_Decl_Array_Access;
Dyn_Var_Bind_Result : Internal_Entity_Basic_Decl_Array_Access;
Let_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_526 is
            begin
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Result_Var_2);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_526;
            procedure Finalizer_Scope_2320 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_2320;
            procedure Finalizer_Scope_2322 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_2322;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4124 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4123 '<bind at ast.py:6993>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6993
--# expr-start 4107 '<FieldAccess for origin_node at ast.py:6994>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6994
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 4107
Bound_Origin := Fld; 
--# expr-start 4109 '<FieldAccess for get_aspect_spec_expr at ast.py:6995>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6995
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4108 '<Symbol Variable_Indexing at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Variable_Indexing); 
--# expr-done 4108
--# property-call-start BasicDecl.get_aspect_spec_expr
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Get_Aspect_Spec_Expr (Node => Ent.Node, Name => Sym, E_Info => Ent.Info);
--# end
--# expr-done 4109
Var_Expr := Fld_1;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind a Var_Expr
--# expr-start 4122 '<FilterMap at ast.py:6996>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6996
   --# expr-start 4111 '<FieldAccess for all_env_elements_internal at ast.py:6996>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6996
--# expr-start 4110 '<Cast to Name at ast.py:6996>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6996
Cast_Expr := Var_Expr; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Name
   then
      Cast_Result := Create_Internal_Entity_Name
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 4110
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.all_env_elements_internal
Fld_2 := Libadalang.Implementation.Name_P_All_Env_Elements_Internal (Node => Cast_Result.Node, Seq => False, E_Info => Cast_Result.Info);
--# end
--# expr-done 4111
      declare
         Map_Result_Vec : Internal_Entity_Basic_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Fld_2;
      begin
         for Item_163 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_163
--# expr-start 4113 '<FieldAccess for subp_spec_or_null at ast.py:7000>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7000
--# expr-start 4112 '<Cast to BasicDecl at ast.py:7000>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7000
Cast_Expr_2 := Item_163; 
   if Cast_Expr_2.Node = null
      or else Cast_Expr_2.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_2 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 4112
      if Cast_Result_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_3 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Cast_Result_2.Node, E_Info => Cast_Result_2.Info);
--# end
--# expr-done 4113
Var_Expr_1 := Fld_3;
if Var_Expr_1 /= No_Entity_Base_Subp_Spec then
   --# scope-start
   --# bind ss Var_Expr_1
--# expr-start 4118 '<FieldAccess for matching_formal_type at ast.py:7002>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7002
--# expr-start 4115 '<at at ast.py:7002>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7002
--# expr-start 4114 '<FieldAccess for unpacked_formal_params at ast.py:7002>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7002
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld_4 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 4114
Get_Result := Get (Fld_4, 0, True); 
--# expr-done 4115
Var_Expr_2 := Get_Result;
if Var_Expr_2 /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 4117 '<FieldAccess for formal_type at ast.py:7002>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7002
--# expr-start 4116 '<FieldAccess for formal_decl at ast.py:7002>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7002
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_5 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Var_Expr_2.Node, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 4116
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamDecl.formal_type
Fld_6 := Libadalang.Implementation.Base_Formal_Param_Decl_P_Formal_Type (Node => Fld_5.Node, Origin => Bound_Origin, E_Info => Fld_5.Info);
--# end
--# expr-done 4117
Scope_Result := Fld_6;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
      if Result_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Cast_Expr_3 := Ent; 
      Cast_Result_3 := Create_Internal_Entity_Base_Type_Decl
        (Node => Cast_Expr_3.Node,
         Info => Cast_Expr_3.Info);
--# property-call-start BaseTypeDecl.matching_formal_type
Fld_7 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Formal_Type (Node => Result_Var.Node, Formal_Type => Cast_Result_3, Origin => Bound_Origin, E_Info => Result_Var.Info);
--# end
--# expr-done 4118
if Fld_7 then
   --# expr-start 4120 '<FieldAccess for is_implicit_deref at ast.py:7004>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7004
--# expr-start 4119 '<FieldAccess for return_type at ast.py:7004>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7004
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.return_type
Fld_8 := Libadalang.Implementation.Base_Subp_Spec_P_Return_Type (Node => Var_Expr_1.Node, Origin => Bound_Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 4119
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_implicit_deref
Fld_9 := Libadalang.Implementation.Base_Type_Decl_P_Is_Implicit_Deref (Node => Fld_8.Node, Origin => Bound_Origin, E_Info => Fld_8.Info);
--# end
--# expr-done 4120
   If_Result := Fld_9;
else
   If_Result := False;
end if;
Scope_Result_1 := If_Result;
   --# end
      Finalizer_Scope_2322;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
         if Result_Var_1 then
      --# expr-start 4121 '<Cast to BasicDecl at ast.py:6998>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6998
Cast_Expr_1 := Item_163; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_1 := No_Entity_Basic_Decl;
   end if;
--# expr-done 4121
         declare
            Item_To_Append : constant Internal_Entity_Basic_Decl := Cast_Result_1;
         begin
            Internal_Entity_Basic_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Basic_Decl_Array
           (Items_Count => Natural (Internal_Entity_Basic_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Basic_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Basic_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Basic_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 4122
Scope_Result_2 := Map_Result;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2320;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Internal_Entity_Basic_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var_2);
Dyn_Var_Bind_Result := Result_Var_2; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 4123
Let_Result := Dyn_Var_Bind_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 4124
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_526;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_526;
               Finalizer_Scope_2320;
               Finalizer_Scope_2322;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Decl_P_Variable_Indexing_Fns;
--# end
--# property-start TypeDecl._env_mappings_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6732
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_3
  (Node : Bare_Type_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Decl :=
     Internal_Entity_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1419 with Inline_Always;
      Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Type_Decl;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1419 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1419;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4127 '<New EnvAssoc at ast.py:6732>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6732
--# expr-start 4125 '<New DesignatedEnv at ast.py:6732>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6732
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 4125
--# expr-start 4126 '<FieldAccess for name_symbol at ast.py:6732>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6732
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 4126
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 4127
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1419;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1419;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_3;
--# end
--# property-start TypeDecl._env_mappings_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6733
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_4
  (Node : Bare_Type_Decl
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1420 with Inline_Always;
      Fld : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1420 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1420;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4128 '<FieldAccess for predefined_operators at ast.py:6733>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6733
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDecl.predefined_operators
Fld := Libadalang.Implementation.Type_Decl_P_Predefined_Operators (Node => Self);
--# end
--# expr-done 4128
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1420;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1420;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_4;
--# end
--# property-start TypeDecl._env_trans_parent_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6731
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_5
  (Node : Bare_Type_Decl
  )
   return Boolean
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_5;
--# end
--# property-start TypeDecl._ref_env_nodes_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6737
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_6
  (Node : Bare_Type_Decl
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1423 with Inline_Always;
      Cast_Expr : Bare_Type_Decl;
Cast_Result : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1423 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1423;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4129 '<Cast to AdaNode at ast.py:6737>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6737
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-done 4129
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Cast_Result;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1423;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1423;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_6;
--# end
--# property-start TypeDecl._env_dest_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6740
pragma Warnings (Off, "is not referenced");
 function Env_Dest_7
  (Node : Bare_Type_Decl
  )
   return Lexical_Env
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1424 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1424 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1424;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4130 '<FieldAccess for node_env at ast.py:6740>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6740
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Node_Env (Node => Self);
--# expr-done 4130
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1424;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1424;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Dest_7;
--# end
--# property-start TypeDecl._ref_cond_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6741
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_8
  (Node : Bare_Type_Decl
  )
   return Boolean
is
   Self : Bare_Type_Decl := Bare_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Type_Def;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4132 '<IsA DerivedTypeDef, InterfaceTypeDef at ast.py:6741>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6741
--# expr-start 4131 '<FieldAccess for type_def at ast.py:6741>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:6741
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Type_Decl_F_Type_Def;
--# expr-done 4131
Is_A := Fld /= null 
and then Fld.Kind in Ada_Derived_Type_Def_Range | Ada_Interface_Type_Def_Range; 
--# expr-done 4132
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_8;
--# end
function Type_Decl_F_Direct_Primitive_Subps
  (E : Entity
  ) return Internal_Inner_Env_Assoc_Array_Access
is
   Result : constant Internal_Inner_Env_Assoc_Array_Access :=
      Type_Decl_F_Direct_Primitive_Subps (E.Node);
begin
      return Result;
end;
function Type_Decl_P_Refined_Parent_Primitives_Env
  (E : Entity
  ) return Lexical_Env
is
      E_Info : Internal_Entity_Info :=
         Shed_Rebindings (E.Info, Node_Env (E.Node));
   Result : constant Lexical_Env :=
      Type_Decl_P_Refined_Parent_Primitives_Env (E.Node, E_Info);
begin
      return Result;
end;
   --
   --  Primitives for Bare_Anonymous_Type_Decl
   --
      procedure Initialize_Fields_For_Anonymous_Type_Decl
        (Self : Bare_Anonymous_Type_Decl
         ; Base_Type_Decl_F_Name : Bare_Defining_Name
         ; Type_Decl_F_Discriminants : Bare_Discriminant_Part
         ; Type_Decl_F_Type_Def : Bare_Type_Def
         ; Type_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Initialize_Fields_For_Type_Decl
              (Self, Base_Type_Decl_F_Name, Type_Decl_F_Discriminants, Type_Decl_F_Type_Def, Type_Decl_F_Aspects);
      end Initialize_Fields_For_Anonymous_Type_Decl;
--# property-start AnonymousTypeDecl.access_def_matches /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7019
pragma Warnings (Off, "is not referenced");
 function Anonymous_Type_Decl_P_Access_Def_Matches
  (Node : Bare_Anonymous_Type_Decl
      ; Other : Internal_Entity_Base_Type_Decl
      ; For_Assignment : Boolean
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Anonymous_Type_Decl := Bare_Anonymous_Type_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Anonymous_Type_Decl :=
     Internal_Entity_Anonymous_Type_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind other Other
   --# bind for_assignment For_Assignment
   --# bind origin Origin
   Property_Result : Boolean;
      Block_Var : Internal_Entity_Access_To_Subp_Def;
Block_Var_1 : Internal_Entity_Access_To_Subp_Def;
Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Entity_Access_Def;
Var_Expr_2 : Internal_Entity_Base_Type_Decl;
Exp : Internal_Entity_Base_Type_Decl;
Act : Internal_Entity_Base_Type_Decl;
Var_Expr_3 : Internal_Entity_Base_Type_Decl;
Var_Expr_4 : Internal_Entity_Classwide_Type_Decl;
Var_Expr_5 : Internal_Entity_Base_Type_Decl;
Var_Expr_6 : Internal_Entity_Classwide_Type_Decl;
Fld : Internal_Entity_Type_Def;
Cast_Result : Internal_Entity_Access_To_Subp_Def;
Fld_1 : Internal_Entity_Access_Def;
Cast_Result_1 : Internal_Entity_Access_To_Subp_Def;
Is_Null : Boolean;
Not_Val : Boolean;
Is_Null_1 : Boolean;
Not_Val_1 : Boolean;
If_Result : Boolean;
Fld_2 : Internal_Entity_Subp_Spec;
Fld_3 : Internal_Entity_Subp_Spec;
Cast_Result_2 : Internal_Entity_Base_Subp_Spec;
Fld_4 : Boolean;
Is_Null_2 : Boolean;
Is_Null_3 : Boolean;
If_Result_1 : Boolean;
Fld_5 : Internal_Entity_Type_Def;
Cast_Result_3 : Internal_Entity_Access_Def;
Fld_6 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Fld_7 : Internal_Entity_Base_Type_Decl;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_4 : Internal_Entity_Classwide_Type_Decl;
Fld_8 : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Result_Var_1 : Internal_Entity_Base_Type_Decl;
Scope_Result_2 : Internal_Entity_Base_Type_Decl;
Result_Var_2 : Internal_Entity_Base_Type_Decl;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_5 : Internal_Entity_Classwide_Type_Decl;
Fld_9 : Internal_Entity_Base_Type_Decl;
Scope_Result_3 : Internal_Entity_Base_Type_Decl;
Result_Var_3 : Internal_Entity_Base_Type_Decl;
Scope_Result_4 : Internal_Entity_Base_Type_Decl;
Result_Var_4 : Internal_Entity_Base_Type_Decl;
Fld_10 : Boolean;
Fld_11 : Boolean;
If_Result_2 : Boolean;
Let_Result : Boolean;
Scope_Result_5 : Boolean;
Result_Var_5 : Boolean;
Scope_Result_6 : Boolean;
Result_Var_6 : Boolean;
If_Result_3 : Boolean;
If_Result_4 : Boolean;
Let_Result_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4159 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 4134 '<Cast to AccessToSubpDef at ast.py:7029>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7029
--# expr-start 4133 '<FieldAccess for type_def at ast.py:7029>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7029
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 4133
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Access_To_Subp_Def_Range
   then
      Cast_Result := Create_Internal_Entity_Access_To_Subp_Def
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Access_To_Subp_Def;
   end if;
--# expr-done 4134
Block_Var := Cast_Result; 
--# bind self_subp_access Block_Var
--# expr-start 4136 '<Cast to AccessToSubpDef at ast.py:7030>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7030
--# expr-start 4135 '<FieldAccess for access_def at ast.py:7030>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7030
      if Other.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.access_def'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Access_Def (Node => Other.Node, Origin => Origin, E_Info => Other.Info);
--# end
--# expr-done 4135
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Access_To_Subp_Def_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Access_To_Subp_Def
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         Cast_Result_1 := No_Entity_Access_To_Subp_Def;
   end if;
--# expr-done 4136
Block_Var_1 := Cast_Result_1; 
--# bind other_subp_access Block_Var_1
--# expr-start 4158 '<Cond at ast.py:7034>' If_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7034
--# expr-start 4138 '<Not at ast.py:7035>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7035
--# expr-start 4137 '<IsNull at ast.py:7035>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7035
Is_Null := Block_Var.Node = null; 
--# expr-done 4137
Not_Val := not (Is_Null); 
--# expr-done 4138
if Not_Val then
   --# expr-start 4140 '<Not at ast.py:7035>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7035
--# expr-start 4139 '<IsNull at ast.py:7035>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7035
Is_Null_1 := Block_Var_1.Node = null; 
--# expr-done 4139
Not_Val_1 := not (Is_Null_1); 
--# expr-done 4140
   If_Result := Not_Val_1;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 4143 '<FieldAccess for match_signature at ast.py:7036>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7036
--# expr-start 4141 '<FieldAccess for subp_spec at ast.py:7036>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7036
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Subp_Spec (Node => Block_Var_1.Node.Access_To_Subp_Def_F_Subp_Spec, Info => Block_Var_1.Info);
--# expr-done 4141
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4142 '<FieldAccess for subp_spec at ast.py:7037>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7037
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Subp_Spec (Node => Block_Var.Node.Access_To_Subp_Def_F_Subp_Spec, Info => Block_Var.Info);
--# expr-done 4142
      Cast_Result_2 := Create_Internal_Entity_Base_Subp_Spec
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
--# property-call-start BaseSubpSpec.match_signature
Fld_4 := Libadalang.Implementation.Base_Subp_Spec_P_Match_Signature (Node => Fld_2.Node, Other => Cast_Result_2, Match_Name => False, E_Info => Fld_2.Info);
--# end
--# expr-done 4143
   If_Result_4 := Fld_4;
else
--# expr-start 4144 '<IsNull at ast.py:7040>' Is_Null_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7040
Is_Null_2 := Block_Var.Node = null; 
--# expr-done 4144
if Is_Null_2 then
   --# expr-start 4145 '<IsNull at ast.py:7040>' Is_Null_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7040
Is_Null_3 := Block_Var_1.Node = null; 
--# expr-done 4145
   If_Result_1 := Is_Null_3;
else
   If_Result_1 := False;
end if;
if If_Result_1 then
--# expr-start 4147 '<Cast to AccessDef at ast.py:7041>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7041
--# expr-start 4146 '<FieldAccess for type_def at ast.py:7041>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7041
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Type_Def (Node => Ent.Node.Type_Decl_F_Type_Def, Info => Ent.Info);
--# expr-done 4146
   if Fld_5.Node = null
      or else Fld_5.Node.Kind in Ada_Access_Def
   then
      Cast_Result_3 := Create_Internal_Entity_Access_Def
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
   else
         Cast_Result_3 := No_Entity_Access_Def;
   end if;
--# expr-done 4147
Var_Expr_1 := Cast_Result_3;
if Var_Expr_1 /= No_Entity_Access_Def then
   --# scope-start
--# expr-start 4148 '<FieldAccess for accessed_type at ast.py:7041>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7041
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.accessed_type'
Fld_6 := Libadalang.Implementation.Dispatcher_Type_Def_P_Accessed_Type (Node => Var_Expr_1.Node, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 4148
Scope_Result := Fld_6;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Var_Expr := Result_Var;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind ast Var_Expr
--# expr-start 4149 '<FieldAccess for accessed_type at ast.py:7042>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7042
      if Other.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_7 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Other.Node, Origin => Origin, E_Info => Other.Info);
--# end
--# expr-done 4149
Var_Expr_2 := Fld_7;
if Var_Expr_2 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind oat Var_Expr_2
--# expr-start 4157 '<Let at ast.py:7043>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7043
--# scope-start
--# expr-start 4150 '<Cast to ClasswideTypeDecl at ast.py:7048>' Cast_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7048
Cast_Expr := Var_Expr; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Classwide_Type_Decl_Range
   then
      Cast_Result_4 := Create_Internal_Entity_Classwide_Type_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result_4 := No_Entity_Classwide_Type_Decl;
   end if;
--# expr-done 4150
Var_Expr_4 := Cast_Result_4;
if Var_Expr_4 /= No_Entity_Classwide_Type_Decl then
   --# scope-start
--# expr-start 4151 '<FieldAccess for typedecl at ast.py:7048>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7048
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ClasswideTypeDecl.typedecl
Fld_8 := Libadalang.Implementation.Classwide_Type_Decl_P_Typedecl (Node => Var_Expr_4.Node, E_Info => Var_Expr_4.Info);
--# end
--# expr-done 4151
Scope_Result_1 := Fld_8;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Base_Type_Decl;
end if;
Var_Expr_3 := Result_Var_1;
if Var_Expr_3 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind e Var_Expr_3
Scope_Result_2 := Var_Expr_3;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := Var_Expr;
end if;
Exp := Result_Var_2; 
--# bind exp Exp
--# expr-start 4152 '<Cast to ClasswideTypeDecl at ast.py:7049>' Cast_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7049
Cast_Expr_1 := Var_Expr_2; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Classwide_Type_Decl_Range
   then
      Cast_Result_5 := Create_Internal_Entity_Classwide_Type_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_5 := No_Entity_Classwide_Type_Decl;
   end if;
--# expr-done 4152
Var_Expr_6 := Cast_Result_5;
if Var_Expr_6 /= No_Entity_Classwide_Type_Decl then
   --# scope-start
--# expr-start 4153 '<FieldAccess for typedecl at ast.py:7049>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7049
      if Var_Expr_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ClasswideTypeDecl.typedecl
Fld_9 := Libadalang.Implementation.Classwide_Type_Decl_P_Typedecl (Node => Var_Expr_6.Node, E_Info => Var_Expr_6.Info);
--# end
--# expr-done 4153
Scope_Result_3 := Fld_9;
   --# end
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := No_Entity_Base_Type_Decl;
end if;
Var_Expr_5 := Result_Var_3;
if Var_Expr_5 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind e Var_Expr_5
Scope_Result_4 := Var_Expr_5;
   --# end
   Result_Var_4 := Scope_Result_4;
else
   Result_Var_4 := Var_Expr_2;
end if;
Act := Result_Var_4; 
--# bind act Act
--# expr-start 4156 '<If at ast.py:7050>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7050
if For_Assignment then
   --# expr-start 4154 '<FieldAccess for matching_assign_type at ast.py:7056>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7056
      if Act.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_assign_type
Fld_10 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Assign_Type (Node => Act.Node, Expected_Type => Var_Expr, Origin => Origin, E_Info => Act.Info);
--# end
--# expr-done 4154
   If_Result_2 := Fld_10;
else
   --# expr-start 4155 '<FieldAccess for matching_type at ast.py:7058>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7058
      if Act.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.matching_type
Fld_11 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Type (Node => Act.Node, Expected_Type => Exp, Origin => Origin, E_Info => Act.Info);
--# end
--# expr-done 4155
   If_Result_2 := Fld_11;
end if;
--# expr-done 4156
Let_Result := If_Result_2; 
--# end
--# expr-done 4157
Scope_Result_5 := Let_Result;
   --# end
   Result_Var_5 := Scope_Result_5;
else
   Result_Var_5 := False;
end if;
Scope_Result_6 := Result_Var_5;
   --# end
   Result_Var_6 := Scope_Result_6;
else
   Result_Var_6 := False;
end if;
   If_Result_3 := Result_Var_6;
else
   If_Result_3 := False;
end if;
   If_Result_4 := If_Result_3;
end if;
--# expr-done 4158
Let_Result_1 := If_Result_4; 
--# end
--# expr-done 4159
      Property_Result := Let_Result_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Anonymous_Type_Decl_P_Access_Def_Matches;
--# end
--# property-start AnonymousTypeDecl.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7067
pragma Warnings (Off, "is not referenced");
 function Anonymous_Type_Decl_P_Xref_Entry_Point
  (Node : Bare_Anonymous_Type_Decl
  )
   return Boolean
is
   Self : Bare_Anonymous_Type_Decl := Bare_Anonymous_Type_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Anonymous_Type_Decl_P_Xref_Entry_Point;
--# end
   --
   --  Primitives for Bare_Synth_Anonymous_Type_Decl
   --
      procedure Initialize_Fields_For_Synth_Anonymous_Type_Decl
        (Self : Bare_Synth_Anonymous_Type_Decl
         ; Base_Type_Decl_F_Name : Bare_Defining_Name
         ; Type_Decl_F_Discriminants : Bare_Discriminant_Part
         ; Type_Decl_F_Type_Def : Bare_Type_Def
         ; Type_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Initialize_Fields_For_Anonymous_Type_Decl
              (Self, Base_Type_Decl_F_Name, Type_Decl_F_Discriminants, Type_Decl_F_Type_Def, Type_Decl_F_Aspects);
      end Initialize_Fields_For_Synth_Anonymous_Type_Decl;
   --
   --  Primitives for Bare_Basic_Subp_Decl
   --
      procedure Basic_Subp_Decl_Pre_Env_Actions
        (Self            : Bare_Basic_Subp_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Dummy : Boolean;
      begin
         Dummy := Libadalang.Implementation.Env_Do_25 (Node => 
Self);
      end;
      declare
         Env : constant Internal_Designated_Env :=
           Libadalang.Implementation.Initial_Env_26 (Node => 
Self);
      begin
         Set_Initial_Env
           (Self,
            State,
            Env,
            DSL_Location => "ast.py:8852");
      end;
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mappings : Internal_Env_Assoc_Array_Access :=
            Libadalang.Implementation.Env_Mappings_27 (Node => 
Self);
      begin
         for Mapping of Mappings.Items loop
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:8859");
         end loop;
         Dec_Ref (Mappings);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_28 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            Libadalang.Implementation.Env_Names_29 (Node => 
Self);
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      declare
         Dummy : Bare_Compilation_Unit_Array_Access;
      begin
         Dummy := Libadalang.Implementation.Env_Do_30 (Node => 
Self);
         Dec_Ref (Dummy);
      end;
      if Libadalang.Implementation.Ref_Cond_32 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_31 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Use_Package_Name_Designated_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_34 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_33 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Name_Designated_Type_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      end;
      procedure Basic_Subp_Decl_Post_Env_Actions
        (Self : Bare_Basic_Subp_Decl; State : in out PLE_Node_State) is
      begin
      declare
         Dummy : Boolean;
      begin
         Dummy := Libadalang.Implementation.Env_Do_35 (Node => 
Self);
      end;
      end;
--# property-start BasicSubpDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8743
pragma Warnings (Off, "is not referenced");
 function Basic_Subp_Decl_P_Defining_Names
  (Node : Bare_Basic_Subp_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Subp_Decl :=
     Internal_Entity_Basic_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_715 with Inline_Always;
      Fld : Internal_Entity_Base_Subp_Spec;
Fld_1 : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_715 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_715;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4161 '<FieldAccess for name at ast.py:8743>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8743
--# expr-start 4160 '<FieldAccess for subp_decl_spec at ast.py:8743>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8743
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicSubpDecl.subp_decl_spec'
Fld := Libadalang.Implementation.Dispatcher_Basic_Subp_Decl_P_Subp_Decl_Spec (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4160
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.name'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Name (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 4161
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld_1;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_715;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_715;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Subp_Decl_P_Defining_Names;
--# end
--# property-start BasicSubpDecl.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8745
pragma Warnings (Off, "is not referenced");
 function Basic_Subp_Decl_P_Defining_Env
  (Node : Bare_Basic_Subp_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Subp_Decl :=
     Internal_Entity_Basic_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_716 with Inline_Always;
      Fld : Internal_Entity_Base_Subp_Spec;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_716 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_716;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4163 '<FieldAccess for defining_env at ast.py:8745>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8745
--# expr-start 4162 '<FieldAccess for subp_decl_spec at ast.py:8745>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8745
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicSubpDecl.subp_decl_spec'
Fld := Libadalang.Implementation.Dispatcher_Basic_Subp_Decl_P_Subp_Decl_Spec (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4162
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.defining_env
Fld_1 := Libadalang.Implementation.Base_Subp_Spec_P_Defining_Env (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 4163
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_716;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_716;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Subp_Decl_P_Defining_Env;
--# end
--# property-start BasicSubpDecl.type_expression /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8747
pragma Warnings (Off, "is not referenced");
 function Basic_Subp_Decl_P_Type_Expression
  (Node : Bare_Basic_Subp_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Expr
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Subp_Decl :=
     Internal_Entity_Basic_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Type_Expr;
      Fld : Internal_Entity_Base_Subp_Spec;
Fld_1 : Internal_Entity_Type_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4165 '<FieldAccess for returns at ast.py:8748>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8748
--# expr-start 4164 '<FieldAccess for subp_decl_spec at ast.py:8748>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8748
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicSubpDecl.subp_decl_spec'
Fld := Libadalang.Implementation.Dispatcher_Basic_Subp_Decl_P_Subp_Decl_Spec (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4164
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.returns'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Returns (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 4165
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Subp_Decl_P_Type_Expression;
--# end
--# property-start BasicSubpDecl.get_body_in_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8755
pragma Warnings (Off, "is not referenced");
 function Basic_Subp_Decl_P_Get_Body_In_Env
  (Node : Bare_Basic_Subp_Decl
      ; Env : Lexical_Env
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Body_Node
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Subp_Decl :=
     Internal_Entity_Basic_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Body_Node;
            procedure Finalizer_Scope_718 with Inline_Always;
      Block_Var : Internal_Entity_Array_Access;
Block_Var_1 : Internal_Entity_Body_Node;
Block_Var_2 : Internal_Entity_Body_Node;
Var_Expr : Internal_Entity_Body_Node;
Var_Expr_1 : Internal_Entity_Body_Node;
Fld : Symbol_Type;
Env_Get_Result : Internal_Entity_Array_Access;
Fld_1 : Bare_Ada_Node;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Cast_Result : Internal_Entity_Body_Node;
Fld_2 : Internal_Entity_Base_Formal_Param_Holder;
Fld_3 : Internal_Entity_Base_Subp_Spec;
Fld_4 : Bare_Base_Subp_Spec;
As_Entity_1 : Internal_Entity_Base_Subp_Spec;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Base_Subp_Spec;
Cast_Result_1 : Internal_Entity_Base_Formal_Param_Holder;
Fld_5 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Find_Result : Internal_Entity;
Cast_Result_2 : Internal_Entity_Body_Node;
Is_A : Boolean;
Find_Result_1 : Internal_Entity;
Cast_Result_3 : Internal_Entity_Body_Node;
Is_Null_2 : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Internal_Entity_Body_Node;
Fld_6 : Bare_Body_Node;
As_Entity_2 : Internal_Entity_Body_Node;
Is_Null_3 : Boolean;
If_Result_4 : Internal_Entity_Body_Node;
Fld_7 : Internal_Entity;
Scope_Result_1 : Internal_Entity;
Result_Var_1 : Internal_Entity;
Cast_Result_4 : Internal_Entity_Body_Node;
Let_Result : Internal_Entity_Body_Node;
            procedure Finalizer_Scope_718 is
            begin
                     Dec_Ref (Env_Get_Result);
                     Dec_Ref (Block_Var);
            end Finalizer_Scope_718;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4187 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4167 '<EnvGet at ast.py:8757>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8757
--# expr-start 4166 '<FieldAccess for name_symbol at ast.py:8757>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8757
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 4166
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Env, Key => Fld, Lookup_Kind => To_Lookup_Kind_Type (Flat), Categories => (Inherited_Primitives => False, Nocat => True))); 
--# expr-done 4167
Block_Var := Env_Get_Result; Inc_Ref (Block_Var);
--# bind elements Block_Var
--# expr-start 4177 '<Cast to Body at ast.py:8759>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8759
--# expr-start 4176 '<Find at ast.py:8759>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8759
      Find_Result := No_Entity;
      declare
         Collection : constant Internal_Entity_Array_Access := Block_Var;
      begin
         for Item_164 of
               Collection.Items
         loop
   --# scope-start
               --# bind ent Item_164
--# expr-start 4170 '<Cast to Body at ast.py:8763>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8763
--# expr-start 4169 '<as_bare_entity at ast.py:8763>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8763
--# expr-start 4168 '<FieldAccess for node at ast.py:8763>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8763
Fld_1 := Item_164.Node;
--# expr-done 4168
Is_Null := Fld_1 = null; 
if Is_Null then
   If_Result := No_Entity;
else
As_Entity := (Info => No_Entity_Info, Node => Fld_1); 
   If_Result := As_Entity;
end if;
--# expr-done 4169
   if If_Result.Node = null
      or else If_Result.Node.Kind in Ada_Body_Node
   then
      Cast_Result := Create_Internal_Entity_Body_Node
        (Node => If_Result.Node,
         Info => If_Result.Info);
   else
         Cast_Result := No_Entity_Body_Node;
   end if;
--# expr-done 4170
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Body_Node then
   --# scope-start
--# expr-start 4175 '<FieldAccess for match_other at ast.py:8763>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8763
--# expr-start 4171 '<FieldAccess for formal_param_holder_or_null at ast.py:8763>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8763
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.formal_param_holder_or_null
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Formal_Param_Holder_Or_Null (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 4171
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4174 '<as_bare_entity at ast.py:8765>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8765
--# expr-start 4173 '<FieldAccess for node at ast.py:8765>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8765
--# expr-start 4172 '<FieldAccess for subp_decl_spec at ast.py:8765>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8765
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicSubpDecl.subp_decl_spec'
Fld_3 := Libadalang.Implementation.Dispatcher_Basic_Subp_Decl_P_Subp_Decl_Spec (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4172
Fld_4 := Fld_3.Node;
--# expr-done 4173
Is_Null_1 := Fld_4 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Base_Subp_Spec;
else
As_Entity_1 := (Info => No_Entity_Info, Node => Fld_4); 
   If_Result_1 := As_Entity_1;
end if;
--# expr-done 4174
      Cast_Result_1 := Create_Internal_Entity_Base_Formal_Param_Holder
        (Node => If_Result_1.Node,
         Info => If_Result_1.Info);
--# property-call-start '[dispatcher]BaseFormalParamHolder.match_other'
Fld_5 := Libadalang.Implementation.Dispatcher_Base_Formal_Param_Holder_P_Match_Other (Node => Fld_2.Node, Other => Cast_Result_1, Match_Names => True, E_Info => Fld_2.Info);
--# end
--# expr-done 4175
Scope_Result := Fld_5;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
      if Result_Var then
         Find_Result := Item_164;
      end if;
   --# end
               exit when Result_Var;
         end loop;
      end;
--# expr-done 4176
   if Find_Result.Node = null
      or else Find_Result.Node.Kind in Ada_Body_Node
   then
      Cast_Result_2 := Create_Internal_Entity_Body_Node
        (Node => Find_Result.Node,
         Info => Find_Result.Info);
   else
         Cast_Result_2 := No_Entity_Body_Node;
   end if;
--# expr-done 4177
Block_Var_1 := Cast_Result_2; 
--# bind precise Block_Var_1
--# expr-start 4182 '<If at ast.py:8769>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8769
--# expr-start 4178 '<IsNull at ast.py:8770>' Is_Null_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8770
Is_Null_2 := Block_Var_1.Node = null; 
--# expr-done 4178
if Is_Null_2 then
   If_Result_2 := Imprecise_Fallback;
else
   If_Result_2 := False;
end if;
if If_Result_2 then
   --# expr-start 4181 '<Cast to Body at ast.py:8771>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8771
--# expr-start 4180 '<Find at ast.py:8771>' Find_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8771
      Find_Result_1 := No_Entity;
      declare
         Collection : constant Internal_Entity_Array_Access := Block_Var;
      begin
         for Item_165 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_165
      --# expr-start 4179 '<IsA Body at ast.py:8771>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8771
Is_A := Item_165.Node /= null 
and then Item_165.Node.Kind in Ada_Body_Node; 
--# expr-done 4179
      if Is_A then
         Find_Result_1 := Item_165;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 4180
   if Find_Result_1.Node = null
      or else Find_Result_1.Node.Kind in Ada_Body_Node
   then
      Cast_Result_3 := Create_Internal_Entity_Body_Node
        (Node => Find_Result_1.Node,
         Info => Find_Result_1.Info);
   else
         Cast_Result_3 := No_Entity_Body_Node;
   end if;
--# expr-done 4181
   If_Result_3 := Cast_Result_3;
else
   If_Result_3 := Block_Var_1;
end if;
--# expr-done 4182
Block_Var_2 := If_Result_3; 
--# bind result Block_Var_2
--# expr-start 4186 '<Cast to Body at ast.py:8776>' Cast_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8776
--# expr-start 4184 '<as_entity at ast.py:8776>' If_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8776
--# expr-start 4183 '<FieldAccess for node at ast.py:8776>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8776
Fld_6 := Block_Var_2.Node;
--# expr-done 4183
Is_Null_3 := Fld_6 = null; 
if Is_Null_3 then
   If_Result_4 := No_Entity_Body_Node;
else
As_Entity_2 := (Info => E_Info, Node => Fld_6); 
   If_Result_4 := As_Entity_2;
end if;
--# expr-done 4184
Var_Expr_1 := If_Result_4;
if Var_Expr_1 /= No_Entity_Body_Node then
   --# scope-start
--# expr-start 4185 '<FieldAccess for without_md at ast.py:8776>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8776
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.without_md
Fld_7 := Libadalang.Implementation.Ada_Node_P_Without_Md (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 4185
Scope_Result_1 := Fld_7;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity;
end if;
   if Result_Var_1.Node = null
      or else Result_Var_1.Node.Kind in Ada_Body_Node
   then
      Cast_Result_4 := Create_Internal_Entity_Body_Node
        (Node => Result_Var_1.Node,
         Info => Result_Var_1.Info);
   else
         Cast_Result_4 := No_Entity_Body_Node;
   end if;
--# expr-done 4186
Let_Result := Cast_Result_4; 
--# end
--# expr-done 4187
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_718;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_718;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Subp_Decl_P_Get_Body_In_Env;
--# end
--# property-start BasicSubpDecl.next_part_for_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8779
pragma Warnings (Off, "is not referenced");
 function Basic_Subp_Decl_P_Next_Part_For_Decl
  (Node : Bare_Basic_Subp_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Subp_Decl :=
     Internal_Entity_Basic_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_719 with Inline_Always;
            procedure Finalizer_Scope_2336 with Inline_Always;
            procedure Finalizer_Scope_2337 with Inline_Always;
            procedure Finalizer_Scope_2339 with Inline_Always;
      Block_Var : Bare_Declarative_Part;
Block_Var_1 : Internal_Entity_Basic_Decl;
Block_Var_2 : Internal_Entity_Basic_Decl;
Var_Expr : Internal_Entity_Declarative_Part;
Var_Expr_1 : Internal_Entity_Basic_Decl;
Var_Expr_2 : Internal_Entity_Body_Node;
Var_Expr_3 : Internal_Entity_Private_Part;
Var_Expr_4 : Internal_Entity_Body_Node;
Fld : Bare_Declarative_Part;
As_Entity : Internal_Entity_Declarative_Part;
Is_Null : Boolean;
If_Result : Internal_Entity_Declarative_Part;
Fld_1 : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity_Basic_Decl;
Is_Null_1 : Boolean;
Not_Val : Boolean;
Is_A : Boolean;
Fld_3 : Internal_Entity_Basic_Decl;
Fld_4 : Lexical_Env := Empty_Env;
Fld_5 : Internal_Entity_Body_Node;
Scope_Result_1 : Internal_Entity_Body_Node;
Result_Var_1 : Internal_Entity_Body_Node;
Is_A_1 : Boolean;
Fld_6 : Internal_Entity_Private_Part;
Fld_7 : Lexical_Env := Empty_Env;
Fld_8 : Internal_Entity_Body_Node;
Scope_Result_2 : Internal_Entity_Body_Node;
Result_Var_2 : Internal_Entity_Body_Node;
Scope_Result_3 : Internal_Entity_Body_Node;
Fld_9 : Internal_Entity_Body_Node;
Fld_10 : Lexical_Env := Empty_Env;
Fld_11 : Internal_Entity_Body_Node;
Scope_Result_4 : Internal_Entity_Body_Node;
Result_Var_3 : Internal_Entity_Body_Node;
Result_Var_4 : Internal_Entity_Body_Node;
Is_Null_2 : Boolean;
Fld_12 : Lexical_Env := Empty_Env;
Fld_13 : Internal_Entity_Body_Node;
Cast_Result_1 : Internal_Entity_Basic_Decl;
Cast_Expr : Internal_Entity_Body_Node;
Cast_Result_2 : Internal_Entity_Basic_Decl;
If_Result_1 : Internal_Entity_Basic_Decl;
Cast_Result_3 : Internal_Entity_Basic_Decl;
If_Result_2 : Internal_Entity_Basic_Decl;
Cast_Result_4 : Internal_Entity_Basic_Decl;
If_Result_3 : Internal_Entity_Basic_Decl;
If_Result_4 : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_719 is
            begin
                     Dec_Ref (Fld_12);
            end Finalizer_Scope_719;
            procedure Finalizer_Scope_2336 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_2336;
            procedure Finalizer_Scope_2337 is
            begin
                     Dec_Ref (Fld_7);
            end Finalizer_Scope_2337;
            procedure Finalizer_Scope_2339 is
            begin
                     Dec_Ref (Fld_10);
            end Finalizer_Scope_2339;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4210 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4188 '<FieldAccess for declarative_scope at ast.py:8780>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8780
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.declarative_scope
Fld := Libadalang.Implementation.Ada_Node_P_Declarative_Scope (Node => Ent.Node);
--# end
--# expr-done 4188
Block_Var := Fld; 
--# bind decl_scope Block_Var
--# expr-start 4189 '<as_entity at ast.py:8781>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8781
Is_Null := Block_Var = null; 
if Is_Null then
   If_Result := No_Entity_Declarative_Part;
else
As_Entity := (Info => E_Info, Node => Block_Var); 
   If_Result := As_Entity;
end if;
--# expr-done 4189
Var_Expr := If_Result;
if Var_Expr /= No_Entity_Declarative_Part then
   --# scope-start
   --# bind ds Var_Expr
--# expr-start 4191 '<Cast to BasicDecl at ast.py:8782>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8782
--# expr-start 4190 '<FieldAccess for semantic_parent at ast.py:8782>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8782
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.semantic_parent
Fld_1 := Libadalang.Implementation.Ada_Node_P_Semantic_Parent (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 4190
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 4191
Scope_Result := Cast_Result;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Basic_Decl;
end if;
Block_Var_1 := Result_Var; 
--# bind parent_decl Block_Var_1
--# expr-start 4192 '<FieldAccess for basic_decl_next_part_for_decl at ast.py:8785>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8785
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.basic_decl_next_part_for_decl
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Basic_Decl_Next_Part_For_Decl (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 4192
Block_Var_2 := Fld_2; 
--# bind default_next_part Block_Var_2
--# expr-start 4209 '<Cond at ast.py:8787>' If_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8787
--# expr-start 4194 '<Not at ast.py:8790>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8790
--# expr-start 4193 '<IsNull at ast.py:8790>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8790
Is_Null_1 := Block_Var_2.Node = null; 
--# expr-done 4193
Not_Val := not (Is_Null_1); 
--# expr-done 4194
if Not_Val then
   If_Result_4 := Block_Var_2;
else
--# expr-start 4195 '<IsA PrivatePart at ast.py:8794>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8794
Is_A := Block_Var /= null 
and then Block_Var.Kind in Ada_Private_Part_Range; 
--# expr-done 4195
if Is_A then
--# expr-start 4196 '<FieldAccess for next_part_for_decl at ast.py:8795>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8795
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.next_part_for_decl'
Fld_3 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Next_Part_For_Decl (Node => Block_Var_1.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Block_Var_1.Info);
--# end
--# expr-done 4196
Var_Expr_1 := Fld_3;
if Var_Expr_1 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind np Var_Expr_1
--# expr-start 4198 '<FieldAccess for get_body_in_env at ast.py:8796>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8796
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4197 '<FieldAccess for children_env at ast.py:8796>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8796
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Children_Env (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# expr-done 4197
--# property-call-start BasicSubpDecl.get_body_in_env
Fld_5 := Libadalang.Implementation.Basic_Subp_Decl_P_Get_Body_In_Env (Node => Ent.Node, Env => Fld_4, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 4198
Scope_Result_1 := Fld_5;
   --# end
      Finalizer_Scope_2336;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Body_Node;
end if;
      Cast_Result_4 := Create_Internal_Entity_Basic_Decl
        (Node => Result_Var_1.Node,
         Info => Result_Var_1.Info);
   If_Result_3 := Cast_Result_4;
else
--# expr-start 4199 '<IsA PublicPart at ast.py:8800>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8800
Is_A_1 := Block_Var /= null 
and then Block_Var.Kind in Ada_Public_Part_Range; 
--# expr-done 4199
if Is_A_1 then
--# expr-start 4200 '<FieldAccess for decl_private_part at ast.py:8803>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8803
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.decl_private_part
Fld_6 := Libadalang.Implementation.Basic_Decl_P_Decl_Private_Part (Node => Block_Var_1.Node, E_Info => Block_Var_1.Info);
--# end
--# expr-done 4200
Var_Expr_3 := Fld_6;
if Var_Expr_3 /= No_Entity_Private_Part then
   --# scope-start
   --# bind dpp Var_Expr_3
--# expr-start 4202 '<FieldAccess for get_body_in_env at ast.py:8804>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8804
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4201 '<FieldAccess for children_env at ast.py:8804>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8804
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Children_Env (Node => Var_Expr_3.Node, E_Info => Var_Expr_3.Info);
--# expr-done 4201
--# property-call-start BasicSubpDecl.get_body_in_env
Fld_8 := Libadalang.Implementation.Basic_Subp_Decl_P_Get_Body_In_Env (Node => Ent.Node, Env => Fld_7, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 4202
Scope_Result_2 := Fld_8;
   --# end
      Finalizer_Scope_2337;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Entity_Body_Node;
end if;
Var_Expr_2 := Result_Var_2;
if Var_Expr_2 /= No_Entity_Body_Node then
   --# scope-start
   --# bind e Var_Expr_2
Scope_Result_3 := Var_Expr_2;
   --# end
   Result_Var_4 := Scope_Result_3;
else
--# expr-start 4203 '<FieldAccess for body_part_for_decl at ast.py:8807>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8807
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.body_part_for_decl
Fld_9 := Libadalang.Implementation.Basic_Decl_P_Body_Part_For_Decl (Node => Block_Var_1.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Block_Var_1.Info);
--# end
--# expr-done 4203
Var_Expr_4 := Fld_9;
if Var_Expr_4 /= No_Entity_Body_Node then
   --# scope-start
   --# bind np Var_Expr_4
--# expr-start 4205 '<FieldAccess for get_body_in_env at ast.py:8808>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8808
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4204 '<FieldAccess for children_env at ast.py:8808>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8808
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Libadalang.Implementation.Children_Env (Node => Var_Expr_4.Node, E_Info => Var_Expr_4.Info);
--# expr-done 4204
--# property-call-start BasicSubpDecl.get_body_in_env
Fld_11 := Libadalang.Implementation.Basic_Subp_Decl_P_Get_Body_In_Env (Node => Ent.Node, Env => Fld_10, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 4205
Scope_Result_4 := Fld_11;
   --# end
      Finalizer_Scope_2339;
   Result_Var_3 := Scope_Result_4;
else
   Result_Var_3 := No_Entity_Body_Node;
end if;
   Result_Var_4 := Result_Var_3;
end if;
      Cast_Result_3 := Create_Internal_Entity_Basic_Decl
        (Node => Result_Var_4.Node,
         Info => Result_Var_4.Info);
   If_Result_2 := Cast_Result_3;
else
--# expr-start 4206 '<IsNull at ast.py:8812>' Is_Null_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8812
Is_Null_2 := Block_Var = null; 
--# expr-done 4206
if Is_Null_2 then
Cast_Expr := No_Entity_Body_Node; 
      Cast_Result_2 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   If_Result_1 := Cast_Result_2;
else
--# expr-start 4208 '<FieldAccess for get_body_in_env at ast.py:8816>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8816
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4207 '<FieldAccess for children_env at ast.py:8816>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8816
      if Block_Var = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Libadalang.Implementation.Children_Env (Node => Block_Var);
--# expr-done 4207
--# property-call-start BasicSubpDecl.get_body_in_env
Fld_13 := Libadalang.Implementation.Basic_Subp_Decl_P_Get_Body_In_Env (Node => Ent.Node, Env => Fld_12, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 4208
      Cast_Result_1 := Create_Internal_Entity_Basic_Decl
        (Node => Fld_13.Node,
         Info => Fld_13.Info);
   If_Result_1 := Cast_Result_1;
end if;
   If_Result_2 := If_Result_1;
end if;
   If_Result_3 := If_Result_2;
end if;
   If_Result_4 := If_Result_3;
end if;
--# expr-done 4209
Let_Result := If_Result_4; 
--# end
--# expr-done 4210
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_719;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_719;
               Finalizer_Scope_2336;
               Finalizer_Scope_2337;
               Finalizer_Scope_2339;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Subp_Decl_P_Next_Part_For_Decl;
--# end
--# property-start BasicSubpDecl.constrain_prefix /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8820
pragma Warnings (Off, "is not referenced");
 function Basic_Subp_Decl_P_Constrain_Prefix
  (Node : Bare_Basic_Subp_Decl
      ; Prefix : Bare_Expr
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Subp_Decl :=
     Internal_Entity_Basic_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind prefix Prefix
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_720 with Inline_Always;
      Var_Expr : Internal_Entity_Defining_Name;
Fld : Logic_Var;
Fld_1 : Internal_Entity_Base_Subp_Spec;
Fld_2 : Internal_Entity_Defining_Name_Array_Access;
Get_Result : Internal_Entity_Defining_Name;
Fld_3 : Internal_Entity_Base_Formal_Param_Decl;
Fld_4 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_5 : Logic_Equation;
And_Pred : Logic_Equation;
True_Rel : Logic_Equation;
Fld_6 : Internal_Entity_Info;
Fld_7 : Internal_Metadata;
Fld_8 : Boolean;
If_Result : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_720 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_5);
                     Dec_Ref (And_Pred);
                     Dec_Ref (True_Rel);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_720;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4224 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4223 '<If at ast.py:8821>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8821
--# expr-start 4213 '<FieldAccess for dottable_subp at ast.py:8825>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8825
--# expr-start 4212 '<FieldAccess for md at ast.py:8825>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8825
--# expr-start 4211 '<FieldAccess for info at ast.py:8825>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8825
Fld_6 := Ent.Info;
--# expr-done 4211
Fld_7 := Fld_6.Md;
--# expr-done 4212
Fld_8 := Fld_7.Dottable_Subp;
--# expr-done 4213
if Fld_8 then
   --# expr-start 4222 '<And at ast.py:8826>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8826
--# expr-start 4220 '<Bind at ast.py:8826>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8826
--# expr-start 4214 '<FieldAccess for expected_type_var at ast.py:8826>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8826
      if Prefix = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Prefix);
--# expr-done 4214
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
--# expr-start 4217 '<at at ast.py:8827>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8827
--# expr-start 4216 '<FieldAccess for unpacked_formal_params at ast.py:8827>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8827
--# expr-start 4215 '<FieldAccess for subp_decl_spec at ast.py:8827>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8827
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicSubpDecl.subp_decl_spec'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Subp_Decl_P_Subp_Decl_Spec (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4215
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld_2 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 4216
Get_Result := Get (Fld_2, 0, True); 
--# expr-done 4217
Var_Expr := Get_Result;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 4219 '<FieldAccess for formal_type at ast.py:8827>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8827
--# expr-start 4218 '<FieldAccess for formal_decl at ast.py:8827>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8827
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_3 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 4218
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamDecl.formal_type
Fld_4 := Libadalang.Implementation.Base_Formal_Param_Decl_P_Formal_Type (Node => Fld_3.Node, Origin => Origin, E_Info => Fld_3.Info);
--# end
--# expr-done 4219
Scope_Result := Fld_4;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
      Cast_Result := Create_Internal_Entity
        (Node => Result_Var.Node,
         Info => Result_Var.Info);
Bind_Result := Solver.Create_Assign (Fld, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8826") else null)); 
--# expr-done 4220
--# expr-start 4221 '<FieldAccess for matches_expected_prefix_type at ast.py:8829>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8829
      if Prefix = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_prefix_type
Fld_5 := Libadalang.Implementation.Expr_P_Matches_Expected_Prefix_Type (Node => Prefix, Origin => Origin);
--# end
--# expr-done 4221
And_Pred := Create_And (Bind_Result, Fld_5, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8826") else null)); 
--# expr-done 4222
   If_Result := And_Pred;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8830") else null)); 
   If_Result := True_Rel;
end if;
      Inc_Ref (If_Result);
--# expr-done 4223
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 4224
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_720;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_720;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Subp_Decl_P_Constrain_Prefix;
--# end
--# property-start BasicSubpDecl.expr_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8834
pragma Warnings (Off, "is not referenced");
 function Basic_Subp_Decl_P_Expr_Type
  (Node : Bare_Basic_Subp_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Subp_Decl :=
     Internal_Entity_Basic_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Base_Subp_Spec;
Fld : Internal_Entity_Base_Subp_Spec;
Fld_1 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4227 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4225 '<FieldAccess for subp_spec_or_null at ast.py:8835>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8835
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4225
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Subp_Spec then
   --# scope-start
--# expr-start 4226 '<FieldAccess for return_type at ast.py:8835>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8835
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.return_type
Fld_1 := Libadalang.Implementation.Base_Subp_Spec_P_Return_Type (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 4226
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 4227
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Subp_Decl_P_Expr_Type;
--# end
--# property-start BasicSubpDecl.env_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8838
pragma Warnings (Off, "is not referenced");
 function Basic_Subp_Decl_P_Env_Names
  (Node : Bare_Basic_Subp_Decl
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_722 with Inline_Always;
            procedure Finalizer_Scope_2342 with Inline_Always;
      Var_Expr : String_Type;
Fld : String_Type;
Sym : Symbol_Type;
Singleton : Symbol_Type_Array_Access;
Scope_Result : Symbol_Type_Array_Access;
Result_Var : Symbol_Type_Array_Access;
Let_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_722 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_722;
            procedure Finalizer_Scope_2342 is
            begin
                     Dec_Ref (Var_Expr);
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2342;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4230 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4228 '<FieldAccess for top_level_env_name at ast.py:8839>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8839
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.top_level_env_name
Fld := Libadalang.Implementation.Basic_Decl_P_Top_Level_Env_Name (Node => Self);
--# end
--# expr-done 4228
Var_Expr := Fld;
      Inc_Ref (Var_Expr);
if Var_Expr /= Empty_String then
   --# scope-start
   --# bind fqn Var_Expr
--# expr-start 4229 '<to_symbol at ast.py:8840>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8840
Sym := String_To_Symbol (Self.Unit.Context, Var_Expr); 
--# expr-done 4229
                Singleton := Create_Symbol_Type_Array (Items_Count => 1);
                Singleton.Items (1) := Sym;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2342;
   Result_Var := Scope_Result;
else
   Result_Var := No_Symbol_Type_Array_Type;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 4230
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_722;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_722;
               Finalizer_Scope_2342;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Basic_Subp_Decl_P_Env_Names;
--# end
--# property-start '[dispatcher]BasicSubpDecl.subp_decl_spec' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Basic_Subp_Decl_P_Subp_Decl_Spec
  (Node : Bare_Basic_Subp_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Subp_Spec
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Subp_Spec;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Basic_Subp_Decl (Self.Kind) is
               when Ada_Abstract_Subp_Decl .. Ada_Subp_Decl =>
                  --# property-call-start ClassicSubpDecl.subp_decl_spec
                  Property_Result := Classic_Subp_Decl_P_Subp_Decl_Spec
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Entry_Decl =>
                  --# property-call-start EntryDecl.subp_decl_spec
                  Property_Result := Entry_Decl_P_Subp_Decl_Spec
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Enum_Literal_Decl .. Ada_Synthetic_Char_Enum_Lit =>
                  --# property-call-start EnumLiteralDecl.subp_decl_spec
                  Property_Result := Enum_Literal_Decl_P_Subp_Decl_Spec
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Subp_Internal =>
                  --# property-call-start GenericSubpInternal.subp_decl_spec
                  Property_Result := Generic_Subp_Internal_P_Subp_Decl_Spec
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Synthetic_Subp_Decl =>
                  --# property-call-start SyntheticSubpDecl.subp_decl_spec
                  Property_Result := Synthetic_Subp_Decl_P_Subp_Decl_Spec
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Basic_Subp_Decl_P_Subp_Decl_Spec;
--# end
--# property-start BasicSubpDecl._env_do_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8850
pragma Warnings (Off, "is not referenced");
 function Env_Do_25
  (Node : Bare_Basic_Subp_Decl
  )
   return Boolean
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4231 '<FieldAccess for env_hook at ast.py:8850>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8850
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.env_hook
Fld := Libadalang.Implementation.Ada_Node_P_Env_Hook (Node => Self);
--# end
--# expr-done 4231
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_25;
--# end
--# property-start BasicSubpDecl._initial_env_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8856
pragma Warnings (Off, "is not referenced");
 function Initial_Env_26
  (Node : Bare_Basic_Subp_Decl
  )
   return Internal_Designated_Env
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_1443 with Inline_Always;
      Fld : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_1443 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1443;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4232 '<FieldAccess for child_decl_initial_env at ast.py:8856>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8856
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.child_decl_initial_env
Fld := Libadalang.Implementation.Basic_Decl_P_Child_Decl_Initial_Env (Node => Self, Private_Part => True);
--# end
--# expr-done 4232
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1443;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1443;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Initial_Env_26;
--# end
--# property-start BasicSubpDecl._env_mappings_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8859
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_27
  (Node : Bare_Basic_Subp_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Basic_Subp_Decl :=
     Internal_Entity_Basic_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1444 with Inline_Always;
      Fld : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1444 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1444;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4233 '<FieldAccess for child_decl_env_assocs at ast.py:8859>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8859
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.child_decl_env_assocs
Fld := Libadalang.Implementation.Basic_Decl_P_Child_Decl_Env_Assocs (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4233
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1444;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1444;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_27;
--# end
--# property-start BasicSubpDecl._env_trans_parent_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8848
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_28
  (Node : Bare_Basic_Subp_Decl
  )
   return Boolean
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_28;
--# end
--# property-start BasicSubpDecl._env_names_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8861
pragma Warnings (Off, "is not referenced");
 function Env_Names_29
  (Node : Bare_Basic_Subp_Decl
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1446 with Inline_Always;
      Fld : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1446 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1446;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4234 '<FieldAccess for env_names at ast.py:8861>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8861
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicSubpDecl.env_names
Fld := Libadalang.Implementation.Basic_Subp_Decl_P_Env_Names (Node => Self);
--# end
--# expr-done 4234
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1446;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1446;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Names_29;
--# end
--# property-start BasicSubpDecl._env_do_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8863
pragma Warnings (Off, "is not referenced");
 function Env_Do_30
  (Node : Bare_Basic_Subp_Decl
  )
   return Bare_Compilation_Unit_Array_Access
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1447 with Inline_Always;
      Fld : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1447 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1447;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4235 '<FieldAccess for populate_dependent_units at ast.py:8863>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8863
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.populate_dependent_units
Fld := Libadalang.Implementation.Basic_Decl_P_Populate_Dependent_Units (Node => Self);
--# end
--# expr-done 4235
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1447;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1447;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_30;
--# end
--# property-start BasicSubpDecl._ref_env_nodes_31 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8866
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_31
  (Node : Bare_Basic_Subp_Decl
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1448 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1448 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1448;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4236 '<FieldAccess for top_level_use_package_clauses at ast.py:8866>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8866
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_package_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Package_Clauses (Node => Self);
--# end
--# expr-done 4236
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1448;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1448;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_31;
--# end
--# property-start BasicSubpDecl._ref_cond_32 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8868
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_32
  (Node : Bare_Basic_Subp_Decl
  )
   return Boolean
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4238 '<IsA LibraryItem, Subunit at ast.py:8868>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8868
--# expr-start 4237 '<parent at ast.py:8868>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8868
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 4237
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 4238
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_32;
--# end
--# property-start BasicSubpDecl._ref_env_nodes_33 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8872
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_33
  (Node : Bare_Basic_Subp_Decl
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1450 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1450 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1450;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4239 '<FieldAccess for top_level_use_type_clauses at ast.py:8872>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8872
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_type_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Type_Clauses (Node => Self);
--# end
--# expr-done 4239
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1450;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1450;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_33;
--# end
--# property-start BasicSubpDecl._ref_cond_34 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8874
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_34
  (Node : Bare_Basic_Subp_Decl
  )
   return Boolean
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4241 '<IsA LibraryItem, Subunit at ast.py:8874>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8874
--# expr-start 4240 '<parent at ast.py:8874>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8874
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 4240
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 4241
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_34;
--# end
--# property-start BasicSubpDecl._env_do_35 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8879
pragma Warnings (Off, "is not referenced");
 function Env_Do_35
  (Node : Bare_Basic_Subp_Decl
  )
   return Boolean
is
   Self : Bare_Basic_Subp_Decl := Bare_Basic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4242 '<FieldAccess for populate_body_unit at ast.py:8879>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8879
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.populate_body_unit
Fld := Libadalang.Implementation.Basic_Decl_P_Populate_Body_Unit (Node => Self);
--# end
--# expr-done 4242
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_35;
--# end
   --
   --  Primitives for Bare_Classic_Subp_Decl
   --
      procedure Initialize_Fields_For_Classic_Subp_Decl
        (Self : Bare_Classic_Subp_Decl
         ; Classic_Subp_Decl_F_Overriding : Bare_Overriding_Node
         ; Classic_Subp_Decl_F_Subp_Spec : Bare_Subp_Spec
        ) is
      begin
            Self.Classic_Subp_Decl_F_Overriding := Classic_Subp_Decl_F_Overriding;
            Self.Classic_Subp_Decl_F_Subp_Spec := Classic_Subp_Decl_F_Subp_Spec;
      end Initialize_Fields_For_Classic_Subp_Decl;
   function Classic_Subp_Decl_F_Overriding
     (Node : Bare_Classic_Subp_Decl) return Bare_Overriding_Node
   is
   begin
         return Node.Classic_Subp_Decl_F_Overriding;
   end;
   function Classic_Subp_Decl_F_Subp_Spec
     (Node : Bare_Classic_Subp_Decl) return Bare_Subp_Spec
   is
   begin
         return Node.Classic_Subp_Decl_F_Subp_Spec;
   end;
--# property-start ClassicSubpDecl.subp_decl_spec /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8893
pragma Warnings (Off, "is not referenced");
 function Classic_Subp_Decl_P_Subp_Decl_Spec
  (Node : Bare_Classic_Subp_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Subp_Spec
is
   Self : Bare_Classic_Subp_Decl := Bare_Classic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Classic_Subp_Decl :=
     Internal_Entity_Classic_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Subp_Spec;
      Fld : Internal_Entity_Subp_Spec;
Cast_Result : Internal_Entity_Base_Subp_Spec;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 4243 '<FieldAccess for subp_spec at ast.py:8893>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8893
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Subp_Spec (Node => Ent.Node.Classic_Subp_Decl_F_Subp_Spec, Info => Ent.Info);
--# expr-done 4243
      Cast_Result := Create_Internal_Entity_Base_Subp_Spec
        (Node => Fld.Node,
         Info => Fld.Info);
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Classic_Subp_Decl_P_Subp_Decl_Spec;
--# end
--# property-start ClassicSubpDecl.body_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8897
pragma Warnings (Off, "is not referenced");
 function Classic_Subp_Decl_P_Body_Part
  (Node : Bare_Classic_Subp_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Subp_Body
is
   Self : Bare_Classic_Subp_Decl := Bare_Classic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Classic_Subp_Decl :=
     Internal_Entity_Classic_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Base_Subp_Body;
      Fld : Internal_Entity_Body_Node;
Cast_Result : Internal_Entity_Base_Subp_Body;
Let_Result : Internal_Entity_Base_Subp_Body;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4246 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4245 '<Cast to BaseSubpBody at ast.py:8901>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8901
--# expr-start 4244 '<FieldAccess for body_part_for_decl at ast.py:8901>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8901
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.body_part_for_decl
Fld := Libadalang.Implementation.Basic_Decl_P_Body_Part_For_Decl (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 4244
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Base_Subp_Body
   then
      Cast_Result := Create_Internal_Entity_Base_Subp_Body
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 4245
Let_Result := Cast_Result; 
--# end
--# expr-done 4246
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Classic_Subp_Decl_P_Body_Part;
--# end
   --
   --  Primitives for Bare_Abstract_Subp_Decl
   --
      procedure Initialize_Fields_For_Abstract_Subp_Decl
        (Self : Bare_Abstract_Subp_Decl
         ; Classic_Subp_Decl_F_Overriding : Bare_Overriding_Node
         ; Classic_Subp_Decl_F_Subp_Spec : Bare_Subp_Spec
         ; Abstract_Subp_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Initialize_Fields_For_Classic_Subp_Decl
              (Self, Classic_Subp_Decl_F_Overriding, Classic_Subp_Decl_F_Subp_Spec);
            Self.Abstract_Subp_Decl_F_Aspects := Abstract_Subp_Decl_F_Aspects;
      end Initialize_Fields_For_Abstract_Subp_Decl;
   --
   --  Primitives for Bare_Formal_Subp_Decl
   --
      procedure Initialize_Fields_For_Formal_Subp_Decl
        (Self : Bare_Formal_Subp_Decl
         ; Classic_Subp_Decl_F_Overriding : Bare_Overriding_Node
         ; Classic_Subp_Decl_F_Subp_Spec : Bare_Subp_Spec
         ; Formal_Subp_Decl_F_Default_Expr : Bare_Expr
         ; Formal_Subp_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Initialize_Fields_For_Classic_Subp_Decl
              (Self, Classic_Subp_Decl_F_Overriding, Classic_Subp_Decl_F_Subp_Spec);
            Self.Formal_Subp_Decl_F_Default_Expr := Formal_Subp_Decl_F_Default_Expr;
            Self.Formal_Subp_Decl_F_Aspects := Formal_Subp_Decl_F_Aspects;
      end Initialize_Fields_For_Formal_Subp_Decl;
   function Formal_Subp_Decl_F_Default_Expr
     (Node : Bare_Formal_Subp_Decl) return Bare_Expr
   is
   begin
         return Node.Formal_Subp_Decl_F_Default_Expr;
   end;
--# property-start FormalSubpDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10611
pragma Warnings (Off, "is not referenced");
 function Formal_Subp_Decl_P_Defining_Names
  (Node : Bare_Formal_Subp_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Formal_Subp_Decl := Bare_Formal_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Formal_Subp_Decl :=
     Internal_Entity_Formal_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_837 with Inline_Always;
      Fld : Internal_Entity_Subp_Spec;
Fld_1 : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_837 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_837;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4248 '<FieldAccess for name at ast.py:10611>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10611
--# expr-start 4247 '<FieldAccess for subp_spec at ast.py:10611>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10611
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Subp_Spec (Node => Ent.Node.Classic_Subp_Decl_F_Subp_Spec, Info => Ent.Info);
--# expr-done 4247
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.name'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Name (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 4248
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld_1;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_837;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_837;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Formal_Subp_Decl_P_Defining_Names;
--# end
--# property-start FormalSubpDecl.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10613
pragma Warnings (Off, "is not referenced");
 function Formal_Subp_Decl_P_Xref_Equation
  (Node : Bare_Formal_Subp_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Formal_Subp_Decl := Bare_Formal_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Formal_Subp_Decl :=
     Internal_Entity_Formal_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_838 with Inline_Always;
            procedure Finalizer_Scope_2343 with Inline_Always;
            procedure Finalizer_Scope_2344 with Inline_Always;
            procedure Finalizer_Scope_2345 with Inline_Always;
            procedure Finalizer_Scope_2346 with Inline_Always;
      Var_Expr : Internal_Entity_Expr;
Match_0 : Internal_Entity_Null_Literal;
Match_1 : Internal_Entity_Box_Expr;
Match_2 : Internal_Entity_Name;
Match_3 : Internal_Entity_Expr;
Fld : Internal_Entity_Expr;
True_Rel : Logic_Equation;
True_Rel_1 : Logic_Equation;
Fld_1 : Logic_Equation;
Fld_2 : Logic_Var;
Cast_Expr : Internal_Entity_Formal_Subp_Decl;
Cast_Result : Internal_Entity_Basic_Decl;
Pred : Logic_Equation;
And_Pred : Logic_Equation;
False_Rel : Logic_Equation;
True_Rel_2 : Logic_Equation;
Is_A : Boolean;
If_Result : Logic_Equation;
And_Pred_1 : Logic_Equation;
Match_Prefix : Internal_Entity_Expr;
Cast_Result_1 : Internal_Entity_Null_Literal;
Let_Result : Logic_Equation;
Scope_Result : Logic_Equation;
Cast_Result_2 : Internal_Entity_Box_Expr;
Let_Result_1 : Logic_Equation;
Scope_Result_1 : Logic_Equation;
Cast_Result_3 : Internal_Entity_Name;
Let_Result_2 : Logic_Equation;
Scope_Result_2 : Logic_Equation;
Cast_Result_4 : Internal_Entity_Expr;
Let_Result_3 : Logic_Equation;
Scope_Result_3 : Logic_Equation;
Match_Result : Logic_Equation;
Scope_Result_4 : Logic_Equation;
True_Rel_3 : Logic_Equation;
Result_Var : Logic_Equation;
            procedure Finalizer_Scope_838 is
            begin
                     Dec_Ref (Scope_Result_4);
                     Dec_Ref (True_Rel_3);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_838;
            procedure Finalizer_Scope_2343 is
            begin
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Let_Result_2);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Let_Result_3);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Match_Result);
            end Finalizer_Scope_2343;
            procedure Finalizer_Scope_2344 is
            begin
                     Dec_Ref (True_Rel);
            end Finalizer_Scope_2344;
            procedure Finalizer_Scope_2345 is
            begin
                     Dec_Ref (True_Rel_1);
            end Finalizer_Scope_2345;
            procedure Finalizer_Scope_2346 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Pred);
                     Dec_Ref (And_Pred);
                     Dec_Ref (False_Rel);
                     Dec_Ref (True_Rel_2);
                     Dec_Ref (If_Result);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_2346;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
--# expr-start 4249 '<FieldAccess for default_expr at ast.py:10614>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10614
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Formal_Subp_Decl_F_Default_Expr, Info => Ent.Info);
--# expr-done 4249
Var_Expr := Fld;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind e Var_Expr
--# expr-start 4258 '<Match at ast.py:10614>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10614
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Var_Expr; 
case Ada_Expr (Match_Prefix.Node.Kind) is
         when Ada_Null_Literal =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Null_Literal
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_1; 
--# bind _ Match_0
True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10615") else null)); 
Let_Result := True_Rel; Inc_Ref (Let_Result);
--# end
Scope_Result := Let_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2344;
            Match_Result := Scope_Result; Inc_Ref (Match_Result);
         when Ada_Box_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Box_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_2; 
--# bind _ Match_1
True_Rel_1 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10617") else null)); 
Let_Result_1 := True_Rel_1; Inc_Ref (Let_Result_1);
--# end
Scope_Result_1 := Let_Result_1;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2345;
            Match_Result := Scope_Result_1; Inc_Ref (Match_Result);
         when Ada_Attribute_Ref .. Ada_String_Literal | Ada_Int_Literal .. Ada_Target_Name =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_3; 
--# bind n Match_2
--# expr-start 4257 '<And at ast.py:10619>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10619
--# expr-start 4254 '<And at ast.py:10620>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10620
--# expr-start 4250 '<FieldAccess for xref_no_overloading at ast.py:10620>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10620
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_1 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Match_2.Node, All_Els => True, Env => Env, Origin => Origin, E_Info => Match_2.Info);
--# end
--# expr-done 4250
--# expr-start 4253 '<Predicate on BasicDecl.subp_decl_match_signature at ast.py:10621>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10621
--# expr-start 4251 '<FieldAccess for ref_var at ast.py:10622>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10622
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Match_2.Node);
--# end
--# expr-done 4251
Fld_2.Value := No_Entity;
Entity_Vars.Reset (Fld_2);
--# expr-start 4252 '<Cast to BasicDecl at ast.py:10622>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10622
Cast_Expr := Ent; 
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
--# expr-done 4252
Pred := Solver.Create_Predicate (Fld_2, Create_Basic_Decl_P_Subp_Decl_Match_Signature_0_Predicate (Cast_Result)); 
--# expr-done 4253
And_Pred := Create_And (Fld_1, Pred, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10620") else null)); 
--# expr-done 4254
--# expr-start 4256 '<If at ast.py:10627>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10627
--# expr-start 4255 '<IsA AttributeRef at ast.py:10627>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10627
Is_A := Match_2.Node /= null 
and then Match_2.Node.Kind in Ada_Attribute_Ref_Range; 
--# expr-done 4255
if Is_A then
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10627") else null)); 
   If_Result := False_Rel;
else
   True_Rel_2 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10627") else null)); 
   If_Result := True_Rel_2;
end if;
      Inc_Ref (If_Result);
--# expr-done 4256
And_Pred_1 := Create_And (And_Pred, If_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10619") else null)); 
--# expr-done 4257
Let_Result_2 := And_Pred_1; Inc_Ref (Let_Result_2);
--# end
Scope_Result_2 := Let_Result_2;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2346;
            Match_Result := Scope_Result_2; Inc_Ref (Match_Result);
         when Ada_Abstract_State_Decl_Expr .. Ada_Relation_Op | Ada_Case_Expr_Alternative .. Ada_Membership_Expr | Ada_Paren_Expr .. Ada_Un_Op =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_4; 
--# bind _ Match_3
Let_Result_3 := raise Property_Error with "Should not happen"; Inc_Ref (Let_Result_3);
--# end
Scope_Result_3 := Let_Result_3;
   Inc_Ref (Scope_Result_3);
   --# end
            Match_Result := Scope_Result_3; Inc_Ref (Match_Result);
end case;
--# expr-done 4258
Scope_Result_4 := Match_Result;
   Inc_Ref (Scope_Result_4);
   --# end
      Finalizer_Scope_2343;
   Result_Var := Scope_Result_4;
else
   True_Rel_3 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10631") else null)); 
   Result_Var := True_Rel_3;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_838;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_838;
               Finalizer_Scope_2343;
               Finalizer_Scope_2344;
               Finalizer_Scope_2345;
               Finalizer_Scope_2346;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Formal_Subp_Decl_P_Xref_Equation;
--# end
--# property-start FormalSubpDecl.designated_subprogram_from /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10635
pragma Warnings (Off, "is not referenced");
 function Formal_Subp_Decl_P_Designated_Subprogram_From
  (Node : Bare_Formal_Subp_Decl
      ; Inst : Internal_Entity_Generic_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Formal_Subp_Decl := Bare_Formal_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Formal_Subp_Decl :=
     Internal_Entity_Formal_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind inst Inst
   Property_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_839 with Inline_Always;
      Block_Var : Internal_Entity_Array_Access;
Block_Var_1 : Internal_Entity_Subp_Spec;
Block_Var_2 : Internal_Entity_Basic_Decl;
Var_Expr : Internal_Entity_Base_Subp_Spec;
Var_Expr_1 : Internal_Entity_Basic_Decl;
Fld : Lexical_Env := Empty_Env;
Fld_1 : Internal_Entity_Defining_Name;
Fld_2 : Symbol_Type;
Fld_3 : Bare_Generic_Instantiation;
Cast_Result : Bare_Ada_Node;
Fld_4 : Internal_Entity_Array_Access;
Fld_5 : Bare_Subp_Spec;
Fld_6 : Internal_Entity_Info;
Fld_7 : Internal_Metadata;
Fld_8 : Internal_Entity_Info;
Fld_9 : Env_Rebindings;
Fld_10 : Bare_Ada_Node;
Fld_11 : Lexical_Env := Empty_Env;
Fld_12 : Lexical_Env := Empty_Env;
Rebinding : Env_Rebindings;
Fld_13 : Internal_Entity_Info;
Fld_14 : Boolean;
New_Struct : Internal_Entity_Info;
New_Struct_1 : Internal_Entity_Subp_Spec;
Cast_Expr : Internal_Entity;
Cast_Result_1 : Internal_Entity_Basic_Decl;
Fld_15 : Internal_Entity_Base_Subp_Spec;
Fld_16 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Find_Result : Internal_Entity;
Cast_Result_2 : Internal_Entity_Basic_Decl;
Fld_17 : Internal_Entity_Basic_Decl;
Scope_Result_1 : Internal_Entity_Basic_Decl;
Result_Var_1 : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_839 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Fld_11);
                     Dec_Ref (Fld_12);
            end Finalizer_Scope_839;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4283 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4263 '<FieldAccess for env_get at ast.py:10641>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10641
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4259 '<FieldAccess for node_env at ast.py:10642>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10642
      if Inst.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Node_Env (Node => Inst.Node, E_Info => Inst.Info);
--# expr-done 4259
--# expr-start 4261 '<FieldAccess for name_symbol at ast.py:10643>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10643
--# expr-start 4260 '<FieldAccess for defining_name at ast.py:10643>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10643
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4260
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_1.Node);
--# end
--# expr-done 4261
--# expr-start 4262 '<FieldAccess for node at ast.py:10644>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10644
Fld_3 := Inst.Node;
--# expr-done 4262
      Cast_Result := Fld_3;
--# property-call-start AdaNode.env_get
Fld_4 := Libadalang.Implementation.Ada_Node_P_Env_Get (Node => Self, Env => Fld, Symbol => Fld_2, From_Node => Cast_Result);
--# end
--# expr-done 4263
Block_Var := Fld_4; Inc_Ref (Block_Var);
--# bind subps Block_Var
--# expr-start 4276 '<New SubpSpec.entity at ???>' New_Struct_1 None
--# expr-start 4274 '<New EntityInfo at ast.py:10662>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10662
--# expr-start 4265 '<FieldAccess for from_rebound at ast.py:10669>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10669
--# expr-start 4264 '<FieldAccess for info at ast.py:10669>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10669
Fld_13 := Ent.Info;
--# expr-done 4264
Fld_14 := Fld_13.From_Rebound;
--# expr-done 4265
--# expr-start 4267 '<FieldAccess for md at ast.py:10663>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10663
--# expr-start 4266 '<FieldAccess for info at ast.py:10663>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10663
Fld_6 := Ent.Info;
--# expr-done 4266
Fld_7 := Fld_6.Md;
--# expr-done 4267
--# expr-start 4273 '<append_rebinding at ast.py:10664>' Rebinding /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10664
--# expr-start 4269 '<FieldAccess for rebindings at ast.py:10664>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10664
--# expr-start 4268 '<FieldAccess for info at ast.py:10664>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10664
Fld_8 := Ent.Info;
--# expr-done 4268
Fld_9 := Fld_8.Rebindings;
--# expr-done 4269
--# expr-start 4271 '<FieldAccess for node_env at ast.py:10667>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10667
--# expr-start 4270 '<parent at ast.py:10667>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10667
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Self.Parent; 
--# expr-done 4270
      if Fld_10 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Libadalang.Implementation.Node_Env (Node => Fld_10);
--# expr-done 4271
--# expr-start 4272 '<FieldAccess for instantiation_env at ast.py:10667>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10667
      if Inst.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericInstantiation.instantiation_env
Fld_12 := Libadalang.Implementation.Generic_Instantiation_F_Instantiation_Env (Node => Inst.Node);
--# end
--# expr-done 4272
Rebinding := AST_Envs.Append_Rebinding (Fld_9, Fld_11, Fld_12); 
--# expr-done 4273
New_Struct := (From_Rebound => Fld_14, Md => Fld_7, Rebindings => Rebinding); 
--# expr-done 4274
--# expr-start 4275 '<FieldAccess for subp_spec at ast.py:10661>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10661
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Self.Classic_Subp_Decl_F_Subp_Spec;
--# expr-done 4275
New_Struct_1 := (Info => New_Struct, Node => Fld_5); 
--# expr-done 4276
Block_Var_1 := New_Struct_1; 
--# bind actual_spec Block_Var_1
--# expr-start 4281 '<Cast to BasicDecl at ast.py:10674>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10674
--# expr-start 4280 '<Find at ast.py:10674>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10674
      Find_Result := No_Entity;
      declare
         Collection : constant Internal_Entity_Array_Access := Block_Var;
      begin
         for Item_166 of
               Collection.Items
         loop
   --# scope-start
               --# bind subp Item_166
--# expr-start 4278 '<FieldAccess for subp_spec_or_null at ast.py:10675>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10675
--# expr-start 4277 '<Cast to BasicDecl at ast.py:10675>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10675
Cast_Expr := Item_166; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result_1 := No_Entity_Basic_Decl;
   end if;
--# expr-done 4277
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_15 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Cast_Result_1.Node, E_Info => Cast_Result_1.Info);
--# end
--# expr-done 4278
Var_Expr := Fld_15;
if Var_Expr /= No_Entity_Base_Subp_Spec then
   --# scope-start
   --# bind spec Var_Expr
--# expr-start 4279 '<FieldAccess for match_signature at ast.py:10676>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10676
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.match_signature
Fld_16 := Libadalang.Implementation.Base_Subp_Spec_P_Match_Signature (Node => Block_Var_1.Node, Other => Var_Expr, Match_Name => False, E_Info => Block_Var_1.Info);
--# end
--# expr-done 4279
Scope_Result := Fld_16;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
      if Result_Var then
         Find_Result := Item_166;
      end if;
   --# end
               exit when Result_Var;
         end loop;
      end;
--# expr-done 4280
   if Find_Result.Node = null
      or else Find_Result.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_2 := Create_Internal_Entity_Basic_Decl
        (Node => Find_Result.Node,
         Info => Find_Result.Info);
   else
         Cast_Result_2 := No_Entity_Basic_Decl;
   end if;
--# expr-done 4281
Block_Var_2 := Cast_Result_2; 
--# bind found Block_Var_2
Var_Expr_1 := Block_Var_2;
if Var_Expr_1 /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 4282 '<FieldAccess for wrap_public_reference at ast.py:10687>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10687
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.wrap_public_reference
Fld_17 := Libadalang.Implementation.Basic_Decl_P_Wrap_Public_Reference (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 4282
Scope_Result_1 := Fld_17;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Basic_Decl;
end if;
Let_Result := Result_Var_1; 
--# end
--# expr-done 4283
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_839;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_839;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Formal_Subp_Decl_P_Designated_Subprogram_From;
--# end
   --
   --  Primitives for Bare_Abstract_Formal_Subp_Decl
   --
      procedure Initialize_Fields_For_Abstract_Formal_Subp_Decl
        (Self : Bare_Abstract_Formal_Subp_Decl
         ; Classic_Subp_Decl_F_Overriding : Bare_Overriding_Node
         ; Classic_Subp_Decl_F_Subp_Spec : Bare_Subp_Spec
         ; Formal_Subp_Decl_F_Default_Expr : Bare_Expr
         ; Formal_Subp_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Initialize_Fields_For_Formal_Subp_Decl
              (Self, Classic_Subp_Decl_F_Overriding, Classic_Subp_Decl_F_Subp_Spec, Formal_Subp_Decl_F_Default_Expr, Formal_Subp_Decl_F_Aspects);
      end Initialize_Fields_For_Abstract_Formal_Subp_Decl;
   --
   --  Primitives for Bare_Concrete_Formal_Subp_Decl
   --
      procedure Initialize_Fields_For_Concrete_Formal_Subp_Decl
        (Self : Bare_Concrete_Formal_Subp_Decl
         ; Classic_Subp_Decl_F_Overriding : Bare_Overriding_Node
         ; Classic_Subp_Decl_F_Subp_Spec : Bare_Subp_Spec
         ; Formal_Subp_Decl_F_Default_Expr : Bare_Expr
         ; Formal_Subp_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Initialize_Fields_For_Formal_Subp_Decl
              (Self, Classic_Subp_Decl_F_Overriding, Classic_Subp_Decl_F_Subp_Spec, Formal_Subp_Decl_F_Default_Expr, Formal_Subp_Decl_F_Aspects);
      end Initialize_Fields_For_Concrete_Formal_Subp_Decl;
   --
   --  Primitives for Bare_Subp_Decl
   --
      procedure Initialize_Fields_For_Subp_Decl
        (Self : Bare_Subp_Decl
         ; Classic_Subp_Decl_F_Overriding : Bare_Overriding_Node
         ; Classic_Subp_Decl_F_Subp_Spec : Bare_Subp_Spec
         ; Subp_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Initialize_Fields_For_Classic_Subp_Decl
              (Self, Classic_Subp_Decl_F_Overriding, Classic_Subp_Decl_F_Subp_Spec);
            Self.Subp_Decl_F_Aspects := Subp_Decl_F_Aspects;
      end Initialize_Fields_For_Subp_Decl;
--# property-start SubpDecl.is_constant_object /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8910
pragma Warnings (Off, "is not referenced");
 function Subp_Decl_P_Is_Constant_Object
  (Node : Bare_Subp_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Subp_Decl := Bare_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subp_Decl_P_Is_Constant_Object;
--# end
   --
   --  Primitives for Bare_Entry_Decl
   --
      procedure Entry_Decl_Pre_Env_Actions
        (Self            : Bare_Entry_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_131 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:16893");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_132 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Entry_Decl
        (Self : Bare_Entry_Decl
         ; Entry_Decl_F_Overriding : Bare_Overriding_Node
         ; Entry_Decl_F_Spec : Bare_Entry_Spec
         ; Entry_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Self.Entry_Decl_F_Overriding := Entry_Decl_F_Overriding;
            Self.Entry_Decl_F_Spec := Entry_Decl_F_Spec;
            Self.Entry_Decl_F_Aspects := Entry_Decl_F_Aspects;
      end Initialize_Fields_For_Entry_Decl;
   function Entry_Decl_F_Overriding
     (Node : Bare_Entry_Decl) return Bare_Overriding_Node
   is
   begin
         return Node.Entry_Decl_F_Overriding;
   end;
   function Entry_Decl_F_Spec
     (Node : Bare_Entry_Decl) return Bare_Entry_Spec
   is
   begin
         return Node.Entry_Decl_F_Spec;
   end;
--# property-start EntryDecl.subp_decl_spec /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16828
pragma Warnings (Off, "is not referenced");
 function Entry_Decl_P_Subp_Decl_Spec
  (Node : Bare_Entry_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Subp_Spec
is
   Self : Bare_Entry_Decl := Bare_Entry_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Decl :=
     Internal_Entity_Entry_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Subp_Spec;
      Fld : Internal_Entity_Entry_Spec;
Cast_Result : Internal_Entity_Base_Subp_Spec;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 4284 '<FieldAccess for spec at ast.py:16828>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16828
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Entry_Spec (Node => Ent.Node.Entry_Decl_F_Spec, Info => Ent.Info);
--# expr-done 4284
      Cast_Result := Create_Internal_Entity_Base_Subp_Spec
        (Node => Fld.Node,
         Info => Fld.Info);
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Decl_P_Subp_Decl_Spec;
--# end
--# property-start EntryDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16830
pragma Warnings (Off, "is not referenced");
 function Entry_Decl_P_Defining_Names
  (Node : Bare_Entry_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Entry_Decl := Bare_Entry_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Decl :=
     Internal_Entity_Entry_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1205 with Inline_Always;
      Fld : Internal_Entity_Entry_Spec;
Fld_1 : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1205 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1205;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4286 '<FieldAccess for name at ast.py:16830>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16830
--# expr-start 4285 '<FieldAccess for spec at ast.py:16830>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16830
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Entry_Spec (Node => Ent.Node.Entry_Decl_F_Spec, Info => Ent.Info);
--# expr-done 4285
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.name'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Name (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 4286
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld_1;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1205;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1205;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Decl_P_Defining_Names;
--# end
--# property-start EntryDecl.body_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16834
pragma Warnings (Off, "is not referenced");
 function Entry_Decl_P_Body_Part
  (Node : Bare_Entry_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Entry_Body
is
   Self : Bare_Entry_Decl := Bare_Entry_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Decl :=
     Internal_Entity_Entry_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Entry_Body;
      Fld : Internal_Entity_Body_Node;
Cast_Result : Internal_Entity_Entry_Body;
Let_Result : Internal_Entity_Entry_Body;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4289 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4288 '<Cast to EntryBody at ast.py:16838>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16838
--# expr-start 4287 '<FieldAccess for body_part_for_decl at ast.py:16838>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16838
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.body_part_for_decl
Fld := Libadalang.Implementation.Basic_Decl_P_Body_Part_For_Decl (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 4287
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Entry_Body_Range
   then
      Cast_Result := Create_Internal_Entity_Entry_Body
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 4288
Let_Result := Cast_Result; 
--# end
--# expr-done 4289
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Decl_P_Body_Part;
--# end
--# property-start EntryDecl.has_family /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16841
pragma Warnings (Off, "is not referenced");
 function Entry_Decl_P_Has_Family
  (Node : Bare_Entry_Decl
  )
   return Boolean
is
   Self : Bare_Entry_Decl := Bare_Entry_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Entry_Spec;
Fld_1 : Bare_Ada_Node;
Is_Null : Boolean;
Not_Val : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4294 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4293 '<Not at ast.py:16845>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16845
--# expr-start 4292 '<IsNull at ast.py:16845>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16845
--# expr-start 4291 '<FieldAccess for family_type at ast.py:16845>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16845
--# expr-start 4290 '<FieldAccess for spec at ast.py:16845>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16845
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Entry_Decl_F_Spec;
--# expr-done 4290
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Entry_Spec_F_Family_Type;
--# expr-done 4291
Is_Null := Fld_1 = null; 
--# expr-done 4292
Not_Val := not (Is_Null); 
--# expr-done 4293
Let_Result := Not_Val; 
--# end
--# expr-done 4294
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Decl_P_Has_Family;
--# end
--# property-start EntryDecl.family_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16848
pragma Warnings (Off, "is not referenced");
 function Entry_Decl_P_Family_Type
  (Node : Bare_Entry_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Entry_Decl := Bare_Entry_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Decl :=
     Internal_Entity_Entry_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Type_Expr;
Fld : Internal_Entity_Entry_Spec;
Fld_1 : Internal_Entity;
Cast_Result : Internal_Entity_Type_Expr;
Fld_2 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4299 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4297 '<Cast to TypeExpr at ast.py:16854>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16854
--# expr-start 4296 '<FieldAccess for family_type at ast.py:16854>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16854
--# expr-start 4295 '<FieldAccess for spec at ast.py:16854>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16854
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Entry_Spec (Node => Ent.Node.Entry_Decl_F_Spec, Info => Ent.Info);
--# expr-done 4295
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity (Node => Fld.Node.Entry_Spec_F_Family_Type, Info => Fld.Info);
--# expr-done 4296
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Type_Expr
   then
      Cast_Result := Create_Internal_Entity_Type_Expr
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         Cast_Result := No_Entity_Type_Expr;
   end if;
--# expr-done 4297
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Type_Expr then
   --# scope-start
--# expr-start 4298 '<FieldAccess for designated_type at ast.py:16854>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16854
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 4298
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 4299
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Decl_P_Family_Type;
--# end
--# property-start EntryDecl.accept_stmts /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16857
pragma Warnings (Off, "is not referenced");
 function Entry_Decl_P_Accept_Stmts
  (Node : Bare_Entry_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Accept_Stmt_Array_Access
is
   Self : Bare_Entry_Decl := Bare_Entry_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Decl :=
     Internal_Entity_Entry_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Accept_Stmt_Array_Access;
            procedure Finalizer_Scope_1209 with Inline_Always;
      Match_0 : Internal_Entity_Single_Task_Type_Decl;
Match_1 : Internal_Entity_Task_Type_Decl;
Match_2 : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Basic_Decl;
Match_Prefix : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity_Single_Task_Type_Decl;
Let_Result : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Cast_Result_1 : Internal_Entity_Task_Type_Decl;
Cast_Expr : Internal_Entity_Task_Type_Decl;
Cast_Result_2 : Internal_Entity_Basic_Decl;
Let_Result_1 : Internal_Entity_Basic_Decl;
Scope_Result_1 : Internal_Entity_Basic_Decl;
Cast_Result_3 : Internal_Entity_Basic_Decl;
Let_Result_2 : Internal_Entity_Basic_Decl;
Scope_Result_2 : Internal_Entity_Basic_Decl;
Match_Result : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity_Body_Node;
Cast_Result_4 : Internal_Entity;
Fld_3 : Internal_Entity_Accept_Stmt_Array_Access;
Let_Result_3 : Internal_Entity_Accept_Stmt_Array_Access;
            procedure Finalizer_Scope_1209 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Let_Result_3);
            end Finalizer_Scope_1209;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4305 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 4304 '<FieldAccess for find_accept_stmts at ast.py:16861>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16861
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4303 '<FieldAccess for body_part_for_decl at ast.py:16863>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16863
--# expr-start 4302 '<Match at ast.py:16863>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16863
--# expr-start 4300 '<FieldAccess for parent_basic_decl at ast.py:16863>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16863
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.parent_basic_decl
Fld := Libadalang.Implementation.Ada_Node_P_Parent_Basic_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4300
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld; 
case Ada_Basic_Decl (Match_Prefix.Node.Kind) is
         when Ada_Single_Task_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Single_Task_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind st Match_0
--# expr-start 4301 '<FieldAccess for parent_basic_decl at ast.py:16864>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16864
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.parent_basic_decl
Fld_1 := Libadalang.Implementation.Ada_Node_P_Parent_Basic_Decl (Node => Match_0.Node, E_Info => Match_0.Info);
--# end
--# expr-done 4301
Let_Result := Fld_1; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Task_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Task_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind tt Match_1
Cast_Expr := Match_1; 
      Cast_Result_2 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Let_Result_1 := Cast_Result_2; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abstract_State_Decl .. Ada_Protected_Type_Decl | Ada_Type_Decl .. Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Basic_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_3; 
--# bind _ Match_2
Let_Result_2 := raise Property_Error with "Can only be called on EntryDecl in Tasks"; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 4302
      if Match_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.body_part_for_decl
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Body_Part_For_Decl (Node => Match_Result.Node, Imprecise_Fallback => False, E_Info => Match_Result.Info);
--# end
--# expr-done 4303
      Cast_Result_4 := Create_Internal_Entity
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
--# property-call-start EntryDecl.find_accept_stmts
Fld_3 := Libadalang.Implementation.Entry_Decl_P_Find_Accept_Stmts (Node => Ent.Node, Root => Cast_Result_4, E_Info => Ent.Info);
--# end
--# expr-done 4304
Let_Result_3 := Fld_3; Inc_Ref (Let_Result_3);
--# end
--# expr-done 4305
      Property_Result := Let_Result_3;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1209;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1209;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Decl_P_Accept_Stmts;
--# end
--# property-start EntryDecl.find_accept_stmts /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16874
pragma Warnings (Off, "is not referenced");
 function Entry_Decl_P_Find_Accept_Stmts
  (Node : Bare_Entry_Decl
      ; Root : Internal_Entity
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Accept_Stmt_Array_Access
is
   Self : Bare_Entry_Decl := Bare_Entry_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Decl :=
     Internal_Entity_Entry_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind root Root
   Property_Result : Internal_Entity_Accept_Stmt_Array_Access;
            procedure Finalizer_Scope_1210 with Inline_Always;
            procedure Finalizer_Scope_2355 with Inline_Always;
            procedure Finalizer_Scope_2357 with Inline_Always;
            procedure Finalizer_Scope_2358 with Inline_Always;
      Var_Expr : Internal_Entity_Array_Access;
Var_Expr_1 : Internal_Entity_Accept_Stmt;
Fld : Internal_Entity_Array_Access;
Is_Null : Boolean;
Not_Val : Boolean;
Map_Result : Internal_Entity_Array_Access;
Fld_1 : Internal_Entity_Accept_Stmt_Array_Access;
Map_Result_1 : Internal_Entity_Accept_Stmt_Array_Access;
Scope_Result : Internal_Entity_Accept_Stmt_Array_Access;
Result_Var : Internal_Entity_Accept_Stmt_Array_Access;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Accept_Stmt;
Singleton : Internal_Entity_Accept_Stmt_Array_Access;
Fld_2 : Internal_Entity_Entry_Decl;
Cast_Result_1 : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Entry_Decl;
Cast_Result_2 : Internal_Entity;
Is_Equiv : Boolean;
If_Result : Internal_Entity_Accept_Stmt_Array_Access;
Scope_Result_1 : Internal_Entity_Accept_Stmt_Array_Access;
Result_Var_1 : Internal_Entity_Accept_Stmt_Array_Access;
Concat_Result : Internal_Entity_Accept_Stmt_Array_Access;
Let_Result : Internal_Entity_Accept_Stmt_Array_Access;
            procedure Finalizer_Scope_1210 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1210;
            procedure Finalizer_Scope_2355 is
            begin
                     Dec_Ref (Var_Expr);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Map_Result_1);
            end Finalizer_Scope_2355;
            procedure Finalizer_Scope_2357 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2357;
            procedure Finalizer_Scope_2358 is
            begin
                     Dec_Ref (Singleton);
                     Dec_Ref (If_Result);
            end Finalizer_Scope_2358;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4317 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4316 '<Concat at ast.py:16879>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16879
--# expr-start 4306 '<children at ast.py:16879>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16879
      if Root.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children (Node => Root.Node, E_Info => Root.Info);
--# expr-done 4306
Var_Expr := Fld;
      Inc_Ref (Var_Expr);
if Var_Expr /= No_Internal_Entity_Array_Type then
   --# scope-start
   --# bind child Var_Expr
--# expr-start 4311 '<Mapcat at ast.py:16880>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16880
   --# expr-start 4309 '<Filter at ast.py:16880>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16880
      declare
         Map_Result_Vec : Internal_Entity_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Var_Expr;
      begin
         for Item_168 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_168
         --# expr-start 4308 '<Not at ast.py:16880>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16880
--# expr-start 4307 '<IsNull at ast.py:16880>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16880
Is_Null := Item_168.Node = null; 
--# expr-done 4307
Not_Val := not (Is_Null); 
--# expr-done 4308
         if Not_Val then
         declare
            Item_To_Append : constant Internal_Entity := Item_168;
         begin
            Internal_Entity_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 4309
      declare
         Map_Result_1_Vec : Internal_Entity_Accept_Stmt_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Map_Result;
      begin
         for Item_167 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_167
      --# expr-start 4310 '<FieldAccess for find_accept_stmts at ast.py:16881>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16881
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start EntryDecl.find_accept_stmts
Fld_1 := Libadalang.Implementation.Entry_Decl_P_Find_Accept_Stmts (Node => Ent.Node, Root => Item_167, E_Info => Ent.Info);
--# end
--# expr-done 4310
         for Item_To_Append of
               Fld_1.Items
         loop
            Internal_Entity_Accept_Stmt_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_2357;
         end loop;
      end;
         Map_Result_1 := Create_Internal_Entity_Accept_Stmt_Array
           (Items_Count => Natural (Internal_Entity_Accept_Stmt_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Entity_Accept_Stmt_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Entity_Accept_Stmt_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Entity_Accept_Stmt_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 4311
Scope_Result := Map_Result_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2355;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Accept_Stmt_Array_Type;
end if;
      Inc_Ref (Result_Var);
--# expr-start 4312 '<Cast to AcceptStmt at ast.py:16883>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16883
Cast_Expr := Root; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Accept_Stmt_Range
   then
      Cast_Result := Create_Internal_Entity_Accept_Stmt
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Accept_Stmt;
   end if;
--# expr-done 4312
Var_Expr_1 := Cast_Result;
if Var_Expr_1 /= No_Entity_Accept_Stmt then
   --# scope-start
   --# bind stmt Var_Expr_1
--# expr-start 4315 '<If at ast.py:16884>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16884
--# expr-start 4314 '<Eq at ast.py:16885>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16885
--# expr-start 4313 '<FieldAccess for corresponding_entry at ast.py:16885>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16885
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AcceptStmt.corresponding_entry
Fld_2 := Libadalang.Implementation.Accept_Stmt_P_Corresponding_Entry (Node => Var_Expr_1.Node, Origin => No_Bare_Ada_Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 4313
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Cast_Expr_1 := Ent; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Is_Equiv := Equivalent (Cast_Result_1, Cast_Result_2); 
--# expr-done 4314
if Is_Equiv then
                Singleton := Create_Internal_Entity_Accept_Stmt_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr_1;
   If_Result := Singleton;
else
   If_Result := No_Internal_Entity_Accept_Stmt_Array_Type;
end if;
      Inc_Ref (If_Result);
--# expr-done 4315
Scope_Result_1 := If_Result;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2358;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Internal_Entity_Accept_Stmt_Array_Type;
end if;
      Inc_Ref (Result_Var_1);
Concat_Result := Concat (Result_Var, Result_Var_1); 
--# expr-done 4316
Let_Result := Concat_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 4317
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1210;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1210;
               Finalizer_Scope_2355;
               Finalizer_Scope_2357;
               Finalizer_Scope_2358;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Decl_P_Find_Accept_Stmts;
--# end
--# property-start EntryDecl._env_mappings_131 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16893
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_131
  (Node : Bare_Entry_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Entry_Decl := Bare_Entry_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Decl :=
     Internal_Entity_Entry_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1548 with Inline_Always;
      Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Entry_Decl;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1548 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1548;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4320 '<New EnvAssoc at ast.py:16893>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16893
--# expr-start 4318 '<New DesignatedEnv at ast.py:16893>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16893
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 4318
--# expr-start 4319 '<FieldAccess for name_symbol at ast.py:16893>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16893
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 4319
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 4320
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1548;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1548;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_131;
--# end
--# property-start EntryDecl._env_trans_parent_132 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16892
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_132
  (Node : Bare_Entry_Decl
  )
   return Boolean
is
   Self : Bare_Entry_Decl := Bare_Entry_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_132;
--# end
   --
   --  Primitives for Bare_Enum_Literal_Decl
   --
      procedure Enum_Literal_Decl_Pre_Env_Actions
        (Self            : Bare_Enum_Literal_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_129 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:16120");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_130 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Enum_Literal_Decl
        (Self : Bare_Enum_Literal_Decl
         ; Enum_Literal_Decl_F_Name : Bare_Defining_Name
        ) is
      begin
            Self.Enum_Literal_Decl_F_Name := Enum_Literal_Decl_F_Name;
      end Initialize_Fields_For_Enum_Literal_Decl;
   function Enum_Literal_Decl_F_Name
     (Node : Bare_Enum_Literal_Decl) return Bare_Defining_Name
   is
   begin
         return Node.Enum_Literal_Decl_F_Name;
   end;
--# property-start EnumLiteralDecl.is_static_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16095
pragma Warnings (Off, "is not referenced");
 function Enum_Literal_Decl_P_Is_Static_Decl
  (Node : Bare_Enum_Literal_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Enum_Literal_Decl := Bare_Enum_Literal_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Literal_Decl_P_Is_Static_Decl;
--# end
--# property-start EnumLiteralDecl.is_constant_object /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16097
pragma Warnings (Off, "is not referenced");
 function Enum_Literal_Decl_P_Is_Constant_Object
  (Node : Bare_Enum_Literal_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Enum_Literal_Decl := Bare_Enum_Literal_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Literal_Decl_P_Is_Constant_Object;
--# end
--# property-start '[dispatcher]EnumLiteralDecl.enum_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Enum_Literal_Decl_P_Enum_Type
  (Node : Bare_Enum_Literal_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Decl
is
   Self : Bare_Enum_Literal_Decl := Bare_Enum_Literal_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Enum_Literal_Decl_Range (Self.Kind) is
               when Ada_Enum_Literal_Decl =>
                  --# property-call-start EnumLiteralDecl.enum_type
                  Property_Result := Enum_Literal_Decl_P_Enum_Type
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Synthetic_Char_Enum_Lit =>
                  --# property-call-start SyntheticCharEnumLit.enum_type
                  Property_Result := Synthetic_Char_Enum_Lit_P_Enum_Type
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Enum_Literal_Decl_P_Enum_Type;
--# end
--# property-start EnumLiteralDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16108
pragma Warnings (Off, "is not referenced");
 function Enum_Literal_Decl_P_Defining_Names
  (Node : Bare_Enum_Literal_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Enum_Literal_Decl := Bare_Enum_Literal_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Enum_Literal_Decl :=
     Internal_Entity_Enum_Literal_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1141 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1141 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1141;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4321 '<FieldAccess for name at ast.py:16108>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16108
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Enum_Literal_Decl_F_Name, Info => Ent.Info);
--# expr-done 4321
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1141;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1141;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Literal_Decl_P_Defining_Names;
--# end
--# property-start EnumLiteralDecl.synth_type_expr /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16111
pragma Warnings (Off, "is not referenced");
 function Enum_Literal_Decl_P_Synth_Type_Expr
  (Node : Bare_Enum_Literal_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Enum_Lit_Synth_Type_Expr
is
   Self : Bare_Enum_Literal_Decl := Bare_Enum_Literal_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Enum_Lit_Synth_Type_Expr;
      New_Node : Bare_Enum_Lit_Synth_Type_Expr;
As_Entity : Internal_Entity_Enum_Lit_Synth_Type_Expr;
Is_Null : Boolean;
If_Result : Internal_Entity_Enum_Lit_Synth_Type_Expr;
Let_Result : Internal_Entity_Enum_Lit_Synth_Type_Expr;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Enum_Literal_Decl_Enum_Literal_Decl_P_Synth_Type_Expr,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Enum_Literal_Decl,
                                As_Bare_Enum_Literal_Decl => Self);
               Mmz_K.Items (2) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Enum_Lit_Synth_Type_Expr;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 4324 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4323 '<as_entity at ast.py:16112>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16112
--# expr-start 4322 '<New EnumLitSynthTypeExpr at ast.py:16112>' New_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16112
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node := new Root_Node_Record
  (Ada_Enum_Lit_Synth_Type_Expr);
Initialize
  (Self => New_Node,
   Kind => Ada_Enum_Lit_Synth_Type_Expr,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node);
--# expr-done 4322
Is_Null := New_Node = null; 
if Is_Null then
   If_Result := No_Entity_Enum_Lit_Synth_Type_Expr;
else
As_Entity := (Info => E_Info, Node => New_Node); 
   If_Result := As_Entity;
end if;
--# expr-done 4323
Let_Result := If_Result; 
--# end
--# expr-done 4324
      Property_Result := Let_Result;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Enum_Lit_Synth_Type_Expr,
                        As_Internal_Entity_Enum_Lit_Synth_Type_Expr => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Literal_Decl_P_Synth_Type_Expr;
--# end
--# property-start EnumLiteralDecl.subp_decl_spec /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16115
pragma Warnings (Off, "is not referenced");
 function Enum_Literal_Decl_P_Subp_Decl_Spec
  (Node : Bare_Enum_Literal_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Subp_Spec
is
   Self : Bare_Enum_Literal_Decl := Bare_Enum_Literal_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Subp_Spec;
      New_Node : Bare_Enum_Subp_Spec;
As_Entity : Internal_Entity_Enum_Subp_Spec;
Is_Null : Boolean;
If_Result : Internal_Entity_Enum_Subp_Spec;
Let_Result : Internal_Entity_Enum_Subp_Spec;
Cast_Result : Internal_Entity_Base_Subp_Spec;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Enum_Literal_Decl_Enum_Literal_Decl_P_Subp_Decl_Spec,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Enum_Literal_Decl,
                                As_Bare_Enum_Literal_Decl => Self);
               Mmz_K.Items (2) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Base_Subp_Spec;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
--# expr-start 4327 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4326 '<as_entity at ???>' If_Result None
--# expr-start 4325 '<New EnumSubpSpec at ???>' New_Node None
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node := new Root_Node_Record
  (Ada_Enum_Subp_Spec);
Initialize
  (Self => New_Node,
   Kind => Ada_Enum_Subp_Spec,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node);
--# expr-done 4325
Is_Null := New_Node = null; 
if Is_Null then
   If_Result := No_Entity_Enum_Subp_Spec;
else
As_Entity := (Info => E_Info, Node => New_Node); 
   If_Result := As_Entity;
end if;
--# expr-done 4326
Let_Result := If_Result; 
--# end
--# expr-done 4327
      Cast_Result := Create_Internal_Entity_Base_Subp_Spec
        (Node => Let_Result.Node,
         Info => Let_Result.Info);
      Property_Result := Cast_Result;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Base_Subp_Spec,
                        As_Internal_Entity_Base_Subp_Spec => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Literal_Decl_P_Subp_Decl_Spec;
--# end
--# property-start '[dispatcher]EnumLiteralDecl.enum_decl_name' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Enum_Literal_Decl_P_Enum_Decl_Name
  (Node : Bare_Enum_Literal_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Enum_Literal_Decl := Bare_Enum_Literal_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Enum_Literal_Decl :=
     Internal_Entity_Enum_Literal_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Enum_Literal_Decl_Range (Self.Kind) is
               when Ada_Enum_Literal_Decl =>
                  --# property-call-start EnumLiteralDecl.enum_decl_name
                  Property_Result := Enum_Literal_Decl_P_Enum_Decl_Name
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Synthetic_Char_Enum_Lit =>
                  --# property-call-start SyntheticCharEnumLit.enum_decl_name
                  Property_Result := Synthetic_Char_Enum_Lit_P_Enum_Decl_Name
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Enum_Literal_Decl_P_Enum_Decl_Name;
--# end
--# property-start EnumLiteralDecl._env_mappings_129 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16120
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_129
  (Node : Bare_Enum_Literal_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Enum_Literal_Decl := Bare_Enum_Literal_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Enum_Literal_Decl :=
     Internal_Entity_Enum_Literal_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1546 with Inline_Always;
      Fld : Symbol_Type;
Fld_1 : Internal_Entity_Type_Decl;
Fld_2 : Lexical_Env := Empty_Env;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Enum_Literal_Decl;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1546 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1546;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4332 '<New EnvAssoc at ast.py:16120>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16120
--# expr-start 4330 '<New DesignatedEnv at ast.py:16121>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16121
--# expr-start 4329 '<FieldAccess for node_env at ast.py:16121>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16121
--# expr-start 4328 '<FieldAccess for enum_type at ast.py:16121>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16121
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]EnumLiteralDecl.enum_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Enum_Literal_Decl_P_Enum_Type (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4328
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Node_Env (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# expr-done 4329
Inc_Ref (Fld_2);
New_Struct := (Direct_Env => Fld_2, Env_Name => null, Kind => Direct_Env); 
--# expr-done 4330
--# expr-start 4331 '<FieldAccess for name_symbol at ast.py:16120>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16120
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Self);
--# end
--# expr-done 4331
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 4332
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1546;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1546;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_129;
--# end
--# property-start EnumLiteralDecl._env_trans_parent_130 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16118
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_130
  (Node : Bare_Enum_Literal_Decl
  )
   return Boolean
is
   Self : Bare_Enum_Literal_Decl := Bare_Enum_Literal_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_130;
--# end
--# property-start EnumLiteralDecl.enum_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16100
pragma Warnings (Off, "is not referenced");
 function Enum_Literal_Decl_P_Enum_Type
  (Node : Bare_Enum_Literal_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Decl
is
   Self : Bare_Enum_Literal_Decl := Bare_Enum_Literal_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Type_Decl;
            procedure Finalizer_Scope_1140 with Inline_Always;
      Node_Parents : Bare_Ada_Node_Array_Access;
Is_A : Boolean;
Find_Result : Bare_Ada_Node;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Cast_Result : Internal_Entity_Type_Decl;
Let_Result : Internal_Entity_Type_Decl;
            procedure Finalizer_Scope_1140 is
            begin
                     Dec_Ref (Node_Parents);
            end Finalizer_Scope_1140;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4338 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4337 '<Cast to TypeDecl at ast.py:16104>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16104
--# expr-start 4336 '<as_entity at ast.py:16104>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16104
--# expr-start 4335 '<Find at ast.py:16104>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16104
   --# expr-start 4333 '<FieldAccess for parents at ast.py:16104>' Node_Parents /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16104
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents := Parents (Self, True); 
--# expr-done 4333
      Find_Result := No_Bare_Ada_Node;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents;
      begin
         for Item_169 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_169
      --# expr-start 4334 '<IsA TypeDecl at ast.py:16105>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16105
Is_A := Item_169 /= null 
and then Item_169.Kind in Ada_Type_Decl_Range; 
--# expr-done 4334
      if Is_A then
         Find_Result := Item_169;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 4335
Is_Null := Find_Result = null; 
if Is_Null then
   If_Result := No_Entity;
else
As_Entity := (Info => E_Info, Node => Find_Result); 
   If_Result := As_Entity;
end if;
--# expr-done 4336
   if If_Result.Node = null
      or else If_Result.Node.Kind in Ada_Type_Decl_Range
   then
      Cast_Result := Create_Internal_Entity_Type_Decl
        (Node => If_Result.Node,
         Info => If_Result.Info);
   else
         Cast_Result := No_Entity_Type_Decl;
   end if;
--# expr-done 4337
Let_Result := Cast_Result; 
--# end
--# expr-done 4338
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1140;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1140;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Literal_Decl_P_Enum_Type;
--# end
--# property-start EnumLiteralDecl.enum_decl_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16129
pragma Warnings (Off, "is not referenced");
 function Enum_Literal_Decl_P_Enum_Decl_Name
  (Node : Bare_Enum_Literal_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Enum_Literal_Decl := Bare_Enum_Literal_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Enum_Literal_Decl :=
     Internal_Entity_Enum_Literal_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name;
      Fld : Internal_Entity_Defining_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4339 '<FieldAccess for name at ast.py:16129>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16129
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Enum_Literal_Decl_F_Name, Info => Ent.Info);
--# expr-done 4339
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Literal_Decl_P_Enum_Decl_Name;
--# end
   --
   --  Primitives for Bare_Synthetic_Char_Enum_Lit
   --
      procedure Initialize_Fields_For_Synthetic_Char_Enum_Lit
        (Self : Bare_Synthetic_Char_Enum_Lit
         ; Enum_Literal_Decl_F_Name : Bare_Defining_Name
        ) is
      begin
            Initialize_Fields_For_Enum_Literal_Decl
              (Self, Enum_Literal_Decl_F_Name);
      Self.Synthetic_Char_Enum_Lit_Char_Symbol := null;
      Self.Synthetic_Char_Enum_Lit_Enum_Type_Decl := No_Entity_Type_Decl;
      end Initialize_Fields_For_Synthetic_Char_Enum_Lit;
--# property-start SyntheticCharEnumLit.enum_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16140
pragma Warnings (Off, "is not referenced");
 function Synthetic_Char_Enum_Lit_P_Enum_Type
  (Node : Bare_Synthetic_Char_Enum_Lit
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Decl
is
   Self : Bare_Synthetic_Char_Enum_Lit := Bare_Synthetic_Char_Enum_Lit (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Synthetic_Char_Enum_Lit :=
     Internal_Entity_Synthetic_Char_Enum_Lit'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Type_Decl;
      Fld : Internal_Entity_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4340 '<FieldAccess for enum_type_decl at ast.py:16140>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16140
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Ent.Node.Synthetic_Char_Enum_Lit_Enum_Type_Decl;
--# expr-done 4340
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Synthetic_Char_Enum_Lit_P_Enum_Type;
--# end
--# property-start SyntheticCharEnumLit.expr /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16142
pragma Warnings (Off, "is not referenced");
 function Synthetic_Char_Enum_Lit_P_Expr
  (Node : Bare_Synthetic_Char_Enum_Lit
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Synthetic_Char_Enum_Lit := Bare_Synthetic_Char_Enum_Lit (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Synthetic_Char_Enum_Lit :=
     Internal_Entity_Synthetic_Char_Enum_Lit'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_1146 with Inline_Always;
      Fld : Internal_Entity_Defining_Name_Array_Access;
Get_Result : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_1146 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1146;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4342 '<at at ast.py:16142>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16142
--# expr-start 4341 '<FieldAccess for defining_names at ast.py:16142>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16142
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_names'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Names (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4341
Get_Result := Get (Fld, 0, True); 
--# expr-done 4342
      Property_Result := Get_Result;
   --# end
      Finalizer_Scope_1146;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1146;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Synthetic_Char_Enum_Lit_P_Expr;
--# end
--# property-start SyntheticCharEnumLit.enum_decl_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16145
pragma Warnings (Off, "is not referenced");
 function Synthetic_Char_Enum_Lit_P_Enum_Decl_Name
  (Node : Bare_Synthetic_Char_Enum_Lit
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Synthetic_Char_Enum_Lit := Bare_Synthetic_Char_Enum_Lit (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Synthetic_Char_Enum_Lit :=
     Internal_Entity_Synthetic_Char_Enum_Lit'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name;
      Fld : Internal_Entity_Defining_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4343 '<FieldAccess for expr at ast.py:16145>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16145
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SyntheticCharEnumLit.expr
Fld := Libadalang.Implementation.Synthetic_Char_Enum_Lit_P_Expr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4343
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Synthetic_Char_Enum_Lit_P_Enum_Decl_Name;
--# end
--# property-start SyntheticCharEnumLit.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16147
pragma Warnings (Off, "is not referenced");
 function Synthetic_Char_Enum_Lit_P_Defining_Names
  (Node : Bare_Synthetic_Char_Enum_Lit
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Synthetic_Char_Enum_Lit := Bare_Synthetic_Char_Enum_Lit (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1148 with Inline_Always;
      Fld : Symbol_Type;
Fld_1 : Bare_Defining_Name;
As_Entity : Internal_Entity_Defining_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Defining_Name;
Array_Lit : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1148 is
            begin
                     Dec_Ref (Array_Lit);
            end Finalizer_Scope_1148;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4347 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 4346 '<as_entity at ast.py:16148>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16148
--# expr-start 4345 '<FieldAccess for synthesize_defining_name at ast.py:16148>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16148
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4344 '<FieldAccess for char_symbol at ast.py:16148>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16148
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Synthetic_Char_Enum_Lit_Char_Symbol;
--# expr-done 4344
--# property-call-start AdaNode.synthesize_defining_name
Fld_1 := Libadalang.Implementation.Ada_Node_P_Synthesize_Defining_Name (Node => Self, Sym => Fld);
--# end
--# expr-done 4345
Is_Null := Fld_1 = null; 
if Is_Null then
   If_Result := No_Entity_Defining_Name;
else
As_Entity := (Info => E_Info, Node => Fld_1); 
   If_Result := As_Entity;
end if;
--# expr-done 4346
Array_Lit := Create_Internal_Entity_Defining_Name_Array (Internal_Internal_Entity_Defining_Name_Array'(1 => If_Result)); 
--# expr-done 4347
      Property_Result := Array_Lit;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1148;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1148;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Synthetic_Char_Enum_Lit_P_Defining_Names;
--# end
   --
   --  Primitives for Bare_Generic_Subp_Internal
   --
      procedure Generic_Subp_Internal_Pre_Env_Actions
        (Self            : Bare_Generic_Subp_Internal;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_105 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            Libadalang.Implementation.Env_Names_106 (Node => 
Self);
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Generic_Subp_Internal
        (Self : Bare_Generic_Subp_Internal
         ; Generic_Subp_Internal_F_Subp_Spec : Bare_Subp_Spec
         ; Generic_Subp_Internal_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Self.Generic_Subp_Internal_F_Subp_Spec := Generic_Subp_Internal_F_Subp_Spec;
            Self.Generic_Subp_Internal_F_Aspects := Generic_Subp_Internal_F_Aspects;
      end Initialize_Fields_For_Generic_Subp_Internal;
   function Generic_Subp_Internal_F_Subp_Spec
     (Node : Bare_Generic_Subp_Internal) return Bare_Subp_Spec
   is
   begin
         return Node.Generic_Subp_Internal_F_Subp_Spec;
   end;
--# property-start GenericSubpInternal.subp_decl_spec /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10780
pragma Warnings (Off, "is not referenced");
 function Generic_Subp_Internal_P_Subp_Decl_Spec
  (Node : Bare_Generic_Subp_Internal
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Subp_Spec
is
   Self : Bare_Generic_Subp_Internal := Bare_Generic_Subp_Internal (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Subp_Internal :=
     Internal_Entity_Generic_Subp_Internal'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Subp_Spec;
      Fld : Internal_Entity_Subp_Spec;
Cast_Result : Internal_Entity_Base_Subp_Spec;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 4348 '<FieldAccess for subp_spec at ast.py:10780>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10780
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Subp_Spec (Node => Ent.Node.Generic_Subp_Internal_F_Subp_Spec, Info => Ent.Info);
--# expr-done 4348
      Cast_Result := Create_Internal_Entity_Base_Subp_Spec
        (Node => Fld.Node,
         Info => Fld.Info);
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Subp_Internal_P_Subp_Decl_Spec;
--# end
--# property-start GenericSubpInternal._env_trans_parent_105 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10782
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_105
  (Node : Bare_Generic_Subp_Internal
  )
   return Boolean
is
   Self : Bare_Generic_Subp_Internal := Bare_Generic_Subp_Internal (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_105;
--# end
--# property-start GenericSubpInternal._env_names_106 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10782
pragma Warnings (Off, "is not referenced");
 function Env_Names_106
  (Node : Bare_Generic_Subp_Internal
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Generic_Subp_Internal := Bare_Generic_Subp_Internal (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1523 with Inline_Always;
      Fld : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1523 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1523;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4349 '<FieldAccess for env_names at ast.py:10782>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10782
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicSubpDecl.env_names
Fld := Libadalang.Implementation.Basic_Subp_Decl_P_Env_Names (Node => Self);
--# end
--# expr-done 4349
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1523;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1523;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Names_106;
--# end
   --
   --  Primitives for Bare_Synthetic_Subp_Decl
   --
      procedure Initialize_Fields_For_Synthetic_Subp_Decl
        (Self : Bare_Synthetic_Subp_Decl
         ; Synthetic_Subp_Decl_F_Spec : Bare_Base_Subp_Spec
        ) is
      begin
            Self.Synthetic_Subp_Decl_F_Spec := Synthetic_Subp_Decl_F_Spec;
      end Initialize_Fields_For_Synthetic_Subp_Decl;
   function Synthetic_Subp_Decl_F_Spec
     (Node : Bare_Synthetic_Subp_Decl) return Bare_Base_Subp_Spec
   is
   begin
         return Node.Synthetic_Subp_Decl_F_Spec;
   end;
--# property-start SyntheticSubpDecl.subp_decl_spec /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16794
pragma Warnings (Off, "is not referenced");
 function Synthetic_Subp_Decl_P_Subp_Decl_Spec
  (Node : Bare_Synthetic_Subp_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Subp_Spec
is
   Self : Bare_Synthetic_Subp_Decl := Bare_Synthetic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Synthetic_Subp_Decl :=
     Internal_Entity_Synthetic_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Subp_Spec;
      Fld : Internal_Entity_Base_Subp_Spec;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4350 '<FieldAccess for spec at ast.py:16794>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16794
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Base_Subp_Spec (Node => Ent.Node.Synthetic_Subp_Decl_F_Spec, Info => Ent.Info);
--# expr-done 4350
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Synthetic_Subp_Decl_P_Subp_Decl_Spec;
--# end
--# property-start SyntheticSubpDecl.next_part_for_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16797
pragma Warnings (Off, "is not referenced");
 function Synthetic_Subp_Decl_P_Next_Part_For_Decl
  (Node : Bare_Synthetic_Subp_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Synthetic_Subp_Decl := Bare_Synthetic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4351 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := No_Entity_Basic_Decl; 
--# end
--# expr-done 4351
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Synthetic_Subp_Decl_P_Next_Part_For_Decl;
--# end
--# property-start SyntheticSubpDecl.is_static_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16801
pragma Warnings (Off, "is not referenced");
 function Synthetic_Subp_Decl_P_Is_Static_Decl
  (Node : Bare_Synthetic_Subp_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Synthetic_Subp_Decl := Bare_Synthetic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4352 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := True; 
--# end
--# expr-done 4352
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Synthetic_Subp_Decl_P_Is_Static_Decl;
--# end
   --
   --  Primitives for Bare_Body_Node
   --
--# property-start Body.env_hook_body /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3324
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Env_Hook_Body
  (Node : Bare_Body_Node
  )
   return Boolean
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
            procedure Finalizer_Scope_213 with Inline_Always;
      Unused_1 : Internal_Unit;
Var_Expr : Bare_Subp_Body;
As_Entity : Internal_Entity_Body_Node;
Is_Null : Boolean;
If_Result : Internal_Entity_Body_Node;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Symbol_Type_Array_Access;
Is_A : Boolean;
Not_Val : Boolean;
Fld_2 : Internal_Unit;
Cast_Expr : Bare_Body_Node;
Cast_Result : Bare_Subp_Body;
Fld_3 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Let_Result : Boolean;
Is_A_1 : Boolean;
If_Result_1 : Boolean;
Let_Result_1 : Boolean;
            procedure Finalizer_Scope_213 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_213;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4364 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 4363 '<If at ast.py:3328>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3328
--# expr-start 4353 '<IsA PackageBody, SubpBody at ast.py:3331>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3331
Is_A_1 := Self /= null 
and then Self.Kind in Ada_Package_Body_Range | Ada_Subp_Body_Range; 
--# expr-done 4353
if Is_A_1 then
   --# expr-start 4362 '<Let at ast.py:3333>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3333
--# scope-start
--# expr-start 4359 '<FieldAccess for get_unit at ast.py:3334>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3334
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4356 '<FieldAccess for as_symbol_array at ast.py:3335>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3335
--# expr-start 4355 '<FieldAccess for defining_name at ast.py:3335>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3335
--# expr-start 4354 '<as_bare_entity at ast.py:3335>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3335
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity_Body_Node;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 4354
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 4355
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.as_symbol_array'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_As_Symbol_Array (Node => Fld.Node);
--# end
--# expr-done 4356
--# expr-start 4358 '<Not at ast.py:3338>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3338
--# expr-start 4357 '<IsA SubpBody at ast.py:3338>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3338
Is_A := Self /= null 
and then Self.Kind in Ada_Subp_Body_Range; 
--# expr-done 4357
Not_Val := not (Is_A); 
--# expr-done 4358
Fld_2 := Libadalang.Implementation.Extensions.Ada_Node_P_Get_Unit (Node => Self, Name => Fld_1, Kind => Unit_Specification, Load_If_Needed => True, Not_Found_Is_Error => Not_Val);
--# expr-done 4359
Unused_1 := Fld_2; 
--# bind _ Unused_1
--# expr-start 4360 '<Cast to SubpBody at ast.py:3343>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3343
Cast_Expr := Self; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_Subp_Body_Range
   then
      Cast_Result := Cast_Expr;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 4360
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind subp_body Var_Expr
--# expr-start 4361 '<FieldAccess for env_hook_basic_decl at ast.py:3344>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3344
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.env_hook_basic_decl
Fld_3 := Libadalang.Implementation.Basic_Decl_P_Env_Hook_Basic_Decl (Node => Var_Expr);
--# end
--# expr-done 4361
Scope_Result := Fld_3;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 4362
   If_Result_1 := Let_Result;
else
   If_Result_1 := False;
end if;
--# expr-done 4363
Let_Result_1 := If_Result_1; 
--# end
--# expr-done 4364
      Property_Result := Let_Result_1;
   --# end
      Finalizer_Scope_213;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_213;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Env_Hook_Body;
--# end
--# property-start Body.subunit_decl_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3352
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Subunit_Decl_Env
  (Node : Bare_Body_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Body_Node :=
     Internal_Entity_Body_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_214 with Inline_Always;
            procedure Finalizer_Scope_2363 with Inline_Always;
            procedure Finalizer_Scope_2364 with Inline_Always;
            procedure Finalizer_Scope_2366 with Inline_Always;
            procedure Finalizer_Scope_2367 with Inline_Always;
      Var_Expr : Lexical_Env := Empty_Env;
Match_0 : Internal_Entity_Generic_Package_Decl;
Match_1 : Internal_Entity_Basic_Decl;
Match_2 : Internal_Entity;
Var_Expr_1 : Internal_Entity;
Fld : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_1 : Lexical_Env := Empty_Env;
Fld_2 : Symbol_Type;
Env_Get_Result : Internal_Entity_Array_Access;
Get_Result : Internal_Entity;
Fld_3 : Lexical_Env := Empty_Env;
Fld_4 : Internal_Entity_Generic_Package_Internal;
Fld_5 : Lexical_Env := Empty_Env;
Array_Lit : Lexical_Env_Array_Access;
Group_Env : Lexical_Env := Empty_Env;
Fld_6 : Lexical_Env := Empty_Env;
Match_Prefix : Internal_Entity;
Cast_Result : Internal_Entity_Generic_Package_Decl;
Let_Result : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Cast_Result_1 : Internal_Entity_Basic_Decl;
Let_Result_1 : Lexical_Env := Empty_Env;
Scope_Result_1 : Lexical_Env := Empty_Env;
Cast_Result_2 : Internal_Entity;
Let_Result_2 : Lexical_Env := Empty_Env;
Scope_Result_2 : Lexical_Env := Empty_Env;
Match_Result : Lexical_Env := Empty_Env;
Sym : Symbol_Type;
Env_Get_Result_1 : Internal_Entity_Array_Access;
Get_Result_1 : Internal_Entity;
Fld_7 : Lexical_Env := Empty_Env;
Array_Lit_1 : Lexical_Env_Array_Access;
Group_Env_1 : Lexical_Env := Empty_Env;
Scope_Result_3 : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Scope_Result_4 : Lexical_Env := Empty_Env;
Result_Var_1 : Lexical_Env := Empty_Env;
Dyn_Var_Bind_Result : Lexical_Env := Empty_Env;
Let_Result_3 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_214 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Env_Get_Result);
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Let_Result_2);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Match_Result);
                     Dec_Ref (Scope_Result_4);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Let_Result_3);
            end Finalizer_Scope_214;
            procedure Finalizer_Scope_2363 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Fld_5);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Group_Env);
            end Finalizer_Scope_2363;
            procedure Finalizer_Scope_2364 is
            begin
                     Dec_Ref (Fld_6);
            end Finalizer_Scope_2364;
            procedure Finalizer_Scope_2366 is
            begin
                     Dec_Ref (Var_Expr);
                     Dec_Ref (Env_Get_Result_1);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_2366;
            procedure Finalizer_Scope_2367 is
            begin
                     Dec_Ref (Fld_7);
                     Dec_Ref (Array_Lit_1);
                     Dec_Ref (Group_Env_1);
            end Finalizer_Scope_2367;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4384 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 4383 '<bind at ast.py:3353>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3353
--# expr-start 4365 '<FieldAccess for default_initial_env at ast.py:3354>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3354
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.default_initial_env
Fld := Libadalang.Implementation.Ada_Node_P_Default_Initial_Env (Node => Self);
--# end
--# expr-done 4365
Bound_Env := Fld; Inc_Ref (Bound_Env);
--# expr-start 4376 '<Match at ast.py:3355>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3355
--# expr-start 4369 '<at at ast.py:3355>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3355
--# expr-start 4368 '<EnvGet at ast.py:3355>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3355
--# expr-start 4366 '<FieldAccess for body_scope at ast.py:3355>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3355
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.body_scope
Fld_1 := Libadalang.Implementation.Body_Node_P_Body_Scope (Node => Ent.Node, Follow_Private => True, Env => Bound_Env, E_Info => Ent.Info);
--# end
--# expr-done 4366
--# expr-start 4367 '<FieldAccess for name_symbol at ast.py:3356>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3356
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 4367
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Fld_1, Key => Fld_2, Lookup_Kind => To_Lookup_Kind_Type (Recursive), Categories => (Inherited_Primitives => False, Nocat => True))); 
--# expr-done 4368
Get_Result := Get (Env_Get_Result, 0, True); 
--# expr-done 4369
      if Get_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Get_Result; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Generic_Package_Decl =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Generic_Package_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind gpd Match_0
--# expr-start 4374 '<env_group at ast.py:3363>' Group_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3363
--# expr-start 4373 '<ArrayLiteral at ast.py:3363>' Array_Lit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3363
--# expr-start 4370 '<FieldAccess for children_env at ast.py:3363>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3363
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Children_Env (Node => Match_0.Node, E_Info => Match_0.Info);
--# expr-done 4370
--# expr-start 4372 '<FieldAccess for children_env at ast.py:3363>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3363
--# expr-start 4371 '<FieldAccess for package_decl at ast.py:3363>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3363
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Generic_Package_Internal (Node => Match_0.Node.Generic_Package_Decl_F_Package_Decl, Info => Match_0.Info);
--# expr-done 4371
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Children_Env (Node => Fld_4.Node, E_Info => Fld_4.Info);
--# expr-done 4372
Array_Lit := Create_Lexical_Env_Array (Internal_Lexical_Env_Array'(1 => Fld_3, 2 => Fld_5)); 
--# expr-done 4373
Group_Env := Group (Array_Lit, No_Metadata); 
--# expr-done 4374
Let_Result := Group_Env; Inc_Ref (Let_Result);
--# end
Scope_Result := Let_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2363;
            Match_Result := Scope_Result; Inc_Ref (Match_Result);
         when Ada_Abstract_State_Decl .. Ada_For_Loop_Var_Decl | Ada_Generic_Subp_Decl .. Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Basic_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind pd Match_1
--# expr-start 4375 '<FieldAccess for children_env at ast.py:3365>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3365
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Children_Env (Node => Match_1.Node, E_Info => Match_1.Info);
--# expr-done 4375
Let_Result_1 := Fld_6; Inc_Ref (Let_Result_1);
--# end
Scope_Result_1 := Let_Result_1;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2364;
            Match_Result := Scope_Result_1; Inc_Ref (Match_Result);
         when Ada_Abort_Absent .. Ada_Param_Assoc | Ada_Case_Stmt_Alternative .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind _ Match_2
Let_Result_2 := raise Property_Error; Inc_Ref (Let_Result_2);
--# end
Scope_Result_2 := Let_Result_2;
   Inc_Ref (Scope_Result_2);
   --# end
            Match_Result := Scope_Result_2; Inc_Ref (Match_Result);
end case;
--# expr-done 4376
Var_Expr := Match_Result;
      Inc_Ref (Var_Expr);
if Var_Expr /= Empty_Env then
   --# scope-start
   --# bind public_part Var_Expr
--# expr-start 4379 '<at at ast.py:3367>' Get_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3367
--# expr-start 4378 '<EnvGet at ast.py:3367>' Env_Get_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3367
--# expr-start 4377 '<Symbol __privatepart at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Privatepart); 
--# expr-done 4377
Env_Get_Result_1 := Construct_Entity_Array (AST_Envs.Get (Self => Var_Expr, Key => Sym, Lookup_Kind => To_Lookup_Kind_Type (Flat), Categories => (Inherited_Primitives => False, Nocat => True))); 
--# expr-done 4378
Get_Result_1 := Get (Env_Get_Result_1, 0, True); 
--# expr-done 4379
Var_Expr_1 := Get_Result_1;
if Var_Expr_1 /= No_Entity then
   --# scope-start
   --# bind pp Var_Expr_1
--# expr-start 4382 '<env_group at ast.py:3371>' Group_Env_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3371
--# expr-start 4381 '<ArrayLiteral at ast.py:3371>' Array_Lit_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3371
--# expr-start 4380 '<FieldAccess for children_env at ast.py:3371>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3371
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Children_Env (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# expr-done 4380
Array_Lit_1 := Create_Lexical_Env_Array (Internal_Lexical_Env_Array'(1 => Fld_7, 2 => Var_Expr)); 
--# expr-done 4381
Group_Env_1 := Group (Array_Lit_1, No_Metadata); 
--# expr-done 4382
Scope_Result_3 := Group_Env_1;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_2367;
   Result_Var := Scope_Result_3;
else
   Result_Var := Var_Expr;
end if;
      Inc_Ref (Result_Var);
Scope_Result_4 := Result_Var;
   Inc_Ref (Scope_Result_4);
   --# end
      Finalizer_Scope_2366;
   Result_Var_1 := Scope_Result_4;
else
   Result_Var_1 := Empty_Env;
end if;
      Inc_Ref (Result_Var_1);
Dyn_Var_Bind_Result := Result_Var_1; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 4383
Let_Result_3 := Dyn_Var_Bind_Result; Inc_Ref (Let_Result_3);
--# end
--# expr-done 4384
      Property_Result := Let_Result_3;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_214;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_214;
               Finalizer_Scope_2363;
               Finalizer_Scope_2364;
               Finalizer_Scope_2366;
               Finalizer_Scope_2367;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Subunit_Decl_Env;
--# end
--# property-start Body.in_scope /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3377
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_In_Scope
  (Node : Bare_Body_Node
      ; Origin : Bare_Ada_Node
  )
   return Boolean
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
            procedure Finalizer_Scope_215 with Inline_Always;
      Is_Null : Boolean;
Not_Val : Boolean;
Fld : Internal_Unit;
Fld_1 : Internal_Unit;
Is_Equal : Boolean;
If_Result : Boolean;
Node_Parents : Bare_Ada_Node_Array_Access;
Cast_Expr : Bare_Body_Node;
Cast_Result : Bare_Ada_Node;
Is_Equal_1 : Boolean;
Find_Result : Bare_Ada_Node;
Is_Null_1 : Boolean;
Not_Val_1 : Boolean;
If_Result_1 : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_215 is
            begin
                     Dec_Ref (Node_Parents);
            end Finalizer_Scope_215;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4395 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4386 '<Not at ast.py:3382>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3382
--# expr-start 4385 '<IsNull at ast.py:3382>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3382
Is_Null := Origin = null; 
--# expr-done 4385
Not_Val := not (Is_Null); 
--# expr-done 4386
if Not_Val then
   --# expr-start 4389 '<Eq at ast.py:3383>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3383
--# expr-start 4387 '<FieldAccess for unit at ast.py:3383>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3383
      if Origin = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Unit (Node => Origin);
--# expr-done 4387
--# expr-start 4388 '<FieldAccess for unit at ast.py:3383>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3383
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Unit (Node => Self);
--# expr-done 4388
Is_Equal := Fld = Fld_1; 
--# expr-done 4389
   If_Result := Is_Equal;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 4394 '<Not at ast.py:3384>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3384
--# expr-start 4393 '<IsNull at ast.py:3384>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3384
--# expr-start 4392 '<Find at ast.py:3384>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3384
   --# expr-start 4390 '<FieldAccess for parents at ast.py:3384>' Node_Parents /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3384
      if Origin = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents := Parents (Origin, True); 
--# expr-done 4390
      Find_Result := No_Bare_Ada_Node;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents;
      begin
         for Item_170 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_170
      --# expr-start 4391 '<Eq at ast.py:3384>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3384
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Is_Equal_1 := Cast_Result = Item_170; 
--# expr-done 4391
      if Is_Equal_1 then
         Find_Result := Item_170;
      end if;
   --# end
               exit when Is_Equal_1;
         end loop;
      end;
--# expr-done 4392
Is_Null_1 := Find_Result = null; 
--# expr-done 4393
Not_Val_1 := not (Is_Null_1); 
--# expr-done 4394
   If_Result_1 := Not_Val_1;
else
   If_Result_1 := False;
end if;
Let_Result := If_Result_1; 
--# end
--# expr-done 4395
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_215;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_215;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_In_Scope;
--# end
--# property-start Body.body_decl_scope /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3388
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Body_Decl_Scope
  (Node : Bare_Body_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Body_Node :=
     Internal_Entity_Body_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_216 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_1 : Lexical_Env := Empty_Env;
Dyn_Var_Bind_Result : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_216 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_216;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4399 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4398 '<bind at ast.py:3392>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3392
--# expr-start 4396 '<FieldAccess for default_initial_env at ast.py:3393>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3393
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.default_initial_env
Fld := Libadalang.Implementation.Ada_Node_P_Default_Initial_Env (Node => Self);
--# end
--# expr-done 4396
Bound_Env := Fld; Inc_Ref (Bound_Env);
--# expr-start 4397 '<FieldAccess for body_scope at ast.py:3394>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3394
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.body_scope
Fld_1 := Libadalang.Implementation.Body_Node_P_Body_Scope (Node => Ent.Node, Follow_Private => True, Force_Decl => True, Env => Bound_Env, E_Info => Ent.Info);
--# end
--# expr-done 4397
Dyn_Var_Bind_Result := Fld_1; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 4398
Let_Result := Dyn_Var_Bind_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 4399
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_216;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_216;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Body_Decl_Scope;
--# end
--# property-start Body.body_initial_env_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3398
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Body_Initial_Env_Name
  (Node : Bare_Body_Node
  )
   return Symbol_Type
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type;
            procedure Finalizer_Scope_217 with Inline_Always;
      Fld : Boolean;
Fld_1 : String_Type;
Str : String_Type;
Concat_Result : String_Type;
Sym : Symbol_Type;
Fld_2 : Boolean;
Fld_3 : String_Type;
Str_1 : String_Type;
Concat_Result_1 : String_Type;
Sym_1 : Symbol_Type;
If_Result : Symbol_Type;
If_Result_1 : Symbol_Type;
Let_Result : Symbol_Type;
            procedure Finalizer_Scope_217 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Str);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Str_1);
                     Dec_Ref (Concat_Result_1);
            end Finalizer_Scope_217;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4411 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4410 '<Cond at ast.py:3404>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3404
--# expr-start 4400 '<FieldAccess for is_library_item at ast.py:3405>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3405
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_library_item'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Library_Item (Node => Self);
--# end
--# expr-done 4400
if Fld then
   --# expr-start 4404 '<to_symbol at ast.py:3406>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3406
--# expr-start 4403 '<Concat at ast.py:3406>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3406
--# expr-start 4401 '<FieldAccess for top_level_env_name at ast.py:3406>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3406
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.top_level_env_name
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Top_Level_Env_Name (Node => Self);
--# end
--# expr-done 4401
--# expr-start 4402 '<String at ast.py:3406>' Str /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3406
Str := Create_String (".__privatepart"); 
--# expr-done 4402
Concat_Result := Concat_String (Fld_1, Str); 
--# expr-done 4403
Sym := String_To_Symbol (Self.Unit.Context, Concat_Result); 
--# expr-done 4404
   If_Result_1 := Sym;
else
--# expr-start 4405 '<FieldAccess for is_subunit at ast.py:3408>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3408
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.is_subunit
Fld_2 := Libadalang.Implementation.Body_Node_P_Is_Subunit (Node => Self);
--# end
--# expr-done 4405
if Fld_2 then
   --# expr-start 4409 '<to_symbol at ast.py:3409>' Sym_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3409
--# expr-start 4408 '<Concat at ast.py:3409>' Concat_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3409
--# expr-start 4406 '<FieldAccess for top_level_env_name at ast.py:3409>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3409
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.top_level_env_name
Fld_3 := Libadalang.Implementation.Basic_Decl_P_Top_Level_Env_Name (Node => Self);
--# end
--# expr-done 4406
--# expr-start 4407 '<String at ast.py:3409>' Str_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3409
Str_1 := Create_String ("__stub"); 
--# expr-done 4407
Concat_Result_1 := Concat_String (Fld_3, Str_1); 
--# expr-done 4408
Sym_1 := String_To_Symbol (Self.Unit.Context, Concat_Result_1); 
--# expr-done 4409
   If_Result := Sym_1;
else
   If_Result := null;
end if;
   If_Result_1 := If_Result;
end if;
--# expr-done 4410
Let_Result := If_Result_1; 
--# end
--# expr-done 4411
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_217;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_217;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Body_Initial_Env_Name;
--# end
--# property-start Body.body_initial_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3415
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Body_Initial_Env
  (Node : Bare_Body_Node
  )
   return Internal_Designated_Env
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_218 with Inline_Always;
            procedure Finalizer_Scope_2369 with Inline_Always;
      Var_Expr : Symbol_Type;
Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Scope_Result : Internal_Designated_Env := No_Designated_Env;
New_Struct_1 : Internal_Designated_Env := No_Designated_Env;
Result_Var : Internal_Designated_Env := No_Designated_Env;
Let_Result : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_218 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (New_Struct_1);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_218;
            procedure Finalizer_Scope_2369 is
            begin
                     Dec_Ref (New_Struct);
            end Finalizer_Scope_2369;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4415 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4412 '<FieldAccess for body_initial_env_name at ast.py:3422>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3422
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.body_initial_env_name
Fld := Libadalang.Implementation.Body_Node_P_Body_Initial_Env_Name (Node => Self);
--# end
--# expr-done 4412
Var_Expr := Fld;
if Var_Expr /= null then
   --# scope-start
   --# bind non_null_name Var_Expr
--# expr-start 4413 '<New DesignatedEnv at ???>' New_Struct None
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => Var_Expr, Kind => Named_Env); 
--# expr-done 4413
Scope_Result := New_Struct;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2369;
   Result_Var := Scope_Result;
else
   --# expr-start 4414 '<New DesignatedEnv at ast.py:3421>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3421
Inc_Ref (Empty_Env);
New_Struct_1 := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 4414
   Result_Var := New_Struct_1;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 4415
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_218;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_218;
               Finalizer_Scope_2369;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Body_Initial_Env;
--# end
--# property-start '[dispatcher]Body.previous_part_env_name' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Body_Node_P_Previous_Part_Env_Name
  (Node : Bare_Body_Node
  )
   return Symbol_Type
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Body_Node (Self.Kind) is
               when Ada_Package_Body_Stub .. Ada_Task_Body =>
                  --# property-call-start Body.previous_part_env_name
                  Property_Result := Body_Node_P_Previous_Part_Env_Name
                    (Self
                    );
                  --# end
               when Ada_Expr_Function .. Ada_Subp_Renaming_Decl =>
                  --# property-call-start BaseSubpBody.previous_part_env_name
                  Property_Result := Base_Subp_Body_P_Previous_Part_Env_Name
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Body_Node_P_Previous_Part_Env_Name;
--# end
--# property-start '[dispatcher]Body.previous_part_link_env_assoc' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Body_Node_P_Previous_Part_Link_Env_Assoc
  (Node : Bare_Body_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Body_Node :=
     Internal_Entity_Body_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Body_Node (Self.Kind) is
               when Ada_Expr_Function .. Ada_Subp_Renaming_Decl | Ada_Entry_Body .. Ada_Task_Body =>
                  --# property-call-start Body.previous_part_link_env_assoc
                  Property_Result := Body_Node_P_Previous_Part_Link_Env_Assoc
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Package_Body_Stub .. Ada_Task_Body_Stub =>
                  --# property-call-start BodyStub.previous_part_link_env_assoc
                  Property_Result := Body_Stub_P_Previous_Part_Link_Env_Assoc
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Body_Node_P_Previous_Part_Link_Env_Assoc;
--# end
--# property-start Body.subp_previous_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3471
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Subp_Previous_Part
  (Node : Bare_Body_Node
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Body_Node :=
     Internal_Entity_Body_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_221 with Inline_Always;
            procedure Finalizer_Scope_2372 with Inline_Always;
      Block_Var : Internal_Entity;
Block_Var_1 : Internal_Entity_Array_Access;
Block_Var_2 : Boolean;
Block_Var_3 : Internal_Entity_Basic_Decl;
Var_Expr : Internal_Entity_Compilation_Unit;
Match_0 : Internal_Entity_Generic_Subp_Decl;
Match_1 : Internal_Entity_Formal_Subp_Decl;
Match_2 : Internal_Entity_Basic_Subp_Decl;
Match_3 : Internal_Entity_Subp_Body_Stub;
Match_4 : Internal_Entity;
Fld : Internal_Entity;
Fld_1 : Internal_Entity;
Fld_2 : Internal_Entity;
Fld_3 : Boolean;
If_Result : Internal_Entity;
Is_Null : Boolean;
Is_A : Boolean;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Compilation_Unit;
Fld_4 : Internal_Entity_Compilation_Unit;
Fld_5 : Bare_Basic_Decl;
Cast_Result_1 : Bare_Ada_Node;
As_Entity : Internal_Entity;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity;
Singleton : Internal_Entity_Array_Access;
Fld_6 : Bare_Basic_Decl;
Is_A_1 : Boolean;
If_Result_2 : Internal_Entity_Array_Access;
Scope_Result : Internal_Entity_Array_Access;
Result_Var : Internal_Entity_Array_Access;
Is_A_2 : Boolean;
Singleton_1 : Internal_Entity_Array_Access;
Fld_7 : Lexical_Env := Empty_Env;
Fld_8 : Symbol_Type;
Env_Get_Result : Internal_Entity_Array_Access;
If_Result_3 : Internal_Entity_Array_Access;
If_Result_4 : Internal_Entity_Array_Access;
If_Result_5 : Internal_Entity_Array_Access;
Is_A_3 : Boolean;
Get_Result : Internal_Entity;
Is_Null_2 : Boolean;
Not_Val : Boolean;
Fld_9 : Bare_Ada_Node;
Cast_Expr_1 : Bare_Body_Node;
Cast_Result_2 : Bare_Ada_Node;
Is_Equal : Boolean;
Not_Val_1 : Boolean;
If_Result_6 : Boolean;
Fld_10 : Internal_Entity_Base_Subp_Spec;
Fld_11 : Internal_Entity_Base_Subp_Spec;
Cast_Result_3 : Internal_Entity_Subp_Spec;
Cast_Result_4 : Internal_Entity_Base_Subp_Spec;
Fld_12 : Boolean;
Fld_13 : Internal_Entity_Subp_Spec;
Fld_14 : Internal_Entity_Base_Subp_Spec;
Cast_Result_5 : Internal_Entity_Subp_Spec;
Cast_Result_6 : Internal_Entity_Base_Subp_Spec;
Fld_15 : Boolean;
Match_Prefix : Internal_Entity;
Cast_Result_7 : Internal_Entity_Generic_Subp_Decl;
Let_Result : Boolean;
Scope_Result_1 : Boolean;
Cast_Result_8 : Internal_Entity_Formal_Subp_Decl;
Let_Result_1 : Boolean;
Scope_Result_2 : Boolean;
Cast_Result_9 : Internal_Entity_Basic_Subp_Decl;
Let_Result_2 : Boolean;
Scope_Result_3 : Boolean;
Cast_Result_10 : Internal_Entity_Subp_Body_Stub;
Let_Result_3 : Boolean;
Scope_Result_4 : Boolean;
Cast_Result_11 : Internal_Entity;
Let_Result_4 : Boolean;
Scope_Result_5 : Boolean;
Match_Result : Boolean;
If_Result_7 : Boolean;
Find_Result : Internal_Entity;
If_Result_8 : Internal_Entity;
Cast_Result_12 : Internal_Entity_Basic_Decl;
Is_Null_3 : Boolean;
Not_Val_2 : Boolean;
Fld_16 : Bare_Ada_Node;
Cast_Expr_2 : Bare_Body_Node;
Cast_Result_13 : Bare_Ada_Node;
Is_Equal_1 : Boolean;
Not_Val_3 : Boolean;
If_Result_9 : Boolean;
Is_A_4 : Boolean;
If_Result_10 : Boolean;
Find_Result_1 : Internal_Entity;
Cast_Result_14 : Internal_Entity_Basic_Decl;
Is_Null_4 : Boolean;
If_Result_11 : Boolean;
If_Result_12 : Internal_Entity_Basic_Decl;
Let_Result_5 : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_221 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Singleton_1);
                     Dec_Ref (Fld_7);
                     Dec_Ref (Env_Get_Result);
                     Dec_Ref (If_Result_3);
                     Dec_Ref (If_Result_4);
                     Dec_Ref (If_Result_5);
                     Dec_Ref (Block_Var_1);
            end Finalizer_Scope_221;
            procedure Finalizer_Scope_2372 is
            begin
                     Dec_Ref (Singleton);
                     Dec_Ref (If_Result_2);
            end Finalizer_Scope_2372;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4465 '<Block at ???>' Let_Result_5 None
--# scope-start
--# expr-start 4420 '<If at ast.py:3481>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3481
--# expr-start 4416 '<FieldAccess for is_library_item at ast.py:3482>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3482
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_library_item'
Fld_3 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Library_Item (Node => Ent.Node);
--# end
--# expr-done 4416
if Fld_3 then
   --# expr-start 4418 '<parent at ast.py:3483>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3483
--# expr-start 4417 '<parent at ast.py:3483>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3483
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 4417
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Parent (Node => Fld.Node, E_Info => Fld.Info);
--# expr-done 4418
   If_Result := Fld_1;
else
   --# expr-start 4419 '<FieldAccess for semantic_parent at ast.py:3484>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3484
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.semantic_parent
Fld_2 := Libadalang.Implementation.Ada_Node_P_Semantic_Parent (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4419
   If_Result := Fld_2;
end if;
--# expr-done 4420
Block_Var := If_Result; 
--# bind parent Block_Var
--# expr-start 4435 '<Cond at ast.py:3487>' If_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3487
--# expr-start 4421 '<IsNull at ast.py:3488>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3488
Is_Null := Block_Var.Node = null; 
--# expr-done 4421
if Is_Null then
   If_Result_5 := No_Internal_Entity_Array_Type;
else
--# expr-start 4422 '<IsA CompilationUnit at ast.py:3493>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3493
Is_A := Block_Var.Node /= null 
and then Block_Var.Node.Kind in Ada_Compilation_Unit_Range; 
--# expr-done 4422
if Is_A then
--# expr-start 4424 '<FieldAccess for other_part at ast.py:3494>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3494
--# expr-start 4423 '<Cast to CompilationUnit at ast.py:3494>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3494
Cast_Expr := Block_Var; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Compilation_Unit_Range
   then
      Cast_Result := Create_Internal_Entity_Compilation_Unit
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Compilation_Unit;
   end if;
--# expr-done 4423
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.other_part
Fld_4 := Libadalang.Implementation.Compilation_Unit_P_Other_Part (Node => Cast_Result.Node);
--# end
--# expr-done 4424
Var_Expr := Fld_4;
if Var_Expr /= No_Entity_Compilation_Unit then
   --# scope-start
   --# bind cu Var_Expr
--# expr-start 4430 '<If at ast.py:3497>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3497
--# expr-start 4426 '<IsA BasicSubpDecl, GenericSubpDecl at ast.py:3498>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3498
--# expr-start 4425 '<FieldAccess for decl at ast.py:3498>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3498
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.decl
Fld_6 := Libadalang.Implementation.Compilation_Unit_P_Decl (Node => Var_Expr.Node);
--# end
--# expr-done 4425
Is_A_1 := Fld_6 /= null 
and then Fld_6.Kind in Ada_Basic_Subp_Decl | Ada_Generic_Subp_Decl_Range; 
--# expr-done 4426
if Is_A_1 then
   --# expr-start 4429 '<as_entity at ast.py:3499>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3499
--# expr-start 4428 '<Cast to AdaNode at ast.py:3499>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3499
--# expr-start 4427 '<FieldAccess for decl at ast.py:3499>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3499
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.decl
Fld_5 := Libadalang.Implementation.Compilation_Unit_P_Decl (Node => Var_Expr.Node);
--# end
--# expr-done 4427
      Cast_Result_1 := Fld_5;
--# expr-done 4428
Is_Null_1 := Cast_Result_1 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity;
else
As_Entity := (Info => E_Info, Node => Cast_Result_1); 
   If_Result_1 := As_Entity;
end if;
--# expr-done 4429
                Singleton := Create_Internal_Entity_Array (Items_Count => 1);
                Singleton.Items (1) := If_Result_1;
   If_Result_2 := Singleton;
else
   If_Result_2 := No_Internal_Entity_Array_Type;
end if;
      Inc_Ref (If_Result_2);
--# expr-done 4430
Scope_Result := If_Result_2;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2372;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Array_Type;
end if;
      Inc_Ref (Result_Var);
   If_Result_4 := Result_Var;
else
--# expr-start 4431 '<IsA BodyStub at ast.py:3506>' Is_A_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3506
Is_A_2 := Block_Var.Node /= null 
and then Block_Var.Node.Kind in Ada_Body_Stub; 
--# expr-done 4431
if Is_A_2 then
                Singleton_1 := Create_Internal_Entity_Array (Items_Count => 1);
                Singleton_1.Items (1) := Block_Var;
   If_Result_3 := Singleton_1;
else
   --# expr-start 4434 '<EnvGet at ast.py:3510>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3510
--# expr-start 4432 '<FieldAccess for immediate_declarative_region at ast.py:3510>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3510
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.immediate_declarative_region'
Fld_7 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Immediate_Declarative_Region (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 4432
--# expr-start 4433 '<FieldAccess for name_symbol at ast.py:3511>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3511
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld_8 := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 4433
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Fld_7, Key => Fld_8, Lookup_Kind => To_Lookup_Kind_Type (Minimal), Categories => (Inherited_Primitives => True, Nocat => True))); 
--# expr-done 4434
   If_Result_3 := Env_Get_Result;
end if;
      Inc_Ref (If_Result_3);
   If_Result_4 := If_Result_3;
end if;
      Inc_Ref (If_Result_4);
   If_Result_5 := If_Result_4;
end if;
      Inc_Ref (If_Result_5);
--# expr-done 4435
Block_Var_1 := If_Result_5; Inc_Ref (Block_Var_1);
--# bind elements Block_Var_1
--# expr-start 4436 '<IsA CompilationUnit, BodyStub at ast.py:3519>' Is_A_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3519
Is_A_3 := Block_Var.Node /= null 
and then Block_Var.Node.Kind in Ada_Compilation_Unit_Range | Ada_Body_Stub; 
--# expr-done 4436
Block_Var_2 := Is_A_3; 
--# bind already_precise Block_Var_2
--# expr-start 4454 '<Cast to BasicDecl.entity at ast.py:3521>' Cast_Result_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3521
--# expr-start 4453 '<If at ast.py:3521>' If_Result_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3521
if Block_Var_2 then
   --# expr-start 4437 '<at at ast.py:3523>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3523
Get_Result := Get (Block_Var_1, 0, True); 
--# expr-done 4437
   If_Result_8 := Get_Result;
else
   --# expr-start 4452 '<Find at ast.py:3524>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3524
      Find_Result := No_Entity;
      declare
         Collection : constant Internal_Entity_Array_Access := Block_Var_1;
      begin
         for Item_171 of
               Collection.Items
         loop
   --# scope-start
               --# bind sp Item_171
--# expr-start 4439 '<Not at ast.py:3525>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3525
--# expr-start 4438 '<IsNull at ast.py:3525>' Is_Null_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3525
Is_Null_2 := Item_171.Node = null; 
--# expr-done 4438
Not_Val := not (Is_Null_2); 
--# expr-done 4439
if Not_Val then
   --# expr-start 4442 '<Not at ast.py:3526>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3526
--# expr-start 4441 '<Eq at ast.py:3526>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3526
--# expr-start 4440 '<FieldAccess for node at ast.py:3526>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3526
Fld_9 := Item_171.Node;
--# expr-done 4440
Cast_Expr_1 := Self; 
      Cast_Result_2 := Cast_Expr_1;
Is_Equal := Fld_9 = Cast_Result_2; 
--# expr-done 4441
Not_Val_1 := not (Is_Equal); 
--# expr-done 4442
   If_Result_6 := Not_Val_1;
else
   If_Result_6 := False;
end if;
if If_Result_6 then
   --# expr-start 4451 '<Match at ast.py:3527>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3527
      if Item_171.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Item_171; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Generic_Subp_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_7 := Create_Internal_Entity_Generic_Subp_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_7; 
--# bind _ Match_0
Let_Result := True; 
--# end
Scope_Result_1 := Let_Result;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abstract_Formal_Subp_Decl .. Ada_Concrete_Formal_Subp_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_8 := Create_Internal_Entity_Formal_Subp_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_8; 
--# bind _ Match_1
Let_Result_1 := False; 
--# end
Scope_Result_2 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Abstract_Subp_Decl | Ada_Subp_Decl .. Ada_Synthetic_Subp_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_9 := Create_Internal_Entity_Basic_Subp_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_9; 
--# bind subp_decl Match_2
--# expr-start 4446 '<FieldAccess for match_signature at ast.py:3537>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3537
--# expr-start 4443 '<FieldAccess for subp_decl_spec at ast.py:3537>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3537
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicSubpDecl.subp_decl_spec'
Fld_10 := Libadalang.Implementation.Dispatcher_Basic_Subp_Decl_P_Subp_Decl_Spec (Node => Match_2.Node, E_Info => Match_2.Info);
--# end
--# expr-done 4443
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4445 '<Cast to SubpSpec at ast.py:3538>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3538
--# expr-start 4444 '<FieldAccess for subp_spec_or_null at ast.py:3538>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3538
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_11 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4444
   if Fld_11.Node = null
      or else Fld_11.Node.Kind in Ada_Subp_Spec_Range
   then
      Cast_Result_3 := Create_Internal_Entity_Subp_Spec
        (Node => Fld_11.Node,
         Info => Fld_11.Info);
   else
         Cast_Result_3 := No_Entity_Subp_Spec;
   end if;
--# expr-done 4445
      Cast_Result_4 := Create_Internal_Entity_Base_Subp_Spec
        (Node => Cast_Result_3.Node,
         Info => Cast_Result_3.Info);
--# property-call-start BaseSubpSpec.match_signature
Fld_12 := Libadalang.Implementation.Base_Subp_Spec_P_Match_Signature (Node => Fld_10.Node, Other => Cast_Result_4, Match_Name => True, Use_Entity_Info => False, E_Info => Fld_10.Info);
--# end
--# expr-done 4446
Let_Result_2 := Fld_12; 
--# end
Scope_Result_3 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_3; 
         when Ada_Subp_Body_Stub =>
   --# scope-start
--# scope-start
      Cast_Result_10 := Create_Internal_Entity_Subp_Body_Stub
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_10; 
--# bind subp_stub Match_3
--# expr-start 4450 '<FieldAccess for match_signature at ast.py:3545>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3545
--# expr-start 4447 '<FieldAccess for subp_spec at ast.py:3545>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3545
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Create_Internal_Entity_Subp_Spec (Node => Match_3.Node.Subp_Body_Stub_F_Subp_Spec, Info => Match_3.Info);
--# expr-done 4447
      if Fld_13.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4449 '<Cast to SubpSpec at ast.py:3546>' Cast_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3546
--# expr-start 4448 '<FieldAccess for subp_spec_or_null at ast.py:3546>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3546
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_14 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4448
   if Fld_14.Node = null
      or else Fld_14.Node.Kind in Ada_Subp_Spec_Range
   then
      Cast_Result_5 := Create_Internal_Entity_Subp_Spec
        (Node => Fld_14.Node,
         Info => Fld_14.Info);
   else
         Cast_Result_5 := No_Entity_Subp_Spec;
   end if;
--# expr-done 4449
      Cast_Result_6 := Create_Internal_Entity_Base_Subp_Spec
        (Node => Cast_Result_5.Node,
         Info => Cast_Result_5.Info);
--# property-call-start BaseSubpSpec.match_signature
Fld_15 := Libadalang.Implementation.Base_Subp_Spec_P_Match_Signature (Node => Fld_13.Node, Other => Cast_Result_6, Match_Name => True, Use_Entity_Info => False, E_Info => Fld_13.Info);
--# end
--# expr-done 4450
Let_Result_3 := Fld_15; 
--# end
Scope_Result_4 := Let_Result_3;
   --# end
            Match_Result := Scope_Result_4; 
         when Ada_Abort_Absent .. Ada_Synth_Anonymous_Type_Decl | Ada_Expr_Function .. Ada_Protected_Body_Stub | Ada_Task_Body_Stub .. Ada_Generic_Package_Decl | Ada_Generic_Package_Instantiation .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_11 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_4 := Cast_Result_11; 
--# bind _ Match_4
Let_Result_4 := False; 
--# end
Scope_Result_5 := Let_Result_4;
   --# end
            Match_Result := Scope_Result_5; 
end case;
--# expr-done 4451
   If_Result_7 := Match_Result;
else
   If_Result_7 := False;
end if;
      if If_Result_7 then
         Find_Result := Item_171;
      end if;
   --# end
               exit when If_Result_7;
         end loop;
      end;
--# expr-done 4452
   If_Result_8 := Find_Result;
end if;
--# expr-done 4453
   if If_Result_8.Node = null
      or else If_Result_8.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_12 := Create_Internal_Entity_Basic_Decl
        (Node => If_Result_8.Node,
         Info => If_Result_8.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 4454
Block_Var_3 := Cast_Result_12; 
--# bind precise Block_Var_3
--# expr-start 4464 '<If at ast.py:3557>' If_Result_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3557
--# expr-start 4455 '<IsNull at ast.py:3558>' Is_Null_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3558
Is_Null_4 := Block_Var_3.Node = null; 
--# expr-done 4455
if Is_Null_4 then
   If_Result_11 := Imprecise_Fallback;
else
   If_Result_11 := False;
end if;
if If_Result_11 then
   --# expr-start 4463 '<Cast to BasicDecl.entity at ast.py:3559>' Cast_Result_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3559
--# expr-start 4462 '<Find at ast.py:3559>' Find_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3559
      Find_Result_1 := No_Entity;
      declare
         Collection : constant Internal_Entity_Array_Access := Block_Var_1;
      begin
         for Item_172 of
               Collection.Items
         loop
   --# scope-start
               --# bind sp Item_172
--# expr-start 4457 '<Not at ast.py:3561>' Not_Val_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3561
--# expr-start 4456 '<IsNull at ast.py:3561>' Is_Null_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3561
Is_Null_3 := Item_172.Node = null; 
--# expr-done 4456
Not_Val_2 := not (Is_Null_3); 
--# expr-done 4457
if Not_Val_2 then
   --# expr-start 4460 '<Not at ast.py:3562>' Not_Val_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3562
--# expr-start 4459 '<Eq at ast.py:3562>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3562
--# expr-start 4458 '<FieldAccess for node at ast.py:3562>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3562
Fld_16 := Item_172.Node;
--# expr-done 4458
Cast_Expr_2 := Self; 
      Cast_Result_13 := Cast_Expr_2;
Is_Equal_1 := Fld_16 = Cast_Result_13; 
--# expr-done 4459
Not_Val_3 := not (Is_Equal_1); 
--# expr-done 4460
   If_Result_9 := Not_Val_3;
else
   If_Result_9 := False;
end if;
if If_Result_9 then
   --# expr-start 4461 '<IsA BasicSubpDecl, SubpBodyStub at ast.py:3563>' Is_A_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3563
Is_A_4 := Item_172.Node /= null 
and then Item_172.Node.Kind in Ada_Basic_Subp_Decl | Ada_Subp_Body_Stub_Range; 
--# expr-done 4461
   If_Result_10 := Is_A_4;
else
   If_Result_10 := False;
end if;
      if If_Result_10 then
         Find_Result_1 := Item_172;
      end if;
   --# end
               exit when If_Result_10;
         end loop;
      end;
--# expr-done 4462
   if Find_Result_1.Node = null
      or else Find_Result_1.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_14 := Create_Internal_Entity_Basic_Decl
        (Node => Find_Result_1.Node,
         Info => Find_Result_1.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 4463
   If_Result_12 := Cast_Result_14;
else
   If_Result_12 := Block_Var_3;
end if;
--# expr-done 4464
Let_Result_5 := If_Result_12; 
--# end
--# expr-done 4465
      Property_Result := Let_Result_5;
   --# end
      Finalizer_Scope_221;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_221;
               Finalizer_Scope_2372;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Subp_Previous_Part;
--# end
--# property-start Body.entry_previous_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3570
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Entry_Previous_Part
  (Node : Bare_Body_Node
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Entry_Decl
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Body_Node :=
     Internal_Entity_Body_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   Property_Result : Internal_Entity_Entry_Decl;
            procedure Finalizer_Scope_222 with Inline_Always;
      Block_Var : Internal_Entity_Entry_Completion_Formal_Params;
Match_0 : Internal_Entity_Entry_Decl;
Match_1 : Internal_Entity;
Cast_Expr : Internal_Entity_Body_Node;
Cast_Result : Internal_Entity_Entry_Body;
Fld : Internal_Entity_Entry_Completion_Formal_Params;
Fld_1 : Symbol_Type;
Env_Get_Result : Internal_Entity_Array_Access;
Is_Null : Boolean;
Not_Val : Boolean;
Fld_2 : Bare_Ada_Node;
Cast_Expr_1 : Bare_Body_Node;
Cast_Result_1 : Bare_Ada_Node;
Is_Equal : Boolean;
Not_Val_1 : Boolean;
If_Result : Boolean;
Fld_3 : Internal_Entity_Entry_Spec;
Cast_Expr_2 : Internal_Entity_Entry_Completion_Formal_Params;
Cast_Result_2 : Internal_Entity_Base_Formal_Param_Holder;
Fld_4 : Boolean;
Match_Prefix : Internal_Entity;
Cast_Result_3 : Internal_Entity_Entry_Decl;
Let_Result : Boolean;
Scope_Result : Boolean;
Cast_Result_4 : Internal_Entity;
Let_Result_1 : Boolean;
Scope_Result_1 : Boolean;
Match_Result : Boolean;
If_Result_1 : Boolean;
Find_Result : Internal_Entity;
Cast_Result_5 : Internal_Entity_Entry_Decl;
Let_Result_2 : Internal_Entity_Entry_Decl;
            procedure Finalizer_Scope_222 is
            begin
                     Dec_Ref (Env_Get_Result);
            end Finalizer_Scope_222;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4480 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 4467 '<FieldAccess for params at ast.py:3574>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3574
--# expr-start 4466 '<Cast to EntryBody at ast.py:3574>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3574
Cast_Expr := Ent; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Entry_Body_Range
   then
      Cast_Result := Create_Internal_Entity_Entry_Body
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 4466
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Entry_Completion_Formal_Params (Node => Cast_Result.Node.Entry_Body_F_Params, Info => Cast_Result.Info);
--# expr-done 4467
Block_Var := Fld; 
--# bind spec Block_Var
--# expr-start 4479 '<Cast to EntryDecl.entity at ast.py:3575>' Cast_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3575
--# expr-start 4478 '<Find at ast.py:3575>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3575
   --# expr-start 4469 '<EnvGet at ast.py:3575>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3575
--# expr-start 4468 '<FieldAccess for name_symbol at ast.py:3575>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3575
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 4468
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Env, Key => Fld_1, Lookup_Kind => To_Lookup_Kind_Type (Recursive), Categories => (Inherited_Primitives => False, Nocat => True))); 
--# expr-done 4469
      Find_Result := No_Entity;
      declare
         Collection : constant Internal_Entity_Array_Access := Env_Get_Result;
      begin
         for Item_173 of
               Collection.Items
         loop
   --# scope-start
               --# bind sp Item_173
--# expr-start 4471 '<Not at ast.py:3577>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3577
--# expr-start 4470 '<IsNull at ast.py:3577>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3577
Is_Null := Item_173.Node = null; 
--# expr-done 4470
Not_Val := not (Is_Null); 
--# expr-done 4471
if Not_Val then
   --# expr-start 4474 '<Not at ast.py:3578>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3578
--# expr-start 4473 '<Eq at ast.py:3578>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3578
--# expr-start 4472 '<FieldAccess for node at ast.py:3578>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3578
Fld_2 := Item_173.Node;
--# expr-done 4472
Cast_Expr_1 := Self; 
      Cast_Result_1 := Cast_Expr_1;
Is_Equal := Fld_2 = Cast_Result_1; 
--# expr-done 4473
Not_Val_1 := not (Is_Equal); 
--# expr-done 4474
   If_Result := Not_Val_1;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 4477 '<Match at ast.py:3579>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3579
      if Item_173.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Item_173; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Entry_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Entry_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_3; 
--# bind entry_decl Match_0
--# expr-start 4476 '<FieldAccess for match_formal_params at ast.py:3581>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3581
--# expr-start 4475 '<FieldAccess for spec at ast.py:3581>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3581
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Entry_Spec (Node => Match_0.Node.Entry_Decl_F_Spec, Info => Match_0.Info);
--# expr-done 4475
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Cast_Expr_2 := Block_Var; 
      Cast_Result_2 := Create_Internal_Entity_Base_Formal_Param_Holder
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
--# property-call-start BaseFormalParamHolder.match_formal_params
Fld_4 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Match_Formal_Params (Node => Fld_3.Node, Other => Cast_Result_2, E_Info => Fld_3.Info);
--# end
--# expr-done 4476
Let_Result := Fld_4; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Abort_Absent .. Ada_Subp_Decl | Ada_Enum_Literal_Decl .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_4; 
--# bind _ Match_1
Let_Result_1 := False; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
end case;
--# expr-done 4477
   If_Result_1 := Match_Result;
else
   If_Result_1 := False;
end if;
      if If_Result_1 then
         Find_Result := Item_173;
      end if;
   --# end
               exit when If_Result_1;
         end loop;
      end;
--# expr-done 4478
   if Find_Result.Node = null
      or else Find_Result.Node.Kind in Ada_Entry_Decl_Range
   then
      Cast_Result_5 := Create_Internal_Entity_Entry_Decl
        (Node => Find_Result.Node,
         Info => Find_Result.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 4479
Let_Result_2 := Cast_Result_5; 
--# end
--# expr-done 4480
      Property_Result := Let_Result_2;
   --# end
      Finalizer_Scope_222;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_222;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Entry_Previous_Part;
--# end
--# property-start Body.package_previous_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3589
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Package_Previous_Part
  (Node : Bare_Body_Node
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Body_Node :=
     Internal_Entity_Body_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   Property_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_223 with Inline_Always;
            procedure Finalizer_Scope_2383 with Inline_Always;
            procedure Finalizer_Scope_2384 with Inline_Always;
      Var_Expr : Internal_Entity_Compilation_Unit;
Var_Expr_1 : Internal_Entity;
Match_0 : Internal_Entity_Package_Decl;
Match_1 : Internal_Entity_Generic_Package_Decl;
Match_2 : Internal_Entity;
Fld : Bare_Compilation_Unit;
Fld_1 : Internal_Entity_Compilation_Unit;
Fld_2 : Bare_Basic_Decl;
As_Entity : Internal_Entity_Basic_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity;
Singleton : Internal_Entity_Array_Access;
Scope_Result : Internal_Entity_Array_Access;
Result_Var : Internal_Entity_Array_Access;
Scope_Result_1 : Internal_Entity_Array_Access;
Result_Var_1 : Internal_Entity_Array_Access;
Fld_3 : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_4 : Internal_Entity_Defining_Name;
Fld_5 : Internal_Entity_Array_Access;
Dyn_Var_Bind_Result : Internal_Entity_Array_Access;
Fld_6 : Boolean;
If_Result_1 : Internal_Entity_Array_Access;
Fld_7 : Internal_Entity_Generic_Package_Internal;
Match_Prefix : Internal_Entity;
Cast_Result_1 : Internal_Entity_Package_Decl;
Cast_Expr : Internal_Entity_Package_Decl;
Cast_Result_2 : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
Scope_Result_2 : Internal_Entity_Basic_Decl;
Cast_Result_3 : Internal_Entity_Generic_Package_Decl;
Cast_Result_4 : Internal_Entity_Basic_Decl;
Let_Result_1 : Internal_Entity_Basic_Decl;
Scope_Result_3 : Internal_Entity_Basic_Decl;
Cast_Result_5 : Internal_Entity;
Let_Result_2 : Internal_Entity_Basic_Decl;
Scope_Result_4 : Internal_Entity_Basic_Decl;
Match_Result : Internal_Entity_Basic_Decl;
Map_Result : Internal_Entity_Basic_Decl_Array_Access;
Is_Null_1 : Boolean;
Not_Val : Boolean;
Find_Result : Internal_Entity_Basic_Decl;
Let_Result_3 : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_223 is
            begin
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Fld_5);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_223;
            procedure Finalizer_Scope_2383 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_2383;
            procedure Finalizer_Scope_2384 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2384;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4498 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 4497 '<Find at ast.py:3596>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3596
   --# expr-start 4494 '<Map at ast.py:3596>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3596
   --# expr-start 4491 '<If at ast.py:3596>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3596
--# expr-start 4481 '<FieldAccess for is_library_item at ast.py:3597>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3597
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_library_item'
Fld_6 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Library_Item (Node => Self);
--# end
--# expr-done 4481
if Fld_6 then
--# expr-start 4483 '<FieldAccess for other_part at ast.py:3598>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3598
--# expr-start 4482 '<FieldAccess for enclosing_compilation_unit at ast.py:3598>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3598
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.enclosing_compilation_unit
Fld := Libadalang.Implementation.Ada_Node_P_Enclosing_Compilation_Unit (Node => Self);
--# end
--# expr-done 4482
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.other_part
Fld_1 := Libadalang.Implementation.Compilation_Unit_P_Other_Part (Node => Fld);
--# end
--# expr-done 4483
Var_Expr := Fld_1;
if Var_Expr /= No_Entity_Compilation_Unit then
   --# scope-start
   --# bind part Var_Expr
--# expr-start 4486 '<Cast to AdaNode at ast.py:3599>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3599
--# expr-start 4485 '<as_entity at ast.py:3599>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3599
--# expr-start 4484 '<FieldAccess for decl at ast.py:3599>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3599
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.decl
Fld_2 := Libadalang.Implementation.Compilation_Unit_P_Decl (Node => Var_Expr.Node);
--# end
--# expr-done 4484
Is_Null := Fld_2 = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Decl;
else
As_Entity := (Info => E_Info, Node => Fld_2); 
   If_Result := As_Entity;
end if;
--# expr-done 4485
      Cast_Result := Create_Internal_Entity
        (Node => If_Result.Node,
         Info => If_Result.Info);
--# expr-done 4486
Var_Expr_1 := Cast_Result;
if Var_Expr_1 /= No_Entity then
   --# scope-start
                Singleton := Create_Internal_Entity_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr_1;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2384;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Array_Type;
end if;
      Inc_Ref (Result_Var);
Scope_Result_1 := Result_Var;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2383;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Internal_Entity_Array_Type;
end if;
      Inc_Ref (Result_Var_1);
   If_Result_1 := Result_Var_1;
else
   --# expr-start 4490 '<bind at ast.py:3601>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3601
--# expr-start 4487 '<FieldAccess for origin_node at ast.py:3602>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3602
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_3 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 4487
Bound_Origin := Fld_3; 
--# expr-start 4489 '<FieldAccess for all_env_els_impl at ast.py:3602>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3602
--# expr-start 4488 '<FieldAccess for defining_name at ast.py:3602>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3602
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4488
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.all_env_els_impl'
Fld_5 := Libadalang.Implementation.Dispatcher_Name_P_All_Env_Els_Impl (Node => Fld_4.Node, Env => Env, Origin => Bound_Origin, E_Info => Fld_4.Info);
--# end
--# expr-done 4489
Dyn_Var_Bind_Result := Fld_5; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 4490
   If_Result_1 := Dyn_Var_Bind_Result;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 4491
      declare
         Map_Result_Vec : Internal_Entity_Basic_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := If_Result_1;
      begin
         for Item_175 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_175
      --# expr-start 4493 '<Match at ast.py:3605>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3605
      if Item_175.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Item_175; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Package_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Package_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_1; 
--# bind pkg_decl Match_0
Cast_Expr := Match_0; 
      Cast_Result_2 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Let_Result := Cast_Result_2; 
--# end
Scope_Result_2 := Let_Result;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Generic_Package_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Generic_Package_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_3; 
--# bind gen_pkg_decl Match_1
--# expr-start 4492 '<FieldAccess for package_decl at ast.py:3608>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3608
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Generic_Package_Internal (Node => Match_1.Node.Generic_Package_Decl_F_Package_Decl, Info => Match_1.Info);
--# expr-done 4492
      Cast_Result_4 := Create_Internal_Entity_Basic_Decl
        (Node => Fld_7.Node,
         Info => Fld_7.Info);
Let_Result_1 := Cast_Result_4; 
--# end
Scope_Result_3 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_3; 
         when Ada_Abort_Absent .. Ada_Generic_Package_Internal | Ada_Discrete_Base_Subtype_Decl .. Ada_For_Loop_Var_Decl | Ada_Generic_Subp_Decl .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_5; 
--# bind _ Match_2
Let_Result_2 := No_Entity_Basic_Decl; 
--# end
Scope_Result_4 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_4; 
end case;
--# expr-done 4493
         declare
            Item_To_Append : constant Internal_Entity_Basic_Decl := Match_Result;
         begin
            Internal_Entity_Basic_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Basic_Decl_Array
           (Items_Count => Natural (Internal_Entity_Basic_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Basic_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Basic_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Basic_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 4494
      Find_Result := No_Entity_Basic_Decl;
      declare
         Collection : constant Internal_Entity_Basic_Decl_Array_Access := Map_Result;
      begin
         for Item_174 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_174
      --# expr-start 4496 '<Not at ast.py:3612>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3612
--# expr-start 4495 '<IsNull at ast.py:3612>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3612
Is_Null_1 := Item_174.Node = null; 
--# expr-done 4495
Not_Val := not (Is_Null_1); 
--# expr-done 4496
      if Not_Val then
         Find_Result := Item_174;
      end if;
   --# end
               exit when Not_Val;
         end loop;
      end;
--# expr-done 4497
Let_Result_3 := Find_Result; 
--# end
--# expr-done 4498
      Property_Result := Let_Result_3;
   --# end
      Finalizer_Scope_223;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_223;
               Finalizer_Scope_2383;
               Finalizer_Scope_2384;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Package_Previous_Part;
--# end
--# property-start Body.task_previous_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3616
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Task_Previous_Part
  (Node : Bare_Body_Node
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Body_Node :=
     Internal_Entity_Body_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   Property_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_224 with Inline_Always;
      Fld : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_1 : Internal_Entity_Defining_Name;
Fld_2 : Internal_Entity_Array_Access;
Dyn_Var_Bind_Result : Internal_Entity_Array_Access;
Get_Result : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_224 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Dyn_Var_Bind_Result);
            end Finalizer_Scope_224;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4505 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4504 '<Cast to BasicDecl at ast.py:3620>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3620
--# expr-start 4503 '<at at ast.py:3620>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3620
--# expr-start 4502 '<bind at ast.py:3620>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3620
--# expr-start 4499 '<FieldAccess for origin_node at ast.py:3621>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3621
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 4499
Bound_Origin := Fld; 
--# expr-start 4501 '<FieldAccess for all_env_els_impl at ast.py:3621>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3621
--# expr-start 4500 '<FieldAccess for defining_name at ast.py:3621>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3621
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4500
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.all_env_els_impl'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_All_Env_Els_Impl (Node => Fld_1.Node, Env => Env, Origin => Bound_Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 4501
Dyn_Var_Bind_Result := Fld_2; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 4502
Get_Result := Get (Dyn_Var_Bind_Result, 0, True); 
--# expr-done 4503
   if Get_Result.Node = null
      or else Get_Result.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Get_Result.Node,
         Info => Get_Result.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 4504
Let_Result := Cast_Result; 
--# end
--# expr-done 4505
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_224;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_224;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Task_Previous_Part;
--# end
--# property-start Body.protected_previous_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3625
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Protected_Previous_Part
  (Node : Bare_Body_Node
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Body_Node :=
     Internal_Entity_Body_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   Property_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_225 with Inline_Always;
      Var_Expr : Internal_Entity;
Match_0 : Internal_Entity_Protected_Type_Decl;
Match_1 : Internal_Entity_Single_Protected_Decl;
Match_2 : Internal_Entity;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Internal_Entity_Array_Access;
Get_Result : Internal_Entity;
Match_Prefix : Internal_Entity;
Cast_Result : Internal_Entity_Protected_Type_Decl;
Cast_Expr : Internal_Entity_Protected_Type_Decl;
Cast_Result_1 : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Cast_Result_2 : Internal_Entity_Single_Protected_Decl;
Cast_Expr_1 : Internal_Entity_Single_Protected_Decl;
Cast_Result_3 : Internal_Entity_Basic_Decl;
Let_Result_1 : Internal_Entity_Basic_Decl;
Scope_Result_1 : Internal_Entity_Basic_Decl;
Cast_Result_4 : Internal_Entity;
Let_Result_2 : Internal_Entity_Basic_Decl;
Scope_Result_2 : Internal_Entity_Basic_Decl;
Match_Result : Internal_Entity_Basic_Decl;
Scope_Result_3 : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Let_Result_3 : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_225 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_225;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4510 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 4508 '<at at ast.py:3629>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3629
--# expr-start 4507 '<FieldAccess for env_elements at ast.py:3629>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3629
--# expr-start 4506 '<FieldAccess for defining_name at ast.py:3629>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3629
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4506
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.env_elements
Fld_1 := Libadalang.Implementation.Expr_P_Env_Elements (Node => Fld.Node, Env => Env, No_Visibility => False, E_Info => Fld.Info);
--# end
--# expr-done 4507
Get_Result := Get (Fld_1, 0, True); 
--# expr-done 4508
Var_Expr := Get_Result;
if Var_Expr /= No_Entity then
   --# scope-start
--# expr-start 4509 '<Match at ast.py:3629>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3629
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Var_Expr; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Protected_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Protected_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind prot_type Match_0
Cast_Expr := Match_0; 
      Cast_Result_1 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Let_Result := Cast_Result_1; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Single_Protected_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Single_Protected_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_2; 
--# bind prot_decl Match_1
Cast_Expr_1 := Match_1; 
      Cast_Result_3 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Let_Result_1 := Cast_Result_3; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abort_Absent .. Ada_Incomplete_Tagged_Type_Decl | Ada_Task_Type_Decl .. Ada_Package_Renaming_Decl | Ada_Single_Task_Decl .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_4; 
--# bind _ Match_2
Let_Result_2 := No_Entity_Basic_Decl; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 4509
Scope_Result_3 := Match_Result;
   --# end
   Result_Var := Scope_Result_3;
else
   Result_Var := No_Entity_Basic_Decl;
end if;
Let_Result_3 := Result_Var; 
--# end
--# expr-done 4510
      Property_Result := Let_Result_3;
   --# end
      Finalizer_Scope_225;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_225;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Protected_Previous_Part;
--# end
--# property-start Body.unbound_previous_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3637
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Unbound_Previous_Part
  (Node : Bare_Body_Node
      ; Env : Lexical_Env
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Body_Node :=
     Internal_Entity_Body_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Block_Var : Internal_Entity_Basic_Decl;
Match_0 : Internal_Entity_Base_Subp_Body;
Match_1 : Internal_Entity_Subp_Body_Stub;
Match_2 : Internal_Entity_Entry_Body;
Match_3 : Internal_Entity_Package_Body;
Match_4 : Internal_Entity_Package_Body_Stub;
Match_5 : Internal_Entity_Protected_Body;
Match_6 : Internal_Entity_Protected_Body_Stub;
Match_7 : Internal_Entity_Task_Body;
Match_8 : Internal_Entity_Task_Body_Stub;
Pp_Next_Part : Internal_Entity_Basic_Decl;
Var_Expr : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity_Entry_Decl;
Fld_3 : Internal_Entity_Basic_Decl;
Fld_4 : Internal_Entity_Basic_Decl;
Fld_5 : Internal_Entity_Basic_Decl;
Fld_6 : Internal_Entity_Basic_Decl;
Fld_7 : Internal_Entity_Basic_Decl;
Fld_8 : Internal_Entity_Basic_Decl;
Match_Prefix : Internal_Entity_Body_Node;
Cast_Result : Internal_Entity_Base_Subp_Body;
Let_Result : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Cast_Result_1 : Internal_Entity_Subp_Body_Stub;
Let_Result_1 : Internal_Entity_Basic_Decl;
Scope_Result_1 : Internal_Entity_Basic_Decl;
Cast_Result_2 : Internal_Entity_Entry_Body;
Cast_Result_3 : Internal_Entity_Basic_Decl;
Let_Result_2 : Internal_Entity_Basic_Decl;
Scope_Result_2 : Internal_Entity_Basic_Decl;
Cast_Result_4 : Internal_Entity_Package_Body;
Let_Result_3 : Internal_Entity_Basic_Decl;
Scope_Result_3 : Internal_Entity_Basic_Decl;
Cast_Result_5 : Internal_Entity_Package_Body_Stub;
Let_Result_4 : Internal_Entity_Basic_Decl;
Scope_Result_4 : Internal_Entity_Basic_Decl;
Cast_Result_6 : Internal_Entity_Protected_Body;
Let_Result_5 : Internal_Entity_Basic_Decl;
Scope_Result_5 : Internal_Entity_Basic_Decl;
Cast_Result_7 : Internal_Entity_Protected_Body_Stub;
Let_Result_6 : Internal_Entity_Basic_Decl;
Scope_Result_6 : Internal_Entity_Basic_Decl;
Cast_Result_8 : Internal_Entity_Task_Body;
Let_Result_7 : Internal_Entity_Basic_Decl;
Scope_Result_7 : Internal_Entity_Basic_Decl;
Cast_Result_9 : Internal_Entity_Task_Body_Stub;
Let_Result_8 : Internal_Entity_Basic_Decl;
Scope_Result_8 : Internal_Entity_Basic_Decl;
Match_Result : Internal_Entity_Basic_Decl;
Fld_9 : Internal_Entity_Basic_Decl;
Scope_Result_9 : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Is_A : Boolean;
If_Result : Internal_Entity_Basic_Decl;
Let_Result_9 : Internal_Entity_Basic_Decl;
Is_A_1 : Boolean;
If_Result_1 : Internal_Entity_Basic_Decl;
Let_Result_10 : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4527 '<Block at ???>' Let_Result_10 None
--# scope-start
--# expr-start 4520 '<Match at ast.py:3642>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3642
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Body_Node (Match_Prefix.Node.Kind) is
         when Ada_Expr_Function .. Ada_Subp_Renaming_Decl =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Base_Subp_Body
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind _ Match_0
--# expr-start 4511 '<FieldAccess for subp_previous_part at ast.py:3643>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3643
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.subp_previous_part
Fld := Libadalang.Implementation.Body_Node_P_Subp_Previous_Part (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 4511
Let_Result := Fld; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Subp_Body_Stub =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Subp_Body_Stub
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind _ Match_1
--# expr-start 4512 '<FieldAccess for subp_previous_part at ast.py:3644>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3644
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.subp_previous_part
Fld_1 := Libadalang.Implementation.Body_Node_P_Subp_Previous_Part (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 4512
Let_Result_1 := Fld_1; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Entry_Body =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Entry_Body
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind _ Match_2
--# expr-start 4513 '<FieldAccess for entry_previous_part at ast.py:3645>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3645
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.entry_previous_part
Fld_2 := Libadalang.Implementation.Body_Node_P_Entry_Previous_Part (Node => Ent.Node, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 4513
      Cast_Result_3 := Create_Internal_Entity_Basic_Decl
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Let_Result_2 := Cast_Result_3; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Package_Body =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Package_Body
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_4; 
--# bind _ Match_3
--# expr-start 4514 '<FieldAccess for package_previous_part at ast.py:3646>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3646
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.package_previous_part
Fld_3 := Libadalang.Implementation.Body_Node_P_Package_Previous_Part (Node => Ent.Node, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 4514
Let_Result_3 := Fld_3; 
--# end
Scope_Result_3 := Let_Result_3;
   --# end
            Match_Result := Scope_Result_3; 
         when Ada_Package_Body_Stub =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Create_Internal_Entity_Package_Body_Stub
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_4 := Cast_Result_5; 
--# bind _ Match_4
--# expr-start 4515 '<FieldAccess for package_previous_part at ast.py:3647>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3647
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.package_previous_part
Fld_4 := Libadalang.Implementation.Body_Node_P_Package_Previous_Part (Node => Ent.Node, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 4515
Let_Result_4 := Fld_4; 
--# end
Scope_Result_4 := Let_Result_4;
   --# end
            Match_Result := Scope_Result_4; 
         when Ada_Protected_Body =>
   --# scope-start
--# scope-start
      Cast_Result_6 := Create_Internal_Entity_Protected_Body
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_5 := Cast_Result_6; 
--# bind _ Match_5
--# expr-start 4516 '<FieldAccess for protected_previous_part at ast.py:3648>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3648
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.protected_previous_part
Fld_5 := Libadalang.Implementation.Body_Node_P_Protected_Previous_Part (Node => Ent.Node, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 4516
Let_Result_5 := Fld_5; 
--# end
Scope_Result_5 := Let_Result_5;
   --# end
            Match_Result := Scope_Result_5; 
         when Ada_Protected_Body_Stub =>
   --# scope-start
--# scope-start
      Cast_Result_7 := Create_Internal_Entity_Protected_Body_Stub
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_6 := Cast_Result_7; 
--# bind _ Match_6
--# expr-start 4517 '<FieldAccess for protected_previous_part at ast.py:3649>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3649
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.protected_previous_part
Fld_6 := Libadalang.Implementation.Body_Node_P_Protected_Previous_Part (Node => Ent.Node, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 4517
Let_Result_6 := Fld_6; 
--# end
Scope_Result_6 := Let_Result_6;
   --# end
            Match_Result := Scope_Result_6; 
         when Ada_Task_Body =>
   --# scope-start
--# scope-start
      Cast_Result_8 := Create_Internal_Entity_Task_Body
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_7 := Cast_Result_8; 
--# bind _ Match_7
--# expr-start 4518 '<FieldAccess for task_previous_part at ast.py:3650>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3650
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.task_previous_part
Fld_7 := Libadalang.Implementation.Body_Node_P_Task_Previous_Part (Node => Ent.Node, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 4518
Let_Result_7 := Fld_7; 
--# end
Scope_Result_7 := Let_Result_7;
   --# end
            Match_Result := Scope_Result_7; 
         when Ada_Task_Body_Stub =>
   --# scope-start
--# scope-start
      Cast_Result_9 := Create_Internal_Entity_Task_Body_Stub
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_8 := Cast_Result_9; 
--# bind _ Match_8
--# expr-start 4519 '<FieldAccess for task_previous_part at ast.py:3651>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3651
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.task_previous_part
Fld_8 := Libadalang.Implementation.Body_Node_P_Task_Previous_Part (Node => Ent.Node, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 4519
Let_Result_8 := Fld_8; 
--# end
Scope_Result_8 := Let_Result_8;
   --# end
            Match_Result := Scope_Result_8; 
end case;
--# expr-done 4520
Block_Var := Match_Result; 
--# bind pp Block_Var
--# expr-start 4526 '<If at ast.py:3658>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3658
--# expr-start 4521 '<IsA BodyStub at ast.py:3659>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3659
Is_A_1 := Ent.Node /= null 
and then Ent.Node.Kind in Ada_Body_Stub; 
--# expr-done 4521
if Is_A_1 then
   If_Result_1 := Block_Var;
else
   --# expr-start 4525 '<Let at ast.py:3661>' Let_Result_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3661
--# scope-start
Var_Expr := Block_Var;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 4522 '<FieldAccess for next_part_for_decl at ast.py:3661>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3661
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.next_part_for_decl'
Fld_9 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Next_Part_For_Decl (Node => Var_Expr.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr.Info);
--# end
--# expr-done 4522
Scope_Result_9 := Fld_9;
   --# end
   Result_Var := Scope_Result_9;
else
   Result_Var := No_Entity_Basic_Decl;
end if;
Pp_Next_Part := Result_Var; 
--# bind pp_next_part Pp_Next_Part
--# expr-start 4524 '<If at ast.py:3661>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3661
--# expr-start 4523 '<IsA BodyStub at ast.py:3662>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3662
Is_A := Pp_Next_Part.Node /= null 
and then Pp_Next_Part.Node.Kind in Ada_Body_Stub; 
--# expr-done 4523
if Is_A then
   If_Result := Pp_Next_Part;
else
   If_Result := Block_Var;
end if;
--# expr-done 4524
Let_Result_9 := If_Result; 
--# end
--# expr-done 4525
   If_Result_1 := Let_Result_9;
end if;
--# expr-done 4526
Let_Result_10 := If_Result_1; 
--# end
--# expr-done 4527
      Property_Result := Let_Result_10;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Unbound_Previous_Part;
--# end
--# property-start Body.previous_part_internal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3670
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Previous_Part_Internal
  (Node : Bare_Body_Node
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_227 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
As_Entity : Internal_Entity_Body_Node;
Is_Null : Boolean;
If_Result : Internal_Entity_Body_Node;
Fld_1 : Internal_Entity_Basic_Decl;
Fld_2 : Bare_Basic_Decl;
As_Entity_1 : Internal_Entity_Basic_Decl;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Basic_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_227 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Bound_Env);
            end Finalizer_Scope_227;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4534 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4533 '<bind at ast.py:3684>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3684
--# expr-start 4528 '<FieldAccess for node_env at ast.py:3685>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3685
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Node_Env (Node => Self);
--# expr-done 4528
Bound_Env := Fld; Inc_Ref (Bound_Env);
--# expr-start 4532 '<as_entity at ast.py:3686>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3686
--# expr-start 4531 '<FieldAccess for node at ast.py:3686>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3686
--# expr-start 4530 '<FieldAccess for unbound_previous_part at ast.py:3686>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3686
--# expr-start 4529 '<as_bare_entity at ast.py:3686>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3686
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity_Body_Node;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 4529
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.unbound_previous_part
Fld_1 := Libadalang.Implementation.Body_Node_P_Unbound_Previous_Part (Node => If_Result.Node, Env => Bound_Env, Imprecise_Fallback => Imprecise_Fallback, E_Info => If_Result.Info);
--# end
--# expr-done 4530
Fld_2 := Fld_1.Node;
--# expr-done 4531
Is_Null_1 := Fld_2 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Basic_Decl;
else
As_Entity_1 := (Info => E_Info, Node => Fld_2); 
   If_Result_1 := As_Entity_1;
end if;
--# expr-done 4532
Dyn_Var_Bind_Result := If_Result_1; 
--# expr-done 4533
Let_Result := Dyn_Var_Bind_Result; 
--# end
--# expr-done 4534
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_227;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_227;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Previous_Part_Internal;
--# end
--# property-start Body.previous_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3692
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Previous_Part
  (Node : Bare_Body_Node
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Body_Node :=
     Internal_Entity_Body_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Body_Node_Body_Node_P_Previous_Part,
            Items    => new Mmz_Key_Array (1 ..  3))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Body_Node,
                                As_Bare_Body_Node => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Boolean,
                                      As_Boolean => Imprecise_Fallback);
               Mmz_K.Items (3) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Basic_Decl;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 4536 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4535 '<FieldAccess for previous_part_internal at ast.py:3697>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3697
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.previous_part_internal
Fld := Libadalang.Implementation.Body_Node_P_Previous_Part_Internal (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 4535
Let_Result := Fld; 
--# end
--# expr-done 4536
      Property_Result := Let_Result;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Basic_Decl,
                        As_Internal_Entity_Basic_Decl => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Previous_Part;
--# end
--# property-start Body.decl_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3700
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Decl_Part
  (Node : Bare_Body_Node
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Body_Node :=
     Internal_Entity_Body_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Var_Expr : Internal_Entity_Basic_Decl;
Match_0 : Internal_Entity_Body_Stub;
Match_1 : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Basic_Decl;
Match_Prefix : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity_Body_Stub;
Let_Result : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Cast_Result_1 : Internal_Entity_Basic_Decl;
Let_Result_1 : Internal_Entity_Basic_Decl;
Scope_Result_1 : Internal_Entity_Basic_Decl;
Match_Result : Internal_Entity_Basic_Decl;
Scope_Result_2 : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Let_Result_2 : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4540 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 4537 '<FieldAccess for previous_part_internal at ast.py:3704>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3704
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.previous_part_internal
Fld := Libadalang.Implementation.Body_Node_P_Previous_Part_Internal (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 4537
Var_Expr := Fld;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind prev_part Var_Expr
--# expr-start 4539 '<Match at ast.py:3705>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3705
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Var_Expr; 
case Ada_Basic_Decl (Match_Prefix.Node.Kind) is
         when Ada_Package_Body_Stub .. Ada_Task_Body_Stub =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Body_Stub
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind stub Match_0
--# expr-start 4538 '<FieldAccess for previous_part_internal at ast.py:3708>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3708
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.previous_part_internal
Fld_1 := Libadalang.Implementation.Body_Node_P_Previous_Part_Internal (Node => Match_0.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Match_0.Info);
--# end
--# expr-done 4538
Let_Result := Fld_1; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Abstract_State_Decl .. Ada_Subp_Renaming_Decl | Ada_Entry_Body .. Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Basic_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind other Match_1
Let_Result_1 := Match_1; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
end case;
--# expr-done 4539
Scope_Result_2 := Match_Result;
   --# end
   Result_Var := Scope_Result_2;
else
   Result_Var := No_Entity_Basic_Decl;
end if;
Let_Result_2 := Result_Var; 
--# end
--# expr-done 4540
      Property_Result := Let_Result_2;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Decl_Part;
--# end
--# property-start Body.safe_generic_decl_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3714
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Safe_Generic_Decl_Part
  (Node : Bare_Body_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Generic_Decl
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Body_Node :=
     Internal_Entity_Body_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Generic_Decl;
            procedure Finalizer_Scope_230 with Inline_Always;
            procedure Finalizer_Scope_2406 with Inline_Always;
      Var_Expr : Internal_Entity_Basic_Decl;
Var_Expr_1 : Internal_Entity_Array_Access;
Var_Expr_2 : Internal_Entity_Generic_Subp_Decl;
Bound_Imprecise_Fallback : Boolean;
Is_A : Boolean;
Fld : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity;
Cast_Result : Internal_Entity_Generic_Package_Decl;
Scope_Result : Internal_Entity_Generic_Package_Decl;
Result_Var : Internal_Entity_Generic_Package_Decl;
Is_A_1 : Boolean;
Fld_2 : Lexical_Env := Empty_Env;
Env_Parent : Lexical_Env := Empty_Env;
Fld_3 : Symbol_Type;
Env_Get_Result : Internal_Entity_Array_Access;
Get_Result : Internal_Entity;
Cast_Result_1 : Internal_Entity_Generic_Subp_Decl;
Scope_Result_1 : Internal_Entity_Generic_Subp_Decl;
Get_Result_1 : Internal_Entity;
Cast_Result_2 : Internal_Entity_Generic_Subp_Decl;
Result_Var_1 : Internal_Entity_Generic_Subp_Decl;
Scope_Result_2 : Internal_Entity_Generic_Subp_Decl;
Result_Var_2 : Internal_Entity_Generic_Subp_Decl;
Cast_Result_3 : Internal_Entity_Generic_Decl;
If_Result : Internal_Entity_Generic_Decl;
Cast_Result_4 : Internal_Entity_Generic_Decl;
If_Result_1 : Internal_Entity_Generic_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Generic_Decl;
Let_Result : Internal_Entity_Generic_Decl;
            procedure Finalizer_Scope_230 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Env_Parent);
                     Dec_Ref (Env_Get_Result);
            end Finalizer_Scope_230;
            procedure Finalizer_Scope_2406 is
            begin
                     Dec_Ref (Var_Expr_1);
            end Finalizer_Scope_2406;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4556 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4555 '<bind at ast.py:3719>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3719
Bound_Imprecise_Fallback := False; 
--# expr-start 4554 '<Cond at ast.py:3721>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3721
--# expr-start 4541 '<IsA PackageBody at ast.py:3722>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3722
Is_A := Self /= null 
and then Self.Kind in Ada_Package_Body_Range; 
--# expr-done 4541
if Is_A then
--# expr-start 4542 '<FieldAccess for decl_part at ast.py:3723>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3723
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.decl_part
Fld := Libadalang.Implementation.Body_Node_P_Decl_Part (Node => Ent.Node, Imprecise_Fallback => Bound_Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 4542
Var_Expr := Fld;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind d Var_Expr
--# expr-start 4544 '<Cast to GenericPackageDecl at ast.py:3724>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3724
--# expr-start 4543 '<parent at ast.py:3724>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3724
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Parent (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# expr-done 4543
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Generic_Package_Decl_Range
   then
      Cast_Result := Create_Internal_Entity_Generic_Package_Decl
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         Cast_Result := No_Entity_Generic_Package_Decl;
   end if;
--# expr-done 4544
Scope_Result := Cast_Result;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Generic_Package_Decl;
end if;
      Cast_Result_4 := Create_Internal_Entity_Generic_Decl
        (Node => Result_Var.Node,
         Info => Result_Var.Info);
   If_Result_1 := Cast_Result_4;
else
--# expr-start 4545 '<IsA BaseSubpBody, SubpBodyStub at ast.py:3727>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3727
Is_A_1 := Self /= null 
and then Self.Kind in Ada_Base_Subp_Body | Ada_Subp_Body_Stub_Range; 
--# expr-done 4545
if Is_A_1 then
--# expr-start 4549 '<EnvGet at ast.py:3733>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3733
--# expr-start 4547 '<env_parent at ast.py:3733>' Env_Parent /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3733
--# expr-start 4546 '<FieldAccess for children_env at ast.py:3733>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3733
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Children_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 4546
Env_Parent := AST_Envs.Parent (Fld_2); 
--# expr-done 4547
--# expr-start 4548 '<FieldAccess for name_symbol at ast.py:3734>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3734
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld_3 := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 4548
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Env_Parent, Key => Fld_3, Lookup_Kind => To_Lookup_Kind_Type (Recursive), Categories => (Inherited_Primitives => False, Nocat => True))); 
--# expr-done 4549
Var_Expr_1 := Env_Get_Result;
      Inc_Ref (Var_Expr_1);
if Var_Expr_1 /= No_Internal_Entity_Array_Type then
   --# scope-start
   --# bind results Var_Expr_1
--# expr-start 4551 '<Cast to GenericSubpDecl at ast.py:3737>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3737
--# expr-start 4550 '<at at ast.py:3737>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3737
Get_Result := Get (Var_Expr_1, 1, True); 
--# expr-done 4550
   if Get_Result.Node = null
      or else Get_Result.Node.Kind in Ada_Generic_Subp_Decl_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Generic_Subp_Decl
        (Node => Get_Result.Node,
         Info => Get_Result.Info);
   else
         Cast_Result_1 := No_Entity_Generic_Subp_Decl;
   end if;
--# expr-done 4551
Var_Expr_2 := Cast_Result_1;
if Var_Expr_2 /= No_Entity_Generic_Subp_Decl then
   --# scope-start
   --# bind e Var_Expr_2
Scope_Result_1 := Var_Expr_2;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   --# expr-start 4553 '<Cast to GenericSubpDecl at ast.py:3738>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3738
--# expr-start 4552 '<at at ast.py:3738>' Get_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3738
Get_Result_1 := Get (Var_Expr_1, 2, True); 
--# expr-done 4552
   if Get_Result_1.Node = null
      or else Get_Result_1.Node.Kind in Ada_Generic_Subp_Decl_Range
   then
      Cast_Result_2 := Create_Internal_Entity_Generic_Subp_Decl
        (Node => Get_Result_1.Node,
         Info => Get_Result_1.Info);
   else
         Cast_Result_2 := No_Entity_Generic_Subp_Decl;
   end if;
--# expr-done 4553
   Result_Var_1 := Cast_Result_2;
end if;
Scope_Result_2 := Result_Var_1;
   --# end
      Finalizer_Scope_2406;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Entity_Generic_Subp_Decl;
end if;
      Cast_Result_3 := Create_Internal_Entity_Generic_Decl
        (Node => Result_Var_2.Node,
         Info => Result_Var_2.Info);
   If_Result := Cast_Result_3;
else
   If_Result := No_Entity_Generic_Decl;
end if;
   If_Result_1 := If_Result;
end if;
--# expr-done 4554
Dyn_Var_Bind_Result := If_Result_1; 
--# expr-done 4555
Let_Result := Dyn_Var_Bind_Result; 
--# end
--# expr-done 4556
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_230;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_230;
               Finalizer_Scope_2406;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Safe_Generic_Decl_Part;
--# end
--# property-start Body.next_part_for_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3747
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Next_Part_For_Decl
  (Node : Bare_Body_Node
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4557 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := No_Entity_Basic_Decl; 
--# end
--# expr-done 4557
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Next_Part_For_Decl;
--# end
--# property-start Body.is_subunit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3755
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Is_Subunit
  (Node : Bare_Body_Node
  )
   return Boolean
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4560 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4559 '<IsA Subunit at ast.py:3756>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3756
--# expr-start 4558 '<parent at ast.py:3756>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3756
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 4558
Is_A := Fld /= null 
and then Fld.Kind in Ada_Subunit_Range; 
--# expr-done 4559
Let_Result := Is_A; 
--# end
--# expr-done 4560
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Is_Subunit;
--# end
--# property-start Body.subunit_root /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3759
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Subunit_Root
  (Node : Bare_Body_Node
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Basic_Decl;
      Var_Expr : Bare_Subunit;
Fld : Bare_Ada_Node;
Cast_Result : Bare_Subunit;
Fld_1 : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4564 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4562 '<Cast to Subunit at ast.py:3763>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3763
--# expr-start 4561 '<parent at ast.py:3763>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3763
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 4561
   if Fld = null
      or else Fld.Kind in Ada_Subunit_Range
   then
      Cast_Result := Fld;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 4562
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind su Var_Expr
--# expr-start 4563 '<FieldAccess for body_root at ast.py:3763>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3763
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Subunit.body_root
Fld_1 := Libadalang.Implementation.Subunit_P_Body_Root (Node => Var_Expr);
--# end
--# expr-done 4563
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Basic_Decl;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 4564
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Subunit_Root;
--# end
--# property-start Body.body_scope /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3766
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Body_Scope
  (Node : Bare_Body_Node
      ; Follow_Private : Boolean
      ; Force_Decl : Boolean
         := False
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Body_Node :=
     Internal_Entity_Body_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind follow_private Follow_Private
   --# bind force_decl Force_Decl
   --# bind env Env
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_234 with Inline_Always;
            procedure Finalizer_Scope_2360 with Inline_Always;
            procedure Finalizer_Scope_2361 with Inline_Always;
      Block_Var : Lexical_Env := Empty_Env;
Block_Var_1 : Lexical_Env := Empty_Env;
Dns : Lexical_Env := Empty_Env;
Var_Expr : Bare_Generic_Package_Decl;
Var_Expr_1 : Bare_Ada_Node;
Var_Expr_2 : Internal_Entity;
Fld : Internal_Entity_Basic_Decl;
Is_Null : Boolean;
Not_Val : Boolean;
Fld_1 : Internal_Entity_Basic_Decl;
Fld_2 : Lexical_Env := Empty_Env;
Fld_3 : Boolean;
Fld_4 : Boolean;
If_Result : Boolean;
Fld_5 : Internal_Entity_Defining_Name;
Fld_6 : Lexical_Env := Empty_Env;
Fld_7 : Internal_Entity_Defining_Name;
Fld_8 : Lexical_Env := Empty_Env;
Is_Equal : Boolean;
Env_Node : Bare_Ada_Node;
Cast_Expr : Bare_Body_Node;
Cast_Result : Bare_Ada_Node;
Is_Equal_1 : Boolean;
If_Result_1 : Boolean;
If_Result_2 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
Fld_9 : Boolean;
If_Result_3 : Boolean;
Fld_10 : Internal_Entity_Defining_Name;
Fld_11 : Lexical_Env := Empty_Env;
Fld_12 : Bare_Ada_Node;
Fld_13 : Lexical_Env := Empty_Env;
If_Result_4 : Lexical_Env := Empty_Env;
If_Result_5 : Lexical_Env := Empty_Env;
If_Result_6 : Lexical_Env := Empty_Env;
Env_Node_1 : Bare_Ada_Node;
Cast_Result_1 : Bare_Generic_Package_Decl;
Fld_14 : Bare_Generic_Package_Internal;
Fld_15 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Sym : Symbol_Type;
Env_Get_Result : Internal_Entity_Array_Access;
Get_Result : Internal_Entity;
Fld_16 : Lexical_Env := Empty_Env;
Scope_Result_1 : Lexical_Env := Empty_Env;
Result_Var_1 : Lexical_Env := Empty_Env;
Env_Node_2 : Bare_Ada_Node;
Is_A : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
If_Result_7 : Boolean;
If_Result_8 : Lexical_Env := Empty_Env;
Let_Result_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_234 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Fld_6);
                     Dec_Ref (Dns);
                     Dec_Ref (Fld_8);
                     Dec_Ref (If_Result_2);
                     Dec_Ref (Let_Result);
                     Dec_Ref (Fld_11);
                     Dec_Ref (Fld_13);
                     Dec_Ref (If_Result_4);
                     Dec_Ref (If_Result_5);
                     Dec_Ref (If_Result_6);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Env_Get_Result);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (If_Result_8);
                     Dec_Ref (Let_Result_1);
            end Finalizer_Scope_234;
            procedure Finalizer_Scope_2360 is
            begin
                     Dec_Ref (Fld_15);
            end Finalizer_Scope_2360;
            procedure Finalizer_Scope_2361 is
            begin
                     Dec_Ref (Fld_16);
            end Finalizer_Scope_2361;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4598 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 4586 '<Cond at ast.py:3775>' If_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3775
--# expr-start 4567 '<Not at ast.py:3780>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3780
--# expr-start 4566 '<IsNull at ast.py:3780>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3780
--# expr-start 4565 '<FieldAccess for subunit_root at ast.py:3780>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3780
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.subunit_root
Fld := Libadalang.Implementation.Body_Node_P_Subunit_Root (Node => Self);
--# end
--# expr-done 4565
Is_Null := Fld.Node = null; 
--# expr-done 4566
Not_Val := not (Is_Null); 
--# expr-done 4567
if Not_Val then
   --# expr-start 4569 '<FieldAccess for children_env at ast.py:3780>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3780
--# expr-start 4568 '<FieldAccess for subunit_root at ast.py:3780>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3780
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.subunit_root
Fld_1 := Libadalang.Implementation.Body_Node_P_Subunit_Root (Node => Self);
--# end
--# expr-done 4568
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Children_Env (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# expr-done 4569
   If_Result_6 := Fld_2;
else
--# expr-start 4570 '<FieldAccess for is_subprogram at ast.py:3785>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3785
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_subprogram
Fld_3 := Libadalang.Implementation.Basic_Decl_P_Is_Subprogram (Node => Self);
--# end
--# expr-done 4570
if Fld_3 then
   --# expr-start 4571 '<FieldAccess for is_compilation_unit_root at ast.py:3785>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3785
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_compilation_unit_root
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Is_Compilation_Unit_Root (Node => Self);
--# end
--# expr-done 4571
   If_Result := Fld_4;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 4580 '<Let at ast.py:3786>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3786
--# scope-start
--# expr-start 4573 '<FieldAccess for scope at ast.py:3786>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3786
--# expr-start 4572 '<FieldAccess for defining_name at ast.py:3786>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3786
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_5 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4572
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.scope'
Fld_6 := Libadalang.Implementation.Dispatcher_Name_P_Scope (Node => Fld_5.Node, Env => Env);
--# end
--# expr-done 4573
Dns := Fld_6; Inc_Ref (Dns);
--# bind dns Dns
--# expr-start 4579 '<If at ast.py:3789>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3789
--# expr-start 4574 '<IsNull at ast.py:3789>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3789
Is_Equal := Equivalent (Dns, Empty_Env); 
--# expr-done 4574
if Is_Equal then
   If_Result_1 := True;
else
   --# expr-start 4576 '<Eq at ast.py:3789>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3789
--# expr-start 4575 '<env_node at ast.py:3789>' Env_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3789
Env_Node := AST_Envs.Env_Node (Dns); 
--# expr-done 4575
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Is_Equal_1 := Env_Node = Cast_Result; 
--# expr-done 4576
   If_Result_1 := Is_Equal_1;
end if;
if If_Result_1 then
   --# expr-start 4578 '<FieldAccess for parent_scope at ast.py:3790>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3790
--# expr-start 4577 '<FieldAccess for defining_name at ast.py:3790>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3790
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_7 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4577
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.parent_scope'
Fld_8 := Libadalang.Implementation.Dispatcher_Name_P_Parent_Scope (Node => Fld_7.Node, Env => Env);
--# end
--# expr-done 4578
   If_Result_2 := Fld_8;
else
   If_Result_2 := Dns;
end if;
      Inc_Ref (If_Result_2);
--# expr-done 4579
Let_Result := If_Result_2; Inc_Ref (Let_Result);
--# end
--# expr-done 4580
   If_Result_5 := Let_Result;
else
--# expr-start 4581 '<FieldAccess for is_compilation_unit_root at ast.py:3795>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3795
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_compilation_unit_root
Fld_9 := Libadalang.Implementation.Basic_Decl_P_Is_Compilation_Unit_Root (Node => Self);
--# end
--# expr-done 4581
if Fld_9 then
   If_Result_3 := True;
else
   If_Result_3 := Force_Decl;
end if;
if If_Result_3 then
   --# expr-start 4583 '<FieldAccess for scope at ast.py:3796>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3796
--# expr-start 4582 '<FieldAccess for defining_name at ast.py:3796>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3796
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_10 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4582
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.scope'
Fld_11 := Libadalang.Implementation.Dispatcher_Name_P_Scope (Node => Fld_10.Node, Env => Env);
--# end
--# expr-done 4583
   If_Result_4 := Fld_11;
else
   --# expr-start 4585 '<FieldAccess for children_env at ast.py:3800>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3800
--# expr-start 4584 '<parent at ast.py:3800>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3800
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Self.Parent; 
--# expr-done 4584
      if Fld_12 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Libadalang.Implementation.Children_Env (Node => Fld_12);
--# expr-done 4585
   If_Result_4 := Fld_13;
end if;
      Inc_Ref (If_Result_4);
   If_Result_5 := If_Result_4;
end if;
      Inc_Ref (If_Result_5);
   If_Result_6 := If_Result_5;
end if;
      Inc_Ref (If_Result_6);
--# expr-done 4586
Block_Var := If_Result_6; Inc_Ref (Block_Var);
--# bind scope Block_Var
--# expr-start 4588 '<Cast to GenericPackageDecl at ast.py:3805>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3805
--# expr-start 4587 '<env_node at ast.py:3805>' Env_Node_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3805
Env_Node_1 := AST_Envs.Env_Node (Block_Var); 
--# expr-done 4587
   if Env_Node_1 = null
      or else Env_Node_1.Kind in Ada_Generic_Package_Decl_Range
   then
      Cast_Result_1 := Env_Node_1;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 4588
Var_Expr := Cast_Result_1;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind gen_pkg_decl Var_Expr
--# expr-start 4590 '<FieldAccess for children_env at ast.py:3806>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3806
--# expr-start 4589 '<FieldAccess for package_decl at ast.py:3806>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3806
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Var_Expr.Generic_Package_Decl_F_Package_Decl;
--# expr-done 4589
      if Fld_14 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Libadalang.Implementation.Children_Env (Node => Fld_14);
--# expr-done 4590
Scope_Result := Fld_15;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2360;
   Result_Var := Scope_Result;
else
   Result_Var := Block_Var;
end if;
      Inc_Ref (Result_Var);
Block_Var_1 := Result_Var; Inc_Ref (Block_Var_1);
--# bind public_scope Block_Var_1
--# expr-start 4597 '<If at ast.py:3812>' If_Result_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3812
if Follow_Private then
--# expr-start 4591 '<env_node at ast.py:3815>' Env_Node_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3815
Env_Node_2 := AST_Envs.Env_Node (Block_Var_1); 
--# expr-done 4591
Var_Expr_1 := Env_Node_2;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 4592 '<IsA BasePackageDecl, SingleProtectedDecl, ProtectedTypeDecl at ast.py:3815>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3815
Is_A := Var_Expr_1 /= null 
and then Var_Expr_1.Kind in Ada_Base_Package_Decl | Ada_Single_Protected_Decl_Range | Ada_Protected_Type_Decl_Range; 
--# expr-done 4592
Scope_Result_2 := Is_A;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := False;
end if;
   If_Result_7 := Result_Var_2;
else
   If_Result_7 := False;
end if;
if If_Result_7 then
--# expr-start 4595 '<at at ast.py:3820>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3820
--# expr-start 4594 '<EnvGet at ast.py:3820>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3820
--# expr-start 4593 '<Symbol __privatepart at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Privatepart); 
--# expr-done 4593
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Block_Var_1, Key => Sym, Lookup_Kind => To_Lookup_Kind_Type (Flat), Categories => (Inherited_Primitives => False, Nocat => True))); 
--# expr-done 4594
Get_Result := Get (Env_Get_Result, 0, True); 
--# expr-done 4595
Var_Expr_2 := Get_Result;
if Var_Expr_2 /= No_Entity then
   --# scope-start
   --# bind pp Var_Expr_2
--# expr-start 4596 '<FieldAccess for children_env at ast.py:3823>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3823
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_16 := Libadalang.Implementation.Children_Env (Node => Var_Expr_2.Node, E_Info => Var_Expr_2.Info);
--# expr-done 4596
Scope_Result_1 := Fld_16;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2361;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := Block_Var_1;
end if;
      Inc_Ref (Result_Var_1);
   If_Result_8 := Result_Var_1;
else
   If_Result_8 := Block_Var_1;
end if;
      Inc_Ref (If_Result_8);
--# expr-done 4597
Let_Result_1 := If_Result_8; Inc_Ref (Let_Result_1);
--# end
--# expr-done 4598
      Property_Result := Let_Result_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_234;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_234;
               Finalizer_Scope_2360;
               Finalizer_Scope_2361;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Body_Scope;
--# end
--# property-start Body.previous_part_env_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3427
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Previous_Part_Env_Name
  (Node : Bare_Body_Node
  )
   return Symbol_Type
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type;
            procedure Finalizer_Scope_219 with Inline_Always;
      Fld : String_Type;
Str : String_Type;
Concat_Result : String_Type;
Sym : Symbol_Type;
Fld_1 : String_Type;
Sym_1 : Symbol_Type;
Fld_2 : Boolean;
If_Result : Symbol_Type;
Let_Result : Symbol_Type;
            procedure Finalizer_Scope_219 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Str);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_219;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4607 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4606 '<If at ast.py:3434>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3434
--# expr-start 4599 '<FieldAccess for is_subunit at ast.py:3435>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3435
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.is_subunit
Fld_2 := Libadalang.Implementation.Body_Node_P_Is_Subunit (Node => Self);
--# end
--# expr-done 4599
if Fld_2 then
   --# expr-start 4603 '<to_symbol at ast.py:3436>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3436
--# expr-start 4602 '<Concat at ast.py:3436>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3436
--# expr-start 4600 '<FieldAccess for top_level_env_name at ast.py:3436>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3436
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.top_level_env_name
Fld := Libadalang.Implementation.Basic_Decl_P_Top_Level_Env_Name (Node => Self);
--# end
--# expr-done 4600
--# expr-start 4601 '<String at ast.py:3436>' Str /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3436
Str := Create_String ("__stub"); 
--# expr-done 4601
Concat_Result := Concat_String (Fld, Str); 
--# expr-done 4602
Sym := String_To_Symbol (Self.Unit.Context, Concat_Result); 
--# expr-done 4603
   If_Result := Sym;
else
   --# expr-start 4605 '<to_symbol at ast.py:3437>' Sym_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3437
--# expr-start 4604 '<FieldAccess for top_level_env_name at ast.py:3437>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3437
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.top_level_env_name
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Top_Level_Env_Name (Node => Self);
--# end
--# expr-done 4604
Sym_1 := String_To_Symbol (Self.Unit.Context, Fld_1); 
--# expr-done 4605
   If_Result := Sym_1;
end if;
--# expr-done 4606
Let_Result := If_Result; 
--# end
--# expr-done 4607
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_219;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_219;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Previous_Part_Env_Name;
--# end
--# property-start Body.previous_part_link_env_assoc /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3441
pragma Warnings (Off, "is not referenced");
 function Body_Node_P_Previous_Part_Link_Env_Assoc
  (Node : Bare_Body_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Body_Node := Bare_Body_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Body_Node :=
     Internal_Entity_Body_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_220 with Inline_Always;
            procedure Finalizer_Scope_2370 with Inline_Always;
            procedure Finalizer_Scope_2371 with Inline_Always;
      Var_Expr : Symbol_Type;
Var_Expr_1 : Lexical_Env := Empty_Env;
Sym : Symbol_Type;
Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Scope_Result : Internal_Designated_Env := No_Designated_Env;
Fld_1 : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_2 : Lexical_Env := Empty_Env;
Dyn_Var_Bind_Result : Lexical_Env := Empty_Env;
New_Struct_1 : Internal_Designated_Env := No_Designated_Env;
Scope_Result_1 : Internal_Designated_Env := No_Designated_Env;
New_Struct_2 : Internal_Designated_Env := No_Designated_Env;
Result_Var : Internal_Designated_Env := No_Designated_Env;
Result_Var_1 : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Body_Node;
Cast_Result : Bare_Ada_Node;
New_Struct_3 : Internal_Env_Assoc := No_Env_Assoc;
Let_Result : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_220 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (New_Struct_2);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (New_Struct_3);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_220;
            procedure Finalizer_Scope_2370 is
            begin
                     Dec_Ref (New_Struct);
            end Finalizer_Scope_2370;
            procedure Finalizer_Scope_2371 is
            begin
                     Dec_Ref (Var_Expr_1);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_2371;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4617 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4616 '<New EnvAssoc at ast.py:3451>' New_Struct_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3451
--# expr-start 4608 '<FieldAccess for previous_part_env_name at ast.py:3454>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3454
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Body.previous_part_env_name'
Fld := Libadalang.Implementation.Dispatcher_Body_Node_P_Previous_Part_Env_Name (Node => Self);
--# end
--# expr-done 4608
Var_Expr := Fld;
if Var_Expr /= null then
   --# scope-start
   --# bind name Var_Expr
--# expr-start 4609 '<New DesignatedEnv at ast.py:3455>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3455
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => Var_Expr, Kind => Named_Env); 
--# expr-done 4609
Scope_Result := New_Struct;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2370;
   Result_Var_1 := Scope_Result;
else
--# expr-start 4612 '<bind at ast.py:3457>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3457
--# expr-start 4610 '<FieldAccess for default_initial_env at ast.py:3458>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3458
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.default_initial_env
Fld_1 := Libadalang.Implementation.Ada_Node_P_Default_Initial_Env (Node => Self);
--# end
--# expr-done 4610
Bound_Env := Fld_1; Inc_Ref (Bound_Env);
--# expr-start 4611 '<FieldAccess for body_scope at ast.py:3459>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3459
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.body_scope
Fld_2 := Libadalang.Implementation.Body_Node_P_Body_Scope (Node => Ent.Node, Follow_Private => False, Force_Decl => True, Env => Bound_Env, E_Info => Ent.Info);
--# end
--# expr-done 4611
Dyn_Var_Bind_Result := Fld_2; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 4612
Var_Expr_1 := Dyn_Var_Bind_Result;
      Inc_Ref (Var_Expr_1);
if Var_Expr_1 /= Empty_Env then
   --# scope-start
   --# bind non_null_env Var_Expr_1
--# expr-start 4613 '<New DesignatedEnv at ???>' New_Struct_1 None
Inc_Ref (Var_Expr_1);
New_Struct_1 := (Direct_Env => Var_Expr_1, Env_Name => null, Kind => Direct_Env); 
--# expr-done 4613
Scope_Result_1 := New_Struct_1;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2371;
   Result_Var := Scope_Result_1;
else
   --# expr-start 4614 '<New DesignatedEnv at ast.py:3456>' New_Struct_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3456
Inc_Ref (Empty_Env);
New_Struct_2 := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 4614
   Result_Var := New_Struct_2;
end if;
      Inc_Ref (Result_Var);
   Result_Var_1 := Result_Var;
end if;
      Inc_Ref (Result_Var_1);
--# expr-start 4615 '<Symbol __nextpart at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Nextpart); 
--# expr-done 4615
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (Result_Var_1);
New_Struct_3 := (Dest_Env => Result_Var_1, Key => Sym, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 4616
Let_Result := New_Struct_3; Inc_Ref (Let_Result);
--# end
--# expr-done 4617
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_220;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_220;
               Finalizer_Scope_2370;
               Finalizer_Scope_2371;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Node_P_Previous_Part_Link_Env_Assoc;
--# end
function Body_Node_P_Subunit_Decl_Env
  (E : Entity
  ) return Lexical_Env
is
      E_Info : Internal_Entity_Info :=
         Shed_Rebindings (E.Info, Node_Env (E.Node));
   Result : constant Lexical_Env :=
      Body_Node_P_Subunit_Decl_Env (E.Node, E_Info);
begin
      return Result;
end;
function Body_Node_P_Body_Decl_Scope
  (E : Entity
  ) return Lexical_Env
is
      E_Info : Internal_Entity_Info :=
         Shed_Rebindings (E.Info, Node_Env (E.Node));
   Result : constant Lexical_Env :=
      Body_Node_P_Body_Decl_Scope (E.Node, E_Info);
begin
      return Result;
end;
   --
   --  Primitives for Bare_Base_Subp_Body
   --
      procedure Base_Subp_Body_Pre_Env_Actions
        (Self            : Bare_Base_Subp_Body;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Dummy : Boolean;
      begin
         Dummy := Libadalang.Implementation.Env_Do_134 (Node => 
Self);
      end;
      declare
         Env : constant Internal_Designated_Env :=
           Libadalang.Implementation.Initial_Env_135 (Node => 
Self);
      begin
         Set_Initial_Env
           (Self,
            State,
            Env,
            DSL_Location => "ast.py:18809");
      end;
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mappings : Internal_Env_Assoc_Array_Access :=
            Libadalang.Implementation.Env_Mappings_136 (Node => 
Self);
      begin
         for Mapping of Mappings.Items loop
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:18816");
         end loop;
         Dec_Ref (Mappings);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_137 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_138 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:18824");
         Dec_Ref (Mapping.Dest_Env);
      end;
      declare
         Dummy : Bare_Compilation_Unit_Array_Access;
      begin
         Dummy := Libadalang.Implementation.Env_Do_139 (Node => 
Self);
         Dec_Ref (Dummy);
      end;
      if Libadalang.Implementation.Ref_Cond_141 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_140 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Use_Package_Name_Designated_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_143 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_142 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Name_Designated_Type_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_145 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_144 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Ada_Node_P_Use_Clauses_In_Spec_Of_Subp_Body'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_147 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_146 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Ada_Node_P_Nested_Generic_Formal_Part'Access,
               Prioritary,
               All_Cats,
               True);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_149 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_148 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Ada_Node_P_Use_Clauses_In_Generic_Formal_Part'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      end;
      procedure Initialize_Fields_For_Base_Subp_Body
        (Self : Bare_Base_Subp_Body
         ; Base_Subp_Body_F_Overriding : Bare_Overriding_Node
         ; Base_Subp_Body_F_Subp_Spec : Bare_Subp_Spec
        ) is
      begin
            Self.Base_Subp_Body_F_Overriding := Base_Subp_Body_F_Overriding;
            Self.Base_Subp_Body_F_Subp_Spec := Base_Subp_Body_F_Subp_Spec;
      end Initialize_Fields_For_Base_Subp_Body;
   function Base_Subp_Body_F_Overriding
     (Node : Bare_Base_Subp_Body) return Bare_Overriding_Node
   is
   begin
         return Node.Base_Subp_Body_F_Overriding;
   end;
   function Base_Subp_Body_F_Subp_Spec
     (Node : Bare_Base_Subp_Body) return Bare_Subp_Spec
   is
   begin
         return Node.Base_Subp_Body_F_Subp_Spec;
   end;
--# property-start BaseSubpBody.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18754
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Body_P_Defining_Names
  (Node : Bare_Base_Subp_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Body :=
     Internal_Entity_Base_Subp_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1321 with Inline_Always;
      Fld : Internal_Entity_Subp_Spec;
Fld_1 : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1321 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1321;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4619 '<FieldAccess for name at ast.py:18754>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18754
--# expr-start 4618 '<FieldAccess for subp_spec at ast.py:18754>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18754
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Subp_Spec (Node => Ent.Node.Base_Subp_Body_F_Subp_Spec, Info => Ent.Info);
--# expr-done 4618
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.name'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Name (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 4619
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld_1;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1321;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1321;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Body_P_Defining_Names;
--# end
--# property-start BaseSubpBody.is_constant_object /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18756
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Body_P_Is_Constant_Object
  (Node : Bare_Base_Subp_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Body_P_Is_Constant_Object;
--# end
--# property-start BaseSubpBody.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18759
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Body_P_Defining_Env
  (Node : Bare_Base_Subp_Body
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Body :=
     Internal_Entity_Base_Subp_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1323 with Inline_Always;
      Var_Expr : Internal_Entity_Base_Subp_Spec;
Fld : Lexical_Env := Empty_Env;
Fld_1 : Internal_Entity_Subp_Spec;
Fld_2 : Lexical_Env := Empty_Env;
Array_Lit : Lexical_Env_Array_Access;
Group_Env : Lexical_Env := Empty_Env;
Fld_3 : Lexical_Env := Empty_Env;
Fld_4 : Internal_Entity_Base_Subp_Spec;
Fld_5 : Internal_Entity_Info;
Fld_6 : Internal_Metadata;
Fld_7 : Boolean;
Fld_8 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result : Lexical_Env := Empty_Env;
Fld_9 : Internal_Entity_Subp_Spec;
Fld_10 : Lexical_Env := Empty_Env;
Fld_11 : Boolean;
If_Result_1 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1323 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Group_Env);
                     Dec_Ref (Fld_3);
                     Dec_Ref (If_Result);
                     Dec_Ref (Fld_10);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1323;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4636 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4635 '<If at ast.py:18760>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18760
--# expr-start 4620 '<FieldAccess for in_scope at ast.py:18761>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18761
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.in_scope
Fld_11 := Libadalang.Implementation.Body_Node_P_In_Scope (Node => Ent.Node, Origin => Origin);
--# end
--# expr-done 4620
if Fld_11 then
   --# expr-start 4632 '<If at ast.py:18763>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18763
--# expr-start 4621 '<FieldAccess for subp_spec_or_null at ast.py:18764>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18764
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4621
Var_Expr := Fld_4;
if Var_Expr /= No_Entity_Base_Subp_Spec then
   --# scope-start
--# expr-start 4625 '<FieldAccess for paramless at ast.py:18764>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18764
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4624 '<FieldAccess for dottable_subp at ast.py:18765>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18765
--# expr-start 4623 '<FieldAccess for md at ast.py:18765>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18765
--# expr-start 4622 '<FieldAccess for info at ast.py:18765>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18765
Fld_5 := Ent.Info;
--# expr-done 4622
Fld_6 := Fld_5.Md;
--# expr-done 4623
Fld_7 := Fld_6.Dottable_Subp;
--# expr-done 4624
--# property-call-start BaseFormalParamHolder.paramless
Fld_8 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Paramless (Node => Var_Expr.Node, Dottable_Subp => Fld_7, Can_Be => True);
--# end
--# expr-done 4625
Scope_Result := Fld_8;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
if Result_Var then
   --# expr-start 4630 '<env_group at ast.py:18766>' Group_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18766
--# expr-start 4629 '<ArrayLiteral at ast.py:18766>' Array_Lit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18766
--# expr-start 4626 '<FieldAccess for children_env at ast.py:18767>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18767
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 4626
--# expr-start 4628 '<FieldAccess for defining_env at ast.py:18767>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18767
--# expr-start 4627 '<FieldAccess for subp_spec at ast.py:18767>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18767
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Subp_Spec (Node => Ent.Node.Base_Subp_Body_F_Subp_Spec, Info => Ent.Info);
--# expr-done 4627
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.defining_env
Fld_2 := Libadalang.Implementation.Base_Subp_Spec_P_Defining_Env (Node => Fld_1.Node, Origin => Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 4628
Array_Lit := Create_Lexical_Env_Array (Internal_Lexical_Env_Array'(1 => Fld, 2 => Fld_2)); 
--# expr-done 4629
Group_Env := Group (Array_Lit, No_Metadata); 
--# expr-done 4630
   If_Result := Group_Env;
else
   --# expr-start 4631 '<FieldAccess for children_env at ast.py:18769>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18769
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Children_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 4631
   If_Result := Fld_3;
end if;
      Inc_Ref (If_Result);
--# expr-done 4632
   If_Result_1 := If_Result;
else
   --# expr-start 4634 '<FieldAccess for defining_env at ast.py:18772>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18772
--# expr-start 4633 '<FieldAccess for subp_spec at ast.py:18772>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18772
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Subp_Spec (Node => Ent.Node.Base_Subp_Body_F_Subp_Spec, Info => Ent.Info);
--# expr-done 4633
      if Fld_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.defining_env
Fld_10 := Libadalang.Implementation.Base_Subp_Spec_P_Defining_Env (Node => Fld_9.Node, Origin => Origin, E_Info => Fld_9.Info);
--# end
--# expr-done 4634
   If_Result_1 := Fld_10;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 4635
Let_Result := If_Result_1; Inc_Ref (Let_Result);
--# end
--# expr-done 4636
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1323;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1323;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Body_P_Defining_Env;
--# end
--# property-start BaseSubpBody.type_expression /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18775
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Body_P_Type_Expression
  (Node : Bare_Base_Subp_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Expr
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Body :=
     Internal_Entity_Base_Subp_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Type_Expr;
      Fld : Internal_Entity_Subp_Spec;
Fld_1 : Internal_Entity_Type_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4638 '<FieldAccess for returns at ast.py:18775>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18775
--# expr-start 4637 '<FieldAccess for subp_spec at ast.py:18775>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18775
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Subp_Spec (Node => Ent.Node.Base_Subp_Body_F_Subp_Spec, Info => Ent.Info);
--# expr-done 4637
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.returns'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Returns (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 4638
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Body_P_Type_Expression;
--# end
--# property-start BaseSubpBody.expr_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18778
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Body_P_Expr_Type
  (Node : Bare_Base_Subp_Body
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Body :=
     Internal_Entity_Base_Subp_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Base_Subp_Spec;
Fld : Internal_Entity_Base_Subp_Spec;
Fld_1 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4641 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4639 '<FieldAccess for subp_spec_or_null at ast.py:18779>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18779
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4639
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Subp_Spec then
   --# scope-start
--# expr-start 4640 '<FieldAccess for return_type at ast.py:18779>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18779
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.return_type
Fld_1 := Libadalang.Implementation.Base_Subp_Spec_P_Return_Type (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 4640
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 4641
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Body_P_Expr_Type;
--# end
--# property-start BaseSubpBody.initial_env_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18782
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Body_P_Initial_Env_Name
  (Node : Bare_Base_Subp_Body
      ; Follow_Private : Boolean
         := False
  )
   return Symbol_Type
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind follow_private Follow_Private
   Property_Result : Symbol_Type;
      Fld : Symbol_Type;
Fld_1 : Symbol_Type;
Fld_2 : Boolean;
If_Result : Symbol_Type;
Let_Result : Symbol_Type;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4646 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4645 '<If at ast.py:18783>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18783
--# expr-start 4642 '<FieldAccess for is_library_item at ast.py:18784>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18784
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_library_item'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Library_Item (Node => Self);
--# end
--# expr-done 4642
if Fld_2 then
   --# expr-start 4643 '<FieldAccess for child_decl_initial_env_name at ast.py:18785>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18785
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.child_decl_initial_env_name'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Child_Decl_Initial_Env_Name (Node => Self, Private_Part => Follow_Private);
--# end
--# expr-done 4643
   If_Result := Fld;
else
   --# expr-start 4644 '<FieldAccess for body_initial_env_name at ast.py:18786>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18786
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.body_initial_env_name
Fld_1 := Libadalang.Implementation.Body_Node_P_Body_Initial_Env_Name (Node => Self);
--# end
--# expr-done 4644
   If_Result := Fld_1;
end if;
--# expr-done 4645
Let_Result := If_Result; 
--# end
--# expr-done 4646
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Body_P_Initial_Env_Name;
--# end
--# property-start BaseSubpBody.previous_part_env_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18790
pragma Warnings (Off, "is not referenced");
 function Base_Subp_Body_P_Previous_Part_Env_Name
  (Node : Bare_Base_Subp_Body
  )
   return Symbol_Type
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type;
            procedure Finalizer_Scope_1327 with Inline_Always;
      Fld : Boolean;
Fld_1 : String_Type;
Str : String_Type;
Concat_Result : String_Type;
Sym : Symbol_Type;
Fld_2 : Boolean;
Fld_3 : String_Type;
Sym_1 : Symbol_Type;
If_Result : Symbol_Type;
If_Result_1 : Symbol_Type;
Let_Result : Symbol_Type;
            procedure Finalizer_Scope_1327 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Str);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_1327;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4656 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4655 '<Cond at ast.py:18796>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18796
--# expr-start 4647 '<FieldAccess for is_subunit at ast.py:18797>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18797
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.is_subunit
Fld := Libadalang.Implementation.Body_Node_P_Is_Subunit (Node => Self);
--# end
--# expr-done 4647
if Fld then
   --# expr-start 4651 '<to_symbol at ast.py:18798>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18798
--# expr-start 4650 '<Concat at ast.py:18798>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18798
--# expr-start 4648 '<FieldAccess for top_level_env_name at ast.py:18798>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18798
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.top_level_env_name
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Top_Level_Env_Name (Node => Self);
--# end
--# expr-done 4648
--# expr-start 4649 '<String at ast.py:18798>' Str /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18798
Str := Create_String ("__stub"); 
--# expr-done 4649
Concat_Result := Concat_String (Fld_1, Str); 
--# expr-done 4650
Sym := String_To_Symbol (Self.Unit.Context, Concat_Result); 
--# expr-done 4651
   If_Result_1 := Sym;
else
--# expr-start 4652 '<FieldAccess for is_library_item at ast.py:18800>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18800
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_library_item'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Library_Item (Node => Self);
--# end
--# expr-done 4652
if Fld_2 then
   --# expr-start 4654 '<to_symbol at ast.py:18801>' Sym_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18801
--# expr-start 4653 '<FieldAccess for top_level_env_name at ast.py:18801>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18801
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.top_level_env_name
Fld_3 := Libadalang.Implementation.Basic_Decl_P_Top_Level_Env_Name (Node => Self);
--# end
--# expr-done 4653
Sym_1 := String_To_Symbol (Self.Unit.Context, Fld_3); 
--# expr-done 4654
   If_Result := Sym_1;
else
   If_Result := null;
end if;
   If_Result_1 := If_Result;
end if;
--# expr-done 4655
Let_Result := If_Result_1; 
--# end
--# expr-done 4656
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1327;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1327;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Subp_Body_P_Previous_Part_Env_Name;
--# end
--# property-start BaseSubpBody._env_do_134 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18807
pragma Warnings (Off, "is not referenced");
 function Env_Do_134
  (Node : Bare_Base_Subp_Body
  )
   return Boolean
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4657 '<FieldAccess for env_hook at ast.py:18807>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18807
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.env_hook
Fld := Libadalang.Implementation.Ada_Node_P_Env_Hook (Node => Self);
--# end
--# expr-done 4657
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_134;
--# end
--# property-start BaseSubpBody._initial_env_135 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18810
pragma Warnings (Off, "is not referenced");
 function Initial_Env_135
  (Node : Bare_Base_Subp_Body
  )
   return Internal_Designated_Env
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_1552 with Inline_Always;
            procedure Finalizer_Scope_2410 with Inline_Always;
      Var_Expr : Symbol_Type;
Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Scope_Result : Internal_Designated_Env := No_Designated_Env;
New_Struct_1 : Internal_Designated_Env := No_Designated_Env;
Result_Var : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_1552 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (New_Struct_1);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1552;
            procedure Finalizer_Scope_2410 is
            begin
                     Dec_Ref (New_Struct);
            end Finalizer_Scope_2410;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
--# expr-start 4658 '<FieldAccess for initial_env_name at ast.py:18811>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18811
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpBody.initial_env_name
Fld := Libadalang.Implementation.Base_Subp_Body_P_Initial_Env_Name (Node => Self, Follow_Private => True);
--# end
--# expr-done 4658
Var_Expr := Fld;
if Var_Expr /= null then
   --# scope-start
   --# bind non_null_name Var_Expr
--# expr-start 4659 '<New DesignatedEnv at ???>' New_Struct None
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => Var_Expr, Kind => Named_Env); 
--# expr-done 4659
Scope_Result := New_Struct;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2410;
   Result_Var := Scope_Result;
else
   --# expr-start 4660 '<New DesignatedEnv at ast.py:18810>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18810
Inc_Ref (Empty_Env);
New_Struct_1 := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 4660
   Result_Var := New_Struct_1;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1552;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1552;
               Finalizer_Scope_2410;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Initial_Env_135;
--# end
--# property-start BaseSubpBody._env_mappings_136 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18817
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_136
  (Node : Bare_Base_Subp_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Body :=
     Internal_Entity_Base_Subp_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1553 with Inline_Always;
            procedure Finalizer_Scope_2411 with Inline_Always;
      Var_Expr : Symbol_Type;
Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Scope_Result : Internal_Designated_Env := No_Designated_Env;
New_Struct_1 : Internal_Designated_Env := No_Designated_Env;
Result_Var : Internal_Designated_Env := No_Designated_Env;
Fld_1 : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1553 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (New_Struct_1);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1553;
            procedure Finalizer_Scope_2411 is
            begin
                     Dec_Ref (New_Struct);
            end Finalizer_Scope_2411;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4664 '<FieldAccess for basic_decl_env_assocs at ast.py:18817>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18817
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4661 '<FieldAccess for initial_env_name at ast.py:18818>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18818
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpBody.initial_env_name
Fld := Libadalang.Implementation.Base_Subp_Body_P_Initial_Env_Name (Node => Self, Follow_Private => False);
--# end
--# expr-done 4661
Var_Expr := Fld;
if Var_Expr /= null then
   --# scope-start
   --# bind non_null_name Var_Expr
--# expr-start 4662 '<New DesignatedEnv at ???>' New_Struct None
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => Var_Expr, Kind => Named_Env); 
--# expr-done 4662
Scope_Result := New_Struct;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2411;
   Result_Var := Scope_Result;
else
   --# expr-start 4663 '<New DesignatedEnv at ast.py:18818>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18818
Inc_Ref (Empty_Env);
New_Struct_1 := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 4663
   Result_Var := New_Struct_1;
end if;
      Inc_Ref (Result_Var);
--# property-call-start BasicDecl.basic_decl_env_assocs
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Basic_Decl_Env_Assocs (Node => Ent.Node, Dest_Env => Result_Var, E_Info => Ent.Info);
--# end
--# expr-done 4664
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1553;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1553;
               Finalizer_Scope_2411;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_136;
--# end
--# property-start BaseSubpBody._env_trans_parent_137 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18806
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_137
  (Node : Bare_Base_Subp_Body
  )
   return Boolean
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_137;
--# end
--# property-start BaseSubpBody._env_mappings_138 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18824
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_138
  (Node : Bare_Base_Subp_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Subp_Body :=
     Internal_Entity_Base_Subp_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1555 with Inline_Always;
            procedure Finalizer_Scope_2412 with Inline_Always;
            procedure Finalizer_Scope_2413 with Inline_Always;
      Var_Expr : Symbol_Type;
Var_Expr_1 : Lexical_Env := Empty_Env;
Sym : Symbol_Type;
Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Scope_Result : Internal_Designated_Env := No_Designated_Env;
New_Struct_1 : Internal_Designated_Env := No_Designated_Env;
Fld_1 : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_2 : Lexical_Env := Empty_Env;
Dyn_Var_Bind_Result : Lexical_Env := Empty_Env;
New_Struct_2 : Internal_Designated_Env := No_Designated_Env;
Scope_Result_1 : Internal_Designated_Env := No_Designated_Env;
New_Struct_3 : Internal_Designated_Env := No_Designated_Env;
Result_Var : Internal_Designated_Env := No_Designated_Env;
Fld_3 : Boolean;
If_Result : Internal_Designated_Env := No_Designated_Env;
Result_Var_1 : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Base_Subp_Body;
Cast_Result : Bare_Ada_Node;
New_Struct_4 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1555 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (New_Struct_1);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (New_Struct_3);
                     Dec_Ref (Result_Var);
                     Dec_Ref (If_Result);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (New_Struct_4);
            end Finalizer_Scope_1555;
            procedure Finalizer_Scope_2412 is
            begin
                     Dec_Ref (New_Struct);
            end Finalizer_Scope_2412;
            procedure Finalizer_Scope_2413 is
            begin
                     Dec_Ref (Var_Expr_1);
                     Dec_Ref (New_Struct_2);
            end Finalizer_Scope_2413;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4676 '<New EnvAssoc at ast.py:18824>' New_Struct_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18824
--# expr-start 4665 '<FieldAccess for previous_part_env_name at ast.py:18827>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18827
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Body.previous_part_env_name'
Fld := Libadalang.Implementation.Dispatcher_Body_Node_P_Previous_Part_Env_Name (Node => Self);
--# end
--# expr-done 4665
Var_Expr := Fld;
if Var_Expr /= null then
   --# scope-start
   --# bind name Var_Expr
--# expr-start 4666 '<New DesignatedEnv at ast.py:18828>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18828
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => Var_Expr, Kind => Named_Env); 
--# expr-done 4666
Scope_Result := New_Struct;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2412;
   Result_Var_1 := Scope_Result;
else
   --# expr-start 4674 '<If at ast.py:18834>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18834
--# expr-start 4667 '<FieldAccess for has_top_level_env_name at ast.py:18835>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18835
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.has_top_level_env_name
Fld_3 := Libadalang.Implementation.Basic_Decl_P_Has_Top_Level_Env_Name (Node => Self);
--# end
--# expr-done 4667
if Fld_3 then
   --# expr-start 4668 '<New DesignatedEnv at ast.py:18836>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18836
Inc_Ref (Empty_Env);
New_Struct_1 := (Direct_Env => Empty_Env, Env_Name => null, Kind => None); 
--# expr-done 4668
   If_Result := New_Struct_1;
else
--# expr-start 4671 '<bind at ast.py:18838>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18838
--# expr-start 4669 '<FieldAccess for default_initial_env at ast.py:18839>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18839
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.default_initial_env
Fld_1 := Libadalang.Implementation.Ada_Node_P_Default_Initial_Env (Node => Self);
--# end
--# expr-done 4669
Bound_Env := Fld_1; Inc_Ref (Bound_Env);
--# expr-start 4670 '<FieldAccess for body_scope at ast.py:18840>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18840
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.body_scope
Fld_2 := Libadalang.Implementation.Body_Node_P_Body_Scope (Node => Ent.Node, Follow_Private => False, Force_Decl => True, Env => Bound_Env, E_Info => Ent.Info);
--# end
--# expr-done 4670
Dyn_Var_Bind_Result := Fld_2; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 4671
Var_Expr_1 := Dyn_Var_Bind_Result;
      Inc_Ref (Var_Expr_1);
if Var_Expr_1 /= Empty_Env then
   --# scope-start
   --# bind non_null_env Var_Expr_1
--# expr-start 4672 '<New DesignatedEnv at ???>' New_Struct_2 None
Inc_Ref (Var_Expr_1);
New_Struct_2 := (Direct_Env => Var_Expr_1, Env_Name => null, Kind => Direct_Env); 
--# expr-done 4672
Scope_Result_1 := New_Struct_2;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2413;
   Result_Var := Scope_Result_1;
else
   --# expr-start 4673 '<New DesignatedEnv at ast.py:18837>' New_Struct_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18837
Inc_Ref (Empty_Env);
New_Struct_3 := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 4673
   Result_Var := New_Struct_3;
end if;
      Inc_Ref (Result_Var);
   If_Result := Result_Var;
end if;
      Inc_Ref (If_Result);
--# expr-done 4674
   Result_Var_1 := If_Result;
end if;
      Inc_Ref (Result_Var_1);
--# expr-start 4675 '<Symbol __nextpart at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Nextpart); 
--# expr-done 4675
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (Result_Var_1);
New_Struct_4 := (Dest_Env => Result_Var_1, Key => Sym, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 4676
      Property_Result := New_Struct_4;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1555;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1555;
               Finalizer_Scope_2412;
               Finalizer_Scope_2413;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_138;
--# end
--# property-start BaseSubpBody._env_do_139 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18851
pragma Warnings (Off, "is not referenced");
 function Env_Do_139
  (Node : Bare_Base_Subp_Body
  )
   return Bare_Compilation_Unit_Array_Access
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1556 with Inline_Always;
      Fld : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1556 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1556;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4677 '<FieldAccess for populate_dependent_units at ast.py:18851>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18851
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.populate_dependent_units
Fld := Libadalang.Implementation.Basic_Decl_P_Populate_Dependent_Units (Node => Self);
--# end
--# expr-done 4677
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1556;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1556;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_139;
--# end
--# property-start BaseSubpBody._ref_env_nodes_140 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18854
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_140
  (Node : Bare_Base_Subp_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1557 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1557 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1557;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4678 '<FieldAccess for top_level_use_package_clauses at ast.py:18854>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18854
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_package_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Package_Clauses (Node => Self);
--# end
--# expr-done 4678
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1557;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1557;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_140;
--# end
--# property-start BaseSubpBody._ref_cond_141 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18856
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_141
  (Node : Bare_Base_Subp_Body
  )
   return Boolean
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4680 '<IsA LibraryItem, Subunit at ast.py:18856>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18856
--# expr-start 4679 '<parent at ast.py:18856>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18856
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 4679
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 4680
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_141;
--# end
--# property-start BaseSubpBody._ref_env_nodes_142 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18860
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_142
  (Node : Bare_Base_Subp_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1559 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1559 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1559;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4681 '<FieldAccess for top_level_use_type_clauses at ast.py:18860>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18860
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_type_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Type_Clauses (Node => Self);
--# end
--# expr-done 4681
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1559;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1559;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_142;
--# end
--# property-start BaseSubpBody._ref_cond_143 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18862
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_143
  (Node : Bare_Base_Subp_Body
  )
   return Boolean
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4683 '<IsA LibraryItem, Subunit at ast.py:18862>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18862
--# expr-start 4682 '<parent at ast.py:18862>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18862
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 4682
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 4683
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_143;
--# end
--# property-start BaseSubpBody._ref_env_nodes_144 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18869
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_144
  (Node : Bare_Base_Subp_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1561 with Inline_Always;
            procedure Finalizer_Scope_2414 with Inline_Always;
      Var_Expr : Bare_Ada_Node;
Cast_Expr : Bare_Base_Subp_Body;
Cast_Result : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
Scope_Result : Bare_Ada_Node_Array_Access;
Result_Var : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1561 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1561;
            procedure Finalizer_Scope_2414 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2414;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 4684 '<Cast to AdaNode at ast.py:18869>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18869
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-done 4684
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2414;
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node_Array_Type;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1561;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1561;
               Finalizer_Scope_2414;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_144;
--# end
--# property-start BaseSubpBody._ref_cond_145 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18871
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_145
  (Node : Bare_Base_Subp_Body
  )
   return Boolean
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4686 '<IsA LibraryItem at ast.py:18871>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18871
--# expr-start 4685 '<parent at ast.py:18871>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18871
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 4685
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range; 
--# expr-done 4686
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_145;
--# end
--# property-start BaseSubpBody._ref_env_nodes_146 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18875
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_146
  (Node : Bare_Base_Subp_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1563 with Inline_Always;
            procedure Finalizer_Scope_2415 with Inline_Always;
      Var_Expr : Bare_Ada_Node;
Cast_Expr : Bare_Base_Subp_Body;
Cast_Result : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
Scope_Result : Bare_Ada_Node_Array_Access;
Result_Var : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1563 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1563;
            procedure Finalizer_Scope_2415 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2415;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 4687 '<Cast to AdaNode at ast.py:18875>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18875
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-done 4687
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2415;
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node_Array_Type;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1563;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1563;
               Finalizer_Scope_2415;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_146;
--# end
--# property-start BaseSubpBody._ref_cond_147 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18877
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_147
  (Node : Bare_Base_Subp_Body
  )
   return Boolean
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4688 '<FieldAccess for should_ref_generic_formals at ast.py:18877>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18877
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.should_ref_generic_formals
Fld := Libadalang.Implementation.Basic_Decl_P_Should_Ref_Generic_Formals (Node => Self);
--# end
--# expr-done 4688
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_147;
--# end
--# property-start BaseSubpBody._ref_env_nodes_148 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18885
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_148
  (Node : Bare_Base_Subp_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1565 with Inline_Always;
            procedure Finalizer_Scope_2416 with Inline_Always;
      Var_Expr : Bare_Ada_Node;
Cast_Expr : Bare_Base_Subp_Body;
Cast_Result : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
Scope_Result : Bare_Ada_Node_Array_Access;
Result_Var : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1565 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1565;
            procedure Finalizer_Scope_2416 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2416;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 4689 '<Cast to AdaNode at ast.py:18885>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18885
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-done 4689
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2416;
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node_Array_Type;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1565;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1565;
               Finalizer_Scope_2416;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_148;
--# end
--# property-start BaseSubpBody._ref_cond_149 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18887
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_149
  (Node : Bare_Base_Subp_Body
  )
   return Boolean
is
   Self : Bare_Base_Subp_Body := Bare_Base_Subp_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4690 '<FieldAccess for should_ref_generic_formals at ast.py:18887>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18887
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.should_ref_generic_formals
Fld := Libadalang.Implementation.Basic_Decl_P_Should_Ref_Generic_Formals (Node => Self);
--# end
--# expr-done 4690
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_149;
--# end
   --
   --  Primitives for Bare_Expr_Function
   --
      procedure Initialize_Fields_For_Expr_Function
        (Self : Bare_Expr_Function
         ; Base_Subp_Body_F_Overriding : Bare_Overriding_Node
         ; Base_Subp_Body_F_Subp_Spec : Bare_Subp_Spec
         ; Expr_Function_F_Expr : Bare_Expr
         ; Expr_Function_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Initialize_Fields_For_Base_Subp_Body
              (Self, Base_Subp_Body_F_Overriding, Base_Subp_Body_F_Subp_Spec);
            Self.Expr_Function_F_Expr := Expr_Function_F_Expr;
            Self.Expr_Function_F_Aspects := Expr_Function_F_Aspects;
      end Initialize_Fields_For_Expr_Function;
   function Expr_Function_F_Expr
     (Node : Bare_Expr_Function) return Bare_Expr
   is
   begin
         return Node.Expr_Function_F_Expr;
   end;
--# property-start ExprFunction.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18901
pragma Warnings (Off, "is not referenced");
 function Expr_Function_P_Xref_Equation
  (Node : Bare_Expr_Function
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Expr_Function := Bare_Expr_Function (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Expr_Function :=
     Internal_Entity_Expr_Function'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1328 with Inline_Always;
      Fld : Internal_Entity_Expr;
Fld_1 : Logic_Var;
Fld_2 : Internal_Entity_Subp_Spec;
Fld_3 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_4 : Internal_Entity_Expr;
Fld_5 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_6 : Bare_Expr;
Fld_7 : Logic_Equation;
And_Pred_1 : Logic_Equation;
            procedure Finalizer_Scope_1328 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_5);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_7);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_1328;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4701 '<And at ast.py:18902>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18902
--# expr-start 4698 '<And at ast.py:18902>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18902
--# expr-start 4695 '<Bind at ast.py:18902>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18902
--# expr-start 4692 '<FieldAccess for expected_type_var at ast.py:18902>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18902
--# expr-start 4691 '<FieldAccess for expr at ast.py:18902>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18902
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Expr_Function_F_Expr, Info => Ent.Info);
--# expr-done 4691
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld.Node);
--# expr-done 4692
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 4694 '<FieldAccess for return_type at ast.py:18902>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18902
--# expr-start 4693 '<FieldAccess for subp_spec at ast.py:18902>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18902
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Subp_Spec (Node => Ent.Node.Base_Subp_Body_F_Subp_Spec, Info => Ent.Info);
--# expr-done 4693
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.return_type
Fld_3 := Libadalang.Implementation.Base_Subp_Spec_P_Return_Type (Node => Fld_2.Node, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 4694
      Cast_Result := Create_Internal_Entity
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
Bind_Result := Solver.Create_Assign (Fld_1, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18902") else null)); 
--# expr-done 4695
--# expr-start 4697 '<FieldAccess for sub_equation at ast.py:18903>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18903
--# expr-start 4696 '<FieldAccess for expr at ast.py:18903>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18903
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Expr (Node => Ent.Node.Expr_Function_F_Expr, Info => Ent.Info);
--# expr-done 4696
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_5 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_4.Node, Env => Env, Origin => Origin, E_Info => Fld_4.Info);
--# end
--# expr-done 4697
And_Pred := Create_And (Bind_Result, Fld_5, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18902") else null)); 
--# expr-done 4698
--# expr-start 4700 '<FieldAccess for matches_expected_assign_type at ast.py:18904>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18904
--# expr-start 4699 '<FieldAccess for expr at ast.py:18904>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18904
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Self.Expr_Function_F_Expr;
--# expr-done 4699
      if Fld_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_assign_type
Fld_7 := Libadalang.Implementation.Expr_P_Matches_Expected_Assign_Type (Node => Fld_6, Origin => Origin);
--# end
--# expr-done 4700
And_Pred_1 := Create_And (And_Pred, Fld_7, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18902") else null)); 
--# expr-done 4701
      Property_Result := And_Pred_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1328;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1328;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_Function_P_Xref_Equation;
--# end
--# property-start ExprFunction.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18907
pragma Warnings (Off, "is not referenced");
 function Expr_Function_P_Xref_Entry_Point
  (Node : Bare_Expr_Function
  )
   return Boolean
is
   Self : Bare_Expr_Function := Bare_Expr_Function (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_Function_P_Xref_Entry_Point;
--# end
   --
   --  Primitives for Bare_Null_Subp_Decl
   --
      procedure Initialize_Fields_For_Null_Subp_Decl
        (Self : Bare_Null_Subp_Decl
         ; Base_Subp_Body_F_Overriding : Bare_Overriding_Node
         ; Base_Subp_Body_F_Subp_Spec : Bare_Subp_Spec
         ; Null_Subp_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Initialize_Fields_For_Base_Subp_Body
              (Self, Base_Subp_Body_F_Overriding, Base_Subp_Body_F_Subp_Spec);
            Self.Null_Subp_Decl_F_Aspects := Null_Subp_Decl_F_Aspects;
      end Initialize_Fields_For_Null_Subp_Decl;
   --
   --  Primitives for Bare_Subp_Body
   --
      procedure Initialize_Fields_For_Subp_Body
        (Self : Bare_Subp_Body
         ; Base_Subp_Body_F_Overriding : Bare_Overriding_Node
         ; Base_Subp_Body_F_Subp_Spec : Bare_Subp_Spec
         ; Subp_Body_F_Aspects : Bare_Aspect_Spec
         ; Subp_Body_F_Decls : Bare_Declarative_Part
         ; Subp_Body_F_Stmts : Bare_Handled_Stmts
         ; Subp_Body_F_End_Name : Bare_End_Name
        ) is
      begin
            Initialize_Fields_For_Base_Subp_Body
              (Self, Base_Subp_Body_F_Overriding, Base_Subp_Body_F_Subp_Spec);
            Self.Subp_Body_F_Aspects := Subp_Body_F_Aspects;
            Self.Subp_Body_F_Decls := Subp_Body_F_Decls;
            Self.Subp_Body_F_Stmts := Subp_Body_F_Stmts;
            Self.Subp_Body_F_End_Name := Subp_Body_F_End_Name;
      end Initialize_Fields_For_Subp_Body;
   function Subp_Body_F_Decls
     (Node : Bare_Subp_Body) return Bare_Declarative_Part
   is
   begin
         return Node.Subp_Body_F_Decls;
   end;
   function Subp_Body_F_Stmts
     (Node : Bare_Subp_Body) return Bare_Handled_Stmts
   is
   begin
         return Node.Subp_Body_F_Stmts;
   end;
   function Subp_Body_F_End_Name
     (Node : Bare_Subp_Body) return Bare_End_Name
   is
   begin
         return Node.Subp_Body_F_End_Name;
   end;
--# property-start SubpBody.declarative_parts /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18970
pragma Warnings (Off, "is not referenced");
 function Subp_Body_P_Declarative_Parts
  (Node : Bare_Subp_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Declarative_Part_Array_Access
is
   Self : Bare_Subp_Body := Bare_Subp_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Subp_Body :=
     Internal_Entity_Subp_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_1332 with Inline_Always;
      Fld : Internal_Entity_Declarative_Part;
Singleton : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_1332 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1332;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4702 '<FieldAccess for decls at ast.py:18970>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18970
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Declarative_Part (Node => Ent.Node.Subp_Body_F_Decls, Info => Ent.Info);
--# expr-done 4702
                Singleton := Create_Internal_Entity_Declarative_Part_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1332;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1332;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subp_Body_P_Declarative_Parts;
--# end
   --
   --  Primitives for Bare_Subp_Renaming_Decl
   --
      procedure Initialize_Fields_For_Subp_Renaming_Decl
        (Self : Bare_Subp_Renaming_Decl
         ; Base_Subp_Body_F_Overriding : Bare_Overriding_Node
         ; Base_Subp_Body_F_Subp_Spec : Bare_Subp_Spec
         ; Subp_Renaming_Decl_F_Renames : Bare_Renaming_Clause
         ; Subp_Renaming_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Initialize_Fields_For_Base_Subp_Body
              (Self, Base_Subp_Body_F_Overriding, Base_Subp_Body_F_Subp_Spec);
            Self.Subp_Renaming_Decl_F_Renames := Subp_Renaming_Decl_F_Renames;
            Self.Subp_Renaming_Decl_F_Aspects := Subp_Renaming_Decl_F_Aspects;
      end Initialize_Fields_For_Subp_Renaming_Decl;
   function Subp_Renaming_Decl_F_Renames
     (Node : Bare_Subp_Renaming_Decl) return Bare_Renaming_Clause
   is
   begin
         return Node.Subp_Renaming_Decl_F_Renames;
   end;
--# property-start SubpRenamingDecl.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18926
pragma Warnings (Off, "is not referenced");
 function Subp_Renaming_Decl_P_Xref_Entry_Point
  (Node : Bare_Subp_Renaming_Decl
  )
   return Boolean
is
   Self : Bare_Subp_Renaming_Decl := Bare_Subp_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subp_Renaming_Decl_P_Xref_Entry_Point;
--# end
--# property-start SubpRenamingDecl.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18927
pragma Warnings (Off, "is not referenced");
 function Subp_Renaming_Decl_P_Xref_Equation
  (Node : Bare_Subp_Renaming_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Subp_Renaming_Decl := Bare_Subp_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Subp_Renaming_Decl :=
     Internal_Entity_Subp_Renaming_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1331 with Inline_Always;
      Fld : Internal_Entity_Renaming_Clause;
Fld_1 : Internal_Entity_Name;
Fld_2 : Logic_Var;
Fld_3 : Internal_Entity_Subp_Spec;
Fld_4 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_5 : Internal_Entity_Renaming_Clause;
Fld_6 : Internal_Entity_Name;
Fld_7 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_8 : Internal_Entity_Renaming_Clause;
Fld_9 : Internal_Entity_Name;
Fld_10 : Logic_Equation;
Fld_11 : Internal_Entity_Renaming_Clause;
Fld_12 : Internal_Entity_Name;
Fld_13 : Logic_Var;
Cast_Expr : Internal_Entity_Subp_Renaming_Decl;
Cast_Result_1 : Internal_Entity_Basic_Decl;
Pred : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_14 : Internal_Entity_Renaming_Clause;
Fld_15 : Internal_Entity_Name;
Is_A : Boolean;
If_Result : Logic_Equation;
False_Rel : Logic_Equation;
True_Rel : Logic_Equation;
Fld_16 : Internal_Entity_Renaming_Clause;
Fld_17 : Internal_Entity_Name;
Is_A_1 : Boolean;
If_Result_1 : Logic_Equation;
And_Pred_2 : Logic_Equation;
True_Rel_1 : Logic_Equation;
False_Rel_1 : Logic_Equation;
Fld_18 : Internal_Entity_Renaming_Clause;
Fld_19 : Internal_Entity_Name;
Fld_20 : Boolean;
If_Result_2 : Logic_Equation;
Or_Pred : Logic_Equation;
            procedure Finalizer_Scope_1331 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_7);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_10);
                     Dec_Ref (Pred);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (If_Result);
                     Dec_Ref (False_Rel);
                     Dec_Ref (True_Rel);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (True_Rel_1);
                     Dec_Ref (False_Rel_1);
                     Dec_Ref (If_Result_2);
                     Dec_Ref (Or_Pred);
            end Finalizer_Scope_1331;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4735 '<Or at ast.py:18927>' Or_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18927
--# expr-start 4730 '<And at ast.py:18928>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18928
--# expr-start 4725 '<If at ast.py:18929>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18929
--# expr-start 4705 '<IsA CharLiteral at ast.py:18930>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18930
--# expr-start 4704 '<FieldAccess for renamed_object at ast.py:18930>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18930
--# expr-start 4703 '<FieldAccess for renames at ast.py:18930>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18930
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Create_Internal_Entity_Renaming_Clause (Node => Ent.Node.Subp_Renaming_Decl_F_Renames, Info => Ent.Info);
--# expr-done 4703
      if Fld_14.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Create_Internal_Entity_Name (Node => Fld_14.Node.Renaming_Clause_F_Renamed_Object, Info => Fld_14.Info);
--# expr-done 4704
Is_A := Fld_15.Node /= null 
and then Fld_15.Node.Kind in Ada_Char_Literal_Range; 
--# expr-done 4705
if Is_A then
   --# expr-start 4715 '<And at ast.py:18934>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18934
--# expr-start 4711 '<Bind at ast.py:18934>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18934
--# expr-start 4708 '<FieldAccess for expected_type_var at ast.py:18934>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18934
--# expr-start 4707 '<FieldAccess for renamed_object at ast.py:18934>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18934
--# expr-start 4706 '<FieldAccess for renames at ast.py:18934>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18934
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Renaming_Clause (Node => Ent.Node.Subp_Renaming_Decl_F_Renames, Info => Ent.Info);
--# expr-done 4706
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Fld.Node.Renaming_Clause_F_Renamed_Object, Info => Fld.Info);
--# expr-done 4707
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_1.Node);
--# expr-done 4708
Fld_2.Value := No_Entity;
Entity_Vars.Reset (Fld_2);
--# expr-start 4710 '<FieldAccess for return_type at ast.py:18935>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18935
--# expr-start 4709 '<FieldAccess for subp_spec at ast.py:18935>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18935
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Subp_Spec (Node => Ent.Node.Base_Subp_Body_F_Subp_Spec, Info => Ent.Info);
--# expr-done 4709
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.return_type
Fld_4 := Libadalang.Implementation.Base_Subp_Spec_P_Return_Type (Node => Fld_3.Node, Origin => Origin, E_Info => Fld_3.Info);
--# end
--# expr-done 4710
      Cast_Result := Create_Internal_Entity
        (Node => Fld_4.Node,
         Info => Fld_4.Info);
Bind_Result := Solver.Create_Assign (Fld_2, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18934") else null)); 
--# expr-done 4711
--# expr-start 4714 '<FieldAccess for sub_equation at ast.py:18936>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18936
--# expr-start 4713 '<FieldAccess for renamed_object at ast.py:18936>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18936
--# expr-start 4712 '<FieldAccess for renames at ast.py:18936>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18936
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Renaming_Clause (Node => Ent.Node.Subp_Renaming_Decl_F_Renames, Info => Ent.Info);
--# expr-done 4712
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Name (Node => Fld_5.Node.Renaming_Clause_F_Renamed_Object, Info => Fld_5.Info);
--# expr-done 4713
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_7 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_6.Node, Env => Env, Origin => Origin, E_Info => Fld_6.Info);
--# end
--# expr-done 4714
And_Pred := Create_And (Bind_Result, Fld_7, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18934") else null)); 
--# expr-done 4715
   If_Result := And_Pred;
else
   --# expr-start 4724 '<And at ast.py:18938>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18938
--# expr-start 4718 '<FieldAccess for xref_no_overloading at ast.py:18938>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18938
--# expr-start 4717 '<FieldAccess for renamed_object at ast.py:18938>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18938
--# expr-start 4716 '<FieldAccess for renames at ast.py:18938>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18938
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Renaming_Clause (Node => Ent.Node.Subp_Renaming_Decl_F_Renames, Info => Ent.Info);
--# expr-done 4716
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Name (Node => Fld_8.Node.Renaming_Clause_F_Renamed_Object, Info => Fld_8.Info);
--# expr-done 4717
      if Fld_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_10 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Fld_9.Node, All_Els => True, Env => Env, Origin => Origin, E_Info => Fld_9.Info);
--# end
--# expr-done 4718
--# expr-start 4723 '<Predicate on BasicDecl.subp_decl_match_signature at ast.py:18939>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18939
--# expr-start 4721 '<FieldAccess for ref_var at ast.py:18940>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18940
--# expr-start 4720 '<FieldAccess for renamed_object at ast.py:18940>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18940
--# expr-start 4719 '<FieldAccess for renames at ast.py:18940>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18940
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Create_Internal_Entity_Renaming_Clause (Node => Ent.Node.Subp_Renaming_Decl_F_Renames, Info => Ent.Info);
--# expr-done 4719
      if Fld_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Create_Internal_Entity_Name (Node => Fld_11.Node.Renaming_Clause_F_Renamed_Object, Info => Fld_11.Info);
--# expr-done 4720
      if Fld_12.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_13 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_12.Node);
--# end
--# expr-done 4721
Fld_13.Value := No_Entity;
Entity_Vars.Reset (Fld_13);
--# expr-start 4722 '<Cast to BasicDecl at ast.py:18941>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18941
Cast_Expr := Ent; 
      Cast_Result_1 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
--# expr-done 4722
Pred := Solver.Create_Predicate (Fld_13, Create_Basic_Decl_P_Subp_Decl_Match_Signature_0_Predicate (Cast_Result_1)); 
--# expr-done 4723
And_Pred_1 := Create_And (Fld_10, Pred, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18938") else null)); 
--# expr-done 4724
   If_Result := And_Pred_1;
end if;
      Inc_Ref (If_Result);
--# expr-done 4725
--# expr-start 4729 '<If at ast.py:18950>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18950
--# expr-start 4728 '<IsA AttributeRef at ast.py:18950>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18950
--# expr-start 4727 '<FieldAccess for renamed_object at ast.py:18950>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18950
--# expr-start 4726 '<FieldAccess for renames at ast.py:18950>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18950
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_16 := Create_Internal_Entity_Renaming_Clause (Node => Ent.Node.Subp_Renaming_Decl_F_Renames, Info => Ent.Info);
--# expr-done 4726
      if Fld_16.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_17 := Create_Internal_Entity_Name (Node => Fld_16.Node.Renaming_Clause_F_Renamed_Object, Info => Fld_16.Info);
--# expr-done 4727
Is_A_1 := Fld_17.Node /= null 
and then Fld_17.Node.Kind in Ada_Attribute_Ref_Range; 
--# expr-done 4728
if Is_A_1 then
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18951") else null)); 
   If_Result_1 := False_Rel;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18951") else null)); 
   If_Result_1 := True_Rel;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 4729
And_Pred_2 := Create_And (If_Result, If_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18928") else null)); 
--# expr-done 4730
--# expr-start 4734 '<If at ast.py:18955>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18955
--# expr-start 4733 '<FieldAccess for is_operator_name at ast.py:18955>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18955
--# expr-start 4732 '<FieldAccess for renamed_object at ast.py:18955>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18955
--# expr-start 4731 '<FieldAccess for renames at ast.py:18955>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18955
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Create_Internal_Entity_Renaming_Clause (Node => Ent.Node.Subp_Renaming_Decl_F_Renames, Info => Ent.Info);
--# expr-done 4731
      if Fld_18.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Create_Internal_Entity_Name (Node => Fld_18.Node.Renaming_Clause_F_Renamed_Object, Info => Fld_18.Info);
--# expr-done 4732
      if Fld_19.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_operator_name
Fld_20 := Libadalang.Implementation.Name_P_Is_Operator_Name (Node => Fld_19.Node, E_Info => Fld_19.Info);
--# end
--# expr-done 4733
if Fld_20 then
   True_Rel_1 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18956") else null)); 
   If_Result_2 := True_Rel_1;
else
   False_Rel_1 := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18956") else null)); 
   If_Result_2 := False_Rel_1;
end if;
      Inc_Ref (If_Result_2);
--# expr-done 4734
Or_Pred := Create_Or (And_Pred_2, If_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18927") else null)); 
--# expr-done 4735
      Property_Result := Or_Pred;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1331;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1331;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subp_Renaming_Decl_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Body_Stub
   --
--# property-start BodyStub.next_part_for_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3837
pragma Warnings (Off, "is not referenced");
 function Body_Stub_P_Next_Part_For_Decl
  (Node : Bare_Body_Stub
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Body_Stub := Bare_Body_Stub (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Body_Stub :=
     Internal_Entity_Body_Stub'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_235 with Inline_Always;
      Fld : Symbol_Type_Array_Access;
Fld_1 : Bare_Basic_Decl;
As_Entity : Internal_Entity_Basic_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_235 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_235;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4739 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4738 '<as_entity at ast.py:3838>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3838
--# expr-start 4737 '<FieldAccess for get_unit_root_decl at ast.py:3838>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3838
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4736 '<FieldAccess for fully_qualified_name_array at ast.py:3839>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3839
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.fully_qualified_name_array
Fld := Libadalang.Implementation.Basic_Decl_P_Fully_Qualified_Name_Array (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4736
--# property-call-start AdaNode.get_unit_root_decl
Fld_1 := Libadalang.Implementation.Ada_Node_P_Get_Unit_Root_Decl (Node => Self, Name => Fld, Kind => Unit_Body, Process_Parents => False);
--# end
--# expr-done 4737
Is_Null := Fld_1 = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Decl;
else
As_Entity := (Info => E_Info, Node => Fld_1); 
   If_Result := As_Entity;
end if;
--# expr-done 4738
Let_Result := If_Result; 
--# end
--# expr-done 4739
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_235;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_235;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Stub_P_Next_Part_For_Decl;
--# end
--# property-start BodyStub.syntactic_fully_qualified_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3846
pragma Warnings (Off, "is not referenced");
 function Body_Stub_P_Syntactic_Fully_Qualified_Name
  (Node : Bare_Body_Stub
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Body_Stub := Bare_Body_Stub (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_236 with Inline_Always;
      Block_Var : Symbol_Type_Array_Access;
Block_Var_1 : Bare_Body_Node;
Block_Var_2 : Bare_Compilation_Unit;
Match_0 : Bare_Subp_Body;
Match_1 : Bare_Protected_Body;
Match_2 : Bare_Package_Body;
Match_3 : Bare_Task_Body;
Match_4 : Bare_Ada_Node;
As_Entity : Internal_Entity_Body_Stub;
Is_Null : Boolean;
If_Result : Internal_Entity_Body_Stub;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Symbol_Type_Array_Access;
Fld_2 : Bare_Ada_Node;
Fld_3 : Bare_Ada_Node;
Fld_4 : Bare_Ada_Node;
Match_Prefix : Bare_Ada_Node;
Cast_Result : Bare_Subp_Body;
Cast_Expr : Bare_Subp_Body;
Cast_Result_1 : Bare_Body_Node;
Let_Result : Bare_Body_Node;
Scope_Result : Bare_Body_Node;
Cast_Result_2 : Bare_Protected_Body;
Cast_Expr_1 : Bare_Protected_Body;
Cast_Result_3 : Bare_Body_Node;
Let_Result_1 : Bare_Body_Node;
Scope_Result_1 : Bare_Body_Node;
Cast_Result_4 : Bare_Package_Body;
Cast_Expr_2 : Bare_Package_Body;
Cast_Result_5 : Bare_Body_Node;
Let_Result_2 : Bare_Body_Node;
Scope_Result_2 : Bare_Body_Node;
Cast_Result_6 : Bare_Task_Body;
Cast_Expr_3 : Bare_Task_Body;
Cast_Result_7 : Bare_Body_Node;
Let_Result_3 : Bare_Body_Node;
Scope_Result_3 : Bare_Body_Node;
Cast_Result_8 : Bare_Ada_Node;
Let_Result_4 : Bare_Body_Node;
Scope_Result_4 : Bare_Body_Node;
Match_Result : Bare_Body_Node;
Fld_5 : Bare_Ada_Node;
Fld_6 : Bare_Ada_Node;
Cast_Result_9 : Bare_Compilation_Unit;
Fld_7 : Bare_Ada_Node;
Is_A : Boolean;
Fld_8 : Bare_Ada_Node;
Is_A_1 : Boolean;
If_Result_1 : Boolean;
If_Result_2 : Bare_Compilation_Unit;
Fld_9 : Symbol_Type_Array_Access;
Concat_Result : Symbol_Type_Array_Access;
Let_Result_5 : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_236 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Fld_9);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Let_Result_5);
            end Finalizer_Scope_236;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4757 '<Block at ???>' Let_Result_5 None
--# scope-start
--# expr-start 4742 '<FieldAccess for as_symbol_array at ast.py:3854>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3854
--# expr-start 4741 '<FieldAccess for defining_name at ast.py:3854>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3854
--# expr-start 4740 '<as_bare_entity at ast.py:3854>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3854
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity_Body_Stub;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 4740
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 4741
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.as_symbol_array'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_As_Symbol_Array (Node => Fld.Node);
--# end
--# expr-done 4742
Block_Var := Fld_1; Inc_Ref (Block_Var);
--# bind rel_name Block_Var
--# expr-start 4746 '<Match at ast.py:3865>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3865
--# expr-start 4745 '<parent at ast.py:3865>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3865
--# expr-start 4744 '<parent at ast.py:3865>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3865
--# expr-start 4743 '<parent at ast.py:3865>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3865
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Parent; 
--# expr-done 4743
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Fld_2.Parent; 
--# expr-done 4744
      if Fld_3 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Fld_3.Parent; 
--# expr-done 4745
      if Fld_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld_4; 
case Ada_Ada_Node (Match_Prefix.Kind) is
         when Ada_Subp_Body =>
   --# scope-start
--# scope-start
      Cast_Result := Match_Prefix;
Match_0 := Cast_Result; 
--# bind b Match_0
Cast_Expr := Match_0; 
      Cast_Result_1 := Cast_Expr;
Let_Result := Cast_Result_1; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Protected_Body =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Match_Prefix;
Match_1 := Cast_Result_2; 
--# bind b Match_1
Cast_Expr_1 := Match_1; 
      Cast_Result_3 := Cast_Expr_1;
Let_Result_1 := Cast_Result_3; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Package_Body =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Match_Prefix;
Match_2 := Cast_Result_4; 
--# bind b Match_2
Cast_Expr_2 := Match_2; 
      Cast_Result_5 := Cast_Expr_2;
Let_Result_2 := Cast_Result_5; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Task_Body =>
   --# scope-start
--# scope-start
      Cast_Result_6 := Match_Prefix;
Match_3 := Cast_Result_6; 
--# bind b Match_3
Cast_Expr_3 := Match_3; 
      Cast_Result_7 := Cast_Expr_3;
Let_Result_3 := Cast_Result_7; 
--# end
Scope_Result_3 := Let_Result_3;
   --# end
            Match_Result := Scope_Result_3; 
         when Ada_Abort_Absent .. Ada_Null_Subp_Decl | Ada_Subp_Renaming_Decl .. Ada_Entry_Body | Ada_Entry_Index_Spec .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_8 := Match_Prefix;
Match_4 := Cast_Result_8; 
--# bind _ Match_4
Let_Result_4 := raise Property_Error with "invalid body stub"; 
--# end
Scope_Result_4 := Let_Result_4;
   --# end
            Match_Result := Scope_Result_4; 
end case;
--# expr-done 4746
Block_Var_1 := Match_Result; 
--# bind top_body Block_Var_1
--# expr-start 4754 '<If at ast.py:3872>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3872
--# expr-start 4748 '<IsA LibraryItem at ast.py:3873>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3873
--# expr-start 4747 '<parent at ast.py:3873>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3873
      if Block_Var_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Block_Var_1.Parent; 
--# expr-done 4747
Is_A := Fld_7 /= null 
and then Fld_7.Kind in Ada_Library_Item_Range; 
--# expr-done 4748
if Is_A then
   If_Result_1 := True;
else
   --# expr-start 4750 '<IsA Subunit at ast.py:3874>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3874
--# expr-start 4749 '<parent at ast.py:3874>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3874
      if Block_Var_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Block_Var_1.Parent; 
--# expr-done 4749
Is_A_1 := Fld_8 /= null 
and then Fld_8.Kind in Ada_Subunit_Range; 
--# expr-done 4750
   If_Result_1 := Is_A_1;
end if;
if If_Result_1 then
   --# expr-start 4753 '<Cast to CompilationUnit at ast.py:3876>' Cast_Result_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3876
--# expr-start 4752 '<parent at ast.py:3876>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3876
--# expr-start 4751 '<parent at ast.py:3876>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3876
      if Block_Var_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Block_Var_1.Parent; 
--# expr-done 4751
      if Fld_5 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Fld_5.Parent; 
--# expr-done 4752
   if Fld_6 = null
      or else Fld_6.Kind in Ada_Compilation_Unit_Range
   then
      Cast_Result_9 := Fld_6;
   else
         Cast_Result_9 := No_Bare_Ada_Node;
   end if;
--# expr-done 4753
   If_Result_2 := Cast_Result_9;
else
   If_Result_2 := raise Property_Error with "invalid body stub";
end if;
--# expr-done 4754
Block_Var_2 := If_Result_2; 
--# bind cu Block_Var_2
--# expr-start 4756 '<Concat at ast.py:3880>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3880
--# expr-start 4755 '<FieldAccess for syntactic_fully_qualified_name at ast.py:3880>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3880
      if Block_Var_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.syntactic_fully_qualified_name
Fld_9 := Libadalang.Implementation.Compilation_Unit_P_Syntactic_Fully_Qualified_Name (Node => Block_Var_2);
--# end
--# expr-done 4755
Concat_Result := Concat (Fld_9, Block_Var); 
--# expr-done 4756
Let_Result_5 := Concat_Result; Inc_Ref (Let_Result_5);
--# end
--# expr-done 4757
      Property_Result := Let_Result_5;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_236;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_236;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Stub_P_Syntactic_Fully_Qualified_Name;
--# end
--# property-start BodyStub.env_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3883
pragma Warnings (Off, "is not referenced");
 function Body_Stub_P_Env_Names
  (Node : Bare_Body_Stub
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Body_Stub := Bare_Body_Stub (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_237 with Inline_Always;
      Fld : String_Type;
Str : String_Type;
Concat_Result : String_Type;
Sym : Symbol_Type;
Array_Lit : Symbol_Type_Array_Access;
Let_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_237 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Str);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_237;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4763 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4762 '<ArrayLiteral at ast.py:3888>' Array_Lit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3888
--# expr-start 4761 '<to_symbol at ast.py:3889>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3889
--# expr-start 4760 '<Concat at ast.py:3889>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3889
--# expr-start 4758 '<FieldAccess for top_level_env_name at ast.py:3889>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3889
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.top_level_env_name
Fld := Libadalang.Implementation.Basic_Decl_P_Top_Level_Env_Name (Node => Self);
--# end
--# expr-done 4758
--# expr-start 4759 '<String at ast.py:3889>' Str /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3889
Str := Create_String ("__stub"); 
--# expr-done 4759
Concat_Result := Concat_String (Fld, Str); 
--# expr-done 4760
Sym := String_To_Symbol (Self.Unit.Context, Concat_Result); 
--# expr-done 4761
Array_Lit := Create_Symbol_Type_Array (Internal_Symbol_Type_Array'(1 => Sym)); 
--# expr-done 4762
Let_Result := Array_Lit; Inc_Ref (Let_Result);
--# end
--# expr-done 4763
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_237;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_237;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Stub_P_Env_Names;
--# end
--# property-start BodyStub.previous_part_link_env_assoc /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3893
pragma Warnings (Off, "is not referenced");
 function Body_Stub_P_Previous_Part_Link_Env_Assoc
  (Node : Bare_Body_Stub
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Body_Stub := Bare_Body_Stub (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_238 with Inline_Always;
      Sym : Symbol_Type;
Fld : String_Type;
Sym_1 : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Body_Stub;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
Let_Result : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_238 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_238;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4769 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4768 '<New EnvAssoc at ast.py:3894>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3894
--# expr-start 4766 '<New DesignatedEnv at ast.py:3897>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3897
--# expr-start 4765 '<to_symbol at ast.py:3897>' Sym_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3897
--# expr-start 4764 '<FieldAccess for top_level_env_name at ast.py:3897>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3897
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.top_level_env_name
Fld := Libadalang.Implementation.Basic_Decl_P_Top_Level_Env_Name (Node => Self);
--# end
--# expr-done 4764
Sym_1 := String_To_Symbol (Self.Unit.Context, Fld); 
--# expr-done 4765
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => Sym_1, Kind => Named_Env); 
--# expr-done 4766
--# expr-start 4767 '<Symbol __nextpart at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Nextpart); 
--# expr-done 4767
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Sym, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 4768
Let_Result := New_Struct_1; Inc_Ref (Let_Result);
--# end
--# expr-done 4769
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_238;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_238;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Body_Stub_P_Previous_Part_Link_Env_Assoc;
--# end
   --
   --  Primitives for Bare_Package_Body_Stub
   --
      procedure Package_Body_Stub_Pre_Env_Actions
        (Self            : Bare_Package_Body_Stub;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_225 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:20122");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_226 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            Libadalang.Implementation.Env_Names_227 (Node => 
Self);
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Package_Body_Stub
        (Self : Bare_Package_Body_Stub
         ; Package_Body_Stub_F_Name : Bare_Defining_Name
         ; Package_Body_Stub_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Self.Package_Body_Stub_F_Name := Package_Body_Stub_F_Name;
            Self.Package_Body_Stub_F_Aspects := Package_Body_Stub_F_Aspects;
      end Initialize_Fields_For_Package_Body_Stub;
   function Package_Body_Stub_F_Name
     (Node : Bare_Package_Body_Stub) return Bare_Defining_Name
   is
   begin
         return Node.Package_Body_Stub_F_Name;
   end;
--# property-start PackageBodyStub.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20119
pragma Warnings (Off, "is not referenced");
 function Package_Body_Stub_P_Defining_Names
  (Node : Bare_Package_Body_Stub
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Package_Body_Stub := Bare_Package_Body_Stub (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Package_Body_Stub :=
     Internal_Entity_Package_Body_Stub'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1402 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1402 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1402;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4770 '<FieldAccess for name at ast.py:20119>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20119
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Package_Body_Stub_F_Name, Info => Ent.Info);
--# expr-done 4770
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1402;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1402;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Package_Body_Stub_P_Defining_Names;
--# end
--# property-start PackageBodyStub._env_mappings_225 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20122
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_225
  (Node : Bare_Package_Body_Stub
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Package_Body_Stub := Bare_Package_Body_Stub (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Package_Body_Stub :=
     Internal_Entity_Package_Body_Stub'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1642 with Inline_Always;
      Fld : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1642 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1642;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4771 '<FieldAccess for previous_part_link_env_assoc at ast.py:20122>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20122
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Body.previous_part_link_env_assoc'
Fld := Libadalang.Implementation.Dispatcher_Body_Node_P_Previous_Part_Link_Env_Assoc (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4771
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1642;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1642;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_225;
--# end
--# property-start PackageBodyStub._env_trans_parent_226 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20121
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_226
  (Node : Bare_Package_Body_Stub
  )
   return Boolean
is
   Self : Bare_Package_Body_Stub := Bare_Package_Body_Stub (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_226;
--# end
--# property-start PackageBodyStub._env_names_227 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20123
pragma Warnings (Off, "is not referenced");
 function Env_Names_227
  (Node : Bare_Package_Body_Stub
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Package_Body_Stub := Bare_Package_Body_Stub (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1644 with Inline_Always;
      Fld : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1644 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1644;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4772 '<FieldAccess for env_names at ast.py:20123>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20123
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BodyStub.env_names
Fld := Libadalang.Implementation.Body_Stub_P_Env_Names (Node => Self);
--# end
--# expr-done 4772
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1644;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1644;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Names_227;
--# end
   --
   --  Primitives for Bare_Protected_Body_Stub
   --
      procedure Protected_Body_Stub_Pre_Env_Actions
        (Self            : Bare_Protected_Body_Stub;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_215 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:20065");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_216 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            Libadalang.Implementation.Env_Names_217 (Node => 
Self);
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Protected_Body_Stub
        (Self : Bare_Protected_Body_Stub
         ; Protected_Body_Stub_F_Name : Bare_Defining_Name
         ; Protected_Body_Stub_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Self.Protected_Body_Stub_F_Name := Protected_Body_Stub_F_Name;
            Self.Protected_Body_Stub_F_Aspects := Protected_Body_Stub_F_Aspects;
      end Initialize_Fields_For_Protected_Body_Stub;
   function Protected_Body_Stub_F_Name
     (Node : Bare_Protected_Body_Stub) return Bare_Defining_Name
   is
   begin
         return Node.Protected_Body_Stub_F_Name;
   end;
--# property-start ProtectedBodyStub.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20062
pragma Warnings (Off, "is not referenced");
 function Protected_Body_Stub_P_Defining_Names
  (Node : Bare_Protected_Body_Stub
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Protected_Body_Stub := Bare_Protected_Body_Stub (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Protected_Body_Stub :=
     Internal_Entity_Protected_Body_Stub'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1399 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1399 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1399;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4773 '<FieldAccess for name at ast.py:20062>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20062
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Protected_Body_Stub_F_Name, Info => Ent.Info);
--# expr-done 4773
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1399;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1399;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Protected_Body_Stub_P_Defining_Names;
--# end
--# property-start ProtectedBodyStub._env_mappings_215 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20065
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_215
  (Node : Bare_Protected_Body_Stub
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Protected_Body_Stub := Bare_Protected_Body_Stub (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Protected_Body_Stub :=
     Internal_Entity_Protected_Body_Stub'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1632 with Inline_Always;
      Fld : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1632 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1632;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4774 '<FieldAccess for previous_part_link_env_assoc at ast.py:20065>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20065
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Body.previous_part_link_env_assoc'
Fld := Libadalang.Implementation.Dispatcher_Body_Node_P_Previous_Part_Link_Env_Assoc (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4774
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1632;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1632;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_215;
--# end
--# property-start ProtectedBodyStub._env_trans_parent_216 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20064
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_216
  (Node : Bare_Protected_Body_Stub
  )
   return Boolean
is
   Self : Bare_Protected_Body_Stub := Bare_Protected_Body_Stub (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_216;
--# end
--# property-start ProtectedBodyStub._env_names_217 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20066
pragma Warnings (Off, "is not referenced");
 function Env_Names_217
  (Node : Bare_Protected_Body_Stub
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Protected_Body_Stub := Bare_Protected_Body_Stub (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1634 with Inline_Always;
      Fld : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1634 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1634;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4775 '<FieldAccess for env_names at ast.py:20066>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20066
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BodyStub.env_names
Fld := Libadalang.Implementation.Body_Stub_P_Env_Names (Node => Self);
--# end
--# expr-done 4775
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1634;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1634;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Names_217;
--# end
   --
   --  Primitives for Bare_Subp_Body_Stub
   --
      procedure Subp_Body_Stub_Pre_Env_Actions
        (Self            : Bare_Subp_Body_Stub;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_218 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:20084");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_219 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            Libadalang.Implementation.Env_Names_220 (Node => 
Self);
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      if Libadalang.Implementation.Ref_Cond_222 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_221 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Ada_Node_P_Nested_Generic_Formal_Part'Access,
               Prioritary,
               All_Cats,
               True);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_224 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_223 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Ada_Node_P_Use_Clauses_In_Generic_Formal_Part'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      end;
      procedure Initialize_Fields_For_Subp_Body_Stub
        (Self : Bare_Subp_Body_Stub
         ; Subp_Body_Stub_F_Overriding : Bare_Overriding_Node
         ; Subp_Body_Stub_F_Subp_Spec : Bare_Subp_Spec
         ; Subp_Body_Stub_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Self.Subp_Body_Stub_F_Overriding := Subp_Body_Stub_F_Overriding;
            Self.Subp_Body_Stub_F_Subp_Spec := Subp_Body_Stub_F_Subp_Spec;
            Self.Subp_Body_Stub_F_Aspects := Subp_Body_Stub_F_Aspects;
      end Initialize_Fields_For_Subp_Body_Stub;
   function Subp_Body_Stub_F_Overriding
     (Node : Bare_Subp_Body_Stub) return Bare_Overriding_Node
   is
   begin
         return Node.Subp_Body_Stub_F_Overriding;
   end;
   function Subp_Body_Stub_F_Subp_Spec
     (Node : Bare_Subp_Body_Stub) return Bare_Subp_Spec
   is
   begin
         return Node.Subp_Body_Stub_F_Subp_Spec;
   end;
--# property-start SubpBodyStub.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20079
pragma Warnings (Off, "is not referenced");
 function Subp_Body_Stub_P_Defining_Names
  (Node : Bare_Subp_Body_Stub
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Subp_Body_Stub := Bare_Subp_Body_Stub (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Subp_Body_Stub :=
     Internal_Entity_Subp_Body_Stub'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1400 with Inline_Always;
      Fld : Internal_Entity_Subp_Spec;
Fld_1 : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1400 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1400;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4777 '<FieldAccess for name at ast.py:20079>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20079
--# expr-start 4776 '<FieldAccess for subp_spec at ast.py:20079>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20079
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Subp_Spec (Node => Ent.Node.Subp_Body_Stub_F_Subp_Spec, Info => Ent.Info);
--# expr-done 4776
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.name'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Name (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 4777
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld_1;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1400;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1400;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subp_Body_Stub_P_Defining_Names;
--# end
--# property-start SubpBodyStub.type_expression /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20108
pragma Warnings (Off, "is not referenced");
 function Subp_Body_Stub_P_Type_Expression
  (Node : Bare_Subp_Body_Stub
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Expr
is
   Self : Bare_Subp_Body_Stub := Bare_Subp_Body_Stub (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Subp_Body_Stub :=
     Internal_Entity_Subp_Body_Stub'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Type_Expr;
      Fld : Internal_Entity_Subp_Spec;
Fld_1 : Internal_Entity_Type_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4779 '<FieldAccess for returns at ast.py:20108>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20108
--# expr-start 4778 '<FieldAccess for subp_spec at ast.py:20108>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20108
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Subp_Spec (Node => Ent.Node.Subp_Body_Stub_F_Subp_Spec, Info => Ent.Info);
--# expr-done 4778
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.returns'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Returns (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 4779
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subp_Body_Stub_P_Type_Expression;
--# end
--# property-start SubpBodyStub._env_mappings_218 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20084
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_218
  (Node : Bare_Subp_Body_Stub
  )
   return Internal_Env_Assoc
is
   Self : Bare_Subp_Body_Stub := Bare_Subp_Body_Stub (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1635 with Inline_Always;
      Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Subp_Body_Stub;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1635 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1635;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4782 '<New EnvAssoc at ast.py:20084>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20084
--# expr-start 4780 '<New DesignatedEnv at ast.py:20084>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20084
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 4780
--# expr-start 4781 '<FieldAccess for name_symbol at ast.py:20085>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20085
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Self);
--# end
--# expr-done 4781
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 4782
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1635;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1635;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_218;
--# end
--# property-start SubpBodyStub._env_trans_parent_219 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20083
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_219
  (Node : Bare_Subp_Body_Stub
  )
   return Boolean
is
   Self : Bare_Subp_Body_Stub := Bare_Subp_Body_Stub (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_219;
--# end
--# property-start SubpBodyStub._env_names_220 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20088
pragma Warnings (Off, "is not referenced");
 function Env_Names_220
  (Node : Bare_Subp_Body_Stub
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Subp_Body_Stub := Bare_Subp_Body_Stub (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1637 with Inline_Always;
      Fld : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1637 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1637;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4783 '<FieldAccess for env_names at ast.py:20088>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20088
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BodyStub.env_names
Fld := Libadalang.Implementation.Body_Stub_P_Env_Names (Node => Self);
--# end
--# expr-done 4783
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1637;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1637;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Names_220;
--# end
--# property-start SubpBodyStub._ref_env_nodes_221 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20091
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_221
  (Node : Bare_Subp_Body_Stub
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Subp_Body_Stub := Bare_Subp_Body_Stub (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1638 with Inline_Always;
            procedure Finalizer_Scope_2422 with Inline_Always;
      Var_Expr : Bare_Ada_Node;
Cast_Expr : Bare_Subp_Body_Stub;
Cast_Result : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
Scope_Result : Bare_Ada_Node_Array_Access;
Result_Var : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1638 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1638;
            procedure Finalizer_Scope_2422 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2422;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 4784 '<Cast to AdaNode at ast.py:20091>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20091
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-done 4784
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2422;
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node_Array_Type;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1638;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1638;
               Finalizer_Scope_2422;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_221;
--# end
--# property-start SubpBodyStub._ref_cond_222 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20093
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_222
  (Node : Bare_Subp_Body_Stub
  )
   return Boolean
is
   Self : Bare_Subp_Body_Stub := Bare_Subp_Body_Stub (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4785 '<FieldAccess for should_ref_generic_formals at ast.py:20093>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20093
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.should_ref_generic_formals
Fld := Libadalang.Implementation.Basic_Decl_P_Should_Ref_Generic_Formals (Node => Self);
--# end
--# expr-done 4785
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_222;
--# end
--# property-start SubpBodyStub._ref_env_nodes_223 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20101
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_223
  (Node : Bare_Subp_Body_Stub
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Subp_Body_Stub := Bare_Subp_Body_Stub (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1640 with Inline_Always;
            procedure Finalizer_Scope_2423 with Inline_Always;
      Var_Expr : Bare_Ada_Node;
Cast_Expr : Bare_Subp_Body_Stub;
Cast_Result : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
Scope_Result : Bare_Ada_Node_Array_Access;
Result_Var : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1640 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1640;
            procedure Finalizer_Scope_2423 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2423;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 4786 '<Cast to AdaNode at ast.py:20101>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20101
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-done 4786
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2423;
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node_Array_Type;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1640;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1640;
               Finalizer_Scope_2423;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_223;
--# end
--# property-start SubpBodyStub._ref_cond_224 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20103
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_224
  (Node : Bare_Subp_Body_Stub
  )
   return Boolean
is
   Self : Bare_Subp_Body_Stub := Bare_Subp_Body_Stub (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4787 '<FieldAccess for should_ref_generic_formals at ast.py:20103>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20103
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.should_ref_generic_formals
Fld := Libadalang.Implementation.Basic_Decl_P_Should_Ref_Generic_Formals (Node => Self);
--# end
--# expr-done 4787
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_224;
--# end
   --
   --  Primitives for Bare_Task_Body_Stub
   --
      procedure Task_Body_Stub_Pre_Env_Actions
        (Self            : Bare_Task_Body_Stub;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_228 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:20138");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_229 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            Libadalang.Implementation.Env_Names_230 (Node => 
Self);
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Task_Body_Stub
        (Self : Bare_Task_Body_Stub
         ; Task_Body_Stub_F_Name : Bare_Defining_Name
         ; Task_Body_Stub_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Self.Task_Body_Stub_F_Name := Task_Body_Stub_F_Name;
            Self.Task_Body_Stub_F_Aspects := Task_Body_Stub_F_Aspects;
      end Initialize_Fields_For_Task_Body_Stub;
   function Task_Body_Stub_F_Name
     (Node : Bare_Task_Body_Stub) return Bare_Defining_Name
   is
   begin
         return Node.Task_Body_Stub_F_Name;
   end;
--# property-start TaskBodyStub.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20135
pragma Warnings (Off, "is not referenced");
 function Task_Body_Stub_P_Defining_Names
  (Node : Bare_Task_Body_Stub
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Task_Body_Stub := Bare_Task_Body_Stub (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Task_Body_Stub :=
     Internal_Entity_Task_Body_Stub'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1403 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1403 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1403;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4788 '<FieldAccess for name at ast.py:20135>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20135
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Task_Body_Stub_F_Name, Info => Ent.Info);
--# expr-done 4788
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1403;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1403;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Task_Body_Stub_P_Defining_Names;
--# end
--# property-start TaskBodyStub._env_mappings_228 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20138
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_228
  (Node : Bare_Task_Body_Stub
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Task_Body_Stub := Bare_Task_Body_Stub (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Task_Body_Stub :=
     Internal_Entity_Task_Body_Stub'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1645 with Inline_Always;
      Fld : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1645 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1645;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4789 '<FieldAccess for previous_part_link_env_assoc at ast.py:20138>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20138
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Body.previous_part_link_env_assoc'
Fld := Libadalang.Implementation.Dispatcher_Body_Node_P_Previous_Part_Link_Env_Assoc (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4789
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1645;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1645;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_228;
--# end
--# property-start TaskBodyStub._env_trans_parent_229 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20137
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_229
  (Node : Bare_Task_Body_Stub
  )
   return Boolean
is
   Self : Bare_Task_Body_Stub := Bare_Task_Body_Stub (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_229;
--# end
--# property-start TaskBodyStub._env_names_230 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20139
pragma Warnings (Off, "is not referenced");
 function Env_Names_230
  (Node : Bare_Task_Body_Stub
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Task_Body_Stub := Bare_Task_Body_Stub (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1647 with Inline_Always;
      Fld : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1647 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1647;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4790 '<FieldAccess for env_names at ast.py:20139>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20139
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BodyStub.env_names
Fld := Libadalang.Implementation.Body_Stub_P_Env_Names (Node => Self);
--# end
--# expr-done 4790
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1647;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1647;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Names_230;
--# end
   --
   --  Primitives for Bare_Entry_Body
   --
      procedure Entry_Body_Pre_Env_Actions
        (Self            : Bare_Entry_Body;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Dummy : Boolean;
      begin
         Dummy := Libadalang.Implementation.Env_Do_210 (Node => 
Self);
      end;
      declare
         Env : constant Internal_Designated_Env :=
           Libadalang.Implementation.Initial_Env_211 (Node => 
Self);
      begin
         Set_Initial_Env
           (Self,
            State,
            Env,
            DSL_Location => "ast.py:19960");
      end;
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mappings : Internal_Env_Assoc_Array_Access :=
            Libadalang.Implementation.Env_Mappings_212 (Node => 
Self);
      begin
         for Mapping of Mappings.Items loop
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:19970");
         end loop;
         Dec_Ref (Mappings);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_213 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Entry_Body
        (Self : Bare_Entry_Body
         ; Entry_Body_F_Entry_Name : Bare_Defining_Name
         ; Entry_Body_F_Index_Spec : Bare_Entry_Index_Spec
         ; Entry_Body_F_Params : Bare_Entry_Completion_Formal_Params
         ; Entry_Body_F_Aspects : Bare_Aspect_Spec
         ; Entry_Body_F_Barrier : Bare_Expr
         ; Entry_Body_F_Decls : Bare_Declarative_Part
         ; Entry_Body_F_Stmts : Bare_Handled_Stmts
         ; Entry_Body_F_End_Name : Bare_End_Name
        ) is
      begin
            Self.Entry_Body_F_Entry_Name := Entry_Body_F_Entry_Name;
            Self.Entry_Body_F_Index_Spec := Entry_Body_F_Index_Spec;
            Self.Entry_Body_F_Params := Entry_Body_F_Params;
            Self.Entry_Body_F_Aspects := Entry_Body_F_Aspects;
            Self.Entry_Body_F_Barrier := Entry_Body_F_Barrier;
            Self.Entry_Body_F_Decls := Entry_Body_F_Decls;
            Self.Entry_Body_F_Stmts := Entry_Body_F_Stmts;
            Self.Entry_Body_F_End_Name := Entry_Body_F_End_Name;
      end Initialize_Fields_For_Entry_Body;
   function Entry_Body_F_Entry_Name
     (Node : Bare_Entry_Body) return Bare_Defining_Name
   is
   begin
         return Node.Entry_Body_F_Entry_Name;
   end;
   function Entry_Body_F_Index_Spec
     (Node : Bare_Entry_Body) return Bare_Entry_Index_Spec
   is
   begin
         return Node.Entry_Body_F_Index_Spec;
   end;
   function Entry_Body_F_Params
     (Node : Bare_Entry_Body) return Bare_Entry_Completion_Formal_Params
   is
   begin
         return Node.Entry_Body_F_Params;
   end;
   function Entry_Body_F_Barrier
     (Node : Bare_Entry_Body) return Bare_Expr
   is
   begin
         return Node.Entry_Body_F_Barrier;
   end;
   function Entry_Body_F_Decls
     (Node : Bare_Entry_Body) return Bare_Declarative_Part
   is
   begin
         return Node.Entry_Body_F_Decls;
   end;
   function Entry_Body_F_Stmts
     (Node : Bare_Entry_Body) return Bare_Handled_Stmts
   is
   begin
         return Node.Entry_Body_F_Stmts;
   end;
   function Entry_Body_F_End_Name
     (Node : Bare_Entry_Body) return Bare_End_Name
   is
   begin
         return Node.Entry_Body_F_End_Name;
   end;
--# property-start EntryBody.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19953
pragma Warnings (Off, "is not referenced");
 function Entry_Body_P_Defining_Names
  (Node : Bare_Entry_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Entry_Body := Bare_Entry_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Body :=
     Internal_Entity_Entry_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1385 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1385 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1385;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4791 '<FieldAccess for entry_name at ast.py:19953>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19953
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Entry_Body_F_Entry_Name, Info => Ent.Info);
--# expr-done 4791
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1385;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1385;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Body_P_Defining_Names;
--# end
--# property-start EntryBody.declarative_parts /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19955
pragma Warnings (Off, "is not referenced");
 function Entry_Body_P_Declarative_Parts
  (Node : Bare_Entry_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Declarative_Part_Array_Access
is
   Self : Bare_Entry_Body := Bare_Entry_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Body :=
     Internal_Entity_Entry_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_1386 with Inline_Always;
      Fld : Internal_Entity_Declarative_Part;
Singleton : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_1386 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1386;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4792 '<FieldAccess for decls at ast.py:19955>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19955
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Declarative_Part (Node => Ent.Node.Entry_Body_F_Decls, Info => Ent.Info);
--# expr-done 4792
                Singleton := Create_Internal_Entity_Declarative_Part_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1386;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1386;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Body_P_Declarative_Parts;
--# end
--# property-start EntryBody.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19977
pragma Warnings (Off, "is not referenced");
 function Entry_Body_P_Xref_Entry_Point
  (Node : Bare_Entry_Body
  )
   return Boolean
is
   Self : Bare_Entry_Body := Bare_Entry_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Body_P_Xref_Entry_Point;
--# end
--# property-start EntryBody.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19980
pragma Warnings (Off, "is not referenced");
 function Entry_Body_P_Xref_Equation
  (Node : Bare_Entry_Body
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Entry_Body := Bare_Entry_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Body :=
     Internal_Entity_Entry_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1388 with Inline_Always;
      Fld : Bare_Expr;
Fld_1 : Logic_Var;
Fld_2 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_3 : Internal_Entity_Expr;
Fld_4 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_5 : Internal_Entity_Expr;
Fld_6 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1388 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_4);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1388;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4803 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4802 '<And at ast.py:19981>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19981
--# expr-start 4799 '<And at ast.py:19981>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19981
--# expr-start 4796 '<Bind at ast.py:19982>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19982
--# expr-start 4794 '<FieldAccess for expected_type_var at ast.py:19982>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19982
--# expr-start 4793 '<FieldAccess for barrier at ast.py:19982>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19982
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Entry_Body_F_Barrier;
--# expr-done 4793
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld);
--# expr-done 4794
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 4795 '<FieldAccess for bool_type at ast.py:19982>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19982
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_2 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 4795
      Cast_Result := Create_Internal_Entity
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Bind_Result := Solver.Create_Assign (Fld_1, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19982") else null)); 
--# expr-done 4796
--# expr-start 4798 '<FieldAccess for sub_equation at ast.py:19983>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19983
--# expr-start 4797 '<FieldAccess for barrier at ast.py:19983>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19983
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Expr (Node => Ent.Node.Entry_Body_F_Barrier, Info => Ent.Info);
--# expr-done 4797
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_4 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_3.Node, Env => Env, Origin => Origin, E_Info => Fld_3.Info);
--# end
--# expr-done 4798
And_Pred := Create_And (Bind_Result, Fld_4, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19981") else null)); 
--# expr-done 4799
--# expr-start 4801 '<FieldAccess for matches_expected_formal_prim_type at ast.py:19984>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19984
--# expr-start 4800 '<FieldAccess for barrier at ast.py:19984>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19984
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Expr (Node => Ent.Node.Entry_Body_F_Barrier, Info => Ent.Info);
--# expr-done 4800
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_prim_type
Fld_6 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Prim_Type (Node => Fld_5.Node, Origin => Origin);
--# end
--# expr-done 4801
And_Pred_1 := Create_And (And_Pred, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19981") else null)); 
--# expr-done 4802
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 4803
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1388;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1388;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Body_P_Xref_Equation;
--# end
--# property-start EntryBody._env_do_210 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19958
pragma Warnings (Off, "is not referenced");
 function Env_Do_210
  (Node : Bare_Entry_Body
  )
   return Boolean
is
   Self : Bare_Entry_Body := Bare_Entry_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4804 '<FieldAccess for env_hook at ast.py:19958>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19958
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.env_hook
Fld := Libadalang.Implementation.Ada_Node_P_Env_Hook (Node => Self);
--# end
--# expr-done 4804
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_210;
--# end
--# property-start EntryBody._initial_env_211 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19961
pragma Warnings (Off, "is not referenced");
 function Initial_Env_211
  (Node : Bare_Entry_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Designated_Env
is
   Self : Bare_Entry_Body := Bare_Entry_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Body :=
     Internal_Entity_Entry_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_1628 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_1 : Lexical_Env := Empty_Env;
Dyn_Var_Bind_Result : Lexical_Env := Empty_Env;
New_Struct : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_1628 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (New_Struct);
            end Finalizer_Scope_1628;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4808 '<New DesignatedEnv at ast.py:19961>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19961
--# expr-start 4807 '<bind at ast.py:19962>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19962
--# expr-start 4805 '<FieldAccess for default_initial_env at ast.py:19963>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19963
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.default_initial_env
Fld := Libadalang.Implementation.Ada_Node_P_Default_Initial_Env (Node => Self);
--# end
--# expr-done 4805
Bound_Env := Fld; Inc_Ref (Bound_Env);
--# expr-start 4806 '<FieldAccess for body_scope at ast.py:19964>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19964
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.body_scope
Fld_1 := Libadalang.Implementation.Body_Node_P_Body_Scope (Node => Ent.Node, Follow_Private => False, Env => Bound_Env, E_Info => Ent.Info);
--# end
--# expr-done 4806
Dyn_Var_Bind_Result := Fld_1; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 4807
Inc_Ref (Dyn_Var_Bind_Result);
New_Struct := (Direct_Env => Dyn_Var_Bind_Result, Env_Name => null, Kind => Direct_Env); 
--# expr-done 4808
      Property_Result := New_Struct;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1628;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1628;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Initial_Env_211;
--# end
--# property-start EntryBody._env_mappings_212 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19971
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_212
  (Node : Bare_Entry_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Entry_Body := Bare_Entry_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Body :=
     Internal_Entity_Entry_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1629 with Inline_Always;
      Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Entry_Body;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
Singleton : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1629 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1629;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4811 '<New EnvAssoc at ast.py:19971>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19971
--# expr-start 4809 '<New DesignatedEnv at ast.py:19971>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19971
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 4809
--# expr-start 4810 '<FieldAccess for name_symbol at ast.py:19971>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19971
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 4810
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 4811
                Singleton := Create_Internal_Env_Assoc_Array (Items_Count => 1);
                Singleton.Items (1) := New_Struct_1;
                Inc_Ref (Singleton.Items (1));
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1629;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1629;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_212;
--# end
--# property-start EntryBody._env_trans_parent_213 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19957
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_213
  (Node : Bare_Entry_Body
  )
   return Boolean
is
   Self : Bare_Entry_Body := Bare_Entry_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_213;
--# end
   --
   --  Primitives for Bare_Package_Body
   --
      procedure Package_Body_Pre_Env_Actions
        (Self            : Bare_Package_Body;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Dummy : Boolean;
      begin
         Dummy := Libadalang.Implementation.Env_Do_159 (Node => 
Self);
      end;
      declare
         Env : constant Internal_Designated_Env :=
           Libadalang.Implementation.Initial_Env_160 (Node => 
Self);
      begin
         Set_Initial_Env
           (Self,
            State,
            Env,
            DSL_Location => "ast.py:19700");
      end;
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_161 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:19702");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_162 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      declare
         Dummy : Bare_Compilation_Unit_Array_Access;
      begin
         Dummy := Libadalang.Implementation.Env_Do_163 (Node => 
Self);
         Dec_Ref (Dummy);
      end;
      if Libadalang.Implementation.Ref_Cond_165 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_164 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Use_Package_Name_Designated_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_167 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_166 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Name_Designated_Type_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_169 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_168 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Ada_Node_P_Nested_Generic_Formal_Part'Access,
               Prioritary,
               All_Cats,
               True);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_171 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_170 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Ada_Node_P_Use_Clauses_In_Generic_Formal_Part'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_173 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_172 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Body_Node_P_Subunit_Decl_Env'Access,
               Prioritary,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_175 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_174 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Body_Node_P_Body_Decl_Scope'Access,
               Prioritary,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_177 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_176 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Package_Body_P_Package_Decl_Uses_Clauses_Envs'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      end;
      procedure Initialize_Fields_For_Package_Body
        (Self : Bare_Package_Body
         ; Package_Body_F_Package_Name : Bare_Defining_Name
         ; Package_Body_F_Aspects : Bare_Aspect_Spec
         ; Package_Body_F_Decls : Bare_Declarative_Part
         ; Package_Body_F_Stmts : Bare_Handled_Stmts
         ; Package_Body_F_End_Name : Bare_End_Name
        ) is
      begin
            Self.Package_Body_F_Package_Name := Package_Body_F_Package_Name;
            Self.Package_Body_F_Aspects := Package_Body_F_Aspects;
            Self.Package_Body_F_Decls := Package_Body_F_Decls;
            Self.Package_Body_F_Stmts := Package_Body_F_Stmts;
            Self.Package_Body_F_End_Name := Package_Body_F_End_Name;
      end Initialize_Fields_For_Package_Body;
   function Package_Body_F_Package_Name
     (Node : Bare_Package_Body) return Bare_Defining_Name
   is
   begin
         return Node.Package_Body_F_Package_Name;
   end;
   function Package_Body_F_Decls
     (Node : Bare_Package_Body) return Bare_Declarative_Part
   is
   begin
         return Node.Package_Body_F_Decls;
   end;
   function Package_Body_F_Stmts
     (Node : Bare_Package_Body) return Bare_Handled_Stmts
   is
   begin
         return Node.Package_Body_F_Stmts;
   end;
   function Package_Body_F_End_Name
     (Node : Bare_Package_Body) return Bare_End_Name
   is
   begin
         return Node.Package_Body_F_End_Name;
   end;
--# property-start PackageBody.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19778
pragma Warnings (Off, "is not referenced");
 function Package_Body_P_Defining_Names
  (Node : Bare_Package_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Package_Body :=
     Internal_Entity_Package_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1375 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1375 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1375;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4812 '<FieldAccess for package_name at ast.py:19778>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19778
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Package_Body_F_Package_Name, Info => Ent.Info);
--# expr-done 4812
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1375;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1375;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Package_Body_P_Defining_Names;
--# end
--# property-start PackageBody.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19779
pragma Warnings (Off, "is not referenced");
 function Package_Body_P_Defining_Env
  (Node : Bare_Package_Body
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Package_Body :=
     Internal_Entity_Package_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1376 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1376 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1376;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4813 '<FieldAccess for children_env at ast.py:19779>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19779
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 4813
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1376;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1376;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Package_Body_P_Defining_Env;
--# end
--# property-start PackageBody.declarative_parts /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19781
pragma Warnings (Off, "is not referenced");
 function Package_Body_P_Declarative_Parts
  (Node : Bare_Package_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Declarative_Part_Array_Access
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Package_Body :=
     Internal_Entity_Package_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_1377 with Inline_Always;
      Fld : Internal_Entity_Declarative_Part;
Singleton : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_1377 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1377;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4814 '<FieldAccess for decls at ast.py:19781>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19781
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Declarative_Part (Node => Ent.Node.Package_Body_F_Decls, Info => Ent.Info);
--# expr-done 4814
                Singleton := Create_Internal_Entity_Declarative_Part_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1377;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1377;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Package_Body_P_Declarative_Parts;
--# end
--# property-start PackageBody.package_decl_uses_clauses_envs /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19784
pragma Warnings (Off, "is not referenced");
 function Package_Body_P_Package_Decl_Uses_Clauses_Envs
  (Node : Bare_Package_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Package_Body :=
     Internal_Entity_Package_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1378 with Inline_Always;
            procedure Finalizer_Scope_2425 with Inline_Always;
      Block_Var : Internal_Entity_Base_Package_Decl;
Var_Expr : Internal_Entity_Private_Part;
Bound_Imprecise_Fallback : Boolean;
Fld : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity_Base_Package_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Base_Package_Decl;
Fld_1 : Internal_Entity_Public_Part;
Fld_2 : Lexical_Env := Empty_Env;
Fld_3 : Internal_Entity_Private_Part;
Fld_4 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Array_Lit : Lexical_Env_Array_Access;
Group_Env : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1378 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Group_Env);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1378;
            procedure Finalizer_Scope_2425 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_2425;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4824 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4817 '<bind at ast.py:19789>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19789
Bound_Imprecise_Fallback := False; 
--# expr-start 4816 '<Cast to BasePackageDecl at ast.py:19790>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19790
--# expr-start 4815 '<FieldAccess for decl_part at ast.py:19790>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19790
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.decl_part
Fld := Libadalang.Implementation.Body_Node_P_Decl_Part (Node => Ent.Node, Imprecise_Fallback => Bound_Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 4815
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Base_Package_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Package_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 4816
Dyn_Var_Bind_Result := Cast_Result; 
--# expr-done 4817
Block_Var := Dyn_Var_Bind_Result; 
--# bind pd Block_Var
--# expr-start 4823 '<env_group at ast.py:19793>' Group_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19793
--# expr-start 4822 '<ArrayLiteral at ast.py:19793>' Array_Lit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19793
--# expr-start 4819 '<FieldAccess for use_clauses_envs at ast.py:19793>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19793
--# expr-start 4818 '<FieldAccess for public_part at ast.py:19793>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19793
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Public_Part (Node => Block_Var.Node.Base_Package_Decl_F_Public_Part, Info => Block_Var.Info);
--# expr-done 4818
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DeclarativePart.use_clauses_envs
Fld_2 := Libadalang.Implementation.Declarative_Part_P_Use_Clauses_Envs (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 4819
--# expr-start 4820 '<FieldAccess for private_part at ast.py:19794>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19794
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Private_Part (Node => Block_Var.Node.Base_Package_Decl_F_Private_Part, Info => Block_Var.Info);
--# expr-done 4820
Var_Expr := Fld_3;
if Var_Expr /= No_Entity_Private_Part then
   --# scope-start
--# expr-start 4821 '<FieldAccess for use_clauses_envs at ast.py:19794>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19794
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DeclarativePart.use_clauses_envs
Fld_4 := Libadalang.Implementation.Declarative_Part_P_Use_Clauses_Envs (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 4821
Scope_Result := Fld_4;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2425;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
Array_Lit := Create_Lexical_Env_Array (Internal_Lexical_Env_Array'(1 => Fld_2, 2 => Result_Var)); 
--# expr-done 4822
Group_Env := Group (Array_Lit, No_Metadata); 
--# expr-done 4823
Let_Result := Group_Env; Inc_Ref (Let_Result);
--# end
--# expr-done 4824
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1378;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1378;
               Finalizer_Scope_2425;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Package_Body_P_Package_Decl_Uses_Clauses_Envs;
--# end
--# property-start PackageBody._env_do_159 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19697
pragma Warnings (Off, "is not referenced");
 function Env_Do_159
  (Node : Bare_Package_Body
  )
   return Boolean
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4825 '<FieldAccess for env_hook at ast.py:19697>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19697
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.env_hook
Fld := Libadalang.Implementation.Ada_Node_P_Env_Hook (Node => Self);
--# end
--# expr-done 4825
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_159;
--# end
--# property-start PackageBody._initial_env_160 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19700
pragma Warnings (Off, "is not referenced");
 function Initial_Env_160
  (Node : Bare_Package_Body
  )
   return Internal_Designated_Env
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_1577 with Inline_Always;
      Fld : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_1577 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1577;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4826 '<FieldAccess for body_initial_env at ast.py:19700>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19700
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.body_initial_env
Fld := Libadalang.Implementation.Body_Node_P_Body_Initial_Env (Node => Self);
--# end
--# expr-done 4826
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1577;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1577;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Initial_Env_160;
--# end
--# property-start PackageBody._env_mappings_161 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19702
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_161
  (Node : Bare_Package_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Package_Body :=
     Internal_Entity_Package_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1578 with Inline_Always;
      Fld : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1578 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1578;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4827 '<FieldAccess for previous_part_link_env_assoc at ast.py:19702>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19702
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Body.previous_part_link_env_assoc'
Fld := Libadalang.Implementation.Dispatcher_Body_Node_P_Previous_Part_Link_Env_Assoc (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4827
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1578;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1578;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_161;
--# end
--# property-start PackageBody._env_trans_parent_162 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19706
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_162
  (Node : Bare_Package_Body
  )
   return Boolean
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4828 '<FieldAccess for is_library_item at ast.py:19706>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19706
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_library_item'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Library_Item (Node => Self);
--# end
--# expr-done 4828
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_162;
--# end
--# property-start PackageBody._env_do_163 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19708
pragma Warnings (Off, "is not referenced");
 function Env_Do_163
  (Node : Bare_Package_Body
  )
   return Bare_Compilation_Unit_Array_Access
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1580 with Inline_Always;
      Fld : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1580 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1580;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4829 '<FieldAccess for populate_dependent_units at ast.py:19708>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19708
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.populate_dependent_units
Fld := Libadalang.Implementation.Basic_Decl_P_Populate_Dependent_Units (Node => Self);
--# end
--# expr-done 4829
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1580;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1580;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_163;
--# end
--# property-start PackageBody._ref_env_nodes_164 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19711
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_164
  (Node : Bare_Package_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1581 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1581 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1581;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4830 '<FieldAccess for top_level_use_package_clauses at ast.py:19711>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19711
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_package_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Package_Clauses (Node => Self);
--# end
--# expr-done 4830
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1581;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1581;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_164;
--# end
--# property-start PackageBody._ref_cond_165 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19713
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_165
  (Node : Bare_Package_Body
  )
   return Boolean
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4832 '<IsA LibraryItem, Subunit at ast.py:19713>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19713
--# expr-start 4831 '<parent at ast.py:19713>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19713
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 4831
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 4832
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_165;
--# end
--# property-start PackageBody._ref_env_nodes_166 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19717
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_166
  (Node : Bare_Package_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1583 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1583 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1583;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4833 '<FieldAccess for top_level_use_type_clauses at ast.py:19717>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19717
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_type_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Type_Clauses (Node => Self);
--# end
--# expr-done 4833
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1583;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1583;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_166;
--# end
--# property-start PackageBody._ref_cond_167 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19719
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_167
  (Node : Bare_Package_Body
  )
   return Boolean
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4835 '<IsA LibraryItem, Subunit at ast.py:19719>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19719
--# expr-start 4834 '<parent at ast.py:19719>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19719
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 4834
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 4835
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_167;
--# end
--# property-start PackageBody._ref_env_nodes_168 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19723
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_168
  (Node : Bare_Package_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1585 with Inline_Always;
            procedure Finalizer_Scope_2426 with Inline_Always;
      Var_Expr : Bare_Ada_Node;
Cast_Expr : Bare_Package_Body;
Cast_Result : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
Scope_Result : Bare_Ada_Node_Array_Access;
Result_Var : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1585 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1585;
            procedure Finalizer_Scope_2426 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2426;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 4836 '<Cast to AdaNode at ast.py:19723>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19723
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-done 4836
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2426;
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node_Array_Type;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1585;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1585;
               Finalizer_Scope_2426;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_168;
--# end
--# property-start PackageBody._ref_cond_169 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19725
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_169
  (Node : Bare_Package_Body
  )
   return Boolean
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4837 '<FieldAccess for should_ref_generic_formals at ast.py:19725>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19725
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.should_ref_generic_formals
Fld := Libadalang.Implementation.Basic_Decl_P_Should_Ref_Generic_Formals (Node => Self);
--# end
--# expr-done 4837
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_169;
--# end
--# property-start PackageBody._ref_env_nodes_170 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19733
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_170
  (Node : Bare_Package_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1587 with Inline_Always;
            procedure Finalizer_Scope_2427 with Inline_Always;
      Var_Expr : Bare_Ada_Node;
Cast_Expr : Bare_Package_Body;
Cast_Result : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
Scope_Result : Bare_Ada_Node_Array_Access;
Result_Var : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1587 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1587;
            procedure Finalizer_Scope_2427 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2427;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 4838 '<Cast to AdaNode at ast.py:19733>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19733
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-done 4838
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2427;
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node_Array_Type;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1587;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1587;
               Finalizer_Scope_2427;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_170;
--# end
--# property-start PackageBody._ref_cond_171 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19735
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_171
  (Node : Bare_Package_Body
  )
   return Boolean
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4839 '<FieldAccess for should_ref_generic_formals at ast.py:19735>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19735
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.should_ref_generic_formals
Fld := Libadalang.Implementation.Basic_Decl_P_Should_Ref_Generic_Formals (Node => Self);
--# end
--# expr-done 4839
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_171;
--# end
--# property-start PackageBody._ref_env_nodes_172 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19746
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_172
  (Node : Bare_Package_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1589 with Inline_Always;
      Cast_Expr : Bare_Package_Body;
Cast_Result : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1589 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1589;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4840 '<Cast to AdaNode at ast.py:19746>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19746
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-done 4840
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Cast_Result;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1589;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1589;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_172;
--# end
--# property-start PackageBody._ref_cond_173 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19748
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_173
  (Node : Bare_Package_Body
  )
   return Boolean
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4841 '<FieldAccess for is_subunit at ast.py:19748>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19748
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.is_subunit
Fld := Libadalang.Implementation.Body_Node_P_Is_Subunit (Node => Self);
--# end
--# expr-done 4841
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_173;
--# end
--# property-start PackageBody._ref_env_nodes_174 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19760
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_174
  (Node : Bare_Package_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1591 with Inline_Always;
      Cast_Expr : Bare_Package_Body;
Cast_Result : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1591 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1591;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4842 '<Cast to AdaNode at ast.py:19760>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19760
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-done 4842
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Cast_Result;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1591;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1591;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_174;
--# end
--# property-start PackageBody._ref_cond_175 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19762
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_175
  (Node : Bare_Package_Body
  )
   return Boolean
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
Not_Val : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4844 '<Not at ast.py:19762>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19762
--# expr-start 4843 '<FieldAccess for is_compilation_unit_root at ast.py:19762>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19762
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_compilation_unit_root
Fld := Libadalang.Implementation.Basic_Decl_P_Is_Compilation_Unit_Root (Node => Self);
--# end
--# expr-done 4843
Not_Val := not (Fld); 
--# expr-done 4844
      Property_Result := Not_Val;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_175;
--# end
--# property-start PackageBody._ref_env_nodes_176 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19767
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_176
  (Node : Bare_Package_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1593 with Inline_Always;
      Cast_Expr : Bare_Package_Body;
Cast_Result : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1593 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1593;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4845 '<Cast to AdaNode at ast.py:19767>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19767
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-done 4845
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Cast_Result;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1593;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1593;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_176;
--# end
--# property-start PackageBody._ref_cond_177 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19769
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_177
  (Node : Bare_Package_Body
  )
   return Boolean
is
   Self : Bare_Package_Body := Bare_Package_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
Not_Val : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4847 '<Not at ast.py:19769>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19769
--# expr-start 4846 '<FieldAccess for is_compilation_unit_root at ast.py:19769>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19769
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_compilation_unit_root
Fld := Libadalang.Implementation.Basic_Decl_P_Is_Compilation_Unit_Root (Node => Self);
--# end
--# expr-done 4846
Not_Val := not (Fld); 
--# expr-done 4847
      Property_Result := Not_Val;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_177;
--# end
function Package_Body_P_Package_Decl_Uses_Clauses_Envs
  (E : Entity
  ) return Lexical_Env
is
      E_Info : Internal_Entity_Info :=
         Shed_Rebindings (E.Info, Node_Env (E.Node));
   Result : constant Lexical_Env :=
      Package_Body_P_Package_Decl_Uses_Clauses_Envs (E.Node, E_Info);
begin
      return Result;
end;
   --
   --  Primitives for Bare_Protected_Body
   --
      procedure Protected_Body_Pre_Env_Actions
        (Self            : Bare_Protected_Body;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Dummy : Boolean;
      begin
         Dummy := Libadalang.Implementation.Env_Do_194 (Node => 
Self);
      end;
      declare
         Env : constant Internal_Designated_Env :=
           Libadalang.Implementation.Initial_Env_195 (Node => 
Self);
      begin
         Set_Initial_Env
           (Self,
            State,
            Env,
            DSL_Location => "ast.py:19882");
      end;
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_196 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:19884");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_197 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      declare
         Dummy : Bare_Compilation_Unit_Array_Access;
      begin
         Dummy := Libadalang.Implementation.Env_Do_198 (Node => 
Self);
         Dec_Ref (Dummy);
      end;
      if Libadalang.Implementation.Ref_Cond_200 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_199 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Use_Package_Name_Designated_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_202 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_201 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Name_Designated_Type_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_204 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_203 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Ada_Node_P_Nested_Generic_Formal_Part'Access,
               Prioritary,
               All_Cats,
               True);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_206 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_205 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Ada_Node_P_Use_Clauses_In_Generic_Formal_Part'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_207 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Body_Node_P_Body_Decl_Scope'Access,
               Prioritary,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      if Libadalang.Implementation.Ref_Cond_209 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_208 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Body_Node_P_Subunit_Decl_Env'Access,
               Prioritary,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      end;
      procedure Initialize_Fields_For_Protected_Body
        (Self : Bare_Protected_Body
         ; Protected_Body_F_Name : Bare_Defining_Name
         ; Protected_Body_F_Aspects : Bare_Aspect_Spec
         ; Protected_Body_F_Decls : Bare_Declarative_Part
         ; Protected_Body_F_End_Name : Bare_End_Name
        ) is
      begin
            Self.Protected_Body_F_Name := Protected_Body_F_Name;
            Self.Protected_Body_F_Aspects := Protected_Body_F_Aspects;
            Self.Protected_Body_F_Decls := Protected_Body_F_Decls;
            Self.Protected_Body_F_End_Name := Protected_Body_F_End_Name;
      end Initialize_Fields_For_Protected_Body;
   function Protected_Body_F_Name
     (Node : Bare_Protected_Body) return Bare_Defining_Name
   is
   begin
         return Node.Protected_Body_F_Name;
   end;
   function Protected_Body_F_Decls
     (Node : Bare_Protected_Body) return Bare_Declarative_Part
   is
   begin
         return Node.Protected_Body_F_Decls;
   end;
   function Protected_Body_F_End_Name
     (Node : Bare_Protected_Body) return Bare_End_Name
   is
   begin
         return Node.Protected_Body_F_End_Name;
   end;
--# property-start ProtectedBody.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19933
pragma Warnings (Off, "is not referenced");
 function Protected_Body_P_Defining_Names
  (Node : Bare_Protected_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Protected_Body := Bare_Protected_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Protected_Body :=
     Internal_Entity_Protected_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1383 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1383 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1383;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4848 '<FieldAccess for name at ast.py:19933>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19933
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Protected_Body_F_Name, Info => Ent.Info);
--# expr-done 4848
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1383;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1383;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Protected_Body_P_Defining_Names;
--# end
--# property-start ProtectedBody.declarative_parts /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19935
pragma Warnings (Off, "is not referenced");
 function Protected_Body_P_Declarative_Parts
  (Node : Bare_Protected_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Declarative_Part_Array_Access
is
   Self : Bare_Protected_Body := Bare_Protected_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Protected_Body :=
     Internal_Entity_Protected_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_1384 with Inline_Always;
      Fld : Internal_Entity_Declarative_Part;
Singleton : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_1384 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1384;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4849 '<FieldAccess for decls at ast.py:19935>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19935
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Declarative_Part (Node => Ent.Node.Protected_Body_F_Decls, Info => Ent.Info);
--# expr-done 4849
                Singleton := Create_Internal_Entity_Declarative_Part_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1384;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1384;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Protected_Body_P_Declarative_Parts;
--# end
--# property-start ProtectedBody._env_do_194 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19880
pragma Warnings (Off, "is not referenced");
 function Env_Do_194
  (Node : Bare_Protected_Body
  )
   return Boolean
is
   Self : Bare_Protected_Body := Bare_Protected_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4850 '<FieldAccess for env_hook at ast.py:19880>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19880
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.env_hook
Fld := Libadalang.Implementation.Ada_Node_P_Env_Hook (Node => Self);
--# end
--# expr-done 4850
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_194;
--# end
--# property-start ProtectedBody._initial_env_195 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19882
pragma Warnings (Off, "is not referenced");
 function Initial_Env_195
  (Node : Bare_Protected_Body
  )
   return Internal_Designated_Env
is
   Self : Bare_Protected_Body := Bare_Protected_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_1612 with Inline_Always;
      Fld : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_1612 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1612;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4851 '<FieldAccess for body_initial_env at ast.py:19882>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19882
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.body_initial_env
Fld := Libadalang.Implementation.Body_Node_P_Body_Initial_Env (Node => Self);
--# end
--# expr-done 4851
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1612;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1612;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Initial_Env_195;
--# end
--# property-start ProtectedBody._env_mappings_196 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19884
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_196
  (Node : Bare_Protected_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Protected_Body := Bare_Protected_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Protected_Body :=
     Internal_Entity_Protected_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1613 with Inline_Always;
      Fld : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1613 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1613;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4852 '<FieldAccess for previous_part_link_env_assoc at ast.py:19884>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19884
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Body.previous_part_link_env_assoc'
Fld := Libadalang.Implementation.Dispatcher_Body_Node_P_Previous_Part_Link_Env_Assoc (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4852
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1613;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1613;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_196;
--# end
--# property-start ProtectedBody._env_trans_parent_197 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19879
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_197
  (Node : Bare_Protected_Body
  )
   return Boolean
is
   Self : Bare_Protected_Body := Bare_Protected_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_197;
--# end
--# property-start ProtectedBody._env_do_198 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19888
pragma Warnings (Off, "is not referenced");
 function Env_Do_198
  (Node : Bare_Protected_Body
  )
   return Bare_Compilation_Unit_Array_Access
is
   Self : Bare_Protected_Body := Bare_Protected_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1615 with Inline_Always;
      Fld : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1615 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1615;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4853 '<FieldAccess for populate_dependent_units at ast.py:19888>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19888
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.populate_dependent_units
Fld := Libadalang.Implementation.Basic_Decl_P_Populate_Dependent_Units (Node => Self);
--# end
--# expr-done 4853
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1615;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1615;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_198;
--# end
--# property-start ProtectedBody._ref_env_nodes_199 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19891
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_199
  (Node : Bare_Protected_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Protected_Body := Bare_Protected_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1616 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1616 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1616;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4854 '<FieldAccess for top_level_use_package_clauses at ast.py:19891>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19891
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_package_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Package_Clauses (Node => Self);
--# end
--# expr-done 4854
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1616;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1616;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_199;
--# end
--# property-start ProtectedBody._ref_cond_200 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19893
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_200
  (Node : Bare_Protected_Body
  )
   return Boolean
is
   Self : Bare_Protected_Body := Bare_Protected_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4856 '<IsA LibraryItem, Subunit at ast.py:19893>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19893
--# expr-start 4855 '<parent at ast.py:19893>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19893
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 4855
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 4856
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_200;
--# end
--# property-start ProtectedBody._ref_env_nodes_201 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19896
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_201
  (Node : Bare_Protected_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Protected_Body := Bare_Protected_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1618 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1618 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1618;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4857 '<FieldAccess for top_level_use_type_clauses at ast.py:19896>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19896
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_type_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Type_Clauses (Node => Self);
--# end
--# expr-done 4857
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1618;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1618;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_201;
--# end
--# property-start ProtectedBody._ref_cond_202 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19898
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_202
  (Node : Bare_Protected_Body
  )
   return Boolean
is
   Self : Bare_Protected_Body := Bare_Protected_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4859 '<IsA LibraryItem, Subunit at ast.py:19898>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19898
--# expr-start 4858 '<parent at ast.py:19898>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19898
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 4858
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 4859
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_202;
--# end
--# property-start ProtectedBody._ref_env_nodes_203 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19901
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_203
  (Node : Bare_Protected_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Protected_Body := Bare_Protected_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1620 with Inline_Always;
            procedure Finalizer_Scope_2428 with Inline_Always;
      Var_Expr : Bare_Ada_Node;
Cast_Expr : Bare_Protected_Body;
Cast_Result : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
Scope_Result : Bare_Ada_Node_Array_Access;
Result_Var : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1620 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1620;
            procedure Finalizer_Scope_2428 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2428;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 4860 '<Cast to AdaNode at ast.py:19901>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19901
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-done 4860
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2428;
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node_Array_Type;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1620;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1620;
               Finalizer_Scope_2428;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_203;
--# end
--# property-start ProtectedBody._ref_cond_204 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19903
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_204
  (Node : Bare_Protected_Body
  )
   return Boolean
is
   Self : Bare_Protected_Body := Bare_Protected_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4861 '<FieldAccess for should_ref_generic_formals at ast.py:19903>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19903
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.should_ref_generic_formals
Fld := Libadalang.Implementation.Basic_Decl_P_Should_Ref_Generic_Formals (Node => Self);
--# end
--# expr-done 4861
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_204;
--# end
--# property-start ProtectedBody._ref_env_nodes_205 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19911
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_205
  (Node : Bare_Protected_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Protected_Body := Bare_Protected_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1622 with Inline_Always;
            procedure Finalizer_Scope_2429 with Inline_Always;
      Var_Expr : Bare_Ada_Node;
Cast_Expr : Bare_Protected_Body;
Cast_Result : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
Scope_Result : Bare_Ada_Node_Array_Access;
Result_Var : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1622 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1622;
            procedure Finalizer_Scope_2429 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2429;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 4862 '<Cast to AdaNode at ast.py:19911>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19911
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-done 4862
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2429;
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node_Array_Type;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1622;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1622;
               Finalizer_Scope_2429;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_205;
--# end
--# property-start ProtectedBody._ref_cond_206 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19913
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_206
  (Node : Bare_Protected_Body
  )
   return Boolean
is
   Self : Bare_Protected_Body := Bare_Protected_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4863 '<FieldAccess for should_ref_generic_formals at ast.py:19913>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19913
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.should_ref_generic_formals
Fld := Libadalang.Implementation.Basic_Decl_P_Should_Ref_Generic_Formals (Node => Self);
--# end
--# expr-done 4863
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_206;
--# end
--# property-start ProtectedBody._ref_env_nodes_207 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19917
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_207
  (Node : Bare_Protected_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Protected_Body := Bare_Protected_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1624 with Inline_Always;
      Cast_Expr : Bare_Protected_Body;
Cast_Result : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1624 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1624;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4864 '<Cast to AdaNode at ast.py:19917>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19917
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-done 4864
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Cast_Result;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1624;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1624;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_207;
--# end
--# property-start ProtectedBody._ref_env_nodes_208 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19922
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_208
  (Node : Bare_Protected_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Protected_Body := Bare_Protected_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1625 with Inline_Always;
      Cast_Expr : Bare_Protected_Body;
Cast_Result : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1625 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1625;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4865 '<Cast to AdaNode at ast.py:19922>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19922
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-done 4865
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Cast_Result;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1625;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1625;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_208;
--# end
--# property-start ProtectedBody._ref_cond_209 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19924
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_209
  (Node : Bare_Protected_Body
  )
   return Boolean
is
   Self : Bare_Protected_Body := Bare_Protected_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4866 '<FieldAccess for is_subunit at ast.py:19924>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19924
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.is_subunit
Fld := Libadalang.Implementation.Body_Node_P_Is_Subunit (Node => Self);
--# end
--# expr-done 4866
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_209;
--# end
   --
   --  Primitives for Bare_Task_Body
   --
      procedure Task_Body_Pre_Env_Actions
        (Self            : Bare_Task_Body;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Dummy : Boolean;
      begin
         Dummy := Libadalang.Implementation.Env_Do_178 (Node => 
Self);
      end;
      declare
         Env : constant Internal_Designated_Env :=
           Libadalang.Implementation.Initial_Env_179 (Node => 
Self);
      begin
         Set_Initial_Env
           (Self,
            State,
            Env,
            DSL_Location => "ast.py:19815");
      end;
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_180 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:19817");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_181 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      declare
         Dummy : Bare_Compilation_Unit_Array_Access;
      begin
         Dummy := Libadalang.Implementation.Env_Do_182 (Node => 
Self);
         Dec_Ref (Dummy);
      end;
      if Libadalang.Implementation.Ref_Cond_184 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_183 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Use_Package_Name_Designated_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_186 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_185 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Name_Designated_Type_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_188 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_187 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Ada_Node_P_Nested_Generic_Formal_Part'Access,
               Prioritary,
               All_Cats,
               True);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_190 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_189 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Ada_Node_P_Use_Clauses_In_Generic_Formal_Part'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_191 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Task_Body_P_Task_Type_Decl_Scope'Access,
               Prioritary,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      if Libadalang.Implementation.Ref_Cond_193 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_192 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Body_Node_P_Subunit_Decl_Env'Access,
               Prioritary,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      end;
      procedure Initialize_Fields_For_Task_Body
        (Self : Bare_Task_Body
         ; Task_Body_F_Name : Bare_Defining_Name
         ; Task_Body_F_Aspects : Bare_Aspect_Spec
         ; Task_Body_F_Decls : Bare_Declarative_Part
         ; Task_Body_F_Stmts : Bare_Handled_Stmts
         ; Task_Body_F_End_Name : Bare_End_Name
        ) is
      begin
            Self.Task_Body_F_Name := Task_Body_F_Name;
            Self.Task_Body_F_Aspects := Task_Body_F_Aspects;
            Self.Task_Body_F_Decls := Task_Body_F_Decls;
            Self.Task_Body_F_Stmts := Task_Body_F_Stmts;
            Self.Task_Body_F_End_Name := Task_Body_F_End_Name;
      end Initialize_Fields_For_Task_Body;
   function Task_Body_F_Name
     (Node : Bare_Task_Body) return Bare_Defining_Name
   is
   begin
         return Node.Task_Body_F_Name;
   end;
   function Task_Body_F_Decls
     (Node : Bare_Task_Body) return Bare_Declarative_Part
   is
   begin
         return Node.Task_Body_F_Decls;
   end;
   function Task_Body_F_Stmts
     (Node : Bare_Task_Body) return Bare_Handled_Stmts
   is
   begin
         return Node.Task_Body_F_Stmts;
   end;
   function Task_Body_F_End_Name
     (Node : Bare_Task_Body) return Bare_End_Name
   is
   begin
         return Node.Task_Body_F_End_Name;
   end;
--# property-start TaskBody.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19808
pragma Warnings (Off, "is not referenced");
 function Task_Body_P_Defining_Names
  (Node : Bare_Task_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Task_Body :=
     Internal_Entity_Task_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1379 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1379 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1379;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4867 '<FieldAccess for name at ast.py:19808>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19808
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Task_Body_F_Name, Info => Ent.Info);
--# expr-done 4867
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1379;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1379;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Task_Body_P_Defining_Names;
--# end
--# property-start TaskBody.declarative_parts /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19810
pragma Warnings (Off, "is not referenced");
 function Task_Body_P_Declarative_Parts
  (Node : Bare_Task_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Declarative_Part_Array_Access
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Task_Body :=
     Internal_Entity_Task_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_1380 with Inline_Always;
      Fld : Internal_Entity_Declarative_Part;
Singleton : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_1380 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1380;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4868 '<FieldAccess for decls at ast.py:19810>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19810
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Declarative_Part (Node => Ent.Node.Task_Body_F_Decls, Info => Ent.Info);
--# expr-done 4868
                Singleton := Create_Internal_Entity_Declarative_Part_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1380;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1380;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Task_Body_P_Declarative_Parts;
--# end
--# property-start TaskBody.task_type_decl_scope /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19863
pragma Warnings (Off, "is not referenced");
 function Task_Body_P_Task_Type_Decl_Scope
  (Node : Bare_Task_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Task_Body :=
     Internal_Entity_Task_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1381 with Inline_Always;
      Fld : Internal_Entity_Task_Type_Decl;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1381 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1381;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4870 '<FieldAccess for children_env at ast.py:19863>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19863
--# expr-start 4869 '<FieldAccess for task_type at ast.py:19863>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19863
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TaskBody.task_type
Fld := Libadalang.Implementation.Task_Body_P_Task_Type (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4869
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Children_Env (Node => Fld.Node, E_Info => Fld.Info);
--# expr-done 4870
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1381;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1381;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Task_Body_P_Task_Type_Decl_Scope;
--# end
--# property-start TaskBody.task_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19866
pragma Warnings (Off, "is not referenced");
 function Task_Body_P_Task_Type
  (Node : Bare_Task_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Task_Type_Decl
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Task_Body :=
     Internal_Entity_Task_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Task_Type_Decl;
      Match_0 : Internal_Entity_Task_Type_Decl;
Match_1 : Internal_Entity_Single_Task_Decl;
Match_2 : Internal_Entity_Basic_Decl;
Bound_Imprecise_Fallback : Boolean;
Fld : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Single_Task_Type_Decl;
Match_Prefix : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity_Task_Type_Decl;
Let_Result : Internal_Entity_Task_Type_Decl;
Scope_Result : Internal_Entity_Task_Type_Decl;
Cast_Result_1 : Internal_Entity_Single_Task_Decl;
Cast_Result_2 : Internal_Entity_Task_Type_Decl;
Let_Result_1 : Internal_Entity_Task_Type_Decl;
Scope_Result_1 : Internal_Entity_Task_Type_Decl;
Cast_Result_3 : Internal_Entity_Basic_Decl;
Let_Result_2 : Internal_Entity_Task_Type_Decl;
Scope_Result_2 : Internal_Entity_Task_Type_Decl;
Match_Result : Internal_Entity_Task_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Task_Type_Decl;
Let_Result_3 : Internal_Entity_Task_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4875 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 4874 '<bind at ast.py:19867>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19867
Bound_Imprecise_Fallback := False; 
--# expr-start 4873 '<Match at ast.py:19867>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19867
--# expr-start 4871 '<FieldAccess for decl_part at ast.py:19867>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19867
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.decl_part
Fld := Libadalang.Implementation.Body_Node_P_Decl_Part (Node => Ent.Node, Imprecise_Fallback => Bound_Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 4871
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld; 
case Ada_Basic_Decl (Match_Prefix.Node.Kind) is
         when Ada_Task_Type_Decl .. Ada_Single_Task_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Task_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind t Match_0
Let_Result := Match_0; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Single_Task_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind t Match_1
--# expr-start 4872 '<FieldAccess for task_type at ast.py:19869>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19869
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Single_Task_Type_Decl (Node => Match_1.Node.Single_Task_Decl_F_Task_Type, Info => Match_1.Info);
--# expr-done 4872
      Cast_Result_2 := Create_Internal_Entity_Task_Type_Decl
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
Let_Result_1 := Cast_Result_2; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abstract_State_Decl .. Ada_Protected_Type_Decl | Ada_Type_Decl .. Ada_Single_Protected_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Basic_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_3; 
--# bind _ Match_2
Let_Result_2 := raise Property_Error with "Should not happen"; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 4873
Dyn_Var_Bind_Result := Match_Result; 
--# expr-done 4874
Let_Result_3 := Dyn_Var_Bind_Result; 
--# end
--# expr-done 4875
      Property_Result := Let_Result_3;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Task_Body_P_Task_Type;
--# end
--# property-start TaskBody._env_do_178 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19813
pragma Warnings (Off, "is not referenced");
 function Env_Do_178
  (Node : Bare_Task_Body
  )
   return Boolean
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4876 '<FieldAccess for env_hook at ast.py:19813>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19813
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.env_hook
Fld := Libadalang.Implementation.Ada_Node_P_Env_Hook (Node => Self);
--# end
--# expr-done 4876
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_178;
--# end
--# property-start TaskBody._initial_env_179 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19815
pragma Warnings (Off, "is not referenced");
 function Initial_Env_179
  (Node : Bare_Task_Body
  )
   return Internal_Designated_Env
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_1596 with Inline_Always;
      Fld : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_1596 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1596;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4877 '<FieldAccess for body_initial_env at ast.py:19815>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19815
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.body_initial_env
Fld := Libadalang.Implementation.Body_Node_P_Body_Initial_Env (Node => Self);
--# end
--# expr-done 4877
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1596;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1596;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Initial_Env_179;
--# end
--# property-start TaskBody._env_mappings_180 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19817
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_180
  (Node : Bare_Task_Body
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Task_Body :=
     Internal_Entity_Task_Body'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1597 with Inline_Always;
      Fld : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1597 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1597;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4878 '<FieldAccess for previous_part_link_env_assoc at ast.py:19817>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19817
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Body.previous_part_link_env_assoc'
Fld := Libadalang.Implementation.Dispatcher_Body_Node_P_Previous_Part_Link_Env_Assoc (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4878
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1597;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1597;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_180;
--# end
--# property-start TaskBody._env_trans_parent_181 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19812
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_181
  (Node : Bare_Task_Body
  )
   return Boolean
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_181;
--# end
--# property-start TaskBody._env_do_182 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19821
pragma Warnings (Off, "is not referenced");
 function Env_Do_182
  (Node : Bare_Task_Body
  )
   return Bare_Compilation_Unit_Array_Access
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1599 with Inline_Always;
      Fld : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1599 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1599;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4879 '<FieldAccess for populate_dependent_units at ast.py:19821>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19821
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.populate_dependent_units
Fld := Libadalang.Implementation.Basic_Decl_P_Populate_Dependent_Units (Node => Self);
--# end
--# expr-done 4879
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1599;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1599;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_182;
--# end
--# property-start TaskBody._ref_env_nodes_183 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19824
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_183
  (Node : Bare_Task_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1600 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1600 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1600;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4880 '<FieldAccess for top_level_use_package_clauses at ast.py:19824>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19824
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_package_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Package_Clauses (Node => Self);
--# end
--# expr-done 4880
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1600;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1600;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_183;
--# end
--# property-start TaskBody._ref_cond_184 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19826
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_184
  (Node : Bare_Task_Body
  )
   return Boolean
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4882 '<IsA LibraryItem, Subunit at ast.py:19826>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19826
--# expr-start 4881 '<parent at ast.py:19826>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19826
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 4881
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 4882
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_184;
--# end
--# property-start TaskBody._ref_env_nodes_185 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19830
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_185
  (Node : Bare_Task_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1602 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1602 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1602;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4883 '<FieldAccess for top_level_use_type_clauses at ast.py:19830>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19830
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_type_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Type_Clauses (Node => Self);
--# end
--# expr-done 4883
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1602;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1602;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_185;
--# end
--# property-start TaskBody._ref_cond_186 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19832
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_186
  (Node : Bare_Task_Body
  )
   return Boolean
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4885 '<IsA LibraryItem, Subunit at ast.py:19832>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19832
--# expr-start 4884 '<parent at ast.py:19832>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19832
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 4884
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 4885
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_186;
--# end
--# property-start TaskBody._ref_env_nodes_187 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19836
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_187
  (Node : Bare_Task_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1604 with Inline_Always;
            procedure Finalizer_Scope_2433 with Inline_Always;
      Var_Expr : Bare_Ada_Node;
Cast_Expr : Bare_Task_Body;
Cast_Result : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
Scope_Result : Bare_Ada_Node_Array_Access;
Result_Var : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1604 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1604;
            procedure Finalizer_Scope_2433 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2433;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 4886 '<Cast to AdaNode at ast.py:19836>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19836
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-done 4886
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2433;
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node_Array_Type;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1604;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1604;
               Finalizer_Scope_2433;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_187;
--# end
--# property-start TaskBody._ref_cond_188 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19838
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_188
  (Node : Bare_Task_Body
  )
   return Boolean
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4887 '<FieldAccess for should_ref_generic_formals at ast.py:19838>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19838
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.should_ref_generic_formals
Fld := Libadalang.Implementation.Basic_Decl_P_Should_Ref_Generic_Formals (Node => Self);
--# end
--# expr-done 4887
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_188;
--# end
--# property-start TaskBody._ref_env_nodes_189 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19846
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_189
  (Node : Bare_Task_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1606 with Inline_Always;
            procedure Finalizer_Scope_2434 with Inline_Always;
      Var_Expr : Bare_Ada_Node;
Cast_Expr : Bare_Task_Body;
Cast_Result : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
Scope_Result : Bare_Ada_Node_Array_Access;
Result_Var : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1606 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1606;
            procedure Finalizer_Scope_2434 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2434;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 4888 '<Cast to AdaNode at ast.py:19846>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19846
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-done 4888
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2434;
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node_Array_Type;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1606;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1606;
               Finalizer_Scope_2434;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_189;
--# end
--# property-start TaskBody._ref_cond_190 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19848
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_190
  (Node : Bare_Task_Body
  )
   return Boolean
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4889 '<FieldAccess for should_ref_generic_formals at ast.py:19848>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19848
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.should_ref_generic_formals
Fld := Libadalang.Implementation.Basic_Decl_P_Should_Ref_Generic_Formals (Node => Self);
--# end
--# expr-done 4889
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_190;
--# end
--# property-start TaskBody._ref_env_nodes_191 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19852
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_191
  (Node : Bare_Task_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1608 with Inline_Always;
      Cast_Expr : Bare_Task_Body;
Cast_Result : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1608 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1608;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4890 '<Cast to AdaNode at ast.py:19852>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19852
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-done 4890
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Cast_Result;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1608;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1608;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_191;
--# end
--# property-start TaskBody._ref_env_nodes_192 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19857
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_192
  (Node : Bare_Task_Body
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1609 with Inline_Always;
      Cast_Expr : Bare_Task_Body;
Cast_Result : Bare_Ada_Node;
Singleton : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1609 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1609;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4891 '<Cast to AdaNode at ast.py:19857>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19857
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# expr-done 4891
                Singleton := Create_Bare_Ada_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Cast_Result;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1609;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1609;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_192;
--# end
--# property-start TaskBody._ref_cond_193 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19859
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_193
  (Node : Bare_Task_Body
  )
   return Boolean
is
   Self : Bare_Task_Body := Bare_Task_Body (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4892 '<FieldAccess for is_subunit at ast.py:19859>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19859
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.is_subunit
Fld := Libadalang.Implementation.Body_Node_P_Is_Subunit (Node => Self);
--# end
--# expr-done 4892
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_193;
--# end
function Task_Body_P_Task_Type_Decl_Scope
  (E : Entity
  ) return Lexical_Env
is
      E_Info : Internal_Entity_Info :=
         Shed_Rebindings (E.Info, Node_Env (E.Node));
   Result : constant Lexical_Env :=
      Task_Body_P_Task_Type_Decl_Scope (E.Node, E_Info);
begin
      return Result;
end;
   --
   --  Primitives for Bare_Entry_Index_Spec
   --
      procedure Entry_Index_Spec_Pre_Env_Actions
        (Self            : Bare_Entry_Index_Spec;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_214 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:19997");
         Dec_Ref (Mapping.Dest_Env);
      end;
      end;
      procedure Initialize_Fields_For_Entry_Index_Spec
        (Self : Bare_Entry_Index_Spec
         ; Entry_Index_Spec_F_Id : Bare_Defining_Name
         ; Entry_Index_Spec_F_Subtype : Bare_Ada_Node
         ; Entry_Index_Spec_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Self.Entry_Index_Spec_F_Id := Entry_Index_Spec_F_Id;
            Self.Entry_Index_Spec_F_Subtype := Entry_Index_Spec_F_Subtype;
            Self.Entry_Index_Spec_F_Aspects := Entry_Index_Spec_F_Aspects;
      end Initialize_Fields_For_Entry_Index_Spec;
   function Entry_Index_Spec_F_Id
     (Node : Bare_Entry_Index_Spec) return Bare_Defining_Name
   is
   begin
         return Node.Entry_Index_Spec_F_Id;
   end;
   function Entry_Index_Spec_F_Subtype
     (Node : Bare_Entry_Index_Spec) return Bare_Ada_Node
   is
   begin
         return Node.Entry_Index_Spec_F_Subtype;
   end;
--# property-start EntryIndexSpec.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19999
pragma Warnings (Off, "is not referenced");
 function Entry_Index_Spec_P_Defining_Names
  (Node : Bare_Entry_Index_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Entry_Index_Spec := Bare_Entry_Index_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Index_Spec :=
     Internal_Entity_Entry_Index_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1389 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1389 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1389;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4893 '<FieldAccess for id at ast.py:19999>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19999
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Entry_Index_Spec_F_Id, Info => Ent.Info);
--# expr-done 4893
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1389;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1389;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Index_Spec_P_Defining_Names;
--# end
--# property-start EntryIndexSpec.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20000
pragma Warnings (Off, "is not referenced");
 function Entry_Index_Spec_P_Defining_Env
  (Node : Bare_Entry_Index_Spec
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Entry_Index_Spec := Bare_Entry_Index_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Index_Spec :=
     Internal_Entity_Entry_Index_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1390 with Inline_Always;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1390 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1390;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4895 '<FieldAccess for defining_env at ast.py:20000>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20000
--# expr-start 4894 '<FieldAccess for expr_type at ast.py:20000>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20000
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.expr_type'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Expr_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 4894
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Fld.Node, Origin => Origin, Include_Ud_Indexing => Include_Ud_Indexing, E_Info => Fld.Info);
--# end
--# expr-done 4895
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1390;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1390;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Index_Spec_P_Defining_Env;
--# end
--# property-start EntryIndexSpec.expr_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20001
pragma Warnings (Off, "is not referenced");
 function Entry_Index_Spec_P_Expr_Type
  (Node : Bare_Entry_Index_Spec
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Entry_Index_Spec := Bare_Entry_Index_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Index_Spec :=
     Internal_Entity_Entry_Index_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Match_0 : Internal_Entity_Subtype_Indication;
Match_1 : Internal_Entity;
Fld : Internal_Entity;
Fld_1 : Internal_Entity_Base_Type_Decl;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Expr;
Fld_2 : Internal_Entity_Base_Type_Decl;
Match_Prefix : Internal_Entity;
Cast_Result_1 : Internal_Entity_Subtype_Indication;
Let_Result : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Let_Result_1 : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Match_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4900 '<Match at ast.py:20001>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20001
--# expr-start 4896 '<FieldAccess for subtype at ast.py:20001>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20001
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity (Node => Ent.Node.Entry_Index_Spec_F_Subtype, Info => Ent.Info);
--# expr-done 4896
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Subtype_Indication .. Ada_Discrete_Subtype_Indication =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Subtype_Indication
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_1; 
--# bind subt Match_0
--# expr-start 4897 '<FieldAccess for designated_type at ast.py:20002>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20002
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Match_0.Node, Origin => Origin, E_Info => Match_0.Info);
--# end
--# expr-done 4897
Let_Result := Fld_1; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Abort_Absent .. Ada_Enum_Lit_Synth_Type_Expr | Ada_Synthetic_Type_Expr .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_2; 
--# bind e Match_1
--# expr-start 4899 '<FieldAccess for expression_type at ast.py:20003>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20003
--# expr-start 4898 '<Cast to Expr at ast.py:20003>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20003
Cast_Expr := Match_1; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Expr
   then
      Cast_Result := Create_Internal_Entity_Expr
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 4898
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.expression_type
Fld_2 := Libadalang.Implementation.Expr_P_Expression_Type (Node => Cast_Result.Node, E_Info => Cast_Result.Info);
--# end
--# expr-done 4899
Let_Result_1 := Fld_2; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
end case;
--# expr-done 4900
      Property_Result := Match_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Index_Spec_P_Expr_Type;
--# end
--# property-start EntryIndexSpec.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20007
pragma Warnings (Off, "is not referenced");
 function Entry_Index_Spec_P_Xref_Equation
  (Node : Bare_Entry_Index_Spec
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Entry_Index_Spec := Bare_Entry_Index_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Index_Spec :=
     Internal_Entity_Entry_Index_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1392 with Inline_Always;
      Fld : Internal_Entity;
Fld_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1392 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1392;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4903 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4902 '<FieldAccess for sub_equation at ast.py:20008>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20008
--# expr-start 4901 '<FieldAccess for subtype at ast.py:20008>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20008
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity (Node => Ent.Node.Entry_Index_Spec_F_Subtype, Info => Ent.Info);
--# expr-done 4901
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 4902
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 4903
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1392;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1392;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Index_Spec_P_Xref_Equation;
--# end
--# property-start EntryIndexSpec.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20010
pragma Warnings (Off, "is not referenced");
 function Entry_Index_Spec_P_Xref_Entry_Point
  (Node : Bare_Entry_Index_Spec
  )
   return Boolean
is
   Self : Bare_Entry_Index_Spec := Bare_Entry_Index_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Entry_Index_Spec_P_Xref_Entry_Point;
--# end
--# property-start EntryIndexSpec._env_mappings_214 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19997
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_214
  (Node : Bare_Entry_Index_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Entry_Index_Spec := Bare_Entry_Index_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Entry_Index_Spec :=
     Internal_Entity_Entry_Index_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1631 with Inline_Always;
      Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Entry_Index_Spec;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1631 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1631;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4906 '<New EnvAssoc at ast.py:19997>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19997
--# expr-start 4904 '<New DesignatedEnv at ast.py:19997>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19997
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 4904
--# expr-start 4905 '<FieldAccess for name_symbol at ast.py:19997>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19997
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 4905
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 4906
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1631;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1631;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_214;
--# end
   --
   --  Primitives for Bare_Error_Decl
   --
--# property-start ErrorDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3304
pragma Warnings (Off, "is not referenced");
 function Error_Decl_P_Defining_Names
  (Node : Bare_Error_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Error_Decl := Bare_Error_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := No_Internal_Entity_Defining_Name_Array_Type;
         Inc_Ref (Property_Result);
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Error_Decl_P_Defining_Names;
--# end
--# property-start ErrorDecl.child_decl_initial_env_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:3307
pragma Warnings (Off, "is not referenced");
 function Error_Decl_P_Child_Decl_Initial_Env_Name
  (Node : Bare_Error_Decl
      ; Private_Part : Boolean
         := False
  )
   return Symbol_Type
is
   Self : Bare_Error_Decl := Bare_Error_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind private_part Private_Part
   Property_Result : Symbol_Type;
      Let_Result : Symbol_Type;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4907 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := null; 
--# end
--# expr-done 4907
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Error_Decl_P_Child_Decl_Initial_Env_Name;
--# end
   --
   --  Primitives for Bare_Exception_Decl
   --
      procedure Exception_Decl_Pre_Env_Actions
        (Self            : Bare_Exception_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mappings : Internal_Env_Assoc_Array_Access :=
            Libadalang.Implementation.Env_Mappings_58 (Node => 
Self);
      begin
         for Mapping of Mappings.Items loop
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:9970");
         end loop;
         Dec_Ref (Mappings);
      end;
      end;
      procedure Initialize_Fields_For_Exception_Decl
        (Self : Bare_Exception_Decl
         ; Exception_Decl_F_Ids : Bare_Defining_Name_List
         ; Exception_Decl_F_Renames : Bare_Renaming_Clause
         ; Exception_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Self.Exception_Decl_F_Ids := Exception_Decl_F_Ids;
            Self.Exception_Decl_F_Renames := Exception_Decl_F_Renames;
            Self.Exception_Decl_F_Aspects := Exception_Decl_F_Aspects;
      end Initialize_Fields_For_Exception_Decl;
   function Exception_Decl_F_Ids
     (Node : Bare_Exception_Decl) return Bare_Defining_Name_List
   is
   begin
         return Node.Exception_Decl_F_Ids;
   end;
   function Exception_Decl_F_Renames
     (Node : Bare_Exception_Decl) return Bare_Renaming_Clause
   is
   begin
         return Node.Exception_Decl_F_Renames;
   end;
--# property-start ExceptionDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9961
pragma Warnings (Off, "is not referenced");
 function Exception_Decl_P_Defining_Names
  (Node : Bare_Exception_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Exception_Decl := Bare_Exception_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Exception_Decl :=
     Internal_Entity_Exception_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_792 with Inline_Always;
      Fld : Internal_Entity_Defining_Name_List;
Fld_1 : Bare_Defining_Name_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity_Defining_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Defining_Name;
Item_176 : Internal_Entity_Defining_Name;
Bare_Item_176 : Bare_Defining_Name;
Map_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_792 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_792;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4909 '<Map at ast.py:9961>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9961
   --# expr-start 4908 '<FieldAccess for ids at ast.py:9961>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9961
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name_List (Node => Ent.Node.Exception_Decl_F_Ids, Info => Ent.Info);
--# expr-done 4908
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Internal_Entity_Defining_Name_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Internal_Entity_Defining_Name_Array (0);
      else
      declare
         Collection : constant Bare_Defining_Name_List := Fld_1;
      begin
         for Untyped_Item_176 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_176 := Untyped_Item_176; 
Is_Null := Bare_Item_176 = null; 
if Is_Null then
   If_Result := No_Entity_Defining_Name;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_176); 
   If_Result := As_Entity;
end if;
                  Item_176 := If_Result; 
   --# scope-start
               --# bind id Item_176
         declare
            Item_To_Append : constant Internal_Entity_Defining_Name := Item_176;
         begin
            Internal_Entity_Defining_Name_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Defining_Name_Array
           (Items_Count => Natural (Internal_Entity_Defining_Name_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Defining_Name_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Defining_Name_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Defining_Name_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 4909
      Property_Result := Map_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_792;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_792;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Exception_Decl_P_Defining_Names;
--# end
--# property-start ExceptionDecl.next_part_for_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9964
pragma Warnings (Off, "is not referenced");
 function Exception_Decl_P_Next_Part_For_Decl
  (Node : Bare_Exception_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Exception_Decl := Bare_Exception_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4910 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := No_Entity_Basic_Decl; 
--# end
--# expr-done 4910
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Exception_Decl_P_Next_Part_For_Decl;
--# end
--# property-start ExceptionDecl.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9972
pragma Warnings (Off, "is not referenced");
 function Exception_Decl_P_Xref_Entry_Point
  (Node : Bare_Exception_Decl
  )
   return Boolean
is
   Self : Bare_Exception_Decl := Bare_Exception_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Exception_Decl_P_Xref_Entry_Point;
--# end
--# property-start ExceptionDecl.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9975
pragma Warnings (Off, "is not referenced");
 function Exception_Decl_P_Xref_Equation
  (Node : Bare_Exception_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Exception_Decl := Bare_Exception_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Exception_Decl :=
     Internal_Entity_Exception_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_795 with Inline_Always;
            procedure Finalizer_Scope_2438 with Inline_Always;
      Var_Expr : Internal_Entity_Renaming_Clause;
Fld : Internal_Entity_Renaming_Clause;
Fld_1 : Internal_Entity_Name;
Fld_2 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_795 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_795;
            procedure Finalizer_Scope_2438 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_2438;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4914 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4911 '<FieldAccess for renames at ast.py:9976>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9976
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Renaming_Clause (Node => Ent.Node.Exception_Decl_F_Renames, Info => Ent.Info);
--# expr-done 4911
Var_Expr := Fld;
if Var_Expr /= No_Entity_Renaming_Clause then
   --# scope-start
   --# bind c Var_Expr
--# expr-start 4913 '<FieldAccess for xref_no_overloading at ast.py:9977>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9977
--# expr-start 4912 '<FieldAccess for renamed_object at ast.py:9977>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9977
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Var_Expr.Node.Renaming_Clause_F_Renamed_Object, Info => Var_Expr.Info);
--# expr-done 4912
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_2 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Fld_1.Node, Env => Env, Origin => Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 4913
Scope_Result := Fld_2;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2438;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9978") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 4914
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_795;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_795;
               Finalizer_Scope_2438;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Exception_Decl_P_Xref_Equation;
--# end
--# property-start ExceptionDecl._env_mappings_58 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9970
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_58
  (Node : Bare_Exception_Decl
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Exception_Decl := Bare_Exception_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1475 with Inline_Always;
      Fld : Bare_Defining_Name_List;
Cast_Expr : Bare_Exception_Decl;
Cast_Result : Bare_Ada_Node;
Fld_1 : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1475 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1475;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4916 '<FieldAccess for env_mappings at ast.py:9970>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9970
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4915 '<FieldAccess for ids at ast.py:9970>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9970
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Exception_Decl_F_Ids;
--# expr-done 4915
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# property-call-start AdaNode.env_mappings
Fld_1 := Libadalang.Implementation.Ada_Node_P_Env_Mappings (Node => Self, Defining_Names => Fld, Value => Cast_Result);
--# end
--# expr-done 4916
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1475;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1475;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_58;
--# end
   --
   --  Primitives for Bare_Exception_Handler
   --
      procedure Exception_Handler_Pre_Env_Actions
        (Self            : Bare_Exception_Handler;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_150 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mappings : Internal_Env_Assoc_Array_Access :=
            Libadalang.Implementation.Env_Mappings_151 (Node => 
Self);
      begin
         for Mapping of Mappings.Items loop
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:18996");
         end loop;
         Dec_Ref (Mappings);
      end;
      end;
      procedure Initialize_Fields_For_Exception_Handler
        (Self : Bare_Exception_Handler
         ; Exception_Handler_F_Exception_Name : Bare_Defining_Name
         ; Exception_Handler_F_Handled_Exceptions : Bare_Alternatives_List
         ; Exception_Handler_F_Stmts : Bare_Stmt_List
        ) is
      begin
            Self.Exception_Handler_F_Exception_Name := Exception_Handler_F_Exception_Name;
            Self.Exception_Handler_F_Handled_Exceptions := Exception_Handler_F_Handled_Exceptions;
            Self.Exception_Handler_F_Stmts := Exception_Handler_F_Stmts;
      end Initialize_Fields_For_Exception_Handler;
   function Exception_Handler_F_Exception_Name
     (Node : Bare_Exception_Handler) return Bare_Defining_Name
   is
   begin
         return Node.Exception_Handler_F_Exception_Name;
   end;
   function Exception_Handler_F_Handled_Exceptions
     (Node : Bare_Exception_Handler) return Bare_Alternatives_List
   is
   begin
         return Node.Exception_Handler_F_Handled_Exceptions;
   end;
   function Exception_Handler_F_Stmts
     (Node : Bare_Exception_Handler) return Bare_Stmt_List
   is
   begin
         return Node.Exception_Handler_F_Stmts;
   end;
--# property-start ExceptionHandler.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19006
pragma Warnings (Off, "is not referenced");
 function Exception_Handler_P_Defining_Names
  (Node : Bare_Exception_Handler
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Exception_Handler := Bare_Exception_Handler (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Exception_Handler :=
     Internal_Entity_Exception_Handler'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1333 with Inline_Always;
            procedure Finalizer_Scope_2439 with Inline_Always;
      Var_Expr : Internal_Entity_Defining_Name;
Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
Scope_Result : Internal_Entity_Defining_Name_Array_Access;
Result_Var : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1333 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1333;
            procedure Finalizer_Scope_2439 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2439;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 4917 '<FieldAccess for exception_name at ast.py:19007>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19007
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Exception_Handler_F_Exception_Name, Info => Ent.Info);
--# expr-done 4917
Var_Expr := Fld;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
   --# bind n Var_Expr
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2439;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Defining_Name_Array_Type;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1333;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1333;
               Finalizer_Scope_2439;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Exception_Handler_P_Defining_Names;
--# end
--# property-start ExceptionHandler.expr_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19011
pragma Warnings (Off, "is not referenced");
 function Exception_Handler_P_Expr_Type
  (Node : Bare_Exception_Handler
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Exception_Handler := Bare_Exception_Handler (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Exception_Handler :=
     Internal_Entity_Exception_Handler'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1334 with Inline_Always;
            procedure Finalizer_Scope_2440 with Inline_Always;
      Var_Expr : Bare_Basic_Decl;
Sym : Symbol_Type;
Sym_1 : Symbol_Type;
Array_Lit : Symbol_Type_Array_Access;
Fld : Bare_Basic_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Sym_2 : Symbol_Type;
Env_Get_Result : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1334 is
            begin
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1334;
            procedure Finalizer_Scope_2440 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2440;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4926 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4925 '<Cast to BaseTypeDecl at ast.py:19013>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19013
--# expr-start 4924 '<EnvGet at ast.py:19013>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19013
--# expr-start 4921 '<FieldAccess for get_unit_root_decl at ast.py:19013>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19013
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 4920 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 4918 '<Symbol Ada at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Ada); 
--# expr-done 4918
--# expr-start 4919 '<Symbol Exceptions at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Exceptions); 
--# expr-done 4919
Array_Lit := Create_Symbol_Type_Array (Internal_Symbol_Type_Array'(1 => Sym, 2 => Sym_1)); 
--# expr-done 4920
--# property-call-start AdaNode.get_unit_root_decl
Fld := Libadalang.Implementation.Ada_Node_P_Get_Unit_Root_Decl (Node => Ent.Node, Name => Array_Lit, Kind => Unit_Specification);
--# end
--# expr-done 4921
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 4922 '<FieldAccess for children_env at ast.py:19013>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19013
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Children_Env (Node => Var_Expr);
--# expr-done 4922
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2440;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
--# expr-start 4923 '<Symbol Exception_Occurrence at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Exception_Occurrence); 
--# expr-done 4923
Env_Get_Result := AST_Envs.Get_First (Self => Result_Var, Key => Sym_2, Lookup_Kind => To_Lookup_Kind_Type (Flat), Categories => (Inherited_Primitives => True, Nocat => True)); 
--# expr-done 4924
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 4925
Let_Result := Cast_Result; 
--# end
--# expr-done 4926
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1334;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1334;
               Finalizer_Scope_2440;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Exception_Handler_P_Expr_Type;
--# end
--# property-start ExceptionHandler.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19019
pragma Warnings (Off, "is not referenced");
 function Exception_Handler_P_Xref_Equation
  (Node : Bare_Exception_Handler
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Exception_Handler := Bare_Exception_Handler (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1335 with Inline_Always;
            procedure Finalizer_Scope_2441 with Inline_Always;
      Fld : Bare_Alternatives_List;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Fld_1 : Logic_Equation;
Item_177 : Bare_Ada_Node;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
            procedure Finalizer_Scope_1335 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
            end Finalizer_Scope_1335;
            procedure Finalizer_Scope_2441 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2441;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4931 '<LogicAll at ast.py:19020>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19020
--# expr-start 4930 '<Map at ast.py:19020>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19020
   --# expr-start 4927 '<FieldAccess for handled_exceptions at ast.py:19020>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19020
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Exception_Handler_F_Handled_Exceptions;
--# expr-done 4927
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Alternatives_List := Fld;
      begin
         for Untyped_Item_177 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_177 := Untyped_Item_177; 
   --# scope-start
               --# bind he Item_177
      --# expr-start 4929 '<FieldAccess for sub_equation at ast.py:19020>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19020
--# expr-start 4928 '<as_entity at ast.py:19020>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19020
Is_Null := Item_177 = null; 
if Is_Null then
   If_Result := No_Entity;
else
As_Entity := (Info => E_Info, Node => Item_177); 
   If_Result := As_Entity;
end if;
--# expr-done 4928
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => If_Result.Node, Env => Env, Origin => Origin, E_Info => If_Result.Info);
--# end
--# expr-done 4929
         declare
            Item_To_Append : constant Logic_Equation := Fld_1;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2441;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 4930
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19020") else null)); 
--# expr-done 4931
      Property_Result := Logic_Boolean_Op;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1335;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1335;
               Finalizer_Scope_2441;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Exception_Handler_P_Xref_Equation;
--# end
--# property-start ExceptionHandler.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19023
pragma Warnings (Off, "is not referenced");
 function Exception_Handler_P_Xref_Entry_Point
  (Node : Bare_Exception_Handler
  )
   return Boolean
is
   Self : Bare_Exception_Handler := Bare_Exception_Handler (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Exception_Handler_P_Xref_Entry_Point;
--# end
--# property-start ExceptionHandler.is_constant_object /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19025
pragma Warnings (Off, "is not referenced");
 function Exception_Handler_P_Is_Constant_Object
  (Node : Bare_Exception_Handler
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Exception_Handler := Bare_Exception_Handler (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Exception_Handler_P_Is_Constant_Object;
--# end
--# property-start ExceptionHandler._env_trans_parent_150 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18994
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_150
  (Node : Bare_Exception_Handler
  )
   return Boolean
is
   Self : Bare_Exception_Handler := Bare_Exception_Handler (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_150;
--# end
--# property-start ExceptionHandler._env_mappings_151 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18997
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_151
  (Node : Bare_Exception_Handler
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Exception_Handler := Bare_Exception_Handler (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Exception_Handler :=
     Internal_Entity_Exception_Handler'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1568 with Inline_Always;
            procedure Finalizer_Scope_2442 with Inline_Always;
            procedure Finalizer_Scope_2443 with Inline_Always;
      Var_Expr : Internal_Entity_Defining_Name;
Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
Fld_1 : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Exception_Handler;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
Map_Result : Internal_Env_Assoc_Array_Access;
Scope_Result : Internal_Env_Assoc_Array_Access;
Result_Var : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1568 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1568;
            procedure Finalizer_Scope_2442 is
            begin
                     Dec_Ref (Singleton);
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_2442;
            procedure Finalizer_Scope_2443 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_2443;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 4932 '<FieldAccess for exception_name at ast.py:18997>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18997
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Exception_Handler_F_Exception_Name, Info => Ent.Info);
--# expr-done 4932
Var_Expr := Fld;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
   --# bind n Var_Expr
--# expr-start 4936 '<Map at ast.py:18997>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18997
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
      declare
         Map_Result_Vec : Internal_Env_Assoc_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Singleton;
      begin
         for Item_178 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_178
      --# expr-start 4935 '<New EnvAssoc at ast.py:18999>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18999
--# expr-start 4933 '<New DesignatedEnv at ast.py:19001>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19001
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 4933
--# expr-start 4934 '<FieldAccess for name_symbol at ast.py:18999>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18999
      if Item_178.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Item_178.Node);
--# end
--# expr-done 4934
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld_1, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 4935
         declare
            Item_To_Append : constant Internal_Env_Assoc := New_Struct_1;
         begin
               Inc_Ref (Item_To_Append);
            Internal_Env_Assoc_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2443;
         end loop;
      end;
         Map_Result := Create_Internal_Env_Assoc_Array
           (Items_Count => Natural (Internal_Env_Assoc_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Env_Assoc_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Env_Assoc_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Env_Assoc_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 4936
Scope_Result := Map_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2442;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Env_Assoc_Array_Type;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1568;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1568;
               Finalizer_Scope_2442;
               Finalizer_Scope_2443;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_151;
--# end
   --
   --  Primitives for Bare_For_Loop_Var_Decl
   --
      procedure For_Loop_Var_Decl_Pre_Env_Actions
        (Self            : Bare_For_Loop_Var_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_133 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:16988");
         Dec_Ref (Mapping.Dest_Env);
      end;
      end;
      procedure Initialize_Fields_For_For_Loop_Var_Decl
        (Self : Bare_For_Loop_Var_Decl
         ; For_Loop_Var_Decl_F_Id : Bare_Defining_Name
         ; For_Loop_Var_Decl_F_Id_Type : Bare_Type_Expr
        ) is
      begin
            Self.For_Loop_Var_Decl_F_Id := For_Loop_Var_Decl_F_Id;
            Self.For_Loop_Var_Decl_F_Id_Type := For_Loop_Var_Decl_F_Id_Type;
      end Initialize_Fields_For_For_Loop_Var_Decl;
   function For_Loop_Var_Decl_F_Id
     (Node : Bare_For_Loop_Var_Decl) return Bare_Defining_Name
   is
   begin
         return Node.For_Loop_Var_Decl_F_Id;
   end;
   function For_Loop_Var_Decl_F_Id_Type
     (Node : Bare_For_Loop_Var_Decl) return Bare_Type_Expr
   is
   begin
         return Node.For_Loop_Var_Decl_F_Id_Type;
   end;
--# property-start ForLoopVarDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16960
pragma Warnings (Off, "is not referenced");
 function For_Loop_Var_Decl_P_Defining_Names
  (Node : Bare_For_Loop_Var_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_For_Loop_Var_Decl := Bare_For_Loop_Var_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_For_Loop_Var_Decl :=
     Internal_Entity_For_Loop_Var_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1215 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1215 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1215;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4937 '<FieldAccess for id at ast.py:16960>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16960
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.For_Loop_Var_Decl_F_Id, Info => Ent.Info);
--# expr-done 4937
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1215;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1215;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end For_Loop_Var_Decl_P_Defining_Names;
--# end
--# property-start ForLoopVarDecl.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16962
pragma Warnings (Off, "is not referenced");
 function For_Loop_Var_Decl_P_Defining_Env
  (Node : Bare_For_Loop_Var_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_For_Loop_Var_Decl := Bare_For_Loop_Var_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_For_Loop_Var_Decl :=
     Internal_Entity_For_Loop_Var_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1216 with Inline_Always;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1216 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1216;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4939 '<FieldAccess for defining_env at ast.py:16962>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16962
--# expr-start 4938 '<FieldAccess for expr_type at ast.py:16962>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16962
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.expr_type'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Expr_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 4938
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Fld.Node, Origin => Origin, Include_Ud_Indexing => Include_Ud_Indexing, E_Info => Fld.Info);
--# end
--# expr-done 4939
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1216;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1216;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end For_Loop_Var_Decl_P_Defining_Env;
--# end
--# property-start ForLoopVarDecl.expr_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16965
pragma Warnings (Off, "is not referenced");
 function For_Loop_Var_Decl_P_Expr_Type
  (Node : Bare_For_Loop_Var_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_For_Loop_Var_Decl := Bare_For_Loop_Var_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_For_Loop_Var_Decl :=
     Internal_Entity_For_Loop_Var_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Bare_Defining_Name;
Fld_1 : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Defining_Name;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Bare_Ada_Node;
Cast_Result_1 : Bare_For_Loop_Spec;
Fld_5 : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
Fld_6 : Internal_Entity_Type_Expr;
Fld_7 : Internal_Entity_Base_Type_Decl;
Fld_8 : Bare_Type_Expr;
Is_Null : Boolean;
If_Result_1 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_For_Loop_Var_Decl_For_Loop_Var_Decl_P_Expr_Type,
            Items    => new Mmz_Key_Array (1 ..  3))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_For_Loop_Var_Decl,
                                As_Bare_For_Loop_Var_Decl => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Bare_Ada_Node,
                                      As_Bare_Ada_Node => Origin);
               Mmz_K.Items (3) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Base_Type_Decl;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 4954 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4953 '<If at ast.py:16966>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16966
--# expr-start 4941 '<IsNull at ast.py:16967>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16967
--# expr-start 4940 '<FieldAccess for id_type at ast.py:16967>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16967
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Self.For_Loop_Var_Decl_F_Id_Type;
--# expr-done 4940
Is_Null := Fld_8 = null; 
--# expr-done 4941
if Is_Null then
   --# expr-start 4950 '<If at ast.py:16978>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16978
--# expr-start 4944 '<FieldAccess for is_iterated_assoc_spec at ast.py:16979>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16979
--# expr-start 4943 '<Cast to ForLoopSpec at ast.py:16979>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16979
--# expr-start 4942 '<parent at ast.py:16979>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16979
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Self.Parent; 
--# expr-done 4942
   if Fld_4 = null
      or else Fld_4.Kind in Ada_For_Loop_Spec_Range
   then
      Cast_Result_1 := Fld_4;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 4943
      if Cast_Result_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ForLoopSpec.is_iterated_assoc_spec
Fld_5 := Libadalang.Implementation.For_Loop_Spec_P_Is_Iterated_Assoc_Spec (Node => Cast_Result_1);
--# end
--# expr-done 4944
if Fld_5 then
   --# expr-start 4947 '<Cast to BaseTypeDecl at ast.py:16980>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16980
--# expr-start 4946 '<FieldAccess for type_val at ast.py:16980>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16980
--# expr-start 4945 '<FieldAccess for id at ast.py:16980>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16980
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.For_Loop_Var_Decl_F_Id;
--# expr-done 4945
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.type_val
Fld_1 := Libadalang.Implementation.Expr_P_Type_Val (Node => Fld);
--# end
--# expr-done 4946
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 4947
   If_Result := Cast_Result;
else
   --# expr-start 4949 '<FieldAccess for expression_type at ast.py:16981>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16981
--# expr-start 4948 '<FieldAccess for id at ast.py:16981>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16981
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Defining_Name (Node => Ent.Node.For_Loop_Var_Decl_F_Id, Info => Ent.Info);
--# expr-done 4948
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.expression_type
Fld_3 := Libadalang.Implementation.Expr_P_Expression_Type (Node => Fld_2.Node, E_Info => Fld_2.Info);
--# end
--# expr-done 4949
   If_Result := Fld_3;
end if;
--# expr-done 4950
   If_Result_1 := If_Result;
else
   --# expr-start 4952 '<FieldAccess for designated_type at ast.py:16985>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16985
--# expr-start 4951 '<FieldAccess for id_type at ast.py:16985>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16985
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Type_Expr (Node => Ent.Node.For_Loop_Var_Decl_F_Id_Type, Info => Ent.Info);
--# expr-done 4951
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_7 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Fld_6.Node, Origin => Origin, E_Info => Fld_6.Info);
--# end
--# expr-done 4952
   If_Result_1 := Fld_7;
end if;
--# expr-done 4953
Let_Result := If_Result_1; 
--# end
--# expr-done 4954
      Property_Result := Let_Result;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Base_Type_Decl,
                        As_Internal_Entity_Base_Type_Decl => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end For_Loop_Var_Decl_P_Expr_Type;
--# end
--# property-start ForLoopVarDecl.is_constant_object /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16991
pragma Warnings (Off, "is not referenced");
 function For_Loop_Var_Decl_P_Is_Constant_Object
  (Node : Bare_For_Loop_Var_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_For_Loop_Var_Decl := Bare_For_Loop_Var_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_For_Loop_Var_Decl :=
     Internal_Entity_For_Loop_Var_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Block_Var : Internal_Entity_For_Loop_Spec;
Fld : Internal_Entity;
Cast_Result : Internal_Entity_For_Loop_Spec;
Fld_1 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Name;
Fld_2 : Boolean;
Fld_3 : Internal_Entity_Iter_Type;
Is_A : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4963 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4956 '<Cast to ForLoopSpec at ast.py:16993>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16993
--# expr-start 4955 '<parent at ast.py:16993>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16993
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 4955
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_For_Loop_Spec_Range
   then
      Cast_Result := Create_Internal_Entity_For_Loop_Spec
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_For_Loop_Spec;
   end if;
--# expr-done 4956
Block_Var := Cast_Result; 
--# bind loop_spec Block_Var
--# expr-start 4962 '<If at ast.py:16994>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16994
--# expr-start 4958 '<IsA IterType.Of at ast.py:16997>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16997
--# expr-start 4957 '<FieldAccess for loop_type at ast.py:16997>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16997
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Iter_Type (Node => Block_Var.Node.For_Loop_Spec_F_Loop_Type, Info => Block_Var.Info);
--# expr-done 4957
Is_A := Fld_3.Node /= null 
and then Fld_3.Node.Kind in Ada_Iter_Type_Of_Range; 
--# expr-done 4958
if Is_A then
   --# expr-start 4961 '<FieldAccess for is_constant at ast.py:16998>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16998
--# expr-start 4960 '<Cast to Name at ast.py:16998>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16998
--# expr-start 4959 '<FieldAccess for iter_expr at ast.py:16998>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16998
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity (Node => Block_Var.Node.For_Loop_Spec_F_Iter_Expr, Info => Block_Var.Info);
--# expr-done 4959
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Name
   then
      Cast_Result_1 := Create_Internal_Entity_Name
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 4960
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.is_constant'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_Is_Constant (Node => Cast_Result_1.Node, E_Info => Cast_Result_1.Info);
--# end
--# expr-done 4961
   If_Result := Fld_2;
else
   If_Result := True;
end if;
--# expr-done 4962
Let_Result := If_Result; 
--# end
--# expr-done 4963
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end For_Loop_Var_Decl_P_Is_Constant_Object;
--# end
--# property-start ForLoopVarDecl.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17004
pragma Warnings (Off, "is not referenced");
 function For_Loop_Var_Decl_P_Xref_Equation
  (Node : Bare_For_Loop_Var_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_For_Loop_Var_Decl := Bare_For_Loop_Var_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_For_Loop_Var_Decl :=
     Internal_Entity_For_Loop_Var_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1219 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Fld_1 : Logic_Equation;
True_Rel : Logic_Equation;
Fld_2 : Internal_Entity_Type_Expr;
Fld_3 : Logic_Equation;
Fld_4 : Internal_Entity_Type_Expr;
Is_Null : Boolean;
If_Result : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1219 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Fld_3);
                     Dec_Ref (If_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1219;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4972 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4971 '<And at ast.py:17005>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17005
--# expr-start 4965 '<FieldAccess for sub_equation at ast.py:17005>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17005
--# expr-start 4964 '<FieldAccess for id at ast.py:17005>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17005
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.For_Loop_Var_Decl_F_Id, Info => Ent.Info);
--# expr-done 4964
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 4965
--# expr-start 4970 '<If at ast.py:17005>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17005
--# expr-start 4967 '<IsNull at ast.py:17006>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17006
--# expr-start 4966 '<FieldAccess for id_type at ast.py:17006>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17006
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Type_Expr (Node => Ent.Node.For_Loop_Var_Decl_F_Id_Type, Info => Ent.Info);
--# expr-done 4966
Is_Null := Fld_4.Node = null; 
--# expr-done 4967
if Is_Null then
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17007") else null)); 
   If_Result := True_Rel;
else
   --# expr-start 4969 '<FieldAccess for sub_equation at ast.py:17008>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17008
--# expr-start 4968 '<FieldAccess for id_type at ast.py:17008>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17008
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Type_Expr (Node => Ent.Node.For_Loop_Var_Decl_F_Id_Type, Info => Ent.Info);
--# expr-done 4968
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_2.Node, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 4969
   If_Result := Fld_3;
end if;
      Inc_Ref (If_Result);
--# expr-done 4970
And_Pred := Create_And (Fld_1, If_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17005") else null)); 
--# expr-done 4971
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 4972
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1219;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1219;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end For_Loop_Var_Decl_P_Xref_Equation;
--# end
--# property-start ForLoopVarDecl._env_mappings_133 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16988
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_133
  (Node : Bare_For_Loop_Var_Decl
  )
   return Internal_Env_Assoc
is
   Self : Bare_For_Loop_Var_Decl := Bare_For_Loop_Var_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1550 with Inline_Always;
      Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_For_Loop_Var_Decl;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1550 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1550;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4975 '<New EnvAssoc at ast.py:16988>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16988
--# expr-start 4973 '<New DesignatedEnv at ast.py:16988>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16988
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 4973
--# expr-start 4974 '<FieldAccess for name_symbol at ast.py:16988>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16988
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Self);
--# end
--# expr-done 4974
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 4975
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1550;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1550;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_133;
--# end
   --
   --  Primitives for Bare_Generic_Decl
   --
      procedure Initialize_Fields_For_Generic_Decl
        (Self : Bare_Generic_Decl
         ; Generic_Decl_F_Formal_Part : Bare_Generic_Formal_Part
        ) is
      begin
            Self.Generic_Decl_F_Formal_Part := Generic_Decl_F_Formal_Part;
      end Initialize_Fields_For_Generic_Decl;
   function Generic_Decl_F_Formal_Part
     (Node : Bare_Generic_Decl) return Bare_Generic_Formal_Part
   is
   begin
         return Node.Generic_Decl_F_Formal_Part;
   end;
--# property-start '[dispatcher]GenericDecl.decl' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Generic_Decl_P_Decl
  (Node : Bare_Generic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Generic_Decl := Bare_Generic_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Generic_Decl (Self.Kind) is
               when Ada_Generic_Package_Decl =>
                  --# property-call-start GenericPackageDecl.decl
                  Property_Result := Generic_Package_Decl_P_Decl
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Subp_Decl =>
                  --# property-call-start GenericSubpDecl.decl
                  Property_Result := Generic_Subp_Decl_P_Decl
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Generic_Decl_P_Decl;
--# end
--# property-start GenericDecl.get_aspect_assoc /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10796
pragma Warnings (Off, "is not referenced");
 function Generic_Decl_P_Get_Aspect_Assoc
  (Node : Bare_Generic_Decl
      ; Name : Symbol_Type
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Aspect_Assoc
is
   Self : Bare_Generic_Decl := Bare_Generic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Decl :=
     Internal_Entity_Generic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind name Name
   Property_Result : Internal_Entity_Aspect_Assoc;
      Fld : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Aspect_Assoc;
Let_Result : Internal_Entity_Aspect_Assoc;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4978 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4977 '<FieldAccess for get_aspect_assoc at ast.py:10799>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10799
--# expr-start 4976 '<FieldAccess for decl at ast.py:10799>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10799
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericDecl.decl'
Fld := Libadalang.Implementation.Dispatcher_Generic_Decl_P_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4976
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.get_aspect_assoc'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Get_Aspect_Assoc (Node => Fld.Node, Name => Name, E_Info => Fld.Info);
--# end
--# expr-done 4977
Let_Result := Fld_1; 
--# end
--# expr-done 4978
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Decl_P_Get_Aspect_Assoc;
--# end
--# property-start GenericDecl.next_part_for_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10802
pragma Warnings (Off, "is not referenced");
 function Generic_Decl_P_Next_Part_For_Decl
  (Node : Bare_Generic_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Generic_Decl := Bare_Generic_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Decl :=
     Internal_Entity_Generic_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4981 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4980 '<FieldAccess for next_part_for_decl at ast.py:10803>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10803
--# expr-start 4979 '<FieldAccess for decl at ast.py:10803>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10803
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericDecl.decl'
Fld := Libadalang.Implementation.Dispatcher_Generic_Decl_P_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4979
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.next_part_for_decl'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Next_Part_For_Decl (Node => Fld.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld.Info);
--# end
--# expr-done 4980
Let_Result := Fld_1; 
--# end
--# expr-done 4981
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Decl_P_Next_Part_For_Decl;
--# end
   --
   --  Primitives for Bare_Generic_Package_Decl
   --
      procedure Generic_Package_Decl_Pre_Env_Actions
        (Self            : Bare_Generic_Package_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Dummy : Boolean;
      begin
         Dummy := Libadalang.Implementation.Env_Do_119 (Node => 
Self);
      end;
      declare
         Env : constant Internal_Designated_Env :=
           Libadalang.Implementation.Initial_Env_120 (Node => 
Self);
      begin
         Set_Initial_Env
           (Self,
            State,
            Env,
            DSL_Location => "ast.py:10890");
      end;
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mappings : Internal_Env_Assoc_Array_Access :=
            Libadalang.Implementation.Env_Mappings_121 (Node => 
Self);
      begin
         for Mapping of Mappings.Items loop
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:10897");
         end loop;
         Dec_Ref (Mappings);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_122 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      declare
         Dummy : Bare_Compilation_Unit_Array_Access;
      begin
         Dummy := Libadalang.Implementation.Env_Do_123 (Node => 
Self);
         Dec_Ref (Dummy);
      end;
      if Libadalang.Implementation.Ref_Cond_125 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_124 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Use_Package_Name_Designated_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_127 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_126 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Name_Designated_Type_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      end;
      procedure Initialize_Fields_For_Generic_Package_Decl
        (Self : Bare_Generic_Package_Decl
         ; Generic_Decl_F_Formal_Part : Bare_Generic_Formal_Part
         ; Generic_Package_Decl_F_Package_Decl : Bare_Generic_Package_Internal
        ) is
      begin
            Initialize_Fields_For_Generic_Decl
              (Self, Generic_Decl_F_Formal_Part);
            Self.Generic_Package_Decl_F_Package_Decl := Generic_Package_Decl_F_Package_Decl;
      end Initialize_Fields_For_Generic_Package_Decl;
   function Generic_Package_Decl_F_Package_Decl
     (Node : Bare_Generic_Package_Decl) return Bare_Generic_Package_Internal
   is
   begin
         return Node.Generic_Package_Decl_F_Package_Decl;
   end;
--# property-start GenericPackageDecl.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10917
pragma Warnings (Off, "is not referenced");
 function Generic_Package_Decl_P_Defining_Env
  (Node : Bare_Generic_Package_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Generic_Package_Decl := Bare_Generic_Package_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Package_Decl :=
     Internal_Entity_Generic_Package_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_856 with Inline_Always;
      Fld : Internal_Entity_Generic_Package_Internal;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_856 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_856;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4983 '<FieldAccess for defining_env at ast.py:10917>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10917
--# expr-start 4982 '<FieldAccess for package_decl at ast.py:10917>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10917
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Generic_Package_Internal (Node => Ent.Node.Generic_Package_Decl_F_Package_Decl, Info => Ent.Info);
--# expr-done 4982
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Fld.Node, Origin => Origin, Include_Ud_Indexing => Include_Ud_Indexing, E_Info => Fld.Info);
--# end
--# expr-done 4983
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_856;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_856;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Package_Decl_P_Defining_Env;
--# end
--# property-start GenericPackageDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10919
pragma Warnings (Off, "is not referenced");
 function Generic_Package_Decl_P_Defining_Names
  (Node : Bare_Generic_Package_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Generic_Package_Decl := Bare_Generic_Package_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_857 with Inline_Always;
      Fld : Bare_Generic_Package_Internal;
Fld_1 : Bare_Defining_Name;
As_Entity : Internal_Entity_Defining_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_857 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_857;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4986 '<as_entity at ast.py:10920>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10920
--# expr-start 4985 '<FieldAccess for package_name at ast.py:10920>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10920
--# expr-start 4984 '<FieldAccess for package_decl at ast.py:10920>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10920
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Generic_Package_Decl_F_Package_Decl;
--# expr-done 4984
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Base_Package_Decl_F_Package_Name;
--# expr-done 4985
Is_Null := Fld_1 = null; 
if Is_Null then
   If_Result := No_Entity_Defining_Name;
else
As_Entity := (Info => E_Info, Node => Fld_1); 
   If_Result := As_Entity;
end if;
--# expr-done 4986
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := If_Result;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_857;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_857;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Package_Decl_P_Defining_Names;
--# end
--# property-start GenericPackageDecl.body_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10924
pragma Warnings (Off, "is not referenced");
 function Generic_Package_Decl_P_Body_Part
  (Node : Bare_Generic_Package_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Package_Body
is
   Self : Bare_Generic_Package_Decl := Bare_Generic_Package_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Package_Decl :=
     Internal_Entity_Generic_Package_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Package_Body;
      Fld : Internal_Entity_Generic_Package_Internal;
Fld_1 : Internal_Entity_Package_Body;
Let_Result : Internal_Entity_Package_Body;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4989 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 4988 '<FieldAccess for body_part at ast.py:10929>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10929
--# expr-start 4987 '<FieldAccess for package_decl at ast.py:10929>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10929
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Generic_Package_Internal (Node => Ent.Node.Generic_Package_Decl_F_Package_Decl, Info => Ent.Info);
--# expr-done 4987
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasePackageDecl.body_part
Fld_1 := Libadalang.Implementation.Base_Package_Decl_P_Body_Part (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 4988
Let_Result := Fld_1; 
--# end
--# expr-done 4989
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Package_Decl_P_Body_Part;
--# end
--# property-start GenericPackageDecl.decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10931
pragma Warnings (Off, "is not referenced");
 function Generic_Package_Decl_P_Decl
  (Node : Bare_Generic_Package_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Generic_Package_Decl := Bare_Generic_Package_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Package_Decl :=
     Internal_Entity_Generic_Package_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Internal_Entity_Generic_Package_Internal;
Cast_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 4990 '<FieldAccess for package_decl at ast.py:10931>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10931
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Generic_Package_Internal (Node => Ent.Node.Generic_Package_Decl_F_Package_Decl, Info => Ent.Info);
--# expr-done 4990
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Package_Decl_P_Decl;
--# end
--# property-start GenericPackageDecl._env_do_119 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10888
pragma Warnings (Off, "is not referenced");
 function Env_Do_119
  (Node : Bare_Generic_Package_Decl
  )
   return Boolean
is
   Self : Bare_Generic_Package_Decl := Bare_Generic_Package_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4991 '<FieldAccess for env_hook at ast.py:10888>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10888
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.env_hook
Fld := Libadalang.Implementation.Ada_Node_P_Env_Hook (Node => Self);
--# end
--# expr-done 4991
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_119;
--# end
--# property-start GenericPackageDecl._initial_env_120 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10894
pragma Warnings (Off, "is not referenced");
 function Initial_Env_120
  (Node : Bare_Generic_Package_Decl
  )
   return Internal_Designated_Env
is
   Self : Bare_Generic_Package_Decl := Bare_Generic_Package_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_1537 with Inline_Always;
      Fld : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_1537 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1537;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4992 '<FieldAccess for child_decl_initial_env at ast.py:10894>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10894
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.child_decl_initial_env
Fld := Libadalang.Implementation.Basic_Decl_P_Child_Decl_Initial_Env (Node => Self, Private_Part => True);
--# end
--# expr-done 4992
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1537;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1537;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Initial_Env_120;
--# end
--# property-start GenericPackageDecl._env_mappings_121 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10897
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_121
  (Node : Bare_Generic_Package_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Generic_Package_Decl := Bare_Generic_Package_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Package_Decl :=
     Internal_Entity_Generic_Package_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1538 with Inline_Always;
      Fld : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1538 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1538;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4993 '<FieldAccess for child_decl_env_assocs at ast.py:10897>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10897
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.child_decl_env_assocs
Fld := Libadalang.Implementation.Basic_Decl_P_Child_Decl_Env_Assocs (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 4993
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1538;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1538;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_121;
--# end
--# property-start GenericPackageDecl._env_trans_parent_122 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10887
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_122
  (Node : Bare_Generic_Package_Decl
  )
   return Boolean
is
   Self : Bare_Generic_Package_Decl := Bare_Generic_Package_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_122;
--# end
--# property-start GenericPackageDecl._env_do_123 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10901
pragma Warnings (Off, "is not referenced");
 function Env_Do_123
  (Node : Bare_Generic_Package_Decl
  )
   return Bare_Compilation_Unit_Array_Access
is
   Self : Bare_Generic_Package_Decl := Bare_Generic_Package_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1540 with Inline_Always;
      Fld : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1540 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1540;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 4994 '<FieldAccess for populate_dependent_units at ast.py:10901>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10901
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.populate_dependent_units
Fld := Libadalang.Implementation.Basic_Decl_P_Populate_Dependent_Units (Node => Self);
--# end
--# expr-done 4994
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1540;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1540;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_123;
--# end
--# property-start GenericPackageDecl._ref_env_nodes_124 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10903
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_124
  (Node : Bare_Generic_Package_Decl
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Generic_Package_Decl := Bare_Generic_Package_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1541 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1541 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1541;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4995 '<FieldAccess for top_level_use_package_clauses at ast.py:10903>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10903
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_package_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Package_Clauses (Node => Self);
--# end
--# expr-done 4995
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1541;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1541;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_124;
--# end
--# property-start GenericPackageDecl._ref_cond_125 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10905
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_125
  (Node : Bare_Generic_Package_Decl
  )
   return Boolean
is
   Self : Bare_Generic_Package_Decl := Bare_Generic_Package_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4997 '<IsA LibraryItem, Subunit at ast.py:10905>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10905
--# expr-start 4996 '<parent at ast.py:10905>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10905
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 4996
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 4997
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_125;
--# end
--# property-start GenericPackageDecl._ref_env_nodes_126 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10908
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_126
  (Node : Bare_Generic_Package_Decl
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Generic_Package_Decl := Bare_Generic_Package_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1543 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1543 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1543;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 4998 '<FieldAccess for top_level_use_type_clauses at ast.py:10908>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10908
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_type_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Type_Clauses (Node => Self);
--# end
--# expr-done 4998
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1543;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1543;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_126;
--# end
--# property-start GenericPackageDecl._ref_cond_127 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10910
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_127
  (Node : Bare_Generic_Package_Decl
  )
   return Boolean
is
   Self : Bare_Generic_Package_Decl := Bare_Generic_Package_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5000 '<IsA LibraryItem, Subunit at ast.py:10910>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10910
--# expr-start 4999 '<parent at ast.py:10910>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10910
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 4999
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 5000
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_127;
--# end
   --
   --  Primitives for Bare_Generic_Subp_Decl
   --
      procedure Generic_Subp_Decl_Pre_Env_Actions
        (Self            : Bare_Generic_Subp_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Dummy : Boolean;
      begin
         Dummy := Libadalang.Implementation.Env_Do_107 (Node => 
Self);
      end;
      declare
         Env : constant Internal_Designated_Env :=
           Libadalang.Implementation.Initial_Env_108 (Node => 
Self);
      begin
         Set_Initial_Env
           (Self,
            State,
            Env,
            DSL_Location => "ast.py:10827");
      end;
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mappings : Internal_Env_Assoc_Array_Access :=
            Libadalang.Implementation.Env_Mappings_109 (Node => 
Self);
      begin
         for Mapping of Mappings.Items loop
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:10834");
         end loop;
         Dec_Ref (Mappings);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_110 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      declare
         Dummy : Bare_Compilation_Unit_Array_Access;
      begin
         Dummy := Libadalang.Implementation.Env_Do_111 (Node => 
Self);
         Dec_Ref (Dummy);
      end;
      if Libadalang.Implementation.Ref_Cond_113 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_112 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Use_Package_Name_Designated_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_115 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_114 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Name_Designated_Type_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      end;
      procedure Generic_Subp_Decl_Post_Env_Actions
        (Self : Bare_Generic_Subp_Decl; State : in out PLE_Node_State) is
      begin
      declare
         Dummy : Boolean;
      begin
         Dummy := Libadalang.Implementation.Env_Do_116 (Node => 
Self);
      end;
      end;
      procedure Initialize_Fields_For_Generic_Subp_Decl
        (Self : Bare_Generic_Subp_Decl
         ; Generic_Decl_F_Formal_Part : Bare_Generic_Formal_Part
         ; Generic_Subp_Decl_F_Subp_Decl : Bare_Generic_Subp_Internal
        ) is
      begin
            Initialize_Fields_For_Generic_Decl
              (Self, Generic_Decl_F_Formal_Part);
            Self.Generic_Subp_Decl_F_Subp_Decl := Generic_Subp_Decl_F_Subp_Decl;
      end Initialize_Fields_For_Generic_Subp_Decl;
   function Generic_Subp_Decl_F_Subp_Decl
     (Node : Bare_Generic_Subp_Decl) return Bare_Generic_Subp_Internal
   is
   begin
         return Node.Generic_Subp_Decl_F_Subp_Decl;
   end;
--# property-start GenericSubpDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10814
pragma Warnings (Off, "is not referenced");
 function Generic_Subp_Decl_P_Defining_Names
  (Node : Bare_Generic_Subp_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Generic_Subp_Decl := Bare_Generic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Subp_Decl :=
     Internal_Entity_Generic_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_850 with Inline_Always;
      Fld : Internal_Entity_Generic_Subp_Internal;
Fld_1 : Internal_Entity_Subp_Spec;
Fld_2 : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_850 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_850;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5003 '<FieldAccess for name at ast.py:10814>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10814
--# expr-start 5002 '<FieldAccess for subp_spec at ast.py:10814>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10814
--# expr-start 5001 '<FieldAccess for subp_decl at ast.py:10814>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10814
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Generic_Subp_Internal (Node => Ent.Node.Generic_Subp_Decl_F_Subp_Decl, Info => Ent.Info);
--# expr-done 5001
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Subp_Spec (Node => Fld.Node.Generic_Subp_Internal_F_Subp_Spec, Info => Fld.Info);
--# expr-done 5002
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.name'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Name (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 5003
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld_2;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_850;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_850;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Subp_Decl_P_Defining_Names;
--# end
--# property-start GenericSubpDecl.body_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10817
pragma Warnings (Off, "is not referenced");
 function Generic_Subp_Decl_P_Body_Part
  (Node : Bare_Generic_Subp_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Subp_Body
is
   Self : Bare_Generic_Subp_Decl := Bare_Generic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Subp_Decl :=
     Internal_Entity_Generic_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Base_Subp_Body;
      Fld : Internal_Entity_Body_Node;
Cast_Result : Internal_Entity_Base_Subp_Body;
Let_Result : Internal_Entity_Base_Subp_Body;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5006 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5005 '<Cast to BaseSubpBody at ast.py:10821>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10821
--# expr-start 5004 '<FieldAccess for body_part_for_decl at ast.py:10821>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10821
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.body_part_for_decl
Fld := Libadalang.Implementation.Basic_Decl_P_Body_Part_For_Decl (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 5004
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Base_Subp_Body
   then
      Cast_Result := Create_Internal_Entity_Base_Subp_Body
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Base_Subp_Body;
   end if;
--# expr-done 5005
Let_Result := Cast_Result; 
--# end
--# expr-done 5006
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Subp_Decl_P_Body_Part;
--# end
--# property-start GenericSubpDecl.decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10856
pragma Warnings (Off, "is not referenced");
 function Generic_Subp_Decl_P_Decl
  (Node : Bare_Generic_Subp_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Generic_Subp_Decl := Bare_Generic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Subp_Decl :=
     Internal_Entity_Generic_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Internal_Entity_Generic_Subp_Internal;
Cast_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 5007 '<FieldAccess for subp_decl at ast.py:10856>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10856
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Generic_Subp_Internal (Node => Ent.Node.Generic_Subp_Decl_F_Subp_Decl, Info => Ent.Info);
--# expr-done 5007
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Subp_Decl_P_Decl;
--# end
--# property-start GenericSubpDecl.is_imported /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10859
pragma Warnings (Off, "is not referenced");
 function Generic_Subp_Decl_P_Is_Imported
  (Node : Bare_Generic_Subp_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Generic_Subp_Decl := Bare_Generic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Subp_Decl :=
     Internal_Entity_Generic_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Internal_Entity_Generic_Subp_Internal;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5009 '<FieldAccess for is_imported at ast.py:10859>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10859
--# expr-start 5008 '<FieldAccess for subp_decl at ast.py:10859>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10859
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Generic_Subp_Internal (Node => Ent.Node.Generic_Subp_Decl_F_Subp_Decl, Info => Ent.Info);
--# expr-done 5008
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_imported'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Imported (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 5009
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Subp_Decl_P_Is_Imported;
--# end
--# property-start GenericSubpDecl.next_part_for_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10862
pragma Warnings (Off, "is not referenced");
 function Generic_Subp_Decl_P_Next_Part_For_Decl
  (Node : Bare_Generic_Subp_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Generic_Subp_Decl := Bare_Generic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Subp_Decl :=
     Internal_Entity_Generic_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Internal_Entity_Generic_Subp_Internal;
Fld_1 : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5012 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5011 '<FieldAccess for next_part_for_decl at ast.py:10863>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10863
--# expr-start 5010 '<FieldAccess for subp_decl at ast.py:10863>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10863
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Generic_Subp_Internal (Node => Ent.Node.Generic_Subp_Decl_F_Subp_Decl, Info => Ent.Info);
--# expr-done 5010
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.next_part_for_decl'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Next_Part_For_Decl (Node => Fld.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld.Info);
--# end
--# expr-done 5011
Let_Result := Fld_1; 
--# end
--# expr-done 5012
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Subp_Decl_P_Next_Part_For_Decl;
--# end
--# property-start GenericSubpDecl._env_do_107 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10825
pragma Warnings (Off, "is not referenced");
 function Env_Do_107
  (Node : Bare_Generic_Subp_Decl
  )
   return Boolean
is
   Self : Bare_Generic_Subp_Decl := Bare_Generic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5013 '<FieldAccess for env_hook at ast.py:10825>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10825
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.env_hook
Fld := Libadalang.Implementation.Ada_Node_P_Env_Hook (Node => Self);
--# end
--# expr-done 5013
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_107;
--# end
--# property-start GenericSubpDecl._initial_env_108 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10831
pragma Warnings (Off, "is not referenced");
 function Initial_Env_108
  (Node : Bare_Generic_Subp_Decl
  )
   return Internal_Designated_Env
is
   Self : Bare_Generic_Subp_Decl := Bare_Generic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_1525 with Inline_Always;
      Fld : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_1525 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1525;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5014 '<FieldAccess for child_decl_initial_env at ast.py:10831>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10831
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.child_decl_initial_env
Fld := Libadalang.Implementation.Basic_Decl_P_Child_Decl_Initial_Env (Node => Self, Private_Part => True);
--# end
--# expr-done 5014
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1525;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1525;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Initial_Env_108;
--# end
--# property-start GenericSubpDecl._env_mappings_109 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10834
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_109
  (Node : Bare_Generic_Subp_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Generic_Subp_Decl := Bare_Generic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Subp_Decl :=
     Internal_Entity_Generic_Subp_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1526 with Inline_Always;
      Fld : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1526 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1526;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5015 '<FieldAccess for child_decl_env_assocs at ast.py:10834>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10834
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.child_decl_env_assocs
Fld := Libadalang.Implementation.Basic_Decl_P_Child_Decl_Env_Assocs (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5015
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1526;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1526;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_109;
--# end
--# property-start GenericSubpDecl._env_trans_parent_110 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10823
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_110
  (Node : Bare_Generic_Subp_Decl
  )
   return Boolean
is
   Self : Bare_Generic_Subp_Decl := Bare_Generic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_110;
--# end
--# property-start GenericSubpDecl._env_do_111 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10838
pragma Warnings (Off, "is not referenced");
 function Env_Do_111
  (Node : Bare_Generic_Subp_Decl
  )
   return Bare_Compilation_Unit_Array_Access
is
   Self : Bare_Generic_Subp_Decl := Bare_Generic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1528 with Inline_Always;
      Fld : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1528 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1528;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5016 '<FieldAccess for populate_dependent_units at ast.py:10838>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10838
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.populate_dependent_units
Fld := Libadalang.Implementation.Basic_Decl_P_Populate_Dependent_Units (Node => Self);
--# end
--# expr-done 5016
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1528;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1528;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_111;
--# end
--# property-start GenericSubpDecl._ref_env_nodes_112 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10841
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_112
  (Node : Bare_Generic_Subp_Decl
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Generic_Subp_Decl := Bare_Generic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1529 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1529 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1529;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5017 '<FieldAccess for top_level_use_package_clauses at ast.py:10841>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10841
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_package_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Package_Clauses (Node => Self);
--# end
--# expr-done 5017
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1529;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1529;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_112;
--# end
--# property-start GenericSubpDecl._ref_cond_113 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10843
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_113
  (Node : Bare_Generic_Subp_Decl
  )
   return Boolean
is
   Self : Bare_Generic_Subp_Decl := Bare_Generic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5019 '<IsA LibraryItem, Subunit at ast.py:10843>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10843
--# expr-start 5018 '<parent at ast.py:10843>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10843
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 5018
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 5019
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_113;
--# end
--# property-start GenericSubpDecl._ref_env_nodes_114 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10846
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_114
  (Node : Bare_Generic_Subp_Decl
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Generic_Subp_Decl := Bare_Generic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1531 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1531 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1531;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5020 '<FieldAccess for top_level_use_type_clauses at ast.py:10846>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10846
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_type_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Type_Clauses (Node => Self);
--# end
--# expr-done 5020
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1531;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1531;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_114;
--# end
--# property-start GenericSubpDecl._ref_cond_115 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10848
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_115
  (Node : Bare_Generic_Subp_Decl
  )
   return Boolean
is
   Self : Bare_Generic_Subp_Decl := Bare_Generic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5022 '<IsA LibraryItem, Subunit at ast.py:10848>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10848
--# expr-start 5021 '<parent at ast.py:10848>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10848
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 5021
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 5022
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_115;
--# end
--# property-start GenericSubpDecl._env_do_116 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10853
pragma Warnings (Off, "is not referenced");
 function Env_Do_116
  (Node : Bare_Generic_Subp_Decl
  )
   return Boolean
is
   Self : Bare_Generic_Subp_Decl := Bare_Generic_Subp_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5023 '<FieldAccess for populate_body_unit at ast.py:10853>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10853
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.populate_body_unit
Fld := Libadalang.Implementation.Basic_Decl_P_Populate_Body_Unit (Node => Self);
--# end
--# expr-done 5023
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_116;
--# end
   --
   --  Primitives for Bare_Generic_Instantiation
   --
      procedure Initialize_Fields_For_Generic_Instantiation
        (Self : Bare_Generic_Instantiation
        ) is
      begin
      Self.Lf_Present_Instantiation_Env := False;
      Self.Lf_Stg_Instantiation_Env := Empty_Env;
      Self.Lf_Present_Actual_Expr_Decls := False;
      Self.Lf_Stg_Actual_Expr_Decls := No_Bare_Anonymous_Expr_Decl_Array_Type;
      end Initialize_Fields_For_Generic_Instantiation;
--# property-start GenericInstantiation.instantiation_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9988
pragma Warnings (Off, "is not referenced");
 function Generic_Instantiation_F_Instantiation_Env
  (Node : Bare_Generic_Instantiation
  )
   return Lexical_Env
is
   Self : Bare_Generic_Instantiation := Bare_Generic_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_796 with Inline_Always;
      Dyn_Env : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_796 is
            begin
                     Dec_Ref (Dyn_Env);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_796;
begin
   --# property-body-start
      if Self.Lf_Present_Instantiation_Env then
         Property_Result := Self.Lf_Stg_Instantiation_Env;
            Inc_Ref (Property_Result);
         return Property_Result;
      end if;
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5025 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5024 '<DynamicLexicalEnv at ast.py:9989>' Dyn_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9989
Dyn_Env := Create_Dynamic_Lexical_Env (Self, Generic_Instantiation_P_Instantiation_Bindings'Access, Ada_Node_P_Resolve_Generic_Actual'Access, False); 
--# expr-done 5024
Let_Result := Dyn_Env; Inc_Ref (Let_Result);
--# end
--# expr-done 5025
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_796;
      Self.Lf_Present_Instantiation_Env := True;
      Self.Lf_Stg_Instantiation_Env := Property_Result;
         Inc_Ref (Property_Result);
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_796;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Instantiation_F_Instantiation_Env;
--# end
--# property-start GenericInstantiation.instantiation_bindings /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9996
pragma Warnings (Off, "is not referenced");
 function Generic_Instantiation_P_Instantiation_Bindings
  (Node : Bare_Generic_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Inner_Env_Assoc_Array_Access
is
   Self : Bare_Generic_Instantiation := Bare_Generic_Instantiation (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Instantiation :=
     Internal_Entity_Generic_Instantiation'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Inner_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_797 with Inline_Always;
      Block_Var : Internal_Entity_Generic_Decl;
Fld : Internal_Entity_Generic_Decl;
Fld_1 : Internal_Inner_Env_Assoc_Array_Access;
Let_Result : Internal_Inner_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_797 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_797;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5028 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5026 '<FieldAccess for nonbound_generic_decl at ast.py:10001>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10001
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericInstantiation.nonbound_generic_decl
Fld := Libadalang.Implementation.Generic_Instantiation_P_Nonbound_Generic_Decl (Node => Self);
--# end
--# expr-done 5026
Block_Var := Fld; 
--# expr-start 5027 '<FieldAccess for instantiation_bindings_internal at ast.py:10002>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10002
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericInstantiation.instantiation_bindings_internal
Fld_1 := Libadalang.Implementation.Generic_Instantiation_P_Instantiation_Bindings_Internal (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5027
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 5028
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_797;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_797;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Instantiation_P_Instantiation_Bindings;
--# end
--# property-start GenericInstantiation.instantiation_bindings_internal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10006
pragma Warnings (Off, "is not referenced");
 function Generic_Instantiation_P_Instantiation_Bindings_Internal
  (Node : Bare_Generic_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Inner_Env_Assoc_Array_Access
is
   Self : Bare_Generic_Instantiation := Bare_Generic_Instantiation (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Instantiation :=
     Internal_Entity_Generic_Instantiation'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Inner_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_798 with Inline_Always;
            procedure Finalizer_Scope_2451 with Inline_Always;
            procedure Finalizer_Scope_2453 with Inline_Always;
      Var_Expr : Internal_Entity_Generic_Decl;
Fld : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_1 : Internal_Entity_Generic_Decl;
Fld_2 : Internal_Entity_Generic_Formal_Part;
Fld_3 : Internal_Entity_Assoc_List;
Fld_4 : Internal_Param_Match_Array_Access;
Scope_Result : Internal_Param_Match_Array_Access;
Result_Var : Internal_Param_Match_Array_Access;
Fld_5 : Internal_Single_Actual;
Fld_6 : Internal_Entity_Basic_Assoc;
Fld_7 : Internal_Entity_Expr;
Is_A : Boolean;
Not_Val : Boolean;
Map_Result : Internal_Param_Match_Array_Access;
Fld_8 : Internal_Entity_Defining_Name;
Fld_9 : Internal_Entity_Name;
Fld_10 : Symbol_Type;
Fld_11 : Bare_Anonymous_Expr_Decl_Array_Access;
Get_Result : Bare_Anonymous_Expr_Decl;
Fld_12 : Internal_Single_Actual;
Fld_13 : Internal_Entity_Basic_Assoc;
Fld_14 : Internal_Entity_Expr;
Fld_15 : Bare_Expr;
Cast_Result : Bare_Ada_Node;
Cast_Result_1 : Bare_Ada_Node;
Fld_16 : Internal_Entity_Defining_Name;
Fld_17 : Internal_Entity_Base_Formal_Param_Decl;
Is_A_1 : Boolean;
If_Result : Bare_Ada_Node;
New_Struct : Internal_Metadata;
New_Struct_1 : Internal_Inner_Env_Assoc;
Index_180 : Integer;
Map_Result_1 : Internal_Inner_Env_Assoc_Array_Access;
Dyn_Var_Bind_Result : Internal_Inner_Env_Assoc_Array_Access;
Fld_18 : Boolean;
If_Result_1 : Internal_Inner_Env_Assoc_Array_Access;
Let_Result : Internal_Inner_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_798 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_798;
            procedure Finalizer_Scope_2451 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_2451;
            procedure Finalizer_Scope_2453 is
            begin
                     Dec_Ref (Fld_11);
            end Finalizer_Scope_2453;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Generic_Instantiation_Generic_Instantiation_P_Instantiation_Bindings_Internal,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Generic_Instantiation,
                                As_Bare_Generic_Instantiation => Self);
               Mmz_K.Items (2) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Inner_Env_Assoc_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
      end if;
   --# scope-start
      --# expr-start 5059 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5058 '<If at ast.py:10007>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10007
--# expr-start 5029 '<FieldAccess for is_any_formal at ast.py:10008>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10008
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericInstantiation.is_any_formal
Fld_18 := Libadalang.Implementation.Generic_Instantiation_P_Is_Any_Formal (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5029
if Fld_18 then
   If_Result_1 := No_Internal_Inner_Env_Assoc_Array_Type;
else
   --# expr-start 5057 '<bind at ast.py:10010>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10010
--# expr-start 5030 '<FieldAccess for default_initial_env at ast.py:10011>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10011
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.default_initial_env
Fld := Libadalang.Implementation.Ada_Node_P_Default_Initial_Env (Node => Self);
--# end
--# expr-done 5030
Bound_Env := Fld; Inc_Ref (Bound_Env);
--# expr-start 5056 '<Map at ast.py:10012>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10012
   --# expr-start 5040 '<Filter at ast.py:10012>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10012
--# expr-start 5031 '<FieldAccess for nonbound_generic_decl at ast.py:10012>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10012
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericInstantiation.nonbound_generic_decl
Fld_1 := Libadalang.Implementation.Generic_Instantiation_P_Nonbound_Generic_Decl (Node => Self);
--# end
--# expr-done 5031
Var_Expr := Fld_1;
if Var_Expr /= No_Entity_Generic_Decl then
   --# scope-start
--# expr-start 5034 '<FieldAccess for match_param_list at ast.py:10012>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10012
--# expr-start 5032 '<FieldAccess for formal_part at ast.py:10012>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10012
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Generic_Formal_Part (Node => Var_Expr.Node.Generic_Decl_F_Formal_Part, Info => Var_Expr.Info);
--# expr-done 5032
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 5033 '<FieldAccess for generic_inst_params at ast.py:10013>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10013
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericInstantiation.generic_inst_params'
Fld_3 := Libadalang.Implementation.Dispatcher_Generic_Instantiation_P_Generic_Inst_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5033
--# property-call-start BaseFormalParamHolder.match_param_list
Fld_4 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Match_Param_List (Node => Fld_2.Node, Params => Fld_3, Is_Dottable_Subp => False, Env => Bound_Env, E_Info => Fld_2.Info);
--# end
--# expr-done 5034
Scope_Result := Fld_4;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2451;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Param_Match_Array_Type;
end if;
      Inc_Ref (Result_Var);
      declare
         Map_Result_Vec : Internal_Param_Match_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Param_Match_Array_Access := Result_Var;
      begin
         for Item_181 of
               Collection.Items
         loop
   --# scope-start
               --# bind pm Item_181
         --# expr-start 5039 '<Not at ast.py:10015>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10015
--# expr-start 5038 '<IsA BoxExpr at ast.py:10015>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10015
--# expr-start 5037 '<FieldAccess for expr at ast.py:10015>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10015
--# expr-start 5036 '<FieldAccess for assoc at ast.py:10015>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10015
--# expr-start 5035 '<FieldAccess for actual at ast.py:10015>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10015
Fld_5 := Item_181.Actual;
--# expr-done 5035
Fld_6 := Fld_5.Assoc;
--# expr-done 5036
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_7 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_6.Node, E_Info => Fld_6.Info);
--# end
--# expr-done 5037
Is_A := Fld_7.Node /= null 
and then Fld_7.Node.Kind in Ada_Box_Expr_Range; 
--# expr-done 5038
Not_Val := not (Is_A); 
--# expr-done 5039
         if Not_Val then
         declare
            Item_To_Append : constant Internal_Param_Match := Item_181;
         begin
            Internal_Param_Match_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Param_Match_Array
           (Items_Count => Natural (Internal_Param_Match_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Param_Match_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Param_Match_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Param_Match_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 5040
      declare
         Map_Result_1_Vec : Internal_Inner_Env_Assoc_Vectors.Vector;
      begin
         Index_180 := 0;
      declare
         Collection : constant Internal_Param_Match_Array_Access := Map_Result;
      begin
         for Item_179 of
               Collection.Items
         loop
   --# scope-start
               --# bind pm Item_179
               --# bind i Index_180
      --# expr-start 5055 '<New InnerEnvAssoc at ast.py:10017>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10017
--# expr-start 5043 '<FieldAccess for name_symbol at ast.py:10018>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10018
--# expr-start 5042 '<FieldAccess for name at ast.py:10018>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10018
--# expr-start 5041 '<FieldAccess for formal at ast.py:10018>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10018
Fld_8 := Item_179.Formal;
--# expr-done 5041
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Name (Node => Fld_8.Node.Defining_Name_F_Name, Info => Fld_8.Info);
--# expr-done 5042
      if Fld_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_10 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_9.Node);
--# end
--# expr-done 5043
--# expr-start 5044 '<New Metadata at ast.py:10024>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10024
New_Struct := (Dottable_Subp => False, Primitive => No_Bare_Ada_Node, Primitive_Real_Type => No_Bare_Ada_Node); 
--# expr-done 5044
--# expr-start 5054 '<If at ast.py:10019>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10019
--# expr-start 5047 '<IsA GenericFormalObjDecl at ast.py:10020>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10020
--# expr-start 5046 '<FieldAccess for formal_decl at ast.py:10020>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10020
--# expr-start 5045 '<FieldAccess for formal at ast.py:10020>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10020
Fld_16 := Item_179.Formal;
--# expr-done 5045
      if Fld_16.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_17 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Fld_16.Node, E_Info => Fld_16.Info);
--# end
--# expr-done 5046
Is_A_1 := Fld_17.Node /= null 
and then Fld_17.Node.Kind in Ada_Generic_Formal_Obj_Decl_Range; 
--# expr-done 5047
if Is_A_1 then
--# expr-start 5049 '<at at ast.py:10021>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10021
--# expr-start 5048 '<FieldAccess for actual_expr_decls at ast.py:10021>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10021
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericInstantiation.actual_expr_decls
Fld_11 := Libadalang.Implementation.Generic_Instantiation_F_Actual_Expr_Decls (Node => Ent.Node);
--# end
--# expr-done 5048
Get_Result := Get (Fld_11, Index_180, True); 
--# expr-done 5049
      Cast_Result := Get_Result;
   If_Result := Cast_Result;
else
--# expr-start 5053 '<FieldAccess for node at ast.py:10022>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10022
--# expr-start 5052 '<FieldAccess for expr at ast.py:10022>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10022
--# expr-start 5051 '<FieldAccess for assoc at ast.py:10022>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10022
--# expr-start 5050 '<FieldAccess for actual at ast.py:10022>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10022
Fld_12 := Item_179.Actual;
--# expr-done 5050
Fld_13 := Fld_12.Assoc;
--# expr-done 5051
      if Fld_13.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_14 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_13.Node, E_Info => Fld_13.Info);
--# end
--# expr-done 5052
Fld_15 := Fld_14.Node;
--# expr-done 5053
      Cast_Result_1 := Fld_15;
   If_Result := Cast_Result_1;
end if;
--# expr-done 5054
New_Struct_1 := (Key => Fld_10, Metadata => New_Struct, Value => If_Result); 
--# expr-done 5055
         declare
            Item_To_Append : constant Internal_Inner_Env_Assoc := New_Struct_1;
         begin
            Internal_Inner_Env_Assoc_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2453;
               Index_180 := Index_180 + 1;
         end loop;
      end;
         Map_Result_1 := Create_Internal_Inner_Env_Assoc_Array
           (Items_Count => Natural (Internal_Inner_Env_Assoc_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Inner_Env_Assoc_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Inner_Env_Assoc_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Inner_Env_Assoc_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 5056
Dyn_Var_Bind_Result := Map_Result_1; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 5057
   If_Result_1 := Dyn_Var_Bind_Result;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 5058
Let_Result := If_Result_1; Inc_Ref (Let_Result);
--# end
--# expr-done 5059
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_798;
      if Self /= null then
            Mmz_Val := (Kind => Mmz_Internal_Inner_Env_Assoc_Array_Access,
                        As_Internal_Inner_Env_Assoc_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_798;
               Finalizer_Scope_2451;
               Finalizer_Scope_2453;
         if Self /= null then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Instantiation_P_Instantiation_Bindings_Internal;
--# end
--# property-start '[dispatcher]GenericInstantiation.generic_entity_name' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Generic_Instantiation_P_Generic_Entity_Name
  (Node : Bare_Generic_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Name
is
   Self : Bare_Generic_Instantiation := Bare_Generic_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Generic_Instantiation (Self.Kind) is
               when Ada_Generic_Package_Instantiation =>
                  --# property-call-start GenericPackageInstantiation.generic_entity_name
                  Property_Result := Generic_Package_Instantiation_P_Generic_Entity_Name
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Subp_Instantiation =>
                  --# property-call-start GenericSubpInstantiation.generic_entity_name
                  Property_Result := Generic_Subp_Instantiation_P_Generic_Entity_Name
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Generic_Instantiation_P_Generic_Entity_Name;
--# end
--# property-start '[dispatcher]GenericInstantiation.generic_inst_params' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Generic_Instantiation_P_Generic_Inst_Params
  (Node : Bare_Generic_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Assoc_List
is
   Self : Bare_Generic_Instantiation := Bare_Generic_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Assoc_List;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Generic_Instantiation (Self.Kind) is
               when Ada_Generic_Package_Instantiation =>
                  --# property-call-start GenericPackageInstantiation.generic_inst_params
                  Property_Result := Generic_Package_Instantiation_P_Generic_Inst_Params
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Subp_Instantiation =>
                  --# property-call-start GenericSubpInstantiation.generic_inst_params
                  Property_Result := Generic_Subp_Instantiation_P_Generic_Inst_Params
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Generic_Instantiation_P_Generic_Inst_Params;
--# end
--# property-start GenericInstantiation.is_any_formal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10043
pragma Warnings (Off, "is not referenced");
 function Generic_Instantiation_P_Is_Any_Formal
  (Node : Bare_Generic_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Generic_Instantiation := Bare_Generic_Instantiation (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Instantiation :=
     Internal_Entity_Generic_Instantiation'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Basic_Assoc;
Fld : Internal_Entity_Assoc_List;
Fld_1 : Bare_Assoc_List;
Fld_2 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_3 : Internal_Entity_Expr;
Scope_Result : Internal_Entity_Expr;
Result_Var : Internal_Entity_Expr;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5062 '<IsA BoxExpr at ast.py:10044>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10044
--# expr-start 5060 '<FieldAccess for generic_inst_params at ast.py:10044>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10044
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericInstantiation.generic_inst_params'
Fld := Libadalang.Implementation.Dispatcher_Generic_Instantiation_P_Generic_Inst_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5060
Fld_1 := Fld.Node;
Get_Result := Get (Fld_1, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Get_Result); 
   If_Result := As_Entity;
end if;
Var_Expr := If_Result;
if Var_Expr /= No_Entity_Basic_Assoc then
   --# scope-start
--# expr-start 5061 '<FieldAccess for expr at ast.py:10044>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10044
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_3 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 5061
Scope_Result := Fld_3;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Expr;
end if;
Is_A := Result_Var.Node /= null 
and then Result_Var.Node.Kind in Ada_Box_Expr_Range; 
--# expr-done 5062
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Instantiation_P_Is_Any_Formal;
--# end
--# property-start GenericInstantiation.nonbound_generic_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10047
pragma Warnings (Off, "is not referenced");
 function Generic_Instantiation_P_Nonbound_Generic_Decl
  (Node : Bare_Generic_Instantiation
  )
   return Internal_Entity_Generic_Decl
is
   Self : Bare_Generic_Instantiation := Bare_Generic_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Generic_Decl;
            procedure Finalizer_Scope_802 with Inline_Always;
      Var_Expr : Internal_Entity_Basic_Decl;
Var_Expr_1 : Internal_Entity;
Match_0 : Internal_Entity_Body_Node;
Match_1 : Internal_Entity_Generic_Renaming_Decl;
Match_2 : Internal_Entity_Basic_Decl;
Match_3 : Internal_Entity;
As_Entity : Internal_Entity_Generic_Instantiation;
Is_Null : Boolean;
If_Result : Internal_Entity_Generic_Instantiation;
Fld : Internal_Entity_Name;
Cast_Expr : Bare_Generic_Instantiation;
Cast_Result : Bare_Ada_Node;
Fld_1 : Internal_Entity_Array_Access;
Fld_2 : Boolean;
Find_Result : Internal_Entity;
Bound_Imprecise_Fallback : Boolean;
Fld_3 : Internal_Entity_Basic_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Basic_Decl;
Fld_4 : Internal_Entity_Generic_Decl;
Match_Prefix : Internal_Entity;
Cast_Result_1 : Internal_Entity_Body_Node;
Let_Result : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Cast_Result_2 : Internal_Entity_Generic_Renaming_Decl;
Cast_Result_3 : Internal_Entity_Basic_Decl;
Let_Result_1 : Internal_Entity_Basic_Decl;
Scope_Result_1 : Internal_Entity_Basic_Decl;
Cast_Result_4 : Internal_Entity_Basic_Decl;
Let_Result_2 : Internal_Entity_Basic_Decl;
Scope_Result_2 : Internal_Entity_Basic_Decl;
Cast_Result_5 : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Generic_Decl;
Cast_Result_6 : Internal_Entity_Basic_Decl;
Let_Result_3 : Internal_Entity_Basic_Decl;
Scope_Result_3 : Internal_Entity_Basic_Decl;
Match_Result : Internal_Entity_Basic_Decl;
Scope_Result_4 : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Cast_Expr_2 : Internal_Entity_Basic_Decl;
Cast_Result_7 : Internal_Entity_Generic_Decl;
Scope_Result_5 : Internal_Entity_Generic_Decl;
Result_Var_1 : Internal_Entity_Generic_Decl;
            procedure Finalizer_Scope_802 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_802;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
--# expr-start 5067 '<Find at ast.py:10048>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10048
   --# expr-start 5065 '<FieldAccess for all_env_elements_internal at ast.py:10048>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10048
--# expr-start 5064 '<FieldAccess for generic_entity_name at ast.py:10048>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10048
--# expr-start 5063 '<as_bare_entity at ast.py:10048>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10048
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity_Generic_Instantiation;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 5063
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericInstantiation.generic_entity_name'
Fld := Libadalang.Implementation.Dispatcher_Generic_Instantiation_P_Generic_Entity_Name (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 5064
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# property-call-start Name.all_env_elements_internal
Fld_1 := Libadalang.Implementation.Name_P_All_Env_Elements_Internal (Node => Fld.Node, Seq => True, Seq_From => Cast_Result, Categories => (Inherited_Primitives => False, Nocat => True), E_Info => Fld.Info);
--# end
--# expr-done 5065
      Find_Result := No_Entity;
      declare
         Collection : constant Internal_Entity_Array_Access := Fld_1;
      begin
         for Item_182 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_182
      --# expr-start 5066 '<FieldAccess for has_visibility at ast.py:10052>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10052
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.has_visibility
Fld_2 := Libadalang.Implementation.Ada_Node_P_Has_Visibility (Node => Self, Other_Entity => Item_182);
--# end
--# expr-done 5066
      if Fld_2 then
         Find_Result := Item_182;
      end if;
   --# end
               exit when Fld_2;
         end loop;
      end;
--# expr-done 5067
Var_Expr_1 := Find_Result;
if Var_Expr_1 /= No_Entity then
   --# scope-start
--# expr-start 5071 '<Match at ast.py:10048>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10048
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Var_Expr_1; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Expr_Function .. Ada_Task_Body =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Body_Node
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_1; 
--# bind b Match_0
--# expr-start 5069 '<bind at ast.py:10054>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10054
Bound_Imprecise_Fallback := False; 
--# expr-start 5068 '<FieldAccess for decl_part at ast.py:10054>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10054
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.decl_part
Fld_3 := Libadalang.Implementation.Body_Node_P_Decl_Part (Node => Match_0.Node, Imprecise_Fallback => Bound_Imprecise_Fallback, E_Info => Match_0.Info);
--# end
--# expr-done 5068
Dyn_Var_Bind_Result := Fld_3; 
--# expr-done 5069
Let_Result := Dyn_Var_Bind_Result; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Generic_Package_Renaming_Decl .. Ada_Generic_Subp_Renaming_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Generic_Renaming_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_2; 
--# bind rd Match_1
--# expr-start 5070 '<FieldAccess for resolve at ast.py:10055>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10055
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericRenamingDecl.resolve
Fld_4 := Libadalang.Implementation.Generic_Renaming_Decl_P_Resolve (Node => Match_1.Node, E_Info => Match_1.Info);
--# end
--# expr-done 5070
      Cast_Result_3 := Create_Internal_Entity_Basic_Decl
        (Node => Fld_4.Node,
         Info => Fld_4.Info);
Let_Result_1 := Cast_Result_3; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abstract_State_Decl .. Ada_Synthetic_Subp_Decl | Ada_Entry_Index_Spec .. Ada_Generic_Subp_Instantiation | Ada_Label_Decl .. Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Basic_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_4; 
--# bind d Match_2
Let_Result_2 := Match_2; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Abort_Absent .. Ada_Param_Assoc | Ada_Case_Stmt_Alternative .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_5; 
--# bind _ Match_3
Cast_Expr_1 := No_Entity_Generic_Decl; 
      Cast_Result_6 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Let_Result_3 := Cast_Result_6; 
--# end
Scope_Result_3 := Let_Result_3;
   --# end
            Match_Result := Scope_Result_3; 
end case;
--# expr-done 5071
Scope_Result_4 := Match_Result;
   --# end
   Result_Var := Scope_Result_4;
else
   Result_Var := No_Entity_Basic_Decl;
end if;
Var_Expr := Result_Var;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 5072 '<Cast to GenericDecl at ast.py:10048>' Cast_Result_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10048
Cast_Expr_2 := Var_Expr; 
   if Cast_Expr_2.Node = null
      or else Cast_Expr_2.Node.Kind in Ada_Generic_Decl
   then
      Cast_Result_7 := Create_Internal_Entity_Generic_Decl
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
   else
         Cast_Result_7 := No_Entity_Generic_Decl;
   end if;
--# expr-done 5072
Scope_Result_5 := Cast_Result_7;
   --# end
   Result_Var_1 := Scope_Result_5;
else
   Result_Var_1 := No_Entity_Generic_Decl;
end if;
      Property_Result := Result_Var_1;
   --# end
      Finalizer_Scope_802;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_802;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Instantiation_P_Nonbound_Generic_Decl;
--# end
--# property-start '[dispatcher]GenericInstantiation.designated_generic_decl' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Generic_Instantiation_P_Designated_Generic_Decl
  (Node : Bare_Generic_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Generic_Instantiation := Bare_Generic_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Generic_Instantiation (Self.Kind) is
               when Ada_Generic_Package_Instantiation =>
                  --# property-call-start GenericPackageInstantiation.designated_generic_decl
                  Property_Result := Generic_Package_Instantiation_P_Designated_Generic_Decl
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Subp_Instantiation =>
                  --# property-call-start GenericSubpInstantiation.designated_generic_decl
                  Property_Result := Generic_Subp_Instantiation_P_Designated_Generic_Decl
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Generic_Instantiation_P_Designated_Generic_Decl;
--# end
--# property-start GenericInstantiation.actual_expr_decls /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10075
pragma Warnings (Off, "is not referenced");
 function Generic_Instantiation_F_Actual_Expr_Decls
  (Node : Bare_Generic_Instantiation
  )
   return Bare_Anonymous_Expr_Decl_Array_Access
is
   Self : Bare_Generic_Instantiation := Bare_Generic_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Anonymous_Expr_Decl_Array_Access;
            procedure Finalizer_Scope_804 with Inline_Always;
      Match_0 : Bare_Generic_Subp_Instantiation;
Match_1 : Bare_Generic_Package_Instantiation;
Fld : Bare_Assoc_List;
Fld_1 : Bare_Assoc_List;
Match_Prefix : Bare_Generic_Instantiation;
Cast_Result : Bare_Generic_Subp_Instantiation;
Let_Result : Bare_Assoc_List;
Scope_Result : Bare_Assoc_List;
Cast_Result_1 : Bare_Generic_Package_Instantiation;
Let_Result_1 : Bare_Assoc_List;
Scope_Result_1 : Bare_Assoc_List;
Match_Result : Bare_Assoc_List;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_2 : Internal_Entity_Expr;
Fld_3 : Bare_Expr;
New_Node : Bare_Anonymous_Expr_Decl;
Item_183 : Bare_Basic_Assoc;
Map_Result : Bare_Anonymous_Expr_Decl_Array_Access;
Let_Result_2 : Bare_Anonymous_Expr_Decl_Array_Access;
            procedure Finalizer_Scope_804 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result_2);
            end Finalizer_Scope_804;
begin
   --# property-body-start
      if Self.Lf_Present_Actual_Expr_Decls then
         Property_Result := Self.Lf_Stg_Actual_Expr_Decls;
            Inc_Ref (Property_Result);
         return Property_Result;
      end if;
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5081 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 5080 '<Map at ast.py:10076>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10076
   --# expr-start 5075 '<Match at ast.py:10076>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10076
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Self; 
case Ada_Generic_Instantiation (Match_Prefix.Kind) is
         when Ada_Generic_Subp_Instantiation =>
   --# scope-start
--# scope-start
      Cast_Result := Match_Prefix;
Match_0 := Cast_Result; 
--# bind subp Match_0
--# expr-start 5073 '<FieldAccess for params at ast.py:10077>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10077
      if Match_0 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Match_0.Generic_Subp_Instantiation_F_Params;
--# expr-done 5073
Let_Result := Fld; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Generic_Package_Instantiation =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Match_Prefix;
Match_1 := Cast_Result_1; 
--# bind pkg Match_1
--# expr-start 5074 '<FieldAccess for params at ast.py:10078>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10078
      if Match_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Match_1.Generic_Package_Instantiation_F_Params;
--# expr-done 5074
Let_Result_1 := Fld_1; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
end case;
--# expr-done 5075
      declare
         Map_Result_Vec : Bare_Anonymous_Expr_Decl_Vectors.Vector;
      begin
      if Match_Result = null then
      Map_Result := Create_Bare_Anonymous_Expr_Decl_Array (0);
      else
      declare
         Collection : constant Bare_Assoc_List := Match_Result;
      begin
         for Untyped_Item_183 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_183 := Untyped_Item_183; 
   --# scope-start
               --# bind assoc Item_183
      --# expr-start 5079 '<New AnonymousExprDecl at ast.py:10080>' New_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10080
--# expr-start 5078 '<FieldAccess for node at ast.py:10081>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10081
--# expr-start 5077 '<FieldAccess for expr at ast.py:10081>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10081
--# expr-start 5076 '<as_bare_entity at ast.py:10081>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10081
Is_Null := Item_183 = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
As_Entity := (Info => No_Entity_Info, Node => Item_183); 
   If_Result := As_Entity;
end if;
--# expr-done 5076
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 5077
Fld_3 := Fld_2.Node;
--# expr-done 5078
if Is_Foreign_Strict (Self.Self_Env, Self) then
   raise Property_Error with
      "synthetic nodes cannot have foreign lexical envs";
end if;
New_Node := new Root_Node_Record
  (Ada_Anonymous_Expr_Decl);
Initialize
  (Self => New_Node,
   Kind => Ada_Anonymous_Expr_Decl,
   Unit => Self.Unit,
   Token_Start_Index => No_Token_Index,
   Token_End_Index   => No_Token_Index,
   Parent => Self,
   Self_Env => Self.Self_Env);
Register_Destroyable (Self.Unit, New_Node);
   Initialize_Fields_For_Anonymous_Expr_Decl
     (Self => New_Node,
      Anonymous_Expr_Decl_F_Expr => Fld_3);
--# expr-done 5079
         declare
            Item_To_Append : constant Bare_Anonymous_Expr_Decl := New_Node;
         begin
            Bare_Anonymous_Expr_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Bare_Anonymous_Expr_Decl_Array
           (Items_Count => Natural (Bare_Anonymous_Expr_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Bare_Anonymous_Expr_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Bare_Anonymous_Expr_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Bare_Anonymous_Expr_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 5080
Let_Result_2 := Map_Result; Inc_Ref (Let_Result_2);
--# end
--# expr-done 5081
      Property_Result := Let_Result_2;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_804;
      Self.Lf_Present_Actual_Expr_Decls := True;
      Self.Lf_Stg_Actual_Expr_Decls := Property_Result;
         Inc_Ref (Property_Result);
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_804;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Instantiation_F_Actual_Expr_Decls;
--# end
--# property-start GenericInstantiation.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10085
pragma Warnings (Off, "is not referenced");
 function Generic_Instantiation_P_Xref_Entry_Point
  (Node : Bare_Generic_Instantiation
  )
   return Boolean
is
   Self : Bare_Generic_Instantiation := Bare_Generic_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Instantiation_P_Xref_Entry_Point;
--# end
--# property-start GenericInstantiation.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10087
pragma Warnings (Off, "is not referenced");
 function Generic_Instantiation_P_Xref_Equation
  (Node : Bare_Generic_Instantiation
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Generic_Instantiation := Bare_Generic_Instantiation (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Instantiation :=
     Internal_Entity_Generic_Instantiation'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_806 with Inline_Always;
            procedure Finalizer_Scope_2458 with Inline_Always;
            procedure Finalizer_Scope_2459 with Inline_Always;
            procedure Finalizer_Scope_2460 with Inline_Always;
            procedure Finalizer_Scope_2462 with Inline_Always;
            procedure Finalizer_Scope_2463 with Inline_Always;
            procedure Finalizer_Scope_2464 with Inline_Always;
            procedure Finalizer_Scope_2465 with Inline_Always;
            procedure Finalizer_Scope_2466 with Inline_Always;
            procedure Finalizer_Scope_2467 with Inline_Always;
            procedure Finalizer_Scope_2468 with Inline_Always;
      Match_0 : Internal_Entity_Dotted_Name;
Match_1 : Internal_Entity_Name;
Var_Expr : Internal_Entity_Generic_Decl;
Actual_Name : Internal_Entity_Name;
Match_0_1 : Internal_Entity_Type_Decl;
Match_1_1 : Internal_Entity_Formal_Subp_Decl;
Match_2 : Internal_Entity_Object_Decl;
Match_3 : Internal_Entity_Basic_Decl;
Var_Expr_1 : Internal_Entity_Name;
Var_Expr_2 : Bare_Base_Id;
Fld : Internal_Entity_Name;
Fld_1 : Logic_Var;
Fld_2 : Internal_Entity_Generic_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_3 : Internal_Entity_Name;
Fld_4 : Internal_Entity_Name;
Fld_5 : Logic_Equation;
True_Rel : Logic_Equation;
Match_Prefix : Internal_Entity_Name;
Cast_Result_1 : Internal_Entity_Dotted_Name;
Let_Result : Logic_Equation;
Scope_Result : Logic_Equation;
Cast_Result_2 : Internal_Entity_Name;
Let_Result_1 : Logic_Equation;
Scope_Result_1 : Logic_Equation;
Match_Result : Logic_Equation;
And_Pred : Logic_Equation;
True_Rel_1 : Logic_Equation;
Fld_6 : Internal_Entity_Basic_Decl;
Cast_Result_3 : Internal_Entity_Generic_Decl;
Fld_7 : Internal_Entity_Generic_Formal_Part;
Fld_8 : Internal_Entity_Assoc_List;
Fld_9 : Internal_Param_Match_Array_Access;
Scope_Result_2 : Internal_Param_Match_Array_Access;
Result_Var : Internal_Param_Match_Array_Access;
Fld_10 : Internal_Single_Actual;
Fld_11 : Internal_Entity_Basic_Assoc;
Fld_12 : Internal_Entity_Expr;
Is_A : Boolean;
Not_Val : Boolean;
Map_Result : Internal_Param_Match_Array_Access;
Fld_13 : Internal_Single_Actual;
Fld_14 : Internal_Entity_Basic_Assoc;
Fld_15 : Internal_Entity_Expr;
Cast_Result_4 : Internal_Entity_Name;
Fld_16 : Internal_Entity_Defining_Name;
Fld_17 : Internal_Entity_Base_Formal_Param_Decl;
Cast_Result_5 : Internal_Entity_Generic_Formal;
Fld_18 : Internal_Entity_Basic_Decl;
Fld_19 : Logic_Equation;
Fld_20 : Logic_Equation;
Fld_21 : Logic_Var;
Cast_Expr : Internal_Entity_Formal_Subp_Decl;
Cast_Result_6 : Internal_Entity_Basic_Decl;
Pred : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_22 : Internal_Entity_Name;
Fld_23 : Logic_Equation;
Scope_Result_3 : Logic_Equation;
True_Rel_2 : Logic_Equation;
Result_Var_1 : Logic_Equation;
True_Rel_3 : Logic_Equation;
Fld_24 : Internal_Entity_Single_Tok_Node;
Fld_25 : Boolean;
If_Result : Logic_Equation;
Or_Pred : Logic_Equation;
Fld_26 : Internal_Single_Actual;
Fld_27 : Internal_Entity_Basic_Assoc;
Fld_28 : Internal_Entity_Expr;
Fld_29 : Logic_Var;
Fld_30 : Internal_Entity_Base_Type_Decl;
Cast_Result_7 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
Fld_31 : Internal_Single_Actual;
Fld_32 : Internal_Entity_Basic_Assoc;
Fld_33 : Internal_Entity_Expr;
Fld_34 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_35 : Internal_Single_Actual;
Fld_36 : Internal_Entity_Basic_Assoc;
Fld_37 : Internal_Entity_Expr;
Fld_38 : Logic_Equation;
And_Pred_3 : Logic_Equation;
True_Rel_4 : Logic_Equation;
Match_Prefix_1 : Internal_Entity_Basic_Decl;
Cast_Result_8 : Internal_Entity_Type_Decl;
Let_Result_2 : Logic_Equation;
Scope_Result_4 : Logic_Equation;
Cast_Result_9 : Internal_Entity_Formal_Subp_Decl;
Let_Result_3 : Logic_Equation;
Scope_Result_5 : Logic_Equation;
Cast_Result_10 : Internal_Entity_Object_Decl;
Let_Result_4 : Logic_Equation;
Scope_Result_6 : Logic_Equation;
Cast_Result_11 : Internal_Entity_Basic_Decl;
Let_Result_5 : Logic_Equation;
Scope_Result_7 : Logic_Equation;
Match_Result_1 : Logic_Equation;
Fld_39 : Internal_Single_Actual;
Fld_40 : Bare_Base_Id;
Fld_41 : Logic_Var;
Fld_42 : Internal_Entity_Defining_Name;
Fld_43 : Bare_Defining_Name;
As_Entity : Internal_Entity_Defining_Name;
Is_Null : Boolean;
If_Result_1 : Internal_Entity_Defining_Name;
Fld_44 : Internal_Entity_Basic_Decl;
Cast_Result_12 : Internal_Entity;
Bind_Result_2 : Logic_Equation;
Scope_Result_8 : Logic_Equation;
True_Rel_5 : Logic_Equation;
Result_Var_2 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Let_Result_6 : Logic_Equation;
Map_Result_1 : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Fld_45 : Boolean;
If_Result_2 : Logic_Equation;
And_Pred_5 : Logic_Equation;
            procedure Finalizer_Scope_806 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Match_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (True_Rel_1);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (If_Result_2);
                     Dec_Ref (And_Pred_5);
            end Finalizer_Scope_806;
            procedure Finalizer_Scope_2458 is
            begin
                     Dec_Ref (Fld_5);
            end Finalizer_Scope_2458;
            procedure Finalizer_Scope_2459 is
            begin
                     Dec_Ref (True_Rel);
            end Finalizer_Scope_2459;
            procedure Finalizer_Scope_2460 is
            begin
                     Dec_Ref (Fld_9);
            end Finalizer_Scope_2460;
            procedure Finalizer_Scope_2462 is
            begin
                     Dec_Ref (Let_Result_2);
                     Dec_Ref (Scope_Result_4);
                     Dec_Ref (Let_Result_3);
                     Dec_Ref (Scope_Result_5);
                     Dec_Ref (Let_Result_4);
                     Dec_Ref (Scope_Result_6);
                     Dec_Ref (Let_Result_5);
                     Dec_Ref (Scope_Result_7);
                     Dec_Ref (Match_Result_1);
                     Dec_Ref (Scope_Result_8);
                     Dec_Ref (True_Rel_5);
                     Dec_Ref (Result_Var_2);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Let_Result_6);
            end Finalizer_Scope_2462;
            procedure Finalizer_Scope_2463 is
            begin
                     Dec_Ref (Fld_19);
            end Finalizer_Scope_2463;
            procedure Finalizer_Scope_2464 is
            begin
                     Dec_Ref (Fld_20);
                     Dec_Ref (Pred);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (True_Rel_2);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (True_Rel_3);
                     Dec_Ref (If_Result);
                     Dec_Ref (Or_Pred);
            end Finalizer_Scope_2464;
            procedure Finalizer_Scope_2465 is
            begin
                     Dec_Ref (Fld_23);
            end Finalizer_Scope_2465;
            procedure Finalizer_Scope_2466 is
            begin
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Fld_34);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Fld_38);
                     Dec_Ref (And_Pred_3);
            end Finalizer_Scope_2466;
            procedure Finalizer_Scope_2467 is
            begin
                     Dec_Ref (True_Rel_4);
            end Finalizer_Scope_2467;
            procedure Finalizer_Scope_2468 is
            begin
                     Dec_Ref (Bind_Result_2);
            end Finalizer_Scope_2468;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5153 '<And at ast.py:10088>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10088
--# expr-start 5090 '<And at ast.py:10088>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10088
--# expr-start 5085 '<Bind at ast.py:10088>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10088
--# expr-start 5083 '<FieldAccess for ref_var at ast.py:10088>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10088
--# expr-start 5082 '<FieldAccess for generic_entity_name at ast.py:10088>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10088
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericInstantiation.generic_entity_name'
Fld := Libadalang.Implementation.Dispatcher_Generic_Instantiation_P_Generic_Entity_Name (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5082
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld.Node);
--# end
--# expr-done 5083
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 5084 '<FieldAccess for nonbound_generic_decl at ast.py:10089>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10089
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericInstantiation.nonbound_generic_decl
Fld_2 := Libadalang.Implementation.Generic_Instantiation_P_Nonbound_Generic_Decl (Node => Ent.Node);
--# end
--# expr-done 5084
      Cast_Result := Create_Internal_Entity
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Bind_Result := Solver.Create_Assign (Fld_1, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10088") else null)); 
--# expr-done 5085
--# expr-start 5089 '<Match at ast.py:10090>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10090
--# expr-start 5086 '<FieldAccess for generic_entity_name at ast.py:10090>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10090
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericInstantiation.generic_entity_name'
Fld_3 := Libadalang.Implementation.Dispatcher_Generic_Instantiation_P_Generic_Entity_Name (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5086
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld_3; 
case Ada_Name (Match_Prefix.Node.Kind) is
         when Ada_Dotted_Name =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Dotted_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_1; 
--# bind dn Match_0
--# expr-start 5088 '<FieldAccess for xref_no_overloading at ast.py:10091>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10091
--# expr-start 5087 '<FieldAccess for prefix at ast.py:10091>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10091
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Name (Node => Match_0.Node.Dotted_Name_F_Prefix, Info => Match_0.Info);
--# expr-done 5087
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_5 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Fld_4.Node, Env => Env, Origin => Origin, E_Info => Fld_4.Info);
--# end
--# expr-done 5088
Let_Result := Fld_5; Inc_Ref (Let_Result);
--# end
Scope_Result := Let_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2458;
            Match_Result := Scope_Result; Inc_Ref (Match_Result);
         when Ada_Attribute_Ref .. Ada_Discrete_Subtype_Name | Ada_End_Name .. Ada_Target_Name =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_2; 
--# bind _ Match_1
True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10092") else null)); 
Let_Result_1 := True_Rel; Inc_Ref (Let_Result_1);
--# end
Scope_Result_1 := Let_Result_1;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2459;
            Match_Result := Scope_Result_1; Inc_Ref (Match_Result);
end case;
--# expr-done 5089
And_Pred := Create_And (Bind_Result, Match_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10088") else null)); 
--# expr-done 5090
--# expr-start 5152 '<If at ast.py:10093>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10093
--# expr-start 5091 '<FieldAccess for is_any_formal at ast.py:10094>' Fld_45 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10094
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericInstantiation.is_any_formal
Fld_45 := Libadalang.Implementation.Generic_Instantiation_P_Is_Any_Formal (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5091
if Fld_45 then
   True_Rel_1 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10095") else null)); 
   If_Result_2 := True_Rel_1;
else
   --# expr-start 5151 '<LogicAll at ast.py:10097>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10097
--# expr-start 5150 '<Map at ast.py:10097>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10097
   --# expr-start 5102 '<Filter at ast.py:10097>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10097
--# expr-start 5093 '<Cast to GenericDecl at ast.py:10097>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10097
--# expr-start 5092 '<FieldAccess for designated_generic_decl at ast.py:10097>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10097
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericInstantiation.designated_generic_decl'
Fld_6 := Libadalang.Implementation.Dispatcher_Generic_Instantiation_P_Designated_Generic_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5092
   if Fld_6.Node = null
      or else Fld_6.Node.Kind in Ada_Generic_Decl
   then
      Cast_Result_3 := Create_Internal_Entity_Generic_Decl
        (Node => Fld_6.Node,
         Info => Fld_6.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 5093
Var_Expr := Cast_Result_3;
if Var_Expr /= No_Entity_Generic_Decl then
   --# scope-start
--# expr-start 5096 '<FieldAccess for match_param_list at ast.py:10097>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10097
--# expr-start 5094 '<FieldAccess for formal_part at ast.py:10097>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10097
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Generic_Formal_Part (Node => Var_Expr.Node.Generic_Decl_F_Formal_Part, Info => Var_Expr.Info);
--# expr-done 5094
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 5095 '<FieldAccess for generic_inst_params at ast.py:10099>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10099
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericInstantiation.generic_inst_params'
Fld_8 := Libadalang.Implementation.Dispatcher_Generic_Instantiation_P_Generic_Inst_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5095
--# property-call-start BaseFormalParamHolder.match_param_list
Fld_9 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Match_Param_List (Node => Fld_7.Node, Params => Fld_8, Is_Dottable_Subp => False, Env => Env, E_Info => Fld_7.Info);
--# end
--# expr-done 5096
Scope_Result_2 := Fld_9;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2460;
   Result_Var := Scope_Result_2;
else
   Result_Var := No_Internal_Param_Match_Array_Type;
end if;
      Inc_Ref (Result_Var);
      declare
         Map_Result_Vec : Internal_Param_Match_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Param_Match_Array_Access := Result_Var;
      begin
         for Item_185 of
               Collection.Items
         loop
   --# scope-start
               --# bind pm Item_185
         --# expr-start 5101 '<Not at ast.py:10101>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10101
--# expr-start 5100 '<IsA BoxExpr at ast.py:10101>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10101
--# expr-start 5099 '<FieldAccess for expr at ast.py:10101>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10101
--# expr-start 5098 '<FieldAccess for assoc at ast.py:10101>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10101
--# expr-start 5097 '<FieldAccess for actual at ast.py:10101>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10101
Fld_10 := Item_185.Actual;
--# expr-done 5097
Fld_11 := Fld_10.Assoc;
--# expr-done 5098
      if Fld_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_12 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_11.Node, E_Info => Fld_11.Info);
--# end
--# expr-done 5099
Is_A := Fld_12.Node /= null 
and then Fld_12.Node.Kind in Ada_Box_Expr_Range; 
--# expr-done 5100
Not_Val := not (Is_A); 
--# expr-done 5101
         if Not_Val then
         declare
            Item_To_Append : constant Internal_Param_Match := Item_185;
         begin
            Internal_Param_Match_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Param_Match_Array
           (Items_Count => Natural (Internal_Param_Match_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Param_Match_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Param_Match_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Param_Match_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 5102
      declare
         Map_Result_1_Vec : Logic_Equation_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Param_Match_Array_Access := Map_Result;
      begin
         for Item_184 of
               Collection.Items
         loop
   --# scope-start
               --# bind pm Item_184
      --# expr-start 5149 '<Let at ast.py:10102>' Let_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10102
--# scope-start
--# expr-start 5106 '<Cast to Name at ast.py:10103>' Cast_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10103
--# expr-start 5105 '<FieldAccess for expr at ast.py:10103>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10103
--# expr-start 5104 '<FieldAccess for assoc at ast.py:10103>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10103
--# expr-start 5103 '<FieldAccess for actual at ast.py:10103>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10103
Fld_13 := Item_184.Actual;
--# expr-done 5103
Fld_14 := Fld_13.Assoc;
--# expr-done 5104
      if Fld_14.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_15 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_14.Node, E_Info => Fld_14.Info);
--# end
--# expr-done 5105
   if Fld_15.Node = null
      or else Fld_15.Node.Kind in Ada_Name
   then
      Cast_Result_4 := Create_Internal_Entity_Name
        (Node => Fld_15.Node,
         Info => Fld_15.Info);
   else
         Cast_Result_4 := No_Entity_Name;
   end if;
--# expr-done 5106
Actual_Name := Cast_Result_4; 
--# bind actual_name Actual_Name
--# expr-start 5148 '<And at ast.py:10104>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10104
--# expr-start 5139 '<Match at ast.py:10104>' Match_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10104
--# expr-start 5110 '<FieldAccess for decl at ast.py:10104>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10104
--# expr-start 5109 '<Cast to GenericFormal at ast.py:10104>' Cast_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10104
--# expr-start 5108 '<FieldAccess for formal_decl at ast.py:10104>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10104
--# expr-start 5107 '<FieldAccess for formal at ast.py:10104>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10104
Fld_16 := Item_184.Formal;
--# expr-done 5107
      if Fld_16.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_17 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Fld_16.Node, E_Info => Fld_16.Info);
--# end
--# expr-done 5108
   if Fld_17.Node = null
      or else Fld_17.Node.Kind in Ada_Generic_Formal
   then
      Cast_Result_5 := Create_Internal_Entity_Generic_Formal
        (Node => Fld_17.Node,
         Info => Fld_17.Info);
   else
         Cast_Result_5 := No_Entity_Generic_Formal;
   end if;
--# expr-done 5109
      if Cast_Result_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Create_Internal_Entity_Basic_Decl (Node => Cast_Result_5.Node.Generic_Formal_F_Decl, Info => Cast_Result_5.Info);
--# expr-done 5110
      if Fld_18.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix_1 := Fld_18; 
case Ada_Basic_Decl (Match_Prefix_1.Node.Kind) is
         when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_8 := Create_Internal_Entity_Type_Decl
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_0_1 := Cast_Result_8; 
--# bind _ Match_0_1
--# expr-start 5111 '<FieldAccess for xref_no_overloading at ast.py:10105>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10105
      if Actual_Name.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_19 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Actual_Name.Node, Env => Env, Origin => Origin, E_Info => Actual_Name.Info);
--# end
--# expr-done 5111
Let_Result_2 := Fld_19; Inc_Ref (Let_Result_2);
--# end
Scope_Result_4 := Let_Result_2;
   Inc_Ref (Scope_Result_4);
   --# end
      Finalizer_Scope_2463;
            Match_Result_1 := Scope_Result_4; Inc_Ref (Match_Result_1);
         when Ada_Abstract_Formal_Subp_Decl .. Ada_Concrete_Formal_Subp_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_9 := Create_Internal_Entity_Formal_Subp_Decl
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_1_1 := Cast_Result_9; 
--# bind subp_decl Match_1_1
--# expr-start 5122 '<Or at ast.py:10108>' Or_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10108
--# expr-start 5116 '<And at ast.py:10109>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10109
--# expr-start 5112 '<FieldAccess for xref_no_overloading at ast.py:10109>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10109
      if Actual_Name.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_20 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Actual_Name.Node, All_Els => True, Env => Env, Origin => Origin, E_Info => Actual_Name.Info);
--# end
--# expr-done 5112
--# expr-start 5115 '<Predicate on BasicDecl.subp_decl_match_signature at ast.py:10110>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10110
--# expr-start 5113 '<FieldAccess for ref_var at ast.py:10111>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10111
      if Actual_Name.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_21 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Actual_Name.Node);
--# end
--# expr-done 5113
Fld_21.Value := No_Entity;
Entity_Vars.Reset (Fld_21);
--# expr-start 5114 '<Cast to BasicDecl at ast.py:10112>' Cast_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10112
Cast_Expr := Match_1_1; 
      Cast_Result_6 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
--# expr-done 5114
Pred := Solver.Create_Predicate (Fld_21, Create_Basic_Decl_P_Subp_Decl_Match_Signature_0_Predicate (Cast_Result_6)); 
--# expr-done 5115
And_Pred_1 := Create_And (Fld_20, Pred, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10109") else null)); 
--# expr-done 5116
--# expr-start 5121 '<If at ast.py:10113>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10113
--# expr-start 5118 '<FieldAccess for is_operator_name at ast.py:10117>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10117
--# expr-start 5117 '<FieldAccess for relative_name at ast.py:10117>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10117
      if Actual_Name.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.relative_name'
Fld_24 := Libadalang.Implementation.Dispatcher_Name_P_Relative_Name (Node => Actual_Name.Node, E_Info => Actual_Name.Info);
--# end
--# expr-done 5117
      if Fld_24.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_operator_name
Fld_25 := Libadalang.Implementation.Name_P_Is_Operator_Name (Node => Fld_24.Node, E_Info => Fld_24.Info);
--# end
--# expr-done 5118
if Fld_25 then
--# expr-start 5119 '<FieldAccess for base_name at ast.py:10118>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10118
      if Actual_Name.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.base_name'
Fld_22 := Libadalang.Implementation.Dispatcher_Name_P_Base_Name (Node => Actual_Name.Node, E_Info => Actual_Name.Info);
--# end
--# expr-done 5119
Var_Expr_1 := Fld_22;
if Var_Expr_1 /= No_Entity_Name then
   --# scope-start
   --# bind e Var_Expr_1
--# expr-start 5120 '<FieldAccess for xref_no_overloading at ast.py:10119>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10119
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_23 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Var_Expr_1.Node, Env => Env, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 5120
Scope_Result_3 := Fld_23;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_2465;
   Result_Var_1 := Scope_Result_3;
else
   True_Rel_2 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10120") else null)); 
   Result_Var_1 := True_Rel_2;
end if;
      Inc_Ref (Result_Var_1);
   If_Result := Result_Var_1;
else
   True_Rel_3 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10122") else null)); 
   If_Result := True_Rel_3;
end if;
      Inc_Ref (If_Result);
--# expr-done 5121
Or_Pred := Create_Or (And_Pred_1, If_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10108") else null)); 
--# expr-done 5122
Let_Result_3 := Or_Pred; Inc_Ref (Let_Result_3);
--# end
Scope_Result_5 := Let_Result_3;
   Inc_Ref (Scope_Result_5);
   --# end
      Finalizer_Scope_2464;
            Match_Result_1 := Scope_Result_5; Inc_Ref (Match_Result_1);
         when Ada_Object_Decl .. Ada_No_Type_Object_Renaming_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_10 := Create_Internal_Entity_Object_Decl
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_2 := Cast_Result_10; 
--# bind obj_decl Match_2
--# expr-start 5138 '<And at ast.py:10127>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10127
--# expr-start 5133 '<And at ast.py:10127>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10127
--# expr-start 5128 '<Bind at ast.py:10127>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10127
--# expr-start 5126 '<FieldAccess for expected_type_var at ast.py:10127>' Fld_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10127
--# expr-start 5125 '<FieldAccess for expr at ast.py:10127>' Fld_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10127
--# expr-start 5124 '<FieldAccess for assoc at ast.py:10127>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10127
--# expr-start 5123 '<FieldAccess for actual at ast.py:10127>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10127
Fld_26 := Item_184.Actual;
--# expr-done 5123
Fld_27 := Fld_26.Assoc;
--# expr-done 5124
      if Fld_27.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_28 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_27.Node, E_Info => Fld_27.Info);
--# end
--# expr-done 5125
      if Fld_28.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_29 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_28.Node);
--# expr-done 5126
Fld_29.Value := No_Entity;
Entity_Vars.Reset (Fld_29);
--# expr-start 5127 '<FieldAccess for expr_type at ast.py:10128>' Fld_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10128
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.expr_type'
Fld_30 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Expr_Type (Node => Match_2.Node, Origin => Origin, E_Info => Match_2.Info);
--# end
--# expr-done 5127
      Cast_Result_7 := Create_Internal_Entity
        (Node => Fld_30.Node,
         Info => Fld_30.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_29, Cast_Result_7, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10127") else null)); 
--# expr-done 5128
--# expr-start 5132 '<FieldAccess for sub_equation at ast.py:10129>' Fld_34 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10129
--# expr-start 5131 '<FieldAccess for expr at ast.py:10129>' Fld_33 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10129
--# expr-start 5130 '<FieldAccess for assoc at ast.py:10129>' Fld_32 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10129
--# expr-start 5129 '<FieldAccess for actual at ast.py:10129>' Fld_31 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10129
Fld_31 := Item_184.Actual;
--# expr-done 5129
Fld_32 := Fld_31.Assoc;
--# expr-done 5130
      if Fld_32.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_33 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_32.Node, E_Info => Fld_32.Info);
--# end
--# expr-done 5131
      if Fld_33.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_34 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_33.Node, Env => Env, Origin => Origin, E_Info => Fld_33.Info);
--# end
--# expr-done 5132
And_Pred_2 := Create_And (Bind_Result_1, Fld_34, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10127") else null)); 
--# expr-done 5133
--# expr-start 5137 '<FieldAccess for matches_expected_assign_type at ast.py:10130>' Fld_38 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10130
--# expr-start 5136 '<FieldAccess for expr at ast.py:10130>' Fld_37 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10130
--# expr-start 5135 '<FieldAccess for assoc at ast.py:10130>' Fld_36 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10130
--# expr-start 5134 '<FieldAccess for actual at ast.py:10130>' Fld_35 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10130
Fld_35 := Item_184.Actual;
--# expr-done 5134
Fld_36 := Fld_35.Assoc;
--# expr-done 5135
      if Fld_36.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_37 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_36.Node, E_Info => Fld_36.Info);
--# end
--# expr-done 5136
      if Fld_37.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_assign_type
Fld_38 := Libadalang.Implementation.Expr_P_Matches_Expected_Assign_Type (Node => Fld_37.Node, Origin => Origin);
--# end
--# expr-done 5137
And_Pred_3 := Create_And (And_Pred_2, Fld_38, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10127") else null)); 
--# expr-done 5138
Let_Result_4 := And_Pred_3; Inc_Ref (Let_Result_4);
--# end
Scope_Result_6 := Let_Result_4;
   Inc_Ref (Scope_Result_6);
   --# end
      Finalizer_Scope_2466;
            Match_Result_1 := Scope_Result_6; Inc_Ref (Match_Result_1);
         when Ada_Abstract_State_Decl .. Ada_Single_Task_Type_Decl | Ada_Abstract_Subp_Decl | Ada_Subp_Decl .. Ada_Number_Decl | Ada_Package_Renaming_Decl .. Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_11 := Create_Internal_Entity_Basic_Decl
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_3 := Cast_Result_11; 
--# bind _ Match_3
True_Rel_4 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10132") else null)); 
Let_Result_5 := True_Rel_4; Inc_Ref (Let_Result_5);
--# end
Scope_Result_7 := Let_Result_5;
   Inc_Ref (Scope_Result_7);
   --# end
      Finalizer_Scope_2467;
            Match_Result_1 := Scope_Result_7; Inc_Ref (Match_Result_1);
end case;
--# expr-done 5139
--# expr-start 5141 '<FieldAccess for name at ast.py:10133>' Fld_40 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10133
--# expr-start 5140 '<FieldAccess for actual at ast.py:10133>' Fld_39 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10133
Fld_39 := Item_184.Actual;
--# expr-done 5140
Fld_40 := Fld_39.Name;
--# expr-done 5141
Var_Expr_2 := Fld_40;
if Var_Expr_2 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind n Var_Expr_2
--# expr-start 5147 '<Bind at ast.py:10135>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10135
--# expr-start 5142 '<FieldAccess for ref_var at ast.py:10135>' Fld_41 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10135
      if Var_Expr_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_41 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Var_Expr_2);
--# end
--# expr-done 5142
Fld_41.Value := No_Entity;
Entity_Vars.Reset (Fld_41);
--# expr-start 5146 '<FieldAccess for basic_decl at ast.py:10136>' Fld_44 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10136
--# expr-start 5145 '<as_bare_entity at ast.py:10136>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10136
--# expr-start 5144 '<FieldAccess for node at ast.py:10136>' Fld_43 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10136
--# expr-start 5143 '<FieldAccess for formal at ast.py:10136>' Fld_42 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10136
Fld_42 := Item_184.Formal;
--# expr-done 5143
Fld_43 := Fld_42.Node;
--# expr-done 5144
Is_Null := Fld_43 = null; 
if Is_Null then
   If_Result_1 := No_Entity_Defining_Name;
else
As_Entity := (Info => No_Entity_Info, Node => Fld_43); 
   If_Result_1 := As_Entity;
end if;
--# expr-done 5145
      if If_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld_44 := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => If_Result_1.Node, E_Info => If_Result_1.Info);
--# end
--# expr-done 5146
      Cast_Result_12 := Create_Internal_Entity
        (Node => Fld_44.Node,
         Info => Fld_44.Info);
Bind_Result_2 := Solver.Create_Assign (Fld_41, Cast_Result_12, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10135") else null)); 
--# expr-done 5147
Scope_Result_8 := Bind_Result_2;
   Inc_Ref (Scope_Result_8);
   --# end
      Finalizer_Scope_2468;
   Result_Var_2 := Scope_Result_8;
else
   True_Rel_5 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10137") else null)); 
   Result_Var_2 := True_Rel_5;
end if;
      Inc_Ref (Result_Var_2);
And_Pred_4 := Create_And (Match_Result_1, Result_Var_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10104") else null)); 
--# expr-done 5148
Let_Result_6 := And_Pred_4; Inc_Ref (Let_Result_6);
--# end
--# expr-done 5149
         declare
            Item_To_Append : constant Logic_Equation := Let_Result_6;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2462;
         end loop;
      end;
         Map_Result_1 := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_1_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 5150
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result_1.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10097") else null)); 
--# expr-done 5151
   If_Result_2 := Logic_Boolean_Op;
end if;
      Inc_Ref (If_Result_2);
--# expr-done 5152
And_Pred_5 := Create_And (And_Pred, If_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:10088") else null)); 
--# expr-done 5153
      Property_Result := And_Pred_5;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_806;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_806;
               Finalizer_Scope_2458;
               Finalizer_Scope_2459;
               Finalizer_Scope_2460;
               Finalizer_Scope_2462;
               Finalizer_Scope_2463;
               Finalizer_Scope_2464;
               Finalizer_Scope_2465;
               Finalizer_Scope_2466;
               Finalizer_Scope_2467;
               Finalizer_Scope_2468;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Instantiation_P_Xref_Equation;
--# end
--# property-start GenericInstantiation.inst_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10144
pragma Warnings (Off, "is not referenced");
 function Generic_Instantiation_P_Inst_Params
  (Node : Bare_Generic_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Param_Actual_Array_Access
is
   Self : Bare_Generic_Instantiation := Bare_Generic_Instantiation (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Instantiation :=
     Internal_Entity_Generic_Instantiation'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Param_Actual_Array_Access;
            procedure Finalizer_Scope_807 with Inline_Always;
            procedure Finalizer_Scope_2470 with Inline_Always;
            procedure Finalizer_Scope_2471 with Inline_Always;
            procedure Finalizer_Scope_2472 with Inline_Always;
            procedure Finalizer_Scope_2473 with Inline_Always;
            procedure Finalizer_Scope_2475 with Inline_Always;
            procedure Finalizer_Scope_2477 with Inline_Always;
      Block_Var : Internal_Entity_Assoc_List;
Var_Expr : Internal_Entity_Generic_Decl;
Match_0 : Internal_Entity_Generic_Formal_Package;
Match_1 : Internal_Entity_Generic_Formal_Type_Decl;
Match_2 : Internal_Entity_Generic_Formal_Obj_Decl;
Match_3 : Internal_Entity_Generic_Formal_Subp_Decl;
Match_4 : Internal_Entity;
Subp : Internal_Entity_Formal_Subp_Decl;
Var_Expr_1 : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Assoc_List;
Fld_1 : Internal_Entity_Generic_Decl;
Fld_2 : Internal_Entity_Generic_Formal_Part;
Fld_3 : Internal_Entity_Ada_Node_List;
Scope_Result : Internal_Entity_Ada_Node_List;
Result_Var : Internal_Entity_Ada_Node_List;
Fld_4 : Bare_Ada_Node_List;
Fld_5 : Internal_Entity_Info;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Fld_6 : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity_Generic_Package_Instantiation;
Fld_7 : Internal_Entity_Defining_Name;
New_Struct : Internal_Param_Actual;
Singleton : Internal_Param_Actual_Array_Access;
Fld_8 : Internal_Entity_Basic_Decl;
Cast_Result_1 : Internal_Entity_Type_Decl;
Fld_9 : Internal_Entity_Defining_Name;
New_Struct_1 : Internal_Param_Actual;
Singleton_1 : Internal_Param_Actual_Array_Access;
Fld_10 : Internal_Entity_Basic_Decl;
Cast_Result_2 : Internal_Entity_Object_Decl;
Fld_11 : Internal_Entity_Defining_Name_List;
Fld_12 : Bare_Defining_Name_List;
Fld_13 : Internal_Entity_Info;
As_Entity_1 : Internal_Entity_Defining_Name;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Defining_Name;
Fld_14 : Internal_Entity_Basic_Decl;
Cast_Result_3 : Internal_Entity_Object_Decl;
Fld_15 : Internal_Entity_Expr;
New_Struct_2 : Internal_Param_Actual;
Item_189 : Internal_Entity_Defining_Name;
Bare_Item_189 : Bare_Defining_Name;
Map_Result : Internal_Param_Actual_Array_Access;
Fld_16 : Internal_Entity_Basic_Decl;
Cast_Result_4 : Internal_Entity_Formal_Subp_Decl;
Fld_17 : Internal_Entity_Defining_Name;
Fld_18 : Internal_Entity_Basic_Decl;
Fld_19 : Internal_Entity_Defining_Name;
Scope_Result_1 : Internal_Entity_Defining_Name;
Result_Var_1 : Internal_Entity_Defining_Name;
Fld_20 : Internal_Entity_Expr;
Cast_Result_5 : Internal_Entity_Expr;
Fld_21 : Internal_Entity_Expr;
Is_A : Boolean;
If_Result_2 : Internal_Entity_Expr;
New_Struct_3 : Internal_Param_Actual;
Let_Result : Internal_Param_Actual;
Singleton_2 : Internal_Param_Actual_Array_Access;
Singleton_3 : Internal_Param_Actual_Array_Access;
Match_Prefix : Internal_Entity;
Cast_Result_6 : Internal_Entity_Generic_Formal_Package;
Let_Result_1 : Internal_Param_Actual_Array_Access;
Scope_Result_2 : Internal_Param_Actual_Array_Access;
Cast_Result_7 : Internal_Entity_Generic_Formal_Type_Decl;
Let_Result_2 : Internal_Param_Actual_Array_Access;
Scope_Result_3 : Internal_Param_Actual_Array_Access;
Cast_Result_8 : Internal_Entity_Generic_Formal_Obj_Decl;
Let_Result_3 : Internal_Param_Actual_Array_Access;
Scope_Result_4 : Internal_Param_Actual_Array_Access;
Cast_Result_9 : Internal_Entity_Generic_Formal_Subp_Decl;
Let_Result_4 : Internal_Param_Actual_Array_Access;
Scope_Result_5 : Internal_Param_Actual_Array_Access;
Cast_Result_10 : Internal_Entity;
Let_Result_5 : Internal_Param_Actual_Array_Access;
Scope_Result_6 : Internal_Param_Actual_Array_Access;
Match_Result : Internal_Param_Actual_Array_Access;
Item_188 : Internal_Entity;
Bare_Item_188 : Bare_Ada_Node;
Map_Result_1 : Internal_Param_Actual_Array_Access;
Fld_22 : Internal_Entity_Defining_Name;
Fld_23 : Internal_Entity_Defining_Name;
Fld_24 : Internal_Entity_Expr;
Fld_25 : Internal_Entity_Expr;
New_Struct_4 : Internal_Param_Actual;
Index_187 : Integer;
Map_Result_2 : Internal_Param_Actual_Array_Access;
Let_Result_6 : Internal_Param_Actual_Array_Access;
            procedure Finalizer_Scope_807 is
            begin
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Map_Result_2);
                     Dec_Ref (Let_Result_6);
            end Finalizer_Scope_807;
            procedure Finalizer_Scope_2470 is
            begin
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Let_Result_2);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Let_Result_3);
                     Dec_Ref (Scope_Result_4);
                     Dec_Ref (Let_Result_4);
                     Dec_Ref (Scope_Result_5);
                     Dec_Ref (Let_Result_5);
                     Dec_Ref (Scope_Result_6);
                     Dec_Ref (Match_Result);
            end Finalizer_Scope_2470;
            procedure Finalizer_Scope_2471 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2471;
            procedure Finalizer_Scope_2472 is
            begin
                     Dec_Ref (Singleton_1);
            end Finalizer_Scope_2472;
            procedure Finalizer_Scope_2473 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_2473;
            procedure Finalizer_Scope_2475 is
            begin
                     Dec_Ref (Singleton_2);
            end Finalizer_Scope_2475;
            procedure Finalizer_Scope_2477 is
            begin
                     Dec_Ref (Singleton_3);
            end Finalizer_Scope_2477;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5193 '<Block at ???>' Let_Result_6 None
--# scope-start
--# expr-start 5154 '<FieldAccess for generic_inst_params at ast.py:10149>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10149
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericInstantiation.generic_inst_params'
Fld := Libadalang.Implementation.Dispatcher_Generic_Instantiation_P_Generic_Inst_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5154
Block_Var := Fld; 
--# bind ap Block_Var
--# expr-start 5192 '<Map at ast.py:10151>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10151
   --# expr-start 5186 '<Mapcat at ast.py:10151>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10151
--# expr-start 5155 '<FieldAccess for nonbound_generic_decl at ast.py:10151>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10151
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericInstantiation.nonbound_generic_decl
Fld_1 := Libadalang.Implementation.Generic_Instantiation_P_Nonbound_Generic_Decl (Node => Ent.Node);
--# end
--# expr-done 5155
Var_Expr := Fld_1;
if Var_Expr /= No_Entity_Generic_Decl then
   --# scope-start
--# expr-start 5157 '<FieldAccess for decls at ast.py:10151>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10151
--# expr-start 5156 '<FieldAccess for formal_part at ast.py:10151>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10151
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Generic_Formal_Part (Node => Var_Expr.Node.Generic_Decl_F_Formal_Part, Info => Var_Expr.Info);
--# expr-done 5156
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Ada_Node_List (Node => Fld_2.Node.Generic_Formal_Part_F_Decls, Info => Fld_2.Info);
--# expr-done 5157
Scope_Result := Fld_3;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Ada_Node_List;
end if;
Fld_4 := Result_Var.Node;
      declare
         Map_Result_1_Vec : Internal_Param_Actual_Vectors.Vector;
      begin
      if Fld_4 = null then
      Map_Result_1 := Create_Internal_Param_Actual_Array (0);
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_4;
      begin
         for Untyped_Item_188 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_188 := Untyped_Item_188; 
Is_Null := Bare_Item_188 = null; 
if Is_Null then
   If_Result := No_Entity;
else
Fld_5 := Result_Var.Info;
As_Entity := (Info => Fld_5, Node => Bare_Item_188); 
   If_Result := As_Entity;
end if;
                  Item_188 := If_Result; 
   --# scope-start
               --# bind d Item_188
      --# expr-start 5185 '<Match at ast.py:10153>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10153
      if Item_188.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Item_188; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Generic_Formal_Package =>
   --# scope-start
--# scope-start
      Cast_Result_6 := Create_Internal_Entity_Generic_Formal_Package
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_6; 
--# bind t Match_0
--# expr-start 5161 '<New ParamActual at ast.py:10154>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10154
--# expr-start 5160 '<FieldAccess for name at ast.py:10155>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10155
--# expr-start 5159 '<Cast to GenericPackageInstantiation at ast.py:10155>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10155
--# expr-start 5158 '<FieldAccess for decl at ast.py:10155>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10155
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Basic_Decl (Node => Match_0.Node.Generic_Formal_F_Decl, Info => Match_0.Info);
--# expr-done 5158
   if Fld_6.Node = null
      or else Fld_6.Node.Kind in Ada_Generic_Package_Instantiation_Range
   then
      Cast_Result := Create_Internal_Entity_Generic_Package_Instantiation
        (Node => Fld_6.Node,
         Info => Fld_6.Info);
   else
         Cast_Result := No_Entity_Generic_Package_Instantiation;
   end if;
--# expr-done 5159
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Defining_Name (Node => Cast_Result.Node.Generic_Package_Instantiation_F_Name, Info => Cast_Result.Info);
--# expr-done 5160
New_Struct := (Actual => No_Entity_Expr, Param => Fld_7); 
--# expr-done 5161
                Singleton := Create_Internal_Param_Actual_Array (Items_Count => 1);
                Singleton.Items (1) := New_Struct;
Let_Result_1 := Singleton; Inc_Ref (Let_Result_1);
--# end
Scope_Result_2 := Let_Result_1;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2471;
            Match_Result := Scope_Result_2; Inc_Ref (Match_Result);
         when Ada_Generic_Formal_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_7 := Create_Internal_Entity_Generic_Formal_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_7; 
--# bind t Match_1
--# expr-start 5165 '<New ParamActual at ast.py:10158>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10158
--# expr-start 5164 '<FieldAccess for name at ast.py:10159>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10159
--# expr-start 5163 '<Cast to TypeDecl at ast.py:10159>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10159
--# expr-start 5162 '<FieldAccess for decl at ast.py:10159>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10159
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Basic_Decl (Node => Match_1.Node.Generic_Formal_F_Decl, Info => Match_1.Info);
--# expr-done 5162
   if Fld_8.Node = null
      or else Fld_8.Node.Kind in Ada_Type_Decl_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Type_Decl
        (Node => Fld_8.Node,
         Info => Fld_8.Info);
   else
         Cast_Result_1 := No_Entity_Type_Decl;
   end if;
--# expr-done 5163
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Defining_Name (Node => Cast_Result_1.Node.Base_Type_Decl_F_Name, Info => Cast_Result_1.Info);
--# expr-done 5164
New_Struct_1 := (Actual => No_Entity_Expr, Param => Fld_9); 
--# expr-done 5165
                Singleton_1 := Create_Internal_Param_Actual_Array (Items_Count => 1);
                Singleton_1.Items (1) := New_Struct_1;
Let_Result_2 := Singleton_1; Inc_Ref (Let_Result_2);
--# end
Scope_Result_3 := Let_Result_2;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_2472;
            Match_Result := Scope_Result_3; Inc_Ref (Match_Result);
         when Ada_Generic_Formal_Obj_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_8 := Create_Internal_Entity_Generic_Formal_Obj_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_8; 
--# bind o Match_2
--# expr-start 5173 '<Map at ast.py:10163>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10163
   --# expr-start 5168 '<FieldAccess for ids at ast.py:10163>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10163
--# expr-start 5167 '<Cast to ObjectDecl at ast.py:10163>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10163
--# expr-start 5166 '<FieldAccess for decl at ast.py:10163>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10163
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Create_Internal_Entity_Basic_Decl (Node => Match_2.Node.Generic_Formal_F_Decl, Info => Match_2.Info);
--# expr-done 5166
   if Fld_10.Node = null
      or else Fld_10.Node.Kind in Ada_Object_Decl_Range
   then
      Cast_Result_2 := Create_Internal_Entity_Object_Decl
        (Node => Fld_10.Node,
         Info => Fld_10.Info);
   else
         Cast_Result_2 := No_Entity_Object_Decl;
   end if;
--# expr-done 5167
      if Cast_Result_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Create_Internal_Entity_Defining_Name_List (Node => Cast_Result_2.Node.Object_Decl_F_Ids, Info => Cast_Result_2.Info);
--# expr-done 5168
Fld_12 := Fld_11.Node;
      declare
         Map_Result_Vec : Internal_Param_Actual_Vectors.Vector;
      begin
      if Fld_12 = null then
      Map_Result := Create_Internal_Param_Actual_Array (0);
      else
      declare
         Collection : constant Bare_Defining_Name_List := Fld_12;
      begin
         for Untyped_Item_189 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_189 := Untyped_Item_189; 
Is_Null_1 := Bare_Item_189 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Defining_Name;
else
Fld_13 := Fld_11.Info;
As_Entity_1 := (Info => Fld_13, Node => Bare_Item_189); 
   If_Result_1 := As_Entity_1;
end if;
                  Item_189 := If_Result_1; 
   --# scope-start
               --# bind i Item_189
      --# expr-start 5172 '<New ParamActual at ast.py:10164>' New_Struct_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10164
--# expr-start 5171 '<FieldAccess for default_expr at ast.py:10166>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10166
--# expr-start 5170 '<Cast to ObjectDecl at ast.py:10166>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10166
--# expr-start 5169 '<FieldAccess for decl at ast.py:10166>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10166
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Create_Internal_Entity_Basic_Decl (Node => Match_2.Node.Generic_Formal_F_Decl, Info => Match_2.Info);
--# expr-done 5169
   if Fld_14.Node = null
      or else Fld_14.Node.Kind in Ada_Object_Decl_Range
   then
      Cast_Result_3 := Create_Internal_Entity_Object_Decl
        (Node => Fld_14.Node,
         Info => Fld_14.Info);
   else
         Cast_Result_3 := No_Entity_Object_Decl;
   end if;
--# expr-done 5170
      if Cast_Result_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Create_Internal_Entity_Expr (Node => Cast_Result_3.Node.Object_Decl_F_Default_Expr, Info => Cast_Result_3.Info);
--# expr-done 5171
New_Struct_2 := (Actual => Fld_15, Param => Item_189); 
--# expr-done 5172
         declare
            Item_To_Append : constant Internal_Param_Actual := New_Struct_2;
         begin
            Internal_Param_Actual_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Param_Actual_Array
           (Items_Count => Natural (Internal_Param_Actual_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Param_Actual_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Param_Actual_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Param_Actual_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 5173
Let_Result_3 := Map_Result; Inc_Ref (Let_Result_3);
--# end
Scope_Result_4 := Let_Result_3;
   Inc_Ref (Scope_Result_4);
   --# end
      Finalizer_Scope_2473;
            Match_Result := Scope_Result_4; Inc_Ref (Match_Result);
         when Ada_Generic_Formal_Subp_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_9 := Create_Internal_Entity_Generic_Formal_Subp_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_9; 
--# bind sp Match_3
--# expr-start 5184 '<Let at ast.py:10169>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10169
--# scope-start
--# expr-start 5175 '<Cast to FormalSubpDecl at ast.py:10170>' Cast_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10170
--# expr-start 5174 '<FieldAccess for decl at ast.py:10170>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10170
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_16 := Create_Internal_Entity_Basic_Decl (Node => Match_3.Node.Generic_Formal_F_Decl, Info => Match_3.Info);
--# expr-done 5174
   if Fld_16.Node = null
      or else Fld_16.Node.Kind in Ada_Formal_Subp_Decl
   then
      Cast_Result_4 := Create_Internal_Entity_Formal_Subp_Decl
        (Node => Fld_16.Node,
         Info => Fld_16.Info);
   else
         Cast_Result_4 := No_Entity_Formal_Subp_Decl;
   end if;
--# expr-done 5175
Subp := Cast_Result_4; 
--# bind subp Subp
--# expr-start 5183 '<New ParamActual at ast.py:10172>' New_Struct_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10172
--# expr-start 5181 '<If at ast.py:10174>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10174
--# expr-start 5177 '<IsA BoxExpr at ast.py:10175>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10175
--# expr-start 5176 '<FieldAccess for default_expr at ast.py:10175>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10175
      if Subp.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_21 := Create_Internal_Entity_Expr (Node => Subp.Node.Formal_Subp_Decl_F_Default_Expr, Info => Subp.Info);
--# expr-done 5176
Is_A := Fld_21.Node /= null 
and then Fld_21.Node.Kind in Ada_Box_Expr_Range; 
--# expr-done 5177
if Is_A then
--# expr-start 5178 '<FieldAccess for designated_subprogram_from at ast.py:10176>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10176
      if Subp.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start FormalSubpDecl.designated_subprogram_from
Fld_18 := Libadalang.Implementation.Formal_Subp_Decl_P_Designated_Subprogram_From (Node => Subp.Node, Inst => Ent, E_Info => Subp.Info);
--# end
--# expr-done 5178
Var_Expr_1 := Fld_18;
if Var_Expr_1 /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 5179 '<FieldAccess for defining_name at ast.py:10176>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10176
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_19 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 5179
Scope_Result_1 := Fld_19;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Defining_Name;
end if;
      Cast_Result_5 := Create_Internal_Entity_Expr
        (Node => Result_Var_1.Node,
         Info => Result_Var_1.Info);
   If_Result_2 := Cast_Result_5;
else
   --# expr-start 5180 '<FieldAccess for default_expr at ast.py:10179>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10179
      if Subp.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_20 := Create_Internal_Entity_Expr (Node => Subp.Node.Formal_Subp_Decl_F_Default_Expr, Info => Subp.Info);
--# expr-done 5180
   If_Result_2 := Fld_20;
end if;
--# expr-done 5181
--# expr-start 5182 '<FieldAccess for defining_name at ast.py:10173>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10173
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_17 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Match_3.Node, E_Info => Match_3.Info);
--# end
--# expr-done 5182
New_Struct_3 := (Actual => If_Result_2, Param => Fld_17); 
--# expr-done 5183
Let_Result := New_Struct_3; 
--# end
--# expr-done 5184
                Singleton_2 := Create_Internal_Param_Actual_Array (Items_Count => 1);
                Singleton_2.Items (1) := Let_Result;
Let_Result_4 := Singleton_2; Inc_Ref (Let_Result_4);
--# end
Scope_Result_5 := Let_Result_4;
   Inc_Ref (Scope_Result_5);
   --# end
      Finalizer_Scope_2475;
            Match_Result := Scope_Result_5; Inc_Ref (Match_Result);
         when Ada_Abort_Absent .. Ada_Discriminant_Spec | Ada_Param_Spec .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_10 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_4 := Cast_Result_10; 
--# bind _ Match_4
                Singleton_3 := Create_Internal_Param_Actual_Array (Items_Count => 1);
                Singleton_3.Items (1) := No_Param_Actual;
Let_Result_5 := Singleton_3; Inc_Ref (Let_Result_5);
--# end
Scope_Result_6 := Let_Result_5;
   Inc_Ref (Scope_Result_6);
   --# end
      Finalizer_Scope_2477;
            Match_Result := Scope_Result_6; Inc_Ref (Match_Result);
end case;
--# expr-done 5185
         for Item_To_Append of
               Match_Result.Items
         loop
            Internal_Param_Actual_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_2470;
         end loop;
      end;
      end if;
         Map_Result_1 := Create_Internal_Param_Actual_Array
           (Items_Count => Natural (Internal_Param_Actual_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Param_Actual_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Param_Actual_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Param_Actual_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 5186
      declare
         Map_Result_2_Vec : Internal_Param_Actual_Vectors.Vector;
      begin
         Index_187 := 0;
      declare
         Collection : constant Internal_Param_Actual_Array_Access := Map_Result_1;
      begin
         for Item_186 of
               Collection.Items
         loop
   --# scope-start
               --# bind dp Item_186
               --# bind i Index_187
      --# expr-start 5191 '<New ParamActual at ast.py:10187>' New_Struct_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10187
--# expr-start 5189 '<FieldAccess for actual_for_param_at at ast.py:10189>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10189
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 5187 '<FieldAccess for param at ast.py:10189>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10189
Fld_23 := Item_186.Param;
--# expr-done 5187
--# expr-start 5188 '<FieldAccess for actual at ast.py:10189>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10189
Fld_24 := Item_186.Actual;
--# expr-done 5188
--# property-call-start AssocList.actual_for_param_at
Fld_25 := Libadalang.Implementation.Assoc_List_P_Actual_For_Param_At (Node => Block_Var.Node, Param => Fld_23, Pos => Index_187, Default_Expr => Fld_24, E_Info => Block_Var.Info);
--# end
--# expr-done 5189
--# expr-start 5190 '<FieldAccess for param at ast.py:10188>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10188
Fld_22 := Item_186.Param;
--# expr-done 5190
New_Struct_4 := (Actual => Fld_25, Param => Fld_22); 
--# expr-done 5191
         declare
            Item_To_Append : constant Internal_Param_Actual := New_Struct_4;
         begin
            Internal_Param_Actual_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
   --# end
               Index_187 := Index_187 + 1;
         end loop;
      end;
         Map_Result_2 := Create_Internal_Param_Actual_Array
           (Items_Count => Natural (Internal_Param_Actual_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Internal_Param_Actual_Vectors.Get
              (Map_Result_2_Vec,
               I + Internal_Param_Actual_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Internal_Param_Actual_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 5192
Let_Result_6 := Map_Result_2; Inc_Ref (Let_Result_6);
--# end
--# expr-done 5193
      Property_Result := Let_Result_6;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_807;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_807;
               Finalizer_Scope_2470;
               Finalizer_Scope_2471;
               Finalizer_Scope_2472;
               Finalizer_Scope_2473;
               Finalizer_Scope_2475;
               Finalizer_Scope_2477;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Instantiation_P_Inst_Params;
--# end
function Generic_Instantiation_P_Instantiation_Bindings
  (E : Entity
  ) return Internal_Inner_Env_Assoc_Array_Access
is
      E_Info : Internal_Entity_Info :=
         Shed_Rebindings (E.Info, Node_Env (E.Node));
   Result : constant Internal_Inner_Env_Assoc_Array_Access :=
      Generic_Instantiation_P_Instantiation_Bindings (E.Node, E_Info);
begin
      return Result;
end;
   --
   --  Primitives for Bare_Generic_Package_Instantiation
   --
      procedure Generic_Package_Instantiation_Pre_Env_Actions
        (Self            : Bare_Generic_Package_Instantiation;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Dummy : Boolean;
      begin
         Dummy := Libadalang.Implementation.Env_Do_68 (Node => 
Self);
      end;
      declare
         Env : constant Internal_Designated_Env :=
           Libadalang.Implementation.Initial_Env_69 (Node => 
Self);
      begin
         Set_Initial_Env
           (Self,
            State,
            Env,
            DSL_Location => "ast.py:10387");
      end;
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_70 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:10394");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_71 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            Libadalang.Implementation.Env_Names_72 (Node => 
Self);
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      declare
         Dummy : Bare_Compilation_Unit_Array_Access;
      begin
         Dummy := Libadalang.Implementation.Env_Do_73 (Node => 
Self);
         Dec_Ref (Dummy);
      end;
      if Libadalang.Implementation.Ref_Cond_75 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_74 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Use_Package_Name_Designated_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_77 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_76 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Name_Designated_Type_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      end;
      procedure Initialize_Fields_For_Generic_Package_Instantiation
        (Self : Bare_Generic_Package_Instantiation
         ; Generic_Package_Instantiation_F_Name : Bare_Defining_Name
         ; Generic_Package_Instantiation_F_Generic_Pkg_Name : Bare_Name
         ; Generic_Package_Instantiation_F_Params : Bare_Assoc_List
         ; Generic_Package_Instantiation_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Initialize_Fields_For_Generic_Instantiation
              (Self);
            Self.Generic_Package_Instantiation_F_Name := Generic_Package_Instantiation_F_Name;
            Self.Generic_Package_Instantiation_F_Generic_Pkg_Name := Generic_Package_Instantiation_F_Generic_Pkg_Name;
            Self.Generic_Package_Instantiation_F_Params := Generic_Package_Instantiation_F_Params;
            Self.Generic_Package_Instantiation_F_Aspects := Generic_Package_Instantiation_F_Aspects;
      end Initialize_Fields_For_Generic_Package_Instantiation;
   function Generic_Package_Instantiation_F_Name
     (Node : Bare_Generic_Package_Instantiation) return Bare_Defining_Name
   is
   begin
         return Node.Generic_Package_Instantiation_F_Name;
   end;
   function Generic_Package_Instantiation_F_Generic_Pkg_Name
     (Node : Bare_Generic_Package_Instantiation) return Bare_Name
   is
   begin
         return Node.Generic_Package_Instantiation_F_Generic_Pkg_Name;
   end;
   function Generic_Package_Instantiation_F_Params
     (Node : Bare_Generic_Package_Instantiation) return Bare_Assoc_List
   is
   begin
         return Node.Generic_Package_Instantiation_F_Params;
   end;
--# property-start GenericPackageInstantiation.generic_entity_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10274
pragma Warnings (Off, "is not referenced");
 function Generic_Package_Instantiation_P_Generic_Entity_Name
  (Node : Bare_Generic_Package_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Name
is
   Self : Bare_Generic_Package_Instantiation := Bare_Generic_Package_Instantiation (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Package_Instantiation :=
     Internal_Entity_Generic_Package_Instantiation'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Name;
      Fld : Internal_Entity_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5194 '<FieldAccess for generic_pkg_name at ast.py:10274>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10274
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Generic_Package_Instantiation_F_Generic_Pkg_Name, Info => Ent.Info);
--# expr-done 5194
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Package_Instantiation_P_Generic_Entity_Name;
--# end
--# property-start GenericPackageInstantiation.generic_inst_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10275
pragma Warnings (Off, "is not referenced");
 function Generic_Package_Instantiation_P_Generic_Inst_Params
  (Node : Bare_Generic_Package_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Assoc_List
is
   Self : Bare_Generic_Package_Instantiation := Bare_Generic_Package_Instantiation (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Package_Instantiation :=
     Internal_Entity_Generic_Package_Instantiation'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Assoc_List;
      Fld : Internal_Entity_Assoc_List;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5195 '<FieldAccess for params at ast.py:10275>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10275
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Assoc_List (Node => Ent.Node.Generic_Package_Instantiation_F_Params, Info => Ent.Info);
--# expr-done 5195
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Package_Instantiation_P_Generic_Inst_Params;
--# end
--# property-start GenericPackageInstantiation.designated_package /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10278
pragma Warnings (Off, "is not referenced");
 function Generic_Package_Instantiation_P_Designated_Package
  (Node : Bare_Generic_Package_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Package_Decl
is
   Self : Bare_Generic_Package_Instantiation := Bare_Generic_Package_Instantiation (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Package_Instantiation :=
     Internal_Entity_Generic_Package_Instantiation'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Package_Decl;
            procedure Finalizer_Scope_2479 with Inline_Always;
      Var_Expr : Internal_Entity_Generic_Decl;
Var_Expr_1 : Internal_Entity_Generic_Decl;
Fld : Internal_Entity_Generic_Decl;
Fld_1 : Bare_Generic_Decl;
Cast_Result : Bare_Generic_Package_Decl;
Fld_2 : Bare_Generic_Package_Internal;
Fld_3 : Internal_Entity_Info;
Fld_4 : Internal_Metadata;
Fld_5 : Internal_Entity_Info;
Fld_6 : Env_Rebindings;
Fld_7 : Internal_Entity_Basic_Decl;
Fld_8 : Internal_Entity_Info;
Scope_Result : Internal_Entity_Info;
Result_Var : Internal_Entity_Info;
Fld_9 : Env_Rebindings;
Rebinding : Env_Rebindings;
Fld_10 : Bare_Generic_Decl;
Fld_11 : Lexical_Env := Empty_Env;
Fld_12 : Lexical_Env := Empty_Env;
Rebinding_1 : Env_Rebindings;
Fld_13 : Internal_Entity_Info;
Fld_14 : Boolean;
New_Struct : Internal_Entity_Info;
Cast_Result_1 : Bare_Base_Package_Decl;
New_Struct_1 : Internal_Entity_Base_Package_Decl;
Scope_Result_1 : Internal_Entity_Base_Package_Decl;
Result_Var_1 : Internal_Entity_Base_Package_Decl;
Let_Result : Internal_Entity_Base_Package_Decl;
            procedure Finalizer_Scope_2479 is
            begin
                     Dec_Ref (Fld_11);
                     Dec_Ref (Fld_12);
            end Finalizer_Scope_2479;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5216 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5196 '<FieldAccess for nonbound_generic_decl at ast.py:10279>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10279
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericInstantiation.nonbound_generic_decl
Fld := Libadalang.Implementation.Generic_Instantiation_P_Nonbound_Generic_Decl (Node => Self);
--# end
--# expr-done 5196
Var_Expr := Fld;
if Var_Expr /= No_Entity_Generic_Decl then
   --# scope-start
   --# bind p Var_Expr
--# expr-start 5215 '<New BasePackageDecl.entity at ???>' New_Struct_1 None
--# expr-start 5211 '<New EntityInfo at ast.py:10282>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10282
--# expr-start 5198 '<FieldAccess for from_rebound at ast.py:10299>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10299
--# expr-start 5197 '<FieldAccess for info at ast.py:10299>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10299
Fld_13 := Var_Expr.Info;
--# expr-done 5197
Fld_14 := Fld_13.From_Rebound;
--# expr-done 5198
--# expr-start 5200 '<FieldAccess for md at ast.py:10283>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10283
--# expr-start 5199 '<FieldAccess for info at ast.py:10283>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10283
Fld_3 := Var_Expr.Info;
--# expr-done 5199
Fld_4 := Fld_3.Md;
--# expr-done 5200
--# expr-start 5210 '<append_rebinding at ast.py:10286>' Rebinding_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10286
--# expr-start 5206 '<concat_rebindings at ast.py:10286>' Rebinding /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10286
--# expr-start 5202 '<FieldAccess for rebindings at ast.py:10286>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10286
--# expr-start 5201 '<FieldAccess for info at ast.py:10286>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10286
Fld_5 := Ent.Info;
--# expr-done 5201
Fld_6 := Fld_5.Rebindings;
--# expr-done 5202
--# expr-start 5205 '<FieldAccess for rebindings at ast.py:10289>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10289
Var_Expr_1 := Var_Expr;
if Var_Expr_1 /= No_Entity_Generic_Decl then
   --# scope-start
--# expr-start 5204 '<FieldAccess for info at ast.py:10289>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10289
--# expr-start 5203 '<FieldAccess for decl at ast.py:10289>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10289
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericDecl.decl'
Fld_7 := Libadalang.Implementation.Dispatcher_Generic_Decl_P_Decl (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 5203
Fld_8 := Fld_7.Info;
--# expr-done 5204
Scope_Result := Fld_8;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Info;
end if;
Fld_9 := Result_Var.Rebindings;
--# expr-done 5205
Rebinding := AST_Envs.Combine (Fld_6, Fld_9); 
--# expr-done 5206
--# expr-start 5208 '<FieldAccess for children_env at ast.py:10297>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10297
--# expr-start 5207 '<FieldAccess for node at ast.py:10297>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10297
Fld_10 := Var_Expr.Node;
--# expr-done 5207
      if Fld_10 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Libadalang.Implementation.Children_Env (Node => Fld_10);
--# expr-done 5208
--# expr-start 5209 '<FieldAccess for instantiation_env at ast.py:10298>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10298
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericInstantiation.instantiation_env
Fld_12 := Libadalang.Implementation.Generic_Instantiation_F_Instantiation_Env (Node => Self);
--# end
--# expr-done 5209
Rebinding_1 := AST_Envs.Append_Rebinding (Rebinding, Fld_11, Fld_12); 
--# expr-done 5210
New_Struct := (From_Rebound => Fld_14, Md => Fld_4, Rebindings => Rebinding_1); 
--# expr-done 5211
--# expr-start 5214 '<FieldAccess for package_decl at ast.py:10281>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10281
--# expr-start 5213 '<Cast to GenericPackageDecl at ast.py:10281>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10281
--# expr-start 5212 '<FieldAccess for node at ast.py:10281>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10281
Fld_1 := Var_Expr.Node;
--# expr-done 5212
   if Fld_1 = null
      or else Fld_1.Kind in Ada_Generic_Package_Decl_Range
   then
      Cast_Result := Fld_1;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 5213
      if Cast_Result = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Cast_Result.Generic_Package_Decl_F_Package_Decl;
--# expr-done 5214
      Cast_Result_1 := Fld_2;
New_Struct_1 := (Info => New_Struct, Node => Cast_Result_1); 
--# expr-done 5215
Scope_Result_1 := New_Struct_1;
   --# end
      Finalizer_Scope_2479;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Base_Package_Decl;
end if;
Let_Result := Result_Var_1; 
--# end
--# expr-done 5216
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_2479;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Package_Instantiation_P_Designated_Package;
--# end
--# property-start GenericPackageInstantiation.designated_generic_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10304
pragma Warnings (Off, "is not referenced");
 function Generic_Package_Instantiation_P_Designated_Generic_Decl
  (Node : Bare_Generic_Package_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Generic_Package_Instantiation := Bare_Generic_Package_Instantiation (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Package_Instantiation :=
     Internal_Entity_Generic_Package_Instantiation'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Internal_Entity_Base_Package_Decl;
Fld_1 : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5219 '<Cast to BasicDecl at ast.py:10305>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10305
--# expr-start 5218 '<parent at ast.py:10305>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10305
--# expr-start 5217 '<FieldAccess for designated_package at ast.py:10305>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10305
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericPackageInstantiation.designated_package
Fld := Libadalang.Implementation.Generic_Package_Instantiation_P_Designated_Package (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5217
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Parent (Node => Fld.Node, E_Info => Fld.Info);
--# expr-done 5218
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 5219
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Package_Instantiation_P_Designated_Generic_Decl;
--# end
--# property-start GenericPackageInstantiation.defining_env_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10309
pragma Warnings (Off, "is not referenced");
 function Generic_Package_Instantiation_P_Defining_Env_Impl
  (Node : Bare_Generic_Package_Instantiation
      ; Inst_From_Formal : Boolean
         := False
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Generic_Package_Instantiation := Bare_Generic_Package_Instantiation (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Package_Instantiation :=
     Internal_Entity_Generic_Package_Instantiation'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind inst_from_formal Inst_From_Formal
   --# bind origin Origin
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_817 with Inline_Always;
      Block_Var : Internal_Entity_Base_Package_Decl;
Fld : Internal_Entity_Base_Package_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Fld_2 : Internal_Entity;
Fld_3 : Lexical_Env := Empty_Env;
Array_Lit : Lexical_Env_Array_Access;
Group_Env : Lexical_Env := Empty_Env;
Fld_4 : Lexical_Env := Empty_Env;
Fld_5 : Boolean;
If_Result : Boolean;
If_Result_1 : Lexical_Env := Empty_Env;
Fld_6 : Lexical_Env := Empty_Env;
Array_Lit_1 : Lexical_Env_Array_Access;
Group_Env_1 : Lexical_Env := Empty_Env;
Is_Null : Boolean;
If_Result_2 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_817 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Group_Env);
                     Dec_Ref (Fld_4);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Fld_6);
                     Dec_Ref (Array_Lit_1);
                     Dec_Ref (Group_Env_1);
                     Dec_Ref (If_Result_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_817;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5234 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5220 '<FieldAccess for designated_package at ast.py:10318>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10318
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericPackageInstantiation.designated_package
Fld := Libadalang.Implementation.Generic_Package_Instantiation_P_Designated_Package (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5220
Block_Var := Fld; 
--# bind dp Block_Var
--# expr-start 5233 '<If at ast.py:10319>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10319
--# expr-start 5221 '<IsNull at ast.py:10339>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10339
Is_Null := Block_Var.Node = null; 
--# expr-done 5221
if Is_Null then
   If_Result_2 := Empty_Env;
else
   --# expr-start 5232 '<env_group at ast.py:10341>' Group_Env_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10341
--# expr-start 5231 '<ArrayLiteral at ast.py:10341>' Array_Lit_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10341
--# expr-start 5229 '<If at ast.py:10342>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10342
--# expr-start 5222 '<FieldAccess for is_formal at ast.py:10343>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10343
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_formal
Fld_5 := Libadalang.Implementation.Basic_Decl_P_Is_Formal (Node => Self);
--# end
--# expr-done 5222
if Fld_5 then
   If_Result := True;
else
   If_Result := Inst_From_Formal;
end if;
if If_Result then
   --# expr-start 5227 '<env_group at ast.py:10344>' Group_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10344
--# expr-start 5226 '<ArrayLiteral at ast.py:10344>' Array_Lit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10344
--# expr-start 5223 '<FieldAccess for children_env at ast.py:10345>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10345
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Children_Env (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# expr-done 5223
--# expr-start 5225 '<FieldAccess for children_env at ast.py:10346>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10346
--# expr-start 5224 '<parent at ast.py:10346>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10346
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Parent (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# expr-done 5224
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Children_Env (Node => Fld_2.Node, E_Info => Fld_2.Info);
--# expr-done 5225
Array_Lit := Create_Lexical_Env_Array (Internal_Lexical_Env_Array'(1 => Fld_1, 2 => Fld_3)); 
--# expr-done 5226
Group_Env := Group (Array_Lit, No_Metadata); 
--# expr-done 5227
   If_Result_1 := Group_Env;
else
   --# expr-start 5228 '<FieldAccess for children_env at ast.py:10348>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10348
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Children_Env (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# expr-done 5228
   If_Result_1 := Fld_4;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 5229
--# expr-start 5230 '<FieldAccess for children_env at ast.py:10360>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10360
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Children_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 5230
Array_Lit_1 := Create_Lexical_Env_Array (Internal_Lexical_Env_Array'(1 => If_Result_1, 2 => Fld_6)); 
--# expr-done 5231
Group_Env_1 := Group (Array_Lit_1, No_Metadata); 
--# expr-done 5232
   If_Result_2 := Group_Env_1;
end if;
      Inc_Ref (If_Result_2);
--# expr-done 5233
Let_Result := If_Result_2; Inc_Ref (Let_Result);
--# end
--# expr-done 5234
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_817;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_817;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Package_Instantiation_P_Defining_Env_Impl;
--# end
--# property-start GenericPackageInstantiation.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10365
pragma Warnings (Off, "is not referenced");
 function Generic_Package_Instantiation_P_Defining_Env
  (Node : Bare_Generic_Package_Instantiation
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Generic_Package_Instantiation := Bare_Generic_Package_Instantiation (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Package_Instantiation :=
     Internal_Entity_Generic_Package_Instantiation'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_818 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_818 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_818;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5236 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5235 '<FieldAccess for defining_env_impl at ast.py:10366>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10366
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericPackageInstantiation.defining_env_impl
Fld := Libadalang.Implementation.Generic_Package_Instantiation_P_Defining_Env_Impl (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 5235
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 5236
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_818;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_818;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Package_Instantiation_P_Defining_Env;
--# end
--# property-start GenericPackageInstantiation.initial_env_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10369
pragma Warnings (Off, "is not referenced");
 function Generic_Package_Instantiation_P_Initial_Env_Name
  (Node : Bare_Generic_Package_Instantiation
  )
   return Symbol_Type
is
   Self : Bare_Generic_Package_Instantiation := Bare_Generic_Package_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type;
      Fld : Symbol_Type;
Fld_1 : Boolean;
If_Result : Symbol_Type;
Let_Result : Symbol_Type;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5240 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5239 '<If at ast.py:10370>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10370
--# expr-start 5237 '<FieldAccess for is_library_item at ast.py:10371>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10371
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_library_item'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Library_Item (Node => Self);
--# end
--# expr-done 5237
if Fld_1 then
   --# expr-start 5238 '<FieldAccess for child_decl_initial_env_name at ast.py:10372>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10372
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.child_decl_initial_env_name'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Child_Decl_Initial_Env_Name (Node => Self);
--# end
--# expr-done 5238
   If_Result := Fld;
else
   If_Result := null;
end if;
--# expr-done 5239
Let_Result := If_Result; 
--# end
--# expr-done 5240
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Package_Instantiation_P_Initial_Env_Name;
--# end
--# property-start GenericPackageInstantiation.env_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10377
pragma Warnings (Off, "is not referenced");
 function Generic_Package_Instantiation_P_Env_Names
  (Node : Bare_Generic_Package_Instantiation
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Generic_Package_Instantiation := Bare_Generic_Package_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_820 with Inline_Always;
            procedure Finalizer_Scope_2481 with Inline_Always;
      Var_Expr : String_Type;
Fld : String_Type;
Sym : Symbol_Type;
Singleton : Symbol_Type_Array_Access;
Scope_Result : Symbol_Type_Array_Access;
Result_Var : Symbol_Type_Array_Access;
Let_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_820 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_820;
            procedure Finalizer_Scope_2481 is
            begin
                     Dec_Ref (Var_Expr);
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2481;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5243 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5241 '<FieldAccess for top_level_env_name at ast.py:10378>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10378
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.top_level_env_name
Fld := Libadalang.Implementation.Basic_Decl_P_Top_Level_Env_Name (Node => Self);
--# end
--# expr-done 5241
Var_Expr := Fld;
      Inc_Ref (Var_Expr);
if Var_Expr /= Empty_String then
   --# scope-start
   --# bind fqn Var_Expr
--# expr-start 5242 '<to_symbol at ast.py:10379>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10379
Sym := String_To_Symbol (Self.Unit.Context, Var_Expr); 
--# expr-done 5242
                Singleton := Create_Symbol_Type_Array (Items_Count => 1);
                Singleton.Items (1) := Sym;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2481;
   Result_Var := Scope_Result;
else
   Result_Var := No_Symbol_Type_Array_Type;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 5243
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_820;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_820;
               Finalizer_Scope_2481;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Package_Instantiation_P_Env_Names;
--# end
--# property-start GenericPackageInstantiation.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10382
pragma Warnings (Off, "is not referenced");
 function Generic_Package_Instantiation_P_Defining_Names
  (Node : Bare_Generic_Package_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Generic_Package_Instantiation := Bare_Generic_Package_Instantiation (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Package_Instantiation :=
     Internal_Entity_Generic_Package_Instantiation'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_821 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_821 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_821;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5244 '<FieldAccess for name at ast.py:10382>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10382
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Generic_Package_Instantiation_F_Name, Info => Ent.Info);
--# expr-done 5244
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_821;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_821;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Package_Instantiation_P_Defining_Names;
--# end
--# property-start GenericPackageInstantiation._env_do_68 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10385
pragma Warnings (Off, "is not referenced");
 function Env_Do_68
  (Node : Bare_Generic_Package_Instantiation
  )
   return Boolean
is
   Self : Bare_Generic_Package_Instantiation := Bare_Generic_Package_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5245 '<FieldAccess for env_hook at ast.py:10385>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10385
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.env_hook
Fld := Libadalang.Implementation.Ada_Node_P_Env_Hook (Node => Self);
--# end
--# expr-done 5245
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_68;
--# end
--# property-start GenericPackageInstantiation._initial_env_69 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10388
pragma Warnings (Off, "is not referenced");
 function Initial_Env_69
  (Node : Bare_Generic_Package_Instantiation
  )
   return Internal_Designated_Env
is
   Self : Bare_Generic_Package_Instantiation := Bare_Generic_Package_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_1486 with Inline_Always;
            procedure Finalizer_Scope_2482 with Inline_Always;
      Var_Expr : Symbol_Type;
Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Scope_Result : Internal_Designated_Env := No_Designated_Env;
New_Struct_1 : Internal_Designated_Env := No_Designated_Env;
Result_Var : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_1486 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (New_Struct_1);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1486;
            procedure Finalizer_Scope_2482 is
            begin
                     Dec_Ref (New_Struct);
            end Finalizer_Scope_2482;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
--# expr-start 5246 '<FieldAccess for initial_env_name at ast.py:10389>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10389
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericPackageInstantiation.initial_env_name
Fld := Libadalang.Implementation.Generic_Package_Instantiation_P_Initial_Env_Name (Node => Self);
--# end
--# expr-done 5246
Var_Expr := Fld;
if Var_Expr /= null then
   --# scope-start
   --# bind non_null_name Var_Expr
--# expr-start 5247 '<New DesignatedEnv at ???>' New_Struct None
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => Var_Expr, Kind => Named_Env); 
--# expr-done 5247
Scope_Result := New_Struct;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2482;
   Result_Var := Scope_Result;
else
   --# expr-start 5248 '<New DesignatedEnv at ast.py:10388>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10388
Inc_Ref (Empty_Env);
New_Struct_1 := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 5248
   Result_Var := New_Struct_1;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1486;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1486;
               Finalizer_Scope_2482;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Initial_Env_69;
--# end
--# property-start GenericPackageInstantiation._env_mappings_70 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10394
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_70
  (Node : Bare_Generic_Package_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Generic_Package_Instantiation := Bare_Generic_Package_Instantiation (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Package_Instantiation :=
     Internal_Entity_Generic_Package_Instantiation'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1487 with Inline_Always;
      Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Generic_Package_Instantiation;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1487 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1487;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5251 '<New EnvAssoc at ast.py:10394>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10394
--# expr-start 5249 '<New DesignatedEnv at ast.py:10394>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10394
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 5249
--# expr-start 5250 '<FieldAccess for name_symbol at ast.py:10394>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10394
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 5250
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 5251
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1487;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1487;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_70;
--# end
--# property-start GenericPackageInstantiation._env_trans_parent_71 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10384
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_71
  (Node : Bare_Generic_Package_Instantiation
  )
   return Boolean
is
   Self : Bare_Generic_Package_Instantiation := Bare_Generic_Package_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_71;
--# end
--# property-start GenericPackageInstantiation._env_names_72 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10396
pragma Warnings (Off, "is not referenced");
 function Env_Names_72
  (Node : Bare_Generic_Package_Instantiation
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Generic_Package_Instantiation := Bare_Generic_Package_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1489 with Inline_Always;
      Fld : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1489 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1489;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5252 '<FieldAccess for env_names at ast.py:10396>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10396
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericPackageInstantiation.env_names
Fld := Libadalang.Implementation.Generic_Package_Instantiation_P_Env_Names (Node => Self);
--# end
--# expr-done 5252
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1489;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1489;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Names_72;
--# end
--# property-start GenericPackageInstantiation._env_do_73 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10398
pragma Warnings (Off, "is not referenced");
 function Env_Do_73
  (Node : Bare_Generic_Package_Instantiation
  )
   return Bare_Compilation_Unit_Array_Access
is
   Self : Bare_Generic_Package_Instantiation := Bare_Generic_Package_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1490 with Inline_Always;
      Fld : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1490 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1490;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5253 '<FieldAccess for populate_dependent_units at ast.py:10398>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10398
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.populate_dependent_units
Fld := Libadalang.Implementation.Basic_Decl_P_Populate_Dependent_Units (Node => Self);
--# end
--# expr-done 5253
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1490;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1490;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_73;
--# end
--# property-start GenericPackageInstantiation._ref_env_nodes_74 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10400
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_74
  (Node : Bare_Generic_Package_Instantiation
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Generic_Package_Instantiation := Bare_Generic_Package_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1491 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1491 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1491;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5254 '<FieldAccess for top_level_use_package_clauses at ast.py:10400>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10400
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_package_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Package_Clauses (Node => Self);
--# end
--# expr-done 5254
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1491;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1491;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_74;
--# end
--# property-start GenericPackageInstantiation._ref_cond_75 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10402
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_75
  (Node : Bare_Generic_Package_Instantiation
  )
   return Boolean
is
   Self : Bare_Generic_Package_Instantiation := Bare_Generic_Package_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5256 '<IsA LibraryItem, Subunit at ast.py:10402>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10402
--# expr-start 5255 '<parent at ast.py:10402>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10402
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 5255
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 5256
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_75;
--# end
--# property-start GenericPackageInstantiation._ref_env_nodes_76 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10405
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_76
  (Node : Bare_Generic_Package_Instantiation
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Generic_Package_Instantiation := Bare_Generic_Package_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1493 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1493 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1493;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5257 '<FieldAccess for top_level_use_type_clauses at ast.py:10405>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10405
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_type_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Type_Clauses (Node => Self);
--# end
--# expr-done 5257
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1493;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1493;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_76;
--# end
--# property-start GenericPackageInstantiation._ref_cond_77 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10407
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_77
  (Node : Bare_Generic_Package_Instantiation
  )
   return Boolean
is
   Self : Bare_Generic_Package_Instantiation := Bare_Generic_Package_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5259 '<IsA LibraryItem, Subunit at ast.py:10407>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10407
--# expr-start 5258 '<parent at ast.py:10407>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10407
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 5258
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 5259
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_77;
--# end
   --
   --  Primitives for Bare_Generic_Subp_Instantiation
   --
      procedure Generic_Subp_Instantiation_Pre_Env_Actions
        (Self            : Bare_Generic_Subp_Instantiation;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Dummy : Boolean;
      begin
         Dummy := Libadalang.Implementation.Env_Do_59 (Node => 
Self);
      end;
      declare
         Env : constant Internal_Designated_Env :=
           Libadalang.Implementation.Initial_Env_60 (Node => 
Self);
      begin
         Set_Initial_Env
           (Self,
            State,
            Env,
            DSL_Location => "ast.py:10239");
      end;
      declare
         Resolver : constant Entity_Resolver :=
            Generic_Subp_Instantiation_P_Designated_Subp'Access;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_61 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:10241");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_62 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      declare
         Dummy : Bare_Compilation_Unit_Array_Access;
      begin
         Dummy := Libadalang.Implementation.Env_Do_63 (Node => 
Self);
         Dec_Ref (Dummy);
      end;
      if Libadalang.Implementation.Ref_Cond_65 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_64 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Use_Package_Name_Designated_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_67 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_66 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Name_Designated_Type_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      end;
      procedure Initialize_Fields_For_Generic_Subp_Instantiation
        (Self : Bare_Generic_Subp_Instantiation
         ; Generic_Subp_Instantiation_F_Overriding : Bare_Overriding_Node
         ; Generic_Subp_Instantiation_F_Kind : Bare_Subp_Kind
         ; Generic_Subp_Instantiation_F_Subp_Name : Bare_Defining_Name
         ; Generic_Subp_Instantiation_F_Generic_Subp_Name : Bare_Name
         ; Generic_Subp_Instantiation_F_Params : Bare_Assoc_List
         ; Generic_Subp_Instantiation_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Initialize_Fields_For_Generic_Instantiation
              (Self);
            Self.Generic_Subp_Instantiation_F_Overriding := Generic_Subp_Instantiation_F_Overriding;
            Self.Generic_Subp_Instantiation_F_Kind := Generic_Subp_Instantiation_F_Kind;
            Self.Generic_Subp_Instantiation_F_Subp_Name := Generic_Subp_Instantiation_F_Subp_Name;
            Self.Generic_Subp_Instantiation_F_Generic_Subp_Name := Generic_Subp_Instantiation_F_Generic_Subp_Name;
            Self.Generic_Subp_Instantiation_F_Params := Generic_Subp_Instantiation_F_Params;
            Self.Generic_Subp_Instantiation_F_Aspects := Generic_Subp_Instantiation_F_Aspects;
      end Initialize_Fields_For_Generic_Subp_Instantiation;
   function Generic_Subp_Instantiation_F_Overriding
     (Node : Bare_Generic_Subp_Instantiation) return Bare_Overriding_Node
   is
   begin
         return Node.Generic_Subp_Instantiation_F_Overriding;
   end;
   function Generic_Subp_Instantiation_F_Kind
     (Node : Bare_Generic_Subp_Instantiation) return Bare_Subp_Kind
   is
   begin
         return Node.Generic_Subp_Instantiation_F_Kind;
   end;
   function Generic_Subp_Instantiation_F_Subp_Name
     (Node : Bare_Generic_Subp_Instantiation) return Bare_Defining_Name
   is
   begin
         return Node.Generic_Subp_Instantiation_F_Subp_Name;
   end;
   function Generic_Subp_Instantiation_F_Generic_Subp_Name
     (Node : Bare_Generic_Subp_Instantiation) return Bare_Name
   is
   begin
         return Node.Generic_Subp_Instantiation_F_Generic_Subp_Name;
   end;
   function Generic_Subp_Instantiation_F_Params
     (Node : Bare_Generic_Subp_Instantiation) return Bare_Assoc_List
   is
   begin
         return Node.Generic_Subp_Instantiation_F_Params;
   end;
--# property-start GenericSubpInstantiation.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10206
pragma Warnings (Off, "is not referenced");
 function Generic_Subp_Instantiation_P_Defining_Names
  (Node : Bare_Generic_Subp_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Generic_Subp_Instantiation := Bare_Generic_Subp_Instantiation (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Subp_Instantiation :=
     Internal_Entity_Generic_Subp_Instantiation'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_808 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_808 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_808;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5260 '<FieldAccess for subp_name at ast.py:10206>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10206
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Generic_Subp_Instantiation_F_Subp_Name, Info => Ent.Info);
--# expr-done 5260
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_808;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_808;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Subp_Instantiation_P_Defining_Names;
--# end
--# property-start GenericSubpInstantiation.generic_entity_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10208
pragma Warnings (Off, "is not referenced");
 function Generic_Subp_Instantiation_P_Generic_Entity_Name
  (Node : Bare_Generic_Subp_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Name
is
   Self : Bare_Generic_Subp_Instantiation := Bare_Generic_Subp_Instantiation (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Subp_Instantiation :=
     Internal_Entity_Generic_Subp_Instantiation'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Name;
      Fld : Internal_Entity_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5261 '<FieldAccess for generic_subp_name at ast.py:10208>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10208
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Generic_Subp_Instantiation_F_Generic_Subp_Name, Info => Ent.Info);
--# expr-done 5261
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Subp_Instantiation_P_Generic_Entity_Name;
--# end
--# property-start GenericSubpInstantiation.generic_inst_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10209
pragma Warnings (Off, "is not referenced");
 function Generic_Subp_Instantiation_P_Generic_Inst_Params
  (Node : Bare_Generic_Subp_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Assoc_List
is
   Self : Bare_Generic_Subp_Instantiation := Bare_Generic_Subp_Instantiation (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Subp_Instantiation :=
     Internal_Entity_Generic_Subp_Instantiation'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Assoc_List;
      Fld : Internal_Entity_Assoc_List;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5262 '<FieldAccess for params at ast.py:10209>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10209
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Assoc_List (Node => Ent.Node.Generic_Subp_Instantiation_F_Params, Info => Ent.Info);
--# expr-done 5262
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Subp_Instantiation_P_Generic_Inst_Params;
--# end
--# property-start GenericSubpInstantiation.designated_subp /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10212
pragma Warnings (Off, "is not referenced");
 function Generic_Subp_Instantiation_P_Designated_Subp
  (Node : Bare_Generic_Subp_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity
is
   Self : Bare_Generic_Subp_Instantiation := Bare_Generic_Subp_Instantiation (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Subp_Instantiation :=
     Internal_Entity_Generic_Subp_Instantiation'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity;
            procedure Finalizer_Scope_2483 with Inline_Always;
      Var_Expr : Internal_Entity_Generic_Decl;
Var_Expr_1 : Internal_Entity_Generic_Decl;
Fld : Internal_Entity_Generic_Decl;
Fld_1 : Bare_Generic_Decl;
Cast_Result : Bare_Generic_Subp_Decl;
Fld_2 : Bare_Generic_Subp_Internal;
Fld_3 : Internal_Entity_Info;
Fld_4 : Internal_Metadata;
Fld_5 : Internal_Entity_Info;
Fld_6 : Env_Rebindings;
Fld_7 : Internal_Entity_Basic_Decl;
Fld_8 : Internal_Entity_Info;
Scope_Result : Internal_Entity_Info;
Result_Var : Internal_Entity_Info;
Fld_9 : Env_Rebindings;
Rebinding : Env_Rebindings;
Fld_10 : Bare_Generic_Decl;
Fld_11 : Lexical_Env := Empty_Env;
Fld_12 : Lexical_Env := Empty_Env;
Rebinding_1 : Env_Rebindings;
Fld_13 : Internal_Entity_Info;
Fld_14 : Boolean;
New_Struct : Internal_Entity_Info;
Cast_Result_1 : Bare_Basic_Subp_Decl;
New_Struct_1 : Internal_Entity_Basic_Subp_Decl;
Cast_Result_2 : Internal_Entity;
Scope_Result_1 : Internal_Entity;
Result_Var_1 : Internal_Entity;
Let_Result : Internal_Entity;
            procedure Finalizer_Scope_2483 is
            begin
                     Dec_Ref (Fld_11);
                     Dec_Ref (Fld_12);
            end Finalizer_Scope_2483;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5284 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5263 '<FieldAccess for nonbound_generic_decl at ast.py:10216>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10216
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericInstantiation.nonbound_generic_decl
Fld := Libadalang.Implementation.Generic_Instantiation_P_Nonbound_Generic_Decl (Node => Self);
--# end
--# expr-done 5263
Var_Expr := Fld;
if Var_Expr /= No_Entity_Generic_Decl then
   --# scope-start
   --# bind p Var_Expr
--# expr-start 5283 '<Cast to AdaNode.entity at ???>' Cast_Result_2 None
--# expr-start 5282 '<New BasicSubpDecl.entity at ???>' New_Struct_1 None
--# expr-start 5278 '<New EntityInfo at ast.py:10219>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10219
--# expr-start 5265 '<FieldAccess for from_rebound at ast.py:10227>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10227
--# expr-start 5264 '<FieldAccess for info at ast.py:10227>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10227
Fld_13 := Var_Expr.Info;
--# expr-done 5264
Fld_14 := Fld_13.From_Rebound;
--# expr-done 5265
--# expr-start 5267 '<FieldAccess for md at ast.py:10220>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10220
--# expr-start 5266 '<FieldAccess for info at ast.py:10220>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10220
Fld_3 := Var_Expr.Info;
--# expr-done 5266
Fld_4 := Fld_3.Md;
--# expr-done 5267
--# expr-start 5277 '<append_rebinding at ast.py:10221>' Rebinding_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10221
--# expr-start 5273 '<concat_rebindings at ast.py:10221>' Rebinding /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10221
--# expr-start 5269 '<FieldAccess for rebindings at ast.py:10221>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10221
--# expr-start 5268 '<FieldAccess for info at ast.py:10221>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10221
Fld_5 := Ent.Info;
--# expr-done 5268
Fld_6 := Fld_5.Rebindings;
--# expr-done 5269
--# expr-start 5272 '<FieldAccess for rebindings at ast.py:10223>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10223
Var_Expr_1 := Var_Expr;
if Var_Expr_1 /= No_Entity_Generic_Decl then
   --# scope-start
--# expr-start 5271 '<FieldAccess for info at ast.py:10223>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10223
--# expr-start 5270 '<FieldAccess for decl at ast.py:10223>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10223
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericDecl.decl'
Fld_7 := Libadalang.Implementation.Dispatcher_Generic_Decl_P_Decl (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 5270
Fld_8 := Fld_7.Info;
--# expr-done 5271
Scope_Result := Fld_8;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Info;
end if;
Fld_9 := Result_Var.Rebindings;
--# expr-done 5272
Rebinding := AST_Envs.Combine (Fld_6, Fld_9); 
--# expr-done 5273
--# expr-start 5275 '<FieldAccess for children_env at ast.py:10225>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10225
--# expr-start 5274 '<FieldAccess for node at ast.py:10225>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10225
Fld_10 := Var_Expr.Node;
--# expr-done 5274
      if Fld_10 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Libadalang.Implementation.Children_Env (Node => Fld_10);
--# expr-done 5275
--# expr-start 5276 '<FieldAccess for instantiation_env at ast.py:10225>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10225
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericInstantiation.instantiation_env
Fld_12 := Libadalang.Implementation.Generic_Instantiation_F_Instantiation_Env (Node => Self);
--# end
--# expr-done 5276
Rebinding_1 := AST_Envs.Append_Rebinding (Rebinding, Fld_11, Fld_12); 
--# expr-done 5277
New_Struct := (From_Rebound => Fld_14, Md => Fld_4, Rebindings => Rebinding_1); 
--# expr-done 5278
--# expr-start 5281 '<FieldAccess for subp_decl at ast.py:10218>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10218
--# expr-start 5280 '<Cast to GenericSubpDecl at ast.py:10218>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10218
--# expr-start 5279 '<FieldAccess for node at ast.py:10218>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10218
Fld_1 := Var_Expr.Node;
--# expr-done 5279
   if Fld_1 = null
      or else Fld_1.Kind in Ada_Generic_Subp_Decl_Range
   then
      Cast_Result := Fld_1;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 5280
      if Cast_Result = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Cast_Result.Generic_Subp_Decl_F_Subp_Decl;
--# expr-done 5281
      Cast_Result_1 := Fld_2;
New_Struct_1 := (Info => New_Struct, Node => Cast_Result_1); 
--# expr-done 5282
      Cast_Result_2 := Create_Internal_Entity
        (Node => New_Struct_1.Node,
         Info => New_Struct_1.Info);
--# expr-done 5283
Scope_Result_1 := Cast_Result_2;
   --# end
      Finalizer_Scope_2483;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity;
end if;
Let_Result := Result_Var_1; 
--# end
--# expr-done 5284
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_2483;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Subp_Instantiation_P_Designated_Subp;
--# end
--# property-start GenericSubpInstantiation.designated_generic_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10232
pragma Warnings (Off, "is not referenced");
 function Generic_Subp_Instantiation_P_Designated_Generic_Decl
  (Node : Bare_Generic_Subp_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Generic_Subp_Instantiation := Bare_Generic_Subp_Instantiation (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Subp_Instantiation :=
     Internal_Entity_Generic_Subp_Instantiation'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Internal_Entity;
Fld_1 : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5287 '<Cast to BasicDecl at ast.py:10233>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10233
--# expr-start 5286 '<parent at ast.py:10233>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10233
--# expr-start 5285 '<FieldAccess for designated_subp at ast.py:10233>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10233
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericSubpInstantiation.designated_subp
Fld := Libadalang.Implementation.Generic_Subp_Instantiation_P_Designated_Subp (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5285
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Parent (Node => Fld.Node, E_Info => Fld.Info);
--# expr-done 5286
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 5287
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Subp_Instantiation_P_Designated_Generic_Decl;
--# end
--# property-start GenericSubpInstantiation._env_do_59 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10237
pragma Warnings (Off, "is not referenced");
 function Env_Do_59
  (Node : Bare_Generic_Subp_Instantiation
  )
   return Boolean
is
   Self : Bare_Generic_Subp_Instantiation := Bare_Generic_Subp_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5288 '<FieldAccess for env_hook at ast.py:10237>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10237
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.env_hook
Fld := Libadalang.Implementation.Ada_Node_P_Env_Hook (Node => Self);
--# end
--# expr-done 5288
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_59;
--# end
--# property-start GenericSubpInstantiation._initial_env_60 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10239
pragma Warnings (Off, "is not referenced");
 function Initial_Env_60
  (Node : Bare_Generic_Subp_Instantiation
  )
   return Internal_Designated_Env
is
   Self : Bare_Generic_Subp_Instantiation := Bare_Generic_Subp_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_1477 with Inline_Always;
      Fld : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_1477 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1477;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5289 '<FieldAccess for child_decl_initial_env at ast.py:10239>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10239
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.child_decl_initial_env
Fld := Libadalang.Implementation.Basic_Decl_P_Child_Decl_Initial_Env (Node => Self);
--# end
--# expr-done 5289
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1477;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1477;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Initial_Env_60;
--# end
--# property-start GenericSubpInstantiation._env_mappings_61 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10241
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_61
  (Node : Bare_Generic_Subp_Instantiation
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Generic_Subp_Instantiation := Bare_Generic_Subp_Instantiation (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Subp_Instantiation :=
     Internal_Entity_Generic_Subp_Instantiation'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1478 with Inline_Always;
      Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Generic_Subp_Instantiation;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1478 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1478;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5292 '<New EnvAssoc at ast.py:10241>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10241
--# expr-start 5290 '<New DesignatedEnv at ast.py:10241>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10241
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 5290
--# expr-start 5291 '<FieldAccess for name_symbol at ast.py:10242>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10242
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 5291
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 5292
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1478;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1478;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_61;
--# end
--# property-start GenericSubpInstantiation._env_trans_parent_62 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10236
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_62
  (Node : Bare_Generic_Subp_Instantiation
  )
   return Boolean
is
   Self : Bare_Generic_Subp_Instantiation := Bare_Generic_Subp_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_62;
--# end
--# property-start GenericSubpInstantiation._env_do_63 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10249
pragma Warnings (Off, "is not referenced");
 function Env_Do_63
  (Node : Bare_Generic_Subp_Instantiation
  )
   return Bare_Compilation_Unit_Array_Access
is
   Self : Bare_Generic_Subp_Instantiation := Bare_Generic_Subp_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1480 with Inline_Always;
      Fld : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1480 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1480;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5293 '<FieldAccess for populate_dependent_units at ast.py:10249>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10249
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.populate_dependent_units
Fld := Libadalang.Implementation.Basic_Decl_P_Populate_Dependent_Units (Node => Self);
--# end
--# expr-done 5293
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1480;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1480;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_63;
--# end
--# property-start GenericSubpInstantiation._ref_env_nodes_64 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10252
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_64
  (Node : Bare_Generic_Subp_Instantiation
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Generic_Subp_Instantiation := Bare_Generic_Subp_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1481 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1481 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1481;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5294 '<FieldAccess for top_level_use_package_clauses at ast.py:10252>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10252
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_package_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Package_Clauses (Node => Self);
--# end
--# expr-done 5294
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1481;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1481;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_64;
--# end
--# property-start GenericSubpInstantiation._ref_cond_65 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10254
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_65
  (Node : Bare_Generic_Subp_Instantiation
  )
   return Boolean
is
   Self : Bare_Generic_Subp_Instantiation := Bare_Generic_Subp_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5296 '<IsA LibraryItem, Subunit at ast.py:10254>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10254
--# expr-start 5295 '<parent at ast.py:10254>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10254
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 5295
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 5296
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_65;
--# end
--# property-start GenericSubpInstantiation._ref_env_nodes_66 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10257
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_66
  (Node : Bare_Generic_Subp_Instantiation
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Generic_Subp_Instantiation := Bare_Generic_Subp_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1483 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1483 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1483;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5297 '<FieldAccess for top_level_use_type_clauses at ast.py:10257>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10257
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_type_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Type_Clauses (Node => Self);
--# end
--# expr-done 5297
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1483;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1483;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_66;
--# end
--# property-start GenericSubpInstantiation._ref_cond_67 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10259
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_67
  (Node : Bare_Generic_Subp_Instantiation
  )
   return Boolean
is
   Self : Bare_Generic_Subp_Instantiation := Bare_Generic_Subp_Instantiation (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5299 '<IsA LibraryItem, Subunit at ast.py:10259>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10259
--# expr-start 5298 '<parent at ast.py:10259>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10259
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 5298
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 5299
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_67;
--# end
function Generic_Subp_Instantiation_P_Designated_Subp
  (E : Entity
  ) return Internal_Entity
is
      E_Info : Internal_Entity_Info :=
         Shed_Rebindings (E.Info, Node_Env (E.Node));
   Result : constant Internal_Entity :=
      Generic_Subp_Instantiation_P_Designated_Subp (E.Node, E_Info);
begin
      return (Node => Result.Node, Info => Result.Info);
end;
   --
   --  Primitives for Bare_Generic_Renaming_Decl
   --
--# property-start '[dispatcher]GenericRenamingDecl.renaming_name' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Generic_Renaming_Decl_P_Renaming_Name
  (Node : Bare_Generic_Renaming_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Name
is
   Self : Bare_Generic_Renaming_Decl := Bare_Generic_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Generic_Renaming_Decl (Self.Kind) is
               when Ada_Generic_Package_Renaming_Decl =>
                  --# property-call-start GenericPackageRenamingDecl.renaming_name
                  Property_Result := Generic_Package_Renaming_Decl_P_Renaming_Name
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Generic_Subp_Renaming_Decl =>
                  --# property-call-start GenericSubpRenamingDecl.renaming_name
                  Property_Result := Generic_Subp_Renaming_Decl_P_Renaming_Name
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Generic_Renaming_Decl_P_Renaming_Name;
--# end
--# property-start GenericRenamingDecl.resolve /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10501
pragma Warnings (Off, "is not referenced");
 function Generic_Renaming_Decl_P_Resolve
  (Node : Bare_Generic_Renaming_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Generic_Decl
is
   Self : Bare_Generic_Renaming_Decl := Bare_Generic_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Renaming_Decl :=
     Internal_Entity_Generic_Renaming_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Generic_Decl;
            procedure Finalizer_Scope_829 with Inline_Always;
      Var_Expr : Internal_Entity;
Match_0 : Internal_Entity_Generic_Decl;
Match_1 : Internal_Entity_Generic_Renaming_Decl;
Match_2 : Internal_Entity;
Fld : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_1 : Internal_Entity_Name;
Fld_2 : Internal_Entity_Array_Access;
Get_Result : Internal_Entity;
Fld_3 : Internal_Entity_Generic_Decl;
Match_Prefix : Internal_Entity;
Cast_Result : Internal_Entity_Generic_Decl;
Let_Result : Internal_Entity_Generic_Decl;
Scope_Result : Internal_Entity_Generic_Decl;
Cast_Result_1 : Internal_Entity_Generic_Renaming_Decl;
Let_Result_1 : Internal_Entity_Generic_Decl;
Scope_Result_1 : Internal_Entity_Generic_Decl;
Cast_Result_2 : Internal_Entity;
Let_Result_2 : Internal_Entity_Generic_Decl;
Scope_Result_2 : Internal_Entity_Generic_Decl;
Match_Result : Internal_Entity_Generic_Decl;
Scope_Result_3 : Internal_Entity_Generic_Decl;
Result_Var : Internal_Entity_Generic_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Generic_Decl;
            procedure Finalizer_Scope_829 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_829;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5306 '<bind at ast.py:10501>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10501
--# expr-start 5300 '<FieldAccess for node_env at ast.py:10502>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10502
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Node_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 5300
Bound_Env := Fld; Inc_Ref (Bound_Env);
--# expr-start 5303 '<at at ast.py:10503>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10503
--# expr-start 5302 '<FieldAccess for env_elements at ast.py:10503>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10503
--# expr-start 5301 '<FieldAccess for renaming_name at ast.py:10503>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10503
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericRenamingDecl.renaming_name'
Fld_1 := Libadalang.Implementation.Dispatcher_Generic_Renaming_Decl_P_Renaming_Name (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5301
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.env_elements
Fld_2 := Libadalang.Implementation.Expr_P_Env_Elements (Node => Fld_1.Node, Env => Bound_Env, No_Visibility => False, E_Info => Fld_1.Info);
--# end
--# expr-done 5302
Get_Result := Get (Fld_2, 0, True); 
--# expr-done 5303
Var_Expr := Get_Result;
if Var_Expr /= No_Entity then
   --# scope-start
--# expr-start 5305 '<Match at ast.py:10503>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10503
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Var_Expr; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Generic_Package_Decl .. Ada_Generic_Subp_Decl =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Generic_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind gd Match_0
Let_Result := Match_0; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Generic_Package_Renaming_Decl .. Ada_Generic_Subp_Renaming_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Generic_Renaming_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind grd Match_1
--# expr-start 5304 '<FieldAccess for resolve at ast.py:10505>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10505
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericRenamingDecl.resolve
Fld_3 := Libadalang.Implementation.Generic_Renaming_Decl_P_Resolve (Node => Match_1.Node, E_Info => Match_1.Info);
--# end
--# expr-done 5304
Let_Result_1 := Fld_3; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abort_Absent .. Ada_For_Loop_Var_Decl | Ada_Generic_Package_Instantiation .. Ada_Generic_Subp_Instantiation | Ada_Label_Decl .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind _ Match_2
Let_Result_2 := No_Entity_Generic_Decl; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 5305
Scope_Result_3 := Match_Result;
   --# end
   Result_Var := Scope_Result_3;
else
   Result_Var := No_Entity_Generic_Decl;
end if;
Dyn_Var_Bind_Result := Result_Var; 
--# expr-done 5306
      Property_Result := Dyn_Var_Bind_Result;
   --# end
      Finalizer_Scope_829;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_829;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Renaming_Decl_P_Resolve;
--# end
--# property-start GenericRenamingDecl.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10512
pragma Warnings (Off, "is not referenced");
 function Generic_Renaming_Decl_P_Xref_Entry_Point
  (Node : Bare_Generic_Renaming_Decl
  )
   return Boolean
is
   Self : Bare_Generic_Renaming_Decl := Bare_Generic_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Renaming_Decl_P_Xref_Entry_Point;
--# end
--# property-start GenericRenamingDecl.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10513
pragma Warnings (Off, "is not referenced");
 function Generic_Renaming_Decl_P_Xref_Equation
  (Node : Bare_Generic_Renaming_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Generic_Renaming_Decl := Bare_Generic_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Renaming_Decl :=
     Internal_Entity_Generic_Renaming_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_831 with Inline_Always;
      Fld : Internal_Entity_Name;
Fld_1 : Logic_Equation;
            procedure Finalizer_Scope_831 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_831;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5308 '<FieldAccess for xref_no_overloading at ast.py:10513>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10513
--# expr-start 5307 '<FieldAccess for renaming_name at ast.py:10513>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10513
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericRenamingDecl.renaming_name'
Fld := Libadalang.Implementation.Dispatcher_Generic_Renaming_Decl_P_Renaming_Name (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5307
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_1 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 5308
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_831;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_831;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Renaming_Decl_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Generic_Package_Renaming_Decl
   --
      procedure Generic_Package_Renaming_Decl_Pre_Env_Actions
        (Self            : Bare_Generic_Package_Renaming_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Dummy : Boolean;
      begin
         Dummy := Libadalang.Implementation.Env_Do_87 (Node => 
Self);
      end;
      declare
         Env : constant Internal_Designated_Env :=
           Libadalang.Implementation.Initial_Env_88 (Node => 
Self);
      begin
         Set_Initial_Env
           (Self,
            State,
            Env,
            DSL_Location => "ast.py:10531");
      end;
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_89 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:10533");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_90 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      declare
         Dummy : Bare_Compilation_Unit_Array_Access;
      begin
         Dummy := Libadalang.Implementation.Env_Do_91 (Node => 
Self);
         Dec_Ref (Dummy);
      end;
      if Libadalang.Implementation.Ref_Cond_93 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_92 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Use_Package_Name_Designated_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_95 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_94 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Name_Designated_Type_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      end;
      procedure Initialize_Fields_For_Generic_Package_Renaming_Decl
        (Self : Bare_Generic_Package_Renaming_Decl
         ; Generic_Package_Renaming_Decl_F_Name : Bare_Defining_Name
         ; Generic_Package_Renaming_Decl_F_Renames : Bare_Name
         ; Generic_Package_Renaming_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Self.Generic_Package_Renaming_Decl_F_Name := Generic_Package_Renaming_Decl_F_Name;
            Self.Generic_Package_Renaming_Decl_F_Renames := Generic_Package_Renaming_Decl_F_Renames;
            Self.Generic_Package_Renaming_Decl_F_Aspects := Generic_Package_Renaming_Decl_F_Aspects;
      end Initialize_Fields_For_Generic_Package_Renaming_Decl;
   function Generic_Package_Renaming_Decl_F_Name
     (Node : Bare_Generic_Package_Renaming_Decl) return Bare_Defining_Name
   is
   begin
         return Node.Generic_Package_Renaming_Decl_F_Name;
   end;
   function Generic_Package_Renaming_Decl_F_Renames
     (Node : Bare_Generic_Package_Renaming_Decl) return Bare_Name
   is
   begin
         return Node.Generic_Package_Renaming_Decl_F_Renames;
   end;
--# property-start GenericPackageRenamingDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10524
pragma Warnings (Off, "is not referenced");
 function Generic_Package_Renaming_Decl_P_Defining_Names
  (Node : Bare_Generic_Package_Renaming_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Generic_Package_Renaming_Decl := Bare_Generic_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Package_Renaming_Decl :=
     Internal_Entity_Generic_Package_Renaming_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_832 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_832 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_832;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5309 '<FieldAccess for name at ast.py:10524>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10524
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Generic_Package_Renaming_Decl_F_Name, Info => Ent.Info);
--# expr-done 5309
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_832;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_832;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Package_Renaming_Decl_P_Defining_Names;
--# end
--# property-start GenericPackageRenamingDecl.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10525
pragma Warnings (Off, "is not referenced");
 function Generic_Package_Renaming_Decl_P_Defining_Env
  (Node : Bare_Generic_Package_Renaming_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Generic_Package_Renaming_Decl := Bare_Generic_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Package_Renaming_Decl :=
     Internal_Entity_Generic_Package_Renaming_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_833 with Inline_Always;
      Fld : Internal_Entity_Generic_Decl;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_833 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_833;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5311 '<FieldAccess for defining_env at ast.py:10525>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10525
--# expr-start 5310 '<FieldAccess for resolve at ast.py:10525>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10525
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericRenamingDecl.resolve
Fld := Libadalang.Implementation.Generic_Renaming_Decl_P_Resolve (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5310
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Fld.Node, Origin => Origin, Include_Ud_Indexing => Include_Ud_Indexing, E_Info => Fld.Info);
--# end
--# expr-done 5311
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_833;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_833;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Package_Renaming_Decl_P_Defining_Env;
--# end
--# property-start GenericPackageRenamingDecl.renaming_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10526
pragma Warnings (Off, "is not referenced");
 function Generic_Package_Renaming_Decl_P_Renaming_Name
  (Node : Bare_Generic_Package_Renaming_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Name
is
   Self : Bare_Generic_Package_Renaming_Decl := Bare_Generic_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Package_Renaming_Decl :=
     Internal_Entity_Generic_Package_Renaming_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Name;
      Fld : Internal_Entity_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5312 '<FieldAccess for renames at ast.py:10526>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10526
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Generic_Package_Renaming_Decl_F_Renames, Info => Ent.Info);
--# expr-done 5312
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Package_Renaming_Decl_P_Renaming_Name;
--# end
--# property-start GenericPackageRenamingDecl._env_do_87 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10529
pragma Warnings (Off, "is not referenced");
 function Env_Do_87
  (Node : Bare_Generic_Package_Renaming_Decl
  )
   return Boolean
is
   Self : Bare_Generic_Package_Renaming_Decl := Bare_Generic_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5313 '<FieldAccess for env_hook at ast.py:10529>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10529
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.env_hook
Fld := Libadalang.Implementation.Ada_Node_P_Env_Hook (Node => Self);
--# end
--# expr-done 5313
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_87;
--# end
--# property-start GenericPackageRenamingDecl._initial_env_88 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10531
pragma Warnings (Off, "is not referenced");
 function Initial_Env_88
  (Node : Bare_Generic_Package_Renaming_Decl
  )
   return Internal_Designated_Env
is
   Self : Bare_Generic_Package_Renaming_Decl := Bare_Generic_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_1505 with Inline_Always;
      Fld : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_1505 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1505;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5314 '<FieldAccess for child_decl_initial_env at ast.py:10531>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10531
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.child_decl_initial_env
Fld := Libadalang.Implementation.Basic_Decl_P_Child_Decl_Initial_Env (Node => Self);
--# end
--# expr-done 5314
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1505;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1505;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Initial_Env_88;
--# end
--# property-start GenericPackageRenamingDecl._env_mappings_89 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10533
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_89
  (Node : Bare_Generic_Package_Renaming_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Generic_Package_Renaming_Decl := Bare_Generic_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Package_Renaming_Decl :=
     Internal_Entity_Generic_Package_Renaming_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1506 with Inline_Always;
      Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Generic_Package_Renaming_Decl;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1506 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1506;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5317 '<New EnvAssoc at ast.py:10533>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10533
--# expr-start 5315 '<New DesignatedEnv at ast.py:10533>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10533
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 5315
--# expr-start 5316 '<FieldAccess for name_symbol at ast.py:10534>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10534
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 5316
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 5317
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1506;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1506;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_89;
--# end
--# property-start GenericPackageRenamingDecl._env_trans_parent_90 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10528
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_90
  (Node : Bare_Generic_Package_Renaming_Decl
  )
   return Boolean
is
   Self : Bare_Generic_Package_Renaming_Decl := Bare_Generic_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_90;
--# end
--# property-start GenericPackageRenamingDecl._env_do_91 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10540
pragma Warnings (Off, "is not referenced");
 function Env_Do_91
  (Node : Bare_Generic_Package_Renaming_Decl
  )
   return Bare_Compilation_Unit_Array_Access
is
   Self : Bare_Generic_Package_Renaming_Decl := Bare_Generic_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1508 with Inline_Always;
      Fld : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1508 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1508;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5318 '<FieldAccess for populate_dependent_units at ast.py:10540>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10540
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.populate_dependent_units
Fld := Libadalang.Implementation.Basic_Decl_P_Populate_Dependent_Units (Node => Self);
--# end
--# expr-done 5318
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1508;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1508;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_91;
--# end
--# property-start GenericPackageRenamingDecl._ref_env_nodes_92 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10543
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_92
  (Node : Bare_Generic_Package_Renaming_Decl
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Generic_Package_Renaming_Decl := Bare_Generic_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1509 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1509 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1509;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5319 '<FieldAccess for top_level_use_package_clauses at ast.py:10543>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10543
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_package_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Package_Clauses (Node => Self);
--# end
--# expr-done 5319
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1509;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1509;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_92;
--# end
--# property-start GenericPackageRenamingDecl._ref_cond_93 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10545
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_93
  (Node : Bare_Generic_Package_Renaming_Decl
  )
   return Boolean
is
   Self : Bare_Generic_Package_Renaming_Decl := Bare_Generic_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5321 '<IsA LibraryItem, Subunit at ast.py:10545>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10545
--# expr-start 5320 '<parent at ast.py:10545>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10545
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 5320
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 5321
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_93;
--# end
--# property-start GenericPackageRenamingDecl._ref_env_nodes_94 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10548
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_94
  (Node : Bare_Generic_Package_Renaming_Decl
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Generic_Package_Renaming_Decl := Bare_Generic_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1511 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1511 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1511;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5322 '<FieldAccess for top_level_use_type_clauses at ast.py:10548>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10548
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_type_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Type_Clauses (Node => Self);
--# end
--# expr-done 5322
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1511;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1511;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_94;
--# end
--# property-start GenericPackageRenamingDecl._ref_cond_95 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10550
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_95
  (Node : Bare_Generic_Package_Renaming_Decl
  )
   return Boolean
is
   Self : Bare_Generic_Package_Renaming_Decl := Bare_Generic_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5324 '<IsA LibraryItem, Subunit at ast.py:10550>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10550
--# expr-start 5323 '<parent at ast.py:10550>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10550
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 5323
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 5324
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_95;
--# end
   --
   --  Primitives for Bare_Generic_Subp_Renaming_Decl
   --
      procedure Generic_Subp_Renaming_Decl_Pre_Env_Actions
        (Self            : Bare_Generic_Subp_Renaming_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Dummy : Boolean;
      begin
         Dummy := Libadalang.Implementation.Env_Do_96 (Node => 
Self);
      end;
      declare
         Env : constant Internal_Designated_Env :=
           Libadalang.Implementation.Initial_Env_97 (Node => 
Self);
      begin
         Set_Initial_Env
           (Self,
            State,
            Env,
            DSL_Location => "ast.py:10570");
      end;
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_98 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:10572");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_99 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      declare
         Dummy : Bare_Compilation_Unit_Array_Access;
      begin
         Dummy := Libadalang.Implementation.Env_Do_100 (Node => 
Self);
         Dec_Ref (Dummy);
      end;
      if Libadalang.Implementation.Ref_Cond_102 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_101 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Use_Package_Name_Designated_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_104 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_103 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Name_Designated_Type_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      end;
      procedure Initialize_Fields_For_Generic_Subp_Renaming_Decl
        (Self : Bare_Generic_Subp_Renaming_Decl
         ; Generic_Subp_Renaming_Decl_F_Kind : Bare_Subp_Kind
         ; Generic_Subp_Renaming_Decl_F_Name : Bare_Defining_Name
         ; Generic_Subp_Renaming_Decl_F_Renames : Bare_Name
         ; Generic_Subp_Renaming_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Self.Generic_Subp_Renaming_Decl_F_Kind := Generic_Subp_Renaming_Decl_F_Kind;
            Self.Generic_Subp_Renaming_Decl_F_Name := Generic_Subp_Renaming_Decl_F_Name;
            Self.Generic_Subp_Renaming_Decl_F_Renames := Generic_Subp_Renaming_Decl_F_Renames;
            Self.Generic_Subp_Renaming_Decl_F_Aspects := Generic_Subp_Renaming_Decl_F_Aspects;
      end Initialize_Fields_For_Generic_Subp_Renaming_Decl;
   function Generic_Subp_Renaming_Decl_F_Kind
     (Node : Bare_Generic_Subp_Renaming_Decl) return Bare_Subp_Kind
   is
   begin
         return Node.Generic_Subp_Renaming_Decl_F_Kind;
   end;
   function Generic_Subp_Renaming_Decl_F_Name
     (Node : Bare_Generic_Subp_Renaming_Decl) return Bare_Defining_Name
   is
   begin
         return Node.Generic_Subp_Renaming_Decl_F_Name;
   end;
   function Generic_Subp_Renaming_Decl_F_Renames
     (Node : Bare_Generic_Subp_Renaming_Decl) return Bare_Name
   is
   begin
         return Node.Generic_Subp_Renaming_Decl_F_Renames;
   end;
--# property-start GenericSubpRenamingDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10598
pragma Warnings (Off, "is not referenced");
 function Generic_Subp_Renaming_Decl_P_Defining_Names
  (Node : Bare_Generic_Subp_Renaming_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Generic_Subp_Renaming_Decl := Bare_Generic_Subp_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Subp_Renaming_Decl :=
     Internal_Entity_Generic_Subp_Renaming_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_835 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_835 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_835;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5325 '<FieldAccess for name at ast.py:10598>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10598
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Generic_Subp_Renaming_Decl_F_Name, Info => Ent.Info);
--# expr-done 5325
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_835;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_835;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Subp_Renaming_Decl_P_Defining_Names;
--# end
--# property-start GenericSubpRenamingDecl.renaming_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10599
pragma Warnings (Off, "is not referenced");
 function Generic_Subp_Renaming_Decl_P_Renaming_Name
  (Node : Bare_Generic_Subp_Renaming_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Name
is
   Self : Bare_Generic_Subp_Renaming_Decl := Bare_Generic_Subp_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Subp_Renaming_Decl :=
     Internal_Entity_Generic_Subp_Renaming_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Name;
      Fld : Internal_Entity_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5326 '<FieldAccess for renames at ast.py:10599>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10599
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Generic_Subp_Renaming_Decl_F_Renames, Info => Ent.Info);
--# expr-done 5326
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Generic_Subp_Renaming_Decl_P_Renaming_Name;
--# end
--# property-start GenericSubpRenamingDecl._env_do_96 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10568
pragma Warnings (Off, "is not referenced");
 function Env_Do_96
  (Node : Bare_Generic_Subp_Renaming_Decl
  )
   return Boolean
is
   Self : Bare_Generic_Subp_Renaming_Decl := Bare_Generic_Subp_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5327 '<FieldAccess for env_hook at ast.py:10568>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10568
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.env_hook
Fld := Libadalang.Implementation.Ada_Node_P_Env_Hook (Node => Self);
--# end
--# expr-done 5327
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_96;
--# end
--# property-start GenericSubpRenamingDecl._initial_env_97 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10570
pragma Warnings (Off, "is not referenced");
 function Initial_Env_97
  (Node : Bare_Generic_Subp_Renaming_Decl
  )
   return Internal_Designated_Env
is
   Self : Bare_Generic_Subp_Renaming_Decl := Bare_Generic_Subp_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_1514 with Inline_Always;
      Fld : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_1514 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1514;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5328 '<FieldAccess for child_decl_initial_env at ast.py:10570>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10570
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.child_decl_initial_env
Fld := Libadalang.Implementation.Basic_Decl_P_Child_Decl_Initial_Env (Node => Self);
--# end
--# expr-done 5328
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1514;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1514;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Initial_Env_97;
--# end
--# property-start GenericSubpRenamingDecl._env_mappings_98 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10572
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_98
  (Node : Bare_Generic_Subp_Renaming_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Generic_Subp_Renaming_Decl := Bare_Generic_Subp_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Generic_Subp_Renaming_Decl :=
     Internal_Entity_Generic_Subp_Renaming_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1515 with Inline_Always;
      Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Generic_Subp_Renaming_Decl;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1515 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1515;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5331 '<New EnvAssoc at ast.py:10572>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10572
--# expr-start 5329 '<New DesignatedEnv at ast.py:10572>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10572
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 5329
--# expr-start 5330 '<FieldAccess for name_symbol at ast.py:10573>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10573
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 5330
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 5331
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1515;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1515;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_98;
--# end
--# property-start GenericSubpRenamingDecl._env_trans_parent_99 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10567
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_99
  (Node : Bare_Generic_Subp_Renaming_Decl
  )
   return Boolean
is
   Self : Bare_Generic_Subp_Renaming_Decl := Bare_Generic_Subp_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_99;
--# end
--# property-start GenericSubpRenamingDecl._env_do_100 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10579
pragma Warnings (Off, "is not referenced");
 function Env_Do_100
  (Node : Bare_Generic_Subp_Renaming_Decl
  )
   return Bare_Compilation_Unit_Array_Access
is
   Self : Bare_Generic_Subp_Renaming_Decl := Bare_Generic_Subp_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1517 with Inline_Always;
      Fld : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1517 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1517;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5332 '<FieldAccess for populate_dependent_units at ast.py:10579>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10579
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.populate_dependent_units
Fld := Libadalang.Implementation.Basic_Decl_P_Populate_Dependent_Units (Node => Self);
--# end
--# expr-done 5332
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1517;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1517;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_100;
--# end
--# property-start GenericSubpRenamingDecl._ref_env_nodes_101 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10582
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_101
  (Node : Bare_Generic_Subp_Renaming_Decl
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Generic_Subp_Renaming_Decl := Bare_Generic_Subp_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1518 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1518 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1518;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5333 '<FieldAccess for top_level_use_package_clauses at ast.py:10582>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10582
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_package_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Package_Clauses (Node => Self);
--# end
--# expr-done 5333
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1518;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1518;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_101;
--# end
--# property-start GenericSubpRenamingDecl._ref_cond_102 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10584
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_102
  (Node : Bare_Generic_Subp_Renaming_Decl
  )
   return Boolean
is
   Self : Bare_Generic_Subp_Renaming_Decl := Bare_Generic_Subp_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5335 '<IsA LibraryItem, Subunit at ast.py:10584>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10584
--# expr-start 5334 '<parent at ast.py:10584>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10584
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 5334
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 5335
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_102;
--# end
--# property-start GenericSubpRenamingDecl._ref_env_nodes_103 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10587
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_103
  (Node : Bare_Generic_Subp_Renaming_Decl
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Generic_Subp_Renaming_Decl := Bare_Generic_Subp_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1520 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1520 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1520;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5336 '<FieldAccess for top_level_use_type_clauses at ast.py:10587>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10587
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_type_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Type_Clauses (Node => Self);
--# end
--# expr-done 5336
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1520;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1520;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_103;
--# end
--# property-start GenericSubpRenamingDecl._ref_cond_104 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10589
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_104
  (Node : Bare_Generic_Subp_Renaming_Decl
  )
   return Boolean
is
   Self : Bare_Generic_Subp_Renaming_Decl := Bare_Generic_Subp_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5338 '<IsA LibraryItem, Subunit at ast.py:10589>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10589
--# expr-start 5337 '<parent at ast.py:10589>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10589
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 5337
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 5338
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_104;
--# end
   --
   --  Primitives for Bare_Label_Decl
   --
      procedure Label_Decl_Pre_Env_Actions
        (Self            : Bare_Label_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_152 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:19380");
         Dec_Ref (Mapping.Dest_Env);
      end;
      end;
      procedure Initialize_Fields_For_Label_Decl
        (Self : Bare_Label_Decl
         ; Label_Decl_F_Name : Bare_Defining_Name
        ) is
      begin
            Self.Label_Decl_F_Name := Label_Decl_F_Name;
      end Initialize_Fields_For_Label_Decl;
   function Label_Decl_F_Name
     (Node : Bare_Label_Decl) return Bare_Defining_Name
   is
   begin
         return Node.Label_Decl_F_Name;
   end;
--# property-start LabelDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19378
pragma Warnings (Off, "is not referenced");
 function Label_Decl_P_Defining_Names
  (Node : Bare_Label_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Label_Decl := Bare_Label_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Label_Decl :=
     Internal_Entity_Label_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1355 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1355 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1355;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5339 '<FieldAccess for name at ast.py:19378>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19378
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Label_Decl_F_Name, Info => Ent.Info);
--# expr-done 5339
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1355;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1355;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Label_Decl_P_Defining_Names;
--# end
--# property-start LabelDecl._env_mappings_152 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19380
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_152
  (Node : Bare_Label_Decl
  )
   return Internal_Env_Assoc
is
   Self : Bare_Label_Decl := Bare_Label_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1569 with Inline_Always;
      Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Label_Decl;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1569 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1569;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5342 '<New EnvAssoc at ast.py:19380>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19380
--# expr-start 5340 '<New DesignatedEnv at ast.py:19380>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19380
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 5340
--# expr-start 5341 '<FieldAccess for name_symbol at ast.py:19380>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19380
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Self);
--# end
--# expr-done 5341
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 5342
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1569;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1569;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_152;
--# end
   --
   --  Primitives for Bare_Named_Stmt_Decl
   --
      procedure Initialize_Fields_For_Named_Stmt_Decl
        (Self : Bare_Named_Stmt_Decl
         ; Named_Stmt_Decl_F_Name : Bare_Defining_Name
        ) is
      begin
            Self.Named_Stmt_Decl_F_Name := Named_Stmt_Decl_F_Name;
      end Initialize_Fields_For_Named_Stmt_Decl;
   function Named_Stmt_Decl_F_Name
     (Node : Bare_Named_Stmt_Decl) return Bare_Defining_Name
   is
   begin
         return Node.Named_Stmt_Decl_F_Name;
   end;
--# property-start NamedStmtDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19418
pragma Warnings (Off, "is not referenced");
 function Named_Stmt_Decl_P_Defining_Names
  (Node : Bare_Named_Stmt_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Named_Stmt_Decl := Bare_Named_Stmt_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Named_Stmt_Decl :=
     Internal_Entity_Named_Stmt_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1358 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1358 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1358;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5343 '<FieldAccess for name at ast.py:19418>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19418
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Named_Stmt_Decl_F_Name, Info => Ent.Info);
--# expr-done 5343
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1358;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1358;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Named_Stmt_Decl_P_Defining_Names;
--# end
--# property-start NamedStmtDecl.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19419
pragma Warnings (Off, "is not referenced");
 function Named_Stmt_Decl_P_Defining_Env
  (Node : Bare_Named_Stmt_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Named_Stmt_Decl := Bare_Named_Stmt_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1359 with Inline_Always;
      Fld : Bare_Ada_Node;
Cast_Result : Bare_Named_Stmt;
Fld_1 : Bare_Composite_Stmt;
Fld_2 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1359 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_1359;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5347 '<FieldAccess for children_env at ast.py:19419>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19419
--# expr-start 5346 '<FieldAccess for stmt at ast.py:19419>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19419
--# expr-start 5345 '<Cast to NamedStmt at ast.py:19419>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19419
--# expr-start 5344 '<parent at ast.py:19419>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19419
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 5344
   if Fld = null
      or else Fld.Kind in Ada_Named_Stmt_Range
   then
      Cast_Result := Fld;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 5345
      if Cast_Result = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Cast_Result.Named_Stmt_F_Stmt;
--# expr-done 5346
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Children_Env (Node => Fld_1);
--# expr-done 5347
      Property_Result := Fld_2;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1359;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1359;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Named_Stmt_Decl_P_Defining_Env;
--# end
   --
   --  Primitives for Bare_Number_Decl
   --
      procedure Number_Decl_Pre_Env_Actions
        (Self            : Bare_Number_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mappings : Internal_Env_Assoc_Array_Access :=
            Libadalang.Implementation.Env_Mappings_43 (Node => 
Self);
      begin
         for Mapping of Mappings.Items loop
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:9588");
         end loop;
         Dec_Ref (Mappings);
      end;
      end;
      procedure Initialize_Fields_For_Number_Decl
        (Self : Bare_Number_Decl
         ; Number_Decl_F_Ids : Bare_Defining_Name_List
         ; Number_Decl_F_Expr : Bare_Expr
        ) is
      begin
            Self.Number_Decl_F_Ids := Number_Decl_F_Ids;
            Self.Number_Decl_F_Expr := Number_Decl_F_Expr;
      end Initialize_Fields_For_Number_Decl;
   function Number_Decl_F_Ids
     (Node : Bare_Number_Decl) return Bare_Defining_Name_List
   is
   begin
         return Node.Number_Decl_F_Ids;
   end;
   function Number_Decl_F_Expr
     (Node : Bare_Number_Decl) return Bare_Expr
   is
   begin
         return Node.Number_Decl_F_Expr;
   end;
--# property-start NumberDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9586
pragma Warnings (Off, "is not referenced");
 function Number_Decl_P_Defining_Names
  (Node : Bare_Number_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Number_Decl := Bare_Number_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Number_Decl :=
     Internal_Entity_Number_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_759 with Inline_Always;
      Fld : Internal_Entity_Defining_Name_List;
Fld_1 : Bare_Defining_Name_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity_Defining_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Defining_Name;
Item_190 : Internal_Entity_Defining_Name;
Bare_Item_190 : Bare_Defining_Name;
Map_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_759 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_759;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5349 '<Map at ast.py:9586>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9586
   --# expr-start 5348 '<FieldAccess for ids at ast.py:9586>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9586
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name_List (Node => Ent.Node.Number_Decl_F_Ids, Info => Ent.Info);
--# expr-done 5348
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Internal_Entity_Defining_Name_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Internal_Entity_Defining_Name_Array (0);
      else
      declare
         Collection : constant Bare_Defining_Name_List := Fld_1;
      begin
         for Untyped_Item_190 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_190 := Untyped_Item_190; 
Is_Null := Bare_Item_190 = null; 
if Is_Null then
   If_Result := No_Entity_Defining_Name;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_190); 
   If_Result := As_Entity;
end if;
                  Item_190 := If_Result; 
   --# scope-start
               --# bind id Item_190
         declare
            Item_To_Append : constant Internal_Entity_Defining_Name := Item_190;
         begin
            Internal_Entity_Defining_Name_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Defining_Name_Array
           (Items_Count => Natural (Internal_Entity_Defining_Name_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Defining_Name_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Defining_Name_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Defining_Name_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 5349
      Property_Result := Map_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_759;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_759;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Number_Decl_P_Defining_Names;
--# end
--# property-start NumberDecl.expr_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9591
pragma Warnings (Off, "is not referenced");
 function Number_Decl_P_Expr_Type
  (Node : Bare_Number_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Number_Decl := Bare_Number_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Number_Decl :=
     Internal_Entity_Number_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Block_Var : Boolean;
Block_Var_1 : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Expr;
Fld_1 : Boolean;
Fld_2 : Bare_Expr;
Fld_3 : Internal_Entity;
Is_Null : Boolean;
If_Result : Boolean;
Fld_4 : Bare_Expr;
Fld_5 : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
If_Result_1 : Internal_Entity_Base_Type_Decl;
Fld_6 : Internal_Entity;
Fld_7 : Internal_Entity;
Fld_8 : Boolean;
If_Result_2 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5365 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5355 '<If at ast.py:9592>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9592
--# expr-start 5352 '<IsNull at ast.py:9592>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9592
--# expr-start 5351 '<FieldAccess for type_val at ast.py:9592>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9592
--# expr-start 5350 '<FieldAccess for expr at ast.py:9592>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9592
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Number_Decl_F_Expr;
--# expr-done 5350
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.type_val
Fld_3 := Libadalang.Implementation.Expr_P_Type_Val (Node => Fld_2);
--# end
--# expr-done 5351
Is_Null := Fld_3.Node = null; 
--# expr-done 5352
if Is_Null then
   --# expr-start 5354 '<FieldAccess for resolve_names at ast.py:9593>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9593
--# expr-start 5353 '<FieldAccess for expr at ast.py:9593>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9593
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Number_Decl_F_Expr, Info => Ent.Info);
--# expr-done 5353
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.resolve_names
Fld_1 := Libadalang.Implementation.Ada_Node_P_Resolve_Names (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 5354
   If_Result := Fld_1;
else
   If_Result := True;
end if;
--# expr-done 5355
Block_Var := If_Result; 
--# bind p Block_Var
--# expr-start 5359 '<If at ast.py:9596>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9596
if Block_Var then
   --# expr-start 5358 '<Cast to BaseTypeDecl.entity at ast.py:9597>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9597
--# expr-start 5357 '<FieldAccess for type_val at ast.py:9597>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9597
--# expr-start 5356 '<FieldAccess for expr at ast.py:9597>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9597
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Self.Number_Decl_F_Expr;
--# expr-done 5356
      if Fld_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.type_val
Fld_5 := Libadalang.Implementation.Expr_P_Type_Val (Node => Fld_4);
--# end
--# expr-done 5357
   if Fld_5.Node = null
      or else Fld_5.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 5358
   If_Result_1 := Cast_Result;
else
   If_Result_1 := No_Entity_Base_Type_Decl;
end if;
--# expr-done 5359
Block_Var_1 := If_Result_1; 
--# bind typ Block_Var_1
--# expr-start 5364 '<Cast to BaseTypeDecl.entity at ast.py:9600>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9600
--# expr-start 5363 '<If at ast.py:9600>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9600
--# expr-start 5360 '<FieldAccess for is_int_type at ast.py:9600>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9600
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_int_type'
Fld_8 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Int_Type (Node => Block_Var_1.Node, Origin => Origin, E_Info => Block_Var_1.Info);
--# end
--# expr-done 5360
if Fld_8 then
   --# expr-start 5361 '<FieldAccess for universal_int_type at ast.py:9601>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9601
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.universal_int_type
Fld_6 := Libadalang.Implementation.Ada_Node_P_Universal_Int_Type (Node => Self);
--# end
--# expr-done 5361
   If_Result_2 := Fld_6;
else
   --# expr-start 5362 '<FieldAccess for universal_real_type at ast.py:9602>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9602
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.universal_real_type
Fld_7 := Libadalang.Implementation.Ada_Node_P_Universal_Real_Type (Node => Self);
--# end
--# expr-done 5362
   If_Result_2 := Fld_7;
end if;
--# expr-done 5363
   if If_Result_2.Node = null
      or else If_Result_2.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Base_Type_Decl
        (Node => If_Result_2.Node,
         Info => If_Result_2.Info);
   else
         Cast_Result_1 := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 5364
Let_Result := Cast_Result_1; 
--# end
--# expr-done 5365
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Number_Decl_P_Expr_Type;
--# end
--# property-start NumberDecl.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9604
pragma Warnings (Off, "is not referenced");
 function Number_Decl_P_Xref_Entry_Point
  (Node : Bare_Number_Decl
  )
   return Boolean
is
   Self : Bare_Number_Decl := Bare_Number_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Number_Decl_P_Xref_Entry_Point;
--# end
--# property-start NumberDecl.is_static_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9606
pragma Warnings (Off, "is not referenced");
 function Number_Decl_P_Is_Static_Decl
  (Node : Bare_Number_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Number_Decl := Bare_Number_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Number_Decl_P_Is_Static_Decl;
--# end
--# property-start NumberDecl.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9608
pragma Warnings (Off, "is not referenced");
 function Number_Decl_P_Xref_Equation
  (Node : Bare_Number_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Number_Decl := Bare_Number_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Number_Decl :=
     Internal_Entity_Number_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_763 with Inline_Always;
      Fld : Bare_Expr;
Fld_1 : Logic_Var;
Node_For_Entity : Bare_Ada_Node;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_2 : Internal_Entity_Expr;
Fld_3 : Logic_Equation;
And_Pred : Logic_Equation;
            procedure Finalizer_Scope_763 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_3);
                     Dec_Ref (And_Pred);
            end Finalizer_Scope_763;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5372 '<And at ast.py:9609>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9609
--# expr-start 5369 '<Bind at ast.py:9609>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9609
--# expr-start 5367 '<FieldAccess for expected_type_var at ast.py:9609>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9609
--# expr-start 5366 '<FieldAccess for expr at ast.py:9609>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9609
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Number_Decl_F_Expr;
--# expr-done 5366
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld);
--# expr-done 5367
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 5368 '<No AdaNode at ast.py:9609>' Node_For_Entity /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9609
Node_For_Entity := No_Bare_Ada_Node;
--# expr-done 5368
Is_Null := Node_For_Entity = null; 
if Is_Null then
   If_Result := No_Entity;
else
As_Entity := (Info => E_Info, Node => Node_For_Entity); 
   If_Result := As_Entity;
end if;
Bind_Result := Solver.Create_Assign (Fld_1, If_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9609") else null)); 
--# expr-done 5369
--# expr-start 5371 '<FieldAccess for sub_equation at ast.py:9610>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9610
--# expr-start 5370 '<FieldAccess for expr at ast.py:9610>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9610
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Expr (Node => Ent.Node.Number_Decl_F_Expr, Info => Ent.Info);
--# expr-done 5370
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_2.Node, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 5371
And_Pred := Create_And (Bind_Result, Fld_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9609") else null)); 
--# expr-done 5372
      Property_Result := And_Pred;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_763;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_763;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Number_Decl_P_Xref_Equation;
--# end
--# property-start NumberDecl.is_constant_object /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9613
pragma Warnings (Off, "is not referenced");
 function Number_Decl_P_Is_Constant_Object
  (Node : Bare_Number_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Number_Decl := Bare_Number_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Number_Decl_P_Is_Constant_Object;
--# end
--# property-start NumberDecl._env_mappings_43 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9588
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_43
  (Node : Bare_Number_Decl
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Number_Decl := Bare_Number_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1460 with Inline_Always;
      Fld : Bare_Defining_Name_List;
Cast_Expr : Bare_Number_Decl;
Cast_Result : Bare_Ada_Node;
Fld_1 : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1460 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1460;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5374 '<FieldAccess for env_mappings at ast.py:9588>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9588
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 5373 '<FieldAccess for ids at ast.py:9588>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9588
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Number_Decl_F_Ids;
--# expr-done 5373
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# property-call-start AdaNode.env_mappings
Fld_1 := Libadalang.Implementation.Ada_Node_P_Env_Mappings (Node => Self, Defining_Names => Fld, Value => Cast_Result);
--# end
--# expr-done 5374
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1460;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1460;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_43;
--# end
   --
   --  Primitives for Bare_Object_Decl
   --
      procedure Object_Decl_Pre_Env_Actions
        (Self            : Bare_Object_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mappings : Internal_Env_Assoc_Array_Access :=
            Libadalang.Implementation.Env_Mappings_44 (Node => 
Self);
      begin
         for Mapping of Mappings.Items loop
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:9632");
         end loop;
         Dec_Ref (Mappings);
      end;
      end;
      procedure Initialize_Fields_For_Object_Decl
        (Self : Bare_Object_Decl
         ; Object_Decl_F_Ids : Bare_Defining_Name_List
         ; Object_Decl_F_Has_Aliased : Bare_Aliased_Node
         ; Object_Decl_F_Has_Constant : Bare_Constant_Node
         ; Object_Decl_F_Mode : Bare_Mode
         ; Object_Decl_F_Type_Expr : Bare_Type_Expr
         ; Object_Decl_F_Default_Expr : Bare_Expr
         ; Object_Decl_F_Renaming_Clause : Bare_Renaming_Clause
         ; Object_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Self.Object_Decl_F_Ids := Object_Decl_F_Ids;
            Self.Object_Decl_F_Has_Aliased := Object_Decl_F_Has_Aliased;
            Self.Object_Decl_F_Has_Constant := Object_Decl_F_Has_Constant;
            Self.Object_Decl_F_Mode := Object_Decl_F_Mode;
            Self.Object_Decl_F_Type_Expr := Object_Decl_F_Type_Expr;
            Self.Object_Decl_F_Default_Expr := Object_Decl_F_Default_Expr;
            Self.Object_Decl_F_Renaming_Clause := Object_Decl_F_Renaming_Clause;
            Self.Object_Decl_F_Aspects := Object_Decl_F_Aspects;
      end Initialize_Fields_For_Object_Decl;
   function Object_Decl_F_Ids
     (Node : Bare_Object_Decl) return Bare_Defining_Name_List
   is
   begin
         return Node.Object_Decl_F_Ids;
   end;
   function Object_Decl_F_Has_Aliased
     (Node : Bare_Object_Decl) return Bare_Aliased_Node
   is
   begin
         return Node.Object_Decl_F_Has_Aliased;
   end;
   function Object_Decl_F_Has_Constant
     (Node : Bare_Object_Decl) return Bare_Constant_Node
   is
   begin
         return Node.Object_Decl_F_Has_Constant;
   end;
   function Object_Decl_F_Mode
     (Node : Bare_Object_Decl) return Bare_Mode
   is
   begin
         return Node.Object_Decl_F_Mode;
   end;
   function Object_Decl_F_Type_Expr
     (Node : Bare_Object_Decl) return Bare_Type_Expr
   is
   begin
         return Node.Object_Decl_F_Type_Expr;
   end;
   function Object_Decl_F_Default_Expr
     (Node : Bare_Object_Decl) return Bare_Expr
   is
   begin
         return Node.Object_Decl_F_Default_Expr;
   end;
   function Object_Decl_F_Renaming_Clause
     (Node : Bare_Object_Decl) return Bare_Renaming_Clause
   is
   begin
         return Node.Object_Decl_F_Renaming_Clause;
   end;
--# property-start ObjectDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9634
pragma Warnings (Off, "is not referenced");
 function Object_Decl_P_Defining_Names
  (Node : Bare_Object_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Object_Decl := Bare_Object_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Object_Decl :=
     Internal_Entity_Object_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_765 with Inline_Always;
      Fld : Internal_Entity_Defining_Name_List;
Fld_1 : Bare_Defining_Name_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity_Defining_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Defining_Name;
Item_191 : Internal_Entity_Defining_Name;
Bare_Item_191 : Bare_Defining_Name;
Map_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_765 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_765;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5376 '<Map at ast.py:9634>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9634
   --# expr-start 5375 '<FieldAccess for ids at ast.py:9634>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9634
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name_List (Node => Ent.Node.Object_Decl_F_Ids, Info => Ent.Info);
--# expr-done 5375
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Internal_Entity_Defining_Name_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Internal_Entity_Defining_Name_Array (0);
      else
      declare
         Collection : constant Bare_Defining_Name_List := Fld_1;
      begin
         for Untyped_Item_191 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_191 := Untyped_Item_191; 
Is_Null := Bare_Item_191 = null; 
if Is_Null then
   If_Result := No_Entity_Defining_Name;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_191); 
   If_Result := As_Entity;
end if;
                  Item_191 := If_Result; 
   --# scope-start
               --# bind id Item_191
         declare
            Item_To_Append : constant Internal_Entity_Defining_Name := Item_191;
         begin
            Internal_Entity_Defining_Name_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Defining_Name_Array
           (Items_Count => Natural (Internal_Entity_Defining_Name_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Defining_Name_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Defining_Name_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Defining_Name_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 5376
      Property_Result := Map_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_765;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_765;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Object_Decl_P_Defining_Names;
--# end
--# property-start ObjectDecl.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9635
pragma Warnings (Off, "is not referenced");
 function Object_Decl_P_Defining_Env
  (Node : Bare_Object_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Object_Decl := Bare_Object_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Object_Decl :=
     Internal_Entity_Object_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_766 with Inline_Always;
      Fld : Internal_Entity_Type_Expr;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_766 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_766;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5378 '<FieldAccess for defining_env at ast.py:9635>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9635
--# expr-start 5377 '<FieldAccess for type_expr at ast.py:9635>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9635
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Expr (Node => Ent.Node.Object_Decl_F_Type_Expr, Info => Ent.Info);
--# expr-done 5377
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeExpr.defining_env
Fld_1 := Libadalang.Implementation.Type_Expr_P_Defining_Env (Node => Fld.Node, Origin => Origin, Include_Ud_Indexing => Include_Ud_Indexing, E_Info => Fld.Info);
--# end
--# expr-done 5378
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_766;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_766;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Object_Decl_P_Defining_Env;
--# end
--# property-start ObjectDecl.type_expression /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9636
pragma Warnings (Off, "is not referenced");
 function Object_Decl_P_Type_Expression
  (Node : Bare_Object_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Expr
is
   Self : Bare_Object_Decl := Bare_Object_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Object_Decl :=
     Internal_Entity_Object_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Type_Expr;
      Fld : Internal_Entity_Type_Expr;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5379 '<FieldAccess for type_expr at ast.py:9636>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9636
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Expr (Node => Ent.Node.Object_Decl_F_Type_Expr, Info => Ent.Info);
--# expr-done 5379
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Object_Decl_P_Type_Expression;
--# end
--# property-start ObjectDecl.is_constant_object /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9639
pragma Warnings (Off, "is not referenced");
 function Object_Decl_P_Is_Constant_Object
  (Node : Bare_Object_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Object_Decl := Bare_Object_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Object_Decl :=
     Internal_Entity_Object_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Constant_Node;
Var_Expr_1 : Internal_Entity_Mode;
Var_Expr_2 : Internal_Entity_Renaming_Clause;
Fld : Internal_Entity_Constant_Node;
Fld_1 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Fld_2 : Internal_Entity_Mode;
Fld_3 : Bare_Ada_Node;
Is_A : Boolean;
Is_A_1 : Boolean;
If_Result : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
If_Result_1 : Boolean;
Fld_4 : Internal_Entity_Renaming_Clause;
Fld_5 : Internal_Entity_Name;
Fld_6 : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
If_Result_2 : Boolean;
Fld_7 : Internal_Entity_Type_Expr;
Fld_8 : Internal_Entity_Base_Type_Decl;
Is_A_2 : Boolean;
If_Result_3 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5392 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5380 '<FieldAccess for has_constant at ast.py:9641>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9641
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Constant_Node (Node => Ent.Node.Object_Decl_F_Has_Constant, Info => Ent.Info);
--# expr-done 5380
Var_Expr := Fld;
if Var_Expr /= No_Entity_Constant_Node then
   --# scope-start
   --# bind c Var_Expr
--# expr-start 5381 '<FieldAccess for as_bool at ast.py:9641>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9641
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Constant.as_bool'
Fld_1 := Libadalang.Implementation.Dispatcher_Constant_Node_P_As_Bool (Node => Var_Expr.Node);
--# end
--# expr-done 5381
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
if Result_Var then
   If_Result_1 := True;
else
--# expr-start 5382 '<FieldAccess for mode at ast.py:9644>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9644
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Mode (Node => Ent.Node.Object_Decl_F_Mode, Info => Ent.Info);
--# expr-done 5382
Var_Expr_1 := Fld_2;
if Var_Expr_1 /= No_Entity_Mode then
   --# scope-start
   --# bind m Var_Expr_1
--# expr-start 5384 '<IsA GenericFormalObjDecl at ast.py:9646>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9646
--# expr-start 5383 '<parent at ast.py:9646>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9646
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Self.Parent; 
--# expr-done 5383
Is_A := Fld_3 /= null 
and then Fld_3.Kind in Ada_Generic_Formal_Obj_Decl_Range; 
--# expr-done 5384
if Is_A then
   --# expr-start 5385 '<IsA Mode.In, Mode.Default at ast.py:9647>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9647
Is_A_1 := Var_Expr_1.Node /= null 
and then Var_Expr_1.Node.Kind in Ada_Mode_In_Range | Ada_Mode_Default_Range; 
--# expr-done 5385
   If_Result := Is_A_1;
else
   If_Result := False;
end if;
Scope_Result_1 := If_Result;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
   If_Result_1 := Result_Var_1;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
--# expr-start 5386 '<FieldAccess for renaming_clause at ast.py:9651>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9651
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Renaming_Clause (Node => Ent.Node.Object_Decl_F_Renaming_Clause, Info => Ent.Info);
--# expr-done 5386
Var_Expr_2 := Fld_4;
if Var_Expr_2 /= No_Entity_Renaming_Clause then
   --# scope-start
--# expr-start 5388 '<FieldAccess for is_constant at ast.py:9651>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9651
--# expr-start 5387 '<FieldAccess for renamed_object at ast.py:9651>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9651
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Name (Node => Var_Expr_2.Node.Renaming_Clause_F_Renamed_Object, Info => Var_Expr_2.Info);
--# expr-done 5387
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.is_constant'
Fld_6 := Libadalang.Implementation.Dispatcher_Name_P_Is_Constant (Node => Fld_5.Node, E_Info => Fld_5.Info);
--# end
--# expr-done 5388
Scope_Result_2 := Fld_6;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := False;
end if;
   If_Result_2 := Result_Var_2;
end if;
if If_Result_2 then
   If_Result_3 := True;
else
   --# expr-start 5391 '<IsA ProtectedTypeDecl at ast.py:9654>' Is_A_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9654
--# expr-start 5390 '<FieldAccess for designated_type_decl at ast.py:9654>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9654
--# expr-start 5389 '<FieldAccess for type_expr at ast.py:9654>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9654
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Type_Expr (Node => Ent.Node.Object_Decl_F_Type_Expr, Info => Ent.Info);
--# expr-done 5389
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeExpr.designated_type_decl
Fld_8 := Libadalang.Implementation.Type_Expr_P_Designated_Type_Decl (Node => Fld_7.Node, E_Info => Fld_7.Info);
--# end
--# expr-done 5390
Is_A_2 := Fld_8.Node /= null 
and then Fld_8.Node.Kind in Ada_Protected_Type_Decl_Range; 
--# expr-done 5391
   If_Result_3 := Is_A_2;
end if;
Let_Result := If_Result_3; 
--# end
--# expr-done 5392
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Object_Decl_P_Is_Constant_Object;
--# end
--# property-start ObjectDecl.is_static_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9658
pragma Warnings (Off, "is not referenced");
 function Object_Decl_P_Is_Static_Decl
  (Node : Bare_Object_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Object_Decl := Bare_Object_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Object_Decl :=
     Internal_Entity_Object_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Expr;
Var_Expr_1 : Internal_Entity_Renaming_Clause;
Fld : Bare_Constant_Node;
Fld_1 : Boolean;
Fld_2 : Internal_Entity_Expr;
Fld_3 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result : Boolean;
Fld_4 : Internal_Entity_Renaming_Clause;
Fld_5 : Internal_Entity_Name;
Fld_6 : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
If_Result_1 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5400 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5394 '<FieldAccess for as_bool at ast.py:9660>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9660
--# expr-start 5393 '<FieldAccess for has_constant at ast.py:9660>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9660
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Object_Decl_F_Has_Constant;
--# expr-done 5393
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Constant.as_bool'
Fld_1 := Libadalang.Implementation.Dispatcher_Constant_Node_P_As_Bool (Node => Fld);
--# end
--# expr-done 5394
if Fld_1 then
--# expr-start 5395 '<FieldAccess for default_expr at ast.py:9660>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9660
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Expr (Node => Ent.Node.Object_Decl_F_Default_Expr, Info => Ent.Info);
--# expr-done 5395
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind expr Var_Expr
--# expr-start 5396 '<FieldAccess for is_static_expr at ast.py:9661>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9661
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_3 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Var_Expr.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr.Info);
--# end
--# expr-done 5396
Scope_Result := Fld_3;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
   If_Result := Result_Var;
else
   If_Result := False;
end if;
if If_Result then
   If_Result_1 := True;
else
--# expr-start 5397 '<FieldAccess for renaming_clause at ast.py:9663>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9663
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Renaming_Clause (Node => Ent.Node.Object_Decl_F_Renaming_Clause, Info => Ent.Info);
--# expr-done 5397
Var_Expr_1 := Fld_4;
if Var_Expr_1 /= No_Entity_Renaming_Clause then
   --# scope-start
--# expr-start 5399 '<FieldAccess for is_static_expr at ast.py:9663>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9663
--# expr-start 5398 '<FieldAccess for renamed_object at ast.py:9663>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9663
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Name (Node => Var_Expr_1.Node.Renaming_Clause_F_Renamed_Object, Info => Var_Expr_1.Info);
--# expr-done 5398
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_6 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_5.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_5.Info);
--# end
--# expr-done 5399
Scope_Result_1 := Fld_6;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
   If_Result_1 := Result_Var_1;
end if;
Let_Result := If_Result_1; 
--# end
--# expr-done 5400
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Object_Decl_P_Is_Static_Decl;
--# end
--# property-start ObjectDecl.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9667
pragma Warnings (Off, "is not referenced");
 function Object_Decl_P_Xref_Equation
  (Node : Bare_Object_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Object_Decl := Bare_Object_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Object_Decl :=
     Internal_Entity_Object_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_770 with Inline_Always;
            procedure Finalizer_Scope_2496 with Inline_Always;
            procedure Finalizer_Scope_2497 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Var_Expr : Internal_Entity_Expr;
Var_Expr_1 : Internal_Entity_Renaming_Clause;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Type_Expr;
Fld_2 : Logic_Equation;
Fld_3 : Internal_Entity_Expr;
Fld_4 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_5 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_6 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_7 : Internal_Entity_Renaming_Clause;
Fld_8 : Internal_Entity_Name;
Fld_9 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
Fld_10 : Internal_Entity_Name;
Fld_11 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_12 : Internal_Entity_Name;
Fld_13 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Scope_Result_1 : Logic_Equation;
True_Rel_1 : Logic_Equation;
Result_Var_1 : Logic_Equation;
And_Pred_5 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_770 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (True_Rel_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_770;
            procedure Finalizer_Scope_2496 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_5);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_2496;
            procedure Finalizer_Scope_2497 is
            begin
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Fld_11);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Fld_13);
                     Dec_Ref (And_Pred_4);
            end Finalizer_Scope_2497;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5423 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5401 '<FieldAccess for expr_type at ast.py:9668>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9668
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.expr_type'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Expr_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 5401
Block_Var := Fld; 
--# bind typ Block_Var
--# expr-start 5422 '<And at ast.py:9670>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9670
--# expr-start 5411 '<And at ast.py:9670>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9670
--# expr-start 5403 '<FieldAccess for sub_equation at ast.py:9670>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9670
--# expr-start 5402 '<FieldAccess for type_expr at ast.py:9670>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9670
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Type_Expr (Node => Ent.Node.Object_Decl_F_Type_Expr, Info => Ent.Info);
--# expr-done 5402
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_2 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_1.Node, Env => Env, Origin => Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 5403
--# expr-start 5404 '<FieldAccess for default_expr at ast.py:9671>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9671
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Expr (Node => Ent.Node.Object_Decl_F_Default_Expr, Info => Ent.Info);
--# expr-done 5404
Var_Expr := Fld_3;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind de Var_Expr
--# expr-start 5410 '<And at ast.py:9673>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9673
--# expr-start 5408 '<And at ast.py:9673>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9673
--# expr-start 5406 '<Bind at ast.py:9673>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9673
--# expr-start 5405 '<FieldAccess for expected_type_var at ast.py:9673>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9673
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Var_Expr.Node);
--# expr-done 5405
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_4, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9673") else null)); 
--# expr-done 5406
--# expr-start 5407 '<FieldAccess for sub_equation at ast.py:9674>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9674
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_5 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 5407
And_Pred := Create_And (Bind_Result, Fld_5, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9673") else null)); 
--# expr-done 5408
--# expr-start 5409 '<FieldAccess for matches_expected_assign_type at ast.py:9675>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9675
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_assign_type
Fld_6 := Libadalang.Implementation.Expr_P_Matches_Expected_Assign_Type (Node => Var_Expr.Node, Origin => Origin);
--# end
--# expr-done 5409
And_Pred_1 := Create_And (And_Pred, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9673") else null)); 
--# expr-done 5410
Scope_Result := And_Pred_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2496;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9676") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred_2 := Create_And (Fld_2, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9670") else null)); 
--# expr-done 5411
--# expr-start 5412 '<FieldAccess for renaming_clause at ast.py:9678>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9678
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Renaming_Clause (Node => Ent.Node.Object_Decl_F_Renaming_Clause, Info => Ent.Info);
--# expr-done 5412
Var_Expr_1 := Fld_7;
if Var_Expr_1 /= No_Entity_Renaming_Clause then
   --# scope-start
   --# bind rc Var_Expr_1
--# expr-start 5421 '<And at ast.py:9680>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9680
--# expr-start 5418 '<And at ast.py:9680>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9680
--# expr-start 5415 '<Bind at ast.py:9680>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9680
--# expr-start 5414 '<FieldAccess for expected_type_var at ast.py:9680>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9680
--# expr-start 5413 '<FieldAccess for renamed_object at ast.py:9680>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9680
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Name (Node => Var_Expr_1.Node.Renaming_Clause_F_Renamed_Object, Info => Var_Expr_1.Info);
--# expr-done 5413
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_8.Node);
--# expr-done 5414
Fld_9.Value := No_Entity;
Entity_Vars.Reset (Fld_9);
Cast_Expr_1 := Block_Var; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_9, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9680") else null)); 
--# expr-done 5415
--# expr-start 5417 '<FieldAccess for sub_equation at ast.py:9681>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9681
--# expr-start 5416 '<FieldAccess for renamed_object at ast.py:9681>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9681
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Create_Internal_Entity_Name (Node => Var_Expr_1.Node.Renaming_Clause_F_Renamed_Object, Info => Var_Expr_1.Info);
--# expr-done 5416
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_11 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_10.Node, Env => Env, Origin => Origin, E_Info => Fld_10.Info);
--# end
--# expr-done 5417
And_Pred_3 := Create_And (Bind_Result_1, Fld_11, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9680") else null)); 
--# expr-done 5418
--# expr-start 5420 '<FieldAccess for matches_expected_assign_type at ast.py:9682>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9682
--# expr-start 5419 '<FieldAccess for renamed_object at ast.py:9682>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9682
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Create_Internal_Entity_Name (Node => Var_Expr_1.Node.Renaming_Clause_F_Renamed_Object, Info => Var_Expr_1.Info);
--# expr-done 5419
      if Fld_12.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_assign_type
Fld_13 := Libadalang.Implementation.Expr_P_Matches_Expected_Assign_Type (Node => Fld_12.Node, Origin => Origin);
--# end
--# expr-done 5420
And_Pred_4 := Create_And (And_Pred_3, Fld_13, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9680") else null)); 
--# expr-done 5421
Scope_Result_1 := And_Pred_4;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2497;
   Result_Var_1 := Scope_Result_1;
else
   True_Rel_1 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9683") else null)); 
   Result_Var_1 := True_Rel_1;
end if;
      Inc_Ref (Result_Var_1);
And_Pred_5 := Create_And (And_Pred_2, Result_Var_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9670") else null)); 
--# expr-done 5422
Let_Result := And_Pred_5; Inc_Ref (Let_Result);
--# end
--# expr-done 5423
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_770;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_770;
               Finalizer_Scope_2496;
               Finalizer_Scope_2497;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Object_Decl_P_Xref_Equation;
--# end
--# property-start ObjectDecl.next_part_for_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9688
pragma Warnings (Off, "is not referenced");
 function Object_Decl_P_Next_Part_For_Name
  (Node : Bare_Object_Decl
      ; Sym : Symbol_Type
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Object_Decl := Bare_Object_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Object_Decl :=
     Internal_Entity_Object_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind sym Sym
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_771 with Inline_Always;
            procedure Finalizer_Scope_2498 with Inline_Always;
      Var_Expr : Bare_Private_Part;
Fld : Bare_Declarative_Part;
Fld_1 : Bare_Ada_Node;
Cast_Result : Bare_Base_Package_Decl;
Fld_2 : Bare_Private_Part;
Fld_3 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Env_Get_Result : Internal_Entity;
Cast_Result_1 : Internal_Entity_Basic_Decl;
Fld_4 : Boolean;
Fld_5 : Bare_Constant_Node;
Fld_6 : Boolean;
If_Result : Boolean;
If_Result_1 : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_771 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_771;
            procedure Finalizer_Scope_2498 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_2498;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5435 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5434 '<If at ast.py:9689>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9689
--# expr-start 5424 '<FieldAccess for is_in_public_part at ast.py:9690>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9690
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_in_public_part
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Is_In_Public_Part (Node => Ent.Node);
--# end
--# expr-done 5424
if Fld_4 then
   --# expr-start 5426 '<FieldAccess for as_bool at ast.py:9690>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9690
--# expr-start 5425 '<FieldAccess for has_constant at ast.py:9690>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9690
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Self.Object_Decl_F_Has_Constant;
--# expr-done 5425
      if Fld_5 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Constant.as_bool'
Fld_6 := Libadalang.Implementation.Dispatcher_Constant_Node_P_As_Bool (Node => Fld_5);
--# end
--# expr-done 5426
   If_Result := Fld_6;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 5433 '<Cast to BasicDecl at ast.py:9691>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9691
--# expr-start 5432 '<EnvGet at ast.py:9691>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9691
--# expr-start 5430 '<FieldAccess for private_part at ast.py:9691>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9691
--# expr-start 5429 '<Cast to BasePackageDecl at ast.py:9691>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9691
--# expr-start 5428 '<parent at ast.py:9691>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9691
--# expr-start 5427 '<FieldAccess for declarative_scope at ast.py:9691>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9691
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.declarative_scope
Fld := Libadalang.Implementation.Ada_Node_P_Declarative_Scope (Node => Self);
--# end
--# expr-done 5427
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Parent; 
--# expr-done 5428
   if Fld_1 = null
      or else Fld_1.Kind in Ada_Base_Package_Decl
   then
      Cast_Result := Fld_1;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 5429
      if Cast_Result = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Cast_Result.Base_Package_Decl_F_Private_Part;
--# expr-done 5430
Var_Expr := Fld_2;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 5431 '<FieldAccess for children_env at ast.py:9691>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9691
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Children_Env (Node => Var_Expr);
--# expr-done 5431
Scope_Result := Fld_3;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2498;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
Env_Get_Result := AST_Envs.Get_First (Self => Result_Var, Key => Sym, Lookup_Kind => To_Lookup_Kind_Type (Minimal), Categories => (Inherited_Primitives => False, Nocat => True)); 
--# expr-done 5432
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Basic_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         Cast_Result_1 := No_Entity_Basic_Decl;
   end if;
--# expr-done 5433
   If_Result_1 := Cast_Result_1;
else
   If_Result_1 := No_Entity_Basic_Decl;
end if;
--# expr-done 5434
Let_Result := If_Result_1; 
--# end
--# expr-done 5435
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_771;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_771;
               Finalizer_Scope_2498;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Object_Decl_P_Next_Part_For_Name;
--# end
--# property-start ObjectDecl.previous_part_for_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9699
pragma Warnings (Off, "is not referenced");
 function Object_Decl_P_Previous_Part_For_Name
  (Node : Bare_Object_Decl
      ; Sym : Symbol_Type
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Object_Decl := Bare_Object_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Object_Decl :=
     Internal_Entity_Object_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind sym Sym
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_772 with Inline_Always;
      Fld : Bare_Declarative_Part;
Fld_1 : Bare_Ada_Node;
Cast_Result : Bare_Base_Package_Decl;
Fld_2 : Bare_Public_Part;
Fld_3 : Lexical_Env := Empty_Env;
Env_Get_Result : Internal_Entity;
Cast_Result_1 : Internal_Entity_Basic_Decl;
Fld_4 : Boolean;
Fld_5 : Bare_Constant_Node;
Fld_6 : Boolean;
If_Result : Boolean;
If_Result_1 : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_772 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_772;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5447 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5446 '<If at ast.py:9700>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9700
--# expr-start 5436 '<FieldAccess for is_in_private_part at ast.py:9701>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9701
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_in_private_part'
Fld_4 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_In_Private_Part (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5436
if Fld_4 then
   --# expr-start 5438 '<FieldAccess for as_bool at ast.py:9701>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9701
--# expr-start 5437 '<FieldAccess for has_constant at ast.py:9701>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9701
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Self.Object_Decl_F_Has_Constant;
--# expr-done 5437
      if Fld_5 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Constant.as_bool'
Fld_6 := Libadalang.Implementation.Dispatcher_Constant_Node_P_As_Bool (Node => Fld_5);
--# end
--# expr-done 5438
   If_Result := Fld_6;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 5445 '<Cast to BasicDecl at ast.py:9702>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9702
--# expr-start 5444 '<EnvGet at ast.py:9702>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9702
--# expr-start 5443 '<FieldAccess for children_env at ast.py:9702>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9702
--# expr-start 5442 '<FieldAccess for public_part at ast.py:9702>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9702
--# expr-start 5441 '<Cast to BasePackageDecl at ast.py:9702>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9702
--# expr-start 5440 '<parent at ast.py:9702>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9702
--# expr-start 5439 '<FieldAccess for declarative_scope at ast.py:9702>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9702
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.declarative_scope
Fld := Libadalang.Implementation.Ada_Node_P_Declarative_Scope (Node => Self);
--# end
--# expr-done 5439
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Parent; 
--# expr-done 5440
   if Fld_1 = null
      or else Fld_1.Kind in Ada_Base_Package_Decl
   then
      Cast_Result := Fld_1;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 5441
      if Cast_Result = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Cast_Result.Base_Package_Decl_F_Public_Part;
--# expr-done 5442
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Children_Env (Node => Fld_2);
--# expr-done 5443
Env_Get_Result := AST_Envs.Get_First (Self => Fld_3, Key => Sym, Lookup_Kind => To_Lookup_Kind_Type (Minimal), Categories => (Inherited_Primitives => False, Nocat => True)); 
--# expr-done 5444
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Basic_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         Cast_Result_1 := No_Entity_Basic_Decl;
   end if;
--# expr-done 5445
   If_Result_1 := Cast_Result_1;
else
   If_Result_1 := No_Entity_Basic_Decl;
end if;
--# expr-done 5446
Let_Result := If_Result_1; 
--# end
--# expr-done 5447
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_772;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_772;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Object_Decl_P_Previous_Part_For_Name;
--# end
--# property-start ObjectDecl.private_part_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9710
pragma Warnings (Off, "is not referenced");
 function Object_Decl_P_Private_Part_Decl
  (Node : Bare_Object_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Object_Decl := Bare_Object_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Object_Decl :=
     Internal_Entity_Object_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Internal_Entity_Defining_Name;
Fld_1 : Symbol_Type;
Fld_2 : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5451 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5450 '<FieldAccess for next_part_for_name at ast.py:9715>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9715
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 5449 '<FieldAccess for name_symbol at ast.py:9716>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9716
--# expr-start 5448 '<FieldAccess for defining_name_or_raise at ast.py:9716>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9716
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name_or_raise
Fld := Libadalang.Implementation.Basic_Decl_P_Defining_Name_Or_Raise (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5448
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld.Node);
--# end
--# expr-done 5449
--# property-call-start '[dispatcher]BasicDecl.next_part_for_name'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Next_Part_For_Name (Node => Ent.Node, Sym => Fld_1, Imprecise_Fallback => False, E_Info => Ent.Info);
--# end
--# expr-done 5450
Let_Result := Fld_2; 
--# end
--# expr-done 5451
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Object_Decl_P_Private_Part_Decl;
--# end
--# property-start ObjectDecl.public_part_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9720
pragma Warnings (Off, "is not referenced");
 function Object_Decl_P_Public_Part_Decl
  (Node : Bare_Object_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Object_Decl := Bare_Object_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Object_Decl :=
     Internal_Entity_Object_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Internal_Entity_Defining_Name;
Fld_1 : Symbol_Type;
Fld_2 : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5455 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5454 '<FieldAccess for previous_part_for_name at ast.py:9725>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9725
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 5453 '<FieldAccess for name_symbol at ast.py:9726>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9726
--# expr-start 5452 '<FieldAccess for defining_name_or_raise at ast.py:9726>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9726
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name_or_raise
Fld := Libadalang.Implementation.Basic_Decl_P_Defining_Name_Or_Raise (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5452
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld.Node);
--# end
--# expr-done 5453
--# property-call-start '[dispatcher]BasicDecl.previous_part_for_name'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Previous_Part_For_Name (Node => Ent.Node, Sym => Fld_1, Imprecise_Fallback => False, E_Info => Ent.Info);
--# end
--# expr-done 5454
Let_Result := Fld_2; 
--# end
--# expr-done 5455
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Object_Decl_P_Public_Part_Decl;
--# end
--# property-start ObjectDecl.next_part_for_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9730
pragma Warnings (Off, "is not referenced");
 function Object_Decl_P_Next_Part_For_Decl
  (Node : Bare_Object_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Object_Decl := Bare_Object_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Object_Decl :=
     Internal_Entity_Object_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5457 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5456 '<FieldAccess for private_part_decl at ast.py:9731>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9731
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ObjectDecl.private_part_decl
Fld := Libadalang.Implementation.Object_Decl_P_Private_Part_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5456
Let_Result := Fld; 
--# end
--# expr-done 5457
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Object_Decl_P_Next_Part_For_Decl;
--# end
--# property-start ObjectDecl.previous_part_for_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9734
pragma Warnings (Off, "is not referenced");
 function Object_Decl_P_Previous_Part_For_Decl
  (Node : Bare_Object_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Object_Decl := Bare_Object_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Object_Decl :=
     Internal_Entity_Object_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5459 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5458 '<FieldAccess for public_part_decl at ast.py:9735>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9735
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ObjectDecl.public_part_decl
Fld := Libadalang.Implementation.Object_Decl_P_Public_Part_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5458
Let_Result := Fld; 
--# end
--# expr-done 5459
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Object_Decl_P_Previous_Part_For_Decl;
--# end
--# property-start ObjectDecl.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9737
pragma Warnings (Off, "is not referenced");
 function Object_Decl_P_Xref_Entry_Point
  (Node : Bare_Object_Decl
  )
   return Boolean
is
   Self : Bare_Object_Decl := Bare_Object_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Object_Decl_P_Xref_Entry_Point;
--# end
--# property-start ObjectDecl._env_mappings_44 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9632
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_44
  (Node : Bare_Object_Decl
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Object_Decl := Bare_Object_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1461 with Inline_Always;
      Fld : Bare_Defining_Name_List;
Cast_Expr : Bare_Object_Decl;
Cast_Result : Bare_Ada_Node;
Fld_1 : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_1461 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1461;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5461 '<FieldAccess for env_mappings at ast.py:9632>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9632
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 5460 '<FieldAccess for ids at ast.py:9632>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9632
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Object_Decl_F_Ids;
--# expr-done 5460
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# property-call-start AdaNode.env_mappings
Fld_1 := Libadalang.Implementation.Ada_Node_P_Env_Mappings (Node => Self, Defining_Names => Fld, Value => Cast_Result);
--# end
--# expr-done 5461
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1461;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1461;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_44;
--# end
   --
   --  Primitives for Bare_Extended_Return_Stmt_Object_Decl
   --
      procedure Initialize_Fields_For_Extended_Return_Stmt_Object_Decl
        (Self : Bare_Extended_Return_Stmt_Object_Decl
         ; Object_Decl_F_Ids : Bare_Defining_Name_List
         ; Object_Decl_F_Has_Aliased : Bare_Aliased_Node
         ; Object_Decl_F_Has_Constant : Bare_Constant_Node
         ; Object_Decl_F_Mode : Bare_Mode
         ; Object_Decl_F_Type_Expr : Bare_Type_Expr
         ; Object_Decl_F_Default_Expr : Bare_Expr
         ; Object_Decl_F_Renaming_Clause : Bare_Renaming_Clause
         ; Object_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Initialize_Fields_For_Object_Decl
              (Self, Object_Decl_F_Ids, Object_Decl_F_Has_Aliased, Object_Decl_F_Has_Constant, Object_Decl_F_Mode, Object_Decl_F_Type_Expr, Object_Decl_F_Default_Expr, Object_Decl_F_Renaming_Clause, Object_Decl_F_Aspects);
      end Initialize_Fields_For_Extended_Return_Stmt_Object_Decl;
   --
   --  Primitives for Bare_No_Type_Object_Renaming_Decl
   --
      procedure Initialize_Fields_For_No_Type_Object_Renaming_Decl
        (Self : Bare_No_Type_Object_Renaming_Decl
         ; Object_Decl_F_Ids : Bare_Defining_Name_List
         ; Object_Decl_F_Has_Aliased : Bare_Aliased_Node
         ; Object_Decl_F_Has_Constant : Bare_Constant_Node
         ; Object_Decl_F_Mode : Bare_Mode
         ; Object_Decl_F_Type_Expr : Bare_Type_Expr
         ; Object_Decl_F_Default_Expr : Bare_Expr
         ; Object_Decl_F_Renaming_Clause : Bare_Renaming_Clause
         ; Object_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Initialize_Fields_For_Object_Decl
              (Self, Object_Decl_F_Ids, Object_Decl_F_Has_Aliased, Object_Decl_F_Has_Constant, Object_Decl_F_Mode, Object_Decl_F_Type_Expr, Object_Decl_F_Default_Expr, Object_Decl_F_Renaming_Clause, Object_Decl_F_Aspects);
      end Initialize_Fields_For_No_Type_Object_Renaming_Decl;
--# property-start NoTypeObjectRenamingDecl.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9804
pragma Warnings (Off, "is not referenced");
 function No_Type_Object_Renaming_Decl_P_Xref_Equation
  (Node : Bare_No_Type_Object_Renaming_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_No_Type_Object_Renaming_Decl := Bare_No_Type_Object_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_No_Type_Object_Renaming_Decl :=
     Internal_Entity_No_Type_Object_Renaming_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_783 with Inline_Always;
      Fld : Internal_Entity_Renaming_Clause;
Fld_1 : Internal_Entity_Name;
Fld_2 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_783 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_783;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5465 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5464 '<FieldAccess for sub_equation at ast.py:9805>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9805
--# expr-start 5463 '<FieldAccess for renamed_object at ast.py:9805>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9805
--# expr-start 5462 '<FieldAccess for renaming_clause at ast.py:9805>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9805
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Renaming_Clause (Node => Ent.Node.Object_Decl_F_Renaming_Clause, Info => Ent.Info);
--# expr-done 5462
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Fld.Node.Renaming_Clause_F_Renamed_Object, Info => Fld.Info);
--# expr-done 5463
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_2 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_1.Node, Env => Env, Origin => Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 5464
Let_Result := Fld_2; Inc_Ref (Let_Result);
--# end
--# expr-done 5465
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_783;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_783;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end No_Type_Object_Renaming_Decl_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Package_Renaming_Decl
   --
      procedure Package_Renaming_Decl_Pre_Env_Actions
        (Self            : Bare_Package_Renaming_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Dummy : Boolean;
      begin
         Dummy := Libadalang.Implementation.Env_Do_78 (Node => 
Self);
      end;
      declare
         Env : constant Internal_Designated_Env :=
           Libadalang.Implementation.Initial_Env_79 (Node => 
Self);
      begin
         Set_Initial_Env
           (Self,
            State,
            Env,
            DSL_Location => "ast.py:10462");
      end;
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_80 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:10464");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_81 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      declare
         Dummy : Bare_Compilation_Unit_Array_Access;
      begin
         Dummy := Libadalang.Implementation.Env_Do_82 (Node => 
Self);
         Dec_Ref (Dummy);
      end;
      if Libadalang.Implementation.Ref_Cond_84 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_83 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Use_Package_Name_Designated_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      if Libadalang.Implementation.Ref_Cond_86 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_85 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Name_Designated_Type_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      end;
      procedure Initialize_Fields_For_Package_Renaming_Decl
        (Self : Bare_Package_Renaming_Decl
         ; Package_Renaming_Decl_F_Name : Bare_Defining_Name
         ; Package_Renaming_Decl_F_Renames : Bare_Renaming_Clause
         ; Package_Renaming_Decl_F_Aspects : Bare_Aspect_Spec
        ) is
      begin
            Self.Package_Renaming_Decl_F_Name := Package_Renaming_Decl_F_Name;
            Self.Package_Renaming_Decl_F_Renames := Package_Renaming_Decl_F_Renames;
            Self.Package_Renaming_Decl_F_Aspects := Package_Renaming_Decl_F_Aspects;
      end Initialize_Fields_For_Package_Renaming_Decl;
   function Package_Renaming_Decl_F_Name
     (Node : Bare_Package_Renaming_Decl) return Bare_Defining_Name
   is
   begin
         return Node.Package_Renaming_Decl_F_Name;
   end;
   function Package_Renaming_Decl_F_Renames
     (Node : Bare_Package_Renaming_Decl) return Bare_Renaming_Clause
   is
   begin
         return Node.Package_Renaming_Decl_F_Renames;
   end;
--# property-start PackageRenamingDecl.renamed_package /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10438
pragma Warnings (Off, "is not referenced");
 function Package_Renaming_Decl_P_Renamed_Package
  (Node : Bare_Package_Renaming_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Package_Renaming_Decl := Bare_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Package_Renaming_Decl :=
     Internal_Entity_Package_Renaming_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_822 with Inline_Always;
      Var_Expr : Internal_Entity;
Fld : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_1 : Internal_Entity_Renaming_Clause;
Fld_2 : Internal_Entity_Name;
Fld_3 : Internal_Entity_Array_Access;
Get_Result : Internal_Entity;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_822 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_822;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5473 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5472 '<bind at ast.py:10442>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10442
--# expr-start 5466 '<FieldAccess for node_env at ast.py:10443>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10443
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Node_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 5466
Bound_Env := Fld; Inc_Ref (Bound_Env);
--# expr-start 5470 '<at at ast.py:10444>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10444
--# expr-start 5469 '<FieldAccess for env_elements at ast.py:10444>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10444
--# expr-start 5468 '<FieldAccess for renamed_object at ast.py:10444>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10444
--# expr-start 5467 '<FieldAccess for renames at ast.py:10444>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10444
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Renaming_Clause (Node => Ent.Node.Package_Renaming_Decl_F_Renames, Info => Ent.Info);
--# expr-done 5467
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Fld_1.Node.Renaming_Clause_F_Renamed_Object, Info => Fld_1.Info);
--# expr-done 5468
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.env_elements
Fld_3 := Libadalang.Implementation.Expr_P_Env_Elements (Node => Fld_2.Node, Env => Bound_Env, No_Visibility => False, E_Info => Fld_2.Info);
--# end
--# expr-done 5469
Get_Result := Get (Fld_3, 0, True); 
--# expr-done 5470
Var_Expr := Get_Result;
if Var_Expr /= No_Entity then
   --# scope-start
--# expr-start 5471 '<Cast to BasicDecl at ast.py:10444>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10444
Cast_Expr := Var_Expr; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 5471
Scope_Result := Cast_Result;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Basic_Decl;
end if;
Dyn_Var_Bind_Result := Result_Var; 
--# expr-done 5472
Let_Result := Dyn_Var_Bind_Result; 
--# end
--# expr-done 5473
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_822;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_822;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Package_Renaming_Decl_P_Renamed_Package;
--# end
--# property-start PackageRenamingDecl.final_renamed_package /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10448
pragma Warnings (Off, "is not referenced");
 function Package_Renaming_Decl_P_Final_Renamed_Package
  (Node : Bare_Package_Renaming_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Package_Renaming_Decl := Bare_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Package_Renaming_Decl :=
     Internal_Entity_Package_Renaming_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Basic_Decl;
      Block_Var : Internal_Entity_Basic_Decl;
Var_Expr : Internal_Entity_Package_Renaming_Decl;
Fld : Internal_Entity_Basic_Decl;
Cast_Expr : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity_Package_Renaming_Decl;
Fld_1 : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5477 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5474 '<FieldAccess for renamed_package at ast.py:10453>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10453
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start PackageRenamingDecl.renamed_package
Fld := Libadalang.Implementation.Package_Renaming_Decl_P_Renamed_Package (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5474
Block_Var := Fld; 
--# bind pkg Block_Var
--# expr-start 5475 '<Cast to PackageRenamingDecl at ast.py:10454>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10454
Cast_Expr := Block_Var; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Package_Renaming_Decl_Range
   then
      Cast_Result := Create_Internal_Entity_Package_Renaming_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Package_Renaming_Decl;
   end if;
--# expr-done 5475
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Package_Renaming_Decl then
   --# scope-start
   --# bind r Var_Expr
--# expr-start 5476 '<FieldAccess for final_renamed_package at ast.py:10455>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10455
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start PackageRenamingDecl.final_renamed_package
Fld_1 := Libadalang.Implementation.Package_Renaming_Decl_P_Final_Renamed_Package (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 5476
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := Block_Var;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 5477
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Package_Renaming_Decl_P_Final_Renamed_Package;
--# end
--# property-start PackageRenamingDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10485
pragma Warnings (Off, "is not referenced");
 function Package_Renaming_Decl_P_Defining_Names
  (Node : Bare_Package_Renaming_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Package_Renaming_Decl := Bare_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Package_Renaming_Decl :=
     Internal_Entity_Package_Renaming_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_824 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_824 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_824;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5478 '<FieldAccess for name at ast.py:10485>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10485
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Package_Renaming_Decl_F_Name, Info => Ent.Info);
--# expr-done 5478
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_824;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_824;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Package_Renaming_Decl_P_Defining_Names;
--# end
--# property-start PackageRenamingDecl.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10486
pragma Warnings (Off, "is not referenced");
 function Package_Renaming_Decl_P_Defining_Env
  (Node : Bare_Package_Renaming_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Package_Renaming_Decl := Bare_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Package_Renaming_Decl :=
     Internal_Entity_Package_Renaming_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_825 with Inline_Always;
      Fld : Internal_Entity_Basic_Decl;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_825 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_825;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5480 '<FieldAccess for defining_env at ast.py:10486>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10486
--# expr-start 5479 '<FieldAccess for renamed_package at ast.py:10486>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10486
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start PackageRenamingDecl.renamed_package
Fld := Libadalang.Implementation.Package_Renaming_Decl_P_Renamed_Package (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5479
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Fld.Node, Origin => Origin, Include_Ud_Indexing => Include_Ud_Indexing, E_Info => Fld.Info);
--# end
--# expr-done 5480
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_825;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_825;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Package_Renaming_Decl_P_Defining_Env;
--# end
--# property-start PackageRenamingDecl.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10488
pragma Warnings (Off, "is not referenced");
 function Package_Renaming_Decl_P_Xref_Entry_Point
  (Node : Bare_Package_Renaming_Decl
  )
   return Boolean
is
   Self : Bare_Package_Renaming_Decl := Bare_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Package_Renaming_Decl_P_Xref_Entry_Point;
--# end
--# property-start PackageRenamingDecl.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10489
pragma Warnings (Off, "is not referenced");
 function Package_Renaming_Decl_P_Xref_Equation
  (Node : Bare_Package_Renaming_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Package_Renaming_Decl := Bare_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Package_Renaming_Decl :=
     Internal_Entity_Package_Renaming_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_827 with Inline_Always;
      Fld : Internal_Entity_Renaming_Clause;
Fld_1 : Internal_Entity_Name;
Fld_2 : Logic_Equation;
            procedure Finalizer_Scope_827 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_827;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5483 '<FieldAccess for xref_no_overloading at ast.py:10490>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10490
--# expr-start 5482 '<FieldAccess for renamed_object at ast.py:10490>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10490
--# expr-start 5481 '<FieldAccess for renames at ast.py:10490>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10490
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Renaming_Clause (Node => Ent.Node.Package_Renaming_Decl_F_Renames, Info => Ent.Info);
--# expr-done 5481
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Fld.Node.Renaming_Clause_F_Renamed_Object, Info => Fld.Info);
--# expr-done 5482
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_2 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Fld_1.Node, Env => Env, Origin => Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 5483
      Property_Result := Fld_2;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_827;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_827;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Package_Renaming_Decl_P_Xref_Equation;
--# end
--# property-start PackageRenamingDecl._env_do_78 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10460
pragma Warnings (Off, "is not referenced");
 function Env_Do_78
  (Node : Bare_Package_Renaming_Decl
  )
   return Boolean
is
   Self : Bare_Package_Renaming_Decl := Bare_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5484 '<FieldAccess for env_hook at ast.py:10460>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10460
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.env_hook
Fld := Libadalang.Implementation.Ada_Node_P_Env_Hook (Node => Self);
--# end
--# expr-done 5484
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_78;
--# end
--# property-start PackageRenamingDecl._initial_env_79 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10462
pragma Warnings (Off, "is not referenced");
 function Initial_Env_79
  (Node : Bare_Package_Renaming_Decl
  )
   return Internal_Designated_Env
is
   Self : Bare_Package_Renaming_Decl := Bare_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_1496 with Inline_Always;
      Fld : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_1496 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1496;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5485 '<FieldAccess for child_decl_initial_env at ast.py:10462>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10462
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.child_decl_initial_env
Fld := Libadalang.Implementation.Basic_Decl_P_Child_Decl_Initial_Env (Node => Self);
--# end
--# expr-done 5485
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1496;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1496;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Initial_Env_79;
--# end
--# property-start PackageRenamingDecl._env_mappings_80 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10464
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_80
  (Node : Bare_Package_Renaming_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Package_Renaming_Decl := Bare_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Package_Renaming_Decl :=
     Internal_Entity_Package_Renaming_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1497 with Inline_Always;
      Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Package_Renaming_Decl;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1497 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1497;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5488 '<New EnvAssoc at ast.py:10464>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10464
--# expr-start 5486 '<New DesignatedEnv at ast.py:10464>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10464
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 5486
--# expr-start 5487 '<FieldAccess for name_symbol at ast.py:10465>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10465
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 5487
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 5488
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1497;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1497;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_80;
--# end
--# property-start PackageRenamingDecl._env_trans_parent_81 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10459
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_81
  (Node : Bare_Package_Renaming_Decl
  )
   return Boolean
is
   Self : Bare_Package_Renaming_Decl := Bare_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_81;
--# end
--# property-start PackageRenamingDecl._env_do_82 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10471
pragma Warnings (Off, "is not referenced");
 function Env_Do_82
  (Node : Bare_Package_Renaming_Decl
  )
   return Bare_Compilation_Unit_Array_Access
is
   Self : Bare_Package_Renaming_Decl := Bare_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1499 with Inline_Always;
      Fld : Bare_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1499 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1499;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5489 '<FieldAccess for populate_dependent_units at ast.py:10471>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10471
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.populate_dependent_units
Fld := Libadalang.Implementation.Basic_Decl_P_Populate_Dependent_Units (Node => Self);
--# end
--# expr-done 5489
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1499;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1499;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Do_82;
--# end
--# property-start PackageRenamingDecl._ref_env_nodes_83 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10474
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_83
  (Node : Bare_Package_Renaming_Decl
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Package_Renaming_Decl := Bare_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1500 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1500 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1500;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5490 '<FieldAccess for top_level_use_package_clauses at ast.py:10474>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10474
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_package_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Package_Clauses (Node => Self);
--# end
--# expr-done 5490
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1500;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1500;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_83;
--# end
--# property-start PackageRenamingDecl._ref_cond_84 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10476
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_84
  (Node : Bare_Package_Renaming_Decl
  )
   return Boolean
is
   Self : Bare_Package_Renaming_Decl := Bare_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5492 '<IsA LibraryItem, Subunit at ast.py:10476>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10476
--# expr-start 5491 '<parent at ast.py:10476>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10476
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 5491
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 5492
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_84;
--# end
--# property-start PackageRenamingDecl._ref_env_nodes_85 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10479
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_85
  (Node : Bare_Package_Renaming_Decl
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Package_Renaming_Decl := Bare_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1502 with Inline_Always;
      Fld : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1502 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1502;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5493 '<FieldAccess for top_level_use_type_clauses at ast.py:10479>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10479
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_use_type_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_Use_Type_Clauses (Node => Self);
--# end
--# expr-done 5493
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1502;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1502;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_85;
--# end
--# property-start PackageRenamingDecl._ref_cond_86 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10481
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_86
  (Node : Bare_Package_Renaming_Decl
  )
   return Boolean
is
   Self : Bare_Package_Renaming_Decl := Bare_Package_Renaming_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5495 '<IsA LibraryItem, Subunit at ast.py:10481>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10481
--# expr-start 5494 '<parent at ast.py:10481>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10481
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 5494
Is_A := Fld /= null 
and then Fld.Kind in Ada_Library_Item_Range | Ada_Subunit_Range; 
--# expr-done 5495
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_86;
--# end
   --
   --  Primitives for Bare_Single_Protected_Decl
   --
      procedure Single_Protected_Decl_Pre_Env_Actions
        (Self            : Bare_Single_Protected_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_40 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:9421");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_41 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            Libadalang.Implementation.Env_Names_42 (Node => 
Self);
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Single_Protected_Decl
        (Self : Bare_Single_Protected_Decl
         ; Single_Protected_Decl_F_Name : Bare_Defining_Name
         ; Single_Protected_Decl_F_Aspects : Bare_Aspect_Spec
         ; Single_Protected_Decl_F_Interfaces : Bare_Parent_List
         ; Single_Protected_Decl_F_Definition : Bare_Protected_Def
        ) is
      begin
            Self.Single_Protected_Decl_F_Name := Single_Protected_Decl_F_Name;
            Self.Single_Protected_Decl_F_Aspects := Single_Protected_Decl_F_Aspects;
            Self.Single_Protected_Decl_F_Interfaces := Single_Protected_Decl_F_Interfaces;
            Self.Single_Protected_Decl_F_Definition := Single_Protected_Decl_F_Definition;
      end Initialize_Fields_For_Single_Protected_Decl;
   function Single_Protected_Decl_F_Name
     (Node : Bare_Single_Protected_Decl) return Bare_Defining_Name
   is
   begin
         return Node.Single_Protected_Decl_F_Name;
   end;
   function Single_Protected_Decl_F_Interfaces
     (Node : Bare_Single_Protected_Decl) return Bare_Parent_List
   is
   begin
         return Node.Single_Protected_Decl_F_Interfaces;
   end;
   function Single_Protected_Decl_F_Definition
     (Node : Bare_Single_Protected_Decl) return Bare_Protected_Def
   is
   begin
         return Node.Single_Protected_Decl_F_Definition;
   end;
--# property-start SingleProtectedDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9397
pragma Warnings (Off, "is not referenced");
 function Single_Protected_Decl_P_Defining_Names
  (Node : Bare_Single_Protected_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Single_Protected_Decl := Bare_Single_Protected_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Single_Protected_Decl :=
     Internal_Entity_Single_Protected_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_748 with Inline_Always;
      Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_748 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_748;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5496 '<FieldAccess for name at ast.py:9397>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9397
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Defining_Name (Node => Ent.Node.Single_Protected_Decl_F_Name, Info => Ent.Info);
--# expr-done 5496
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_748;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_748;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Single_Protected_Decl_P_Defining_Names;
--# end
--# property-start SingleProtectedDecl.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9399
pragma Warnings (Off, "is not referenced");
 function Single_Protected_Decl_P_Defining_Env
  (Node : Bare_Single_Protected_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Single_Protected_Decl := Bare_Single_Protected_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Single_Protected_Decl :=
     Internal_Entity_Single_Protected_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_749 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_749 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_749;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5497 '<FieldAccess for children_env at ast.py:9399>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9399
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 5497
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_749;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_749;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Single_Protected_Decl_P_Defining_Env;
--# end
--# property-start SingleProtectedDecl.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9401
pragma Warnings (Off, "is not referenced");
 function Single_Protected_Decl_P_Xref_Entry_Point
  (Node : Bare_Single_Protected_Decl
  )
   return Boolean
is
   Self : Bare_Single_Protected_Decl := Bare_Single_Protected_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Single_Protected_Decl_P_Xref_Entry_Point;
--# end
--# property-start SingleProtectedDecl.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9404
pragma Warnings (Off, "is not referenced");
 function Single_Protected_Decl_P_Xref_Equation
  (Node : Bare_Single_Protected_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Single_Protected_Decl := Bare_Single_Protected_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Single_Protected_Decl :=
     Internal_Entity_Single_Protected_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_751 with Inline_Always;
            procedure Finalizer_Scope_2501 with Inline_Always;
      Fld : Internal_Entity_Parent_List;
Fld_1 : Bare_Parent_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld_3 : Logic_Equation;
Item_192 : Internal_Entity_Name;
Bare_Item_192 : Bare_Name;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_751 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_751;
            procedure Finalizer_Scope_2501 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_2501;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5502 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5501 '<LogicAll at ast.py:9405>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9405
--# expr-start 5500 '<Map at ast.py:9405>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9405
   --# expr-start 5498 '<FieldAccess for interfaces at ast.py:9405>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9405
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Parent_List (Node => Ent.Node.Single_Protected_Decl_F_Interfaces, Info => Ent.Info);
--# expr-done 5498
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Parent_List := Fld_1;
      begin
         for Untyped_Item_192 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_192 := Untyped_Item_192; 
Is_Null := Bare_Item_192 = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_192); 
   If_Result := As_Entity;
end if;
                  Item_192 := If_Result; 
   --# scope-start
               --# bind ifc Item_192
      --# expr-start 5499 '<FieldAccess for xref_equation at ast.py:9405>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9405
      if Item_192.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_3 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Item_192.Node, Env => Env, Origin => Origin, E_Info => Item_192.Info);
--# end
--# expr-done 5499
         declare
            Item_To_Append : constant Logic_Equation := Fld_3;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2501;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 5500
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9405") else null)); 
--# expr-done 5501
Let_Result := Logic_Boolean_Op; Inc_Ref (Let_Result);
--# end
--# expr-done 5502
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_751;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_751;
               Finalizer_Scope_2501;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Single_Protected_Decl_P_Xref_Equation;
--# end
--# property-start SingleProtectedDecl.env_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9408
pragma Warnings (Off, "is not referenced");
 function Single_Protected_Decl_P_Env_Names
  (Node : Bare_Single_Protected_Decl
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Single_Protected_Decl := Bare_Single_Protected_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_752 with Inline_Always;
            procedure Finalizer_Scope_2502 with Inline_Always;
      Var_Expr : String_Type;
Fld : String_Type;
Sym : Symbol_Type;
Singleton : Symbol_Type_Array_Access;
Scope_Result : Symbol_Type_Array_Access;
Result_Var : Symbol_Type_Array_Access;
Let_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_752 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_752;
            procedure Finalizer_Scope_2502 is
            begin
                     Dec_Ref (Var_Expr);
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2502;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5505 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5503 '<FieldAccess for top_level_env_name at ast.py:9409>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9409
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.top_level_env_name
Fld := Libadalang.Implementation.Basic_Decl_P_Top_Level_Env_Name (Node => Self);
--# end
--# expr-done 5503
Var_Expr := Fld;
      Inc_Ref (Var_Expr);
if Var_Expr /= Empty_String then
   --# scope-start
   --# bind fqn Var_Expr
--# expr-start 5504 '<to_symbol at ast.py:9410>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9410
Sym := String_To_Symbol (Self.Unit.Context, Var_Expr); 
--# expr-done 5504
                Singleton := Create_Symbol_Type_Array (Items_Count => 1);
                Singleton.Items (1) := Sym;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2502;
   Result_Var := Scope_Result;
else
   Result_Var := No_Symbol_Type_Array_Type;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 5505
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_752;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_752;
               Finalizer_Scope_2502;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Single_Protected_Decl_P_Env_Names;
--# end
--# property-start SingleProtectedDecl.declarative_parts /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9414
pragma Warnings (Off, "is not referenced");
 function Single_Protected_Decl_P_Declarative_Parts
  (Node : Bare_Single_Protected_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Declarative_Part_Array_Access
is
   Self : Bare_Single_Protected_Decl := Bare_Single_Protected_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Single_Protected_Decl :=
     Internal_Entity_Single_Protected_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_753 with Inline_Always;
            procedure Finalizer_Scope_2503 with Inline_Always;
      Block_Var : Internal_Entity_Protected_Def;
Var_Expr : Internal_Entity_Declarative_Part;
Fld : Internal_Entity_Protected_Def;
Fld_1 : Internal_Entity_Public_Part;
Cast_Result : Internal_Entity_Declarative_Part;
Singleton : Internal_Entity_Declarative_Part_Array_Access;
Fld_2 : Internal_Entity_Private_Part;
Cast_Result_1 : Internal_Entity_Declarative_Part;
Singleton_1 : Internal_Entity_Declarative_Part_Array_Access;
Scope_Result : Internal_Entity_Declarative_Part_Array_Access;
Result_Var : Internal_Entity_Declarative_Part_Array_Access;
Concat_Result : Internal_Entity_Declarative_Part_Array_Access;
Let_Result : Internal_Entity_Declarative_Part_Array_Access;
            procedure Finalizer_Scope_753 is
            begin
                     Dec_Ref (Singleton);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_753;
            procedure Finalizer_Scope_2503 is
            begin
                     Dec_Ref (Singleton_1);
            end Finalizer_Scope_2503;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5512 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5506 '<FieldAccess for definition at ast.py:9415>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9415
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Protected_Def (Node => Ent.Node.Single_Protected_Decl_F_Definition, Info => Ent.Info);
--# expr-done 5506
Block_Var := Fld; 
--# bind pdef Block_Var
--# expr-start 5511 '<Concat at ast.py:9416>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9416
--# expr-start 5508 '<Cast to DeclarativePart at ast.py:9416>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9416
--# expr-start 5507 '<FieldAccess for public_part at ast.py:9416>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9416
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Public_Part (Node => Block_Var.Node.Protected_Def_F_Public_Part, Info => Block_Var.Info);
--# expr-done 5507
      Cast_Result := Create_Internal_Entity_Declarative_Part
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
--# expr-done 5508
                Singleton := Create_Internal_Entity_Declarative_Part_Array (Items_Count => 1);
                Singleton.Items (1) := Cast_Result;
--# expr-start 5510 '<Cast to DeclarativePart at ast.py:9417>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9417
--# expr-start 5509 '<FieldAccess for private_part at ast.py:9417>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9417
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Private_Part (Node => Block_Var.Node.Protected_Def_F_Private_Part, Info => Block_Var.Info);
--# expr-done 5509
      Cast_Result_1 := Create_Internal_Entity_Declarative_Part
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
--# expr-done 5510
Var_Expr := Cast_Result_1;
if Var_Expr /= No_Entity_Declarative_Part then
   --# scope-start
                Singleton_1 := Create_Internal_Entity_Declarative_Part_Array (Items_Count => 1);
                Singleton_1.Items (1) := Var_Expr;
Scope_Result := Singleton_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2503;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Declarative_Part_Array_Type;
end if;
      Inc_Ref (Result_Var);
Concat_Result := Concat (Singleton, Result_Var); 
--# expr-done 5511
Let_Result := Concat_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 5512
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_753;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_753;
               Finalizer_Scope_2503;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Single_Protected_Decl_P_Declarative_Parts;
--# end
--# property-start SingleProtectedDecl._env_mappings_40 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9421
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_40
  (Node : Bare_Single_Protected_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Env_Assoc
is
   Self : Bare_Single_Protected_Decl := Bare_Single_Protected_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Single_Protected_Decl :=
     Internal_Entity_Single_Protected_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1457 with Inline_Always;
      Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Single_Protected_Decl;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1457 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1457;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5515 '<New EnvAssoc at ast.py:9421>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9421
--# expr-start 5513 '<New DesignatedEnv at ast.py:9421>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9421
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 5513
--# expr-start 5514 '<FieldAccess for name_symbol at ast.py:9421>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9421
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 5514
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 5515
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1457;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1457;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_40;
--# end
--# property-start SingleProtectedDecl._env_trans_parent_41 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9420
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_41
  (Node : Bare_Single_Protected_Decl
  )
   return Boolean
is
   Self : Bare_Single_Protected_Decl := Bare_Single_Protected_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_41;
--# end
--# property-start SingleProtectedDecl._env_names_42 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9422
pragma Warnings (Off, "is not referenced");
 function Env_Names_42
  (Node : Bare_Single_Protected_Decl
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Single_Protected_Decl := Bare_Single_Protected_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1459 with Inline_Always;
      Fld : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1459 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1459;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5516 '<FieldAccess for env_names at ast.py:9422>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9422
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleProtectedDecl.env_names
Fld := Libadalang.Implementation.Single_Protected_Decl_P_Env_Names (Node => Self);
--# end
--# expr-done 5516
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1459;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1459;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Names_42;
--# end
   --
   --  Primitives for Bare_Single_Task_Decl
   --
      procedure Single_Task_Decl_Pre_Env_Actions
        (Self            : Bare_Single_Task_Decl;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_37 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:9383");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_38 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            Libadalang.Implementation.Env_Names_39 (Node => 
Self);
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Single_Task_Decl
        (Self : Bare_Single_Task_Decl
         ; Single_Task_Decl_F_Task_Type : Bare_Single_Task_Type_Decl
        ) is
      begin
            Self.Single_Task_Decl_F_Task_Type := Single_Task_Decl_F_Task_Type;
      end Initialize_Fields_For_Single_Task_Decl;
   function Single_Task_Decl_F_Task_Type
     (Node : Bare_Single_Task_Decl) return Bare_Single_Task_Type_Decl
   is
   begin
         return Node.Single_Task_Decl_F_Task_Type;
   end;
--# property-start SingleTaskDecl.defining_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9371
pragma Warnings (Off, "is not referenced");
 function Single_Task_Decl_P_Defining_Names
  (Node : Bare_Single_Task_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Single_Task_Decl := Bare_Single_Task_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Single_Task_Decl :=
     Internal_Entity_Single_Task_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_744 with Inline_Always;
      Fld : Internal_Entity_Single_Task_Type_Decl;
Fld_1 : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_744 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_744;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5518 '<FieldAccess for defining_names at ast.py:9371>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9371
--# expr-start 5517 '<FieldAccess for task_type at ast.py:9371>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9371
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Single_Task_Type_Decl (Node => Ent.Node.Single_Task_Decl_F_Task_Type, Info => Ent.Info);
--# expr-done 5517
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_names'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Names (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 5518
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_744;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_744;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Single_Task_Decl_P_Defining_Names;
--# end
--# property-start SingleTaskDecl.expr_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9372
pragma Warnings (Off, "is not referenced");
 function Single_Task_Decl_P_Expr_Type
  (Node : Bare_Single_Task_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Single_Task_Decl := Bare_Single_Task_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Single_Task_Decl :=
     Internal_Entity_Single_Task_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Single_Task_Type_Decl;
Cast_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 5519 '<FieldAccess for task_type at ast.py:9372>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9372
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Single_Task_Type_Decl (Node => Ent.Node.Single_Task_Decl_F_Task_Type, Info => Ent.Info);
--# expr-done 5519
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Single_Task_Decl_P_Expr_Type;
--# end
--# property-start SingleTaskDecl.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9374
pragma Warnings (Off, "is not referenced");
 function Single_Task_Decl_P_Defining_Env
  (Node : Bare_Single_Task_Decl
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Single_Task_Decl := Bare_Single_Task_Decl (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Single_Task_Decl :=
     Internal_Entity_Single_Task_Decl'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_746 with Inline_Always;
      Fld : Internal_Entity_Single_Task_Type_Decl;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_746 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_746;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5521 '<FieldAccess for defining_env at ast.py:9374>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9374
--# expr-start 5520 '<FieldAccess for task_type at ast.py:9374>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9374
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Single_Task_Type_Decl (Node => Ent.Node.Single_Task_Decl_F_Task_Type, Info => Ent.Info);
--# expr-done 5520
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Fld.Node, Origin => Origin, Include_Ud_Indexing => Include_Ud_Indexing, E_Info => Fld.Info);
--# end
--# expr-done 5521
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_746;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_746;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Single_Task_Decl_P_Defining_Env;
--# end
--# property-start SingleTaskDecl.env_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9377
pragma Warnings (Off, "is not referenced");
 function Single_Task_Decl_P_Env_Names
  (Node : Bare_Single_Task_Decl
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Single_Task_Decl := Bare_Single_Task_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_747 with Inline_Always;
            procedure Finalizer_Scope_2504 with Inline_Always;
      Var_Expr : String_Type;
Fld : String_Type;
Sym : Symbol_Type;
Singleton : Symbol_Type_Array_Access;
Scope_Result : Symbol_Type_Array_Access;
Result_Var : Symbol_Type_Array_Access;
Let_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_747 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_747;
            procedure Finalizer_Scope_2504 is
            begin
                     Dec_Ref (Var_Expr);
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2504;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5524 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5522 '<FieldAccess for top_level_env_name at ast.py:9378>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9378
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.top_level_env_name
Fld := Libadalang.Implementation.Basic_Decl_P_Top_Level_Env_Name (Node => Self);
--# end
--# expr-done 5522
Var_Expr := Fld;
      Inc_Ref (Var_Expr);
if Var_Expr /= Empty_String then
   --# scope-start
   --# bind fqn Var_Expr
--# expr-start 5523 '<to_symbol at ast.py:9379>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9379
Sym := String_To_Symbol (Self.Unit.Context, Var_Expr); 
--# expr-done 5523
                Singleton := Create_Symbol_Type_Array (Items_Count => 1);
                Singleton.Items (1) := Sym;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2504;
   Result_Var := Scope_Result;
else
   Result_Var := No_Symbol_Type_Array_Type;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 5524
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_747;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_747;
               Finalizer_Scope_2504;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Single_Task_Decl_P_Env_Names;
--# end
--# property-start SingleTaskDecl._env_mappings_37 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9383
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_37
  (Node : Bare_Single_Task_Decl
  )
   return Internal_Env_Assoc
is
   Self : Bare_Single_Task_Decl := Bare_Single_Task_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1454 with Inline_Always;
      Fld : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Single_Task_Decl;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1454 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1454;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5527 '<New EnvAssoc at ast.py:9383>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9383
--# expr-start 5525 '<New DesignatedEnv at ast.py:9383>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9383
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 5525
--# expr-start 5526 '<FieldAccess for name_symbol at ast.py:9383>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9383
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Self);
--# end
--# expr-done 5526
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 5527
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1454;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1454;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_37;
--# end
--# property-start SingleTaskDecl._env_trans_parent_38 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9382
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_38
  (Node : Bare_Single_Task_Decl
  )
   return Boolean
is
   Self : Bare_Single_Task_Decl := Bare_Single_Task_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_38;
--# end
--# property-start SingleTaskDecl._env_names_39 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9384
pragma Warnings (Off, "is not referenced");
 function Env_Names_39
  (Node : Bare_Single_Task_Decl
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Single_Task_Decl := Bare_Single_Task_Decl (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1456 with Inline_Always;
      Fld : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1456 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1456;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5528 '<FieldAccess for env_names at ast.py:9384>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9384
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTaskDecl.env_names
Fld := Libadalang.Implementation.Single_Task_Decl_P_Env_Names (Node => Self);
--# end
--# expr-done 5528
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1456;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1456;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Names_39;
--# end
   --
   --  Primitives for Bare_Case_Stmt_Alternative
   --
      procedure Initialize_Fields_For_Case_Stmt_Alternative
        (Self : Bare_Case_Stmt_Alternative
         ; Case_Stmt_Alternative_F_Choices : Bare_Alternatives_List
         ; Case_Stmt_Alternative_F_Stmts : Bare_Stmt_List
        ) is
      begin
            Self.Case_Stmt_Alternative_F_Choices := Case_Stmt_Alternative_F_Choices;
            Self.Case_Stmt_Alternative_F_Stmts := Case_Stmt_Alternative_F_Stmts;
      end Initialize_Fields_For_Case_Stmt_Alternative;
   function Case_Stmt_Alternative_F_Choices
     (Node : Bare_Case_Stmt_Alternative) return Bare_Alternatives_List
   is
   begin
         return Node.Case_Stmt_Alternative_F_Choices;
   end;
   function Case_Stmt_Alternative_F_Stmts
     (Node : Bare_Case_Stmt_Alternative) return Bare_Stmt_List
   is
   begin
         return Node.Case_Stmt_Alternative_F_Stmts;
   end;
--# property-start CaseStmtAlternative.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19559
pragma Warnings (Off, "is not referenced");
 function Case_Stmt_Alternative_P_Xref_Entry_Point
  (Node : Bare_Case_Stmt_Alternative
  )
   return Boolean
is
   Self : Bare_Case_Stmt_Alternative := Bare_Case_Stmt_Alternative (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Case_Stmt_Alternative_P_Xref_Entry_Point;
--# end
--# property-start CaseStmtAlternative.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19562
pragma Warnings (Off, "is not referenced");
 function Case_Stmt_Alternative_P_Xref_Equation
  (Node : Bare_Case_Stmt_Alternative
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Case_Stmt_Alternative := Bare_Case_Stmt_Alternative (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Case_Stmt_Alternative :=
     Internal_Entity_Case_Stmt_Alternative'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1367 with Inline_Always;
            procedure Finalizer_Scope_2505 with Inline_Always;
            procedure Finalizer_Scope_2506 with Inline_Always;
            procedure Finalizer_Scope_2508 with Inline_Always;
            procedure Finalizer_Scope_2509 with Inline_Always;
      Block_Var : Internal_Entity_Case_Stmt;
Block_Var_1 : Internal_Entity_Base_Type_Decl;
Match_0 : Internal_Entity_Expr;
Match_1 : Internal_Entity_Subtype_Indication;
Match_2 : Internal_Entity_Others_Designator;
Match_3 : Internal_Entity;
Var_Expr : Internal_Entity_Name;
Fld : Internal_Entity;
Fld_1 : Internal_Entity;
Cast_Result : Internal_Entity_Case_Stmt;
Fld_2 : Internal_Entity_Expr;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Internal_Entity_Alternatives_List;
Fld_5 : Bare_Alternatives_List;
Fld_6 : Internal_Entity_Info;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Cast_Expr : Internal_Entity_Expr;
Cast_Result_1 : Internal_Entity_Name;
Fld_7 : Logic_Equation;
Fld_8 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_9 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_10 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Cast_Expr_2 : Internal_Entity_Expr;
Cast_Result_3 : Internal_Entity_Name;
Fld_11 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Is_Null_1 : Boolean;
Not_Val : Boolean;
If_Result_1 : Logic_Equation;
Fld_12 : Logic_Equation;
True_Rel : Logic_Equation;
Match_Prefix : Internal_Entity;
Cast_Result_4 : Internal_Entity_Expr;
Let_Result : Logic_Equation;
Scope_Result_1 : Logic_Equation;
Cast_Result_5 : Internal_Entity_Subtype_Indication;
Let_Result_1 : Logic_Equation;
Scope_Result_2 : Logic_Equation;
Cast_Result_6 : Internal_Entity_Others_Designator;
Let_Result_2 : Logic_Equation;
Scope_Result_3 : Logic_Equation;
Cast_Result_7 : Internal_Entity;
Let_Result_3 : Logic_Equation;
Scope_Result_4 : Logic_Equation;
Match_Result : Logic_Equation;
Item_193 : Internal_Entity;
Bare_Item_193 : Bare_Ada_Node;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Let_Result_4 : Logic_Equation;
            procedure Finalizer_Scope_1367 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Let_Result_4);
            end Finalizer_Scope_1367;
            procedure Finalizer_Scope_2505 is
            begin
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Let_Result_2);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Let_Result_3);
                     Dec_Ref (Scope_Result_4);
                     Dec_Ref (Match_Result);
            end Finalizer_Scope_2505;
            procedure Finalizer_Scope_2506 is
            begin
                     Dec_Ref (Fld_7);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_9);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_10);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (If_Result_1);
            end Finalizer_Scope_2506;
            procedure Finalizer_Scope_2508 is
            begin
                     Dec_Ref (Fld_12);
            end Finalizer_Scope_2508;
            procedure Finalizer_Scope_2509 is
            begin
                     Dec_Ref (True_Rel);
            end Finalizer_Scope_2509;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5552 '<Block at ???>' Let_Result_4 None
--# scope-start
--# expr-start 5531 '<Cast to CaseStmt at ast.py:19563>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19563
--# expr-start 5530 '<parent at ast.py:19563>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19563
--# expr-start 5529 '<parent at ast.py:19563>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19563
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 5529
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Parent (Node => Fld.Node, E_Info => Fld.Info);
--# expr-done 5530
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Case_Stmt_Range
   then
      Cast_Result := Create_Internal_Entity_Case_Stmt
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 5531
Block_Var := Cast_Result; 
--# bind case_stmt Block_Var
--# expr-start 5533 '<FieldAccess for expression_type at ast.py:19566>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19566
--# expr-start 5532 '<FieldAccess for expr at ast.py:19566>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19566
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Expr (Node => Block_Var.Node.Case_Stmt_F_Expr, Info => Block_Var.Info);
--# expr-done 5532
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.expression_type
Fld_3 := Libadalang.Implementation.Expr_P_Expression_Type (Node => Fld_2.Node, E_Info => Fld_2.Info);
--# end
--# expr-done 5533
Block_Var_1 := Fld_3; 
--# bind selected_type Block_Var_1
--# expr-start 5551 '<LogicAll at ast.py:19568>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19568
--# expr-start 5550 '<Map at ast.py:19568>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19568
   --# expr-start 5534 '<FieldAccess for choices at ast.py:19568>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19568
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Alternatives_List (Node => Ent.Node.Case_Stmt_Alternative_F_Choices, Info => Ent.Info);
--# expr-done 5534
Fld_5 := Fld_4.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_5 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Alternatives_List := Fld_5;
      begin
         for Untyped_Item_193 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_193 := Untyped_Item_193; 
Is_Null := Bare_Item_193 = null; 
if Is_Null then
   If_Result := No_Entity;
else
Fld_6 := Fld_4.Info;
As_Entity := (Info => Fld_6, Node => Bare_Item_193); 
   If_Result := As_Entity;
end if;
                  Item_193 := If_Result; 
   --# scope-start
               --# bind c Item_193
      --# expr-start 5549 '<Match at ast.py:19568>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19568
      if Item_193.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Item_193; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Abstract_State_Decl_Expr .. Ada_Un_Op =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_4; 
--# bind e Match_0
--# expr-start 5547 '<If at ast.py:19571>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19571
--# expr-start 5538 '<Not at ast.py:19572>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19572
--# expr-start 5537 '<IsNull at ast.py:19572>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19572
--# expr-start 5535 '<Cast to Name at ast.py:19572>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19572
Cast_Expr_2 := Match_0; 
   if Cast_Expr_2.Node = null
      or else Cast_Expr_2.Node.Kind in Ada_Name
   then
      Cast_Result_3 := Create_Internal_Entity_Name
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
   else
         Cast_Result_3 := No_Entity_Name;
   end if;
--# expr-done 5535
Var_Expr := Cast_Result_3;
if Var_Expr /= No_Entity_Name then
   --# scope-start
--# expr-start 5536 '<FieldAccess for name_designated_type at ast.py:19572>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19572
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_11 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 5536
Scope_Result := Fld_11;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Is_Null_1 := Result_Var.Node = null; 
--# expr-done 5537
Not_Val := not (Is_Null_1); 
--# expr-done 5538
if Not_Val then
   --# expr-start 5540 '<FieldAccess for xref_no_overloading at ast.py:19574>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19574
--# expr-start 5539 '<Cast to Name at ast.py:19574>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19574
Cast_Expr := Match_0; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Name
   then
      Cast_Result_1 := Create_Internal_Entity_Name
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result_1 := No_Entity_Name;
   end if;
--# expr-done 5539
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_7 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Cast_Result_1.Node, Env => Env, Origin => Origin, E_Info => Cast_Result_1.Info);
--# end
--# expr-done 5540
   If_Result_1 := Fld_7;
else
   --# expr-start 5546 '<And at ast.py:19576>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19576
--# expr-start 5544 '<And at ast.py:19576>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19576
--# expr-start 5542 '<Bind at ast.py:19576>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19576
--# expr-start 5541 '<FieldAccess for expected_type_var at ast.py:19576>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19576
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Match_0.Node);
--# expr-done 5541
Fld_8.Value := No_Entity;
Entity_Vars.Reset (Fld_8);
Cast_Expr_1 := Block_Var_1; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result := Solver.Create_Assign (Fld_8, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19576") else null)); 
--# expr-done 5542
--# expr-start 5543 '<FieldAccess for sub_equation at ast.py:19577>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19577
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_9 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Match_0.Node, Env => Env, Origin => Origin, E_Info => Match_0.Info);
--# end
--# expr-done 5543
And_Pred := Create_And (Bind_Result, Fld_9, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19576") else null)); 
--# expr-done 5544
--# expr-start 5545 '<FieldAccess for matches_expected_type at ast.py:19578>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19578
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_10 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Match_0.Node, Origin => Origin);
--# end
--# expr-done 5545
And_Pred_1 := Create_And (And_Pred, Fld_10, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19576") else null)); 
--# expr-done 5546
   If_Result_1 := And_Pred_1;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 5547
Let_Result := If_Result_1; Inc_Ref (Let_Result);
--# end
Scope_Result_1 := Let_Result;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2506;
            Match_Result := Scope_Result_1; Inc_Ref (Match_Result);
         when Ada_Subtype_Indication .. Ada_Discrete_Subtype_Indication =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Create_Internal_Entity_Subtype_Indication
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_5; 
--# bind t Match_1
--# expr-start 5548 '<FieldAccess for xref_equation at ast.py:19582>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19582
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_12 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Match_1.Node, Env => Env, Origin => Origin, E_Info => Match_1.Info);
--# end
--# expr-done 5548
Let_Result_1 := Fld_12; Inc_Ref (Let_Result_1);
--# end
Scope_Result_2 := Let_Result_1;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2508;
            Match_Result := Scope_Result_2; Inc_Ref (Match_Result);
         when Ada_Others_Designator =>
   --# scope-start
--# scope-start
      Cast_Result_6 := Create_Internal_Entity_Others_Designator
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_6; 
--# bind _ Match_2
True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19584") else null)); 
Let_Result_2 := True_Rel; Inc_Ref (Let_Result_2);
--# end
Scope_Result_3 := Let_Result_2;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_2509;
            Match_Result := Scope_Result_3; Inc_Ref (Match_Result);
         when Ada_Abort_Absent .. Ada_Elsif_Stmt_Part | Ada_Handled_Stmts .. Ada_Null_Component_Decl | Ada_Overriding_Not_Overriding .. Ada_Enum_Lit_Synth_Type_Expr | Ada_Synthetic_Type_Expr .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_7 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_7; 
--# bind _ Match_3
Let_Result_3 := raise Property_Error with "Should not happen"; Inc_Ref (Let_Result_3);
--# end
Scope_Result_4 := Let_Result_3;
   Inc_Ref (Scope_Result_4);
   --# end
            Match_Result := Scope_Result_4; Inc_Ref (Match_Result);
end case;
--# expr-done 5549
         declare
            Item_To_Append : constant Logic_Equation := Match_Result;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2505;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 5550
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19568") else null)); 
--# expr-done 5551
Let_Result_4 := Logic_Boolean_Op; Inc_Ref (Let_Result_4);
--# end
--# expr-done 5552
      Property_Result := Let_Result_4;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1367;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1367;
               Finalizer_Scope_2505;
               Finalizer_Scope_2506;
               Finalizer_Scope_2508;
               Finalizer_Scope_2509;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Case_Stmt_Alternative_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Compilation_Unit
   --
      procedure Initialize_Fields_For_Compilation_Unit
        (Self : Bare_Compilation_Unit
         ; Compilation_Unit_F_Prelude : Bare_Ada_Node_List
         ; Compilation_Unit_F_Body : Bare_Ada_Node
         ; Compilation_Unit_F_Pragmas : Bare_Pragma_Node_List
        ) is
      begin
            Self.Compilation_Unit_F_Prelude := Compilation_Unit_F_Prelude;
            Self.Compilation_Unit_F_Body := Compilation_Unit_F_Body;
            Self.Compilation_Unit_F_Pragmas := Compilation_Unit_F_Pragmas;
      Self.Compilation_Unit_No_Env := Empty_Env;
      end Initialize_Fields_For_Compilation_Unit;
   function Compilation_Unit_F_Prelude
     (Node : Bare_Compilation_Unit) return Bare_Ada_Node_List
   is
   begin
         return Node.Compilation_Unit_F_Prelude;
   end;
   function Compilation_Unit_F_Body
     (Node : Bare_Compilation_Unit) return Bare_Ada_Node
   is
   begin
         return Node.Compilation_Unit_F_Body;
   end;
   function Compilation_Unit_F_Pragmas
     (Node : Bare_Compilation_Unit) return Bare_Pragma_Node_List
   is
   begin
         return Node.Compilation_Unit_F_Pragmas;
   end;
--# property-start CompilationUnit.syntactic_fully_qualified_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18393
pragma Warnings (Off, "is not referenced");
 function Compilation_Unit_P_Syntactic_Fully_Qualified_Name
  (Node : Bare_Compilation_Unit
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Compilation_Unit := Bare_Compilation_Unit (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1303 with Inline_Always;
            procedure Finalizer_Scope_1719 with Inline_Always;
            procedure Finalizer_Scope_1720 with Inline_Always;
            procedure Finalizer_Scope_1721 with Inline_Always;
      Match_0 : Internal_Entity_Library_Item;
Match_1 : Internal_Entity_Subunit;
Match_2 : Internal_Entity;
Var_Expr : Internal_Entity_Defining_Name;
As_Entity : Internal_Entity_Compilation_Unit;
Is_Null : Boolean;
If_Result : Internal_Entity_Compilation_Unit;
Fld : Internal_Entity;
Fld_1 : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity_Defining_Name;
Fld_3 : Symbol_Type_Array_Access;
Scope_Result : Symbol_Type_Array_Access;
Result_Var : Symbol_Type_Array_Access;
Fld_4 : Internal_Entity_Name;
Fld_5 : Symbol_Type_Array_Access;
Fld_6 : Internal_Entity_Body_Node;
Fld_7 : Internal_Entity_Defining_Name;
Fld_8 : Symbol_Type_Array_Access;
Concat_Result : Symbol_Type_Array_Access;
Match_Prefix : Internal_Entity;
Cast_Result : Internal_Entity_Library_Item;
Let_Result : Symbol_Type_Array_Access;
Scope_Result_1 : Symbol_Type_Array_Access;
Cast_Result_1 : Internal_Entity_Subunit;
Let_Result_1 : Symbol_Type_Array_Access;
Scope_Result_2 : Symbol_Type_Array_Access;
Cast_Result_2 : Internal_Entity;
Let_Result_2 : Symbol_Type_Array_Access;
Scope_Result_3 : Symbol_Type_Array_Access;
Match_Result : Symbol_Type_Array_Access;
Let_Result_3 : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1303 is
            begin
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Let_Result_2);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Match_Result);
                     Dec_Ref (Let_Result_3);
            end Finalizer_Scope_1303;
            procedure Finalizer_Scope_1719 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1719;
            procedure Finalizer_Scope_1720 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_1720;
            procedure Finalizer_Scope_1721 is
            begin
                     Dec_Ref (Fld_5);
                     Dec_Ref (Fld_8);
                     Dec_Ref (Concat_Result);
            end Finalizer_Scope_1721;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5565 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 5564 '<Match at ast.py:18397>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18397
--# expr-start 5554 '<FieldAccess for body at ast.py:18397>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18397
--# expr-start 5553 '<as_bare_entity at ast.py:18397>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18397
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity_Compilation_Unit;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 5553
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity (Node => If_Result.Node.Compilation_Unit_F_Body, Info => If_Result.Info);
--# expr-done 5554
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Library_Item =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Library_Item
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind li Match_0
--# expr-start 5556 '<FieldAccess for defining_name at ast.py:18398>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18398
--# expr-start 5555 '<FieldAccess for item at ast.py:18398>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18398
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Basic_Decl (Node => Match_0.Node.Library_Item_F_Item, Info => Match_0.Info);
--# expr-done 5555
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 5556
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 5557 '<FieldAccess for as_symbol_array at ast.py:18398>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18398
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.as_symbol_array'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_As_Symbol_Array (Node => Var_Expr.Node);
--# end
--# expr-done 5557
Scope_Result := Fld_3;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_1720;
   Result_Var := Scope_Result;
else
   Result_Var := No_Symbol_Type_Array_Type;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
Scope_Result_1 := Let_Result;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_1719;
            Match_Result := Scope_Result_1; Inc_Ref (Match_Result);
         when Ada_Subunit =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Subunit
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind su Match_1
--# expr-start 5563 '<Concat at ast.py:18399>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18399
--# expr-start 5559 '<FieldAccess for as_symbol_array at ast.py:18399>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18399
--# expr-start 5558 '<FieldAccess for name at ast.py:18399>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18399
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Name (Node => Match_1.Node.Subunit_F_Name, Info => Match_1.Info);
--# expr-done 5558
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.as_symbol_array'
Fld_5 := Libadalang.Implementation.Dispatcher_Name_P_As_Symbol_Array (Node => Fld_4.Node);
--# end
--# expr-done 5559
--# expr-start 5562 '<FieldAccess for as_symbol_array at ast.py:18400>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18400
--# expr-start 5561 '<FieldAccess for defining_name at ast.py:18400>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18400
--# expr-start 5560 '<FieldAccess for body at ast.py:18400>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18400
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Body_Node (Node => Match_1.Node.Subunit_F_Body, Info => Match_1.Info);
--# expr-done 5560
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_7 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Fld_6.Node, E_Info => Fld_6.Info);
--# end
--# expr-done 5561
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.as_symbol_array'
Fld_8 := Libadalang.Implementation.Dispatcher_Name_P_As_Symbol_Array (Node => Fld_7.Node);
--# end
--# expr-done 5562
Concat_Result := Concat (Fld_5, Fld_8); 
--# expr-done 5563
Let_Result_1 := Concat_Result; Inc_Ref (Let_Result_1);
--# end
Scope_Result_2 := Let_Result_1;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_1721;
            Match_Result := Scope_Result_2; Inc_Ref (Match_Result);
         when Ada_Abort_Absent .. Ada_Iter_Type_Of | Ada_Limited_Absent .. Ada_Subp_Kind_Procedure | Ada_Synchronized_Absent .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind _ Match_2
Let_Result_2 := raise Property_Error with "Unexpected CompilationUnit.f_body attribute"; Inc_Ref (Let_Result_2);
--# end
Scope_Result_3 := Let_Result_2;
   Inc_Ref (Scope_Result_3);
   --# end
            Match_Result := Scope_Result_3; Inc_Ref (Match_Result);
end case;
--# expr-done 5564
Let_Result_3 := Match_Result; Inc_Ref (Let_Result_3);
--# end
--# expr-done 5565
      Property_Result := Let_Result_3;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1303;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1303;
               Finalizer_Scope_1719;
               Finalizer_Scope_1720;
               Finalizer_Scope_1721;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Compilation_Unit_P_Syntactic_Fully_Qualified_Name;
--# end
--# property-start CompilationUnit.unit_kind /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18408
pragma Warnings (Off, "is not referenced");
 function Compilation_Unit_P_Unit_Kind
  (Node : Bare_Compilation_Unit
  )
   return Analysis_Unit_Kind
is
   Self : Bare_Compilation_Unit := Bare_Compilation_Unit (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Analysis_Unit_Kind;
      Match_0 : Bare_Library_Item;
Match_1 : Bare_Subunit;
Match_2 : Bare_Ada_Node;
Match_0_1 : Bare_Body_Node;
Match_1_1 : Bare_Basic_Decl;
Fld : Bare_Ada_Node;
Fld_1 : Bare_Basic_Decl;
Match_Prefix : Bare_Basic_Decl;
Cast_Result : Bare_Body_Node;
Let_Result : Analysis_Unit_Kind;
Scope_Result : Analysis_Unit_Kind;
Cast_Result_1 : Bare_Basic_Decl;
Let_Result_1 : Analysis_Unit_Kind;
Scope_Result_1 : Analysis_Unit_Kind;
Match_Result : Analysis_Unit_Kind;
Match_Prefix_1 : Bare_Ada_Node;
Cast_Result_2 : Bare_Library_Item;
Let_Result_2 : Analysis_Unit_Kind;
Scope_Result_2 : Analysis_Unit_Kind;
Cast_Result_3 : Bare_Subunit;
Let_Result_3 : Analysis_Unit_Kind;
Scope_Result_3 : Analysis_Unit_Kind;
Cast_Result_4 : Bare_Ada_Node;
Let_Result_4 : Analysis_Unit_Kind;
Scope_Result_4 : Analysis_Unit_Kind;
Match_Result_1 : Analysis_Unit_Kind;
Let_Result_5 : Analysis_Unit_Kind;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5570 '<Block at ???>' Let_Result_5 None
--# scope-start
--# expr-start 5569 '<Match at ast.py:18412>' Match_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18412
--# expr-start 5566 '<FieldAccess for body at ast.py:18412>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18412
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Compilation_Unit_F_Body;
--# expr-done 5566
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix_1 := Fld; 
case Ada_Ada_Node (Match_Prefix_1.Kind) is
         when Ada_Library_Item =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Match_Prefix_1;
Match_0 := Cast_Result_2; 
--# bind li Match_0
--# expr-start 5568 '<Match at ast.py:18413>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18413
--# expr-start 5567 '<FieldAccess for item at ast.py:18413>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18413
      if Match_0 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Match_0.Library_Item_F_Item;
--# expr-done 5567
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld_1; 
case Ada_Basic_Decl (Match_Prefix.Kind) is
         when Ada_Expr_Function .. Ada_Task_Body =>
   --# scope-start
--# scope-start
      Cast_Result := Match_Prefix;
Match_0_1 := Cast_Result; 
--# bind _ Match_0_1
Let_Result := Unit_Body; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Abstract_State_Decl .. Ada_Synthetic_Subp_Decl | Ada_Entry_Index_Spec .. Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Match_Prefix;
Match_1_1 := Cast_Result_1; 
--# bind _ Match_1_1
Let_Result_1 := Unit_Specification; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
end case;
--# expr-done 5568
Let_Result_2 := Match_Result; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result_1 := Scope_Result_2; 
         when Ada_Subunit =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Match_Prefix_1;
Match_1 := Cast_Result_3; 
--# bind _ Match_1
Let_Result_3 := Unit_Body; 
--# end
Scope_Result_3 := Let_Result_3;
   --# end
            Match_Result_1 := Scope_Result_3; 
         when Ada_Abort_Absent .. Ada_Iter_Type_Of | Ada_Limited_Absent .. Ada_Subp_Kind_Procedure | Ada_Synchronized_Absent .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Match_Prefix_1;
Match_2 := Cast_Result_4; 
--# bind _ Match_2
Let_Result_4 := raise Property_Error with "Unexpected CompilationUnit.f_body attribute"; 
--# end
Scope_Result_4 := Let_Result_4;
   --# end
            Match_Result_1 := Scope_Result_4; 
end case;
--# expr-done 5569
Let_Result_5 := Match_Result_1; 
--# end
--# expr-done 5570
      Property_Result := Let_Result_5;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Compilation_Unit_P_Unit_Kind;
--# end
--# property-start CompilationUnit.withed_units /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18425
pragma Warnings (Off, "is not referenced");
 function Compilation_Unit_P_Withed_Units
  (Node : Bare_Compilation_Unit
  )
   return Internal_Entity_Compilation_Unit_Array_Access
is
   Self : Bare_Compilation_Unit := Bare_Compilation_Unit (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1305 with Inline_Always;
      Fld : Bare_Name_Array_Access;
Fld_1 : Bare_Compilation_Unit;
As_Entity : Internal_Entity_Compilation_Unit;
Is_Null : Boolean;
If_Result : Internal_Entity_Compilation_Unit;
Map_Result : Internal_Entity_Compilation_Unit_Array_Access;
Let_Result : Internal_Entity_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1305 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1305;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Compilation_Unit_Compilation_Unit_P_Withed_Units,
            Items    => new Mmz_Key_Array (1 ..  1))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Compilation_Unit,
                                As_Bare_Compilation_Unit => Self);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Compilation_Unit_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 5575 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5574 '<Map at ast.py:18432>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18432
   --# expr-start 5571 '<FieldAccess for top_level_with_package_clauses at ast.py:18432>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18432
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.top_level_with_package_clauses
Fld := Libadalang.Implementation.Ada_Node_P_Top_Level_With_Package_Clauses (Node => Self);
--# end
--# expr-done 5571
      declare
         Map_Result_Vec : Internal_Entity_Compilation_Unit_Vectors.Vector;
      begin
      declare
         Collection : constant Bare_Name_Array_Access := Fld;
      begin
         for Item_194 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_194
      --# expr-start 5573 '<as_bare_entity at ast.py:18436>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18436
--# expr-start 5572 '<FieldAccess for withed_unit_helper at ast.py:18436>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18436
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.withed_unit_helper
Fld_1 := Libadalang.Implementation.Ada_Node_P_Withed_Unit_Helper (Node => Self, Unit_Name => Item_194);
--# end
--# expr-done 5572
Is_Null := Fld_1 = null; 
if Is_Null then
   If_Result := No_Entity_Compilation_Unit;
else
As_Entity := (Info => No_Entity_Info, Node => Fld_1); 
   If_Result := As_Entity;
end if;
--# expr-done 5573
         declare
            Item_To_Append : constant Internal_Entity_Compilation_Unit := If_Result;
         begin
            Internal_Entity_Compilation_Unit_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Compilation_Unit_Array
           (Items_Count => Natural (Internal_Entity_Compilation_Unit_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Compilation_Unit_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Compilation_Unit_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Compilation_Unit_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 5574
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 5575
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1305;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Compilation_Unit_Array_Access,
                        As_Internal_Entity_Compilation_Unit_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1305;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Compilation_Unit_P_Withed_Units;
--# end
--# property-start CompilationUnit.imported_units /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18441
pragma Warnings (Off, "is not referenced");
 function Compilation_Unit_P_Imported_Units
  (Node : Bare_Compilation_Unit
  )
   return Internal_Entity_Compilation_Unit_Array_Access
is
   Self : Bare_Compilation_Unit := Bare_Compilation_Unit (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1306 with Inline_Always;
            procedure Finalizer_Scope_2512 with Inline_Always;
            procedure Finalizer_Scope_2513 with Inline_Always;
            procedure Finalizer_Scope_2514 with Inline_Always;
            procedure Finalizer_Scope_2515 with Inline_Always;
            procedure Finalizer_Scope_2516 with Inline_Always;
            procedure Finalizer_Scope_2517 with Inline_Always;
      Var_Expr : Internal_Entity_Compilation_Unit_Array_Access;
Var_Expr_1 : Bare_Base_Subp_Body;
Var_Expr_2 : Internal_Entity_Basic_Decl;
Var_Expr_3 : Bare_Ada_Node;
Var_Expr_4 : Lexical_Env := Empty_Env;
Var_Expr_5 : Bare_Basic_Decl;
Fld : Internal_Entity_Compilation_Unit_Array_Access;
Fld_1 : Bare_Basic_Decl;
Cast_Result : Bare_Base_Subp_Body;
As_Entity : Internal_Entity_Base_Subp_Body;
Is_Null : Boolean;
If_Result : Internal_Entity_Base_Subp_Body;
Fld_2 : Internal_Entity_Basic_Decl;
Fld_3 : Bare_Compilation_Unit;
As_Entity_1 : Internal_Entity_Compilation_Unit;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Compilation_Unit;
Singleton : Internal_Entity_Compilation_Unit_Array_Access;
Scope_Result : Internal_Entity_Compilation_Unit_Array_Access;
Result_Var : Internal_Entity_Compilation_Unit_Array_Access;
Scope_Result_1 : Internal_Entity_Compilation_Unit_Array_Access;
Result_Var_1 : Internal_Entity_Compilation_Unit_Array_Access;
Scope_Result_2 : Internal_Entity_Compilation_Unit_Array_Access;
Fld_4 : Bare_Basic_Decl;
Fld_5 : Lexical_Env := Empty_Env;
Scope_Result_3 : Lexical_Env := Empty_Env;
Result_Var_2 : Lexical_Env := Empty_Env;
Env_Node : Bare_Ada_Node;
Scope_Result_4 : Bare_Ada_Node;
Result_Var_3 : Bare_Ada_Node;
Fld_6 : Bare_Compilation_Unit;
As_Entity_2 : Internal_Entity_Compilation_Unit;
Is_Null_2 : Boolean;
If_Result_2 : Internal_Entity_Compilation_Unit;
Singleton_1 : Internal_Entity_Compilation_Unit_Array_Access;
Scope_Result_5 : Internal_Entity_Compilation_Unit_Array_Access;
Result_Var_4 : Internal_Entity_Compilation_Unit_Array_Access;
Result_Var_5 : Internal_Entity_Compilation_Unit_Array_Access;
Concat_Result : Internal_Entity_Compilation_Unit_Array_Access;
Let_Result : Internal_Entity_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1306 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Result_Var_2);
                     Dec_Ref (Scope_Result_5);
                     Dec_Ref (Result_Var_4);
                     Dec_Ref (Result_Var_5);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1306;
            procedure Finalizer_Scope_2512 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_2512;
            procedure Finalizer_Scope_2513 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2513;
            procedure Finalizer_Scope_2514 is
            begin
                     Dec_Ref (Var_Expr);
            end Finalizer_Scope_2514;
            procedure Finalizer_Scope_2515 is
            begin
                     Dec_Ref (Fld_5);
            end Finalizer_Scope_2515;
            procedure Finalizer_Scope_2516 is
            begin
                     Dec_Ref (Var_Expr_4);
            end Finalizer_Scope_2516;
            procedure Finalizer_Scope_2517 is
            begin
                     Dec_Ref (Singleton_1);
            end Finalizer_Scope_2517;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Compilation_Unit_Compilation_Unit_P_Imported_Units,
            Items    => new Mmz_Key_Array (1 ..  1))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Compilation_Unit,
                                As_Bare_Compilation_Unit => Self);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Compilation_Unit_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 5589 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5588 '<Concat at ast.py:18446>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18446
--# expr-start 5576 '<FieldAccess for withed_units at ast.py:18446>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18446
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.withed_units
Fld := Libadalang.Implementation.Compilation_Unit_P_Withed_Units (Node => Self);
--# end
--# expr-done 5576
--# expr-start 5578 '<Cast to BaseSubpBody at ast.py:18451>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18451
--# expr-start 5577 '<FieldAccess for decl at ast.py:18451>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18451
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.decl
Fld_1 := Libadalang.Implementation.Compilation_Unit_P_Decl (Node => Self);
--# end
--# expr-done 5577
   if Fld_1 = null
      or else Fld_1.Kind in Ada_Base_Subp_Body
   then
      Cast_Result := Fld_1;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 5578
Var_Expr_1 := Cast_Result;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind subp Var_Expr_1
--# expr-start 5580 '<FieldAccess for subp_previous_part at ast.py:18455>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18455
--# expr-start 5579 '<as_bare_entity at ast.py:18455>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18455
Is_Null := Var_Expr_1 = null; 
if Is_Null then
   If_Result := No_Entity_Base_Subp_Body;
else
As_Entity := (Info => No_Entity_Info, Node => Var_Expr_1); 
   If_Result := As_Entity;
end if;
--# expr-done 5579
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.subp_previous_part
Fld_2 := Libadalang.Implementation.Body_Node_P_Subp_Previous_Part (Node => If_Result.Node, Imprecise_Fallback => False, E_Info => If_Result.Info);
--# end
--# expr-done 5580
Var_Expr_2 := Fld_2;
if Var_Expr_2 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind pp Var_Expr_2
--# expr-start 5582 '<as_bare_entity at ast.py:18457>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18457
--# expr-start 5581 '<FieldAccess for enclosing_compilation_unit at ast.py:18457>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18457
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.enclosing_compilation_unit
Fld_3 := Libadalang.Implementation.Ada_Node_P_Enclosing_Compilation_Unit (Node => Var_Expr_2.Node);
--# end
--# expr-done 5581
Is_Null_1 := Fld_3 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Compilation_Unit;
else
As_Entity_1 := (Info => No_Entity_Info, Node => Fld_3); 
   If_Result_1 := As_Entity_1;
end if;
--# expr-done 5582
                Singleton := Create_Internal_Entity_Compilation_Unit_Array (Items_Count => 1);
                Singleton.Items (1) := If_Result_1;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2513;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Compilation_Unit_Array_Type;
end if;
      Inc_Ref (Result_Var);
Scope_Result_1 := Result_Var;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2512;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Internal_Entity_Compilation_Unit_Array_Type;
end if;
      Inc_Ref (Result_Var_1);
Var_Expr := Result_Var_1;
      Inc_Ref (Var_Expr);
if Var_Expr /= No_Internal_Entity_Compilation_Unit_Array_Type then
   --# scope-start
   --# bind e Var_Expr
Scope_Result_2 := Var_Expr;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2514;
   Result_Var_5 := Scope_Result_2;
else
--# expr-start 5583 '<FieldAccess for decl at ast.py:18459>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18459
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.decl
Fld_4 := Libadalang.Implementation.Compilation_Unit_P_Decl (Node => Self);
--# end
--# expr-done 5583
Var_Expr_5 := Fld_4;
if Var_Expr_5 /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 5584 '<FieldAccess for node_env at ast.py:18459>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18459
      if Var_Expr_5 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Node_Env (Node => Var_Expr_5);
--# expr-done 5584
Scope_Result_3 := Fld_5;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_2515;
   Result_Var_2 := Scope_Result_3;
else
   Result_Var_2 := Empty_Env;
end if;
      Inc_Ref (Result_Var_2);
Var_Expr_4 := Result_Var_2;
      Inc_Ref (Var_Expr_4);
if Var_Expr_4 /= Empty_Env then
   --# scope-start
--# expr-start 5585 '<env_node at ast.py:18459>' Env_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18459
Env_Node := AST_Envs.Env_Node (Var_Expr_4); 
--# expr-done 5585
Scope_Result_4 := Env_Node;
   --# end
      Finalizer_Scope_2516;
   Result_Var_3 := Scope_Result_4;
else
   Result_Var_3 := No_Bare_Ada_Node;
end if;
Var_Expr_3 := Result_Var_3;
if Var_Expr_3 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind n Var_Expr_3
--# expr-start 5587 '<as_bare_entity at ast.py:18461>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18461
--# expr-start 5586 '<FieldAccess for enclosing_compilation_unit at ast.py:18461>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18461
      if Var_Expr_3 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.enclosing_compilation_unit
Fld_6 := Libadalang.Implementation.Ada_Node_P_Enclosing_Compilation_Unit (Node => Var_Expr_3);
--# end
--# expr-done 5586
Is_Null_2 := Fld_6 = null; 
if Is_Null_2 then
   If_Result_2 := No_Entity_Compilation_Unit;
else
As_Entity_2 := (Info => No_Entity_Info, Node => Fld_6); 
   If_Result_2 := As_Entity_2;
end if;
--# expr-done 5587
                Singleton_1 := Create_Internal_Entity_Compilation_Unit_Array (Items_Count => 1);
                Singleton_1.Items (1) := If_Result_2;
Scope_Result_5 := Singleton_1;
   Inc_Ref (Scope_Result_5);
   --# end
      Finalizer_Scope_2517;
   Result_Var_4 := Scope_Result_5;
else
   Result_Var_4 := No_Internal_Entity_Compilation_Unit_Array_Type;
end if;
      Inc_Ref (Result_Var_4);
   Result_Var_5 := Result_Var_4;
end if;
      Inc_Ref (Result_Var_5);
Concat_Result := Concat (Fld, Result_Var_5); 
--# expr-done 5588
Let_Result := Concat_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 5589
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1306;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Compilation_Unit_Array_Access,
                        As_Internal_Entity_Compilation_Unit_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1306;
               Finalizer_Scope_2512;
               Finalizer_Scope_2513;
               Finalizer_Scope_2514;
               Finalizer_Scope_2515;
               Finalizer_Scope_2516;
               Finalizer_Scope_2517;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Compilation_Unit_P_Imported_Units;
--# end
--# property-start CompilationUnit.unit_dependencies_helper /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18466
pragma Warnings (Off, "is not referenced");
 function Compilation_Unit_P_Unit_Dependencies_Helper
  (Node : Bare_Compilation_Unit
      ; Visited : Internal_Entity_Compilation_Unit_Array_Access
      ; To_Visit : Internal_Entity_Compilation_Unit_Array_Access
  )
   return Internal_Entity_Compilation_Unit_Array_Access
is
   Self : Bare_Compilation_Unit := Bare_Compilation_Unit (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind visited Visited
   --# bind to_visit To_Visit
   Property_Result : Internal_Entity_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1307 with Inline_Always;
            procedure Finalizer_Scope_2518 with Inline_Always;
      Block_Var : Internal_Entity_Compilation_Unit_Array_Access;
Block_Var_1 : Internal_Entity_Compilation_Unit_Array_Access;
Block_Var_2 : Internal_Entity_Compilation_Unit_Array_Access;
Concat_Result : Internal_Entity_Compilation_Unit_Array_Access;
Fld : Internal_Entity_Compilation_Unit_Array_Access;
Map_Result : Internal_Entity_Compilation_Unit_Array_Access;
Unique_Array : Internal_Entity_Compilation_Unit_Array_Access;
Cast_Expr : Internal_Entity_Compilation_Unit;
Cast_Result : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Compilation_Unit;
Cast_Result_1 : Internal_Entity;
Is_Equiv : Boolean;
Quantifier_Result : Boolean;
Not_Val : Boolean;
Map_Result_1 : Internal_Entity_Compilation_Unit_Array_Access;
Fld_1 : Internal_Entity_Compilation_Unit_Array_Access;
Len : Integer;
Comp_Result : Boolean;
If_Result : Internal_Entity_Compilation_Unit_Array_Access;
Let_Result : Internal_Entity_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1307 is
            begin
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Unique_Array);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Block_Var_2);
                     Dec_Ref (Fld_1);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1307;
            procedure Finalizer_Scope_2518 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_2518;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5601 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5590 '<Concat at ast.py:18475>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18475
Concat_Result := Concat (Visited, To_Visit); 
--# expr-done 5590
Block_Var := Concat_Result; Inc_Ref (Block_Var);
--# bind now_visited Block_Var
--# expr-start 5593 '<unique at ast.py:18476>' Unique_Array /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18476
--# expr-start 5592 '<Mapcat at ast.py:18476>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18476
      declare
         Map_Result_Vec : Internal_Entity_Compilation_Unit_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Compilation_Unit_Array_Access := To_Visit;
      begin
         for Item_195 of
               Collection.Items
         loop
   --# scope-start
               --# bind c Item_195
      --# expr-start 5591 '<FieldAccess for imported_units at ast.py:18476>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18476
      if Item_195.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.imported_units
Fld := Libadalang.Implementation.Compilation_Unit_P_Imported_Units (Node => Item_195.Node);
--# end
--# expr-done 5591
         for Item_To_Append of
               Fld.Items
         loop
            Internal_Entity_Compilation_Unit_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_2518;
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Compilation_Unit_Array
           (Items_Count => Natural (Internal_Entity_Compilation_Unit_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Compilation_Unit_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Compilation_Unit_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Compilation_Unit_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 5592
Unique_Array := Make_Unique (Map_Result); 
--# expr-done 5593
Block_Var_1 := Unique_Array; Inc_Ref (Block_Var_1);
--# bind new_imports Block_Var_1
--# expr-start 5597 '<Filter at ast.py:18478>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18478
      declare
         Map_Result_1_Vec : Internal_Entity_Compilation_Unit_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Compilation_Unit_Array_Access := Block_Var_1;
      begin
         for Item_196 of
               Collection.Items
         loop
   --# scope-start
               --# bind c Item_196
         --# expr-start 5596 '<Not at ast.py:18478>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18478
--# expr-start 5595 '<Contains at ast.py:18478>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18478
      Quantifier_Result := False;
      declare
         Collection : constant Internal_Entity_Compilation_Unit_Array_Access := Block_Var;
      begin
         for Item_197 of
               Collection.Items
         loop
   --# scope-start
               --# bind item Item_197
      --# expr-start 5594 '<Eq at ???>' Is_Equiv None
Cast_Expr := Item_197; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Cast_Expr_1 := Item_196; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Is_Equiv := Equivalent (Cast_Result, Cast_Result_1); 
--# expr-done 5594
      Quantifier_Result := Is_Equiv;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
--# expr-done 5595
Not_Val := not (Quantifier_Result); 
--# expr-done 5596
         if Not_Val then
         declare
            Item_To_Append : constant Internal_Entity_Compilation_Unit := Item_196;
         begin
            Internal_Entity_Compilation_Unit_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result_1 := Create_Internal_Entity_Compilation_Unit_Array
           (Items_Count => Natural (Internal_Entity_Compilation_Unit_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Entity_Compilation_Unit_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Entity_Compilation_Unit_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Entity_Compilation_Unit_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 5597
Block_Var_2 := Map_Result_1; Inc_Ref (Block_Var_2);
--# bind to_visit_next Block_Var_2
--# expr-start 5600 '<If at ast.py:18480>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18480
--# expr-start 5598 '<length at ast.py:18481>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18481
Len := Length (Block_Var_2); 
--# expr-done 5598
Comp_Result := Len > 0; 
if Comp_Result then
   --# expr-start 5599 '<FieldAccess for unit_dependencies_helper at ast.py:18482>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18482
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.unit_dependencies_helper
Fld_1 := Libadalang.Implementation.Compilation_Unit_P_Unit_Dependencies_Helper (Node => Self, Visited => Block_Var, To_Visit => Block_Var_2);
--# end
--# expr-done 5599
   If_Result := Fld_1;
else
   If_Result := Block_Var;
end if;
      Inc_Ref (If_Result);
--# expr-done 5600
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 5601
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1307;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1307;
               Finalizer_Scope_2518;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Compilation_Unit_P_Unit_Dependencies_Helper;
--# end
--# property-start CompilationUnit.unit_dependencies /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18488
pragma Warnings (Off, "is not referenced");
 function Compilation_Unit_P_Unit_Dependencies
  (Node : Bare_Compilation_Unit
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Compilation_Unit_Array_Access
is
   Self : Bare_Compilation_Unit := Bare_Compilation_Unit (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Compilation_Unit :=
     Internal_Entity_Compilation_Unit'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1308 with Inline_Always;
      Singleton : Internal_Entity_Compilation_Unit_Array_Access;
Fld : Internal_Entity_Compilation_Unit_Array_Access;
Unique_Array : Internal_Entity_Compilation_Unit_Array_Access;
Fld_1 : Bare_Compilation_Unit;
Cast_Result : Bare_Compilation_Unit;
Is_Equal : Boolean;
Not_Val : Boolean;
Map_Result : Internal_Entity_Compilation_Unit_Array_Access;
Let_Result : Internal_Entity_Compilation_Unit_Array_Access;
            procedure Finalizer_Scope_1308 is
            begin
                     Dec_Ref (Singleton);
                     Dec_Ref (Fld);
                     Dec_Ref (Unique_Array);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1308;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Compilation_Unit_Compilation_Unit_P_Unit_Dependencies,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Compilation_Unit,
                                As_Bare_Compilation_Unit => Self);
               Mmz_K.Items (2) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Compilation_Unit_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 5608 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5607 '<Filter at ast.py:18495>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18495
   --# expr-start 5603 '<unique at ast.py:18495>' Unique_Array /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18495
--# expr-start 5602 '<FieldAccess for unit_dependencies_helper at ast.py:18495>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18495
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
                Singleton := Create_Internal_Entity_Compilation_Unit_Array (Items_Count => 1);
                Singleton.Items (1) := Ent;
--# property-call-start CompilationUnit.unit_dependencies_helper
Fld := Libadalang.Implementation.Compilation_Unit_P_Unit_Dependencies_Helper (Node => Self, Visited => No_Internal_Entity_Compilation_Unit_Array_Type, To_Visit => Singleton);
--# end
--# expr-done 5602
Unique_Array := Make_Unique (Fld); 
--# expr-done 5603
      declare
         Map_Result_Vec : Internal_Entity_Compilation_Unit_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Compilation_Unit_Array_Access := Unique_Array;
      begin
         for Item_198 of
               Collection.Items
         loop
   --# scope-start
               --# bind u Item_198
         --# expr-start 5606 '<Not at ast.py:18500>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18500
--# expr-start 5605 '<Eq at ast.py:18500>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18500
--# expr-start 5604 '<FieldAccess for node at ast.py:18500>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18500
Fld_1 := Item_198.Node;
--# expr-done 5604
      Cast_Result := Fld_1;
Is_Equal := Cast_Result = Self; 
--# expr-done 5605
Not_Val := not (Is_Equal); 
--# expr-done 5606
         if Not_Val then
         declare
            Item_To_Append : constant Internal_Entity_Compilation_Unit := Item_198;
         begin
            Internal_Entity_Compilation_Unit_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Compilation_Unit_Array
           (Items_Count => Natural (Internal_Entity_Compilation_Unit_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Compilation_Unit_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Compilation_Unit_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Compilation_Unit_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 5607
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 5608
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1308;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Compilation_Unit_Array_Access,
                        As_Internal_Entity_Compilation_Unit_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1308;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Compilation_Unit_P_Unit_Dependencies;
--# end
--# property-start CompilationUnit.decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18505
pragma Warnings (Off, "is not referenced");
 function Compilation_Unit_P_Decl
  (Node : Bare_Compilation_Unit
  )
   return Bare_Basic_Decl
is
   Self : Bare_Compilation_Unit := Bare_Compilation_Unit (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Basic_Decl;
      Match_0 : Bare_Library_Item;
Match_1 : Bare_Subunit;
Match_2 : Bare_Ada_Node;
Fld : Bare_Ada_Node;
Fld_1 : Bare_Basic_Decl;
Fld_2 : Bare_Body_Node;
Match_Prefix : Bare_Ada_Node;
Cast_Result : Bare_Library_Item;
Let_Result : Bare_Basic_Decl;
Scope_Result : Bare_Basic_Decl;
Cast_Result_1 : Bare_Subunit;
Cast_Result_2 : Bare_Basic_Decl;
Let_Result_1 : Bare_Basic_Decl;
Scope_Result_1 : Bare_Basic_Decl;
Cast_Result_3 : Bare_Ada_Node;
Let_Result_2 : Bare_Basic_Decl;
Scope_Result_2 : Bare_Basic_Decl;
Match_Result : Bare_Basic_Decl;
Let_Result_3 : Bare_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5613 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 5612 '<Match at ast.py:18509>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18509
--# expr-start 5609 '<FieldAccess for body at ast.py:18509>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18509
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Compilation_Unit_F_Body;
--# expr-done 5609
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld; 
case Ada_Ada_Node (Match_Prefix.Kind) is
         when Ada_Library_Item =>
   --# scope-start
--# scope-start
      Cast_Result := Match_Prefix;
Match_0 := Cast_Result; 
--# bind li Match_0
--# expr-start 5610 '<FieldAccess for item at ast.py:18510>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18510
      if Match_0 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Match_0.Library_Item_F_Item;
--# expr-done 5610
Let_Result := Fld_1; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Subunit =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Match_Prefix;
Match_1 := Cast_Result_1; 
--# bind su Match_1
--# expr-start 5611 '<FieldAccess for body at ast.py:18511>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18511
      if Match_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Match_1.Subunit_F_Body;
--# expr-done 5611
      Cast_Result_2 := Fld_2;
Let_Result_1 := Cast_Result_2; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abort_Absent .. Ada_Iter_Type_Of | Ada_Limited_Absent .. Ada_Subp_Kind_Procedure | Ada_Synchronized_Absent .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Match_Prefix;
Match_2 := Cast_Result_3; 
--# bind _ Match_2
Let_Result_2 := No_Bare_Ada_Node; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 5612
Let_Result_3 := Match_Result; 
--# end
--# expr-done 5613
      Property_Result := Let_Result_3;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Compilation_Unit_P_Decl;
--# end
--# property-start CompilationUnit.is_preelaborable_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18517
pragma Warnings (Off, "is not referenced");
 function Compilation_Unit_P_Is_Preelaborable_Impl
  (Node : Bare_Compilation_Unit
      ; From_Body : Boolean
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Compilation_Unit := Bare_Compilation_Unit (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Compilation_Unit :=
     Internal_Entity_Compilation_Unit'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind from_body From_Body
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Match_0 : Internal_Entity_Subunit;
Match_1 : Internal_Entity_Library_Item;
Match_2 : Internal_Entity;
Match_0_1 : Internal_Entity_Subp_Decl;
Match_1_1 : Internal_Entity_Generic_Subp_Decl;
Match_2_1 : Internal_Entity_Subp_Body;
Match_3 : Internal_Entity_Package_Decl;
Match_4 : Internal_Entity_Generic_Package_Decl;
Match_5 : Internal_Entity_Package_Body;
Match_6 : Internal_Entity_Basic_Decl;
Var_Expr : Internal_Entity_Basic_Decl;
Fld : Internal_Entity;
Fld_1 : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity;
Fld_3 : Internal_Entity;
Cast_Result : Internal_Entity_Compilation_Unit;
Fld_4 : Boolean;
Fld_5 : Internal_Entity_Basic_Decl;
Fld_6 : Boolean;
Fld_7 : Internal_Entity_Basic_Decl;
Fld_8 : Boolean;
Fld_9 : Boolean;
Fld_10 : Internal_Entity_Basic_Decl;
Fld_11 : Internal_Entity;
Fld_12 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Compilation_Unit;
Fld_13 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result : Boolean;
Fld_14 : Boolean;
Fld_15 : Internal_Entity_Generic_Package_Internal;
Fld_16 : Boolean;
Fld_17 : Internal_Entity_Basic_Decl;
Fld_18 : Internal_Unit;
Fld_19 : Bare_Ada_Node;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result_1 : Internal_Entity;
Cast_Result_2 : Internal_Entity_Compilation_Unit;
Fld_20 : Boolean;
Match_Prefix : Internal_Entity_Basic_Decl;
Cast_Result_3 : Internal_Entity_Subp_Decl;
Let_Result : Boolean;
Scope_Result_1 : Boolean;
Cast_Result_4 : Internal_Entity_Generic_Subp_Decl;
Let_Result_1 : Boolean;
Scope_Result_2 : Boolean;
Cast_Result_5 : Internal_Entity_Subp_Body;
Let_Result_2 : Boolean;
Scope_Result_3 : Boolean;
Cast_Result_6 : Internal_Entity_Package_Decl;
Let_Result_3 : Boolean;
Scope_Result_4 : Boolean;
Cast_Result_7 : Internal_Entity_Generic_Package_Decl;
Let_Result_4 : Boolean;
Scope_Result_5 : Boolean;
Cast_Result_8 : Internal_Entity_Package_Body;
Let_Result_5 : Boolean;
Scope_Result_6 : Boolean;
Cast_Result_9 : Internal_Entity_Basic_Decl;
Let_Result_6 : Boolean;
Scope_Result_7 : Boolean;
Match_Result : Boolean;
Match_Prefix_1 : Internal_Entity;
Cast_Result_10 : Internal_Entity_Subunit;
Let_Result_7 : Boolean;
Scope_Result_8 : Boolean;
Cast_Result_11 : Internal_Entity_Library_Item;
Let_Result_8 : Boolean;
Scope_Result_9 : Boolean;
Cast_Result_12 : Internal_Entity;
Let_Result_9 : Boolean;
Scope_Result_10 : Boolean;
Match_Result_1 : Boolean;
Let_Result_10 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5641 '<Block at ???>' Let_Result_10 None
--# scope-start
--# expr-start 5640 '<Match at ast.py:18525>' Match_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18525
--# expr-start 5614 '<FieldAccess for body at ast.py:18525>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18525
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity (Node => Ent.Node.Compilation_Unit_F_Body, Info => Ent.Info);
--# expr-done 5614
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix_1 := Fld; 
case Ada_Ada_Node (Match_Prefix_1.Node.Kind) is
         when Ada_Subunit =>
   --# scope-start
--# scope-start
      Cast_Result_10 := Create_Internal_Entity_Subunit
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_0 := Cast_Result_10; 
--# bind su Match_0
--# expr-start 5619 '<FieldAccess for is_preelaborable_impl at ast.py:18529>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18529
--# expr-start 5618 '<Cast to CompilationUnit at ast.py:18529>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18529
--# expr-start 5617 '<parent at ast.py:18529>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18529
--# expr-start 5616 '<parent at ast.py:18529>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18529
--# expr-start 5615 '<FieldAccess for body_root at ast.py:18529>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18529
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Subunit.body_root
Fld_1 := Libadalang.Implementation.Subunit_P_Body_Root (Node => Match_0.Node);
--# end
--# expr-done 5615
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Parent (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# expr-done 5616
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Parent (Node => Fld_2.Node, E_Info => Fld_2.Info);
--# expr-done 5617
   if Fld_3.Node = null
      or else Fld_3.Node.Kind in Ada_Compilation_Unit_Range
   then
      Cast_Result := Create_Internal_Entity_Compilation_Unit
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 5618
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.is_preelaborable_impl
Fld_4 := Libadalang.Implementation.Compilation_Unit_P_Is_Preelaborable_Impl (Node => Cast_Result.Node, From_Body => True, Imprecise_Fallback => Imprecise_Fallback, E_Info => Cast_Result.Info);
--# end
--# expr-done 5619
Let_Result_7 := Fld_4; 
--# end
Scope_Result_8 := Let_Result_7;
   --# end
            Match_Result_1 := Scope_Result_8; 
         when Ada_Library_Item =>
   --# scope-start
--# scope-start
      Cast_Result_11 := Create_Internal_Entity_Library_Item
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_1 := Cast_Result_11; 
--# bind li Match_1
--# expr-start 5639 '<Match at ast.py:18533>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18533
--# expr-start 5620 '<FieldAccess for item at ast.py:18533>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18533
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Basic_Decl (Node => Match_1.Node.Library_Item_F_Item, Info => Match_1.Info);
--# expr-done 5620
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld_5; 
case Ada_Basic_Decl (Match_Prefix.Node.Kind) is
         when Ada_Subp_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Subp_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0_1 := Cast_Result_3; 
--# bind subp_decl Match_0_1
--# expr-start 5621 '<FieldAccess for does_aspects_make_preelaborable at ast.py:18535>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18535
      if Match_0_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.does_aspects_make_preelaborable
Fld_6 := Libadalang.Implementation.Basic_Decl_P_Does_Aspects_Make_Preelaborable (Node => Match_0_1.Node, From_Body => From_Body, E_Info => Match_0_1.Info);
--# end
--# expr-done 5621
Let_Result := Fld_6; 
--# end
Scope_Result_1 := Let_Result;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Generic_Subp_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Generic_Subp_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1_1 := Cast_Result_4; 
--# bind gen_subp_decl Match_1_1
--# expr-start 5623 '<FieldAccess for does_aspects_make_preelaborable at ast.py:18538>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18538
--# expr-start 5622 '<FieldAccess for decl at ast.py:18538>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18538
      if Match_1_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericDecl.decl'
Fld_7 := Libadalang.Implementation.Dispatcher_Generic_Decl_P_Decl (Node => Match_1_1.Node, E_Info => Match_1_1.Info);
--# end
--# expr-done 5622
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.does_aspects_make_preelaborable
Fld_8 := Libadalang.Implementation.Basic_Decl_P_Does_Aspects_Make_Preelaborable (Node => Fld_7.Node, From_Body => From_Body, E_Info => Fld_7.Info);
--# end
--# expr-done 5623
Let_Result_1 := Fld_8; 
--# end
Scope_Result_2 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Subp_Body =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Create_Internal_Entity_Subp_Body
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2_1 := Cast_Result_5; 
--# bind subp_body Match_2_1
--# expr-start 5624 '<FieldAccess for does_aspects_make_preelaborable at ast.py:18544>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18544
      if Match_2_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.does_aspects_make_preelaborable
Fld_9 := Libadalang.Implementation.Basic_Decl_P_Does_Aspects_Make_Preelaborable (Node => Match_2_1.Node, From_Body => False, E_Info => Match_2_1.Info);
--# end
--# expr-done 5624
if Fld_9 then
   If_Result := True;
else
--# expr-start 5625 '<FieldAccess for decl_part at ast.py:18548>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18548
      if Match_2_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.decl_part
Fld_10 := Libadalang.Implementation.Body_Node_P_Decl_Part (Node => Match_2_1.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Match_2_1.Info);
--# end
--# expr-done 5625
Var_Expr := Fld_10;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind dp Var_Expr
--# expr-start 5629 '<FieldAccess for is_preelaborable_impl at ast.py:18549>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18549
--# expr-start 5628 '<Cast to CompilationUnit at ast.py:18549>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18549
--# expr-start 5627 '<parent at ast.py:18549>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18549
--# expr-start 5626 '<parent at ast.py:18549>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18549
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Libadalang.Implementation.Parent (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# expr-done 5626
      if Fld_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Libadalang.Implementation.Parent (Node => Fld_11.Node, E_Info => Fld_11.Info);
--# expr-done 5627
   if Fld_12.Node = null
      or else Fld_12.Node.Kind in Ada_Compilation_Unit_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Compilation_Unit
        (Node => Fld_12.Node,
         Info => Fld_12.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 5628
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.is_preelaborable_impl
Fld_13 := Libadalang.Implementation.Compilation_Unit_P_Is_Preelaborable_Impl (Node => Cast_Result_1.Node, From_Body => True, Imprecise_Fallback => Imprecise_Fallback, E_Info => Cast_Result_1.Info);
--# end
--# expr-done 5629
Scope_Result := Fld_13;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
   If_Result := Result_Var;
end if;
Let_Result_2 := If_Result; 
--# end
Scope_Result_3 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_3; 
         when Ada_Package_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_6 := Create_Internal_Entity_Package_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_6; 
--# bind pkg_decl Match_3
--# expr-start 5630 '<FieldAccess for does_aspects_make_preelaborable at ast.py:18555>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18555
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.does_aspects_make_preelaborable
Fld_14 := Libadalang.Implementation.Basic_Decl_P_Does_Aspects_Make_Preelaborable (Node => Match_3.Node, From_Body => From_Body, E_Info => Match_3.Info);
--# end
--# expr-done 5630
Let_Result_3 := Fld_14; 
--# end
Scope_Result_4 := Let_Result_3;
   --# end
            Match_Result := Scope_Result_4; 
         when Ada_Generic_Package_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_7 := Create_Internal_Entity_Generic_Package_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_4 := Cast_Result_7; 
--# bind gen_pkg_decl Match_4
--# expr-start 5632 '<FieldAccess for does_aspects_make_preelaborable at ast.py:18558>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18558
--# expr-start 5631 '<FieldAccess for package_decl at ast.py:18558>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18558
      if Match_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Create_Internal_Entity_Generic_Package_Internal (Node => Match_4.Node.Generic_Package_Decl_F_Package_Decl, Info => Match_4.Info);
--# expr-done 5631
      if Fld_15.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.does_aspects_make_preelaborable
Fld_16 := Libadalang.Implementation.Basic_Decl_P_Does_Aspects_Make_Preelaborable (Node => Fld_15.Node, From_Body => From_Body, E_Info => Fld_15.Info);
--# end
--# expr-done 5632
Let_Result_4 := Fld_16; 
--# end
Scope_Result_5 := Let_Result_4;
   --# end
            Match_Result := Scope_Result_5; 
         when Ada_Package_Body =>
   --# scope-start
--# scope-start
      Cast_Result_8 := Create_Internal_Entity_Package_Body
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_5 := Cast_Result_8; 
--# bind pkg_body Match_5
--# expr-start 5638 '<FieldAccess for is_preelaborable_impl at ast.py:18564>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18564
--# expr-start 5637 '<Cast to CompilationUnit at ast.py:18564>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18564
--# expr-start 5636 '<as_bare_entity at ast.py:18564>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18564
--# expr-start 5635 '<FieldAccess for root at ast.py:18564>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18564
--# expr-start 5634 '<FieldAccess for unit at ast.py:18564>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18564
--# expr-start 5633 '<FieldAccess for decl_part at ast.py:18564>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18564
      if Match_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.decl_part
Fld_17 := Libadalang.Implementation.Body_Node_P_Decl_Part (Node => Match_5.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Match_5.Info);
--# end
--# expr-done 5633
      if Fld_17.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Libadalang.Implementation.Unit (Node => Fld_17.Node);
--# expr-done 5634
      if Fld_18 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Fld_18.Ast_Root;
--# expr-done 5635
Is_Null := Fld_19 = null; 
if Is_Null then
   If_Result_1 := No_Entity;
else
As_Entity := (Info => No_Entity_Info, Node => Fld_19); 
   If_Result_1 := As_Entity;
end if;
--# expr-done 5636
   if If_Result_1.Node = null
      or else If_Result_1.Node.Kind in Ada_Compilation_Unit_Range
   then
      Cast_Result_2 := Create_Internal_Entity_Compilation_Unit
        (Node => If_Result_1.Node,
         Info => If_Result_1.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 5637
      if Cast_Result_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.is_preelaborable_impl
Fld_20 := Libadalang.Implementation.Compilation_Unit_P_Is_Preelaborable_Impl (Node => Cast_Result_2.Node, From_Body => True, Imprecise_Fallback => Imprecise_Fallback, E_Info => Cast_Result_2.Info);
--# end
--# expr-done 5638
Let_Result_5 := Fld_20; 
--# end
Scope_Result_6 := Let_Result_5;
   --# end
            Match_Result := Scope_Result_6; 
         when Ada_Abstract_State_Decl .. Ada_Generic_Package_Internal | Ada_Discrete_Base_Subtype_Decl .. Ada_Concrete_Formal_Subp_Decl | Ada_Entry_Decl .. Ada_Null_Subp_Decl | Ada_Subp_Renaming_Decl .. Ada_Entry_Body | Ada_Protected_Body .. Ada_For_Loop_Var_Decl | Ada_Generic_Package_Instantiation .. Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_9 := Create_Internal_Entity_Basic_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_6 := Cast_Result_9; 
--# bind _ Match_6
Let_Result_6 := False; 
--# end
Scope_Result_7 := Let_Result_6;
   --# end
            Match_Result := Scope_Result_7; 
end case;
--# expr-done 5639
Let_Result_8 := Match_Result; 
--# end
Scope_Result_9 := Let_Result_8;
   --# end
            Match_Result_1 := Scope_Result_9; 
         when Ada_Abort_Absent .. Ada_Iter_Type_Of | Ada_Limited_Absent .. Ada_Subp_Kind_Procedure | Ada_Synchronized_Absent .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_12 := Create_Internal_Entity
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_2 := Cast_Result_12; 
--# bind _ Match_2
Let_Result_9 := False; 
--# end
Scope_Result_10 := Let_Result_9;
   --# end
            Match_Result_1 := Scope_Result_10; 
end case;
--# expr-done 5640
Let_Result_10 := Match_Result_1; 
--# end
--# expr-done 5641
      Property_Result := Let_Result_10;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Compilation_Unit_P_Is_Preelaborable_Impl;
--# end
--# property-start CompilationUnit.is_preelaborable /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18577
pragma Warnings (Off, "is not referenced");
 function Compilation_Unit_P_Is_Preelaborable
  (Node : Bare_Compilation_Unit
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Compilation_Unit := Bare_Compilation_Unit (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Compilation_Unit :=
     Internal_Entity_Compilation_Unit'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Fld : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5643 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5642 '<FieldAccess for is_preelaborable_impl at ast.py:18581>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18581
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.is_preelaborable_impl
Fld := Libadalang.Implementation.Compilation_Unit_P_Is_Preelaborable_Impl (Node => Ent.Node, From_Body => False, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 5642
Let_Result := Fld; 
--# end
--# expr-done 5643
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Compilation_Unit_P_Is_Preelaborable;
--# end
--# property-start CompilationUnit.self_restrictions /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18584
pragma Warnings (Off, "is not referenced");
 function Compilation_Unit_P_Self_Restrictions
  (Node : Bare_Compilation_Unit
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Compilation_Unit := Bare_Compilation_Unit (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1312 with Inline_Always;
      Var_Expr : Internal_Entity_Base_Assoc;
Var_Expr_1 : Internal_Entity_Pragma_Node;
As_Entity : Internal_Entity_Compilation_Unit;
Is_Null : Boolean;
If_Result : Internal_Entity_Compilation_Unit;
Fld : Internal_Entity_Ada_Node_List;
Fld_1 : Bare_Ada_Node_List;
Fld_2 : Internal_Entity_Info;
As_Entity_1 : Internal_Entity;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Pragma_Node;
Fld_3 : Internal_Entity_Base_Assoc_List;
Fld_4 : Bare_Base_Assoc_List;
Fld_5 : Internal_Entity_Info;
Get_Result : Bare_Base_Assoc;
As_Entity_2 : Internal_Entity_Base_Assoc;
Is_Null_2 : Boolean;
If_Result_2 : Internal_Entity_Base_Assoc;
Fld_6 : Internal_Entity_Expr;
Scope_Result : Internal_Entity_Expr;
Result_Var : Internal_Entity_Expr;
Cast_Result_1 : Internal_Entity_Base_Id;
Fld_7 : Symbol_Type;
Item_199 : Internal_Entity;
Bare_Item_199 : Bare_Ada_Node;
Cast_Expr_1 : Internal_Entity;
Cast_Result_2 : Internal_Entity_Pragma_Node;
Fld_8 : Internal_Entity_Identifier;
Fld_9 : Symbol_Type;
Sym : Symbol_Type;
Is_Equal : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
Map_Result : Symbol_Type_Array_Access;
Let_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1312 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1312;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5657 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5656 '<FilterMap at ast.py:18590>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18590
   --# expr-start 5645 '<FieldAccess for prelude at ast.py:18590>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18590
--# expr-start 5644 '<as_bare_entity at ast.py:18590>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18590
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity_Compilation_Unit;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 5644
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Ada_Node_List (Node => If_Result.Node.Compilation_Unit_F_Prelude, Info => If_Result.Info);
--# expr-done 5645
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Symbol_Type_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Symbol_Type_Array (0);
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_1;
      begin
         for Untyped_Item_199 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_199 := Untyped_Item_199; 
Is_Null_1 := Bare_Item_199 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity;
else
Fld_2 := Fld.Info;
As_Entity_1 := (Info => Fld_2, Node => Bare_Item_199); 
   If_Result_1 := As_Entity_1;
end if;
                  Item_199 := If_Result_1; 
   --# scope-start
               --# bind n Item_199
--# expr-start 5646 '<Cast to Pragma at ast.py:18594>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18594
Cast_Expr_1 := Item_199; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Pragma_Node_Range
   then
      Cast_Result_2 := Create_Internal_Entity_Pragma_Node
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_2 := No_Entity_Pragma_Node;
   end if;
--# expr-done 5646
Var_Expr_1 := Cast_Result_2;
if Var_Expr_1 /= No_Entity_Pragma_Node then
   --# scope-start
   --# bind p Var_Expr_1
--# expr-start 5650 '<Eq at ast.py:18595>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18595
--# expr-start 5648 '<FieldAccess for name_symbol at ast.py:18595>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18595
--# expr-start 5647 '<FieldAccess for id at ast.py:18595>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18595
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Identifier (Node => Var_Expr_1.Node.Pragma_Node_F_Id, Info => Var_Expr_1.Info);
--# expr-done 5647
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_9 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_8.Node);
--# end
--# expr-done 5648
--# expr-start 5649 '<Symbol Restrictions at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Restrictions); 
--# expr-done 5649
Is_Equal := Fld_9 = Sym; 
--# expr-done 5650
Scope_Result_1 := Is_Equal;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
         if Result_Var_1 then
      --# expr-start 5655 '<FieldAccess for name_symbol at ast.py:18592>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18592
--# expr-start 5654 '<Cast to BaseId at ast.py:18592>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18592
--# expr-start 5652 '<FieldAccess for args at ast.py:18592>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18592
--# expr-start 5651 '<Cast to Pragma at ast.py:18592>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18592
Cast_Expr := Item_199; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Pragma_Node_Range
   then
      Cast_Result := Create_Internal_Entity_Pragma_Node
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Pragma_Node;
   end if;
--# expr-done 5651
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Base_Assoc_List (Node => Cast_Result.Node.Pragma_Node_F_Args, Info => Cast_Result.Info);
--# expr-done 5652
Fld_4 := Fld_3.Node;
Get_Result := Get (Fld_4, 0, True); 
Is_Null_2 := Get_Result = null; 
if Is_Null_2 then
   If_Result_2 := No_Entity_Base_Assoc;
else
Fld_5 := Fld_3.Info;
As_Entity_2 := (Info => Fld_5, Node => Get_Result); 
   If_Result_2 := As_Entity_2;
end if;
Var_Expr := If_Result_2;
if Var_Expr /= No_Entity_Base_Assoc then
   --# scope-start
--# expr-start 5653 '<FieldAccess for assoc_expr at ast.py:18592>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18592
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseAssoc.assoc_expr'
Fld_6 := Libadalang.Implementation.Dispatcher_Base_Assoc_P_Assoc_Expr (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 5653
Scope_Result := Fld_6;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Expr;
end if;
   if Result_Var.Node = null
      or else Result_Var.Node.Kind in Ada_Base_Id
   then
      Cast_Result_1 := Create_Internal_Entity_Base_Id
        (Node => Result_Var.Node,
         Info => Result_Var.Info);
   else
         Cast_Result_1 := No_Entity_Base_Id;
   end if;
--# expr-done 5654
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_7 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Cast_Result_1.Node);
--# end
--# expr-done 5655
         declare
            Item_To_Append : constant Symbol_Type := Fld_7;
         begin
            Symbol_Type_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Symbol_Type_Array
           (Items_Count => Natural (Symbol_Type_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Symbol_Type_Vectors.Get
              (Map_Result_Vec,
               I + Symbol_Type_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Symbol_Type_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 5656
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 5657
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1312;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1312;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Compilation_Unit_P_Self_Restrictions;
--# end
--# property-start CompilationUnit.other_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18600
pragma Warnings (Off, "is not referenced");
 function Compilation_Unit_P_Other_Part
  (Node : Bare_Compilation_Unit
  )
   return Internal_Entity_Compilation_Unit
is
   Self : Bare_Compilation_Unit := Bare_Compilation_Unit (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Compilation_Unit;
            procedure Finalizer_Scope_1313 with Inline_Always;
      Block_Var : Analysis_Unit_Kind;
Var_Expr : Bare_Compilation_Unit;
Fld : Analysis_Unit_Kind;
Is_Equal : Boolean;
If_Result : Analysis_Unit_Kind;
Fld_1 : Symbol_Type_Array_Access;
Fld_2 : Bare_Compilation_Unit;
As_Entity : Internal_Entity_Compilation_Unit;
Is_Null : Boolean;
If_Result_1 : Internal_Entity_Compilation_Unit;
Scope_Result : Internal_Entity_Compilation_Unit;
Result_Var : Internal_Entity_Compilation_Unit;
Let_Result : Internal_Entity_Compilation_Unit;
            procedure Finalizer_Scope_1313 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1313;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5664 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5660 '<If at ast.py:18605>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18605
--# expr-start 5659 '<Eq at ast.py:18606>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18606
--# expr-start 5658 '<FieldAccess for unit_kind at ast.py:18606>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18606
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.unit_kind
Fld := Libadalang.Implementation.Compilation_Unit_P_Unit_Kind (Node => Self);
--# end
--# expr-done 5658
Is_Equal := Fld = Unit_Specification; 
--# expr-done 5659
if Is_Equal then
   If_Result := Unit_Body;
else
   If_Result := Unit_Specification;
end if;
--# expr-done 5660
Block_Var := If_Result; 
--# bind other_kind Block_Var
--# expr-start 5662 '<FieldAccess for designated_compilation_unit at ast.py:18610>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18610
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 5661 '<FieldAccess for syntactic_fully_qualified_name at ast.py:18611>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18611
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.syntactic_fully_qualified_name
Fld_1 := Libadalang.Implementation.Compilation_Unit_P_Syntactic_Fully_Qualified_Name (Node => Self);
--# end
--# expr-done 5661
--# property-call-start AdaNode.designated_compilation_unit
Fld_2 := Libadalang.Implementation.Ada_Node_P_Designated_Compilation_Unit (Node => Self, Name => Fld_1, Kind => Block_Var, Not_Found_Is_Error => False);
--# end
--# expr-done 5662
Var_Expr := Fld_2;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 5663 '<as_bare_entity at ast.py:18610>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18610
Is_Null := Var_Expr = null; 
if Is_Null then
   If_Result_1 := No_Entity_Compilation_Unit;
else
As_Entity := (Info => No_Entity_Info, Node => Var_Expr); 
   If_Result_1 := As_Entity;
end if;
--# expr-done 5663
Scope_Result := If_Result_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Compilation_Unit;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 5664
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1313;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1313;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Compilation_Unit_P_Other_Part;
--# end
--# property-start CompilationUnit.has_restriction /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18617
pragma Warnings (Off, "is not referenced");
 function Compilation_Unit_P_Has_Restriction
  (Node : Bare_Compilation_Unit
      ; Name : Symbol_Type
  )
   return Boolean
is
   Self : Bare_Compilation_Unit := Bare_Compilation_Unit (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind name Name
   Property_Result : Boolean;
            procedure Finalizer_Scope_2540 with Inline_Always;
            procedure Finalizer_Scope_2542 with Inline_Always;
      Match_0 : Bare_Library_Item;
Match_1 : Bare_Subunit;
Match_2 : Bare_Ada_Node;
Var_Expr : Internal_Entity_Compilation_Unit;
Var_Expr_1 : Internal_Entity_Compilation_Unit;
Sym : Symbol_Type;
Is_Equal : Boolean;
Fld : Bare_Ada_Node;
Fld_1 : Symbol_Type_Array_Access;
Is_Equal_1 : Boolean;
Quantifier_Result : Boolean;
Fld_2 : Internal_Entity_Compilation_Unit;
Fld_3 : Symbol_Type_Array_Access;
Scope_Result : Symbol_Type_Array_Access;
Result_Var : Symbol_Type_Array_Access;
Is_Equal_2 : Boolean;
Quantifier_Result_1 : Boolean;
If_Result : Boolean;
Fld_4 : Internal_Entity_Compilation_Unit;
Fld_5 : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
Match_Prefix : Bare_Ada_Node;
Cast_Result : Bare_Library_Item;
Let_Result : Boolean;
Scope_Result_2 : Boolean;
Cast_Result_1 : Bare_Subunit;
Let_Result_1 : Boolean;
Scope_Result_3 : Boolean;
Cast_Result_2 : Bare_Ada_Node;
Let_Result_2 : Boolean;
Scope_Result_4 : Boolean;
Match_Result : Boolean;
If_Result_1 : Boolean;
Let_Result_3 : Boolean;
            procedure Finalizer_Scope_2540 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_2540;
            procedure Finalizer_Scope_2542 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_2542;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5679 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 5678 '<Cond at ast.py:18626>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18626
--# expr-start 5666 '<Eq at ast.py:18627>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18627
--# expr-start 5665 '<Symbol No_Elaboration_Code at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_No_Elaboration_Code); 
--# expr-done 5665
Is_Equal := Name = Sym; 
--# expr-done 5666
if Is_Equal then
   --# expr-start 5677 '<Match at ast.py:18629>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18629
--# expr-start 5667 '<FieldAccess for body at ast.py:18629>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18629
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Compilation_Unit_F_Body;
--# expr-done 5667
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld; 
case Ada_Ada_Node (Match_Prefix.Kind) is
         when Ada_Library_Item =>
   --# scope-start
--# scope-start
      Cast_Result := Match_Prefix;
Match_0 := Cast_Result; 
--# bind _ Match_0
--# expr-start 5670 '<Contains at ast.py:18633>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18633
   --# expr-start 5668 '<FieldAccess for self_restrictions at ast.py:18633>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18633
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.self_restrictions
Fld_1 := Libadalang.Implementation.Compilation_Unit_P_Self_Restrictions (Node => Self);
--# end
--# expr-done 5668
      Quantifier_Result := False;
      declare
         Collection : constant Symbol_Type_Array_Access := Fld_1;
      begin
         for Item_200 of
               Collection.Items
         loop
   --# scope-start
               --# bind item Item_200
      --# expr-start 5669 '<Eq at ???>' Is_Equal_1 None
Is_Equal_1 := Item_200 = Name; 
--# expr-done 5669
      Quantifier_Result := Is_Equal_1;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
--# expr-done 5670
if Quantifier_Result then
   If_Result := True;
else
   --# expr-start 5674 '<Contains at ast.py:18634>' Quantifier_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18634
--# expr-start 5671 '<FieldAccess for other_part at ast.py:18634>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18634
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.other_part
Fld_2 := Libadalang.Implementation.Compilation_Unit_P_Other_Part (Node => Self);
--# end
--# expr-done 5671
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Compilation_Unit then
   --# scope-start
--# expr-start 5672 '<FieldAccess for self_restrictions at ast.py:18634>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18634
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.self_restrictions
Fld_3 := Libadalang.Implementation.Compilation_Unit_P_Self_Restrictions (Node => Var_Expr.Node);
--# end
--# expr-done 5672
Scope_Result := Fld_3;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2542;
   Result_Var := Scope_Result;
else
   Result_Var := No_Symbol_Type_Array_Type;
end if;
      Inc_Ref (Result_Var);
      Quantifier_Result_1 := False;
      declare
         Collection : constant Symbol_Type_Array_Access := Result_Var;
      begin
         for Item_201 of
               Collection.Items
         loop
   --# scope-start
               --# bind item Item_201
      --# expr-start 5673 '<Eq at ???>' Is_Equal_2 None
Is_Equal_2 := Item_201 = Name; 
--# expr-done 5673
      Quantifier_Result_1 := Is_Equal_2;
   --# end
               exit when Quantifier_Result_1;
         end loop;
      end;
--# expr-done 5674
   If_Result := Quantifier_Result_1;
end if;
Let_Result := If_Result; 
--# end
Scope_Result_2 := Let_Result;
   --# end
      Finalizer_Scope_2540;
            Match_Result := Scope_Result_2; 
         when Ada_Subunit =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Match_Prefix;
Match_1 := Cast_Result_1; 
--# bind su Match_1
--# expr-start 5675 '<FieldAccess for root_unit at ast.py:18638>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18638
      if Match_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Subunit.root_unit
Fld_4 := Libadalang.Implementation.Subunit_P_Root_Unit (Node => Match_1);
--# end
--# expr-done 5675
Var_Expr_1 := Fld_4;
if Var_Expr_1 /= No_Entity_Compilation_Unit then
   --# scope-start
--# expr-start 5676 '<FieldAccess for has_restriction at ast.py:18638>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18638
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.has_restriction
Fld_5 := Libadalang.Implementation.Compilation_Unit_P_Has_Restriction (Node => Var_Expr_1.Node, Name => Name);
--# end
--# expr-done 5676
Scope_Result_1 := Fld_5;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
Let_Result_1 := Result_Var_1; 
--# end
Scope_Result_3 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_3; 
         when Ada_Abort_Absent .. Ada_Iter_Type_Of | Ada_Limited_Absent .. Ada_Subp_Kind_Procedure | Ada_Synchronized_Absent .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Match_Prefix;
Match_2 := Cast_Result_2; 
--# bind _ Match_2
Let_Result_2 := raise Property_Error with "Unexpected CompilationUnit.f_body attribute"; 
--# end
Scope_Result_4 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_4; 
end case;
--# expr-done 5677
   If_Result_1 := Match_Result;
else
   If_Result_1 := raise Property_Error with "Unsupported restriction";
end if;
--# expr-done 5678
Let_Result_3 := If_Result_1; 
--# end
--# expr-done 5679
      Property_Result := Let_Result_3;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_2540;
               Finalizer_Scope_2542;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Compilation_Unit_P_Has_Restriction;
--# end
--# property-start CompilationUnit.is_text_io_child /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18649
pragma Warnings (Off, "is not referenced");
 function Compilation_Unit_P_Is_Text_Io_Child
  (Node : Bare_Compilation_Unit
  )
   return Boolean
is
   Self : Bare_Compilation_Unit := Bare_Compilation_Unit (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
            procedure Finalizer_Scope_1315 with Inline_Always;
      Block_Var : Symbol_Type_Array_Access;
Fld : Symbol_Type_Array_Access;
Len : Integer;
Is_Equal : Boolean;
Get_Result : Symbol_Type;
Sym : Symbol_Type;
Is_Equal_1 : Boolean;
If_Result : Boolean;
Get_Result_1 : Symbol_Type;
Sym_1 : Symbol_Type;
Is_Equal_2 : Boolean;
Get_Result_2 : Symbol_Type;
Sym_2 : Symbol_Type;
Is_Equal_3 : Boolean;
If_Result_1 : Boolean;
Get_Result_3 : Symbol_Type;
Sym_3 : Symbol_Type;
Is_Equal_4 : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_1315 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Block_Var);
            end Finalizer_Scope_1315;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5695 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5680 '<FieldAccess for syntactic_fully_qualified_name at ast.py:18654>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18654
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.syntactic_fully_qualified_name
Fld := Libadalang.Implementation.Compilation_Unit_P_Syntactic_Fully_Qualified_Name (Node => Self);
--# end
--# expr-done 5680
Block_Var := Fld; Inc_Ref (Block_Var);
--# bind name_parts Block_Var
--# expr-start 5682 '<Eq at ast.py:18656>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18656
--# expr-start 5681 '<length at ast.py:18656>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18656
Len := Length (Block_Var); 
--# expr-done 5681
Is_Equal := Len = 3; 
--# expr-done 5682
if Is_Equal then
   --# expr-start 5685 '<Eq at ast.py:18657>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18657
--# expr-start 5683 '<at at ast.py:18657>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18657
Get_Result := Get (Block_Var, 0, True); 
--# expr-done 5683
--# expr-start 5684 '<Symbol ada at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Ada_261); 
--# expr-done 5684
Is_Equal_1 := Get_Result = Sym; 
--# expr-done 5685
   If_Result := Is_Equal_1;
else
   If_Result := False;
end if;
if If_Result then
--# expr-start 5688 '<Eq at ast.py:18658>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18658
--# expr-start 5686 '<at at ast.py:18658>' Get_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18658
Get_Result_1 := Get (Block_Var, 1, True); 
--# expr-done 5686
--# expr-start 5687 '<Symbol text_io at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Text_Io); 
--# expr-done 5687
Is_Equal_2 := Get_Result_1 = Sym_1; 
--# expr-done 5688
if Is_Equal_2 then
   If_Result_1 := True;
else
   --# expr-start 5691 '<Eq at ast.py:18658>' Is_Equal_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18658
--# expr-start 5689 '<at at ast.py:18658>' Get_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18658
Get_Result_2 := Get (Block_Var, 1, True); 
--# expr-done 5689
--# expr-start 5690 '<Symbol wide_text_io at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Wide_Text_Io); 
--# expr-done 5690
Is_Equal_3 := Get_Result_2 = Sym_2; 
--# expr-done 5691
   If_Result_1 := Is_Equal_3;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
   --# expr-start 5694 '<Eq at ast.py:18658>' Is_Equal_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18658
--# expr-start 5692 '<at at ast.py:18658>' Get_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18658
Get_Result_3 := Get (Block_Var, 1, True); 
--# expr-done 5692
--# expr-start 5693 '<Symbol wide_wide_text_io at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Wide_Wide_Text_Io); 
--# expr-done 5693
Is_Equal_4 := Get_Result_3 = Sym_3; 
--# expr-done 5694
   If_Result_2 := Is_Equal_4;
end if;
   If_Result_3 := If_Result_2;
else
   If_Result_3 := False;
end if;
Let_Result := If_Result_3; 
--# end
--# expr-done 5695
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1315;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1315;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Compilation_Unit_P_Is_Text_Io_Child;
--# end
--# property-start CompilationUnit.local_config_pragmas /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18677
pragma Warnings (Off, "is not referenced");
 function Compilation_Unit_P_Local_Config_Pragmas
  (Node : Bare_Compilation_Unit
  )
   return Bare_Pragma_Node_Array_Access
is
   Self : Bare_Compilation_Unit := Bare_Compilation_Unit (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Pragma_Node_Array_Access;
            procedure Finalizer_Scope_1317 with Inline_Always;
      Fld : Bare_Ada_Node_List;
Cast_Expr : Bare_Ada_Node;
Cast_Result : Bare_Pragma_Node;
Item_202 : Bare_Ada_Node;
Is_A : Boolean;
Map_Result : Bare_Pragma_Node_Array_Access;
Let_Result : Bare_Pragma_Node_Array_Access;
            procedure Finalizer_Scope_1317 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1317;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5700 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5699 '<FilterMap at ast.py:18682>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18682
   --# expr-start 5696 '<FieldAccess for prelude at ast.py:18682>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18682
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Compilation_Unit_F_Prelude;
--# expr-done 5696
      declare
         Map_Result_Vec : Bare_Pragma_Node_Vectors.Vector;
      begin
      if Fld = null then
      Map_Result := Create_Bare_Pragma_Node_Array (0);
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld;
      begin
         for Untyped_Item_202 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_202 := Untyped_Item_202; 
   --# scope-start
               --# bind n Item_202
         --# expr-start 5697 '<IsA Pragma at ast.py:18684>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18684
Is_A := Item_202 /= null 
and then Item_202.Kind in Ada_Pragma_Node_Range; 
--# expr-done 5697
         if Is_A then
      --# expr-start 5698 '<Cast to Pragma at ast.py:18683>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18683
Cast_Expr := Item_202; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_Pragma_Node_Range
   then
      Cast_Result := Cast_Expr;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 5698
         declare
            Item_To_Append : constant Bare_Pragma_Node := Cast_Result;
         begin
            Bare_Pragma_Node_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Bare_Pragma_Node_Array
           (Items_Count => Natural (Bare_Pragma_Node_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Bare_Pragma_Node_Vectors.Get
              (Map_Result_Vec,
               I + Bare_Pragma_Node_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Bare_Pragma_Node_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 5699
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 5700
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1317;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1317;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Compilation_Unit_P_Local_Config_Pragmas;
--# end
--# property-start CompilationUnit.sources_config_pragmas /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18688
pragma Warnings (Off, "is not referenced");
 function Compilation_Unit_P_Sources_Config_Pragmas
  (Node : Bare_Compilation_Unit
  )
   return Bare_Pragma_Node_Array_Access
is
   Self : Bare_Compilation_Unit := Bare_Compilation_Unit (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Pragma_Node_Array_Access;
            procedure Finalizer_Scope_1318 with Inline_Always;
            procedure Finalizer_Scope_2548 with Inline_Always;
            procedure Finalizer_Scope_2549 with Inline_Always;
      Block_Var : Bare_Pragma_Node_Array_Access;
Block_Var_1 : Bare_Pragma_Node_Array_Access;
Var_Expr : Internal_Entity_Compilation_Unit;
Var_Expr_1 : Internal_Entity_Compilation_Unit;
Fld : Bare_Pragma_Node_Array_Access;
Fld_1 : Bare_Ada_Node;
Is_A : Boolean;
Fld_2 : Internal_Entity_Compilation_Unit;
Fld_3 : Bare_Pragma_Node_Array_Access;
Scope_Result : Bare_Pragma_Node_Array_Access;
Result_Var : Bare_Pragma_Node_Array_Access;
Fld_4 : Bare_Ada_Node;
Is_A_1 : Boolean;
Fld_5 : Bare_Ada_Node;
Cast_Result : Bare_Subunit;
Fld_6 : Internal_Entity_Compilation_Unit;
Fld_7 : Bare_Pragma_Node_Array_Access;
Scope_Result_1 : Bare_Pragma_Node_Array_Access;
Result_Var_1 : Bare_Pragma_Node_Array_Access;
If_Result : Bare_Pragma_Node_Array_Access;
If_Result_1 : Bare_Pragma_Node_Array_Access;
Concat_Result : Bare_Pragma_Node_Array_Access;
Let_Result : Bare_Pragma_Node_Array_Access;
            procedure Finalizer_Scope_1318 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (If_Result);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1318;
            procedure Finalizer_Scope_2548 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_2548;
            procedure Finalizer_Scope_2549 is
            begin
                     Dec_Ref (Fld_7);
            end Finalizer_Scope_2549;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5714 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5701 '<FieldAccess for local_config_pragmas at ast.py:18694>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18694
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.local_config_pragmas
Fld := Libadalang.Implementation.Compilation_Unit_P_Local_Config_Pragmas (Node => Self);
--# end
--# expr-done 5701
Block_Var := Fld; Inc_Ref (Block_Var);
--# bind current_unit Block_Var
--# expr-start 5712 '<Cond at ast.py:18698>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18698
--# expr-start 5703 '<IsA LibraryItem at ast.py:18701>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18701
--# expr-start 5702 '<FieldAccess for body at ast.py:18701>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18701
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Self.Compilation_Unit_F_Body;
--# expr-done 5702
Is_A := Fld_1 /= null 
and then Fld_1.Kind in Ada_Library_Item_Range; 
--# expr-done 5703
if Is_A then
--# expr-start 5704 '<FieldAccess for other_part at ast.py:18702>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18702
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.other_part
Fld_2 := Libadalang.Implementation.Compilation_Unit_P_Other_Part (Node => Self);
--# end
--# expr-done 5704
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Compilation_Unit then
   --# scope-start
--# expr-start 5705 '<FieldAccess for local_config_pragmas at ast.py:18702>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18702
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.local_config_pragmas
Fld_3 := Libadalang.Implementation.Compilation_Unit_P_Local_Config_Pragmas (Node => Var_Expr.Node);
--# end
--# expr-done 5705
Scope_Result := Fld_3;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2548;
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Pragma_Node_Array_Type;
end if;
      Inc_Ref (Result_Var);
   If_Result_1 := Result_Var;
else
--# expr-start 5707 '<IsA Subunit at ast.py:18706>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18706
--# expr-start 5706 '<FieldAccess for body at ast.py:18706>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18706
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Self.Compilation_Unit_F_Body;
--# expr-done 5706
Is_A_1 := Fld_4 /= null 
and then Fld_4.Kind in Ada_Subunit_Range; 
--# expr-done 5707
if Is_A_1 then
--# expr-start 5710 '<FieldAccess for root_unit at ast.py:18707>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18707
--# expr-start 5709 '<Cast to Subunit at ast.py:18707>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18707
--# expr-start 5708 '<FieldAccess for body at ast.py:18707>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18707
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Self.Compilation_Unit_F_Body;
--# expr-done 5708
   if Fld_5 = null
      or else Fld_5.Kind in Ada_Subunit_Range
   then
      Cast_Result := Fld_5;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 5709
      if Cast_Result = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Subunit.root_unit
Fld_6 := Libadalang.Implementation.Subunit_P_Root_Unit (Node => Cast_Result);
--# end
--# expr-done 5710
Var_Expr_1 := Fld_6;
if Var_Expr_1 /= No_Entity_Compilation_Unit then
   --# scope-start
--# expr-start 5711 '<FieldAccess for sources_config_pragmas at ast.py:18707>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18707
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.sources_config_pragmas
Fld_7 := Libadalang.Implementation.Compilation_Unit_P_Sources_Config_Pragmas (Node => Var_Expr_1.Node);
--# end
--# expr-done 5711
Scope_Result_1 := Fld_7;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2549;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Bare_Pragma_Node_Array_Type;
end if;
      Inc_Ref (Result_Var_1);
   If_Result := Result_Var_1;
else
   If_Result := No_Bare_Pragma_Node_Array_Type;
end if;
      Inc_Ref (If_Result);
   If_Result_1 := If_Result;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 5712
Block_Var_1 := If_Result_1; Inc_Ref (Block_Var_1);
--# bind related_units Block_Var_1
--# expr-start 5713 '<Concat at ast.py:18713>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18713
Concat_Result := Concat (Block_Var, Block_Var_1); 
--# expr-done 5713
Let_Result := Concat_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 5714
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1318;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1318;
               Finalizer_Scope_2548;
               Finalizer_Scope_2549;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Compilation_Unit_P_Sources_Config_Pragmas;
--# end
--# property-start CompilationUnit.all_config_pragmas /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18716
pragma Warnings (Off, "is not referenced");
 function Compilation_Unit_P_All_Config_Pragmas
  (Node : Bare_Compilation_Unit
  )
   return Internal_Entity_Pragma_Node_Array_Access
is
   Self : Bare_Compilation_Unit := Bare_Compilation_Unit (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Pragma_Node_Array_Access;
            procedure Finalizer_Scope_1319 with Inline_Always;
      Fld : Bare_Pragma_Node_Array_Access;
Fld_1 : Bare_Pragma_Node_Array_Access;
Concat_Result : Bare_Pragma_Node_Array_Access;
As_Entity : Internal_Entity_Pragma_Node;
Is_Null : Boolean;
If_Result : Internal_Entity_Pragma_Node;
Map_Result : Internal_Entity_Pragma_Node_Array_Access;
Let_Result : Internal_Entity_Pragma_Node_Array_Access;
            procedure Finalizer_Scope_1319 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1319;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5720 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5719 '<Map at ast.py:18726>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18726
   --# expr-start 5717 '<Concat at ast.py:18726>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18726
--# expr-start 5715 '<FieldAccess for sources_config_pragmas at ast.py:18726>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18726
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.sources_config_pragmas
Fld := Libadalang.Implementation.Compilation_Unit_P_Sources_Config_Pragmas (Node => Self);
--# end
--# expr-done 5715
--# expr-start 5716 '<FieldAccess for external_config_pragmas at ast.py:18726>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18726
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Compilation_Unit_P_External_Config_Pragmas (Node => Self);
--# expr-done 5716
Concat_Result := Concat (Fld, Fld_1); 
--# expr-done 5717
      declare
         Map_Result_Vec : Internal_Entity_Pragma_Node_Vectors.Vector;
      begin
      declare
         Collection : constant Bare_Pragma_Node_Array_Access := Concat_Result;
      begin
         for Item_203 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_203
      --# expr-start 5718 '<as_bare_entity at ast.py:18727>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18727
Is_Null := Item_203 = null; 
if Is_Null then
   If_Result := No_Entity_Pragma_Node;
else
As_Entity := (Info => No_Entity_Info, Node => Item_203); 
   If_Result := As_Entity;
end if;
--# expr-done 5718
         declare
            Item_To_Append : constant Internal_Entity_Pragma_Node := If_Result;
         begin
            Internal_Entity_Pragma_Node_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Pragma_Node_Array
           (Items_Count => Natural (Internal_Entity_Pragma_Node_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Pragma_Node_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Pragma_Node_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Pragma_Node_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 5719
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 5720
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1319;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1319;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Compilation_Unit_P_All_Config_Pragmas;
--# end
--# property-start CompilationUnit.config_pragmas /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18731
pragma Warnings (Off, "is not referenced");
 function Compilation_Unit_P_Config_Pragmas
  (Node : Bare_Compilation_Unit
      ; Name : Symbol_Type
  )
   return Internal_Entity_Pragma_Node_Array_Access
is
   Self : Bare_Compilation_Unit := Bare_Compilation_Unit (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind name Name
   Property_Result : Internal_Entity_Pragma_Node_Array_Access;
            procedure Finalizer_Scope_1320 with Inline_Always;
      Fld : Internal_Entity_Pragma_Node_Array_Access;
Fld_1 : Internal_Entity_Identifier;
Fld_2 : Symbol_Type;
Is_Equal : Boolean;
Map_Result : Internal_Entity_Pragma_Node_Array_Access;
Let_Result : Internal_Entity_Pragma_Node_Array_Access;
            procedure Finalizer_Scope_1320 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1320;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5726 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5725 '<Filter at ast.py:18740>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18740
   --# expr-start 5721 '<FieldAccess for all_config_pragmas at ast.py:18740>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18740
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.all_config_pragmas
Fld := Libadalang.Implementation.Compilation_Unit_P_All_Config_Pragmas (Node => Self);
--# end
--# expr-done 5721
      declare
         Map_Result_Vec : Internal_Entity_Pragma_Node_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Pragma_Node_Array_Access := Fld;
      begin
         for Item_204 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_204
         --# expr-start 5724 '<Eq at ast.py:18741>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18741
--# expr-start 5723 '<FieldAccess for name_symbol at ast.py:18741>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18741
--# expr-start 5722 '<FieldAccess for id at ast.py:18741>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18741
      if Item_204.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Identifier (Node => Item_204.Node.Pragma_Node_F_Id, Info => Item_204.Info);
--# expr-done 5722
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_1.Node);
--# end
--# expr-done 5723
Is_Equal := Fld_2 = Name; 
--# expr-done 5724
         if Is_Equal then
         declare
            Item_To_Append : constant Internal_Entity_Pragma_Node := Item_204;
         begin
            Internal_Entity_Pragma_Node_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Pragma_Node_Array
           (Items_Count => Natural (Internal_Entity_Pragma_Node_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Pragma_Node_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Pragma_Node_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Pragma_Node_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 5725
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 5726
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1320;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1320;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Compilation_Unit_P_Config_Pragmas;
--# end
   --
   --  Primitives for Bare_Component_Clause
   --
      procedure Initialize_Fields_For_Component_Clause
        (Self : Bare_Component_Clause
         ; Component_Clause_F_Id : Bare_Identifier
         ; Component_Clause_F_Position : Bare_Expr
         ; Component_Clause_F_Range : Bare_Range_Spec
        ) is
      begin
            Self.Component_Clause_F_Id := Component_Clause_F_Id;
            Self.Component_Clause_F_Position := Component_Clause_F_Position;
            Self.Component_Clause_F_Range := Component_Clause_F_Range;
      end Initialize_Fields_For_Component_Clause;
   function Component_Clause_F_Id
     (Node : Bare_Component_Clause) return Bare_Identifier
   is
   begin
         return Node.Component_Clause_F_Id;
   end;
   function Component_Clause_F_Position
     (Node : Bare_Component_Clause) return Bare_Expr
   is
   begin
         return Node.Component_Clause_F_Position;
   end;
   function Component_Clause_F_Range
     (Node : Bare_Component_Clause) return Bare_Range_Spec
   is
   begin
         return Node.Component_Clause_F_Range;
   end;
--# property-start ComponentClause.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9313
pragma Warnings (Off, "is not referenced");
 function Component_Clause_P_Xref_Entry_Point
  (Node : Bare_Component_Clause
  )
   return Boolean
is
   Self : Bare_Component_Clause := Bare_Component_Clause (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_Clause_P_Xref_Entry_Point;
--# end
--# property-start ComponentClause.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9316
pragma Warnings (Off, "is not referenced");
 function Component_Clause_P_Xref_Equation
  (Node : Bare_Component_Clause
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Component_Clause := Bare_Component_Clause (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Component_Clause :=
     Internal_Entity_Component_Clause'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_741 with Inline_Always;
      Block_Var : Internal_Entity_Record_Rep_Clause;
Block_Var_1 : Lexical_Env := Empty_Env;
Fld : Internal_Entity;
Fld_1 : Internal_Entity;
Cast_Result : Internal_Entity_Record_Rep_Clause;
Fld_2 : Internal_Entity_Name;
Fld_3 : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_4 : Internal_Entity_Identifier;
Fld_5 : Logic_Equation;
Dyn_Var_Bind_Result : Logic_Equation;
Fld_6 : Internal_Entity_Expr;
Fld_7 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_8 : Internal_Entity_Range_Spec;
Fld_9 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_741 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_5);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Fld_7);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_9);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_741;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5741 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5729 '<Cast to RecordRepClause at ast.py:9319>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9319
--# expr-start 5728 '<parent at ast.py:9319>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9319
--# expr-start 5727 '<parent at ast.py:9319>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9319
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 5727
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Parent (Node => Fld.Node, E_Info => Fld.Info);
--# expr-done 5728
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Record_Rep_Clause_Range
   then
      Cast_Result := Create_Internal_Entity_Record_Rep_Clause
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 5729
Block_Var := Cast_Result; 
--# bind rep_clause Block_Var
--# expr-start 5731 '<FieldAccess for designated_env_no_overloading at ast.py:9323>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9323
--# expr-start 5730 '<FieldAccess for name at ast.py:9323>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9323
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Block_Var.Node.Record_Rep_Clause_F_Name, Info => Block_Var.Info);
--# expr-done 5730
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.designated_env_no_overloading'
Fld_3 := Libadalang.Implementation.Dispatcher_Expr_P_Designated_Env_No_Overloading (Node => Fld_2.Node, Env => Env, Origin => Origin, No_Visibility => False, E_Info => Fld_2.Info);
--# end
--# expr-done 5731
Block_Var_1 := Fld_3; Inc_Ref (Block_Var_1);
--# bind record_env Block_Var_1
--# expr-start 5740 '<And at ast.py:9325>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9325
--# expr-start 5737 '<And at ast.py:9325>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9325
--# expr-start 5734 '<bind at ast.py:9327>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9327
Bound_Env := Block_Var_1; Inc_Ref (Bound_Env);
--# expr-start 5733 '<FieldAccess for xref_equation at ast.py:9327>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9327
--# expr-start 5732 '<FieldAccess for id at ast.py:9327>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9327
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Identifier (Node => Ent.Node.Component_Clause_F_Id, Info => Ent.Info);
--# expr-done 5732
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_5 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Fld_4.Node, Env => Bound_Env, Origin => Origin, E_Info => Fld_4.Info);
--# end
--# expr-done 5733
Dyn_Var_Bind_Result := Fld_5; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 5734
--# expr-start 5736 '<FieldAccess for sub_equation at ast.py:9328>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9328
--# expr-start 5735 '<FieldAccess for position at ast.py:9328>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9328
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Expr (Node => Ent.Node.Component_Clause_F_Position, Info => Ent.Info);
--# expr-done 5735
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_7 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_6.Node, Env => Env, Origin => Origin, E_Info => Fld_6.Info);
--# end
--# expr-done 5736
And_Pred := Create_And (Dyn_Var_Bind_Result, Fld_7, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9325") else null)); 
--# expr-done 5737
--# expr-start 5739 '<FieldAccess for sub_equation at ast.py:9329>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9329
--# expr-start 5738 '<FieldAccess for range at ast.py:9329>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9329
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Range_Spec (Node => Ent.Node.Component_Clause_F_Range, Info => Ent.Info);
--# expr-done 5738
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_9 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_8.Node, Env => Env, Origin => Origin, E_Info => Fld_8.Info);
--# end
--# expr-done 5739
And_Pred_1 := Create_And (And_Pred, Fld_9, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9325") else null)); 
--# expr-done 5740
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 5741
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_741;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_741;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_Clause_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Component_Def
   --
      procedure Initialize_Fields_For_Component_Def
        (Self : Bare_Component_Def
         ; Component_Def_F_Has_Aliased : Bare_Aliased_Node
         ; Component_Def_F_Has_Constant : Bare_Constant_Node
         ; Component_Def_F_Type_Expr : Bare_Type_Expr
        ) is
      begin
            Self.Component_Def_F_Has_Aliased := Component_Def_F_Has_Aliased;
            Self.Component_Def_F_Has_Constant := Component_Def_F_Has_Constant;
            Self.Component_Def_F_Type_Expr := Component_Def_F_Type_Expr;
      end Initialize_Fields_For_Component_Def;
   function Component_Def_F_Has_Aliased
     (Node : Bare_Component_Def) return Bare_Aliased_Node
   is
   begin
         return Node.Component_Def_F_Has_Aliased;
   end;
   function Component_Def_F_Has_Constant
     (Node : Bare_Component_Def) return Bare_Constant_Node
   is
   begin
         return Node.Component_Def_F_Has_Constant;
   end;
   function Component_Def_F_Type_Expr
     (Node : Bare_Component_Def) return Bare_Type_Expr
   is
   begin
         return Node.Component_Def_F_Type_Expr;
   end;
--# property-start ComponentDef.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7806
pragma Warnings (Off, "is not referenced");
 function Component_Def_P_Xref_Equation
  (Node : Bare_Component_Def
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Component_Def := Bare_Component_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Component_Def :=
     Internal_Entity_Component_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_598 with Inline_Always;
      Fld : Internal_Entity_Type_Expr;
Fld_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_598 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_598;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5744 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5743 '<FieldAccess for sub_equation at ast.py:7807>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7807
--# expr-start 5742 '<FieldAccess for type_expr at ast.py:7807>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7807
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Type_Expr (Node => Ent.Node.Component_Def_F_Type_Expr, Info => Ent.Info);
--# expr-done 5742
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 5743
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 5744
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_598;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_598;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Component_Def_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Constant_Node
   --
--# property-start '[dispatcher]Constant.as_bool' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Constant_Node_P_As_Bool
  (Node : Bare_Constant_Node
  )
   return Boolean
is
   Self : Bare_Constant_Node := Bare_Constant_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Constant_Node (Self.Kind) is
               when Ada_Constant_Absent =>
                  --# property-call-start Constant.Absent.as_bool
                  Property_Result := Constant_Absent_P_As_Bool
                    (Self
                    );
                  --# end
               when Ada_Constant_Present =>
                  --# property-call-start Constant.Present.as_bool
                  Property_Result := Constant_Present_P_As_Bool
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Constant_Node_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Constant_Absent
   --
--# property-start Constant.Absent.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4906
pragma Warnings (Off, "is not referenced");
 function Constant_Absent_P_As_Bool
  (Node : Bare_Constant_Absent
  )
   return Boolean
is
   Self : Bare_Constant_Absent := Bare_Constant_Absent (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Constant_Absent_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Constant_Present
   --
--# property-start Constant.Present.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4906
pragma Warnings (Off, "is not referenced");
 function Constant_Present_P_As_Bool
  (Node : Bare_Constant_Present
  )
   return Boolean
is
   Self : Bare_Constant_Present := Bare_Constant_Present (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Constant_Present_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Constraint
   --
--# property-start Constraint.subtype /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7313
pragma Warnings (Off, "is not referenced");
 function Constraint_P_Subtype
  (Node : Bare_Constraint
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Constraint := Bare_Constraint (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_1 : Bare_Ada_Node;
Cast_Result : Bare_Subtype_Indication;
As_Entity : Internal_Entity_Subtype_Indication;
Is_Null : Boolean;
If_Result : Internal_Entity_Subtype_Indication;
Fld_2 : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5750 '<bind at ast.py:7313>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7313
--# expr-start 5745 '<FieldAccess for origin_node at ast.py:7314>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7314
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 5745
Bound_Origin := Fld; 
--# expr-start 5749 '<FieldAccess for designated_type at ast.py:7315>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7315
--# expr-start 5748 '<as_entity at ast.py:7315>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7315
--# expr-start 5747 '<Cast to SubtypeIndication at ast.py:7315>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7315
--# expr-start 5746 '<parent at ast.py:7315>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7315
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Self.Parent; 
--# expr-done 5746
   if Fld_1 = null
      or else Fld_1.Kind in Ada_Subtype_Indication_Range
   then
      Cast_Result := Fld_1;
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 5747
Is_Null := Cast_Result = null; 
if Is_Null then
   If_Result := No_Entity_Subtype_Indication;
else
As_Entity := (Info => E_Info, Node => Cast_Result); 
   If_Result := As_Entity;
end if;
--# expr-done 5748
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => If_Result.Node, Origin => Bound_Origin, E_Info => If_Result.Info);
--# end
--# expr-done 5749
Dyn_Var_Bind_Result := Fld_2; 
--# expr-done 5750
      Property_Result := Dyn_Var_Bind_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Constraint_P_Subtype;
--# end
--# property-start Constraint.is_static /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7320
pragma Warnings (Off, "is not referenced");
 function Constraint_P_Is_Static
  (Node : Bare_Constraint
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Constraint := Bare_Constraint (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Constraint :=
     Internal_Entity_Constraint'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Match_0 : Internal_Entity_Range_Constraint;
Match_1 : Internal_Entity_Composite_Constraint;
Match_2 : Internal_Entity_Digits_Constraint;
Match_3 : Internal_Entity_Delta_Constraint;
Match_0_1 : Internal_Entity_Subtype_Indication;
Match_1_1 : Internal_Entity_Expr;
Match_2_1 : Internal_Entity;
Var_Expr : Internal_Entity_Range_Spec;
Var_Expr_1 : Internal_Entity_Range_Spec;
Fld : Internal_Entity_Range_Spec;
Fld_1 : Internal_Entity_Expr;
Fld_2 : Boolean;
Fld_3 : Internal_Entity_Assoc_List;
Fld_4 : Bare_Assoc_List;
Fld_5 : Internal_Entity_Info;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Cast_Expr : Internal_Entity_Basic_Assoc;
Cast_Result : Internal_Entity_Composite_Constraint_Assoc;
Fld_6 : Internal_Entity;
Fld_7 : Boolean;
Fld_8 : Boolean;
Match_Prefix : Internal_Entity;
Cast_Result_1 : Internal_Entity_Subtype_Indication;
Let_Result : Boolean;
Scope_Result : Boolean;
Cast_Result_2 : Internal_Entity_Expr;
Let_Result_1 : Boolean;
Scope_Result_1 : Boolean;
Cast_Result_3 : Internal_Entity;
Let_Result_2 : Boolean;
Scope_Result_2 : Boolean;
Match_Result : Boolean;
Item_205 : Internal_Entity_Basic_Assoc;
Bare_Item_205 : Bare_Basic_Assoc;
Quantifier_Result : Boolean;
Fld_9 : Internal_Entity_Assoc_List;
Fld_10 : Bare_Assoc_List;
Fld_11 : Internal_Entity_Info;
As_Entity_1 : Internal_Entity_Basic_Assoc;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Basic_Assoc;
Fld_12 : Internal_Entity_Expr;
Fld_13 : Boolean;
Item_206 : Internal_Entity_Basic_Assoc;
Bare_Item_206 : Bare_Basic_Assoc;
Quantifier_Result_1 : Boolean;
Fld_14 : Boolean;
If_Result_2 : Boolean;
Fld_15 : Internal_Entity_Range_Spec;
Fld_16 : Internal_Entity_Expr;
Fld_17 : Boolean;
Scope_Result_3 : Boolean;
Result_Var : Boolean;
Fld_18 : Internal_Entity_Range_Spec;
Fld_19 : Internal_Entity_Expr;
Fld_20 : Boolean;
Scope_Result_4 : Boolean;
Result_Var_1 : Boolean;
Match_Prefix_1 : Internal_Entity_Constraint;
Cast_Result_4 : Internal_Entity_Range_Constraint;
Let_Result_3 : Boolean;
Scope_Result_5 : Boolean;
Cast_Result_5 : Internal_Entity_Composite_Constraint;
Let_Result_4 : Boolean;
Scope_Result_6 : Boolean;
Cast_Result_6 : Internal_Entity_Digits_Constraint;
Let_Result_5 : Boolean;
Scope_Result_7 : Boolean;
Cast_Result_7 : Internal_Entity_Delta_Constraint;
Let_Result_6 : Boolean;
Scope_Result_8 : Boolean;
Match_Result_1 : Boolean;
Let_Result_7 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5774 '<Block at ???>' Let_Result_7 None
--# scope-start
--# expr-start 5773 '<Match at ast.py:7321>' Match_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7321
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix_1 := Ent; 
case Ada_Constraint (Match_Prefix_1.Node.Kind) is
         when Ada_Range_Constraint =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Range_Constraint
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_0 := Cast_Result_4; 
--# bind rc Match_0
--# expr-start 5753 '<FieldAccess for is_static_expr at ast.py:7322>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7322
--# expr-start 5752 '<FieldAccess for range at ast.py:7322>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7322
--# expr-start 5751 '<FieldAccess for range at ast.py:7322>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7322
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Range_Spec (Node => Match_0.Node.Range_Constraint_F_Range, Info => Match_0.Info);
--# expr-done 5751
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Expr (Node => Fld.Node.Range_Spec_F_Range, Info => Fld.Info);
--# expr-done 5752
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_2 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_1.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_1.Info);
--# end
--# expr-done 5753
Let_Result_3 := Fld_2; 
--# end
Scope_Result_5 := Let_Result_3;
   --# end
            Match_Result_1 := Scope_Result_5; 
         when Ada_Composite_Constraint =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Create_Internal_Entity_Composite_Constraint
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_1 := Cast_Result_5; 
--# bind cc Match_1
--# expr-start 5766 '<If at ast.py:7323>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7323
--# expr-start 5754 '<FieldAccess for is_index_constraint at ast.py:7324>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7324
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompositeConstraint.is_index_constraint
Fld_14 := Libadalang.Implementation.Composite_Constraint_P_Is_Index_Constraint (Node => Match_1.Node, E_Info => Match_1.Info);
--# end
--# expr-done 5754
if Fld_14 then
   --# expr-start 5761 '<AllQuantifier at ast.py:7325>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7325
   --# expr-start 5755 '<FieldAccess for constraints at ast.py:7325>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7325
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Assoc_List (Node => Match_1.Node.Composite_Constraint_F_Constraints, Info => Match_1.Info);
--# expr-done 5755
Fld_4 := Fld_3.Node;
      Quantifier_Result := True;
      if Fld_4 = null then
      null;
      else
      declare
         Collection : constant Bare_Assoc_List := Fld_4;
      begin
         for Untyped_Item_205 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_205 := Untyped_Item_205; 
Is_Null := Bare_Item_205 = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_5 := Fld_3.Info;
As_Entity := (Info => Fld_5, Node => Bare_Item_205); 
   If_Result := As_Entity;
end if;
                  Item_205 := If_Result; 
   --# scope-start
               --# bind c Item_205
      --# expr-start 5760 '<Match at ast.py:7326>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7326
--# expr-start 5757 '<FieldAccess for constraint_expr at ast.py:7326>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7326
--# expr-start 5756 '<Cast to CompositeConstraintAssoc at ast.py:7326>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7326
Cast_Expr := Item_205; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Composite_Constraint_Assoc_Range
   then
      Cast_Result := Create_Internal_Entity_Composite_Constraint_Assoc
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Composite_Constraint_Assoc;
   end if;
--# expr-done 5756
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity (Node => Cast_Result.Node.Composite_Constraint_Assoc_F_Constraint_Expr, Info => Cast_Result.Info);
--# expr-done 5757
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld_6; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Subtype_Indication .. Ada_Discrete_Subtype_Indication =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Subtype_Indication
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0_1 := Cast_Result_1; 
--# bind st Match_0_1
--# expr-start 5758 '<FieldAccess for is_static_subtype at ast.py:7328>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7328
      if Match_0_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SubtypeIndication.is_static_subtype
Fld_7 := Libadalang.Implementation.Subtype_Indication_P_Is_Static_Subtype (Node => Match_0_1.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Match_0_1.Info);
--# end
--# expr-done 5758
Let_Result := Fld_7; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Abstract_State_Decl_Expr .. Ada_Un_Op =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1_1 := Cast_Result_2; 
--# bind e Match_1_1
--# expr-start 5759 '<FieldAccess for is_static_expr at ast.py:7329>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7329
      if Match_1_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_8 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Match_1_1.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Match_1_1.Info);
--# end
--# expr-done 5759
Let_Result_1 := Fld_8; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abort_Absent .. Ada_Elsif_Stmt_Part | Ada_Handled_Stmts .. Ada_Enum_Lit_Synth_Type_Expr | Ada_Synthetic_Type_Expr .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2_1 := Cast_Result_3; 
--# bind _ Match_2_1
Let_Result_2 := False; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 5760
      Quantifier_Result := Match_Result;
   --# end
               exit when not Quantifier_Result;
         end loop;
      end;
      end if;
--# expr-done 5761
   If_Result_2 := Quantifier_Result;
else
   --# expr-start 5765 '<AllQuantifier at ast.py:7333>' Quantifier_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7333
   --# expr-start 5762 '<FieldAccess for constraints at ast.py:7333>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7333
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Assoc_List (Node => Match_1.Node.Composite_Constraint_F_Constraints, Info => Match_1.Info);
--# expr-done 5762
Fld_10 := Fld_9.Node;
      Quantifier_Result_1 := True;
      if Fld_10 = null then
      null;
      else
      declare
         Collection : constant Bare_Assoc_List := Fld_10;
      begin
         for Untyped_Item_206 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_206 := Untyped_Item_206; 
Is_Null_1 := Bare_Item_206 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Basic_Assoc;
else
Fld_11 := Fld_9.Info;
As_Entity_1 := (Info => Fld_11, Node => Bare_Item_206); 
   If_Result_1 := As_Entity_1;
end if;
                  Item_206 := If_Result_1; 
   --# scope-start
               --# bind c Item_206
      --# expr-start 5764 '<FieldAccess for is_static_expr at ast.py:7334>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7334
--# expr-start 5763 '<FieldAccess for expr at ast.py:7334>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7334
      if Item_206.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_12 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Item_206.Node, E_Info => Item_206.Info);
--# end
--# expr-done 5763
      if Fld_12.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_13 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_12.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_12.Info);
--# end
--# expr-done 5764
      Quantifier_Result_1 := Fld_13;
   --# end
               exit when not Quantifier_Result_1;
         end loop;
      end;
      end if;
--# expr-done 5765
   If_Result_2 := Quantifier_Result_1;
end if;
--# expr-done 5766
Let_Result_4 := If_Result_2; 
--# end
Scope_Result_6 := Let_Result_4;
   --# end
            Match_Result_1 := Scope_Result_6; 
         when Ada_Digits_Constraint =>
   --# scope-start
--# scope-start
      Cast_Result_6 := Create_Internal_Entity_Digits_Constraint
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_2 := Cast_Result_6; 
--# bind dc Match_2
--# expr-start 5767 '<FieldAccess for range at ast.py:7337>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7337
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Create_Internal_Entity_Range_Spec (Node => Match_2.Node.Digits_Constraint_F_Range, Info => Match_2.Info);
--# expr-done 5767
Var_Expr := Fld_15;
if Var_Expr /= No_Entity_Range_Spec then
   --# scope-start
   --# bind range Var_Expr
--# expr-start 5769 '<FieldAccess for is_static_expr at ast.py:7338>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7338
--# expr-start 5768 '<FieldAccess for range at ast.py:7338>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7338
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_16 := Create_Internal_Entity_Expr (Node => Var_Expr.Node.Range_Spec_F_Range, Info => Var_Expr.Info);
--# expr-done 5768
      if Fld_16.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_17 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_16.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_16.Info);
--# end
--# expr-done 5769
Scope_Result_3 := Fld_17;
   --# end
   Result_Var := Scope_Result_3;
else
   Result_Var := True;
end if;
Let_Result_5 := Result_Var; 
--# end
Scope_Result_7 := Let_Result_5;
   --# end
            Match_Result_1 := Scope_Result_7; 
         when Ada_Delta_Constraint =>
   --# scope-start
--# scope-start
      Cast_Result_7 := Create_Internal_Entity_Delta_Constraint
        (Node => Match_Prefix_1.Node,
         Info => Match_Prefix_1.Info);
Match_3 := Cast_Result_7; 
--# bind dc Match_3
--# expr-start 5770 '<FieldAccess for range at ast.py:7341>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7341
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Create_Internal_Entity_Range_Spec (Node => Match_3.Node.Delta_Constraint_F_Range, Info => Match_3.Info);
--# expr-done 5770
Var_Expr_1 := Fld_18;
if Var_Expr_1 /= No_Entity_Range_Spec then
   --# scope-start
   --# bind range Var_Expr_1
--# expr-start 5772 '<FieldAccess for is_static_expr at ast.py:7342>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7342
--# expr-start 5771 '<FieldAccess for range at ast.py:7342>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7342
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Create_Internal_Entity_Expr (Node => Var_Expr_1.Node.Range_Spec_F_Range, Info => Var_Expr_1.Info);
--# expr-done 5771
      if Fld_19.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_20 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_19.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_19.Info);
--# end
--# expr-done 5772
Scope_Result_4 := Fld_20;
   --# end
   Result_Var_1 := Scope_Result_4;
else
   Result_Var_1 := True;
end if;
Let_Result_6 := Result_Var_1; 
--# end
Scope_Result_8 := Let_Result_6;
   --# end
            Match_Result_1 := Scope_Result_8; 
end case;
--# expr-done 5773
Let_Result_7 := Match_Result_1; 
--# end
--# expr-done 5774
      Property_Result := Let_Result_7;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Constraint_P_Is_Static;
--# end
   --
   --  Primitives for Bare_Composite_Constraint
   --
      procedure Initialize_Fields_For_Composite_Constraint
        (Self : Bare_Composite_Constraint
         ; Composite_Constraint_F_Constraints : Bare_Assoc_List
        ) is
      begin
            Self.Composite_Constraint_F_Constraints := Composite_Constraint_F_Constraints;
      end Initialize_Fields_For_Composite_Constraint;
   function Composite_Constraint_F_Constraints
     (Node : Bare_Composite_Constraint) return Bare_Assoc_List
   is
   begin
         return Node.Composite_Constraint_F_Constraints;
   end;
--# property-start CompositeConstraint.is_index_constraint /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7405
pragma Warnings (Off, "is not referenced");
 function Composite_Constraint_P_Is_Index_Constraint
  (Node : Bare_Composite_Constraint
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Composite_Constraint := Bare_Composite_Constraint (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Composite_Constraint :=
     Internal_Entity_Composite_Constraint'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
Fld_2 : Internal_Entity_Base_Type_Decl;
Fld_3 : Boolean;
Fld_4 : Internal_Entity_Base_Type_Decl;
Fld_5 : Internal_Entity_Base_Type_Decl;
Fld_6 : Boolean;
If_Result : Boolean;
If_Result_1 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5782 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5776 '<FieldAccess for is_array_type at ast.py:7410>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7410
--# expr-start 5775 '<FieldAccess for subtype at ast.py:7410>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7410
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Constraint.subtype
Fld := Libadalang.Implementation.Constraint_P_Subtype (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5775
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_array_type
Fld_1 := Libadalang.Implementation.Base_Type_Decl_P_Is_Array_Type (Node => Fld.Node, Origin => No_Bare_Ada_Node, E_Info => Fld.Info);
--# end
--# expr-done 5776
if Fld_1 then
   If_Result_1 := True;
else
--# expr-start 5778 '<FieldAccess for is_access_type at ast.py:7411>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7411
--# expr-start 5777 '<FieldAccess for subtype at ast.py:7411>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7411
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Constraint.subtype
Fld_2 := Libadalang.Implementation.Constraint_P_Subtype (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5777
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_access_type'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Access_Type (Node => Fld_2.Node, Origin => No_Bare_Ada_Node, E_Info => Fld_2.Info);
--# end
--# expr-done 5778
if Fld_3 then
   --# expr-start 5781 '<FieldAccess for is_array_type at ast.py:7412>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7412
--# expr-start 5780 '<FieldAccess for accessed_type at ast.py:7412>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7412
--# expr-start 5779 '<FieldAccess for subtype at ast.py:7412>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7412
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Constraint.subtype
Fld_4 := Libadalang.Implementation.Constraint_P_Subtype (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5779
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_5 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Fld_4.Node, Origin => No_Bare_Ada_Node, E_Info => Fld_4.Info);
--# end
--# expr-done 5780
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_array_type
Fld_6 := Libadalang.Implementation.Base_Type_Decl_P_Is_Array_Type (Node => Fld_5.Node, Origin => No_Bare_Ada_Node, E_Info => Fld_5.Info);
--# end
--# expr-done 5781
   If_Result := Fld_6;
else
   If_Result := False;
end if;
   If_Result_1 := If_Result;
end if;
Let_Result := If_Result_1; 
--# end
--# expr-done 5782
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Composite_Constraint_P_Is_Index_Constraint;
--# end
--# property-start CompositeConstraint.is_discriminant_constraint /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7416
pragma Warnings (Off, "is not referenced");
 function Composite_Constraint_P_Is_Discriminant_Constraint
  (Node : Bare_Composite_Constraint
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Composite_Constraint := Bare_Composite_Constraint (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Composite_Constraint :=
     Internal_Entity_Composite_Constraint'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Boolean;
Not_Val : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5785 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5784 '<Not at ast.py:7420>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7420
--# expr-start 5783 '<FieldAccess for is_index_constraint at ast.py:7420>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7420
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompositeConstraint.is_index_constraint
Fld := Libadalang.Implementation.Composite_Constraint_P_Is_Index_Constraint (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5783
Not_Val := not (Fld); 
--# expr-done 5784
Let_Result := Not_Val; 
--# end
--# expr-done 5785
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Composite_Constraint_P_Is_Discriminant_Constraint;
--# end
--# property-start CompositeConstraint.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7423
pragma Warnings (Off, "is not referenced");
 function Composite_Constraint_P_Xref_Equation
  (Node : Bare_Composite_Constraint
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Composite_Constraint := Bare_Composite_Constraint (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Composite_Constraint :=
     Internal_Entity_Composite_Constraint'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_551 with Inline_Always;
            procedure Finalizer_Scope_2563 with Inline_Always;
            procedure Finalizer_Scope_2564 with Inline_Always;
            procedure Finalizer_Scope_2565 with Inline_Always;
            procedure Finalizer_Scope_2566 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Ex : Internal_Entity;
Var_Expr : Internal_Entity_Expr;
Var_Expr_1 : Bare_Base_Id;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Assoc_List;
Fld_2 : Bare_Assoc_List;
Fld_3 : Internal_Entity_Info;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Cast_Expr : Internal_Entity_Basic_Assoc;
Cast_Result : Internal_Entity_Composite_Constraint_Assoc;
Fld_4 : Internal_Entity;
Cast_Expr_1 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Expr;
Fld_5 : Logic_Var;
Fld_6 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
Fld_7 : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
Item_207 : Internal_Entity_Basic_Assoc;
Bare_Item_207 : Bare_Basic_Assoc;
Index_208 : Integer;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Fld_8 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_9 : Internal_Entity_Assoc_List;
Fld_10 : Internal_Param_Match_Array_Access;
Fld_11 : Internal_Single_Actual;
Fld_12 : Internal_Entity_Basic_Assoc;
Fld_13 : Internal_Entity_Expr;
Fld_14 : Logic_Var;
Fld_15 : Internal_Entity_Defining_Name;
Fld_16 : Internal_Entity_Base_Formal_Param_Decl;
Fld_17 : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
Fld_18 : Internal_Single_Actual;
Fld_19 : Internal_Entity_Basic_Assoc;
Fld_20 : Internal_Entity_Expr;
Fld_21 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_22 : Internal_Single_Actual;
Fld_23 : Internal_Entity_Basic_Assoc;
Fld_24 : Internal_Entity_Expr;
Fld_25 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_26 : Internal_Single_Actual;
Fld_27 : Bare_Base_Id;
Fld_28 : Logic_Var;
Fld_29 : Internal_Entity_Defining_Name;
Fld_30 : Internal_Entity_Base_Formal_Param_Decl;
Cast_Result_4 : Internal_Entity;
Bind_Result_2 : Logic_Equation;
Scope_Result_1 : Logic_Equation;
True_Rel_1 : Logic_Equation;
Result_Var_1 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Map_Result_1 : Logic_Equation_Array_Access;
Logic_Boolean_Op_1 : Logic_Equation;
Fld_31 : Boolean;
If_Result_1 : Logic_Equation;
Let_Result_1 : Logic_Equation;
            procedure Finalizer_Scope_551 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Fld_8);
                     Dec_Ref (Fld_10);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Logic_Boolean_Op_1);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Let_Result_1);
            end Finalizer_Scope_551;
            procedure Finalizer_Scope_2563 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Fld_7);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_2563;
            procedure Finalizer_Scope_2564 is
            begin
                     Dec_Ref (Bind_Result);
            end Finalizer_Scope_2564;
            procedure Finalizer_Scope_2565 is
            begin
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Fld_21);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_25);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (True_Rel_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (And_Pred_3);
            end Finalizer_Scope_2565;
            procedure Finalizer_Scope_2566 is
            begin
                     Dec_Ref (Bind_Result_2);
            end Finalizer_Scope_2566;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5831 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 5786 '<FieldAccess for subtype at ast.py:7424>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7424
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Constraint.subtype
Fld := Libadalang.Implementation.Constraint_P_Subtype (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5786
Block_Var := Fld; 
--# bind typ Block_Var
--# expr-start 5830 '<If at ast.py:7425>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7425
--# expr-start 5787 '<FieldAccess for is_index_constraint at ast.py:7426>' Fld_31 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7426
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompositeConstraint.is_index_constraint
Fld_31 := Libadalang.Implementation.Composite_Constraint_P_Is_Index_Constraint (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5787
if Fld_31 then
   --# expr-start 5799 '<LogicAll at ast.py:7427>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7427
--# expr-start 5798 '<Map at ast.py:7427>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7427
   --# expr-start 5788 '<FieldAccess for constraints at ast.py:7427>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7427
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Assoc_List (Node => Ent.Node.Composite_Constraint_F_Constraints, Info => Ent.Info);
--# expr-done 5788
Fld_2 := Fld_1.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_2 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
         Index_208 := 0;
      declare
         Collection : constant Bare_Assoc_List := Fld_2;
      begin
         for Untyped_Item_207 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_207 := Untyped_Item_207; 
Is_Null := Bare_Item_207 = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_3 := Fld_1.Info;
As_Entity := (Info => Fld_3, Node => Bare_Item_207); 
   If_Result := As_Entity;
end if;
                  Item_207 := If_Result; 
   --# scope-start
               --# bind c Item_207
               --# bind i Index_208
      --# expr-start 5797 '<Let at ast.py:7427>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7427
--# scope-start
--# expr-start 5790 '<FieldAccess for constraint_expr at ast.py:7428>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7428
--# expr-start 5789 '<Cast to CompositeConstraintAssoc at ast.py:7428>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7428
Cast_Expr := Item_207; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Composite_Constraint_Assoc_Range
   then
      Cast_Result := Create_Internal_Entity_Composite_Constraint_Assoc
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Composite_Constraint_Assoc;
   end if;
--# expr-done 5789
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity (Node => Cast_Result.Node.Composite_Constraint_Assoc_F_Constraint_Expr, Info => Cast_Result.Info);
--# expr-done 5790
Ex := Fld_4; 
--# bind ex Ex
--# expr-start 5796 '<And at ast.py:7448>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7448
--# expr-start 5791 '<Cast to Expr at ast.py:7448>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7448
Cast_Expr_1 := Ex; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Expr
   then
      Cast_Result_1 := Create_Internal_Entity_Expr
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_1 := No_Entity_Expr;
   end if;
--# expr-done 5791
Var_Expr := Cast_Result_1;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind e Var_Expr
--# expr-start 5794 '<Bind at ast.py:7450>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7450
--# expr-start 5792 '<FieldAccess for expected_type_var at ast.py:7450>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7450
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Var_Expr.Node);
--# expr-done 5792
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
--# expr-start 5793 '<FieldAccess for index_type at ast.py:7450>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7450
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.index_type
Fld_6 := Libadalang.Implementation.Base_Type_Decl_P_Index_Type (Node => Block_Var.Node, Dim => Index_208, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 5793
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_6.Node,
         Info => Fld_6.Info);
Bind_Result := Solver.Create_Assign (Fld_5, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7450") else null)); 
--# expr-done 5794
Scope_Result := Bind_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2564;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7451") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
--# expr-start 5795 '<FieldAccess for sub_equation at ast.py:7452>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7452
      if Ex.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_7 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Ex.Node, Env => Env, Origin => Origin, E_Info => Ex.Info);
--# end
--# expr-done 5795
And_Pred := Create_And (Result_Var, Fld_7, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7448") else null)); 
--# expr-done 5796
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 5797
         declare
            Item_To_Append : constant Logic_Equation := Let_Result;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2563;
               Index_208 := Index_208 + 1;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 5798
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7427") else null)); 
--# expr-done 5799
   If_Result_1 := Logic_Boolean_Op;
else
   --# expr-start 5829 '<LogicAll at ast.py:7456>' Logic_Boolean_Op_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7456
--# expr-start 5828 '<Map at ast.py:7456>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7456
   --# expr-start 5802 '<FieldAccess for match_formals at ast.py:7456>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7456
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 5800 '<FieldAccess for discriminants_list at ast.py:7457>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7457
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.discriminants_list'
Fld_8 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Discriminants_List (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 5800
--# expr-start 5801 '<FieldAccess for constraints at ast.py:7457>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7457
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Assoc_List (Node => Ent.Node.Composite_Constraint_F_Constraints, Info => Ent.Info);
--# expr-done 5801
--# property-call-start AdaNode.match_formals
Fld_10 := Libadalang.Implementation.Ada_Node_P_Match_Formals (Node => Self, Formal_Params => Fld_8, Params => Fld_9, Is_Dottable_Subp => False);
--# end
--# expr-done 5802
      declare
         Map_Result_1_Vec : Logic_Equation_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Param_Match_Array_Access := Fld_10;
      begin
         for Item_209 of
               Collection.Items
         loop
   --# scope-start
               --# bind pm Item_209
      --# expr-start 5827 '<And at ast.py:7460>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7460
--# expr-start 5820 '<And at ast.py:7460>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7460
--# expr-start 5815 '<And at ast.py:7460>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7460
--# expr-start 5810 '<Bind at ast.py:7460>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7460
--# expr-start 5806 '<FieldAccess for expected_type_var at ast.py:7460>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7460
--# expr-start 5805 '<FieldAccess for expr at ast.py:7460>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7460
--# expr-start 5804 '<FieldAccess for assoc at ast.py:7460>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7460
--# expr-start 5803 '<FieldAccess for actual at ast.py:7460>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7460
Fld_11 := Item_209.Actual;
--# expr-done 5803
Fld_12 := Fld_11.Assoc;
--# expr-done 5804
      if Fld_12.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_13 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_12.Node, E_Info => Fld_12.Info);
--# end
--# expr-done 5805
      if Fld_13.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_13.Node);
--# expr-done 5806
Fld_14.Value := No_Entity;
Entity_Vars.Reset (Fld_14);
--# expr-start 5809 '<FieldAccess for formal_type at ast.py:7461>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7461
--# expr-start 5808 '<FieldAccess for formal_decl at ast.py:7461>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7461
--# expr-start 5807 '<FieldAccess for formal at ast.py:7461>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7461
Fld_15 := Item_209.Formal;
--# expr-done 5807
      if Fld_15.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_16 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Fld_15.Node, E_Info => Fld_15.Info);
--# end
--# expr-done 5808
      if Fld_16.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamDecl.formal_type
Fld_17 := Libadalang.Implementation.Base_Formal_Param_Decl_P_Formal_Type (Node => Fld_16.Node, Origin => Origin, E_Info => Fld_16.Info);
--# end
--# expr-done 5809
      Cast_Result_3 := Create_Internal_Entity
        (Node => Fld_17.Node,
         Info => Fld_17.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_14, Cast_Result_3, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7460") else null)); 
--# expr-done 5810
--# expr-start 5814 '<FieldAccess for sub_equation at ast.py:7462>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7462
--# expr-start 5813 '<FieldAccess for expr at ast.py:7462>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7462
--# expr-start 5812 '<FieldAccess for assoc at ast.py:7462>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7462
--# expr-start 5811 '<FieldAccess for actual at ast.py:7462>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7462
Fld_18 := Item_209.Actual;
--# expr-done 5811
Fld_19 := Fld_18.Assoc;
--# expr-done 5812
      if Fld_19.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_20 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_19.Node, E_Info => Fld_19.Info);
--# end
--# expr-done 5813
      if Fld_20.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_21 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_20.Node, Env => Env, Origin => Origin, E_Info => Fld_20.Info);
--# end
--# expr-done 5814
And_Pred_1 := Create_And (Bind_Result_1, Fld_21, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7460") else null)); 
--# expr-done 5815
--# expr-start 5819 '<FieldAccess for matches_expected_formal_type at ast.py:7463>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7463
--# expr-start 5818 '<FieldAccess for expr at ast.py:7463>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7463
--# expr-start 5817 '<FieldAccess for assoc at ast.py:7463>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7463
--# expr-start 5816 '<FieldAccess for actual at ast.py:7463>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7463
Fld_22 := Item_209.Actual;
--# expr-done 5816
Fld_23 := Fld_22.Assoc;
--# expr-done 5817
      if Fld_23.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_24 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_23.Node, E_Info => Fld_23.Info);
--# end
--# expr-done 5818
      if Fld_24.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_25 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_24.Node, Origin => Origin);
--# end
--# expr-done 5819
And_Pred_2 := Create_And (And_Pred_1, Fld_25, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7460") else null)); 
--# expr-done 5820
--# expr-start 5822 '<FieldAccess for name at ast.py:7464>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7464
--# expr-start 5821 '<FieldAccess for actual at ast.py:7464>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7464
Fld_26 := Item_209.Actual;
--# expr-done 5821
Fld_27 := Fld_26.Name;
--# expr-done 5822
Var_Expr_1 := Fld_27;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind name Var_Expr_1
--# expr-start 5826 '<Bind at ast.py:7465>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7465
--# expr-start 5823 '<FieldAccess for ref_var at ast.py:7465>' Fld_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7465
      if Var_Expr_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_28 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Var_Expr_1);
--# end
--# expr-done 5823
Fld_28.Value := No_Entity;
Entity_Vars.Reset (Fld_28);
--# expr-start 5825 '<FieldAccess for formal_decl at ast.py:7465>' Fld_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7465
--# expr-start 5824 '<FieldAccess for formal at ast.py:7465>' Fld_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7465
Fld_29 := Item_209.Formal;
--# expr-done 5824
      if Fld_29.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_30 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Fld_29.Node, E_Info => Fld_29.Info);
--# end
--# expr-done 5825
      Cast_Result_4 := Create_Internal_Entity
        (Node => Fld_30.Node,
         Info => Fld_30.Info);
Bind_Result_2 := Solver.Create_Assign (Fld_28, Cast_Result_4, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7465") else null)); 
--# expr-done 5826
Scope_Result_1 := Bind_Result_2;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2566;
   Result_Var_1 := Scope_Result_1;
else
   True_Rel_1 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7466") else null)); 
   Result_Var_1 := True_Rel_1;
end if;
      Inc_Ref (Result_Var_1);
And_Pred_3 := Create_And (And_Pred_2, Result_Var_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7460") else null)); 
--# expr-done 5827
         declare
            Item_To_Append : constant Logic_Equation := And_Pred_3;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2565;
         end loop;
      end;
         Map_Result_1 := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_1_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 5828
Logic_Boolean_Op_1 := Solver.Create_All (Relation_Array (Map_Result_1.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7456") else null)); 
--# expr-done 5829
   If_Result_1 := Logic_Boolean_Op_1;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 5830
Let_Result_1 := If_Result_1; Inc_Ref (Let_Result_1);
--# end
--# expr-done 5831
      Property_Result := Let_Result_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_551;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_551;
               Finalizer_Scope_2563;
               Finalizer_Scope_2564;
               Finalizer_Scope_2565;
               Finalizer_Scope_2566;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Composite_Constraint_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Delta_Constraint
   --
      procedure Initialize_Fields_For_Delta_Constraint
        (Self : Bare_Delta_Constraint
         ; Delta_Constraint_F_Digits : Bare_Expr
         ; Delta_Constraint_F_Range : Bare_Range_Spec
        ) is
      begin
            Self.Delta_Constraint_F_Digits := Delta_Constraint_F_Digits;
            Self.Delta_Constraint_F_Range := Delta_Constraint_F_Range;
      end Initialize_Fields_For_Delta_Constraint;
   function Delta_Constraint_F_Digits
     (Node : Bare_Delta_Constraint) return Bare_Expr
   is
   begin
         return Node.Delta_Constraint_F_Digits;
   end;
   function Delta_Constraint_F_Range
     (Node : Bare_Delta_Constraint) return Bare_Range_Spec
   is
   begin
         return Node.Delta_Constraint_F_Range;
   end;
--# property-start DeltaConstraint.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7386
pragma Warnings (Off, "is not referenced");
 function Delta_Constraint_P_Xref_Equation
  (Node : Bare_Delta_Constraint
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Delta_Constraint := Bare_Delta_Constraint (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Delta_Constraint :=
     Internal_Entity_Delta_Constraint'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_548 with Inline_Always;
            procedure Finalizer_Scope_2567 with Inline_Always;
      Var_Expr : Internal_Entity_Range_Spec;
Fld : Internal_Entity_Expr;
Fld_1 : Logic_Equation;
Fld_2 : Internal_Entity_Range_Spec;
Fld_3 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred : Logic_Equation;
            procedure Finalizer_Scope_548 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred);
            end Finalizer_Scope_548;
            procedure Finalizer_Scope_2567 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_2567;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5836 '<And at ast.py:7387>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7387
--# expr-start 5833 '<FieldAccess for sub_equation at ast.py:7387>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7387
--# expr-start 5832 '<FieldAccess for digits at ast.py:7387>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7387
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Delta_Constraint_F_Digits, Info => Ent.Info);
--# expr-done 5832
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 5833
--# expr-start 5834 '<FieldAccess for range at ast.py:7387>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7387
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Range_Spec (Node => Ent.Node.Delta_Constraint_F_Range, Info => Ent.Info);
--# expr-done 5834
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Range_Spec then
   --# scope-start
   --# bind range Var_Expr
--# expr-start 5835 '<FieldAccess for sub_equation at ast.py:7388>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7388
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 5835
Scope_Result := Fld_3;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2567;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7389") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred := Create_And (Fld_1, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7387") else null)); 
--# expr-done 5836
      Property_Result := And_Pred;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_548;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_548;
               Finalizer_Scope_2567;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Delta_Constraint_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Digits_Constraint
   --
      procedure Initialize_Fields_For_Digits_Constraint
        (Self : Bare_Digits_Constraint
         ; Digits_Constraint_F_Digits : Bare_Expr
         ; Digits_Constraint_F_Range : Bare_Range_Spec
        ) is
      begin
            Self.Digits_Constraint_F_Digits := Digits_Constraint_F_Digits;
            Self.Digits_Constraint_F_Range := Digits_Constraint_F_Range;
      end Initialize_Fields_For_Digits_Constraint;
   function Digits_Constraint_F_Digits
     (Node : Bare_Digits_Constraint) return Bare_Expr
   is
   begin
         return Node.Digits_Constraint_F_Digits;
   end;
   function Digits_Constraint_F_Range
     (Node : Bare_Digits_Constraint) return Bare_Range_Spec
   is
   begin
         return Node.Digits_Constraint_F_Range;
   end;
--# property-start DigitsConstraint.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7371
pragma Warnings (Off, "is not referenced");
 function Digits_Constraint_P_Xref_Equation
  (Node : Bare_Digits_Constraint
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Digits_Constraint := Bare_Digits_Constraint (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Digits_Constraint :=
     Internal_Entity_Digits_Constraint'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_547 with Inline_Always;
            procedure Finalizer_Scope_2568 with Inline_Always;
      Var_Expr : Internal_Entity_Range_Spec;
Fld : Internal_Entity_Expr;
Fld_1 : Logic_Equation;
Fld_2 : Internal_Entity_Range_Spec;
Fld_3 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred : Logic_Equation;
            procedure Finalizer_Scope_547 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred);
            end Finalizer_Scope_547;
            procedure Finalizer_Scope_2568 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_2568;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5841 '<And at ast.py:7372>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7372
--# expr-start 5838 '<FieldAccess for sub_equation at ast.py:7372>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7372
--# expr-start 5837 '<FieldAccess for digits at ast.py:7372>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7372
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Digits_Constraint_F_Digits, Info => Ent.Info);
--# expr-done 5837
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 5838
--# expr-start 5839 '<FieldAccess for range at ast.py:7372>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7372
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Range_Spec (Node => Ent.Node.Digits_Constraint_F_Range, Info => Ent.Info);
--# expr-done 5839
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Range_Spec then
   --# scope-start
   --# bind range Var_Expr
--# expr-start 5840 '<FieldAccess for sub_equation at ast.py:7373>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7373
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 5840
Scope_Result := Fld_3;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2568;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7374") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred := Create_And (Fld_1, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7372") else null)); 
--# expr-done 5841
      Property_Result := And_Pred;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_547;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_547;
               Finalizer_Scope_2568;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Digits_Constraint_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Range_Constraint
   --
      procedure Initialize_Fields_For_Range_Constraint
        (Self : Bare_Range_Constraint
         ; Range_Constraint_F_Range : Bare_Range_Spec
        ) is
      begin
            Self.Range_Constraint_F_Range := Range_Constraint_F_Range;
      end Initialize_Fields_For_Range_Constraint;
   function Range_Constraint_F_Range
     (Node : Bare_Range_Constraint) return Bare_Range_Spec
   is
   begin
         return Node.Range_Constraint_F_Range;
   end;
--# property-start RangeConstraint.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7355
pragma Warnings (Off, "is not referenced");
 function Range_Constraint_P_Xref_Equation
  (Node : Bare_Range_Constraint
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Range_Constraint := Bare_Range_Constraint (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Range_Constraint :=
     Internal_Entity_Range_Constraint'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_546 with Inline_Always;
      Fld : Bare_Range_Spec;
Fld_1 : Bare_Expr;
Fld_2 : Logic_Var;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_5 : Internal_Entity_Range_Spec;
Fld_6 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_7 : Internal_Entity_Range_Spec;
Fld_8 : Internal_Entity_Expr;
Fld_9 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_546 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_9);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_546;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5855 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5854 '<And at ast.py:7356>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7356
--# expr-start 5850 '<And at ast.py:7356>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7356
--# expr-start 5847 '<Bind at ast.py:7357>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7357
--# expr-start 5844 '<FieldAccess for expected_type_var at ast.py:7357>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7357
--# expr-start 5843 '<FieldAccess for range at ast.py:7357>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7357
--# expr-start 5842 '<FieldAccess for range at ast.py:7357>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7357
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Range_Constraint_F_Range;
--# expr-done 5842
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Range_Spec_F_Range;
--# expr-done 5843
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_1);
--# expr-done 5844
Fld_2.Value := No_Entity;
Entity_Vars.Reset (Fld_2);
--# expr-start 5846 '<FieldAccess for base_subtype at ast.py:7358>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7358
--# expr-start 5845 '<FieldAccess for subtype at ast.py:7358>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7358
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Constraint.subtype
Fld_3 := Libadalang.Implementation.Constraint_P_Subtype (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5845
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.base_subtype
Fld_4 := Libadalang.Implementation.Base_Type_Decl_P_Base_Subtype (Node => Fld_3.Node, Origin => Origin, E_Info => Fld_3.Info);
--# end
--# expr-done 5846
      Cast_Result := Create_Internal_Entity
        (Node => Fld_4.Node,
         Info => Fld_4.Info);
Bind_Result := Solver.Create_Assign (Fld_2, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7357") else null)); 
--# expr-done 5847
--# expr-start 5849 '<FieldAccess for sub_equation at ast.py:7359>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7359
--# expr-start 5848 '<FieldAccess for range at ast.py:7359>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7359
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Range_Spec (Node => Ent.Node.Range_Constraint_F_Range, Info => Ent.Info);
--# expr-done 5848
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_6 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_5.Node, Env => Env, Origin => Origin, E_Info => Fld_5.Info);
--# end
--# expr-done 5849
And_Pred := Create_And (Bind_Result, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7356") else null)); 
--# expr-done 5850
--# expr-start 5853 '<FieldAccess for matches_expected_type at ast.py:7360>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7360
--# expr-start 5852 '<FieldAccess for range at ast.py:7360>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7360
--# expr-start 5851 '<FieldAccess for range at ast.py:7360>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7360
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Range_Spec (Node => Ent.Node.Range_Constraint_F_Range, Info => Ent.Info);
--# expr-done 5851
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Expr (Node => Fld_7.Node.Range_Spec_F_Range, Info => Fld_7.Info);
--# expr-done 5852
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_9 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_8.Node, Origin => Origin);
--# end
--# expr-done 5853
And_Pred_1 := Create_And (And_Pred, Fld_9, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7356") else null)); 
--# expr-done 5854
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 5855
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_546;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_546;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Range_Constraint_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Declarative_Part
   --
      procedure Initialize_Fields_For_Declarative_Part
        (Self : Bare_Declarative_Part
         ; Declarative_Part_F_Decls : Bare_Ada_Node_List
        ) is
      begin
            Self.Declarative_Part_F_Decls := Declarative_Part_F_Decls;
      end Initialize_Fields_For_Declarative_Part;
   function Declarative_Part_F_Decls
     (Node : Bare_Declarative_Part) return Bare_Ada_Node_List
   is
   begin
         return Node.Declarative_Part_F_Decls;
   end;
--# property-start DeclarativePart.use_clauses_envs /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9817
pragma Warnings (Off, "is not referenced");
 function Declarative_Part_P_Use_Clauses_Envs
  (Node : Bare_Declarative_Part
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Declarative_Part := Bare_Declarative_Part (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Declarative_Part :=
     Internal_Entity_Declarative_Part'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_784 with Inline_Always;
            procedure Finalizer_Scope_2424 with Inline_Always;
      Fld : Internal_Entity_Ada_Node_List;
Fld_1 : Internal_Entity_Array_Access;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Use_Clause;
Fld_2 : Lexical_Env := Empty_Env;
Is_A : Boolean;
Map_Result : Lexical_Env_Array_Access;
Group_Env : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_784 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Group_Env);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_784;
            procedure Finalizer_Scope_2424 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_2424;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5863 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5862 '<env_group at ast.py:9822>' Group_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9822
--# expr-start 5861 '<FilterMap at ast.py:9822>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9822
   --# expr-start 5857 '<children at ast.py:9822>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9822
--# expr-start 5856 '<FieldAccess for decls at ast.py:9822>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9822
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Ada_Node_List (Node => Ent.Node.Declarative_Part_F_Decls, Info => Ent.Info);
--# expr-done 5856
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Children (Node => Fld.Node, E_Info => Fld.Info);
--# expr-done 5857
      declare
         Map_Result_Vec : Lexical_Env_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Fld_1;
      begin
         for Item_210 of
               Collection.Items
         loop
   --# scope-start
               --# bind u Item_210
         --# expr-start 5858 '<IsA UseClause at ast.py:9824>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9824
Is_A := Item_210.Node /= null 
and then Item_210.Node.Kind in Ada_Use_Clause; 
--# expr-done 5858
         if Is_A then
      --# expr-start 5860 '<FieldAccess for used_envs at ast.py:9823>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9823
--# expr-start 5859 '<Cast to UseClause at ast.py:9823>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9823
Cast_Expr := Item_210; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Use_Clause
   then
      Cast_Result := Create_Internal_Entity_Use_Clause
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Use_Clause;
   end if;
--# expr-done 5859
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start UseClause.used_envs
Fld_2 := Libadalang.Implementation.Use_Clause_P_Used_Envs (Node => Cast_Result.Node, E_Info => Cast_Result.Info);
--# end
--# expr-done 5860
         declare
            Item_To_Append : constant Lexical_Env := Fld_2;
         begin
               Inc_Ref (Item_To_Append);
            Lexical_Env_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
      Finalizer_Scope_2424;
         end loop;
      end;
         Map_Result := Create_Lexical_Env_Array
           (Items_Count => Natural (Lexical_Env_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Lexical_Env_Vectors.Get
              (Map_Result_Vec,
               I + Lexical_Env_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Lexical_Env_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 5861
Group_Env := Group (Map_Result, No_Metadata); 
--# expr-done 5862
Let_Result := Group_Env; Inc_Ref (Let_Result);
--# end
--# expr-done 5863
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_784;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_784;
               Finalizer_Scope_2424;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Declarative_Part_P_Use_Clauses_Envs;
--# end
   --
   --  Primitives for Bare_Private_Part
   --
      procedure Private_Part_Pre_Env_Actions
        (Self            : Bare_Private_Part;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_45 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:9852");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_46 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            Libadalang.Implementation.Env_Names_47 (Node => 
Self);
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Private_Part
        (Self : Bare_Private_Part
         ; Declarative_Part_F_Decls : Bare_Ada_Node_List
        ) is
      begin
            Initialize_Fields_For_Declarative_Part
              (Self, Declarative_Part_F_Decls);
      end Initialize_Fields_For_Private_Part;
--# property-start PrivatePart.env_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9833
pragma Warnings (Off, "is not referenced");
 function Private_Part_P_Env_Names
  (Node : Bare_Private_Part
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Private_Part := Bare_Private_Part (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_785 with Inline_Always;
            procedure Finalizer_Scope_2569 with Inline_Always;
            procedure Finalizer_Scope_2570 with Inline_Always;
      Var_Expr : Bare_Base_Package_Decl;
Var_Expr_1 : String_Type;
Fld : Bare_Ada_Node;
Cast_Result : Bare_Base_Package_Decl;
Fld_1 : String_Type;
Str : String_Type;
Concat_Result : String_Type;
Sym : Symbol_Type;
Array_Lit : Symbol_Type_Array_Access;
Scope_Result : Symbol_Type_Array_Access;
Result_Var : Symbol_Type_Array_Access;
Scope_Result_1 : Symbol_Type_Array_Access;
Result_Var_1 : Symbol_Type_Array_Access;
Let_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_785 is
            begin
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_785;
            procedure Finalizer_Scope_2569 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_2569;
            procedure Finalizer_Scope_2570 is
            begin
                     Dec_Ref (Var_Expr_1);
                     Dec_Ref (Str);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Array_Lit);
            end Finalizer_Scope_2570;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5871 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5865 '<Cast to BasePackageDecl at ast.py:9839>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9839
--# expr-start 5864 '<parent at ast.py:9839>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9839
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 5864
   if Fld = null
      or else Fld.Kind in Ada_Base_Package_Decl
   then
      Cast_Result := Fld;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 5865
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind pkg Var_Expr
--# expr-start 5866 '<FieldAccess for top_level_env_name at ast.py:9840>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9840
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.top_level_env_name
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Top_Level_Env_Name (Node => Var_Expr);
--# end
--# expr-done 5866
Var_Expr_1 := Fld_1;
      Inc_Ref (Var_Expr_1);
if Var_Expr_1 /= Empty_String then
   --# scope-start
   --# bind name Var_Expr_1
--# expr-start 5870 '<ArrayLiteral at ast.py:9841>' Array_Lit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9841
--# expr-start 5869 '<to_symbol at ast.py:9842>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9842
--# expr-start 5868 '<Concat at ast.py:9842>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9842
--# expr-start 5867 '<String at ast.py:9842>' Str /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9842
Str := Create_String (".__privatepart"); 
--# expr-done 5867
Concat_Result := Concat_String (Var_Expr_1, Str); 
--# expr-done 5868
Sym := String_To_Symbol (Self.Unit.Context, Concat_Result); 
--# expr-done 5869
Array_Lit := Create_Symbol_Type_Array (Internal_Symbol_Type_Array'(1 => Sym)); 
--# expr-done 5870
Scope_Result := Array_Lit;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2570;
   Result_Var := Scope_Result;
else
   Result_Var := No_Symbol_Type_Array_Type;
end if;
      Inc_Ref (Result_Var);
Scope_Result_1 := Result_Var;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2569;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Symbol_Type_Array_Type;
end if;
      Inc_Ref (Result_Var_1);
Let_Result := Result_Var_1; Inc_Ref (Let_Result);
--# end
--# expr-done 5871
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_785;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_785;
               Finalizer_Scope_2569;
               Finalizer_Scope_2570;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Private_Part_P_Env_Names;
--# end
--# property-start PrivatePart.immediate_declarative_region /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9848
pragma Warnings (Off, "is not referenced");
 function Private_Part_P_Immediate_Declarative_Region
  (Node : Bare_Private_Part
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Private_Part := Bare_Private_Part (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Private_Part :=
     Internal_Entity_Private_Part'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_786 with Inline_Always;
      Fld : Internal_Entity;
Fld_1 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_786 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_786;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5874 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5873 '<FieldAccess for immediate_declarative_region at ast.py:9849>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9849
--# expr-start 5872 '<FieldAccess for semantic_parent at ast.py:9849>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9849
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.semantic_parent
Fld := Libadalang.Implementation.Ada_Node_P_Semantic_Parent (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5872
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.immediate_declarative_region'
Fld_1 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Immediate_Declarative_Region (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 5873
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 5874
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_786;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_786;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Private_Part_P_Immediate_Declarative_Region;
--# end
--# property-start PrivatePart._env_mappings_45 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9852
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_45
  (Node : Bare_Private_Part
  )
   return Internal_Env_Assoc
is
   Self : Bare_Private_Part := Bare_Private_Part (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1462 with Inline_Always;
      Sym : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Expr : Bare_Private_Part;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1462 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1462;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5877 '<New EnvAssoc at ast.py:9852>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9852
--# expr-start 5875 '<New DesignatedEnv at ast.py:9852>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9852
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 5875
--# expr-start 5876 '<Symbol __privatepart at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Privatepart); 
--# expr-done 5876
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Sym, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 5877
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1462;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1462;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_45;
--# end
--# property-start PrivatePart._env_trans_parent_46 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9851
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_46
  (Node : Bare_Private_Part
  )
   return Boolean
is
   Self : Bare_Private_Part := Bare_Private_Part (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_46;
--# end
--# property-start PrivatePart._env_names_47 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9853
pragma Warnings (Off, "is not referenced");
 function Env_Names_47
  (Node : Bare_Private_Part
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Private_Part := Bare_Private_Part (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1464 with Inline_Always;
      Fld : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1464 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1464;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5878 '<FieldAccess for env_names at ast.py:9853>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9853
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start PrivatePart.env_names
Fld := Libadalang.Implementation.Private_Part_P_Env_Names (Node => Self);
--# end
--# expr-done 5878
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1464;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1464;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Names_47;
--# end
   --
   --  Primitives for Bare_Public_Part
   --
      procedure Initialize_Fields_For_Public_Part
        (Self : Bare_Public_Part
         ; Declarative_Part_F_Decls : Bare_Ada_Node_List
        ) is
      begin
            Initialize_Fields_For_Declarative_Part
              (Self, Declarative_Part_F_Decls);
      end Initialize_Fields_For_Public_Part;
   --
   --  Primitives for Bare_Elsif_Expr_Part
   --
      procedure Initialize_Fields_For_Elsif_Expr_Part
        (Self : Bare_Elsif_Expr_Part
         ; Elsif_Expr_Part_F_Cond_Expr : Bare_Expr
         ; Elsif_Expr_Part_F_Then_Expr : Bare_Expr
        ) is
      begin
            Self.Elsif_Expr_Part_F_Cond_Expr := Elsif_Expr_Part_F_Cond_Expr;
            Self.Elsif_Expr_Part_F_Then_Expr := Elsif_Expr_Part_F_Then_Expr;
      end Initialize_Fields_For_Elsif_Expr_Part;
   function Elsif_Expr_Part_F_Cond_Expr
     (Node : Bare_Elsif_Expr_Part) return Bare_Expr
   is
   begin
         return Node.Elsif_Expr_Part_F_Cond_Expr;
   end;
   function Elsif_Expr_Part_F_Then_Expr
     (Node : Bare_Elsif_Expr_Part) return Bare_Expr
   is
   begin
         return Node.Elsif_Expr_Part_F_Then_Expr;
   end;
   --
   --  Primitives for Bare_Elsif_Stmt_Part
   --
      procedure Initialize_Fields_For_Elsif_Stmt_Part
        (Self : Bare_Elsif_Stmt_Part
         ; Elsif_Stmt_Part_F_Cond_Expr : Bare_Expr
         ; Elsif_Stmt_Part_F_Stmts : Bare_Stmt_List
        ) is
      begin
            Self.Elsif_Stmt_Part_F_Cond_Expr := Elsif_Stmt_Part_F_Cond_Expr;
            Self.Elsif_Stmt_Part_F_Stmts := Elsif_Stmt_Part_F_Stmts;
      end Initialize_Fields_For_Elsif_Stmt_Part;
   function Elsif_Stmt_Part_F_Cond_Expr
     (Node : Bare_Elsif_Stmt_Part) return Bare_Expr
   is
   begin
         return Node.Elsif_Stmt_Part_F_Cond_Expr;
   end;
   function Elsif_Stmt_Part_F_Stmts
     (Node : Bare_Elsif_Stmt_Part) return Bare_Stmt_List
   is
   begin
         return Node.Elsif_Stmt_Part_F_Stmts;
   end;
--# property-start ElsifStmtPart.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19359
pragma Warnings (Off, "is not referenced");
 function Elsif_Stmt_Part_P_Xref_Entry_Point
  (Node : Bare_Elsif_Stmt_Part
  )
   return Boolean
is
   Self : Bare_Elsif_Stmt_Part := Bare_Elsif_Stmt_Part (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Elsif_Stmt_Part_P_Xref_Entry_Point;
--# end
--# property-start ElsifStmtPart.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19362
pragma Warnings (Off, "is not referenced");
 function Elsif_Stmt_Part_P_Xref_Equation
  (Node : Bare_Elsif_Stmt_Part
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Elsif_Stmt_Part := Bare_Elsif_Stmt_Part (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Elsif_Stmt_Part :=
     Internal_Entity_Elsif_Stmt_Part'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1354 with Inline_Always;
      Fld : Bare_Expr;
Fld_1 : Logic_Var;
Fld_2 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_3 : Internal_Entity_Expr;
Fld_4 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_5 : Bare_Expr;
Fld_6 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1354 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_4);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1354;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5889 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5888 '<And at ast.py:19364>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19364
--# expr-start 5885 '<And at ast.py:19364>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19364
--# expr-start 5882 '<Bind at ast.py:19364>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19364
--# expr-start 5880 '<FieldAccess for expected_type_var at ast.py:19364>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19364
--# expr-start 5879 '<FieldAccess for cond_expr at ast.py:19364>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19364
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Elsif_Stmt_Part_F_Cond_Expr;
--# expr-done 5879
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld);
--# expr-done 5880
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 5881 '<FieldAccess for bool_type at ast.py:19364>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19364
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_2 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 5881
      Cast_Result := Create_Internal_Entity
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Bind_Result := Solver.Create_Assign (Fld_1, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19364") else null)); 
--# expr-done 5882
--# expr-start 5884 '<FieldAccess for sub_equation at ast.py:19365>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19365
--# expr-start 5883 '<FieldAccess for cond_expr at ast.py:19365>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19365
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Expr (Node => Ent.Node.Elsif_Stmt_Part_F_Cond_Expr, Info => Ent.Info);
--# expr-done 5883
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_4 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_3.Node, Env => Env, Origin => Origin, E_Info => Fld_3.Info);
--# end
--# expr-done 5884
And_Pred := Create_And (Bind_Result, Fld_4, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19364") else null)); 
--# expr-done 5885
--# expr-start 5887 '<FieldAccess for matches_expected_formal_prim_type at ast.py:19366>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19366
--# expr-start 5886 '<FieldAccess for cond_expr at ast.py:19366>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19366
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Self.Elsif_Stmt_Part_F_Cond_Expr;
--# expr-done 5886
      if Fld_5 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_prim_type
Fld_6 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Prim_Type (Node => Fld_5, Origin => Origin);
--# end
--# expr-done 5887
And_Pred_1 := Create_And (And_Pred, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19364") else null)); 
--# expr-done 5888
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 5889
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1354;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1354;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Elsif_Stmt_Part_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Expr
   --
      procedure Initialize_Fields_For_Expr
        (Self : Bare_Expr
        ) is
      begin
      Self.Expr_Logic_Vars := System.Null_Address;
      end Initialize_Fields_For_Expr;
--# property-start Expr.type_val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10974
pragma Warnings (Off, "is not referenced");
 function Expr_P_Type_Val
  (Node : Bare_Expr
  )
   return Internal_Entity
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity;
      Fld : Logic_Var;
Is_Logic_Var_Defined : Boolean;
Eq_Solution : Internal_Entity;
If_Result : Internal_Entity;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5891 '<get_value at ast.py:10974>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10974
--# expr-start 5890 '<FieldAccess for type_var at ast.py:10974>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10974
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 5890
Is_Logic_Var_Defined := Entity_Vars.Is_Defined (Fld); 
if Is_Logic_Var_Defined then
Eq_Solution := Entity_Vars.Get_Value (Fld); 
   If_Result := Eq_Solution;
else
   If_Result := No_Entity;
end if;
--# expr-done 5891
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Type_Val;
--# end
--# property-start Expr.expression_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10976
pragma Warnings (Off, "is not referenced");
 function Expr_P_Expression_Type
  (Node : Bare_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Expr :=
     Internal_Entity_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Type_Decl;
      Cast_Expr : Internal_Entity_Expr;
Cast_Result : Internal_Entity;
Fld : Logic_Var;
Fld_1 : Internal_Logic_Val_Result;
Fld_2 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5895 '<Cast to BaseTypeDecl at ast.py:10977>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10977
--# expr-start 5894 '<FieldAccess for value at ast.py:10977>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10977
--# expr-start 5893 '<FieldAccess for logic_val at ast.py:10977>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10977
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Cast_Expr := Ent; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
--# expr-start 5892 '<FieldAccess for type_var at ast.py:10977>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10977
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 5892
--# property-call-start AdaNode.logic_val
Fld_1 := Libadalang.Implementation.Ada_Node_P_Logic_Val (Node => Self, From_Node => Cast_Result, Lvar => Fld);
--# end
--# expr-done 5893
Fld_2 := Fld_1.Value;
--# expr-done 5894
   if Fld_2.Node = null
      or else Fld_2.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 5895
      Property_Result := Cast_Result_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Expression_Type;
--# end
--# property-start Expr.expected_expression_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10986
pragma Warnings (Off, "is not referenced");
 function Expr_P_Expected_Expression_Type
  (Node : Bare_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Expr :=
     Internal_Entity_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Type_Decl;
      Cast_Expr : Internal_Entity_Expr;
Cast_Result : Internal_Entity;
Fld : Logic_Var;
Fld_1 : Internal_Logic_Val_Result;
Fld_2 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5899 '<Cast to BaseTypeDecl at ast.py:10987>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10987
--# expr-start 5898 '<FieldAccess for value at ast.py:10987>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10987
--# expr-start 5897 '<FieldAccess for logic_val at ast.py:10987>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10987
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Cast_Expr := Ent; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
--# expr-start 5896 '<FieldAccess for expected_type_var at ast.py:10987>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10987
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 5896
--# property-call-start AdaNode.logic_val
Fld_1 := Libadalang.Implementation.Ada_Node_P_Logic_Val (Node => Self, From_Node => Cast_Result, Lvar => Fld);
--# end
--# expr-done 5897
Fld_2 := Fld_1.Value;
--# expr-done 5898
   if Fld_2.Node = null
      or else Fld_2.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 5899
      Property_Result := Cast_Result_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Expected_Expression_Type;
--# end
--# property-start Expr.xref_stop_resolution /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:10996
pragma Warnings (Off, "is not referenced");
 function Expr_P_Xref_Stop_Resolution
  (Node : Bare_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Var_Expr : Bare_For_Loop_Spec;
Fld : Bare_Ada_Node;
Cast_Result : Bare_For_Loop_Spec;
Fld_1 : Bare_Expr;
Cast_Result_1 : Bare_Expr;
Is_Equal : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 5901 '<Cast to ForLoopSpec at ast.py:11001>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11001
--# expr-start 5900 '<parent at ast.py:11001>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11001
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 5900
   if Fld = null
      or else Fld.Kind in Ada_For_Loop_Spec_Range
   then
      Cast_Result := Fld;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 5901
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind spec Var_Expr
--# expr-start 5903 '<Eq at ast.py:11002>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11002
--# expr-start 5902 '<FieldAccess for iter_filter at ast.py:11002>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11002
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Var_Expr.For_Loop_Spec_F_Iter_Filter;
--# expr-done 5902
      Cast_Result_1 := Fld_1;
Is_Equal := Cast_Result_1 = Self; 
--# expr-done 5903
Scope_Result := Is_Equal;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
      Property_Result := Result_Var;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Xref_Stop_Resolution;
--# end
--# property-start '[dispatcher]Expr.has_context_free_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Expr_P_Has_Context_Free_Type
  (Node : Bare_Expr
  )
   return Boolean
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Expr (Self.Kind) is
               when Ada_Abstract_State_Decl_Expr | Ada_Box_Expr .. Ada_Case_Expr_Alternative | Ada_Contract_Cases | Ada_Membership_Expr | Ada_Call_Expr .. Ada_Discrete_Subtype_Name | Ada_End_Name .. Ada_Reduce_Attribute_Ref | Ada_Identifier .. Ada_Op_Xor | Ada_Int_Literal .. Ada_Target_Name | Ada_Quantified_Expr =>
                  --# property-call-start Expr.has_context_free_type
                  Property_Result := Expr_P_Has_Context_Free_Type
                    (Self
                    );
                  --# end
               when Ada_Allocator =>
                  --# property-call-start Allocator.has_context_free_type
                  Property_Result := Allocator_P_Has_Context_Free_Type
                    (Self
                    );
                  --# end
               when Ada_Aggregate .. Ada_Null_Record_Aggregate =>
                  --# property-call-start BaseAggregate.has_context_free_type
                  Property_Result := Base_Aggregate_P_Has_Context_Free_Type
                    (Self
                    );
                  --# end
               when Ada_Bin_Op =>
                  --# property-call-start BinOp.has_context_free_type
                  Property_Result := Bin_Op_P_Has_Context_Free_Type
                    (Self
                    );
                  --# end
               when Ada_Relation_Op =>
                  --# property-call-start RelationOp.has_context_free_type
                  Property_Result := Relation_Op_P_Has_Context_Free_Type
                    (Self
                    );
                  --# end
               when Ada_Case_Expr =>
                  --# property-call-start CaseExpr.has_context_free_type
                  Property_Result := Case_Expr_P_Has_Context_Free_Type
                    (Self
                    );
                  --# end
               when Ada_If_Expr =>
                  --# property-call-start IfExpr.has_context_free_type
                  Property_Result := If_Expr_P_Has_Context_Free_Type
                    (Self
                    );
                  --# end
               when Ada_Decl_Expr =>
                  --# property-call-start DeclExpr.has_context_free_type
                  Property_Result := Decl_Expr_P_Has_Context_Free_Type
                    (Self
                    );
                  --# end
               when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref =>
                  --# property-call-start AttributeRef.has_context_free_type
                  Property_Result := Attribute_Ref_P_Has_Context_Free_Type
                    (Self
                    );
                  --# end
               when Ada_Dotted_Name =>
                  --# property-call-start DottedName.has_context_free_type
                  Property_Result := Dotted_Name_P_Has_Context_Free_Type
                    (Self
                    );
                  --# end
               when Ada_Char_Literal =>
                  --# property-call-start CharLiteral.has_context_free_type
                  Property_Result := Char_Literal_P_Has_Context_Free_Type
                    (Self
                    );
                  --# end
               when Ada_String_Literal =>
                  --# property-call-start StringLiteral.has_context_free_type
                  Property_Result := String_Literal_P_Has_Context_Free_Type
                    (Self
                    );
                  --# end
               when Ada_Null_Literal =>
                  --# property-call-start NullLiteral.has_context_free_type
                  Property_Result := Null_Literal_P_Has_Context_Free_Type
                    (Self
                    );
                  --# end
               when Ada_Paren_Expr =>
                  --# property-call-start ParenExpr.has_context_free_type
                  Property_Result := Paren_Expr_P_Has_Context_Free_Type
                    (Self
                    );
                  --# end
               when Ada_Raise_Expr =>
                  --# property-call-start RaiseExpr.has_context_free_type
                  Property_Result := Raise_Expr_P_Has_Context_Free_Type
                    (Self
                    );
                  --# end
               when Ada_Un_Op =>
                  --# property-call-start UnOp.has_context_free_type
                  Property_Result := Un_Op_P_Has_Context_Free_Type
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Expr_P_Has_Context_Free_Type;
--# end
--# property-start Expr.matches_expected_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11022
pragma Warnings (Off, "is not referenced");
 function Expr_P_Matches_Expected_Type
  (Node : Bare_Expr
      ; Origin : Bare_Ada_Node
  )
   return Logic_Equation
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_867 with Inline_Always;
      Fld : Logic_Var;
Fld_1 : Logic_Var;
Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_867 is
            begin
                     Dec_Ref (Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_867;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5907 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5906 '<Predicate on BaseTypeDecl.matching_type at ast.py:11023>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11023
--# expr-start 5904 '<FieldAccess for type_var at ast.py:11025>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11025
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 5904
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
--# expr-start 5905 '<FieldAccess for expected_type_var at ast.py:11026>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11026
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 5905
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
Pred := Solver.Create_N_Predicate ((Fld, Fld_1), Create_Base_Type_Decl_P_Matching_Type_0_Predicate (Origin)); 
--# expr-done 5906
Let_Result := Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 5907
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_867;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_867;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Matches_Expected_Type;
--# end
--# property-start Expr.matches_expected_assign_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11030
pragma Warnings (Off, "is not referenced");
 function Expr_P_Matches_Expected_Assign_Type
  (Node : Bare_Expr
      ; Origin : Bare_Ada_Node
  )
   return Logic_Equation
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_868 with Inline_Always;
      Fld : Logic_Var;
Fld_1 : Logic_Var;
Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_868 is
            begin
                     Dec_Ref (Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_868;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5911 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5910 '<Predicate on BaseTypeDecl.matching_assign_type at ast.py:11031>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11031
--# expr-start 5908 '<FieldAccess for type_var at ast.py:11033>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11033
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 5908
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
--# expr-start 5909 '<FieldAccess for expected_type_var at ast.py:11034>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11034
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 5909
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
Pred := Solver.Create_N_Predicate ((Fld, Fld_1), Create_Base_Type_Decl_P_Matching_Assign_Type_0_Predicate (Origin)); 
--# expr-done 5910
Let_Result := Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 5911
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_868;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_868;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Matches_Expected_Assign_Type;
--# end
--# property-start Expr.matches_expected_formal_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11038
pragma Warnings (Off, "is not referenced");
 function Expr_P_Matches_Expected_Formal_Type
  (Node : Bare_Expr
      ; Origin : Bare_Ada_Node
  )
   return Logic_Equation
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_869 with Inline_Always;
      Fld : Logic_Var;
Fld_1 : Logic_Var;
Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_869 is
            begin
                     Dec_Ref (Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_869;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5915 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5914 '<Predicate on BaseTypeDecl.matching_formal_type at ast.py:11039>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11039
--# expr-start 5912 '<FieldAccess for type_var at ast.py:11041>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11041
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 5912
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
--# expr-start 5913 '<FieldAccess for expected_type_var at ast.py:11042>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11042
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 5913
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
Pred := Solver.Create_N_Predicate ((Fld, Fld_1), Create_Base_Type_Decl_P_Matching_Formal_Type_0_Predicate (Origin)); 
--# expr-done 5914
Let_Result := Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 5915
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_869;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_869;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Matches_Expected_Formal_Type;
--# end
--# property-start Expr.matches_expected_formal_prim_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11046
pragma Warnings (Off, "is not referenced");
 function Expr_P_Matches_Expected_Formal_Prim_Type
  (Node : Bare_Expr
      ; Origin : Bare_Ada_Node
  )
   return Logic_Equation
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_870 with Inline_Always;
      Fld : Logic_Var;
Fld_1 : Logic_Var;
Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_870 is
            begin
                     Dec_Ref (Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_870;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5919 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5918 '<Predicate on BaseTypeDecl.matching_formal_prim_type at ast.py:11047>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11047
--# expr-start 5916 '<FieldAccess for type_var at ast.py:11049>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11049
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 5916
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
--# expr-start 5917 '<FieldAccess for expected_type_var at ast.py:11050>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11050
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 5917
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
Pred := Solver.Create_N_Predicate ((Fld, Fld_1), Create_Base_Type_Decl_P_Matching_Formal_Prim_Type_0_Predicate (Origin)); 
--# expr-done 5918
Let_Result := Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 5919
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_870;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_870;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Matches_Expected_Formal_Prim_Type;
--# end
--# property-start Expr.matches_expected_prefix_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11054
pragma Warnings (Off, "is not referenced");
 function Expr_P_Matches_Expected_Prefix_Type
  (Node : Bare_Expr
      ; Origin : Bare_Ada_Node
  )
   return Logic_Equation
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_871 with Inline_Always;
      Fld : Logic_Var;
Fld_1 : Logic_Var;
Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_871 is
            begin
                     Dec_Ref (Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_871;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5923 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5922 '<Predicate on BaseTypeDecl.matching_prefix_type at ast.py:11055>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11055
--# expr-start 5920 '<FieldAccess for type_var at ast.py:11057>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11057
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 5920
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
--# expr-start 5921 '<FieldAccess for expected_type_var at ast.py:11058>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11058
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 5921
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
Pred := Solver.Create_N_Predicate ((Fld, Fld_1), Create_Base_Type_Decl_P_Matching_Prefix_Type_0_Predicate (Origin)); 
--# expr-done 5922
Let_Result := Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 5923
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_871;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_871;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Matches_Expected_Prefix_Type;
--# end
--# property-start Expr.is_dynamically_tagged /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11063
pragma Warnings (Off, "is not referenced");
 function Expr_P_Is_Dynamically_Tagged
  (Node : Bare_Expr
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Expr :=
     Internal_Entity_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
            procedure Finalizer_Scope_2576 with Inline_Always;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Match_0 : Internal_Entity_Qual_Expr;
Match_1 : Internal_Entity_Name;
Match_2 : Internal_Entity_Cond_Expr;
Match_3 : Internal_Entity_Decl_Expr;
Match_4 : Internal_Entity_Paren_Expr;
Match_5 : Internal_Entity_Expr;
Var_Expr_1 : Internal_Entity_Base_Subp_Spec;
Fld : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Boolean;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Internal_Entity_Base_Type_Decl;
Fld_5 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result : Boolean;
Fld_6 : Internal_Entity_Expr;
Fld_7 : Boolean;
Fld_8 : Boolean;
Fld_9 : Internal_Entity_Base_Formal_Param_Holder;
Cast_Result : Internal_Entity_Base_Subp_Spec;
Fld_10 : Boolean;
Fld_11 : Boolean;
If_Result_1 : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
If_Result_2 : Boolean;
Fld_12 : Internal_Entity_Expr_Array_Access;
Fld_13 : Boolean;
Quantifier_Result : Boolean;
Fld_14 : Internal_Entity_Expr;
Fld_15 : Boolean;
Fld_16 : Internal_Entity_Expr;
Fld_17 : Boolean;
Match_Prefix : Internal_Entity_Expr;
Cast_Result_1 : Internal_Entity_Qual_Expr;
Let_Result : Boolean;
Scope_Result_2 : Boolean;
Cast_Result_2 : Internal_Entity_Name;
Let_Result_1 : Boolean;
Scope_Result_3 : Boolean;
Cast_Result_3 : Internal_Entity_Cond_Expr;
Let_Result_2 : Boolean;
Scope_Result_4 : Boolean;
Cast_Result_4 : Internal_Entity_Decl_Expr;
Let_Result_3 : Boolean;
Scope_Result_5 : Boolean;
Cast_Result_5 : Internal_Entity_Paren_Expr;
Let_Result_4 : Boolean;
Scope_Result_6 : Boolean;
Cast_Result_6 : Internal_Entity_Expr;
Let_Result_5 : Boolean;
Scope_Result_7 : Boolean;
Match_Result : Boolean;
If_Result_3 : Boolean;
Dyn_Var_Bind_Result : Boolean;
Let_Result_6 : Boolean;
            procedure Finalizer_Scope_2576 is
            begin
                     Dec_Ref (Fld_12);
            end Finalizer_Scope_2576;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5946 '<Block at ???>' Let_Result_6 None
--# scope-start
--# expr-start 5945 '<bind at ast.py:11069>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11069
--# expr-start 5924 '<FieldAccess for origin_node at ast.py:11069>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11069
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 5924
Bound_Origin := Fld; 
--# expr-start 5926 '<FieldAccess for is_classwide at ast.py:11070>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11070
--# expr-start 5925 '<FieldAccess for expression_type at ast.py:11070>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11070
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.expression_type
Fld_1 := Libadalang.Implementation.Expr_P_Expression_Type (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5925
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_classwide'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Classwide (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 5926
if Fld_2 then
   If_Result := True;
else
--# expr-start 5928 '<FieldAccess for accessed_type at ast.py:11071>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11071
--# expr-start 5927 '<FieldAccess for expression_type at ast.py:11071>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11071
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.expression_type
Fld_3 := Libadalang.Implementation.Expr_P_Expression_Type (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5927
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_4 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Fld_3.Node, Origin => Bound_Origin, E_Info => Fld_3.Info);
--# end
--# expr-done 5928
Var_Expr := Fld_4;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 5929 '<FieldAccess for is_classwide at ast.py:11071>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11071
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_classwide'
Fld_5 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Classwide (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 5929
Scope_Result := Fld_5;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
   If_Result := Result_Var;
end if;
if If_Result then
   If_Result_3 := True;
else
   --# expr-start 5944 '<Match at ast.py:11073>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11073
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Expr (Match_Prefix.Node.Kind) is
         when Ada_Qual_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Qual_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_1; 
--# bind qual_expr Match_0
--# expr-start 5931 '<FieldAccess for is_dynamically_tagged at ast.py:11075>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11075
--# expr-start 5930 '<FieldAccess for suffix at ast.py:11075>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11075
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Expr (Node => Match_0.Node.Qual_Expr_F_Suffix, Info => Match_0.Info);
--# expr-done 5930
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_dynamically_tagged
Fld_7 := Libadalang.Implementation.Expr_P_Is_Dynamically_Tagged (Node => Fld_6.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_6.Info);
--# end
--# expr-done 5931
Let_Result := Fld_7; 
--# end
Scope_Result_2 := Let_Result;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Attribute_Ref .. Ada_Explicit_Deref | Ada_Reduce_Attribute_Ref .. Ada_Target_Name =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_2; 
--# bind n Match_1
--# expr-start 5932 '<FieldAccess for is_direct_call at ast.py:11081>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11081
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_direct_call
Fld_8 := Libadalang.Implementation.Name_P_Is_Direct_Call (Node => Match_1.Node, E_Info => Match_1.Info);
--# end
--# expr-done 5932
if Fld_8 then
--# expr-start 5934 '<Cast to BaseSubpSpec at ast.py:11082>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11082
--# expr-start 5933 '<FieldAccess for called_subp_spec at ast.py:11082>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11082
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.called_subp_spec
Fld_9 := Libadalang.Implementation.Name_P_Called_Subp_Spec (Node => Match_1.Node, E_Info => Match_1.Info);
--# end
--# expr-done 5933
   if Fld_9.Node = null
      or else Fld_9.Node.Kind in Ada_Base_Subp_Spec
   then
      Cast_Result := Create_Internal_Entity_Base_Subp_Spec
        (Node => Fld_9.Node,
         Info => Fld_9.Info);
   else
         Cast_Result := No_Entity_Base_Subp_Spec;
   end if;
--# expr-done 5934
Var_Expr_1 := Cast_Result;
if Var_Expr_1 /= No_Entity_Base_Subp_Spec then
   --# scope-start
   --# bind spec Var_Expr_1
--# expr-start 5935 '<FieldAccess for has_controlling_result at ast.py:11084>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11084
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.has_controlling_result
Fld_10 := Libadalang.Implementation.Base_Subp_Spec_P_Has_Controlling_Result (Node => Var_Expr_1.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 5935
if Fld_10 then
   --# expr-start 5936 '<FieldAccess for has_dynamic_controlling_operand at ast.py:11085>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11085
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.has_dynamic_controlling_operand
Fld_11 := Libadalang.Implementation.Expr_P_Has_Dynamic_Controlling_Operand (Node => Match_1.Node, Spec => Var_Expr_1, Imprecise_Fallback => Imprecise_Fallback, E_Info => Match_1.Info);
--# end
--# expr-done 5936
   If_Result_1 := Fld_11;
else
   If_Result_1 := False;
end if;
Scope_Result_1 := If_Result_1;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
   If_Result_2 := Result_Var_1;
else
   If_Result_2 := False;
end if;
Let_Result_1 := If_Result_2; 
--# end
Scope_Result_3 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_3; 
         when Ada_Case_Expr .. Ada_If_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Cond_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_3; 
--# bind cond_expr Match_2
--# expr-start 5939 '<AllQuantifier at ast.py:11089>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11089
   --# expr-start 5937 '<FieldAccess for dependent_exprs at ast.py:11089>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11089
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]CondExpr.dependent_exprs'
Fld_12 := Libadalang.Implementation.Dispatcher_Cond_Expr_P_Dependent_Exprs (Node => Match_2.Node, E_Info => Match_2.Info);
--# end
--# expr-done 5937
      Quantifier_Result := True;
      declare
         Collection : constant Internal_Entity_Expr_Array_Access := Fld_12;
      begin
         for Item_211 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_211
      --# expr-start 5938 '<FieldAccess for is_dynamically_tagged at ast.py:11090>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11090
      if Item_211.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_dynamically_tagged
Fld_13 := Libadalang.Implementation.Expr_P_Is_Dynamically_Tagged (Node => Item_211.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Item_211.Info);
--# end
--# expr-done 5938
      Quantifier_Result := Fld_13;
   --# end
               exit when not Quantifier_Result;
         end loop;
      end;
--# expr-done 5939
Let_Result_2 := Quantifier_Result; 
--# end
Scope_Result_4 := Let_Result_2;
   --# end
      Finalizer_Scope_2576;
            Match_Result := Scope_Result_4; 
         when Ada_Decl_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Decl_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_4; 
--# bind decl_expr Match_3
--# expr-start 5941 '<FieldAccess for is_dynamically_tagged at ast.py:11094>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11094
--# expr-start 5940 '<FieldAccess for expr at ast.py:11094>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11094
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Create_Internal_Entity_Expr (Node => Match_3.Node.Decl_Expr_F_Expr, Info => Match_3.Info);
--# expr-done 5940
      if Fld_14.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_dynamically_tagged
Fld_15 := Libadalang.Implementation.Expr_P_Is_Dynamically_Tagged (Node => Fld_14.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_14.Info);
--# end
--# expr-done 5941
Let_Result_3 := Fld_15; 
--# end
Scope_Result_5 := Let_Result_3;
   --# end
            Match_Result := Scope_Result_5; 
         when Ada_Paren_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Create_Internal_Entity_Paren_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_4 := Cast_Result_5; 
--# bind paren_expr Match_4
--# expr-start 5943 '<FieldAccess for is_dynamically_tagged at ast.py:11097>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11097
--# expr-start 5942 '<FieldAccess for expr at ast.py:11097>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11097
      if Match_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_16 := Create_Internal_Entity_Expr (Node => Match_4.Node.Paren_Expr_F_Expr, Info => Match_4.Info);
--# expr-done 5942
      if Fld_16.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_dynamically_tagged
Fld_17 := Libadalang.Implementation.Expr_P_Is_Dynamically_Tagged (Node => Fld_16.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_16.Info);
--# end
--# expr-done 5943
Let_Result_4 := Fld_17; 
--# end
Scope_Result_6 := Let_Result_4;
   --# end
            Match_Result := Scope_Result_6; 
         when Ada_Abstract_State_Decl_Expr .. Ada_Case_Expr_Alternative | Ada_Contract_Cases | Ada_Membership_Expr | Ada_Quantified_Expr .. Ada_Un_Op =>
   --# scope-start
--# scope-start
      Cast_Result_6 := Create_Internal_Entity_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_5 := Cast_Result_6; 
--# bind _ Match_5
Let_Result_5 := False; 
--# end
Scope_Result_7 := Let_Result_5;
   --# end
            Match_Result := Scope_Result_7; 
end case;
--# expr-done 5944
   If_Result_3 := Match_Result;
end if;
Dyn_Var_Bind_Result := If_Result_3; 
--# expr-done 5945
Let_Result_6 := Dyn_Var_Bind_Result; 
--# end
--# expr-done 5946
      Property_Result := Let_Result_6;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_2576;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Is_Dynamically_Tagged;
--# end
--# property-start '[dispatcher]Expr.potential_actuals_for_dispatch' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Expr_P_Potential_Actuals_For_Dispatch
  (Node : Bare_Expr
      ; Spec : Internal_Entity_Base_Subp_Spec
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Expected_Type_For_Expr_Array_Access
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind spec Spec
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Expected_Type_For_Expr_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Expr (Self.Kind) is
               when Ada_Abstract_State_Decl_Expr .. Ada_Null_Record_Aggregate | Ada_Box_Expr .. Ada_Membership_Expr | Ada_Paren_Expr .. Ada_Raise_Expr =>
                  --# property-call-start Expr.potential_actuals_for_dispatch
                  Property_Result := Expr_P_Potential_Actuals_For_Dispatch
                    (Self
                        , Spec
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Bin_Op .. Ada_Relation_Op =>
                  --# property-call-start BinOp.potential_actuals_for_dispatch
                  Property_Result := Bin_Op_P_Potential_Actuals_For_Dispatch
                    (Self
                        , Spec
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Attribute_Ref .. Ada_Target_Name =>
                  --# property-call-start Name.potential_actuals_for_dispatch
                  Property_Result := Name_P_Potential_Actuals_For_Dispatch
                    (Self
                        , Spec
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Un_Op =>
                  --# property-call-start UnOp.potential_actuals_for_dispatch
                  Property_Result := Un_Op_P_Potential_Actuals_For_Dispatch
                    (Self
                        , Spec
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Expr_P_Potential_Actuals_For_Dispatch;
--# end
--# property-start Expr.has_dynamic_controlling_operand /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11122
pragma Warnings (Off, "is not referenced");
 function Expr_P_Has_Dynamic_Controlling_Operand
  (Node : Bare_Expr
      ; Spec : Internal_Entity_Base_Subp_Spec
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Expr :=
     Internal_Entity_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind spec Spec
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
            procedure Finalizer_Scope_874 with Inline_Always;
      Fld : Internal_Expected_Type_For_Expr_Array_Access;
Fld_1 : Internal_Entity_Type_Expr;
Fld_2 : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
Not_Val : Boolean;
Fld_3 : Internal_Entity_Expr;
Fld_4 : Boolean;
If_Result : Boolean;
Quantifier_Result : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_874 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_874;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5955 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5954 '<AnyQuantifier at ast.py:11133>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11133
   --# expr-start 5947 '<FieldAccess for potential_actuals_for_dispatch at ast.py:11133>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11133
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.potential_actuals_for_dispatch'
Fld := Libadalang.Implementation.Dispatcher_Expr_P_Potential_Actuals_For_Dispatch (Node => Ent.Node, Spec => Spec, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 5947
      Quantifier_Result := False;
      declare
         Collection : constant Internal_Expected_Type_For_Expr_Array_Access := Fld;
      begin
         for Item_212 of
               Collection.Items
         loop
   --# scope-start
               --# bind c Item_212
--# expr-start 5951 '<Not at ast.py:11135>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11135
--# expr-start 5950 '<IsNull at ast.py:11135>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11135
--# expr-start 5949 '<FieldAccess for get_candidate_type_for_primitive at ast.py:11135>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11135
      if Spec.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 5948 '<FieldAccess for expected_type at ast.py:11136>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11136
Fld_1 := Item_212.Expected_Type;
--# expr-done 5948
--# property-call-start BaseSubpSpec.get_candidate_type_for_primitive
Fld_2 := Libadalang.Implementation.Base_Subp_Spec_P_Get_Candidate_Type_For_Primitive (Node => Spec.Node, Type_Expr => Fld_1);
--# end
--# expr-done 5949
Is_Null := Fld_2.Node = null; 
--# expr-done 5950
Not_Val := not (Is_Null); 
--# expr-done 5951
if Not_Val then
   --# expr-start 5953 '<FieldAccess for is_dynamically_tagged at ast.py:11137>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11137
--# expr-start 5952 '<FieldAccess for expr at ast.py:11137>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11137
Fld_3 := Item_212.Expr;
--# expr-done 5952
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_dynamically_tagged
Fld_4 := Libadalang.Implementation.Expr_P_Is_Dynamically_Tagged (Node => Fld_3.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_3.Info);
--# end
--# expr-done 5953
   If_Result := Fld_4;
else
   If_Result := False;
end if;
      Quantifier_Result := If_Result;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
--# expr-done 5954
Let_Result := Quantifier_Result; 
--# end
--# expr-done 5955
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_874;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_874;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Has_Dynamic_Controlling_Operand;
--# end
--# property-start Expr.parent_candidate_dispatching_call /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11142
pragma Warnings (Off, "is not referenced");
 function Expr_P_Parent_Candidate_Dispatching_Call
  (Node : Bare_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Expr
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Expr :=
     Internal_Entity_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Expr;
            procedure Finalizer_Scope_2586 with Inline_Always;
      Match_0 : Internal_Entity_Call_Expr;
Match_1 : Internal_Entity_Dotted_Name;
Match_2 : Internal_Entity_Un_Op;
Match_3 : Internal_Entity_Bin_Op;
Match_4 : Internal_Entity_Cond_Expr;
Match_5 : Internal_Entity_Paren_Expr;
Match_6 : Internal_Entity_Qual_Expr;
Match_7 : Internal_Entity_Decl_Expr;
Match_8 : Internal_Entity_Param_Assoc;
Match_9 : Internal_Entity;
Fld : Internal_Entity;
Fld_1 : Internal_Entity_Expr;
Cast_Expr : Internal_Entity_Call_Expr;
Cast_Result : Internal_Entity_Expr;
Fld_2 : Internal_Entity_Name;
Cast_Expr_1 : Internal_Entity_Expr;
Cast_Result_1 : Internal_Entity;
Cast_Result_2 : Internal_Entity;
Is_Equiv : Boolean;
If_Result : Internal_Entity_Expr;
Fld_3 : Internal_Entity_Expr;
Cast_Expr_2 : Internal_Entity_Dotted_Name;
Cast_Result_3 : Internal_Entity_Expr;
Fld_4 : Internal_Entity_Base_Id;
Cast_Expr_3 : Internal_Entity_Expr;
Cast_Result_4 : Internal_Entity;
Cast_Result_5 : Internal_Entity;
Is_Equiv_1 : Boolean;
If_Result_1 : Internal_Entity_Expr;
Fld_5 : Internal_Entity_Expr;
Cast_Expr_4 : Internal_Entity_Un_Op;
Cast_Result_6 : Internal_Entity_Expr;
Fld_6 : Internal_Entity_Op;
Cast_Expr_5 : Internal_Entity_Expr;
Cast_Result_7 : Internal_Entity;
Cast_Result_8 : Internal_Entity;
Is_Equiv_2 : Boolean;
If_Result_2 : Internal_Entity_Expr;
Fld_7 : Internal_Entity_Expr;
Cast_Expr_6 : Internal_Entity_Bin_Op;
Cast_Result_9 : Internal_Entity_Expr;
Fld_8 : Internal_Entity_Op;
Cast_Expr_7 : Internal_Entity_Expr;
Cast_Result_10 : Internal_Entity;
Cast_Result_11 : Internal_Entity;
Is_Equiv_3 : Boolean;
If_Result_3 : Internal_Entity_Expr;
Fld_9 : Internal_Entity_Expr;
Fld_10 : Internal_Entity_Expr_Array_Access;
Cast_Expr_8 : Internal_Entity_Expr;
Cast_Result_12 : Internal_Entity;
Cast_Expr_9 : Internal_Entity_Expr;
Cast_Result_13 : Internal_Entity;
Is_Equiv_4 : Boolean;
Quantifier_Result : Boolean;
If_Result_4 : Internal_Entity_Expr;
Fld_11 : Internal_Entity_Expr;
Fld_12 : Internal_Entity_Expr;
Fld_13 : Internal_Entity_Expr;
Fld_14 : Internal_Entity;
Fld_15 : Internal_Entity;
Cast_Result_14 : Internal_Entity_Call_Expr;
Match_Prefix : Internal_Entity;
Cast_Result_15 : Internal_Entity_Call_Expr;
Let_Result : Internal_Entity_Expr;
Scope_Result : Internal_Entity_Expr;
Cast_Result_16 : Internal_Entity_Dotted_Name;
Let_Result_1 : Internal_Entity_Expr;
Scope_Result_1 : Internal_Entity_Expr;
Cast_Result_17 : Internal_Entity_Un_Op;
Let_Result_2 : Internal_Entity_Expr;
Scope_Result_2 : Internal_Entity_Expr;
Cast_Result_18 : Internal_Entity_Bin_Op;
Let_Result_3 : Internal_Entity_Expr;
Scope_Result_3 : Internal_Entity_Expr;
Cast_Result_19 : Internal_Entity_Cond_Expr;
Let_Result_4 : Internal_Entity_Expr;
Scope_Result_4 : Internal_Entity_Expr;
Cast_Result_20 : Internal_Entity_Paren_Expr;
Let_Result_5 : Internal_Entity_Expr;
Scope_Result_5 : Internal_Entity_Expr;
Cast_Result_21 : Internal_Entity_Qual_Expr;
Let_Result_6 : Internal_Entity_Expr;
Scope_Result_6 : Internal_Entity_Expr;
Cast_Result_22 : Internal_Entity_Decl_Expr;
Let_Result_7 : Internal_Entity_Expr;
Scope_Result_7 : Internal_Entity_Expr;
Cast_Result_23 : Internal_Entity_Param_Assoc;
Cast_Result_24 : Internal_Entity_Expr;
Let_Result_8 : Internal_Entity_Expr;
Scope_Result_8 : Internal_Entity_Expr;
Cast_Result_25 : Internal_Entity;
Let_Result_9 : Internal_Entity_Expr;
Scope_Result_9 : Internal_Entity_Expr;
Match_Result : Internal_Entity_Expr;
Let_Result_10 : Internal_Entity_Expr;
            procedure Finalizer_Scope_2586 is
            begin
                     Dec_Ref (Fld_10);
            end Finalizer_Scope_2586;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 5985 '<Block at ???>' Let_Result_10 None
--# scope-start
--# expr-start 5984 '<Match at ast.py:11148>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11148
--# expr-start 5956 '<parent at ast.py:11148>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11148
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 5956
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Call_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_15 := Create_Internal_Entity_Call_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_15; 
--# bind ce Match_0
--# expr-start 5960 '<If at ast.py:11149>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11149
--# expr-start 5958 '<Eq at ast.py:11150>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11150
Cast_Expr_1 := Ent; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
--# expr-start 5957 '<FieldAccess for name at ast.py:11150>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11150
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Match_0.Node.Call_Expr_F_Name, Info => Match_0.Info);
--# expr-done 5957
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Is_Equiv := Equivalent (Cast_Result_1, Cast_Result_2); 
--# expr-done 5958
if Is_Equiv then
   --# expr-start 5959 '<FieldAccess for parent_candidate_dispatching_call at ast.py:11151>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11151
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.parent_candidate_dispatching_call
Fld_1 := Libadalang.Implementation.Expr_P_Parent_Candidate_Dispatching_Call (Node => Match_0.Node, E_Info => Match_0.Info);
--# end
--# expr-done 5959
   If_Result := Fld_1;
else
Cast_Expr := Match_0; 
      Cast_Result := Create_Internal_Entity_Expr
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   If_Result := Cast_Result;
end if;
--# expr-done 5960
Let_Result := If_Result; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Dotted_Name =>
   --# scope-start
--# scope-start
      Cast_Result_16 := Create_Internal_Entity_Dotted_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_16; 
--# bind dn Match_1
--# expr-start 5964 '<If at ast.py:11154>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11154
--# expr-start 5962 '<Eq at ast.py:11155>' Is_Equiv_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11155
Cast_Expr_3 := Ent; 
      Cast_Result_4 := Create_Internal_Entity
        (Node => Cast_Expr_3.Node,
         Info => Cast_Expr_3.Info);
--# expr-start 5961 '<FieldAccess for suffix at ast.py:11155>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11155
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Base_Id (Node => Match_1.Node.Dotted_Name_F_Suffix, Info => Match_1.Info);
--# expr-done 5961
      Cast_Result_5 := Create_Internal_Entity
        (Node => Fld_4.Node,
         Info => Fld_4.Info);
Is_Equiv_1 := Equivalent (Cast_Result_4, Cast_Result_5); 
--# expr-done 5962
if Is_Equiv_1 then
   --# expr-start 5963 '<FieldAccess for parent_candidate_dispatching_call at ast.py:11156>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11156
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.parent_candidate_dispatching_call
Fld_3 := Libadalang.Implementation.Expr_P_Parent_Candidate_Dispatching_Call (Node => Match_1.Node, E_Info => Match_1.Info);
--# end
--# expr-done 5963
   If_Result_1 := Fld_3;
else
Cast_Expr_2 := Match_1; 
      Cast_Result_3 := Create_Internal_Entity_Expr
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
   If_Result_1 := Cast_Result_3;
end if;
--# expr-done 5964
Let_Result_1 := If_Result_1; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Un_Op =>
   --# scope-start
--# scope-start
      Cast_Result_17 := Create_Internal_Entity_Un_Op
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_17; 
--# bind uo Match_2
--# expr-start 5968 '<If at ast.py:11159>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11159
--# expr-start 5966 '<Eq at ast.py:11160>' Is_Equiv_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11160
Cast_Expr_5 := Ent; 
      Cast_Result_7 := Create_Internal_Entity
        (Node => Cast_Expr_5.Node,
         Info => Cast_Expr_5.Info);
--# expr-start 5965 '<FieldAccess for op at ast.py:11160>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11160
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Op (Node => Match_2.Node.Un_Op_F_Op, Info => Match_2.Info);
--# expr-done 5965
      Cast_Result_8 := Create_Internal_Entity
        (Node => Fld_6.Node,
         Info => Fld_6.Info);
Is_Equiv_2 := Equivalent (Cast_Result_7, Cast_Result_8); 
--# expr-done 5966
if Is_Equiv_2 then
   --# expr-start 5967 '<FieldAccess for parent_candidate_dispatching_call at ast.py:11161>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11161
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.parent_candidate_dispatching_call
Fld_5 := Libadalang.Implementation.Expr_P_Parent_Candidate_Dispatching_Call (Node => Match_2.Node, E_Info => Match_2.Info);
--# end
--# expr-done 5967
   If_Result_2 := Fld_5;
else
Cast_Expr_4 := Match_2; 
      Cast_Result_6 := Create_Internal_Entity_Expr
        (Node => Cast_Expr_4.Node,
         Info => Cast_Expr_4.Info);
   If_Result_2 := Cast_Result_6;
end if;
--# expr-done 5968
Let_Result_2 := If_Result_2; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Bin_Op .. Ada_Relation_Op =>
   --# scope-start
--# scope-start
      Cast_Result_18 := Create_Internal_Entity_Bin_Op
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_18; 
--# bind bo Match_3
--# expr-start 5972 '<If at ast.py:11164>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11164
--# expr-start 5970 '<Eq at ast.py:11165>' Is_Equiv_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11165
Cast_Expr_7 := Ent; 
      Cast_Result_10 := Create_Internal_Entity
        (Node => Cast_Expr_7.Node,
         Info => Cast_Expr_7.Info);
--# expr-start 5969 '<FieldAccess for op at ast.py:11165>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11165
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Op (Node => Match_3.Node.Bin_Op_F_Op, Info => Match_3.Info);
--# expr-done 5969
      Cast_Result_11 := Create_Internal_Entity
        (Node => Fld_8.Node,
         Info => Fld_8.Info);
Is_Equiv_3 := Equivalent (Cast_Result_10, Cast_Result_11); 
--# expr-done 5970
if Is_Equiv_3 then
   --# expr-start 5971 '<FieldAccess for parent_candidate_dispatching_call at ast.py:11166>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11166
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.parent_candidate_dispatching_call
Fld_7 := Libadalang.Implementation.Expr_P_Parent_Candidate_Dispatching_Call (Node => Match_3.Node, E_Info => Match_3.Info);
--# end
--# expr-done 5971
   If_Result_3 := Fld_7;
else
Cast_Expr_6 := Match_3; 
      Cast_Result_9 := Create_Internal_Entity_Expr
        (Node => Cast_Expr_6.Node,
         Info => Cast_Expr_6.Info);
   If_Result_3 := Cast_Result_9;
end if;
--# expr-done 5972
Let_Result_3 := If_Result_3; 
--# end
Scope_Result_3 := Let_Result_3;
   --# end
            Match_Result := Scope_Result_3; 
         when Ada_Case_Expr .. Ada_If_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_19 := Create_Internal_Entity_Cond_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_4 := Cast_Result_19; 
--# bind ce Match_4
--# expr-start 5977 '<If at ast.py:11169>' If_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11169
--# expr-start 5975 '<Contains at ast.py:11170>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11170
   --# expr-start 5973 '<FieldAccess for dependent_exprs at ast.py:11170>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11170
      if Match_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]CondExpr.dependent_exprs'
Fld_10 := Libadalang.Implementation.Dispatcher_Cond_Expr_P_Dependent_Exprs (Node => Match_4.Node, E_Info => Match_4.Info);
--# end
--# expr-done 5973
      Quantifier_Result := False;
      declare
         Collection : constant Internal_Entity_Expr_Array_Access := Fld_10;
      begin
         for Item_213 of
               Collection.Items
         loop
   --# scope-start
               --# bind item Item_213
      --# expr-start 5974 '<Eq at ???>' Is_Equiv_4 None
Cast_Expr_8 := Item_213; 
      Cast_Result_12 := Create_Internal_Entity
        (Node => Cast_Expr_8.Node,
         Info => Cast_Expr_8.Info);
Cast_Expr_9 := Ent; 
      Cast_Result_13 := Create_Internal_Entity
        (Node => Cast_Expr_9.Node,
         Info => Cast_Expr_9.Info);
Is_Equiv_4 := Equivalent (Cast_Result_12, Cast_Result_13); 
--# expr-done 5974
      Quantifier_Result := Is_Equiv_4;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
--# expr-done 5975
if Quantifier_Result then
   --# expr-start 5976 '<FieldAccess for parent_candidate_dispatching_call at ast.py:11171>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11171
      if Match_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.parent_candidate_dispatching_call
Fld_9 := Libadalang.Implementation.Expr_P_Parent_Candidate_Dispatching_Call (Node => Match_4.Node, E_Info => Match_4.Info);
--# end
--# expr-done 5976
   If_Result_4 := Fld_9;
else
   If_Result_4 := No_Entity_Expr;
end if;
--# expr-done 5977
Let_Result_4 := If_Result_4; 
--# end
Scope_Result_4 := Let_Result_4;
   --# end
      Finalizer_Scope_2586;
            Match_Result := Scope_Result_4; 
         when Ada_Paren_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_20 := Create_Internal_Entity_Paren_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_5 := Cast_Result_20; 
--# bind pe Match_5
--# expr-start 5978 '<FieldAccess for parent_candidate_dispatching_call at ast.py:11174>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11174
      if Match_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.parent_candidate_dispatching_call
Fld_11 := Libadalang.Implementation.Expr_P_Parent_Candidate_Dispatching_Call (Node => Match_5.Node, E_Info => Match_5.Info);
--# end
--# expr-done 5978
Let_Result_5 := Fld_11; 
--# end
Scope_Result_5 := Let_Result_5;
   --# end
            Match_Result := Scope_Result_5; 
         when Ada_Qual_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_21 := Create_Internal_Entity_Qual_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_6 := Cast_Result_21; 
--# bind qe Match_6
--# expr-start 5979 '<FieldAccess for parent_candidate_dispatching_call at ast.py:11175>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11175
      if Match_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.parent_candidate_dispatching_call
Fld_12 := Libadalang.Implementation.Expr_P_Parent_Candidate_Dispatching_Call (Node => Match_6.Node, E_Info => Match_6.Info);
--# end
--# expr-done 5979
Let_Result_6 := Fld_12; 
--# end
Scope_Result_6 := Let_Result_6;
   --# end
            Match_Result := Scope_Result_6; 
         when Ada_Decl_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_22 := Create_Internal_Entity_Decl_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_7 := Cast_Result_22; 
--# bind de Match_7
--# expr-start 5980 '<FieldAccess for parent_candidate_dispatching_call at ast.py:11176>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11176
      if Match_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.parent_candidate_dispatching_call
Fld_13 := Libadalang.Implementation.Expr_P_Parent_Candidate_Dispatching_Call (Node => Match_7.Node, E_Info => Match_7.Info);
--# end
--# expr-done 5980
Let_Result_7 := Fld_13; 
--# end
Scope_Result_7 := Let_Result_7;
   --# end
            Match_Result := Scope_Result_7; 
         when Ada_Param_Assoc =>
   --# scope-start
--# scope-start
      Cast_Result_23 := Create_Internal_Entity_Param_Assoc
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_8 := Cast_Result_23; 
--# bind pa Match_8
--# expr-start 5983 '<Cast to CallExpr at ast.py:11177>' Cast_Result_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11177
--# expr-start 5982 '<parent at ast.py:11177>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11177
--# expr-start 5981 '<parent at ast.py:11177>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11177
      if Match_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Libadalang.Implementation.Parent (Node => Match_8.Node, E_Info => Match_8.Info);
--# expr-done 5981
      if Fld_14.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Libadalang.Implementation.Parent (Node => Fld_14.Node, E_Info => Fld_14.Info);
--# expr-done 5982
   if Fld_15.Node = null
      or else Fld_15.Node.Kind in Ada_Call_Expr_Range
   then
      Cast_Result_14 := Create_Internal_Entity_Call_Expr
        (Node => Fld_15.Node,
         Info => Fld_15.Info);
   else
         Cast_Result_14 := No_Entity_Call_Expr;
   end if;
--# expr-done 5983
      Cast_Result_24 := Create_Internal_Entity_Expr
        (Node => Cast_Result_14.Node,
         Info => Cast_Result_14.Info);
Let_Result_8 := Cast_Result_24; 
--# end
Scope_Result_8 := Let_Result_8;
   --# end
            Match_Result := Scope_Result_8; 
         when Ada_Abort_Absent .. Ada_Iterated_Assoc | Ada_Abstract_State_Decl .. Ada_Null_Record_Aggregate | Ada_Box_Expr .. Ada_Case_Expr_Alternative | Ada_Contract_Cases | Ada_Membership_Expr .. Ada_Update_Attribute_Ref | Ada_Defining_Name .. Ada_Discrete_Subtype_Name | Ada_End_Name .. Ada_Explicit_Deref | Ada_Reduce_Attribute_Ref .. Ada_Target_Name | Ada_Quantified_Expr .. Ada_Raise_Expr | Ada_Handled_Stmts .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_25 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_9 := Cast_Result_25; 
--# bind _ Match_9
Let_Result_9 := No_Entity_Expr; 
--# end
Scope_Result_9 := Let_Result_9;
   --# end
            Match_Result := Scope_Result_9; 
end case;
--# expr-done 5984
Let_Result_10 := Match_Result; 
--# end
--# expr-done 5985
      Property_Result := Let_Result_10;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_2586;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Parent_Candidate_Dispatching_Call;
--# end
--# property-start Expr.has_dynamic_context /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11183
pragma Warnings (Off, "is not referenced");
 function Expr_P_Has_Dynamic_Context
  (Node : Bare_Expr
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Expr :=
     Internal_Entity_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Assign_Stmt;
Var_Expr_1 : Internal_Entity_Expr;
Fld : Internal_Entity;
Cast_Result : Internal_Entity_Assign_Stmt;
Fld_1 : Internal_Entity_Expr;
Cast_Result_1 : Internal_Entity;
Cast_Expr : Internal_Entity_Expr;
Cast_Result_2 : Internal_Entity;
Is_Equiv : Boolean;
Fld_2 : Internal_Entity_Base_Type_Decl;
Fld_3 : Boolean;
If_Result : Boolean;
Scope_Result : Boolean;
Fld_4 : Internal_Entity_Expr;
Fld_5 : Boolean;
Scope_Result_1 : Boolean;
Result_Var : Boolean;
Result_Var_1 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 5994 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5987 '<Cast to AssignStmt at ast.py:11190>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11190
--# expr-start 5986 '<parent at ast.py:11190>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11190
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 5986
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Assign_Stmt_Range
   then
      Cast_Result := Create_Internal_Entity_Assign_Stmt
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Assign_Stmt;
   end if;
--# expr-done 5987
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Assign_Stmt then
   --# scope-start
   --# bind a Var_Expr
--# expr-start 5989 '<Eq at ast.py:11195>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11195
--# expr-start 5988 '<FieldAccess for expr at ast.py:11195>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11195
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Expr (Node => Var_Expr.Node.Assign_Stmt_F_Expr, Info => Var_Expr.Info);
--# expr-done 5988
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
Cast_Expr := Ent; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Is_Equiv := Equivalent (Cast_Result_1, Cast_Result_2); 
--# expr-done 5989
if Is_Equiv then
   --# expr-start 5991 '<FieldAccess for is_classwide at ast.py:11196>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11196
--# expr-start 5990 '<FieldAccess for expected_expression_type at ast.py:11196>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11196
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.expected_expression_type
Fld_2 := Libadalang.Implementation.Expr_P_Expected_Expression_Type (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5990
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_classwide'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Classwide (Node => Fld_2.Node, E_Info => Fld_2.Info);
--# end
--# expr-done 5991
   If_Result := Fld_3;
else
   If_Result := False;
end if;
Scope_Result := If_Result;
   --# end
   Result_Var_1 := Scope_Result;
else
--# expr-start 5992 '<FieldAccess for parent_candidate_dispatching_call at ast.py:11202>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11202
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.parent_candidate_dispatching_call
Fld_4 := Libadalang.Implementation.Expr_P_Parent_Candidate_Dispatching_Call (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 5992
Var_Expr_1 := Fld_4;
if Var_Expr_1 /= No_Entity_Expr then
   --# scope-start
--# expr-start 5993 '<FieldAccess for is_dispatching_call at ast.py:11202>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11202
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.is_dispatching_call'
Fld_5 := Libadalang.Implementation.Dispatcher_Expr_P_Is_Dispatching_Call (Node => Var_Expr_1.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 5993
Scope_Result_1 := Fld_5;
   --# end
   Result_Var := Scope_Result_1;
else
   Result_Var := False;
end if;
   Result_Var_1 := Result_Var;
end if;
Let_Result := Result_Var_1; 
--# end
--# expr-done 5994
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Has_Dynamic_Context;
--# end
--# property-start Expr.is_dispatching_call_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11208
pragma Warnings (Off, "is not referenced");
 function Expr_P_Is_Dispatching_Call_Impl
  (Node : Bare_Expr
      ; Decl : Internal_Entity_Basic_Decl
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Expr :=
     Internal_Entity_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind decl Decl
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Block_Var : Internal_Entity_Base_Subp_Spec;
Fld : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Base_Subp_Spec;
Is_A : Boolean;
Fld_2 : Boolean;
If_Result : Boolean;
Fld_3 : Boolean;
Fld_4 : Boolean;
If_Result_1 : Boolean;
If_Result_2 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6001 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 5996 '<FieldAccess for subp_spec_or_null at ast.py:11214>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11214
--# expr-start 5995 '<FieldAccess for canonical_part at ast.py:11214>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11214
      if Decl.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.canonical_part
Fld := Libadalang.Implementation.Basic_Decl_P_Canonical_Part (Node => Decl.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Decl.Info);
--# end
--# expr-done 5995
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Fld.Node, Follow_Generic => True, E_Info => Fld.Info);
--# end
--# expr-done 5996
Block_Var := Fld_1; 
--# bind spec Block_Var
--# expr-start 5997 '<IsA AbstractFormalSubpDecl at ast.py:11218>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11218
Is_A := Decl.Node /= null 
and then Decl.Node.Kind in Ada_Abstract_Formal_Subp_Decl_Range; 
--# expr-done 5997
if Is_A then
   If_Result := True;
else
   --# expr-start 5998 '<FieldAccess for has_dynamic_controlling_operand at ast.py:11222>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11222
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.has_dynamic_controlling_operand
Fld_2 := Libadalang.Implementation.Expr_P_Has_Dynamic_Controlling_Operand (Node => Ent.Node, Spec => Block_Var, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 5998
   If_Result := Fld_2;
end if;
if If_Result then
   If_Result_2 := True;
else
--# expr-start 5999 '<FieldAccess for has_controlling_result at ast.py:11228>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11228
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.has_controlling_result
Fld_3 := Libadalang.Implementation.Base_Subp_Spec_P_Has_Controlling_Result (Node => Block_Var.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Block_Var.Info);
--# end
--# expr-done 5999
if Fld_3 then
   --# expr-start 6000 '<FieldAccess for has_dynamic_context at ast.py:11228>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11228
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.has_dynamic_context
Fld_4 := Libadalang.Implementation.Expr_P_Has_Dynamic_Context (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 6000
   If_Result_1 := Fld_4;
else
   If_Result_1 := False;
end if;
   If_Result_2 := If_Result_1;
end if;
Let_Result := If_Result_2; 
--# end
--# expr-done 6001
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Is_Dispatching_Call_Impl;
--# end
--# property-start '[dispatcher]Expr.is_dispatching_call' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Expr_P_Is_Dispatching_Call
  (Node : Bare_Expr
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Expr (Self.Kind) is
               when Ada_Abstract_State_Decl_Expr .. Ada_Null_Record_Aggregate | Ada_Box_Expr .. Ada_Membership_Expr | Ada_Paren_Expr .. Ada_Raise_Expr =>
                  --# property-call-start Expr.is_dispatching_call
                  Property_Result := Expr_P_Is_Dispatching_Call
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Bin_Op .. Ada_Relation_Op =>
                  --# property-call-start BinOp.is_dispatching_call
                  Property_Result := Bin_Op_P_Is_Dispatching_Call
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Attribute_Ref .. Ada_Identifier | Ada_String_Literal .. Ada_Target_Name =>
                  --# property-call-start Name.is_dispatching_call
                  Property_Result := Name_P_Is_Dispatching_Call
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Op_Abs .. Ada_Op_Xor =>
                  --# property-call-start Op.is_dispatching_call
                  Property_Result := Op_P_Is_Dispatching_Call
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Un_Op =>
                  --# property-call-start UnOp.is_dispatching_call
                  Property_Result := Un_Op_P_Is_Dispatching_Call
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Expr_P_Is_Dispatching_Call;
--# end
--# property-start Expr.is_static_expr /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11252
pragma Warnings (Off, "is not referenced");
 function Expr_P_Is_Static_Expr
  (Node : Bare_Expr
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Expr :=
     Internal_Entity_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Match_0 : Internal_Entity_Num_Literal;
Match_1 : Internal_Entity_String_Literal;
Match_2 : Internal_Entity_Attribute_Ref;
Match_3 : Internal_Entity_Call_Expr;
Match_4 : Internal_Entity_Qual_Expr;
Match_5 : Internal_Entity_Name;
Match_6 : Internal_Entity_Membership_Expr;
Match_7 : Internal_Entity_Bin_Op;
Match_8 : Internal_Entity_Un_Op;
Match_9 : Internal_Entity_If_Expr;
Match_10 : Internal_Entity_Paren_Expr;
Match_11 : Internal_Entity_Expr;
Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Entity_Base_Type_Decl;
Var_Expr_2 : Internal_Entity_Base_Type_Decl;
Var_Expr_3 : Internal_Entity_Basic_Decl;
Var_Expr_4 : Internal_Entity_Assoc_List;
Var_Expr_5 : Internal_Entity_Base_Type_Decl;
Var_Expr_6 : Internal_Entity_Base_Type_Decl;
Var_Expr_7 : Internal_Entity_Basic_Decl;
Var_Expr_8 : Internal_Entity_Basic_Decl;
Var_Expr_9 : Internal_Entity_Basic_Decl;
Fld : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_1 : Internal_Entity_Name;
Fld_2 : Boolean;
Fld_3 : Internal_Entity_Name;
Fld_4 : Internal_Entity_Base_Type_Decl;
Fld_5 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Fld_6 : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
Not_Val : Boolean;
Fld_7 : Internal_Entity_Identifier;
Fld_8 : Symbol_Type;
Sym : Symbol_Type;
Is_Equal : Boolean;
If_Result : Boolean;
Fld_9 : Internal_Entity_Name;
Fld_10 : Internal_Entity_Base_Type_Decl;
Fld_11 : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
Fld_12 : Internal_Entity_Identifier;
Fld_13 : Symbol_Type;
Sym_1 : Symbol_Type;
Is_Equal_1 : Boolean;
Fld_14 : Internal_Entity_Identifier;
Fld_15 : Symbol_Type;
Sym_2 : Symbol_Type;
Is_Equal_2 : Boolean;
If_Result_1 : Boolean;
Fld_16 : Internal_Entity_Identifier;
Fld_17 : Symbol_Type;
Sym_3 : Symbol_Type;
Is_Equal_3 : Boolean;
If_Result_2 : Boolean;
Fld_18 : Internal_Entity_Identifier;
Fld_19 : Symbol_Type;
Sym_4 : Symbol_Type;
Is_Equal_4 : Boolean;
If_Result_3 : Boolean;
If_Result_4 : Boolean;
If_Result_5 : Boolean;
Fld_20 : Internal_Entity_Name;
Fld_21 : Internal_Entity_Basic_Decl;
Fld_22 : Boolean;
Scope_Result_3 : Boolean;
Result_Var_3 : Boolean;
Fld_23 : Internal_Entity_Identifier;
Fld_24 : Symbol_Type;
Sym_5 : Symbol_Type;
Is_Equal_5 : Boolean;
Fld_25 : Internal_Entity_Identifier;
Fld_26 : Symbol_Type;
Sym_6 : Symbol_Type;
Is_Equal_6 : Boolean;
If_Result_6 : Boolean;
Fld_27 : Internal_Entity_Identifier;
Fld_28 : Symbol_Type;
Sym_7 : Symbol_Type;
Is_Equal_7 : Boolean;
If_Result_7 : Boolean;
Fld_29 : Internal_Entity_Identifier;
Fld_30 : Symbol_Type;
Sym_8 : Symbol_Type;
Is_Equal_8 : Boolean;
If_Result_8 : Boolean;
If_Result_9 : Boolean;
If_Result_10 : Boolean;
If_Result_11 : Boolean;
Fld_31 : Internal_Entity_Assoc_List;
Saved : Internal_Entity_Assoc_List;
Fld_32 : Bare_Assoc_List;
Fld_33 : Internal_Entity_Info;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result_12 : Internal_Entity_Basic_Assoc;
Fld_34 : Internal_Entity_Expr;
Fld_35 : Boolean;
Item_214 : Internal_Entity_Basic_Assoc;
Bare_Item_214 : Bare_Basic_Assoc;
Quantifier_Result : Boolean;
Scope_Result_4 : Boolean;
Result_Var_4 : Boolean;
If_Result_13 : Boolean;
Fld_36 : Internal_Entity_Name;
Fld_37 : Internal_Entity_Base_Type_Decl;
Fld_38 : Boolean;
Scope_Result_5 : Boolean;
Result_Var_5 : Boolean;
Fld_39 : Internal_Entity_Assoc_List;
Fld_40 : Bare_Assoc_List;
Fld_41 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity_1 : Internal_Entity_Basic_Assoc;
Is_Null_1 : Boolean;
If_Result_14 : Internal_Entity_Basic_Assoc;
Fld_42 : Internal_Entity_Expr;
Fld_43 : Boolean;
If_Result_15 : Boolean;
Fld_44 : Internal_Entity_Name;
Fld_45 : Internal_Entity_Base_Type_Decl;
Fld_46 : Boolean;
Scope_Result_6 : Boolean;
Result_Var_6 : Boolean;
Fld_47 : Internal_Entity_Expr;
Fld_48 : Boolean;
If_Result_16 : Boolean;
Fld_49 : Internal_Entity_Basic_Decl;
Fld_50 : Boolean;
Scope_Result_7 : Boolean;
Result_Var_7 : Boolean;
Fld_51 : Internal_Entity_Expr;
Fld_52 : Boolean;
Fld_53 : Internal_Entity_Expr_Alternatives_List;
Fld_54 : Bare_Expr_Alternatives_List;
Fld_55 : Internal_Entity_Info;
As_Entity_2 : Internal_Entity_Expr;
Is_Null_2 : Boolean;
If_Result_17 : Internal_Entity_Expr;
Fld_56 : Boolean;
Item_215 : Internal_Entity_Expr;
Bare_Item_215 : Bare_Expr;
Quantifier_Result_1 : Boolean;
If_Result_18 : Boolean;
Fld_57 : Internal_Entity_Expr;
Fld_58 : Boolean;
Fld_59 : Internal_Entity_Expr;
Fld_60 : Boolean;
If_Result_19 : Boolean;
Fld_61 : Internal_Entity_Op;
Fld_62 : Internal_Entity_Basic_Decl;
Fld_63 : Boolean;
Scope_Result_8 : Boolean;
Result_Var_8 : Boolean;
If_Result_20 : Boolean;
Fld_64 : Internal_Entity_Expr;
Fld_65 : Boolean;
Fld_66 : Internal_Entity_Op;
Fld_67 : Internal_Entity_Basic_Decl;
Fld_68 : Boolean;
Scope_Result_9 : Boolean;
Result_Var_9 : Boolean;
If_Result_21 : Boolean;
Fld_69 : Internal_Entity_Expr;
Fld_70 : Boolean;
Fld_71 : Internal_Entity_Expr;
Fld_72 : Boolean;
If_Result_22 : Boolean;
Fld_73 : Internal_Entity_Elsif_Expr_Part_List;
Fld_74 : Bare_Elsif_Expr_Part_List;
Fld_75 : Internal_Entity_Info;
As_Entity_3 : Internal_Entity_Elsif_Expr_Part;
Is_Null_3 : Boolean;
If_Result_23 : Internal_Entity_Elsif_Expr_Part;
Fld_76 : Internal_Entity_Expr;
Fld_77 : Boolean;
Fld_78 : Internal_Entity_Expr;
Fld_79 : Boolean;
If_Result_24 : Boolean;
Item_216 : Internal_Entity_Elsif_Expr_Part;
Bare_Item_216 : Bare_Elsif_Expr_Part;
Quantifier_Result_2 : Boolean;
If_Result_25 : Boolean;
Fld_80 : Internal_Entity_Expr;
Fld_81 : Boolean;
If_Result_26 : Boolean;
Fld_82 : Internal_Entity_Expr;
Fld_83 : Boolean;
Match_Prefix : Internal_Entity_Expr;
Cast_Result : Internal_Entity_Num_Literal;
Let_Result : Boolean;
Scope_Result_10 : Boolean;
Cast_Result_1 : Internal_Entity_String_Literal;
Let_Result_1 : Boolean;
Scope_Result_11 : Boolean;
Cast_Result_2 : Internal_Entity_Attribute_Ref;
Let_Result_2 : Boolean;
Scope_Result_12 : Boolean;
Cast_Result_3 : Internal_Entity_Call_Expr;
Let_Result_3 : Boolean;
Scope_Result_13 : Boolean;
Cast_Result_4 : Internal_Entity_Qual_Expr;
Let_Result_4 : Boolean;
Scope_Result_14 : Boolean;
Cast_Result_5 : Internal_Entity_Name;
Let_Result_5 : Boolean;
Scope_Result_15 : Boolean;
Cast_Result_6 : Internal_Entity_Membership_Expr;
Let_Result_6 : Boolean;
Scope_Result_16 : Boolean;
Cast_Result_7 : Internal_Entity_Bin_Op;
Let_Result_7 : Boolean;
Scope_Result_17 : Boolean;
Cast_Result_8 : Internal_Entity_Un_Op;
Let_Result_8 : Boolean;
Scope_Result_18 : Boolean;
Cast_Result_9 : Internal_Entity_If_Expr;
Let_Result_9 : Boolean;
Scope_Result_19 : Boolean;
Cast_Result_10 : Internal_Entity_Paren_Expr;
Let_Result_10 : Boolean;
Scope_Result_20 : Boolean;
Cast_Result_11 : Internal_Entity_Expr;
Let_Result_11 : Boolean;
Scope_Result_21 : Boolean;
Match_Result : Boolean;
Dyn_Var_Bind_Result : Boolean;
Let_Result_12 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6102 '<Block at ???>' Let_Result_12 None
--# scope-start
--# expr-start 6101 '<bind at ast.py:11257>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11257
--# expr-start 6002 '<FieldAccess for origin_node at ast.py:11257>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11257
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 6002
Bound_Origin := Fld; 
--# expr-start 6100 '<Match at ast.py:11257>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11257
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Expr (Match_Prefix.Node.Kind) is
         when Ada_Int_Literal .. Ada_Real_Literal =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Num_Literal
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind _ Match_0
Let_Result := True; 
--# end
Scope_Result_10 := Let_Result;
   --# end
            Match_Result := Scope_Result_10; 
         when Ada_String_Literal =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_String_Literal
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind _ Match_1
Let_Result_1 := True; 
--# end
Scope_Result_11 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_11; 
         when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Attribute_Ref
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind ar Match_2
--# expr-start 6004 '<FieldAccess for is_static_expr at ast.py:11261>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11261
--# expr-start 6003 '<FieldAccess for prefix at ast.py:11261>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11261
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Match_2.Node.Attribute_Ref_F_Prefix, Info => Match_2.Info);
--# expr-done 6003
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_2 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_1.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_1.Info);
--# end
--# expr-done 6004
if Fld_2 then
--# expr-start 6009 '<Not at ast.py:11262>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11262
--# expr-start 6006 '<FieldAccess for name_designated_type at ast.py:11262>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11262
--# expr-start 6005 '<FieldAccess for prefix at ast.py:11262>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11262
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Name (Node => Match_2.Node.Attribute_Ref_F_Prefix, Info => Match_2.Info);
--# expr-done 6005
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_4 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld_3.Node, E_Info => Fld_3.Info);
--# end
--# expr-done 6006
Var_Expr_1 := Fld_4;
if Var_Expr_1 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 6007 '<FieldAccess for root_type at ast.py:11262>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11262
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.root_type'
Fld_5 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Root_Type (Node => Var_Expr_1.Node, Origin => Bound_Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 6007
Scope_Result := Fld_5;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Var_Expr := Result_Var;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 6008 '<FieldAccess for is_formal at ast.py:11262>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11262
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_formal
Fld_6 := Libadalang.Implementation.Basic_Decl_P_Is_Formal (Node => Var_Expr.Node);
--# end
--# expr-done 6008
Scope_Result_1 := Fld_6;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
Not_Val := not (Result_Var_1); 
--# expr-done 6009
if Not_Val then
   --# expr-start 6013 '<Eq at ast.py:11264>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11264
--# expr-start 6011 '<FieldAccess for name_symbol at ast.py:11264>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11264
--# expr-start 6010 '<FieldAccess for attribute at ast.py:11264>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11264
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Identifier (Node => Match_2.Node.Attribute_Ref_F_Attribute, Info => Match_2.Info);
--# expr-done 6010
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_8 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_7.Node);
--# end
--# expr-done 6011
--# expr-start 6012 '<Symbol Base at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Base); 
--# expr-done 6012
Is_Equal := Fld_8 = Sym; 
--# expr-done 6013
   If_Result := Is_Equal;
else
   If_Result := False;
end if;
if If_Result then
   If_Result_5 := True;
else
--# expr-start 6015 '<FieldAccess for name_designated_type at ast.py:11266>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11266
--# expr-start 6014 '<FieldAccess for prefix at ast.py:11266>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11266
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Name (Node => Match_2.Node.Attribute_Ref_F_Prefix, Info => Match_2.Info);
--# expr-done 6014
      if Fld_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_10 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld_9.Node, E_Info => Fld_9.Info);
--# end
--# expr-done 6015
Var_Expr_2 := Fld_10;
if Var_Expr_2 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 6016 '<FieldAccess for is_static_decl at ast.py:11266>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11266
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_static_decl'
Fld_11 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Static_Decl (Node => Var_Expr_2.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 6016
Scope_Result_2 := Fld_11;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := False;
end if;
if Result_Var_2 then
--# expr-start 6020 '<Eq at ast.py:11267>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11267
--# expr-start 6018 '<FieldAccess for name_symbol at ast.py:11267>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11267
--# expr-start 6017 '<FieldAccess for attribute at ast.py:11267>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11267
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Create_Internal_Entity_Identifier (Node => Match_2.Node.Attribute_Ref_F_Attribute, Info => Match_2.Info);
--# expr-done 6017
      if Fld_12.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_13 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_12.Node);
--# end
--# expr-done 6018
--# expr-start 6019 '<Symbol First at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_First); 
--# expr-done 6019
Is_Equal_1 := Fld_13 = Sym_1; 
--# expr-done 6020
if Is_Equal_1 then
   If_Result_1 := True;
else
   --# expr-start 6024 '<Eq at ast.py:11267>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11267
--# expr-start 6022 '<FieldAccess for name_symbol at ast.py:11267>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11267
--# expr-start 6021 '<FieldAccess for attribute at ast.py:11267>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11267
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Create_Internal_Entity_Identifier (Node => Match_2.Node.Attribute_Ref_F_Attribute, Info => Match_2.Info);
--# expr-done 6021
      if Fld_14.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_15 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_14.Node);
--# end
--# expr-done 6022
--# expr-start 6023 '<Symbol Last at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Last); 
--# expr-done 6023
Is_Equal_2 := Fld_15 = Sym_2; 
--# expr-done 6024
   If_Result_1 := Is_Equal_2;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
   --# expr-start 6028 '<Eq at ast.py:11267>' Is_Equal_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11267
--# expr-start 6026 '<FieldAccess for name_symbol at ast.py:11267>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11267
--# expr-start 6025 '<FieldAccess for attribute at ast.py:11267>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11267
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_16 := Create_Internal_Entity_Identifier (Node => Match_2.Node.Attribute_Ref_F_Attribute, Info => Match_2.Info);
--# expr-done 6025
      if Fld_16.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_17 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_16.Node);
--# end
--# expr-done 6026
--# expr-start 6027 '<Symbol Range at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Range); 
--# expr-done 6027
Is_Equal_3 := Fld_17 = Sym_3; 
--# expr-done 6028
   If_Result_2 := Is_Equal_3;
end if;
if If_Result_2 then
   If_Result_3 := True;
else
   --# expr-start 6032 '<Eq at ast.py:11267>' Is_Equal_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11267
--# expr-start 6030 '<FieldAccess for name_symbol at ast.py:11267>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11267
--# expr-start 6029 '<FieldAccess for attribute at ast.py:11267>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11267
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Create_Internal_Entity_Identifier (Node => Match_2.Node.Attribute_Ref_F_Attribute, Info => Match_2.Info);
--# expr-done 6029
      if Fld_18.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_19 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_18.Node);
--# end
--# expr-done 6030
--# expr-start 6031 '<Symbol Val at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Val); 
--# expr-done 6031
Is_Equal_4 := Fld_19 = Sym_4; 
--# expr-done 6032
   If_Result_3 := Is_Equal_4;
end if;
   If_Result_4 := If_Result_3;
else
   If_Result_4 := False;
end if;
   If_Result_5 := If_Result_4;
end if;
if If_Result_5 then
   If_Result_10 := True;
else
--# expr-start 6034 '<FieldAccess for referenced_decl at ast.py:11271>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11271
--# expr-start 6033 '<FieldAccess for prefix at ast.py:11271>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11271
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_20 := Create_Internal_Entity_Name (Node => Match_2.Node.Attribute_Ref_F_Prefix, Info => Match_2.Info);
--# expr-done 6033
      if Fld_20.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld_21 := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Fld_20.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_20.Info);
--# end
--# expr-done 6034
Var_Expr_3 := Fld_21;
if Var_Expr_3 /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 6035 '<FieldAccess for is_array at ast.py:11271>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11271
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_array
Fld_22 := Libadalang.Implementation.Basic_Decl_P_Is_Array (Node => Var_Expr_3.Node, Origin => Bound_Origin, E_Info => Var_Expr_3.Info);
--# end
--# expr-done 6035
Scope_Result_3 := Fld_22;
   --# end
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := False;
end if;
if Result_Var_3 then
--# expr-start 6039 '<Eq at ast.py:11272>' Is_Equal_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11272
--# expr-start 6037 '<FieldAccess for name_symbol at ast.py:11272>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11272
--# expr-start 6036 '<FieldAccess for attribute at ast.py:11272>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11272
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_23 := Create_Internal_Entity_Identifier (Node => Match_2.Node.Attribute_Ref_F_Attribute, Info => Match_2.Info);
--# expr-done 6036
      if Fld_23.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_24 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_23.Node);
--# end
--# expr-done 6037
--# expr-start 6038 '<Symbol First at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_First); 
--# expr-done 6038
Is_Equal_5 := Fld_24 = Sym_5; 
--# expr-done 6039
if Is_Equal_5 then
   If_Result_6 := True;
else
   --# expr-start 6043 '<Eq at ast.py:11272>' Is_Equal_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11272
--# expr-start 6041 '<FieldAccess for name_symbol at ast.py:11272>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11272
--# expr-start 6040 '<FieldAccess for attribute at ast.py:11272>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11272
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_25 := Create_Internal_Entity_Identifier (Node => Match_2.Node.Attribute_Ref_F_Attribute, Info => Match_2.Info);
--# expr-done 6040
      if Fld_25.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_26 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_25.Node);
--# end
--# expr-done 6041
--# expr-start 6042 '<Symbol Last at ???>' Sym_6 None
Sym_6 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Last); 
--# expr-done 6042
Is_Equal_6 := Fld_26 = Sym_6; 
--# expr-done 6043
   If_Result_6 := Is_Equal_6;
end if;
if If_Result_6 then
   If_Result_7 := True;
else
   --# expr-start 6047 '<Eq at ast.py:11272>' Is_Equal_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11272
--# expr-start 6045 '<FieldAccess for name_symbol at ast.py:11272>' Fld_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11272
--# expr-start 6044 '<FieldAccess for attribute at ast.py:11272>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11272
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_27 := Create_Internal_Entity_Identifier (Node => Match_2.Node.Attribute_Ref_F_Attribute, Info => Match_2.Info);
--# expr-done 6044
      if Fld_27.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_28 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_27.Node);
--# end
--# expr-done 6045
--# expr-start 6046 '<Symbol Length at ???>' Sym_7 None
Sym_7 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Length); 
--# expr-done 6046
Is_Equal_7 := Fld_28 = Sym_7; 
--# expr-done 6047
   If_Result_7 := Is_Equal_7;
end if;
if If_Result_7 then
   If_Result_8 := True;
else
   --# expr-start 6051 '<Eq at ast.py:11272>' Is_Equal_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11272
--# expr-start 6049 '<FieldAccess for name_symbol at ast.py:11272>' Fld_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11272
--# expr-start 6048 '<FieldAccess for attribute at ast.py:11272>' Fld_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11272
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_29 := Create_Internal_Entity_Identifier (Node => Match_2.Node.Attribute_Ref_F_Attribute, Info => Match_2.Info);
--# expr-done 6048
      if Fld_29.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_30 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_29.Node);
--# end
--# expr-done 6049
--# expr-start 6050 '<Symbol Range at ???>' Sym_8 None
Sym_8 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Range); 
--# expr-done 6050
Is_Equal_8 := Fld_30 = Sym_8; 
--# expr-done 6051
   If_Result_8 := Is_Equal_8;
end if;
   If_Result_9 := If_Result_8;
else
   If_Result_9 := False;
end if;
   If_Result_10 := If_Result_9;
end if;
   If_Result_11 := If_Result_10;
else
   If_Result_11 := False;
end if;
if If_Result_11 then
--# expr-start 6052 '<FieldAccess for args_list at ast.py:11275>' Fld_31 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11275
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_31 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Match_2.Node, E_Info => Match_2.Info);
--# end
--# expr-done 6052
Var_Expr_4 := Fld_31;
if Var_Expr_4 /= No_Entity_Assoc_List then
   --# scope-start
   --# bind args Var_Expr_4
--# expr-start 6055 '<AllQuantifier at ast.py:11278>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11278
Saved := Var_Expr_4; 
Fld_32 := Saved.Node;
      Quantifier_Result := True;
      if Fld_32 = null then
      null;
      else
      declare
         Collection : constant Bare_Assoc_List := Fld_32;
      begin
         for Untyped_Item_214 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_214 := Untyped_Item_214; 
Is_Null := Bare_Item_214 = null; 
if Is_Null then
   If_Result_12 := No_Entity_Basic_Assoc;
else
Fld_33 := Saved.Info;
As_Entity := (Info => Fld_33, Node => Bare_Item_214); 
   If_Result_12 := As_Entity;
end if;
                  Item_214 := If_Result_12; 
   --# scope-start
               --# bind arg Item_214
      --# expr-start 6054 '<FieldAccess for is_static_expr at ast.py:11278>' Fld_35 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11278
--# expr-start 6053 '<FieldAccess for expr at ast.py:11278>' Fld_34 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11278
      if Item_214.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_34 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Item_214.Node, E_Info => Item_214.Info);
--# end
--# expr-done 6053
      if Fld_34.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_35 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_34.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_34.Info);
--# end
--# expr-done 6054
      Quantifier_Result := Fld_35;
   --# end
               exit when not Quantifier_Result;
         end loop;
      end;
      end if;
--# expr-done 6055
Scope_Result_4 := Quantifier_Result;
   --# end
   Result_Var_4 := Scope_Result_4;
else
   Result_Var_4 := True;
end if;
   If_Result_13 := Result_Var_4;
else
   If_Result_13 := False;
end if;
Let_Result_2 := If_Result_13; 
--# end
Scope_Result_12 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_12; 
         when Ada_Call_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Call_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_3; 
--# bind ce Match_3
--# expr-start 6057 '<FieldAccess for name_designated_type at ast.py:11283>' Fld_37 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11283
--# expr-start 6056 '<FieldAccess for name at ast.py:11283>' Fld_36 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11283
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_36 := Create_Internal_Entity_Name (Node => Match_3.Node.Call_Expr_F_Name, Info => Match_3.Info);
--# expr-done 6056
      if Fld_36.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_37 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld_36.Node, E_Info => Fld_36.Info);
--# end
--# expr-done 6057
Var_Expr_5 := Fld_37;
if Var_Expr_5 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 6058 '<FieldAccess for is_static_decl at ast.py:11283>' Fld_38 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11283
      if Var_Expr_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_static_decl'
Fld_38 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Static_Decl (Node => Var_Expr_5.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr_5.Info);
--# end
--# expr-done 6058
Scope_Result_5 := Fld_38;
   --# end
   Result_Var_5 := Scope_Result_5;
else
   Result_Var_5 := False;
end if;
if Result_Var_5 then
   --# expr-start 6061 '<FieldAccess for is_static_expr at ast.py:11284>' Fld_43 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11284
--# expr-start 6060 '<FieldAccess for expr at ast.py:11284>' Fld_42 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11284
--# expr-start 6059 '<FieldAccess for params at ast.py:11284>' Fld_39 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11284
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.params
Fld_39 := Libadalang.Implementation.Call_Expr_P_Params (Node => Match_3.Node, E_Info => Match_3.Info);
--# end
--# expr-done 6059
Fld_40 := Fld_39.Node;
Get_Result := Get (Fld_40, 0, True); 
Is_Null_1 := Get_Result = null; 
if Is_Null_1 then
   If_Result_14 := No_Entity_Basic_Assoc;
else
Fld_41 := Fld_39.Info;
As_Entity_1 := (Info => Fld_41, Node => Get_Result); 
   If_Result_14 := As_Entity_1;
end if;
      if If_Result_14.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_42 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result_14.Node, E_Info => If_Result_14.Info);
--# end
--# expr-done 6060
      if Fld_42.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_43 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_42.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_42.Info);
--# end
--# expr-done 6061
   If_Result_15 := Fld_43;
else
   If_Result_15 := False;
end if;
Let_Result_3 := If_Result_15; 
--# end
Scope_Result_13 := Let_Result_3;
   --# end
            Match_Result := Scope_Result_13; 
         when Ada_Qual_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Qual_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_4 := Cast_Result_4; 
--# bind qe Match_4
--# expr-start 6063 '<FieldAccess for name_designated_type at ast.py:11287>' Fld_45 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11287
--# expr-start 6062 '<FieldAccess for prefix at ast.py:11287>' Fld_44 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11287
      if Match_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_44 := Create_Internal_Entity_Name (Node => Match_4.Node.Qual_Expr_F_Prefix, Info => Match_4.Info);
--# expr-done 6062
      if Fld_44.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_45 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld_44.Node, E_Info => Fld_44.Info);
--# end
--# expr-done 6063
Var_Expr_6 := Fld_45;
if Var_Expr_6 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 6064 '<FieldAccess for is_static_decl at ast.py:11287>' Fld_46 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11287
      if Var_Expr_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_static_decl'
Fld_46 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Static_Decl (Node => Var_Expr_6.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr_6.Info);
--# end
--# expr-done 6064
Scope_Result_6 := Fld_46;
   --# end
   Result_Var_6 := Scope_Result_6;
else
   Result_Var_6 := False;
end if;
if Result_Var_6 then
   --# expr-start 6066 '<FieldAccess for is_static_expr at ast.py:11288>' Fld_48 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11288
--# expr-start 6065 '<FieldAccess for suffix at ast.py:11288>' Fld_47 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11288
      if Match_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_47 := Create_Internal_Entity_Expr (Node => Match_4.Node.Qual_Expr_F_Suffix, Info => Match_4.Info);
--# expr-done 6065
      if Fld_47.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_48 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_47.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_47.Info);
--# end
--# expr-done 6066
   If_Result_16 := Fld_48;
else
   If_Result_16 := False;
end if;
Let_Result_4 := If_Result_16; 
--# end
Scope_Result_14 := Let_Result_4;
   --# end
            Match_Result := Scope_Result_14; 
         when Ada_Defining_Name .. Ada_Explicit_Deref | Ada_Reduce_Attribute_Ref .. Ada_Op_Xor | Ada_Null_Literal | Ada_Synthetic_Identifier .. Ada_Target_Name =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Create_Internal_Entity_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_5 := Cast_Result_5; 
--# bind n Match_5
--# expr-start 6067 '<FieldAccess for referenced_decl at ast.py:11291>' Fld_49 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11291
      if Match_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld_49 := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Match_5.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Match_5.Info);
--# end
--# expr-done 6067
Var_Expr_7 := Fld_49;
if Var_Expr_7 /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 6068 '<FieldAccess for is_static_decl at ast.py:11291>' Fld_50 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11291
      if Var_Expr_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_static_decl'
Fld_50 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Static_Decl (Node => Var_Expr_7.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr_7.Info);
--# end
--# expr-done 6068
Scope_Result_7 := Fld_50;
   --# end
   Result_Var_7 := Scope_Result_7;
else
   Result_Var_7 := False;
end if;
Let_Result_5 := Result_Var_7; 
--# end
Scope_Result_15 := Let_Result_5;
   --# end
            Match_Result := Scope_Result_15; 
         when Ada_Membership_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_6 := Create_Internal_Entity_Membership_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_6 := Cast_Result_6; 
--# bind me Match_6
--# expr-start 6070 '<FieldAccess for is_static_expr at ast.py:11294>' Fld_52 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11294
--# expr-start 6069 '<FieldAccess for expr at ast.py:11294>' Fld_51 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11294
      if Match_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_51 := Create_Internal_Entity_Expr (Node => Match_6.Node.Membership_Expr_F_Expr, Info => Match_6.Info);
--# expr-done 6069
      if Fld_51.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_52 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_51.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_51.Info);
--# end
--# expr-done 6070
if Fld_52 then
   --# expr-start 6073 '<AllQuantifier at ast.py:11294>' Quantifier_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11294
   --# expr-start 6071 '<FieldAccess for membership_exprs at ast.py:11294>' Fld_53 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11294
      if Match_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_53 := Create_Internal_Entity_Expr_Alternatives_List (Node => Match_6.Node.Membership_Expr_F_Membership_Exprs, Info => Match_6.Info);
--# expr-done 6071
Fld_54 := Fld_53.Node;
      Quantifier_Result_1 := True;
      if Fld_54 = null then
      null;
      else
      declare
         Collection : constant Bare_Expr_Alternatives_List := Fld_54;
      begin
         for Untyped_Item_215 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_215 := Untyped_Item_215; 
Is_Null_2 := Bare_Item_215 = null; 
if Is_Null_2 then
   If_Result_17 := No_Entity_Expr;
else
Fld_55 := Fld_53.Info;
As_Entity_2 := (Info => Fld_55, Node => Bare_Item_215); 
   If_Result_17 := As_Entity_2;
end if;
                  Item_215 := If_Result_17; 
   --# scope-start
               --# bind e Item_215
      --# expr-start 6072 '<FieldAccess for is_static_expr at ast.py:11295>' Fld_56 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11295
      if Item_215.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_56 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Item_215.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Item_215.Info);
--# end
--# expr-done 6072
      Quantifier_Result_1 := Fld_56;
   --# end
               exit when not Quantifier_Result_1;
         end loop;
      end;
      end if;
--# expr-done 6073
   If_Result_18 := Quantifier_Result_1;
else
   If_Result_18 := False;
end if;
Let_Result_6 := If_Result_18; 
--# end
Scope_Result_16 := Let_Result_6;
   --# end
            Match_Result := Scope_Result_16; 
         when Ada_Bin_Op .. Ada_Relation_Op =>
   --# scope-start
--# scope-start
      Cast_Result_7 := Create_Internal_Entity_Bin_Op
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_7 := Cast_Result_7; 
--# bind bo Match_7
--# expr-start 6075 '<FieldAccess for is_static_expr at ast.py:11299>' Fld_58 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11299
--# expr-start 6074 '<FieldAccess for left at ast.py:11299>' Fld_57 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11299
      if Match_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_57 := Create_Internal_Entity_Expr (Node => Match_7.Node.Bin_Op_F_Left, Info => Match_7.Info);
--# expr-done 6074
      if Fld_57.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_58 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_57.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_57.Info);
--# end
--# expr-done 6075
if Fld_58 then
   --# expr-start 6077 '<FieldAccess for is_static_expr at ast.py:11300>' Fld_60 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11300
--# expr-start 6076 '<FieldAccess for right at ast.py:11300>' Fld_59 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11300
      if Match_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_59 := Create_Internal_Entity_Expr (Node => Match_7.Node.Bin_Op_F_Right, Info => Match_7.Info);
--# expr-done 6076
      if Fld_59.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_60 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_59.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_59.Info);
--# end
--# expr-done 6077
   If_Result_19 := Fld_60;
else
   If_Result_19 := False;
end if;
if If_Result_19 then
--# expr-start 6079 '<FieldAccess for referenced_decl at ast.py:11301>' Fld_62 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11301
--# expr-start 6078 '<FieldAccess for op at ast.py:11301>' Fld_61 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11301
      if Match_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_61 := Create_Internal_Entity_Op (Node => Match_7.Node.Bin_Op_F_Op, Info => Match_7.Info);
--# expr-done 6078
      if Fld_61.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld_62 := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Fld_61.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_61.Info);
--# end
--# expr-done 6079
Var_Expr_8 := Fld_62;
if Var_Expr_8 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind decl Var_Expr_8
--# expr-start 6080 '<FieldAccess for is_static_decl at ast.py:11302>' Fld_63 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11302
      if Var_Expr_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_static_decl'
Fld_63 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Static_Decl (Node => Var_Expr_8.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr_8.Info);
--# end
--# expr-done 6080
Scope_Result_8 := Fld_63;
   --# end
   Result_Var_8 := Scope_Result_8;
else
   Result_Var_8 := True;
end if;
   If_Result_20 := Result_Var_8;
else
   If_Result_20 := False;
end if;
Let_Result_7 := If_Result_20; 
--# end
Scope_Result_17 := Let_Result_7;
   --# end
            Match_Result := Scope_Result_17; 
         when Ada_Un_Op =>
   --# scope-start
--# scope-start
      Cast_Result_8 := Create_Internal_Entity_Un_Op
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_8 := Cast_Result_8; 
--# bind uo Match_8
--# expr-start 6082 '<FieldAccess for is_static_expr at ast.py:11307>' Fld_65 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11307
--# expr-start 6081 '<FieldAccess for expr at ast.py:11307>' Fld_64 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11307
      if Match_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_64 := Create_Internal_Entity_Expr (Node => Match_8.Node.Un_Op_F_Expr, Info => Match_8.Info);
--# expr-done 6081
      if Fld_64.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_65 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_64.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_64.Info);
--# end
--# expr-done 6082
if Fld_65 then
--# expr-start 6084 '<FieldAccess for referenced_decl at ast.py:11308>' Fld_67 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11308
--# expr-start 6083 '<FieldAccess for op at ast.py:11308>' Fld_66 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11308
      if Match_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_66 := Create_Internal_Entity_Op (Node => Match_8.Node.Un_Op_F_Op, Info => Match_8.Info);
--# expr-done 6083
      if Fld_66.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld_67 := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Fld_66.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_66.Info);
--# end
--# expr-done 6084
Var_Expr_9 := Fld_67;
if Var_Expr_9 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind decl Var_Expr_9
--# expr-start 6085 '<FieldAccess for is_static_decl at ast.py:11309>' Fld_68 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11309
      if Var_Expr_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_static_decl'
Fld_68 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Static_Decl (Node => Var_Expr_9.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr_9.Info);
--# end
--# expr-done 6085
Scope_Result_9 := Fld_68;
   --# end
   Result_Var_9 := Scope_Result_9;
else
   Result_Var_9 := True;
end if;
   If_Result_21 := Result_Var_9;
else
   If_Result_21 := False;
end if;
Let_Result_8 := If_Result_21; 
--# end
Scope_Result_18 := Let_Result_8;
   --# end
            Match_Result := Scope_Result_18; 
         when Ada_If_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_9 := Create_Internal_Entity_If_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_9 := Cast_Result_9; 
--# bind i Match_9
--# expr-start 6087 '<FieldAccess for is_static_expr at ast.py:11314>' Fld_70 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11314
--# expr-start 6086 '<FieldAccess for cond_expr at ast.py:11314>' Fld_69 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11314
      if Match_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_69 := Create_Internal_Entity_Expr (Node => Match_9.Node.If_Expr_F_Cond_Expr, Info => Match_9.Info);
--# expr-done 6086
      if Fld_69.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_70 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_69.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_69.Info);
--# end
--# expr-done 6087
if Fld_70 then
   --# expr-start 6089 '<FieldAccess for is_static_expr at ast.py:11314>' Fld_72 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11314
--# expr-start 6088 '<FieldAccess for then_expr at ast.py:11314>' Fld_71 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11314
      if Match_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_71 := Create_Internal_Entity_Expr (Node => Match_9.Node.If_Expr_F_Then_Expr, Info => Match_9.Info);
--# expr-done 6088
      if Fld_71.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_72 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_71.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_71.Info);
--# end
--# expr-done 6089
   If_Result_22 := Fld_72;
else
   If_Result_22 := False;
end if;
if If_Result_22 then
   --# expr-start 6095 '<AllQuantifier at ast.py:11315>' Quantifier_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11315
   --# expr-start 6090 '<FieldAccess for alternatives at ast.py:11315>' Fld_73 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11315
      if Match_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_73 := Create_Internal_Entity_Elsif_Expr_Part_List (Node => Match_9.Node.If_Expr_F_Alternatives, Info => Match_9.Info);
--# expr-done 6090
Fld_74 := Fld_73.Node;
      Quantifier_Result_2 := True;
      if Fld_74 = null then
      null;
      else
      declare
         Collection : constant Bare_Elsif_Expr_Part_List := Fld_74;
      begin
         for Untyped_Item_216 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_216 := Untyped_Item_216; 
Is_Null_3 := Bare_Item_216 = null; 
if Is_Null_3 then
   If_Result_23 := No_Entity_Elsif_Expr_Part;
else
Fld_75 := Fld_73.Info;
As_Entity_3 := (Info => Fld_75, Node => Bare_Item_216); 
   If_Result_23 := As_Entity_3;
end if;
                  Item_216 := If_Result_23; 
   --# scope-start
               --# bind a Item_216
--# expr-start 6092 '<FieldAccess for is_static_expr at ast.py:11317>' Fld_77 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11317
--# expr-start 6091 '<FieldAccess for cond_expr at ast.py:11317>' Fld_76 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11317
      if Item_216.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_76 := Create_Internal_Entity_Expr (Node => Item_216.Node.Elsif_Expr_Part_F_Cond_Expr, Info => Item_216.Info);
--# expr-done 6091
      if Fld_76.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_77 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_76.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_76.Info);
--# end
--# expr-done 6092
if Fld_77 then
   --# expr-start 6094 '<FieldAccess for is_static_expr at ast.py:11318>' Fld_79 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11318
--# expr-start 6093 '<FieldAccess for then_expr at ast.py:11318>' Fld_78 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11318
      if Item_216.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_78 := Create_Internal_Entity_Expr (Node => Item_216.Node.Elsif_Expr_Part_F_Then_Expr, Info => Item_216.Info);
--# expr-done 6093
      if Fld_78.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_79 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_78.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_78.Info);
--# end
--# expr-done 6094
   If_Result_24 := Fld_79;
else
   If_Result_24 := False;
end if;
      Quantifier_Result_2 := If_Result_24;
   --# end
               exit when not Quantifier_Result_2;
         end loop;
      end;
      end if;
--# expr-done 6095
   If_Result_25 := Quantifier_Result_2;
else
   If_Result_25 := False;
end if;
if If_Result_25 then
   --# expr-start 6097 '<FieldAccess for is_static_expr at ast.py:11320>' Fld_81 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11320
--# expr-start 6096 '<FieldAccess for else_expr at ast.py:11320>' Fld_80 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11320
      if Match_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_80 := Create_Internal_Entity_Expr (Node => Match_9.Node.If_Expr_F_Else_Expr, Info => Match_9.Info);
--# expr-done 6096
      if Fld_80.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_81 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_80.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_80.Info);
--# end
--# expr-done 6097
   If_Result_26 := Fld_81;
else
   If_Result_26 := False;
end if;
Let_Result_9 := If_Result_26; 
--# end
Scope_Result_19 := Let_Result_9;
   --# end
            Match_Result := Scope_Result_19; 
         when Ada_Paren_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_10 := Create_Internal_Entity_Paren_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_10 := Cast_Result_10; 
--# bind pe Match_10
--# expr-start 6099 '<FieldAccess for is_static_expr at ast.py:11322>' Fld_83 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11322
--# expr-start 6098 '<FieldAccess for expr at ast.py:11322>' Fld_82 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11322
      if Match_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_82 := Create_Internal_Entity_Expr (Node => Match_10.Node.Paren_Expr_F_Expr, Info => Match_10.Info);
--# expr-done 6098
      if Fld_82.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_83 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_82.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_82.Info);
--# end
--# expr-done 6099
Let_Result_10 := Fld_83; 
--# end
Scope_Result_20 := Let_Result_10;
   --# end
            Match_Result := Scope_Result_20; 
         when Ada_Abstract_State_Decl_Expr .. Ada_Null_Record_Aggregate | Ada_Box_Expr .. Ada_Case_Expr | Ada_Contract_Cases .. Ada_Decl_Expr | Ada_Quantified_Expr .. Ada_Raise_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_11 := Create_Internal_Entity_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_11 := Cast_Result_11; 
--# bind _ Match_11
Let_Result_11 := False; 
--# end
Scope_Result_21 := Let_Result_11;
   --# end
            Match_Result := Scope_Result_21; 
end case;
--# expr-done 6100
Dyn_Var_Bind_Result := Match_Result; 
--# expr-done 6101
Let_Result_12 := Dyn_Var_Bind_Result; 
--# end
--# expr-done 6102
      Property_Result := Let_Result_12;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Is_Static_Expr;
--# end
--# property-start '[dispatcher]Expr.first_corresponding_decl' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Expr_P_First_Corresponding_Decl
  (Node : Bare_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Expr (Self.Kind) is
               when Ada_Abstract_State_Decl_Expr .. Ada_Membership_Expr | Ada_Paren_Expr .. Ada_Un_Op =>
                  --# property-call-start Expr.first_corresponding_decl
                  Property_Result := Expr_P_First_Corresponding_Decl
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Attribute_Ref .. Ada_Target_Name =>
                  --# property-call-start Name.first_corresponding_decl
                  Property_Result := Name_P_First_Corresponding_Decl
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Expr_P_First_Corresponding_Decl;
--# end
--# property-start Expr.eval_as_int /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11336
pragma Warnings (Off, "is not referenced");
 function Expr_P_Eval_As_Int
  (Node : Bare_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Big_Integer_Type
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Expr :=
     Internal_Entity_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Big_Integer_Type;
            procedure Finalizer_Scope_881 with Inline_Always;
      Fld : Big_Integer_Type;
Let_Result : Big_Integer_Type;
            procedure Finalizer_Scope_881 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_881;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 6104 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6103 '<FieldAccess for eval_as_int_in_env at ast.py:11352>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11352
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Eval_As_Int_In_Env (Node => Ent.Node, Env => No_Internal_Substitution_Array_Type, E_Info => Ent.Info);
--# expr-done 6103
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 6104
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_881;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_881;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Eval_As_Int;
--# end
--# property-start Expr.eval_as_string /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11376
pragma Warnings (Off, "is not referenced");
 function Expr_P_Eval_As_String
  (Node : Bare_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return String_Type
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Expr :=
     Internal_Entity_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : String_Type;
            procedure Finalizer_Scope_884 with Inline_Always;
      Fld : String_Type;
Let_Result : String_Type;
            procedure Finalizer_Scope_884 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_884;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 6106 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6105 '<FieldAccess for eval_as_string_in_env at ast.py:11392>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11392
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Eval_As_String_In_Env (Node => Ent.Node, Env => No_Internal_Substitution_Array_Type, E_Info => Ent.Info);
--# expr-done 6105
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 6106
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_884;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_884;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Eval_As_String;
--# end
--# property-start Expr.discrete_range /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11416
pragma Warnings (Off, "is not referenced");
 function Expr_P_Discrete_Range
  (Node : Bare_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Discrete_Range
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Expr :=
     Internal_Entity_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Discrete_Range;
      Match_0 : Internal_Entity_Attribute_Ref;
Match_1 : Internal_Entity_Name;
Match_2 : Internal_Entity_Bin_Op;
Match_3 : Internal_Entity_Expr;
Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Discrete_Range;
Fld_2 : Internal_Entity_Base_Type_Decl;
Fld_3 : Internal_Discrete_Range;
Scope_Result : Internal_Discrete_Range;
Result_Var : Internal_Discrete_Range;
Fld_4 : Internal_Entity_Expr;
Fld_5 : Internal_Entity_Expr;
New_Struct : Internal_Discrete_Range;
Match_Prefix : Internal_Entity_Expr;
Cast_Result : Internal_Entity_Attribute_Ref;
Let_Result : Internal_Discrete_Range;
Scope_Result_1 : Internal_Discrete_Range;
Cast_Result_1 : Internal_Entity_Name;
Let_Result_1 : Internal_Discrete_Range;
Scope_Result_2 : Internal_Discrete_Range;
Cast_Result_2 : Internal_Entity_Bin_Op;
Let_Result_2 : Internal_Discrete_Range;
Scope_Result_3 : Internal_Discrete_Range;
Cast_Result_3 : Internal_Entity_Expr;
Let_Result_3 : Internal_Discrete_Range;
Scope_Result_4 : Internal_Discrete_Range;
Match_Result : Internal_Discrete_Range;
Let_Result_4 : Internal_Discrete_Range;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6115 '<Block at ???>' Let_Result_4 None
--# scope-start
--# expr-start 6114 '<Match at ast.py:11420>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11420
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Expr (Match_Prefix.Node.Kind) is
         when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Attribute_Ref
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind ar Match_0
--# expr-start 6108 '<FieldAccess for discrete_range at ast.py:11422>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11422
--# expr-start 6107 '<FieldAccess for prefix at ast.py:11422>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11422
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Match_0.Node.Attribute_Ref_F_Prefix, Info => Match_0.Info);
--# expr-done 6107
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.discrete_range
Fld_1 := Libadalang.Implementation.Expr_P_Discrete_Range (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 6108
Let_Result := Fld_1; 
--# end
Scope_Result_1 := Let_Result;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Call_Expr .. Ada_Target_Name =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind n Match_1
--# expr-start 6109 '<FieldAccess for name_designated_type at ast.py:11424>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11424
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_2 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Match_1.Node, E_Info => Match_1.Info);
--# end
--# expr-done 6109
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind dt Var_Expr
--# expr-start 6110 '<FieldAccess for discrete_range at ast.py:11425>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11425
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.discrete_range'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Discrete_Range (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 6110
Scope_Result := Fld_3;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Discrete_Range;
end if;
Let_Result_1 := Result_Var; 
--# end
Scope_Result_2 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Bin_Op .. Ada_Relation_Op =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Bin_Op
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind bo Match_2
--# expr-start 6113 '<New DiscreteRange at ast.py:11428>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11428
--# expr-start 6111 '<FieldAccess for right at ast.py:11429>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11429
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Expr (Node => Match_2.Node.Bin_Op_F_Right, Info => Match_2.Info);
--# expr-done 6111
--# expr-start 6112 '<FieldAccess for left at ast.py:11429>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11429
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Expr (Node => Match_2.Node.Bin_Op_F_Left, Info => Match_2.Info);
--# expr-done 6112
New_Struct := (High_Bound => Fld_5, Low_Bound => Fld_4); 
--# expr-done 6113
Let_Result_2 := New_Struct; 
--# end
Scope_Result_3 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_3; 
         when Ada_Abstract_State_Decl_Expr .. Ada_Null_Record_Aggregate | Ada_Box_Expr .. Ada_Membership_Expr | Ada_Paren_Expr .. Ada_Un_Op =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_3; 
--# bind _ Match_3
Let_Result_3 := No_Discrete_Range; 
--# end
Scope_Result_4 := Let_Result_3;
   --# end
            Match_Result := Scope_Result_4; 
end case;
--# expr-done 6114
Let_Result_4 := Match_Result; 
--# end
--# expr-done 6115
      Property_Result := Let_Result_4;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Discrete_Range;
--# end
--# property-start '[dispatcher]Expr.designated_env_no_overloading' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Expr_P_Designated_Env_No_Overloading
  (Node : Bare_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
      ; No_Visibility : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Expr :=
     Internal_Entity_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   --# bind no_visibility No_Visibility
   Property_Result : Lexical_Env;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Expr (Self.Kind) is
               when Ada_Abstract_State_Decl_Expr .. Ada_Discrete_Subtype_Name | Ada_End_Name .. Ada_Reduce_Attribute_Ref | Ada_Null_Literal .. Ada_Un_Op =>
                  --# property-call-start Expr.designated_env_no_overloading
                  Property_Result := Expr_P_Designated_Env_No_Overloading
                    (Self
                        , Env
                        , Origin
                        , No_Visibility
                        , E_Info
                    );
                  --# end
               when Ada_Dotted_Name =>
                  --# property-call-start DottedName.designated_env_no_overloading
                  Property_Result := Dotted_Name_P_Designated_Env_No_Overloading
                    (Self
                        , Env
                        , Origin
                        , No_Visibility
                        , E_Info
                    );
                  --# end
               when Ada_Char_Literal .. Ada_String_Literal =>
                  --# property-call-start BaseId.designated_env_no_overloading
                  Property_Result := Base_Id_P_Designated_Env_No_Overloading
                    (Self
                        , Env
                        , Origin
                        , No_Visibility
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Expr_P_Designated_Env_No_Overloading;
--# end
--# property-start '[dispatcher]Expr.designated_env' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Expr_P_Designated_Env
  (Node : Bare_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
      ; No_Visibility : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   --# bind no_visibility No_Visibility
   Property_Result : Lexical_Env;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Expr (Self.Kind) is
               when Ada_Abstract_State_Decl_Expr .. Ada_Membership_Expr | Ada_Defining_Name .. Ada_Discrete_Subtype_Name | Ada_End_Name | Ada_Reduce_Attribute_Ref | Ada_Null_Literal .. Ada_Un_Op =>
                  --# property-call-start Expr.designated_env
                  Property_Result := Expr_P_Designated_Env
                    (Self
                        , Env
                        , Origin
                        , No_Visibility
                        , E_Info
                    );
                  --# end
               when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref =>
                  --# property-call-start AttributeRef.designated_env
                  Property_Result := Attribute_Ref_P_Designated_Env
                    (Self
                        , Env
                        , Origin
                        , No_Visibility
                        , E_Info
                    );
                  --# end
               when Ada_Call_Expr =>
                  --# property-call-start CallExpr.designated_env
                  Property_Result := Call_Expr_P_Designated_Env
                    (Self
                        , Env
                        , Origin
                        , No_Visibility
                        , E_Info
                    );
                  --# end
               when Ada_Dotted_Name =>
                  --# property-call-start DottedName.designated_env
                  Property_Result := Dotted_Name_P_Designated_Env
                    (Self
                        , Env
                        , Origin
                        , No_Visibility
                        , E_Info
                    );
                  --# end
               when Ada_Explicit_Deref =>
                  --# property-call-start ExplicitDeref.designated_env
                  Property_Result := Explicit_Deref_P_Designated_Env
                    (Self
                        , Env
                        , Origin
                        , No_Visibility
                        , E_Info
                    );
                  --# end
               when Ada_Qual_Expr =>
                  --# property-call-start QualExpr.designated_env
                  Property_Result := Qual_Expr_P_Designated_Env
                    (Self
                        , Env
                        , Origin
                        , No_Visibility
                        , E_Info
                    );
                  --# end
               when Ada_Char_Literal .. Ada_String_Literal =>
                  --# property-call-start BaseId.designated_env
                  Property_Result := Base_Id_P_Designated_Env
                    (Self
                        , Env
                        , Origin
                        , No_Visibility
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Expr_P_Designated_Env;
--# end
--# property-start Expr.env_elements /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11459
pragma Warnings (Off, "is not referenced");
 function Expr_P_Env_Elements
  (Node : Bare_Expr
      ; Env : Lexical_Env
      ; No_Visibility : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Expr :=
     Internal_Entity_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind no_visibility No_Visibility
   Property_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_889 with Inline_Always;
      Fld : Internal_Entity_Array_Access;
Fld_1 : Boolean;
Map_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_889 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_889;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6118 '<Filter at ast.py:11460>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11460
   --# expr-start 6116 '<FieldAccess for env_elements_impl at ast.py:11460>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11460
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.env_elements_impl'
Fld := Libadalang.Implementation.Dispatcher_Expr_P_Env_Elements_Impl (Node => Ent.Node, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 6116
      declare
         Map_Result_Vec : Internal_Entity_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Fld;
      begin
         for Item_217 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_217
         --# expr-start 6117 '<FieldAccess for has_visibility at ast.py:11460>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11460
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.has_visibility
Fld_1 := Libadalang.Implementation.Ada_Node_P_Has_Visibility (Node => Self, Other_Entity => Item_217);
--# end
--# expr-done 6117
         if Fld_1 then
         declare
            Item_To_Append : constant Internal_Entity := Item_217;
         begin
            Internal_Entity_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 6118
      Property_Result := Map_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_889;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_889;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Env_Elements;
--# end
--# property-start '[dispatcher]Expr.env_elements_impl' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Expr_P_Env_Elements_Impl
  (Node : Bare_Expr
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   Property_Result : Internal_Entity_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Expr (Self.Kind) is
               when Ada_Abstract_State_Decl_Expr .. Ada_Membership_Expr | Ada_Discrete_Subtype_Name | Ada_Reduce_Attribute_Ref | Ada_Null_Literal .. Ada_Un_Op =>
                  --# property-call-start Expr.env_elements_impl
                  Property_Result := Expr_P_Env_Elements_Impl
                    (Self
                        , Env
                        , E_Info
                    );
                  --# end
               when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref =>
                  --# property-call-start AttributeRef.env_elements_impl
                  Property_Result := Attribute_Ref_P_Env_Elements_Impl
                    (Self
                        , Env
                        , E_Info
                    );
                  --# end
               when Ada_Call_Expr =>
                  --# property-call-start CallExpr.env_elements_impl
                  Property_Result := Call_Expr_P_Env_Elements_Impl
                    (Self
                        , Env
                        , E_Info
                    );
                  --# end
               when Ada_Defining_Name .. Ada_Synthetic_Defining_Name =>
                  --# property-call-start DefiningName.env_elements_impl
                  Property_Result := Defining_Name_P_Env_Elements_Impl
                    (Self
                        , Env
                        , E_Info
                    );
                  --# end
               when Ada_Dotted_Name =>
                  --# property-call-start DottedName.env_elements_impl
                  Property_Result := Dotted_Name_P_Env_Elements_Impl
                    (Self
                        , Env
                        , E_Info
                    );
                  --# end
               when Ada_End_Name =>
                  --# property-call-start EndName.env_elements_impl
                  Property_Result := End_Name_P_Env_Elements_Impl
                    (Self
                        , Env
                        , E_Info
                    );
                  --# end
               when Ada_Explicit_Deref =>
                  --# property-call-start ExplicitDeref.env_elements_impl
                  Property_Result := Explicit_Deref_P_Env_Elements_Impl
                    (Self
                        , Env
                        , E_Info
                    );
                  --# end
               when Ada_Qual_Expr =>
                  --# property-call-start QualExpr.env_elements_impl
                  Property_Result := Qual_Expr_P_Env_Elements_Impl
                    (Self
                        , Env
                        , E_Info
                    );
                  --# end
               when Ada_Char_Literal .. Ada_String_Literal =>
                  --# property-call-start BaseId.env_elements_impl
                  Property_Result := Base_Id_P_Env_Elements_Impl
                    (Self
                        , Env
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Expr_P_Env_Elements_Impl;
--# end
--# property-start Expr.matching_nodes /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11475
pragma Warnings (Off, "is not referenced");
 function Expr_P_Matching_Nodes
  (Node : Bare_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Expr :=
     Internal_Entity_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_891 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_1 : Internal_Entity_Array_Access;
Dyn_Var_Bind_Result : Internal_Entity_Array_Access;
Let_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_891 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_891;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6122 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6121 '<bind at ast.py:11480>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11480
--# expr-start 6119 '<FieldAccess for node_env at ast.py:11480>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11480
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Node_Env (Node => Self);
--# expr-done 6119
Bound_Env := Fld; Inc_Ref (Bound_Env);
--# expr-start 6120 '<FieldAccess for env_elements at ast.py:11480>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11480
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.env_elements
Fld_1 := Libadalang.Implementation.Expr_P_Env_Elements (Node => Ent.Node, Env => Bound_Env, No_Visibility => False, E_Info => Ent.Info);
--# end
--# expr-done 6120
Dyn_Var_Bind_Result := Fld_1; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 6121
Let_Result := Dyn_Var_Bind_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 6122
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_891;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_891;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Matching_Nodes;
--# end
--# property-start Expr.has_context_free_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11007
pragma Warnings (Off, "is not referenced");
 function Expr_P_Has_Context_Free_Type
  (Node : Bare_Expr
  )
   return Boolean
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 6123 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := True; 
--# end
--# expr-done 6123
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Has_Context_Free_Type;
--# end
 function Expr_P_Potential_Actuals_For_Dispatch
  (Node : Bare_Expr
      ; Spec : Internal_Entity_Base_Subp_Spec
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Expected_Type_For_Expr_Array_Access
is (raise Property_Error
    with "Property Expr.potential_actuals_for_dispatch not implemented on type "
    & Kind_Name (Node));
--# property-start Expr.is_dispatching_call /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11233
pragma Warnings (Off, "is not referenced");
 function Expr_P_Is_Dispatching_Call
  (Node : Bare_Expr
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 6124 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := raise Property_Error with "Invalid node type: expected Name, UnOp or BinOp"; 
--# end
--# expr-done 6124
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Is_Dispatching_Call;
--# end
--# property-start Expr.first_corresponding_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11328
pragma Warnings (Off, "is not referenced");
 function Expr_P_First_Corresponding_Decl
  (Node : Bare_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Basic_Decl;
      Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 6125 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := No_Entity_Basic_Decl; 
--# end
--# expr-done 6125
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_First_Corresponding_Decl;
--# end
--# property-start Expr.designated_env_no_overloading /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11436
pragma Warnings (Off, "is not referenced");
 function Expr_P_Designated_Env_No_Overloading
  (Node : Bare_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
      ; No_Visibility : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Expr := Bare_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Expr :=
     Internal_Entity_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   --# bind no_visibility No_Visibility
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_887 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_887 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_887;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6127 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6126 '<FieldAccess for designated_env at ast.py:11443>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11443
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.designated_env'
Fld := Libadalang.Implementation.Dispatcher_Expr_P_Designated_Env (Node => Ent.Node, Env => Env, Origin => Origin, No_Visibility => No_Visibility, E_Info => Ent.Info);
--# end
--# expr-done 6126
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 6127
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_887;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_887;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Expr_P_Designated_Env_No_Overloading;
--# end
 function Expr_P_Designated_Env
  (Node : Bare_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
      ; No_Visibility : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is (raise Property_Error
    with "Property Expr.designated_env not implemented on type "
    & Kind_Name (Node));
 function Expr_P_Env_Elements_Impl
  (Node : Bare_Expr
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is (raise Property_Error
    with "Property Expr.env_elements_impl not implemented on type "
    & Kind_Name (Node));
   --
   --  Primitives for Bare_Abstract_State_Decl_Expr
   --
      procedure Initialize_Fields_For_Abstract_State_Decl_Expr
        (Self : Bare_Abstract_State_Decl_Expr
         ; Abstract_State_Decl_Expr_F_State_Decl : Bare_Ada_Node
        ) is
      begin
            Initialize_Fields_For_Expr
              (Self);
            Self.Abstract_State_Decl_Expr_F_State_Decl := Abstract_State_Decl_Expr_F_State_Decl;
      end Initialize_Fields_For_Abstract_State_Decl_Expr;
   function Abstract_State_Decl_Expr_F_State_Decl
     (Node : Bare_Abstract_State_Decl_Expr) return Bare_Ada_Node
   is
   begin
         return Node.Abstract_State_Decl_Expr_F_State_Decl;
   end;
--# property-start AbstractStateDeclExpr.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20366
pragma Warnings (Off, "is not referenced");
 function Abstract_State_Decl_Expr_P_Xref_Equation
  (Node : Bare_Abstract_State_Decl_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Abstract_State_Decl_Expr := Bare_Abstract_State_Decl_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1415 with Inline_Always;
      True_Rel : Logic_Equation;
            procedure Finalizer_Scope_1415 is
            begin
                     Dec_Ref (True_Rel);
            end Finalizer_Scope_1415;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:20366") else null)); 
      Property_Result := True_Rel;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1415;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1415;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Abstract_State_Decl_Expr_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Allocator
   --
      procedure Initialize_Fields_For_Allocator
        (Self : Bare_Allocator
         ; Allocator_F_Subpool : Bare_Name
         ; Allocator_F_Type_Or_Expr : Bare_Ada_Node
        ) is
      begin
            Initialize_Fields_For_Expr
              (Self);
            Self.Allocator_F_Subpool := Allocator_F_Subpool;
            Self.Allocator_F_Type_Or_Expr := Allocator_F_Type_Or_Expr;
      end Initialize_Fields_For_Allocator;
   function Allocator_F_Subpool
     (Node : Bare_Allocator) return Bare_Name
   is
   begin
         return Node.Allocator_F_Subpool;
   end;
   function Allocator_F_Type_Or_Expr
     (Node : Bare_Allocator) return Bare_Ada_Node
   is
   begin
         return Node.Allocator_F_Type_Or_Expr;
   end;
--# property-start Allocator.has_context_free_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17178
pragma Warnings (Off, "is not referenced");
 function Allocator_P_Has_Context_Free_Type
  (Node : Bare_Allocator
  )
   return Boolean
is
   Self : Bare_Allocator := Bare_Allocator (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Allocator_P_Has_Context_Free_Type;
--# end
--# property-start Allocator.get_allocated_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17181
pragma Warnings (Off, "is not referenced");
 function Allocator_P_Get_Allocated_Type
  (Node : Bare_Allocator
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Allocator := Bare_Allocator (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Allocator :=
     Internal_Entity_Allocator'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Type_Decl;
      Match_0 : Internal_Entity_Subtype_Indication;
Match_1 : Internal_Entity_Qual_Expr;
Match_2 : Internal_Entity;
Fld : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_1 : Internal_Entity;
Fld_2 : Internal_Entity_Base_Type_Decl;
Fld_3 : Internal_Entity_Base_Type_Decl;
Match_Prefix : Internal_Entity;
Cast_Result : Internal_Entity_Subtype_Indication;
Let_Result : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity_Qual_Expr;
Let_Result_1 : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Let_Result_2 : Internal_Entity_Base_Type_Decl;
Scope_Result_2 : Internal_Entity_Base_Type_Decl;
Match_Result : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Base_Type_Decl;
Let_Result_3 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6134 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 6133 '<bind at ast.py:17185>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17185
--# expr-start 6128 '<FieldAccess for origin_node at ast.py:17185>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17185
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 6128
Bound_Origin := Fld; 
--# expr-start 6132 '<Match at ast.py:17185>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17185
--# expr-start 6129 '<FieldAccess for type_or_expr at ast.py:17185>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17185
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity (Node => Ent.Node.Allocator_F_Type_Or_Expr, Info => Ent.Info);
--# expr-done 6129
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld_1; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Subtype_Indication .. Ada_Discrete_Subtype_Indication =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Subtype_Indication
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind t Match_0
--# expr-start 6130 '<FieldAccess for designated_type at ast.py:17186>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17186
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Match_0.Node, Origin => Bound_Origin, E_Info => Match_0.Info);
--# end
--# expr-done 6130
Let_Result := Fld_2; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Qual_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Qual_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind q Match_1
--# expr-start 6131 '<FieldAccess for designated_type at ast.py:17187>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17187
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start QualExpr.designated_type
Fld_3 := Libadalang.Implementation.Qual_Expr_P_Designated_Type (Node => Match_1.Node, E_Info => Match_1.Info);
--# end
--# expr-done 6131
Let_Result_1 := Fld_3; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abort_Absent .. Ada_Explicit_Deref | Ada_Reduce_Attribute_Ref .. Ada_Enum_Lit_Synth_Type_Expr | Ada_Synthetic_Type_Expr .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind _ Match_2
Let_Result_2 := No_Entity_Base_Type_Decl; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 6132
Dyn_Var_Bind_Result := Match_Result; 
--# expr-done 6133
Let_Result_3 := Dyn_Var_Bind_Result; 
--# end
--# expr-done 6134
      Property_Result := Let_Result_3;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Allocator_P_Get_Allocated_Type;
--# end
--# property-start Allocator.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17192
pragma Warnings (Off, "is not referenced");
 function Allocator_P_Xref_Equation
  (Node : Bare_Allocator
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Allocator := Bare_Allocator (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Allocator :=
     Internal_Entity_Allocator'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1228 with Inline_Always;
      Fld : Internal_Entity;
Fld_1 : Logic_Equation;
Fld_2 : Logic_Var;
Fld_3 : Logic_Var;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_4 : Logic_Var;
Fld_5 : Internal_Entity_Base_Type_Decl;
Pred : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1228 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Pred);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1228;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6145 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6144 '<And at ast.py:17194>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17194
--# expr-start 6140 '<And at ast.py:17194>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17194
--# expr-start 6136 '<FieldAccess for sub_equation at ast.py:17194>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17194
--# expr-start 6135 '<FieldAccess for type_or_expr at ast.py:17194>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17194
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity (Node => Ent.Node.Allocator_F_Type_Or_Expr, Info => Ent.Info);
--# expr-done 6135
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 6136
--# expr-start 6139 '<Bind at ast.py:17195>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17195
--# expr-start 6137 '<FieldAccess for expected_type_var at ast.py:17195>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17195
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6137
Fld_2.Value := No_Entity;
Entity_Vars.Reset (Fld_2);
--# expr-start 6138 '<FieldAccess for type_var at ast.py:17195>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17195
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6138
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
Bind_Result := Solver.Create_Unify (Fld_2, Fld_3, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17195") else null)); 
--# expr-done 6139
And_Pred := Create_And (Fld_1, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17194") else null)); 
--# expr-done 6140
--# expr-start 6143 '<Predicate on BaseTypeDecl.matching_allocator_type at ast.py:17196>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17196
--# expr-start 6141 '<FieldAccess for type_var at ast.py:17197>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17197
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6141
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
--# expr-start 6142 '<FieldAccess for get_allocated_type at ast.py:17197>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17197
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Allocator.get_allocated_type
Fld_5 := Libadalang.Implementation.Allocator_P_Get_Allocated_Type (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 6142
Pred := Solver.Create_Predicate (Fld_4, Create_Base_Type_Decl_P_Matching_Allocator_Type_0_Predicate (Fld_5, Origin)); 
--# expr-done 6143
And_Pred_1 := Create_And (And_Pred, Pred, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17194") else null)); 
--# expr-done 6144
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 6145
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1228;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1228;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Allocator_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Base_Aggregate
   --
      procedure Initialize_Fields_For_Base_Aggregate
        (Self : Bare_Base_Aggregate
         ; Base_Aggregate_F_Ancestor_Expr : Bare_Expr
         ; Base_Aggregate_F_Assocs : Bare_Assoc_List
        ) is
      begin
            Initialize_Fields_For_Expr
              (Self);
            Self.Base_Aggregate_F_Ancestor_Expr := Base_Aggregate_F_Ancestor_Expr;
            Self.Base_Aggregate_F_Assocs := Base_Aggregate_F_Assocs;
      end Initialize_Fields_For_Base_Aggregate;
   function Base_Aggregate_F_Ancestor_Expr
     (Node : Bare_Base_Aggregate) return Bare_Expr
   is
   begin
         return Node.Base_Aggregate_F_Ancestor_Expr;
   end;
   function Base_Aggregate_F_Assocs
     (Node : Bare_Base_Aggregate) return Bare_Assoc_List
   is
   begin
         return Node.Base_Aggregate_F_Assocs;
   end;
--# property-start BaseAggregate.has_context_free_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12076
pragma Warnings (Off, "is not referenced");
 function Base_Aggregate_P_Has_Context_Free_Type
  (Node : Bare_Base_Aggregate
  )
   return Boolean
is
   Self : Bare_Base_Aggregate := Bare_Base_Aggregate (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Aggregate_P_Has_Context_Free_Type;
--# end
--# property-start BaseAggregate.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12079
pragma Warnings (Off, "is not referenced");
 function Base_Aggregate_P_Xref_Equation
  (Node : Bare_Base_Aggregate
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Base_Aggregate := Bare_Base_Aggregate (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Aggregate :=
     Internal_Entity_Base_Aggregate'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_919 with Inline_Always;
            procedure Finalizer_Scope_2629 with Inline_Always;
            procedure Finalizer_Scope_2630 with Inline_Always;
      Block_Var : Logic_Equation;
Var_Expr : Internal_Entity_Expr;
True_Rel : Logic_Equation;
Fld : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_1 : Logic_Var;
Pred : Logic_Equation;
Fld_2 : Logic_Var;
Fld_3 : Logic_Var;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Dyn_Var_Bind_Result : Logic_Equation;
Sym : Symbol_Type;
Fld_4 : Boolean;
Sym_1 : Symbol_Type;
Fld_5 : Boolean;
If_Result : Boolean;
Sym_2 : Symbol_Type;
Fld_6 : Boolean;
If_Result_1 : Boolean;
Sym_3 : Symbol_Type;
Fld_7 : Boolean;
If_Result_2 : Boolean;
Sym_4 : Symbol_Type;
Fld_8 : Boolean;
If_Result_3 : Boolean;
Sym_5 : Symbol_Type;
Fld_9 : Boolean;
If_Result_4 : Boolean;
Fld_10 : Boolean;
If_Result_5 : Boolean;
If_Result_6 : Logic_Equation;
Fld_11 : Internal_Entity_Expr;
Fld_12 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel_1 : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_13 : Internal_Entity_Assoc_List;
Fld_14 : Bare_Assoc_List;
Fld_15 : Internal_Entity_Info;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result_7 : Internal_Entity_Basic_Assoc;
Fld_16 : Logic_Equation;
Item_218 : Internal_Entity_Basic_Assoc;
Bare_Item_218 : Bare_Basic_Assoc;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
And_Pred_2 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_919 is
            begin
                     Dec_Ref (True_Rel);
                     Dec_Ref (Pred);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (If_Result_6);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel_1);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_919;
            procedure Finalizer_Scope_2629 is
            begin
                     Dec_Ref (Fld_12);
            end Finalizer_Scope_2629;
            procedure Finalizer_Scope_2630 is
            begin
                     Dec_Ref (Fld_16);
            end Finalizer_Scope_2630;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6176 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6167 '<If at ast.py:12085>' If_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12085
--# expr-start 6147 '<FieldAccess for in_aspect at ast.py:12089>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12089
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6146 '<Symbol Global at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Global); 
--# expr-done 6146
--# property-call-start AdaNode.in_aspect
Fld_4 := Libadalang.Implementation.Ada_Node_P_In_Aspect (Node => Self, Name => Sym);
--# end
--# expr-done 6147
if Fld_4 then
   If_Result := True;
else
   --# expr-start 6149 '<FieldAccess for in_aspect at ast.py:12089>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12089
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6148 '<Symbol Refined_Global at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Refined_Global); 
--# expr-done 6148
--# property-call-start AdaNode.in_aspect
Fld_5 := Libadalang.Implementation.Ada_Node_P_In_Aspect (Node => Self, Name => Sym_1);
--# end
--# expr-done 6149
   If_Result := Fld_5;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 6151 '<FieldAccess for in_aspect at ast.py:12090>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12090
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6150 '<Symbol Depends at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Depends); 
--# expr-done 6150
--# property-call-start AdaNode.in_aspect
Fld_6 := Libadalang.Implementation.Ada_Node_P_In_Aspect (Node => Self, Name => Sym_2);
--# end
--# expr-done 6151
   If_Result_1 := Fld_6;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
   --# expr-start 6153 '<FieldAccess for in_aspect at ast.py:12090>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12090
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6152 '<Symbol Refined_Depends at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Refined_Depends); 
--# expr-done 6152
--# property-call-start AdaNode.in_aspect
Fld_7 := Libadalang.Implementation.Ada_Node_P_In_Aspect (Node => Self, Name => Sym_3);
--# end
--# expr-done 6153
   If_Result_2 := Fld_7;
end if;
if If_Result_2 then
   If_Result_3 := True;
else
   --# expr-start 6155 '<FieldAccess for in_aspect at ast.py:12091>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12091
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6154 '<Symbol Test_Case at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Test_Case); 
--# expr-done 6154
--# property-call-start AdaNode.in_aspect
Fld_8 := Libadalang.Implementation.Ada_Node_P_In_Aspect (Node => Self, Name => Sym_4);
--# end
--# expr-done 6155
   If_Result_3 := Fld_8;
end if;
if If_Result_3 then
   If_Result_4 := True;
else
   --# expr-start 6157 '<FieldAccess for in_aspect at ast.py:12091>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12091
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6156 '<Symbol Refined_State at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Refined_State); 
--# expr-done 6156
--# property-call-start AdaNode.in_aspect
Fld_9 := Libadalang.Implementation.Ada_Node_P_In_Aspect (Node => Self, Name => Sym_5);
--# end
--# expr-done 6157
   If_Result_4 := Fld_9;
end if;
if If_Result_4 then
   If_Result_5 := True;
else
   --# expr-start 6158 '<FieldAccess for is_contract_cases_base_aggregate at ast.py:12095>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12095
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseAggregate.is_contract_cases_base_aggregate
Fld_10 := Libadalang.Implementation.Base_Aggregate_P_Is_Contract_Cases_Base_Aggregate (Node => Self);
--# end
--# expr-done 6158
   If_Result_5 := Fld_10;
end if;
if If_Result_5 then
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12096") else null)); 
   If_Result_6 := True_Rel;
else
   --# expr-start 6166 '<bind at ast.py:12097>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12097
--# expr-start 6159 '<FieldAccess for origin_node at ast.py:12098>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12098
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 6159
Bound_Origin := Fld; 
--# expr-start 6165 '<And at ast.py:12099>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12099
--# expr-start 6161 '<Predicate on BaseTypeDecl.is_array_or_rec at ast.py:12099>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12099
--# expr-start 6160 '<FieldAccess for expected_type_var at ast.py:12099>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12099
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6160
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
Pred := Solver.Create_Predicate (Fld_1, Create_Base_Type_Decl_P_Is_Array_Or_Rec_0_Predicate (Bound_Origin)); 
--# expr-done 6161
--# expr-start 6164 '<Bind at ast.py:12100>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12100
--# expr-start 6162 '<FieldAccess for expected_type_var at ast.py:12100>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12100
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6162
Fld_2.Value := No_Entity;
Entity_Vars.Reset (Fld_2);
--# expr-start 6163 '<FieldAccess for type_var at ast.py:12100>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12100
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6163
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
Bind_Result := Solver.Create_Unify (Fld_2, Fld_3, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12100") else null)); 
--# expr-done 6164
And_Pred := Create_And (Pred, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12099") else null)); 
--# expr-done 6165
Dyn_Var_Bind_Result := And_Pred; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 6166
   If_Result_6 := Dyn_Var_Bind_Result;
end if;
      Inc_Ref (If_Result_6);
--# expr-done 6167
Block_Var := If_Result_6; Inc_Ref (Block_Var);
--# bind type_constraint Block_Var
--# expr-start 6175 '<And at ast.py:12104>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12104
--# expr-start 6170 '<And at ast.py:12104>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12104
--# expr-start 6168 '<FieldAccess for ancestor_expr at ast.py:12104>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12104
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Create_Internal_Entity_Expr (Node => Ent.Node.Base_Aggregate_F_Ancestor_Expr, Info => Ent.Info);
--# expr-done 6168
Var_Expr := Fld_11;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind ae Var_Expr
--# expr-start 6169 '<FieldAccess for sub_equation at ast.py:12105>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12105
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_12 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 6169
Scope_Result := Fld_12;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2629;
   Result_Var := Scope_Result;
else
   True_Rel_1 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12105") else null)); 
   Result_Var := True_Rel_1;
end if;
      Inc_Ref (Result_Var);
And_Pred_1 := Create_And (Block_Var, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12104") else null)); 
--# expr-done 6170
--# expr-start 6174 '<LogicAll at ast.py:12106>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12106
--# expr-start 6173 '<Map at ast.py:12106>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12106
   --# expr-start 6171 '<FieldAccess for assocs at ast.py:12106>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12106
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Create_Internal_Entity_Assoc_List (Node => Ent.Node.Base_Aggregate_F_Assocs, Info => Ent.Info);
--# expr-done 6171
Fld_14 := Fld_13.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_14 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Assoc_List := Fld_14;
      begin
         for Untyped_Item_218 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_218 := Untyped_Item_218; 
Is_Null := Bare_Item_218 = null; 
if Is_Null then
   If_Result_7 := No_Entity_Basic_Assoc;
else
Fld_15 := Fld_13.Info;
As_Entity := (Info => Fld_15, Node => Bare_Item_218); 
   If_Result_7 := As_Entity;
end if;
                  Item_218 := If_Result_7; 
   --# scope-start
               --# bind assoc Item_218
      --# expr-start 6172 '<FieldAccess for sub_equation at ast.py:12107>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12107
      if Item_218.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_16 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Item_218.Node, Env => Env, Origin => Origin, E_Info => Item_218.Info);
--# end
--# expr-done 6172
         declare
            Item_To_Append : constant Logic_Equation := Fld_16;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2630;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 6173
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12106") else null)); 
--# expr-done 6174
And_Pred_2 := Create_And (And_Pred_1, Logic_Boolean_Op, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12104") else null)); 
--# expr-done 6175
Let_Result := And_Pred_2; Inc_Ref (Let_Result);
--# end
--# expr-done 6176
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_919;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_919;
               Finalizer_Scope_2629;
               Finalizer_Scope_2630;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Aggregate_P_Xref_Equation;
--# end
--# property-start BaseAggregate.is_contract_cases_base_aggregate /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12111
pragma Warnings (Off, "is not referenced");
 function Base_Aggregate_P_Is_Contract_Cases_Base_Aggregate
  (Node : Bare_Base_Aggregate
  )
   return Boolean
is
   Self : Bare_Base_Aggregate := Bare_Base_Aggregate (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Var_Expr : Bare_Aspect_Assoc;
Fld : Bare_Ada_Node;
Cast_Result : Bare_Aspect_Assoc;
Fld_1 : Bare_Name;
Sym : Symbol_Type;
Fld_2 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 6182 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6178 '<Cast to AspectAssoc at ast.py:12116>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12116
--# expr-start 6177 '<parent at ast.py:12116>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12116
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 6177
   if Fld = null
      or else Fld.Kind in Ada_Aspect_Assoc_Range
   then
      Cast_Result := Fld;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 6178
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind aspect Var_Expr
--# expr-start 6181 '<FieldAccess for name_is at ast.py:12117>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12117
--# expr-start 6179 '<FieldAccess for id at ast.py:12117>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12117
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Var_Expr.Aspect_Assoc_F_Id;
--# expr-done 6179
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6180 '<Symbol Contract_Cases at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Contract_Cases); 
--# expr-done 6180
--# property-call-start Name.name_is
Fld_2 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_1, Sym => Sym);
--# end
--# expr-done 6181
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 6182
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Aggregate_P_Is_Contract_Cases_Base_Aggregate;
--# end
--# property-start BaseAggregate.multidim_root_aggregate /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12122
pragma Warnings (Off, "is not referenced");
 function Base_Aggregate_P_Multidim_Root_Aggregate
  (Node : Bare_Base_Aggregate
      ; R : Integer
         := 0
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Multidim_Aggregate_Info
is
   Self : Bare_Base_Aggregate := Bare_Base_Aggregate (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Aggregate :=
     Internal_Entity_Base_Aggregate'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind r R
   --# bind origin Origin
   Property_Result : Internal_Multidim_Aggregate_Info;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Entity_Aggregate;
Fld : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
New_Struct : Internal_Multidim_Aggregate_Info;
Fld_1 : Integer;
Comp_Result : Boolean;
If_Result : Internal_Multidim_Aggregate_Info;
Scope_Result : Internal_Multidim_Aggregate_Info;
Fld_2 : Internal_Entity;
Fld_3 : Internal_Entity;
Fld_4 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Aggregate;
Arith_Result : Integer;
Fld_5 : Internal_Multidim_Aggregate_Info;
Scope_Result_1 : Internal_Multidim_Aggregate_Info;
Result_Var : Internal_Multidim_Aggregate_Info;
Result_Var_1 : Internal_Multidim_Aggregate_Info;
Let_Result : Internal_Multidim_Aggregate_Info;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Aggregate_Base_Aggregate_P_Multidim_Root_Aggregate,
            Items    => new Mmz_Key_Array (1 ..  4))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Aggregate,
                                As_Bare_Base_Aggregate => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Integer,
                                      As_Integer => R);
               Mmz_K.Items (3) := (Kind => Mmz_Bare_Ada_Node,
                                      As_Bare_Ada_Node => Origin);
               Mmz_K.Items (4) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Multidim_Aggregate_Info;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 6194 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6184 '<Cast to BaseTypeDecl at ast.py:12129>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12129
--# expr-start 6183 '<FieldAccess for type_val at ast.py:12129>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12129
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.type_val
Fld := Libadalang.Implementation.Expr_P_Type_Val (Node => Ent.Node);
--# end
--# expr-done 6183
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 6184
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind tv Var_Expr
--# expr-start 6187 '<If at ast.py:12130>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12130
--# expr-start 6185 '<FieldAccess for array_ndims at ast.py:12133>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12133
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.array_ndims'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Array_Ndims (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 6185
Comp_Result := Fld_1 > 1; 
if Comp_Result then
   --# expr-start 6186 '<New MultidimAggregateInfo at ast.py:12134>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12134
New_Struct := (Agg => Ent, Rank => R, Typ => Var_Expr); 
--# expr-done 6186
   If_Result := New_Struct;
else
   If_Result := No_Multidim_Aggregate_Info;
end if;
--# expr-done 6187
Scope_Result := If_Result;
   --# end
   Result_Var_1 := Scope_Result;
else
--# expr-start 6191 '<Cast to Aggregate at ast.py:12142>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12142
--# expr-start 6190 '<parent at ast.py:12142>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12142
--# expr-start 6189 '<parent at ast.py:12142>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12142
--# expr-start 6188 '<parent at ast.py:12142>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12142
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 6188
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Parent (Node => Fld_2.Node, E_Info => Fld_2.Info);
--# expr-done 6189
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Parent (Node => Fld_3.Node, E_Info => Fld_3.Info);
--# expr-done 6190
   if Fld_4.Node = null
      or else Fld_4.Node.Kind in Ada_Aggregate_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Aggregate
        (Node => Fld_4.Node,
         Info => Fld_4.Info);
   else
         Cast_Result_1 := No_Entity_Aggregate;
   end if;
--# expr-done 6191
Var_Expr_1 := Cast_Result_1;
if Var_Expr_1 /= No_Entity_Aggregate then
   --# scope-start
--# expr-start 6193 '<FieldAccess for multidim_root_aggregate at ast.py:12142>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12142
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6192 '<Arithmetic + at ast.py:12143>' Arith_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12143
Arith_Result := (R + 1); 
--# expr-done 6192
--# property-call-start BaseAggregate.multidim_root_aggregate
Fld_5 := Libadalang.Implementation.Base_Aggregate_P_Multidim_Root_Aggregate (Node => Var_Expr_1.Node, R => Arith_Result, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 6193
Scope_Result_1 := Fld_5;
   --# end
   Result_Var := Scope_Result_1;
else
   Result_Var := No_Multidim_Aggregate_Info;
end if;
   Result_Var_1 := Result_Var;
end if;
Let_Result := Result_Var_1; 
--# end
--# expr-done 6194
      Property_Result := Let_Result;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Multidim_Aggregate_Info,
                        As_Internal_Multidim_Aggregate_Info => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Aggregate_P_Multidim_Root_Aggregate;
--# end
--# property-start BaseAggregate.all_discriminants /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12149
pragma Warnings (Off, "is not referenced");
 function Base_Aggregate_P_All_Discriminants
  (Node : Bare_Base_Aggregate
      ; Origin : Bare_Ada_Node
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Base_Aggregate := Bare_Base_Aggregate (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_922 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Block_Var_1 : Internal_Entity_Type_Decl;
Fld : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Record_Def;
Fld_2 : Internal_Entity_Component_List;
Fld_3 : Internal_Entity_Type_Decl;
Fld_4 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Let_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_922 is
            begin
                     Dec_Ref (Fld_4);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_922;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Aggregate_Base_Aggregate_P_All_Discriminants,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Aggregate,
                                As_Bare_Base_Aggregate => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Bare_Ada_Node,
                                      As_Bare_Ada_Node => Origin);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Base_Formal_Param_Decl_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 6201 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6196 '<Cast to BaseTypeDecl at ast.py:12165>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12165
--# expr-start 6195 '<FieldAccess for type_val at ast.py:12165>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12165
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.type_val
Fld := Libadalang.Implementation.Expr_P_Type_Val (Node => Self);
--# end
--# expr-done 6195
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 6196
Block_Var := Cast_Result; 
--# bind td Block_Var
--# expr-start 6199 '<FieldAccess for type_decl at ast.py:12166>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12166
--# expr-start 6198 '<FieldAccess for comps at ast.py:12166>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12166
--# expr-start 6197 '<FieldAccess for record_def at ast.py:12166>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12166
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.record_def'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Record_Def (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 6197
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseRecordDef.comps
Fld_2 := Libadalang.Implementation.Base_Record_Def_P_Comps (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 6198
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.type_decl
Fld_3 := Libadalang.Implementation.Component_List_P_Type_Decl (Node => Fld_2.Node, E_Info => Fld_2.Info);
--# end
--# expr-done 6199
Block_Var_1 := Fld_3; 
--# bind record_decl Block_Var_1
--# expr-start 6200 '<FieldAccess for discriminants_list at ast.py:12167>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12167
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.discriminants_list'
Fld_4 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Discriminants_List (Node => Block_Var_1.Node, Origin => Origin, E_Info => Block_Var_1.Info);
--# end
--# expr-done 6200
Let_Result := Fld_4; Inc_Ref (Let_Result);
--# end
--# expr-done 6201
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_922;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Base_Formal_Param_Decl_Array_Access,
                        As_Internal_Entity_Base_Formal_Param_Decl_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_922;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Aggregate_P_All_Discriminants;
--# end
--# property-start BaseAggregate.all_components /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12172
pragma Warnings (Off, "is not referenced");
 function Base_Aggregate_P_All_Components
  (Node : Bare_Base_Aggregate
      ; Origin : Bare_Ada_Node
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Base_Aggregate := Bare_Base_Aggregate (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Aggregate :=
     Internal_Entity_Base_Aggregate'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind env Env
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_923 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Block_Var_1 : Internal_Entity_Component_List;
Fld : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Record_Def;
Fld_2 : Internal_Entity_Component_List;
Fld_3 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_4 : Internal_Entity_Assoc_List;
Fld_5 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Is_A : Boolean;
If_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Let_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_923 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Fld_5);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_923;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Aggregate_Base_Aggregate_P_All_Components,
            Items    => new Mmz_Key_Array (1 ..  4))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Aggregate,
                                As_Bare_Base_Aggregate => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Bare_Ada_Node,
                                      As_Bare_Ada_Node => Origin);
               Mmz_K.Items (3) := (Kind => Mmz_Lexical_Env,
                                      As_Lexical_Env => Env);
                  Inc_Ref (Mmz_K.Items (3).As_Lexical_Env);
               Mmz_K.Items (4) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Base_Formal_Param_Decl_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 6211 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6203 '<Cast to BaseTypeDecl at ast.py:12181>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12181
--# expr-start 6202 '<FieldAccess for type_val at ast.py:12181>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12181
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.type_val
Fld := Libadalang.Implementation.Expr_P_Type_Val (Node => Self);
--# end
--# expr-done 6202
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 6203
Block_Var := Cast_Result; 
--# bind td Block_Var
--# expr-start 6205 '<FieldAccess for comps at ast.py:12182>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12182
--# expr-start 6204 '<FieldAccess for record_def at ast.py:12182>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12182
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.record_def'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Record_Def (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 6204
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseRecordDef.comps
Fld_2 := Libadalang.Implementation.Base_Record_Def_P_Comps (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 6205
Block_Var_1 := Fld_2; 
--# bind comp_list Block_Var_1
--# expr-start 6210 '<If at ast.py:12183>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12183
--# expr-start 6206 '<IsA DeltaAggregate at ast.py:12184>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12184
Is_A := Ent.Node /= null 
and then Ent.Node.Kind in Ada_Delta_Aggregate_Range; 
--# expr-done 6206
if Is_A then
   --# expr-start 6207 '<FieldAccess for abstract_formal_params_for_delta_assocs at ast.py:12187>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12187
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.abstract_formal_params_for_delta_assocs
Fld_3 := Libadalang.Implementation.Component_List_P_Abstract_Formal_Params_For_Delta_Assocs (Node => Block_Var_1.Node, E_Info => Block_Var_1.Info);
--# end
--# expr-done 6207
   If_Result := Fld_3;
else
   --# expr-start 6209 '<FieldAccess for abstract_formal_params_for_assocs at ast.py:12188>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12188
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6208 '<FieldAccess for assocs at ast.py:12188>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12188
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Assoc_List (Node => Ent.Node.Base_Aggregate_F_Assocs, Info => Ent.Info);
--# expr-done 6208
--# property-call-start ComponentList.abstract_formal_params_for_assocs
Fld_5 := Libadalang.Implementation.Component_List_P_Abstract_Formal_Params_For_Assocs (Node => Block_Var_1.Node, Assocs => Fld_4, Env => Env, Origin => Origin, E_Info => Block_Var_1.Info);
--# end
--# expr-done 6209
   If_Result := Fld_5;
end if;
      Inc_Ref (If_Result);
--# expr-done 6210
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 6211
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_923;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Base_Formal_Param_Decl_Array_Access,
                        As_Internal_Entity_Base_Formal_Param_Decl_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_923;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Aggregate_P_All_Components;
--# end
--# property-start BaseAggregate.matched_discriminants /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12193
pragma Warnings (Off, "is not referenced");
 function Base_Aggregate_P_Matched_Discriminants
  (Node : Bare_Base_Aggregate
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Param_Match_Array_Access
is
   Self : Bare_Base_Aggregate := Bare_Base_Aggregate (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Aggregate :=
     Internal_Entity_Base_Aggregate'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Param_Match_Array_Access;
            procedure Finalizer_Scope_924 with Inline_Always;
      Fld : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_1 : Internal_Entity_Assoc_List;
Fld_2 : Internal_Param_Match_Array_Access;
Let_Result : Internal_Param_Match_Array_Access;
            procedure Finalizer_Scope_924 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_924;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Aggregate_Base_Aggregate_P_Matched_Discriminants,
            Items    => new Mmz_Key_Array (1 ..  3))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Aggregate,
                                As_Bare_Base_Aggregate => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Bare_Ada_Node,
                                      As_Bare_Ada_Node => Origin);
               Mmz_K.Items (3) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Param_Match_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 6215 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6214 '<FieldAccess for match_formals at ast.py:12202>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12202
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6212 '<FieldAccess for all_discriminants at ast.py:12203>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12203
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseAggregate.all_discriminants
Fld := Libadalang.Implementation.Base_Aggregate_P_All_Discriminants (Node => Ent.Node, Origin => Origin);
--# end
--# expr-done 6212
--# expr-start 6213 '<FieldAccess for assocs at ast.py:12203>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12203
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Assoc_List (Node => Ent.Node.Base_Aggregate_F_Assocs, Info => Ent.Info);
--# expr-done 6213
--# property-call-start AdaNode.match_formals
Fld_2 := Libadalang.Implementation.Ada_Node_P_Match_Formals (Node => Self, Formal_Params => Fld, Params => Fld_1, Is_Dottable_Subp => False);
--# end
--# expr-done 6214
Let_Result := Fld_2; Inc_Ref (Let_Result);
--# end
--# expr-done 6215
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_924;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Param_Match_Array_Access,
                        As_Internal_Param_Match_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_924;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Aggregate_P_Matched_Discriminants;
--# end
--# property-start BaseAggregate.matched_components /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12208
pragma Warnings (Off, "is not referenced");
 function Base_Aggregate_P_Matched_Components
  (Node : Bare_Base_Aggregate
      ; Origin : Bare_Ada_Node
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Param_Match_Array_Access
is
   Self : Bare_Base_Aggregate := Bare_Base_Aggregate (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Aggregate :=
     Internal_Entity_Base_Aggregate'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind env Env
   Property_Result : Internal_Param_Match_Array_Access;
            procedure Finalizer_Scope_925 with Inline_Always;
      Fld : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_1 : Internal_Entity_Assoc_List;
Fld_2 : Internal_Param_Match_Array_Access;
Let_Result : Internal_Param_Match_Array_Access;
            procedure Finalizer_Scope_925 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_925;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Aggregate_Base_Aggregate_P_Matched_Components,
            Items    => new Mmz_Key_Array (1 ..  4))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Aggregate,
                                As_Bare_Base_Aggregate => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Bare_Ada_Node,
                                      As_Bare_Ada_Node => Origin);
               Mmz_K.Items (3) := (Kind => Mmz_Lexical_Env,
                                      As_Lexical_Env => Env);
                  Inc_Ref (Mmz_K.Items (3).As_Lexical_Env);
               Mmz_K.Items (4) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Param_Match_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 6219 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6218 '<FieldAccess for match_formals at ast.py:12217>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12217
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6216 '<FieldAccess for all_components at ast.py:12218>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12218
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseAggregate.all_components
Fld := Libadalang.Implementation.Base_Aggregate_P_All_Components (Node => Ent.Node, Origin => Origin, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 6216
--# expr-start 6217 '<FieldAccess for assocs at ast.py:12218>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12218
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Assoc_List (Node => Ent.Node.Base_Aggregate_F_Assocs, Info => Ent.Info);
--# expr-done 6217
--# property-call-start AdaNode.match_formals
Fld_2 := Libadalang.Implementation.Ada_Node_P_Match_Formals (Node => Self, Formal_Params => Fld, Params => Fld_1, Is_Dottable_Subp => False);
--# end
--# expr-done 6218
Let_Result := Fld_2; Inc_Ref (Let_Result);
--# end
--# expr-done 6219
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_925;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Param_Match_Array_Access,
                        As_Internal_Param_Match_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_925;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Aggregate_P_Matched_Components;
--# end
--# property-start BaseAggregate.first_unmatched_formal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12223
pragma Warnings (Off, "is not referenced");
 function Base_Aggregate_P_First_Unmatched_Formal
  (Node : Bare_Base_Aggregate
      ; Origin : Bare_Ada_Node
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Base_Aggregate := Bare_Base_Aggregate (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Aggregate :=
     Internal_Entity_Base_Aggregate'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind env Env
   Property_Result : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_926 with Inline_Always;
            procedure Finalizer_Scope_2634 with Inline_Always;
            procedure Finalizer_Scope_2636 with Inline_Always;
      Block_Var : Internal_Entity_Defining_Name;
Fld : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_1 : Internal_Entity_Defining_Name_Array_Access;
Fld_2 : Internal_Param_Match_Array_Access;
Fld_3 : Internal_Entity_Defining_Name;
Cast_Result : Internal_Entity;
Cast_Expr : Internal_Entity_Defining_Name;
Cast_Result_1 : Internal_Entity;
Is_Equiv : Boolean;
Quantifier_Result : Boolean;
Not_Val : Boolean;
Find_Result : Internal_Entity_Defining_Name;
Fld_4 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_5 : Internal_Entity_Defining_Name_Array_Access;
Fld_6 : Internal_Param_Match_Array_Access;
Fld_7 : Internal_Entity_Defining_Name;
Cast_Result_2 : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Defining_Name;
Cast_Result_3 : Internal_Entity;
Is_Equiv_1 : Boolean;
Quantifier_Result_1 : Boolean;
Not_Val_1 : Boolean;
Find_Result_1 : Internal_Entity_Defining_Name;
Is_Null : Boolean;
Not_Val_2 : Boolean;
If_Result : Internal_Entity_Defining_Name;
Let_Result : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_926 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Fld_5);
            end Finalizer_Scope_926;
            procedure Finalizer_Scope_2634 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_2634;
            procedure Finalizer_Scope_2636 is
            begin
                     Dec_Ref (Fld_6);
            end Finalizer_Scope_2636;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6239 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6227 '<Find at ast.py:12234>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12234
   --# expr-start 6221 '<FieldAccess for unpack_formals at ast.py:12234>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12234
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6220 '<FieldAccess for all_discriminants at ast.py:12234>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12234
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseAggregate.all_discriminants
Fld := Libadalang.Implementation.Base_Aggregate_P_All_Discriminants (Node => Ent.Node, Origin => Origin);
--# end
--# expr-done 6220
--# property-call-start AdaNode.unpack_formals
Fld_1 := Libadalang.Implementation.Ada_Node_P_Unpack_Formals (Node => Self, Formal_Params => Fld);
--# end
--# expr-done 6221
      Find_Result := No_Entity_Defining_Name;
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Fld_1;
      begin
         for Item_219 of
               Collection.Items
         loop
   --# scope-start
               --# bind f Item_219
      --# expr-start 6226 '<Not at ast.py:12235>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12235
--# expr-start 6225 '<AnyQuantifier at ast.py:12235>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12235
   --# expr-start 6222 '<FieldAccess for matched_discriminants at ast.py:12235>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12235
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseAggregate.matched_discriminants
Fld_2 := Libadalang.Implementation.Base_Aggregate_P_Matched_Discriminants (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 6222
      Quantifier_Result := False;
      declare
         Collection : constant Internal_Param_Match_Array_Access := Fld_2;
      begin
         for Item_220 of
               Collection.Items
         loop
   --# scope-start
               --# bind m Item_220
      --# expr-start 6224 '<Eq at ast.py:12236>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12236
--# expr-start 6223 '<FieldAccess for formal at ast.py:12236>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12236
Fld_3 := Item_220.Formal;
--# expr-done 6223
      Cast_Result := Create_Internal_Entity
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
Cast_Expr := Item_219; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Is_Equiv := Equivalent (Cast_Result, Cast_Result_1); 
--# expr-done 6224
      Quantifier_Result := Is_Equiv;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
--# expr-done 6225
Not_Val := not (Quantifier_Result); 
--# expr-done 6226
      if Not_Val then
         Find_Result := Item_219;
      end if;
   --# end
      Finalizer_Scope_2634;
               exit when Not_Val;
         end loop;
      end;
--# expr-done 6227
Block_Var := Find_Result; 
--# bind unmatched_discr Block_Var
--# expr-start 6238 '<If at ast.py:12240>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12240
--# expr-start 6229 '<Not at ast.py:12241>' Not_Val_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12241
--# expr-start 6228 '<IsNull at ast.py:12241>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12241
Is_Null := Block_Var.Node = null; 
--# expr-done 6228
Not_Val_2 := not (Is_Null); 
--# expr-done 6229
if Not_Val_2 then
   If_Result := Block_Var;
else
   --# expr-start 6237 '<Find at ast.py:12250>' Find_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12250
   --# expr-start 6231 '<FieldAccess for unpack_formals at ast.py:12250>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12250
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6230 '<FieldAccess for all_components at ast.py:12250>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12250
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseAggregate.all_components
Fld_4 := Libadalang.Implementation.Base_Aggregate_P_All_Components (Node => Ent.Node, Origin => Origin, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 6230
--# property-call-start AdaNode.unpack_formals
Fld_5 := Libadalang.Implementation.Ada_Node_P_Unpack_Formals (Node => Self, Formal_Params => Fld_4);
--# end
--# expr-done 6231
      Find_Result_1 := No_Entity_Defining_Name;
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Fld_5;
      begin
         for Item_221 of
               Collection.Items
         loop
   --# scope-start
               --# bind f Item_221
      --# expr-start 6236 '<Not at ast.py:12251>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12251
--# expr-start 6235 '<AnyQuantifier at ast.py:12251>' Quantifier_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12251
   --# expr-start 6232 '<FieldAccess for matched_components at ast.py:12251>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12251
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseAggregate.matched_components
Fld_6 := Libadalang.Implementation.Base_Aggregate_P_Matched_Components (Node => Ent.Node, Origin => Origin, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 6232
      Quantifier_Result_1 := False;
      declare
         Collection : constant Internal_Param_Match_Array_Access := Fld_6;
      begin
         for Item_222 of
               Collection.Items
         loop
   --# scope-start
               --# bind m Item_222
      --# expr-start 6234 '<Eq at ast.py:12252>' Is_Equiv_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12252
--# expr-start 6233 '<FieldAccess for formal at ast.py:12252>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12252
Fld_7 := Item_222.Formal;
--# expr-done 6233
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_7.Node,
         Info => Fld_7.Info);
Cast_Expr_1 := Item_221; 
      Cast_Result_3 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Is_Equiv_1 := Equivalent (Cast_Result_2, Cast_Result_3); 
--# expr-done 6234
      Quantifier_Result_1 := Is_Equiv_1;
   --# end
               exit when Quantifier_Result_1;
         end loop;
      end;
--# expr-done 6235
Not_Val_1 := not (Quantifier_Result_1); 
--# expr-done 6236
      if Not_Val_1 then
         Find_Result_1 := Item_221;
      end if;
   --# end
      Finalizer_Scope_2636;
               exit when Not_Val_1;
         end loop;
      end;
--# expr-done 6237
   If_Result := Find_Result_1;
end if;
--# expr-done 6238
Let_Result := If_Result; 
--# end
--# expr-done 6239
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_926;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_926;
               Finalizer_Scope_2634;
               Finalizer_Scope_2636;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Aggregate_P_First_Unmatched_Formal;
--# end
--# property-start BaseAggregate.aggregate_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12258
pragma Warnings (Off, "is not referenced");
 function Base_Aggregate_P_Aggregate_Params
  (Node : Bare_Base_Aggregate
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Param_Actual_Array_Access
is
   Self : Bare_Base_Aggregate := Bare_Base_Aggregate (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Aggregate :=
     Internal_Entity_Base_Aggregate'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Param_Actual_Array_Access;
            procedure Finalizer_Scope_928 with Inline_Always;
      Fld : Internal_Entity_Assoc_List;
Fld_1 : Internal_Param_Actual_Array_Access;
Let_Result : Internal_Param_Actual_Array_Access;
            procedure Finalizer_Scope_928 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_928;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6242 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6241 '<FieldAccess for zip_with_params at ast.py:12263>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12263
--# expr-start 6240 '<FieldAccess for assocs at ast.py:12263>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12263
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Assoc_List (Node => Ent.Node.Base_Aggregate_F_Assocs, Info => Ent.Info);
--# expr-done 6240
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AssocList.zip_with_params
Fld_1 := Libadalang.Implementation.Assoc_List_P_Zip_With_Params (Node => Fld.Node, Imprecise_Fallback => False, E_Info => Fld.Info);
--# end
--# expr-done 6241
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 6242
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_928;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_928;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Aggregate_P_Aggregate_Params;
--# end
--# property-start BaseAggregate.is_subaggregate /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12266
pragma Warnings (Off, "is not referenced");
 function Base_Aggregate_P_Is_Subaggregate
  (Node : Bare_Base_Aggregate
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Aggregate := Bare_Base_Aggregate (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Aggregate :=
     Internal_Entity_Base_Aggregate'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Block_Var : Boolean;
Fld : Boolean;
Cast_Expr : Bare_Base_Aggregate;
Cast_Result : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_1 : Internal_Multidim_Aggregate_Info;
Fld_2 : Integer;
Comp_Result : Boolean;
Dyn_Var_Bind_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6247 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6243 '<FieldAccess for resolve_names_from_closest_entry_point at ast.py:12273>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12273
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.resolve_names_from_closest_entry_point
Fld := Libadalang.Implementation.Ada_Node_P_Resolve_Names_From_Closest_Entry_Point (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 6243
Block_Var := Fld; 
--# expr-start 6246 '<bind at ast.py:12274>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12274
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Bound_Origin := Cast_Result; 
--# expr-start 6245 '<FieldAccess for rank at ast.py:12274>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12274
--# expr-start 6244 '<FieldAccess for multidim_root_aggregate at ast.py:12274>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12274
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseAggregate.multidim_root_aggregate
Fld_1 := Libadalang.Implementation.Base_Aggregate_P_Multidim_Root_Aggregate (Node => Ent.Node, Origin => Bound_Origin, E_Info => Ent.Info);
--# end
--# expr-done 6244
Fld_2 := Fld_1.Rank;
--# expr-done 6245
Comp_Result := Fld_2 > 0; 
Dyn_Var_Bind_Result := Comp_Result; 
--# expr-done 6246
Let_Result := Dyn_Var_Bind_Result; 
--# end
--# expr-done 6247
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Aggregate_P_Is_Subaggregate;
--# end
   --
   --  Primitives for Bare_Aggregate
   --
      procedure Initialize_Fields_For_Aggregate
        (Self : Bare_Aggregate
         ; Base_Aggregate_F_Ancestor_Expr : Bare_Expr
         ; Base_Aggregate_F_Assocs : Bare_Assoc_List
        ) is
      begin
            Initialize_Fields_For_Base_Aggregate
              (Self, Base_Aggregate_F_Ancestor_Expr, Base_Aggregate_F_Assocs);
      end Initialize_Fields_For_Aggregate;
   --
   --  Primitives for Bare_Bracket_Aggregate
   --
      procedure Initialize_Fields_For_Bracket_Aggregate
        (Self : Bare_Bracket_Aggregate
         ; Base_Aggregate_F_Ancestor_Expr : Bare_Expr
         ; Base_Aggregate_F_Assocs : Bare_Assoc_List
        ) is
      begin
            Initialize_Fields_For_Aggregate
              (Self, Base_Aggregate_F_Ancestor_Expr, Base_Aggregate_F_Assocs);
      end Initialize_Fields_For_Bracket_Aggregate;
   --
   --  Primitives for Bare_Delta_Aggregate
   --
      procedure Initialize_Fields_For_Delta_Aggregate
        (Self : Bare_Delta_Aggregate
         ; Base_Aggregate_F_Ancestor_Expr : Bare_Expr
         ; Base_Aggregate_F_Assocs : Bare_Assoc_List
        ) is
      begin
            Initialize_Fields_For_Base_Aggregate
              (Self, Base_Aggregate_F_Ancestor_Expr, Base_Aggregate_F_Assocs);
      end Initialize_Fields_For_Delta_Aggregate;
   --
   --  Primitives for Bare_Bracket_Delta_Aggregate
   --
      procedure Initialize_Fields_For_Bracket_Delta_Aggregate
        (Self : Bare_Bracket_Delta_Aggregate
         ; Base_Aggregate_F_Ancestor_Expr : Bare_Expr
         ; Base_Aggregate_F_Assocs : Bare_Assoc_List
        ) is
      begin
            Initialize_Fields_For_Delta_Aggregate
              (Self, Base_Aggregate_F_Ancestor_Expr, Base_Aggregate_F_Assocs);
      end Initialize_Fields_For_Bracket_Delta_Aggregate;
   --
   --  Primitives for Bare_Null_Record_Aggregate
   --
      procedure Initialize_Fields_For_Null_Record_Aggregate
        (Self : Bare_Null_Record_Aggregate
         ; Base_Aggregate_F_Ancestor_Expr : Bare_Expr
         ; Base_Aggregate_F_Assocs : Bare_Assoc_List
        ) is
      begin
            Initialize_Fields_For_Base_Aggregate
              (Self, Base_Aggregate_F_Ancestor_Expr, Base_Aggregate_F_Assocs);
      end Initialize_Fields_For_Null_Record_Aggregate;
   --
   --  Primitives for Bare_Bin_Op
   --
      procedure Initialize_Fields_For_Bin_Op
        (Self : Bare_Bin_Op
         ; Bin_Op_F_Left : Bare_Expr
         ; Bin_Op_F_Op : Bare_Op
         ; Bin_Op_F_Right : Bare_Expr
        ) is
      begin
            Initialize_Fields_For_Expr
              (Self);
            Self.Bin_Op_F_Left := Bin_Op_F_Left;
            Self.Bin_Op_F_Op := Bin_Op_F_Op;
            Self.Bin_Op_F_Right := Bin_Op_F_Right;
      end Initialize_Fields_For_Bin_Op;
   function Bin_Op_F_Left
     (Node : Bare_Bin_Op) return Bare_Expr
   is
   begin
         return Node.Bin_Op_F_Left;
   end;
   function Bin_Op_F_Op
     (Node : Bare_Bin_Op) return Bare_Op
   is
   begin
         return Node.Bin_Op_F_Op;
   end;
   function Bin_Op_F_Right
     (Node : Bare_Bin_Op) return Bare_Expr
   is
   begin
         return Node.Bin_Op_F_Right;
   end;
--# property-start BinOp.has_context_free_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11620
pragma Warnings (Off, "is not referenced");
 function Bin_Op_P_Has_Context_Free_Type
  (Node : Bare_Bin_Op
  )
   return Boolean
is
   Self : Bare_Bin_Op := Bare_Bin_Op (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Op;
Is_A : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 6250 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6249 '<IsA Op.AndThen, Op.OrElse at ast.py:11621>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11621
--# expr-start 6248 '<FieldAccess for op at ast.py:11621>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11621
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Bin_Op_F_Op;
--# expr-done 6248
Is_A := Fld /= null 
and then Fld.Kind in Ada_Op_And_Then_Range | Ada_Op_Or_Else_Range; 
--# expr-done 6249
Let_Result := Is_A; 
--# end
--# expr-done 6250
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Bin_Op_P_Has_Context_Free_Type;
--# end
--# property-start BinOp.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11624
pragma Warnings (Off, "is not referenced");
 function Bin_Op_P_Xref_Equation
  (Node : Bare_Bin_Op
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Bin_Op := Bare_Bin_Op (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Bin_Op :=
     Internal_Entity_Bin_Op'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_904 with Inline_Always;
      Fld : Internal_Entity_Expr;
Fld_1 : Logic_Equation;
Fld_2 : Internal_Entity_Expr;
Fld_3 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_4 : Bare_Op;
Is_A : Boolean;
Fld_5 : Logic_Equation;
Fld_6 : Bare_Op;
Is_A_1 : Boolean;
Fld_7 : Logic_Equation;
Fld_8 : Logic_Equation;
Fld_9 : Logic_Equation;
Or_Pred : Logic_Equation;
If_Result : Logic_Equation;
If_Result_1 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_904 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Fld_3);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_5);
                     Dec_Ref (Fld_7);
                     Dec_Ref (Fld_8);
                     Dec_Ref (Fld_9);
                     Dec_Ref (Or_Pred);
                     Dec_Ref (If_Result);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_904;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6267 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6266 '<And at ast.py:11625>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11625
--# expr-start 6255 '<And at ast.py:11625>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11625
--# expr-start 6252 '<FieldAccess for sub_equation at ast.py:11626>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11626
--# expr-start 6251 '<FieldAccess for left at ast.py:11626>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11626
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Bin_Op_F_Left, Info => Ent.Info);
--# expr-done 6251
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 6252
--# expr-start 6254 '<FieldAccess for sub_equation at ast.py:11627>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11627
--# expr-start 6253 '<FieldAccess for right at ast.py:11627>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11627
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Expr (Node => Ent.Node.Bin_Op_F_Right, Info => Ent.Info);
--# expr-done 6253
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_2.Node, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 6254
And_Pred := Create_And (Fld_1, Fld_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11625") else null)); 
--# expr-done 6255
--# expr-start 6265 '<Cond at ast.py:11628>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11628
--# expr-start 6257 '<IsA Op.DoubleDot at ast.py:11629>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11629
--# expr-start 6256 '<FieldAccess for op at ast.py:11629>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11629
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Self.Bin_Op_F_Op;
--# expr-done 6256
Is_A := Fld_4 /= null 
and then Fld_4.Kind in Ada_Op_Double_Dot_Range; 
--# expr-done 6257
if Is_A then
   --# expr-start 6258 '<FieldAccess for double_dot_equation at ast.py:11630>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11630
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BinOp.double_dot_equation
Fld_5 := Libadalang.Implementation.Bin_Op_P_Double_Dot_Equation (Node => Ent.Node, Origin => Origin, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 6258
   If_Result_1 := Fld_5;
else
--# expr-start 6260 '<IsA Op.AndThen, Op.OrElse at ast.py:11632>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11632
--# expr-start 6259 '<FieldAccess for op at ast.py:11632>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11632
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Self.Bin_Op_F_Op;
--# expr-done 6259
Is_A_1 := Fld_6 /= null 
and then Fld_6.Kind in Ada_Op_And_Then_Range | Ada_Op_Or_Else_Range; 
--# expr-done 6260
if Is_A_1 then
   --# expr-start 6261 '<FieldAccess for test_eq at ast.py:11633>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11633
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BinOp.test_eq
Fld_7 := Libadalang.Implementation.Bin_Op_P_Test_Eq (Node => Self, Origin => Origin);
--# end
--# expr-done 6261
   If_Result := Fld_7;
else
   --# expr-start 6264 '<Or at ast.py:11635>' Or_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11635
--# expr-start 6262 '<FieldAccess for overload_equation at ast.py:11635>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11635
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BinOp.overload_equation
Fld_8 := Libadalang.Implementation.Bin_Op_P_Overload_Equation (Node => Ent.Node, Origin => Origin, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 6262
--# expr-start 6263 '<FieldAccess for no_overload_equation at ast.py:11635>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11635
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BinOp.no_overload_equation'
Fld_9 := Libadalang.Implementation.Dispatcher_Bin_Op_P_No_Overload_Equation (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 6263
Or_Pred := Create_Or (Fld_8, Fld_9, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11635") else null)); 
--# expr-done 6264
   If_Result := Or_Pred;
end if;
      Inc_Ref (If_Result);
   If_Result_1 := If_Result;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 6265
And_Pred_1 := Create_And (And_Pred, If_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11625") else null)); 
--# expr-done 6266
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 6267
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_904;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_904;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Bin_Op_P_Xref_Equation;
--# end
--# property-start BinOp.double_dot_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11639
pragma Warnings (Off, "is not referenced");
 function Bin_Op_P_Double_Dot_Equation
  (Node : Bare_Bin_Op
      ; Origin : Bare_Ada_Node
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Bin_Op := Bare_Bin_Op (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Bin_Op :=
     Internal_Entity_Bin_Op'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind env Env
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_905 with Inline_Always;
      True_Rel : Logic_Equation;
Fld : Logic_Var;
Fld_1 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Or_Pred : Logic_Equation;
Fld_2 : Logic_Var;
Fld_3 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
Or_Pred_1 : Logic_Equation;
Fld_4 : Logic_Var;
Pred : Logic_Equation;
Fld_5 : Logic_Var;
Fld_6 : Logic_Var;
Bind_Result_2 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_7 : Logic_Var;
Fld_8 : Bare_Expr;
Fld_9 : Logic_Var;
Bind_Result_3 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_10 : Logic_Var;
Fld_11 : Bare_Expr;
Fld_12 : Logic_Var;
Bind_Result_4 : Logic_Equation;
And_Pred_2 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_13 : Bare_Expr;
Fld_14 : Logic_Equation;
Fld_15 : Bare_Expr;
Fld_16 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Fld_17 : Logic_Var;
Fld_18 : Logic_Equation;
Or_Pred_2 : Logic_Equation;
And_Pred_5 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_905 is
            begin
                     Dec_Ref (True_Rel);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Or_Pred);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Or_Pred_1);
                     Dec_Ref (Pred);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Bind_Result_3);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Bind_Result_4);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Fld_14);
                     Dec_Ref (Fld_16);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Fld_18);
                     Dec_Ref (Or_Pred_2);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_905;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6302 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6301 '<And at ast.py:11640>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11640
--# expr-start 6292 '<And at ast.py:11640>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11640
--# expr-start 6275 '<Or at ast.py:11640>' Or_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11640
--# expr-start 6271 '<Or at ast.py:11640>' Or_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11640
True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11641") else null)); 
--# expr-start 6270 '<Bind at ast.py:11642>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11642
--# expr-start 6268 '<FieldAccess for expected_type_var at ast.py:11642>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11642
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6268
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
--# expr-start 6269 '<FieldAccess for root_int_type at ast.py:11642>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11642
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.root_int_type
Fld_1 := Libadalang.Implementation.Ada_Node_P_Root_Int_Type (Node => Self);
--# end
--# expr-done 6269
      Cast_Result := Create_Internal_Entity
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
Bind_Result := Solver.Create_Assign (Fld, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11642") else null)); 
--# expr-done 6270
Or_Pred := Create_Or (True_Rel, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11640") else null)); 
--# expr-done 6271
--# expr-start 6274 '<Bind at ast.py:11643>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11643
--# expr-start 6272 '<FieldAccess for expected_type_var at ast.py:11643>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11643
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6272
Fld_2.Value := No_Entity;
Entity_Vars.Reset (Fld_2);
--# expr-start 6273 '<FieldAccess for int_type at ast.py:11643>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11643
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.int_type
Fld_3 := Libadalang.Implementation.Ada_Node_P_Int_Type (Node => Self);
--# end
--# expr-done 6273
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_2, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11643") else null)); 
--# expr-done 6274
Or_Pred_1 := Create_Or (Or_Pred, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11640") else null)); 
--# expr-done 6275
--# expr-start 6291 '<And at ast.py:11644>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11644
--# expr-start 6286 '<And at ast.py:11644>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11644
--# expr-start 6281 '<And at ast.py:11644>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11644
--# expr-start 6277 '<Predicate on AdaNode.is_not_null at ast.py:11645>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11645
--# expr-start 6276 '<FieldAccess for expected_type_var at ast.py:11645>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11645
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6276
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
Pred := Solver.Create_Predicate (Fld_4, Create_Ada_Node_P_Is_Not_Null_0_Predicate); 
--# expr-done 6277
--# expr-start 6280 '<Bind at ast.py:11646>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11646
--# expr-start 6278 '<FieldAccess for expected_type_var at ast.py:11646>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11646
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6278
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
--# expr-start 6279 '<FieldAccess for type_var at ast.py:11646>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11646
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6279
Fld_6.Value := No_Entity;
Entity_Vars.Reset (Fld_6);
Bind_Result_2 := Solver.Create_Unify (Fld_5, Fld_6, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11646") else null)); 
--# expr-done 6280
And_Pred := Create_And (Pred, Bind_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11644") else null)); 
--# expr-done 6281
--# expr-start 6285 '<Bind at ast.py:11647>' Bind_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11647
--# expr-start 6282 '<FieldAccess for type_var at ast.py:11647>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11647
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6282
Fld_7.Value := No_Entity;
Entity_Vars.Reset (Fld_7);
--# expr-start 6284 '<FieldAccess for expected_type_var at ast.py:11647>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11647
--# expr-start 6283 '<FieldAccess for left at ast.py:11647>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11647
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Self.Bin_Op_F_Left;
--# expr-done 6283
      if Fld_8 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_8);
--# expr-done 6284
Fld_9.Value := No_Entity;
Entity_Vars.Reset (Fld_9);
Bind_Result_3 := Solver.Create_Unify (Fld_7, Fld_9, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11647") else null)); 
--# expr-done 6285
And_Pred_1 := Create_And (And_Pred, Bind_Result_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11644") else null)); 
--# expr-done 6286
--# expr-start 6290 '<Bind at ast.py:11648>' Bind_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11648
--# expr-start 6287 '<FieldAccess for type_var at ast.py:11648>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11648
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6287
Fld_10.Value := No_Entity;
Entity_Vars.Reset (Fld_10);
--# expr-start 6289 '<FieldAccess for expected_type_var at ast.py:11648>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11648
--# expr-start 6288 '<FieldAccess for right at ast.py:11648>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11648
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Self.Bin_Op_F_Right;
--# expr-done 6288
      if Fld_11 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_11);
--# expr-done 6289
Fld_12.Value := No_Entity;
Entity_Vars.Reset (Fld_12);
Bind_Result_4 := Solver.Create_Unify (Fld_10, Fld_12, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11648") else null)); 
--# expr-done 6290
And_Pred_2 := Create_And (And_Pred_1, Bind_Result_4, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11644") else null)); 
--# expr-done 6291
And_Pred_3 := Create_And (Or_Pred_1, And_Pred_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11640") else null)); 
--# expr-done 6292
--# expr-start 6300 '<Or at ast.py:11649>' Or_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11649
--# expr-start 6297 '<And at ast.py:11654>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11654
--# expr-start 6294 '<FieldAccess for matches_expected_formal_type at ast.py:11654>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11654
--# expr-start 6293 '<FieldAccess for left at ast.py:11654>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11654
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Self.Bin_Op_F_Left;
--# expr-done 6293
      if Fld_13 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_14 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_13, Origin => Origin);
--# end
--# expr-done 6294
--# expr-start 6296 '<FieldAccess for matches_expected_formal_type at ast.py:11655>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11655
--# expr-start 6295 '<FieldAccess for right at ast.py:11655>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11655
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Self.Bin_Op_F_Right;
--# expr-done 6295
      if Fld_15 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_16 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_15, Origin => Origin);
--# end
--# expr-done 6296
And_Pred_4 := Create_And (Fld_14, Fld_16, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11654") else null)); 
--# expr-done 6297
--# expr-start 6299 '<FieldAccess for numeric_type_from_operands_equation at ast.py:11659>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11659
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6298 '<FieldAccess for type_var at ast.py:11659>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11659
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_17 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6298
--# property-call-start BinOp.numeric_type_from_operands_equation
Fld_18 := Libadalang.Implementation.Bin_Op_P_Numeric_Type_From_Operands_Equation (Node => Ent.Node, Dest_Var => Fld_17, Origin => Origin);
--# end
--# expr-done 6299
Or_Pred_2 := Create_Or (And_Pred_4, Fld_18, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11649") else null)); 
--# expr-done 6300
And_Pred_5 := Create_And (And_Pred_3, Or_Pred_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11640") else null)); 
--# expr-done 6301
Let_Result := And_Pred_5; Inc_Ref (Let_Result);
--# end
--# expr-done 6302
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_905;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_905;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Bin_Op_P_Double_Dot_Equation;
--# end
--# property-start BinOp.test_eq /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11663
pragma Warnings (Off, "is not referenced");
 function Bin_Op_P_Test_Eq
  (Node : Bare_Bin_Op
      ; Origin : Bare_Ada_Node
  )
   return Logic_Equation
is
   Self : Bare_Bin_Op := Bare_Bin_Op (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_906 with Inline_Always;
      Fld : Bare_Expr;
Fld_1 : Logic_Var;
Fld_2 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_3 : Bare_Expr;
Fld_4 : Logic_Var;
Fld_5 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_6 : Bare_Expr;
Fld_7 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_8 : Bare_Expr;
Fld_9 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_10 : Logic_Var;
Fld_11 : Bare_Expr;
Fld_12 : Logic_Var;
Bind_Result_2 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_906 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_7);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_9);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_906;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6323 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6322 '<And at ast.py:11664>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11664
--# expr-start 6317 '<And at ast.py:11664>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11664
--# expr-start 6314 '<And at ast.py:11664>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11664
--# expr-start 6311 '<And at ast.py:11664>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11664
--# expr-start 6306 '<Bind at ast.py:11665>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11665
--# expr-start 6304 '<FieldAccess for expected_type_var at ast.py:11665>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11665
--# expr-start 6303 '<FieldAccess for left at ast.py:11665>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11665
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Bin_Op_F_Left;
--# expr-done 6303
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld);
--# expr-done 6304
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 6305 '<FieldAccess for bool_type at ast.py:11665>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11665
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_2 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 6305
      Cast_Result := Create_Internal_Entity
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Bind_Result := Solver.Create_Assign (Fld_1, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11665") else null)); 
--# expr-done 6306
--# expr-start 6310 '<Bind at ast.py:11666>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11666
--# expr-start 6308 '<FieldAccess for expected_type_var at ast.py:11666>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11666
--# expr-start 6307 '<FieldAccess for right at ast.py:11666>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11666
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Self.Bin_Op_F_Right;
--# expr-done 6307
      if Fld_3 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_3);
--# expr-done 6308
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
--# expr-start 6309 '<FieldAccess for bool_type at ast.py:11666>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11666
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_5 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 6309
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_4, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11666") else null)); 
--# expr-done 6310
And_Pred := Create_And (Bind_Result, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11664") else null)); 
--# expr-done 6311
--# expr-start 6313 '<FieldAccess for matches_expected_formal_prim_type at ast.py:11668>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11668
--# expr-start 6312 '<FieldAccess for left at ast.py:11668>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11668
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Self.Bin_Op_F_Left;
--# expr-done 6312
      if Fld_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_prim_type
Fld_7 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Prim_Type (Node => Fld_6, Origin => Origin);
--# end
--# expr-done 6313
And_Pred_1 := Create_And (And_Pred, Fld_7, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11664") else null)); 
--# expr-done 6314
--# expr-start 6316 '<FieldAccess for matches_expected_formal_prim_type at ast.py:11669>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11669
--# expr-start 6315 '<FieldAccess for right at ast.py:11669>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11669
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Self.Bin_Op_F_Right;
--# expr-done 6315
      if Fld_8 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_prim_type
Fld_9 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Prim_Type (Node => Fld_8, Origin => Origin);
--# end
--# expr-done 6316
And_Pred_2 := Create_And (And_Pred_1, Fld_9, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11664") else null)); 
--# expr-done 6317
--# expr-start 6321 '<Bind at ast.py:11671>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11671
--# expr-start 6318 '<FieldAccess for type_var at ast.py:11671>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11671
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6318
Fld_10.Value := No_Entity;
Entity_Vars.Reset (Fld_10);
--# expr-start 6320 '<FieldAccess for type_var at ast.py:11671>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11671
--# expr-start 6319 '<FieldAccess for left at ast.py:11671>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11671
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Self.Bin_Op_F_Left;
--# expr-done 6319
      if Fld_11 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_11);
--# expr-done 6320
Fld_12.Value := No_Entity;
Entity_Vars.Reset (Fld_12);
Bind_Result_2 := Solver.Create_Unify (Fld_10, Fld_12, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11671") else null)); 
--# expr-done 6321
And_Pred_3 := Create_And (And_Pred_2, Bind_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11664") else null)); 
--# expr-done 6322
Let_Result := And_Pred_3; Inc_Ref (Let_Result);
--# end
--# expr-done 6323
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_906;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_906;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Bin_Op_P_Test_Eq;
--# end
--# property-start BinOp.arguments_eq /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11675
pragma Warnings (Off, "is not referenced");
 function Bin_Op_P_Arguments_Eq
  (Node : Bare_Bin_Op
      ; Spec : Internal_Entity_Base_Subp_Spec
      ; Origin : Bare_Ada_Node
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Bin_Op := Bare_Bin_Op (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Bin_Op :=
     Internal_Entity_Bin_Op'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind spec Spec
   --# bind origin Origin
   --# bind env Env
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_907 with Inline_Always;
      Block_Var : Internal_Entity_Defining_Name_Array_Access;
Fld : Internal_Entity_Defining_Name_Array_Access;
Get_Result : Internal_Entity_Defining_Name;
Fld_1 : Internal_Entity_Base_Formal_Param_Decl;
Fld_2 : Internal_Entity_Expr;
Fld_3 : Logic_Equation;
Get_Result_1 : Internal_Entity_Defining_Name;
Fld_4 : Internal_Entity_Base_Formal_Param_Decl;
Fld_5 : Internal_Entity_Expr;
Fld_6 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_7 : Logic_Var;
Fld_8 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred_1 : Logic_Equation;
False_Rel : Logic_Equation;
Len : Integer;
Is_Equal : Boolean;
If_Result : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_907 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (False_Rel);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_907;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6341 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6324 '<FieldAccess for unpacked_formal_params at ast.py:11676>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11676
      if Spec.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => Spec.Node, E_Info => Spec.Info);
--# end
--# expr-done 6324
Block_Var := Fld; Inc_Ref (Block_Var);
--# bind ps Block_Var
--# expr-start 6340 '<If at ast.py:11677>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11677
--# expr-start 6326 '<Eq at ast.py:11678>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11678
--# expr-start 6325 '<length at ast.py:11678>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11678
Len := Length (Block_Var); 
--# expr-done 6325
Is_Equal := Len = 2; 
--# expr-done 6326
if Is_Equal then
   --# expr-start 6339 '<And at ast.py:11682>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11682
--# expr-start 6335 '<And at ast.py:11682>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11682
--# expr-start 6330 '<FieldAccess for call_argument_equation at ast.py:11682>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11682
      if Spec.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6328 '<FieldAccess for formal_decl at ast.py:11682>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11682
--# expr-start 6327 '<at at ast.py:11682>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11682
Get_Result := Get (Block_Var, 0, True); 
--# expr-done 6327
      if Get_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_1 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Get_Result.Node, E_Info => Get_Result.Info);
--# end
--# expr-done 6328
--# expr-start 6329 '<FieldAccess for left at ast.py:11682>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11682
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Expr (Node => Ent.Node.Bin_Op_F_Left, Info => Ent.Info);
--# expr-done 6329
--# property-call-start BaseFormalParamHolder.call_argument_equation
Fld_3 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Call_Argument_Equation (Node => Spec.Node, Param => Fld_1, Arg => Fld_2, Env => Env, Origin => Origin, E_Info => Spec.Info);
--# end
--# expr-done 6330
--# expr-start 6334 '<FieldAccess for call_argument_equation at ast.py:11686>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11686
      if Spec.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6332 '<FieldAccess for formal_decl at ast.py:11686>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11686
--# expr-start 6331 '<at at ast.py:11686>' Get_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11686
Get_Result_1 := Get (Block_Var, 1, True); 
--# expr-done 6331
      if Get_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_4 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Get_Result_1.Node, E_Info => Get_Result_1.Info);
--# end
--# expr-done 6332
--# expr-start 6333 '<FieldAccess for right at ast.py:11686>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11686
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Expr (Node => Ent.Node.Bin_Op_F_Right, Info => Ent.Info);
--# expr-done 6333
--# property-call-start BaseFormalParamHolder.call_argument_equation
Fld_6 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Call_Argument_Equation (Node => Spec.Node, Param => Fld_4, Arg => Fld_5, Env => Env, Origin => Origin, E_Info => Spec.Info);
--# end
--# expr-done 6334
And_Pred := Create_And (Fld_3, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11682") else null)); 
--# expr-done 6335
--# expr-start 6338 '<Bind at ast.py:11689>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11689
--# expr-start 6336 '<FieldAccess for type_var at ast.py:11689>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11689
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6336
Fld_7.Value := No_Entity;
Entity_Vars.Reset (Fld_7);
--# expr-start 6337 '<FieldAccess for return_type at ast.py:11689>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11689
      if Spec.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.return_type
Fld_8 := Libadalang.Implementation.Base_Subp_Spec_P_Return_Type (Node => Spec.Node, Origin => Origin, E_Info => Spec.Info);
--# end
--# expr-done 6337
      Cast_Result := Create_Internal_Entity
        (Node => Fld_8.Node,
         Info => Fld_8.Info);
Bind_Result := Solver.Create_Assign (Fld_7, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11689") else null)); 
--# expr-done 6338
And_Pred_1 := Create_And (And_Pred, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11682") else null)); 
--# expr-done 6339
   If_Result := And_Pred_1;
else
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11691") else null)); 
   If_Result := False_Rel;
end if;
      Inc_Ref (If_Result);
--# expr-done 6340
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 6341
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_907;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_907;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Bin_Op_P_Arguments_Eq;
--# end
--# property-start BinOp.entity_eq /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11695
pragma Warnings (Off, "is not referenced");
 function Bin_Op_P_Entity_Eq
  (Node : Bare_Bin_Op
      ; Subp : Internal_Entity_Basic_Decl
      ; Origin : Bare_Ada_Node
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Bin_Op := Bare_Bin_Op (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Bin_Op :=
     Internal_Entity_Bin_Op'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind subp Subp
   --# bind origin Origin
   --# bind env Env
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_908 with Inline_Always;
      Block_Var : Internal_Entity_Base_Subp_Spec;
Fld : Internal_Entity_Base_Subp_Spec;
Fld_1 : Logic_Equation;
Fld_2 : Bare_Op;
Fld_3 : Logic_Var;
Cast_Expr : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_4 : Bare_Op;
Fld_5 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Base_Subp_Spec;
Cast_Result_1 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_908 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_908;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6352 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6342 '<FieldAccess for subp_spec_or_null at ast.py:11696>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11696
      if Subp.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Subp.Node, E_Info => Subp.Info);
--# end
--# expr-done 6342
Block_Var := Fld; 
--# bind spec Block_Var
--# expr-start 6351 '<And at ast.py:11697>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11697
--# expr-start 6347 '<And at ast.py:11697>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11697
--# expr-start 6343 '<FieldAccess for arguments_eq at ast.py:11698>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11698
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BinOp.arguments_eq
Fld_1 := Libadalang.Implementation.Bin_Op_P_Arguments_Eq (Node => Ent.Node, Spec => Block_Var, Origin => Origin, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 6343
--# expr-start 6346 '<Bind at ast.py:11701>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11701
--# expr-start 6345 '<FieldAccess for ref_var at ast.py:11701>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11701
--# expr-start 6344 '<FieldAccess for op at ast.py:11701>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11701
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Bin_Op_F_Op;
--# expr-done 6344
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_2);
--# end
--# expr-done 6345
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
Cast_Expr := Subp; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_3, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11701") else null)); 
--# expr-done 6346
And_Pred := Create_And (Fld_1, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11697") else null)); 
--# expr-done 6347
--# expr-start 6350 '<Bind at ast.py:11702>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11702
--# expr-start 6349 '<FieldAccess for subp_spec_var at ast.py:11702>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11702
--# expr-start 6348 '<FieldAccess for op at ast.py:11702>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11702
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Self.Bin_Op_F_Op;
--# expr-done 6348
      if Fld_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.subp_spec_var'
Fld_5 := Libadalang.Implementation.Dispatcher_Name_P_Subp_Spec_Var (Node => Fld_4);
--# end
--# expr-done 6349
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
Cast_Expr_1 := Block_Var; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_5, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11702") else null)); 
--# expr-done 6350
And_Pred_1 := Create_And (And_Pred, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11697") else null)); 
--# expr-done 6351
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 6352
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_908;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_908;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Bin_Op_P_Entity_Eq;
--# end
--# property-start BinOp.overload_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11706
pragma Warnings (Off, "is not referenced");
 function Bin_Op_P_Overload_Equation
  (Node : Bare_Bin_Op
      ; Origin : Bare_Ada_Node
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Bin_Op := Bare_Bin_Op (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Bin_Op :=
     Internal_Entity_Bin_Op'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind env Env
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_909 with Inline_Always;
            procedure Finalizer_Scope_2638 with Inline_Always;
      Fld : Internal_Entity_Op;
Fld_1 : Internal_Entity_Basic_Decl_Array_Access;
Fld_2 : Logic_Equation;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_909 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_909;
            procedure Finalizer_Scope_2638 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_2638;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6358 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6357 '<LogicAny at ast.py:11707>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11707
--# expr-start 6356 '<Map at ast.py:11707>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11707
   --# expr-start 6354 '<FieldAccess for subprograms at ast.py:11707>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11707
--# expr-start 6353 '<FieldAccess for op at ast.py:11707>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11707
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Op (Node => Ent.Node.Bin_Op_F_Op, Info => Ent.Info);
--# expr-done 6353
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Op.subprograms
Fld_1 := Libadalang.Implementation.Op_P_Subprograms (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 6354
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Basic_Decl_Array_Access := Fld_1;
      begin
         for Item_223 of
               Collection.Items
         loop
   --# scope-start
               --# bind subp Item_223
      --# expr-start 6355 '<FieldAccess for entity_eq at ast.py:11708>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11708
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BinOp.entity_eq
Fld_2 := Libadalang.Implementation.Bin_Op_P_Entity_Eq (Node => Ent.Node, Subp => Item_223, Origin => Origin, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 6355
         declare
            Item_To_Append : constant Logic_Equation := Fld_2;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2638;
         end loop;
      end;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 6356
Logic_Boolean_Op := Solver.Create_Any (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11707") else null)); 
--# expr-done 6357
Let_Result := Logic_Boolean_Op; Inc_Ref (Let_Result);
--# end
--# expr-done 6358
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_909;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_909;
               Finalizer_Scope_2638;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Bin_Op_P_Overload_Equation;
--# end
--# property-start '[dispatcher]BinOp.no_overload_equation' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Bin_Op_P_No_Overload_Equation
  (Node : Bare_Bin_Op
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Bin_Op := Bare_Bin_Op (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Bin_Op :=
     Internal_Entity_Bin_Op'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Logic_Equation;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Bin_Op_Range (Self.Kind) is
               when Ada_Bin_Op =>
                  --# property-call-start BinOp.no_overload_equation
                  Property_Result := Bin_Op_P_No_Overload_Equation
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Relation_Op =>
                  --# property-call-start RelationOp.no_overload_equation
                  Property_Result := Relation_Op_P_No_Overload_Equation
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Bin_Op_P_No_Overload_Equation;
--# end
--# property-start BinOp.array_concat_expected_type_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11808
pragma Warnings (Off, "is not referenced");
 function Bin_Op_P_Array_Concat_Expected_Type_Equation
  (Node : Bare_Bin_Op
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Bin_Op := Bare_Bin_Op (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Bin_Op :=
     Internal_Entity_Bin_Op'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_911 with Inline_Always;
      Block_Var : Boolean;
Block_Var_1 : Boolean;
Fld : Bare_Expr;
Fld_1 : Boolean;
Fld_2 : Bare_Expr;
Fld_3 : Boolean;
If_Result : Boolean;
Fld_4 : Logic_Var;
Fld_5 : Bare_Expr;
Fld_6 : Logic_Var;
Fld_7 : Bare_Expr;
Fld_8 : Logic_Var;
Bind_Result : Logic_Equation;
Fld_9 : Bare_Expr;
Fld_10 : Logic_Var;
Fld_11 : Logic_Var;
Fld_12 : Bare_Expr;
Fld_13 : Logic_Var;
Bind_Result_1 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_14 : Bare_Expr;
Fld_15 : Logic_Var;
Fld_16 : Logic_Var;
Fld_17 : Bare_Expr;
Fld_18 : Logic_Var;
Bind_Result_2 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_19 : Logic_Var;
Fld_20 : Logic_Var;
Bind_Result_3 : Logic_Equation;
Fld_21 : Logic_Var;
Fld_22 : Bare_Expr;
Fld_23 : Logic_Var;
Fld_24 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_25 : Logic_Var;
Fld_26 : Bare_Expr;
Fld_27 : Logic_Var;
Fld_28 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_29 : Internal_Entity_Expr;
Fld_30 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Fld_31 : Internal_Entity_Expr;
Fld_32 : Logic_Equation;
And_Pred_5 : Logic_Equation;
Or_Pred : Logic_Equation;
Fld_33 : Bare_Expr;
Fld_34 : Logic_Var;
Fld_35 : Logic_Var;
Bind_Result_4 : Logic_Equation;
Fld_36 : Logic_Var;
Fld_37 : Bare_Expr;
Fld_38 : Logic_Var;
Bind_Result_5 : Logic_Equation;
And_Pred_6 : Logic_Equation;
Fld_39 : Logic_Var;
Fld_40 : Logic_Var;
Bind_Result_6 : Logic_Equation;
Fld_41 : Logic_Var;
Fld_42 : Bare_Expr;
Fld_43 : Logic_Var;
Fld_44 : Logic_Equation;
And_Pred_7 : Logic_Equation;
Fld_45 : Internal_Entity_Expr;
Fld_46 : Logic_Equation;
And_Pred_8 : Logic_Equation;
Or_Pred_1 : Logic_Equation;
Fld_47 : Logic_Var;
Fld_48 : Bare_Expr;
Fld_49 : Logic_Var;
Bind_Result_7 : Logic_Equation;
Fld_50 : Logic_Var;
Fld_51 : Bare_Expr;
Fld_52 : Logic_Var;
Fld_53 : Logic_Equation;
Or_Pred_2 : Logic_Equation;
And_Pred_9 : Logic_Equation;
Fld_54 : Internal_Entity_Expr;
Fld_55 : Logic_Equation;
And_Pred_10 : Logic_Equation;
Fld_56 : Bare_Expr;
Fld_57 : Logic_Var;
Fld_58 : Logic_Var;
Bind_Result_8 : Logic_Equation;
Fld_59 : Logic_Var;
Fld_60 : Bare_Expr;
Fld_61 : Logic_Var;
Bind_Result_9 : Logic_Equation;
And_Pred_11 : Logic_Equation;
Fld_62 : Logic_Var;
Fld_63 : Logic_Var;
Bind_Result_10 : Logic_Equation;
Fld_64 : Logic_Var;
Fld_65 : Bare_Expr;
Fld_66 : Logic_Var;
Fld_67 : Logic_Equation;
And_Pred_12 : Logic_Equation;
Fld_68 : Internal_Entity_Expr;
Fld_69 : Logic_Equation;
And_Pred_13 : Logic_Equation;
Or_Pred_3 : Logic_Equation;
Fld_70 : Logic_Var;
Fld_71 : Bare_Expr;
Fld_72 : Logic_Var;
Bind_Result_11 : Logic_Equation;
Fld_73 : Logic_Var;
Fld_74 : Bare_Expr;
Fld_75 : Logic_Var;
Fld_76 : Logic_Equation;
Or_Pred_4 : Logic_Equation;
And_Pred_14 : Logic_Equation;
Fld_77 : Internal_Entity_Expr;
Fld_78 : Logic_Equation;
And_Pred_15 : Logic_Equation;
Fld_79 : Logic_Var;
Fld_80 : Logic_Var;
Bind_Result_12 : Logic_Equation;
Fld_81 : Logic_Var;
Fld_82 : Bare_Expr;
Fld_83 : Logic_Var;
Bind_Result_13 : Logic_Equation;
Fld_84 : Logic_Var;
Fld_85 : Bare_Expr;
Fld_86 : Logic_Var;
Fld_87 : Logic_Equation;
Or_Pred_5 : Logic_Equation;
And_Pred_16 : Logic_Equation;
Fld_88 : Logic_Var;
Fld_89 : Bare_Expr;
Fld_90 : Logic_Var;
Bind_Result_14 : Logic_Equation;
Fld_91 : Logic_Var;
Fld_92 : Bare_Expr;
Fld_93 : Logic_Var;
Fld_94 : Logic_Equation;
Or_Pred_6 : Logic_Equation;
And_Pred_17 : Logic_Equation;
Fld_95 : Internal_Entity_Expr;
Fld_96 : Logic_Equation;
And_Pred_18 : Logic_Equation;
Fld_97 : Internal_Entity_Expr;
Fld_98 : Logic_Equation;
And_Pred_19 : Logic_Equation;
If_Result_1 : Logic_Equation;
If_Result_2 : Logic_Equation;
If_Result_3 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_911 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Bind_Result_3);
                     Dec_Ref (Fld_24);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Fld_28);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Fld_30);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Fld_32);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (Or_Pred);
                     Dec_Ref (Bind_Result_4);
                     Dec_Ref (Bind_Result_5);
                     Dec_Ref (And_Pred_6);
                     Dec_Ref (Bind_Result_6);
                     Dec_Ref (Fld_44);
                     Dec_Ref (And_Pred_7);
                     Dec_Ref (Fld_46);
                     Dec_Ref (And_Pred_8);
                     Dec_Ref (Or_Pred_1);
                     Dec_Ref (Bind_Result_7);
                     Dec_Ref (Fld_53);
                     Dec_Ref (Or_Pred_2);
                     Dec_Ref (And_Pred_9);
                     Dec_Ref (Fld_55);
                     Dec_Ref (And_Pred_10);
                     Dec_Ref (Bind_Result_8);
                     Dec_Ref (Bind_Result_9);
                     Dec_Ref (And_Pred_11);
                     Dec_Ref (Bind_Result_10);
                     Dec_Ref (Fld_67);
                     Dec_Ref (And_Pred_12);
                     Dec_Ref (Fld_69);
                     Dec_Ref (And_Pred_13);
                     Dec_Ref (Or_Pred_3);
                     Dec_Ref (Bind_Result_11);
                     Dec_Ref (Fld_76);
                     Dec_Ref (Or_Pred_4);
                     Dec_Ref (And_Pred_14);
                     Dec_Ref (Fld_78);
                     Dec_Ref (And_Pred_15);
                     Dec_Ref (Bind_Result_12);
                     Dec_Ref (Bind_Result_13);
                     Dec_Ref (Fld_87);
                     Dec_Ref (Or_Pred_5);
                     Dec_Ref (And_Pred_16);
                     Dec_Ref (Bind_Result_14);
                     Dec_Ref (Fld_94);
                     Dec_Ref (Or_Pred_6);
                     Dec_Ref (And_Pred_17);
                     Dec_Ref (Fld_96);
                     Dec_Ref (And_Pred_18);
                     Dec_Ref (Fld_98);
                     Dec_Ref (And_Pred_19);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (If_Result_2);
                     Dec_Ref (If_Result_3);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_911;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6501 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6360 '<FieldAccess for has_context_free_type at ast.py:11815>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11815
--# expr-start 6359 '<FieldAccess for left at ast.py:11815>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11815
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Bin_Op_F_Left;
--# expr-done 6359
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.has_context_free_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Expr_P_Has_Context_Free_Type (Node => Fld);
--# end
--# expr-done 6360
Block_Var := Fld_1; 
--# bind left_ctx_free Block_Var
--# expr-start 6362 '<FieldAccess for has_context_free_type at ast.py:11816>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11816
--# expr-start 6361 '<FieldAccess for right at ast.py:11816>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11816
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Bin_Op_F_Right;
--# expr-done 6361
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.has_context_free_type'
Fld_3 := Libadalang.Implementation.Dispatcher_Expr_P_Has_Context_Free_Type (Node => Fld_2);
--# end
--# expr-done 6362
Block_Var_1 := Fld_3; 
--# bind right_ctx_free Block_Var_1
--# expr-start 6500 '<Cond at ast.py:11821>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11821
if Block_Var then
   If_Result := Block_Var_1;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 6402 '<Or at ast.py:11824>' Or_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11824
--# expr-start 6382 '<And at ast.py:11831>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11831
--# expr-start 6375 '<And at ast.py:11831>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11831
--# expr-start 6368 '<NPropagate at ast.py:11831>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11831
--# expr-start 6363 '<FieldAccess for type_var at ast.py:11831>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11831
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6363
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
--# expr-start 6365 '<FieldAccess for type_var at ast.py:11833>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11833
--# expr-start 6364 '<FieldAccess for left at ast.py:11833>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11833
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Self.Bin_Op_F_Left;
--# expr-done 6364
      if Fld_5 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_5);
--# expr-done 6365
Fld_6.Value := No_Entity;
Entity_Vars.Reset (Fld_6);
--# expr-start 6367 '<FieldAccess for type_var at ast.py:11833>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11833
--# expr-start 6366 '<FieldAccess for right at ast.py:11833>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11833
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Self.Bin_Op_F_Right;
--# expr-done 6366
      if Fld_7 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_7);
--# expr-done 6367
Fld_8.Value := No_Entity;
Entity_Vars.Reset (Fld_8);
Bind_Result := Solver.Create_N_Propagate (Fld_4, Logic_Functor_581'(N => 2, Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Origin), (1 => Fld_6, 2 => Fld_8), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11831") else null)); 
--# expr-done 6368
--# expr-start 6374 '<NPropagate at ast.py:11834>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11834
--# expr-start 6370 '<FieldAccess for expected_type_var at ast.py:11834>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11834
--# expr-start 6369 '<FieldAccess for left at ast.py:11834>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11834
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Self.Bin_Op_F_Left;
--# expr-done 6369
      if Fld_9 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_9);
--# expr-done 6370
Fld_10.Value := No_Entity;
Entity_Vars.Reset (Fld_10);
--# expr-start 6371 '<FieldAccess for type_var at ast.py:11836>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11836
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6371
Fld_11.Value := No_Entity;
Entity_Vars.Reset (Fld_11);
--# expr-start 6373 '<FieldAccess for type_var at ast.py:11836>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11836
--# expr-start 6372 '<FieldAccess for left at ast.py:11836>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11836
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Self.Bin_Op_F_Left;
--# expr-done 6372
      if Fld_12 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_12);
--# expr-done 6373
Fld_13.Value := No_Entity;
Entity_Vars.Reset (Fld_13);
Bind_Result_1 := Solver.Create_N_Propagate (Fld_10, Logic_Functor_582'(N => 2, Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Origin), (1 => Fld_11, 2 => Fld_13), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11834") else null)); 
--# expr-done 6374
And_Pred := Create_And (Bind_Result, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11831") else null)); 
--# expr-done 6375
--# expr-start 6381 '<NPropagate at ast.py:11837>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11837
--# expr-start 6377 '<FieldAccess for expected_type_var at ast.py:11837>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11837
--# expr-start 6376 '<FieldAccess for right at ast.py:11837>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11837
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Self.Bin_Op_F_Right;
--# expr-done 6376
      if Fld_14 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_14);
--# expr-done 6377
Fld_15.Value := No_Entity;
Entity_Vars.Reset (Fld_15);
--# expr-start 6378 '<FieldAccess for type_var at ast.py:11839>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11839
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_16 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6378
Fld_16.Value := No_Entity;
Entity_Vars.Reset (Fld_16);
--# expr-start 6380 '<FieldAccess for type_var at ast.py:11839>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11839
--# expr-start 6379 '<FieldAccess for right at ast.py:11839>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11839
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_17 := Self.Bin_Op_F_Right;
--# expr-done 6379
      if Fld_17 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_17);
--# expr-done 6380
Fld_18.Value := No_Entity;
Entity_Vars.Reset (Fld_18);
Bind_Result_2 := Solver.Create_N_Propagate (Fld_15, Logic_Functor_582'(N => 2, Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Origin), (1 => Fld_16, 2 => Fld_18), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11837") else null)); 
--# expr-done 6381
And_Pred_1 := Create_And (And_Pred, Bind_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11831") else null)); 
--# expr-done 6382
--# expr-start 6401 '<And at ast.py:11846>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11846
--# expr-start 6398 '<And at ast.py:11846>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11846
--# expr-start 6395 '<And at ast.py:11846>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11846
--# expr-start 6390 '<And at ast.py:11846>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11846
--# expr-start 6385 '<Bind at ast.py:11846>' Bind_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11846
--# expr-start 6383 '<FieldAccess for expected_type_var at ast.py:11846>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11846
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6383
Fld_19.Value := No_Entity;
Entity_Vars.Reset (Fld_19);
--# expr-start 6384 '<FieldAccess for type_var at ast.py:11846>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11846
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_20 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6384
Fld_20.Value := No_Entity;
Entity_Vars.Reset (Fld_20);
Bind_Result_3 := Solver.Create_Propagate (Fld_19, Fld_20, Logic_Functor_578'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Origin), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11846") else null)); 
--# expr-done 6385
--# expr-start 6389 '<FieldAccess for comp_bind at ast.py:11848>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11848
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6386 '<FieldAccess for type_var at ast.py:11848>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11848
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_21 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6386
--# expr-start 6388 '<FieldAccess for expected_type_var at ast.py:11848>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11848
--# expr-start 6387 '<FieldAccess for left at ast.py:11848>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11848
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_22 := Self.Bin_Op_F_Left;
--# expr-done 6387
      if Fld_22 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_23 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_22);
--# expr-done 6388
--# property-call-start AdaNode.comp_bind
Fld_24 := Libadalang.Implementation.Ada_Node_P_Comp_Bind (Node => Self, Left => Fld_21, Right => Fld_23, Origin => Origin);
--# end
--# expr-done 6389
And_Pred_2 := Create_And (Bind_Result_3, Fld_24, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11846") else null)); 
--# expr-done 6390
--# expr-start 6394 '<FieldAccess for comp_bind at ast.py:11849>' Fld_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11849
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6391 '<FieldAccess for type_var at ast.py:11849>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11849
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_25 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6391
--# expr-start 6393 '<FieldAccess for expected_type_var at ast.py:11849>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11849
--# expr-start 6392 '<FieldAccess for right at ast.py:11849>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11849
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_26 := Self.Bin_Op_F_Right;
--# expr-done 6392
      if Fld_26 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_27 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_26);
--# expr-done 6393
--# property-call-start AdaNode.comp_bind
Fld_28 := Libadalang.Implementation.Ada_Node_P_Comp_Bind (Node => Self, Left => Fld_25, Right => Fld_27, Origin => Origin);
--# end
--# expr-done 6394
And_Pred_3 := Create_And (And_Pred_2, Fld_28, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11846") else null)); 
--# expr-done 6395
--# expr-start 6397 '<FieldAccess for matches_expected_formal_type at ast.py:11850>' Fld_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11850
--# expr-start 6396 '<FieldAccess for left at ast.py:11850>' Fld_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11850
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_29 := Create_Internal_Entity_Expr (Node => Ent.Node.Bin_Op_F_Left, Info => Ent.Info);
--# expr-done 6396
      if Fld_29.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_30 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_29.Node, Origin => Origin);
--# end
--# expr-done 6397
And_Pred_4 := Create_And (And_Pred_3, Fld_30, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11846") else null)); 
--# expr-done 6398
--# expr-start 6400 '<FieldAccess for matches_expected_formal_type at ast.py:11851>' Fld_32 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11851
--# expr-start 6399 '<FieldAccess for right at ast.py:11851>' Fld_31 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11851
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_31 := Create_Internal_Entity_Expr (Node => Ent.Node.Bin_Op_F_Right, Info => Ent.Info);
--# expr-done 6399
      if Fld_31.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_32 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_31.Node, Origin => Origin);
--# end
--# expr-done 6400
And_Pred_5 := Create_And (And_Pred_4, Fld_32, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11846") else null)); 
--# expr-done 6401
Or_Pred := Create_Or (And_Pred_1, And_Pred_5, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11824") else null)); 
--# expr-done 6402
   If_Result_3 := Or_Pred;
else
if Block_Var then
   --# expr-start 6436 '<And at ast.py:11855>' And_Pred_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11855
--# expr-start 6433 '<And at ast.py:11855>' And_Pred_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11855
--# expr-start 6423 '<Or at ast.py:11855>' Or_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11855
--# expr-start 6411 '<And at ast.py:11859>' And_Pred_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11859
--# expr-start 6406 '<Bind at ast.py:11859>' Bind_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11859
--# expr-start 6404 '<FieldAccess for type_var at ast.py:11859>' Fld_34 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11859
--# expr-start 6403 '<FieldAccess for left at ast.py:11859>' Fld_33 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11859
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_33 := Self.Bin_Op_F_Left;
--# expr-done 6403
      if Fld_33 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_34 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_33);
--# expr-done 6404
Fld_34.Value := No_Entity;
Entity_Vars.Reset (Fld_34);
--# expr-start 6405 '<FieldAccess for type_var at ast.py:11859>' Fld_35 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11859
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_35 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6405
Fld_35.Value := No_Entity;
Entity_Vars.Reset (Fld_35);
Bind_Result_4 := Solver.Create_Propagate (Fld_34, Fld_35, Logic_Functor_578'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Origin), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11859") else null)); 
--# expr-done 6406
--# expr-start 6410 '<Bind at ast.py:11861>' Bind_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11861
--# expr-start 6407 '<FieldAccess for type_var at ast.py:11861>' Fld_36 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11861
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_36 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6407
Fld_36.Value := No_Entity;
Entity_Vars.Reset (Fld_36);
--# expr-start 6409 '<FieldAccess for expected_type_var at ast.py:11861>' Fld_38 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11861
--# expr-start 6408 '<FieldAccess for left at ast.py:11861>' Fld_37 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11861
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_37 := Self.Bin_Op_F_Left;
--# expr-done 6408
      if Fld_37 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_38 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_37);
--# expr-done 6409
Fld_38.Value := No_Entity;
Entity_Vars.Reset (Fld_38);
Bind_Result_5 := Solver.Create_Unify (Fld_36, Fld_38, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11861") else null)); 
--# expr-done 6410
And_Pred_6 := Create_And (Bind_Result_4, Bind_Result_5, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11859") else null)); 
--# expr-done 6411
--# expr-start 6422 '<And at ast.py:11868>' And_Pred_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11868
--# expr-start 6419 '<And at ast.py:11868>' And_Pred_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11868
--# expr-start 6414 '<Bind at ast.py:11868>' Bind_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11868
--# expr-start 6412 '<FieldAccess for expected_type_var at ast.py:11868>' Fld_39 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11868
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_39 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6412
Fld_39.Value := No_Entity;
Entity_Vars.Reset (Fld_39);
--# expr-start 6413 '<FieldAccess for type_var at ast.py:11868>' Fld_40 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11868
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_40 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6413
Fld_40.Value := No_Entity;
Entity_Vars.Reset (Fld_40);
Bind_Result_6 := Solver.Create_Propagate (Fld_39, Fld_40, Logic_Functor_578'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Origin), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11868") else null)); 
--# expr-done 6414
--# expr-start 6418 '<FieldAccess for comp_bind at ast.py:11870>' Fld_44 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11870
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6415 '<FieldAccess for type_var at ast.py:11870>' Fld_41 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11870
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_41 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6415
--# expr-start 6417 '<FieldAccess for expected_type_var at ast.py:11870>' Fld_43 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11870
--# expr-start 6416 '<FieldAccess for left at ast.py:11870>' Fld_42 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11870
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_42 := Self.Bin_Op_F_Left;
--# expr-done 6416
      if Fld_42 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_43 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_42);
--# expr-done 6417
--# property-call-start AdaNode.comp_bind
Fld_44 := Libadalang.Implementation.Ada_Node_P_Comp_Bind (Node => Self, Left => Fld_41, Right => Fld_43, Origin => Origin);
--# end
--# expr-done 6418
And_Pred_7 := Create_And (Bind_Result_6, Fld_44, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11868") else null)); 
--# expr-done 6419
--# expr-start 6421 '<FieldAccess for matches_expected_formal_type at ast.py:11871>' Fld_46 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11871
--# expr-start 6420 '<FieldAccess for left at ast.py:11871>' Fld_45 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11871
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_45 := Create_Internal_Entity_Expr (Node => Ent.Node.Bin_Op_F_Left, Info => Ent.Info);
--# expr-done 6420
      if Fld_45.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_46 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_45.Node, Origin => Origin);
--# end
--# expr-done 6421
And_Pred_8 := Create_And (And_Pred_7, Fld_46, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11868") else null)); 
--# expr-done 6422
Or_Pred_1 := Create_Or (And_Pred_6, And_Pred_8, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11855") else null)); 
--# expr-done 6423
--# expr-start 6432 '<Or at ast.py:11875>' Or_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11875
--# expr-start 6427 '<Bind at ast.py:11876>' Bind_Result_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11876
--# expr-start 6424 '<FieldAccess for type_var at ast.py:11876>' Fld_47 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11876
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_47 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6424
Fld_47.Value := No_Entity;
Entity_Vars.Reset (Fld_47);
--# expr-start 6426 '<FieldAccess for expected_type_var at ast.py:11876>' Fld_49 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11876
--# expr-start 6425 '<FieldAccess for right at ast.py:11876>' Fld_48 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11876
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_48 := Self.Bin_Op_F_Right;
--# expr-done 6425
      if Fld_48 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_49 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_48);
--# expr-done 6426
Fld_49.Value := No_Entity;
Entity_Vars.Reset (Fld_49);
Bind_Result_7 := Solver.Create_Unify (Fld_47, Fld_49, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11876") else null)); 
--# expr-done 6427
--# expr-start 6431 '<FieldAccess for comp_bind at ast.py:11877>' Fld_53 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11877
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6428 '<FieldAccess for type_var at ast.py:11877>' Fld_50 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11877
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_50 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6428
--# expr-start 6430 '<FieldAccess for expected_type_var at ast.py:11877>' Fld_52 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11877
--# expr-start 6429 '<FieldAccess for right at ast.py:11877>' Fld_51 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11877
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_51 := Self.Bin_Op_F_Right;
--# expr-done 6429
      if Fld_51 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_52 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_51);
--# expr-done 6430
--# property-call-start AdaNode.comp_bind
Fld_53 := Libadalang.Implementation.Ada_Node_P_Comp_Bind (Node => Self, Left => Fld_50, Right => Fld_52, Origin => Origin);
--# end
--# expr-done 6431
Or_Pred_2 := Create_Or (Bind_Result_7, Fld_53, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11875") else null)); 
--# expr-done 6432
And_Pred_9 := Create_And (Or_Pred_1, Or_Pred_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11855") else null)); 
--# expr-done 6433
--# expr-start 6435 '<FieldAccess for matches_expected_formal_type at ast.py:11879>' Fld_55 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11879
--# expr-start 6434 '<FieldAccess for right at ast.py:11879>' Fld_54 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11879
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_54 := Create_Internal_Entity_Expr (Node => Ent.Node.Bin_Op_F_Right, Info => Ent.Info);
--# expr-done 6434
      if Fld_54.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_55 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_54.Node, Origin => Origin);
--# end
--# expr-done 6435
And_Pred_10 := Create_And (And_Pred_9, Fld_55, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11855") else null)); 
--# expr-done 6436
   If_Result_2 := And_Pred_10;
else
if Block_Var_1 then
   --# expr-start 6470 '<And at ast.py:11882>' And_Pred_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11882
--# expr-start 6467 '<And at ast.py:11882>' And_Pred_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11882
--# expr-start 6457 '<Or at ast.py:11882>' Or_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11882
--# expr-start 6445 '<And at ast.py:11886>' And_Pred_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11886
--# expr-start 6440 '<Bind at ast.py:11886>' Bind_Result_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11886
--# expr-start 6438 '<FieldAccess for type_var at ast.py:11886>' Fld_57 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11886
--# expr-start 6437 '<FieldAccess for right at ast.py:11886>' Fld_56 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11886
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_56 := Self.Bin_Op_F_Right;
--# expr-done 6437
      if Fld_56 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_57 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_56);
--# expr-done 6438
Fld_57.Value := No_Entity;
Entity_Vars.Reset (Fld_57);
--# expr-start 6439 '<FieldAccess for type_var at ast.py:11886>' Fld_58 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11886
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_58 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6439
Fld_58.Value := No_Entity;
Entity_Vars.Reset (Fld_58);
Bind_Result_8 := Solver.Create_Propagate (Fld_57, Fld_58, Logic_Functor_578'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Origin), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11886") else null)); 
--# expr-done 6440
--# expr-start 6444 '<Bind at ast.py:11888>' Bind_Result_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11888
--# expr-start 6441 '<FieldAccess for type_var at ast.py:11888>' Fld_59 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11888
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_59 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6441
Fld_59.Value := No_Entity;
Entity_Vars.Reset (Fld_59);
--# expr-start 6443 '<FieldAccess for expected_type_var at ast.py:11888>' Fld_61 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11888
--# expr-start 6442 '<FieldAccess for right at ast.py:11888>' Fld_60 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11888
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_60 := Self.Bin_Op_F_Right;
--# expr-done 6442
      if Fld_60 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_61 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_60);
--# expr-done 6443
Fld_61.Value := No_Entity;
Entity_Vars.Reset (Fld_61);
Bind_Result_9 := Solver.Create_Unify (Fld_59, Fld_61, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11888") else null)); 
--# expr-done 6444
And_Pred_11 := Create_And (Bind_Result_8, Bind_Result_9, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11886") else null)); 
--# expr-done 6445
--# expr-start 6456 '<And at ast.py:11895>' And_Pred_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11895
--# expr-start 6453 '<And at ast.py:11895>' And_Pred_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11895
--# expr-start 6448 '<Bind at ast.py:11895>' Bind_Result_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11895
--# expr-start 6446 '<FieldAccess for expected_type_var at ast.py:11895>' Fld_62 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11895
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_62 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6446
Fld_62.Value := No_Entity;
Entity_Vars.Reset (Fld_62);
--# expr-start 6447 '<FieldAccess for type_var at ast.py:11895>' Fld_63 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11895
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_63 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6447
Fld_63.Value := No_Entity;
Entity_Vars.Reset (Fld_63);
Bind_Result_10 := Solver.Create_Propagate (Fld_62, Fld_63, Logic_Functor_578'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Origin), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11895") else null)); 
--# expr-done 6448
--# expr-start 6452 '<FieldAccess for comp_bind at ast.py:11897>' Fld_67 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11897
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6449 '<FieldAccess for type_var at ast.py:11897>' Fld_64 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11897
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_64 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6449
--# expr-start 6451 '<FieldAccess for expected_type_var at ast.py:11897>' Fld_66 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11897
--# expr-start 6450 '<FieldAccess for right at ast.py:11897>' Fld_65 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11897
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_65 := Self.Bin_Op_F_Right;
--# expr-done 6450
      if Fld_65 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_66 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_65);
--# expr-done 6451
--# property-call-start AdaNode.comp_bind
Fld_67 := Libadalang.Implementation.Ada_Node_P_Comp_Bind (Node => Self, Left => Fld_64, Right => Fld_66, Origin => Origin);
--# end
--# expr-done 6452
And_Pred_12 := Create_And (Bind_Result_10, Fld_67, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11895") else null)); 
--# expr-done 6453
--# expr-start 6455 '<FieldAccess for matches_expected_formal_type at ast.py:11898>' Fld_69 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11898
--# expr-start 6454 '<FieldAccess for right at ast.py:11898>' Fld_68 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11898
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_68 := Create_Internal_Entity_Expr (Node => Ent.Node.Bin_Op_F_Right, Info => Ent.Info);
--# expr-done 6454
      if Fld_68.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_69 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_68.Node, Origin => Origin);
--# end
--# expr-done 6455
And_Pred_13 := Create_And (And_Pred_12, Fld_69, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11895") else null)); 
--# expr-done 6456
Or_Pred_3 := Create_Or (And_Pred_11, And_Pred_13, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11882") else null)); 
--# expr-done 6457
--# expr-start 6466 '<Or at ast.py:11902>' Or_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11902
--# expr-start 6461 '<Bind at ast.py:11903>' Bind_Result_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11903
--# expr-start 6458 '<FieldAccess for type_var at ast.py:11903>' Fld_70 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11903
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_70 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6458
Fld_70.Value := No_Entity;
Entity_Vars.Reset (Fld_70);
--# expr-start 6460 '<FieldAccess for expected_type_var at ast.py:11903>' Fld_72 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11903
--# expr-start 6459 '<FieldAccess for left at ast.py:11903>' Fld_71 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11903
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_71 := Self.Bin_Op_F_Left;
--# expr-done 6459
      if Fld_71 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_72 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_71);
--# expr-done 6460
Fld_72.Value := No_Entity;
Entity_Vars.Reset (Fld_72);
Bind_Result_11 := Solver.Create_Unify (Fld_70, Fld_72, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11903") else null)); 
--# expr-done 6461
--# expr-start 6465 '<FieldAccess for comp_bind at ast.py:11904>' Fld_76 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11904
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6462 '<FieldAccess for type_var at ast.py:11904>' Fld_73 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11904
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_73 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6462
--# expr-start 6464 '<FieldAccess for expected_type_var at ast.py:11904>' Fld_75 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11904
--# expr-start 6463 '<FieldAccess for left at ast.py:11904>' Fld_74 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11904
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_74 := Self.Bin_Op_F_Left;
--# expr-done 6463
      if Fld_74 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_75 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_74);
--# expr-done 6464
--# property-call-start AdaNode.comp_bind
Fld_76 := Libadalang.Implementation.Ada_Node_P_Comp_Bind (Node => Self, Left => Fld_73, Right => Fld_75, Origin => Origin);
--# end
--# expr-done 6465
Or_Pred_4 := Create_Or (Bind_Result_11, Fld_76, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11902") else null)); 
--# expr-done 6466
And_Pred_14 := Create_And (Or_Pred_3, Or_Pred_4, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11882") else null)); 
--# expr-done 6467
--# expr-start 6469 '<FieldAccess for matches_expected_formal_type at ast.py:11906>' Fld_78 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11906
--# expr-start 6468 '<FieldAccess for left at ast.py:11906>' Fld_77 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11906
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_77 := Create_Internal_Entity_Expr (Node => Ent.Node.Bin_Op_F_Left, Info => Ent.Info);
--# expr-done 6468
      if Fld_77.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_78 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_77.Node, Origin => Origin);
--# end
--# expr-done 6469
And_Pred_15 := Create_And (And_Pred_14, Fld_78, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11882") else null)); 
--# expr-done 6470
   If_Result_1 := And_Pred_15;
else
   --# expr-start 6499 '<And at ast.py:11912>' And_Pred_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11912
--# expr-start 6496 '<And at ast.py:11912>' And_Pred_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11912
--# expr-start 6493 '<And at ast.py:11912>' And_Pred_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11912
--# expr-start 6483 '<And at ast.py:11912>' And_Pred_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11912
--# expr-start 6473 '<Bind at ast.py:11912>' Bind_Result_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11912
--# expr-start 6471 '<FieldAccess for expected_type_var at ast.py:11912>' Fld_79 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11912
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_79 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6471
Fld_79.Value := No_Entity;
Entity_Vars.Reset (Fld_79);
--# expr-start 6472 '<FieldAccess for type_var at ast.py:11912>' Fld_80 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11912
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_80 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6472
Fld_80.Value := No_Entity;
Entity_Vars.Reset (Fld_80);
Bind_Result_12 := Solver.Create_Propagate (Fld_79, Fld_80, Logic_Functor_578'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Origin), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11912") else null)); 
--# expr-done 6473
--# expr-start 6482 '<Or at ast.py:11914>' Or_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11914
--# expr-start 6477 '<Bind at ast.py:11915>' Bind_Result_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11915
--# expr-start 6474 '<FieldAccess for type_var at ast.py:11915>' Fld_81 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11915
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_81 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6474
Fld_81.Value := No_Entity;
Entity_Vars.Reset (Fld_81);
--# expr-start 6476 '<FieldAccess for expected_type_var at ast.py:11915>' Fld_83 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11915
--# expr-start 6475 '<FieldAccess for left at ast.py:11915>' Fld_82 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11915
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_82 := Self.Bin_Op_F_Left;
--# expr-done 6475
      if Fld_82 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_83 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_82);
--# expr-done 6476
Fld_83.Value := No_Entity;
Entity_Vars.Reset (Fld_83);
Bind_Result_13 := Solver.Create_Unify (Fld_81, Fld_83, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11915") else null)); 
--# expr-done 6477
--# expr-start 6481 '<FieldAccess for comp_bind at ast.py:11916>' Fld_87 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11916
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6478 '<FieldAccess for type_var at ast.py:11916>' Fld_84 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11916
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_84 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6478
--# expr-start 6480 '<FieldAccess for expected_type_var at ast.py:11916>' Fld_86 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11916
--# expr-start 6479 '<FieldAccess for left at ast.py:11916>' Fld_85 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11916
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_85 := Self.Bin_Op_F_Left;
--# expr-done 6479
      if Fld_85 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_86 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_85);
--# expr-done 6480
--# property-call-start AdaNode.comp_bind
Fld_87 := Libadalang.Implementation.Ada_Node_P_Comp_Bind (Node => Self, Left => Fld_84, Right => Fld_86, Origin => Origin);
--# end
--# expr-done 6481
Or_Pred_5 := Create_Or (Bind_Result_13, Fld_87, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11914") else null)); 
--# expr-done 6482
And_Pred_16 := Create_And (Bind_Result_12, Or_Pred_5, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11912") else null)); 
--# expr-done 6483
--# expr-start 6492 '<Or at ast.py:11918>' Or_Pred_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11918
--# expr-start 6487 '<Bind at ast.py:11919>' Bind_Result_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11919
--# expr-start 6484 '<FieldAccess for type_var at ast.py:11919>' Fld_88 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11919
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_88 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6484
Fld_88.Value := No_Entity;
Entity_Vars.Reset (Fld_88);
--# expr-start 6486 '<FieldAccess for expected_type_var at ast.py:11919>' Fld_90 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11919
--# expr-start 6485 '<FieldAccess for right at ast.py:11919>' Fld_89 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11919
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_89 := Self.Bin_Op_F_Right;
--# expr-done 6485
      if Fld_89 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_90 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_89);
--# expr-done 6486
Fld_90.Value := No_Entity;
Entity_Vars.Reset (Fld_90);
Bind_Result_14 := Solver.Create_Unify (Fld_88, Fld_90, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11919") else null)); 
--# expr-done 6487
--# expr-start 6491 '<FieldAccess for comp_bind at ast.py:11920>' Fld_94 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11920
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6488 '<FieldAccess for type_var at ast.py:11920>' Fld_91 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11920
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_91 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6488
--# expr-start 6490 '<FieldAccess for expected_type_var at ast.py:11920>' Fld_93 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11920
--# expr-start 6489 '<FieldAccess for right at ast.py:11920>' Fld_92 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11920
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_92 := Self.Bin_Op_F_Right;
--# expr-done 6489
      if Fld_92 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_93 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_92);
--# expr-done 6490
--# property-call-start AdaNode.comp_bind
Fld_94 := Libadalang.Implementation.Ada_Node_P_Comp_Bind (Node => Self, Left => Fld_91, Right => Fld_93, Origin => Origin);
--# end
--# expr-done 6491
Or_Pred_6 := Create_Or (Bind_Result_14, Fld_94, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11918") else null)); 
--# expr-done 6492
And_Pred_17 := Create_And (And_Pred_16, Or_Pred_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11912") else null)); 
--# expr-done 6493
--# expr-start 6495 '<FieldAccess for matches_expected_formal_type at ast.py:11922>' Fld_96 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11922
--# expr-start 6494 '<FieldAccess for left at ast.py:11922>' Fld_95 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11922
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_95 := Create_Internal_Entity_Expr (Node => Ent.Node.Bin_Op_F_Left, Info => Ent.Info);
--# expr-done 6494
      if Fld_95.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_96 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_95.Node, Origin => Origin);
--# end
--# expr-done 6495
And_Pred_18 := Create_And (And_Pred_17, Fld_96, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11912") else null)); 
--# expr-done 6496
--# expr-start 6498 '<FieldAccess for matches_expected_formal_type at ast.py:11923>' Fld_98 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11923
--# expr-start 6497 '<FieldAccess for right at ast.py:11923>' Fld_97 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11923
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_97 := Create_Internal_Entity_Expr (Node => Ent.Node.Bin_Op_F_Right, Info => Ent.Info);
--# expr-done 6497
      if Fld_97.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_98 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_97.Node, Origin => Origin);
--# end
--# expr-done 6498
And_Pred_19 := Create_And (And_Pred_18, Fld_98, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11912") else null)); 
--# expr-done 6499
   If_Result_1 := And_Pred_19;
end if;
      Inc_Ref (If_Result_1);
   If_Result_2 := If_Result_1;
end if;
      Inc_Ref (If_Result_2);
   If_Result_3 := If_Result_2;
end if;
      Inc_Ref (If_Result_3);
--# expr-done 6500
Let_Result := If_Result_3; Inc_Ref (Let_Result);
--# end
--# expr-done 6501
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_911;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_911;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Bin_Op_P_Array_Concat_Expected_Type_Equation;
--# end
--# property-start BinOp.numeric_type_from_operands_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11927
pragma Warnings (Off, "is not referenced");
 function Bin_Op_P_Numeric_Type_From_Operands_Equation
  (Node : Bare_Bin_Op
      ; Dest_Var : Logic_Var
      ; Origin : Bare_Ada_Node
  )
   return Logic_Equation
is
   Self : Bare_Bin_Op := Bare_Bin_Op (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind dest_var Dest_Var
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_912 with Inline_Always;
      Block_Var : Boolean;
Block_Var_1 : Boolean;
Infer_Left : Logic_Equation;
Infer_Right : Logic_Equation;
Fld : Bare_Expr;
Fld_1 : Boolean;
Fld_2 : Bare_Expr;
Fld_3 : Boolean;
Fld_4 : Bare_Expr;
Fld_5 : Logic_Var;
Fld_6 : Bare_Expr;
Fld_7 : Logic_Var;
Pred : Logic_Equation;
Fld_8 : Bare_Expr;
Fld_9 : Logic_Var;
Fld_10 : Bare_Expr;
Fld_11 : Logic_Var;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_12 : Bare_Expr;
Fld_13 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_14 : Bare_Expr;
Fld_15 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_16 : Bare_Expr;
Fld_17 : Logic_Var;
Pred_1 : Logic_Equation;
Fld_18 : Bare_Expr;
Fld_19 : Logic_Var;
Bind_Result_1 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_20 : Bare_Expr;
Fld_21 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Fld_22 : Bare_Expr;
Fld_23 : Logic_Var;
Pred_2 : Logic_Equation;
Fld_24 : Bare_Expr;
Fld_25 : Logic_Var;
Bind_Result_2 : Logic_Equation;
And_Pred_5 : Logic_Equation;
Fld_26 : Bare_Expr;
Fld_27 : Logic_Equation;
And_Pred_6 : Logic_Equation;
Or_Pred : Logic_Equation;
Or_Pred_1 : Logic_Equation;
If_Result : Logic_Equation;
Let_Result : Logic_Equation;
If_Result_1 : Boolean;
If_Result_2 : Logic_Equation;
Let_Result_1 : Logic_Equation;
            procedure Finalizer_Scope_912 is
            begin
                     Dec_Ref (Pred);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_13);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_15);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Pred_1);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Fld_21);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Infer_Left);
                     Dec_Ref (Pred_2);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (Fld_27);
                     Dec_Ref (And_Pred_6);
                     Dec_Ref (Infer_Right);
                     Dec_Ref (Or_Pred);
                     Dec_Ref (Or_Pred_1);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
                     Dec_Ref (If_Result_2);
                     Dec_Ref (Let_Result_1);
            end Finalizer_Scope_912;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6548 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 6503 '<FieldAccess for has_context_free_type at ast.py:11933>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11933
--# expr-start 6502 '<FieldAccess for left at ast.py:11933>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11933
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Bin_Op_F_Left;
--# expr-done 6502
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.has_context_free_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Expr_P_Has_Context_Free_Type (Node => Fld);
--# end
--# expr-done 6503
Block_Var := Fld_1; 
--# bind left_ctx_free Block_Var
--# expr-start 6505 '<FieldAccess for has_context_free_type at ast.py:11934>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11934
--# expr-start 6504 '<FieldAccess for right at ast.py:11934>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11934
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Bin_Op_F_Right;
--# expr-done 6504
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.has_context_free_type'
Fld_3 := Libadalang.Implementation.Dispatcher_Expr_P_Has_Context_Free_Type (Node => Fld_2);
--# end
--# expr-done 6505
Block_Var_1 := Fld_3; 
--# bind right_ctx_free Block_Var_1
--# expr-start 6547 '<If at ast.py:11935>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11935
if Block_Var then
   If_Result_1 := Block_Var_1;
else
   If_Result_1 := False;
end if;
if If_Result_1 then
   --# expr-start 6522 '<And at ast.py:11942>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11942
--# expr-start 6519 '<And at ast.py:11942>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11942
--# expr-start 6516 '<And at ast.py:11942>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11942
--# expr-start 6510 '<Predicate on BaseTypeDecl.one_non_universal at ast.py:11942>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11942
--# expr-start 6507 '<FieldAccess for type_var at ast.py:11943>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11943
--# expr-start 6506 '<FieldAccess for left at ast.py:11943>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11943
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Self.Bin_Op_F_Left;
--# expr-done 6506
      if Fld_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_4);
--# expr-done 6507
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
--# expr-start 6509 '<FieldAccess for type_var at ast.py:11943>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11943
--# expr-start 6508 '<FieldAccess for right at ast.py:11943>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11943
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Self.Bin_Op_F_Right;
--# expr-done 6508
      if Fld_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_6);
--# expr-done 6509
Fld_7.Value := No_Entity;
Entity_Vars.Reset (Fld_7);
Pred := Solver.Create_N_Predicate ((Fld_5, Fld_7), Create_Base_Type_Decl_P_One_Non_Universal_0_Predicate); 
--# expr-done 6510
--# expr-start 6515 '<NPropagate at ast.py:11944>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11944
Dest_Var.Value := No_Entity;
Entity_Vars.Reset (Dest_Var);
--# expr-start 6512 '<FieldAccess for type_var at ast.py:11946>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11946
--# expr-start 6511 '<FieldAccess for left at ast.py:11946>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11946
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Self.Bin_Op_F_Left;
--# expr-done 6511
      if Fld_8 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_8);
--# expr-done 6512
Fld_9.Value := No_Entity;
Entity_Vars.Reset (Fld_9);
--# expr-start 6514 '<FieldAccess for type_var at ast.py:11946>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11946
--# expr-start 6513 '<FieldAccess for right at ast.py:11946>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11946
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Self.Bin_Op_F_Right;
--# expr-done 6513
      if Fld_10 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_10);
--# expr-done 6514
Fld_11.Value := No_Entity;
Entity_Vars.Reset (Fld_11);
Bind_Result := Solver.Create_N_Propagate (Dest_Var, Logic_Functor_580'(N => 2, Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Origin), (1 => Fld_9, 2 => Fld_11), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11944") else null)); 
--# expr-done 6515
And_Pred := Create_And (Pred, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11942") else null)); 
--# expr-done 6516
--# expr-start 6518 '<FieldAccess for matches_expected_formal_type at ast.py:11947>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11947
--# expr-start 6517 '<FieldAccess for left at ast.py:11947>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11947
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Self.Bin_Op_F_Left;
--# expr-done 6517
      if Fld_12 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_13 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_12, Origin => Origin);
--# end
--# expr-done 6518
And_Pred_1 := Create_And (And_Pred, Fld_13, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11942") else null)); 
--# expr-done 6519
--# expr-start 6521 '<FieldAccess for matches_expected_formal_type at ast.py:11948>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11948
--# expr-start 6520 '<FieldAccess for right at ast.py:11948>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11948
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Self.Bin_Op_F_Right;
--# expr-done 6520
      if Fld_14 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_15 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_14, Origin => Origin);
--# end
--# expr-done 6521
And_Pred_2 := Create_And (And_Pred_1, Fld_15, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11942") else null)); 
--# expr-done 6522
   If_Result_2 := And_Pred_2;
else
   --# expr-start 6546 '<Let at ast.py:11950>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11950
--# scope-start
--# expr-start 6532 '<And at ast.py:11952>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11952
--# expr-start 6529 '<And at ast.py:11952>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11952
--# expr-start 6525 '<Predicate on BaseTypeDecl.is_not_universal_type at ast.py:11952>' Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11952
--# expr-start 6524 '<FieldAccess for type_var at ast.py:11953>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11953
--# expr-start 6523 '<FieldAccess for left at ast.py:11953>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11953
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_16 := Self.Bin_Op_F_Left;
--# expr-done 6523
      if Fld_16 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_17 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_16);
--# expr-done 6524
Fld_17.Value := No_Entity;
Entity_Vars.Reset (Fld_17);
Pred_1 := Solver.Create_Predicate (Fld_17, Create_Base_Type_Decl_P_Is_Not_Universal_Type_0_Predicate); 
--# expr-done 6525
--# expr-start 6528 '<Bind at ast.py:11954>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11954
--# expr-start 6527 '<FieldAccess for type_var at ast.py:11954>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11954
--# expr-start 6526 '<FieldAccess for left at ast.py:11954>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11954
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Self.Bin_Op_F_Left;
--# expr-done 6526
      if Fld_18 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_18);
--# expr-done 6527
Fld_19.Value := No_Entity;
Entity_Vars.Reset (Fld_19);
Dest_Var.Value := No_Entity;
Entity_Vars.Reset (Dest_Var);
Bind_Result_1 := Solver.Create_Propagate (Fld_19, Dest_Var, Logic_Functor_578'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Origin), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11954") else null)); 
--# expr-done 6528
And_Pred_3 := Create_And (Pred_1, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11952") else null)); 
--# expr-done 6529
--# expr-start 6531 '<FieldAccess for matches_expected_formal_type at ast.py:11956>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11956
--# expr-start 6530 '<FieldAccess for left at ast.py:11956>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11956
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_20 := Self.Bin_Op_F_Left;
--# expr-done 6530
      if Fld_20 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_21 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_20, Origin => Origin);
--# end
--# expr-done 6531
And_Pred_4 := Create_And (And_Pred_3, Fld_21, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11952") else null)); 
--# expr-done 6532
Infer_Left := And_Pred_4; Inc_Ref (Infer_Left);
--# bind infer_left Infer_Left
--# expr-start 6542 '<And at ast.py:11958>' And_Pred_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11958
--# expr-start 6539 '<And at ast.py:11958>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11958
--# expr-start 6535 '<Predicate on BaseTypeDecl.is_not_universal_type at ast.py:11958>' Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11958
--# expr-start 6534 '<FieldAccess for type_var at ast.py:11959>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11959
--# expr-start 6533 '<FieldAccess for right at ast.py:11959>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11959
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_22 := Self.Bin_Op_F_Right;
--# expr-done 6533
      if Fld_22 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_23 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_22);
--# expr-done 6534
Fld_23.Value := No_Entity;
Entity_Vars.Reset (Fld_23);
Pred_2 := Solver.Create_Predicate (Fld_23, Create_Base_Type_Decl_P_Is_Not_Universal_Type_0_Predicate); 
--# expr-done 6535
--# expr-start 6538 '<Bind at ast.py:11960>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11960
--# expr-start 6537 '<FieldAccess for type_var at ast.py:11960>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11960
--# expr-start 6536 '<FieldAccess for right at ast.py:11960>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11960
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_24 := Self.Bin_Op_F_Right;
--# expr-done 6536
      if Fld_24 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_25 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_24);
--# expr-done 6537
Fld_25.Value := No_Entity;
Entity_Vars.Reset (Fld_25);
Dest_Var.Value := No_Entity;
Entity_Vars.Reset (Dest_Var);
Bind_Result_2 := Solver.Create_Propagate (Fld_25, Dest_Var, Logic_Functor_578'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Origin), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11960") else null)); 
--# expr-done 6538
And_Pred_5 := Create_And (Pred_2, Bind_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11958") else null)); 
--# expr-done 6539
--# expr-start 6541 '<FieldAccess for matches_expected_formal_type at ast.py:11962>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11962
--# expr-start 6540 '<FieldAccess for right at ast.py:11962>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11962
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_26 := Self.Bin_Op_F_Right;
--# expr-done 6540
      if Fld_26 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_27 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_26, Origin => Origin);
--# end
--# expr-done 6541
And_Pred_6 := Create_And (And_Pred_5, Fld_27, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11958") else null)); 
--# expr-done 6542
Infer_Right := And_Pred_6; Inc_Ref (Infer_Right);
--# bind infer_right Infer_Right
--# expr-start 6545 '<If at ast.py:11964>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11964
if Block_Var then
   --# expr-start 6543 '<Or at ast.py:11965>' Or_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11965
Or_Pred := Create_Or (Infer_Left, Infer_Right, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11965") else null)); 
--# expr-done 6543
   If_Result := Or_Pred;
else
   --# expr-start 6544 '<Or at ast.py:11966>' Or_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11966
Or_Pred_1 := Create_Or (Infer_Right, Infer_Left, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11966") else null)); 
--# expr-done 6544
   If_Result := Or_Pred_1;
end if;
      Inc_Ref (If_Result);
--# expr-done 6545
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 6546
   If_Result_2 := Let_Result;
end if;
      Inc_Ref (If_Result_2);
--# expr-done 6547
Let_Result_1 := If_Result_2; Inc_Ref (Let_Result_1);
--# end
--# expr-done 6548
      Property_Result := Let_Result_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_912;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_912;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Bin_Op_P_Numeric_Type_From_Operands_Equation;
--# end
--# property-start BinOp.potential_actuals_for_dispatch /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11971
pragma Warnings (Off, "is not referenced");
 function Bin_Op_P_Potential_Actuals_For_Dispatch
  (Node : Bare_Bin_Op
      ; Spec : Internal_Entity_Base_Subp_Spec
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Expected_Type_For_Expr_Array_Access
is
   Self : Bare_Bin_Op := Bare_Bin_Op (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Bin_Op :=
     Internal_Entity_Bin_Op'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind spec Spec
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Expected_Type_For_Expr_Array_Access;
            procedure Finalizer_Scope_913 with Inline_Always;
      Block_Var : Internal_Entity_Defining_Name_Array_Access;
Fld : Internal_Entity_Defining_Name_Array_Access;
Get_Result : Internal_Entity_Defining_Name;
Fld_1 : Internal_Entity_Base_Formal_Param_Decl;
Fld_2 : Internal_Entity_Type_Expr;
Fld_3 : Internal_Entity_Expr;
New_Struct : Internal_Expected_Type_For_Expr;
Get_Result_1 : Internal_Entity_Defining_Name;
Fld_4 : Internal_Entity_Base_Formal_Param_Decl;
Fld_5 : Internal_Entity_Type_Expr;
Fld_6 : Internal_Entity_Expr;
New_Struct_1 : Internal_Expected_Type_For_Expr;
Array_Lit : Internal_Expected_Type_For_Expr_Array_Access;
Let_Result : Internal_Expected_Type_For_Expr_Array_Access;
            procedure Finalizer_Scope_913 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_913;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6561 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6549 '<FieldAccess for unpacked_formal_params at ast.py:11972>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11972
      if Spec.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => Spec.Node, E_Info => Spec.Info);
--# end
--# expr-done 6549
Block_Var := Fld; Inc_Ref (Block_Var);
--# bind params Block_Var
--# expr-start 6560 '<ArrayLiteral at ast.py:11973>' Array_Lit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11973
--# expr-start 6554 '<New ExpectedTypeForExpr at ast.py:11974>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11974
--# expr-start 6552 '<FieldAccess for type_expression at ast.py:11975>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11975
--# expr-start 6551 '<FieldAccess for formal_decl at ast.py:11975>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11975
--# expr-start 6550 '<at at ast.py:11975>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11975
Get_Result := Get (Block_Var, 0, True); 
--# expr-done 6550
      if Get_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_1 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Get_Result.Node, E_Info => Get_Result.Info);
--# end
--# expr-done 6551
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.type_expression'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Type_Expression (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 6552
--# expr-start 6553 '<FieldAccess for left at ast.py:11976>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11976
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Expr (Node => Ent.Node.Bin_Op_F_Left, Info => Ent.Info);
--# expr-done 6553
New_Struct := (Expected_Type => Fld_2, Expr => Fld_3); 
--# expr-done 6554
--# expr-start 6559 '<New ExpectedTypeForExpr at ast.py:11978>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11978
--# expr-start 6557 '<FieldAccess for type_expression at ast.py:11979>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11979
--# expr-start 6556 '<FieldAccess for formal_decl at ast.py:11979>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11979
--# expr-start 6555 '<at at ast.py:11979>' Get_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11979
Get_Result_1 := Get (Block_Var, 1, True); 
--# expr-done 6555
      if Get_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_4 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Get_Result_1.Node, E_Info => Get_Result_1.Info);
--# end
--# expr-done 6556
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.type_expression'
Fld_5 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Type_Expression (Node => Fld_4.Node, E_Info => Fld_4.Info);
--# end
--# expr-done 6557
--# expr-start 6558 '<FieldAccess for right at ast.py:11980>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11980
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Expr (Node => Ent.Node.Bin_Op_F_Right, Info => Ent.Info);
--# expr-done 6558
New_Struct_1 := (Expected_Type => Fld_5, Expr => Fld_6); 
--# expr-done 6559
Array_Lit := Create_Internal_Expected_Type_For_Expr_Array (Internal_Internal_Expected_Type_For_Expr_Array'(1 => New_Struct, 2 => New_Struct_1)); 
--# expr-done 6560
Let_Result := Array_Lit; Inc_Ref (Let_Result);
--# end
--# expr-done 6561
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_913;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_913;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Bin_Op_P_Potential_Actuals_For_Dispatch;
--# end
--# property-start BinOp.is_dispatching_call /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11985
pragma Warnings (Off, "is not referenced");
 function Bin_Op_P_Is_Dispatching_Call
  (Node : Bare_Bin_Op
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Bin_Op := Bare_Bin_Op (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Bin_Op :=
     Internal_Entity_Bin_Op'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Op;
Fld_1 : Internal_Entity_Basic_Decl;
Fld_2 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6565 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6563 '<FieldAccess for referenced_decl at ast.py:11986>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11986
--# expr-start 6562 '<FieldAccess for op at ast.py:11986>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11986
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Op (Node => Ent.Node.Bin_Op_F_Op, Info => Ent.Info);
--# expr-done 6562
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld_1 := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Fld.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld.Info);
--# end
--# expr-done 6563
Var_Expr := Fld_1;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind decl Var_Expr
--# expr-start 6564 '<FieldAccess for is_dispatching_call_impl at ast.py:11987>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11987
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_dispatching_call_impl
Fld_2 := Libadalang.Implementation.Expr_P_Is_Dispatching_Call_Impl (Node => Ent.Node, Decl => Var_Expr, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 6564
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 6565
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Bin_Op_P_Is_Dispatching_Call;
--# end
--# property-start BinOp.no_overload_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11712
pragma Warnings (Off, "is not referenced");
 function Bin_Op_P_No_Overload_Equation
  (Node : Bare_Bin_Op
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Bin_Op := Bare_Bin_Op (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Bin_Op :=
     Internal_Entity_Bin_Op'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_910 with Inline_Always;
            procedure Finalizer_Scope_2639 with Inline_Always;
            procedure Finalizer_Scope_2640 with Inline_Always;
      Match_0 : Bare_Op_Concat;
Match_1 : Bare_Op;
Fld : Bare_Op;
Fld_1 : Logic_Var;
Pred : Logic_Equation;
Fld_2 : Logic_Var;
Pred_1 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_3 : Logic_Var;
Pred_2 : Logic_Equation;
Fld_4 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_5 : Logic_Var;
Node_For_Entity : Bare_Base_Type_Decl;
As_Entity : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_6 : Bare_Expr;
Fld_7 : Logic_Var;
Fld_8 : Bare_Expr;
Fld_9 : Logic_Var;
Bind_Result_1 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_10 : Bare_Expr;
Fld_11 : Logic_Var;
Fld_12 : Bare_Expr;
Fld_13 : Logic_Var;
Bind_Result_2 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_14 : Bare_Expr;
Fld_15 : Logic_Var;
Fld_16 : Logic_Var;
Bind_Result_3 : Logic_Equation;
Fld_17 : Logic_Var;
Fld_18 : Bare_Expr;
Fld_19 : Logic_Var;
Fld_20 : Logic_Equation;
Or_Pred : Logic_Equation;
And_Pred_4 : Logic_Equation;
Fld_21 : Bare_Expr;
Fld_22 : Logic_Var;
Fld_23 : Logic_Var;
Bind_Result_4 : Logic_Equation;
Fld_24 : Logic_Var;
Fld_25 : Bare_Expr;
Fld_26 : Logic_Var;
Fld_27 : Logic_Equation;
Or_Pred_1 : Logic_Equation;
And_Pred_5 : Logic_Equation;
Or_Pred_2 : Logic_Equation;
And_Pred_6 : Logic_Equation;
Fld_28 : Logic_Var;
Fld_29 : Bare_Expr;
Fld_30 : Logic_Var;
Bind_Result_5 : Logic_Equation;
Fld_31 : Logic_Var;
Fld_32 : Bare_Expr;
Fld_33 : Logic_Var;
Bind_Result_6 : Logic_Equation;
And_Pred_7 : Logic_Equation;
Fld_34 : Bare_Op;
Is_A : Boolean;
Fld_35 : Bare_Expr;
Fld_36 : Logic_Var;
Fld_37 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_7 : Logic_Equation;
Fld_38 : Bare_Expr;
Fld_39 : Logic_Var;
Fld_40 : Logic_Var;
Bind_Result_8 : Logic_Equation;
And_Pred_8 : Logic_Equation;
Fld_41 : Bare_Expr;
Fld_42 : Logic_Var;
Fld_43 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result_9 : Logic_Equation;
Fld_44 : Bare_Expr;
Fld_45 : Logic_Var;
Fld_46 : Logic_Var;
Bind_Result_10 : Logic_Equation;
And_Pred_9 : Logic_Equation;
Or_Pred_3 : Logic_Equation;
Fld_47 : Bare_Op;
Is_A_1 : Boolean;
Fld_48 : Bare_Expr;
Fld_49 : Logic_Var;
Fld_50 : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity;
Bind_Result_11 : Logic_Equation;
Fld_51 : Bare_Expr;
Fld_52 : Logic_Var;
Fld_53 : Logic_Var;
Bind_Result_12 : Logic_Equation;
And_Pred_10 : Logic_Equation;
False_Rel : Logic_Equation;
If_Result_1 : Logic_Equation;
If_Result_2 : Logic_Equation;
Or_Pred_4 : Logic_Equation;
Fld_54 : Logic_Var;
Pred_3 : Logic_Equation;
Fld_55 : Logic_Var;
Fld_56 : Logic_Var;
Bind_Result_13 : Logic_Equation;
And_Pred_11 : Logic_Equation;
Fld_57 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_4 : Internal_Entity;
Bind_Result_14 : Logic_Equation;
Or_Pred_5 : Logic_Equation;
And_Pred_12 : Logic_Equation;
Fld_58 : Logic_Var;
Pred_4 : Logic_Equation;
Fld_59 : Bare_Expr;
Fld_60 : Logic_Equation;
And_Pred_13 : Logic_Equation;
Fld_61 : Bare_Expr;
Fld_62 : Logic_Equation;
And_Pred_14 : Logic_Equation;
Fld_63 : Logic_Var;
Fld_64 : Logic_Equation;
Or_Pred_6 : Logic_Equation;
And_Pred_15 : Logic_Equation;
Match_Prefix : Bare_Op;
Cast_Result_5 : Bare_Op_Concat;
Let_Result : Logic_Equation;
Scope_Result : Logic_Equation;
Cast_Result_6 : Bare_Op;
Let_Result_1 : Logic_Equation;
Scope_Result_1 : Logic_Equation;
Match_Result : Logic_Equation;
Let_Result_2 : Logic_Equation;
            procedure Finalizer_Scope_910 is
            begin
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Match_Result);
                     Dec_Ref (Let_Result_2);
            end Finalizer_Scope_910;
            procedure Finalizer_Scope_2639 is
            begin
                     Dec_Ref (Pred);
                     Dec_Ref (Pred_1);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Pred_2);
                     Dec_Ref (Fld_4);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Bind_Result_3);
                     Dec_Ref (Fld_20);
                     Dec_Ref (Or_Pred);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Bind_Result_4);
                     Dec_Ref (Fld_27);
                     Dec_Ref (Or_Pred_1);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (Or_Pred_2);
                     Dec_Ref (And_Pred_6);
            end Finalizer_Scope_2639;
            procedure Finalizer_Scope_2640 is
            begin
                     Dec_Ref (Bind_Result_5);
                     Dec_Ref (Bind_Result_6);
                     Dec_Ref (And_Pred_7);
                     Dec_Ref (Bind_Result_7);
                     Dec_Ref (Bind_Result_8);
                     Dec_Ref (And_Pred_8);
                     Dec_Ref (Bind_Result_9);
                     Dec_Ref (Bind_Result_10);
                     Dec_Ref (And_Pred_9);
                     Dec_Ref (Or_Pred_3);
                     Dec_Ref (Bind_Result_11);
                     Dec_Ref (Bind_Result_12);
                     Dec_Ref (And_Pred_10);
                     Dec_Ref (False_Rel);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (If_Result_2);
                     Dec_Ref (Or_Pred_4);
                     Dec_Ref (Pred_3);
                     Dec_Ref (Bind_Result_13);
                     Dec_Ref (And_Pred_11);
                     Dec_Ref (Bind_Result_14);
                     Dec_Ref (Or_Pred_5);
                     Dec_Ref (And_Pred_12);
                     Dec_Ref (Pred_4);
                     Dec_Ref (Fld_60);
                     Dec_Ref (And_Pred_13);
                     Dec_Ref (Fld_62);
                     Dec_Ref (And_Pred_14);
                     Dec_Ref (Fld_64);
                     Dec_Ref (Or_Pred_6);
                     Dec_Ref (And_Pred_15);
            end Finalizer_Scope_2640;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6679 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 6678 '<Match at ast.py:11717>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11717
--# expr-start 6566 '<FieldAccess for op at ast.py:11717>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11717
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Bin_Op_F_Op;
--# expr-done 6566
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld; 
case Ada_Op (Match_Prefix.Kind) is
         when Ada_Op_Concat =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Match_Prefix;
Match_0 := Cast_Result_5; 
--# bind _ Match_0
--# expr-start 6612 '<And at ast.py:11719>' And_Pred_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11719
--# expr-start 6571 '<And at ast.py:11719>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11719
--# expr-start 6568 '<Predicate on BaseTypeDecl.is_array_def_with_deref_or_null at ast.py:11719>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11719
--# expr-start 6567 '<FieldAccess for expected_type_var at ast.py:11720>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11720
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6567
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
Pred := Solver.Create_Predicate (Fld_1, Create_Base_Type_Decl_P_Is_Array_Def_With_Deref_Or_Null_0_Predicate (Origin)); 
--# expr-done 6568
--# expr-start 6570 '<Predicate on BaseTypeDecl.is_array_def_with_deref at ast.py:11721>' Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11721
--# expr-start 6569 '<FieldAccess for type_var at ast.py:11721>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11721
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6569
Fld_2.Value := No_Entity;
Entity_Vars.Reset (Fld_2);
Pred_1 := Solver.Create_Predicate (Fld_2, Create_Base_Type_Decl_P_Is_Array_Def_With_Deref_0_Predicate (Origin)); 
--# expr-done 6570
And_Pred := Create_And (Pred, Pred_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11719") else null)); 
--# expr-done 6571
--# expr-start 6611 '<Or at ast.py:11722>' Or_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11722
--# expr-start 6575 '<And at ast.py:11725>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11725
--# expr-start 6573 '<Predicate on AdaNode.is_not_null at ast.py:11725>' Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11725
--# expr-start 6572 '<FieldAccess for expected_type_var at ast.py:11725>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11725
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6572
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
Pred_2 := Solver.Create_Predicate (Fld_3, Create_Ada_Node_P_Is_Not_Null_0_Predicate); 
--# expr-done 6573
--# expr-start 6574 '<FieldAccess for array_concat_expected_type_equation at ast.py:11726>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11726
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BinOp.array_concat_expected_type_equation
Fld_4 := Libadalang.Implementation.Bin_Op_P_Array_Concat_Expected_Type_Equation (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 6574
And_Pred_1 := Create_And (Pred_2, Fld_4, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11725") else null)); 
--# expr-done 6575
--# expr-start 6610 '<And at ast.py:11732>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11732
--# expr-start 6600 '<And at ast.py:11732>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11732
--# expr-start 6590 '<And at ast.py:11732>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11732
--# expr-start 6584 '<And at ast.py:11732>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11732
--# expr-start 6578 '<Bind at ast.py:11732>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11732
--# expr-start 6576 '<FieldAccess for expected_type_var at ast.py:11732>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11732
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6576
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
--# expr-start 6577 '<No BaseTypeDecl at ast.py:11732>' Node_For_Entity /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11732
Node_For_Entity := No_Bare_Ada_Node;
--# expr-done 6577
Is_Null := Node_For_Entity = null; 
if Is_Null then
   If_Result := No_Entity_Base_Type_Decl;
else
As_Entity := (Info => E_Info, Node => Node_For_Entity); 
   If_Result := As_Entity;
end if;
      Cast_Result := Create_Internal_Entity
        (Node => If_Result.Node,
         Info => If_Result.Info);
Bind_Result := Solver.Create_Assign (Fld_5, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11732") else null)); 
--# expr-done 6578
--# expr-start 6583 '<Bind at ast.py:11733>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11733
--# expr-start 6580 '<FieldAccess for type_var at ast.py:11733>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11733
--# expr-start 6579 '<FieldAccess for left at ast.py:11733>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11733
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Self.Bin_Op_F_Left;
--# expr-done 6579
      if Fld_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_6);
--# expr-done 6580
Fld_7.Value := No_Entity;
Entity_Vars.Reset (Fld_7);
--# expr-start 6582 '<FieldAccess for expected_type_var at ast.py:11733>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11733
--# expr-start 6581 '<FieldAccess for left at ast.py:11733>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11733
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Self.Bin_Op_F_Left;
--# expr-done 6581
      if Fld_8 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_8);
--# expr-done 6582
Fld_9.Value := No_Entity;
Entity_Vars.Reset (Fld_9);
Bind_Result_1 := Solver.Create_Unify (Fld_7, Fld_9, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11733") else null)); 
--# expr-done 6583
And_Pred_2 := Create_And (Bind_Result, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11732") else null)); 
--# expr-done 6584
--# expr-start 6589 '<Bind at ast.py:11734>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11734
--# expr-start 6586 '<FieldAccess for type_var at ast.py:11734>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11734
--# expr-start 6585 '<FieldAccess for right at ast.py:11734>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11734
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Self.Bin_Op_F_Right;
--# expr-done 6585
      if Fld_10 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_10);
--# expr-done 6586
Fld_11.Value := No_Entity;
Entity_Vars.Reset (Fld_11);
--# expr-start 6588 '<FieldAccess for expected_type_var at ast.py:11734>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11734
--# expr-start 6587 '<FieldAccess for right at ast.py:11734>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11734
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Self.Bin_Op_F_Right;
--# expr-done 6587
      if Fld_12 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_12);
--# expr-done 6588
Fld_13.Value := No_Entity;
Entity_Vars.Reset (Fld_13);
Bind_Result_2 := Solver.Create_Unify (Fld_11, Fld_13, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11734") else null)); 
--# expr-done 6589
And_Pred_3 := Create_And (And_Pred_2, Bind_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11732") else null)); 
--# expr-done 6590
--# expr-start 6599 '<Or at ast.py:11735>' Or_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11735
--# expr-start 6594 '<Bind at ast.py:11736>' Bind_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11736
--# expr-start 6592 '<FieldAccess for type_var at ast.py:11736>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11736
--# expr-start 6591 '<FieldAccess for left at ast.py:11736>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11736
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Self.Bin_Op_F_Left;
--# expr-done 6591
      if Fld_14 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_14);
--# expr-done 6592
Fld_15.Value := No_Entity;
Entity_Vars.Reset (Fld_15);
--# expr-start 6593 '<FieldAccess for type_var at ast.py:11736>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11736
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_16 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6593
Fld_16.Value := No_Entity;
Entity_Vars.Reset (Fld_16);
Bind_Result_3 := Solver.Create_Unify (Fld_15, Fld_16, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11736") else null)); 
--# expr-done 6594
--# expr-start 6598 '<FieldAccess for comp_bind at ast.py:11737>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11737
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6595 '<FieldAccess for type_var at ast.py:11737>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11737
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_17 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6595
--# expr-start 6597 '<FieldAccess for type_var at ast.py:11737>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11737
--# expr-start 6596 '<FieldAccess for left at ast.py:11737>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11737
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Self.Bin_Op_F_Left;
--# expr-done 6596
      if Fld_18 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_18);
--# expr-done 6597
--# property-call-start AdaNode.comp_bind
Fld_20 := Libadalang.Implementation.Ada_Node_P_Comp_Bind (Node => Self, Left => Fld_17, Right => Fld_19, Origin => Origin);
--# end
--# expr-done 6598
Or_Pred := Create_Or (Bind_Result_3, Fld_20, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11735") else null)); 
--# expr-done 6599
And_Pred_4 := Create_And (And_Pred_3, Or_Pred, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11732") else null)); 
--# expr-done 6600
--# expr-start 6609 '<Or at ast.py:11739>' Or_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11739
--# expr-start 6604 '<Bind at ast.py:11740>' Bind_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11740
--# expr-start 6602 '<FieldAccess for type_var at ast.py:11740>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11740
--# expr-start 6601 '<FieldAccess for right at ast.py:11740>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11740
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_21 := Self.Bin_Op_F_Right;
--# expr-done 6601
      if Fld_21 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_22 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_21);
--# expr-done 6602
Fld_22.Value := No_Entity;
Entity_Vars.Reset (Fld_22);
--# expr-start 6603 '<FieldAccess for type_var at ast.py:11740>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11740
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_23 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6603
Fld_23.Value := No_Entity;
Entity_Vars.Reset (Fld_23);
Bind_Result_4 := Solver.Create_Unify (Fld_22, Fld_23, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11740") else null)); 
--# expr-done 6604
--# expr-start 6608 '<FieldAccess for comp_bind at ast.py:11741>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11741
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6605 '<FieldAccess for type_var at ast.py:11741>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11741
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_24 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6605
--# expr-start 6607 '<FieldAccess for type_var at ast.py:11741>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11741
--# expr-start 6606 '<FieldAccess for right at ast.py:11741>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11741
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_25 := Self.Bin_Op_F_Right;
--# expr-done 6606
      if Fld_25 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_26 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_25);
--# expr-done 6607
--# property-call-start AdaNode.comp_bind
Fld_27 := Libadalang.Implementation.Ada_Node_P_Comp_Bind (Node => Self, Left => Fld_24, Right => Fld_26, Origin => Origin);
--# end
--# expr-done 6608
Or_Pred_1 := Create_Or (Bind_Result_4, Fld_27, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11739") else null)); 
--# expr-done 6609
And_Pred_5 := Create_And (And_Pred_4, Or_Pred_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11732") else null)); 
--# expr-done 6610
Or_Pred_2 := Create_Or (And_Pred_1, And_Pred_5, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11722") else null)); 
--# expr-done 6611
And_Pred_6 := Create_And (And_Pred, Or_Pred_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11719") else null)); 
--# expr-done 6612
Let_Result := And_Pred_6; Inc_Ref (Let_Result);
--# end
Scope_Result := Let_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2639;
            Match_Result := Scope_Result; Inc_Ref (Match_Result);
         when Ada_Op_Abs .. Ada_Op_And_Then | Ada_Op_Div .. Ada_Op_Xor =>
   --# scope-start
--# scope-start
      Cast_Result_6 := Match_Prefix;
Match_1 := Cast_Result_6; 
--# bind _ Match_1
--# expr-start 6677 '<And at ast.py:11746>' And_Pred_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11746
--# expr-start 6665 '<And at ast.py:11746>' And_Pred_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11746
--# expr-start 6655 '<Or at ast.py:11746>' Or_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11746
--# expr-start 6621 '<And at ast.py:11750>' And_Pred_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11750
--# expr-start 6616 '<Bind at ast.py:11751>' Bind_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11751
--# expr-start 6613 '<FieldAccess for type_var at ast.py:11751>' Fld_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11751
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_28 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6613
Fld_28.Value := No_Entity;
Entity_Vars.Reset (Fld_28);
--# expr-start 6615 '<FieldAccess for expected_type_var at ast.py:11751>' Fld_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11751
--# expr-start 6614 '<FieldAccess for left at ast.py:11751>' Fld_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11751
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_29 := Self.Bin_Op_F_Left;
--# expr-done 6614
      if Fld_29 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_30 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_29);
--# expr-done 6615
Fld_30.Value := No_Entity;
Entity_Vars.Reset (Fld_30);
Bind_Result_5 := Solver.Create_Unify (Fld_28, Fld_30, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11751") else null)); 
--# expr-done 6616
--# expr-start 6620 '<Bind at ast.py:11752>' Bind_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11752
--# expr-start 6617 '<FieldAccess for type_var at ast.py:11752>' Fld_31 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11752
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_31 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6617
Fld_31.Value := No_Entity;
Entity_Vars.Reset (Fld_31);
--# expr-start 6619 '<FieldAccess for expected_type_var at ast.py:11752>' Fld_33 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11752
--# expr-start 6618 '<FieldAccess for right at ast.py:11752>' Fld_32 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11752
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_32 := Self.Bin_Op_F_Right;
--# expr-done 6618
      if Fld_32 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_33 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_32);
--# expr-done 6619
Fld_33.Value := No_Entity;
Entity_Vars.Reset (Fld_33);
Bind_Result_6 := Solver.Create_Unify (Fld_31, Fld_33, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11752") else null)); 
--# expr-done 6620
And_Pred_7 := Create_And (Bind_Result_5, Bind_Result_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11750") else null)); 
--# expr-done 6621
--# expr-start 6654 '<Cond at ast.py:11756>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11756
--# expr-start 6623 '<IsA Op.Mult, Op.Div at ast.py:11760>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11760
--# expr-start 6622 '<FieldAccess for op at ast.py:11760>' Fld_34 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11760
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_34 := Self.Bin_Op_F_Op;
--# expr-done 6622
Is_A := Fld_34 /= null 
and then Fld_34.Kind in Ada_Op_Mult_Range | Ada_Op_Div_Range; 
--# expr-done 6623
if Is_A then
   --# expr-start 6642 '<Or at ast.py:11761>' Or_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11761
--# expr-start 6632 '<And at ast.py:11762>' And_Pred_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11762
--# expr-start 6627 '<Bind at ast.py:11762>' Bind_Result_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11762
--# expr-start 6625 '<FieldAccess for expected_type_var at ast.py:11762>' Fld_36 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11762
--# expr-start 6624 '<FieldAccess for left at ast.py:11762>' Fld_35 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11762
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_35 := Self.Bin_Op_F_Left;
--# expr-done 6624
      if Fld_35 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_36 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_35);
--# expr-done 6625
Fld_36.Value := No_Entity;
Entity_Vars.Reset (Fld_36);
--# expr-start 6626 '<FieldAccess for int_type at ast.py:11762>' Fld_37 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11762
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.int_type
Fld_37 := Libadalang.Implementation.Ada_Node_P_Int_Type (Node => Self);
--# end
--# expr-done 6626
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_37.Node,
         Info => Fld_37.Info);
Bind_Result_7 := Solver.Create_Assign (Fld_36, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11762") else null)); 
--# expr-done 6627
--# expr-start 6631 '<Bind at ast.py:11763>' Bind_Result_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11763
--# expr-start 6629 '<FieldAccess for expected_type_var at ast.py:11763>' Fld_39 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11763
--# expr-start 6628 '<FieldAccess for right at ast.py:11763>' Fld_38 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11763
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_38 := Self.Bin_Op_F_Right;
--# expr-done 6628
      if Fld_38 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_39 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_38);
--# expr-done 6629
Fld_39.Value := No_Entity;
Entity_Vars.Reset (Fld_39);
--# expr-start 6630 '<FieldAccess for type_var at ast.py:11763>' Fld_40 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11763
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_40 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6630
Fld_40.Value := No_Entity;
Entity_Vars.Reset (Fld_40);
Bind_Result_8 := Solver.Create_Unify (Fld_39, Fld_40, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11763") else null)); 
--# expr-done 6631
And_Pred_8 := Create_And (Bind_Result_7, Bind_Result_8, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11762") else null)); 
--# expr-done 6632
--# expr-start 6641 '<And at ast.py:11765>' And_Pred_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11765
--# expr-start 6636 '<Bind at ast.py:11765>' Bind_Result_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11765
--# expr-start 6634 '<FieldAccess for expected_type_var at ast.py:11765>' Fld_42 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11765
--# expr-start 6633 '<FieldAccess for right at ast.py:11765>' Fld_41 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11765
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_41 := Self.Bin_Op_F_Right;
--# expr-done 6633
      if Fld_41 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_42 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_41);
--# expr-done 6634
Fld_42.Value := No_Entity;
Entity_Vars.Reset (Fld_42);
--# expr-start 6635 '<FieldAccess for int_type at ast.py:11765>' Fld_43 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11765
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.int_type
Fld_43 := Libadalang.Implementation.Ada_Node_P_Int_Type (Node => Self);
--# end
--# expr-done 6635
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_43.Node,
         Info => Fld_43.Info);
Bind_Result_9 := Solver.Create_Assign (Fld_42, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11765") else null)); 
--# expr-done 6636
--# expr-start 6640 '<Bind at ast.py:11766>' Bind_Result_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11766
--# expr-start 6638 '<FieldAccess for expected_type_var at ast.py:11766>' Fld_45 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11766
--# expr-start 6637 '<FieldAccess for left at ast.py:11766>' Fld_44 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11766
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_44 := Self.Bin_Op_F_Left;
--# expr-done 6637
      if Fld_44 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_45 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_44);
--# expr-done 6638
Fld_45.Value := No_Entity;
Entity_Vars.Reset (Fld_45);
--# expr-start 6639 '<FieldAccess for type_var at ast.py:11766>' Fld_46 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11766
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_46 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6639
Fld_46.Value := No_Entity;
Entity_Vars.Reset (Fld_46);
Bind_Result_10 := Solver.Create_Unify (Fld_45, Fld_46, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11766") else null)); 
--# expr-done 6640
And_Pred_9 := Create_And (Bind_Result_9, Bind_Result_10, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11765") else null)); 
--# expr-done 6641
Or_Pred_3 := Create_Or (And_Pred_8, And_Pred_9, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11761") else null)); 
--# expr-done 6642
   If_Result_2 := Or_Pred_3;
else
--# expr-start 6644 '<IsA Op.Pow at ast.py:11771>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11771
--# expr-start 6643 '<FieldAccess for op at ast.py:11771>' Fld_47 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11771
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_47 := Self.Bin_Op_F_Op;
--# expr-done 6643
Is_A_1 := Fld_47 /= null 
and then Fld_47.Kind in Ada_Op_Pow_Range; 
--# expr-done 6644
if Is_A_1 then
   --# expr-start 6653 '<And at ast.py:11772>' And_Pred_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11772
--# expr-start 6648 '<Bind at ast.py:11772>' Bind_Result_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11772
--# expr-start 6646 '<FieldAccess for expected_type_var at ast.py:11772>' Fld_49 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11772
--# expr-start 6645 '<FieldAccess for right at ast.py:11772>' Fld_48 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11772
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_48 := Self.Bin_Op_F_Right;
--# expr-done 6645
      if Fld_48 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_49 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_48);
--# expr-done 6646
Fld_49.Value := No_Entity;
Entity_Vars.Reset (Fld_49);
--# expr-start 6647 '<FieldAccess for int_type at ast.py:11772>' Fld_50 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11772
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.int_type
Fld_50 := Libadalang.Implementation.Ada_Node_P_Int_Type (Node => Self);
--# end
--# expr-done 6647
      Cast_Result_3 := Create_Internal_Entity
        (Node => Fld_50.Node,
         Info => Fld_50.Info);
Bind_Result_11 := Solver.Create_Assign (Fld_49, Cast_Result_3, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11772") else null)); 
--# expr-done 6648
--# expr-start 6652 '<Bind at ast.py:11773>' Bind_Result_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11773
--# expr-start 6650 '<FieldAccess for expected_type_var at ast.py:11773>' Fld_52 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11773
--# expr-start 6649 '<FieldAccess for left at ast.py:11773>' Fld_51 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11773
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_51 := Self.Bin_Op_F_Left;
--# expr-done 6649
      if Fld_51 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_52 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_51);
--# expr-done 6650
Fld_52.Value := No_Entity;
Entity_Vars.Reset (Fld_52);
--# expr-start 6651 '<FieldAccess for type_var at ast.py:11773>' Fld_53 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11773
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_53 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6651
Fld_53.Value := No_Entity;
Entity_Vars.Reset (Fld_53);
Bind_Result_12 := Solver.Create_Unify (Fld_52, Fld_53, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11773") else null)); 
--# expr-done 6652
And_Pred_10 := Create_And (Bind_Result_11, Bind_Result_12, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11772") else null)); 
--# expr-done 6653
   If_Result_1 := And_Pred_10;
else
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11775") else null)); 
   If_Result_1 := False_Rel;
end if;
      Inc_Ref (If_Result_1);
   If_Result_2 := If_Result_1;
end if;
      Inc_Ref (If_Result_2);
--# expr-done 6654
Or_Pred_4 := Create_Or (And_Pred_7, If_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11746") else null)); 
--# expr-done 6655
--# expr-start 6664 '<Or at ast.py:11778>' Or_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11778
--# expr-start 6661 '<And at ast.py:11782>' And_Pred_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11782
--# expr-start 6657 '<Predicate on AdaNode.is_not_null at ast.py:11782>' Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11782
--# expr-start 6656 '<FieldAccess for expected_type_var at ast.py:11782>' Fld_54 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11782
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_54 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6656
Fld_54.Value := No_Entity;
Entity_Vars.Reset (Fld_54);
Pred_3 := Solver.Create_Predicate (Fld_54, Create_Ada_Node_P_Is_Not_Null_0_Predicate); 
--# expr-done 6657
--# expr-start 6660 '<Bind at ast.py:11783>' Bind_Result_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11783
--# expr-start 6658 '<FieldAccess for expected_type_var at ast.py:11783>' Fld_55 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11783
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_55 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6658
Fld_55.Value := No_Entity;
Entity_Vars.Reset (Fld_55);
--# expr-start 6659 '<FieldAccess for type_var at ast.py:11783>' Fld_56 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11783
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_56 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6659
Fld_56.Value := No_Entity;
Entity_Vars.Reset (Fld_56);
Bind_Result_13 := Solver.Create_Propagate (Fld_55, Fld_56, Logic_Functor_578'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Origin), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11783") else null)); 
--# expr-done 6660
And_Pred_11 := Create_And (Pred_3, Bind_Result_13, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11782") else null)); 
--# expr-done 6661
--# expr-start 6663 '<Bind at ast.py:11790>' Bind_Result_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11790
--# expr-start 6662 '<FieldAccess for expected_type_var at ast.py:11790>' Fld_57 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11790
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_57 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6662
Fld_57.Value := No_Entity;
Entity_Vars.Reset (Fld_57);
Cast_Expr := No_Entity_Base_Type_Decl; 
      Cast_Result_4 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result_14 := Solver.Create_Assign (Fld_57, Cast_Result_4, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11790") else null)); 
--# expr-done 6663
Or_Pred_5 := Create_Or (And_Pred_11, Bind_Result_14, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11778") else null)); 
--# expr-done 6664
And_Pred_12 := Create_And (Or_Pred_4, Or_Pred_5, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11746") else null)); 
--# expr-done 6665
--# expr-start 6676 '<Or at ast.py:11792>' Or_Pred_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11792
--# expr-start 6673 '<And at ast.py:11795>' And_Pred_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11795
--# expr-start 6670 '<And at ast.py:11795>' And_Pred_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11795
--# expr-start 6667 '<Predicate on AdaNode.is_not_null at ast.py:11795>' Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11795
--# expr-start 6666 '<FieldAccess for expected_type_var at ast.py:11795>' Fld_58 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11795
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_58 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6666
Fld_58.Value := No_Entity;
Entity_Vars.Reset (Fld_58);
Pred_4 := Solver.Create_Predicate (Fld_58, Create_Ada_Node_P_Is_Not_Null_0_Predicate); 
--# expr-done 6667
--# expr-start 6669 '<FieldAccess for matches_expected_formal_type at ast.py:11796>' Fld_60 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11796
--# expr-start 6668 '<FieldAccess for left at ast.py:11796>' Fld_59 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11796
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_59 := Self.Bin_Op_F_Left;
--# expr-done 6668
      if Fld_59 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_60 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_59, Origin => Origin);
--# end
--# expr-done 6669
And_Pred_13 := Create_And (Pred_4, Fld_60, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11795") else null)); 
--# expr-done 6670
--# expr-start 6672 '<FieldAccess for matches_expected_formal_type at ast.py:11797>' Fld_62 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11797
--# expr-start 6671 '<FieldAccess for right at ast.py:11797>' Fld_61 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11797
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_61 := Self.Bin_Op_F_Right;
--# expr-done 6671
      if Fld_61 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_62 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_61, Origin => Origin);
--# end
--# expr-done 6672
And_Pred_14 := Create_And (And_Pred_13, Fld_62, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11795") else null)); 
--# expr-done 6673
--# expr-start 6675 '<FieldAccess for numeric_type_from_operands_equation at ast.py:11803>' Fld_64 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11803
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6674 '<FieldAccess for type_var at ast.py:11803>' Fld_63 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11803
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_63 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6674
--# property-call-start BinOp.numeric_type_from_operands_equation
Fld_64 := Libadalang.Implementation.Bin_Op_P_Numeric_Type_From_Operands_Equation (Node => Ent.Node, Dest_Var => Fld_63, Origin => Origin);
--# end
--# expr-done 6675
Or_Pred_6 := Create_Or (And_Pred_14, Fld_64, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11792") else null)); 
--# expr-done 6676
And_Pred_15 := Create_And (And_Pred_12, Or_Pred_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11746") else null)); 
--# expr-done 6677
Let_Result_1 := And_Pred_15; Inc_Ref (Let_Result_1);
--# end
Scope_Result_1 := Let_Result_1;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2640;
            Match_Result := Scope_Result_1; Inc_Ref (Match_Result);
end case;
--# expr-done 6678
Let_Result_2 := Match_Result; Inc_Ref (Let_Result_2);
--# end
--# expr-done 6679
      Property_Result := Let_Result_2;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_910;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_910;
               Finalizer_Scope_2639;
               Finalizer_Scope_2640;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Bin_Op_P_No_Overload_Equation;
--# end
   --
   --  Primitives for Bare_Relation_Op
   --
      procedure Initialize_Fields_For_Relation_Op
        (Self : Bare_Relation_Op
         ; Bin_Op_F_Left : Bare_Expr
         ; Bin_Op_F_Op : Bare_Op
         ; Bin_Op_F_Right : Bare_Expr
        ) is
      begin
            Initialize_Fields_For_Bin_Op
              (Self, Bin_Op_F_Left, Bin_Op_F_Op, Bin_Op_F_Right);
      end Initialize_Fields_For_Relation_Op;
--# property-start RelationOp.has_context_free_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11996
pragma Warnings (Off, "is not referenced");
 function Relation_Op_P_Has_Context_Free_Type
  (Node : Bare_Relation_Op
  )
   return Boolean
is
   Self : Bare_Relation_Op := Bare_Relation_Op (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Relation_Op_P_Has_Context_Free_Type;
--# end
--# property-start RelationOp.no_overload_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11998
pragma Warnings (Off, "is not referenced");
 function Relation_Op_P_No_Overload_Equation
  (Node : Bare_Relation_Op
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Relation_Op := Bare_Relation_Op (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Relation_Op :=
     Internal_Entity_Relation_Op'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_916 with Inline_Always;
      Fld : Logic_Var;
Fld_1 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_2 : Bare_Expr;
Fld_3 : Logic_Var;
Fld_4 : Bare_Expr;
Fld_5 : Logic_Var;
Bind_Result_1 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_6 : Bare_Expr;
Fld_7 : Logic_Var;
Fld_8 : Logic_Equation;
And_Pred_1 : Logic_Equation;
            procedure Finalizer_Scope_916 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_8);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_916;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6692 '<And at ast.py:11999>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11999
--# expr-start 6688 '<And at ast.py:11999>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11999
--# expr-start 6682 '<Bind at ast.py:11999>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11999
--# expr-start 6680 '<FieldAccess for type_var at ast.py:11999>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11999
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6680
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
--# expr-start 6681 '<FieldAccess for bool_type at ast.py:11999>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11999
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_1 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 6681
      Cast_Result := Create_Internal_Entity
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
Bind_Result := Solver.Create_Assign (Fld, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11999") else null)); 
--# expr-done 6682
--# expr-start 6687 '<Bind at ast.py:12000>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12000
--# expr-start 6684 '<FieldAccess for expected_type_var at ast.py:12000>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12000
--# expr-start 6683 '<FieldAccess for left at ast.py:12000>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12000
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Bin_Op_F_Left;
--# expr-done 6683
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_2);
--# expr-done 6684
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
--# expr-start 6686 '<FieldAccess for expected_type_var at ast.py:12000>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12000
--# expr-start 6685 '<FieldAccess for right at ast.py:12000>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12000
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Self.Bin_Op_F_Right;
--# expr-done 6685
      if Fld_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_4);
--# expr-done 6686
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
Bind_Result_1 := Solver.Create_Unify (Fld_3, Fld_5, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12000") else null)); 
--# expr-done 6687
And_Pred := Create_And (Bind_Result, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11999") else null)); 
--# expr-done 6688
--# expr-start 6691 '<FieldAccess for numeric_type_from_operands_equation at ast.py:12001>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12001
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6690 '<FieldAccess for expected_type_var at ast.py:12002>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12002
--# expr-start 6689 '<FieldAccess for left at ast.py:12002>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12002
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Self.Bin_Op_F_Left;
--# expr-done 6689
      if Fld_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_6);
--# expr-done 6690
--# property-call-start BinOp.numeric_type_from_operands_equation
Fld_8 := Libadalang.Implementation.Bin_Op_P_Numeric_Type_From_Operands_Equation (Node => Ent.Node, Dest_Var => Fld_7, Origin => Origin);
--# end
--# expr-done 6691
And_Pred_1 := Create_And (And_Pred, Fld_8, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11999") else null)); 
--# expr-done 6692
      Property_Result := And_Pred_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_916;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_916;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Relation_Op_P_No_Overload_Equation;
--# end
   --
   --  Primitives for Bare_Box_Expr
   --
      procedure Initialize_Fields_For_Box_Expr
        (Self : Bare_Box_Expr
        ) is
      begin
            Initialize_Fields_For_Expr
              (Self);
      end Initialize_Fields_For_Box_Expr;
--# property-start BoxExpr.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14571
pragma Warnings (Off, "is not referenced");
 function Box_Expr_P_Xref_Equation
  (Node : Bare_Box_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Box_Expr := Bare_Box_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1048 with Inline_Always;
      Fld : Logic_Var;
Fld_1 : Logic_Var;
Bind_Result : Logic_Equation;
            procedure Finalizer_Scope_1048 is
            begin
                     Dec_Ref (Bind_Result);
            end Finalizer_Scope_1048;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 6695 '<Bind at ast.py:14571>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14571
--# expr-start 6693 '<FieldAccess for type_var at ast.py:14571>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14571
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6693
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
--# expr-start 6694 '<FieldAccess for expected_type_var at ast.py:14571>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14571
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6694
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
Bind_Result := Solver.Create_Unify (Fld, Fld_1, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14571") else null)); 
--# expr-done 6695
      Property_Result := Bind_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1048;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1048;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Box_Expr_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Case_Expr_Alternative
   --
      procedure Initialize_Fields_For_Case_Expr_Alternative
        (Self : Bare_Case_Expr_Alternative
         ; Case_Expr_Alternative_F_Choices : Bare_Alternatives_List
         ; Case_Expr_Alternative_F_Expr : Bare_Expr
        ) is
      begin
            Initialize_Fields_For_Expr
              (Self);
            Self.Case_Expr_Alternative_F_Choices := Case_Expr_Alternative_F_Choices;
            Self.Case_Expr_Alternative_F_Expr := Case_Expr_Alternative_F_Expr;
      end Initialize_Fields_For_Case_Expr_Alternative;
   function Case_Expr_Alternative_F_Choices
     (Node : Bare_Case_Expr_Alternative) return Bare_Alternatives_List
   is
   begin
         return Node.Case_Expr_Alternative_F_Choices;
   end;
   function Case_Expr_Alternative_F_Expr
     (Node : Bare_Case_Expr_Alternative) return Bare_Expr
   is
   begin
         return Node.Case_Expr_Alternative_F_Expr;
   end;
   --
   --  Primitives for Bare_Cond_Expr
   --
      procedure Initialize_Fields_For_Cond_Expr
        (Self : Bare_Cond_Expr
        ) is
      begin
            Initialize_Fields_For_Expr
              (Self);
      end Initialize_Fields_For_Cond_Expr;
--# property-start '[dispatcher]CondExpr.dependent_exprs' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Cond_Expr_P_Dependent_Exprs
  (Node : Bare_Cond_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Expr_Array_Access
is
   Self : Bare_Cond_Expr := Bare_Cond_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Expr_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Cond_Expr (Self.Kind) is
               when Ada_Case_Expr =>
                  --# property-call-start CaseExpr.dependent_exprs
                  Property_Result := Case_Expr_P_Dependent_Exprs
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_If_Expr =>
                  --# property-call-start IfExpr.dependent_exprs
                  Property_Result := If_Expr_P_Dependent_Exprs
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Cond_Expr_P_Dependent_Exprs;
--# end
   --
   --  Primitives for Bare_Case_Expr
   --
      procedure Initialize_Fields_For_Case_Expr
        (Self : Bare_Case_Expr
         ; Case_Expr_F_Expr : Bare_Expr
         ; Case_Expr_F_Cases : Bare_Case_Expr_Alternative_List
        ) is
      begin
            Initialize_Fields_For_Cond_Expr
              (Self);
            Self.Case_Expr_F_Expr := Case_Expr_F_Expr;
            Self.Case_Expr_F_Cases := Case_Expr_F_Cases;
      end Initialize_Fields_For_Case_Expr;
   function Case_Expr_F_Expr
     (Node : Bare_Case_Expr) return Bare_Expr
   is
   begin
         return Node.Case_Expr_F_Expr;
   end;
   function Case_Expr_F_Cases
     (Node : Bare_Case_Expr) return Bare_Case_Expr_Alternative_List
   is
   begin
         return Node.Case_Expr_F_Cases;
   end;
--# property-start CaseExpr.has_context_free_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14714
pragma Warnings (Off, "is not referenced");
 function Case_Expr_P_Has_Context_Free_Type
  (Node : Bare_Case_Expr
  )
   return Boolean
is
   Self : Bare_Case_Expr := Bare_Case_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Case_Expr_P_Has_Context_Free_Type;
--# end
--# property-start CaseExpr.dependent_exprs /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14717
pragma Warnings (Off, "is not referenced");
 function Case_Expr_P_Dependent_Exprs
  (Node : Bare_Case_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Expr_Array_Access
is
   Self : Bare_Case_Expr := Bare_Case_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Case_Expr :=
     Internal_Entity_Case_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Expr_Array_Access;
            procedure Finalizer_Scope_1057 with Inline_Always;
      Fld : Internal_Entity_Case_Expr_Alternative_List;
Fld_1 : Bare_Case_Expr_Alternative_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity_Case_Expr_Alternative;
Is_Null : Boolean;
If_Result : Internal_Entity_Case_Expr_Alternative;
Fld_3 : Internal_Entity_Expr;
Item_224 : Internal_Entity_Case_Expr_Alternative;
Bare_Item_224 : Bare_Case_Expr_Alternative;
Map_Result : Internal_Entity_Expr_Array_Access;
Let_Result : Internal_Entity_Expr_Array_Access;
            procedure Finalizer_Scope_1057 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1057;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 6699 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6698 '<Map at ast.py:14718>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14718
   --# expr-start 6696 '<FieldAccess for cases at ast.py:14718>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14718
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Case_Expr_Alternative_List (Node => Ent.Node.Case_Expr_F_Cases, Info => Ent.Info);
--# expr-done 6696
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Internal_Entity_Expr_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Internal_Entity_Expr_Array (0);
      else
      declare
         Collection : constant Bare_Case_Expr_Alternative_List := Fld_1;
      begin
         for Untyped_Item_224 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_224 := Untyped_Item_224; 
Is_Null := Bare_Item_224 = null; 
if Is_Null then
   If_Result := No_Entity_Case_Expr_Alternative;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_224); 
   If_Result := As_Entity;
end if;
                  Item_224 := If_Result; 
   --# scope-start
               --# bind c Item_224
      --# expr-start 6697 '<FieldAccess for expr at ast.py:14718>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14718
      if Item_224.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Expr (Node => Item_224.Node.Case_Expr_Alternative_F_Expr, Info => Item_224.Info);
--# expr-done 6697
         declare
            Item_To_Append : constant Internal_Entity_Expr := Fld_3;
         begin
            Internal_Entity_Expr_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Expr_Array
           (Items_Count => Natural (Internal_Entity_Expr_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Expr_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Expr_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Expr_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 6698
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 6699
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1057;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1057;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Case_Expr_P_Dependent_Exprs;
--# end
--# property-start CaseExpr.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14721
pragma Warnings (Off, "is not referenced");
 function Case_Expr_P_Xref_Equation
  (Node : Bare_Case_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Case_Expr := Bare_Case_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Case_Expr :=
     Internal_Entity_Case_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1058 with Inline_Always;
            procedure Finalizer_Scope_2643 with Inline_Always;
            procedure Finalizer_Scope_2644 with Inline_Always;
            procedure Finalizer_Scope_2645 with Inline_Always;
            procedure Finalizer_Scope_2647 with Inline_Always;
            procedure Finalizer_Scope_2648 with Inline_Always;
      Block_Var : Boolean;
Match_0 : Internal_Entity_Expr;
Match_1 : Internal_Entity_Subtype_Indication;
Match_2 : Internal_Entity_Others_Designator;
Match_3 : Internal_Entity;
Var_Expr : Internal_Entity_Name;
Fld : Internal_Entity_Expr;
Fld_1 : Bare_Expr;
Fld_2 : Logic_Var;
Pred : Logic_Equation;
Fld_3 : Boolean;
Fld_4 : Logic_Var;
Fld_5 : Logic_Var;
Bind_Result : Logic_Equation;
Fld_6 : Internal_Entity_Case_Expr_Alternative_List;
Fld_7 : Bare_Case_Expr_Alternative_List;
Fld_8 : Internal_Entity_Info;
As_Entity : Internal_Entity_Case_Expr_Alternative;
Is_Null : Boolean;
If_Result : Internal_Entity_Case_Expr_Alternative;
Fld_9 : Internal_Entity_Alternatives_List;
Fld_10 : Bare_Alternatives_List;
Fld_11 : Internal_Entity_Info;
As_Entity_1 : Internal_Entity;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity;
Cast_Expr : Internal_Entity_Expr;
Cast_Result : Internal_Entity_Name;
Fld_12 : Logic_Equation;
Fld_13 : Logic_Var;
Fld_14 : Bare_Expr;
Fld_15 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
Fld_16 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_17 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Cast_Expr_1 : Internal_Entity_Expr;
Cast_Result_1 : Internal_Entity_Name;
Fld_18 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Is_Null_2 : Boolean;
Not_Val : Boolean;
If_Result_2 : Logic_Equation;
Fld_19 : Logic_Equation;
True_Rel : Logic_Equation;
Match_Prefix : Internal_Entity;
Cast_Result_2 : Internal_Entity_Expr;
Let_Result : Logic_Equation;
Scope_Result_1 : Logic_Equation;
Cast_Result_3 : Internal_Entity_Subtype_Indication;
Let_Result_1 : Logic_Equation;
Scope_Result_2 : Logic_Equation;
Cast_Result_4 : Internal_Entity_Others_Designator;
Let_Result_2 : Logic_Equation;
Scope_Result_3 : Logic_Equation;
Cast_Result_5 : Internal_Entity;
Let_Result_3 : Logic_Equation;
Scope_Result_4 : Logic_Equation;
Match_Result : Logic_Equation;
Item_226 : Internal_Entity;
Bare_Item_226 : Bare_Ada_Node;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Fld_20 : Logic_Var;
Fld_21 : Internal_Entity_Expr;
Fld_22 : Logic_Var;
Bind_Result_2 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_23 : Internal_Entity_Expr;
Fld_24 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_25 : Internal_Entity_Expr;
Fld_26 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Item_225 : Internal_Entity_Case_Expr_Alternative;
Bare_Item_225 : Bare_Case_Expr_Alternative;
Map_Result_1 : Logic_Equation_Array_Access;
Logic_Boolean_Op_1 : Logic_Equation;
And_Pred_5 : Logic_Equation;
Let_Result_4 : Logic_Equation;
            procedure Finalizer_Scope_1058 is
            begin
                     Dec_Ref (Pred);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Logic_Boolean_Op_1);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (Let_Result_4);
            end Finalizer_Scope_1058;
            procedure Finalizer_Scope_2643 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Fld_24);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Fld_26);
                     Dec_Ref (And_Pred_4);
            end Finalizer_Scope_2643;
            procedure Finalizer_Scope_2644 is
            begin
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Let_Result_2);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Let_Result_3);
                     Dec_Ref (Scope_Result_4);
                     Dec_Ref (Match_Result);
            end Finalizer_Scope_2644;
            procedure Finalizer_Scope_2645 is
            begin
                     Dec_Ref (Fld_12);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Fld_16);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_17);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (If_Result_2);
            end Finalizer_Scope_2645;
            procedure Finalizer_Scope_2647 is
            begin
                     Dec_Ref (Fld_19);
            end Finalizer_Scope_2647;
            procedure Finalizer_Scope_2648 is
            begin
                     Dec_Ref (True_Rel);
            end Finalizer_Scope_2648;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6743 '<Block at ???>' Let_Result_4 None
--# scope-start
--# expr-start 6704 '<FieldAccess for resolve_names_internal_with_eq at ast.py:14724>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14724
--# expr-start 6700 '<FieldAccess for expr at ast.py:14724>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14724
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Case_Expr_F_Expr, Info => Ent.Info);
--# expr-done 6700
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 6703 '<Predicate on BaseTypeDecl.is_discrete_type at ast.py:14725>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14725
--# expr-start 6702 '<FieldAccess for type_var at ast.py:14725>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14725
--# expr-start 6701 '<FieldAccess for expr at ast.py:14725>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14725
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Self.Case_Expr_F_Expr;
--# expr-done 6701
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_1);
--# expr-done 6702
Fld_2.Value := No_Entity;
Entity_Vars.Reset (Fld_2);
Pred := Solver.Create_Predicate (Fld_2, Create_Base_Type_Decl_P_Is_Discrete_Type_0_Predicate (Origin)); 
--# expr-done 6703
--# property-call-start AdaNode.resolve_names_internal_with_eq
Fld_3 := Libadalang.Implementation.Ada_Node_P_Resolve_Names_Internal_With_Eq (Node => Fld.Node, Additional_Equation => Pred, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 6704
Block_Var := Fld_3; 
--# expr-start 6742 '<And at ast.py:14728>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14728
--# expr-start 6707 '<Bind at ast.py:14729>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14729
--# expr-start 6705 '<FieldAccess for type_var at ast.py:14729>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14729
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6705
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
--# expr-start 6706 '<FieldAccess for expected_type_var at ast.py:14729>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14729
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6706
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
Bind_Result := Solver.Create_Unify (Fld_4, Fld_5, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14729") else null)); 
--# expr-done 6707
--# expr-start 6741 '<LogicAll at ast.py:14730>' Logic_Boolean_Op_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14730
--# expr-start 6740 '<Map at ast.py:14730>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14730
   --# expr-start 6708 '<FieldAccess for cases at ast.py:14730>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14730
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Case_Expr_Alternative_List (Node => Ent.Node.Case_Expr_F_Cases, Info => Ent.Info);
--# expr-done 6708
Fld_7 := Fld_6.Node;
      declare
         Map_Result_1_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_7 = null then
      Map_Result_1 := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Case_Expr_Alternative_List := Fld_7;
      begin
         for Untyped_Item_225 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_225 := Untyped_Item_225; 
Is_Null := Bare_Item_225 = null; 
if Is_Null then
   If_Result := No_Entity_Case_Expr_Alternative;
else
Fld_8 := Fld_6.Info;
As_Entity := (Info => Fld_8, Node => Bare_Item_225); 
   If_Result := As_Entity;
end if;
                  Item_225 := If_Result; 
   --# scope-start
               --# bind alt Item_225
      --# expr-start 6739 '<And at ast.py:14733>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14733
--# expr-start 6736 '<And at ast.py:14733>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14733
--# expr-start 6733 '<And at ast.py:14733>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14733
--# expr-start 6728 '<LogicAll at ast.py:14733>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14733
--# expr-start 6727 '<Map at ast.py:14733>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14733
   --# expr-start 6709 '<FieldAccess for choices at ast.py:14733>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14733
      if Item_225.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Alternatives_List (Node => Item_225.Node.Case_Expr_Alternative_F_Choices, Info => Item_225.Info);
--# expr-done 6709
Fld_10 := Fld_9.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_10 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Alternatives_List := Fld_10;
      begin
         for Untyped_Item_226 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_226 := Untyped_Item_226; 
Is_Null_1 := Bare_Item_226 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity;
else
Fld_11 := Fld_9.Info;
As_Entity_1 := (Info => Fld_11, Node => Bare_Item_226); 
   If_Result_1 := As_Entity_1;
end if;
                  Item_226 := If_Result_1; 
   --# scope-start
               --# bind c Item_226
      --# expr-start 6726 '<Match at ast.py:14733>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14733
      if Item_226.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Item_226; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Abstract_State_Decl_Expr .. Ada_Un_Op =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_2; 
--# bind e Match_0
--# expr-start 6724 '<If at ast.py:14735>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14735
--# expr-start 6713 '<Not at ast.py:14736>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14736
--# expr-start 6712 '<IsNull at ast.py:14736>' Is_Null_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14736
--# expr-start 6710 '<Cast to Name at ast.py:14736>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14736
Cast_Expr_1 := Match_0; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Name
   then
      Cast_Result_1 := Create_Internal_Entity_Name
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_1 := No_Entity_Name;
   end if;
--# expr-done 6710
Var_Expr := Cast_Result_1;
if Var_Expr /= No_Entity_Name then
   --# scope-start
--# expr-start 6711 '<FieldAccess for name_designated_type at ast.py:14736>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14736
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_18 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 6711
Scope_Result := Fld_18;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Is_Null_2 := Result_Var.Node = null; 
--# expr-done 6712
Not_Val := not (Is_Null_2); 
--# expr-done 6713
if Not_Val then
   --# expr-start 6715 '<FieldAccess for xref_no_overloading at ast.py:14738>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14738
--# expr-start 6714 '<Cast to Name at ast.py:14738>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14738
Cast_Expr := Match_0; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Name
   then
      Cast_Result := Create_Internal_Entity_Name
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Name;
   end if;
--# expr-done 6714
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_12 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Cast_Result.Node, Env => Env, Origin => Origin, E_Info => Cast_Result.Info);
--# end
--# expr-done 6715
   If_Result_2 := Fld_12;
else
   --# expr-start 6723 '<And at ast.py:14740>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14740
--# expr-start 6721 '<And at ast.py:14740>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14740
--# expr-start 6719 '<Bind at ast.py:14740>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14740
--# expr-start 6716 '<FieldAccess for expected_type_var at ast.py:14740>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14740
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Match_0.Node);
--# expr-done 6716
Fld_13.Value := No_Entity;
Entity_Vars.Reset (Fld_13);
--# expr-start 6718 '<FieldAccess for type_val at ast.py:14740>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14740
--# expr-start 6717 '<FieldAccess for expr at ast.py:14740>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14740
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Self.Case_Expr_F_Expr;
--# expr-done 6717
      if Fld_14 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.type_val
Fld_15 := Libadalang.Implementation.Expr_P_Type_Val (Node => Fld_14);
--# end
--# expr-done 6718
Bind_Result_1 := Solver.Create_Assign (Fld_13, Fld_15, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14740") else null)); 
--# expr-done 6719
--# expr-start 6720 '<FieldAccess for sub_equation at ast.py:14741>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14741
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_16 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Match_0.Node, Env => Env, Origin => Origin, E_Info => Match_0.Info);
--# end
--# expr-done 6720
And_Pred := Create_And (Bind_Result_1, Fld_16, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14740") else null)); 
--# expr-done 6721
--# expr-start 6722 '<FieldAccess for matches_expected_type at ast.py:14742>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14742
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_17 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Match_0.Node, Origin => Origin);
--# end
--# expr-done 6722
And_Pred_1 := Create_And (And_Pred, Fld_17, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14740") else null)); 
--# expr-done 6723
   If_Result_2 := And_Pred_1;
end if;
      Inc_Ref (If_Result_2);
--# expr-done 6724
Let_Result := If_Result_2; Inc_Ref (Let_Result);
--# end
Scope_Result_1 := Let_Result;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2645;
            Match_Result := Scope_Result_1; Inc_Ref (Match_Result);
         when Ada_Subtype_Indication .. Ada_Discrete_Subtype_Indication =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Subtype_Indication
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_3; 
--# bind t Match_1
--# expr-start 6725 '<FieldAccess for xref_equation at ast.py:14746>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14746
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_19 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Match_1.Node, Env => Env, Origin => Origin, E_Info => Match_1.Info);
--# end
--# expr-done 6725
Let_Result_1 := Fld_19; Inc_Ref (Let_Result_1);
--# end
Scope_Result_2 := Let_Result_1;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2647;
            Match_Result := Scope_Result_2; Inc_Ref (Match_Result);
         when Ada_Others_Designator =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Others_Designator
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_4; 
--# bind _ Match_2
True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14748") else null)); 
Let_Result_2 := True_Rel; Inc_Ref (Let_Result_2);
--# end
Scope_Result_3 := Let_Result_2;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_2648;
            Match_Result := Scope_Result_3; Inc_Ref (Match_Result);
         when Ada_Abort_Absent .. Ada_Elsif_Stmt_Part | Ada_Handled_Stmts .. Ada_Null_Component_Decl | Ada_Overriding_Not_Overriding .. Ada_Enum_Lit_Synth_Type_Expr | Ada_Synthetic_Type_Expr .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_5; 
--# bind _ Match_3
Let_Result_3 := raise Property_Error with "Should not happen"; Inc_Ref (Let_Result_3);
--# end
Scope_Result_4 := Let_Result_3;
   Inc_Ref (Scope_Result_4);
   --# end
            Match_Result := Scope_Result_4; Inc_Ref (Match_Result);
end case;
--# expr-done 6726
         declare
            Item_To_Append : constant Logic_Equation := Match_Result;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2644;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 6727
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14733") else null)); 
--# expr-done 6728
--# expr-start 6732 '<Bind at ast.py:14754>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14754
--# expr-start 6729 '<FieldAccess for expected_type_var at ast.py:14754>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14754
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_20 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6729
Fld_20.Value := No_Entity;
Entity_Vars.Reset (Fld_20);
--# expr-start 6731 '<FieldAccess for expected_type_var at ast.py:14754>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14754
--# expr-start 6730 '<FieldAccess for expr at ast.py:14754>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14754
      if Item_225.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_21 := Create_Internal_Entity_Expr (Node => Item_225.Node.Case_Expr_Alternative_F_Expr, Info => Item_225.Info);
--# expr-done 6730
      if Fld_21.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_22 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_21.Node);
--# expr-done 6731
Fld_22.Value := No_Entity;
Entity_Vars.Reset (Fld_22);
Bind_Result_2 := Solver.Create_Unify (Fld_20, Fld_22, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14754") else null)); 
--# expr-done 6732
And_Pred_2 := Create_And (Logic_Boolean_Op, Bind_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14733") else null)); 
--# expr-done 6733
--# expr-start 6735 '<FieldAccess for sub_equation at ast.py:14755>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14755
--# expr-start 6734 '<FieldAccess for expr at ast.py:14755>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14755
      if Item_225.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_23 := Create_Internal_Entity_Expr (Node => Item_225.Node.Case_Expr_Alternative_F_Expr, Info => Item_225.Info);
--# expr-done 6734
      if Fld_23.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_24 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_23.Node, Env => Env, Origin => Origin, E_Info => Fld_23.Info);
--# end
--# expr-done 6735
And_Pred_3 := Create_And (And_Pred_2, Fld_24, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14733") else null)); 
--# expr-done 6736
--# expr-start 6738 '<FieldAccess for matches_expected_type at ast.py:14756>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14756
--# expr-start 6737 '<FieldAccess for expr at ast.py:14756>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14756
      if Item_225.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_25 := Create_Internal_Entity_Expr (Node => Item_225.Node.Case_Expr_Alternative_F_Expr, Info => Item_225.Info);
--# expr-done 6737
      if Fld_25.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_26 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_25.Node, Origin => Origin);
--# end
--# expr-done 6738
And_Pred_4 := Create_And (And_Pred_3, Fld_26, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14733") else null)); 
--# expr-done 6739
         declare
            Item_To_Append : constant Logic_Equation := And_Pred_4;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2643;
         end loop;
      end;
      end if;
         Map_Result_1 := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_1_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 6740
Logic_Boolean_Op_1 := Solver.Create_All (Relation_Array (Map_Result_1.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14730") else null)); 
--# expr-done 6741
And_Pred_5 := Create_And (Bind_Result, Logic_Boolean_Op_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14728") else null)); 
--# expr-done 6742
Let_Result_4 := And_Pred_5; Inc_Ref (Let_Result_4);
--# end
--# expr-done 6743
      Property_Result := Let_Result_4;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1058;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1058;
               Finalizer_Scope_2643;
               Finalizer_Scope_2644;
               Finalizer_Scope_2645;
               Finalizer_Scope_2647;
               Finalizer_Scope_2648;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Case_Expr_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_If_Expr
   --
      procedure Initialize_Fields_For_If_Expr
        (Self : Bare_If_Expr
         ; If_Expr_F_Cond_Expr : Bare_Expr
         ; If_Expr_F_Then_Expr : Bare_Expr
         ; If_Expr_F_Alternatives : Bare_Elsif_Expr_Part_List
         ; If_Expr_F_Else_Expr : Bare_Expr
        ) is
      begin
            Initialize_Fields_For_Cond_Expr
              (Self);
            Self.If_Expr_F_Cond_Expr := If_Expr_F_Cond_Expr;
            Self.If_Expr_F_Then_Expr := If_Expr_F_Then_Expr;
            Self.If_Expr_F_Alternatives := If_Expr_F_Alternatives;
            Self.If_Expr_F_Else_Expr := If_Expr_F_Else_Expr;
      end Initialize_Fields_For_If_Expr;
   function If_Expr_F_Cond_Expr
     (Node : Bare_If_Expr) return Bare_Expr
   is
   begin
         return Node.If_Expr_F_Cond_Expr;
   end;
   function If_Expr_F_Then_Expr
     (Node : Bare_If_Expr) return Bare_Expr
   is
   begin
         return Node.If_Expr_F_Then_Expr;
   end;
   function If_Expr_F_Alternatives
     (Node : Bare_If_Expr) return Bare_Elsif_Expr_Part_List
   is
   begin
         return Node.If_Expr_F_Alternatives;
   end;
   function If_Expr_F_Else_Expr
     (Node : Bare_If_Expr) return Bare_Expr
   is
   begin
         return Node.If_Expr_F_Else_Expr;
   end;
--# property-start IfExpr.has_context_free_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14607
pragma Warnings (Off, "is not referenced");
 function If_Expr_P_Has_Context_Free_Type
  (Node : Bare_If_Expr
  )
   return Boolean
is
   Self : Bare_If_Expr := Bare_If_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end If_Expr_P_Has_Context_Free_Type;
--# end
--# property-start IfExpr.dependent_exprs /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14610
pragma Warnings (Off, "is not referenced");
 function If_Expr_P_Dependent_Exprs
  (Node : Bare_If_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Expr_Array_Access
is
   Self : Bare_If_Expr := Bare_If_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_If_Expr :=
     Internal_Entity_If_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Expr_Array_Access;
            procedure Finalizer_Scope_1052 with Inline_Always;
      Fld : Internal_Entity_Expr;
Singleton : Internal_Entity_Expr_Array_Access;
Fld_1 : Internal_Entity_Elsif_Expr_Part_List;
Fld_2 : Bare_Elsif_Expr_Part_List;
Fld_3 : Internal_Entity_Info;
As_Entity : Internal_Entity_Elsif_Expr_Part;
Is_Null : Boolean;
If_Result : Internal_Entity_Elsif_Expr_Part;
Fld_4 : Internal_Entity_Expr;
Item_227 : Internal_Entity_Elsif_Expr_Part;
Bare_Item_227 : Bare_Elsif_Expr_Part;
Map_Result : Internal_Entity_Expr_Array_Access;
Concat_Result : Internal_Entity_Expr_Array_Access;
Fld_5 : Internal_Entity_Expr;
Singleton_1 : Internal_Entity_Expr_Array_Access;
Concat_Result_1 : Internal_Entity_Expr_Array_Access;
Let_Result : Internal_Entity_Expr_Array_Access;
            procedure Finalizer_Scope_1052 is
            begin
                     Dec_Ref (Singleton);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Singleton_1);
                     Dec_Ref (Concat_Result_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1052;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 6751 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6750 '<Concat at ast.py:14611>' Concat_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14611
--# expr-start 6748 '<Concat at ast.py:14611>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14611
--# expr-start 6744 '<FieldAccess for then_expr at ast.py:14611>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14611
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.If_Expr_F_Then_Expr, Info => Ent.Info);
--# expr-done 6744
                Singleton := Create_Internal_Entity_Expr_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
--# expr-start 6747 '<Map at ast.py:14612>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14612
   --# expr-start 6745 '<FieldAccess for alternatives at ast.py:14612>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14612
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Elsif_Expr_Part_List (Node => Ent.Node.If_Expr_F_Alternatives, Info => Ent.Info);
--# expr-done 6745
Fld_2 := Fld_1.Node;
      declare
         Map_Result_Vec : Internal_Entity_Expr_Vectors.Vector;
      begin
      if Fld_2 = null then
      Map_Result := Create_Internal_Entity_Expr_Array (0);
      else
      declare
         Collection : constant Bare_Elsif_Expr_Part_List := Fld_2;
      begin
         for Untyped_Item_227 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_227 := Untyped_Item_227; 
Is_Null := Bare_Item_227 = null; 
if Is_Null then
   If_Result := No_Entity_Elsif_Expr_Part;
else
Fld_3 := Fld_1.Info;
As_Entity := (Info => Fld_3, Node => Bare_Item_227); 
   If_Result := As_Entity;
end if;
                  Item_227 := If_Result; 
   --# scope-start
               --# bind a Item_227
      --# expr-start 6746 '<FieldAccess for then_expr at ast.py:14612>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14612
      if Item_227.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Expr (Node => Item_227.Node.Elsif_Expr_Part_F_Then_Expr, Info => Item_227.Info);
--# expr-done 6746
         declare
            Item_To_Append : constant Internal_Entity_Expr := Fld_4;
         begin
            Internal_Entity_Expr_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Expr_Array
           (Items_Count => Natural (Internal_Entity_Expr_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Expr_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Expr_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Expr_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 6747
Concat_Result := Concat (Singleton, Map_Result); 
--# expr-done 6748
--# expr-start 6749 '<FieldAccess for else_expr at ast.py:14613>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14613
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Expr (Node => Ent.Node.If_Expr_F_Else_Expr, Info => Ent.Info);
--# expr-done 6749
                Singleton_1 := Create_Internal_Entity_Expr_Array (Items_Count => 1);
                Singleton_1.Items (1) := Fld_5;
Concat_Result_1 := Concat (Concat_Result, Singleton_1); 
--# expr-done 6750
Let_Result := Concat_Result_1; Inc_Ref (Let_Result);
--# end
--# expr-done 6751
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1052;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1052;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end If_Expr_P_Dependent_Exprs;
--# end
--# property-start IfExpr.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14616
pragma Warnings (Off, "is not referenced");
 function If_Expr_P_Xref_Equation
  (Node : Bare_If_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_If_Expr := Bare_If_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_If_Expr :=
     Internal_Entity_If_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1053 with Inline_Always;
            procedure Finalizer_Scope_2651 with Inline_Always;
      Fld : Bare_Expr;
Fld_1 : Logic_Var;
Fld_2 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_3 : Internal_Entity_Expr;
Fld_4 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_5 : Internal_Entity_Expr;
Fld_6 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_7 : Internal_Entity_Expr;
Fld_8 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_9 : Internal_Entity_Elsif_Expr_Part_List;
Fld_10 : Bare_Elsif_Expr_Part_List;
Fld_11 : Internal_Entity_Info;
As_Entity : Internal_Entity_Elsif_Expr_Part;
Is_Null : Boolean;
If_Result : Internal_Entity_Elsif_Expr_Part;
Fld_12 : Internal_Entity_Expr;
Fld_13 : Logic_Var;
Fld_14 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
Fld_15 : Internal_Entity_Expr;
Fld_16 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_17 : Internal_Entity_Expr;
Fld_18 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Fld_19 : Internal_Entity_Expr;
Fld_20 : Logic_Equation;
And_Pred_5 : Logic_Equation;
Item_228 : Internal_Entity_Elsif_Expr_Part;
Bare_Item_228 : Bare_Elsif_Expr_Part;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
And_Pred_6 : Logic_Equation;
Fld_21 : Internal_Entity_Expr;
Fld_22 : Logic_Equation;
Fld_23 : Bare_Expr;
Fld_24 : Logic_Var;
Fld_25 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result_2 : Logic_Equation;
Fld_26 : Internal_Entity_Expr;
Fld_27 : Logic_Equation;
And_Pred_7 : Logic_Equation;
Fld_28 : Bare_Expr;
Is_Null_1 : Boolean;
Not_Val : Boolean;
If_Result_1 : Logic_Equation;
And_Pred_8 : Logic_Equation;
Fld_29 : Logic_Var;
Pred : Logic_Equation;
Fld_30 : Logic_Equation;
And_Pred_9 : Logic_Equation;
Fld_31 : Logic_Var;
Node_For_Entity : Bare_Base_Type_Decl;
As_Entity_1 : Internal_Entity_Base_Type_Decl;
Is_Null_2 : Boolean;
If_Result_2 : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity;
Bind_Result_3 : Logic_Equation;
Fld_32 : Logic_Equation;
And_Pred_10 : Logic_Equation;
Or_Pred : Logic_Equation;
And_Pred_11 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1053 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_4);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_8);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (And_Pred_6);
                     Dec_Ref (Fld_22);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (Fld_27);
                     Dec_Ref (And_Pred_7);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (And_Pred_8);
                     Dec_Ref (Pred);
                     Dec_Ref (Fld_30);
                     Dec_Ref (And_Pred_9);
                     Dec_Ref (Bind_Result_3);
                     Dec_Ref (Fld_32);
                     Dec_Ref (And_Pred_10);
                     Dec_Ref (Or_Pred);
                     Dec_Ref (And_Pred_11);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1053;
            procedure Finalizer_Scope_2651 is
            begin
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Fld_16);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Fld_18);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Fld_20);
                     Dec_Ref (And_Pred_5);
            end Finalizer_Scope_2651;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6807 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6806 '<And at ast.py:14619>' And_Pred_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14619
--# expr-start 6795 '<And at ast.py:14619>' And_Pred_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14619
--# expr-start 6781 '<And at ast.py:14619>' And_Pred_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14619
--# expr-start 6764 '<And at ast.py:14619>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14619
--# expr-start 6761 '<And at ast.py:14619>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14619
--# expr-start 6758 '<And at ast.py:14619>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14619
--# expr-start 6755 '<Bind at ast.py:14619>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14619
--# expr-start 6753 '<FieldAccess for expected_type_var at ast.py:14619>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14619
--# expr-start 6752 '<FieldAccess for cond_expr at ast.py:14619>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14619
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.If_Expr_F_Cond_Expr;
--# expr-done 6752
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld);
--# expr-done 6753
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 6754 '<FieldAccess for bool_type at ast.py:14619>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14619
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_2 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 6754
      Cast_Result := Create_Internal_Entity
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Bind_Result := Solver.Create_Assign (Fld_1, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14619") else null)); 
--# expr-done 6755
--# expr-start 6757 '<FieldAccess for sub_equation at ast.py:14620>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14620
--# expr-start 6756 '<FieldAccess for cond_expr at ast.py:14620>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14620
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Expr (Node => Ent.Node.If_Expr_F_Cond_Expr, Info => Ent.Info);
--# expr-done 6756
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_4 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_3.Node, Env => Env, Origin => Origin, E_Info => Fld_3.Info);
--# end
--# expr-done 6757
And_Pred := Create_And (Bind_Result, Fld_4, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14619") else null)); 
--# expr-done 6758
--# expr-start 6760 '<FieldAccess for matches_expected_formal_prim_type at ast.py:14621>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14621
--# expr-start 6759 '<FieldAccess for cond_expr at ast.py:14621>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14621
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Expr (Node => Ent.Node.If_Expr_F_Cond_Expr, Info => Ent.Info);
--# expr-done 6759
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_prim_type
Fld_6 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Prim_Type (Node => Fld_5.Node, Origin => Origin);
--# end
--# expr-done 6760
And_Pred_1 := Create_And (And_Pred, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14619") else null)); 
--# expr-done 6761
--# expr-start 6763 '<FieldAccess for sub_equation at ast.py:14624>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14624
--# expr-start 6762 '<FieldAccess for then_expr at ast.py:14624>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14624
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Expr (Node => Ent.Node.If_Expr_F_Then_Expr, Info => Ent.Info);
--# expr-done 6762
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_8 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_7.Node, Env => Env, Origin => Origin, E_Info => Fld_7.Info);
--# end
--# expr-done 6763
And_Pred_2 := Create_And (And_Pred_1, Fld_8, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14619") else null)); 
--# expr-done 6764
--# expr-start 6780 '<LogicAll at ast.py:14626>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14626
--# expr-start 6779 '<Map at ast.py:14626>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14626
   --# expr-start 6765 '<FieldAccess for alternatives at ast.py:14626>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14626
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Elsif_Expr_Part_List (Node => Ent.Node.If_Expr_F_Alternatives, Info => Ent.Info);
--# expr-done 6765
Fld_10 := Fld_9.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_10 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Elsif_Expr_Part_List := Fld_10;
      begin
         for Untyped_Item_228 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_228 := Untyped_Item_228; 
Is_Null := Bare_Item_228 = null; 
if Is_Null then
   If_Result := No_Entity_Elsif_Expr_Part;
else
Fld_11 := Fld_9.Info;
As_Entity := (Info => Fld_11, Node => Bare_Item_228); 
   If_Result := As_Entity;
end if;
                  Item_228 := If_Result; 
   --# scope-start
               --# bind elsif Item_228
      --# expr-start 6778 '<And at ast.py:14629>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14629
--# expr-start 6775 '<And at ast.py:14629>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14629
--# expr-start 6772 '<And at ast.py:14629>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14629
--# expr-start 6769 '<Bind at ast.py:14629>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14629
--# expr-start 6767 '<FieldAccess for expected_type_var at ast.py:14629>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14629
--# expr-start 6766 '<FieldAccess for cond_expr at ast.py:14629>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14629
      if Item_228.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Create_Internal_Entity_Expr (Node => Item_228.Node.Elsif_Expr_Part_F_Cond_Expr, Info => Item_228.Info);
--# expr-done 6766
      if Fld_12.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_12.Node);
--# expr-done 6767
Fld_13.Value := No_Entity;
Entity_Vars.Reset (Fld_13);
--# expr-start 6768 '<FieldAccess for bool_type at ast.py:14629>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14629
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_14 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 6768
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_14.Node,
         Info => Fld_14.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_13, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14629") else null)); 
--# expr-done 6769
--# expr-start 6771 '<FieldAccess for sub_equation at ast.py:14630>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14630
--# expr-start 6770 '<FieldAccess for cond_expr at ast.py:14630>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14630
      if Item_228.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Create_Internal_Entity_Expr (Node => Item_228.Node.Elsif_Expr_Part_F_Cond_Expr, Info => Item_228.Info);
--# expr-done 6770
      if Fld_15.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_16 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_15.Node, Env => Env, Origin => Origin, E_Info => Fld_15.Info);
--# end
--# expr-done 6771
And_Pred_3 := Create_And (Bind_Result_1, Fld_16, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14629") else null)); 
--# expr-done 6772
--# expr-start 6774 '<FieldAccess for matches_expected_formal_prim_type at ast.py:14631>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14631
--# expr-start 6773 '<FieldAccess for cond_expr at ast.py:14631>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14631
      if Item_228.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_17 := Create_Internal_Entity_Expr (Node => Item_228.Node.Elsif_Expr_Part_F_Cond_Expr, Info => Item_228.Info);
--# expr-done 6773
      if Fld_17.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_prim_type
Fld_18 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Prim_Type (Node => Fld_17.Node, Origin => Origin);
--# end
--# expr-done 6774
And_Pred_4 := Create_And (And_Pred_3, Fld_18, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14629") else null)); 
--# expr-done 6775
--# expr-start 6777 '<FieldAccess for sub_equation at ast.py:14632>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14632
--# expr-start 6776 '<FieldAccess for then_expr at ast.py:14632>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14632
      if Item_228.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Create_Internal_Entity_Expr (Node => Item_228.Node.Elsif_Expr_Part_F_Then_Expr, Info => Item_228.Info);
--# expr-done 6776
      if Fld_19.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_20 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_19.Node, Env => Env, Origin => Origin, E_Info => Fld_19.Info);
--# end
--# expr-done 6777
And_Pred_5 := Create_And (And_Pred_4, Fld_20, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14629") else null)); 
--# expr-done 6778
         declare
            Item_To_Append : constant Logic_Equation := And_Pred_5;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2651;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 6779
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14626") else null)); 
--# expr-done 6780
And_Pred_6 := Create_And (And_Pred_2, Logic_Boolean_Op, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14619") else null)); 
--# expr-done 6781
--# expr-start 6794 '<If at ast.py:14635>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14635
--# expr-start 6784 '<Not at ast.py:14636>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14636
--# expr-start 6783 '<IsNull at ast.py:14636>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14636
--# expr-start 6782 '<FieldAccess for else_expr at ast.py:14636>' Fld_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14636
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_28 := Self.If_Expr_F_Else_Expr;
--# expr-done 6782
Is_Null_1 := Fld_28 = null; 
--# expr-done 6783
Not_Val := not (Is_Null_1); 
--# expr-done 6784
if Not_Val then
   --# expr-start 6786 '<FieldAccess for sub_equation at ast.py:14639>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14639
--# expr-start 6785 '<FieldAccess for else_expr at ast.py:14639>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14639
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_21 := Create_Internal_Entity_Expr (Node => Ent.Node.If_Expr_F_Else_Expr, Info => Ent.Info);
--# expr-done 6785
      if Fld_21.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_22 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_21.Node, Env => Env, Origin => Origin, E_Info => Fld_21.Info);
--# end
--# expr-done 6786
   If_Result_1 := Fld_22;
else
   --# expr-start 6793 '<And at ast.py:14642>' And_Pred_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14642
--# expr-start 6790 '<Bind at ast.py:14642>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14642
--# expr-start 6788 '<FieldAccess for expected_type_var at ast.py:14642>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14642
--# expr-start 6787 '<FieldAccess for then_expr at ast.py:14642>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14642
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_23 := Self.If_Expr_F_Then_Expr;
--# expr-done 6787
      if Fld_23 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_24 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_23);
--# expr-done 6788
Fld_24.Value := No_Entity;
Entity_Vars.Reset (Fld_24);
--# expr-start 6789 '<FieldAccess for bool_type at ast.py:14642>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14642
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_25 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 6789
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_25.Node,
         Info => Fld_25.Info);
Bind_Result_2 := Solver.Create_Assign (Fld_24, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14642") else null)); 
--# expr-done 6790
--# expr-start 6792 '<FieldAccess for matches_expected_formal_prim_type at ast.py:14643>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14643
--# expr-start 6791 '<FieldAccess for then_expr at ast.py:14643>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14643
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_26 := Create_Internal_Entity_Expr (Node => Ent.Node.If_Expr_F_Then_Expr, Info => Ent.Info);
--# expr-done 6791
      if Fld_26.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_prim_type
Fld_27 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Prim_Type (Node => Fld_26.Node, Origin => Origin);
--# end
--# expr-done 6792
And_Pred_7 := Create_And (Bind_Result_2, Fld_27, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14642") else null)); 
--# expr-done 6793
   If_Result_1 := And_Pred_7;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 6794
And_Pred_8 := Create_And (And_Pred_6, If_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14619") else null)); 
--# expr-done 6795
--# expr-start 6805 '<Or at ast.py:14646>' Or_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14646
--# expr-start 6799 '<And at ast.py:14647>' And_Pred_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14647
--# expr-start 6797 '<Predicate on AdaNode.is_not_null at ast.py:14647>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14647
--# expr-start 6796 '<FieldAccess for expected_type_var at ast.py:14647>' Fld_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14647
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_29 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6796
Fld_29.Value := No_Entity;
Entity_Vars.Reset (Fld_29);
Pred := Solver.Create_Predicate (Fld_29, Create_Ada_Node_P_Is_Not_Null_0_Predicate); 
--# expr-done 6797
--# expr-start 6798 '<FieldAccess for expected_type_equation at ast.py:14648>' Fld_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14648
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start IfExpr.expected_type_equation
Fld_30 := Libadalang.Implementation.If_Expr_P_Expected_Type_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 6798
And_Pred_9 := Create_And (Pred, Fld_30, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14647") else null)); 
--# expr-done 6799
--# expr-start 6804 '<And at ast.py:14650>' And_Pred_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14650
--# expr-start 6802 '<Bind at ast.py:14650>' Bind_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14650
--# expr-start 6800 '<FieldAccess for expected_type_var at ast.py:14650>' Fld_31 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14650
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_31 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6800
Fld_31.Value := No_Entity;
Entity_Vars.Reset (Fld_31);
--# expr-start 6801 '<No BaseTypeDecl at ast.py:14650>' Node_For_Entity /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14650
Node_For_Entity := No_Bare_Ada_Node;
--# expr-done 6801
Is_Null_2 := Node_For_Entity = null; 
if Is_Null_2 then
   If_Result_2 := No_Entity_Base_Type_Decl;
else
As_Entity_1 := (Info => E_Info, Node => Node_For_Entity); 
   If_Result_2 := As_Entity_1;
end if;
      Cast_Result_3 := Create_Internal_Entity
        (Node => If_Result_2.Node,
         Info => If_Result_2.Info);
Bind_Result_3 := Solver.Create_Assign (Fld_31, Cast_Result_3, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14650") else null)); 
--# expr-done 6802
--# expr-start 6803 '<FieldAccess for no_expected_type_equation at ast.py:14651>' Fld_32 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14651
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start IfExpr.no_expected_type_equation
Fld_32 := Libadalang.Implementation.If_Expr_P_No_Expected_Type_Equation (Node => Ent.Node, Env => Env, Origin => Origin);
--# end
--# expr-done 6803
And_Pred_10 := Create_And (Bind_Result_3, Fld_32, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14650") else null)); 
--# expr-done 6804
Or_Pred := Create_Or (And_Pred_9, And_Pred_10, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14646") else null)); 
--# expr-done 6805
And_Pred_11 := Create_And (And_Pred_8, Or_Pred, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14619") else null)); 
--# expr-done 6806
Let_Result := And_Pred_11; Inc_Ref (Let_Result);
--# end
--# expr-done 6807
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1053;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1053;
               Finalizer_Scope_2651;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end If_Expr_P_Xref_Equation;
--# end
--# property-start IfExpr.expected_type_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14656
pragma Warnings (Off, "is not referenced");
 function If_Expr_P_Expected_Type_Equation
  (Node : Bare_If_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_If_Expr := Bare_If_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_If_Expr :=
     Internal_Entity_If_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1054 with Inline_Always;
            procedure Finalizer_Scope_2652 with Inline_Always;
      Fld : Logic_Var;
Fld_1 : Logic_Var;
Bind_Result : Logic_Equation;
Fld_2 : Bare_Expr;
Fld_3 : Logic_Var;
Fld_4 : Logic_Var;
Bind_Result_1 : Logic_Equation;
Fld_5 : Bare_Expr;
Fld_6 : Logic_Var;
Fld_7 : Logic_Var;
Bind_Result_2 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_8 : Internal_Entity_Expr;
Fld_9 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_10 : Internal_Entity_Expr;
Fld_11 : Logic_Equation;
And_Pred_2 : Logic_Equation;
True_Rel : Logic_Equation;
Fld_12 : Bare_Expr;
Is_Null : Boolean;
Not_Val : Boolean;
If_Result : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_13 : Internal_Entity_Elsif_Expr_Part_List;
Fld_14 : Bare_Elsif_Expr_Part_List;
Fld_15 : Internal_Entity_Info;
As_Entity : Internal_Entity_Elsif_Expr_Part;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Elsif_Expr_Part;
Fld_16 : Internal_Entity_Expr;
Fld_17 : Logic_Var;
Fld_18 : Logic_Var;
Bind_Result_3 : Logic_Equation;
Fld_19 : Internal_Entity_Expr;
Fld_20 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Item_229 : Internal_Entity_Elsif_Expr_Part;
Bare_Item_229 : Bare_Elsif_Expr_Part;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
And_Pred_5 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1054 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_9);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_11);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (True_Rel);
                     Dec_Ref (If_Result);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1054;
            procedure Finalizer_Scope_2652 is
            begin
                     Dec_Ref (Bind_Result_3);
                     Dec_Ref (Fld_20);
                     Dec_Ref (And_Pred_4);
            end Finalizer_Scope_2652;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6842 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6841 '<And at ast.py:14662>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14662
--# expr-start 6830 '<And at ast.py:14662>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14662
--# expr-start 6810 '<Bind at ast.py:14662>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14662
--# expr-start 6808 '<FieldAccess for type_var at ast.py:14662>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14662
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6808
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
--# expr-start 6809 '<FieldAccess for expected_type_var at ast.py:14662>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14662
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6809
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
Bind_Result := Solver.Create_Unify (Fld, Fld_1, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14662") else null)); 
--# expr-done 6810
--# expr-start 6829 '<If at ast.py:14662>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14662
--# expr-start 6813 '<Not at ast.py:14663>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14663
--# expr-start 6812 '<IsNull at ast.py:14663>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14663
--# expr-start 6811 '<FieldAccess for else_expr at ast.py:14663>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14663
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Self.If_Expr_F_Else_Expr;
--# expr-done 6811
Is_Null := Fld_12 = null; 
--# expr-done 6812
Not_Val := not (Is_Null); 
--# expr-done 6813
if Not_Val then
   --# expr-start 6828 '<And at ast.py:14665>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14665
--# expr-start 6825 '<And at ast.py:14665>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14665
--# expr-start 6822 '<And at ast.py:14665>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14665
--# expr-start 6817 '<Bind at ast.py:14665>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14665
--# expr-start 6815 '<FieldAccess for expected_type_var at ast.py:14665>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14665
--# expr-start 6814 '<FieldAccess for else_expr at ast.py:14665>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14665
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.If_Expr_F_Else_Expr;
--# expr-done 6814
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_2);
--# expr-done 6815
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
--# expr-start 6816 '<FieldAccess for expected_type_var at ast.py:14665>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14665
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6816
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
Bind_Result_1 := Solver.Create_Unify (Fld_3, Fld_4, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14665") else null)); 
--# expr-done 6817
--# expr-start 6821 '<Bind at ast.py:14666>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14666
--# expr-start 6819 '<FieldAccess for expected_type_var at ast.py:14666>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14666
--# expr-start 6818 '<FieldAccess for then_expr at ast.py:14666>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14666
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Self.If_Expr_F_Then_Expr;
--# expr-done 6818
      if Fld_5 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_5);
--# expr-done 6819
Fld_6.Value := No_Entity;
Entity_Vars.Reset (Fld_6);
--# expr-start 6820 '<FieldAccess for expected_type_var at ast.py:14666>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14666
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6820
Fld_7.Value := No_Entity;
Entity_Vars.Reset (Fld_7);
Bind_Result_2 := Solver.Create_Unify (Fld_6, Fld_7, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14666") else null)); 
--# expr-done 6821
And_Pred := Create_And (Bind_Result_1, Bind_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14665") else null)); 
--# expr-done 6822
--# expr-start 6824 '<FieldAccess for matches_expected_formal_type at ast.py:14667>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14667
--# expr-start 6823 '<FieldAccess for else_expr at ast.py:14667>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14667
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Expr (Node => Ent.Node.If_Expr_F_Else_Expr, Info => Ent.Info);
--# expr-done 6823
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_9 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_8.Node, Origin => Origin);
--# end
--# expr-done 6824
And_Pred_1 := Create_And (And_Pred, Fld_9, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14665") else null)); 
--# expr-done 6825
--# expr-start 6827 '<FieldAccess for matches_expected_formal_type at ast.py:14668>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14668
--# expr-start 6826 '<FieldAccess for then_expr at ast.py:14668>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14668
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Create_Internal_Entity_Expr (Node => Ent.Node.If_Expr_F_Then_Expr, Info => Ent.Info);
--# expr-done 6826
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_11 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_10.Node, Origin => Origin);
--# end
--# expr-done 6827
And_Pred_2 := Create_And (And_Pred_1, Fld_11, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14665") else null)); 
--# expr-done 6828
   If_Result := And_Pred_2;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14670") else null)); 
   If_Result := True_Rel;
end if;
      Inc_Ref (If_Result);
--# expr-done 6829
And_Pred_3 := Create_And (Bind_Result, If_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14662") else null)); 
--# expr-done 6830
--# expr-start 6840 '<LogicAll at ast.py:14671>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14671
--# expr-start 6839 '<Map at ast.py:14671>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14671
   --# expr-start 6831 '<FieldAccess for alternatives at ast.py:14671>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14671
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Create_Internal_Entity_Elsif_Expr_Part_List (Node => Ent.Node.If_Expr_F_Alternatives, Info => Ent.Info);
--# expr-done 6831
Fld_14 := Fld_13.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_14 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Elsif_Expr_Part_List := Fld_14;
      begin
         for Untyped_Item_229 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_229 := Untyped_Item_229; 
Is_Null_1 := Bare_Item_229 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Elsif_Expr_Part;
else
Fld_15 := Fld_13.Info;
As_Entity := (Info => Fld_15, Node => Bare_Item_229); 
   If_Result_1 := As_Entity;
end if;
                  Item_229 := If_Result_1; 
   --# scope-start
               --# bind elsif Item_229
      --# expr-start 6838 '<And at ast.py:14673>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14673
--# expr-start 6835 '<Bind at ast.py:14673>' Bind_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14673
--# expr-start 6833 '<FieldAccess for expected_type_var at ast.py:14673>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14673
--# expr-start 6832 '<FieldAccess for then_expr at ast.py:14673>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14673
      if Item_229.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_16 := Create_Internal_Entity_Expr (Node => Item_229.Node.Elsif_Expr_Part_F_Then_Expr, Info => Item_229.Info);
--# expr-done 6832
      if Fld_16.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_17 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_16.Node);
--# expr-done 6833
Fld_17.Value := No_Entity;
Entity_Vars.Reset (Fld_17);
--# expr-start 6834 '<FieldAccess for expected_type_var at ast.py:14673>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14673
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6834
Fld_18.Value := No_Entity;
Entity_Vars.Reset (Fld_18);
Bind_Result_3 := Solver.Create_Unify (Fld_17, Fld_18, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14673") else null)); 
--# expr-done 6835
--# expr-start 6837 '<FieldAccess for matches_expected_formal_type at ast.py:14674>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14674
--# expr-start 6836 '<FieldAccess for then_expr at ast.py:14674>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14674
      if Item_229.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Create_Internal_Entity_Expr (Node => Item_229.Node.Elsif_Expr_Part_F_Then_Expr, Info => Item_229.Info);
--# expr-done 6836
      if Fld_19.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_20 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_19.Node, Origin => Origin);
--# end
--# expr-done 6837
And_Pred_4 := Create_And (Bind_Result_3, Fld_20, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14673") else null)); 
--# expr-done 6838
         declare
            Item_To_Append : constant Logic_Equation := And_Pred_4;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2652;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 6839
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14671") else null)); 
--# expr-done 6840
And_Pred_5 := Create_And (And_Pred_3, Logic_Boolean_Op, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14662") else null)); 
--# expr-done 6841
Let_Result := And_Pred_5; Inc_Ref (Let_Result);
--# end
--# expr-done 6842
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1054;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1054;
               Finalizer_Scope_2652;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end If_Expr_P_Expected_Type_Equation;
--# end
--# property-start IfExpr.no_expected_type_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14678
pragma Warnings (Off, "is not referenced");
 function If_Expr_P_No_Expected_Type_Equation
  (Node : Bare_If_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
  )
   return Logic_Equation
is
   Self : Bare_If_Expr := Bare_If_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1055 with Inline_Always;
            procedure Finalizer_Scope_2655 with Inline_Always;
            procedure Finalizer_Scope_2657 with Inline_Always;
      Var_Expr : Bare_Expr;
Fld : Bare_Expr;
Singleton : Bare_Expr_Array_Access;
Fld_1 : Bare_Elsif_Expr_Part_List;
Fld_2 : Bare_Expr;
Item_232 : Bare_Elsif_Expr_Part;
Map_Result : Bare_Expr_Array_Access;
Concat_Result : Bare_Expr_Array_Access;
Fld_3 : Bare_Expr;
Singleton_1 : Bare_Expr_Array_Access;
Scope_Result : Bare_Expr_Array_Access;
Result_Var : Bare_Expr_Array_Access;
Concat_Result_1 : Bare_Expr_Array_Access;
Fld_4 : Boolean;
Map_Result_1 : Bare_Expr_Array_Access;
Fld_5 : Logic_Var;
Pred : Logic_Equation;
Fld_6 : Logic_Var;
Fld_7 : Logic_Var;
Bind_Result : Logic_Equation;
Or_Pred : Logic_Equation;
Map_Result_2 : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1055 is
            begin
                     Dec_Ref (Singleton);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Concat_Result_1);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Map_Result_2);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1055;
            procedure Finalizer_Scope_2655 is
            begin
                     Dec_Ref (Singleton_1);
            end Finalizer_Scope_2655;
            procedure Finalizer_Scope_2657 is
            begin
                     Dec_Ref (Pred);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Or_Pred);
            end Finalizer_Scope_2657;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6860 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6859 '<LogicAll at ast.py:14685>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14685
--# expr-start 6858 '<Map at ast.py:14685>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14685
   --# expr-start 6851 '<Filter at ast.py:14685>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14685
   --# expr-start 6849 '<Concat at ast.py:14685>' Concat_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14685
--# expr-start 6847 '<Concat at ast.py:14685>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14685
--# expr-start 6843 '<FieldAccess for then_expr at ast.py:14685>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14685
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.If_Expr_F_Then_Expr;
--# expr-done 6843
                Singleton := Create_Bare_Expr_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
--# expr-start 6846 '<Map at ast.py:14686>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14686
   --# expr-start 6844 '<FieldAccess for alternatives at ast.py:14686>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14686
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Self.If_Expr_F_Alternatives;
--# expr-done 6844
      declare
         Map_Result_Vec : Bare_Expr_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Bare_Expr_Array (0);
      else
      declare
         Collection : constant Bare_Elsif_Expr_Part_List := Fld_1;
      begin
         for Untyped_Item_232 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_232 := Untyped_Item_232; 
   --# scope-start
               --# bind a Item_232
      --# expr-start 6845 '<FieldAccess for then_expr at ast.py:14686>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14686
      if Item_232 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Item_232.Elsif_Expr_Part_F_Then_Expr;
--# expr-done 6845
         declare
            Item_To_Append : constant Bare_Expr := Fld_2;
         begin
            Bare_Expr_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Bare_Expr_Array
           (Items_Count => Natural (Bare_Expr_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Bare_Expr_Vectors.Get
              (Map_Result_Vec,
               I + Bare_Expr_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Bare_Expr_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 6846
Concat_Result := Concat (Singleton, Map_Result); 
--# expr-done 6847
--# expr-start 6848 '<FieldAccess for else_expr at ast.py:14688>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14688
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Self.If_Expr_F_Else_Expr;
--# expr-done 6848
Var_Expr := Fld_3;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
                Singleton_1 := Create_Bare_Expr_Array (Items_Count => 1);
                Singleton_1.Items (1) := Var_Expr;
Scope_Result := Singleton_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2655;
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Expr_Array_Type;
end if;
      Inc_Ref (Result_Var);
Concat_Result_1 := Concat (Concat_Result, Result_Var); 
--# expr-done 6849
      declare
         Map_Result_1_Vec : Bare_Expr_Vectors.Vector;
      begin
      declare
         Collection : constant Bare_Expr_Array_Access := Concat_Result_1;
      begin
         for Item_231 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_231
         --# expr-start 6850 '<FieldAccess for has_context_free_type at ast.py:14690>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14690
      if Item_231 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.has_context_free_type'
Fld_4 := Libadalang.Implementation.Dispatcher_Expr_P_Has_Context_Free_Type (Node => Item_231);
--# end
--# expr-done 6850
         if Fld_4 then
         declare
            Item_To_Append : constant Bare_Expr := Item_231;
         begin
            Bare_Expr_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result_1 := Create_Bare_Expr_Array
           (Items_Count => Natural (Bare_Expr_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Bare_Expr_Vectors.Get
              (Map_Result_1_Vec,
               I + Bare_Expr_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Bare_Expr_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 6851
      declare
         Map_Result_2_Vec : Logic_Equation_Vectors.Vector;
      begin
      declare
         Collection : constant Bare_Expr_Array_Access := Map_Result_1;
      begin
         for Item_230 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_230
      --# expr-start 6857 '<Or at ast.py:14693>' Or_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14693
--# expr-start 6853 '<Predicate on BaseTypeDecl.is_universal_type at ast.py:14693>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14693
--# expr-start 6852 '<FieldAccess for type_var at ast.py:14693>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14693
      if Item_230 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Item_230);
--# expr-done 6852
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
Pred := Solver.Create_Predicate (Fld_5, Create_Base_Type_Decl_P_Is_Universal_Type_0_Predicate); 
--# expr-done 6853
--# expr-start 6856 '<Bind at ast.py:14694>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14694
--# expr-start 6854 '<FieldAccess for type_var at ast.py:14694>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14694
      if Item_230 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Item_230);
--# expr-done 6854
Fld_6.Value := No_Entity;
Entity_Vars.Reset (Fld_6);
--# expr-start 6855 '<FieldAccess for type_var at ast.py:14694>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14694
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6855
Fld_7.Value := No_Entity;
Entity_Vars.Reset (Fld_7);
Bind_Result := Solver.Create_Propagate (Fld_6, Fld_7, Logic_Functor_552'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Origin), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14694") else null)); 
--# expr-done 6856
Or_Pred := Create_Or (Pred, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14693") else null)); 
--# expr-done 6857
         declare
            Item_To_Append : constant Logic_Equation := Or_Pred;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2657;
         end loop;
      end;
         Map_Result_2 := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_2_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 6858
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result_2.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14685") else null)); 
--# expr-done 6859
Let_Result := Logic_Boolean_Op; Inc_Ref (Let_Result);
--# end
--# expr-done 6860
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1055;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1055;
               Finalizer_Scope_2655;
               Finalizer_Scope_2657;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end If_Expr_P_No_Expected_Type_Equation;
--# end
   --
   --  Primitives for Bare_Contract_Cases
   --
      procedure Initialize_Fields_For_Contract_Cases
        (Self : Bare_Contract_Cases
         ; Contract_Cases_F_Contract_Cases : Bare_Contract_Case_Assoc_List
        ) is
      begin
            Initialize_Fields_For_Expr
              (Self);
            Self.Contract_Cases_F_Contract_Cases := Contract_Cases_F_Contract_Cases;
      end Initialize_Fields_For_Contract_Cases;
   function Contract_Cases_F_Contract_Cases
     (Node : Bare_Contract_Cases) return Bare_Contract_Case_Assoc_List
   is
   begin
         return Node.Contract_Cases_F_Contract_Cases;
   end;
   --
   --  Primitives for Bare_Decl_Expr
   --
      procedure Decl_Expr_Pre_Env_Actions
        (Self            : Bare_Decl_Expr;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_128 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Decl_Expr
        (Self : Bare_Decl_Expr
         ; Decl_Expr_F_Decls : Bare_Basic_Decl_List
         ; Decl_Expr_F_Expr : Bare_Expr
        ) is
      begin
            Initialize_Fields_For_Expr
              (Self);
            Self.Decl_Expr_F_Decls := Decl_Expr_F_Decls;
            Self.Decl_Expr_F_Expr := Decl_Expr_F_Expr;
      end Initialize_Fields_For_Decl_Expr;
   function Decl_Expr_F_Decls
     (Node : Bare_Decl_Expr) return Bare_Basic_Decl_List
   is
   begin
         return Node.Decl_Expr_F_Decls;
   end;
   function Decl_Expr_F_Expr
     (Node : Bare_Decl_Expr) return Bare_Expr
   is
   begin
         return Node.Decl_Expr_F_Expr;
   end;
--# property-start DeclExpr.has_context_free_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11505
pragma Warnings (Off, "is not referenced");
 function Decl_Expr_P_Has_Context_Free_Type
  (Node : Bare_Decl_Expr
  )
   return Boolean
is
   Self : Bare_Decl_Expr := Bare_Decl_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Expr;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 6862 '<FieldAccess for has_context_free_type at ast.py:11505>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11505
--# expr-start 6861 '<FieldAccess for expr at ast.py:11505>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11505
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Decl_Expr_F_Expr;
--# expr-done 6861
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.has_context_free_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Expr_P_Has_Context_Free_Type (Node => Fld);
--# end
--# expr-done 6862
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Decl_Expr_P_Has_Context_Free_Type;
--# end
--# property-start DeclExpr.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11508
pragma Warnings (Off, "is not referenced");
 function Decl_Expr_P_Xref_Equation
  (Node : Bare_Decl_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Decl_Expr := Bare_Decl_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Decl_Expr :=
     Internal_Entity_Decl_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_894 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_1 : Internal_Entity_Expr;
Fld_2 : Logic_Equation;
Dyn_Var_Bind_Result : Logic_Equation;
Fld_3 : Bare_Expr;
Fld_4 : Logic_Var;
Fld_5 : Logic_Var;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_6 : Bare_Expr;
Fld_7 : Logic_Var;
Fld_8 : Logic_Var;
Bind_Result_1 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_894 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_894;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6877 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6876 '<And at ast.py:11509>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11509
--# expr-start 6871 '<And at ast.py:11509>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11509
--# expr-start 6866 '<bind at ast.py:11510>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11510
--# expr-start 6863 '<FieldAccess for children_env at ast.py:11510>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11510
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 6863
Bound_Env := Fld; Inc_Ref (Bound_Env);
--# expr-start 6865 '<FieldAccess for sub_equation at ast.py:11510>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11510
--# expr-start 6864 '<FieldAccess for expr at ast.py:11510>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11510
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Expr (Node => Ent.Node.Decl_Expr_F_Expr, Info => Ent.Info);
--# expr-done 6864
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_2 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_1.Node, Env => Bound_Env, Origin => Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 6865
Dyn_Var_Bind_Result := Fld_2; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 6866
--# expr-start 6870 '<Bind at ast.py:11511>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11511
--# expr-start 6868 '<FieldAccess for expected_type_var at ast.py:11511>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11511
--# expr-start 6867 '<FieldAccess for expr at ast.py:11511>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11511
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Self.Decl_Expr_F_Expr;
--# expr-done 6867
      if Fld_3 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_3);
--# expr-done 6868
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
--# expr-start 6869 '<FieldAccess for expected_type_var at ast.py:11511>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11511
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 6869
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
Bind_Result := Solver.Create_Unify (Fld_4, Fld_5, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11511") else null)); 
--# expr-done 6870
And_Pred := Create_And (Dyn_Var_Bind_Result, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11509") else null)); 
--# expr-done 6871
--# expr-start 6875 '<Bind at ast.py:11512>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11512
--# expr-start 6873 '<FieldAccess for type_var at ast.py:11512>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11512
--# expr-start 6872 '<FieldAccess for expr at ast.py:11512>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11512
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Self.Decl_Expr_F_Expr;
--# expr-done 6872
      if Fld_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_6);
--# expr-done 6873
Fld_7.Value := No_Entity;
Entity_Vars.Reset (Fld_7);
--# expr-start 6874 '<FieldAccess for type_var at ast.py:11512>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11512
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6874
Fld_8.Value := No_Entity;
Entity_Vars.Reset (Fld_8);
Bind_Result_1 := Solver.Create_Unify (Fld_7, Fld_8, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11512") else null)); 
--# expr-done 6875
And_Pred_1 := Create_And (And_Pred, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11509") else null)); 
--# expr-done 6876
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 6877
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_894;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_894;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Decl_Expr_P_Xref_Equation;
--# end
--# property-start DeclExpr._env_trans_parent_128 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11501
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_128
  (Node : Bare_Decl_Expr
  )
   return Boolean
is
   Self : Bare_Decl_Expr := Bare_Decl_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_128;
--# end
   --
   --  Primitives for Bare_Membership_Expr
   --
      procedure Initialize_Fields_For_Membership_Expr
        (Self : Bare_Membership_Expr
         ; Membership_Expr_F_Expr : Bare_Expr
         ; Membership_Expr_F_Op : Bare_Op
         ; Membership_Expr_F_Membership_Exprs : Bare_Expr_Alternatives_List
        ) is
      begin
            Initialize_Fields_For_Expr
              (Self);
            Self.Membership_Expr_F_Expr := Membership_Expr_F_Expr;
            Self.Membership_Expr_F_Op := Membership_Expr_F_Op;
            Self.Membership_Expr_F_Membership_Exprs := Membership_Expr_F_Membership_Exprs;
      end Initialize_Fields_For_Membership_Expr;
   function Membership_Expr_F_Expr
     (Node : Bare_Membership_Expr) return Bare_Expr
   is
   begin
         return Node.Membership_Expr_F_Expr;
   end;
   function Membership_Expr_F_Op
     (Node : Bare_Membership_Expr) return Bare_Op
   is
   begin
         return Node.Membership_Expr_F_Op;
   end;
   function Membership_Expr_F_Membership_Exprs
     (Node : Bare_Membership_Expr) return Bare_Expr_Alternatives_List
   is
   begin
         return Node.Membership_Expr_F_Membership_Exprs;
   end;
--# property-start MembershipExpr.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12018
pragma Warnings (Off, "is not referenced");
 function Membership_Expr_P_Xref_Equation
  (Node : Bare_Membership_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Membership_Expr := Bare_Membership_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Membership_Expr :=
     Internal_Entity_Membership_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_917 with Inline_Always;
            procedure Finalizer_Scope_2658 with Inline_Always;
            procedure Finalizer_Scope_2660 with Inline_Always;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Entity_Name;
Fld : Logic_Var;
Fld_1 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_2 : Internal_Entity_Expr;
Fld_3 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_4 : Internal_Entity_Expr_Alternatives_List;
Fld_5 : Bare_Expr_Alternatives_List;
Fld_6 : Internal_Entity_Info;
As_Entity : Internal_Entity_Expr;
Is_Null : Boolean;
If_Result : Internal_Entity_Expr;
Cast_Expr : Internal_Entity_Expr;
Cast_Result_1 : Internal_Entity_Name;
Fld_7 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Cast_Expr_1 : Internal_Entity_Expr;
Cast_Result_2 : Internal_Entity_Name;
Fld_8 : Logic_Equation;
Fld_9 : Boolean;
Fld_10 : Bare_Expr;
Fld_11 : Logic_Var;
Cast_Expr_2 : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
Fld_12 : Bare_Expr;
Fld_13 : Logic_Var;
Pred : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_14 : Bare_Expr;
Fld_15 : Logic_Var;
Fld_16 : Internal_Entity_Base_Type_Decl;
Cast_Result_4 : Internal_Entity;
Bind_Result_2 : Logic_Equation;
Fld_17 : Bare_Expr;
Fld_18 : Logic_Equation;
And_Pred_2 : Logic_Equation;
If_Result_1 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Scope_Result_1 : Logic_Equation;
Fld_19 : Logic_Var;
Fld_20 : Bare_Expr;
Fld_21 : Logic_Var;
Bind_Result_3 : Logic_Equation;
Fld_22 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Fld_23 : Logic_Equation;
And_Pred_5 : Logic_Equation;
Result_Var_1 : Logic_Equation;
Item_233 : Internal_Entity_Expr;
Bare_Item_233 : Bare_Expr;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
And_Pred_6 : Logic_Equation;
            procedure Finalizer_Scope_917 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_3);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (And_Pred_6);
            end Finalizer_Scope_917;
            procedure Finalizer_Scope_2658 is
            begin
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Bind_Result_3);
                     Dec_Ref (Fld_22);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Fld_23);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (Result_Var_1);
            end Finalizer_Scope_2658;
            procedure Finalizer_Scope_2660 is
            begin
                     Dec_Ref (Fld_8);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Pred);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (Fld_18);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (And_Pred_3);
            end Finalizer_Scope_2660;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6916 '<And at ast.py:12019>' And_Pred_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12019
--# expr-start 6883 '<And at ast.py:12019>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12019
--# expr-start 6880 '<Bind at ast.py:12019>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12019
--# expr-start 6878 '<FieldAccess for type_var at ast.py:12019>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12019
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 6878
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
--# expr-start 6879 '<FieldAccess for bool_type at ast.py:12019>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12019
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_1 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 6879
      Cast_Result := Create_Internal_Entity
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
Bind_Result := Solver.Create_Assign (Fld, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12019") else null)); 
--# expr-done 6880
--# expr-start 6882 '<FieldAccess for sub_equation at ast.py:12020>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12020
--# expr-start 6881 '<FieldAccess for expr at ast.py:12020>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12020
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Expr (Node => Ent.Node.Membership_Expr_F_Expr, Info => Ent.Info);
--# expr-done 6881
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_2.Node, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 6882
And_Pred := Create_And (Bind_Result, Fld_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12019") else null)); 
--# expr-done 6883
--# expr-start 6915 '<LogicAll at ast.py:12021>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12021
--# expr-start 6914 '<Map at ast.py:12021>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12021
   --# expr-start 6884 '<FieldAccess for membership_exprs at ast.py:12021>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12021
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Expr_Alternatives_List (Node => Ent.Node.Membership_Expr_F_Membership_Exprs, Info => Ent.Info);
--# expr-done 6884
Fld_5 := Fld_4.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_5 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Expr_Alternatives_List := Fld_5;
      begin
         for Untyped_Item_233 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_233 := Untyped_Item_233; 
Is_Null := Bare_Item_233 = null; 
if Is_Null then
   If_Result := No_Entity_Expr;
else
Fld_6 := Fld_4.Info;
As_Entity := (Info => Fld_6, Node => Bare_Item_233); 
   If_Result := As_Entity;
end if;
                  Item_233 := If_Result; 
   --# scope-start
               --# bind m Item_233
--# expr-start 6885 '<Cast to Name at ast.py:12022>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12022
Cast_Expr := Item_233; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Name
   then
      Cast_Result_1 := Create_Internal_Entity_Name
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result_1 := No_Entity_Name;
   end if;
--# expr-done 6885
Var_Expr_1 := Cast_Result_1;
if Var_Expr_1 /= No_Entity_Name then
   --# scope-start
--# expr-start 6886 '<FieldAccess for name_designated_type at ast.py:12022>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12022
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_7 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 6886
Scope_Result := Fld_7;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Var_Expr := Result_Var;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind typ Var_Expr
--# expr-start 6905 '<And at ast.py:12025>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12025
--# expr-start 6888 '<FieldAccess for xref_no_overloading at ast.py:12025>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12025
--# expr-start 6887 '<Cast to Name at ast.py:12025>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12025
Cast_Expr_1 := Item_233; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Name
   then
      Cast_Result_2 := Create_Internal_Entity_Name
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_2 := No_Entity_Name;
   end if;
--# expr-done 6887
      if Cast_Result_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_8 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Cast_Result_2.Node, Env => Env, Origin => Origin, E_Info => Cast_Result_2.Info);
--# end
--# expr-done 6888
--# expr-start 6904 '<Cond at ast.py:12026>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12026
--# expr-start 6889 '<FieldAccess for is_tagged_type at ast.py:12034>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12034
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_tagged_type'
Fld_9 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Tagged_Type (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 6889
if Fld_9 then
   --# expr-start 6896 '<And at ast.py:12035>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12035
--# expr-start 6892 '<Bind at ast.py:12035>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12035
--# expr-start 6891 '<FieldAccess for expected_type_var at ast.py:12035>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12035
--# expr-start 6890 '<FieldAccess for expr at ast.py:12035>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12035
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Self.Membership_Expr_F_Expr;
--# expr-done 6890
      if Fld_10 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_10);
--# expr-done 6891
Fld_11.Value := No_Entity;
Entity_Vars.Reset (Fld_11);
Cast_Expr_2 := No_Entity_Base_Type_Decl; 
      Cast_Result_3 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_11, Cast_Result_3, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12035") else null)); 
--# expr-done 6892
--# expr-start 6895 '<Predicate on BaseTypeDecl.is_tagged_type_with_deref at ast.py:12036>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12036
--# expr-start 6894 '<FieldAccess for type_var at ast.py:12037>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12037
--# expr-start 6893 '<FieldAccess for expr at ast.py:12037>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12037
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Self.Membership_Expr_F_Expr;
--# expr-done 6893
      if Fld_12 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_12);
--# expr-done 6894
Fld_13.Value := No_Entity;
Entity_Vars.Reset (Fld_13);
Pred := Solver.Create_Predicate (Fld_13, Create_Base_Type_Decl_P_Is_Tagged_Type_With_Deref_0_Predicate (Origin)); 
--# expr-done 6895
And_Pred_1 := Create_And (Bind_Result_1, Pred, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12035") else null)); 
--# expr-done 6896
   If_Result_1 := And_Pred_1;
else
   --# expr-start 6903 '<And at ast.py:12041>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12041
--# expr-start 6900 '<Bind at ast.py:12041>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12041
--# expr-start 6898 '<FieldAccess for expected_type_var at ast.py:12041>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12041
--# expr-start 6897 '<FieldAccess for expr at ast.py:12041>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12041
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Self.Membership_Expr_F_Expr;
--# expr-done 6897
      if Fld_14 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_14);
--# expr-done 6898
Fld_15.Value := No_Entity;
Entity_Vars.Reset (Fld_15);
--# expr-start 6899 '<FieldAccess for base_subtype at ast.py:12042>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12042
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.base_subtype
Fld_16 := Libadalang.Implementation.Base_Type_Decl_P_Base_Subtype (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 6899
      Cast_Result_4 := Create_Internal_Entity
        (Node => Fld_16.Node,
         Info => Fld_16.Info);
Bind_Result_2 := Solver.Create_Assign (Fld_15, Cast_Result_4, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12041") else null)); 
--# expr-done 6900
--# expr-start 6902 '<FieldAccess for matches_expected_formal_type at ast.py:12043>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12043
--# expr-start 6901 '<FieldAccess for expr at ast.py:12043>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12043
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_17 := Self.Membership_Expr_F_Expr;
--# expr-done 6901
      if Fld_17 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_18 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_17, Origin => Origin);
--# end
--# expr-done 6902
And_Pred_2 := Create_And (Bind_Result_2, Fld_18, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12041") else null)); 
--# expr-done 6903
   If_Result_1 := And_Pred_2;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 6904
And_Pred_3 := Create_And (Fld_8, If_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12025") else null)); 
--# expr-done 6905
Scope_Result_1 := And_Pred_3;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2660;
   Result_Var_1 := Scope_Result_1;
else
   --# expr-start 6913 '<And at ast.py:12047>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12047
--# expr-start 6911 '<And at ast.py:12047>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12047
--# expr-start 6909 '<Bind at ast.py:12048>' Bind_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12048
--# expr-start 6906 '<FieldAccess for expected_type_var at ast.py:12048>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12048
      if Item_233.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Item_233.Node);
--# expr-done 6906
Fld_19.Value := No_Entity;
Entity_Vars.Reset (Fld_19);
--# expr-start 6908 '<FieldAccess for type_var at ast.py:12048>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12048
--# expr-start 6907 '<FieldAccess for expr at ast.py:12048>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12048
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_20 := Self.Membership_Expr_F_Expr;
--# expr-done 6907
      if Fld_20 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_21 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_20);
--# expr-done 6908
Fld_21.Value := No_Entity;
Entity_Vars.Reset (Fld_21);
Bind_Result_3 := Solver.Create_Unify (Fld_19, Fld_21, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12048") else null)); 
--# expr-done 6909
--# expr-start 6910 '<FieldAccess for sub_equation at ast.py:12049>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12049
      if Item_233.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_22 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Item_233.Node, Env => Env, Origin => Origin, E_Info => Item_233.Info);
--# end
--# expr-done 6910
And_Pred_4 := Create_And (Bind_Result_3, Fld_22, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12047") else null)); 
--# expr-done 6911
--# expr-start 6912 '<FieldAccess for matches_expected_type at ast.py:12050>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12050
      if Item_233.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_23 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Item_233.Node, Origin => Origin);
--# end
--# expr-done 6912
And_Pred_5 := Create_And (And_Pred_4, Fld_23, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12047") else null)); 
--# expr-done 6913
   Result_Var_1 := And_Pred_5;
end if;
      Inc_Ref (Result_Var_1);
         declare
            Item_To_Append : constant Logic_Equation := Result_Var_1;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2658;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 6914
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12021") else null)); 
--# expr-done 6915
And_Pred_6 := Create_And (And_Pred, Logic_Boolean_Op, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12019") else null)); 
--# expr-done 6916
      Property_Result := And_Pred_6;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_917;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_917;
               Finalizer_Scope_2658;
               Finalizer_Scope_2660;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Membership_Expr_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Name
   --
      procedure Initialize_Fields_For_Name
        (Self : Bare_Name
        ) is
      begin
            Initialize_Fields_For_Expr
              (Self);
      end Initialize_Fields_For_Name;
--# property-start Name.enclosing_defining_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12367
pragma Warnings (Off, "is not referenced");
 function Name_P_Enclosing_Defining_Name
  (Node : Bare_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_930 with Inline_Always;
      Fld : Internal_Entity_Array_Access;
Is_A : Boolean;
Find_Result : Internal_Entity;
Cast_Result : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_930 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_930;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 6920 '<Cast to DefiningName at ast.py:12368>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12368
--# expr-start 6919 '<Find at ast.py:12368>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12368
   --# expr-start 6917 '<FieldAccess for parents at ast.py:12368>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12368
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parents (Node => Ent.Node, With_Self => True, E_Info => Ent.Info);
--# expr-done 6917
      Find_Result := No_Entity;
      declare
         Collection : constant Internal_Entity_Array_Access := Fld;
      begin
         for Item_234 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_234
      --# expr-start 6918 '<IsA DefiningName at ast.py:12368>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12368
Is_A := Item_234.Node /= null 
and then Item_234.Node.Kind in Ada_Defining_Name_Range; 
--# expr-done 6918
      if Is_A then
         Find_Result := Item_234;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 6919
   if Find_Result.Node = null
      or else Find_Result.Node.Kind in Ada_Defining_Name_Range
   then
      Cast_Result := Create_Internal_Entity_Defining_Name
        (Node => Find_Result.Node,
         Info => Find_Result.Info);
   else
         Cast_Result := No_Entity_Defining_Name;
   end if;
--# expr-done 6920
      Property_Result := Cast_Result;
   --# end
      Finalizer_Scope_930;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_930;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Enclosing_Defining_Name;
--# end
--# property-start Name.is_defining /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12377
pragma Warnings (Off, "is not referenced");
 function Name_P_Is_Defining
  (Node : Bare_Name
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Var_Expr : Bare_Dotted_Name;
Is_A : Boolean;
Fld : Bare_Ada_Node;
Is_A_1 : Boolean;
If_Result : Boolean;
Fld_1 : Bare_Ada_Node;
Fld_2 : Bare_Ada_Node;
Is_A_2 : Boolean;
Fld_3 : Bare_Ada_Node;
Cast_Result : Bare_Dotted_Name;
Fld_4 : Bare_Base_Id;
Scope_Result : Bare_Base_Id;
Result_Var : Bare_Base_Id;
Cast_Result_1 : Bare_Name;
Is_Equal : Boolean;
If_Result_1 : Boolean;
If_Result_2 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 6931 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6921 '<IsA DefiningName at ast.py:12383>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12383
Is_A := Self /= null 
and then Self.Kind in Ada_Defining_Name_Range; 
--# expr-done 6921
if Is_A then
   If_Result := True;
else
   --# expr-start 6923 '<IsA DefiningName at ast.py:12387>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12387
--# expr-start 6922 '<parent at ast.py:12387>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12387
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 6922
Is_A_1 := Fld /= null 
and then Fld.Kind in Ada_Defining_Name_Range; 
--# expr-done 6923
   If_Result := Is_A_1;
end if;
if If_Result then
   If_Result_2 := True;
else
--# expr-start 6926 '<IsA DefiningName at ast.py:12391>' Is_A_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12391
--# expr-start 6925 '<parent at ast.py:12391>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12391
--# expr-start 6924 '<parent at ast.py:12391>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12391
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Self.Parent; 
--# expr-done 6924
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Fld_1.Parent; 
--# expr-done 6925
Is_A_2 := Fld_2 /= null 
and then Fld_2.Kind in Ada_Defining_Name_Range; 
--# expr-done 6926
if Is_A_2 then
   --# expr-start 6930 '<Eq at ast.py:12392>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12392
--# expr-start 6928 '<Cast to DottedName at ast.py:12392>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12392
--# expr-start 6927 '<parent at ast.py:12392>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12392
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Self.Parent; 
--# expr-done 6927
   if Fld_3 = null
      or else Fld_3.Kind in Ada_Dotted_Name_Range
   then
      Cast_Result := Fld_3;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 6928
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 6929 '<FieldAccess for suffix at ast.py:12392>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12392
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Var_Expr.Dotted_Name_F_Suffix;
--# expr-done 6929
Scope_Result := Fld_4;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node;
end if;
      Cast_Result_1 := Result_Var;
Is_Equal := Cast_Result_1 = Self; 
--# expr-done 6930
   If_Result_1 := Is_Equal;
else
   If_Result_1 := False;
end if;
   If_Result_2 := If_Result_1;
end if;
Let_Result := If_Result_2; 
--# end
--# expr-done 6931
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Is_Defining;
--# end
--# property-start '[dispatcher]Name.parent_scope' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Name_P_Parent_Scope
  (Node : Bare_Name
      ; Env : Lexical_Env
  )
   return Lexical_Env
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   Property_Result : Lexical_Env;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Name (Self.Kind) is
               when Ada_Attribute_Ref .. Ada_Call_Expr | Ada_Discrete_Subtype_Name | Ada_Explicit_Deref .. Ada_Reduce_Attribute_Ref | Ada_Null_Literal .. Ada_Target_Name =>
                  --# property-call-start Name.parent_scope
                  Property_Result := Name_P_Parent_Scope
                    (Self
                        , Env
                    );
                  --# end
               when Ada_Defining_Name .. Ada_Synthetic_Defining_Name =>
                  --# property-call-start DefiningName.parent_scope
                  Property_Result := Defining_Name_P_Parent_Scope
                    (Self
                        , Env
                    );
                  --# end
               when Ada_Dotted_Name =>
                  --# property-call-start DottedName.parent_scope
                  Property_Result := Dotted_Name_P_Parent_Scope
                    (Self
                        , Env
                    );
                  --# end
               when Ada_End_Name =>
                  --# property-call-start EndName.parent_scope
                  Property_Result := End_Name_P_Parent_Scope
                    (Self
                        , Env
                    );
                  --# end
               when Ada_Char_Literal .. Ada_String_Literal =>
                  --# property-call-start BaseId.parent_scope
                  Property_Result := Base_Id_P_Parent_Scope
                    (Self
                        , Env
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Name_P_Parent_Scope;
--# end
--# property-start Name.name_is /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12405
pragma Warnings (Off, "is not referenced");
 function Name_P_Name_Is
  (Node : Bare_Name
      ; Sym : Symbol_Type
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind sym Sym
   Property_Result : Boolean;
      Var_Expr : Symbol_Type;
Fld : Symbol_Type;
Is_Equal : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 6934 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6932 '<FieldAccess for name_symbol at ast.py:12409>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12409
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Self);
--# end
--# expr-done 6932
Var_Expr := Fld;
if Var_Expr /= null then
   --# scope-start
   --# bind ns Var_Expr
--# expr-start 6933 '<Eq at ast.py:12409>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12409
Is_Equal := Var_Expr = Sym; 
--# expr-done 6933
Scope_Result := Is_Equal;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 6934
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Name_Is;
--# end
--# property-start Name.is_direct_call /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12412
pragma Warnings (Off, "is not referenced");
 function Name_P_Is_Direct_Call
  (Node : Bare_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Boolean;
Fld_1 : Internal_Entity_Base_Formal_Param_Holder;
Fld_2 : Internal_Entity;
Is_A : Boolean;
Not_Val : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6940 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6935 '<FieldAccess for is_call at ast.py:12418>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12418
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_call
Fld := Libadalang.Implementation.Name_P_Is_Call (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 6935
if Fld then
   --# expr-start 6939 '<Not at ast.py:12419>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12419
--# expr-start 6938 '<IsA AccessToSubpDef at ast.py:12419>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12419
--# expr-start 6937 '<parent at ast.py:12419>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12419
--# expr-start 6936 '<FieldAccess for called_subp_spec at ast.py:12419>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12419
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.called_subp_spec
Fld_1 := Libadalang.Implementation.Name_P_Called_Subp_Spec (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 6936
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Parent (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# expr-done 6937
Is_A := Fld_2.Node /= null 
and then Fld_2.Node.Kind in Ada_Access_To_Subp_Def_Range; 
--# expr-done 6938
Not_Val := not (Is_A); 
--# expr-done 6939
   If_Result := Not_Val;
else
   If_Result := False;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 6940
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Is_Direct_Call;
--# end
--# property-start Name.is_access_call /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12423
pragma Warnings (Off, "is not referenced");
 function Name_P_Is_Access_Call
  (Node : Bare_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Boolean;
Fld_1 : Internal_Entity_Base_Formal_Param_Holder;
Fld_2 : Internal_Entity;
Is_A : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6945 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6941 '<FieldAccess for is_call at ast.py:12429>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12429
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_call
Fld := Libadalang.Implementation.Name_P_Is_Call (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 6941
if Fld then
   --# expr-start 6944 '<IsA AccessToSubpDef at ast.py:12430>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12430
--# expr-start 6943 '<parent at ast.py:12430>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12430
--# expr-start 6942 '<FieldAccess for called_subp_spec at ast.py:12430>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12430
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.called_subp_spec
Fld_1 := Libadalang.Implementation.Name_P_Called_Subp_Spec (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 6942
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Parent (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# expr-done 6943
Is_A := Fld_2.Node /= null 
and then Fld_2.Node.Kind in Ada_Access_To_Subp_Def_Range; 
--# expr-done 6944
   If_Result := Is_A;
else
   If_Result := False;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 6945
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Is_Access_Call;
--# end
--# property-start Name.is_call /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12434
pragma Warnings (Off, "is not referenced");
 function Name_P_Is_Call
  (Node : Bare_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Boolean;
Not_Val : Boolean;
Fld_1 : Internal_Entity_Base_Formal_Param_Holder;
Is_Null : Boolean;
Not_Val_1 : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6951 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6947 '<Not at ast.py:12439>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12439
--# expr-start 6946 '<FieldAccess for is_defining at ast.py:12439>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12439
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_defining
Fld := Libadalang.Implementation.Name_P_Is_Defining (Node => Ent.Node);
--# end
--# expr-done 6946
Not_Val := not (Fld); 
--# expr-done 6947
if Not_Val then
   --# expr-start 6950 '<Not at ast.py:12440>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12440
--# expr-start 6949 '<IsNull at ast.py:12440>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12440
--# expr-start 6948 '<FieldAccess for called_subp_spec at ast.py:12440>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12440
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.called_subp_spec
Fld_1 := Libadalang.Implementation.Name_P_Called_Subp_Spec (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 6948
Is_Null := Fld_1.Node = null; 
--# expr-done 6949
Not_Val_1 := not (Is_Null); 
--# expr-done 6950
   If_Result := Not_Val_1;
else
   If_Result := False;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 6951
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Is_Call;
--# end
--# property-start Name.is_dot_call /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12445
pragma Warnings (Off, "is not referenced");
 function Name_P_Is_Dot_Call
  (Node : Bare_Name
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Fld : Boolean;
Not_Val : Boolean;
Fld_1 : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity_Info;
Fld_3 : Internal_Metadata;
Fld_4 : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6958 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6953 '<Not at ast.py:12450>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12450
--# expr-start 6952 '<FieldAccess for is_defining at ast.py:12450>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12450
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_defining
Fld := Libadalang.Implementation.Name_P_Is_Defining (Node => Ent.Node);
--# end
--# expr-done 6952
Not_Val := not (Fld); 
--# expr-done 6953
if Not_Val then
   --# expr-start 6957 '<FieldAccess for dottable_subp at ast.py:12451>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12451
--# expr-start 6956 '<FieldAccess for md at ast.py:12451>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12451
--# expr-start 6955 '<FieldAccess for info at ast.py:12451>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12451
--# expr-start 6954 '<FieldAccess for referenced_decl at ast.py:12451>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12451
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld_1 := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 6954
Fld_2 := Fld_1.Info;
--# expr-done 6955
Fld_3 := Fld_2.Md;
--# expr-done 6956
Fld_4 := Fld_3.Dottable_Subp;
--# expr-done 6957
   If_Result := Fld_4;
else
   If_Result := False;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 6958
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Is_Dot_Call;
--# end
--# property-start Name.failsafe_referenced_def_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12456
pragma Warnings (Off, "is not referenced");
 function Name_P_Failsafe_Referenced_Def_Name
  (Node : Bare_Name
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Refd_Def
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Refd_Def;
            procedure Finalizer_Scope_2664 with Inline_Always;
      Block_Var : Internal_Refd_Decl;
Block_Var_1 : Internal_Entity_Defining_Name;
Var_Expr : Internal_Entity_Basic_Decl;
Var_Expr_1 : Internal_Entity_Defining_Name;
Var_Expr_2 : Symbol_Type;
Fld : Internal_Refd_Decl;
Fld_1 : Internal_Entity_Basic_Decl;
Fld_2 : Symbol_Type;
Fld_3 : Internal_Entity_Defining_Name_Array_Access;
Fld_4 : Boolean;
Find_Result : Internal_Entity_Defining_Name;
Scope_Result : Internal_Entity_Defining_Name;
Result_Var : Internal_Entity_Defining_Name;
Scope_Result_1 : Internal_Entity_Defining_Name;
Fld_5 : Internal_Entity_Defining_Name;
Result_Var_1 : Internal_Entity_Defining_Name;
Scope_Result_2 : Internal_Entity_Defining_Name;
Result_Var_2 : Internal_Entity_Defining_Name;
Fld_6 : Ref_Result_Kind;
New_Struct : Internal_Refd_Def;
Let_Result : Internal_Refd_Def;
            procedure Finalizer_Scope_2664 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_2664;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6968 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6959 '<FieldAccess for failsafe_referenced_decl at ast.py:12461>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12461
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.failsafe_referenced_decl
Fld := Libadalang.Implementation.Name_P_Failsafe_Referenced_Decl (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 6959
Block_Var := Fld; 
--# bind ref_decl Block_Var
--# expr-start 6960 '<FieldAccess for decl at ast.py:12463>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12463
Fld_1 := Block_Var.Decl;
--# expr-done 6960
Var_Expr := Fld_1;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind ref_decl Var_Expr
--# expr-start 6961 '<FieldAccess for name_symbol at ast.py:12464>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12464
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 6961
Var_Expr_2 := Fld_2;
if Var_Expr_2 /= null then
   --# scope-start
   --# bind rel_name Var_Expr_2
--# expr-start 6964 '<Find at ast.py:12465>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12465
   --# expr-start 6962 '<FieldAccess for defining_names at ast.py:12465>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12465
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_names'
Fld_3 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Names (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 6962
      Find_Result := No_Entity_Defining_Name;
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Fld_3;
      begin
         for Item_235 of
               Collection.Items
         loop
   --# scope-start
               --# bind dn Item_235
      --# expr-start 6963 '<FieldAccess for name_is at ast.py:12466>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12466
      if Item_235.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_is
Fld_4 := Libadalang.Implementation.Name_P_Name_Is (Node => Item_235.Node, Sym => Var_Expr_2);
--# end
--# expr-done 6963
      if Fld_4 then
         Find_Result := Item_235;
      end if;
   --# end
               exit when Fld_4;
         end loop;
      end;
--# expr-done 6964
Scope_Result := Find_Result;
   --# end
      Finalizer_Scope_2664;
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Defining_Name;
end if;
Var_Expr_1 := Result_Var;
if Var_Expr_1 /= No_Entity_Defining_Name then
   --# scope-start
   --# bind e Var_Expr_1
Scope_Result_1 := Var_Expr_1;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   --# expr-start 6965 '<FieldAccess for defining_name at ast.py:12468>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12468
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_5 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 6965
   Result_Var_1 := Fld_5;
end if;
Scope_Result_2 := Result_Var_1;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Entity_Defining_Name;
end if;
Block_Var_1 := Result_Var_2; 
--# bind def_name Block_Var_1
--# expr-start 6967 '<New RefdDef at ast.py:12472>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12472
--# expr-start 6966 '<FieldAccess for kind at ast.py:12472>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12472
Fld_6 := Block_Var.Kind;
--# expr-done 6966
New_Struct := (Def_Name => Block_Var_1, Kind => Fld_6); 
--# expr-done 6967
Let_Result := New_Struct; 
--# end
--# expr-done 6968
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_2664;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Failsafe_Referenced_Def_Name;
--# end
--# property-start Name.referenced_defining_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12476
pragma Warnings (Off, "is not referenced");
 function Name_P_Referenced_Defining_Name
  (Node : Bare_Name
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_2668 with Inline_Always;
      Block_Var : Internal_Entity_Basic_Decl;
Var_Expr : Internal_Entity_Basic_Decl;
Var_Expr_1 : Internal_Entity_Defining_Name;
Var_Expr_2 : Symbol_Type;
Fld : Internal_Entity_Basic_Decl;
Fld_1 : Symbol_Type;
Fld_2 : Internal_Entity_Defining_Name_Array_Access;
Fld_3 : Boolean;
Find_Result : Internal_Entity_Defining_Name;
Scope_Result : Internal_Entity_Defining_Name;
Result_Var : Internal_Entity_Defining_Name;
Scope_Result_1 : Internal_Entity_Defining_Name;
Fld_4 : Internal_Entity_Defining_Name;
Result_Var_1 : Internal_Entity_Defining_Name;
Scope_Result_2 : Internal_Entity_Defining_Name;
Result_Var_2 : Internal_Entity_Defining_Name;
Let_Result : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_2668 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_2668;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 6975 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6969 '<FieldAccess for referenced_decl at ast.py:12481>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12481
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 6969
Block_Var := Fld; 
--# bind ref_decl Block_Var
Var_Expr := Block_Var;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind ref_decl Var_Expr
--# expr-start 6970 '<FieldAccess for name_symbol at ast.py:12482>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12482
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 6970
Var_Expr_2 := Fld_1;
if Var_Expr_2 /= null then
   --# scope-start
   --# bind rel_name Var_Expr_2
--# expr-start 6973 '<Find at ast.py:12483>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12483
   --# expr-start 6971 '<FieldAccess for defining_names at ast.py:12483>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12483
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_names'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Names (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 6971
      Find_Result := No_Entity_Defining_Name;
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Fld_2;
      begin
         for Item_236 of
               Collection.Items
         loop
   --# scope-start
               --# bind dn Item_236
      --# expr-start 6972 '<FieldAccess for name_is at ast.py:12484>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12484
      if Item_236.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_is
Fld_3 := Libadalang.Implementation.Name_P_Name_Is (Node => Item_236.Node, Sym => Var_Expr_2);
--# end
--# expr-done 6972
      if Fld_3 then
         Find_Result := Item_236;
      end if;
   --# end
               exit when Fld_3;
         end loop;
      end;
--# expr-done 6973
Scope_Result := Find_Result;
   --# end
      Finalizer_Scope_2668;
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Defining_Name;
end if;
Var_Expr_1 := Result_Var;
if Var_Expr_1 /= No_Entity_Defining_Name then
   --# scope-start
   --# bind e Var_Expr_1
Scope_Result_1 := Var_Expr_1;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   --# expr-start 6974 '<FieldAccess for defining_name at ast.py:12486>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12486
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 6974
   Result_Var_1 := Fld_4;
end if;
Scope_Result_2 := Result_Var_1;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Entity_Defining_Name;
end if;
Let_Result := Result_Var_2; 
--# end
--# expr-done 6975
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_2668;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Referenced_Defining_Name;
--# end
--# property-start Name.gnat_xref_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12490
pragma Warnings (Off, "is not referenced");
 function Name_P_Gnat_Xref_Decl
  (Node : Bare_Name
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Defining_Name;
      Block_Var : Internal_Entity_Defining_Name;
Block_Var_1 : Internal_Entity_Basic_Decl;
Var_Expr : Boolean;
Var_Expr_1 : Internal_Entity_Defining_Name;
Var_Expr_2 : Internal_Entity_Basic_Decl;
Var_Expr_3 : Internal_Entity_Basic_Decl;
Var_Expr_4 : Internal_Entity_Basic_Decl;
Var_Expr_5 : Internal_Entity_Basic_Decl;
Var_Expr_6 : Internal_Entity_Base_Type_Decl;
Var_Expr_7 : Internal_Entity_Basic_Decl;
Var_Expr_8 : Internal_Entity_Basic_Decl;
Var_Expr_9 : Internal_Entity_Basic_Decl;
Var_Expr_10 : Internal_Entity_Defining_Name;
Var_Expr_11 : Internal_Entity_Body_Node;
Fld : Boolean;
Fld_1 : Internal_Entity_Defining_Name;
Scope_Result : Internal_Entity_Defining_Name;
Result_Var : Internal_Entity_Defining_Name;
Fld_2 : Internal_Entity_Basic_Decl;
Scope_Result_1 : Internal_Entity_Basic_Decl;
Result_Var_1 : Internal_Entity_Basic_Decl;
Is_A : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
Cast_Expr : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity_Param_Spec;
Fld_3 : Internal_Entity_Defining_Name;
Is_A_1 : Boolean;
Scope_Result_3 : Boolean;
Result_Var_3 : Boolean;
Cast_Expr_1 : Internal_Entity_Basic_Decl;
Cast_Result_1 : Internal_Entity_Body_Node;
Fld_4 : Internal_Entity_Basic_Decl;
Fld_5 : Internal_Entity_Defining_Name;
Scope_Result_4 : Internal_Entity_Defining_Name;
Result_Var_4 : Internal_Entity_Defining_Name;
Is_A_2 : Boolean;
Cast_Expr_2 : Internal_Entity_Basic_Decl;
Cast_Result_2 : Internal_Entity_Base_Type_Decl;
Fld_6 : Boolean;
If_Result : Boolean;
Scope_Result_5 : Boolean;
Result_Var_5 : Boolean;
Cast_Expr_3 : Internal_Entity_Basic_Decl;
Cast_Result_3 : Internal_Entity_Base_Type_Decl;
Fld_7 : Internal_Entity_Base_Type_Decl;
Fld_8 : Internal_Entity_Defining_Name;
Scope_Result_6 : Internal_Entity_Defining_Name;
Result_Var_6 : Internal_Entity_Defining_Name;
Is_A_3 : Boolean;
Scope_Result_7 : Boolean;
Result_Var_7 : Boolean;
Is_A_4 : Boolean;
Scope_Result_8 : Boolean;
Result_Var_8 : Boolean;
Is_A_5 : Boolean;
Scope_Result_9 : Boolean;
Result_Var_9 : Boolean;
Fld_9 : Internal_Entity_Body_Node;
Fld_10 : Internal_Entity_Defining_Name;
Scope_Result_10 : Internal_Entity_Defining_Name;
Result_Var_10 : Internal_Entity_Defining_Name;
Scope_Result_11 : Internal_Entity_Defining_Name;
Fld_11 : Internal_Entity_Defining_Name;
Result_Var_11 : Internal_Entity_Defining_Name;
Fld_12 : Internal_Entity_Defining_Name;
If_Result_1 : Internal_Entity_Defining_Name;
If_Result_2 : Internal_Entity_Defining_Name;
If_Result_3 : Internal_Entity_Defining_Name;
If_Result_4 : Internal_Entity_Defining_Name;
If_Result_5 : Internal_Entity_Defining_Name;
If_Result_6 : Internal_Entity_Defining_Name;
Let_Result : Internal_Entity_Defining_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7000 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 6976 '<FieldAccess for is_defining at ast.py:12495>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12495
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_defining
Fld := Libadalang.Implementation.Name_P_Is_Defining (Node => Ent.Node);
--# end
--# expr-done 6976
Var_Expr := Fld;
if Var_Expr /= False then
   --# scope-start
   --# bind _ Var_Expr
--# expr-start 6977 '<FieldAccess for enclosing_defining_name at ast.py:12496>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12496
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.enclosing_defining_name
Fld_1 := Libadalang.Implementation.Name_P_Enclosing_Defining_Name (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 6977
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Defining_Name;
end if;
Block_Var := Result_Var; 
--# bind dn Block_Var
Var_Expr_1 := Block_Var;
if Var_Expr_1 /= No_Entity_Defining_Name then
   --# scope-start
   --# bind dn Var_Expr_1
--# expr-start 6978 '<FieldAccess for basic_decl at ast.py:12498>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12498
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld_2 := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 6978
Scope_Result_1 := Fld_2;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Basic_Decl;
end if;
Block_Var_1 := Result_Var_1; 
--# bind bd Block_Var_1
--# expr-start 6999 '<Cond at ast.py:12500>' If_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12500
Var_Expr_2 := Block_Var_1;
if Var_Expr_2 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind bd Var_Expr_2
--# expr-start 6979 '<IsA ParamSpec at ast.py:12501>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12501
Is_A := Var_Expr_2.Node /= null 
and then Var_Expr_2.Node.Kind in Ada_Param_Spec_Range; 
--# expr-done 6979
Scope_Result_2 := Is_A;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := False;
end if;
if Result_Var_2 then
   --# expr-start 6981 '<FieldAccess for decl_param at ast.py:12502>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12502
--# expr-start 6980 '<Cast to ParamSpec at ast.py:12502>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12502
Cast_Expr := Block_Var_1; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Param_Spec_Range
   then
      Cast_Result := Create_Internal_Entity_Param_Spec
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Param_Spec;
   end if;
--# expr-done 6980
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamDecl.decl_param
Fld_3 := Libadalang.Implementation.Base_Formal_Param_Decl_P_Decl_Param (Node => Cast_Result.Node, Param => Block_Var, Imprecise_Fallback => Imprecise_Fallback, E_Info => Cast_Result.Info);
--# end
--# expr-done 6981
   If_Result_6 := Fld_3;
else
Var_Expr_3 := Block_Var_1;
if Var_Expr_3 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind bd Var_Expr_3
--# expr-start 6982 '<IsA Body at ast.py:12504>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12504
Is_A_1 := Var_Expr_3.Node /= null 
and then Var_Expr_3.Node.Kind in Ada_Body_Node; 
--# expr-done 6982
Scope_Result_3 := Is_A_1;
   --# end
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := False;
end if;
if Result_Var_3 then
--# expr-start 6984 '<FieldAccess for decl_part at ast.py:12505>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12505
--# expr-start 6983 '<Cast to Body at ast.py:12505>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12505
Cast_Expr_1 := Block_Var_1; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Body_Node
   then
      Cast_Result_1 := Create_Internal_Entity_Body_Node
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_1 := No_Entity_Body_Node;
   end if;
--# expr-done 6983
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.decl_part
Fld_4 := Libadalang.Implementation.Body_Node_P_Decl_Part (Node => Cast_Result_1.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Cast_Result_1.Info);
--# end
--# expr-done 6984
Var_Expr_4 := Fld_4;
if Var_Expr_4 /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 6985 '<FieldAccess for defining_name at ast.py:12505>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12505
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_5 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Var_Expr_4.Node, E_Info => Var_Expr_4.Info);
--# end
--# expr-done 6985
Scope_Result_4 := Fld_5;
   --# end
   Result_Var_4 := Scope_Result_4;
else
   Result_Var_4 := No_Entity_Defining_Name;
end if;
   If_Result_5 := Result_Var_4;
else
Var_Expr_5 := Block_Var_1;
if Var_Expr_5 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind bd Var_Expr_5
--# expr-start 6986 '<IsA BaseTypeDecl at ast.py:12507>' Is_A_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12507
Is_A_2 := Var_Expr_5.Node /= null 
and then Var_Expr_5.Node.Kind in Ada_Base_Type_Decl; 
--# expr-done 6986
if Is_A_2 then
   --# expr-start 6988 '<FieldAccess for is_in_private_part at ast.py:12508>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12508
--# expr-start 6987 '<Cast to BaseTypeDecl at ast.py:12508>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12508
Cast_Expr_2 := Var_Expr_5; 
   if Cast_Expr_2.Node = null
      or else Cast_Expr_2.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_2 := Create_Internal_Entity_Base_Type_Decl
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
   else
         Cast_Result_2 := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 6987
      if Cast_Result_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_in_private_part'
Fld_6 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_In_Private_Part (Node => Cast_Result_2.Node, E_Info => Cast_Result_2.Info);
--# end
--# expr-done 6988
   If_Result := Fld_6;
else
   If_Result := False;
end if;
Scope_Result_5 := If_Result;
   --# end
   Result_Var_5 := Scope_Result_5;
else
   Result_Var_5 := False;
end if;
if Result_Var_5 then
--# expr-start 6990 '<FieldAccess for previous_part at ast.py:12509>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12509
--# expr-start 6989 '<Cast to BaseTypeDecl at ast.py:12509>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12509
Cast_Expr_3 := Block_Var_1; 
   if Cast_Expr_3.Node = null
      or else Cast_Expr_3.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_3 := Create_Internal_Entity_Base_Type_Decl
        (Node => Cast_Expr_3.Node,
         Info => Cast_Expr_3.Info);
   else
         Cast_Result_3 := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 6989
      if Cast_Result_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.previous_part'
Fld_7 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Previous_Part (Node => Cast_Result_3.Node, Go_To_Incomplete => True, E_Info => Cast_Result_3.Info);
--# end
--# expr-done 6990
Var_Expr_6 := Fld_7;
if Var_Expr_6 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 6991 '<FieldAccess for defining_name at ast.py:12509>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12509
      if Var_Expr_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_8 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Var_Expr_6.Node, E_Info => Var_Expr_6.Info);
--# end
--# expr-done 6991
Scope_Result_6 := Fld_8;
   --# end
   Result_Var_6 := Scope_Result_6;
else
   Result_Var_6 := No_Entity_Defining_Name;
end if;
   If_Result_4 := Result_Var_6;
else
Var_Expr_7 := Block_Var_1;
if Var_Expr_7 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind bd Var_Expr_7
--# expr-start 6992 '<IsA ObjectDecl at ast.py:12511>' Is_A_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12511
Is_A_3 := Var_Expr_7.Node /= null 
and then Var_Expr_7.Node.Kind in Ada_Object_Decl_Range; 
--# expr-done 6992
Scope_Result_7 := Is_A_3;
   --# end
   Result_Var_7 := Scope_Result_7;
else
   Result_Var_7 := False;
end if;
if Result_Var_7 then
   If_Result_3 := No_Entity_Defining_Name;
else
Var_Expr_8 := Block_Var_1;
if Var_Expr_8 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind bd Var_Expr_8
--# expr-start 6993 '<IsA NumberDecl at ast.py:12516>' Is_A_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12516
Is_A_4 := Var_Expr_8.Node /= null 
and then Var_Expr_8.Node.Kind in Ada_Number_Decl_Range; 
--# expr-done 6993
Scope_Result_8 := Is_A_4;
   --# end
   Result_Var_8 := Scope_Result_8;
else
   Result_Var_8 := False;
end if;
if Result_Var_8 then
   If_Result_2 := No_Entity_Defining_Name;
else
Var_Expr_9 := Block_Var_1;
if Var_Expr_9 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind bd Var_Expr_9
--# expr-start 6994 '<IsA BasicDecl at ast.py:12520>' Is_A_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12520
Is_A_5 := Var_Expr_9.Node /= null 
and then Var_Expr_9.Node.Kind in Ada_Basic_Decl; 
--# expr-done 6994
Scope_Result_9 := Is_A_5;
   --# end
   Result_Var_9 := Scope_Result_9;
else
   Result_Var_9 := False;
end if;
if Result_Var_9 then
--# expr-start 6995 '<FieldAccess for body_part_for_decl at ast.py:12521>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12521
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.body_part_for_decl
Fld_9 := Libadalang.Implementation.Basic_Decl_P_Body_Part_For_Decl (Node => Block_Var_1.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Block_Var_1.Info);
--# end
--# expr-done 6995
Var_Expr_11 := Fld_9;
if Var_Expr_11 /= No_Entity_Body_Node then
   --# scope-start
   --# bind bpe Var_Expr_11
--# expr-start 6996 '<FieldAccess for defining_name at ast.py:12521>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12521
      if Var_Expr_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_10 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Var_Expr_11.Node, E_Info => Var_Expr_11.Info);
--# end
--# expr-done 6996
Scope_Result_10 := Fld_10;
   --# end
   Result_Var_10 := Scope_Result_10;
else
   Result_Var_10 := No_Entity_Defining_Name;
end if;
Var_Expr_10 := Result_Var_10;
if Var_Expr_10 /= No_Entity_Defining_Name then
   --# scope-start
   --# bind e Var_Expr_10
Scope_Result_11 := Var_Expr_10;
   --# end
   Result_Var_11 := Scope_Result_11;
else
   --# expr-start 6997 '<FieldAccess for defining_name at ast.py:12522>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12522
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_11 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Block_Var_1.Node, E_Info => Block_Var_1.Info);
--# end
--# expr-done 6997
   Result_Var_11 := Fld_11;
end if;
   If_Result_1 := Result_Var_11;
else
   --# expr-start 6998 '<FieldAccess for referenced_defining_name at ast.py:12524>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12524
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_defining_name
Fld_12 := Libadalang.Implementation.Name_P_Referenced_Defining_Name (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 6998
   If_Result_1 := Fld_12;
end if;
   If_Result_2 := If_Result_1;
end if;
   If_Result_3 := If_Result_2;
end if;
   If_Result_4 := If_Result_3;
end if;
   If_Result_5 := If_Result_4;
end if;
   If_Result_6 := If_Result_5;
end if;
--# expr-done 6999
Let_Result := If_Result_6; 
--# end
--# expr-done 7000
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Gnat_Xref_Decl;
--# end
--# property-start '[dispatcher]Name.all_env_els_impl' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Name_P_All_Env_Els_Impl
  (Node : Bare_Name
      ; Seq : Boolean
         := True
      ; Seq_From : Bare_Ada_Node
         := No_Bare_Ada_Node
      ; Categories : Ref_Categories
         := (Inherited_Primitives => True, Nocat => True)
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind seq Seq
   --# bind seq_from Seq_From
   --# bind categories Categories
   --# bind env Env
   --# bind origin Origin
   Property_Result : Internal_Entity_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Name (Self.Kind) is
               when Ada_Attribute_Ref .. Ada_Call_Expr | Ada_Discrete_Subtype_Name | Ada_End_Name .. Ada_Reduce_Attribute_Ref | Ada_Null_Literal .. Ada_Target_Name =>
                  --# property-call-start Name.all_env_els_impl
                  Property_Result := Name_P_All_Env_Els_Impl
                    (Self
                        , Seq
                        , Seq_From
                        , Categories
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Defining_Name .. Ada_Synthetic_Defining_Name =>
                  --# property-call-start DefiningName.all_env_els_impl
                  Property_Result := Defining_Name_P_All_Env_Els_Impl
                    (Self
                        , Seq
                        , Seq_From
                        , Categories
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Dotted_Name =>
                  --# property-call-start DottedName.all_env_els_impl
                  Property_Result := Dotted_Name_P_All_Env_Els_Impl
                    (Self
                        , Seq
                        , Seq_From
                        , Categories
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Char_Literal .. Ada_String_Literal =>
                  --# property-call-start BaseId.all_env_els_impl
                  Property_Result := Base_Id_P_All_Env_Els_Impl
                    (Self
                        , Seq
                        , Seq_From
                        , Categories
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Name_P_All_Env_Els_Impl;
--# end
--# property-start Name.all_env_elements_internal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12538
pragma Warnings (Off, "is not referenced");
 function Name_P_All_Env_Elements_Internal
  (Node : Bare_Name
      ; Seq : Boolean
         := True
      ; Seq_From : Bare_Ada_Node
         := No_Bare_Ada_Node
      ; Categories : Ref_Categories
         := (Inherited_Primitives => True, Nocat => True)
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind seq Seq
   --# bind seq_from Seq_From
   --# bind categories Categories
   Property_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_942 with Inline_Always;
      Fld : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_1 : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_2 : Internal_Entity_Array_Access;
Dyn_Var_Bind_Result : Internal_Entity_Array_Access;
Dyn_Var_Bind_Result_1 : Internal_Entity_Array_Access;
Let_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_942 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Dyn_Var_Bind_Result_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_942;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7006 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7005 '<bind at ast.py:12547>' Dyn_Var_Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12547
--# expr-start 7001 '<FieldAccess for origin_node at ast.py:12548>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12548
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 7001
Bound_Origin := Fld; 
--# expr-start 7004 '<bind at ast.py:12549>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12549
--# expr-start 7002 '<FieldAccess for node_env at ast.py:12550>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12550
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Node_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 7002
Bound_Env := Fld_1; Inc_Ref (Bound_Env);
--# expr-start 7003 '<FieldAccess for all_env_els_impl at ast.py:12551>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12551
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.all_env_els_impl'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_All_Env_Els_Impl (Node => Ent.Node, Seq => Seq, Seq_From => Seq_From, Categories => Categories, Env => Bound_Env, Origin => Bound_Origin, E_Info => Ent.Info);
--# end
--# expr-done 7003
Dyn_Var_Bind_Result := Fld_2; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 7004
Dyn_Var_Bind_Result_1 := Dyn_Var_Bind_Result; Inc_Ref (Dyn_Var_Bind_Result_1);
--# expr-done 7005
Let_Result := Dyn_Var_Bind_Result_1; Inc_Ref (Let_Result);
--# end
--# expr-done 7006
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_942;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_942;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_All_Env_Elements_Internal;
--# end
--# property-start Name.all_env_elements /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12560
pragma Warnings (Off, "is not referenced");
 function Name_P_All_Env_Elements
  (Node : Bare_Name
      ; Seq : Boolean
         := True
      ; Seq_From : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind seq Seq
   --# bind seq_from Seq_From
   Property_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_943 with Inline_Always;
      Fld : Internal_Entity_Array_Access;
Let_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_943 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_943;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7008 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7007 '<FieldAccess for all_env_elements_internal at ast.py:12567>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12567
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.all_env_elements_internal
Fld := Libadalang.Implementation.Name_P_All_Env_Elements_Internal (Node => Ent.Node, Seq => Seq, Seq_From => Seq_From, E_Info => Ent.Info);
--# end
--# expr-done 7007
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 7008
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_943;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_943;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_All_Env_Elements;
--# end
--# property-start Name.first_corresponding_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12570
pragma Warnings (Off, "is not referenced");
 function Name_P_First_Corresponding_Decl
  (Node : Bare_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_944 with Inline_Always;
      Fld : Internal_Entity;
Cast_Result : Internal_Entity_Dotted_Name;
Fld_1 : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity_Array_Access;
Get_Result : Internal_Entity;
Cast_Result_1 : Internal_Entity_Basic_Decl;
Fld_3 : Bare_Ada_Node;
Is_A : Boolean;
Fld_4 : Boolean;
If_Result : Boolean;
If_Result_1 : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_944 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_944;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7019 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7018 '<If at ast.py:12571>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12571
--# expr-start 7010 '<IsA DottedName at ast.py:12572>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12572
--# expr-start 7009 '<parent at ast.py:12572>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12572
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Self.Parent; 
--# expr-done 7009
Is_A := Fld_3 /= null 
and then Fld_3.Kind in Ada_Dotted_Name_Range; 
--# expr-done 7010
if Is_A then
   --# expr-start 7011 '<FieldAccess for is_suffix at ast.py:12572>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12572
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_suffix
Fld_4 := Libadalang.Implementation.Name_P_Is_Suffix (Node => Self);
--# end
--# expr-done 7011
   If_Result := Fld_4;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 7014 '<FieldAccess for first_corresponding_decl at ast.py:12573>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12573
--# expr-start 7013 '<Cast to DottedName at ast.py:12573>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12573
--# expr-start 7012 '<parent at ast.py:12573>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12573
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 7012
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Dotted_Name_Range
   then
      Cast_Result := Create_Internal_Entity_Dotted_Name
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Dotted_Name;
   end if;
--# expr-done 7013
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.first_corresponding_decl'
Fld_1 := Libadalang.Implementation.Dispatcher_Expr_P_First_Corresponding_Decl (Node => Cast_Result.Node, E_Info => Cast_Result.Info);
--# end
--# expr-done 7014
   If_Result_1 := Fld_1;
else
   --# expr-start 7017 '<Cast to BasicDecl at ast.py:12574>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12574
--# expr-start 7016 '<at at ast.py:12574>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12574
--# expr-start 7015 '<FieldAccess for all_env_elements_internal at ast.py:12574>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12574
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.all_env_elements_internal
Fld_2 := Libadalang.Implementation.Name_P_All_Env_Elements_Internal (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 7015
Get_Result := Get (Fld_2, 0, True); 
--# expr-done 7016
   if Get_Result.Node = null
      or else Get_Result.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Basic_Decl
        (Node => Get_Result.Node,
         Info => Get_Result.Info);
   else
         Cast_Result_1 := No_Entity_Basic_Decl;
   end if;
--# expr-done 7017
   If_Result_1 := Cast_Result_1;
end if;
--# expr-done 7018
Let_Result := If_Result_1; 
--# end
--# expr-done 7019
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_944;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_944;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_First_Corresponding_Decl;
--# end
--# property-start Name.bottom_up_name_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12578
pragma Warnings (Off, "is not referenced");
 function Name_P_Bottom_Up_Name_Equation
  (Node : Bare_Name
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_945 with Inline_Always;
            procedure Finalizer_Scope_2683 with Inline_Always;
            procedure Finalizer_Scope_2684 with Inline_Always;
            procedure Finalizer_Scope_2685 with Inline_Always;
            procedure Finalizer_Scope_2686 with Inline_Always;
      Match_0 : Internal_Entity_Call_Expr;
Match_1 : Internal_Entity_Explicit_Deref;
Match_2 : Internal_Entity_Qual_Expr;
Match_3 : Internal_Entity_Name;
Fld : Bare_Name;
As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld_1 : Logic_Equation;
Fld_2 : Logic_Equation;
Fld_3 : Logic_Equation;
False_Rel : Logic_Equation;
Match_Prefix : Internal_Entity_Name;
Cast_Result : Internal_Entity_Call_Expr;
Let_Result : Logic_Equation;
Scope_Result : Logic_Equation;
Cast_Result_1 : Internal_Entity_Explicit_Deref;
Let_Result_1 : Logic_Equation;
Scope_Result_1 : Logic_Equation;
Cast_Result_2 : Internal_Entity_Qual_Expr;
Let_Result_2 : Logic_Equation;
Scope_Result_2 : Logic_Equation;
Cast_Result_3 : Internal_Entity_Name;
Let_Result_3 : Logic_Equation;
Scope_Result_3 : Logic_Equation;
Match_Result : Logic_Equation;
Let_Result_4 : Logic_Equation;
            procedure Finalizer_Scope_945 is
            begin
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Let_Result_2);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Let_Result_3);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Match_Result);
                     Dec_Ref (Let_Result_4);
            end Finalizer_Scope_945;
            procedure Finalizer_Scope_2683 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2683;
            procedure Finalizer_Scope_2684 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_2684;
            procedure Finalizer_Scope_2685 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_2685;
            procedure Finalizer_Scope_2686 is
            begin
                     Dec_Ref (False_Rel);
            end Finalizer_Scope_2686;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7026 '<Block at ???>' Let_Result_4 None
--# scope-start
--# expr-start 7025 '<Match at ast.py:12579>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12579
--# expr-start 7021 '<as_entity at ast.py:12579>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12579
--# expr-start 7020 '<FieldAccess for innermost_name at ast.py:12579>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12579
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.innermost_name
Fld := Libadalang.Implementation.Name_P_Innermost_Name (Node => Self);
--# end
--# expr-done 7020
Is_Null := Fld = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
As_Entity := (Info => E_Info, Node => Fld); 
   If_Result := As_Entity;
end if;
--# expr-done 7021
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := If_Result; 
case Ada_Name (Match_Prefix.Node.Kind) is
         when Ada_Call_Expr =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Call_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind ce Match_0
--# expr-start 7022 '<FieldAccess for general_xref_equation at ast.py:12580>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12580
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.general_xref_equation
Fld_1 := Libadalang.Implementation.Call_Expr_P_General_Xref_Equation (Node => Match_0.Node, Root => Self, Env => Env, Origin => Origin, E_Info => Match_0.Info);
--# end
--# expr-done 7022
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
Scope_Result := Let_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2683;
            Match_Result := Scope_Result; Inc_Ref (Match_Result);
         when Ada_Explicit_Deref =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Explicit_Deref
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind ed Match_1
--# expr-start 7023 '<FieldAccess for general_xref_equation at ast.py:12581>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12581
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ExplicitDeref.general_xref_equation
Fld_2 := Libadalang.Implementation.Explicit_Deref_P_General_Xref_Equation (Node => Match_1.Node, Root => Self, Env => Env, Origin => Origin, E_Info => Match_1.Info);
--# end
--# expr-done 7023
Let_Result_1 := Fld_2; Inc_Ref (Let_Result_1);
--# end
Scope_Result_1 := Let_Result_1;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2684;
            Match_Result := Scope_Result_1; Inc_Ref (Match_Result);
         when Ada_Qual_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Qual_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind qe Match_2
--# expr-start 7024 '<FieldAccess for general_xref_equation at ast.py:12582>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12582
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start QualExpr.general_xref_equation
Fld_3 := Libadalang.Implementation.Qual_Expr_P_General_Xref_Equation (Node => Match_2.Node, Root => Self, Env => Env, Origin => Origin, E_Info => Match_2.Info);
--# end
--# expr-done 7024
Let_Result_2 := Fld_3; Inc_Ref (Let_Result_2);
--# end
Scope_Result_2 := Let_Result_2;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2685;
            Match_Result := Scope_Result_2; Inc_Ref (Match_Result);
         when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref | Ada_Defining_Name .. Ada_End_Name | Ada_Reduce_Attribute_Ref .. Ada_Target_Name =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_3; 
--# bind _ Match_3
False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12583") else null)); 
Let_Result_3 := False_Rel; Inc_Ref (Let_Result_3);
--# end
Scope_Result_3 := Let_Result_3;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_2686;
            Match_Result := Scope_Result_3; Inc_Ref (Match_Result);
end case;
--# expr-done 7025
Let_Result_4 := Match_Result; Inc_Ref (Let_Result_4);
--# end
--# expr-done 7026
      Property_Result := Let_Result_4;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_945;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_945;
               Finalizer_Scope_2683;
               Finalizer_Scope_2684;
               Finalizer_Scope_2685;
               Finalizer_Scope_2686;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Bottom_Up_Name_Equation;
--# end
--# property-start Name.innermost_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12587
pragma Warnings (Off, "is not referenced");
 function Name_P_Innermost_Name
  (Node : Bare_Name
  )
   return Bare_Name
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Name;
      Block_Var : Bare_Name;
Match_0 : Bare_Call_Expr;
Match_1 : Bare_Explicit_Deref;
Match_2 : Bare_Name;
Fld : Bare_Name;
Fld_1 : Bare_Name;
Match_Prefix : Bare_Name;
Cast_Result : Bare_Call_Expr;
Let_Result : Bare_Name;
Scope_Result : Bare_Name;
Cast_Result_1 : Bare_Explicit_Deref;
Let_Result_1 : Bare_Name;
Scope_Result_1 : Bare_Name;
Cast_Result_2 : Bare_Name;
Let_Result_2 : Bare_Name;
Scope_Result_2 : Bare_Name;
Match_Result : Bare_Name;
Is_A : Boolean;
Fld_2 : Bare_Name;
Is_A_1 : Boolean;
If_Result : Bare_Name;
If_Result_1 : Bare_Name;
Let_Result_3 : Bare_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 7034 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 7029 '<Match at ast.py:12600>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12600
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Self; 
case Ada_Name (Match_Prefix.Kind) is
         when Ada_Call_Expr =>
   --# scope-start
--# scope-start
      Cast_Result := Match_Prefix;
Match_0 := Cast_Result; 
--# bind ce Match_0
--# expr-start 7027 '<FieldAccess for name at ast.py:12601>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12601
      if Match_0 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Match_0.Call_Expr_F_Name;
--# expr-done 7027
Let_Result := Fld; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Explicit_Deref =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Match_Prefix;
Match_1 := Cast_Result_1; 
--# bind ed Match_1
--# expr-start 7028 '<FieldAccess for prefix at ast.py:12602>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12602
      if Match_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Match_1.Explicit_Deref_F_Prefix;
--# expr-done 7028
Let_Result_1 := Fld_1; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref | Ada_Defining_Name .. Ada_End_Name | Ada_Qual_Expr .. Ada_Target_Name =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Match_Prefix;
Match_2 := Cast_Result_2; 
--# bind _ Match_2
Let_Result_2 := No_Bare_Ada_Node; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 7029
Block_Var := Match_Result; 
--# bind name Block_Var
--# expr-start 7033 '<Cond at ast.py:12607>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12607
--# expr-start 7030 '<IsA CallExpr, ExplicitDeref at ast.py:12608>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12608
Is_A := Block_Var /= null 
and then Block_Var.Kind in Ada_Call_Expr_Range | Ada_Explicit_Deref_Range; 
--# expr-done 7030
if Is_A then
   --# expr-start 7031 '<FieldAccess for innermost_name at ast.py:12608>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12608
      if Block_Var = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.innermost_name
Fld_2 := Libadalang.Implementation.Name_P_Innermost_Name (Node => Block_Var);
--# end
--# expr-done 7031
   If_Result_1 := Fld_2;
else
--# expr-start 7032 '<IsA QualExpr at ast.py:12609>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12609
Is_A_1 := Block_Var /= null 
and then Block_Var.Kind in Ada_Qual_Expr_Range; 
--# expr-done 7032
if Is_A_1 then
   If_Result := Block_Var;
else
   If_Result := Self;
end if;
   If_Result_1 := If_Result;
end if;
--# expr-done 7033
Let_Result_3 := If_Result_1; 
--# end
--# expr-done 7034
      Property_Result := Let_Result_3;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Innermost_Name;
--# end
--# property-start Name.parent_name_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12614
pragma Warnings (Off, "is not referenced");
 function Name_P_Parent_Name_Equation
  (Node : Bare_Name
      ; Typ : Internal_Entity_Base_Type_Decl
      ; Root : Bare_Name
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind typ Typ
   --# bind root Root
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_947 with Inline_Always;
            procedure Finalizer_Scope_2694 with Inline_Always;
            procedure Finalizer_Scope_2695 with Inline_Always;
            procedure Finalizer_Scope_2697 with Inline_Always;
            procedure Finalizer_Scope_2698 with Inline_Always;
      Block_Var : Internal_Entity_Access_To_Subp_Def;
Block_Var_1 : Boolean;
Block_Var_2 : Boolean;
Block_Var_3 : Internal_Entity_Base_Type_Decl;
Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Entity_Access_To_Subp_Def;
Var_Expr_2 : Internal_Entity_Access_To_Subp_Def;
Var_Expr_3 : Internal_Entity_Base_Type_Decl;
Match_0 : Bare_Call_Expr;
Match_1 : Bare_Explicit_Deref;
Match_2 : Bare_Name;
Var_Expr_4 : Internal_Entity_Access_To_Subp_Def;
Var_Expr_5 : Internal_Entity_Name;
Var_Expr_6 : Internal_Entity_Call_Expr;
Fld : Internal_Entity_Access_Def;
Scope_Result : Internal_Entity_Access_Def;
Result_Var : Internal_Entity_Access_Def;
Cast_Result : Internal_Entity_Access_To_Subp_Def;
Fld_1 : Internal_Entity_Subp_Spec;
Fld_2 : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
Fld_3 : Internal_Entity_Subp_Spec;
Fld_4 : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
Is_A : Boolean;
Not_Val : Boolean;
Fld_5 : Internal_Entity_Base_Type_Decl;
Scope_Result_3 : Internal_Entity_Base_Type_Decl;
Result_Var_3 : Internal_Entity_Base_Type_Decl;
False_Rel : Logic_Equation;
As_Entity : Internal_Entity_Call_Expr;
Is_Null : Boolean;
If_Result : Internal_Entity_Call_Expr;
Fld_6 : Logic_Equation;
As_Entity_1 : Internal_Entity_Explicit_Deref;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Explicit_Deref;
Fld_7 : Logic_Equation;
Fld_8 : Logic_Var;
Fld_9 : Internal_Entity_Subp_Spec;
Scope_Result_4 : Internal_Entity_Subp_Spec;
Result_Var_4 : Internal_Entity_Subp_Spec;
Cast_Result_1 : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_10 : Logic_Var;
Fld_11 : Bare_Ada_Node;
As_Entity_2 : Internal_Entity;
Is_Null_2 : Boolean;
If_Result_2 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
Match_Prefix : Bare_Name;
Cast_Result_2 : Bare_Call_Expr;
Let_Result : Logic_Equation;
Scope_Result_5 : Logic_Equation;
Cast_Result_3 : Bare_Explicit_Deref;
Let_Result_1 : Logic_Equation;
Scope_Result_6 : Logic_Equation;
Cast_Result_4 : Bare_Name;
Let_Result_2 : Logic_Equation;
Scope_Result_7 : Logic_Equation;
Match_Result : Logic_Equation;
Fld_12 : Bare_Name;
As_Entity_3 : Internal_Entity_Name;
Is_Null_3 : Boolean;
If_Result_3 : Internal_Entity_Name;
Is_A_1 : Boolean;
Is_Null_4 : Boolean;
Not_Val_1 : Boolean;
If_Result_4 : Boolean;
Fld_13 : Logic_Equation;
Fld_14 : Logic_Equation;
Fld_15 : Logic_Equation;
Or_Pred : Logic_Equation;
Fld_16 : Logic_Equation;
If_Result_5 : Logic_Equation;
If_Result_6 : Logic_Equation;
Cast_Expr : Internal_Entity_Name;
Cast_Result_5 : Internal_Entity_Call_Expr;
Fld_17 : Boolean;
Scope_Result_8 : Boolean;
Result_Var_5 : Boolean;
Fld_18 : Logic_Equation;
Fld_19 : Logic_Equation;
If_Result_7 : Logic_Equation;
If_Result_8 : Logic_Equation;
Scope_Result_9 : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var_6 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Is_Null_5 : Boolean;
If_Result_9 : Logic_Equation;
Let_Result_3 : Logic_Equation;
            procedure Finalizer_Scope_947 is
            begin
                     Dec_Ref (False_Rel);
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result_5);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_6);
                     Dec_Ref (Let_Result_2);
                     Dec_Ref (Scope_Result_7);
                     Dec_Ref (Match_Result);
                     Dec_Ref (Scope_Result_9);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var_6);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (If_Result_9);
                     Dec_Ref (Let_Result_3);
            end Finalizer_Scope_947;
            procedure Finalizer_Scope_2694 is
            begin
                     Dec_Ref (Fld_6);
            end Finalizer_Scope_2694;
            procedure Finalizer_Scope_2695 is
            begin
                     Dec_Ref (Fld_7);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
            end Finalizer_Scope_2695;
            procedure Finalizer_Scope_2697 is
            begin
                     Dec_Ref (Bind_Result_1);
            end Finalizer_Scope_2697;
            procedure Finalizer_Scope_2698 is
            begin
                     Dec_Ref (Fld_13);
                     Dec_Ref (Fld_14);
                     Dec_Ref (Fld_15);
                     Dec_Ref (Or_Pred);
                     Dec_Ref (Fld_16);
                     Dec_Ref (If_Result_5);
                     Dec_Ref (If_Result_6);
                     Dec_Ref (Fld_18);
                     Dec_Ref (Fld_19);
                     Dec_Ref (If_Result_7);
                     Dec_Ref (If_Result_8);
            end Finalizer_Scope_2698;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7075 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 7036 '<Cast to AccessToSubpDef at ast.py:12618>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12618
Var_Expr := Typ;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 7035 '<FieldAccess for access_def at ast.py:12618>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12618
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.access_def'
Fld := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Access_Def (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 7035
Scope_Result := Fld;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Access_Def;
end if;
   if Result_Var.Node = null
      or else Result_Var.Node.Kind in Ada_Access_To_Subp_Def_Range
   then
      Cast_Result := Create_Internal_Entity_Access_To_Subp_Def
        (Node => Result_Var.Node,
         Info => Result_Var.Info);
   else
         Cast_Result := No_Entity_Access_To_Subp_Def;
   end if;
--# expr-done 7036
Block_Var := Cast_Result; 
--# bind as_subp_access Block_Var
Var_Expr_1 := Block_Var;
if Var_Expr_1 /= No_Entity_Access_To_Subp_Def then
   --# scope-start
--# expr-start 7038 '<FieldAccess for paramless at ast.py:12619>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12619
--# expr-start 7037 '<FieldAccess for subp_spec at ast.py:12619>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12619
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Subp_Spec (Node => Var_Expr_1.Node.Access_To_Subp_Def_F_Subp_Spec, Info => Var_Expr_1.Info);
--# expr-done 7037
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.paramless
Fld_2 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Paramless (Node => Fld_1.Node, Dottable_Subp => False, Can_Be => False);
--# end
--# expr-done 7038
Scope_Result_1 := Fld_2;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
Block_Var_1 := Result_Var_1; 
--# bind is_paramless Block_Var_1
Var_Expr_2 := Block_Var;
if Var_Expr_2 /= No_Entity_Access_To_Subp_Def then
   --# scope-start
--# expr-start 7040 '<FieldAccess for paramless at ast.py:12623>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12623
--# expr-start 7039 '<FieldAccess for subp_spec at ast.py:12623>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12623
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Subp_Spec (Node => Var_Expr_2.Node.Access_To_Subp_Def_F_Subp_Spec, Info => Var_Expr_2.Info);
--# expr-done 7039
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.paramless
Fld_4 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Paramless (Node => Fld_3.Node, Dottable_Subp => False, Can_Be => True);
--# end
--# expr-done 7040
Scope_Result_2 := Fld_4;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := False;
end if;
Block_Var_2 := Result_Var_2; 
--# bind can_be_paramless Block_Var_2
Var_Expr_3 := Typ;
if Var_Expr_3 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 7043 '<FieldAccess for comp_type at ast.py:12628>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12628
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7042 '<Not at ast.py:12628>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12628
--# expr-start 7041 '<IsA ExplicitDeref at ast.py:12628>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12628
Is_A := Self /= null 
and then Self.Kind in Ada_Explicit_Deref_Range; 
--# expr-done 7041
Not_Val := not (Is_A); 
--# expr-done 7042
--# property-call-start BaseTypeDecl.comp_type
Fld_5 := Libadalang.Implementation.Base_Type_Decl_P_Comp_Type (Node => Var_Expr_3.Node, Is_Subscript => Not_Val, Origin => Origin, E_Info => Var_Expr_3.Info);
--# end
--# expr-done 7043
Scope_Result_3 := Fld_5;
   --# end
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := No_Entity_Base_Type_Decl;
end if;
Block_Var_3 := Result_Var_3; 
--# bind comp_type Block_Var_3
--# expr-start 7074 '<If at ast.py:12630>' If_Result_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12630
--# expr-start 7044 '<IsNull at ast.py:12631>' Is_Null_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12631
Is_Null_5 := Typ.Node = null; 
--# expr-done 7044
if Is_Null_5 then
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12632") else null)); 
   If_Result_9 := False_Rel;
else
   --# expr-start 7073 '<And at ast.py:12634>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12634
--# expr-start 7056 '<Match at ast.py:12634>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12634
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Self; 
case Ada_Name (Match_Prefix.Kind) is
         when Ada_Call_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Match_Prefix;
Match_0 := Cast_Result_2; 
--# bind ce Match_0
--# expr-start 7046 '<FieldAccess for subscriptable_type_equation at ast.py:12636>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12636
--# expr-start 7045 '<as_entity at ast.py:12636>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12636
Is_Null := Match_0 = null; 
if Is_Null then
   If_Result := No_Entity_Call_Expr;
else
As_Entity := (Info => E_Info, Node => Match_0); 
   If_Result := As_Entity;
end if;
--# expr-done 7045
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.subscriptable_type_equation
Fld_6 := Libadalang.Implementation.Call_Expr_P_Subscriptable_Type_Equation (Node => If_Result.Node, Typ => Typ, Env => Env, Origin => Origin, E_Info => If_Result.Info);
--# end
--# expr-done 7046
Let_Result := Fld_6; Inc_Ref (Let_Result);
--# end
Scope_Result_5 := Let_Result;
   Inc_Ref (Scope_Result_5);
   --# end
      Finalizer_Scope_2694;
            Match_Result := Scope_Result_5; Inc_Ref (Match_Result);
         when Ada_Explicit_Deref =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Match_Prefix;
Match_1 := Cast_Result_3; 
--# bind ed Match_1
--# expr-start 7052 '<And at ast.py:12639>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12639
--# expr-start 7048 '<FieldAccess for eq_for_type at ast.py:12639>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12639
--# expr-start 7047 '<as_entity at ast.py:12639>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12639
Is_Null_1 := Match_1 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Explicit_Deref;
else
As_Entity_1 := (Info => E_Info, Node => Match_1); 
   If_Result_1 := As_Entity_1;
end if;
--# expr-done 7047
      if If_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ExplicitDeref.eq_for_type
Fld_7 := Libadalang.Implementation.Explicit_Deref_P_Eq_For_Type (Node => If_Result_1.Node, Typ => Typ, Env => Env, Origin => Origin, E_Info => If_Result_1.Info);
--# end
--# expr-done 7048
--# expr-start 7051 '<Bind at ast.py:12644>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12644
--# expr-start 7049 '<FieldAccess for subp_spec_var at ast.py:12644>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12644
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.subp_spec_var'
Fld_8 := Libadalang.Implementation.Dispatcher_Name_P_Subp_Spec_Var (Node => Self);
--# end
--# expr-done 7049
Fld_8.Value := No_Entity;
Entity_Vars.Reset (Fld_8);
Var_Expr_4 := Block_Var;
if Var_Expr_4 /= No_Entity_Access_To_Subp_Def then
   --# scope-start
--# expr-start 7050 '<FieldAccess for subp_spec at ast.py:12644>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12644
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Subp_Spec (Node => Var_Expr_4.Node.Access_To_Subp_Def_F_Subp_Spec, Info => Var_Expr_4.Info);
--# expr-done 7050
Scope_Result_4 := Fld_9;
   --# end
   Result_Var_4 := Scope_Result_4;
else
   Result_Var_4 := No_Entity_Subp_Spec;
end if;
      Cast_Result_1 := Create_Internal_Entity
        (Node => Result_Var_4.Node,
         Info => Result_Var_4.Info);
Bind_Result := Solver.Create_Assign (Fld_8, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12644") else null)); 
--# expr-done 7051
And_Pred := Create_And (Fld_7, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12639") else null)); 
--# expr-done 7052
Let_Result_1 := And_Pred; Inc_Ref (Let_Result_1);
--# end
Scope_Result_6 := Let_Result_1;
   Inc_Ref (Scope_Result_6);
   --# end
      Finalizer_Scope_2695;
            Match_Result := Scope_Result_6; Inc_Ref (Match_Result);
         when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref | Ada_Defining_Name .. Ada_End_Name | Ada_Qual_Expr .. Ada_Target_Name =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Match_Prefix;
Match_2 := Cast_Result_4; 
--# bind _ Match_2
--# expr-start 7055 '<Bind at ast.py:12646>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12646
--# expr-start 7053 '<FieldAccess for type_var at ast.py:12646>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12646
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 7053
Fld_10.Value := No_Entity;
Entity_Vars.Reset (Fld_10);
--# expr-start 7054 '<FieldAccess for node at ast.py:12646>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12646
Fld_11 := No_Entity.Node;
--# expr-done 7054
Is_Null_2 := Fld_11 = null; 
if Is_Null_2 then
   If_Result_2 := No_Entity;
else
As_Entity_2 := (Info => E_Info, Node => Fld_11); 
   If_Result_2 := As_Entity_2;
end if;
Bind_Result_1 := Solver.Create_Assign (Fld_10, If_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12646") else null)); 
--# expr-done 7055
Let_Result_2 := Bind_Result_1; Inc_Ref (Let_Result_2);
--# end
Scope_Result_7 := Let_Result_2;
   Inc_Ref (Scope_Result_7);
   --# end
      Finalizer_Scope_2697;
            Match_Result := Scope_Result_7; Inc_Ref (Match_Result);
end case;
--# expr-done 7056
--# expr-start 7058 '<as_entity at ast.py:12647>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12647
--# expr-start 7057 '<FieldAccess for parent_name at ast.py:12647>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12647
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.parent_name
Fld_12 := Libadalang.Implementation.Name_P_Parent_Name (Node => Self, Stop_At => Root);
--# end
--# expr-done 7057
Is_Null_3 := Fld_12 = null; 
if Is_Null_3 then
   If_Result_3 := No_Entity_Name;
else
As_Entity_3 := (Info => E_Info, Node => Fld_12); 
   If_Result_3 := As_Entity_3;
end if;
--# expr-done 7058
Var_Expr_5 := If_Result_3;
if Var_Expr_5 /= No_Entity_Name then
   --# scope-start
   --# bind pn Var_Expr_5
--# expr-start 7072 '<Cond at ast.py:12648>' If_Result_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12648
--# expr-start 7059 '<IsA ExplicitDeref at ast.py:12649>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12649
Is_A_1 := Self /= null 
and then Self.Kind in Ada_Explicit_Deref_Range; 
--# expr-done 7059
if Is_A_1 then
   --# expr-start 7061 '<Not at ast.py:12649>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12649
--# expr-start 7060 '<IsNull at ast.py:12649>' Is_Null_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12649
Is_Null_4 := Block_Var.Node = null; 
--# expr-done 7060
Not_Val_1 := not (Is_Null_4); 
--# expr-done 7061
   If_Result_4 := Not_Val_1;
else
   If_Result_4 := False;
end if;
if If_Result_4 then
   --# expr-start 7067 '<Cond at ast.py:12653>' If_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12653
if Block_Var_1 then
   --# expr-start 7062 '<FieldAccess for parent_name_equation at ast.py:12659>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12659
      if Var_Expr_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.parent_name_equation
Fld_13 := Libadalang.Implementation.Name_P_Parent_Name_Equation (Node => Var_Expr_5.Node, Typ => Block_Var_3, Root => Root, Env => Env, Origin => Origin, E_Info => Var_Expr_5.Info);
--# end
--# expr-done 7062
   If_Result_6 := Fld_13;
else
if Block_Var_2 then
   --# expr-start 7065 '<Or at ast.py:12666>' Or_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12666
--# expr-start 7063 '<FieldAccess for parent_name_equation at ast.py:12667>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12667
      if Var_Expr_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.parent_name_equation
Fld_14 := Libadalang.Implementation.Name_P_Parent_Name_Equation (Node => Var_Expr_5.Node, Typ => Block_Var_3, Root => Root, Env => Env, Origin => Origin, E_Info => Var_Expr_5.Info);
--# end
--# expr-done 7063
--# expr-start 7064 '<FieldAccess for parent_name_equation at ast.py:12668>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12668
      if Var_Expr_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.parent_name_equation
Fld_15 := Libadalang.Implementation.Name_P_Parent_Name_Equation (Node => Var_Expr_5.Node, Typ => Typ, Root => Root, Env => Env, Origin => Origin, E_Info => Var_Expr_5.Info);
--# end
--# expr-done 7064
Or_Pred := Create_Or (Fld_14, Fld_15, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12666") else null)); 
--# expr-done 7065
   If_Result_5 := Or_Pred;
else
   --# expr-start 7066 '<FieldAccess for parent_name_equation at ast.py:12674>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12674
      if Var_Expr_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.parent_name_equation
Fld_16 := Libadalang.Implementation.Name_P_Parent_Name_Equation (Node => Var_Expr_5.Node, Typ => Typ, Root => Root, Env => Env, Origin => Origin, E_Info => Var_Expr_5.Info);
--# end
--# expr-done 7066
   If_Result_5 := Fld_16;
end if;
      Inc_Ref (If_Result_5);
   If_Result_6 := If_Result_5;
end if;
      Inc_Ref (If_Result_6);
--# expr-done 7067
   If_Result_8 := If_Result_6;
else
--# expr-start 7068 '<Cast to CallExpr at ast.py:12678>' Cast_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12678
Cast_Expr := Ent; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Call_Expr_Range
   then
      Cast_Result_5 := Create_Internal_Entity_Call_Expr
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result_5 := No_Entity_Call_Expr;
   end if;
--# expr-done 7068
Var_Expr_6 := Cast_Result_5;
if Var_Expr_6 /= No_Entity_Call_Expr then
   --# scope-start
--# expr-start 7069 '<FieldAccess for check_array_slice at ast.py:12678>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12678
      if Var_Expr_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.check_array_slice
Fld_17 := Libadalang.Implementation.Call_Expr_P_Check_Array_Slice (Node => Var_Expr_6.Node, Typ => Typ, Origin => Origin, E_Info => Var_Expr_6.Info);
--# end
--# expr-done 7069
Scope_Result_8 := Fld_17;
   --# end
   Result_Var_5 := Scope_Result_8;
else
   Result_Var_5 := False;
end if;
if Result_Var_5 then
   --# expr-start 7070 '<FieldAccess for parent_name_equation at ast.py:12679>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12679
      if Var_Expr_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.parent_name_equation
Fld_18 := Libadalang.Implementation.Name_P_Parent_Name_Equation (Node => Var_Expr_5.Node, Typ => Typ, Root => Root, Env => Env, Origin => Origin, E_Info => Var_Expr_5.Info);
--# end
--# expr-done 7070
   If_Result_7 := Fld_18;
else
   --# expr-start 7071 '<FieldAccess for parent_name_equation at ast.py:12683>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12683
      if Var_Expr_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.parent_name_equation
Fld_19 := Libadalang.Implementation.Name_P_Parent_Name_Equation (Node => Var_Expr_5.Node, Typ => Block_Var_3, Root => Root, Env => Env, Origin => Origin, E_Info => Var_Expr_5.Info);
--# end
--# expr-done 7071
   If_Result_7 := Fld_19;
end if;
      Inc_Ref (If_Result_7);
   If_Result_8 := If_Result_7;
end if;
      Inc_Ref (If_Result_8);
--# expr-done 7072
Scope_Result_9 := If_Result_8;
   Inc_Ref (Scope_Result_9);
   --# end
      Finalizer_Scope_2698;
   Result_Var_6 := Scope_Result_9;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12685") else null)); 
   Result_Var_6 := True_Rel;
end if;
      Inc_Ref (Result_Var_6);
And_Pred_1 := Create_And (Match_Result, Result_Var_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:12634") else null)); 
--# expr-done 7073
   If_Result_9 := And_Pred_1;
end if;
      Inc_Ref (If_Result_9);
--# expr-done 7074
Let_Result_3 := If_Result_9; Inc_Ref (Let_Result_3);
--# end
--# expr-done 7075
      Property_Result := Let_Result_3;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_947;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_947;
               Finalizer_Scope_2694;
               Finalizer_Scope_2695;
               Finalizer_Scope_2697;
               Finalizer_Scope_2698;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Parent_Name_Equation;
--# end
--# property-start Name.subtype_indication_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12690
pragma Warnings (Off, "is not referenced");
 function Name_P_Subtype_Indication_Equation
  (Node : Bare_Name
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_948 with Inline_Always;
      Fld : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_948 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_948;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7077 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7076 '<FieldAccess for xref_no_overloading at ast.py:12691>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12691
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7076
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 7077
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_948;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_948;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Subtype_Indication_Equation;
--# end
--# property-start Name.can_designate_primitive /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12694
pragma Warnings (Off, "is not referenced");
 function Name_P_Can_Designate_Primitive
  (Node : Bare_Name
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Match_0 : Bare_Name;
Match_1 : Bare_Renaming_Clause;
Match_2 : Bare_Ada_Node;
Fld : Bare_Ada_Node;
Fld_1 : Boolean;
Fld_2 : Bare_Ada_Node;
Is_A : Boolean;
Not_Val : Boolean;
Match_Prefix : Bare_Ada_Node;
Cast_Result : Bare_Name;
Let_Result : Boolean;
Scope_Result : Boolean;
Cast_Result_1 : Bare_Renaming_Clause;
Let_Result_1 : Boolean;
Scope_Result_1 : Boolean;
Cast_Result_2 : Bare_Ada_Node;
Let_Result_2 : Boolean;
Scope_Result_2 : Boolean;
Match_Result : Boolean;
Let_Result_3 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 7084 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 7083 '<Match at ast.py:12703>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12703
--# expr-start 7078 '<parent at ast.py:12703>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12703
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 7078
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld; 
case Ada_Ada_Node (Match_Prefix.Kind) is
         when Ada_Attribute_Ref .. Ada_Target_Name =>
   --# scope-start
--# scope-start
      Cast_Result := Match_Prefix;
Match_0 := Cast_Result; 
--# bind n Match_0
--# expr-start 7079 '<FieldAccess for can_designate_primitive at ast.py:12704>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12704
      if Match_0 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.can_designate_primitive
Fld_1 := Libadalang.Implementation.Name_P_Can_Designate_Primitive (Node => Match_0);
--# end
--# expr-done 7079
Let_Result := Fld_1; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Renaming_Clause .. Ada_Synthetic_Renaming_Clause =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Match_Prefix;
Match_1 := Cast_Result_1; 
--# bind r Match_1
--# expr-start 7082 '<Not at ast.py:12705>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12705
--# expr-start 7081 '<IsA PackageRenamingDecl at ast.py:12705>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12705
--# expr-start 7080 '<parent at ast.py:12705>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12705
      if Match_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Match_1.Parent; 
--# expr-done 7080
Is_A := Fld_2 /= null 
and then Fld_2.Kind in Ada_Package_Renaming_Decl_Range; 
--# expr-done 7081
Not_Val := not (Is_A); 
--# expr-done 7082
Let_Result_1 := Not_Val; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abort_Absent .. Ada_Membership_Expr | Ada_Paren_Expr .. Ada_Range_Spec | Ada_Reverse_Absent .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Match_Prefix;
Match_2 := Cast_Result_2; 
--# bind _ Match_2
Let_Result_2 := True; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 7083
Let_Result_3 := Match_Result; 
--# end
--# expr-done 7084
      Property_Result := Let_Result_3;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Can_Designate_Primitive;
--# end
--# property-start Name.parent_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12710
pragma Warnings (Off, "is not referenced");
 function Name_P_Parent_Name
  (Node : Bare_Name
      ; Stop_At : Bare_Name
  )
   return Bare_Name
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind stop_at Stop_At
   Property_Result : Bare_Name;
      Fld : Bare_Ada_Node;
Cast_Result : Bare_Name;
Is_Null : Boolean;
Cast_Expr : Bare_Name;
Cast_Result_1 : Bare_Name;
Is_Equal : Boolean;
If_Result : Boolean;
If_Result_1 : Bare_Name;
Let_Result : Bare_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 7090 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7089 '<If at ast.py:12714>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12714
--# expr-start 7085 '<IsNull at ast.py:12714>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12714
Is_Null := Stop_At = null; 
--# expr-done 7085
if Is_Null then
   If_Result := True;
else
   --# expr-start 7086 '<Eq at ast.py:12714>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12714
Cast_Expr := Self; 
      Cast_Result_1 := Cast_Expr;
Is_Equal := Cast_Result_1 = Stop_At; 
--# expr-done 7086
   If_Result := Is_Equal;
end if;
if If_Result then
   If_Result_1 := No_Bare_Ada_Node;
else
   --# expr-start 7088 '<Cast to Name at ast.py:12716>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12716
--# expr-start 7087 '<parent at ast.py:12716>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12716
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 7087
   if Fld = null
      or else Fld.Kind in Ada_Name
   then
      Cast_Result := Fld;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 7088
   If_Result_1 := Cast_Result;
end if;
--# expr-done 7089
Let_Result := If_Result_1; 
--# end
--# expr-done 7090
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Parent_Name;
--# end
--# property-start Name.parent_callexpr /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12719
pragma Warnings (Off, "is not referenced");
 function Name_P_Parent_Callexpr
  (Node : Bare_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Call_Expr
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Call_Expr;
            procedure Finalizer_Scope_951 with Inline_Always;
      Match_0 : Internal_Entity_Dotted_Name;
Match_1 : Internal_Entity_Call_Expr;
Match_2 : Internal_Entity;
Fld : Internal_Entity_Array_Access;
Is_A : Boolean;
Is_A_1 : Boolean;
Fld_1 : Internal_Entity;
Fld_2 : Internal_Entity_Base_Id;
Cast_Result : Internal_Entity;
Is_Equiv : Boolean;
Fld_3 : Internal_Entity_Name;
Cast_Result_1 : Internal_Entity;
Is_Equiv_1 : Boolean;
Match_Prefix : Internal_Entity;
Cast_Result_2 : Internal_Entity_Dotted_Name;
Let_Result : Boolean;
Scope_Result : Boolean;
Cast_Result_3 : Internal_Entity_Call_Expr;
Let_Result_1 : Boolean;
Scope_Result_1 : Boolean;
Cast_Result_4 : Internal_Entity;
Let_Result_2 : Boolean;
Scope_Result_2 : Boolean;
Match_Result : Boolean;
If_Result : Boolean;
If_Result_1 : Boolean;
Map_Result : Internal_Entity_Array_Access;
Is_A_2 : Boolean;
Find_Result : Internal_Entity;
Cast_Result_5 : Internal_Entity_Call_Expr;
Let_Result_3 : Internal_Entity_Call_Expr;
            procedure Finalizer_Scope_951 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_951;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 7104 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 7103 '<Cast to CallExpr at ast.py:12736>' Cast_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12736
--# expr-start 7102 '<Find at ast.py:12736>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12736
   --# expr-start 7100 '<TakeWhile at ast.py:12736>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12736
   --# expr-start 7091 '<FieldAccess for parents at ast.py:12736>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12736
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parents (Node => Ent.Node, With_Self => True, E_Info => Ent.Info);
--# expr-done 7091
      declare
         Map_Result_Vec : Internal_Entity_Vectors.Vector;
            Take_While_Ok : Boolean := False;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Fld;
      begin
         for Item_238 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_238
--# expr-start 7092 '<IsA CallExpr at ast.py:12737>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12737
Is_A := Item_238.Node /= null 
and then Item_238.Node.Kind in Ada_Call_Expr_Range; 
--# expr-done 7092
if Is_A then
   If_Result_1 := True;
else
--# expr-start 7093 '<IsA DottedName, BaseId at ast.py:12738>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12738
Is_A_1 := Item_238.Node /= null 
and then Item_238.Node.Kind in Ada_Dotted_Name_Range | Ada_Base_Id; 
--# expr-done 7093
if Is_A_1 then
   --# expr-start 7099 '<Match at ast.py:12738>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12738
--# expr-start 7094 '<parent at ast.py:12738>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12738
      if Item_238.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Parent (Node => Item_238.Node, E_Info => Item_238.Info);
--# expr-done 7094
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld_1; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Dotted_Name =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Dotted_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_2; 
--# bind pfx Match_0
--# expr-start 7096 '<Eq at ast.py:12739>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12739
--# expr-start 7095 '<FieldAccess for suffix at ast.py:12739>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12739
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Base_Id (Node => Match_0.Node.Dotted_Name_F_Suffix, Info => Match_0.Info);
--# expr-done 7095
      Cast_Result := Create_Internal_Entity
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Is_Equiv := Equivalent (Cast_Result, Item_238); 
--# expr-done 7096
Let_Result := Is_Equiv; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Call_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Call_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_3; 
--# bind ce Match_1
--# expr-start 7098 '<Eq at ast.py:12740>' Is_Equiv_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12740
--# expr-start 7097 '<FieldAccess for name at ast.py:12740>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12740
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Name (Node => Match_1.Node.Call_Expr_F_Name, Info => Match_1.Info);
--# expr-done 7097
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
Is_Equiv_1 := Equivalent (Cast_Result_1, Item_238); 
--# expr-done 7098
Let_Result_1 := Is_Equiv_1; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abort_Absent .. Ada_Update_Attribute_Ref | Ada_Defining_Name .. Ada_Discrete_Subtype_Name | Ada_End_Name .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_4; 
--# bind _ Match_2
Let_Result_2 := False; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 7099
   If_Result := Match_Result;
else
   If_Result := False;
end if;
   If_Result_1 := If_Result;
end if;
         Take_While_Ok := If_Result_1;
         if Take_While_Ok then
         declare
            Item_To_Append : constant Internal_Entity := Item_238;
         begin
            Internal_Entity_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
               exit when not Take_While_Ok;
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 7100
      Find_Result := No_Entity;
      declare
         Collection : constant Internal_Entity_Array_Access := Map_Result;
      begin
         for Item_237 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_237
      --# expr-start 7101 '<IsA CallExpr at ast.py:12743>' Is_A_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12743
Is_A_2 := Item_237.Node /= null 
and then Item_237.Node.Kind in Ada_Call_Expr_Range; 
--# expr-done 7101
      if Is_A_2 then
         Find_Result := Item_237;
      end if;
   --# end
               exit when Is_A_2;
         end loop;
      end;
--# expr-done 7102
   if Find_Result.Node = null
      or else Find_Result.Node.Kind in Ada_Call_Expr_Range
   then
      Cast_Result_5 := Create_Internal_Entity_Call_Expr
        (Node => Find_Result.Node,
         Info => Find_Result.Info);
   else
         Cast_Result_5 := No_Entity_Call_Expr;
   end if;
--# expr-done 7103
Let_Result_3 := Cast_Result_5; 
--# end
--# expr-done 7104
      Property_Result := Let_Result_3;
   --# end
      Finalizer_Scope_951;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_951;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Parent_Callexpr;
--# end
--# property-start Name.is_range_attribute /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12746
pragma Warnings (Off, "is not referenced");
 function Name_P_Is_Range_Attribute
  (Node : Bare_Name
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Var_Expr : Bare_Attribute_Ref;
Cast_Expr : Bare_Name;
Cast_Result : Bare_Attribute_Ref;
As_Entity : Internal_Entity_Attribute_Ref;
Is_Null : Boolean;
If_Result : Internal_Entity_Attribute_Ref;
Fld : Internal_Entity_Identifier;
Sym : Symbol_Type;
Fld_1 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 7110 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7105 '<Cast to AttributeRef at ast.py:12750>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12750
Cast_Expr := Self; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_Attribute_Ref_Range
   then
      Cast_Result := Cast_Expr;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 7105
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind attr_ref Var_Expr
--# expr-start 7109 '<FieldAccess for name_is at ast.py:12752>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12752
--# expr-start 7107 '<FieldAccess for attribute at ast.py:12752>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12752
--# expr-start 7106 '<as_bare_entity at ast.py:12752>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12752
Is_Null := Var_Expr = null; 
if Is_Null then
   If_Result := No_Entity_Attribute_Ref;
else
As_Entity := (Info => No_Entity_Info, Node => Var_Expr); 
   If_Result := As_Entity;
end if;
--# expr-done 7106
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Identifier (Node => If_Result.Node.Attribute_Ref_F_Attribute, Info => If_Result.Info);
--# expr-done 7107
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7108 '<Symbol Range at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Range); 
--# expr-done 7108
--# property-call-start Name.name_is
Fld_1 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld.Node, Sym => Sym);
--# end
--# expr-done 7109
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 7110
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Is_Range_Attribute;
--# end
--# property-start '[dispatcher]Name.scope' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Name_P_Scope
  (Node : Bare_Name
      ; Env : Lexical_Env
  )
   return Lexical_Env
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   Property_Result : Lexical_Env;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Name (Self.Kind) is
               when Ada_Attribute_Ref .. Ada_Call_Expr | Ada_Discrete_Subtype_Name | Ada_Explicit_Deref .. Ada_Reduce_Attribute_Ref | Ada_Null_Literal .. Ada_Target_Name =>
                  --# property-call-start Name.scope
                  Property_Result := Name_P_Scope
                    (Self
                        , Env
                    );
                  --# end
               when Ada_Defining_Name .. Ada_Synthetic_Defining_Name =>
                  --# property-call-start DefiningName.scope
                  Property_Result := Defining_Name_P_Scope
                    (Self
                        , Env
                    );
                  --# end
               when Ada_Dotted_Name =>
                  --# property-call-start DottedName.scope
                  Property_Result := Dotted_Name_P_Scope
                    (Self
                        , Env
                    );
                  --# end
               when Ada_End_Name =>
                  --# property-call-start EndName.scope
                  Property_Result := End_Name_P_Scope
                    (Self
                        , Env
                    );
                  --# end
               when Ada_Char_Literal .. Ada_String_Literal =>
                  --# property-call-start BaseId.scope
                  Property_Result := Base_Id_P_Scope
                    (Self
                        , Env
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Name_P_Scope;
--# end
--# property-start '[dispatcher]Name.ref_var' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Name_P_Ref_Var
  (Node : Bare_Name
  )
   return Logic_Var
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Logic_Var;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Name (Self.Kind) is
               when Ada_Discrete_Subtype_Name | Ada_Synthetic_Identifier =>
                  --# property-call-start Name.ref_var
                  Property_Result := Name_P_Ref_Var
                    (Self
                    );
                  --# end
               when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref =>
                  --# property-call-start AttributeRef.ref_var
                  Property_Result := Attribute_Ref_P_Ref_Var
                    (Self
                    );
                  --# end
               when Ada_Call_Expr =>
                  --# property-call-start CallExpr.ref_var
                  Property_Result := Call_Expr_P_Ref_Var
                    (Self
                    );
                  --# end
               when Ada_Defining_Name .. Ada_Synthetic_Defining_Name =>
                  --# property-call-start DefiningName.ref_var
                  Property_Result := Defining_Name_P_Ref_Var
                    (Self
                    );
                  --# end
               when Ada_Dotted_Name =>
                  --# property-call-start DottedName.ref_var
                  Property_Result := Dotted_Name_P_Ref_Var
                    (Self
                    );
                  --# end
               when Ada_End_Name =>
                  --# property-call-start EndName.ref_var
                  Property_Result := End_Name_P_Ref_Var
                    (Self
                    );
                  --# end
               when Ada_Explicit_Deref =>
                  --# property-call-start ExplicitDeref.ref_var
                  Property_Result := Explicit_Deref_P_Ref_Var
                    (Self
                    );
                  --# end
               when Ada_Qual_Expr =>
                  --# property-call-start QualExpr.ref_var
                  Property_Result := Qual_Expr_P_Ref_Var
                    (Self
                    );
                  --# end
               when Ada_Reduce_Attribute_Ref =>
                  --# property-call-start ReduceAttributeRef.ref_var
                  Property_Result := Reduce_Attribute_Ref_P_Ref_Var
                    (Self
                    );
                  --# end
               when Ada_Char_Literal .. Ada_Real_Literal =>
                  --# property-call-start SingleTokNode.ref_var
                  Property_Result := Single_Tok_Node_P_Ref_Var
                    (Self
                    );
                  --# end
               when Ada_Target_Name =>
                  --# property-call-start TargetName.ref_var
                  Property_Result := Target_Name_P_Ref_Var
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Name_P_Ref_Var;
--# end
--# property-start '[dispatcher]Name.subp_spec_var' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Name_P_Subp_Spec_Var
  (Node : Bare_Name
  )
   return Logic_Var
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Logic_Var;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Name (Self.Kind) is
               when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref | Ada_Defining_Name .. Ada_Discrete_Subtype_Name | Ada_End_Name | Ada_Qual_Expr .. Ada_Reduce_Attribute_Ref | Ada_Synthetic_Identifier .. Ada_Target_Name =>
                  --# property-call-start Name.subp_spec_var
                  Property_Result := Name_P_Subp_Spec_Var
                    (Self
                    );
                  --# end
               when Ada_Call_Expr =>
                  --# property-call-start CallExpr.subp_spec_var
                  Property_Result := Call_Expr_P_Subp_Spec_Var
                    (Self
                    );
                  --# end
               when Ada_Dotted_Name =>
                  --# property-call-start DottedName.subp_spec_var
                  Property_Result := Dotted_Name_P_Subp_Spec_Var
                    (Self
                    );
                  --# end
               when Ada_Explicit_Deref =>
                  --# property-call-start ExplicitDeref.subp_spec_var
                  Property_Result := Explicit_Deref_P_Subp_Spec_Var
                    (Self
                    );
                  --# end
               when Ada_Char_Literal .. Ada_Real_Literal =>
                  --# property-call-start SingleTokNode.subp_spec_var
                  Property_Result := Single_Tok_Node_P_Subp_Spec_Var
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Name_P_Subp_Spec_Var;
--# end
--# property-start '[dispatcher]Name.defines_subp_spec_var' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Name_P_Defines_Subp_Spec_Var
  (Node : Bare_Name
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Name (Self.Kind) is
               when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref | Ada_Defining_Name .. Ada_Discrete_Subtype_Name | Ada_End_Name | Ada_Qual_Expr .. Ada_Reduce_Attribute_Ref | Ada_Synthetic_Identifier .. Ada_Target_Name =>
                  --# property-call-start Name.defines_subp_spec_var
                  Property_Result := Name_P_Defines_Subp_Spec_Var
                    (Self
                    );
                  --# end
               when Ada_Call_Expr =>
                  --# property-call-start CallExpr.defines_subp_spec_var
                  Property_Result := Call_Expr_P_Defines_Subp_Spec_Var
                    (Self
                    );
                  --# end
               when Ada_Dotted_Name =>
                  --# property-call-start DottedName.defines_subp_spec_var
                  Property_Result := Dotted_Name_P_Defines_Subp_Spec_Var
                    (Self
                    );
                  --# end
               when Ada_Explicit_Deref =>
                  --# property-call-start ExplicitDeref.defines_subp_spec_var
                  Property_Result := Explicit_Deref_P_Defines_Subp_Spec_Var
                    (Self
                    );
                  --# end
               when Ada_Char_Literal .. Ada_Real_Literal =>
                  --# property-call-start SingleTokNode.defines_subp_spec_var
                  Property_Result := Single_Tok_Node_P_Defines_Subp_Spec_Var
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Name_P_Defines_Subp_Spec_Var;
--# end
--# property-start Name.called_subp_spec /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12798
pragma Warnings (Off, "is not referenced");
 function Name_P_Called_Subp_Spec
  (Node : Bare_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Holder
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Formal_Param_Holder;
      Cast_Expr : Internal_Entity_Name;
Cast_Result : Internal_Entity;
Fld : Logic_Var;
Fld_1 : Internal_Logic_Val_Result;
Fld_2 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Base_Formal_Param_Holder;
Fld_3 : Boolean;
If_Result : Internal_Entity_Base_Formal_Param_Holder;
Let_Result : Internal_Entity_Base_Formal_Param_Holder;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7117 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7116 '<If at ast.py:12803>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12803
--# expr-start 7111 '<FieldAccess for defines_subp_spec_var at ast.py:12804>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12804
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.defines_subp_spec_var'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Defines_Subp_Spec_Var (Node => Self);
--# end
--# expr-done 7111
if Fld_3 then
   --# expr-start 7115 '<Cast to BaseFormalParamHolder at ast.py:12805>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12805
--# expr-start 7114 '<FieldAccess for value at ast.py:12805>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12805
--# expr-start 7113 '<FieldAccess for logic_val at ast.py:12805>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12805
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Cast_Expr := Ent; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
--# expr-start 7112 '<FieldAccess for subp_spec_var at ast.py:12807>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12807
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.subp_spec_var'
Fld := Libadalang.Implementation.Dispatcher_Name_P_Subp_Spec_Var (Node => Self);
--# end
--# expr-done 7112
--# property-call-start AdaNode.logic_val
Fld_1 := Libadalang.Implementation.Ada_Node_P_Logic_Val (Node => Self, From_Node => Cast_Result, Lvar => Fld);
--# end
--# expr-done 7113
Fld_2 := Fld_1.Value;
--# expr-done 7114
   if Fld_2.Node = null
      or else Fld_2.Node.Kind in Ada_Base_Formal_Param_Holder
   then
      Cast_Result_1 := Create_Internal_Entity_Base_Formal_Param_Holder
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
   else
         Cast_Result_1 := No_Entity_Base_Formal_Param_Holder;
   end if;
--# expr-done 7115
   If_Result := Cast_Result_1;
else
   If_Result := No_Entity_Base_Formal_Param_Holder;
end if;
--# expr-done 7116
Let_Result := If_Result; 
--# end
--# expr-done 7117
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Called_Subp_Spec;
--# end
--# property-start Name.referenced_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12814
pragma Warnings (Off, "is not referenced");
 function Name_P_Referenced_Decl
  (Node : Bare_Name
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Internal_Refd_Decl;
Fld_1 : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7120 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7119 '<FieldAccess for decl at ast.py:12821>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12821
--# expr-start 7118 '<FieldAccess for referenced_decl_internal at ast.py:12821>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12821
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl_internal
Fld := Libadalang.Implementation.Name_P_Referenced_Decl_Internal (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 7118
Fld_1 := Fld.Decl;
--# expr-done 7119
Let_Result := Fld_1; 
--# end
--# expr-done 7120
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Referenced_Decl;
--# end
--# property-start Name.failsafe_referenced_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12825
pragma Warnings (Off, "is not referenced");
 function Name_P_Failsafe_Referenced_Decl
  (Node : Bare_Name
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Refd_Decl
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Refd_Decl;
      Fld : Internal_Refd_Decl;
New_Struct : Internal_Refd_Decl;
Try_Result : Internal_Refd_Decl;
Let_Result : Internal_Refd_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7124 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7123 '<Try at ast.py:12830>' Try_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12830
begin
   --# expr-start 7121 '<FieldAccess for referenced_decl_internal at ast.py:12831>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12831
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl_internal
Fld := Libadalang.Implementation.Name_P_Referenced_Decl_Internal (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 7121
   Try_Result := Fld;
exception
   when Property_Error =>
      --# expr-start 7122 '<New RefdDecl at ast.py:12832>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12832
New_Struct := (Decl => No_Entity_Basic_Decl, Kind => Error); 
--# expr-done 7122
      Try_Result := New_Struct;
end;
--# expr-done 7123
Let_Result := Try_Result; 
--# end
--# expr-done 7124
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Failsafe_Referenced_Decl;
--# end
--# property-start Name.referenced_decl_internal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12837
pragma Warnings (Off, "is not referenced");
 function Name_P_Referenced_Decl_Internal
  (Node : Bare_Name
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Refd_Decl
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Refd_Decl;
      Var_Expr : Internal_Refd_Decl;
V : Internal_Logic_Val_Result;
Var_Expr_1 : Internal_Entity_Basic_Decl;
Var_Expr_2 : Internal_Entity_Name;
V_1 : Internal_Logic_Val_Result;
Var_Expr_3 : Internal_Refd_Decl;
Var_Expr_4 : Internal_Entity_Basic_Decl;
Cast_Expr : Internal_Entity_Name;
Cast_Result : Internal_Entity;
Fld : Logic_Var;
Fld_1 : Internal_Logic_Val_Result;
New_Struct : Internal_Logic_Val_Result;
Try_Result : Internal_Logic_Val_Result;
Fld_2 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Basic_Decl;
New_Struct_1 : Internal_Refd_Decl;
Fld_3 : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
New_Struct_2 : Internal_Refd_Decl;
Scope_Result_1 : Internal_Refd_Decl;
New_Struct_3 : Internal_Refd_Decl;
Result_Var_1 : Internal_Refd_Decl;
Fld_4 : Boolean;
If_Result : Internal_Refd_Decl;
Let_Result : Internal_Refd_Decl;
Cast_Expr_1 : Internal_Entity_Name;
Cast_Result_2 : Internal_Entity;
Fld_5 : Logic_Var;
Fld_6 : Internal_Logic_Val_Result;
Fld_7 : Internal_Entity;
Cast_Result_3 : Internal_Entity_Basic_Decl;
New_Struct_4 : Internal_Refd_Decl;
New_Struct_5 : Internal_Refd_Decl;
Fld_8 : Boolean;
If_Result_1 : Internal_Refd_Decl;
Let_Result_1 : Internal_Refd_Decl;
If_Result_2 : Internal_Refd_Decl;
Fld_9 : Internal_Entity_Basic_Decl;
Fld_10 : Internal_Entity_Basic_Decl;
Scope_Result_2 : Internal_Entity_Basic_Decl;
Result_Var_2 : Internal_Entity_Basic_Decl;
Fld_11 : Ref_Result_Kind;
New_Struct_6 : Internal_Refd_Decl;
Fld_12 : Internal_Entity_Basic_Decl;
Is_Null : Boolean;
Fld_13 : Ref_Result_Kind;
Is_Equal : Boolean;
If_Result_3 : Boolean;
If_Result_4 : Internal_Refd_Decl;
Scope_Result_3 : Internal_Refd_Decl;
Result_Var_3 : Internal_Refd_Decl;
Scope_Result_4 : Internal_Refd_Decl;
Result_Var_4 : Internal_Refd_Decl;
Fld_14 : Boolean;
If_Result_5 : Internal_Refd_Decl;
Let_Result_2 : Internal_Refd_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7159 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 7158 '<If at ast.py:12844>' If_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12844
--# expr-start 7125 '<FieldAccess for is_defining at ast.py:12847>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12847
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_defining
Fld_14 := Libadalang.Implementation.Name_P_Is_Defining (Node => Ent.Node);
--# end
--# expr-done 7125
if Fld_14 then
   If_Result_5 := No_Refd_Decl;
else
--# expr-start 7148 '<If at ast.py:12850>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12850
if Imprecise_Fallback then
   --# expr-start 7138 '<Let at ast.py:12854>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12854
--# scope-start
--# expr-start 7129 '<Try at ast.py:12854>' Try_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12854
begin
   --# expr-start 7127 '<FieldAccess for logic_val at ast.py:12855>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12855
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Cast_Expr := Ent; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
--# expr-start 7126 '<FieldAccess for ref_var at ast.py:12855>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12855
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Self);
--# end
--# expr-done 7126
--# property-call-start AdaNode.logic_val
Fld_1 := Libadalang.Implementation.Ada_Node_P_Logic_Val (Node => Self, From_Node => Cast_Result, Lvar => Fld);
--# end
--# expr-done 7127
   Try_Result := Fld_1;
exception
   when Property_Error =>
      --# expr-start 7128 '<New LogicValResult at ast.py:12856>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12856
New_Struct := (Success => False, Value => No_Entity); 
--# expr-done 7128
      Try_Result := New_Struct;
end;
--# expr-done 7129
V := Try_Result; 
--# bind v V
--# expr-start 7137 '<If at ast.py:12857>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12857
--# expr-start 7130 '<FieldAccess for success at ast.py:12858>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12858
Fld_4 := V.Success;
--# expr-done 7130
if Fld_4 then
   --# expr-start 7133 '<New RefdDecl at ast.py:12861>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12861
--# expr-start 7132 '<Cast to BasicDecl.entity at ast.py:12862>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12862
--# expr-start 7131 '<FieldAccess for value at ast.py:12862>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12862
Fld_2 := V.Value;
--# expr-done 7131
   if Fld_2.Node = null
      or else Fld_2.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Basic_Decl
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
   else
         Cast_Result_1 := No_Entity_Basic_Decl;
   end if;
--# expr-done 7132
New_Struct_1 := (Decl => Cast_Result_1, Kind => Precise); 
--# expr-done 7133
   If_Result := New_Struct_1;
else
Var_Expr_2 := Ent;
if Var_Expr_2 /= No_Entity_Name then
   --# scope-start
--# expr-start 7134 '<FieldAccess for first_corresponding_decl at ast.py:12868>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12868
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.first_corresponding_decl'
Fld_3 := Libadalang.Implementation.Dispatcher_Expr_P_First_Corresponding_Decl (Node => Var_Expr_2.Node, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 7134
Scope_Result := Fld_3;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Basic_Decl;
end if;
Var_Expr_1 := Result_Var;
if Var_Expr_1 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind fcd Var_Expr_1
--# expr-start 7135 '<New RefdDecl at ast.py:12870>' New_Struct_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12870
New_Struct_2 := (Decl => Var_Expr_1, Kind => Imprecise); 
--# expr-done 7135
Scope_Result_1 := New_Struct_2;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   --# expr-start 7136 '<New RefdDecl at ast.py:12871>' New_Struct_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12871
New_Struct_3 := (Decl => No_Entity_Basic_Decl, Kind => Error); 
--# expr-done 7136
   Result_Var_1 := New_Struct_3;
end if;
   If_Result := Result_Var_1;
end if;
--# expr-done 7137
Let_Result := If_Result; 
--# end
--# expr-done 7138
   If_Result_2 := Let_Result;
else
   --# expr-start 7147 '<Let at ast.py:12876>' Let_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12876
--# scope-start
--# expr-start 7140 '<FieldAccess for logic_val at ast.py:12876>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12876
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Cast_Expr_1 := Ent; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
--# expr-start 7139 '<FieldAccess for ref_var at ast.py:12876>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12876
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_5 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Self);
--# end
--# expr-done 7139
--# property-call-start AdaNode.logic_val
Fld_6 := Libadalang.Implementation.Ada_Node_P_Logic_Val (Node => Self, From_Node => Cast_Result_2, Lvar => Fld_5);
--# end
--# expr-done 7140
V_1 := Fld_6; 
--# bind v V_1
--# expr-start 7146 '<If at ast.py:12876>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12876
--# expr-start 7141 '<FieldAccess for success at ast.py:12877>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12877
Fld_8 := V_1.Success;
--# expr-done 7141
if Fld_8 then
   --# expr-start 7144 '<New RefdDecl at ast.py:12878>' New_Struct_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12878
--# expr-start 7143 '<Cast to BasicDecl.entity at ast.py:12879>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12879
--# expr-start 7142 '<FieldAccess for value at ast.py:12879>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12879
Fld_7 := V_1.Value;
--# expr-done 7142
   if Fld_7.Node = null
      or else Fld_7.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_3 := Create_Internal_Entity_Basic_Decl
        (Node => Fld_7.Node,
         Info => Fld_7.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 7143
New_Struct_4 := (Decl => Cast_Result_3, Kind => Precise); 
--# expr-done 7144
   If_Result_1 := New_Struct_4;
else
   --# expr-start 7145 '<New RefdDecl at ast.py:12882>' New_Struct_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12882
New_Struct_5 := (Decl => No_Entity_Basic_Decl, Kind => Error); 
--# expr-done 7145
   If_Result_1 := New_Struct_5;
end if;
--# expr-done 7146
Let_Result_1 := If_Result_1; 
--# end
--# expr-done 7147
   If_Result_2 := Let_Result_1;
end if;
--# expr-done 7148
Var_Expr := If_Result_2;
if Var_Expr /= No_Refd_Decl then
   --# scope-start
   --# bind res Var_Expr
--# expr-start 7152 '<New RefdDecl at ast.py:12887>' New_Struct_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12887
--# expr-start 7149 '<FieldAccess for decl at ast.py:12888>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12888
Fld_9 := Var_Expr.Decl;
--# expr-done 7149
Var_Expr_4 := Fld_9;
if Var_Expr_4 /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 7150 '<FieldAccess for wrap_public_reference at ast.py:12888>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12888
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.wrap_public_reference
Fld_10 := Libadalang.Implementation.Basic_Decl_P_Wrap_Public_Reference (Node => Var_Expr_4.Node, E_Info => Var_Expr_4.Info);
--# end
--# expr-done 7150
Scope_Result_2 := Fld_10;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Entity_Basic_Decl;
end if;
--# expr-start 7151 '<FieldAccess for kind at ast.py:12889>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12889
Fld_11 := Var_Expr.Kind;
--# expr-done 7151
New_Struct_6 := (Decl => Result_Var_2, Kind => Fld_11); 
--# expr-done 7152
Var_Expr_3 := New_Struct_6;
if Var_Expr_3 /= No_Refd_Decl then
   --# scope-start
   --# bind res Var_Expr_3
--# expr-start 7157 '<If at ast.py:12890>' If_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12890
--# expr-start 7154 '<IsNull at ast.py:12893>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12893
--# expr-start 7153 '<FieldAccess for decl at ast.py:12893>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12893
Fld_12 := Var_Expr_3.Decl;
--# expr-done 7153
Is_Null := Fld_12.Node = null; 
--# expr-done 7154
if Is_Null then
   --# expr-start 7156 '<Eq at ast.py:12893>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12893
--# expr-start 7155 '<FieldAccess for kind at ast.py:12893>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12893
Fld_13 := Var_Expr_3.Kind;
--# expr-done 7155
Is_Equal := Fld_13 = Precise; 
--# expr-done 7156
   If_Result_3 := Is_Equal;
else
   If_Result_3 := False;
end if;
if If_Result_3 then
   If_Result_4 := No_Refd_Decl;
else
   If_Result_4 := Var_Expr_3;
end if;
--# expr-done 7157
Scope_Result_3 := If_Result_4;
   --# end
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := No_Refd_Decl;
end if;
Scope_Result_4 := Result_Var_3;
   --# end
   Result_Var_4 := Scope_Result_4;
else
   Result_Var_4 := No_Refd_Decl;
end if;
   If_Result_5 := Result_Var_4;
end if;
--# expr-done 7158
Let_Result_2 := If_Result_5; 
--# end
--# expr-done 7159
      Property_Result := Let_Result_2;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Referenced_Decl_Internal;
--# end
--# property-start '[dispatcher]Name.designated_type_impl' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Name_P_Designated_Type_Impl
  (Node : Bare_Name
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Name (Self.Kind) is
               when Ada_Defining_Name .. Ada_Discrete_Subtype_Name | Ada_End_Name .. Ada_Reduce_Attribute_Ref | Ada_Null_Literal .. Ada_Target_Name =>
                  --# property-call-start Name.designated_type_impl
                  Property_Result := Name_P_Designated_Type_Impl
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref =>
                  --# property-call-start AttributeRef.designated_type_impl
                  Property_Result := Attribute_Ref_P_Designated_Type_Impl
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Call_Expr =>
                  --# property-call-start CallExpr.designated_type_impl
                  Property_Result := Call_Expr_P_Designated_Type_Impl
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Dotted_Name =>
                  --# property-call-start DottedName.designated_type_impl
                  Property_Result := Dotted_Name_P_Designated_Type_Impl
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Char_Literal .. Ada_String_Literal =>
                  --# property-call-start BaseId.designated_type_impl
                  Property_Result := Base_Id_P_Designated_Type_Impl
                    (Self
                        , Env
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Name_P_Designated_Type_Impl;
--# end
--# property-start Name.name_designated_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12910
pragma Warnings (Off, "is not referenced");
 function Name_P_Name_Designated_Type
  (Node : Bare_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_962 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_1 : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_2 : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result_1 : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_962 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Bound_Env);
            end Finalizer_Scope_962;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7164 '<bind at ast.py:12911>' Dyn_Var_Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12911
--# expr-start 7160 '<FieldAccess for node_env at ast.py:12911>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12911
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Node_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 7160
Bound_Env := Fld; Inc_Ref (Bound_Env);
--# expr-start 7163 '<bind at ast.py:12912>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12912
--# expr-start 7161 '<FieldAccess for origin_node at ast.py:12912>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12912
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_1 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 7161
Bound_Origin := Fld_1; 
--# expr-start 7162 '<FieldAccess for designated_type_impl at ast.py:12912>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12912
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.designated_type_impl'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_Designated_Type_Impl (Node => Ent.Node, Env => Bound_Env, Origin => Bound_Origin, E_Info => Ent.Info);
--# end
--# expr-done 7162
Dyn_Var_Bind_Result := Fld_2; 
--# expr-done 7163
Dyn_Var_Bind_Result_1 := Dyn_Var_Bind_Result; 
--# expr-done 7164
      Property_Result := Dyn_Var_Bind_Result_1;
   --# end
      Finalizer_Scope_962;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_962;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Name_Designated_Type;
--# end
--# property-start Name.is_static_subtype /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12924
pragma Warnings (Off, "is not referenced");
 function Name_P_Is_Static_Subtype
  (Node : Bare_Name
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7167 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7166 '<FieldAccess for is_static_decl at ast.py:12928>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12928
--# expr-start 7165 '<FieldAccess for name_designated_type at ast.py:12928>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12928
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 7165
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_static_decl'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Static_Decl (Node => Fld.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld.Info);
--# end
--# expr-done 7166
Let_Result := Fld_1; 
--# end
--# expr-done 7167
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Is_Static_Subtype;
--# end
--# property-start Name.name_designated_type_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12931
pragma Warnings (Off, "is not referenced");
 function Name_P_Name_Designated_Type_Env
  (Node : Bare_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_964 with Inline_Always;
            procedure Finalizer_Scope_2714 with Inline_Always;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_964 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_964;
            procedure Finalizer_Scope_2714 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2714;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Name_Name_P_Name_Designated_Type_Env,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Name,
                                As_Bare_Name => Self);
               Mmz_K.Items (2) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Lexical_Env;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 7170 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7168 '<FieldAccess for name_designated_type at ast.py:12932>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12932
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 7168
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 7169 '<FieldAccess for primitives_env at ast.py:12932>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12932
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.primitives_env'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Primitives_Env (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 7169
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2714;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 7170
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_964;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Lexical_Env,
                        As_Lexical_Env => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_964;
               Finalizer_Scope_2714;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Name_Designated_Type_Env;
--# end
--# property-start Name.referenced_unit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12935
pragma Warnings (Off, "is not referenced");
 function Name_P_Referenced_Unit
  (Node : Bare_Name
      ; Kind : Analysis_Unit_Kind
      ; Not_Found_Is_Error : Boolean
         := True
  )
   return Internal_Unit
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind kind Kind
   --# bind not_found_is_error Not_Found_Is_Error
   Property_Result : Internal_Unit;
            procedure Finalizer_Scope_965 with Inline_Always;
      Fld : Symbol_Type_Array_Access;
Fld_1 : Internal_Unit;
Let_Result : Internal_Unit;
            procedure Finalizer_Scope_965 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_965;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 7173 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7172 '<FieldAccess for get_unit at ast.py:12941>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12941
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7171 '<FieldAccess for as_symbol_array at ast.py:12942>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12942
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.as_symbol_array'
Fld := Libadalang.Implementation.Dispatcher_Name_P_As_Symbol_Array (Node => Self);
--# end
--# expr-done 7171
Fld_1 := Libadalang.Implementation.Extensions.Ada_Node_P_Get_Unit (Node => Self, Name => Fld, Kind => Kind, Load_If_Needed => True, Not_Found_Is_Error => Not_Found_Is_Error);
--# expr-done 7172
Let_Result := Fld_1; 
--# end
--# expr-done 7173
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_965;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_965;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Referenced_Unit;
--# end
--# property-start Name.matches /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12946
pragma Warnings (Off, "is not referenced");
 function Name_P_Matches
  (Node : Bare_Name
      ; N : Bare_Name
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind n N
   Property_Result : Boolean;
      Match_0 : Bare_Synthetic_Identifier;
Match_1 : Bare_Identifier;
Match_2 : Bare_String_Literal;
Match_3 : Bare_Dotted_Name;
Match_4 : Bare_Defining_Name;
Match_5 : Bare_Name;
Var_Expr : Bare_Identifier;
Var_Expr_1 : Bare_Identifier;
Var_Expr_2 : Bare_String_Literal;
Var_Expr_3 : Bare_Dotted_Name;
Cast_Expr : Bare_Name;
Cast_Result : Bare_Identifier;
Fld : Symbol_Type;
Fld_1 : Symbol_Type;
Is_Equal : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Cast_Expr_1 : Bare_Name;
Cast_Result_1 : Bare_Identifier;
Fld_2 : Symbol_Type;
Fld_3 : Symbol_Type;
Is_Equal_1 : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
Cast_Expr_2 : Bare_Name;
Cast_Result_2 : Bare_String_Literal;
Fld_4 : Symbol_Type;
Fld_5 : Symbol_Type;
Is_Equal_2 : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
Cast_Expr_3 : Bare_Name;
Cast_Result_3 : Bare_Dotted_Name;
Fld_6 : Bare_Name;
Fld_7 : Bare_Name;
Fld_8 : Boolean;
Fld_9 : Bare_Base_Id;
Fld_10 : Bare_Base_Id;
Cast_Result_4 : Bare_Name;
Fld_11 : Boolean;
If_Result : Boolean;
Scope_Result_3 : Boolean;
Result_Var_3 : Boolean;
Fld_12 : Bare_Name;
Fld_13 : Boolean;
Match_Prefix : Bare_Name;
Cast_Result_5 : Bare_Synthetic_Identifier;
Let_Result : Boolean;
Scope_Result_4 : Boolean;
Cast_Result_6 : Bare_Identifier;
Let_Result_1 : Boolean;
Scope_Result_5 : Boolean;
Cast_Result_7 : Bare_String_Literal;
Let_Result_2 : Boolean;
Scope_Result_6 : Boolean;
Cast_Result_8 : Bare_Dotted_Name;
Let_Result_3 : Boolean;
Scope_Result_7 : Boolean;
Cast_Result_9 : Bare_Defining_Name;
Let_Result_4 : Boolean;
Scope_Result_8 : Boolean;
Cast_Result_10 : Bare_Name;
Let_Result_5 : Boolean;
Scope_Result_9 : Boolean;
Match_Result : Boolean;
Let_Result_6 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 7196 '<Block at ???>' Let_Result_6 None
--# scope-start
--# expr-start 7195 '<Match at ast.py:12953>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12953
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Self; 
case Ada_Name (Match_Prefix.Kind) is
         when Ada_Synthetic_Identifier =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Match_Prefix;
Match_0 := Cast_Result_5; 
--# bind id Match_0
--# expr-start 7174 '<Cast to Identifier at ast.py:12954>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12954
Cast_Expr := N; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_Identifier_Range
   then
      Cast_Result := Cast_Expr;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 7174
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind other_id Var_Expr
--# expr-start 7177 '<Eq at ast.py:12961>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12961
--# expr-start 7175 '<FieldAccess for sym at ast.py:12961>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12961
      if Match_0 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Match_0.Synthetic_Identifier_Sym;
--# expr-done 7175
--# expr-start 7176 '<FieldAccess for sym at ast.py:12961>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12961
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTokNode.sym
Fld_1 := Libadalang.Implementation.Single_Tok_Node_P_Sym (Node => Var_Expr);
--# end
--# expr-done 7176
Is_Equal := Fld = Fld_1; 
--# expr-done 7177
Scope_Result := Is_Equal;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Let_Result := Result_Var; 
--# end
Scope_Result_4 := Let_Result;
   --# end
            Match_Result := Scope_Result_4; 
         when Ada_Identifier =>
   --# scope-start
--# scope-start
      Cast_Result_6 := Match_Prefix;
Match_1 := Cast_Result_6; 
--# bind id Match_1
--# expr-start 7178 '<Cast to Identifier at ast.py:12963>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12963
Cast_Expr_1 := N; 
   if Cast_Expr_1 = null
      or else Cast_Expr_1.Kind in Ada_Identifier_Range
   then
      Cast_Result_1 := Cast_Expr_1;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 7178
Var_Expr_1 := Cast_Result_1;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind other_id Var_Expr_1
--# expr-start 7181 '<Eq at ast.py:12964>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12964
--# expr-start 7179 '<FieldAccess for sym at ast.py:12964>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12964
      if Match_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTokNode.sym
Fld_2 := Libadalang.Implementation.Single_Tok_Node_P_Sym (Node => Match_1);
--# end
--# expr-done 7179
--# expr-start 7180 '<FieldAccess for sym at ast.py:12964>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12964
      if Var_Expr_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTokNode.sym
Fld_3 := Libadalang.Implementation.Single_Tok_Node_P_Sym (Node => Var_Expr_1);
--# end
--# expr-done 7180
Is_Equal_1 := Fld_2 = Fld_3; 
--# expr-done 7181
Scope_Result_1 := Is_Equal_1;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
Let_Result_1 := Result_Var_1; 
--# end
Scope_Result_5 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_5; 
         when Ada_String_Literal =>
   --# scope-start
--# scope-start
      Cast_Result_7 := Match_Prefix;
Match_2 := Cast_Result_7; 
--# bind sl Match_2
--# expr-start 7182 '<Cast to StringLiteral at ast.py:12966>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12966
Cast_Expr_2 := N; 
   if Cast_Expr_2 = null
      or else Cast_Expr_2.Kind in Ada_String_Literal_Range
   then
      Cast_Result_2 := Cast_Expr_2;
   else
         Cast_Result_2 := No_Bare_Ada_Node;
   end if;
--# expr-done 7182
Var_Expr_2 := Cast_Result_2;
if Var_Expr_2 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind other_sl Var_Expr_2
--# expr-start 7185 '<Eq at ast.py:12967>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12967
--# expr-start 7183 '<FieldAccess for sym at ast.py:12967>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12967
      if Match_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTokNode.sym
Fld_4 := Libadalang.Implementation.Single_Tok_Node_P_Sym (Node => Match_2);
--# end
--# expr-done 7183
--# expr-start 7184 '<FieldAccess for sym at ast.py:12967>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12967
      if Var_Expr_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTokNode.sym
Fld_5 := Libadalang.Implementation.Single_Tok_Node_P_Sym (Node => Var_Expr_2);
--# end
--# expr-done 7184
Is_Equal_2 := Fld_4 = Fld_5; 
--# expr-done 7185
Scope_Result_2 := Is_Equal_2;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := False;
end if;
Let_Result_2 := Result_Var_2; 
--# end
Scope_Result_6 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_6; 
         when Ada_Dotted_Name =>
   --# scope-start
--# scope-start
      Cast_Result_8 := Match_Prefix;
Match_3 := Cast_Result_8; 
--# bind dn Match_3
--# expr-start 7186 '<Cast to DottedName at ast.py:12969>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12969
Cast_Expr_3 := N; 
   if Cast_Expr_3 = null
      or else Cast_Expr_3.Kind in Ada_Dotted_Name_Range
   then
      Cast_Result_3 := Cast_Expr_3;
   else
         Cast_Result_3 := No_Bare_Ada_Node;
   end if;
--# expr-done 7186
Var_Expr_3 := Cast_Result_3;
if Var_Expr_3 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind sn Var_Expr_3
--# expr-start 7189 '<FieldAccess for matches at ast.py:12971>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12971
--# expr-start 7187 '<FieldAccess for prefix at ast.py:12971>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12971
      if Var_Expr_3 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Var_Expr_3.Dotted_Name_F_Prefix;
--# expr-done 7187
      if Fld_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7188 '<FieldAccess for prefix at ast.py:12971>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12971
      if Match_3 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Match_3.Dotted_Name_F_Prefix;
--# expr-done 7188
--# property-call-start Name.matches
Fld_8 := Libadalang.Implementation.Name_P_Matches (Node => Fld_6, N => Fld_7);
--# end
--# expr-done 7189
if Fld_8 then
   --# expr-start 7192 '<FieldAccess for matches at ast.py:12972>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12972
--# expr-start 7190 '<FieldAccess for suffix at ast.py:12972>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12972
      if Var_Expr_3 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Var_Expr_3.Dotted_Name_F_Suffix;
--# expr-done 7190
      if Fld_9 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7191 '<FieldAccess for suffix at ast.py:12972>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12972
      if Match_3 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Match_3.Dotted_Name_F_Suffix;
--# expr-done 7191
      Cast_Result_4 := Fld_10;
--# property-call-start Name.matches
Fld_11 := Libadalang.Implementation.Name_P_Matches (Node => Fld_9, N => Cast_Result_4);
--# end
--# expr-done 7192
   If_Result := Fld_11;
else
   If_Result := False;
end if;
Scope_Result_3 := If_Result;
   --# end
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := False;
end if;
Let_Result_3 := Result_Var_3; 
--# end
Scope_Result_7 := Let_Result_3;
   --# end
            Match_Result := Scope_Result_7; 
         when Ada_Defining_Name .. Ada_Synthetic_Defining_Name =>
   --# scope-start
--# scope-start
      Cast_Result_9 := Match_Prefix;
Match_4 := Cast_Result_9; 
--# bind di Match_4
--# expr-start 7194 '<FieldAccess for matches at ast.py:12975>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12975
      if N = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7193 '<FieldAccess for name at ast.py:12975>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12975
      if Match_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Match_4.Defining_Name_F_Name;
--# expr-done 7193
--# property-call-start Name.matches
Fld_13 := Libadalang.Implementation.Name_P_Matches (Node => N, N => Fld_12);
--# end
--# expr-done 7194
Let_Result_4 := Fld_13; 
--# end
Scope_Result_8 := Let_Result_4;
   --# end
            Match_Result := Scope_Result_8; 
         when Ada_Attribute_Ref .. Ada_Call_Expr | Ada_Discrete_Subtype_Name | Ada_End_Name .. Ada_Char_Literal | Ada_Op_Abs .. Ada_Op_Xor | Ada_Null_Literal .. Ada_Real_Literal | Ada_Target_Name =>
   --# scope-start
--# scope-start
      Cast_Result_10 := Match_Prefix;
Match_5 := Cast_Result_10; 
--# bind _ Match_5
Let_Result_5 := False; 
--# end
Scope_Result_9 := Let_Result_5;
   --# end
            Match_Result := Scope_Result_9; 
end case;
--# expr-done 7195
Let_Result_6 := Match_Result; 
--# end
--# expr-done 7196
      Property_Result := Let_Result_6;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Matches;
--# end
--# property-start Name.name_matches /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12980
pragma Warnings (Off, "is not referenced");
 function Name_P_Name_Matches
  (Node : Bare_Name
      ; N : Internal_Entity_Name
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind n N
   Property_Result : Boolean;
      Fld : Bare_Name;
Fld_1 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 7199 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7198 '<FieldAccess for matches at ast.py:12987>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12987
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7197 '<FieldAccess for node at ast.py:12987>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12987
Fld := N.Node;
--# expr-done 7197
--# property-call-start Name.matches
Fld_1 := Libadalang.Implementation.Name_P_Matches (Node => Self, N => Fld);
--# end
--# expr-done 7198
Let_Result := Fld_1; 
--# end
--# expr-done 7199
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Name_Matches;
--# end
--# property-start Name.use_package_name_designated_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12990
pragma Warnings (Off, "is not referenced");
 function Name_P_Use_Package_Name_Designated_Env
  (Node : Bare_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_968 with Inline_Always;
      Fld : Internal_Entity;
Fld_1 : Internal_Entity;
Cast_Result : Internal_Entity_Use_Package_Clause;
Fld_2 : Integer;
Fld_3 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_968 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_968;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7205 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7204 '<FieldAccess for designated_env at ast.py:12996>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12996
--# expr-start 7202 '<Cast to UsePackageClause at ast.py:12996>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12996
--# expr-start 7201 '<parent at ast.py:12996>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12996
--# expr-start 7200 '<parent at ast.py:12996>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12996
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 7200
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Parent (Node => Fld.Node, E_Info => Fld.Info);
--# expr-done 7201
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Use_Package_Clause_Range
   then
      Cast_Result := Create_Internal_Entity_Use_Package_Clause
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 7202
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7203 '<FieldAccess for child_index at ast.py:12997>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12997
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Child_Index (Node => Self);
--# expr-done 7203
--# property-call-start UsePackageClause.designated_env
Fld_3 := Libadalang.Implementation.Use_Package_Clause_P_Designated_Env (Node => Cast_Result.Node, Index => Fld_2, E_Info => Cast_Result.Info);
--# end
--# expr-done 7204
Let_Result := Fld_3; Inc_Ref (Let_Result);
--# end
--# expr-done 7205
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_968;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_968;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Use_Package_Name_Designated_Env;
--# end
--# property-start '[dispatcher]Name.relative_name' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Name_P_Relative_Name
  (Node : Bare_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Single_Tok_Node
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Single_Tok_Node;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Name (Self.Kind) is
               when Ada_Discrete_Subtype_Name | Ada_Reduce_Attribute_Ref | Ada_Synthetic_Identifier =>
                  --# property-call-start Name.relative_name
                  Property_Result := Name_P_Relative_Name
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref =>
                  --# property-call-start AttributeRef.relative_name
                  Property_Result := Attribute_Ref_P_Relative_Name
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Call_Expr =>
                  --# property-call-start CallExpr.relative_name
                  Property_Result := Call_Expr_P_Relative_Name
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Defining_Name .. Ada_Synthetic_Defining_Name =>
                  --# property-call-start DefiningName.relative_name
                  Property_Result := Defining_Name_P_Relative_Name
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Dotted_Name =>
                  --# property-call-start DottedName.relative_name
                  Property_Result := Dotted_Name_P_Relative_Name
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_End_Name =>
                  --# property-call-start EndName.relative_name
                  Property_Result := End_Name_P_Relative_Name
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Explicit_Deref =>
                  --# property-call-start ExplicitDeref.relative_name
                  Property_Result := Explicit_Deref_P_Relative_Name
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Qual_Expr =>
                  --# property-call-start QualExpr.relative_name
                  Property_Result := Qual_Expr_P_Relative_Name
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Char_Literal .. Ada_Real_Literal =>
                  --# property-call-start SingleTokNode.relative_name
                  Property_Result := Single_Tok_Node_P_Relative_Name
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Target_Name =>
                  --# property-call-start TargetName.relative_name
                  Property_Result := Target_Name_P_Relative_Name
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Name_P_Relative_Name;
--# end
--# property-start '[dispatcher]Name.name_symbol' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Name_P_Name_Symbol
  (Node : Bare_Name
  )
   return Symbol_Type
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Name (Self.Kind) is
               when Ada_Attribute_Ref .. Ada_Defining_Name | Ada_Discrete_Subtype_Name .. Ada_Identifier | Ada_String_Literal .. Ada_Real_Literal | Ada_Target_Name =>
                  --# property-call-start Name.name_symbol
                  Property_Result := Name_P_Name_Symbol
                    (Self
                    );
                  --# end
               when Ada_Synthetic_Defining_Name =>
                  --# property-call-start SyntheticDefiningName.name_symbol
                  Property_Result := Synthetic_Defining_Name_P_Name_Symbol
                    (Self
                    );
                  --# end
               when Ada_Op_Abs .. Ada_Op_Xor =>
                  --# property-call-start Op.name_symbol
                  Property_Result := Op_P_Name_Symbol
                    (Self
                    );
                  --# end
               when Ada_Synthetic_Identifier =>
                  --# property-call-start SyntheticIdentifier.name_symbol
                  Property_Result := Synthetic_Identifier_P_Name_Symbol
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Name_P_Name_Symbol;
--# end
--# property-start '[dispatcher]Name.base_name' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Name_P_Base_Name
  (Node : Bare_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Name
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Name (Self.Kind) is
               when Ada_Attribute_Ref .. Ada_Discrete_Subtype_Name | Ada_End_Name .. Ada_Target_Name =>
                  --# property-call-start Name.base_name
                  Property_Result := Name_P_Base_Name
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Dotted_Name =>
                  --# property-call-start DottedName.base_name
                  Property_Result := Dotted_Name_P_Base_Name
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Name_P_Base_Name;
--# end
--# property-start Name.xref_no_overloading /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13018
pragma Warnings (Off, "is not referenced");
 function Name_P_Xref_No_Overloading
  (Node : Bare_Name
      ; Sequential : Boolean
         := True
      ; All_Els : Boolean
         := False
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind sequential Sequential
   --# bind all_els All_Els
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_972 with Inline_Always;
            procedure Finalizer_Scope_2725 with Inline_Always;
            procedure Finalizer_Scope_2726 with Inline_Always;
            procedure Finalizer_Scope_2728 with Inline_Always;
            procedure Finalizer_Scope_2729 with Inline_Always;
      Match_0 : Internal_Entity_Dotted_Name;
Match_1 : Internal_Entity_Base_Id;
Match_2 : Internal_Entity_Attribute_Ref;
Match_3 : Internal_Entity_Name;
Fld : Internal_Entity_Name;
Fld_1 : Logic_Equation;
Fld_2 : Internal_Entity_Name;
Fld_3 : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_4 : Internal_Entity_Base_Id;
Fld_5 : Logic_Equation;
Dyn_Var_Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_6 : Symbol_Type;
Fld_7 : Boolean;
If_Result : Lookup_Kind;
Fld_8 : Bare_Name;
If_Result_1 : Bare_Name;
Cast_Result : Bare_Ada_Node;
Fld_9 : Internal_Entity_Array_Access;
Fld_10 : Logic_Var;
Domain_Equation : Logic_Equation;
Fld_11 : Logic_Var;
Fld_12 : Boolean;
If_Result_2 : Lookup_Kind;
Fld_13 : Bare_Name;
If_Result_3 : Bare_Name;
Cast_Result_1 : Bare_Ada_Node;
Fld_14 : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_15 : Boolean;
If_Result_4 : Boolean;
If_Result_5 : Logic_Equation;
Fld_16 : Internal_Entity_Name;
Fld_17 : Logic_Equation;
Fld_18 : Internal_Entity_Identifier;
Fld_19 : Symbol_Type;
Sym : Symbol_Type;
Is_Equal : Boolean;
Fld_20 : Internal_Entity_Name;
Fld_21 : Logic_Var;
Fld_22 : Logic_Var;
Bind_Result_1 : Logic_Equation;
Fld_23 : Internal_Entity_Identifier;
Fld_24 : Symbol_Type;
Sym_1 : Symbol_Type;
Is_Equal_1 : Boolean;
Fld_25 : Internal_Entity_Name;
Fld_26 : Logic_Var;
Fld_27 : Logic_Var;
Bind_Result_2 : Logic_Equation;
True_Rel : Logic_Equation;
If_Result_6 : Logic_Equation;
If_Result_7 : Logic_Equation;
And_Pred_1 : Logic_Equation;
False_Rel : Logic_Equation;
Match_Prefix : Internal_Entity_Name;
Cast_Result_2 : Internal_Entity_Dotted_Name;
Let_Result : Logic_Equation;
Scope_Result : Logic_Equation;
Cast_Result_3 : Internal_Entity_Base_Id;
Let_Result_1 : Logic_Equation;
Scope_Result_1 : Logic_Equation;
Cast_Result_4 : Internal_Entity_Attribute_Ref;
Let_Result_2 : Logic_Equation;
Scope_Result_2 : Logic_Equation;
Cast_Result_5 : Internal_Entity_Name;
Let_Result_3 : Logic_Equation;
Scope_Result_3 : Logic_Equation;
Match_Result : Logic_Equation;
Let_Result_4 : Logic_Equation;
            procedure Finalizer_Scope_972 is
            begin
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Let_Result_2);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Let_Result_3);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Match_Result);
                     Dec_Ref (Let_Result_4);
            end Finalizer_Scope_972;
            procedure Finalizer_Scope_2725 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_5);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (And_Pred);
            end Finalizer_Scope_2725;
            procedure Finalizer_Scope_2726 is
            begin
                     Dec_Ref (Fld_9);
                     Dec_Ref (Domain_Equation);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (If_Result_5);
            end Finalizer_Scope_2726;
            procedure Finalizer_Scope_2728 is
            begin
                     Dec_Ref (Fld_17);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (True_Rel);
                     Dec_Ref (If_Result_6);
                     Dec_Ref (If_Result_7);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_2728;
            procedure Finalizer_Scope_2729 is
            begin
                     Dec_Ref (False_Rel);
            end Finalizer_Scope_2729;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7252 '<Block at ???>' Let_Result_4 None
--# scope-start
--# expr-start 7251 '<Match at ast.py:13027>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13027
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Name (Match_Prefix.Node.Kind) is
         when Ada_Dotted_Name =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Dotted_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_2; 
--# bind dn Match_0
--# expr-start 7213 '<And at ast.py:13030>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13030
--# expr-start 7207 '<FieldAccess for xref_no_overloading at ast.py:13030>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13030
--# expr-start 7206 '<FieldAccess for prefix at ast.py:13030>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13030
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Match_0.Node.Dotted_Name_F_Prefix, Info => Match_0.Info);
--# expr-done 7206
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_1 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Fld.Node, Sequential => Sequential, All_Els => All_Els, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 7207
--# expr-start 7212 '<bind at ast.py:13031>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13031
--# expr-start 7209 '<FieldAccess for designated_env at ast.py:13032>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13032
--# expr-start 7208 '<FieldAccess for prefix at ast.py:13032>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13032
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Match_0.Node.Dotted_Name_F_Prefix, Info => Match_0.Info);
--# expr-done 7208
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.designated_env'
Fld_3 := Libadalang.Implementation.Dispatcher_Expr_P_Designated_Env (Node => Fld_2.Node, Env => Env, Origin => Origin, No_Visibility => False, E_Info => Fld_2.Info);
--# end
--# expr-done 7209
Bound_Env := Fld_3; Inc_Ref (Bound_Env);
--# expr-start 7211 '<FieldAccess for xref_no_overloading at ast.py:13033>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13033
--# expr-start 7210 '<FieldAccess for suffix at ast.py:13033>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13033
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Base_Id (Node => Match_0.Node.Dotted_Name_F_Suffix, Info => Match_0.Info);
--# expr-done 7210
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_5 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Fld_4.Node, Sequential => Sequential, All_Els => All_Els, Env => Bound_Env, Origin => Origin, E_Info => Fld_4.Info);
--# end
--# expr-done 7211
Dyn_Var_Bind_Result := Fld_5; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 7212
And_Pred := Create_And (Fld_1, Dyn_Var_Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13030") else null)); 
--# expr-done 7213
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
Scope_Result := Let_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2725;
            Match_Result := Scope_Result; Inc_Ref (Match_Result);
         when Ada_Char_Literal .. Ada_String_Literal =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Base_Id
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_3; 
--# bind i Match_1
--# expr-start 7230 '<If at ast.py:13035>' If_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13035
if All_Els then
   --# expr-start 7214 '<FieldAccess for is_suffix at ast.py:13036>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13036
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_suffix
Fld_15 := Libadalang.Implementation.Name_P_Is_Suffix (Node => Self);
--# end
--# expr-done 7214
   If_Result_4 := Fld_15;
else
   If_Result_4 := False;
end if;
if If_Result_4 then
   --# expr-start 7222 '<domain at ast.py:13037>' Domain_Equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13037
--# expr-start 7220 '<FieldAccess for env_get at ast.py:13038>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13038
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7215 '<FieldAccess for name_symbol at ast.py:13040>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13040
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_6 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Match_1.Node);
--# end
--# expr-done 7215
--# expr-start 7217 '<If at ast.py:13042>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13042
--# expr-start 7216 '<FieldAccess for is_prefix at ast.py:13042>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13042
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_prefix
Fld_7 := Libadalang.Implementation.Name_P_Is_Prefix (Node => Self);
--# end
--# expr-done 7216
if Fld_7 then
   If_Result := Recursive;
else
   If_Result := Flat;
end if;
--# expr-done 7217
--# expr-start 7219 '<If at ast.py:13041>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13041
if Sequential then
   --# expr-start 7218 '<FieldAccess for node at ast.py:13041>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13041
Fld_8 := Ent.Node;
--# expr-done 7218
   If_Result_1 := Fld_8;
else
   If_Result_1 := No_Bare_Ada_Node;
end if;
--# expr-done 7219
      Cast_Result := If_Result_1;
--# property-call-start AdaNode.env_get
Fld_9 := Libadalang.Implementation.Ada_Node_P_Env_Get (Node => Self, Env => Env, Symbol => Fld_6, Lookup => If_Result, From_Node => Cast_Result);
--# end
--# expr-done 7220
--# expr-start 7221 '<FieldAccess for ref_var at ast.py:13037>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13037
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_10 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Match_1.Node);
--# end
--# expr-done 7221
Fld_10.Value := No_Entity;
Entity_Vars.Reset (Fld_10);
declare
   Dom : Internal_Entity_Array_Access := Fld_9;
   A   : Entity_Vars.Value_Array (1 .. Length (Dom));
begin
   for J in 0 .. Length (Dom) - 1 loop
      declare
         Item : constant Internal_Entity := Get (Dom, J);
      begin
         A (J + 1) := (Node => Item.Node, Info => Item.Info);
      end;
   end loop;
   Domain_Equation := Solver.Create_Domain
     (Fld_10, A, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13037") else null));
end;
--# expr-done 7222
   If_Result_5 := Domain_Equation;
else
   --# expr-start 7229 '<Bind at ast.py:13045>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13045
--# expr-start 7223 '<FieldAccess for ref_var at ast.py:13046>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13046
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_11 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Match_1.Node);
--# end
--# expr-done 7223
Fld_11.Value := No_Entity;
Entity_Vars.Reset (Fld_11);
--# expr-start 7228 '<FieldAccess for env_get_first_visible at ast.py:13047>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13047
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7225 '<If at ast.py:13050>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13050
--# expr-start 7224 '<FieldAccess for is_prefix at ast.py:13050>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13050
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_prefix
Fld_12 := Libadalang.Implementation.Name_P_Is_Prefix (Node => Self);
--# end
--# expr-done 7224
if Fld_12 then
   If_Result_2 := Recursive;
else
   If_Result_2 := Flat;
end if;
--# expr-done 7225
--# expr-start 7227 '<If at ast.py:13049>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13049
if Sequential then
   --# expr-start 7226 '<FieldAccess for node at ast.py:13049>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13049
Fld_13 := Ent.Node;
--# expr-done 7226
   If_Result_3 := Fld_13;
else
   If_Result_3 := No_Bare_Ada_Node;
end if;
--# expr-done 7227
      Cast_Result_1 := If_Result_3;
--# property-call-start SingleTokNode.env_get_first_visible
Fld_14 := Libadalang.Implementation.Single_Tok_Node_P_Env_Get_First_Visible (Node => Match_1.Node, Lex_Env => Env, Lookup_Type => If_Result_2, From_Node => Cast_Result_1, No_Visibility => False);
--# end
--# expr-done 7228
Bind_Result := Solver.Create_Assign (Fld_11, Fld_14, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13045") else null)); 
--# expr-done 7229
   If_Result_5 := Bind_Result;
end if;
      Inc_Ref (If_Result_5);
--# expr-done 7230
Let_Result_1 := If_Result_5; Inc_Ref (Let_Result_1);
--# end
Scope_Result_1 := Let_Result_1;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2726;
            Match_Result := Scope_Result_1; Inc_Ref (Match_Result);
         when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Attribute_Ref
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_4; 
--# bind ar Match_2
--# expr-start 7250 '<And at ast.py:13059>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13059
--# expr-start 7232 '<FieldAccess for xref_no_overloading at ast.py:13059>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13059
--# expr-start 7231 '<FieldAccess for prefix at ast.py:13059>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13059
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_16 := Create_Internal_Entity_Name (Node => Match_2.Node.Attribute_Ref_F_Prefix, Info => Match_2.Info);
--# expr-done 7231
      if Fld_16.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_17 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Fld_16.Node, Sequential => Sequential, All_Els => All_Els, Env => Env, Origin => Origin, E_Info => Fld_16.Info);
--# end
--# expr-done 7232
--# expr-start 7249 '<Cond at ast.py:13060>' If_Result_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13060
--# expr-start 7236 '<Eq at ast.py:13061>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13061
--# expr-start 7234 '<FieldAccess for sym at ast.py:13061>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13061
--# expr-start 7233 '<FieldAccess for attribute at ast.py:13061>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13061
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Create_Internal_Entity_Identifier (Node => Match_2.Node.Attribute_Ref_F_Attribute, Info => Match_2.Info);
--# expr-done 7233
      if Fld_18.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTokNode.sym
Fld_19 := Libadalang.Implementation.Single_Tok_Node_P_Sym (Node => Fld_18.Node);
--# end
--# expr-done 7234
--# expr-start 7235 '<Symbol Class at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Class); 
--# expr-done 7235
Is_Equal := Fld_19 = Sym; 
--# expr-done 7236
if Is_Equal then
   --# expr-start 7240 '<Bind at ast.py:13062>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13062
--# expr-start 7238 '<FieldAccess for ref_var at ast.py:13062>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13062
--# expr-start 7237 '<FieldAccess for prefix at ast.py:13062>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13062
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_20 := Create_Internal_Entity_Name (Node => Match_2.Node.Attribute_Ref_F_Prefix, Info => Match_2.Info);
--# expr-done 7237
      if Fld_20.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_21 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_20.Node);
--# end
--# expr-done 7238
Fld_21.Value := No_Entity;
Entity_Vars.Reset (Fld_21);
--# expr-start 7239 '<FieldAccess for ref_var at ast.py:13062>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13062
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_22 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Match_2.Node);
--# end
--# expr-done 7239
Fld_22.Value := No_Entity;
Entity_Vars.Reset (Fld_22);
Bind_Result_1 := Solver.Create_Propagate (Fld_21, Fld_22, Logic_Functor_586'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13062") else null)); 
--# expr-done 7240
   If_Result_7 := Bind_Result_1;
else
--# expr-start 7244 '<Eq at ast.py:13065>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13065
--# expr-start 7242 '<FieldAccess for sym at ast.py:13065>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13065
--# expr-start 7241 '<FieldAccess for attribute at ast.py:13065>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13065
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_23 := Create_Internal_Entity_Identifier (Node => Match_2.Node.Attribute_Ref_F_Attribute, Info => Match_2.Info);
--# expr-done 7241
      if Fld_23.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTokNode.sym
Fld_24 := Libadalang.Implementation.Single_Tok_Node_P_Sym (Node => Fld_23.Node);
--# end
--# expr-done 7242
--# expr-start 7243 '<Symbol Base at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Base); 
--# expr-done 7243
Is_Equal_1 := Fld_24 = Sym_1; 
--# expr-done 7244
if Is_Equal_1 then
   --# expr-start 7248 '<Bind at ast.py:13066>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13066
--# expr-start 7246 '<FieldAccess for ref_var at ast.py:13066>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13066
--# expr-start 7245 '<FieldAccess for prefix at ast.py:13066>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13066
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_25 := Create_Internal_Entity_Name (Node => Match_2.Node.Attribute_Ref_F_Prefix, Info => Match_2.Info);
--# expr-done 7245
      if Fld_25.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_26 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_25.Node);
--# end
--# expr-done 7246
Fld_26.Value := No_Entity;
Entity_Vars.Reset (Fld_26);
--# expr-start 7247 '<FieldAccess for ref_var at ast.py:13066>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13066
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_27 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Match_2.Node);
--# end
--# expr-done 7247
Fld_27.Value := No_Entity;
Entity_Vars.Reset (Fld_27);
Bind_Result_2 := Solver.Create_Propagate (Fld_26, Fld_27, Logic_Functor_587'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13066") else null)); 
--# expr-done 7248
   If_Result_6 := Bind_Result_2;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13069") else null)); 
   If_Result_6 := True_Rel;
end if;
      Inc_Ref (If_Result_6);
   If_Result_7 := If_Result_6;
end if;
      Inc_Ref (If_Result_7);
--# expr-done 7249
And_Pred_1 := Create_And (Fld_17, If_Result_7, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13059") else null)); 
--# expr-done 7250
Let_Result_2 := And_Pred_1; Inc_Ref (Let_Result_2);
--# end
Scope_Result_2 := Let_Result_2;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2728;
            Match_Result := Scope_Result_2; Inc_Ref (Match_Result);
         when Ada_Call_Expr .. Ada_Discrete_Subtype_Name | Ada_End_Name .. Ada_Reduce_Attribute_Ref | Ada_Null_Literal .. Ada_Target_Name =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Create_Internal_Entity_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_5; 
--# bind _ Match_3
False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13072") else null)); 
Let_Result_3 := False_Rel; Inc_Ref (Let_Result_3);
--# end
Scope_Result_3 := Let_Result_3;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_2729;
            Match_Result := Scope_Result_3; Inc_Ref (Match_Result);
end case;
--# expr-done 7251
Let_Result_4 := Match_Result; Inc_Ref (Let_Result_4);
--# end
--# expr-done 7252
      Property_Result := Let_Result_4;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_972;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_972;
               Finalizer_Scope_2725;
               Finalizer_Scope_2726;
               Finalizer_Scope_2728;
               Finalizer_Scope_2729;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Xref_No_Overloading;
--# end
--# property-start Name.is_prefix /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13076
pragma Warnings (Off, "is not referenced");
 function Name_P_Is_Prefix
  (Node : Bare_Name
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
Fld_1 : Bare_Ada_Node;
Cast_Result : Bare_Dotted_Name;
Fld_2 : Bare_Name;
Cast_Result_1 : Bare_Name;
Is_Equal : Boolean;
If_Result : Boolean;
Fld_3 : Bare_Ada_Node;
Cast_Result_2 : Bare_Name;
Fld_4 : Boolean;
If_Result_1 : Boolean;
Fld_5 : Bare_Ada_Node;
Is_A_1 : Boolean;
Not_Val : Boolean;
If_Result_2 : Boolean;
Let_Result : Boolean;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Name_Name_P_Is_Prefix,
            Items    => new Mmz_Key_Array (1 ..  1))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Name,
                                As_Bare_Name => Self);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Boolean;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 7265 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7254 '<IsA DottedName at ast.py:13083>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13083
--# expr-start 7253 '<parent at ast.py:13083>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13083
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 7253
Is_A := Fld /= null 
and then Fld.Kind in Ada_Dotted_Name_Range; 
--# expr-done 7254
if Is_A then
   --# expr-start 7258 '<Eq at ast.py:13084>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13084
--# expr-start 7257 '<FieldAccess for prefix at ast.py:13084>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13084
--# expr-start 7256 '<Cast to DottedName at ast.py:13084>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13084
--# expr-start 7255 '<parent at ast.py:13084>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13084
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Self.Parent; 
--# expr-done 7255
   if Fld_1 = null
      or else Fld_1.Kind in Ada_Dotted_Name_Range
   then
      Cast_Result := Fld_1;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 7256
      if Cast_Result = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Cast_Result.Dotted_Name_F_Prefix;
--# expr-done 7257
      Cast_Result_1 := Fld_2;
Is_Equal := Cast_Result_1 = Self; 
--# expr-done 7258
   If_Result := Is_Equal;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 7261 '<FieldAccess for is_prefix at ast.py:13085>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13085
--# expr-start 7260 '<Cast to Name at ast.py:13085>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13085
--# expr-start 7259 '<parent at ast.py:13085>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13085
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Self.Parent; 
--# expr-done 7259
   if Fld_3 = null
      or else Fld_3.Kind in Ada_Name
   then
      Cast_Result_2 := Fld_3;
   else
         Cast_Result_2 := No_Bare_Ada_Node;
   end if;
--# expr-done 7260
      if Cast_Result_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_prefix
Fld_4 := Libadalang.Implementation.Name_P_Is_Prefix (Node => Cast_Result_2);
--# end
--# expr-done 7261
   If_Result_1 := Fld_4;
else
   If_Result_1 := False;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
   --# expr-start 7264 '<Not at ast.py:13086>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13086
--# expr-start 7263 '<IsA DottedName at ast.py:13086>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13086
--# expr-start 7262 '<parent at ast.py:13086>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13086
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Self.Parent; 
--# expr-done 7262
Is_A_1 := Fld_5 /= null 
and then Fld_5.Kind in Ada_Dotted_Name_Range; 
--# expr-done 7263
Not_Val := not (Is_A_1); 
--# expr-done 7264
   If_Result_2 := Not_Val;
end if;
Let_Result := If_Result_2; 
--# end
--# expr-done 7265
      Property_Result := Let_Result;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Boolean,
                        As_Boolean => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Is_Prefix;
--# end
--# property-start Name.is_suffix /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13090
pragma Warnings (Off, "is not referenced");
 function Name_P_Is_Suffix
  (Node : Bare_Name
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
Fld_1 : Bare_Ada_Node;
Cast_Result : Bare_Dotted_Name;
Fld_2 : Bare_Base_Id;
Cast_Result_1 : Bare_Name;
Is_Equal : Boolean;
If_Result : Boolean;
Fld_3 : Bare_Ada_Node;
Cast_Result_2 : Bare_Name;
Fld_4 : Boolean;
If_Result_1 : Boolean;
Fld_5 : Bare_Ada_Node;
Is_A_1 : Boolean;
Not_Val : Boolean;
If_Result_2 : Boolean;
Let_Result : Boolean;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Name_Name_P_Is_Suffix,
            Items    => new Mmz_Key_Array (1 ..  1))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Name,
                                As_Bare_Name => Self);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Boolean;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 7278 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7267 '<IsA DottedName at ast.py:13095>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13095
--# expr-start 7266 '<parent at ast.py:13095>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13095
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 7266
Is_A := Fld /= null 
and then Fld.Kind in Ada_Dotted_Name_Range; 
--# expr-done 7267
if Is_A then
   --# expr-start 7271 '<Eq at ast.py:13096>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13096
--# expr-start 7270 '<FieldAccess for suffix at ast.py:13096>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13096
--# expr-start 7269 '<Cast to DottedName at ast.py:13096>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13096
--# expr-start 7268 '<parent at ast.py:13096>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13096
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Self.Parent; 
--# expr-done 7268
   if Fld_1 = null
      or else Fld_1.Kind in Ada_Dotted_Name_Range
   then
      Cast_Result := Fld_1;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 7269
      if Cast_Result = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Cast_Result.Dotted_Name_F_Suffix;
--# expr-done 7270
      Cast_Result_1 := Fld_2;
Is_Equal := Cast_Result_1 = Self; 
--# expr-done 7271
   If_Result := Is_Equal;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 7274 '<FieldAccess for is_suffix at ast.py:13097>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13097
--# expr-start 7273 '<Cast to Name at ast.py:13097>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13097
--# expr-start 7272 '<parent at ast.py:13097>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13097
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Self.Parent; 
--# expr-done 7272
   if Fld_3 = null
      or else Fld_3.Kind in Ada_Name
   then
      Cast_Result_2 := Fld_3;
   else
         Cast_Result_2 := No_Bare_Ada_Node;
   end if;
--# expr-done 7273
      if Cast_Result_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_suffix
Fld_4 := Libadalang.Implementation.Name_P_Is_Suffix (Node => Cast_Result_2);
--# end
--# expr-done 7274
   If_Result_1 := Fld_4;
else
   If_Result_1 := False;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
   --# expr-start 7277 '<Not at ast.py:13099>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13099
--# expr-start 7276 '<IsA DottedName at ast.py:13099>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13099
--# expr-start 7275 '<parent at ast.py:13099>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13099
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Self.Parent; 
--# expr-done 7275
Is_A_1 := Fld_5 /= null 
and then Fld_5.Kind in Ada_Dotted_Name_Range; 
--# expr-done 7276
Not_Val := not (Is_A_1); 
--# expr-done 7277
   If_Result_2 := Not_Val;
end if;
Let_Result := If_Result_2; 
--# end
--# expr-done 7278
      Property_Result := Let_Result;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Boolean,
                        As_Boolean => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Is_Suffix;
--# end
--# property-start Name.is_operator_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13103
pragma Warnings (Off, "is not referenced");
 function Name_P_Is_Operator_Name
  (Node : Bare_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Symbol_Type;
Sym : Symbol_Type;
Is_Equal : Boolean;
Fld_1 : Symbol_Type;
Sym_1 : Symbol_Type;
Is_Equal_1 : Boolean;
If_Result : Boolean;
Fld_2 : Symbol_Type;
Sym_2 : Symbol_Type;
Is_Equal_2 : Boolean;
If_Result_1 : Boolean;
Fld_3 : Symbol_Type;
Sym_3 : Symbol_Type;
Is_Equal_3 : Boolean;
If_Result_2 : Boolean;
Fld_4 : Symbol_Type;
Sym_4 : Symbol_Type;
Is_Equal_4 : Boolean;
If_Result_3 : Boolean;
Fld_5 : Symbol_Type;
Sym_5 : Symbol_Type;
Is_Equal_5 : Boolean;
If_Result_4 : Boolean;
Fld_6 : Symbol_Type;
Sym_6 : Symbol_Type;
Is_Equal_6 : Boolean;
If_Result_5 : Boolean;
Fld_7 : Symbol_Type;
Sym_7 : Symbol_Type;
Is_Equal_7 : Boolean;
If_Result_6 : Boolean;
Fld_8 : Symbol_Type;
Sym_8 : Symbol_Type;
Is_Equal_8 : Boolean;
If_Result_7 : Boolean;
Fld_9 : Symbol_Type;
Sym_9 : Symbol_Type;
Is_Equal_9 : Boolean;
If_Result_8 : Boolean;
Fld_10 : Symbol_Type;
Sym_10 : Symbol_Type;
Is_Equal_10 : Boolean;
If_Result_9 : Boolean;
Fld_11 : Symbol_Type;
Sym_11 : Symbol_Type;
Is_Equal_11 : Boolean;
If_Result_10 : Boolean;
Fld_12 : Symbol_Type;
Sym_12 : Symbol_Type;
Is_Equal_12 : Boolean;
If_Result_11 : Boolean;
Fld_13 : Symbol_Type;
Sym_13 : Symbol_Type;
Is_Equal_13 : Boolean;
If_Result_12 : Boolean;
Fld_14 : Symbol_Type;
Sym_14 : Symbol_Type;
Is_Equal_14 : Boolean;
If_Result_13 : Boolean;
Fld_15 : Symbol_Type;
Sym_15 : Symbol_Type;
Is_Equal_15 : Boolean;
If_Result_14 : Boolean;
Fld_16 : Symbol_Type;
Sym_16 : Symbol_Type;
Is_Equal_16 : Boolean;
If_Result_15 : Boolean;
Fld_17 : Symbol_Type;
Sym_17 : Symbol_Type;
Is_Equal_17 : Boolean;
If_Result_16 : Boolean;
Fld_18 : Symbol_Type;
Sym_18 : Symbol_Type;
Is_Equal_18 : Boolean;
If_Result_17 : Boolean;
Fld_19 : Symbol_Type;
Sym_19 : Symbol_Type;
Is_Equal_19 : Boolean;
If_Result_18 : Boolean;
Fld_20 : Symbol_Type;
Sym_20 : Symbol_Type;
Is_Equal_20 : Boolean;
If_Result_19 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 7342 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7281 '<Eq at ast.py:13107>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7279 '<FieldAccess for name_symbol at ast.py:13107>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7279
--# expr-start 7280 '<Symbol "=" at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_10); 
--# expr-done 7280
Is_Equal := Fld = Sym; 
--# expr-done 7281
if Is_Equal then
   If_Result := True;
else
   --# expr-start 7284 '<Eq at ast.py:13107>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7282 '<FieldAccess for name_symbol at ast.py:13107>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7282
--# expr-start 7283 '<Symbol "=" at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_10); 
--# expr-done 7283
Is_Equal_1 := Fld_1 = Sym_1; 
--# expr-done 7284
   If_Result := Is_Equal_1;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 7287 '<Eq at ast.py:13107>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7285 '<FieldAccess for name_symbol at ast.py:13107>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7285
--# expr-start 7286 '<Symbol "/=" at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_7); 
--# expr-done 7286
Is_Equal_2 := Fld_2 = Sym_2; 
--# expr-done 7287
   If_Result_1 := Is_Equal_2;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
   --# expr-start 7290 '<Eq at ast.py:13107>' Is_Equal_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7288 '<FieldAccess for name_symbol at ast.py:13107>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7288
--# expr-start 7289 '<Symbol "<" at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_8); 
--# expr-done 7289
Is_Equal_3 := Fld_3 = Sym_3; 
--# expr-done 7290
   If_Result_2 := Is_Equal_3;
end if;
if If_Result_2 then
   If_Result_3 := True;
else
   --# expr-start 7293 '<Eq at ast.py:13107>' Is_Equal_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7291 '<FieldAccess for name_symbol at ast.py:13107>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_4 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7291
--# expr-start 7292 '<Symbol "<=" at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_9); 
--# expr-done 7292
Is_Equal_4 := Fld_4 = Sym_4; 
--# expr-done 7293
   If_Result_3 := Is_Equal_4;
end if;
if If_Result_3 then
   If_Result_4 := True;
else
   --# expr-start 7296 '<Eq at ast.py:13107>' Is_Equal_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7294 '<FieldAccess for name_symbol at ast.py:13107>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_5 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7294
--# expr-start 7295 '<Symbol ">" at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_11); 
--# expr-done 7295
Is_Equal_5 := Fld_5 = Sym_5; 
--# expr-done 7296
   If_Result_4 := Is_Equal_5;
end if;
if If_Result_4 then
   If_Result_5 := True;
else
   --# expr-start 7299 '<Eq at ast.py:13107>' Is_Equal_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7297 '<FieldAccess for name_symbol at ast.py:13107>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_6 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7297
--# expr-start 7298 '<Symbol ">=" at ???>' Sym_6 None
Sym_6 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_12); 
--# expr-done 7298
Is_Equal_6 := Fld_6 = Sym_6; 
--# expr-done 7299
   If_Result_5 := Is_Equal_6;
end if;
if If_Result_5 then
   If_Result_6 := True;
else
   --# expr-start 7302 '<Eq at ast.py:13107>' Is_Equal_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7300 '<FieldAccess for name_symbol at ast.py:13107>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_7 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7300
--# expr-start 7301 '<Symbol "and" at ???>' Sym_7 None
Sym_7 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_And); 
--# expr-done 7301
Is_Equal_7 := Fld_7 = Sym_7; 
--# expr-done 7302
   If_Result_6 := Is_Equal_7;
end if;
if If_Result_6 then
   If_Result_7 := True;
else
   --# expr-start 7305 '<Eq at ast.py:13107>' Is_Equal_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7303 '<FieldAccess for name_symbol at ast.py:13107>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_8 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7303
--# expr-start 7304 '<Symbol "or" at ???>' Sym_8 None
Sym_8 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Or); 
--# expr-done 7304
Is_Equal_8 := Fld_8 = Sym_8; 
--# expr-done 7305
   If_Result_7 := Is_Equal_8;
end if;
if If_Result_7 then
   If_Result_8 := True;
else
   --# expr-start 7308 '<Eq at ast.py:13107>' Is_Equal_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7306 '<FieldAccess for name_symbol at ast.py:13107>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_9 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7306
--# expr-start 7307 '<Symbol "xor" at ???>' Sym_9 None
Sym_9 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Xor); 
--# expr-done 7307
Is_Equal_9 := Fld_9 = Sym_9; 
--# expr-done 7308
   If_Result_8 := Is_Equal_9;
end if;
if If_Result_8 then
   If_Result_9 := True;
else
   --# expr-start 7311 '<Eq at ast.py:13107>' Is_Equal_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7309 '<FieldAccess for name_symbol at ast.py:13107>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_10 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7309
--# expr-start 7310 '<Symbol "abs" at ???>' Sym_10 None
Sym_10 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Abs); 
--# expr-done 7310
Is_Equal_10 := Fld_10 = Sym_10; 
--# expr-done 7311
   If_Result_9 := Is_Equal_10;
end if;
if If_Result_9 then
   If_Result_10 := True;
else
   --# expr-start 7314 '<Eq at ast.py:13107>' Is_Equal_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7312 '<FieldAccess for name_symbol at ast.py:13107>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_11 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7312
--# expr-start 7313 '<Symbol "*" at ???>' Sym_11 None
Sym_11 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_2); 
--# expr-done 7313
Is_Equal_11 := Fld_11 = Sym_11; 
--# expr-done 7314
   If_Result_10 := Is_Equal_11;
end if;
if If_Result_10 then
   If_Result_11 := True;
else
   --# expr-start 7317 '<Eq at ast.py:13107>' Is_Equal_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7315 '<FieldAccess for name_symbol at ast.py:13107>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_12 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7315
--# expr-start 7316 '<Symbol "/" at ???>' Sym_12 None
Sym_12 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_6); 
--# expr-done 7316
Is_Equal_12 := Fld_12 = Sym_12; 
--# expr-done 7317
   If_Result_11 := Is_Equal_12;
end if;
if If_Result_11 then
   If_Result_12 := True;
else
   --# expr-start 7320 '<Eq at ast.py:13107>' Is_Equal_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7318 '<FieldAccess for name_symbol at ast.py:13107>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_13 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7318
--# expr-start 7319 '<Symbol "mod" at ???>' Sym_13 None
Sym_13 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Mod); 
--# expr-done 7319
Is_Equal_13 := Fld_13 = Sym_13; 
--# expr-done 7320
   If_Result_12 := Is_Equal_13;
end if;
if If_Result_12 then
   If_Result_13 := True;
else
   --# expr-start 7323 '<Eq at ast.py:13107>' Is_Equal_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7321 '<FieldAccess for name_symbol at ast.py:13107>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_14 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7321
--# expr-start 7322 '<Symbol "rem" at ???>' Sym_14 None
Sym_14 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Rem); 
--# expr-done 7322
Is_Equal_14 := Fld_14 = Sym_14; 
--# expr-done 7323
   If_Result_13 := Is_Equal_14;
end if;
if If_Result_13 then
   If_Result_14 := True;
else
   --# expr-start 7326 '<Eq at ast.py:13107>' Is_Equal_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7324 '<FieldAccess for name_symbol at ast.py:13107>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_15 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7324
--# expr-start 7325 '<Symbol "+" at ???>' Sym_15 None
Sym_15 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_4); 
--# expr-done 7325
Is_Equal_15 := Fld_15 = Sym_15; 
--# expr-done 7326
   If_Result_14 := Is_Equal_15;
end if;
if If_Result_14 then
   If_Result_15 := True;
else
   --# expr-start 7329 '<Eq at ast.py:13107>' Is_Equal_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7327 '<FieldAccess for name_symbol at ast.py:13107>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_16 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7327
--# expr-start 7328 '<Symbol "-" at ???>' Sym_16 None
Sym_16 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_5); 
--# expr-done 7328
Is_Equal_16 := Fld_16 = Sym_16; 
--# expr-done 7329
   If_Result_15 := Is_Equal_16;
end if;
if If_Result_15 then
   If_Result_16 := True;
else
   --# expr-start 7332 '<Eq at ast.py:13107>' Is_Equal_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7330 '<FieldAccess for name_symbol at ast.py:13107>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_17 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7330
--# expr-start 7331 '<Symbol "&""+" at ???>' Sym_17 None
Sym_17 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_1); 
--# expr-done 7331
Is_Equal_17 := Fld_17 = Sym_17; 
--# expr-done 7332
   If_Result_16 := Is_Equal_17;
end if;
if If_Result_16 then
   If_Result_17 := True;
else
   --# expr-start 7335 '<Eq at ast.py:13107>' Is_Equal_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7333 '<FieldAccess for name_symbol at ast.py:13107>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_18 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7333
--# expr-start 7334 '<Symbol "-" at ???>' Sym_18 None
Sym_18 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_5); 
--# expr-done 7334
Is_Equal_18 := Fld_18 = Sym_18; 
--# expr-done 7335
   If_Result_17 := Is_Equal_18;
end if;
if If_Result_17 then
   If_Result_18 := True;
else
   --# expr-start 7338 '<Eq at ast.py:13107>' Is_Equal_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7336 '<FieldAccess for name_symbol at ast.py:13107>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_19 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7336
--# expr-start 7337 '<Symbol "not" at ???>' Sym_19 None
Sym_19 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Not); 
--# expr-done 7337
Is_Equal_19 := Fld_19 = Sym_19; 
--# expr-done 7338
   If_Result_18 := Is_Equal_19;
end if;
if If_Result_18 then
   If_Result_19 := True;
else
   --# expr-start 7341 '<Eq at ast.py:13107>' Is_Equal_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
--# expr-start 7339 '<FieldAccess for name_symbol at ast.py:13107>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_20 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 7339
--# expr-start 7340 '<Symbol "abs" at ???>' Sym_20 None
Sym_20 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Abs); 
--# expr-done 7340
Is_Equal_20 := Fld_20 = Sym_20; 
--# expr-done 7341
   If_Result_19 := Is_Equal_20;
end if;
Let_Result := If_Result_19; 
--# end
--# expr-done 7342
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Is_Operator_Name;
--# end
--# property-start Name.is_write_reference /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13115
pragma Warnings (Off, "is not referenced");
 function Name_P_Is_Write_Reference
  (Node : Bare_Name
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
            procedure Finalizer_Scope_2733 with Inline_Always;
            procedure Finalizer_Scope_2736 with Inline_Always;
      Match_0 : Internal_Entity_Assign_Stmt;
Match_1 : Internal_Entity_Call_Expr;
Match_2 : Internal_Entity_Dotted_Name;
Match_3 : Internal_Entity_Param_Assoc;
Match_4 : Internal_Entity_Attribute_Ref;
Match_5 : Internal_Entity_Alternatives_List;
Match_6 : Internal_Entity;
Var_Expr : Internal_Entity_Base_Formal_Param_Holder;
Var_Expr_1 : Internal_Entity_Mode;
Var_Expr_2 : Internal_Entity_Param_Spec;
Var_Expr_3 : Internal_Entity_Call_Expr;
Var_Expr_4 : Internal_Entity_Mode;
Var_Expr_5 : Internal_Entity_Param_Spec;
Fld : Internal_Entity;
Fld_1 : Internal_Entity_Name;
Cast_Result : Internal_Entity;
Cast_Expr : Internal_Entity_Name;
Cast_Result_1 : Internal_Entity;
Is_Equiv : Boolean;
Fld_2 : Internal_Entity_Name;
Cast_Result_2 : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Name;
Cast_Result_3 : Internal_Entity;
Is_Equiv_1 : Boolean;
Fld_3 : Boolean;
If_Result : Boolean;
Fld_4 : Boolean;
Fld_5 : Internal_Entity_Name;
Cast_Result_4 : Internal_Entity;
Cast_Expr_2 : Internal_Entity_Name;
Cast_Result_5 : Internal_Entity;
Is_Equiv_2 : Boolean;
Fld_6 : Internal_Entity_Base_Id;
Fld_7 : Internal_Entity_Base_Formal_Param_Holder;
Fld_8 : Internal_Entity_Info;
Fld_9 : Internal_Metadata;
Fld_10 : Boolean;
Fld_11 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Get_Result : Internal_Entity_Base_Formal_Param_Decl;
Cast_Result_6 : Internal_Entity_Param_Spec;
Fld_12 : Internal_Entity_Mode;
Scope_Result : Internal_Entity_Mode;
Result_Var : Internal_Entity_Mode;
Fld_13 : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
If_Result_1 : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Boolean;
Fld_14 : Internal_Entity;
Fld_15 : Internal_Entity;
Cast_Result_7 : Internal_Entity_Call_Expr;
Fld_16 : Boolean;
Fld_17 : Internal_Entity_Defining_Name_Array_Access;
Fld_18 : Internal_Entity_Basic_Decl;
Cast_Result_8 : Internal_Entity_Param_Spec;
Fld_19 : Internal_Entity_Mode;
Scope_Result_3 : Internal_Entity_Mode;
Result_Var_3 : Internal_Entity_Mode;
Fld_20 : Boolean;
Scope_Result_4 : Boolean;
Result_Var_4 : Boolean;
Quantifier_Result : Boolean;
Fld_21 : Internal_Entity;
Fld_22 : Internal_Entity;
Cast_Result_9 : Internal_Entity_Call_Expr;
Fld_23 : Boolean;
Scope_Result_5 : Boolean;
Result_Var_5 : Boolean;
If_Result_4 : Boolean;
Fld_24 : Internal_Entity_Name;
Cast_Result_10 : Internal_Entity;
Cast_Expr_3 : Internal_Entity_Name;
Cast_Result_11 : Internal_Entity;
Is_Equiv_3 : Boolean;
Fld_25 : Boolean;
If_Result_5 : Boolean;
Fld_26 : Internal_Entity;
Is_A : Boolean;
Match_Prefix : Internal_Entity;
Cast_Result_12 : Internal_Entity_Assign_Stmt;
Let_Result : Boolean;
Scope_Result_6 : Boolean;
Cast_Result_13 : Internal_Entity_Call_Expr;
Let_Result_1 : Boolean;
Scope_Result_7 : Boolean;
Cast_Result_14 : Internal_Entity_Dotted_Name;
Let_Result_2 : Boolean;
Scope_Result_8 : Boolean;
Cast_Result_15 : Internal_Entity_Param_Assoc;
Let_Result_3 : Boolean;
Scope_Result_9 : Boolean;
Cast_Result_16 : Internal_Entity_Attribute_Ref;
Let_Result_4 : Boolean;
Scope_Result_10 : Boolean;
Cast_Result_17 : Internal_Entity_Alternatives_List;
Let_Result_5 : Boolean;
Scope_Result_11 : Boolean;
Cast_Result_18 : Internal_Entity;
Let_Result_6 : Boolean;
Scope_Result_12 : Boolean;
Match_Result : Boolean;
Let_Result_7 : Boolean;
            procedure Finalizer_Scope_2733 is
            begin
                     Dec_Ref (Fld_11);
            end Finalizer_Scope_2733;
            procedure Finalizer_Scope_2736 is
            begin
                     Dec_Ref (Fld_17);
            end Finalizer_Scope_2736;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7383 '<Block at ???>' Let_Result_7 None
--# scope-start
--# expr-start 7382 '<Match at ast.py:13132>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13132
--# expr-start 7343 '<parent at ast.py:13132>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13132
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 7343
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Assign_Stmt =>
   --# scope-start
--# scope-start
      Cast_Result_12 := Create_Internal_Entity_Assign_Stmt
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_12; 
--# bind a Match_0
--# expr-start 7345 '<Eq at ast.py:13135>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13135
--# expr-start 7344 '<FieldAccess for dest at ast.py:13135>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13135
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Match_0.Node.Assign_Stmt_F_Dest, Info => Match_0.Info);
--# expr-done 7344
      Cast_Result := Create_Internal_Entity
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
Cast_Expr := Ent; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Is_Equiv := Equivalent (Cast_Result, Cast_Result_1); 
--# expr-done 7345
Let_Result := Is_Equiv; 
--# end
Scope_Result_6 := Let_Result;
   --# end
            Match_Result := Scope_Result_6; 
         when Ada_Call_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_13 := Create_Internal_Entity_Call_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_13; 
--# bind c Match_1
--# expr-start 7347 '<Eq at ast.py:13140>' Is_Equiv_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13140
--# expr-start 7346 '<FieldAccess for name at ast.py:13140>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13140
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Match_1.Node.Call_Expr_F_Name, Info => Match_1.Info);
--# expr-done 7346
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Cast_Expr_1 := Ent; 
      Cast_Result_3 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Is_Equiv_1 := Equivalent (Cast_Result_2, Cast_Result_3); 
--# expr-done 7347
if Is_Equiv_1 then
   --# expr-start 7348 '<FieldAccess for is_write_reference at ast.py:13141>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13141
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_write_reference
Fld_3 := Libadalang.Implementation.Name_P_Is_Write_Reference (Node => Match_1.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Match_1.Info);
--# end
--# expr-done 7348
   If_Result := Fld_3;
else
   If_Result := False;
end if;
Let_Result_1 := If_Result; 
--# end
Scope_Result_7 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_7; 
         when Ada_Dotted_Name =>
   --# scope-start
--# scope-start
      Cast_Result_14 := Create_Internal_Entity_Dotted_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_14; 
--# bind d Match_2
--# expr-start 7349 '<FieldAccess for is_write_reference at ast.py:13151>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13151
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_write_reference
Fld_4 := Libadalang.Implementation.Name_P_Is_Write_Reference (Node => Match_2.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Match_2.Info);
--# end
--# expr-done 7349
if Fld_4 then
   If_Result_3 := True;
else
--# expr-start 7351 '<Eq at ast.py:13154>' Is_Equiv_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13154
--# expr-start 7350 '<FieldAccess for prefix at ast.py:13154>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13154
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Name (Node => Match_2.Node.Dotted_Name_F_Prefix, Info => Match_2.Info);
--# expr-done 7350
      Cast_Result_4 := Create_Internal_Entity
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
Cast_Expr_2 := Ent; 
      Cast_Result_5 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
Is_Equiv_2 := Equivalent (Cast_Result_4, Cast_Result_5); 
--# expr-done 7351
if Is_Equiv_2 then
--# expr-start 7353 '<FieldAccess for called_subp_spec at ast.py:13154>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13154
--# expr-start 7352 '<FieldAccess for suffix at ast.py:13154>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13154
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Base_Id (Node => Match_2.Node.Dotted_Name_F_Suffix, Info => Match_2.Info);
--# expr-done 7352
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.called_subp_spec
Fld_7 := Libadalang.Implementation.Name_P_Called_Subp_Spec (Node => Fld_6.Node, E_Info => Fld_6.Info);
--# end
--# expr-done 7353
Var_Expr := Fld_7;
if Var_Expr /= No_Entity_Base_Formal_Param_Holder then
   --# scope-start
   --# bind spec Var_Expr
--# expr-start 7356 '<FieldAccess for dottable_subp at ast.py:13156>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13156
--# expr-start 7355 '<FieldAccess for md at ast.py:13156>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13156
--# expr-start 7354 '<FieldAccess for info at ast.py:13156>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13156
Fld_8 := Var_Expr.Info;
--# expr-done 7354
Fld_9 := Fld_8.Md;
--# expr-done 7355
Fld_10 := Fld_9.Dottable_Subp;
--# expr-done 7356
if Fld_10 then
--# expr-start 7359 '<Cast to ParamSpec at ast.py:13157>' Cast_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13157
--# expr-start 7358 '<at at ast.py:13157>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13157
--# expr-start 7357 '<FieldAccess for abstract_formal_params at ast.py:13157>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13157
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseFormalParamHolder.abstract_formal_params'
Fld_11 := Libadalang.Implementation.Dispatcher_Base_Formal_Param_Holder_P_Abstract_Formal_Params (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 7357
Get_Result := Get (Fld_11, 0, True); 
--# expr-done 7358
   if Get_Result.Node = null
      or else Get_Result.Node.Kind in Ada_Param_Spec_Range
   then
      Cast_Result_6 := Create_Internal_Entity_Param_Spec
        (Node => Get_Result.Node,
         Info => Get_Result.Info);
   else
         Cast_Result_6 := No_Entity_Param_Spec;
   end if;
--# expr-done 7359
Var_Expr_2 := Cast_Result_6;
if Var_Expr_2 /= No_Entity_Param_Spec then
   --# scope-start
--# expr-start 7360 '<FieldAccess for mode at ast.py:13157>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13157
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Create_Internal_Entity_Mode (Node => Var_Expr_2.Node.Param_Spec_F_Mode, Info => Var_Expr_2.Info);
--# expr-done 7360
Scope_Result := Fld_12;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Mode;
end if;
Var_Expr_1 := Result_Var;
if Var_Expr_1 /= No_Entity_Mode then
   --# scope-start
--# expr-start 7361 '<FieldAccess for is_writable at ast.py:13157>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13157
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Mode.is_writable
Fld_13 := Libadalang.Implementation.Mode_P_Is_Writable (Node => Var_Expr_1.Node);
--# end
--# expr-done 7361
Scope_Result_1 := Fld_13;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
   If_Result_1 := Result_Var_1;
else
   If_Result_1 := False;
end if;
Scope_Result_2 := If_Result_1;
   --# end
      Finalizer_Scope_2733;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := False;
end if;
   If_Result_2 := Result_Var_2;
else
   If_Result_2 := False;
end if;
   If_Result_3 := If_Result_2;
end if;
Let_Result_2 := If_Result_3; 
--# end
Scope_Result_8 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_8; 
         when Ada_Param_Assoc =>
   --# scope-start
--# scope-start
      Cast_Result_15 := Create_Internal_Entity_Param_Assoc
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_15; 
--# bind p Match_3
--# expr-start 7376 '<If at ast.py:13163>' If_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13163
--# expr-start 7364 '<Cast to CallExpr at ast.py:13164>' Cast_Result_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13164
--# expr-start 7363 '<parent at ast.py:13164>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13164
--# expr-start 7362 '<parent at ast.py:13164>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13164
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_21 := Libadalang.Implementation.Parent (Node => Match_3.Node, E_Info => Match_3.Info);
--# expr-done 7362
      if Fld_21.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_22 := Libadalang.Implementation.Parent (Node => Fld_21.Node, E_Info => Fld_21.Info);
--# expr-done 7363
   if Fld_22.Node = null
      or else Fld_22.Node.Kind in Ada_Call_Expr_Range
   then
      Cast_Result_9 := Create_Internal_Entity_Call_Expr
        (Node => Fld_22.Node,
         Info => Fld_22.Info);
   else
         Cast_Result_9 := No_Entity_Call_Expr;
   end if;
--# expr-done 7364
Var_Expr_3 := Cast_Result_9;
if Var_Expr_3 /= No_Entity_Call_Expr then
   --# scope-start
--# expr-start 7365 '<FieldAccess for is_type_conversion at ast.py:13164>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13164
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.is_type_conversion
Fld_23 := Libadalang.Implementation.Call_Expr_P_Is_Type_Conversion (Node => Var_Expr_3.Node, E_Info => Var_Expr_3.Info);
--# end
--# expr-done 7365
Scope_Result_5 := Fld_23;
   --# end
   Result_Var_5 := Scope_Result_5;
else
   Result_Var_5 := False;
end if;
if Result_Var_5 then
   --# expr-start 7369 '<FieldAccess for is_write_reference at ast.py:13165>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13165
--# expr-start 7368 '<Cast to CallExpr at ast.py:13165>' Cast_Result_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13165
--# expr-start 7367 '<parent at ast.py:13165>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13165
--# expr-start 7366 '<parent at ast.py:13165>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13165
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Libadalang.Implementation.Parent (Node => Match_3.Node, E_Info => Match_3.Info);
--# expr-done 7366
      if Fld_14.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Libadalang.Implementation.Parent (Node => Fld_14.Node, E_Info => Fld_14.Info);
--# expr-done 7367
   if Fld_15.Node = null
      or else Fld_15.Node.Kind in Ada_Call_Expr_Range
   then
      Cast_Result_7 := Create_Internal_Entity_Call_Expr
        (Node => Fld_15.Node,
         Info => Fld_15.Info);
   else
         Cast_Result_7 := No_Entity_Call_Expr;
   end if;
--# expr-done 7368
      if Cast_Result_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_write_reference
Fld_16 := Libadalang.Implementation.Name_P_Is_Write_Reference (Node => Cast_Result_7.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Cast_Result_7.Info);
--# end
--# expr-done 7369
   If_Result_4 := Fld_16;
else
   --# expr-start 7375 '<AnyQuantifier at ast.py:13166>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13166
   --# expr-start 7370 '<FieldAccess for get_params at ast.py:13166>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13166
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicAssoc.get_params
Fld_17 := Libadalang.Implementation.Basic_Assoc_P_Get_Params (Node => Match_3.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Match_3.Info);
--# end
--# expr-done 7370
      Quantifier_Result := False;
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Fld_17;
      begin
         for Item_239 of
               Collection.Items
         loop
   --# scope-start
               --# bind m Item_239
--# expr-start 7372 '<Cast to ParamSpec at ast.py:13168>' Cast_Result_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13168
--# expr-start 7371 '<FieldAccess for basic_decl at ast.py:13168>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13168
      if Item_239.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld_18 := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Item_239.Node, E_Info => Item_239.Info);
--# end
--# expr-done 7371
   if Fld_18.Node = null
      or else Fld_18.Node.Kind in Ada_Param_Spec_Range
   then
      Cast_Result_8 := Create_Internal_Entity_Param_Spec
        (Node => Fld_18.Node,
         Info => Fld_18.Info);
   else
         Cast_Result_8 := No_Entity_Param_Spec;
   end if;
--# expr-done 7372
Var_Expr_5 := Cast_Result_8;
if Var_Expr_5 /= No_Entity_Param_Spec then
   --# scope-start
--# expr-start 7373 '<FieldAccess for mode at ast.py:13168>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13168
      if Var_Expr_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Create_Internal_Entity_Mode (Node => Var_Expr_5.Node.Param_Spec_F_Mode, Info => Var_Expr_5.Info);
--# expr-done 7373
Scope_Result_3 := Fld_19;
   --# end
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := No_Entity_Mode;
end if;
Var_Expr_4 := Result_Var_3;
if Var_Expr_4 /= No_Entity_Mode then
   --# scope-start
--# expr-start 7374 '<FieldAccess for is_writable at ast.py:13168>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13168
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Mode.is_writable
Fld_20 := Libadalang.Implementation.Mode_P_Is_Writable (Node => Var_Expr_4.Node);
--# end
--# expr-done 7374
Scope_Result_4 := Fld_20;
   --# end
   Result_Var_4 := Scope_Result_4;
else
   Result_Var_4 := False;
end if;
      Quantifier_Result := Result_Var_4;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
--# expr-done 7375
   If_Result_4 := Quantifier_Result;
end if;
--# expr-done 7376
Let_Result_3 := If_Result_4; 
--# end
Scope_Result_9 := Let_Result_3;
   --# end
      Finalizer_Scope_2736;
            Match_Result := Scope_Result_9; 
         when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref =>
   --# scope-start
--# scope-start
      Cast_Result_16 := Create_Internal_Entity_Attribute_Ref
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_4 := Cast_Result_16; 
--# bind a Match_4
--# expr-start 7378 '<Eq at ast.py:13173>' Is_Equiv_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13173
--# expr-start 7377 '<FieldAccess for prefix at ast.py:13173>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13173
      if Match_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_24 := Create_Internal_Entity_Name (Node => Match_4.Node.Attribute_Ref_F_Prefix, Info => Match_4.Info);
--# expr-done 7377
      Cast_Result_10 := Create_Internal_Entity
        (Node => Fld_24.Node,
         Info => Fld_24.Info);
Cast_Expr_3 := Ent; 
      Cast_Result_11 := Create_Internal_Entity
        (Node => Cast_Expr_3.Node,
         Info => Cast_Expr_3.Info);
Is_Equiv_3 := Equivalent (Cast_Result_10, Cast_Result_11); 
--# expr-done 7378
if Is_Equiv_3 then
   --# expr-start 7379 '<FieldAccess for is_access_attr at ast.py:13173>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13173
      if Match_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.is_access_attr
Fld_25 := Libadalang.Implementation.Attribute_Ref_P_Is_Access_Attr (Node => Match_4.Node);
--# end
--# expr-done 7379
   If_Result_5 := Fld_25;
else
   If_Result_5 := False;
end if;
Let_Result_4 := If_Result_5; 
--# end
Scope_Result_10 := Let_Result_4;
   --# end
            Match_Result := Scope_Result_10; 
         when Ada_Alternatives_List =>
   --# scope-start
--# scope-start
      Cast_Result_17 := Create_Internal_Entity_Alternatives_List
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_5 := Cast_Result_17; 
--# bind l Match_5
--# expr-start 7381 '<IsA AggregateAssoc at ast.py:13175>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13175
--# expr-start 7380 '<parent at ast.py:13175>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13175
      if Match_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_26 := Libadalang.Implementation.Parent (Node => Match_5.Node, E_Info => Match_5.Info);
--# expr-done 7380
Is_A := Fld_26.Node /= null 
and then Fld_26.Node.Kind in Ada_Aggregate_Assoc_Range; 
--# expr-done 7381
Let_Result_5 := Is_A; 
--# end
Scope_Result_11 := Let_Result_5;
   --# end
            Match_Result := Scope_Result_11; 
         when Ada_Abort_Absent .. Ada_Abstract_State_Decl_List | Ada_Constraint_List .. Ada_Iterated_Assoc | Ada_Abstract_State_Decl .. Ada_Membership_Expr | Ada_Defining_Name .. Ada_Discrete_Subtype_Name | Ada_End_Name .. Ada_Abort_Stmt | Ada_Call_Stmt .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_18 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_6 := Cast_Result_18; 
--# bind _ Match_6
Let_Result_6 := False; 
--# end
Scope_Result_12 := Let_Result_6;
   --# end
            Match_Result := Scope_Result_12; 
end case;
--# expr-done 7382
Let_Result_7 := Match_Result; 
--# end
--# expr-done 7383
      Property_Result := Let_Result_7;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_2733;
               Finalizer_Scope_2736;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Is_Write_Reference;
--# end
--# property-start Name.potential_actuals_for_dispatch /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13181
pragma Warnings (Off, "is not referenced");
 function Name_P_Potential_Actuals_For_Dispatch
  (Node : Bare_Name
      ; Spec : Internal_Entity_Base_Subp_Spec
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Expected_Type_For_Expr_Array_Access
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind spec Spec
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Expected_Type_For_Expr_Array_Access;
            procedure Finalizer_Scope_977 with Inline_Always;
            procedure Finalizer_Scope_2744 with Inline_Always;
            procedure Finalizer_Scope_2749 with Inline_Always;
            procedure Finalizer_Scope_2751 with Inline_Always;
      Block_Var : Internal_Expected_Type_For_Expr_Array_Access;
Block_Var_1 : Internal_Expected_Type_For_Expr_Array_Access;
Var_Expr : Boolean;
Var_Expr_1 : Internal_Entity_Dotted_Name;
Match_0 : Internal_Entity_Call_Expr;
Match_1 : Internal_Entity_Dotted_Name;
Match_2 : Internal_Entity_Identifier;
Match_3 : Internal_Entity_Name;
Var_Expr_2 : Internal_Entity_Assoc_List;
Var_Expr_3 : Internal_Entity_Call_Expr;
Fld : Boolean;
Fld_1 : Internal_Entity_Name;
Cast_Result : Internal_Entity_Dotted_Name;
Fld_2 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Dotted_Name;
Match_Prefix : Internal_Entity_Name;
Cast_Result_2 : Internal_Entity_Call_Expr;
Let_Result : Internal_Entity_Dotted_Name;
Scope_Result : Internal_Entity_Dotted_Name;
Cast_Result_3 : Internal_Entity_Dotted_Name;
Let_Result_1 : Internal_Entity_Dotted_Name;
Scope_Result_1 : Internal_Entity_Dotted_Name;
Cast_Result_4 : Internal_Entity_Identifier;
Let_Result_2 : Internal_Entity_Dotted_Name;
Scope_Result_2 : Internal_Entity_Dotted_Name;
Cast_Result_5 : Internal_Entity_Name;
Let_Result_3 : Internal_Entity_Dotted_Name;
Scope_Result_3 : Internal_Entity_Dotted_Name;
Match_Result : Internal_Entity_Dotted_Name;
Fld_3 : Internal_Entity_Defining_Name_Array_Access;
Get_Result : Internal_Entity_Defining_Name;
Fld_4 : Internal_Entity_Base_Formal_Param_Decl;
Fld_5 : Internal_Entity_Type_Expr;
Fld_6 : Internal_Entity_Name;
Cast_Result_6 : Internal_Entity_Expr;
New_Struct : Internal_Expected_Type_For_Expr;
Singleton : Internal_Expected_Type_For_Expr_Array_Access;
Scope_Result_4 : Internal_Expected_Type_For_Expr_Array_Access;
Result_Var : Internal_Expected_Type_For_Expr_Array_Access;
Scope_Result_5 : Internal_Expected_Type_For_Expr_Array_Access;
Result_Var_1 : Internal_Expected_Type_For_Expr_Array_Access;
Cast_Expr : Internal_Entity_Name;
Cast_Result_7 : Internal_Entity_Call_Expr;
Fld_7 : Internal_Entity_Assoc_List;
Scope_Result_6 : Internal_Entity_Assoc_List;
Result_Var_2 : Internal_Entity_Assoc_List;
Fld_8 : Internal_Param_Actual_Array_Access;
Scope_Result_7 : Internal_Param_Actual_Array_Access;
Result_Var_3 : Internal_Param_Actual_Array_Access;
Fld_9 : Internal_Entity_Defining_Name;
Fld_10 : Internal_Entity_Basic_Decl;
Fld_11 : Internal_Entity_Type_Expr;
Fld_12 : Internal_Entity_Expr;
New_Struct_1 : Internal_Expected_Type_For_Expr;
Map_Result : Internal_Expected_Type_For_Expr_Array_Access;
Concat_Result : Internal_Expected_Type_For_Expr_Array_Access;
Let_Result_4 : Internal_Expected_Type_For_Expr_Array_Access;
            procedure Finalizer_Scope_977 is
            begin
                     Dec_Ref (Scope_Result_5);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Scope_Result_7);
                     Dec_Ref (Result_Var_3);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Let_Result_4);
            end Finalizer_Scope_977;
            procedure Finalizer_Scope_2744 is
            begin
                     Dec_Ref (Scope_Result_4);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_2744;
            procedure Finalizer_Scope_2749 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2749;
            procedure Finalizer_Scope_2751 is
            begin
                     Dec_Ref (Fld_8);
            end Finalizer_Scope_2751;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7406 '<Block at ???>' Let_Result_4 None
--# scope-start
--# expr-start 7384 '<FieldAccess for is_dot_call at ast.py:13191>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13191
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_dot_call
Fld := Libadalang.Implementation.Name_P_Is_Dot_Call (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 7384
Var_Expr := Fld;
if Var_Expr /= False then
   --# scope-start
   --# bind _ Var_Expr
--# expr-start 7389 '<Match at ast.py:13191>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13191
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Name (Match_Prefix.Node.Kind) is
         when Ada_Call_Expr =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Call_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_2; 
--# bind c Match_0
--# expr-start 7386 '<Cast to DottedName at ast.py:13192>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13192
--# expr-start 7385 '<FieldAccess for name at ast.py:13192>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13192
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Match_0.Node.Call_Expr_F_Name, Info => Match_0.Info);
--# expr-done 7385
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Dotted_Name_Range
   then
      Cast_Result := Create_Internal_Entity_Dotted_Name
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 7386
Let_Result := Cast_Result; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Dotted_Name =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Dotted_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_3; 
--# bind d Match_1
Let_Result_1 := Match_1; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Identifier =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Identifier
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_4; 
--# bind i Match_2
--# expr-start 7388 '<Cast to DottedName at ast.py:13194>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13194
--# expr-start 7387 '<parent at ast.py:13194>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13194
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Parent (Node => Match_2.Node, E_Info => Match_2.Info);
--# expr-done 7387
   if Fld_2.Node = null
      or else Fld_2.Node.Kind in Ada_Dotted_Name_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Dotted_Name
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 7388
Let_Result_2 := Cast_Result_1; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref | Ada_Defining_Name .. Ada_Discrete_Subtype_Name | Ada_End_Name .. Ada_Char_Literal | Ada_Op_Abs .. Ada_Target_Name =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Create_Internal_Entity_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_5; 
--# bind _ Match_3
Let_Result_3 := No_Entity_Dotted_Name; 
--# end
Scope_Result_3 := Let_Result_3;
   --# end
            Match_Result := Scope_Result_3; 
end case;
--# expr-done 7389
Var_Expr_1 := Match_Result;
if Var_Expr_1 /= No_Entity_Dotted_Name then
   --# scope-start
   --# bind d Var_Expr_1
--# expr-start 7395 '<New ExpectedTypeForExpr at ast.py:13196>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13196
--# expr-start 7393 '<FieldAccess for type_expression at ast.py:13197>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13197
--# expr-start 7392 '<FieldAccess for formal_decl at ast.py:13197>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13197
--# expr-start 7391 '<at at ast.py:13197>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13197
--# expr-start 7390 '<FieldAccess for unpacked_formal_params at ast.py:13197>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13197
      if Spec.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld_3 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => Spec.Node, E_Info => Spec.Info);
--# end
--# expr-done 7390
Get_Result := Get (Fld_3, 0, True); 
--# expr-done 7391
      if Get_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_4 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Get_Result.Node, E_Info => Get_Result.Info);
--# end
--# expr-done 7392
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.type_expression'
Fld_5 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Type_Expression (Node => Fld_4.Node, E_Info => Fld_4.Info);
--# end
--# expr-done 7393
--# expr-start 7394 '<FieldAccess for prefix at ast.py:13199>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13199
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Name (Node => Var_Expr_1.Node.Dotted_Name_F_Prefix, Info => Var_Expr_1.Info);
--# expr-done 7394
      Cast_Result_6 := Create_Internal_Entity_Expr
        (Node => Fld_6.Node,
         Info => Fld_6.Info);
New_Struct := (Expected_Type => Fld_5, Expr => Cast_Result_6); 
--# expr-done 7395
                Singleton := Create_Internal_Expected_Type_For_Expr_Array (Items_Count => 1);
                Singleton.Items (1) := New_Struct;
Scope_Result_4 := Singleton;
   Inc_Ref (Scope_Result_4);
   --# end
      Finalizer_Scope_2749;
   Result_Var := Scope_Result_4;
else
   Result_Var := No_Internal_Expected_Type_For_Expr_Array_Type;
end if;
      Inc_Ref (Result_Var);
Scope_Result_5 := Result_Var;
   Inc_Ref (Scope_Result_5);
   --# end
      Finalizer_Scope_2744;
   Result_Var_1 := Scope_Result_5;
else
   Result_Var_1 := No_Internal_Expected_Type_For_Expr_Array_Type;
end if;
      Inc_Ref (Result_Var_1);
Block_Var := Result_Var_1; Inc_Ref (Block_Var);
--# bind prefix Block_Var
--# expr-start 7404 '<Map at ast.py:13206>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13206
--# expr-start 7396 '<Cast to CallExpr at ast.py:13206>' Cast_Result_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13206
Cast_Expr := Ent; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Call_Expr_Range
   then
      Cast_Result_7 := Create_Internal_Entity_Call_Expr
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result_7 := No_Entity_Call_Expr;
   end if;
--# expr-done 7396
Var_Expr_3 := Cast_Result_7;
if Var_Expr_3 /= No_Entity_Call_Expr then
   --# scope-start
--# expr-start 7397 '<FieldAccess for params at ast.py:13206>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13206
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.params
Fld_7 := Libadalang.Implementation.Call_Expr_P_Params (Node => Var_Expr_3.Node, E_Info => Var_Expr_3.Info);
--# end
--# expr-done 7397
Scope_Result_6 := Fld_7;
   --# end
   Result_Var_2 := Scope_Result_6;
else
   Result_Var_2 := No_Entity_Assoc_List;
end if;
Var_Expr_2 := Result_Var_2;
if Var_Expr_2 /= No_Entity_Assoc_List then
   --# scope-start
--# expr-start 7398 '<FieldAccess for zip_with_params at ast.py:13206>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13206
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AssocList.zip_with_params
Fld_8 := Libadalang.Implementation.Assoc_List_P_Zip_With_Params (Node => Var_Expr_2.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 7398
Scope_Result_7 := Fld_8;
   Inc_Ref (Scope_Result_7);
   --# end
      Finalizer_Scope_2751;
   Result_Var_3 := Scope_Result_7;
else
   Result_Var_3 := No_Internal_Param_Actual_Array_Type;
end if;
      Inc_Ref (Result_Var_3);
      declare
         Map_Result_Vec : Internal_Expected_Type_For_Expr_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Param_Actual_Array_Access := Result_Var_3;
      begin
         for Item_240 of
               Collection.Items
         loop
   --# scope-start
               --# bind pm Item_240
      --# expr-start 7403 '<New ExpectedTypeForExpr at ast.py:13207>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13207
--# expr-start 7401 '<FieldAccess for type_expression at ast.py:13208>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13208
--# expr-start 7400 '<FieldAccess for basic_decl at ast.py:13208>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13208
--# expr-start 7399 '<FieldAccess for param at ast.py:13208>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13208
Fld_9 := Item_240.Param;
--# expr-done 7399
      if Fld_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld_10 := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Fld_9.Node, E_Info => Fld_9.Info);
--# end
--# expr-done 7400
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.type_expression'
Fld_11 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Type_Expression (Node => Fld_10.Node, E_Info => Fld_10.Info);
--# end
--# expr-done 7401
--# expr-start 7402 '<FieldAccess for actual at ast.py:13209>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13209
Fld_12 := Item_240.Actual;
--# expr-done 7402
New_Struct_1 := (Expected_Type => Fld_11, Expr => Fld_12); 
--# expr-done 7403
         declare
            Item_To_Append : constant Internal_Expected_Type_For_Expr := New_Struct_1;
         begin
            Internal_Expected_Type_For_Expr_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Expected_Type_For_Expr_Array
           (Items_Count => Natural (Internal_Expected_Type_For_Expr_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Expected_Type_For_Expr_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Expected_Type_For_Expr_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Expected_Type_For_Expr_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 7404
Block_Var_1 := Map_Result; Inc_Ref (Block_Var_1);
--# bind args Block_Var_1
--# expr-start 7405 '<Concat at ast.py:13214>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13214
Concat_Result := Concat (Block_Var, Block_Var_1); 
--# expr-done 7405
Let_Result_4 := Concat_Result; Inc_Ref (Let_Result_4);
--# end
--# expr-done 7406
      Property_Result := Let_Result_4;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_977;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_977;
               Finalizer_Scope_2744;
               Finalizer_Scope_2749;
               Finalizer_Scope_2751;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Potential_Actuals_For_Dispatch;
--# end
--# property-start Name.is_dispatching_call /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13217
pragma Warnings (Off, "is not referenced");
 function Name_P_Is_Dispatching_Call
  (Node : Bare_Name
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Name;
Fld : Boolean;
Fld_1 : Boolean;
Fld_2 : Internal_Entity_Call_Expr;
Cast_Result : Internal_Entity_Name;
Scope_Result : Internal_Entity_Name;
Result_Var : Internal_Entity_Name;
Fld_3 : Internal_Entity_Basic_Decl;
Fld_4 : Boolean;
If_Result : Boolean;
If_Result_1 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7413 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7407 '<FieldAccess for is_access_call at ast.py:13219>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13219
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_access_call
Fld := Libadalang.Implementation.Name_P_Is_Access_Call (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 7407
if Fld then
   If_Result_1 := True;
else
--# expr-start 7408 '<FieldAccess for is_direct_call at ast.py:13220>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13220
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_direct_call
Fld_1 := Libadalang.Implementation.Name_P_Is_Direct_Call (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 7408
if Fld_1 then
   --# expr-start 7412 '<FieldAccess for is_dispatching_call_impl at ast.py:13221>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13221
--# expr-start 7410 '<Cast to Name.entity at ast.py:13221>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13221
--# expr-start 7409 '<FieldAccess for parent_callexpr at ast.py:13221>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13221
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.parent_callexpr
Fld_2 := Libadalang.Implementation.Name_P_Parent_Callexpr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 7409
      Cast_Result := Create_Internal_Entity_Name
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
--# expr-done 7410
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Name then
   --# scope-start
   --# bind e Var_Expr
Scope_Result := Var_Expr;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := Ent;
end if;
      if Result_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7411 '<FieldAccess for referenced_decl at ast.py:13222>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13222
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld_3 := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 7411
--# property-call-start Expr.is_dispatching_call_impl
Fld_4 := Libadalang.Implementation.Expr_P_Is_Dispatching_Call_Impl (Node => Result_Var.Node, Decl => Fld_3, Imprecise_Fallback => Imprecise_Fallback, E_Info => Result_Var.Info);
--# end
--# expr-done 7412
   If_Result := Fld_4;
else
   If_Result := False;
end if;
   If_Result_1 := If_Result;
end if;
Let_Result := If_Result_1; 
--# end
--# expr-done 7413
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Is_Dispatching_Call;
--# end
--# property-start Name.is_static_call /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13227
pragma Warnings (Off, "is not referenced");
 function Name_P_Is_Static_Call
  (Node : Bare_Name
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Fld : Boolean;
Fld_1 : Boolean;
Not_Val : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7417 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7414 '<FieldAccess for is_call at ast.py:13236>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13236
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_call
Fld := Libadalang.Implementation.Name_P_Is_Call (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 7414
if Fld then
   --# expr-start 7416 '<Not at ast.py:13236>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13236
--# expr-start 7415 '<FieldAccess for is_dispatching_call at ast.py:13236>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13236
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.is_dispatching_call'
Fld_1 := Libadalang.Implementation.Dispatcher_Expr_P_Is_Dispatching_Call (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 7415
Not_Val := not (Fld_1); 
--# expr-done 7416
   If_Result := Not_Val;
else
   If_Result := False;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 7417
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Is_Static_Call;
--# end
--# property-start Name.as_single_tok_node_array /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13239
pragma Warnings (Off, "is not referenced");
 function Name_P_As_Single_Tok_Node_Array
  (Node : Bare_Name
  )
   return Bare_Single_Tok_Node_Array_Access
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Single_Tok_Node_Array_Access;
            procedure Finalizer_Scope_980 with Inline_Always;
            procedure Finalizer_Scope_2754 with Inline_Always;
            procedure Finalizer_Scope_2755 with Inline_Always;
            procedure Finalizer_Scope_2756 with Inline_Always;
      Match_0 : Bare_Defining_Name;
Match_1 : Bare_Single_Tok_Node;
Match_2 : Bare_Dotted_Name;
Match_3 : Bare_Name;
Fld : Bare_Name;
Fld_1 : Bare_Single_Tok_Node_Array_Access;
Singleton : Bare_Single_Tok_Node_Array_Access;
Fld_2 : Bare_Name;
Fld_3 : Bare_Single_Tok_Node_Array_Access;
Fld_4 : Bare_Base_Id;
Fld_5 : Bare_Single_Tok_Node_Array_Access;
Concat_Result : Bare_Single_Tok_Node_Array_Access;
Match_Prefix : Bare_Name;
Cast_Result : Bare_Defining_Name;
Let_Result : Bare_Single_Tok_Node_Array_Access;
Scope_Result : Bare_Single_Tok_Node_Array_Access;
Cast_Result_1 : Bare_Single_Tok_Node;
Let_Result_1 : Bare_Single_Tok_Node_Array_Access;
Scope_Result_1 : Bare_Single_Tok_Node_Array_Access;
Cast_Result_2 : Bare_Dotted_Name;
Let_Result_2 : Bare_Single_Tok_Node_Array_Access;
Scope_Result_2 : Bare_Single_Tok_Node_Array_Access;
Cast_Result_3 : Bare_Name;
Let_Result_3 : Bare_Single_Tok_Node_Array_Access;
Scope_Result_3 : Bare_Single_Tok_Node_Array_Access;
Match_Result : Bare_Single_Tok_Node_Array_Access;
Let_Result_4 : Bare_Single_Tok_Node_Array_Access;
            procedure Finalizer_Scope_980 is
            begin
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Let_Result_2);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Let_Result_3);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Match_Result);
                     Dec_Ref (Let_Result_4);
            end Finalizer_Scope_980;
            procedure Finalizer_Scope_2754 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2754;
            procedure Finalizer_Scope_2755 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2755;
            procedure Finalizer_Scope_2756 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Fld_5);
                     Dec_Ref (Concat_Result);
            end Finalizer_Scope_2756;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 7426 '<Block at ???>' Let_Result_4 None
--# scope-start
--# expr-start 7425 '<Match at ast.py:13246>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13246
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Self; 
case Ada_Name (Match_Prefix.Kind) is
         when Ada_Defining_Name .. Ada_Synthetic_Defining_Name =>
   --# scope-start
--# scope-start
      Cast_Result := Match_Prefix;
Match_0 := Cast_Result; 
--# bind dname Match_0
--# expr-start 7419 '<FieldAccess for as_single_tok_node_array at ast.py:13247>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13247
--# expr-start 7418 '<FieldAccess for name at ast.py:13247>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13247
      if Match_0 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Match_0.Defining_Name_F_Name;
--# expr-done 7418
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.as_single_tok_node_array
Fld_1 := Libadalang.Implementation.Name_P_As_Single_Tok_Node_Array (Node => Fld);
--# end
--# expr-done 7419
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
Scope_Result := Let_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2754;
            Match_Result := Scope_Result; Inc_Ref (Match_Result);
         when Ada_Char_Literal .. Ada_Real_Literal =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Match_Prefix;
Match_1 := Cast_Result_1; 
--# bind tok Match_1
                Singleton := Create_Bare_Single_Tok_Node_Array (Items_Count => 1);
                Singleton.Items (1) := Match_1;
Let_Result_1 := Singleton; Inc_Ref (Let_Result_1);
--# end
Scope_Result_1 := Let_Result_1;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2755;
            Match_Result := Scope_Result_1; Inc_Ref (Match_Result);
         when Ada_Dotted_Name =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Match_Prefix;
Match_2 := Cast_Result_2; 
--# bind dot Match_2
--# expr-start 7424 '<Concat at ast.py:13250>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13250
--# expr-start 7421 '<FieldAccess for as_single_tok_node_array at ast.py:13250>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13250
--# expr-start 7420 '<FieldAccess for prefix at ast.py:13250>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13250
      if Match_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Match_2.Dotted_Name_F_Prefix;
--# expr-done 7420
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.as_single_tok_node_array
Fld_3 := Libadalang.Implementation.Name_P_As_Single_Tok_Node_Array (Node => Fld_2);
--# end
--# expr-done 7421
--# expr-start 7423 '<FieldAccess for as_single_tok_node_array at ast.py:13251>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13251
--# expr-start 7422 '<FieldAccess for suffix at ast.py:13251>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13251
      if Match_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Match_2.Dotted_Name_F_Suffix;
--# expr-done 7422
      if Fld_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.as_single_tok_node_array
Fld_5 := Libadalang.Implementation.Name_P_As_Single_Tok_Node_Array (Node => Fld_4);
--# end
--# expr-done 7423
Concat_Result := Concat (Fld_3, Fld_5); 
--# expr-done 7424
Let_Result_2 := Concat_Result; Inc_Ref (Let_Result_2);
--# end
Scope_Result_2 := Let_Result_2;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2756;
            Match_Result := Scope_Result_2; Inc_Ref (Match_Result);
         when Ada_Attribute_Ref .. Ada_Call_Expr | Ada_Discrete_Subtype_Name | Ada_End_Name .. Ada_Reduce_Attribute_Ref | Ada_Synthetic_Identifier .. Ada_Target_Name =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Match_Prefix;
Match_3 := Cast_Result_3; 
--# bind _ Match_3
Let_Result_3 := raise Property_Error; Inc_Ref (Let_Result_3);
--# end
Scope_Result_3 := Let_Result_3;
   Inc_Ref (Scope_Result_3);
   --# end
            Match_Result := Scope_Result_3; Inc_Ref (Match_Result);
end case;
--# expr-done 7425
Let_Result_4 := Match_Result; Inc_Ref (Let_Result_4);
--# end
--# expr-done 7426
      Property_Result := Let_Result_4;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_980;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_980;
               Finalizer_Scope_2754;
               Finalizer_Scope_2755;
               Finalizer_Scope_2756;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_As_Single_Tok_Node_Array;
--# end
--# property-start '[dispatcher]Name.as_symbol_array' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Name_P_As_Symbol_Array
  (Node : Bare_Name
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Name (Self.Kind) is
               when Ada_Attribute_Ref .. Ada_Defining_Name | Ada_Discrete_Subtype_Name .. Ada_Target_Name =>
                  --# property-call-start Name.as_symbol_array
                  Property_Result := Name_P_As_Symbol_Array
                    (Self
                    );
                  --# end
               when Ada_Synthetic_Defining_Name =>
                  --# property-call-start SyntheticDefiningName.as_symbol_array
                  Property_Result := Synthetic_Defining_Name_P_As_Symbol_Array
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Name_P_As_Symbol_Array;
--# end
--# property-start '[dispatcher]Name.canonical_text' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Name_P_Canonical_Text
  (Node : Bare_Name
  )
   return Symbol_Type
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Name (Self.Kind) is
               when Ada_Attribute_Ref .. Ada_Reduce_Attribute_Ref | Ada_Synthetic_Identifier .. Ada_Target_Name =>
                  --# property-call-start Name.canonical_text
                  Property_Result := Name_P_Canonical_Text
                    (Self
                    );
                  --# end
               when Ada_Char_Literal .. Ada_Real_Literal =>
                  --# property-call-start SingleTokNode.canonical_text
                  Property_Result := Single_Tok_Node_P_Canonical_Text
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Name_P_Canonical_Text;
--# end
--# property-start '[dispatcher]Name.is_constant' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Name_P_Is_Constant
  (Node : Bare_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Name (Self.Kind) is
               when Ada_Defining_Name .. Ada_Discrete_Subtype_Name | Ada_End_Name | Ada_Reduce_Attribute_Ref .. Ada_Char_Literal | Ada_Op_Abs .. Ada_Null_Literal | Ada_Synthetic_Identifier .. Ada_Target_Name =>
                  --# property-call-start Name.is_constant
                  Property_Result := Name_P_Is_Constant
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref =>
                  --# property-call-start AttributeRef.is_constant
                  Property_Result := Attribute_Ref_P_Is_Constant
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Call_Expr =>
                  --# property-call-start CallExpr.is_constant
                  Property_Result := Call_Expr_P_Is_Constant
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Dotted_Name =>
                  --# property-call-start DottedName.is_constant
                  Property_Result := Dotted_Name_P_Is_Constant
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Explicit_Deref =>
                  --# property-call-start ExplicitDeref.is_constant
                  Property_Result := Explicit_Deref_P_Is_Constant
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Qual_Expr =>
                  --# property-call-start QualExpr.is_constant
                  Property_Result := Qual_Expr_P_Is_Constant
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Identifier =>
                  --# property-call-start Identifier.is_constant
                  Property_Result := Identifier_P_Is_Constant
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Int_Literal .. Ada_Real_Literal =>
                  --# property-call-start NumLiteral.is_constant
                  Property_Result := Num_Literal_P_Is_Constant
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Name_P_Is_Constant;
--# end
--# property-start Name.call_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13288
pragma Warnings (Off, "is not referenced");
 function Name_P_Call_Params
  (Node : Bare_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Param_Actual_Array_Access
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Name :=
     Internal_Entity_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Param_Actual_Array_Access;
            procedure Finalizer_Scope_984 with Inline_Always;
            procedure Finalizer_Scope_2762 with Inline_Always;
            procedure Finalizer_Scope_2765 with Inline_Always;
      Offset : Integer;
Aparams : Internal_Entity_Assoc_List;
Dparams : Internal_Param_Actual_Array_Access;
Var_Expr : Internal_Entity_Call_Expr;
Var_Expr_1 : Internal_Entity_Params;
Var_Expr_2 : Internal_Entity_Subp_Spec;
Var_Expr_3 : Internal_Entity_Call_Expr;
Var_Expr_4 : Internal_Entity_Assoc_List;
Fld : Boolean;
If_Result : Integer;
Cast_Expr : Internal_Entity_Name;
Cast_Result : Internal_Entity_Call_Expr;
Fld_1 : Internal_Entity_Assoc_List;
Scope_Result : Internal_Entity_Assoc_List;
Result_Var : Internal_Entity_Assoc_List;
Fld_2 : Internal_Entity_Base_Formal_Param_Holder;
Cast_Result_1 : Internal_Entity_Subp_Spec;
Fld_3 : Internal_Entity_Params;
Scope_Result_1 : Internal_Entity_Params;
Result_Var_1 : Internal_Entity_Params;
Fld_4 : Internal_Entity_Param_Spec_List;
Scope_Result_2 : Internal_Entity_Param_Spec_List;
Result_Var_2 : Internal_Entity_Param_Spec_List;
Fld_5 : Bare_Param_Spec_List;
Fld_6 : Internal_Entity_Info;
As_Entity : Internal_Entity_Param_Spec;
Is_Null : Boolean;
If_Result_1 : Internal_Entity_Param_Spec;
Fld_7 : Internal_Entity_Defining_Name_Array_Access;
Cast_Expr_1 : Internal_Entity_Name;
Cast_Result_2 : Internal_Entity_Call_Expr;
Fld_8 : Internal_Entity_Name;
Cast_Result_3 : Internal_Entity_Dotted_Name;
Fld_9 : Internal_Entity_Name;
Scope_Result_3 : Internal_Entity_Name;
Cast_Expr_2 : Internal_Entity_Name;
Cast_Result_4 : Internal_Entity_Dotted_Name;
Fld_10 : Internal_Entity_Name;
Result_Var_3 : Internal_Entity_Name;
Fld_11 : Internal_Entity_Expr;
Cast_Result_5 : Internal_Entity_Expr;
Fld_12 : Boolean;
Is_Equal : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Internal_Entity_Expr;
New_Struct : Internal_Param_Actual;
Map_Result : Internal_Param_Actual_Array_Access;
Item_242 : Internal_Entity_Param_Spec;
Bare_Item_242 : Bare_Param_Spec;
Index_243 : Integer;
Map_Result_1 : Internal_Param_Actual_Array_Access;
Fld_13 : Internal_Entity_Defining_Name;
Fld_14 : Internal_Entity_Expr;
Fld_15 : Internal_Entity_Defining_Name;
Arith_Result : Integer;
Fld_16 : Internal_Entity_Expr;
Fld_17 : Internal_Entity_Expr;
Fld_18 : Boolean;
Is_Equal_1 : Boolean;
If_Result_4 : Boolean;
If_Result_5 : Internal_Entity_Expr;
New_Struct_1 : Internal_Param_Actual;
Index_246 : Integer;
Map_Result_2 : Internal_Param_Actual_Array_Access;
Scope_Result_4 : Internal_Param_Actual_Array_Access;
Result_Var_4 : Internal_Param_Actual_Array_Access;
Let_Result : Internal_Param_Actual_Array_Access;
Fld_19 : Boolean;
If_Result_6 : Internal_Param_Actual_Array_Access;
Let_Result_1 : Internal_Param_Actual_Array_Access;
            procedure Finalizer_Scope_984 is
            begin
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Dparams);
                     Dec_Ref (Scope_Result_4);
                     Dec_Ref (Result_Var_4);
                     Dec_Ref (Let_Result);
                     Dec_Ref (If_Result_6);
                     Dec_Ref (Let_Result_1);
            end Finalizer_Scope_984;
            procedure Finalizer_Scope_2762 is
            begin
                     Dec_Ref (Fld_7);
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_2762;
            procedure Finalizer_Scope_2765 is
            begin
                     Dec_Ref (Map_Result_2);
            end Finalizer_Scope_2765;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7463 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 7462 '<If at ast.py:13293>' If_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13293
--# expr-start 7427 '<FieldAccess for is_call at ast.py:13294>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13294
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_call
Fld_19 := Libadalang.Implementation.Name_P_Is_Call (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 7427
if Fld_19 then
   --# expr-start 7461 '<Let at ast.py:13296>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13296
--# scope-start
--# expr-start 7429 '<If at ast.py:13297>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13297
--# expr-start 7428 '<FieldAccess for is_dot_call at ast.py:13297>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13297
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_dot_call
Fld := Libadalang.Implementation.Name_P_Is_Dot_Call (Node => Ent.Node, Imprecise_Fallback => False, E_Info => Ent.Info);
--# end
--# expr-done 7428
if Fld then
   If_Result := 1;
else
   If_Result := 0;
end if;
--# expr-done 7429
Offset := If_Result; 
--# bind offset Offset
--# expr-start 7430 '<Cast to CallExpr at ast.py:13299>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13299
Cast_Expr := Ent; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Call_Expr_Range
   then
      Cast_Result := Create_Internal_Entity_Call_Expr
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Call_Expr;
   end if;
--# expr-done 7430
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Call_Expr then
   --# scope-start
--# expr-start 7431 '<FieldAccess for params at ast.py:13299>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13299
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.params
Fld_1 := Libadalang.Implementation.Call_Expr_P_Params (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 7431
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Assoc_List;
end if;
Aparams := Result_Var; 
--# bind aparams Aparams
--# expr-start 7449 '<Mapcat at ast.py:13302>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13302
--# expr-start 7433 '<Cast to SubpSpec at ast.py:13302>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13302
--# expr-start 7432 '<FieldAccess for called_subp_spec at ast.py:13302>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13302
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.called_subp_spec
Fld_2 := Libadalang.Implementation.Name_P_Called_Subp_Spec (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 7432
   if Fld_2.Node = null
      or else Fld_2.Node.Kind in Ada_Subp_Spec_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Subp_Spec
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
   else
         Cast_Result_1 := No_Entity_Subp_Spec;
   end if;
--# expr-done 7433
Var_Expr_2 := Cast_Result_1;
if Var_Expr_2 /= No_Entity_Subp_Spec then
   --# scope-start
--# expr-start 7434 '<FieldAccess for subp_params at ast.py:13302>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13302
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Params (Node => Var_Expr_2.Node.Subp_Spec_F_Subp_Params, Info => Var_Expr_2.Info);
--# expr-done 7434
Scope_Result_1 := Fld_3;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Params;
end if;
Var_Expr_1 := Result_Var_1;
if Var_Expr_1 /= No_Entity_Params then
   --# scope-start
--# expr-start 7435 '<FieldAccess for params at ast.py:13302>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13302
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Param_Spec_List (Node => Var_Expr_1.Node.Params_F_Params, Info => Var_Expr_1.Info);
--# expr-done 7435
Scope_Result_2 := Fld_4;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Entity_Param_Spec_List;
end if;
Fld_5 := Result_Var_2.Node;
      declare
         Map_Result_1_Vec : Internal_Param_Actual_Vectors.Vector;
      begin
      if Fld_5 = null then
      Map_Result_1 := Create_Internal_Param_Actual_Array (0);
      else
         Index_243 := 0;
      declare
         Collection : constant Bare_Param_Spec_List := Fld_5;
      begin
         for Untyped_Item_242 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_242 := Untyped_Item_242; 
Is_Null := Bare_Item_242 = null; 
if Is_Null then
   If_Result_1 := No_Entity_Param_Spec;
else
Fld_6 := Result_Var_2.Info;
As_Entity := (Info => Fld_6, Node => Bare_Item_242); 
   If_Result_1 := As_Entity;
end if;
                  Item_242 := If_Result_1; 
   --# scope-start
               --# bind p Item_242
               --# bind i Index_243
      --# expr-start 7448 '<Map at ast.py:13304>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13304
   --# expr-start 7436 '<FieldAccess for defining_names at ast.py:13304>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13304
      if Item_242.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_names'
Fld_7 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Names (Node => Item_242.Node, E_Info => Item_242.Info);
--# end
--# expr-done 7436
      declare
         Map_Result_Vec : Internal_Param_Actual_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Fld_7;
      begin
         for Item_244 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_244
      --# expr-start 7447 '<New ParamActual at ast.py:13305>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13305
--# expr-start 7446 '<If at ast.py:13307>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13307
--# expr-start 7437 '<FieldAccess for is_dot_call at ast.py:13310>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13310
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_dot_call
Fld_12 := Libadalang.Implementation.Name_P_Is_Dot_Call (Node => Ent.Node, Imprecise_Fallback => False, E_Info => Ent.Info);
--# end
--# expr-done 7437
if Fld_12 then
   --# expr-start 7438 '<Eq at ast.py:13310>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13310
Is_Equal := Index_243 = 0; 
--# expr-done 7438
   If_Result_2 := Is_Equal;
else
   If_Result_2 := False;
end if;
if If_Result_2 then
--# expr-start 7439 '<Cast to CallExpr at ast.py:13312>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13312
Cast_Expr_1 := Ent; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Call_Expr_Range
   then
      Cast_Result_2 := Create_Internal_Entity_Call_Expr
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_2 := No_Entity_Call_Expr;
   end if;
--# expr-done 7439
Var_Expr_3 := Cast_Result_2;
if Var_Expr_3 /= No_Entity_Call_Expr then
   --# scope-start
   --# bind c Var_Expr_3
--# expr-start 7442 '<FieldAccess for prefix at ast.py:13313>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13313
--# expr-start 7441 '<Cast to DottedName at ast.py:13313>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13313
--# expr-start 7440 '<FieldAccess for name at ast.py:13313>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13313
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Name (Node => Var_Expr_3.Node.Call_Expr_F_Name, Info => Var_Expr_3.Info);
--# expr-done 7440
   if Fld_8.Node = null
      or else Fld_8.Node.Kind in Ada_Dotted_Name_Range
   then
      Cast_Result_3 := Create_Internal_Entity_Dotted_Name
        (Node => Fld_8.Node,
         Info => Fld_8.Info);
   else
         Cast_Result_3 := No_Entity_Dotted_Name;
   end if;
--# expr-done 7441
      if Cast_Result_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Name (Node => Cast_Result_3.Node.Dotted_Name_F_Prefix, Info => Cast_Result_3.Info);
--# expr-done 7442
Scope_Result_3 := Fld_9;
   --# end
   Result_Var_3 := Scope_Result_3;
else
   --# expr-start 7444 '<FieldAccess for prefix at ast.py:13314>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13314
--# expr-start 7443 '<Cast to DottedName at ast.py:13314>' Cast_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13314
Cast_Expr_2 := Ent; 
   if Cast_Expr_2.Node = null
      or else Cast_Expr_2.Node.Kind in Ada_Dotted_Name_Range
   then
      Cast_Result_4 := Create_Internal_Entity_Dotted_Name
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
   else
         Cast_Result_4 := No_Entity_Dotted_Name;
   end if;
--# expr-done 7443
      if Cast_Result_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Create_Internal_Entity_Name (Node => Cast_Result_4.Node.Dotted_Name_F_Prefix, Info => Cast_Result_4.Info);
--# expr-done 7444
   Result_Var_3 := Fld_10;
end if;
      Cast_Result_5 := Create_Internal_Entity_Expr
        (Node => Result_Var_3.Node,
         Info => Result_Var_3.Info);
   If_Result_3 := Cast_Result_5;
else
   --# expr-start 7445 '<FieldAccess for default_expr at ast.py:13317>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13317
      if Item_242.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Create_Internal_Entity_Expr (Node => Item_242.Node.Param_Spec_F_Default_Expr, Info => Item_242.Info);
--# expr-done 7445
   If_Result_3 := Fld_11;
end if;
--# expr-done 7446
New_Struct := (Actual => If_Result_3, Param => Item_244); 
--# expr-done 7447
         declare
            Item_To_Append : constant Internal_Param_Actual := New_Struct;
         begin
            Internal_Param_Actual_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Param_Actual_Array
           (Items_Count => Natural (Internal_Param_Actual_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Param_Actual_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Param_Actual_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Param_Actual_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 7448
         for Item_To_Append of
               Map_Result.Items
         loop
            Internal_Param_Actual_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_2762;
               Index_243 := Index_243 + 1;
         end loop;
      end;
      end if;
         Map_Result_1 := Create_Internal_Param_Actual_Array
           (Items_Count => Natural (Internal_Param_Actual_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Param_Actual_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Param_Actual_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Param_Actual_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 7449
Dparams := Map_Result_1; Inc_Ref (Dparams);
--# bind dparams Dparams
Var_Expr_4 := Aparams;
if Var_Expr_4 /= No_Entity_Assoc_List then
   --# scope-start
   --# bind ap Var_Expr_4
--# expr-start 7460 '<Map at ast.py:13326>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13326
      declare
         Map_Result_2_Vec : Internal_Param_Actual_Vectors.Vector;
      begin
         Index_246 := 0;
      declare
         Collection : constant Internal_Param_Actual_Array_Access := Dparams;
      begin
         for Item_245 of
               Collection.Items
         loop
   --# scope-start
               --# bind dp Item_245
               --# bind i Index_246
      --# expr-start 7459 '<New ParamActual at ast.py:13327>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13327
--# expr-start 7457 '<If at ast.py:13331>' If_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13331
--# expr-start 7450 '<FieldAccess for is_dot_call at ast.py:13334>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13334
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_dot_call
Fld_18 := Libadalang.Implementation.Name_P_Is_Dot_Call (Node => Ent.Node, Imprecise_Fallback => False, E_Info => Ent.Info);
--# end
--# expr-done 7450
if Fld_18 then
   --# expr-start 7451 '<Eq at ast.py:13334>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13334
Is_Equal_1 := Index_246 = 0; 
--# expr-done 7451
   If_Result_4 := Is_Equal_1;
else
   If_Result_4 := False;
end if;
if If_Result_4 then
   --# expr-start 7452 '<FieldAccess for actual at ast.py:13336>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13336
Fld_14 := Item_245.Actual;
--# expr-done 7452
   If_Result_5 := Fld_14;
else
   --# expr-start 7456 '<FieldAccess for actual_for_param_at at ast.py:13338>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13338
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7453 '<FieldAccess for param at ast.py:13339>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13339
Fld_15 := Item_245.Param;
--# expr-done 7453
--# expr-start 7454 '<Arithmetic - at ast.py:13339>' Arith_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13339
Arith_Result := (Index_246 - Offset); 
--# expr-done 7454
--# expr-start 7455 '<FieldAccess for actual at ast.py:13339>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13339
Fld_16 := Item_245.Actual;
--# expr-done 7455
--# property-call-start AssocList.actual_for_param_at
Fld_17 := Libadalang.Implementation.Assoc_List_P_Actual_For_Param_At (Node => Var_Expr_4.Node, Param => Fld_15, Pos => Arith_Result, Default_Expr => Fld_16, E_Info => Var_Expr_4.Info);
--# end
--# expr-done 7456
   If_Result_5 := Fld_17;
end if;
--# expr-done 7457
--# expr-start 7458 '<FieldAccess for param at ast.py:13328>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13328
Fld_13 := Item_245.Param;
--# expr-done 7458
New_Struct_1 := (Actual => If_Result_5, Param => Fld_13); 
--# expr-done 7459
         declare
            Item_To_Append : constant Internal_Param_Actual := New_Struct_1;
         begin
            Internal_Param_Actual_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
   --# end
               Index_246 := Index_246 + 1;
         end loop;
      end;
         Map_Result_2 := Create_Internal_Param_Actual_Array
           (Items_Count => Natural (Internal_Param_Actual_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Internal_Param_Actual_Vectors.Get
              (Map_Result_2_Vec,
               I + Internal_Param_Actual_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Internal_Param_Actual_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 7460
Scope_Result_4 := Map_Result_2;
   Inc_Ref (Scope_Result_4);
   --# end
      Finalizer_Scope_2765;
   Result_Var_4 := Scope_Result_4;
else
   Result_Var_4 := Dparams;
end if;
      Inc_Ref (Result_Var_4);
Let_Result := Result_Var_4; Inc_Ref (Let_Result);
--# end
--# expr-done 7461
   If_Result_6 := Let_Result;
else
   If_Result_6 := raise Property_Error with "this name doesn't reference a call expression";
end if;
      Inc_Ref (If_Result_6);
--# expr-done 7462
Let_Result_1 := If_Result_6; Inc_Ref (Let_Result_1);
--# end
--# expr-done 7463
      Property_Result := Let_Result_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_984;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_984;
               Finalizer_Scope_2762;
               Finalizer_Scope_2765;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Call_Params;
--# end
 function Name_P_Parent_Scope
  (Node : Bare_Name
      ; Env : Lexical_Env
  )
   return Lexical_Env
is (raise Property_Error
    with "Property Name.parent_scope not implemented on type "
    & Kind_Name (Node));
 function Name_P_All_Env_Els_Impl
  (Node : Bare_Name
      ; Seq : Boolean
         := True
      ; Seq_From : Bare_Ada_Node
         := No_Bare_Ada_Node
      ; Categories : Ref_Categories
         := (Inherited_Primitives => True, Nocat => True)
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is (raise Property_Error
    with "Property Name.all_env_els_impl not implemented on type "
    & Kind_Name (Node));
--# property-start Name.scope /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12755
pragma Warnings (Off, "is not referenced");
 function Name_P_Scope
  (Node : Bare_Name
      ; Env : Lexical_Env
  )
   return Lexical_Env
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   Property_Result : Lexical_Env;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := Empty_Env;
         Inc_Ref (Property_Result);
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Scope;
--# end
 function Name_P_Ref_Var
  (Node : Bare_Name
  )
   return Logic_Var
is (raise Property_Error
    with "Property Name.ref_var not implemented on type "
    & Kind_Name (Node));
 function Name_P_Subp_Spec_Var
  (Node : Bare_Name
  )
   return Logic_Var
is (raise Property_Error
    with "Property Name.subp_spec_var not implemented on type "
    & Kind_Name (Node));
--# property-start Name.defines_subp_spec_var /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12788
pragma Warnings (Off, "is not referenced");
 function Name_P_Defines_Subp_Spec_Var
  (Node : Bare_Name
  )
   return Boolean
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 7464 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := False; 
--# end
--# expr-done 7464
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Defines_Subp_Spec_Var;
--# end
--# property-start Name.designated_type_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:12899
pragma Warnings (Off, "is not referenced");
 function Name_P_Designated_Type_Impl
  (Node : Bare_Name
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := No_Entity_Base_Type_Decl;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Designated_Type_Impl;
--# end
 function Name_P_Relative_Name
  (Node : Bare_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Single_Tok_Node
is (raise Property_Error
    with "Property Name.relative_name not implemented on type "
    & Kind_Name (Node));
--# property-start Name.name_symbol /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13007
pragma Warnings (Off, "is not referenced");
 function Name_P_Name_Symbol
  (Node : Bare_Name
  )
   return Symbol_Type
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type;
      As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld : Internal_Entity_Single_Tok_Node;
Fld_1 : Bare_Single_Tok_Node;
Sym : Symbol_Type;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 7467 '<GetSymbol at ast.py:13007>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13007
--# expr-start 7466 '<FieldAccess for relative_name at ast.py:13007>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13007
--# expr-start 7465 '<as_bare_entity at ast.py:13007>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13007
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 7465
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.relative_name'
Fld := Libadalang.Implementation.Dispatcher_Name_P_Relative_Name (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 7466
Fld_1 := Fld.Node; 
Sym := Get_Symbol (Fld_1); 
--# expr-done 7467
      Property_Result := Sym;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Name_Symbol;
--# end
--# property-start Name.base_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13009
pragma Warnings (Off, "is not referenced");
 function Name_P_Base_Name
  (Node : Bare_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Name
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := No_Entity_Name;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Base_Name;
--# end
--# property-start Name.as_symbol_array /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13257
pragma Warnings (Off, "is not referenced");
 function Name_P_As_Symbol_Array
  (Node : Bare_Name
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_981 with Inline_Always;
      Fld : Bare_Single_Tok_Node_Array_Access;
Sym : Symbol_Type;
Map_Result : Symbol_Type_Array_Access;
Let_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_981 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_981;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 7471 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7470 '<Map at ast.py:13267>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13267
   --# expr-start 7468 '<FieldAccess for as_single_tok_node_array at ast.py:13267>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13267
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.as_single_tok_node_array
Fld := Libadalang.Implementation.Name_P_As_Single_Tok_Node_Array (Node => Self);
--# end
--# expr-done 7468
      declare
         Map_Result_Vec : Symbol_Type_Vectors.Vector;
      begin
      declare
         Collection : constant Bare_Single_Tok_Node_Array_Access := Fld;
      begin
         for Item_241 of
               Collection.Items
         loop
   --# scope-start
               --# bind t Item_241
      --# expr-start 7469 '<GetSymbol at ast.py:13267>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13267
Sym := Get_Symbol (Item_241); 
--# expr-done 7469
         declare
            Item_To_Append : constant Symbol_Type := Sym;
         begin
            Symbol_Type_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Symbol_Type_Array
           (Items_Count => Natural (Symbol_Type_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Symbol_Type_Vectors.Get
              (Map_Result_Vec,
               I + Symbol_Type_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Symbol_Type_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 7470
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 7471
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_981;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_981;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_As_Symbol_Array;
--# end
--# property-start Name.canonical_text /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13270
pragma Warnings (Off, "is not referenced");
 function Name_P_Canonical_Text
  (Node : Bare_Name
  )
   return Symbol_Type
is
   Self : Bare_Name := Bare_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type;
            procedure Finalizer_Scope_982 with Inline_Always;
      Fld : Symbol_Type_Array_Access;
Str : String_Type;
Fld_1 : String_Type;
Sym : Symbol_Type;
Let_Result : Symbol_Type;
            procedure Finalizer_Scope_982 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Str);
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_982;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7476 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7475 '<to_symbol at ast.py:13277>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13277
--# expr-start 7474 '<FieldAccess for sym_join at ast.py:13277>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13277
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7472 '<FieldAccess for as_symbol_array at ast.py:13277>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13277
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.as_symbol_array'
Fld := Libadalang.Implementation.Dispatcher_Name_P_As_Symbol_Array (Node => Self);
--# end
--# expr-done 7472
--# expr-start 7473 '<String at ast.py:13277>' Str /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13277
Str := Create_String ("."); 
--# expr-done 7473
--# property-call-start AdaNode.sym_join
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sym_Join (Node => Self, Syms => Fld, Sep => Str);
--# end
--# expr-done 7474
Sym := String_To_Symbol (Self.Unit.Context, Fld_1); 
--# expr-done 7475
Let_Result := Sym; 
--# end
--# expr-done 7476
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_982;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_982;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Name_P_Canonical_Text;
--# end
 function Name_P_Is_Constant
  (Node : Bare_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is (raise Property_Error
    with "Property Name.is_constant not implemented on type "
    & Kind_Name (Node));
function Name_P_Name_Designated_Type_Env
  (E : Entity
  ) return Lexical_Env
is
      E_Info : Internal_Entity_Info :=
         Shed_Rebindings (E.Info, Node_Env (E.Node));
   Result : constant Lexical_Env :=
      Name_P_Name_Designated_Type_Env (E.Node, E_Info);
begin
      return Result;
end;
function Name_P_Use_Package_Name_Designated_Env
  (E : Entity
  ) return Lexical_Env
is
      E_Info : Internal_Entity_Info :=
         Shed_Rebindings (E.Info, Node_Env (E.Node));
   Result : constant Lexical_Env :=
      Name_P_Use_Package_Name_Designated_Env (E.Node, E_Info);
begin
      return Result;
end;
   --
   --  Primitives for Bare_Attribute_Ref
   --
      procedure Initialize_Fields_For_Attribute_Ref
        (Self : Bare_Attribute_Ref
         ; Attribute_Ref_F_Prefix : Bare_Name
         ; Attribute_Ref_F_Attribute : Bare_Identifier
         ; Attribute_Ref_F_Args : Bare_Ada_Node
        ) is
      begin
            Initialize_Fields_For_Name
              (Self);
            Self.Attribute_Ref_F_Prefix := Attribute_Ref_F_Prefix;
            Self.Attribute_Ref_F_Attribute := Attribute_Ref_F_Attribute;
            Self.Attribute_Ref_F_Args := Attribute_Ref_F_Args;
      Self.Attribute_Ref_R_Ref_Var := Null_Var_Record;
      end Initialize_Fields_For_Attribute_Ref;
   function Attribute_Ref_F_Prefix
     (Node : Bare_Attribute_Ref) return Bare_Name
   is
   begin
         return Node.Attribute_Ref_F_Prefix;
   end;
   function Attribute_Ref_F_Attribute
     (Node : Bare_Attribute_Ref) return Bare_Identifier
   is
   begin
         return Node.Attribute_Ref_F_Attribute;
   end;
   function Attribute_Ref_F_Args
     (Node : Bare_Attribute_Ref) return Bare_Ada_Node
   is
   begin
         return Node.Attribute_Ref_F_Args;
   end;
--# property-start AttributeRef.ref_var /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17268
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Ref_Var
  (Node : Bare_Attribute_Ref
  )
   return Logic_Var
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Logic_Var;
      Fld : Logic_Var;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 7477 '<FieldAccess for r_ref_var at ast.py:17268>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17268
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Attribute_Ref_R_Ref_Var'Unrestricted_Access;
--# expr-done 7477
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Ref_Var;
--# end
--# property-start AttributeRef.relative_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17271
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Relative_Name
  (Node : Bare_Attribute_Ref
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Single_Tok_Node
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Single_Tok_Node;
      Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Single_Tok_Node;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 7479 '<FieldAccess for relative_name at ast.py:17271>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17271
--# expr-start 7478 '<FieldAccess for prefix at ast.py:17271>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17271
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7478
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.relative_name'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Relative_Name (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 7479
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Relative_Name;
--# end
--# property-start AttributeRef.is_constant /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17273
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Is_Constant
  (Node : Bare_Attribute_Ref
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Is_Constant;
--# end
--# property-start AttributeRef.designated_type_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17275
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Designated_Type_Impl
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Bare_Identifier;
Fld_1 : Symbol_Type;
Sym : Symbol_Type;
Is_Equal : Boolean;
Fld_2 : Internal_Entity_Name;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Internal_Entity_Classwide_Type_Decl;
Scope_Result : Internal_Entity_Classwide_Type_Decl;
Result_Var : Internal_Entity_Classwide_Type_Decl;
Fld_5 : Bare_Identifier;
Fld_6 : Symbol_Type;
Sym_1 : Symbol_Type;
Is_Equal_1 : Boolean;
Fld_7 : Internal_Entity_Name;
Fld_8 : Internal_Entity_Base_Type_Decl;
Fld_9 : Internal_Entity_Discrete_Base_Subtype_Decl;
Cast_Result : Internal_Entity_Base_Type_Decl;
If_Result : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity_Base_Type_Decl;
If_Result_1 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7494 '<Cond at ast.py:17275>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17275
--# expr-start 7483 '<Eq at ast.py:17276>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17276
--# expr-start 7481 '<FieldAccess for sym at ast.py:17276>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17276
--# expr-start 7480 '<FieldAccess for attribute at ast.py:17276>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17276
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Attribute_Ref_F_Attribute;
--# expr-done 7480
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTokNode.sym
Fld_1 := Libadalang.Implementation.Single_Tok_Node_P_Sym (Node => Fld);
--# end
--# expr-done 7481
--# expr-start 7482 '<Symbol Class at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Class); 
--# expr-done 7482
Is_Equal := Fld_1 = Sym; 
--# expr-done 7483
if Is_Equal then
--# expr-start 7485 '<FieldAccess for designated_type_impl at ast.py:17277>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17277
--# expr-start 7484 '<FieldAccess for prefix at ast.py:17277>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17277
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7484
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.designated_type_impl'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Designated_Type_Impl (Node => Fld_2.Node, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 7485
Var_Expr := Fld_3;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 7486 '<FieldAccess for classwide_type at ast.py:17277>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17277
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.classwide_type'
Fld_4 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Classwide_Type (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 7486
Scope_Result := Fld_4;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Classwide_Type_Decl;
end if;
      Cast_Result_1 := Create_Internal_Entity_Base_Type_Decl
        (Node => Result_Var.Node,
         Info => Result_Var.Info);
   If_Result_1 := Cast_Result_1;
else
--# expr-start 7490 '<Eq at ast.py:17279>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17279
--# expr-start 7488 '<FieldAccess for sym at ast.py:17279>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17279
--# expr-start 7487 '<FieldAccess for attribute at ast.py:17279>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17279
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Self.Attribute_Ref_F_Attribute;
--# expr-done 7487
      if Fld_5 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTokNode.sym
Fld_6 := Libadalang.Implementation.Single_Tok_Node_P_Sym (Node => Fld_5);
--# end
--# expr-done 7488
--# expr-start 7489 '<Symbol Base at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Base); 
--# expr-done 7489
Is_Equal_1 := Fld_6 = Sym_1; 
--# expr-done 7490
if Is_Equal_1 then
--# expr-start 7493 '<FieldAccess for scalar_base_subtype at ast.py:17280>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17280
--# expr-start 7492 '<FieldAccess for name_designated_type at ast.py:17280>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17280
--# expr-start 7491 '<FieldAccess for prefix at ast.py:17280>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17280
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7491
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_8 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld_7.Node, E_Info => Fld_7.Info);
--# end
--# expr-done 7492
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.scalar_base_subtype
Fld_9 := Libadalang.Implementation.Base_Type_Decl_P_Scalar_Base_Subtype (Node => Fld_8.Node, E_Info => Fld_8.Info);
--# end
--# expr-done 7493
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld_9.Node,
         Info => Fld_9.Info);
   If_Result := Cast_Result;
else
   If_Result := No_Entity_Base_Type_Decl;
end if;
   If_Result_1 := If_Result;
end if;
--# expr-done 7494
      Property_Result := If_Result_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Designated_Type_Impl;
--# end
--# property-start AttributeRef.args_list /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17286
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Args_List
  (Node : Bare_Attribute_Ref
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Assoc_List
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Assoc_List;
      Block_Var : Internal_Entity_Assoc_List;
Var_Expr : Internal_Entity;
Fld : Internal_Entity;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Assoc_List;
Scope_Result : Internal_Entity_Assoc_List;
Result_Var : Internal_Entity_Assoc_List;
Fld_1 : Bare_Assoc_List;
Len : Integer;
Comp_Result : Boolean;
If_Result : Internal_Entity_Assoc_List;
Let_Result : Internal_Entity_Assoc_List;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 7499 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7495 '<FieldAccess for args at ast.py:17287>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17287
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity (Node => Ent.Node.Attribute_Ref_F_Args, Info => Ent.Info);
--# expr-done 7495
Var_Expr := Fld;
if Var_Expr /= No_Entity then
   --# scope-start
--# expr-start 7496 '<Cast to AssocList at ast.py:17287>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17287
Cast_Expr := Var_Expr; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Assoc_List_Range
   then
      Cast_Result := Create_Internal_Entity_Assoc_List
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 7496
Scope_Result := Cast_Result;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Assoc_List;
end if;
Block_Var := Result_Var; 
--# bind assoc_list Block_Var
--# expr-start 7498 '<If at ast.py:17288>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17288
--# expr-start 7497 '<length at ast.py:17289>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17289
Fld_1 := Block_Var.Node; 
Len := Length (Fld_1); 
--# expr-done 7497
Comp_Result := Len > 0; 
if Comp_Result then
   If_Result := Block_Var;
else
   If_Result := No_Entity_Assoc_List;
end if;
--# expr-done 7498
Let_Result := If_Result; 
--# end
--# expr-done 7499
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Args_List;
--# end
--# property-start AttributeRef.has_context_free_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17294
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Has_Context_Free_Type
  (Node : Bare_Attribute_Ref
  )
   return Boolean
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
Not_Val : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 7501 '<Not at ast.py:17294>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17294
--# expr-start 7500 '<FieldAccess for is_access_attr at ast.py:17294>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17294
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.is_access_attr
Fld := Libadalang.Implementation.Attribute_Ref_P_Is_Access_Attr (Node => Self);
--# end
--# expr-done 7500
Not_Val := not (Fld); 
--# expr-done 7501
      Property_Result := Not_Val;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Has_Context_Free_Type;
--# end
--# property-start AttributeRef.env_elements_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17297
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Env_Elements_Impl
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   Property_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1243 with Inline_Always;
      Fld : Bare_Identifier;
Fld_1 : Symbol_Type;
Sym : Symbol_Type;
Is_Equal : Boolean;
Fld_2 : Internal_Entity_Name;
Fld_3 : Internal_Entity_Array_Access;
If_Result : Internal_Entity_Array_Access;
Let_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1243 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1243;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7509 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7508 '<Cond at ast.py:17298>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17298
--# expr-start 7505 '<Eq at ast.py:17299>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17299
--# expr-start 7503 '<FieldAccess for sym at ast.py:17299>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17299
--# expr-start 7502 '<FieldAccess for attribute at ast.py:17299>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17299
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Attribute_Ref_F_Attribute;
--# expr-done 7502
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTokNode.sym
Fld_1 := Libadalang.Implementation.Single_Tok_Node_P_Sym (Node => Fld);
--# end
--# expr-done 7503
--# expr-start 7504 '<Symbol Unrestricted_Access at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Unrestricted_Access); 
--# expr-done 7504
Is_Equal := Fld_1 = Sym; 
--# expr-done 7505
if Is_Equal then
   --# expr-start 7507 '<FieldAccess for env_elements_impl at ast.py:17304>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17304
--# expr-start 7506 '<FieldAccess for prefix at ast.py:17304>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17304
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7506
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.env_elements_impl'
Fld_3 := Libadalang.Implementation.Dispatcher_Expr_P_Env_Elements_Impl (Node => Fld_2.Node, Env => Env, E_Info => Fld_2.Info);
--# end
--# expr-done 7507
   If_Result := Fld_3;
else
   If_Result := No_Internal_Entity_Array_Type;
end if;
      Inc_Ref (If_Result);
--# expr-done 7508
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 7509
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1243;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1243;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Env_Elements_Impl;
--# end
--# property-start AttributeRef.is_access_attr /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17309
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Is_Access_Attr
  (Node : Bare_Attribute_Ref
  )
   return Boolean
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Identifier;
Fld_1 : Symbol_Type;
Sym : Symbol_Type;
Is_Equal : Boolean;
Fld_2 : Bare_Identifier;
Fld_3 : Symbol_Type;
Sym_1 : Symbol_Type;
Is_Equal_1 : Boolean;
If_Result : Boolean;
Fld_4 : Bare_Identifier;
Fld_5 : Symbol_Type;
Sym_2 : Symbol_Type;
Is_Equal_2 : Boolean;
If_Result_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 7513 '<Eq at ast.py:17310>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17310
--# expr-start 7511 '<FieldAccess for name_symbol at ast.py:17310>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17310
--# expr-start 7510 '<FieldAccess for attribute at ast.py:17310>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17310
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Attribute_Ref_F_Attribute;
--# expr-done 7510
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld);
--# end
--# expr-done 7511
--# expr-start 7512 '<Symbol Access at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Access); 
--# expr-done 7512
Is_Equal := Fld_1 = Sym; 
--# expr-done 7513
if Is_Equal then
   If_Result := True;
else
   --# expr-start 7517 '<Eq at ast.py:17310>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17310
--# expr-start 7515 '<FieldAccess for name_symbol at ast.py:17310>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17310
--# expr-start 7514 '<FieldAccess for attribute at ast.py:17310>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17310
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Attribute_Ref_F_Attribute;
--# expr-done 7514
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_2);
--# end
--# expr-done 7515
--# expr-start 7516 '<Symbol Unchecked_Access at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Unchecked_Access); 
--# expr-done 7516
Is_Equal_1 := Fld_3 = Sym_1; 
--# expr-done 7517
   If_Result := Is_Equal_1;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 7521 '<Eq at ast.py:17310>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17310
--# expr-start 7519 '<FieldAccess for name_symbol at ast.py:17310>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17310
--# expr-start 7518 '<FieldAccess for attribute at ast.py:17310>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17310
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Self.Attribute_Ref_F_Attribute;
--# expr-done 7518
      if Fld_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_5 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_4);
--# end
--# expr-done 7519
--# expr-start 7520 '<Symbol Unrestricted_Access at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Unrestricted_Access); 
--# expr-done 7520
Is_Equal_2 := Fld_5 = Sym_2; 
--# expr-done 7521
   If_Result_1 := Is_Equal_2;
end if;
      Property_Result := If_Result_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Is_Access_Attr;
--# end
--# property-start AttributeRef.designated_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17316
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Designated_Env
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
      ; No_Visibility : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   --# bind no_visibility No_Visibility
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1245 with Inline_Always;
      Fld : Boolean;
Fld_1 : Internal_Entity_Name;
Fld_2 : Lexical_Env := Empty_Env;
Fld_3 : Internal_Entity_Identifier;
Sym : Symbol_Type;
Fld_4 : Boolean;
Node_Parents : Bare_Ada_Node_Array_Access;
Is_A : Boolean;
Find_Result : Bare_Ada_Node;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Fld_5 : Internal_Entity_Base_Subp_Spec;
Fld_6 : Internal_Entity_Base_Type_Decl;
Fld_7 : Lexical_Env := Empty_Env;
If_Result_1 : Lexical_Env := Empty_Env;
If_Result_2 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1245 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Node_Parents);
                     Dec_Ref (Fld_7);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (If_Result_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1245;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7537 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7536 '<Cond at ast.py:17317>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17317
--# expr-start 7522 '<FieldAccess for is_access_attr at ast.py:17318>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17318
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.is_access_attr
Fld := Libadalang.Implementation.Attribute_Ref_P_Is_Access_Attr (Node => Ent.Node);
--# end
--# expr-done 7522
if Fld then
   --# expr-start 7524 '<FieldAccess for designated_env at ast.py:17319>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17319
--# expr-start 7523 '<FieldAccess for prefix at ast.py:17319>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17319
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7523
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.designated_env'
Fld_2 := Libadalang.Implementation.Dispatcher_Expr_P_Designated_Env (Node => Fld_1.Node, Env => Env, Origin => Origin, No_Visibility => No_Visibility, E_Info => Fld_1.Info);
--# end
--# expr-done 7524
   If_Result_2 := Fld_2;
else
--# expr-start 7527 '<FieldAccess for name_is at ast.py:17321>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17321
--# expr-start 7525 '<FieldAccess for attribute at ast.py:17321>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17321
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Identifier (Node => Ent.Node.Attribute_Ref_F_Attribute, Info => Ent.Info);
--# expr-done 7525
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7526 '<Symbol Result at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Result); 
--# expr-done 7526
--# property-call-start Name.name_is
Fld_4 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_3.Node, Sym => Sym);
--# end
--# expr-done 7527
if Fld_4 then
   --# expr-start 7535 '<FieldAccess for defining_env at ast.py:17322>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17322
--# expr-start 7534 '<FieldAccess for return_type at ast.py:17322>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17322
--# expr-start 7533 '<FieldAccess for subp_spec_or_null at ast.py:17322>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17322
--# expr-start 7532 '<Cast to BasicDecl at ast.py:17322>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17322
--# expr-start 7531 '<as_entity at ast.py:17322>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17322
--# expr-start 7530 '<Find at ast.py:17322>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17322
   --# expr-start 7528 '<FieldAccess for parents at ast.py:17322>' Node_Parents /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17322
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents := Parents (Self, True); 
--# expr-done 7528
      Find_Result := No_Bare_Ada_Node;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents;
      begin
         for Item_247 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_247
      --# expr-start 7529 '<IsA BasicSubpDecl, SubpBody at ast.py:17322>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17322
Is_A := Item_247 /= null 
and then Item_247.Kind in Ada_Basic_Subp_Decl | Ada_Subp_Body_Range; 
--# expr-done 7529
      if Is_A then
         Find_Result := Item_247;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 7530
Is_Null := Find_Result = null; 
if Is_Null then
   If_Result := No_Entity;
else
As_Entity := (Info => E_Info, Node => Find_Result); 
   If_Result := As_Entity;
end if;
--# expr-done 7531
   if If_Result.Node = null
      or else If_Result.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => If_Result.Node,
         Info => If_Result.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 7532
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_5 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Cast_Result.Node, E_Info => Cast_Result.Info);
--# end
--# expr-done 7533
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.return_type
Fld_6 := Libadalang.Implementation.Base_Subp_Spec_P_Return_Type (Node => Fld_5.Node, Origin => Origin, E_Info => Fld_5.Info);
--# end
--# expr-done 7534
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_7 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Fld_6.Node, Origin => Origin, Include_Ud_Indexing => False, E_Info => Fld_6.Info);
--# end
--# expr-done 7535
   If_Result_1 := Fld_7;
else
   If_Result_1 := Empty_Env;
end if;
      Inc_Ref (If_Result_1);
   If_Result_2 := If_Result_1;
end if;
      Inc_Ref (If_Result_2);
--# expr-done 7536
Let_Result := If_Result_2; Inc_Ref (Let_Result);
--# end
--# expr-done 7537
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1245;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1245;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Designated_Env;
--# end
--# property-start AttributeRef.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17330
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Xref_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1246 with Inline_Always;
      Block_Var : Symbol_Type;
Fld : Internal_Entity_Identifier;
Fld_1 : Symbol_Type;
Sym : Symbol_Type;
Is_Equal : Boolean;
Sym_1 : Symbol_Type;
Is_Equal_1 : Boolean;
If_Result : Boolean;
Fld_2 : Logic_Equation;
Sym_2 : Symbol_Type;
Is_Equal_2 : Boolean;
Sym_3 : Symbol_Type;
Is_Equal_3 : Boolean;
If_Result_1 : Boolean;
Fld_3 : Logic_Equation;
Sym_4 : Symbol_Type;
Is_Equal_4 : Boolean;
Sym_5 : Symbol_Type;
Is_Equal_5 : Boolean;
If_Result_2 : Boolean;
Sym_6 : Symbol_Type;
Is_Equal_6 : Boolean;
If_Result_3 : Boolean;
Sym_7 : Symbol_Type;
Is_Equal_7 : Boolean;
If_Result_4 : Boolean;
Fld_4 : Logic_Equation;
Sym_8 : Symbol_Type;
Is_Equal_8 : Boolean;
Sym_9 : Symbol_Type;
Is_Equal_9 : Boolean;
If_Result_5 : Boolean;
Fld_5 : Logic_Equation;
Sym_10 : Symbol_Type;
Is_Equal_10 : Boolean;
Fld_6 : Logic_Equation;
Sym_11 : Symbol_Type;
Is_Equal_11 : Boolean;
Sym_12 : Symbol_Type;
Is_Equal_12 : Boolean;
If_Result_6 : Boolean;
Fld_7 : Logic_Equation;
Sym_13 : Symbol_Type;
Is_Equal_13 : Boolean;
Sym_14 : Symbol_Type;
Is_Equal_14 : Boolean;
If_Result_7 : Boolean;
Sym_15 : Symbol_Type;
Is_Equal_15 : Boolean;
If_Result_8 : Boolean;
Sym_16 : Symbol_Type;
Is_Equal_16 : Boolean;
If_Result_9 : Boolean;
Sym_17 : Symbol_Type;
Is_Equal_17 : Boolean;
If_Result_10 : Boolean;
Fld_8 : Logic_Equation;
Sym_18 : Symbol_Type;
Is_Equal_18 : Boolean;
Sym_19 : Symbol_Type;
Is_Equal_19 : Boolean;
If_Result_11 : Boolean;
Sym_20 : Symbol_Type;
Is_Equal_20 : Boolean;
If_Result_12 : Boolean;
Fld_9 : Logic_Equation;
Sym_21 : Symbol_Type;
Is_Equal_21 : Boolean;
Sym_22 : Symbol_Type;
Fld_10 : Internal_Entity;
Fld_11 : Logic_Equation;
Sym_23 : Symbol_Type;
Is_Equal_22 : Boolean;
Sym_24 : Symbol_Type;
Fld_12 : Internal_Entity;
Fld_13 : Logic_Equation;
Sym_25 : Symbol_Type;
Is_Equal_23 : Boolean;
Sym_26 : Symbol_Type;
Fld_14 : Internal_Entity;
Fld_15 : Logic_Equation;
Sym_27 : Symbol_Type;
Is_Equal_24 : Boolean;
Sym_28 : Symbol_Type;
Fld_16 : Internal_Entity;
Fld_17 : Logic_Equation;
Sym_29 : Symbol_Type;
Is_Equal_25 : Boolean;
Sym_30 : Symbol_Type;
Fld_18 : Internal_Entity;
Fld_19 : Logic_Equation;
Sym_31 : Symbol_Type;
Is_Equal_26 : Boolean;
Sym_32 : Symbol_Type;
Fld_20 : Internal_Entity;
Fld_21 : Logic_Equation;
Sym_33 : Symbol_Type;
Is_Equal_27 : Boolean;
Fld_22 : Internal_Entity;
Fld_23 : Logic_Equation;
Sym_34 : Symbol_Type;
Is_Equal_28 : Boolean;
Fld_24 : Internal_Entity;
Fld_25 : Logic_Equation;
Sym_35 : Symbol_Type;
Is_Equal_29 : Boolean;
Fld_26 : Logic_Equation;
Sym_36 : Symbol_Type;
Is_Equal_30 : Boolean;
Fld_27 : Logic_Equation;
Sym_37 : Symbol_Type;
Is_Equal_31 : Boolean;
Fld_28 : Logic_Equation;
Sym_38 : Symbol_Type;
Is_Equal_32 : Boolean;
Sym_39 : Symbol_Type;
Is_Equal_33 : Boolean;
If_Result_13 : Boolean;
Sym_40 : Symbol_Type;
Is_Equal_34 : Boolean;
If_Result_14 : Boolean;
Sym_41 : Symbol_Type;
Is_Equal_35 : Boolean;
If_Result_15 : Boolean;
Sym_42 : Symbol_Type;
Is_Equal_36 : Boolean;
If_Result_16 : Boolean;
Sym_43 : Symbol_Type;
Is_Equal_37 : Boolean;
If_Result_17 : Boolean;
Sym_44 : Symbol_Type;
Is_Equal_38 : Boolean;
If_Result_18 : Boolean;
Sym_45 : Symbol_Type;
Is_Equal_39 : Boolean;
If_Result_19 : Boolean;
Fld_29 : Logic_Equation;
Sym_46 : Symbol_Type;
Is_Equal_40 : Boolean;
Sym_47 : Symbol_Type;
Fld_30 : Internal_Entity;
Fld_31 : Logic_Equation;
Sym_48 : Symbol_Type;
Is_Equal_41 : Boolean;
Sym_49 : Symbol_Type;
Is_Equal_42 : Boolean;
If_Result_20 : Boolean;
Sym_50 : Symbol_Type;
Is_Equal_43 : Boolean;
If_Result_21 : Boolean;
Fld_32 : Logic_Equation;
Sym_51 : Symbol_Type;
Is_Equal_44 : Boolean;
Fld_33 : Logic_Equation;
Sym_52 : Symbol_Type;
Is_Equal_45 : Boolean;
Fld_34 : Logic_Equation;
Sym_53 : Symbol_Type;
Is_Equal_46 : Boolean;
Fld_35 : Logic_Equation;
Sym_54 : Symbol_Type;
Is_Equal_47 : Boolean;
Sym_55 : Symbol_Type;
Is_Equal_48 : Boolean;
If_Result_22 : Boolean;
Fld_36 : Logic_Equation;
Sym_56 : Symbol_Type;
Is_Equal_49 : Boolean;
Sym_57 : Symbol_Type;
Is_Equal_50 : Boolean;
If_Result_23 : Boolean;
Fld_37 : Internal_Entity_Name;
Fld_38 : Logic_Equation;
Sym_58 : Symbol_Type;
Is_Equal_51 : Boolean;
Sym_59 : Symbol_Type;
Is_Equal_52 : Boolean;
If_Result_24 : Boolean;
Sym_60 : Symbol_Type;
Is_Equal_53 : Boolean;
If_Result_25 : Boolean;
Sym_61 : Symbol_Type;
Is_Equal_54 : Boolean;
If_Result_26 : Boolean;
Sym_62 : Symbol_Type;
Is_Equal_55 : Boolean;
If_Result_27 : Boolean;
Sym_63 : Symbol_Type;
Is_Equal_56 : Boolean;
If_Result_28 : Boolean;
Sym_64 : Symbol_Type;
Is_Equal_57 : Boolean;
If_Result_29 : Boolean;
Sym_65 : Symbol_Type;
Is_Equal_58 : Boolean;
If_Result_30 : Boolean;
Sym_66 : Symbol_Type;
Is_Equal_59 : Boolean;
If_Result_31 : Boolean;
Fld_39 : Internal_Entity_Name;
Fld_40 : Logic_Equation;
Fld_41 : Logic_Var;
Fld_42 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Sym_67 : Symbol_Type;
Is_Equal_60 : Boolean;
Sym_68 : Symbol_Type;
Is_Equal_61 : Boolean;
If_Result_32 : Boolean;
Sym_69 : Symbol_Type;
Is_Equal_62 : Boolean;
If_Result_33 : Boolean;
Sym_70 : Symbol_Type;
Is_Equal_63 : Boolean;
If_Result_34 : Boolean;
Sym_71 : Symbol_Type;
Is_Equal_64 : Boolean;
If_Result_35 : Boolean;
Sym_72 : Symbol_Type;
Is_Equal_65 : Boolean;
If_Result_36 : Boolean;
Sym_73 : Symbol_Type;
Is_Equal_66 : Boolean;
If_Result_37 : Boolean;
Sym_74 : Symbol_Type;
Is_Equal_67 : Boolean;
If_Result_38 : Boolean;
Sym_75 : Symbol_Type;
Is_Equal_68 : Boolean;
If_Result_39 : Boolean;
Sym_76 : Symbol_Type;
Is_Equal_69 : Boolean;
If_Result_40 : Boolean;
Sym_77 : Symbol_Type;
Is_Equal_70 : Boolean;
If_Result_41 : Boolean;
Sym_78 : Symbol_Type;
Is_Equal_71 : Boolean;
If_Result_42 : Boolean;
Sym_79 : Symbol_Type;
Is_Equal_72 : Boolean;
If_Result_43 : Boolean;
Sym_80 : Symbol_Type;
Is_Equal_73 : Boolean;
If_Result_44 : Boolean;
Sym_81 : Symbol_Type;
Is_Equal_74 : Boolean;
If_Result_45 : Boolean;
Sym_82 : Symbol_Type;
Is_Equal_75 : Boolean;
If_Result_46 : Boolean;
Sym_83 : Symbol_Type;
Is_Equal_76 : Boolean;
If_Result_47 : Boolean;
Sym_84 : Symbol_Type;
Is_Equal_77 : Boolean;
If_Result_48 : Boolean;
Sym_85 : Symbol_Type;
Is_Equal_78 : Boolean;
If_Result_49 : Boolean;
Sym_86 : Symbol_Type;
Is_Equal_79 : Boolean;
If_Result_50 : Boolean;
Sym_87 : Symbol_Type;
Is_Equal_80 : Boolean;
If_Result_51 : Boolean;
Sym_88 : Symbol_Type;
Is_Equal_81 : Boolean;
If_Result_52 : Boolean;
Sym_89 : Symbol_Type;
Is_Equal_82 : Boolean;
If_Result_53 : Boolean;
Sym_90 : Symbol_Type;
Is_Equal_83 : Boolean;
If_Result_54 : Boolean;
Sym_91 : Symbol_Type;
Is_Equal_84 : Boolean;
If_Result_55 : Boolean;
Sym_92 : Symbol_Type;
Is_Equal_85 : Boolean;
If_Result_56 : Boolean;
Sym_93 : Symbol_Type;
Is_Equal_86 : Boolean;
If_Result_57 : Boolean;
Fld_43 : Internal_Entity_Name;
Fld_44 : Logic_Equation;
Fld_45 : Logic_Var;
Fld_46 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Sym_94 : Symbol_Type;
Is_Equal_87 : Boolean;
Fld_47 : Logic_Var;
Sym_95 : Symbol_Type;
Fld_48 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
Sym_96 : Symbol_Type;
Is_Equal_88 : Boolean;
Fld_49 : Logic_Equation;
Sym_97 : Symbol_Type;
Is_Equal_89 : Boolean;
Fld_50 : Logic_Equation;
Sym_98 : Symbol_Type;
Is_Equal_90 : Boolean;
Fld_51 : Logic_Equation;
Sym_99 : Symbol_Type;
Is_Equal_91 : Boolean;
Sym_100 : Symbol_Type;
Is_Equal_92 : Boolean;
If_Result_58 : Boolean;
Fld_52 : Internal_Entity_Name;
Fld_53 : Logic_Equation;
Fld_54 : Logic_Var;
Fld_55 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_2 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Sym_101 : Symbol_Type;
Is_Equal_93 : Boolean;
Fld_56 : Internal_Entity_Name;
Fld_57 : Logic_Equation;
Fld_58 : Logic_Var;
Fld_59 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Sym_102 : Symbol_Type;
Is_Equal_94 : Boolean;
Fld_60 : Internal_Entity_Name;
Fld_61 : Logic_Equation;
Fld_62 : Logic_Var;
Fld_63 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result_3 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Sym_103 : Symbol_Type;
Is_Equal_95 : Boolean;
Sym_104 : Symbol_Type;
Is_Equal_96 : Boolean;
If_Result_59 : Boolean;
Sym_105 : Symbol_Type;
Is_Equal_97 : Boolean;
If_Result_60 : Boolean;
Sym_106 : Symbol_Type;
Is_Equal_98 : Boolean;
If_Result_61 : Boolean;
Sym_107 : Symbol_Type;
Is_Equal_99 : Boolean;
If_Result_62 : Boolean;
Sym_108 : Symbol_Type;
Is_Equal_100 : Boolean;
If_Result_63 : Boolean;
Sym_109 : Symbol_Type;
Is_Equal_101 : Boolean;
If_Result_64 : Boolean;
Sym_110 : Symbol_Type;
Is_Equal_102 : Boolean;
If_Result_65 : Boolean;
Sym_111 : Symbol_Type;
Is_Equal_103 : Boolean;
If_Result_66 : Boolean;
Sym_112 : Symbol_Type;
Is_Equal_104 : Boolean;
If_Result_67 : Boolean;
Fld_64 : Logic_Equation;
Sym_113 : Symbol_Type;
Is_Equal_105 : Boolean;
Sym_114 : Symbol_Type;
Is_Equal_106 : Boolean;
If_Result_68 : Boolean;
Sym_115 : Symbol_Type;
Is_Equal_107 : Boolean;
If_Result_69 : Boolean;
Fld_65 : Logic_Equation;
Sym_116 : Symbol_Type;
Is_Equal_108 : Boolean;
Sym_117 : Symbol_Type;
Is_Equal_109 : Boolean;
If_Result_70 : Boolean;
Fld_66 : Logic_Equation;
Sym_118 : Symbol_Type;
Is_Equal_110 : Boolean;
Fld_67 : Logic_Equation;
Sym_119 : Symbol_Type;
Is_Equal_111 : Boolean;
Sym_120 : Symbol_Type;
Is_Equal_112 : Boolean;
If_Result_71 : Boolean;
Fld_68 : Logic_Equation;
Sym_121 : Symbol_Type;
Is_Equal_113 : Boolean;
Fld_69 : Logic_Equation;
Sym_122 : Symbol_Type;
Is_Equal_114 : Boolean;
Fld_70 : Logic_Equation;
Sym_123 : Symbol_Type;
Is_Equal_115 : Boolean;
Fld_71 : Logic_Equation;
Sym_124 : Symbol_Type;
Is_Equal_116 : Boolean;
Fld_72 : Logic_Var;
Sym_125 : Symbol_Type;
Fld_73 : Internal_Entity;
Bind_Result_4 : Logic_Equation;
Fld_74 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity;
Bind_Result_5 : Logic_Equation;
And_Pred_5 : Logic_Equation;
If_Result_72 : Logic_Equation;
If_Result_73 : Logic_Equation;
If_Result_74 : Logic_Equation;
If_Result_75 : Logic_Equation;
If_Result_76 : Logic_Equation;
If_Result_77 : Logic_Equation;
If_Result_78 : Logic_Equation;
If_Result_79 : Logic_Equation;
If_Result_80 : Logic_Equation;
If_Result_81 : Logic_Equation;
If_Result_82 : Logic_Equation;
If_Result_83 : Logic_Equation;
If_Result_84 : Logic_Equation;
If_Result_85 : Logic_Equation;
If_Result_86 : Logic_Equation;
If_Result_87 : Logic_Equation;
If_Result_88 : Logic_Equation;
If_Result_89 : Logic_Equation;
If_Result_90 : Logic_Equation;
If_Result_91 : Logic_Equation;
If_Result_92 : Logic_Equation;
If_Result_93 : Logic_Equation;
If_Result_94 : Logic_Equation;
If_Result_95 : Logic_Equation;
If_Result_96 : Logic_Equation;
If_Result_97 : Logic_Equation;
If_Result_98 : Logic_Equation;
If_Result_99 : Logic_Equation;
If_Result_100 : Logic_Equation;
If_Result_101 : Logic_Equation;
If_Result_102 : Logic_Equation;
If_Result_103 : Logic_Equation;
If_Result_104 : Logic_Equation;
If_Result_105 : Logic_Equation;
If_Result_106 : Logic_Equation;
If_Result_107 : Logic_Equation;
If_Result_108 : Logic_Equation;
If_Result_109 : Logic_Equation;
If_Result_110 : Logic_Equation;
If_Result_111 : Logic_Equation;
If_Result_112 : Logic_Equation;
If_Result_113 : Logic_Equation;
If_Result_114 : Logic_Equation;
If_Result_115 : Logic_Equation;
If_Result_116 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1246 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Fld_5);
                     Dec_Ref (Fld_6);
                     Dec_Ref (Fld_7);
                     Dec_Ref (Fld_8);
                     Dec_Ref (Fld_9);
                     Dec_Ref (Fld_11);
                     Dec_Ref (Fld_13);
                     Dec_Ref (Fld_15);
                     Dec_Ref (Fld_17);
                     Dec_Ref (Fld_19);
                     Dec_Ref (Fld_21);
                     Dec_Ref (Fld_23);
                     Dec_Ref (Fld_25);
                     Dec_Ref (Fld_26);
                     Dec_Ref (Fld_27);
                     Dec_Ref (Fld_28);
                     Dec_Ref (Fld_29);
                     Dec_Ref (Fld_31);
                     Dec_Ref (Fld_32);
                     Dec_Ref (Fld_33);
                     Dec_Ref (Fld_34);
                     Dec_Ref (Fld_35);
                     Dec_Ref (Fld_36);
                     Dec_Ref (Fld_38);
                     Dec_Ref (Fld_40);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_44);
                     Dec_Ref (Fld_46);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Fld_49);
                     Dec_Ref (Fld_50);
                     Dec_Ref (Fld_51);
                     Dec_Ref (Fld_53);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Fld_57);
                     Dec_Ref (Fld_59);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Fld_61);
                     Dec_Ref (Bind_Result_3);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Fld_64);
                     Dec_Ref (Fld_65);
                     Dec_Ref (Fld_66);
                     Dec_Ref (Fld_67);
                     Dec_Ref (Fld_68);
                     Dec_Ref (Fld_69);
                     Dec_Ref (Fld_70);
                     Dec_Ref (Fld_71);
                     Dec_Ref (Bind_Result_4);
                     Dec_Ref (Bind_Result_5);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (If_Result_72);
                     Dec_Ref (If_Result_73);
                     Dec_Ref (If_Result_74);
                     Dec_Ref (If_Result_75);
                     Dec_Ref (If_Result_76);
                     Dec_Ref (If_Result_77);
                     Dec_Ref (If_Result_78);
                     Dec_Ref (If_Result_79);
                     Dec_Ref (If_Result_80);
                     Dec_Ref (If_Result_81);
                     Dec_Ref (If_Result_82);
                     Dec_Ref (If_Result_83);
                     Dec_Ref (If_Result_84);
                     Dec_Ref (If_Result_85);
                     Dec_Ref (If_Result_86);
                     Dec_Ref (If_Result_87);
                     Dec_Ref (If_Result_88);
                     Dec_Ref (If_Result_89);
                     Dec_Ref (If_Result_90);
                     Dec_Ref (If_Result_91);
                     Dec_Ref (If_Result_92);
                     Dec_Ref (If_Result_93);
                     Dec_Ref (If_Result_94);
                     Dec_Ref (If_Result_95);
                     Dec_Ref (If_Result_96);
                     Dec_Ref (If_Result_97);
                     Dec_Ref (If_Result_98);
                     Dec_Ref (If_Result_99);
                     Dec_Ref (If_Result_100);
                     Dec_Ref (If_Result_101);
                     Dec_Ref (If_Result_102);
                     Dec_Ref (If_Result_103);
                     Dec_Ref (If_Result_104);
                     Dec_Ref (If_Result_105);
                     Dec_Ref (If_Result_106);
                     Dec_Ref (If_Result_107);
                     Dec_Ref (If_Result_108);
                     Dec_Ref (If_Result_109);
                     Dec_Ref (If_Result_110);
                     Dec_Ref (If_Result_111);
                     Dec_Ref (If_Result_112);
                     Dec_Ref (If_Result_113);
                     Dec_Ref (If_Result_114);
                     Dec_Ref (If_Result_115);
                     Dec_Ref (If_Result_116);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1246;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7869 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7539 '<FieldAccess for name_symbol at ast.py:17331>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17331
--# expr-start 7538 '<FieldAccess for attribute at ast.py:17331>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17331
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Identifier (Node => Ent.Node.Attribute_Ref_F_Attribute, Info => Ent.Info);
--# expr-done 7538
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld.Node);
--# end
--# expr-done 7539
Block_Var := Fld_1; 
--# bind rel_name Block_Var
--# expr-start 7868 '<Cond at ast.py:17332>' If_Result_116 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17332
--# expr-start 7541 '<Eq at ast.py:17333>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17333
--# expr-start 7540 '<Symbol Succ at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Succ); 
--# expr-done 7540
Is_Equal := Block_Var = Sym; 
--# expr-done 7541
if Is_Equal then
   If_Result := True;
else
   --# expr-start 7543 '<Eq at ast.py:17333>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17333
--# expr-start 7542 '<Symbol Pred at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Pred); 
--# expr-done 7542
Is_Equal_1 := Block_Var = Sym_1; 
--# expr-done 7543
   If_Result := Is_Equal_1;
end if;
if If_Result then
   --# expr-start 7544 '<FieldAccess for succpred_xref_equation at ast.py:17333>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17333
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.succpred_xref_equation
Fld_2 := Libadalang.Implementation.Attribute_Ref_P_Succpred_Xref_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7544
   If_Result_116 := Fld_2;
else
--# expr-start 7546 '<Eq at ast.py:17334>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17334
--# expr-start 7545 '<Symbol Min at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Min); 
--# expr-done 7545
Is_Equal_2 := Block_Var = Sym_2; 
--# expr-done 7546
if Is_Equal_2 then
   If_Result_1 := True;
else
   --# expr-start 7548 '<Eq at ast.py:17334>' Is_Equal_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17334
--# expr-start 7547 '<Symbol Max at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Max); 
--# expr-done 7547
Is_Equal_3 := Block_Var = Sym_3; 
--# expr-done 7548
   If_Result_1 := Is_Equal_3;
end if;
if If_Result_1 then
   --# expr-start 7549 '<FieldAccess for minmax_equation at ast.py:17334>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17334
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.minmax_equation
Fld_3 := Libadalang.Implementation.Attribute_Ref_P_Minmax_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7549
   If_Result_115 := Fld_3;
else
--# expr-start 7551 '<Eq at ast.py:17336>' Is_Equal_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17336
--# expr-start 7550 '<Symbol First at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_First); 
--# expr-done 7550
Is_Equal_4 := Block_Var = Sym_4; 
--# expr-done 7551
if Is_Equal_4 then
   If_Result_2 := True;
else
   --# expr-start 7553 '<Eq at ast.py:17336>' Is_Equal_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17336
--# expr-start 7552 '<Symbol Last at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Last); 
--# expr-done 7552
Is_Equal_5 := Block_Var = Sym_5; 
--# expr-done 7553
   If_Result_2 := Is_Equal_5;
end if;
if If_Result_2 then
   If_Result_3 := True;
else
   --# expr-start 7555 '<Eq at ast.py:17336>' Is_Equal_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17336
--# expr-start 7554 '<Symbol Range at ???>' Sym_6 None
Sym_6 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Range); 
--# expr-done 7554
Is_Equal_6 := Block_Var = Sym_6; 
--# expr-done 7555
   If_Result_3 := Is_Equal_6;
end if;
if If_Result_3 then
   If_Result_4 := True;
else
   --# expr-start 7557 '<Eq at ast.py:17336>' Is_Equal_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17336
--# expr-start 7556 '<Symbol Length at ???>' Sym_7 None
Sym_7 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Length); 
--# expr-done 7556
Is_Equal_7 := Block_Var = Sym_7; 
--# expr-done 7557
   If_Result_4 := Is_Equal_7;
end if;
if If_Result_4 then
   --# expr-start 7558 '<FieldAccess for array_attr_equation at ast.py:17337>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17337
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.array_attr_equation
Fld_4 := Libadalang.Implementation.Attribute_Ref_P_Array_Attr_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7558
   If_Result_114 := Fld_4;
else
--# expr-start 7560 '<Eq at ast.py:17339>' Is_Equal_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17339
--# expr-start 7559 '<Symbol Size at ???>' Sym_8 None
Sym_8 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Size); 
--# expr-done 7559
Is_Equal_8 := Block_Var = Sym_8; 
--# expr-done 7560
if Is_Equal_8 then
   If_Result_5 := True;
else
   --# expr-start 7562 '<Eq at ast.py:17339>' Is_Equal_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17339
--# expr-start 7561 '<Symbol VADS_Size at ???>' Sym_9 None
Sym_9 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Vads_Size); 
--# expr-done 7561
Is_Equal_9 := Block_Var = Sym_9; 
--# expr-done 7562
   If_Result_5 := Is_Equal_9;
end if;
if If_Result_5 then
   --# expr-start 7563 '<FieldAccess for size_equation at ast.py:17339>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17339
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.size_equation
Fld_5 := Libadalang.Implementation.Attribute_Ref_P_Size_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7563
   If_Result_113 := Fld_5;
else
--# expr-start 7565 '<Eq at ast.py:17340>' Is_Equal_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17340
--# expr-start 7564 '<Symbol Pos at ???>' Sym_10 None
Sym_10 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Pos); 
--# expr-done 7564
Is_Equal_10 := Block_Var = Sym_10; 
--# expr-done 7565
if Is_Equal_10 then
   --# expr-start 7566 '<FieldAccess for pos_equation at ast.py:17340>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17340
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.pos_equation
Fld_6 := Libadalang.Implementation.Attribute_Ref_P_Pos_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7566
   If_Result_112 := Fld_6;
else
--# expr-start 7568 '<Eq at ast.py:17341>' Is_Equal_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17341
--# expr-start 7567 '<Symbol Val at ???>' Sym_11 None
Sym_11 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Val); 
--# expr-done 7567
Is_Equal_11 := Block_Var = Sym_11; 
--# expr-done 7568
if Is_Equal_11 then
   If_Result_6 := True;
else
   --# expr-start 7570 '<Eq at ast.py:17341>' Is_Equal_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17341
--# expr-start 7569 '<Symbol Enum_Val at ???>' Sym_12 None
Sym_12 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Enum_Val); 
--# expr-done 7569
Is_Equal_12 := Block_Var = Sym_12; 
--# expr-done 7570
   If_Result_6 := Is_Equal_12;
end if;
if If_Result_6 then
   --# expr-start 7571 '<FieldAccess for val_equation at ast.py:17341>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17341
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.val_equation
Fld_7 := Libadalang.Implementation.Attribute_Ref_P_Val_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7571
   If_Result_111 := Fld_7;
else
--# expr-start 7573 '<Eq at ast.py:17343>' Is_Equal_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17343
--# expr-start 7572 '<Symbol Max_Size_In_Storage_Elements at ???>' Sym_13 None
Sym_13 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Max_Size_In_Storage_Elements); 
--# expr-done 7572
Is_Equal_13 := Block_Var = Sym_13; 
--# expr-done 7573
if Is_Equal_13 then
   If_Result_7 := True;
else
   --# expr-start 7575 '<Eq at ast.py:17343>' Is_Equal_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17343
--# expr-start 7574 '<Symbol Aft at ???>' Sym_14 None
Sym_14 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Aft); 
--# expr-done 7574
Is_Equal_14 := Block_Var = Sym_14; 
--# expr-done 7575
   If_Result_7 := Is_Equal_14;
end if;
if If_Result_7 then
   If_Result_8 := True;
else
   --# expr-start 7577 '<Eq at ast.py:17343>' Is_Equal_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17343
--# expr-start 7576 '<Symbol Object_Size at ???>' Sym_15 None
Sym_15 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Object_Size); 
--# expr-done 7576
Is_Equal_15 := Block_Var = Sym_15; 
--# expr-done 7577
   If_Result_8 := Is_Equal_15;
end if;
if If_Result_8 then
   If_Result_9 := True;
else
   --# expr-start 7579 '<Eq at ast.py:17343>' Is_Equal_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17343
--# expr-start 7578 '<Symbol Value_Size at ???>' Sym_16 None
Sym_16 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Value_Size); 
--# expr-done 7578
Is_Equal_16 := Block_Var = Sym_16; 
--# expr-done 7579
   If_Result_9 := Is_Equal_16;
end if;
if If_Result_9 then
   If_Result_10 := True;
else
   --# expr-start 7581 '<Eq at ast.py:17343>' Is_Equal_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17343
--# expr-start 7580 '<Symbol Storage_Size at ???>' Sym_17 None
Sym_17 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Storage_Size); 
--# expr-done 7580
Is_Equal_17 := Block_Var = Sym_17; 
--# expr-done 7581
   If_Result_10 := Is_Equal_17;
end if;
if If_Result_10 then
   --# expr-start 7582 '<FieldAccess for subtype_attr_equation at ast.py:17345>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17345
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.subtype_attr_equation
Fld_8 := Libadalang.Implementation.Attribute_Ref_P_Subtype_Attr_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7582
   If_Result_110 := Fld_8;
else
--# expr-start 7584 '<Eq at ast.py:17347>' Is_Equal_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17347
--# expr-start 7583 '<Symbol Access at ???>' Sym_18 None
Sym_18 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Access); 
--# expr-done 7583
Is_Equal_18 := Block_Var = Sym_18; 
--# expr-done 7584
if Is_Equal_18 then
   If_Result_11 := True;
else
   --# expr-start 7586 '<Eq at ast.py:17347>' Is_Equal_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17347
--# expr-start 7585 '<Symbol Unchecked_Access at ???>' Sym_19 None
Sym_19 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Unchecked_Access); 
--# expr-done 7585
Is_Equal_19 := Block_Var = Sym_19; 
--# expr-done 7586
   If_Result_11 := Is_Equal_19;
end if;
if If_Result_11 then
   If_Result_12 := True;
else
   --# expr-start 7588 '<Eq at ast.py:17347>' Is_Equal_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17347
--# expr-start 7587 '<Symbol Unrestricted_Access at ???>' Sym_20 None
Sym_20 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Unrestricted_Access); 
--# expr-done 7587
Is_Equal_20 := Block_Var = Sym_20; 
--# expr-done 7588
   If_Result_12 := Is_Equal_20;
end if;
if If_Result_12 then
   --# expr-start 7589 '<FieldAccess for access_equation at ast.py:17349>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17349
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.access_equation
Fld_9 := Libadalang.Implementation.Attribute_Ref_P_Access_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7589
   If_Result_109 := Fld_9;
else
--# expr-start 7591 '<Eq at ast.py:17351>' Is_Equal_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17351
--# expr-start 7590 '<Symbol Image at ???>' Sym_21 None
Sym_21 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Image); 
--# expr-done 7590
Is_Equal_21 := Block_Var = Sym_21; 
--# expr-done 7591
if Is_Equal_21 then
   --# expr-start 7594 '<FieldAccess for image_equation at ast.py:17352>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17352
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7593 '<FieldAccess for std_entity at ast.py:17352>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17352
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7592 '<Symbol String at ???>' Sym_22 None
Sym_22 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_String); 
--# expr-done 7592
--# property-call-start AdaNode.std_entity
Fld_10 := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym_22);
--# end
--# expr-done 7593
--# property-call-start AttributeRef.image_equation
Fld_11 := Libadalang.Implementation.Attribute_Ref_P_Image_Equation (Node => Ent.Node, Str_Type => Fld_10, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7594
   If_Result_108 := Fld_11;
else
--# expr-start 7596 '<Eq at ast.py:17354>' Is_Equal_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17354
--# expr-start 7595 '<Symbol Wide_Image at ???>' Sym_23 None
Sym_23 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Wide_Image); 
--# expr-done 7595
Is_Equal_22 := Block_Var = Sym_23; 
--# expr-done 7596
if Is_Equal_22 then
   --# expr-start 7599 '<FieldAccess for image_equation at ast.py:17355>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17355
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7598 '<FieldAccess for std_entity at ast.py:17355>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17355
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7597 '<Symbol Wide_String at ???>' Sym_24 None
Sym_24 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Wide_String); 
--# expr-done 7597
--# property-call-start AdaNode.std_entity
Fld_12 := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym_24);
--# end
--# expr-done 7598
--# property-call-start AttributeRef.image_equation
Fld_13 := Libadalang.Implementation.Attribute_Ref_P_Image_Equation (Node => Ent.Node, Str_Type => Fld_12, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7599
   If_Result_107 := Fld_13;
else
--# expr-start 7601 '<Eq at ast.py:17357>' Is_Equal_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17357
--# expr-start 7600 '<Symbol Wide_Wide_Image at ???>' Sym_25 None
Sym_25 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Wide_Wide_Image); 
--# expr-done 7600
Is_Equal_23 := Block_Var = Sym_25; 
--# expr-done 7601
if Is_Equal_23 then
   --# expr-start 7604 '<FieldAccess for image_equation at ast.py:17358>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17358
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7603 '<FieldAccess for std_entity at ast.py:17358>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17358
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7602 '<Symbol Wide_Wide_String at ???>' Sym_26 None
Sym_26 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Wide_Wide_String); 
--# expr-done 7602
--# property-call-start AdaNode.std_entity
Fld_14 := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym_26);
--# end
--# expr-done 7603
--# property-call-start AttributeRef.image_equation
Fld_15 := Libadalang.Implementation.Attribute_Ref_P_Image_Equation (Node => Ent.Node, Str_Type => Fld_14, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7604
   If_Result_106 := Fld_15;
else
--# expr-start 7606 '<Eq at ast.py:17360>' Is_Equal_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17360
--# expr-start 7605 '<Symbol Value at ???>' Sym_27 None
Sym_27 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Value); 
--# expr-done 7605
Is_Equal_24 := Block_Var = Sym_27; 
--# expr-done 7606
if Is_Equal_24 then
   --# expr-start 7609 '<FieldAccess for value_equation at ast.py:17361>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17361
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7608 '<FieldAccess for std_entity at ast.py:17361>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17361
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7607 '<Symbol String at ???>' Sym_28 None
Sym_28 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_String); 
--# expr-done 7607
--# property-call-start AdaNode.std_entity
Fld_16 := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym_28);
--# end
--# expr-done 7608
--# property-call-start AttributeRef.value_equation
Fld_17 := Libadalang.Implementation.Attribute_Ref_P_Value_Equation (Node => Ent.Node, Str_Type => Fld_16, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7609
   If_Result_105 := Fld_17;
else
--# expr-start 7611 '<Eq at ast.py:17363>' Is_Equal_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17363
--# expr-start 7610 '<Symbol Wide_Value at ???>' Sym_29 None
Sym_29 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Wide_Value); 
--# expr-done 7610
Is_Equal_25 := Block_Var = Sym_29; 
--# expr-done 7611
if Is_Equal_25 then
   --# expr-start 7614 '<FieldAccess for value_equation at ast.py:17364>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17364
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7613 '<FieldAccess for std_entity at ast.py:17364>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17364
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7612 '<Symbol Wide_String at ???>' Sym_30 None
Sym_30 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Wide_String); 
--# expr-done 7612
--# property-call-start AdaNode.std_entity
Fld_18 := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym_30);
--# end
--# expr-done 7613
--# property-call-start AttributeRef.value_equation
Fld_19 := Libadalang.Implementation.Attribute_Ref_P_Value_Equation (Node => Ent.Node, Str_Type => Fld_18, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7614
   If_Result_104 := Fld_19;
else
--# expr-start 7616 '<Eq at ast.py:17366>' Is_Equal_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17366
--# expr-start 7615 '<Symbol Wide_Wide_Value at ???>' Sym_31 None
Sym_31 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Wide_Wide_Value); 
--# expr-done 7615
Is_Equal_26 := Block_Var = Sym_31; 
--# expr-done 7616
if Is_Equal_26 then
   --# expr-start 7619 '<FieldAccess for value_equation at ast.py:17367>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17367
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7618 '<FieldAccess for std_entity at ast.py:17367>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17367
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7617 '<Symbol Wide_Wide_String at ???>' Sym_32 None
Sym_32 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Wide_Wide_String); 
--# expr-done 7617
--# property-call-start AdaNode.std_entity
Fld_20 := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym_32);
--# end
--# expr-done 7618
--# property-call-start AttributeRef.value_equation
Fld_21 := Libadalang.Implementation.Attribute_Ref_P_Value_Equation (Node => Ent.Node, Str_Type => Fld_20, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7619
   If_Result_103 := Fld_21;
else
--# expr-start 7621 '<Eq at ast.py:17369>' Is_Equal_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17369
--# expr-start 7620 '<Symbol Fixed_Value at ???>' Sym_33 None
Sym_33 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Fixed_Value); 
--# expr-done 7620
Is_Equal_27 := Block_Var = Sym_33; 
--# expr-done 7621
if Is_Equal_27 then
   --# expr-start 7623 '<FieldAccess for fixed_value_equation at ast.py:17370>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17370
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7622 '<FieldAccess for universal_int_type at ast.py:17370>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17370
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.universal_int_type
Fld_22 := Libadalang.Implementation.Ada_Node_P_Universal_Int_Type (Node => Self);
--# end
--# expr-done 7622
--# property-call-start AttributeRef.fixed_value_equation
Fld_23 := Libadalang.Implementation.Attribute_Ref_P_Fixed_Value_Equation (Node => Ent.Node, Universal_Type => Fld_22, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7623
   If_Result_102 := Fld_23;
else
--# expr-start 7625 '<Eq at ast.py:17372>' Is_Equal_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17372
--# expr-start 7624 '<Symbol Integer_Value at ???>' Sym_34 None
Sym_34 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Integer_Value); 
--# expr-done 7624
Is_Equal_28 := Block_Var = Sym_34; 
--# expr-done 7625
if Is_Equal_28 then
   --# expr-start 7627 '<FieldAccess for fixed_value_equation at ast.py:17373>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17373
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7626 '<FieldAccess for universal_real_type at ast.py:17373>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17373
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.universal_real_type
Fld_24 := Libadalang.Implementation.Ada_Node_P_Universal_Real_Type (Node => Self);
--# end
--# expr-done 7626
--# property-call-start AttributeRef.fixed_value_equation
Fld_25 := Libadalang.Implementation.Attribute_Ref_P_Fixed_Value_Equation (Node => Ent.Node, Universal_Type => Fld_24, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7627
   If_Result_101 := Fld_25;
else
--# expr-start 7629 '<Eq at ast.py:17375>' Is_Equal_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17375
--# expr-start 7628 '<Symbol Invalid_Value at ???>' Sym_35 None
Sym_35 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Invalid_Value); 
--# expr-done 7628
Is_Equal_29 := Block_Var = Sym_35; 
--# expr-done 7629
if Is_Equal_29 then
   --# expr-start 7630 '<FieldAccess for invalid_value_equation at ast.py:17376>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17376
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.invalid_value_equation
Fld_26 := Libadalang.Implementation.Attribute_Ref_P_Invalid_Value_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7630
   If_Result_100 := Fld_26;
else
--# expr-start 7632 '<Eq at ast.py:17378>' Is_Equal_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17378
--# expr-start 7631 '<Symbol Identity at ???>' Sym_36 None
Sym_36 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Identity); 
--# expr-done 7631
Is_Equal_30 := Block_Var = Sym_36; 
--# expr-done 7632
if Is_Equal_30 then
   --# expr-start 7633 '<FieldAccess for identity_equation at ast.py:17378>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17378
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.identity_equation
Fld_27 := Libadalang.Implementation.Attribute_Ref_P_Identity_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7633
   If_Result_99 := Fld_27;
else
--# expr-start 7635 '<Eq at ast.py:17379>' Is_Equal_31 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17379
--# expr-start 7634 '<Symbol Address at ???>' Sym_37 None
Sym_37 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Address); 
--# expr-done 7634
Is_Equal_31 := Block_Var = Sym_37; 
--# expr-done 7635
if Is_Equal_31 then
   --# expr-start 7636 '<FieldAccess for address_equation at ast.py:17379>' Fld_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17379
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.address_equation
Fld_28 := Libadalang.Implementation.Attribute_Ref_P_Address_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7636
   If_Result_98 := Fld_28;
else
--# expr-start 7638 '<Eq at ast.py:17381>' Is_Equal_32 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17381
--# expr-start 7637 '<Symbol Small at ???>' Sym_38 None
Sym_38 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Small); 
--# expr-done 7637
Is_Equal_32 := Block_Var = Sym_38; 
--# expr-done 7638
if Is_Equal_32 then
   If_Result_13 := True;
else
   --# expr-start 7640 '<Eq at ast.py:17381>' Is_Equal_33 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17381
--# expr-start 7639 '<Symbol Model_Small at ???>' Sym_39 None
Sym_39 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Model_Small); 
--# expr-done 7639
Is_Equal_33 := Block_Var = Sym_39; 
--# expr-done 7640
   If_Result_13 := Is_Equal_33;
end if;
if If_Result_13 then
   If_Result_14 := True;
else
   --# expr-start 7642 '<Eq at ast.py:17381>' Is_Equal_34 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17381
--# expr-start 7641 '<Symbol Safe_Small at ???>' Sym_40 None
Sym_40 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Safe_Small); 
--# expr-done 7641
Is_Equal_34 := Block_Var = Sym_40; 
--# expr-done 7642
   If_Result_14 := Is_Equal_34;
end if;
if If_Result_14 then
   If_Result_15 := True;
else
   --# expr-start 7644 '<Eq at ast.py:17381>' Is_Equal_35 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17381
--# expr-start 7643 '<Symbol Epsilon at ???>' Sym_41 None
Sym_41 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Epsilon); 
--# expr-done 7643
Is_Equal_35 := Block_Var = Sym_41; 
--# expr-done 7644
   If_Result_15 := Is_Equal_35;
end if;
if If_Result_15 then
   If_Result_16 := True;
else
   --# expr-start 7646 '<Eq at ast.py:17381>' Is_Equal_36 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17381
--# expr-start 7645 '<Symbol Model_Epsilon at ???>' Sym_42 None
Sym_42 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Model_Epsilon); 
--# expr-done 7645
Is_Equal_36 := Block_Var = Sym_42; 
--# expr-done 7646
   If_Result_16 := Is_Equal_36;
end if;
if If_Result_16 then
   If_Result_17 := True;
else
   --# expr-start 7648 '<Eq at ast.py:17381>' Is_Equal_37 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17381
--# expr-start 7647 '<Symbol Large at ???>' Sym_43 None
Sym_43 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Large); 
--# expr-done 7647
Is_Equal_37 := Block_Var = Sym_43; 
--# expr-done 7648
   If_Result_17 := Is_Equal_37;
end if;
if If_Result_17 then
   If_Result_18 := True;
else
   --# expr-start 7650 '<Eq at ast.py:17381>' Is_Equal_38 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17381
--# expr-start 7649 '<Symbol Safe_Large at ???>' Sym_44 None
Sym_44 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Safe_Large); 
--# expr-done 7649
Is_Equal_38 := Block_Var = Sym_44; 
--# expr-done 7650
   If_Result_18 := Is_Equal_38;
end if;
if If_Result_18 then
   If_Result_19 := True;
else
   --# expr-start 7652 '<Eq at ast.py:17381>' Is_Equal_39 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17381
--# expr-start 7651 '<Symbol Delta at ???>' Sym_45 None
Sym_45 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Delta); 
--# expr-done 7651
Is_Equal_39 := Block_Var = Sym_45; 
--# expr-done 7652
   If_Result_19 := Is_Equal_39;
end if;
if If_Result_19 then
   --# expr-start 7653 '<FieldAccess for universal_real_equation at ast.py:17385>' Fld_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17385
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.universal_real_equation
Fld_29 := Libadalang.Implementation.Attribute_Ref_P_Universal_Real_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7653
   If_Result_97 := Fld_29;
else
--# expr-start 7655 '<Eq at ast.py:17387>' Is_Equal_40 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17387
--# expr-start 7654 '<Symbol Img at ???>' Sym_46 None
Sym_46 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Img); 
--# expr-done 7654
Is_Equal_40 := Block_Var = Sym_46; 
--# expr-done 7655
if Is_Equal_40 then
   --# expr-start 7658 '<FieldAccess for img_equation at ast.py:17388>' Fld_31 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17388
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7657 '<FieldAccess for std_entity at ast.py:17388>' Fld_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17388
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7656 '<Symbol String at ???>' Sym_47 None
Sym_47 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_String); 
--# expr-done 7656
--# property-call-start AdaNode.std_entity
Fld_30 := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym_47);
--# end
--# expr-done 7657
--# property-call-start AttributeRef.img_equation
Fld_31 := Libadalang.Implementation.Attribute_Ref_P_Img_Equation (Node => Ent.Node, Str_Type => Fld_30, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7658
   If_Result_96 := Fld_31;
else
--# expr-start 7660 '<Eq at ast.py:17390>' Is_Equal_41 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17390
--# expr-start 7659 '<Symbol Write at ???>' Sym_48 None
Sym_48 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Write); 
--# expr-done 7659
Is_Equal_41 := Block_Var = Sym_48; 
--# expr-done 7660
if Is_Equal_41 then
   If_Result_20 := True;
else
   --# expr-start 7662 '<Eq at ast.py:17390>' Is_Equal_42 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17390
--# expr-start 7661 '<Symbol Read at ???>' Sym_49 None
Sym_49 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Read); 
--# expr-done 7661
Is_Equal_42 := Block_Var = Sym_49; 
--# expr-done 7662
   If_Result_20 := Is_Equal_42;
end if;
if If_Result_20 then
   If_Result_21 := True;
else
   --# expr-start 7664 '<Eq at ast.py:17390>' Is_Equal_43 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17390
--# expr-start 7663 '<Symbol Output at ???>' Sym_50 None
Sym_50 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Output); 
--# expr-done 7663
Is_Equal_43 := Block_Var = Sym_50; 
--# expr-done 7664
   If_Result_21 := Is_Equal_43;
end if;
if If_Result_21 then
   --# expr-start 7665 '<FieldAccess for stream_attrs_equation at ast.py:17391>' Fld_32 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17391
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.stream_attrs_equation
Fld_32 := Libadalang.Implementation.Attribute_Ref_P_Stream_Attrs_Equation (Node => Ent.Node, Return_Obj => False, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7665
   If_Result_95 := Fld_32;
else
--# expr-start 7667 '<Eq at ast.py:17393>' Is_Equal_44 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17393
--# expr-start 7666 '<Symbol Input at ???>' Sym_51 None
Sym_51 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Input); 
--# expr-done 7666
Is_Equal_44 := Block_Var = Sym_51; 
--# expr-done 7667
if Is_Equal_44 then
   --# expr-start 7668 '<FieldAccess for stream_attrs_equation at ast.py:17393>' Fld_33 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17393
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.stream_attrs_equation
Fld_33 := Libadalang.Implementation.Attribute_Ref_P_Stream_Attrs_Equation (Node => Ent.Node, Return_Obj => True, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7668
   If_Result_94 := Fld_33;
else
--# expr-start 7670 '<Eq at ast.py:17395>' Is_Equal_45 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17395
--# expr-start 7669 '<Symbol Tag at ???>' Sym_52 None
Sym_52 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Tag); 
--# expr-done 7669
Is_Equal_45 := Block_Var = Sym_52; 
--# expr-done 7670
if Is_Equal_45 then
   --# expr-start 7671 '<FieldAccess for tag_attr_equation at ast.py:17395>' Fld_34 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17395
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.tag_attr_equation
Fld_34 := Libadalang.Implementation.Attribute_Ref_P_Tag_Attr_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7671
   If_Result_93 := Fld_34;
else
--# expr-start 7673 '<Eq at ast.py:17397>' Is_Equal_46 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17397
--# expr-start 7672 '<Symbol Result at ???>' Sym_53 None
Sym_53 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Result); 
--# expr-done 7672
Is_Equal_46 := Block_Var = Sym_53; 
--# expr-done 7673
if Is_Equal_46 then
   --# expr-start 7674 '<FieldAccess for result_attr_equation at ast.py:17397>' Fld_35 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17397
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.result_attr_equation
Fld_35 := Libadalang.Implementation.Attribute_Ref_P_Result_Attr_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7674
   If_Result_92 := Fld_35;
else
--# expr-start 7676 '<Eq at ast.py:17399>' Is_Equal_47 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17399
--# expr-start 7675 '<Symbol Old at ???>' Sym_54 None
Sym_54 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Old); 
--# expr-done 7675
Is_Equal_47 := Block_Var = Sym_54; 
--# expr-done 7676
if Is_Equal_47 then
   If_Result_22 := True;
else
   --# expr-start 7678 '<Eq at ast.py:17399>' Is_Equal_48 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17399
--# expr-start 7677 '<Symbol Loop_Entry at ???>' Sym_55 None
Sym_55 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Loop_Entry); 
--# expr-done 7677
Is_Equal_48 := Block_Var = Sym_55; 
--# expr-done 7678
   If_Result_22 := Is_Equal_48;
end if;
if If_Result_22 then
   --# expr-start 7679 '<FieldAccess for bind_to_prefix_eq at ast.py:17400>' Fld_36 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17400
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.bind_to_prefix_eq
Fld_36 := Libadalang.Implementation.Attribute_Ref_P_Bind_To_Prefix_Eq (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7679
   If_Result_91 := Fld_36;
else
--# expr-start 7681 '<Eq at ast.py:17402>' Is_Equal_49 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17402
--# expr-start 7680 '<Symbol Class at ???>' Sym_56 None
Sym_56 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Class); 
--# expr-done 7680
Is_Equal_49 := Block_Var = Sym_56; 
--# expr-done 7681
if Is_Equal_49 then
   If_Result_23 := True;
else
   --# expr-start 7683 '<Eq at ast.py:17402>' Is_Equal_50 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17402
--# expr-start 7682 '<Symbol Base at ???>' Sym_57 None
Sym_57 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Base); 
--# expr-done 7682
Is_Equal_50 := Block_Var = Sym_57; 
--# expr-done 7683
   If_Result_23 := Is_Equal_50;
end if;
if If_Result_23 then
   --# expr-start 7685 '<FieldAccess for sub_equation at ast.py:17403>' Fld_38 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17403
--# expr-start 7684 '<FieldAccess for prefix at ast.py:17403>' Fld_37 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17403
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_37 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7684
      if Fld_37.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_38 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_37.Node, Env => Env, Origin => Origin, E_Info => Fld_37.Info);
--# end
--# expr-done 7685
   If_Result_90 := Fld_38;
else
--# expr-start 7687 '<Eq at ast.py:17405>' Is_Equal_51 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17405
--# expr-start 7686 '<Symbol Valid at ???>' Sym_58 None
Sym_58 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Valid); 
--# expr-done 7686
Is_Equal_51 := Block_Var = Sym_58; 
--# expr-done 7687
if Is_Equal_51 then
   If_Result_24 := True;
else
   --# expr-start 7689 '<Eq at ast.py:17405>' Is_Equal_52 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17405
--# expr-start 7688 '<Symbol Machine_Overflows at ???>' Sym_59 None
Sym_59 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Machine_Overflows); 
--# expr-done 7688
Is_Equal_52 := Block_Var = Sym_59; 
--# expr-done 7689
   If_Result_24 := Is_Equal_52;
end if;
if If_Result_24 then
   If_Result_25 := True;
else
   --# expr-start 7691 '<Eq at ast.py:17405>' Is_Equal_53 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17405
--# expr-start 7690 '<Symbol Machine_Rounds at ???>' Sym_60 None
Sym_60 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Machine_Rounds); 
--# expr-done 7690
Is_Equal_53 := Block_Var = Sym_60; 
--# expr-done 7691
   If_Result_25 := Is_Equal_53;
end if;
if If_Result_25 then
   If_Result_26 := True;
else
   --# expr-start 7693 '<Eq at ast.py:17405>' Is_Equal_54 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17405
--# expr-start 7692 '<Symbol Has_Access_Values at ???>' Sym_61 None
Sym_61 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Has_Access_Values); 
--# expr-done 7692
Is_Equal_54 := Block_Var = Sym_61; 
--# expr-done 7693
   If_Result_26 := Is_Equal_54;
end if;
if If_Result_26 then
   If_Result_27 := True;
else
   --# expr-start 7695 '<Eq at ast.py:17405>' Is_Equal_55 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17405
--# expr-start 7694 '<Symbol Has_Discriminants at ???>' Sym_62 None
Sym_62 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Has_Discriminants); 
--# expr-done 7694
Is_Equal_55 := Block_Var = Sym_62; 
--# expr-done 7695
   If_Result_27 := Is_Equal_55;
end if;
if If_Result_27 then
   If_Result_28 := True;
else
   --# expr-start 7697 '<Eq at ast.py:17405>' Is_Equal_56 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17405
--# expr-start 7696 '<Symbol Has_Tagged_Values at ???>' Sym_63 None
Sym_63 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Has_Tagged_Values); 
--# expr-done 7696
Is_Equal_56 := Block_Var = Sym_63; 
--# expr-done 7697
   If_Result_28 := Is_Equal_56;
end if;
if If_Result_28 then
   If_Result_29 := True;
else
   --# expr-start 7699 '<Eq at ast.py:17405>' Is_Equal_57 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17405
--# expr-start 7698 '<Symbol Definite at ???>' Sym_64 None
Sym_64 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Definite); 
--# expr-done 7698
Is_Equal_57 := Block_Var = Sym_64; 
--# expr-done 7699
   If_Result_29 := Is_Equal_57;
end if;
if If_Result_29 then
   If_Result_30 := True;
else
   --# expr-start 7701 '<Eq at ast.py:17405>' Is_Equal_58 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17405
--# expr-start 7700 '<Symbol Constrained at ???>' Sym_65 None
Sym_65 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Constrained); 
--# expr-done 7700
Is_Equal_58 := Block_Var = Sym_65; 
--# expr-done 7701
   If_Result_30 := Is_Equal_58;
end if;
if If_Result_30 then
   If_Result_31 := True;
else
   --# expr-start 7703 '<Eq at ast.py:17405>' Is_Equal_59 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17405
--# expr-start 7702 '<Symbol Initialized at ???>' Sym_66 None
Sym_66 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Initialized); 
--# expr-done 7702
Is_Equal_59 := Block_Var = Sym_66; 
--# expr-done 7703
   If_Result_31 := Is_Equal_59;
end if;
if If_Result_31 then
   --# expr-start 7709 '<And at ast.py:17409>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17409
--# expr-start 7705 '<FieldAccess for sub_equation at ast.py:17409>' Fld_40 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17409
--# expr-start 7704 '<FieldAccess for prefix at ast.py:17409>' Fld_39 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17409
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_39 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7704
      if Fld_39.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_40 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_39.Node, Env => Env, Origin => Origin, E_Info => Fld_39.Info);
--# end
--# expr-done 7705
--# expr-start 7708 '<Bind at ast.py:17410>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17410
--# expr-start 7706 '<FieldAccess for type_var at ast.py:17410>' Fld_41 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17410
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_41 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 7706
Fld_41.Value := No_Entity;
Entity_Vars.Reset (Fld_41);
--# expr-start 7707 '<FieldAccess for bool_type at ast.py:17410>' Fld_42 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17410
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_42 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 7707
      Cast_Result := Create_Internal_Entity
        (Node => Fld_42.Node,
         Info => Fld_42.Info);
Bind_Result := Solver.Create_Assign (Fld_41, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17410") else null)); 
--# expr-done 7708
And_Pred := Create_And (Fld_40, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17409") else null)); 
--# expr-done 7709
   If_Result_89 := And_Pred;
else
--# expr-start 7711 '<Eq at ast.py:17412>' Is_Equal_60 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7710 '<Symbol Width at ???>' Sym_67 None
Sym_67 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Width); 
--# expr-done 7710
Is_Equal_60 := Block_Var = Sym_67; 
--# expr-done 7711
if Is_Equal_60 then
   If_Result_32 := True;
else
   --# expr-start 7713 '<Eq at ast.py:17412>' Is_Equal_61 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7712 '<Symbol Component_Size at ???>' Sym_68 None
Sym_68 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Component_Size); 
--# expr-done 7712
Is_Equal_61 := Block_Var = Sym_68; 
--# expr-done 7713
   If_Result_32 := Is_Equal_61;
end if;
if If_Result_32 then
   If_Result_33 := True;
else
   --# expr-start 7715 '<Eq at ast.py:17412>' Is_Equal_62 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7714 '<Symbol Position at ???>' Sym_69 None
Sym_69 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Position); 
--# expr-done 7714
Is_Equal_62 := Block_Var = Sym_69; 
--# expr-done 7715
   If_Result_33 := Is_Equal_62;
end if;
if If_Result_33 then
   If_Result_34 := True;
else
   --# expr-start 7717 '<Eq at ast.py:17412>' Is_Equal_63 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7716 '<Symbol Mantissa at ???>' Sym_70 None
Sym_70 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Mantissa); 
--# expr-done 7716
Is_Equal_63 := Block_Var = Sym_70; 
--# expr-done 7717
   If_Result_34 := Is_Equal_63;
end if;
if If_Result_34 then
   If_Result_35 := True;
else
   --# expr-start 7719 '<Eq at ast.py:17412>' Is_Equal_64 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7718 '<Symbol Model_Mantissa at ???>' Sym_71 None
Sym_71 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Model_Mantissa); 
--# expr-done 7718
Is_Equal_64 := Block_Var = Sym_71; 
--# expr-done 7719
   If_Result_35 := Is_Equal_64;
end if;
if If_Result_35 then
   If_Result_36 := True;
else
   --# expr-start 7721 '<Eq at ast.py:17412>' Is_Equal_65 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7720 '<Symbol Machine_Mantissa at ???>' Sym_72 None
Sym_72 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Machine_Mantissa); 
--# expr-done 7720
Is_Equal_65 := Block_Var = Sym_72; 
--# expr-done 7721
   If_Result_36 := Is_Equal_65;
end if;
if If_Result_36 then
   If_Result_37 := True;
else
   --# expr-start 7723 '<Eq at ast.py:17412>' Is_Equal_66 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7722 '<Symbol Fore at ???>' Sym_73 None
Sym_73 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Fore); 
--# expr-done 7722
Is_Equal_66 := Block_Var = Sym_73; 
--# expr-done 7723
   If_Result_37 := Is_Equal_66;
end if;
if If_Result_37 then
   If_Result_38 := True;
else
   --# expr-start 7725 '<Eq at ast.py:17412>' Is_Equal_67 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7724 '<Symbol Aft at ???>' Sym_74 None
Sym_74 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Aft); 
--# expr-done 7724
Is_Equal_67 := Block_Var = Sym_74; 
--# expr-done 7725
   If_Result_38 := Is_Equal_67;
end if;
if If_Result_38 then
   If_Result_39 := True;
else
   --# expr-start 7727 '<Eq at ast.py:17412>' Is_Equal_68 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7726 '<Symbol Digits at ???>' Sym_75 None
Sym_75 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Digits); 
--# expr-done 7726
Is_Equal_68 := Block_Var = Sym_75; 
--# expr-done 7727
   If_Result_39 := Is_Equal_68;
end if;
if If_Result_39 then
   If_Result_40 := True;
else
   --# expr-start 7729 '<Eq at ast.py:17412>' Is_Equal_69 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7728 '<Symbol Modulus at ???>' Sym_76 None
Sym_76 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Modulus); 
--# expr-done 7728
Is_Equal_69 := Block_Var = Sym_76; 
--# expr-done 7729
   If_Result_40 := Is_Equal_69;
end if;
if If_Result_40 then
   If_Result_41 := True;
else
   --# expr-start 7731 '<Eq at ast.py:17412>' Is_Equal_70 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7730 '<Symbol Word_Size at ???>' Sym_77 None
Sym_77 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Word_Size); 
--# expr-done 7730
Is_Equal_70 := Block_Var = Sym_77; 
--# expr-done 7731
   If_Result_41 := Is_Equal_70;
end if;
if If_Result_41 then
   If_Result_42 := True;
else
   --# expr-start 7733 '<Eq at ast.py:17412>' Is_Equal_71 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7732 '<Symbol Max_Integer_Size at ???>' Sym_78 None
Sym_78 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Max_Integer_Size); 
--# expr-done 7732
Is_Equal_71 := Block_Var = Sym_78; 
--# expr-done 7733
   If_Result_42 := Is_Equal_71;
end if;
if If_Result_42 then
   If_Result_43 := True;
else
   --# expr-start 7735 '<Eq at ast.py:17412>' Is_Equal_72 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7734 '<Symbol Address_Size at ???>' Sym_79 None
Sym_79 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Address_Size); 
--# expr-done 7734
Is_Equal_72 := Block_Var = Sym_79; 
--# expr-done 7735
   If_Result_43 := Is_Equal_72;
end if;
if If_Result_43 then
   If_Result_44 := True;
else
   --# expr-start 7737 '<Eq at ast.py:17412>' Is_Equal_73 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7736 '<Symbol Maximum_Alignment at ???>' Sym_80 None
Sym_80 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Maximum_Alignment); 
--# expr-done 7736
Is_Equal_73 := Block_Var = Sym_80; 
--# expr-done 7737
   If_Result_44 := Is_Equal_73;
end if;
if If_Result_44 then
   If_Result_45 := True;
else
   --# expr-start 7739 '<Eq at ast.py:17412>' Is_Equal_74 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7738 '<Symbol System_Allocator_Alignment at ???>' Sym_81 None
Sym_81 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_System_Allocator_Alignment); 
--# expr-done 7738
Is_Equal_74 := Block_Var = Sym_81; 
--# expr-done 7739
   If_Result_45 := Is_Equal_74;
end if;
if If_Result_45 then
   If_Result_46 := True;
else
   --# expr-start 7741 '<Eq at ast.py:17412>' Is_Equal_75 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7740 '<Symbol Finalization_Size at ???>' Sym_82 None
Sym_82 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Finalization_Size); 
--# expr-done 7740
Is_Equal_75 := Block_Var = Sym_82; 
--# expr-done 7741
   If_Result_46 := Is_Equal_75;
end if;
if If_Result_46 then
   If_Result_47 := True;
else
   --# expr-start 7743 '<Eq at ast.py:17412>' Is_Equal_76 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7742 '<Symbol Descriptor_Size at ???>' Sym_83 None
Sym_83 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Descriptor_Size); 
--# expr-done 7742
Is_Equal_76 := Block_Var = Sym_83; 
--# expr-done 7743
   If_Result_47 := Is_Equal_76;
end if;
if If_Result_47 then
   If_Result_48 := True;
else
   --# expr-start 7745 '<Eq at ast.py:17412>' Is_Equal_77 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7744 '<Symbol Alignment at ???>' Sym_84 None
Sym_84 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Alignment); 
--# expr-done 7744
Is_Equal_77 := Block_Var = Sym_84; 
--# expr-done 7745
   If_Result_48 := Is_Equal_77;
end if;
if If_Result_48 then
   If_Result_49 := True;
else
   --# expr-start 7747 '<Eq at ast.py:17412>' Is_Equal_78 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7746 '<Symbol First_Bit at ???>' Sym_85 None
Sym_85 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_First_Bit); 
--# expr-done 7746
Is_Equal_78 := Block_Var = Sym_85; 
--# expr-done 7747
   If_Result_49 := Is_Equal_78;
end if;
if If_Result_49 then
   If_Result_50 := True;
else
   --# expr-start 7749 '<Eq at ast.py:17412>' Is_Equal_79 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7748 '<Symbol Last_Bit at ???>' Sym_86 None
Sym_86 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Last_Bit); 
--# expr-done 7748
Is_Equal_79 := Block_Var = Sym_86; 
--# expr-done 7749
   If_Result_50 := Is_Equal_79;
end if;
if If_Result_50 then
   If_Result_51 := True;
else
   --# expr-start 7751 '<Eq at ast.py:17412>' Is_Equal_80 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7750 '<Symbol Default_Bit_Order at ???>' Sym_87 None
Sym_87 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Default_Bit_Order); 
--# expr-done 7750
Is_Equal_80 := Block_Var = Sym_87; 
--# expr-done 7751
   If_Result_51 := Is_Equal_80;
end if;
if If_Result_51 then
   If_Result_52 := True;
else
   --# expr-start 7753 '<Eq at ast.py:17412>' Is_Equal_81 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7752 '<Symbol Range_Length at ???>' Sym_88 None
Sym_88 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Range_Length); 
--# expr-done 7752
Is_Equal_81 := Block_Var = Sym_88; 
--# expr-done 7753
   If_Result_52 := Is_Equal_81;
end if;
if If_Result_52 then
   If_Result_53 := True;
else
   --# expr-start 7755 '<Eq at ast.py:17412>' Is_Equal_82 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7754 '<Symbol Storage_Unit at ???>' Sym_89 None
Sym_89 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Storage_Unit); 
--# expr-done 7754
Is_Equal_82 := Block_Var = Sym_89; 
--# expr-done 7755
   If_Result_53 := Is_Equal_82;
end if;
if If_Result_53 then
   If_Result_54 := True;
else
   --# expr-start 7757 '<Eq at ast.py:17412>' Is_Equal_83 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7756 '<Symbol Small_Numerator at ???>' Sym_90 None
Sym_90 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Small_Numerator); 
--# expr-done 7756
Is_Equal_83 := Block_Var = Sym_90; 
--# expr-done 7757
   If_Result_54 := Is_Equal_83;
end if;
if If_Result_54 then
   If_Result_55 := True;
else
   --# expr-start 7759 '<Eq at ast.py:17412>' Is_Equal_84 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7758 '<Symbol Small_Denominator at ???>' Sym_91 None
Sym_91 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Small_Denominator); 
--# expr-done 7758
Is_Equal_84 := Block_Var = Sym_91; 
--# expr-done 7759
   If_Result_55 := Is_Equal_84;
end if;
if If_Result_55 then
   If_Result_56 := True;
else
   --# expr-start 7761 '<Eq at ast.py:17412>' Is_Equal_85 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7760 '<Symbol Machine_Emin at ???>' Sym_92 None
Sym_92 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Machine_Emin); 
--# expr-done 7760
Is_Equal_85 := Block_Var = Sym_92; 
--# expr-done 7761
   If_Result_56 := Is_Equal_85;
end if;
if If_Result_56 then
   If_Result_57 := True;
else
   --# expr-start 7763 '<Eq at ast.py:17412>' Is_Equal_86 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17412
--# expr-start 7762 '<Symbol Machine_Emax at ???>' Sym_93 None
Sym_93 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Machine_Emax); 
--# expr-done 7762
Is_Equal_86 := Block_Var = Sym_93; 
--# expr-done 7763
   If_Result_57 := Is_Equal_86;
end if;
if If_Result_57 then
   --# expr-start 7768 '<And at ast.py:17423>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17423
--# expr-start 7765 '<FieldAccess for sub_equation at ast.py:17423>' Fld_44 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17423
--# expr-start 7764 '<FieldAccess for prefix at ast.py:17423>' Fld_43 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17423
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_43 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7764
      if Fld_43.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_44 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_43.Node, Env => Env, Origin => Origin, E_Info => Fld_43.Info);
--# end
--# expr-done 7765
--# expr-start 7767 '<FieldAccess for universal_int_bind at ast.py:17424>' Fld_46 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17424
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7766 '<FieldAccess for type_var at ast.py:17424>' Fld_45 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17424
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_45 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 7766
--# property-call-start AdaNode.universal_int_bind
Fld_46 := Libadalang.Implementation.Ada_Node_P_Universal_Int_Bind (Node => Self, Type_Var => Fld_45, Origin => Origin);
--# end
--# expr-done 7767
And_Pred_1 := Create_And (Fld_44, Fld_46, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17423") else null)); 
--# expr-done 7768
   If_Result_88 := And_Pred_1;
else
--# expr-start 7770 '<Eq at ast.py:17426>' Is_Equal_87 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17426
--# expr-start 7769 '<Symbol Target_Name at ???>' Sym_94 None
Sym_94 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Target_Name); 
--# expr-done 7769
Is_Equal_87 := Block_Var = Sym_94; 
--# expr-done 7770
if Is_Equal_87 then
   --# expr-start 7774 '<Bind at ast.py:17427>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17427
--# expr-start 7771 '<FieldAccess for type_var at ast.py:17427>' Fld_47 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17427
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_47 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 7771
Fld_47.Value := No_Entity;
Entity_Vars.Reset (Fld_47);
--# expr-start 7773 '<FieldAccess for std_entity at ast.py:17427>' Fld_48 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17427
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7772 '<Symbol String at ???>' Sym_95 None
Sym_95 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_String); 
--# expr-done 7772
--# property-call-start AdaNode.std_entity
Fld_48 := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym_95);
--# end
--# expr-done 7773
Bind_Result_1 := Solver.Create_Assign (Fld_47, Fld_48, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17427") else null)); 
--# expr-done 7774
   If_Result_87 := Bind_Result_1;
else
--# expr-start 7776 '<Eq at ast.py:17429>' Is_Equal_88 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17429
--# expr-start 7775 '<Symbol Storage_Pool at ???>' Sym_96 None
Sym_96 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Storage_Pool); 
--# expr-done 7775
Is_Equal_88 := Block_Var = Sym_96; 
--# expr-done 7776
if Is_Equal_88 then
   --# expr-start 7777 '<FieldAccess for storage_pool_equation at ast.py:17429>' Fld_49 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17429
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.storage_pool_equation
Fld_49 := Libadalang.Implementation.Attribute_Ref_P_Storage_Pool_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7777
   If_Result_86 := Fld_49;
else
--# expr-start 7779 '<Eq at ast.py:17431>' Is_Equal_89 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17431
--# expr-start 7778 '<Symbol Scalar_Storage_Order at ???>' Sym_97 None
Sym_97 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Scalar_Storage_Order); 
--# expr-done 7778
Is_Equal_89 := Block_Var = Sym_97; 
--# expr-done 7779
if Is_Equal_89 then
   --# expr-start 7780 '<FieldAccess for scalar_storage_order_equation at ast.py:17432>' Fld_50 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17432
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.scalar_storage_order_equation
Fld_50 := Libadalang.Implementation.Attribute_Ref_P_Scalar_Storage_Order_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7780
   If_Result_85 := Fld_50;
else
--# expr-start 7782 '<Eq at ast.py:17434>' Is_Equal_90 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17434
--# expr-start 7781 '<Symbol Type_Class at ???>' Sym_98 None
Sym_98 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Type_Class); 
--# expr-done 7781
Is_Equal_90 := Block_Var = Sym_98; 
--# expr-done 7782
if Is_Equal_90 then
   --# expr-start 7783 '<FieldAccess for type_class_equation at ast.py:17434>' Fld_51 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17434
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.type_class_equation
Fld_51 := Libadalang.Implementation.Attribute_Ref_P_Type_Class_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7783
   If_Result_84 := Fld_51;
else
--# expr-start 7785 '<Eq at ast.py:17437>' Is_Equal_91 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17437
--# expr-start 7784 '<Symbol Callable at ???>' Sym_99 None
Sym_99 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Callable); 
--# expr-done 7784
Is_Equal_91 := Block_Var = Sym_99; 
--# expr-done 7785
if Is_Equal_91 then
   If_Result_58 := True;
else
   --# expr-start 7787 '<Eq at ast.py:17437>' Is_Equal_92 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17437
--# expr-start 7786 '<Symbol Terminated at ???>' Sym_100 None
Sym_100 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Terminated); 
--# expr-done 7786
Is_Equal_92 := Block_Var = Sym_100; 
--# expr-done 7787
   If_Result_58 := Is_Equal_92;
end if;
if If_Result_58 then
   --# expr-start 7793 '<And at ast.py:17438>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17438
--# expr-start 7789 '<FieldAccess for sub_equation at ast.py:17438>' Fld_53 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17438
--# expr-start 7788 '<FieldAccess for prefix at ast.py:17438>' Fld_52 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17438
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_52 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7788
      if Fld_52.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_53 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_52.Node, Env => Env, Origin => Origin, E_Info => Fld_52.Info);
--# end
--# expr-done 7789
--# expr-start 7792 '<Bind at ast.py:17438>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17438
--# expr-start 7790 '<FieldAccess for type_var at ast.py:17438>' Fld_54 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17438
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_54 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 7790
Fld_54.Value := No_Entity;
Entity_Vars.Reset (Fld_54);
--# expr-start 7791 '<FieldAccess for bool_type at ast.py:17438>' Fld_55 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17438
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_55 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 7791
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_55.Node,
         Info => Fld_55.Info);
Bind_Result_2 := Solver.Create_Assign (Fld_54, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17438") else null)); 
--# expr-done 7792
And_Pred_2 := Create_And (Fld_53, Bind_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17438") else null)); 
--# expr-done 7793
   If_Result_83 := And_Pred_2;
else
--# expr-start 7795 '<Eq at ast.py:17441>' Is_Equal_93 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17441
--# expr-start 7794 '<Symbol Count at ???>' Sym_101 None
Sym_101 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Count); 
--# expr-done 7794
Is_Equal_93 := Block_Var = Sym_101; 
--# expr-done 7795
if Is_Equal_93 then
   --# expr-start 7800 '<And at ast.py:17442>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17442
--# expr-start 7797 '<FieldAccess for xref_no_overloading at ast.py:17442>' Fld_57 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17442
--# expr-start 7796 '<FieldAccess for prefix at ast.py:17442>' Fld_56 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17442
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_56 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7796
      if Fld_56.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_57 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Fld_56.Node, Env => Env, Origin => Origin, E_Info => Fld_56.Info);
--# end
--# expr-done 7797
--# expr-start 7799 '<FieldAccess for universal_int_bind at ast.py:17443>' Fld_59 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17443
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7798 '<FieldAccess for type_var at ast.py:17443>' Fld_58 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17443
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_58 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 7798
--# property-call-start AdaNode.universal_int_bind
Fld_59 := Libadalang.Implementation.Ada_Node_P_Universal_Int_Bind (Node => Self, Type_Var => Fld_58, Origin => Origin);
--# end
--# expr-done 7799
And_Pred_3 := Create_And (Fld_57, Fld_59, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17442") else null)); 
--# expr-done 7800
   If_Result_82 := And_Pred_3;
else
--# expr-start 7802 '<Eq at ast.py:17445>' Is_Equal_94 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17445
--# expr-start 7801 '<Symbol Caller at ???>' Sym_102 None
Sym_102 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Caller); 
--# expr-done 7801
Is_Equal_94 := Block_Var = Sym_102; 
--# expr-done 7802
if Is_Equal_94 then
   --# expr-start 7808 '<And at ast.py:17446>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17446
--# expr-start 7804 '<FieldAccess for xref_no_overloading at ast.py:17446>' Fld_61 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17446
--# expr-start 7803 '<FieldAccess for prefix at ast.py:17446>' Fld_60 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17446
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_60 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7803
      if Fld_60.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_61 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Fld_60.Node, Env => Env, Origin => Origin, E_Info => Fld_60.Info);
--# end
--# expr-done 7804
--# expr-start 7807 '<Bind at ast.py:17447>' Bind_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17447
--# expr-start 7805 '<FieldAccess for type_var at ast.py:17447>' Fld_62 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17447
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_62 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 7805
Fld_62.Value := No_Entity;
Entity_Vars.Reset (Fld_62);
--# expr-start 7806 '<FieldAccess for task_id_type at ast.py:17447>' Fld_63 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17447
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.task_id_type
Fld_63 := Libadalang.Implementation.Ada_Node_P_Task_Id_Type (Node => Self);
--# end
--# expr-done 7806
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_63.Node,
         Info => Fld_63.Info);
Bind_Result_3 := Solver.Create_Assign (Fld_62, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17447") else null)); 
--# expr-done 7807
And_Pred_4 := Create_And (Fld_61, Bind_Result_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17446") else null)); 
--# expr-done 7808
   If_Result_81 := And_Pred_4;
else
--# expr-start 7810 '<Eq at ast.py:17451>' Is_Equal_95 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17451
--# expr-start 7809 '<Symbol Ceiling at ???>' Sym_103 None
Sym_103 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Ceiling); 
--# expr-done 7809
Is_Equal_95 := Block_Var = Sym_103; 
--# expr-done 7810
if Is_Equal_95 then
   If_Result_59 := True;
else
   --# expr-start 7812 '<Eq at ast.py:17451>' Is_Equal_96 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17451
--# expr-start 7811 '<Symbol Floor at ???>' Sym_104 None
Sym_104 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Floor); 
--# expr-done 7811
Is_Equal_96 := Block_Var = Sym_104; 
--# expr-done 7812
   If_Result_59 := Is_Equal_96;
end if;
if If_Result_59 then
   If_Result_60 := True;
else
   --# expr-start 7814 '<Eq at ast.py:17451>' Is_Equal_97 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17451
--# expr-start 7813 '<Symbol Rounding at ???>' Sym_105 None
Sym_105 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Rounding); 
--# expr-done 7813
Is_Equal_97 := Block_Var = Sym_105; 
--# expr-done 7814
   If_Result_60 := Is_Equal_97;
end if;
if If_Result_60 then
   If_Result_61 := True;
else
   --# expr-start 7816 '<Eq at ast.py:17451>' Is_Equal_98 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17451
--# expr-start 7815 '<Symbol Truncation at ???>' Sym_106 None
Sym_106 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Truncation); 
--# expr-done 7815
Is_Equal_98 := Block_Var = Sym_106; 
--# expr-done 7816
   If_Result_61 := Is_Equal_98;
end if;
if If_Result_61 then
   If_Result_62 := True;
else
   --# expr-start 7818 '<Eq at ast.py:17451>' Is_Equal_99 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17451
--# expr-start 7817 '<Symbol Copy_Sign at ???>' Sym_107 None
Sym_107 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Copy_Sign); 
--# expr-done 7817
Is_Equal_99 := Block_Var = Sym_107; 
--# expr-done 7818
   If_Result_62 := Is_Equal_99;
end if;
if If_Result_62 then
   If_Result_63 := True;
else
   --# expr-start 7820 '<Eq at ast.py:17451>' Is_Equal_100 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17451
--# expr-start 7819 '<Symbol Remainder at ???>' Sym_108 None
Sym_108 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Remainder); 
--# expr-done 7819
Is_Equal_100 := Block_Var = Sym_108; 
--# expr-done 7820
   If_Result_63 := Is_Equal_100;
end if;
if If_Result_63 then
   If_Result_64 := True;
else
   --# expr-start 7822 '<Eq at ast.py:17451>' Is_Equal_101 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17451
--# expr-start 7821 '<Symbol Adjacent at ???>' Sym_109 None
Sym_109 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Adjacent); 
--# expr-done 7821
Is_Equal_101 := Block_Var = Sym_109; 
--# expr-done 7822
   If_Result_64 := Is_Equal_101;
end if;
if If_Result_64 then
   If_Result_65 := True;
else
   --# expr-start 7824 '<Eq at ast.py:17451>' Is_Equal_102 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17451
--# expr-start 7823 '<Symbol Machine at ???>' Sym_110 None
Sym_110 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Machine); 
--# expr-done 7823
Is_Equal_102 := Block_Var = Sym_110; 
--# expr-done 7824
   If_Result_65 := Is_Equal_102;
end if;
if If_Result_65 then
   If_Result_66 := True;
else
   --# expr-start 7826 '<Eq at ast.py:17451>' Is_Equal_103 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17451
--# expr-start 7825 '<Symbol Machine_Rounding at ???>' Sym_111 None
Sym_111 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Machine_Rounding); 
--# expr-done 7825
Is_Equal_103 := Block_Var = Sym_111; 
--# expr-done 7826
   If_Result_66 := Is_Equal_103;
end if;
if If_Result_66 then
   If_Result_67 := True;
else
   --# expr-start 7828 '<Eq at ast.py:17451>' Is_Equal_104 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17451
--# expr-start 7827 '<Symbol Fraction at ???>' Sym_112 None
Sym_112 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Fraction); 
--# expr-done 7827
Is_Equal_104 := Block_Var = Sym_112; 
--# expr-done 7828
   If_Result_67 := Is_Equal_104;
end if;
if If_Result_67 then
   --# expr-start 7829 '<FieldAccess for float_funcs_equation at ast.py:17454>' Fld_64 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17454
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.float_funcs_equation
Fld_64 := Libadalang.Implementation.Attribute_Ref_P_Float_Funcs_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7829
   If_Result_80 := Fld_64;
else
--# expr-start 7831 '<Eq at ast.py:17456>' Is_Equal_105 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17456
--# expr-start 7830 '<Symbol Exponent at ???>' Sym_113 None
Sym_113 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Exponent); 
--# expr-done 7830
Is_Equal_105 := Block_Var = Sym_113; 
--# expr-done 7831
if Is_Equal_105 then
   If_Result_68 := True;
else
   --# expr-start 7833 '<Eq at ast.py:17456>' Is_Equal_106 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17456
--# expr-start 7832 '<Symbol Machine_Radix at ???>' Sym_114 None
Sym_114 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Machine_Radix); 
--# expr-done 7832
Is_Equal_106 := Block_Var = Sym_114; 
--# expr-done 7833
   If_Result_68 := Is_Equal_106;
end if;
if If_Result_68 then
   If_Result_69 := True;
else
   --# expr-start 7835 '<Eq at ast.py:17456>' Is_Equal_107 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17456
--# expr-start 7834 '<Symbol Enum_Rep at ???>' Sym_115 None
Sym_115 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Enum_Rep); 
--# expr-done 7834
Is_Equal_107 := Block_Var = Sym_115; 
--# expr-done 7835
   If_Result_69 := Is_Equal_107;
end if;
if If_Result_69 then
   --# expr-start 7836 '<FieldAccess for universal_int_equation at ast.py:17457>' Fld_65 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17457
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.universal_int_equation
Fld_65 := Libadalang.Implementation.Attribute_Ref_P_Universal_Int_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7836
   If_Result_79 := Fld_65;
else
--# expr-start 7838 '<Eq at ast.py:17459>' Is_Equal_108 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17459
--# expr-start 7837 '<Symbol Scaling at ???>' Sym_116 None
Sym_116 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Scaling); 
--# expr-done 7837
Is_Equal_108 := Block_Var = Sym_116; 
--# expr-done 7838
if Is_Equal_108 then
   If_Result_70 := True;
else
   --# expr-start 7840 '<Eq at ast.py:17459>' Is_Equal_109 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17459
--# expr-start 7839 '<Symbol Compose at ???>' Sym_117 None
Sym_117 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Compose); 
--# expr-done 7839
Is_Equal_109 := Block_Var = Sym_117; 
--# expr-done 7840
   If_Result_70 := Is_Equal_109;
end if;
if If_Result_70 then
   --# expr-start 7841 '<FieldAccess for float_universal_int_equation at ast.py:17460>' Fld_66 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17460
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.float_universal_int_equation
Fld_66 := Libadalang.Implementation.Attribute_Ref_P_Float_Universal_Int_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7841
   If_Result_78 := Fld_66;
else
--# expr-start 7843 '<Eq at ast.py:17462>' Is_Equal_110 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17462
--# expr-start 7842 '<Symbol Mod at ???>' Sym_118 None
Sym_118 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Mod_141); 
--# expr-done 7842
Is_Equal_110 := Block_Var = Sym_118; 
--# expr-done 7843
if Is_Equal_110 then
   --# expr-start 7844 '<FieldAccess for mod_equation at ast.py:17463>' Fld_67 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17463
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.mod_equation
Fld_67 := Libadalang.Implementation.Attribute_Ref_P_Mod_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7844
   If_Result_77 := Fld_67;
else
--# expr-start 7846 '<Eq at ast.py:17465>' Is_Equal_111 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17465
--# expr-start 7845 '<Symbol Asm_Input at ???>' Sym_119 None
Sym_119 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Asm_Input); 
--# expr-done 7845
Is_Equal_111 := Block_Var = Sym_119; 
--# expr-done 7846
if Is_Equal_111 then
   If_Result_71 := True;
else
   --# expr-start 7848 '<Eq at ast.py:17465>' Is_Equal_112 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17465
--# expr-start 7847 '<Symbol Asm_Output at ???>' Sym_120 None
Sym_120 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Asm_Output); 
--# expr-done 7847
Is_Equal_112 := Block_Var = Sym_120; 
--# expr-done 7848
   If_Result_71 := Is_Equal_112;
end if;
if If_Result_71 then
   --# expr-start 7849 '<FieldAccess for inline_asm_equation at ast.py:17466>' Fld_68 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17466
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.inline_asm_equation
Fld_68 := Libadalang.Implementation.Attribute_Ref_P_Inline_Asm_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7849
   If_Result_76 := Fld_68;
else
--# expr-start 7851 '<Eq at ast.py:17468>' Is_Equal_113 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17468
--# expr-start 7850 '<Symbol To_Address at ???>' Sym_121 None
Sym_121 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_To_Address); 
--# expr-done 7850
Is_Equal_113 := Block_Var = Sym_121; 
--# expr-done 7851
if Is_Equal_113 then
   --# expr-start 7852 '<FieldAccess for to_address_equation at ast.py:17469>' Fld_69 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17469
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.to_address_equation
Fld_69 := Libadalang.Implementation.Attribute_Ref_P_To_Address_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7852
   If_Result_75 := Fld_69;
else
--# expr-start 7854 '<Eq at ast.py:17471>' Is_Equal_114 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17471
--# expr-start 7853 '<Symbol Put_Image at ???>' Sym_122 None
Sym_122 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Put_Image); 
--# expr-done 7853
Is_Equal_114 := Block_Var = Sym_122; 
--# expr-done 7854
if Is_Equal_114 then
   --# expr-start 7855 '<FieldAccess for put_image_equation at ast.py:17472>' Fld_70 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17472
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.put_image_equation
Fld_70 := Libadalang.Implementation.Attribute_Ref_P_Put_Image_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7855
   If_Result_74 := Fld_70;
else
--# expr-start 7857 '<Eq at ast.py:17474>' Is_Equal_115 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17474
--# expr-start 7856 '<Symbol Index at ???>' Sym_123 None
Sym_123 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Index); 
--# expr-done 7856
Is_Equal_115 := Block_Var = Sym_123; 
--# expr-done 7857
if Is_Equal_115 then
   --# expr-start 7858 '<FieldAccess for index_equation at ast.py:17475>' Fld_71 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17475
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.index_equation
Fld_71 := Libadalang.Implementation.Attribute_Ref_P_Index_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 7858
   If_Result_73 := Fld_71;
else
--# expr-start 7860 '<Eq at ast.py:17477>' Is_Equal_116 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17477
--# expr-start 7859 '<Symbol Abort_Signal at ???>' Sym_124 None
Sym_124 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Abort_Signal); 
--# expr-done 7859
Is_Equal_116 := Block_Var = Sym_124; 
--# expr-done 7860
if Is_Equal_116 then
   --# expr-start 7867 '<And at ast.py:17478>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17478
--# expr-start 7864 '<Bind at ast.py:17478>' Bind_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17478
--# expr-start 7861 '<FieldAccess for ref_var at ast.py:17478>' Fld_72 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17478
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_72 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Self);
--# end
--# expr-done 7861
Fld_72.Value := No_Entity;
Entity_Vars.Reset (Fld_72);
--# expr-start 7863 '<FieldAccess for std_entity at ast.py:17478>' Fld_73 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17478
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7862 '<Symbol abort_signal_ at ???>' Sym_125 None
Sym_125 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Abort_Signal_256); 
--# expr-done 7862
--# property-call-start AdaNode.std_entity
Fld_73 := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym_125);
--# end
--# expr-done 7863
Bind_Result_4 := Solver.Create_Assign (Fld_72, Fld_73, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17478") else null)); 
--# expr-done 7864
--# expr-start 7866 '<Bind at ast.py:17479>' Bind_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17479
--# expr-start 7865 '<FieldAccess for type_var at ast.py:17479>' Fld_74 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17479
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_74 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 7865
Fld_74.Value := No_Entity;
Entity_Vars.Reset (Fld_74);
Cast_Expr := No_Entity_Base_Type_Decl; 
      Cast_Result_3 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result_5 := Solver.Create_Assign (Fld_74, Cast_Result_3, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17479") else null)); 
--# expr-done 7866
And_Pred_5 := Create_And (Bind_Result_4, Bind_Result_5, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17478") else null)); 
--# expr-done 7867
   If_Result_72 := And_Pred_5;
else
   If_Result_72 := raise Property_Error with "Unhandled attribute";
end if;
      Inc_Ref (If_Result_72);
   If_Result_73 := If_Result_72;
end if;
      Inc_Ref (If_Result_73);
   If_Result_74 := If_Result_73;
end if;
      Inc_Ref (If_Result_74);
   If_Result_75 := If_Result_74;
end if;
      Inc_Ref (If_Result_75);
   If_Result_76 := If_Result_75;
end if;
      Inc_Ref (If_Result_76);
   If_Result_77 := If_Result_76;
end if;
      Inc_Ref (If_Result_77);
   If_Result_78 := If_Result_77;
end if;
      Inc_Ref (If_Result_78);
   If_Result_79 := If_Result_78;
end if;
      Inc_Ref (If_Result_79);
   If_Result_80 := If_Result_79;
end if;
      Inc_Ref (If_Result_80);
   If_Result_81 := If_Result_80;
end if;
      Inc_Ref (If_Result_81);
   If_Result_82 := If_Result_81;
end if;
      Inc_Ref (If_Result_82);
   If_Result_83 := If_Result_82;
end if;
      Inc_Ref (If_Result_83);
   If_Result_84 := If_Result_83;
end if;
      Inc_Ref (If_Result_84);
   If_Result_85 := If_Result_84;
end if;
      Inc_Ref (If_Result_85);
   If_Result_86 := If_Result_85;
end if;
      Inc_Ref (If_Result_86);
   If_Result_87 := If_Result_86;
end if;
      Inc_Ref (If_Result_87);
   If_Result_88 := If_Result_87;
end if;
      Inc_Ref (If_Result_88);
   If_Result_89 := If_Result_88;
end if;
      Inc_Ref (If_Result_89);
   If_Result_90 := If_Result_89;
end if;
      Inc_Ref (If_Result_90);
   If_Result_91 := If_Result_90;
end if;
      Inc_Ref (If_Result_91);
   If_Result_92 := If_Result_91;
end if;
      Inc_Ref (If_Result_92);
   If_Result_93 := If_Result_92;
end if;
      Inc_Ref (If_Result_93);
   If_Result_94 := If_Result_93;
end if;
      Inc_Ref (If_Result_94);
   If_Result_95 := If_Result_94;
end if;
      Inc_Ref (If_Result_95);
   If_Result_96 := If_Result_95;
end if;
      Inc_Ref (If_Result_96);
   If_Result_97 := If_Result_96;
end if;
      Inc_Ref (If_Result_97);
   If_Result_98 := If_Result_97;
end if;
      Inc_Ref (If_Result_98);
   If_Result_99 := If_Result_98;
end if;
      Inc_Ref (If_Result_99);
   If_Result_100 := If_Result_99;
end if;
      Inc_Ref (If_Result_100);
   If_Result_101 := If_Result_100;
end if;
      Inc_Ref (If_Result_101);
   If_Result_102 := If_Result_101;
end if;
      Inc_Ref (If_Result_102);
   If_Result_103 := If_Result_102;
end if;
      Inc_Ref (If_Result_103);
   If_Result_104 := If_Result_103;
end if;
      Inc_Ref (If_Result_104);
   If_Result_105 := If_Result_104;
end if;
      Inc_Ref (If_Result_105);
   If_Result_106 := If_Result_105;
end if;
      Inc_Ref (If_Result_106);
   If_Result_107 := If_Result_106;
end if;
      Inc_Ref (If_Result_107);
   If_Result_108 := If_Result_107;
end if;
      Inc_Ref (If_Result_108);
   If_Result_109 := If_Result_108;
end if;
      Inc_Ref (If_Result_109);
   If_Result_110 := If_Result_109;
end if;
      Inc_Ref (If_Result_110);
   If_Result_111 := If_Result_110;
end if;
      Inc_Ref (If_Result_111);
   If_Result_112 := If_Result_111;
end if;
      Inc_Ref (If_Result_112);
   If_Result_113 := If_Result_112;
end if;
      Inc_Ref (If_Result_113);
   If_Result_114 := If_Result_113;
end if;
      Inc_Ref (If_Result_114);
   If_Result_115 := If_Result_114;
end if;
      Inc_Ref (If_Result_115);
   If_Result_116 := If_Result_115;
end if;
      Inc_Ref (If_Result_116);
--# expr-done 7868
Let_Result := If_Result_116; Inc_Ref (Let_Result);
--# end
--# expr-done 7869
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1246;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1246;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Xref_Equation;
--# end
--# property-start AttributeRef.float_funcs_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17485
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Float_Funcs_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1247 with Inline_Always;
            procedure Finalizer_Scope_2769 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Name;
Fld_3 : Logic_Equation;
Fld_4 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_5 : Internal_Entity_Assoc_List;
Fld_6 : Bare_Assoc_List;
Fld_7 : Internal_Entity_Info;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_8 : Internal_Entity_Expr;
Fld_9 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
Fld_10 : Internal_Entity_Expr;
Fld_11 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_12 : Internal_Entity_Expr;
Fld_13 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Item_248 : Internal_Entity_Basic_Assoc;
Bare_Item_248 : Bare_Basic_Assoc;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
And_Pred_3 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1247 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1247;
            procedure Finalizer_Scope_2769 is
            begin
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Fld_11);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_13);
                     Dec_Ref (And_Pred_2);
            end Finalizer_Scope_2769;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7890 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7871 '<FieldAccess for name_designated_type at ast.py:17490>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17490
--# expr-start 7870 '<FieldAccess for prefix at ast.py:17490>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17490
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7870
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_1 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 7871
Block_Var := Fld_1; 
--# bind typ Block_Var
--# expr-start 7889 '<And at ast.py:17493>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17493
--# expr-start 7876 '<And at ast.py:17493>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17493
--# expr-start 7873 '<FieldAccess for sub_equation at ast.py:17493>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17493
--# expr-start 7872 '<FieldAccess for prefix at ast.py:17493>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17493
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7872
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_2.Node, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 7873
--# expr-start 7875 '<Bind at ast.py:17494>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17494
--# expr-start 7874 '<FieldAccess for type_var at ast.py:17494>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17494
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 7874
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_4, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17494") else null)); 
--# expr-done 7875
And_Pred := Create_And (Fld_3, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17493") else null)); 
--# expr-done 7876
--# expr-start 7888 '<LogicAll at ast.py:17495>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17495
--# expr-start 7887 '<Map at ast.py:17495>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17495
   --# expr-start 7877 '<FieldAccess for args_list at ast.py:17495>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17495
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_5 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 7877
Fld_6 := Fld_5.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_6 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Assoc_List := Fld_6;
      begin
         for Untyped_Item_248 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_248 := Untyped_Item_248; 
Is_Null := Bare_Item_248 = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_7 := Fld_5.Info;
As_Entity := (Info => Fld_7, Node => Bare_Item_248); 
   If_Result := As_Entity;
end if;
                  Item_248 := If_Result; 
   --# scope-start
               --# bind arg Item_248
      --# expr-start 7886 '<And at ast.py:17497>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17497
--# expr-start 7883 '<And at ast.py:17497>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17497
--# expr-start 7880 '<Bind at ast.py:17497>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17497
--# expr-start 7879 '<FieldAccess for expected_type_var at ast.py:17497>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17497
--# expr-start 7878 '<FieldAccess for expr at ast.py:17497>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17497
      if Item_248.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_8 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Item_248.Node, E_Info => Item_248.Info);
--# end
--# expr-done 7878
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_8.Node);
--# expr-done 7879
Fld_9.Value := No_Entity;
Entity_Vars.Reset (Fld_9);
Cast_Expr_1 := Block_Var; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_9, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17497") else null)); 
--# expr-done 7880
--# expr-start 7882 '<FieldAccess for sub_equation at ast.py:17498>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17498
--# expr-start 7881 '<FieldAccess for expr at ast.py:17498>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17498
      if Item_248.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_10 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Item_248.Node, E_Info => Item_248.Info);
--# end
--# expr-done 7881
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_11 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_10.Node, Env => Env, Origin => Origin, E_Info => Fld_10.Info);
--# end
--# expr-done 7882
And_Pred_1 := Create_And (Bind_Result_1, Fld_11, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17497") else null)); 
--# expr-done 7883
--# expr-start 7885 '<FieldAccess for matches_expected_type at ast.py:17499>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17499
--# expr-start 7884 '<FieldAccess for expr at ast.py:17499>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17499
      if Item_248.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_12 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Item_248.Node, E_Info => Item_248.Info);
--# end
--# expr-done 7884
      if Fld_12.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_13 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_12.Node, Origin => Origin);
--# end
--# expr-done 7885
And_Pred_2 := Create_And (And_Pred_1, Fld_13, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17497") else null)); 
--# expr-done 7886
         declare
            Item_To_Append : constant Logic_Equation := And_Pred_2;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2769;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 7887
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17495") else null)); 
--# expr-done 7888
And_Pred_3 := Create_And (And_Pred, Logic_Boolean_Op, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17493") else null)); 
--# expr-done 7889
Let_Result := And_Pred_3; Inc_Ref (Let_Result);
--# end
--# expr-done 7890
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1247;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1247;
               Finalizer_Scope_2769;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Float_Funcs_Equation;
--# end
--# property-start AttributeRef.float_universal_int_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17504
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Float_Universal_Int_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1248 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Block_Var_1 : Internal_Entity_Expr;
Block_Var_2 : Internal_Entity_Expr;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Assoc_List;
Fld_3 : Bare_Assoc_List;
Fld_4 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_5 : Internal_Entity_Expr;
Fld_6 : Internal_Entity_Assoc_List;
Fld_7 : Bare_Assoc_List;
Fld_8 : Internal_Entity_Info;
Get_Result_1 : Bare_Basic_Assoc;
As_Entity_1 : Internal_Entity_Basic_Assoc;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Basic_Assoc;
Fld_9 : Internal_Entity_Expr;
Fld_10 : Internal_Entity_Name;
Fld_11 : Logic_Equation;
Fld_12 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_13 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_14 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_15 : Logic_Var;
Fld_16 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_17 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Fld_18 : Logic_Equation;
And_Pred_5 : Logic_Equation;
Fld_19 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred_6 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1248 is
            begin
                     Dec_Ref (Fld_11);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_13);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_14);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Fld_16);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Fld_17);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Fld_18);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_6);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1248;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7916 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7892 '<FieldAccess for name_designated_type at ast.py:17509>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17509
--# expr-start 7891 '<FieldAccess for prefix at ast.py:17509>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17509
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7891
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_1 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 7892
Block_Var := Fld_1; 
--# bind typ Block_Var
--# expr-start 7894 '<FieldAccess for expr at ast.py:17510>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17510
--# expr-start 7893 '<FieldAccess for args_list at ast.py:17510>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17510
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_2 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 7893
Fld_3 := Fld_2.Node;
Get_Result := Get (Fld_3, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_4 := Fld_2.Info;
As_Entity := (Info => Fld_4, Node => Get_Result); 
   If_Result := As_Entity;
end if;
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_5 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 7894
Block_Var_1 := Fld_5; 
--# bind float_expr Block_Var_1
--# expr-start 7896 '<FieldAccess for expr at ast.py:17511>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17511
--# expr-start 7895 '<FieldAccess for args_list at ast.py:17511>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17511
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_6 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 7895
Fld_7 := Fld_6.Node;
Get_Result_1 := Get (Fld_7, 1, True); 
Is_Null_1 := Get_Result_1 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Basic_Assoc;
else
Fld_8 := Fld_6.Info;
As_Entity_1 := (Info => Fld_8, Node => Get_Result_1); 
   If_Result_1 := As_Entity_1;
end if;
      if If_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_9 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result_1.Node, E_Info => If_Result_1.Info);
--# end
--# expr-done 7896
Block_Var_2 := Fld_9; 
--# bind int_expr Block_Var_2
--# expr-start 7915 '<And at ast.py:17513>' And_Pred_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17513
--# expr-start 7912 '<And at ast.py:17513>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17513
--# expr-start 7910 '<And at ast.py:17513>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17513
--# expr-start 7908 '<And at ast.py:17513>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17513
--# expr-start 7905 '<And at ast.py:17513>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17513
--# expr-start 7903 '<And at ast.py:17513>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17513
--# expr-start 7901 '<And at ast.py:17513>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17513
--# expr-start 7898 '<FieldAccess for sub_equation at ast.py:17514>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17514
--# expr-start 7897 '<FieldAccess for prefix at ast.py:17514>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17514
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7897
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_11 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_10.Node, Env => Env, Origin => Origin, E_Info => Fld_10.Info);
--# end
--# expr-done 7898
--# expr-start 7900 '<Bind at ast.py:17516>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17516
--# expr-start 7899 '<FieldAccess for expected_type_var at ast.py:17516>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17516
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Block_Var_1.Node);
--# expr-done 7899
Fld_12.Value := No_Entity;
Entity_Vars.Reset (Fld_12);
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_12, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17516") else null)); 
--# expr-done 7900
And_Pred := Create_And (Fld_11, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17513") else null)); 
--# expr-done 7901
--# expr-start 7902 '<FieldAccess for sub_equation at ast.py:17517>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17517
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_13 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Block_Var_1.Node, Env => Env, Origin => Origin, E_Info => Block_Var_1.Info);
--# end
--# expr-done 7902
And_Pred_1 := Create_And (And_Pred, Fld_13, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17513") else null)); 
--# expr-done 7903
--# expr-start 7904 '<FieldAccess for matches_expected_type at ast.py:17518>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17518
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_14 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Block_Var_1.Node, Origin => Origin);
--# end
--# expr-done 7904
And_Pred_2 := Create_And (And_Pred_1, Fld_14, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17513") else null)); 
--# expr-done 7905
--# expr-start 7907 '<FieldAccess for universal_int_bind at ast.py:17520>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17520
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7906 '<FieldAccess for expected_type_var at ast.py:17520>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17520
      if Block_Var_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Block_Var_2.Node);
--# expr-done 7906
--# property-call-start AdaNode.universal_int_bind
Fld_16 := Libadalang.Implementation.Ada_Node_P_Universal_Int_Bind (Node => Self, Type_Var => Fld_15, Origin => Origin);
--# end
--# expr-done 7907
And_Pred_3 := Create_And (And_Pred_2, Fld_16, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17513") else null)); 
--# expr-done 7908
--# expr-start 7909 '<FieldAccess for sub_equation at ast.py:17521>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17521
      if Block_Var_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_17 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Block_Var_2.Node, Env => Env, Origin => Origin, E_Info => Block_Var_2.Info);
--# end
--# expr-done 7909
And_Pred_4 := Create_And (And_Pred_3, Fld_17, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17513") else null)); 
--# expr-done 7910
--# expr-start 7911 '<FieldAccess for matches_expected_type at ast.py:17522>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17522
      if Block_Var_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_18 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Block_Var_2.Node, Origin => Origin);
--# end
--# expr-done 7911
And_Pred_5 := Create_And (And_Pred_4, Fld_18, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17513") else null)); 
--# expr-done 7912
--# expr-start 7914 '<Bind at ast.py:17524>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17524
--# expr-start 7913 '<FieldAccess for type_var at ast.py:17524>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17524
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 7913
Fld_19.Value := No_Entity;
Entity_Vars.Reset (Fld_19);
Cast_Expr_1 := Block_Var; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_19, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17524") else null)); 
--# expr-done 7914
And_Pred_6 := Create_And (And_Pred_5, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17513") else null)); 
--# expr-done 7915
Let_Result := And_Pred_6; Inc_Ref (Let_Result);
--# end
--# expr-done 7916
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1248;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1248;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Float_Universal_Int_Equation;
--# end
--# property-start AttributeRef.mod_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17528
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Mod_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1249 with Inline_Always;
            procedure Finalizer_Scope_2770 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Name;
Fld_3 : Logic_Equation;
Fld_4 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_5 : Internal_Entity_Assoc_List;
Fld_6 : Bare_Assoc_List;
Fld_7 : Internal_Entity_Info;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_8 : Internal_Entity_Expr;
Fld_9 : Logic_Equation;
Item_249 : Internal_Entity_Basic_Assoc;
Bare_Item_249 : Bare_Basic_Assoc;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1249 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1249;
            procedure Finalizer_Scope_2770 is
            begin
                     Dec_Ref (Fld_9);
            end Finalizer_Scope_2770;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7930 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7918 '<FieldAccess for name_designated_type at ast.py:17533>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17533
--# expr-start 7917 '<FieldAccess for prefix at ast.py:17533>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17533
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7917
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_1 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 7918
Block_Var := Fld_1; 
--# bind typ Block_Var
--# expr-start 7929 '<And at ast.py:17534>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17534
--# expr-start 7923 '<And at ast.py:17534>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17534
--# expr-start 7920 '<FieldAccess for sub_equation at ast.py:17535>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17535
--# expr-start 7919 '<FieldAccess for prefix at ast.py:17535>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17535
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7919
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_2.Node, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 7920
--# expr-start 7922 '<Bind at ast.py:17536>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17536
--# expr-start 7921 '<FieldAccess for type_var at ast.py:17536>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17536
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 7921
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_4, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17536") else null)); 
--# expr-done 7922
And_Pred := Create_And (Fld_3, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17534") else null)); 
--# expr-done 7923
--# expr-start 7928 '<LogicAll at ast.py:17537>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17537
--# expr-start 7927 '<Map at ast.py:17537>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17537
   --# expr-start 7924 '<FieldAccess for args_list at ast.py:17537>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17537
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_5 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 7924
Fld_6 := Fld_5.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_6 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Assoc_List := Fld_6;
      begin
         for Untyped_Item_249 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_249 := Untyped_Item_249; 
Is_Null := Bare_Item_249 = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_7 := Fld_5.Info;
As_Entity := (Info => Fld_7, Node => Bare_Item_249); 
   If_Result := As_Entity;
end if;
                  Item_249 := If_Result; 
   --# scope-start
               --# bind arg Item_249
      --# expr-start 7926 '<FieldAccess for sub_equation at ast.py:17538>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17538
--# expr-start 7925 '<FieldAccess for expr at ast.py:17538>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17538
      if Item_249.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_8 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Item_249.Node, E_Info => Item_249.Info);
--# end
--# expr-done 7925
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_9 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_8.Node, Env => Env, Origin => Origin, E_Info => Fld_8.Info);
--# end
--# expr-done 7926
         declare
            Item_To_Append : constant Logic_Equation := Fld_9;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2770;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 7927
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17537") else null)); 
--# expr-done 7928
And_Pred_1 := Create_And (And_Pred, Logic_Boolean_Op, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17534") else null)); 
--# expr-done 7929
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 7930
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1249;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1249;
               Finalizer_Scope_2770;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Mod_Equation;
--# end
--# property-start AttributeRef.type_class_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17543
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Type_Class_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1250 with Inline_Always;
            procedure Finalizer_Scope_2771 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Var_Expr : Bare_Basic_Decl;
Sym : Symbol_Type;
Sym_1 : Symbol_Type;
Array_Lit : Symbol_Type_Array_Access;
Fld : Bare_Basic_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Sym_2 : Symbol_Type;
Env_Get_Result : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Name;
Fld_3 : Logic_Equation;
Fld_4 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1250 is
            begin
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1250;
            procedure Finalizer_Scope_2771 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2771;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7944 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7938 '<Cast to BaseTypeDecl at ast.py:17549>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17549
--# expr-start 7937 '<EnvGet at ast.py:17549>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17549
--# expr-start 7934 '<FieldAccess for get_unit_root_decl at ast.py:17549>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17549
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7933 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 7931 '<Symbol System at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_System); 
--# expr-done 7931
--# expr-start 7932 '<Symbol Aux_DEC at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Aux_Dec); 
--# expr-done 7932
Array_Lit := Create_Symbol_Type_Array (Internal_Symbol_Type_Array'(1 => Sym, 2 => Sym_1)); 
--# expr-done 7933
--# property-call-start AdaNode.get_unit_root_decl
Fld := Libadalang.Implementation.Ada_Node_P_Get_Unit_Root_Decl (Node => Ent.Node, Name => Array_Lit, Kind => Unit_Specification);
--# end
--# expr-done 7934
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 7935 '<FieldAccess for children_env at ast.py:17549>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17549
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Children_Env (Node => Var_Expr);
--# expr-done 7935
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2771;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
--# expr-start 7936 '<Symbol Type_Class at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Type_Class); 
--# expr-done 7936
Env_Get_Result := AST_Envs.Get_First (Self => Result_Var, Key => Sym_2, Lookup_Kind => To_Lookup_Kind_Type (Flat), Categories => (Inherited_Primitives => True, Nocat => True)); 
--# expr-done 7937
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 7938
Block_Var := Cast_Result; 
--# bind typ Block_Var
--# expr-start 7943 '<And at ast.py:17555>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17555
--# expr-start 7940 '<FieldAccess for xref_equation at ast.py:17555>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17555
--# expr-start 7939 '<FieldAccess for prefix at ast.py:17555>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17555
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7939
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_3 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Fld_2.Node, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 7940
--# expr-start 7942 '<Bind at ast.py:17555>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17555
--# expr-start 7941 '<FieldAccess for type_var at ast.py:17555>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17555
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 7941
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
Cast_Expr := Block_Var; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_4, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17555") else null)); 
--# expr-done 7942
And_Pred := Create_And (Fld_3, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17555") else null)); 
--# expr-done 7943
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 7944
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1250;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1250;
               Finalizer_Scope_2771;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Type_Class_Equation;
--# end
--# property-start AttributeRef.storage_pool_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17558
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Storage_Pool_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1251 with Inline_Always;
            procedure Finalizer_Scope_2772 with Inline_Always;
      Block_Var : Internal_Entity_Classwide_Type_Decl;
Var_Expr : Bare_Basic_Decl;
Sym : Symbol_Type;
Sym_1 : Symbol_Type;
Array_Lit : Symbol_Type_Array_Access;
Fld : Bare_Basic_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Sym_2 : Symbol_Type;
Env_Get_Result : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Classwide_Type_Decl;
Fld_3 : Internal_Entity_Name;
Fld_4 : Logic_Equation;
Fld_5 : Logic_Var;
Cast_Expr : Internal_Entity_Classwide_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1251 is
            begin
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1251;
            procedure Finalizer_Scope_2772 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2772;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7959 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7953 '<FieldAccess for classwide_type at ast.py:17563>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17563
--# expr-start 7952 '<Cast to BaseTypeDecl at ast.py:17563>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17563
--# expr-start 7951 '<EnvGet at ast.py:17563>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17563
--# expr-start 7948 '<FieldAccess for get_unit_root_decl at ast.py:17563>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17563
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7947 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 7945 '<Symbol System at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_System); 
--# expr-done 7945
--# expr-start 7946 '<Symbol Storage_Pools at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Storage_Pools); 
--# expr-done 7946
Array_Lit := Create_Symbol_Type_Array (Internal_Symbol_Type_Array'(1 => Sym, 2 => Sym_1)); 
--# expr-done 7947
--# property-call-start AdaNode.get_unit_root_decl
Fld := Libadalang.Implementation.Ada_Node_P_Get_Unit_Root_Decl (Node => Ent.Node, Name => Array_Lit, Kind => Unit_Specification);
--# end
--# expr-done 7948
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 7949 '<FieldAccess for children_env at ast.py:17563>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17563
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Children_Env (Node => Var_Expr);
--# expr-done 7949
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2772;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
--# expr-start 7950 '<Symbol Root_Storage_Pool at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Root_Storage_Pool); 
--# expr-done 7950
Env_Get_Result := AST_Envs.Get_First (Self => Result_Var, Key => Sym_2, Lookup_Kind => To_Lookup_Kind_Type (Flat), Categories => (Inherited_Primitives => True, Nocat => True)); 
--# expr-done 7951
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 7952
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.classwide_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Classwide_Type (Node => Cast_Result.Node, E_Info => Cast_Result.Info);
--# end
--# expr-done 7953
Block_Var := Fld_2; 
--# bind typ Block_Var
--# expr-start 7958 '<And at ast.py:17569>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17569
--# expr-start 7955 '<FieldAccess for xref_equation at ast.py:17569>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17569
--# expr-start 7954 '<FieldAccess for prefix at ast.py:17569>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17569
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7954
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_4 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Fld_3.Node, Env => Env, Origin => Origin, E_Info => Fld_3.Info);
--# end
--# expr-done 7955
--# expr-start 7957 '<Bind at ast.py:17569>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17569
--# expr-start 7956 '<FieldAccess for type_var at ast.py:17569>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17569
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 7956
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
Cast_Expr := Block_Var; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_5, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17569") else null)); 
--# expr-done 7957
And_Pred := Create_And (Fld_4, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17569") else null)); 
--# expr-done 7958
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 7959
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1251;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1251;
               Finalizer_Scope_2772;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Storage_Pool_Equation;
--# end
--# property-start AttributeRef.scalar_storage_order_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17572
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Scalar_Storage_Order_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1252 with Inline_Always;
            procedure Finalizer_Scope_2773 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Var_Expr : Bare_Basic_Decl;
Sym : Symbol_Type;
Array_Lit : Symbol_Type_Array_Access;
Fld : Bare_Basic_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Sym_1 : Symbol_Type;
Env_Get_Result : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Name;
Fld_3 : Logic_Equation;
Fld_4 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1252 is
            begin
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1252;
            procedure Finalizer_Scope_2773 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2773;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7972 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7966 '<Cast to BaseTypeDecl at ast.py:17577>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17577
--# expr-start 7965 '<EnvGet at ast.py:17577>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17577
--# expr-start 7962 '<FieldAccess for get_unit_root_decl at ast.py:17577>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17577
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 7961 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 7960 '<Symbol System at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_System); 
--# expr-done 7960
Array_Lit := Create_Symbol_Type_Array (Internal_Symbol_Type_Array'(1 => Sym)); 
--# expr-done 7961
--# property-call-start AdaNode.get_unit_root_decl
Fld := Libadalang.Implementation.Ada_Node_P_Get_Unit_Root_Decl (Node => Ent.Node, Name => Array_Lit, Kind => Unit_Specification);
--# end
--# expr-done 7962
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 7963 '<FieldAccess for children_env at ast.py:17577>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17577
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Children_Env (Node => Var_Expr);
--# expr-done 7963
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2773;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
--# expr-start 7964 '<Symbol Bit_Order at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Bit_Order); 
--# expr-done 7964
Env_Get_Result := AST_Envs.Get_First (Self => Result_Var, Key => Sym_1, Lookup_Kind => To_Lookup_Kind_Type (Flat), Categories => (Inherited_Primitives => True, Nocat => True)); 
--# expr-done 7965
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 7966
Block_Var := Cast_Result; 
--# bind typ Block_Var
--# expr-start 7971 '<And at ast.py:17583>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17583
--# expr-start 7968 '<FieldAccess for xref_equation at ast.py:17583>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17583
--# expr-start 7967 '<FieldAccess for prefix at ast.py:17583>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17583
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7967
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_3 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Fld_2.Node, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 7968
--# expr-start 7970 '<Bind at ast.py:17583>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17583
--# expr-start 7969 '<FieldAccess for type_var at ast.py:17583>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17583
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 7969
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
Cast_Expr := Block_Var; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_4, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17583") else null)); 
--# expr-done 7970
And_Pred := Create_And (Fld_3, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17583") else null)); 
--# expr-done 7971
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 7972
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1252;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1252;
               Finalizer_Scope_2773;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Scalar_Storage_Order_Equation;
--# end
--# property-start AttributeRef.bind_to_prefix_eq /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17586
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Bind_To_Prefix_Eq
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1253 with Inline_Always;
      Fld : Bare_Name;
Fld_1 : Logic_Var;
Fld_2 : Logic_Var;
Bind_Result : Logic_Equation;
Fld_3 : Internal_Entity_Name;
Fld_4 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_5 : Logic_Var;
Fld_6 : Bare_Name;
Fld_7 : Logic_Var;
Bind_Result_1 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1253 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_4);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1253;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 7985 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7984 '<And at ast.py:17587>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17587
--# expr-start 7979 '<And at ast.py:17587>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17587
--# expr-start 7976 '<Bind at ast.py:17588>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17588
--# expr-start 7974 '<FieldAccess for expected_type_var at ast.py:17588>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17588
--# expr-start 7973 '<FieldAccess for prefix at ast.py:17588>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17588
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Attribute_Ref_F_Prefix;
--# expr-done 7973
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld);
--# expr-done 7974
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 7975 '<FieldAccess for expected_type_var at ast.py:17588>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17588
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 7975
Fld_2.Value := No_Entity;
Entity_Vars.Reset (Fld_2);
Bind_Result := Solver.Create_Unify (Fld_1, Fld_2, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17588") else null)); 
--# expr-done 7976
--# expr-start 7978 '<FieldAccess for sub_equation at ast.py:17589>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17589
--# expr-start 7977 '<FieldAccess for prefix at ast.py:17589>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17589
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7977
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_4 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_3.Node, Env => Env, Origin => Origin, E_Info => Fld_3.Info);
--# end
--# expr-done 7978
And_Pred := Create_And (Bind_Result, Fld_4, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17587") else null)); 
--# expr-done 7979
--# expr-start 7983 '<Bind at ast.py:17590>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17590
--# expr-start 7980 '<FieldAccess for type_var at ast.py:17590>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17590
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 7980
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
--# expr-start 7982 '<FieldAccess for type_var at ast.py:17590>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17590
--# expr-start 7981 '<FieldAccess for prefix at ast.py:17590>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17590
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Self.Attribute_Ref_F_Prefix;
--# expr-done 7981
      if Fld_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_6);
--# expr-done 7982
Fld_7.Value := No_Entity;
Entity_Vars.Reset (Fld_7);
Bind_Result_1 := Solver.Create_Unify (Fld_5, Fld_7, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17590") else null)); 
--# expr-done 7983
And_Pred_1 := Create_And (And_Pred, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17587") else null)); 
--# expr-done 7984
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 7985
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1253;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1253;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Bind_To_Prefix_Eq;
--# end
--# property-start AttributeRef.result_attr_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17594
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Result_Attr_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1254 with Inline_Always;
      Block_Var : Internal_Entity_Basic_Decl;
Block_Var_1 : Internal_Entity_Base_Type_Decl;
Var_Expr : Internal_Entity_Base_Subp_Spec;
Env_Node : Bare_Ada_Node;
Node_Parents : Bare_Ada_Node_Array_Access;
Is_A : Boolean;
Find_Result : Bare_Ada_Node;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Base_Subp_Spec;
Fld_1 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Fld_2 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_3 : Internal_Entity_Name;
Fld_4 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Basic_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1254 is
            begin
                     Dec_Ref (Node_Parents);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1254;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8000 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 7991 '<Cast to BasicDecl at ast.py:17598>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17598
--# expr-start 7990 '<as_entity at ast.py:17598>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17598
--# expr-start 7989 '<Find at ast.py:17598>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17598
   --# expr-start 7987 '<FieldAccess for parents at ast.py:17598>' Node_Parents /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17598
--# expr-start 7986 '<env_node at ast.py:17598>' Env_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17598
Env_Node := AST_Envs.Env_Node (Env); 
--# expr-done 7986
      if Env_Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents := Parents (Env_Node, True); 
--# expr-done 7987
      Find_Result := No_Bare_Ada_Node;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents;
      begin
         for Item_250 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_250
      --# expr-start 7988 '<IsA BasicSubpDecl, BaseSubpBody at ast.py:17599>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17599
Is_A := Item_250 /= null 
and then Item_250.Kind in Ada_Basic_Subp_Decl | Ada_Base_Subp_Body; 
--# expr-done 7988
      if Is_A then
         Find_Result := Item_250;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 7989
Is_Null := Find_Result = null; 
if Is_Null then
   If_Result := No_Entity;
else
As_Entity := (Info => E_Info, Node => Find_Result); 
   If_Result := As_Entity;
end if;
--# expr-done 7990
   if If_Result.Node = null
      or else If_Result.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => If_Result.Node,
         Info => If_Result.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 7991
Block_Var := Cast_Result; 
--# bind containing_subp Block_Var
--# expr-start 7992 '<FieldAccess for subp_spec_or_null at ast.py:17602>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17602
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 7992
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Subp_Spec then
   --# scope-start
   --# bind ss Var_Expr
--# expr-start 7993 '<FieldAccess for return_type at ast.py:17603>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17603
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.return_type
Fld_1 := Libadalang.Implementation.Base_Subp_Spec_P_Return_Type (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 7993
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Block_Var_1 := Result_Var; 
--# bind returns Block_Var_1
--# expr-start 7999 '<And at ast.py:17606>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17606
--# expr-start 7995 '<Bind at ast.py:17607>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17607
--# expr-start 7994 '<FieldAccess for type_var at ast.py:17607>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17607
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 7994
Fld_2.Value := No_Entity;
Entity_Vars.Reset (Fld_2);
Cast_Expr := Block_Var_1; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_2, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17607") else null)); 
--# expr-done 7995
--# expr-start 7998 '<Bind at ast.py:17608>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17608
--# expr-start 7997 '<FieldAccess for ref_var at ast.py:17608>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17608
--# expr-start 7996 '<FieldAccess for prefix at ast.py:17608>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17608
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 7996
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_4 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_3.Node);
--# end
--# expr-done 7997
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
Cast_Expr_1 := Block_Var; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_4, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17608") else null)); 
--# expr-done 7998
And_Pred := Create_And (Bind_Result, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17606") else null)); 
--# expr-done 7999
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 8000
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1254;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1254;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Result_Attr_Equation;
--# end
--# property-start AttributeRef.tag_attr_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17612
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Tag_Attr_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1255 with Inline_Always;
            procedure Finalizer_Scope_2776 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Var_Expr : Bare_Basic_Decl;
Sym : Symbol_Type;
Sym_1 : Symbol_Type;
Array_Lit : Symbol_Type_Array_Access;
Fld : Bare_Basic_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Sym_2 : Symbol_Type;
Env_Get_Result : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Name;
Fld_3 : Logic_Equation;
Fld_4 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1255 is
            begin
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1255;
            procedure Finalizer_Scope_2776 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2776;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8014 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8008 '<Cast to BaseTypeDecl at ast.py:17614>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17614
--# expr-start 8007 '<EnvGet at ast.py:17614>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17614
--# expr-start 8004 '<FieldAccess for get_unit_root_decl at ast.py:17614>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17614
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8003 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 8001 '<Symbol Ada at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Ada); 
--# expr-done 8001
--# expr-start 8002 '<Symbol Tags at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Tags); 
--# expr-done 8002
Array_Lit := Create_Symbol_Type_Array (Internal_Symbol_Type_Array'(1 => Sym, 2 => Sym_1)); 
--# expr-done 8003
--# property-call-start AdaNode.get_unit_root_decl
Fld := Libadalang.Implementation.Ada_Node_P_Get_Unit_Root_Decl (Node => Ent.Node, Name => Array_Lit, Kind => Unit_Specification);
--# end
--# expr-done 8004
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 8005 '<FieldAccess for children_env at ast.py:17614>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17614
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Children_Env (Node => Var_Expr);
--# expr-done 8005
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2776;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
--# expr-start 8006 '<Symbol Tag at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Tag); 
--# expr-done 8006
Env_Get_Result := AST_Envs.Get_First (Self => Result_Var, Key => Sym_2, Lookup_Kind => To_Lookup_Kind_Type (Flat), Categories => (Inherited_Primitives => True, Nocat => True)); 
--# expr-done 8007
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 8008
Block_Var := Cast_Result; 
--# bind tag_type Block_Var
--# expr-start 8013 '<And at ast.py:17622>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17622
--# expr-start 8010 '<FieldAccess for xref_equation at ast.py:17622>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17622
--# expr-start 8009 '<FieldAccess for prefix at ast.py:17622>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17622
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8009
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_3 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Fld_2.Node, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 8010
--# expr-start 8012 '<Bind at ast.py:17625>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17625
--# expr-start 8011 '<FieldAccess for type_var at ast.py:17625>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17625
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8011
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
Cast_Expr := Block_Var; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_4, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17625") else null)); 
--# expr-done 8012
And_Pred := Create_And (Fld_3, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17622") else null)); 
--# expr-done 8013
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 8014
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1255;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1255;
               Finalizer_Scope_2776;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Tag_Attr_Equation;
--# end
--# property-start AttributeRef.stream_attrs_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17629
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Stream_Attrs_Equation
  (Node : Bare_Attribute_Ref
      ; Return_Obj : Boolean
         := False
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind return_obj Return_Obj
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1256 with Inline_Always;
            procedure Finalizer_Scope_2777 with Inline_Always;
            procedure Finalizer_Scope_2779 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Block_Var_1 : Internal_Entity_Base_Type_Decl;
Block_Var_2 : Internal_Entity_Expr;
Block_Var_3 : Internal_Entity_Expr;
Var_Expr : Bare_Basic_Decl;
Var_Expr_1 : Internal_Entity_Basic_Assoc;
Var_Expr_2 : Internal_Entity_Attribute_Def_Clause;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Base_Type_Decl;
Sym : Symbol_Type;
Sym_1 : Symbol_Type;
Array_Lit : Symbol_Type_Array_Access;
Fld_2 : Bare_Basic_Decl;
Fld_3 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Sym_2 : Symbol_Type;
Env_Get_Result : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
Fld_4 : Internal_Entity_Classwide_Type_Decl;
Cast_Result_1 : Internal_Entity_Base_Type_Decl;
Fld_5 : Internal_Entity_Assoc_List;
Fld_6 : Bare_Assoc_List;
Fld_7 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_8 : Internal_Entity_Expr;
Fld_9 : Internal_Entity_Assoc_List;
Fld_10 : Bare_Assoc_List;
Fld_11 : Internal_Entity_Info;
Get_Result_1 : Bare_Basic_Assoc;
As_Entity_1 : Internal_Entity_Basic_Assoc;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Basic_Assoc;
Fld_12 : Internal_Entity_Expr;
Scope_Result_1 : Internal_Entity_Expr;
Result_Var_1 : Internal_Entity_Expr;
Fld_13 : Internal_Entity_Name;
Fld_14 : Logic_Equation;
Fld_15 : Logic_Var;
Fld_16 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_17 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_18 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_19 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
Fld_20 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_4 : Internal_Entity;
Bind_Result_2 : Logic_Equation;
Fld_21 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_22 : Logic_Equation;
And_Pred_4 : Logic_Equation;
If_Result_2 : Logic_Equation;
And_Pred_5 : Logic_Equation;
Bound_Imprecise_Fallback : Boolean;
Fld_23 : Internal_Entity_Identifier;
Fld_24 : Symbol_Type;
Fld_25 : Internal_Entity_Attribute_Def_Clause;
Fld_26 : Internal_Entity_Identifier;
Fld_27 : Logic_Var;
Fld_28 : Internal_Entity_Expr;
Cast_Result_5 : Internal_Entity_Name;
Fld_29 : Internal_Entity_Basic_Decl;
Cast_Result_6 : Internal_Entity;
Bind_Result_3 : Logic_Equation;
Scope_Result_2 : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var_2 : Logic_Equation;
Dyn_Var_Bind_Result : Logic_Equation;
And_Pred_6 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1256 is
            begin
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Fld_14);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_17);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_18);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (Fld_21);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Fld_22);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (If_Result_2);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var_2);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (And_Pred_6);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1256;
            procedure Finalizer_Scope_2777 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_2777;
            procedure Finalizer_Scope_2779 is
            begin
                     Dec_Ref (Bind_Result_3);
            end Finalizer_Scope_2779;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8062 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8016 '<FieldAccess for name_designated_type at ast.py:17630>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17630
--# expr-start 8015 '<FieldAccess for prefix at ast.py:17630>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17630
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8015
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_1 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 8016
Block_Var := Fld_1; 
--# bind typ Block_Var
--# expr-start 8026 '<Cast to BaseTypeDecl at ast.py:17633>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17633
--# expr-start 8025 '<FieldAccess for classwide_type at ast.py:17633>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17633
--# expr-start 8024 '<Cast to BaseTypeDecl at ast.py:17633>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17633
--# expr-start 8023 '<EnvGet at ast.py:17633>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17633
--# expr-start 8020 '<FieldAccess for get_unit_root_decl at ast.py:17633>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17633
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8019 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 8017 '<Symbol Ada at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Ada); 
--# expr-done 8017
--# expr-start 8018 '<Symbol Streams at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Streams); 
--# expr-done 8018
Array_Lit := Create_Symbol_Type_Array (Internal_Symbol_Type_Array'(1 => Sym, 2 => Sym_1)); 
--# expr-done 8019
--# property-call-start AdaNode.get_unit_root_decl
Fld_2 := Libadalang.Implementation.Ada_Node_P_Get_Unit_Root_Decl (Node => Ent.Node, Name => Array_Lit, Kind => Unit_Specification);
--# end
--# expr-done 8020
Var_Expr := Fld_2;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 8021 '<FieldAccess for children_env at ast.py:17633>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17633
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Children_Env (Node => Var_Expr);
--# expr-done 8021
Scope_Result := Fld_3;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2777;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
--# expr-start 8022 '<Symbol Root_Stream_Type at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Root_Stream_Type); 
--# expr-done 8022
Env_Get_Result := AST_Envs.Get_First (Self => Result_Var, Key => Sym_2, Lookup_Kind => To_Lookup_Kind_Type (Flat), Categories => (Inherited_Primitives => True, Nocat => True)); 
--# expr-done 8023
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 8024
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.classwide_type'
Fld_4 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Classwide_Type (Node => Cast_Result.Node, E_Info => Cast_Result.Info);
--# end
--# expr-done 8025
      Cast_Result_1 := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld_4.Node,
         Info => Fld_4.Info);
--# expr-done 8026
Block_Var_1 := Cast_Result_1; 
--# bind root_stream_type Block_Var_1
--# expr-start 8028 '<FieldAccess for expr at ast.py:17639>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17639
--# expr-start 8027 '<FieldAccess for args_list at ast.py:17639>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17639
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_5 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8027
Fld_6 := Fld_5.Node;
Get_Result := Get (Fld_6, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_7 := Fld_5.Info;
As_Entity := (Info => Fld_7, Node => Get_Result); 
   If_Result := As_Entity;
end if;
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_8 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 8028
Block_Var_2 := Fld_8; 
--# bind stream_arg Block_Var_2
--# expr-start 8029 '<FieldAccess for args_list at ast.py:17640>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17640
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_9 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8029
Fld_10 := Fld_9.Node;
Get_Result_1 := Get (Fld_10, 1, True); 
Is_Null_1 := Get_Result_1 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Basic_Assoc;
else
Fld_11 := Fld_9.Info;
As_Entity_1 := (Info => Fld_11, Node => Get_Result_1); 
   If_Result_1 := As_Entity_1;
end if;
Var_Expr_1 := If_Result_1;
if Var_Expr_1 /= No_Entity_Basic_Assoc then
   --# scope-start
--# expr-start 8030 '<FieldAccess for expr at ast.py:17640>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17640
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_12 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 8030
Scope_Result_1 := Fld_12;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Expr;
end if;
Block_Var_3 := Result_Var_1; 
--# bind obj_arg Block_Var_3
--# expr-start 8061 '<And at ast.py:17643>' And_Pred_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17643
--# expr-start 8050 '<And at ast.py:17643>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17643
--# expr-start 8040 '<And at ast.py:17643>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17643
--# expr-start 8038 '<And at ast.py:17643>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17643
--# expr-start 8036 '<And at ast.py:17643>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17643
--# expr-start 8032 '<FieldAccess for sub_equation at ast.py:17643>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17643
--# expr-start 8031 '<FieldAccess for prefix at ast.py:17643>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17643
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8031
      if Fld_13.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_14 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_13.Node, Env => Env, Origin => Origin, E_Info => Fld_13.Info);
--# end
--# expr-done 8032
--# expr-start 8035 '<Bind at ast.py:17644>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17644
--# expr-start 8033 '<FieldAccess for expected_type_var at ast.py:17644>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17644
      if Block_Var_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Block_Var_2.Node);
--# expr-done 8033
Fld_15.Value := No_Entity;
Entity_Vars.Reset (Fld_15);
--# expr-start 8034 '<FieldAccess for anonymous_access_type at ast.py:17645>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17645
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.anonymous_access_type
Fld_16 := Libadalang.Implementation.Base_Type_Decl_P_Anonymous_Access_Type (Node => Block_Var_1.Node, E_Info => Block_Var_1.Info);
--# end
--# expr-done 8034
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_16.Node,
         Info => Fld_16.Info);
Bind_Result := Solver.Create_Assign (Fld_15, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17644") else null)); 
--# expr-done 8035
And_Pred := Create_And (Fld_14, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17643") else null)); 
--# expr-done 8036
--# expr-start 8037 '<FieldAccess for sub_equation at ast.py:17646>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17646
      if Block_Var_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_17 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Block_Var_2.Node, Env => Env, Origin => Origin, E_Info => Block_Var_2.Info);
--# end
--# expr-done 8037
And_Pred_1 := Create_And (And_Pred, Fld_17, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17643") else null)); 
--# expr-done 8038
--# expr-start 8039 '<FieldAccess for matches_expected_type at ast.py:17647>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17647
      if Block_Var_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_18 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Block_Var_2.Node, Origin => Origin);
--# end
--# expr-done 8039
And_Pred_2 := Create_And (And_Pred_1, Fld_18, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17643") else null)); 
--# expr-done 8040
--# expr-start 8049 '<If at ast.py:17649>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17649
if Return_Obj then
   --# expr-start 8042 '<Bind at ast.py:17651>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17651
--# expr-start 8041 '<FieldAccess for type_var at ast.py:17651>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17651
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8041
Fld_19.Value := No_Entity;
Entity_Vars.Reset (Fld_19);
Cast_Expr := Block_Var; 
      Cast_Result_3 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_19, Cast_Result_3, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17651") else null)); 
--# expr-done 8042
   If_Result_2 := Bind_Result_1;
else
   --# expr-start 8048 '<And at ast.py:17652>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17652
--# expr-start 8046 '<And at ast.py:17652>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17652
--# expr-start 8044 '<Bind at ast.py:17652>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17652
--# expr-start 8043 '<FieldAccess for expected_type_var at ast.py:17652>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17652
      if Block_Var_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_20 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Block_Var_3.Node);
--# expr-done 8043
Fld_20.Value := No_Entity;
Entity_Vars.Reset (Fld_20);
Cast_Expr_1 := Block_Var; 
      Cast_Result_4 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_2 := Solver.Create_Assign (Fld_20, Cast_Result_4, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17652") else null)); 
--# expr-done 8044
--# expr-start 8045 '<FieldAccess for sub_equation at ast.py:17653>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17653
      if Block_Var_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_21 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Block_Var_3.Node, Env => Env, Origin => Origin, E_Info => Block_Var_3.Info);
--# end
--# expr-done 8045
And_Pred_3 := Create_And (Bind_Result_2, Fld_21, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17652") else null)); 
--# expr-done 8046
--# expr-start 8047 '<FieldAccess for matches_expected_type at ast.py:17654>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17654
      if Block_Var_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_22 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Block_Var_3.Node, Origin => Origin);
--# end
--# expr-done 8047
And_Pred_4 := Create_And (And_Pred_3, Fld_22, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17652") else null)); 
--# expr-done 8048
   If_Result_2 := And_Pred_4;
end if;
      Inc_Ref (If_Result_2);
--# expr-done 8049
And_Pred_5 := Create_And (And_Pred_2, If_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17643") else null)); 
--# expr-done 8050
--# expr-start 8060 '<bind at ast.py:17660>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17660
Bound_Imprecise_Fallback := False; 
--# expr-start 8053 '<FieldAccess for get_representation_clause at ast.py:17662>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17662
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8052 '<FieldAccess for name_symbol at ast.py:17663>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17663
--# expr-start 8051 '<FieldAccess for attribute at ast.py:17663>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17663
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_23 := Create_Internal_Entity_Identifier (Node => Ent.Node.Attribute_Ref_F_Attribute, Info => Ent.Info);
--# expr-done 8051
      if Fld_23.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_24 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_23.Node);
--# end
--# expr-done 8052
--# property-call-start BasicDecl.get_representation_clause
Fld_25 := Libadalang.Implementation.Basic_Decl_P_Get_Representation_Clause (Node => Block_Var.Node, Name => Fld_24, Imprecise_Fallback => Bound_Imprecise_Fallback, E_Info => Block_Var.Info);
--# end
--# expr-done 8053
Var_Expr_2 := Fld_25;
if Var_Expr_2 /= No_Entity_Attribute_Def_Clause then
   --# scope-start
   --# bind x Var_Expr_2
--# expr-start 8059 '<Bind at ast.py:17665>' Bind_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17665
--# expr-start 8055 '<FieldAccess for ref_var at ast.py:17666>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17666
--# expr-start 8054 '<FieldAccess for attribute at ast.py:17666>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17666
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_26 := Create_Internal_Entity_Identifier (Node => Ent.Node.Attribute_Ref_F_Attribute, Info => Ent.Info);
--# expr-done 8054
      if Fld_26.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_27 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_26.Node);
--# end
--# expr-done 8055
Fld_27.Value := No_Entity;
Entity_Vars.Reset (Fld_27);
--# expr-start 8058 '<FieldAccess for referenced_decl at ast.py:17667>' Fld_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17667
--# expr-start 8057 '<Cast to Name at ast.py:17667>' Cast_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17667
--# expr-start 8056 '<FieldAccess for expr at ast.py:17667>' Fld_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17667
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_28 := Create_Internal_Entity_Expr (Node => Var_Expr_2.Node.Attribute_Def_Clause_F_Expr, Info => Var_Expr_2.Info);
--# expr-done 8056
   if Fld_28.Node = null
      or else Fld_28.Node.Kind in Ada_Name
   then
      Cast_Result_5 := Create_Internal_Entity_Name
        (Node => Fld_28.Node,
         Info => Fld_28.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 8057
      if Cast_Result_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld_29 := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Cast_Result_5.Node, Imprecise_Fallback => Bound_Imprecise_Fallback, E_Info => Cast_Result_5.Info);
--# end
--# expr-done 8058
      Cast_Result_6 := Create_Internal_Entity
        (Node => Fld_29.Node,
         Info => Fld_29.Info);
Bind_Result_3 := Solver.Create_Assign (Fld_27, Cast_Result_6, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17665") else null)); 
--# expr-done 8059
Scope_Result_2 := Bind_Result_3;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2779;
   Result_Var_2 := Scope_Result_2;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17669") else null)); 
   Result_Var_2 := True_Rel;
end if;
      Inc_Ref (Result_Var_2);
Dyn_Var_Bind_Result := Result_Var_2; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 8060
And_Pred_6 := Create_And (And_Pred_5, Dyn_Var_Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17643") else null)); 
--# expr-done 8061
Let_Result := And_Pred_6; Inc_Ref (Let_Result);
--# end
--# expr-done 8062
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1256;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1256;
               Finalizer_Scope_2777;
               Finalizer_Scope_2779;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Stream_Attrs_Equation;
--# end
--# property-start AttributeRef.address_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17675
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Address_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1257 with Inline_Always;
            procedure Finalizer_Scope_2780 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Var_Expr : Bare_Basic_Decl;
Sym : Symbol_Type;
Array_Lit : Symbol_Type_Array_Access;
Fld : Bare_Basic_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Sym_1 : Symbol_Type;
Env_Get_Result : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Name;
Fld_3 : Logic_Equation;
Fld_4 : Bare_Name;
Fld_5 : Logic_Var;
Pred : Logic_Equation;
And_Pred : Logic_Equation;
Fld_6 : Internal_Entity_Name;
Fld_7 : Logic_Equation;
Or_Pred : Logic_Equation;
Fld_8 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1257 is
            begin
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Pred);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_7);
                     Dec_Ref (Or_Pred);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1257;
            procedure Finalizer_Scope_2780 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2780;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8082 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8069 '<Cast to BaseTypeDecl at ast.py:17677>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17677
--# expr-start 8068 '<EnvGet at ast.py:17677>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17677
--# expr-start 8065 '<FieldAccess for get_unit_root_decl at ast.py:17677>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17677
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8064 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 8063 '<Symbol System at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_System); 
--# expr-done 8063
Array_Lit := Create_Symbol_Type_Array (Internal_Symbol_Type_Array'(1 => Sym)); 
--# expr-done 8064
--# property-call-start AdaNode.get_unit_root_decl
Fld := Libadalang.Implementation.Ada_Node_P_Get_Unit_Root_Decl (Node => Ent.Node, Name => Array_Lit, Kind => Unit_Specification);
--# end
--# expr-done 8065
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 8066 '<FieldAccess for children_env at ast.py:17677>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17677
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Children_Env (Node => Var_Expr);
--# expr-done 8066
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2780;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
--# expr-start 8067 '<Symbol Address at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Address); 
--# expr-done 8067
Env_Get_Result := AST_Envs.Get_First (Self => Result_Var, Key => Sym_1, Lookup_Kind => To_Lookup_Kind_Type (Flat), Categories => (Inherited_Primitives => True, Nocat => True)); 
--# expr-done 8068
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 8069
Block_Var := Cast_Result; 
--# bind address_type Block_Var
--# expr-start 8081 '<And at ast.py:17685>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17685
--# expr-start 8078 '<Or at ast.py:17685>' Or_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17685
--# expr-start 8075 '<And at ast.py:17686>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17686
--# expr-start 8071 '<FieldAccess for xref_no_overloading at ast.py:17686>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17686
--# expr-start 8070 '<FieldAccess for prefix at ast.py:17686>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17686
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8070
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_3 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Fld_2.Node, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 8071
--# expr-start 8074 '<Predicate on BasicDecl.is_subprogram at ast.py:17687>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17687
--# expr-start 8073 '<FieldAccess for ref_var at ast.py:17687>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17687
--# expr-start 8072 '<FieldAccess for prefix at ast.py:17687>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17687
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Self.Attribute_Ref_F_Prefix;
--# expr-done 8072
      if Fld_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_5 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_4);
--# end
--# expr-done 8073
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
Pred := Solver.Create_Predicate (Fld_5, Create_Basic_Decl_P_Is_Subprogram_0_Predicate); 
--# expr-done 8074
And_Pred := Create_And (Fld_3, Pred, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17686") else null)); 
--# expr-done 8075
--# expr-start 8077 '<FieldAccess for sub_equation at ast.py:17689>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17689
--# expr-start 8076 '<FieldAccess for prefix at ast.py:17689>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17689
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8076
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_7 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_6.Node, Env => Env, Origin => Origin, E_Info => Fld_6.Info);
--# end
--# expr-done 8077
Or_Pred := Create_Or (And_Pred, Fld_7, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17685") else null)); 
--# expr-done 8078
--# expr-start 8080 '<Bind at ast.py:17690>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17690
--# expr-start 8079 '<FieldAccess for type_var at ast.py:17690>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17690
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8079
Fld_8.Value := No_Entity;
Entity_Vars.Reset (Fld_8);
Cast_Expr := Block_Var; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_8, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17690") else null)); 
--# expr-done 8080
And_Pred_1 := Create_And (Or_Pred, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17685") else null)); 
--# expr-done 8081
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 8082
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1257;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1257;
               Finalizer_Scope_2780;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Address_Equation;
--# end
--# property-start AttributeRef.identity_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17693
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Identity_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1258 with Inline_Always;
      Fld : Internal_Entity_Name;
Fld_1 : Logic_Equation;
Fld_2 : Bare_Name;
Fld_3 : Logic_Var;
Fld_4 : Logic_Var;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1258 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1258;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8090 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8089 '<And at ast.py:17696>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17696
--# expr-start 8084 '<FieldAccess for sub_equation at ast.py:17697>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17697
--# expr-start 8083 '<FieldAccess for prefix at ast.py:17697>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17697
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8083
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 8084
--# expr-start 8088 '<Bind at ast.py:17698>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17698
--# expr-start 8086 '<FieldAccess for ref_var at ast.py:17698>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17698
--# expr-start 8085 '<FieldAccess for prefix at ast.py:17698>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17698
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Attribute_Ref_F_Prefix;
--# expr-done 8085
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_2);
--# end
--# expr-done 8086
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
--# expr-start 8087 '<FieldAccess for type_var at ast.py:17698>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17698
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8087
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
Bind_Result := Solver.Create_Propagate (Fld_3, Fld_4, Logic_Functor_204'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Origin), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17698") else null)); 
--# expr-done 8088
And_Pred := Create_And (Fld_1, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17696") else null)); 
--# expr-done 8089
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 8090
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1258;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1258;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Identity_Equation;
--# end
--# property-start AttributeRef.universal_real_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17703
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Universal_Real_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1259 with Inline_Always;
      Fld : Logic_Var;
Fld_1 : Logic_Equation;
Fld_2 : Internal_Entity_Name;
Fld_3 : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1259 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Fld_3);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1259;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8096 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8095 '<And at ast.py:17705>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17705
--# expr-start 8092 '<FieldAccess for universal_real_bind at ast.py:17705>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17705
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8091 '<FieldAccess for type_var at ast.py:17705>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17705
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8091
--# property-call-start AdaNode.universal_real_bind
Fld_1 := Libadalang.Implementation.Ada_Node_P_Universal_Real_Bind (Node => Self, Type_Var => Fld, Origin => Origin);
--# end
--# expr-done 8092
--# expr-start 8094 '<FieldAccess for sub_equation at ast.py:17706>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17706
--# expr-start 8093 '<FieldAccess for prefix at ast.py:17706>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17706
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8093
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_2.Node, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 8094
And_Pred := Create_And (Fld_1, Fld_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17705") else null)); 
--# expr-done 8095
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 8096
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1259;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1259;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Universal_Real_Equation;
--# end
--# property-start AttributeRef.universal_int_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17710
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Universal_Int_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1260 with Inline_Always;
            procedure Finalizer_Scope_2781 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Name;
Fld_3 : Logic_Equation;
Fld_4 : Logic_Var;
Fld_5 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_6 : Internal_Entity_Assoc_List;
Fld_7 : Bare_Assoc_List;
Fld_8 : Internal_Entity_Info;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_9 : Internal_Entity_Expr;
Fld_10 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_11 : Internal_Entity_Expr;
Fld_12 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_13 : Internal_Entity_Expr;
Fld_14 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Item_251 : Internal_Entity_Basic_Assoc;
Bare_Item_251 : Bare_Basic_Assoc;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
And_Pred_3 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1260 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Fld_5);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1260;
            procedure Finalizer_Scope_2781 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_12);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_14);
                     Dec_Ref (And_Pred_2);
            end Finalizer_Scope_2781;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8117 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8098 '<FieldAccess for name_designated_type at ast.py:17711>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17711
--# expr-start 8097 '<FieldAccess for prefix at ast.py:17711>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17711
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8097
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_1 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 8098
Block_Var := Fld_1; 
--# bind typ Block_Var
--# expr-start 8116 '<And at ast.py:17714>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17714
--# expr-start 8103 '<And at ast.py:17714>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17714
--# expr-start 8100 '<FieldAccess for sub_equation at ast.py:17714>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17714
--# expr-start 8099 '<FieldAccess for prefix at ast.py:17714>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17714
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8099
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_2.Node, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 8100
--# expr-start 8102 '<FieldAccess for universal_int_bind at ast.py:17715>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17715
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8101 '<FieldAccess for type_var at ast.py:17715>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17715
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8101
--# property-call-start AdaNode.universal_int_bind
Fld_5 := Libadalang.Implementation.Ada_Node_P_Universal_Int_Bind (Node => Self, Type_Var => Fld_4, Origin => Origin);
--# end
--# expr-done 8102
And_Pred := Create_And (Fld_3, Fld_5, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17714") else null)); 
--# expr-done 8103
--# expr-start 8115 '<LogicAll at ast.py:17716>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17716
--# expr-start 8114 '<Map at ast.py:17716>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17716
   --# expr-start 8104 '<FieldAccess for args_list at ast.py:17716>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17716
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_6 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8104
Fld_7 := Fld_6.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_7 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Assoc_List := Fld_7;
      begin
         for Untyped_Item_251 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_251 := Untyped_Item_251; 
Is_Null := Bare_Item_251 = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_8 := Fld_6.Info;
As_Entity := (Info => Fld_8, Node => Bare_Item_251); 
   If_Result := As_Entity;
end if;
                  Item_251 := If_Result; 
   --# scope-start
               --# bind arg Item_251
      --# expr-start 8113 '<And at ast.py:17718>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17718
--# expr-start 8110 '<And at ast.py:17718>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17718
--# expr-start 8107 '<Bind at ast.py:17718>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17718
--# expr-start 8106 '<FieldAccess for expected_type_var at ast.py:17718>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17718
--# expr-start 8105 '<FieldAccess for expr at ast.py:17718>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17718
      if Item_251.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_9 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Item_251.Node, E_Info => Item_251.Info);
--# end
--# expr-done 8105
      if Fld_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_9.Node);
--# expr-done 8106
Fld_10.Value := No_Entity;
Entity_Vars.Reset (Fld_10);
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_10, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17718") else null)); 
--# expr-done 8107
--# expr-start 8109 '<FieldAccess for sub_equation at ast.py:17719>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17719
--# expr-start 8108 '<FieldAccess for expr at ast.py:17719>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17719
      if Item_251.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_11 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Item_251.Node, E_Info => Item_251.Info);
--# end
--# expr-done 8108
      if Fld_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_12 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_11.Node, Env => Env, Origin => Origin, E_Info => Fld_11.Info);
--# end
--# expr-done 8109
And_Pred_1 := Create_And (Bind_Result, Fld_12, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17718") else null)); 
--# expr-done 8110
--# expr-start 8112 '<FieldAccess for matches_expected_type at ast.py:17720>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17720
--# expr-start 8111 '<FieldAccess for expr at ast.py:17720>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17720
      if Item_251.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_13 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Item_251.Node, E_Info => Item_251.Info);
--# end
--# expr-done 8111
      if Fld_13.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_14 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_13.Node, Origin => Origin);
--# end
--# expr-done 8112
And_Pred_2 := Create_And (And_Pred_1, Fld_14, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17718") else null)); 
--# expr-done 8113
         declare
            Item_To_Append : constant Logic_Equation := And_Pred_2;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2781;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 8114
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17716") else null)); 
--# expr-done 8115
And_Pred_3 := Create_And (And_Pred, Logic_Boolean_Op, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17714") else null)); 
--# expr-done 8116
Let_Result := And_Pred_3; Inc_Ref (Let_Result);
--# end
--# expr-done 8117
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1260;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1260;
               Finalizer_Scope_2781;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Universal_Int_Equation;
--# end
--# property-start AttributeRef.succpred_xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17725
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Succpred_Xref_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1261 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Block_Var_1 : Internal_Entity_Expr;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Assoc_List;
Fld_3 : Bare_Assoc_List;
Fld_4 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_5 : Internal_Entity_Expr;
Fld_6 : Bare_Name;
Fld_7 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_8 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_9 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_10 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_11 : Logic_Var;
Cast_Expr_2 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result_2 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1261 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_9);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_10);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1261;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8135 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8119 '<FieldAccess for name_designated_type at ast.py:17726>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17726
--# expr-start 8118 '<FieldAccess for prefix at ast.py:17726>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17726
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8118
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_1 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 8119
Block_Var := Fld_1; 
--# bind typ Block_Var
--# expr-start 8121 '<FieldAccess for expr at ast.py:17727>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17727
--# expr-start 8120 '<FieldAccess for args_list at ast.py:17727>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17727
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_2 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8120
Fld_3 := Fld_2.Node;
Get_Result := Get (Fld_3, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_4 := Fld_2.Info;
As_Entity := (Info => Fld_4, Node => Get_Result); 
   If_Result := As_Entity;
end if;
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_5 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 8121
Block_Var_1 := Fld_5; 
--# bind arg Block_Var_1
--# expr-start 8134 '<And at ast.py:17730>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17730
--# expr-start 8131 '<And at ast.py:17730>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17730
--# expr-start 8129 '<And at ast.py:17730>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17730
--# expr-start 8127 '<And at ast.py:17730>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17730
--# expr-start 8124 '<Bind at ast.py:17730>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17730
--# expr-start 8123 '<FieldAccess for ref_var at ast.py:17730>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17730
--# expr-start 8122 '<FieldAccess for prefix at ast.py:17730>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17730
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Self.Attribute_Ref_F_Prefix;
--# expr-done 8122
      if Fld_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_7 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_6);
--# end
--# expr-done 8123
Fld_7.Value := No_Entity;
Entity_Vars.Reset (Fld_7);
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_7, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17730") else null)); 
--# expr-done 8124
--# expr-start 8126 '<Bind at ast.py:17731>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17731
--# expr-start 8125 '<FieldAccess for expected_type_var at ast.py:17731>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17731
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Block_Var_1.Node);
--# expr-done 8125
Fld_8.Value := No_Entity;
Entity_Vars.Reset (Fld_8);
Cast_Expr_1 := Block_Var; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_8, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17731") else null)); 
--# expr-done 8126
And_Pred := Create_And (Bind_Result, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17730") else null)); 
--# expr-done 8127
--# expr-start 8128 '<FieldAccess for sub_equation at ast.py:17732>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17732
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_9 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Block_Var_1.Node, Env => Env, Origin => Origin, E_Info => Block_Var_1.Info);
--# end
--# expr-done 8128
And_Pred_1 := Create_And (And_Pred, Fld_9, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17730") else null)); 
--# expr-done 8129
--# expr-start 8130 '<FieldAccess for matches_expected_type at ast.py:17733>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17733
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_10 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Block_Var_1.Node, Origin => Origin);
--# end
--# expr-done 8130
And_Pred_2 := Create_And (And_Pred_1, Fld_10, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17730") else null)); 
--# expr-done 8131
--# expr-start 8133 '<Bind at ast.py:17734>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17734
--# expr-start 8132 '<FieldAccess for type_var at ast.py:17734>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17734
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8132
Fld_11.Value := No_Entity;
Entity_Vars.Reset (Fld_11);
Cast_Expr_2 := Block_Var; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
Bind_Result_2 := Solver.Create_Assign (Fld_11, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17734") else null)); 
--# expr-done 8133
And_Pred_3 := Create_And (And_Pred_2, Bind_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17730") else null)); 
--# expr-done 8134
Let_Result := And_Pred_3; Inc_Ref (Let_Result);
--# end
--# expr-done 8135
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1261;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1261;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Succpred_Xref_Equation;
--# end
--# property-start AttributeRef.minmax_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17738
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Minmax_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1262 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Block_Var_1 : Internal_Entity_Expr;
Block_Var_2 : Internal_Entity_Expr;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Assoc_List;
Fld_3 : Bare_Assoc_List;
Fld_4 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_5 : Internal_Entity_Expr;
Fld_6 : Internal_Entity_Assoc_List;
Fld_7 : Bare_Assoc_List;
Fld_8 : Internal_Entity_Info;
Get_Result_1 : Bare_Basic_Assoc;
As_Entity_1 : Internal_Entity_Basic_Assoc;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Basic_Assoc;
Fld_9 : Internal_Entity_Expr;
Fld_10 : Internal_Entity_Name;
Fld_11 : Logic_Equation;
Fld_12 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_13 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_14 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_15 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_16 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Fld_17 : Logic_Equation;
And_Pred_5 : Logic_Equation;
Fld_18 : Logic_Var;
Cast_Expr_2 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result_2 : Logic_Equation;
And_Pred_6 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1262 is
            begin
                     Dec_Ref (Fld_11);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_13);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_14);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Fld_16);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Fld_17);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (And_Pred_6);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1262;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8161 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8137 '<FieldAccess for name_designated_type at ast.py:17739>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17739
--# expr-start 8136 '<FieldAccess for prefix at ast.py:17739>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17739
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8136
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_1 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 8137
Block_Var := Fld_1; 
--# bind typ Block_Var
--# expr-start 8139 '<FieldAccess for expr at ast.py:17740>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17740
--# expr-start 8138 '<FieldAccess for args_list at ast.py:17740>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17740
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_2 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8138
Fld_3 := Fld_2.Node;
Get_Result := Get (Fld_3, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_4 := Fld_2.Info;
As_Entity := (Info => Fld_4, Node => Get_Result); 
   If_Result := As_Entity;
end if;
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_5 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 8139
Block_Var_1 := Fld_5; 
--# bind left Block_Var_1
--# expr-start 8141 '<FieldAccess for expr at ast.py:17741>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17741
--# expr-start 8140 '<FieldAccess for args_list at ast.py:17741>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17741
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_6 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8140
Fld_7 := Fld_6.Node;
Get_Result_1 := Get (Fld_7, 1, True); 
Is_Null_1 := Get_Result_1 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Basic_Assoc;
else
Fld_8 := Fld_6.Info;
As_Entity_1 := (Info => Fld_8, Node => Get_Result_1); 
   If_Result_1 := As_Entity_1;
end if;
      if If_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_9 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result_1.Node, E_Info => If_Result_1.Info);
--# end
--# expr-done 8141
Block_Var_2 := Fld_9; 
--# bind right Block_Var_2
--# expr-start 8160 '<And at ast.py:17744>' And_Pred_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17744
--# expr-start 8157 '<And at ast.py:17744>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17744
--# expr-start 8155 '<And at ast.py:17744>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17744
--# expr-start 8153 '<And at ast.py:17744>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17744
--# expr-start 8150 '<And at ast.py:17744>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17744
--# expr-start 8148 '<And at ast.py:17744>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17744
--# expr-start 8146 '<And at ast.py:17744>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17744
--# expr-start 8143 '<FieldAccess for sub_equation at ast.py:17744>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17744
--# expr-start 8142 '<FieldAccess for prefix at ast.py:17744>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17744
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8142
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_11 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_10.Node, Env => Env, Origin => Origin, E_Info => Fld_10.Info);
--# end
--# expr-done 8143
--# expr-start 8145 '<Bind at ast.py:17745>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17745
--# expr-start 8144 '<FieldAccess for expected_type_var at ast.py:17745>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17745
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Block_Var_1.Node);
--# expr-done 8144
Fld_12.Value := No_Entity;
Entity_Vars.Reset (Fld_12);
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_12, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17745") else null)); 
--# expr-done 8145
And_Pred := Create_And (Fld_11, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17744") else null)); 
--# expr-done 8146
--# expr-start 8147 '<FieldAccess for sub_equation at ast.py:17746>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17746
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_13 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Block_Var_1.Node, Env => Env, Origin => Origin, E_Info => Block_Var_1.Info);
--# end
--# expr-done 8147
And_Pred_1 := Create_And (And_Pred, Fld_13, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17744") else null)); 
--# expr-done 8148
--# expr-start 8149 '<FieldAccess for matches_expected_type at ast.py:17747>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17747
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_14 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Block_Var_1.Node, Origin => Origin);
--# end
--# expr-done 8149
And_Pred_2 := Create_And (And_Pred_1, Fld_14, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17744") else null)); 
--# expr-done 8150
--# expr-start 8152 '<Bind at ast.py:17749>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17749
--# expr-start 8151 '<FieldAccess for expected_type_var at ast.py:17749>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17749
      if Block_Var_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Block_Var_2.Node);
--# expr-done 8151
Fld_15.Value := No_Entity;
Entity_Vars.Reset (Fld_15);
Cast_Expr_1 := Block_Var; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_15, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17749") else null)); 
--# expr-done 8152
And_Pred_3 := Create_And (And_Pred_2, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17744") else null)); 
--# expr-done 8153
--# expr-start 8154 '<FieldAccess for sub_equation at ast.py:17750>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17750
      if Block_Var_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_16 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Block_Var_2.Node, Env => Env, Origin => Origin, E_Info => Block_Var_2.Info);
--# end
--# expr-done 8154
And_Pred_4 := Create_And (And_Pred_3, Fld_16, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17744") else null)); 
--# expr-done 8155
--# expr-start 8156 '<FieldAccess for matches_expected_type at ast.py:17751>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17751
      if Block_Var_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_17 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Block_Var_2.Node, Origin => Origin);
--# end
--# expr-done 8156
And_Pred_5 := Create_And (And_Pred_4, Fld_17, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17744") else null)); 
--# expr-done 8157
--# expr-start 8159 '<Bind at ast.py:17753>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17753
--# expr-start 8158 '<FieldAccess for type_var at ast.py:17753>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17753
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8158
Fld_18.Value := No_Entity;
Entity_Vars.Reset (Fld_18);
Cast_Expr_2 := Block_Var; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
Bind_Result_2 := Solver.Create_Assign (Fld_18, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17753") else null)); 
--# expr-done 8159
And_Pred_6 := Create_And (And_Pred_5, Bind_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17744") else null)); 
--# expr-done 8160
Let_Result := And_Pred_6; Inc_Ref (Let_Result);
--# end
--# expr-done 8161
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1262;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1262;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Minmax_Equation;
--# end
--# property-start AttributeRef.value_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17757
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Value_Equation
  (Node : Bare_Attribute_Ref
      ; Str_Type : Internal_Entity
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind str_type Str_Type
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1263 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Block_Var_1 : Internal_Entity_Expr;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Assoc_List;
Fld_3 : Bare_Assoc_List;
Fld_4 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_5 : Internal_Entity_Expr;
Fld_6 : Internal_Entity_Name;
Fld_7 : Logic_Equation;
Fld_8 : Logic_Var;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_9 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_10 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_11 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1263 is
            begin
                     Dec_Ref (Fld_7);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_9);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_10);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1263;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8178 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8163 '<FieldAccess for name_designated_type at ast.py:17758>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17758
--# expr-start 8162 '<FieldAccess for prefix at ast.py:17758>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17758
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8162
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_1 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 8163
Block_Var := Fld_1; 
--# bind typ Block_Var
--# expr-start 8165 '<FieldAccess for expr at ast.py:17759>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17759
--# expr-start 8164 '<FieldAccess for args_list at ast.py:17759>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17759
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_2 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8164
Fld_3 := Fld_2.Node;
Get_Result := Get (Fld_3, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_4 := Fld_2.Info;
As_Entity := (Info => Fld_4, Node => Get_Result); 
   If_Result := As_Entity;
end if;
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_5 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 8165
Block_Var_1 := Fld_5; 
--# bind expr Block_Var_1
--# expr-start 8177 '<And at ast.py:17763>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17763
--# expr-start 8174 '<And at ast.py:17763>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17763
--# expr-start 8172 '<And at ast.py:17763>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17763
--# expr-start 8170 '<And at ast.py:17763>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17763
--# expr-start 8167 '<FieldAccess for sub_equation at ast.py:17763>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17763
--# expr-start 8166 '<FieldAccess for prefix at ast.py:17763>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17763
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8166
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_7 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_6.Node, Env => Env, Origin => Origin, E_Info => Fld_6.Info);
--# end
--# expr-done 8167
--# expr-start 8169 '<Bind at ast.py:17766>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17766
--# expr-start 8168 '<FieldAccess for expected_type_var at ast.py:17766>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17766
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Block_Var_1.Node);
--# expr-done 8168
Fld_8.Value := No_Entity;
Entity_Vars.Reset (Fld_8);
Bind_Result := Solver.Create_Assign (Fld_8, Str_Type, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17766") else null)); 
--# expr-done 8169
And_Pred := Create_And (Fld_7, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17763") else null)); 
--# expr-done 8170
--# expr-start 8171 '<FieldAccess for sub_equation at ast.py:17767>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17767
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_9 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Block_Var_1.Node, Env => Env, Origin => Origin, E_Info => Block_Var_1.Info);
--# end
--# expr-done 8171
And_Pred_1 := Create_And (And_Pred, Fld_9, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17763") else null)); 
--# expr-done 8172
--# expr-start 8173 '<FieldAccess for matches_expected_type at ast.py:17768>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17768
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_10 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Block_Var_1.Node, Origin => Origin);
--# end
--# expr-done 8173
And_Pred_2 := Create_And (And_Pred_1, Fld_10, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17763") else null)); 
--# expr-done 8174
--# expr-start 8176 '<Bind at ast.py:17771>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17771
--# expr-start 8175 '<FieldAccess for type_var at ast.py:17771>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17771
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8175
Fld_11.Value := No_Entity;
Entity_Vars.Reset (Fld_11);
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_11, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17771") else null)); 
--# expr-done 8176
And_Pred_3 := Create_And (And_Pred_2, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17763") else null)); 
--# expr-done 8177
Let_Result := And_Pred_3; Inc_Ref (Let_Result);
--# end
--# expr-done 8178
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1263;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1263;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Value_Equation;
--# end
--# property-start AttributeRef.fixed_value_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17775
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Fixed_Value_Equation
  (Node : Bare_Attribute_Ref
      ; Universal_Type : Internal_Entity
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind universal_type Universal_Type
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1264 with Inline_Always;
            procedure Finalizer_Scope_2782 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Var_Expr : Internal_Entity_Expr;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Name;
Fld_3 : Logic_Equation;
Fld_4 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_5 : Internal_Entity_Assoc_List;
Fld_6 : Bare_Assoc_List;
Fld_7 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_8 : Internal_Entity_Expr;
Fld_9 : Logic_Var;
Bind_Result_1 : Logic_Equation;
Fld_10 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_11 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Scope_Result : Logic_Equation;
False_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred_3 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1264 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (False_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1264;
            procedure Finalizer_Scope_2782 is
            begin
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Fld_10);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_11);
                     Dec_Ref (And_Pred_2);
            end Finalizer_Scope_2782;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8195 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8180 '<FieldAccess for name_designated_type at ast.py:17781>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17781
--# expr-start 8179 '<FieldAccess for prefix at ast.py:17781>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17781
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8179
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_1 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 8180
Block_Var := Fld_1; 
--# bind typ Block_Var
--# expr-start 8194 '<And at ast.py:17782>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17782
--# expr-start 8185 '<And at ast.py:17782>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17782
--# expr-start 8182 '<FieldAccess for sub_equation at ast.py:17783>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17783
--# expr-start 8181 '<FieldAccess for prefix at ast.py:17783>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17783
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8181
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_2.Node, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 8182
--# expr-start 8184 '<Bind at ast.py:17784>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17784
--# expr-start 8183 '<FieldAccess for type_var at ast.py:17784>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17784
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8183
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_4, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17784") else null)); 
--# expr-done 8184
And_Pred := Create_And (Fld_3, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17782") else null)); 
--# expr-done 8185
--# expr-start 8187 '<FieldAccess for expr at ast.py:17785>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17785
--# expr-start 8186 '<FieldAccess for args_list at ast.py:17785>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17785
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_5 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8186
Fld_6 := Fld_5.Node;
Get_Result := Get (Fld_6, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_7 := Fld_5.Info;
As_Entity := (Info => Fld_7, Node => Get_Result); 
   If_Result := As_Entity;
end if;
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_8 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 8187
Var_Expr := Fld_8;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind e Var_Expr
--# expr-start 8193 '<And at ast.py:17786>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17786
--# expr-start 8191 '<And at ast.py:17786>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17786
--# expr-start 8189 '<Bind at ast.py:17787>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17787
--# expr-start 8188 '<FieldAccess for expected_type_var at ast.py:17787>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17787
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Var_Expr.Node);
--# expr-done 8188
Fld_9.Value := No_Entity;
Entity_Vars.Reset (Fld_9);
Bind_Result_1 := Solver.Create_Assign (Fld_9, Universal_Type, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17787") else null)); 
--# expr-done 8189
--# expr-start 8190 '<FieldAccess for sub_equation at ast.py:17788>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17788
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_10 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 8190
And_Pred_1 := Create_And (Bind_Result_1, Fld_10, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17786") else null)); 
--# expr-done 8191
--# expr-start 8192 '<FieldAccess for matches_expected_type at ast.py:17789>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17789
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_11 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Var_Expr.Node, Origin => Origin);
--# end
--# expr-done 8192
And_Pred_2 := Create_And (And_Pred_1, Fld_11, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17786") else null)); 
--# expr-done 8193
Scope_Result := And_Pred_2;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2782;
   Result_Var := Scope_Result;
else
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17791") else null)); 
   Result_Var := False_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred_3 := Create_And (And_Pred, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17782") else null)); 
--# expr-done 8194
Let_Result := And_Pred_3; Inc_Ref (Let_Result);
--# end
--# expr-done 8195
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1264;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1264;
               Finalizer_Scope_2782;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Fixed_Value_Equation;
--# end
--# property-start AttributeRef.image_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17796
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Image_Equation
  (Node : Bare_Attribute_Ref
      ; Str_Type : Internal_Entity
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind str_type Str_Type
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1265 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Block_Var_1 : Internal_Entity_Expr;
Var_Expr : Internal_Entity_Assoc_List;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Assoc_List;
Saved : Internal_Entity_Assoc_List;
Fld_3 : Bare_Assoc_List;
Fld_4 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_5 : Internal_Entity_Expr;
Scope_Result : Internal_Entity_Expr;
Result_Var : Internal_Entity_Expr;
Fld_6 : Internal_Entity_Name;
Fld_7 : Logic_Equation;
Fld_8 : Logic_Var;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_9 : Logic_Equation;
Fld_10 : Bare_Name;
Fld_11 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_12 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_2 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_13 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_14 : Logic_Var;
Bind_Result_3 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Is_Null_1 : Boolean;
If_Result_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1265 is
            begin
                     Dec_Ref (Fld_7);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_9);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Fld_13);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Bind_Result_3);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1265;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8220 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8197 '<FieldAccess for name_designated_type at ast.py:17797>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17797
--# expr-start 8196 '<FieldAccess for prefix at ast.py:17797>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17797
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8196
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_1 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 8197
Block_Var := Fld_1; 
--# bind typ Block_Var
--# expr-start 8198 '<FieldAccess for args_list at ast.py:17798>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17798
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_2 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8198
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Assoc_List then
   --# scope-start
   --# bind al Var_Expr
--# expr-start 8199 '<FieldAccess for expr at ast.py:17798>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17798
Saved := Var_Expr; 
Fld_3 := Saved.Node;
Get_Result := Get (Fld_3, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_4 := Saved.Info;
As_Entity := (Info => Fld_4, Node => Get_Result); 
   If_Result := As_Entity;
end if;
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_5 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 8199
Scope_Result := Fld_5;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Expr;
end if;
Block_Var_1 := Result_Var; 
--# bind expr Block_Var_1
--# expr-start 8219 '<If at ast.py:17800>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17800
--# expr-start 8200 '<IsNull at ast.py:17801>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17801
Is_Null_1 := Block_Var.Node = null; 
--# expr-done 8200
if Is_Null_1 then
   --# expr-start 8205 '<And at ast.py:17804>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17804
--# expr-start 8202 '<FieldAccess for sub_equation at ast.py:17804>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17804
--# expr-start 8201 '<FieldAccess for prefix at ast.py:17804>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17804
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8201
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_7 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_6.Node, Env => Env, Origin => Origin, E_Info => Fld_6.Info);
--# end
--# expr-done 8202
--# expr-start 8204 '<Bind at ast.py:17805>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17805
--# expr-start 8203 '<FieldAccess for type_var at ast.py:17805>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17805
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8203
Fld_8.Value := No_Entity;
Entity_Vars.Reset (Fld_8);
Bind_Result := Solver.Create_Assign (Fld_8, Str_Type, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17805") else null)); 
--# expr-done 8204
And_Pred := Create_And (Fld_7, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17804") else null)); 
--# expr-done 8205
   If_Result_1 := And_Pred;
else
   --# expr-start 8218 '<And at ast.py:17807>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17807
--# expr-start 8215 '<And at ast.py:17807>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17807
--# expr-start 8213 '<And at ast.py:17807>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17807
--# expr-start 8210 '<And at ast.py:17807>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17807
--# expr-start 8206 '<FieldAccess for sub_equation at ast.py:17807>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17807
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_9 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Block_Var_1.Node, Env => Env, Origin => Origin, E_Info => Block_Var_1.Info);
--# end
--# expr-done 8206
--# expr-start 8209 '<Bind at ast.py:17809>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17809
--# expr-start 8208 '<FieldAccess for ref_var at ast.py:17809>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17809
--# expr-start 8207 '<FieldAccess for prefix at ast.py:17809>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17809
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Self.Attribute_Ref_F_Prefix;
--# expr-done 8207
      if Fld_10 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_11 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_10);
--# end
--# expr-done 8208
Fld_11.Value := No_Entity;
Entity_Vars.Reset (Fld_11);
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_11, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17809") else null)); 
--# expr-done 8209
And_Pred_1 := Create_And (Fld_9, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17807") else null)); 
--# expr-done 8210
--# expr-start 8212 '<Bind at ast.py:17811>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17811
--# expr-start 8211 '<FieldAccess for expected_type_var at ast.py:17811>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17811
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Block_Var_1.Node);
--# expr-done 8211
Fld_12.Value := No_Entity;
Entity_Vars.Reset (Fld_12);
Cast_Expr_1 := Block_Var; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_2 := Solver.Create_Assign (Fld_12, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17811") else null)); 
--# expr-done 8212
And_Pred_2 := Create_And (And_Pred_1, Bind_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17807") else null)); 
--# expr-done 8213
--# expr-start 8214 '<FieldAccess for matches_expected_type at ast.py:17812>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17812
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_13 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Block_Var_1.Node, Origin => Origin);
--# end
--# expr-done 8214
And_Pred_3 := Create_And (And_Pred_2, Fld_13, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17807") else null)); 
--# expr-done 8215
--# expr-start 8217 '<Bind at ast.py:17814>' Bind_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17814
--# expr-start 8216 '<FieldAccess for type_var at ast.py:17814>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17814
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8216
Fld_14.Value := No_Entity;
Entity_Vars.Reset (Fld_14);
Bind_Result_3 := Solver.Create_Assign (Fld_14, Str_Type, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17814") else null)); 
--# expr-done 8217
And_Pred_4 := Create_And (And_Pred_3, Bind_Result_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17807") else null)); 
--# expr-done 8218
   If_Result_1 := And_Pred_4;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 8219
Let_Result := If_Result_1; Inc_Ref (Let_Result);
--# end
--# expr-done 8220
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1265;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1265;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Image_Equation;
--# end
--# property-start AttributeRef.img_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17818
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Img_Equation
  (Node : Bare_Attribute_Ref
      ; Str_Type : Internal_Entity
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind str_type Str_Type
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1266 with Inline_Always;
      Fld : Internal_Entity_Name;
Fld_1 : Logic_Equation;
Fld_2 : Logic_Var;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1266 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1266;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8226 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8225 '<And at ast.py:17821>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17821
--# expr-start 8222 '<FieldAccess for xref_equation at ast.py:17821>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17821
--# expr-start 8221 '<FieldAccess for prefix at ast.py:17821>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17821
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8221
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_1 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 8222
--# expr-start 8224 '<Bind at ast.py:17824>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17824
--# expr-start 8223 '<FieldAccess for type_var at ast.py:17824>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17824
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8223
Fld_2.Value := No_Entity;
Entity_Vars.Reset (Fld_2);
Bind_Result := Solver.Create_Assign (Fld_2, Str_Type, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17824") else null)); 
--# expr-done 8224
And_Pred := Create_And (Fld_1, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17821") else null)); 
--# expr-done 8225
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 8226
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1266;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1266;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Img_Equation;
--# end
--# property-start AttributeRef.pos_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17828
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Pos_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1267 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Block_Var_1 : Internal_Entity_Expr;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Assoc_List;
Fld_3 : Bare_Assoc_List;
Fld_4 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_5 : Internal_Entity_Expr;
Fld_6 : Bare_Name;
Fld_7 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_8 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_9 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_10 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_11 : Logic_Var;
Fld_12 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1267 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_9);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_10);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Fld_12);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1267;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8244 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8228 '<FieldAccess for name_designated_type at ast.py:17829>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17829
--# expr-start 8227 '<FieldAccess for prefix at ast.py:17829>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17829
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8227
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_1 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 8228
Block_Var := Fld_1; 
--# bind typ Block_Var
--# expr-start 8230 '<FieldAccess for expr at ast.py:17830>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17830
--# expr-start 8229 '<FieldAccess for args_list at ast.py:17830>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17830
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_2 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8229
Fld_3 := Fld_2.Node;
Get_Result := Get (Fld_3, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_4 := Fld_2.Info;
As_Entity := (Info => Fld_4, Node => Get_Result); 
   If_Result := As_Entity;
end if;
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_5 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 8230
Block_Var_1 := Fld_5; 
--# bind expr Block_Var_1
--# expr-start 8243 '<And at ast.py:17834>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17834
--# expr-start 8240 '<And at ast.py:17834>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17834
--# expr-start 8238 '<And at ast.py:17834>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17834
--# expr-start 8236 '<And at ast.py:17834>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17834
--# expr-start 8233 '<Bind at ast.py:17834>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17834
--# expr-start 8232 '<FieldAccess for ref_var at ast.py:17834>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17834
--# expr-start 8231 '<FieldAccess for prefix at ast.py:17834>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17834
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Self.Attribute_Ref_F_Prefix;
--# expr-done 8231
      if Fld_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_7 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_6);
--# end
--# expr-done 8232
Fld_7.Value := No_Entity;
Entity_Vars.Reset (Fld_7);
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_7, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17834") else null)); 
--# expr-done 8233
--# expr-start 8235 '<Bind at ast.py:17836>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17836
--# expr-start 8234 '<FieldAccess for expected_type_var at ast.py:17836>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17836
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Block_Var_1.Node);
--# expr-done 8234
Fld_8.Value := No_Entity;
Entity_Vars.Reset (Fld_8);
Cast_Expr_1 := Block_Var; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_8, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17836") else null)); 
--# expr-done 8235
And_Pred := Create_And (Bind_Result, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17834") else null)); 
--# expr-done 8236
--# expr-start 8237 '<FieldAccess for sub_equation at ast.py:17837>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17837
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_9 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Block_Var_1.Node, Env => Env, Origin => Origin, E_Info => Block_Var_1.Info);
--# end
--# expr-done 8237
And_Pred_1 := Create_And (And_Pred, Fld_9, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17834") else null)); 
--# expr-done 8238
--# expr-start 8239 '<FieldAccess for matches_expected_type at ast.py:17838>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17838
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_10 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Block_Var_1.Node, Origin => Origin);
--# end
--# expr-done 8239
And_Pred_2 := Create_And (And_Pred_1, Fld_10, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17834") else null)); 
--# expr-done 8240
--# expr-start 8242 '<FieldAccess for universal_int_bind at ast.py:17840>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17840
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8241 '<FieldAccess for type_var at ast.py:17840>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17840
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8241
--# property-call-start AdaNode.universal_int_bind
Fld_12 := Libadalang.Implementation.Ada_Node_P_Universal_Int_Bind (Node => Self, Type_Var => Fld_11, Origin => Origin);
--# end
--# expr-done 8242
And_Pred_3 := Create_And (And_Pred_2, Fld_12, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17834") else null)); 
--# expr-done 8243
Let_Result := And_Pred_3; Inc_Ref (Let_Result);
--# end
--# expr-done 8244
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1267;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1267;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Pos_Equation;
--# end
--# property-start AttributeRef.val_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17844
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Val_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1268 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Block_Var_1 : Internal_Entity_Expr;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Assoc_List;
Fld_3 : Bare_Assoc_List;
Fld_4 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_5 : Internal_Entity_Expr;
Fld_6 : Bare_Name;
Fld_7 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_8 : Logic_Var;
Fld_9 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_10 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_11 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_12 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1268 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_9);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_10);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_11);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1268;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8262 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8246 '<FieldAccess for name_designated_type at ast.py:17845>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17845
--# expr-start 8245 '<FieldAccess for prefix at ast.py:17845>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17845
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8245
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_1 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 8246
Block_Var := Fld_1; 
--# bind typ Block_Var
--# expr-start 8248 '<FieldAccess for expr at ast.py:17846>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17846
--# expr-start 8247 '<FieldAccess for args_list at ast.py:17846>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17846
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_2 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8247
Fld_3 := Fld_2.Node;
Get_Result := Get (Fld_3, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_4 := Fld_2.Info;
As_Entity := (Info => Fld_4, Node => Get_Result); 
   If_Result := As_Entity;
end if;
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_5 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 8248
Block_Var_1 := Fld_5; 
--# bind expr Block_Var_1
--# expr-start 8261 '<And at ast.py:17849>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17849
--# expr-start 8258 '<And at ast.py:17849>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17849
--# expr-start 8256 '<And at ast.py:17849>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17849
--# expr-start 8254 '<And at ast.py:17849>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17849
--# expr-start 8251 '<Bind at ast.py:17849>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17849
--# expr-start 8250 '<FieldAccess for ref_var at ast.py:17849>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17849
--# expr-start 8249 '<FieldAccess for prefix at ast.py:17849>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17849
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Self.Attribute_Ref_F_Prefix;
--# expr-done 8249
      if Fld_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_7 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_6);
--# end
--# expr-done 8250
Fld_7.Value := No_Entity;
Entity_Vars.Reset (Fld_7);
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_7, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17849") else null)); 
--# expr-done 8251
--# expr-start 8253 '<FieldAccess for universal_int_bind at ast.py:17851>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17851
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8252 '<FieldAccess for expected_type_var at ast.py:17851>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17851
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Block_Var_1.Node);
--# expr-done 8252
--# property-call-start AdaNode.universal_int_bind
Fld_9 := Libadalang.Implementation.Ada_Node_P_Universal_Int_Bind (Node => Self, Type_Var => Fld_8, Origin => Origin);
--# end
--# expr-done 8253
And_Pred := Create_And (Bind_Result, Fld_9, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17849") else null)); 
--# expr-done 8254
--# expr-start 8255 '<FieldAccess for sub_equation at ast.py:17852>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17852
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_10 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Block_Var_1.Node, Env => Env, Origin => Origin, E_Info => Block_Var_1.Info);
--# end
--# expr-done 8255
And_Pred_1 := Create_And (And_Pred, Fld_10, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17849") else null)); 
--# expr-done 8256
--# expr-start 8257 '<FieldAccess for matches_expected_type at ast.py:17853>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17853
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_11 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Block_Var_1.Node, Origin => Origin);
--# end
--# expr-done 8257
And_Pred_2 := Create_And (And_Pred_1, Fld_11, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17849") else null)); 
--# expr-done 8258
--# expr-start 8260 '<Bind at ast.py:17855>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17855
--# expr-start 8259 '<FieldAccess for type_var at ast.py:17855>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17855
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8259
Fld_12.Value := No_Entity;
Entity_Vars.Reset (Fld_12);
Cast_Expr_1 := Block_Var; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_12, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17855") else null)); 
--# expr-done 8260
And_Pred_3 := Create_And (And_Pred_2, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17849") else null)); 
--# expr-done 8261
Let_Result := And_Pred_3; Inc_Ref (Let_Result);
--# end
--# expr-done 8262
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1268;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1268;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Val_Equation;
--# end
--# property-start AttributeRef.invalid_value_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17859
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Invalid_Value_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1269 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Bare_Name;
Fld_3 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_4 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1269 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1269;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8271 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8264 '<FieldAccess for name_designated_type at ast.py:17860>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17860
--# expr-start 8263 '<FieldAccess for prefix at ast.py:17860>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17860
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8263
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_1 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 8264
Block_Var := Fld_1; 
--# bind typ Block_Var
--# expr-start 8270 '<And at ast.py:17861>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17861
--# expr-start 8267 '<Bind at ast.py:17862>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17862
--# expr-start 8266 '<FieldAccess for ref_var at ast.py:17862>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17862
--# expr-start 8265 '<FieldAccess for prefix at ast.py:17862>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17862
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Attribute_Ref_F_Prefix;
--# expr-done 8265
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_2);
--# end
--# expr-done 8266
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_3, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17862") else null)); 
--# expr-done 8267
--# expr-start 8269 '<Bind at ast.py:17863>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17863
--# expr-start 8268 '<FieldAccess for type_var at ast.py:17863>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17863
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8268
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
Cast_Expr_1 := Block_Var; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_4, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17863") else null)); 
--# expr-done 8269
And_Pred := Create_And (Bind_Result, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17861") else null)); 
--# expr-done 8270
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 8271
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1269;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1269;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Invalid_Value_Equation;
--# end
--# property-start AttributeRef.access_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17867
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Access_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1270 with Inline_Always;
      Fld : Internal_Entity_Name;
Fld_1 : Logic_Equation;
Fld_2 : Logic_Var;
Fld_3 : Bare_Name;
Fld_4 : Logic_Var;
Pred : Logic_Equation;
And_Pred : Logic_Equation;
Fld_5 : Logic_Var;
Fld_6 : Logic_Var;
Bind_Result : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_7 : Internal_Entity_Name;
Fld_8 : Logic_Equation;
Fld_9 : Logic_Var;
Fld_10 : Logic_Var;
Bind_Result_1 : Logic_Equation;
Fld_11 : Logic_Var;
Fld_12 : Bare_Name;
Fld_13 : Logic_Var;
Bind_Result_2 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_14 : Bare_Name;
Fld_15 : Logic_Var;
Fld_16 : Bare_Name;
Fld_17 : Logic_Var;
Bind_Result_3 : Logic_Equation;
Fld_18 : Bare_Name;
Fld_19 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result_4 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_20 : Internal_Entity_Name;
Fld_21 : Logic_Equation;
Or_Pred : Logic_Equation;
And_Pred_4 : Logic_Equation;
Fld_22 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_5 : Logic_Equation;
Fld_23 : Bare_Name;
Fld_24 : Logic_Var;
Fld_25 : Logic_Var;
Bind_Result_6 : Logic_Equation;
And_Pred_5 : Logic_Equation;
Or_Pred_1 : Logic_Equation;
And_Pred_6 : Logic_Equation;
Or_Pred_2 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1270 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Pred);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_8);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Bind_Result_3);
                     Dec_Ref (Bind_Result_4);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Fld_21);
                     Dec_Ref (Or_Pred);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Bind_Result_5);
                     Dec_Ref (Bind_Result_6);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (Or_Pred_1);
                     Dec_Ref (And_Pred_6);
                     Dec_Ref (Or_Pred_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1270;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8316 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8315 '<Or at ast.py:17868>' Or_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17868
--# expr-start 8282 '<And at ast.py:17870>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17870
--# expr-start 8278 '<And at ast.py:17870>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17870
--# expr-start 8273 '<FieldAccess for xref_no_overloading at ast.py:17870>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17870
--# expr-start 8272 '<FieldAccess for prefix at ast.py:17870>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17870
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8272
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_1 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Fld.Node, All_Els => True, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 8273
--# expr-start 8277 '<Predicate on BaseTypeDecl.is_subp_access_of at ast.py:17871>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17871
--# expr-start 8274 '<FieldAccess for type_var at ast.py:17872>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17872
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8274
Fld_2.Value := No_Entity;
Entity_Vars.Reset (Fld_2);
--# expr-start 8276 '<FieldAccess for ref_var at ast.py:17873>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17873
--# expr-start 8275 '<FieldAccess for prefix at ast.py:17873>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17873
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Self.Attribute_Ref_F_Prefix;
--# expr-done 8275
      if Fld_3 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_4 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_3);
--# end
--# expr-done 8276
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
Pred := Solver.Create_N_Predicate ((Fld_2, Fld_4), Create_Base_Type_Decl_P_Is_Subp_Access_Of_0_Predicate (Origin)); 
--# expr-done 8277
And_Pred := Create_And (Fld_1, Pred, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17870") else null)); 
--# expr-done 8278
--# expr-start 8281 '<Bind at ast.py:17874>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17874
--# expr-start 8279 '<FieldAccess for type_var at ast.py:17874>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17874
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8279
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
--# expr-start 8280 '<FieldAccess for expected_type_var at ast.py:17874>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17874
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 8280
Fld_6.Value := No_Entity;
Entity_Vars.Reset (Fld_6);
Bind_Result := Solver.Create_Unify (Fld_5, Fld_6, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17874") else null)); 
--# expr-done 8281
And_Pred_1 := Create_And (And_Pred, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17870") else null)); 
--# expr-done 8282
--# expr-start 8314 '<And at ast.py:17877>' And_Pred_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17877
--# expr-start 8284 '<FieldAccess for sub_equation at ast.py:17877>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17877
--# expr-start 8283 '<FieldAccess for prefix at ast.py:17877>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17877
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8283
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_8 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_7.Node, Env => Env, Origin => Origin, E_Info => Fld_7.Info);
--# end
--# expr-done 8284
--# expr-start 8313 '<Or at ast.py:17878>' Or_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17878
--# expr-start 8305 '<And at ast.py:17888>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17888
--# expr-start 8292 '<And at ast.py:17888>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17888
--# expr-start 8287 '<Bind at ast.py:17888>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17888
--# expr-start 8285 '<FieldAccess for expected_type_var at ast.py:17888>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17888
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 8285
Fld_9.Value := No_Entity;
Entity_Vars.Reset (Fld_9);
--# expr-start 8286 '<FieldAccess for type_var at ast.py:17888>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17888
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8286
Fld_10.Value := No_Entity;
Entity_Vars.Reset (Fld_10);
Bind_Result_1 := Solver.Create_Unify (Fld_9, Fld_10, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17888") else null)); 
--# expr-done 8287
--# expr-start 8291 '<Bind at ast.py:17889>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17889
--# expr-start 8288 '<FieldAccess for expected_type_var at ast.py:17889>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17889
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 8288
Fld_11.Value := No_Entity;
Entity_Vars.Reset (Fld_11);
--# expr-start 8290 '<FieldAccess for expected_type_var at ast.py:17889>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17889
--# expr-start 8289 '<FieldAccess for prefix at ast.py:17889>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17889
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Self.Attribute_Ref_F_Prefix;
--# expr-done 8289
      if Fld_12 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_12);
--# expr-done 8290
Fld_13.Value := No_Entity;
Entity_Vars.Reset (Fld_13);
Bind_Result_2 := Solver.Create_Propagate (Fld_11, Fld_13, Logic_Functor_604'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Origin), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17889") else null)); 
--# expr-done 8291
And_Pred_2 := Create_And (Bind_Result_1, Bind_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17888") else null)); 
--# expr-done 8292
--# expr-start 8304 '<Or at ast.py:17891>' Or_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17891
--# expr-start 8301 '<And at ast.py:17896>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17896
--# expr-start 8297 '<Bind at ast.py:17896>' Bind_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17896
--# expr-start 8294 '<FieldAccess for expected_type_var at ast.py:17896>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17896
--# expr-start 8293 '<FieldAccess for prefix at ast.py:17896>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17896
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Self.Attribute_Ref_F_Prefix;
--# expr-done 8293
      if Fld_14 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_14);
--# expr-done 8294
Fld_15.Value := No_Entity;
Entity_Vars.Reset (Fld_15);
--# expr-start 8296 '<FieldAccess for type_var at ast.py:17896>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17896
--# expr-start 8295 '<FieldAccess for prefix at ast.py:17896>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17896
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_16 := Self.Attribute_Ref_F_Prefix;
--# expr-done 8295
      if Fld_16 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_17 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_16);
--# expr-done 8296
Fld_17.Value := No_Entity;
Entity_Vars.Reset (Fld_17);
Bind_Result_3 := Solver.Create_Unify (Fld_15, Fld_17, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17896") else null)); 
--# expr-done 8297
--# expr-start 8300 '<Bind at ast.py:17897>' Bind_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17897
--# expr-start 8299 '<FieldAccess for type_var at ast.py:17897>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17897
--# expr-start 8298 '<FieldAccess for prefix at ast.py:17897>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17897
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Self.Attribute_Ref_F_Prefix;
--# expr-done 8298
      if Fld_18 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_18);
--# expr-done 8299
Fld_19.Value := No_Entity;
Entity_Vars.Reset (Fld_19);
Cast_Expr := No_Entity_Base_Type_Decl; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result_4 := Solver.Create_Assign (Fld_19, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17897") else null)); 
--# expr-done 8300
And_Pred_3 := Create_And (Bind_Result_3, Bind_Result_4, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17896") else null)); 
--# expr-done 8301
--# expr-start 8303 '<FieldAccess for matches_expected_formal_type at ast.py:17901>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17901
--# expr-start 8302 '<FieldAccess for prefix at ast.py:17901>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17901
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_20 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8302
      if Fld_20.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_21 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Fld_20.Node, Origin => Origin);
--# end
--# expr-done 8303
Or_Pred := Create_Or (And_Pred_3, Fld_21, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17891") else null)); 
--# expr-done 8304
And_Pred_4 := Create_And (And_Pred_2, Or_Pred, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17888") else null)); 
--# expr-done 8305
--# expr-start 8312 '<And at ast.py:17906>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17906
--# expr-start 8307 '<Bind at ast.py:17906>' Bind_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17906
--# expr-start 8306 '<FieldAccess for expected_type_var at ast.py:17906>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17906
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_22 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 8306
Fld_22.Value := No_Entity;
Entity_Vars.Reset (Fld_22);
Cast_Expr_1 := No_Entity_Base_Type_Decl; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_5 := Solver.Create_Assign (Fld_22, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17906") else null)); 
--# expr-done 8307
--# expr-start 8311 '<Bind at ast.py:17907>' Bind_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17907
--# expr-start 8309 '<FieldAccess for type_var at ast.py:17907>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17907
--# expr-start 8308 '<FieldAccess for prefix at ast.py:17907>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17907
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_23 := Self.Attribute_Ref_F_Prefix;
--# expr-done 8308
      if Fld_23 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_24 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_23);
--# expr-done 8309
Fld_24.Value := No_Entity;
Entity_Vars.Reset (Fld_24);
--# expr-start 8310 '<FieldAccess for type_var at ast.py:17908>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17908
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_25 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8310
Fld_25.Value := No_Entity;
Entity_Vars.Reset (Fld_25);
Bind_Result_6 := Solver.Create_Propagate (Fld_24, Fld_25, Logic_Functor_554'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17907") else null)); 
--# expr-done 8311
And_Pred_5 := Create_And (Bind_Result_5, Bind_Result_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17906") else null)); 
--# expr-done 8312
Or_Pred_1 := Create_Or (And_Pred_4, And_Pred_5, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17878") else null)); 
--# expr-done 8313
And_Pred_6 := Create_And (Fld_8, Or_Pred_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17877") else null)); 
--# expr-done 8314
Or_Pred_2 := Create_Or (And_Pred_1, And_Pred_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17868") else null)); 
--# expr-done 8315
Let_Result := Or_Pred_2; Inc_Ref (Let_Result);
--# end
--# expr-done 8316
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1270;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1270;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Access_Equation;
--# end
--# property-start AttributeRef.size_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17914
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Size_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1271 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Bare_Name;
Fld_3 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_4 : Logic_Var;
Fld_5 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_6 : Internal_Entity_Name;
Fld_7 : Logic_Equation;
Fld_8 : Logic_Var;
Fld_9 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Is_Null : Boolean;
Not_Val : Boolean;
If_Result : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1271 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_5);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_7);
                     Dec_Ref (Fld_9);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1271;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8333 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8318 '<FieldAccess for name_designated_type at ast.py:17915>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17915
--# expr-start 8317 '<FieldAccess for prefix at ast.py:17915>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17915
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8317
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_1 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 8318
Block_Var := Fld_1; 
--# bind typ Block_Var
--# expr-start 8332 '<If at ast.py:17916>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17916
--# expr-start 8320 '<Not at ast.py:17917>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17917
--# expr-start 8319 '<IsNull at ast.py:17917>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17917
Is_Null := Block_Var.Node = null; 
--# expr-done 8319
Not_Val := not (Is_Null); 
--# expr-done 8320
if Not_Val then
   --# expr-start 8326 '<And at ast.py:17919>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17919
--# expr-start 8323 '<Bind at ast.py:17919>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17919
--# expr-start 8322 '<FieldAccess for ref_var at ast.py:17919>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17919
--# expr-start 8321 '<FieldAccess for prefix at ast.py:17919>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17919
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Attribute_Ref_F_Prefix;
--# expr-done 8321
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_2);
--# end
--# expr-done 8322
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_3, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17919") else null)); 
--# expr-done 8323
--# expr-start 8325 '<FieldAccess for universal_int_bind at ast.py:17920>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17920
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8324 '<FieldAccess for type_var at ast.py:17920>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17920
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8324
--# property-call-start AdaNode.universal_int_bind
Fld_5 := Libadalang.Implementation.Ada_Node_P_Universal_Int_Bind (Node => Self, Type_Var => Fld_4, Origin => Origin);
--# end
--# expr-done 8325
And_Pred := Create_And (Bind_Result, Fld_5, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17919") else null)); 
--# expr-done 8326
   If_Result := And_Pred;
else
   --# expr-start 8331 '<And at ast.py:17922>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17922
--# expr-start 8328 '<FieldAccess for sub_equation at ast.py:17922>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17922
--# expr-start 8327 '<FieldAccess for prefix at ast.py:17922>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17922
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8327
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_7 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_6.Node, Env => Env, Origin => Origin, E_Info => Fld_6.Info);
--# end
--# expr-done 8328
--# expr-start 8330 '<FieldAccess for universal_int_bind at ast.py:17923>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17923
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8329 '<FieldAccess for type_var at ast.py:17923>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17923
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8329
--# property-call-start AdaNode.universal_int_bind
Fld_9 := Libadalang.Implementation.Ada_Node_P_Universal_Int_Bind (Node => Self, Type_Var => Fld_8, Origin => Origin);
--# end
--# expr-done 8330
And_Pred_1 := Create_And (Fld_7, Fld_9, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17922") else null)); 
--# expr-done 8331
   If_Result := And_Pred_1;
end if;
      Inc_Ref (If_Result);
--# expr-done 8332
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 8333
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1271;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1271;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Size_Equation;
--# end
--# property-start AttributeRef.array_attr_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17927
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Array_Attr_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1272 with Inline_Always;
            procedure Finalizer_Scope_2785 with Inline_Always;
      Block_Var : Boolean;
Block_Var_1 : Internal_Entity_Base_Type_Decl;
Block_Var_2 : Integer;
Var_Expr : Internal_Entity_Assoc_List;
Var_Expr_1 : Internal_Entity_Expr;
Unused_2 : Boolean;
Res : Boolean;
Pfx_Typ : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Identifier;
Sym : Symbol_Type;
Fld_1 : Boolean;
Fld_2 : Internal_Entity_Name;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Internal_Entity_Assoc_List;
Saved : Internal_Entity_Assoc_List;
Fld_5 : Bare_Assoc_List;
Fld_6 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_7 : Internal_Entity_Expr;
Fld_8 : Boolean;
Fld_9 : Big_Integer_Type;
Small_Int : Integer;
Let_Result : Integer;
Scope_Result : Integer;
Result_Var : Integer;
Scope_Result_1 : Integer;
Result_Var_1 : Integer;
Arith_Result : Integer;
Fld_10 : Internal_Entity_Name;
Fld_11 : Logic_Equation;
Fld_12 : Boolean;
If_Result_1 : Boolean;
Fld_13 : Logic_Var;
Fld_14 : Logic_Equation;
Fld_15 : Boolean;
Fld_16 : Logic_Var;
Fld_17 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_18 : Boolean;
Fld_19 : Boolean;
Not_Val : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Boolean;
Fld_20 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
False_Rel : Logic_Equation;
If_Result_4 : Logic_Equation;
If_Result_5 : Logic_Equation;
If_Result_6 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_21 : Internal_Entity_Name;
Fld_22 : Internal_Entity_Name;
Fld_23 : Logic_Var;
Pred : Logic_Equation;
Fld_24 : Boolean;
Fld_25 : Internal_Entity_Name;
Fld_26 : Internal_Entity;
Cast_Result_2 : Internal_Entity_Base_Type_Decl;
Fld_27 : Logic_Var;
Fld_28 : Logic_Equation;
Fld_29 : Logic_Var;
Fld_30 : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity;
Bind_Result_2 : Logic_Equation;
If_Result_7 : Logic_Equation;
False_Rel_1 : Logic_Equation;
If_Result_8 : Logic_Equation;
Let_Result_1 : Logic_Equation;
Is_Null_1 : Boolean;
Not_Val_1 : Boolean;
If_Result_9 : Logic_Equation;
Let_Result_2 : Logic_Equation;
            procedure Finalizer_Scope_1272 is
            begin
                     Dec_Ref (Fld_11);
                     Dec_Ref (Fld_14);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (False_Rel);
                     Dec_Ref (If_Result_4);
                     Dec_Ref (If_Result_5);
                     Dec_Ref (If_Result_6);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Pred);
                     Dec_Ref (Fld_28);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (If_Result_7);
                     Dec_Ref (False_Rel_1);
                     Dec_Ref (If_Result_8);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (If_Result_9);
                     Dec_Ref (Let_Result_2);
            end Finalizer_Scope_1272;
            procedure Finalizer_Scope_2785 is
            begin
                     Dec_Ref (Fld_9);
            end Finalizer_Scope_2785;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8381 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 8336 '<FieldAccess for name_is at ast.py:17928>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17928
--# expr-start 8334 '<FieldAccess for attribute at ast.py:17928>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17928
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Identifier (Node => Ent.Node.Attribute_Ref_F_Attribute, Info => Ent.Info);
--# expr-done 8334
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8335 '<Symbol Length at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Length); 
--# expr-done 8335
--# property-call-start Name.name_is
Fld_1 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld.Node, Sym => Sym);
--# end
--# expr-done 8336
Block_Var := Fld_1; 
--# bind is_length Block_Var
--# expr-start 8338 '<FieldAccess for name_designated_type at ast.py:17929>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17929
--# expr-start 8337 '<FieldAccess for prefix at ast.py:17929>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17929
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8337
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_3 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld_2.Node, E_Info => Fld_2.Info);
--# end
--# expr-done 8338
Block_Var_1 := Fld_3; 
--# bind typ Block_Var_1
--# expr-start 8345 '<Arithmetic - at ast.py:17933>' Arith_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17933
--# expr-start 8339 '<FieldAccess for args_list at ast.py:17933>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17933
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_4 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8339
Var_Expr := Fld_4;
if Var_Expr /= No_Entity_Assoc_List then
   --# scope-start
   --# bind a Var_Expr
--# expr-start 8340 '<FieldAccess for expr at ast.py:17933>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17933
Saved := Var_Expr; 
Fld_5 := Saved.Node;
Get_Result := Get (Fld_5, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_6 := Saved.Info;
As_Entity := (Info => Fld_6, Node => Get_Result); 
   If_Result := As_Entity;
end if;
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_7 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 8340
Var_Expr_1 := Fld_7;
if Var_Expr_1 /= No_Entity_Expr then
   --# scope-start
   --# bind expr Var_Expr_1
--# expr-start 8344 '<Let at ast.py:17934>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17934
--# scope-start
--# expr-start 8341 '<FieldAccess for resolve_names_internal at ast.py:17935>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17935
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.resolve_names_internal
Fld_8 := Libadalang.Implementation.Ada_Node_P_Resolve_Names_Internal (Node => Var_Expr_1.Node, Env => Env, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 8341
Unused_2 := Fld_8; 
--# bind _ Unused_2
--# expr-start 8343 '<as_int at ast.py:17936>' Small_Int /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17936
--# expr-start 8342 '<FieldAccess for eval_as_int at ast.py:17936>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17936
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.eval_as_int
Fld_9 := Libadalang.Implementation.Expr_P_Eval_As_Int (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 8342
Small_Int := To_Integer (Fld_9); 
--# expr-done 8343
Let_Result := Small_Int; 
--# end
--# expr-done 8344
Scope_Result := Let_Result;
   --# end
      Finalizer_Scope_2785;
   Result_Var := Scope_Result;
else
   Result_Var := 1;
end if;
Scope_Result_1 := Result_Var;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := 1;
end if;
Arith_Result := (Result_Var_1 - 1); 
--# expr-done 8345
Block_Var_2 := Arith_Result; 
--# bind dim Block_Var_2
--# expr-start 8380 '<If at ast.py:17940>' If_Result_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17940
--# expr-start 8347 '<Not at ast.py:17941>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17941
--# expr-start 8346 '<IsNull at ast.py:17941>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17941
Is_Null_1 := Block_Var_1.Node = null; 
--# expr-done 8346
Not_Val_1 := not (Is_Null_1); 
--# expr-done 8347
if Not_Val_1 then
   --# expr-start 8363 '<And at ast.py:17944>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17944
--# expr-start 8349 '<FieldAccess for xref_no_overloading at ast.py:17944>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17944
--# expr-start 8348 '<FieldAccess for prefix at ast.py:17944>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17944
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8348
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_11 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Fld_10.Node, Env => Env, Origin => Origin, E_Info => Fld_10.Info);
--# end
--# expr-done 8349
--# expr-start 8362 '<Cond at ast.py:17944>' If_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17944
--# expr-start 8350 '<FieldAccess for is_array_def_with_deref at ast.py:17945>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17945
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_array_def_with_deref
Fld_12 := Libadalang.Implementation.Base_Type_Decl_P_Is_Array_Def_With_Deref (Node => Block_Var_1.Node, Origin => Origin, E_Info => Block_Var_1.Info);
--# end
--# expr-done 8350
if Fld_12 then
   If_Result_1 := Block_Var;
else
   If_Result_1 := False;
end if;
if If_Result_1 then
   --# expr-start 8352 '<FieldAccess for universal_int_bind at ast.py:17946>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17946
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8351 '<FieldAccess for type_var at ast.py:17946>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17946
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8351
--# property-call-start AdaNode.universal_int_bind
Fld_14 := Libadalang.Implementation.Ada_Node_P_Universal_Int_Bind (Node => Self, Type_Var => Fld_13, Origin => Origin);
--# end
--# expr-done 8352
   If_Result_6 := Fld_14;
else
--# expr-start 8353 '<FieldAccess for is_array_def_with_deref at ast.py:17949>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17949
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_array_def_with_deref
Fld_15 := Libadalang.Implementation.Base_Type_Decl_P_Is_Array_Def_With_Deref (Node => Block_Var_1.Node, Origin => Origin, E_Info => Block_Var_1.Info);
--# end
--# expr-done 8353
if Fld_15 then
   --# expr-start 8356 '<Bind at ast.py:17950>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17950
--# expr-start 8354 '<FieldAccess for type_var at ast.py:17950>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17950
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_16 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8354
Fld_16.Value := No_Entity;
Entity_Vars.Reset (Fld_16);
--# expr-start 8355 '<FieldAccess for index_type at ast.py:17950>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17950
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.index_type
Fld_17 := Libadalang.Implementation.Base_Type_Decl_P_Index_Type (Node => Block_Var_1.Node, Dim => Block_Var_2, Origin => Origin, E_Info => Block_Var_1.Info);
--# end
--# expr-done 8355
      Cast_Result := Create_Internal_Entity
        (Node => Fld_17.Node,
         Info => Fld_17.Info);
Bind_Result := Solver.Create_Assign (Fld_16, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17950") else null)); 
--# expr-done 8356
   If_Result_5 := Bind_Result;
else
--# expr-start 8357 '<FieldAccess for is_discrete_type at ast.py:17953>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17953
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_discrete_type'
Fld_18 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Discrete_Type (Node => Block_Var_1.Node, Origin => Origin, E_Info => Block_Var_1.Info);
--# end
--# expr-done 8357
if Fld_18 then
   If_Result_3 := True;
else
--# expr-start 8358 '<FieldAccess for is_real_type at ast.py:17953>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17953
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_real_type'
Fld_19 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Real_Type (Node => Block_Var_1.Node, Origin => Origin, E_Info => Block_Var_1.Info);
--# end
--# expr-done 8358
if Fld_19 then
   --# expr-start 8359 '<Not at ast.py:17953>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17953
Not_Val := not (Block_Var); 
--# expr-done 8359
   If_Result_2 := Not_Val;
else
   If_Result_2 := False;
end if;
   If_Result_3 := If_Result_2;
end if;
if If_Result_3 then
   --# expr-start 8361 '<Bind at ast.py:17954>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17954
--# expr-start 8360 '<FieldAccess for type_var at ast.py:17954>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17954
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_20 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8360
Fld_20.Value := No_Entity;
Entity_Vars.Reset (Fld_20);
Cast_Expr := Block_Var_1; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_20, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17954") else null)); 
--# expr-done 8361
   If_Result_4 := Bind_Result_1;
else
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17956") else null)); 
   If_Result_4 := False_Rel;
end if;
      Inc_Ref (If_Result_4);
   If_Result_5 := If_Result_4;
end if;
      Inc_Ref (If_Result_5);
   If_Result_6 := If_Result_5;
end if;
      Inc_Ref (If_Result_6);
--# expr-done 8362
And_Pred := Create_And (Fld_11, If_Result_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17944") else null)); 
--# expr-done 8363
   If_Result_9 := And_Pred;
else
   --# expr-start 8379 '<Let at ast.py:17961>' Let_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17961
--# scope-start
--# expr-start 8368 '<FieldAccess for resolve_names_internal_with_eq at ast.py:17963>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17963
--# expr-start 8364 '<FieldAccess for prefix at ast.py:17963>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17963
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_21 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8364
      if Fld_21.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8367 '<Predicate on BaseTypeDecl.is_array_def_with_deref at ast.py:17964>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17964
--# expr-start 8366 '<FieldAccess for type_var at ast.py:17965>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17965
--# expr-start 8365 '<FieldAccess for prefix at ast.py:17965>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17965
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_22 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8365
      if Fld_22.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_23 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_22.Node);
--# expr-done 8366
Fld_23.Value := No_Entity;
Entity_Vars.Reset (Fld_23);
Pred := Solver.Create_Predicate (Fld_23, Create_Base_Type_Decl_P_Is_Array_Def_With_Deref_0_Predicate (Origin)); 
--# expr-done 8367
--# property-call-start AdaNode.resolve_names_internal_with_eq
Fld_24 := Libadalang.Implementation.Ada_Node_P_Resolve_Names_Internal_With_Eq (Node => Fld_21.Node, Additional_Equation => Pred, Env => Env, Origin => Origin, E_Info => Fld_21.Info);
--# end
--# expr-done 8368
Res := Fld_24; 
--# bind res Res
--# expr-start 8371 '<Cast to BaseTypeDecl at ast.py:17967>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17967
--# expr-start 8370 '<FieldAccess for type_val at ast.py:17967>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17967
--# expr-start 8369 '<FieldAccess for prefix at ast.py:17967>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17967
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_25 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8369
      if Fld_25.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.type_val
Fld_26 := Libadalang.Implementation.Expr_P_Type_Val (Node => Fld_25.Node);
--# end
--# expr-done 8370
   if Fld_26.Node = null
      or else Fld_26.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_2 := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld_26.Node,
         Info => Fld_26.Info);
   else
         Cast_Result_2 := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 8371
Pfx_Typ := Cast_Result_2; 
--# bind pfx_typ Pfx_Typ
--# expr-start 8378 '<If at ast.py:17969>' If_Result_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17969
if Res then
   --# expr-start 8377 '<If at ast.py:17971>' If_Result_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17971
if Block_Var then
   --# expr-start 8373 '<FieldAccess for universal_int_bind at ast.py:17973>' Fld_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17973
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8372 '<FieldAccess for type_var at ast.py:17973>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17973
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_27 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8372
--# property-call-start AdaNode.universal_int_bind
Fld_28 := Libadalang.Implementation.Ada_Node_P_Universal_Int_Bind (Node => Self, Type_Var => Fld_27, Origin => Origin);
--# end
--# expr-done 8373
   If_Result_7 := Fld_28;
else
   --# expr-start 8376 '<Bind at ast.py:17974>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17974
--# expr-start 8374 '<FieldAccess for type_var at ast.py:17974>' Fld_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17974
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_29 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8374
Fld_29.Value := No_Entity;
Entity_Vars.Reset (Fld_29);
--# expr-start 8375 '<FieldAccess for index_type at ast.py:17974>' Fld_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17974
      if Pfx_Typ.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.index_type
Fld_30 := Libadalang.Implementation.Base_Type_Decl_P_Index_Type (Node => Pfx_Typ.Node, Dim => Block_Var_2, Origin => Origin, E_Info => Pfx_Typ.Info);
--# end
--# expr-done 8375
      Cast_Result_3 := Create_Internal_Entity
        (Node => Fld_30.Node,
         Info => Fld_30.Info);
Bind_Result_2 := Solver.Create_Assign (Fld_29, Cast_Result_3, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17974") else null)); 
--# expr-done 8376
   If_Result_7 := Bind_Result_2;
end if;
      Inc_Ref (If_Result_7);
--# expr-done 8377
   If_Result_8 := If_Result_7;
else
   False_Rel_1 := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17976") else null)); 
   If_Result_8 := False_Rel_1;
end if;
      Inc_Ref (If_Result_8);
--# expr-done 8378
Let_Result_1 := If_Result_8; Inc_Ref (Let_Result_1);
--# end
--# expr-done 8379
   If_Result_9 := Let_Result_1;
end if;
      Inc_Ref (If_Result_9);
--# expr-done 8380
Let_Result_2 := If_Result_9; Inc_Ref (Let_Result_2);
--# end
--# expr-done 8381
      Property_Result := Let_Result_2;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1272;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1272;
               Finalizer_Scope_2785;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Array_Attr_Equation;
--# end
--# property-start AttributeRef.subtype_attr_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17982
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Subtype_Attr_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1273 with Inline_Always;
      Fld : Bare_Name;
Fld_1 : Logic_Var;
Fld_2 : Internal_Entity_Name;
Fld_3 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_4 : Logic_Var;
Fld_5 : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1273 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_5);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1273;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8390 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8389 '<And at ast.py:17988>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17988
--# expr-start 8386 '<Bind at ast.py:17988>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17988
--# expr-start 8383 '<FieldAccess for ref_var at ast.py:17988>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17988
--# expr-start 8382 '<FieldAccess for prefix at ast.py:17988>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17988
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Attribute_Ref_F_Prefix;
--# expr-done 8382
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld);
--# end
--# expr-done 8383
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 8385 '<FieldAccess for name_designated_type at ast.py:17988>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17988
--# expr-start 8384 '<FieldAccess for prefix at ast.py:17988>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17988
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8384
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_3 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld_2.Node, E_Info => Fld_2.Info);
--# end
--# expr-done 8385
      Cast_Result := Create_Internal_Entity
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
Bind_Result := Solver.Create_Assign (Fld_1, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17988") else null)); 
--# expr-done 8386
--# expr-start 8388 '<FieldAccess for universal_int_bind at ast.py:17989>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17989
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8387 '<FieldAccess for type_var at ast.py:17989>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17989
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8387
--# property-call-start AdaNode.universal_int_bind
Fld_5 := Libadalang.Implementation.Ada_Node_P_Universal_Int_Bind (Node => Self, Type_Var => Fld_4, Origin => Origin);
--# end
--# expr-done 8388
And_Pred := Create_And (Bind_Result, Fld_5, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17988") else null)); 
--# expr-done 8389
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 8390
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1273;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1273;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Subtype_Attr_Equation;
--# end
--# property-start AttributeRef.inline_asm_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17993
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Inline_Asm_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1274 with Inline_Always;
            procedure Finalizer_Scope_2786 with Inline_Always;
      Block_Var : Symbol_Type;
Block_Var_1 : Internal_Entity_Base_Type_Decl;
Block_Var_2 : Internal_Entity_Expr;
Block_Var_3 : Internal_Entity_Expr;
Var_Expr : Bare_Basic_Decl;
Sym : Symbol_Type;
Sym_1 : Symbol_Type;
Fld : Bare_Identifier;
Sym_2 : Symbol_Type;
Fld_1 : Boolean;
If_Result : Symbol_Type;
Sym_3 : Symbol_Type;
Sym_4 : Symbol_Type;
Array_Lit : Symbol_Type_Array_Access;
Fld_2 : Bare_Basic_Decl;
Fld_3 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Env_Get_Result : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
Fld_4 : Internal_Entity_Assoc_List;
Fld_5 : Bare_Assoc_List;
Fld_6 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result_1 : Internal_Entity_Basic_Assoc;
Fld_7 : Internal_Entity_Expr;
Fld_8 : Internal_Entity_Assoc_List;
Fld_9 : Bare_Assoc_List;
Fld_10 : Internal_Entity_Info;
Get_Result_1 : Bare_Basic_Assoc;
As_Entity_1 : Internal_Entity_Basic_Assoc;
Is_Null_1 : Boolean;
If_Result_2 : Internal_Entity_Basic_Assoc;
Fld_11 : Internal_Entity_Expr;
Fld_12 : Internal_Entity_Name;
Fld_13 : Logic_Equation;
Fld_14 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_15 : Logic_Var;
Sym_5 : Symbol_Type;
Fld_16 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_17 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_18 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_19 : Logic_Var;
Fld_20 : Internal_Entity_Name;
Fld_21 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result_2 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Fld_22 : Logic_Equation;
And_Pred_5 : Logic_Equation;
Fld_23 : Logic_Equation;
And_Pred_6 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1274 is
            begin
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Fld_13);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_17);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Fld_18);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Fld_22);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (Fld_23);
                     Dec_Ref (And_Pred_6);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1274;
            procedure Finalizer_Scope_2786 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_2786;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8431 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8396 '<If at ast.py:17998>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17998
--# expr-start 8393 '<FieldAccess for name_is at ast.py:17999>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17999
--# expr-start 8391 '<FieldAccess for attribute at ast.py:17999>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17999
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Attribute_Ref_F_Attribute;
--# expr-done 8391
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8392 '<Symbol Asm_Input at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Asm_Input); 
--# expr-done 8392
--# property-call-start Name.name_is
Fld_1 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld, Sym => Sym_2);
--# end
--# expr-done 8393
if Fld_1 then
   --# expr-start 8394 '<Symbol Asm_Input_Operand at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Asm_Input_Operand); 
--# expr-done 8394
   If_Result := Sym;
else
   --# expr-start 8395 '<Symbol Asm_Output_Operand at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Asm_Output_Operand); 
--# expr-done 8395
   If_Result := Sym_1;
end if;
--# expr-done 8396
Block_Var := If_Result; 
--# bind return_type_name Block_Var
--# expr-start 8403 '<Cast to BaseTypeDecl at ast.py:18007>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18007
--# expr-start 8402 '<EnvGet at ast.py:18007>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18007
--# expr-start 8400 '<FieldAccess for get_unit_root_decl at ast.py:18007>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18007
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8399 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 8397 '<Symbol System at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_System); 
--# expr-done 8397
--# expr-start 8398 '<Symbol Machine_Code at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Machine_Code); 
--# expr-done 8398
Array_Lit := Create_Symbol_Type_Array (Internal_Symbol_Type_Array'(1 => Sym_3, 2 => Sym_4)); 
--# expr-done 8399
--# property-call-start AdaNode.get_unit_root_decl
Fld_2 := Libadalang.Implementation.Ada_Node_P_Get_Unit_Root_Decl (Node => Ent.Node, Name => Array_Lit, Kind => Unit_Specification);
--# end
--# expr-done 8400
Var_Expr := Fld_2;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 8401 '<FieldAccess for children_env at ast.py:18007>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18007
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Children_Env (Node => Var_Expr);
--# expr-done 8401
Scope_Result := Fld_3;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2786;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
Env_Get_Result := AST_Envs.Get_First (Self => Result_Var, Key => Block_Var, Lookup_Kind => To_Lookup_Kind_Type (Flat), Categories => (Inherited_Primitives => True, Nocat => True)); 
--# expr-done 8402
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 8403
Block_Var_1 := Cast_Result; 
--# bind return_type Block_Var_1
--# expr-start 8405 '<FieldAccess for expr at ast.py:18013>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18013
--# expr-start 8404 '<FieldAccess for args_list at ast.py:18013>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18013
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_4 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8404
Fld_5 := Fld_4.Node;
Get_Result := Get (Fld_5, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result_1 := No_Entity_Basic_Assoc;
else
Fld_6 := Fld_4.Info;
As_Entity := (Info => Fld_6, Node => Get_Result); 
   If_Result_1 := As_Entity;
end if;
      if If_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_7 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result_1.Node, E_Info => If_Result_1.Info);
--# end
--# expr-done 8405
Block_Var_2 := Fld_7; 
--# bind first_arg Block_Var_2
--# expr-start 8407 '<FieldAccess for expr at ast.py:18014>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18014
--# expr-start 8406 '<FieldAccess for args_list at ast.py:18014>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18014
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_8 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8406
Fld_9 := Fld_8.Node;
Get_Result_1 := Get (Fld_9, 1, True); 
Is_Null_1 := Get_Result_1 = null; 
if Is_Null_1 then
   If_Result_2 := No_Entity_Basic_Assoc;
else
Fld_10 := Fld_8.Info;
As_Entity_1 := (Info => Fld_10, Node => Get_Result_1); 
   If_Result_2 := As_Entity_1;
end if;
      if If_Result_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_11 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result_2.Node, E_Info => If_Result_2.Info);
--# end
--# expr-done 8407
Block_Var_3 := Fld_11; 
--# bind second_arg Block_Var_3
--# expr-start 8430 '<And at ast.py:18016>' And_Pred_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18016
--# expr-start 8428 '<And at ast.py:18016>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18016
--# expr-start 8426 '<And at ast.py:18016>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18016
--# expr-start 8421 '<And at ast.py:18016>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18016
--# expr-start 8419 '<And at ast.py:18016>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18016
--# expr-start 8417 '<And at ast.py:18016>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18016
--# expr-start 8412 '<And at ast.py:18016>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18016
--# expr-start 8409 '<FieldAccess for xref_no_overloading at ast.py:18017>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18017
--# expr-start 8408 '<FieldAccess for prefix at ast.py:18017>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18017
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8408
      if Fld_12.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_13 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Fld_12.Node, Env => Env, Origin => Origin, E_Info => Fld_12.Info);
--# end
--# expr-done 8409
--# expr-start 8411 '<Bind at ast.py:18018>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18018
--# expr-start 8410 '<FieldAccess for type_var at ast.py:18018>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18018
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8410
Fld_14.Value := No_Entity;
Entity_Vars.Reset (Fld_14);
Cast_Expr := Block_Var_1; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_14, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18018") else null)); 
--# expr-done 8411
And_Pred := Create_And (Fld_13, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18016") else null)); 
--# expr-done 8412
--# expr-start 8416 '<Bind at ast.py:18021>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18021
--# expr-start 8413 '<FieldAccess for expected_type_var at ast.py:18021>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18021
      if Block_Var_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Block_Var_2.Node);
--# expr-done 8413
Fld_15.Value := No_Entity;
Entity_Vars.Reset (Fld_15);
--# expr-start 8415 '<FieldAccess for std_entity at ast.py:18021>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18021
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8414 '<Symbol String at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_String); 
--# expr-done 8414
--# property-call-start AdaNode.std_entity
Fld_16 := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym_5);
--# end
--# expr-done 8415
Bind_Result_1 := Solver.Create_Assign (Fld_15, Fld_16, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18021") else null)); 
--# expr-done 8416
And_Pred_1 := Create_And (And_Pred, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18016") else null)); 
--# expr-done 8417
--# expr-start 8418 '<FieldAccess for sub_equation at ast.py:18022>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18022
      if Block_Var_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_17 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Block_Var_2.Node, Env => Env, Origin => Origin, E_Info => Block_Var_2.Info);
--# end
--# expr-done 8418
And_Pred_2 := Create_And (And_Pred_1, Fld_17, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18016") else null)); 
--# expr-done 8419
--# expr-start 8420 '<FieldAccess for matches_expected_type at ast.py:18023>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18023
      if Block_Var_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_18 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Block_Var_2.Node, Origin => Origin);
--# end
--# expr-done 8420
And_Pred_3 := Create_And (And_Pred_2, Fld_18, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18016") else null)); 
--# expr-done 8421
--# expr-start 8425 '<Bind at ast.py:18026>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18026
--# expr-start 8422 '<FieldAccess for expected_type_var at ast.py:18026>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18026
      if Block_Var_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Block_Var_3.Node);
--# expr-done 8422
Fld_19.Value := No_Entity;
Entity_Vars.Reset (Fld_19);
--# expr-start 8424 '<FieldAccess for name_designated_type at ast.py:18027>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18027
--# expr-start 8423 '<FieldAccess for prefix at ast.py:18027>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18027
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_20 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8423
      if Fld_20.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_21 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld_20.Node, E_Info => Fld_20.Info);
--# end
--# expr-done 8424
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_21.Node,
         Info => Fld_21.Info);
Bind_Result_2 := Solver.Create_Assign (Fld_19, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18026") else null)); 
--# expr-done 8425
And_Pred_4 := Create_And (And_Pred_3, Bind_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18016") else null)); 
--# expr-done 8426
--# expr-start 8427 '<FieldAccess for sub_equation at ast.py:18028>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18028
      if Block_Var_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_22 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Block_Var_3.Node, Env => Env, Origin => Origin, E_Info => Block_Var_3.Info);
--# end
--# expr-done 8427
And_Pred_5 := Create_And (And_Pred_4, Fld_22, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18016") else null)); 
--# expr-done 8428
--# expr-start 8429 '<FieldAccess for matches_expected_type at ast.py:18029>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18029
      if Block_Var_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_23 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Block_Var_3.Node, Origin => Origin);
--# end
--# expr-done 8429
And_Pred_6 := Create_And (And_Pred_5, Fld_23, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18016") else null)); 
--# expr-done 8430
Let_Result := And_Pred_6; Inc_Ref (Let_Result);
--# end
--# expr-done 8431
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1274;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1274;
               Finalizer_Scope_2786;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Inline_Asm_Equation;
--# end
--# property-start AttributeRef.to_address_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18033
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_To_Address_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1275 with Inline_Always;
            procedure Finalizer_Scope_2787 with Inline_Always;
      Block_Var : Internal_Entity_Basic_Subp_Decl;
Block_Var_1 : Internal_Entity_Base_Subp_Spec;
Var_Expr : Bare_Basic_Decl;
Sym : Symbol_Type;
Sym_1 : Symbol_Type;
Array_Lit : Symbol_Type_Array_Access;
Fld : Bare_Basic_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Sym_2 : Symbol_Type;
Env_Get_Result : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Subp_Decl;
Fld_2 : Internal_Entity_Base_Subp_Spec;
Fld_3 : Internal_Entity_Name;
Fld_4 : Logic_Equation;
Fld_5 : Internal_Entity_Identifier;
Fld_6 : Logic_Var;
Cast_Expr : Internal_Entity_Basic_Subp_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_7 : Internal_Entity_Assoc_List;
Fld_8 : Bare_Assoc_List;
Fld_9 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_10 : Internal_Entity_Expr;
Fld_11 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_12 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Get_Result_1 : Internal_Entity_Base_Formal_Param_Decl;
Fld_13 : Internal_Entity_Assoc_List;
Fld_14 : Bare_Assoc_List;
Fld_15 : Internal_Entity_Info;
Get_Result_2 : Bare_Basic_Assoc;
As_Entity_1 : Internal_Entity_Basic_Assoc;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Basic_Assoc;
Fld_16 : Internal_Entity_Expr;
Fld_17 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_18 : Logic_Var;
Fld_19 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1275 is
            begin
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_11);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_12);
                     Dec_Ref (Fld_17);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1275;
            procedure Finalizer_Scope_2787 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2787;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8461 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8439 '<Cast to BasicSubpDecl at ast.py:18040>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18040
--# expr-start 8438 '<EnvGet at ast.py:18040>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18040
--# expr-start 8435 '<FieldAccess for get_unit_root_decl at ast.py:18040>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18040
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8434 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 8432 '<Symbol System at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_System); 
--# expr-done 8432
--# expr-start 8433 '<Symbol Storage_Elements at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Storage_Elements); 
--# expr-done 8433
Array_Lit := Create_Symbol_Type_Array (Internal_Symbol_Type_Array'(1 => Sym, 2 => Sym_1)); 
--# expr-done 8434
--# property-call-start AdaNode.get_unit_root_decl
Fld := Libadalang.Implementation.Ada_Node_P_Get_Unit_Root_Decl (Node => Ent.Node, Name => Array_Lit, Kind => Unit_Specification);
--# end
--# expr-done 8435
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 8436 '<FieldAccess for children_env at ast.py:18040>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18040
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Children_Env (Node => Var_Expr);
--# expr-done 8436
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2787;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
--# expr-start 8437 '<Symbol To_Address at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_To_Address); 
--# expr-done 8437
Env_Get_Result := AST_Envs.Get_First (Self => Result_Var, Key => Sym_2, Lookup_Kind => To_Lookup_Kind_Type (Minimal), Categories => (Inherited_Primitives => True, Nocat => True)); 
--# expr-done 8438
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Basic_Subp_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Subp_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         Cast_Result := No_Entity_Basic_Subp_Decl;
   end if;
--# expr-done 8439
Block_Var := Cast_Result; 
--# bind to_address_subp Block_Var
--# expr-start 8440 '<FieldAccess for subp_decl_spec at ast.py:18046>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18046
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicSubpDecl.subp_decl_spec'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Subp_Decl_P_Subp_Decl_Spec (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 8440
Block_Var_1 := Fld_2; 
--# bind spec Block_Var_1
--# expr-start 8460 '<And at ast.py:18047>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18047
--# expr-start 8456 '<And at ast.py:18047>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18047
--# expr-start 8450 '<And at ast.py:18047>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18047
--# expr-start 8446 '<And at ast.py:18047>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18047
--# expr-start 8442 '<FieldAccess for sub_equation at ast.py:18048>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18048
--# expr-start 8441 '<FieldAccess for prefix at ast.py:18048>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18048
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8441
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_4 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_3.Node, Env => Env, Origin => Origin, E_Info => Fld_3.Info);
--# end
--# expr-done 8442
--# expr-start 8445 '<Bind at ast.py:18049>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18049
--# expr-start 8444 '<FieldAccess for ref_var at ast.py:18049>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18049
--# expr-start 8443 '<FieldAccess for attribute at ast.py:18049>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18049
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Identifier (Node => Ent.Node.Attribute_Ref_F_Attribute, Info => Ent.Info);
--# expr-done 8443
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_6 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_5.Node);
--# end
--# expr-done 8444
Fld_6.Value := No_Entity;
Entity_Vars.Reset (Fld_6);
Cast_Expr := Block_Var; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_6, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18049") else null)); 
--# expr-done 8445
And_Pred := Create_And (Fld_4, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18047") else null)); 
--# expr-done 8446
--# expr-start 8449 '<FieldAccess for sub_equation at ast.py:18051>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18051
--# expr-start 8448 '<FieldAccess for expr at ast.py:18051>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18051
--# expr-start 8447 '<FieldAccess for args_list at ast.py:18051>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18051
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_7 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8447
Fld_8 := Fld_7.Node;
Get_Result := Get (Fld_8, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_9 := Fld_7.Info;
As_Entity := (Info => Fld_9, Node => Get_Result); 
   If_Result := As_Entity;
end if;
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_10 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 8448
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_11 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_10.Node, Env => Env, Origin => Origin, E_Info => Fld_10.Info);
--# end
--# expr-done 8449
And_Pred_1 := Create_And (And_Pred, Fld_11, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18047") else null)); 
--# expr-done 8450
--# expr-start 8455 '<FieldAccess for call_argument_equation at ast.py:18052>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18052
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8452 '<at at ast.py:18053>' Get_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18053
--# expr-start 8451 '<FieldAccess for abstract_formal_params at ast.py:18053>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18053
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseFormalParamHolder.abstract_formal_params'
Fld_12 := Libadalang.Implementation.Dispatcher_Base_Formal_Param_Holder_P_Abstract_Formal_Params (Node => Block_Var_1.Node, E_Info => Block_Var_1.Info);
--# end
--# expr-done 8451
Get_Result_1 := Get (Fld_12, 0, True); 
--# expr-done 8452
--# expr-start 8454 '<FieldAccess for expr at ast.py:18054>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18054
--# expr-start 8453 '<FieldAccess for args_list at ast.py:18054>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18054
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_13 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8453
Fld_14 := Fld_13.Node;
Get_Result_2 := Get (Fld_14, 0, True); 
Is_Null_1 := Get_Result_2 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Basic_Assoc;
else
Fld_15 := Fld_13.Info;
As_Entity_1 := (Info => Fld_15, Node => Get_Result_2); 
   If_Result_1 := As_Entity_1;
end if;
      if If_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_16 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result_1.Node, E_Info => If_Result_1.Info);
--# end
--# expr-done 8454
--# property-call-start BaseFormalParamHolder.call_argument_equation
Fld_17 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Call_Argument_Equation (Node => Block_Var_1.Node, Param => Get_Result_1, Arg => Fld_16, Env => Env, Origin => Origin, E_Info => Block_Var_1.Info);
--# end
--# expr-done 8455
And_Pred_2 := Create_And (And_Pred_1, Fld_17, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18047") else null)); 
--# expr-done 8456
--# expr-start 8459 '<Bind at ast.py:18057>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18057
--# expr-start 8457 '<FieldAccess for type_var at ast.py:18057>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18057
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Ent.Node);
--# expr-done 8457
Fld_18.Value := No_Entity;
Entity_Vars.Reset (Fld_18);
--# expr-start 8458 '<FieldAccess for expr_type at ast.py:18057>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18057
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.expr_type'
Fld_19 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Expr_Type (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 8458
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_19.Node,
         Info => Fld_19.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_18, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18057") else null)); 
--# expr-done 8459
And_Pred_3 := Create_And (And_Pred_2, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18047") else null)); 
--# expr-done 8460
Let_Result := And_Pred_3; Inc_Ref (Let_Result);
--# end
--# expr-done 8461
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1275;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1275;
               Finalizer_Scope_2787;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_To_Address_Equation;
--# end
--# property-start AttributeRef.put_image_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18061
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Put_Image_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1276 with Inline_Always;
            procedure Finalizer_Scope_2790 with Inline_Always;
      Block_Var : Internal_Entity_Classwide_Type_Decl;
Block_Var_1 : Internal_Entity_Base_Type_Decl;
Block_Var_2 : Internal_Aspect;
Block_Var_3 : Internal_Entity_Basic_Decl;
Block_Var_4 : Internal_Entity_Expr;
Block_Var_5 : Internal_Entity_Expr;
Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Entity_Name;
Var_Expr_2 : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Classwide_Type_Decl;
Fld_2 : Internal_Entity_Name;
Fld_3 : Internal_Entity_Base_Type_Decl;
Sym : Symbol_Type;
Fld_4 : Internal_Aspect;
Scope_Result : Internal_Aspect;
Result_Var : Internal_Aspect;
Fld_5 : Internal_Entity_Expr;
Cast_Result : Internal_Entity_Name;
Fld_6 : Internal_Entity_Basic_Decl;
Scope_Result_1 : Internal_Entity_Basic_Decl;
Result_Var_1 : Internal_Entity_Basic_Decl;
Fld_7 : Internal_Entity_Assoc_List;
Fld_8 : Bare_Assoc_List;
Fld_9 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_10 : Internal_Entity_Expr;
Fld_11 : Internal_Entity_Assoc_List;
Fld_12 : Bare_Assoc_List;
Fld_13 : Internal_Entity_Info;
Get_Result_1 : Bare_Basic_Assoc;
As_Entity_1 : Internal_Entity_Basic_Assoc;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Basic_Assoc;
Fld_14 : Internal_Entity_Expr;
Fld_15 : Internal_Entity_Name;
Fld_16 : Logic_Equation;
Fld_17 : Internal_Entity_Identifier;
Fld_18 : Logic_Var;
Cast_Expr : Internal_Entity_Basic_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result : Logic_Equation;
Scope_Result_2 : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var_2 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_19 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Classwide_Type_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_20 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_21 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_22 : Logic_Var;
Cast_Expr_2 : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity;
Bind_Result_2 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Fld_23 : Logic_Equation;
And_Pred_5 : Logic_Equation;
Fld_24 : Logic_Equation;
And_Pred_6 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1276 is
            begin
                     Dec_Ref (Fld_16);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var_2);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_20);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Fld_21);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Fld_23);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (Fld_24);
                     Dec_Ref (And_Pred_6);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1276;
            procedure Finalizer_Scope_2790 is
            begin
                     Dec_Ref (Bind_Result);
            end Finalizer_Scope_2790;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8495 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8463 '<FieldAccess for classwide_type at ast.py:18065>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18065
--# expr-start 8462 '<FieldAccess for root_buffer_type at ast.py:18065>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18065
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.root_buffer_type
Fld := Libadalang.Implementation.Ada_Node_P_Root_Buffer_Type (Node => Self);
--# end
--# expr-done 8462
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.classwide_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Classwide_Type (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 8463
Block_Var := Fld_1; 
--# bind root_buffer_type Block_Var
--# expr-start 8465 '<FieldAccess for name_designated_type at ast.py:18066>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18066
--# expr-start 8464 '<FieldAccess for prefix at ast.py:18066>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18066
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8464
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_3 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld_2.Node, E_Info => Fld_2.Info);
--# end
--# expr-done 8465
Block_Var_1 := Fld_3; 
--# bind typ Block_Var_1
Var_Expr := Block_Var_1;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 8467 '<FieldAccess for get_aspect at ast.py:18067>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18067
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8466 '<Symbol put_image at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Put_Image_307); 
--# expr-done 8466
--# property-call-start BasicDecl.get_aspect
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Get_Aspect (Node => Var_Expr.Node, Name => Sym, Imprecise_Fallback => False, E_Info => Var_Expr.Info);
--# end
--# expr-done 8467
Scope_Result := Fld_4;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Aspect;
end if;
Block_Var_2 := Result_Var; 
--# bind put_image_aspect Block_Var_2
--# expr-start 8469 '<Cast to Name at ast.py:18069>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18069
--# expr-start 8468 '<FieldAccess for value at ast.py:18069>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18069
Fld_5 := Block_Var_2.Value;
--# expr-done 8468
   if Fld_5.Node = null
      or else Fld_5.Node.Kind in Ada_Name
   then
      Cast_Result := Create_Internal_Entity_Name
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
   else
         Cast_Result := No_Entity_Name;
   end if;
--# expr-done 8469
Var_Expr_1 := Cast_Result;
if Var_Expr_1 /= No_Entity_Name then
   --# scope-start
--# expr-start 8470 '<FieldAccess for referenced_decl at ast.py:18069>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18069
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld_6 := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Var_Expr_1.Node, Imprecise_Fallback => False, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 8470
Scope_Result_1 := Fld_6;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Basic_Decl;
end if;
Block_Var_3 := Result_Var_1; 
--# bind put_image_subp Block_Var_3
--# expr-start 8472 '<FieldAccess for expr at ast.py:18072>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18072
--# expr-start 8471 '<FieldAccess for args_list at ast.py:18072>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18072
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_7 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8471
Fld_8 := Fld_7.Node;
Get_Result := Get (Fld_8, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_9 := Fld_7.Info;
As_Entity := (Info => Fld_9, Node => Get_Result); 
   If_Result := As_Entity;
end if;
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_10 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 8472
Block_Var_4 := Fld_10; 
--# bind first_arg Block_Var_4
--# expr-start 8474 '<FieldAccess for expr at ast.py:18073>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18073
--# expr-start 8473 '<FieldAccess for args_list at ast.py:18073>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18073
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.args_list
Fld_11 := Libadalang.Implementation.Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8473
Fld_12 := Fld_11.Node;
Get_Result_1 := Get (Fld_12, 1, True); 
Is_Null_1 := Get_Result_1 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Basic_Assoc;
else
Fld_13 := Fld_11.Info;
As_Entity_1 := (Info => Fld_13, Node => Get_Result_1); 
   If_Result_1 := As_Entity_1;
end if;
      if If_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_14 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result_1.Node, E_Info => If_Result_1.Info);
--# end
--# expr-done 8474
Block_Var_5 := Fld_14; 
--# bind second_arg Block_Var_5
--# expr-start 8494 '<And at ast.py:18075>' And_Pred_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18075
--# expr-start 8492 '<And at ast.py:18075>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18075
--# expr-start 8490 '<And at ast.py:18075>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18075
--# expr-start 8487 '<And at ast.py:18075>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18075
--# expr-start 8485 '<And at ast.py:18075>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18075
--# expr-start 8483 '<And at ast.py:18075>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18075
--# expr-start 8480 '<And at ast.py:18075>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18075
--# expr-start 8476 '<FieldAccess for xref_no_overloading at ast.py:18076>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18076
--# expr-start 8475 '<FieldAccess for prefix at ast.py:18076>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18076
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8475
      if Fld_15.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_16 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Fld_15.Node, Env => Env, Origin => Origin, E_Info => Fld_15.Info);
--# end
--# expr-done 8476
Var_Expr_2 := Block_Var_3;
if Var_Expr_2 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind subp Var_Expr_2
--# expr-start 8479 '<Bind at ast.py:18078>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18078
--# expr-start 8478 '<FieldAccess for ref_var at ast.py:18078>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18078
--# expr-start 8477 '<FieldAccess for attribute at ast.py:18078>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18078
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_17 := Create_Internal_Entity_Identifier (Node => Ent.Node.Attribute_Ref_F_Attribute, Info => Ent.Info);
--# expr-done 8477
      if Fld_17.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_18 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_17.Node);
--# end
--# expr-done 8478
Fld_18.Value := No_Entity;
Entity_Vars.Reset (Fld_18);
Cast_Expr := Var_Expr_2; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_18, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18078") else null)); 
--# expr-done 8479
Scope_Result_2 := Bind_Result;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2790;
   Result_Var_2 := Scope_Result_2;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18079") else null)); 
   Result_Var_2 := True_Rel;
end if;
      Inc_Ref (Result_Var_2);
And_Pred := Create_And (Fld_16, Result_Var_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18075") else null)); 
--# expr-done 8480
--# expr-start 8482 '<Bind at ast.py:18082>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18082
--# expr-start 8481 '<FieldAccess for expected_type_var at ast.py:18082>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18082
      if Block_Var_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Block_Var_4.Node);
--# expr-done 8481
Fld_19.Value := No_Entity;
Entity_Vars.Reset (Fld_19);
Cast_Expr_1 := Block_Var; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_19, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18082") else null)); 
--# expr-done 8482
And_Pred_1 := Create_And (And_Pred, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18075") else null)); 
--# expr-done 8483
--# expr-start 8484 '<FieldAccess for sub_equation at ast.py:18083>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18083
      if Block_Var_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_20 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Block_Var_4.Node, Env => Env, Origin => Origin, E_Info => Block_Var_4.Info);
--# end
--# expr-done 8484
And_Pred_2 := Create_And (And_Pred_1, Fld_20, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18075") else null)); 
--# expr-done 8485
--# expr-start 8486 '<FieldAccess for matches_expected_formal_type at ast.py:18084>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18084
      if Block_Var_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_21 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Block_Var_4.Node, Origin => Origin);
--# end
--# expr-done 8486
And_Pred_3 := Create_And (And_Pred_2, Fld_21, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18075") else null)); 
--# expr-done 8487
--# expr-start 8489 '<Bind at ast.py:18086>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18086
--# expr-start 8488 '<FieldAccess for expected_type_var at ast.py:18086>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18086
      if Block_Var_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_22 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Block_Var_5.Node);
--# expr-done 8488
Fld_22.Value := No_Entity;
Entity_Vars.Reset (Fld_22);
Cast_Expr_2 := Block_Var_1; 
      Cast_Result_3 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
Bind_Result_2 := Solver.Create_Assign (Fld_22, Cast_Result_3, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18086") else null)); 
--# expr-done 8489
And_Pred_4 := Create_And (And_Pred_3, Bind_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18075") else null)); 
--# expr-done 8490
--# expr-start 8491 '<FieldAccess for sub_equation at ast.py:18087>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18087
      if Block_Var_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_23 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Block_Var_5.Node, Env => Env, Origin => Origin, E_Info => Block_Var_5.Info);
--# end
--# expr-done 8491
And_Pred_5 := Create_And (And_Pred_4, Fld_23, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18075") else null)); 
--# expr-done 8492
--# expr-start 8493 '<FieldAccess for matches_expected_formal_type at ast.py:18088>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18088
      if Block_Var_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_24 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Block_Var_5.Node, Origin => Origin);
--# end
--# expr-done 8493
And_Pred_6 := Create_And (And_Pred_5, Fld_24, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18075") else null)); 
--# expr-done 8494
Let_Result := And_Pred_6; Inc_Ref (Let_Result);
--# end
--# expr-done 8495
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1276;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1276;
               Finalizer_Scope_2790;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Put_Image_Equation;
--# end
--# property-start AttributeRef.index_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18092
pragma Warnings (Off, "is not referenced");
 function Attribute_Ref_P_Index_Equation
  (Node : Bare_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Attribute_Ref := Bare_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Attribute_Ref :=
     Internal_Entity_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1277 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Name;
Fld_1 : Symbol_Type;
Env_Get_Result : Internal_Entity;
Cast_Result : Internal_Entity_Entry_Decl;
Fld_2 : Internal_Entity_Base_Type_Decl;
Fld_3 : Internal_Entity_Name;
Fld_4 : Logic_Equation;
Fld_5 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1277 is
            begin
                     Dec_Ref (Fld_4);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1277;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8506 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8500 '<FieldAccess for family_type at ast.py:18097>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18097
--# expr-start 8499 '<Cast to EntryDecl at ast.py:18097>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18097
--# expr-start 8498 '<EnvGet at ast.py:18097>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18097
--# expr-start 8497 '<FieldAccess for name_symbol at ast.py:18097>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18097
--# expr-start 8496 '<FieldAccess for prefix at ast.py:18097>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18097
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8496
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld.Node);
--# end
--# expr-done 8497
Env_Get_Result := AST_Envs.Get_First (Self => Env, Key => Fld_1, Lookup_Kind => To_Lookup_Kind_Type (Recursive), Categories => (Inherited_Primitives => True, Nocat => True)); 
--# expr-done 8498
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Entry_Decl_Range
   then
      Cast_Result := Create_Internal_Entity_Entry_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 8499
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start EntryDecl.family_type
Fld_2 := Libadalang.Implementation.Entry_Decl_P_Family_Type (Node => Cast_Result.Node, Origin => Origin, E_Info => Cast_Result.Info);
--# end
--# expr-done 8500
Block_Var := Fld_2; 
--# bind typ Block_Var
--# expr-start 8505 '<And at ast.py:18100>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18100
--# expr-start 8502 '<FieldAccess for sub_equation at ast.py:18100>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18100
--# expr-start 8501 '<FieldAccess for prefix at ast.py:18100>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18100
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8501
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_4 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_3.Node, Env => Env, Origin => Origin, E_Info => Fld_3.Info);
--# end
--# expr-done 8502
--# expr-start 8504 '<Bind at ast.py:18100>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18100
--# expr-start 8503 '<FieldAccess for type_var at ast.py:18100>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18100
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8503
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
Cast_Expr := Block_Var; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_5, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18100") else null)); 
--# expr-done 8504
And_Pred := Create_And (Fld_4, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18100") else null)); 
--# expr-done 8505
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 8506
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1277;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1277;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Attribute_Ref_P_Index_Equation;
--# end
   --
   --  Primitives for Bare_Update_Attribute_Ref
   --
      procedure Initialize_Fields_For_Update_Attribute_Ref
        (Self : Bare_Update_Attribute_Ref
         ; Attribute_Ref_F_Prefix : Bare_Name
         ; Attribute_Ref_F_Attribute : Bare_Identifier
         ; Attribute_Ref_F_Args : Bare_Ada_Node
        ) is
      begin
            Initialize_Fields_For_Attribute_Ref
              (Self, Attribute_Ref_F_Prefix, Attribute_Ref_F_Attribute, Attribute_Ref_F_Args);
      end Initialize_Fields_For_Update_Attribute_Ref;
--# property-start UpdateAttributeRef.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18199
pragma Warnings (Off, "is not referenced");
 function Update_Attribute_Ref_P_Xref_Equation
  (Node : Bare_Update_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Update_Attribute_Ref := Bare_Update_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Update_Attribute_Ref :=
     Internal_Entity_Update_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1283 with Inline_Always;
      Block_Var : Boolean;
Block_Var_1 : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Name;
Fld_1 : Boolean;
Fld_2 : Internal_Entity_Name;
Fld_3 : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
Fld_4 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Aggregate;
Fld_5 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_6 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1283 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1283;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8519 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8508 '<FieldAccess for resolve_names_internal at ast.py:18204>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18204
--# expr-start 8507 '<FieldAccess for prefix at ast.py:18204>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18204
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8507
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.resolve_names_internal
Fld_1 := Libadalang.Implementation.Ada_Node_P_Resolve_Names_Internal (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 8508
Block_Var := Fld_1; 
--# expr-start 8511 '<Cast to BaseTypeDecl at ast.py:18205>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18205
--# expr-start 8510 '<FieldAccess for type_val at ast.py:18205>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18205
--# expr-start 8509 '<FieldAccess for prefix at ast.py:18205>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18205
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 8509
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.type_val
Fld_3 := Libadalang.Implementation.Expr_P_Type_Val (Node => Fld_2.Node);
--# end
--# expr-done 8510
   if Fld_3.Node = null
      or else Fld_3.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
   else
         Cast_Result := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 8511
Block_Var_1 := Cast_Result; 
--# bind prefix_type Block_Var_1
--# expr-start 8518 '<And at ast.py:18206>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18206
--# expr-start 8515 '<Bind at ast.py:18207>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18207
--# expr-start 8514 '<FieldAccess for type_var at ast.py:18207>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18207
--# expr-start 8513 '<Cast to Aggregate at ast.py:18207>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18207
--# expr-start 8512 '<FieldAccess for args at ast.py:18207>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18207
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity (Node => Ent.Node.Attribute_Ref_F_Args, Info => Ent.Info);
--# expr-done 8512
   if Fld_4.Node = null
      or else Fld_4.Node.Kind in Ada_Aggregate_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Aggregate
        (Node => Fld_4.Node,
         Info => Fld_4.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 8513
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Cast_Result_1.Node);
--# expr-done 8514
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
Cast_Expr := Block_Var_1; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_5, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18207") else null)); 
--# expr-done 8515
--# expr-start 8517 '<Bind at ast.py:18208>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18208
--# expr-start 8516 '<FieldAccess for type_var at ast.py:18208>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18208
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8516
Fld_6.Value := No_Entity;
Entity_Vars.Reset (Fld_6);
Cast_Expr_1 := Block_Var_1; 
      Cast_Result_3 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_6, Cast_Result_3, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18208") else null)); 
--# expr-done 8517
And_Pred := Create_And (Bind_Result, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18206") else null)); 
--# expr-done 8518
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 8519
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1283;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1283;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Update_Attribute_Ref_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Call_Expr
   --
      procedure Initialize_Fields_For_Call_Expr
        (Self : Bare_Call_Expr
         ; Call_Expr_F_Name : Bare_Name
         ; Call_Expr_F_Suffix : Bare_Ada_Node
        ) is
      begin
            Initialize_Fields_For_Name
              (Self);
            Self.Call_Expr_F_Name := Call_Expr_F_Name;
            Self.Call_Expr_F_Suffix := Call_Expr_F_Suffix;
      Self.Call_Expr_R_Called_Spec := Null_Var_Record;
      end Initialize_Fields_For_Call_Expr;
   function Call_Expr_F_Name
     (Node : Bare_Call_Expr) return Bare_Name
   is
   begin
         return Node.Call_Expr_F_Name;
   end;
   function Call_Expr_F_Suffix
     (Node : Bare_Call_Expr) return Bare_Ada_Node
   is
   begin
         return Node.Call_Expr_F_Suffix;
   end;
--# property-start CallExpr.ref_var /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13416
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Ref_Var
  (Node : Bare_Call_Expr
  )
   return Logic_Var
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Logic_Var;
      Fld : Bare_Name;
Fld_1 : Logic_Var;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 8521 '<FieldAccess for ref_var at ast.py:13416>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13416
--# expr-start 8520 '<FieldAccess for name at ast.py:13416>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13416
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Call_Expr_F_Name;
--# expr-done 8520
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld);
--# end
--# expr-done 8521
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Ref_Var;
--# end
--# property-start CallExpr.subp_spec_var /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13420
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Subp_Spec_Var
  (Node : Bare_Call_Expr
  )
   return Logic_Var
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Logic_Var;
      Fld : Logic_Var;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 8522 '<FieldAccess for r_called_spec at ast.py:13420>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13420
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Call_Expr_R_Called_Spec'Unrestricted_Access;
--# expr-done 8522
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Subp_Spec_Var;
--# end
--# property-start CallExpr.defines_subp_spec_var /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13421
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Defines_Subp_Spec_Var
  (Node : Bare_Call_Expr
  )
   return Boolean
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Defines_Subp_Spec_Var;
--# end
--# property-start CallExpr.relative_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13423
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Relative_Name
  (Node : Bare_Call_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Single_Tok_Node
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Single_Tok_Node;
      Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Single_Tok_Node;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 8524 '<FieldAccess for relative_name at ast.py:13423>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13423
--# expr-start 8523 '<FieldAccess for name at ast.py:13423>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13423
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Call_Expr_F_Name, Info => Ent.Info);
--# expr-done 8523
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.relative_name'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Relative_Name (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 8524
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Relative_Name;
--# end
--# property-start CallExpr.kind /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13426
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Kind
  (Node : Bare_Call_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Call_Expr_Kind
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Call_Expr_Kind;
      Fld : Boolean;
Fld_1 : Boolean;
Fld_2 : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_3 : Internal_Entity_Name;
Fld_4 : Internal_Entity_Base_Type_Decl;
Fld_5 : Internal_Entity_Array_Type_Def;
Is_Null : Boolean;
Not_Val : Boolean;
Dyn_Var_Bind_Result : Boolean;
Fld_6 : Internal_Entity_Assoc_List;
Fld_7 : Bare_Assoc_List;
Len : Integer;
Is_Equal : Boolean;
Fld_8 : Internal_Entity_Name;
Fld_9 : Internal_Entity_Basic_Decl;
Is_A : Boolean;
If_Result : Boolean;
If_Result_1 : Call_Expr_Kind;
If_Result_2 : Call_Expr_Kind;
If_Result_3 : Call_Expr_Kind;
If_Result_4 : Call_Expr_Kind;
Let_Result : Call_Expr_Kind;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8541 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8540 '<Cond at ast.py:13431>' If_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13431
--# expr-start 8525 '<FieldAccess for is_call at ast.py:13432>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13432
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_call
Fld := Libadalang.Implementation.Name_P_Is_Call (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8525
if Fld then
   If_Result_4 := Call;
else
--# expr-start 8526 '<FieldAccess for is_array_slice at ast.py:13435>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13435
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.is_array_slice
Fld_1 := Libadalang.Implementation.Call_Expr_P_Is_Array_Slice (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8526
if Fld_1 then
   If_Result_3 := Array_Slice;
else
--# expr-start 8533 '<bind at ast.py:13438>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13438
--# expr-start 8527 '<FieldAccess for origin_node at ast.py:13439>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13439
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_2 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 8527
Bound_Origin := Fld_2; 
--# expr-start 8532 '<Not at ast.py:13440>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13440
--# expr-start 8531 '<IsNull at ast.py:13440>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13440
--# expr-start 8530 '<FieldAccess for array_def_with_deref at ast.py:13440>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13440
--# expr-start 8529 '<FieldAccess for expression_type at ast.py:13440>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13440
--# expr-start 8528 '<FieldAccess for name at ast.py:13440>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13440
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Name (Node => Ent.Node.Call_Expr_F_Name, Info => Ent.Info);
--# expr-done 8528
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.expression_type
Fld_4 := Libadalang.Implementation.Expr_P_Expression_Type (Node => Fld_3.Node, E_Info => Fld_3.Info);
--# end
--# expr-done 8529
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.array_def_with_deref
Fld_5 := Libadalang.Implementation.Base_Type_Decl_P_Array_Def_With_Deref (Node => Fld_4.Node, Origin => Bound_Origin, E_Info => Fld_4.Info);
--# end
--# expr-done 8530
Is_Null := Fld_5.Node = null; 
--# expr-done 8531
Not_Val := not (Is_Null); 
--# expr-done 8532
Dyn_Var_Bind_Result := Not_Val; 
--# expr-done 8533
if Dyn_Var_Bind_Result then
   If_Result_2 := Array_Index;
else
--# expr-start 8536 '<Eq at ast.py:13447>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13447
--# expr-start 8535 '<length at ast.py:13447>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13447
--# expr-start 8534 '<FieldAccess for params at ast.py:13447>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13447
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.params
Fld_6 := Libadalang.Implementation.Call_Expr_P_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8534
Fld_7 := Fld_6.Node; 
Len := Length (Fld_7); 
--# expr-done 8535
Is_Equal := Len = 1; 
--# expr-done 8536
if Is_Equal then
   --# expr-start 8539 '<IsA TypeDecl at ast.py:13448>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13448
--# expr-start 8538 '<FieldAccess for referenced_decl at ast.py:13448>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13448
--# expr-start 8537 '<FieldAccess for name at ast.py:13448>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13448
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Name (Node => Ent.Node.Call_Expr_F_Name, Info => Ent.Info);
--# expr-done 8537
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld_9 := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Fld_8.Node, Imprecise_Fallback => False, E_Info => Fld_8.Info);
--# end
--# expr-done 8538
Is_A := Fld_9.Node /= null 
and then Fld_9.Node.Kind in Ada_Type_Decl_Range; 
--# expr-done 8539
   If_Result := Is_A;
else
   If_Result := False;
end if;
if If_Result then
   If_Result_1 := Type_Conversion;
else
   If_Result_1 := raise Property_Error with "undetermined CallExpr kind";
end if;
   If_Result_2 := If_Result_1;
end if;
   If_Result_3 := If_Result_2;
end if;
   If_Result_4 := If_Result_3;
end if;
--# expr-done 8540
Let_Result := If_Result_4; 
--# end
--# expr-done 8541
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Kind;
--# end
--# property-start CallExpr.is_constant /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13457
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Is_Constant
  (Node : Bare_Call_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Match_0 : Internal_Entity_Name;
Match_1 : Internal_Entity_Expr;
Fld : Call_Expr_Kind;
Is_Equal : Boolean;
Fld_1 : Internal_Entity_Assoc_List;
Fld_2 : Bare_Assoc_List;
Fld_3 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_4 : Internal_Entity_Expr;
Fld_5 : Boolean;
Match_Prefix : Internal_Entity_Expr;
Cast_Result : Internal_Entity_Name;
Let_Result : Boolean;
Scope_Result : Boolean;
Cast_Result_1 : Internal_Entity_Expr;
Let_Result_1 : Boolean;
Scope_Result_1 : Boolean;
Match_Result : Boolean;
Fld_6 : Internal_Entity_Basic_Decl;
Fld_7 : Boolean;
If_Result_1 : Boolean;
Let_Result_2 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8551 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 8550 '<Cond at ast.py:13458>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13458
--# expr-start 8543 '<Eq at ast.py:13459>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13459
--# expr-start 8542 '<FieldAccess for kind at ast.py:13459>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13459
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.kind
Fld := Libadalang.Implementation.Call_Expr_P_Kind (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8542
Is_Equal := Fld = Type_Conversion; 
--# expr-done 8543
if Is_Equal then
   --# expr-start 8547 '<Match at ast.py:13460>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13460
--# expr-start 8545 '<FieldAccess for expr at ast.py:13460>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13460
--# expr-start 8544 '<FieldAccess for params at ast.py:13460>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13460
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.params
Fld_1 := Libadalang.Implementation.Call_Expr_P_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8544
Fld_2 := Fld_1.Node;
Get_Result := Get (Fld_2, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_3 := Fld_1.Info;
As_Entity := (Info => Fld_3, Node => Get_Result); 
   If_Result := As_Entity;
end if;
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_4 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 8545
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld_4; 
case Ada_Expr (Match_Prefix.Node.Kind) is
         when Ada_Attribute_Ref .. Ada_Target_Name =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind n Match_0
--# expr-start 8546 '<FieldAccess for is_constant at ast.py:13463>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13463
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.is_constant'
Fld_5 := Libadalang.Implementation.Dispatcher_Name_P_Is_Constant (Node => Match_0.Node, E_Info => Match_0.Info);
--# end
--# expr-done 8546
Let_Result := Fld_5; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Abstract_State_Decl_Expr .. Ada_Membership_Expr | Ada_Paren_Expr .. Ada_Un_Op =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind _ Match_1
Let_Result_1 := True; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
end case;
--# expr-done 8547
   If_Result_1 := Match_Result;
else
   --# expr-start 8549 '<FieldAccess for is_constant_object at ast.py:13469>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13469
--# expr-start 8548 '<FieldAccess for referenced_decl at ast.py:13469>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13469
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld_6 := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Ent.Node, Imprecise_Fallback => False, E_Info => Ent.Info);
--# end
--# expr-done 8548
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_constant_object'
Fld_7 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Constant_Object (Node => Fld_6.Node, E_Info => Fld_6.Info);
--# end
--# expr-done 8549
   If_Result_1 := Fld_7;
end if;
--# expr-done 8550
Let_Result_2 := If_Result_1; 
--# end
--# expr-done 8551
      Property_Result := Let_Result_2;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Is_Constant;
--# end
--# property-start CallExpr.designated_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13473
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Designated_Env
  (Node : Bare_Call_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
      ; No_Visibility : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   --# bind no_visibility No_Visibility
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_995 with Inline_Always;
            procedure Finalizer_Scope_2793 with Inline_Always;
            procedure Finalizer_Scope_2794 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Match_0 : Internal_Entity_Basic_Decl;
Match_1 : Internal_Entity;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Lexical_Env := Empty_Env;
Bound_Include_Ud_Indexing : Boolean;
Fld_3 : Internal_Entity_Array_Access;
Fld_4 : Lexical_Env := Empty_Env;
Match_Prefix : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Let_Result : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Cast_Result_1 : Internal_Entity;
Let_Result_1 : Lexical_Env := Empty_Env;
Scope_Result_1 : Lexical_Env := Empty_Env;
Match_Result : Lexical_Env := Empty_Env;
Map_Result : Lexical_Env_Array_Access;
Group_Env : Lexical_Env := Empty_Env;
Dyn_Var_Bind_Result : Lexical_Env := Empty_Env;
Is_Null : Boolean;
Not_Val : Boolean;
If_Result : Lexical_Env := Empty_Env;
Let_Result_2 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_995 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Group_Env);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result_2);
            end Finalizer_Scope_995;
            procedure Finalizer_Scope_2793 is
            begin
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Match_Result);
            end Finalizer_Scope_2793;
            procedure Finalizer_Scope_2794 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_2794;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8564 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 8553 '<FieldAccess for name_designated_type at ast.py:13474>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13474
--# expr-start 8552 '<FieldAccess for name at ast.py:13474>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13474
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Call_Expr_F_Name, Info => Ent.Info);
--# expr-done 8552
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_1 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 8553
Block_Var := Fld_1; 
--# bind typ Block_Var
--# expr-start 8563 '<If at ast.py:13476>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13476
--# expr-start 8555 '<Not at ast.py:13477>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13477
--# expr-start 8554 '<IsNull at ast.py:13477>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13477
Is_Null := Block_Var.Node = null; 
--# expr-done 8554
Not_Val := not (Is_Null); 
--# expr-done 8555
if Not_Val then
   --# expr-start 8556 '<FieldAccess for defining_env at ast.py:13479>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13479
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Block_Var.Node, Origin => Origin, Include_Ud_Indexing => False, E_Info => Block_Var.Info);
--# end
--# expr-done 8556
   If_Result := Fld_2;
else
   --# expr-start 8562 '<bind at ast.py:13484>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13484
Bound_Include_Ud_Indexing := True; 
--# expr-start 8561 '<env_group at ast.py:13486>' Group_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13486
--# expr-start 8560 '<Map at ast.py:13486>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13486
   --# expr-start 8557 '<FieldAccess for env_elements at ast.py:13486>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13486
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.env_elements
Fld_3 := Libadalang.Implementation.Expr_P_Env_Elements (Node => Ent.Node, Env => Env, No_Visibility => No_Visibility, E_Info => Ent.Info);
--# end
--# expr-done 8557
      declare
         Map_Result_Vec : Lexical_Env_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Fld_3;
      begin
         for Item_252 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_252
      --# expr-start 8559 '<Match at ast.py:13486>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13486
      if Item_252.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Item_252; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Abstract_State_Decl .. Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind bd Match_0
--# expr-start 8558 '<FieldAccess for defining_env at ast.py:13487>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13487
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_4 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Match_0.Node, Origin => Origin, Include_Ud_Indexing => Bound_Include_Ud_Indexing, E_Info => Match_0.Info);
--# end
--# expr-done 8558
Let_Result := Fld_4; Inc_Ref (Let_Result);
--# end
Scope_Result := Let_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2794;
            Match_Result := Scope_Result; Inc_Ref (Match_Result);
         when Ada_Abort_Absent .. Ada_Param_Assoc | Ada_Case_Stmt_Alternative .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind _ Match_1
Let_Result_1 := Empty_Env; Inc_Ref (Let_Result_1);
--# end
Scope_Result_1 := Let_Result_1;
   Inc_Ref (Scope_Result_1);
   --# end
            Match_Result := Scope_Result_1; Inc_Ref (Match_Result);
end case;
--# expr-done 8559
         declare
            Item_To_Append : constant Lexical_Env := Match_Result;
         begin
               Inc_Ref (Item_To_Append);
            Lexical_Env_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2793;
         end loop;
      end;
         Map_Result := Create_Lexical_Env_Array
           (Items_Count => Natural (Lexical_Env_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Lexical_Env_Vectors.Get
              (Map_Result_Vec,
               I + Lexical_Env_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Lexical_Env_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 8560
Group_Env := Group (Map_Result, No_Metadata); 
--# expr-done 8561
Dyn_Var_Bind_Result := Group_Env; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 8562
   If_Result := Dyn_Var_Bind_Result;
end if;
      Inc_Ref (If_Result);
--# expr-done 8563
Let_Result_2 := If_Result; Inc_Ref (Let_Result_2);
--# end
--# expr-done 8564
      Property_Result := Let_Result_2;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_995;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_995;
               Finalizer_Scope_2793;
               Finalizer_Scope_2794;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Designated_Env;
--# end
--# property-start CallExpr.env_elements_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13494
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Env_Elements_Impl
  (Node : Bare_Call_Expr
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   Property_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_996 with Inline_Always;
      Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Array_Access;
Let_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_996 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_996;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8567 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8566 '<FieldAccess for env_elements_impl at ast.py:13495>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13495
--# expr-start 8565 '<FieldAccess for name at ast.py:13495>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13495
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Call_Expr_F_Name, Info => Ent.Info);
--# expr-done 8565
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.env_elements_impl'
Fld_1 := Libadalang.Implementation.Dispatcher_Expr_P_Env_Elements_Impl (Node => Fld.Node, Env => Env, E_Info => Fld.Info);
--# end
--# expr-done 8566
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 8567
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_996;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_996;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Env_Elements_Impl;
--# end
--# property-start CallExpr.designated_type_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13500
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Designated_Type_Impl
  (Node : Bare_Call_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_2796 with Inline_Always;
            procedure Finalizer_Scope_2797 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Block_Var_1 : Boolean;
Var_Expr : Internal_Entity_Assoc_List;
Var_Expr_1 : Internal_Entity_Base_Type_Decl;
Var_Expr_2 : Internal_Entity_Name;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Assoc_List;
Fld_3 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Scope_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Result_Var : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Fld_4 : Internal_Param_Match_Array_Access;
Fld_5 : Boolean;
Fld_6 : Internal_Entity_Defining_Name;
Fld_7 : Internal_Entity_Base_Formal_Param_Decl;
Fld_8 : Internal_Entity_Base_Type_Decl;
Fld_9 : Internal_Single_Actual;
Fld_10 : Internal_Entity_Basic_Assoc;
Fld_11 : Internal_Entity_Expr;
Cast_Result : Internal_Entity_Name;
Fld_12 : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Result_Var_1 : Internal_Entity_Base_Type_Decl;
Fld_13 : Boolean;
If_Result : Boolean;
Quantifier_Result : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
If_Result_1 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_2796 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_2796;
            procedure Finalizer_Scope_2797 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_2797;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8585 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8569 '<FieldAccess for designated_type_impl at ast.py:13502>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13502
--# expr-start 8568 '<FieldAccess for name at ast.py:13502>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13502
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Call_Expr_F_Name, Info => Ent.Info);
--# expr-done 8568
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.designated_type_impl'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Designated_Type_Impl (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 8569
Block_Var := Fld_1; 
--# bind prefix_tpe Block_Var
--# expr-start 8570 '<FieldAccess for params at ast.py:13507>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13507
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.params
Fld_2 := Libadalang.Implementation.Call_Expr_P_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8570
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Assoc_List then
   --# scope-start
   --# bind ps Var_Expr
--# expr-start 8583 '<AllQuantifier at ast.py:13508>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13508
   --# expr-start 8572 '<FieldAccess for match_formals at ast.py:13508>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13508
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Var_Expr_1 := Block_Var;
if Var_Expr_1 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 8571 '<FieldAccess for discriminants_list at ast.py:13509>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13509
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.discriminants_list'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Discriminants_List (Node => Var_Expr_1.Node, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 8571
Scope_Result := Fld_3;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2797;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Base_Formal_Param_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var);
--# property-call-start AdaNode.match_formals
Fld_4 := Libadalang.Implementation.Ada_Node_P_Match_Formals (Node => Self, Formal_Params => Result_Var, Params => Var_Expr, Is_Dottable_Subp => False);
--# end
--# expr-done 8572
      Quantifier_Result := True;
      declare
         Collection : constant Internal_Param_Match_Array_Access := Fld_4;
      begin
         for Item_253 of
               Collection.Items
         loop
   --# scope-start
               --# bind pm Item_253
--# expr-start 8573 '<FieldAccess for has_matched at ast.py:13512>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13512
Fld_5 := Item_253.Has_Matched;
--# expr-done 8573
if Fld_5 then
   --# expr-start 8582 '<FieldAccess for matching_type at ast.py:13513>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13513
--# expr-start 8576 '<FieldAccess for formal_type at ast.py:13513>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13513
--# expr-start 8575 '<FieldAccess for formal_decl at ast.py:13513>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13513
--# expr-start 8574 '<FieldAccess for formal at ast.py:13513>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13513
Fld_6 := Item_253.Formal;
--# expr-done 8574
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_7 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Fld_6.Node, E_Info => Fld_6.Info);
--# end
--# expr-done 8575
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamDecl.formal_type
Fld_8 := Libadalang.Implementation.Base_Formal_Param_Decl_P_Formal_Type (Node => Fld_7.Node, Origin => Origin, E_Info => Fld_7.Info);
--# end
--# expr-done 8576
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8580 '<Cast to Name at ast.py:13514>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13514
--# expr-start 8579 '<FieldAccess for expr at ast.py:13514>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13514
--# expr-start 8578 '<FieldAccess for assoc at ast.py:13514>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13514
--# expr-start 8577 '<FieldAccess for actual at ast.py:13514>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13514
Fld_9 := Item_253.Actual;
--# expr-done 8577
Fld_10 := Fld_9.Assoc;
--# expr-done 8578
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_11 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_10.Node, E_Info => Fld_10.Info);
--# end
--# expr-done 8579
   if Fld_11.Node = null
      or else Fld_11.Node.Kind in Ada_Name
   then
      Cast_Result := Create_Internal_Entity_Name
        (Node => Fld_11.Node,
         Info => Fld_11.Info);
   else
         Cast_Result := No_Entity_Name;
   end if;
--# expr-done 8580
Var_Expr_2 := Cast_Result;
if Var_Expr_2 /= No_Entity_Name then
   --# scope-start
--# expr-start 8581 '<FieldAccess for name_designated_type at ast.py:13514>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13514
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_12 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Var_Expr_2.Node, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 8581
Scope_Result_1 := Fld_12;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Base_Type_Decl;
end if;
--# property-call-start BaseTypeDecl.matching_type
Fld_13 := Libadalang.Implementation.Base_Type_Decl_P_Matching_Type (Node => Fld_8.Node, Expected_Type => Result_Var_1, Origin => Origin, E_Info => Fld_8.Info);
--# end
--# expr-done 8582
   If_Result := Fld_13;
else
   If_Result := False;
end if;
      Quantifier_Result := If_Result;
   --# end
               exit when not Quantifier_Result;
         end loop;
      end;
--# expr-done 8583
Scope_Result_2 := Quantifier_Result;
   --# end
      Finalizer_Scope_2796;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := True;
end if;
Block_Var_1 := Result_Var_2; 
--# bind matches_formals Block_Var_1
--# expr-start 8584 '<If at ast.py:13523>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13523
if Block_Var_1 then
   If_Result_1 := Block_Var;
else
   If_Result_1 := No_Entity_Base_Type_Decl;
end if;
--# expr-done 8584
Let_Result := If_Result_1; 
--# end
--# expr-done 8585
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_2796;
               Finalizer_Scope_2797;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Designated_Type_Impl;
--# end
--# property-start CallExpr.params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13525
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Params
  (Node : Bare_Call_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Assoc_List
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Assoc_List;
      Fld : Internal_Entity;
Cast_Result : Internal_Entity_Assoc_List;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 8587 '<Cast to AssocList at ast.py:13525>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13525
--# expr-start 8586 '<FieldAccess for suffix at ast.py:13525>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13525
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity (Node => Ent.Node.Call_Expr_F_Suffix, Info => Ent.Info);
--# expr-done 8586
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Assoc_List_Range
   then
      Cast_Result := Create_Internal_Entity_Assoc_List
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Assoc_List;
   end if;
--# expr-done 8587
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Params;
--# end
--# property-start CallExpr.check_array_slice /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13528
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Check_Array_Slice
  (Node : Bare_Call_Expr
      ; Typ : Internal_Entity_Base_Type_Decl
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind typ Typ
   --# bind origin Origin
   Property_Result : Boolean;
      Block_Var : Internal_Entity_Array_Type_Def;
Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Entity;
Var_Expr_2 : Internal_Entity_Assoc_List;
Var_Expr_3 : Internal_Entity_Name;
Fld : Internal_Entity_Array_Type_Def;
Scope_Result : Internal_Entity_Array_Type_Def;
Result_Var : Internal_Entity_Array_Type_Def;
Is_Null : Boolean;
Not_Val : Boolean;
Fld_1 : Internal_Entity;
Is_A : Boolean;
Is_A_1 : Boolean;
If_Result : Boolean;
Is_A_2 : Boolean;
If_Result_1 : Boolean;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Assoc_List;
Fld_2 : Bare_Assoc_List;
Len : Integer;
Is_Equal : Boolean;
Saved : Internal_Entity_Assoc_List;
Fld_3 : Bare_Assoc_List;
Fld_4 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null_1 : Boolean;
If_Result_2 : Internal_Entity_Basic_Assoc;
Fld_5 : Internal_Entity_Expr;
Cast_Result_1 : Internal_Entity_Name;
Fld_6 : Internal_Entity_Base_Type_Decl;
Is_Null_2 : Boolean;
Not_Val_1 : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
If_Result_3 : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
If_Result_4 : Boolean;
Scope_Result_3 : Boolean;
Result_Var_3 : Boolean;
If_Result_5 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8603 '<Block at ???>' Let_Result None
--# scope-start
Var_Expr := Typ;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind t Var_Expr
--# expr-start 8588 '<FieldAccess for array_def_with_deref at ast.py:13533>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13533
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.array_def_with_deref
Fld := Libadalang.Implementation.Base_Type_Decl_P_Array_Def_With_Deref (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 8588
Scope_Result := Fld;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Array_Type_Def;
end if;
Block_Var := Result_Var; 
--# bind atd Block_Var
--# expr-start 8590 '<Not at ast.py:13535>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13535
--# expr-start 8589 '<IsNull at ast.py:13535>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13535
Is_Null := Block_Var.Node = null; 
--# expr-done 8589
Not_Val := not (Is_Null); 
--# expr-done 8590
if Not_Val then
--# expr-start 8591 '<FieldAccess for suffix at ast.py:13536>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13536
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity (Node => Ent.Node.Call_Expr_F_Suffix, Info => Ent.Info);
--# expr-done 8591
Var_Expr_1 := Fld_1;
if Var_Expr_1 /= No_Entity then
   --# scope-start
   --# bind sfx Var_Expr_1
--# expr-start 8592 '<IsA BinOp at ast.py:13539>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13539
Is_A := Var_Expr_1.Node /= null 
and then Var_Expr_1.Node.Kind in Ada_Bin_Op_Range; 
--# expr-done 8592
if Is_A then
   If_Result := True;
else
   --# expr-start 8593 '<IsA AttributeRef at ast.py:13541>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13541
Is_A_1 := Var_Expr_1.Node /= null 
and then Var_Expr_1.Node.Kind in Ada_Attribute_Ref_Range; 
--# expr-done 8593
   If_Result := Is_A_1;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 8594 '<IsA SubtypeIndication at ast.py:13543>' Is_A_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13543
Is_A_2 := Var_Expr_1.Node /= null 
and then Var_Expr_1.Node.Kind in Ada_Subtype_Indication_Range; 
--# expr-done 8594
   If_Result_1 := Is_A_2;
end if;
if If_Result_1 then
   If_Result_4 := True;
else
--# expr-start 8595 '<Cast to AssocList at ast.py:13545>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13545
Cast_Expr := Var_Expr_1; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Assoc_List_Range
   then
      Cast_Result := Create_Internal_Entity_Assoc_List
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Assoc_List;
   end if;
--# expr-done 8595
Var_Expr_2 := Cast_Result;
if Var_Expr_2 /= No_Entity_Assoc_List then
   --# scope-start
   --# bind al Var_Expr_2
--# expr-start 8597 '<Eq at ast.py:13547>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13547
--# expr-start 8596 '<length at ast.py:13547>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13547
Fld_2 := Var_Expr_2.Node; 
Len := Length (Fld_2); 
--# expr-done 8596
Is_Equal := Len = 1; 
--# expr-done 8597
if Is_Equal then
--# expr-start 8599 '<Cast to Name at ast.py:13548>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13548
--# expr-start 8598 '<FieldAccess for expr at ast.py:13548>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13548
Saved := Var_Expr_2; 
Fld_3 := Saved.Node;
Get_Result := Get (Fld_3, 0, True); 
Is_Null_1 := Get_Result = null; 
if Is_Null_1 then
   If_Result_2 := No_Entity_Basic_Assoc;
else
Fld_4 := Saved.Info;
As_Entity := (Info => Fld_4, Node => Get_Result); 
   If_Result_2 := As_Entity;
end if;
      if If_Result_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_5 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result_2.Node, E_Info => If_Result_2.Info);
--# end
--# expr-done 8598
   if Fld_5.Node = null
      or else Fld_5.Node.Kind in Ada_Name
   then
      Cast_Result_1 := Create_Internal_Entity_Name
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
   else
         Cast_Result_1 := No_Entity_Name;
   end if;
--# expr-done 8599
Var_Expr_3 := Cast_Result_1;
if Var_Expr_3 /= No_Entity_Name then
   --# scope-start
   --# bind n Var_Expr_3
--# expr-start 8602 '<Not at ast.py:13549>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13549
--# expr-start 8601 '<IsNull at ast.py:13549>' Is_Null_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13549
--# expr-start 8600 '<FieldAccess for name_designated_type at ast.py:13549>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13549
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_6 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Var_Expr_3.Node, E_Info => Var_Expr_3.Info);
--# end
--# expr-done 8600
Is_Null_2 := Fld_6.Node = null; 
--# expr-done 8601
Not_Val_1 := not (Is_Null_2); 
--# expr-done 8602
Scope_Result_1 := Not_Val_1;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
   If_Result_3 := Result_Var_1;
else
   If_Result_3 := False;
end if;
Scope_Result_2 := If_Result_3;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := False;
end if;
   If_Result_4 := Result_Var_2;
end if;
Scope_Result_3 := If_Result_4;
   --# end
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := False;
end if;
   If_Result_5 := Result_Var_3;
else
   If_Result_5 := False;
end if;
Let_Result := If_Result_5; 
--# end
--# expr-done 8603
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Check_Array_Slice;
--# end
--# property-start CallExpr.is_array_slice /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13558
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Is_Array_Slice
  (Node : Bare_Call_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_1 : Internal_Entity_Name;
Fld_2 : Internal_Entity_Base_Type_Decl;
Fld_3 : Boolean;
Dyn_Var_Bind_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8609 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8608 '<bind at ast.py:13563>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13563
--# expr-start 8604 '<FieldAccess for origin_node at ast.py:13564>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13564
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 8604
Bound_Origin := Fld; 
--# expr-start 8607 '<FieldAccess for check_array_slice at ast.py:13565>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13565
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8606 '<FieldAccess for expression_type at ast.py:13565>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13565
--# expr-start 8605 '<FieldAccess for name at ast.py:13565>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13565
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Ent.Node.Call_Expr_F_Name, Info => Ent.Info);
--# expr-done 8605
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.expression_type
Fld_2 := Libadalang.Implementation.Expr_P_Expression_Type (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 8606
--# property-call-start CallExpr.check_array_slice
Fld_3 := Libadalang.Implementation.Call_Expr_P_Check_Array_Slice (Node => Ent.Node, Typ => Fld_2, Origin => Bound_Origin, E_Info => Ent.Info);
--# end
--# expr-done 8607
Dyn_Var_Bind_Result := Fld_3; 
--# expr-done 8608
Let_Result := Dyn_Var_Bind_Result; 
--# end
--# expr-done 8609
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Is_Array_Slice;
--# end
--# property-start CallExpr.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13569
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Xref_Equation
  (Node : Bare_Call_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1001 with Inline_Always;
      Fld : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1001 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1001;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8611 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8610 '<FieldAccess for bottom_up_name_equation at ast.py:13570>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13570
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.bottom_up_name_equation
Fld := Libadalang.Implementation.Name_P_Bottom_Up_Name_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 8610
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 8611
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1001;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1001;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Xref_Equation;
--# end
--# property-start CallExpr.type_conv_self_xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13573
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Type_Conv_Self_Xref_Equation
  (Node : Bare_Call_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1002 with Inline_Always;
      Fld : Internal_Entity_Name;
Fld_1 : Logic_Equation;
Fld_2 : Bare_Name;
Fld_3 : Logic_Var;
Fld_4 : Bare_Name;
Fld_5 : Logic_Var;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_6 : Logic_Var;
Fld_7 : Bare_Name;
Fld_8 : Logic_Var;
Bind_Result_1 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1002 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1002;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8625 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8624 '<And at ast.py:13579>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13579
--# expr-start 8619 '<And at ast.py:13579>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13579
--# expr-start 8613 '<FieldAccess for subtype_indication_equation at ast.py:13580>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13580
--# expr-start 8612 '<FieldAccess for name at ast.py:13580>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13580
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Call_Expr_F_Name, Info => Ent.Info);
--# expr-done 8612
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.subtype_indication_equation
Fld_1 := Libadalang.Implementation.Name_P_Subtype_Indication_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 8613
--# expr-start 8618 '<Bind at ast.py:13581>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13581
--# expr-start 8615 '<FieldAccess for ref_var at ast.py:13581>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13581
--# expr-start 8614 '<FieldAccess for name at ast.py:13581>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13581
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Call_Expr_F_Name;
--# expr-done 8614
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_2);
--# end
--# expr-done 8615
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
--# expr-start 8617 '<FieldAccess for type_var at ast.py:13581>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13581
--# expr-start 8616 '<FieldAccess for name at ast.py:13581>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13581
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Self.Call_Expr_F_Name;
--# expr-done 8616
      if Fld_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_4);
--# expr-done 8617
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
Bind_Result := Solver.Create_Unify (Fld_3, Fld_5, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13581") else null)); 
--# expr-done 8618
And_Pred := Create_And (Fld_1, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13579") else null)); 
--# expr-done 8619
--# expr-start 8623 '<Bind at ast.py:13582>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13582
--# expr-start 8620 '<FieldAccess for type_var at ast.py:13582>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13582
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8620
Fld_6.Value := No_Entity;
Entity_Vars.Reset (Fld_6);
--# expr-start 8622 '<FieldAccess for ref_var at ast.py:13582>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13582
--# expr-start 8621 '<FieldAccess for name at ast.py:13582>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13582
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Self.Call_Expr_F_Name;
--# expr-done 8621
      if Fld_7 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_8 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_7);
--# end
--# expr-done 8622
Fld_8.Value := No_Entity;
Entity_Vars.Reset (Fld_8);
Bind_Result_1 := Solver.Create_Unify (Fld_6, Fld_8, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13582") else null)); 
--# expr-done 8623
And_Pred_1 := Create_And (And_Pred, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13579") else null)); 
--# expr-done 8624
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 8625
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1002;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1002;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Type_Conv_Self_Xref_Equation;
--# end
--# property-start CallExpr.entry_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13586
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Entry_Equation
  (Node : Bare_Call_Expr
      ; E : Internal_Entity_Entry_Decl
      ; Root : Bare_Name
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind e E
   --# bind root Root
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1003 with Inline_Always;
            procedure Finalizer_Scope_2804 with Inline_Always;
            procedure Finalizer_Scope_2805 with Inline_Always;
            procedure Finalizer_Scope_2806 with Inline_Always;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Entity_Call_Expr;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Assoc_List;
Fld_2 : Bare_Assoc_List;
Fld_3 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_4 : Internal_Entity_Expr;
Fld_5 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_6 : Internal_Entity_Assoc_List;
Fld_7 : Bare_Assoc_List;
Fld_8 : Internal_Entity_Info;
Get_Result_1 : Bare_Basic_Assoc;
As_Entity_1 : Internal_Entity_Basic_Assoc;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Basic_Assoc;
Fld_9 : Internal_Entity_Expr;
Fld_10 : Logic_Equation;
And_Pred : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
Fld_11 : Bare_Name;
Cast_Result_1 : Bare_Call_Expr;
As_Entity_2 : Internal_Entity_Call_Expr;
Is_Null_2 : Boolean;
If_Result_2 : Internal_Entity_Call_Expr;
Fld_12 : Internal_Entity_Assoc_List;
Fld_13 : Bare_Assoc_List;
Fld_14 : Internal_Entity_Info;
As_Entity_3 : Internal_Entity_Basic_Assoc;
Is_Null_3 : Boolean;
If_Result_3 : Internal_Entity_Basic_Assoc;
Fld_15 : Internal_Entity_Expr;
Fld_16 : Logic_Equation;
Item_254 : Internal_Entity_Basic_Assoc;
Bare_Item_254 : Bare_Basic_Assoc;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Cast_Expr_1 : Internal_Entity_Entry_Decl;
Cast_Result_2 : Internal_Entity_Basic_Decl;
Fld_17 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Scope_Result_1 : Logic_Equation;
True_Rel_1 : Logic_Equation;
Result_Var_1 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Cast_Expr_2 : Internal_Entity_Entry_Decl;
Cast_Result_3 : Internal_Entity_Basic_Decl;
Fld_18 : Logic_Equation;
Fld_19 : Boolean;
If_Result_4 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1003 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (True_Rel_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Fld_18);
                     Dec_Ref (If_Result_4);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1003;
            procedure Finalizer_Scope_2804 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_10);
                     Dec_Ref (And_Pred);
            end Finalizer_Scope_2804;
            procedure Finalizer_Scope_2805 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Fld_17);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_2805;
            procedure Finalizer_Scope_2806 is
            begin
                     Dec_Ref (Fld_16);
            end Finalizer_Scope_2806;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8649 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8648 '<If at ast.py:13591>' If_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13591
--# expr-start 8626 '<FieldAccess for has_family at ast.py:13592>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13592
      if E.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start EntryDecl.has_family
Fld_19 := Libadalang.Implementation.Entry_Decl_P_Has_Family (Node => E.Node);
--# end
--# expr-done 8626
if Fld_19 then
   --# expr-start 8646 '<And at ast.py:13595>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13595
--# expr-start 8627 '<FieldAccess for family_type at ast.py:13595>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13595
      if E.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start EntryDecl.family_type
Fld := Libadalang.Implementation.Entry_Decl_P_Family_Type (Node => E.Node, Origin => Origin, E_Info => E.Info);
--# end
--# expr-done 8627
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind ft Var_Expr
--# expr-start 8635 '<And at ast.py:13597>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13597
--# expr-start 8631 '<Bind at ast.py:13597>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13597
--# expr-start 8630 '<FieldAccess for expected_type_var at ast.py:13597>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13597
--# expr-start 8629 '<FieldAccess for expr at ast.py:13597>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13597
--# expr-start 8628 '<FieldAccess for params at ast.py:13597>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13597
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.params
Fld_1 := Libadalang.Implementation.Call_Expr_P_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8628
Fld_2 := Fld_1.Node;
Get_Result := Get (Fld_2, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_3 := Fld_1.Info;
As_Entity := (Info => Fld_3, Node => Get_Result); 
   If_Result := As_Entity;
end if;
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_4 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 8629
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_4.Node);
--# expr-done 8630
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
Cast_Expr := Var_Expr; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_5, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13597") else null)); 
--# expr-done 8631
--# expr-start 8634 '<FieldAccess for matches_expected_type at ast.py:13598>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13598
--# expr-start 8633 '<FieldAccess for expr at ast.py:13598>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13598
--# expr-start 8632 '<FieldAccess for params at ast.py:13598>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13598
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.params
Fld_6 := Libadalang.Implementation.Call_Expr_P_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8632
Fld_7 := Fld_6.Node;
Get_Result_1 := Get (Fld_7, 0, True); 
Is_Null_1 := Get_Result_1 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Basic_Assoc;
else
Fld_8 := Fld_6.Info;
As_Entity_1 := (Info => Fld_8, Node => Get_Result_1); 
   If_Result_1 := As_Entity_1;
end if;
      if If_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_9 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result_1.Node, E_Info => If_Result_1.Info);
--# end
--# expr-done 8633
      if Fld_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_10 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_9.Node, Origin => Origin);
--# end
--# expr-done 8634
And_Pred := Create_And (Bind_Result, Fld_10, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13597") else null)); 
--# expr-done 8635
Scope_Result := And_Pred;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2804;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13602") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
--# expr-start 8638 '<as_entity at ast.py:13604>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13604
--# expr-start 8637 '<Cast to CallExpr at ast.py:13604>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13604
--# expr-start 8636 '<FieldAccess for parent_name at ast.py:13604>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13604
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.parent_name
Fld_11 := Libadalang.Implementation.Name_P_Parent_Name (Node => Self, Stop_At => Root);
--# end
--# expr-done 8636
   if Fld_11 = null
      or else Fld_11.Kind in Ada_Call_Expr_Range
   then
      Cast_Result_1 := Fld_11;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 8637
Is_Null_2 := Cast_Result_1 = null; 
if Is_Null_2 then
   If_Result_2 := No_Entity_Call_Expr;
else
As_Entity_2 := (Info => E_Info, Node => Cast_Result_1); 
   If_Result_2 := As_Entity_2;
end if;
--# expr-done 8638
Var_Expr_1 := If_Result_2;
if Var_Expr_1 /= No_Entity_Call_Expr then
   --# scope-start
   --# bind c Var_Expr_1
--# expr-start 8645 '<And at ast.py:13606>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13606
--# expr-start 8643 '<LogicAll at ast.py:13606>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13606
--# expr-start 8642 '<Map at ast.py:13606>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13606
   --# expr-start 8639 '<FieldAccess for params at ast.py:13606>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13606
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.params
Fld_12 := Libadalang.Implementation.Call_Expr_P_Params (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 8639
Fld_13 := Fld_12.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_13 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Assoc_List := Fld_13;
      begin
         for Untyped_Item_254 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_254 := Untyped_Item_254; 
Is_Null_3 := Bare_Item_254 = null; 
if Is_Null_3 then
   If_Result_3 := No_Entity_Basic_Assoc;
else
Fld_14 := Fld_12.Info;
As_Entity_3 := (Info => Fld_14, Node => Bare_Item_254); 
   If_Result_3 := As_Entity_3;
end if;
                  Item_254 := If_Result_3; 
   --# scope-start
               --# bind pa Item_254
      --# expr-start 8641 '<FieldAccess for sub_equation at ast.py:13606>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13606
--# expr-start 8640 '<FieldAccess for expr at ast.py:13606>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13606
      if Item_254.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_15 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Item_254.Node, E_Info => Item_254.Info);
--# end
--# expr-done 8640
      if Fld_15.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_16 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_15.Node, Env => Env, Origin => Origin, E_Info => Fld_15.Info);
--# end
--# expr-done 8641
         declare
            Item_To_Append : constant Logic_Equation := Fld_16;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2806;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 8642
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13606") else null)); 
--# expr-done 8643
--# expr-start 8644 '<FieldAccess for entity_equation at ast.py:13607>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13607
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Cast_Expr_1 := E; 
      Cast_Result_2 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
--# property-call-start CallExpr.entity_equation
Fld_17 := Libadalang.Implementation.Call_Expr_P_Entity_Equation (Node => Var_Expr_1.Node, S => Cast_Result_2, Root => Root, Env => Env, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 8644
And_Pred_1 := Create_And (Logic_Boolean_Op, Fld_17, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13606") else null)); 
--# expr-done 8645
Scope_Result_1 := And_Pred_1;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2805;
   Result_Var_1 := Scope_Result_1;
else
   True_Rel_1 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13611") else null)); 
   Result_Var_1 := True_Rel_1;
end if;
      Inc_Ref (Result_Var_1);
And_Pred_2 := Create_And (Result_Var, Result_Var_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13595") else null)); 
--# expr-done 8646
   If_Result_4 := And_Pred_2;
else
   --# expr-start 8647 '<FieldAccess for entity_equation at ast.py:13616>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13616
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Cast_Expr_2 := E; 
      Cast_Result_3 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
--# property-call-start CallExpr.entity_equation
Fld_18 := Libadalang.Implementation.Call_Expr_P_Entity_Equation (Node => Ent.Node, S => Cast_Result_3, Root => Root, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 8647
   If_Result_4 := Fld_18;
end if;
      Inc_Ref (If_Result_4);
--# expr-done 8648
Let_Result := If_Result_4; Inc_Ref (Let_Result);
--# end
--# expr-done 8649
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1003;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1003;
               Finalizer_Scope_2804;
               Finalizer_Scope_2805;
               Finalizer_Scope_2806;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Entry_Equation;
--# end
--# property-start CallExpr.entity_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13620
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Entity_Equation
  (Node : Bare_Call_Expr
      ; S : Internal_Entity_Basic_Decl
      ; Root : Bare_Name
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind s S
   --# bind root Root
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1004 with Inline_Always;
            procedure Finalizer_Scope_2807 with Inline_Always;
            procedure Finalizer_Scope_2808 with Inline_Always;
      Var_Expr : Internal_Entity_Name;
Var_Expr_1 : Internal_Entity_Name;
Fld : Bare_Name;
Fld_1 : Logic_Var;
Cast_Expr : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_2 : Boolean;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Logic_Equation;
Fld_5 : Boolean;
Fld_6 : Internal_Entity_Base_Type_Decl;
Fld_7 : Logic_Equation;
Fld_8 : Internal_Entity_Base_Subp_Spec;
Cast_Result_1 : Internal_Entity_Base_Formal_Param_Holder;
Fld_9 : Internal_Entity_Info;
Fld_10 : Internal_Metadata;
Fld_11 : Boolean;
Fld_12 : Logic_Equation;
Fld_13 : Bare_Name;
As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld_14 : Internal_Entity_Base_Type_Decl;
Fld_15 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred : Logic_Equation;
Or_Pred : Logic_Equation;
Fld_16 : Internal_Entity_Base_Subp_Spec;
Cast_Result_2 : Internal_Entity_Base_Formal_Param_Holder;
Fld_17 : Internal_Entity_Info;
Fld_18 : Internal_Metadata;
Fld_19 : Boolean;
Fld_20 : Logic_Equation;
Fld_21 : Bare_Name;
As_Entity_1 : Internal_Entity_Name;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Name;
Fld_22 : Internal_Entity_Base_Type_Decl;
Fld_23 : Logic_Equation;
Scope_Result_1 : Logic_Equation;
True_Rel_1 : Logic_Equation;
Result_Var_1 : Logic_Equation;
And_Pred_1 : Logic_Equation;
If_Result_2 : Logic_Equation;
If_Result_3 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1004 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Fld_7);
                     Dec_Ref (Fld_12);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Or_Pred);
                     Dec_Ref (Fld_20);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (True_Rel_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (If_Result_2);
                     Dec_Ref (If_Result_3);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1004;
            procedure Finalizer_Scope_2807 is
            begin
                     Dec_Ref (Fld_15);
            end Finalizer_Scope_2807;
            procedure Finalizer_Scope_2808 is
            begin
                     Dec_Ref (Fld_23);
            end Finalizer_Scope_2808;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8682 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8681 '<And at ast.py:13622>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13622
--# expr-start 8652 '<Bind at ast.py:13622>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13622
--# expr-start 8651 '<FieldAccess for ref_var at ast.py:13622>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13622
--# expr-start 8650 '<FieldAccess for name at ast.py:13622>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13622
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Call_Expr_F_Name;
--# expr-done 8650
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld);
--# end
--# expr-done 8651
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
Cast_Expr := S; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_1, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13622") else null)); 
--# expr-done 8652
--# expr-start 8680 '<Cond at ast.py:13622>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13622
--# expr-start 8653 '<FieldAccess for is_paramless at ast.py:13627>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13627
      if S.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_paramless
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Is_Paramless (Node => S.Node, E_Info => S.Info);
--# end
--# expr-done 8653
if Fld_2 then
   --# expr-start 8655 '<FieldAccess for parent_name_equation at ast.py:13627>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13627
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8654 '<FieldAccess for expr_type at ast.py:13628>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13628
      if S.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.expr_type'
Fld_3 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Expr_Type (Node => S.Node, Origin => Origin, E_Info => S.Info);
--# end
--# expr-done 8654
--# property-call-start Name.parent_name_equation
Fld_4 := Libadalang.Implementation.Name_P_Parent_Name_Equation (Node => Ent.Node, Typ => Fld_3, Root => Root, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 8655
   If_Result_3 := Fld_4;
else
--# expr-start 8656 '<FieldAccess for can_be_paramless at ast.py:13633>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13633
      if S.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.can_be_paramless
Fld_5 := Libadalang.Implementation.Basic_Decl_P_Can_Be_Paramless (Node => S.Node, E_Info => S.Info);
--# end
--# expr-done 8656
if Fld_5 then
   --# expr-start 8669 '<Or at ast.py:13633>' Or_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13633
--# expr-start 8658 '<FieldAccess for parent_name_equation at ast.py:13634>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13634
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8657 '<FieldAccess for expr_type at ast.py:13635>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13635
      if S.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.expr_type'
Fld_6 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Expr_Type (Node => S.Node, Origin => Origin, E_Info => S.Info);
--# end
--# expr-done 8657
--# property-call-start Name.parent_name_equation
Fld_7 := Libadalang.Implementation.Name_P_Parent_Name_Equation (Node => Ent.Node, Typ => Fld_6, Root => Root, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 8658
--# expr-start 8668 '<And at ast.py:13638>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13638
--# expr-start 8663 '<FieldAccess for subprogram_equation at ast.py:13639>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13639
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8659 '<FieldAccess for subp_spec_or_null at ast.py:13640>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13640
      if S.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_8 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => S.Node, E_Info => S.Info);
--# end
--# expr-done 8659
      Cast_Result_1 := Create_Internal_Entity_Base_Formal_Param_Holder
        (Node => Fld_8.Node,
         Info => Fld_8.Info);
--# expr-start 8662 '<FieldAccess for dottable_subp at ast.py:13641>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13641
--# expr-start 8661 '<FieldAccess for md at ast.py:13641>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13641
--# expr-start 8660 '<FieldAccess for info at ast.py:13641>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13641
Fld_9 := S.Info;
--# expr-done 8660
Fld_10 := Fld_9.Md;
--# expr-done 8661
Fld_11 := Fld_10.Dottable_Subp;
--# expr-done 8662
--# property-call-start CallExpr.subprogram_equation
Fld_12 := Libadalang.Implementation.Call_Expr_P_Subprogram_Equation (Node => Ent.Node, Subp_Spec => Cast_Result_1, Dottable_Subp => Fld_11, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 8663
--# expr-start 8665 '<as_entity at ast.py:13643>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13643
--# expr-start 8664 '<FieldAccess for parent_name at ast.py:13643>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13643
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.parent_name
Fld_13 := Libadalang.Implementation.Name_P_Parent_Name (Node => Ent.Node, Stop_At => Root);
--# end
--# expr-done 8664
Is_Null := Fld_13 = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
As_Entity := (Info => E_Info, Node => Fld_13); 
   If_Result := As_Entity;
end if;
--# expr-done 8665
Var_Expr := If_Result;
if Var_Expr /= No_Entity_Name then
   --# scope-start
   --# bind pn Var_Expr
--# expr-start 8667 '<FieldAccess for parent_name_equation at ast.py:13645>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13645
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8666 '<FieldAccess for expr_type at ast.py:13645>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13645
      if S.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.expr_type'
Fld_14 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Expr_Type (Node => S.Node, Origin => Origin, E_Info => S.Info);
--# end
--# expr-done 8666
--# property-call-start Name.parent_name_equation
Fld_15 := Libadalang.Implementation.Name_P_Parent_Name_Equation (Node => Var_Expr.Node, Typ => Fld_14, Root => Root, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 8667
Scope_Result := Fld_15;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2807;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13646") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred := Create_And (Fld_12, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13638") else null)); 
--# expr-done 8668
Or_Pred := Create_Or (Fld_7, And_Pred, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13633") else null)); 
--# expr-done 8669
   If_Result_2 := Or_Pred;
else
   --# expr-start 8679 '<And at ast.py:13651>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13651
--# expr-start 8674 '<FieldAccess for subprogram_equation at ast.py:13652>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13652
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8670 '<FieldAccess for subp_spec_or_null at ast.py:13653>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13653
      if S.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_16 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => S.Node, E_Info => S.Info);
--# end
--# expr-done 8670
      Cast_Result_2 := Create_Internal_Entity_Base_Formal_Param_Holder
        (Node => Fld_16.Node,
         Info => Fld_16.Info);
--# expr-start 8673 '<FieldAccess for dottable_subp at ast.py:13654>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13654
--# expr-start 8672 '<FieldAccess for md at ast.py:13654>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13654
--# expr-start 8671 '<FieldAccess for info at ast.py:13654>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13654
Fld_17 := S.Info;
--# expr-done 8671
Fld_18 := Fld_17.Md;
--# expr-done 8672
Fld_19 := Fld_18.Dottable_Subp;
--# expr-done 8673
--# property-call-start CallExpr.subprogram_equation
Fld_20 := Libadalang.Implementation.Call_Expr_P_Subprogram_Equation (Node => Ent.Node, Subp_Spec => Cast_Result_2, Dottable_Subp => Fld_19, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 8674
--# expr-start 8676 '<as_entity at ast.py:13656>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13656
--# expr-start 8675 '<FieldAccess for parent_name at ast.py:13656>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13656
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.parent_name
Fld_21 := Libadalang.Implementation.Name_P_Parent_Name (Node => Ent.Node, Stop_At => Root);
--# end
--# expr-done 8675
Is_Null_1 := Fld_21 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Name;
else
As_Entity_1 := (Info => E_Info, Node => Fld_21); 
   If_Result_1 := As_Entity_1;
end if;
--# expr-done 8676
Var_Expr_1 := If_Result_1;
if Var_Expr_1 /= No_Entity_Name then
   --# scope-start
   --# bind pn Var_Expr_1
--# expr-start 8678 '<FieldAccess for parent_name_equation at ast.py:13658>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13658
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8677 '<FieldAccess for expr_type at ast.py:13658>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13658
      if S.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.expr_type'
Fld_22 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Expr_Type (Node => S.Node, Origin => Origin, E_Info => S.Info);
--# end
--# expr-done 8677
--# property-call-start Name.parent_name_equation
Fld_23 := Libadalang.Implementation.Name_P_Parent_Name_Equation (Node => Var_Expr_1.Node, Typ => Fld_22, Root => Root, Env => Env, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 8678
Scope_Result_1 := Fld_23;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2808;
   Result_Var_1 := Scope_Result_1;
else
   True_Rel_1 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13659") else null)); 
   Result_Var_1 := True_Rel_1;
end if;
      Inc_Ref (Result_Var_1);
And_Pred_1 := Create_And (Fld_20, Result_Var_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13651") else null)); 
--# expr-done 8679
   If_Result_2 := And_Pred_1;
end if;
      Inc_Ref (If_Result_2);
   If_Result_3 := If_Result_2;
end if;
      Inc_Ref (If_Result_3);
--# expr-done 8680
And_Pred_2 := Create_And (Bind_Result, If_Result_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13622") else null)); 
--# expr-done 8681
Let_Result := And_Pred_2; Inc_Ref (Let_Result);
--# end
--# expr-done 8682
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1004;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1004;
               Finalizer_Scope_2807;
               Finalizer_Scope_2808;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Entity_Equation;
--# end
--# property-start CallExpr.is_type_conversion /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13665
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Is_Type_Conversion
  (Node : Bare_Call_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Internal_Entity_Name;
Is_A : Boolean;
Not_Val : Boolean;
Fld_1 : Internal_Entity_Name;
Fld_2 : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
Not_Val_1 : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8690 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8685 '<Not at ast.py:13670>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13670
--# expr-start 8684 '<IsA QualExpr at ast.py:13670>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13670
--# expr-start 8683 '<FieldAccess for name at ast.py:13670>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13670
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Call_Expr_F_Name, Info => Ent.Info);
--# expr-done 8683
Is_A := Fld.Node /= null 
and then Fld.Node.Kind in Ada_Qual_Expr_Range; 
--# expr-done 8684
Not_Val := not (Is_A); 
--# expr-done 8685
if Not_Val then
   --# expr-start 8689 '<Not at ast.py:13671>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13671
--# expr-start 8688 '<IsNull at ast.py:13671>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13671
--# expr-start 8687 '<FieldAccess for name_designated_type at ast.py:13671>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13671
--# expr-start 8686 '<FieldAccess for name at ast.py:13671>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13671
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Ent.Node.Call_Expr_F_Name, Info => Ent.Info);
--# expr-done 8686
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_2 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 8687
Is_Null := Fld_2.Node = null; 
--# expr-done 8688
Not_Val_1 := not (Is_Null); 
--# expr-done 8689
   If_Result := Not_Val_1;
else
   If_Result := False;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 8690
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Is_Type_Conversion;
--# end
--# property-start CallExpr.xref_stop_resolution /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13674
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Xref_Stop_Resolution
  (Node : Bare_Call_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Boolean;
Fld_1 : Boolean;
If_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
--# expr-start 8691 '<Super at ast.py:13675>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13675
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.xref_stop_resolution
Fld := Libadalang.Implementation.Expr_P_Xref_Stop_Resolution (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8691
if Fld then
   If_Result := True;
else
   --# expr-start 8692 '<FieldAccess for is_type_conversion at ast.py:13675>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13675
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.is_type_conversion
Fld_1 := Libadalang.Implementation.Call_Expr_P_Is_Type_Conversion (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8692
   If_Result := Fld_1;
end if;
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Xref_Stop_Resolution;
--# end
--# property-start CallExpr.stop_resolution_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13678
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Stop_Resolution_Equation
  (Node : Bare_Call_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1007 with Inline_Always;
      Fld : Logic_Equation;
Fld_1 : Logic_Equation;
Fld_2 : Boolean;
If_Result : Logic_Equation;
            procedure Finalizer_Scope_1007 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_1);
                     Dec_Ref (If_Result);
            end Finalizer_Scope_1007;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8696 '<If at ast.py:13678>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13678
--# expr-start 8693 '<FieldAccess for is_type_conversion at ast.py:13679>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13679
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.is_type_conversion
Fld_2 := Libadalang.Implementation.Call_Expr_P_Is_Type_Conversion (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8693
if Fld_2 then
   --# expr-start 8694 '<FieldAccess for type_conv_self_xref_equation at ast.py:13680>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13680
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.type_conv_self_xref_equation
Fld := Libadalang.Implementation.Call_Expr_P_Type_Conv_Self_Xref_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 8694
   If_Result := Fld;
else
   --# expr-start 8695 '<Super at ast.py:13681>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13681
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.stop_resolution_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Stop_Resolution_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 8695
   If_Result := Fld_1;
end if;
      Inc_Ref (If_Result);
--# expr-done 8696
      Property_Result := If_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1007;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1007;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Stop_Resolution_Equation;
--# end
--# property-start CallExpr.general_xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13685
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_General_Xref_Equation
  (Node : Bare_Call_Expr
      ; Root : Bare_Name
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind root Root
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1008 with Inline_Always;
            procedure Finalizer_Scope_2809 with Inline_Always;
            procedure Finalizer_Scope_2810 with Inline_Always;
            procedure Finalizer_Scope_2811 with Inline_Always;
            procedure Finalizer_Scope_2812 with Inline_Always;
            procedure Finalizer_Scope_2813 with Inline_Always;
      Var_Expr : Internal_Entity_Name;
Var_Expr_1 : Boolean;
Subps : Internal_Entity_Array_Access;
Var_Expr_2 : Internal_Entity_Entry_Decl;
Fld : Boolean;
Fld_1 : Logic_Equation;
Fld_2 : Internal_Entity_Assoc_List;
Fld_3 : Bare_Assoc_List;
Fld_4 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_5 : Internal_Entity_Expr;
Fld_6 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_7 : Internal_Entity_Assoc_List;
Fld_8 : Bare_Assoc_List;
Fld_9 : Internal_Entity_Info;
Get_Result_1 : Bare_Basic_Assoc;
As_Entity_1 : Internal_Entity_Basic_Assoc;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Basic_Assoc;
Fld_10 : Internal_Entity_Expr;
Fld_11 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_12 : Bare_Name;
As_Entity_2 : Internal_Entity_Name;
Is_Null_2 : Boolean;
If_Result_2 : Internal_Entity_Name;
Fld_13 : Internal_Entity_Name;
Fld_14 : Internal_Entity_Base_Type_Decl;
Fld_15 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_16 : Internal_Entity_Name;
Is_A : Boolean;
Fld_17 : Internal_Entity_Name;
Fld_18 : Boolean;
Fld_19 : Internal_Entity_Name;
Fld_20 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Base_Type_Decl;
Fld_21 : Logic_Equation;
Scope_Result_1 : Logic_Equation;
False_Rel : Logic_Equation;
Result_Var_1 : Logic_Equation;
Fld_22 : Internal_Entity_Assoc_List;
Fld_23 : Bare_Assoc_List;
Fld_24 : Internal_Entity_Info;
As_Entity_3 : Internal_Entity_Basic_Assoc;
Is_Null_3 : Boolean;
If_Result_3 : Internal_Entity_Basic_Assoc;
Fld_25 : Internal_Entity_Expr;
Fld_26 : Logic_Equation;
Item_255 : Internal_Entity_Basic_Assoc;
Bare_Item_255 : Bare_Basic_Assoc;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Fld_27 : Internal_Entity_Array_Access;
Cast_Expr_1 : Internal_Entity;
Cast_Result_2 : Internal_Entity_Entry_Decl;
Fld_28 : Logic_Equation;
Scope_Result_2 : Logic_Equation;
Cast_Expr_2 : Internal_Entity;
Cast_Result_3 : Internal_Entity_Basic_Decl;
Fld_29 : Logic_Equation;
Result_Var_2 : Logic_Equation;
Map_Result_1 : Logic_Equation_Array_Access;
Logic_Boolean_Op_1 : Logic_Equation;
Fld_30 : Internal_Entity_Name;
Fld_31 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Let_Result : Logic_Equation;
And_Pred_4 : Logic_Equation;
If_Result_4 : Logic_Equation;
If_Result_5 : Logic_Equation;
Let_Result_1 : Logic_Equation;
            procedure Finalizer_Scope_1008 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_11);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (False_Rel);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Fld_27);
                     Dec_Ref (Subps);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Logic_Boolean_Op_1);
                     Dec_Ref (Fld_31);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Let_Result);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (If_Result_4);
                     Dec_Ref (If_Result_5);
                     Dec_Ref (Let_Result_1);
            end Finalizer_Scope_1008;
            procedure Finalizer_Scope_2809 is
            begin
                     Dec_Ref (Fld_15);
            end Finalizer_Scope_2809;
            procedure Finalizer_Scope_2810 is
            begin
                     Dec_Ref (Fld_21);
            end Finalizer_Scope_2810;
            procedure Finalizer_Scope_2811 is
            begin
                     Dec_Ref (Fld_26);
            end Finalizer_Scope_2811;
            procedure Finalizer_Scope_2812 is
            begin
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Fld_29);
                     Dec_Ref (Result_Var_2);
            end Finalizer_Scope_2812;
            procedure Finalizer_Scope_2813 is
            begin
                     Dec_Ref (Fld_28);
            end Finalizer_Scope_2813;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8740 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 8739 '<Cond at ast.py:13690>' If_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13690
--# expr-start 8697 '<FieldAccess for is_type_conversion at ast.py:13691>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13691
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.is_type_conversion
Fld := Libadalang.Implementation.Call_Expr_P_Is_Type_Conversion (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8697
if Fld then
   --# expr-start 8713 '<And at ast.py:13694>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13694
--# expr-start 8707 '<And at ast.py:13694>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13694
--# expr-start 8703 '<And at ast.py:13694>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13694
--# expr-start 8698 '<FieldAccess for type_conv_self_xref_equation at ast.py:13694>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13694
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.type_conv_self_xref_equation
Fld_1 := Libadalang.Implementation.Call_Expr_P_Type_Conv_Self_Xref_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 8698
--# expr-start 8702 '<Bind at ast.py:13695>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13695
--# expr-start 8701 '<FieldAccess for expected_type_var at ast.py:13695>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13695
--# expr-start 8700 '<FieldAccess for expr at ast.py:13695>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13695
--# expr-start 8699 '<FieldAccess for params at ast.py:13695>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13695
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.params
Fld_2 := Libadalang.Implementation.Call_Expr_P_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8699
Fld_3 := Fld_2.Node;
Get_Result := Get (Fld_3, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_4 := Fld_2.Info;
As_Entity := (Info => Fld_4, Node => Get_Result); 
   If_Result := As_Entity;
end if;
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_5 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 8700
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_5.Node);
--# expr-done 8701
Fld_6.Value := No_Entity;
Entity_Vars.Reset (Fld_6);
Cast_Expr := No_Entity_Base_Type_Decl; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_6, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13695") else null)); 
--# expr-done 8702
And_Pred := Create_And (Fld_1, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13694") else null)); 
--# expr-done 8703
--# expr-start 8706 '<FieldAccess for sub_equation at ast.py:13697>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13697
--# expr-start 8705 '<FieldAccess for expr at ast.py:13697>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13697
--# expr-start 8704 '<FieldAccess for params at ast.py:13697>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13697
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.params
Fld_7 := Libadalang.Implementation.Call_Expr_P_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8704
Fld_8 := Fld_7.Node;
Get_Result_1 := Get (Fld_8, 0, True); 
Is_Null_1 := Get_Result_1 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Basic_Assoc;
else
Fld_9 := Fld_7.Info;
As_Entity_1 := (Info => Fld_9, Node => Get_Result_1); 
   If_Result_1 := As_Entity_1;
end if;
      if If_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_10 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result_1.Node, E_Info => If_Result_1.Info);
--# end
--# expr-done 8705
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_11 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_10.Node, Env => Env, Origin => Origin, E_Info => Fld_10.Info);
--# end
--# expr-done 8706
And_Pred_1 := Create_And (And_Pred, Fld_11, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13694") else null)); 
--# expr-done 8707
--# expr-start 8709 '<as_entity at ast.py:13698>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13698
--# expr-start 8708 '<FieldAccess for parent_name at ast.py:13698>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13698
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.parent_name
Fld_12 := Libadalang.Implementation.Name_P_Parent_Name (Node => Ent.Node, Stop_At => Root);
--# end
--# expr-done 8708
Is_Null_2 := Fld_12 = null; 
if Is_Null_2 then
   If_Result_2 := No_Entity_Name;
else
As_Entity_2 := (Info => E_Info, Node => Fld_12); 
   If_Result_2 := As_Entity_2;
end if;
--# expr-done 8709
Var_Expr := If_Result_2;
if Var_Expr /= No_Entity_Name then
   --# scope-start
   --# bind pn Var_Expr
--# expr-start 8712 '<FieldAccess for parent_name_equation at ast.py:13699>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13699
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8711 '<FieldAccess for name_designated_type at ast.py:13700>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13700
--# expr-start 8710 '<FieldAccess for name at ast.py:13700>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13700
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Create_Internal_Entity_Name (Node => Ent.Node.Call_Expr_F_Name, Info => Ent.Info);
--# expr-done 8710
      if Fld_13.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_14 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Fld_13.Node, E_Info => Fld_13.Info);
--# end
--# expr-done 8711
--# property-call-start Name.parent_name_equation
Fld_15 := Libadalang.Implementation.Name_P_Parent_Name_Equation (Node => Var_Expr.Node, Typ => Fld_14, Root => Root, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 8712
Scope_Result := Fld_15;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2809;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13703") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred_2 := Create_And (And_Pred_1, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13694") else null)); 
--# expr-done 8713
   If_Result_5 := And_Pred_2;
else
--# expr-start 8715 '<IsA AttributeRef at ast.py:13709>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13709
--# expr-start 8714 '<FieldAccess for name at ast.py:13709>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13709
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_16 := Create_Internal_Entity_Name (Node => Ent.Node.Call_Expr_F_Name, Info => Ent.Info);
--# expr-done 8714
Is_A := Fld_16.Node /= null 
and then Fld_16.Node.Kind in Ada_Attribute_Ref_Range; 
--# expr-done 8715
if Is_A then
--# expr-start 8717 '<FieldAccess for resolve_names_internal at ast.py:13710>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13710
--# expr-start 8716 '<FieldAccess for name at ast.py:13710>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13710
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_17 := Create_Internal_Entity_Name (Node => Ent.Node.Call_Expr_F_Name, Info => Ent.Info);
--# expr-done 8716
      if Fld_17.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.resolve_names_internal
Fld_18 := Libadalang.Implementation.Ada_Node_P_Resolve_Names_Internal (Node => Fld_17.Node, Env => Env, Origin => Origin, E_Info => Fld_17.Info);
--# end
--# expr-done 8717
Var_Expr_1 := Fld_18;
if Var_Expr_1 /= False then
   --# scope-start
   --# bind _ Var_Expr_1
--# expr-start 8721 '<FieldAccess for parent_name_equation at ast.py:13711>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13711
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8720 '<Cast to BaseTypeDecl at ast.py:13712>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13712
--# expr-start 8719 '<FieldAccess for type_val at ast.py:13712>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13712
--# expr-start 8718 '<FieldAccess for name at ast.py:13712>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13712
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Create_Internal_Entity_Name (Node => Ent.Node.Call_Expr_F_Name, Info => Ent.Info);
--# expr-done 8718
      if Fld_19.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.type_val
Fld_20 := Libadalang.Implementation.Expr_P_Type_Val (Node => Fld_19.Node);
--# end
--# expr-done 8719
   if Fld_20.Node = null
      or else Fld_20.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld_20.Node,
         Info => Fld_20.Info);
   else
         Cast_Result_1 := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 8720
--# property-call-start Name.parent_name_equation
Fld_21 := Libadalang.Implementation.Name_P_Parent_Name_Equation (Node => Ent.Node, Typ => Cast_Result_1, Root => Root, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 8721
Scope_Result_1 := Fld_21;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2810;
   Result_Var_1 := Scope_Result_1;
else
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13714") else null)); 
   Result_Var_1 := False_Rel;
end if;
      Inc_Ref (Result_Var_1);
   If_Result_4 := Result_Var_1;
else
   --# expr-start 8738 '<And at ast.py:13717>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13717
--# expr-start 8726 '<LogicAll at ast.py:13718>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13718
--# expr-start 8725 '<Map at ast.py:13718>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13718
   --# expr-start 8722 '<FieldAccess for params at ast.py:13718>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13718
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.params
Fld_22 := Libadalang.Implementation.Call_Expr_P_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8722
Fld_23 := Fld_22.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_23 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Assoc_List := Fld_23;
      begin
         for Untyped_Item_255 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_255 := Untyped_Item_255; 
Is_Null_3 := Bare_Item_255 = null; 
if Is_Null_3 then
   If_Result_3 := No_Entity_Basic_Assoc;
else
Fld_24 := Fld_22.Info;
As_Entity_3 := (Info => Fld_24, Node => Bare_Item_255); 
   If_Result_3 := As_Entity_3;
end if;
                  Item_255 := If_Result_3; 
   --# scope-start
               --# bind pa Item_255
      --# expr-start 8724 '<FieldAccess for sub_equation at ast.py:13718>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13718
--# expr-start 8723 '<FieldAccess for expr at ast.py:13718>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13718
      if Item_255.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_25 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Item_255.Node, E_Info => Item_255.Info);
--# end
--# expr-done 8723
      if Fld_25.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_26 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_25.Node, Env => Env, Origin => Origin, E_Info => Fld_25.Info);
--# end
--# expr-done 8724
         declare
            Item_To_Append : constant Logic_Equation := Fld_26;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2811;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 8725
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13718") else null)); 
--# expr-done 8726
--# expr-start 8737 '<Let at ast.py:13722>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13722
--# scope-start
--# expr-start 8727 '<FieldAccess for env_elements at ast.py:13722>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13722
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.env_elements
Fld_27 := Libadalang.Implementation.Expr_P_Env_Elements (Node => Ent.Node, Env => Env, No_Visibility => False, E_Info => Ent.Info);
--# end
--# expr-done 8727
Subps := Fld_27; Inc_Ref (Subps);
--# bind subps Subps
--# expr-start 8736 '<And at ast.py:13722>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13722
--# expr-start 8733 '<LogicAny at ast.py:13723>' Logic_Boolean_Op_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13723
--# expr-start 8732 '<Map at ast.py:13723>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13723
      declare
         Map_Result_1_Vec : Logic_Equation_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Subps;
      begin
         for Item_256 of
               Collection.Items
         loop
   --# scope-start
               --# bind s Item_256
--# expr-start 8728 '<Cast to EntryDecl at ast.py:13724>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13724
Cast_Expr_1 := Item_256; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Entry_Decl_Range
   then
      Cast_Result_2 := Create_Internal_Entity_Entry_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_2 := No_Entity_Entry_Decl;
   end if;
--# expr-done 8728
Var_Expr_2 := Cast_Result_2;
if Var_Expr_2 /= No_Entity_Entry_Decl then
   --# scope-start
   --# bind e Var_Expr_2
--# expr-start 8729 '<FieldAccess for entry_equation at ast.py:13725>' Fld_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13725
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.entry_equation
Fld_28 := Libadalang.Implementation.Call_Expr_P_Entry_Equation (Node => Ent.Node, E => Var_Expr_2, Root => Root, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 8729
Scope_Result_2 := Fld_28;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2813;
   Result_Var_2 := Scope_Result_2;
else
   --# expr-start 8731 '<FieldAccess for entity_equation at ast.py:13726>' Fld_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13726
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8730 '<Cast to BasicDecl at ast.py:13727>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13727
Cast_Expr_2 := Item_256; 
   if Cast_Expr_2.Node = null
      or else Cast_Expr_2.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_3 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 8730
--# property-call-start CallExpr.entity_equation
Fld_29 := Libadalang.Implementation.Call_Expr_P_Entity_Equation (Node => Ent.Node, S => Cast_Result_3, Root => Root, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 8731
   Result_Var_2 := Fld_29;
end if;
      Inc_Ref (Result_Var_2);
         declare
            Item_To_Append : constant Logic_Equation := Result_Var_2;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2812;
         end loop;
      end;
         Map_Result_1 := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_1_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 8732
Logic_Boolean_Op_1 := Solver.Create_Any (Relation_Array (Map_Result_1.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13723") else null)); 
--# expr-done 8733
--# expr-start 8735 '<FieldAccess for sub_equation at ast.py:13732>' Fld_31 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13732
--# expr-start 8734 '<FieldAccess for name at ast.py:13732>' Fld_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13732
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_30 := Create_Internal_Entity_Name (Node => Ent.Node.Call_Expr_F_Name, Info => Ent.Info);
--# expr-done 8734
      if Fld_30.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_31 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_30.Node, Env => Env, Origin => Origin, E_Info => Fld_30.Info);
--# end
--# expr-done 8735
And_Pred_3 := Create_And (Logic_Boolean_Op_1, Fld_31, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13722") else null)); 
--# expr-done 8736
Let_Result := And_Pred_3; Inc_Ref (Let_Result);
--# end
--# expr-done 8737
And_Pred_4 := Create_And (Logic_Boolean_Op, Let_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13717") else null)); 
--# expr-done 8738
   If_Result_4 := And_Pred_4;
end if;
      Inc_Ref (If_Result_4);
   If_Result_5 := If_Result_4;
end if;
      Inc_Ref (If_Result_5);
--# expr-done 8739
Let_Result_1 := If_Result_5; Inc_Ref (Let_Result_1);
--# end
--# expr-done 8740
      Property_Result := Let_Result_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1008;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1008;
               Finalizer_Scope_2809;
               Finalizer_Scope_2810;
               Finalizer_Scope_2811;
               Finalizer_Scope_2812;
               Finalizer_Scope_2813;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_General_Xref_Equation;
--# end
--# property-start CallExpr.subscriptable_type_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13740
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Subscriptable_Type_Equation
  (Node : Bare_Call_Expr
      ; Typ : Internal_Entity_Base_Type_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind typ Typ
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1009 with Inline_Always;
            procedure Finalizer_Scope_2816 with Inline_Always;
            procedure Finalizer_Scope_2817 with Inline_Always;
            procedure Finalizer_Scope_2819 with Inline_Always;
            procedure Finalizer_Scope_2821 with Inline_Always;
            procedure Finalizer_Scope_2822 with Inline_Always;
            procedure Finalizer_Scope_2823 with Inline_Always;
            procedure Finalizer_Scope_2824 with Inline_Always;
            procedure Finalizer_Scope_2825 with Inline_Always;
            procedure Finalizer_Scope_2826 with Inline_Always;
            procedure Finalizer_Scope_2827 with Inline_Always;
            procedure Finalizer_Scope_2828 with Inline_Always;
      Block_Var : Internal_Entity_Array_Type_Def;
Block_Var_1 : Internal_Entity_Base_Type_Decl;
Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Entity_Base_Type_Decl;
Var_Expr_2 : Internal_Entity_Access_To_Subp_Def;
Var_Expr_3 : Internal_Entity_Array_Type_Def;
Match_0 : Internal_Entity_Assoc_List;
Match_1 : Internal_Entity_Bin_Op;
Match_2 : Internal_Entity_Attribute_Ref;
Match_3 : Internal_Entity_Subtype_Indication;
Match_4 : Internal_Entity;
Var_Expr_4 : Internal_Entity_Name;
Var_Expr_5 : Internal_Entity_Basic_Assoc;
Var_Expr_6 : Internal_Entity_Assoc_List;
Formal : Internal_Entity_Defining_Name;
Ret_Type : Internal_Entity_Base_Type_Decl;
Param : Internal_Entity_Expr;
Fld : Internal_Entity_Array_Type_Def;
Scope_Result : Internal_Entity_Array_Type_Def;
Result_Var : Internal_Entity_Array_Type_Def;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Result_Var_1 : Internal_Entity_Base_Type_Decl;
Fld_3 : Internal_Entity_Access_Def;
Is_A : Boolean;
Fld_4 : Internal_Entity_Access_Def;
Cast_Result : Internal_Entity_Access_To_Subp_Def;
Fld_5 : Internal_Entity_Subp_Spec;
Cast_Result_1 : Internal_Entity_Base_Formal_Param_Holder;
Fld_6 : Logic_Equation;
Fld_7 : Internal_Entity_Assoc_List;
Fld_8 : Bare_Assoc_List;
Fld_9 : Internal_Entity_Info;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result_1 : Internal_Entity_Basic_Assoc;
Fld_10 : Internal_Entity_Expr;
Fld_11 : Logic_Equation;
Item_257 : Internal_Entity_Basic_Assoc;
Bare_Item_257 : Bare_Basic_Assoc;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
And_Pred : Logic_Equation;
Scope_Result_2 : Logic_Equation;
False_Rel : Logic_Equation;
Result_Var_2 : Logic_Equation;
Fld_12 : Internal_Entity_Array_Indices;
Scope_Result_3 : Internal_Entity_Array_Indices;
Result_Var_3 : Internal_Entity_Array_Indices;
Is_Null_1 : Boolean;
Not_Val : Boolean;
Fld_13 : Internal_Entity;
Fld_14 : Internal_Entity_Assoc_List;
Fld_15 : Bare_Assoc_List;
Fld_16 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity_1 : Internal_Entity_Basic_Assoc;
Is_Null_2 : Boolean;
If_Result_2 : Internal_Entity_Basic_Assoc;
Fld_17 : Internal_Entity_Expr;
Scope_Result_4 : Internal_Entity_Expr;
Result_Var_4 : Internal_Entity_Expr;
Cast_Result_2 : Internal_Entity_Name;
False_Rel_1 : Logic_Equation;
Fld_18 : Logic_Equation;
Fld_19 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_20 : Internal_Entity_Base_Type_Decl;
Is_Null_3 : Boolean;
If_Result_3 : Logic_Equation;
Scope_Result_5 : Logic_Equation;
False_Rel_2 : Logic_Equation;
Result_Var_5 : Logic_Equation;
Fld_21 : Internal_Entity_Assoc_List;
Saved : Internal_Entity_Assoc_List;
Fld_22 : Bare_Assoc_List;
Fld_23 : Internal_Entity_Info;
As_Entity_2 : Internal_Entity_Basic_Assoc;
Is_Null_4 : Boolean;
If_Result_4 : Internal_Entity_Basic_Assoc;
Fld_24 : Internal_Entity_Expr;
Fld_25 : Logic_Equation;
Fld_26 : Internal_Entity_Array_Indices;
Fld_27 : Internal_Entity_Expr;
Fld_28 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Item_258 : Internal_Entity_Basic_Assoc;
Bare_Item_258 : Bare_Basic_Assoc;
Index_259 : Integer;
Map_Result_1 : Logic_Equation_Array_Access;
Scope_Result_6 : Logic_Equation_Array_Access;
Result_Var_6 : Logic_Equation_Array_Access;
Logic_Boolean_Op_1 : Logic_Equation;
Fld_29 : Logic_Var;
Fld_30 : Internal_Entity_Base_Type_Decl;
Cast_Result_4 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Or_Pred : Logic_Equation;
Fld_31 : Internal_Entity_Array_Indices;
Fld_32 : Internal_Entity_Expr;
Fld_33 : Logic_Equation;
Fld_34 : Internal_Entity_Array_Indices;
Fld_35 : Internal_Entity_Expr;
Fld_36 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Fld_37 : Logic_Var;
Fld_38 : Internal_Entity_Expr;
Fld_39 : Logic_Var;
Bind_Result_2 : Logic_Equation;
And_Pred_5 : Logic_Equation;
Fld_40 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_5 : Internal_Entity;
Bind_Result_3 : Logic_Equation;
And_Pred_6 : Logic_Equation;
Fld_41 : Internal_Entity_Expr;
Fld_42 : Logic_Equation;
And_Pred_7 : Logic_Equation;
Fld_43 : Internal_Entity_Expr;
Fld_44 : Logic_Equation;
And_Pred_8 : Logic_Equation;
Fld_45 : Logic_Equation;
Fld_46 : Internal_Entity_Array_Indices;
Cast_Expr_2 : Internal_Entity_Attribute_Ref;
Cast_Result_6 : Internal_Entity_Expr;
Fld_47 : Logic_Equation;
And_Pred_9 : Logic_Equation;
Fld_48 : Logic_Var;
Cast_Expr_3 : Internal_Entity_Base_Type_Decl;
Cast_Result_7 : Internal_Entity;
Bind_Result_4 : Logic_Equation;
And_Pred_10 : Logic_Equation;
Fld_49 : Logic_Equation;
Fld_50 : Logic_Var;
Cast_Expr_4 : Internal_Entity_Base_Type_Decl;
Cast_Result_8 : Internal_Entity;
Bind_Result_5 : Logic_Equation;
And_Pred_11 : Logic_Equation;
False_Rel_3 : Logic_Equation;
Match_Prefix : Internal_Entity;
Cast_Result_9 : Internal_Entity_Assoc_List;
Let_Result : Logic_Equation;
Scope_Result_7 : Logic_Equation;
Cast_Result_10 : Internal_Entity_Bin_Op;
Let_Result_1 : Logic_Equation;
Scope_Result_8 : Logic_Equation;
Cast_Result_11 : Internal_Entity_Attribute_Ref;
Let_Result_2 : Logic_Equation;
Scope_Result_9 : Logic_Equation;
Cast_Result_12 : Internal_Entity_Subtype_Indication;
Let_Result_3 : Logic_Equation;
Scope_Result_10 : Logic_Equation;
Cast_Result_13 : Internal_Entity;
Let_Result_4 : Logic_Equation;
Scope_Result_11 : Logic_Equation;
Match_Result : Logic_Equation;
Is_Null_5 : Boolean;
Not_Val_1 : Boolean;
Fld_51 : Boolean;
If_Result_5 : Boolean;
Fld_52 : Internal_Entity_Basic_Decl_Array_Access;
Fld_53 : Internal_Entity_Basic_Decl_Array_Access;
Concat_Result : Internal_Entity_Basic_Decl_Array_Access;
Fld_54 : Internal_Entity_Base_Subp_Spec;
Fld_55 : Internal_Entity_Defining_Name_Array_Access;
Get_Result_1 : Internal_Entity_Defining_Name;
Fld_56 : Internal_Entity_Base_Subp_Spec;
Fld_57 : Internal_Entity_Base_Type_Decl;
Fld_58 : Internal_Entity_Assoc_List;
Fld_59 : Bare_Assoc_List;
Fld_60 : Internal_Entity_Info;
Get_Result_2 : Bare_Basic_Assoc;
As_Entity_3 : Internal_Entity_Basic_Assoc;
Is_Null_6 : Boolean;
If_Result_6 : Internal_Entity_Basic_Assoc;
Fld_61 : Internal_Entity_Expr;
Fld_62 : Logic_Var;
Cast_Expr_5 : Internal_Entity_Base_Type_Decl;
Cast_Result_14 : Internal_Entity;
Bind_Result_6 : Logic_Equation;
Fld_63 : Logic_Var;
Fld_64 : Internal_Entity_Base_Formal_Param_Decl;
Fld_65 : Internal_Entity_Base_Type_Decl;
Cast_Result_15 : Internal_Entity;
Bind_Result_7 : Logic_Equation;
And_Pred_12 : Logic_Equation;
Fld_66 : Logic_Equation;
And_Pred_13 : Logic_Equation;
Fld_67 : Logic_Equation;
And_Pred_14 : Logic_Equation;
Let_Result_5 : Logic_Equation;
Map_Result_2 : Logic_Equation_Array_Access;
Logic_Boolean_Op_2 : Logic_Equation;
False_Rel_4 : Logic_Equation;
If_Result_7 : Logic_Equation;
If_Result_8 : Logic_Equation;
If_Result_9 : Logic_Equation;
Let_Result_6 : Logic_Equation;
            procedure Finalizer_Scope_1009 is
            begin
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (False_Rel);
                     Dec_Ref (Result_Var_2);
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result_7);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_8);
                     Dec_Ref (Let_Result_2);
                     Dec_Ref (Scope_Result_9);
                     Dec_Ref (Let_Result_3);
                     Dec_Ref (Scope_Result_10);
                     Dec_Ref (Let_Result_4);
                     Dec_Ref (Scope_Result_11);
                     Dec_Ref (Match_Result);
                     Dec_Ref (Fld_52);
                     Dec_Ref (Fld_53);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Map_Result_2);
                     Dec_Ref (Logic_Boolean_Op_2);
                     Dec_Ref (False_Rel_4);
                     Dec_Ref (If_Result_7);
                     Dec_Ref (If_Result_8);
                     Dec_Ref (If_Result_9);
                     Dec_Ref (Let_Result_6);
            end Finalizer_Scope_1009;
            procedure Finalizer_Scope_2816 is
            begin
                     Dec_Ref (Fld_6);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (And_Pred);
            end Finalizer_Scope_2816;
            procedure Finalizer_Scope_2817 is
            begin
                     Dec_Ref (Fld_11);
            end Finalizer_Scope_2817;
            procedure Finalizer_Scope_2819 is
            begin
                     Dec_Ref (Scope_Result_5);
                     Dec_Ref (False_Rel_2);
                     Dec_Ref (Result_Var_5);
                     Dec_Ref (Scope_Result_6);
                     Dec_Ref (Result_Var_6);
                     Dec_Ref (Logic_Boolean_Op_1);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Or_Pred);
            end Finalizer_Scope_2819;
            procedure Finalizer_Scope_2821 is
            begin
                     Dec_Ref (False_Rel_1);
                     Dec_Ref (Fld_18);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (If_Result_3);
            end Finalizer_Scope_2821;
            procedure Finalizer_Scope_2822 is
            begin
                     Dec_Ref (Map_Result_1);
            end Finalizer_Scope_2822;
            procedure Finalizer_Scope_2823 is
            begin
                     Dec_Ref (Fld_25);
                     Dec_Ref (Fld_28);
                     Dec_Ref (And_Pred_2);
            end Finalizer_Scope_2823;
            procedure Finalizer_Scope_2824 is
            begin
                     Dec_Ref (Fld_33);
                     Dec_Ref (Fld_36);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (Bind_Result_3);
                     Dec_Ref (And_Pred_6);
                     Dec_Ref (Fld_42);
                     Dec_Ref (And_Pred_7);
                     Dec_Ref (Fld_44);
                     Dec_Ref (And_Pred_8);
            end Finalizer_Scope_2824;
            procedure Finalizer_Scope_2825 is
            begin
                     Dec_Ref (Fld_45);
                     Dec_Ref (Fld_47);
                     Dec_Ref (And_Pred_9);
                     Dec_Ref (Bind_Result_4);
                     Dec_Ref (And_Pred_10);
            end Finalizer_Scope_2825;
            procedure Finalizer_Scope_2826 is
            begin
                     Dec_Ref (Fld_49);
                     Dec_Ref (Bind_Result_5);
                     Dec_Ref (And_Pred_11);
            end Finalizer_Scope_2826;
            procedure Finalizer_Scope_2827 is
            begin
                     Dec_Ref (False_Rel_3);
            end Finalizer_Scope_2827;
            procedure Finalizer_Scope_2828 is
            begin
                     Dec_Ref (Fld_55);
                     Dec_Ref (Bind_Result_6);
                     Dec_Ref (Bind_Result_7);
                     Dec_Ref (And_Pred_12);
                     Dec_Ref (Fld_66);
                     Dec_Ref (And_Pred_13);
                     Dec_Ref (Fld_67);
                     Dec_Ref (And_Pred_14);
                     Dec_Ref (Let_Result_5);
            end Finalizer_Scope_2828;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8846 '<Block at ???>' Let_Result_6 None
--# scope-start
Var_Expr := Typ;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind t Var_Expr
--# expr-start 8741 '<FieldAccess for array_def_with_deref at ast.py:13745>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13745
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.array_def_with_deref
Fld := Libadalang.Implementation.Base_Type_Decl_P_Array_Def_With_Deref (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 8741
Scope_Result := Fld;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Array_Type_Def;
end if;
Block_Var := Result_Var; 
--# bind atd Block_Var
Var_Expr_1 := Typ;
if Var_Expr_1 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind t Var_Expr_1
--# expr-start 8744 '<If at ast.py:13747>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13747
--# expr-start 8742 '<FieldAccess for is_implicit_deref at ast.py:13747>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13747
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_implicit_deref
Fld_2 := Libadalang.Implementation.Base_Type_Decl_P_Is_Implicit_Deref (Node => Var_Expr_1.Node, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 8742
if Fld_2 then
   --# expr-start 8743 '<FieldAccess for accessed_type at ast.py:13747>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13747
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Var_Expr_1.Node, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 8743
   If_Result := Fld_1;
else
   If_Result := Var_Expr_1;
end if;
--# expr-done 8744
Scope_Result_1 := If_Result;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Base_Type_Decl;
end if;
Block_Var_1 := Result_Var_1; 
--# bind real_typ Block_Var_1
--# expr-start 8845 '<Cond at ast.py:13750>' If_Result_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13750
--# expr-start 8746 '<IsA AccessToSubpDef at ast.py:13752>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13752
--# expr-start 8745 '<FieldAccess for access_def at ast.py:13752>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13752
      if Typ.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.access_def'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Access_Def (Node => Typ.Node, Origin => Origin, E_Info => Typ.Info);
--# end
--# expr-done 8745
Is_A := Fld_3.Node /= null 
and then Fld_3.Node.Kind in Ada_Access_To_Subp_Def_Range; 
--# expr-done 8746
if Is_A then
--# expr-start 8748 '<Cast to AccessToSubpDef at ast.py:13753>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13753
--# expr-start 8747 '<FieldAccess for access_def at ast.py:13753>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13753
      if Typ.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.access_def'
Fld_4 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Access_Def (Node => Typ.Node, Origin => Origin, E_Info => Typ.Info);
--# end
--# expr-done 8747
   if Fld_4.Node = null
      or else Fld_4.Node.Kind in Ada_Access_To_Subp_Def_Range
   then
      Cast_Result := Create_Internal_Entity_Access_To_Subp_Def
        (Node => Fld_4.Node,
         Info => Fld_4.Info);
   else
         Cast_Result := No_Entity_Access_To_Subp_Def;
   end if;
--# expr-done 8748
Var_Expr_2 := Cast_Result;
if Var_Expr_2 /= No_Entity_Access_To_Subp_Def then
   --# scope-start
   --# bind asd Var_Expr_2
--# expr-start 8756 '<And at ast.py:13755>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13755
--# expr-start 8750 '<FieldAccess for subprogram_equation at ast.py:13755>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13755
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8749 '<FieldAccess for subp_spec at ast.py:13755>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13755
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Subp_Spec (Node => Var_Expr_2.Node.Access_To_Subp_Def_F_Subp_Spec, Info => Var_Expr_2.Info);
--# expr-done 8749
      Cast_Result_1 := Create_Internal_Entity_Base_Formal_Param_Holder
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
--# property-call-start CallExpr.subprogram_equation
Fld_6 := Libadalang.Implementation.Call_Expr_P_Subprogram_Equation (Node => Ent.Node, Subp_Spec => Cast_Result_1, Dottable_Subp => False, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 8750
--# expr-start 8755 '<LogicAll at ast.py:13756>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13756
--# expr-start 8754 '<Map at ast.py:13756>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13756
   --# expr-start 8751 '<FieldAccess for params at ast.py:13756>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13756
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.params
Fld_7 := Libadalang.Implementation.Call_Expr_P_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8751
Fld_8 := Fld_7.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_8 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Assoc_List := Fld_8;
      begin
         for Untyped_Item_257 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_257 := Untyped_Item_257; 
Is_Null := Bare_Item_257 = null; 
if Is_Null then
   If_Result_1 := No_Entity_Basic_Assoc;
else
Fld_9 := Fld_7.Info;
As_Entity := (Info => Fld_9, Node => Bare_Item_257); 
   If_Result_1 := As_Entity;
end if;
                  Item_257 := If_Result_1; 
   --# scope-start
               --# bind pa Item_257
      --# expr-start 8753 '<FieldAccess for sub_equation at ast.py:13757>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13757
--# expr-start 8752 '<FieldAccess for expr at ast.py:13757>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13757
      if Item_257.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_10 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Item_257.Node, E_Info => Item_257.Info);
--# end
--# expr-done 8752
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_11 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_10.Node, Env => Env, Origin => Origin, E_Info => Fld_10.Info);
--# end
--# expr-done 8753
         declare
            Item_To_Append : constant Logic_Equation := Fld_11;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2817;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 8754
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13756") else null)); 
--# expr-done 8755
And_Pred := Create_And (Fld_6, Logic_Boolean_Op, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13755") else null)); 
--# expr-done 8756
Scope_Result_2 := And_Pred;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2816;
   Result_Var_2 := Scope_Result_2;
else
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13759") else null)); 
   Result_Var_2 := False_Rel;
end if;
      Inc_Ref (Result_Var_2);
   If_Result_9 := Result_Var_2;
else
--# expr-start 8759 '<Not at ast.py:13762>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13762
--# expr-start 8758 '<IsNull at ast.py:13762>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13762
Var_Expr_3 := Block_Var;
if Var_Expr_3 /= No_Entity_Array_Type_Def then
   --# scope-start
--# expr-start 8757 '<FieldAccess for indices at ast.py:13762>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13762
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Create_Internal_Entity_Array_Indices (Node => Var_Expr_3.Node.Array_Type_Def_F_Indices, Info => Var_Expr_3.Info);
--# expr-done 8757
Scope_Result_3 := Fld_12;
   --# end
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := No_Entity_Array_Indices;
end if;
Is_Null_1 := Result_Var_3.Node = null; 
--# expr-done 8758
Not_Val := not (Is_Null_1); 
--# expr-done 8759
if Not_Val then
   --# expr-start 8817 '<Match at ast.py:13762>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13762
--# expr-start 8760 '<FieldAccess for suffix at ast.py:13762>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13762
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Create_Internal_Entity (Node => Ent.Node.Call_Expr_F_Suffix, Info => Ent.Info);
--# expr-done 8760
      if Fld_13.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld_13; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Assoc_List =>
   --# scope-start
--# scope-start
      Cast_Result_9 := Create_Internal_Entity_Assoc_List
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_9; 
--# bind _ Match_0
--# expr-start 8784 '<Or at ast.py:13763>' Or_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13763
--# expr-start 8763 '<Cast to Name at ast.py:13765>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13765
--# expr-start 8761 '<FieldAccess for params at ast.py:13765>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13765
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.params
Fld_14 := Libadalang.Implementation.Call_Expr_P_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8761
Fld_15 := Fld_14.Node;
Get_Result := Get (Fld_15, 0, True); 
Is_Null_2 := Get_Result = null; 
if Is_Null_2 then
   If_Result_2 := No_Entity_Basic_Assoc;
else
Fld_16 := Fld_14.Info;
As_Entity_1 := (Info => Fld_16, Node => Get_Result); 
   If_Result_2 := As_Entity_1;
end if;
Var_Expr_5 := If_Result_2;
if Var_Expr_5 /= No_Entity_Basic_Assoc then
   --# scope-start
--# expr-start 8762 '<FieldAccess for expr at ast.py:13765>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13765
      if Var_Expr_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_17 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Var_Expr_5.Node, E_Info => Var_Expr_5.Info);
--# end
--# expr-done 8762
Scope_Result_4 := Fld_17;
   --# end
   Result_Var_4 := Scope_Result_4;
else
   Result_Var_4 := No_Entity_Expr;
end if;
   if Result_Var_4.Node = null
      or else Result_Var_4.Node.Kind in Ada_Name
   then
      Cast_Result_2 := Create_Internal_Entity_Name
        (Node => Result_Var_4.Node,
         Info => Result_Var_4.Info);
   else
         Cast_Result_2 := No_Entity_Name;
   end if;
--# expr-done 8763
Var_Expr_4 := Cast_Result_2;
if Var_Expr_4 /= No_Entity_Name then
   --# scope-start
   --# bind name Var_Expr_4
--# expr-start 8770 '<If at ast.py:13766>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13766
--# expr-start 8765 '<IsNull at ast.py:13767>' Is_Null_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13767
--# expr-start 8764 '<FieldAccess for name_designated_type at ast.py:13767>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13767
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_20 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Var_Expr_4.Node, E_Info => Var_Expr_4.Info);
--# end
--# expr-done 8764
Is_Null_3 := Fld_20.Node = null; 
--# expr-done 8765
if Is_Null_3 then
   False_Rel_1 := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13768") else null)); 
   If_Result_3 := False_Rel_1;
else
   --# expr-start 8769 '<And at ast.py:13769>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13769
--# expr-start 8766 '<FieldAccess for xref_no_overloading at ast.py:13769>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13769
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_18 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Var_Expr_4.Node, Env => Env, Origin => Origin, E_Info => Var_Expr_4.Info);
--# end
--# expr-done 8766
--# expr-start 8768 '<Bind at ast.py:13770>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13770
--# expr-start 8767 '<FieldAccess for type_var at ast.py:13770>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13770
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8767
Fld_19.Value := No_Entity;
Entity_Vars.Reset (Fld_19);
Cast_Expr := Block_Var_1; 
      Cast_Result_3 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_19, Cast_Result_3, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13770") else null)); 
--# expr-done 8768
And_Pred_1 := Create_And (Fld_18, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13769") else null)); 
--# expr-done 8769
   If_Result_3 := And_Pred_1;
end if;
      Inc_Ref (If_Result_3);
--# expr-done 8770
Scope_Result_5 := If_Result_3;
   Inc_Ref (Scope_Result_5);
   --# end
      Finalizer_Scope_2821;
   Result_Var_5 := Scope_Result_5;
else
   False_Rel_2 := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13772") else null)); 
   Result_Var_5 := False_Rel_2;
end if;
      Inc_Ref (Result_Var_5);
--# expr-start 8783 '<And at ast.py:13776>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13776
--# expr-start 8779 '<LogicAll at ast.py:13776>' Logic_Boolean_Op_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13776
--# expr-start 8771 '<FieldAccess for params at ast.py:13776>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13776
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.params
Fld_21 := Libadalang.Implementation.Call_Expr_P_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8771
Var_Expr_6 := Fld_21;
if Var_Expr_6 /= No_Entity_Assoc_List then
   --# scope-start
--# expr-start 8778 '<Map at ast.py:13776>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13776
Saved := Var_Expr_6; 
Fld_22 := Saved.Node;
      declare
         Map_Result_1_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_22 = null then
      Map_Result_1 := Create_Logic_Equation_Array (0);
      else
         Index_259 := 0;
      declare
         Collection : constant Bare_Assoc_List := Fld_22;
      begin
         for Untyped_Item_258 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_258 := Untyped_Item_258; 
Is_Null_4 := Bare_Item_258 = null; 
if Is_Null_4 then
   If_Result_4 := No_Entity_Basic_Assoc;
else
Fld_23 := Saved.Info;
As_Entity_2 := (Info => Fld_23, Node => Bare_Item_258); 
   If_Result_4 := As_Entity_2;
end if;
                  Item_258 := If_Result_4; 
   --# scope-start
               --# bind pa Item_258
               --# bind i Index_259
      --# expr-start 8777 '<And at ast.py:13778>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13778
--# expr-start 8773 '<FieldAccess for sub_equation at ast.py:13778>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13778
--# expr-start 8772 '<FieldAccess for expr at ast.py:13778>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13778
      if Item_258.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_24 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Item_258.Node, E_Info => Item_258.Info);
--# end
--# expr-done 8772
      if Fld_24.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_25 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_24.Node, Env => Env, Origin => Origin, E_Info => Fld_24.Info);
--# end
--# expr-done 8773
--# expr-start 8776 '<FieldAccess for constrain_index_expr at ast.py:13779>' Fld_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13779
--# expr-start 8774 '<FieldAccess for indices at ast.py:13779>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13779
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_26 := Create_Internal_Entity_Array_Indices (Node => Block_Var.Node.Array_Type_Def_F_Indices, Info => Block_Var.Info);
--# expr-done 8774
      if Fld_26.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8775 '<FieldAccess for expr at ast.py:13779>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13779
      if Item_258.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_27 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Item_258.Node, E_Info => Item_258.Info);
--# end
--# expr-done 8775
--# property-call-start '[dispatcher]ArrayIndices.constrain_index_expr'
Fld_28 := Libadalang.Implementation.Dispatcher_Array_Indices_P_Constrain_Index_Expr (Node => Fld_26.Node, Index_Expr => Fld_27, Dim => Index_259, Origin => Origin, E_Info => Fld_26.Info);
--# end
--# expr-done 8776
And_Pred_2 := Create_And (Fld_25, Fld_28, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13778") else null)); 
--# expr-done 8777
         declare
            Item_To_Append : constant Logic_Equation := And_Pred_2;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2823;
               Index_259 := Index_259 + 1;
         end loop;
      end;
      end if;
         Map_Result_1 := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_1_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 8778
Scope_Result_6 := Map_Result_1;
   Inc_Ref (Scope_Result_6);
   --# end
      Finalizer_Scope_2822;
   Result_Var_6 := Scope_Result_6;
else
   Result_Var_6 := No_Logic_Equation_Array_Type;
end if;
      Inc_Ref (Result_Var_6);
Logic_Boolean_Op_1 := Solver.Create_All (Relation_Array (Result_Var_6.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13776") else null)); 
--# expr-done 8779
--# expr-start 8782 '<Bind at ast.py:13781>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13781
--# expr-start 8780 '<FieldAccess for type_var at ast.py:13781>' Fld_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13781
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_29 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8780
Fld_29.Value := No_Entity;
Entity_Vars.Reset (Fld_29);
--# expr-start 8781 '<FieldAccess for comp_type at ast.py:13781>' Fld_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13781
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ArrayTypeDef.comp_type
Fld_30 := Libadalang.Implementation.Array_Type_Def_P_Comp_Type (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 8781
      Cast_Result_4 := Create_Internal_Entity
        (Node => Fld_30.Node,
         Info => Fld_30.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_29, Cast_Result_4, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13781") else null)); 
--# expr-done 8782
And_Pred_3 := Create_And (Logic_Boolean_Op_1, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13776") else null)); 
--# expr-done 8783
Or_Pred := Create_Or (Result_Var_5, And_Pred_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13763") else null)); 
--# expr-done 8784
Let_Result := Or_Pred; Inc_Ref (Let_Result);
--# end
Scope_Result_7 := Let_Result;
   Inc_Ref (Scope_Result_7);
   --# end
      Finalizer_Scope_2819;
            Match_Result := Scope_Result_7; Inc_Ref (Match_Result);
         when Ada_Bin_Op .. Ada_Relation_Op =>
   --# scope-start
--# scope-start
      Cast_Result_10 := Create_Internal_Entity_Bin_Op
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_10; 
--# bind bo Match_1
--# expr-start 8805 '<And at ast.py:13786>' And_Pred_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13786
--# expr-start 8802 '<And at ast.py:13786>' And_Pred_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13786
--# expr-start 8799 '<And at ast.py:13786>' And_Pred_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13786
--# expr-start 8796 '<And at ast.py:13786>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13786
--# expr-start 8791 '<And at ast.py:13786>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13786
--# expr-start 8787 '<FieldAccess for constrain_index_expr at ast.py:13786>' Fld_33 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13786
--# expr-start 8785 '<FieldAccess for indices at ast.py:13786>' Fld_31 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13786
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_31 := Create_Internal_Entity_Array_Indices (Node => Block_Var.Node.Array_Type_Def_F_Indices, Info => Block_Var.Info);
--# expr-done 8785
      if Fld_31.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8786 '<FieldAccess for left at ast.py:13786>' Fld_32 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13786
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_32 := Create_Internal_Entity_Expr (Node => Match_1.Node.Bin_Op_F_Left, Info => Match_1.Info);
--# expr-done 8786
--# property-call-start '[dispatcher]ArrayIndices.constrain_index_expr'
Fld_33 := Libadalang.Implementation.Dispatcher_Array_Indices_P_Constrain_Index_Expr (Node => Fld_31.Node, Index_Expr => Fld_32, Dim => 0, Origin => Origin, E_Info => Fld_31.Info);
--# end
--# expr-done 8787
--# expr-start 8790 '<FieldAccess for constrain_index_expr at ast.py:13787>' Fld_36 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13787
--# expr-start 8788 '<FieldAccess for indices at ast.py:13787>' Fld_34 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13787
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_34 := Create_Internal_Entity_Array_Indices (Node => Block_Var.Node.Array_Type_Def_F_Indices, Info => Block_Var.Info);
--# expr-done 8788
      if Fld_34.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8789 '<FieldAccess for right at ast.py:13787>' Fld_35 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13787
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_35 := Create_Internal_Entity_Expr (Node => Match_1.Node.Bin_Op_F_Right, Info => Match_1.Info);
--# expr-done 8789
--# property-call-start '[dispatcher]ArrayIndices.constrain_index_expr'
Fld_36 := Libadalang.Implementation.Dispatcher_Array_Indices_P_Constrain_Index_Expr (Node => Fld_34.Node, Index_Expr => Fld_35, Dim => 0, Origin => Origin, E_Info => Fld_34.Info);
--# end
--# expr-done 8790
And_Pred_4 := Create_And (Fld_33, Fld_36, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13786") else null)); 
--# expr-done 8791
--# expr-start 8795 '<Bind at ast.py:13788>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13788
--# expr-start 8792 '<FieldAccess for expected_type_var at ast.py:13788>' Fld_37 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13788
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_37 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Match_1.Node);
--# expr-done 8792
Fld_37.Value := No_Entity;
Entity_Vars.Reset (Fld_37);
--# expr-start 8794 '<FieldAccess for expected_type_var at ast.py:13788>' Fld_39 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13788
--# expr-start 8793 '<FieldAccess for right at ast.py:13788>' Fld_38 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13788
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_38 := Create_Internal_Entity_Expr (Node => Match_1.Node.Bin_Op_F_Right, Info => Match_1.Info);
--# expr-done 8793
      if Fld_38.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_39 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_38.Node);
--# expr-done 8794
Fld_39.Value := No_Entity;
Entity_Vars.Reset (Fld_39);
Bind_Result_2 := Solver.Create_Unify (Fld_37, Fld_39, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13788") else null)); 
--# expr-done 8795
And_Pred_5 := Create_And (And_Pred_4, Bind_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13786") else null)); 
--# expr-done 8796
--# expr-start 8798 '<Bind at ast.py:13789>' Bind_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13789
--# expr-start 8797 '<FieldAccess for type_var at ast.py:13789>' Fld_40 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13789
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_40 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8797
Fld_40.Value := No_Entity;
Entity_Vars.Reset (Fld_40);
Cast_Expr_1 := Block_Var_1; 
      Cast_Result_5 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_3 := Solver.Create_Assign (Fld_40, Cast_Result_5, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13789") else null)); 
--# expr-done 8798
And_Pred_6 := Create_And (And_Pred_5, Bind_Result_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13786") else null)); 
--# expr-done 8799
--# expr-start 8801 '<FieldAccess for sub_equation at ast.py:13790>' Fld_42 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13790
--# expr-start 8800 '<FieldAccess for left at ast.py:13790>' Fld_41 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13790
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_41 := Create_Internal_Entity_Expr (Node => Match_1.Node.Bin_Op_F_Left, Info => Match_1.Info);
--# expr-done 8800
      if Fld_41.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_42 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_41.Node, Env => Env, Origin => Origin, E_Info => Fld_41.Info);
--# end
--# expr-done 8801
And_Pred_7 := Create_And (And_Pred_6, Fld_42, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13786") else null)); 
--# expr-done 8802
--# expr-start 8804 '<FieldAccess for sub_equation at ast.py:13791>' Fld_44 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13791
--# expr-start 8803 '<FieldAccess for right at ast.py:13791>' Fld_43 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13791
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_43 := Create_Internal_Entity_Expr (Node => Match_1.Node.Bin_Op_F_Right, Info => Match_1.Info);
--# expr-done 8803
      if Fld_43.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_44 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_43.Node, Env => Env, Origin => Origin, E_Info => Fld_43.Info);
--# end
--# expr-done 8804
And_Pred_8 := Create_And (And_Pred_7, Fld_44, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13786") else null)); 
--# expr-done 8805
Let_Result_1 := And_Pred_8; Inc_Ref (Let_Result_1);
--# end
Scope_Result_8 := Let_Result_1;
   Inc_Ref (Scope_Result_8);
   --# end
      Finalizer_Scope_2824;
            Match_Result := Scope_Result_8; Inc_Ref (Match_Result);
         when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref =>
   --# scope-start
--# scope-start
      Cast_Result_11 := Create_Internal_Entity_Attribute_Ref
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_11; 
--# bind ar Match_2
--# expr-start 8812 '<And at ast.py:13795>' And_Pred_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13795
--# expr-start 8809 '<And at ast.py:13795>' And_Pred_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13795
--# expr-start 8806 '<FieldAccess for sub_equation at ast.py:13795>' Fld_45 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13795
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_45 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Match_2.Node, Env => Env, Origin => Origin, E_Info => Match_2.Info);
--# end
--# expr-done 8806
--# expr-start 8808 '<FieldAccess for constrain_index_expr at ast.py:13796>' Fld_47 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13796
--# expr-start 8807 '<FieldAccess for indices at ast.py:13796>' Fld_46 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13796
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_46 := Create_Internal_Entity_Array_Indices (Node => Block_Var.Node.Array_Type_Def_F_Indices, Info => Block_Var.Info);
--# expr-done 8807
      if Fld_46.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Cast_Expr_2 := Match_2; 
      Cast_Result_6 := Create_Internal_Entity_Expr
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
--# property-call-start '[dispatcher]ArrayIndices.constrain_index_expr'
Fld_47 := Libadalang.Implementation.Dispatcher_Array_Indices_P_Constrain_Index_Expr (Node => Fld_46.Node, Index_Expr => Cast_Result_6, Dim => 0, Origin => Origin, E_Info => Fld_46.Info);
--# end
--# expr-done 8808
And_Pred_9 := Create_And (Fld_45, Fld_47, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13795") else null)); 
--# expr-done 8809
--# expr-start 8811 '<Bind at ast.py:13797>' Bind_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13797
--# expr-start 8810 '<FieldAccess for type_var at ast.py:13797>' Fld_48 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13797
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_48 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8810
Fld_48.Value := No_Entity;
Entity_Vars.Reset (Fld_48);
Cast_Expr_3 := Block_Var_1; 
      Cast_Result_7 := Create_Internal_Entity
        (Node => Cast_Expr_3.Node,
         Info => Cast_Expr_3.Info);
Bind_Result_4 := Solver.Create_Assign (Fld_48, Cast_Result_7, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13797") else null)); 
--# expr-done 8811
And_Pred_10 := Create_And (And_Pred_9, Bind_Result_4, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13795") else null)); 
--# expr-done 8812
Let_Result_2 := And_Pred_10; Inc_Ref (Let_Result_2);
--# end
Scope_Result_9 := Let_Result_2;
   Inc_Ref (Scope_Result_9);
   --# end
      Finalizer_Scope_2825;
            Match_Result := Scope_Result_9; Inc_Ref (Match_Result);
         when Ada_Subtype_Indication .. Ada_Discrete_Subtype_Indication =>
   --# scope-start
--# scope-start
      Cast_Result_12 := Create_Internal_Entity_Subtype_Indication
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_12; 
--# bind st Match_3
--# expr-start 8816 '<And at ast.py:13801>' And_Pred_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13801
--# expr-start 8813 '<FieldAccess for sub_equation at ast.py:13801>' Fld_49 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13801
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_49 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Match_3.Node, Env => Env, Origin => Origin, E_Info => Match_3.Info);
--# end
--# expr-done 8813
--# expr-start 8815 '<Bind at ast.py:13802>' Bind_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13802
--# expr-start 8814 '<FieldAccess for type_var at ast.py:13802>' Fld_50 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13802
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_50 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8814
Fld_50.Value := No_Entity;
Entity_Vars.Reset (Fld_50);
Cast_Expr_4 := Block_Var_1; 
      Cast_Result_8 := Create_Internal_Entity
        (Node => Cast_Expr_4.Node,
         Info => Cast_Expr_4.Info);
Bind_Result_5 := Solver.Create_Assign (Fld_50, Cast_Result_8, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13802") else null)); 
--# expr-done 8815
And_Pred_11 := Create_And (Fld_49, Bind_Result_5, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13801") else null)); 
--# expr-done 8816
Let_Result_3 := And_Pred_11; Inc_Ref (Let_Result_3);
--# end
Scope_Result_10 := Let_Result_3;
   Inc_Ref (Scope_Result_10);
   --# end
      Finalizer_Scope_2826;
            Match_Result := Scope_Result_10; Inc_Ref (Match_Result);
         when Ada_Abort_Absent .. Ada_Base_Assoc_List | Ada_Basic_Decl_List .. Ada_Null_Record_Aggregate | Ada_Box_Expr .. Ada_Membership_Expr | Ada_Call_Expr .. Ada_Enum_Lit_Synth_Type_Expr | Ada_Synthetic_Type_Expr .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_13 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_4 := Cast_Result_13; 
--# bind _ Match_4
False_Rel_3 := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13804") else null)); 
Let_Result_4 := False_Rel_3; Inc_Ref (Let_Result_4);
--# end
Scope_Result_11 := Let_Result_4;
   Inc_Ref (Scope_Result_11);
   --# end
      Finalizer_Scope_2827;
            Match_Result := Scope_Result_11; Inc_Ref (Match_Result);
end case;
--# expr-done 8817
   If_Result_8 := Match_Result;
else
--# expr-start 8819 '<Not at ast.py:13808>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13808
--# expr-start 8818 '<IsNull at ast.py:13808>' Is_Null_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13808
Is_Null_5 := Typ.Node = null; 
--# expr-done 8818
Not_Val_1 := not (Is_Null_5); 
--# expr-done 8819
if Not_Val_1 then
   --# expr-start 8820 '<FieldAccess for has_ud_indexing at ast.py:13808>' Fld_51 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13808
      if Typ.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.has_ud_indexing'
Fld_51 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Has_Ud_Indexing (Node => Typ.Node, E_Info => Typ.Info);
--# end
--# expr-done 8820
   If_Result_5 := Fld_51;
else
   If_Result_5 := False;
end if;
if If_Result_5 then
   --# expr-start 8844 '<LogicAny at ast.py:13809>' Logic_Boolean_Op_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13809
--# expr-start 8843 '<Map at ast.py:13809>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13809
   --# expr-start 8823 '<Concat at ast.py:13809>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13809
--# expr-start 8821 '<FieldAccess for constant_indexing_fns at ast.py:13809>' Fld_52 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13809
      if Typ.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.constant_indexing_fns'
Fld_52 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Constant_Indexing_Fns (Node => Typ.Node, E_Info => Typ.Info);
--# end
--# expr-done 8821
--# expr-start 8822 '<FieldAccess for variable_indexing_fns at ast.py:13809>' Fld_53 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13809
      if Typ.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.variable_indexing_fns'
Fld_53 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Variable_Indexing_Fns (Node => Typ.Node, E_Info => Typ.Info);
--# end
--# expr-done 8822
Concat_Result := Concat (Fld_52, Fld_53); 
--# expr-done 8823
      declare
         Map_Result_2_Vec : Logic_Equation_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Basic_Decl_Array_Access := Concat_Result;
      begin
         for Item_260 of
               Collection.Items
         loop
   --# scope-start
               --# bind fn Item_260
      --# expr-start 8842 '<Let at ast.py:13810>' Let_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13810
--# scope-start
--# expr-start 8826 '<at at ast.py:13812>' Get_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13812
--# expr-start 8825 '<FieldAccess for unpacked_formal_params at ast.py:13812>' Fld_55 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13812
--# expr-start 8824 '<FieldAccess for subp_spec_or_null at ast.py:13812>' Fld_54 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13812
      if Item_260.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_54 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Item_260.Node, E_Info => Item_260.Info);
--# end
--# expr-done 8824
      if Fld_54.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld_55 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => Fld_54.Node, E_Info => Fld_54.Info);
--# end
--# expr-done 8825
Get_Result_1 := Get (Fld_55, 1, True); 
--# expr-done 8826
Formal := Get_Result_1; 
--# bind formal Formal
--# expr-start 8828 '<FieldAccess for return_type at ast.py:13813>' Fld_57 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13813
--# expr-start 8827 '<FieldAccess for subp_spec_or_null at ast.py:13813>' Fld_56 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13813
      if Item_260.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_56 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Item_260.Node, E_Info => Item_260.Info);
--# end
--# expr-done 8827
      if Fld_56.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.return_type
Fld_57 := Libadalang.Implementation.Base_Subp_Spec_P_Return_Type (Node => Fld_56.Node, Origin => Origin, E_Info => Fld_56.Info);
--# end
--# expr-done 8828
Ret_Type := Fld_57; 
--# bind ret_type Ret_Type
--# expr-start 8830 '<FieldAccess for expr at ast.py:13814>' Fld_61 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13814
--# expr-start 8829 '<FieldAccess for params at ast.py:13814>' Fld_58 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13814
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.params
Fld_58 := Libadalang.Implementation.Call_Expr_P_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8829
Fld_59 := Fld_58.Node;
Get_Result_2 := Get (Fld_59, 0, True); 
Is_Null_6 := Get_Result_2 = null; 
if Is_Null_6 then
   If_Result_6 := No_Entity_Basic_Assoc;
else
Fld_60 := Fld_58.Info;
As_Entity_3 := (Info => Fld_60, Node => Get_Result_2); 
   If_Result_6 := As_Entity_3;
end if;
      if If_Result_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_61 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result_6.Node, E_Info => If_Result_6.Info);
--# end
--# expr-done 8830
Param := Fld_61; 
--# bind param Param
--# expr-start 8841 '<And at ast.py:13816>' And_Pred_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13816
--# expr-start 8839 '<And at ast.py:13816>' And_Pred_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13816
--# expr-start 8837 '<And at ast.py:13816>' And_Pred_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13816
--# expr-start 8832 '<Bind at ast.py:13816>' Bind_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13816
--# expr-start 8831 '<FieldAccess for type_var at ast.py:13816>' Fld_62 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13816
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_62 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8831
Fld_62.Value := No_Entity;
Entity_Vars.Reset (Fld_62);
Cast_Expr_5 := Ret_Type; 
      Cast_Result_14 := Create_Internal_Entity
        (Node => Cast_Expr_5.Node,
         Info => Cast_Expr_5.Info);
Bind_Result_6 := Solver.Create_Assign (Fld_62, Cast_Result_14, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13816") else null)); 
--# expr-done 8832
--# expr-start 8836 '<Bind at ast.py:13817>' Bind_Result_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13817
--# expr-start 8833 '<FieldAccess for expected_type_var at ast.py:13817>' Fld_63 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13817
      if Param.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_63 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Param.Node);
--# expr-done 8833
Fld_63.Value := No_Entity;
Entity_Vars.Reset (Fld_63);
--# expr-start 8835 '<FieldAccess for formal_type at ast.py:13817>' Fld_65 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13817
--# expr-start 8834 '<FieldAccess for formal_decl at ast.py:13817>' Fld_64 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13817
      if Formal.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_64 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Formal.Node, E_Info => Formal.Info);
--# end
--# expr-done 8834
      if Fld_64.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamDecl.formal_type
Fld_65 := Libadalang.Implementation.Base_Formal_Param_Decl_P_Formal_Type (Node => Fld_64.Node, Origin => Origin, E_Info => Fld_64.Info);
--# end
--# expr-done 8835
      Cast_Result_15 := Create_Internal_Entity
        (Node => Fld_65.Node,
         Info => Fld_65.Info);
Bind_Result_7 := Solver.Create_Assign (Fld_63, Cast_Result_15, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13817") else null)); 
--# expr-done 8836
And_Pred_12 := Create_And (Bind_Result_6, Bind_Result_7, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13816") else null)); 
--# expr-done 8837
--# expr-start 8838 '<FieldAccess for sub_equation at ast.py:13818>' Fld_66 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13818
      if Param.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_66 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Param.Node, Env => Env, Origin => Origin, E_Info => Param.Info);
--# end
--# expr-done 8838
And_Pred_13 := Create_And (And_Pred_12, Fld_66, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13816") else null)); 
--# expr-done 8839
--# expr-start 8840 '<FieldAccess for matches_expected_type at ast.py:13819>' Fld_67 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13819
      if Param.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_67 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Param.Node, Origin => Origin);
--# end
--# expr-done 8840
And_Pred_14 := Create_And (And_Pred_13, Fld_67, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13816") else null)); 
--# expr-done 8841
Let_Result_5 := And_Pred_14; Inc_Ref (Let_Result_5);
--# end
--# expr-done 8842
         declare
            Item_To_Append : constant Logic_Equation := Let_Result_5;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2828;
         end loop;
      end;
         Map_Result_2 := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_2_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 8843
Logic_Boolean_Op_2 := Solver.Create_Any (Relation_Array (Map_Result_2.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13809") else null)); 
--# expr-done 8844
   If_Result_7 := Logic_Boolean_Op_2;
else
   False_Rel_4 := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13822") else null)); 
   If_Result_7 := False_Rel_4;
end if;
      Inc_Ref (If_Result_7);
   If_Result_8 := If_Result_7;
end if;
      Inc_Ref (If_Result_8);
   If_Result_9 := If_Result_8;
end if;
      Inc_Ref (If_Result_9);
--# expr-done 8845
Let_Result_6 := If_Result_9; Inc_Ref (Let_Result_6);
--# end
--# expr-done 8846
      Property_Result := Let_Result_6;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1009;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1009;
               Finalizer_Scope_2816;
               Finalizer_Scope_2817;
               Finalizer_Scope_2819;
               Finalizer_Scope_2821;
               Finalizer_Scope_2822;
               Finalizer_Scope_2823;
               Finalizer_Scope_2824;
               Finalizer_Scope_2825;
               Finalizer_Scope_2826;
               Finalizer_Scope_2827;
               Finalizer_Scope_2828;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Subscriptable_Type_Equation;
--# end
--# property-start CallExpr.subprogram_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13826
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Subprogram_Equation
  (Node : Bare_Call_Expr
      ; Subp_Spec : Internal_Entity_Base_Formal_Param_Holder
      ; Dottable_Subp : Boolean
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind subp_spec Subp_Spec
   --# bind dottable_subp Dottable_Subp
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1010 with Inline_Always;
            procedure Finalizer_Scope_2829 with Inline_Always;
            procedure Finalizer_Scope_2831 with Inline_Always;
      Var_Expr : Internal_Entity_Base_Formal_Param_Holder;
Var_Expr_1 : Internal_Entity_Base_Subp_Spec;
N : Internal_Entity_Base_Formal_Param_Decl;
Fld : Logic_Var;
Cast_Expr : Internal_Entity_Base_Formal_Param_Holder;
Cast_Result : Internal_Entity_Base_Subp_Spec;
Fld_1 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_2 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Base_Formal_Param_Holder;
Cast_Result_2 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_3 : Internal_Entity_Assoc_List;
Fld_4 : Internal_Param_Match_Array_Access;
Fld_5 : Internal_Entity_Defining_Name;
Fld_6 : Internal_Entity_Base_Formal_Param_Decl;
Fld_7 : Internal_Single_Actual;
Fld_8 : Internal_Entity_Basic_Assoc;
Fld_9 : Internal_Entity_Expr;
Fld_10 : Logic_Equation;
True_Rel : Logic_Equation;
Fld_11 : Internal_Single_Actual;
Fld_12 : Bare_Base_Id;
Fld_13 : Logic_Var;
Fld_14 : Internal_Entity_Defining_Name;
Fld_15 : Internal_Entity_Base_Formal_Param_Decl;
Fld_16 : Bare_Base_Formal_Param_Decl;
Cast_Result_3 : Bare_Ada_Node;
Fld_17 : Internal_Entity_Info;
Fld_18 : Env_Rebindings;
Fld_19 : Internal_Entity_Info;
Fld_20 : Boolean;
Fld_21 : Internal_Entity;
Let_Result : Internal_Entity;
Bind_Result_2 : Logic_Equation;
Fld_22 : Internal_Single_Actual;
Fld_23 : Bare_Base_Id;
Is_Null : Boolean;
If_Result : Logic_Equation;
And_Pred_1 : Logic_Equation;
False_Rel : Logic_Equation;
Fld_24 : Boolean;
If_Result_1 : Logic_Equation;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
And_Pred_2 : Logic_Equation;
Scope_Result_1 : Logic_Equation;
False_Rel_1 : Logic_Equation;
Result_Var_1 : Logic_Equation;
Let_Result_1 : Logic_Equation;
            procedure Finalizer_Scope_1010 is
            begin
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (False_Rel_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Let_Result_1);
            end Finalizer_Scope_1010;
            procedure Finalizer_Scope_2829 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (And_Pred_2);
            end Finalizer_Scope_2829;
            procedure Finalizer_Scope_2831 is
            begin
                     Dec_Ref (Fld_10);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (If_Result);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (False_Rel);
                     Dec_Ref (If_Result_1);
            end Finalizer_Scope_2831;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8885 '<Block at ???>' Let_Result_1 None
--# scope-start
Var_Expr := Subp_Spec;
if Var_Expr /= No_Entity_Base_Formal_Param_Holder then
   --# scope-start
   --# bind subp_spec Var_Expr
--# expr-start 8884 '<And at ast.py:13832>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13832
--# expr-start 8853 '<And at ast.py:13832>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13832
--# expr-start 8850 '<Bind at ast.py:13832>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13832
--# expr-start 8847 '<FieldAccess for type_var at ast.py:13832>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13832
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 8847
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
--# expr-start 8848 '<Cast to BaseSubpSpec at ast.py:13832>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13832
Cast_Expr := Var_Expr; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Base_Subp_Spec
   then
      Cast_Result := Create_Internal_Entity_Base_Subp_Spec
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Base_Subp_Spec;
   end if;
--# expr-done 8848
Var_Expr_1 := Cast_Result;
if Var_Expr_1 /= No_Entity_Base_Subp_Spec then
   --# scope-start
--# expr-start 8849 '<FieldAccess for return_type at ast.py:13832>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13832
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.return_type
Fld_1 := Libadalang.Implementation.Base_Subp_Spec_P_Return_Type (Node => Var_Expr_1.Node, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 8849
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
      Cast_Result_1 := Create_Internal_Entity
        (Node => Result_Var.Node,
         Info => Result_Var.Info);
Bind_Result := Solver.Create_Assign (Fld, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13832") else null)); 
--# expr-done 8850
--# expr-start 8852 '<Bind at ast.py:13836>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13836
--# expr-start 8851 '<FieldAccess for subp_spec_var at ast.py:13836>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13836
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.subp_spec_var'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_Subp_Spec_Var (Node => Self);
--# end
--# expr-done 8851
Fld_2.Value := No_Entity;
Entity_Vars.Reset (Fld_2);
Cast_Expr_1 := Var_Expr; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_2, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13836") else null)); 
--# expr-done 8852
And_Pred := Create_And (Bind_Result, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13832") else null)); 
--# expr-done 8853
--# expr-start 8883 '<LogicAll at ast.py:13840>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13840
--# expr-start 8882 '<Map at ast.py:13840>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13840
   --# expr-start 8855 '<FieldAccess for match_param_list at ast.py:13840>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13840
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8854 '<FieldAccess for params at ast.py:13841>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13841
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.params
Fld_3 := Libadalang.Implementation.Call_Expr_P_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8854
--# property-call-start BaseFormalParamHolder.match_param_list
Fld_4 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Match_Param_List (Node => Var_Expr.Node, Params => Fld_3, Is_Dottable_Subp => Dottable_Subp, Env => Env, E_Info => Var_Expr.Info);
--# end
--# expr-done 8855
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Param_Match_Array_Access := Fld_4;
      begin
         for Item_261 of
               Collection.Items
         loop
   --# scope-start
               --# bind pm Item_261
      --# expr-start 8881 '<If at ast.py:13843>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13843
--# expr-start 8856 '<FieldAccess for has_matched at ast.py:13844>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13844
Fld_24 := Item_261.Has_Matched;
--# expr-done 8856
if Fld_24 then
   --# expr-start 8880 '<And at ast.py:13845>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13845
--# expr-start 8862 '<FieldAccess for call_argument_equation at ast.py:13845>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13845
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8858 '<FieldAccess for formal_decl at ast.py:13846>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13846
--# expr-start 8857 '<FieldAccess for formal at ast.py:13846>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13846
Fld_5 := Item_261.Formal;
--# expr-done 8857
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_6 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Fld_5.Node, E_Info => Fld_5.Info);
--# end
--# expr-done 8858
--# expr-start 8861 '<FieldAccess for expr at ast.py:13847>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13847
--# expr-start 8860 '<FieldAccess for assoc at ast.py:13847>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13847
--# expr-start 8859 '<FieldAccess for actual at ast.py:13847>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13847
Fld_7 := Item_261.Actual;
--# expr-done 8859
Fld_8 := Fld_7.Assoc;
--# expr-done 8860
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_9 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_8.Node, E_Info => Fld_8.Info);
--# end
--# expr-done 8861
--# property-call-start BaseFormalParamHolder.call_argument_equation
Fld_10 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Call_Argument_Equation (Node => Var_Expr.Node, Param => Fld_6, Arg => Fld_9, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 8862
--# expr-start 8879 '<If at ast.py:13848>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13848
--# expr-start 8865 '<IsNull at ast.py:13851>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13851
--# expr-start 8864 '<FieldAccess for name at ast.py:13851>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13851
--# expr-start 8863 '<FieldAccess for actual at ast.py:13851>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13851
Fld_22 := Item_261.Actual;
--# expr-done 8863
Fld_23 := Fld_22.Name;
--# expr-done 8864
Is_Null := Fld_23 = null; 
--# expr-done 8865
if Is_Null then
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13852") else null)); 
   If_Result := True_Rel;
else
   --# expr-start 8878 '<Bind at ast.py:13853>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13853
--# expr-start 8868 '<FieldAccess for ref_var at ast.py:13854>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13854
--# expr-start 8867 '<FieldAccess for name at ast.py:13854>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13854
--# expr-start 8866 '<FieldAccess for actual at ast.py:13854>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13854
Fld_11 := Item_261.Actual;
--# expr-done 8866
Fld_12 := Fld_11.Name;
--# expr-done 8867
      if Fld_12 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_13 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_12);
--# end
--# expr-done 8868
Fld_13.Value := No_Entity;
Entity_Vars.Reset (Fld_13);
--# expr-start 8877 '<Let at ast.py:13855>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13855
--# scope-start
--# expr-start 8870 '<FieldAccess for formal_decl at ast.py:13855>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13855
--# expr-start 8869 '<FieldAccess for formal at ast.py:13855>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13855
Fld_14 := Item_261.Formal;
--# expr-done 8869
      if Fld_14.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_15 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Fld_14.Node, E_Info => Fld_14.Info);
--# end
--# expr-done 8870
N := Fld_15; 
--# bind n N
--# expr-start 8876 '<FieldAccess for entity_no_md at ast.py:13856>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13856
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8871 '<FieldAccess for node at ast.py:13857>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13857
Fld_16 := N.Node;
--# expr-done 8871
      Cast_Result_3 := Fld_16;
--# expr-start 8873 '<FieldAccess for rebindings at ast.py:13858>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13858
--# expr-start 8872 '<FieldAccess for info at ast.py:13858>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13858
Fld_17 := N.Info;
--# expr-done 8872
Fld_18 := Fld_17.Rebindings;
--# expr-done 8873
--# expr-start 8875 '<FieldAccess for from_rebound at ast.py:13859>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13859
--# expr-start 8874 '<FieldAccess for info at ast.py:13859>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13859
Fld_19 := N.Info;
--# expr-done 8874
Fld_20 := Fld_19.From_Rebound;
--# expr-done 8875
--# property-call-start AdaNode.entity_no_md
Fld_21 := Libadalang.Implementation.Ada_Node_P_Entity_No_Md (Node => Ent.Node, N => Cast_Result_3, Rebindings => Fld_18, From_Rebound => Fld_20);
--# end
--# expr-done 8876
Let_Result := Fld_21; 
--# end
--# expr-done 8877
Bind_Result_2 := Solver.Create_Assign (Fld_13, Let_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13853") else null)); 
--# expr-done 8878
   If_Result := Bind_Result_2;
end if;
      Inc_Ref (If_Result);
--# expr-done 8879
And_Pred_1 := Create_And (Fld_10, If_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13845") else null)); 
--# expr-done 8880
   If_Result_1 := And_Pred_1;
else
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13863") else null)); 
   If_Result_1 := False_Rel;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 8881
         declare
            Item_To_Append : constant Logic_Equation := If_Result_1;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2831;
         end loop;
      end;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 8882
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13840") else null)); 
--# expr-done 8883
And_Pred_2 := Create_And (And_Pred, Logic_Boolean_Op, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13832") else null)); 
--# expr-done 8884
Scope_Result_1 := And_Pred_2;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2829;
   Result_Var_1 := Scope_Result_1;
else
   False_Rel_1 := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13866") else null)); 
   Result_Var_1 := False_Rel_1;
end if;
      Inc_Ref (Result_Var_1);
Let_Result_1 := Result_Var_1; Inc_Ref (Let_Result_1);
--# end
--# expr-done 8885
      Property_Result := Let_Result_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1010;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1010;
               Finalizer_Scope_2829;
               Finalizer_Scope_2831;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Subprogram_Equation;
--# end
--# property-start CallExpr.check_for_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13870
pragma Warnings (Off, "is not referenced");
 function Call_Expr_P_Check_For_Type
  (Node : Bare_Call_Expr
      ; Typ : Internal_Entity_Base_Type_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Call_Expr := Bare_Call_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Expr :=
     Internal_Entity_Call_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind typ Typ
   --# bind env Env
   --# bind origin Origin
   Property_Result : Boolean;
      Block_Var : Internal_Entity_Array_Type_Def;
Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Entity_Base_Type_Decl;
Var_Expr_2 : Internal_Entity_Array_Type_Def;
Match_0 : Bare_Assoc_List;
Match_1 : Bare_Bin_Op;
Match_2 : Bare_Subtype_Indication;
Match_3 : Bare_Attribute_Ref;
Match_4 : Bare_Ada_Node;
Var_Expr_3 : Internal_Entity_Access_To_Subp_Def;
Var_Expr_4 : Bare_Assoc_List;
Var_Expr_5 : Internal_Entity_Call_Expr;
Fld : Internal_Entity_Array_Type_Def;
Scope_Result : Internal_Entity_Array_Type_Def;
Result_Var : Internal_Entity_Array_Type_Def;
Fld_1 : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_2 : Bare_Ada_Node;
Fld_3 : Integer;
Len : Integer;
Is_Equal : Boolean;
Fld_4 : Integer;
Is_Equal_1 : Boolean;
Fld_5 : Integer;
Is_Equal_2 : Boolean;
Fld_6 : Integer;
Is_Equal_3 : Boolean;
Match_Prefix : Bare_Ada_Node;
Cast_Result : Bare_Assoc_List;
Let_Result : Boolean;
Scope_Result_1 : Boolean;
Cast_Result_1 : Bare_Bin_Op;
Let_Result_1 : Boolean;
Scope_Result_2 : Boolean;
Cast_Result_2 : Bare_Subtype_Indication;
Let_Result_2 : Boolean;
Scope_Result_3 : Boolean;
Cast_Result_3 : Bare_Attribute_Ref;
Let_Result_3 : Boolean;
Scope_Result_4 : Boolean;
Cast_Result_4 : Bare_Ada_Node;
Let_Result_4 : Boolean;
Scope_Result_5 : Boolean;
Match_Result : Boolean;
Scope_Result_6 : Boolean;
Result_Var_1 : Boolean;
Fld_7 : Internal_Entity_Access_Def;
Cast_Result_5 : Internal_Entity_Access_To_Subp_Def;
Fld_8 : Internal_Entity_Subp_Spec;
Fld_9 : Internal_Entity_Assoc_List;
Fld_10 : Boolean;
Scope_Result_7 : Boolean;
Result_Var_2 : Boolean;
If_Result : Boolean;
Fld_11 : Boolean;
Fld_12 : Bare_Ada_Node;
Cast_Result_6 : Bare_Assoc_List;
Len_1 : Integer;
Is_Equal_4 : Boolean;
Scope_Result_8 : Boolean;
Result_Var_3 : Boolean;
If_Result_1 : Boolean;
If_Result_2 : Boolean;
Fld_13 : Internal_Entity;
Cast_Result_7 : Internal_Entity_Call_Expr;
Fld_14 : Internal_Entity_Base_Type_Decl;
Fld_15 : Boolean;
If_Result_3 : Internal_Entity_Base_Type_Decl;
Fld_16 : Boolean;
Scope_Result_9 : Boolean;
Result_Var_4 : Boolean;
If_Result_4 : Boolean;
Scope_Result_10 : Boolean;
Result_Var_5 : Boolean;
Dyn_Var_Bind_Result : Boolean;
Let_Result_5 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8916 '<Block at ???>' Let_Result_5 None
--# scope-start
Var_Expr := Typ;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind t Var_Expr
--# expr-start 8886 '<FieldAccess for array_def_with_deref at ast.py:13881>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13881
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.array_def_with_deref
Fld := Libadalang.Implementation.Base_Type_Decl_P_Array_Def_With_Deref (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 8886
Scope_Result := Fld;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Array_Type_Def;
end if;
Block_Var := Result_Var; 
--# bind atd Block_Var
--# expr-start 8915 '<bind at ast.py:13883>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13883
--# expr-start 8887 '<FieldAccess for origin_node at ast.py:13883>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13883
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_1 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 8887
Bound_Origin := Fld_1; 
Var_Expr_1 := Typ;
if Var_Expr_1 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind typ Var_Expr_1
Var_Expr_2 := Block_Var;
if Var_Expr_2 /= No_Entity_Array_Type_Def then
   --# scope-start
   --# bind _ Var_Expr_2
--# expr-start 8898 '<Match at ast.py:13886>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13886
--# expr-start 8888 '<FieldAccess for suffix at ast.py:13886>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13886
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Call_Expr_F_Suffix;
--# expr-done 8888
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld_2; 
case Ada_Ada_Node (Match_Prefix.Kind) is
         when Ada_Assoc_List =>
   --# scope-start
--# scope-start
      Cast_Result := Match_Prefix;
Match_0 := Cast_Result; 
--# bind al Match_0
--# expr-start 8891 '<Eq at ast.py:13888>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13888
--# expr-start 8889 '<FieldAccess for array_ndims at ast.py:13888>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13888
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.array_ndims'
Fld_3 := Libadalang.Implementation.Dispatcher_Type_Def_P_Array_Ndims (Node => Block_Var.Node, Origin => Bound_Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 8889
--# expr-start 8890 '<length at ast.py:13888>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13888
Len := Length (Match_0); 
--# expr-done 8890
Is_Equal := Fld_3 = Len; 
--# expr-done 8891
Let_Result := Is_Equal; 
--# end
Scope_Result_1 := Let_Result;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Bin_Op .. Ada_Relation_Op =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Match_Prefix;
Match_1 := Cast_Result_1; 
--# bind _ Match_1
--# expr-start 8893 '<Eq at ast.py:13891>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13891
--# expr-start 8892 '<FieldAccess for array_ndims at ast.py:13891>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13891
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.array_ndims'
Fld_4 := Libadalang.Implementation.Dispatcher_Type_Def_P_Array_Ndims (Node => Block_Var.Node, Origin => Bound_Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 8892
Is_Equal_1 := Fld_4 = 1; 
--# expr-done 8893
Let_Result_1 := Is_Equal_1; 
--# end
Scope_Result_2 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Subtype_Indication .. Ada_Discrete_Subtype_Indication =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Match_Prefix;
Match_2 := Cast_Result_2; 
--# bind _ Match_2
--# expr-start 8895 '<Eq at ast.py:13892>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13892
--# expr-start 8894 '<FieldAccess for array_ndims at ast.py:13892>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13892
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.array_ndims'
Fld_5 := Libadalang.Implementation.Dispatcher_Type_Def_P_Array_Ndims (Node => Block_Var.Node, Origin => Bound_Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 8894
Is_Equal_2 := Fld_5 = 1; 
--# expr-done 8895
Let_Result_2 := Is_Equal_2; 
--# end
Scope_Result_3 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_3; 
         when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Match_Prefix;
Match_3 := Cast_Result_3; 
--# bind _ Match_3
--# expr-start 8897 '<Eq at ast.py:13893>' Is_Equal_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13893
--# expr-start 8896 '<FieldAccess for array_ndims at ast.py:13893>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13893
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.array_ndims'
Fld_6 := Libadalang.Implementation.Dispatcher_Type_Def_P_Array_Ndims (Node => Block_Var.Node, Origin => Bound_Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 8896
Is_Equal_3 := Fld_6 = 1; 
--# expr-done 8897
Let_Result_3 := Is_Equal_3; 
--# end
Scope_Result_4 := Let_Result_3;
   --# end
            Match_Result := Scope_Result_4; 
         when Ada_Abort_Absent .. Ada_Base_Assoc_List | Ada_Basic_Decl_List .. Ada_Null_Record_Aggregate | Ada_Box_Expr .. Ada_Membership_Expr | Ada_Call_Expr .. Ada_Enum_Lit_Synth_Type_Expr | Ada_Synthetic_Type_Expr .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Match_Prefix;
Match_4 := Cast_Result_4; 
--# bind _ Match_4
Let_Result_4 := False; 
--# end
Scope_Result_5 := Let_Result_4;
   --# end
            Match_Result := Scope_Result_5; 
end case;
--# expr-done 8898
Scope_Result_6 := Match_Result;
   --# end
   Result_Var_1 := Scope_Result_6;
else
   Result_Var_1 := False;
end if;
if Result_Var_1 then
   If_Result := True;
else
--# expr-start 8900 '<Cast to AccessToSubpDef at ast.py:13899>' Cast_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13899
--# expr-start 8899 '<FieldAccess for access_def at ast.py:13899>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13899
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.access_def'
Fld_7 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Access_Def (Node => Var_Expr_1.Node, Origin => Bound_Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 8899
   if Fld_7.Node = null
      or else Fld_7.Node.Kind in Ada_Access_To_Subp_Def_Range
   then
      Cast_Result_5 := Create_Internal_Entity_Access_To_Subp_Def
        (Node => Fld_7.Node,
         Info => Fld_7.Info);
   else
         Cast_Result_5 := No_Entity_Access_To_Subp_Def;
   end if;
--# expr-done 8900
Var_Expr_3 := Cast_Result_5;
if Var_Expr_3 /= No_Entity_Access_To_Subp_Def then
   --# scope-start
   --# bind sa Var_Expr_3
--# expr-start 8903 '<FieldAccess for is_matching_param_list at ast.py:13901>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13901
--# expr-start 8901 '<FieldAccess for subp_spec at ast.py:13901>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13901
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Subp_Spec (Node => Var_Expr_3.Node.Access_To_Subp_Def_F_Subp_Spec, Info => Var_Expr_3.Info);
--# expr-done 8901
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8902 '<FieldAccess for params at ast.py:13901>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13901
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.params
Fld_9 := Libadalang.Implementation.Call_Expr_P_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8902
--# property-call-start BaseFormalParamHolder.is_matching_param_list
Fld_10 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Is_Matching_Param_List (Node => Fld_8.Node, Params => Fld_9, Is_Dottable_Subp => False, Env => Env);
--# end
--# expr-done 8903
Scope_Result_7 := Fld_10;
   --# end
   Result_Var_2 := Scope_Result_7;
else
   Result_Var_2 := False;
end if;
   If_Result := Result_Var_2;
end if;
if If_Result then
   If_Result_2 := True;
else
--# expr-start 8904 '<FieldAccess for has_ud_indexing at ast.py:13905>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13905
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.has_ud_indexing'
Fld_11 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Has_Ud_Indexing (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 8904
if Fld_11 then
--# expr-start 8906 '<Cast to AssocList at ast.py:13906>' Cast_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13906
--# expr-start 8905 '<FieldAccess for suffix at ast.py:13906>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13906
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Self.Call_Expr_F_Suffix;
--# expr-done 8905
   if Fld_12 = null
      or else Fld_12.Kind in Ada_Assoc_List_Range
   then
      Cast_Result_6 := Fld_12;
   else
         Cast_Result_6 := No_Bare_Ada_Node;
   end if;
--# expr-done 8906
Var_Expr_4 := Cast_Result_6;
if Var_Expr_4 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind al Var_Expr_4
--# expr-start 8908 '<Eq at ast.py:13906>' Is_Equal_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13906
--# expr-start 8907 '<length at ast.py:13906>' Len_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13906
Len_1 := Length (Var_Expr_4); 
--# expr-done 8907
Is_Equal_4 := Len_1 = 1; 
--# expr-done 8908
Scope_Result_8 := Is_Equal_4;
   --# end
   Result_Var_3 := Scope_Result_8;
else
   Result_Var_3 := False;
end if;
   If_Result_1 := Result_Var_3;
else
   If_Result_1 := False;
end if;
   If_Result_2 := If_Result_1;
end if;
if If_Result_2 then
--# expr-start 8910 '<Cast to CallExpr at ast.py:13909>' Cast_Result_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13909
--# expr-start 8909 '<parent at ast.py:13909>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13909
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 8909
   if Fld_13.Node = null
      or else Fld_13.Node.Kind in Ada_Call_Expr_Range
   then
      Cast_Result_7 := Create_Internal_Entity_Call_Expr
        (Node => Fld_13.Node,
         Info => Fld_13.Info);
   else
         Cast_Result_7 := No_Entity_Call_Expr;
   end if;
--# expr-done 8910
Var_Expr_5 := Cast_Result_7;
if Var_Expr_5 /= No_Entity_Call_Expr then
   --# scope-start
   --# bind ce Var_Expr_5
--# expr-start 8914 '<FieldAccess for check_for_type at ast.py:13915>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13915
      if Var_Expr_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8913 '<If at ast.py:13915>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13915
--# expr-start 8911 '<FieldAccess for check_array_slice at ast.py:13916>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13916
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.check_array_slice
Fld_15 := Libadalang.Implementation.Call_Expr_P_Check_Array_Slice (Node => Ent.Node, Typ => Var_Expr_1, Origin => Bound_Origin, E_Info => Ent.Info);
--# end
--# expr-done 8911
if Fld_15 then
   If_Result_3 := Var_Expr_1;
else
   --# expr-start 8912 '<FieldAccess for comp_type at ast.py:13918>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13918
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.comp_type
Fld_14 := Libadalang.Implementation.Base_Type_Decl_P_Comp_Type (Node => Var_Expr_1.Node, Is_Subscript => True, Origin => Bound_Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 8912
   If_Result_3 := Fld_14;
end if;
--# expr-done 8913
--# property-call-start CallExpr.check_for_type
Fld_16 := Libadalang.Implementation.Call_Expr_P_Check_For_Type (Node => Var_Expr_5.Node, Typ => If_Result_3, Env => Env, Origin => Bound_Origin, E_Info => Var_Expr_5.Info);
--# end
--# expr-done 8914
Scope_Result_9 := Fld_16;
   --# end
   Result_Var_4 := Scope_Result_9;
else
   Result_Var_4 := True;
end if;
   If_Result_4 := Result_Var_4;
else
   If_Result_4 := False;
end if;
Scope_Result_10 := If_Result_4;
   --# end
   Result_Var_5 := Scope_Result_10;
else
   Result_Var_5 := False;
end if;
Dyn_Var_Bind_Result := Result_Var_5; 
--# expr-done 8915
Let_Result_5 := Dyn_Var_Bind_Result; 
--# end
--# expr-done 8916
      Property_Result := Let_Result_5;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Expr_P_Check_For_Type;
--# end
   --
   --  Primitives for Bare_Defining_Name
   --
      procedure Initialize_Fields_For_Defining_Name
        (Self : Bare_Defining_Name
         ; Defining_Name_F_Name : Bare_Name
        ) is
      begin
            Initialize_Fields_For_Name
              (Self);
            Self.Defining_Name_F_Name := Defining_Name_F_Name;
      end Initialize_Fields_For_Defining_Name;
   function Defining_Name_F_Name
     (Node : Bare_Defining_Name) return Bare_Name
   is
   begin
         return Node.Defining_Name_F_Name;
   end;
--# property-start DefiningName.parent_scope /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14831
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Parent_Scope
  (Node : Bare_Defining_Name
      ; Env : Lexical_Env
  )
   return Lexical_Env
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1066 with Inline_Always;
      Fld : Bare_Name;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1066 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1066;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8918 '<FieldAccess for parent_scope at ast.py:14831>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14831
--# expr-start 8917 '<FieldAccess for name at ast.py:14831>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14831
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Defining_Name_F_Name;
--# expr-done 8917
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.parent_scope'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Parent_Scope (Node => Fld, Env => Env);
--# end
--# expr-done 8918
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1066;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1066;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Parent_Scope;
--# end
--# property-start DefiningName.scope /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14832
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Scope
  (Node : Bare_Defining_Name
      ; Env : Lexical_Env
  )
   return Lexical_Env
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1067 with Inline_Always;
      Fld : Bare_Name;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1067 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1067;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8920 '<FieldAccess for scope at ast.py:14832>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14832
--# expr-start 8919 '<FieldAccess for name at ast.py:14832>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14832
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Defining_Name_F_Name;
--# expr-done 8919
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.scope'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Scope (Node => Fld, Env => Env);
--# end
--# expr-done 8920
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1067;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1067;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Scope;
--# end
--# property-start DefiningName.relative_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14833
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Relative_Name
  (Node : Bare_Defining_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Single_Tok_Node
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Single_Tok_Node;
      Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Single_Tok_Node;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 8922 '<FieldAccess for relative_name at ast.py:14833>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14833
--# expr-start 8921 '<FieldAccess for name at ast.py:14833>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14833
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Defining_Name_F_Name, Info => Ent.Info);
--# expr-done 8921
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.relative_name'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Relative_Name (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 8922
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Relative_Name;
--# end
--# property-start DefiningName.ref_var /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14834
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Ref_Var
  (Node : Bare_Defining_Name
  )
   return Logic_Var
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Logic_Var;
      Fld : Bare_Name;
Fld_1 : Logic_Var;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 8924 '<FieldAccess for ref_var at ast.py:14834>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14834
--# expr-start 8923 '<FieldAccess for name at ast.py:14834>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14834
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Defining_Name_F_Name;
--# expr-done 8923
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld);
--# end
--# expr-done 8924
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Ref_Var;
--# end
--# property-start DefiningName.env_elements_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14835
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Env_Elements_Impl
  (Node : Bare_Defining_Name
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   Property_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1070 with Inline_Always;
      Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1070 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1070;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8926 '<FieldAccess for env_elements_impl at ast.py:14835>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14835
--# expr-start 8925 '<FieldAccess for name at ast.py:14835>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14835
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Defining_Name_F_Name, Info => Ent.Info);
--# expr-done 8925
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.env_elements_impl'
Fld_1 := Libadalang.Implementation.Dispatcher_Expr_P_Env_Elements_Impl (Node => Fld.Node, Env => Env, E_Info => Fld.Info);
--# end
--# expr-done 8926
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1070;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1070;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Env_Elements_Impl;
--# end
--# property-start DefiningName.formal_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14838
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Formal_Decl
  (Node : Bare_Defining_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Formal_Param_Decl;
            procedure Finalizer_Scope_1071 with Inline_Always;
      Fld : Internal_Entity_Array_Access;
Is_A : Boolean;
Find_Result : Internal_Entity;
Cast_Result : Internal_Entity_Base_Formal_Param_Decl;
Let_Result : Internal_Entity_Base_Formal_Param_Decl;
            procedure Finalizer_Scope_1071 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1071;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 8931 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8930 '<Cast to BaseFormalParamDecl at ast.py:14843>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14843
--# expr-start 8929 '<Find at ast.py:14843>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14843
   --# expr-start 8927 '<FieldAccess for parents at ast.py:14843>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14843
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parents (Node => Ent.Node, With_Self => True, E_Info => Ent.Info);
--# expr-done 8927
      Find_Result := No_Entity;
      declare
         Collection : constant Internal_Entity_Array_Access := Fld;
      begin
         for Item_262 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_262
      --# expr-start 8928 '<IsA BaseFormalParamDecl at ast.py:14844>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14844
Is_A := Item_262.Node /= null 
and then Item_262.Node.Kind in Ada_Base_Formal_Param_Decl; 
--# expr-done 8928
      if Is_A then
         Find_Result := Item_262;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 8929
   if Find_Result.Node = null
      or else Find_Result.Node.Kind in Ada_Base_Formal_Param_Decl
   then
      Cast_Result := Create_Internal_Entity_Base_Formal_Param_Decl
        (Node => Find_Result.Node,
         Info => Find_Result.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 8930
Let_Result := Cast_Result; 
--# end
--# expr-done 8931
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1071;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1071;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Formal_Decl;
--# end
--# property-start DefiningName.canonical_fully_qualified_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14848
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Canonical_Fully_Qualified_Name
  (Node : Bare_Defining_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return String_Type
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : String_Type;
            procedure Finalizer_Scope_1072 with Inline_Always;
      Fld : Internal_Entity_Basic_Decl;
Fld_1 : String_Type;
Let_Result : String_Type;
            procedure Finalizer_Scope_1072 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1072;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8934 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8933 '<FieldAccess for canonical_fully_qualified_name_impl at ast.py:14853>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14853
--# expr-start 8932 '<FieldAccess for basic_decl at ast.py:14853>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14853
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8932
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.canonical_fully_qualified_name_impl
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Canonical_Fully_Qualified_Name_Impl (Node => Fld.Node, Dn => Ent, E_Info => Fld.Info);
--# end
--# expr-done 8933
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 8934
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1072;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1072;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Canonical_Fully_Qualified_Name;
--# end
--# property-start DefiningName.unique_identifying_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14856
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Unique_Identifying_Name
  (Node : Bare_Defining_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return String_Type
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : String_Type;
            procedure Finalizer_Scope_1073 with Inline_Always;
      Fld : Internal_Entity_Basic_Decl;
Fld_1 : String_Type;
Let_Result : String_Type;
            procedure Finalizer_Scope_1073 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1073;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8937 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8936 '<FieldAccess for unique_identifying_name_impl at ast.py:14867>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14867
--# expr-start 8935 '<FieldAccess for basic_decl at ast.py:14867>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14867
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8935
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.unique_identifying_name_impl
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Unique_Identifying_Name_Impl (Node => Fld.Node, Dn => Ent, E_Info => Fld.Info);
--# end
--# expr-done 8936
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 8937
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1073;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1073;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Unique_Identifying_Name;
--# end
--# property-start DefiningName.fully_qualified_name_array /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14870
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Fully_Qualified_Name_Array
  (Node : Bare_Defining_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1074 with Inline_Always;
      Fld : Internal_Entity_Basic_Decl;
Fld_1 : String_Type_Array_Access;
Sym : Symbol_Type;
Map_Result : Symbol_Type_Array_Access;
Let_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1074 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1074;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8942 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8941 '<Map at ast.py:14875>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14875
   --# expr-start 8939 '<FieldAccess for fully_qualified_name_impl at ast.py:14875>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14875
--# expr-start 8938 '<FieldAccess for basic_decl at ast.py:14875>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14875
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8938
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.fully_qualified_name_impl
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Fully_Qualified_Name_Impl (Node => Fld.Node, Dn => Ent, E_Info => Fld.Info);
--# end
--# expr-done 8939
      declare
         Map_Result_Vec : Symbol_Type_Vectors.Vector;
      begin
      declare
         Collection : constant String_Type_Array_Access := Fld_1;
      begin
         for Item_263 of
               Collection.Items
         loop
   --# scope-start
               --# bind t Item_263
      --# expr-start 8940 '<to_symbol at ast.py:14877>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14877
Sym := String_To_Symbol (Self.Unit.Context, Item_263); 
--# expr-done 8940
         declare
            Item_To_Append : constant Symbol_Type := Sym;
         begin
            Symbol_Type_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Symbol_Type_Array
           (Items_Count => Natural (Symbol_Type_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Symbol_Type_Vectors.Get
              (Map_Result_Vec,
               I + Symbol_Type_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Symbol_Type_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 8941
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 8942
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1074;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1074;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Fully_Qualified_Name_Array;
--# end
--# property-start DefiningName.fully_qualified_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14880
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Fully_Qualified_Name
  (Node : Bare_Defining_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return String_Type
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : String_Type;
            procedure Finalizer_Scope_1075 with Inline_Always;
      Str : String_Type;
Fld : Internal_Entity_Basic_Decl;
Fld_1 : String_Type_Array_Access;
Join_Result : String_Type;
Let_Result : String_Type;
            procedure Finalizer_Scope_1075 is
            begin
                     Dec_Ref (Str);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Join_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1075;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8947 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8946 '<Join at ast.py:14884>' Join_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14884
--# expr-start 8943 '<String at ast.py:14884>' Str /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14884
Str := Create_String ("."); 
--# expr-done 8943
--# expr-start 8945 '<FieldAccess for fully_qualified_name_impl at ast.py:14885>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14885
--# expr-start 8944 '<FieldAccess for basic_decl at ast.py:14885>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14885
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8944
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.fully_qualified_name_impl
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Fully_Qualified_Name_Impl (Node => Fld.Node, Dn => Ent, E_Info => Fld.Info);
--# end
--# expr-done 8945
Join_Result := Join_Strings (Str, Fld_1); 
--# expr-done 8946
Let_Result := Join_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 8947
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1075;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1075;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Fully_Qualified_Name;
--# end
--# property-start DefiningName.all_env_els_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14889
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_All_Env_Els_Impl
  (Node : Bare_Defining_Name
      ; Seq : Boolean
         := True
      ; Seq_From : Bare_Ada_Node
         := No_Bare_Ada_Node
      ; Categories : Ref_Categories
         := (Inherited_Primitives => True, Nocat => True)
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind seq Seq
   --# bind seq_from Seq_From
   --# bind categories Categories
   --# bind env Env
   --# bind origin Origin
   Property_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1076 with Inline_Always;
      Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Array_Access;
Let_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1076 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1076;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8950 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8949 '<FieldAccess for all_env_els_impl at ast.py:14894>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14894
--# expr-start 8948 '<FieldAccess for name at ast.py:14894>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14894
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Defining_Name_F_Name, Info => Ent.Info);
--# expr-done 8948
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.all_env_els_impl'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_All_Env_Els_Impl (Node => Fld.Node, Seq => Seq, Seq_From => Seq_From, Categories => Categories, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 8949
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 8950
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1076;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1076;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_All_Env_Els_Impl;
--# end
--# property-start DefiningName.basic_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14896
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Basic_Decl
  (Node : Bare_Defining_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_1077 with Inline_Always;
      Node_Parents : Bare_Ada_Node_Array_Access;
Is_A : Boolean;
Find_Result : Bare_Ada_Node;
Cast_Result : Bare_Basic_Decl;
As_Entity : Internal_Entity_Basic_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_1077 is
            begin
                     Dec_Ref (Node_Parents);
            end Finalizer_Scope_1077;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Defining_Name_Defining_Name_P_Basic_Decl,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Defining_Name,
                                As_Bare_Defining_Name => Self);
               Mmz_K.Items (2) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Basic_Decl;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 8955 '<as_entity at ast.py:14897>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14897
--# expr-start 8954 '<Cast to BasicDecl at ast.py:14897>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14897
--# expr-start 8953 '<Find at ast.py:14897>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14897
   --# expr-start 8951 '<FieldAccess for parents at ast.py:14897>' Node_Parents /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14897
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents := Parents (Self, True); 
--# expr-done 8951
      Find_Result := No_Bare_Ada_Node;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents;
      begin
         for Item_264 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_264
      --# expr-start 8952 '<IsA BasicDecl at ast.py:14897>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14897
Is_A := Item_264 /= null 
and then Item_264.Kind in Ada_Basic_Decl; 
--# expr-done 8952
      if Is_A then
         Find_Result := Item_264;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 8953
   if Find_Result = null
      or else Find_Result.Kind in Ada_Basic_Decl
   then
      Cast_Result := Find_Result;
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 8954
Is_Null := Cast_Result = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Decl;
else
As_Entity := (Info => E_Info, Node => Cast_Result); 
   If_Result := As_Entity;
end if;
--# expr-done 8955
      Property_Result := If_Result;
   --# end
      Finalizer_Scope_1077;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Basic_Decl,
                        As_Internal_Entity_Basic_Decl => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1077;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Basic_Decl;
--# end
--# property-start DefiningName.basic_decl_no_internal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14903
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Basic_Decl_No_Internal
  (Node : Bare_Defining_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Basic_Decl;
      Var_Expr : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Is_A : Boolean;
If_Result : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 8956 '<FieldAccess for basic_decl at ast.py:14904>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14904
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8956
Var_Expr := Fld;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind bd Var_Expr
--# expr-start 8960 '<If at ast.py:14905>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14905
--# expr-start 8957 '<IsA GenericPackageInternal, GenericSubpInternal, SingleTaskTypeDecl at ast.py:14906>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14906
Is_A := Var_Expr.Node /= null 
and then Var_Expr.Node.Kind in Ada_Generic_Package_Internal_Range | Ada_Generic_Subp_Internal_Range | Ada_Single_Task_Type_Decl_Range; 
--# expr-done 8957
if Is_A then
   --# expr-start 8959 '<Cast to BasicDecl at ast.py:14908>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14908
--# expr-start 8958 '<parent at ast.py:14908>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14908
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Parent (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# expr-done 8958
   if Fld_1.Node = null
      or else Fld_1.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 8959
   If_Result := Cast_Result;
else
   If_Result := Var_Expr;
end if;
--# expr-done 8960
Scope_Result := If_Result;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Basic_Decl;
end if;
      Property_Result := Result_Var;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Basic_Decl_No_Internal;
--# end
--# property-start DefiningName.find_refs /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14919
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Find_Refs
  (Node : Bare_Defining_Name
      ; Root : Internal_Entity
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Ref_Result_Array_Access
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind root Root
   --# bind origin Origin
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Ref_Result_Array_Access;
            procedure Finalizer_Scope_1080 with Inline_Always;
            procedure Finalizer_Scope_2846 with Inline_Always;
            procedure Finalizer_Scope_2848 with Inline_Always;
            procedure Finalizer_Scope_2849 with Inline_Always;
            procedure Finalizer_Scope_2850 with Inline_Always;
      Var_Expr : Internal_Entity_Array_Access;
Var_Expr_1 : Internal_Entity_Base_Id;
Var_Expr_2 : Ref_Result_Kind;
Fld : Internal_Entity_Array_Access;
Is_Null : Boolean;
Fld_1 : Bare_Ada_Node;
Cast_Result : Bare_Ada_Node;
Is_Equal : Boolean;
If_Result : Boolean;
Not_Val : Boolean;
Map_Result : Internal_Entity_Array_Access;
Fld_2 : Internal_Ref_Result_Array_Access;
Map_Result_1 : Internal_Ref_Result_Array_Access;
Scope_Result : Internal_Ref_Result_Array_Access;
Result_Var : Internal_Ref_Result_Array_Access;
Cast_Expr : Internal_Entity;
Cast_Result_1 : Internal_Entity_Base_Id;
Fld_3 : Ref_Result_Kind;
New_Struct : Internal_Ref_Result;
Singleton : Internal_Ref_Result_Array_Access;
Is_Equal_1 : Boolean;
Is_Equal_2 : Boolean;
If_Result_1 : Boolean;
If_Result_2 : Internal_Ref_Result_Array_Access;
Scope_Result_1 : Internal_Ref_Result_Array_Access;
Result_Var_1 : Internal_Ref_Result_Array_Access;
Scope_Result_2 : Internal_Ref_Result_Array_Access;
Result_Var_2 : Internal_Ref_Result_Array_Access;
Concat_Result : Internal_Ref_Result_Array_Access;
Let_Result : Internal_Ref_Result_Array_Access;
            procedure Finalizer_Scope_1080 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Result_Var_2);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1080;
            procedure Finalizer_Scope_2846 is
            begin
                     Dec_Ref (Var_Expr);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Map_Result_1);
            end Finalizer_Scope_2846;
            procedure Finalizer_Scope_2848 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_2848;
            procedure Finalizer_Scope_2849 is
            begin
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
            end Finalizer_Scope_2849;
            procedure Finalizer_Scope_2850 is
            begin
                     Dec_Ref (Singleton);
                     Dec_Ref (If_Result_2);
            end Finalizer_Scope_2850;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 8976 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8975 '<Concat at ast.py:14925>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14925
--# expr-start 8961 '<children at ast.py:14925>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14925
      if Root.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children (Node => Root.Node, E_Info => Root.Info);
--# expr-done 8961
Var_Expr := Fld;
      Inc_Ref (Var_Expr);
if Var_Expr /= No_Internal_Entity_Array_Type then
   --# scope-start
   --# bind c Var_Expr
--# expr-start 8968 '<Mapcat at ast.py:14926>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14926
   --# expr-start 8966 '<Filter at ast.py:14926>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14926
      declare
         Map_Result_Vec : Internal_Entity_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Var_Expr;
      begin
         for Item_266 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_266
         --# expr-start 8965 '<Not at ast.py:14926>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14926
--# expr-start 8962 '<IsNull at ast.py:14926>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14926
Is_Null := Item_266.Node = null; 
--# expr-done 8962
if Is_Null then
   If_Result := True;
else
   --# expr-start 8964 '<Eq at ast.py:14926>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14926
--# expr-start 8963 '<FieldAccess for node at ast.py:14926>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14926
Fld_1 := Item_266.Node;
--# expr-done 8963
      Cast_Result := Fld_1;
Is_Equal := Cast_Result = Origin; 
--# expr-done 8964
   If_Result := Is_Equal;
end if;
Not_Val := not (If_Result); 
--# expr-done 8965
         if Not_Val then
         declare
            Item_To_Append : constant Internal_Entity := Item_266;
         begin
            Internal_Entity_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 8966
      declare
         Map_Result_1_Vec : Internal_Ref_Result_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Map_Result;
      begin
         for Item_265 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_265
      --# expr-start 8967 '<FieldAccess for find_refs at ast.py:14927>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14927
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.find_refs
Fld_2 := Libadalang.Implementation.Defining_Name_P_Find_Refs (Node => Ent.Node, Root => Item_265, Origin => Origin, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 8967
         for Item_To_Append of
               Fld_2.Items
         loop
            Internal_Ref_Result_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_2848;
         end loop;
      end;
         Map_Result_1 := Create_Internal_Ref_Result_Array
           (Items_Count => Natural (Internal_Ref_Result_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Ref_Result_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Ref_Result_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Ref_Result_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 8968
Scope_Result := Map_Result_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2846;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Ref_Result_Array_Type;
end if;
      Inc_Ref (Result_Var);
--# expr-start 8969 '<Cast to BaseId at ast.py:14929>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14929
Cast_Expr := Root; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Base_Id
   then
      Cast_Result_1 := Create_Internal_Entity_Base_Id
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result_1 := No_Entity_Base_Id;
   end if;
--# expr-done 8969
Var_Expr_1 := Cast_Result_1;
if Var_Expr_1 /= No_Entity_Base_Id then
   --# scope-start
   --# bind id Var_Expr_1
--# expr-start 8970 '<FieldAccess for is_referenced_by at ast.py:14930>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14930
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.is_referenced_by
Fld_3 := Libadalang.Implementation.Defining_Name_P_Is_Referenced_By (Node => Ent.Node, Id => Var_Expr_1, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 8970
Var_Expr_2 := Fld_3;
if Var_Expr_2 /= No_Ref then
   --# scope-start
   --# bind ref_kind Var_Expr_2
--# expr-start 8974 '<If at ast.py:14931>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14931
--# expr-start 8971 '<Eq at ast.py:14932>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14932
Is_Equal_1 := Var_Expr_2 = Precise; 
--# expr-done 8971
if Is_Equal_1 then
   If_Result_1 := True;
else
   --# expr-start 8972 '<Eq at ast.py:14932>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14932
Is_Equal_2 := Var_Expr_2 = Imprecise; 
--# expr-done 8972
   If_Result_1 := Is_Equal_2;
end if;
if If_Result_1 then
   --# expr-start 8973 '<New RefResult at ast.py:14935>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14935
New_Struct := (Kind => Var_Expr_2, Ref => Var_Expr_1); 
--# expr-done 8973
                Singleton := Create_Internal_Ref_Result_Array (Items_Count => 1);
                Singleton.Items (1) := New_Struct;
   If_Result_2 := Singleton;
else
   If_Result_2 := No_Internal_Ref_Result_Array_Type;
end if;
      Inc_Ref (If_Result_2);
--# expr-done 8974
Scope_Result_1 := If_Result_2;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2850;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Internal_Ref_Result_Array_Type;
end if;
      Inc_Ref (Result_Var_1);
Scope_Result_2 := Result_Var_1;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2849;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Internal_Ref_Result_Array_Type;
end if;
      Inc_Ref (Result_Var_2);
Concat_Result := Concat (Result_Var, Result_Var_2); 
--# expr-done 8975
Let_Result := Concat_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 8976
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1080;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1080;
               Finalizer_Scope_2846;
               Finalizer_Scope_2848;
               Finalizer_Scope_2849;
               Finalizer_Scope_2850;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Find_Refs;
--# end
--# property-start DefiningName.is_derivable_equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14942
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Is_Derivable_Equal
  (Node : Bare_Defining_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Base_Subp_Spec;
Var_Expr_1 : Internal_Entity_Type_Expr;
Sym : Symbol_Type;
Fld : Boolean;
Fld_1 : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity_Base_Subp_Spec;
Fld_3 : Internal_Entity_Type_Expr;
Fld_4 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Fld_5 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Cast_Result_1 : Internal_Entity;
Is_Equiv : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Defining_Name_Defining_Name_P_Is_Derivable_Equal,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Defining_Name,
                                As_Bare_Defining_Name => Self);
               Mmz_K.Items (2) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Boolean;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 8985 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8978 '<FieldAccess for name_is at ast.py:14950>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14950
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8977 '<Symbol "=" at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_10); 
--# expr-done 8977
--# property-call-start Name.name_is
Fld := Libadalang.Implementation.Name_P_Name_Is (Node => Self, Sym => Sym);
--# end
--# expr-done 8978
if Fld then
--# expr-start 8980 '<FieldAccess for subp_spec_or_null at ast.py:14951>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14951
--# expr-start 8979 '<FieldAccess for basic_decl at ast.py:14951>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14951
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld_1 := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8979
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 8980
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Base_Subp_Spec then
   --# scope-start
   --# bind s Var_Expr
--# expr-start 8984 '<Eq at ast.py:14952>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14952
--# expr-start 8981 '<FieldAccess for returns at ast.py:14952>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14952
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.returns'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Returns (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 8981
Var_Expr_1 := Fld_3;
if Var_Expr_1 /= No_Entity_Type_Expr then
   --# scope-start
--# expr-start 8982 '<FieldAccess for designated_type_decl at ast.py:14952>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14952
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeExpr.designated_type_decl
Fld_4 := Libadalang.Implementation.Type_Expr_P_Designated_Type_Decl (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 8982
Scope_Result := Fld_4;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
      Cast_Result := Create_Internal_Entity
        (Node => Result_Var.Node,
         Info => Result_Var.Info);
--# expr-start 8983 '<FieldAccess for bool_type at ast.py:14952>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14952
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_5 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Ent.Node);
--# end
--# expr-done 8983
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
Is_Equiv := Equivalent (Cast_Result, Cast_Result_1); 
--# expr-done 8984
Scope_Result_1 := Is_Equiv;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
   If_Result := Result_Var_1;
else
   If_Result := False;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 8985
      Property_Result := Let_Result;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Boolean,
                        As_Boolean => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Is_Derivable_Equal;
--# end
--# property-start DefiningName.is_potential_reference /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14957
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Is_Potential_Reference
  (Node : Bare_Defining_Name
      ; Symbol : Symbol_Type
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind symbol Symbol
   Property_Result : Boolean;
      Fld : Boolean;
Fld_1 : Boolean;
Sym : Symbol_Type;
Is_Equal : Boolean;
If_Result : Boolean;
If_Result_1 : Boolean;
Let_Result : Boolean;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Defining_Name_Defining_Name_P_Is_Potential_Reference,
            Items    => new Mmz_Key_Array (1 ..  3))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Defining_Name,
                                As_Bare_Defining_Name => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Symbol_Type,
                                      As_Symbol_Type => Symbol);
               Mmz_K.Items (3) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Boolean;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 8990 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 8986 '<FieldAccess for name_is at ast.py:14963>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14963
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_is
Fld := Libadalang.Implementation.Name_P_Name_Is (Node => Self, Sym => Symbol);
--# end
--# expr-done 8986
if Fld then
   If_Result_1 := True;
else
--# expr-start 8987 '<FieldAccess for is_derivable_equal at ast.py:14964>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14964
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.is_derivable_equal
Fld_1 := Libadalang.Implementation.Defining_Name_P_Is_Derivable_Equal (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 8987
if Fld_1 then
   --# expr-start 8989 '<Eq at ast.py:14964>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14964
--# expr-start 8988 '<Symbol "/=" at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_7); 
--# expr-done 8988
Is_Equal := Symbol = Sym; 
--# expr-done 8989
   If_Result := Is_Equal;
else
   If_Result := False;
end if;
   If_Result_1 := If_Result;
end if;
Let_Result := If_Result_1; 
--# end
--# expr-done 8990
      Property_Result := Let_Result;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Boolean,
                        As_Boolean => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Is_Potential_Reference;
--# end
--# property-start DefiningName.is_referenced_by /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14969
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Is_Referenced_By
  (Node : Bare_Defining_Name
      ; Id : Internal_Entity_Base_Id
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Ref_Result_Kind
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind id Id
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Ref_Result_Kind;
            procedure Finalizer_Scope_2853 with Inline_Always;
            procedure Finalizer_Scope_2858 with Inline_Always;
      Var_Expr : Internal_Refd_Def;
Canon : Bare_Defining_Name;
Var_Expr_1 : Internal_Entity_Defining_Name;
Var_Expr_2 : Internal_Entity_Defining_Name;
Var_Expr_3 : Internal_Entity_Basic_Decl_Array_Access;
Fld : Internal_Entity_Defining_Name;
New_Struct : Internal_Refd_Def;
Fld_1 : Internal_Refd_Def;
Fld_2 : Boolean;
If_Result : Internal_Refd_Def;
Fld_3 : Internal_Entity_Defining_Name;
Fld_4 : Internal_Entity_Defining_Name;
Scope_Result : Internal_Entity_Defining_Name;
Result_Var : Internal_Entity_Defining_Name;
Fld_5 : Bare_Defining_Name;
Scope_Result_1 : Bare_Defining_Name;
Result_Var_1 : Bare_Defining_Name;
Fld_6 : Ref_Result_Kind;
Cast_Expr : Bare_Defining_Name;
Cast_Result : Bare_Defining_Name;
Is_Equal : Boolean;
Fld_7 : Internal_Entity_Basic_Decl;
Fld_8 : Internal_Entity_Basic_Decl_Array_Access;
Fld_9 : Internal_Entity_Defining_Name;
Fld_10 : Bare_Defining_Name;
Cast_Result_1 : Bare_Defining_Name;
Is_Equal_1 : Boolean;
Quantifier_Result : Boolean;
Fld_11 : Boolean;
If_Result_1 : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Ref_Result_Kind;
Let_Result : Ref_Result_Kind;
Scope_Result_3 : Ref_Result_Kind;
Result_Var_3 : Ref_Result_Kind;
Fld_12 : Symbol_Type;
Fld_13 : Boolean;
If_Result_4 : Ref_Result_Kind;
Let_Result_1 : Ref_Result_Kind;
            procedure Finalizer_Scope_2853 is
            begin
                     Dec_Ref (Fld_8);
            end Finalizer_Scope_2853;
            procedure Finalizer_Scope_2858 is
            begin
                     Dec_Ref (Var_Expr_3);
            end Finalizer_Scope_2858;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9013 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 9012 '<If at ast.py:14981>' If_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14981
--# expr-start 8992 '<FieldAccess for is_potential_reference at ast.py:14982>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14982
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 8991 '<FieldAccess for name_symbol at ast.py:14982>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14982
      if Id.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_12 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Id.Node);
--# end
--# expr-done 8991
--# property-call-start DefiningName.is_potential_reference
Fld_13 := Libadalang.Implementation.Defining_Name_P_Is_Potential_Reference (Node => Ent.Node, Symbol => Fld_12, E_Info => Ent.Info);
--# end
--# expr-done 8992
if Fld_13 then
--# expr-start 8997 '<If at ast.py:14984>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14984
--# expr-start 8993 '<FieldAccess for is_defining at ast.py:14984>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14984
      if Id.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_defining
Fld_2 := Libadalang.Implementation.Name_P_Is_Defining (Node => Id.Node);
--# end
--# expr-done 8993
if Fld_2 then
   --# expr-start 8995 '<New RefdDef at ast.py:14985>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14985
--# expr-start 8994 '<FieldAccess for enclosing_defining_name at ast.py:14985>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14985
      if Id.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.enclosing_defining_name
Fld := Libadalang.Implementation.Name_P_Enclosing_Defining_Name (Node => Id.Node, E_Info => Id.Info);
--# end
--# expr-done 8994
New_Struct := (Def_Name => Fld, Kind => Precise); 
--# expr-done 8995
   If_Result := New_Struct;
else
   --# expr-start 8996 '<FieldAccess for failsafe_referenced_def_name at ast.py:14987>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14987
      if Id.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.failsafe_referenced_def_name
Fld_1 := Libadalang.Implementation.Name_P_Failsafe_Referenced_Def_Name (Node => Id.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Id.Info);
--# end
--# expr-done 8996
   If_Result := Fld_1;
end if;
--# expr-done 8997
Var_Expr := If_Result;
if Var_Expr /= No_Refd_Def then
   --# scope-start
   --# bind def_res Var_Expr
--# expr-start 9011 '<Let at ast.py:14990>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14990
--# scope-start
--# expr-start 8998 '<FieldAccess for def_name at ast.py:14991>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14991
Fld_3 := Var_Expr.Def_Name;
--# expr-done 8998
Var_Expr_2 := Fld_3;
if Var_Expr_2 /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 8999 '<FieldAccess for canonical_part at ast.py:14991>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14991
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.canonical_part
Fld_4 := Libadalang.Implementation.Defining_Name_P_Canonical_Part (Node => Var_Expr_2.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 8999
Scope_Result := Fld_4;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Defining_Name;
end if;
Var_Expr_1 := Result_Var;
if Var_Expr_1 /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 9000 '<FieldAccess for node at ast.py:14991>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14991
Fld_5 := Var_Expr_1.Node;
--# expr-done 9000
Scope_Result_1 := Fld_5;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Bare_Ada_Node;
end if;
Canon := Result_Var_1; 
--# bind canon Canon
--# expr-start 9010 '<If at ast.py:14993>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14993
--# expr-start 9001 '<Eq at ast.py:14996>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14996
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Is_Equal := Cast_Result = Canon; 
--# expr-done 9001
if Is_Equal then
   If_Result_2 := True;
else
--# expr-start 9003 '<FieldAccess for base_subp_declarations at ast.py:15001>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15001
--# expr-start 9002 '<FieldAccess for basic_decl at ast.py:15001>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15001
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld_7 := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 9002
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.base_subp_declarations
Fld_8 := Libadalang.Implementation.Basic_Decl_P_Base_Subp_Declarations (Node => Fld_7.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_7.Info);
--# end
--# expr-done 9003
Var_Expr_3 := Fld_8;
      Inc_Ref (Var_Expr_3);
if Var_Expr_3 /= No_Internal_Entity_Basic_Decl_Array_Type then
   --# scope-start
   --# bind decls Var_Expr_3
--# expr-start 9007 '<AnyQuantifier at ast.py:15002>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15002
      Quantifier_Result := False;
      declare
         Collection : constant Internal_Entity_Basic_Decl_Array_Access := Var_Expr_3;
      begin
         for Item_267 of
               Collection.Items
         loop
   --# scope-start
               --# bind d Item_267
      --# expr-start 9006 '<Eq at ast.py:15003>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15003
--# expr-start 9005 '<FieldAccess for node at ast.py:15003>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15003
--# expr-start 9004 '<FieldAccess for defining_name at ast.py:15003>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15003
      if Item_267.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_9 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Item_267.Node, E_Info => Item_267.Info);
--# end
--# expr-done 9004
Fld_10 := Fld_9.Node;
--# expr-done 9005
      Cast_Result_1 := Fld_10;
Is_Equal_1 := Cast_Result_1 = Canon; 
--# expr-done 9006
      Quantifier_Result := Is_Equal_1;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
--# expr-done 9007
if Quantifier_Result then
   --# expr-start 9008 '<FieldAccess for is_dispatching_call at ast.py:15005>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15005
      if Id.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.is_dispatching_call'
Fld_11 := Libadalang.Implementation.Dispatcher_Expr_P_Is_Dispatching_Call (Node => Id.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Id.Info);
--# end
--# expr-done 9008
   If_Result_1 := Fld_11;
else
   If_Result_1 := False;
end if;
Scope_Result_2 := If_Result_1;
   --# end
      Finalizer_Scope_2858;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := False;
end if;
   If_Result_2 := Result_Var_2;
end if;
if If_Result_2 then
   --# expr-start 9009 '<FieldAccess for kind at ast.py:15008>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15008
Fld_6 := Var_Expr.Kind;
--# expr-done 9009
   If_Result_3 := Fld_6;
else
   If_Result_3 := No_Ref;
end if;
--# expr-done 9010
Let_Result := If_Result_3; 
--# end
--# expr-done 9011
Scope_Result_3 := Let_Result;
   --# end
      Finalizer_Scope_2853;
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := No_Ref;
end if;
   If_Result_4 := Result_Var_3;
else
   If_Result_4 := No_Ref;
end if;
--# expr-done 9012
Let_Result_1 := If_Result_4; 
--# end
--# expr-done 9013
      Property_Result := Let_Result_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_2853;
               Finalizer_Scope_2858;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Is_Referenced_By;
--# end
--# property-start DefiningName.find_all_references /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15019
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Find_All_References
  (Node : Bare_Defining_Name
      ; Units : Internal_Unit_Array_Access
      ; Follow_Renamings : Boolean
         := False
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Ref_Result_Array_Access
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind units Units
   --# bind follow_renamings Follow_Renamings
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Ref_Result_Array_Access;
            procedure Finalizer_Scope_1084 with Inline_Always;
            procedure Finalizer_Scope_2860 with Inline_Always;
            procedure Finalizer_Scope_2861 with Inline_Always;
            procedure Finalizer_Scope_2862 with Inline_Always;
            procedure Finalizer_Scope_2863 with Inline_Always;
            procedure Finalizer_Scope_2864 with Inline_Always;
            procedure Finalizer_Scope_2867 with Inline_Always;
            procedure Finalizer_Scope_2869 with Inline_Always;
      Block_Var : Internal_Entity_Defining_Name;
Block_Var_1 : Internal_Entity_Basic_Decl_Array_Access;
Block_Var_2 : Internal_Unit_Array_Access;
Block_Var_3 : Internal_Ref_Result_Array_Access;
Var_Expr : Internal_Entity_Basic_Decl_Array_Access;
Var_Expr_1 : Bare_Ada_Node;
Var_Expr_2 : Internal_Entity_Renaming_Clause;
Var_Expr_3 : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Defining_Name;
Cast_Expr : Bare_Defining_Name;
Cast_Result : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_1 : Internal_Entity_Basic_Decl;
Fld_2 : Internal_Entity_Basic_Decl_Array_Access;
Scope_Result : Internal_Entity_Basic_Decl_Array_Access;
Fld_3 : Internal_Entity_Basic_Decl;
Singleton : Internal_Entity_Basic_Decl_Array_Access;
Result_Var : Internal_Entity_Basic_Decl_Array_Access;
Dyn_Var_Bind_Result : Internal_Entity_Basic_Decl_Array_Access;
Fld_4 : Internal_Unit_Array_Access;
Map_Result : Internal_Unit_Array_Access;
Unique_Array : Internal_Unit_Array_Access;
Cast_Expr_1 : Bare_Defining_Name;
Cast_Result_1 : Bare_Ada_Node;
Bound_Origin_1 : Bare_Ada_Node;
Fld_5 : Bare_Ada_Node;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Fld_6 : Internal_Ref_Result_Array_Access;
Scope_Result_1 : Internal_Ref_Result_Array_Access;
Result_Var_1 : Internal_Ref_Result_Array_Access;
Map_Result_1 : Internal_Ref_Result_Array_Access;
Dyn_Var_Bind_Result_1 : Internal_Ref_Result_Array_Access;
Fld_7 : Internal_Entity_Base_Id;
Fld_8 : Internal_Entity_Array_Access;
Is_A : Boolean;
Find_Result : Internal_Entity;
Cast_Result_2 : Internal_Entity_Renaming_Clause;
Fld_9 : Internal_Entity_Name;
Fld_10 : Internal_Entity_Defining_Name;
Cast_Result_3 : Internal_Entity;
Cast_Expr_2 : Internal_Entity_Defining_Name;
Cast_Result_4 : Internal_Entity;
Is_Equiv : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
Map_Result_2 : Internal_Ref_Result_Array_Access;
Fld_11 : Internal_Entity_Base_Id;
Fld_12 : Internal_Entity_Array_Access;
Is_A_1 : Boolean;
Find_Result_1 : Internal_Entity;
Fld_13 : Internal_Entity;
Cast_Result_5 : Internal_Entity_Basic_Decl;
Fld_14 : Internal_Entity_Defining_Name;
Fld_15 : Internal_Ref_Result_Array_Access;
Scope_Result_3 : Internal_Ref_Result_Array_Access;
Result_Var_3 : Internal_Ref_Result_Array_Access;
Map_Result_3 : Internal_Ref_Result_Array_Access;
Concat_Result : Internal_Ref_Result_Array_Access;
If_Result_1 : Internal_Ref_Result_Array_Access;
Let_Result : Internal_Ref_Result_Array_Access;
            procedure Finalizer_Scope_1084 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Singleton);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Unique_Array);
                     Dec_Ref (Block_Var_2);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Dyn_Var_Bind_Result_1);
                     Dec_Ref (Block_Var_3);
                     Dec_Ref (Map_Result_2);
                     Dec_Ref (Map_Result_3);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1084;
            procedure Finalizer_Scope_2860 is
            begin
                     Dec_Ref (Var_Expr);
            end Finalizer_Scope_2860;
            procedure Finalizer_Scope_2861 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_2861;
            procedure Finalizer_Scope_2862 is
            begin
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
            end Finalizer_Scope_2862;
            procedure Finalizer_Scope_2863 is
            begin
                     Dec_Ref (Fld_6);
            end Finalizer_Scope_2863;
            procedure Finalizer_Scope_2864 is
            begin
                     Dec_Ref (Fld_8);
            end Finalizer_Scope_2864;
            procedure Finalizer_Scope_2867 is
            begin
                     Dec_Ref (Fld_12);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Result_Var_3);
            end Finalizer_Scope_2867;
            procedure Finalizer_Scope_2869 is
            begin
                     Dec_Ref (Fld_15);
            end Finalizer_Scope_2869;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9047 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9014 '<FieldAccess for canonical_part at ast.py:15029>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15029
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.canonical_part
Fld := Libadalang.Implementation.Defining_Name_P_Canonical_Part (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 9014
Block_Var := Fld; 
--# bind dn Block_Var
--# expr-start 9018 '<bind at ast.py:15039>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15039
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Bound_Origin := Cast_Result; 
--# expr-start 9016 '<FieldAccess for root_subp_declarations at ast.py:15041>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15041
--# expr-start 9015 '<FieldAccess for basic_decl at ast.py:15041>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15041
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld_1 := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 9015
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.root_subp_declarations
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Root_Subp_Declarations (Node => Fld_1.Node, Origin => Bound_Origin, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_1.Info);
--# end
--# expr-done 9016
Var_Expr := Fld_2;
      Inc_Ref (Var_Expr);
if Var_Expr /= No_Internal_Entity_Basic_Decl_Array_Type then
   --# scope-start
   --# bind e Var_Expr
Scope_Result := Var_Expr;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2860;
   Result_Var := Scope_Result;
else
   --# expr-start 9017 '<FieldAccess for basic_decl at ast.py:15041>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15041
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld_3 := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 9017
                Singleton := Create_Internal_Entity_Basic_Decl_Array (Items_Count => 1);
                Singleton.Items (1) := Fld_3;
   Result_Var := Singleton;
end if;
      Inc_Ref (Result_Var);
Dyn_Var_Bind_Result := Result_Var; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 9018
Block_Var_1 := Dyn_Var_Bind_Result; Inc_Ref (Block_Var_1);
--# bind bases Block_Var_1
--# expr-start 9021 '<unique at ast.py:15044>' Unique_Array /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15044
--# expr-start 9020 '<Mapcat at ast.py:15044>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15044
      declare
         Map_Result_Vec : Internal_Unit_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Basic_Decl_Array_Access := Block_Var_1;
      begin
         for Item_268 of
               Collection.Items
         loop
   --# scope-start
               --# bind base Item_268
      --# expr-start 9019 '<FieldAccess for filter_is_imported_by at ast.py:15045>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15045
      if Item_268.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Ada_Node_P_Filter_Is_Imported_By (Node => Item_268.Node, Units => Units, Transitive => True);
--# expr-done 9019
         for Item_To_Append of
               Fld_4.Items
         loop
            Internal_Unit_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_2861;
         end loop;
      end;
         Map_Result := Create_Internal_Unit_Array
           (Items_Count => Natural (Internal_Unit_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Unit_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Unit_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Unit_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 9020
Unique_Array := Make_Unique (Map_Result); 
--# expr-done 9021
Block_Var_2 := Unique_Array; Inc_Ref (Block_Var_2);
--# bind all_units Block_Var_2
--# expr-start 9026 '<bind at ast.py:15048>' Dyn_Var_Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15048
Cast_Expr_1 := Self; 
      Cast_Result_1 := Cast_Expr_1;
Bound_Origin_1 := Cast_Result_1; 
--# expr-start 9025 '<Mapcat at ast.py:15048>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15048
      declare
         Map_Result_1_Vec : Internal_Ref_Result_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Unit_Array_Access := Block_Var_2;
      begin
         for Item_269 of
               Collection.Items
         loop
   --# scope-start
               --# bind u Item_269
--# expr-start 9022 '<FieldAccess for root at ast.py:15048>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15048
      if Item_269 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Item_269.Ast_Root;
--# expr-done 9022
Var_Expr_1 := Fld_5;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind r Var_Expr_1
--# expr-start 9024 '<FieldAccess for find_refs at ast.py:15049>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15049
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9023 '<as_bare_entity at ast.py:15049>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15049
Is_Null := Var_Expr_1 = null; 
if Is_Null then
   If_Result := No_Entity;
else
As_Entity := (Info => No_Entity_Info, Node => Var_Expr_1); 
   If_Result := As_Entity;
end if;
--# expr-done 9023
--# property-call-start DefiningName.find_refs
Fld_6 := Libadalang.Implementation.Defining_Name_P_Find_Refs (Node => Block_Var.Node, Root => If_Result, Origin => Bound_Origin_1, Imprecise_Fallback => Imprecise_Fallback, E_Info => Block_Var.Info);
--# end
--# expr-done 9024
Scope_Result_1 := Fld_6;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2863;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Internal_Ref_Result_Array_Type;
end if;
      Inc_Ref (Result_Var_1);
         for Item_To_Append of
               Result_Var_1.Items
         loop
            Internal_Ref_Result_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_2862;
         end loop;
      end;
         Map_Result_1 := Create_Internal_Ref_Result_Array
           (Items_Count => Natural (Internal_Ref_Result_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Ref_Result_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Ref_Result_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Ref_Result_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 9025
Dyn_Var_Bind_Result_1 := Map_Result_1; Inc_Ref (Dyn_Var_Bind_Result_1);
--# expr-done 9026
Block_Var_3 := Dyn_Var_Bind_Result_1; Inc_Ref (Block_Var_3);
--# bind refs Block_Var_3
--# expr-start 9046 '<If at ast.py:15052>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15052
if Follow_Renamings then
   --# expr-start 9045 '<Concat at ast.py:15054>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15054
--# expr-start 9044 '<Mapcat at ast.py:15055>' Map_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15055
   --# expr-start 9035 '<Filter at ast.py:15055>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15055
      declare
         Map_Result_2_Vec : Internal_Ref_Result_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Ref_Result_Array_Access := Block_Var_3;
      begin
         for Item_271 of
               Collection.Items
         loop
   --# scope-start
               --# bind f Item_271
--# expr-start 9031 '<Cast to RenamingClause at ast.py:15061>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15061
--# expr-start 9030 '<Find at ast.py:15061>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15061
   --# expr-start 9028 '<FieldAccess for parents at ast.py:15061>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15061
--# expr-start 9027 '<FieldAccess for ref at ast.py:15061>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15061
Fld_7 := Item_271.Ref;
--# expr-done 9027
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Parents (Node => Fld_7.Node, With_Self => True, E_Info => Fld_7.Info);
--# expr-done 9028
      Find_Result := No_Entity;
      declare
         Collection : constant Internal_Entity_Array_Access := Fld_8;
      begin
         for Item_272 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_272
      --# expr-start 9029 '<IsA RenamingClause at ast.py:15062>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15062
Is_A := Item_272.Node /= null 
and then Item_272.Node.Kind in Ada_Renaming_Clause_Range; 
--# expr-done 9029
      if Is_A then
         Find_Result := Item_272;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 9030
   if Find_Result.Node = null
      or else Find_Result.Node.Kind in Ada_Renaming_Clause_Range
   then
      Cast_Result_2 := Create_Internal_Entity_Renaming_Clause
        (Node => Find_Result.Node,
         Info => Find_Result.Info);
   else
         Cast_Result_2 := No_Entity_Renaming_Clause;
   end if;
--# expr-done 9031
Var_Expr_2 := Cast_Result_2;
if Var_Expr_2 /= No_Entity_Renaming_Clause then
   --# scope-start
   --# bind r Var_Expr_2
--# expr-start 9034 '<Eq at ast.py:15065>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15065
--# expr-start 9033 '<FieldAccess for referenced_defining_name at ast.py:15065>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15065
--# expr-start 9032 '<FieldAccess for renamed_object at ast.py:15065>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15065
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Name (Node => Var_Expr_2.Node.Renaming_Clause_F_Renamed_Object, Info => Var_Expr_2.Info);
--# expr-done 9032
      if Fld_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_defining_name
Fld_10 := Libadalang.Implementation.Name_P_Referenced_Defining_Name (Node => Fld_9.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_9.Info);
--# end
--# expr-done 9033
      Cast_Result_3 := Create_Internal_Entity
        (Node => Fld_10.Node,
         Info => Fld_10.Info);
Cast_Expr_2 := Ent; 
      Cast_Result_4 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
Is_Equiv := Equivalent (Cast_Result_3, Cast_Result_4); 
--# expr-done 9034
Scope_Result_2 := Is_Equiv;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := False;
end if;
         if Result_Var_2 then
         declare
            Item_To_Append : constant Internal_Ref_Result := Item_271;
         begin
            Internal_Ref_Result_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
         end if;
   --# end
      Finalizer_Scope_2864;
         end loop;
      end;
         Map_Result_2 := Create_Internal_Ref_Result_Array
           (Items_Count => Natural (Internal_Ref_Result_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Internal_Ref_Result_Vectors.Get
              (Map_Result_2_Vec,
               I + Internal_Ref_Result_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Internal_Ref_Result_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 9035
      declare
         Map_Result_3_Vec : Internal_Ref_Result_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Ref_Result_Array_Access := Map_Result_2;
      begin
         for Item_270 of
               Collection.Items
         loop
   --# scope-start
               --# bind f Item_270
--# expr-start 9041 '<Cast to BasicDecl at ast.py:15071>' Cast_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15071
--# expr-start 9040 '<parent at ast.py:15071>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15071
--# expr-start 9039 '<Find at ast.py:15071>' Find_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15071
   --# expr-start 9037 '<FieldAccess for parents at ast.py:15071>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15071
--# expr-start 9036 '<FieldAccess for ref at ast.py:15071>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15071
Fld_11 := Item_270.Ref;
--# expr-done 9036
      if Fld_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Libadalang.Implementation.Parents (Node => Fld_11.Node, With_Self => True, E_Info => Fld_11.Info);
--# expr-done 9037
      Find_Result_1 := No_Entity;
      declare
         Collection : constant Internal_Entity_Array_Access := Fld_12;
      begin
         for Item_273 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_273
      --# expr-start 9038 '<IsA RenamingClause at ast.py:15072>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15072
Is_A_1 := Item_273.Node /= null 
and then Item_273.Node.Kind in Ada_Renaming_Clause_Range; 
--# expr-done 9038
      if Is_A_1 then
         Find_Result_1 := Item_273;
      end if;
   --# end
               exit when Is_A_1;
         end loop;
      end;
--# expr-done 9039
      if Find_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Libadalang.Implementation.Parent (Node => Find_Result_1.Node, E_Info => Find_Result_1.Info);
--# expr-done 9040
   if Fld_13.Node = null
      or else Fld_13.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_5 := Create_Internal_Entity_Basic_Decl
        (Node => Fld_13.Node,
         Info => Fld_13.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 9041
Var_Expr_3 := Cast_Result_5;
if Var_Expr_3 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind bd Var_Expr_3
--# expr-start 9043 '<FieldAccess for find_all_references at ast.py:15074>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15074
--# expr-start 9042 '<FieldAccess for defining_name at ast.py:15074>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15074
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_14 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Var_Expr_3.Node, E_Info => Var_Expr_3.Info);
--# end
--# expr-done 9042
      if Fld_14.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.find_all_references
Fld_15 := Libadalang.Implementation.Defining_Name_P_Find_All_References (Node => Fld_14.Node, Units => Units, Follow_Renamings => True, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_14.Info);
--# end
--# expr-done 9043
Scope_Result_3 := Fld_15;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_2869;
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := No_Internal_Ref_Result_Array_Type;
end if;
      Inc_Ref (Result_Var_3);
         for Item_To_Append of
               Result_Var_3.Items
         loop
            Internal_Ref_Result_Vectors.Append (Map_Result_3_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_2867;
         end loop;
      end;
         Map_Result_3 := Create_Internal_Ref_Result_Array
           (Items_Count => Natural (Internal_Ref_Result_Vectors.Length (Map_Result_3_Vec)));
         for I in Map_Result_3.Items'Range loop
            Map_Result_3.Items (I) := Internal_Ref_Result_Vectors.Get
              (Map_Result_3_Vec,
               I + Internal_Ref_Result_Vectors.Index_Type'First - Map_Result_3.Items'First);
         end loop;
         Internal_Ref_Result_Vectors.Destroy (Map_Result_3_Vec);
      end;
--# expr-done 9044
Concat_Result := Concat (Block_Var_3, Map_Result_3); 
--# expr-done 9045
   If_Result_1 := Concat_Result;
else
   If_Result_1 := Block_Var_3;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 9046
Let_Result := If_Result_1; Inc_Ref (Let_Result);
--# end
--# expr-done 9047
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1084;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1084;
               Finalizer_Scope_2860;
               Finalizer_Scope_2861;
               Finalizer_Scope_2862;
               Finalizer_Scope_2863;
               Finalizer_Scope_2864;
               Finalizer_Scope_2867;
               Finalizer_Scope_2869;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Find_All_References;
--# end
--# property-start DefiningName.find_matching_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15085
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Find_Matching_Name
  (Node : Bare_Defining_Name
      ; Bd : Internal_Entity_Basic_Decl
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind bd Bd
   Property_Result : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_1085 with Inline_Always;
            procedure Finalizer_Scope_2855 with Inline_Always;
      Var_Expr : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Defining_Name_Array_Access;
Scope_Result : Internal_Entity_Defining_Name_Array_Access;
Result_Var : Internal_Entity_Defining_Name_Array_Access;
Fld_1 : Internal_Entity_Name;
Fld_2 : Symbol_Type;
Fld_3 : Boolean;
Find_Result : Internal_Entity_Defining_Name;
Let_Result : Internal_Entity_Defining_Name;
            procedure Finalizer_Scope_1085 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1085;
            procedure Finalizer_Scope_2855 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_2855;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9053 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9052 '<Find at ast.py:15090>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15090
Var_Expr := Bd;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 9048 '<FieldAccess for defining_names at ast.py:15090>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15090
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_names'
Fld := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Names (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 9048
Scope_Result := Fld;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2855;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Defining_Name_Array_Type;
end if;
      Inc_Ref (Result_Var);
      Find_Result := No_Entity_Defining_Name;
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Result_Var;
      begin
         for Item_274 of
               Collection.Items
         loop
   --# scope-start
               --# bind di Item_274
      --# expr-start 9051 '<FieldAccess for name_is at ast.py:15091>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15091
--# expr-start 9049 '<FieldAccess for name at ast.py:15091>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15091
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Ent.Node.Defining_Name_F_Name, Info => Ent.Info);
--# expr-done 9049
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9050 '<FieldAccess for name_symbol at ast.py:15091>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15091
      if Item_274.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Item_274.Node);
--# end
--# expr-done 9050
--# property-call-start Name.name_is
Fld_3 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_1.Node, Sym => Fld_2);
--# end
--# expr-done 9051
      if Fld_3 then
         Find_Result := Item_274;
      end if;
   --# end
               exit when Fld_3;
         end loop;
      end;
--# expr-done 9052
Let_Result := Find_Result; 
--# end
--# expr-done 9053
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1085;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1085;
               Finalizer_Scope_2855;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Find_Matching_Name;
--# end
--# property-start DefiningName.find_all_calls /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15096
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Find_All_Calls
  (Node : Bare_Defining_Name
      ; Units : Internal_Unit_Array_Access
      ; Follow_Renamings : Boolean
         := False
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Ref_Result_Array_Access
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind units Units
   --# bind follow_renamings Follow_Renamings
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Ref_Result_Array_Access;
            procedure Finalizer_Scope_1086 with Inline_Always;
      Fld : Internal_Ref_Result_Array_Access;
Fld_1 : Internal_Entity_Base_Id;
Fld_2 : Boolean;
Map_Result : Internal_Ref_Result_Array_Access;
Let_Result : Internal_Ref_Result_Array_Access;
            procedure Finalizer_Scope_1086 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1086;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9058 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9057 '<Filter at ast.py:15108>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15108
   --# expr-start 9054 '<FieldAccess for find_all_references at ast.py:15108>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15108
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.find_all_references
Fld := Libadalang.Implementation.Defining_Name_P_Find_All_References (Node => Ent.Node, Units => Units, Follow_Renamings => Follow_Renamings, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 9054
      declare
         Map_Result_Vec : Internal_Ref_Result_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Ref_Result_Array_Access := Fld;
      begin
         for Item_275 of
               Collection.Items
         loop
   --# scope-start
               --# bind r Item_275
         --# expr-start 9056 '<FieldAccess for is_direct_call at ast.py:15109>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15109
--# expr-start 9055 '<FieldAccess for ref at ast.py:15109>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15109
Fld_1 := Item_275.Ref;
--# expr-done 9055
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_direct_call
Fld_2 := Libadalang.Implementation.Name_P_Is_Direct_Call (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 9056
         if Fld_2 then
         declare
            Item_To_Append : constant Internal_Ref_Result := Item_275;
         begin
            Internal_Ref_Result_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Ref_Result_Array
           (Items_Count => Natural (Internal_Ref_Result_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Ref_Result_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Ref_Result_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Ref_Result_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 9057
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 9058
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1086;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1086;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Find_All_Calls;
--# end
--# property-start DefiningName.next_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15112
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Next_Part
  (Node : Bare_Defining_Name
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Defining_Name;
      Fld : Internal_Entity_Basic_Decl;
Fld_1 : Symbol_Type;
Fld_2 : Internal_Entity_Basic_Decl;
Fld_3 : Internal_Entity_Defining_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9062 '<FieldAccess for find_matching_name at ast.py:15113>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15113
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9061 '<FieldAccess for next_part_for_name at ast.py:15114>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15114
--# expr-start 9059 '<FieldAccess for basic_decl at ast.py:15114>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15114
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 9059
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9060 '<FieldAccess for name_symbol at ast.py:15114>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15114
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 9060
--# property-call-start '[dispatcher]BasicDecl.next_part_for_name'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Next_Part_For_Name (Node => Fld.Node, Sym => Fld_1, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld.Info);
--# end
--# expr-done 9061
--# property-call-start DefiningName.find_matching_name
Fld_3 := Libadalang.Implementation.Defining_Name_P_Find_Matching_Name (Node => Ent.Node, Bd => Fld_2, E_Info => Ent.Info);
--# end
--# expr-done 9062
      Property_Result := Fld_3;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Next_Part;
--# end
--# property-start DefiningName.previous_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15121
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Previous_Part
  (Node : Bare_Defining_Name
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Defining_Name;
      Fld : Internal_Entity_Basic_Decl;
Fld_1 : Symbol_Type;
Fld_2 : Internal_Entity_Basic_Decl;
Fld_3 : Internal_Entity_Defining_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9066 '<FieldAccess for find_matching_name at ast.py:15122>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15122
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9065 '<FieldAccess for previous_part_for_name at ast.py:15123>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15123
--# expr-start 9063 '<FieldAccess for basic_decl at ast.py:15123>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15123
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 9063
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9064 '<FieldAccess for name_symbol at ast.py:15123>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15123
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 9064
--# property-call-start '[dispatcher]BasicDecl.previous_part_for_name'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Previous_Part_For_Name (Node => Fld.Node, Sym => Fld_1, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld.Info);
--# end
--# expr-done 9065
--# property-call-start DefiningName.find_matching_name
Fld_3 := Libadalang.Implementation.Defining_Name_P_Find_Matching_Name (Node => Ent.Node, Bd => Fld_2, E_Info => Ent.Info);
--# end
--# expr-done 9066
      Property_Result := Fld_3;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Previous_Part;
--# end
--# property-start DefiningName.canonical_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15130
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Canonical_Part
  (Node : Bare_Defining_Name
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Defining_Name;
      Fld : Internal_Entity_Basic_Decl;
Fld_1 : Symbol_Type;
Fld_2 : Internal_Entity_Basic_Decl;
Fld_3 : Internal_Entity_Defining_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9070 '<FieldAccess for find_matching_name at ast.py:15131>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15131
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9069 '<FieldAccess for canonical_part_for_name at ast.py:15132>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15132
--# expr-start 9067 '<FieldAccess for basic_decl at ast.py:15132>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15132
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 9067
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9068 '<FieldAccess for name_symbol at ast.py:15132>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15132
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 9068
--# property-call-start BasicDecl.canonical_part_for_name
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Canonical_Part_For_Name (Node => Fld.Node, Sym => Fld_1, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld.Info);
--# end
--# expr-done 9069
--# property-call-start DefiningName.find_matching_name
Fld_3 := Libadalang.Implementation.Defining_Name_P_Find_Matching_Name (Node => Ent.Node, Bd => Fld_2, E_Info => Ent.Info);
--# end
--# expr-done 9070
      Property_Result := Fld_3;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Canonical_Part;
--# end
--# property-start DefiningName.most_visible_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15141
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Most_Visible_Part
  (Node : Bare_Defining_Name
      ; Origin : Bare_Ada_Node
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Defining_Name;
      Fld : Internal_Entity_Basic_Decl;
Fld_1 : Symbol_Type;
Fld_2 : Internal_Entity_Basic_Decl;
Fld_3 : Internal_Entity_Defining_Name;
Let_Result : Internal_Entity_Defining_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9075 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9074 '<FieldAccess for find_matching_name at ast.py:15147>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15147
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9073 '<FieldAccess for most_visible_part_for_name at ast.py:15148>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15148
--# expr-start 9071 '<FieldAccess for basic_decl at ast.py:15148>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15148
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 9071
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9072 '<FieldAccess for name_symbol at ast.py:15148>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15148
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Ent.Node);
--# end
--# expr-done 9072
--# property-call-start BasicDecl.most_visible_part_for_name
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Most_Visible_Part_For_Name (Node => Fld.Node, Sym => Fld_1, Origin => Origin, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld.Info);
--# end
--# expr-done 9073
--# property-call-start DefiningName.find_matching_name
Fld_3 := Libadalang.Implementation.Defining_Name_P_Find_Matching_Name (Node => Ent.Node, Bd => Fld_2, E_Info => Ent.Info);
--# end
--# expr-done 9074
Let_Result := Fld_3; 
--# end
--# expr-done 9075
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Most_Visible_Part;
--# end
--# property-start DefiningName.all_previous_parts /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15153
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_All_Previous_Parts
  (Node : Bare_Defining_Name
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1091 with Inline_Always;
            procedure Finalizer_Scope_2871 with Inline_Always;
      Var_Expr : Internal_Entity_Defining_Name;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Internal_Entity_Defining_Name_Array_Access;
Singleton : Internal_Entity_Defining_Name_Array_Access;
Concat_Result : Internal_Entity_Defining_Name_Array_Access;
Cast_Expr : Internal_Entity_Defining_Name;
Cast_Result : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Defining_Name;
Cast_Result_1 : Internal_Entity;
Is_Equiv : Boolean;
If_Result : Internal_Entity_Defining_Name_Array_Access;
Scope_Result : Internal_Entity_Defining_Name_Array_Access;
Result_Var : Internal_Entity_Defining_Name_Array_Access;
Let_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1091 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1091;
            procedure Finalizer_Scope_2871 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Singleton);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (If_Result);
            end Finalizer_Scope_2871;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9081 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9076 '<FieldAccess for previous_part at ast.py:15158>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15158
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.previous_part
Fld := Libadalang.Implementation.Defining_Name_P_Previous_Part (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 9076
Var_Expr := Fld;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
   --# bind pp Var_Expr
--# expr-start 9080 '<If at ast.py:15159>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15159
--# expr-start 9077 '<Eq at ast.py:15160>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15160
Cast_Expr := Ent; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Cast_Expr_1 := Var_Expr; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Is_Equiv := Equivalent (Cast_Result, Cast_Result_1); 
--# expr-done 9077
if Is_Equiv then
   If_Result := No_Internal_Entity_Defining_Name_Array_Type;
else
   --# expr-start 9079 '<Concat at ast.py:15162>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15162
--# expr-start 9078 '<FieldAccess for all_previous_parts at ast.py:15162>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15162
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.all_previous_parts
Fld_1 := Libadalang.Implementation.Defining_Name_P_All_Previous_Parts (Node => Var_Expr.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr.Info);
--# end
--# expr-done 9078
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
Concat_Result := Concat (Fld_1, Singleton); 
--# expr-done 9079
   If_Result := Concat_Result;
end if;
      Inc_Ref (If_Result);
--# expr-done 9080
Scope_Result := If_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2871;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Defining_Name_Array_Type;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 9081
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1091;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1091;
               Finalizer_Scope_2871;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_All_Previous_Parts;
--# end
--# property-start DefiningName.all_next_parts /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15168
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_All_Next_Parts
  (Node : Bare_Defining_Name
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1092 with Inline_Always;
            procedure Finalizer_Scope_2872 with Inline_Always;
      Var_Expr : Internal_Entity_Defining_Name;
Fld : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
Fld_1 : Internal_Entity_Defining_Name_Array_Access;
Concat_Result : Internal_Entity_Defining_Name_Array_Access;
Cast_Expr : Internal_Entity_Defining_Name;
Cast_Result : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Defining_Name;
Cast_Result_1 : Internal_Entity;
Is_Equiv : Boolean;
If_Result : Internal_Entity_Defining_Name_Array_Access;
Scope_Result : Internal_Entity_Defining_Name_Array_Access;
Result_Var : Internal_Entity_Defining_Name_Array_Access;
Let_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1092 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1092;
            procedure Finalizer_Scope_2872 is
            begin
                     Dec_Ref (Singleton);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (If_Result);
            end Finalizer_Scope_2872;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9087 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9082 '<FieldAccess for next_part at ast.py:15173>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15173
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.next_part
Fld := Libadalang.Implementation.Defining_Name_P_Next_Part (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 9082
Var_Expr := Fld;
if Var_Expr /= No_Entity_Defining_Name then
   --# scope-start
   --# bind np Var_Expr
--# expr-start 9086 '<If at ast.py:15174>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15174
--# expr-start 9083 '<Eq at ast.py:15175>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15175
Cast_Expr := Ent; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Cast_Expr_1 := Var_Expr; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Is_Equiv := Equivalent (Cast_Result, Cast_Result_1); 
--# expr-done 9083
if Is_Equiv then
   If_Result := No_Internal_Entity_Defining_Name_Array_Type;
else
   --# expr-start 9085 '<Concat at ast.py:15177>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15177
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
--# expr-start 9084 '<FieldAccess for all_next_parts at ast.py:15177>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15177
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.all_next_parts
Fld_1 := Libadalang.Implementation.Defining_Name_P_All_Next_Parts (Node => Var_Expr.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr.Info);
--# end
--# expr-done 9084
Concat_Result := Concat (Singleton, Fld_1); 
--# expr-done 9085
   If_Result := Concat_Result;
end if;
      Inc_Ref (If_Result);
--# expr-done 9086
Scope_Result := If_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2872;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Defining_Name_Array_Type;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 9087
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1092;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1092;
               Finalizer_Scope_2872;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_All_Next_Parts;
--# end
--# property-start DefiningName.all_parts /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15183
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_All_Parts
  (Node : Bare_Defining_Name
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1093 with Inline_Always;
      Block_Var : Internal_Entity_Defining_Name_Array_Access;
Block_Var_1 : Internal_Entity_Defining_Name_Array_Access;
Fld : Internal_Entity_Defining_Name_Array_Access;
Fld_1 : Internal_Entity_Defining_Name_Array_Access;
Singleton : Internal_Entity_Defining_Name_Array_Access;
Concat_Result : Internal_Entity_Defining_Name_Array_Access;
Concat_Result_1 : Internal_Entity_Defining_Name_Array_Access;
Let_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_1093 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Singleton);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Concat_Result_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1093;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9092 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9088 '<FieldAccess for all_previous_parts at ast.py:15188>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15188
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.all_previous_parts
Fld := Libadalang.Implementation.Defining_Name_P_All_Previous_Parts (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 9088
Block_Var := Fld; Inc_Ref (Block_Var);
--# bind prevs Block_Var
--# expr-start 9089 '<FieldAccess for all_next_parts at ast.py:15189>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15189
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.all_next_parts
Fld_1 := Libadalang.Implementation.Defining_Name_P_All_Next_Parts (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 9089
Block_Var_1 := Fld_1; Inc_Ref (Block_Var_1);
--# bind nexts Block_Var_1
--# expr-start 9091 '<Concat at ast.py:15190>' Concat_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15190
--# expr-start 9090 '<Concat at ast.py:15190>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15190
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Ent;
Concat_Result := Concat (Block_Var, Singleton); 
--# expr-done 9090
Concat_Result_1 := Concat (Concat_Result, Block_Var_1); 
--# expr-done 9091
Let_Result := Concat_Result_1; Inc_Ref (Let_Result);
--# end
--# expr-done 9092
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1093;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1093;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_All_Parts;
--# end
--# property-start DefiningName.get_aspect_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15194
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Get_Aspect_Impl
  (Node : Bare_Defining_Name
      ; Name : Symbol_Type
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Aspect
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind name Name
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Aspect;
      Var_Expr : Internal_Aspect;
Var_Expr_1 : Internal_Aspect;
Var_Expr_2 : Internal_Aspect;
Var_Expr_3 : Internal_Entity_Pragma_Node;
Var_Expr_4 : Internal_Entity_Base_Assoc;
Var_Expr_5 : Internal_Entity_Base_Assoc_List;
Var_Expr_6 : Internal_Entity_Aspect_Assoc;
Var_Expr_7 : Internal_Entity_Attribute_Def_Clause;
Var_Expr_8 : Internal_Entity_At_Clause;
Fld : Internal_Entity_Pragma_Node;
Fld_1 : Internal_Entity_Base_Assoc_List;
Saved : Internal_Entity_Base_Assoc_List;
Fld_2 : Bare_Base_Assoc_List;
Fld_3 : Internal_Entity_Info;
Get_Result : Bare_Base_Assoc;
As_Entity : Internal_Entity_Base_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Base_Assoc;
Scope_Result : Internal_Entity_Base_Assoc;
Result_Var : Internal_Entity_Base_Assoc;
Fld_4 : Internal_Entity_Expr;
Scope_Result_1 : Internal_Entity_Expr;
Result_Var_1 : Internal_Entity_Expr;
Cast_Expr : Internal_Entity_Pragma_Node;
Cast_Result : Internal_Entity;
New_Struct : Internal_Aspect;
Scope_Result_2 : Internal_Aspect;
Result_Var_2 : Internal_Aspect;
Scope_Result_3 : Internal_Aspect;
Fld_5 : Internal_Entity_Basic_Decl;
Fld_6 : Internal_Entity_Aspect_Assoc;
Fld_7 : Internal_Entity_Expr;
Cast_Expr_1 : Internal_Entity_Aspect_Assoc;
Cast_Result_1 : Internal_Entity;
New_Struct_1 : Internal_Aspect;
Scope_Result_4 : Internal_Aspect;
Result_Var_3 : Internal_Aspect;
Result_Var_4 : Internal_Aspect;
Scope_Result_5 : Internal_Aspect;
Fld_8 : Internal_Entity_Attribute_Def_Clause;
Fld_9 : Internal_Entity_Expr;
Cast_Expr_2 : Internal_Entity_Attribute_Def_Clause;
Cast_Result_2 : Internal_Entity;
New_Struct_2 : Internal_Aspect;
Scope_Result_6 : Internal_Aspect;
Result_Var_5 : Internal_Aspect;
Result_Var_6 : Internal_Aspect;
Scope_Result_7 : Internal_Aspect;
Fld_10 : Internal_Entity_At_Clause;
Fld_11 : Internal_Entity_Expr;
Cast_Expr_3 : Internal_Entity_At_Clause;
Cast_Result_3 : Internal_Entity;
New_Struct_3 : Internal_Aspect;
Scope_Result_8 : Internal_Aspect;
Result_Var_7 : Internal_Aspect;
Sym : Symbol_Type;
Is_Equal : Boolean;
If_Result_1 : Internal_Aspect;
Result_Var_8 : Internal_Aspect;
Let_Result : Internal_Aspect;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9110 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9093 '<FieldAccess for get_pragma at ast.py:15199>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15199
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.get_pragma
Fld := Libadalang.Implementation.Defining_Name_P_Get_Pragma (Node => Ent.Node, Name => Name, E_Info => Ent.Info);
--# end
--# expr-done 9093
Var_Expr_3 := Fld;
if Var_Expr_3 /= No_Entity_Pragma_Node then
   --# scope-start
   --# bind p Var_Expr_3
--# expr-start 9096 '<New Aspect at ast.py:15200>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15200
Cast_Expr := Var_Expr_3; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
--# expr-start 9094 '<FieldAccess for args at ast.py:15201>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15201
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Base_Assoc_List (Node => Var_Expr_3.Node.Pragma_Node_F_Args, Info => Var_Expr_3.Info);
--# expr-done 9094
Var_Expr_5 := Fld_1;
if Var_Expr_5 /= No_Entity_Base_Assoc_List then
   --# scope-start
Saved := Var_Expr_5; 
Fld_2 := Saved.Node;
Get_Result := Get (Fld_2, 1, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Base_Assoc;
else
Fld_3 := Saved.Info;
As_Entity := (Info => Fld_3, Node => Get_Result); 
   If_Result := As_Entity;
end if;
Scope_Result := If_Result;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Assoc;
end if;
Var_Expr_4 := Result_Var;
if Var_Expr_4 /= No_Entity_Base_Assoc then
   --# scope-start
--# expr-start 9095 '<FieldAccess for assoc_expr at ast.py:15201>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15201
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseAssoc.assoc_expr'
Fld_4 := Libadalang.Implementation.Dispatcher_Base_Assoc_P_Assoc_Expr (Node => Var_Expr_4.Node, E_Info => Var_Expr_4.Info);
--# end
--# expr-done 9095
Scope_Result_1 := Fld_4;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Expr;
end if;
New_Struct := (Exists => True, Node => Cast_Result, Value => Result_Var_1); 
--# expr-done 9096
Scope_Result_2 := New_Struct;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Aspect;
end if;
Var_Expr_2 := Result_Var_2;
if Var_Expr_2 /= No_Aspect then
   --# scope-start
   --# bind e Var_Expr_2
Scope_Result_3 := Var_Expr_2;
   --# end
   Result_Var_4 := Scope_Result_3;
else
--# expr-start 9098 '<FieldAccess for get_aspect_assoc at ast.py:15203>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15203
--# expr-start 9097 '<FieldAccess for basic_decl at ast.py:15203>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15203
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl
Fld_5 := Libadalang.Implementation.Defining_Name_P_Basic_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 9097
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.get_aspect_assoc'
Fld_6 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Get_Aspect_Assoc (Node => Fld_5.Node, Name => Name, E_Info => Fld_5.Info);
--# end
--# expr-done 9098
Var_Expr_6 := Fld_6;
if Var_Expr_6 /= No_Entity_Aspect_Assoc then
   --# scope-start
   --# bind aa Var_Expr_6
--# expr-start 9100 '<New Aspect at ast.py:15204>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15204
Cast_Expr_1 := Var_Expr_6; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
--# expr-start 9099 '<FieldAccess for expr at ast.py:15204>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15204
      if Var_Expr_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Expr (Node => Var_Expr_6.Node.Aspect_Assoc_F_Expr, Info => Var_Expr_6.Info);
--# expr-done 9099
New_Struct_1 := (Exists => True, Node => Cast_Result_1, Value => Fld_7); 
--# expr-done 9100
Scope_Result_4 := New_Struct_1;
   --# end
   Result_Var_3 := Scope_Result_4;
else
   Result_Var_3 := No_Aspect;
end if;
   Result_Var_4 := Result_Var_3;
end if;
Var_Expr_1 := Result_Var_4;
if Var_Expr_1 /= No_Aspect then
   --# scope-start
   --# bind e Var_Expr_1
Scope_Result_5 := Var_Expr_1;
   --# end
   Result_Var_6 := Scope_Result_5;
else
--# expr-start 9101 '<FieldAccess for get_representation_clause at ast.py:15205>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15205
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.get_representation_clause
Fld_8 := Libadalang.Implementation.Defining_Name_P_Get_Representation_Clause (Node => Ent.Node, Name => Name, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 9101
Var_Expr_7 := Fld_8;
if Var_Expr_7 /= No_Entity_Attribute_Def_Clause then
   --# scope-start
   --# bind rc Var_Expr_7
--# expr-start 9103 '<New Aspect at ast.py:15206>' New_Struct_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15206
Cast_Expr_2 := Var_Expr_7; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
--# expr-start 9102 '<FieldAccess for expr at ast.py:15206>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15206
      if Var_Expr_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Expr (Node => Var_Expr_7.Node.Attribute_Def_Clause_F_Expr, Info => Var_Expr_7.Info);
--# expr-done 9102
New_Struct_2 := (Exists => True, Node => Cast_Result_2, Value => Fld_9); 
--# expr-done 9103
Scope_Result_6 := New_Struct_2;
   --# end
   Result_Var_5 := Scope_Result_6;
else
   Result_Var_5 := No_Aspect;
end if;
   Result_Var_6 := Result_Var_5;
end if;
Var_Expr := Result_Var_6;
if Var_Expr /= No_Aspect then
   --# scope-start
   --# bind e Var_Expr
Scope_Result_7 := Var_Expr;
   --# end
   Result_Var_8 := Scope_Result_7;
else
   --# expr-start 9109 '<If at ast.py:15207>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15207
--# expr-start 9105 '<Eq at ast.py:15208>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15208
--# expr-start 9104 '<Symbol Address at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Address); 
--# expr-done 9104
Is_Equal := Name = Sym; 
--# expr-done 9105
if Is_Equal then
--# expr-start 9106 '<FieldAccess for get_at_clause at ast.py:15209>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15209
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.get_at_clause
Fld_10 := Libadalang.Implementation.Defining_Name_P_Get_At_Clause (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 9106
Var_Expr_8 := Fld_10;
if Var_Expr_8 /= No_Entity_At_Clause then
   --# scope-start
   --# bind atc Var_Expr_8
--# expr-start 9108 '<New Aspect at ast.py:15210>' New_Struct_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15210
Cast_Expr_3 := Var_Expr_8; 
      Cast_Result_3 := Create_Internal_Entity
        (Node => Cast_Expr_3.Node,
         Info => Cast_Expr_3.Info);
--# expr-start 9107 '<FieldAccess for expr at ast.py:15210>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15210
      if Var_Expr_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Create_Internal_Entity_Expr (Node => Var_Expr_8.Node.At_Clause_F_Expr, Info => Var_Expr_8.Info);
--# expr-done 9107
New_Struct_3 := (Exists => True, Node => Cast_Result_3, Value => Fld_11); 
--# expr-done 9108
Scope_Result_8 := New_Struct_3;
   --# end
   Result_Var_7 := Scope_Result_8;
else
   Result_Var_7 := No_Aspect;
end if;
   If_Result_1 := Result_Var_7;
else
   If_Result_1 := No_Aspect;
end if;
--# expr-done 9109
   Result_Var_8 := If_Result_1;
end if;
Let_Result := Result_Var_8; 
--# end
--# expr-done 9110
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Get_Aspect_Impl;
--# end
--# property-start DefiningName.get_aspect /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15217
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Get_Aspect
  (Node : Bare_Defining_Name
      ; Name : Symbol_Type
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Aspect
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind name Name
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Aspect;
            procedure Finalizer_Scope_1095 with Inline_Always;
      Block_Var : Internal_Entity_Defining_Name_Array_Access;
Fld : Internal_Entity_Defining_Name_Array_Access;
Singleton : Internal_Entity_Defining_Name_Array_Access;
Sym : Symbol_Type;
Is_Equal : Boolean;
Sym_1 : Symbol_Type;
Is_Equal_1 : Boolean;
If_Result : Boolean;
Sym_2 : Symbol_Type;
Is_Equal_2 : Boolean;
If_Result_1 : Boolean;
If_Result_2 : Internal_Entity_Defining_Name_Array_Access;
Fld_1 : Internal_Aspect;
Map_Result : Internal_Aspect_Array_Access;
Fld_2 : Boolean;
Find_Result : Internal_Aspect;
Let_Result : Internal_Aspect;
            procedure Finalizer_Scope_1095 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Singleton);
                     Dec_Ref (If_Result_2);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_1095;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9123 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9118 '<If at ast.py:15231>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15231
--# expr-start 9112 '<Eq at ast.py:15232>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15232
--# expr-start 9111 '<Symbol Inline at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Inline); 
--# expr-done 9111
Is_Equal := Name = Sym; 
--# expr-done 9112
if Is_Equal then
   If_Result := True;
else
   --# expr-start 9114 '<Eq at ast.py:15232>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15232
--# expr-start 9113 '<Symbol Ghost at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Ghost); 
--# expr-done 9113
Is_Equal_1 := Name = Sym_1; 
--# expr-done 9114
   If_Result := Is_Equal_1;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 9116 '<Eq at ast.py:15232>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15232
--# expr-start 9115 '<Symbol Default_Initial_Condition at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Default_Initial_Condition); 
--# expr-done 9115
Is_Equal_2 := Name = Sym_2; 
--# expr-done 9116
   If_Result_1 := Is_Equal_2;
end if;
if If_Result_1 then
   --# expr-start 9117 '<FieldAccess for all_parts at ast.py:15239>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15239
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.all_parts
Fld := Libadalang.Implementation.Defining_Name_P_All_Parts (Node => Ent.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 9117
   If_Result_2 := Fld;
else
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Ent;
   If_Result_2 := Singleton;
end if;
      Inc_Ref (If_Result_2);
--# expr-done 9118
Block_Var := If_Result_2; Inc_Ref (Block_Var);
--# bind parts_to_check Block_Var
--# expr-start 9122 '<Find at ast.py:15242>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15242
   --# expr-start 9120 '<Map at ast.py:15242>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15242
      declare
         Map_Result_Vec : Internal_Aspect_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Block_Var;
      begin
         for Item_277 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_277
      --# expr-start 9119 '<FieldAccess for get_aspect_impl at ast.py:15243>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15243
      if Item_277.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.get_aspect_impl
Fld_1 := Libadalang.Implementation.Defining_Name_P_Get_Aspect_Impl (Node => Item_277.Node, Name => Name, Imprecise_Fallback => Imprecise_Fallback, E_Info => Item_277.Info);
--# end
--# expr-done 9119
         declare
            Item_To_Append : constant Internal_Aspect := Fld_1;
         begin
            Internal_Aspect_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Aspect_Array
           (Items_Count => Natural (Internal_Aspect_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Aspect_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Aspect_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Aspect_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 9120
      Find_Result := No_Aspect;
      declare
         Collection : constant Internal_Aspect_Array_Access := Map_Result;
      begin
         for Item_276 of
               Collection.Items
         loop
   --# scope-start
               --# bind a Item_276
      --# expr-start 9121 '<FieldAccess for exists at ast.py:15245>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15245
Fld_2 := Item_276.Exists;
--# expr-done 9121
      if Fld_2 then
         Find_Result := Item_276;
      end if;
   --# end
               exit when Fld_2;
         end loop;
      end;
--# expr-done 9122
Let_Result := Find_Result; 
--# end
--# expr-done 9123
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1095;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1095;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Get_Aspect;
--# end
--# property-start DefiningName.has_aspect /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15250
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Has_Aspect
  (Node : Bare_Defining_Name
      ; Name : Symbol_Type
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind name Name
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
            procedure Finalizer_Scope_2885 with Inline_Always;
      Block_Var : Internal_Aspect;
Var_Expr : Internal_Entity_Expr;
Fld : Internal_Aspect;
Fld_1 : Boolean;
Fld_2 : Internal_Entity_Expr;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Cast_Result_1 : Internal_Entity;
Is_Equiv : Boolean;
Not_Val : Boolean;
Fld_5 : Big_Integer_Type;
Big_Int : Big_Integer_Type;
Is_Equal : Boolean;
If_Result : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Fld_6 : Boolean;
If_Result_1 : Boolean;
If_Result_2 : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_2885 is
            begin
                     Dec_Ref (Fld_5);
                     Dec_Ref (Big_Int);
            end Finalizer_Scope_2885;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9136 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9124 '<FieldAccess for get_aspect at ast.py:15257>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15257
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.get_aspect
Fld := Libadalang.Implementation.Defining_Name_P_Get_Aspect (Node => Ent.Node, Name => Name, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 9124
Block_Var := Fld; 
--# bind a Block_Var
--# expr-start 9125 '<FieldAccess for exists at ast.py:15259>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15259
Fld_1 := Block_Var.Exists;
--# expr-done 9125
if Fld_1 then
   --# expr-start 9135 '<If at ast.py:15259>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15259
--# expr-start 9126 '<FieldAccess for is_contract_aspect at ast.py:15260>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15260
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.is_contract_aspect
Fld_6 := Libadalang.Implementation.Ada_Node_P_Is_Contract_Aspect (Node => Self, Name => Name);
--# end
--# expr-done 9126
if Fld_6 then
   If_Result_1 := True;
else
--# expr-start 9127 '<FieldAccess for value at ast.py:15266>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15266
Fld_2 := Block_Var.Value;
--# expr-done 9127
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind val Var_Expr
--# expr-start 9131 '<Not at ast.py:15270>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15270
--# expr-start 9130 '<Eq at ast.py:15270>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15270
--# expr-start 9128 '<FieldAccess for expression_type at ast.py:15270>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15270
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.expression_type
Fld_3 := Libadalang.Implementation.Expr_P_Expression_Type (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 9128
      Cast_Result := Create_Internal_Entity
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
--# expr-start 9129 '<FieldAccess for bool_type at ast.py:15270>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15270
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_4 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 9129
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_4.Node,
         Info => Fld_4.Info);
Is_Equiv := Equivalent (Cast_Result, Cast_Result_1); 
--# expr-done 9130
Not_Val := not (Is_Equiv); 
--# expr-done 9131
if Not_Val then
   If_Result := True;
else
   --# expr-start 9134 '<Eq at ast.py:15271>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15271
--# expr-start 9132 '<FieldAccess for eval_as_int at ast.py:15271>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15271
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.eval_as_int
Fld_5 := Libadalang.Implementation.Expr_P_Eval_As_Int (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 9132
--# expr-start 9133 '<BigInteger 1 at ast.py:15271>' Big_Int /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15271
Big_Int := Create_Big_Integer ("1"); 
--# expr-done 9133
Is_Equal := Equivalent (Fld_5, Big_Int); 
--# expr-done 9134
   If_Result := Is_Equal;
end if;
Scope_Result := If_Result;
   --# end
      Finalizer_Scope_2885;
   Result_Var := Scope_Result;
else
   Result_Var := True;
end if;
   If_Result_1 := Result_Var;
end if;
--# expr-done 9135
   If_Result_2 := If_Result_1;
else
   If_Result_2 := False;
end if;
Let_Result := If_Result_2; 
--# end
--# expr-done 9136
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_2885;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Has_Aspect;
--# end
--# property-start DefiningName.is_valid_pragma_for_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15276
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Is_Valid_Pragma_For_Name
  (Node : Bare_Defining_Name
      ; Name : Symbol_Type
      ; Decl : Internal_Entity
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind name Name
   --# bind decl Decl
   Property_Result : Boolean;
            procedure Finalizer_Scope_2886 with Inline_Always;
      Var_Expr : Internal_Entity_Pragma_Node;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Pragma_Node;
Fld : Internal_Entity_Identifier;
Fld_1 : Boolean;
Fld_2 : Internal_Entity_Defining_Name_Array_Access;
Cast_Expr_1 : Internal_Entity_Defining_Name;
Cast_Result_1 : Internal_Entity;
Cast_Expr_2 : Internal_Entity_Defining_Name;
Cast_Result_2 : Internal_Entity;
Is_Equiv : Boolean;
Find_Result : Internal_Entity_Defining_Name;
Is_Null : Boolean;
Not_Val : Boolean;
If_Result : Boolean;
Fld_3 : Bare_Pragma_Node;
Node_Comp : Boolean;
If_Result_1 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_2886 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_2886;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9147 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9137 '<Cast to Pragma at ast.py:15282>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15282
Cast_Expr := Decl; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Pragma_Node_Range
   then
      Cast_Result := Create_Internal_Entity_Pragma_Node
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Pragma_Node;
   end if;
--# expr-done 9137
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Pragma_Node then
   --# scope-start
   --# bind p Var_Expr
--# expr-start 9139 '<FieldAccess for name_is at ast.py:15285>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15285
--# expr-start 9138 '<FieldAccess for id at ast.py:15285>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15285
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Identifier (Node => Var_Expr.Node.Pragma_Node_F_Id, Info => Var_Expr.Info);
--# expr-done 9138
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_is
Fld_1 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld.Node, Sym => Name);
--# end
--# expr-done 9139
if Fld_1 then
   --# expr-start 9144 '<Not at ast.py:15287>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15287
--# expr-start 9143 '<IsNull at ast.py:15287>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15287
--# expr-start 9142 '<Find at ast.py:15287>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15287
   --# expr-start 9140 '<FieldAccess for associated_entities at ast.py:15287>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15287
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Pragma.associated_entities
Fld_2 := Libadalang.Implementation.Pragma_Node_P_Associated_Entities (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 9140
      Find_Result := No_Entity_Defining_Name;
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Fld_2;
      begin
         for Item_278 of
               Collection.Items
         loop
   --# scope-start
               --# bind d Item_278
      --# expr-start 9141 '<Eq at ast.py:15287>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15287
Cast_Expr_1 := Ent; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Cast_Expr_2 := Item_278; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
Is_Equiv := Equivalent (Cast_Result_1, Cast_Result_2); 
--# expr-done 9141
      if Is_Equiv then
         Find_Result := Item_278;
      end if;
   --# end
               exit when Is_Equiv;
         end loop;
      end;
--# expr-done 9142
Is_Null := Find_Result.Node = null; 
--# expr-done 9143
Not_Val := not (Is_Null); 
--# expr-done 9144
   If_Result := Not_Val;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 9146 '<OrderingTest '"'"'lt'"'"' at ast.py:15290>' Node_Comp /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15290
--# expr-start 9145 '<FieldAccess for node at ast.py:15290>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15290
Fld_3 := Var_Expr.Node;
--# expr-done 9145
Node_Comp := Compare (Self, Fld_3, Less_Than); 
--# expr-done 9146
   If_Result_1 := Node_Comp;
else
   If_Result_1 := False;
end if;
Scope_Result := If_Result_1;
   --# end
      Finalizer_Scope_2886;
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 9147
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_2886;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Is_Valid_Pragma_For_Name;
--# end
--# property-start DefiningName.get_pragma /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15295
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Get_Pragma
  (Node : Bare_Defining_Name
      ; Name : Symbol_Type
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Pragma_Node
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind name Name
   Property_Result : Internal_Entity_Pragma_Node;
            procedure Finalizer_Scope_1098 with Inline_Always;
      Block_Var : Internal_Entity_Basic_Decl;
Var_Expr : Internal_Entity_Pragma_Node;
Var_Expr_1 : Internal_Entity_Pragma_Node_List;
Var_Expr_2 : Internal_Entity;
Var_Expr_3 : Internal_Entity;
Var_Expr_4 : Internal_Entity_Ada_Node_List;
Var_Expr_5 : Bare_Public_Part;
Var_Expr_6 : Bare_Private_Part;
Match_0 : Bare_Base_Package_Decl;
Match_1 : Bare_Protected_Def;
Match_2 : Bare_Ada_Node;
Var_Expr_7 : Internal_Entity_Declarative_Part;
Fld : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Pragma_Node_List;
Saved : Internal_Entity_Pragma_Node_List;
Fld_2 : Bare_Pragma_Node_List;
Fld_3 : Internal_Entity_Info;
As_Entity : Internal_Entity_Pragma_Node;
Is_Null : Boolean;
If_Result : Internal_Entity_Pragma_Node;
Fld_4 : Internal_Entity_Identifier;
Fld_5 : Boolean;
Item_279 : Internal_Entity_Pragma_Node;
Bare_Item_279 : Bare_Pragma_Node;
Find_Result : Internal_Entity_Pragma_Node;
Scope_Result : Internal_Entity_Pragma_Node;
Result_Var : Internal_Entity_Pragma_Node;
Scope_Result_1 : Internal_Entity_Pragma_Node;
Fld_6 : Internal_Entity;
Cast_Result : Internal_Entity_Ada_Node_List;
Saved_1 : Internal_Entity_Ada_Node_List;
Fld_7 : Bare_Ada_Node_List;
Fld_8 : Internal_Entity_Info;
As_Entity_1 : Internal_Entity;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity;
Fld_9 : Boolean;
Item_280 : Internal_Entity;
Bare_Item_280 : Bare_Ada_Node;
Find_Result_1 : Internal_Entity;
Scope_Result_2 : Internal_Entity;
Result_Var_1 : Internal_Entity;
Scope_Result_3 : Internal_Entity;
Fld_10 : Bare_Declarative_Part;
Cast_Result_1 : Bare_Public_Part;
Fld_11 : Bare_Ada_Node;
Fld_12 : Bare_Private_Part;
Fld_13 : Bare_Private_Part;
Match_Prefix : Bare_Ada_Node;
Cast_Result_2 : Bare_Base_Package_Decl;
Let_Result : Bare_Private_Part;
Scope_Result_4 : Bare_Private_Part;
Cast_Result_3 : Bare_Protected_Def;
Let_Result_1 : Bare_Private_Part;
Scope_Result_5 : Bare_Private_Part;
Cast_Result_4 : Bare_Ada_Node;
Let_Result_2 : Bare_Private_Part;
Scope_Result_6 : Bare_Private_Part;
Match_Result : Bare_Private_Part;
Fld_14 : Bare_Ada_Node_List;
Scope_Result_7 : Bare_Ada_Node_List;
Result_Var_2 : Bare_Ada_Node_List;
As_Entity_2 : Internal_Entity_Ada_Node_List;
Is_Null_2 : Boolean;
If_Result_2 : Internal_Entity_Ada_Node_List;
Fld_15 : Bare_Ada_Node_List;
Fld_16 : Internal_Entity_Info;
As_Entity_3 : Internal_Entity;
Is_Null_3 : Boolean;
If_Result_3 : Internal_Entity;
Fld_17 : Boolean;
Item_281 : Internal_Entity;
Bare_Item_281 : Bare_Ada_Node;
Find_Result_2 : Internal_Entity;
Scope_Result_8 : Internal_Entity;
Result_Var_3 : Internal_Entity;
Result_Var_4 : Internal_Entity;
Scope_Result_9 : Internal_Entity;
Fld_18 : Internal_Entity_Declarative_Part_Array_Access;
Get_Result : Internal_Entity_Declarative_Part;
Fld_19 : Internal_Entity_Ada_Node_List;
Scope_Result_10 : Internal_Entity_Ada_Node_List;
Result_Var_5 : Internal_Entity_Ada_Node_List;
Fld_20 : Bare_Ada_Node_List;
Fld_21 : Internal_Entity_Info;
As_Entity_4 : Internal_Entity;
Is_Null_4 : Boolean;
If_Result_4 : Internal_Entity;
Fld_22 : Boolean;
Item_282 : Internal_Entity;
Bare_Item_282 : Bare_Ada_Node;
Find_Result_3 : Internal_Entity;
Result_Var_6 : Internal_Entity;
Cast_Result_5 : Internal_Entity_Pragma_Node;
Result_Var_7 : Internal_Entity_Pragma_Node;
Let_Result_3 : Internal_Entity_Pragma_Node;
            procedure Finalizer_Scope_1098 is
            begin
                     Dec_Ref (Fld_18);
            end Finalizer_Scope_1098;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9173 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 9148 '<FieldAccess for basic_decl_no_internal at ast.py:15303>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15303
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl_no_internal
Fld := Libadalang.Implementation.Defining_Name_P_Basic_Decl_No_Internal (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 9148
Block_Var := Fld; 
--# bind bd Block_Var
--# expr-start 9149 '<FieldAccess for library_item_pragmas at ast.py:15305>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15305
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.library_item_pragmas
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Library_Item_Pragmas (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 9149
Var_Expr_1 := Fld_1;
if Var_Expr_1 /= No_Entity_Pragma_Node_List then
   --# scope-start
   --# bind plist Var_Expr_1
--# expr-start 9152 '<Find at ast.py:15307>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15307
Saved := Var_Expr_1; 
Fld_2 := Saved.Node;
      Find_Result := No_Entity_Pragma_Node;
      if Fld_2 = null then
      Find_Result := No_Entity_Pragma_Node;
      else
      declare
         Collection : constant Bare_Pragma_Node_List := Fld_2;
      begin
         for Untyped_Item_279 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_279 := Untyped_Item_279; 
Is_Null := Bare_Item_279 = null; 
if Is_Null then
   If_Result := No_Entity_Pragma_Node;
else
Fld_3 := Saved.Info;
As_Entity := (Info => Fld_3, Node => Bare_Item_279); 
   If_Result := As_Entity;
end if;
                  Item_279 := If_Result; 
   --# scope-start
               --# bind p Item_279
      --# expr-start 9151 '<FieldAccess for name_is at ast.py:15307>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15307
--# expr-start 9150 '<FieldAccess for id at ast.py:15307>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15307
      if Item_279.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Identifier (Node => Item_279.Node.Pragma_Node_F_Id, Info => Item_279.Info);
--# expr-done 9150
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_is
Fld_5 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_4.Node, Sym => Name);
--# end
--# expr-done 9151
      if Fld_5 then
         Find_Result := Item_279;
      end if;
   --# end
               exit when Fld_5;
         end loop;
      end;
      end if;
--# expr-done 9152
Scope_Result := Find_Result;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Pragma_Node;
end if;
Var_Expr := Result_Var;
if Var_Expr /= No_Entity_Pragma_Node then
   --# scope-start
   --# bind e Var_Expr
Scope_Result_1 := Var_Expr;
   --# end
   Result_Var_7 := Scope_Result_1;
else
   --# expr-start 9172 '<Cast to Pragma at ast.py:15313>' Cast_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15313
--# expr-start 9154 '<Cast to AdaNode.list at ast.py:15313>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15313
--# expr-start 9153 '<parent at ast.py:15313>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15313
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Parent (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# expr-done 9153
   if Fld_6.Node = null
      or else Fld_6.Node.Kind in Ada_Ada_Node_List_Range
   then
      Cast_Result := Create_Internal_Entity_Ada_Node_List
        (Node => Fld_6.Node,
         Info => Fld_6.Info);
   else
         Cast_Result := No_Entity_Ada_Node_List;
   end if;
--# expr-done 9154
Var_Expr_4 := Cast_Result;
if Var_Expr_4 /= No_Entity_Ada_Node_List then
   --# scope-start
--# expr-start 9156 '<Find at ast.py:15313>' Find_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15313
Saved_1 := Var_Expr_4; 
Fld_7 := Saved_1.Node;
      Find_Result_1 := No_Entity;
      if Fld_7 = null then
      Find_Result_1 := No_Entity;
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_7;
      begin
         for Untyped_Item_280 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_280 := Untyped_Item_280; 
Is_Null_1 := Bare_Item_280 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity;
else
Fld_8 := Saved_1.Info;
As_Entity_1 := (Info => Fld_8, Node => Bare_Item_280); 
   If_Result_1 := As_Entity_1;
end if;
                  Item_280 := If_Result_1; 
   --# scope-start
               --# bind d Item_280
      --# expr-start 9155 '<FieldAccess for is_valid_pragma_for_name at ast.py:15314>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15314
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.is_valid_pragma_for_name
Fld_9 := Libadalang.Implementation.Defining_Name_P_Is_Valid_Pragma_For_Name (Node => Ent.Node, Name => Name, Decl => Item_280, E_Info => Ent.Info);
--# end
--# expr-done 9155
      if Fld_9 then
         Find_Result_1 := Item_280;
      end if;
   --# end
               exit when Fld_9;
         end loop;
      end;
      end if;
--# expr-done 9156
Scope_Result_2 := Find_Result_1;
   --# end
   Result_Var_1 := Scope_Result_2;
else
   Result_Var_1 := No_Entity;
end if;
Var_Expr_3 := Result_Var_1;
if Var_Expr_3 /= No_Entity then
   --# scope-start
   --# bind e Var_Expr_3
Scope_Result_3 := Var_Expr_3;
   --# end
   Result_Var_4 := Scope_Result_3;
else
--# expr-start 9158 '<Cast to PublicPart at ast.py:15319>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15319
--# expr-start 9157 '<FieldAccess for declarative_scope at ast.py:15319>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15319
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.declarative_scope
Fld_10 := Libadalang.Implementation.Ada_Node_P_Declarative_Scope (Node => Block_Var.Node);
--# end
--# expr-done 9157
   if Fld_10 = null
      or else Fld_10.Kind in Ada_Public_Part_Range
   then
      Cast_Result_1 := Fld_10;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 9158
Var_Expr_5 := Cast_Result_1;
if Var_Expr_5 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind pp Var_Expr_5
--# expr-start 9166 '<Find at ast.py:15320>' Find_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15320
   --# expr-start 9164 '<as_entity at ast.py:15320>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15320
--# expr-start 9162 '<Match at ast.py:15320>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15320
--# expr-start 9159 '<parent at ast.py:15320>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15320
      if Var_Expr_5 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Var_Expr_5.Parent; 
--# expr-done 9159
      if Fld_11 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld_11; 
case Ada_Ada_Node (Match_Prefix.Kind) is
         when Ada_Generic_Package_Internal .. Ada_Package_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Match_Prefix;
Match_0 := Cast_Result_2; 
--# bind pkg Match_0
--# expr-start 9160 '<FieldAccess for private_part at ast.py:15321>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15321
      if Match_0 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Match_0.Base_Package_Decl_F_Private_Part;
--# expr-done 9160
Let_Result := Fld_12; 
--# end
Scope_Result_4 := Let_Result;
   --# end
            Match_Result := Scope_Result_4; 
         when Ada_Protected_Def =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Match_Prefix;
Match_1 := Cast_Result_3; 
--# bind ptd Match_1
--# expr-start 9161 '<FieldAccess for private_part at ast.py:15322>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15322
      if Match_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Match_1.Protected_Def_F_Private_Part;
--# expr-done 9161
Let_Result_1 := Fld_13; 
--# end
Scope_Result_5 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_5; 
         when Ada_Abort_Absent .. Ada_Synthetic_Formal_Param_Decl | Ada_Discrete_Base_Subtype_Decl .. Ada_Private_Present | Ada_Protected_Absent .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Match_Prefix;
Match_2 := Cast_Result_4; 
--# bind _ Match_2
Let_Result_2 := No_Bare_Ada_Node; 
--# end
Scope_Result_6 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_6; 
end case;
--# expr-done 9162
Var_Expr_6 := Match_Result;
if Var_Expr_6 /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 9163 '<FieldAccess for decls at ast.py:15320>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15320
      if Var_Expr_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Var_Expr_6.Declarative_Part_F_Decls;
--# expr-done 9163
Scope_Result_7 := Fld_14;
   --# end
   Result_Var_2 := Scope_Result_7;
else
   Result_Var_2 := No_Bare_Ada_Node;
end if;
Is_Null_2 := Result_Var_2 = null; 
if Is_Null_2 then
   If_Result_2 := No_Entity_Ada_Node_List;
else
As_Entity_2 := (Info => E_Info, Node => Result_Var_2); 
   If_Result_2 := As_Entity_2;
end if;
--# expr-done 9164
Fld_15 := If_Result_2.Node;
      Find_Result_2 := No_Entity;
      if Fld_15 = null then
      Find_Result_2 := No_Entity;
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_15;
      begin
         for Untyped_Item_281 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_281 := Untyped_Item_281; 
Is_Null_3 := Bare_Item_281 = null; 
if Is_Null_3 then
   If_Result_3 := No_Entity;
else
Fld_16 := If_Result_2.Info;
As_Entity_3 := (Info => Fld_16, Node => Bare_Item_281); 
   If_Result_3 := As_Entity_3;
end if;
                  Item_281 := If_Result_3; 
   --# scope-start
               --# bind d Item_281
      --# expr-start 9165 '<FieldAccess for is_valid_pragma_for_name at ast.py:15325>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15325
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.is_valid_pragma_for_name
Fld_17 := Libadalang.Implementation.Defining_Name_P_Is_Valid_Pragma_For_Name (Node => Ent.Node, Name => Name, Decl => Item_281, E_Info => Ent.Info);
--# end
--# expr-done 9165
      if Fld_17 then
         Find_Result_2 := Item_281;
      end if;
   --# end
               exit when Fld_17;
         end loop;
      end;
      end if;
--# expr-done 9166
Scope_Result_8 := Find_Result_2;
   --# end
   Result_Var_3 := Scope_Result_8;
else
   Result_Var_3 := No_Entity;
end if;
   Result_Var_4 := Result_Var_3;
end if;
Var_Expr_2 := Result_Var_4;
if Var_Expr_2 /= No_Entity then
   --# scope-start
   --# bind e Var_Expr_2
Scope_Result_9 := Var_Expr_2;
   --# end
   Result_Var_6 := Scope_Result_9;
else
   --# expr-start 9171 '<Find at ast.py:15330>' Find_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15330
--# expr-start 9168 '<at at ast.py:15330>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15330
--# expr-start 9167 '<FieldAccess for declarative_parts at ast.py:15330>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15330
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.declarative_parts'
Fld_18 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Declarative_Parts (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 9167
Get_Result := Get (Fld_18, 0, True); 
--# expr-done 9168
Var_Expr_7 := Get_Result;
if Var_Expr_7 /= No_Entity_Declarative_Part then
   --# scope-start
--# expr-start 9169 '<FieldAccess for decls at ast.py:15330>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15330
      if Var_Expr_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Create_Internal_Entity_Ada_Node_List (Node => Var_Expr_7.Node.Declarative_Part_F_Decls, Info => Var_Expr_7.Info);
--# expr-done 9169
Scope_Result_10 := Fld_19;
   --# end
   Result_Var_5 := Scope_Result_10;
else
   Result_Var_5 := No_Entity_Ada_Node_List;
end if;
Fld_20 := Result_Var_5.Node;
      Find_Result_3 := No_Entity;
      if Fld_20 = null then
      Find_Result_3 := No_Entity;
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_20;
      begin
         for Untyped_Item_282 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_282 := Untyped_Item_282; 
Is_Null_4 := Bare_Item_282 = null; 
if Is_Null_4 then
   If_Result_4 := No_Entity;
else
Fld_21 := Result_Var_5.Info;
As_Entity_4 := (Info => Fld_21, Node => Bare_Item_282); 
   If_Result_4 := As_Entity_4;
end if;
                  Item_282 := If_Result_4; 
   --# scope-start
               --# bind d Item_282
      --# expr-start 9170 '<FieldAccess for is_valid_pragma_for_name at ast.py:15331>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15331
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.is_valid_pragma_for_name
Fld_22 := Libadalang.Implementation.Defining_Name_P_Is_Valid_Pragma_For_Name (Node => Ent.Node, Name => Name, Decl => Item_282, E_Info => Ent.Info);
--# end
--# expr-done 9170
      if Fld_22 then
         Find_Result_3 := Item_282;
      end if;
   --# end
               exit when Fld_22;
         end loop;
      end;
      end if;
--# expr-done 9171
   Result_Var_6 := Find_Result_3;
end if;
   if Result_Var_6.Node = null
      or else Result_Var_6.Node.Kind in Ada_Pragma_Node_Range
   then
      Cast_Result_5 := Create_Internal_Entity_Pragma_Node
        (Node => Result_Var_6.Node,
         Info => Result_Var_6.Info);
   else
         Cast_Result_5 := No_Entity_Pragma_Node;
   end if;
--# expr-done 9172
   Result_Var_7 := Cast_Result_5;
end if;
Let_Result_3 := Result_Var_7; 
--# end
--# expr-done 9173
      Property_Result := Let_Result_3;
   --# end
      Finalizer_Scope_1098;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1098;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Get_Pragma;
--# end
--# property-start DefiningName.get_representation_clause /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15339
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Get_Representation_Clause
  (Node : Bare_Defining_Name
      ; Name : Symbol_Type
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Attribute_Def_Clause
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind name Name
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_Attribute_Def_Clause;
      Var_Expr : Bare_Declarative_Part;
Var_Expr_1 : Internal_Entity_Attribute_Def_Clause;
Attr : Internal_Entity_Attribute_Ref;
Fld : Bare_Declarative_Part;
Fld_1 : Bare_Ada_Node_List;
Scope_Result : Bare_Ada_Node_List;
Result_Var : Bare_Ada_Node_List;
As_Entity : Internal_Entity_Ada_Node_List;
Is_Null : Boolean;
If_Result : Internal_Entity_Ada_Node_List;
Fld_2 : Bare_Ada_Node_List;
Fld_3 : Internal_Entity_Info;
As_Entity_1 : Internal_Entity;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Attribute_Def_Clause;
Fld_4 : Internal_Entity_Name;
Cast_Result_1 : Internal_Entity_Attribute_Ref;
Fld_5 : Internal_Entity_Identifier;
Fld_6 : Boolean;
Fld_7 : Internal_Entity_Name;
Fld_8 : Internal_Entity_Defining_Name;
Cast_Result_2 : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Defining_Name;
Cast_Result_3 : Internal_Entity;
Is_Equiv : Boolean;
If_Result_2 : Boolean;
Let_Result : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
Item_283 : Internal_Entity;
Bare_Item_283 : Bare_Ada_Node;
Find_Result : Internal_Entity;
Cast_Result_4 : Internal_Entity_Attribute_Def_Clause;
Let_Result_1 : Internal_Entity_Attribute_Def_Clause;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9188 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 9187 '<Cast to AttributeDefClause.entity at ast.py:15344>' Cast_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15344
--# expr-start 9186 '<Find at ast.py:15344>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15344
   --# expr-start 9176 '<as_entity at ast.py:15344>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15344
--# expr-start 9174 '<FieldAccess for declarative_scope at ast.py:15344>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15344
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.declarative_scope
Fld := Libadalang.Implementation.Ada_Node_P_Declarative_Scope (Node => Ent.Node);
--# end
--# expr-done 9174
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 9175 '<FieldAccess for decls at ast.py:15344>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15344
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Var_Expr.Declarative_Part_F_Decls;
--# expr-done 9175
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node;
end if;
Is_Null := Result_Var = null; 
if Is_Null then
   If_Result := No_Entity_Ada_Node_List;
else
As_Entity := (Info => E_Info, Node => Result_Var); 
   If_Result := As_Entity;
end if;
--# expr-done 9176
Fld_2 := If_Result.Node;
      Find_Result := No_Entity;
      if Fld_2 = null then
      Find_Result := No_Entity;
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_2;
      begin
         for Untyped_Item_283 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_283 := Untyped_Item_283; 
Is_Null_1 := Bare_Item_283 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity;
else
Fld_3 := If_Result.Info;
As_Entity_1 := (Info => Fld_3, Node => Bare_Item_283); 
   If_Result_1 := As_Entity_1;
end if;
                  Item_283 := If_Result_1; 
   --# scope-start
               --# bind d Item_283
--# expr-start 9177 '<Cast to AttributeDefClause at ast.py:15345>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15345
Cast_Expr := Item_283; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Attribute_Def_Clause_Range
   then
      Cast_Result := Create_Internal_Entity_Attribute_Def_Clause
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Attribute_Def_Clause;
   end if;
--# expr-done 9177
Var_Expr_1 := Cast_Result;
if Var_Expr_1 /= No_Entity_Attribute_Def_Clause then
   --# scope-start
   --# bind p Var_Expr_1
--# expr-start 9185 '<Let at ast.py:15346>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15346
--# scope-start
--# expr-start 9179 '<Cast to AttributeRef at ast.py:15347>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15347
--# expr-start 9178 '<FieldAccess for attribute_expr at ast.py:15347>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15347
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Name (Node => Var_Expr_1.Node.Attribute_Def_Clause_F_Attribute_Expr, Info => Var_Expr_1.Info);
--# expr-done 9178
   if Fld_4.Node = null
      or else Fld_4.Node.Kind in Ada_Attribute_Ref_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Attribute_Ref
        (Node => Fld_4.Node,
         Info => Fld_4.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 9179
Attr := Cast_Result_1; 
--# bind attr Attr
--# expr-start 9181 '<FieldAccess for name_is at ast.py:15348>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15348
--# expr-start 9180 '<FieldAccess for attribute at ast.py:15348>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15348
      if Attr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Identifier (Node => Attr.Node.Attribute_Ref_F_Attribute, Info => Attr.Info);
--# expr-done 9180
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_is
Fld_6 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_5.Node, Sym => Name);
--# end
--# expr-done 9181
if Fld_6 then
   --# expr-start 9184 '<Eq at ast.py:15349>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15349
--# expr-start 9183 '<FieldAccess for referenced_defining_name at ast.py:15349>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15349
--# expr-start 9182 '<FieldAccess for prefix at ast.py:15349>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15349
      if Attr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Name (Node => Attr.Node.Attribute_Ref_F_Prefix, Info => Attr.Info);
--# expr-done 9182
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_defining_name
Fld_8 := Libadalang.Implementation.Name_P_Referenced_Defining_Name (Node => Fld_7.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_7.Info);
--# end
--# expr-done 9183
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_8.Node,
         Info => Fld_8.Info);
Cast_Expr_1 := Ent; 
      Cast_Result_3 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Is_Equiv := Equivalent (Cast_Result_2, Cast_Result_3); 
--# expr-done 9184
   If_Result_2 := Is_Equiv;
else
   If_Result_2 := False;
end if;
Let_Result := If_Result_2; 
--# end
--# expr-done 9185
Scope_Result_1 := Let_Result;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
      if Result_Var_1 then
         Find_Result := Item_283;
      end if;
   --# end
               exit when Result_Var_1;
         end loop;
      end;
      end if;
--# expr-done 9186
   if Find_Result.Node = null
      or else Find_Result.Node.Kind in Ada_Attribute_Def_Clause_Range
   then
      Cast_Result_4 := Create_Internal_Entity_Attribute_Def_Clause
        (Node => Find_Result.Node,
         Info => Find_Result.Info);
   else
         Cast_Result_4 := No_Entity_Attribute_Def_Clause;
   end if;
--# expr-done 9187
Let_Result_1 := Cast_Result_4; 
--# end
--# expr-done 9188
      Property_Result := Let_Result_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Get_Representation_Clause;
--# end
--# property-start DefiningName.get_at_clause /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15356
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Get_At_Clause
  (Node : Bare_Defining_Name
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_At_Clause
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Entity_At_Clause;
      Var_Expr : Bare_Declarative_Part;
Var_Expr_1 : Internal_Entity_At_Clause;
Fld : Bare_Declarative_Part;
Fld_1 : Bare_Ada_Node_List;
Scope_Result : Bare_Ada_Node_List;
Result_Var : Bare_Ada_Node_List;
As_Entity : Internal_Entity_Ada_Node_List;
Is_Null : Boolean;
If_Result : Internal_Entity_Ada_Node_List;
Fld_2 : Bare_Ada_Node_List;
Fld_3 : Internal_Entity_Info;
As_Entity_1 : Internal_Entity;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_At_Clause;
Fld_4 : Internal_Entity_Base_Id;
Fld_5 : Internal_Entity_Defining_Name;
Cast_Result_1 : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Defining_Name;
Cast_Result_2 : Internal_Entity;
Is_Equiv : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
Item_284 : Internal_Entity;
Bare_Item_284 : Bare_Ada_Node;
Find_Result : Internal_Entity;
Cast_Result_3 : Internal_Entity_At_Clause;
Let_Result : Internal_Entity_At_Clause;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9198 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9197 '<Cast to AtClause.entity at ast.py:15360>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15360
--# expr-start 9196 '<Find at ast.py:15360>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15360
   --# expr-start 9191 '<as_entity at ast.py:15360>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15360
--# expr-start 9189 '<FieldAccess for declarative_scope at ast.py:15360>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15360
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.declarative_scope
Fld := Libadalang.Implementation.Ada_Node_P_Declarative_Scope (Node => Ent.Node);
--# end
--# expr-done 9189
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 9190 '<FieldAccess for decls at ast.py:15360>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15360
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Var_Expr.Declarative_Part_F_Decls;
--# expr-done 9190
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node;
end if;
Is_Null := Result_Var = null; 
if Is_Null then
   If_Result := No_Entity_Ada_Node_List;
else
As_Entity := (Info => E_Info, Node => Result_Var); 
   If_Result := As_Entity;
end if;
--# expr-done 9191
Fld_2 := If_Result.Node;
      Find_Result := No_Entity;
      if Fld_2 = null then
      Find_Result := No_Entity;
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_2;
      begin
         for Untyped_Item_284 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_284 := Untyped_Item_284; 
Is_Null_1 := Bare_Item_284 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity;
else
Fld_3 := If_Result.Info;
As_Entity_1 := (Info => Fld_3, Node => Bare_Item_284); 
   If_Result_1 := As_Entity_1;
end if;
                  Item_284 := If_Result_1; 
   --# scope-start
               --# bind d Item_284
--# expr-start 9192 '<Cast to AtClause at ast.py:15361>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15361
Cast_Expr := Item_284; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_At_Clause_Range
   then
      Cast_Result := Create_Internal_Entity_At_Clause
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_At_Clause;
   end if;
--# expr-done 9192
Var_Expr_1 := Cast_Result;
if Var_Expr_1 /= No_Entity_At_Clause then
   --# scope-start
   --# bind p Var_Expr_1
--# expr-start 9195 '<Eq at ast.py:15362>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15362
--# expr-start 9194 '<FieldAccess for referenced_defining_name at ast.py:15362>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15362
--# expr-start 9193 '<FieldAccess for name at ast.py:15362>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15362
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Base_Id (Node => Var_Expr_1.Node.At_Clause_F_Name, Info => Var_Expr_1.Info);
--# expr-done 9193
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_defining_name
Fld_5 := Libadalang.Implementation.Name_P_Referenced_Defining_Name (Node => Fld_4.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_4.Info);
--# end
--# expr-done 9194
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
Cast_Expr_1 := Ent; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Is_Equiv := Equivalent (Cast_Result_1, Cast_Result_2); 
--# expr-done 9195
Scope_Result_1 := Is_Equiv;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
      if Result_Var_1 then
         Find_Result := Item_284;
      end if;
   --# end
               exit when Result_Var_1;
         end loop;
      end;
      end if;
--# expr-done 9196
   if Find_Result.Node = null
      or else Find_Result.Node.Kind in Ada_At_Clause_Range
   then
      Cast_Result_3 := Create_Internal_Entity_At_Clause
        (Node => Find_Result.Node,
         Info => Find_Result.Info);
   else
         Cast_Result_3 := No_Entity_At_Clause;
   end if;
--# expr-done 9197
Let_Result := Cast_Result_3; 
--# end
--# expr-done 9198
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Get_At_Clause;
--# end
--# property-start DefiningName.is_imported /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15367
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Is_Imported
  (Node : Bare_Defining_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Sym : Symbol_Type;
Fld : Boolean;
Sym_1 : Symbol_Type;
Fld_1 : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9203 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9200 '<FieldAccess for has_aspect at ast.py:15372>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15372
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9199 '<Symbol Import at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Import); 
--# expr-done 9199
--# property-call-start DefiningName.has_aspect
Fld := Libadalang.Implementation.Defining_Name_P_Has_Aspect (Node => Ent.Node, Name => Sym, Imprecise_Fallback => False, E_Info => Ent.Info);
--# end
--# expr-done 9200
if Fld then
   If_Result := True;
else
   --# expr-start 9202 '<FieldAccess for has_aspect at ast.py:15372>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15372
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9201 '<Symbol Interface at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Interface); 
--# expr-done 9201
--# property-call-start DefiningName.has_aspect
Fld_1 := Libadalang.Implementation.Defining_Name_P_Has_Aspect (Node => Ent.Node, Name => Sym_1, Imprecise_Fallback => False, E_Info => Ent.Info);
--# end
--# expr-done 9202
   If_Result := Fld_1;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 9203
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Is_Imported;
--# end
--# property-start DefiningName.is_ghost_code /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15375
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Is_Ghost_Code
  (Node : Bare_Defining_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Block_Var : Internal_Entity_Basic_Decl;
Var_Expr : Internal_Entity_Basic_Decl;
Var_Expr_1 : Internal_Entity_Generic_Instantiation;
Var_Expr_2 : Internal_Entity_Name;
Match_0 : Internal_Entity_Subp_Renaming_Decl;
Match_1 : Internal_Entity_Package_Renaming_Decl;
Match_2 : Internal_Entity_Generic_Renaming_Decl;
Match_3 : Internal_Entity_Basic_Decl;
Var_Expr_3 : Internal_Entity_Defining_Name;
Fld : Internal_Entity_Basic_Decl;
Sym : Symbol_Type;
Fld_1 : Boolean;
Fld_2 : Internal_Entity_Basic_Decl;
Fld_3 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result : Boolean;
Cast_Expr : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity_Generic_Instantiation;
Fld_4 : Internal_Entity_Basic_Decl;
Fld_5 : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
If_Result_1 : Boolean;
Fld_6 : Internal_Entity_Renaming_Clause;
Fld_7 : Internal_Entity_Name;
Fld_8 : Internal_Entity_Renaming_Clause;
Fld_9 : Internal_Entity_Name;
Fld_10 : Internal_Entity_Name;
Match_Prefix : Internal_Entity_Basic_Decl;
Cast_Result_1 : Internal_Entity_Subp_Renaming_Decl;
Let_Result : Internal_Entity_Name;
Scope_Result_2 : Internal_Entity_Name;
Cast_Result_2 : Internal_Entity_Package_Renaming_Decl;
Let_Result_1 : Internal_Entity_Name;
Scope_Result_3 : Internal_Entity_Name;
Cast_Result_3 : Internal_Entity_Generic_Renaming_Decl;
Let_Result_2 : Internal_Entity_Name;
Scope_Result_4 : Internal_Entity_Name;
Cast_Result_4 : Internal_Entity_Basic_Decl;
Let_Result_3 : Internal_Entity_Name;
Scope_Result_5 : Internal_Entity_Name;
Match_Result : Internal_Entity_Name;
Fld_11 : Internal_Entity_Defining_Name;
Fld_12 : Boolean;
Scope_Result_6 : Boolean;
Result_Var_2 : Boolean;
Scope_Result_7 : Boolean;
Result_Var_3 : Boolean;
If_Result_2 : Boolean;
Let_Result_4 : Boolean;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Defining_Name_Defining_Name_P_Is_Ghost_Code,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Defining_Name,
                                As_Bare_Defining_Name => Self);
               Mmz_K.Items (2) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Boolean;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 9220 '<Block at ???>' Let_Result_4 None
--# scope-start
--# expr-start 9204 '<FieldAccess for basic_decl_no_internal at ast.py:15380>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15380
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.basic_decl_no_internal
Fld := Libadalang.Implementation.Defining_Name_P_Basic_Decl_No_Internal (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 9204
Block_Var := Fld; 
--# bind bd Block_Var
--# expr-start 9206 '<FieldAccess for has_aspect at ast.py:15382>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15382
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9205 '<Symbol Ghost at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Ghost); 
--# expr-done 9205
--# property-call-start DefiningName.has_aspect
Fld_1 := Libadalang.Implementation.Defining_Name_P_Has_Aspect (Node => Ent.Node, Name => Sym, Imprecise_Fallback => False, E_Info => Ent.Info);
--# end
--# expr-done 9206
if Fld_1 then
   If_Result := True;
else
--# expr-start 9207 '<FieldAccess for parent_basic_decl at ast.py:15383>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15383
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.parent_basic_decl
Fld_2 := Libadalang.Implementation.Ada_Node_P_Parent_Basic_Decl (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 9207
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 9208 '<FieldAccess for is_ghost_code at ast.py:15383>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15383
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_ghost_code
Fld_3 := Libadalang.Implementation.Basic_Decl_P_Is_Ghost_Code (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 9208
Scope_Result := Fld_3;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
   If_Result := Result_Var;
end if;
if If_Result then
   If_Result_1 := True;
else
--# expr-start 9209 '<Cast to GenericInstantiation at ast.py:15386>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15386
Cast_Expr := Block_Var; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Generic_Instantiation
   then
      Cast_Result := Create_Internal_Entity_Generic_Instantiation
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Generic_Instantiation;
   end if;
--# expr-done 9209
Var_Expr_1 := Cast_Result;
if Var_Expr_1 /= No_Entity_Generic_Instantiation then
   --# scope-start
   --# bind gi Var_Expr_1
--# expr-start 9211 '<FieldAccess for is_ghost_code at ast.py:15387>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15387
--# expr-start 9210 '<FieldAccess for designated_generic_decl at ast.py:15387>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15387
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericInstantiation.designated_generic_decl'
Fld_4 := Libadalang.Implementation.Dispatcher_Generic_Instantiation_P_Designated_Generic_Decl (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 9210
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_ghost_code
Fld_5 := Libadalang.Implementation.Basic_Decl_P_Is_Ghost_Code (Node => Fld_4.Node, E_Info => Fld_4.Info);
--# end
--# expr-done 9211
Scope_Result_1 := Fld_5;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
   If_Result_1 := Result_Var_1;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
--# expr-start 9217 '<Match at ast.py:15391>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15391
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Block_Var; 
case Ada_Basic_Decl (Match_Prefix.Node.Kind) is
         when Ada_Subp_Renaming_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Subp_Renaming_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_1; 
--# bind sr Match_0
--# expr-start 9213 '<FieldAccess for renamed_object at ast.py:15392>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15392
--# expr-start 9212 '<FieldAccess for renames at ast.py:15392>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15392
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Renaming_Clause (Node => Match_0.Node.Subp_Renaming_Decl_F_Renames, Info => Match_0.Info);
--# expr-done 9212
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Name (Node => Fld_6.Node.Renaming_Clause_F_Renamed_Object, Info => Fld_6.Info);
--# expr-done 9213
Let_Result := Fld_7; 
--# end
Scope_Result_2 := Let_Result;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Package_Renaming_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Package_Renaming_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_2; 
--# bind pr Match_1
--# expr-start 9215 '<FieldAccess for renamed_object at ast.py:15393>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15393
--# expr-start 9214 '<FieldAccess for renames at ast.py:15393>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15393
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Renaming_Clause (Node => Match_1.Node.Package_Renaming_Decl_F_Renames, Info => Match_1.Info);
--# expr-done 9214
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Name (Node => Fld_8.Node.Renaming_Clause_F_Renamed_Object, Info => Fld_8.Info);
--# expr-done 9215
Let_Result_1 := Fld_9; 
--# end
Scope_Result_3 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_3; 
         when Ada_Generic_Package_Renaming_Decl .. Ada_Generic_Subp_Renaming_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Generic_Renaming_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_3; 
--# bind gr Match_2
--# expr-start 9216 '<FieldAccess for renaming_name at ast.py:15394>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15394
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]GenericRenamingDecl.renaming_name'
Fld_10 := Libadalang.Implementation.Dispatcher_Generic_Renaming_Decl_P_Renaming_Name (Node => Match_2.Node, E_Info => Match_2.Info);
--# end
--# expr-done 9216
Let_Result_2 := Fld_10; 
--# end
Scope_Result_4 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_4; 
         when Ada_Abstract_State_Decl .. Ada_Subp_Body | Ada_Package_Body_Stub .. Ada_Generic_Subp_Instantiation | Ada_Label_Decl .. Ada_No_Type_Object_Renaming_Decl | Ada_Single_Protected_Decl .. Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Basic_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_4; 
--# bind _ Match_3
Let_Result_3 := No_Entity_Name; 
--# end
Scope_Result_5 := Let_Result_3;
   --# end
            Match_Result := Scope_Result_5; 
end case;
--# expr-done 9217
Var_Expr_2 := Match_Result;
if Var_Expr_2 /= No_Entity_Name then
   --# scope-start
   --# bind c Var_Expr_2
--# expr-start 9218 '<FieldAccess for referenced_defining_name at ast.py:15396>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15396
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_defining_name
Fld_11 := Libadalang.Implementation.Name_P_Referenced_Defining_Name (Node => Var_Expr_2.Node, Imprecise_Fallback => False, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 9218
Var_Expr_3 := Fld_11;
if Var_Expr_3 /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 9219 '<FieldAccess for is_ghost_code at ast.py:15396>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15396
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.is_ghost_code
Fld_12 := Libadalang.Implementation.Defining_Name_P_Is_Ghost_Code (Node => Var_Expr_3.Node, E_Info => Var_Expr_3.Info);
--# end
--# expr-done 9219
Scope_Result_6 := Fld_12;
   --# end
   Result_Var_2 := Scope_Result_6;
else
   Result_Var_2 := False;
end if;
Scope_Result_7 := Result_Var_2;
   --# end
   Result_Var_3 := Scope_Result_7;
else
   Result_Var_3 := False;
end if;
   If_Result_2 := Result_Var_3;
end if;
Let_Result_4 := If_Result_2; 
--# end
--# expr-done 9220
      Property_Result := Let_Result_4;
   --# end
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Boolean,
                        As_Boolean => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Is_Ghost_Code;
--# end
--# property-start DefiningName.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15400
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Xref_Equation
  (Node : Bare_Defining_Name
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Defining_Name :=
     Internal_Entity_Defining_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1103 with Inline_Always;
            procedure Finalizer_Scope_2933 with Inline_Always;
      Var_Expr : Internal_Entity_Dotted_Name;
Fld : Internal_Entity_Name;
Cast_Result : Internal_Entity_Dotted_Name;
Fld_1 : Internal_Entity_Name;
Fld_2 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1103 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1103;
            procedure Finalizer_Scope_2933 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_2933;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9225 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9222 '<Cast to DottedName at ast.py:15404>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15404
--# expr-start 9221 '<FieldAccess for name at ast.py:15404>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15404
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Defining_Name_F_Name, Info => Ent.Info);
--# expr-done 9221
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Dotted_Name_Range
   then
      Cast_Result := Create_Internal_Entity_Dotted_Name
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Dotted_Name;
   end if;
--# expr-done 9222
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Dotted_Name then
   --# scope-start
   --# bind dn Var_Expr
--# expr-start 9224 '<FieldAccess for xref_equation at ast.py:15405>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15405
--# expr-start 9223 '<FieldAccess for prefix at ast.py:15405>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15405
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Var_Expr.Node.Dotted_Name_F_Prefix, Info => Var_Expr.Info);
--# expr-done 9223
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_2 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Fld_1.Node, Env => Env, Origin => Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 9224
Scope_Result := Fld_2;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2933;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:15406") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 9225
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1103;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1103;
               Finalizer_Scope_2933;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Xref_Equation;
--# end
--# property-start DefiningName.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15411
pragma Warnings (Off, "is not referenced");
 function Defining_Name_P_Xref_Entry_Point
  (Node : Bare_Defining_Name
  )
   return Boolean
is
   Self : Bare_Defining_Name := Bare_Defining_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Name;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9227 '<IsA DottedName at ast.py:15411>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15411
--# expr-start 9226 '<FieldAccess for name at ast.py:15411>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15411
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Defining_Name_F_Name;
--# expr-done 9226
Is_A := Fld /= null 
and then Fld.Kind in Ada_Dotted_Name_Range; 
--# expr-done 9227
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Defining_Name_P_Xref_Entry_Point;
--# end
   --
   --  Primitives for Bare_Synthetic_Defining_Name
   --
      procedure Initialize_Fields_For_Synthetic_Defining_Name
        (Self : Bare_Synthetic_Defining_Name
         ; Defining_Name_F_Name : Bare_Name
        ) is
      begin
            Initialize_Fields_For_Defining_Name
              (Self, Defining_Name_F_Name);
      end Initialize_Fields_For_Synthetic_Defining_Name;
--# property-start SyntheticDefiningName.name_symbol /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16679
pragma Warnings (Off, "is not referenced");
 function Synthetic_Defining_Name_P_Name_Symbol
  (Node : Bare_Synthetic_Defining_Name
  )
   return Symbol_Type
is
   Self : Bare_Synthetic_Defining_Name := Bare_Synthetic_Defining_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type;
      Fld : Bare_Name;
Fld_1 : Symbol_Type;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9229 '<FieldAccess for name_symbol at ast.py:16679>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16679
--# expr-start 9228 '<FieldAccess for name at ast.py:16679>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16679
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Defining_Name_F_Name;
--# expr-done 9228
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld);
--# end
--# expr-done 9229
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Synthetic_Defining_Name_P_Name_Symbol;
--# end
--# property-start SyntheticDefiningName.as_symbol_array /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16681
pragma Warnings (Off, "is not referenced");
 function Synthetic_Defining_Name_P_As_Symbol_Array
  (Node : Bare_Synthetic_Defining_Name
  )
   return Symbol_Type_Array_Access
is
   Self : Bare_Synthetic_Defining_Name := Bare_Synthetic_Defining_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1188 with Inline_Always;
      Fld : Symbol_Type;
Singleton : Symbol_Type_Array_Access;
            procedure Finalizer_Scope_1188 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_1188;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9230 '<FieldAccess for name_symbol at ast.py:16681>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16681
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Self);
--# end
--# expr-done 9230
                Singleton := Create_Symbol_Type_Array (Items_Count => 1);
                Singleton.Items (1) := Fld;
      Property_Result := Singleton;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1188;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1188;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Synthetic_Defining_Name_P_As_Symbol_Array;
--# end
   --
   --  Primitives for Bare_Discrete_Subtype_Name
   --
      procedure Initialize_Fields_For_Discrete_Subtype_Name
        (Self : Bare_Discrete_Subtype_Name
         ; Discrete_Subtype_Name_F_Subtype : Bare_Discrete_Subtype_Indication
        ) is
      begin
            Initialize_Fields_For_Name
              (Self);
            Self.Discrete_Subtype_Name_F_Subtype := Discrete_Subtype_Name_F_Subtype;
      end Initialize_Fields_For_Discrete_Subtype_Name;
   function Discrete_Subtype_Name_F_Subtype
     (Node : Bare_Discrete_Subtype_Name) return Bare_Discrete_Subtype_Indication
   is
   begin
         return Node.Discrete_Subtype_Name_F_Subtype;
   end;
   --
   --  Primitives for Bare_Dotted_Name
   --
      procedure Initialize_Fields_For_Dotted_Name
        (Self : Bare_Dotted_Name
         ; Dotted_Name_F_Prefix : Bare_Name
         ; Dotted_Name_F_Suffix : Bare_Base_Id
        ) is
      begin
            Initialize_Fields_For_Name
              (Self);
            Self.Dotted_Name_F_Prefix := Dotted_Name_F_Prefix;
            Self.Dotted_Name_F_Suffix := Dotted_Name_F_Suffix;
      end Initialize_Fields_For_Dotted_Name;
   function Dotted_Name_F_Prefix
     (Node : Bare_Dotted_Name) return Bare_Name
   is
   begin
         return Node.Dotted_Name_F_Prefix;
   end;
   function Dotted_Name_F_Suffix
     (Node : Bare_Dotted_Name) return Bare_Base_Id
   is
   begin
         return Node.Dotted_Name_F_Suffix;
   end;
--# property-start DottedName.ref_var /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18246
pragma Warnings (Off, "is not referenced");
 function Dotted_Name_P_Ref_Var
  (Node : Bare_Dotted_Name
  )
   return Logic_Var
is
   Self : Bare_Dotted_Name := Bare_Dotted_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Logic_Var;
      Fld : Bare_Base_Id;
Fld_1 : Logic_Var;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9232 '<FieldAccess for ref_var at ast.py:18246>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18246
--# expr-start 9231 '<FieldAccess for suffix at ast.py:18246>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18246
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Dotted_Name_F_Suffix;
--# expr-done 9231
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld);
--# end
--# expr-done 9232
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dotted_Name_P_Ref_Var;
--# end
--# property-start DottedName.subp_spec_var /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18248
pragma Warnings (Off, "is not referenced");
 function Dotted_Name_P_Subp_Spec_Var
  (Node : Bare_Dotted_Name
  )
   return Logic_Var
is
   Self : Bare_Dotted_Name := Bare_Dotted_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Logic_Var;
      Fld : Bare_Base_Id;
Fld_1 : Logic_Var;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9234 '<FieldAccess for subp_spec_var at ast.py:18248>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18248
--# expr-start 9233 '<FieldAccess for suffix at ast.py:18248>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18248
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Dotted_Name_F_Suffix;
--# expr-done 9233
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.subp_spec_var'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Subp_Spec_Var (Node => Fld);
--# end
--# expr-done 9234
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dotted_Name_P_Subp_Spec_Var;
--# end
--# property-start DottedName.defines_subp_spec_var /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18249
pragma Warnings (Off, "is not referenced");
 function Dotted_Name_P_Defines_Subp_Spec_Var
  (Node : Bare_Dotted_Name
  )
   return Boolean
is
   Self : Bare_Dotted_Name := Bare_Dotted_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dotted_Name_P_Defines_Subp_Spec_Var;
--# end
--# property-start DottedName.has_context_free_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18251
pragma Warnings (Off, "is not referenced");
 function Dotted_Name_P_Has_Context_Free_Type
  (Node : Bare_Dotted_Name
  )
   return Boolean
is
   Self : Bare_Dotted_Name := Bare_Dotted_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Base_Id;
Is_A : Boolean;
Not_Val : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9237 '<Not at ast.py:18251>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18251
--# expr-start 9236 '<IsA CharLiteral at ast.py:18251>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18251
--# expr-start 9235 '<FieldAccess for suffix at ast.py:18251>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18251
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Dotted_Name_F_Suffix;
--# expr-done 9235
Is_A := Fld /= null 
and then Fld.Kind in Ada_Char_Literal_Range; 
--# expr-done 9236
Not_Val := not (Is_A); 
--# expr-done 9237
      Property_Result := Not_Val;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dotted_Name_P_Has_Context_Free_Type;
--# end
--# property-start DottedName.complete_items /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18254
pragma Warnings (Off, "is not referenced");
 function Dotted_Name_P_Complete_Items
  (Node : Bare_Dotted_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Completion_Item_Array_Access
is
   Self : Bare_Dotted_Name := Bare_Dotted_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Dotted_Name :=
     Internal_Entity_Dotted_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Completion_Item_Array_Access;
            procedure Finalizer_Scope_1290 with Inline_Always;
      Fld : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_1 : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Bound_No_Visibility : Boolean;
Fld_2 : Internal_Entity_Name;
Fld_3 : Lexical_Env := Empty_Env;
Fld_4 : Internal_Entity_Array_Access;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Fld_5 : Internal_Entity_Info;
Fld_6 : Internal_Metadata;
Fld_7 : Boolean;
Fld_8 : Internal_Entity_Info;
Fld_9 : Internal_Metadata;
Fld_10 : Boolean;
Fld_11 : Internal_Unit;
Fld_12 : Boolean;
If_Result : Boolean;
New_Struct : Internal_Completion_Item;
Is_A : Boolean;
Not_Val : Boolean;
Fld_13 : Analysis_Unit_Kind;
Is_Equal : Boolean;
Fld_14 : Boolean;
If_Result_1 : Boolean;
If_Result_2 : Boolean;
Map_Result : Internal_Completion_Item_Array_Access;
Dyn_Var_Bind_Result : Internal_Completion_Item_Array_Access;
Dyn_Var_Bind_Result_1 : Internal_Completion_Item_Array_Access;
Dyn_Var_Bind_Result_2 : Internal_Completion_Item_Array_Access;
Let_Result : Internal_Completion_Item_Array_Access;
            procedure Finalizer_Scope_1290 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Dyn_Var_Bind_Result_1);
                     Dec_Ref (Dyn_Var_Bind_Result_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1290;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9262 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9261 '<bind at ast.py:18255>' Dyn_Var_Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18255
--# expr-start 9238 '<FieldAccess for origin_node at ast.py:18255>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18255
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 9238
Bound_Origin := Fld; 
--# expr-start 9260 '<bind at ast.py:18255>' Dyn_Var_Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18255
--# expr-start 9239 '<FieldAccess for node_env at ast.py:18256>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18256
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Node_Env (Node => Self);
--# expr-done 9239
Bound_Env := Fld_1; Inc_Ref (Bound_Env);
--# expr-start 9259 '<bind at ast.py:18260>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18260
Bound_No_Visibility := True; 
--# expr-start 9258 '<FilterMap at ast.py:18262>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18262
   --# expr-start 9242 '<FieldAccess for env_get_public at ast.py:18262>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18262
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9241 '<FieldAccess for designated_env at ast.py:18263>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18263
--# expr-start 9240 '<FieldAccess for prefix at ast.py:18263>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18263
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Dotted_Name_F_Prefix, Info => Ent.Info);
--# expr-done 9240
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.designated_env'
Fld_3 := Libadalang.Implementation.Dispatcher_Expr_P_Designated_Env (Node => Fld_2.Node, Env => Bound_Env, Origin => Bound_Origin, No_Visibility => Bound_No_Visibility, E_Info => Fld_2.Info);
--# end
--# expr-done 9241
--# property-call-start AdaNode.env_get_public
Fld_4 := Libadalang.Implementation.Ada_Node_P_Env_Get_Public (Node => Self, Env => Fld_3, Symbol => null, Lookup => Flat);
--# end
--# expr-done 9242
      declare
         Map_Result_Vec : Internal_Completion_Item_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Fld_4;
      begin
         for Item_285 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_285
--# expr-start 9244 '<Not at ast.py:18287>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18287
--# expr-start 9243 '<IsA PrivatePart at ast.py:18287>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18287
Is_A := Item_285.Node /= null 
and then Item_285.Node.Kind in Ada_Private_Part_Range; 
--# expr-done 9243
Not_Val := not (Is_A); 
--# expr-done 9244
if Not_Val then
--# expr-start 9246 '<Eq at ast.py:18289>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18289
--# expr-start 9245 '<FieldAccess for owning_unit_kind at ast.py:18289>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18289
      if Item_285.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.owning_unit_kind
Fld_13 := Libadalang.Implementation.Ada_Node_P_Owning_Unit_Kind (Node => Item_285.Node);
--# end
--# expr-done 9245
Is_Equal := Fld_13 = Unit_Specification; 
--# expr-done 9246
if Is_Equal then
   If_Result_1 := True;
else
   --# expr-start 9247 '<FieldAccess for has_visibility at ast.py:18290>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18290
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.has_visibility
Fld_14 := Libadalang.Implementation.Ada_Node_P_Has_Visibility (Node => Self, Other_Entity => Item_285);
--# end
--# expr-done 9247
   If_Result_1 := Fld_14;
end if;
   If_Result_2 := If_Result_1;
else
   If_Result_2 := False;
end if;
         if If_Result_2 then
      --# expr-start 9257 '<New CompletionItem at ast.py:18267>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18267
--# expr-start 9248 '<Cast to BasicDecl at ast.py:18268>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18268
Cast_Expr := Item_285; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 9248
--# expr-start 9251 '<FieldAccess for dottable_subp at ast.py:18269>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18269
--# expr-start 9250 '<FieldAccess for md at ast.py:18269>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18269
--# expr-start 9249 '<FieldAccess for info at ast.py:18269>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18269
Fld_5 := Item_285.Info;
--# expr-done 9249
Fld_6 := Fld_5.Md;
--# expr-done 9250
Fld_7 := Fld_6.Dottable_Subp;
--# expr-done 9251
--# expr-start 9254 '<FieldAccess for dottable_subp at ast.py:18273>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18273
--# expr-start 9253 '<FieldAccess for md at ast.py:18273>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18273
--# expr-start 9252 '<FieldAccess for info at ast.py:18273>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18273
Fld_8 := Item_285.Info;
--# expr-done 9252
Fld_9 := Fld_8.Md;
--# expr-done 9253
Fld_10 := Fld_9.Dottable_Subp;
--# expr-done 9254
if Fld_10 then
   If_Result := True;
else
   --# expr-start 9256 '<FieldAccess for has_with_visibility at ast.py:18276>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18276
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9255 '<FieldAccess for unit at ast.py:18276>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18276
      if Item_285.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Libadalang.Implementation.Unit (Node => Item_285.Node);
--# expr-done 9255
--# property-call-start AdaNode.has_with_visibility
Fld_12 := Libadalang.Implementation.Ada_Node_P_Has_With_Visibility (Node => Self, Refd_Unit => Fld_11);
--# end
--# expr-done 9256
   If_Result := Fld_12;
end if;
New_Struct := (Decl => Cast_Result, Is_Dot_Call => Fld_7, Is_Visible => If_Result, Weight => 0); 
--# expr-done 9257
         declare
            Item_To_Append : constant Internal_Completion_Item := New_Struct;
         begin
            Internal_Completion_Item_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Completion_Item_Array
           (Items_Count => Natural (Internal_Completion_Item_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Completion_Item_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Completion_Item_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Completion_Item_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 9258
Dyn_Var_Bind_Result := Map_Result; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 9259
Dyn_Var_Bind_Result_1 := Dyn_Var_Bind_Result; Inc_Ref (Dyn_Var_Bind_Result_1);
--# expr-done 9260
Dyn_Var_Bind_Result_2 := Dyn_Var_Bind_Result_1; Inc_Ref (Dyn_Var_Bind_Result_2);
--# expr-done 9261
Let_Result := Dyn_Var_Bind_Result_2; Inc_Ref (Let_Result);
--# end
--# expr-done 9262
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1290;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1290;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dotted_Name_P_Complete_Items;
--# end
--# property-start DottedName.designated_env_no_overloading /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18298
pragma Warnings (Off, "is not referenced");
 function Dotted_Name_P_Designated_Env_No_Overloading
  (Node : Bare_Dotted_Name
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
      ; No_Visibility : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Dotted_Name := Bare_Dotted_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Dotted_Name :=
     Internal_Entity_Dotted_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   --# bind no_visibility No_Visibility
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1291 with Inline_Always;
      Block_Var : Lexical_Env := Empty_Env;
Fld : Internal_Entity_Name;
Fld_1 : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_2 : Internal_Entity_Base_Id;
Fld_3 : Lexical_Env := Empty_Env;
Dyn_Var_Bind_Result : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1291 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1291;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9268 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9264 '<FieldAccess for designated_env_no_overloading at ast.py:18299>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18299
--# expr-start 9263 '<FieldAccess for prefix at ast.py:18299>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18299
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Dotted_Name_F_Prefix, Info => Ent.Info);
--# expr-done 9263
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.designated_env_no_overloading'
Fld_1 := Libadalang.Implementation.Dispatcher_Expr_P_Designated_Env_No_Overloading (Node => Fld.Node, Env => Env, Origin => Origin, No_Visibility => No_Visibility, E_Info => Fld.Info);
--# end
--# expr-done 9264
Block_Var := Fld_1; Inc_Ref (Block_Var);
--# bind pfx_env Block_Var
--# expr-start 9267 '<bind at ast.py:18300>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18300
Bound_Env := Block_Var; Inc_Ref (Bound_Env);
--# expr-start 9266 '<FieldAccess for designated_env_no_overloading at ast.py:18301>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18301
--# expr-start 9265 '<FieldAccess for suffix at ast.py:18301>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18301
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Base_Id (Node => Ent.Node.Dotted_Name_F_Suffix, Info => Ent.Info);
--# expr-done 9265
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.designated_env_no_overloading'
Fld_3 := Libadalang.Implementation.Dispatcher_Expr_P_Designated_Env_No_Overloading (Node => Fld_2.Node, Env => Bound_Env, Origin => Origin, No_Visibility => No_Visibility, E_Info => Fld_2.Info);
--# end
--# expr-done 9266
Dyn_Var_Bind_Result := Fld_3; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 9267
Let_Result := Dyn_Var_Bind_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 9268
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1291;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1291;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dotted_Name_P_Designated_Env_No_Overloading;
--# end
--# property-start DottedName.designated_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18304
pragma Warnings (Off, "is not referenced");
 function Dotted_Name_P_Designated_Env
  (Node : Bare_Dotted_Name
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
      ; No_Visibility : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Dotted_Name := Bare_Dotted_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Dotted_Name :=
     Internal_Entity_Dotted_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   --# bind no_visibility No_Visibility
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1292 with Inline_Always;
      Block_Var : Lexical_Env := Empty_Env;
Fld : Internal_Entity_Name;
Fld_1 : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_2 : Internal_Entity_Base_Id;
Fld_3 : Lexical_Env := Empty_Env;
Dyn_Var_Bind_Result : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1292 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1292;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9274 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9270 '<FieldAccess for designated_env at ast.py:18305>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18305
--# expr-start 9269 '<FieldAccess for prefix at ast.py:18305>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18305
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Dotted_Name_F_Prefix, Info => Ent.Info);
--# expr-done 9269
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.designated_env'
Fld_1 := Libadalang.Implementation.Dispatcher_Expr_P_Designated_Env (Node => Fld.Node, Env => Env, Origin => Origin, No_Visibility => No_Visibility, E_Info => Fld.Info);
--# end
--# expr-done 9270
Block_Var := Fld_1; Inc_Ref (Block_Var);
--# bind pfx_env Block_Var
--# expr-start 9273 '<bind at ast.py:18306>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18306
Bound_Env := Block_Var; Inc_Ref (Bound_Env);
--# expr-start 9272 '<FieldAccess for designated_env at ast.py:18306>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18306
--# expr-start 9271 '<FieldAccess for suffix at ast.py:18306>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18306
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Base_Id (Node => Ent.Node.Dotted_Name_F_Suffix, Info => Ent.Info);
--# expr-done 9271
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.designated_env'
Fld_3 := Libadalang.Implementation.Dispatcher_Expr_P_Designated_Env (Node => Fld_2.Node, Env => Bound_Env, Origin => Origin, No_Visibility => No_Visibility, E_Info => Fld_2.Info);
--# end
--# expr-done 9272
Dyn_Var_Bind_Result := Fld_3; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 9273
Let_Result := Dyn_Var_Bind_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 9274
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1292;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1292;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dotted_Name_P_Designated_Env;
--# end
--# property-start DottedName.all_env_els_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18309
pragma Warnings (Off, "is not referenced");
 function Dotted_Name_P_All_Env_Els_Impl
  (Node : Bare_Dotted_Name
      ; Seq : Boolean
         := True
      ; Seq_From : Bare_Ada_Node
         := No_Bare_Ada_Node
      ; Categories : Ref_Categories
         := (Inherited_Primitives => True, Nocat => True)
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_Dotted_Name := Bare_Dotted_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Dotted_Name :=
     Internal_Entity_Dotted_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind seq Seq
   --# bind seq_from Seq_From
   --# bind categories Categories
   --# bind env Env
   --# bind origin Origin
   Property_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1293 with Inline_Always;
      Block_Var : Lexical_Env := Empty_Env;
Fld : Internal_Entity_Name;
Fld_1 : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_2 : Internal_Entity_Base_Id;
Fld_3 : Internal_Entity_Array_Access;
Dyn_Var_Bind_Result : Internal_Entity_Array_Access;
Let_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1293 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1293;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9280 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9276 '<FieldAccess for designated_env at ast.py:18314>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18314
--# expr-start 9275 '<FieldAccess for prefix at ast.py:18314>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18314
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Dotted_Name_F_Prefix, Info => Ent.Info);
--# expr-done 9275
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.designated_env'
Fld_1 := Libadalang.Implementation.Dispatcher_Expr_P_Designated_Env (Node => Fld.Node, Env => Env, Origin => Origin, No_Visibility => False, E_Info => Fld.Info);
--# end
--# expr-done 9276
Block_Var := Fld_1; Inc_Ref (Block_Var);
--# bind pfx_env Block_Var
--# expr-start 9279 '<bind at ast.py:18315>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18315
Bound_Env := Block_Var; Inc_Ref (Bound_Env);
--# expr-start 9278 '<FieldAccess for all_env_els_impl at ast.py:18317>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18317
--# expr-start 9277 '<FieldAccess for suffix at ast.py:18317>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18317
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Base_Id (Node => Ent.Node.Dotted_Name_F_Suffix, Info => Ent.Info);
--# expr-done 9277
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.all_env_els_impl'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_All_Env_Els_Impl (Node => Fld_2.Node, Seq => Seq, Seq_From => Seq_From, Categories => Categories, Env => Bound_Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 9278
Dyn_Var_Bind_Result := Fld_3; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 9279
Let_Result := Dyn_Var_Bind_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 9280
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1293;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1293;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dotted_Name_P_All_Env_Els_Impl;
--# end
--# property-start DottedName.scope /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18320
pragma Warnings (Off, "is not referenced");
 function Dotted_Name_P_Scope
  (Node : Bare_Dotted_Name
      ; Env : Lexical_Env
  )
   return Lexical_Env
is
   Self : Bare_Dotted_Name := Bare_Dotted_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1294 with Inline_Always;
            procedure Finalizer_Scope_2935 with Inline_Always;
      Var_Expr : Bare_Base_Id;
Fld : Bare_Base_Id;
Fld_1 : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_2 : Lexical_Env := Empty_Env;
Dyn_Var_Bind_Result : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1294 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_1294;
            procedure Finalizer_Scope_2935 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Dyn_Var_Bind_Result);
            end Finalizer_Scope_2935;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
--# expr-start 9281 '<FieldAccess for suffix at ast.py:18320>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18320
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Dotted_Name_F_Suffix;
--# expr-done 9281
Var_Expr := Fld;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind sfx Var_Expr
--# expr-start 9284 '<bind at ast.py:18321>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18321
--# expr-start 9282 '<FieldAccess for parent_scope at ast.py:18321>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18321
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.parent_scope'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Parent_Scope (Node => Self, Env => Env);
--# end
--# expr-done 9282
Bound_Env := Fld_1; Inc_Ref (Bound_Env);
--# expr-start 9283 '<FieldAccess for scope at ast.py:18321>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18321
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.scope'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_Scope (Node => Var_Expr, Env => Bound_Env);
--# end
--# expr-done 9283
Dyn_Var_Bind_Result := Fld_2; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 9284
Scope_Result := Dyn_Var_Bind_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2935;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1294;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1294;
               Finalizer_Scope_2935;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dotted_Name_P_Scope;
--# end
--# property-start DottedName.parent_scope /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18325
pragma Warnings (Off, "is not referenced");
 function Dotted_Name_P_Parent_Scope
  (Node : Bare_Dotted_Name
      ; Env : Lexical_Env
  )
   return Lexical_Env
is
   Self : Bare_Dotted_Name := Bare_Dotted_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1295 with Inline_Always;
      Fld : Bare_Name;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1295 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1295;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9286 '<FieldAccess for scope at ast.py:18325>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18325
--# expr-start 9285 '<FieldAccess for prefix at ast.py:18325>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18325
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Dotted_Name_F_Prefix;
--# expr-done 9285
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.scope'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Scope (Node => Fld, Env => Env);
--# end
--# expr-done 9286
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1295;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1295;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dotted_Name_P_Parent_Scope;
--# end
--# property-start DottedName.relative_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18327
pragma Warnings (Off, "is not referenced");
 function Dotted_Name_P_Relative_Name
  (Node : Bare_Dotted_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Single_Tok_Node
is
   Self : Bare_Dotted_Name := Bare_Dotted_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Dotted_Name :=
     Internal_Entity_Dotted_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Single_Tok_Node;
      Fld : Internal_Entity_Base_Id;
Fld_1 : Internal_Entity_Single_Tok_Node;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9288 '<FieldAccess for relative_name at ast.py:18327>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18327
--# expr-start 9287 '<FieldAccess for suffix at ast.py:18327>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18327
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Base_Id (Node => Ent.Node.Dotted_Name_F_Suffix, Info => Ent.Info);
--# expr-done 9287
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.relative_name'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Relative_Name (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 9288
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dotted_Name_P_Relative_Name;
--# end
--# property-start DottedName.base_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18328
pragma Warnings (Off, "is not referenced");
 function Dotted_Name_P_Base_Name
  (Node : Bare_Dotted_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Name
is
   Self : Bare_Dotted_Name := Bare_Dotted_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Dotted_Name :=
     Internal_Entity_Dotted_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Name;
      Fld : Internal_Entity_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9289 '<FieldAccess for prefix at ast.py:18328>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18328
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Dotted_Name_F_Prefix, Info => Ent.Info);
--# expr-done 9289
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dotted_Name_P_Base_Name;
--# end
--# property-start DottedName.env_elements_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18331
pragma Warnings (Off, "is not referenced");
 function Dotted_Name_P_Env_Elements_Impl
  (Node : Bare_Dotted_Name
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_Dotted_Name := Bare_Dotted_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Dotted_Name :=
     Internal_Entity_Dotted_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   Property_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1298 with Inline_Always;
      Block_Var : Lexical_Env := Empty_Env;
Fld : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_1 : Internal_Entity_Name;
Fld_2 : Lexical_Env := Empty_Env;
Dyn_Var_Bind_Result : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_3 : Internal_Entity_Base_Id;
Fld_4 : Internal_Entity_Array_Access;
Dyn_Var_Bind_Result_1 : Internal_Entity_Array_Access;
Let_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1298 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Dyn_Var_Bind_Result_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1298;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9297 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9293 '<bind at ast.py:18332>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18332
--# expr-start 9290 '<FieldAccess for origin_node at ast.py:18332>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18332
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 9290
Bound_Origin := Fld; 
--# expr-start 9292 '<FieldAccess for designated_env at ast.py:18333>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18333
--# expr-start 9291 '<FieldAccess for prefix at ast.py:18333>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18333
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Ent.Node.Dotted_Name_F_Prefix, Info => Ent.Info);
--# expr-done 9291
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.designated_env'
Fld_2 := Libadalang.Implementation.Dispatcher_Expr_P_Designated_Env (Node => Fld_1.Node, Env => Env, Origin => Bound_Origin, No_Visibility => False, E_Info => Fld_1.Info);
--# end
--# expr-done 9292
Dyn_Var_Bind_Result := Fld_2; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 9293
Block_Var := Dyn_Var_Bind_Result; Inc_Ref (Block_Var);
--# bind pfx_env Block_Var
--# expr-start 9296 '<bind at ast.py:18334>' Dyn_Var_Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18334
Bound_Env := Block_Var; Inc_Ref (Bound_Env);
--# expr-start 9295 '<FieldAccess for env_elements_baseid at ast.py:18334>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18334
--# expr-start 9294 '<FieldAccess for suffix at ast.py:18334>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18334
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Base_Id (Node => Ent.Node.Dotted_Name_F_Suffix, Info => Ent.Info);
--# expr-done 9294
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseId.env_elements_baseid
Fld_4 := Libadalang.Implementation.Base_Id_P_Env_Elements_Baseid (Node => Fld_3.Node, Env => Bound_Env, E_Info => Fld_3.Info);
--# end
--# expr-done 9295
Dyn_Var_Bind_Result_1 := Fld_4; Inc_Ref (Dyn_Var_Bind_Result_1);
--# expr-done 9296
Let_Result := Dyn_Var_Bind_Result_1; Inc_Ref (Let_Result);
--# end
--# expr-done 9297
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1298;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1298;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dotted_Name_P_Env_Elements_Impl;
--# end
--# property-start DottedName.designated_type_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18337
pragma Warnings (Off, "is not referenced");
 function Dotted_Name_P_Designated_Type_Impl
  (Node : Bare_Dotted_Name
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Dotted_Name := Bare_Dotted_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Dotted_Name :=
     Internal_Entity_Dotted_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1299 with Inline_Always;
      Fld : Internal_Entity_Name;
Fld_1 : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_2 : Internal_Entity_Base_Id;
Fld_3 : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1299 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Bound_Env);
            end Finalizer_Scope_1299;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9303 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9302 '<bind at ast.py:18338>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18338
--# expr-start 9299 '<FieldAccess for designated_env_no_overloading at ast.py:18338>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18338
--# expr-start 9298 '<FieldAccess for prefix at ast.py:18338>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18338
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Dotted_Name_F_Prefix, Info => Ent.Info);
--# expr-done 9298
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.designated_env_no_overloading'
Fld_1 := Libadalang.Implementation.Dispatcher_Expr_P_Designated_Env_No_Overloading (Node => Fld.Node, Env => Env, Origin => Origin, No_Visibility => False, E_Info => Fld.Info);
--# end
--# expr-done 9299
Bound_Env := Fld_1; Inc_Ref (Bound_Env);
--# expr-start 9301 '<FieldAccess for designated_type_impl at ast.py:18339>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18339
--# expr-start 9300 '<FieldAccess for suffix at ast.py:18339>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18339
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Base_Id (Node => Ent.Node.Dotted_Name_F_Suffix, Info => Ent.Info);
--# expr-done 9300
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.designated_type_impl'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Designated_Type_Impl (Node => Fld_2.Node, Env => Bound_Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 9301
Dyn_Var_Bind_Result := Fld_3; 
--# expr-done 9302
Let_Result := Dyn_Var_Bind_Result; 
--# end
--# expr-done 9303
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1299;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1299;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dotted_Name_P_Designated_Type_Impl;
--# end
--# property-start DottedName.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18342
pragma Warnings (Off, "is not referenced");
 function Dotted_Name_P_Xref_Equation
  (Node : Bare_Dotted_Name
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Dotted_Name := Bare_Dotted_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Dotted_Name :=
     Internal_Entity_Dotted_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1300 with Inline_Always;
            procedure Finalizer_Scope_2947 with Inline_Always;
      Block_Var : Logic_Equation;
Fld : Internal_Entity_Name;
Fld_1 : Logic_Equation;
Fld_2 : Internal_Entity_Name;
Fld_3 : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_4 : Internal_Entity_Base_Id;
Fld_5 : Logic_Equation;
Dyn_Var_Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_6 : Logic_Var;
Fld_7 : Bare_Base_Id;
Fld_8 : Logic_Var;
Bind_Result : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_9 : Logic_Var;
Fld_10 : Bare_Base_Id;
Fld_11 : Logic_Var;
Bind_Result_1 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_12 : Internal_Entity_Array_Access;
Fld_13 : Bare_Base_Id;
Fld_14 : Logic_Var;
Bind_Result_2 : Logic_Equation;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Fld_15 : Bare_Name;
Cast_Result_1 : Bare_Expr;
Fld_16 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
And_Pred_4 : Logic_Equation;
Fld_17 : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
Not_Val : Boolean;
If_Result : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1300 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_5);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Fld_12);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1300;
            procedure Finalizer_Scope_2947 is
            begin
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (Fld_16);
                     Dec_Ref (And_Pred_3);
            end Finalizer_Scope_2947;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9337 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9311 '<And at ast.py:18343>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18343
--# expr-start 9305 '<FieldAccess for sub_equation at ast.py:18343>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18343
--# expr-start 9304 '<FieldAccess for prefix at ast.py:18343>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18343
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Dotted_Name_F_Prefix, Info => Ent.Info);
--# expr-done 9304
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 9305
--# expr-start 9310 '<bind at ast.py:18344>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18344
--# expr-start 9307 '<FieldAccess for designated_env at ast.py:18344>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18344
--# expr-start 9306 '<FieldAccess for prefix at ast.py:18344>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18344
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Dotted_Name_F_Prefix, Info => Ent.Info);
--# expr-done 9306
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.designated_env'
Fld_3 := Libadalang.Implementation.Dispatcher_Expr_P_Designated_Env (Node => Fld_2.Node, Env => Env, Origin => Origin, No_Visibility => False, E_Info => Fld_2.Info);
--# end
--# expr-done 9307
Bound_Env := Fld_3; Inc_Ref (Bound_Env);
--# expr-start 9309 '<FieldAccess for sub_equation at ast.py:18345>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18345
--# expr-start 9308 '<FieldAccess for suffix at ast.py:18345>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18345
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Base_Id (Node => Ent.Node.Dotted_Name_F_Suffix, Info => Ent.Info);
--# expr-done 9308
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_5 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_4.Node, Env => Bound_Env, Origin => Origin, E_Info => Fld_4.Info);
--# end
--# expr-done 9309
Dyn_Var_Bind_Result := Fld_5; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 9310
And_Pred := Create_And (Fld_1, Dyn_Var_Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18343") else null)); 
--# expr-done 9311
Block_Var := And_Pred; Inc_Ref (Block_Var);
--# bind base Block_Var
--# expr-start 9336 '<If at ast.py:18347>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18347
--# expr-start 9314 '<Not at ast.py:18348>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18348
--# expr-start 9313 '<IsNull at ast.py:18348>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18348
--# expr-start 9312 '<FieldAccess for designated_type_impl at ast.py:18348>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18348
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.designated_type_impl'
Fld_17 := Libadalang.Implementation.Dispatcher_Name_P_Designated_Type_Impl (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 9312
Is_Null := Fld_17.Node = null; 
--# expr-done 9313
Not_Val := not (Is_Null); 
--# expr-done 9314
if Not_Val then
   If_Result := Block_Var;
else
   --# expr-start 9335 '<And at ast.py:18350>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18350
--# expr-start 9324 '<And at ast.py:18350>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18350
--# expr-start 9319 '<And at ast.py:18350>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18350
--# expr-start 9318 '<Bind at ast.py:18351>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18351
--# expr-start 9315 '<FieldAccess for expected_type_var at ast.py:18351>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18351
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 9315
Fld_6.Value := No_Entity;
Entity_Vars.Reset (Fld_6);
--# expr-start 9317 '<FieldAccess for expected_type_var at ast.py:18351>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18351
--# expr-start 9316 '<FieldAccess for suffix at ast.py:18351>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18351
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Self.Dotted_Name_F_Suffix;
--# expr-done 9316
      if Fld_7 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_7);
--# expr-done 9317
Fld_8.Value := No_Entity;
Entity_Vars.Reset (Fld_8);
Bind_Result := Solver.Create_Unify (Fld_6, Fld_8, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18351") else null)); 
--# expr-done 9318
And_Pred_1 := Create_And (Block_Var, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18350") else null)); 
--# expr-done 9319
--# expr-start 9323 '<Bind at ast.py:18352>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18352
--# expr-start 9320 '<FieldAccess for type_var at ast.py:18352>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18352
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 9320
Fld_9.Value := No_Entity;
Entity_Vars.Reset (Fld_9);
--# expr-start 9322 '<FieldAccess for type_var at ast.py:18352>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18352
--# expr-start 9321 '<FieldAccess for suffix at ast.py:18352>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18352
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Self.Dotted_Name_F_Suffix;
--# expr-done 9321
      if Fld_10 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_10);
--# expr-done 9322
Fld_11.Value := No_Entity;
Entity_Vars.Reset (Fld_11);
Bind_Result_1 := Solver.Create_Unify (Fld_9, Fld_11, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18352") else null)); 
--# expr-done 9323
And_Pred_2 := Create_And (And_Pred_1, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18350") else null)); 
--# expr-done 9324
--# expr-start 9334 '<LogicAny at ast.py:18353>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18353
--# expr-start 9333 '<Map at ast.py:18353>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18353
   --# expr-start 9325 '<FieldAccess for env_elements at ast.py:18353>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18353
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.env_elements
Fld_12 := Libadalang.Implementation.Expr_P_Env_Elements (Node => Ent.Node, Env => Env, No_Visibility => False, E_Info => Ent.Info);
--# end
--# expr-done 9325
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Fld_12;
      begin
         for Item_286 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_286
      --# expr-start 9332 '<And at ast.py:18355>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18355
--# expr-start 9328 '<Bind at ast.py:18355>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18355
--# expr-start 9327 '<FieldAccess for ref_var at ast.py:18355>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18355
--# expr-start 9326 '<FieldAccess for suffix at ast.py:18355>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18355
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Self.Dotted_Name_F_Suffix;
--# expr-done 9326
      if Fld_13 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_14 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_13);
--# end
--# expr-done 9327
Fld_14.Value := No_Entity;
Entity_Vars.Reset (Fld_14);
Bind_Result_2 := Solver.Create_Assign (Fld_14, Item_286, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18355") else null)); 
--# expr-done 9328
--# expr-start 9331 '<FieldAccess for constrain_prefix at ast.py:18356>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18356
--# expr-start 9329 '<Cast to BasicDecl.entity at ast.py:18356>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18356
Cast_Expr := Item_286; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 9329
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9330 '<FieldAccess for prefix at ast.py:18356>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18356
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Self.Dotted_Name_F_Prefix;
--# expr-done 9330
      Cast_Result_1 := Fld_15;
--# property-call-start '[dispatcher]BasicDecl.constrain_prefix'
Fld_16 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Constrain_Prefix (Node => Cast_Result.Node, Prefix => Cast_Result_1, Origin => Origin, E_Info => Cast_Result.Info);
--# end
--# expr-done 9331
And_Pred_3 := Create_And (Bind_Result_2, Fld_16, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18355") else null)); 
--# expr-done 9332
         declare
            Item_To_Append : constant Logic_Equation := And_Pred_3;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2947;
         end loop;
      end;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 9333
Logic_Boolean_Op := Solver.Create_Any (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18353") else null)); 
--# expr-done 9334
And_Pred_4 := Create_And (And_Pred_2, Logic_Boolean_Op, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18350") else null)); 
--# expr-done 9335
   If_Result := And_Pred_4;
end if;
      Inc_Ref (If_Result);
--# expr-done 9336
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 9337
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1300;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1300;
               Finalizer_Scope_2947;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dotted_Name_P_Xref_Equation;
--# end
--# property-start DottedName.is_constant /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18361
pragma Warnings (Off, "is not referenced");
 function Dotted_Name_P_Is_Constant
  (Node : Bare_Dotted_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Dotted_Name := Bare_Dotted_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Dotted_Name :=
     Internal_Entity_Dotted_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Internal_Entity_Name;
Fld_1 : Boolean;
Fld_2 : Internal_Entity_Base_Id;
Fld_3 : Boolean;
If_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9342 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9339 '<FieldAccess for is_constant at ast.py:18365>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18365
--# expr-start 9338 '<FieldAccess for prefix at ast.py:18365>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18365
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Dotted_Name_F_Prefix, Info => Ent.Info);
--# expr-done 9338
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.is_constant'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Is_Constant (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 9339
if Fld_1 then
   If_Result := True;
else
   --# expr-start 9341 '<FieldAccess for is_constant at ast.py:18366>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18366
--# expr-start 9340 '<FieldAccess for suffix at ast.py:18366>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18366
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Base_Id (Node => Ent.Node.Dotted_Name_F_Suffix, Info => Ent.Info);
--# expr-done 9340
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.is_constant'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Is_Constant (Node => Fld_2.Node, E_Info => Fld_2.Info);
--# end
--# expr-done 9341
   If_Result := Fld_3;
end if;
Let_Result := If_Result; 
--# end
--# expr-done 9342
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dotted_Name_P_Is_Constant;
--# end
   --
   --  Primitives for Bare_End_Name
   --
      procedure Initialize_Fields_For_End_Name
        (Self : Bare_End_Name
         ; End_Name_F_Name : Bare_Name
        ) is
      begin
            Initialize_Fields_For_Name
              (Self);
            Self.End_Name_F_Name := End_Name_F_Name;
      end Initialize_Fields_For_End_Name;
   function End_Name_F_Name
     (Node : Bare_End_Name) return Bare_Name
   is
   begin
         return Node.End_Name_F_Name;
   end;
--# property-start EndName.parent_scope /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15421
pragma Warnings (Off, "is not referenced");
 function End_Name_P_Parent_Scope
  (Node : Bare_End_Name
      ; Env : Lexical_Env
  )
   return Lexical_Env
is
   Self : Bare_End_Name := Bare_End_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1105 with Inline_Always;
      Fld : Bare_Name;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1105 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1105;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9344 '<FieldAccess for parent_scope at ast.py:15421>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15421
--# expr-start 9343 '<FieldAccess for name at ast.py:15421>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15421
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.End_Name_F_Name;
--# expr-done 9343
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.parent_scope'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Parent_Scope (Node => Fld, Env => Env);
--# end
--# expr-done 9344
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1105;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1105;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end End_Name_P_Parent_Scope;
--# end
--# property-start EndName.scope /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15422
pragma Warnings (Off, "is not referenced");
 function End_Name_P_Scope
  (Node : Bare_End_Name
      ; Env : Lexical_Env
  )
   return Lexical_Env
is
   Self : Bare_End_Name := Bare_End_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1106 with Inline_Always;
      Fld : Bare_Name;
Fld_1 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1106 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1106;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9346 '<FieldAccess for scope at ast.py:15422>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15422
--# expr-start 9345 '<FieldAccess for name at ast.py:15422>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15422
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.End_Name_F_Name;
--# expr-done 9345
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.scope'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Scope (Node => Fld, Env => Env);
--# end
--# expr-done 9346
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1106;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1106;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end End_Name_P_Scope;
--# end
--# property-start EndName.relative_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15423
pragma Warnings (Off, "is not referenced");
 function End_Name_P_Relative_Name
  (Node : Bare_End_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Single_Tok_Node
is
   Self : Bare_End_Name := Bare_End_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_End_Name :=
     Internal_Entity_End_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Single_Tok_Node;
      Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Single_Tok_Node;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9348 '<FieldAccess for relative_name at ast.py:15423>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15423
--# expr-start 9347 '<FieldAccess for name at ast.py:15423>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15423
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.End_Name_F_Name, Info => Ent.Info);
--# expr-done 9347
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.relative_name'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Relative_Name (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 9348
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end End_Name_P_Relative_Name;
--# end
--# property-start EndName.ref_var /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15424
pragma Warnings (Off, "is not referenced");
 function End_Name_P_Ref_Var
  (Node : Bare_End_Name
  )
   return Logic_Var
is
   Self : Bare_End_Name := Bare_End_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Logic_Var;
      Fld : Bare_Name;
Fld_1 : Logic_Var;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9350 '<FieldAccess for ref_var at ast.py:15424>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15424
--# expr-start 9349 '<FieldAccess for name at ast.py:15424>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15424
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.End_Name_F_Name;
--# expr-done 9349
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld);
--# end
--# expr-done 9350
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end End_Name_P_Ref_Var;
--# end
--# property-start EndName.env_elements_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15425
pragma Warnings (Off, "is not referenced");
 function End_Name_P_Env_Elements_Impl
  (Node : Bare_End_Name
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_End_Name := Bare_End_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_End_Name :=
     Internal_Entity_End_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   Property_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1109 with Inline_Always;
      Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1109 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1109;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9352 '<FieldAccess for env_elements_impl at ast.py:15425>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15425
--# expr-start 9351 '<FieldAccess for name at ast.py:15425>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15425
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.End_Name_F_Name, Info => Ent.Info);
--# expr-done 9351
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.env_elements_impl'
Fld_1 := Libadalang.Implementation.Dispatcher_Expr_P_Env_Elements_Impl (Node => Fld.Node, Env => Env, E_Info => Fld.Info);
--# end
--# expr-done 9352
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1109;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1109;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end End_Name_P_Env_Elements_Impl;
--# end
--# property-start EndName.basic_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15427
pragma Warnings (Off, "is not referenced");
 function End_Name_P_Basic_Decl
  (Node : Bare_End_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_End_Name := Bare_End_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_1110 with Inline_Always;
      Var_Expr : Internal_Entity_Basic_Decl;
Var_Expr_1 : Bare_Ada_Node;
Node_Parents : Bare_Ada_Node_Array_Access;
Is_A : Boolean;
Find_Result : Bare_Ada_Node;
Cast_Expr : Bare_Ada_Node;
Cast_Result : Bare_Named_Stmt;
Fld : Bare_Named_Stmt_Decl;
Cast_Result_1 : Bare_Basic_Decl;
As_Entity : Internal_Entity_Basic_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Scope_Result_1 : Internal_Entity_Basic_Decl;
Node_Parents_1 : Bare_Ada_Node_Array_Access;
Is_A_1 : Boolean;
Find_Result_1 : Bare_Ada_Node;
Cast_Result_2 : Bare_Basic_Decl;
As_Entity_1 : Internal_Entity_Basic_Decl;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Basic_Decl;
Result_Var_1 : Internal_Entity_Basic_Decl;
            procedure Finalizer_Scope_1110 is
            begin
                     Dec_Ref (Node_Parents);
                     Dec_Ref (Node_Parents_1);
            end Finalizer_Scope_1110;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_End_Name_End_Name_P_Basic_Decl,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_End_Name,
                                As_Bare_End_Name => Self);
               Mmz_K.Items (2) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Basic_Decl;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
--# expr-start 9355 '<Find at ast.py:15428>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15428
   --# expr-start 9353 '<FieldAccess for parents at ast.py:15428>' Node_Parents /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15428
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents := Parents (Self, True); 
--# expr-done 9353
      Find_Result := No_Bare_Ada_Node;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents;
      begin
         for Item_287 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_287
      --# expr-start 9354 '<IsA NamedStmt at ast.py:15428>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15428
Is_A := Item_287 /= null 
and then Item_287.Kind in Ada_Named_Stmt_Range; 
--# expr-done 9354
      if Is_A then
         Find_Result := Item_287;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 9355
Var_Expr_1 := Find_Result;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind ns Var_Expr_1
--# expr-start 9359 '<as_entity at ast.py:15430>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15430
--# expr-start 9358 '<Cast to BasicDecl at ast.py:15430>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15430
--# expr-start 9357 '<FieldAccess for decl at ast.py:15430>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15430
--# expr-start 9356 '<Cast to NamedStmt at ast.py:15430>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15430
Cast_Expr := Var_Expr_1; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_Named_Stmt_Range
   then
      Cast_Result := Cast_Expr;
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 9356
      if Cast_Result = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Cast_Result.Named_Stmt_F_Decl;
--# expr-done 9357
      Cast_Result_1 := Fld;
--# expr-done 9358
Is_Null := Cast_Result_1 = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Decl;
else
As_Entity := (Info => E_Info, Node => Cast_Result_1); 
   If_Result := As_Entity;
end if;
--# expr-done 9359
Scope_Result := If_Result;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Basic_Decl;
end if;
Var_Expr := Result_Var;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind e Var_Expr
Scope_Result_1 := Var_Expr;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   --# expr-start 9364 '<as_entity at ast.py:15432>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15432
--# expr-start 9363 '<Cast to BasicDecl at ast.py:15432>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15432
--# expr-start 9362 '<Find at ast.py:15432>' Find_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15432
   --# expr-start 9360 '<FieldAccess for parents at ast.py:15432>' Node_Parents_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15432
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents_1 := Parents (Self, True); 
--# expr-done 9360
      Find_Result_1 := No_Bare_Ada_Node;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents_1;
      begin
         for Item_288 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_288
      --# expr-start 9361 '<IsA BasicDecl at ast.py:15432>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15432
Is_A_1 := Item_288 /= null 
and then Item_288.Kind in Ada_Basic_Decl; 
--# expr-done 9361
      if Is_A_1 then
         Find_Result_1 := Item_288;
      end if;
   --# end
               exit when Is_A_1;
         end loop;
      end;
--# expr-done 9362
   if Find_Result_1 = null
      or else Find_Result_1.Kind in Ada_Basic_Decl
   then
      Cast_Result_2 := Find_Result_1;
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 9363
Is_Null_1 := Cast_Result_2 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Basic_Decl;
else
As_Entity_1 := (Info => E_Info, Node => Cast_Result_2); 
   If_Result_1 := As_Entity_1;
end if;
--# expr-done 9364
   Result_Var_1 := If_Result_1;
end if;
      Property_Result := Result_Var_1;
   --# end
      Finalizer_Scope_1110;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Basic_Decl,
                        As_Internal_Entity_Basic_Decl => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1110;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end End_Name_P_Basic_Decl;
--# end
--# property-start EndName.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15440
pragma Warnings (Off, "is not referenced");
 function End_Name_P_Xref_Equation
  (Node : Bare_End_Name
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_End_Name := Bare_End_Name (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_End_Name :=
     Internal_Entity_End_Name'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1111 with Inline_Always;
            procedure Finalizer_Scope_2952 with Inline_Always;
            procedure Finalizer_Scope_2953 with Inline_Always;
      Var_Expr : Internal_Entity_Accept_Stmt_With_Stmts;
Var_Expr_1 : Internal_Entity_Dotted_Name;
Fld : Internal_Entity;
Cast_Result : Internal_Entity_Accept_Stmt_With_Stmts;
Fld_1 : Logic_Var;
Fld_2 : Internal_Entity_Entry_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result : Logic_Equation;
Scope_Result : Logic_Equation;
Fld_3 : Logic_Var;
Fld_4 : Internal_Entity_Basic_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
Result_Var : Logic_Equation;
Fld_5 : Internal_Entity_Name;
Cast_Result_3 : Internal_Entity_Dotted_Name;
Fld_6 : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_7 : Internal_Entity_Name;
Fld_8 : Logic_Equation;
Dyn_Var_Bind_Result : Logic_Equation;
Scope_Result_1 : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var_1 : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1111 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1111;
            procedure Finalizer_Scope_2952 is
            begin
                     Dec_Ref (Bind_Result);
            end Finalizer_Scope_2952;
            procedure Finalizer_Scope_2953 is
            begin
                     Dec_Ref (Fld_6);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_8);
                     Dec_Ref (Dyn_Var_Bind_Result);
            end Finalizer_Scope_2953;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9380 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9379 '<And at ast.py:15441>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15441
--# expr-start 9366 '<Cast to AcceptStmtWithStmts at ast.py:15441>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15441
--# expr-start 9365 '<parent at ast.py:15441>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15441
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 9365
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Accept_Stmt_With_Stmts_Range
   then
      Cast_Result := Create_Internal_Entity_Accept_Stmt_With_Stmts
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Accept_Stmt_With_Stmts;
   end if;
--# expr-done 9366
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Accept_Stmt_With_Stmts then
   --# scope-start
   --# bind stmt Var_Expr
--# expr-start 9369 '<Bind at ast.py:15442>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15442
--# expr-start 9367 '<FieldAccess for ref_var at ast.py:15442>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15442
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Self);
--# end
--# expr-done 9367
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 9368 '<FieldAccess for designated_entry at ast.py:15442>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15442
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AcceptStmt.designated_entry
Fld_2 := Libadalang.Implementation.Accept_Stmt_P_Designated_Entry (Node => Var_Expr.Node, Origin => Origin, Env => Env, E_Info => Var_Expr.Info);
--# end
--# expr-done 9368
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Bind_Result := Solver.Create_Assign (Fld_1, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:15442") else null)); 
--# expr-done 9369
Scope_Result := Bind_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2952;
   Result_Var := Scope_Result;
else
   --# expr-start 9372 '<Bind at ast.py:15443>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15443
--# expr-start 9370 '<FieldAccess for ref_var at ast.py:15443>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15443
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Self);
--# end
--# expr-done 9370
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
--# expr-start 9371 '<FieldAccess for basic_decl at ast.py:15443>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15443
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start EndName.basic_decl
Fld_4 := Libadalang.Implementation.End_Name_P_Basic_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 9371
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_4.Node,
         Info => Fld_4.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_3, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:15443") else null)); 
--# expr-done 9372
   Result_Var := Bind_Result_1;
end if;
      Inc_Ref (Result_Var);
--# expr-start 9374 '<Cast to DottedName at ast.py:15444>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15444
--# expr-start 9373 '<FieldAccess for name at ast.py:15444>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15444
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Name (Node => Ent.Node.End_Name_F_Name, Info => Ent.Info);
--# expr-done 9373
   if Fld_5.Node = null
      or else Fld_5.Node.Kind in Ada_Dotted_Name_Range
   then
      Cast_Result_3 := Create_Internal_Entity_Dotted_Name
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
   else
         Cast_Result_3 := No_Entity_Dotted_Name;
   end if;
--# expr-done 9374
Var_Expr_1 := Cast_Result_3;
if Var_Expr_1 /= No_Entity_Dotted_Name then
   --# scope-start
   --# bind dn Var_Expr_1
--# expr-start 9378 '<bind at ast.py:15448>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15448
--# expr-start 9375 '<FieldAccess for std_env at ast.py:15449>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15449
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.std_env
Fld_6 := Libadalang.Implementation.Ada_Node_P_Std_Env (Node => Ent.Node);
--# end
--# expr-done 9375
Bound_Env := Fld_6; Inc_Ref (Bound_Env);
--# expr-start 9377 '<FieldAccess for xref_no_overloading at ast.py:15450>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15450
--# expr-start 9376 '<FieldAccess for prefix at ast.py:15450>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15450
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Name (Node => Var_Expr_1.Node.Dotted_Name_F_Prefix, Info => Var_Expr_1.Info);
--# expr-done 9376
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_8 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Fld_7.Node, Env => Bound_Env, Origin => Origin, E_Info => Fld_7.Info);
--# end
--# expr-done 9377
Dyn_Var_Bind_Result := Fld_8; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 9378
Scope_Result_1 := Dyn_Var_Bind_Result;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2953;
   Result_Var_1 := Scope_Result_1;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:15452") else null)); 
   Result_Var_1 := True_Rel;
end if;
      Inc_Ref (Result_Var_1);
And_Pred := Create_And (Result_Var, Result_Var_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:15441") else null)); 
--# expr-done 9379
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 9380
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1111;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1111;
               Finalizer_Scope_2952;
               Finalizer_Scope_2953;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end End_Name_P_Xref_Equation;
--# end
--# property-start EndName.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15455
pragma Warnings (Off, "is not referenced");
 function End_Name_P_Xref_Entry_Point
  (Node : Bare_End_Name
  )
   return Boolean
is
   Self : Bare_End_Name := Bare_End_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end End_Name_P_Xref_Entry_Point;
--# end
   --
   --  Primitives for Bare_Explicit_Deref
   --
      procedure Initialize_Fields_For_Explicit_Deref
        (Self : Bare_Explicit_Deref
         ; Explicit_Deref_F_Prefix : Bare_Name
        ) is
      begin
            Initialize_Fields_For_Name
              (Self);
            Self.Explicit_Deref_F_Prefix := Explicit_Deref_F_Prefix;
      Self.Explicit_Deref_R_Called_Spec := Null_Var_Record;
      end Initialize_Fields_For_Explicit_Deref;
   function Explicit_Deref_F_Prefix
     (Node : Bare_Explicit_Deref) return Bare_Name
   is
   begin
         return Node.Explicit_Deref_F_Prefix;
   end;
--# property-start ExplicitDeref.ref_var /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14488
pragma Warnings (Off, "is not referenced");
 function Explicit_Deref_P_Ref_Var
  (Node : Bare_Explicit_Deref
  )
   return Logic_Var
is
   Self : Bare_Explicit_Deref := Bare_Explicit_Deref (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Logic_Var;
      Fld : Bare_Name;
Fld_1 : Logic_Var;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9382 '<FieldAccess for ref_var at ast.py:14488>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14488
--# expr-start 9381 '<FieldAccess for prefix at ast.py:14488>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14488
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Explicit_Deref_F_Prefix;
--# expr-done 9381
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld);
--# end
--# expr-done 9382
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Explicit_Deref_P_Ref_Var;
--# end
--# property-start ExplicitDeref.subp_spec_var /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14492
pragma Warnings (Off, "is not referenced");
 function Explicit_Deref_P_Subp_Spec_Var
  (Node : Bare_Explicit_Deref
  )
   return Logic_Var
is
   Self : Bare_Explicit_Deref := Bare_Explicit_Deref (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Logic_Var;
      Fld : Logic_Var;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9383 '<FieldAccess for r_called_spec at ast.py:14492>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14492
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Explicit_Deref_R_Called_Spec'Unrestricted_Access;
--# expr-done 9383
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Explicit_Deref_P_Subp_Spec_Var;
--# end
--# property-start ExplicitDeref.defines_subp_spec_var /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14493
pragma Warnings (Off, "is not referenced");
 function Explicit_Deref_P_Defines_Subp_Spec_Var
  (Node : Bare_Explicit_Deref
  )
   return Boolean
is
   Self : Bare_Explicit_Deref := Bare_Explicit_Deref (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Explicit_Deref_P_Defines_Subp_Spec_Var;
--# end
--# property-start ExplicitDeref.relative_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14495
pragma Warnings (Off, "is not referenced");
 function Explicit_Deref_P_Relative_Name
  (Node : Bare_Explicit_Deref
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Single_Tok_Node
is
   Self : Bare_Explicit_Deref := Bare_Explicit_Deref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Explicit_Deref :=
     Internal_Entity_Explicit_Deref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Single_Tok_Node;
      Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Single_Tok_Node;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9385 '<FieldAccess for relative_name at ast.py:14495>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14495
--# expr-start 9384 '<FieldAccess for prefix at ast.py:14495>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14495
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Explicit_Deref_F_Prefix, Info => Ent.Info);
--# expr-done 9384
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.relative_name'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Relative_Name (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 9385
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Explicit_Deref_P_Relative_Name;
--# end
--# property-start ExplicitDeref.designated_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14498
pragma Warnings (Off, "is not referenced");
 function Explicit_Deref_P_Designated_Env
  (Node : Bare_Explicit_Deref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
      ; No_Visibility : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Explicit_Deref := Bare_Explicit_Deref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Explicit_Deref :=
     Internal_Entity_Explicit_Deref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   --# bind no_visibility No_Visibility
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1042 with Inline_Always;
      Fld : Internal_Entity_Name;
Fld_1 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1042 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1042;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9388 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9387 '<FieldAccess for designated_env at ast.py:14501>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14501
--# expr-start 9386 '<FieldAccess for prefix at ast.py:14501>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14501
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Explicit_Deref_F_Prefix, Info => Ent.Info);
--# expr-done 9386
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.designated_env'
Fld_1 := Libadalang.Implementation.Dispatcher_Expr_P_Designated_Env (Node => Fld.Node, Env => Env, Origin => Origin, No_Visibility => No_Visibility, E_Info => Fld.Info);
--# end
--# expr-done 9387
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 9388
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1042;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1042;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Explicit_Deref_P_Designated_Env;
--# end
--# property-start ExplicitDeref.env_elements_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14504
pragma Warnings (Off, "is not referenced");
 function Explicit_Deref_P_Env_Elements_Impl
  (Node : Bare_Explicit_Deref
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_Explicit_Deref := Bare_Explicit_Deref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Explicit_Deref :=
     Internal_Entity_Explicit_Deref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   Property_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1043 with Inline_Always;
      Block_Var : Internal_Entity_Name;
Block_Var_1 : Internal_Entity_Array_Access;
Var_Expr : Internal_Entity_Attribute_Ref;
Var_Expr_1 : Internal_Entity_Base_Type_Decl;
Var_Expr_2 : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Array_Access;
Fld_2 : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Fld_3 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Fld_4 : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
Map_Result : Internal_Entity_Array_Access;
Dyn_Var_Bind_Result : Internal_Entity_Array_Access;
Cast_Expr_1 : Internal_Entity_Name;
Cast_Result_1 : Internal_Entity_Attribute_Ref;
Fld_5 : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
If_Result : Internal_Entity_Array_Access;
Let_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1043 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1043;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9400 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9389 '<FieldAccess for prefix at ast.py:14505>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14505
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Explicit_Deref_F_Prefix, Info => Ent.Info);
--# expr-done 9389
Block_Var := Fld; 
--# bind prefix Block_Var
--# expr-start 9390 '<FieldAccess for env_elements_impl at ast.py:14506>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14506
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.env_elements_impl'
Fld_1 := Libadalang.Implementation.Dispatcher_Expr_P_Env_Elements_Impl (Node => Block_Var.Node, Env => Env, E_Info => Block_Var.Info);
--# end
--# expr-done 9390
Block_Var_1 := Fld_1; Inc_Ref (Block_Var_1);
--# bind env_els Block_Var_1
--# expr-start 9399 '<If at ast.py:14507>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14507
--# expr-start 9391 '<Cast to AttributeRef at ast.py:14508>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14508
Cast_Expr_1 := Block_Var; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Attribute_Ref_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Attribute_Ref
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_1 := No_Entity_Attribute_Ref;
   end if;
--# expr-done 9391
Var_Expr := Cast_Result_1;
if Var_Expr /= No_Entity_Attribute_Ref then
   --# scope-start
--# expr-start 9392 '<FieldAccess for is_access_attr at ast.py:14508>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14508
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.is_access_attr
Fld_5 := Libadalang.Implementation.Attribute_Ref_P_Is_Access_Attr (Node => Var_Expr.Node);
--# end
--# expr-done 9392
Scope_Result_2 := Fld_5;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := False;
end if;
if Result_Var_2 then
   If_Result := Block_Var_1;
else
   --# expr-start 9398 '<bind at ast.py:14510>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14510
--# expr-start 9393 '<FieldAccess for origin_node at ast.py:14510>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14510
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_2 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 9393
Bound_Origin := Fld_2; 
--# expr-start 9397 '<Filter at ast.py:14510>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14510
      declare
         Map_Result_Vec : Internal_Entity_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Block_Var_1;
      begin
         for Item_289 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_289
--# expr-start 9394 '<Cast to BasicDecl at ast.py:14512>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14512
Cast_Expr := Item_289; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 9394
Var_Expr_2 := Cast_Result;
if Var_Expr_2 /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 9395 '<FieldAccess for expr_type at ast.py:14512>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14512
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.expr_type'
Fld_3 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Expr_Type (Node => Var_Expr_2.Node, Origin => Bound_Origin, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 9395
Scope_Result := Fld_3;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Var_Expr_1 := Result_Var;
if Var_Expr_1 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind t Var_Expr_1
--# expr-start 9396 '<FieldAccess for is_access_type at ast.py:14513>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14513
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_access_type'
Fld_4 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Access_Type (Node => Var_Expr_1.Node, Origin => Bound_Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 9396
Scope_Result_1 := Fld_4;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
         if Result_Var_1 then
         declare
            Item_To_Append : constant Internal_Entity := Item_289;
         begin
            Internal_Entity_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 9397
Dyn_Var_Bind_Result := Map_Result; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 9398
   If_Result := Dyn_Var_Bind_Result;
end if;
      Inc_Ref (If_Result);
--# expr-done 9399
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 9400
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1043;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1043;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Explicit_Deref_P_Env_Elements_Impl;
--# end
--# property-start ExplicitDeref.eq_for_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14519
pragma Warnings (Off, "is not referenced");
 function Explicit_Deref_P_Eq_For_Type
  (Node : Bare_Explicit_Deref
      ; Typ : Internal_Entity_Base_Type_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Explicit_Deref := Bare_Explicit_Deref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Explicit_Deref :=
     Internal_Entity_Explicit_Deref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind typ Typ
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1044 with Inline_Always;
      Fld : Bare_Name;
Fld_1 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_2 : Internal_Entity_Name;
Fld_3 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_4 : Logic_Var;
Fld_5 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred_1 : Logic_Equation;
False_Rel : Logic_Equation;
Fld_6 : Boolean;
If_Result : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1044 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_3);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (False_Rel);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1044;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9413 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9412 '<If at ast.py:14520>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14520
--# expr-start 9401 '<FieldAccess for is_access_type at ast.py:14521>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14521
      if Typ.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_access_type'
Fld_6 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Access_Type (Node => Typ.Node, Origin => Origin, E_Info => Typ.Info);
--# end
--# expr-done 9401
if Fld_6 then
   --# expr-start 9411 '<And at ast.py:14523>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14523
--# expr-start 9407 '<And at ast.py:14523>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14523
--# expr-start 9404 '<Bind at ast.py:14523>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14523
--# expr-start 9403 '<FieldAccess for expected_type_var at ast.py:14523>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14523
--# expr-start 9402 '<FieldAccess for prefix at ast.py:14523>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14523
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Explicit_Deref_F_Prefix;
--# expr-done 9402
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld);
--# expr-done 9403
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
Cast_Expr := Typ; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_1, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14523") else null)); 
--# expr-done 9404
--# expr-start 9406 '<FieldAccess for matches_expected_type at ast.py:14524>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14524
--# expr-start 9405 '<FieldAccess for prefix at ast.py:14524>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14524
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Explicit_Deref_F_Prefix, Info => Ent.Info);
--# expr-done 9405
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_3 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_2.Node, Origin => Origin);
--# end
--# expr-done 9406
And_Pred := Create_And (Bind_Result, Fld_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14523") else null)); 
--# expr-done 9407
--# expr-start 9410 '<Bind at ast.py:14525>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14525
--# expr-start 9408 '<FieldAccess for type_var at ast.py:14525>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14525
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 9408
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
--# expr-start 9409 '<FieldAccess for accessed_type at ast.py:14525>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14525
      if Typ.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_5 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Typ.Node, Origin => Origin, E_Info => Typ.Info);
--# end
--# expr-done 9409
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_4, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14525") else null)); 
--# expr-done 9410
And_Pred_1 := Create_And (And_Pred, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14523") else null)); 
--# expr-done 9411
   If_Result := And_Pred_1;
else
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14527") else null)); 
   If_Result := False_Rel;
end if;
      Inc_Ref (If_Result);
--# expr-done 9412
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 9413
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1044;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1044;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Explicit_Deref_P_Eq_For_Type;
--# end
--# property-start ExplicitDeref.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14531
pragma Warnings (Off, "is not referenced");
 function Explicit_Deref_P_Xref_Equation
  (Node : Bare_Explicit_Deref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Explicit_Deref := Bare_Explicit_Deref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Explicit_Deref :=
     Internal_Entity_Explicit_Deref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1045 with Inline_Always;
      Fld : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1045 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1045;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9415 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9414 '<FieldAccess for bottom_up_name_equation at ast.py:14532>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14532
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.bottom_up_name_equation
Fld := Libadalang.Implementation.Name_P_Bottom_Up_Name_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 9414
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 9415
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1045;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1045;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Explicit_Deref_P_Xref_Equation;
--# end
--# property-start ExplicitDeref.general_xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14535
pragma Warnings (Off, "is not referenced");
 function Explicit_Deref_P_General_Xref_Equation
  (Node : Bare_Explicit_Deref
      ; Root : Bare_Name
         := No_Bare_Ada_Node
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Explicit_Deref := Bare_Explicit_Deref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Explicit_Deref :=
     Internal_Entity_Explicit_Deref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind root Root
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1046 with Inline_Always;
            procedure Finalizer_Scope_2958 with Inline_Always;
            procedure Finalizer_Scope_2959 with Inline_Always;
      Block_Var : Internal_Entity_Array_Access;
Var_Expr : Internal_Entity_Base_Type_Decl;
Var_Expr_1 : Internal_Entity_Attribute_Ref;
Fld : Internal_Entity_Array_Access;
Fld_1 : Internal_Entity_Name;
Fld_2 : Logic_Equation;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Logic_Var;
Bind_Result : Logic_Equation;
Fld_5 : Logic_Var;
Fld_6 : Bare_Name;
Fld_7 : Logic_Var;
Bind_Result_1 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_8 : Internal_Entity_Base_Type_Decl;
Fld_9 : Internal_Entity_Name;
Cast_Result_1 : Internal_Entity_Attribute_Ref;
Fld_10 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
Fld_11 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Scope_Result_1 : Logic_Equation;
False_Rel : Logic_Equation;
Result_Var_1 : Logic_Equation;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
And_Pred_2 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1046 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1046;
            procedure Finalizer_Scope_2958 is
            begin
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (False_Rel);
                     Dec_Ref (Result_Var_1);
            end Finalizer_Scope_2958;
            procedure Finalizer_Scope_2959 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_11);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_2959;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9438 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9416 '<FieldAccess for env_elements at ast.py:14536>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14536
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.env_elements
Fld := Libadalang.Implementation.Expr_P_Env_Elements (Node => Ent.Node, Env => Env, No_Visibility => False, E_Info => Ent.Info);
--# end
--# expr-done 9416
Block_Var := Fld; Inc_Ref (Block_Var);
--# bind env_els Block_Var
--# expr-start 9437 '<And at ast.py:14538>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14538
--# expr-start 9418 '<FieldAccess for sub_equation at ast.py:14538>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14538
--# expr-start 9417 '<FieldAccess for prefix at ast.py:14538>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14538
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Ent.Node.Explicit_Deref_F_Prefix, Info => Ent.Info);
--# expr-done 9417
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_2 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_1.Node, Env => Env, Origin => Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 9418
--# expr-start 9436 '<LogicAny at ast.py:14538>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14538
--# expr-start 9435 '<Map at ast.py:14538>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14538
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Block_Var;
      begin
         for Item_290 of
               Collection.Items
         loop
   --# scope-start
               --# bind el Item_290
--# expr-start 9420 '<FieldAccess for expr_type at ast.py:14539>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14539
--# expr-start 9419 '<Cast to BasicDecl at ast.py:14539>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14539
Cast_Expr := Item_290; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 9419
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.expr_type'
Fld_3 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Expr_Type (Node => Cast_Result.Node, Origin => Origin, E_Info => Cast_Result.Info);
--# end
--# expr-done 9420
Var_Expr := Fld_3;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind typ Var_Expr
--# expr-start 9434 '<And at ast.py:14541>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14541
--# expr-start 9427 '<And at ast.py:14541>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14541
--# expr-start 9422 '<Bind at ast.py:14541>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14541
--# expr-start 9421 '<FieldAccess for ref_var at ast.py:14541>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14541
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_4 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Self);
--# end
--# expr-done 9421
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
Bind_Result := Solver.Create_Assign (Fld_4, Item_290, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14541") else null)); 
--# expr-done 9422
--# expr-start 9426 '<Bind at ast.py:14542>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14542
--# expr-start 9423 '<FieldAccess for ref_var at ast.py:14542>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14542
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_5 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Self);
--# end
--# expr-done 9423
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
--# expr-start 9425 '<FieldAccess for ref_var at ast.py:14542>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14542
--# expr-start 9424 '<FieldAccess for prefix at ast.py:14542>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14542
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Self.Explicit_Deref_F_Prefix;
--# expr-done 9424
      if Fld_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_7 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_6);
--# end
--# expr-done 9425
Fld_7.Value := No_Entity;
Entity_Vars.Reset (Fld_7);
Bind_Result_1 := Solver.Create_Unify (Fld_5, Fld_7, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14542") else null)); 
--# expr-done 9426
And_Pred := Create_And (Bind_Result, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14541") else null)); 
--# expr-done 9427
--# expr-start 9433 '<FieldAccess for parent_name_equation at ast.py:14543>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14543
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9432 '<If at ast.py:14544>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14544
--# expr-start 9429 '<Cast to AttributeRef at ast.py:14544>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14544
--# expr-start 9428 '<FieldAccess for prefix at ast.py:14544>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14544
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Name (Node => Ent.Node.Explicit_Deref_F_Prefix, Info => Ent.Info);
--# expr-done 9428
   if Fld_9.Node = null
      or else Fld_9.Node.Kind in Ada_Attribute_Ref_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Attribute_Ref
        (Node => Fld_9.Node,
         Info => Fld_9.Info);
   else
         Cast_Result_1 := No_Entity_Attribute_Ref;
   end if;
--# expr-done 9429
Var_Expr_1 := Cast_Result_1;
if Var_Expr_1 /= No_Entity_Attribute_Ref then
   --# scope-start
--# expr-start 9430 '<FieldAccess for is_access_attr at ast.py:14544>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14544
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AttributeRef.is_access_attr
Fld_10 := Libadalang.Implementation.Attribute_Ref_P_Is_Access_Attr (Node => Var_Expr_1.Node);
--# end
--# expr-done 9430
Scope_Result := Fld_10;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
if Result_Var then
   --# expr-start 9431 '<FieldAccess for anonymous_access_type at ast.py:14545>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14545
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.anonymous_access_type
Fld_8 := Libadalang.Implementation.Base_Type_Decl_P_Anonymous_Access_Type (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 9431
   If_Result := Fld_8;
else
   If_Result := Var_Expr;
end if;
--# expr-done 9432
--# property-call-start Name.parent_name_equation
Fld_11 := Libadalang.Implementation.Name_P_Parent_Name_Equation (Node => Ent.Node, Typ => If_Result, Root => Root, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 9433
And_Pred_1 := Create_And (And_Pred, Fld_11, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14541") else null)); 
--# expr-done 9434
Scope_Result_1 := And_Pred_1;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2959;
   Result_Var_1 := Scope_Result_1;
else
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14549") else null)); 
   Result_Var_1 := False_Rel;
end if;
      Inc_Ref (Result_Var_1);
         declare
            Item_To_Append : constant Logic_Equation := Result_Var_1;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2958;
         end loop;
      end;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 9435
Logic_Boolean_Op := Solver.Create_Any (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14538") else null)); 
--# expr-done 9436
And_Pred_2 := Create_And (Fld_2, Logic_Boolean_Op, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14538") else null)); 
--# expr-done 9437
Let_Result := And_Pred_2; Inc_Ref (Let_Result);
--# end
--# expr-done 9438
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1046;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1046;
               Finalizer_Scope_2958;
               Finalizer_Scope_2959;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Explicit_Deref_P_General_Xref_Equation;
--# end
--# property-start ExplicitDeref.is_constant /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14554
pragma Warnings (Off, "is not referenced");
 function Explicit_Deref_P_Is_Constant
  (Node : Bare_Explicit_Deref
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Explicit_Deref := Bare_Explicit_Deref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Explicit_Deref :=
     Internal_Entity_Explicit_Deref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Type_Access_Def;
Var_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Expr : Bare_Explicit_Deref;
Cast_Result : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Access_Def;
Scope_Result : Internal_Entity_Access_Def;
Result_Var : Internal_Entity_Access_Def;
Cast_Result_1 : Internal_Entity_Type_Access_Def;
Fld_3 : Internal_Entity_Constant_Node;
Fld_4 : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
Dyn_Var_Bind_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9446 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9445 '<bind at ast.py:14557>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14557
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Bound_Origin := Cast_Result; 
--# expr-start 9442 '<Cast to TypeAccessDef at ast.py:14559>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14559
--# expr-start 9440 '<FieldAccess for expression_type at ast.py:14559>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14559
--# expr-start 9439 '<FieldAccess for prefix at ast.py:14559>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14559
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Explicit_Deref_F_Prefix, Info => Ent.Info);
--# expr-done 9439
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.expression_type
Fld_1 := Libadalang.Implementation.Expr_P_Expression_Type (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 9440
Var_Expr_1 := Fld_1;
if Var_Expr_1 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 9441 '<FieldAccess for access_def at ast.py:14559>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14559
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.access_def'
Fld_2 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Access_Def (Node => Var_Expr_1.Node, Origin => Bound_Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 9441
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Access_Def;
end if;
   if Result_Var.Node = null
      or else Result_Var.Node.Kind in Ada_Type_Access_Def_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Type_Access_Def
        (Node => Result_Var.Node,
         Info => Result_Var.Info);
   else
         Cast_Result_1 := No_Entity_Type_Access_Def;
   end if;
--# expr-done 9442
Var_Expr := Cast_Result_1;
if Var_Expr /= No_Entity_Type_Access_Def then
   --# scope-start
--# expr-start 9444 '<FieldAccess for as_bool at ast.py:14559>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14559
--# expr-start 9443 '<FieldAccess for has_constant at ast.py:14559>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14559
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Constant_Node (Node => Var_Expr.Node.Type_Access_Def_F_Has_Constant, Info => Var_Expr.Info);
--# expr-done 9443
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Constant.as_bool'
Fld_4 := Libadalang.Implementation.Dispatcher_Constant_Node_P_As_Bool (Node => Fld_3.Node);
--# end
--# expr-done 9444
Scope_Result_1 := Fld_4;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
Dyn_Var_Bind_Result := Result_Var_1; 
--# expr-done 9445
Let_Result := Dyn_Var_Bind_Result; 
--# end
--# expr-done 9446
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Explicit_Deref_P_Is_Constant;
--# end
   --
   --  Primitives for Bare_Qual_Expr
   --
      procedure Initialize_Fields_For_Qual_Expr
        (Self : Bare_Qual_Expr
         ; Qual_Expr_F_Prefix : Bare_Name
         ; Qual_Expr_F_Suffix : Bare_Expr
        ) is
      begin
            Initialize_Fields_For_Name
              (Self);
            Self.Qual_Expr_F_Prefix := Qual_Expr_F_Prefix;
            Self.Qual_Expr_F_Suffix := Qual_Expr_F_Suffix;
      end Initialize_Fields_For_Qual_Expr;
   function Qual_Expr_F_Prefix
     (Node : Bare_Qual_Expr) return Bare_Name
   is
   begin
         return Node.Qual_Expr_F_Prefix;
   end;
   function Qual_Expr_F_Suffix
     (Node : Bare_Qual_Expr) return Bare_Expr
   is
   begin
         return Node.Qual_Expr_F_Suffix;
   end;
--# property-start QualExpr.ref_var /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17209
pragma Warnings (Off, "is not referenced");
 function Qual_Expr_P_Ref_Var
  (Node : Bare_Qual_Expr
  )
   return Logic_Var
is
   Self : Bare_Qual_Expr := Bare_Qual_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Logic_Var;
      Fld : Bare_Name;
Fld_1 : Logic_Var;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9448 '<FieldAccess for ref_var at ast.py:17209>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17209
--# expr-start 9447 '<FieldAccess for prefix at ast.py:17209>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17209
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Qual_Expr_F_Prefix;
--# expr-done 9447
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld);
--# end
--# expr-done 9448
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Qual_Expr_P_Ref_Var;
--# end
--# property-start QualExpr.relative_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17211
pragma Warnings (Off, "is not referenced");
 function Qual_Expr_P_Relative_Name
  (Node : Bare_Qual_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Single_Tok_Node
is
   Self : Bare_Qual_Expr := Bare_Qual_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Qual_Expr :=
     Internal_Entity_Qual_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Single_Tok_Node;
      Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Single_Tok_Node;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9450 '<FieldAccess for relative_name at ast.py:17211>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17211
--# expr-start 9449 '<FieldAccess for prefix at ast.py:17211>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17211
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Qual_Expr_F_Prefix, Info => Ent.Info);
--# expr-done 9449
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.relative_name'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Relative_Name (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 9450
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Qual_Expr_P_Relative_Name;
--# end
--# property-start QualExpr.is_constant /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17213
pragma Warnings (Off, "is not referenced");
 function Qual_Expr_P_Is_Constant
  (Node : Bare_Qual_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Qual_Expr := Bare_Qual_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Qual_Expr_P_Is_Constant;
--# end
--# property-start QualExpr.general_xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17216
pragma Warnings (Off, "is not referenced");
 function Qual_Expr_P_General_Xref_Equation
  (Node : Bare_Qual_Expr
      ; Root : Bare_Name
         := No_Bare_Ada_Node
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Qual_Expr := Bare_Qual_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Qual_Expr :=
     Internal_Entity_Qual_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind root Root
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1232 with Inline_Always;
            procedure Finalizer_Scope_2963 with Inline_Always;
      Var_Expr : Internal_Entity_Name;
Fld : Logic_Equation;
Fld_1 : Bare_Name;
As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld_2 : Internal_Entity_Name;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1232 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1232;
            procedure Finalizer_Scope_2963 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_2963;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9458 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9457 '<And at ast.py:17217>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17217
--# expr-start 9451 '<FieldAccess for xref_equation at ast.py:17218>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17218
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 9451
--# expr-start 9453 '<as_entity at ast.py:17219>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17219
--# expr-start 9452 '<FieldAccess for parent_name at ast.py:17219>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17219
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.parent_name
Fld_1 := Libadalang.Implementation.Name_P_Parent_Name (Node => Self, Stop_At => Root);
--# end
--# expr-done 9452
Is_Null := Fld_1 = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
As_Entity := (Info => E_Info, Node => Fld_1); 
   If_Result := As_Entity;
end if;
--# expr-done 9453
Var_Expr := If_Result;
if Var_Expr /= No_Entity_Name then
   --# scope-start
   --# bind pn Var_Expr
--# expr-start 9456 '<FieldAccess for parent_name_equation at ast.py:17221>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17221
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9455 '<FieldAccess for designated_type_impl at ast.py:17222>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17222
--# expr-start 9454 '<FieldAccess for prefix at ast.py:17222>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17222
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Qual_Expr_F_Prefix, Info => Ent.Info);
--# expr-done 9454
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.designated_type_impl'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Designated_Type_Impl (Node => Fld_2.Node, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 9455
--# property-call-start Name.parent_name_equation
Fld_4 := Libadalang.Implementation.Name_P_Parent_Name_Equation (Node => Var_Expr.Node, Typ => Fld_3, Root => Root, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 9456
Scope_Result := Fld_4;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2963;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17225") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred := Create_And (Fld, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17217") else null)); 
--# expr-done 9457
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 9458
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1232;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1232;
               Finalizer_Scope_2963;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Qual_Expr_P_General_Xref_Equation;
--# end
--# property-start QualExpr.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17230
pragma Warnings (Off, "is not referenced");
 function Qual_Expr_P_Xref_Equation
  (Node : Bare_Qual_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Qual_Expr := Bare_Qual_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Qual_Expr :=
     Internal_Entity_Qual_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1233 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Expr;
Fld_3 : Logic_Equation;
Fld_4 : Bare_Name;
Fld_5 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_6 : Bare_Expr;
Fld_7 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_8 : Internal_Entity_Expr;
Fld_9 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_10 : Logic_Var;
Cast_Expr_2 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result_2 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1233 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_9);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1233;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9477 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9460 '<FieldAccess for designated_type_impl at ast.py:17231>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17231
--# expr-start 9459 '<FieldAccess for prefix at ast.py:17231>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17231
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Qual_Expr_F_Prefix, Info => Ent.Info);
--# expr-done 9459
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.designated_type_impl'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Designated_Type_Impl (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 9460
Block_Var := Fld_1; 
--# bind typ Block_Var
--# expr-start 9476 '<And at ast.py:17234>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17234
--# expr-start 9473 '<And at ast.py:17234>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17234
--# expr-start 9470 '<And at ast.py:17234>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17234
--# expr-start 9466 '<And at ast.py:17234>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17234
--# expr-start 9462 '<FieldAccess for sub_equation at ast.py:17234>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17234
--# expr-start 9461 '<FieldAccess for suffix at ast.py:17234>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17234
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Expr (Node => Ent.Node.Qual_Expr_F_Suffix, Info => Ent.Info);
--# expr-done 9461
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_2.Node, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 9462
--# expr-start 9465 '<Bind at ast.py:17235>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17235
--# expr-start 9464 '<FieldAccess for ref_var at ast.py:17235>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17235
--# expr-start 9463 '<FieldAccess for prefix at ast.py:17235>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17235
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Self.Qual_Expr_F_Prefix;
--# expr-done 9463
      if Fld_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_5 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_4);
--# end
--# expr-done 9464
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
Cast_Expr := Block_Var; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_5, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17235") else null)); 
--# expr-done 9465
And_Pred := Create_And (Fld_3, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17234") else null)); 
--# expr-done 9466
--# expr-start 9469 '<Bind at ast.py:17236>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17236
--# expr-start 9468 '<FieldAccess for expected_type_var at ast.py:17236>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17236
--# expr-start 9467 '<FieldAccess for suffix at ast.py:17236>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17236
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Self.Qual_Expr_F_Suffix;
--# expr-done 9467
      if Fld_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_6);
--# expr-done 9468
Fld_7.Value := No_Entity;
Entity_Vars.Reset (Fld_7);
Cast_Expr_1 := Block_Var; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_7, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17236") else null)); 
--# expr-done 9469
And_Pred_1 := Create_And (And_Pred, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17234") else null)); 
--# expr-done 9470
--# expr-start 9472 '<FieldAccess for matches_expected_type at ast.py:17237>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17237
--# expr-start 9471 '<FieldAccess for suffix at ast.py:17237>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17237
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Expr (Node => Ent.Node.Qual_Expr_F_Suffix, Info => Ent.Info);
--# expr-done 9471
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_9 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_8.Node, Origin => Origin);
--# end
--# expr-done 9472
And_Pred_2 := Create_And (And_Pred_1, Fld_9, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17234") else null)); 
--# expr-done 9473
--# expr-start 9475 '<Bind at ast.py:17238>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17238
--# expr-start 9474 '<FieldAccess for type_var at ast.py:17238>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17238
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 9474
Fld_10.Value := No_Entity;
Entity_Vars.Reset (Fld_10);
Cast_Expr_2 := Block_Var; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
Bind_Result_2 := Solver.Create_Assign (Fld_10, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17238") else null)); 
--# expr-done 9475
And_Pred_3 := Create_And (And_Pred_2, Bind_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17234") else null)); 
--# expr-done 9476
Let_Result := And_Pred_3; Inc_Ref (Let_Result);
--# end
--# expr-done 9477
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1233;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1233;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Qual_Expr_P_Xref_Equation;
--# end
--# property-start QualExpr.designated_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17244
pragma Warnings (Off, "is not referenced");
 function Qual_Expr_P_Designated_Type
  (Node : Bare_Qual_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Qual_Expr := Bare_Qual_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Qual_Expr :=
     Internal_Entity_Qual_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1234 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_1 : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_2 : Internal_Entity_Name;
Fld_3 : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result_1 : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1234 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Bound_Env);
            end Finalizer_Scope_1234;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9483 '<bind at ast.py:17245>' Dyn_Var_Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17245
--# expr-start 9478 '<FieldAccess for node_env at ast.py:17245>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17245
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Node_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 9478
Bound_Env := Fld; Inc_Ref (Bound_Env);
--# expr-start 9482 '<bind at ast.py:17246>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17246
--# expr-start 9479 '<FieldAccess for origin_node at ast.py:17246>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17246
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_1 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 9479
Bound_Origin := Fld_1; 
--# expr-start 9481 '<FieldAccess for designated_type_impl at ast.py:17247>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17247
--# expr-start 9480 '<FieldAccess for prefix at ast.py:17247>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17247
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Qual_Expr_F_Prefix, Info => Ent.Info);
--# expr-done 9480
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.designated_type_impl'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Designated_Type_Impl (Node => Fld_2.Node, Env => Bound_Env, Origin => Bound_Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 9481
Dyn_Var_Bind_Result := Fld_3; 
--# expr-done 9482
Dyn_Var_Bind_Result_1 := Dyn_Var_Bind_Result; 
--# expr-done 9483
      Property_Result := Dyn_Var_Bind_Result_1;
   --# end
      Finalizer_Scope_1234;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1234;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Qual_Expr_P_Designated_Type;
--# end
--# property-start QualExpr.designated_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17251
pragma Warnings (Off, "is not referenced");
 function Qual_Expr_P_Designated_Env
  (Node : Bare_Qual_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
      ; No_Visibility : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Qual_Expr := Bare_Qual_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Qual_Expr :=
     Internal_Entity_Qual_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   --# bind no_visibility No_Visibility
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1235 with Inline_Always;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1235 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1235;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9486 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9485 '<FieldAccess for defining_env at ast.py:17252>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17252
--# expr-start 9484 '<FieldAccess for designated_type at ast.py:17252>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17252
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start QualExpr.designated_type
Fld := Libadalang.Implementation.Qual_Expr_P_Designated_Type (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 9484
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Fld.Node, Origin => Origin, Include_Ud_Indexing => False, E_Info => Fld.Info);
--# end
--# expr-done 9485
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 9486
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1235;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1235;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Qual_Expr_P_Designated_Env;
--# end
--# property-start QualExpr.env_elements_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17255
pragma Warnings (Off, "is not referenced");
 function Qual_Expr_P_Env_Elements_Impl
  (Node : Bare_Qual_Expr
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_Qual_Expr := Bare_Qual_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Qual_Expr :=
     Internal_Entity_Qual_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   Property_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1236 with Inline_Always;
      Fld : Internal_Entity_Name;
Fld_1 : Internal_Entity_Array_Access;
Let_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1236 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1236;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9489 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9488 '<FieldAccess for env_elements_impl at ast.py:17256>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17256
--# expr-start 9487 '<FieldAccess for prefix at ast.py:17256>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17256
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Qual_Expr_F_Prefix, Info => Ent.Info);
--# expr-done 9487
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.env_elements_impl'
Fld_1 := Libadalang.Implementation.Dispatcher_Expr_P_Env_Elements_Impl (Node => Fld.Node, Env => Env, E_Info => Fld.Info);
--# end
--# expr-done 9488
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 9489
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1236;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1236;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Qual_Expr_P_Env_Elements_Impl;
--# end
   --
   --  Primitives for Bare_Reduce_Attribute_Ref
   --
      procedure Initialize_Fields_For_Reduce_Attribute_Ref
        (Self : Bare_Reduce_Attribute_Ref
         ; Reduce_Attribute_Ref_F_Prefix : Bare_Ada_Node
         ; Reduce_Attribute_Ref_F_Attribute : Bare_Identifier
         ; Reduce_Attribute_Ref_F_Args : Bare_Basic_Assoc_List
        ) is
      begin
            Initialize_Fields_For_Name
              (Self);
            Self.Reduce_Attribute_Ref_F_Prefix := Reduce_Attribute_Ref_F_Prefix;
            Self.Reduce_Attribute_Ref_F_Attribute := Reduce_Attribute_Ref_F_Attribute;
            Self.Reduce_Attribute_Ref_F_Args := Reduce_Attribute_Ref_F_Args;
      Self.Reduce_Attribute_Ref_R_Ref_Var := Null_Var_Record;
      end Initialize_Fields_For_Reduce_Attribute_Ref;
   function Reduce_Attribute_Ref_F_Prefix
     (Node : Bare_Reduce_Attribute_Ref) return Bare_Ada_Node
   is
   begin
         return Node.Reduce_Attribute_Ref_F_Prefix;
   end;
   function Reduce_Attribute_Ref_F_Attribute
     (Node : Bare_Reduce_Attribute_Ref) return Bare_Identifier
   is
   begin
         return Node.Reduce_Attribute_Ref_F_Attribute;
   end;
   function Reduce_Attribute_Ref_F_Args
     (Node : Bare_Reduce_Attribute_Ref) return Bare_Basic_Assoc_List
   is
   begin
         return Node.Reduce_Attribute_Ref_F_Args;
   end;
--# property-start ReduceAttributeRef.ref_var /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18128
pragma Warnings (Off, "is not referenced");
 function Reduce_Attribute_Ref_P_Ref_Var
  (Node : Bare_Reduce_Attribute_Ref
  )
   return Logic_Var
is
   Self : Bare_Reduce_Attribute_Ref := Bare_Reduce_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Logic_Var;
      Fld : Logic_Var;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9490 '<FieldAccess for r_ref_var at ast.py:18128>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18128
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Reduce_Attribute_Ref_R_Ref_Var'Unrestricted_Access;
--# expr-done 9490
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Reduce_Attribute_Ref_P_Ref_Var;
--# end
--# property-start ReduceAttributeRef.args_list /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18131
pragma Warnings (Off, "is not referenced");
 function Reduce_Attribute_Ref_P_Args_List
  (Node : Bare_Reduce_Attribute_Ref
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Assoc_List
is
   Self : Bare_Reduce_Attribute_Ref := Bare_Reduce_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Reduce_Attribute_Ref :=
     Internal_Entity_Reduce_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Assoc_List;
      Var_Expr : Internal_Entity_Basic_Assoc_List;
Fld : Internal_Entity_Basic_Assoc_List;
Cast_Expr : Internal_Entity_Basic_Assoc_List;
Cast_Result : Internal_Entity_Assoc_List;
Scope_Result : Internal_Entity_Assoc_List;
Result_Var : Internal_Entity_Assoc_List;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 9491 '<FieldAccess for args at ast.py:18131>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18131
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Basic_Assoc_List (Node => Ent.Node.Reduce_Attribute_Ref_F_Args, Info => Ent.Info);
--# expr-done 9491
Var_Expr := Fld;
if Var_Expr /= No_Entity_Basic_Assoc_List then
   --# scope-start
--# expr-start 9492 '<Cast to AssocList at ast.py:18131>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18131
Cast_Expr := Var_Expr; 
      Cast_Result := Create_Internal_Entity_Assoc_List
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
--# expr-done 9492
Scope_Result := Cast_Result;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Assoc_List;
end if;
      Property_Result := Result_Var;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Reduce_Attribute_Ref_P_Args_List;
--# end
--# property-start ReduceAttributeRef.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18134
pragma Warnings (Off, "is not referenced");
 function Reduce_Attribute_Ref_P_Xref_Equation
  (Node : Bare_Reduce_Attribute_Ref
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Reduce_Attribute_Ref := Bare_Reduce_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Reduce_Attribute_Ref :=
     Internal_Entity_Reduce_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1281 with Inline_Always;
            procedure Finalizer_Scope_2965 with Inline_Always;
            procedure Finalizer_Scope_2967 with Inline_Always;
      Block_Var : Internal_Entity_Expr;
Var_Expr : Boolean;
Var_Expr_1 : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Assoc_List;
Fld_1 : Bare_Assoc_List;
Fld_2 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_3 : Internal_Entity_Expr;
Cast_Expr : Internal_Entity_Expr;
Cast_Result : Internal_Entity_Base_Id;
Fld_4 : Symbol_Type;
Fld_5 : Bare_Ada_Node;
Fld_6 : Internal_Entity_Array_Access;
Cast_Expr_1 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Basic_Decl;
Fld_7 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Cast_Expr_2 : Internal_Entity;
Cast_Result_2 : Internal_Entity_Basic_Decl;
Fld_8 : Logic_Equation;
Scope_Result_1 : Logic_Equation;
False_Rel : Logic_Equation;
Result_Var_1 : Logic_Equation;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1281 is
            begin
                     Dec_Ref (Fld_6);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1281;
            procedure Finalizer_Scope_2965 is
            begin
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (False_Rel);
                     Dec_Ref (Result_Var_1);
            end Finalizer_Scope_2965;
            procedure Finalizer_Scope_2967 is
            begin
                     Dec_Ref (Fld_8);
            end Finalizer_Scope_2967;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9505 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9494 '<FieldAccess for expr at ast.py:18143>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18143
--# expr-start 9493 '<FieldAccess for args_list at ast.py:18143>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18143
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ReduceAttributeRef.args_list
Fld := Libadalang.Implementation.Reduce_Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 9493
Fld_1 := Fld.Node;
Get_Result := Get (Fld_1, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Get_Result); 
   If_Result := As_Entity;
end if;
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_3 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 9494
Block_Var := Fld_3; 
--# bind reducer Block_Var
--# expr-start 9504 '<LogicAny at ast.py:18145>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18145
--# expr-start 9503 '<Map at ast.py:18145>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18145
   --# expr-start 9498 '<FieldAccess for env_get at ast.py:18145>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18145
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9496 '<FieldAccess for sym at ast.py:18147>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18147
--# expr-start 9495 '<Cast to BaseId at ast.py:18147>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18147
Cast_Expr := Block_Var; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Base_Id
   then
      Cast_Result := Create_Internal_Entity_Base_Id
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Base_Id;
   end if;
--# expr-done 9495
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTokNode.sym
Fld_4 := Libadalang.Implementation.Single_Tok_Node_P_Sym (Node => Cast_Result.Node);
--# end
--# expr-done 9496
--# expr-start 9497 '<FieldAccess for origin_node at ast.py:18148>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18148
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_5 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 9497
--# property-call-start AdaNode.env_get
Fld_6 := Libadalang.Implementation.Ada_Node_P_Env_Get (Node => Self, Env => Env, Symbol => Fld_4, From_Node => Fld_5);
--# end
--# expr-done 9498
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Fld_6;
      begin
         for Item_291 of
               Collection.Items
         loop
   --# scope-start
               --# bind subp Item_291
--# expr-start 9499 '<Cast to BasicDecl at ast.py:18151>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18151
Cast_Expr_1 := Item_291; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_1 := No_Entity_Basic_Decl;
   end if;
--# expr-done 9499
Var_Expr_1 := Cast_Result_1;
if Var_Expr_1 /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 9500 '<FieldAccess for is_valid_reducer_candidate at ast.py:18151>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18151
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_valid_reducer_candidate
Fld_7 := Libadalang.Implementation.Basic_Decl_P_Is_Valid_Reducer_Candidate (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 9500
Scope_Result := Fld_7;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Var_Expr := Result_Var;
if Var_Expr /= False then
   --# scope-start
   --# bind _ Var_Expr
--# expr-start 9502 '<FieldAccess for xref_equation_for_reducer_candidate at ast.py:18153>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18153
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9501 '<Cast to BasicDecl at ast.py:18153>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18153
Cast_Expr_2 := Item_291; 
   if Cast_Expr_2.Node = null
      or else Cast_Expr_2.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_2 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
   else
         Cast_Result_2 := No_Entity_Basic_Decl;
   end if;
--# expr-done 9501
--# property-call-start ReduceAttributeRef.xref_equation_for_reducer_candidate
Fld_8 := Libadalang.Implementation.Reduce_Attribute_Ref_P_Xref_Equation_For_Reducer_Candidate (Node => Ent.Node, Subp => Cast_Result_2, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 9502
Scope_Result_1 := Fld_8;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2967;
   Result_Var_1 := Scope_Result_1;
else
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18155") else null)); 
   Result_Var_1 := False_Rel;
end if;
      Inc_Ref (Result_Var_1);
         declare
            Item_To_Append : constant Logic_Equation := Result_Var_1;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2965;
         end loop;
      end;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 9503
Logic_Boolean_Op := Solver.Create_Any (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18145") else null)); 
--# expr-done 9504
Let_Result := Logic_Boolean_Op; Inc_Ref (Let_Result);
--# end
--# expr-done 9505
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1281;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1281;
               Finalizer_Scope_2965;
               Finalizer_Scope_2967;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Reduce_Attribute_Ref_P_Xref_Equation;
--# end
--# property-start ReduceAttributeRef.xref_equation_for_reducer_candidate /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18160
pragma Warnings (Off, "is not referenced");
 function Reduce_Attribute_Ref_P_Xref_Equation_For_Reducer_Candidate
  (Node : Bare_Reduce_Attribute_Ref
      ; Subp : Internal_Entity_Basic_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Reduce_Attribute_Ref := Bare_Reduce_Attribute_Ref (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Reduce_Attribute_Ref :=
     Internal_Entity_Reduce_Attribute_Ref'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind subp Subp
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1282 with Inline_Always;
            procedure Finalizer_Scope_2968 with Inline_Always;
      Block_Var : Internal_Entity_Base_Id;
Block_Var_1 : Internal_Entity_Expr;
Block_Var_2 : Internal_Entity_Base_Type_Decl_Array_Access;
Block_Var_3 : Internal_Entity_Base_Type_Decl;
Block_Var_4 : Internal_Entity_Base_Type_Decl;
Var_Expr : Internal_Entity_Value_Sequence;
Fld : Internal_Entity_Assoc_List;
Fld_1 : Bare_Assoc_List;
Fld_2 : Internal_Entity_Info;
Get_Result : Bare_Basic_Assoc;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Assoc;
Fld_3 : Internal_Entity_Expr;
Cast_Result : Internal_Entity_Base_Id;
Fld_4 : Internal_Entity_Assoc_List;
Fld_5 : Bare_Assoc_List;
Fld_6 : Internal_Entity_Info;
Get_Result_1 : Bare_Basic_Assoc;
As_Entity_1 : Internal_Entity_Basic_Assoc;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Basic_Assoc;
Fld_7 : Internal_Entity_Expr;
Fld_8 : Internal_Entity_Base_Subp_Spec;
Fld_9 : Internal_Entity_Base_Type_Decl_Array_Access;
Get_Result_2 : Internal_Entity_Base_Type_Decl;
Get_Result_3 : Internal_Entity_Base_Type_Decl;
Fld_10 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Value_Sequence;
Fld_11 : Internal_Entity_Iterated_Assoc;
Fld_12 : Internal_Entity_Expr;
Fld_13 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_14 : Internal_Entity_Iterated_Assoc;
Fld_15 : Internal_Entity_Expr;
Fld_16 : Logic_Equation;
And_Pred : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
Fld_17 : Internal_Entity;
Fld_18 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_19 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_20 : Logic_Var;
Cast_Expr_2 : Internal_Entity_Basic_Decl;
Cast_Result_4 : Internal_Entity;
Bind_Result_2 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_21 : Logic_Var;
Cast_Expr_3 : Internal_Entity_Base_Type_Decl;
Cast_Result_5 : Internal_Entity;
Bind_Result_3 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Fld_22 : Logic_Equation;
And_Pred_5 : Logic_Equation;
Fld_23 : Logic_Equation;
And_Pred_6 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1282 is
            begin
                     Dec_Ref (Fld_9);
                     Dec_Ref (Block_Var_2);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Fld_18);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Bind_Result_3);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Fld_22);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (Fld_23);
                     Dec_Ref (And_Pred_6);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1282;
            procedure Finalizer_Scope_2968 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_16);
                     Dec_Ref (And_Pred);
            end Finalizer_Scope_2968;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9541 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9508 '<Cast to BaseId at ast.py:18164>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18164
--# expr-start 9507 '<FieldAccess for expr at ast.py:18164>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18164
--# expr-start 9506 '<FieldAccess for args_list at ast.py:18164>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18164
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ReduceAttributeRef.args_list
Fld := Libadalang.Implementation.Reduce_Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 9506
Fld_1 := Fld.Node;
Get_Result := Get (Fld_1, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Assoc;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Get_Result); 
   If_Result := As_Entity;
end if;
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_3 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 9507
   if Fld_3.Node = null
      or else Fld_3.Node.Kind in Ada_Base_Id
   then
      Cast_Result := Create_Internal_Entity_Base_Id
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
   else
         Cast_Result := No_Entity_Base_Id;
   end if;
--# expr-done 9508
Block_Var := Cast_Result; 
--# bind reducer Block_Var
--# expr-start 9510 '<FieldAccess for expr at ast.py:18165>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18165
--# expr-start 9509 '<FieldAccess for args_list at ast.py:18165>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18165
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ReduceAttributeRef.args_list
Fld_4 := Libadalang.Implementation.Reduce_Attribute_Ref_P_Args_List (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 9509
Fld_5 := Fld_4.Node;
Get_Result_1 := Get (Fld_5, 1, True); 
Is_Null_1 := Get_Result_1 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Basic_Assoc;
else
Fld_6 := Fld_4.Info;
As_Entity_1 := (Info => Fld_6, Node => Get_Result_1); 
   If_Result_1 := As_Entity_1;
end if;
      if If_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_7 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => If_Result_1.Node, E_Info => If_Result_1.Info);
--# end
--# expr-done 9510
Block_Var_1 := Fld_7; 
--# bind initial_value_expression Block_Var_1
--# expr-start 9512 '<FieldAccess for param_types at ast.py:18170>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18170
--# expr-start 9511 '<FieldAccess for subp_spec_or_null at ast.py:18170>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18170
      if Subp.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_8 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Subp.Node, E_Info => Subp.Info);
--# end
--# expr-done 9511
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.param_types
Fld_9 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Param_Types (Node => Fld_8.Node, Origin => Origin, E_Info => Fld_8.Info);
--# end
--# expr-done 9512
Block_Var_2 := Fld_9; Inc_Ref (Block_Var_2);
--# bind param_types Block_Var_2
--# expr-start 9513 '<at at ast.py:18171>' Get_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18171
Get_Result_2 := Get (Block_Var_2, 0, True); 
--# expr-done 9513
Block_Var_3 := Get_Result_2; 
--# bind accum_type Block_Var_3
--# expr-start 9514 '<at at ast.py:18172>' Get_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18172
Get_Result_3 := Get (Block_Var_2, 1, True); 
--# expr-done 9514
Block_Var_4 := Get_Result_3; 
--# bind value_type Block_Var_4
--# expr-start 9540 '<And at ast.py:18174>' And_Pred_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18174
--# expr-start 9538 '<And at ast.py:18174>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18174
--# expr-start 9536 '<And at ast.py:18174>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18174
--# expr-start 9533 '<And at ast.py:18174>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18174
--# expr-start 9530 '<And at ast.py:18174>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18174
--# expr-start 9527 '<And at ast.py:18174>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18174
--# expr-start 9516 '<Cast to ValueSequence at ast.py:18175>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18175
--# expr-start 9515 '<FieldAccess for prefix at ast.py:18175>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18175
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Create_Internal_Entity (Node => Ent.Node.Reduce_Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 9515
   if Fld_10.Node = null
      or else Fld_10.Node.Kind in Ada_Value_Sequence_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Value_Sequence
        (Node => Fld_10.Node,
         Info => Fld_10.Info);
   else
         Cast_Result_1 := No_Entity_Value_Sequence;
   end if;
--# expr-done 9516
Var_Expr := Cast_Result_1;
if Var_Expr /= No_Entity_Value_Sequence then
   --# scope-start
   --# bind vs Var_Expr
--# expr-start 9524 '<And at ast.py:18177>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18177
--# expr-start 9520 '<Bind at ast.py:18177>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18177
--# expr-start 9519 '<FieldAccess for expected_type_var at ast.py:18177>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18177
--# expr-start 9518 '<FieldAccess for expr at ast.py:18177>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18177
--# expr-start 9517 '<FieldAccess for iter_assoc at ast.py:18177>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18177
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Create_Internal_Entity_Iterated_Assoc (Node => Var_Expr.Node.Value_Sequence_F_Iter_Assoc, Info => Var_Expr.Info);
--# expr-done 9517
      if Fld_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_12 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_11.Node, E_Info => Fld_11.Info);
--# end
--# expr-done 9518
      if Fld_12.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_12.Node);
--# expr-done 9519
Fld_13.Value := No_Entity;
Entity_Vars.Reset (Fld_13);
Cast_Expr := Block_Var_4; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_13, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18177") else null)); 
--# expr-done 9520
--# expr-start 9523 '<FieldAccess for matches_expected_type at ast.py:18178>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18178
--# expr-start 9522 '<FieldAccess for expr at ast.py:18178>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18178
--# expr-start 9521 '<FieldAccess for iter_assoc at ast.py:18178>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18178
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Create_Internal_Entity_Iterated_Assoc (Node => Var_Expr.Node.Value_Sequence_F_Iter_Assoc, Info => Var_Expr.Info);
--# expr-done 9521
      if Fld_14.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_15 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_14.Node, E_Info => Fld_14.Info);
--# end
--# expr-done 9522
      if Fld_15.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_16 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_15.Node, Origin => Origin);
--# end
--# expr-done 9523
And_Pred := Create_And (Bind_Result, Fld_16, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18177") else null)); 
--# expr-done 9524
Scope_Result := And_Pred;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2968;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18179") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
--# expr-start 9526 '<FieldAccess for sub_equation at ast.py:18181>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18181
--# expr-start 9525 '<FieldAccess for prefix at ast.py:18181>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18181
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_17 := Create_Internal_Entity (Node => Ent.Node.Reduce_Attribute_Ref_F_Prefix, Info => Ent.Info);
--# expr-done 9525
      if Fld_17.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_18 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_17.Node, Env => Env, Origin => Origin, E_Info => Fld_17.Info);
--# end
--# expr-done 9526
And_Pred_1 := Create_And (Result_Var, Fld_18, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18174") else null)); 
--# expr-done 9527
--# expr-start 9529 '<Bind at ast.py:18183>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18183
--# expr-start 9528 '<FieldAccess for type_var at ast.py:18183>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18183
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 9528
Fld_19.Value := No_Entity;
Entity_Vars.Reset (Fld_19);
Cast_Expr_1 := Block_Var_3; 
      Cast_Result_3 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_19, Cast_Result_3, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18183") else null)); 
--# expr-done 9529
And_Pred_2 := Create_And (And_Pred_1, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18174") else null)); 
--# expr-done 9530
--# expr-start 9532 '<Bind at ast.py:18185>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18185
--# expr-start 9531 '<FieldAccess for ref_var at ast.py:18185>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18185
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_20 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Block_Var.Node);
--# end
--# expr-done 9531
Fld_20.Value := No_Entity;
Entity_Vars.Reset (Fld_20);
Cast_Expr_2 := Subp; 
      Cast_Result_4 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
Bind_Result_2 := Solver.Create_Assign (Fld_20, Cast_Result_4, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18185") else null)); 
--# expr-done 9532
And_Pred_3 := Create_And (And_Pred_2, Bind_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18174") else null)); 
--# expr-done 9533
--# expr-start 9535 '<Bind at ast.py:18187>' Bind_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18187
--# expr-start 9534 '<FieldAccess for expected_type_var at ast.py:18187>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18187
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_21 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Block_Var_1.Node);
--# expr-done 9534
Fld_21.Value := No_Entity;
Entity_Vars.Reset (Fld_21);
Cast_Expr_3 := Block_Var_3; 
      Cast_Result_5 := Create_Internal_Entity
        (Node => Cast_Expr_3.Node,
         Info => Cast_Expr_3.Info);
Bind_Result_3 := Solver.Create_Assign (Fld_21, Cast_Result_5, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18187") else null)); 
--# expr-done 9535
And_Pred_4 := Create_And (And_Pred_3, Bind_Result_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18174") else null)); 
--# expr-done 9536
--# expr-start 9537 '<FieldAccess for sub_equation at ast.py:18188>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18188
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_22 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Block_Var_1.Node, Env => Env, Origin => Origin, E_Info => Block_Var_1.Info);
--# end
--# expr-done 9537
And_Pred_5 := Create_And (And_Pred_4, Fld_22, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18174") else null)); 
--# expr-done 9538
--# expr-start 9539 '<FieldAccess for matches_expected_formal_type at ast.py:18189>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18189
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_type
Fld_23 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Type (Node => Block_Var_1.Node, Origin => Origin);
--# end
--# expr-done 9539
And_Pred_6 := Create_And (And_Pred_5, Fld_23, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18174") else null)); 
--# expr-done 9540
Let_Result := And_Pred_6; Inc_Ref (Let_Result);
--# end
--# expr-done 9541
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1282;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1282;
               Finalizer_Scope_2968;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Reduce_Attribute_Ref_P_Xref_Equation_For_Reducer_Candidate;
--# end
   --
   --  Primitives for Bare_Single_Tok_Node
   --
      procedure Initialize_Fields_For_Single_Tok_Node
        (Self : Bare_Single_Tok_Node
        ) is
      begin
            Initialize_Fields_For_Name
              (Self);
      end Initialize_Fields_For_Single_Tok_Node;
--# property-start SingleTokNode.relative_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14777
pragma Warnings (Off, "is not referenced");
 function Single_Tok_Node_P_Relative_Name
  (Node : Bare_Single_Tok_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Single_Tok_Node
is
   Self : Bare_Single_Tok_Node := Bare_Single_Tok_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Single_Tok_Node :=
     Internal_Entity_Single_Tok_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Single_Tok_Node;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := Ent;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Single_Tok_Node_P_Relative_Name;
--# end
--# property-start SingleTokNode.defines_subp_spec_var /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14789
pragma Warnings (Off, "is not referenced");
 function Single_Tok_Node_P_Defines_Subp_Spec_Var
  (Node : Bare_Single_Tok_Node
  )
   return Boolean
is
   Self : Bare_Single_Tok_Node := Bare_Single_Tok_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Single_Tok_Node_P_Defines_Subp_Spec_Var;
--# end
--# property-start SingleTokNode.sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14791
pragma Warnings (Off, "is not referenced");
 function Single_Tok_Node_P_Sym
  (Node : Bare_Single_Tok_Node
  )
   return Symbol_Type
is
   Self : Bare_Single_Tok_Node := Bare_Single_Tok_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type;
      Sym : Symbol_Type;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9542 '<GetSymbol at ast.py:14792>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14792
Sym := Get_Symbol (Self); 
--# expr-done 9542
      Property_Result := Sym;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Single_Tok_Node_P_Sym;
--# end
--# property-start SingleTokNode.canonical_text /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14801
pragma Warnings (Off, "is not referenced");
 function Single_Tok_Node_P_Canonical_Text
  (Node : Bare_Single_Tok_Node
  )
   return Symbol_Type
is
   Self : Bare_Single_Tok_Node := Bare_Single_Tok_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type;
      Fld : Symbol_Type;
Let_Result : Symbol_Type;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9544 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9543 '<FieldAccess for sym at ast.py:14802>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14802
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTokNode.sym
Fld := Libadalang.Implementation.Single_Tok_Node_P_Sym (Node => Self);
--# end
--# expr-done 9543
Let_Result := Fld; 
--# end
--# expr-done 9544
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Single_Tok_Node_P_Canonical_Text;
--# end
--# property-start SingleTokNode.env_get_first_visible /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14805
pragma Warnings (Off, "is not referenced");
 function Single_Tok_Node_P_Env_Get_First_Visible
  (Node : Bare_Single_Tok_Node
      ; Lex_Env : Lexical_Env
      ; Lookup_Type : Lookup_Kind
      ; From_Node : Bare_Ada_Node
      ; No_Visibility : Boolean
         := False
  )
   return Internal_Entity
is
   Self : Bare_Single_Tok_Node := Bare_Single_Tok_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind lex_env Lex_Env
   --# bind lookup_type Lookup_Type
   --# bind from_node From_Node
   --# bind no_visibility No_Visibility
   Property_Result : Internal_Entity;
            procedure Finalizer_Scope_1065 with Inline_Always;
      Sym : Symbol_Type;
Fld : Internal_Entity_Array_Access;
Fld_1 : Boolean;
If_Result : Boolean;
Find_Result : Internal_Entity;
Let_Result : Internal_Entity;
            procedure Finalizer_Scope_1065 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1065;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9549 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9548 '<Find at ast.py:14814>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14814
   --# expr-start 9546 '<FieldAccess for env_get at ast.py:14814>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14814
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9545 '<GetSymbol at ast.py:14816>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14816
Sym := Get_Symbol (Self); 
--# expr-done 9545
--# property-call-start AdaNode.env_get
Fld := Libadalang.Implementation.Ada_Node_P_Env_Get (Node => Self, Env => Lex_Env, Symbol => Sym, Lookup => Lookup_Type, From_Node => From_Node, Categories => (Inherited_Primitives => False, Nocat => True));
--# end
--# expr-done 9546
      Find_Result := No_Entity;
      declare
         Collection : constant Internal_Entity_Array_Access := Fld;
      begin
         for Item_292 of
               Collection.Items
         loop
   --# scope-start
               --# bind el Item_292
if No_Visibility then
   If_Result := True;
else
   --# expr-start 9547 '<FieldAccess for has_visibility at ast.py:14822>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14822
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.has_visibility
Fld_1 := Libadalang.Implementation.Ada_Node_P_Has_Visibility (Node => Self, Other_Entity => Item_292);
--# end
--# expr-done 9547
   If_Result := Fld_1;
end if;
      if If_Result then
         Find_Result := Item_292;
      end if;
   --# end
               exit when If_Result;
         end loop;
      end;
--# expr-done 9548
Let_Result := Find_Result; 
--# end
--# expr-done 9549
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1065;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1065;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Single_Tok_Node_P_Env_Get_First_Visible;
--# end
   --
   --  Primitives for Bare_Base_Id
   --
      procedure Initialize_Fields_For_Base_Id
        (Self : Bare_Base_Id
        ) is
      begin
            Initialize_Fields_For_Single_Tok_Node
              (Self);
      end Initialize_Fields_For_Base_Id;
--# property-start BaseId.scope /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15467
pragma Warnings (Off, "is not referenced");
 function Base_Id_P_Scope
  (Node : Bare_Base_Id
      ; Env : Lexical_Env
  )
   return Lexical_Env
is
   Self : Bare_Base_Id := Bare_Base_Id (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1113 with Inline_Always;
            procedure Finalizer_Scope_2969 with Inline_Always;
      Block_Var : Internal_Entity;
Block_Var_1 : Lexical_Env := Empty_Env;
Var_Expr : Bare_Generic_Package_Decl;
Sym : Symbol_Type;
Fld : Boolean;
If_Result : Lookup_Kind;
Env_Get_Result : Internal_Entity;
Fld_1 : Lexical_Env := Empty_Env;
Is_Null : Boolean;
Not_Val : Boolean;
Fld_2 : Bare_Ada_Node;
Is_A : Boolean;
If_Result_1 : Boolean;
If_Result_2 : Lexical_Env := Empty_Env;
Env_Node : Bare_Ada_Node;
Cast_Result : Bare_Generic_Package_Decl;
Fld_3 : Bare_Generic_Package_Internal;
Fld_4 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1113 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (If_Result_2);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1113;
            procedure Finalizer_Scope_2969 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_2969;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Id_Base_Id_P_Scope,
            Items    => new Mmz_Key_Array (1 ..  2))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Id,
                                As_Bare_Base_Id => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Lexical_Env,
                                      As_Lexical_Env => Env);
                  Inc_Ref (Mmz_K.Items (2).As_Lexical_Env);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Lexical_Env;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 9563 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9552 '<EnvGet at ast.py:15468>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15468
Sym := Get_Symbol (Self); 
--# expr-start 9551 '<If at ast.py:15470>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15470
--# expr-start 9550 '<FieldAccess for is_prefix at ast.py:15470>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15470
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_prefix
Fld := Libadalang.Implementation.Name_P_Is_Prefix (Node => Self);
--# end
--# expr-done 9550
if Fld then
   If_Result := Recursive;
else
   If_Result := Flat;
end if;
--# expr-done 9551
Env_Get_Result := AST_Envs.Get_First (Self => Env, Key => Sym, Lookup_Kind => To_Lookup_Kind_Type (If_Result), Categories => (Inherited_Primitives => False, Nocat => True)); 
--# expr-done 9552
Block_Var := Env_Get_Result; 
--# bind elt Block_Var
--# expr-start 9558 '<If at ast.py:15473>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15473
--# expr-start 9554 '<Not at ast.py:15474>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15474
--# expr-start 9553 '<IsNull at ast.py:15474>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15474
Is_Null := Block_Var.Node = null; 
--# expr-done 9553
Not_Val := not (Is_Null); 
--# expr-done 9554
if Not_Val then
   --# expr-start 9556 '<IsA BasicDecl at ast.py:15474>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15474
--# expr-start 9555 '<FieldAccess for node at ast.py:15474>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15474
Fld_2 := Block_Var.Node;
--# expr-done 9555
Is_A := Fld_2 /= null 
and then Fld_2.Kind in Ada_Basic_Decl; 
--# expr-done 9556
   If_Result_1 := Is_A;
else
   If_Result_1 := False;
end if;
if If_Result_1 then
   --# expr-start 9557 '<FieldAccess for children_env at ast.py:15477>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15477
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Children_Env (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# expr-done 9557
   If_Result_2 := Fld_1;
else
   If_Result_2 := Empty_Env;
end if;
      Inc_Ref (If_Result_2);
--# expr-done 9558
Block_Var_1 := If_Result_2; Inc_Ref (Block_Var_1);
--# bind ret Block_Var_1
--# expr-start 9560 '<Cast to GenericPackageDecl at ast.py:15483>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15483
--# expr-start 9559 '<env_node at ast.py:15483>' Env_Node /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15483
Env_Node := AST_Envs.Env_Node (Block_Var_1); 
--# expr-done 9559
   if Env_Node = null
      or else Env_Node.Kind in Ada_Generic_Package_Decl_Range
   then
      Cast_Result := Env_Node;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 9560
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind gen_pkg_decl Var_Expr
--# expr-start 9562 '<FieldAccess for children_env at ast.py:15484>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15484
--# expr-start 9561 '<FieldAccess for package_decl at ast.py:15484>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15484
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Var_Expr.Generic_Package_Decl_F_Package_Decl;
--# expr-done 9561
      if Fld_3 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Children_Env (Node => Fld_3);
--# expr-done 9562
Scope_Result := Fld_4;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2969;
   Result_Var := Scope_Result;
else
   Result_Var := Block_Var_1;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 9563
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1113;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Lexical_Env,
                        As_Lexical_Env => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1113;
               Finalizer_Scope_2969;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Id_P_Scope;
--# end
--# property-start BaseId.designated_env_no_overloading /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15489
pragma Warnings (Off, "is not referenced");
 function Base_Id_P_Designated_Env_No_Overloading
  (Node : Bare_Base_Id
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
      ; No_Visibility : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Base_Id := Bare_Base_Id (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Id :=
     Internal_Entity_Base_Id'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   --# bind no_visibility No_Visibility
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1114 with Inline_Always;
            procedure Finalizer_Scope_2970 with Inline_Always;
      Var_Expr : Internal_Entity_Basic_Decl;
Var_Expr_1 : Internal_Entity_Basic_Decl;
Fld : Boolean;
If_Result : Lookup_Kind;
Fld_1 : Bare_Ada_Node;
Fld_2 : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Fld_3 : Lexical_Env := Empty_Env;
Fld_4 : Lexical_Env := Empty_Env;
Fld_5 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result_1 : Lexical_Env := Empty_Env;
Scope_Result_1 : Lexical_Env := Empty_Env;
Result_Var_1 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1114 is
            begin
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1114;
            procedure Finalizer_Scope_2970 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Fld_4);
                     Dec_Ref (If_Result_1);
            end Finalizer_Scope_2970;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9573 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9568 '<Cast to BasicDecl at ast.py:15490>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15490
--# expr-start 9567 '<FieldAccess for env_get_first_visible at ast.py:15490>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15490
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9565 '<If at ast.py:15492>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15492
--# expr-start 9564 '<FieldAccess for is_prefix at ast.py:15492>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15492
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_prefix
Fld := Libadalang.Implementation.Name_P_Is_Prefix (Node => Self);
--# end
--# expr-done 9564
if Fld then
   If_Result := Recursive;
else
   If_Result := Flat;
end if;
--# expr-done 9565
--# expr-start 9566 '<FieldAccess for origin_node at ast.py:15493>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15493
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_1 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 9566
--# property-call-start SingleTokNode.env_get_first_visible
Fld_2 := Libadalang.Implementation.Single_Tok_Node_P_Env_Get_First_Visible (Node => Self, Lex_Env => Env, Lookup_Type => If_Result, From_Node => Fld_1, No_Visibility => No_Visibility);
--# end
--# expr-done 9567
   if Fld_2.Node = null
      or else Fld_2.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 9568
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind bd Var_Expr
--# expr-start 9572 '<If at ast.py:15495>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15495
Var_Expr_1 := Var_Expr;
if Var_Expr_1 /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 9569 '<FieldAccess for is_package at ast.py:15496>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15496
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.is_package
Fld_5 := Libadalang.Implementation.Ada_Node_P_Is_Package (Node => Var_Expr_1.Node);
--# end
--# expr-done 9569
Scope_Result := Fld_5;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
if Result_Var then
   --# expr-start 9570 '<FieldAccess for pkg_env at ast.py:15496>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15496
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseId.pkg_env
Fld_3 := Libadalang.Implementation.Base_Id_P_Pkg_Env (Node => Ent.Node, From_Pkg => Var_Expr, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 9570
   If_Result_1 := Fld_3;
else
   --# expr-start 9571 '<FieldAccess for defining_env at ast.py:15496>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15496
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_4 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Var_Expr.Node, Origin => Origin, Include_Ud_Indexing => False, E_Info => Var_Expr.Info);
--# end
--# expr-done 9571
   If_Result_1 := Fld_4;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 9572
Scope_Result_1 := If_Result_1;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2970;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := Empty_Env;
end if;
      Inc_Ref (Result_Var_1);
Let_Result := Result_Var_1; Inc_Ref (Let_Result);
--# end
--# expr-done 9573
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1114;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1114;
               Finalizer_Scope_2970;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Id_P_Designated_Env_No_Overloading;
--# end
--# property-start BaseId.designated_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15501
pragma Warnings (Off, "is not referenced");
 function Base_Id_P_Designated_Env
  (Node : Bare_Base_Id
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
      ; No_Visibility : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Base_Id := Bare_Base_Id (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Id :=
     Internal_Entity_Base_Id'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   --# bind no_visibility No_Visibility
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1115 with Inline_Always;
            procedure Finalizer_Scope_2980 with Inline_Always;
            procedure Finalizer_Scope_2982 with Inline_Always;
      Block_Var : Bare_Ada_Node;
Block_Var_1 : Internal_Entity;
Var_Expr : Internal_Entity_Basic_Decl;
Var_Expr_1 : Internal_Entity_Array_Access;
Node_Parents : Bare_Ada_Node_Array_Access;
Is_A : Boolean;
Find_Result : Bare_Ada_Node;
Fld : Boolean;
If_Result : Lookup_Kind;
Fld_1 : Bare_Ada_Node;
Fld_2 : Internal_Entity;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Fld_3 : Lexical_Env := Empty_Env;
Fld_4 : Internal_Entity_Array_Access;
Fld_5 : Bare_Ada_Node;
Cast_Result_1 : Bare_Ada_Node;
Is_Equal : Boolean;
Not_Val : Boolean;
Fld_6 : Boolean;
If_Result_1 : Boolean;
Map_Result : Internal_Entity_Array_Access;
Scope_Result : Internal_Entity_Array_Access;
Result_Var : Internal_Entity_Array_Access;
Cast_Expr_1 : Internal_Entity;
Cast_Result_2 : Internal_Entity_Basic_Decl;
Fld_7 : Lexical_Env := Empty_Env;
Map_Result_1 : Lexical_Env_Array_Access;
Group_Env : Lexical_Env := Empty_Env;
Cast_Expr_2 : Internal_Entity;
Cast_Result_3 : Internal_Entity_Basic_Decl;
Fld_8 : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
Cast_Expr_3 : Internal_Entity;
Cast_Result_4 : Internal_Entity_Basic_Decl;
Fld_9 : Bare_Basic_Decl;
Cast_Result_5 : Bare_Ada_Node;
Is_Equal_1 : Boolean;
Not_Val_1 : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1115 is
            begin
                     Dec_Ref (Node_Parents);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Group_Env);
                     Dec_Ref (If_Result_3);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1115;
            procedure Finalizer_Scope_2980 is
            begin
                     Dec_Ref (Var_Expr_1);
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_2980;
            procedure Finalizer_Scope_2982 is
            begin
                     Dec_Ref (Fld_7);
            end Finalizer_Scope_2982;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9600 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9576 '<Find at ast.py:15506>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15506
   --# expr-start 9574 '<FieldAccess for parents at ast.py:15506>' Node_Parents /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15506
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents := Parents (Self, True); 
--# expr-done 9574
      Find_Result := No_Bare_Ada_Node;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents;
      begin
         for Item_293 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_293
      --# expr-start 9575 '<IsA GenericPackageInstantiation at ast.py:15507>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15507
Is_A := Item_293 /= null 
and then Item_293.Kind in Ada_Generic_Package_Instantiation_Range; 
--# expr-done 9575
      if Is_A then
         Find_Result := Item_293;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 9576
Block_Var := Find_Result; 
--# bind bd Block_Var
--# expr-start 9580 '<FieldAccess for env_get_first_visible at ast.py:15510>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15510
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9578 '<If at ast.py:15512>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15512
--# expr-start 9577 '<FieldAccess for is_prefix at ast.py:15512>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15512
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_prefix
Fld := Libadalang.Implementation.Name_P_Is_Prefix (Node => Self);
--# end
--# expr-done 9577
if Fld then
   If_Result := Recursive;
else
   If_Result := Flat;
end if;
--# expr-done 9578
--# expr-start 9579 '<FieldAccess for origin_node at ast.py:15513>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15513
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_1 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 9579
--# property-call-start SingleTokNode.env_get_first_visible
Fld_2 := Libadalang.Implementation.Single_Tok_Node_P_Env_Get_First_Visible (Node => Self, Lex_Env => Env, Lookup_Type => If_Result, From_Node => Fld_1, No_Visibility => No_Visibility);
--# end
--# expr-done 9580
Block_Var_1 := Fld_2; 
--# expr-start 9599 '<If at ast.py:15516>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15516
--# expr-start 9581 '<Cast to BasicDecl at ast.py:15510>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15510
Cast_Expr_2 := Block_Var_1; 
   if Cast_Expr_2.Node = null
      or else Cast_Expr_2.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_3 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
   else
         Cast_Result_3 := No_Entity_Basic_Decl;
   end if;
--# expr-done 9581
Var_Expr := Cast_Result_3;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
--# expr-start 9582 '<FieldAccess for is_package at ast.py:15518>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15518
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.is_package
Fld_8 := Libadalang.Implementation.Ada_Node_P_Is_Package (Node => Var_Expr.Node);
--# end
--# expr-done 9582
Scope_Result_1 := Fld_8;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
if Result_Var_1 then
   --# expr-start 9586 '<Not at ast.py:15518>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15518
--# expr-start 9585 '<Eq at ast.py:15518>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15518
--# expr-start 9584 '<FieldAccess for node at ast.py:15518>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15518
--# expr-start 9583 '<Cast to BasicDecl at ast.py:15510>' Cast_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15510
Cast_Expr_3 := Block_Var_1; 
   if Cast_Expr_3.Node = null
      or else Cast_Expr_3.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_4 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_3.Node,
         Info => Cast_Expr_3.Info);
   else
         Cast_Result_4 := No_Entity_Basic_Decl;
   end if;
--# expr-done 9583
Fld_9 := Cast_Result_4.Node;
--# expr-done 9584
      Cast_Result_5 := Fld_9;
Is_Equal_1 := Cast_Result_5 = Block_Var; 
--# expr-done 9585
Not_Val_1 := not (Is_Equal_1); 
--# expr-done 9586
   If_Result_2 := Not_Val_1;
else
   If_Result_2 := False;
end if;
if If_Result_2 then
   --# expr-start 9588 '<FieldAccess for pkg_env at ast.py:15519>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15519
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9587 '<Cast to BasicDecl at ast.py:15510>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15510
Cast_Expr := Block_Var_1; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 9587
--# property-call-start BaseId.pkg_env
Fld_3 := Libadalang.Implementation.Base_Id_P_Pkg_Env (Node => Ent.Node, From_Pkg => Cast_Result, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 9588
   If_Result_3 := Fld_3;
else
   --# expr-start 9598 '<env_group at ast.py:15521>' Group_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15521
--# expr-start 9597 '<Map at ast.py:15521>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15521
--# expr-start 9589 '<FieldAccess for env_elements_baseid at ast.py:15521>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15521
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseId.env_elements_baseid
Fld_4 := Libadalang.Implementation.Base_Id_P_Env_Elements_Baseid (Node => Ent.Node, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 9589
Var_Expr_1 := Fld_4;
      Inc_Ref (Var_Expr_1);
if Var_Expr_1 /= No_Internal_Entity_Array_Type then
   --# scope-start
   --# bind all_env_els Var_Expr_1
--# expr-start 9594 '<Filter at ast.py:15523>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15523
      declare
         Map_Result_Vec : Internal_Entity_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Var_Expr_1;
      begin
         for Item_295 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_295
--# expr-start 9592 '<Not at ast.py:15525>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15525
--# expr-start 9591 '<Eq at ast.py:15525>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15525
--# expr-start 9590 '<FieldAccess for node at ast.py:15525>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15525
Fld_5 := Item_295.Node;
--# expr-done 9590
      Cast_Result_1 := Fld_5;
Is_Equal := Cast_Result_1 = Block_Var; 
--# expr-done 9591
Not_Val := not (Is_Equal); 
--# expr-done 9592
if Not_Val then
   --# expr-start 9593 '<FieldAccess for has_visibility at ast.py:15527>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15527
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.has_visibility
Fld_6 := Libadalang.Implementation.Ada_Node_P_Has_Visibility (Node => Self, Other_Entity => Item_295);
--# end
--# expr-done 9593
   If_Result_1 := Fld_6;
else
   If_Result_1 := False;
end if;
         if If_Result_1 then
         declare
            Item_To_Append : constant Internal_Entity := Item_295;
         begin
            Internal_Entity_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 9594
Scope_Result := Map_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2980;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Array_Type;
end if;
      Inc_Ref (Result_Var);
      declare
         Map_Result_1_Vec : Lexical_Env_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Result_Var;
      begin
         for Item_294 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_294
      --# expr-start 9596 '<FieldAccess for defining_env at ast.py:15529>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15529
--# expr-start 9595 '<Cast to BasicDecl at ast.py:15529>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15529
Cast_Expr_1 := Item_294; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_2 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_2 := No_Entity_Basic_Decl;
   end if;
--# expr-done 9595
      if Cast_Result_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_7 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Cast_Result_2.Node, Origin => Origin, Include_Ud_Indexing => False, E_Info => Cast_Result_2.Info);
--# end
--# expr-done 9596
         declare
            Item_To_Append : constant Lexical_Env := Fld_7;
         begin
               Inc_Ref (Item_To_Append);
            Lexical_Env_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2982;
         end loop;
      end;
         Map_Result_1 := Create_Lexical_Env_Array
           (Items_Count => Natural (Lexical_Env_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Lexical_Env_Vectors.Get
              (Map_Result_1_Vec,
               I + Lexical_Env_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Lexical_Env_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 9597
Group_Env := Group (Map_Result_1, No_Metadata); 
--# expr-done 9598
   If_Result_3 := Group_Env;
end if;
      Inc_Ref (If_Result_3);
--# expr-done 9599
Let_Result := If_Result_3; Inc_Ref (Let_Result);
--# end
--# expr-done 9600
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1115;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1115;
               Finalizer_Scope_2980;
               Finalizer_Scope_2982;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Id_P_Designated_Env;
--# end
--# property-start BaseId.pkg_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15533
pragma Warnings (Off, "is not referenced");
 function Base_Id_P_Pkg_Env
  (Node : Bare_Base_Id
      ; From_Pkg : Internal_Entity_Basic_Decl
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Base_Id := Bare_Base_Id (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Id :=
     Internal_Entity_Base_Id'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind from_pkg From_Pkg
   --# bind env Env
   --# bind origin Origin
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1116 with Inline_Always;
            procedure Finalizer_Scope_2972 with Inline_Always;
            procedure Finalizer_Scope_2973 with Inline_Always;
            procedure Finalizer_Scope_2974 with Inline_Always;
            procedure Finalizer_Scope_2975 with Inline_Always;
      Block_Var : Internal_Entity_Basic_Decl;
Block_Var_1 : Internal_Entity_Basic_Decl;
Block_Var_2 : Boolean;
Block_Var_3 : Lexical_Env := Empty_Env;
Block_Var_4 : Lexical_Env := Empty_Env;
Block_Var_5 : Lexical_Env := Empty_Env;
Block_Var_6 : Lexical_Env := Empty_Env;
Var_Expr : Internal_Entity_Package_Renaming_Decl;
Var_Expr_1 : Internal_Entity;
Var_Expr_2 : Internal_Entity;
Var_Expr_3 : Internal_Entity;
Var_Expr_4 : Internal_Entity_Generic_Package_Decl;
Var_Expr_5 : Bare_Ada_Node;
Var_Expr_6 : Bare_Ada_Node;
Cast_Expr : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity_Package_Renaming_Decl;
Fld : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Info;
Fld_2 : Env_Rebindings;
Fld_3 : Internal_Entity_Basic_Decl;
Is_A : Boolean;
Fld_4 : Internal_Entity_Info;
Fld_5 : Boolean;
If_Result : Boolean;
Cast_Expr_1 : Internal_Entity_Basic_Decl;
Cast_Result_1 : Internal_Entity_Generic_Package_Instantiation;
Fld_6 : Lexical_Env := Empty_Env;
Fld_7 : Lexical_Env := Empty_Env;
Is_A_1 : Boolean;
If_Result_1 : Boolean;
If_Result_2 : Lexical_Env := Empty_Env;
Sym : Symbol_Type;
Env_Get_Result : Internal_Entity_Array_Access;
Get_Result : Internal_Entity;
Fld_8 : Lexical_Env := Empty_Env;
Scope_Result_1 : Lexical_Env := Empty_Env;
Result_Var_1 : Lexical_Env := Empty_Env;
Sym_1 : Symbol_Type;
Env_Get_Result_1 : Internal_Entity_Array_Access;
Get_Result_1 : Internal_Entity;
Fld_9 : Lexical_Env := Empty_Env;
Sym_2 : Symbol_Type;
Env_Get_Result_2 : Internal_Entity_Array_Access;
Get_Result_2 : Internal_Entity;
Fld_10 : Lexical_Env := Empty_Env;
Scope_Result_2 : Lexical_Env := Empty_Env;
Result_Var_2 : Lexical_Env := Empty_Env;
Fld_11 : Lexical_Env := Empty_Env;
Is_A_2 : Boolean;
If_Result_3 : Lexical_Env := Empty_Env;
Scope_Result_3 : Lexical_Env := Empty_Env;
Result_Var_3 : Lexical_Env := Empty_Env;
Cast_Expr_2 : Internal_Entity_Basic_Decl;
Cast_Result_2 : Internal_Entity_Generic_Package_Decl;
Fld_12 : Internal_Entity_Generic_Formal_Part;
Fld_13 : Lexical_Env := Empty_Env;
Scope_Result_4 : Lexical_Env := Empty_Env;
Result_Var_4 : Lexical_Env := Empty_Env;
Is_Equal : Boolean;
Not_Val : Boolean;
Scope_Result_5 : Bare_Ada_Node;
Cast_Expr_3 : Bare_Base_Id;
Cast_Result_3 : Bare_Ada_Node;
Result_Var_5 : Bare_Ada_Node;
Fld_14 : Lexical_Env := Empty_Env;
Fld_15 : Boolean;
If_Result_4 : Boolean;
Array_Lit : Lexical_Env_Array_Access;
Group_Env : Lexical_Env := Empty_Env;
Scope_Result_6 : Bare_Ada_Node;
Cast_Expr_4 : Bare_Base_Id;
Cast_Result_4 : Bare_Ada_Node;
Result_Var_6 : Bare_Ada_Node;
Fld_16 : Lexical_Env := Empty_Env;
Fld_17 : Boolean;
Array_Lit_1 : Lexical_Env_Array_Access;
Group_Env_1 : Lexical_Env := Empty_Env;
If_Result_5 : Lexical_Env := Empty_Env;
If_Result_6 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1116 is
            begin
                     Dec_Ref (Fld_6);
                     Dec_Ref (Fld_7);
                     Dec_Ref (If_Result_2);
                     Dec_Ref (Block_Var_3);
                     Dec_Ref (Env_Get_Result);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Block_Var_4);
                     Dec_Ref (Env_Get_Result_1);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Result_Var_3);
                     Dec_Ref (Block_Var_5);
                     Dec_Ref (Scope_Result_4);
                     Dec_Ref (Result_Var_4);
                     Dec_Ref (Block_Var_6);
                     Dec_Ref (Fld_14);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Group_Env);
                     Dec_Ref (Fld_16);
                     Dec_Ref (Array_Lit_1);
                     Dec_Ref (Group_Env_1);
                     Dec_Ref (If_Result_5);
                     Dec_Ref (If_Result_6);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1116;
            procedure Finalizer_Scope_2972 is
            begin
                     Dec_Ref (Fld_8);
            end Finalizer_Scope_2972;
            procedure Finalizer_Scope_2973 is
            begin
                     Dec_Ref (Fld_9);
                     Dec_Ref (Env_Get_Result_2);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Result_Var_2);
                     Dec_Ref (Fld_11);
                     Dec_Ref (If_Result_3);
            end Finalizer_Scope_2973;
            procedure Finalizer_Scope_2974 is
            begin
                     Dec_Ref (Fld_10);
            end Finalizer_Scope_2974;
            procedure Finalizer_Scope_2975 is
            begin
                     Dec_Ref (Fld_13);
            end Finalizer_Scope_2975;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9643 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9601 '<Cast to PackageRenamingDecl at ast.py:15548>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15548
Cast_Expr := From_Pkg; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Package_Renaming_Decl_Range
   then
      Cast_Result := Create_Internal_Entity_Package_Renaming_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Package_Renaming_Decl;
   end if;
--# expr-done 9601
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Package_Renaming_Decl then
   --# scope-start
   --# bind r Var_Expr
--# expr-start 9602 '<FieldAccess for final_renamed_package at ast.py:15549>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15549
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start PackageRenamingDecl.final_renamed_package
Fld := Libadalang.Implementation.Package_Renaming_Decl_P_Final_Renamed_Package (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 9602
Scope_Result := Fld;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := From_Pkg;
end if;
Block_Var := Result_Var; 
--# bind pkg Block_Var
--# expr-start 9605 '<FieldAccess for unshed_rebindings at ast.py:15560>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15560
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9604 '<FieldAccess for rebindings at ast.py:15560>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15560
--# expr-start 9603 '<FieldAccess for info at ast.py:15560>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15560
Fld_1 := Ent.Info;
--# expr-done 9603
Fld_2 := Fld_1.Rebindings;
--# expr-done 9604
--# property-call-start BasicDecl.unshed_rebindings
Fld_3 := Libadalang.Implementation.Basic_Decl_P_Unshed_Rebindings (Node => Block_Var.Node, Rebindings => Fld_2, E_Info => Block_Var.Info);
--# end
--# expr-done 9605
Block_Var_1 := Fld_3; 
--# bind bd Block_Var_1
--# expr-start 9606 '<IsA GenericPackageInstantiation at ast.py:15562>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15562
Is_A := Block_Var.Node /= null 
and then Block_Var.Node.Kind in Ada_Generic_Package_Instantiation_Range; 
--# expr-done 9606
if Is_A then
   --# expr-start 9608 '<FieldAccess for from_rebound at ast.py:15563>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15563
--# expr-start 9607 '<FieldAccess for info at ast.py:15563>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15563
Fld_4 := From_Pkg.Info;
--# expr-done 9607
Fld_5 := Fld_4.From_Rebound;
--# expr-done 9608
   If_Result := Fld_5;
else
   If_Result := False;
end if;
Block_Var_2 := If_Result; 
--# bind is_inst_from_formal Block_Var_2
--# expr-start 9613 '<If at ast.py:15565>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15565
--# expr-start 9609 '<IsA GenericPackageInstantiation at ast.py:15566>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15566
Is_A_1 := Block_Var_1.Node /= null 
and then Block_Var_1.Node.Kind in Ada_Generic_Package_Instantiation_Range; 
--# expr-done 9609
if Is_A_1 then
   If_Result_1 := Block_Var_2;
else
   If_Result_1 := False;
end if;
if If_Result_1 then
   --# expr-start 9611 '<FieldAccess for defining_env_impl at ast.py:15567>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15567
--# expr-start 9610 '<Cast to GenericPackageInstantiation at ast.py:15567>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15567
Cast_Expr_1 := Block_Var_1; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Generic_Package_Instantiation_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Generic_Package_Instantiation
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_1 := No_Entity_Generic_Package_Instantiation;
   end if;
--# expr-done 9610
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start GenericPackageInstantiation.defining_env_impl
Fld_6 := Libadalang.Implementation.Generic_Package_Instantiation_P_Defining_Env_Impl (Node => Cast_Result_1.Node, Inst_From_Formal => True, Origin => Origin, E_Info => Cast_Result_1.Info);
--# end
--# expr-done 9611
   If_Result_2 := Fld_6;
else
   --# expr-start 9612 '<FieldAccess for defining_env at ast.py:15568>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15568
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_7 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Block_Var_1.Node, Origin => Origin, Include_Ud_Indexing => False, E_Info => Block_Var_1.Info);
--# end
--# expr-done 9612
   If_Result_2 := Fld_7;
end if;
      Inc_Ref (If_Result_2);
--# expr-done 9613
Block_Var_3 := If_Result_2; Inc_Ref (Block_Var_3);
--# bind env Block_Var_3
--# expr-start 9616 '<at at ast.py:15574>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15574
--# expr-start 9615 '<EnvGet at ast.py:15574>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15574
--# expr-start 9614 '<Symbol __privatepart at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Privatepart); 
--# expr-done 9614
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Block_Var_3, Key => Sym, Lookup_Kind => To_Lookup_Kind_Type (Flat), Categories => (Inherited_Primitives => False, Nocat => True))); 
--# expr-done 9615
Get_Result := Get (Env_Get_Result, 0, True); 
--# expr-done 9616
Var_Expr_1 := Get_Result;
if Var_Expr_1 /= No_Entity then
   --# scope-start
   --# bind pp Var_Expr_1
--# expr-start 9617 '<FieldAccess for children_env at ast.py:15575>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15575
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Children_Env (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# expr-done 9617
Scope_Result_1 := Fld_8;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2972;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := Block_Var_3;
end if;
      Inc_Ref (Result_Var_1);
Block_Var_4 := Result_Var_1; Inc_Ref (Block_Var_4);
--# bind private_part_env Block_Var_4
--# expr-start 9620 '<at at ast.py:15580>' Get_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15580
--# expr-start 9619 '<EnvGet at ast.py:15580>' Env_Get_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15580
--# expr-start 9618 '<Symbol __nextpart at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Nextpart); 
--# expr-done 9618
Env_Get_Result_1 := Construct_Entity_Array (AST_Envs.Get (Self => Block_Var_4, Key => Sym_1, Lookup_Kind => To_Lookup_Kind_Type (Flat), Categories => (Inherited_Primitives => False, Nocat => True))); 
--# expr-done 9619
Get_Result_1 := Get (Env_Get_Result_1, 0, True); 
--# expr-done 9620
Var_Expr_2 := Get_Result_1;
if Var_Expr_2 /= No_Entity then
   --# scope-start
   --# bind pb Var_Expr_2
--# expr-start 9628 '<If at ast.py:15582>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15582
--# expr-start 9621 '<IsA PackageBodyStub at ast.py:15585>' Is_A_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15585
Is_A_2 := Var_Expr_2.Node /= null 
and then Var_Expr_2.Node.Kind in Ada_Package_Body_Stub_Range; 
--# expr-done 9621
if Is_A_2 then
--# expr-start 9625 '<at at ast.py:15587>' Get_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15587
--# expr-start 9624 '<EnvGet at ast.py:15587>' Env_Get_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15587
--# expr-start 9622 '<FieldAccess for children_env at ast.py:15587>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15587
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Libadalang.Implementation.Children_Env (Node => Var_Expr_2.Node, E_Info => Var_Expr_2.Info);
--# expr-done 9622
--# expr-start 9623 '<Symbol __nextpart at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Nextpart); 
--# expr-done 9623
Env_Get_Result_2 := Construct_Entity_Array (AST_Envs.Get (Self => Fld_9, Key => Sym_2, Lookup_Kind => To_Lookup_Kind_Type (Flat), Categories => (Inherited_Primitives => False, Nocat => True))); 
--# expr-done 9624
Get_Result_2 := Get (Env_Get_Result_2, 0, True); 
--# expr-done 9625
Var_Expr_3 := Get_Result_2;
if Var_Expr_3 /= No_Entity then
   --# scope-start
   --# bind pb Var_Expr_3
--# expr-start 9626 '<FieldAccess for children_env at ast.py:15589>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15589
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Libadalang.Implementation.Children_Env (Node => Var_Expr_3.Node, E_Info => Var_Expr_3.Info);
--# expr-done 9626
Scope_Result_2 := Fld_10;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2974;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := Empty_Env;
end if;
      Inc_Ref (Result_Var_2);
   If_Result_3 := Result_Var_2;
else
   --# expr-start 9627 '<FieldAccess for children_env at ast.py:15591>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15591
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Libadalang.Implementation.Children_Env (Node => Var_Expr_2.Node, E_Info => Var_Expr_2.Info);
--# expr-done 9627
   If_Result_3 := Fld_11;
end if;
      Inc_Ref (If_Result_3);
--# expr-done 9628
Scope_Result_3 := If_Result_3;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_2973;
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := Empty_Env;
end if;
      Inc_Ref (Result_Var_3);
Block_Var_5 := Result_Var_3; Inc_Ref (Block_Var_5);
--# bind package_body_env Block_Var_5
--# expr-start 9629 '<Cast to GenericPackageDecl at ast.py:15596>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15596
Cast_Expr_2 := Block_Var_1; 
   if Cast_Expr_2.Node = null
      or else Cast_Expr_2.Node.Kind in Ada_Generic_Package_Decl_Range
   then
      Cast_Result_2 := Create_Internal_Entity_Generic_Package_Decl
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
   else
         Cast_Result_2 := No_Entity_Generic_Package_Decl;
   end if;
--# expr-done 9629
Var_Expr_4 := Cast_Result_2;
if Var_Expr_4 /= No_Entity_Generic_Package_Decl then
   --# scope-start
   --# bind pkg_g Var_Expr_4
--# expr-start 9631 '<FieldAccess for children_env at ast.py:15597>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15597
--# expr-start 9630 '<FieldAccess for formal_part at ast.py:15597>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15597
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Create_Internal_Entity_Generic_Formal_Part (Node => Var_Expr_4.Node.Generic_Decl_F_Formal_Part, Info => Var_Expr_4.Info);
--# expr-done 9630
      if Fld_12.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Libadalang.Implementation.Children_Env (Node => Fld_12.Node, E_Info => Fld_12.Info);
--# expr-done 9631
Scope_Result_4 := Fld_13;
   Inc_Ref (Scope_Result_4);
   --# end
      Finalizer_Scope_2975;
   Result_Var_4 := Scope_Result_4;
else
   Result_Var_4 := Empty_Env;
end if;
      Inc_Ref (Result_Var_4);
Block_Var_6 := Result_Var_4; Inc_Ref (Block_Var_6);
--# bind formals_env Block_Var_6
--# expr-start 9642 '<Cond at ast.py:15601>' If_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15601
--# expr-start 9633 '<Not at ast.py:15605>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15605
--# expr-start 9632 '<Eq at ast.py:15605>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15605
Is_Equal := Equivalent (Block_Var_5, Empty_Env); 
--# expr-done 9632
Not_Val := not (Is_Equal); 
--# expr-done 9633
if Not_Val then
   --# expr-start 9635 '<FieldAccess for is_children_env at ast.py:15606>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15606
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9634 '<FieldAccess for node_env at ast.py:15607>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15607
Var_Expr_5 := Origin;
if Var_Expr_5 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind e Var_Expr_5
Scope_Result_5 := Var_Expr_5;
   --# end
   Result_Var_5 := Scope_Result_5;
else
Cast_Expr_3 := Self; 
      Cast_Result_3 := Cast_Expr_3;
   Result_Var_5 := Cast_Result_3;
end if;
      if Result_Var_5 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Libadalang.Implementation.Node_Env (Node => Result_Var_5);
--# expr-done 9634
--# property-call-start AdaNode.is_children_env
Fld_15 := Libadalang.Implementation.Ada_Node_P_Is_Children_Env (Node => Self, Parent => Block_Var_5, Current_Env => Fld_14);
--# end
--# expr-done 9635
   If_Result_4 := Fld_15;
else
   If_Result_4 := False;
end if;
if If_Result_4 then
   --# expr-start 9637 '<env_group at ast.py:15608>' Group_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15608
--# expr-start 9636 '<ArrayLiteral at ast.py:15608>' Array_Lit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15608
Array_Lit := Create_Lexical_Env_Array (Internal_Lexical_Env_Array'(1 => Block_Var_5, 2 => Block_Var_4, 3 => Block_Var_3, 4 => Block_Var_6)); 
--# expr-done 9636
Group_Env := Group (Array_Lit, No_Metadata); 
--# expr-done 9637
   If_Result_6 := Group_Env;
else
--# expr-start 9639 '<FieldAccess for is_children_env at ast.py:15614>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15614
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9638 '<FieldAccess for node_env at ast.py:15615>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15615
Var_Expr_6 := Origin;
if Var_Expr_6 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind e Var_Expr_6
Scope_Result_6 := Var_Expr_6;
   --# end
   Result_Var_6 := Scope_Result_6;
else
Cast_Expr_4 := Self; 
      Cast_Result_4 := Cast_Expr_4;
   Result_Var_6 := Cast_Result_4;
end if;
      if Result_Var_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_16 := Libadalang.Implementation.Node_Env (Node => Result_Var_6);
--# expr-done 9638
--# property-call-start AdaNode.is_children_env
Fld_17 := Libadalang.Implementation.Ada_Node_P_Is_Children_Env (Node => Self, Parent => Block_Var_4, Current_Env => Fld_16);
--# end
--# expr-done 9639
if Fld_17 then
   --# expr-start 9641 '<env_group at ast.py:15616>' Group_Env_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15616
--# expr-start 9640 '<ArrayLiteral at ast.py:15616>' Array_Lit_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15616
Array_Lit_1 := Create_Lexical_Env_Array (Internal_Lexical_Env_Array'(1 => Block_Var_4, 2 => Block_Var_3, 3 => Block_Var_6)); 
--# expr-done 9640
Group_Env_1 := Group (Array_Lit_1, No_Metadata); 
--# expr-done 9641
   If_Result_5 := Group_Env_1;
else
   If_Result_5 := Block_Var_3;
end if;
      Inc_Ref (If_Result_5);
   If_Result_6 := If_Result_5;
end if;
      Inc_Ref (If_Result_6);
--# expr-done 9642
Let_Result := If_Result_6; Inc_Ref (Let_Result);
--# end
--# expr-done 9643
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1116;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1116;
               Finalizer_Scope_2972;
               Finalizer_Scope_2973;
               Finalizer_Scope_2974;
               Finalizer_Scope_2975;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Id_P_Pkg_Env;
--# end
--# property-start BaseId.parent_scope /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15625
pragma Warnings (Off, "is not referenced");
 function Base_Id_P_Parent_Scope
  (Node : Bare_Base_Id
      ; Env : Lexical_Env
  )
   return Lexical_Env
is
   Self : Bare_Base_Id := Bare_Base_Id (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   Property_Result : Lexical_Env;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := Env;
         Inc_Ref (Property_Result);
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Id_P_Parent_Scope;
--# end
--# property-start BaseId.designated_type_impl_get_real_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15628
pragma Warnings (Off, "is not referenced");
 function Base_Id_P_Designated_Type_Impl_Get_Real_Type
  (Node : Bare_Base_Id
      ; N : Internal_Entity
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Id := Bare_Base_Id (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind n N
   Property_Result : Internal_Entity_Base_Type_Decl;
      Match_0 : Internal_Entity_Base_Type_Decl;
Match_1 : Internal_Entity_Task_Body;
Match_2 : Internal_Entity;
Fld : Internal_Entity_Task_Type_Decl;
Match_Prefix : Internal_Entity;
Cast_Result : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity_Task_Body;
Cast_Result_2 : Internal_Entity_Base_Type_Decl;
Let_Result_1 : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity;
Let_Result_2 : Internal_Entity_Base_Type_Decl;
Scope_Result_2 : Internal_Entity_Base_Type_Decl;
Match_Result : Internal_Entity_Base_Type_Decl;
Let_Result_3 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9646 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 9645 '<Match at ast.py:15633>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15633
      if N.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := N; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Discrete_Base_Subtype_Decl .. Ada_Synth_Anonymous_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind t Match_0
Let_Result := Match_0; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Task_Body =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Task_Body
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind tb Match_1
--# expr-start 9644 '<FieldAccess for task_type at ast.py:15635>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15635
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TaskBody.task_type
Fld := Libadalang.Implementation.Task_Body_P_Task_Type (Node => Match_1.Node, E_Info => Match_1.Info);
--# end
--# expr-done 9644
      Cast_Result_2 := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
Let_Result_1 := Cast_Result_2; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abort_Absent .. Ada_Package_Decl | Ada_Abstract_Subp_Decl .. Ada_Protected_Body | Ada_Entry_Index_Spec .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_3; 
--# bind _ Match_2
Let_Result_2 := No_Entity_Base_Type_Decl; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 9645
Let_Result_3 := Match_Result; 
--# end
--# expr-done 9646
      Property_Result := Let_Result_3;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Id_P_Designated_Type_Impl_Get_Real_Type;
--# end
--# property-start BaseId.designated_type_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15640
pragma Warnings (Off, "is not referenced");
 function Base_Id_P_Designated_Type_Impl
  (Node : Bare_Base_Id
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Base_Id := Bare_Base_Id (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1119 with Inline_Always;
            procedure Finalizer_Scope_2990 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl;
Block_Var_1 : Internal_Entity_Base_Type_Decl;
Block_Var_2 : Internal_Entity_Base_Type_Decl;
Block_Var_3 : Internal_Entity_Base_Type_Decl;
Var_Expr : Internal_Entity;
Var_Expr_1 : Internal_Entity;
Var_Expr_2 : Internal_Entity_Base_Type_Decl;
Var_Expr_3 : Bare_Ada_Node;
Var_Expr_4 : Internal_Entity_Base_Type_Decl;
Fld : Boolean;
If_Result : Lookup_Kind;
Cast_Expr : Bare_Base_Id;
Cast_Result : Bare_Ada_Node;
Fld_1 : Internal_Entity;
Fld_2 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Fld_3 : Boolean;
If_Result_1 : Lookup_Kind;
Fld_4 : Internal_Entity;
Fld_5 : Internal_Entity_Base_Type_Decl;
Scope_Result_1 : Internal_Entity_Base_Type_Decl;
Result_Var_1 : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Cast_Expr_2 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Is_Equiv : Boolean;
Fld_6 : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
If_Result_2 : Internal_Entity_Base_Type_Decl;
If_Result_3 : Internal_Entity_Base_Type_Decl;
If_Result_4 : Internal_Entity_Base_Type_Decl;
Fld_7 : Lexical_Env := Empty_Env;
Sym : Symbol_Type;
Fld_8 : Internal_Entity_Array_Access;
Scope_Result_3 : Internal_Entity_Array_Access;
Result_Var_3 : Internal_Entity_Array_Access;
Cast_Expr_3 : Internal_Entity;
Cast_Result_3 : Internal_Entity_Base_Type_Decl;
Is_A : Boolean;
Cast_Expr_4 : Internal_Entity;
Cast_Result_4 : Internal_Entity_Base_Type_Decl;
Fld_9 : Boolean;
Scope_Result_4 : Boolean;
Result_Var_4 : Boolean;
If_Result_5 : Boolean;
Map_Result : Internal_Entity_Base_Type_Decl_Array_Access;
Get_Result : Internal_Entity_Base_Type_Decl;
Is_Null_1 : Boolean;
Not_Val : Boolean;
If_Result_6 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1119 is
            begin
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Result_Var_3);
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_1119;
            procedure Finalizer_Scope_2990 is
            begin
                     Dec_Ref (Fld_7);
                     Dec_Ref (Fld_8);
            end Finalizer_Scope_2990;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9671 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9649 '<FieldAccess for env_get_first_visible at ast.py:15642>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15642
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9648 '<If at ast.py:15645>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15645
--# expr-start 9647 '<FieldAccess for is_prefix at ast.py:15645>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15645
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_prefix
Fld := Libadalang.Implementation.Name_P_Is_Prefix (Node => Self);
--# end
--# expr-done 9647
if Fld then
   If_Result := Recursive;
else
   If_Result := Minimal;
end if;
--# expr-done 9648
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
--# property-call-start SingleTokNode.env_get_first_visible
Fld_1 := Libadalang.Implementation.Single_Tok_Node_P_Env_Get_First_Visible (Node => Self, Lex_Env => Env, Lookup_Type => If_Result, From_Node => Cast_Result, No_Visibility => False);
--# end
--# expr-done 9649
Var_Expr := Fld_1;
if Var_Expr /= No_Entity then
   --# scope-start
   --# bind env_el Var_Expr
--# expr-start 9650 '<FieldAccess for designated_type_impl_get_real_type at ast.py:15647>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15647
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseId.designated_type_impl_get_real_type
Fld_2 := Libadalang.Implementation.Base_Id_P_Designated_Type_Impl_Get_Real_Type (Node => Self, N => Var_Expr);
--# end
--# expr-done 9650
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Block_Var := Result_Var; 
--# bind des_type_1 Block_Var
--# expr-start 9653 '<FieldAccess for env_get_first_visible at ast.py:15651>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15651
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9652 '<If at ast.py:15654>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15654
--# expr-start 9651 '<FieldAccess for is_prefix at ast.py:15654>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15654
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_prefix
Fld_3 := Libadalang.Implementation.Name_P_Is_Prefix (Node => Self);
--# end
--# expr-done 9651
if Fld_3 then
   If_Result_1 := Recursive;
else
   If_Result_1 := Minimal;
end if;
--# expr-done 9652
--# property-call-start SingleTokNode.env_get_first_visible
Fld_4 := Libadalang.Implementation.Single_Tok_Node_P_Env_Get_First_Visible (Node => Self, Lex_Env => Env, Lookup_Type => If_Result_1, From_Node => Origin, No_Visibility => False);
--# end
--# expr-done 9653
Var_Expr_1 := Fld_4;
if Var_Expr_1 /= No_Entity then
   --# scope-start
   --# bind env_el Var_Expr_1
--# expr-start 9654 '<FieldAccess for designated_type_impl_get_real_type at ast.py:15656>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15656
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseId.designated_type_impl_get_real_type
Fld_5 := Libadalang.Implementation.Base_Id_P_Designated_Type_Impl_Get_Real_Type (Node => Self, N => Var_Expr_1);
--# end
--# expr-done 9654
Scope_Result_1 := Fld_5;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity_Base_Type_Decl;
end if;
Block_Var_1 := Result_Var_1; 
--# bind des_type_2 Block_Var_1
--# expr-start 9658 '<Cond at ast.py:15659>' If_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15659
--# expr-start 9655 '<IsNull at ast.py:15661>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15661
Is_Null := Block_Var.Node = null; 
--# expr-done 9655
if Is_Null then
   If_Result_4 := Block_Var_1;
else
--# expr-start 9656 '<Eq at ast.py:15665>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15665
Cast_Expr_1 := Block_Var; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Cast_Expr_2 := Block_Var_1; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
Is_Equiv := Equivalent (Cast_Result_1, Cast_Result_2); 
--# expr-done 9656
if Is_Equiv then
   If_Result_3 := Block_Var;
else
Var_Expr_2 := Block_Var;
if Var_Expr_2 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind d Var_Expr_2
--# expr-start 9657 '<FieldAccess for is_view_of_type at ast.py:15669>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15669
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_view_of_type
Fld_6 := Libadalang.Implementation.Base_Type_Decl_P_Is_View_Of_Type (Node => Var_Expr_2.Node, Comp_View => Block_Var_1, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 9657
Scope_Result_2 := Fld_6;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := False;
end if;
if Result_Var_2 then
   If_Result_2 := Block_Var_1;
else
   If_Result_2 := Block_Var;
end if;
   If_Result_3 := If_Result_2;
end if;
   If_Result_4 := If_Result_3;
end if;
--# expr-done 9658
Block_Var_2 := If_Result_4; 
--# bind des_type Block_Var_2
--# expr-start 9667 '<at at ast.py:15679>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15679
--# expr-start 9666 '<FilterMap at ast.py:15679>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15679
Var_Expr_3 := Origin;
if Var_Expr_3 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind o Var_Expr_3
--# expr-start 9661 '<FieldAccess for env_get at ast.py:15679>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15679
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9659 '<FieldAccess for children_env at ast.py:15680>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15680
      if Var_Expr_3 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Children_Env (Node => Var_Expr_3);
--# expr-done 9659
--# expr-start 9660 '<GetSymbol at ast.py:15680>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15680
Sym := Get_Symbol (Self); 
--# expr-done 9660
--# property-call-start AdaNode.env_get
Fld_8 := Libadalang.Implementation.Ada_Node_P_Env_Get (Node => Self, Env => Fld_7, Symbol => Sym, From_Node => Origin, Categories => (Inherited_Primitives => False, Nocat => True));
--# end
--# expr-done 9661
Scope_Result_3 := Fld_8;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_2990;
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := No_Internal_Entity_Array_Type;
end if;
      Inc_Ref (Result_Var_3);
      declare
         Map_Result_Vec : Internal_Entity_Base_Type_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Result_Var_3;
      begin
         for Item_296 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_296
--# expr-start 9662 '<IsA BaseTypeDecl at ast.py:15684>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15684
Is_A := Item_296.Node /= null 
and then Item_296.Node.Kind in Ada_Base_Type_Decl; 
--# expr-done 9662
if Is_A then
Var_Expr_4 := Block_Var_2;
if Var_Expr_4 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind d Var_Expr_4
--# expr-start 9664 '<FieldAccess for is_view_of_type at ast.py:15685>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15685
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9663 '<Cast to BaseTypeDecl at ast.py:15685>' Cast_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15685
Cast_Expr_4 := Item_296; 
   if Cast_Expr_4.Node = null
      or else Cast_Expr_4.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_4 := Create_Internal_Entity_Base_Type_Decl
        (Node => Cast_Expr_4.Node,
         Info => Cast_Expr_4.Info);
   else
         Cast_Result_4 := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 9663
--# property-call-start BaseTypeDecl.is_view_of_type
Fld_9 := Libadalang.Implementation.Base_Type_Decl_P_Is_View_Of_Type (Node => Var_Expr_4.Node, Comp_View => Cast_Result_4, E_Info => Var_Expr_4.Info);
--# end
--# expr-done 9664
Scope_Result_4 := Fld_9;
   --# end
   Result_Var_4 := Scope_Result_4;
else
   Result_Var_4 := False;
end if;
   If_Result_5 := Result_Var_4;
else
   If_Result_5 := False;
end if;
         if If_Result_5 then
      --# expr-start 9665 '<Cast to BaseTypeDecl at ast.py:15682>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15682
Cast_Expr_3 := Item_296; 
   if Cast_Expr_3.Node = null
      or else Cast_Expr_3.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_3 := Create_Internal_Entity_Base_Type_Decl
        (Node => Cast_Expr_3.Node,
         Info => Cast_Expr_3.Info);
   else
         Cast_Result_3 := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 9665
         declare
            Item_To_Append : constant Internal_Entity_Base_Type_Decl := Cast_Result_3;
         begin
            Internal_Entity_Base_Type_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Base_Type_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Type_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Base_Type_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Base_Type_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Base_Type_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 9666
Get_Result := Get (Map_Result, 0, True); 
--# expr-done 9667
Block_Var_3 := Get_Result; 
--# bind completer_view Block_Var_3
--# expr-start 9670 '<If at ast.py:15690>' If_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15690
--# expr-start 9669 '<Not at ast.py:15691>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15691
--# expr-start 9668 '<IsNull at ast.py:15691>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15691
Is_Null_1 := Block_Var_3.Node = null; 
--# expr-done 9668
Not_Val := not (Is_Null_1); 
--# expr-done 9669
if Not_Val then
   If_Result_6 := Block_Var_3;
else
   If_Result_6 := Block_Var_2;
end if;
--# expr-done 9670
Let_Result := If_Result_6; 
--# end
--# expr-done 9671
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1119;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1119;
               Finalizer_Scope_2990;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Id_P_Designated_Type_Impl;
--# end
--# property-start BaseId.env_elements_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15697
pragma Warnings (Off, "is not referenced");
 function Base_Id_P_Env_Elements_Impl
  (Node : Bare_Base_Id
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_Base_Id := Bare_Base_Id (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Id :=
     Internal_Entity_Base_Id'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   Property_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1120 with Inline_Always;
      Fld : Internal_Entity_Array_Access;
Let_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1120 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1120;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9673 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9672 '<FieldAccess for env_elements_baseid at ast.py:15698>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15698
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseId.env_elements_baseid
Fld := Libadalang.Implementation.Base_Id_P_Env_Elements_Baseid (Node => Ent.Node, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 9672
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 9673
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1120;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1120;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Id_P_Env_Elements_Impl;
--# end
--# property-start BaseId.all_env_els_impl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15701
pragma Warnings (Off, "is not referenced");
 function Base_Id_P_All_Env_Els_Impl
  (Node : Bare_Base_Id
      ; Seq : Boolean
         := True
      ; Seq_From : Bare_Ada_Node
         := No_Bare_Ada_Node
      ; Categories : Ref_Categories
         := (Inherited_Primitives => True, Nocat => True)
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_Base_Id := Bare_Base_Id (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind seq Seq
   --# bind seq_from Seq_From
   --# bind categories Categories
   --# bind env Env
   --# bind origin Origin
   Property_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1121 with Inline_Always;
      Fld : Symbol_Type;
Fld_1 : Boolean;
If_Result : Lookup_Kind;
Cast_Expr : Bare_Base_Id;
Cast_Result : Bare_Ada_Node;
Is_Null : Boolean;
Not_Val : Boolean;
If_Result_1 : Bare_Ada_Node;
If_Result_2 : Bare_Ada_Node;
Fld_2 : Internal_Entity_Array_Access;
Let_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1121 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1121;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9682 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9681 '<FieldAccess for env_get at ast.py:15706>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15706
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9674 '<FieldAccess for name_symbol at ast.py:15708>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15708
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Self);
--# end
--# expr-done 9674
--# expr-start 9676 '<If at ast.py:15709>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15709
--# expr-start 9675 '<FieldAccess for is_prefix at ast.py:15709>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15709
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_prefix
Fld_1 := Libadalang.Implementation.Name_P_Is_Prefix (Node => Self);
--# end
--# expr-done 9675
if Fld_1 then
   If_Result := Recursive;
else
   If_Result := Flat;
end if;
--# expr-done 9676
--# expr-start 9680 '<If at ast.py:15710>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15710
if Seq then
   --# expr-start 9679 '<If at ast.py:15710>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15710
--# expr-start 9678 '<Not at ast.py:15710>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15710
--# expr-start 9677 '<IsNull at ast.py:15710>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15710
Is_Null := Seq_From = null; 
--# expr-done 9677
Not_Val := not (Is_Null); 
--# expr-done 9678
if Not_Val then
   If_Result_1 := Seq_From;
else
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
   If_Result_1 := Cast_Result;
end if;
--# expr-done 9679
   If_Result_2 := If_Result_1;
else
   If_Result_2 := No_Bare_Ada_Node;
end if;
--# expr-done 9680
--# property-call-start AdaNode.env_get
Fld_2 := Libadalang.Implementation.Ada_Node_P_Env_Get (Node => Self, Env => Env, Symbol => Fld, Lookup => If_Result, From_Node => If_Result_2, Categories => Categories);
--# end
--# expr-done 9681
Let_Result := Fld_2; Inc_Ref (Let_Result);
--# end
--# expr-done 9682
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1121;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1121;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Id_P_All_Env_Els_Impl;
--# end
--# property-start BaseId.env_elements_baseid /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15716
pragma Warnings (Off, "is not referenced");
 function Base_Id_P_Env_Elements_Baseid
  (Node : Bare_Base_Id
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Array_Access
is
   Self : Bare_Base_Id := Bare_Base_Id (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Id :=
     Internal_Entity_Base_Id'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   Property_Result : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1122 with Inline_Always;
            procedure Finalizer_Scope_2938 with Inline_Always;
      Block_Var : Internal_Entity_Array_Access;
Block_Var_1 : Internal_Entity_Call_Expr;
Block_Var_2 : Boolean;
Var_Expr : Internal_Entity_Base_Subp_Body;
Var_Expr_1 : Internal_Entity_Assoc_List;
Match_0 : Internal_Entity_Base_Type_Decl;
Match_1 : Internal_Entity_Basic_Decl;
Match_2 : Internal_Entity;
Var_Expr_2 : Internal_Entity_Base_Subp_Spec;
Real_Pc : Internal_Entity_Call_Expr;
Var_Expr_3 : Internal_Entity_Entry_Spec;
Var_Expr_4 : Internal_Entity_Call_Expr;
Sym : Symbol_Type;
Fld : Boolean;
If_Result : Lookup_Kind;
Fld_1 : Bare_Ada_Node;
Fld_2 : Boolean;
If_Result_1 : Ref_Categories;
Fld_3 : Internal_Entity_Array_Access;
Fld_4 : Internal_Entity_Call_Expr;
Fld_5 : Boolean;
Not_Val : Boolean;
Fld_6 : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Is_Null : Boolean;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Fld_7 : Boolean;
Cast_Expr_1 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Base_Subp_Body;
Fld_8 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Boolean;
Map_Result : Internal_Entity_Array_Access;
Fld_9 : Internal_Entity;
Cast_Result_2 : Internal_Entity_Assoc_List;
Fld_10 : Bare_Assoc_List;
Len : Integer;
Is_Equal : Boolean;
Fld_11 : Internal_Entity_Base_Subp_Spec;
Fld_12 : Internal_Entity;
Cast_Result_3 : Internal_Entity_Call_Expr;
Cast_Expr_2 : Internal_Entity_Base_Subp_Spec;
Cast_Result_4 : Internal_Entity_Entry_Spec;
Fld_13 : Internal_Entity;
Scope_Result_1 : Internal_Entity;
Result_Var_1 : Internal_Entity;
Is_Null_1 : Boolean;
If_Result_4 : Internal_Entity_Call_Expr;
Is_Null_2 : Boolean;
Fld_14 : Internal_Entity_Info;
Fld_15 : Internal_Metadata;
Fld_16 : Boolean;
Fld_17 : Boolean;
Fld_18 : Internal_Entity;
Cast_Result_5 : Internal_Entity_Call_Expr;
Fld_19 : Internal_Entity_Base_Type_Decl;
Fld_20 : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
If_Result_5 : Boolean;
If_Result_6 : Boolean;
Fld_21 : Internal_Entity_Base_Type_Decl;
Fld_22 : Boolean;
Fld_23 : Internal_Entity_Info;
Fld_24 : Internal_Metadata;
Fld_25 : Boolean;
Fld_26 : Boolean;
If_Result_7 : Boolean;
If_Result_8 : Boolean;
Let_Result : Boolean;
Scope_Result_3 : Boolean;
Fld_27 : Internal_Entity_Base_Type_Decl;
Fld_28 : Boolean;
Result_Var_3 : Boolean;
Match_Prefix : Internal_Entity;
Cast_Result_6 : Internal_Entity_Base_Type_Decl;
Let_Result_1 : Boolean;
Scope_Result_4 : Boolean;
Cast_Result_7 : Internal_Entity_Basic_Decl;
Let_Result_2 : Boolean;
Scope_Result_5 : Boolean;
Cast_Result_8 : Internal_Entity;
Let_Result_3 : Boolean;
Scope_Result_6 : Boolean;
Match_Result : Boolean;
Map_Result_1 : Internal_Entity_Array_Access;
Scope_Result_7 : Internal_Entity_Array_Access;
Is_A : Boolean;
Not_Val_1 : Boolean;
Map_Result_2 : Internal_Entity_Array_Access;
Result_Var_4 : Internal_Entity_Array_Access;
If_Result_9 : Internal_Entity_Array_Access;
Dyn_Var_Bind_Result : Internal_Entity_Array_Access;
Let_Result_4 : Internal_Entity_Array_Access;
            procedure Finalizer_Scope_1122 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Scope_Result_7);
                     Dec_Ref (Map_Result_2);
                     Dec_Ref (Result_Var_4);
                     Dec_Ref (If_Result_9);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Let_Result_4);
            end Finalizer_Scope_1122;
            procedure Finalizer_Scope_2938 is
            begin
                     Dec_Ref (Map_Result_1);
            end Finalizer_Scope_2938;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Base_Id_Base_Id_P_Env_Elements_Baseid,
            Items    => new Mmz_Key_Array (1 ..  3))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Base_Id,
                                As_Bare_Base_Id => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Lexical_Env,
                                      As_Lexical_Env => Env);
                  Inc_Ref (Mmz_K.Items (2).As_Lexical_Env);
               Mmz_K.Items (3) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 9736 '<Block at ???>' Let_Result_4 None
--# scope-start
--# expr-start 9689 '<FieldAccess for env_get at ast.py:15721>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15721
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9683 '<GetSymbol at ast.py:15723>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15723
Sym := Get_Symbol (Self); 
--# expr-done 9683
--# expr-start 9685 '<If at ast.py:15724>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15724
--# expr-start 9684 '<FieldAccess for is_prefix at ast.py:15724>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15724
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_prefix
Fld := Libadalang.Implementation.Name_P_Is_Prefix (Node => Self);
--# end
--# expr-done 9684
if Fld then
   If_Result := Recursive;
else
   If_Result := Flat;
end if;
--# expr-done 9685
--# expr-start 9686 '<FieldAccess for origin_node at ast.py:15726>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15726
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_1 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 9686
--# expr-start 9688 '<If at ast.py:15727>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15727
--# expr-start 9687 '<FieldAccess for can_designate_primitive at ast.py:15728>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15728
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.can_designate_primitive
Fld_2 := Libadalang.Implementation.Name_P_Can_Designate_Primitive (Node => Self);
--# end
--# expr-done 9687
if Fld_2 then
   If_Result_1 := (Inherited_Primitives => True, Nocat => True);
else
   If_Result_1 := (Inherited_Primitives => False, Nocat => True);
end if;
--# expr-done 9688
--# property-call-start AdaNode.env_get
Fld_3 := Libadalang.Implementation.Ada_Node_P_Env_Get (Node => Self, Env => Env, Symbol => Sym, Lookup => If_Result, From_Node => Fld_1, Categories => If_Result_1);
--# end
--# expr-done 9689
Block_Var := Fld_3; Inc_Ref (Block_Var);
--# bind items Block_Var
--# expr-start 9690 '<FieldAccess for parent_callexpr at ast.py:15737>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15737
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.parent_callexpr
Fld_4 := Libadalang.Implementation.Name_P_Parent_Callexpr (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 9690
Block_Var_1 := Fld_4; 
--# bind pc Block_Var_1
--# expr-start 9692 '<Not at ast.py:15738>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15738
--# expr-start 9691 '<FieldAccess for is_suffix at ast.py:15738>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15738
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_suffix
Fld_5 := Libadalang.Implementation.Name_P_Is_Suffix (Node => Self);
--# end
--# expr-done 9691
Not_Val := not (Fld_5); 
--# expr-done 9692
Block_Var_2 := Not_Val; 
--# bind is_prefix Block_Var_2
--# expr-start 9735 '<bind at ast.py:15740>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15740
--# expr-start 9693 '<FieldAccess for origin_node at ast.py:15740>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15740
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_6 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 9693
Bound_Origin := Fld_6; 
--# expr-start 9734 '<Cond at ast.py:15740>' If_Result_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15740
--# expr-start 9694 '<IsNull at ast.py:15741>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15741
Is_Null := Block_Var_1.Node = null; 
--# expr-done 9694
if Is_Null then
   --# expr-start 9699 '<Filter at ast.py:15747>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15747
      declare
         Map_Result_Vec : Internal_Entity_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Block_Var;
      begin
         for Item_297 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_297
--# expr-start 9696 '<FieldAccess for can_be_paramless at ast.py:15752>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15752
--# expr-start 9695 '<Cast to BasicDecl at ast.py:15752>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15752
Cast_Expr := Item_297; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 9695
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.can_be_paramless
Fld_7 := Libadalang.Implementation.Basic_Decl_P_Can_Be_Paramless (Node => Cast_Result.Node, E_Info => Cast_Result.Info);
--# end
--# expr-done 9696
if Fld_7 then
   If_Result_3 := True;
else
if Block_Var_2 then
--# expr-start 9697 '<Cast to BaseSubpBody at ast.py:15757>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15757
Cast_Expr_1 := Item_297; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Base_Subp_Body
   then
      Cast_Result_1 := Create_Internal_Entity_Base_Subp_Body
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_1 := No_Entity_Base_Subp_Body;
   end if;
--# expr-done 9697
Var_Expr := Cast_Result_1;
if Var_Expr /= No_Entity_Base_Subp_Body then
   --# scope-start
--# expr-start 9698 '<FieldAccess for in_scope at ast.py:15757>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15757
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.in_scope
Fld_8 := Libadalang.Implementation.Body_Node_P_In_Scope (Node => Var_Expr.Node, Origin => Bound_Origin);
--# end
--# expr-done 9698
Scope_Result := Fld_8;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
   If_Result_2 := Result_Var;
else
   If_Result_2 := False;
end if;
   If_Result_3 := If_Result_2;
end if;
         if If_Result_3 then
         declare
            Item_To_Append : constant Internal_Entity := Item_297;
         begin
            Internal_Entity_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 9699
   If_Result_9 := Map_Result;
else
--# expr-start 9701 '<Cast to AssocList at ast.py:15767>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15767
--# expr-start 9700 '<FieldAccess for suffix at ast.py:15767>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15767
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity (Node => Block_Var_1.Node.Call_Expr_F_Suffix, Info => Block_Var_1.Info);
--# expr-done 9700
   if Fld_9.Node = null
      or else Fld_9.Node.Kind in Ada_Assoc_List_Range
   then
      Cast_Result_2 := Create_Internal_Entity_Assoc_List
        (Node => Fld_9.Node,
         Info => Fld_9.Info);
   else
         Cast_Result_2 := No_Entity_Assoc_List;
   end if;
--# expr-done 9701
Var_Expr_1 := Cast_Result_2;
if Var_Expr_1 /= No_Entity_Assoc_List then
   --# scope-start
   --# bind params Var_Expr_1
--# expr-start 9730 '<Filter at ast.py:15768>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15768
      declare
         Map_Result_1_Vec : Internal_Entity_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Block_Var;
      begin
         for Item_299 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_299
         --# expr-start 9729 '<Match at ast.py:15768>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15768
      if Item_299.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Item_299; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Discrete_Base_Subtype_Decl .. Ada_Synth_Anonymous_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_6 := Create_Internal_Entity_Base_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_6; 
--# bind _ Match_0
--# expr-start 9703 '<Eq at ast.py:15770>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15770
--# expr-start 9702 '<length at ast.py:15770>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15770
Fld_10 := Var_Expr_1.Node; 
Len := Length (Fld_10); 
--# expr-done 9702
Is_Equal := Len = 1; 
--# expr-done 9703
Let_Result_1 := Is_Equal; 
--# end
Scope_Result_4 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_4; 
         when Ada_Abstract_State_Decl .. Ada_Package_Decl | Ada_Abstract_Subp_Decl .. Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_7 := Create_Internal_Entity_Basic_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_7; 
--# bind b Match_1
--# expr-start 9704 '<FieldAccess for subp_spec_or_null at ast.py:15773>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15773
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_11 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Match_1.Node, E_Info => Match_1.Info);
--# end
--# expr-done 9704
Var_Expr_2 := Fld_11;
if Var_Expr_2 /= No_Entity_Base_Subp_Spec then
   --# scope-start
   --# bind spec Var_Expr_2
--# expr-start 9726 '<Let at ast.py:15774>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15774
--# scope-start
--# expr-start 9710 '<If at ast.py:15775>' If_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15775
--# expr-start 9707 '<IsNull at ast.py:15776>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15776
--# expr-start 9705 '<Cast to EntrySpec at ast.py:15776>' Cast_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15776
Cast_Expr_2 := Var_Expr_2; 
   if Cast_Expr_2.Node = null
      or else Cast_Expr_2.Node.Kind in Ada_Entry_Spec_Range
   then
      Cast_Result_4 := Create_Internal_Entity_Entry_Spec
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
   else
         Cast_Result_4 := No_Entity_Entry_Spec;
   end if;
--# expr-done 9705
Var_Expr_3 := Cast_Result_4;
if Var_Expr_3 /= No_Entity_Entry_Spec then
   --# scope-start
--# expr-start 9706 '<FieldAccess for family_type at ast.py:15776>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15776
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Create_Internal_Entity (Node => Var_Expr_3.Node.Entry_Spec_F_Family_Type, Info => Var_Expr_3.Info);
--# expr-done 9706
Scope_Result_1 := Fld_13;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Entity;
end if;
Is_Null_1 := Result_Var_1.Node = null; 
--# expr-done 9707
if Is_Null_1 then
   If_Result_4 := Block_Var_1;
else
   --# expr-start 9709 '<Cast to CallExpr at ast.py:15777>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15777
--# expr-start 9708 '<parent at ast.py:15777>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15777
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Libadalang.Implementation.Parent (Node => Block_Var_1.Node, E_Info => Block_Var_1.Info);
--# expr-done 9708
   if Fld_12.Node = null
      or else Fld_12.Node.Kind in Ada_Call_Expr_Range
   then
      Cast_Result_3 := Create_Internal_Entity_Call_Expr
        (Node => Fld_12.Node,
         Info => Fld_12.Info);
   else
         Cast_Result_3 := No_Entity_Call_Expr;
   end if;
--# expr-done 9709
   If_Result_4 := Cast_Result_3;
end if;
--# expr-done 9710
Real_Pc := If_Result_4; 
--# bind real_pc Real_Pc
--# expr-start 9711 '<IsNull at ast.py:15783>' Is_Null_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15783
Is_Null_2 := Real_Pc.Node = null; 
--# expr-done 9711
if Is_Null_2 then
   If_Result_6 := True;
else
--# expr-start 9715 '<FieldAccess for is_matching_param_list at ast.py:15788>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15788
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9714 '<FieldAccess for dottable_subp at ast.py:15789>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15789
--# expr-start 9713 '<FieldAccess for md at ast.py:15789>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15789
--# expr-start 9712 '<FieldAccess for info at ast.py:15789>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15789
Fld_14 := Match_1.Info;
--# expr-done 9712
Fld_15 := Fld_14.Md;
--# expr-done 9713
Fld_16 := Fld_15.Dottable_Subp;
--# expr-done 9714
--# property-call-start BaseFormalParamHolder.is_matching_param_list
Fld_17 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Is_Matching_Param_List (Node => Var_Expr_2.Node, Params => Var_Expr_1, Is_Dottable_Subp => Fld_16, Env => Env);
--# end
--# expr-done 9715
if Fld_17 then
--# expr-start 9717 '<Cast to CallExpr at ast.py:15791>' Cast_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15791
--# expr-start 9716 '<parent at ast.py:15791>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15791
      if Real_Pc.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Libadalang.Implementation.Parent (Node => Real_Pc.Node, E_Info => Real_Pc.Info);
--# expr-done 9716
   if Fld_18.Node = null
      or else Fld_18.Node.Kind in Ada_Call_Expr_Range
   then
      Cast_Result_5 := Create_Internal_Entity_Call_Expr
        (Node => Fld_18.Node,
         Info => Fld_18.Info);
   else
         Cast_Result_5 := No_Entity_Call_Expr;
   end if;
--# expr-done 9717
Var_Expr_4 := Cast_Result_5;
if Var_Expr_4 /= No_Entity_Call_Expr then
   --# scope-start
   --# bind ce Var_Expr_4
--# expr-start 9719 '<FieldAccess for check_for_type at ast.py:15792>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15792
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9718 '<FieldAccess for expr_type at ast.py:15792>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15792
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.expr_type'
Fld_19 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Expr_Type (Node => Match_1.Node, Origin => Bound_Origin, E_Info => Match_1.Info);
--# end
--# expr-done 9718
--# property-call-start CallExpr.check_for_type
Fld_20 := Libadalang.Implementation.Call_Expr_P_Check_For_Type (Node => Var_Expr_4.Node, Typ => Fld_19, Env => Env, Origin => Bound_Origin, E_Info => Var_Expr_4.Info);
--# end
--# expr-done 9719
Scope_Result_2 := Fld_20;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := True;
end if;
   If_Result_5 := Result_Var_2;
else
   If_Result_5 := False;
end if;
   If_Result_6 := If_Result_5;
end if;
if If_Result_6 then
   If_Result_8 := True;
else
--# expr-start 9721 '<FieldAccess for check_for_type at ast.py:15799>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15799
      if Real_Pc.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9720 '<FieldAccess for expr_type at ast.py:15799>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15799
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.expr_type'
Fld_21 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Expr_Type (Node => Match_1.Node, Origin => Bound_Origin, E_Info => Match_1.Info);
--# end
--# expr-done 9720
--# property-call-start CallExpr.check_for_type
Fld_22 := Libadalang.Implementation.Call_Expr_P_Check_For_Type (Node => Real_Pc.Node, Typ => Fld_21, Env => Env, Origin => Bound_Origin, E_Info => Real_Pc.Info);
--# end
--# expr-done 9721
if Fld_22 then
   --# expr-start 9725 '<FieldAccess for paramless at ast.py:15800>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15800
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9724 '<FieldAccess for dottable_subp at ast.py:15800>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15800
--# expr-start 9723 '<FieldAccess for md at ast.py:15800>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15800
--# expr-start 9722 '<FieldAccess for info at ast.py:15800>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15800
Fld_23 := Match_1.Info;
--# expr-done 9722
Fld_24 := Fld_23.Md;
--# expr-done 9723
Fld_25 := Fld_24.Dottable_Subp;
--# expr-done 9724
--# property-call-start BaseFormalParamHolder.paramless
Fld_26 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Paramless (Node => Var_Expr_2.Node, Dottable_Subp => Fld_25);
--# end
--# expr-done 9725
   If_Result_7 := Fld_26;
else
   If_Result_7 := False;
end if;
   If_Result_8 := If_Result_7;
end if;
Let_Result := If_Result_8; 
--# end
--# expr-done 9726
Scope_Result_3 := Let_Result;
   --# end
   Result_Var_3 := Scope_Result_3;
else
   --# expr-start 9728 '<FieldAccess for check_for_type at ast.py:15806>' Fld_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15806
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9727 '<FieldAccess for expr_type at ast.py:15806>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15806
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.expr_type'
Fld_27 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Expr_Type (Node => Match_1.Node, Origin => Bound_Origin, E_Info => Match_1.Info);
--# end
--# expr-done 9727
--# property-call-start CallExpr.check_for_type
Fld_28 := Libadalang.Implementation.Call_Expr_P_Check_For_Type (Node => Block_Var_1.Node, Typ => Fld_27, Env => Env, Origin => Bound_Origin, E_Info => Block_Var_1.Info);
--# end
--# expr-done 9728
   Result_Var_3 := Fld_28;
end if;
Let_Result_2 := Result_Var_3; 
--# end
Scope_Result_5 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_5; 
         when Ada_Abort_Absent .. Ada_Param_Assoc | Ada_Case_Stmt_Alternative .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_8 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_8; 
--# bind _ Match_2
Let_Result_3 := False; 
--# end
Scope_Result_6 := Let_Result_3;
   --# end
            Match_Result := Scope_Result_6; 
end case;
--# expr-done 9729
         if Match_Result then
         declare
            Item_To_Append : constant Internal_Entity := Item_299;
         begin
            Internal_Entity_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result_1 := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 9730
Scope_Result_7 := Map_Result_1;
   Inc_Ref (Scope_Result_7);
   --# end
      Finalizer_Scope_2938;
   Result_Var_4 := Scope_Result_7;
else
   --# expr-start 9733 '<Filter at ast.py:15814>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15814
      declare
         Map_Result_2_Vec : Internal_Entity_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Block_Var;
      begin
         for Item_298 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_298
         --# expr-start 9732 '<Not at ast.py:15814>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15814
--# expr-start 9731 '<IsA BaseTypeDecl at ast.py:15814>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15814
Is_A := Item_298.Node /= null 
and then Item_298.Node.Kind in Ada_Base_Type_Decl; 
--# expr-done 9731
Not_Val_1 := not (Is_A); 
--# expr-done 9732
         if Not_Val_1 then
         declare
            Item_To_Append : constant Internal_Entity := Item_298;
         begin
            Internal_Entity_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result_2 := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_2_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 9733
   Result_Var_4 := Map_Result_2;
end if;
      Inc_Ref (Result_Var_4);
   If_Result_9 := Result_Var_4;
end if;
      Inc_Ref (If_Result_9);
--# expr-done 9734
Dyn_Var_Bind_Result := If_Result_9; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 9735
Let_Result_4 := Dyn_Var_Bind_Result; Inc_Ref (Let_Result_4);
--# end
--# expr-done 9736
      Property_Result := Let_Result_4;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1122;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Array_Access,
                        As_Internal_Entity_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1122;
               Finalizer_Scope_2938;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Id_P_Env_Elements_Baseid;
--# end
--# property-start BaseId.denotes_the_property_function /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15819
pragma Warnings (Off, "is not referenced");
 function Base_Id_P_Denotes_The_Property_Function
  (Node : Bare_Base_Id
      ; Subp_Spec : Internal_Entity_Base_Subp_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Base_Id := Bare_Base_Id (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Id :=
     Internal_Entity_Base_Id'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind subp_spec Subp_Spec
   Property_Result : Boolean;
            procedure Finalizer_Scope_1123 with Inline_Always;
            procedure Finalizer_Scope_2995 with Inline_Always;
      Block_Var : Internal_Entity_Base_Type_Decl_Array_Access;
Match_0 : Internal_Entity_Type_Decl;
Match_1 : Internal_Entity_Subp_Decl;
Match_2 : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Base_Type_Decl_Array_Access;
Fld_1 : Internal_Entity_Param_Spec_Array_Access;
Len : Integer;
Is_Equal : Boolean;
Fld_2 : Internal_Entity_Basic_Decl;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Type_Decl;
Cast_Result_1 : Internal_Entity;
Is_Equiv : Boolean;
Find_Result : Internal_Entity_Base_Type_Decl;
Is_Null : Boolean;
Not_Val : Boolean;
Fld_3 : Internal_Entity_Subp_Spec;
Fld_4 : Internal_Entity_Base_Type_Decl_Array_Access;
Cast_Expr_2 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Cast_Expr_3 : Internal_Entity_Base_Type_Decl;
Cast_Result_3 : Internal_Entity;
Is_Equiv_1 : Boolean;
Find_Result_1 : Internal_Entity_Base_Type_Decl;
Is_Null_1 : Boolean;
Not_Val_1 : Boolean;
Map_Result : Internal_Entity_Base_Type_Decl_Array_Access;
Is_Equal_1 : Boolean;
Not_Val_2 : Boolean;
Match_Prefix : Internal_Entity_Basic_Decl;
Cast_Result_4 : Internal_Entity_Type_Decl;
Let_Result : Boolean;
Scope_Result : Boolean;
Cast_Result_5 : Internal_Entity_Subp_Decl;
Let_Result_1 : Boolean;
Scope_Result_1 : Boolean;
Cast_Result_6 : Internal_Entity_Basic_Decl;
Let_Result_2 : Boolean;
Scope_Result_2 : Boolean;
Match_Result : Boolean;
If_Result : Boolean;
Let_Result_3 : Boolean;
            procedure Finalizer_Scope_1123 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1123;
            procedure Finalizer_Scope_2995 is
            begin
                     Dec_Ref (Fld_4);
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_2995;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9756 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 9737 '<FieldAccess for primitive_subp_types at ast.py:15825>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15825
      if Subp_Spec.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.primitive_subp_types
Fld := Libadalang.Implementation.Base_Subp_Spec_P_Primitive_Subp_Types (Node => Subp_Spec.Node, Imprecise_Fallback => False, E_Info => Subp_Spec.Info);
--# end
--# expr-done 9737
Block_Var := Fld; Inc_Ref (Block_Var);
--# bind primitive_types Block_Var
--# expr-start 9740 '<Eq at ast.py:15831>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15831
--# expr-start 9739 '<length at ast.py:15831>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15831
--# expr-start 9738 '<FieldAccess for params at ast.py:15831>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15831
      if Subp_Spec.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.params'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Params (Node => Subp_Spec.Node, E_Info => Subp_Spec.Info);
--# end
--# expr-done 9738
Len := Length (Fld_1); 
--# expr-done 9739
Is_Equal := Len = 1; 
--# expr-done 9740
if Is_Equal then
   --# expr-start 9755 '<Match at ast.py:15835>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15835
--# expr-start 9741 '<FieldAccess for parent_basic_decl at ast.py:15835>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15835
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.parent_basic_decl
Fld_2 := Libadalang.Implementation.Ada_Node_P_Parent_Basic_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 9741
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld_2; 
case Ada_Basic_Decl (Match_Prefix.Node.Kind) is
         when Ada_Type_Decl .. Ada_Synth_Anonymous_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_4; 
--# bind td Match_0
--# expr-start 9745 '<Not at ast.py:15838>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15838
--# expr-start 9744 '<IsNull at ast.py:15839>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15839
--# expr-start 9743 '<Find at ast.py:15839>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15839
      Find_Result := No_Entity_Base_Type_Decl;
      declare
         Collection : constant Internal_Entity_Base_Type_Decl_Array_Access := Block_Var;
      begin
         for Item_300 of
               Collection.Items
         loop
   --# scope-start
               --# bind t Item_300
      --# expr-start 9742 '<Eq at ast.py:15840>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15840
Cast_Expr := Item_300; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Cast_Expr_1 := Match_0; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Is_Equiv := Equivalent (Cast_Result, Cast_Result_1); 
--# expr-done 9742
      if Is_Equiv then
         Find_Result := Item_300;
      end if;
   --# end
               exit when Is_Equiv;
         end loop;
      end;
--# expr-done 9743
Is_Null := Find_Result.Node = null; 
--# expr-done 9744
Not_Val := not (Is_Null); 
--# expr-done 9745
Let_Result := Not_Val; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Subp_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Create_Internal_Entity_Subp_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_5; 
--# bind sd Match_1
--# expr-start 9754 '<Not at ast.py:15844>' Not_Val_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15844
--# expr-start 9753 '<IsNull at ast.py:15845>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15845
--# expr-start 9752 '<Filter at ast.py:15845>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15845
   --# expr-start 9747 '<FieldAccess for primitive_subp_types at ast.py:15845>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15845
--# expr-start 9746 '<FieldAccess for subp_spec at ast.py:15845>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15845
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Subp_Spec (Node => Match_1.Node.Classic_Subp_Decl_F_Subp_Spec, Info => Match_1.Info);
--# expr-done 9746
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.primitive_subp_types
Fld_4 := Libadalang.Implementation.Base_Subp_Spec_P_Primitive_Subp_Types (Node => Fld_3.Node, Imprecise_Fallback => False, E_Info => Fld_3.Info);
--# end
--# expr-done 9747
      declare
         Map_Result_Vec : Internal_Entity_Base_Type_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Base_Type_Decl_Array_Access := Fld_4;
      begin
         for Item_301 of
               Collection.Items
         loop
   --# scope-start
               --# bind t1 Item_301
         --# expr-start 9751 '<Not at ast.py:15846>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15846
--# expr-start 9750 '<IsNull at ast.py:15847>' Is_Null_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15847
--# expr-start 9749 '<Find at ast.py:15847>' Find_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15847
      Find_Result_1 := No_Entity_Base_Type_Decl;
      declare
         Collection : constant Internal_Entity_Base_Type_Decl_Array_Access := Block_Var;
      begin
         for Item_302 of
               Collection.Items
         loop
   --# scope-start
               --# bind t2 Item_302
      --# expr-start 9748 '<Eq at ast.py:15848>' Is_Equiv_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15848
Cast_Expr_2 := Item_301; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
Cast_Expr_3 := Item_302; 
      Cast_Result_3 := Create_Internal_Entity
        (Node => Cast_Expr_3.Node,
         Info => Cast_Expr_3.Info);
Is_Equiv_1 := Equivalent (Cast_Result_2, Cast_Result_3); 
--# expr-done 9748
      if Is_Equiv_1 then
         Find_Result_1 := Item_302;
      end if;
   --# end
               exit when Is_Equiv_1;
         end loop;
      end;
--# expr-done 9749
Is_Null_1 := Find_Result_1.Node = null; 
--# expr-done 9750
Not_Val_1 := not (Is_Null_1); 
--# expr-done 9751
         if Not_Val_1 then
         declare
            Item_To_Append : constant Internal_Entity_Base_Type_Decl := Item_301;
         begin
            Internal_Entity_Base_Type_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Base_Type_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Type_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Base_Type_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Base_Type_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Base_Type_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 9752
Is_Equal_1 := Equivalent (Map_Result, No_Internal_Entity_Base_Type_Decl_Array_Type); 
--# expr-done 9753
Not_Val_2 := not (Is_Equal_1); 
--# expr-done 9754
Let_Result_1 := Not_Val_2; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
      Finalizer_Scope_2995;
            Match_Result := Scope_Result_1; 
         when Ada_Abstract_State_Decl .. Ada_Single_Task_Type_Decl | Ada_Abstract_Subp_Decl .. Ada_Concrete_Formal_Subp_Decl | Ada_Entry_Decl .. Ada_Single_Task_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_6 := Create_Internal_Entity_Basic_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_6; 
--# bind _ Match_2
Let_Result_2 := False; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 9755
   If_Result := Match_Result;
else
   If_Result := False;
end if;
Let_Result_3 := If_Result; 
--# end
--# expr-done 9756
      Property_Result := Let_Result_3;
   --# end
      Finalizer_Scope_1123;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1123;
               Finalizer_Scope_2995;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Id_P_Denotes_The_Property_Function;
--# end
--# property-start BaseId.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15858
pragma Warnings (Off, "is not referenced");
 function Base_Id_P_Xref_Equation
  (Node : Bare_Base_Id
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Base_Id := Bare_Base_Id (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Id :=
     Internal_Entity_Base_Id'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1124 with Inline_Always;
      Fld : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1124 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1124;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9758 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9757 '<FieldAccess for base_id_xref_equation at ast.py:15859>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15859
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseId.base_id_xref_equation
Fld := Libadalang.Implementation.Base_Id_P_Base_Id_Xref_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 9757
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 9758
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1124;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1124;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Id_P_Xref_Equation;
--# end
--# property-start BaseId.base_id_xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15862
pragma Warnings (Off, "is not referenced");
 function Base_Id_P_Base_Id_Xref_Equation
  (Node : Bare_Base_Id
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Base_Id := Bare_Base_Id (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Id :=
     Internal_Entity_Base_Id'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1125 with Inline_Always;
            procedure Finalizer_Scope_2999 with Inline_Always;
      Block_Var : Internal_Entity_Array_Access;
Block_Var_1 : Boolean;
Var_Expr : Internal_Entity_Base_Subp_Body;
Fld : Internal_Entity_Array_Access;
Fld_1 : Boolean;
Not_Val : Boolean;
Fld_2 : Logic_Var;
Bind_Result : Logic_Equation;
Fld_3 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result_1 : Logic_Equation;
Fld_4 : Logic_Var;
Fld_5 : Logic_Var;
Bind_Result_2 : Logic_Equation;
Fld_6 : Logic_Var;
Fld_7 : Logic_Var;
Bind_Result_3 : Logic_Equation;
And_Pred : Logic_Equation;
Cast_Expr_1 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Base_Subp_Body;
Fld_8 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
If_Result : Boolean;
If_Result_1 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1125 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1125;
            procedure Finalizer_Scope_2999 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (Bind_Result_3);
                     Dec_Ref (And_Pred);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_2999;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9779 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9759 '<FieldAccess for env_elements at ast.py:15863>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15863
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.env_elements
Fld := Libadalang.Implementation.Expr_P_Env_Elements (Node => Ent.Node, Env => Env, No_Visibility => False, E_Info => Ent.Info);
--# end
--# expr-done 9759
Block_Var := Fld; Inc_Ref (Block_Var);
--# bind env_els Block_Var
--# expr-start 9761 '<Not at ast.py:15864>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15864
--# expr-start 9760 '<FieldAccess for is_suffix at ast.py:15864>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15864
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.is_suffix
Fld_1 := Libadalang.Implementation.Name_P_Is_Suffix (Node => Self);
--# end
--# expr-done 9760
Not_Val := not (Fld_1); 
--# expr-done 9761
Block_Var_1 := Not_Val; 
--# bind is_prefix Block_Var_1
--# expr-start 9778 '<LogicAny at ast.py:15866>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15866
--# expr-start 9777 '<Map at ast.py:15866>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15866
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Block_Var;
      begin
         for Item_303 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_303
      --# expr-start 9776 '<And at ast.py:15868>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15868
--# expr-start 9763 '<Bind at ast.py:15868>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15868
--# expr-start 9762 '<FieldAccess for ref_var at ast.py:15868>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15868
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Self);
--# end
--# expr-done 9762
Fld_2.Value := No_Entity;
Entity_Vars.Reset (Fld_2);
Bind_Result := Solver.Create_Assign (Fld_2, Item_303, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:15868") else null)); 
--# expr-done 9763
--# expr-start 9775 '<If at ast.py:15869>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15869
if Block_Var_1 then
--# expr-start 9764 '<Cast to BaseSubpBody at ast.py:15872>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15872
Cast_Expr_1 := Item_303; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Base_Subp_Body
   then
      Cast_Result_1 := Create_Internal_Entity_Base_Subp_Body
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_1 := No_Entity_Base_Subp_Body;
   end if;
--# expr-done 9764
Var_Expr := Cast_Result_1;
if Var_Expr /= No_Entity_Base_Subp_Body then
   --# scope-start
--# expr-start 9765 '<FieldAccess for in_scope at ast.py:15872>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15872
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Body.in_scope
Fld_8 := Libadalang.Implementation.Body_Node_P_In_Scope (Node => Var_Expr.Node, Origin => Origin);
--# end
--# expr-done 9765
Scope_Result := Fld_8;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
   If_Result := Result_Var;
else
   If_Result := False;
end if;
if If_Result then
   --# expr-start 9767 '<Bind at ast.py:15874>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15874
--# expr-start 9766 '<FieldAccess for type_var at ast.py:15874>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15874
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 9766
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
Cast_Expr := No_Entity_Base_Type_Decl; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_3, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:15874") else null)); 
--# expr-done 9767
   If_Result_1 := Bind_Result_1;
else
   --# expr-start 9774 '<And at ast.py:15881>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15881
--# expr-start 9770 '<Bind at ast.py:15881>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15881
--# expr-start 9768 '<FieldAccess for ref_var at ast.py:15881>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15881
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_4 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Self);
--# end
--# expr-done 9768
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
--# expr-start 9769 '<FieldAccess for type_var at ast.py:15881>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15881
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 9769
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
Bind_Result_2 := Solver.Create_Propagate (Fld_4, Fld_5, Logic_Functor_207'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Origin), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:15881") else null)); 
--# expr-done 9770
--# expr-start 9773 '<Bind at ast.py:15883>' Bind_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15883
--# expr-start 9771 '<FieldAccess for ref_var at ast.py:15883>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15883
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_6 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Self);
--# end
--# expr-done 9771
Fld_6.Value := No_Entity;
Entity_Vars.Reset (Fld_6);
--# expr-start 9772 '<FieldAccess for subp_spec_var at ast.py:15883>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15883
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.subp_spec_var'
Fld_7 := Libadalang.Implementation.Dispatcher_Name_P_Subp_Spec_Var (Node => Self);
--# end
--# expr-done 9772
Fld_7.Value := No_Entity;
Entity_Vars.Reset (Fld_7);
Bind_Result_3 := Solver.Create_Propagate (Fld_6, Fld_7, Logic_Functor_209'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:15883") else null)); 
--# expr-done 9773
And_Pred := Create_And (Bind_Result_2, Bind_Result_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:15881") else null)); 
--# expr-done 9774
   If_Result_1 := And_Pred;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 9775
And_Pred_1 := Create_And (Bind_Result, If_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:15868") else null)); 
--# expr-done 9776
         declare
            Item_To_Append : constant Logic_Equation := And_Pred_1;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2999;
         end loop;
      end;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 9777
Logic_Boolean_Op := Solver.Create_Any (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:15866") else null)); 
--# expr-done 9778
Let_Result := Logic_Boolean_Op; Inc_Ref (Let_Result);
--# end
--# expr-done 9779
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1125;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1125;
               Finalizer_Scope_2999;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Id_P_Base_Id_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Char_Literal
   --
      procedure Initialize_Fields_For_Char_Literal
        (Self : Bare_Char_Literal
        ) is
      begin
            Initialize_Fields_For_Base_Id
              (Self);
      end Initialize_Fields_For_Char_Literal;
--# property-start CharLiteral.has_context_free_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16167
pragma Warnings (Off, "is not referenced");
 function Char_Literal_P_Has_Context_Free_Type
  (Node : Bare_Char_Literal
  )
   return Boolean
is
   Self : Bare_Char_Literal := Bare_Char_Literal (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Char_Literal_P_Has_Context_Free_Type;
--# end
--# property-start CharLiteral.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16170
pragma Warnings (Off, "is not referenced");
 function Char_Literal_P_Xref_Equation
  (Node : Bare_Char_Literal
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Char_Literal := Bare_Char_Literal (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1151 with Inline_Always;
      Fld : Logic_Var;
Pred : Logic_Equation;
Fld_1 : Logic_Var;
Fld_2 : Logic_Var;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Cast_Expr : Bare_Char_Literal;
Cast_Result : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_3 : Logic_Var;
Fld_4 : Logic_Var;
Bind_Result_1 : Logic_Equation;
Dyn_Var_Bind_Result : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1151 is
            begin
                     Dec_Ref (Pred);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1151;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9791 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9790 '<And at ast.py:16171>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16171
--# expr-start 9785 '<And at ast.py:16171>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16171
--# expr-start 9781 '<Predicate on BaseTypeDecl.is_non_null_char_type at ast.py:16172>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16172
--# expr-start 9780 '<FieldAccess for expected_type_var at ast.py:16173>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16173
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 9780
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
Pred := Solver.Create_Predicate (Fld, Create_Base_Type_Decl_P_Is_Non_Null_Char_Type_0_Predicate (Origin)); 
--# expr-done 9781
--# expr-start 9784 '<Bind at ast.py:16174>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16174
--# expr-start 9782 '<FieldAccess for expected_type_var at ast.py:16174>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16174
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 9782
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 9783 '<FieldAccess for type_var at ast.py:16174>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16174
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 9783
Fld_2.Value := No_Entity;
Entity_Vars.Reset (Fld_2);
Bind_Result := Solver.Create_Unify (Fld_1, Fld_2, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:16174") else null)); 
--# expr-done 9784
And_Pred := Create_And (Pred, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:16171") else null)); 
--# expr-done 9785
--# expr-start 9789 '<bind at ast.py:16183>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16183
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Bound_Origin := Cast_Result; 
--# expr-start 9788 '<Bind at ast.py:16185>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16185
--# expr-start 9786 '<FieldAccess for type_var at ast.py:16185>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16185
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 9786
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
--# expr-start 9787 '<FieldAccess for ref_var at ast.py:16185>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16185
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_4 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Self);
--# end
--# expr-done 9787
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
Bind_Result_1 := Solver.Create_Propagate (Fld_3, Fld_4, Logic_Functor_645'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Bound_Origin), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:16185") else null)); 
--# expr-done 9788
Dyn_Var_Bind_Result := Bind_Result_1; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 9789
And_Pred_1 := Create_And (And_Pred, Dyn_Var_Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:16171") else null)); 
--# expr-done 9790
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 9791
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1151;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1151;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Char_Literal_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Identifier
   --
      procedure Initialize_Fields_For_Identifier
        (Self : Bare_Identifier
        ) is
      begin
            Initialize_Fields_For_Base_Id
              (Self);
      end Initialize_Fields_For_Identifier;
--# property-start Identifier.is_attr_with_args /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15997
pragma Warnings (Off, "is not referenced");
 function Identifier_P_Is_Attr_With_Args
  (Node : Bare_Identifier
  )
   return Boolean
is
   Self : Bare_Identifier := Bare_Identifier (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Sym : Symbol_Type;
Sym_1 : Symbol_Type;
Is_Equal : Boolean;
Sym_2 : Symbol_Type;
Sym_3 : Symbol_Type;
Is_Equal_1 : Boolean;
If_Result : Boolean;
Sym_4 : Symbol_Type;
Sym_5 : Symbol_Type;
Is_Equal_2 : Boolean;
If_Result_1 : Boolean;
Sym_6 : Symbol_Type;
Sym_7 : Symbol_Type;
Is_Equal_3 : Boolean;
If_Result_2 : Boolean;
Sym_8 : Symbol_Type;
Sym_9 : Symbol_Type;
Is_Equal_4 : Boolean;
If_Result_3 : Boolean;
Sym_10 : Symbol_Type;
Sym_11 : Symbol_Type;
Is_Equal_5 : Boolean;
If_Result_4 : Boolean;
Sym_12 : Symbol_Type;
Sym_13 : Symbol_Type;
Is_Equal_6 : Boolean;
If_Result_5 : Boolean;
Sym_14 : Symbol_Type;
Sym_15 : Symbol_Type;
Is_Equal_7 : Boolean;
If_Result_6 : Boolean;
Sym_16 : Symbol_Type;
Sym_17 : Symbol_Type;
Is_Equal_8 : Boolean;
If_Result_7 : Boolean;
Sym_18 : Symbol_Type;
Sym_19 : Symbol_Type;
Is_Equal_9 : Boolean;
If_Result_8 : Boolean;
Sym_20 : Symbol_Type;
Sym_21 : Symbol_Type;
Is_Equal_10 : Boolean;
If_Result_9 : Boolean;
Sym_22 : Symbol_Type;
Sym_23 : Symbol_Type;
Is_Equal_11 : Boolean;
If_Result_10 : Boolean;
Sym_24 : Symbol_Type;
Sym_25 : Symbol_Type;
Is_Equal_12 : Boolean;
If_Result_11 : Boolean;
Sym_26 : Symbol_Type;
Sym_27 : Symbol_Type;
Is_Equal_13 : Boolean;
If_Result_12 : Boolean;
Sym_28 : Symbol_Type;
Sym_29 : Symbol_Type;
Is_Equal_14 : Boolean;
If_Result_13 : Boolean;
Sym_30 : Symbol_Type;
Sym_31 : Symbol_Type;
Is_Equal_15 : Boolean;
If_Result_14 : Boolean;
Sym_32 : Symbol_Type;
Sym_33 : Symbol_Type;
Is_Equal_16 : Boolean;
If_Result_15 : Boolean;
Sym_34 : Symbol_Type;
Sym_35 : Symbol_Type;
Is_Equal_17 : Boolean;
If_Result_16 : Boolean;
Sym_36 : Symbol_Type;
Sym_37 : Symbol_Type;
Is_Equal_18 : Boolean;
If_Result_17 : Boolean;
Sym_38 : Symbol_Type;
Sym_39 : Symbol_Type;
Is_Equal_19 : Boolean;
If_Result_18 : Boolean;
Sym_40 : Symbol_Type;
Sym_41 : Symbol_Type;
Is_Equal_20 : Boolean;
If_Result_19 : Boolean;
Sym_42 : Symbol_Type;
Sym_43 : Symbol_Type;
Is_Equal_21 : Boolean;
If_Result_20 : Boolean;
Sym_44 : Symbol_Type;
Sym_45 : Symbol_Type;
Is_Equal_22 : Boolean;
If_Result_21 : Boolean;
Sym_46 : Symbol_Type;
Sym_47 : Symbol_Type;
Is_Equal_23 : Boolean;
If_Result_22 : Boolean;
Sym_48 : Symbol_Type;
Sym_49 : Symbol_Type;
Is_Equal_24 : Boolean;
If_Result_23 : Boolean;
Sym_50 : Symbol_Type;
Sym_51 : Symbol_Type;
Is_Equal_25 : Boolean;
If_Result_24 : Boolean;
Sym_52 : Symbol_Type;
Sym_53 : Symbol_Type;
Is_Equal_26 : Boolean;
If_Result_25 : Boolean;
Sym_54 : Symbol_Type;
Sym_55 : Symbol_Type;
Is_Equal_27 : Boolean;
If_Result_26 : Boolean;
Sym_56 : Symbol_Type;
Sym_57 : Symbol_Type;
Is_Equal_28 : Boolean;
If_Result_27 : Boolean;
Sym_58 : Symbol_Type;
Sym_59 : Symbol_Type;
Is_Equal_29 : Boolean;
If_Result_28 : Boolean;
Sym_60 : Symbol_Type;
Sym_61 : Symbol_Type;
Is_Equal_30 : Boolean;
If_Result_29 : Boolean;
Sym_62 : Symbol_Type;
Sym_63 : Symbol_Type;
Is_Equal_31 : Boolean;
If_Result_30 : Boolean;
Sym_64 : Symbol_Type;
Sym_65 : Symbol_Type;
Is_Equal_32 : Boolean;
If_Result_31 : Boolean;
Sym_66 : Symbol_Type;
Sym_67 : Symbol_Type;
Is_Equal_33 : Boolean;
If_Result_32 : Boolean;
Sym_68 : Symbol_Type;
Sym_69 : Symbol_Type;
Is_Equal_34 : Boolean;
If_Result_33 : Boolean;
Sym_70 : Symbol_Type;
Sym_71 : Symbol_Type;
Is_Equal_35 : Boolean;
If_Result_34 : Boolean;
Sym_72 : Symbol_Type;
Sym_73 : Symbol_Type;
Is_Equal_36 : Boolean;
If_Result_35 : Boolean;
Sym_74 : Symbol_Type;
Sym_75 : Symbol_Type;
Is_Equal_37 : Boolean;
If_Result_36 : Boolean;
Sym_76 : Symbol_Type;
Sym_77 : Symbol_Type;
Is_Equal_38 : Boolean;
If_Result_37 : Boolean;
Sym_78 : Symbol_Type;
Sym_79 : Symbol_Type;
Is_Equal_39 : Boolean;
If_Result_38 : Boolean;
Sym_80 : Symbol_Type;
Sym_81 : Symbol_Type;
Is_Equal_40 : Boolean;
If_Result_39 : Boolean;
Sym_82 : Symbol_Type;
Sym_83 : Symbol_Type;
Is_Equal_41 : Boolean;
If_Result_40 : Boolean;
Sym_84 : Symbol_Type;
Sym_85 : Symbol_Type;
Is_Equal_42 : Boolean;
If_Result_41 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 9794 '<Eq at ast.py:15998>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9792 '<GetSymbol at ast.py:15998>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym := Get_Symbol (Self); 
--# expr-done 9792
--# expr-start 9793 '<Symbol Write at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Write); 
--# expr-done 9793
Is_Equal := Sym = Sym_1; 
--# expr-done 9794
if Is_Equal then
   If_Result := True;
else
   --# expr-start 9797 '<Eq at ast.py:15998>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9795 '<GetSymbol at ast.py:15998>' Sym_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_2 := Get_Symbol (Self); 
--# expr-done 9795
--# expr-start 9796 '<Symbol Read at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Read); 
--# expr-done 9796
Is_Equal_1 := Sym_2 = Sym_3; 
--# expr-done 9797
   If_Result := Is_Equal_1;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 9800 '<Eq at ast.py:15998>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9798 '<GetSymbol at ast.py:15998>' Sym_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_4 := Get_Symbol (Self); 
--# expr-done 9798
--# expr-start 9799 '<Symbol Input at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Input); 
--# expr-done 9799
Is_Equal_2 := Sym_4 = Sym_5; 
--# expr-done 9800
   If_Result_1 := Is_Equal_2;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
   --# expr-start 9803 '<Eq at ast.py:15998>' Is_Equal_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9801 '<GetSymbol at ast.py:15998>' Sym_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_6 := Get_Symbol (Self); 
--# expr-done 9801
--# expr-start 9802 '<Symbol Output at ???>' Sym_7 None
Sym_7 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Output); 
--# expr-done 9802
Is_Equal_3 := Sym_6 = Sym_7; 
--# expr-done 9803
   If_Result_2 := Is_Equal_3;
end if;
if If_Result_2 then
   If_Result_3 := True;
else
   --# expr-start 9806 '<Eq at ast.py:15998>' Is_Equal_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9804 '<GetSymbol at ast.py:15998>' Sym_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_8 := Get_Symbol (Self); 
--# expr-done 9804
--# expr-start 9805 '<Symbol Succ at ???>' Sym_9 None
Sym_9 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Succ); 
--# expr-done 9805
Is_Equal_4 := Sym_8 = Sym_9; 
--# expr-done 9806
   If_Result_3 := Is_Equal_4;
end if;
if If_Result_3 then
   If_Result_4 := True;
else
   --# expr-start 9809 '<Eq at ast.py:15998>' Is_Equal_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9807 '<GetSymbol at ast.py:15998>' Sym_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_10 := Get_Symbol (Self); 
--# expr-done 9807
--# expr-start 9808 '<Symbol Pred at ???>' Sym_11 None
Sym_11 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Pred); 
--# expr-done 9808
Is_Equal_5 := Sym_10 = Sym_11; 
--# expr-done 9809
   If_Result_4 := Is_Equal_5;
end if;
if If_Result_4 then
   If_Result_5 := True;
else
   --# expr-start 9812 '<Eq at ast.py:15998>' Is_Equal_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9810 '<GetSymbol at ast.py:15998>' Sym_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_12 := Get_Symbol (Self); 
--# expr-done 9810
--# expr-start 9811 '<Symbol Min at ???>' Sym_13 None
Sym_13 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Min); 
--# expr-done 9811
Is_Equal_6 := Sym_12 = Sym_13; 
--# expr-done 9812
   If_Result_5 := Is_Equal_6;
end if;
if If_Result_5 then
   If_Result_6 := True;
else
   --# expr-start 9815 '<Eq at ast.py:15998>' Is_Equal_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9813 '<GetSymbol at ast.py:15998>' Sym_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_14 := Get_Symbol (Self); 
--# expr-done 9813
--# expr-start 9814 '<Symbol Max at ???>' Sym_15 None
Sym_15 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Max); 
--# expr-done 9814
Is_Equal_7 := Sym_14 = Sym_15; 
--# expr-done 9815
   If_Result_6 := Is_Equal_7;
end if;
if If_Result_6 then
   If_Result_7 := True;
else
   --# expr-start 9818 '<Eq at ast.py:15998>' Is_Equal_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9816 '<GetSymbol at ast.py:15998>' Sym_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_16 := Get_Symbol (Self); 
--# expr-done 9816
--# expr-start 9817 '<Symbol Value at ???>' Sym_17 None
Sym_17 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Value); 
--# expr-done 9817
Is_Equal_8 := Sym_16 = Sym_17; 
--# expr-done 9818
   If_Result_7 := Is_Equal_8;
end if;
if If_Result_7 then
   If_Result_8 := True;
else
   --# expr-start 9821 '<Eq at ast.py:15998>' Is_Equal_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9819 '<GetSymbol at ast.py:15998>' Sym_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_18 := Get_Symbol (Self); 
--# expr-done 9819
--# expr-start 9820 '<Symbol Wide_Value at ???>' Sym_19 None
Sym_19 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Wide_Value); 
--# expr-done 9820
Is_Equal_9 := Sym_18 = Sym_19; 
--# expr-done 9821
   If_Result_8 := Is_Equal_9;
end if;
if If_Result_8 then
   If_Result_9 := True;
else
   --# expr-start 9824 '<Eq at ast.py:15998>' Is_Equal_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9822 '<GetSymbol at ast.py:15998>' Sym_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_20 := Get_Symbol (Self); 
--# expr-done 9822
--# expr-start 9823 '<Symbol Wide_Wide_Value at ???>' Sym_21 None
Sym_21 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Wide_Wide_Value); 
--# expr-done 9823
Is_Equal_10 := Sym_20 = Sym_21; 
--# expr-done 9824
   If_Result_9 := Is_Equal_10;
end if;
if If_Result_9 then
   If_Result_10 := True;
else
   --# expr-start 9827 '<Eq at ast.py:15998>' Is_Equal_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9825 '<GetSymbol at ast.py:15998>' Sym_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_22 := Get_Symbol (Self); 
--# expr-done 9825
--# expr-start 9826 '<Symbol Fixed_Value at ???>' Sym_23 None
Sym_23 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Fixed_Value); 
--# expr-done 9826
Is_Equal_11 := Sym_22 = Sym_23; 
--# expr-done 9827
   If_Result_10 := Is_Equal_11;
end if;
if If_Result_10 then
   If_Result_11 := True;
else
   --# expr-start 9830 '<Eq at ast.py:15998>' Is_Equal_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9828 '<GetSymbol at ast.py:15998>' Sym_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_24 := Get_Symbol (Self); 
--# expr-done 9828
--# expr-start 9829 '<Symbol Integer_Value at ???>' Sym_25 None
Sym_25 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Integer_Value); 
--# expr-done 9829
Is_Equal_12 := Sym_24 = Sym_25; 
--# expr-done 9830
   If_Result_11 := Is_Equal_12;
end if;
if If_Result_11 then
   If_Result_12 := True;
else
   --# expr-start 9833 '<Eq at ast.py:15998>' Is_Equal_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9831 '<GetSymbol at ast.py:15998>' Sym_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_26 := Get_Symbol (Self); 
--# expr-done 9831
--# expr-start 9832 '<Symbol Pos at ???>' Sym_27 None
Sym_27 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Pos); 
--# expr-done 9832
Is_Equal_13 := Sym_26 = Sym_27; 
--# expr-done 9833
   If_Result_12 := Is_Equal_13;
end if;
if If_Result_12 then
   If_Result_13 := True;
else
   --# expr-start 9836 '<Eq at ast.py:15998>' Is_Equal_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9834 '<GetSymbol at ast.py:15998>' Sym_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_28 := Get_Symbol (Self); 
--# expr-done 9834
--# expr-start 9835 '<Symbol Val at ???>' Sym_29 None
Sym_29 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Val); 
--# expr-done 9835
Is_Equal_14 := Sym_28 = Sym_29; 
--# expr-done 9836
   If_Result_13 := Is_Equal_14;
end if;
if If_Result_13 then
   If_Result_14 := True;
else
   --# expr-start 9839 '<Eq at ast.py:15998>' Is_Equal_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9837 '<GetSymbol at ast.py:15998>' Sym_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_30 := Get_Symbol (Self); 
--# expr-done 9837
--# expr-start 9838 '<Symbol Enum_Rep at ???>' Sym_31 None
Sym_31 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Enum_Rep); 
--# expr-done 9838
Is_Equal_15 := Sym_30 = Sym_31; 
--# expr-done 9839
   If_Result_14 := Is_Equal_15;
end if;
if If_Result_14 then
   If_Result_15 := True;
else
   --# expr-start 9842 '<Eq at ast.py:15998>' Is_Equal_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9840 '<GetSymbol at ast.py:15998>' Sym_32 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_32 := Get_Symbol (Self); 
--# expr-done 9840
--# expr-start 9841 '<Symbol Enum_Val at ???>' Sym_33 None
Sym_33 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Enum_Val); 
--# expr-done 9841
Is_Equal_16 := Sym_32 = Sym_33; 
--# expr-done 9842
   If_Result_15 := Is_Equal_16;
end if;
if If_Result_15 then
   If_Result_16 := True;
else
   --# expr-start 9845 '<Eq at ast.py:15998>' Is_Equal_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9843 '<GetSymbol at ast.py:15998>' Sym_34 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_34 := Get_Symbol (Self); 
--# expr-done 9843
--# expr-start 9844 '<Symbol First at ???>' Sym_35 None
Sym_35 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_First); 
--# expr-done 9844
Is_Equal_17 := Sym_34 = Sym_35; 
--# expr-done 9845
   If_Result_16 := Is_Equal_17;
end if;
if If_Result_16 then
   If_Result_17 := True;
else
   --# expr-start 9848 '<Eq at ast.py:15998>' Is_Equal_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9846 '<GetSymbol at ast.py:15998>' Sym_36 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_36 := Get_Symbol (Self); 
--# expr-done 9846
--# expr-start 9847 '<Symbol Last at ???>' Sym_37 None
Sym_37 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Last); 
--# expr-done 9847
Is_Equal_18 := Sym_36 = Sym_37; 
--# expr-done 9848
   If_Result_17 := Is_Equal_18;
end if;
if If_Result_17 then
   If_Result_18 := True;
else
   --# expr-start 9851 '<Eq at ast.py:15998>' Is_Equal_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9849 '<GetSymbol at ast.py:15998>' Sym_38 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_38 := Get_Symbol (Self); 
--# expr-done 9849
--# expr-start 9850 '<Symbol Range at ???>' Sym_39 None
Sym_39 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Range); 
--# expr-done 9850
Is_Equal_19 := Sym_38 = Sym_39; 
--# expr-done 9851
   If_Result_18 := Is_Equal_19;
end if;
if If_Result_18 then
   If_Result_19 := True;
else
   --# expr-start 9854 '<Eq at ast.py:15998>' Is_Equal_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9852 '<GetSymbol at ast.py:15998>' Sym_40 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_40 := Get_Symbol (Self); 
--# expr-done 9852
--# expr-start 9853 '<Symbol Length at ???>' Sym_41 None
Sym_41 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Length); 
--# expr-done 9853
Is_Equal_20 := Sym_40 = Sym_41; 
--# expr-done 9854
   If_Result_19 := Is_Equal_20;
end if;
if If_Result_19 then
   If_Result_20 := True;
else
   --# expr-start 9857 '<Eq at ast.py:15998>' Is_Equal_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9855 '<GetSymbol at ast.py:15998>' Sym_42 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_42 := Get_Symbol (Self); 
--# expr-done 9855
--# expr-start 9856 '<Symbol Image at ???>' Sym_43 None
Sym_43 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Image); 
--# expr-done 9856
Is_Equal_21 := Sym_42 = Sym_43; 
--# expr-done 9857
   If_Result_20 := Is_Equal_21;
end if;
if If_Result_20 then
   If_Result_21 := True;
else
   --# expr-start 9860 '<Eq at ast.py:15998>' Is_Equal_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9858 '<GetSymbol at ast.py:15998>' Sym_44 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_44 := Get_Symbol (Self); 
--# expr-done 9858
--# expr-start 9859 '<Symbol Wide_Image at ???>' Sym_45 None
Sym_45 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Wide_Image); 
--# expr-done 9859
Is_Equal_22 := Sym_44 = Sym_45; 
--# expr-done 9860
   If_Result_21 := Is_Equal_22;
end if;
if If_Result_21 then
   If_Result_22 := True;
else
   --# expr-start 9863 '<Eq at ast.py:15998>' Is_Equal_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9861 '<GetSymbol at ast.py:15998>' Sym_46 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_46 := Get_Symbol (Self); 
--# expr-done 9861
--# expr-start 9862 '<Symbol Wide_Wide_Image at ???>' Sym_47 None
Sym_47 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Wide_Wide_Image); 
--# expr-done 9862
Is_Equal_23 := Sym_46 = Sym_47; 
--# expr-done 9863
   If_Result_22 := Is_Equal_23;
end if;
if If_Result_22 then
   If_Result_23 := True;
else
   --# expr-start 9866 '<Eq at ast.py:15998>' Is_Equal_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9864 '<GetSymbol at ast.py:15998>' Sym_48 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_48 := Get_Symbol (Self); 
--# expr-done 9864
--# expr-start 9865 '<Symbol Put_Image at ???>' Sym_49 None
Sym_49 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Put_Image); 
--# expr-done 9865
Is_Equal_24 := Sym_48 = Sym_49; 
--# expr-done 9866
   If_Result_23 := Is_Equal_24;
end if;
if If_Result_23 then
   If_Result_24 := True;
else
   --# expr-start 9869 '<Eq at ast.py:15998>' Is_Equal_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9867 '<GetSymbol at ast.py:15998>' Sym_50 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_50 := Get_Symbol (Self); 
--# expr-done 9867
--# expr-start 9868 '<Symbol Asm_Input at ???>' Sym_51 None
Sym_51 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Asm_Input); 
--# expr-done 9868
Is_Equal_25 := Sym_50 = Sym_51; 
--# expr-done 9869
   If_Result_24 := Is_Equal_25;
end if;
if If_Result_24 then
   If_Result_25 := True;
else
   --# expr-start 9872 '<Eq at ast.py:15998>' Is_Equal_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9870 '<GetSymbol at ast.py:15998>' Sym_52 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_52 := Get_Symbol (Self); 
--# expr-done 9870
--# expr-start 9871 '<Symbol Asm_Output at ???>' Sym_53 None
Sym_53 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Asm_Output); 
--# expr-done 9871
Is_Equal_26 := Sym_52 = Sym_53; 
--# expr-done 9872
   If_Result_25 := Is_Equal_26;
end if;
if If_Result_25 then
   If_Result_26 := True;
else
   --# expr-start 9875 '<Eq at ast.py:15998>' Is_Equal_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9873 '<GetSymbol at ast.py:15998>' Sym_54 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_54 := Get_Symbol (Self); 
--# expr-done 9873
--# expr-start 9874 '<Symbol To_Address at ???>' Sym_55 None
Sym_55 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_To_Address); 
--# expr-done 9874
Is_Equal_27 := Sym_54 = Sym_55; 
--# expr-done 9875
   If_Result_26 := Is_Equal_27;
end if;
if If_Result_26 then
   If_Result_27 := True;
else
   --# expr-start 9878 '<Eq at ast.py:15998>' Is_Equal_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9876 '<GetSymbol at ast.py:15998>' Sym_56 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_56 := Get_Symbol (Self); 
--# expr-done 9876
--# expr-start 9877 '<Symbol Rounding at ???>' Sym_57 None
Sym_57 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Rounding); 
--# expr-done 9877
Is_Equal_28 := Sym_56 = Sym_57; 
--# expr-done 9878
   If_Result_27 := Is_Equal_28;
end if;
if If_Result_27 then
   If_Result_28 := True;
else
   --# expr-start 9881 '<Eq at ast.py:15998>' Is_Equal_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9879 '<GetSymbol at ast.py:15998>' Sym_58 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_58 := Get_Symbol (Self); 
--# expr-done 9879
--# expr-start 9880 '<Symbol Round at ???>' Sym_59 None
Sym_59 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Round); 
--# expr-done 9880
Is_Equal_29 := Sym_58 = Sym_59; 
--# expr-done 9881
   If_Result_28 := Is_Equal_29;
end if;
if If_Result_28 then
   If_Result_29 := True;
else
   --# expr-start 9884 '<Eq at ast.py:15998>' Is_Equal_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9882 '<GetSymbol at ast.py:15998>' Sym_60 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_60 := Get_Symbol (Self); 
--# expr-done 9882
--# expr-start 9883 '<Symbol Ceiling at ???>' Sym_61 None
Sym_61 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Ceiling); 
--# expr-done 9883
Is_Equal_30 := Sym_60 = Sym_61; 
--# expr-done 9884
   If_Result_29 := Is_Equal_30;
end if;
if If_Result_29 then
   If_Result_30 := True;
else
   --# expr-start 9887 '<Eq at ast.py:15998>' Is_Equal_31 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9885 '<GetSymbol at ast.py:15998>' Sym_62 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_62 := Get_Symbol (Self); 
--# expr-done 9885
--# expr-start 9886 '<Symbol Floor at ???>' Sym_63 None
Sym_63 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Floor); 
--# expr-done 9886
Is_Equal_31 := Sym_62 = Sym_63; 
--# expr-done 9887
   If_Result_30 := Is_Equal_31;
end if;
if If_Result_30 then
   If_Result_31 := True;
else
   --# expr-start 9890 '<Eq at ast.py:15998>' Is_Equal_32 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9888 '<GetSymbol at ast.py:15998>' Sym_64 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_64 := Get_Symbol (Self); 
--# expr-done 9888
--# expr-start 9889 '<Symbol Truncation at ???>' Sym_65 None
Sym_65 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Truncation); 
--# expr-done 9889
Is_Equal_32 := Sym_64 = Sym_65; 
--# expr-done 9890
   If_Result_31 := Is_Equal_32;
end if;
if If_Result_31 then
   If_Result_32 := True;
else
   --# expr-start 9893 '<Eq at ast.py:15998>' Is_Equal_33 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9891 '<GetSymbol at ast.py:15998>' Sym_66 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_66 := Get_Symbol (Self); 
--# expr-done 9891
--# expr-start 9892 '<Symbol Copy_Sign at ???>' Sym_67 None
Sym_67 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Copy_Sign); 
--# expr-done 9892
Is_Equal_33 := Sym_66 = Sym_67; 
--# expr-done 9893
   If_Result_32 := Is_Equal_33;
end if;
if If_Result_32 then
   If_Result_33 := True;
else
   --# expr-start 9896 '<Eq at ast.py:15998>' Is_Equal_34 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9894 '<GetSymbol at ast.py:15998>' Sym_68 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_68 := Get_Symbol (Self); 
--# expr-done 9894
--# expr-start 9895 '<Symbol Remainder at ???>' Sym_69 None
Sym_69 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Remainder); 
--# expr-done 9895
Is_Equal_34 := Sym_68 = Sym_69; 
--# expr-done 9896
   If_Result_33 := Is_Equal_34;
end if;
if If_Result_33 then
   If_Result_34 := True;
else
   --# expr-start 9899 '<Eq at ast.py:15998>' Is_Equal_35 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9897 '<GetSymbol at ast.py:15998>' Sym_70 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_70 := Get_Symbol (Self); 
--# expr-done 9897
--# expr-start 9898 '<Symbol Adjacent at ???>' Sym_71 None
Sym_71 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Adjacent); 
--# expr-done 9898
Is_Equal_35 := Sym_70 = Sym_71; 
--# expr-done 9899
   If_Result_34 := Is_Equal_35;
end if;
if If_Result_34 then
   If_Result_35 := True;
else
   --# expr-start 9902 '<Eq at ast.py:15998>' Is_Equal_36 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9900 '<GetSymbol at ast.py:15998>' Sym_72 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_72 := Get_Symbol (Self); 
--# expr-done 9900
--# expr-start 9901 '<Symbol Mod at ???>' Sym_73 None
Sym_73 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Mod_141); 
--# expr-done 9901
Is_Equal_36 := Sym_72 = Sym_73; 
--# expr-done 9902
   If_Result_35 := Is_Equal_36;
end if;
if If_Result_35 then
   If_Result_36 := True;
else
   --# expr-start 9905 '<Eq at ast.py:15998>' Is_Equal_37 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9903 '<GetSymbol at ast.py:15998>' Sym_74 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_74 := Get_Symbol (Self); 
--# expr-done 9903
--# expr-start 9904 '<Symbol Machine at ???>' Sym_75 None
Sym_75 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Machine); 
--# expr-done 9904
Is_Equal_37 := Sym_74 = Sym_75; 
--# expr-done 9905
   If_Result_36 := Is_Equal_37;
end if;
if If_Result_36 then
   If_Result_37 := True;
else
   --# expr-start 9908 '<Eq at ast.py:15998>' Is_Equal_38 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9906 '<GetSymbol at ast.py:15998>' Sym_76 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_76 := Get_Symbol (Self); 
--# expr-done 9906
--# expr-start 9907 '<Symbol Machine_Rounding at ???>' Sym_77 None
Sym_77 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Machine_Rounding); 
--# expr-done 9907
Is_Equal_38 := Sym_76 = Sym_77; 
--# expr-done 9908
   If_Result_37 := Is_Equal_38;
end if;
if If_Result_37 then
   If_Result_38 := True;
else
   --# expr-start 9911 '<Eq at ast.py:15998>' Is_Equal_39 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9909 '<GetSymbol at ast.py:15998>' Sym_78 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_78 := Get_Symbol (Self); 
--# expr-done 9909
--# expr-start 9910 '<Symbol Exponent at ???>' Sym_79 None
Sym_79 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Exponent); 
--# expr-done 9910
Is_Equal_39 := Sym_78 = Sym_79; 
--# expr-done 9911
   If_Result_38 := Is_Equal_39;
end if;
if If_Result_38 then
   If_Result_39 := True;
else
   --# expr-start 9914 '<Eq at ast.py:15998>' Is_Equal_40 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9912 '<GetSymbol at ast.py:15998>' Sym_80 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_80 := Get_Symbol (Self); 
--# expr-done 9912
--# expr-start 9913 '<Symbol Scaling at ???>' Sym_81 None
Sym_81 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Scaling); 
--# expr-done 9913
Is_Equal_40 := Sym_80 = Sym_81; 
--# expr-done 9914
   If_Result_39 := Is_Equal_40;
end if;
if If_Result_39 then
   If_Result_40 := True;
else
   --# expr-start 9917 '<Eq at ast.py:15998>' Is_Equal_41 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9915 '<GetSymbol at ast.py:15998>' Sym_82 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_82 := Get_Symbol (Self); 
--# expr-done 9915
--# expr-start 9916 '<Symbol Compose at ???>' Sym_83 None
Sym_83 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Compose); 
--# expr-done 9916
Is_Equal_41 := Sym_82 = Sym_83; 
--# expr-done 9917
   If_Result_40 := Is_Equal_41;
end if;
if If_Result_40 then
   If_Result_41 := True;
else
   --# expr-start 9920 '<Eq at ast.py:15998>' Is_Equal_42 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
--# expr-start 9918 '<GetSymbol at ast.py:15998>' Sym_84 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15998
Sym_84 := Get_Symbol (Self); 
--# expr-done 9918
--# expr-start 9919 '<Symbol Fraction at ???>' Sym_85 None
Sym_85 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Fraction); 
--# expr-done 9919
Is_Equal_42 := Sym_84 = Sym_85; 
--# expr-done 9920
   If_Result_41 := Is_Equal_42;
end if;
      Property_Result := If_Result_41;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Identifier_P_Is_Attr_With_Args;
--# end
--# property-start Identifier.complete_items /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16019
pragma Warnings (Off, "is not referenced");
 function Identifier_P_Complete_Items
  (Node : Bare_Identifier
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Completion_Item_Array_Access
is
   Self : Bare_Identifier := Bare_Identifier (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Identifier :=
     Internal_Entity_Identifier'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Completion_Item_Array_Access;
            procedure Finalizer_Scope_1133 with Inline_Always;
      Fld : Internal_Entity;
Fld_1 : Internal_Completion_Item_Array_Access;
Let_Result : Internal_Completion_Item_Array_Access;
            procedure Finalizer_Scope_1133 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1133;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9923 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9922 '<FieldAccess for complete_items at ast.py:16020>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16020
--# expr-start 9921 '<parent at ast.py:16020>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16020
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 9921
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.complete_items'
Fld_1 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Complete_Items (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 9922
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 9923
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1133;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1133;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Identifier_P_Complete_Items;
--# end
--# property-start Identifier.is_constant /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16023
pragma Warnings (Off, "is not referenced");
 function Identifier_P_Is_Constant
  (Node : Bare_Identifier
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Identifier := Bare_Identifier (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Identifier :=
     Internal_Entity_Identifier'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
            procedure Finalizer_Scope_1134 with Inline_Always;
            procedure Finalizer_Scope_3003 with Inline_Always;
      Block_Var : Internal_Entity_Basic_Decl;
Var_Expr : Internal_Entity;
Var_Expr_1 : Internal_Entity_Base_Subp_Body;
Fld : Internal_Entity_Basic_Decl;
Fld_1 : Boolean;
Fld_2 : Boolean;
Not_Val : Boolean;
Is_A : Boolean;
If_Result : Boolean;
If_Result_1 : Boolean;
Fld_3 : Internal_Entity_Array_Access;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Base_Subp_Body;
Fld_4 : Internal_Entity_Subp_Spec;
Fld_5 : Internal_Entity_Subp_Kind;
Scope_Result : Internal_Entity_Subp_Kind;
Result_Var : Internal_Entity_Subp_Kind;
Is_A_1 : Boolean;
Find_Result : Internal_Entity;
Fld_6 : Boolean;
Fld_7 : Internal_Entity_Array_Access;
Is_A_2 : Boolean;
Find_Result_1 : Internal_Entity;
Is_Null : Boolean;
Not_Val_1 : Boolean;
If_Result_2 : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
If_Result_3 : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_1134 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_1134;
            procedure Finalizer_Scope_3003 is
            begin
                     Dec_Ref (Fld_7);
            end Finalizer_Scope_3003;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9942 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9924 '<FieldAccess for referenced_decl at ast.py:16024>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16024
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Ent.Node, Imprecise_Fallback => False, E_Info => Ent.Info);
--# end
--# expr-done 9924
Block_Var := Fld; 
--# bind rd Block_Var
--# expr-start 9929 '<If at ast.py:16029>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16029
--# expr-start 9926 '<Not at ast.py:16030>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16030
--# expr-start 9925 '<FieldAccess for is_subprogram at ast.py:16030>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16030
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_subprogram
Fld_2 := Libadalang.Implementation.Basic_Decl_P_Is_Subprogram (Node => Block_Var.Node);
--# end
--# expr-done 9925
Not_Val := not (Fld_2); 
--# expr-done 9926
if Not_Val then
   If_Result := True;
else
   --# expr-start 9927 '<IsA EnumLiteralDecl at ast.py:16030>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16030
Is_A := Block_Var.Node /= null 
and then Block_Var.Node.Kind in Ada_Enum_Literal_Decl_Range; 
--# expr-done 9927
   If_Result := Is_A;
end if;
if If_Result then
   --# expr-start 9928 '<FieldAccess for is_constant_object at ast.py:16031>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16031
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_constant_object'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Constant_Object (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 9928
   If_Result_1 := Fld_1;
else
   If_Result_1 := False;
end if;
--# expr-done 9929
if If_Result_1 then
   If_Result_3 := True;
else
--# expr-start 9935 '<Find at ast.py:16037>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16037
   --# expr-start 9930 '<FieldAccess for parents at ast.py:16037>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16037
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Parents (Node => Ent.Node, With_Self => True, E_Info => Ent.Info);
--# expr-done 9930
      Find_Result := No_Entity;
      declare
         Collection : constant Internal_Entity_Array_Access := Fld_3;
      begin
         for Item_304 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_304
      --# expr-start 9934 '<IsA SubpKind.Function at ast.py:16042>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16042
--# expr-start 9931 '<Cast to BaseSubpBody at ast.py:16042>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16042
Cast_Expr := Item_304; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Base_Subp_Body
   then
      Cast_Result := Create_Internal_Entity_Base_Subp_Body
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Base_Subp_Body;
   end if;
--# expr-done 9931
Var_Expr_1 := Cast_Result;
if Var_Expr_1 /= No_Entity_Base_Subp_Body then
   --# scope-start
--# expr-start 9933 '<FieldAccess for subp_kind at ast.py:16042>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16042
--# expr-start 9932 '<FieldAccess for subp_spec at ast.py:16042>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16042
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Subp_Spec (Node => Var_Expr_1.Node.Base_Subp_Body_F_Subp_Spec, Info => Var_Expr_1.Info);
--# expr-done 9932
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Subp_Kind (Node => Fld_4.Node.Subp_Spec_F_Subp_Kind, Info => Fld_4.Info);
--# expr-done 9933
Scope_Result := Fld_5;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Subp_Kind;
end if;
Is_A_1 := Result_Var.Node /= null 
and then Result_Var.Node.Kind in Ada_Subp_Kind_Function_Range; 
--# expr-done 9934
      if Is_A_1 then
         Find_Result := Item_304;
      end if;
   --# end
               exit when Is_A_1;
         end loop;
      end;
--# expr-done 9935
Var_Expr := Find_Result;
if Var_Expr /= No_Entity then
   --# scope-start
   --# bind _ Var_Expr
--# expr-start 9936 '<FieldAccess for is_in_private_part at ast.py:16046>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16046
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_in_private_part'
Fld_6 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_In_Private_Part (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 9936
if Fld_6 then
   --# expr-start 9941 '<Not at ast.py:16047>' Not_Val_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16047
--# expr-start 9940 '<IsNull at ast.py:16048>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16048
--# expr-start 9939 '<Find at ast.py:16048>' Find_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16048
   --# expr-start 9937 '<FieldAccess for parents at ast.py:16048>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16048
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Parents (Node => Block_Var.Node, With_Self => True, E_Info => Block_Var.Info);
--# expr-done 9937
      Find_Result_1 := No_Entity;
      declare
         Collection : constant Internal_Entity_Array_Access := Fld_7;
      begin
         for Item_305 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_305
      --# expr-start 9938 '<IsA ProtectedTypeDecl at ast.py:16049>' Is_A_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16049
Is_A_2 := Item_305.Node /= null 
and then Item_305.Node.Kind in Ada_Protected_Type_Decl_Range; 
--# expr-done 9938
      if Is_A_2 then
         Find_Result_1 := Item_305;
      end if;
   --# end
               exit when Is_A_2;
         end loop;
      end;
--# expr-done 9939
Is_Null := Find_Result_1.Node = null; 
--# expr-done 9940
Not_Val_1 := not (Is_Null); 
--# expr-done 9941
   If_Result_2 := Not_Val_1;
else
   If_Result_2 := False;
end if;
Scope_Result_1 := If_Result_2;
   --# end
      Finalizer_Scope_3003;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
   If_Result_3 := Result_Var_1;
end if;
Let_Result := If_Result_3; 
--# end
--# expr-done 9942
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1134;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1134;
               Finalizer_Scope_3003;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Identifier_P_Is_Constant;
--# end
   --
   --  Primitives for Bare_Op
   --
      procedure Initialize_Fields_For_Op
        (Self : Bare_Op
        ) is
      begin
            Initialize_Fields_For_Base_Id
              (Self);
      end Initialize_Fields_For_Op;
--# property-start Op.subprogram_symbol /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15903
pragma Warnings (Off, "is not referenced");
 function Op_P_Subprogram_Symbol
  (Node : Bare_Op
  )
   return Symbol_Type
is
   Self : Bare_Op := Bare_Op (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type;
      Match_0 : Bare_Op_And;
Match_1 : Bare_Op_Or;
Match_2 : Bare_Op_Xor;
Match_3 : Bare_Op_Abs;
Match_4 : Bare_Op_Not;
Match_5 : Bare_Op_Pow;
Match_6 : Bare_Op_Mult;
Match_7 : Bare_Op_Div;
Match_8 : Bare_Op_Mod;
Match_9 : Bare_Op_Rem;
Match_10 : Bare_Op_Plus;
Match_11 : Bare_Op_Minus;
Match_12 : Bare_Op_Concat;
Match_13 : Bare_Op_Eq;
Match_14 : Bare_Op_Neq;
Match_15 : Bare_Op_Lt;
Match_16 : Bare_Op_Lte;
Match_17 : Bare_Op_Gt;
Match_18 : Bare_Op_Gte;
Match_19 : Bare_Op;
Sym : Symbol_Type;
Sym_1 : Symbol_Type;
Sym_2 : Symbol_Type;
Sym_3 : Symbol_Type;
Sym_4 : Symbol_Type;
Sym_5 : Symbol_Type;
Sym_6 : Symbol_Type;
Sym_7 : Symbol_Type;
Sym_8 : Symbol_Type;
Sym_9 : Symbol_Type;
Sym_10 : Symbol_Type;
Sym_11 : Symbol_Type;
Sym_12 : Symbol_Type;
Sym_13 : Symbol_Type;
Sym_14 : Symbol_Type;
Sym_15 : Symbol_Type;
Sym_16 : Symbol_Type;
Sym_17 : Symbol_Type;
Sym_18 : Symbol_Type;
Sym_19 : Symbol_Type;
Match_Prefix : Bare_Op;
Cast_Result : Bare_Op_And;
Let_Result : Symbol_Type;
Scope_Result : Symbol_Type;
Cast_Result_1 : Bare_Op_Or;
Let_Result_1 : Symbol_Type;
Scope_Result_1 : Symbol_Type;
Cast_Result_2 : Bare_Op_Xor;
Let_Result_2 : Symbol_Type;
Scope_Result_2 : Symbol_Type;
Cast_Result_3 : Bare_Op_Abs;
Let_Result_3 : Symbol_Type;
Scope_Result_3 : Symbol_Type;
Cast_Result_4 : Bare_Op_Not;
Let_Result_4 : Symbol_Type;
Scope_Result_4 : Symbol_Type;
Cast_Result_5 : Bare_Op_Pow;
Let_Result_5 : Symbol_Type;
Scope_Result_5 : Symbol_Type;
Cast_Result_6 : Bare_Op_Mult;
Let_Result_6 : Symbol_Type;
Scope_Result_6 : Symbol_Type;
Cast_Result_7 : Bare_Op_Div;
Let_Result_7 : Symbol_Type;
Scope_Result_7 : Symbol_Type;
Cast_Result_8 : Bare_Op_Mod;
Let_Result_8 : Symbol_Type;
Scope_Result_8 : Symbol_Type;
Cast_Result_9 : Bare_Op_Rem;
Let_Result_9 : Symbol_Type;
Scope_Result_9 : Symbol_Type;
Cast_Result_10 : Bare_Op_Plus;
Let_Result_10 : Symbol_Type;
Scope_Result_10 : Symbol_Type;
Cast_Result_11 : Bare_Op_Minus;
Let_Result_11 : Symbol_Type;
Scope_Result_11 : Symbol_Type;
Cast_Result_12 : Bare_Op_Concat;
Let_Result_12 : Symbol_Type;
Scope_Result_12 : Symbol_Type;
Cast_Result_13 : Bare_Op_Eq;
Let_Result_13 : Symbol_Type;
Scope_Result_13 : Symbol_Type;
Cast_Result_14 : Bare_Op_Neq;
Let_Result_14 : Symbol_Type;
Scope_Result_14 : Symbol_Type;
Cast_Result_15 : Bare_Op_Lt;
Let_Result_15 : Symbol_Type;
Scope_Result_15 : Symbol_Type;
Cast_Result_16 : Bare_Op_Lte;
Let_Result_16 : Symbol_Type;
Scope_Result_16 : Symbol_Type;
Cast_Result_17 : Bare_Op_Gt;
Let_Result_17 : Symbol_Type;
Scope_Result_17 : Symbol_Type;
Cast_Result_18 : Bare_Op_Gte;
Let_Result_18 : Symbol_Type;
Scope_Result_18 : Symbol_Type;
Cast_Result_19 : Bare_Op;
Let_Result_19 : Symbol_Type;
Scope_Result_19 : Symbol_Type;
Match_Result : Symbol_Type;
Let_Result_20 : Symbol_Type;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9964 '<Block at ???>' Let_Result_20 None
--# scope-start
--# expr-start 9963 '<Match at ast.py:15908>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15908
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Self; 
case Ada_Op (Match_Prefix.Kind) is
         when Ada_Op_And =>
   --# scope-start
--# scope-start
      Cast_Result := Match_Prefix;
Match_0 := Cast_Result; 
--# bind _ Match_0
--# expr-start 9943 '<Symbol "and" at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_And); 
--# expr-done 9943
Let_Result := Sym; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Op_Or =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Match_Prefix;
Match_1 := Cast_Result_1; 
--# bind _ Match_1
--# expr-start 9944 '<Symbol "or" at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Or); 
--# expr-done 9944
Let_Result_1 := Sym_1; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Op_Xor =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Match_Prefix;
Match_2 := Cast_Result_2; 
--# bind _ Match_2
--# expr-start 9945 '<Symbol "xor" at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Xor); 
--# expr-done 9945
Let_Result_2 := Sym_2; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Op_Abs =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Match_Prefix;
Match_3 := Cast_Result_3; 
--# bind _ Match_3
--# expr-start 9946 '<Symbol "abs" at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Abs); 
--# expr-done 9946
Let_Result_3 := Sym_3; 
--# end
Scope_Result_3 := Let_Result_3;
   --# end
            Match_Result := Scope_Result_3; 
         when Ada_Op_Not =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Match_Prefix;
Match_4 := Cast_Result_4; 
--# bind _ Match_4
--# expr-start 9947 '<Symbol "not" at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Not); 
--# expr-done 9947
Let_Result_4 := Sym_4; 
--# end
Scope_Result_4 := Let_Result_4;
   --# end
            Match_Result := Scope_Result_4; 
         when Ada_Op_Pow =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Match_Prefix;
Match_5 := Cast_Result_5; 
--# bind _ Match_5
--# expr-start 9948 '<Symbol "**" at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_3); 
--# expr-done 9948
Let_Result_5 := Sym_5; 
--# end
Scope_Result_5 := Let_Result_5;
   --# end
            Match_Result := Scope_Result_5; 
         when Ada_Op_Mult =>
   --# scope-start
--# scope-start
      Cast_Result_6 := Match_Prefix;
Match_6 := Cast_Result_6; 
--# bind _ Match_6
--# expr-start 9949 '<Symbol "*" at ???>' Sym_6 None
Sym_6 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_2); 
--# expr-done 9949
Let_Result_6 := Sym_6; 
--# end
Scope_Result_6 := Let_Result_6;
   --# end
            Match_Result := Scope_Result_6; 
         when Ada_Op_Div =>
   --# scope-start
--# scope-start
      Cast_Result_7 := Match_Prefix;
Match_7 := Cast_Result_7; 
--# bind _ Match_7
--# expr-start 9950 '<Symbol "/" at ???>' Sym_7 None
Sym_7 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_6); 
--# expr-done 9950
Let_Result_7 := Sym_7; 
--# end
Scope_Result_7 := Let_Result_7;
   --# end
            Match_Result := Scope_Result_7; 
         when Ada_Op_Mod =>
   --# scope-start
--# scope-start
      Cast_Result_8 := Match_Prefix;
Match_8 := Cast_Result_8; 
--# bind _ Match_8
--# expr-start 9951 '<Symbol "mod" at ???>' Sym_8 None
Sym_8 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Mod); 
--# expr-done 9951
Let_Result_8 := Sym_8; 
--# end
Scope_Result_8 := Let_Result_8;
   --# end
            Match_Result := Scope_Result_8; 
         when Ada_Op_Rem =>
   --# scope-start
--# scope-start
      Cast_Result_9 := Match_Prefix;
Match_9 := Cast_Result_9; 
--# bind _ Match_9
--# expr-start 9952 '<Symbol "rem" at ???>' Sym_9 None
Sym_9 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Rem); 
--# expr-done 9952
Let_Result_9 := Sym_9; 
--# end
Scope_Result_9 := Let_Result_9;
   --# end
            Match_Result := Scope_Result_9; 
         when Ada_Op_Plus =>
   --# scope-start
--# scope-start
      Cast_Result_10 := Match_Prefix;
Match_10 := Cast_Result_10; 
--# bind _ Match_10
--# expr-start 9953 '<Symbol "+" at ???>' Sym_10 None
Sym_10 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_4); 
--# expr-done 9953
Let_Result_10 := Sym_10; 
--# end
Scope_Result_10 := Let_Result_10;
   --# end
            Match_Result := Scope_Result_10; 
         when Ada_Op_Minus =>
   --# scope-start
--# scope-start
      Cast_Result_11 := Match_Prefix;
Match_11 := Cast_Result_11; 
--# bind _ Match_11
--# expr-start 9954 '<Symbol "-" at ???>' Sym_11 None
Sym_11 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_5); 
--# expr-done 9954
Let_Result_11 := Sym_11; 
--# end
Scope_Result_11 := Let_Result_11;
   --# end
            Match_Result := Scope_Result_11; 
         when Ada_Op_Concat =>
   --# scope-start
--# scope-start
      Cast_Result_12 := Match_Prefix;
Match_12 := Cast_Result_12; 
--# bind _ Match_12
--# expr-start 9955 '<Symbol "&" at ???>' Sym_12 None
Sym_12 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym); 
--# expr-done 9955
Let_Result_12 := Sym_12; 
--# end
Scope_Result_12 := Let_Result_12;
   --# end
            Match_Result := Scope_Result_12; 
         when Ada_Op_Eq =>
   --# scope-start
--# scope-start
      Cast_Result_13 := Match_Prefix;
Match_13 := Cast_Result_13; 
--# bind _ Match_13
--# expr-start 9956 '<Symbol "=" at ???>' Sym_13 None
Sym_13 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_10); 
--# expr-done 9956
Let_Result_13 := Sym_13; 
--# end
Scope_Result_13 := Let_Result_13;
   --# end
            Match_Result := Scope_Result_13; 
         when Ada_Op_Neq =>
   --# scope-start
--# scope-start
      Cast_Result_14 := Match_Prefix;
Match_14 := Cast_Result_14; 
--# bind _ Match_14
--# expr-start 9957 '<Symbol "/=" at ???>' Sym_14 None
Sym_14 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_7); 
--# expr-done 9957
Let_Result_14 := Sym_14; 
--# end
Scope_Result_14 := Let_Result_14;
   --# end
            Match_Result := Scope_Result_14; 
         when Ada_Op_Lt =>
   --# scope-start
--# scope-start
      Cast_Result_15 := Match_Prefix;
Match_15 := Cast_Result_15; 
--# bind _ Match_15
--# expr-start 9958 '<Symbol "<" at ???>' Sym_15 None
Sym_15 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_8); 
--# expr-done 9958
Let_Result_15 := Sym_15; 
--# end
Scope_Result_15 := Let_Result_15;
   --# end
            Match_Result := Scope_Result_15; 
         when Ada_Op_Lte =>
   --# scope-start
--# scope-start
      Cast_Result_16 := Match_Prefix;
Match_16 := Cast_Result_16; 
--# bind _ Match_16
--# expr-start 9959 '<Symbol "<=" at ???>' Sym_16 None
Sym_16 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_9); 
--# expr-done 9959
Let_Result_16 := Sym_16; 
--# end
Scope_Result_16 := Let_Result_16;
   --# end
            Match_Result := Scope_Result_16; 
         when Ada_Op_Gt =>
   --# scope-start
--# scope-start
      Cast_Result_17 := Match_Prefix;
Match_17 := Cast_Result_17; 
--# bind _ Match_17
--# expr-start 9960 '<Symbol ">" at ???>' Sym_17 None
Sym_17 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_11); 
--# expr-done 9960
Let_Result_17 := Sym_17; 
--# end
Scope_Result_17 := Let_Result_17;
   --# end
            Match_Result := Scope_Result_17; 
         when Ada_Op_Gte =>
   --# scope-start
--# scope-start
      Cast_Result_18 := Match_Prefix;
Match_18 := Cast_Result_18; 
--# bind _ Match_18
--# expr-start 9961 '<Symbol ">=" at ???>' Sym_18 None
Sym_18 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_12); 
--# expr-done 9961
Let_Result_18 := Sym_18; 
--# end
Scope_Result_18 := Let_Result_18;
   --# end
            Match_Result := Scope_Result_18; 
         when Ada_Op_And_Then | Ada_Op_Double_Dot | Ada_Op_In | Ada_Op_Not_In | Ada_Op_Or_Else =>
   --# scope-start
--# scope-start
      Cast_Result_19 := Match_Prefix;
Match_19 := Cast_Result_19; 
--# bind _ Match_19
--# expr-start 9962 '<Symbol <<>> at ???>' Sym_19 None
Sym_19 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_20); 
--# expr-done 9962
Let_Result_19 := Sym_19; 
--# end
Scope_Result_19 := Let_Result_19;
   --# end
            Match_Result := Scope_Result_19; 
end case;
--# expr-done 9963
Let_Result_20 := Match_Result; 
--# end
--# expr-done 9964
      Property_Result := Let_Result_20;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Op_P_Subprogram_Symbol;
--# end
--# property-start Op.subprograms_for_symbol /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15932
pragma Warnings (Off, "is not referenced");
 function Op_P_Subprograms_For_Symbol
  (Node : Bare_Op
      ; Sym : Symbol_Type
      ; From_Node : Internal_Entity
  )
   return Internal_Entity_Basic_Decl_Array_Access
is
   Self : Bare_Op := Bare_Op (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind sym Sym
   --# bind from_node From_Node
   Property_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_1127 with Inline_Always;
      Fld : Internal_Entity_Basic_Decl_Array_Access;
Fld_1 : Lexical_Env := Empty_Env;
Fld_2 : Bare_Ada_Node;
Fld_3 : Internal_Entity_Array_Access;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Cast_Expr_1 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Basic_Decl;
Fld_4 : Boolean;
Is_A : Boolean;
Not_Val : Boolean;
If_Result : Boolean;
Map_Result : Internal_Entity_Basic_Decl_Array_Access;
Concat_Result : Internal_Entity_Basic_Decl_Array_Access;
Let_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_1127 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1127;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9976 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9975 '<Concat at ast.py:15939>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15939
--# expr-start 9965 '<FieldAccess for root_type_ops at ast.py:15939>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15939
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.root_type_ops
Fld := Libadalang.Implementation.Ada_Node_P_Root_Type_Ops (Node => Self, Sym => Sym);
--# end
--# expr-done 9965
--# expr-start 9974 '<FilterMap at ast.py:15939>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15939
   --# expr-start 9968 '<FieldAccess for env_get at ast.py:15939>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15939
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9966 '<FieldAccess for node_env at ast.py:15940>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15940
      if From_Node.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Node_Env (Node => From_Node.Node, E_Info => From_Node.Info);
--# expr-done 9966
--# expr-start 9967 '<FieldAccess for node at ast.py:15940>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15940
Fld_2 := From_Node.Node;
--# expr-done 9967
--# property-call-start AdaNode.env_get
Fld_3 := Libadalang.Implementation.Ada_Node_P_Env_Get (Node => Self, Env => Fld_1, Symbol => Sym, From_Node => Fld_2);
--# end
--# expr-done 9968
      declare
         Map_Result_Vec : Internal_Entity_Basic_Decl_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Fld_3;
      begin
         for Item_306 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_306
--# expr-start 9970 '<FieldAccess for is_subprogram at ast.py:15944>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15944
--# expr-start 9969 '<Cast to BasicDecl at ast.py:15944>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15944
Cast_Expr_1 := Item_306; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 9969
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_subprogram
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Is_Subprogram (Node => Cast_Result_1.Node);
--# end
--# expr-done 9970
if Fld_4 then
   --# expr-start 9972 '<Not at ast.py:15950>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15950
--# expr-start 9971 '<IsA SyntheticSubpDecl at ast.py:15950>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15950
Is_A := Item_306.Node /= null 
and then Item_306.Node.Kind in Ada_Synthetic_Subp_Decl_Range; 
--# expr-done 9971
Not_Val := not (Is_A); 
--# expr-done 9972
   If_Result := Not_Val;
else
   If_Result := False;
end if;
         if If_Result then
      --# expr-start 9973 '<Cast to BasicDecl at ast.py:15942>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15942
Cast_Expr := Item_306; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 9973
         declare
            Item_To_Append : constant Internal_Entity_Basic_Decl := Cast_Result;
         begin
            Internal_Entity_Basic_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Basic_Decl_Array
           (Items_Count => Natural (Internal_Entity_Basic_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Basic_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Basic_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Basic_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 9974
Concat_Result := Concat (Fld, Map_Result); 
--# expr-done 9975
Let_Result := Concat_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 9976
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1127;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1127;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Op_P_Subprograms_For_Symbol;
--# end
--# property-start Op.subprograms /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15954
pragma Warnings (Off, "is not referenced");
 function Op_P_Subprograms
  (Node : Bare_Op
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Basic_Decl_Array_Access
is
   Self : Bare_Op := Bare_Op (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Op :=
     Internal_Entity_Op'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_1128 with Inline_Always;
      Fld : Symbol_Type;
Cast_Expr : Internal_Entity_Op;
Cast_Result : Internal_Entity;
Fld_1 : Internal_Entity_Basic_Decl_Array_Access;
            procedure Finalizer_Scope_1128 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1128;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9978 '<FieldAccess for subprograms_for_symbol at ast.py:15955>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15955
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 9977 '<FieldAccess for subprogram_symbol at ast.py:15955>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15955
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Op.subprogram_symbol
Fld := Libadalang.Implementation.Op_P_Subprogram_Symbol (Node => Self);
--# end
--# expr-done 9977
Cast_Expr := Ent; 
      Cast_Result := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
--# property-call-start Op.subprograms_for_symbol
Fld_1 := Libadalang.Implementation.Op_P_Subprograms_For_Symbol (Node => Self, Sym => Fld, From_Node => Cast_Result);
--# end
--# expr-done 9978
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1128;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1128;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Op_P_Subprograms;
--# end
--# property-start Op.name_symbol /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15962
pragma Warnings (Off, "is not referenced");
 function Op_P_Name_Symbol
  (Node : Bare_Op
  )
   return Symbol_Type
is
   Self : Bare_Op := Bare_Op (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type;
      Fld : Symbol_Type;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9979 '<FieldAccess for subprogram_symbol at ast.py:15962>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15962
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Op.subprogram_symbol
Fld := Libadalang.Implementation.Op_P_Subprogram_Symbol (Node => Self);
--# end
--# expr-done 9979
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Op_P_Name_Symbol;
--# end
--# property-start Op.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15965
pragma Warnings (Off, "is not referenced");
 function Op_P_Xref_Equation
  (Node : Bare_Op
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Op := Bare_Op (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1130 with Inline_Always;
      False_Rel : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1130 is
            begin
                     Dec_Ref (False_Rel);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1130;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 9980 '<Block at ???>' Let_Result None
--# scope-start
False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:15968") else null)); 
Let_Result := False_Rel; Inc_Ref (Let_Result);
--# end
--# expr-done 9980
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1130;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1130;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Op_P_Xref_Equation;
--# end
--# property-start Op.is_dispatching_call /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15971
pragma Warnings (Off, "is not referenced");
 function Op_P_Is_Dispatching_Call
  (Node : Bare_Op
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Op := Bare_Op (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Op :=
     Internal_Entity_Op'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Fld : Internal_Entity;
Cast_Result : Internal_Entity_Expr;
Fld_1 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9984 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9983 '<FieldAccess for is_dispatching_call at ast.py:15974>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15974
--# expr-start 9982 '<Cast to Expr at ast.py:15974>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15974
--# expr-start 9981 '<parent at ast.py:15974>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:15974
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 9981
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Expr
   then
      Cast_Result := Create_Internal_Entity_Expr
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Expr;
   end if;
--# expr-done 9982
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.is_dispatching_call'
Fld_1 := Libadalang.Implementation.Dispatcher_Expr_P_Is_Dispatching_Call (Node => Cast_Result.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Cast_Result.Info);
--# end
--# expr-done 9983
Let_Result := Fld_1; 
--# end
--# expr-done 9984
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Op_P_Is_Dispatching_Call;
--# end
   --
   --  Primitives for Bare_Op_Abs
   --
      procedure Initialize_Fields_For_Op_Abs
        (Self : Bare_Op_Abs
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Abs;
   --
   --  Primitives for Bare_Op_And
   --
      procedure Initialize_Fields_For_Op_And
        (Self : Bare_Op_And
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_And;
   --
   --  Primitives for Bare_Op_And_Then
   --
      procedure Initialize_Fields_For_Op_And_Then
        (Self : Bare_Op_And_Then
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_And_Then;
   --
   --  Primitives for Bare_Op_Concat
   --
      procedure Initialize_Fields_For_Op_Concat
        (Self : Bare_Op_Concat
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Concat;
   --
   --  Primitives for Bare_Op_Div
   --
      procedure Initialize_Fields_For_Op_Div
        (Self : Bare_Op_Div
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Div;
   --
   --  Primitives for Bare_Op_Double_Dot
   --
      procedure Initialize_Fields_For_Op_Double_Dot
        (Self : Bare_Op_Double_Dot
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Double_Dot;
   --
   --  Primitives for Bare_Op_Eq
   --
      procedure Initialize_Fields_For_Op_Eq
        (Self : Bare_Op_Eq
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Eq;
   --
   --  Primitives for Bare_Op_Gt
   --
      procedure Initialize_Fields_For_Op_Gt
        (Self : Bare_Op_Gt
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Gt;
   --
   --  Primitives for Bare_Op_Gte
   --
      procedure Initialize_Fields_For_Op_Gte
        (Self : Bare_Op_Gte
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Gte;
   --
   --  Primitives for Bare_Op_In
   --
      procedure Initialize_Fields_For_Op_In
        (Self : Bare_Op_In
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_In;
   --
   --  Primitives for Bare_Op_Lt
   --
      procedure Initialize_Fields_For_Op_Lt
        (Self : Bare_Op_Lt
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Lt;
   --
   --  Primitives for Bare_Op_Lte
   --
      procedure Initialize_Fields_For_Op_Lte
        (Self : Bare_Op_Lte
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Lte;
   --
   --  Primitives for Bare_Op_Minus
   --
      procedure Initialize_Fields_For_Op_Minus
        (Self : Bare_Op_Minus
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Minus;
   --
   --  Primitives for Bare_Op_Mod
   --
      procedure Initialize_Fields_For_Op_Mod
        (Self : Bare_Op_Mod
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Mod;
   --
   --  Primitives for Bare_Op_Mult
   --
      procedure Initialize_Fields_For_Op_Mult
        (Self : Bare_Op_Mult
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Mult;
   --
   --  Primitives for Bare_Op_Neq
   --
      procedure Initialize_Fields_For_Op_Neq
        (Self : Bare_Op_Neq
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Neq;
   --
   --  Primitives for Bare_Op_Not
   --
      procedure Initialize_Fields_For_Op_Not
        (Self : Bare_Op_Not
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Not;
   --
   --  Primitives for Bare_Op_Not_In
   --
      procedure Initialize_Fields_For_Op_Not_In
        (Self : Bare_Op_Not_In
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Not_In;
   --
   --  Primitives for Bare_Op_Or
   --
      procedure Initialize_Fields_For_Op_Or
        (Self : Bare_Op_Or
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Or;
   --
   --  Primitives for Bare_Op_Or_Else
   --
      procedure Initialize_Fields_For_Op_Or_Else
        (Self : Bare_Op_Or_Else
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Or_Else;
   --
   --  Primitives for Bare_Op_Plus
   --
      procedure Initialize_Fields_For_Op_Plus
        (Self : Bare_Op_Plus
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Plus;
   --
   --  Primitives for Bare_Op_Pow
   --
      procedure Initialize_Fields_For_Op_Pow
        (Self : Bare_Op_Pow
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Pow;
   --
   --  Primitives for Bare_Op_Rem
   --
      procedure Initialize_Fields_For_Op_Rem
        (Self : Bare_Op_Rem
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Rem;
   --
   --  Primitives for Bare_Op_Xor
   --
      procedure Initialize_Fields_For_Op_Xor
        (Self : Bare_Op_Xor
        ) is
      begin
            Initialize_Fields_For_Op
              (Self);
      end Initialize_Fields_For_Op_Xor;
   --
   --  Primitives for Bare_String_Literal
   --
      procedure Initialize_Fields_For_String_Literal
        (Self : Bare_String_Literal
        ) is
      begin
            Initialize_Fields_For_Base_Id
              (Self);
      end Initialize_Fields_For_String_Literal;
--# property-start StringLiteral.has_context_free_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16072
pragma Warnings (Off, "is not referenced");
 function String_Literal_P_Has_Context_Free_Type
  (Node : Bare_String_Literal
  )
   return Boolean
is
   Self : Bare_String_Literal := Bare_String_Literal (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end String_Literal_P_Has_Context_Free_Type;
--# end
--# property-start StringLiteral.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16075
pragma Warnings (Off, "is not referenced");
 function String_Literal_P_Xref_Equation
  (Node : Bare_String_Literal
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_String_Literal := Bare_String_Literal (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_String_Literal :=
     Internal_Entity_String_Literal'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1137 with Inline_Always;
      Fld : Logic_Equation;
Fld_1 : Logic_Var;
Pred : Logic_Equation;
Fld_2 : Logic_Var;
Fld_3 : Logic_Var;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_4 : Bare_Ada_Node;
Is_A : Boolean;
If_Result : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1137 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Pred);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1137;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 9995 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 9994 '<If at ast.py:16076>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16076
--# expr-start 9986 '<IsA Name at ast.py:16079>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16079
--# expr-start 9985 '<parent at ast.py:16079>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16079
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Self.Parent; 
--# expr-done 9985
Is_A := Fld_4 /= null 
and then Fld_4.Kind in Ada_Name; 
--# expr-done 9986
if Is_A then
   --# expr-start 9987 '<FieldAccess for base_id_xref_equation at ast.py:16080>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16080
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseId.base_id_xref_equation
Fld := Libadalang.Implementation.Base_Id_P_Base_Id_Xref_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 9987
   If_Result := Fld;
else
   --# expr-start 9993 '<And at ast.py:16081>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16081
--# expr-start 9989 '<Predicate on BaseTypeDecl.allows_string_literal at ast.py:16081>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16081
--# expr-start 9988 '<FieldAccess for expected_type_var at ast.py:16082>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16082
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 9988
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
Pred := Solver.Create_Predicate (Fld_1, Create_Base_Type_Decl_P_Allows_String_Literal_0_Predicate (Origin)); 
--# expr-done 9989
--# expr-start 9992 '<Bind at ast.py:16083>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16083
--# expr-start 9990 '<FieldAccess for expected_type_var at ast.py:16083>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16083
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 9990
Fld_2.Value := No_Entity;
Entity_Vars.Reset (Fld_2);
--# expr-start 9991 '<FieldAccess for type_var at ast.py:16083>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16083
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 9991
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
Bind_Result := Solver.Create_Unify (Fld_2, Fld_3, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:16083") else null)); 
--# expr-done 9992
And_Pred := Create_And (Pred, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:16081") else null)); 
--# expr-done 9993
   If_Result := And_Pred;
end if;
      Inc_Ref (If_Result);
--# expr-done 9994
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 9995
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1137;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1137;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end String_Literal_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Null_Literal
   --
      procedure Initialize_Fields_For_Null_Literal
        (Self : Bare_Null_Literal
        ) is
      begin
            Initialize_Fields_For_Single_Tok_Node
              (Self);
      end Initialize_Fields_For_Null_Literal;
--# property-start NullLiteral.has_context_free_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16241
pragma Warnings (Off, "is not referenced");
 function Null_Literal_P_Has_Context_Free_Type
  (Node : Bare_Null_Literal
  )
   return Boolean
is
   Self : Bare_Null_Literal := Bare_Null_Literal (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Null_Literal_P_Has_Context_Free_Type;
--# end
--# property-start NullLiteral.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16244
pragma Warnings (Off, "is not referenced");
 function Null_Literal_P_Xref_Equation
  (Node : Bare_Null_Literal
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Null_Literal := Bare_Null_Literal (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1157 with Inline_Always;
      Fld : Logic_Var;
Pred : Logic_Equation;
Fld_1 : Logic_Var;
Fld_2 : Logic_Var;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1157 is
            begin
                     Dec_Ref (Pred);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1157;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10002 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10001 '<And at ast.py:16245>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16245
--# expr-start 9997 '<Predicate on BaseTypeDecl.is_access_type_predicate at ast.py:16246>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16246
--# expr-start 9996 '<FieldAccess for expected_type_var at ast.py:16247>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16247
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 9996
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
Pred := Solver.Create_Predicate (Fld, Create_Base_Type_Decl_P_Is_Access_Type_Predicate_0_Predicate (Origin)); 
--# expr-done 9997
--# expr-start 10000 '<Bind at ast.py:16248>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16248
--# expr-start 9998 '<FieldAccess for expected_type_var at ast.py:16248>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16248
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 9998
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 9999 '<FieldAccess for type_var at ast.py:16248>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16248
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 9999
Fld_2.Value := No_Entity;
Entity_Vars.Reset (Fld_2);
Bind_Result := Solver.Create_Unify (Fld_1, Fld_2, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:16248") else null)); 
--# expr-done 10000
And_Pred := Create_And (Pred, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:16245") else null)); 
--# expr-done 10001
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 10002
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1157;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1157;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Null_Literal_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Num_Literal
   --
      procedure Initialize_Fields_For_Num_Literal
        (Self : Bare_Num_Literal
        ) is
      begin
            Initialize_Fields_For_Single_Tok_Node
              (Self);
      end Initialize_Fields_For_Num_Literal;
--# property-start NumLiteral.is_constant /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16199
pragma Warnings (Off, "is not referenced");
 function Num_Literal_P_Is_Constant
  (Node : Bare_Num_Literal
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Num_Literal := Bare_Num_Literal (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Num_Literal_P_Is_Constant;
--# end
   --
   --  Primitives for Bare_Int_Literal
   --
      procedure Initialize_Fields_For_Int_Literal
        (Self : Bare_Int_Literal
        ) is
      begin
            Initialize_Fields_For_Num_Literal
              (Self);
      end Initialize_Fields_For_Int_Literal;
--# property-start IntLiteral.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16222
pragma Warnings (Off, "is not referenced");
 function Int_Literal_P_Xref_Equation
  (Node : Bare_Int_Literal
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Int_Literal := Bare_Int_Literal (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1154 with Inline_Always;
      Fld : Logic_Var;
Fld_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1154 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1154;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10005 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10004 '<FieldAccess for universal_int_bind at ast.py:16223>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16223
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10003 '<FieldAccess for type_var at ast.py:16223>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16223
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 10003
--# property-call-start AdaNode.universal_int_bind
Fld_1 := Libadalang.Implementation.Ada_Node_P_Universal_Int_Bind (Node => Self, Type_Var => Fld, Origin => Origin);
--# end
--# expr-done 10004
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 10005
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1154;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1154;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Int_Literal_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Real_Literal
   --
      procedure Initialize_Fields_For_Real_Literal
        (Self : Bare_Real_Literal
        ) is
      begin
            Initialize_Fields_For_Num_Literal
              (Self);
      end Initialize_Fields_For_Real_Literal;
--# property-start RealLiteral.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16210
pragma Warnings (Off, "is not referenced");
 function Real_Literal_P_Xref_Equation
  (Node : Bare_Real_Literal
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Real_Literal := Bare_Real_Literal (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1153 with Inline_Always;
      Fld : Logic_Var;
Fld_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1153 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1153;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10008 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10007 '<FieldAccess for universal_real_bind at ast.py:16211>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16211
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10006 '<FieldAccess for type_var at ast.py:16211>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16211
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 10006
--# property-call-start AdaNode.universal_real_bind
Fld_1 := Libadalang.Implementation.Ada_Node_P_Universal_Real_Bind (Node => Self, Type_Var => Fld, Origin => Origin);
--# end
--# expr-done 10007
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 10008
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1153;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1153;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Real_Literal_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Synthetic_Identifier
   --
      procedure Initialize_Fields_For_Synthetic_Identifier
        (Self : Bare_Synthetic_Identifier
        ) is
      begin
            Initialize_Fields_For_Name
              (Self);
      Self.Synthetic_Identifier_Sym := null;
      end Initialize_Fields_For_Synthetic_Identifier;
--# property-start SyntheticIdentifier.name_symbol /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16669
pragma Warnings (Off, "is not referenced");
 function Synthetic_Identifier_P_Name_Symbol
  (Node : Bare_Synthetic_Identifier
  )
   return Symbol_Type
is
   Self : Bare_Synthetic_Identifier := Bare_Synthetic_Identifier (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Symbol_Type;
      Fld : Symbol_Type;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 10009 '<FieldAccess for sym at ast.py:16669>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16669
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Synthetic_Identifier_Sym;
--# expr-done 10009
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Synthetic_Identifier_P_Name_Symbol;
--# end
   --
   --  Primitives for Bare_Target_Name
   --
      procedure Initialize_Fields_For_Target_Name
        (Self : Bare_Target_Name
        ) is
      begin
            Initialize_Fields_For_Name
              (Self);
      Self.Target_Name_R_Ref_Var := Null_Var_Record;
      end Initialize_Fields_For_Target_Name;
--# property-start TargetName.ref_var /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13369
pragma Warnings (Off, "is not referenced");
 function Target_Name_P_Ref_Var
  (Node : Bare_Target_Name
  )
   return Logic_Var
is
   Self : Bare_Target_Name := Bare_Target_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Logic_Var;
      Fld : Logic_Var;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 10010 '<FieldAccess for r_ref_var at ast.py:13369>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13369
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Target_Name_R_Ref_Var'Unrestricted_Access;
--# expr-done 10010
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Target_Name_P_Ref_Var;
--# end
--# property-start TargetName.assign_statement /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13371
pragma Warnings (Off, "is not referenced");
 function Target_Name_P_Assign_Statement
  (Node : Bare_Target_Name
  )
   return Bare_Assign_Stmt
is
   Self : Bare_Target_Name := Bare_Target_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Assign_Stmt;
            procedure Finalizer_Scope_986 with Inline_Always;
      Node_Parents : Bare_Ada_Node_Array_Access;
Is_A : Boolean;
Find_Result : Bare_Ada_Node;
Cast_Result : Bare_Assign_Stmt;
            procedure Finalizer_Scope_986 is
            begin
                     Dec_Ref (Node_Parents);
            end Finalizer_Scope_986;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 10014 '<Cast to AssignStmt at ast.py:13372>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13372
--# expr-start 10013 '<Find at ast.py:13372>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13372
   --# expr-start 10011 '<FieldAccess for parents at ast.py:13372>' Node_Parents /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13372
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents := Parents (Self, True); 
--# expr-done 10011
      Find_Result := No_Bare_Ada_Node;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents;
      begin
         for Item_307 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_307
      --# expr-start 10012 '<IsA AssignStmt at ast.py:13372>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13372
Is_A := Item_307 /= null 
and then Item_307.Kind in Ada_Assign_Stmt_Range; 
--# expr-done 10012
      if Is_A then
         Find_Result := Item_307;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 10013
   if Find_Result = null
      or else Find_Result.Kind in Ada_Assign_Stmt_Range
   then
      Cast_Result := Find_Result;
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 10014
      Property_Result := Cast_Result;
   --# end
      Finalizer_Scope_986;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_986;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Target_Name_P_Assign_Statement;
--# end
--# property-start TargetName.relative_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13377
pragma Warnings (Off, "is not referenced");
 function Target_Name_P_Relative_Name
  (Node : Bare_Target_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Single_Tok_Node
is
   Self : Bare_Target_Name := Bare_Target_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Single_Tok_Node;
      Fld : Bare_Assign_Stmt;
Fld_1 : Bare_Name;
As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld_2 : Internal_Entity_Single_Tok_Node;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 10018 '<FieldAccess for relative_name at ast.py:13378>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13378
--# expr-start 10017 '<as_entity at ast.py:13378>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13378
--# expr-start 10016 '<FieldAccess for dest at ast.py:13378>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13378
--# expr-start 10015 '<FieldAccess for assign_statement at ast.py:13378>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13378
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TargetName.assign_statement
Fld := Libadalang.Implementation.Target_Name_P_Assign_Statement (Node => Self);
--# end
--# expr-done 10015
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Assign_Stmt_F_Dest;
--# expr-done 10016
Is_Null := Fld_1 = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
As_Entity := (Info => E_Info, Node => Fld_1); 
   If_Result := As_Entity;
end if;
--# expr-done 10017
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.relative_name'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_Relative_Name (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 10018
      Property_Result := Fld_2;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Target_Name_P_Relative_Name;
--# end
--# property-start TargetName.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13382
pragma Warnings (Off, "is not referenced");
 function Target_Name_P_Xref_Equation
  (Node : Bare_Target_Name
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Target_Name := Bare_Target_Name (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_988 with Inline_Always;
      Fld : Logic_Var;
Fld_1 : Bare_Assign_Stmt;
Fld_2 : Bare_Name;
Fld_3 : Logic_Var;
Bind_Result : Logic_Equation;
Fld_4 : Logic_Var;
Fld_5 : Bare_Assign_Stmt;
Fld_6 : Bare_Name;
Fld_7 : Logic_Var;
Bind_Result_1 : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_988 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_988;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 10030 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10029 '<And at ast.py:13383>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13383
--# expr-start 10023 '<Bind at ast.py:13384>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13384
--# expr-start 10019 '<FieldAccess for type_var at ast.py:13384>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13384
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 10019
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
--# expr-start 10022 '<FieldAccess for type_var at ast.py:13384>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13384
--# expr-start 10021 '<FieldAccess for dest at ast.py:13384>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13384
--# expr-start 10020 '<FieldAccess for assign_statement at ast.py:13384>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13384
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TargetName.assign_statement
Fld_1 := Libadalang.Implementation.Target_Name_P_Assign_Statement (Node => Self);
--# end
--# expr-done 10020
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Fld_1.Assign_Stmt_F_Dest;
--# expr-done 10021
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_2);
--# expr-done 10022
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
Bind_Result := Solver.Create_Unify (Fld, Fld_3, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13384") else null)); 
--# expr-done 10023
--# expr-start 10028 '<Bind at ast.py:13385>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13385
--# expr-start 10024 '<FieldAccess for ref_var at ast.py:13385>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13385
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_4 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Self);
--# end
--# expr-done 10024
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
--# expr-start 10027 '<FieldAccess for ref_var at ast.py:13385>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13385
--# expr-start 10026 '<FieldAccess for dest at ast.py:13385>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13385
--# expr-start 10025 '<FieldAccess for assign_statement at ast.py:13385>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:13385
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TargetName.assign_statement
Fld_5 := Libadalang.Implementation.Target_Name_P_Assign_Statement (Node => Self);
--# end
--# expr-done 10025
      if Fld_5 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Fld_5.Assign_Stmt_F_Dest;
--# expr-done 10026
      if Fld_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_7 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_6);
--# end
--# expr-done 10027
Fld_7.Value := No_Entity;
Entity_Vars.Reset (Fld_7);
Bind_Result_1 := Solver.Create_Unify (Fld_4, Fld_7, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13385") else null)); 
--# expr-done 10028
And_Pred := Create_And (Bind_Result, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:13383") else null)); 
--# expr-done 10029
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 10030
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_988;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_988;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Target_Name_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Paren_Expr
   --
      procedure Initialize_Fields_For_Paren_Expr
        (Self : Bare_Paren_Expr
         ; Paren_Expr_F_Expr : Bare_Expr
        ) is
      begin
            Initialize_Fields_For_Expr
              (Self);
            Self.Paren_Expr_F_Expr := Paren_Expr_F_Expr;
      end Initialize_Fields_For_Paren_Expr;
   function Paren_Expr_F_Expr
     (Node : Bare_Paren_Expr) return Bare_Expr
   is
   begin
         return Node.Paren_Expr_F_Expr;
   end;
--# property-start ParenExpr.has_context_free_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11533
pragma Warnings (Off, "is not referenced");
 function Paren_Expr_P_Has_Context_Free_Type
  (Node : Bare_Paren_Expr
  )
   return Boolean
is
   Self : Bare_Paren_Expr := Bare_Paren_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Expr;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 10032 '<FieldAccess for has_context_free_type at ast.py:11533>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11533
--# expr-start 10031 '<FieldAccess for expr at ast.py:11533>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11533
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Paren_Expr_F_Expr;
--# expr-done 10031
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.has_context_free_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Expr_P_Has_Context_Free_Type (Node => Fld);
--# end
--# expr-done 10032
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Paren_Expr_P_Has_Context_Free_Type;
--# end
--# property-start ParenExpr.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11536
pragma Warnings (Off, "is not referenced");
 function Paren_Expr_P_Xref_Equation
  (Node : Bare_Paren_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Paren_Expr := Bare_Paren_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Paren_Expr :=
     Internal_Entity_Paren_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_896 with Inline_Always;
      Fld : Internal_Entity_Expr;
Fld_1 : Logic_Equation;
Fld_2 : Bare_Expr;
Fld_3 : Logic_Var;
Fld_4 : Logic_Var;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_5 : Bare_Expr;
Fld_6 : Logic_Var;
Fld_7 : Logic_Var;
Bind_Result_1 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_896 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_896;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10045 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10044 '<And at ast.py:11538>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11538
--# expr-start 10039 '<And at ast.py:11538>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11538
--# expr-start 10034 '<FieldAccess for sub_equation at ast.py:11538>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11538
--# expr-start 10033 '<FieldAccess for expr at ast.py:11538>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11538
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Paren_Expr_F_Expr, Info => Ent.Info);
--# expr-done 10033
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 10034
--# expr-start 10038 '<Bind at ast.py:11539>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11539
--# expr-start 10036 '<FieldAccess for expected_type_var at ast.py:11539>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11539
--# expr-start 10035 '<FieldAccess for expr at ast.py:11539>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11539
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Paren_Expr_F_Expr;
--# expr-done 10035
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_2);
--# expr-done 10036
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
--# expr-start 10037 '<FieldAccess for expected_type_var at ast.py:11539>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11539
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 10037
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
Bind_Result := Solver.Create_Unify (Fld_3, Fld_4, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11539") else null)); 
--# expr-done 10038
And_Pred := Create_And (Fld_1, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11538") else null)); 
--# expr-done 10039
--# expr-start 10043 '<Bind at ast.py:11540>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11540
--# expr-start 10041 '<FieldAccess for type_var at ast.py:11540>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11540
--# expr-start 10040 '<FieldAccess for expr at ast.py:11540>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11540
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Self.Paren_Expr_F_Expr;
--# expr-done 10040
      if Fld_5 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_5);
--# expr-done 10041
Fld_6.Value := No_Entity;
Entity_Vars.Reset (Fld_6);
--# expr-start 10042 '<FieldAccess for type_var at ast.py:11540>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11540
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 10042
Fld_7.Value := No_Entity;
Entity_Vars.Reset (Fld_7);
Bind_Result_1 := Solver.Create_Unify (Fld_6, Fld_7, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11540") else null)); 
--# expr-done 10043
And_Pred_1 := Create_And (And_Pred, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11538") else null)); 
--# expr-done 10044
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 10045
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_896;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_896;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Paren_Expr_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Quantified_Expr
   --
      procedure Initialize_Fields_For_Quantified_Expr
        (Self : Bare_Quantified_Expr
         ; Quantified_Expr_F_Quantifier : Bare_Quantifier
         ; Quantified_Expr_F_Loop_Spec : Bare_For_Loop_Spec
         ; Quantified_Expr_F_Expr : Bare_Expr
        ) is
      begin
            Initialize_Fields_For_Expr
              (Self);
            Self.Quantified_Expr_F_Quantifier := Quantified_Expr_F_Quantifier;
            Self.Quantified_Expr_F_Loop_Spec := Quantified_Expr_F_Loop_Spec;
            Self.Quantified_Expr_F_Expr := Quantified_Expr_F_Expr;
      end Initialize_Fields_For_Quantified_Expr;
   function Quantified_Expr_F_Quantifier
     (Node : Bare_Quantified_Expr) return Bare_Quantifier
   is
   begin
         return Node.Quantified_Expr_F_Quantifier;
   end;
   function Quantified_Expr_F_Loop_Spec
     (Node : Bare_Quantified_Expr) return Bare_For_Loop_Spec
   is
   begin
         return Node.Quantified_Expr_F_Loop_Spec;
   end;
   function Quantified_Expr_F_Expr
     (Node : Bare_Quantified_Expr) return Bare_Expr
   is
   begin
         return Node.Quantified_Expr_F_Expr;
   end;
--# property-start QuantifiedExpr.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17155
pragma Warnings (Off, "is not referenced");
 function Quantified_Expr_P_Xref_Equation
  (Node : Bare_Quantified_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Quantified_Expr := Bare_Quantified_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Quantified_Expr :=
     Internal_Entity_Quantified_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1225 with Inline_Always;
      Block_Var : Boolean;
Fld : Internal_Entity_For_Loop_Spec;
Fld_1 : Boolean;
Fld_2 : Bare_Expr;
Fld_3 : Logic_Var;
Fld_4 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_5 : Internal_Entity_Expr;
Fld_6 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_7 : Internal_Entity_Expr;
Fld_8 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_9 : Logic_Var;
Fld_10 : Bare_Expr;
Fld_11 : Logic_Var;
Bind_Result_1 : Logic_Equation;
And_Pred_2 : Logic_Equation;
False_Rel : Logic_Equation;
If_Result : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1225 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_8);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (False_Rel);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1225;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10064 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10047 '<FieldAccess for resolve_names at ast.py:17158>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17158
--# expr-start 10046 '<FieldAccess for loop_spec at ast.py:17158>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17158
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_For_Loop_Spec (Node => Ent.Node.Quantified_Expr_F_Loop_Spec, Info => Ent.Info);
--# expr-done 10046
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.resolve_names
Fld_1 := Libadalang.Implementation.Ada_Node_P_Resolve_Names (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 10047
Block_Var := Fld_1; 
--# bind spec_success Block_Var
--# expr-start 10063 '<If at ast.py:17160>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17160
if Block_Var then
   --# expr-start 10062 '<And at ast.py:17162>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17162
--# expr-start 10057 '<And at ast.py:17162>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17162
--# expr-start 10054 '<And at ast.py:17162>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17162
--# expr-start 10051 '<Bind at ast.py:17162>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17162
--# expr-start 10049 '<FieldAccess for expected_type_var at ast.py:17162>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17162
--# expr-start 10048 '<FieldAccess for expr at ast.py:17162>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17162
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Quantified_Expr_F_Expr;
--# expr-done 10048
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_2);
--# expr-done 10049
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
--# expr-start 10050 '<FieldAccess for bool_type at ast.py:17162>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17162
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_4 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 10050
      Cast_Result := Create_Internal_Entity
        (Node => Fld_4.Node,
         Info => Fld_4.Info);
Bind_Result := Solver.Create_Assign (Fld_3, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17162") else null)); 
--# expr-done 10051
--# expr-start 10053 '<FieldAccess for sub_equation at ast.py:17163>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17163
--# expr-start 10052 '<FieldAccess for expr at ast.py:17163>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17163
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Expr (Node => Ent.Node.Quantified_Expr_F_Expr, Info => Ent.Info);
--# expr-done 10052
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_6 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_5.Node, Env => Env, Origin => Origin, E_Info => Fld_5.Info);
--# end
--# expr-done 10053
And_Pred := Create_And (Bind_Result, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17162") else null)); 
--# expr-done 10054
--# expr-start 10056 '<FieldAccess for matches_expected_formal_prim_type at ast.py:17164>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17164
--# expr-start 10055 '<FieldAccess for expr at ast.py:17164>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17164
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Expr (Node => Ent.Node.Quantified_Expr_F_Expr, Info => Ent.Info);
--# expr-done 10055
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_prim_type
Fld_8 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Prim_Type (Node => Fld_7.Node, Origin => Origin);
--# end
--# expr-done 10056
And_Pred_1 := Create_And (And_Pred, Fld_8, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17162") else null)); 
--# expr-done 10057
--# expr-start 10061 '<Bind at ast.py:17165>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17165
--# expr-start 10058 '<FieldAccess for type_var at ast.py:17165>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17165
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 10058
Fld_9.Value := No_Entity;
Entity_Vars.Reset (Fld_9);
--# expr-start 10060 '<FieldAccess for type_var at ast.py:17165>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17165
--# expr-start 10059 '<FieldAccess for expr at ast.py:17165>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17165
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Self.Quantified_Expr_F_Expr;
--# expr-done 10059
      if Fld_10 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_10);
--# expr-done 10060
Fld_11.Value := No_Entity;
Entity_Vars.Reset (Fld_11);
Bind_Result_1 := Solver.Create_Unify (Fld_9, Fld_11, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17165") else null)); 
--# expr-done 10061
And_Pred_2 := Create_And (And_Pred_1, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17162") else null)); 
--# expr-done 10062
   If_Result := And_Pred_2;
else
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17166") else null)); 
   If_Result := False_Rel;
end if;
      Inc_Ref (If_Result);
--# expr-done 10063
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 10064
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1225;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1225;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Quantified_Expr_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Raise_Expr
   --
      procedure Initialize_Fields_For_Raise_Expr
        (Self : Bare_Raise_Expr
         ; Raise_Expr_F_Exception_Name : Bare_Name
         ; Raise_Expr_F_Error_Message : Bare_Expr
        ) is
      begin
            Initialize_Fields_For_Expr
              (Self);
            Self.Raise_Expr_F_Exception_Name := Raise_Expr_F_Exception_Name;
            Self.Raise_Expr_F_Error_Message := Raise_Expr_F_Error_Message;
      end Initialize_Fields_For_Raise_Expr;
   function Raise_Expr_F_Exception_Name
     (Node : Bare_Raise_Expr) return Bare_Name
   is
   begin
         return Node.Raise_Expr_F_Exception_Name;
   end;
   function Raise_Expr_F_Error_Message
     (Node : Bare_Raise_Expr) return Bare_Expr
   is
   begin
         return Node.Raise_Expr_F_Error_Message;
   end;
--# property-start RaiseExpr.has_context_free_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18220
pragma Warnings (Off, "is not referenced");
 function Raise_Expr_P_Has_Context_Free_Type
  (Node : Bare_Raise_Expr
  )
   return Boolean
is
   Self : Bare_Raise_Expr := Bare_Raise_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Raise_Expr_P_Has_Context_Free_Type;
--# end
--# property-start RaiseExpr.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18223
pragma Warnings (Off, "is not referenced");
 function Raise_Expr_P_Xref_Equation
  (Node : Bare_Raise_Expr
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Raise_Expr := Bare_Raise_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Raise_Expr :=
     Internal_Entity_Raise_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1285 with Inline_Always;
            procedure Finalizer_Scope_3027 with Inline_Always;
      Var_Expr : Internal_Entity_Expr;
Fld : Internal_Entity_Name;
Fld_1 : Logic_Equation;
Fld_2 : Logic_Var;
Fld_3 : Logic_Var;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_4 : Internal_Entity_Expr;
Fld_5 : Logic_Var;
Sym : Symbol_Type;
Fld_6 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
Fld_7 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred_2 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1285 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1285;
            procedure Finalizer_Scope_3027 is
            begin
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Fld_7);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_3027;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10079 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10078 '<And at ast.py:18224>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18224
--# expr-start 10070 '<And at ast.py:18224>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18224
--# expr-start 10066 '<FieldAccess for sub_equation at ast.py:18225>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18225
--# expr-start 10065 '<FieldAccess for exception_name at ast.py:18225>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18225
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Raise_Expr_F_Exception_Name, Info => Ent.Info);
--# expr-done 10065
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 10066
--# expr-start 10069 '<Bind at ast.py:18226>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18226
--# expr-start 10067 '<FieldAccess for expected_type_var at ast.py:18226>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18226
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 10067
Fld_2.Value := No_Entity;
Entity_Vars.Reset (Fld_2);
--# expr-start 10068 '<FieldAccess for type_var at ast.py:18226>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18226
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 10068
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
Bind_Result := Solver.Create_Unify (Fld_2, Fld_3, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18226") else null)); 
--# expr-done 10069
And_Pred := Create_And (Fld_1, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18224") else null)); 
--# expr-done 10070
--# expr-start 10071 '<FieldAccess for error_message at ast.py:18227>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18227
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Expr (Node => Ent.Node.Raise_Expr_F_Error_Message, Info => Ent.Info);
--# expr-done 10071
Var_Expr := Fld_4;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind er Var_Expr
--# expr-start 10077 '<And at ast.py:18228>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18228
--# expr-start 10075 '<Bind at ast.py:18231>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18231
--# expr-start 10072 '<FieldAccess for expected_type_var at ast.py:18231>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18231
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Var_Expr.Node);
--# expr-done 10072
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
--# expr-start 10074 '<FieldAccess for std_entity at ast.py:18231>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18231
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10073 '<Symbol String at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_String); 
--# expr-done 10073
--# property-call-start AdaNode.std_entity
Fld_6 := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym);
--# end
--# expr-done 10074
Bind_Result_1 := Solver.Create_Assign (Fld_5, Fld_6, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18231") else null)); 
--# expr-done 10075
--# expr-start 10076 '<FieldAccess for sub_equation at ast.py:18232>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18232
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_7 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 10076
And_Pred_1 := Create_And (Bind_Result_1, Fld_7, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18228") else null)); 
--# expr-done 10077
Scope_Result := And_Pred_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_3027;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18234") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred_2 := Create_And (And_Pred, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:18224") else null)); 
--# expr-done 10078
Let_Result := And_Pred_2; Inc_Ref (Let_Result);
--# end
--# expr-done 10079
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1285;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1285;
               Finalizer_Scope_3027;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Raise_Expr_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Un_Op
   --
      procedure Initialize_Fields_For_Un_Op
        (Self : Bare_Un_Op
         ; Un_Op_F_Op : Bare_Op
         ; Un_Op_F_Expr : Bare_Expr
        ) is
      begin
            Initialize_Fields_For_Expr
              (Self);
            Self.Un_Op_F_Op := Un_Op_F_Op;
            Self.Un_Op_F_Expr := Un_Op_F_Expr;
      end Initialize_Fields_For_Un_Op;
   function Un_Op_F_Op
     (Node : Bare_Un_Op) return Bare_Op
   is
   begin
         return Node.Un_Op_F_Op;
   end;
   function Un_Op_F_Expr
     (Node : Bare_Un_Op) return Bare_Expr
   is
   begin
         return Node.Un_Op_F_Expr;
   end;
--# property-start UnOp.has_context_free_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11556
pragma Warnings (Off, "is not referenced");
 function Un_Op_P_Has_Context_Free_Type
  (Node : Bare_Un_Op
  )
   return Boolean
is
   Self : Bare_Un_Op := Bare_Un_Op (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Un_Op_P_Has_Context_Free_Type;
--# end
--# property-start UnOp.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11559
pragma Warnings (Off, "is not referenced");
 function Un_Op_P_Xref_Equation
  (Node : Bare_Un_Op
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Un_Op := Bare_Un_Op (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Un_Op :=
     Internal_Entity_Un_Op'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_898 with Inline_Always;
      Fld : Internal_Entity_Expr;
Fld_1 : Logic_Equation;
True_Rel : Logic_Equation;
Fld_2 : Logic_Equation;
Fld_3 : Logic_Var;
Fld_4 : Bare_Expr;
Fld_5 : Logic_Var;
Bind_Result : Logic_Equation;
Fld_6 : Logic_Var;
Fld_7 : Bare_Expr;
Fld_8 : Logic_Var;
Bind_Result_1 : Logic_Equation;
And_Pred : Logic_Equation;
Or_Pred : Logic_Equation;
Sym : Symbol_Type;
Fld_9 : Boolean;
Sym_1 : Symbol_Type;
Fld_10 : Boolean;
If_Result : Boolean;
If_Result_1 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_898 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Or_Pred);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_898;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10099 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10098 '<And at ast.py:11560>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11560
--# expr-start 10081 '<FieldAccess for sub_equation at ast.py:11560>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11560
--# expr-start 10080 '<FieldAccess for expr at ast.py:11560>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11560
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Un_Op_F_Expr, Info => Ent.Info);
--# expr-done 10080
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 10081
--# expr-start 10097 '<If at ast.py:11560>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11560
--# expr-start 10083 '<FieldAccess for in_aspect at ast.py:11561>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11561
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10082 '<Symbol Depends at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Depends); 
--# expr-done 10082
--# property-call-start AdaNode.in_aspect
Fld_9 := Libadalang.Implementation.Ada_Node_P_In_Aspect (Node => Self, Name => Sym);
--# end
--# expr-done 10083
if Fld_9 then
   If_Result := True;
else
   --# expr-start 10085 '<FieldAccess for in_aspect at ast.py:11561>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11561
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10084 '<Symbol Refined_Depends at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Refined_Depends); 
--# expr-done 10084
--# property-call-start AdaNode.in_aspect
Fld_10 := Libadalang.Implementation.Ada_Node_P_In_Aspect (Node => Self, Name => Sym_1);
--# end
--# expr-done 10085
   If_Result := Fld_10;
end if;
if If_Result then
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11562") else null)); 
   If_Result_1 := True_Rel;
else
   --# expr-start 10096 '<Or at ast.py:11563>' Or_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11563
--# expr-start 10086 '<FieldAccess for overload_equation at ast.py:11563>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11563
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start UnOp.overload_equation
Fld_2 := Libadalang.Implementation.Un_Op_P_Overload_Equation (Node => Ent.Node, Origin => Origin, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 10086
--# expr-start 10095 '<And at ast.py:11564>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11564
--# expr-start 10090 '<Bind at ast.py:11564>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11564
--# expr-start 10087 '<FieldAccess for expected_type_var at ast.py:11564>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11564
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Self);
--# expr-done 10087
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
--# expr-start 10089 '<FieldAccess for expected_type_var at ast.py:11564>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11564
--# expr-start 10088 '<FieldAccess for expr at ast.py:11564>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11564
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Self.Un_Op_F_Expr;
--# expr-done 10088
      if Fld_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_4);
--# expr-done 10089
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
Bind_Result := Solver.Create_Unify (Fld_3, Fld_5, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11564") else null)); 
--# expr-done 10090
--# expr-start 10094 '<Bind at ast.py:11565>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11565
--# expr-start 10091 '<FieldAccess for type_var at ast.py:11565>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11565
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 10091
Fld_6.Value := No_Entity;
Entity_Vars.Reset (Fld_6);
--# expr-start 10093 '<FieldAccess for type_var at ast.py:11565>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11565
--# expr-start 10092 '<FieldAccess for expr at ast.py:11565>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11565
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Self.Un_Op_F_Expr;
--# expr-done 10092
      if Fld_7 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_7);
--# expr-done 10093
Fld_8.Value := No_Entity;
Entity_Vars.Reset (Fld_8);
Bind_Result_1 := Solver.Create_Unify (Fld_6, Fld_8, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11565") else null)); 
--# expr-done 10094
And_Pred := Create_And (Bind_Result, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11564") else null)); 
--# expr-done 10095
Or_Pred := Create_Or (Fld_2, And_Pred, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11563") else null)); 
--# expr-done 10096
   If_Result_1 := Or_Pred;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 10097
And_Pred_1 := Create_And (Fld_1, If_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11560") else null)); 
--# expr-done 10098
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 10099
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_898;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_898;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Un_Op_P_Xref_Equation;
--# end
--# property-start UnOp.overload_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11569
pragma Warnings (Off, "is not referenced");
 function Un_Op_P_Overload_Equation
  (Node : Bare_Un_Op
      ; Origin : Bare_Ada_Node
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Un_Op := Bare_Un_Op (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Un_Op :=
     Internal_Entity_Un_Op'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind env Env
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_899 with Inline_Always;
            procedure Finalizer_Scope_3028 with Inline_Always;
      Fld : Internal_Entity_Op;
Fld_1 : Internal_Entity_Basic_Decl_Array_Access;
Fld_2 : Logic_Equation;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_899 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_899;
            procedure Finalizer_Scope_3028 is
            begin
                     Dec_Ref (Fld_2);
            end Finalizer_Scope_3028;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10105 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10104 '<LogicAny at ast.py:11570>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11570
--# expr-start 10103 '<Map at ast.py:11570>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11570
   --# expr-start 10101 '<FieldAccess for subprograms at ast.py:11570>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11570
--# expr-start 10100 '<FieldAccess for op at ast.py:11570>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11570
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Op (Node => Ent.Node.Un_Op_F_Op, Info => Ent.Info);
--# expr-done 10100
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Op.subprograms
Fld_1 := Libadalang.Implementation.Op_P_Subprograms (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 10101
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Basic_Decl_Array_Access := Fld_1;
      begin
         for Item_308 of
               Collection.Items
         loop
   --# scope-start
               --# bind subp Item_308
      --# expr-start 10102 '<FieldAccess for entity_eq at ast.py:11571>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11571
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start UnOp.entity_eq
Fld_2 := Libadalang.Implementation.Un_Op_P_Entity_Eq (Node => Ent.Node, Subp => Item_308, Origin => Origin, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 10102
         declare
            Item_To_Append : constant Logic_Equation := Fld_2;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_3028;
         end loop;
      end;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 10103
Logic_Boolean_Op := Solver.Create_Any (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11570") else null)); 
--# expr-done 10104
Let_Result := Logic_Boolean_Op; Inc_Ref (Let_Result);
--# end
--# expr-done 10105
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_899;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_899;
               Finalizer_Scope_3028;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Un_Op_P_Overload_Equation;
--# end
--# property-start UnOp.entity_eq /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11575
pragma Warnings (Off, "is not referenced");
 function Un_Op_P_Entity_Eq
  (Node : Bare_Un_Op
      ; Subp : Internal_Entity_Basic_Decl
      ; Origin : Bare_Ada_Node
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Un_Op := Bare_Un_Op (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Un_Op :=
     Internal_Entity_Un_Op'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind subp Subp
   --# bind origin Origin
   --# bind env Env
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_900 with Inline_Always;
      Block_Var : Internal_Entity_Base_Subp_Spec;
Block_Var_1 : Internal_Entity_Defining_Name_Array_Access;
Fld : Internal_Entity_Base_Subp_Spec;
Fld_1 : Internal_Entity_Defining_Name_Array_Access;
Get_Result : Internal_Entity_Defining_Name;
Fld_2 : Internal_Entity_Base_Formal_Param_Decl;
Fld_3 : Internal_Entity_Expr;
Fld_4 : Logic_Equation;
Fld_5 : Logic_Var;
Fld_6 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_7 : Bare_Op;
Fld_8 : Logic_Var;
Cast_Expr : Internal_Entity_Basic_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_9 : Bare_Op;
Fld_10 : Logic_Var;
Cast_Expr_1 : Internal_Entity_Base_Subp_Spec;
Cast_Result_2 : Internal_Entity;
Bind_Result_2 : Logic_Equation;
And_Pred_2 : Logic_Equation;
False_Rel : Logic_Equation;
Len : Integer;
Is_Equal : Boolean;
If_Result : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_900 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (False_Rel);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_900;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10127 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10106 '<FieldAccess for subp_spec_or_null at ast.py:11576>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11576
      if Subp.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Subp.Node, E_Info => Subp.Info);
--# end
--# expr-done 10106
Block_Var := Fld; 
--# bind spec Block_Var
--# expr-start 10107 '<FieldAccess for unpacked_formal_params at ast.py:11577>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11577
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.unpacked_formal_params
Fld_1 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Unpacked_Formal_Params (Node => Block_Var.Node, E_Info => Block_Var.Info);
--# end
--# expr-done 10107
Block_Var_1 := Fld_1; Inc_Ref (Block_Var_1);
--# bind ps Block_Var_1
--# expr-start 10126 '<If at ast.py:11578>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11578
--# expr-start 10109 '<Eq at ast.py:11579>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11579
--# expr-start 10108 '<length at ast.py:11579>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11579
Len := Length (Block_Var_1); 
--# expr-done 10108
Is_Equal := Len = 1; 
--# expr-done 10109
if Is_Equal then
   --# expr-start 10125 '<And at ast.py:11582>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11582
--# expr-start 10121 '<And at ast.py:11582>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11582
--# expr-start 10117 '<And at ast.py:11582>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11582
--# expr-start 10113 '<FieldAccess for call_argument_equation at ast.py:11582>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11582
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10111 '<FieldAccess for formal_decl at ast.py:11582>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11582
--# expr-start 10110 '<at at ast.py:11582>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11582
Get_Result := Get (Block_Var_1, 0, True); 
--# expr-done 10110
      if Get_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.formal_decl
Fld_2 := Libadalang.Implementation.Defining_Name_P_Formal_Decl (Node => Get_Result.Node, E_Info => Get_Result.Info);
--# end
--# expr-done 10111
--# expr-start 10112 '<FieldAccess for expr at ast.py:11582>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11582
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Expr (Node => Ent.Node.Un_Op_F_Expr, Info => Ent.Info);
--# expr-done 10112
--# property-call-start BaseFormalParamHolder.call_argument_equation
Fld_4 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Call_Argument_Equation (Node => Block_Var.Node, Param => Fld_2, Arg => Fld_3, Env => Env, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 10113
--# expr-start 10116 '<Bind at ast.py:11585>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11585
--# expr-start 10114 '<FieldAccess for type_var at ast.py:11585>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11585
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Self);
--# expr-done 10114
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
--# expr-start 10115 '<FieldAccess for return_type at ast.py:11585>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11585
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.return_type
Fld_6 := Libadalang.Implementation.Base_Subp_Spec_P_Return_Type (Node => Block_Var.Node, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 10115
      Cast_Result := Create_Internal_Entity
        (Node => Fld_6.Node,
         Info => Fld_6.Info);
Bind_Result := Solver.Create_Assign (Fld_5, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11585") else null)); 
--# expr-done 10116
And_Pred := Create_And (Fld_4, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11582") else null)); 
--# expr-done 10117
--# expr-start 10120 '<Bind at ast.py:11588>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11588
--# expr-start 10119 '<FieldAccess for ref_var at ast.py:11588>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11588
--# expr-start 10118 '<FieldAccess for op at ast.py:11588>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11588
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Self.Un_Op_F_Op;
--# expr-done 10118
      if Fld_7 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_8 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld_7);
--# end
--# expr-done 10119
Fld_8.Value := No_Entity;
Entity_Vars.Reset (Fld_8);
Cast_Expr := Subp; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_8, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11588") else null)); 
--# expr-done 10120
And_Pred_1 := Create_And (And_Pred, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11582") else null)); 
--# expr-done 10121
--# expr-start 10124 '<Bind at ast.py:11589>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11589
--# expr-start 10123 '<FieldAccess for subp_spec_var at ast.py:11589>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11589
--# expr-start 10122 '<FieldAccess for op at ast.py:11589>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11589
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Self.Un_Op_F_Op;
--# expr-done 10122
      if Fld_9 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.subp_spec_var'
Fld_10 := Libadalang.Implementation.Dispatcher_Name_P_Subp_Spec_Var (Node => Fld_9);
--# end
--# expr-done 10123
Fld_10.Value := No_Entity;
Entity_Vars.Reset (Fld_10);
Cast_Expr_1 := Block_Var; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Bind_Result_2 := Solver.Create_Assign (Fld_10, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11589") else null)); 
--# expr-done 10124
And_Pred_2 := Create_And (And_Pred_1, Bind_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11582") else null)); 
--# expr-done 10125
   If_Result := And_Pred_2;
else
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:11590") else null)); 
   If_Result := False_Rel;
end if;
      Inc_Ref (If_Result);
--# expr-done 10126
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 10127
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_900;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_900;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Un_Op_P_Entity_Eq;
--# end
--# property-start UnOp.potential_actuals_for_dispatch /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11594
pragma Warnings (Off, "is not referenced");
 function Un_Op_P_Potential_Actuals_For_Dispatch
  (Node : Bare_Un_Op
      ; Spec : Internal_Entity_Base_Subp_Spec
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Expected_Type_For_Expr_Array_Access
is
   Self : Bare_Un_Op := Bare_Un_Op (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Un_Op :=
     Internal_Entity_Un_Op'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind spec Spec
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Internal_Expected_Type_For_Expr_Array_Access;
            procedure Finalizer_Scope_901 with Inline_Always;
      Fld : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Get_Result : Internal_Entity_Base_Formal_Param_Decl;
Fld_1 : Internal_Entity_Type_Expr;
Fld_2 : Internal_Entity_Expr;
New_Struct : Internal_Expected_Type_For_Expr;
Singleton : Internal_Expected_Type_For_Expr_Array_Access;
Let_Result : Internal_Expected_Type_For_Expr_Array_Access;
            procedure Finalizer_Scope_901 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Singleton);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_901;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10133 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10132 '<New ExpectedTypeForExpr at ast.py:11595>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11595
--# expr-start 10130 '<FieldAccess for type_expression at ast.py:11596>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11596
--# expr-start 10129 '<at at ast.py:11596>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11596
--# expr-start 10128 '<FieldAccess for abstract_formal_params at ast.py:11596>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11596
      if Spec.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseFormalParamHolder.abstract_formal_params'
Fld := Libadalang.Implementation.Dispatcher_Base_Formal_Param_Holder_P_Abstract_Formal_Params (Node => Spec.Node, E_Info => Spec.Info);
--# end
--# expr-done 10128
Get_Result := Get (Fld, 0, True); 
--# expr-done 10129
      if Get_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.type_expression'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Type_Expression (Node => Get_Result.Node, E_Info => Get_Result.Info);
--# end
--# expr-done 10130
--# expr-start 10131 '<FieldAccess for expr at ast.py:11597>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11597
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Expr (Node => Ent.Node.Un_Op_F_Expr, Info => Ent.Info);
--# expr-done 10131
New_Struct := (Expected_Type => Fld_1, Expr => Fld_2); 
--# expr-done 10132
                Singleton := Create_Internal_Expected_Type_For_Expr_Array (Items_Count => 1);
                Singleton.Items (1) := New_Struct;
Let_Result := Singleton; Inc_Ref (Let_Result);
--# end
--# expr-done 10133
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_901;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_901;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Un_Op_P_Potential_Actuals_For_Dispatch;
--# end
--# property-start UnOp.is_dispatching_call /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11601
pragma Warnings (Off, "is not referenced");
 function Un_Op_P_Is_Dispatching_Call
  (Node : Bare_Un_Op
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Un_Op := Bare_Un_Op (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Un_Op :=
     Internal_Entity_Un_Op'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Basic_Decl;
Fld : Internal_Entity_Op;
Fld_1 : Internal_Entity_Basic_Decl;
Fld_2 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10137 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10135 '<FieldAccess for referenced_decl at ast.py:11602>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11602
--# expr-start 10134 '<FieldAccess for op at ast.py:11602>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11602
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Op (Node => Ent.Node.Un_Op_F_Op, Info => Ent.Info);
--# expr-done 10134
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld_1 := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Fld.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld.Info);
--# end
--# expr-done 10135
Var_Expr := Fld_1;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind decl Var_Expr
--# expr-start 10136 '<FieldAccess for is_dispatching_call_impl at ast.py:11603>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:11603
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_dispatching_call_impl
Fld_2 := Libadalang.Implementation.Expr_P_Is_Dispatching_Call_Impl (Node => Ent.Node, Decl => Var_Expr, Imprecise_Fallback => Imprecise_Fallback, E_Info => Ent.Info);
--# end
--# expr-done 10136
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 10137
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Un_Op_P_Is_Dispatching_Call;
--# end
   --
   --  Primitives for Bare_Handled_Stmts
   --
      procedure Initialize_Fields_For_Handled_Stmts
        (Self : Bare_Handled_Stmts
         ; Handled_Stmts_F_Stmts : Bare_Stmt_List
         ; Handled_Stmts_F_Exceptions : Bare_Ada_Node_List
        ) is
      begin
            Self.Handled_Stmts_F_Stmts := Handled_Stmts_F_Stmts;
            Self.Handled_Stmts_F_Exceptions := Handled_Stmts_F_Exceptions;
      end Initialize_Fields_For_Handled_Stmts;
   function Handled_Stmts_F_Stmts
     (Node : Bare_Handled_Stmts) return Bare_Stmt_List
   is
   begin
         return Node.Handled_Stmts_F_Stmts;
   end;
   function Handled_Stmts_F_Exceptions
     (Node : Bare_Handled_Stmts) return Bare_Ada_Node_List
   is
   begin
         return Node.Handled_Stmts_F_Exceptions;
   end;
   --
   --  Primitives for Bare_Interface_Kind
   --
   --
   --  Primitives for Bare_Interface_Kind_Limited
   --
   --
   --  Primitives for Bare_Interface_Kind_Protected
   --
   --
   --  Primitives for Bare_Interface_Kind_Synchronized
   --
   --
   --  Primitives for Bare_Interface_Kind_Task
   --
   --
   --  Primitives for Bare_Iter_Type
   --
   --
   --  Primitives for Bare_Iter_Type_In
   --
   --
   --  Primitives for Bare_Iter_Type_Of
   --
   --
   --  Primitives for Bare_Library_Item
   --
      procedure Initialize_Fields_For_Library_Item
        (Self : Bare_Library_Item
         ; Library_Item_F_Has_Private : Bare_Private_Node
         ; Library_Item_F_Item : Bare_Basic_Decl
        ) is
      begin
            Self.Library_Item_F_Has_Private := Library_Item_F_Has_Private;
            Self.Library_Item_F_Item := Library_Item_F_Item;
      end Initialize_Fields_For_Library_Item;
   function Library_Item_F_Has_Private
     (Node : Bare_Library_Item) return Bare_Private_Node
   is
   begin
         return Node.Library_Item_F_Has_Private;
   end;
   function Library_Item_F_Item
     (Node : Bare_Library_Item) return Bare_Basic_Decl
   is
   begin
         return Node.Library_Item_F_Item;
   end;
   --
   --  Primitives for Bare_Limited_Node
   --
--# property-start '[dispatcher]Limited.as_bool' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Limited_Node_P_As_Bool
  (Node : Bare_Limited_Node
  )
   return Boolean
is
   Self : Bare_Limited_Node := Bare_Limited_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Limited_Node (Self.Kind) is
               when Ada_Limited_Absent =>
                  --# property-call-start Limited.Absent.as_bool
                  Property_Result := Limited_Absent_P_As_Bool
                    (Self
                    );
                  --# end
               when Ada_Limited_Present =>
                  --# property-call-start Limited.Present.as_bool
                  Property_Result := Limited_Present_P_As_Bool
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Limited_Node_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Limited_Absent
   --
--# property-start Limited.Absent.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4874
pragma Warnings (Off, "is not referenced");
 function Limited_Absent_P_As_Bool
  (Node : Bare_Limited_Absent
  )
   return Boolean
is
   Self : Bare_Limited_Absent := Bare_Limited_Absent (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Limited_Absent_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Limited_Present
   --
--# property-start Limited.Present.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4874
pragma Warnings (Off, "is not referenced");
 function Limited_Present_P_As_Bool
  (Node : Bare_Limited_Present
  )
   return Boolean
is
   Self : Bare_Limited_Present := Bare_Limited_Present (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Limited_Present_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Loop_Spec
   --
   --
   --  Primitives for Bare_For_Loop_Spec
   --
      procedure Initialize_Fields_For_For_Loop_Spec
        (Self : Bare_For_Loop_Spec
         ; For_Loop_Spec_F_Var_Decl : Bare_For_Loop_Var_Decl
         ; For_Loop_Spec_F_Loop_Type : Bare_Iter_Type
         ; For_Loop_Spec_F_Has_Reverse : Bare_Reverse_Node
         ; For_Loop_Spec_F_Iter_Expr : Bare_Ada_Node
         ; For_Loop_Spec_F_Iter_Filter : Bare_Expr
        ) is
      begin
            Self.For_Loop_Spec_F_Var_Decl := For_Loop_Spec_F_Var_Decl;
            Self.For_Loop_Spec_F_Loop_Type := For_Loop_Spec_F_Loop_Type;
            Self.For_Loop_Spec_F_Has_Reverse := For_Loop_Spec_F_Has_Reverse;
            Self.For_Loop_Spec_F_Iter_Expr := For_Loop_Spec_F_Iter_Expr;
            Self.For_Loop_Spec_F_Iter_Filter := For_Loop_Spec_F_Iter_Filter;
      end Initialize_Fields_For_For_Loop_Spec;
   function For_Loop_Spec_F_Var_Decl
     (Node : Bare_For_Loop_Spec) return Bare_For_Loop_Var_Decl
   is
   begin
         return Node.For_Loop_Spec_F_Var_Decl;
   end;
   function For_Loop_Spec_F_Loop_Type
     (Node : Bare_For_Loop_Spec) return Bare_Iter_Type
   is
   begin
         return Node.For_Loop_Spec_F_Loop_Type;
   end;
   function For_Loop_Spec_F_Has_Reverse
     (Node : Bare_For_Loop_Spec) return Bare_Reverse_Node
   is
   begin
         return Node.For_Loop_Spec_F_Has_Reverse;
   end;
   function For_Loop_Spec_F_Iter_Expr
     (Node : Bare_For_Loop_Spec) return Bare_Ada_Node
   is
   begin
         return Node.For_Loop_Spec_F_Iter_Expr;
   end;
   function For_Loop_Spec_F_Iter_Filter
     (Node : Bare_For_Loop_Spec) return Bare_Expr
   is
   begin
         return Node.For_Loop_Spec_F_Iter_Filter;
   end;
--# property-start ForLoopSpec.is_iterated_assoc_spec /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17024
pragma Warnings (Off, "is not referenced");
 function For_Loop_Spec_P_Is_Iterated_Assoc_Spec
  (Node : Bare_For_Loop_Spec
  )
   return Boolean
is
   Self : Bare_For_Loop_Spec := Bare_For_Loop_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 10140 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10139 '<IsA IteratedAssoc at ast.py:17029>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17029
--# expr-start 10138 '<parent at ast.py:17029>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17029
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 10138
Is_A := Fld /= null 
and then Fld.Kind in Ada_Iterated_Assoc_Range; 
--# expr-done 10139
Let_Result := Is_A; 
--# end
--# expr-done 10140
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end For_Loop_Spec_P_Is_Iterated_Assoc_Spec;
--# end
--# property-start ForLoopSpec.iter_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17033
pragma Warnings (Off, "is not referenced");
 function For_Loop_Spec_P_Iter_Type
  (Node : Bare_For_Loop_Spec
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_For_Loop_Spec := Bare_For_Loop_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_For_Loop_Spec :=
     Internal_Entity_For_Loop_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1221 with Inline_Always;
      Block_Var : Logic_Var;
Block_Var_1 : Boolean;
Block_Var_2 : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity;
Cast_Result : Internal_Entity_Expr;
Fld_1 : Logic_Var;
Fld_2 : Internal_Entity;
Pred : Logic_Equation;
True_Rel : Logic_Equation;
Fld_3 : Bare_Ada_Node;
Cast_Result_1 : Bare_Expr;
Fld_4 : Logic_Var;
Fld_5 : Internal_Entity_For_Loop_Var_Decl;
Fld_6 : Internal_Entity_Type_Expr;
Fld_7 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_8 : Bare_For_Loop_Var_Decl;
Fld_9 : Bare_Type_Expr;
Is_Null : Boolean;
If_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_10 : Boolean;
Is_Logic_Var_Defined : Boolean;
Eq_Solution : Internal_Entity;
If_Result_1 : Internal_Entity;
Cast_Result_3 : Internal_Entity_Base_Type_Decl;
If_Result_2 : Internal_Entity_Base_Type_Decl;
Fld_11 : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_12 : Internal_Entity_Base_Type_Decl;
Fld_13 : Boolean;
If_Result_3 : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1221 is
            begin
                     Dec_Ref (Pred);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (If_Result);
                     Dec_Ref (And_Pred);
            end Finalizer_Scope_1221;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_For_Loop_Spec_For_Loop_Spec_P_Iter_Type,
            Items    => new Mmz_Key_Array (1 ..  4))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_For_Loop_Spec,
                                As_Bare_For_Loop_Spec => Self);
               Mmz_K.Items (2) := (Kind => Mmz_Lexical_Env,
                                      As_Lexical_Env => Env);
                  Inc_Ref (Mmz_K.Items (2).As_Lexical_Env);
               Mmz_K.Items (3) := (Kind => Mmz_Bare_Ada_Node,
                                      As_Bare_Ada_Node => Origin);
               Mmz_K.Items (4) :=
                 (Kind => Mmz_Internal_Entity_Info,
                  As_Internal_Entity_Info => E_Info);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Entity_Base_Type_Decl;
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 10167 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10143 '<FieldAccess for type_var at ast.py:17034>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17034
--# expr-start 10142 '<Cast to Expr at ast.py:17034>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17034
--# expr-start 10141 '<FieldAccess for iter_expr at ast.py:17034>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17034
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity (Node => Ent.Node.For_Loop_Spec_F_Iter_Expr, Info => Ent.Info);
--# expr-done 10141
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Expr
   then
      Cast_Result := Create_Internal_Entity_Expr
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 10142
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Cast_Result.Node);
--# expr-done 10143
Block_Var := Fld_1; 
--# bind type_var Block_Var
--# expr-start 10158 '<FieldAccess for resolve_names_internal_with_eq at ast.py:17036>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17036
--# expr-start 10144 '<FieldAccess for iter_expr at ast.py:17036>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17036
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity (Node => Ent.Node.For_Loop_Spec_F_Iter_Expr, Info => Ent.Info);
--# expr-done 10144
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10157 '<And at ast.py:17038>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17038
--# expr-start 10145 '<Predicate on AdaNode.is_not_null at ast.py:17038>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17038
Block_Var.Value := No_Entity;
Entity_Vars.Reset (Block_Var);
Pred := Solver.Create_Predicate (Block_Var, Create_Ada_Node_P_Is_Not_Null_0_Predicate); 
--# expr-done 10145
--# expr-start 10156 '<If at ast.py:17042>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17042
--# expr-start 10148 '<IsNull at ast.py:17043>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17043
--# expr-start 10147 '<FieldAccess for id_type at ast.py:17043>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17043
--# expr-start 10146 '<FieldAccess for var_decl at ast.py:17043>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17043
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Self.For_Loop_Spec_F_Var_Decl;
--# expr-done 10146
      if Fld_8 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Fld_8.For_Loop_Var_Decl_F_Id_Type;
--# expr-done 10147
Is_Null := Fld_9 = null; 
--# expr-done 10148
if Is_Null then
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17044") else null)); 
   If_Result := True_Rel;
else
   --# expr-start 10155 '<Bind at ast.py:17045>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17045
--# expr-start 10151 '<FieldAccess for expected_type_var at ast.py:17045>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17045
--# expr-start 10150 '<Cast to Expr at ast.py:17045>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17045
--# expr-start 10149 '<FieldAccess for iter_expr at ast.py:17045>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17045
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Self.For_Loop_Spec_F_Iter_Expr;
--# expr-done 10149
   if Fld_3 = null
      or else Fld_3.Kind in Ada_Expr
   then
      Cast_Result_1 := Fld_3;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 10150
      if Cast_Result_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Cast_Result_1);
--# expr-done 10151
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
--# expr-start 10154 '<FieldAccess for designated_type at ast.py:17046>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17046
--# expr-start 10153 '<FieldAccess for id_type at ast.py:17046>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17046
--# expr-start 10152 '<FieldAccess for var_decl at ast.py:17046>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17046
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_For_Loop_Var_Decl (Node => Ent.Node.For_Loop_Spec_F_Var_Decl, Info => Ent.Info);
--# expr-done 10152
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Type_Expr (Node => Fld_5.Node.For_Loop_Var_Decl_F_Id_Type, Info => Fld_5.Info);
--# expr-done 10153
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_7 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Fld_6.Node, Origin => Origin, E_Info => Fld_6.Info);
--# end
--# expr-done 10154
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_7.Node,
         Info => Fld_7.Info);
Bind_Result := Solver.Create_Assign (Fld_4, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17045") else null)); 
--# expr-done 10155
   If_Result := Bind_Result;
end if;
      Inc_Ref (If_Result);
--# expr-done 10156
And_Pred := Create_And (Pred, If_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17038") else null)); 
--# expr-done 10157
--# property-call-start AdaNode.resolve_names_internal_with_eq
Fld_10 := Libadalang.Implementation.Ada_Node_P_Resolve_Names_Internal_With_Eq (Node => Fld_2.Node, Additional_Equation => And_Pred, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 10158
Block_Var_1 := Fld_10; 
--# bind p Block_Var_1
--# expr-start 10161 '<If at ast.py:17050>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17050
if Block_Var_1 then
   --# expr-start 10160 '<Cast to BaseTypeDecl at ast.py:17052>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17052
--# expr-start 10159 '<get_value at ast.py:17052>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17052
Is_Logic_Var_Defined := Entity_Vars.Is_Defined (Block_Var); 
if Is_Logic_Var_Defined then
Eq_Solution := Entity_Vars.Get_Value (Block_Var); 
   If_Result_1 := Eq_Solution;
else
   If_Result_1 := No_Entity;
end if;
--# expr-done 10159
   if If_Result_1.Node = null
      or else If_Result_1.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_3 := Create_Internal_Entity_Base_Type_Decl
        (Node => If_Result_1.Node,
         Info => If_Result_1.Info);
   else
         Cast_Result_3 := No_Entity_Base_Type_Decl;
   end if;
--# expr-done 10160
   If_Result_2 := Cast_Result_3;
else
   If_Result_2 := No_Entity_Base_Type_Decl;
end if;
--# expr-done 10161
Block_Var_2 := If_Result_2; 
--# bind typ Block_Var_2
--# expr-start 10166 '<bind at ast.py:17056>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17056
--# expr-start 10162 '<FieldAccess for origin_node at ast.py:17056>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17056
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_11 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 10162
Bound_Origin := Fld_11; 
--# expr-start 10165 '<If at ast.py:17056>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17056
--# expr-start 10163 '<FieldAccess for is_implicit_deref at ast.py:17057>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17057
      if Block_Var_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.is_implicit_deref
Fld_13 := Libadalang.Implementation.Base_Type_Decl_P_Is_Implicit_Deref (Node => Block_Var_2.Node, Origin => Bound_Origin, E_Info => Block_Var_2.Info);
--# end
--# expr-done 10163
if Fld_13 then
   --# expr-start 10164 '<FieldAccess for accessed_type at ast.py:17058>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17058
      if Block_Var_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_12 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Block_Var_2.Node, Origin => Bound_Origin, E_Info => Block_Var_2.Info);
--# end
--# expr-done 10164
   If_Result_3 := Fld_12;
else
   If_Result_3 := Block_Var_2;
end if;
--# expr-done 10165
Dyn_Var_Bind_Result := If_Result_3; 
--# expr-done 10166
Let_Result := Dyn_Var_Bind_Result; 
--# end
--# expr-done 10167
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1221;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Entity_Base_Type_Decl,
                        As_Internal_Entity_Base_Type_Decl => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1221;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end For_Loop_Spec_P_Iter_Type;
--# end
--# property-start ForLoopSpec.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17063
pragma Warnings (Off, "is not referenced");
 function For_Loop_Spec_P_Xref_Equation
  (Node : Bare_For_Loop_Spec
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_For_Loop_Spec := Bare_For_Loop_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_For_Loop_Spec :=
     Internal_Entity_For_Loop_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1222 with Inline_Always;
            procedure Finalizer_Scope_3030 with Inline_Always;
            procedure Finalizer_Scope_3031 with Inline_Always;
            procedure Finalizer_Scope_3032 with Inline_Always;
            procedure Finalizer_Scope_3033 with Inline_Always;
            procedure Finalizer_Scope_3034 with Inline_Always;
            procedure Finalizer_Scope_3035 with Inline_Always;
            procedure Finalizer_Scope_3036 with Inline_Always;
            procedure Finalizer_Scope_3037 with Inline_Always;
      Match_0 : Bare_Iter_Type_In;
Match_1 : Bare_Iter_Type_Of;
Match_0_1 : Internal_Entity_Bin_Op;
Match_1_1 : Internal_Entity_Subtype_Indication;
Match_2 : Internal_Entity_Attribute_Ref;
Match_3 : Internal_Entity_Name;
Match_4 : Internal_Entity;
Var_Expr : Internal_Entity_Base_Type_Decl;
It_Typ : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_For_Loop_Var_Decl;
Fld_1 : Logic_Equation;
Fld_2 : Bare_Iter_Type;
Fld_3 : Internal_Entity;
Fld_4 : Logic_Equation;
Fld_5 : Logic_Var;
Pred : Logic_Equation;
And_Pred : Logic_Equation;
Fld_6 : Bare_For_Loop_Var_Decl;
Fld_7 : Bare_Defining_Name;
Fld_8 : Logic_Var;
Fld_9 : Logic_Var;
Bind_Result : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_10 : Logic_Equation;
Fld_11 : Bare_For_Loop_Var_Decl;
Fld_12 : Bare_Defining_Name;
Fld_13 : Logic_Var;
Fld_14 : Internal_Entity_Base_Type_Decl;
Fld_15 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result_1 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_16 : Logic_Equation;
Fld_17 : Bare_For_Loop_Var_Decl;
Fld_18 : Bare_Defining_Name;
Fld_19 : Logic_Var;
Fld_20 : Logic_Var;
Bind_Result_2 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_21 : Internal_Entity_Base_Type_Decl;
Fld_22 : Logic_Equation;
Fld_23 : Bare_For_Loop_Var_Decl;
Fld_24 : Bare_Defining_Name;
Fld_25 : Logic_Var;
Fld_26 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Bind_Result_3 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Scope_Result : Logic_Equation;
Fld_27 : Logic_Equation;
Result_Var : Logic_Equation;
True_Rel : Logic_Equation;
Match_Prefix : Internal_Entity;
Cast_Result_2 : Internal_Entity_Bin_Op;
Let_Result : Logic_Equation;
Scope_Result_1 : Logic_Equation;
Cast_Result_3 : Internal_Entity_Subtype_Indication;
Let_Result_1 : Logic_Equation;
Scope_Result_2 : Logic_Equation;
Cast_Result_4 : Internal_Entity_Attribute_Ref;
Let_Result_2 : Logic_Equation;
Scope_Result_3 : Logic_Equation;
Cast_Result_5 : Internal_Entity_Name;
Let_Result_3 : Logic_Equation;
Scope_Result_4 : Logic_Equation;
Cast_Result_6 : Internal_Entity;
Let_Result_4 : Logic_Equation;
Scope_Result_5 : Logic_Equation;
Match_Result : Logic_Equation;
Fld_28 : Internal_Entity_Base_Type_Decl;
Fld_29 : Bare_For_Loop_Var_Decl;
Fld_30 : Bare_Defining_Name;
Fld_31 : Logic_Var;
Fld_32 : Internal_Entity_Base_Type_Decl;
Cast_Result_7 : Internal_Entity;
Bind_Result_4 : Logic_Equation;
False_Rel : Logic_Equation;
Fld_33 : Boolean;
If_Result : Logic_Equation;
Let_Result_5 : Logic_Equation;
Match_Prefix_1 : Bare_Iter_Type;
Cast_Result_8 : Bare_Iter_Type_In;
Let_Result_6 : Logic_Equation;
Scope_Result_6 : Logic_Equation;
Cast_Result_9 : Bare_Iter_Type_Of;
Let_Result_7 : Logic_Equation;
Scope_Result_7 : Logic_Equation;
Match_Result_1 : Logic_Equation;
And_Pred_5 : Logic_Equation;
True_Rel_1 : Logic_Equation;
Fld_34 : Bare_Expr;
Fld_35 : Logic_Var;
Fld_36 : Internal_Entity_Base_Type_Decl;
Cast_Result_10 : Internal_Entity;
Bind_Result_5 : Logic_Equation;
Fld_37 : Bare_Expr;
Fld_38 : Logic_Var;
Fld_39 : Internal_Entity_Base_Type_Decl;
Cast_Result_11 : Internal_Entity;
Bind_Result_6 : Logic_Equation;
And_Pred_6 : Logic_Equation;
Fld_40 : Internal_Entity_Expr;
Fld_41 : Logic_Equation;
And_Pred_7 : Logic_Equation;
Fld_42 : Internal_Entity_Expr;
Is_Null : Boolean;
If_Result_1 : Logic_Equation;
And_Pred_8 : Logic_Equation;
Let_Result_8 : Logic_Equation;
            procedure Finalizer_Scope_1222 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result_6);
                     Dec_Ref (Scope_Result_6);
                     Dec_Ref (Let_Result_7);
                     Dec_Ref (Scope_Result_7);
                     Dec_Ref (Match_Result_1);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (True_Rel_1);
                     Dec_Ref (Bind_Result_5);
                     Dec_Ref (Bind_Result_6);
                     Dec_Ref (And_Pred_6);
                     Dec_Ref (Fld_41);
                     Dec_Ref (And_Pred_7);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (And_Pred_8);
                     Dec_Ref (Let_Result_8);
            end Finalizer_Scope_1222;
            procedure Finalizer_Scope_3030 is
            begin
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Let_Result_2);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Let_Result_3);
                     Dec_Ref (Scope_Result_4);
                     Dec_Ref (Let_Result_4);
                     Dec_Ref (Scope_Result_5);
                     Dec_Ref (Match_Result);
            end Finalizer_Scope_3030;
            procedure Finalizer_Scope_3031 is
            begin
                     Dec_Ref (Fld_4);
                     Dec_Ref (Pred);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_3031;
            procedure Finalizer_Scope_3032 is
            begin
                     Dec_Ref (Fld_10);
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (And_Pred_2);
            end Finalizer_Scope_3032;
            procedure Finalizer_Scope_3033 is
            begin
                     Dec_Ref (Fld_16);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (And_Pred_3);
            end Finalizer_Scope_3033;
            procedure Finalizer_Scope_3034 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Fld_27);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_3034;
            procedure Finalizer_Scope_3035 is
            begin
                     Dec_Ref (Fld_22);
                     Dec_Ref (Bind_Result_3);
                     Dec_Ref (And_Pred_4);
            end Finalizer_Scope_3035;
            procedure Finalizer_Scope_3036 is
            begin
                     Dec_Ref (True_Rel);
            end Finalizer_Scope_3036;
            procedure Finalizer_Scope_3037 is
            begin
                     Dec_Ref (Bind_Result_4);
                     Dec_Ref (False_Rel);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result_5);
            end Finalizer_Scope_3037;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10234 '<Block at ???>' Let_Result_8 None
--# scope-start
--# expr-start 10233 '<And at ast.py:17064>' And_Pred_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17064
--# expr-start 10217 '<And at ast.py:17064>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17064
--# expr-start 10169 '<FieldAccess for sub_equation at ast.py:17064>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17064
--# expr-start 10168 '<FieldAccess for var_decl at ast.py:17064>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17064
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_For_Loop_Var_Decl (Node => Ent.Node.For_Loop_Spec_F_Var_Decl, Info => Ent.Info);
--# expr-done 10168
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 10169
--# expr-start 10216 '<Match at ast.py:17064>' Match_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17064
--# expr-start 10170 '<FieldAccess for loop_type at ast.py:17064>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17064
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.For_Loop_Spec_F_Loop_Type;
--# expr-done 10170
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix_1 := Fld_2; 
case Ada_Iter_Type (Match_Prefix_1.Kind) is
         when Ada_Iter_Type_In =>
   --# scope-start
--# scope-start
      Cast_Result_8 := Match_Prefix_1;
Match_0 := Cast_Result_8; 
--# bind _ Match_0
--# expr-start 10206 '<Match at ast.py:17070>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17070
--# expr-start 10171 '<FieldAccess for iter_expr at ast.py:17070>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17070
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity (Node => Ent.Node.For_Loop_Spec_F_Iter_Expr, Info => Ent.Info);
--# expr-done 10171
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Fld_3; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Bin_Op .. Ada_Relation_Op =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Bin_Op
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0_1 := Cast_Result_2; 
--# bind binop Match_0_1
--# expr-start 10181 '<And at ast.py:17073>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17073
--# expr-start 10175 '<And at ast.py:17073>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17073
--# expr-start 10172 '<FieldAccess for sub_equation at ast.py:17073>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17073
      if Match_0_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_4 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Match_0_1.Node, Env => Env, Origin => Origin, E_Info => Match_0_1.Info);
--# end
--# expr-done 10172
--# expr-start 10174 '<Predicate on BaseTypeDecl.is_not_root_int_type at ast.py:17076>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17076
--# expr-start 10173 '<FieldAccess for type_var at ast.py:17076>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17076
      if Match_0_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Match_0_1.Node);
--# expr-done 10173
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
Pred := Solver.Create_Predicate (Fld_5, Create_Base_Type_Decl_P_Is_Not_Root_Int_Type_0_Predicate); 
--# expr-done 10174
And_Pred := Create_And (Fld_4, Pred, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17073") else null)); 
--# expr-done 10175
--# expr-start 10180 '<Bind at ast.py:17077>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17077
--# expr-start 10178 '<FieldAccess for type_var at ast.py:17077>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17077
--# expr-start 10177 '<FieldAccess for id at ast.py:17077>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17077
--# expr-start 10176 '<FieldAccess for var_decl at ast.py:17077>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17077
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Self.For_Loop_Spec_F_Var_Decl;
--# expr-done 10176
      if Fld_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Fld_6.For_Loop_Var_Decl_F_Id;
--# expr-done 10177
      if Fld_7 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_7);
--# expr-done 10178
Fld_8.Value := No_Entity;
Entity_Vars.Reset (Fld_8);
--# expr-start 10179 '<FieldAccess for type_var at ast.py:17077>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17077
      if Match_0_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Match_0_1.Node);
--# expr-done 10179
Fld_9.Value := No_Entity;
Entity_Vars.Reset (Fld_9);
Bind_Result := Solver.Create_Unify (Fld_8, Fld_9, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17077") else null)); 
--# expr-done 10180
And_Pred_1 := Create_And (And_Pred, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17073") else null)); 
--# expr-done 10181
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
Scope_Result_1 := Let_Result;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_3031;
            Match_Result := Scope_Result_1; Inc_Ref (Match_Result);
         when Ada_Subtype_Indication .. Ada_Discrete_Subtype_Indication =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Subtype_Indication
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1_1 := Cast_Result_3; 
--# bind t Match_1_1
--# expr-start 10189 '<And at ast.py:17082>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17082
--# expr-start 10182 '<FieldAccess for sub_equation at ast.py:17082>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17082
      if Match_1_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_10 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Match_1_1.Node, Env => Env, Origin => Origin, E_Info => Match_1_1.Info);
--# end
--# expr-done 10182
--# expr-start 10188 '<Bind at ast.py:17083>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17083
--# expr-start 10185 '<FieldAccess for type_var at ast.py:17083>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17083
--# expr-start 10184 '<FieldAccess for id at ast.py:17083>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17083
--# expr-start 10183 '<FieldAccess for var_decl at ast.py:17083>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17083
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Self.For_Loop_Spec_F_Var_Decl;
--# expr-done 10183
      if Fld_11 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Fld_11.For_Loop_Var_Decl_F_Id;
--# expr-done 10184
      if Fld_12 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_12);
--# expr-done 10185
Fld_13.Value := No_Entity;
Entity_Vars.Reset (Fld_13);
--# expr-start 10187 '<FieldAccess for canonical_type at ast.py:17084>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17084
--# expr-start 10186 '<FieldAccess for designated_type at ast.py:17084>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17084
      if Match_1_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_14 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Match_1_1.Node, Origin => Origin, E_Info => Match_1_1.Info);
--# end
--# expr-done 10186
      if Fld_14.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.canonical_type'
Fld_15 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Canonical_Type (Node => Fld_14.Node, Origin => Origin, E_Info => Fld_14.Info);
--# end
--# expr-done 10187
      Cast_Result := Create_Internal_Entity
        (Node => Fld_15.Node,
         Info => Fld_15.Info);
Bind_Result_1 := Solver.Create_Assign (Fld_13, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17083") else null)); 
--# expr-done 10188
And_Pred_2 := Create_And (Fld_10, Bind_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17082") else null)); 
--# expr-done 10189
Let_Result_1 := And_Pred_2; Inc_Ref (Let_Result_1);
--# end
Scope_Result_2 := Let_Result_1;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_3032;
            Match_Result := Scope_Result_2; Inc_Ref (Match_Result);
         when Ada_Attribute_Ref .. Ada_Update_Attribute_Ref =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Attribute_Ref
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_4; 
--# bind r Match_2
--# expr-start 10196 '<And at ast.py:17087>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17087
--# expr-start 10190 '<FieldAccess for sub_equation at ast.py:17087>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17087
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_16 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Match_2.Node, Env => Env, Origin => Origin, E_Info => Match_2.Info);
--# end
--# expr-done 10190
--# expr-start 10195 '<Bind at ast.py:17088>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17088
--# expr-start 10193 '<FieldAccess for type_var at ast.py:17088>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17088
--# expr-start 10192 '<FieldAccess for id at ast.py:17088>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17088
--# expr-start 10191 '<FieldAccess for var_decl at ast.py:17088>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17088
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_17 := Self.For_Loop_Spec_F_Var_Decl;
--# expr-done 10191
      if Fld_17 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Fld_17.For_Loop_Var_Decl_F_Id;
--# expr-done 10192
      if Fld_18 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_18);
--# expr-done 10193
Fld_19.Value := No_Entity;
Entity_Vars.Reset (Fld_19);
--# expr-start 10194 '<FieldAccess for type_var at ast.py:17088>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17088
      if Match_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_20 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Match_2.Node);
--# expr-done 10194
Fld_20.Value := No_Entity;
Entity_Vars.Reset (Fld_20);
Bind_Result_2 := Solver.Create_Unify (Fld_19, Fld_20, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17088") else null)); 
--# expr-done 10195
And_Pred_3 := Create_And (Fld_16, Bind_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17087") else null)); 
--# expr-done 10196
Let_Result_2 := And_Pred_3; Inc_Ref (Let_Result_2);
--# end
Scope_Result_3 := Let_Result_2;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_3033;
            Match_Result := Scope_Result_3; Inc_Ref (Match_Result);
         when Ada_Call_Expr .. Ada_Target_Name =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Create_Internal_Entity_Name
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_5; 
--# bind t Match_3
--# expr-start 10197 '<FieldAccess for name_designated_type at ast.py:17094>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17094
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_21 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Match_3.Node, E_Info => Match_3.Info);
--# end
--# expr-done 10197
Var_Expr := Fld_21;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind typ Var_Expr
--# expr-start 10204 '<And at ast.py:17096>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17096
--# expr-start 10198 '<FieldAccess for sub_equation at ast.py:17096>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17096
      if Match_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_22 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Match_3.Node, Env => Env, Origin => Origin, E_Info => Match_3.Info);
--# end
--# expr-done 10198
--# expr-start 10203 '<Bind at ast.py:17097>' Bind_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17097
--# expr-start 10201 '<FieldAccess for type_var at ast.py:17097>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17097
--# expr-start 10200 '<FieldAccess for id at ast.py:17097>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17097
--# expr-start 10199 '<FieldAccess for var_decl at ast.py:17097>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17097
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_23 := Self.For_Loop_Spec_F_Var_Decl;
--# expr-done 10199
      if Fld_23 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_24 := Fld_23.For_Loop_Var_Decl_F_Id;
--# expr-done 10200
      if Fld_24 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_25 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_24);
--# expr-done 10201
Fld_25.Value := No_Entity;
Entity_Vars.Reset (Fld_25);
--# expr-start 10202 '<FieldAccess for canonical_type at ast.py:17097>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17097
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.canonical_type'
Fld_26 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Canonical_Type (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 10202
      Cast_Result_1 := Create_Internal_Entity
        (Node => Fld_26.Node,
         Info => Fld_26.Info);
Bind_Result_3 := Solver.Create_Assign (Fld_25, Cast_Result_1, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17097") else null)); 
--# expr-done 10203
And_Pred_4 := Create_And (Fld_22, Bind_Result_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17096") else null)); 
--# expr-done 10204
Scope_Result := And_Pred_4;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_3035;
   Result_Var := Scope_Result;
else
   --# expr-start 10205 '<FieldAccess for iterator_xref_equation at ast.py:17098>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17098
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ForLoopSpec.iterator_xref_equation
Fld_27 := Libadalang.Implementation.For_Loop_Spec_P_Iterator_Xref_Equation (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 10205
   Result_Var := Fld_27;
end if;
      Inc_Ref (Result_Var);
Let_Result_3 := Result_Var; Inc_Ref (Let_Result_3);
--# end
Scope_Result_4 := Let_Result_3;
   Inc_Ref (Scope_Result_4);
   --# end
      Finalizer_Scope_3034;
            Match_Result := Scope_Result_4; Inc_Ref (Match_Result);
         when Ada_Abort_Absent .. Ada_Null_Record_Aggregate | Ada_Box_Expr .. Ada_Membership_Expr | Ada_Paren_Expr .. Ada_Enum_Lit_Synth_Type_Expr | Ada_Synthetic_Type_Expr .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_6 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_4 := Cast_Result_6; 
--# bind _ Match_4
True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17101") else null)); 
Let_Result_4 := True_Rel; Inc_Ref (Let_Result_4);
--# end
Scope_Result_5 := Let_Result_4;
   Inc_Ref (Scope_Result_5);
   --# end
      Finalizer_Scope_3036;
            Match_Result := Scope_Result_5; Inc_Ref (Match_Result);
end case;
--# expr-done 10206
Let_Result_6 := Match_Result; Inc_Ref (Let_Result_6);
--# end
Scope_Result_6 := Let_Result_6;
   Inc_Ref (Scope_Result_6);
   --# end
      Finalizer_Scope_3030;
            Match_Result_1 := Scope_Result_6; Inc_Ref (Match_Result_1);
         when Ada_Iter_Type_Of =>
   --# scope-start
--# scope-start
      Cast_Result_9 := Match_Prefix_1;
Match_1 := Cast_Result_9; 
--# bind _ Match_1
--# expr-start 10215 '<Let at ast.py:17105>' Let_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17105
--# scope-start
--# expr-start 10207 '<FieldAccess for iter_type at ast.py:17105>' Fld_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17105
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ForLoopSpec.iter_type
Fld_28 := Libadalang.Implementation.For_Loop_Spec_P_Iter_Type (Node => Ent.Node, Env => Env, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 10207
It_Typ := Fld_28; 
--# bind it_typ It_Typ
--# expr-start 10214 '<If at ast.py:17105>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17105
--# expr-start 10208 '<FieldAccess for is_iterable_type at ast.py:17106>' Fld_33 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17106
      if It_Typ.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_iterable_type'
Fld_33 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Iterable_Type (Node => It_Typ.Node, Origin => Origin, E_Info => It_Typ.Info);
--# end
--# expr-done 10208
if Fld_33 then
   --# expr-start 10213 '<Bind at ast.py:17110>' Bind_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17110
--# expr-start 10211 '<FieldAccess for type_var at ast.py:17110>' Fld_31 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17110
--# expr-start 10210 '<FieldAccess for id at ast.py:17110>' Fld_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17110
--# expr-start 10209 '<FieldAccess for var_decl at ast.py:17110>' Fld_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17110
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_29 := Self.For_Loop_Spec_F_Var_Decl;
--# expr-done 10209
      if Fld_29 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_30 := Fld_29.For_Loop_Var_Decl_F_Id;
--# expr-done 10210
      if Fld_30 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_31 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_30);
--# expr-done 10211
Fld_31.Value := No_Entity;
Entity_Vars.Reset (Fld_31);
--# expr-start 10212 '<FieldAccess for iterable_comp_type at ast.py:17110>' Fld_32 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17110
      if It_Typ.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.iterable_comp_type'
Fld_32 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Iterable_Comp_Type (Node => It_Typ.Node, Origin => Origin, E_Info => It_Typ.Info);
--# end
--# expr-done 10212
      Cast_Result_7 := Create_Internal_Entity
        (Node => Fld_32.Node,
         Info => Fld_32.Info);
Bind_Result_4 := Solver.Create_Assign (Fld_31, Cast_Result_7, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17110") else null)); 
--# expr-done 10213
   If_Result := Bind_Result_4;
else
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17112") else null)); 
   If_Result := False_Rel;
end if;
      Inc_Ref (If_Result);
--# expr-done 10214
Let_Result_5 := If_Result; Inc_Ref (Let_Result_5);
--# end
--# expr-done 10215
Let_Result_7 := Let_Result_5; Inc_Ref (Let_Result_7);
--# end
Scope_Result_7 := Let_Result_7;
   Inc_Ref (Scope_Result_7);
   --# end
      Finalizer_Scope_3037;
            Match_Result_1 := Scope_Result_7; Inc_Ref (Match_Result_1);
end case;
--# expr-done 10216
And_Pred_5 := Create_And (Fld_1, Match_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17064") else null)); 
--# expr-done 10217
--# expr-start 10232 '<If at ast.py:17114>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17114
--# expr-start 10219 '<IsNull at ast.py:17115>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17115
--# expr-start 10218 '<FieldAccess for iter_filter at ast.py:17115>' Fld_42 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17115
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_42 := Create_Internal_Entity_Expr (Node => Ent.Node.For_Loop_Spec_F_Iter_Filter, Info => Ent.Info);
--# expr-done 10218
Is_Null := Fld_42.Node = null; 
--# expr-done 10219
if Is_Null then
   True_Rel_1 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17116") else null)); 
   If_Result_1 := True_Rel_1;
else
   --# expr-start 10231 '<And at ast.py:17117>' And_Pred_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17117
--# expr-start 10228 '<And at ast.py:17117>' And_Pred_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17117
--# expr-start 10223 '<Bind at ast.py:17117>' Bind_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17117
--# expr-start 10221 '<FieldAccess for expected_type_var at ast.py:17117>' Fld_35 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17117
--# expr-start 10220 '<FieldAccess for iter_filter at ast.py:17117>' Fld_34 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17117
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_34 := Self.For_Loop_Spec_F_Iter_Filter;
--# expr-done 10220
      if Fld_34 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_35 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_34);
--# expr-done 10221
Fld_35.Value := No_Entity;
Entity_Vars.Reset (Fld_35);
--# expr-start 10222 '<FieldAccess for bool_type at ast.py:17117>' Fld_36 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17117
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_36 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 10222
      Cast_Result_10 := Create_Internal_Entity
        (Node => Fld_36.Node,
         Info => Fld_36.Info);
Bind_Result_5 := Solver.Create_Assign (Fld_35, Cast_Result_10, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17117") else null)); 
--# expr-done 10223
--# expr-start 10227 '<Bind at ast.py:17118>' Bind_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17118
--# expr-start 10225 '<FieldAccess for type_var at ast.py:17118>' Fld_38 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17118
--# expr-start 10224 '<FieldAccess for iter_filter at ast.py:17118>' Fld_37 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17118
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_37 := Self.For_Loop_Spec_F_Iter_Filter;
--# expr-done 10224
      if Fld_37 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_38 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_37);
--# expr-done 10225
Fld_38.Value := No_Entity;
Entity_Vars.Reset (Fld_38);
--# expr-start 10226 '<FieldAccess for bool_type at ast.py:17118>' Fld_39 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17118
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_39 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 10226
      Cast_Result_11 := Create_Internal_Entity
        (Node => Fld_39.Node,
         Info => Fld_39.Info);
Bind_Result_6 := Solver.Create_Assign (Fld_38, Cast_Result_11, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17118") else null)); 
--# expr-done 10227
And_Pred_6 := Create_And (Bind_Result_5, Bind_Result_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17117") else null)); 
--# expr-done 10228
--# expr-start 10230 '<FieldAccess for sub_equation at ast.py:17119>' Fld_41 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17119
--# expr-start 10229 '<FieldAccess for iter_filter at ast.py:17119>' Fld_40 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17119
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_40 := Create_Internal_Entity_Expr (Node => Ent.Node.For_Loop_Spec_F_Iter_Filter, Info => Ent.Info);
--# expr-done 10229
      if Fld_40.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_41 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_40.Node, Env => Env, Origin => Origin, E_Info => Fld_40.Info);
--# end
--# expr-done 10230
And_Pred_7 := Create_And (And_Pred_6, Fld_41, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17117") else null)); 
--# expr-done 10231
   If_Result_1 := And_Pred_7;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 10232
And_Pred_8 := Create_And (And_Pred_5, If_Result_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17064") else null)); 
--# expr-done 10233
Let_Result_8 := And_Pred_8; Inc_Ref (Let_Result_8);
--# end
--# expr-done 10234
      Property_Result := Let_Result_8;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1222;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1222;
               Finalizer_Scope_3030;
               Finalizer_Scope_3031;
               Finalizer_Scope_3032;
               Finalizer_Scope_3033;
               Finalizer_Scope_3034;
               Finalizer_Scope_3035;
               Finalizer_Scope_3036;
               Finalizer_Scope_3037;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end For_Loop_Spec_P_Xref_Equation;
--# end
--# property-start ForLoopSpec.iterator_xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17123
pragma Warnings (Off, "is not referenced");
 function For_Loop_Spec_P_Iterator_Xref_Equation
  (Node : Bare_For_Loop_Spec
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_For_Loop_Spec := Bare_For_Loop_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_For_Loop_Spec :=
     Internal_Entity_For_Loop_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1223 with Inline_Always;
      Block_Var : Internal_Entity_Expr;
Block_Var_1 : Boolean;
Block_Var_2 : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity;
Cast_Result : Internal_Entity_Expr;
Fld_1 : Logic_Var;
Pred : Logic_Equation;
Fld_2 : Boolean;
Fld_3 : Internal_Entity;
Fld_4 : Lexical_Env := Empty_Env;
Sym : Symbol_Type;
Env_Get_Result : Internal_Entity;
Cast_Result_1 : Internal_Entity_Base_Type_Decl;
Fld_5 : Bare_For_Loop_Var_Decl;
Fld_6 : Bare_Defining_Name;
Fld_7 : Logic_Var;
Cast_Expr : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result : Logic_Equation;
False_Rel : Logic_Equation;
If_Result : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1223 is
            begin
                     Dec_Ref (Pred);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (False_Rel);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1223;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10250 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10236 '<Cast to Expr at ast.py:17124>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17124
--# expr-start 10235 '<FieldAccess for iter_expr at ast.py:17124>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17124
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity (Node => Ent.Node.For_Loop_Spec_F_Iter_Expr, Info => Ent.Info);
--# expr-done 10235
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Expr
   then
      Cast_Result := Create_Internal_Entity_Expr
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 10236
Block_Var := Cast_Result; 
--# bind iter_expr Block_Var
--# expr-start 10239 '<FieldAccess for resolve_names_internal_with_eq at ast.py:17126>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17126
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10238 '<Predicate on BaseTypeDecl.is_iterator_type at ast.py:17127>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17127
--# expr-start 10237 '<FieldAccess for type_var at ast.py:17128>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17128
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Block_Var.Node);
--# expr-done 10237
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
Pred := Solver.Create_Predicate (Fld_1, Create_Base_Type_Decl_P_Is_Iterator_Type_0_Predicate (Origin)); 
--# expr-done 10238
--# property-call-start AdaNode.resolve_names_internal_with_eq
Fld_2 := Libadalang.Implementation.Ada_Node_P_Resolve_Names_Internal_With_Eq (Node => Block_Var.Node, Additional_Equation => Pred, Env => Env, Origin => Origin, E_Info => Block_Var.Info);
--# end
--# expr-done 10239
Block_Var_1 := Fld_2; 
--# bind p Block_Var_1
--# expr-start 10244 '<Cast to BaseTypeDecl at ast.py:17131>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17131
--# expr-start 10243 '<EnvGet at ast.py:17131>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17131
--# expr-start 10241 '<FieldAccess for children_env at ast.py:17131>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17131
--# expr-start 10240 '<FieldAccess for type_val at ast.py:17131>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17131
      if Block_Var.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.type_val
Fld_3 := Libadalang.Implementation.Expr_P_Type_Val (Node => Block_Var.Node);
--# end
--# expr-done 10240
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Children_Env (Node => Fld_3.Node, E_Info => Fld_3.Info);
--# expr-done 10241
--# expr-start 10242 '<Symbol Cursor at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Cursor); 
--# expr-done 10242
Env_Get_Result := AST_Envs.Get_First (Self => Fld_4, Key => Sym, Lookup_Kind => To_Lookup_Kind_Type (Recursive), Categories => (Inherited_Primitives => True, Nocat => True)); 
--# expr-done 10243
   if Env_Get_Result.Node = null
      or else Env_Get_Result.Node.Kind in Ada_Base_Type_Decl
   then
      Cast_Result_1 := Create_Internal_Entity_Base_Type_Decl
        (Node => Env_Get_Result.Node,
         Info => Env_Get_Result.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 10244
Block_Var_2 := Cast_Result_1; 
--# bind cursor_type Block_Var_2
--# expr-start 10249 '<If at ast.py:17134>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17134
if Block_Var_1 then
   --# expr-start 10248 '<Bind at ast.py:17136>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17136
--# expr-start 10247 '<FieldAccess for type_var at ast.py:17136>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17136
--# expr-start 10246 '<FieldAccess for id at ast.py:17136>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17136
--# expr-start 10245 '<FieldAccess for var_decl at ast.py:17136>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17136
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Self.For_Loop_Spec_F_Var_Decl;
--# expr-done 10245
      if Fld_5 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Fld_5.For_Loop_Var_Decl_F_Id;
--# expr-done 10246
      if Fld_6 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_6);
--# expr-done 10247
Fld_7.Value := No_Entity;
Entity_Vars.Reset (Fld_7);
Cast_Expr := Block_Var_2; 
      Cast_Result_2 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Bind_Result := Solver.Create_Assign (Fld_7, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17136") else null)); 
--# expr-done 10248
   If_Result := Bind_Result;
else
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:17137") else null)); 
   If_Result := False_Rel;
end if;
      Inc_Ref (If_Result);
--# expr-done 10249
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 10250
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1223;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1223;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end For_Loop_Spec_P_Iterator_Xref_Equation;
--# end
--# property-start ForLoopSpec.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17143
pragma Warnings (Off, "is not referenced");
 function For_Loop_Spec_P_Xref_Entry_Point
  (Node : Bare_For_Loop_Spec
  )
   return Boolean
is
   Self : Bare_For_Loop_Spec := Bare_For_Loop_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Boolean;
Not_Val : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 10252 '<Not at ast.py:17143>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17143
--# expr-start 10251 '<FieldAccess for is_iterated_assoc_spec at ast.py:17143>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:17143
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ForLoopSpec.is_iterated_assoc_spec
Fld := Libadalang.Implementation.For_Loop_Spec_P_Is_Iterated_Assoc_Spec (Node => Self);
--# end
--# expr-done 10251
Not_Val := not (Fld); 
--# expr-done 10252
      Property_Result := Not_Val;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end For_Loop_Spec_P_Xref_Entry_Point;
--# end
   --
   --  Primitives for Bare_While_Loop_Spec
   --
      procedure Initialize_Fields_For_While_Loop_Spec
        (Self : Bare_While_Loop_Spec
         ; While_Loop_Spec_F_Expr : Bare_Expr
        ) is
      begin
            Self.While_Loop_Spec_F_Expr := While_Loop_Spec_F_Expr;
      end Initialize_Fields_For_While_Loop_Spec;
   function While_Loop_Spec_F_Expr
     (Node : Bare_While_Loop_Spec) return Bare_Expr
   is
   begin
         return Node.While_Loop_Spec_F_Expr;
   end;
--# property-start WhileLoopSpec.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19403
pragma Warnings (Off, "is not referenced");
 function While_Loop_Spec_P_Xref_Equation
  (Node : Bare_While_Loop_Spec
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_While_Loop_Spec := Bare_While_Loop_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_While_Loop_Spec :=
     Internal_Entity_While_Loop_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1357 with Inline_Always;
      Fld : Bare_Expr;
Fld_1 : Logic_Var;
Fld_2 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_3 : Internal_Entity_Expr;
Fld_4 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_5 : Internal_Entity_Expr;
Fld_6 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1357 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_4);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1357;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10263 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10262 '<And at ast.py:19404>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19404
--# expr-start 10259 '<And at ast.py:19404>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19404
--# expr-start 10256 '<Bind at ast.py:19405>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19405
--# expr-start 10254 '<FieldAccess for expected_type_var at ast.py:19405>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19405
--# expr-start 10253 '<FieldAccess for expr at ast.py:19405>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19405
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.While_Loop_Spec_F_Expr;
--# expr-done 10253
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld);
--# expr-done 10254
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 10255 '<FieldAccess for bool_type at ast.py:19405>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19405
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_2 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 10255
      Cast_Result := Create_Internal_Entity
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Bind_Result := Solver.Create_Assign (Fld_1, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19405") else null)); 
--# expr-done 10256
--# expr-start 10258 '<FieldAccess for sub_equation at ast.py:19406>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19406
--# expr-start 10257 '<FieldAccess for expr at ast.py:19406>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19406
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Expr (Node => Ent.Node.While_Loop_Spec_F_Expr, Info => Ent.Info);
--# expr-done 10257
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_4 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_3.Node, Env => Env, Origin => Origin, E_Info => Fld_3.Info);
--# end
--# expr-done 10258
And_Pred := Create_And (Bind_Result, Fld_4, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19404") else null)); 
--# expr-done 10259
--# expr-start 10261 '<FieldAccess for matches_expected_formal_prim_type at ast.py:19407>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19407
--# expr-start 10260 '<FieldAccess for expr at ast.py:19407>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19407
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Expr (Node => Ent.Node.While_Loop_Spec_F_Expr, Info => Ent.Info);
--# expr-done 10260
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_prim_type
Fld_6 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Prim_Type (Node => Fld_5.Node, Origin => Origin);
--# end
--# expr-done 10261
And_Pred_1 := Create_And (And_Pred, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19404") else null)); 
--# expr-done 10262
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 10263
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1357;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1357;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end While_Loop_Spec_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Mode
   --
--# property-start Mode.is_writable /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8652
pragma Warnings (Off, "is not referenced");
 function Mode_P_Is_Writable
  (Node : Bare_Mode
  )
   return Boolean
is
   Self : Bare_Mode := Bare_Mode (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Is_A : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 10265 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10264 '<IsA Mode.Out, Mode.InOut at ast.py:8657>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8657
Is_A := Self /= null 
and then Self.Kind in Ada_Mode_Out_Range | Ada_Mode_In_Out_Range; 
--# expr-done 10264
Let_Result := Is_A; 
--# end
--# expr-done 10265
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Mode_P_Is_Writable;
--# end
   --
   --  Primitives for Bare_Mode_Default
   --
   --
   --  Primitives for Bare_Mode_In
   --
   --
   --  Primitives for Bare_Mode_In_Out
   --
   --
   --  Primitives for Bare_Mode_Out
   --
   --
   --  Primitives for Bare_Multi_Abstract_State_Decl
   --
      procedure Initialize_Fields_For_Multi_Abstract_State_Decl
        (Self : Bare_Multi_Abstract_State_Decl
         ; Multi_Abstract_State_Decl_F_Decls : Bare_Abstract_State_Decl_List
        ) is
      begin
            Self.Multi_Abstract_State_Decl_F_Decls := Multi_Abstract_State_Decl_F_Decls;
      end Initialize_Fields_For_Multi_Abstract_State_Decl;
   function Multi_Abstract_State_Decl_F_Decls
     (Node : Bare_Multi_Abstract_State_Decl) return Bare_Abstract_State_Decl_List
   is
   begin
         return Node.Multi_Abstract_State_Decl_F_Decls;
   end;
   --
   --  Primitives for Bare_Not_Null
   --
--# property-start '[dispatcher]NotNull.as_bool' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Not_Null_P_As_Bool
  (Node : Bare_Not_Null
  )
   return Boolean
is
   Self : Bare_Not_Null := Bare_Not_Null (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Not_Null (Self.Kind) is
               when Ada_Not_Null_Absent =>
                  --# property-call-start NotNull.Absent.as_bool
                  Property_Result := Not_Null_Absent_P_As_Bool
                    (Self
                    );
                  --# end
               when Ada_Not_Null_Present =>
                  --# property-call-start NotNull.Present.as_bool
                  Property_Result := Not_Null_Present_P_As_Bool
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Not_Null_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Not_Null_Absent
   --
--# property-start NotNull.Absent.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4898
pragma Warnings (Off, "is not referenced");
 function Not_Null_Absent_P_As_Bool
  (Node : Bare_Not_Null_Absent
  )
   return Boolean
is
   Self : Bare_Not_Null_Absent := Bare_Not_Null_Absent (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Not_Null_Absent_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Not_Null_Present
   --
--# property-start NotNull.Present.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4898
pragma Warnings (Off, "is not referenced");
 function Not_Null_Present_P_As_Bool
  (Node : Bare_Not_Null_Present
  )
   return Boolean
is
   Self : Bare_Not_Null_Present := Bare_Not_Null_Present (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Not_Null_Present_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Null_Component_Decl
   --
   --
   --  Primitives for Bare_Others_Designator
   --
--# property-start OthersDesignator.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:14578
pragma Warnings (Off, "is not referenced");
 function Others_Designator_P_Xref_Equation
  (Node : Bare_Others_Designator
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Others_Designator := Bare_Others_Designator (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1049 with Inline_Always;
      True_Rel : Logic_Equation;
            procedure Finalizer_Scope_1049 is
            begin
                     Dec_Ref (True_Rel);
            end Finalizer_Scope_1049;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:14578") else null)); 
      Property_Result := True_Rel;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1049;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1049;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Others_Designator_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Overriding_Node
   --
   --
   --  Primitives for Bare_Overriding_Not_Overriding
   --
   --
   --  Primitives for Bare_Overriding_Overriding
   --
   --
   --  Primitives for Bare_Overriding_Unspecified
   --
   --
   --  Primitives for Bare_Params
   --
      procedure Initialize_Fields_For_Params
        (Self : Bare_Params
         ; Params_F_Params : Bare_Param_Spec_List
        ) is
      begin
            Self.Params_F_Params := Params_F_Params;
      end Initialize_Fields_For_Params;
   function Params_F_Params
     (Node : Bare_Params) return Bare_Param_Spec_List
   is
   begin
         return Node.Params_F_Params;
   end;
   --
   --  Primitives for Bare_Paren_Abstract_State_Decl
   --
      procedure Initialize_Fields_For_Paren_Abstract_State_Decl
        (Self : Bare_Paren_Abstract_State_Decl
         ; Paren_Abstract_State_Decl_F_Decl : Bare_Ada_Node
        ) is
      begin
            Self.Paren_Abstract_State_Decl_F_Decl := Paren_Abstract_State_Decl_F_Decl;
      end Initialize_Fields_For_Paren_Abstract_State_Decl;
   function Paren_Abstract_State_Decl_F_Decl
     (Node : Bare_Paren_Abstract_State_Decl) return Bare_Ada_Node
   is
   begin
         return Node.Paren_Abstract_State_Decl_F_Decl;
   end;
   --
   --  Primitives for Bare_Pp_Directive
   --
   --
   --  Primitives for Bare_Pp_Else_Directive
   --
   --
   --  Primitives for Bare_Pp_Elsif_Directive
   --
      procedure Initialize_Fields_For_Pp_Elsif_Directive
        (Self : Bare_Pp_Elsif_Directive
         ; Pp_Elsif_Directive_F_Expr : Bare_Expr
         ; Pp_Elsif_Directive_F_Then_Kw : Bare_Pp_Then_Kw
        ) is
      begin
            Self.Pp_Elsif_Directive_F_Expr := Pp_Elsif_Directive_F_Expr;
            Self.Pp_Elsif_Directive_F_Then_Kw := Pp_Elsif_Directive_F_Then_Kw;
      end Initialize_Fields_For_Pp_Elsif_Directive;
   function Pp_Elsif_Directive_F_Expr
     (Node : Bare_Pp_Elsif_Directive) return Bare_Expr
   is
   begin
         return Node.Pp_Elsif_Directive_F_Expr;
   end;
   function Pp_Elsif_Directive_F_Then_Kw
     (Node : Bare_Pp_Elsif_Directive) return Bare_Pp_Then_Kw
   is
   begin
         return Node.Pp_Elsif_Directive_F_Then_Kw;
   end;
   --
   --  Primitives for Bare_Pp_End_If_Directive
   --
   --
   --  Primitives for Bare_Pp_If_Directive
   --
      procedure Initialize_Fields_For_Pp_If_Directive
        (Self : Bare_Pp_If_Directive
         ; Pp_If_Directive_F_Expr : Bare_Expr
         ; Pp_If_Directive_F_Then_Kw : Bare_Pp_Then_Kw
        ) is
      begin
            Self.Pp_If_Directive_F_Expr := Pp_If_Directive_F_Expr;
            Self.Pp_If_Directive_F_Then_Kw := Pp_If_Directive_F_Then_Kw;
      end Initialize_Fields_For_Pp_If_Directive;
   function Pp_If_Directive_F_Expr
     (Node : Bare_Pp_If_Directive) return Bare_Expr
   is
   begin
         return Node.Pp_If_Directive_F_Expr;
   end;
   function Pp_If_Directive_F_Then_Kw
     (Node : Bare_Pp_If_Directive) return Bare_Pp_Then_Kw
   is
   begin
         return Node.Pp_If_Directive_F_Then_Kw;
   end;
   --
   --  Primitives for Bare_Pp_Then_Kw
   --
   --
   --  Primitives for Bare_Pragma_Node
   --
      procedure Pragma_Pre_Env_Actions
        (Self            : Bare_Pragma_Node;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Env : constant Internal_Designated_Env :=
           Libadalang.Implementation.Initial_Env_36 (Node => 
Self);
      begin
         Set_Initial_Env
           (Self,
            State,
            Env,
            DSL_Location => "ast.py:9209");
      end;
      end;
      procedure Initialize_Fields_For_Pragma_Node
        (Self : Bare_Pragma_Node
         ; Pragma_Node_F_Id : Bare_Identifier
         ; Pragma_Node_F_Args : Bare_Base_Assoc_List
        ) is
      begin
            Self.Pragma_Node_F_Id := Pragma_Node_F_Id;
            Self.Pragma_Node_F_Args := Pragma_Node_F_Args;
      end Initialize_Fields_For_Pragma_Node;
   function Pragma_Node_F_Id
     (Node : Bare_Pragma_Node) return Bare_Identifier
   is
   begin
         return Node.Pragma_Node_F_Id;
   end;
   function Pragma_Node_F_Args
     (Node : Bare_Pragma_Node) return Bare_Base_Assoc_List
   is
   begin
         return Node.Pragma_Node_F_Args;
   end;
--# property-start Pragma.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8928
pragma Warnings (Off, "is not referenced");
 function Pragma_Node_P_Xref_Entry_Point
  (Node : Bare_Pragma_Node
  )
   return Boolean
is
   Self : Bare_Pragma_Node := Bare_Pragma_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Pragma_Node_P_Xref_Entry_Point;
--# end
--# property-start Pragma.is_ghost_code /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8931
pragma Warnings (Off, "is not referenced");
 function Pragma_Node_P_Is_Ghost_Code
  (Node : Bare_Pragma_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Pragma_Node := Bare_Pragma_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Pragma_Node :=
     Internal_Entity_Pragma_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Fld : Internal_Entity_Identifier;
Fld_1 : Symbol_Type;
Sym : Symbol_Type;
Is_Equal : Boolean;
Fld_2 : Internal_Entity_Identifier;
Fld_3 : Symbol_Type;
Sym_1 : Symbol_Type;
Is_Equal_1 : Boolean;
If_Result : Boolean;
Fld_4 : Internal_Entity_Identifier;
Fld_5 : Symbol_Type;
Sym_2 : Symbol_Type;
Is_Equal_2 : Boolean;
If_Result_1 : Boolean;
Fld_6 : Internal_Entity_Identifier;
Fld_7 : Symbol_Type;
Sym_3 : Symbol_Type;
Is_Equal_3 : Boolean;
If_Result_2 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 10282 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10269 '<Eq at ast.py:8937>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8937
--# expr-start 10267 '<FieldAccess for sym at ast.py:8937>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8937
--# expr-start 10266 '<FieldAccess for id at ast.py:8937>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8937
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10266
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTokNode.sym
Fld_1 := Libadalang.Implementation.Single_Tok_Node_P_Sym (Node => Fld.Node);
--# end
--# expr-done 10267
--# expr-start 10268 '<Symbol Assert at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Assert); 
--# expr-done 10268
Is_Equal := Fld_1 = Sym; 
--# expr-done 10269
if Is_Equal then
   If_Result := True;
else
   --# expr-start 10273 '<Eq at ast.py:8937>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8937
--# expr-start 10271 '<FieldAccess for sym at ast.py:8937>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8937
--# expr-start 10270 '<FieldAccess for id at ast.py:8937>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8937
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10270
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTokNode.sym
Fld_3 := Libadalang.Implementation.Single_Tok_Node_P_Sym (Node => Fld_2.Node);
--# end
--# expr-done 10271
--# expr-start 10272 '<Symbol Assert_And_Cut at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Assert_And_Cut); 
--# expr-done 10272
Is_Equal_1 := Fld_3 = Sym_1; 
--# expr-done 10273
   If_Result := Is_Equal_1;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 10277 '<Eq at ast.py:8937>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8937
--# expr-start 10275 '<FieldAccess for sym at ast.py:8937>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8937
--# expr-start 10274 '<FieldAccess for id at ast.py:8937>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8937
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10274
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTokNode.sym
Fld_5 := Libadalang.Implementation.Single_Tok_Node_P_Sym (Node => Fld_4.Node);
--# end
--# expr-done 10275
--# expr-start 10276 '<Symbol Assume at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Assume); 
--# expr-done 10276
Is_Equal_2 := Fld_5 = Sym_2; 
--# expr-done 10277
   If_Result_1 := Is_Equal_2;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
   --# expr-start 10281 '<Eq at ast.py:8937>' Is_Equal_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8937
--# expr-start 10279 '<FieldAccess for sym at ast.py:8937>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8937
--# expr-start 10278 '<FieldAccess for id at ast.py:8937>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8937
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10278
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SingleTokNode.sym
Fld_7 := Libadalang.Implementation.Single_Tok_Node_P_Sym (Node => Fld_6.Node);
--# end
--# expr-done 10279
--# expr-start 10280 '<Symbol Loop_Invariant at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Loop_Invariant); 
--# expr-done 10280
Is_Equal_3 := Fld_7 = Sym_3; 
--# expr-done 10281
   If_Result_2 := Is_Equal_3;
end if;
Let_Result := If_Result_2; 
--# end
--# expr-done 10282
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Pragma_Node_P_Is_Ghost_Code;
--# end
--# property-start Pragma.xref_initial_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8942
pragma Warnings (Off, "is not referenced");
 function Pragma_Node_P_Xref_Initial_Env
  (Node : Bare_Pragma_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Pragma_Node := Bare_Pragma_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Pragma_Node :=
     Internal_Entity_Pragma_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_729 with Inline_Always;
      Fld : Internal_Entity_Defining_Name_Array_Access;
Get_Result : Internal_Entity_Defining_Name;
Fld_1 : Lexical_Env := Empty_Env;
Fld_2 : Lexical_Env := Empty_Env;
Fld_3 : Internal_Entity_Identifier;
Fld_4 : Symbol_Type;
Sym : Symbol_Type;
Is_Equal : Boolean;
Fld_5 : Internal_Entity_Identifier;
Fld_6 : Symbol_Type;
Sym_1 : Symbol_Type;
Is_Equal_1 : Boolean;
If_Result : Boolean;
Fld_7 : Internal_Entity_Identifier;
Fld_8 : Symbol_Type;
Sym_2 : Symbol_Type;
Is_Equal_2 : Boolean;
If_Result_1 : Boolean;
Fld_9 : Internal_Entity_Identifier;
Fld_10 : Symbol_Type;
Sym_3 : Symbol_Type;
Is_Equal_3 : Boolean;
If_Result_2 : Boolean;
Fld_11 : Internal_Entity_Identifier;
Fld_12 : Symbol_Type;
Sym_4 : Symbol_Type;
Is_Equal_4 : Boolean;
If_Result_3 : Boolean;
Fld_13 : Internal_Entity_Identifier;
Fld_14 : Symbol_Type;
Sym_5 : Symbol_Type;
Is_Equal_5 : Boolean;
If_Result_4 : Boolean;
Fld_15 : Internal_Entity_Identifier;
Fld_16 : Symbol_Type;
Sym_6 : Symbol_Type;
Is_Equal_6 : Boolean;
If_Result_5 : Boolean;
Fld_17 : Internal_Entity_Identifier;
Fld_18 : Symbol_Type;
Sym_7 : Symbol_Type;
Is_Equal_7 : Boolean;
If_Result_6 : Boolean;
Fld_19 : Internal_Entity_Identifier;
Fld_20 : Symbol_Type;
Sym_8 : Symbol_Type;
Is_Equal_8 : Boolean;
If_Result_7 : Boolean;
Fld_21 : Internal_Entity_Identifier;
Fld_22 : Symbol_Type;
Sym_9 : Symbol_Type;
Is_Equal_9 : Boolean;
If_Result_8 : Boolean;
If_Result_9 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_729 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Fld_2);
                     Dec_Ref (If_Result_9);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_729;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10328 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10327 '<If at ast.py:8950>' If_Result_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8950
--# expr-start 10286 '<Eq at ast.py:8951>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10284 '<FieldAccess for name_symbol at ast.py:8951>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10283 '<FieldAccess for id at ast.py:8951>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10283
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_4 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_3.Node);
--# end
--# expr-done 10284
--# expr-start 10285 '<Symbol Pre at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Pre); 
--# expr-done 10285
Is_Equal := Fld_4 = Sym; 
--# expr-done 10286
if Is_Equal then
   If_Result := True;
else
   --# expr-start 10290 '<Eq at ast.py:8951>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10288 '<FieldAccess for name_symbol at ast.py:8951>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10287 '<FieldAccess for id at ast.py:8951>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10287
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_6 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_5.Node);
--# end
--# expr-done 10288
--# expr-start 10289 '<Symbol Post at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Post); 
--# expr-done 10289
Is_Equal_1 := Fld_6 = Sym_1; 
--# expr-done 10290
   If_Result := Is_Equal_1;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 10294 '<Eq at ast.py:8951>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10292 '<FieldAccess for name_symbol at ast.py:8951>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10291 '<FieldAccess for id at ast.py:8951>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10291
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_8 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_7.Node);
--# end
--# expr-done 10292
--# expr-start 10293 '<Symbol Pre'"'"'Class at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Pre_Class); 
--# expr-done 10293
Is_Equal_2 := Fld_8 = Sym_2; 
--# expr-done 10294
   If_Result_1 := Is_Equal_2;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
   --# expr-start 10298 '<Eq at ast.py:8951>' Is_Equal_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10296 '<FieldAccess for name_symbol at ast.py:8951>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10295 '<FieldAccess for id at ast.py:8951>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10295
      if Fld_9.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_10 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_9.Node);
--# end
--# expr-done 10296
--# expr-start 10297 '<Symbol Post'"'"'Class at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Post_Class); 
--# expr-done 10297
Is_Equal_3 := Fld_10 = Sym_3; 
--# expr-done 10298
   If_Result_2 := Is_Equal_3;
end if;
if If_Result_2 then
   If_Result_3 := True;
else
   --# expr-start 10302 '<Eq at ast.py:8951>' Is_Equal_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10300 '<FieldAccess for name_symbol at ast.py:8951>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10299 '<FieldAccess for id at ast.py:8951>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10299
      if Fld_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_12 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_11.Node);
--# end
--# expr-done 10300
--# expr-start 10301 '<Symbol Precondition at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Precondition); 
--# expr-done 10301
Is_Equal_4 := Fld_12 = Sym_4; 
--# expr-done 10302
   If_Result_3 := Is_Equal_4;
end if;
if If_Result_3 then
   If_Result_4 := True;
else
   --# expr-start 10306 '<Eq at ast.py:8951>' Is_Equal_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10304 '<FieldAccess for name_symbol at ast.py:8951>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10303 '<FieldAccess for id at ast.py:8951>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10303
      if Fld_13.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_14 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_13.Node);
--# end
--# expr-done 10304
--# expr-start 10305 '<Symbol Postcondition at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Postcondition); 
--# expr-done 10305
Is_Equal_5 := Fld_14 = Sym_5; 
--# expr-done 10306
   If_Result_4 := Is_Equal_5;
end if;
if If_Result_4 then
   If_Result_5 := True;
else
   --# expr-start 10310 '<Eq at ast.py:8951>' Is_Equal_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10308 '<FieldAccess for name_symbol at ast.py:8951>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10307 '<FieldAccess for id at ast.py:8951>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10307
      if Fld_15.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_16 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_15.Node);
--# end
--# expr-done 10308
--# expr-start 10309 '<Symbol Precondition'"'"'Class at ???>' Sym_6 None
Sym_6 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Precondition_Class); 
--# expr-done 10309
Is_Equal_6 := Fld_16 = Sym_6; 
--# expr-done 10310
   If_Result_5 := Is_Equal_6;
end if;
if If_Result_5 then
   If_Result_6 := True;
else
   --# expr-start 10314 '<Eq at ast.py:8951>' Is_Equal_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10312 '<FieldAccess for name_symbol at ast.py:8951>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10311 '<FieldAccess for id at ast.py:8951>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_17 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10311
      if Fld_17.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_18 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_17.Node);
--# end
--# expr-done 10312
--# expr-start 10313 '<Symbol Postcondition'"'"'Class at ???>' Sym_7 None
Sym_7 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Postcondition_Class); 
--# expr-done 10313
Is_Equal_7 := Fld_18 = Sym_7; 
--# expr-done 10314
   If_Result_6 := Is_Equal_7;
end if;
if If_Result_6 then
   If_Result_7 := True;
else
   --# expr-start 10318 '<Eq at ast.py:8951>' Is_Equal_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10316 '<FieldAccess for name_symbol at ast.py:8951>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10315 '<FieldAccess for id at ast.py:8951>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10315
      if Fld_19.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_20 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_19.Node);
--# end
--# expr-done 10316
--# expr-start 10317 '<Symbol Test_Case at ???>' Sym_8 None
Sym_8 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Test_Case); 
--# expr-done 10317
Is_Equal_8 := Fld_20 = Sym_8; 
--# expr-done 10318
   If_Result_7 := Is_Equal_8;
end if;
if If_Result_7 then
   If_Result_8 := True;
else
   --# expr-start 10322 '<Eq at ast.py:8951>' Is_Equal_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10320 '<FieldAccess for name_symbol at ast.py:8951>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
--# expr-start 10319 '<FieldAccess for id at ast.py:8951>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8951
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_21 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10319
      if Fld_21.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_22 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_21.Node);
--# end
--# expr-done 10320
--# expr-start 10321 '<Symbol Contract_Cases at ???>' Sym_9 None
Sym_9 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Contract_Cases); 
--# expr-done 10321
Is_Equal_9 := Fld_22 = Sym_9; 
--# expr-done 10322
   If_Result_8 := Is_Equal_9;
end if;
if If_Result_8 then
   --# expr-start 10325 '<FieldAccess for children_env at ast.py:8957>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8957
--# expr-start 10324 '<at at ast.py:8957>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8957
--# expr-start 10323 '<FieldAccess for associated_entities at ast.py:8957>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8957
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Pragma.associated_entities
Fld := Libadalang.Implementation.Pragma_Node_P_Associated_Entities (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 10323
Get_Result := Get (Fld, 0, True); 
--# expr-done 10324
      if Get_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Children_Env (Node => Get_Result.Node, E_Info => Get_Result.Info);
--# expr-done 10325
   If_Result_9 := Fld_1;
else
   --# expr-start 10326 '<FieldAccess for children_env at ast.py:8958>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8958
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Children_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 10326
   If_Result_9 := Fld_2;
end if;
      Inc_Ref (If_Result_9);
--# expr-done 10327
Let_Result := If_Result_9; Inc_Ref (Let_Result);
--# end
--# expr-done 10328
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_729;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_729;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Pragma_Node_P_Xref_Initial_Env;
--# end
--# property-start Pragma.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8962
pragma Warnings (Off, "is not referenced");
 function Pragma_Node_P_Xref_Equation
  (Node : Bare_Pragma_Node
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Pragma_Node := Bare_Pragma_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Pragma_Node :=
     Internal_Entity_Pragma_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_730 with Inline_Always;
            procedure Finalizer_Scope_3039 with Inline_Always;
            procedure Finalizer_Scope_3040 with Inline_Always;
            procedure Finalizer_Scope_3041 with Inline_Always;
            procedure Finalizer_Scope_3043 with Inline_Always;
            procedure Finalizer_Scope_3044 with Inline_Always;
            procedure Finalizer_Scope_3048 with Inline_Always;
            procedure Finalizer_Scope_3049 with Inline_Always;
            procedure Finalizer_Scope_3051 with Inline_Always;
      Expr : Internal_Entity_Expr;
Var_Expr : Internal_Entity_Expr;
Var_Expr_1 : Internal_Entity_Base_Assoc;
Var_Expr_2 : Internal_Entity_Expr;
Var_Expr_3 : Internal_Entity_Base_Assoc;
Var_Expr_4 : Internal_Entity_Expr;
Var_Expr_5 : Internal_Entity_Pragma_Argument_Assoc;
Expr_1 : Internal_Entity_Expr;
Proc : Internal_Entity_Expr;
Var_Expr_6 : Internal_Entity_Base_Assoc;
Fld : Internal_Entity_Identifier;
Sym : Symbol_Type;
Fld_1 : Boolean;
Fld_2 : Internal_Entity_Identifier;
Sym_1 : Symbol_Type;
Fld_3 : Boolean;
If_Result : Boolean;
Fld_4 : Internal_Entity_Identifier;
Sym_2 : Symbol_Type;
Fld_5 : Boolean;
If_Result_1 : Boolean;
Fld_6 : Internal_Entity_Identifier;
Sym_3 : Symbol_Type;
Fld_7 : Boolean;
If_Result_2 : Boolean;
Fld_8 : Internal_Entity_Base_Assoc_List;
Fld_9 : Bare_Base_Assoc_List;
Fld_10 : Internal_Entity_Info;
Get_Result : Bare_Base_Assoc;
As_Entity : Internal_Entity_Base_Assoc;
Is_Null : Boolean;
If_Result_3 : Internal_Entity_Base_Assoc;
Fld_11 : Internal_Entity_Expr;
Fld_12 : Logic_Var;
Fld_13 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_14 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_15 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
Fld_16 : Internal_Entity_Base_Assoc_List;
Fld_17 : Bare_Base_Assoc_List;
Fld_18 : Internal_Entity_Info;
Get_Result_1 : Bare_Base_Assoc;
As_Entity_1 : Internal_Entity_Base_Assoc;
Is_Null_1 : Boolean;
If_Result_4 : Internal_Entity_Base_Assoc;
Fld_19 : Internal_Entity_Expr;
Scope_Result : Internal_Entity_Expr;
Result_Var : Internal_Entity_Expr;
Fld_20 : Logic_Var;
Sym_4 : Symbol_Type;
Fld_21 : Internal_Entity;
Bind_Result_1 : Logic_Equation;
Fld_22 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_23 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Scope_Result_1 : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var_1 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Fld_24 : Internal_Entity_Identifier;
Sym_5 : Symbol_Type;
Fld_25 : Boolean;
Fld_26 : Internal_Entity_Base_Assoc_List;
Fld_27 : Bare_Base_Assoc_List;
Fld_28 : Internal_Entity_Info;
As_Entity_2 : Internal_Entity_Base_Assoc;
Is_Null_2 : Boolean;
If_Result_5 : Internal_Entity_Base_Assoc;
Fld_29 : Internal_Entity_Expr;
Cast_Result_1 : Internal_Entity_Name;
Fld_30 : Logic_Equation;
Item_309 : Internal_Entity_Base_Assoc;
Bare_Item_309 : Bare_Base_Assoc;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Fld_31 : Internal_Entity_Identifier;
Fld_32 : Symbol_Type;
Sym_6 : Symbol_Type;
Is_Equal : Boolean;
Fld_33 : Internal_Entity_Identifier;
Fld_34 : Symbol_Type;
Sym_7 : Symbol_Type;
Is_Equal_1 : Boolean;
If_Result_6 : Boolean;
Fld_35 : Internal_Entity_Identifier;
Fld_36 : Symbol_Type;
Sym_8 : Symbol_Type;
Is_Equal_2 : Boolean;
If_Result_7 : Boolean;
Fld_37 : Internal_Entity_Identifier;
Fld_38 : Symbol_Type;
Sym_9 : Symbol_Type;
Is_Equal_3 : Boolean;
If_Result_8 : Boolean;
Fld_39 : Internal_Entity_Identifier;
Fld_40 : Symbol_Type;
Sym_10 : Symbol_Type;
Is_Equal_4 : Boolean;
If_Result_9 : Boolean;
Fld_41 : Internal_Entity_Identifier;
Fld_42 : Symbol_Type;
Sym_11 : Symbol_Type;
Is_Equal_5 : Boolean;
If_Result_10 : Boolean;
Fld_43 : Internal_Entity_Identifier;
Fld_44 : Symbol_Type;
Sym_12 : Symbol_Type;
Is_Equal_6 : Boolean;
If_Result_11 : Boolean;
Fld_45 : Internal_Entity_Identifier;
Fld_46 : Symbol_Type;
Sym_13 : Symbol_Type;
Is_Equal_7 : Boolean;
If_Result_12 : Boolean;
Fld_47 : Internal_Entity_Identifier;
Fld_48 : Symbol_Type;
Sym_14 : Symbol_Type;
Is_Equal_8 : Boolean;
If_Result_13 : Boolean;
Fld_49 : Internal_Entity_Identifier;
Fld_50 : Symbol_Type;
Sym_15 : Symbol_Type;
Is_Equal_9 : Boolean;
If_Result_14 : Boolean;
Fld_51 : Internal_Entity_Name_Array_Access;
Fld_52 : Logic_Equation;
Map_Result_1 : Logic_Equation_Array_Access;
Logic_Boolean_Op_1 : Logic_Equation;
Fld_53 : Internal_Entity_Identifier;
Sym_16 : Symbol_Type;
Fld_54 : Boolean;
Fld_55 : Internal_Entity_Base_Assoc_List;
Fld_56 : Bare_Base_Assoc_List;
Fld_57 : Internal_Entity_Info;
Get_Result_2 : Bare_Base_Assoc;
As_Entity_3 : Internal_Entity_Base_Assoc;
Is_Null_3 : Boolean;
If_Result_15 : Internal_Entity_Base_Assoc;
Fld_58 : Internal_Entity_Expr;
Scope_Result_2 : Internal_Entity_Expr;
Result_Var_2 : Internal_Entity_Expr;
Fld_59 : Logic_Equation;
Fld_60 : Logic_Var;
Sym_17 : Symbol_Type;
Fld_61 : Internal_Entity;
Bind_Result_2 : Logic_Equation;
Fld_62 : Logic_Equation;
And_Pred_5 : Logic_Equation;
Fld_63 : Logic_Equation;
And_Pred_6 : Logic_Equation;
Is_A : Boolean;
If_Result_16 : Logic_Equation;
Scope_Result_3 : Logic_Equation;
True_Rel_1 : Logic_Equation;
Result_Var_3 : Logic_Equation;
Fld_64 : Internal_Entity_Identifier;
Fld_65 : Symbol_Type;
Sym_18 : Symbol_Type;
Is_Equal_10 : Boolean;
Fld_66 : Internal_Entity_Identifier;
Fld_67 : Symbol_Type;
Sym_19 : Symbol_Type;
Is_Equal_11 : Boolean;
If_Result_17 : Boolean;
Fld_68 : Internal_Entity_Identifier;
Fld_69 : Symbol_Type;
Sym_20 : Symbol_Type;
Is_Equal_12 : Boolean;
If_Result_18 : Boolean;
True_Rel_2 : Logic_Equation;
Fld_70 : Internal_Entity_Identifier;
Fld_71 : Symbol_Type;
Sym_21 : Symbol_Type;
Is_Equal_13 : Boolean;
Fld_72 : Internal_Entity_Identifier;
Fld_73 : Symbol_Type;
Sym_22 : Symbol_Type;
Is_Equal_14 : Boolean;
If_Result_19 : Boolean;
Fld_74 : Internal_Entity_Identifier;
Fld_75 : Symbol_Type;
Sym_23 : Symbol_Type;
Is_Equal_15 : Boolean;
If_Result_20 : Boolean;
Fld_76 : Internal_Entity_Identifier;
Fld_77 : Symbol_Type;
Sym_24 : Symbol_Type;
Is_Equal_16 : Boolean;
If_Result_21 : Boolean;
Fld_78 : Internal_Entity_Identifier;
Fld_79 : Symbol_Type;
Sym_25 : Symbol_Type;
Is_Equal_17 : Boolean;
If_Result_22 : Boolean;
Fld_80 : Internal_Entity_Identifier;
Fld_81 : Symbol_Type;
Sym_26 : Symbol_Type;
Is_Equal_18 : Boolean;
If_Result_23 : Boolean;
Fld_82 : Internal_Entity_Identifier;
Fld_83 : Symbol_Type;
Sym_27 : Symbol_Type;
Is_Equal_19 : Boolean;
If_Result_24 : Boolean;
Fld_84 : Internal_Entity_Identifier;
Fld_85 : Symbol_Type;
Sym_28 : Symbol_Type;
Is_Equal_20 : Boolean;
If_Result_25 : Boolean;
Fld_86 : Internal_Entity_Base_Assoc_List;
Fld_87 : Bare_Base_Assoc_List;
Fld_88 : Internal_Entity_Info;
Get_Result_3 : Bare_Base_Assoc;
As_Entity_4 : Internal_Entity_Base_Assoc;
Is_Null_4 : Boolean;
If_Result_26 : Internal_Entity_Base_Assoc;
Fld_89 : Internal_Entity_Expr;
Fld_90 : Logic_Var;
Fld_91 : Internal_Entity_Base_Type_Decl;
Cast_Result_2 : Internal_Entity;
Bind_Result_3 : Logic_Equation;
Fld_92 : Logic_Equation;
And_Pred_7 : Logic_Equation;
Fld_93 : Logic_Equation;
And_Pred_8 : Logic_Equation;
Scope_Result_4 : Logic_Equation;
False_Rel : Logic_Equation;
Result_Var_4 : Logic_Equation;
Fld_94 : Internal_Entity_Identifier;
Sym_29 : Symbol_Type;
Fld_95 : Boolean;
Fld_96 : Internal_Entity_Base_Assoc_List;
Fld_97 : Bare_Base_Assoc_List;
Fld_98 : Internal_Entity_Info;
As_Entity_5 : Internal_Entity_Base_Assoc;
Is_Null_5 : Boolean;
If_Result_27 : Internal_Entity_Base_Assoc;
Item_312 : Internal_Entity_Base_Assoc;
Bare_Item_312 : Bare_Base_Assoc;
Cast_Expr : Internal_Entity_Base_Assoc;
Cast_Result_3 : Internal_Entity_Pragma_Argument_Assoc;
Fld_99 : Internal_Entity_Name;
Fld_100 : Symbol_Type;
Sym_30 : Symbol_Type;
Is_Equal_21 : Boolean;
Fld_101 : Internal_Entity_Name;
Fld_102 : Symbol_Type;
Sym_31 : Symbol_Type;
Is_Equal_22 : Boolean;
If_Result_28 : Boolean;
Scope_Result_5 : Boolean;
Result_Var_5 : Boolean;
Map_Result_2 : Internal_Entity_Base_Assoc_Array_Access;
Fld_103 : Internal_Entity_Expr;
Fld_104 : Logic_Var;
Fld_105 : Internal_Entity_Base_Type_Decl;
Cast_Result_4 : Internal_Entity;
Bind_Result_4 : Logic_Equation;
Fld_106 : Internal_Entity_Expr;
Fld_107 : Logic_Equation;
And_Pred_9 : Logic_Equation;
Fld_108 : Internal_Entity_Expr;
Fld_109 : Logic_Equation;
And_Pred_10 : Logic_Equation;
Map_Result_3 : Logic_Equation_Array_Access;
Logic_Boolean_Op_2 : Logic_Equation;
Fld_110 : Internal_Entity_Identifier;
Sym_32 : Symbol_Type;
Fld_111 : Boolean;
Fld_112 : Internal_Entity_Base_Assoc_List;
Fld_113 : Bare_Base_Assoc_List;
Fld_114 : Internal_Entity_Info;
Get_Result_4 : Bare_Base_Assoc;
As_Entity_6 : Internal_Entity_Base_Assoc;
Is_Null_6 : Boolean;
If_Result_29 : Internal_Entity_Base_Assoc;
Fld_115 : Internal_Entity_Expr;
Cast_Result_5 : Internal_Entity_Base_Aggregate;
Fld_116 : Internal_Entity_Assoc_List;
Fld_117 : Bare_Assoc_List;
Fld_118 : Internal_Entity_Info;
As_Entity_7 : Internal_Entity_Basic_Assoc;
Is_Null_7 : Boolean;
If_Result_30 : Internal_Entity_Basic_Assoc;
Cast_Expr_1 : Internal_Entity_Basic_Assoc;
Cast_Result_6 : Internal_Entity_Aggregate_Assoc;
Fld_119 : Logic_Equation;
Item_313 : Internal_Entity_Basic_Assoc;
Bare_Item_313 : Bare_Basic_Assoc;
Map_Result_4 : Logic_Equation_Array_Access;
Logic_Boolean_Op_3 : Logic_Equation;
Fld_120 : Internal_Entity_Identifier;
Sym_33 : Symbol_Type;
Fld_121 : Boolean;
Fld_122 : Internal_Entity_Base_Assoc_List;
Fld_123 : Bare_Base_Assoc_List;
Fld_124 : Internal_Entity_Info;
Get_Result_5 : Bare_Base_Assoc;
As_Entity_8 : Internal_Entity_Base_Assoc;
Is_Null_8 : Boolean;
If_Result_31 : Internal_Entity_Base_Assoc;
Fld_125 : Internal_Entity_Expr;
Fld_126 : Logic_Var;
Fld_127 : Internal_Entity_Base_Type_Decl;
Cast_Result_7 : Internal_Entity;
Bind_Result_5 : Logic_Equation;
Fld_128 : Logic_Equation;
And_Pred_11 : Logic_Equation;
Fld_129 : Logic_Equation;
And_Pred_12 : Logic_Equation;
Let_Result_1 : Logic_Equation;
True_Rel_3 : Logic_Equation;
Fld_130 : Internal_Entity_Base_Assoc_List;
Fld_131 : Bare_Base_Assoc_List;
Len : Integer;
Is_Equal_23 : Boolean;
If_Result_32 : Logic_Equation;
Fld_132 : Internal_Entity_Base_Assoc_List;
Fld_133 : Bare_Base_Assoc_List;
Len_1 : Integer;
Arith_Result : Integer;
Fld_134 : Internal_Entity_Base_Assoc_List;
Fld_135 : Bare_Base_Assoc_List;
Fld_136 : Internal_Entity_Info;
Get_Result_6 : Bare_Base_Assoc;
As_Entity_9 : Internal_Entity_Base_Assoc;
Is_Null_9 : Boolean;
If_Result_33 : Internal_Entity_Base_Assoc;
Fld_137 : Internal_Entity_Expr;
Scope_Result_6 : Internal_Entity_Expr;
Result_Var_6 : Internal_Entity_Expr;
Fld_138 : Logic_Var;
Cast_Expr_2 : Internal_Entity_Base_Type_Decl;
Cast_Result_8 : Internal_Entity;
Bind_Result_6 : Logic_Equation;
Fld_139 : Logic_Equation;
And_Pred_13 : Logic_Equation;
Let_Result_2 : Logic_Equation;
And_Pred_14 : Logic_Equation;
Fld_140 : Internal_Entity_Base_Assoc_List;
Fld_141 : Bare_Base_Assoc_List;
Fld_142 : Internal_Entity_Info;
As_Entity_10 : Internal_Entity_Base_Assoc;
Is_Null_10 : Boolean;
If_Result_34 : Internal_Entity_Base_Assoc;
Fld_143 : Internal_Entity_Expr;
Fld_144 : Logic_Equation;
True_Rel_4 : Logic_Equation;
Or_Pred : Logic_Equation;
Item_314 : Internal_Entity_Base_Assoc;
Bare_Item_314 : Bare_Base_Assoc;
Map_Result_5 : Logic_Equation_Array_Access;
Logic_Boolean_Op_4 : Logic_Equation;
If_Result_35 : Logic_Equation;
If_Result_36 : Logic_Equation;
If_Result_37 : Logic_Equation;
If_Result_38 : Logic_Equation;
If_Result_39 : Logic_Equation;
If_Result_40 : Logic_Equation;
If_Result_41 : Logic_Equation;
If_Result_42 : Logic_Equation;
If_Result_43 : Logic_Equation;
Let_Result_3 : Logic_Equation;
            procedure Finalizer_Scope_730 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_14);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_15);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Fld_51);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Logic_Boolean_Op_1);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (True_Rel_1);
                     Dec_Ref (Result_Var_3);
                     Dec_Ref (True_Rel_2);
                     Dec_Ref (Scope_Result_4);
                     Dec_Ref (False_Rel);
                     Dec_Ref (Result_Var_4);
                     Dec_Ref (Map_Result_2);
                     Dec_Ref (Map_Result_3);
                     Dec_Ref (Logic_Boolean_Op_2);
                     Dec_Ref (Map_Result_4);
                     Dec_Ref (Logic_Boolean_Op_3);
                     Dec_Ref (Bind_Result_5);
                     Dec_Ref (Fld_128);
                     Dec_Ref (And_Pred_11);
                     Dec_Ref (Fld_129);
                     Dec_Ref (And_Pred_12);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (True_Rel_3);
                     Dec_Ref (If_Result_32);
                     Dec_Ref (Bind_Result_6);
                     Dec_Ref (Fld_139);
                     Dec_Ref (And_Pred_13);
                     Dec_Ref (Let_Result_2);
                     Dec_Ref (And_Pred_14);
                     Dec_Ref (Map_Result_5);
                     Dec_Ref (Logic_Boolean_Op_4);
                     Dec_Ref (If_Result_35);
                     Dec_Ref (If_Result_36);
                     Dec_Ref (If_Result_37);
                     Dec_Ref (If_Result_38);
                     Dec_Ref (If_Result_39);
                     Dec_Ref (If_Result_40);
                     Dec_Ref (If_Result_41);
                     Dec_Ref (If_Result_42);
                     Dec_Ref (If_Result_43);
                     Dec_Ref (Let_Result_3);
            end Finalizer_Scope_730;
            procedure Finalizer_Scope_3039 is
            begin
                     Dec_Ref (Bind_Result_1);
                     Dec_Ref (Fld_22);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Fld_23);
                     Dec_Ref (And_Pred_3);
            end Finalizer_Scope_3039;
            procedure Finalizer_Scope_3040 is
            begin
                     Dec_Ref (Fld_30);
            end Finalizer_Scope_3040;
            procedure Finalizer_Scope_3041 is
            begin
                     Dec_Ref (Fld_52);
            end Finalizer_Scope_3041;
            procedure Finalizer_Scope_3043 is
            begin
                     Dec_Ref (Fld_59);
                     Dec_Ref (Bind_Result_2);
                     Dec_Ref (Fld_62);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (Fld_63);
                     Dec_Ref (And_Pred_6);
                     Dec_Ref (If_Result_16);
            end Finalizer_Scope_3043;
            procedure Finalizer_Scope_3044 is
            begin
                     Dec_Ref (Bind_Result_3);
                     Dec_Ref (Fld_92);
                     Dec_Ref (And_Pred_7);
                     Dec_Ref (Fld_93);
                     Dec_Ref (And_Pred_8);
            end Finalizer_Scope_3044;
            procedure Finalizer_Scope_3048 is
            begin
                     Dec_Ref (Bind_Result_4);
                     Dec_Ref (Fld_107);
                     Dec_Ref (And_Pred_9);
                     Dec_Ref (Fld_109);
                     Dec_Ref (And_Pred_10);
            end Finalizer_Scope_3048;
            procedure Finalizer_Scope_3049 is
            begin
                     Dec_Ref (Fld_119);
            end Finalizer_Scope_3049;
            procedure Finalizer_Scope_3051 is
            begin
                     Dec_Ref (Fld_144);
                     Dec_Ref (True_Rel_4);
                     Dec_Ref (Or_Pred);
            end Finalizer_Scope_3051;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10556 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 10555 '<Cond at ast.py:8963>' If_Result_43 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8963
--# expr-start 10331 '<FieldAccess for name_is at ast.py:8965>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8965
--# expr-start 10329 '<FieldAccess for id at ast.py:8965>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8965
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10329
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10330 '<Symbol Assert at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Assert); 
--# expr-done 10330
--# property-call-start Name.name_is
Fld_1 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld.Node, Sym => Sym);
--# end
--# expr-done 10331
if Fld_1 then
   If_Result := True;
else
   --# expr-start 10334 '<FieldAccess for name_is at ast.py:8966>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8966
--# expr-start 10332 '<FieldAccess for id at ast.py:8966>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8966
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10332
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10333 '<Symbol Loop_Invariant at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Loop_Invariant); 
--# expr-done 10333
--# property-call-start Name.name_is
Fld_3 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_2.Node, Sym => Sym_1);
--# end
--# expr-done 10334
   If_Result := Fld_3;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 10337 '<FieldAccess for name_is at ast.py:8967>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8967
--# expr-start 10335 '<FieldAccess for id at ast.py:8967>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8967
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10335
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10336 '<Symbol Compile_Time_Warning at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Compile_Time_Warning); 
--# expr-done 10336
--# property-call-start Name.name_is
Fld_5 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_4.Node, Sym => Sym_2);
--# end
--# expr-done 10337
   If_Result_1 := Fld_5;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
   --# expr-start 10340 '<FieldAccess for name_is at ast.py:8968>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8968
--# expr-start 10338 '<FieldAccess for id at ast.py:8968>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8968
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10338
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10339 '<Symbol Compile_Time_Error at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Compile_Time_Error); 
--# expr-done 10339
--# property-call-start Name.name_is
Fld_7 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_6.Node, Sym => Sym_3);
--# end
--# expr-done 10340
   If_Result_2 := Fld_7;
end if;
if If_Result_2 then
   --# expr-start 10361 '<And at ast.py:8970>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8970
--# expr-start 10350 '<Let at ast.py:8970>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8970
--# scope-start
--# expr-start 10342 '<FieldAccess for assoc_expr at ast.py:8971>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8971
--# expr-start 10341 '<FieldAccess for args at ast.py:8971>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8971
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Base_Assoc_List (Node => Ent.Node.Pragma_Node_F_Args, Info => Ent.Info);
--# expr-done 10341
Fld_9 := Fld_8.Node;
Get_Result := Get (Fld_9, 0, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result_3 := No_Entity_Base_Assoc;
else
Fld_10 := Fld_8.Info;
As_Entity := (Info => Fld_10, Node => Get_Result); 
   If_Result_3 := As_Entity;
end if;
      if If_Result_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseAssoc.assoc_expr'
Fld_11 := Libadalang.Implementation.Dispatcher_Base_Assoc_P_Assoc_Expr (Node => If_Result_3.Node, E_Info => If_Result_3.Info);
--# end
--# expr-done 10342
Expr := Fld_11; 
--# bind expr Expr
--# expr-start 10349 '<And at ast.py:8972>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8972
--# expr-start 10347 '<And at ast.py:8972>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8972
--# expr-start 10345 '<Bind at ast.py:8972>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8972
--# expr-start 10343 '<FieldAccess for expected_type_var at ast.py:8972>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8972
      if Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Expr.Node);
--# expr-done 10343
Fld_12.Value := No_Entity;
Entity_Vars.Reset (Fld_12);
--# expr-start 10344 '<FieldAccess for bool_type at ast.py:8972>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8972
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_13 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 10344
      Cast_Result := Create_Internal_Entity
        (Node => Fld_13.Node,
         Info => Fld_13.Info);
Bind_Result := Solver.Create_Assign (Fld_12, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8972") else null)); 
--# expr-done 10345
--# expr-start 10346 '<FieldAccess for sub_equation at ast.py:8973>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8973
      if Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_14 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Expr.Node, Env => Env, Origin => Origin, E_Info => Expr.Info);
--# end
--# expr-done 10346
And_Pred := Create_And (Bind_Result, Fld_14, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8972") else null)); 
--# expr-done 10347
--# expr-start 10348 '<FieldAccess for matches_expected_formal_prim_type at ast.py:8974>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8974
      if Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_prim_type
Fld_15 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Prim_Type (Node => Expr.Node, Origin => Origin);
--# end
--# expr-done 10348
And_Pred_1 := Create_And (And_Pred, Fld_15, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8972") else null)); 
--# expr-done 10349
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 10350
--# expr-start 10351 '<FieldAccess for args at ast.py:8975>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8975
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_16 := Create_Internal_Entity_Base_Assoc_List (Node => Ent.Node.Pragma_Node_F_Args, Info => Ent.Info);
--# expr-done 10351
Fld_17 := Fld_16.Node;
Get_Result_1 := Get (Fld_17, 1, True); 
Is_Null_1 := Get_Result_1 = null; 
if Is_Null_1 then
   If_Result_4 := No_Entity_Base_Assoc;
else
Fld_18 := Fld_16.Info;
As_Entity_1 := (Info => Fld_18, Node => Get_Result_1); 
   If_Result_4 := As_Entity_1;
end if;
Var_Expr_1 := If_Result_4;
if Var_Expr_1 /= No_Entity_Base_Assoc then
   --# scope-start
--# expr-start 10352 '<FieldAccess for assoc_expr at ast.py:8975>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8975
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseAssoc.assoc_expr'
Fld_19 := Libadalang.Implementation.Dispatcher_Base_Assoc_P_Assoc_Expr (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 10352
Scope_Result := Fld_19;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Expr;
end if;
Var_Expr := Result_Var;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind msg Var_Expr
--# expr-start 10360 '<And at ast.py:8977>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8977
--# expr-start 10358 '<And at ast.py:8977>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8977
--# expr-start 10356 '<Bind at ast.py:8977>' Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8977
--# expr-start 10353 '<FieldAccess for expected_type_var at ast.py:8977>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8977
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_20 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Var_Expr.Node);
--# expr-done 10353
Fld_20.Value := No_Entity;
Entity_Vars.Reset (Fld_20);
--# expr-start 10355 '<FieldAccess for std_entity at ast.py:8977>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8977
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10354 '<Symbol String at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_String); 
--# expr-done 10354
--# property-call-start AdaNode.std_entity
Fld_21 := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym_4);
--# end
--# expr-done 10355
Bind_Result_1 := Solver.Create_Assign (Fld_20, Fld_21, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8977") else null)); 
--# expr-done 10356
--# expr-start 10357 '<FieldAccess for sub_equation at ast.py:8978>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8978
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_22 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 10357
And_Pred_2 := Create_And (Bind_Result_1, Fld_22, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8977") else null)); 
--# expr-done 10358
--# expr-start 10359 '<FieldAccess for matches_expected_type at ast.py:8979>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8979
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_23 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Var_Expr.Node, Origin => Origin);
--# end
--# expr-done 10359
And_Pred_3 := Create_And (And_Pred_2, Fld_23, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8977") else null)); 
--# expr-done 10360
Scope_Result_1 := And_Pred_3;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_3039;
   Result_Var_1 := Scope_Result_1;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8980") else null)); 
   Result_Var_1 := True_Rel;
end if;
      Inc_Ref (Result_Var_1);
And_Pred_4 := Create_And (Let_Result, Result_Var_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8970") else null)); 
--# expr-done 10361
   If_Result_43 := And_Pred_4;
else
--# expr-start 10364 '<FieldAccess for name_is at ast.py:8983>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8983
--# expr-start 10362 '<FieldAccess for id at ast.py:8983>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8983
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_24 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10362
      if Fld_24.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10363 '<Symbol Unreferenced at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Unreferenced); 
--# expr-done 10363
--# property-call-start Name.name_is
Fld_25 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_24.Node, Sym => Sym_5);
--# end
--# expr-done 10364
if Fld_25 then
   --# expr-start 10370 '<LogicAll at ast.py:8984>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8984
--# expr-start 10369 '<Map at ast.py:8984>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8984
   --# expr-start 10365 '<FieldAccess for args at ast.py:8984>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8984
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_26 := Create_Internal_Entity_Base_Assoc_List (Node => Ent.Node.Pragma_Node_F_Args, Info => Ent.Info);
--# expr-done 10365
Fld_27 := Fld_26.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_27 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Base_Assoc_List := Fld_27;
      begin
         for Untyped_Item_309 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_309 := Untyped_Item_309; 
Is_Null_2 := Bare_Item_309 = null; 
if Is_Null_2 then
   If_Result_5 := No_Entity_Base_Assoc;
else
Fld_28 := Fld_26.Info;
As_Entity_2 := (Info => Fld_28, Node => Bare_Item_309); 
   If_Result_5 := As_Entity_2;
end if;
                  Item_309 := If_Result_5; 
   --# scope-start
               --# bind assoc Item_309
      --# expr-start 10368 '<FieldAccess for xref_no_overloading at ast.py:8986>' Fld_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8986
--# expr-start 10367 '<Cast to Name at ast.py:8986>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8986
--# expr-start 10366 '<FieldAccess for assoc_expr at ast.py:8986>' Fld_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8986
      if Item_309.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseAssoc.assoc_expr'
Fld_29 := Libadalang.Implementation.Dispatcher_Base_Assoc_P_Assoc_Expr (Node => Item_309.Node, E_Info => Item_309.Info);
--# end
--# expr-done 10366
   if Fld_29.Node = null
      or else Fld_29.Node.Kind in Ada_Name
   then
      Cast_Result_1 := Create_Internal_Entity_Name
        (Node => Fld_29.Node,
         Info => Fld_29.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 10367
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_30 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Cast_Result_1.Node, Env => Env, Origin => Origin, E_Info => Cast_Result_1.Info);
--# end
--# expr-done 10368
         declare
            Item_To_Append : constant Logic_Equation := Fld_30;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_3040;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 10369
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8984") else null)); 
--# expr-done 10370
   If_Result_42 := Logic_Boolean_Op;
else
--# expr-start 10374 '<Eq at ast.py:8989>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10372 '<FieldAccess for name_symbol at ast.py:8989>' Fld_32 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10371 '<FieldAccess for id at ast.py:8989>' Fld_31 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_31 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10371
      if Fld_31.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_32 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_31.Node);
--# end
--# expr-done 10372
--# expr-start 10373 '<Symbol Import at ???>' Sym_6 None
Sym_6 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Import); 
--# expr-done 10373
Is_Equal := Fld_32 = Sym_6; 
--# expr-done 10374
if Is_Equal then
   If_Result_6 := True;
else
   --# expr-start 10378 '<Eq at ast.py:8989>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10376 '<FieldAccess for name_symbol at ast.py:8989>' Fld_34 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10375 '<FieldAccess for id at ast.py:8989>' Fld_33 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_33 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10375
      if Fld_33.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_34 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_33.Node);
--# end
--# expr-done 10376
--# expr-start 10377 '<Symbol Export at ???>' Sym_7 None
Sym_7 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Export); 
--# expr-done 10377
Is_Equal_1 := Fld_34 = Sym_7; 
--# expr-done 10378
   If_Result_6 := Is_Equal_1;
end if;
if If_Result_6 then
   If_Result_7 := True;
else
   --# expr-start 10382 '<Eq at ast.py:8989>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10380 '<FieldAccess for name_symbol at ast.py:8989>' Fld_36 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10379 '<FieldAccess for id at ast.py:8989>' Fld_35 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_35 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10379
      if Fld_35.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_36 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_35.Node);
--# end
--# expr-done 10380
--# expr-start 10381 '<Symbol Interface at ???>' Sym_8 None
Sym_8 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Interface); 
--# expr-done 10381
Is_Equal_2 := Fld_36 = Sym_8; 
--# expr-done 10382
   If_Result_7 := Is_Equal_2;
end if;
if If_Result_7 then
   If_Result_8 := True;
else
   --# expr-start 10386 '<Eq at ast.py:8989>' Is_Equal_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10384 '<FieldAccess for name_symbol at ast.py:8989>' Fld_38 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10383 '<FieldAccess for id at ast.py:8989>' Fld_37 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_37 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10383
      if Fld_37.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_38 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_37.Node);
--# end
--# expr-done 10384
--# expr-start 10385 '<Symbol Convention at ???>' Sym_9 None
Sym_9 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Convention); 
--# expr-done 10385
Is_Equal_3 := Fld_38 = Sym_9; 
--# expr-done 10386
   If_Result_8 := Is_Equal_3;
end if;
if If_Result_8 then
   If_Result_9 := True;
else
   --# expr-start 10390 '<Eq at ast.py:8989>' Is_Equal_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10388 '<FieldAccess for name_symbol at ast.py:8989>' Fld_40 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10387 '<FieldAccess for id at ast.py:8989>' Fld_39 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_39 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10387
      if Fld_39.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_40 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_39.Node);
--# end
--# expr-done 10388
--# expr-start 10389 '<Symbol Pack at ???>' Sym_10 None
Sym_10 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Pack); 
--# expr-done 10389
Is_Equal_4 := Fld_40 = Sym_10; 
--# expr-done 10390
   If_Result_9 := Is_Equal_4;
end if;
if If_Result_9 then
   If_Result_10 := True;
else
   --# expr-start 10394 '<Eq at ast.py:8989>' Is_Equal_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10392 '<FieldAccess for name_symbol at ast.py:8989>' Fld_42 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10391 '<FieldAccess for id at ast.py:8989>' Fld_41 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_41 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10391
      if Fld_41.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_42 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_41.Node);
--# end
--# expr-done 10392
--# expr-start 10393 '<Symbol Pure at ???>' Sym_11 None
Sym_11 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Pure); 
--# expr-done 10393
Is_Equal_5 := Fld_42 = Sym_11; 
--# expr-done 10394
   If_Result_10 := Is_Equal_5;
end if;
if If_Result_10 then
   If_Result_11 := True;
else
   --# expr-start 10398 '<Eq at ast.py:8989>' Is_Equal_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10396 '<FieldAccess for name_symbol at ast.py:8989>' Fld_44 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10395 '<FieldAccess for id at ast.py:8989>' Fld_43 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_43 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10395
      if Fld_43.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_44 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_43.Node);
--# end
--# expr-done 10396
--# expr-start 10397 '<Symbol Preelaborate at ???>' Sym_12 None
Sym_12 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Preelaborate); 
--# expr-done 10397
Is_Equal_6 := Fld_44 = Sym_12; 
--# expr-done 10398
   If_Result_11 := Is_Equal_6;
end if;
if If_Result_11 then
   If_Result_12 := True;
else
   --# expr-start 10402 '<Eq at ast.py:8989>' Is_Equal_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10400 '<FieldAccess for name_symbol at ast.py:8989>' Fld_46 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10399 '<FieldAccess for id at ast.py:8989>' Fld_45 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_45 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10399
      if Fld_45.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_46 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_45.Node);
--# end
--# expr-done 10400
--# expr-start 10401 '<Symbol Elaborate_Body at ???>' Sym_13 None
Sym_13 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Elaborate_Body); 
--# expr-done 10401
Is_Equal_7 := Fld_46 = Sym_13; 
--# expr-done 10402
   If_Result_12 := Is_Equal_7;
end if;
if If_Result_12 then
   If_Result_13 := True;
else
   --# expr-start 10406 '<Eq at ast.py:8989>' Is_Equal_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10404 '<FieldAccess for name_symbol at ast.py:8989>' Fld_48 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10403 '<FieldAccess for id at ast.py:8989>' Fld_47 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_47 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10403
      if Fld_47.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_48 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_47.Node);
--# end
--# expr-done 10404
--# expr-start 10405 '<Symbol Inline at ???>' Sym_14 None
Sym_14 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Inline); 
--# expr-done 10405
Is_Equal_8 := Fld_48 = Sym_14; 
--# expr-done 10406
   If_Result_13 := Is_Equal_8;
end if;
if If_Result_13 then
   If_Result_14 := True;
else
   --# expr-start 10410 '<Eq at ast.py:8989>' Is_Equal_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10408 '<FieldAccess for name_symbol at ast.py:8989>' Fld_50 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
--# expr-start 10407 '<FieldAccess for id at ast.py:8989>' Fld_49 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8989
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_49 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10407
      if Fld_49.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_50 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_49.Node);
--# end
--# expr-done 10408
--# expr-start 10409 '<Symbol Volatile at ???>' Sym_15 None
Sym_15 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Volatile); 
--# expr-done 10409
Is_Equal_9 := Fld_50 = Sym_15; 
--# expr-done 10410
   If_Result_14 := Is_Equal_9;
end if;
if If_Result_14 then
   --# expr-start 10414 '<LogicAll at ast.py:8994>' Logic_Boolean_Op_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8994
--# expr-start 10413 '<Map at ast.py:8994>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8994
   --# expr-start 10411 '<FieldAccess for associated_entity_names at ast.py:8994>' Fld_51 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8994
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Pragma.associated_entity_names
Fld_51 := Libadalang.Implementation.Pragma_Node_P_Associated_Entity_Names (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 10411
      declare
         Map_Result_1_Vec : Logic_Equation_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Name_Array_Access := Fld_51;
      begin
         for Item_310 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_310
      --# expr-start 10412 '<FieldAccess for xref_no_overloading at ast.py:8995>' Fld_52 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8995
      if Item_310.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_52 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Item_310.Node, Env => Env, Origin => Origin, E_Info => Item_310.Info);
--# end
--# expr-done 10412
         declare
            Item_To_Append : constant Logic_Equation := Fld_52;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_3041;
         end loop;
      end;
         Map_Result_1 := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_1_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 10413
Logic_Boolean_Op_1 := Solver.Create_All (Relation_Array (Map_Result_1.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8994") else null)); 
--# expr-done 10414
   If_Result_41 := Logic_Boolean_Op_1;
else
--# expr-start 10417 '<FieldAccess for name_is at ast.py:8998>' Fld_54 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8998
--# expr-start 10415 '<FieldAccess for id at ast.py:8998>' Fld_53 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8998
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_53 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10415
      if Fld_53.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10416 '<Symbol Warnings at ???>' Sym_16 None
Sym_16 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Warnings); 
--# expr-done 10416
--# property-call-start Name.name_is
Fld_54 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_53.Node, Sym => Sym_16);
--# end
--# expr-done 10417
if Fld_54 then
--# expr-start 10418 '<FieldAccess for args at ast.py:8999>' Fld_55 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8999
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_55 := Create_Internal_Entity_Base_Assoc_List (Node => Ent.Node.Pragma_Node_F_Args, Info => Ent.Info);
--# expr-done 10418
Fld_56 := Fld_55.Node;
Get_Result_2 := Get (Fld_56, 1, True); 
Is_Null_3 := Get_Result_2 = null; 
if Is_Null_3 then
   If_Result_15 := No_Entity_Base_Assoc;
else
Fld_57 := Fld_55.Info;
As_Entity_3 := (Info => Fld_57, Node => Get_Result_2); 
   If_Result_15 := As_Entity_3;
end if;
Var_Expr_3 := If_Result_15;
if Var_Expr_3 /= No_Entity_Base_Assoc then
   --# scope-start
--# expr-start 10419 '<FieldAccess for assoc_expr at ast.py:8999>' Fld_58 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8999
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseAssoc.assoc_expr'
Fld_58 := Libadalang.Implementation.Dispatcher_Base_Assoc_P_Assoc_Expr (Node => Var_Expr_3.Node, E_Info => Var_Expr_3.Info);
--# end
--# expr-done 10419
Scope_Result_2 := Fld_58;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := No_Entity_Expr;
end if;
Var_Expr_2 := Result_Var_2;
if Var_Expr_2 /= No_Entity_Expr then
   --# scope-start
   --# bind expr Var_Expr_2
--# expr-start 10430 '<If at ast.py:9000>' If_Result_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9000
--# expr-start 10420 '<IsA Identifier at ast.py:9001>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9001
Is_A := Var_Expr_2.Node /= null 
and then Var_Expr_2.Node.Kind in Ada_Identifier_Range; 
--# expr-done 10420
if Is_A then
   --# expr-start 10421 '<FieldAccess for sub_equation at ast.py:9003>' Fld_59 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9003
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_59 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr_2.Node, Env => Env, Origin => Origin, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 10421
   If_Result_16 := Fld_59;
else
   --# expr-start 10429 '<And at ast.py:9005>' And_Pred_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9005
--# expr-start 10427 '<And at ast.py:9005>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9005
--# expr-start 10425 '<Bind at ast.py:9005>' Bind_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9005
--# expr-start 10422 '<FieldAccess for expected_type_var at ast.py:9005>' Fld_60 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9005
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_60 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Var_Expr_2.Node);
--# expr-done 10422
Fld_60.Value := No_Entity;
Entity_Vars.Reset (Fld_60);
--# expr-start 10424 '<FieldAccess for std_entity at ast.py:9005>' Fld_61 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9005
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10423 '<Symbol String at ???>' Sym_17 None
Sym_17 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_String); 
--# expr-done 10423
--# property-call-start AdaNode.std_entity
Fld_61 := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym_17);
--# end
--# expr-done 10424
Bind_Result_2 := Solver.Create_Assign (Fld_60, Fld_61, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9005") else null)); 
--# expr-done 10425
--# expr-start 10426 '<FieldAccess for sub_equation at ast.py:9006>' Fld_62 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9006
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_62 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr_2.Node, Env => Env, Origin => Origin, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 10426
And_Pred_5 := Create_And (Bind_Result_2, Fld_62, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9005") else null)); 
--# expr-done 10427
--# expr-start 10428 '<FieldAccess for matches_expected_type at ast.py:9007>' Fld_63 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9007
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_63 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Var_Expr_2.Node, Origin => Origin);
--# end
--# expr-done 10428
And_Pred_6 := Create_And (And_Pred_5, Fld_63, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9005") else null)); 
--# expr-done 10429
   If_Result_16 := And_Pred_6;
end if;
      Inc_Ref (If_Result_16);
--# expr-done 10430
Scope_Result_3 := If_Result_16;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_3043;
   Result_Var_3 := Scope_Result_3;
else
   True_Rel_1 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9009") else null)); 
   Result_Var_3 := True_Rel_1;
end if;
      Inc_Ref (Result_Var_3);
   If_Result_40 := Result_Var_3;
else
--# expr-start 10434 '<Eq at ast.py:9017>' Is_Equal_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9017
--# expr-start 10432 '<FieldAccess for name_symbol at ast.py:9017>' Fld_65 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9017
--# expr-start 10431 '<FieldAccess for id at ast.py:9017>' Fld_64 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9017
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_64 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10431
      if Fld_64.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_65 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_64.Node);
--# end
--# expr-done 10432
--# expr-start 10433 '<Symbol Style_Checks at ???>' Sym_18 None
Sym_18 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Style_Checks); 
--# expr-done 10433
Is_Equal_10 := Fld_65 = Sym_18; 
--# expr-done 10434
if Is_Equal_10 then
   If_Result_17 := True;
else
   --# expr-start 10438 '<Eq at ast.py:9017>' Is_Equal_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9017
--# expr-start 10436 '<FieldAccess for name_symbol at ast.py:9017>' Fld_67 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9017
--# expr-start 10435 '<FieldAccess for id at ast.py:9017>' Fld_66 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9017
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_66 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10435
      if Fld_66.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_67 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_66.Node);
--# end
--# expr-done 10436
--# expr-start 10437 '<Symbol Import_Function at ???>' Sym_19 None
Sym_19 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Import_Function); 
--# expr-done 10437
Is_Equal_11 := Fld_67 = Sym_19; 
--# expr-done 10438
   If_Result_17 := Is_Equal_11;
end if;
if If_Result_17 then
   If_Result_18 := True;
else
   --# expr-start 10442 '<Eq at ast.py:9017>' Is_Equal_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9017
--# expr-start 10440 '<FieldAccess for name_symbol at ast.py:9017>' Fld_69 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9017
--# expr-start 10439 '<FieldAccess for id at ast.py:9017>' Fld_68 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9017
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_68 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10439
      if Fld_68.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_69 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_68.Node);
--# end
--# expr-done 10440
--# expr-start 10441 '<Symbol Import_Procedure at ???>' Sym_20 None
Sym_20 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Import_Procedure); 
--# expr-done 10441
Is_Equal_12 := Fld_69 = Sym_20; 
--# expr-done 10442
   If_Result_18 := Is_Equal_12;
end if;
if If_Result_18 then
   True_Rel_2 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9020") else null)); 
   If_Result_39 := True_Rel_2;
else
--# expr-start 10446 '<Eq at ast.py:9022>' Is_Equal_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
--# expr-start 10444 '<FieldAccess for name_symbol at ast.py:9022>' Fld_71 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
--# expr-start 10443 '<FieldAccess for id at ast.py:9022>' Fld_70 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_70 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10443
      if Fld_70.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_71 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_70.Node);
--# end
--# expr-done 10444
--# expr-start 10445 '<Symbol Pre at ???>' Sym_21 None
Sym_21 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Pre); 
--# expr-done 10445
Is_Equal_13 := Fld_71 = Sym_21; 
--# expr-done 10446
if Is_Equal_13 then
   If_Result_19 := True;
else
   --# expr-start 10450 '<Eq at ast.py:9022>' Is_Equal_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
--# expr-start 10448 '<FieldAccess for name_symbol at ast.py:9022>' Fld_73 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
--# expr-start 10447 '<FieldAccess for id at ast.py:9022>' Fld_72 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_72 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10447
      if Fld_72.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_73 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_72.Node);
--# end
--# expr-done 10448
--# expr-start 10449 '<Symbol Post at ???>' Sym_22 None
Sym_22 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Post); 
--# expr-done 10449
Is_Equal_14 := Fld_73 = Sym_22; 
--# expr-done 10450
   If_Result_19 := Is_Equal_14;
end if;
if If_Result_19 then
   If_Result_20 := True;
else
   --# expr-start 10454 '<Eq at ast.py:9022>' Is_Equal_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
--# expr-start 10452 '<FieldAccess for name_symbol at ast.py:9022>' Fld_75 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
--# expr-start 10451 '<FieldAccess for id at ast.py:9022>' Fld_74 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_74 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10451
      if Fld_74.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_75 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_74.Node);
--# end
--# expr-done 10452
--# expr-start 10453 '<Symbol Pre'"'"'Class at ???>' Sym_23 None
Sym_23 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Pre_Class); 
--# expr-done 10453
Is_Equal_15 := Fld_75 = Sym_23; 
--# expr-done 10454
   If_Result_20 := Is_Equal_15;
end if;
if If_Result_20 then
   If_Result_21 := True;
else
   --# expr-start 10458 '<Eq at ast.py:9022>' Is_Equal_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
--# expr-start 10456 '<FieldAccess for name_symbol at ast.py:9022>' Fld_77 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
--# expr-start 10455 '<FieldAccess for id at ast.py:9022>' Fld_76 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_76 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10455
      if Fld_76.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_77 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_76.Node);
--# end
--# expr-done 10456
--# expr-start 10457 '<Symbol Post'"'"'Class at ???>' Sym_24 None
Sym_24 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Post_Class); 
--# expr-done 10457
Is_Equal_16 := Fld_77 = Sym_24; 
--# expr-done 10458
   If_Result_21 := Is_Equal_16;
end if;
if If_Result_21 then
   If_Result_22 := True;
else
   --# expr-start 10462 '<Eq at ast.py:9022>' Is_Equal_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
--# expr-start 10460 '<FieldAccess for name_symbol at ast.py:9022>' Fld_79 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
--# expr-start 10459 '<FieldAccess for id at ast.py:9022>' Fld_78 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_78 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10459
      if Fld_78.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_79 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_78.Node);
--# end
--# expr-done 10460
--# expr-start 10461 '<Symbol Precondition at ???>' Sym_25 None
Sym_25 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Precondition); 
--# expr-done 10461
Is_Equal_17 := Fld_79 = Sym_25; 
--# expr-done 10462
   If_Result_22 := Is_Equal_17;
end if;
if If_Result_22 then
   If_Result_23 := True;
else
   --# expr-start 10466 '<Eq at ast.py:9022>' Is_Equal_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
--# expr-start 10464 '<FieldAccess for name_symbol at ast.py:9022>' Fld_81 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
--# expr-start 10463 '<FieldAccess for id at ast.py:9022>' Fld_80 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_80 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10463
      if Fld_80.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_81 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_80.Node);
--# end
--# expr-done 10464
--# expr-start 10465 '<Symbol Postcondition at ???>' Sym_26 None
Sym_26 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Postcondition); 
--# expr-done 10465
Is_Equal_18 := Fld_81 = Sym_26; 
--# expr-done 10466
   If_Result_23 := Is_Equal_18;
end if;
if If_Result_23 then
   If_Result_24 := True;
else
   --# expr-start 10470 '<Eq at ast.py:9022>' Is_Equal_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
--# expr-start 10468 '<FieldAccess for name_symbol at ast.py:9022>' Fld_83 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
--# expr-start 10467 '<FieldAccess for id at ast.py:9022>' Fld_82 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_82 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10467
      if Fld_82.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_83 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_82.Node);
--# end
--# expr-done 10468
--# expr-start 10469 '<Symbol Precondition'"'"'Class at ???>' Sym_27 None
Sym_27 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Precondition_Class); 
--# expr-done 10469
Is_Equal_19 := Fld_83 = Sym_27; 
--# expr-done 10470
   If_Result_24 := Is_Equal_19;
end if;
if If_Result_24 then
   If_Result_25 := True;
else
   --# expr-start 10474 '<Eq at ast.py:9022>' Is_Equal_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
--# expr-start 10472 '<FieldAccess for name_symbol at ast.py:9022>' Fld_85 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
--# expr-start 10471 '<FieldAccess for id at ast.py:9022>' Fld_84 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9022
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_84 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10471
      if Fld_84.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_85 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_84.Node);
--# end
--# expr-done 10472
--# expr-start 10473 '<Symbol Postcondition'"'"'Class at ???>' Sym_28 None
Sym_28 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Postcondition_Class); 
--# expr-done 10473
Is_Equal_20 := Fld_85 = Sym_28; 
--# expr-done 10474
   If_Result_25 := Is_Equal_20;
end if;
if If_Result_25 then
--# expr-start 10476 '<FieldAccess for assoc_expr at ast.py:9027>' Fld_89 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9027
--# expr-start 10475 '<FieldAccess for args at ast.py:9027>' Fld_86 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9027
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_86 := Create_Internal_Entity_Base_Assoc_List (Node => Ent.Node.Pragma_Node_F_Args, Info => Ent.Info);
--# expr-done 10475
Fld_87 := Fld_86.Node;
Get_Result_3 := Get (Fld_87, 0, True); 
Is_Null_4 := Get_Result_3 = null; 
if Is_Null_4 then
   If_Result_26 := No_Entity_Base_Assoc;
else
Fld_88 := Fld_86.Info;
As_Entity_4 := (Info => Fld_88, Node => Get_Result_3); 
   If_Result_26 := As_Entity_4;
end if;
      if If_Result_26.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseAssoc.assoc_expr'
Fld_89 := Libadalang.Implementation.Dispatcher_Base_Assoc_P_Assoc_Expr (Node => If_Result_26.Node, E_Info => If_Result_26.Info);
--# end
--# expr-done 10476
Var_Expr_4 := Fld_89;
if Var_Expr_4 /= No_Entity_Expr then
   --# scope-start
   --# bind expr Var_Expr_4
--# expr-start 10483 '<And at ast.py:9029>' And_Pred_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9029
--# expr-start 10481 '<And at ast.py:9029>' And_Pred_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9029
--# expr-start 10479 '<Bind at ast.py:9029>' Bind_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9029
--# expr-start 10477 '<FieldAccess for expected_type_var at ast.py:9029>' Fld_90 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9029
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_90 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Var_Expr_4.Node);
--# expr-done 10477
Fld_90.Value := No_Entity;
Entity_Vars.Reset (Fld_90);
--# expr-start 10478 '<FieldAccess for bool_type at ast.py:9029>' Fld_91 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9029
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_91 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 10478
      Cast_Result_2 := Create_Internal_Entity
        (Node => Fld_91.Node,
         Info => Fld_91.Info);
Bind_Result_3 := Solver.Create_Assign (Fld_90, Cast_Result_2, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9029") else null)); 
--# expr-done 10479
--# expr-start 10480 '<FieldAccess for sub_equation at ast.py:9030>' Fld_92 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9030
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_92 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr_4.Node, Env => Env, Origin => Origin, E_Info => Var_Expr_4.Info);
--# end
--# expr-done 10480
And_Pred_7 := Create_And (Bind_Result_3, Fld_92, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9029") else null)); 
--# expr-done 10481
--# expr-start 10482 '<FieldAccess for matches_expected_formal_prim_type at ast.py:9031>' Fld_93 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9031
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_prim_type
Fld_93 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Prim_Type (Node => Var_Expr_4.Node, Origin => Origin);
--# end
--# expr-done 10482
And_Pred_8 := Create_And (And_Pred_7, Fld_93, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9029") else null)); 
--# expr-done 10483
Scope_Result_4 := And_Pred_8;
   Inc_Ref (Scope_Result_4);
   --# end
      Finalizer_Scope_3044;
   Result_Var_4 := Scope_Result_4;
else
   False_Rel := Solver.Create_False ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9032") else null)); 
   Result_Var_4 := False_Rel;
end if;
      Inc_Ref (Result_Var_4);
   If_Result_38 := Result_Var_4;
else
--# expr-start 10486 '<FieldAccess for name_is at ast.py:9035>' Fld_95 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9035
--# expr-start 10484 '<FieldAccess for id at ast.py:9035>' Fld_94 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9035
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_94 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10484
      if Fld_94.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10485 '<Symbol Test_Case at ???>' Sym_29 None
Sym_29 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Test_Case); 
--# expr-done 10485
--# property-call-start Name.name_is
Fld_95 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_94.Node, Sym => Sym_29);
--# end
--# expr-done 10486
if Fld_95 then
   --# expr-start 10509 '<LogicAll at ast.py:9036>' Logic_Boolean_Op_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9036
--# expr-start 10508 '<Map at ast.py:9036>' Map_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9036
   --# expr-start 10497 '<Filter at ast.py:9036>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9036
   --# expr-start 10487 '<FieldAccess for args at ast.py:9036>' Fld_96 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9036
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_96 := Create_Internal_Entity_Base_Assoc_List (Node => Ent.Node.Pragma_Node_F_Args, Info => Ent.Info);
--# expr-done 10487
Fld_97 := Fld_96.Node;
      declare
         Map_Result_2_Vec : Internal_Entity_Base_Assoc_Vectors.Vector;
      begin
      if Fld_97 = null then
      Map_Result_2 := Create_Internal_Entity_Base_Assoc_Array (0);
      else
      declare
         Collection : constant Bare_Base_Assoc_List := Fld_97;
      begin
         for Untyped_Item_312 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_312 := Untyped_Item_312; 
Is_Null_5 := Bare_Item_312 = null; 
if Is_Null_5 then
   If_Result_27 := No_Entity_Base_Assoc;
else
Fld_98 := Fld_96.Info;
As_Entity_5 := (Info => Fld_98, Node => Bare_Item_312); 
   If_Result_27 := As_Entity_5;
end if;
                  Item_312 := If_Result_27; 
   --# scope-start
               --# bind arg Item_312
--# expr-start 10488 '<Cast to PragmaArgumentAssoc at ast.py:9037>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9037
Cast_Expr := Item_312; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Pragma_Argument_Assoc_Range
   then
      Cast_Result_3 := Create_Internal_Entity_Pragma_Argument_Assoc
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result_3 := No_Entity_Pragma_Argument_Assoc;
   end if;
--# expr-done 10488
Var_Expr_5 := Cast_Result_3;
if Var_Expr_5 /= No_Entity_Pragma_Argument_Assoc then
   --# scope-start
   --# bind parg Var_Expr_5
--# expr-start 10492 '<Eq at ast.py:9038>' Is_Equal_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9038
--# expr-start 10490 '<FieldAccess for name_symbol at ast.py:9038>' Fld_100 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9038
--# expr-start 10489 '<FieldAccess for name at ast.py:9038>' Fld_99 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9038
      if Var_Expr_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_99 := Create_Internal_Entity_Name (Node => Var_Expr_5.Node.Pragma_Argument_Assoc_F_Name, Info => Var_Expr_5.Info);
--# expr-done 10489
      if Fld_99.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_100 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_99.Node);
--# end
--# expr-done 10490
--# expr-start 10491 '<Symbol Requires at ???>' Sym_30 None
Sym_30 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Requires); 
--# expr-done 10491
Is_Equal_21 := Fld_100 = Sym_30; 
--# expr-done 10492
if Is_Equal_21 then
   If_Result_28 := True;
else
   --# expr-start 10496 '<Eq at ast.py:9038>' Is_Equal_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9038
--# expr-start 10494 '<FieldAccess for name_symbol at ast.py:9038>' Fld_102 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9038
--# expr-start 10493 '<FieldAccess for name at ast.py:9038>' Fld_101 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9038
      if Var_Expr_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_101 := Create_Internal_Entity_Name (Node => Var_Expr_5.Node.Pragma_Argument_Assoc_F_Name, Info => Var_Expr_5.Info);
--# expr-done 10493
      if Fld_101.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_102 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_101.Node);
--# end
--# expr-done 10494
--# expr-start 10495 '<Symbol Ensures at ???>' Sym_31 None
Sym_31 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Ensures); 
--# expr-done 10495
Is_Equal_22 := Fld_102 = Sym_31; 
--# expr-done 10496
   If_Result_28 := Is_Equal_22;
end if;
Scope_Result_5 := If_Result_28;
   --# end
   Result_Var_5 := Scope_Result_5;
else
   Result_Var_5 := False;
end if;
         if Result_Var_5 then
         declare
            Item_To_Append : constant Internal_Entity_Base_Assoc := Item_312;
         begin
            Internal_Entity_Base_Assoc_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
      end if;
         Map_Result_2 := Create_Internal_Entity_Base_Assoc_Array
           (Items_Count => Natural (Internal_Entity_Base_Assoc_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Internal_Entity_Base_Assoc_Vectors.Get
              (Map_Result_2_Vec,
               I + Internal_Entity_Base_Assoc_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Internal_Entity_Base_Assoc_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 10497
      declare
         Map_Result_3_Vec : Logic_Equation_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Base_Assoc_Array_Access := Map_Result_2;
      begin
         for Item_311 of
               Collection.Items
         loop
   --# scope-start
               --# bind arg Item_311
      --# expr-start 10507 '<And at ast.py:9044>' And_Pred_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9044
--# expr-start 10504 '<And at ast.py:9044>' And_Pred_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9044
--# expr-start 10501 '<Bind at ast.py:9044>' Bind_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9044
--# expr-start 10499 '<FieldAccess for expected_type_var at ast.py:9044>' Fld_104 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9044
--# expr-start 10498 '<FieldAccess for assoc_expr at ast.py:9044>' Fld_103 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9044
      if Item_311.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseAssoc.assoc_expr'
Fld_103 := Libadalang.Implementation.Dispatcher_Base_Assoc_P_Assoc_Expr (Node => Item_311.Node, E_Info => Item_311.Info);
--# end
--# expr-done 10498
      if Fld_103.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_104 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_103.Node);
--# expr-done 10499
Fld_104.Value := No_Entity;
Entity_Vars.Reset (Fld_104);
--# expr-start 10500 '<FieldAccess for bool_type at ast.py:9044>' Fld_105 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9044
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_105 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 10500
      Cast_Result_4 := Create_Internal_Entity
        (Node => Fld_105.Node,
         Info => Fld_105.Info);
Bind_Result_4 := Solver.Create_Assign (Fld_104, Cast_Result_4, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9044") else null)); 
--# expr-done 10501
--# expr-start 10503 '<FieldAccess for sub_equation at ast.py:9045>' Fld_107 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9045
--# expr-start 10502 '<FieldAccess for assoc_expr at ast.py:9045>' Fld_106 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9045
      if Item_311.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseAssoc.assoc_expr'
Fld_106 := Libadalang.Implementation.Dispatcher_Base_Assoc_P_Assoc_Expr (Node => Item_311.Node, E_Info => Item_311.Info);
--# end
--# expr-done 10502
      if Fld_106.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_107 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_106.Node, Env => Env, Origin => Origin, E_Info => Fld_106.Info);
--# end
--# expr-done 10503
And_Pred_9 := Create_And (Bind_Result_4, Fld_107, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9044") else null)); 
--# expr-done 10504
--# expr-start 10506 '<FieldAccess for matches_expected_formal_prim_type at ast.py:9046>' Fld_109 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9046
--# expr-start 10505 '<FieldAccess for assoc_expr at ast.py:9046>' Fld_108 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9046
      if Item_311.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseAssoc.assoc_expr'
Fld_108 := Libadalang.Implementation.Dispatcher_Base_Assoc_P_Assoc_Expr (Node => Item_311.Node, E_Info => Item_311.Info);
--# end
--# expr-done 10505
      if Fld_108.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_prim_type
Fld_109 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Prim_Type (Node => Fld_108.Node, Origin => Origin);
--# end
--# expr-done 10506
And_Pred_10 := Create_And (And_Pred_9, Fld_109, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9044") else null)); 
--# expr-done 10507
         declare
            Item_To_Append : constant Logic_Equation := And_Pred_10;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_3_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_3048;
         end loop;
      end;
         Map_Result_3 := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_3_Vec)));
         for I in Map_Result_3.Items'Range loop
            Map_Result_3.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_3_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result_3.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_3_Vec);
      end;
--# expr-done 10508
Logic_Boolean_Op_2 := Solver.Create_All (Relation_Array (Map_Result_3.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9036") else null)); 
--# expr-done 10509
   If_Result_37 := Logic_Boolean_Op_2;
else
--# expr-start 10512 '<FieldAccess for name_is at ast.py:9049>' Fld_111 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9049
--# expr-start 10510 '<FieldAccess for id at ast.py:9049>' Fld_110 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9049
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_110 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10510
      if Fld_110.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10511 '<Symbol Contract_Cases at ???>' Sym_32 None
Sym_32 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Contract_Cases); 
--# expr-done 10511
--# property-call-start Name.name_is
Fld_111 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_110.Node, Sym => Sym_32);
--# end
--# expr-done 10512
if Fld_111 then
   --# expr-start 10520 '<LogicAll at ast.py:9050>' Logic_Boolean_Op_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9050
--# expr-start 10519 '<Map at ast.py:9050>' Map_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9050
   --# expr-start 10516 '<FieldAccess for assocs at ast.py:9050>' Fld_116 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9050
--# expr-start 10515 '<Cast to BaseAggregate at ast.py:9050>' Cast_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9050
--# expr-start 10514 '<FieldAccess for assoc_expr at ast.py:9050>' Fld_115 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9050
--# expr-start 10513 '<FieldAccess for args at ast.py:9050>' Fld_112 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9050
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_112 := Create_Internal_Entity_Base_Assoc_List (Node => Ent.Node.Pragma_Node_F_Args, Info => Ent.Info);
--# expr-done 10513
Fld_113 := Fld_112.Node;
Get_Result_4 := Get (Fld_113, 0, True); 
Is_Null_6 := Get_Result_4 = null; 
if Is_Null_6 then
   If_Result_29 := No_Entity_Base_Assoc;
else
Fld_114 := Fld_112.Info;
As_Entity_6 := (Info => Fld_114, Node => Get_Result_4); 
   If_Result_29 := As_Entity_6;
end if;
      if If_Result_29.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseAssoc.assoc_expr'
Fld_115 := Libadalang.Implementation.Dispatcher_Base_Assoc_P_Assoc_Expr (Node => If_Result_29.Node, E_Info => If_Result_29.Info);
--# end
--# expr-done 10514
   if Fld_115.Node = null
      or else Fld_115.Node.Kind in Ada_Base_Aggregate
   then
      Cast_Result_5 := Create_Internal_Entity_Base_Aggregate
        (Node => Fld_115.Node,
         Info => Fld_115.Info);
   else
         Cast_Result_5 := No_Entity_Base_Aggregate;
   end if;
--# expr-done 10515
      if Cast_Result_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_116 := Create_Internal_Entity_Assoc_List (Node => Cast_Result_5.Node.Base_Aggregate_F_Assocs, Info => Cast_Result_5.Info);
--# expr-done 10516
Fld_117 := Fld_116.Node;
      declare
         Map_Result_4_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_117 = null then
      Map_Result_4 := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Assoc_List := Fld_117;
      begin
         for Untyped_Item_313 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_313 := Untyped_Item_313; 
Is_Null_7 := Bare_Item_313 = null; 
if Is_Null_7 then
   If_Result_30 := No_Entity_Basic_Assoc;
else
Fld_118 := Fld_116.Info;
As_Entity_7 := (Info => Fld_118, Node => Bare_Item_313); 
   If_Result_30 := As_Entity_7;
end if;
                  Item_313 := If_Result_30; 
   --# scope-start
               --# bind assoc Item_313
      --# expr-start 10518 '<FieldAccess for contract_cases_assoc_equation at ast.py:9052>' Fld_119 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9052
--# expr-start 10517 '<Cast to AggregateAssoc at ast.py:9052>' Cast_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9052
Cast_Expr_1 := Item_313; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Aggregate_Assoc_Range
   then
      Cast_Result_6 := Create_Internal_Entity_Aggregate_Assoc
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_6 := No_Entity_Aggregate_Assoc;
   end if;
--# expr-done 10517
      if Cast_Result_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AggregateAssoc.contract_cases_assoc_equation
Fld_119 := Libadalang.Implementation.Aggregate_Assoc_P_Contract_Cases_Assoc_Equation (Node => Cast_Result_6.Node, Env => Env, Origin => Origin, E_Info => Cast_Result_6.Info);
--# end
--# expr-done 10518
         declare
            Item_To_Append : constant Logic_Equation := Fld_119;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_4_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_3049;
         end loop;
      end;
      end if;
         Map_Result_4 := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_4_Vec)));
         for I in Map_Result_4.Items'Range loop
            Map_Result_4.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_4_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result_4.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_4_Vec);
      end;
--# expr-done 10519
Logic_Boolean_Op_3 := Solver.Create_All (Relation_Array (Map_Result_4.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9050") else null)); 
--# expr-done 10520
   If_Result_36 := Logic_Boolean_Op_3;
else
--# expr-start 10523 '<FieldAccess for name_is at ast.py:9055>' Fld_121 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9055
--# expr-start 10521 '<FieldAccess for id at ast.py:9055>' Fld_120 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9055
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_120 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10521
      if Fld_120.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10522 '<Symbol Debug at ???>' Sym_33 None
Sym_33 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Debug); 
--# expr-done 10522
--# property-call-start Name.name_is
Fld_121 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_120.Node, Sym => Sym_33);
--# end
--# expr-done 10523
if Fld_121 then
   --# expr-start 10548 '<And at ast.py:9056>' And_Pred_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9056
--# expr-start 10537 '<If at ast.py:9056>' If_Result_32 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9056
--# expr-start 10526 '<Eq at ast.py:9059>' Is_Equal_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9059
--# expr-start 10525 '<length at ast.py:9059>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9059
--# expr-start 10524 '<FieldAccess for args at ast.py:9059>' Fld_130 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9059
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_130 := Create_Internal_Entity_Base_Assoc_List (Node => Ent.Node.Pragma_Node_F_Args, Info => Ent.Info);
--# expr-done 10524
Fld_131 := Fld_130.Node; 
Len := Length (Fld_131); 
--# expr-done 10525
Is_Equal_23 := Len = 2; 
--# expr-done 10526
if Is_Equal_23 then
   --# expr-start 10536 '<Let at ast.py:9060>' Let_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9060
--# scope-start
--# expr-start 10528 '<FieldAccess for assoc_expr at ast.py:9061>' Fld_125 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9061
--# expr-start 10527 '<FieldAccess for args at ast.py:9061>' Fld_122 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9061
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_122 := Create_Internal_Entity_Base_Assoc_List (Node => Ent.Node.Pragma_Node_F_Args, Info => Ent.Info);
--# expr-done 10527
Fld_123 := Fld_122.Node;
Get_Result_5 := Get (Fld_123, 0, True); 
Is_Null_8 := Get_Result_5 = null; 
if Is_Null_8 then
   If_Result_31 := No_Entity_Base_Assoc;
else
Fld_124 := Fld_122.Info;
As_Entity_8 := (Info => Fld_124, Node => Get_Result_5); 
   If_Result_31 := As_Entity_8;
end if;
      if If_Result_31.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseAssoc.assoc_expr'
Fld_125 := Libadalang.Implementation.Dispatcher_Base_Assoc_P_Assoc_Expr (Node => If_Result_31.Node, E_Info => If_Result_31.Info);
--# end
--# expr-done 10528
Expr_1 := Fld_125; 
--# bind expr Expr_1
--# expr-start 10535 '<And at ast.py:9062>' And_Pred_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9062
--# expr-start 10533 '<And at ast.py:9062>' And_Pred_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9062
--# expr-start 10531 '<Bind at ast.py:9062>' Bind_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9062
--# expr-start 10529 '<FieldAccess for expected_type_var at ast.py:9062>' Fld_126 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9062
      if Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_126 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Expr_1.Node);
--# expr-done 10529
Fld_126.Value := No_Entity;
Entity_Vars.Reset (Fld_126);
--# expr-start 10530 '<FieldAccess for bool_type at ast.py:9062>' Fld_127 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9062
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_127 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 10530
      Cast_Result_7 := Create_Internal_Entity
        (Node => Fld_127.Node,
         Info => Fld_127.Info);
Bind_Result_5 := Solver.Create_Assign (Fld_126, Cast_Result_7, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9062") else null)); 
--# expr-done 10531
--# expr-start 10532 '<FieldAccess for sub_equation at ast.py:9063>' Fld_128 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9063
      if Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_128 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Expr_1.Node, Env => Env, Origin => Origin, E_Info => Expr_1.Info);
--# end
--# expr-done 10532
And_Pred_11 := Create_And (Bind_Result_5, Fld_128, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9062") else null)); 
--# expr-done 10533
--# expr-start 10534 '<FieldAccess for matches_expected_formal_prim_type at ast.py:9064>' Fld_129 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9064
      if Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_prim_type
Fld_129 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Prim_Type (Node => Expr_1.Node, Origin => Origin);
--# end
--# expr-done 10534
And_Pred_12 := Create_And (And_Pred_11, Fld_129, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9062") else null)); 
--# expr-done 10535
Let_Result_1 := And_Pred_12; Inc_Ref (Let_Result_1);
--# end
--# expr-done 10536
   If_Result_32 := Let_Result_1;
else
   True_Rel_3 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9066") else null)); 
   If_Result_32 := True_Rel_3;
end if;
      Inc_Ref (If_Result_32);
--# expr-done 10537
--# expr-start 10547 '<Let at ast.py:9067>' Let_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9067
--# scope-start
--# expr-start 10538 '<FieldAccess for args at ast.py:9069>' Fld_134 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9069
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_134 := Create_Internal_Entity_Base_Assoc_List (Node => Ent.Node.Pragma_Node_F_Args, Info => Ent.Info);
--# expr-done 10538
Fld_135 := Fld_134.Node;
--# expr-start 10541 '<Arithmetic - at ast.py:9069>' Arith_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9069
--# expr-start 10540 '<length at ast.py:9069>' Len_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9069
--# expr-start 10539 '<FieldAccess for args at ast.py:9069>' Fld_132 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9069
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_132 := Create_Internal_Entity_Base_Assoc_List (Node => Ent.Node.Pragma_Node_F_Args, Info => Ent.Info);
--# expr-done 10539
Fld_133 := Fld_132.Node; 
Len_1 := Length (Fld_133); 
--# expr-done 10540
Arith_Result := (Len_1 - 1); 
--# expr-done 10541
Get_Result_6 := Get (Fld_135, Arith_Result, True); 
Is_Null_9 := Get_Result_6 = null; 
if Is_Null_9 then
   If_Result_33 := No_Entity_Base_Assoc;
else
Fld_136 := Fld_134.Info;
As_Entity_9 := (Info => Fld_136, Node => Get_Result_6); 
   If_Result_33 := As_Entity_9;
end if;
Var_Expr_6 := If_Result_33;
if Var_Expr_6 /= No_Entity_Base_Assoc then
   --# scope-start
--# expr-start 10542 '<FieldAccess for assoc_expr at ast.py:9069>' Fld_137 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9069
      if Var_Expr_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseAssoc.assoc_expr'
Fld_137 := Libadalang.Implementation.Dispatcher_Base_Assoc_P_Assoc_Expr (Node => Var_Expr_6.Node, E_Info => Var_Expr_6.Info);
--# end
--# expr-done 10542
Scope_Result_6 := Fld_137;
   --# end
   Result_Var_6 := Scope_Result_6;
else
   Result_Var_6 := No_Entity_Expr;
end if;
Proc := Result_Var_6; 
--# bind proc Proc
--# expr-start 10546 '<And at ast.py:9070>' And_Pred_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9070
--# expr-start 10544 '<Bind at ast.py:9070>' Bind_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9070
--# expr-start 10543 '<FieldAccess for type_var at ast.py:9070>' Fld_138 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9070
      if Proc.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_138 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Proc.Node);
--# expr-done 10543
Fld_138.Value := No_Entity;
Entity_Vars.Reset (Fld_138);
Cast_Expr_2 := No_Entity_Base_Type_Decl; 
      Cast_Result_8 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
Bind_Result_6 := Solver.Create_Assign (Fld_138, Cast_Result_8, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9070") else null)); 
--# expr-done 10544
--# expr-start 10545 '<FieldAccess for sub_equation at ast.py:9071>' Fld_139 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9071
      if Proc.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_139 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Proc.Node, Env => Env, Origin => Origin, E_Info => Proc.Info);
--# end
--# expr-done 10545
And_Pred_13 := Create_And (Bind_Result_6, Fld_139, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9070") else null)); 
--# expr-done 10546
Let_Result_2 := And_Pred_13; Inc_Ref (Let_Result_2);
--# end
--# expr-done 10547
And_Pred_14 := Create_And (If_Result_32, Let_Result_2, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9056") else null)); 
--# expr-done 10548
   If_Result_35 := And_Pred_14;
else
   --# expr-start 10554 '<LogicAll at ast.py:9074>' Logic_Boolean_Op_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9074
--# expr-start 10553 '<Map at ast.py:9074>' Map_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9074
   --# expr-start 10549 '<FieldAccess for args at ast.py:9074>' Fld_140 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9074
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_140 := Create_Internal_Entity_Base_Assoc_List (Node => Ent.Node.Pragma_Node_F_Args, Info => Ent.Info);
--# expr-done 10549
Fld_141 := Fld_140.Node;
      declare
         Map_Result_5_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_141 = null then
      Map_Result_5 := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Base_Assoc_List := Fld_141;
      begin
         for Untyped_Item_314 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_314 := Untyped_Item_314; 
Is_Null_10 := Bare_Item_314 = null; 
if Is_Null_10 then
   If_Result_34 := No_Entity_Base_Assoc;
else
Fld_142 := Fld_140.Info;
As_Entity_10 := (Info => Fld_142, Node => Bare_Item_314); 
   If_Result_34 := As_Entity_10;
end if;
                  Item_314 := If_Result_34; 
   --# scope-start
               --# bind a Item_314
      --# expr-start 10552 '<Or at ast.py:9078>' Or_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9078
--# expr-start 10551 '<FieldAccess for sub_equation at ast.py:9078>' Fld_144 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9078
--# expr-start 10550 '<FieldAccess for assoc_expr at ast.py:9078>' Fld_143 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9078
      if Item_314.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseAssoc.assoc_expr'
Fld_143 := Libadalang.Implementation.Dispatcher_Base_Assoc_P_Assoc_Expr (Node => Item_314.Node, E_Info => Item_314.Info);
--# end
--# expr-done 10550
      if Fld_143.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_144 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_143.Node, Env => Env, Origin => Origin, E_Info => Fld_143.Info);
--# end
--# expr-done 10551
True_Rel_4 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9078") else null)); 
Or_Pred := Create_Or (Fld_144, True_Rel_4, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9078") else null)); 
--# expr-done 10552
         declare
            Item_To_Append : constant Logic_Equation := Or_Pred;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_5_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_3051;
         end loop;
      end;
      end if;
         Map_Result_5 := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_5_Vec)));
         for I in Map_Result_5.Items'Range loop
            Map_Result_5.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_5_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result_5.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_5_Vec);
      end;
--# expr-done 10553
Logic_Boolean_Op_4 := Solver.Create_All (Relation_Array (Map_Result_5.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:9074") else null)); 
--# expr-done 10554
   If_Result_35 := Logic_Boolean_Op_4;
end if;
      Inc_Ref (If_Result_35);
   If_Result_36 := If_Result_35;
end if;
      Inc_Ref (If_Result_36);
   If_Result_37 := If_Result_36;
end if;
      Inc_Ref (If_Result_37);
   If_Result_38 := If_Result_37;
end if;
      Inc_Ref (If_Result_38);
   If_Result_39 := If_Result_38;
end if;
      Inc_Ref (If_Result_39);
   If_Result_40 := If_Result_39;
end if;
      Inc_Ref (If_Result_40);
   If_Result_41 := If_Result_40;
end if;
      Inc_Ref (If_Result_41);
   If_Result_42 := If_Result_41;
end if;
      Inc_Ref (If_Result_42);
   If_Result_43 := If_Result_42;
end if;
      Inc_Ref (If_Result_43);
--# expr-done 10555
Let_Result_3 := If_Result_43; Inc_Ref (Let_Result_3);
--# end
--# expr-done 10556
      Property_Result := Let_Result_3;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_730;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_730;
               Finalizer_Scope_3039;
               Finalizer_Scope_3040;
               Finalizer_Scope_3041;
               Finalizer_Scope_3043;
               Finalizer_Scope_3044;
               Finalizer_Scope_3048;
               Finalizer_Scope_3049;
               Finalizer_Scope_3051;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Pragma_Node_P_Xref_Equation;
--# end
--# property-start Pragma.associated_entity_names /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9083
pragma Warnings (Off, "is not referenced");
 function Pragma_Node_P_Associated_Entity_Names
  (Node : Bare_Pragma_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Name_Array_Access
is
   Self : Bare_Pragma_Node := Bare_Pragma_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Pragma_Node :=
     Internal_Entity_Pragma_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Name_Array_Access;
            procedure Finalizer_Scope_731 with Inline_Always;
            procedure Finalizer_Scope_3054 with Inline_Always;
      Var_Expr : Internal_Entity_Name;
Var_Expr_1 : Internal_Entity_Base_Assoc;
Fld : Internal_Entity_Identifier;
Fld_1 : Symbol_Type;
Sym : Symbol_Type;
Is_Equal : Boolean;
Fld_2 : Internal_Entity_Identifier;
Fld_3 : Symbol_Type;
Sym_1 : Symbol_Type;
Is_Equal_1 : Boolean;
If_Result : Boolean;
Fld_4 : Internal_Entity_Identifier;
Fld_5 : Symbol_Type;
Sym_2 : Symbol_Type;
Is_Equal_2 : Boolean;
If_Result_1 : Boolean;
Fld_6 : Internal_Entity_Identifier;
Fld_7 : Symbol_Type;
Sym_3 : Symbol_Type;
Is_Equal_3 : Boolean;
If_Result_2 : Boolean;
Fld_8 : Internal_Entity_Base_Assoc_List;
Fld_9 : Bare_Base_Assoc_List;
Fld_10 : Internal_Entity_Info;
Get_Result : Bare_Base_Assoc;
As_Entity : Internal_Entity_Base_Assoc;
Is_Null : Boolean;
If_Result_3 : Internal_Entity_Base_Assoc;
Fld_11 : Internal_Entity_Expr;
Cast_Result : Internal_Entity_Name;
Singleton : Internal_Entity_Name_Array_Access;
Fld_12 : Internal_Entity_Identifier;
Sym_4 : Symbol_Type;
Fld_13 : Boolean;
Fld_14 : Internal_Entity_Base_Assoc_List;
Fld_15 : Bare_Base_Assoc_List;
Fld_16 : Internal_Entity_Info;
As_Entity_1 : Internal_Entity_Base_Assoc;
Is_Null_1 : Boolean;
If_Result_4 : Internal_Entity_Base_Assoc;
Fld_17 : Internal_Entity_Expr;
Cast_Result_1 : Internal_Entity_Name;
Item_315 : Internal_Entity_Base_Assoc;
Bare_Item_315 : Bare_Base_Assoc;
Map_Result : Internal_Entity_Name_Array_Access;
Fld_18 : Internal_Entity_Identifier;
Fld_19 : Symbol_Type;
Sym_5 : Symbol_Type;
Is_Equal_4 : Boolean;
Fld_20 : Internal_Entity_Identifier;
Fld_21 : Symbol_Type;
Sym_6 : Symbol_Type;
Is_Equal_5 : Boolean;
If_Result_5 : Boolean;
Fld_22 : Internal_Entity_Identifier;
Fld_23 : Symbol_Type;
Sym_7 : Symbol_Type;
Is_Equal_6 : Boolean;
If_Result_6 : Boolean;
Fld_24 : Internal_Entity_Identifier;
Fld_25 : Symbol_Type;
Sym_8 : Symbol_Type;
Is_Equal_7 : Boolean;
If_Result_7 : Boolean;
Fld_26 : Internal_Entity_Identifier;
Fld_27 : Symbol_Type;
Sym_9 : Symbol_Type;
Is_Equal_8 : Boolean;
If_Result_8 : Boolean;
Fld_28 : Internal_Entity_Identifier;
Fld_29 : Symbol_Type;
Sym_10 : Symbol_Type;
Is_Equal_9 : Boolean;
If_Result_9 : Boolean;
Fld_30 : Internal_Entity_Identifier;
Fld_31 : Symbol_Type;
Sym_11 : Symbol_Type;
Is_Equal_10 : Boolean;
If_Result_10 : Boolean;
Fld_32 : Internal_Entity_Identifier;
Fld_33 : Symbol_Type;
Sym_12 : Symbol_Type;
Is_Equal_11 : Boolean;
If_Result_11 : Boolean;
Fld_34 : Internal_Entity_Identifier;
Fld_35 : Symbol_Type;
Sym_13 : Symbol_Type;
Is_Equal_12 : Boolean;
If_Result_12 : Boolean;
Fld_36 : Internal_Entity_Identifier;
Fld_37 : Symbol_Type;
Sym_14 : Symbol_Type;
Is_Equal_13 : Boolean;
If_Result_13 : Boolean;
Fld_38 : Internal_Entity_Identifier;
Fld_39 : Symbol_Type;
Sym_15 : Symbol_Type;
Is_Equal_14 : Boolean;
If_Result_14 : Boolean;
Fld_40 : Internal_Entity_Identifier;
Fld_41 : Symbol_Type;
Sym_16 : Symbol_Type;
Is_Equal_15 : Boolean;
If_Result_15 : Boolean;
Fld_42 : Internal_Entity_Identifier;
Fld_43 : Symbol_Type;
Sym_17 : Symbol_Type;
Is_Equal_16 : Boolean;
If_Result_16 : Boolean;
Fld_44 : Internal_Entity_Identifier;
Fld_45 : Symbol_Type;
Sym_18 : Symbol_Type;
Is_Equal_17 : Boolean;
If_Result_17 : Boolean;
Fld_46 : Internal_Entity_Identifier;
Fld_47 : Symbol_Type;
Sym_19 : Symbol_Type;
Is_Equal_18 : Boolean;
If_Result_18 : Boolean;
Fld_48 : Internal_Entity_Identifier;
Fld_49 : Symbol_Type;
Sym_20 : Symbol_Type;
Is_Equal_19 : Boolean;
If_Result_19 : Boolean;
Fld_50 : Internal_Entity_Base_Assoc_List;
Fld_51 : Bare_Base_Assoc_List;
Fld_52 : Internal_Entity_Info;
Get_Result_1 : Bare_Base_Assoc;
As_Entity_2 : Internal_Entity_Base_Assoc;
Is_Null_2 : Boolean;
If_Result_20 : Internal_Entity_Base_Assoc;
Fld_53 : Internal_Entity_Expr;
Scope_Result : Internal_Entity_Expr;
Result_Var : Internal_Entity_Expr;
Cast_Result_2 : Internal_Entity_Name;
Singleton_1 : Internal_Entity_Name_Array_Access;
Scope_Result_1 : Internal_Entity_Name_Array_Access;
Result_Var_1 : Internal_Entity_Name_Array_Access;
If_Result_21 : Internal_Entity_Name_Array_Access;
If_Result_22 : Internal_Entity_Name_Array_Access;
If_Result_23 : Internal_Entity_Name_Array_Access;
Let_Result : Internal_Entity_Name_Array_Access;
            procedure Finalizer_Scope_731 is
            begin
                     Dec_Ref (Singleton);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (If_Result_21);
                     Dec_Ref (If_Result_22);
                     Dec_Ref (If_Result_23);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_731;
            procedure Finalizer_Scope_3054 is
            begin
                     Dec_Ref (Singleton_1);
            end Finalizer_Scope_3054;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 10651 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10650 '<Cond at ast.py:9084>' If_Result_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9084
--# expr-start 10560 '<Eq at ast.py:9085>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9085
--# expr-start 10558 '<FieldAccess for name_symbol at ast.py:9085>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9085
--# expr-start 10557 '<FieldAccess for id at ast.py:9085>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9085
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10557
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld.Node);
--# end
--# expr-done 10558
--# expr-start 10559 '<Symbol Import at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Import); 
--# expr-done 10559
Is_Equal := Fld_1 = Sym; 
--# expr-done 10560
if Is_Equal then
   If_Result := True;
else
   --# expr-start 10564 '<Eq at ast.py:9085>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9085
--# expr-start 10562 '<FieldAccess for name_symbol at ast.py:9085>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9085
--# expr-start 10561 '<FieldAccess for id at ast.py:9085>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9085
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10561
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_2.Node);
--# end
--# expr-done 10562
--# expr-start 10563 '<Symbol Export at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Export); 
--# expr-done 10563
Is_Equal_1 := Fld_3 = Sym_1; 
--# expr-done 10564
   If_Result := Is_Equal_1;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 10568 '<Eq at ast.py:9085>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9085
--# expr-start 10566 '<FieldAccess for name_symbol at ast.py:9085>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9085
--# expr-start 10565 '<FieldAccess for id at ast.py:9085>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9085
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10565
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_5 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_4.Node);
--# end
--# expr-done 10566
--# expr-start 10567 '<Symbol Interface at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Interface); 
--# expr-done 10567
Is_Equal_2 := Fld_5 = Sym_2; 
--# expr-done 10568
   If_Result_1 := Is_Equal_2;
end if;
if If_Result_1 then
   If_Result_2 := True;
else
   --# expr-start 10572 '<Eq at ast.py:9085>' Is_Equal_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9085
--# expr-start 10570 '<FieldAccess for name_symbol at ast.py:9085>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9085
--# expr-start 10569 '<FieldAccess for id at ast.py:9085>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9085
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10569
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_7 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_6.Node);
--# end
--# expr-done 10570
--# expr-start 10571 '<Symbol Convention at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Convention); 
--# expr-done 10571
Is_Equal_3 := Fld_7 = Sym_3; 
--# expr-done 10572
   If_Result_2 := Is_Equal_3;
end if;
if If_Result_2 then
   --# expr-start 10575 '<Cast to Name at ast.py:9088>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9088
--# expr-start 10574 '<FieldAccess for assoc_expr at ast.py:9088>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9088
--# expr-start 10573 '<FieldAccess for args at ast.py:9088>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9088
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Base_Assoc_List (Node => Ent.Node.Pragma_Node_F_Args, Info => Ent.Info);
--# expr-done 10573
Fld_9 := Fld_8.Node;
Get_Result := Get (Fld_9, 1, True); 
Is_Null := Get_Result = null; 
if Is_Null then
   If_Result_3 := No_Entity_Base_Assoc;
else
Fld_10 := Fld_8.Info;
As_Entity := (Info => Fld_10, Node => Get_Result); 
   If_Result_3 := As_Entity;
end if;
      if If_Result_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseAssoc.assoc_expr'
Fld_11 := Libadalang.Implementation.Dispatcher_Base_Assoc_P_Assoc_Expr (Node => If_Result_3.Node, E_Info => If_Result_3.Info);
--# end
--# expr-done 10574
   if Fld_11.Node = null
      or else Fld_11.Node.Kind in Ada_Name
   then
      Cast_Result := Create_Internal_Entity_Name
        (Node => Fld_11.Node,
         Info => Fld_11.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 10575
                Singleton := Create_Internal_Entity_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Cast_Result;
   If_Result_23 := Singleton;
else
--# expr-start 10578 '<FieldAccess for name_is at ast.py:9090>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9090
--# expr-start 10576 '<FieldAccess for id at ast.py:9090>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9090
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10576
      if Fld_12.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10577 '<Symbol Inline at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Inline); 
--# expr-done 10577
--# property-call-start Name.name_is
Fld_13 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_12.Node, Sym => Sym_4);
--# end
--# expr-done 10578
if Fld_13 then
   --# expr-start 10582 '<Map at ast.py:9091>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9091
   --# expr-start 10579 '<FieldAccess for args at ast.py:9091>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9091
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Create_Internal_Entity_Base_Assoc_List (Node => Ent.Node.Pragma_Node_F_Args, Info => Ent.Info);
--# expr-done 10579
Fld_15 := Fld_14.Node;
      declare
         Map_Result_Vec : Internal_Entity_Name_Vectors.Vector;
      begin
      if Fld_15 = null then
      Map_Result := Create_Internal_Entity_Name_Array (0);
      else
      declare
         Collection : constant Bare_Base_Assoc_List := Fld_15;
      begin
         for Untyped_Item_315 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_315 := Untyped_Item_315; 
Is_Null_1 := Bare_Item_315 = null; 
if Is_Null_1 then
   If_Result_4 := No_Entity_Base_Assoc;
else
Fld_16 := Fld_14.Info;
As_Entity_1 := (Info => Fld_16, Node => Bare_Item_315); 
   If_Result_4 := As_Entity_1;
end if;
                  Item_315 := If_Result_4; 
   --# scope-start
               --# bind a Item_315
      --# expr-start 10581 '<Cast to Name at ast.py:9091>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9091
--# expr-start 10580 '<FieldAccess for assoc_expr at ast.py:9091>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9091
      if Item_315.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseAssoc.assoc_expr'
Fld_17 := Libadalang.Implementation.Dispatcher_Base_Assoc_P_Assoc_Expr (Node => Item_315.Node, E_Info => Item_315.Info);
--# end
--# expr-done 10580
   if Fld_17.Node = null
      or else Fld_17.Node.Kind in Ada_Name
   then
      Cast_Result_1 := Create_Internal_Entity_Name
        (Node => Fld_17.Node,
         Info => Fld_17.Info);
   else
         Cast_Result_1 := No_Entity_Name;
   end if;
--# expr-done 10581
         declare
            Item_To_Append : constant Internal_Entity_Name := Cast_Result_1;
         begin
            Internal_Entity_Name_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Name_Array
           (Items_Count => Natural (Internal_Entity_Name_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Name_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Name_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Name_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 10582
   If_Result_22 := Map_Result;
else
--# expr-start 10586 '<Eq at ast.py:9093>' Is_Equal_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10584 '<FieldAccess for name_symbol at ast.py:9093>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10583 '<FieldAccess for id at ast.py:9093>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10583
      if Fld_18.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_19 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_18.Node);
--# end
--# expr-done 10584
--# expr-start 10585 '<Symbol Pack at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Pack); 
--# expr-done 10585
Is_Equal_4 := Fld_19 = Sym_5; 
--# expr-done 10586
if Is_Equal_4 then
   If_Result_5 := True;
else
   --# expr-start 10590 '<Eq at ast.py:9093>' Is_Equal_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10588 '<FieldAccess for name_symbol at ast.py:9093>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10587 '<FieldAccess for id at ast.py:9093>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_20 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10587
      if Fld_20.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_21 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_20.Node);
--# end
--# expr-done 10588
--# expr-start 10589 '<Symbol Pure at ???>' Sym_6 None
Sym_6 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Pure); 
--# expr-done 10589
Is_Equal_5 := Fld_21 = Sym_6; 
--# expr-done 10590
   If_Result_5 := Is_Equal_5;
end if;
if If_Result_5 then
   If_Result_6 := True;
else
   --# expr-start 10594 '<Eq at ast.py:9093>' Is_Equal_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10592 '<FieldAccess for name_symbol at ast.py:9093>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10591 '<FieldAccess for id at ast.py:9093>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_22 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10591
      if Fld_22.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_23 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_22.Node);
--# end
--# expr-done 10592
--# expr-start 10593 '<Symbol Preelaborate at ???>' Sym_7 None
Sym_7 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Preelaborate); 
--# expr-done 10593
Is_Equal_6 := Fld_23 = Sym_7; 
--# expr-done 10594
   If_Result_6 := Is_Equal_6;
end if;
if If_Result_6 then
   If_Result_7 := True;
else
   --# expr-start 10598 '<Eq at ast.py:9093>' Is_Equal_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10596 '<FieldAccess for name_symbol at ast.py:9093>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10595 '<FieldAccess for id at ast.py:9093>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_24 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10595
      if Fld_24.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_25 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_24.Node);
--# end
--# expr-done 10596
--# expr-start 10597 '<Symbol Elaborate_Body at ???>' Sym_8 None
Sym_8 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Elaborate_Body); 
--# expr-done 10597
Is_Equal_7 := Fld_25 = Sym_8; 
--# expr-done 10598
   If_Result_7 := Is_Equal_7;
end if;
if If_Result_7 then
   If_Result_8 := True;
else
   --# expr-start 10602 '<Eq at ast.py:9093>' Is_Equal_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10600 '<FieldAccess for name_symbol at ast.py:9093>' Fld_27 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10599 '<FieldAccess for id at ast.py:9093>' Fld_26 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_26 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10599
      if Fld_26.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_27 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_26.Node);
--# end
--# expr-done 10600
--# expr-start 10601 '<Symbol Volatile at ???>' Sym_9 None
Sym_9 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Volatile); 
--# expr-done 10601
Is_Equal_8 := Fld_27 = Sym_9; 
--# expr-done 10602
   If_Result_8 := Is_Equal_8;
end if;
if If_Result_8 then
   If_Result_9 := True;
else
   --# expr-start 10606 '<Eq at ast.py:9093>' Is_Equal_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10604 '<FieldAccess for name_symbol at ast.py:9093>' Fld_29 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10603 '<FieldAccess for id at ast.py:9093>' Fld_28 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_28 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10603
      if Fld_28.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_29 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_28.Node);
--# end
--# expr-done 10604
--# expr-start 10605 '<Symbol Volatile_Components at ???>' Sym_10 None
Sym_10 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Volatile_Components); 
--# expr-done 10605
Is_Equal_9 := Fld_29 = Sym_10; 
--# expr-done 10606
   If_Result_9 := Is_Equal_9;
end if;
if If_Result_9 then
   If_Result_10 := True;
else
   --# expr-start 10610 '<Eq at ast.py:9093>' Is_Equal_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10608 '<FieldAccess for name_symbol at ast.py:9093>' Fld_31 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10607 '<FieldAccess for id at ast.py:9093>' Fld_30 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_30 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10607
      if Fld_30.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_31 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_30.Node);
--# end
--# expr-done 10608
--# expr-start 10609 '<Symbol Unchecked_Union at ???>' Sym_11 None
Sym_11 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Unchecked_Union); 
--# expr-done 10609
Is_Equal_10 := Fld_31 = Sym_11; 
--# expr-done 10610
   If_Result_10 := Is_Equal_10;
end if;
if If_Result_10 then
   If_Result_11 := True;
else
   --# expr-start 10614 '<Eq at ast.py:9093>' Is_Equal_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10612 '<FieldAccess for name_symbol at ast.py:9093>' Fld_33 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10611 '<FieldAccess for id at ast.py:9093>' Fld_32 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_32 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10611
      if Fld_32.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_33 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_32.Node);
--# end
--# expr-done 10612
--# expr-start 10613 '<Symbol Atomic at ???>' Sym_12 None
Sym_12 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Atomic); 
--# expr-done 10613
Is_Equal_11 := Fld_33 = Sym_12; 
--# expr-done 10614
   If_Result_11 := Is_Equal_11;
end if;
if If_Result_11 then
   If_Result_12 := True;
else
   --# expr-start 10618 '<Eq at ast.py:9093>' Is_Equal_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10616 '<FieldAccess for name_symbol at ast.py:9093>' Fld_35 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10615 '<FieldAccess for id at ast.py:9093>' Fld_34 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_34 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10615
      if Fld_34.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_35 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_34.Node);
--# end
--# expr-done 10616
--# expr-start 10617 '<Symbol Atomic_Components at ???>' Sym_13 None
Sym_13 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Atomic_Components); 
--# expr-done 10617
Is_Equal_12 := Fld_35 = Sym_13; 
--# expr-done 10618
   If_Result_12 := Is_Equal_12;
end if;
if If_Result_12 then
   If_Result_13 := True;
else
   --# expr-start 10622 '<Eq at ast.py:9093>' Is_Equal_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10620 '<FieldAccess for name_symbol at ast.py:9093>' Fld_37 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10619 '<FieldAccess for id at ast.py:9093>' Fld_36 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_36 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10619
      if Fld_36.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_37 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_36.Node);
--# end
--# expr-done 10620
--# expr-start 10621 '<Symbol No_Return at ???>' Sym_14 None
Sym_14 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_No_Return); 
--# expr-done 10621
Is_Equal_13 := Fld_37 = Sym_14; 
--# expr-done 10622
   If_Result_13 := Is_Equal_13;
end if;
if If_Result_13 then
   If_Result_14 := True;
else
   --# expr-start 10626 '<Eq at ast.py:9093>' Is_Equal_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10624 '<FieldAccess for name_symbol at ast.py:9093>' Fld_39 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10623 '<FieldAccess for id at ast.py:9093>' Fld_38 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_38 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10623
      if Fld_38.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_39 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_38.Node);
--# end
--# expr-done 10624
--# expr-start 10625 '<Symbol Discard_Names at ???>' Sym_15 None
Sym_15 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Discard_Names); 
--# expr-done 10625
Is_Equal_14 := Fld_39 = Sym_15; 
--# expr-done 10626
   If_Result_14 := Is_Equal_14;
end if;
if If_Result_14 then
   If_Result_15 := True;
else
   --# expr-start 10630 '<Eq at ast.py:9093>' Is_Equal_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10628 '<FieldAccess for name_symbol at ast.py:9093>' Fld_41 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10627 '<FieldAccess for id at ast.py:9093>' Fld_40 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_40 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10627
      if Fld_40.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_41 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_40.Node);
--# end
--# expr-done 10628
--# expr-start 10629 '<Symbol Independent at ???>' Sym_16 None
Sym_16 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Independent); 
--# expr-done 10629
Is_Equal_15 := Fld_41 = Sym_16; 
--# expr-done 10630
   If_Result_15 := Is_Equal_15;
end if;
if If_Result_15 then
   If_Result_16 := True;
else
   --# expr-start 10634 '<Eq at ast.py:9093>' Is_Equal_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10632 '<FieldAccess for name_symbol at ast.py:9093>' Fld_43 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10631 '<FieldAccess for id at ast.py:9093>' Fld_42 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_42 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10631
      if Fld_42.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_43 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_42.Node);
--# end
--# expr-done 10632
--# expr-start 10633 '<Symbol Independent_Components at ???>' Sym_17 None
Sym_17 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Independent_Components); 
--# expr-done 10633
Is_Equal_16 := Fld_43 = Sym_17; 
--# expr-done 10634
   If_Result_16 := Is_Equal_16;
end if;
if If_Result_16 then
   If_Result_17 := True;
else
   --# expr-start 10638 '<Eq at ast.py:9093>' Is_Equal_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10636 '<FieldAccess for name_symbol at ast.py:9093>' Fld_45 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10635 '<FieldAccess for id at ast.py:9093>' Fld_44 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_44 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10635
      if Fld_44.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_45 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_44.Node);
--# end
--# expr-done 10636
--# expr-start 10637 '<Symbol Asynchronous at ???>' Sym_18 None
Sym_18 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Asynchronous); 
--# expr-done 10637
Is_Equal_17 := Fld_45 = Sym_18; 
--# expr-done 10638
   If_Result_17 := Is_Equal_17;
end if;
if If_Result_17 then
   If_Result_18 := True;
else
   --# expr-start 10642 '<Eq at ast.py:9093>' Is_Equal_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10640 '<FieldAccess for name_symbol at ast.py:9093>' Fld_47 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10639 '<FieldAccess for id at ast.py:9093>' Fld_46 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_46 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10639
      if Fld_46.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_47 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_46.Node);
--# end
--# expr-done 10640
--# expr-start 10641 '<Symbol Interrupt_Handler at ???>' Sym_19 None
Sym_19 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Interrupt_Handler); 
--# expr-done 10641
Is_Equal_18 := Fld_47 = Sym_19; 
--# expr-done 10642
   If_Result_18 := Is_Equal_18;
end if;
if If_Result_18 then
   If_Result_19 := True;
else
   --# expr-start 10646 '<Eq at ast.py:9093>' Is_Equal_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10644 '<FieldAccess for name_symbol at ast.py:9093>' Fld_49 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
--# expr-start 10643 '<FieldAccess for id at ast.py:9093>' Fld_48 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9093
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_48 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10643
      if Fld_48.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_49 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_48.Node);
--# end
--# expr-done 10644
--# expr-start 10645 '<Symbol Attach_Handler at ???>' Sym_20 None
Sym_20 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Attach_Handler); 
--# expr-done 10645
Is_Equal_19 := Fld_49 = Sym_20; 
--# expr-done 10646
   If_Result_19 := Is_Equal_19;
end if;
if If_Result_19 then
--# expr-start 10649 '<Cast to Name at ast.py:9100>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9100
--# expr-start 10647 '<FieldAccess for args at ast.py:9100>' Fld_50 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9100
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_50 := Create_Internal_Entity_Base_Assoc_List (Node => Ent.Node.Pragma_Node_F_Args, Info => Ent.Info);
--# expr-done 10647
Fld_51 := Fld_50.Node;
Get_Result_1 := Get (Fld_51, 0, True); 
Is_Null_2 := Get_Result_1 = null; 
if Is_Null_2 then
   If_Result_20 := No_Entity_Base_Assoc;
else
Fld_52 := Fld_50.Info;
As_Entity_2 := (Info => Fld_52, Node => Get_Result_1); 
   If_Result_20 := As_Entity_2;
end if;
Var_Expr_1 := If_Result_20;
if Var_Expr_1 /= No_Entity_Base_Assoc then
   --# scope-start
--# expr-start 10648 '<FieldAccess for assoc_expr at ast.py:9100>' Fld_53 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9100
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseAssoc.assoc_expr'
Fld_53 := Libadalang.Implementation.Dispatcher_Base_Assoc_P_Assoc_Expr (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 10648
Scope_Result := Fld_53;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Expr;
end if;
   if Result_Var.Node = null
      or else Result_Var.Node.Kind in Ada_Name
   then
      Cast_Result_2 := Create_Internal_Entity_Name
        (Node => Result_Var.Node,
         Info => Result_Var.Info);
   else
         Cast_Result_2 := No_Entity_Name;
   end if;
--# expr-done 10649
Var_Expr := Cast_Result_2;
if Var_Expr /= No_Entity_Name then
   --# scope-start
                Singleton_1 := Create_Internal_Entity_Name_Array (Items_Count => 1);
                Singleton_1.Items (1) := Var_Expr;
Scope_Result_1 := Singleton_1;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_3054;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Internal_Entity_Name_Array_Type;
end if;
      Inc_Ref (Result_Var_1);
   If_Result_21 := Result_Var_1;
else
   If_Result_21 := No_Internal_Entity_Name_Array_Type;
end if;
      Inc_Ref (If_Result_21);
   If_Result_22 := If_Result_21;
end if;
      Inc_Ref (If_Result_22);
   If_Result_23 := If_Result_22;
end if;
      Inc_Ref (If_Result_23);
--# expr-done 10650
Let_Result := If_Result_23; Inc_Ref (Let_Result);
--# end
--# expr-done 10651
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_731;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_731;
               Finalizer_Scope_3054;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Pragma_Node_P_Associated_Entity_Names;
--# end
--# property-start Pragma.associated_entities_helper /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9106
pragma Warnings (Off, "is not referenced");
 function Pragma_Node_P_Associated_Entities_Helper
  (Node : Bare_Pragma_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Pragma_Node := Bare_Pragma_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Pragma_Node :=
     Internal_Entity_Pragma_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_732 with Inline_Always;
            procedure Finalizer_Scope_2890 with Inline_Always;
            procedure Finalizer_Scope_2891 with Inline_Always;
            procedure Finalizer_Scope_2892 with Inline_Always;
      Var_Expr : Internal_Entity;
Fld : Internal_Entity_Name_Array_Access;
Fld_1 : Internal_Entity;
Fld_2 : Lexical_Env := Empty_Env;
Fld_3 : Symbol_Type;
Env_Get_Result : Internal_Entity_Array_Access;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Fld_4 : Internal_Entity_Basic_Decl;
Fld_5 : Internal_Entity_Defining_Name_Array_Access;
Fld_6 : Symbol_Type;
Fld_7 : Boolean;
Find_Result : Internal_Entity_Defining_Name;
Map_Result : Internal_Entity_Defining_Name_Array_Access;
Scope_Result : Internal_Entity_Defining_Name_Array_Access;
Result_Var : Internal_Entity_Defining_Name_Array_Access;
Fld_8 : Internal_Unit;
Fld_9 : Internal_Unit;
Is_Equal : Boolean;
Fld_10 : Bare_Defining_Name;
Node_Comp : Boolean;
If_Result : Boolean;
Map_Result_1 : Internal_Entity_Defining_Name_Array_Access;
Map_Result_2 : Internal_Entity_Defining_Name_Array_Access;
Let_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_732 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Map_Result_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_732;
            procedure Finalizer_Scope_2890 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Map_Result_1);
            end Finalizer_Scope_2890;
            procedure Finalizer_Scope_2891 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Env_Get_Result);
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_2891;
            procedure Finalizer_Scope_2892 is
            begin
                     Dec_Ref (Fld_5);
            end Finalizer_Scope_2892;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10671 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10670 '<Mapcat at ast.py:9107>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9107
   --# expr-start 10652 '<FieldAccess for associated_entity_names at ast.py:9107>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9107
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Pragma.associated_entity_names
Fld := Libadalang.Implementation.Pragma_Node_P_Associated_Entity_Names (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 10652
      declare
         Map_Result_2_Vec : Internal_Entity_Defining_Name_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Name_Array_Access := Fld;
      begin
         for Item_316 of
               Collection.Items
         loop
   --# scope-start
               --# bind name Item_316
      --# expr-start 10669 '<Filter at ast.py:9111>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9111
--# expr-start 10653 '<FieldAccess for semantic_parent at ast.py:9111>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9111
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.semantic_parent
Fld_1 := Libadalang.Implementation.Ada_Node_P_Semantic_Parent (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 10653
Var_Expr := Fld_1;
if Var_Expr /= No_Entity then
   --# scope-start
   --# bind parent Var_Expr
--# expr-start 10663 '<Map at ast.py:9113>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9113
   --# expr-start 10656 '<EnvGet at ast.py:9113>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9113
--# expr-start 10654 '<FieldAccess for children_env at ast.py:9113>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9113
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Children_Env (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# expr-done 10654
--# expr-start 10655 '<FieldAccess for name_symbol at ast.py:9114>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9114
      if Item_316.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_3 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Item_316.Node);
--# end
--# expr-done 10655
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Fld_2, Key => Fld_3, Lookup_Kind => To_Lookup_Kind_Type (Flat), Categories => (Inherited_Primitives => False, Nocat => True))); 
--# expr-done 10656
      declare
         Map_Result_Vec : Internal_Entity_Defining_Name_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Env_Get_Result;
      begin
         for Item_318 of
               Collection.Items
         loop
   --# scope-start
               --# bind node Item_318
      --# expr-start 10662 '<Find at ast.py:9120>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9120
   --# expr-start 10659 '<FieldAccess for defining_names at ast.py:9120>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9120
--# expr-start 10658 '<FieldAccess for wrap_public_reference at ast.py:9120>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9120
--# expr-start 10657 '<Cast to BasicDecl at ast.py:9120>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9120
Cast_Expr := Item_318; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 10657
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.wrap_public_reference
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Wrap_Public_Reference (Node => Cast_Result.Node, E_Info => Cast_Result.Info);
--# end
--# expr-done 10658
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_names'
Fld_5 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Names (Node => Fld_4.Node, E_Info => Fld_4.Info);
--# end
--# expr-done 10659
      Find_Result := No_Entity_Defining_Name;
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Fld_5;
      begin
         for Item_319 of
               Collection.Items
         loop
   --# scope-start
               --# bind dn Item_319
      --# expr-start 10661 '<FieldAccess for name_is at ast.py:9122>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9122
      if Item_319.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10660 '<FieldAccess for name_symbol at ast.py:9122>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9122
      if Item_316.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_6 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Item_316.Node);
--# end
--# expr-done 10660
--# property-call-start Name.name_is
Fld_7 := Libadalang.Implementation.Name_P_Name_Is (Node => Item_319.Node, Sym => Fld_6);
--# end
--# expr-done 10661
      if Fld_7 then
         Find_Result := Item_319;
      end if;
   --# end
               exit when Fld_7;
         end loop;
      end;
--# expr-done 10662
         declare
            Item_To_Append : constant Internal_Entity_Defining_Name := Find_Result;
         begin
            Internal_Entity_Defining_Name_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2892;
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Defining_Name_Array
           (Items_Count => Natural (Internal_Entity_Defining_Name_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Defining_Name_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Defining_Name_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Defining_Name_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 10663
Scope_Result := Map_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2891;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Defining_Name_Array_Type;
end if;
      Inc_Ref (Result_Var);
      declare
         Map_Result_1_Vec : Internal_Entity_Defining_Name_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Defining_Name_Array_Access := Result_Var;
      begin
         for Item_317 of
               Collection.Items
         loop
   --# scope-start
               --# bind ent Item_317
--# expr-start 10666 '<Eq at ast.py:9128>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9128
--# expr-start 10664 '<FieldAccess for unit at ast.py:9128>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9128
      if Item_317.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Unit (Node => Item_317.Node);
--# expr-done 10664
--# expr-start 10665 '<FieldAccess for unit at ast.py:9128>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9128
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Libadalang.Implementation.Unit (Node => Self);
--# expr-done 10665
Is_Equal := Fld_8 = Fld_9; 
--# expr-done 10666
if Is_Equal then
   --# expr-start 10668 '<OrderingTest '"'"'lt'"'"' at ast.py:9128>' Node_Comp /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9128
--# expr-start 10667 '<FieldAccess for node at ast.py:9128>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9128
Fld_10 := Item_317.Node;
--# expr-done 10667
Node_Comp := Compare (Fld_10, Self, Less_Than); 
--# expr-done 10668
   If_Result := Node_Comp;
else
   If_Result := False;
end if;
         if If_Result then
         declare
            Item_To_Append : constant Internal_Entity_Defining_Name := Item_317;
         begin
            Internal_Entity_Defining_Name_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result_1 := Create_Internal_Entity_Defining_Name_Array
           (Items_Count => Natural (Internal_Entity_Defining_Name_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Entity_Defining_Name_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Entity_Defining_Name_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Entity_Defining_Name_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 10669
         for Item_To_Append of
               Map_Result_1.Items
         loop
            Internal_Entity_Defining_Name_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_2890;
         end loop;
      end;
         Map_Result_2 := Create_Internal_Entity_Defining_Name_Array
           (Items_Count => Natural (Internal_Entity_Defining_Name_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Internal_Entity_Defining_Name_Vectors.Get
              (Map_Result_2_Vec,
               I + Internal_Entity_Defining_Name_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Internal_Entity_Defining_Name_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 10670
Let_Result := Map_Result_2; Inc_Ref (Let_Result);
--# end
--# expr-done 10671
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_732;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_732;
               Finalizer_Scope_2890;
               Finalizer_Scope_2891;
               Finalizer_Scope_2892;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Pragma_Node_P_Associated_Entities_Helper;
--# end
--# property-start Pragma.associated_entities /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9132
pragma Warnings (Off, "is not referenced");
 function Pragma_Node_P_Associated_Entities
  (Node : Bare_Pragma_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Defining_Name_Array_Access
is
   Self : Bare_Pragma_Node := Bare_Pragma_Node (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Pragma_Node :=
     Internal_Entity_Pragma_Node'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_733 with Inline_Always;
            procedure Finalizer_Scope_2887 with Inline_Always;
            procedure Finalizer_Scope_2889 with Inline_Always;
            procedure Finalizer_Scope_2895 with Inline_Always;
            procedure Finalizer_Scope_2896 with Inline_Always;
            procedure Finalizer_Scope_2897 with Inline_Always;
            procedure Finalizer_Scope_2899 with Inline_Always;
            procedure Finalizer_Scope_2900 with Inline_Always;
            procedure Finalizer_Scope_2901 with Inline_Always;
      Block_Var : Internal_Entity_Defining_Name_Array_Access;
Block_Var_1 : Internal_Entity_Basic_Decl;
Var_Expr : Bare_Compilation_Unit;
Var_Expr_1 : Internal_Entity_Name_Array_Access;
Var_Expr_2 : Internal_Entity_Basic_Decl_Array_Access;
Var_Expr_3 : Bare_Declarative_Part;
Var_Expr_4 : Internal_Entity_Basic_Decl;
Var_Expr_5 : Bare_Ada_Node_Array_Access;
P : Internal_Entity_Defining_Name_Array_Access;
Var_Expr_6 : Internal_Entity_Defining_Name_Array_Access;
Var_Expr_7 : Internal_Entity_Basic_Decl;
Fld : Bare_Ada_Node;
Fld_1 : Bare_Ada_Node;
Cast_Result : Bare_Compilation_Unit;
Fld_2 : Bare_Ada_Node;
Cast_Result_1 : Bare_Library_Item;
Fld_3 : Bare_Basic_Decl;
As_Entity : Internal_Entity_Basic_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Decl;
Fld_4 : Internal_Entity_Defining_Name;
Singleton : Internal_Entity_Defining_Name_Array_Access;
Scope_Result : Internal_Entity_Defining_Name_Array_Access;
Result_Var : Internal_Entity_Defining_Name_Array_Access;
Node_Parents : Bare_Ada_Node_Array_Access;
Is_A : Boolean;
Find_Result : Bare_Ada_Node;
Cast_Result_2 : Bare_Basic_Decl;
As_Entity_1 : Internal_Entity_Basic_Decl;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity_Basic_Decl;
Fld_5 : Internal_Entity_Name_Array_Access;
Fld_6 : Internal_Entity_Defining_Name_Array_Access;
Scope_Result_1 : Internal_Entity_Defining_Name_Array_Access;
Result_Var_1 : Internal_Entity_Defining_Name_Array_Access;
Fld_7 : Internal_Entity_Defining_Name;
Singleton_1 : Internal_Entity_Defining_Name_Array_Access;
Len : Integer;
Is_Equal : Boolean;
Get_Result : Internal_Entity_Name;
Fld_8 : Internal_Entity_Basic_Decl;
Cast_Result_3 : Internal_Entity;
Cast_Expr : Internal_Entity_Basic_Decl;
Cast_Result_4 : Internal_Entity;
Is_Equiv : Boolean;
If_Result_2 : Boolean;
If_Result_3 : Internal_Entity_Defining_Name_Array_Access;
Scope_Result_2 : Internal_Entity_Defining_Name_Array_Access;
Result_Var_2 : Internal_Entity_Defining_Name_Array_Access;
Is_Equal_1 : Boolean;
Not_Val : Boolean;
If_Result_4 : Internal_Entity_Defining_Name_Array_Access;
Let_Result : Internal_Entity_Defining_Name_Array_Access;
Scope_Result_3 : Internal_Entity_Defining_Name_Array_Access;
Fld_9 : Bare_Declarative_Part;
Fld_10 : Bare_Ada_Node_List;
Item_322 : Bare_Ada_Node;
Is_A_1 : Boolean;
Node_Comp : Boolean;
If_Result_5 : Boolean;
Map_Result : Bare_Ada_Node_Array_Access;
Len_1 : Integer;
Arith_Result : Integer;
Get_Result_1 : Bare_Ada_Node;
Scope_Result_4 : Bare_Ada_Node;
Result_Var_3 : Bare_Ada_Node;
Cast_Result_5 : Bare_Basic_Decl;
As_Entity_2 : Internal_Entity_Basic_Decl;
Is_Null_2 : Boolean;
If_Result_6 : Internal_Entity_Basic_Decl;
Singleton_2 : Internal_Entity_Basic_Decl_Array_Access;
Scope_Result_5 : Internal_Entity_Basic_Decl_Array_Access;
Result_Var_4 : Internal_Entity_Basic_Decl_Array_Access;
Scope_Result_6 : Internal_Entity_Basic_Decl_Array_Access;
Result_Var_5 : Internal_Entity_Basic_Decl_Array_Access;
Scope_Result_7 : Internal_Entity_Basic_Decl_Array_Access;
Singleton_3 : Internal_Entity_Basic_Decl_Array_Access;
Result_Var_6 : Internal_Entity_Basic_Decl_Array_Access;
Singleton_4 : Internal_Entity_Basic_Decl_Array_Access;
Fld_11 : Internal_Entity_Identifier;
Fld_12 : Symbol_Type;
Fld_13 : Boolean;
If_Result_7 : Internal_Entity_Basic_Decl_Array_Access;
Fld_14 : Internal_Entity_Defining_Name;
Map_Result_1 : Internal_Entity_Defining_Name_Array_Access;
Result_Var_7 : Internal_Entity_Defining_Name_Array_Access;
Let_Result_1 : Internal_Entity_Defining_Name_Array_Access;
            procedure Finalizer_Scope_733 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Node_Parents);
                     Dec_Ref (Fld_5);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Scope_Result_6);
                     Dec_Ref (Result_Var_5);
                     Dec_Ref (Scope_Result_7);
                     Dec_Ref (Singleton_3);
                     Dec_Ref (Result_Var_6);
                     Dec_Ref (Singleton_4);
                     Dec_Ref (If_Result_7);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Result_Var_7);
                     Dec_Ref (Let_Result_1);
            end Finalizer_Scope_733;
            procedure Finalizer_Scope_2887 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2887;
            procedure Finalizer_Scope_2889 is
            begin
                     Dec_Ref (Var_Expr_1);
                     Dec_Ref (Fld_6);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (P);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Result_Var_2);
                     Dec_Ref (If_Result_4);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_2889;
            procedure Finalizer_Scope_2895 is
            begin
                     Dec_Ref (Var_Expr_6);
            end Finalizer_Scope_2895;
            procedure Finalizer_Scope_2896 is
            begin
                     Dec_Ref (Singleton_1);
                     Dec_Ref (If_Result_3);
            end Finalizer_Scope_2896;
            procedure Finalizer_Scope_2897 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Scope_Result_5);
                     Dec_Ref (Result_Var_4);
            end Finalizer_Scope_2897;
            procedure Finalizer_Scope_2899 is
            begin
                     Dec_Ref (Var_Expr_5);
            end Finalizer_Scope_2899;
            procedure Finalizer_Scope_2900 is
            begin
                     Dec_Ref (Singleton_2);
            end Finalizer_Scope_2900;
            procedure Finalizer_Scope_2901 is
            begin
                     Dec_Ref (Var_Expr_2);
            end Finalizer_Scope_2901;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10714 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 10674 '<Cast to CompilationUnit at ast.py:9137>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9137
--# expr-start 10673 '<parent at ast.py:9137>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9137
--# expr-start 10672 '<parent at ast.py:9137>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9137
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 10672
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Parent; 
--# expr-done 10673
   if Fld_1 = null
      or else Fld_1.Kind in Ada_Compilation_Unit_Range
   then
      Cast_Result := Fld_1;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 10674
Var_Expr := Cast_Result;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
   --# bind cu Var_Expr
--# expr-start 10679 '<FieldAccess for defining_name at ast.py:9138>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9138
--# expr-start 10678 '<as_entity at ast.py:9138>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9138
--# expr-start 10677 '<FieldAccess for item at ast.py:9138>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9138
--# expr-start 10676 '<Cast to LibraryItem at ast.py:9138>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9138
--# expr-start 10675 '<FieldAccess for body at ast.py:9138>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9138
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Var_Expr.Compilation_Unit_F_Body;
--# expr-done 10675
   if Fld_2 = null
      or else Fld_2.Kind in Ada_Library_Item_Range
   then
      Cast_Result_1 := Fld_2;
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 10676
      if Cast_Result_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Cast_Result_1.Library_Item_F_Item;
--# expr-done 10677
Is_Null := Fld_3 = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Decl;
else
As_Entity := (Info => E_Info, Node => Fld_3); 
   If_Result := As_Entity;
end if;
--# expr-done 10678
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => If_Result.Node, E_Info => If_Result.Info);
--# end
--# expr-done 10679
                Singleton := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton.Items (1) := Fld_4;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2887;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Defining_Name_Array_Type;
end if;
      Inc_Ref (Result_Var);
Block_Var := Result_Var; Inc_Ref (Block_Var);
--# bind top_level_decl Block_Var
--# expr-start 10684 '<as_entity at ast.py:9143>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9143
--# expr-start 10683 '<Cast to BasicDecl at ast.py:9143>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9143
--# expr-start 10682 '<Find at ast.py:9143>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9143
   --# expr-start 10680 '<FieldAccess for parents at ast.py:9143>' Node_Parents /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9143
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents := Parents (Self, True); 
--# expr-done 10680
      Find_Result := No_Bare_Ada_Node;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents;
      begin
         for Item_320 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_320
      --# expr-start 10681 '<IsA BasicDecl at ast.py:9144>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9144
Is_A := Item_320 /= null 
and then Item_320.Kind in Ada_Basic_Decl; 
--# expr-done 10681
      if Is_A then
         Find_Result := Item_320;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 10682
   if Find_Result = null
      or else Find_Result.Kind in Ada_Basic_Decl
   then
      Cast_Result_2 := Find_Result;
   else
         Cast_Result_2 := No_Bare_Ada_Node;
   end if;
--# expr-done 10683
Is_Null_1 := Cast_Result_2 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity_Basic_Decl;
else
As_Entity_1 := (Info => E_Info, Node => Cast_Result_2); 
   If_Result_1 := As_Entity_1;
end if;
--# expr-done 10684
Block_Var_1 := If_Result_1; 
--# bind enclosing_program_unit Block_Var_1
--# expr-start 10685 '<FieldAccess for associated_entity_names at ast.py:9158>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9158
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Pragma.associated_entity_names
Fld_5 := Libadalang.Implementation.Pragma_Node_P_Associated_Entity_Names (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 10685
Var_Expr_1 := Fld_5;
      Inc_Ref (Var_Expr_1);
if Var_Expr_1 /= No_Internal_Entity_Name_Array_Type then
   --# scope-start
   --# bind names Var_Expr_1
--# expr-start 10697 '<Let at ast.py:9159>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9159
--# scope-start
--# expr-start 10686 '<FieldAccess for associated_entities_helper at ast.py:9160>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9160
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Pragma.associated_entities_helper
Fld_6 := Libadalang.Implementation.Pragma_Node_P_Associated_Entities_Helper (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 10686
Var_Expr_6 := Fld_6;
      Inc_Ref (Var_Expr_6);
if Var_Expr_6 /= No_Internal_Entity_Defining_Name_Array_Type then
   --# scope-start
   --# bind e Var_Expr_6
Scope_Result_1 := Var_Expr_6;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2895;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := Block_Var;
end if;
      Inc_Ref (Result_Var_1);
P := Result_Var_1; Inc_Ref (P);
--# bind p P
--# expr-start 10696 '<If at ast.py:9161>' If_Result_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9161
--# expr-start 10688 '<Not at ast.py:9162>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9162
--# expr-start 10687 '<Eq at ast.py:9162>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9162
Is_Equal_1 := Equivalent (P, No_Internal_Entity_Defining_Name_Array_Type); 
--# expr-done 10687
Not_Val := not (Is_Equal_1); 
--# expr-done 10688
if Not_Val then
   If_Result_4 := P;
else
Var_Expr_7 := Block_Var_1;
if Var_Expr_7 /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind epu Var_Expr_7
--# expr-start 10695 '<If at ast.py:9164>' If_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9164
--# expr-start 10690 '<Eq at ast.py:9166>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9166
--# expr-start 10689 '<length at ast.py:9166>' Len /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9166
Len := Length (Var_Expr_1); 
--# expr-done 10689
Is_Equal := Len = 1; 
--# expr-done 10690
if Is_Equal then
   --# expr-start 10693 '<Eq at ast.py:9167>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9167
--# expr-start 10692 '<FieldAccess for referenced_decl at ast.py:9167>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9167
--# expr-start 10691 '<at at ast.py:9167>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9167
Get_Result := Get (Var_Expr_1, 0, True); 
--# expr-done 10691
      if Get_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.referenced_decl
Fld_8 := Libadalang.Implementation.Name_P_Referenced_Decl (Node => Get_Result.Node, Imprecise_Fallback => False, E_Info => Get_Result.Info);
--# end
--# expr-done 10692
      Cast_Result_3 := Create_Internal_Entity
        (Node => Fld_8.Node,
         Info => Fld_8.Info);
Cast_Expr := Var_Expr_7; 
      Cast_Result_4 := Create_Internal_Entity
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
Is_Equiv := Equivalent (Cast_Result_3, Cast_Result_4); 
--# expr-done 10693
   If_Result_2 := Is_Equiv;
else
   If_Result_2 := False;
end if;
if If_Result_2 then
   --# expr-start 10694 '<FieldAccess for defining_name at ast.py:9169>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9169
      if Var_Expr_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_7 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Var_Expr_7.Node, E_Info => Var_Expr_7.Info);
--# end
--# expr-done 10694
                Singleton_1 := Create_Internal_Entity_Defining_Name_Array (Items_Count => 1);
                Singleton_1.Items (1) := Fld_7;
   If_Result_3 := Singleton_1;
else
   If_Result_3 := No_Internal_Entity_Defining_Name_Array_Type;
end if;
      Inc_Ref (If_Result_3);
--# expr-done 10695
Scope_Result_2 := If_Result_3;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2896;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := Block_Var;
end if;
      Inc_Ref (Result_Var_2);
   If_Result_4 := Result_Var_2;
end if;
      Inc_Ref (If_Result_4);
--# expr-done 10696
Let_Result := If_Result_4; Inc_Ref (Let_Result);
--# end
--# expr-done 10697
Scope_Result_3 := Let_Result;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_2889;
   Result_Var_7 := Scope_Result_3;
else
   --# expr-start 10713 '<Map at ast.py:9174>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9174
   --# expr-start 10711 '<If at ast.py:9174>' If_Result_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9174
--# expr-start 10700 '<FieldAccess for is_contract_aspect at ast.py:9176>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9176
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10699 '<FieldAccess for name_symbol at ast.py:9176>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9176
--# expr-start 10698 '<FieldAccess for id at ast.py:9176>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9176
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Create_Internal_Entity_Identifier (Node => Ent.Node.Pragma_Node_F_Id, Info => Ent.Info);
--# expr-done 10698
      if Fld_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.name_symbol'
Fld_12 := Libadalang.Implementation.Dispatcher_Name_P_Name_Symbol (Node => Fld_11.Node);
--# end
--# expr-done 10699
--# property-call-start AdaNode.is_contract_aspect
Fld_13 := Libadalang.Implementation.Ada_Node_P_Is_Contract_Aspect (Node => Self, Name => Fld_12);
--# end
--# expr-done 10700
if Fld_13 then
--# expr-start 10701 '<FieldAccess for declarative_scope at ast.py:9180>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9180
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.declarative_scope
Fld_9 := Libadalang.Implementation.Ada_Node_P_Declarative_Scope (Node => Ent.Node);
--# end
--# expr-done 10701
Var_Expr_3 := Fld_9;
if Var_Expr_3 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind decl_scope Var_Expr_3
--# expr-start 10710 '<as_entity at ast.py:9181>' If_Result_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9181
--# expr-start 10709 '<Cast to BasicDecl at ast.py:9181>' Cast_Result_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9181
--# expr-start 10705 '<Filter at ast.py:9181>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9181
   --# expr-start 10702 '<FieldAccess for decls at ast.py:9181>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9181
      if Var_Expr_3 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Var_Expr_3.Declarative_Part_F_Decls;
--# expr-done 10702
      declare
         Map_Result_Vec : Bare_Ada_Node_Vectors.Vector;
      begin
      if Fld_10 = null then
      Map_Result := Create_Bare_Ada_Node_Array (0);
      else
      declare
         Collection : constant Bare_Ada_Node_List := Fld_10;
      begin
         for Untyped_Item_322 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_322 := Untyped_Item_322; 
   --# scope-start
               --# bind decl Item_322
--# expr-start 10703 '<IsA BasicDecl at ast.py:9182>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9182
Is_A_1 := Item_322 /= null 
and then Item_322.Kind in Ada_Basic_Decl; 
--# expr-done 10703
if Is_A_1 then
   --# expr-start 10704 '<OrderingTest '"'"'gt'"'"' at ast.py:9182>' Node_Comp /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9182
Node_Comp := Compare (Self, Item_322, Greater_Than); 
--# expr-done 10704
   If_Result_5 := Node_Comp;
else
   If_Result_5 := False;
end if;
         if If_Result_5 then
         declare
            Item_To_Append : constant Bare_Ada_Node := Item_322;
         begin
            Bare_Ada_Node_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Bare_Ada_Node_Array
           (Items_Count => Natural (Bare_Ada_Node_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Bare_Ada_Node_Vectors.Get
              (Map_Result_Vec,
               I + Bare_Ada_Node_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Bare_Ada_Node_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 10705
Var_Expr_5 := Map_Result;
      Inc_Ref (Var_Expr_5);
if Var_Expr_5 /= No_Bare_Ada_Node_Array_Type then
   --# scope-start
   --# bind decls Var_Expr_5
--# expr-start 10708 '<at at ast.py:9184>' Get_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9184
--# expr-start 10707 '<Arithmetic - at ast.py:9184>' Arith_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9184
--# expr-start 10706 '<length at ast.py:9184>' Len_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9184
Len_1 := Length (Var_Expr_5); 
--# expr-done 10706
Arith_Result := (Len_1 - 1); 
--# expr-done 10707
Get_Result_1 := Get (Var_Expr_5, Arith_Result, True); 
--# expr-done 10708
Scope_Result_4 := Get_Result_1;
   --# end
      Finalizer_Scope_2899;
   Result_Var_3 := Scope_Result_4;
else
   Result_Var_3 := No_Bare_Ada_Node;
end if;
   if Result_Var_3 = null
      or else Result_Var_3.Kind in Ada_Basic_Decl
   then
      Cast_Result_5 := Result_Var_3;
   else
         Cast_Result_5 := No_Bare_Ada_Node;
   end if;
--# expr-done 10709
Is_Null_2 := Cast_Result_5 = null; 
if Is_Null_2 then
   If_Result_6 := No_Entity_Basic_Decl;
else
As_Entity_2 := (Info => E_Info, Node => Cast_Result_5); 
   If_Result_6 := As_Entity_2;
end if;
--# expr-done 10710
Var_Expr_4 := If_Result_6;
if Var_Expr_4 /= No_Entity_Basic_Decl then
   --# scope-start
                Singleton_2 := Create_Internal_Entity_Basic_Decl_Array (Items_Count => 1);
                Singleton_2.Items (1) := Var_Expr_4;
Scope_Result_5 := Singleton_2;
   Inc_Ref (Scope_Result_5);
   --# end
      Finalizer_Scope_2900;
   Result_Var_4 := Scope_Result_5;
else
   Result_Var_4 := No_Internal_Entity_Basic_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var_4);
Scope_Result_6 := Result_Var_4;
   Inc_Ref (Scope_Result_6);
   --# end
      Finalizer_Scope_2897;
   Result_Var_5 := Scope_Result_6;
else
   Result_Var_5 := No_Internal_Entity_Basic_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var_5);
Var_Expr_2 := Result_Var_5;
      Inc_Ref (Var_Expr_2);
if Var_Expr_2 /= No_Internal_Entity_Basic_Decl_Array_Type then
   --# scope-start
   --# bind e Var_Expr_2
Scope_Result_7 := Var_Expr_2;
   Inc_Ref (Scope_Result_7);
   --# end
      Finalizer_Scope_2901;
   Result_Var_6 := Scope_Result_7;
else
                Singleton_3 := Create_Internal_Entity_Basic_Decl_Array (Items_Count => 1);
                Singleton_3.Items (1) := Block_Var_1;
   Result_Var_6 := Singleton_3;
end if;
      Inc_Ref (Result_Var_6);
   If_Result_7 := Result_Var_6;
else
                Singleton_4 := Create_Internal_Entity_Basic_Decl_Array (Items_Count => 1);
                Singleton_4.Items (1) := Block_Var_1;
   If_Result_7 := Singleton_4;
end if;
      Inc_Ref (If_Result_7);
--# expr-done 10711
      declare
         Map_Result_1_Vec : Internal_Entity_Defining_Name_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Basic_Decl_Array_Access := If_Result_7;
      begin
         for Item_321 of
               Collection.Items
         loop
   --# scope-start
               --# bind bd Item_321
      --# expr-start 10712 '<FieldAccess for defining_name at ast.py:9193>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9193
      if Item_321.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_14 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Item_321.Node, E_Info => Item_321.Info);
--# end
--# expr-done 10712
         declare
            Item_To_Append : constant Internal_Entity_Defining_Name := Fld_14;
         begin
            Internal_Entity_Defining_Name_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result_1 := Create_Internal_Entity_Defining_Name_Array
           (Items_Count => Natural (Internal_Entity_Defining_Name_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Entity_Defining_Name_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Entity_Defining_Name_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Entity_Defining_Name_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 10713
   Result_Var_7 := Map_Result_1;
end if;
      Inc_Ref (Result_Var_7);
Let_Result_1 := Result_Var_7; Inc_Ref (Let_Result_1);
--# end
--# expr-done 10714
      Property_Result := Let_Result_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_733;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_733;
               Finalizer_Scope_2887;
               Finalizer_Scope_2889;
               Finalizer_Scope_2895;
               Finalizer_Scope_2896;
               Finalizer_Scope_2897;
               Finalizer_Scope_2899;
               Finalizer_Scope_2900;
               Finalizer_Scope_2901;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Pragma_Node_P_Associated_Entities;
--# end
--# property-start Pragma.initial_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9197
pragma Warnings (Off, "is not referenced");
 function Pragma_Node_P_Initial_Env
  (Node : Bare_Pragma_Node
  )
   return Internal_Designated_Env
is
   Self : Bare_Pragma_Node := Bare_Pragma_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_734 with Inline_Always;
      Sym : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
New_Struct_1 : Internal_Designated_Env := No_Designated_Env;
Fld : Bare_Ada_Node;
Fld_1 : Bare_Ada_Node;
Is_A : Boolean;
If_Result : Internal_Designated_Env := No_Designated_Env;
Let_Result : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_734 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_734;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 10722 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10721 '<If at ast.py:9202>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9202
--# expr-start 10717 '<IsA CompilationUnit at ast.py:9203>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9203
--# expr-start 10716 '<parent at ast.py:9203>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9203
--# expr-start 10715 '<parent at ast.py:9203>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9203
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 10715
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Parent; 
--# expr-done 10716
Is_A := Fld_1 /= null 
and then Fld_1.Kind in Ada_Compilation_Unit_Range; 
--# expr-done 10717
if Is_A then
   --# expr-start 10719 '<New DesignatedEnv at ast.py:9204>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9204
--# expr-start 10718 '<Symbol Standard at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Standard); 
--# expr-done 10718
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => Sym, Kind => Named_Env); 
--# expr-done 10719
   If_Result := New_Struct;
else
   --# expr-start 10720 '<New DesignatedEnv at ast.py:9205>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9205
Inc_Ref (Empty_Env);
New_Struct_1 := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 10720
   If_Result := New_Struct_1;
end if;
      Inc_Ref (If_Result);
--# expr-done 10721
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 10722
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_734;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_734;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Pragma_Node_P_Initial_Env;
--# end
--# property-start Pragma._initial_env_36 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9209
pragma Warnings (Off, "is not referenced");
 function Initial_Env_36
  (Node : Bare_Pragma_Node
  )
   return Internal_Designated_Env
is
   Self : Bare_Pragma_Node := Bare_Pragma_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_1453 with Inline_Always;
      Fld : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_1453 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1453;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 10723 '<FieldAccess for initial_env at ast.py:9209>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:9209
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Pragma.initial_env
Fld := Libadalang.Implementation.Pragma_Node_P_Initial_Env (Node => Self);
--# end
--# expr-done 10723
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1453;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1453;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Initial_Env_36;
--# end
   --
   --  Primitives for Bare_Private_Node
   --
--# property-start '[dispatcher]Private.as_bool' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Private_Node_P_As_Bool
  (Node : Bare_Private_Node
  )
   return Boolean
is
   Self : Bare_Private_Node := Bare_Private_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Private_Node (Self.Kind) is
               when Ada_Private_Absent =>
                  --# property-call-start Private.Absent.as_bool
                  Property_Result := Private_Absent_P_As_Bool
                    (Self
                    );
                  --# end
               when Ada_Private_Present =>
                  --# property-call-start Private.Present.as_bool
                  Property_Result := Private_Present_P_As_Bool
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Private_Node_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Private_Absent
   --
--# property-start Private.Absent.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4882
pragma Warnings (Off, "is not referenced");
 function Private_Absent_P_As_Bool
  (Node : Bare_Private_Absent
  )
   return Boolean
is
   Self : Bare_Private_Absent := Bare_Private_Absent (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Private_Absent_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Private_Present
   --
--# property-start Private.Present.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4882
pragma Warnings (Off, "is not referenced");
 function Private_Present_P_As_Bool
  (Node : Bare_Private_Present
  )
   return Boolean
is
   Self : Bare_Private_Present := Bare_Private_Present (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Private_Present_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Protected_Def
   --
      procedure Initialize_Fields_For_Protected_Def
        (Self : Bare_Protected_Def
         ; Protected_Def_F_Public_Part : Bare_Public_Part
         ; Protected_Def_F_Private_Part : Bare_Private_Part
         ; Protected_Def_F_End_Name : Bare_End_Name
        ) is
      begin
            Self.Protected_Def_F_Public_Part := Protected_Def_F_Public_Part;
            Self.Protected_Def_F_Private_Part := Protected_Def_F_Private_Part;
            Self.Protected_Def_F_End_Name := Protected_Def_F_End_Name;
      end Initialize_Fields_For_Protected_Def;
   function Protected_Def_F_Public_Part
     (Node : Bare_Protected_Def) return Bare_Public_Part
   is
   begin
         return Node.Protected_Def_F_Public_Part;
   end;
   function Protected_Def_F_Private_Part
     (Node : Bare_Protected_Def) return Bare_Private_Part
   is
   begin
         return Node.Protected_Def_F_Private_Part;
   end;
   function Protected_Def_F_End_Name
     (Node : Bare_Protected_Def) return Bare_End_Name
   is
   begin
         return Node.Protected_Def_F_End_Name;
   end;
   --
   --  Primitives for Bare_Protected_Node
   --
--# property-start '[dispatcher]Protected.as_bool' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Protected_Node_P_As_Bool
  (Node : Bare_Protected_Node
  )
   return Boolean
is
   Self : Bare_Protected_Node := Bare_Protected_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Protected_Node (Self.Kind) is
               when Ada_Protected_Absent =>
                  --# property-call-start Protected.Absent.as_bool
                  Property_Result := Protected_Absent_P_As_Bool
                    (Self
                    );
                  --# end
               when Ada_Protected_Present =>
                  --# property-call-start Protected.Present.as_bool
                  Property_Result := Protected_Present_P_As_Bool
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Protected_Node_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Protected_Absent
   --
--# property-start Protected.Absent.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4962
pragma Warnings (Off, "is not referenced");
 function Protected_Absent_P_As_Bool
  (Node : Bare_Protected_Absent
  )
   return Boolean
is
   Self : Bare_Protected_Absent := Bare_Protected_Absent (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Protected_Absent_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Protected_Present
   --
--# property-start Protected.Present.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4962
pragma Warnings (Off, "is not referenced");
 function Protected_Present_P_As_Bool
  (Node : Bare_Protected_Present
  )
   return Boolean
is
   Self : Bare_Protected_Present := Bare_Protected_Present (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Protected_Present_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Quantifier
   --
   --
   --  Primitives for Bare_Quantifier_All
   --
   --
   --  Primitives for Bare_Quantifier_Some
   --
   --
   --  Primitives for Bare_Range_Spec
   --
      procedure Initialize_Fields_For_Range_Spec
        (Self : Bare_Range_Spec
         ; Range_Spec_F_Range : Bare_Expr
        ) is
      begin
            Self.Range_Spec_F_Range := Range_Spec_F_Range;
      end Initialize_Fields_For_Range_Spec;
   function Range_Spec_F_Range
     (Node : Bare_Range_Spec) return Bare_Expr
   is
   begin
         return Node.Range_Spec_F_Range;
   end;
--# property-start RangeSpec.xref_stop_resolution /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20159
pragma Warnings (Off, "is not referenced");
 function Range_Spec_P_Xref_Stop_Resolution
  (Node : Bare_Range_Spec
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Range_Spec := Bare_Range_Spec (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Is_A : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 10725 '<IsA ComponentClause at ast.py:20159>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20159
--# expr-start 10724 '<parent at ast.py:20159>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20159
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 10724
Is_A := Fld /= null 
and then Fld.Kind in Ada_Component_Clause_Range; 
--# expr-done 10725
      Property_Result := Is_A;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Range_Spec_P_Xref_Stop_Resolution;
--# end
--# property-start RangeSpec.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20160
pragma Warnings (Off, "is not referenced");
 function Range_Spec_P_Xref_Equation
  (Node : Bare_Range_Spec
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Range_Spec := Bare_Range_Spec (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Range_Spec :=
     Internal_Entity_Range_Spec'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1405 with Inline_Always;
      Fld : Internal_Entity_Expr;
Fld_1 : Logic_Equation;
Fld_2 : Bare_Expr;
Fld_3 : Logic_Var;
Fld_4 : Logic_Equation;
Fld_5 : Internal_Entity_Expr;
Fld_6 : Logic_Equation;
And_Pred : Logic_Equation;
True_Rel : Logic_Equation;
Fld_7 : Bare_Ada_Node;
Is_A : Boolean;
If_Result : Logic_Equation;
And_Pred_1 : Logic_Equation;
            procedure Finalizer_Scope_1405 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred);
                     Dec_Ref (True_Rel);
                     Dec_Ref (If_Result);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_1405;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10737 '<And at ast.py:20160>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20160
--# expr-start 10727 '<FieldAccess for xref_equation at ast.py:20160>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20160
--# expr-start 10726 '<FieldAccess for range at ast.py:20160>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20160
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Range_Spec_F_Range, Info => Ent.Info);
--# expr-done 10726
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_1 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 10727
--# expr-start 10736 '<If at ast.py:20160>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20160
--# expr-start 10729 '<IsA ComponentClause, SignedIntTypeDef at ast.py:20163>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20163
--# expr-start 10728 '<parent at ast.py:20163>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20163
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Self.Parent; 
--# expr-done 10728
Is_A := Fld_7 /= null 
and then Fld_7.Kind in Ada_Component_Clause_Range | Ada_Signed_Int_Type_Def_Range; 
--# expr-done 10729
if Is_A then
   --# expr-start 10735 '<And at ast.py:20165>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20165
--# expr-start 10732 '<FieldAccess for universal_int_bind at ast.py:20165>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20165
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10731 '<FieldAccess for expected_type_var at ast.py:20165>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20165
--# expr-start 10730 '<FieldAccess for range at ast.py:20165>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20165
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Range_Spec_F_Range;
--# expr-done 10730
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_2);
--# expr-done 10731
--# property-call-start AdaNode.universal_int_bind
Fld_4 := Libadalang.Implementation.Ada_Node_P_Universal_Int_Bind (Node => Self, Type_Var => Fld_3, Origin => Origin);
--# end
--# expr-done 10732
--# expr-start 10734 '<FieldAccess for matches_expected_type at ast.py:20166>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20166
--# expr-start 10733 '<FieldAccess for range at ast.py:20166>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20166
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Expr (Node => Ent.Node.Range_Spec_F_Range, Info => Ent.Info);
--# expr-done 10733
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_6 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_5.Node, Origin => Origin);
--# end
--# expr-done 10734
And_Pred := Create_And (Fld_4, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:20165") else null)); 
--# expr-done 10735
   If_Result := And_Pred;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:20168") else null)); 
   If_Result := True_Rel;
end if;
      Inc_Ref (If_Result);
--# expr-done 10736
And_Pred_1 := Create_And (Fld_1, If_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:20160") else null)); 
--# expr-done 10737
      Property_Result := And_Pred_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1405;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1405;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Range_Spec_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Renaming_Clause
   --
      procedure Initialize_Fields_For_Renaming_Clause
        (Self : Bare_Renaming_Clause
         ; Renaming_Clause_F_Renamed_Object : Bare_Name
        ) is
      begin
            Self.Renaming_Clause_F_Renamed_Object := Renaming_Clause_F_Renamed_Object;
      end Initialize_Fields_For_Renaming_Clause;
   function Renaming_Clause_F_Renamed_Object
     (Node : Bare_Renaming_Clause) return Bare_Name
   is
   begin
         return Node.Renaming_Clause_F_Renamed_Object;
   end;
   --
   --  Primitives for Bare_Synthetic_Renaming_Clause
   --
      procedure Initialize_Fields_For_Synthetic_Renaming_Clause
        (Self : Bare_Synthetic_Renaming_Clause
         ; Renaming_Clause_F_Renamed_Object : Bare_Name
        ) is
      begin
            Initialize_Fields_For_Renaming_Clause
              (Self, Renaming_Clause_F_Renamed_Object);
      end Initialize_Fields_For_Synthetic_Renaming_Clause;
   --
   --  Primitives for Bare_Reverse_Node
   --
--# property-start '[dispatcher]Reverse.as_bool' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Reverse_Node_P_As_Bool
  (Node : Bare_Reverse_Node
  )
   return Boolean
is
   Self : Bare_Reverse_Node := Bare_Reverse_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Reverse_Node (Self.Kind) is
               when Ada_Reverse_Absent =>
                  --# property-call-start Reverse.Absent.as_bool
                  Property_Result := Reverse_Absent_P_As_Bool
                    (Self
                    );
                  --# end
               when Ada_Reverse_Present =>
                  --# property-call-start Reverse.Present.as_bool
                  Property_Result := Reverse_Present_P_As_Bool
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Reverse_Node_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Reverse_Absent
   --
--# property-start Reverse.Absent.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4930
pragma Warnings (Off, "is not referenced");
 function Reverse_Absent_P_As_Bool
  (Node : Bare_Reverse_Absent
  )
   return Boolean
is
   Self : Bare_Reverse_Absent := Bare_Reverse_Absent (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Reverse_Absent_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Reverse_Present
   --
--# property-start Reverse.Present.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4930
pragma Warnings (Off, "is not referenced");
 function Reverse_Present_P_As_Bool
  (Node : Bare_Reverse_Present
  )
   return Boolean
is
   Self : Bare_Reverse_Present := Bare_Reverse_Present (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Reverse_Present_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Select_When_Part
   --
      procedure Initialize_Fields_For_Select_When_Part
        (Self : Bare_Select_When_Part
         ; Select_When_Part_F_Cond_Expr : Bare_Expr
         ; Select_When_Part_F_Stmts : Bare_Stmt_List
        ) is
      begin
            Self.Select_When_Part_F_Cond_Expr := Select_When_Part_F_Cond_Expr;
            Self.Select_When_Part_F_Stmts := Select_When_Part_F_Stmts;
      end Initialize_Fields_For_Select_When_Part;
   function Select_When_Part_F_Cond_Expr
     (Node : Bare_Select_When_Part) return Bare_Expr
   is
   begin
         return Node.Select_When_Part_F_Cond_Expr;
   end;
   function Select_When_Part_F_Stmts
     (Node : Bare_Select_When_Part) return Bare_Stmt_List
   is
   begin
         return Node.Select_When_Part_F_Stmts;
   end;
--# property-start SelectWhenPart.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19673
pragma Warnings (Off, "is not referenced");
 function Select_When_Part_P_Xref_Equation
  (Node : Bare_Select_When_Part
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Select_When_Part := Bare_Select_When_Part (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Select_When_Part :=
     Internal_Entity_Select_When_Part'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1373 with Inline_Always;
            procedure Finalizer_Scope_3055 with Inline_Always;
      Var_Expr : Internal_Entity_Expr;
Fld : Internal_Entity_Expr;
Fld_1 : Logic_Var;
Fld_2 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_3 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_4 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1373 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1373;
            procedure Finalizer_Scope_3055 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_3);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_4);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_3055;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10746 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10738 '<FieldAccess for cond_expr at ast.py:19674>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19674
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Select_When_Part_F_Cond_Expr, Info => Ent.Info);
--# expr-done 10738
Var_Expr := Fld;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind c Var_Expr
--# expr-start 10745 '<And at ast.py:19675>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19675
--# expr-start 10743 '<And at ast.py:19675>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19675
--# expr-start 10741 '<Bind at ast.py:19676>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19676
--# expr-start 10739 '<FieldAccess for expected_type_var at ast.py:19676>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19676
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Var_Expr.Node);
--# expr-done 10739
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 10740 '<FieldAccess for bool_type at ast.py:19676>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19676
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_2 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 10740
      Cast_Result := Create_Internal_Entity
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Bind_Result := Solver.Create_Assign (Fld_1, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19676") else null)); 
--# expr-done 10741
--# expr-start 10742 '<FieldAccess for sub_equation at ast.py:19677>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19677
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 10742
And_Pred := Create_And (Bind_Result, Fld_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19675") else null)); 
--# expr-done 10743
--# expr-start 10744 '<FieldAccess for matches_expected_formal_prim_type at ast.py:19678>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19678
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_prim_type
Fld_4 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Prim_Type (Node => Var_Expr.Node, Origin => Origin);
--# end
--# expr-done 10744
And_Pred_1 := Create_And (And_Pred, Fld_4, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19675") else null)); 
--# expr-done 10745
Scope_Result := And_Pred_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_3055;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19680") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 10746
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1373;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1373;
               Finalizer_Scope_3055;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Select_When_Part_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Stmt
   --
--# property-start Stmt.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19034
pragma Warnings (Off, "is not referenced");
 function Stmt_P_Xref_Entry_Point
  (Node : Bare_Stmt
  )
   return Boolean
is
   Self : Bare_Stmt := Bare_Stmt (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Stmt_P_Xref_Entry_Point;
--# end
--# property-start Stmt.is_ghost_code /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19037
pragma Warnings (Off, "is not referenced");
 function Stmt_P_Is_Ghost_Code
  (Node : Bare_Stmt
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Stmt := Bare_Stmt (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Stmt :=
     Internal_Entity_Stmt'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Basic_Decl;
Var_Expr_1 : Internal_Refd_Def;
Match_0 : Internal_Entity_Assign_Stmt;
Match_1 : Internal_Entity_Call_Stmt;
Match_2 : Internal_Entity_Stmt;
Var_Expr_2 : Internal_Entity_Defining_Name;
Fld : Internal_Entity_Basic_Decl;
Fld_1 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Fld_2 : Internal_Entity_Name;
Fld_3 : Internal_Refd_Def;
Fld_4 : Internal_Entity_Name;
Fld_5 : Internal_Refd_Def;
Match_Prefix : Internal_Entity_Stmt;
Cast_Result : Internal_Entity_Assign_Stmt;
Let_Result : Internal_Refd_Def;
Scope_Result_1 : Internal_Refd_Def;
Cast_Result_1 : Internal_Entity_Call_Stmt;
Let_Result_1 : Internal_Refd_Def;
Scope_Result_2 : Internal_Refd_Def;
Cast_Result_2 : Internal_Entity_Stmt;
Let_Result_2 : Internal_Refd_Def;
Scope_Result_3 : Internal_Refd_Def;
Match_Result : Internal_Refd_Def;
Fld_6 : Internal_Entity_Defining_Name;
Fld_7 : Boolean;
Scope_Result_4 : Boolean;
Result_Var_1 : Boolean;
Fld_8 : Ref_Result_Kind;
Is_Equal : Boolean;
If_Result : Boolean;
Scope_Result_5 : Boolean;
Result_Var_2 : Boolean;
If_Result_1 : Boolean;
Let_Result_3 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10759 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 10747 '<FieldAccess for parent_basic_decl at ast.py:19044>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19044
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.parent_basic_decl
Fld := Libadalang.Implementation.Ada_Node_P_Parent_Basic_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 10747
Var_Expr := Fld;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind bd Var_Expr
--# expr-start 10748 '<FieldAccess for is_ghost_code at ast.py:19045>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19045
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.is_ghost_code
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Is_Ghost_Code (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 10748
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
if Result_Var then
   If_Result_1 := True;
else
--# expr-start 10753 '<Match at ast.py:19050>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19050
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Stmt (Match_Prefix.Node.Kind) is
         when Ada_Assign_Stmt =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Assign_Stmt
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind ass Match_0
--# expr-start 10750 '<FieldAccess for failsafe_referenced_def_name at ast.py:19051>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19051
--# expr-start 10749 '<FieldAccess for dest at ast.py:19051>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19051
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Match_0.Node.Assign_Stmt_F_Dest, Info => Match_0.Info);
--# expr-done 10749
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.failsafe_referenced_def_name
Fld_3 := Libadalang.Implementation.Name_P_Failsafe_Referenced_Def_Name (Node => Fld_2.Node, Imprecise_Fallback => False, E_Info => Fld_2.Info);
--# end
--# expr-done 10750
Let_Result := Fld_3; 
--# end
Scope_Result_1 := Let_Result;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Call_Stmt =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Call_Stmt
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind call Match_1
--# expr-start 10752 '<FieldAccess for failsafe_referenced_def_name at ast.py:19052>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19052
--# expr-start 10751 '<FieldAccess for call at ast.py:19052>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19052
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Name (Node => Match_1.Node.Call_Stmt_F_Call, Info => Match_1.Info);
--# expr-done 10751
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.failsafe_referenced_def_name
Fld_5 := Libadalang.Implementation.Name_P_Failsafe_Referenced_Def_Name (Node => Fld_4.Node, Imprecise_Fallback => False, E_Info => Fld_4.Info);
--# end
--# expr-done 10752
Let_Result_1 := Fld_5; 
--# end
Scope_Result_2 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_2; 
         when Ada_Accept_Stmt .. Ada_Abort_Stmt | Ada_Delay_Stmt .. Ada_Terminate_Alternative =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Stmt
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind _ Match_2
Let_Result_2 := No_Refd_Def; 
--# end
Scope_Result_3 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_3; 
end case;
--# expr-done 10753
Var_Expr_1 := Match_Result;
if Var_Expr_1 /= No_Refd_Def then
   --# scope-start
   --# bind res Var_Expr_1
--# expr-start 10758 '<If at ast.py:19054>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19054
--# expr-start 10755 '<Eq at ast.py:19061>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19061
--# expr-start 10754 '<FieldAccess for kind at ast.py:19061>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19061
Fld_8 := Var_Expr_1.Kind;
--# expr-done 10754
Is_Equal := Fld_8 = Error; 
--# expr-done 10755
if Is_Equal then
   If_Result := raise Property_Error with "Name resolution error";
else
--# expr-start 10756 '<FieldAccess for def_name at ast.py:19063>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19063
Fld_6 := Var_Expr_1.Def_Name;
--# expr-done 10756
Var_Expr_2 := Fld_6;
if Var_Expr_2 /= No_Entity_Defining_Name then
   --# scope-start
--# expr-start 10757 '<FieldAccess for is_ghost_code at ast.py:19063>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19063
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start DefiningName.is_ghost_code
Fld_7 := Libadalang.Implementation.Defining_Name_P_Is_Ghost_Code (Node => Var_Expr_2.Node, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 10757
Scope_Result_4 := Fld_7;
   --# end
   Result_Var_1 := Scope_Result_4;
else
   Result_Var_1 := False;
end if;
   If_Result := Result_Var_1;
end if;
--# expr-done 10758
Scope_Result_5 := If_Result;
   --# end
   Result_Var_2 := Scope_Result_5;
else
   Result_Var_2 := False;
end if;
   If_Result_1 := Result_Var_2;
end if;
Let_Result_3 := If_Result_1; 
--# end
--# expr-done 10759
      Property_Result := Let_Result_3;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Stmt_P_Is_Ghost_Code;
--# end
   --
   --  Primitives for Bare_Composite_Stmt
   --
   --
   --  Primitives for Bare_Accept_Stmt
   --
      procedure Accept_Stmt_Pre_Env_Actions
        (Self            : Bare_Accept_Stmt;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_158 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Accept_Stmt
        (Self : Bare_Accept_Stmt
         ; Accept_Stmt_F_Name : Bare_Identifier
         ; Accept_Stmt_F_Entry_Index_Expr : Bare_Expr
         ; Accept_Stmt_F_Params : Bare_Entry_Completion_Formal_Params
        ) is
      begin
            Self.Accept_Stmt_F_Name := Accept_Stmt_F_Name;
            Self.Accept_Stmt_F_Entry_Index_Expr := Accept_Stmt_F_Entry_Index_Expr;
            Self.Accept_Stmt_F_Params := Accept_Stmt_F_Params;
      end Initialize_Fields_For_Accept_Stmt;
   function Accept_Stmt_F_Name
     (Node : Bare_Accept_Stmt) return Bare_Identifier
   is
   begin
         return Node.Accept_Stmt_F_Name;
   end;
   function Accept_Stmt_F_Entry_Index_Expr
     (Node : Bare_Accept_Stmt) return Bare_Expr
   is
   begin
         return Node.Accept_Stmt_F_Entry_Index_Expr;
   end;
   function Accept_Stmt_F_Params
     (Node : Bare_Accept_Stmt) return Bare_Entry_Completion_Formal_Params
   is
   begin
         return Node.Accept_Stmt_F_Params;
   end;
--# property-start AcceptStmt.designated_entry /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19615
pragma Warnings (Off, "is not referenced");
 function Accept_Stmt_P_Designated_Entry
  (Node : Bare_Accept_Stmt
      ; Origin : Bare_Ada_Node
      ; Env : Lexical_Env
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Entry_Decl
is
   Self : Bare_Accept_Stmt := Bare_Accept_Stmt (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Accept_Stmt :=
     Internal_Entity_Accept_Stmt'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind env Env
   Property_Result : Internal_Entity_Entry_Decl;
            procedure Finalizer_Scope_1369 with Inline_Always;
      Var_Expr : Internal_Entity_Entry_Decl;
Fld : Internal_Entity_Identifier;
Fld_1 : Internal_Entity_Array_Access;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Entry_Decl;
Fld_2 : Internal_Entity_Entry_Spec;
Fld_3 : Internal_Entity_Entry_Completion_Formal_Params;
Cast_Result_1 : Internal_Entity_Base_Formal_Param_Holder;
Fld_4 : Boolean;
Scope_Result : Boolean;
Result_Var : Boolean;
Find_Result : Internal_Entity;
Cast_Result_2 : Internal_Entity_Entry_Decl;
Let_Result : Internal_Entity_Entry_Decl;
            procedure Finalizer_Scope_1369 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1369;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10768 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10767 '<Cast to EntryDecl at ast.py:19616>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19616
--# expr-start 10766 '<Find at ast.py:19616>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19616
   --# expr-start 10761 '<FieldAccess for all_env_els_impl at ast.py:19616>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19616
--# expr-start 10760 '<FieldAccess for name at ast.py:19616>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19616
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Identifier (Node => Ent.Node.Accept_Stmt_F_Name, Info => Ent.Info);
--# expr-done 10760
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.all_env_els_impl'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_All_Env_Els_Impl (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 10761
      Find_Result := No_Entity;
      declare
         Collection : constant Internal_Entity_Array_Access := Fld_1;
      begin
         for Item_323 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_323
--# expr-start 10762 '<Cast to EntryDecl at ast.py:19617>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19617
Cast_Expr := Item_323; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Entry_Decl_Range
   then
      Cast_Result := Create_Internal_Entity_Entry_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Entry_Decl;
   end if;
--# expr-done 10762
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Entry_Decl then
   --# scope-start
   --# bind d Var_Expr
--# expr-start 10765 '<FieldAccess for match_formal_params at ast.py:19618>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19618
--# expr-start 10763 '<FieldAccess for spec at ast.py:19618>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19618
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Entry_Spec (Node => Var_Expr.Node.Entry_Decl_F_Spec, Info => Var_Expr.Info);
--# expr-done 10763
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10764 '<FieldAccess for params at ast.py:19618>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19618
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Entry_Completion_Formal_Params (Node => Ent.Node.Accept_Stmt_F_Params, Info => Ent.Info);
--# expr-done 10764
      Cast_Result_1 := Create_Internal_Entity_Base_Formal_Param_Holder
        (Node => Fld_3.Node,
         Info => Fld_3.Info);
--# property-call-start BaseFormalParamHolder.match_formal_params
Fld_4 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Match_Formal_Params (Node => Fld_2.Node, Other => Cast_Result_1, E_Info => Fld_2.Info);
--# end
--# expr-done 10765
Scope_Result := Fld_4;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := False;
end if;
      if Result_Var then
         Find_Result := Item_323;
      end if;
   --# end
               exit when Result_Var;
         end loop;
      end;
--# expr-done 10766
   if Find_Result.Node = null
      or else Find_Result.Node.Kind in Ada_Entry_Decl_Range
   then
      Cast_Result_2 := Create_Internal_Entity_Entry_Decl
        (Node => Find_Result.Node,
         Info => Find_Result.Info);
   else
         Cast_Result_2 := No_Entity_Entry_Decl;
   end if;
--# expr-done 10767
Let_Result := Cast_Result_2; 
--# end
--# expr-done 10768
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1369;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1369;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Accept_Stmt_P_Designated_Entry;
--# end
--# property-start AcceptStmt.corresponding_entry /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19624
pragma Warnings (Off, "is not referenced");
 function Accept_Stmt_P_Corresponding_Entry
  (Node : Bare_Accept_Stmt
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Entry_Decl
is
   Self : Bare_Accept_Stmt := Bare_Accept_Stmt (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Accept_Stmt :=
     Internal_Entity_Accept_Stmt'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Entry_Decl;
            procedure Finalizer_Scope_1370 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_1 : Internal_Entity_Entry_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Entry_Decl;
Let_Result : Internal_Entity_Entry_Decl;
            procedure Finalizer_Scope_1370 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Bound_Env);
            end Finalizer_Scope_1370;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10772 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10771 '<bind at ast.py:19628>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19628
--# expr-start 10769 '<FieldAccess for node_env at ast.py:19628>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19628
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Node_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 10769
Bound_Env := Fld; Inc_Ref (Bound_Env);
--# expr-start 10770 '<FieldAccess for designated_entry at ast.py:19628>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19628
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AcceptStmt.designated_entry
Fld_1 := Libadalang.Implementation.Accept_Stmt_P_Designated_Entry (Node => Ent.Node, Origin => Origin, Env => Bound_Env, E_Info => Ent.Info);
--# end
--# expr-done 10770
Dyn_Var_Bind_Result := Fld_1; 
--# expr-done 10771
Let_Result := Dyn_Var_Bind_Result; 
--# end
--# expr-done 10772
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1370;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1370;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Accept_Stmt_P_Corresponding_Entry;
--# end
--# property-start AcceptStmt.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19631
pragma Warnings (Off, "is not referenced");
 function Accept_Stmt_P_Xref_Equation
  (Node : Bare_Accept_Stmt
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Accept_Stmt := Bare_Accept_Stmt (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Accept_Stmt :=
     Internal_Entity_Accept_Stmt'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1371 with Inline_Always;
            procedure Finalizer_Scope_3064 with Inline_Always;
      Var_Expr : Internal_Entity_Expr;
Fld : Bare_Identifier;
Fld_1 : Logic_Var;
Fld_2 : Internal_Entity_Entry_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_3 : Internal_Entity_Expr;
Fld_4 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1371 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1371;
            procedure Finalizer_Scope_3064 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_3064;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10780 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10779 '<And at ast.py:19632>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19632
--# expr-start 10776 '<Bind at ast.py:19633>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19633
--# expr-start 10774 '<FieldAccess for ref_var at ast.py:19633>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19633
--# expr-start 10773 '<FieldAccess for name at ast.py:19633>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19633
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Accept_Stmt_F_Name;
--# expr-done 10773
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld);
--# end
--# expr-done 10774
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 10775 '<FieldAccess for designated_entry at ast.py:19633>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19633
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AcceptStmt.designated_entry
Fld_2 := Libadalang.Implementation.Accept_Stmt_P_Designated_Entry (Node => Ent.Node, Origin => Origin, Env => Env, E_Info => Ent.Info);
--# end
--# expr-done 10775
      Cast_Result := Create_Internal_Entity
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Bind_Result := Solver.Create_Assign (Fld_1, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19633") else null)); 
--# expr-done 10776
--# expr-start 10777 '<FieldAccess for entry_index_expr at ast.py:19634>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19634
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Expr (Node => Ent.Node.Accept_Stmt_F_Entry_Index_Expr, Info => Ent.Info);
--# expr-done 10777
Var_Expr := Fld_3;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind e Var_Expr
--# expr-start 10778 '<FieldAccess for sub_equation at ast.py:19635>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19635
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_4 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 10778
Scope_Result := Fld_4;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_3064;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19636") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred := Create_And (Bind_Result, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19632") else null)); 
--# expr-done 10779
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 10780
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1371;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1371;
               Finalizer_Scope_3064;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Accept_Stmt_P_Xref_Equation;
--# end
--# property-start AcceptStmt._env_trans_parent_158 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19611
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_158
  (Node : Bare_Accept_Stmt
  )
   return Boolean
is
   Self : Bare_Accept_Stmt := Bare_Accept_Stmt (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_158;
--# end
   --
   --  Primitives for Bare_Accept_Stmt_With_Stmts
   --
      procedure Initialize_Fields_For_Accept_Stmt_With_Stmts
        (Self : Bare_Accept_Stmt_With_Stmts
         ; Accept_Stmt_F_Name : Bare_Identifier
         ; Accept_Stmt_F_Entry_Index_Expr : Bare_Expr
         ; Accept_Stmt_F_Params : Bare_Entry_Completion_Formal_Params
         ; Accept_Stmt_With_Stmts_F_Stmts : Bare_Handled_Stmts
         ; Accept_Stmt_With_Stmts_F_End_Name : Bare_End_Name
        ) is
      begin
            Initialize_Fields_For_Accept_Stmt
              (Self, Accept_Stmt_F_Name, Accept_Stmt_F_Entry_Index_Expr, Accept_Stmt_F_Params);
            Self.Accept_Stmt_With_Stmts_F_Stmts := Accept_Stmt_With_Stmts_F_Stmts;
            Self.Accept_Stmt_With_Stmts_F_End_Name := Accept_Stmt_With_Stmts_F_End_Name;
      end Initialize_Fields_For_Accept_Stmt_With_Stmts;
   function Accept_Stmt_With_Stmts_F_Stmts
     (Node : Bare_Accept_Stmt_With_Stmts) return Bare_Handled_Stmts
   is
   begin
         return Node.Accept_Stmt_With_Stmts_F_Stmts;
   end;
   function Accept_Stmt_With_Stmts_F_End_Name
     (Node : Bare_Accept_Stmt_With_Stmts) return Bare_End_Name
   is
   begin
         return Node.Accept_Stmt_With_Stmts_F_End_Name;
   end;
   --
   --  Primitives for Bare_Base_Loop_Stmt
   --
      procedure Initialize_Fields_For_Base_Loop_Stmt
        (Self : Bare_Base_Loop_Stmt
         ; Base_Loop_Stmt_F_Spec : Bare_Loop_Spec
         ; Base_Loop_Stmt_F_Stmts : Bare_Stmt_List
         ; Base_Loop_Stmt_F_End_Name : Bare_End_Name
        ) is
      begin
            Self.Base_Loop_Stmt_F_Spec := Base_Loop_Stmt_F_Spec;
            Self.Base_Loop_Stmt_F_Stmts := Base_Loop_Stmt_F_Stmts;
            Self.Base_Loop_Stmt_F_End_Name := Base_Loop_Stmt_F_End_Name;
      end Initialize_Fields_For_Base_Loop_Stmt;
   function Base_Loop_Stmt_F_Spec
     (Node : Bare_Base_Loop_Stmt) return Bare_Loop_Spec
   is
   begin
         return Node.Base_Loop_Stmt_F_Spec;
   end;
   function Base_Loop_Stmt_F_Stmts
     (Node : Bare_Base_Loop_Stmt) return Bare_Stmt_List
   is
   begin
         return Node.Base_Loop_Stmt_F_Stmts;
   end;
   function Base_Loop_Stmt_F_End_Name
     (Node : Bare_Base_Loop_Stmt) return Bare_End_Name
   is
   begin
         return Node.Base_Loop_Stmt_F_End_Name;
   end;
--# property-start BaseLoopStmt.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19450
pragma Warnings (Off, "is not referenced");
 function Base_Loop_Stmt_P_Xref_Equation
  (Node : Bare_Base_Loop_Stmt
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Base_Loop_Stmt := Bare_Base_Loop_Stmt (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Base_Loop_Stmt :=
     Internal_Entity_Base_Loop_Stmt'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1361 with Inline_Always;
            procedure Finalizer_Scope_3065 with Inline_Always;
      Var_Expr : Internal_Entity_Loop_Spec;
Fld : Internal_Entity_Loop_Spec;
Fld_1 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1361 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1361;
            procedure Finalizer_Scope_3065 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_3065;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10783 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10781 '<FieldAccess for spec at ast.py:19451>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19451
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Loop_Spec (Node => Ent.Node.Base_Loop_Stmt_F_Spec, Info => Ent.Info);
--# expr-done 10781
Var_Expr := Fld;
if Var_Expr /= No_Entity_Loop_Spec then
   --# scope-start
   --# bind s Var_Expr
--# expr-start 10782 '<FieldAccess for xref_equation at ast.py:19451>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19451
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_1 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 10782
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_3065;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19452") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 10783
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1361;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1361;
               Finalizer_Scope_3065;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Base_Loop_Stmt_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_For_Loop_Stmt
   --
      procedure For_Loop_Stmt_Pre_Env_Actions
        (Self            : Bare_For_Loop_Stmt;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_155 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_For_Loop_Stmt
        (Self : Bare_For_Loop_Stmt
         ; Base_Loop_Stmt_F_Spec : Bare_Loop_Spec
         ; Base_Loop_Stmt_F_Stmts : Bare_Stmt_List
         ; Base_Loop_Stmt_F_End_Name : Bare_End_Name
        ) is
      begin
            Initialize_Fields_For_Base_Loop_Stmt
              (Self, Base_Loop_Stmt_F_Spec, Base_Loop_Stmt_F_Stmts, Base_Loop_Stmt_F_End_Name);
      end Initialize_Fields_For_For_Loop_Stmt;
--# property-start ForLoopStmt._env_trans_parent_155 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19469
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_155
  (Node : Bare_For_Loop_Stmt
  )
   return Boolean
is
   Self : Bare_For_Loop_Stmt := Bare_For_Loop_Stmt (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_155;
--# end
   --
   --  Primitives for Bare_Loop_Stmt
   --
      procedure Initialize_Fields_For_Loop_Stmt
        (Self : Bare_Loop_Stmt
         ; Base_Loop_Stmt_F_Spec : Bare_Loop_Spec
         ; Base_Loop_Stmt_F_Stmts : Bare_Stmt_List
         ; Base_Loop_Stmt_F_End_Name : Bare_End_Name
        ) is
      begin
            Initialize_Fields_For_Base_Loop_Stmt
              (Self, Base_Loop_Stmt_F_Spec, Base_Loop_Stmt_F_Stmts, Base_Loop_Stmt_F_End_Name);
      end Initialize_Fields_For_Loop_Stmt;
   --
   --  Primitives for Bare_While_Loop_Stmt
   --
      procedure Initialize_Fields_For_While_Loop_Stmt
        (Self : Bare_While_Loop_Stmt
         ; Base_Loop_Stmt_F_Spec : Bare_Loop_Spec
         ; Base_Loop_Stmt_F_Stmts : Bare_Stmt_List
         ; Base_Loop_Stmt_F_End_Name : Bare_End_Name
        ) is
      begin
            Initialize_Fields_For_Base_Loop_Stmt
              (Self, Base_Loop_Stmt_F_Spec, Base_Loop_Stmt_F_Stmts, Base_Loop_Stmt_F_End_Name);
      end Initialize_Fields_For_While_Loop_Stmt;
   --
   --  Primitives for Bare_Block_Stmt
   --
      procedure Block_Stmt_Pre_Env_Actions
        (Self            : Bare_Block_Stmt;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_156 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
--# property-start BlockStmt.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19489
pragma Warnings (Off, "is not referenced");
 function Block_Stmt_P_Xref_Equation
  (Node : Bare_Block_Stmt
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Block_Stmt := Bare_Block_Stmt (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1362 with Inline_Always;
      True_Rel : Logic_Equation;
            procedure Finalizer_Scope_1362 is
            begin
                     Dec_Ref (True_Rel);
            end Finalizer_Scope_1362;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19489") else null)); 
      Property_Result := True_Rel;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1362;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1362;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Block_Stmt_P_Xref_Equation;
--# end
--# property-start BlockStmt._env_trans_parent_156 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19487
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_156
  (Node : Bare_Block_Stmt
  )
   return Boolean
is
   Self : Bare_Block_Stmt := Bare_Block_Stmt (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_156;
--# end
   --
   --  Primitives for Bare_Begin_Block
   --
      procedure Initialize_Fields_For_Begin_Block
        (Self : Bare_Begin_Block
         ; Begin_Block_F_Stmts : Bare_Handled_Stmts
         ; Begin_Block_F_End_Name : Bare_End_Name
        ) is
      begin
            Self.Begin_Block_F_Stmts := Begin_Block_F_Stmts;
            Self.Begin_Block_F_End_Name := Begin_Block_F_End_Name;
      end Initialize_Fields_For_Begin_Block;
   function Begin_Block_F_Stmts
     (Node : Bare_Begin_Block) return Bare_Handled_Stmts
   is
   begin
         return Node.Begin_Block_F_Stmts;
   end;
   function Begin_Block_F_End_Name
     (Node : Bare_Begin_Block) return Bare_End_Name
   is
   begin
         return Node.Begin_Block_F_End_Name;
   end;
   --
   --  Primitives for Bare_Decl_Block
   --
      procedure Initialize_Fields_For_Decl_Block
        (Self : Bare_Decl_Block
         ; Decl_Block_F_Decls : Bare_Declarative_Part
         ; Decl_Block_F_Stmts : Bare_Handled_Stmts
         ; Decl_Block_F_End_Name : Bare_End_Name
        ) is
      begin
            Self.Decl_Block_F_Decls := Decl_Block_F_Decls;
            Self.Decl_Block_F_Stmts := Decl_Block_F_Stmts;
            Self.Decl_Block_F_End_Name := Decl_Block_F_End_Name;
      end Initialize_Fields_For_Decl_Block;
   function Decl_Block_F_Decls
     (Node : Bare_Decl_Block) return Bare_Declarative_Part
   is
   begin
         return Node.Decl_Block_F_Decls;
   end;
   function Decl_Block_F_Stmts
     (Node : Bare_Decl_Block) return Bare_Handled_Stmts
   is
   begin
         return Node.Decl_Block_F_Stmts;
   end;
   function Decl_Block_F_End_Name
     (Node : Bare_Decl_Block) return Bare_End_Name
   is
   begin
         return Node.Decl_Block_F_End_Name;
   end;
--# property-start DeclBlock.immediate_declarative_region /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19502
pragma Warnings (Off, "is not referenced");
 function Decl_Block_P_Immediate_Declarative_Region
  (Node : Bare_Decl_Block
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Decl_Block := Bare_Decl_Block (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Decl_Block :=
     Internal_Entity_Decl_Block'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_1363 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_1363 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1363;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10785 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10784 '<FieldAccess for children_env at ast.py:19503>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19503
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 10784
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 10785
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1363;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1363;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Decl_Block_P_Immediate_Declarative_Region;
--# end
   --
   --  Primitives for Bare_Case_Stmt
   --
      procedure Initialize_Fields_For_Case_Stmt
        (Self : Bare_Case_Stmt
         ; Case_Stmt_F_Expr : Bare_Expr
         ; Case_Stmt_F_Pragmas : Bare_Pragma_Node_List
         ; Case_Stmt_F_Alternatives : Bare_Case_Stmt_Alternative_List
        ) is
      begin
            Self.Case_Stmt_F_Expr := Case_Stmt_F_Expr;
            Self.Case_Stmt_F_Pragmas := Case_Stmt_F_Pragmas;
            Self.Case_Stmt_F_Alternatives := Case_Stmt_F_Alternatives;
      end Initialize_Fields_For_Case_Stmt;
   function Case_Stmt_F_Expr
     (Node : Bare_Case_Stmt) return Bare_Expr
   is
   begin
         return Node.Case_Stmt_F_Expr;
   end;
   function Case_Stmt_F_Pragmas
     (Node : Bare_Case_Stmt) return Bare_Pragma_Node_List
   is
   begin
         return Node.Case_Stmt_F_Pragmas;
   end;
   function Case_Stmt_F_Alternatives
     (Node : Bare_Case_Stmt) return Bare_Case_Stmt_Alternative_List
   is
   begin
         return Node.Case_Stmt_F_Alternatives;
   end;
--# property-start CaseStmt.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19540
pragma Warnings (Off, "is not referenced");
 function Case_Stmt_P_Xref_Equation
  (Node : Bare_Case_Stmt
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Case_Stmt := Bare_Case_Stmt (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Case_Stmt :=
     Internal_Entity_Case_Stmt'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1365 with Inline_Always;
      Fld : Internal_Entity_Expr;
Fld_1 : Logic_Equation;
Fld_2 : Bare_Expr;
Fld_3 : Logic_Var;
Pred : Logic_Equation;
Fld_4 : Bare_Expr;
Fld_5 : Logic_Var;
Pred_1 : Logic_Equation;
And_Pred : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1365 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Pred);
                     Dec_Ref (Pred_1);
                     Dec_Ref (And_Pred);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1365;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10796 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10795 '<And at ast.py:19541>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19541
--# expr-start 10787 '<FieldAccess for sub_equation at ast.py:19541>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19541
--# expr-start 10786 '<FieldAccess for expr at ast.py:19541>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19541
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Case_Stmt_F_Expr, Info => Ent.Info);
--# expr-done 10786
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 10787
--# expr-start 10794 '<And at ast.py:19545>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19545
--# expr-start 10790 '<Predicate on AdaNode.is_not_null at ast.py:19545>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19545
--# expr-start 10789 '<FieldAccess for type_var at ast.py:19545>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19545
--# expr-start 10788 '<FieldAccess for expr at ast.py:19545>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19545
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Case_Stmt_F_Expr;
--# expr-done 10788
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_2);
--# expr-done 10789
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
Pred := Solver.Create_Predicate (Fld_3, Create_Ada_Node_P_Is_Not_Null_0_Predicate); 
--# expr-done 10790
--# expr-start 10793 '<Predicate on BaseTypeDecl.is_discrete_type at ast.py:19547>' Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19547
--# expr-start 10792 '<FieldAccess for type_var at ast.py:19547>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19547
--# expr-start 10791 '<FieldAccess for expr at ast.py:19547>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19547
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Self.Case_Stmt_F_Expr;
--# expr-done 10791
      if Fld_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_4);
--# expr-done 10792
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
Pred_1 := Solver.Create_Predicate (Fld_5, Create_Base_Type_Decl_P_Is_Discrete_Type_0_Predicate (Origin)); 
--# expr-done 10793
And_Pred := Create_And (Pred, Pred_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19545") else null)); 
--# expr-done 10794
And_Pred_1 := Create_And (Fld_1, And_Pred, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19541") else null)); 
--# expr-done 10795
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 10796
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1365;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1365;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Case_Stmt_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Extended_Return_Stmt
   --
      procedure Extended_Return_Stmt_Pre_Env_Actions
        (Self            : Bare_Extended_Return_Stmt;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_157 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Extended_Return_Stmt
        (Self : Bare_Extended_Return_Stmt
         ; Extended_Return_Stmt_F_Decl : Bare_Extended_Return_Stmt_Object_Decl
         ; Extended_Return_Stmt_F_Stmts : Bare_Handled_Stmts
        ) is
      begin
            Self.Extended_Return_Stmt_F_Decl := Extended_Return_Stmt_F_Decl;
            Self.Extended_Return_Stmt_F_Stmts := Extended_Return_Stmt_F_Stmts;
      end Initialize_Fields_For_Extended_Return_Stmt;
   function Extended_Return_Stmt_F_Decl
     (Node : Bare_Extended_Return_Stmt) return Bare_Extended_Return_Stmt_Object_Decl
   is
   begin
         return Node.Extended_Return_Stmt_F_Decl;
   end;
   function Extended_Return_Stmt_F_Stmts
     (Node : Bare_Extended_Return_Stmt) return Bare_Handled_Stmts
   is
   begin
         return Node.Extended_Return_Stmt_F_Stmts;
   end;
--# property-start ExtendedReturnStmt.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19524
pragma Warnings (Off, "is not referenced");
 function Extended_Return_Stmt_P_Xref_Equation
  (Node : Bare_Extended_Return_Stmt
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Extended_Return_Stmt := Bare_Extended_Return_Stmt (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1364 with Inline_Always;
      True_Rel : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1364 is
            begin
                     Dec_Ref (True_Rel);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1364;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 10797 '<Block at ???>' Let_Result None
--# scope-start
True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19525") else null)); 
Let_Result := True_Rel; Inc_Ref (Let_Result);
--# end
--# expr-done 10797
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1364;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1364;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Extended_Return_Stmt_P_Xref_Equation;
--# end
--# property-start ExtendedReturnStmt._env_trans_parent_157 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19527
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_157
  (Node : Bare_Extended_Return_Stmt
  )
   return Boolean
is
   Self : Bare_Extended_Return_Stmt := Bare_Extended_Return_Stmt (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_157;
--# end
   --
   --  Primitives for Bare_If_Stmt
   --
      procedure Initialize_Fields_For_If_Stmt
        (Self : Bare_If_Stmt
         ; If_Stmt_F_Cond_Expr : Bare_Expr
         ; If_Stmt_F_Then_Stmts : Bare_Stmt_List
         ; If_Stmt_F_Alternatives : Bare_Elsif_Stmt_Part_List
         ; If_Stmt_F_Else_Stmts : Bare_Stmt_List
        ) is
      begin
            Self.If_Stmt_F_Cond_Expr := If_Stmt_F_Cond_Expr;
            Self.If_Stmt_F_Then_Stmts := If_Stmt_F_Then_Stmts;
            Self.If_Stmt_F_Alternatives := If_Stmt_F_Alternatives;
            Self.If_Stmt_F_Else_Stmts := If_Stmt_F_Else_Stmts;
      end Initialize_Fields_For_If_Stmt;
   function If_Stmt_F_Cond_Expr
     (Node : Bare_If_Stmt) return Bare_Expr
   is
   begin
         return Node.If_Stmt_F_Cond_Expr;
   end;
   function If_Stmt_F_Then_Stmts
     (Node : Bare_If_Stmt) return Bare_Stmt_List
   is
   begin
         return Node.If_Stmt_F_Then_Stmts;
   end;
   function If_Stmt_F_Alternatives
     (Node : Bare_If_Stmt) return Bare_Elsif_Stmt_Part_List
   is
   begin
         return Node.If_Stmt_F_Alternatives;
   end;
   function If_Stmt_F_Else_Stmts
     (Node : Bare_If_Stmt) return Bare_Stmt_List
   is
   begin
         return Node.If_Stmt_F_Else_Stmts;
   end;
--# property-start IfStmt.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19343
pragma Warnings (Off, "is not referenced");
 function If_Stmt_P_Xref_Equation
  (Node : Bare_If_Stmt
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_If_Stmt := Bare_If_Stmt (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_If_Stmt :=
     Internal_Entity_If_Stmt'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1352 with Inline_Always;
      Fld : Bare_Expr;
Fld_1 : Logic_Var;
Fld_2 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_3 : Internal_Entity_Expr;
Fld_4 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_5 : Bare_Expr;
Fld_6 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1352 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_4);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1352;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10808 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10807 '<And at ast.py:19345>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19345
--# expr-start 10804 '<And at ast.py:19345>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19345
--# expr-start 10801 '<Bind at ast.py:19345>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19345
--# expr-start 10799 '<FieldAccess for expected_type_var at ast.py:19345>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19345
--# expr-start 10798 '<FieldAccess for cond_expr at ast.py:19345>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19345
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.If_Stmt_F_Cond_Expr;
--# expr-done 10798
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld);
--# expr-done 10799
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 10800 '<FieldAccess for bool_type at ast.py:19345>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19345
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_2 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 10800
      Cast_Result := Create_Internal_Entity
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Bind_Result := Solver.Create_Assign (Fld_1, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19345") else null)); 
--# expr-done 10801
--# expr-start 10803 '<FieldAccess for sub_equation at ast.py:19346>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19346
--# expr-start 10802 '<FieldAccess for cond_expr at ast.py:19346>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19346
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Expr (Node => Ent.Node.If_Stmt_F_Cond_Expr, Info => Ent.Info);
--# expr-done 10802
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_4 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_3.Node, Env => Env, Origin => Origin, E_Info => Fld_3.Info);
--# end
--# expr-done 10803
And_Pred := Create_And (Bind_Result, Fld_4, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19345") else null)); 
--# expr-done 10804
--# expr-start 10806 '<FieldAccess for matches_expected_formal_prim_type at ast.py:19347>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19347
--# expr-start 10805 '<FieldAccess for cond_expr at ast.py:19347>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19347
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Self.If_Stmt_F_Cond_Expr;
--# expr-done 10805
      if Fld_5 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_prim_type
Fld_6 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Prim_Type (Node => Fld_5, Origin => Origin);
--# end
--# expr-done 10806
And_Pred_1 := Create_And (And_Pred, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19345") else null)); 
--# expr-done 10807
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 10808
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1352;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1352;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end If_Stmt_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Named_Stmt
   --
      procedure Named_Stmt_Pre_Env_Actions
        (Self            : Bare_Named_Stmt;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Resolver : constant Entity_Resolver :=
            null;
         Mapping : Internal_Env_Assoc :=
            Libadalang.Implementation.Env_Mappings_153 (Node => 
Self);
      begin
         Add_To_Env
           (Self,
            State,
            Mapping.Key,
            Mapping.Value,
            Mapping.Metadata,
            Resolver,
            Mapping.Dest_Env,
            DSL_Location => "ast.py:19432");
         Dec_Ref (Mapping.Dest_Env);
      end;
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_154 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Named_Stmt
        (Self : Bare_Named_Stmt
         ; Named_Stmt_F_Decl : Bare_Named_Stmt_Decl
         ; Named_Stmt_F_Stmt : Bare_Composite_Stmt
        ) is
      begin
            Self.Named_Stmt_F_Decl := Named_Stmt_F_Decl;
            Self.Named_Stmt_F_Stmt := Named_Stmt_F_Stmt;
      end Initialize_Fields_For_Named_Stmt;
   function Named_Stmt_F_Decl
     (Node : Bare_Named_Stmt) return Bare_Named_Stmt_Decl
   is
   begin
         return Node.Named_Stmt_F_Decl;
   end;
   function Named_Stmt_F_Stmt
     (Node : Bare_Named_Stmt) return Bare_Composite_Stmt
   is
   begin
         return Node.Named_Stmt_F_Stmt;
   end;
--# property-start NamedStmt.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19436
pragma Warnings (Off, "is not referenced");
 function Named_Stmt_P_Xref_Equation
  (Node : Bare_Named_Stmt
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Named_Stmt := Bare_Named_Stmt (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1360 with Inline_Always;
      True_Rel : Logic_Equation;
            procedure Finalizer_Scope_1360 is
            begin
                     Dec_Ref (True_Rel);
            end Finalizer_Scope_1360;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19436") else null)); 
      Property_Result := True_Rel;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1360;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1360;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Named_Stmt_P_Xref_Equation;
--# end
--# property-start NamedStmt._env_mappings_153 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19432
pragma Warnings (Off, "is not referenced");
 function Env_Mappings_153
  (Node : Bare_Named_Stmt
  )
   return Internal_Env_Assoc
is
   Self : Bare_Named_Stmt := Bare_Named_Stmt (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc;
            procedure Finalizer_Scope_1570 with Inline_Always;
      Fld : Bare_Named_Stmt_Decl;
Fld_1 : Symbol_Type;
Fld_2 : Bare_Named_Stmt_Decl;
New_Struct : Internal_Designated_Env := No_Designated_Env;
Cast_Result : Bare_Ada_Node;
New_Struct_1 : Internal_Env_Assoc := No_Env_Assoc;
            procedure Finalizer_Scope_1570 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
            end Finalizer_Scope_1570;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10813 '<New EnvAssoc at ast.py:19432>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19432
--# expr-start 10809 '<New DesignatedEnv at ast.py:19432>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19432
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 10809
--# expr-start 10811 '<FieldAccess for name_symbol at ast.py:19432>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19432
--# expr-start 10810 '<FieldAccess for decl at ast.py:19432>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19432
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Named_Stmt_F_Decl;
--# expr-done 10810
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.name_symbol
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Name_Symbol (Node => Fld);
--# end
--# expr-done 10811
--# expr-start 10812 '<FieldAccess for decl at ast.py:19432>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19432
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Named_Stmt_F_Decl;
--# expr-done 10812
      Cast_Result := Fld_2;
Inc_Ref (New_Struct);
New_Struct_1 := (Dest_Env => New_Struct, Key => Fld_1, Metadata => No_Metadata, Value => Cast_Result); 
--# expr-done 10813
      Property_Result := New_Struct_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1570;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1570;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Mappings_153;
--# end
--# property-start NamedStmt._env_trans_parent_154 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19431
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_154
  (Node : Bare_Named_Stmt
  )
   return Boolean
is
   Self : Bare_Named_Stmt := Bare_Named_Stmt (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_154;
--# end
   --
   --  Primitives for Bare_Select_Stmt
   --
      procedure Initialize_Fields_For_Select_Stmt
        (Self : Bare_Select_Stmt
         ; Select_Stmt_F_Guards : Bare_Select_When_Part_List
         ; Select_Stmt_F_Else_Stmts : Bare_Stmt_List
         ; Select_Stmt_F_Abort_Stmts : Bare_Stmt_List
        ) is
      begin
            Self.Select_Stmt_F_Guards := Select_Stmt_F_Guards;
            Self.Select_Stmt_F_Else_Stmts := Select_Stmt_F_Else_Stmts;
            Self.Select_Stmt_F_Abort_Stmts := Select_Stmt_F_Abort_Stmts;
      end Initialize_Fields_For_Select_Stmt;
   function Select_Stmt_F_Guards
     (Node : Bare_Select_Stmt) return Bare_Select_When_Part_List
   is
   begin
         return Node.Select_Stmt_F_Guards;
   end;
   function Select_Stmt_F_Else_Stmts
     (Node : Bare_Select_Stmt) return Bare_Stmt_List
   is
   begin
         return Node.Select_Stmt_F_Else_Stmts;
   end;
   function Select_Stmt_F_Abort_Stmts
     (Node : Bare_Select_Stmt) return Bare_Stmt_List
   is
   begin
         return Node.Select_Stmt_F_Abort_Stmts;
   end;
--# property-start SelectStmt.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19660
pragma Warnings (Off, "is not referenced");
 function Select_Stmt_P_Xref_Equation
  (Node : Bare_Select_Stmt
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Select_Stmt := Bare_Select_Stmt (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Select_Stmt :=
     Internal_Entity_Select_Stmt'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1372 with Inline_Always;
            procedure Finalizer_Scope_3066 with Inline_Always;
      Fld : Internal_Entity_Select_When_Part_List;
Fld_1 : Bare_Select_When_Part_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity_Select_When_Part;
Is_Null : Boolean;
If_Result : Internal_Entity_Select_When_Part;
Fld_3 : Logic_Equation;
Item_324 : Internal_Entity_Select_When_Part;
Bare_Item_324 : Bare_Select_When_Part;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1372 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1372;
            procedure Finalizer_Scope_3066 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_3066;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10818 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10817 '<LogicAll at ast.py:19661>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19661
--# expr-start 10816 '<Map at ast.py:19661>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19661
   --# expr-start 10814 '<FieldAccess for guards at ast.py:19661>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19661
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Select_When_Part_List (Node => Ent.Node.Select_Stmt_F_Guards, Info => Ent.Info);
--# expr-done 10814
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Select_When_Part_List := Fld_1;
      begin
         for Untyped_Item_324 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_324 := Untyped_Item_324; 
Is_Null := Bare_Item_324 = null; 
if Is_Null then
   If_Result := No_Entity_Select_When_Part;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_324); 
   If_Result := As_Entity;
end if;
                  Item_324 := If_Result; 
   --# scope-start
               --# bind wp Item_324
      --# expr-start 10815 '<FieldAccess for sub_equation at ast.py:19661>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19661
      if Item_324.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Item_324.Node, Env => Env, Origin => Origin, E_Info => Item_324.Info);
--# end
--# expr-done 10815
         declare
            Item_To_Append : constant Logic_Equation := Fld_3;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_3066;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 10816
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19661") else null)); 
--# expr-done 10817
Let_Result := Logic_Boolean_Op; Inc_Ref (Let_Result);
--# end
--# expr-done 10818
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1372;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1372;
               Finalizer_Scope_3066;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Select_Stmt_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Error_Stmt
   --
   --
   --  Primitives for Bare_Simple_Stmt
   --
   --
   --  Primitives for Bare_Abort_Stmt
   --
      procedure Initialize_Fields_For_Abort_Stmt
        (Self : Bare_Abort_Stmt
         ; Abort_Stmt_F_Names : Bare_Name_List
        ) is
      begin
            Self.Abort_Stmt_F_Names := Abort_Stmt_F_Names;
      end Initialize_Fields_For_Abort_Stmt;
   function Abort_Stmt_F_Names
     (Node : Bare_Abort_Stmt) return Bare_Name_List
   is
   begin
         return Node.Abort_Stmt_F_Names;
   end;
--# property-start AbortStmt.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19279
pragma Warnings (Off, "is not referenced");
 function Abort_Stmt_P_Xref_Equation
  (Node : Bare_Abort_Stmt
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Abort_Stmt := Bare_Abort_Stmt (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Abort_Stmt :=
     Internal_Entity_Abort_Stmt'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1349 with Inline_Always;
            procedure Finalizer_Scope_3067 with Inline_Always;
      Fld : Internal_Entity_Name_List;
Fld_1 : Bare_Name_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld_3 : Logic_Equation;
Fld_4 : Logic_Var;
Pred : Logic_Equation;
And_Pred : Logic_Equation;
Item_325 : Internal_Entity_Name;
Bare_Item_325 : Bare_Name;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1349 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1349;
            procedure Finalizer_Scope_3067 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Pred);
                     Dec_Ref (And_Pred);
            end Finalizer_Scope_3067;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10826 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10825 '<LogicAll at ast.py:19280>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19280
--# expr-start 10824 '<Map at ast.py:19280>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19280
   --# expr-start 10819 '<FieldAccess for names at ast.py:19280>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19280
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name_List (Node => Ent.Node.Abort_Stmt_F_Names, Info => Ent.Info);
--# expr-done 10819
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Name_List := Fld_1;
      begin
         for Untyped_Item_325 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_325 := Untyped_Item_325; 
Is_Null := Bare_Item_325 = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_325); 
   If_Result := As_Entity;
end if;
                  Item_325 := If_Result; 
   --# scope-start
               --# bind name Item_325
      --# expr-start 10823 '<And at ast.py:19282>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19282
--# expr-start 10820 '<FieldAccess for sub_equation at ast.py:19282>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19282
      if Item_325.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Item_325.Node, Env => Env, Origin => Origin, E_Info => Item_325.Info);
--# end
--# expr-done 10820
--# expr-start 10822 '<Predicate on BaseTypeDecl.is_task_type at ast.py:19282>' Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19282
--# expr-start 10821 '<FieldAccess for type_var at ast.py:19283>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19283
      if Item_325.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Item_325.Node);
--# expr-done 10821
Fld_4.Value := No_Entity;
Entity_Vars.Reset (Fld_4);
Pred := Solver.Create_Predicate (Fld_4, Create_Base_Type_Decl_P_Is_Task_Type_0_Predicate); 
--# expr-done 10822
And_Pred := Create_And (Fld_3, Pred, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19282") else null)); 
--# expr-done 10823
         declare
            Item_To_Append : constant Logic_Equation := And_Pred;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_3067;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 10824
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19280") else null)); 
--# expr-done 10825
Let_Result := Logic_Boolean_Op; Inc_Ref (Let_Result);
--# end
--# expr-done 10826
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1349;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1349;
               Finalizer_Scope_3067;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Abort_Stmt_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Assign_Stmt
   --
      procedure Initialize_Fields_For_Assign_Stmt
        (Self : Bare_Assign_Stmt
         ; Assign_Stmt_F_Dest : Bare_Name
         ; Assign_Stmt_F_Expr : Bare_Expr
        ) is
      begin
            Self.Assign_Stmt_F_Dest := Assign_Stmt_F_Dest;
            Self.Assign_Stmt_F_Expr := Assign_Stmt_F_Expr;
      end Initialize_Fields_For_Assign_Stmt;
   function Assign_Stmt_F_Dest
     (Node : Bare_Assign_Stmt) return Bare_Name
   is
   begin
         return Node.Assign_Stmt_F_Dest;
   end;
   function Assign_Stmt_F_Expr
     (Node : Bare_Assign_Stmt) return Bare_Expr
   is
   begin
         return Node.Assign_Stmt_F_Expr;
   end;
--# property-start AssignStmt.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19129
pragma Warnings (Off, "is not referenced");
 function Assign_Stmt_P_Xref_Equation
  (Node : Bare_Assign_Stmt
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Assign_Stmt := Bare_Assign_Stmt (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Assign_Stmt :=
     Internal_Entity_Assign_Stmt'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1342 with Inline_Always;
      Fld : Internal_Entity_Name;
Fld_1 : Logic_Equation;
Fld_2 : Bare_Name;
Fld_3 : Logic_Var;
Fld_4 : Bare_Expr;
Fld_5 : Logic_Var;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Fld_6 : Internal_Entity_Expr;
Fld_7 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_8 : Bare_Expr;
Fld_9 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1342 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_7);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_9);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1342;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10841 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10840 '<And at ast.py:19131>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19131
--# expr-start 10837 '<And at ast.py:19131>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19131
--# expr-start 10834 '<And at ast.py:19131>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19131
--# expr-start 10828 '<FieldAccess for sub_equation at ast.py:19131>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19131
--# expr-start 10827 '<FieldAccess for dest at ast.py:19131>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19131
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Assign_Stmt_F_Dest, Info => Ent.Info);
--# expr-done 10827
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 10828
--# expr-start 10833 '<Bind at ast.py:19132>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19132
--# expr-start 10830 '<FieldAccess for type_var at ast.py:19132>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19132
--# expr-start 10829 '<FieldAccess for dest at ast.py:19132>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19132
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Assign_Stmt_F_Dest;
--# expr-done 10829
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_2);
--# expr-done 10830
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
--# expr-start 10832 '<FieldAccess for expected_type_var at ast.py:19132>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19132
--# expr-start 10831 '<FieldAccess for expr at ast.py:19132>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19132
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Self.Assign_Stmt_F_Expr;
--# expr-done 10831
      if Fld_4 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_4);
--# expr-done 10832
Fld_5.Value := No_Entity;
Entity_Vars.Reset (Fld_5);
Bind_Result := Solver.Create_Propagate (Fld_3, Fld_5, Logic_Functor_577'(Ref_Count => 1, Cache_Set => False, Cache_Key => <>, Cache_Value => <>, Origin => Origin), Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19132") else null)); 
--# expr-done 10833
And_Pred := Create_And (Fld_1, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19131") else null)); 
--# expr-done 10834
--# expr-start 10836 '<FieldAccess for sub_equation at ast.py:19134>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19134
--# expr-start 10835 '<FieldAccess for expr at ast.py:19134>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19134
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Create_Internal_Entity_Expr (Node => Ent.Node.Assign_Stmt_F_Expr, Info => Ent.Info);
--# expr-done 10835
      if Fld_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_7 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_6.Node, Env => Env, Origin => Origin, E_Info => Fld_6.Info);
--# end
--# expr-done 10836
And_Pred_1 := Create_And (And_Pred, Fld_7, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19131") else null)); 
--# expr-done 10837
--# expr-start 10839 '<FieldAccess for matches_expected_assign_type at ast.py:19135>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19135
--# expr-start 10838 '<FieldAccess for expr at ast.py:19135>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19135
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Self.Assign_Stmt_F_Expr;
--# expr-done 10838
      if Fld_8 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_assign_type
Fld_9 := Libadalang.Implementation.Expr_P_Matches_Expected_Assign_Type (Node => Fld_8, Origin => Origin);
--# end
--# expr-done 10839
And_Pred_2 := Create_And (And_Pred_1, Fld_9, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19131") else null)); 
--# expr-done 10840
Let_Result := And_Pred_2; Inc_Ref (Let_Result);
--# end
--# expr-done 10841
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1342;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1342;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Assign_Stmt_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Call_Stmt
   --
      procedure Initialize_Fields_For_Call_Stmt
        (Self : Bare_Call_Stmt
         ; Call_Stmt_F_Call : Bare_Name
        ) is
      begin
            Self.Call_Stmt_F_Call := Call_Stmt_F_Call;
      end Initialize_Fields_For_Call_Stmt;
   function Call_Stmt_F_Call
     (Node : Bare_Call_Stmt) return Bare_Name
   is
   begin
         return Node.Call_Stmt_F_Call;
   end;
--# property-start CallStmt.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19101
pragma Warnings (Off, "is not referenced");
 function Call_Stmt_P_Xref_Equation
  (Node : Bare_Call_Stmt
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Call_Stmt := Bare_Call_Stmt (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Call_Stmt :=
     Internal_Entity_Call_Stmt'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1340 with Inline_Always;
      Fld : Internal_Entity_Name;
Fld_1 : Logic_Equation;
Fld_2 : Bare_Name;
Fld_3 : Logic_Var;
Bind_Result : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1340 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1340;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10848 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10847 '<And at ast.py:19103>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19103
--# expr-start 10843 '<FieldAccess for sub_equation at ast.py:19103>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19103
--# expr-start 10842 '<FieldAccess for call at ast.py:19103>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19103
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Call_Stmt_F_Call, Info => Ent.Info);
--# expr-done 10842
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 10843
--# expr-start 10846 '<Bind at ast.py:19106>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19106
--# expr-start 10845 '<FieldAccess for type_var at ast.py:19106>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19106
--# expr-start 10844 '<FieldAccess for call at ast.py:19106>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19106
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Call_Stmt_F_Call;
--# expr-done 10844
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Type_Var (Node => Fld_2);
--# expr-done 10845
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
Bind_Result := Solver.Create_Assign (Fld_3, No_Entity, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19106") else null)); 
--# expr-done 10846
And_Pred := Create_And (Fld_1, Bind_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19103") else null)); 
--# expr-done 10847
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 10848
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1340;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1340;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Call_Stmt_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Delay_Stmt
   --
      procedure Initialize_Fields_For_Delay_Stmt
        (Self : Bare_Delay_Stmt
         ; Delay_Stmt_F_Has_Until : Bare_Until_Node
         ; Delay_Stmt_F_Expr : Bare_Expr
        ) is
      begin
            Self.Delay_Stmt_F_Has_Until := Delay_Stmt_F_Has_Until;
            Self.Delay_Stmt_F_Expr := Delay_Stmt_F_Expr;
      end Initialize_Fields_For_Delay_Stmt;
   function Delay_Stmt_F_Has_Until
     (Node : Bare_Delay_Stmt) return Bare_Until_Node
   is
   begin
         return Node.Delay_Stmt_F_Has_Until;
   end;
   function Delay_Stmt_F_Expr
     (Node : Bare_Delay_Stmt) return Bare_Expr
   is
   begin
         return Node.Delay_Stmt_F_Expr;
   end;
--# property-start DelayStmt.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19296
pragma Warnings (Off, "is not referenced");
 function Delay_Stmt_P_Xref_Equation
  (Node : Bare_Delay_Stmt
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Delay_Stmt := Bare_Delay_Stmt (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Delay_Stmt :=
     Internal_Entity_Delay_Stmt'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1350 with Inline_Always;
      Fld : Internal_Entity_Expr;
Fld_1 : Logic_Equation;
True_Rel : Logic_Equation;
Fld_2 : Bare_Expr;
Fld_3 : Logic_Var;
Sym : Symbol_Type;
Fld_4 : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_5 : Internal_Entity_Expr;
Fld_6 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_7 : Bare_Until_Node;
Fld_8 : Boolean;
If_Result : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1350 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred);
                     Dec_Ref (If_Result);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1350;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10863 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10862 '<And at ast.py:19297>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19297
--# expr-start 10850 '<FieldAccess for sub_equation at ast.py:19297>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19297
--# expr-start 10849 '<FieldAccess for expr at ast.py:19297>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19297
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Delay_Stmt_F_Expr, Info => Ent.Info);
--# expr-done 10849
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 10850
--# expr-start 10861 '<If at ast.py:19297>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19297
--# expr-start 10852 '<FieldAccess for as_bool at ast.py:19298>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19298
--# expr-start 10851 '<FieldAccess for has_until at ast.py:19298>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19298
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Self.Delay_Stmt_F_Has_Until;
--# expr-done 10851
      if Fld_7 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Until.as_bool'
Fld_8 := Libadalang.Implementation.Dispatcher_Until_Node_P_As_Bool (Node => Fld_7);
--# end
--# expr-done 10852
if Fld_8 then
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19299") else null)); 
   If_Result := True_Rel;
else
   --# expr-start 10860 '<And at ast.py:19300>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19300
--# expr-start 10857 '<Bind at ast.py:19300>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19300
--# expr-start 10854 '<FieldAccess for expected_type_var at ast.py:19300>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19300
--# expr-start 10853 '<FieldAccess for expr at ast.py:19300>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19300
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Delay_Stmt_F_Expr;
--# expr-done 10853
      if Fld_2 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_2);
--# expr-done 10854
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
--# expr-start 10856 '<FieldAccess for std_entity at ast.py:19300>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19300
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10855 '<Symbol Duration at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Duration); 
--# expr-done 10855
--# property-call-start AdaNode.std_entity
Fld_4 := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym);
--# end
--# expr-done 10856
Bind_Result := Solver.Create_Assign (Fld_3, Fld_4, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19300") else null)); 
--# expr-done 10857
--# expr-start 10859 '<FieldAccess for matches_expected_type at ast.py:19301>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19301
--# expr-start 10858 '<FieldAccess for expr at ast.py:19301>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19301
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Expr (Node => Ent.Node.Delay_Stmt_F_Expr, Info => Ent.Info);
--# expr-done 10858
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_6 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_5.Node, Origin => Origin);
--# end
--# expr-done 10859
And_Pred := Create_And (Bind_Result, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19300") else null)); 
--# expr-done 10860
   If_Result := And_Pred;
end if;
      Inc_Ref (If_Result);
--# expr-done 10861
And_Pred_1 := Create_And (Fld_1, If_Result, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19297") else null)); 
--# expr-done 10862
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 10863
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1350;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1350;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Delay_Stmt_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Exit_Stmt
   --
      procedure Initialize_Fields_For_Exit_Stmt
        (Self : Bare_Exit_Stmt
         ; Exit_Stmt_F_Loop_Name : Bare_Name
         ; Exit_Stmt_F_Cond_Expr : Bare_Expr
        ) is
      begin
            Self.Exit_Stmt_F_Loop_Name := Exit_Stmt_F_Loop_Name;
            Self.Exit_Stmt_F_Cond_Expr := Exit_Stmt_F_Cond_Expr;
      end Initialize_Fields_For_Exit_Stmt;
   function Exit_Stmt_F_Loop_Name
     (Node : Bare_Exit_Stmt) return Bare_Name
   is
   begin
         return Node.Exit_Stmt_F_Loop_Name;
   end;
   function Exit_Stmt_F_Cond_Expr
     (Node : Bare_Exit_Stmt) return Bare_Expr
   is
   begin
         return Node.Exit_Stmt_F_Cond_Expr;
   end;
--# property-start ExitStmt.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19160
pragma Warnings (Off, "is not referenced");
 function Exit_Stmt_P_Xref_Equation
  (Node : Bare_Exit_Stmt
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Exit_Stmt := Bare_Exit_Stmt (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Exit_Stmt :=
     Internal_Entity_Exit_Stmt'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1344 with Inline_Always;
            procedure Finalizer_Scope_3068 with Inline_Always;
            procedure Finalizer_Scope_3069 with Inline_Always;
      Var_Expr : Internal_Entity_Expr;
Var_Expr_1 : Internal_Entity_Name;
Fld : Internal_Entity_Expr;
Fld_1 : Logic_Var;
Fld_2 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_3 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_4 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
Fld_5 : Internal_Entity_Name;
Fld_6 : Logic_Equation;
Scope_Result_1 : Logic_Equation;
True_Rel_1 : Logic_Equation;
Result_Var_1 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1344 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (True_Rel_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1344;
            procedure Finalizer_Scope_3068 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_3);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_4);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_3068;
            procedure Finalizer_Scope_3069 is
            begin
                     Dec_Ref (Fld_6);
            end Finalizer_Scope_3069;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10875 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10874 '<And at ast.py:19161>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19161
--# expr-start 10864 '<FieldAccess for cond_expr at ast.py:19162>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19162
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Exit_Stmt_F_Cond_Expr, Info => Ent.Info);
--# expr-done 10864
Var_Expr := Fld;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind cond Var_Expr
--# expr-start 10871 '<And at ast.py:19164>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19164
--# expr-start 10869 '<And at ast.py:19164>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19164
--# expr-start 10867 '<Bind at ast.py:19164>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19164
--# expr-start 10865 '<FieldAccess for expected_type_var at ast.py:19164>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19164
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Var_Expr.Node);
--# expr-done 10865
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 10866 '<FieldAccess for bool_type at ast.py:19164>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19164
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_2 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 10866
      Cast_Result := Create_Internal_Entity
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Bind_Result := Solver.Create_Assign (Fld_1, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19164") else null)); 
--# expr-done 10867
--# expr-start 10868 '<FieldAccess for sub_equation at ast.py:19165>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19165
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 10868
And_Pred := Create_And (Bind_Result, Fld_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19164") else null)); 
--# expr-done 10869
--# expr-start 10870 '<FieldAccess for matches_expected_formal_prim_type at ast.py:19166>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19166
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_formal_prim_type
Fld_4 := Libadalang.Implementation.Expr_P_Matches_Expected_Formal_Prim_Type (Node => Var_Expr.Node, Origin => Origin);
--# end
--# expr-done 10870
And_Pred_1 := Create_And (And_Pred, Fld_4, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19164") else null)); 
--# expr-done 10871
Scope_Result := And_Pred_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_3068;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19167") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
--# expr-start 10872 '<FieldAccess for loop_name at ast.py:19170>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19170
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Name (Node => Ent.Node.Exit_Stmt_F_Loop_Name, Info => Ent.Info);
--# expr-done 10872
Var_Expr_1 := Fld_5;
if Var_Expr_1 /= No_Entity_Name then
   --# scope-start
   --# bind ln Var_Expr_1
--# expr-start 10873 '<FieldAccess for xref_no_overloading at ast.py:19171>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19171
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_6 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Var_Expr_1.Node, Env => Env, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 10873
Scope_Result_1 := Fld_6;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_3069;
   Result_Var_1 := Scope_Result_1;
else
   True_Rel_1 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19172") else null)); 
   Result_Var_1 := True_Rel_1;
end if;
      Inc_Ref (Result_Var_1);
And_Pred_2 := Create_And (Result_Var, Result_Var_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19161") else null)); 
--# expr-done 10874
Let_Result := And_Pred_2; Inc_Ref (Let_Result);
--# end
--# expr-done 10875
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1344;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1344;
               Finalizer_Scope_3068;
               Finalizer_Scope_3069;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Exit_Stmt_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Goto_Stmt
   --
      procedure Initialize_Fields_For_Goto_Stmt
        (Self : Bare_Goto_Stmt
         ; Goto_Stmt_F_Label_Name : Bare_Name
        ) is
      begin
            Self.Goto_Stmt_F_Label_Name := Goto_Stmt_F_Label_Name;
      end Initialize_Fields_For_Goto_Stmt;
   function Goto_Stmt_F_Label_Name
     (Node : Bare_Goto_Stmt) return Bare_Name
   is
   begin
         return Node.Goto_Stmt_F_Label_Name;
   end;
--# property-start GotoStmt.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19147
pragma Warnings (Off, "is not referenced");
 function Goto_Stmt_P_Xref_Equation
  (Node : Bare_Goto_Stmt
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Goto_Stmt := Bare_Goto_Stmt (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Goto_Stmt :=
     Internal_Entity_Goto_Stmt'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1343 with Inline_Always;
      Fld : Internal_Entity_Name;
Fld_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1343 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1343;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10878 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10877 '<FieldAccess for xref_no_overloading at ast.py:19148>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19148
--# expr-start 10876 '<FieldAccess for label_name at ast.py:19148>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19148
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Goto_Stmt_F_Label_Name, Info => Ent.Info);
--# expr-done 10876
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_1 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Fld.Node, Sequential => False, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 10877
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 10878
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1343;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1343;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Goto_Stmt_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Label
   --
      procedure Initialize_Fields_For_Label
        (Self : Bare_Label
         ; Label_F_Decl : Bare_Label_Decl
        ) is
      begin
            Self.Label_F_Decl := Label_F_Decl;
      end Initialize_Fields_For_Label;
   function Label_F_Decl
     (Node : Bare_Label) return Bare_Label_Decl
   is
   begin
         return Node.Label_F_Decl;
   end;
--# property-start Label.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19391
pragma Warnings (Off, "is not referenced");
 function Label_P_Xref_Equation
  (Node : Bare_Label
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Label := Bare_Label (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1356 with Inline_Always;
      True_Rel : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1356 is
            begin
                     Dec_Ref (True_Rel);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1356;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 10879 '<Block at ???>' Let_Result None
--# scope-start
True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19392") else null)); 
Let_Result := True_Rel; Inc_Ref (Let_Result);
--# end
--# expr-done 10879
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1356;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1356;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Label_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Null_Stmt
   --
--# property-start NullStmt.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19116
pragma Warnings (Off, "is not referenced");
 function Null_Stmt_P_Xref_Equation
  (Node : Bare_Null_Stmt
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Null_Stmt := Bare_Null_Stmt (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1341 with Inline_Always;
      True_Rel : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1341 is
            begin
                     Dec_Ref (True_Rel);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1341;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 10880 '<Block at ???>' Let_Result None
--# scope-start
True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19117") else null)); 
Let_Result := True_Rel; Inc_Ref (Let_Result);
--# end
--# expr-done 10880
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1341;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1341;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Null_Stmt_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Raise_Stmt
   --
      procedure Initialize_Fields_For_Raise_Stmt
        (Self : Bare_Raise_Stmt
         ; Raise_Stmt_F_Exception_Name : Bare_Name
         ; Raise_Stmt_F_Error_Message : Bare_Expr
        ) is
      begin
            Self.Raise_Stmt_F_Exception_Name := Raise_Stmt_F_Exception_Name;
            Self.Raise_Stmt_F_Error_Message := Raise_Stmt_F_Error_Message;
      end Initialize_Fields_For_Raise_Stmt;
   function Raise_Stmt_F_Exception_Name
     (Node : Bare_Raise_Stmt) return Bare_Name
   is
   begin
         return Node.Raise_Stmt_F_Exception_Name;
   end;
   function Raise_Stmt_F_Error_Message
     (Node : Bare_Raise_Stmt) return Bare_Expr
   is
   begin
         return Node.Raise_Stmt_F_Error_Message;
   end;
--# property-start RaiseStmt.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19314
pragma Warnings (Off, "is not referenced");
 function Raise_Stmt_P_Xref_Equation
  (Node : Bare_Raise_Stmt
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Raise_Stmt := Bare_Raise_Stmt (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Raise_Stmt :=
     Internal_Entity_Raise_Stmt'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1351 with Inline_Always;
            procedure Finalizer_Scope_3070 with Inline_Always;
            procedure Finalizer_Scope_3071 with Inline_Always;
      Var_Expr : Internal_Entity_Name;
Var_Expr_1 : Internal_Entity_Expr;
Fld : Internal_Entity_Name;
Fld_1 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
Fld_2 : Internal_Entity_Expr;
Fld_3 : Logic_Var;
Sym : Symbol_Type;
Fld_4 : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_5 : Logic_Equation;
And_Pred : Logic_Equation;
Scope_Result_1 : Logic_Equation;
True_Rel_1 : Logic_Equation;
Result_Var_1 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1351 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (True_Rel_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1351;
            procedure Finalizer_Scope_3070 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_3070;
            procedure Finalizer_Scope_3071 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_5);
                     Dec_Ref (And_Pred);
            end Finalizer_Scope_3071;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10891 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10890 '<And at ast.py:19315>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19315
--# expr-start 10881 '<FieldAccess for exception_name at ast.py:19316>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19316
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Raise_Stmt_F_Exception_Name, Info => Ent.Info);
--# expr-done 10881
Var_Expr := Fld;
if Var_Expr /= No_Entity_Name then
   --# scope-start
   --# bind en Var_Expr
--# expr-start 10882 '<FieldAccess for sub_equation at ast.py:19317>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19317
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 10882
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_3070;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19318") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
--# expr-start 10883 '<FieldAccess for error_message at ast.py:19320>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19320
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Expr (Node => Ent.Node.Raise_Stmt_F_Error_Message, Info => Ent.Info);
--# expr-done 10883
Var_Expr_1 := Fld_2;
if Var_Expr_1 /= No_Entity_Expr then
   --# scope-start
   --# bind er Var_Expr_1
--# expr-start 10889 '<And at ast.py:19321>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19321
--# expr-start 10887 '<Bind at ast.py:19324>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19324
--# expr-start 10884 '<FieldAccess for expected_type_var at ast.py:19324>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19324
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Var_Expr_1.Node);
--# expr-done 10884
Fld_3.Value := No_Entity;
Entity_Vars.Reset (Fld_3);
--# expr-start 10886 '<FieldAccess for std_entity at ast.py:19324>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19324
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10885 '<Symbol String at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_String); 
--# expr-done 10885
--# property-call-start AdaNode.std_entity
Fld_4 := Libadalang.Implementation.Ada_Node_P_Std_Entity (Node => Self, Sym => Sym);
--# end
--# expr-done 10886
Bind_Result := Solver.Create_Assign (Fld_3, Fld_4, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19324") else null)); 
--# expr-done 10887
--# expr-start 10888 '<FieldAccess for sub_equation at ast.py:19325>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19325
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_5 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr_1.Node, Env => Env, Origin => Origin, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 10888
And_Pred := Create_And (Bind_Result, Fld_5, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19321") else null)); 
--# expr-done 10889
Scope_Result_1 := And_Pred;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_3071;
   Result_Var_1 := Scope_Result_1;
else
   True_Rel_1 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19327") else null)); 
   Result_Var_1 := True_Rel_1;
end if;
      Inc_Ref (Result_Var_1);
And_Pred_1 := Create_And (Result_Var, Result_Var_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19315") else null)); 
--# expr-done 10890
Let_Result := And_Pred_1; Inc_Ref (Let_Result);
--# end
--# expr-done 10891
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1351;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1351;
               Finalizer_Scope_3070;
               Finalizer_Scope_3071;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Raise_Stmt_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Requeue_Stmt
   --
      procedure Initialize_Fields_For_Requeue_Stmt
        (Self : Bare_Requeue_Stmt
         ; Requeue_Stmt_F_Call_Name : Bare_Name
         ; Requeue_Stmt_F_Has_Abort : Bare_Abort_Node
        ) is
      begin
            Self.Requeue_Stmt_F_Call_Name := Requeue_Stmt_F_Call_Name;
            Self.Requeue_Stmt_F_Has_Abort := Requeue_Stmt_F_Has_Abort;
      end Initialize_Fields_For_Requeue_Stmt;
   function Requeue_Stmt_F_Call_Name
     (Node : Bare_Requeue_Stmt) return Bare_Name
   is
   begin
         return Node.Requeue_Stmt_F_Call_Name;
   end;
   function Requeue_Stmt_F_Has_Abort
     (Node : Bare_Requeue_Stmt) return Bare_Abort_Node
   is
   begin
         return Node.Requeue_Stmt_F_Has_Abort;
   end;
--# property-start RequeueStmt.innermost_entry_or_accept_stmt_params /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19209
pragma Warnings (Off, "is not referenced");
 function Requeue_Stmt_P_Innermost_Entry_Or_Accept_Stmt_Params
  (Node : Bare_Requeue_Stmt
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Holder
is
   Self : Bare_Requeue_Stmt := Bare_Requeue_Stmt (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Requeue_Stmt :=
     Internal_Entity_Requeue_Stmt'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Formal_Param_Holder;
            procedure Finalizer_Scope_1347 with Inline_Always;
      Match_0 : Internal_Entity_Accept_Stmt_With_Stmts;
Match_1 : Internal_Entity_Entry_Body;
Match_2 : Internal_Entity;
Fld : Internal_Entity_Array_Access;
Is_A : Boolean;
Find_Result : Internal_Entity;
Fld_1 : Internal_Entity_Entry_Completion_Formal_Params;
Fld_2 : Internal_Entity_Entry_Completion_Formal_Params;
Match_Prefix : Internal_Entity;
Cast_Result : Internal_Entity_Accept_Stmt_With_Stmts;
Let_Result : Internal_Entity_Entry_Completion_Formal_Params;
Scope_Result : Internal_Entity_Entry_Completion_Formal_Params;
Cast_Result_1 : Internal_Entity_Entry_Body;
Let_Result_1 : Internal_Entity_Entry_Completion_Formal_Params;
Scope_Result_1 : Internal_Entity_Entry_Completion_Formal_Params;
Cast_Result_2 : Internal_Entity;
Let_Result_2 : Internal_Entity_Entry_Completion_Formal_Params;
Scope_Result_2 : Internal_Entity_Entry_Completion_Formal_Params;
Match_Result : Internal_Entity_Entry_Completion_Formal_Params;
Cast_Result_3 : Internal_Entity_Base_Formal_Param_Holder;
            procedure Finalizer_Scope_1347 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1347;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 10898 '<Cast to BaseFormalParamHolder.entity at ast.py:19210>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19210
--# expr-start 10897 '<Match at ast.py:19210>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19210
--# expr-start 10894 '<Find at ast.py:19210>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19210
   --# expr-start 10892 '<FieldAccess for parents at ast.py:19210>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19210
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parents (Node => Ent.Node, With_Self => True, E_Info => Ent.Info);
--# expr-done 10892
      Find_Result := No_Entity;
      declare
         Collection : constant Internal_Entity_Array_Access := Fld;
      begin
         for Item_326 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_326
      --# expr-start 10893 '<IsA AcceptStmtWithStmts, EntryBody at ast.py:19211>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19211
Is_A := Item_326.Node /= null 
and then Item_326.Node.Kind in Ada_Accept_Stmt_With_Stmts_Range | Ada_Entry_Body_Range; 
--# expr-done 10893
      if Is_A then
         Find_Result := Item_326;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 10894
      if Find_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Find_Result; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Accept_Stmt_With_Stmts =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Accept_Stmt_With_Stmts
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind a Match_0
--# expr-start 10895 '<FieldAccess for params at ast.py:19213>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19213
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Entry_Completion_Formal_Params (Node => Match_0.Node.Accept_Stmt_F_Params, Info => Match_0.Info);
--# expr-done 10895
Let_Result := Fld_1; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Entry_Body =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Entry_Body
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind b Match_1
--# expr-start 10896 '<FieldAccess for params at ast.py:19214>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19214
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Entry_Completion_Formal_Params (Node => Match_1.Node.Entry_Body_F_Params, Info => Match_1.Info);
--# expr-done 10896
Let_Result_1 := Fld_2; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abort_Absent .. Ada_Task_Body_Stub | Ada_Package_Body .. Ada_Accept_Stmt | Ada_For_Loop_Stmt .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind _ Match_2
Let_Result_2 := No_Entity_Entry_Completion_Formal_Params; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 10897
      Cast_Result_3 := Create_Internal_Entity_Base_Formal_Param_Holder
        (Node => Match_Result.Node,
         Info => Match_Result.Info);
--# expr-done 10898
      Property_Result := Cast_Result_3;
   --# end
      Finalizer_Scope_1347;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1347;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Requeue_Stmt_P_Innermost_Entry_Or_Accept_Stmt_Params;
--# end
--# property-start RequeueStmt.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19220
pragma Warnings (Off, "is not referenced");
 function Requeue_Stmt_P_Xref_Equation
  (Node : Bare_Requeue_Stmt
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Requeue_Stmt := Bare_Requeue_Stmt (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Requeue_Stmt :=
     Internal_Entity_Requeue_Stmt'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1348 with Inline_Always;
            procedure Finalizer_Scope_3080 with Inline_Always;
            procedure Finalizer_Scope_3087 with Inline_Always;
            procedure Finalizer_Scope_3088 with Inline_Always;
      Block_Var : Internal_Entity_Call_Expr;
Block_Var_1 : Internal_Entity_Name;
Block_Var_2 : Internal_Entity_Array_Access;
Var_Expr : Internal_Entity_Call_Expr;
Var_Expr_1 : Internal_Entity_Entry_Decl;
Var_Expr_2 : Internal_Entity_Base_Subp_Spec;
Fam_Type : Internal_Entity_Base_Type_Decl;
First_Param : Internal_Entity_Expr;
Var_Expr_3 : Internal_Entity_Subtype_Indication;
Var_Expr_4 : Internal_Entity_Entry_Decl;
Var_Expr_5 : Internal_Entity_Basic_Assoc;
Var_Expr_6 : Internal_Entity_Call_Expr;
Var_Expr_7 : Internal_Entity_Expr;
Var_Expr_8 : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Name;
Cast_Result : Internal_Entity_Call_Expr;
Fld_1 : Internal_Entity_Name;
Scope_Result : Internal_Entity_Name;
Fld_2 : Internal_Entity_Name;
Result_Var : Internal_Entity_Name;
Fld_3 : Internal_Entity_Array_Access;
Cast_Expr : Internal_Entity;
Cast_Result_1 : Internal_Entity_Entry_Decl;
Fld_4 : Internal_Entity_Base_Subp_Spec;
Fld_5 : Integer;
Is_Equal : Boolean;
Fld_6 : Internal_Entity_Type_Expr;
Is_Null : Boolean;
If_Result : Boolean;
Scope_Result_1 : Boolean;
Result_Var_1 : Boolean;
Fld_7 : Internal_Entity_Entry_Spec;
Fld_8 : Internal_Entity_Base_Formal_Param_Holder;
Fld_9 : Boolean;
If_Result_1 : Boolean;
Scope_Result_2 : Boolean;
Result_Var_2 : Boolean;
Map_Result : Internal_Entity_Array_Access;
Fld_10 : Logic_Equation;
Cast_Expr_1 : Internal_Entity;
Cast_Result_2 : Internal_Entity_Entry_Decl;
Fld_11 : Internal_Entity_Entry_Spec;
Fld_12 : Internal_Entity;
Scope_Result_3 : Internal_Entity;
Result_Var_3 : Internal_Entity;
Cast_Result_3 : Internal_Entity_Subtype_Indication;
Fld_13 : Internal_Entity_Base_Type_Decl;
Scope_Result_4 : Internal_Entity_Base_Type_Decl;
Result_Var_4 : Internal_Entity_Base_Type_Decl;
Fld_14 : Internal_Entity_Assoc_List;
Scope_Result_5 : Internal_Entity_Assoc_List;
Result_Var_5 : Internal_Entity_Assoc_List;
Fld_15 : Bare_Assoc_List;
Fld_16 : Internal_Entity_Info;
As_Entity : Internal_Entity_Basic_Assoc;
Is_Null_1 : Boolean;
If_Result_2 : Internal_Entity_Basic_Assoc;
Item_361 : Internal_Entity_Basic_Assoc;
Bare_Item_361 : Bare_Basic_Assoc;
Map_Result_1 : Internal_Entity_Basic_Assoc_Array_Access;
Get_Result : Internal_Entity_Basic_Assoc;
Fld_17 : Internal_Entity_Expr;
Scope_Result_6 : Internal_Entity_Expr;
Result_Var_6 : Internal_Entity_Expr;
Fld_18 : Logic_Equation;
Fld_19 : Logic_Var;
Cast_Expr_2 : Internal_Entity_Base_Type_Decl;
Cast_Result_4 : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_20 : Logic_Equation;
And_Pred : Logic_Equation;
Scope_Result_7 : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var_7 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Scope_Result_8 : Logic_Equation;
True_Rel_1 : Logic_Equation;
Result_Var_8 : Logic_Equation;
Let_Result : Logic_Equation;
Map_Result_2 : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
And_Pred_2 : Logic_Equation;
Let_Result_1 : Logic_Equation;
            procedure Finalizer_Scope_1348 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Block_Var_2);
                     Dec_Ref (Fld_10);
                     Dec_Ref (Map_Result_2);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Let_Result_1);
            end Finalizer_Scope_1348;
            procedure Finalizer_Scope_3080 is
            begin
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Scope_Result_8);
                     Dec_Ref (True_Rel_1);
                     Dec_Ref (Result_Var_8);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_3080;
            procedure Finalizer_Scope_3087 is
            begin
                     Dec_Ref (Fld_18);
                     Dec_Ref (Scope_Result_7);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var_7);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_3087;
            procedure Finalizer_Scope_3088 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_20);
                     Dec_Ref (And_Pred);
            end Finalizer_Scope_3088;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10934 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 10900 '<Cast to CallExpr at ast.py:19221>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19221
--# expr-start 10899 '<FieldAccess for call_name at ast.py:19221>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19221
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Requeue_Stmt_F_Call_Name, Info => Ent.Info);
--# expr-done 10899
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Call_Expr_Range
   then
      Cast_Result := Create_Internal_Entity_Call_Expr
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Call_Expr;
   end if;
--# expr-done 10900
Block_Var := Cast_Result; 
--# bind ce Block_Var
Var_Expr := Block_Var;
if Var_Expr /= No_Entity_Call_Expr then
   --# scope-start
   --# bind ce Var_Expr
--# expr-start 10901 '<FieldAccess for name at ast.py:19222>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19222
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Var_Expr.Node.Call_Expr_F_Name, Info => Var_Expr.Info);
--# expr-done 10901
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   --# expr-start 10902 '<FieldAccess for call_name at ast.py:19223>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19223
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Name (Node => Ent.Node.Requeue_Stmt_F_Call_Name, Info => Ent.Info);
--# expr-done 10902
   Result_Var := Fld_2;
end if;
Block_Var_1 := Result_Var; 
--# bind name Block_Var_1
--# expr-start 10913 '<Filter at ast.py:19225>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19225
   --# expr-start 10903 '<FieldAccess for all_env_elements_internal at ast.py:19225>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19225
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.all_env_elements_internal
Fld_3 := Libadalang.Implementation.Name_P_All_Env_Elements_Internal (Node => Block_Var_1.Node, E_Info => Block_Var_1.Info);
--# end
--# expr-done 10903
      declare
         Map_Result_Vec : Internal_Entity_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Fld_3;
      begin
         for Item_327 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_327
--# expr-start 10904 '<Cast to EntryDecl at ast.py:19228>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19228
Cast_Expr := Item_327; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Entry_Decl_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Entry_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result_1 := No_Entity_Entry_Decl;
   end if;
--# expr-done 10904
Var_Expr_1 := Cast_Result_1;
if Var_Expr_1 /= No_Entity_Entry_Decl then
   --# scope-start
   --# bind e Var_Expr_1
--# expr-start 10905 '<FieldAccess for subp_spec_or_null at ast.py:19230>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19230
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.subp_spec_or_null
Fld_4 := Libadalang.Implementation.Basic_Decl_P_Subp_Spec_Or_Null (Node => Var_Expr_1.Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 10905
Var_Expr_2 := Fld_4;
if Var_Expr_2 /= No_Entity_Base_Subp_Spec then
   --# scope-start
   --# bind ss Var_Expr_2
--# expr-start 10907 '<Eq at ast.py:19231>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19231
--# expr-start 10906 '<FieldAccess for nb_max_params at ast.py:19231>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19231
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseFormalParamHolder.nb_max_params
Fld_5 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Nb_Max_Params (Node => Var_Expr_2.Node);
--# end
--# expr-done 10906
Is_Equal := Fld_5 = 0; 
--# expr-done 10907
if Is_Equal then
   --# expr-start 10909 '<IsNull at ast.py:19232>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19232
--# expr-start 10908 '<FieldAccess for returns at ast.py:19232>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19232
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.returns'
Fld_6 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Returns (Node => Var_Expr_2.Node, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 10908
Is_Null := Fld_6.Node = null; 
--# expr-done 10909
   If_Result := Is_Null;
else
   If_Result := False;
end if;
Scope_Result_1 := If_Result;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := False;
end if;
if Result_Var_1 then
   If_Result_1 := True;
else
   --# expr-start 10912 '<FieldAccess for match_formal_params at ast.py:19235>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19235
--# expr-start 10910 '<FieldAccess for spec at ast.py:19235>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19235
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Entry_Spec (Node => Var_Expr_1.Node.Entry_Decl_F_Spec, Info => Var_Expr_1.Info);
--# expr-done 10910
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10911 '<FieldAccess for innermost_entry_or_accept_stmt_params at ast.py:19236>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19236
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start RequeueStmt.innermost_entry_or_accept_stmt_params
Fld_8 := Libadalang.Implementation.Requeue_Stmt_P_Innermost_Entry_Or_Accept_Stmt_Params (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 10911
--# property-call-start BaseFormalParamHolder.match_formal_params
Fld_9 := Libadalang.Implementation.Base_Formal_Param_Holder_P_Match_Formal_Params (Node => Fld_7.Node, Other => Fld_8, E_Info => Fld_7.Info);
--# end
--# expr-done 10912
   If_Result_1 := Fld_9;
end if;
Scope_Result_2 := If_Result_1;
   --# end
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := False;
end if;
         if Result_Var_2 then
         declare
            Item_To_Append : constant Internal_Entity := Item_327;
         begin
            Internal_Entity_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
         end if;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Entity_Array
           (Items_Count => Natural (Internal_Entity_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 10913
Block_Var_2 := Map_Result; Inc_Ref (Block_Var_2);
--# bind entries Block_Var_2
--# expr-start 10933 '<And at ast.py:19241>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19241
--# expr-start 10914 '<FieldAccess for xref_no_overloading at ast.py:19244>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19244
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_10 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Block_Var_1.Node, Env => Env, Origin => Origin, E_Info => Block_Var_1.Info);
--# end
--# expr-done 10914
--# expr-start 10932 '<LogicAny at ast.py:19247>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19247
--# expr-start 10931 '<Map at ast.py:19247>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19247
      declare
         Map_Result_2_Vec : Logic_Equation_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Block_Var_2;
      begin
         for Item_328 of
               Collection.Items
         loop
   --# scope-start
               --# bind e Item_328
      --# expr-start 10930 '<Let at ast.py:19247>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19247
--# scope-start
--# expr-start 10918 '<Cast to SubtypeIndication at ast.py:19251>' Cast_Result_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19251
--# expr-start 10915 '<Cast to EntryDecl at ast.py:19251>' Cast_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19251
Cast_Expr_1 := Item_328; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Entry_Decl_Range
   then
      Cast_Result_2 := Create_Internal_Entity_Entry_Decl
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_2 := No_Entity_Entry_Decl;
   end if;
--# expr-done 10915
Var_Expr_4 := Cast_Result_2;
if Var_Expr_4 /= No_Entity_Entry_Decl then
   --# scope-start
--# expr-start 10917 '<FieldAccess for family_type at ast.py:19251>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19251
--# expr-start 10916 '<FieldAccess for spec at ast.py:19251>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19251
      if Var_Expr_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Create_Internal_Entity_Entry_Spec (Node => Var_Expr_4.Node.Entry_Decl_F_Spec, Info => Var_Expr_4.Info);
--# expr-done 10916
      if Fld_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Create_Internal_Entity (Node => Fld_11.Node.Entry_Spec_F_Family_Type, Info => Fld_11.Info);
--# expr-done 10917
Scope_Result_3 := Fld_12;
   --# end
   Result_Var_3 := Scope_Result_3;
else
   Result_Var_3 := No_Entity;
end if;
   if Result_Var_3.Node = null
      or else Result_Var_3.Node.Kind in Ada_Subtype_Indication_Range
   then
      Cast_Result_3 := Create_Internal_Entity_Subtype_Indication
        (Node => Result_Var_3.Node,
         Info => Result_Var_3.Info);
   else
         Cast_Result_3 := No_Entity_Subtype_Indication;
   end if;
--# expr-done 10918
Var_Expr_3 := Cast_Result_3;
if Var_Expr_3 /= No_Entity_Subtype_Indication then
   --# scope-start
--# expr-start 10919 '<FieldAccess for designated_type at ast.py:19251>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19251
      if Var_Expr_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_13 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Var_Expr_3.Node, Origin => Origin, E_Info => Var_Expr_3.Info);
--# end
--# expr-done 10919
Scope_Result_4 := Fld_13;
   --# end
   Result_Var_4 := Scope_Result_4;
else
   Result_Var_4 := No_Entity_Base_Type_Decl;
end if;
Fam_Type := Result_Var_4; 
--# bind fam_type Fam_Type
--# expr-start 10922 '<at at ast.py:19256>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19256
--# expr-start 10921 '<TakeWhile at ???>' Map_Result_1 None
Var_Expr_6 := Block_Var;
if Var_Expr_6 /= No_Entity_Call_Expr then
   --# scope-start
--# expr-start 10920 '<FieldAccess for params at ast.py:19256>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19256
      if Var_Expr_6.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CallExpr.params
Fld_14 := Libadalang.Implementation.Call_Expr_P_Params (Node => Var_Expr_6.Node, E_Info => Var_Expr_6.Info);
--# end
--# expr-done 10920
Scope_Result_5 := Fld_14;
   --# end
   Result_Var_5 := Scope_Result_5;
else
   Result_Var_5 := No_Entity_Assoc_List;
end if;
Fld_15 := Result_Var_5.Node;
      declare
         Map_Result_1_Vec : Internal_Entity_Basic_Assoc_Vectors.Vector;
      begin
      if Fld_15 = null then
      Map_Result_1 := Create_Internal_Entity_Basic_Assoc_Array (0);
      else
      declare
         Collection : constant Bare_Assoc_List := Fld_15;
      begin
         for Untyped_Item_361 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_361 := Untyped_Item_361; 
Is_Null_1 := Bare_Item_361 = null; 
if Is_Null_1 then
   If_Result_2 := No_Entity_Basic_Assoc;
else
Fld_16 := Result_Var_5.Info;
As_Entity := (Info => Fld_16, Node => Bare_Item_361); 
   If_Result_2 := As_Entity;
end if;
                  Item_361 := If_Result_2; 
   --# scope-start
         declare
            Item_To_Append : constant Internal_Entity_Basic_Assoc := Item_361;
         begin
            Internal_Entity_Basic_Assoc_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result_1 := Create_Internal_Entity_Basic_Assoc_Array
           (Items_Count => Natural (Internal_Entity_Basic_Assoc_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Entity_Basic_Assoc_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Entity_Basic_Assoc_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Entity_Basic_Assoc_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 10921
Get_Result := Get (Map_Result_1, 0, True); 
--# expr-done 10922
Var_Expr_5 := Get_Result;
if Var_Expr_5 /= No_Entity_Basic_Assoc then
   --# scope-start
--# expr-start 10923 '<FieldAccess for expr at ast.py:19256>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19256
      if Var_Expr_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_17 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Var_Expr_5.Node, E_Info => Var_Expr_5.Info);
--# end
--# expr-done 10923
Scope_Result_6 := Fld_17;
   --# end
   Result_Var_6 := Scope_Result_6;
else
   Result_Var_6 := No_Entity_Expr;
end if;
First_Param := Result_Var_6; 
--# bind first_param First_Param
Var_Expr_7 := First_Param;
if Var_Expr_7 /= No_Entity_Expr then
   --# scope-start
   --# bind p Var_Expr_7
--# expr-start 10929 '<And at ast.py:19259>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19259
--# expr-start 10924 '<FieldAccess for sub_equation at ast.py:19259>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19259
      if Var_Expr_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_18 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr_7.Node, Env => Env, Origin => Origin, E_Info => Var_Expr_7.Info);
--# end
--# expr-done 10924
Var_Expr_8 := Fam_Type;
if Var_Expr_8 /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind eft Var_Expr_8
--# expr-start 10928 '<And at ast.py:19261>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19261
--# expr-start 10926 '<Bind at ast.py:19261>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19261
--# expr-start 10925 '<FieldAccess for expected_type_var at ast.py:19261>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19261
      if Var_Expr_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_19 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Var_Expr_7.Node);
--# expr-done 10925
Fld_19.Value := No_Entity;
Entity_Vars.Reset (Fld_19);
Cast_Expr_2 := Var_Expr_8; 
      Cast_Result_4 := Create_Internal_Entity
        (Node => Cast_Expr_2.Node,
         Info => Cast_Expr_2.Info);
Bind_Result := Solver.Create_Assign (Fld_19, Cast_Result_4, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19261") else null)); 
--# expr-done 10926
--# expr-start 10927 '<FieldAccess for matches_expected_type at ast.py:19262>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19262
      if Var_Expr_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_20 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Var_Expr_7.Node, Origin => Origin);
--# end
--# expr-done 10927
And_Pred := Create_And (Bind_Result, Fld_20, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19261") else null)); 
--# expr-done 10928
Scope_Result_7 := And_Pred;
   Inc_Ref (Scope_Result_7);
   --# end
      Finalizer_Scope_3088;
   Result_Var_7 := Scope_Result_7;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19263") else null)); 
   Result_Var_7 := True_Rel;
end if;
      Inc_Ref (Result_Var_7);
And_Pred_1 := Create_And (Fld_18, Result_Var_7, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19259") else null)); 
--# expr-done 10929
Scope_Result_8 := And_Pred_1;
   Inc_Ref (Scope_Result_8);
   --# end
      Finalizer_Scope_3087;
   Result_Var_8 := Scope_Result_8;
else
   True_Rel_1 := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19265") else null)); 
   Result_Var_8 := True_Rel_1;
end if;
      Inc_Ref (Result_Var_8);
Let_Result := Result_Var_8; Inc_Ref (Let_Result);
--# end
--# expr-done 10930
         declare
            Item_To_Append : constant Logic_Equation := Let_Result;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_3080;
         end loop;
      end;
         Map_Result_2 := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_2_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 10931
Logic_Boolean_Op := Solver.Create_Any (Relation_Array (Map_Result_2.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19247") else null)); 
--# expr-done 10932
And_Pred_2 := Create_And (Fld_10, Logic_Boolean_Op, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19241") else null)); 
--# expr-done 10933
Let_Result_1 := And_Pred_2; Inc_Ref (Let_Result_1);
--# end
--# expr-done 10934
      Property_Result := Let_Result_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1348;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1348;
               Finalizer_Scope_3080;
               Finalizer_Scope_3087;
               Finalizer_Scope_3088;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Requeue_Stmt_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Return_Stmt
   --
      procedure Initialize_Fields_For_Return_Stmt
        (Self : Bare_Return_Stmt
         ; Return_Stmt_F_Return_Expr : Bare_Expr
        ) is
      begin
            Self.Return_Stmt_F_Return_Expr := Return_Stmt_F_Return_Expr;
      end Initialize_Fields_For_Return_Stmt;
   function Return_Stmt_F_Return_Expr
     (Node : Bare_Return_Stmt) return Bare_Expr
   is
   begin
         return Node.Return_Stmt_F_Return_Expr;
   end;
--# property-start ReturnStmt.subp /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19184
pragma Warnings (Off, "is not referenced");
 function Return_Stmt_P_Subp
  (Node : Bare_Return_Stmt
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Subp_Body
is
   Self : Bare_Return_Stmt := Bare_Return_Stmt (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Subp_Body;
            procedure Finalizer_Scope_1345 with Inline_Always;
      Node_Parents : Bare_Ada_Node_Array_Access;
Is_A : Boolean;
Find_Result : Bare_Ada_Node;
Cast_Result : Bare_Subp_Body;
As_Entity : Internal_Entity_Subp_Body;
Is_Null : Boolean;
If_Result : Internal_Entity_Subp_Body;
            procedure Finalizer_Scope_1345 is
            begin
                     Dec_Ref (Node_Parents);
            end Finalizer_Scope_1345;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 10939 '<as_entity at ast.py:19185>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19185
--# expr-start 10938 '<Cast to SubpBody at ast.py:19185>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19185
--# expr-start 10937 '<Find at ast.py:19185>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19185
   --# expr-start 10935 '<FieldAccess for parents at ast.py:19185>' Node_Parents /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19185
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents := Parents (Self, True); 
--# expr-done 10935
      Find_Result := No_Bare_Ada_Node;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents;
      begin
         for Item_329 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_329
      --# expr-start 10936 '<IsA SubpBody at ast.py:19185>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19185
Is_A := Item_329 /= null 
and then Item_329.Kind in Ada_Subp_Body_Range; 
--# expr-done 10936
      if Is_A then
         Find_Result := Item_329;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 10937
   if Find_Result = null
      or else Find_Result.Kind in Ada_Subp_Body_Range
   then
      Cast_Result := Find_Result;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 10938
Is_Null := Cast_Result = null; 
if Is_Null then
   If_Result := No_Entity_Subp_Body;
else
As_Entity := (Info => E_Info, Node => Cast_Result); 
   If_Result := As_Entity;
end if;
--# expr-done 10939
      Property_Result := If_Result;
   --# end
      Finalizer_Scope_1345;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1345;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Return_Stmt_P_Subp;
--# end
--# property-start ReturnStmt.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19190
pragma Warnings (Off, "is not referenced");
 function Return_Stmt_P_Xref_Equation
  (Node : Bare_Return_Stmt
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Return_Stmt := Bare_Return_Stmt (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Return_Stmt :=
     Internal_Entity_Return_Stmt'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1346 with Inline_Always;
            procedure Finalizer_Scope_3090 with Inline_Always;
      Var_Expr : Internal_Entity_Expr;
Fld : Internal_Entity_Expr;
Fld_1 : Logic_Var;
Fld_2 : Internal_Entity_Subp_Body;
Fld_3 : Internal_Entity_Subp_Spec;
Fld_4 : Internal_Entity_Type_Expr;
Fld_5 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_6 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_7 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1346 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1346;
            procedure Finalizer_Scope_3090 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_7);
                     Dec_Ref (And_Pred_1);
            end Finalizer_Scope_3090;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10951 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10940 '<FieldAccess for return_expr at ast.py:19191>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19191
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Return_Stmt_F_Return_Expr, Info => Ent.Info);
--# expr-done 10940
Var_Expr := Fld;
if Var_Expr /= No_Entity_Expr then
   --# scope-start
   --# bind rexpr Var_Expr
--# expr-start 10950 '<And at ast.py:19193>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19193
--# expr-start 10948 '<And at ast.py:19193>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19193
--# expr-start 10946 '<Bind at ast.py:19193>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19193
--# expr-start 10941 '<FieldAccess for expected_type_var at ast.py:19193>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19193
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Var_Expr.Node);
--# expr-done 10941
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 10945 '<FieldAccess for designated_type at ast.py:19194>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19194
--# expr-start 10944 '<FieldAccess for returns at ast.py:19194>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19194
--# expr-start 10943 '<FieldAccess for subp_spec at ast.py:19194>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19194
--# expr-start 10942 '<FieldAccess for subp at ast.py:19194>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19194
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ReturnStmt.subp
Fld_2 := Libadalang.Implementation.Return_Stmt_P_Subp (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 10942
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Subp_Spec (Node => Fld_2.Node.Base_Subp_Body_F_Subp_Spec, Info => Fld_2.Info);
--# expr-done 10943
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseSubpSpec.returns'
Fld_4 := Libadalang.Implementation.Dispatcher_Base_Subp_Spec_P_Returns (Node => Fld_3.Node, E_Info => Fld_3.Info);
--# end
--# expr-done 10944
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_5 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Fld_4.Node, Origin => Origin, E_Info => Fld_4.Info);
--# end
--# expr-done 10945
      Cast_Result := Create_Internal_Entity
        (Node => Fld_5.Node,
         Info => Fld_5.Info);
Bind_Result := Solver.Create_Assign (Fld_1, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19193") else null)); 
--# expr-done 10946
--# expr-start 10947 '<FieldAccess for sub_equation at ast.py:19195>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19195
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_6 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 10947
And_Pred := Create_And (Bind_Result, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19193") else null)); 
--# expr-done 10948
--# expr-start 10949 '<FieldAccess for matches_expected_assign_type at ast.py:19196>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19196
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_assign_type
Fld_7 := Libadalang.Implementation.Expr_P_Matches_Expected_Assign_Type (Node => Var_Expr.Node, Origin => Origin);
--# end
--# expr-done 10949
And_Pred_1 := Create_And (And_Pred, Fld_7, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19193") else null)); 
--# expr-done 10950
Scope_Result := And_Pred_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_3090;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19197") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
Let_Result := Result_Var; Inc_Ref (Let_Result);
--# end
--# expr-done 10951
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1346;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1346;
               Finalizer_Scope_3090;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Return_Stmt_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Terminate_Alternative
   --
--# property-start TerminateAlternative.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:19689
pragma Warnings (Off, "is not referenced");
 function Terminate_Alternative_P_Xref_Equation
  (Node : Bare_Terminate_Alternative
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Terminate_Alternative := Bare_Terminate_Alternative (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1374 with Inline_Always;
      True_Rel : Logic_Equation;
            procedure Finalizer_Scope_1374 is
            begin
                     Dec_Ref (True_Rel);
            end Finalizer_Scope_1374;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:19689") else null)); 
      Property_Result := True_Rel;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1374;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1374;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Terminate_Alternative_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Subp_Kind
   --
   --
   --  Primitives for Bare_Subp_Kind_Function
   --
   --
   --  Primitives for Bare_Subp_Kind_Procedure
   --
   --
   --  Primitives for Bare_Subunit
   --
      procedure Initialize_Fields_For_Subunit
        (Self : Bare_Subunit
         ; Subunit_F_Name : Bare_Name
         ; Subunit_F_Body : Bare_Body_Node
        ) is
      begin
            Self.Subunit_F_Name := Subunit_F_Name;
            Self.Subunit_F_Body := Subunit_F_Body;
      end Initialize_Fields_For_Subunit;
   function Subunit_F_Name
     (Node : Bare_Subunit) return Bare_Name
   is
   begin
         return Node.Subunit_F_Name;
   end;
   function Subunit_F_Body
     (Node : Bare_Subunit) return Bare_Body_Node
   is
   begin
         return Node.Subunit_F_Body;
   end;
--# property-start Subunit.root_unit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20022
pragma Warnings (Off, "is not referenced");
 function Subunit_P_Root_Unit
  (Node : Bare_Subunit
  )
   return Internal_Entity_Compilation_Unit
is
   Self : Bare_Subunit := Bare_Subunit (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Compilation_Unit;
            procedure Finalizer_Scope_1394 with Inline_Always;
      Fld : Bare_Name;
Fld_1 : Symbol_Type_Array_Access;
Fld_2 : Bare_Compilation_Unit;
As_Entity : Internal_Entity_Compilation_Unit;
Is_Null : Boolean;
If_Result : Internal_Entity_Compilation_Unit;
Let_Result : Internal_Entity_Compilation_Unit;
            procedure Finalizer_Scope_1394 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_1394;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10956 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10955 '<as_bare_entity at ast.py:20026>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20026
--# expr-start 10954 '<FieldAccess for designated_compilation_unit at ast.py:20026>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20026
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10953 '<FieldAccess for as_symbol_array at ast.py:20027>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20027
--# expr-start 10952 '<FieldAccess for name at ast.py:20027>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20027
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Subunit_F_Name;
--# expr-done 10952
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.as_symbol_array'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_As_Symbol_Array (Node => Fld);
--# end
--# expr-done 10953
--# property-call-start AdaNode.designated_compilation_unit
Fld_2 := Libadalang.Implementation.Ada_Node_P_Designated_Compilation_Unit (Node => Self, Name => Fld_1, Kind => Unit_Body, Load_If_Needed => True, Not_Found_Is_Error => True);
--# end
--# expr-done 10954
Is_Null := Fld_2 = null; 
if Is_Null then
   If_Result := No_Entity_Compilation_Unit;
else
As_Entity := (Info => No_Entity_Info, Node => Fld_2); 
   If_Result := As_Entity;
end if;
--# expr-done 10955
Let_Result := If_Result; 
--# end
--# expr-done 10956
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1394;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1394;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subunit_P_Root_Unit;
--# end
--# property-start Subunit.env_hook_subunit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20034
pragma Warnings (Off, "is not referenced");
 function Subunit_P_Env_Hook_Subunit
  (Node : Bare_Subunit
  )
   return Boolean
is
   Self : Bare_Subunit := Bare_Subunit (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Block_Var : Internal_Entity_Compilation_Unit;
Fld : Internal_Entity_Compilation_Unit;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10958 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10957 '<FieldAccess for root_unit at ast.py:20040>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20040
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Subunit.root_unit
Fld := Libadalang.Implementation.Subunit_P_Root_Unit (Node => Self);
--# end
--# expr-done 10957
Block_Var := Fld; 
Let_Result := False; 
--# end
--# expr-done 10958
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subunit_P_Env_Hook_Subunit;
--# end
--# property-start Subunit.body_root /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20044
pragma Warnings (Off, "is not referenced");
 function Subunit_P_Body_Root
  (Node : Bare_Subunit
  )
   return Internal_Entity_Basic_Decl
is
   Self : Bare_Subunit := Bare_Subunit (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Basic_Decl;
      Fld : Internal_Entity_Compilation_Unit;
Fld_1 : Bare_Basic_Decl;
As_Entity : Internal_Entity_Basic_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Decl;
Let_Result : Internal_Entity_Basic_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10962 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10961 '<as_bare_entity at ast.py:20048>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20048
--# expr-start 10960 '<FieldAccess for decl at ast.py:20048>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20048
--# expr-start 10959 '<FieldAccess for root_unit at ast.py:20048>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20048
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Subunit.root_unit
Fld := Libadalang.Implementation.Subunit_P_Root_Unit (Node => Self);
--# end
--# expr-done 10959
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.decl
Fld_1 := Libadalang.Implementation.Compilation_Unit_P_Decl (Node => Fld.Node);
--# end
--# expr-done 10960
Is_Null := Fld_1 = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Decl;
else
As_Entity := (Info => No_Entity_Info, Node => Fld_1); 
   If_Result := As_Entity;
end if;
--# expr-done 10961
Let_Result := If_Result; 
--# end
--# expr-done 10962
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subunit_P_Body_Root;
--# end
--# property-start Subunit.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20050
pragma Warnings (Off, "is not referenced");
 function Subunit_P_Xref_Entry_Point
  (Node : Bare_Subunit
  )
   return Boolean
is
   Self : Bare_Subunit := Bare_Subunit (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subunit_P_Xref_Entry_Point;
--# end
--# property-start Subunit.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20051
pragma Warnings (Off, "is not referenced");
 function Subunit_P_Xref_Equation
  (Node : Bare_Subunit
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Subunit := Bare_Subunit (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Subunit :=
     Internal_Entity_Subunit'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1398 with Inline_Always;
      Fld : Bare_Name;
Fld_1 : Logic_Var;
Fld_2 : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
            procedure Finalizer_Scope_1398 is
            begin
                     Dec_Ref (Bind_Result);
            end Finalizer_Scope_1398;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10966 '<Bind at ast.py:20051>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20051
--# expr-start 10964 '<FieldAccess for ref_var at ast.py:20051>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20051
--# expr-start 10963 '<FieldAccess for name at ast.py:20051>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20051
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Subunit_F_Name;
--# expr-done 10963
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld_1 := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Fld);
--# end
--# expr-done 10964
Fld_1.Value := No_Entity;
Entity_Vars.Reset (Fld_1);
--# expr-start 10965 '<FieldAccess for body_root at ast.py:20051>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20051
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Subunit.body_root
Fld_2 := Libadalang.Implementation.Subunit_P_Body_Root (Node => Ent.Node);
--# end
--# expr-done 10965
      Cast_Result := Create_Internal_Entity
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
Bind_Result := Solver.Create_Assign (Fld_1, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:20051") else null)); 
--# expr-done 10966
      Property_Result := Bind_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1398;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1398;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subunit_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Synchronized_Node
   --
--# property-start '[dispatcher]Synchronized.as_bool' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Synchronized_Node_P_As_Bool
  (Node : Bare_Synchronized_Node
  )
   return Boolean
is
   Self : Bare_Synchronized_Node := Bare_Synchronized_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Synchronized_Node (Self.Kind) is
               when Ada_Synchronized_Absent =>
                  --# property-call-start Synchronized.Absent.as_bool
                  Property_Result := Synchronized_Absent_P_As_Bool
                    (Self
                    );
                  --# end
               when Ada_Synchronized_Present =>
                  --# property-call-start Synchronized.Present.as_bool
                  Property_Result := Synchronized_Present_P_As_Bool
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Synchronized_Node_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Synchronized_Absent
   --
--# property-start Synchronized.Absent.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4954
pragma Warnings (Off, "is not referenced");
 function Synchronized_Absent_P_As_Bool
  (Node : Bare_Synchronized_Absent
  )
   return Boolean
is
   Self : Bare_Synchronized_Absent := Bare_Synchronized_Absent (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Synchronized_Absent_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Synchronized_Present
   --
--# property-start Synchronized.Present.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4954
pragma Warnings (Off, "is not referenced");
 function Synchronized_Present_P_As_Bool
  (Node : Bare_Synchronized_Present
  )
   return Boolean
is
   Self : Bare_Synchronized_Present := Bare_Synchronized_Present (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Synchronized_Present_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Tagged_Node
   --
--# property-start '[dispatcher]Tagged.as_bool' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Tagged_Node_P_As_Bool
  (Node : Bare_Tagged_Node
  )
   return Boolean
is
   Self : Bare_Tagged_Node := Bare_Tagged_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Tagged_Node (Self.Kind) is
               when Ada_Tagged_Absent =>
                  --# property-call-start Tagged.Absent.as_bool
                  Property_Result := Tagged_Absent_P_As_Bool
                    (Self
                    );
                  --# end
               when Ada_Tagged_Present =>
                  --# property-call-start Tagged.Present.as_bool
                  Property_Result := Tagged_Present_P_As_Bool
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Tagged_Node_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Tagged_Absent
   --
--# property-start Tagged.Absent.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4858
pragma Warnings (Off, "is not referenced");
 function Tagged_Absent_P_As_Bool
  (Node : Bare_Tagged_Absent
  )
   return Boolean
is
   Self : Bare_Tagged_Absent := Bare_Tagged_Absent (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Tagged_Absent_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Tagged_Present
   --
--# property-start Tagged.Present.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4858
pragma Warnings (Off, "is not referenced");
 function Tagged_Present_P_As_Bool
  (Node : Bare_Tagged_Present
  )
   return Boolean
is
   Self : Bare_Tagged_Present := Bare_Tagged_Present (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Tagged_Present_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Task_Def
   --
      procedure Initialize_Fields_For_Task_Def
        (Self : Bare_Task_Def
         ; Task_Def_F_Interfaces : Bare_Parent_List
         ; Task_Def_F_Public_Part : Bare_Public_Part
         ; Task_Def_F_Private_Part : Bare_Private_Part
         ; Task_Def_F_End_Name : Bare_End_Name
        ) is
      begin
            Self.Task_Def_F_Interfaces := Task_Def_F_Interfaces;
            Self.Task_Def_F_Public_Part := Task_Def_F_Public_Part;
            Self.Task_Def_F_Private_Part := Task_Def_F_Private_Part;
            Self.Task_Def_F_End_Name := Task_Def_F_End_Name;
      end Initialize_Fields_For_Task_Def;
   function Task_Def_F_Interfaces
     (Node : Bare_Task_Def) return Bare_Parent_List
   is
   begin
         return Node.Task_Def_F_Interfaces;
   end;
   function Task_Def_F_Public_Part
     (Node : Bare_Task_Def) return Bare_Public_Part
   is
   begin
         return Node.Task_Def_F_Public_Part;
   end;
   function Task_Def_F_Private_Part
     (Node : Bare_Task_Def) return Bare_Private_Part
   is
   begin
         return Node.Task_Def_F_Private_Part;
   end;
   function Task_Def_F_End_Name
     (Node : Bare_Task_Def) return Bare_End_Name
   is
   begin
         return Node.Task_Def_F_End_Name;
   end;
--# property-start TaskDef.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8026
pragma Warnings (Off, "is not referenced");
 function Task_Def_P_Xref_Equation
  (Node : Bare_Task_Def
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Task_Def := Bare_Task_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Task_Def :=
     Internal_Entity_Task_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_648 with Inline_Always;
            procedure Finalizer_Scope_3091 with Inline_Always;
      Fld : Internal_Entity_Parent_List;
Fld_1 : Bare_Parent_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld_3 : Logic_Equation;
Item_330 : Internal_Entity_Name;
Bare_Item_330 : Bare_Name;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_648 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_648;
            procedure Finalizer_Scope_3091 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_3091;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10971 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10970 '<LogicAll at ast.py:8027>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8027
--# expr-start 10969 '<Map at ast.py:8027>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8027
   --# expr-start 10967 '<FieldAccess for interfaces at ast.py:8027>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8027
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Parent_List (Node => Ent.Node.Task_Def_F_Interfaces, Info => Ent.Info);
--# expr-done 10967
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Parent_List := Fld_1;
      begin
         for Untyped_Item_330 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_330 := Untyped_Item_330; 
Is_Null := Bare_Item_330 = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_330); 
   If_Result := As_Entity;
end if;
                  Item_330 := If_Result; 
   --# scope-start
               --# bind ifc Item_330
      --# expr-start 10968 '<FieldAccess for xref_equation at ast.py:8027>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8027
      if Item_330.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_3 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Item_330.Node, Env => Env, Origin => Origin, E_Info => Item_330.Info);
--# end
--# expr-done 10968
         declare
            Item_To_Append : constant Logic_Equation := Fld_3;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_3091;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 10969
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8027") else null)); 
--# expr-done 10970
Let_Result := Logic_Boolean_Op; Inc_Ref (Let_Result);
--# end
--# expr-done 10971
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_648;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_648;
               Finalizer_Scope_3091;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Task_Def_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Type_Def
   --
--# property-start '[dispatcher]TypeDef.discrete_range' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Type_Def_P_Discrete_Range
  (Node : Bare_Type_Def
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Discrete_Range
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Discrete_Range;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Type_Def (Self.Kind) is
               when Ada_Access_To_Subp_Def .. Ada_Array_Type_Def | Ada_Enum_Type_Def .. Ada_Interface_Type_Def | Ada_Private_Type_Def .. Ada_Record_Type_Def =>
                  --# property-call-start TypeDef.discrete_range
                  Property_Result := Type_Def_P_Discrete_Range
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Derived_Type_Def =>
                  --# property-call-start DerivedTypeDef.discrete_range
                  Property_Result := Derived_Type_Def_P_Discrete_Range
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Mod_Int_Type_Def =>
                  --# property-call-start ModIntTypeDef.discrete_range
                  Property_Result := Mod_Int_Type_Def_P_Discrete_Range
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Signed_Int_Type_Def =>
                  --# property-call-start SignedIntTypeDef.discrete_range
                  Property_Result := Signed_Int_Type_Def_P_Discrete_Range
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Type_Def_P_Discrete_Range;
--# end
--# property-start '[dispatcher]TypeDef.array_ndims' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Type_Def_P_Array_Ndims
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Integer
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Integer;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Type_Def (Self.Kind) is
               when Ada_Access_To_Subp_Def .. Ada_Type_Access_Def | Ada_Enum_Type_Def .. Ada_Signed_Int_Type_Def =>
                  --# property-call-start TypeDef.array_ndims
                  Property_Result := Type_Def_P_Array_Ndims
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Array_Type_Def =>
                  --# property-call-start ArrayTypeDef.array_ndims
                  Property_Result := Array_Type_Def_P_Array_Ndims
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Derived_Type_Def =>
                  --# property-call-start DerivedTypeDef.array_ndims
                  Property_Result := Derived_Type_Def_P_Array_Ndims
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Type_Def_P_Array_Ndims;
--# end
--# property-start TypeDef.is_real_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4375
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Is_Real_Type
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Def :=
     Internal_Entity_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Boolean;
Fld_1 : Boolean;
If_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
--# expr-start 10972 '<FieldAccess for is_float_type at ast.py:4376>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4376
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.is_float_type'
Fld := Libadalang.Implementation.Dispatcher_Type_Def_P_Is_Float_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 10972
if Fld then
   If_Result := True;
else
   --# expr-start 10973 '<FieldAccess for is_fixed_point at ast.py:4376>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4376
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.is_fixed_point'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Is_Fixed_Point (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 10973
   If_Result := Fld_1;
end if;
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Is_Real_Type;
--# end
--# property-start '[dispatcher]TypeDef.is_float_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Type_Def_P_Is_Float_Type
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Type_Def (Self.Kind) is
               when Ada_Access_To_Subp_Def .. Ada_Array_Type_Def | Ada_Enum_Type_Def .. Ada_Decimal_Fixed_Point_Def | Ada_Ordinary_Fixed_Point_Def .. Ada_Signed_Int_Type_Def =>
                  --# property-call-start TypeDef.is_float_type
                  Property_Result := Type_Def_P_Is_Float_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Derived_Type_Def =>
                  --# property-call-start DerivedTypeDef.is_float_type
                  Property_Result := Derived_Type_Def_P_Is_Float_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Floating_Point_Def =>
                  --# property-call-start FloatingPointDef.is_float_type
                  Property_Result := Floating_Point_Def_P_Is_Float_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Type_Def_P_Is_Float_Type;
--# end
--# property-start '[dispatcher]TypeDef.is_fixed_point' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Type_Def_P_Is_Fixed_Point
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Type_Def (Self.Kind) is
               when Ada_Access_To_Subp_Def .. Ada_Array_Type_Def | Ada_Enum_Type_Def .. Ada_Private_Type_Def | Ada_Floating_Point_Def | Ada_Record_Type_Def .. Ada_Signed_Int_Type_Def =>
                  --# property-call-start TypeDef.is_fixed_point
                  Property_Result := Type_Def_P_Is_Fixed_Point
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Derived_Type_Def =>
                  --# property-call-start DerivedTypeDef.is_fixed_point
                  Property_Result := Derived_Type_Def_P_Is_Fixed_Point
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Decimal_Fixed_Point_Def =>
                  --# property-call-start DecimalFixedPointDef.is_fixed_point
                  Property_Result := Decimal_Fixed_Point_Def_P_Is_Fixed_Point
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Ordinary_Fixed_Point_Def =>
                  --# property-call-start OrdinaryFixedPointDef.is_fixed_point
                  Property_Result := Ordinary_Fixed_Point_Def_P_Is_Fixed_Point
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Type_Def_P_Is_Fixed_Point;
--# end
--# property-start '[dispatcher]TypeDef.predefined_operators' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Type_Def_P_Predefined_Operators
  (Node : Bare_Type_Def
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Type_Def (Self.Kind) is
               when Ada_Derived_Type_Def | Ada_Interface_Type_Def =>
                  --# property-call-start TypeDef.predefined_operators
                  Property_Result := Type_Def_P_Predefined_Operators
                    (Self
                    );
                  --# end
               when Ada_Access_To_Subp_Def .. Ada_Type_Access_Def =>
                  --# property-call-start AccessDef.predefined_operators
                  Property_Result := Access_Def_P_Predefined_Operators
                    (Self
                    );
                  --# end
               when Ada_Array_Type_Def =>
                  --# property-call-start ArrayTypeDef.predefined_operators
                  Property_Result := Array_Type_Def_P_Predefined_Operators
                    (Self
                    );
                  --# end
               when Ada_Enum_Type_Def =>
                  --# property-call-start EnumTypeDef.predefined_operators
                  Property_Result := Enum_Type_Def_P_Predefined_Operators
                    (Self
                    );
                  --# end
               when Ada_Formal_Discrete_Type_Def =>
                  --# property-call-start FormalDiscreteTypeDef.predefined_operators
                  Property_Result := Formal_Discrete_Type_Def_P_Predefined_Operators
                    (Self
                    );
                  --# end
               when Ada_Mod_Int_Type_Def =>
                  --# property-call-start ModIntTypeDef.predefined_operators
                  Property_Result := Mod_Int_Type_Def_P_Predefined_Operators
                    (Self
                    );
                  --# end
               when Ada_Private_Type_Def =>
                  --# property-call-start PrivateTypeDef.predefined_operators
                  Property_Result := Private_Type_Def_P_Predefined_Operators
                    (Self
                    );
                  --# end
               when Ada_Floating_Point_Def =>
                  --# property-call-start RealTypeDef.predefined_operators
                  Property_Result := Real_Type_Def_P_Predefined_Operators
                    (Self
                    );
                  --# end
               when Ada_Decimal_Fixed_Point_Def =>
                  --# property-call-start DecimalFixedPointDef.predefined_operators
                  Property_Result := Decimal_Fixed_Point_Def_P_Predefined_Operators
                    (Self
                    );
                  --# end
               when Ada_Ordinary_Fixed_Point_Def =>
                  --# property-call-start OrdinaryFixedPointDef.predefined_operators
                  Property_Result := Ordinary_Fixed_Point_Def_P_Predefined_Operators
                    (Self
                    );
                  --# end
               when Ada_Record_Type_Def =>
                  --# property-call-start RecordTypeDef.predefined_operators
                  Property_Result := Record_Type_Def_P_Predefined_Operators
                    (Self
                    );
                  --# end
               when Ada_Signed_Int_Type_Def =>
                  --# property-call-start SignedIntTypeDef.predefined_operators
                  Property_Result := Signed_Int_Type_Def_P_Predefined_Operators
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Type_Def_P_Predefined_Operators;
--# end
--# property-start TypeDef.predefined_equality_operators /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4402
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Predefined_Equality_Operators
  (Node : Bare_Type_Def
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_277 with Inline_Always;
      Block_Var : Bare_Type_Decl;
Block_Var_1 : Bare_Base_Type_Decl;
Fld : Bare_Ada_Node;
Cast_Result : Bare_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Bare_Base_Type_Decl;
Sym : Symbol_Type;
Cast_Expr : Bare_Type_Decl;
Cast_Result_1 : Bare_Base_Type_Decl;
Cast_Expr_1 : Bare_Type_Decl;
Cast_Result_2 : Bare_Base_Type_Decl;
Fld_3 : Internal_Env_Assoc := No_Env_Assoc;
Sym_1 : Symbol_Type;
Cast_Expr_2 : Bare_Type_Decl;
Cast_Result_3 : Bare_Base_Type_Decl;
Cast_Expr_3 : Bare_Type_Decl;
Cast_Result_4 : Bare_Base_Type_Decl;
Fld_4 : Internal_Env_Assoc := No_Env_Assoc;
Array_Lit : Internal_Env_Assoc_Array_Access;
Let_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_277 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_277;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10983 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 10975 '<Cast to TypeDecl at ast.py:4403>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4403
--# expr-start 10974 '<parent at ast.py:4403>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4403
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 10974
   if Fld = null
      or else Fld.Kind in Ada_Type_Decl_Range
   then
      Cast_Result := Fld;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 10975
Block_Var := Cast_Result; 
--# bind self_type Block_Var
--# expr-start 10977 '<FieldAccess for node at ast.py:4404>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4404
--# expr-start 10976 '<FieldAccess for bool_type at ast.py:4404>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4404
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_1 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 10976
Fld_2 := Fld_1.Node;
--# expr-done 10977
Block_Var_1 := Fld_2; 
--# bind bool_type Block_Var_1
--# expr-start 10982 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 10979 '<FieldAccess for create_binop_assoc at ast.py:4406>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4406
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10978 '<Symbol "=" at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_10); 
--# expr-done 10978
Cast_Expr := Block_Var; 
      Cast_Result_1 := Cast_Expr;
Cast_Expr_1 := Block_Var; 
      Cast_Result_2 := Cast_Expr_1;
--# property-call-start AdaNode.create_binop_assoc
Fld_3 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym, Lhs => Cast_Result_1, Rhs => Cast_Result_2, Ret => Block_Var_1);
--# end
--# expr-done 10979
--# expr-start 10981 '<FieldAccess for create_binop_assoc at ast.py:4407>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4407
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 10980 '<Symbol "/=" at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_7); 
--# expr-done 10980
Cast_Expr_2 := Block_Var; 
      Cast_Result_3 := Cast_Expr_2;
Cast_Expr_3 := Block_Var; 
      Cast_Result_4 := Cast_Expr_3;
--# property-call-start AdaNode.create_binop_assoc
Fld_4 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_1, Lhs => Cast_Result_3, Rhs => Cast_Result_4, Ret => Block_Var_1);
--# end
--# expr-done 10981
Array_Lit := Create_Internal_Env_Assoc_Array (Internal_Internal_Env_Assoc_Array'(1 => Fld_3, 2 => Fld_4)); 
--# expr-done 10982
Let_Result := Array_Lit; Inc_Ref (Let_Result);
--# end
--# expr-done 10983
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_277;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_277;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Predefined_Equality_Operators;
--# end
--# property-start '[dispatcher]TypeDef.is_discrete_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Type_Def_P_Is_Discrete_Type
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Def :=
     Internal_Entity_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Type_Def (Self.Kind) is
               when Ada_Access_To_Subp_Def .. Ada_Enum_Type_Def | Ada_Interface_Type_Def .. Ada_Signed_Int_Type_Def =>
                  --# property-call-start TypeDef.is_discrete_type
                  Property_Result := Type_Def_P_Is_Discrete_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Formal_Discrete_Type_Def =>
                  --# property-call-start FormalDiscreteTypeDef.is_discrete_type
                  Property_Result := Formal_Discrete_Type_Def_P_Is_Discrete_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Type_Def_P_Is_Discrete_Type;
--# end
--# property-start '[dispatcher]TypeDef.is_int_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Type_Def_P_Is_Int_Type
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Type_Def (Self.Kind) is
               when Ada_Access_To_Subp_Def .. Ada_Array_Type_Def | Ada_Enum_Type_Def .. Ada_Interface_Type_Def | Ada_Private_Type_Def .. Ada_Record_Type_Def =>
                  --# property-call-start TypeDef.is_int_type
                  Property_Result := Type_Def_P_Is_Int_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Derived_Type_Def =>
                  --# property-call-start DerivedTypeDef.is_int_type
                  Property_Result := Derived_Type_Def_P_Is_Int_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Mod_Int_Type_Def =>
                  --# property-call-start ModIntTypeDef.is_int_type
                  Property_Result := Mod_Int_Type_Def_P_Is_Int_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Signed_Int_Type_Def =>
                  --# property-call-start SignedIntTypeDef.is_int_type
                  Property_Result := Signed_Int_Type_Def_P_Is_Int_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Type_Def_P_Is_Int_Type;
--# end
--# property-start '[dispatcher]TypeDef.is_access_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Type_Def_P_Is_Access_Type
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
  )
   return Boolean
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Type_Def (Self.Kind) is
               when Ada_Array_Type_Def | Ada_Enum_Type_Def .. Ada_Signed_Int_Type_Def =>
                  --# property-call-start TypeDef.is_access_type
                  Property_Result := Type_Def_P_Is_Access_Type
                    (Self
                        , Origin
                    );
                  --# end
               when Ada_Access_To_Subp_Def .. Ada_Type_Access_Def =>
                  --# property-call-start AccessDef.is_access_type
                  Property_Result := Access_Def_P_Is_Access_Type
                    (Self
                        , Origin
                    );
                  --# end
               when Ada_Derived_Type_Def =>
                  --# property-call-start DerivedTypeDef.is_access_type
                  Property_Result := Derived_Type_Def_P_Is_Access_Type
                    (Self
                        , Origin
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Type_Def_P_Is_Access_Type;
--# end
--# property-start '[dispatcher]TypeDef.is_char_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Type_Def_P_Is_Char_Type
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Type_Def (Self.Kind) is
               when Ada_Access_To_Subp_Def .. Ada_Array_Type_Def | Ada_Formal_Discrete_Type_Def .. Ada_Signed_Int_Type_Def =>
                  --# property-call-start TypeDef.is_char_type
                  Property_Result := Type_Def_P_Is_Char_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Derived_Type_Def =>
                  --# property-call-start DerivedTypeDef.is_char_type
                  Property_Result := Derived_Type_Def_P_Is_Char_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Enum_Type_Def =>
                  --# property-call-start EnumTypeDef.is_char_type
                  Property_Result := Enum_Type_Def_P_Is_Char_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Type_Def_P_Is_Char_Type;
--# end
--# property-start '[dispatcher]TypeDef.is_enum_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Type_Def_P_Is_Enum_Type
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Type_Def (Self.Kind) is
               when Ada_Access_To_Subp_Def .. Ada_Array_Type_Def | Ada_Formal_Discrete_Type_Def .. Ada_Signed_Int_Type_Def =>
                  --# property-call-start TypeDef.is_enum_type
                  Property_Result := Type_Def_P_Is_Enum_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Derived_Type_Def =>
                  --# property-call-start DerivedTypeDef.is_enum_type
                  Property_Result := Derived_Type_Def_P_Is_Enum_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Enum_Type_Def =>
                  --# property-call-start EnumTypeDef.is_enum_type
                  Property_Result := Enum_Type_Def_P_Is_Enum_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Type_Def_P_Is_Enum_Type;
--# end
--# property-start '[dispatcher]TypeDef.accessed_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Type_Def_P_Accessed_Type
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Type_Def (Self.Kind) is
               when Ada_Array_Type_Def | Ada_Enum_Type_Def .. Ada_Signed_Int_Type_Def =>
                  --# property-call-start TypeDef.accessed_type
                  Property_Result := Type_Def_P_Accessed_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Access_To_Subp_Def =>
                  --# property-call-start AccessToSubpDef.accessed_type
                  Property_Result := Access_To_Subp_Def_P_Accessed_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Anonymous_Type_Access_Def =>
                  --# property-call-start AnonymousTypeAccessDef.accessed_type
                  Property_Result := Anonymous_Type_Access_Def_P_Accessed_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Type_Access_Def =>
                  --# property-call-start TypeAccessDef.accessed_type
                  Property_Result := Type_Access_Def_P_Accessed_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Derived_Type_Def =>
                  --# property-call-start DerivedTypeDef.accessed_type
                  Property_Result := Derived_Type_Def_P_Accessed_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Type_Def_P_Accessed_Type;
--# end
--# property-start '[dispatcher]TypeDef.is_tagged_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Type_Def_P_Is_Tagged_Type
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Type_Def (Self.Kind) is
               when Ada_Access_To_Subp_Def .. Ada_Array_Type_Def | Ada_Enum_Type_Def .. Ada_Formal_Discrete_Type_Def | Ada_Mod_Int_Type_Def | Ada_Decimal_Fixed_Point_Def .. Ada_Ordinary_Fixed_Point_Def | Ada_Signed_Int_Type_Def =>
                  --# property-call-start TypeDef.is_tagged_type
                  Property_Result := Type_Def_P_Is_Tagged_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Derived_Type_Def =>
                  --# property-call-start DerivedTypeDef.is_tagged_type
                  Property_Result := Derived_Type_Def_P_Is_Tagged_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Interface_Type_Def =>
                  --# property-call-start InterfaceTypeDef.is_tagged_type
                  Property_Result := Interface_Type_Def_P_Is_Tagged_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Private_Type_Def =>
                  --# property-call-start PrivateTypeDef.is_tagged_type
                  Property_Result := Private_Type_Def_P_Is_Tagged_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Record_Type_Def =>
                  --# property-call-start RecordTypeDef.is_tagged_type
                  Property_Result := Record_Type_Def_P_Is_Tagged_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Type_Def_P_Is_Tagged_Type;
--# end
--# property-start '[dispatcher]TypeDef.base_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Type_Def_P_Base_Type
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Type_Def (Self.Kind) is
               when Ada_Access_To_Subp_Def .. Ada_Array_Type_Def | Ada_Enum_Type_Def .. Ada_Signed_Int_Type_Def =>
                  --# property-call-start TypeDef.base_type
                  Property_Result := Type_Def_P_Base_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Derived_Type_Def =>
                  --# property-call-start DerivedTypeDef.base_type
                  Property_Result := Derived_Type_Def_P_Base_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Type_Def_P_Base_Type;
--# end
--# property-start TypeDef.base_types /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4451
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Base_Types
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl_Array_Access
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Def :=
     Internal_Entity_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_286 with Inline_Always;
            procedure Finalizer_Scope_2282 with Inline_Always;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Singleton : Internal_Entity_Base_Type_Decl_Array_Access;
Scope_Result : Internal_Entity_Base_Type_Decl_Array_Access;
Result_Var : Internal_Entity_Base_Type_Decl_Array_Access;
Fld_1 : Internal_Entity_Base_Type_Decl_Array_Access;
Concat_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_286 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Concat_Result);
            end Finalizer_Scope_286;
            procedure Finalizer_Scope_2282 is
            begin
                     Dec_Ref (Singleton);
            end Finalizer_Scope_2282;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 10986 '<Concat at ast.py:4452>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4452
--# expr-start 10984 '<FieldAccess for base_type at ast.py:4452>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4452
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 10984
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind bt Var_Expr
                Singleton := Create_Internal_Entity_Base_Type_Decl_Array (Items_Count => 1);
                Singleton.Items (1) := Var_Expr;
Scope_Result := Singleton;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2282;
   Result_Var := Scope_Result;
else
   Result_Var := No_Internal_Entity_Base_Type_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var);
--# expr-start 10985 '<FieldAccess for base_interfaces at ast.py:4453>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4453
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_interfaces'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Interfaces (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 10985
Concat_Result := Concat (Result_Var, Fld_1); 
--# expr-done 10986
      Property_Result := Concat_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_286;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_286;
               Finalizer_Scope_2282;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Base_Types;
--# end
--# property-start '[dispatcher]TypeDef.base_interfaces' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Type_Def_P_Base_Interfaces
  (Node : Bare_Type_Def
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl_Array_Access
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Type_Decl_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Type_Def (Self.Kind) is
               when Ada_Access_To_Subp_Def .. Ada_Array_Type_Def | Ada_Enum_Type_Def .. Ada_Formal_Discrete_Type_Def | Ada_Mod_Int_Type_Def .. Ada_Signed_Int_Type_Def =>
                  --# property-call-start TypeDef.base_interfaces
                  Property_Result := Type_Def_P_Base_Interfaces
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Derived_Type_Def =>
                  --# property-call-start DerivedTypeDef.base_interfaces
                  Property_Result := Derived_Type_Def_P_Base_Interfaces
                    (Self
                        , E_Info
                    );
                  --# end
               when Ada_Interface_Type_Def =>
                  --# property-call-start InterfaceTypeDef.base_interfaces
                  Property_Result := Interface_Type_Def_P_Base_Interfaces
                    (Self
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Type_Def_P_Base_Interfaces;
--# end
--# property-start TypeDef.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4467
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Defining_Env
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Def :=
     Internal_Entity_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_288 with Inline_Always;
            procedure Finalizer_Scope_2283 with Inline_Always;
            procedure Finalizer_Scope_2284 with Inline_Always;
            procedure Finalizer_Scope_2285 with Inline_Always;
            procedure Finalizer_Scope_2286 with Inline_Always;
            procedure Finalizer_Scope_2287 with Inline_Always;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Match_0 : Internal_Entity_Array_Type_Def;
Match_1 : Internal_Entity_Access_Def;
Match_2 : Internal_Entity_Type_Def;
Var_Expr_1 : Internal_Entity_Base_Type_Decl;
Is_A : Boolean;
Fld : Lexical_Env := Empty_Env;
Fld_1 : Lexical_Env := Empty_Env;
Fld_2 : Lexical_Env := Empty_Env;
Array_Lit : Lexical_Env_Array_Access;
Group_Env : Lexical_Env := Empty_Env;
Is_A_1 : Boolean;
Fld_3 : Lexical_Env := Empty_Env;
Fld_4 : Lexical_Env := Empty_Env;
Fld_5 : Lexical_Env := Empty_Env;
Array_Lit_1 : Lexical_Env_Array_Access;
Fld_6 : Internal_Entity_Base_Type_Decl_Array_Access;
Fld_7 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
Map_Result : Lexical_Env_Array_Access;
Concat_Result : Lexical_Env_Array_Access;
Group_Env_1 : Lexical_Env := Empty_Env;
Fld_8 : Internal_Entity_Base_Type_Decl;
Fld_9 : Lexical_Env := Empty_Env;
Fld_10 : Internal_Entity_Base_Type_Decl;
Fld_11 : Lexical_Env := Empty_Env;
Scope_Result_1 : Lexical_Env := Empty_Env;
Result_Var_1 : Lexical_Env := Empty_Env;
Match_Prefix : Internal_Entity_Type_Def;
Cast_Result : Internal_Entity_Array_Type_Def;
Let_Result : Lexical_Env := Empty_Env;
Scope_Result_2 : Lexical_Env := Empty_Env;
Cast_Result_1 : Internal_Entity_Access_Def;
Let_Result_1 : Lexical_Env := Empty_Env;
Scope_Result_3 : Lexical_Env := Empty_Env;
Cast_Result_2 : Internal_Entity_Type_Def;
Let_Result_2 : Lexical_Env := Empty_Env;
Scope_Result_4 : Lexical_Env := Empty_Env;
Match_Result : Lexical_Env := Empty_Env;
If_Result : Lexical_Env := Empty_Env;
If_Result_1 : Lexical_Env := Empty_Env;
Let_Result_3 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_288 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Fld_1);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Group_Env);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Fld_5);
                     Dec_Ref (Array_Lit_1);
                     Dec_Ref (Fld_6);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Group_Env_1);
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Let_Result_2);
                     Dec_Ref (Scope_Result_4);
                     Dec_Ref (Match_Result);
                     Dec_Ref (If_Result);
                     Dec_Ref (If_Result_1);
                     Dec_Ref (Let_Result_3);
            end Finalizer_Scope_288;
            procedure Finalizer_Scope_2283 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_2283;
            procedure Finalizer_Scope_2284 is
            begin
                     Dec_Ref (Fld_7);
            end Finalizer_Scope_2284;
            procedure Finalizer_Scope_2285 is
            begin
                     Dec_Ref (Fld_9);
            end Finalizer_Scope_2285;
            procedure Finalizer_Scope_2286 is
            begin
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
            end Finalizer_Scope_2286;
            procedure Finalizer_Scope_2287 is
            begin
                     Dec_Ref (Fld_11);
            end Finalizer_Scope_2287;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11009 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 11008 '<Cond at ast.py:4470>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4470
--# expr-start 10987 '<IsA RecordTypeDef, PrivateTypeDef at ast.py:4474>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4474
Is_A := Self /= null 
and then Self.Kind in Ada_Record_Type_Def_Range | Ada_Private_Type_Def_Range; 
--# expr-done 10987
if Is_A then
   --# expr-start 10992 '<env_group at ast.py:4475>' Group_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4475
--# expr-start 10991 '<ArrayLiteral at ast.py:4475>' Array_Lit /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4475
--# expr-start 10988 '<FieldAccess for children_env at ast.py:4476>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4476
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 10988
--# expr-start 10989 '<FieldAccess for dottable_subps_env at ast.py:4477>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4477
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDef.dottable_subps_env
Fld_1 := Libadalang.Implementation.Type_Def_P_Dottable_Subps_Env (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 10989
--# expr-start 10990 '<FieldAccess for previous_part_env at ast.py:4478>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4478
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDef.previous_part_env
Fld_2 := Libadalang.Implementation.Type_Def_P_Previous_Part_Env (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 10990
Array_Lit := Create_Lexical_Env_Array (Internal_Lexical_Env_Array'(1 => Fld, 2 => Fld_1, 3 => Fld_2)); 
--# expr-done 10991
Group_Env := Group (Array_Lit, No_Metadata); 
--# expr-done 10992
   If_Result_1 := Group_Env;
else
--# expr-start 10993 '<IsA DerivedTypeDef, InterfaceTypeDef at ast.py:4483>' Is_A_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4483
Is_A_1 := Self /= null 
and then Self.Kind in Ada_Derived_Type_Def_Range | Ada_Interface_Type_Def_Range; 
--# expr-done 10993
if Is_A_1 then
   --# expr-start 11002 '<env_group at ast.py:4487>' Group_Env_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4487
--# expr-start 11001 '<Concat at ast.py:4487>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4487
--# expr-start 10997 '<ArrayLiteral at ast.py:4487>' Array_Lit_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4487
--# expr-start 10994 '<FieldAccess for children_env at ast.py:4488>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4488
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Libadalang.Implementation.Children_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 10994
--# expr-start 10995 '<FieldAccess for dottable_subps_env at ast.py:4489>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4489
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDef.dottable_subps_env
Fld_4 := Libadalang.Implementation.Type_Def_P_Dottable_Subps_Env (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 10995
--# expr-start 10996 '<FieldAccess for previous_part_env at ast.py:4490>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4490
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDef.previous_part_env
Fld_5 := Libadalang.Implementation.Type_Def_P_Previous_Part_Env (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 10996
Array_Lit_1 := Create_Lexical_Env_Array (Internal_Lexical_Env_Array'(1 => Fld_3, 2 => Fld_4, 3 => Fld_5)); 
--# expr-done 10997
--# expr-start 11000 '<Map at ast.py:4492>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4492
   --# expr-start 10998 '<FieldAccess for base_types at ast.py:4492>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4492
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDef.base_types
Fld_6 := Libadalang.Implementation.Type_Def_P_Base_Types (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 10998
      declare
         Map_Result_Vec : Lexical_Env_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Base_Type_Decl_Array_Access := Fld_6;
      begin
         for Item_331 of
               Collection.Items
         loop
   --# scope-start
               --# bind bt Item_331
Var_Expr := Item_331;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 10999 '<FieldAccess for defining_env at ast.py:4492>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4492
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_7 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Var_Expr.Node, Origin => Origin, Include_Ud_Indexing => Include_Ud_Indexing, E_Info => Var_Expr.Info);
--# end
--# expr-done 10999
Scope_Result := Fld_7;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2284;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
         declare
            Item_To_Append : constant Lexical_Env := Result_Var;
         begin
               Inc_Ref (Item_To_Append);
            Lexical_Env_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_2283;
         end loop;
      end;
         Map_Result := Create_Lexical_Env_Array
           (Items_Count => Natural (Lexical_Env_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Lexical_Env_Vectors.Get
              (Map_Result_Vec,
               I + Lexical_Env_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Lexical_Env_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 11000
Concat_Result := Concat (Array_Lit_1, Map_Result); 
--# expr-done 11001
Group_Env_1 := Group (Concat_Result, No_Metadata); 
--# expr-done 11002
   If_Result := Group_Env_1;
else
   --# expr-start 11007 '<Match at ast.py:4495>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4495
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Type_Def (Match_Prefix.Node.Kind) is
         when Ada_Array_Type_Def =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Array_Type_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind ar Match_0
--# expr-start 11004 '<FieldAccess for defining_env at ast.py:4496>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4496
--# expr-start 11003 '<FieldAccess for comp_type at ast.py:4496>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4496
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ArrayTypeDef.comp_type
Fld_8 := Libadalang.Implementation.Array_Type_Def_P_Comp_Type (Node => Match_0.Node, Origin => Origin, E_Info => Match_0.Info);
--# end
--# expr-done 11003
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_9 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Fld_8.Node, Origin => Origin, Include_Ud_Indexing => Include_Ud_Indexing, E_Info => Fld_8.Info);
--# end
--# expr-done 11004
Let_Result := Fld_9; Inc_Ref (Let_Result);
--# end
Scope_Result_2 := Let_Result;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_2285;
            Match_Result := Scope_Result_2; Inc_Ref (Match_Result);
         when Ada_Access_To_Subp_Def .. Ada_Type_Access_Def =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Access_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind ac Match_1
--# expr-start 11005 '<FieldAccess for accessed_type at ast.py:4500>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4500
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.accessed_type'
Fld_10 := Libadalang.Implementation.Dispatcher_Type_Def_P_Accessed_Type (Node => Match_1.Node, Origin => Origin, E_Info => Match_1.Info);
--# end
--# expr-done 11005
Var_Expr_1 := Fld_10;
if Var_Expr_1 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 11006 '<FieldAccess for defining_env at ast.py:4500>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4500
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_11 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Var_Expr_1.Node, Origin => Origin, Include_Ud_Indexing => Include_Ud_Indexing, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 11006
Scope_Result_1 := Fld_11;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_2287;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := Empty_Env;
end if;
      Inc_Ref (Result_Var_1);
Let_Result_1 := Result_Var_1; Inc_Ref (Let_Result_1);
--# end
Scope_Result_3 := Let_Result_1;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_2286;
            Match_Result := Scope_Result_3; Inc_Ref (Match_Result);
         when Ada_Derived_Type_Def .. Ada_Signed_Int_Type_Def =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Type_Def
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_2; 
--# bind _ Match_2
Let_Result_2 := Empty_Env; Inc_Ref (Let_Result_2);
--# end
Scope_Result_4 := Let_Result_2;
   Inc_Ref (Scope_Result_4);
   --# end
            Match_Result := Scope_Result_4; Inc_Ref (Match_Result);
end case;
--# expr-done 11007
   If_Result := Match_Result;
end if;
      Inc_Ref (If_Result);
   If_Result_1 := If_Result;
end if;
      Inc_Ref (If_Result_1);
--# expr-done 11008
Let_Result_3 := If_Result_1; Inc_Ref (Let_Result_3);
--# end
--# expr-done 11009
      Property_Result := Let_Result_3;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_288;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_288;
               Finalizer_Scope_2283;
               Finalizer_Scope_2284;
               Finalizer_Scope_2285;
               Finalizer_Scope_2286;
               Finalizer_Scope_2287;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Defining_Env;
--# end
--# property-start TypeDef.containing_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4506
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Containing_Type
  (Node : Bare_Type_Def
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Type_Decl
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Def :=
     Internal_Entity_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Type_Decl;
      Fld : Internal_Entity;
Cast_Result : Internal_Entity_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11011 '<Cast to TypeDecl at ast.py:4507>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4507
--# expr-start 11010 '<parent at ast.py:4507>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4507
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 11010
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Type_Decl_Range
   then
      Cast_Result := Create_Internal_Entity_Type_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 11011
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Containing_Type;
--# end
--# property-start TypeDef.previous_part /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4512
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Previous_Part
  (Node : Bare_Type_Def
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Def :=
     Internal_Entity_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11013 '<FieldAccess for previous_part at ast.py:4512>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4512
--# expr-start 11012 '<FieldAccess for containing_type at ast.py:4512>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4512
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDef.containing_type
Fld := Libadalang.Implementation.Type_Def_P_Containing_Type (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 11012
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.previous_part'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Previous_Part (Node => Fld.Node, Go_To_Incomplete => True, E_Info => Fld.Info);
--# end
--# expr-done 11013
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Previous_Part;
--# end
--# property-start TypeDef.previous_part_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4514
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Previous_Part_Env
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Def :=
     Internal_Entity_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_291 with Inline_Always;
            procedure Finalizer_Scope_2281 with Inline_Always;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Result_Var : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_291 is
            begin
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Result_Var);
            end Finalizer_Scope_291;
            procedure Finalizer_Scope_2281 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_2281;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
--# expr-start 11014 '<FieldAccess for previous_part at ast.py:4515>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4515
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDef.previous_part
Fld := Libadalang.Implementation.Type_Def_P_Previous_Part (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 11014
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 11015 '<FieldAccess for defining_env at ast.py:4515>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4515
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Var_Expr.Node, Origin => Origin, Include_Ud_Indexing => False, E_Info => Var_Expr.Info);
--# end
--# expr-done 11015
Scope_Result := Fld_1;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_2281;
   Result_Var := Scope_Result;
else
   Result_Var := Empty_Env;
end if;
      Inc_Ref (Result_Var);
      Property_Result := Result_Var;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_291;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_291;
               Finalizer_Scope_2281;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Previous_Part_Env;
--# end
--# property-start TypeDef.dottable_subps_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4519
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Dottable_Subps_Env
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Def :=
     Internal_Entity_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_292 with Inline_Always;
      Fld : Internal_Entity_Type_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Fld_2 : Internal_Entity_Info;
Fld_3 : Env_Rebindings;
Rebound_Env : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_292 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Rebound_Env);
            end Finalizer_Scope_292;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11020 '<rebind_env at ast.py:4525>' Rebound_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4525
--# expr-start 11017 '<FieldAccess for dottable_subps_env at ast.py:4525>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4525
--# expr-start 11016 '<FieldAccess for containing_type at ast.py:4525>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4525
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDef.containing_type
Fld := Libadalang.Implementation.Type_Def_P_Containing_Type (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 11016
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseTypeDecl.dottable_subps_env
Fld_1 := Libadalang.Implementation.Base_Type_Decl_F_Dottable_Subps_Env (Node => Fld.Node);
--# end
--# expr-done 11017
--# expr-start 11019 '<FieldAccess for rebindings at ast.py:4526>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4526
--# expr-start 11018 '<FieldAccess for info at ast.py:4526>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4526
Fld_2 := Ent.Info;
--# expr-done 11018
Fld_3 := Fld_2.Rebindings;
--# expr-done 11019
Rebound_Env := Rebind_Env (Fld_1, Fld_3); 
--# expr-done 11020
      Property_Result := Rebound_Env;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_292;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_292;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Dottable_Subps_Env;
--# end
--# property-start '[dispatcher]TypeDef.is_static' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Type_Def_P_Is_Static
  (Node : Bare_Type_Def
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Type_Def (Self.Kind) is
               when Ada_Access_To_Subp_Def .. Ada_Type_Access_Def | Ada_Formal_Discrete_Type_Def .. Ada_Interface_Type_Def | Ada_Private_Type_Def | Ada_Record_Type_Def =>
                  --# property-call-start TypeDef.is_static
                  Property_Result := Type_Def_P_Is_Static
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Array_Type_Def =>
                  --# property-call-start ArrayTypeDef.is_static
                  Property_Result := Array_Type_Def_P_Is_Static
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Derived_Type_Def =>
                  --# property-call-start DerivedTypeDef.is_static
                  Property_Result := Derived_Type_Def_P_Is_Static
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Enum_Type_Def =>
                  --# property-call-start EnumTypeDef.is_static
                  Property_Result := Enum_Type_Def_P_Is_Static
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Mod_Int_Type_Def =>
                  --# property-call-start ModIntTypeDef.is_static
                  Property_Result := Mod_Int_Type_Def_P_Is_Static
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Decimal_Fixed_Point_Def .. Ada_Ordinary_Fixed_Point_Def =>
                  --# property-call-start RealTypeDef.is_static
                  Property_Result := Real_Type_Def_P_Is_Static
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
               when Ada_Signed_Int_Type_Def =>
                  --# property-call-start SignedIntTypeDef.is_static
                  Property_Result := Signed_Int_Type_Def_P_Is_Static
                    (Self
                        , Imprecise_Fallback
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Type_Def_P_Is_Static;
--# end
--# property-start TypeDef.discrete_range /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4360
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Discrete_Range
  (Node : Bare_Type_Def
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Discrete_Range
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Discrete_Range;
      Let_Result : Internal_Discrete_Range;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11021 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := No_Discrete_Range; 
--# end
--# expr-done 11021
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Discrete_Range;
--# end
--# property-start TypeDef.array_ndims /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4366
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Array_Ndims
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Integer
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Integer;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := 0;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Array_Ndims;
--# end
--# property-start TypeDef.is_float_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4381
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Is_Float_Type
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Is_Float_Type;
--# end
--# property-start TypeDef.is_fixed_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4385
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Is_Fixed_Point
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Is_Fixed_Point;
--# end
--# property-start TypeDef.predefined_operators /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4391
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Predefined_Operators
  (Node : Bare_Type_Def
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_276 with Inline_Always;
      Let_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_276 is
            begin
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_276;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11022 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := No_Internal_Env_Assoc_Array_Type; Inc_Ref (Let_Result);
--# end
--# expr-done 11022
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_276;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_276;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Predefined_Operators;
--# end
--# property-start TypeDef.is_discrete_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4411
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Is_Discrete_Type
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Def :=
     Internal_Entity_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
Scope_Result : Boolean;
Fld_2 : Boolean;
Fld_3 : Boolean;
If_Result : Boolean;
Fld_4 : Boolean;
If_Result_1 : Boolean;
Result_Var : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11028 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11023 '<FieldAccess for base_type at ast.py:4412>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4412
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 11023
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
   --# bind bt Var_Expr
--# expr-start 11024 '<FieldAccess for is_discrete_type at ast.py:4413>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4413
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_discrete_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Discrete_Type (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 11024
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
--# expr-start 11025 '<FieldAccess for is_int_type at ast.py:4415>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4415
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.is_int_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Type_Def_P_Is_Int_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 11025
if Fld_2 then
   If_Result := True;
else
   --# expr-start 11026 '<FieldAccess for is_enum_type at ast.py:4415>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4415
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.is_enum_type'
Fld_3 := Libadalang.Implementation.Dispatcher_Type_Def_P_Is_Enum_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 11026
   If_Result := Fld_3;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 11027 '<FieldAccess for is_char_type at ast.py:4415>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4415
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.is_char_type'
Fld_4 := Libadalang.Implementation.Dispatcher_Type_Def_P_Is_Char_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 11027
   If_Result_1 := Fld_4;
end if;
   Result_Var := If_Result_1;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 11028
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Is_Discrete_Type;
--# end
--# property-start TypeDef.is_int_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4420
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Is_Int_Type
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
      Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11029 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := False; 
--# end
--# expr-done 11029
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Is_Int_Type;
--# end
--# property-start TypeDef.is_access_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4424
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Is_Access_Type
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
  )
   return Boolean
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Is_Access_Type;
--# end
--# property-start TypeDef.is_char_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4430
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Is_Char_Type
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Is_Char_Type;
--# end
--# property-start TypeDef.is_enum_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4431
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Is_Enum_Type
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Is_Enum_Type;
--# end
--# property-start TypeDef.accessed_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4434
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Accessed_Type
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11030 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := No_Entity_Base_Type_Decl; 
--# end
--# expr-done 11030
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Accessed_Type;
--# end
--# property-start TypeDef.is_tagged_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4438
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Is_Tagged_Type
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
      Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11031 '<Block at ???>' Let_Result None
--# scope-start
Let_Result := False; 
--# end
--# expr-done 11031
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Is_Tagged_Type;
--# end
--# property-start TypeDef.base_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4444
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Base_Type
  (Node : Bare_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := No_Entity_Base_Type_Decl;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Base_Type;
--# end
--# property-start TypeDef.base_interfaces /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4460
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Base_Interfaces
  (Node : Bare_Type_Def
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl_Array_Access
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Entity_Base_Type_Decl_Array_Access;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := No_Internal_Entity_Base_Type_Decl_Array_Type;
         Inc_Ref (Property_Result);
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Base_Interfaces;
--# end
--# property-start TypeDef.is_static /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4531
pragma Warnings (Off, "is not referenced");
 function Type_Def_P_Is_Static
  (Node : Bare_Type_Def
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Type_Def := Bare_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Def_P_Is_Static;
--# end
   --
   --  Primitives for Bare_Access_Def
   --
      procedure Initialize_Fields_For_Access_Def
        (Self : Bare_Access_Def
         ; Access_Def_F_Has_Not_Null : Bare_Not_Null
        ) is
      begin
            Self.Access_Def_F_Has_Not_Null := Access_Def_F_Has_Not_Null;
      end Initialize_Fields_For_Access_Def;
   function Access_Def_F_Has_Not_Null
     (Node : Bare_Access_Def) return Bare_Not_Null
   is
   begin
         return Node.Access_Def_F_Has_Not_Null;
   end;
--# property-start AccessDef.is_access_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8151
pragma Warnings (Off, "is not referenced");
 function Access_Def_P_Is_Access_Type
  (Node : Bare_Access_Def
      ; Origin : Bare_Ada_Node
  )
   return Boolean
is
   Self : Bare_Access_Def := Bare_Access_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Access_Def_P_Is_Access_Type;
--# end
--# property-start AccessDef.predefined_operators /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8154
pragma Warnings (Off, "is not referenced");
 function Access_Def_P_Predefined_Operators
  (Node : Bare_Access_Def
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Access_Def := Bare_Access_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_666 with Inline_Always;
      Fld : Internal_Env_Assoc_Array_Access;
Let_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_666 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_666;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Access_Def_Access_Def_P_Predefined_Operators,
            Items    => new Mmz_Key_Array (1 ..  1))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Access_Def,
                                As_Bare_Access_Def => Self);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Env_Assoc_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 11033 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11032 '<FieldAccess for predefined_equality_operators at ast.py:8155>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8155
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDef.predefined_equality_operators
Fld := Libadalang.Implementation.Type_Def_P_Predefined_Equality_Operators (Node => Self);
--# end
--# expr-done 11032
Let_Result := Fld; Inc_Ref (Let_Result);
--# end
--# expr-done 11033
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_666;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Env_Assoc_Array_Access,
                        As_Internal_Env_Assoc_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_666;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Access_Def_P_Predefined_Operators;
--# end
   --
   --  Primitives for Bare_Access_To_Subp_Def
   --
      procedure Access_To_Subp_Def_Pre_Env_Actions
        (Self            : Bare_Access_To_Subp_Def;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      if Add_To_Env_Only then
         return;
      end if;
      declare
         No_Parent         : constant Boolean :=
            False;
         Transitive_Parent : constant Boolean :=
            Libadalang.Implementation.Env_Trans_Parent_16 (Node => 
Self);
         Names             : Symbol_Type_Array_Access :=
            null;
      begin
         Add_Env (Self, State, No_Parent, Transitive_Parent, Names);
      end;
      end;
      procedure Initialize_Fields_For_Access_To_Subp_Def
        (Self : Bare_Access_To_Subp_Def
         ; Access_Def_F_Has_Not_Null : Bare_Not_Null
         ; Access_To_Subp_Def_F_Has_Protected : Bare_Protected_Node
         ; Access_To_Subp_Def_F_Subp_Spec : Bare_Subp_Spec
        ) is
      begin
            Initialize_Fields_For_Access_Def
              (Self, Access_Def_F_Has_Not_Null);
            Self.Access_To_Subp_Def_F_Has_Protected := Access_To_Subp_Def_F_Has_Protected;
            Self.Access_To_Subp_Def_F_Subp_Spec := Access_To_Subp_Def_F_Subp_Spec;
      end Initialize_Fields_For_Access_To_Subp_Def;
   function Access_To_Subp_Def_F_Has_Protected
     (Node : Bare_Access_To_Subp_Def) return Bare_Protected_Node
   is
   begin
         return Node.Access_To_Subp_Def_F_Has_Protected;
   end;
   function Access_To_Subp_Def_F_Subp_Spec
     (Node : Bare_Access_To_Subp_Def) return Bare_Subp_Spec
   is
   begin
         return Node.Access_To_Subp_Def_F_Subp_Spec;
   end;
--# property-start AccessToSubpDef.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8165
pragma Warnings (Off, "is not referenced");
 function Access_To_Subp_Def_P_Xref_Equation
  (Node : Bare_Access_To_Subp_Def
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Access_To_Subp_Def := Bare_Access_To_Subp_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_667 with Inline_Always;
      True_Rel : Logic_Equation;
            procedure Finalizer_Scope_667 is
            begin
                     Dec_Ref (True_Rel);
            end Finalizer_Scope_667;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8165") else null)); 
      Property_Result := True_Rel;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_667;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_667;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Access_To_Subp_Def_P_Xref_Equation;
--# end
--# property-start AccessToSubpDef.accessed_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8167
pragma Warnings (Off, "is not referenced");
 function Access_To_Subp_Def_P_Accessed_Type
  (Node : Bare_Access_To_Subp_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Access_To_Subp_Def := Bare_Access_To_Subp_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Access_To_Subp_Def :=
     Internal_Entity_Access_To_Subp_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Subp_Spec;
Fld_1 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11035 '<FieldAccess for return_type at ast.py:8167>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8167
--# expr-start 11034 '<FieldAccess for subp_spec at ast.py:8167>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8167
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Subp_Spec (Node => Ent.Node.Access_To_Subp_Def_F_Subp_Spec, Info => Ent.Info);
--# expr-done 11034
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BaseSubpSpec.return_type
Fld_1 := Libadalang.Implementation.Base_Subp_Spec_P_Return_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11035
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Access_To_Subp_Def_P_Accessed_Type;
--# end
--# property-start AccessToSubpDef._env_trans_parent_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8171
pragma Warnings (Off, "is not referenced");
 function Env_Trans_Parent_16
  (Node : Bare_Access_To_Subp_Def
  )
   return Boolean
is
   Self : Bare_Access_To_Subp_Def := Bare_Access_To_Subp_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Env_Trans_Parent_16;
--# end
   --
   --  Primitives for Bare_Base_Type_Access_Def
   --
      procedure Initialize_Fields_For_Base_Type_Access_Def
        (Self : Bare_Base_Type_Access_Def
         ; Access_Def_F_Has_Not_Null : Bare_Not_Null
        ) is
      begin
            Initialize_Fields_For_Access_Def
              (Self, Access_Def_F_Has_Not_Null);
      end Initialize_Fields_For_Base_Type_Access_Def;
   --
   --  Primitives for Bare_Anonymous_Type_Access_Def
   --
      procedure Initialize_Fields_For_Anonymous_Type_Access_Def
        (Self : Bare_Anonymous_Type_Access_Def
         ; Access_Def_F_Has_Not_Null : Bare_Not_Null
         ; Anonymous_Type_Access_Def_F_Type_Decl : Bare_Base_Type_Decl
        ) is
      begin
            Initialize_Fields_For_Base_Type_Access_Def
              (Self, Access_Def_F_Has_Not_Null);
            Self.Anonymous_Type_Access_Def_F_Type_Decl := Anonymous_Type_Access_Def_F_Type_Decl;
      end Initialize_Fields_For_Anonymous_Type_Access_Def;
   function Anonymous_Type_Access_Def_F_Type_Decl
     (Node : Bare_Anonymous_Type_Access_Def) return Bare_Base_Type_Decl
   is
   begin
         return Node.Anonymous_Type_Access_Def_F_Type_Decl;
   end;
--# property-start AnonymousTypeAccessDef.accessed_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8202
pragma Warnings (Off, "is not referenced");
 function Anonymous_Type_Access_Def_P_Accessed_Type
  (Node : Bare_Anonymous_Type_Access_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Anonymous_Type_Access_Def := Bare_Anonymous_Type_Access_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Anonymous_Type_Access_Def :=
     Internal_Entity_Anonymous_Type_Access_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11036 '<FieldAccess for type_decl at ast.py:8202>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8202
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Base_Type_Decl (Node => Ent.Node.Anonymous_Type_Access_Def_F_Type_Decl, Info => Ent.Info);
--# expr-done 11036
      Property_Result := Fld;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Anonymous_Type_Access_Def_P_Accessed_Type;
--# end
   --
   --  Primitives for Bare_Type_Access_Def
   --
      procedure Initialize_Fields_For_Type_Access_Def
        (Self : Bare_Type_Access_Def
         ; Access_Def_F_Has_Not_Null : Bare_Not_Null
         ; Type_Access_Def_F_Has_All : Bare_All_Node
         ; Type_Access_Def_F_Has_Constant : Bare_Constant_Node
         ; Type_Access_Def_F_Subtype_Indication : Bare_Subtype_Indication
        ) is
      begin
            Initialize_Fields_For_Base_Type_Access_Def
              (Self, Access_Def_F_Has_Not_Null);
            Self.Type_Access_Def_F_Has_All := Type_Access_Def_F_Has_All;
            Self.Type_Access_Def_F_Has_Constant := Type_Access_Def_F_Has_Constant;
            Self.Type_Access_Def_F_Subtype_Indication := Type_Access_Def_F_Subtype_Indication;
      end Initialize_Fields_For_Type_Access_Def;
   function Type_Access_Def_F_Has_All
     (Node : Bare_Type_Access_Def) return Bare_All_Node
   is
   begin
         return Node.Type_Access_Def_F_Has_All;
   end;
   function Type_Access_Def_F_Has_Constant
     (Node : Bare_Type_Access_Def) return Bare_Constant_Node
   is
   begin
         return Node.Type_Access_Def_F_Has_Constant;
   end;
   function Type_Access_Def_F_Subtype_Indication
     (Node : Bare_Type_Access_Def) return Bare_Subtype_Indication
   is
   begin
         return Node.Type_Access_Def_F_Subtype_Indication;
   end;
--# property-start TypeAccessDef.accessed_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8190
pragma Warnings (Off, "is not referenced");
 function Type_Access_Def_P_Accessed_Type
  (Node : Bare_Type_Access_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Type_Access_Def := Bare_Type_Access_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Access_Def :=
     Internal_Entity_Type_Access_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Subtype_Indication;
Fld_1 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11038 '<FieldAccess for designated_type at ast.py:8190>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8190
--# expr-start 11037 '<FieldAccess for subtype_indication at ast.py:8190>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8190
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Subtype_Indication (Node => Ent.Node.Type_Access_Def_F_Subtype_Indication, Info => Ent.Info);
--# expr-done 11037
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11038
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Access_Def_P_Accessed_Type;
--# end
--# property-start TypeAccessDef.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8191
pragma Warnings (Off, "is not referenced");
 function Type_Access_Def_P_Xref_Equation
  (Node : Bare_Type_Access_Def
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Type_Access_Def := Bare_Type_Access_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Access_Def :=
     Internal_Entity_Type_Access_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_670 with Inline_Always;
      Fld : Internal_Entity_Subtype_Indication;
Fld_1 : Logic_Equation;
            procedure Finalizer_Scope_670 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_670;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11040 '<FieldAccess for xref_equation at ast.py:8191>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8191
--# expr-start 11039 '<FieldAccess for subtype_indication at ast.py:8191>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8191
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Subtype_Indication (Node => Ent.Node.Type_Access_Def_F_Subtype_Indication, Info => Ent.Info);
--# expr-done 11039
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_1 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11040
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_670;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_670;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Access_Def_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Array_Type_Def
   --
      procedure Initialize_Fields_For_Array_Type_Def
        (Self : Bare_Array_Type_Def
         ; Array_Type_Def_F_Indices : Bare_Array_Indices
         ; Array_Type_Def_F_Component_Type : Bare_Component_Def
        ) is
      begin
            Self.Array_Type_Def_F_Indices := Array_Type_Def_F_Indices;
            Self.Array_Type_Def_F_Component_Type := Array_Type_Def_F_Component_Type;
      end Initialize_Fields_For_Array_Type_Def;
   function Array_Type_Def_F_Indices
     (Node : Bare_Array_Type_Def) return Bare_Array_Indices
   is
   begin
         return Node.Array_Type_Def_F_Indices;
   end;
   function Array_Type_Def_F_Component_Type
     (Node : Bare_Array_Type_Def) return Bare_Component_Def
   is
   begin
         return Node.Array_Type_Def_F_Component_Type;
   end;
--# property-start ArrayTypeDef.comp_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7818
pragma Warnings (Off, "is not referenced");
 function Array_Type_Def_P_Comp_Type
  (Node : Bare_Array_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Array_Type_Def := Bare_Array_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Array_Type_Def :=
     Internal_Entity_Array_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Component_Def;
Fld_1 : Internal_Entity_Type_Expr;
Fld_2 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11044 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11043 '<FieldAccess for designated_type at ast.py:7820>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7820
--# expr-start 11042 '<FieldAccess for type_expr at ast.py:7820>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7820
--# expr-start 11041 '<FieldAccess for component_type at ast.py:7820>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7820
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Component_Def (Node => Ent.Node.Array_Type_Def_F_Component_Type, Info => Ent.Info);
--# expr-done 11041
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Type_Expr (Node => Fld.Node.Component_Def_F_Type_Expr, Info => Fld.Info);
--# expr-done 11042
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Fld_1.Node, Origin => Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 11043
Let_Result := Fld_2; 
--# end
--# expr-done 11044
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Array_Type_Def_P_Comp_Type;
--# end
--# property-start ArrayTypeDef.index_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7823
pragma Warnings (Off, "is not referenced");
 function Array_Type_Def_P_Index_Type
  (Node : Bare_Array_Type_Def
      ; Dim : Integer
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Array_Type_Def := Bare_Array_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Array_Type_Def :=
     Internal_Entity_Array_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind dim Dim
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Array_Indices;
Fld_1 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11047 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11046 '<FieldAccess for index_type at ast.py:7824>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7824
--# expr-start 11045 '<FieldAccess for indices at ast.py:7824>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7824
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Array_Indices (Node => Ent.Node.Array_Type_Def_F_Indices, Info => Ent.Info);
--# expr-done 11045
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]ArrayIndices.index_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Array_Indices_P_Index_Type (Node => Fld.Node, Dim => Dim, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11046
Let_Result := Fld_1; 
--# end
--# expr-done 11047
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Array_Type_Def_P_Index_Type;
--# end
--# property-start ArrayTypeDef.array_ndims /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7826
pragma Warnings (Off, "is not referenced");
 function Array_Type_Def_P_Array_Ndims
  (Node : Bare_Array_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Integer
is
   Self : Bare_Array_Type_Def := Bare_Array_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Integer;
      Fld : Bare_Array_Indices;
Fld_1 : Integer;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11049 '<FieldAccess for ndims at ast.py:7826>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7826
--# expr-start 11048 '<FieldAccess for indices at ast.py:7826>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7826
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Array_Type_Def_F_Indices;
--# expr-done 11048
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]ArrayIndices.ndims'
Fld_1 := Libadalang.Implementation.Dispatcher_Array_Indices_P_Ndims (Node => Fld);
--# end
--# expr-done 11049
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Array_Type_Def_P_Array_Ndims;
--# end
--# property-start ArrayTypeDef.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7829
pragma Warnings (Off, "is not referenced");
 function Array_Type_Def_P_Xref_Equation
  (Node : Bare_Array_Type_Def
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Array_Type_Def := Bare_Array_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Array_Type_Def :=
     Internal_Entity_Array_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_602 with Inline_Always;
      Fld : Internal_Entity_Array_Indices;
Fld_1 : Logic_Equation;
Fld_2 : Internal_Entity_Component_Def;
Fld_3 : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_602 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Fld_3);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_602;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11055 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11054 '<And at ast.py:7830>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7830
--# expr-start 11051 '<FieldAccess for sub_equation at ast.py:7831>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7831
--# expr-start 11050 '<FieldAccess for indices at ast.py:7831>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7831
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Array_Indices (Node => Ent.Node.Array_Type_Def_F_Indices, Info => Ent.Info);
--# expr-done 11050
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11051
--# expr-start 11053 '<FieldAccess for sub_equation at ast.py:7832>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7832
--# expr-start 11052 '<FieldAccess for component_type at ast.py:7832>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7832
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Component_Def (Node => Ent.Node.Array_Type_Def_F_Component_Type, Info => Ent.Info);
--# expr-done 11052
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_2.Node, Env => Env, Origin => Origin, E_Info => Fld_2.Info);
--# end
--# expr-done 11053
And_Pred := Create_And (Fld_1, Fld_3, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7830") else null)); 
--# expr-done 11054
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 11055
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_602;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_602;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Array_Type_Def_P_Xref_Equation;
--# end
--# property-start ArrayTypeDef.predefined_operators /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7836
pragma Warnings (Off, "is not referenced");
 function Array_Type_Def_P_Predefined_Operators
  (Node : Bare_Array_Type_Def
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Array_Type_Def := Bare_Array_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_603 with Inline_Always;
      Block_Var : Bare_Type_Decl;
Block_Var_1 : Bare_Base_Type_Decl;
Block_Var_2 : Bare_Type_Expr;
Fld : Bare_Ada_Node;
Cast_Result : Bare_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Bare_Base_Type_Decl;
Fld_3 : Bare_Component_Def;
Fld_4 : Bare_Type_Expr;
Sym : Symbol_Type;
Cast_Expr : Bare_Type_Decl;
Cast_Result_1 : Bare_Base_Type_Decl;
Cast_Expr_1 : Bare_Type_Decl;
Cast_Result_2 : Bare_Base_Type_Decl;
Fld_5 : Internal_Env_Assoc := No_Env_Assoc;
Sym_1 : Symbol_Type;
Cast_Expr_2 : Bare_Type_Decl;
Cast_Result_3 : Bare_Base_Type_Decl;
Cast_Expr_3 : Bare_Type_Decl;
Cast_Result_4 : Bare_Base_Type_Decl;
Fld_6 : Internal_Env_Assoc := No_Env_Assoc;
Sym_2 : Symbol_Type;
Cast_Expr_4 : Bare_Type_Decl;
Cast_Result_5 : Bare_Base_Type_Decl;
Cast_Expr_5 : Bare_Type_Decl;
Cast_Result_6 : Bare_Base_Type_Decl;
Fld_7 : Internal_Env_Assoc := No_Env_Assoc;
Sym_3 : Symbol_Type;
Cast_Expr_6 : Bare_Type_Decl;
Cast_Result_7 : Bare_Base_Type_Decl;
Cast_Expr_7 : Bare_Type_Decl;
Cast_Result_8 : Bare_Base_Type_Decl;
Fld_8 : Internal_Env_Assoc := No_Env_Assoc;
Sym_4 : Symbol_Type;
Cast_Expr_8 : Bare_Type_Decl;
Cast_Result_9 : Bare_Base_Type_Decl;
Cast_Expr_9 : Bare_Type_Decl;
Cast_Result_10 : Bare_Base_Type_Decl;
Fld_9 : Internal_Env_Assoc := No_Env_Assoc;
Sym_5 : Symbol_Type;
Cast_Expr_10 : Bare_Type_Decl;
Cast_Result_11 : Bare_Base_Type_Decl;
Cast_Expr_11 : Bare_Type_Decl;
Cast_Result_12 : Bare_Base_Type_Decl;
Fld_10 : Internal_Env_Assoc := No_Env_Assoc;
Sym_6 : Symbol_Type;
Cast_Expr_12 : Bare_Type_Decl;
Cast_Result_13 : Bare_Base_Type_Decl;
Cast_Expr_13 : Bare_Type_Decl;
Cast_Result_14 : Bare_Base_Type_Decl;
Cast_Expr_14 : Bare_Type_Decl;
Cast_Result_15 : Bare_Base_Type_Decl;
Fld_11 : Internal_Env_Assoc := No_Env_Assoc;
Sym_7 : Symbol_Type;
Cast_Expr_15 : Bare_Type_Decl;
Cast_Result_16 : Bare_Base_Type_Decl;
Cast_Expr_16 : Bare_Type_Decl;
Cast_Result_17 : Bare_Base_Type_Decl;
Cast_Expr_17 : Bare_Type_Decl;
Cast_Result_18 : Bare_Base_Type_Decl;
Fld_12 : Internal_Env_Assoc := No_Env_Assoc;
Sym_8 : Symbol_Type;
Cast_Expr_18 : Bare_Type_Decl;
Cast_Result_19 : Bare_Base_Type_Decl;
Cast_Expr_19 : Bare_Type_Decl;
Cast_Result_20 : Bare_Base_Type_Decl;
Cast_Expr_20 : Bare_Type_Decl;
Cast_Result_21 : Bare_Base_Type_Decl;
Fld_13 : Internal_Env_Assoc := No_Env_Assoc;
Sym_9 : Symbol_Type;
Cast_Expr_21 : Bare_Type_Decl;
Cast_Result_22 : Bare_Base_Type_Decl;
Cast_Expr_22 : Bare_Type_Decl;
Cast_Result_23 : Bare_Base_Type_Decl;
Fld_14 : Internal_Env_Assoc := No_Env_Assoc;
Sym_10 : Symbol_Type;
Cast_Expr_23 : Bare_Type_Decl;
Cast_Result_24 : Bare_Base_Type_Decl;
Fld_15 : Internal_Env_Assoc := No_Env_Assoc;
Sym_11 : Symbol_Type;
Cast_Expr_24 : Bare_Type_Decl;
Cast_Result_25 : Bare_Base_Type_Decl;
Cast_Expr_25 : Bare_Type_Decl;
Cast_Result_26 : Bare_Base_Type_Decl;
Fld_16 : Internal_Env_Assoc := No_Env_Assoc;
Sym_12 : Symbol_Type;
Cast_Expr_26 : Bare_Type_Decl;
Cast_Result_27 : Bare_Base_Type_Decl;
Cast_Expr_27 : Bare_Type_Decl;
Cast_Result_28 : Bare_Base_Type_Decl;
Fld_17 : Internal_Env_Assoc := No_Env_Assoc;
Sym_13 : Symbol_Type;
Cast_Expr_28 : Bare_Type_Decl;
Cast_Result_29 : Bare_Base_Type_Decl;
Cast_Expr_29 : Bare_Type_Decl;
Cast_Result_30 : Bare_Base_Type_Decl;
Cast_Expr_30 : Bare_Type_Decl;
Cast_Result_31 : Bare_Base_Type_Decl;
Fld_18 : Internal_Env_Assoc := No_Env_Assoc;
Array_Lit : Internal_Env_Assoc_Array_Access;
Let_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_603 is
            begin
                     Dec_Ref (Fld_5);
                     Dec_Ref (Fld_6);
                     Dec_Ref (Fld_7);
                     Dec_Ref (Fld_8);
                     Dec_Ref (Fld_9);
                     Dec_Ref (Fld_10);
                     Dec_Ref (Fld_11);
                     Dec_Ref (Fld_12);
                     Dec_Ref (Fld_13);
                     Dec_Ref (Fld_14);
                     Dec_Ref (Fld_15);
                     Dec_Ref (Fld_16);
                     Dec_Ref (Fld_17);
                     Dec_Ref (Fld_18);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_603;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Array_Type_Def_Array_Type_Def_P_Predefined_Operators,
            Items    => new Mmz_Key_Array (1 ..  1))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Array_Type_Def,
                                As_Bare_Array_Type_Def => Self);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Env_Assoc_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 11091 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11057 '<Cast to TypeDecl at ast.py:7837>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7837
--# expr-start 11056 '<parent at ast.py:7837>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7837
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 11056
   if Fld = null
      or else Fld.Kind in Ada_Type_Decl_Range
   then
      Cast_Result := Fld;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 11057
Block_Var := Cast_Result; 
--# bind self_type Block_Var
--# expr-start 11059 '<FieldAccess for node at ast.py:7838>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7838
--# expr-start 11058 '<FieldAccess for bool_type at ast.py:7838>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7838
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_1 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 11058
Fld_2 := Fld_1.Node;
--# expr-done 11059
Block_Var_1 := Fld_2; 
--# bind bool_type Block_Var_1
--# expr-start 11061 '<FieldAccess for type_expr at ast.py:7839>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7839
--# expr-start 11060 '<FieldAccess for component_type at ast.py:7839>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7839
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Self.Array_Type_Def_F_Component_Type;
--# expr-done 11060
      if Fld_3 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Fld_3.Component_Def_F_Type_Expr;
--# expr-done 11061
Block_Var_2 := Fld_4; 
--# bind comp_type_expr Block_Var_2
--# expr-start 11090 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 11063 '<FieldAccess for create_binop_assoc at ast.py:7850>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7850
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11062 '<Symbol "<" at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_8); 
--# expr-done 11062
Cast_Expr := Block_Var; 
      Cast_Result_1 := Cast_Expr;
Cast_Expr_1 := Block_Var; 
      Cast_Result_2 := Cast_Expr_1;
--# property-call-start AdaNode.create_binop_assoc
Fld_5 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym, Lhs => Cast_Result_1, Rhs => Cast_Result_2, Ret => Block_Var_1);
--# end
--# expr-done 11063
--# expr-start 11065 '<FieldAccess for create_binop_assoc at ast.py:7851>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7851
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11064 '<Symbol "<=" at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_9); 
--# expr-done 11064
Cast_Expr_2 := Block_Var; 
      Cast_Result_3 := Cast_Expr_2;
Cast_Expr_3 := Block_Var; 
      Cast_Result_4 := Cast_Expr_3;
--# property-call-start AdaNode.create_binop_assoc
Fld_6 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_1, Lhs => Cast_Result_3, Rhs => Cast_Result_4, Ret => Block_Var_1);
--# end
--# expr-done 11065
--# expr-start 11067 '<FieldAccess for create_binop_assoc at ast.py:7852>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7852
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11066 '<Symbol "=" at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_10); 
--# expr-done 11066
Cast_Expr_4 := Block_Var; 
      Cast_Result_5 := Cast_Expr_4;
Cast_Expr_5 := Block_Var; 
      Cast_Result_6 := Cast_Expr_5;
--# property-call-start AdaNode.create_binop_assoc
Fld_7 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_2, Lhs => Cast_Result_5, Rhs => Cast_Result_6, Ret => Block_Var_1);
--# end
--# expr-done 11067
--# expr-start 11069 '<FieldAccess for create_binop_assoc at ast.py:7853>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7853
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11068 '<Symbol "/=" at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_7); 
--# expr-done 11068
Cast_Expr_6 := Block_Var; 
      Cast_Result_7 := Cast_Expr_6;
Cast_Expr_7 := Block_Var; 
      Cast_Result_8 := Cast_Expr_7;
--# property-call-start AdaNode.create_binop_assoc
Fld_8 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_3, Lhs => Cast_Result_7, Rhs => Cast_Result_8, Ret => Block_Var_1);
--# end
--# expr-done 11069
--# expr-start 11071 '<FieldAccess for create_binop_assoc at ast.py:7854>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7854
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11070 '<Symbol ">" at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_11); 
--# expr-done 11070
Cast_Expr_8 := Block_Var; 
      Cast_Result_9 := Cast_Expr_8;
Cast_Expr_9 := Block_Var; 
      Cast_Result_10 := Cast_Expr_9;
--# property-call-start AdaNode.create_binop_assoc
Fld_9 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_4, Lhs => Cast_Result_9, Rhs => Cast_Result_10, Ret => Block_Var_1);
--# end
--# expr-done 11071
--# expr-start 11073 '<FieldAccess for create_binop_assoc at ast.py:7855>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7855
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11072 '<Symbol ">=" at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_12); 
--# expr-done 11072
Cast_Expr_10 := Block_Var; 
      Cast_Result_11 := Cast_Expr_10;
Cast_Expr_11 := Block_Var; 
      Cast_Result_12 := Cast_Expr_11;
--# property-call-start AdaNode.create_binop_assoc
Fld_10 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_5, Lhs => Cast_Result_11, Rhs => Cast_Result_12, Ret => Block_Var_1);
--# end
--# expr-done 11073
--# expr-start 11075 '<FieldAccess for create_binop_assoc at ast.py:7856>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7856
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11074 '<Symbol "and" at ???>' Sym_6 None
Sym_6 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_And); 
--# expr-done 11074
Cast_Expr_12 := Block_Var; 
      Cast_Result_13 := Cast_Expr_12;
Cast_Expr_13 := Block_Var; 
      Cast_Result_14 := Cast_Expr_13;
Cast_Expr_14 := Block_Var; 
      Cast_Result_15 := Cast_Expr_14;
--# property-call-start AdaNode.create_binop_assoc
Fld_11 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_6, Lhs => Cast_Result_13, Rhs => Cast_Result_14, Ret => Cast_Result_15);
--# end
--# expr-done 11075
--# expr-start 11077 '<FieldAccess for create_binop_assoc at ast.py:7857>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7857
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11076 '<Symbol "or" at ???>' Sym_7 None
Sym_7 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Or); 
--# expr-done 11076
Cast_Expr_15 := Block_Var; 
      Cast_Result_16 := Cast_Expr_15;
Cast_Expr_16 := Block_Var; 
      Cast_Result_17 := Cast_Expr_16;
Cast_Expr_17 := Block_Var; 
      Cast_Result_18 := Cast_Expr_17;
--# property-call-start AdaNode.create_binop_assoc
Fld_12 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_7, Lhs => Cast_Result_16, Rhs => Cast_Result_17, Ret => Cast_Result_18);
--# end
--# expr-done 11077
--# expr-start 11079 '<FieldAccess for create_binop_assoc at ast.py:7858>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7858
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11078 '<Symbol "xor" at ???>' Sym_8 None
Sym_8 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Xor); 
--# expr-done 11078
Cast_Expr_18 := Block_Var; 
      Cast_Result_19 := Cast_Expr_18;
Cast_Expr_19 := Block_Var; 
      Cast_Result_20 := Cast_Expr_19;
Cast_Expr_20 := Block_Var; 
      Cast_Result_21 := Cast_Expr_20;
--# property-call-start AdaNode.create_binop_assoc
Fld_13 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_8, Lhs => Cast_Result_19, Rhs => Cast_Result_20, Ret => Cast_Result_21);
--# end
--# expr-done 11079
--# expr-start 11081 '<FieldAccess for create_unop_assoc at ast.py:7859>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7859
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11080 '<Symbol "not" at ???>' Sym_9 None
Sym_9 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Not); 
--# expr-done 11080
Cast_Expr_21 := Block_Var; 
      Cast_Result_22 := Cast_Expr_21;
Cast_Expr_22 := Block_Var; 
      Cast_Result_23 := Cast_Expr_22;
--# property-call-start AdaNode.create_unop_assoc
Fld_14 := Libadalang.Implementation.Ada_Node_P_Create_Unop_Assoc (Node => Self, Op => Sym_9, Rhs => Cast_Result_22, Ret => Cast_Result_23);
--# end
--# expr-done 11081
--# expr-start 11083 '<FieldAccess for create_binop_assoc_l_r_expr at ast.py:7862>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7862
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11082 '<Symbol "&" at ???>' Sym_10 None
Sym_10 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym); 
--# expr-done 11082
Cast_Expr_23 := Block_Var; 
      Cast_Result_24 := Cast_Expr_23;
--# property-call-start AdaNode.create_binop_assoc_l_r_expr
Fld_15 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc_L_R_Expr (Node => Self, Op => Sym_10, Lhs => Block_Var_2, Rhs => Block_Var_2, Ret => Cast_Result_24);
--# end
--# expr-done 11083
--# expr-start 11085 '<FieldAccess for create_binop_assoc_l_expr at ast.py:7865>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7865
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11084 '<Symbol "&" at ???>' Sym_11 None
Sym_11 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym); 
--# expr-done 11084
Cast_Expr_24 := Block_Var; 
      Cast_Result_25 := Cast_Expr_24;
Cast_Expr_25 := Block_Var; 
      Cast_Result_26 := Cast_Expr_25;
--# property-call-start AdaNode.create_binop_assoc_l_expr
Fld_16 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc_L_Expr (Node => Self, Op => Sym_11, Lhs => Block_Var_2, Rhs => Cast_Result_25, Ret => Cast_Result_26);
--# end
--# expr-done 11085
--# expr-start 11087 '<FieldAccess for create_binop_assoc_r_expr at ast.py:7868>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7868
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11086 '<Symbol "&" at ???>' Sym_12 None
Sym_12 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym); 
--# expr-done 11086
Cast_Expr_26 := Block_Var; 
      Cast_Result_27 := Cast_Expr_26;
Cast_Expr_27 := Block_Var; 
      Cast_Result_28 := Cast_Expr_27;
--# property-call-start AdaNode.create_binop_assoc_r_expr
Fld_17 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc_R_Expr (Node => Self, Op => Sym_12, Lhs => Cast_Result_27, Rhs => Block_Var_2, Ret => Cast_Result_28);
--# end
--# expr-done 11087
--# expr-start 11089 '<FieldAccess for create_binop_assoc at ast.py:7871>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7871
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11088 '<Symbol "&" at ???>' Sym_13 None
Sym_13 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym); 
--# expr-done 11088
Cast_Expr_28 := Block_Var; 
      Cast_Result_29 := Cast_Expr_28;
Cast_Expr_29 := Block_Var; 
      Cast_Result_30 := Cast_Expr_29;
Cast_Expr_30 := Block_Var; 
      Cast_Result_31 := Cast_Expr_30;
--# property-call-start AdaNode.create_binop_assoc
Fld_18 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_13, Lhs => Cast_Result_29, Rhs => Cast_Result_30, Ret => Cast_Result_31);
--# end
--# expr-done 11089
Array_Lit := Create_Internal_Env_Assoc_Array (Internal_Internal_Env_Assoc_Array'(1 => Fld_5, 2 => Fld_6, 3 => Fld_7, 4 => Fld_8, 5 => Fld_9, 6 => Fld_10, 7 => Fld_11, 8 => Fld_12, 9 => Fld_13, 10 => Fld_14, 11 => Fld_15, 12 => Fld_16, 13 => Fld_17, 14 => Fld_18)); 
--# expr-done 11090
Let_Result := Array_Lit; Inc_Ref (Let_Result);
--# end
--# expr-done 11091
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_603;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Env_Assoc_Array_Access,
                        As_Internal_Env_Assoc_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_603;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Array_Type_Def_P_Predefined_Operators;
--# end
--# property-start ArrayTypeDef.is_static /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7876
pragma Warnings (Off, "is not referenced");
 function Array_Type_Def_P_Is_Static
  (Node : Bare_Array_Type_Def
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Array_Type_Def := Bare_Array_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Array_Type_Def :=
     Internal_Entity_Array_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Fld : Internal_Entity_Array_Indices;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11093 '<FieldAccess for is_static at ast.py:7876>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7876
--# expr-start 11092 '<FieldAccess for indices at ast.py:7876>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7876
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Array_Indices (Node => Ent.Node.Array_Type_Def_F_Indices, Info => Ent.Info);
--# expr-done 11092
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]ArrayIndices.is_static'
Fld_1 := Libadalang.Implementation.Dispatcher_Array_Indices_P_Is_Static (Node => Fld.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld.Info);
--# end
--# expr-done 11093
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Array_Type_Def_P_Is_Static;
--# end
--# property-start ArrayTypeDef.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7878
pragma Warnings (Off, "is not referenced");
 function Array_Type_Def_P_Xref_Entry_Point
  (Node : Bare_Array_Type_Def
  )
   return Boolean
is
   Self : Bare_Array_Type_Def := Bare_Array_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Array_Type_Def_P_Xref_Entry_Point;
--# end
   --
   --  Primitives for Bare_Derived_Type_Def
   --
      procedure Initialize_Fields_For_Derived_Type_Def
        (Self : Bare_Derived_Type_Def
         ; Derived_Type_Def_F_Has_Abstract : Bare_Abstract_Node
         ; Derived_Type_Def_F_Has_Limited : Bare_Limited_Node
         ; Derived_Type_Def_F_Has_Synchronized : Bare_Synchronized_Node
         ; Derived_Type_Def_F_Subtype_Indication : Bare_Subtype_Indication
         ; Derived_Type_Def_F_Interfaces : Bare_Parent_List
         ; Derived_Type_Def_F_Record_Extension : Bare_Base_Record_Def
         ; Derived_Type_Def_F_Has_With_Private : Bare_With_Private
        ) is
      begin
            Self.Derived_Type_Def_F_Has_Abstract := Derived_Type_Def_F_Has_Abstract;
            Self.Derived_Type_Def_F_Has_Limited := Derived_Type_Def_F_Has_Limited;
            Self.Derived_Type_Def_F_Has_Synchronized := Derived_Type_Def_F_Has_Synchronized;
            Self.Derived_Type_Def_F_Subtype_Indication := Derived_Type_Def_F_Subtype_Indication;
            Self.Derived_Type_Def_F_Interfaces := Derived_Type_Def_F_Interfaces;
            Self.Derived_Type_Def_F_Record_Extension := Derived_Type_Def_F_Record_Extension;
            Self.Derived_Type_Def_F_Has_With_Private := Derived_Type_Def_F_Has_With_Private;
      end Initialize_Fields_For_Derived_Type_Def;
   function Derived_Type_Def_F_Has_Abstract
     (Node : Bare_Derived_Type_Def) return Bare_Abstract_Node
   is
   begin
         return Node.Derived_Type_Def_F_Has_Abstract;
   end;
   function Derived_Type_Def_F_Has_Limited
     (Node : Bare_Derived_Type_Def) return Bare_Limited_Node
   is
   begin
         return Node.Derived_Type_Def_F_Has_Limited;
   end;
   function Derived_Type_Def_F_Has_Synchronized
     (Node : Bare_Derived_Type_Def) return Bare_Synchronized_Node
   is
   begin
         return Node.Derived_Type_Def_F_Has_Synchronized;
   end;
   function Derived_Type_Def_F_Subtype_Indication
     (Node : Bare_Derived_Type_Def) return Bare_Subtype_Indication
   is
   begin
         return Node.Derived_Type_Def_F_Subtype_Indication;
   end;
   function Derived_Type_Def_F_Interfaces
     (Node : Bare_Derived_Type_Def) return Bare_Parent_List
   is
   begin
         return Node.Derived_Type_Def_F_Interfaces;
   end;
   function Derived_Type_Def_F_Record_Extension
     (Node : Bare_Derived_Type_Def) return Bare_Base_Record_Def
   is
   begin
         return Node.Derived_Type_Def_F_Record_Extension;
   end;
   function Derived_Type_Def_F_Has_With_Private
     (Node : Bare_Derived_Type_Def) return Bare_With_Private
   is
   begin
         return Node.Derived_Type_Def_F_Has_With_Private;
   end;
--# property-start DerivedTypeDef.array_ndims /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7519
pragma Warnings (Off, "is not referenced");
 function Derived_Type_Def_P_Array_Ndims
  (Node : Bare_Derived_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Integer
is
   Self : Bare_Derived_Type_Def := Bare_Derived_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Derived_Type_Def :=
     Internal_Entity_Derived_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Integer;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Integer;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11095 '<FieldAccess for array_ndims at ast.py:7519>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7519
--# expr-start 11094 '<FieldAccess for base_type at ast.py:7519>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7519
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 11094
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.array_ndims'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Array_Ndims (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11095
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Derived_Type_Def_P_Array_Ndims;
--# end
--# property-start DerivedTypeDef.base_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7522
pragma Warnings (Off, "is not referenced");
 function Derived_Type_Def_P_Base_Type
  (Node : Bare_Derived_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Derived_Type_Def := Bare_Derived_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Derived_Type_Def :=
     Internal_Entity_Derived_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_1 : Internal_Entity_Subtype_Indication;
Fld_2 : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11099 '<bind at ast.py:7523>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7523
--# expr-start 11096 '<FieldAccess for origin_node at ast.py:7523>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7523
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 11096
Bound_Origin := Fld; 
--# expr-start 11098 '<FieldAccess for designated_type at ast.py:7524>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7524
--# expr-start 11097 '<FieldAccess for subtype_indication at ast.py:7524>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7524
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Subtype_Indication (Node => Ent.Node.Derived_Type_Def_F_Subtype_Indication, Info => Ent.Info);
--# expr-done 11097
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Fld_1.Node, Origin => Bound_Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 11098
Dyn_Var_Bind_Result := Fld_2; 
--# expr-done 11099
      Property_Result := Dyn_Var_Bind_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Derived_Type_Def_P_Base_Type;
--# end
--# property-start DerivedTypeDef.base_interfaces /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7527
pragma Warnings (Off, "is not referenced");
 function Derived_Type_Def_P_Base_Interfaces
  (Node : Bare_Derived_Type_Def
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl_Array_Access
is
   Self : Bare_Derived_Type_Def := Bare_Derived_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Derived_Type_Def :=
     Internal_Entity_Derived_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_559 with Inline_Always;
      Fld : Internal_Entity_Parent_List;
Fld_1 : Bare_Parent_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld_3 : Internal_Entity_Base_Type_Decl;
Item_332 : Internal_Entity_Name;
Bare_Item_332 : Bare_Name;
Map_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_559 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_559;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11102 '<Map at ast.py:7528>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7528
   --# expr-start 11100 '<FieldAccess for interfaces at ast.py:7528>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7528
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Parent_List (Node => Ent.Node.Derived_Type_Def_F_Interfaces, Info => Ent.Info);
--# expr-done 11100
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Internal_Entity_Base_Type_Decl_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Internal_Entity_Base_Type_Decl_Array (0);
      else
      declare
         Collection : constant Bare_Parent_List := Fld_1;
      begin
         for Untyped_Item_332 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_332 := Untyped_Item_332; 
Is_Null := Bare_Item_332 = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_332); 
   If_Result := As_Entity;
end if;
                  Item_332 := If_Result; 
   --# scope-start
               --# bind i Item_332
      --# expr-start 11101 '<FieldAccess for name_designated_type at ast.py:7528>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7528
      if Item_332.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_3 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Item_332.Node, E_Info => Item_332.Info);
--# end
--# expr-done 11101
         declare
            Item_To_Append : constant Internal_Entity_Base_Type_Decl := Fld_3;
         begin
            Internal_Entity_Base_Type_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Base_Type_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Type_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Base_Type_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Base_Type_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Base_Type_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 11102
      Property_Result := Map_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_559;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_559;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Derived_Type_Def_P_Base_Interfaces;
--# end
--# property-start DerivedTypeDef.is_int_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7531
pragma Warnings (Off, "is not referenced");
 function Derived_Type_Def_P_Is_Int_Type
  (Node : Bare_Derived_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Derived_Type_Def := Bare_Derived_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Derived_Type_Def :=
     Internal_Entity_Derived_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11104 '<FieldAccess for is_int_type at ast.py:7531>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7531
--# expr-start 11103 '<FieldAccess for base_type at ast.py:7531>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7531
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 11103
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_int_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Int_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11104
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Derived_Type_Def_P_Is_Int_Type;
--# end
--# property-start DerivedTypeDef.is_access_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7532
pragma Warnings (Off, "is not referenced");
 function Derived_Type_Def_P_Is_Access_Type
  (Node : Bare_Derived_Type_Def
      ; Origin : Bare_Ada_Node
  )
   return Boolean
is
   Self : Bare_Derived_Type_Def := Bare_Derived_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
      As_Entity : Internal_Entity_Derived_Type_Def;
Is_Null : Boolean;
If_Result : Internal_Entity_Derived_Type_Def;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11107 '<FieldAccess for is_access_type at ast.py:7532>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7532
--# expr-start 11106 '<FieldAccess for base_type at ast.py:7532>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7532
--# expr-start 11105 '<as_bare_entity at ast.py:7532>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7532
Is_Null := Self = null; 
if Is_Null then
   If_Result := No_Entity_Derived_Type_Def;
else
As_Entity := (Info => No_Entity_Info, Node => Self); 
   If_Result := As_Entity;
end if;
--# expr-done 11105
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => If_Result.Node, Origin => Origin, E_Info => If_Result.Info);
--# end
--# expr-done 11106
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_access_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Access_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11107
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Derived_Type_Def_P_Is_Access_Type;
--# end
--# property-start DerivedTypeDef.is_char_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7533
pragma Warnings (Off, "is not referenced");
 function Derived_Type_Def_P_Is_Char_Type
  (Node : Bare_Derived_Type_Def
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Derived_Type_Def := Bare_Derived_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Derived_Type_Def :=
     Internal_Entity_Derived_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11109 '<FieldAccess for is_char_type at ast.py:7533>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7533
--# expr-start 11108 '<FieldAccess for base_type at ast.py:7533>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7533
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 11108
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_char_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Char_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11109
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Derived_Type_Def_P_Is_Char_Type;
--# end
--# property-start DerivedTypeDef.is_float_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7534
pragma Warnings (Off, "is not referenced");
 function Derived_Type_Def_P_Is_Float_Type
  (Node : Bare_Derived_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Derived_Type_Def := Bare_Derived_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Derived_Type_Def :=
     Internal_Entity_Derived_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11111 '<FieldAccess for is_float_type at ast.py:7534>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7534
--# expr-start 11110 '<FieldAccess for base_type at ast.py:7534>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7534
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 11110
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_float_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Float_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11111
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Derived_Type_Def_P_Is_Float_Type;
--# end
--# property-start DerivedTypeDef.is_fixed_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7535
pragma Warnings (Off, "is not referenced");
 function Derived_Type_Def_P_Is_Fixed_Point
  (Node : Bare_Derived_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Derived_Type_Def := Bare_Derived_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Derived_Type_Def :=
     Internal_Entity_Derived_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11113 '<FieldAccess for is_fixed_point at ast.py:7535>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7535
--# expr-start 11112 '<FieldAccess for base_type at ast.py:7535>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7535
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 11112
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_fixed_point'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Fixed_Point (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11113
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Derived_Type_Def_P_Is_Fixed_Point;
--# end
--# property-start DerivedTypeDef.accessed_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7536
pragma Warnings (Off, "is not referenced");
 function Derived_Type_Def_P_Accessed_Type
  (Node : Bare_Derived_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Derived_Type_Def := Bare_Derived_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Derived_Type_Def :=
     Internal_Entity_Derived_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11115 '<FieldAccess for accessed_type at ast.py:7536>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7536
--# expr-start 11114 '<FieldAccess for base_type at ast.py:7536>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7536
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 11114
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11115
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Derived_Type_Def_P_Accessed_Type;
--# end
--# property-start DerivedTypeDef.is_tagged_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7537
pragma Warnings (Off, "is not referenced");
 function Derived_Type_Def_P_Is_Tagged_Type
  (Node : Bare_Derived_Type_Def
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Derived_Type_Def := Bare_Derived_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Derived_Type_Def :=
     Internal_Entity_Derived_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Record_Def;
Is_Null : Boolean;
Not_Val : Boolean;
Fld_1 : Internal_Entity_With_Private;
Fld_2 : Boolean;
If_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 11118 '<Not at ast.py:7538>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7538
--# expr-start 11117 '<IsNull at ast.py:7538>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7538
--# expr-start 11116 '<FieldAccess for record_extension at ast.py:7538>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7538
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Base_Record_Def (Node => Ent.Node.Derived_Type_Def_F_Record_Extension, Info => Ent.Info);
--# expr-done 11116
Is_Null := Fld.Node = null; 
--# expr-done 11117
Not_Val := not (Is_Null); 
--# expr-done 11118
if Not_Val then
   If_Result := True;
else
   --# expr-start 11120 '<FieldAccess for as_bool at ast.py:7538>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7538
--# expr-start 11119 '<FieldAccess for has_with_private at ast.py:7538>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7538
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_With_Private (Node => Ent.Node.Derived_Type_Def_F_Has_With_Private, Info => Ent.Info);
--# expr-done 11119
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]WithPrivate.as_bool'
Fld_2 := Libadalang.Implementation.Dispatcher_With_Private_P_As_Bool (Node => Fld_1.Node);
--# end
--# expr-done 11120
   If_Result := Fld_2;
end if;
      Property_Result := If_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Derived_Type_Def_P_Is_Tagged_Type;
--# end
--# property-start DerivedTypeDef.is_enum_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7541
pragma Warnings (Off, "is not referenced");
 function Derived_Type_Def_P_Is_Enum_Type
  (Node : Bare_Derived_Type_Def
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Derived_Type_Def := Bare_Derived_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Derived_Type_Def :=
     Internal_Entity_Derived_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11122 '<FieldAccess for is_enum_type at ast.py:7541>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7541
--# expr-start 11121 '<FieldAccess for base_type at ast.py:7541>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7541
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeDef.base_type'
Fld := Libadalang.Implementation.Dispatcher_Type_Def_P_Base_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 11121
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.is_enum_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Is_Enum_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11122
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Derived_Type_Def_P_Is_Enum_Type;
--# end
--# property-start DerivedTypeDef.is_static /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7542
pragma Warnings (Off, "is not referenced");
 function Derived_Type_Def_P_Is_Static
  (Node : Bare_Derived_Type_Def
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Derived_Type_Def := Bare_Derived_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Derived_Type_Def :=
     Internal_Entity_Derived_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Fld : Internal_Entity_Subtype_Indication;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11124 '<FieldAccess for is_static_subtype at ast.py:7542>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7542
--# expr-start 11123 '<FieldAccess for subtype_indication at ast.py:7542>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7542
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Subtype_Indication (Node => Ent.Node.Derived_Type_Def_F_Subtype_Indication, Info => Ent.Info);
--# expr-done 11123
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SubtypeIndication.is_static_subtype
Fld_1 := Libadalang.Implementation.Subtype_Indication_P_Is_Static_Subtype (Node => Fld.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld.Info);
--# end
--# expr-done 11124
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Derived_Type_Def_P_Is_Static;
--# end
--# property-start DerivedTypeDef.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7545
pragma Warnings (Off, "is not referenced");
 function Derived_Type_Def_P_Xref_Equation
  (Node : Bare_Derived_Type_Def
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Derived_Type_Def := Bare_Derived_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Derived_Type_Def :=
     Internal_Entity_Derived_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_569 with Inline_Always;
            procedure Finalizer_Scope_3093 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_1 : Internal_Entity_Subtype_Indication;
Fld_2 : Logic_Equation;
Fld_3 : Internal_Entity_Parent_List;
Fld_4 : Bare_Parent_List;
Fld_5 : Internal_Entity_Info;
As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld_6 : Logic_Equation;
Item_333 : Internal_Entity_Name;
Bare_Item_333 : Bare_Name;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
And_Pred : Logic_Equation;
Dyn_Var_Bind_Result : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_569 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_2);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_569;
            procedure Finalizer_Scope_3093 is
            begin
                     Dec_Ref (Fld_6);
            end Finalizer_Scope_3093;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11134 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11133 '<bind at ast.py:7548>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7548
--# expr-start 11125 '<FieldAccess for children_env at ast.py:7548>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7548
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 11125
Bound_Env := Fld; Inc_Ref (Bound_Env);
--# expr-start 11132 '<And at ast.py:7549>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7549
--# expr-start 11127 '<FieldAccess for xref_equation at ast.py:7549>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7549
--# expr-start 11126 '<FieldAccess for subtype_indication at ast.py:7549>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7549
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Subtype_Indication (Node => Ent.Node.Derived_Type_Def_F_Subtype_Indication, Info => Ent.Info);
--# expr-done 11126
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_2 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Fld_1.Node, Env => Bound_Env, Origin => Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 11127
--# expr-start 11131 '<LogicAll at ast.py:7550>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7550
--# expr-start 11130 '<Map at ast.py:7550>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7550
   --# expr-start 11128 '<FieldAccess for interfaces at ast.py:7550>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7550
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Parent_List (Node => Ent.Node.Derived_Type_Def_F_Interfaces, Info => Ent.Info);
--# expr-done 11128
Fld_4 := Fld_3.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_4 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Parent_List := Fld_4;
      begin
         for Untyped_Item_333 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_333 := Untyped_Item_333; 
Is_Null := Bare_Item_333 = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
Fld_5 := Fld_3.Info;
As_Entity := (Info => Fld_5, Node => Bare_Item_333); 
   If_Result := As_Entity;
end if;
                  Item_333 := If_Result; 
   --# scope-start
               --# bind ifc Item_333
      --# expr-start 11129 '<FieldAccess for xref_equation at ast.py:7550>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7550
      if Item_333.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_6 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Item_333.Node, Env => Bound_Env, Origin => Origin, E_Info => Item_333.Info);
--# end
--# expr-done 11129
         declare
            Item_To_Append : constant Logic_Equation := Fld_6;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_3093;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 11130
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7550") else null)); 
--# expr-done 11131
And_Pred := Create_And (Fld_2, Logic_Boolean_Op, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7549") else null)); 
--# expr-done 11132
Dyn_Var_Bind_Result := And_Pred; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 11133
Let_Result := Dyn_Var_Bind_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 11134
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_569;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_569;
               Finalizer_Scope_3093;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Derived_Type_Def_P_Xref_Equation;
--# end
--# property-start DerivedTypeDef.discrete_range /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7554
pragma Warnings (Off, "is not referenced");
 function Derived_Type_Def_P_Discrete_Range
  (Node : Bare_Derived_Type_Def
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Discrete_Range
is
   Self : Bare_Derived_Type_Def := Bare_Derived_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Derived_Type_Def :=
     Internal_Entity_Derived_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Discrete_Range;
      Fld : Internal_Entity_Subtype_Indication;
Fld_1 : Internal_Discrete_Range;
Let_Result : Internal_Discrete_Range;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11137 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11136 '<FieldAccess for discrete_range at ast.py:7555>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7555
--# expr-start 11135 '<FieldAccess for subtype_indication at ast.py:7555>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7555
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Subtype_Indication (Node => Ent.Node.Derived_Type_Def_F_Subtype_Indication, Info => Ent.Info);
--# expr-done 11135
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start SubtypeIndication.discrete_range
Fld_1 := Libadalang.Implementation.Subtype_Indication_P_Discrete_Range (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 11136
Let_Result := Fld_1; 
--# end
--# expr-done 11137
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Derived_Type_Def_P_Discrete_Range;
--# end
   --
   --  Primitives for Bare_Enum_Type_Def
   --
      procedure Initialize_Fields_For_Enum_Type_Def
        (Self : Bare_Enum_Type_Def
         ; Enum_Type_Def_F_Enum_Literals : Bare_Enum_Literal_Decl_List
        ) is
      begin
            Self.Enum_Type_Def_F_Enum_Literals := Enum_Type_Def_F_Enum_Literals;
      end Initialize_Fields_For_Enum_Type_Def;
   function Enum_Type_Def_F_Enum_Literals
     (Node : Bare_Enum_Type_Def) return Bare_Enum_Literal_Decl_List
   is
   begin
         return Node.Enum_Type_Def_F_Enum_Literals;
   end;
--# property-start EnumTypeDef.is_char_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7088
pragma Warnings (Off, "is not referenced");
 function Enum_Type_Def_P_Is_Char_Type
  (Node : Bare_Enum_Type_Def
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Enum_Type_Def := Bare_Enum_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Bare_Enum_Literal_Decl_List;
Fld_1 : Bare_Defining_Name;
Fld_2 : Bare_Name;
Is_A : Boolean;
Item_334 : Bare_Enum_Literal_Decl;
Quantifier_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11142 '<AnyQuantifier at ast.py:7088>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7088
   --# expr-start 11138 '<FieldAccess for enum_literals at ast.py:7088>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7088
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Enum_Type_Def_F_Enum_Literals;
--# expr-done 11138
      Quantifier_Result := False;
      if Fld = null then
      null;
      else
      declare
         Collection : constant Bare_Enum_Literal_Decl_List := Fld;
      begin
         for Untyped_Item_334 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_334 := Untyped_Item_334; 
   --# scope-start
               --# bind lit Item_334
      --# expr-start 11141 '<IsA CharLiteral at ast.py:7089>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7089
--# expr-start 11140 '<FieldAccess for name at ast.py:7089>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7089
--# expr-start 11139 '<FieldAccess for name at ast.py:7089>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7089
      if Item_334 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Item_334.Enum_Literal_Decl_F_Name;
--# expr-done 11139
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Fld_1.Defining_Name_F_Name;
--# expr-done 11140
Is_A := Fld_2 /= null 
and then Fld_2.Kind in Ada_Char_Literal_Range; 
--# expr-done 11141
      Quantifier_Result := Is_A;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
      end if;
--# expr-done 11142
      Property_Result := Quantifier_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Type_Def_P_Is_Char_Type;
--# end
--# property-start EnumTypeDef.is_enum_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7092
pragma Warnings (Off, "is not referenced");
 function Enum_Type_Def_P_Is_Enum_Type
  (Node : Bare_Enum_Type_Def
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Enum_Type_Def := Bare_Enum_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Type_Def_P_Is_Enum_Type;
--# end
--# property-start EnumTypeDef.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7094
pragma Warnings (Off, "is not referenced");
 function Enum_Type_Def_P_Xref_Equation
  (Node : Bare_Enum_Type_Def
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Enum_Type_Def := Bare_Enum_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_531 with Inline_Always;
      True_Rel : Logic_Equation;
            procedure Finalizer_Scope_531 is
            begin
                     Dec_Ref (True_Rel);
            end Finalizer_Scope_531;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7094") else null)); 
      Property_Result := True_Rel;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_531;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_531;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Type_Def_P_Xref_Equation;
--# end
--# property-start EnumTypeDef.is_static /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7096
pragma Warnings (Off, "is not referenced");
 function Enum_Type_Def_P_Is_Static
  (Node : Bare_Enum_Type_Def
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Enum_Type_Def := Bare_Enum_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Type_Def_P_Is_Static;
--# end
--# property-start EnumTypeDef.is_std_char_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7099
pragma Warnings (Off, "is not referenced");
 function Enum_Type_Def_P_Is_Std_Char_Type
  (Node : Bare_Enum_Type_Def
  )
   return Boolean
is
   Self : Bare_Enum_Type_Def := Bare_Enum_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Block_Var : Bare_Type_Decl;
Fld : Bare_Ada_Node;
Cast_Result : Bare_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Bare_Base_Type_Decl;
Cast_Expr : Bare_Type_Decl;
Cast_Result_1 : Bare_Base_Type_Decl;
Is_Equal : Boolean;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Bare_Base_Type_Decl;
Cast_Expr_1 : Bare_Type_Decl;
Cast_Result_2 : Bare_Base_Type_Decl;
Is_Equal_1 : Boolean;
If_Result : Boolean;
Fld_5 : Internal_Entity_Base_Type_Decl;
Fld_6 : Bare_Base_Type_Decl;
Cast_Expr_2 : Bare_Type_Decl;
Cast_Result_3 : Bare_Base_Type_Decl;
Is_Equal_2 : Boolean;
If_Result_1 : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11154 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11144 '<Cast to TypeDecl at ast.py:7100>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7100
--# expr-start 11143 '<parent at ast.py:7100>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7100
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 11143
   if Fld = null
      or else Fld.Kind in Ada_Type_Decl_Range
   then
      Cast_Result := Fld;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 11144
Block_Var := Cast_Result; 
--# bind self_type Block_Var
--# expr-start 11147 '<Eq at ast.py:7101>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7101
Cast_Expr := Block_Var; 
      Cast_Result_1 := Cast_Expr;
--# expr-start 11146 '<FieldAccess for node at ast.py:7102>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7102
--# expr-start 11145 '<FieldAccess for std_char_type at ast.py:7102>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7102
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.std_char_type
Fld_1 := Libadalang.Implementation.Ada_Node_P_Std_Char_Type (Node => Self);
--# end
--# expr-done 11145
Fld_2 := Fld_1.Node;
--# expr-done 11146
Is_Equal := Cast_Result_1 = Fld_2; 
--# expr-done 11147
if Is_Equal then
   If_Result := True;
else
   --# expr-start 11150 '<Eq at ast.py:7101>' Is_Equal_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7101
Cast_Expr_1 := Block_Var; 
      Cast_Result_2 := Cast_Expr_1;
--# expr-start 11149 '<FieldAccess for node at ast.py:7103>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7103
--# expr-start 11148 '<FieldAccess for std_wide_char_type at ast.py:7103>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7103
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.std_wide_char_type
Fld_3 := Libadalang.Implementation.Ada_Node_P_Std_Wide_Char_Type (Node => Self);
--# end
--# expr-done 11148
Fld_4 := Fld_3.Node;
--# expr-done 11149
Is_Equal_1 := Cast_Result_2 = Fld_4; 
--# expr-done 11150
   If_Result := Is_Equal_1;
end if;
if If_Result then
   If_Result_1 := True;
else
   --# expr-start 11153 '<Eq at ast.py:7101>' Is_Equal_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7101
Cast_Expr_2 := Block_Var; 
      Cast_Result_3 := Cast_Expr_2;
--# expr-start 11152 '<FieldAccess for node at ast.py:7104>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7104
--# expr-start 11151 '<FieldAccess for std_wide_wide_char_type at ast.py:7104>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7104
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.std_wide_wide_char_type
Fld_5 := Libadalang.Implementation.Ada_Node_P_Std_Wide_Wide_Char_Type (Node => Self);
--# end
--# expr-done 11151
Fld_6 := Fld_5.Node;
--# expr-done 11152
Is_Equal_2 := Cast_Result_3 = Fld_6; 
--# expr-done 11153
   If_Result_1 := Is_Equal_2;
end if;
Let_Result := If_Result_1; 
--# end
--# expr-done 11154
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Type_Def_P_Is_Std_Char_Type;
--# end
--# property-start EnumTypeDef.predefined_operators /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7108
pragma Warnings (Off, "is not referenced");
 function Enum_Type_Def_P_Predefined_Operators
  (Node : Bare_Enum_Type_Def
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Enum_Type_Def := Bare_Enum_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_534 with Inline_Always;
      Block_Var : Bare_Type_Decl;
Block_Var_1 : Bare_Base_Type_Decl;
Block_Var_2 : Internal_Env_Assoc_Array_Access;
Block_Var_3 : Internal_Env_Assoc_Array_Access;
Fld : Bare_Ada_Node;
Cast_Result : Bare_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Bare_Base_Type_Decl;
Sym : Symbol_Type;
Cast_Expr : Bare_Type_Decl;
Cast_Result_1 : Bare_Base_Type_Decl;
Cast_Expr_1 : Bare_Type_Decl;
Cast_Result_2 : Bare_Base_Type_Decl;
Fld_3 : Internal_Env_Assoc := No_Env_Assoc;
Sym_1 : Symbol_Type;
Cast_Expr_2 : Bare_Type_Decl;
Cast_Result_3 : Bare_Base_Type_Decl;
Cast_Expr_3 : Bare_Type_Decl;
Cast_Result_4 : Bare_Base_Type_Decl;
Fld_4 : Internal_Env_Assoc := No_Env_Assoc;
Sym_2 : Symbol_Type;
Cast_Expr_4 : Bare_Type_Decl;
Cast_Result_5 : Bare_Base_Type_Decl;
Cast_Expr_5 : Bare_Type_Decl;
Cast_Result_6 : Bare_Base_Type_Decl;
Fld_5 : Internal_Env_Assoc := No_Env_Assoc;
Sym_3 : Symbol_Type;
Cast_Expr_6 : Bare_Type_Decl;
Cast_Result_7 : Bare_Base_Type_Decl;
Cast_Expr_7 : Bare_Type_Decl;
Cast_Result_8 : Bare_Base_Type_Decl;
Fld_6 : Internal_Env_Assoc := No_Env_Assoc;
Sym_4 : Symbol_Type;
Cast_Expr_8 : Bare_Type_Decl;
Cast_Result_9 : Bare_Base_Type_Decl;
Cast_Expr_9 : Bare_Type_Decl;
Cast_Result_10 : Bare_Base_Type_Decl;
Fld_7 : Internal_Env_Assoc := No_Env_Assoc;
Sym_5 : Symbol_Type;
Cast_Expr_10 : Bare_Type_Decl;
Cast_Result_11 : Bare_Base_Type_Decl;
Cast_Expr_11 : Bare_Type_Decl;
Cast_Result_12 : Bare_Base_Type_Decl;
Fld_8 : Internal_Env_Assoc := No_Env_Assoc;
Array_Lit : Internal_Env_Assoc_Array_Access;
Sym_6 : Symbol_Type;
Cast_Expr_12 : Bare_Type_Decl;
Cast_Result_13 : Bare_Base_Type_Decl;
Cast_Expr_13 : Bare_Type_Decl;
Cast_Result_14 : Bare_Base_Type_Decl;
Cast_Expr_14 : Bare_Type_Decl;
Cast_Result_15 : Bare_Base_Type_Decl;
Fld_9 : Internal_Env_Assoc := No_Env_Assoc;
Sym_7 : Symbol_Type;
Cast_Expr_15 : Bare_Type_Decl;
Cast_Result_16 : Bare_Base_Type_Decl;
Cast_Expr_16 : Bare_Type_Decl;
Cast_Result_17 : Bare_Base_Type_Decl;
Cast_Expr_17 : Bare_Type_Decl;
Cast_Result_18 : Bare_Base_Type_Decl;
Fld_10 : Internal_Env_Assoc := No_Env_Assoc;
Sym_8 : Symbol_Type;
Cast_Expr_18 : Bare_Type_Decl;
Cast_Result_19 : Bare_Base_Type_Decl;
Cast_Expr_19 : Bare_Type_Decl;
Cast_Result_20 : Bare_Base_Type_Decl;
Cast_Expr_20 : Bare_Type_Decl;
Cast_Result_21 : Bare_Base_Type_Decl;
Fld_11 : Internal_Env_Assoc := No_Env_Assoc;
Sym_9 : Symbol_Type;
Cast_Expr_21 : Bare_Type_Decl;
Cast_Result_22 : Bare_Base_Type_Decl;
Cast_Expr_22 : Bare_Type_Decl;
Cast_Result_23 : Bare_Base_Type_Decl;
Fld_12 : Internal_Env_Assoc := No_Env_Assoc;
Array_Lit_1 : Internal_Env_Assoc_Array_Access;
Cast_Expr_23 : Bare_Type_Decl;
Cast_Result_24 : Bare_Base_Type_Decl;
Is_Equal : Boolean;
If_Result : Internal_Env_Assoc_Array_Access;
Concat_Result : Internal_Env_Assoc_Array_Access;
Let_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_534 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Fld_5);
                     Dec_Ref (Fld_6);
                     Dec_Ref (Fld_7);
                     Dec_Ref (Fld_8);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Block_Var_2);
                     Dec_Ref (Fld_9);
                     Dec_Ref (Fld_10);
                     Dec_Ref (Fld_11);
                     Dec_Ref (Fld_12);
                     Dec_Ref (Array_Lit_1);
                     Dec_Ref (If_Result);
                     Dec_Ref (Block_Var_3);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_534;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Enum_Type_Def_Enum_Type_Def_P_Predefined_Operators,
            Items    => new Mmz_Key_Array (1 ..  1))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Enum_Type_Def,
                                As_Bare_Enum_Type_Def => Self);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Env_Assoc_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 11184 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11156 '<Cast to TypeDecl at ast.py:7109>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7109
--# expr-start 11155 '<parent at ast.py:7109>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7109
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 11155
   if Fld = null
      or else Fld.Kind in Ada_Type_Decl_Range
   then
      Cast_Result := Fld;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 11156
Block_Var := Cast_Result; 
--# bind self_type Block_Var
--# expr-start 11158 '<FieldAccess for node at ast.py:7110>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7110
--# expr-start 11157 '<FieldAccess for bool_type at ast.py:7110>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7110
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_1 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 11157
Fld_2 := Fld_1.Node;
--# expr-done 11158
Block_Var_1 := Fld_2; 
--# bind bool_type Block_Var_1
--# expr-start 11171 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 11160 '<FieldAccess for create_binop_assoc at ast.py:7113>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7113
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11159 '<Symbol "<" at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_8); 
--# expr-done 11159
Cast_Expr := Block_Var; 
      Cast_Result_1 := Cast_Expr;
Cast_Expr_1 := Block_Var; 
      Cast_Result_2 := Cast_Expr_1;
--# property-call-start AdaNode.create_binop_assoc
Fld_3 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym, Lhs => Cast_Result_1, Rhs => Cast_Result_2, Ret => Block_Var_1);
--# end
--# expr-done 11160
--# expr-start 11162 '<FieldAccess for create_binop_assoc at ast.py:7114>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7114
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11161 '<Symbol "<=" at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_9); 
--# expr-done 11161
Cast_Expr_2 := Block_Var; 
      Cast_Result_3 := Cast_Expr_2;
Cast_Expr_3 := Block_Var; 
      Cast_Result_4 := Cast_Expr_3;
--# property-call-start AdaNode.create_binop_assoc
Fld_4 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_1, Lhs => Cast_Result_3, Rhs => Cast_Result_4, Ret => Block_Var_1);
--# end
--# expr-done 11162
--# expr-start 11164 '<FieldAccess for create_binop_assoc at ast.py:7115>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7115
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11163 '<Symbol "=" at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_10); 
--# expr-done 11163
Cast_Expr_4 := Block_Var; 
      Cast_Result_5 := Cast_Expr_4;
Cast_Expr_5 := Block_Var; 
      Cast_Result_6 := Cast_Expr_5;
--# property-call-start AdaNode.create_binop_assoc
Fld_5 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_2, Lhs => Cast_Result_5, Rhs => Cast_Result_6, Ret => Block_Var_1);
--# end
--# expr-done 11164
--# expr-start 11166 '<FieldAccess for create_binop_assoc at ast.py:7116>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7116
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11165 '<Symbol "/=" at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_7); 
--# expr-done 11165
Cast_Expr_6 := Block_Var; 
      Cast_Result_7 := Cast_Expr_6;
Cast_Expr_7 := Block_Var; 
      Cast_Result_8 := Cast_Expr_7;
--# property-call-start AdaNode.create_binop_assoc
Fld_6 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_3, Lhs => Cast_Result_7, Rhs => Cast_Result_8, Ret => Block_Var_1);
--# end
--# expr-done 11166
--# expr-start 11168 '<FieldAccess for create_binop_assoc at ast.py:7117>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7117
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11167 '<Symbol ">" at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_11); 
--# expr-done 11167
Cast_Expr_8 := Block_Var; 
      Cast_Result_9 := Cast_Expr_8;
Cast_Expr_9 := Block_Var; 
      Cast_Result_10 := Cast_Expr_9;
--# property-call-start AdaNode.create_binop_assoc
Fld_7 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_4, Lhs => Cast_Result_9, Rhs => Cast_Result_10, Ret => Block_Var_1);
--# end
--# expr-done 11168
--# expr-start 11170 '<FieldAccess for create_binop_assoc at ast.py:7118>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7118
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11169 '<Symbol ">=" at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_12); 
--# expr-done 11169
Cast_Expr_10 := Block_Var; 
      Cast_Result_11 := Cast_Expr_10;
Cast_Expr_11 := Block_Var; 
      Cast_Result_12 := Cast_Expr_11;
--# property-call-start AdaNode.create_binop_assoc
Fld_8 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_5, Lhs => Cast_Result_11, Rhs => Cast_Result_12, Ret => Block_Var_1);
--# end
--# expr-done 11170
Array_Lit := Create_Internal_Env_Assoc_Array (Internal_Internal_Env_Assoc_Array'(1 => Fld_3, 2 => Fld_4, 3 => Fld_5, 4 => Fld_6, 5 => Fld_7, 6 => Fld_8)); 
--# expr-done 11171
Block_Var_2 := Array_Lit; Inc_Ref (Block_Var_2);
--# bind defaults Block_Var_2
--# expr-start 11182 '<If at ast.py:7122>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7122
--# expr-start 11172 '<Eq at ast.py:7123>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7123
Cast_Expr_23 := Block_Var; 
      Cast_Result_24 := Cast_Expr_23;
Is_Equal := Cast_Result_24 = Block_Var_1; 
--# expr-done 11172
if Is_Equal then
   --# expr-start 11181 '<ArrayLiteral at ???>' Array_Lit_1 None
--# expr-start 11174 '<FieldAccess for create_binop_assoc at ast.py:7124>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7124
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11173 '<Symbol "and" at ???>' Sym_6 None
Sym_6 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_And); 
--# expr-done 11173
Cast_Expr_12 := Block_Var; 
      Cast_Result_13 := Cast_Expr_12;
Cast_Expr_13 := Block_Var; 
      Cast_Result_14 := Cast_Expr_13;
Cast_Expr_14 := Block_Var; 
      Cast_Result_15 := Cast_Expr_14;
--# property-call-start AdaNode.create_binop_assoc
Fld_9 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_6, Lhs => Cast_Result_13, Rhs => Cast_Result_14, Ret => Cast_Result_15);
--# end
--# expr-done 11174
--# expr-start 11176 '<FieldAccess for create_binop_assoc at ast.py:7125>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7125
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11175 '<Symbol "or" at ???>' Sym_7 None
Sym_7 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Or); 
--# expr-done 11175
Cast_Expr_15 := Block_Var; 
      Cast_Result_16 := Cast_Expr_15;
Cast_Expr_16 := Block_Var; 
      Cast_Result_17 := Cast_Expr_16;
Cast_Expr_17 := Block_Var; 
      Cast_Result_18 := Cast_Expr_17;
--# property-call-start AdaNode.create_binop_assoc
Fld_10 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_7, Lhs => Cast_Result_16, Rhs => Cast_Result_17, Ret => Cast_Result_18);
--# end
--# expr-done 11176
--# expr-start 11178 '<FieldAccess for create_binop_assoc at ast.py:7126>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7126
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11177 '<Symbol "xor" at ???>' Sym_8 None
Sym_8 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Xor); 
--# expr-done 11177
Cast_Expr_18 := Block_Var; 
      Cast_Result_19 := Cast_Expr_18;
Cast_Expr_19 := Block_Var; 
      Cast_Result_20 := Cast_Expr_19;
Cast_Expr_20 := Block_Var; 
      Cast_Result_21 := Cast_Expr_20;
--# property-call-start AdaNode.create_binop_assoc
Fld_11 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_8, Lhs => Cast_Result_19, Rhs => Cast_Result_20, Ret => Cast_Result_21);
--# end
--# expr-done 11178
--# expr-start 11180 '<FieldAccess for create_unop_assoc at ast.py:7127>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7127
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11179 '<Symbol "not" at ???>' Sym_9 None
Sym_9 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Not); 
--# expr-done 11179
Cast_Expr_21 := Block_Var; 
      Cast_Result_22 := Cast_Expr_21;
Cast_Expr_22 := Block_Var; 
      Cast_Result_23 := Cast_Expr_22;
--# property-call-start AdaNode.create_unop_assoc
Fld_12 := Libadalang.Implementation.Ada_Node_P_Create_Unop_Assoc (Node => Self, Op => Sym_9, Rhs => Cast_Result_22, Ret => Cast_Result_23);
--# end
--# expr-done 11180
Array_Lit_1 := Create_Internal_Env_Assoc_Array (Internal_Internal_Env_Assoc_Array'(1 => Fld_9, 2 => Fld_10, 3 => Fld_11, 4 => Fld_12)); 
--# expr-done 11181
   If_Result := Array_Lit_1;
else
   If_Result := No_Internal_Env_Assoc_Array_Type;
end if;
      Inc_Ref (If_Result);
--# expr-done 11182
Block_Var_3 := If_Result; Inc_Ref (Block_Var_3);
--# bind specials Block_Var_3
--# expr-start 11183 '<Concat at ast.py:7131>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7131
Concat_Result := Concat (Block_Var_2, Block_Var_3); 
--# expr-done 11183
Let_Result := Concat_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 11184
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_534;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Env_Assoc_Array_Access,
                        As_Internal_Env_Assoc_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_534;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Type_Def_P_Predefined_Operators;
--# end
   --
   --  Primitives for Bare_Formal_Discrete_Type_Def
   --
--# property-start FormalDiscreteTypeDef.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8210
pragma Warnings (Off, "is not referenced");
 function Formal_Discrete_Type_Def_P_Xref_Equation
  (Node : Bare_Formal_Discrete_Type_Def
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Formal_Discrete_Type_Def := Bare_Formal_Discrete_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_672 with Inline_Always;
      True_Rel : Logic_Equation;
            procedure Finalizer_Scope_672 is
            begin
                     Dec_Ref (True_Rel);
            end Finalizer_Scope_672;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8210") else null)); 
      Property_Result := True_Rel;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_672;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_672;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Formal_Discrete_Type_Def_P_Xref_Equation;
--# end
--# property-start FormalDiscreteTypeDef.is_discrete_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8212
pragma Warnings (Off, "is not referenced");
 function Formal_Discrete_Type_Def_P_Is_Discrete_Type
  (Node : Bare_Formal_Discrete_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Formal_Discrete_Type_Def := Bare_Formal_Discrete_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Formal_Discrete_Type_Def_P_Is_Discrete_Type;
--# end
--# property-start FormalDiscreteTypeDef.predefined_operators /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8215
pragma Warnings (Off, "is not referenced");
 function Formal_Discrete_Type_Def_P_Predefined_Operators
  (Node : Bare_Formal_Discrete_Type_Def
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Formal_Discrete_Type_Def := Bare_Formal_Discrete_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_674 with Inline_Always;
      Block_Var : Bare_Type_Decl;
Block_Var_1 : Bare_Base_Type_Decl;
Fld : Bare_Ada_Node;
Cast_Result : Bare_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Bare_Base_Type_Decl;
Sym : Symbol_Type;
Cast_Expr : Bare_Type_Decl;
Cast_Result_1 : Bare_Base_Type_Decl;
Cast_Expr_1 : Bare_Type_Decl;
Cast_Result_2 : Bare_Base_Type_Decl;
Fld_3 : Internal_Env_Assoc := No_Env_Assoc;
Sym_1 : Symbol_Type;
Cast_Expr_2 : Bare_Type_Decl;
Cast_Result_3 : Bare_Base_Type_Decl;
Cast_Expr_3 : Bare_Type_Decl;
Cast_Result_4 : Bare_Base_Type_Decl;
Fld_4 : Internal_Env_Assoc := No_Env_Assoc;
Sym_2 : Symbol_Type;
Cast_Expr_4 : Bare_Type_Decl;
Cast_Result_5 : Bare_Base_Type_Decl;
Cast_Expr_5 : Bare_Type_Decl;
Cast_Result_6 : Bare_Base_Type_Decl;
Fld_5 : Internal_Env_Assoc := No_Env_Assoc;
Sym_3 : Symbol_Type;
Cast_Expr_6 : Bare_Type_Decl;
Cast_Result_7 : Bare_Base_Type_Decl;
Cast_Expr_7 : Bare_Type_Decl;
Cast_Result_8 : Bare_Base_Type_Decl;
Fld_6 : Internal_Env_Assoc := No_Env_Assoc;
Sym_4 : Symbol_Type;
Cast_Expr_8 : Bare_Type_Decl;
Cast_Result_9 : Bare_Base_Type_Decl;
Cast_Expr_9 : Bare_Type_Decl;
Cast_Result_10 : Bare_Base_Type_Decl;
Fld_7 : Internal_Env_Assoc := No_Env_Assoc;
Sym_5 : Symbol_Type;
Cast_Expr_10 : Bare_Type_Decl;
Cast_Result_11 : Bare_Base_Type_Decl;
Cast_Expr_11 : Bare_Type_Decl;
Cast_Result_12 : Bare_Base_Type_Decl;
Fld_8 : Internal_Env_Assoc := No_Env_Assoc;
Array_Lit : Internal_Env_Assoc_Array_Access;
Let_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_674 is
            begin
                     Dec_Ref (Fld_3);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Fld_5);
                     Dec_Ref (Fld_6);
                     Dec_Ref (Fld_7);
                     Dec_Ref (Fld_8);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_674;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Formal_Discrete_Type_Def_Formal_Discrete_Type_Def_P_Predefined_Operators,
            Items    => new Mmz_Key_Array (1 ..  1))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Formal_Discrete_Type_Def,
                                As_Bare_Formal_Discrete_Type_Def => Self);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Env_Assoc_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 11202 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11186 '<Cast to TypeDecl at ast.py:8216>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8216
--# expr-start 11185 '<parent at ast.py:8216>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8216
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 11185
   if Fld = null
      or else Fld.Kind in Ada_Type_Decl_Range
   then
      Cast_Result := Fld;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 11186
Block_Var := Cast_Result; 
--# bind self_type Block_Var
--# expr-start 11188 '<FieldAccess for node at ast.py:8217>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8217
--# expr-start 11187 '<FieldAccess for bool_type at ast.py:8217>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8217
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_1 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 11187
Fld_2 := Fld_1.Node;
--# expr-done 11188
Block_Var_1 := Fld_2; 
--# bind bool_type Block_Var_1
--# expr-start 11201 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 11190 '<FieldAccess for create_binop_assoc at ast.py:8219>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8219
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11189 '<Symbol "<" at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_8); 
--# expr-done 11189
Cast_Expr := Block_Var; 
      Cast_Result_1 := Cast_Expr;
Cast_Expr_1 := Block_Var; 
      Cast_Result_2 := Cast_Expr_1;
--# property-call-start AdaNode.create_binop_assoc
Fld_3 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym, Lhs => Cast_Result_1, Rhs => Cast_Result_2, Ret => Block_Var_1);
--# end
--# expr-done 11190
--# expr-start 11192 '<FieldAccess for create_binop_assoc at ast.py:8220>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8220
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11191 '<Symbol "<=" at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_9); 
--# expr-done 11191
Cast_Expr_2 := Block_Var; 
      Cast_Result_3 := Cast_Expr_2;
Cast_Expr_3 := Block_Var; 
      Cast_Result_4 := Cast_Expr_3;
--# property-call-start AdaNode.create_binop_assoc
Fld_4 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_1, Lhs => Cast_Result_3, Rhs => Cast_Result_4, Ret => Block_Var_1);
--# end
--# expr-done 11192
--# expr-start 11194 '<FieldAccess for create_binop_assoc at ast.py:8221>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8221
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11193 '<Symbol "=" at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_10); 
--# expr-done 11193
Cast_Expr_4 := Block_Var; 
      Cast_Result_5 := Cast_Expr_4;
Cast_Expr_5 := Block_Var; 
      Cast_Result_6 := Cast_Expr_5;
--# property-call-start AdaNode.create_binop_assoc
Fld_5 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_2, Lhs => Cast_Result_5, Rhs => Cast_Result_6, Ret => Block_Var_1);
--# end
--# expr-done 11194
--# expr-start 11196 '<FieldAccess for create_binop_assoc at ast.py:8222>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8222
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11195 '<Symbol "/=" at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_7); 
--# expr-done 11195
Cast_Expr_6 := Block_Var; 
      Cast_Result_7 := Cast_Expr_6;
Cast_Expr_7 := Block_Var; 
      Cast_Result_8 := Cast_Expr_7;
--# property-call-start AdaNode.create_binop_assoc
Fld_6 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_3, Lhs => Cast_Result_7, Rhs => Cast_Result_8, Ret => Block_Var_1);
--# end
--# expr-done 11196
--# expr-start 11198 '<FieldAccess for create_binop_assoc at ast.py:8223>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8223
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11197 '<Symbol ">" at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_11); 
--# expr-done 11197
Cast_Expr_8 := Block_Var; 
      Cast_Result_9 := Cast_Expr_8;
Cast_Expr_9 := Block_Var; 
      Cast_Result_10 := Cast_Expr_9;
--# property-call-start AdaNode.create_binop_assoc
Fld_7 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_4, Lhs => Cast_Result_9, Rhs => Cast_Result_10, Ret => Block_Var_1);
--# end
--# expr-done 11198
--# expr-start 11200 '<FieldAccess for create_binop_assoc at ast.py:8224>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8224
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11199 '<Symbol ">=" at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_12); 
--# expr-done 11199
Cast_Expr_10 := Block_Var; 
      Cast_Result_11 := Cast_Expr_10;
Cast_Expr_11 := Block_Var; 
      Cast_Result_12 := Cast_Expr_11;
--# property-call-start AdaNode.create_binop_assoc
Fld_8 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_5, Lhs => Cast_Result_11, Rhs => Cast_Result_12, Ret => Block_Var_1);
--# end
--# expr-done 11200
Array_Lit := Create_Internal_Env_Assoc_Array (Internal_Internal_Env_Assoc_Array'(1 => Fld_3, 2 => Fld_4, 3 => Fld_5, 4 => Fld_6, 5 => Fld_7, 6 => Fld_8)); 
--# expr-done 11201
Let_Result := Array_Lit; Inc_Ref (Let_Result);
--# end
--# expr-done 11202
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_674;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Env_Assoc_Array_Access,
                        As_Internal_Env_Assoc_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_674;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Formal_Discrete_Type_Def_P_Predefined_Operators;
--# end
   --
   --  Primitives for Bare_Interface_Type_Def
   --
      procedure Initialize_Fields_For_Interface_Type_Def
        (Self : Bare_Interface_Type_Def
         ; Interface_Type_Def_F_Interface_Kind : Bare_Interface_Kind
         ; Interface_Type_Def_F_Interfaces : Bare_Parent_List
        ) is
      begin
            Self.Interface_Type_Def_F_Interface_Kind := Interface_Type_Def_F_Interface_Kind;
            Self.Interface_Type_Def_F_Interfaces := Interface_Type_Def_F_Interfaces;
      end Initialize_Fields_For_Interface_Type_Def;
   function Interface_Type_Def_F_Interface_Kind
     (Node : Bare_Interface_Type_Def) return Bare_Interface_Kind
   is
   begin
         return Node.Interface_Type_Def_F_Interface_Kind;
   end;
   function Interface_Type_Def_F_Interfaces
     (Node : Bare_Interface_Type_Def) return Bare_Parent_List
   is
   begin
         return Node.Interface_Type_Def_F_Interfaces;
   end;
--# property-start InterfaceTypeDef.is_tagged_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7896
pragma Warnings (Off, "is not referenced");
 function Interface_Type_Def_P_Is_Tagged_Type
  (Node : Bare_Interface_Type_Def
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Interface_Type_Def := Bare_Interface_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Interface_Type_Def_P_Is_Tagged_Type;
--# end
--# property-start InterfaceTypeDef.base_interfaces /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7898
pragma Warnings (Off, "is not referenced");
 function Interface_Type_Def_P_Base_Interfaces
  (Node : Bare_Interface_Type_Def
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl_Array_Access
is
   Self : Bare_Interface_Type_Def := Bare_Interface_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Interface_Type_Def :=
     Internal_Entity_Interface_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_607 with Inline_Always;
      Fld : Internal_Entity_Parent_List;
Fld_1 : Bare_Parent_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld_3 : Internal_Entity_Base_Type_Decl;
Item_335 : Internal_Entity_Name;
Bare_Item_335 : Bare_Name;
Map_Result : Internal_Entity_Base_Type_Decl_Array_Access;
            procedure Finalizer_Scope_607 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_607;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11205 '<Map at ast.py:7899>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7899
   --# expr-start 11203 '<FieldAccess for interfaces at ast.py:7899>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7899
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Parent_List (Node => Ent.Node.Interface_Type_Def_F_Interfaces, Info => Ent.Info);
--# expr-done 11203
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Internal_Entity_Base_Type_Decl_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Internal_Entity_Base_Type_Decl_Array (0);
      else
      declare
         Collection : constant Bare_Parent_List := Fld_1;
      begin
         for Untyped_Item_335 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_335 := Untyped_Item_335; 
Is_Null := Bare_Item_335 = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_335); 
   If_Result := As_Entity;
end if;
                  Item_335 := If_Result; 
   --# scope-start
               --# bind i Item_335
      --# expr-start 11204 '<FieldAccess for name_designated_type at ast.py:7899>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7899
      if Item_335.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_3 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Item_335.Node, E_Info => Item_335.Info);
--# end
--# expr-done 11204
         declare
            Item_To_Append : constant Internal_Entity_Base_Type_Decl := Fld_3;
         begin
            Internal_Entity_Base_Type_Decl_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Entity_Base_Type_Decl_Array
           (Items_Count => Natural (Internal_Entity_Base_Type_Decl_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Entity_Base_Type_Decl_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Entity_Base_Type_Decl_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Entity_Base_Type_Decl_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 11205
      Property_Result := Map_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_607;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_607;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Interface_Type_Def_P_Base_Interfaces;
--# end
--# property-start InterfaceTypeDef.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7903
pragma Warnings (Off, "is not referenced");
 function Interface_Type_Def_P_Xref_Equation
  (Node : Bare_Interface_Type_Def
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Interface_Type_Def := Bare_Interface_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Interface_Type_Def :=
     Internal_Entity_Interface_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_608 with Inline_Always;
            procedure Finalizer_Scope_3096 with Inline_Always;
      Fld : Internal_Entity_Parent_List;
Fld_1 : Bare_Parent_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld_3 : Logic_Equation;
Item_336 : Internal_Entity_Name;
Bare_Item_336 : Bare_Name;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_608 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_608;
            procedure Finalizer_Scope_3096 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_3096;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11210 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11209 '<LogicAll at ast.py:7904>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7904
--# expr-start 11208 '<Map at ast.py:7904>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7904
   --# expr-start 11206 '<FieldAccess for interfaces at ast.py:7904>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7904
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Parent_List (Node => Ent.Node.Interface_Type_Def_F_Interfaces, Info => Ent.Info);
--# expr-done 11206
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Parent_List := Fld_1;
      begin
         for Untyped_Item_336 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_336 := Untyped_Item_336; 
Is_Null := Bare_Item_336 = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_336); 
   If_Result := As_Entity;
end if;
                  Item_336 := If_Result; 
   --# scope-start
               --# bind ifc Item_336
      --# expr-start 11207 '<FieldAccess for xref_equation at ast.py:7904>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7904
      if Item_336.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_3 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Item_336.Node, Env => Env, Origin => Origin, E_Info => Item_336.Info);
--# end
--# expr-done 11207
         declare
            Item_To_Append : constant Logic_Equation := Fld_3;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_3096;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 11208
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7904") else null)); 
--# expr-done 11209
Let_Result := Logic_Boolean_Op; Inc_Ref (Let_Result);
--# end
--# expr-done 11210
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_608;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_608;
               Finalizer_Scope_3096;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Interface_Type_Def_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Mod_Int_Type_Def
   --
      procedure Initialize_Fields_For_Mod_Int_Type_Def
        (Self : Bare_Mod_Int_Type_Def
         ; Mod_Int_Type_Def_F_Expr : Bare_Expr
        ) is
      begin
            Self.Mod_Int_Type_Def_F_Expr := Mod_Int_Type_Def_F_Expr;
      end Initialize_Fields_For_Mod_Int_Type_Def;
   function Mod_Int_Type_Def_F_Expr
     (Node : Bare_Mod_Int_Type_Def) return Bare_Expr
   is
   begin
         return Node.Mod_Int_Type_Def_F_Expr;
   end;
--# property-start ModIntTypeDef.is_int_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7634
pragma Warnings (Off, "is not referenced");
 function Mod_Int_Type_Def_P_Is_Int_Type
  (Node : Bare_Mod_Int_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Mod_Int_Type_Def := Bare_Mod_Int_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Mod_Int_Type_Def_P_Is_Int_Type;
--# end
--# property-start ModIntTypeDef.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7636
pragma Warnings (Off, "is not referenced");
 function Mod_Int_Type_Def_P_Xref_Equation
  (Node : Bare_Mod_Int_Type_Def
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Mod_Int_Type_Def := Bare_Mod_Int_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Mod_Int_Type_Def :=
     Internal_Entity_Mod_Int_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_580 with Inline_Always;
      Fld : Internal_Entity_Expr;
Fld_1 : Logic_Equation;
            procedure Finalizer_Scope_580 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_580;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11212 '<FieldAccess for sub_equation at ast.py:7636>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7636
--# expr-start 11211 '<FieldAccess for expr at ast.py:7636>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7636
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Mod_Int_Type_Def_F_Expr, Info => Ent.Info);
--# expr-done 11211
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11212
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_580;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_580;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Mod_Int_Type_Def_P_Xref_Equation;
--# end
--# property-start ModIntTypeDef.is_static /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7638
pragma Warnings (Off, "is not referenced");
 function Mod_Int_Type_Def_P_Is_Static
  (Node : Bare_Mod_Int_Type_Def
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Mod_Int_Type_Def := Bare_Mod_Int_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Mod_Int_Type_Def :=
     Internal_Entity_Mod_Int_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Fld : Internal_Entity_Expr;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11214 '<FieldAccess for is_static_expr at ast.py:7638>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7638
--# expr-start 11213 '<FieldAccess for expr at ast.py:7638>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7638
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Mod_Int_Type_Def_F_Expr, Info => Ent.Info);
--# expr-done 11213
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_1 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld.Info);
--# end
--# expr-done 11214
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Mod_Int_Type_Def_P_Is_Static;
--# end
--# property-start ModIntTypeDef.discrete_range /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7641
pragma Warnings (Off, "is not referenced");
 function Mod_Int_Type_Def_P_Discrete_Range
  (Node : Bare_Mod_Int_Type_Def
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Discrete_Range
is
   Self : Bare_Mod_Int_Type_Def := Bare_Mod_Int_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Mod_Int_Type_Def :=
     Internal_Entity_Mod_Int_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Discrete_Range;
      Fld : Internal_Entity_Expr;
New_Struct : Internal_Discrete_Range;
Let_Result : Internal_Discrete_Range;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11217 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11216 '<New DiscreteRange at ast.py:7642>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7642
--# expr-start 11215 '<FieldAccess for expr at ast.py:7643>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7643
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Mod_Int_Type_Def_F_Expr, Info => Ent.Info);
--# expr-done 11215
New_Struct := (High_Bound => Fld, Low_Bound => No_Entity_Expr); 
--# expr-done 11216
Let_Result := New_Struct; 
--# end
--# expr-done 11217
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Mod_Int_Type_Def_P_Discrete_Range;
--# end
--# property-start ModIntTypeDef.predefined_operators /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7646
pragma Warnings (Off, "is not referenced");
 function Mod_Int_Type_Def_P_Predefined_Operators
  (Node : Bare_Mod_Int_Type_Def
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Mod_Int_Type_Def := Bare_Mod_Int_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_583 with Inline_Always;
      Block_Var : Bare_Type_Decl;
Block_Var_1 : Bare_Base_Type_Decl;
Block_Var_2 : Bare_Base_Type_Decl;
Fld : Bare_Ada_Node;
Cast_Result : Bare_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Bare_Base_Type_Decl;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Bare_Base_Type_Decl;
Sym : Symbol_Type;
Cast_Expr : Bare_Type_Decl;
Cast_Result_1 : Bare_Base_Type_Decl;
Cast_Expr_1 : Bare_Type_Decl;
Cast_Result_2 : Bare_Base_Type_Decl;
Cast_Expr_2 : Bare_Type_Decl;
Cast_Result_3 : Bare_Base_Type_Decl;
Fld_5 : Internal_Env_Assoc := No_Env_Assoc;
Sym_1 : Symbol_Type;
Cast_Expr_3 : Bare_Type_Decl;
Cast_Result_4 : Bare_Base_Type_Decl;
Cast_Expr_4 : Bare_Type_Decl;
Cast_Result_5 : Bare_Base_Type_Decl;
Cast_Expr_5 : Bare_Type_Decl;
Cast_Result_6 : Bare_Base_Type_Decl;
Fld_6 : Internal_Env_Assoc := No_Env_Assoc;
Sym_2 : Symbol_Type;
Cast_Expr_6 : Bare_Type_Decl;
Cast_Result_7 : Bare_Base_Type_Decl;
Cast_Expr_7 : Bare_Type_Decl;
Cast_Result_8 : Bare_Base_Type_Decl;
Cast_Expr_8 : Bare_Type_Decl;
Cast_Result_9 : Bare_Base_Type_Decl;
Fld_7 : Internal_Env_Assoc := No_Env_Assoc;
Sym_3 : Symbol_Type;
Cast_Expr_9 : Bare_Type_Decl;
Cast_Result_10 : Bare_Base_Type_Decl;
Cast_Expr_10 : Bare_Type_Decl;
Cast_Result_11 : Bare_Base_Type_Decl;
Cast_Expr_11 : Bare_Type_Decl;
Cast_Result_12 : Bare_Base_Type_Decl;
Fld_8 : Internal_Env_Assoc := No_Env_Assoc;
Sym_4 : Symbol_Type;
Cast_Expr_12 : Bare_Type_Decl;
Cast_Result_13 : Bare_Base_Type_Decl;
Cast_Expr_13 : Bare_Type_Decl;
Cast_Result_14 : Bare_Base_Type_Decl;
Cast_Expr_14 : Bare_Type_Decl;
Cast_Result_15 : Bare_Base_Type_Decl;
Fld_9 : Internal_Env_Assoc := No_Env_Assoc;
Sym_5 : Symbol_Type;
Cast_Expr_15 : Bare_Type_Decl;
Cast_Result_16 : Bare_Base_Type_Decl;
Cast_Expr_16 : Bare_Type_Decl;
Cast_Result_17 : Bare_Base_Type_Decl;
Cast_Expr_17 : Bare_Type_Decl;
Cast_Result_18 : Bare_Base_Type_Decl;
Fld_10 : Internal_Env_Assoc := No_Env_Assoc;
Sym_6 : Symbol_Type;
Cast_Expr_18 : Bare_Type_Decl;
Cast_Result_19 : Bare_Base_Type_Decl;
Cast_Expr_19 : Bare_Type_Decl;
Cast_Result_20 : Bare_Base_Type_Decl;
Cast_Expr_20 : Bare_Type_Decl;
Cast_Result_21 : Bare_Base_Type_Decl;
Fld_11 : Internal_Env_Assoc := No_Env_Assoc;
Sym_7 : Symbol_Type;
Cast_Expr_21 : Bare_Type_Decl;
Cast_Result_22 : Bare_Base_Type_Decl;
Cast_Expr_22 : Bare_Type_Decl;
Cast_Result_23 : Bare_Base_Type_Decl;
Cast_Expr_23 : Bare_Type_Decl;
Cast_Result_24 : Bare_Base_Type_Decl;
Fld_12 : Internal_Env_Assoc := No_Env_Assoc;
Sym_8 : Symbol_Type;
Cast_Expr_24 : Bare_Type_Decl;
Cast_Result_25 : Bare_Base_Type_Decl;
Cast_Expr_25 : Bare_Type_Decl;
Cast_Result_26 : Bare_Base_Type_Decl;
Fld_13 : Internal_Env_Assoc := No_Env_Assoc;
Sym_9 : Symbol_Type;
Cast_Expr_26 : Bare_Type_Decl;
Cast_Result_27 : Bare_Base_Type_Decl;
Cast_Expr_27 : Bare_Type_Decl;
Cast_Result_28 : Bare_Base_Type_Decl;
Fld_14 : Internal_Env_Assoc := No_Env_Assoc;
Sym_10 : Symbol_Type;
Cast_Expr_28 : Bare_Type_Decl;
Cast_Result_29 : Bare_Base_Type_Decl;
Cast_Expr_29 : Bare_Type_Decl;
Cast_Result_30 : Bare_Base_Type_Decl;
Fld_15 : Internal_Env_Assoc := No_Env_Assoc;
Sym_11 : Symbol_Type;
Cast_Expr_30 : Bare_Type_Decl;
Cast_Result_31 : Bare_Base_Type_Decl;
Cast_Expr_31 : Bare_Type_Decl;
Cast_Result_32 : Bare_Base_Type_Decl;
Fld_16 : Internal_Env_Assoc := No_Env_Assoc;
Sym_12 : Symbol_Type;
Cast_Expr_32 : Bare_Type_Decl;
Cast_Result_33 : Bare_Base_Type_Decl;
Cast_Expr_33 : Bare_Type_Decl;
Cast_Result_34 : Bare_Base_Type_Decl;
Fld_17 : Internal_Env_Assoc := No_Env_Assoc;
Sym_13 : Symbol_Type;
Cast_Expr_34 : Bare_Type_Decl;
Cast_Result_35 : Bare_Base_Type_Decl;
Cast_Expr_35 : Bare_Type_Decl;
Cast_Result_36 : Bare_Base_Type_Decl;
Fld_18 : Internal_Env_Assoc := No_Env_Assoc;
Sym_14 : Symbol_Type;
Cast_Expr_36 : Bare_Type_Decl;
Cast_Result_37 : Bare_Base_Type_Decl;
Cast_Expr_37 : Bare_Type_Decl;
Cast_Result_38 : Bare_Base_Type_Decl;
Fld_19 : Internal_Env_Assoc := No_Env_Assoc;
Sym_15 : Symbol_Type;
Cast_Expr_38 : Bare_Type_Decl;
Cast_Result_39 : Bare_Base_Type_Decl;
Cast_Expr_39 : Bare_Type_Decl;
Cast_Result_40 : Bare_Base_Type_Decl;
Fld_20 : Internal_Env_Assoc := No_Env_Assoc;
Sym_16 : Symbol_Type;
Cast_Expr_40 : Bare_Type_Decl;
Cast_Result_41 : Bare_Base_Type_Decl;
Cast_Expr_41 : Bare_Type_Decl;
Cast_Result_42 : Bare_Base_Type_Decl;
Fld_21 : Internal_Env_Assoc := No_Env_Assoc;
Sym_17 : Symbol_Type;
Cast_Expr_42 : Bare_Type_Decl;
Cast_Result_43 : Bare_Base_Type_Decl;
Cast_Expr_43 : Bare_Type_Decl;
Cast_Result_44 : Bare_Base_Type_Decl;
Fld_22 : Internal_Env_Assoc := No_Env_Assoc;
Sym_18 : Symbol_Type;
Cast_Expr_44 : Bare_Type_Decl;
Cast_Result_45 : Bare_Base_Type_Decl;
Cast_Expr_45 : Bare_Type_Decl;
Cast_Result_46 : Bare_Base_Type_Decl;
Fld_23 : Internal_Env_Assoc := No_Env_Assoc;
Array_Lit : Internal_Env_Assoc_Array_Access;
Let_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_583 is
            begin
                     Dec_Ref (Fld_5);
                     Dec_Ref (Fld_6);
                     Dec_Ref (Fld_7);
                     Dec_Ref (Fld_8);
                     Dec_Ref (Fld_9);
                     Dec_Ref (Fld_10);
                     Dec_Ref (Fld_11);
                     Dec_Ref (Fld_12);
                     Dec_Ref (Fld_13);
                     Dec_Ref (Fld_14);
                     Dec_Ref (Fld_15);
                     Dec_Ref (Fld_16);
                     Dec_Ref (Fld_17);
                     Dec_Ref (Fld_18);
                     Dec_Ref (Fld_19);
                     Dec_Ref (Fld_20);
                     Dec_Ref (Fld_21);
                     Dec_Ref (Fld_22);
                     Dec_Ref (Fld_23);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_583;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Mod_Int_Type_Def_Mod_Int_Type_Def_P_Predefined_Operators,
            Items    => new Mmz_Key_Array (1 ..  1))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Mod_Int_Type_Def,
                                As_Bare_Mod_Int_Type_Def => Self);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Env_Assoc_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 11263 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11219 '<Cast to TypeDecl at ast.py:7647>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7647
--# expr-start 11218 '<parent at ast.py:7647>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7647
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 11218
   if Fld = null
      or else Fld.Kind in Ada_Type_Decl_Range
   then
      Cast_Result := Fld;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 11219
Block_Var := Cast_Result; 
--# bind self_type Block_Var
--# expr-start 11221 '<FieldAccess for node at ast.py:7648>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7648
--# expr-start 11220 '<FieldAccess for bool_type at ast.py:7648>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7648
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_1 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 11220
Fld_2 := Fld_1.Node;
--# expr-done 11221
Block_Var_1 := Fld_2; 
--# bind bool_type Block_Var_1
--# expr-start 11223 '<FieldAccess for node at ast.py:7649>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7649
--# expr-start 11222 '<FieldAccess for int_type at ast.py:7649>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7649
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.int_type
Fld_3 := Libadalang.Implementation.Ada_Node_P_Int_Type (Node => Self);
--# end
--# expr-done 11222
Fld_4 := Fld_3.Node;
--# expr-done 11223
Block_Var_2 := Fld_4; 
--# bind int_type Block_Var_2
--# expr-start 11262 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 11225 '<FieldAccess for create_binop_assoc at ast.py:7652>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7652
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11224 '<Symbol "+" at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_4); 
--# expr-done 11224
Cast_Expr := Block_Var; 
      Cast_Result_1 := Cast_Expr;
Cast_Expr_1 := Block_Var; 
      Cast_Result_2 := Cast_Expr_1;
Cast_Expr_2 := Block_Var; 
      Cast_Result_3 := Cast_Expr_2;
--# property-call-start AdaNode.create_binop_assoc
Fld_5 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym, Lhs => Cast_Result_1, Rhs => Cast_Result_2, Ret => Cast_Result_3);
--# end
--# expr-done 11225
--# expr-start 11227 '<FieldAccess for create_binop_assoc at ast.py:7653>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7653
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11226 '<Symbol "-" at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_5); 
--# expr-done 11226
Cast_Expr_3 := Block_Var; 
      Cast_Result_4 := Cast_Expr_3;
Cast_Expr_4 := Block_Var; 
      Cast_Result_5 := Cast_Expr_4;
Cast_Expr_5 := Block_Var; 
      Cast_Result_6 := Cast_Expr_5;
--# property-call-start AdaNode.create_binop_assoc
Fld_6 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_1, Lhs => Cast_Result_4, Rhs => Cast_Result_5, Ret => Cast_Result_6);
--# end
--# expr-done 11227
--# expr-start 11229 '<FieldAccess for create_binop_assoc at ast.py:7654>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7654
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11228 '<Symbol "*" at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_2); 
--# expr-done 11228
Cast_Expr_6 := Block_Var; 
      Cast_Result_7 := Cast_Expr_6;
Cast_Expr_7 := Block_Var; 
      Cast_Result_8 := Cast_Expr_7;
Cast_Expr_8 := Block_Var; 
      Cast_Result_9 := Cast_Expr_8;
--# property-call-start AdaNode.create_binop_assoc
Fld_7 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_2, Lhs => Cast_Result_7, Rhs => Cast_Result_8, Ret => Cast_Result_9);
--# end
--# expr-done 11229
--# expr-start 11231 '<FieldAccess for create_binop_assoc at ast.py:7655>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7655
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11230 '<Symbol "/" at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_6); 
--# expr-done 11230
Cast_Expr_9 := Block_Var; 
      Cast_Result_10 := Cast_Expr_9;
Cast_Expr_10 := Block_Var; 
      Cast_Result_11 := Cast_Expr_10;
Cast_Expr_11 := Block_Var; 
      Cast_Result_12 := Cast_Expr_11;
--# property-call-start AdaNode.create_binop_assoc
Fld_8 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_3, Lhs => Cast_Result_10, Rhs => Cast_Result_11, Ret => Cast_Result_12);
--# end
--# expr-done 11231
--# expr-start 11233 '<FieldAccess for create_binop_assoc at ast.py:7656>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7656
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11232 '<Symbol "mod" at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Mod); 
--# expr-done 11232
Cast_Expr_12 := Block_Var; 
      Cast_Result_13 := Cast_Expr_12;
Cast_Expr_13 := Block_Var; 
      Cast_Result_14 := Cast_Expr_13;
Cast_Expr_14 := Block_Var; 
      Cast_Result_15 := Cast_Expr_14;
--# property-call-start AdaNode.create_binop_assoc
Fld_9 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_4, Lhs => Cast_Result_13, Rhs => Cast_Result_14, Ret => Cast_Result_15);
--# end
--# expr-done 11233
--# expr-start 11235 '<FieldAccess for create_binop_assoc at ast.py:7657>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7657
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11234 '<Symbol "rem" at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Rem); 
--# expr-done 11234
Cast_Expr_15 := Block_Var; 
      Cast_Result_16 := Cast_Expr_15;
Cast_Expr_16 := Block_Var; 
      Cast_Result_17 := Cast_Expr_16;
Cast_Expr_17 := Block_Var; 
      Cast_Result_18 := Cast_Expr_17;
--# property-call-start AdaNode.create_binop_assoc
Fld_10 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_5, Lhs => Cast_Result_16, Rhs => Cast_Result_17, Ret => Cast_Result_18);
--# end
--# expr-done 11235
--# expr-start 11237 '<FieldAccess for create_binop_assoc at ast.py:7658>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7658
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11236 '<Symbol "and" at ???>' Sym_6 None
Sym_6 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_And); 
--# expr-done 11236
Cast_Expr_18 := Block_Var; 
      Cast_Result_19 := Cast_Expr_18;
Cast_Expr_19 := Block_Var; 
      Cast_Result_20 := Cast_Expr_19;
Cast_Expr_20 := Block_Var; 
      Cast_Result_21 := Cast_Expr_20;
--# property-call-start AdaNode.create_binop_assoc
Fld_11 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_6, Lhs => Cast_Result_19, Rhs => Cast_Result_20, Ret => Cast_Result_21);
--# end
--# expr-done 11237
--# expr-start 11239 '<FieldAccess for create_binop_assoc at ast.py:7659>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7659
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11238 '<Symbol "or" at ???>' Sym_7 None
Sym_7 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Or); 
--# expr-done 11238
Cast_Expr_21 := Block_Var; 
      Cast_Result_22 := Cast_Expr_21;
Cast_Expr_22 := Block_Var; 
      Cast_Result_23 := Cast_Expr_22;
Cast_Expr_23 := Block_Var; 
      Cast_Result_24 := Cast_Expr_23;
--# property-call-start AdaNode.create_binop_assoc
Fld_12 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_7, Lhs => Cast_Result_22, Rhs => Cast_Result_23, Ret => Cast_Result_24);
--# end
--# expr-done 11239
--# expr-start 11241 '<FieldAccess for create_binop_assoc at ast.py:7661>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7661
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11240 '<Symbol "**" at ???>' Sym_8 None
Sym_8 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_3); 
--# expr-done 11240
Cast_Expr_24 := Block_Var; 
      Cast_Result_25 := Cast_Expr_24;
Cast_Expr_25 := Block_Var; 
      Cast_Result_26 := Cast_Expr_25;
--# property-call-start AdaNode.create_binop_assoc
Fld_13 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_8, Lhs => Cast_Result_25, Rhs => Block_Var_2, Ret => Cast_Result_26);
--# end
--# expr-done 11241
--# expr-start 11243 '<FieldAccess for create_binop_assoc at ast.py:7663>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7663
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11242 '<Symbol "<" at ???>' Sym_9 None
Sym_9 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_8); 
--# expr-done 11242
Cast_Expr_26 := Block_Var; 
      Cast_Result_27 := Cast_Expr_26;
Cast_Expr_27 := Block_Var; 
      Cast_Result_28 := Cast_Expr_27;
--# property-call-start AdaNode.create_binop_assoc
Fld_14 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_9, Lhs => Cast_Result_27, Rhs => Cast_Result_28, Ret => Block_Var_1);
--# end
--# expr-done 11243
--# expr-start 11245 '<FieldAccess for create_binop_assoc at ast.py:7664>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7664
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11244 '<Symbol "<=" at ???>' Sym_10 None
Sym_10 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_9); 
--# expr-done 11244
Cast_Expr_28 := Block_Var; 
      Cast_Result_29 := Cast_Expr_28;
Cast_Expr_29 := Block_Var; 
      Cast_Result_30 := Cast_Expr_29;
--# property-call-start AdaNode.create_binop_assoc
Fld_15 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_10, Lhs => Cast_Result_29, Rhs => Cast_Result_30, Ret => Block_Var_1);
--# end
--# expr-done 11245
--# expr-start 11247 '<FieldAccess for create_binop_assoc at ast.py:7665>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7665
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11246 '<Symbol "=" at ???>' Sym_11 None
Sym_11 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_10); 
--# expr-done 11246
Cast_Expr_30 := Block_Var; 
      Cast_Result_31 := Cast_Expr_30;
Cast_Expr_31 := Block_Var; 
      Cast_Result_32 := Cast_Expr_31;
--# property-call-start AdaNode.create_binop_assoc
Fld_16 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_11, Lhs => Cast_Result_31, Rhs => Cast_Result_32, Ret => Block_Var_1);
--# end
--# expr-done 11247
--# expr-start 11249 '<FieldAccess for create_binop_assoc at ast.py:7666>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7666
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11248 '<Symbol "/=" at ???>' Sym_12 None
Sym_12 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_7); 
--# expr-done 11248
Cast_Expr_32 := Block_Var; 
      Cast_Result_33 := Cast_Expr_32;
Cast_Expr_33 := Block_Var; 
      Cast_Result_34 := Cast_Expr_33;
--# property-call-start AdaNode.create_binop_assoc
Fld_17 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_12, Lhs => Cast_Result_33, Rhs => Cast_Result_34, Ret => Block_Var_1);
--# end
--# expr-done 11249
--# expr-start 11251 '<FieldAccess for create_binop_assoc at ast.py:7667>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7667
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11250 '<Symbol ">" at ???>' Sym_13 None
Sym_13 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_11); 
--# expr-done 11250
Cast_Expr_34 := Block_Var; 
      Cast_Result_35 := Cast_Expr_34;
Cast_Expr_35 := Block_Var; 
      Cast_Result_36 := Cast_Expr_35;
--# property-call-start AdaNode.create_binop_assoc
Fld_18 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_13, Lhs => Cast_Result_35, Rhs => Cast_Result_36, Ret => Block_Var_1);
--# end
--# expr-done 11251
--# expr-start 11253 '<FieldAccess for create_binop_assoc at ast.py:7668>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7668
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11252 '<Symbol ">=" at ???>' Sym_14 None
Sym_14 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_12); 
--# expr-done 11252
Cast_Expr_36 := Block_Var; 
      Cast_Result_37 := Cast_Expr_36;
Cast_Expr_37 := Block_Var; 
      Cast_Result_38 := Cast_Expr_37;
--# property-call-start AdaNode.create_binop_assoc
Fld_19 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_14, Lhs => Cast_Result_37, Rhs => Cast_Result_38, Ret => Block_Var_1);
--# end
--# expr-done 11253
--# expr-start 11255 '<FieldAccess for create_unop_assoc at ast.py:7670>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7670
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11254 '<Symbol "+" at ???>' Sym_15 None
Sym_15 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_4); 
--# expr-done 11254
Cast_Expr_38 := Block_Var; 
      Cast_Result_39 := Cast_Expr_38;
Cast_Expr_39 := Block_Var; 
      Cast_Result_40 := Cast_Expr_39;
--# property-call-start AdaNode.create_unop_assoc
Fld_20 := Libadalang.Implementation.Ada_Node_P_Create_Unop_Assoc (Node => Self, Op => Sym_15, Rhs => Cast_Result_39, Ret => Cast_Result_40);
--# end
--# expr-done 11255
--# expr-start 11257 '<FieldAccess for create_unop_assoc at ast.py:7671>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7671
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11256 '<Symbol "-" at ???>' Sym_16 None
Sym_16 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_5); 
--# expr-done 11256
Cast_Expr_40 := Block_Var; 
      Cast_Result_41 := Cast_Expr_40;
Cast_Expr_41 := Block_Var; 
      Cast_Result_42 := Cast_Expr_41;
--# property-call-start AdaNode.create_unop_assoc
Fld_21 := Libadalang.Implementation.Ada_Node_P_Create_Unop_Assoc (Node => Self, Op => Sym_16, Rhs => Cast_Result_41, Ret => Cast_Result_42);
--# end
--# expr-done 11257
--# expr-start 11259 '<FieldAccess for create_unop_assoc at ast.py:7672>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7672
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11258 '<Symbol "abs" at ???>' Sym_17 None
Sym_17 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Abs); 
--# expr-done 11258
Cast_Expr_42 := Block_Var; 
      Cast_Result_43 := Cast_Expr_42;
Cast_Expr_43 := Block_Var; 
      Cast_Result_44 := Cast_Expr_43;
--# property-call-start AdaNode.create_unop_assoc
Fld_22 := Libadalang.Implementation.Ada_Node_P_Create_Unop_Assoc (Node => Self, Op => Sym_17, Rhs => Cast_Result_43, Ret => Cast_Result_44);
--# end
--# expr-done 11259
--# expr-start 11261 '<FieldAccess for create_unop_assoc at ast.py:7673>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7673
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11260 '<Symbol "not" at ???>' Sym_18 None
Sym_18 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Not); 
--# expr-done 11260
Cast_Expr_44 := Block_Var; 
      Cast_Result_45 := Cast_Expr_44;
Cast_Expr_45 := Block_Var; 
      Cast_Result_46 := Cast_Expr_45;
--# property-call-start AdaNode.create_unop_assoc
Fld_23 := Libadalang.Implementation.Ada_Node_P_Create_Unop_Assoc (Node => Self, Op => Sym_18, Rhs => Cast_Result_45, Ret => Cast_Result_46);
--# end
--# expr-done 11261
Array_Lit := Create_Internal_Env_Assoc_Array (Internal_Internal_Env_Assoc_Array'(1 => Fld_5, 2 => Fld_6, 3 => Fld_7, 4 => Fld_8, 5 => Fld_9, 6 => Fld_10, 7 => Fld_11, 8 => Fld_12, 9 => Fld_13, 10 => Fld_14, 11 => Fld_15, 12 => Fld_16, 13 => Fld_17, 14 => Fld_18, 15 => Fld_19, 16 => Fld_20, 17 => Fld_21, 18 => Fld_22, 19 => Fld_23)); 
--# expr-done 11262
Let_Result := Array_Lit; Inc_Ref (Let_Result);
--# end
--# expr-done 11263
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_583;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Env_Assoc_Array_Access,
                        As_Internal_Env_Assoc_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_583;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Mod_Int_Type_Def_P_Predefined_Operators;
--# end
   --
   --  Primitives for Bare_Private_Type_Def
   --
      procedure Initialize_Fields_For_Private_Type_Def
        (Self : Bare_Private_Type_Def
         ; Private_Type_Def_F_Has_Abstract : Bare_Abstract_Node
         ; Private_Type_Def_F_Has_Tagged : Bare_Tagged_Node
         ; Private_Type_Def_F_Has_Limited : Bare_Limited_Node
        ) is
      begin
            Self.Private_Type_Def_F_Has_Abstract := Private_Type_Def_F_Has_Abstract;
            Self.Private_Type_Def_F_Has_Tagged := Private_Type_Def_F_Has_Tagged;
            Self.Private_Type_Def_F_Has_Limited := Private_Type_Def_F_Has_Limited;
      end Initialize_Fields_For_Private_Type_Def;
   function Private_Type_Def_F_Has_Abstract
     (Node : Bare_Private_Type_Def) return Bare_Abstract_Node
   is
   begin
         return Node.Private_Type_Def_F_Has_Abstract;
   end;
   function Private_Type_Def_F_Has_Tagged
     (Node : Bare_Private_Type_Def) return Bare_Tagged_Node
   is
   begin
         return Node.Private_Type_Def_F_Has_Tagged;
   end;
   function Private_Type_Def_F_Has_Limited
     (Node : Bare_Private_Type_Def) return Bare_Limited_Node
   is
   begin
         return Node.Private_Type_Def_F_Has_Limited;
   end;
--# property-start PrivateTypeDef.is_tagged_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7570
pragma Warnings (Off, "is not referenced");
 function Private_Type_Def_P_Is_Tagged_Type
  (Node : Bare_Private_Type_Def
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Private_Type_Def := Bare_Private_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Bare_Tagged_Node;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11265 '<FieldAccess for as_bool at ast.py:7570>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7570
--# expr-start 11264 '<FieldAccess for has_tagged at ast.py:7570>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7570
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Private_Type_Def_F_Has_Tagged;
--# expr-done 11264
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Tagged.as_bool'
Fld_1 := Libadalang.Implementation.Dispatcher_Tagged_Node_P_As_Bool (Node => Fld);
--# end
--# expr-done 11265
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Private_Type_Def_P_Is_Tagged_Type;
--# end
--# property-start PrivateTypeDef.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7572
pragma Warnings (Off, "is not referenced");
 function Private_Type_Def_P_Xref_Equation
  (Node : Bare_Private_Type_Def
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Private_Type_Def := Bare_Private_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_572 with Inline_Always;
      True_Rel : Logic_Equation;
            procedure Finalizer_Scope_572 is
            begin
                     Dec_Ref (True_Rel);
            end Finalizer_Scope_572;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7572") else null)); 
      Property_Result := True_Rel;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_572;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_572;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Private_Type_Def_P_Xref_Equation;
--# end
--# property-start PrivateTypeDef.predefined_operators /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7575
pragma Warnings (Off, "is not referenced");
 function Private_Type_Def_P_Predefined_Operators
  (Node : Bare_Private_Type_Def
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Private_Type_Def := Bare_Private_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_573 with Inline_Always;
      Fld : Internal_Env_Assoc_Array_Access;
Fld_1 : Bare_Limited_Node;
Fld_2 : Boolean;
If_Result : Internal_Env_Assoc_Array_Access;
Let_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_573 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_573;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Private_Type_Def_Private_Type_Def_P_Predefined_Operators,
            Items    => new Mmz_Key_Array (1 ..  1))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Private_Type_Def,
                                As_Bare_Private_Type_Def => Self);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Env_Assoc_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 11270 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11269 '<If at ast.py:7576>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7576
--# expr-start 11267 '<FieldAccess for as_bool at ast.py:7577>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7577
--# expr-start 11266 '<FieldAccess for has_limited at ast.py:7577>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7577
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Self.Private_Type_Def_F_Has_Limited;
--# expr-done 11266
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Limited.as_bool'
Fld_2 := Libadalang.Implementation.Dispatcher_Limited_Node_P_As_Bool (Node => Fld_1);
--# end
--# expr-done 11267
if Fld_2 then
   If_Result := No_Internal_Env_Assoc_Array_Type;
else
   --# expr-start 11268 '<FieldAccess for predefined_equality_operators at ast.py:7579>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7579
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDef.predefined_equality_operators
Fld := Libadalang.Implementation.Type_Def_P_Predefined_Equality_Operators (Node => Self);
--# end
--# expr-done 11268
   If_Result := Fld;
end if;
      Inc_Ref (If_Result);
--# expr-done 11269
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 11270
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_573;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Env_Assoc_Array_Access,
                        As_Internal_Env_Assoc_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_573;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Private_Type_Def_P_Predefined_Operators;
--# end
   --
   --  Primitives for Bare_Real_Type_Def
   --
--# property-start RealTypeDef.is_static /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4997
pragma Warnings (Off, "is not referenced");
 function Real_Type_Def_P_Is_Static
  (Node : Bare_Real_Type_Def
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Real_Type_Def := Bare_Real_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Real_Type_Def_P_Is_Static;
--# end
--# property-start RealTypeDef.predefined_operators /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5000
pragma Warnings (Off, "is not referenced");
 function Real_Type_Def_P_Predefined_Operators
  (Node : Bare_Real_Type_Def
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Real_Type_Def := Bare_Real_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_363 with Inline_Always;
      Block_Var : Bare_Type_Decl;
Block_Var_1 : Bare_Base_Type_Decl;
Block_Var_2 : Bare_Base_Type_Decl;
Block_Var_3 : Bare_Base_Type_Decl;
Block_Var_4 : Internal_Env_Assoc_Array_Access;
Block_Var_5 : Internal_Env_Assoc_Array_Access;
Fld : Bare_Ada_Node;
Cast_Result : Bare_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Bare_Base_Type_Decl;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Bare_Base_Type_Decl;
Fld_5 : Internal_Entity_Base_Type_Decl;
Fld_6 : Bare_Base_Type_Decl;
Sym : Symbol_Type;
Cast_Expr : Bare_Type_Decl;
Cast_Result_1 : Bare_Base_Type_Decl;
Cast_Expr_1 : Bare_Type_Decl;
Cast_Result_2 : Bare_Base_Type_Decl;
Cast_Expr_2 : Bare_Type_Decl;
Cast_Result_3 : Bare_Base_Type_Decl;
Fld_7 : Internal_Env_Assoc := No_Env_Assoc;
Sym_1 : Symbol_Type;
Cast_Expr_3 : Bare_Type_Decl;
Cast_Result_4 : Bare_Base_Type_Decl;
Cast_Expr_4 : Bare_Type_Decl;
Cast_Result_5 : Bare_Base_Type_Decl;
Cast_Expr_5 : Bare_Type_Decl;
Cast_Result_6 : Bare_Base_Type_Decl;
Fld_8 : Internal_Env_Assoc := No_Env_Assoc;
Sym_2 : Symbol_Type;
Cast_Expr_6 : Bare_Type_Decl;
Cast_Result_7 : Bare_Base_Type_Decl;
Cast_Expr_7 : Bare_Type_Decl;
Cast_Result_8 : Bare_Base_Type_Decl;
Cast_Expr_8 : Bare_Type_Decl;
Cast_Result_9 : Bare_Base_Type_Decl;
Fld_9 : Internal_Env_Assoc := No_Env_Assoc;
Sym_3 : Symbol_Type;
Cast_Expr_9 : Bare_Type_Decl;
Cast_Result_10 : Bare_Base_Type_Decl;
Cast_Expr_10 : Bare_Type_Decl;
Cast_Result_11 : Bare_Base_Type_Decl;
Cast_Expr_11 : Bare_Type_Decl;
Cast_Result_12 : Bare_Base_Type_Decl;
Fld_10 : Internal_Env_Assoc := No_Env_Assoc;
Sym_4 : Symbol_Type;
Cast_Expr_12 : Bare_Type_Decl;
Cast_Result_13 : Bare_Base_Type_Decl;
Cast_Expr_13 : Bare_Type_Decl;
Cast_Result_14 : Bare_Base_Type_Decl;
Fld_11 : Internal_Env_Assoc := No_Env_Assoc;
Sym_5 : Symbol_Type;
Cast_Expr_14 : Bare_Type_Decl;
Cast_Result_15 : Bare_Base_Type_Decl;
Cast_Expr_15 : Bare_Type_Decl;
Cast_Result_16 : Bare_Base_Type_Decl;
Fld_12 : Internal_Env_Assoc := No_Env_Assoc;
Sym_6 : Symbol_Type;
Cast_Expr_16 : Bare_Type_Decl;
Cast_Result_17 : Bare_Base_Type_Decl;
Cast_Expr_17 : Bare_Type_Decl;
Cast_Result_18 : Bare_Base_Type_Decl;
Fld_13 : Internal_Env_Assoc := No_Env_Assoc;
Sym_7 : Symbol_Type;
Cast_Expr_18 : Bare_Type_Decl;
Cast_Result_19 : Bare_Base_Type_Decl;
Cast_Expr_19 : Bare_Type_Decl;
Cast_Result_20 : Bare_Base_Type_Decl;
Fld_14 : Internal_Env_Assoc := No_Env_Assoc;
Sym_8 : Symbol_Type;
Cast_Expr_20 : Bare_Type_Decl;
Cast_Result_21 : Bare_Base_Type_Decl;
Cast_Expr_21 : Bare_Type_Decl;
Cast_Result_22 : Bare_Base_Type_Decl;
Fld_15 : Internal_Env_Assoc := No_Env_Assoc;
Sym_9 : Symbol_Type;
Cast_Expr_22 : Bare_Type_Decl;
Cast_Result_23 : Bare_Base_Type_Decl;
Cast_Expr_23 : Bare_Type_Decl;
Cast_Result_24 : Bare_Base_Type_Decl;
Fld_16 : Internal_Env_Assoc := No_Env_Assoc;
Sym_10 : Symbol_Type;
Cast_Expr_24 : Bare_Type_Decl;
Cast_Result_25 : Bare_Base_Type_Decl;
Cast_Expr_25 : Bare_Type_Decl;
Cast_Result_26 : Bare_Base_Type_Decl;
Fld_17 : Internal_Env_Assoc := No_Env_Assoc;
Sym_11 : Symbol_Type;
Cast_Expr_26 : Bare_Type_Decl;
Cast_Result_27 : Bare_Base_Type_Decl;
Cast_Expr_27 : Bare_Type_Decl;
Cast_Result_28 : Bare_Base_Type_Decl;
Fld_18 : Internal_Env_Assoc := No_Env_Assoc;
Sym_12 : Symbol_Type;
Cast_Expr_28 : Bare_Type_Decl;
Cast_Result_29 : Bare_Base_Type_Decl;
Cast_Expr_29 : Bare_Type_Decl;
Cast_Result_30 : Bare_Base_Type_Decl;
Fld_19 : Internal_Env_Assoc := No_Env_Assoc;
Sym_13 : Symbol_Type;
Cast_Expr_30 : Bare_Type_Decl;
Cast_Result_31 : Bare_Base_Type_Decl;
Cast_Expr_31 : Bare_Type_Decl;
Cast_Result_32 : Bare_Base_Type_Decl;
Fld_20 : Internal_Env_Assoc := No_Env_Assoc;
Array_Lit : Internal_Env_Assoc_Array_Access;
Sym_14 : Symbol_Type;
Cast_Expr_32 : Bare_Type_Decl;
Cast_Result_33 : Bare_Base_Type_Decl;
Cast_Expr_33 : Bare_Type_Decl;
Cast_Result_34 : Bare_Base_Type_Decl;
Fld_21 : Internal_Env_Assoc := No_Env_Assoc;
Sym_15 : Symbol_Type;
Cast_Expr_34 : Bare_Type_Decl;
Cast_Result_35 : Bare_Base_Type_Decl;
Cast_Expr_35 : Bare_Type_Decl;
Cast_Result_36 : Bare_Base_Type_Decl;
Fld_22 : Internal_Env_Assoc := No_Env_Assoc;
Sym_16 : Symbol_Type;
Cast_Expr_36 : Bare_Type_Decl;
Cast_Result_37 : Bare_Base_Type_Decl;
Cast_Expr_37 : Bare_Type_Decl;
Cast_Result_38 : Bare_Base_Type_Decl;
Fld_23 : Internal_Env_Assoc := No_Env_Assoc;
Array_Lit_1 : Internal_Env_Assoc_Array_Access;
Fld_24 : Internal_Entity_Base_Type_Decl;
Fld_25 : Bare_Base_Type_Decl;
Cast_Expr_38 : Bare_Type_Decl;
Cast_Result_39 : Bare_Base_Type_Decl;
Is_Equal : Boolean;
If_Result : Internal_Env_Assoc_Array_Access;
Concat_Result : Internal_Env_Assoc_Array_Access;
Let_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_363 is
            begin
                     Dec_Ref (Fld_7);
                     Dec_Ref (Fld_8);
                     Dec_Ref (Fld_9);
                     Dec_Ref (Fld_10);
                     Dec_Ref (Fld_11);
                     Dec_Ref (Fld_12);
                     Dec_Ref (Fld_13);
                     Dec_Ref (Fld_14);
                     Dec_Ref (Fld_15);
                     Dec_Ref (Fld_16);
                     Dec_Ref (Fld_17);
                     Dec_Ref (Fld_18);
                     Dec_Ref (Fld_19);
                     Dec_Ref (Fld_20);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Block_Var_4);
                     Dec_Ref (Fld_21);
                     Dec_Ref (Fld_22);
                     Dec_Ref (Fld_23);
                     Dec_Ref (Array_Lit_1);
                     Dec_Ref (If_Result);
                     Dec_Ref (Block_Var_5);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_363;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Real_Type_Def_Real_Type_Def_P_Predefined_Operators,
            Items    => new Mmz_Key_Array (1 ..  1))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Real_Type_Def,
                                As_Bare_Real_Type_Def => Self);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Env_Assoc_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 11320 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11272 '<Cast to TypeDecl at ast.py:5001>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5001
--# expr-start 11271 '<parent at ast.py:5001>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5001
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 11271
   if Fld = null
      or else Fld.Kind in Ada_Type_Decl_Range
   then
      Cast_Result := Fld;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 11272
Block_Var := Cast_Result; 
--# bind self_type Block_Var
--# expr-start 11274 '<FieldAccess for node at ast.py:5002>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5002
--# expr-start 11273 '<FieldAccess for bool_type at ast.py:5002>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5002
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_1 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 11273
Fld_2 := Fld_1.Node;
--# expr-done 11274
Block_Var_1 := Fld_2; 
--# bind bool_type Block_Var_1
--# expr-start 11276 '<FieldAccess for node at ast.py:5003>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5003
--# expr-start 11275 '<FieldAccess for int_type at ast.py:5003>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5003
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.int_type
Fld_3 := Libadalang.Implementation.Ada_Node_P_Int_Type (Node => Self);
--# end
--# expr-done 11275
Fld_4 := Fld_3.Node;
--# expr-done 11276
Block_Var_2 := Fld_4; 
--# bind int_type Block_Var_2
--# expr-start 11278 '<FieldAccess for node at ast.py:5004>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5004
--# expr-start 11277 '<FieldAccess for root_int_type at ast.py:5004>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5004
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.root_int_type
Fld_5 := Libadalang.Implementation.Ada_Node_P_Root_Int_Type (Node => Self);
--# end
--# expr-done 11277
Fld_6 := Fld_5.Node;
--# expr-done 11278
Block_Var_3 := Fld_6; 
--# bind root_int_type Block_Var_3
--# expr-start 11307 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 11280 '<FieldAccess for create_binop_assoc at ast.py:5007>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5007
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11279 '<Symbol "+" at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_4); 
--# expr-done 11279
Cast_Expr := Block_Var; 
      Cast_Result_1 := Cast_Expr;
Cast_Expr_1 := Block_Var; 
      Cast_Result_2 := Cast_Expr_1;
Cast_Expr_2 := Block_Var; 
      Cast_Result_3 := Cast_Expr_2;
--# property-call-start AdaNode.create_binop_assoc
Fld_7 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym, Lhs => Cast_Result_1, Rhs => Cast_Result_2, Ret => Cast_Result_3);
--# end
--# expr-done 11280
--# expr-start 11282 '<FieldAccess for create_binop_assoc at ast.py:5008>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5008
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11281 '<Symbol "-" at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_5); 
--# expr-done 11281
Cast_Expr_3 := Block_Var; 
      Cast_Result_4 := Cast_Expr_3;
Cast_Expr_4 := Block_Var; 
      Cast_Result_5 := Cast_Expr_4;
Cast_Expr_5 := Block_Var; 
      Cast_Result_6 := Cast_Expr_5;
--# property-call-start AdaNode.create_binop_assoc
Fld_8 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_1, Lhs => Cast_Result_4, Rhs => Cast_Result_5, Ret => Cast_Result_6);
--# end
--# expr-done 11282
--# expr-start 11284 '<FieldAccess for create_binop_assoc at ast.py:5009>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5009
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11283 '<Symbol "*" at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_2); 
--# expr-done 11283
Cast_Expr_6 := Block_Var; 
      Cast_Result_7 := Cast_Expr_6;
Cast_Expr_7 := Block_Var; 
      Cast_Result_8 := Cast_Expr_7;
Cast_Expr_8 := Block_Var; 
      Cast_Result_9 := Cast_Expr_8;
--# property-call-start AdaNode.create_binop_assoc
Fld_9 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_2, Lhs => Cast_Result_7, Rhs => Cast_Result_8, Ret => Cast_Result_9);
--# end
--# expr-done 11284
--# expr-start 11286 '<FieldAccess for create_binop_assoc at ast.py:5010>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5010
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11285 '<Symbol "/" at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_6); 
--# expr-done 11285
Cast_Expr_9 := Block_Var; 
      Cast_Result_10 := Cast_Expr_9;
Cast_Expr_10 := Block_Var; 
      Cast_Result_11 := Cast_Expr_10;
Cast_Expr_11 := Block_Var; 
      Cast_Result_12 := Cast_Expr_11;
--# property-call-start AdaNode.create_binop_assoc
Fld_10 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_3, Lhs => Cast_Result_10, Rhs => Cast_Result_11, Ret => Cast_Result_12);
--# end
--# expr-done 11286
--# expr-start 11288 '<FieldAccess for create_binop_assoc at ast.py:5012>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5012
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11287 '<Symbol "**" at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_3); 
--# expr-done 11287
Cast_Expr_12 := Block_Var; 
      Cast_Result_13 := Cast_Expr_12;
Cast_Expr_13 := Block_Var; 
      Cast_Result_14 := Cast_Expr_13;
--# property-call-start AdaNode.create_binop_assoc
Fld_11 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_4, Lhs => Cast_Result_13, Rhs => Block_Var_2, Ret => Cast_Result_14);
--# end
--# expr-done 11288
--# expr-start 11290 '<FieldAccess for create_binop_assoc at ast.py:5014>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5014
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11289 '<Symbol "<" at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_8); 
--# expr-done 11289
Cast_Expr_14 := Block_Var; 
      Cast_Result_15 := Cast_Expr_14;
Cast_Expr_15 := Block_Var; 
      Cast_Result_16 := Cast_Expr_15;
--# property-call-start AdaNode.create_binop_assoc
Fld_12 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_5, Lhs => Cast_Result_15, Rhs => Cast_Result_16, Ret => Block_Var_1);
--# end
--# expr-done 11290
--# expr-start 11292 '<FieldAccess for create_binop_assoc at ast.py:5015>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5015
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11291 '<Symbol "<=" at ???>' Sym_6 None
Sym_6 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_9); 
--# expr-done 11291
Cast_Expr_16 := Block_Var; 
      Cast_Result_17 := Cast_Expr_16;
Cast_Expr_17 := Block_Var; 
      Cast_Result_18 := Cast_Expr_17;
--# property-call-start AdaNode.create_binop_assoc
Fld_13 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_6, Lhs => Cast_Result_17, Rhs => Cast_Result_18, Ret => Block_Var_1);
--# end
--# expr-done 11292
--# expr-start 11294 '<FieldAccess for create_binop_assoc at ast.py:5016>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5016
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11293 '<Symbol "=" at ???>' Sym_7 None
Sym_7 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_10); 
--# expr-done 11293
Cast_Expr_18 := Block_Var; 
      Cast_Result_19 := Cast_Expr_18;
Cast_Expr_19 := Block_Var; 
      Cast_Result_20 := Cast_Expr_19;
--# property-call-start AdaNode.create_binop_assoc
Fld_14 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_7, Lhs => Cast_Result_19, Rhs => Cast_Result_20, Ret => Block_Var_1);
--# end
--# expr-done 11294
--# expr-start 11296 '<FieldAccess for create_binop_assoc at ast.py:5017>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5017
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11295 '<Symbol "/=" at ???>' Sym_8 None
Sym_8 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_7); 
--# expr-done 11295
Cast_Expr_20 := Block_Var; 
      Cast_Result_21 := Cast_Expr_20;
Cast_Expr_21 := Block_Var; 
      Cast_Result_22 := Cast_Expr_21;
--# property-call-start AdaNode.create_binop_assoc
Fld_15 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_8, Lhs => Cast_Result_21, Rhs => Cast_Result_22, Ret => Block_Var_1);
--# end
--# expr-done 11296
--# expr-start 11298 '<FieldAccess for create_binop_assoc at ast.py:5018>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5018
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11297 '<Symbol ">" at ???>' Sym_9 None
Sym_9 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_11); 
--# expr-done 11297
Cast_Expr_22 := Block_Var; 
      Cast_Result_23 := Cast_Expr_22;
Cast_Expr_23 := Block_Var; 
      Cast_Result_24 := Cast_Expr_23;
--# property-call-start AdaNode.create_binop_assoc
Fld_16 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_9, Lhs => Cast_Result_23, Rhs => Cast_Result_24, Ret => Block_Var_1);
--# end
--# expr-done 11298
--# expr-start 11300 '<FieldAccess for create_binop_assoc at ast.py:5019>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5019
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11299 '<Symbol ">=" at ???>' Sym_10 None
Sym_10 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_12); 
--# expr-done 11299
Cast_Expr_24 := Block_Var; 
      Cast_Result_25 := Cast_Expr_24;
Cast_Expr_25 := Block_Var; 
      Cast_Result_26 := Cast_Expr_25;
--# property-call-start AdaNode.create_binop_assoc
Fld_17 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_10, Lhs => Cast_Result_25, Rhs => Cast_Result_26, Ret => Block_Var_1);
--# end
--# expr-done 11300
--# expr-start 11302 '<FieldAccess for create_unop_assoc at ast.py:5021>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5021
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11301 '<Symbol "+" at ???>' Sym_11 None
Sym_11 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_4); 
--# expr-done 11301
Cast_Expr_26 := Block_Var; 
      Cast_Result_27 := Cast_Expr_26;
Cast_Expr_27 := Block_Var; 
      Cast_Result_28 := Cast_Expr_27;
--# property-call-start AdaNode.create_unop_assoc
Fld_18 := Libadalang.Implementation.Ada_Node_P_Create_Unop_Assoc (Node => Self, Op => Sym_11, Rhs => Cast_Result_27, Ret => Cast_Result_28);
--# end
--# expr-done 11302
--# expr-start 11304 '<FieldAccess for create_unop_assoc at ast.py:5022>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5022
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11303 '<Symbol "-" at ???>' Sym_12 None
Sym_12 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_5); 
--# expr-done 11303
Cast_Expr_28 := Block_Var; 
      Cast_Result_29 := Cast_Expr_28;
Cast_Expr_29 := Block_Var; 
      Cast_Result_30 := Cast_Expr_29;
--# property-call-start AdaNode.create_unop_assoc
Fld_19 := Libadalang.Implementation.Ada_Node_P_Create_Unop_Assoc (Node => Self, Op => Sym_12, Rhs => Cast_Result_29, Ret => Cast_Result_30);
--# end
--# expr-done 11304
--# expr-start 11306 '<FieldAccess for create_unop_assoc at ast.py:5023>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5023
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11305 '<Symbol "abs" at ???>' Sym_13 None
Sym_13 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Abs); 
--# expr-done 11305
Cast_Expr_30 := Block_Var; 
      Cast_Result_31 := Cast_Expr_30;
Cast_Expr_31 := Block_Var; 
      Cast_Result_32 := Cast_Expr_31;
--# property-call-start AdaNode.create_unop_assoc
Fld_20 := Libadalang.Implementation.Ada_Node_P_Create_Unop_Assoc (Node => Self, Op => Sym_13, Rhs => Cast_Result_31, Ret => Cast_Result_32);
--# end
--# expr-done 11306
Array_Lit := Create_Internal_Env_Assoc_Array (Internal_Internal_Env_Assoc_Array'(1 => Fld_7, 2 => Fld_8, 3 => Fld_9, 4 => Fld_10, 5 => Fld_11, 6 => Fld_12, 7 => Fld_13, 8 => Fld_14, 9 => Fld_15, 10 => Fld_16, 11 => Fld_17, 12 => Fld_18, 13 => Fld_19, 14 => Fld_20)); 
--# expr-done 11307
Block_Var_4 := Array_Lit; Inc_Ref (Block_Var_4);
--# bind defaults Block_Var_4
--# expr-start 11318 '<If at ast.py:5027>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5027
--# expr-start 11310 '<Eq at ast.py:5028>' Is_Equal /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5028
Cast_Expr_38 := Block_Var; 
      Cast_Result_39 := Cast_Expr_38;
--# expr-start 11309 '<FieldAccess for node at ast.py:5028>' Fld_25 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5028
--# expr-start 11308 '<FieldAccess for root_real_type at ast.py:5028>' Fld_24 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5028
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.root_real_type
Fld_24 := Libadalang.Implementation.Ada_Node_P_Root_Real_Type (Node => Self);
--# end
--# expr-done 11308
Fld_25 := Fld_24.Node;
--# expr-done 11309
Is_Equal := Cast_Result_39 = Fld_25; 
--# expr-done 11310
if Is_Equal then
   --# expr-start 11317 '<ArrayLiteral at ???>' Array_Lit_1 None
--# expr-start 11312 '<FieldAccess for create_binop_assoc at ast.py:5030>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5030
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11311 '<Symbol "*" at ???>' Sym_14 None
Sym_14 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_2); 
--# expr-done 11311
Cast_Expr_32 := Block_Var; 
      Cast_Result_33 := Cast_Expr_32;
Cast_Expr_33 := Block_Var; 
      Cast_Result_34 := Cast_Expr_33;
--# property-call-start AdaNode.create_binop_assoc
Fld_21 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_14, Lhs => Block_Var_3, Rhs => Cast_Result_33, Ret => Cast_Result_34);
--# end
--# expr-done 11312
--# expr-start 11314 '<FieldAccess for create_binop_assoc at ast.py:5033>' Fld_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5033
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11313 '<Symbol "*" at ???>' Sym_15 None
Sym_15 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_2); 
--# expr-done 11313
Cast_Expr_34 := Block_Var; 
      Cast_Result_35 := Cast_Expr_34;
Cast_Expr_35 := Block_Var; 
      Cast_Result_36 := Cast_Expr_35;
--# property-call-start AdaNode.create_binop_assoc
Fld_22 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_15, Lhs => Cast_Result_35, Rhs => Block_Var_3, Ret => Cast_Result_36);
--# end
--# expr-done 11314
--# expr-start 11316 '<FieldAccess for create_binop_assoc at ast.py:5036>' Fld_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5036
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11315 '<Symbol "/" at ???>' Sym_16 None
Sym_16 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_6); 
--# expr-done 11315
Cast_Expr_36 := Block_Var; 
      Cast_Result_37 := Cast_Expr_36;
Cast_Expr_37 := Block_Var; 
      Cast_Result_38 := Cast_Expr_37;
--# property-call-start AdaNode.create_binop_assoc
Fld_23 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_16, Lhs => Cast_Result_37, Rhs => Block_Var_3, Ret => Cast_Result_38);
--# end
--# expr-done 11316
Array_Lit_1 := Create_Internal_Env_Assoc_Array (Internal_Internal_Env_Assoc_Array'(1 => Fld_21, 2 => Fld_22, 3 => Fld_23)); 
--# expr-done 11317
   If_Result := Array_Lit_1;
else
   If_Result := No_Internal_Env_Assoc_Array_Type;
end if;
      Inc_Ref (If_Result);
--# expr-done 11318
Block_Var_5 := If_Result; Inc_Ref (Block_Var_5);
--# bind specials Block_Var_5
--# expr-start 11319 '<Concat at ast.py:5043>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:5043
Concat_Result := Concat (Block_Var_4, Block_Var_5); 
--# expr-done 11319
Let_Result := Concat_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 11320
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_363;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Env_Assoc_Array_Access,
                        As_Internal_Env_Assoc_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_363;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Real_Type_Def_P_Predefined_Operators;
--# end
   --
   --  Primitives for Bare_Decimal_Fixed_Point_Def
   --
      procedure Initialize_Fields_For_Decimal_Fixed_Point_Def
        (Self : Bare_Decimal_Fixed_Point_Def
         ; Decimal_Fixed_Point_Def_F_Delta : Bare_Expr
         ; Decimal_Fixed_Point_Def_F_Digits : Bare_Expr
         ; Decimal_Fixed_Point_Def_F_Range : Bare_Range_Spec
        ) is
      begin
            Self.Decimal_Fixed_Point_Def_F_Delta := Decimal_Fixed_Point_Def_F_Delta;
            Self.Decimal_Fixed_Point_Def_F_Digits := Decimal_Fixed_Point_Def_F_Digits;
            Self.Decimal_Fixed_Point_Def_F_Range := Decimal_Fixed_Point_Def_F_Range;
      end Initialize_Fields_For_Decimal_Fixed_Point_Def;
   function Decimal_Fixed_Point_Def_F_Delta
     (Node : Bare_Decimal_Fixed_Point_Def) return Bare_Expr
   is
   begin
         return Node.Decimal_Fixed_Point_Def_F_Delta;
   end;
   function Decimal_Fixed_Point_Def_F_Digits
     (Node : Bare_Decimal_Fixed_Point_Def) return Bare_Expr
   is
   begin
         return Node.Decimal_Fixed_Point_Def_F_Digits;
   end;
   function Decimal_Fixed_Point_Def_F_Range
     (Node : Bare_Decimal_Fixed_Point_Def) return Bare_Range_Spec
   is
   begin
         return Node.Decimal_Fixed_Point_Def_F_Range;
   end;
--# property-start DecimalFixedPointDef.is_fixed_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7236
pragma Warnings (Off, "is not referenced");
 function Decimal_Fixed_Point_Def_P_Is_Fixed_Point
  (Node : Bare_Decimal_Fixed_Point_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Decimal_Fixed_Point_Def := Bare_Decimal_Fixed_Point_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Decimal_Fixed_Point_Def_P_Is_Fixed_Point;
--# end
--# property-start DecimalFixedPointDef.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7239
pragma Warnings (Off, "is not referenced");
 function Decimal_Fixed_Point_Def_P_Xref_Equation
  (Node : Bare_Decimal_Fixed_Point_Def
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Decimal_Fixed_Point_Def := Bare_Decimal_Fixed_Point_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Decimal_Fixed_Point_Def :=
     Internal_Entity_Decimal_Fixed_Point_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_541 with Inline_Always;
            procedure Finalizer_Scope_3097 with Inline_Always;
      Var_Expr : Internal_Entity_Range_Spec;
Fld : Internal_Entity_Expr;
Fld_1 : Logic_Var;
Fld_2 : Logic_Equation;
Fld_3 : Internal_Entity_Expr;
Fld_4 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_5 : Internal_Entity_Expr;
Fld_6 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_7 : Internal_Entity_Expr;
Fld_8 : Logic_Var;
Fld_9 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_10 : Internal_Entity_Expr;
Fld_11 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Fld_12 : Internal_Entity_Expr;
Fld_13 : Logic_Equation;
And_Pred_4 : Logic_Equation;
Fld_14 : Internal_Entity_Range_Spec;
Fld_15 : Internal_Entity_Expr;
Fld_16 : Logic_Var;
Fld_17 : Logic_Equation;
Fld_18 : Internal_Entity_Expr;
Fld_19 : Logic_Equation;
And_Pred_5 : Logic_Equation;
Fld_20 : Internal_Entity_Expr;
Fld_21 : Logic_Equation;
And_Pred_6 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred_7 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_541 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Fld_4);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Fld_9);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Fld_11);
                     Dec_Ref (And_Pred_3);
                     Dec_Ref (Fld_13);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred_7);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_541;
            procedure Finalizer_Scope_3097 is
            begin
                     Dec_Ref (Fld_17);
                     Dec_Ref (Fld_19);
                     Dec_Ref (And_Pred_5);
                     Dec_Ref (Fld_21);
                     Dec_Ref (And_Pred_6);
            end Finalizer_Scope_3097;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11351 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11350 '<And at ast.py:7243>' And_Pred_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7243
--# expr-start 11339 '<And at ast.py:7243>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7243
--# expr-start 11336 '<And at ast.py:7243>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7243
--# expr-start 11333 '<And at ast.py:7243>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7243
--# expr-start 11329 '<And at ast.py:7243>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7243
--# expr-start 11326 '<And at ast.py:7243>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7243
--# expr-start 11323 '<FieldAccess for universal_real_bind at ast.py:7246>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7246
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11322 '<FieldAccess for expected_type_var at ast.py:7246>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7246
--# expr-start 11321 '<FieldAccess for delta at ast.py:7246>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7246
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Decimal_Fixed_Point_Def_F_Delta, Info => Ent.Info);
--# expr-done 11321
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld.Node);
--# expr-done 11322
--# property-call-start AdaNode.universal_real_bind
Fld_2 := Libadalang.Implementation.Ada_Node_P_Universal_Real_Bind (Node => Ent.Node, Type_Var => Fld_1, Origin => Origin);
--# end
--# expr-done 11323
--# expr-start 11325 '<FieldAccess for sub_equation at ast.py:7247>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7247
--# expr-start 11324 '<FieldAccess for delta at ast.py:7247>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7247
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Expr (Node => Ent.Node.Decimal_Fixed_Point_Def_F_Delta, Info => Ent.Info);
--# expr-done 11324
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_4 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_3.Node, Env => Env, Origin => Origin, E_Info => Fld_3.Info);
--# end
--# expr-done 11325
And_Pred := Create_And (Fld_2, Fld_4, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7243") else null)); 
--# expr-done 11326
--# expr-start 11328 '<FieldAccess for matches_expected_type at ast.py:7248>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7248
--# expr-start 11327 '<FieldAccess for delta at ast.py:7248>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7248
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Expr (Node => Ent.Node.Decimal_Fixed_Point_Def_F_Delta, Info => Ent.Info);
--# expr-done 11327
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_6 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_5.Node, Origin => Origin);
--# end
--# expr-done 11328
And_Pred_1 := Create_And (And_Pred, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7243") else null)); 
--# expr-done 11329
--# expr-start 11332 '<FieldAccess for universal_int_bind at ast.py:7251>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7251
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11331 '<FieldAccess for expected_type_var at ast.py:7251>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7251
--# expr-start 11330 '<FieldAccess for digits at ast.py:7251>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7251
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Expr (Node => Ent.Node.Decimal_Fixed_Point_Def_F_Digits, Info => Ent.Info);
--# expr-done 11330
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_7.Node);
--# expr-done 11331
--# property-call-start AdaNode.universal_int_bind
Fld_9 := Libadalang.Implementation.Ada_Node_P_Universal_Int_Bind (Node => Ent.Node, Type_Var => Fld_8, Origin => Origin);
--# end
--# expr-done 11332
And_Pred_2 := Create_And (And_Pred_1, Fld_9, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7243") else null)); 
--# expr-done 11333
--# expr-start 11335 '<FieldAccess for sub_equation at ast.py:7252>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7252
--# expr-start 11334 '<FieldAccess for digits at ast.py:7252>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7252
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Create_Internal_Entity_Expr (Node => Ent.Node.Decimal_Fixed_Point_Def_F_Digits, Info => Ent.Info);
--# expr-done 11334
      if Fld_10.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_11 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_10.Node, Env => Env, Origin => Origin, E_Info => Fld_10.Info);
--# end
--# expr-done 11335
And_Pred_3 := Create_And (And_Pred_2, Fld_11, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7243") else null)); 
--# expr-done 11336
--# expr-start 11338 '<FieldAccess for matches_expected_type at ast.py:7253>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7253
--# expr-start 11337 '<FieldAccess for digits at ast.py:7253>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7253
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_12 := Create_Internal_Entity_Expr (Node => Ent.Node.Decimal_Fixed_Point_Def_F_Digits, Info => Ent.Info);
--# expr-done 11337
      if Fld_12.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_13 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_12.Node, Origin => Origin);
--# end
--# expr-done 11338
And_Pred_4 := Create_And (And_Pred_3, Fld_13, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7243") else null)); 
--# expr-done 11339
--# expr-start 11340 '<FieldAccess for range at ast.py:7257>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7257
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_14 := Create_Internal_Entity_Range_Spec (Node => Ent.Node.Decimal_Fixed_Point_Def_F_Range, Info => Ent.Info);
--# expr-done 11340
Var_Expr := Fld_14;
if Var_Expr /= No_Entity_Range_Spec then
   --# scope-start
   --# bind r Var_Expr
--# expr-start 11349 '<And at ast.py:7259>' And_Pred_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7259
--# expr-start 11346 '<And at ast.py:7259>' And_Pred_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7259
--# expr-start 11343 '<FieldAccess for universal_real_bind at ast.py:7259>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7259
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11342 '<FieldAccess for expected_type_var at ast.py:7259>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7259
--# expr-start 11341 '<FieldAccess for range at ast.py:7259>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7259
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_15 := Create_Internal_Entity_Expr (Node => Var_Expr.Node.Range_Spec_F_Range, Info => Var_Expr.Info);
--# expr-done 11341
      if Fld_15.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_16 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_15.Node);
--# expr-done 11342
--# property-call-start AdaNode.universal_real_bind
Fld_17 := Libadalang.Implementation.Ada_Node_P_Universal_Real_Bind (Node => Ent.Node, Type_Var => Fld_16, Origin => Origin);
--# end
--# expr-done 11343
--# expr-start 11345 '<FieldAccess for sub_equation at ast.py:7260>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7260
--# expr-start 11344 '<FieldAccess for range at ast.py:7260>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7260
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_18 := Create_Internal_Entity_Expr (Node => Var_Expr.Node.Range_Spec_F_Range, Info => Var_Expr.Info);
--# expr-done 11344
      if Fld_18.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_19 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_18.Node, Env => Env, Origin => Origin, E_Info => Fld_18.Info);
--# end
--# expr-done 11345
And_Pred_5 := Create_And (Fld_17, Fld_19, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7259") else null)); 
--# expr-done 11346
--# expr-start 11348 '<FieldAccess for matches_expected_type at ast.py:7261>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7261
--# expr-start 11347 '<FieldAccess for range at ast.py:7261>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7261
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_20 := Create_Internal_Entity_Expr (Node => Var_Expr.Node.Range_Spec_F_Range, Info => Var_Expr.Info);
--# expr-done 11347
      if Fld_20.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_21 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_20.Node, Origin => Origin);
--# end
--# expr-done 11348
And_Pred_6 := Create_And (And_Pred_5, Fld_21, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7259") else null)); 
--# expr-done 11349
Scope_Result := And_Pred_6;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_3097;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7262") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred_7 := Create_And (And_Pred_4, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7243") else null)); 
--# expr-done 11350
Let_Result := And_Pred_7; Inc_Ref (Let_Result);
--# end
--# expr-done 11351
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_541;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_541;
               Finalizer_Scope_3097;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Decimal_Fixed_Point_Def_P_Xref_Equation;
--# end
--# property-start DecimalFixedPointDef.predefined_operators /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7267
pragma Warnings (Off, "is not referenced");
 function Decimal_Fixed_Point_Def_P_Predefined_Operators
  (Node : Bare_Decimal_Fixed_Point_Def
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Decimal_Fixed_Point_Def := Bare_Decimal_Fixed_Point_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_542 with Inline_Always;
      Block_Var : Bare_Type_Decl;
Block_Var_1 : Bare_Base_Type_Decl;
Block_Var_2 : Bare_Base_Type_Decl;
Fld : Bare_Ada_Node;
Cast_Result : Bare_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Bare_Base_Type_Decl;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Bare_Base_Type_Decl;
Sym : Symbol_Type;
Cast_Expr : Bare_Type_Decl;
Cast_Result_1 : Bare_Base_Type_Decl;
Cast_Expr_1 : Bare_Type_Decl;
Cast_Result_2 : Bare_Base_Type_Decl;
Cast_Expr_2 : Bare_Type_Decl;
Cast_Result_3 : Bare_Base_Type_Decl;
Fld_5 : Internal_Env_Assoc := No_Env_Assoc;
Sym_1 : Symbol_Type;
Cast_Expr_3 : Bare_Type_Decl;
Cast_Result_4 : Bare_Base_Type_Decl;
Cast_Expr_4 : Bare_Type_Decl;
Cast_Result_5 : Bare_Base_Type_Decl;
Cast_Expr_5 : Bare_Type_Decl;
Cast_Result_6 : Bare_Base_Type_Decl;
Fld_6 : Internal_Env_Assoc := No_Env_Assoc;
Sym_2 : Symbol_Type;
Cast_Expr_6 : Bare_Type_Decl;
Cast_Result_7 : Bare_Base_Type_Decl;
Cast_Expr_7 : Bare_Type_Decl;
Cast_Result_8 : Bare_Base_Type_Decl;
Cast_Expr_8 : Bare_Type_Decl;
Cast_Result_9 : Bare_Base_Type_Decl;
Fld_7 : Internal_Env_Assoc := No_Env_Assoc;
Sym_3 : Symbol_Type;
Cast_Expr_9 : Bare_Type_Decl;
Cast_Result_10 : Bare_Base_Type_Decl;
Cast_Expr_10 : Bare_Type_Decl;
Cast_Result_11 : Bare_Base_Type_Decl;
Cast_Expr_11 : Bare_Type_Decl;
Cast_Result_12 : Bare_Base_Type_Decl;
Fld_8 : Internal_Env_Assoc := No_Env_Assoc;
Sym_4 : Symbol_Type;
Cast_Expr_12 : Bare_Type_Decl;
Cast_Result_13 : Bare_Base_Type_Decl;
Cast_Expr_13 : Bare_Type_Decl;
Cast_Result_14 : Bare_Base_Type_Decl;
Fld_9 : Internal_Env_Assoc := No_Env_Assoc;
Sym_5 : Symbol_Type;
Cast_Expr_14 : Bare_Type_Decl;
Cast_Result_15 : Bare_Base_Type_Decl;
Cast_Expr_15 : Bare_Type_Decl;
Cast_Result_16 : Bare_Base_Type_Decl;
Fld_10 : Internal_Env_Assoc := No_Env_Assoc;
Sym_6 : Symbol_Type;
Cast_Expr_16 : Bare_Type_Decl;
Cast_Result_17 : Bare_Base_Type_Decl;
Cast_Expr_17 : Bare_Type_Decl;
Cast_Result_18 : Bare_Base_Type_Decl;
Fld_11 : Internal_Env_Assoc := No_Env_Assoc;
Sym_7 : Symbol_Type;
Cast_Expr_18 : Bare_Type_Decl;
Cast_Result_19 : Bare_Base_Type_Decl;
Cast_Expr_19 : Bare_Type_Decl;
Cast_Result_20 : Bare_Base_Type_Decl;
Fld_12 : Internal_Env_Assoc := No_Env_Assoc;
Sym_8 : Symbol_Type;
Cast_Expr_20 : Bare_Type_Decl;
Cast_Result_21 : Bare_Base_Type_Decl;
Cast_Expr_21 : Bare_Type_Decl;
Cast_Result_22 : Bare_Base_Type_Decl;
Fld_13 : Internal_Env_Assoc := No_Env_Assoc;
Sym_9 : Symbol_Type;
Cast_Expr_22 : Bare_Type_Decl;
Cast_Result_23 : Bare_Base_Type_Decl;
Cast_Expr_23 : Bare_Type_Decl;
Cast_Result_24 : Bare_Base_Type_Decl;
Fld_14 : Internal_Env_Assoc := No_Env_Assoc;
Sym_10 : Symbol_Type;
Cast_Expr_24 : Bare_Type_Decl;
Cast_Result_25 : Bare_Base_Type_Decl;
Cast_Expr_25 : Bare_Type_Decl;
Cast_Result_26 : Bare_Base_Type_Decl;
Fld_15 : Internal_Env_Assoc := No_Env_Assoc;
Sym_11 : Symbol_Type;
Cast_Expr_26 : Bare_Type_Decl;
Cast_Result_27 : Bare_Base_Type_Decl;
Cast_Expr_27 : Bare_Type_Decl;
Cast_Result_28 : Bare_Base_Type_Decl;
Fld_16 : Internal_Env_Assoc := No_Env_Assoc;
Sym_12 : Symbol_Type;
Cast_Expr_28 : Bare_Type_Decl;
Cast_Result_29 : Bare_Base_Type_Decl;
Cast_Expr_29 : Bare_Type_Decl;
Cast_Result_30 : Bare_Base_Type_Decl;
Fld_17 : Internal_Env_Assoc := No_Env_Assoc;
Sym_13 : Symbol_Type;
Cast_Expr_30 : Bare_Type_Decl;
Cast_Result_31 : Bare_Base_Type_Decl;
Cast_Expr_31 : Bare_Type_Decl;
Cast_Result_32 : Bare_Base_Type_Decl;
Fld_18 : Internal_Env_Assoc := No_Env_Assoc;
Sym_14 : Symbol_Type;
Cast_Expr_32 : Bare_Type_Decl;
Cast_Result_33 : Bare_Base_Type_Decl;
Cast_Expr_33 : Bare_Type_Decl;
Cast_Result_34 : Bare_Base_Type_Decl;
Fld_19 : Internal_Env_Assoc := No_Env_Assoc;
Sym_15 : Symbol_Type;
Cast_Expr_34 : Bare_Type_Decl;
Cast_Result_35 : Bare_Base_Type_Decl;
Cast_Expr_35 : Bare_Type_Decl;
Cast_Result_36 : Bare_Base_Type_Decl;
Fld_20 : Internal_Env_Assoc := No_Env_Assoc;
Sym_16 : Symbol_Type;
Cast_Expr_36 : Bare_Type_Decl;
Cast_Result_37 : Bare_Base_Type_Decl;
Cast_Expr_37 : Bare_Type_Decl;
Cast_Result_38 : Bare_Base_Type_Decl;
Fld_21 : Internal_Env_Assoc := No_Env_Assoc;
Array_Lit : Internal_Env_Assoc_Array_Access;
Let_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_542 is
            begin
                     Dec_Ref (Fld_5);
                     Dec_Ref (Fld_6);
                     Dec_Ref (Fld_7);
                     Dec_Ref (Fld_8);
                     Dec_Ref (Fld_9);
                     Dec_Ref (Fld_10);
                     Dec_Ref (Fld_11);
                     Dec_Ref (Fld_12);
                     Dec_Ref (Fld_13);
                     Dec_Ref (Fld_14);
                     Dec_Ref (Fld_15);
                     Dec_Ref (Fld_16);
                     Dec_Ref (Fld_17);
                     Dec_Ref (Fld_18);
                     Dec_Ref (Fld_19);
                     Dec_Ref (Fld_20);
                     Dec_Ref (Fld_21);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_542;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Decimal_Fixed_Point_Def_Decimal_Fixed_Point_Def_P_Predefined_Operators,
            Items    => new Mmz_Key_Array (1 ..  1))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Decimal_Fixed_Point_Def,
                                As_Bare_Decimal_Fixed_Point_Def => Self);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Env_Assoc_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 11393 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11353 '<Cast to TypeDecl at ast.py:7268>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7268
--# expr-start 11352 '<parent at ast.py:7268>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7268
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 11352
   if Fld = null
      or else Fld.Kind in Ada_Type_Decl_Range
   then
      Cast_Result := Fld;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 11353
Block_Var := Cast_Result; 
--# bind self_type Block_Var
--# expr-start 11355 '<FieldAccess for node at ast.py:7269>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7269
--# expr-start 11354 '<FieldAccess for bool_type at ast.py:7269>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7269
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_1 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 11354
Fld_2 := Fld_1.Node;
--# expr-done 11355
Block_Var_1 := Fld_2; 
--# bind bool_type Block_Var_1
--# expr-start 11357 '<FieldAccess for node at ast.py:7270>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7270
--# expr-start 11356 '<FieldAccess for int_type at ast.py:7270>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7270
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.int_type
Fld_3 := Libadalang.Implementation.Ada_Node_P_Int_Type (Node => Self);
--# end
--# expr-done 11356
Fld_4 := Fld_3.Node;
--# expr-done 11357
Block_Var_2 := Fld_4; 
--# bind int_type Block_Var_2
--# expr-start 11392 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 11359 '<FieldAccess for create_binop_assoc at ast.py:7273>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7273
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11358 '<Symbol "+" at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_4); 
--# expr-done 11358
Cast_Expr := Block_Var; 
      Cast_Result_1 := Cast_Expr;
Cast_Expr_1 := Block_Var; 
      Cast_Result_2 := Cast_Expr_1;
Cast_Expr_2 := Block_Var; 
      Cast_Result_3 := Cast_Expr_2;
--# property-call-start AdaNode.create_binop_assoc
Fld_5 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym, Lhs => Cast_Result_1, Rhs => Cast_Result_2, Ret => Cast_Result_3);
--# end
--# expr-done 11359
--# expr-start 11361 '<FieldAccess for create_binop_assoc at ast.py:7274>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7274
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11360 '<Symbol "-" at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_5); 
--# expr-done 11360
Cast_Expr_3 := Block_Var; 
      Cast_Result_4 := Cast_Expr_3;
Cast_Expr_4 := Block_Var; 
      Cast_Result_5 := Cast_Expr_4;
Cast_Expr_5 := Block_Var; 
      Cast_Result_6 := Cast_Expr_5;
--# property-call-start AdaNode.create_binop_assoc
Fld_6 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_1, Lhs => Cast_Result_4, Rhs => Cast_Result_5, Ret => Cast_Result_6);
--# end
--# expr-done 11361
--# expr-start 11363 '<FieldAccess for create_binop_assoc at ast.py:7275>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7275
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11362 '<Symbol "*" at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_2); 
--# expr-done 11362
Cast_Expr_6 := Block_Var; 
      Cast_Result_7 := Cast_Expr_6;
Cast_Expr_7 := Block_Var; 
      Cast_Result_8 := Cast_Expr_7;
Cast_Expr_8 := Block_Var; 
      Cast_Result_9 := Cast_Expr_8;
--# property-call-start AdaNode.create_binop_assoc
Fld_7 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_2, Lhs => Cast_Result_7, Rhs => Cast_Result_8, Ret => Cast_Result_9);
--# end
--# expr-done 11363
--# expr-start 11365 '<FieldAccess for create_binop_assoc at ast.py:7276>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7276
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11364 '<Symbol "/" at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_6); 
--# expr-done 11364
Cast_Expr_9 := Block_Var; 
      Cast_Result_10 := Cast_Expr_9;
Cast_Expr_10 := Block_Var; 
      Cast_Result_11 := Cast_Expr_10;
Cast_Expr_11 := Block_Var; 
      Cast_Result_12 := Cast_Expr_11;
--# property-call-start AdaNode.create_binop_assoc
Fld_8 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_3, Lhs => Cast_Result_10, Rhs => Cast_Result_11, Ret => Cast_Result_12);
--# end
--# expr-done 11365
--# expr-start 11367 '<FieldAccess for create_binop_assoc at ast.py:7278>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7278
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11366 '<Symbol "*" at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_2); 
--# expr-done 11366
Cast_Expr_12 := Block_Var; 
      Cast_Result_13 := Cast_Expr_12;
Cast_Expr_13 := Block_Var; 
      Cast_Result_14 := Cast_Expr_13;
--# property-call-start AdaNode.create_binop_assoc
Fld_9 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_4, Lhs => Block_Var_2, Rhs => Cast_Result_13, Ret => Cast_Result_14);
--# end
--# expr-done 11367
--# expr-start 11369 '<FieldAccess for create_binop_assoc at ast.py:7279>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7279
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11368 '<Symbol "*" at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_2); 
--# expr-done 11368
Cast_Expr_14 := Block_Var; 
      Cast_Result_15 := Cast_Expr_14;
Cast_Expr_15 := Block_Var; 
      Cast_Result_16 := Cast_Expr_15;
--# property-call-start AdaNode.create_binop_assoc
Fld_10 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_5, Lhs => Cast_Result_15, Rhs => Block_Var_2, Ret => Cast_Result_16);
--# end
--# expr-done 11369
--# expr-start 11371 '<FieldAccess for create_binop_assoc at ast.py:7280>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7280
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11370 '<Symbol "/" at ???>' Sym_6 None
Sym_6 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_6); 
--# expr-done 11370
Cast_Expr_16 := Block_Var; 
      Cast_Result_17 := Cast_Expr_16;
Cast_Expr_17 := Block_Var; 
      Cast_Result_18 := Cast_Expr_17;
--# property-call-start AdaNode.create_binop_assoc
Fld_11 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_6, Lhs => Cast_Result_17, Rhs => Block_Var_2, Ret => Cast_Result_18);
--# end
--# expr-done 11371
--# expr-start 11373 '<FieldAccess for create_binop_assoc at ast.py:7281>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7281
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11372 '<Symbol "**" at ???>' Sym_7 None
Sym_7 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_3); 
--# expr-done 11372
Cast_Expr_18 := Block_Var; 
      Cast_Result_19 := Cast_Expr_18;
Cast_Expr_19 := Block_Var; 
      Cast_Result_20 := Cast_Expr_19;
--# property-call-start AdaNode.create_binop_assoc
Fld_12 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_7, Lhs => Cast_Result_19, Rhs => Block_Var_2, Ret => Cast_Result_20);
--# end
--# expr-done 11373
--# expr-start 11375 '<FieldAccess for create_binop_assoc at ast.py:7283>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7283
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11374 '<Symbol "<" at ???>' Sym_8 None
Sym_8 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_8); 
--# expr-done 11374
Cast_Expr_20 := Block_Var; 
      Cast_Result_21 := Cast_Expr_20;
Cast_Expr_21 := Block_Var; 
      Cast_Result_22 := Cast_Expr_21;
--# property-call-start AdaNode.create_binop_assoc
Fld_13 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_8, Lhs => Cast_Result_21, Rhs => Cast_Result_22, Ret => Block_Var_1);
--# end
--# expr-done 11375
--# expr-start 11377 '<FieldAccess for create_binop_assoc at ast.py:7284>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7284
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11376 '<Symbol "<=" at ???>' Sym_9 None
Sym_9 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_9); 
--# expr-done 11376
Cast_Expr_22 := Block_Var; 
      Cast_Result_23 := Cast_Expr_22;
Cast_Expr_23 := Block_Var; 
      Cast_Result_24 := Cast_Expr_23;
--# property-call-start AdaNode.create_binop_assoc
Fld_14 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_9, Lhs => Cast_Result_23, Rhs => Cast_Result_24, Ret => Block_Var_1);
--# end
--# expr-done 11377
--# expr-start 11379 '<FieldAccess for create_binop_assoc at ast.py:7285>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7285
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11378 '<Symbol "=" at ???>' Sym_10 None
Sym_10 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_10); 
--# expr-done 11378
Cast_Expr_24 := Block_Var; 
      Cast_Result_25 := Cast_Expr_24;
Cast_Expr_25 := Block_Var; 
      Cast_Result_26 := Cast_Expr_25;
--# property-call-start AdaNode.create_binop_assoc
Fld_15 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_10, Lhs => Cast_Result_25, Rhs => Cast_Result_26, Ret => Block_Var_1);
--# end
--# expr-done 11379
--# expr-start 11381 '<FieldAccess for create_binop_assoc at ast.py:7286>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7286
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11380 '<Symbol "/=" at ???>' Sym_11 None
Sym_11 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_7); 
--# expr-done 11380
Cast_Expr_26 := Block_Var; 
      Cast_Result_27 := Cast_Expr_26;
Cast_Expr_27 := Block_Var; 
      Cast_Result_28 := Cast_Expr_27;
--# property-call-start AdaNode.create_binop_assoc
Fld_16 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_11, Lhs => Cast_Result_27, Rhs => Cast_Result_28, Ret => Block_Var_1);
--# end
--# expr-done 11381
--# expr-start 11383 '<FieldAccess for create_binop_assoc at ast.py:7287>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7287
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11382 '<Symbol ">" at ???>' Sym_12 None
Sym_12 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_11); 
--# expr-done 11382
Cast_Expr_28 := Block_Var; 
      Cast_Result_29 := Cast_Expr_28;
Cast_Expr_29 := Block_Var; 
      Cast_Result_30 := Cast_Expr_29;
--# property-call-start AdaNode.create_binop_assoc
Fld_17 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_12, Lhs => Cast_Result_29, Rhs => Cast_Result_30, Ret => Block_Var_1);
--# end
--# expr-done 11383
--# expr-start 11385 '<FieldAccess for create_binop_assoc at ast.py:7288>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7288
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11384 '<Symbol ">=" at ???>' Sym_13 None
Sym_13 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_12); 
--# expr-done 11384
Cast_Expr_30 := Block_Var; 
      Cast_Result_31 := Cast_Expr_30;
Cast_Expr_31 := Block_Var; 
      Cast_Result_32 := Cast_Expr_31;
--# property-call-start AdaNode.create_binop_assoc
Fld_18 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_13, Lhs => Cast_Result_31, Rhs => Cast_Result_32, Ret => Block_Var_1);
--# end
--# expr-done 11385
--# expr-start 11387 '<FieldAccess for create_unop_assoc at ast.py:7290>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7290
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11386 '<Symbol "+" at ???>' Sym_14 None
Sym_14 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_4); 
--# expr-done 11386
Cast_Expr_32 := Block_Var; 
      Cast_Result_33 := Cast_Expr_32;
Cast_Expr_33 := Block_Var; 
      Cast_Result_34 := Cast_Expr_33;
--# property-call-start AdaNode.create_unop_assoc
Fld_19 := Libadalang.Implementation.Ada_Node_P_Create_Unop_Assoc (Node => Self, Op => Sym_14, Rhs => Cast_Result_33, Ret => Cast_Result_34);
--# end
--# expr-done 11387
--# expr-start 11389 '<FieldAccess for create_unop_assoc at ast.py:7291>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7291
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11388 '<Symbol "-" at ???>' Sym_15 None
Sym_15 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_5); 
--# expr-done 11388
Cast_Expr_34 := Block_Var; 
      Cast_Result_35 := Cast_Expr_34;
Cast_Expr_35 := Block_Var; 
      Cast_Result_36 := Cast_Expr_35;
--# property-call-start AdaNode.create_unop_assoc
Fld_20 := Libadalang.Implementation.Ada_Node_P_Create_Unop_Assoc (Node => Self, Op => Sym_15, Rhs => Cast_Result_35, Ret => Cast_Result_36);
--# end
--# expr-done 11389
--# expr-start 11391 '<FieldAccess for create_unop_assoc at ast.py:7292>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7292
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11390 '<Symbol "abs" at ???>' Sym_16 None
Sym_16 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Abs); 
--# expr-done 11390
Cast_Expr_36 := Block_Var; 
      Cast_Result_37 := Cast_Expr_36;
Cast_Expr_37 := Block_Var; 
      Cast_Result_38 := Cast_Expr_37;
--# property-call-start AdaNode.create_unop_assoc
Fld_21 := Libadalang.Implementation.Ada_Node_P_Create_Unop_Assoc (Node => Self, Op => Sym_16, Rhs => Cast_Result_37, Ret => Cast_Result_38);
--# end
--# expr-done 11391
Array_Lit := Create_Internal_Env_Assoc_Array (Internal_Internal_Env_Assoc_Array'(1 => Fld_5, 2 => Fld_6, 3 => Fld_7, 4 => Fld_8, 5 => Fld_9, 6 => Fld_10, 7 => Fld_11, 8 => Fld_12, 9 => Fld_13, 10 => Fld_14, 11 => Fld_15, 12 => Fld_16, 13 => Fld_17, 14 => Fld_18, 15 => Fld_19, 16 => Fld_20, 17 => Fld_21)); 
--# expr-done 11392
Let_Result := Array_Lit; Inc_Ref (Let_Result);
--# end
--# expr-done 11393
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_542;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Env_Assoc_Array_Access,
                        As_Internal_Env_Assoc_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_542;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Decimal_Fixed_Point_Def_P_Predefined_Operators;
--# end
   --
   --  Primitives for Bare_Floating_Point_Def
   --
      procedure Initialize_Fields_For_Floating_Point_Def
        (Self : Bare_Floating_Point_Def
         ; Floating_Point_Def_F_Num_Digits : Bare_Expr
         ; Floating_Point_Def_F_Range : Bare_Range_Spec
        ) is
      begin
            Self.Floating_Point_Def_F_Num_Digits := Floating_Point_Def_F_Num_Digits;
            Self.Floating_Point_Def_F_Range := Floating_Point_Def_F_Range;
      end Initialize_Fields_For_Floating_Point_Def;
   function Floating_Point_Def_F_Num_Digits
     (Node : Bare_Floating_Point_Def) return Bare_Expr
   is
   begin
         return Node.Floating_Point_Def_F_Num_Digits;
   end;
   function Floating_Point_Def_F_Range
     (Node : Bare_Floating_Point_Def) return Bare_Range_Spec
   is
   begin
         return Node.Floating_Point_Def_F_Range;
   end;
--# property-start FloatingPointDef.is_float_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7140
pragma Warnings (Off, "is not referenced");
 function Floating_Point_Def_P_Is_Float_Type
  (Node : Bare_Floating_Point_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Floating_Point_Def := Bare_Floating_Point_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Floating_Point_Def_P_Is_Float_Type;
--# end
--# property-start FloatingPointDef.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7143
pragma Warnings (Off, "is not referenced");
 function Floating_Point_Def_P_Xref_Equation
  (Node : Bare_Floating_Point_Def
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Floating_Point_Def := Bare_Floating_Point_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Floating_Point_Def :=
     Internal_Entity_Floating_Point_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_536 with Inline_Always;
            procedure Finalizer_Scope_3098 with Inline_Always;
      Var_Expr : Internal_Entity_Range_Spec;
Fld : Internal_Entity_Expr;
Fld_1 : Logic_Var;
Fld_2 : Logic_Equation;
Fld_3 : Internal_Entity_Expr;
Fld_4 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_5 : Internal_Entity_Expr;
Fld_6 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_7 : Internal_Entity_Range_Spec;
Fld_8 : Internal_Entity_Expr;
Fld_9 : Logic_Var;
Fld_10 : Logic_Equation;
Fld_11 : Internal_Entity_Expr;
Fld_12 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_13 : Internal_Entity_Expr;
Fld_14 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred_4 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_536 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Fld_4);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_536;
            procedure Finalizer_Scope_3098 is
            begin
                     Dec_Ref (Fld_10);
                     Dec_Ref (Fld_12);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Fld_14);
                     Dec_Ref (And_Pred_3);
            end Finalizer_Scope_3098;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11414 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11413 '<And at ast.py:7147>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7147
--# expr-start 11402 '<And at ast.py:7147>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7147
--# expr-start 11399 '<And at ast.py:7147>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7147
--# expr-start 11396 '<FieldAccess for universal_int_bind at ast.py:7150>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7150
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11395 '<FieldAccess for expected_type_var at ast.py:7150>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7150
--# expr-start 11394 '<FieldAccess for num_digits at ast.py:7150>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7150
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Floating_Point_Def_F_Num_Digits, Info => Ent.Info);
--# expr-done 11394
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld.Node);
--# expr-done 11395
--# property-call-start AdaNode.universal_int_bind
Fld_2 := Libadalang.Implementation.Ada_Node_P_Universal_Int_Bind (Node => Ent.Node, Type_Var => Fld_1, Origin => Origin);
--# end
--# expr-done 11396
--# expr-start 11398 '<FieldAccess for sub_equation at ast.py:7151>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7151
--# expr-start 11397 '<FieldAccess for num_digits at ast.py:7151>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7151
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Expr (Node => Ent.Node.Floating_Point_Def_F_Num_Digits, Info => Ent.Info);
--# expr-done 11397
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_4 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_3.Node, Env => Env, Origin => Origin, E_Info => Fld_3.Info);
--# end
--# expr-done 11398
And_Pred := Create_And (Fld_2, Fld_4, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7147") else null)); 
--# expr-done 11399
--# expr-start 11401 '<FieldAccess for matches_expected_type at ast.py:7152>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7152
--# expr-start 11400 '<FieldAccess for num_digits at ast.py:7152>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7152
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Expr (Node => Ent.Node.Floating_Point_Def_F_Num_Digits, Info => Ent.Info);
--# expr-done 11400
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_6 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_5.Node, Origin => Origin);
--# end
--# expr-done 11401
And_Pred_1 := Create_And (And_Pred, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7147") else null)); 
--# expr-done 11402
--# expr-start 11403 '<FieldAccess for range at ast.py:7156>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7156
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Range_Spec (Node => Ent.Node.Floating_Point_Def_F_Range, Info => Ent.Info);
--# expr-done 11403
Var_Expr := Fld_7;
if Var_Expr /= No_Entity_Range_Spec then
   --# scope-start
   --# bind r Var_Expr
--# expr-start 11412 '<And at ast.py:7158>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7158
--# expr-start 11409 '<And at ast.py:7158>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7158
--# expr-start 11406 '<FieldAccess for universal_real_bind at ast.py:7158>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7158
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11405 '<FieldAccess for expected_type_var at ast.py:7158>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7158
--# expr-start 11404 '<FieldAccess for range at ast.py:7158>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7158
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Expr (Node => Var_Expr.Node.Range_Spec_F_Range, Info => Var_Expr.Info);
--# expr-done 11404
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_8.Node);
--# expr-done 11405
--# property-call-start AdaNode.universal_real_bind
Fld_10 := Libadalang.Implementation.Ada_Node_P_Universal_Real_Bind (Node => Ent.Node, Type_Var => Fld_9, Origin => Origin);
--# end
--# expr-done 11406
--# expr-start 11408 '<FieldAccess for sub_equation at ast.py:7159>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7159
--# expr-start 11407 '<FieldAccess for range at ast.py:7159>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7159
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Create_Internal_Entity_Expr (Node => Var_Expr.Node.Range_Spec_F_Range, Info => Var_Expr.Info);
--# expr-done 11407
      if Fld_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_12 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_11.Node, Env => Env, Origin => Origin, E_Info => Fld_11.Info);
--# end
--# expr-done 11408
And_Pred_2 := Create_And (Fld_10, Fld_12, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7158") else null)); 
--# expr-done 11409
--# expr-start 11411 '<FieldAccess for matches_expected_type at ast.py:7160>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7160
--# expr-start 11410 '<FieldAccess for range at ast.py:7160>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7160
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Create_Internal_Entity_Expr (Node => Var_Expr.Node.Range_Spec_F_Range, Info => Var_Expr.Info);
--# expr-done 11410
      if Fld_13.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_14 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_13.Node, Origin => Origin);
--# end
--# expr-done 11411
And_Pred_3 := Create_And (And_Pred_2, Fld_14, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7158") else null)); 
--# expr-done 11412
Scope_Result := And_Pred_3;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_3098;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7161") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred_4 := Create_And (And_Pred_1, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7147") else null)); 
--# expr-done 11413
Let_Result := And_Pred_4; Inc_Ref (Let_Result);
--# end
--# expr-done 11414
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_536;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_536;
               Finalizer_Scope_3098;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Floating_Point_Def_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Ordinary_Fixed_Point_Def
   --
      procedure Initialize_Fields_For_Ordinary_Fixed_Point_Def
        (Self : Bare_Ordinary_Fixed_Point_Def
         ; Ordinary_Fixed_Point_Def_F_Delta : Bare_Expr
         ; Ordinary_Fixed_Point_Def_F_Range : Bare_Range_Spec
        ) is
      begin
            Self.Ordinary_Fixed_Point_Def_F_Delta := Ordinary_Fixed_Point_Def_F_Delta;
            Self.Ordinary_Fixed_Point_Def_F_Range := Ordinary_Fixed_Point_Def_F_Range;
      end Initialize_Fields_For_Ordinary_Fixed_Point_Def;
   function Ordinary_Fixed_Point_Def_F_Delta
     (Node : Bare_Ordinary_Fixed_Point_Def) return Bare_Expr
   is
   begin
         return Node.Ordinary_Fixed_Point_Def_F_Delta;
   end;
   function Ordinary_Fixed_Point_Def_F_Range
     (Node : Bare_Ordinary_Fixed_Point_Def) return Bare_Range_Spec
   is
   begin
         return Node.Ordinary_Fixed_Point_Def_F_Range;
   end;
--# property-start OrdinaryFixedPointDef.is_fixed_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7173
pragma Warnings (Off, "is not referenced");
 function Ordinary_Fixed_Point_Def_P_Is_Fixed_Point
  (Node : Bare_Ordinary_Fixed_Point_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Ordinary_Fixed_Point_Def := Bare_Ordinary_Fixed_Point_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ordinary_Fixed_Point_Def_P_Is_Fixed_Point;
--# end
--# property-start OrdinaryFixedPointDef.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7176
pragma Warnings (Off, "is not referenced");
 function Ordinary_Fixed_Point_Def_P_Xref_Equation
  (Node : Bare_Ordinary_Fixed_Point_Def
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Ordinary_Fixed_Point_Def := Bare_Ordinary_Fixed_Point_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Ordinary_Fixed_Point_Def :=
     Internal_Entity_Ordinary_Fixed_Point_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_538 with Inline_Always;
            procedure Finalizer_Scope_3099 with Inline_Always;
      Var_Expr : Internal_Entity_Range_Spec;
Fld : Internal_Entity_Expr;
Fld_1 : Logic_Var;
Fld_2 : Logic_Equation;
Fld_3 : Internal_Entity_Expr;
Fld_4 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_5 : Internal_Entity_Expr;
Fld_6 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Fld_7 : Internal_Entity_Range_Spec;
Fld_8 : Internal_Entity_Expr;
Fld_9 : Logic_Var;
Fld_10 : Logic_Equation;
Fld_11 : Internal_Entity_Expr;
Fld_12 : Logic_Equation;
And_Pred_2 : Logic_Equation;
Fld_13 : Internal_Entity_Expr;
Fld_14 : Logic_Equation;
And_Pred_3 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred_4 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_538 is
            begin
                     Dec_Ref (Fld_2);
                     Dec_Ref (Fld_4);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_6);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred_4);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_538;
            procedure Finalizer_Scope_3099 is
            begin
                     Dec_Ref (Fld_10);
                     Dec_Ref (Fld_12);
                     Dec_Ref (And_Pred_2);
                     Dec_Ref (Fld_14);
                     Dec_Ref (And_Pred_3);
            end Finalizer_Scope_3099;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11435 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11434 '<And at ast.py:7180>' And_Pred_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7180
--# expr-start 11423 '<And at ast.py:7180>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7180
--# expr-start 11420 '<And at ast.py:7180>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7180
--# expr-start 11417 '<FieldAccess for universal_real_bind at ast.py:7183>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7183
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11416 '<FieldAccess for expected_type_var at ast.py:7183>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7183
--# expr-start 11415 '<FieldAccess for delta at ast.py:7183>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7183
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Expr (Node => Ent.Node.Ordinary_Fixed_Point_Def_F_Delta, Info => Ent.Info);
--# expr-done 11415
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld.Node);
--# expr-done 11416
--# property-call-start AdaNode.universal_real_bind
Fld_2 := Libadalang.Implementation.Ada_Node_P_Universal_Real_Bind (Node => Ent.Node, Type_Var => Fld_1, Origin => Origin);
--# end
--# expr-done 11417
--# expr-start 11419 '<FieldAccess for sub_equation at ast.py:7184>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7184
--# expr-start 11418 '<FieldAccess for delta at ast.py:7184>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7184
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Expr (Node => Ent.Node.Ordinary_Fixed_Point_Def_F_Delta, Info => Ent.Info);
--# expr-done 11418
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_4 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_3.Node, Env => Env, Origin => Origin, E_Info => Fld_3.Info);
--# end
--# expr-done 11419
And_Pred := Create_And (Fld_2, Fld_4, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7180") else null)); 
--# expr-done 11420
--# expr-start 11422 '<FieldAccess for matches_expected_type at ast.py:7185>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7185
--# expr-start 11421 '<FieldAccess for delta at ast.py:7185>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7185
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Expr (Node => Ent.Node.Ordinary_Fixed_Point_Def_F_Delta, Info => Ent.Info);
--# expr-done 11421
      if Fld_5.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_6 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_5.Node, Origin => Origin);
--# end
--# expr-done 11422
And_Pred_1 := Create_And (And_Pred, Fld_6, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7180") else null)); 
--# expr-done 11423
--# expr-start 11424 '<FieldAccess for range at ast.py:7189>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7189
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Range_Spec (Node => Ent.Node.Ordinary_Fixed_Point_Def_F_Range, Info => Ent.Info);
--# expr-done 11424
Var_Expr := Fld_7;
if Var_Expr /= No_Entity_Range_Spec then
   --# scope-start
   --# bind r Var_Expr
--# expr-start 11433 '<And at ast.py:7191>' And_Pred_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7191
--# expr-start 11430 '<And at ast.py:7191>' And_Pred_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7191
--# expr-start 11427 '<FieldAccess for universal_real_bind at ast.py:7191>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7191
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11426 '<FieldAccess for expected_type_var at ast.py:7191>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7191
--# expr-start 11425 '<FieldAccess for range at ast.py:7191>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7191
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Create_Internal_Entity_Expr (Node => Var_Expr.Node.Range_Spec_F_Range, Info => Var_Expr.Info);
--# expr-done 11425
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Fld_8.Node);
--# expr-done 11426
--# property-call-start AdaNode.universal_real_bind
Fld_10 := Libadalang.Implementation.Ada_Node_P_Universal_Real_Bind (Node => Ent.Node, Type_Var => Fld_9, Origin => Origin);
--# end
--# expr-done 11427
--# expr-start 11429 '<FieldAccess for sub_equation at ast.py:7192>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7192
--# expr-start 11428 '<FieldAccess for range at ast.py:7192>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7192
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Create_Internal_Entity_Expr (Node => Var_Expr.Node.Range_Spec_F_Range, Info => Var_Expr.Info);
--# expr-done 11428
      if Fld_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_12 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld_11.Node, Env => Env, Origin => Origin, E_Info => Fld_11.Info);
--# end
--# expr-done 11429
And_Pred_2 := Create_And (Fld_10, Fld_12, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7191") else null)); 
--# expr-done 11430
--# expr-start 11432 '<FieldAccess for matches_expected_type at ast.py:7193>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7193
--# expr-start 11431 '<FieldAccess for range at ast.py:7193>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7193
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_13 := Create_Internal_Entity_Expr (Node => Var_Expr.Node.Range_Spec_F_Range, Info => Var_Expr.Info);
--# expr-done 11431
      if Fld_13.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_14 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Fld_13.Node, Origin => Origin);
--# end
--# expr-done 11432
And_Pred_3 := Create_And (And_Pred_2, Fld_14, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7191") else null)); 
--# expr-done 11433
Scope_Result := And_Pred_3;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_3099;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7194") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred_4 := Create_And (And_Pred_1, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:7180") else null)); 
--# expr-done 11434
Let_Result := And_Pred_4; Inc_Ref (Let_Result);
--# end
--# expr-done 11435
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_538;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_538;
               Finalizer_Scope_3099;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ordinary_Fixed_Point_Def_P_Xref_Equation;
--# end
--# property-start OrdinaryFixedPointDef.predefined_operators /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7199
pragma Warnings (Off, "is not referenced");
 function Ordinary_Fixed_Point_Def_P_Predefined_Operators
  (Node : Bare_Ordinary_Fixed_Point_Def
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Ordinary_Fixed_Point_Def := Bare_Ordinary_Fixed_Point_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_539 with Inline_Always;
      Block_Var : Bare_Type_Decl;
Block_Var_1 : Bare_Base_Type_Decl;
Block_Var_2 : Bare_Base_Type_Decl;
Fld : Bare_Ada_Node;
Cast_Result : Bare_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Bare_Base_Type_Decl;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Bare_Base_Type_Decl;
Sym : Symbol_Type;
Cast_Expr : Bare_Type_Decl;
Cast_Result_1 : Bare_Base_Type_Decl;
Cast_Expr_1 : Bare_Type_Decl;
Cast_Result_2 : Bare_Base_Type_Decl;
Cast_Expr_2 : Bare_Type_Decl;
Cast_Result_3 : Bare_Base_Type_Decl;
Fld_5 : Internal_Env_Assoc := No_Env_Assoc;
Sym_1 : Symbol_Type;
Cast_Expr_3 : Bare_Type_Decl;
Cast_Result_4 : Bare_Base_Type_Decl;
Cast_Expr_4 : Bare_Type_Decl;
Cast_Result_5 : Bare_Base_Type_Decl;
Cast_Expr_5 : Bare_Type_Decl;
Cast_Result_6 : Bare_Base_Type_Decl;
Fld_6 : Internal_Env_Assoc := No_Env_Assoc;
Sym_2 : Symbol_Type;
Cast_Expr_6 : Bare_Type_Decl;
Cast_Result_7 : Bare_Base_Type_Decl;
Cast_Expr_7 : Bare_Type_Decl;
Cast_Result_8 : Bare_Base_Type_Decl;
Cast_Expr_8 : Bare_Type_Decl;
Cast_Result_9 : Bare_Base_Type_Decl;
Fld_7 : Internal_Env_Assoc := No_Env_Assoc;
Sym_3 : Symbol_Type;
Cast_Expr_9 : Bare_Type_Decl;
Cast_Result_10 : Bare_Base_Type_Decl;
Cast_Expr_10 : Bare_Type_Decl;
Cast_Result_11 : Bare_Base_Type_Decl;
Cast_Expr_11 : Bare_Type_Decl;
Cast_Result_12 : Bare_Base_Type_Decl;
Fld_8 : Internal_Env_Assoc := No_Env_Assoc;
Sym_4 : Symbol_Type;
Cast_Expr_12 : Bare_Type_Decl;
Cast_Result_13 : Bare_Base_Type_Decl;
Cast_Expr_13 : Bare_Type_Decl;
Cast_Result_14 : Bare_Base_Type_Decl;
Fld_9 : Internal_Env_Assoc := No_Env_Assoc;
Sym_5 : Symbol_Type;
Cast_Expr_14 : Bare_Type_Decl;
Cast_Result_15 : Bare_Base_Type_Decl;
Cast_Expr_15 : Bare_Type_Decl;
Cast_Result_16 : Bare_Base_Type_Decl;
Fld_10 : Internal_Env_Assoc := No_Env_Assoc;
Sym_6 : Symbol_Type;
Cast_Expr_16 : Bare_Type_Decl;
Cast_Result_17 : Bare_Base_Type_Decl;
Cast_Expr_17 : Bare_Type_Decl;
Cast_Result_18 : Bare_Base_Type_Decl;
Fld_11 : Internal_Env_Assoc := No_Env_Assoc;
Sym_7 : Symbol_Type;
Cast_Expr_18 : Bare_Type_Decl;
Cast_Result_19 : Bare_Base_Type_Decl;
Cast_Expr_19 : Bare_Type_Decl;
Cast_Result_20 : Bare_Base_Type_Decl;
Fld_12 : Internal_Env_Assoc := No_Env_Assoc;
Sym_8 : Symbol_Type;
Cast_Expr_20 : Bare_Type_Decl;
Cast_Result_21 : Bare_Base_Type_Decl;
Cast_Expr_21 : Bare_Type_Decl;
Cast_Result_22 : Bare_Base_Type_Decl;
Fld_13 : Internal_Env_Assoc := No_Env_Assoc;
Sym_9 : Symbol_Type;
Cast_Expr_22 : Bare_Type_Decl;
Cast_Result_23 : Bare_Base_Type_Decl;
Cast_Expr_23 : Bare_Type_Decl;
Cast_Result_24 : Bare_Base_Type_Decl;
Fld_14 : Internal_Env_Assoc := No_Env_Assoc;
Sym_10 : Symbol_Type;
Cast_Expr_24 : Bare_Type_Decl;
Cast_Result_25 : Bare_Base_Type_Decl;
Cast_Expr_25 : Bare_Type_Decl;
Cast_Result_26 : Bare_Base_Type_Decl;
Fld_15 : Internal_Env_Assoc := No_Env_Assoc;
Sym_11 : Symbol_Type;
Cast_Expr_26 : Bare_Type_Decl;
Cast_Result_27 : Bare_Base_Type_Decl;
Cast_Expr_27 : Bare_Type_Decl;
Cast_Result_28 : Bare_Base_Type_Decl;
Fld_16 : Internal_Env_Assoc := No_Env_Assoc;
Sym_12 : Symbol_Type;
Cast_Expr_28 : Bare_Type_Decl;
Cast_Result_29 : Bare_Base_Type_Decl;
Cast_Expr_29 : Bare_Type_Decl;
Cast_Result_30 : Bare_Base_Type_Decl;
Fld_17 : Internal_Env_Assoc := No_Env_Assoc;
Sym_13 : Symbol_Type;
Cast_Expr_30 : Bare_Type_Decl;
Cast_Result_31 : Bare_Base_Type_Decl;
Cast_Expr_31 : Bare_Type_Decl;
Cast_Result_32 : Bare_Base_Type_Decl;
Fld_18 : Internal_Env_Assoc := No_Env_Assoc;
Sym_14 : Symbol_Type;
Cast_Expr_32 : Bare_Type_Decl;
Cast_Result_33 : Bare_Base_Type_Decl;
Cast_Expr_33 : Bare_Type_Decl;
Cast_Result_34 : Bare_Base_Type_Decl;
Fld_19 : Internal_Env_Assoc := No_Env_Assoc;
Sym_15 : Symbol_Type;
Cast_Expr_34 : Bare_Type_Decl;
Cast_Result_35 : Bare_Base_Type_Decl;
Cast_Expr_35 : Bare_Type_Decl;
Cast_Result_36 : Bare_Base_Type_Decl;
Fld_20 : Internal_Env_Assoc := No_Env_Assoc;
Sym_16 : Symbol_Type;
Cast_Expr_36 : Bare_Type_Decl;
Cast_Result_37 : Bare_Base_Type_Decl;
Cast_Expr_37 : Bare_Type_Decl;
Cast_Result_38 : Bare_Base_Type_Decl;
Fld_21 : Internal_Env_Assoc := No_Env_Assoc;
Array_Lit : Internal_Env_Assoc_Array_Access;
Let_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_539 is
            begin
                     Dec_Ref (Fld_5);
                     Dec_Ref (Fld_6);
                     Dec_Ref (Fld_7);
                     Dec_Ref (Fld_8);
                     Dec_Ref (Fld_9);
                     Dec_Ref (Fld_10);
                     Dec_Ref (Fld_11);
                     Dec_Ref (Fld_12);
                     Dec_Ref (Fld_13);
                     Dec_Ref (Fld_14);
                     Dec_Ref (Fld_15);
                     Dec_Ref (Fld_16);
                     Dec_Ref (Fld_17);
                     Dec_Ref (Fld_18);
                     Dec_Ref (Fld_19);
                     Dec_Ref (Fld_20);
                     Dec_Ref (Fld_21);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_539;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Ordinary_Fixed_Point_Def_Ordinary_Fixed_Point_Def_P_Predefined_Operators,
            Items    => new Mmz_Key_Array (1 ..  1))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Ordinary_Fixed_Point_Def,
                                As_Bare_Ordinary_Fixed_Point_Def => Self);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Env_Assoc_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 11477 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11437 '<Cast to TypeDecl at ast.py:7200>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7200
--# expr-start 11436 '<parent at ast.py:7200>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7200
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 11436
   if Fld = null
      or else Fld.Kind in Ada_Type_Decl_Range
   then
      Cast_Result := Fld;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 11437
Block_Var := Cast_Result; 
--# bind self_type Block_Var
--# expr-start 11439 '<FieldAccess for node at ast.py:7201>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7201
--# expr-start 11438 '<FieldAccess for bool_type at ast.py:7201>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7201
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_1 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 11438
Fld_2 := Fld_1.Node;
--# expr-done 11439
Block_Var_1 := Fld_2; 
--# bind bool_type Block_Var_1
--# expr-start 11441 '<FieldAccess for node at ast.py:7202>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7202
--# expr-start 11440 '<FieldAccess for int_type at ast.py:7202>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7202
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.int_type
Fld_3 := Libadalang.Implementation.Ada_Node_P_Int_Type (Node => Self);
--# end
--# expr-done 11440
Fld_4 := Fld_3.Node;
--# expr-done 11441
Block_Var_2 := Fld_4; 
--# bind int_type Block_Var_2
--# expr-start 11476 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 11443 '<FieldAccess for create_binop_assoc at ast.py:7205>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7205
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11442 '<Symbol "+" at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_4); 
--# expr-done 11442
Cast_Expr := Block_Var; 
      Cast_Result_1 := Cast_Expr;
Cast_Expr_1 := Block_Var; 
      Cast_Result_2 := Cast_Expr_1;
Cast_Expr_2 := Block_Var; 
      Cast_Result_3 := Cast_Expr_2;
--# property-call-start AdaNode.create_binop_assoc
Fld_5 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym, Lhs => Cast_Result_1, Rhs => Cast_Result_2, Ret => Cast_Result_3);
--# end
--# expr-done 11443
--# expr-start 11445 '<FieldAccess for create_binop_assoc at ast.py:7206>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7206
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11444 '<Symbol "-" at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_5); 
--# expr-done 11444
Cast_Expr_3 := Block_Var; 
      Cast_Result_4 := Cast_Expr_3;
Cast_Expr_4 := Block_Var; 
      Cast_Result_5 := Cast_Expr_4;
Cast_Expr_5 := Block_Var; 
      Cast_Result_6 := Cast_Expr_5;
--# property-call-start AdaNode.create_binop_assoc
Fld_6 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_1, Lhs => Cast_Result_4, Rhs => Cast_Result_5, Ret => Cast_Result_6);
--# end
--# expr-done 11445
--# expr-start 11447 '<FieldAccess for create_binop_assoc at ast.py:7207>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7207
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11446 '<Symbol "*" at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_2); 
--# expr-done 11446
Cast_Expr_6 := Block_Var; 
      Cast_Result_7 := Cast_Expr_6;
Cast_Expr_7 := Block_Var; 
      Cast_Result_8 := Cast_Expr_7;
Cast_Expr_8 := Block_Var; 
      Cast_Result_9 := Cast_Expr_8;
--# property-call-start AdaNode.create_binop_assoc
Fld_7 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_2, Lhs => Cast_Result_7, Rhs => Cast_Result_8, Ret => Cast_Result_9);
--# end
--# expr-done 11447
--# expr-start 11449 '<FieldAccess for create_binop_assoc at ast.py:7208>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7208
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11448 '<Symbol "/" at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_6); 
--# expr-done 11448
Cast_Expr_9 := Block_Var; 
      Cast_Result_10 := Cast_Expr_9;
Cast_Expr_10 := Block_Var; 
      Cast_Result_11 := Cast_Expr_10;
Cast_Expr_11 := Block_Var; 
      Cast_Result_12 := Cast_Expr_11;
--# property-call-start AdaNode.create_binop_assoc
Fld_8 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_3, Lhs => Cast_Result_10, Rhs => Cast_Result_11, Ret => Cast_Result_12);
--# end
--# expr-done 11449
--# expr-start 11451 '<FieldAccess for create_binop_assoc at ast.py:7210>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7210
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11450 '<Symbol "*" at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_2); 
--# expr-done 11450
Cast_Expr_12 := Block_Var; 
      Cast_Result_13 := Cast_Expr_12;
Cast_Expr_13 := Block_Var; 
      Cast_Result_14 := Cast_Expr_13;
--# property-call-start AdaNode.create_binop_assoc
Fld_9 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_4, Lhs => Block_Var_2, Rhs => Cast_Result_13, Ret => Cast_Result_14);
--# end
--# expr-done 11451
--# expr-start 11453 '<FieldAccess for create_binop_assoc at ast.py:7211>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7211
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11452 '<Symbol "*" at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_2); 
--# expr-done 11452
Cast_Expr_14 := Block_Var; 
      Cast_Result_15 := Cast_Expr_14;
Cast_Expr_15 := Block_Var; 
      Cast_Result_16 := Cast_Expr_15;
--# property-call-start AdaNode.create_binop_assoc
Fld_10 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_5, Lhs => Cast_Result_15, Rhs => Block_Var_2, Ret => Cast_Result_16);
--# end
--# expr-done 11453
--# expr-start 11455 '<FieldAccess for create_binop_assoc at ast.py:7212>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7212
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11454 '<Symbol "/" at ???>' Sym_6 None
Sym_6 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_6); 
--# expr-done 11454
Cast_Expr_16 := Block_Var; 
      Cast_Result_17 := Cast_Expr_16;
Cast_Expr_17 := Block_Var; 
      Cast_Result_18 := Cast_Expr_17;
--# property-call-start AdaNode.create_binop_assoc
Fld_11 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_6, Lhs => Cast_Result_17, Rhs => Block_Var_2, Ret => Cast_Result_18);
--# end
--# expr-done 11455
--# expr-start 11457 '<FieldAccess for create_binop_assoc at ast.py:7213>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7213
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11456 '<Symbol "**" at ???>' Sym_7 None
Sym_7 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_3); 
--# expr-done 11456
Cast_Expr_18 := Block_Var; 
      Cast_Result_19 := Cast_Expr_18;
Cast_Expr_19 := Block_Var; 
      Cast_Result_20 := Cast_Expr_19;
--# property-call-start AdaNode.create_binop_assoc
Fld_12 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_7, Lhs => Cast_Result_19, Rhs => Block_Var_2, Ret => Cast_Result_20);
--# end
--# expr-done 11457
--# expr-start 11459 '<FieldAccess for create_binop_assoc at ast.py:7215>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7215
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11458 '<Symbol "<" at ???>' Sym_8 None
Sym_8 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_8); 
--# expr-done 11458
Cast_Expr_20 := Block_Var; 
      Cast_Result_21 := Cast_Expr_20;
Cast_Expr_21 := Block_Var; 
      Cast_Result_22 := Cast_Expr_21;
--# property-call-start AdaNode.create_binop_assoc
Fld_13 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_8, Lhs => Cast_Result_21, Rhs => Cast_Result_22, Ret => Block_Var_1);
--# end
--# expr-done 11459
--# expr-start 11461 '<FieldAccess for create_binop_assoc at ast.py:7216>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7216
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11460 '<Symbol "<=" at ???>' Sym_9 None
Sym_9 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_9); 
--# expr-done 11460
Cast_Expr_22 := Block_Var; 
      Cast_Result_23 := Cast_Expr_22;
Cast_Expr_23 := Block_Var; 
      Cast_Result_24 := Cast_Expr_23;
--# property-call-start AdaNode.create_binop_assoc
Fld_14 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_9, Lhs => Cast_Result_23, Rhs => Cast_Result_24, Ret => Block_Var_1);
--# end
--# expr-done 11461
--# expr-start 11463 '<FieldAccess for create_binop_assoc at ast.py:7217>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7217
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11462 '<Symbol "=" at ???>' Sym_10 None
Sym_10 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_10); 
--# expr-done 11462
Cast_Expr_24 := Block_Var; 
      Cast_Result_25 := Cast_Expr_24;
Cast_Expr_25 := Block_Var; 
      Cast_Result_26 := Cast_Expr_25;
--# property-call-start AdaNode.create_binop_assoc
Fld_15 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_10, Lhs => Cast_Result_25, Rhs => Cast_Result_26, Ret => Block_Var_1);
--# end
--# expr-done 11463
--# expr-start 11465 '<FieldAccess for create_binop_assoc at ast.py:7218>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7218
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11464 '<Symbol "/=" at ???>' Sym_11 None
Sym_11 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_7); 
--# expr-done 11464
Cast_Expr_26 := Block_Var; 
      Cast_Result_27 := Cast_Expr_26;
Cast_Expr_27 := Block_Var; 
      Cast_Result_28 := Cast_Expr_27;
--# property-call-start AdaNode.create_binop_assoc
Fld_16 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_11, Lhs => Cast_Result_27, Rhs => Cast_Result_28, Ret => Block_Var_1);
--# end
--# expr-done 11465
--# expr-start 11467 '<FieldAccess for create_binop_assoc at ast.py:7219>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7219
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11466 '<Symbol ">" at ???>' Sym_12 None
Sym_12 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_11); 
--# expr-done 11466
Cast_Expr_28 := Block_Var; 
      Cast_Result_29 := Cast_Expr_28;
Cast_Expr_29 := Block_Var; 
      Cast_Result_30 := Cast_Expr_29;
--# property-call-start AdaNode.create_binop_assoc
Fld_17 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_12, Lhs => Cast_Result_29, Rhs => Cast_Result_30, Ret => Block_Var_1);
--# end
--# expr-done 11467
--# expr-start 11469 '<FieldAccess for create_binop_assoc at ast.py:7220>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7220
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11468 '<Symbol ">=" at ???>' Sym_13 None
Sym_13 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_12); 
--# expr-done 11468
Cast_Expr_30 := Block_Var; 
      Cast_Result_31 := Cast_Expr_30;
Cast_Expr_31 := Block_Var; 
      Cast_Result_32 := Cast_Expr_31;
--# property-call-start AdaNode.create_binop_assoc
Fld_18 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_13, Lhs => Cast_Result_31, Rhs => Cast_Result_32, Ret => Block_Var_1);
--# end
--# expr-done 11469
--# expr-start 11471 '<FieldAccess for create_unop_assoc at ast.py:7222>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7222
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11470 '<Symbol "+" at ???>' Sym_14 None
Sym_14 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_4); 
--# expr-done 11470
Cast_Expr_32 := Block_Var; 
      Cast_Result_33 := Cast_Expr_32;
Cast_Expr_33 := Block_Var; 
      Cast_Result_34 := Cast_Expr_33;
--# property-call-start AdaNode.create_unop_assoc
Fld_19 := Libadalang.Implementation.Ada_Node_P_Create_Unop_Assoc (Node => Self, Op => Sym_14, Rhs => Cast_Result_33, Ret => Cast_Result_34);
--# end
--# expr-done 11471
--# expr-start 11473 '<FieldAccess for create_unop_assoc at ast.py:7223>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7223
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11472 '<Symbol "-" at ???>' Sym_15 None
Sym_15 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_5); 
--# expr-done 11472
Cast_Expr_34 := Block_Var; 
      Cast_Result_35 := Cast_Expr_34;
Cast_Expr_35 := Block_Var; 
      Cast_Result_36 := Cast_Expr_35;
--# property-call-start AdaNode.create_unop_assoc
Fld_20 := Libadalang.Implementation.Ada_Node_P_Create_Unop_Assoc (Node => Self, Op => Sym_15, Rhs => Cast_Result_35, Ret => Cast_Result_36);
--# end
--# expr-done 11473
--# expr-start 11475 '<FieldAccess for create_unop_assoc at ast.py:7224>' Fld_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7224
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11474 '<Symbol "abs" at ???>' Sym_16 None
Sym_16 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Abs); 
--# expr-done 11474
Cast_Expr_36 := Block_Var; 
      Cast_Result_37 := Cast_Expr_36;
Cast_Expr_37 := Block_Var; 
      Cast_Result_38 := Cast_Expr_37;
--# property-call-start AdaNode.create_unop_assoc
Fld_21 := Libadalang.Implementation.Ada_Node_P_Create_Unop_Assoc (Node => Self, Op => Sym_16, Rhs => Cast_Result_37, Ret => Cast_Result_38);
--# end
--# expr-done 11475
Array_Lit := Create_Internal_Env_Assoc_Array (Internal_Internal_Env_Assoc_Array'(1 => Fld_5, 2 => Fld_6, 3 => Fld_7, 4 => Fld_8, 5 => Fld_9, 6 => Fld_10, 7 => Fld_11, 8 => Fld_12, 9 => Fld_13, 10 => Fld_14, 11 => Fld_15, 12 => Fld_16, 13 => Fld_17, 14 => Fld_18, 15 => Fld_19, 16 => Fld_20, 17 => Fld_21)); 
--# expr-done 11476
Let_Result := Array_Lit; Inc_Ref (Let_Result);
--# end
--# expr-done 11477
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_539;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Env_Assoc_Array_Access,
                        As_Internal_Env_Assoc_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_539;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ordinary_Fixed_Point_Def_P_Predefined_Operators;
--# end
   --
   --  Primitives for Bare_Record_Type_Def
   --
      procedure Initialize_Fields_For_Record_Type_Def
        (Self : Bare_Record_Type_Def
         ; Record_Type_Def_F_Has_Abstract : Bare_Abstract_Node
         ; Record_Type_Def_F_Has_Tagged : Bare_Tagged_Node
         ; Record_Type_Def_F_Has_Limited : Bare_Limited_Node
         ; Record_Type_Def_F_Record_Def : Bare_Base_Record_Def
        ) is
      begin
            Self.Record_Type_Def_F_Has_Abstract := Record_Type_Def_F_Has_Abstract;
            Self.Record_Type_Def_F_Has_Tagged := Record_Type_Def_F_Has_Tagged;
            Self.Record_Type_Def_F_Has_Limited := Record_Type_Def_F_Has_Limited;
            Self.Record_Type_Def_F_Record_Def := Record_Type_Def_F_Record_Def;
      end Initialize_Fields_For_Record_Type_Def;
   function Record_Type_Def_F_Has_Abstract
     (Node : Bare_Record_Type_Def) return Bare_Abstract_Node
   is
   begin
         return Node.Record_Type_Def_F_Has_Abstract;
   end;
   function Record_Type_Def_F_Has_Tagged
     (Node : Bare_Record_Type_Def) return Bare_Tagged_Node
   is
   begin
         return Node.Record_Type_Def_F_Has_Tagged;
   end;
   function Record_Type_Def_F_Has_Limited
     (Node : Bare_Record_Type_Def) return Bare_Limited_Node
   is
   begin
         return Node.Record_Type_Def_F_Has_Limited;
   end;
   function Record_Type_Def_F_Record_Def
     (Node : Bare_Record_Type_Def) return Bare_Base_Record_Def
   is
   begin
         return Node.Record_Type_Def_F_Record_Def;
   end;
--# property-start RecordTypeDef.is_tagged_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4979
pragma Warnings (Off, "is not referenced");
 function Record_Type_Def_P_Is_Tagged_Type
  (Node : Bare_Record_Type_Def
      ; Origin : Bare_Ada_Node
         := No_Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Record_Type_Def := Bare_Record_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
      Fld : Bare_Tagged_Node;
Fld_1 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11479 '<FieldAccess for as_bool at ast.py:4979>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4979
--# expr-start 11478 '<FieldAccess for has_tagged at ast.py:4979>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4979
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Record_Type_Def_F_Has_Tagged;
--# expr-done 11478
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Tagged.as_bool'
Fld_1 := Libadalang.Implementation.Dispatcher_Tagged_Node_P_As_Bool (Node => Fld);
--# end
--# expr-done 11479
      Property_Result := Fld_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Record_Type_Def_P_Is_Tagged_Type;
--# end
--# property-start RecordTypeDef.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4981
pragma Warnings (Off, "is not referenced");
 function Record_Type_Def_P_Xref_Equation
  (Node : Bare_Record_Type_Def
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Record_Type_Def := Bare_Record_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_360 with Inline_Always;
      True_Rel : Logic_Equation;
            procedure Finalizer_Scope_360 is
            begin
                     Dec_Ref (True_Rel);
            end Finalizer_Scope_360;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:4981") else null)); 
      Property_Result := True_Rel;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_360;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_360;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Record_Type_Def_P_Xref_Equation;
--# end
--# property-start RecordTypeDef.predefined_operators /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4984
pragma Warnings (Off, "is not referenced");
 function Record_Type_Def_P_Predefined_Operators
  (Node : Bare_Record_Type_Def
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Record_Type_Def := Bare_Record_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_361 with Inline_Always;
      Fld : Internal_Env_Assoc_Array_Access;
Fld_1 : Bare_Limited_Node;
Fld_2 : Boolean;
If_Result : Internal_Env_Assoc_Array_Access;
Let_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_361 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_361;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Record_Type_Def_Record_Type_Def_P_Predefined_Operators,
            Items    => new Mmz_Key_Array (1 ..  1))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Record_Type_Def,
                                As_Bare_Record_Type_Def => Self);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Env_Assoc_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 11484 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11483 '<If at ast.py:4985>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4985
--# expr-start 11481 '<FieldAccess for as_bool at ast.py:4986>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4986
--# expr-start 11480 '<FieldAccess for has_limited at ast.py:4986>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4986
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Self.Record_Type_Def_F_Has_Limited;
--# expr-done 11480
      if Fld_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Limited.as_bool'
Fld_2 := Libadalang.Implementation.Dispatcher_Limited_Node_P_As_Bool (Node => Fld_1);
--# end
--# expr-done 11481
if Fld_2 then
   If_Result := No_Internal_Env_Assoc_Array_Type;
else
   --# expr-start 11482 '<FieldAccess for predefined_equality_operators at ast.py:4988>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4988
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeDef.predefined_equality_operators
Fld := Libadalang.Implementation.Type_Def_P_Predefined_Equality_Operators (Node => Self);
--# end
--# expr-done 11482
   If_Result := Fld;
end if;
      Inc_Ref (If_Result);
--# expr-done 11483
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 11484
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_361;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Env_Assoc_Array_Access,
                        As_Internal_Env_Assoc_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_361;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Record_Type_Def_P_Predefined_Operators;
--# end
   --
   --  Primitives for Bare_Signed_Int_Type_Def
   --
      procedure Initialize_Fields_For_Signed_Int_Type_Def
        (Self : Bare_Signed_Int_Type_Def
         ; Signed_Int_Type_Def_F_Range : Bare_Range_Spec
        ) is
      begin
            Self.Signed_Int_Type_Def_F_Range := Signed_Int_Type_Def_F_Range;
      end Initialize_Fields_For_Signed_Int_Type_Def;
   function Signed_Int_Type_Def_F_Range
     (Node : Bare_Signed_Int_Type_Def) return Bare_Range_Spec
   is
   begin
         return Node.Signed_Int_Type_Def_F_Range;
   end;
--# property-start SignedIntTypeDef.is_int_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7588
pragma Warnings (Off, "is not referenced");
 function Signed_Int_Type_Def_P_Is_Int_Type
  (Node : Bare_Signed_Int_Type_Def
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Signed_Int_Type_Def := Bare_Signed_Int_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Signed_Int_Type_Def_P_Is_Int_Type;
--# end
--# property-start SignedIntTypeDef.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7590
pragma Warnings (Off, "is not referenced");
 function Signed_Int_Type_Def_P_Xref_Equation
  (Node : Bare_Signed_Int_Type_Def
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Signed_Int_Type_Def := Bare_Signed_Int_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Signed_Int_Type_Def :=
     Internal_Entity_Signed_Int_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_575 with Inline_Always;
      Fld : Internal_Entity_Range_Spec;
Fld_1 : Logic_Equation;
            procedure Finalizer_Scope_575 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_575;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11486 '<FieldAccess for xref_equation at ast.py:7590>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7590
--# expr-start 11485 '<FieldAccess for range at ast.py:7590>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7590
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Range_Spec (Node => Ent.Node.Signed_Int_Type_Def_F_Range, Info => Ent.Info);
--# expr-done 11485
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_1 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11486
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_575;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_575;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Signed_Int_Type_Def_P_Xref_Equation;
--# end
--# property-start SignedIntTypeDef.discrete_range /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7593
pragma Warnings (Off, "is not referenced");
 function Signed_Int_Type_Def_P_Discrete_Range
  (Node : Bare_Signed_Int_Type_Def
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Discrete_Range
is
   Self : Bare_Signed_Int_Type_Def := Bare_Signed_Int_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Signed_Int_Type_Def :=
     Internal_Entity_Signed_Int_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Discrete_Range;
      Fld : Internal_Entity_Range_Spec;
Fld_1 : Internal_Entity_Expr;
Fld_2 : Internal_Discrete_Range;
Let_Result : Internal_Discrete_Range;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11490 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11489 '<FieldAccess for discrete_range at ast.py:7594>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7594
--# expr-start 11488 '<FieldAccess for range at ast.py:7594>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7594
--# expr-start 11487 '<FieldAccess for range at ast.py:7594>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7594
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Range_Spec (Node => Ent.Node.Signed_Int_Type_Def_F_Range, Info => Ent.Info);
--# expr-done 11487
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Expr (Node => Fld.Node.Range_Spec_F_Range, Info => Fld.Info);
--# expr-done 11488
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.discrete_range
Fld_2 := Libadalang.Implementation.Expr_P_Discrete_Range (Node => Fld_1.Node, E_Info => Fld_1.Info);
--# end
--# expr-done 11489
Let_Result := Fld_2; 
--# end
--# expr-done 11490
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Signed_Int_Type_Def_P_Discrete_Range;
--# end
--# property-start SignedIntTypeDef.predefined_operators /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7597
pragma Warnings (Off, "is not referenced");
 function Signed_Int_Type_Def_P_Predefined_Operators
  (Node : Bare_Signed_Int_Type_Def
  )
   return Internal_Env_Assoc_Array_Access
is
   Self : Bare_Signed_Int_Type_Def := Bare_Signed_Int_Type_Def (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_577 with Inline_Always;
      Block_Var : Bare_Type_Decl;
Block_Var_1 : Bare_Base_Type_Decl;
Block_Var_2 : Bare_Base_Type_Decl;
Block_Var_3 : Internal_Env_Assoc_Array_Access;
Fld : Bare_Ada_Node;
Cast_Result : Bare_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Bare_Base_Type_Decl;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Bare_Base_Type_Decl;
Sym : Symbol_Type;
Cast_Expr : Bare_Type_Decl;
Cast_Result_1 : Bare_Base_Type_Decl;
Cast_Expr_1 : Bare_Type_Decl;
Cast_Result_2 : Bare_Base_Type_Decl;
Cast_Expr_2 : Bare_Type_Decl;
Cast_Result_3 : Bare_Base_Type_Decl;
Fld_5 : Internal_Env_Assoc := No_Env_Assoc;
Sym_1 : Symbol_Type;
Cast_Expr_3 : Bare_Type_Decl;
Cast_Result_4 : Bare_Base_Type_Decl;
Cast_Expr_4 : Bare_Type_Decl;
Cast_Result_5 : Bare_Base_Type_Decl;
Cast_Expr_5 : Bare_Type_Decl;
Cast_Result_6 : Bare_Base_Type_Decl;
Fld_6 : Internal_Env_Assoc := No_Env_Assoc;
Sym_2 : Symbol_Type;
Cast_Expr_6 : Bare_Type_Decl;
Cast_Result_7 : Bare_Base_Type_Decl;
Cast_Expr_7 : Bare_Type_Decl;
Cast_Result_8 : Bare_Base_Type_Decl;
Cast_Expr_8 : Bare_Type_Decl;
Cast_Result_9 : Bare_Base_Type_Decl;
Fld_7 : Internal_Env_Assoc := No_Env_Assoc;
Sym_3 : Symbol_Type;
Cast_Expr_9 : Bare_Type_Decl;
Cast_Result_10 : Bare_Base_Type_Decl;
Cast_Expr_10 : Bare_Type_Decl;
Cast_Result_11 : Bare_Base_Type_Decl;
Cast_Expr_11 : Bare_Type_Decl;
Cast_Result_12 : Bare_Base_Type_Decl;
Fld_8 : Internal_Env_Assoc := No_Env_Assoc;
Sym_4 : Symbol_Type;
Cast_Expr_12 : Bare_Type_Decl;
Cast_Result_13 : Bare_Base_Type_Decl;
Cast_Expr_13 : Bare_Type_Decl;
Cast_Result_14 : Bare_Base_Type_Decl;
Cast_Expr_14 : Bare_Type_Decl;
Cast_Result_15 : Bare_Base_Type_Decl;
Fld_9 : Internal_Env_Assoc := No_Env_Assoc;
Sym_5 : Symbol_Type;
Cast_Expr_15 : Bare_Type_Decl;
Cast_Result_16 : Bare_Base_Type_Decl;
Cast_Expr_16 : Bare_Type_Decl;
Cast_Result_17 : Bare_Base_Type_Decl;
Cast_Expr_17 : Bare_Type_Decl;
Cast_Result_18 : Bare_Base_Type_Decl;
Fld_10 : Internal_Env_Assoc := No_Env_Assoc;
Sym_6 : Symbol_Type;
Cast_Expr_18 : Bare_Type_Decl;
Cast_Result_19 : Bare_Base_Type_Decl;
Cast_Expr_19 : Bare_Type_Decl;
Cast_Result_20 : Bare_Base_Type_Decl;
Fld_11 : Internal_Env_Assoc := No_Env_Assoc;
Sym_7 : Symbol_Type;
Cast_Expr_20 : Bare_Type_Decl;
Cast_Result_21 : Bare_Base_Type_Decl;
Cast_Expr_21 : Bare_Type_Decl;
Cast_Result_22 : Bare_Base_Type_Decl;
Fld_12 : Internal_Env_Assoc := No_Env_Assoc;
Sym_8 : Symbol_Type;
Cast_Expr_22 : Bare_Type_Decl;
Cast_Result_23 : Bare_Base_Type_Decl;
Cast_Expr_23 : Bare_Type_Decl;
Cast_Result_24 : Bare_Base_Type_Decl;
Fld_13 : Internal_Env_Assoc := No_Env_Assoc;
Sym_9 : Symbol_Type;
Cast_Expr_24 : Bare_Type_Decl;
Cast_Result_25 : Bare_Base_Type_Decl;
Cast_Expr_25 : Bare_Type_Decl;
Cast_Result_26 : Bare_Base_Type_Decl;
Fld_14 : Internal_Env_Assoc := No_Env_Assoc;
Sym_10 : Symbol_Type;
Cast_Expr_26 : Bare_Type_Decl;
Cast_Result_27 : Bare_Base_Type_Decl;
Cast_Expr_27 : Bare_Type_Decl;
Cast_Result_28 : Bare_Base_Type_Decl;
Fld_15 : Internal_Env_Assoc := No_Env_Assoc;
Sym_11 : Symbol_Type;
Cast_Expr_28 : Bare_Type_Decl;
Cast_Result_29 : Bare_Base_Type_Decl;
Cast_Expr_29 : Bare_Type_Decl;
Cast_Result_30 : Bare_Base_Type_Decl;
Fld_16 : Internal_Env_Assoc := No_Env_Assoc;
Sym_12 : Symbol_Type;
Cast_Expr_30 : Bare_Type_Decl;
Cast_Result_31 : Bare_Base_Type_Decl;
Cast_Expr_31 : Bare_Type_Decl;
Cast_Result_32 : Bare_Base_Type_Decl;
Fld_17 : Internal_Env_Assoc := No_Env_Assoc;
Sym_13 : Symbol_Type;
Cast_Expr_32 : Bare_Type_Decl;
Cast_Result_33 : Bare_Base_Type_Decl;
Cast_Expr_33 : Bare_Type_Decl;
Cast_Result_34 : Bare_Base_Type_Decl;
Fld_18 : Internal_Env_Assoc := No_Env_Assoc;
Sym_14 : Symbol_Type;
Cast_Expr_34 : Bare_Type_Decl;
Cast_Result_35 : Bare_Base_Type_Decl;
Cast_Expr_35 : Bare_Type_Decl;
Cast_Result_36 : Bare_Base_Type_Decl;
Fld_19 : Internal_Env_Assoc := No_Env_Assoc;
Sym_15 : Symbol_Type;
Cast_Expr_36 : Bare_Type_Decl;
Cast_Result_37 : Bare_Base_Type_Decl;
Cast_Expr_37 : Bare_Type_Decl;
Cast_Result_38 : Bare_Base_Type_Decl;
Fld_20 : Internal_Env_Assoc := No_Env_Assoc;
Array_Lit : Internal_Env_Assoc_Array_Access;
Let_Result : Internal_Env_Assoc_Array_Access;
            procedure Finalizer_Scope_577 is
            begin
                     Dec_Ref (Fld_5);
                     Dec_Ref (Fld_6);
                     Dec_Ref (Fld_7);
                     Dec_Ref (Fld_8);
                     Dec_Ref (Fld_9);
                     Dec_Ref (Fld_10);
                     Dec_Ref (Fld_11);
                     Dec_Ref (Fld_12);
                     Dec_Ref (Fld_13);
                     Dec_Ref (Fld_14);
                     Dec_Ref (Fld_15);
                     Dec_Ref (Fld_16);
                     Dec_Ref (Fld_17);
                     Dec_Ref (Fld_18);
                     Dec_Ref (Fld_19);
                     Dec_Ref (Fld_20);
                     Dec_Ref (Array_Lit);
                     Dec_Ref (Block_Var_3);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_577;
      Mmz_Handle : Memoization_Handle;
      Mmz_Val    : Mmz_Value;
      Mmz_Stored : Boolean;
      --  Whether the memoization couple was actually stored. Used to determine
      --  whether to inc-ref the memoized value.
      function Create_Mmz_Key return Mmz_Key;
      --  Create a memoization key for this property call and return it
      --------------------
      -- Create_Mmz_Key --
      --------------------
      function Create_Mmz_Key return Mmz_Key is
      begin
         return Mmz_K : Mmz_Key :=
           (Property => Mmz_Bare_Signed_Int_Type_Def_Signed_Int_Type_Def_P_Predefined_Operators,
            Items    => new Mmz_Key_Array (1 ..  1))
         do
            Mmz_K.Items (1) := (Kind => Mmz_Bare_Signed_Int_Type_Def,
                                As_Bare_Signed_Int_Type_Def => Self);
         end return;
      end Create_Mmz_Key;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            if Find_Memoized_Value
              (Self.Unit, Mmz_Handle, Mmz_Val, Create_Mmz_Key'Access)
            then
               --# memoization-lookup
               if Mmz_Val.Kind = Mmz_Evaluating then
                  --# memoization-return
                  raise Property_Error with "Infinite recursion detected";
               elsif Mmz_Val.Kind = Mmz_Error then
                  --# memoization-return
                  Ada.Exceptions.Raise_Exception
                    (Mmz_Val.Exc_Id, "Memoized error");
               else
                  Property_Result := Mmz_Val.As_Internal_Env_Assoc_Array_Access;
                     Inc_Ref (Property_Result);
                  --# memoization-return
                  Exit_Call (Self.Unit.Context, Call_Depth);
                  return Property_Result;
               end if;
               --# end
            end if;
         end if;
      end if;
   --# scope-start
      --# expr-start 11530 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11492 '<Cast to TypeDecl at ast.py:7598>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7598
--# expr-start 11491 '<parent at ast.py:7598>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7598
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 11491
   if Fld = null
      or else Fld.Kind in Ada_Type_Decl_Range
   then
      Cast_Result := Fld;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 11492
Block_Var := Cast_Result; 
--# bind self_type Block_Var
--# expr-start 11494 '<FieldAccess for node at ast.py:7599>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7599
--# expr-start 11493 '<FieldAccess for bool_type at ast.py:7599>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7599
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.bool_type
Fld_1 := Libadalang.Implementation.Ada_Node_P_Bool_Type (Node => Self);
--# end
--# expr-done 11493
Fld_2 := Fld_1.Node;
--# expr-done 11494
Block_Var_1 := Fld_2; 
--# bind bool_type Block_Var_1
--# expr-start 11496 '<FieldAccess for node at ast.py:7600>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7600
--# expr-start 11495 '<FieldAccess for int_type at ast.py:7600>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7600
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.int_type
Fld_3 := Libadalang.Implementation.Ada_Node_P_Int_Type (Node => Self);
--# end
--# expr-done 11495
Fld_4 := Fld_3.Node;
--# expr-done 11496
Block_Var_2 := Fld_4; 
--# bind int_type Block_Var_2
--# expr-start 11529 '<ArrayLiteral at ???>' Array_Lit None
--# expr-start 11498 '<FieldAccess for create_binop_assoc at ast.py:7603>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7603
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11497 '<Symbol "+" at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_4); 
--# expr-done 11497
Cast_Expr := Block_Var; 
      Cast_Result_1 := Cast_Expr;
Cast_Expr_1 := Block_Var; 
      Cast_Result_2 := Cast_Expr_1;
Cast_Expr_2 := Block_Var; 
      Cast_Result_3 := Cast_Expr_2;
--# property-call-start AdaNode.create_binop_assoc
Fld_5 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym, Lhs => Cast_Result_1, Rhs => Cast_Result_2, Ret => Cast_Result_3);
--# end
--# expr-done 11498
--# expr-start 11500 '<FieldAccess for create_binop_assoc at ast.py:7604>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7604
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11499 '<Symbol "-" at ???>' Sym_1 None
Sym_1 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_5); 
--# expr-done 11499
Cast_Expr_3 := Block_Var; 
      Cast_Result_4 := Cast_Expr_3;
Cast_Expr_4 := Block_Var; 
      Cast_Result_5 := Cast_Expr_4;
Cast_Expr_5 := Block_Var; 
      Cast_Result_6 := Cast_Expr_5;
--# property-call-start AdaNode.create_binop_assoc
Fld_6 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_1, Lhs => Cast_Result_4, Rhs => Cast_Result_5, Ret => Cast_Result_6);
--# end
--# expr-done 11500
--# expr-start 11502 '<FieldAccess for create_binop_assoc at ast.py:7605>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7605
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11501 '<Symbol "*" at ???>' Sym_2 None
Sym_2 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_2); 
--# expr-done 11501
Cast_Expr_6 := Block_Var; 
      Cast_Result_7 := Cast_Expr_6;
Cast_Expr_7 := Block_Var; 
      Cast_Result_8 := Cast_Expr_7;
Cast_Expr_8 := Block_Var; 
      Cast_Result_9 := Cast_Expr_8;
--# property-call-start AdaNode.create_binop_assoc
Fld_7 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_2, Lhs => Cast_Result_7, Rhs => Cast_Result_8, Ret => Cast_Result_9);
--# end
--# expr-done 11502
--# expr-start 11504 '<FieldAccess for create_binop_assoc at ast.py:7606>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7606
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11503 '<Symbol "/" at ???>' Sym_3 None
Sym_3 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_6); 
--# expr-done 11503
Cast_Expr_9 := Block_Var; 
      Cast_Result_10 := Cast_Expr_9;
Cast_Expr_10 := Block_Var; 
      Cast_Result_11 := Cast_Expr_10;
Cast_Expr_11 := Block_Var; 
      Cast_Result_12 := Cast_Expr_11;
--# property-call-start AdaNode.create_binop_assoc
Fld_8 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_3, Lhs => Cast_Result_10, Rhs => Cast_Result_11, Ret => Cast_Result_12);
--# end
--# expr-done 11504
--# expr-start 11506 '<FieldAccess for create_binop_assoc at ast.py:7607>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7607
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11505 '<Symbol "mod" at ???>' Sym_4 None
Sym_4 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Mod); 
--# expr-done 11505
Cast_Expr_12 := Block_Var; 
      Cast_Result_13 := Cast_Expr_12;
Cast_Expr_13 := Block_Var; 
      Cast_Result_14 := Cast_Expr_13;
Cast_Expr_14 := Block_Var; 
      Cast_Result_15 := Cast_Expr_14;
--# property-call-start AdaNode.create_binop_assoc
Fld_9 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_4, Lhs => Cast_Result_13, Rhs => Cast_Result_14, Ret => Cast_Result_15);
--# end
--# expr-done 11506
--# expr-start 11508 '<FieldAccess for create_binop_assoc at ast.py:7608>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7608
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11507 '<Symbol "rem" at ???>' Sym_5 None
Sym_5 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Rem); 
--# expr-done 11507
Cast_Expr_15 := Block_Var; 
      Cast_Result_16 := Cast_Expr_15;
Cast_Expr_16 := Block_Var; 
      Cast_Result_17 := Cast_Expr_16;
Cast_Expr_17 := Block_Var; 
      Cast_Result_18 := Cast_Expr_17;
--# property-call-start AdaNode.create_binop_assoc
Fld_10 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_5, Lhs => Cast_Result_16, Rhs => Cast_Result_17, Ret => Cast_Result_18);
--# end
--# expr-done 11508
--# expr-start 11510 '<FieldAccess for create_binop_assoc at ast.py:7610>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7610
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11509 '<Symbol "**" at ???>' Sym_6 None
Sym_6 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_3); 
--# expr-done 11509
Cast_Expr_18 := Block_Var; 
      Cast_Result_19 := Cast_Expr_18;
Cast_Expr_19 := Block_Var; 
      Cast_Result_20 := Cast_Expr_19;
--# property-call-start AdaNode.create_binop_assoc
Fld_11 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_6, Lhs => Cast_Result_19, Rhs => Block_Var_2, Ret => Cast_Result_20);
--# end
--# expr-done 11510
--# expr-start 11512 '<FieldAccess for create_binop_assoc at ast.py:7612>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7612
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11511 '<Symbol "<" at ???>' Sym_7 None
Sym_7 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_8); 
--# expr-done 11511
Cast_Expr_20 := Block_Var; 
      Cast_Result_21 := Cast_Expr_20;
Cast_Expr_21 := Block_Var; 
      Cast_Result_22 := Cast_Expr_21;
--# property-call-start AdaNode.create_binop_assoc
Fld_12 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_7, Lhs => Cast_Result_21, Rhs => Cast_Result_22, Ret => Block_Var_1);
--# end
--# expr-done 11512
--# expr-start 11514 '<FieldAccess for create_binop_assoc at ast.py:7613>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7613
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11513 '<Symbol "<=" at ???>' Sym_8 None
Sym_8 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_9); 
--# expr-done 11513
Cast_Expr_22 := Block_Var; 
      Cast_Result_23 := Cast_Expr_22;
Cast_Expr_23 := Block_Var; 
      Cast_Result_24 := Cast_Expr_23;
--# property-call-start AdaNode.create_binop_assoc
Fld_13 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_8, Lhs => Cast_Result_23, Rhs => Cast_Result_24, Ret => Block_Var_1);
--# end
--# expr-done 11514
--# expr-start 11516 '<FieldAccess for create_binop_assoc at ast.py:7614>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7614
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11515 '<Symbol "=" at ???>' Sym_9 None
Sym_9 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_10); 
--# expr-done 11515
Cast_Expr_24 := Block_Var; 
      Cast_Result_25 := Cast_Expr_24;
Cast_Expr_25 := Block_Var; 
      Cast_Result_26 := Cast_Expr_25;
--# property-call-start AdaNode.create_binop_assoc
Fld_14 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_9, Lhs => Cast_Result_25, Rhs => Cast_Result_26, Ret => Block_Var_1);
--# end
--# expr-done 11516
--# expr-start 11518 '<FieldAccess for create_binop_assoc at ast.py:7615>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7615
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11517 '<Symbol "/=" at ???>' Sym_10 None
Sym_10 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_7); 
--# expr-done 11517
Cast_Expr_26 := Block_Var; 
      Cast_Result_27 := Cast_Expr_26;
Cast_Expr_27 := Block_Var; 
      Cast_Result_28 := Cast_Expr_27;
--# property-call-start AdaNode.create_binop_assoc
Fld_15 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_10, Lhs => Cast_Result_27, Rhs => Cast_Result_28, Ret => Block_Var_1);
--# end
--# expr-done 11518
--# expr-start 11520 '<FieldAccess for create_binop_assoc at ast.py:7616>' Fld_16 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7616
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11519 '<Symbol ">" at ???>' Sym_11 None
Sym_11 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_11); 
--# expr-done 11519
Cast_Expr_28 := Block_Var; 
      Cast_Result_29 := Cast_Expr_28;
Cast_Expr_29 := Block_Var; 
      Cast_Result_30 := Cast_Expr_29;
--# property-call-start AdaNode.create_binop_assoc
Fld_16 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_11, Lhs => Cast_Result_29, Rhs => Cast_Result_30, Ret => Block_Var_1);
--# end
--# expr-done 11520
--# expr-start 11522 '<FieldAccess for create_binop_assoc at ast.py:7617>' Fld_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7617
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11521 '<Symbol ">=" at ???>' Sym_12 None
Sym_12 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_12); 
--# expr-done 11521
Cast_Expr_30 := Block_Var; 
      Cast_Result_31 := Cast_Expr_30;
Cast_Expr_31 := Block_Var; 
      Cast_Result_32 := Cast_Expr_31;
--# property-call-start AdaNode.create_binop_assoc
Fld_17 := Libadalang.Implementation.Ada_Node_P_Create_Binop_Assoc (Node => Self, Op => Sym_12, Lhs => Cast_Result_31, Rhs => Cast_Result_32, Ret => Block_Var_1);
--# end
--# expr-done 11522
--# expr-start 11524 '<FieldAccess for create_unop_assoc at ast.py:7619>' Fld_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7619
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11523 '<Symbol "+" at ???>' Sym_13 None
Sym_13 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_4); 
--# expr-done 11523
Cast_Expr_32 := Block_Var; 
      Cast_Result_33 := Cast_Expr_32;
Cast_Expr_33 := Block_Var; 
      Cast_Result_34 := Cast_Expr_33;
--# property-call-start AdaNode.create_unop_assoc
Fld_18 := Libadalang.Implementation.Ada_Node_P_Create_Unop_Assoc (Node => Self, Op => Sym_13, Rhs => Cast_Result_33, Ret => Cast_Result_34);
--# end
--# expr-done 11524
--# expr-start 11526 '<FieldAccess for create_unop_assoc at ast.py:7620>' Fld_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7620
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11525 '<Symbol "-" at ???>' Sym_14 None
Sym_14 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_5); 
--# expr-done 11525
Cast_Expr_34 := Block_Var; 
      Cast_Result_35 := Cast_Expr_34;
Cast_Expr_35 := Block_Var; 
      Cast_Result_36 := Cast_Expr_35;
--# property-call-start AdaNode.create_unop_assoc
Fld_19 := Libadalang.Implementation.Ada_Node_P_Create_Unop_Assoc (Node => Self, Op => Sym_14, Rhs => Cast_Result_35, Ret => Cast_Result_36);
--# end
--# expr-done 11526
--# expr-start 11528 '<FieldAccess for create_unop_assoc at ast.py:7621>' Fld_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7621
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11527 '<Symbol "abs" at ???>' Sym_15 None
Sym_15 := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Abs); 
--# expr-done 11527
Cast_Expr_36 := Block_Var; 
      Cast_Result_37 := Cast_Expr_36;
Cast_Expr_37 := Block_Var; 
      Cast_Result_38 := Cast_Expr_37;
--# property-call-start AdaNode.create_unop_assoc
Fld_20 := Libadalang.Implementation.Ada_Node_P_Create_Unop_Assoc (Node => Self, Op => Sym_15, Rhs => Cast_Result_37, Ret => Cast_Result_38);
--# end
--# expr-done 11528
Array_Lit := Create_Internal_Env_Assoc_Array (Internal_Internal_Env_Assoc_Array'(1 => Fld_5, 2 => Fld_6, 3 => Fld_7, 4 => Fld_8, 5 => Fld_9, 6 => Fld_10, 7 => Fld_11, 8 => Fld_12, 9 => Fld_13, 10 => Fld_14, 11 => Fld_15, 12 => Fld_16, 13 => Fld_17, 14 => Fld_18, 15 => Fld_19, 16 => Fld_20)); 
--# expr-done 11529
Block_Var_3 := Array_Lit; Inc_Ref (Block_Var_3);
--# bind defaults Block_Var_3
Let_Result := Block_Var_3; Inc_Ref (Let_Result);
--# end
--# expr-done 11530
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_577;
      if Self /= null then
         if not Self.Unit.Context.In_Populate_Lexical_Env then
            Mmz_Val := (Kind => Mmz_Internal_Env_Assoc_Array_Access,
                        As_Internal_Env_Assoc_Array_Access => Property_Result);
            Add_Memoized_Value (Self.Unit, Mmz_Handle, Mmz_Val, Mmz_Stored);
               if Mmz_Stored then
                  Inc_Ref (Property_Result);
               end if;
         end if;
      end if;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_577;
         if Self /= null then
            if not Self.Unit.Context.In_Populate_Lexical_Env then
               Add_Memoized_Value
                 (Self.Unit,
                  Mmz_Handle,
                  (Kind   => Mmz_Error,
                   Exc_Id => Ada.Exceptions.Exception_Identity (Exc)),
                  Mmz_Stored);
            end if;
         end if;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Signed_Int_Type_Def_P_Predefined_Operators;
--# end
--# property-start SignedIntTypeDef.is_static /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7626
pragma Warnings (Off, "is not referenced");
 function Signed_Int_Type_Def_P_Is_Static
  (Node : Bare_Signed_Int_Type_Def
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Signed_Int_Type_Def := Bare_Signed_Int_Type_Def (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Signed_Int_Type_Def :=
     Internal_Entity_Signed_Int_Type_Def'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Fld : Internal_Entity_Range_Spec;
Fld_1 : Internal_Entity_Expr;
Fld_2 : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11533 '<FieldAccess for is_static_expr at ast.py:7626>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7626
--# expr-start 11532 '<FieldAccess for range at ast.py:7626>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7626
--# expr-start 11531 '<FieldAccess for range at ast.py:7626>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:7626
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Range_Spec (Node => Ent.Node.Signed_Int_Type_Def_F_Range, Info => Ent.Info);
--# expr-done 11531
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Expr (Node => Fld.Node.Range_Spec_F_Range, Info => Fld.Info);
--# expr-done 11532
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.is_static_expr
Fld_2 := Libadalang.Implementation.Expr_P_Is_Static_Expr (Node => Fld_1.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_1.Info);
--# end
--# expr-done 11533
      Property_Result := Fld_2;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Signed_Int_Type_Def_P_Is_Static;
--# end
   --
   --  Primitives for Bare_Type_Expr
   --
--# property-start TypeExpr.array_ndims /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8388
pragma Warnings (Off, "is not referenced");
 function Type_Expr_P_Array_Ndims
  (Node : Bare_Type_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Integer
is
   Self : Bare_Type_Expr := Bare_Type_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Expr :=
     Internal_Entity_Type_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Integer;
      Fld : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Integer;
Dyn_Var_Bind_Result : Integer;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11537 '<bind at ast.py:8389>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8389
--# expr-start 11534 '<FieldAccess for origin_node at ast.py:8389>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8389
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 11534
Bound_Origin := Fld; 
--# expr-start 11536 '<FieldAccess for array_ndims at ast.py:8389>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8389
--# expr-start 11535 '<FieldAccess for designated_type at ast.py:8389>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8389
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Ent.Node, Origin => Bound_Origin, E_Info => Ent.Info);
--# end
--# expr-done 11535
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.array_ndims'
Fld_2 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Array_Ndims (Node => Fld_1.Node, Origin => Bound_Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 11536
Dyn_Var_Bind_Result := Fld_2; 
--# expr-done 11537
      Property_Result := Dyn_Var_Bind_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Expr_P_Array_Ndims;
--# end
--# property-start TypeExpr.type_name /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8392
pragma Warnings (Off, "is not referenced");
 function Type_Expr_P_Type_Name
  (Node : Bare_Type_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Name
is
   Self : Bare_Type_Expr := Bare_Type_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Expr :=
     Internal_Entity_Type_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Name;
      Var_Expr : Internal_Entity_Subtype_Indication;
Cast_Expr : Internal_Entity_Type_Expr;
Cast_Result : Internal_Entity_Subtype_Indication;
Fld : Internal_Entity_Name;
Scope_Result : Internal_Entity_Name;
Result_Var : Internal_Entity_Name;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 11538 '<Cast to SubtypeIndication at ast.py:8393>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8393
Cast_Expr := Ent; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Subtype_Indication_Range
   then
      Cast_Result := Create_Internal_Entity_Subtype_Indication
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Subtype_Indication;
   end if;
--# expr-done 11538
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Subtype_Indication then
   --# scope-start
   --# bind sti Var_Expr
--# expr-start 11539 '<FieldAccess for name at ast.py:8393>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8393
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Var_Expr.Node.Subtype_Indication_F_Name, Info => Var_Expr.Info);
--# expr-done 11539
Scope_Result := Fld;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Name;
end if;
      Property_Result := Result_Var;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Expr_P_Type_Name;
--# end
--# property-start TypeExpr.accessed_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8400
pragma Warnings (Off, "is not referenced");
 function Type_Expr_P_Accessed_Type
  (Node : Bare_Type_Expr
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Type_Expr := Bare_Type_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Expr :=
     Internal_Entity_Type_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11542 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11540 '<FieldAccess for designated_type at ast.py:8401>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8401
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 11540
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 11541 '<FieldAccess for accessed_type at ast.py:8401>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8401
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.accessed_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Accessed_Type (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 11541
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 11542
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Expr_P_Accessed_Type;
--# end
--# property-start TypeExpr.defining_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8404
pragma Warnings (Off, "is not referenced");
 function Type_Expr_P_Defining_Env
  (Node : Bare_Type_Expr
      ; Origin : Bare_Ada_Node
      ; Include_Ud_Indexing : Boolean
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Type_Expr := Bare_Type_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Expr :=
     Internal_Entity_Type_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   --# bind include_ud_indexing Include_Ud_Indexing
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_688 with Inline_Always;
      Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_688 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_688;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11545 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11544 '<FieldAccess for defining_env at ast.py:8405>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8405
--# expr-start 11543 '<FieldAccess for designated_type at ast.py:8405>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8405
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 11543
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.defining_env'
Fld_1 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Defining_Env (Node => Fld.Node, Origin => Origin, Include_Ud_Indexing => Include_Ud_Indexing, E_Info => Fld.Info);
--# end
--# expr-done 11544
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 11545
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_688;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_688;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Expr_P_Defining_Env;
--# end
--# property-start '[dispatcher]TypeExpr.designated_type' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Type_Expr_P_Designated_Type
  (Node : Bare_Type_Expr
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Type_Expr := Bare_Type_Expr (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Type_Expr (Self.Kind) is
               when Ada_Anonymous_Type =>
                  --# property-call-start AnonymousType.designated_type
                  Property_Result := Anonymous_Type_P_Designated_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Enum_Lit_Synth_Type_Expr =>
                  --# property-call-start EnumLitSynthTypeExpr.designated_type
                  Property_Result := Enum_Lit_Synth_Type_Expr_P_Designated_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Subtype_Indication .. Ada_Discrete_Subtype_Indication =>
                  --# property-call-start SubtypeIndication.designated_type
                  Property_Result := Subtype_Indication_P_Designated_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
               when Ada_Synthetic_Type_Expr =>
                  --# property-call-start SyntheticTypeExpr.designated_type
                  Property_Result := Synthetic_Type_Expr_P_Designated_Type
                    (Self
                        , Origin
                        , E_Info
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Type_Expr_P_Designated_Type;
--# end
--# property-start TypeExpr.designated_type_decl /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8415
pragma Warnings (Off, "is not referenced");
 function Type_Expr_P_Designated_Type_Decl
  (Node : Bare_Type_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Type_Expr := Bare_Type_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Expr :=
     Internal_Entity_Type_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_1 : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11548 '<bind at ast.py:8416>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8416
--# expr-start 11546 '<FieldAccess for origin_node at ast.py:8416>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8416
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 11546
Bound_Origin := Fld; 
--# expr-start 11547 '<FieldAccess for designated_type at ast.py:8416>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8416
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Ent.Node, Origin => Bound_Origin, E_Info => Ent.Info);
--# end
--# expr-done 11547
Dyn_Var_Bind_Result := Fld_1; 
--# expr-done 11548
      Property_Result := Dyn_Var_Bind_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Expr_P_Designated_Type_Decl;
--# end
--# property-start TypeExpr.designated_type_decl_from /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8424
pragma Warnings (Off, "is not referenced");
 function Type_Expr_P_Designated_Type_Decl_From
  (Node : Bare_Type_Expr
      ; Origin_Node : Internal_Entity
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Type_Expr := Bare_Type_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Expr :=
     Internal_Entity_Type_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin_node Origin_Node
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Bare_Ada_Node;
Fld_1 : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_2 : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11553 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11552 '<bind at ast.py:8429>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8429
--# expr-start 11550 '<FieldAccess for origin_node at ast.py:8429>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8429
--# expr-start 11549 '<FieldAccess for node at ast.py:8429>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8429
Fld := Origin_Node.Node;
--# expr-done 11549
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_1 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Fld);
--# end
--# expr-done 11550
Bound_Origin := Fld_1; 
--# expr-start 11551 '<FieldAccess for designated_type at ast.py:8430>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8430
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Ent.Node, Origin => Bound_Origin, E_Info => Ent.Info);
--# end
--# expr-done 11551
Dyn_Var_Bind_Result := Fld_2; 
--# expr-done 11552
Let_Result := Dyn_Var_Bind_Result; 
--# end
--# expr-done 11553
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Expr_P_Designated_Type_Decl_From;
--# end
--# property-start TypeExpr.element_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8433
pragma Warnings (Off, "is not referenced");
 function Type_Expr_P_Element_Type
  (Node : Bare_Type_Expr
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Type_Expr := Bare_Type_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Expr :=
     Internal_Entity_Type_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Anonymous_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Fld_2 : Internal_Entity_Base_Type_Decl;
Cast_Result : Internal_Entity_Anonymous_Type_Decl;
Fld_3 : Internal_Entity_Type_Def;
Scope_Result : Internal_Entity_Type_Def;
Result_Var : Internal_Entity_Type_Def;
Cast_Result_1 : Internal_Entity_Access_Def;
Is_Null : Boolean;
If_Result : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11562 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11561 '<If at ast.py:8439>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8439
--# expr-start 11558 '<IsNull at ast.py:8440>' Is_Null /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8440
--# expr-start 11557 '<Cast to AccessDef at ast.py:8440>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8440
--# expr-start 11555 '<Cast to AnonymousTypeDecl at ast.py:8440>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8440
--# expr-start 11554 '<FieldAccess for designated_type at ast.py:8438>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8438
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_2 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 11554
   if Fld_2.Node = null
      or else Fld_2.Node.Kind in Ada_Anonymous_Type_Decl_Range
   then
      Cast_Result := Create_Internal_Entity_Anonymous_Type_Decl
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
   else
         Cast_Result := No_Entity_Anonymous_Type_Decl;
   end if;
--# expr-done 11555
Var_Expr := Cast_Result;
if Var_Expr /= No_Entity_Anonymous_Type_Decl then
   --# scope-start
--# expr-start 11556 '<FieldAccess for type_def at ast.py:8440>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8440
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Create_Internal_Entity_Type_Def (Node => Var_Expr.Node.Type_Decl_F_Type_Def, Info => Var_Expr.Info);
--# expr-done 11556
Scope_Result := Fld_3;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Type_Def;
end if;
   if Result_Var.Node = null
      or else Result_Var.Node.Kind in Ada_Access_Def
   then
      Cast_Result_1 := Create_Internal_Entity_Access_Def
        (Node => Result_Var.Node,
         Info => Result_Var.Info);
   else
         Cast_Result_1 := No_Entity_Access_Def;
   end if;
--# expr-done 11557
Is_Null := Cast_Result_1.Node = null; 
--# expr-done 11558
if Is_Null then
   --# expr-start 11559 '<FieldAccess for designated_type at ast.py:8438>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8438
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 11559
   If_Result := Fld;
else
   --# expr-start 11560 '<FieldAccess for accessed_type at ast.py:8442>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8442
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeExpr.accessed_type
Fld_1 := Libadalang.Implementation.Type_Expr_P_Accessed_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 11560
   If_Result := Fld_1;
end if;
--# expr-done 11561
Let_Result := If_Result; 
--# end
--# expr-done 11562
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Expr_P_Element_Type;
--# end
--# property-start TypeExpr.canonical_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8447
pragma Warnings (Off, "is not referenced");
 function Type_Expr_P_Canonical_Type
  (Node : Bare_Type_Expr
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Type_Expr := Bare_Type_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Type_Expr :=
     Internal_Entity_Type_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Var_Expr : Internal_Entity_Base_Type_Decl;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11565 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11563 '<FieldAccess for designated_type at ast.py:8448>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8448
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Ent.Node, Origin => Origin, E_Info => Ent.Info);
--# end
--# expr-done 11563
Var_Expr := Fld;
if Var_Expr /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 11564 '<FieldAccess for canonical_type at ast.py:8448>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8448
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.canonical_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Canonical_Type (Node => Var_Expr.Node, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 11564
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 11565
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Type_Expr_P_Canonical_Type;
--# end
   --
   --  Primitives for Bare_Anonymous_Type
   --
      procedure Initialize_Fields_For_Anonymous_Type
        (Self : Bare_Anonymous_Type
         ; Anonymous_Type_F_Type_Decl : Bare_Anonymous_Type_Decl
        ) is
      begin
            Self.Anonymous_Type_F_Type_Decl := Anonymous_Type_F_Type_Decl;
      end Initialize_Fields_For_Anonymous_Type;
   function Anonymous_Type_F_Type_Decl
     (Node : Bare_Anonymous_Type) return Bare_Anonymous_Type_Decl
   is
   begin
         return Node.Anonymous_Type_F_Type_Decl;
   end;
--# property-start AnonymousType.designated_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8484
pragma Warnings (Off, "is not referenced");
 function Anonymous_Type_P_Designated_Type
  (Node : Bare_Anonymous_Type
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Anonymous_Type := Bare_Anonymous_Type (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Anonymous_Type :=
     Internal_Entity_Anonymous_Type'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Anonymous_Type_Decl;
Cast_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 11566 '<FieldAccess for type_decl at ast.py:8484>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8484
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Anonymous_Type_Decl (Node => Ent.Node.Anonymous_Type_F_Type_Decl, Info => Ent.Info);
--# expr-done 11566
      Cast_Result := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
      Property_Result := Cast_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Anonymous_Type_P_Designated_Type;
--# end
--# property-start AnonymousType.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8485
pragma Warnings (Off, "is not referenced");
 function Anonymous_Type_P_Xref_Equation
  (Node : Bare_Anonymous_Type
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Anonymous_Type := Bare_Anonymous_Type (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Anonymous_Type :=
     Internal_Entity_Anonymous_Type'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_697 with Inline_Always;
      Fld : Internal_Entity_Anonymous_Type_Decl;
Fld_1 : Logic_Equation;
            procedure Finalizer_Scope_697 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_697;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11568 '<FieldAccess for sub_equation at ast.py:8485>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8485
--# expr-start 11567 '<FieldAccess for type_decl at ast.py:8485>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8485
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Anonymous_Type_Decl (Node => Ent.Node.Anonymous_Type_F_Type_Decl, Info => Ent.Info);
--# expr-done 11567
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_1 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11568
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_697;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_697;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Anonymous_Type_P_Xref_Equation;
--# end
--# property-start AnonymousType.custom_id_text /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8492
pragma Warnings (Off, "is not referenced");
 function Anonymous_Type_P_Custom_Id_Text
  (Node : Bare_Anonymous_Type
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return String_Type
is
   Self : Bare_Anonymous_Type := Bare_Anonymous_Type (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Anonymous_Type :=
     Internal_Entity_Anonymous_Type'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : String_Type;
            procedure Finalizer_Scope_698 with Inline_Always;
      Fld : Internal_Entity_Anonymous_Type_Decl;
Fld_1 : String_Type;
            procedure Finalizer_Scope_698 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_698;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11570 '<FieldAccess for text at ast.py:8492>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8492
--# expr-start 11569 '<FieldAccess for type_decl at ast.py:8492>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8492
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Anonymous_Type_Decl (Node => Ent.Node.Anonymous_Type_F_Type_Decl, Info => Ent.Info);
--# expr-done 11569
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Text (Node => Fld.Node);
--# expr-done 11570
      Property_Result := Fld_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_698;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_698;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Anonymous_Type_P_Custom_Id_Text;
--# end
   --
   --  Primitives for Bare_Enum_Lit_Synth_Type_Expr
   --
--# property-start EnumLitSynthTypeExpr.designated_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8456
pragma Warnings (Off, "is not referenced");
 function Enum_Lit_Synth_Type_Expr_P_Designated_Type
  (Node : Bare_Enum_Lit_Synth_Type_Expr
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Enum_Lit_Synth_Type_Expr := Bare_Enum_Lit_Synth_Type_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Enum_Lit_Synth_Type_Expr :=
     Internal_Entity_Enum_Lit_Synth_Type_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity;
Cast_Result : Internal_Entity_Enum_Literal_Decl;
Fld_1 : Internal_Entity_Type_Decl;
Cast_Result_1 : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
--# expr-start 11573 '<FieldAccess for enum_type at ast.py:8457>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8457
--# expr-start 11572 '<Cast to EnumLiteralDecl at ast.py:8457>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8457
--# expr-start 11571 '<parent at ast.py:8457>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8457
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 11571
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Enum_Literal_Decl_Range
   then
      Cast_Result := Create_Internal_Entity_Enum_Literal_Decl
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         Cast_Result := No_Entity_Enum_Literal_Decl;
   end if;
--# expr-done 11572
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]EnumLiteralDecl.enum_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Enum_Literal_Decl_P_Enum_Type (Node => Cast_Result.Node, E_Info => Cast_Result.Info);
--# end
--# expr-done 11573
      Cast_Result_1 := Create_Internal_Entity_Base_Type_Decl
        (Node => Fld_1.Node,
         Info => Fld_1.Info);
      Property_Result := Cast_Result_1;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Lit_Synth_Type_Expr_P_Designated_Type;
--# end
--# property-start EnumLitSynthTypeExpr.custom_id_text /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8460
pragma Warnings (Off, "is not referenced");
 function Enum_Lit_Synth_Type_Expr_P_Custom_Id_Text
  (Node : Bare_Enum_Lit_Synth_Type_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return String_Type
is
   Self : Bare_Enum_Lit_Synth_Type_Expr := Bare_Enum_Lit_Synth_Type_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Enum_Lit_Synth_Type_Expr :=
     Internal_Entity_Enum_Lit_Synth_Type_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : String_Type;
            procedure Finalizer_Scope_695 with Inline_Always;
      Cast_Expr : Bare_Enum_Lit_Synth_Type_Expr;
Cast_Result : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : String_Type;
Str : String_Type;
Concat_Result : String_Type;
Fld_2 : Internal_Entity;
Cast_Result_1 : Internal_Entity_Enum_Literal_Decl;
Fld_3 : Internal_Entity_Defining_Name;
Fld_4 : Symbol_Type_Array_Access;
Str_1 : String_Type;
Fld_5 : String_Type;
Concat_Result_1 : String_Type;
Dyn_Var_Bind_Result : String_Type;
            procedure Finalizer_Scope_695 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Str);
                     Dec_Ref (Concat_Result);
                     Dec_Ref (Fld_4);
                     Dec_Ref (Str_1);
                     Dec_Ref (Fld_5);
                     Dec_Ref (Concat_Result_1);
                     Dec_Ref (Dyn_Var_Bind_Result);
            end Finalizer_Scope_695;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11585 '<bind at ast.py:8463>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8463
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Bound_Origin := Cast_Result; 
--# expr-start 11584 '<Concat at ast.py:8465>' Concat_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8465
--# expr-start 11577 '<Concat at ast.py:8465>' Concat_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8465
--# expr-start 11575 '<FieldAccess for canonical_fully_qualified_name at ast.py:8465>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8465
--# expr-start 11574 '<FieldAccess for designated_type at ast.py:8465>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8465
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Ent.Node, Origin => Bound_Origin, E_Info => Ent.Info);
--# end
--# expr-done 11574
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.canonical_fully_qualified_name
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Canonical_Fully_Qualified_Name (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 11575
--# expr-start 11576 '<String at ast.py:8466>' Str /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8466
Str := Create_String ("."); 
--# expr-done 11576
Concat_Result := Concat_String (Fld_1, Str); 
--# expr-done 11577
--# expr-start 11583 '<FieldAccess for sym_join at ast.py:8468>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8468
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11581 '<FieldAccess for as_symbol_array at ast.py:8469>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8469
--# expr-start 11580 '<FieldAccess for defining_name at ast.py:8469>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8469
--# expr-start 11579 '<Cast to EnumLiteralDecl at ast.py:8469>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8469
--# expr-start 11578 '<parent at ast.py:8469>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8469
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 11578
   if Fld_2.Node = null
      or else Fld_2.Node.Kind in Ada_Enum_Literal_Decl_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Enum_Literal_Decl
        (Node => Fld_2.Node,
         Info => Fld_2.Info);
   else
         Cast_Result_1 := No_Entity_Enum_Literal_Decl;
   end if;
--# expr-done 11579
      if Cast_Result_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.defining_name
Fld_3 := Libadalang.Implementation.Basic_Decl_P_Defining_Name (Node => Cast_Result_1.Node, E_Info => Cast_Result_1.Info);
--# end
--# expr-done 11580
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.as_symbol_array'
Fld_4 := Libadalang.Implementation.Dispatcher_Name_P_As_Symbol_Array (Node => Fld_3.Node);
--# end
--# expr-done 11581
--# expr-start 11582 '<String at ast.py:8471>' Str_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8471
Str_1 := Create_String (""); 
--# expr-done 11582
--# property-call-start AdaNode.sym_join
Fld_5 := Libadalang.Implementation.Ada_Node_P_Sym_Join (Node => Ent.Node, Syms => Fld_4, Sep => Str_1);
--# end
--# expr-done 11583
Concat_Result_1 := Concat_String (Concat_Result, Fld_5); 
--# expr-done 11584
Dyn_Var_Bind_Result := Concat_Result_1; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 11585
      Property_Result := Dyn_Var_Bind_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_695;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_695;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Enum_Lit_Synth_Type_Expr_P_Custom_Id_Text;
--# end
   --
   --  Primitives for Bare_Subtype_Indication
   --
      procedure Initialize_Fields_For_Subtype_Indication
        (Self : Bare_Subtype_Indication
         ; Subtype_Indication_F_Has_Not_Null : Bare_Not_Null
         ; Subtype_Indication_F_Name : Bare_Name
         ; Subtype_Indication_F_Constraint : Bare_Constraint
        ) is
      begin
            Self.Subtype_Indication_F_Has_Not_Null := Subtype_Indication_F_Has_Not_Null;
            Self.Subtype_Indication_F_Name := Subtype_Indication_F_Name;
            Self.Subtype_Indication_F_Constraint := Subtype_Indication_F_Constraint;
      end Initialize_Fields_For_Subtype_Indication;
   function Subtype_Indication_F_Has_Not_Null
     (Node : Bare_Subtype_Indication) return Bare_Not_Null
   is
   begin
         return Node.Subtype_Indication_F_Has_Not_Null;
   end;
   function Subtype_Indication_F_Name
     (Node : Bare_Subtype_Indication) return Bare_Name
   is
   begin
         return Node.Subtype_Indication_F_Name;
   end;
   function Subtype_Indication_F_Constraint
     (Node : Bare_Subtype_Indication) return Bare_Constraint
   is
   begin
         return Node.Subtype_Indication_F_Constraint;
   end;
--# property-start SubtypeIndication.designated_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8519
pragma Warnings (Off, "is not referenced");
 function Subtype_Indication_P_Designated_Type
  (Node : Bare_Subtype_Indication
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Subtype_Indication := Bare_Subtype_Indication (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Subtype_Indication :=
     Internal_Entity_Subtype_Indication'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_699 with Inline_Always;
      Fld : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_1 : Internal_Entity_Name;
Fld_2 : Internal_Entity_Base_Type_Decl;
Dyn_Var_Bind_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_699 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Bound_Env);
            end Finalizer_Scope_699;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11589 '<bind at ast.py:8520>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8520
--# expr-start 11586 '<FieldAccess for node_env at ast.py:8520>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8520
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Node_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 11586
Bound_Env := Fld; Inc_Ref (Bound_Env);
--# expr-start 11588 '<FieldAccess for designated_type_impl at ast.py:8520>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8520
--# expr-start 11587 '<FieldAccess for name at ast.py:8520>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8520
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Ent.Node.Subtype_Indication_F_Name, Info => Ent.Info);
--# expr-done 11587
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.designated_type_impl'
Fld_2 := Libadalang.Implementation.Dispatcher_Name_P_Designated_Type_Impl (Node => Fld_1.Node, Env => Bound_Env, Origin => Origin, E_Info => Fld_1.Info);
--# end
--# expr-done 11588
Dyn_Var_Bind_Result := Fld_2; 
--# expr-done 11589
      Property_Result := Dyn_Var_Bind_Result;
   --# end
      Finalizer_Scope_699;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_699;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subtype_Indication_P_Designated_Type;
--# end
--# property-start SubtypeIndication.complete_items /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8524
pragma Warnings (Off, "is not referenced");
 function Subtype_Indication_P_Complete_Items
  (Node : Bare_Subtype_Indication
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Completion_Item_Array_Access
is
   Self : Bare_Subtype_Indication := Bare_Subtype_Indication (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Subtype_Indication :=
     Internal_Entity_Subtype_Indication'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Completion_Item_Array_Access;
            procedure Finalizer_Scope_700 with Inline_Always;
      Match_0 : Internal_Entity_Base_Type_Decl;
Match_1 : Internal_Entity_Package_Decl;
Match_2 : Internal_Entity;
Fld : Lexical_Env := Empty_Env;
Env_Get_Result : Internal_Entity_Array_Access;
Cast_Expr : Internal_Entity;
Cast_Result : Internal_Entity_Basic_Decl;
Fld_1 : Internal_Entity_Info;
Fld_2 : Internal_Metadata;
Fld_3 : Boolean;
Fld_4 : Internal_Unit;
Fld_5 : Boolean;
Fld_6 : Internal_Entity;
Is_A : Boolean;
Fld_7 : Internal_Entity;
Fld_8 : Internal_Entity;
Cast_Expr_1 : Internal_Entity_Base_Type_Decl;
Cast_Result_1 : Internal_Entity;
Is_Equiv : Boolean;
If_Result : Boolean;
If_Result_1 : Integer;
Match_Prefix : Internal_Entity;
Cast_Result_2 : Internal_Entity_Base_Type_Decl;
Let_Result : Integer;
Scope_Result : Integer;
Cast_Result_3 : Internal_Entity_Package_Decl;
Let_Result_1 : Integer;
Scope_Result_1 : Integer;
Cast_Result_4 : Internal_Entity;
Let_Result_2 : Integer;
Scope_Result_2 : Integer;
Match_Result : Integer;
New_Struct : Internal_Completion_Item;
Map_Result : Internal_Completion_Item_Array_Access;
Let_Result_3 : Internal_Completion_Item_Array_Access;
            procedure Finalizer_Scope_700 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Env_Get_Result);
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result_3);
            end Finalizer_Scope_700;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11607 '<Block at ???>' Let_Result_3 None
--# scope-start
--# expr-start 11606 '<Map at ast.py:8531>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8531
   --# expr-start 11591 '<EnvGet at ast.py:8531>' Env_Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8531
--# expr-start 11590 '<FieldAccess for children_env at ast.py:8531>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8531
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Libadalang.Implementation.Children_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 11590
Env_Get_Result := Construct_Entity_Array (AST_Envs.Get (Self => Fld, Key => null, Lookup_Kind => To_Lookup_Kind_Type (Recursive), Categories => (Inherited_Primitives => True, Nocat => True))); 
--# expr-done 11591
      declare
         Map_Result_Vec : Internal_Completion_Item_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Array_Access := Env_Get_Result;
      begin
         for Item_337 of
               Collection.Items
         loop
   --# scope-start
               --# bind n Item_337
      --# expr-start 11605 '<New CompletionItem at ast.py:8532>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8532
--# expr-start 11592 '<Cast to BasicDecl at ast.py:8533>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8533
Cast_Expr := Item_337; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Basic_Decl
   then
      Cast_Result := Create_Internal_Entity_Basic_Decl
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Basic_Decl;
   end if;
--# expr-done 11592
--# expr-start 11595 '<FieldAccess for dottable_subp at ast.py:8534>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8534
--# expr-start 11594 '<FieldAccess for md at ast.py:8534>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8534
--# expr-start 11593 '<FieldAccess for info at ast.py:8534>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8534
Fld_1 := Item_337.Info;
--# expr-done 11593
Fld_2 := Fld_1.Md;
--# expr-done 11594
Fld_3 := Fld_2.Dottable_Subp;
--# expr-done 11595
--# expr-start 11597 '<FieldAccess for has_with_visibility at ast.py:8535>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8535
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11596 '<FieldAccess for unit at ast.py:8535>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8535
      if Item_337.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Libadalang.Implementation.Unit (Node => Item_337.Node);
--# expr-done 11596
--# property-call-start AdaNode.has_with_visibility
Fld_5 := Libadalang.Implementation.Ada_Node_P_Has_With_Visibility (Node => Self, Refd_Unit => Fld_4);
--# end
--# expr-done 11597
--# expr-start 11604 '<Match at ast.py:8536>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8536
      if Item_337.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Item_337; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Discrete_Base_Subtype_Decl .. Ada_Synth_Anonymous_Type_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Base_Type_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_2; 
--# bind btd Match_0
--# expr-start 11603 '<If at ast.py:8537>' If_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8537
--# expr-start 11599 '<IsA DerivedTypeDef at ast.py:8544>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8544
--# expr-start 11598 '<parent at ast.py:8544>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8544
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 11598
Is_A := Fld_6.Node /= null 
and then Fld_6.Node.Kind in Ada_Derived_Type_Def_Range; 
--# expr-done 11599
if Is_A then
   --# expr-start 11602 '<Eq at ast.py:8545>' Is_Equiv /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8545
--# expr-start 11601 '<parent at ast.py:8545>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8545
--# expr-start 11600 '<parent at ast.py:8545>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8545
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Libadalang.Implementation.Parent (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 11600
      if Fld_7.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_8 := Libadalang.Implementation.Parent (Node => Fld_7.Node, E_Info => Fld_7.Info);
--# expr-done 11601
Cast_Expr_1 := Match_0; 
      Cast_Result_1 := Create_Internal_Entity
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
Is_Equiv := Equivalent (Fld_8, Cast_Result_1); 
--# expr-done 11602
   If_Result := Is_Equiv;
else
   If_Result := False;
end if;
if If_Result then
   If_Result_1 := 0;
else
   If_Result_1 := 100;
end if;
--# expr-done 11603
Let_Result := If_Result_1; 
--# end
Scope_Result := Let_Result;
   --# end
            Match_Result := Scope_Result; 
         when Ada_Package_Decl =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Package_Decl
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_3; 
--# bind _ Match_1
Let_Result_1 := 50; 
--# end
Scope_Result_1 := Let_Result_1;
   --# end
            Match_Result := Scope_Result_1; 
         when Ada_Abort_Absent .. Ada_Generic_Package_Internal | Ada_Abstract_Subp_Decl .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_4; 
--# bind _ Match_2
Let_Result_2 := 0; 
--# end
Scope_Result_2 := Let_Result_2;
   --# end
            Match_Result := Scope_Result_2; 
end case;
--# expr-done 11604
New_Struct := (Decl => Cast_Result, Is_Dot_Call => Fld_3, Is_Visible => Fld_5, Weight => Match_Result); 
--# expr-done 11605
         declare
            Item_To_Append : constant Internal_Completion_Item := New_Struct;
         begin
            Internal_Completion_Item_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
         Map_Result := Create_Internal_Completion_Item_Array
           (Items_Count => Natural (Internal_Completion_Item_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Completion_Item_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Completion_Item_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Completion_Item_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 11606
Let_Result_3 := Map_Result; Inc_Ref (Let_Result_3);
--# end
--# expr-done 11607
      Property_Result := Let_Result_3;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_700;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_700;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subtype_Indication_P_Complete_Items;
--# end
--# property-start SubtypeIndication.subtype_constraints /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8557
pragma Warnings (Off, "is not referenced");
 function Subtype_Indication_P_Subtype_Constraints
  (Node : Bare_Subtype_Indication
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Param_Actual_Array_Access
is
   Self : Bare_Subtype_Indication := Bare_Subtype_Indication (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Subtype_Indication :=
     Internal_Entity_Subtype_Indication'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Param_Actual_Array_Access;
            procedure Finalizer_Scope_701 with Inline_Always;
            procedure Finalizer_Scope_3109 with Inline_Always;
            procedure Finalizer_Scope_3110 with Inline_Always;
            procedure Finalizer_Scope_3112 with Inline_Always;
      Block_Var : Internal_Entity_Assoc_List;
Block_Var_1 : Internal_Param_Actual_Array_Access;
Var_Expr : Internal_Entity_Constraint;
Var_Expr_1 : Internal_Entity_Base_Type_Decl;
Ds : Internal_Entity_Discriminant_Spec;
Var_Expr_2 : Internal_Entity_Assoc_List;
Fld : Internal_Entity_Constraint;
Cast_Expr : Internal_Entity_Constraint;
Cast_Result : Internal_Entity_Composite_Constraint;
Fld_1 : Internal_Entity_Assoc_List;
Scope_Result : Internal_Entity_Assoc_List;
Result_Var : Internal_Entity_Assoc_List;
Fld_2 : Internal_Entity_Base_Type_Decl;
Fld_3 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Scope_Result_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Result_Var_1 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Cast_Expr_1 : Internal_Entity_Base_Formal_Param_Decl;
Cast_Result_1 : Internal_Entity_Discriminant_Spec;
Fld_4 : Internal_Entity_Defining_Name_List;
Fld_5 : Bare_Defining_Name_List;
Fld_6 : Internal_Entity_Info;
As_Entity : Internal_Entity_Defining_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Defining_Name;
Fld_7 : Internal_Entity_Expr;
New_Struct : Internal_Param_Actual;
Item_339 : Internal_Entity_Defining_Name;
Bare_Item_339 : Bare_Defining_Name;
Map_Result : Internal_Param_Actual_Array_Access;
Let_Result : Internal_Param_Actual_Array_Access;
Map_Result_1 : Internal_Param_Actual_Array_Access;
Fld_8 : Internal_Entity_Defining_Name;
Fld_9 : Internal_Entity_Defining_Name;
Fld_10 : Internal_Entity_Expr;
Fld_11 : Internal_Entity_Expr;
New_Struct_1 : Internal_Param_Actual;
Index_341 : Integer;
Map_Result_2 : Internal_Param_Actual_Array_Access;
Scope_Result_2 : Internal_Param_Actual_Array_Access;
Result_Var_2 : Internal_Param_Actual_Array_Access;
Let_Result_1 : Internal_Param_Actual_Array_Access;
            procedure Finalizer_Scope_701 is
            begin
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Block_Var_1);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Result_Var_2);
                     Dec_Ref (Let_Result_1);
            end Finalizer_Scope_701;
            procedure Finalizer_Scope_3109 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_3109;
            procedure Finalizer_Scope_3110 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_3110;
            procedure Finalizer_Scope_3112 is
            begin
                     Dec_Ref (Map_Result_2);
            end Finalizer_Scope_3112;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11626 '<Block at ???>' Let_Result_1 None
--# scope-start
--# expr-start 11608 '<FieldAccess for constraint at ast.py:8563>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8563
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Constraint (Node => Ent.Node.Subtype_Indication_F_Constraint, Info => Ent.Info);
--# expr-done 11608
Var_Expr := Fld;
if Var_Expr /= No_Entity_Constraint then
   --# scope-start
--# expr-start 11610 '<FieldAccess for constraints at ast.py:8563>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8563
--# expr-start 11609 '<Cast to CompositeConstraint at ast.py:8563>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8563
Cast_Expr := Var_Expr; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Composite_Constraint_Range
   then
      Cast_Result := Create_Internal_Entity_Composite_Constraint
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Composite_Constraint;
   end if;
--# expr-done 11609
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Assoc_List (Node => Cast_Result.Node.Composite_Constraint_F_Constraints, Info => Cast_Result.Info);
--# expr-done 11610
Scope_Result := Fld_1;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Assoc_List;
end if;
Block_Var := Result_Var; 
--# bind constraints Block_Var
--# expr-start 11619 '<Mapcat at ast.py:8567>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8567
--# expr-start 11611 '<FieldAccess for designated_type_decl at ast.py:8567>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8567
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeExpr.designated_type_decl
Fld_2 := Libadalang.Implementation.Type_Expr_P_Designated_Type_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 11611
Var_Expr_1 := Fld_2;
if Var_Expr_1 /= No_Entity_Base_Type_Decl then
   --# scope-start
--# expr-start 11612 '<FieldAccess for discriminants_list at ast.py:8567>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8567
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.discriminants_list'
Fld_3 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Discriminants_List (Node => Var_Expr_1.Node, Origin => No_Bare_Ada_Node, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 11612
Scope_Result_1 := Fld_3;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_3109;
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := No_Internal_Entity_Base_Formal_Param_Decl_Array_Type;
end if;
      Inc_Ref (Result_Var_1);
      declare
         Map_Result_1_Vec : Internal_Param_Actual_Vectors.Vector;
      begin
      declare
         Collection : constant Internal_Entity_Base_Formal_Param_Decl_Array_Access := Result_Var_1;
      begin
         for Item_338 of
               Collection.Items
         loop
   --# scope-start
               --# bind d Item_338
      --# expr-start 11618 '<Let at ast.py:8568>' Let_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8568
--# scope-start
--# expr-start 11613 '<Cast to DiscriminantSpec at ast.py:8569>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8569
Cast_Expr_1 := Item_338; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Discriminant_Spec_Range
   then
      Cast_Result_1 := Create_Internal_Entity_Discriminant_Spec
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_1 := No_Entity_Discriminant_Spec;
   end if;
--# expr-done 11613
Ds := Cast_Result_1; 
--# bind ds Ds
--# expr-start 11617 '<Map at ast.py:8570>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8570
   --# expr-start 11614 '<FieldAccess for ids at ast.py:8570>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8570
      if Ds.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Defining_Name_List (Node => Ds.Node.Discriminant_Spec_F_Ids, Info => Ds.Info);
--# expr-done 11614
Fld_5 := Fld_4.Node;
      declare
         Map_Result_Vec : Internal_Param_Actual_Vectors.Vector;
      begin
      if Fld_5 = null then
      Map_Result := Create_Internal_Param_Actual_Array (0);
      else
      declare
         Collection : constant Bare_Defining_Name_List := Fld_5;
      begin
         for Untyped_Item_339 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_339 := Untyped_Item_339; 
Is_Null := Bare_Item_339 = null; 
if Is_Null then
   If_Result := No_Entity_Defining_Name;
else
Fld_6 := Fld_4.Info;
As_Entity := (Info => Fld_6, Node => Bare_Item_339); 
   If_Result := As_Entity;
end if;
                  Item_339 := If_Result; 
   --# scope-start
               --# bind i Item_339
      --# expr-start 11616 '<New ParamActual at ast.py:8571>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8571
--# expr-start 11615 '<FieldAccess for default_expr at ast.py:8573>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8573
      if Ds.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_7 := Create_Internal_Entity_Expr (Node => Ds.Node.Discriminant_Spec_F_Default_Expr, Info => Ds.Info);
--# expr-done 11615
New_Struct := (Actual => Fld_7, Param => Item_339); 
--# expr-done 11616
         declare
            Item_To_Append : constant Internal_Param_Actual := New_Struct;
         begin
            Internal_Param_Actual_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Internal_Param_Actual_Array
           (Items_Count => Natural (Internal_Param_Actual_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Internal_Param_Actual_Vectors.Get
              (Map_Result_Vec,
               I + Internal_Param_Actual_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Internal_Param_Actual_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 11617
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 11618
         for Item_To_Append of
               Let_Result.Items
         loop
            Internal_Param_Actual_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end loop;
   --# end
      Finalizer_Scope_3110;
         end loop;
      end;
         Map_Result_1 := Create_Internal_Param_Actual_Array
           (Items_Count => Natural (Internal_Param_Actual_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Internal_Param_Actual_Vectors.Get
              (Map_Result_1_Vec,
               I + Internal_Param_Actual_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Internal_Param_Actual_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 11619
Block_Var_1 := Map_Result_1; Inc_Ref (Block_Var_1);
--# bind discrs Block_Var_1
Var_Expr_2 := Block_Var;
if Var_Expr_2 /= No_Entity_Assoc_List then
   --# scope-start
   --# bind c Var_Expr_2
--# expr-start 11625 '<Map at ast.py:8582>' Map_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8582
      declare
         Map_Result_2_Vec : Internal_Param_Actual_Vectors.Vector;
      begin
         Index_341 := 0;
      declare
         Collection : constant Internal_Param_Actual_Array_Access := Block_Var_1;
      begin
         for Item_340 of
               Collection.Items
         loop
   --# scope-start
               --# bind dp Item_340
               --# bind i Index_341
      --# expr-start 11624 '<New ParamActual at ast.py:8583>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8583
--# expr-start 11622 '<FieldAccess for actual_for_param_at at ast.py:8585>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8585
      if Var_Expr_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11620 '<FieldAccess for param at ast.py:8585>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8585
Fld_9 := Item_340.Param;
--# expr-done 11620
--# expr-start 11621 '<FieldAccess for actual at ast.py:8585>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8585
Fld_10 := Item_340.Actual;
--# expr-done 11621
--# property-call-start AssocList.actual_for_param_at
Fld_11 := Libadalang.Implementation.Assoc_List_P_Actual_For_Param_At (Node => Var_Expr_2.Node, Param => Fld_9, Pos => Index_341, Default_Expr => Fld_10, E_Info => Var_Expr_2.Info);
--# end
--# expr-done 11622
--# expr-start 11623 '<FieldAccess for param at ast.py:8584>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8584
Fld_8 := Item_340.Param;
--# expr-done 11623
New_Struct_1 := (Actual => Fld_11, Param => Fld_8); 
--# expr-done 11624
         declare
            Item_To_Append : constant Internal_Param_Actual := New_Struct_1;
         begin
            Internal_Param_Actual_Vectors.Append (Map_Result_2_Vec, Item_To_Append);
         end;
   --# end
               Index_341 := Index_341 + 1;
         end loop;
      end;
         Map_Result_2 := Create_Internal_Param_Actual_Array
           (Items_Count => Natural (Internal_Param_Actual_Vectors.Length (Map_Result_2_Vec)));
         for I in Map_Result_2.Items'Range loop
            Map_Result_2.Items (I) := Internal_Param_Actual_Vectors.Get
              (Map_Result_2_Vec,
               I + Internal_Param_Actual_Vectors.Index_Type'First - Map_Result_2.Items'First);
         end loop;
         Internal_Param_Actual_Vectors.Destroy (Map_Result_2_Vec);
      end;
--# expr-done 11625
Scope_Result_2 := Map_Result_2;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_3112;
   Result_Var_2 := Scope_Result_2;
else
   Result_Var_2 := Block_Var_1;
end if;
      Inc_Ref (Result_Var_2);
Let_Result_1 := Result_Var_2; Inc_Ref (Let_Result_1);
--# end
--# expr-done 11626
      Property_Result := Let_Result_1;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_701;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_701;
               Finalizer_Scope_3109;
               Finalizer_Scope_3110;
               Finalizer_Scope_3112;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subtype_Indication_P_Subtype_Constraints;
--# end
--# property-start SubtypeIndication.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8592
pragma Warnings (Off, "is not referenced");
 function Subtype_Indication_P_Xref_Equation
  (Node : Bare_Subtype_Indication
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Subtype_Indication := Bare_Subtype_Indication (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Subtype_Indication :=
     Internal_Entity_Subtype_Indication'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_702 with Inline_Always;
            procedure Finalizer_Scope_3114 with Inline_Always;
      Var_Expr : Internal_Entity_Constraint;
Fld : Internal_Entity_Name;
Fld_1 : Logic_Equation;
Fld_2 : Internal_Entity_Constraint;
Fld_3 : Logic_Equation;
Scope_Result : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_702 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_702;
            procedure Finalizer_Scope_3114 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_3114;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11632 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11631 '<And at ast.py:8595>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8595
--# expr-start 11628 '<FieldAccess for subtype_indication_equation at ast.py:8596>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8596
--# expr-start 11627 '<FieldAccess for name at ast.py:8596>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8596
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name (Node => Ent.Node.Subtype_Indication_F_Name, Info => Ent.Info);
--# expr-done 11627
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.subtype_indication_equation
Fld_1 := Libadalang.Implementation.Name_P_Subtype_Indication_Equation (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11628
--# expr-start 11629 '<FieldAccess for constraint at ast.py:8597>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8597
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Constraint (Node => Ent.Node.Subtype_Indication_F_Constraint, Info => Ent.Info);
--# expr-done 11629
Var_Expr := Fld_2;
if Var_Expr /= No_Entity_Constraint then
   --# scope-start
   --# bind c Var_Expr
--# expr-start 11630 '<FieldAccess for sub_equation at ast.py:8598>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8598
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_3 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Var_Expr.Node, Env => Env, Origin => Origin, E_Info => Var_Expr.Info);
--# end
--# expr-done 11630
Scope_Result := Fld_3;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_3114;
   Result_Var := Scope_Result;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8598") else null)); 
   Result_Var := True_Rel;
end if;
      Inc_Ref (Result_Var);
And_Pred := Create_And (Fld_1, Result_Var, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8595") else null)); 
--# expr-done 11631
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 11632
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_702;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_702;
               Finalizer_Scope_3114;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subtype_Indication_P_Xref_Equation;
--# end
--# property-start SubtypeIndication.discrete_range /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8603
pragma Warnings (Off, "is not referenced");
 function Subtype_Indication_P_Discrete_Range
  (Node : Bare_Subtype_Indication
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Discrete_Range
is
   Self : Bare_Subtype_Indication := Bare_Subtype_Indication (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Subtype_Indication :=
     Internal_Entity_Subtype_Indication'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Internal_Discrete_Range;
      Block_Var : Internal_Entity_Range_Constraint;
Var_Expr : Internal_Entity_Range_Constraint;
Fld : Internal_Entity_Constraint;
Cast_Result : Internal_Entity_Range_Constraint;
Fld_1 : Internal_Entity_Range_Spec;
Fld_2 : Internal_Entity_Expr;
Fld_3 : Internal_Discrete_Range;
Scope_Result : Internal_Discrete_Range;
Fld_4 : Internal_Entity_Base_Type_Decl;
Fld_5 : Internal_Discrete_Range;
Result_Var : Internal_Discrete_Range;
Let_Result : Internal_Discrete_Range;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11640 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11634 '<Cast to RangeConstraint at ast.py:8604>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8604
--# expr-start 11633 '<FieldAccess for constraint at ast.py:8604>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8604
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Constraint (Node => Ent.Node.Subtype_Indication_F_Constraint, Info => Ent.Info);
--# expr-done 11633
   if Fld.Node = null
      or else Fld.Node.Kind in Ada_Range_Constraint_Range
   then
      Cast_Result := Create_Internal_Entity_Range_Constraint
        (Node => Fld.Node,
         Info => Fld.Info);
   else
         raise Property_Error with "invalid object cast";
   end if;
--# expr-done 11634
Block_Var := Cast_Result; 
--# bind rc Block_Var
Var_Expr := Block_Var;
if Var_Expr /= No_Entity_Range_Constraint then
   --# scope-start
   --# bind r Var_Expr
--# expr-start 11637 '<FieldAccess for discrete_range at ast.py:8606>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8606
--# expr-start 11636 '<FieldAccess for range at ast.py:8606>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8606
--# expr-start 11635 '<FieldAccess for range at ast.py:8606>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8606
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Range_Spec (Node => Var_Expr.Node.Range_Constraint_F_Range, Info => Var_Expr.Info);
--# expr-done 11635
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Expr (Node => Fld_1.Node.Range_Spec_F_Range, Info => Fld_1.Info);
--# expr-done 11636
      if Fld_2.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.discrete_range
Fld_3 := Libadalang.Implementation.Expr_P_Discrete_Range (Node => Fld_2.Node, E_Info => Fld_2.Info);
--# end
--# expr-done 11637
Scope_Result := Fld_3;
   --# end
   Result_Var := Scope_Result;
else
   --# expr-start 11639 '<FieldAccess for discrete_range at ast.py:8609>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8609
--# expr-start 11638 '<FieldAccess for designated_type_decl at ast.py:8609>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8609
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start TypeExpr.designated_type_decl
Fld_4 := Libadalang.Implementation.Type_Expr_P_Designated_Type_Decl (Node => Ent.Node, E_Info => Ent.Info);
--# end
--# expr-done 11638
      if Fld_4.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BaseTypeDecl.discrete_range'
Fld_5 := Libadalang.Implementation.Dispatcher_Base_Type_Decl_P_Discrete_Range (Node => Fld_4.Node, E_Info => Fld_4.Info);
--# end
--# expr-done 11639
   Result_Var := Fld_5;
end if;
Let_Result := Result_Var; 
--# end
--# expr-done 11640
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subtype_Indication_P_Discrete_Range;
--# end
--# property-start SubtypeIndication.is_static_subtype /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8615
pragma Warnings (Off, "is not referenced");
 function Subtype_Indication_P_Is_Static_Subtype
  (Node : Bare_Subtype_Indication
      ; Imprecise_Fallback : Boolean
         := False
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Subtype_Indication := Bare_Subtype_Indication (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Subtype_Indication :=
     Internal_Entity_Subtype_Indication'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind imprecise_fallback Imprecise_Fallback
   Property_Result : Boolean;
      Var_Expr : Internal_Entity_Constraint;
Fld : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld_1 : Internal_Entity_Constraint;
Fld_2 : Boolean;
Scope_Result : Boolean;
Fld_3 : Internal_Entity_Base_Type_Decl;
Fld_4 : Boolean;
Result_Var : Boolean;
Dyn_Var_Bind_Result : Boolean;
Let_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
            Populate_Lexical_Env (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11647 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11646 '<bind at ast.py:8619>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8619
--# expr-start 11641 '<FieldAccess for origin_node at ast.py:8619>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8619
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Self);
--# end
--# expr-done 11641
Bound_Origin := Fld; 
--# expr-start 11642 '<FieldAccess for constraint at ast.py:8619>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8619
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Constraint (Node => Ent.Node.Subtype_Indication_F_Constraint, Info => Ent.Info);
--# expr-done 11642
Var_Expr := Fld_1;
if Var_Expr /= No_Entity_Constraint then
   --# scope-start
   --# bind c Var_Expr
--# expr-start 11643 '<FieldAccess for is_static at ast.py:8620>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8620
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Constraint.is_static
Fld_2 := Libadalang.Implementation.Constraint_P_Is_Static (Node => Var_Expr.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Var_Expr.Info);
--# end
--# expr-done 11643
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   --# expr-start 11645 '<FieldAccess for is_static_decl at ast.py:8621>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8621
--# expr-start 11644 '<FieldAccess for designated_type at ast.py:8621>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8621
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_3 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Ent.Node, Origin => Bound_Origin, E_Info => Ent.Info);
--# end
--# expr-done 11644
      if Fld_3.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicDecl.is_static_decl'
Fld_4 := Libadalang.Implementation.Dispatcher_Basic_Decl_P_Is_Static_Decl (Node => Fld_3.Node, Imprecise_Fallback => Imprecise_Fallback, E_Info => Fld_3.Info);
--# end
--# expr-done 11645
   Result_Var := Fld_4;
end if;
Dyn_Var_Bind_Result := Result_Var; 
--# expr-done 11646
Let_Result := Dyn_Var_Bind_Result; 
--# end
--# expr-done 11647
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subtype_Indication_P_Is_Static_Subtype;
--# end
--# property-start SubtypeIndication.custom_id_text /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8624
pragma Warnings (Off, "is not referenced");
 function Subtype_Indication_P_Custom_Id_Text
  (Node : Bare_Subtype_Indication
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return String_Type
is
   Self : Bare_Subtype_Indication := Bare_Subtype_Indication (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Subtype_Indication :=
     Internal_Entity_Subtype_Indication'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : String_Type;
            procedure Finalizer_Scope_705 with Inline_Always;
      Cast_Expr : Bare_Subtype_Indication;
Cast_Result : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
Fld : Internal_Entity_Base_Type_Decl;
Fld_1 : String_Type;
Dyn_Var_Bind_Result : String_Type;
            procedure Finalizer_Scope_705 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Dyn_Var_Bind_Result);
            end Finalizer_Scope_705;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11650 '<bind at ast.py:8624>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8624
Cast_Expr := Self; 
      Cast_Result := Cast_Expr;
Bound_Origin := Cast_Result; 
--# expr-start 11649 '<FieldAccess for canonical_fully_qualified_name at ast.py:8626>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8626
--# expr-start 11648 '<FieldAccess for designated_type at ast.py:8626>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8626
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Ent.Node, Origin => Bound_Origin, E_Info => Ent.Info);
--# end
--# expr-done 11648
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start BasicDecl.canonical_fully_qualified_name
Fld_1 := Libadalang.Implementation.Basic_Decl_P_Canonical_Fully_Qualified_Name (Node => Fld.Node, E_Info => Fld.Info);
--# end
--# expr-done 11649
Dyn_Var_Bind_Result := Fld_1; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 11650
      Property_Result := Dyn_Var_Bind_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_705;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_705;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Subtype_Indication_P_Custom_Id_Text;
--# end
   --
   --  Primitives for Bare_Constrained_Subtype_Indication
   --
      procedure Initialize_Fields_For_Constrained_Subtype_Indication
        (Self : Bare_Constrained_Subtype_Indication
         ; Subtype_Indication_F_Has_Not_Null : Bare_Not_Null
         ; Subtype_Indication_F_Name : Bare_Name
         ; Subtype_Indication_F_Constraint : Bare_Constraint
        ) is
      begin
            Initialize_Fields_For_Subtype_Indication
              (Self, Subtype_Indication_F_Has_Not_Null, Subtype_Indication_F_Name, Subtype_Indication_F_Constraint);
      end Initialize_Fields_For_Constrained_Subtype_Indication;
   --
   --  Primitives for Bare_Discrete_Subtype_Indication
   --
      procedure Initialize_Fields_For_Discrete_Subtype_Indication
        (Self : Bare_Discrete_Subtype_Indication
         ; Subtype_Indication_F_Has_Not_Null : Bare_Not_Null
         ; Subtype_Indication_F_Name : Bare_Name
         ; Subtype_Indication_F_Constraint : Bare_Constraint
        ) is
      begin
            Initialize_Fields_For_Subtype_Indication
              (Self, Subtype_Indication_F_Has_Not_Null, Subtype_Indication_F_Name, Subtype_Indication_F_Constraint);
      end Initialize_Fields_For_Discrete_Subtype_Indication;
   --
   --  Primitives for Bare_Synthetic_Type_Expr
   --
      procedure Initialize_Fields_For_Synthetic_Type_Expr
        (Self : Bare_Synthetic_Type_Expr
         ; Synthetic_Type_Expr_F_Target_Type : Bare_Base_Type_Decl
        ) is
      begin
            Self.Synthetic_Type_Expr_F_Target_Type := Synthetic_Type_Expr_F_Target_Type;
      end Initialize_Fields_For_Synthetic_Type_Expr;
   function Synthetic_Type_Expr_F_Target_Type
     (Node : Bare_Synthetic_Type_Expr) return Bare_Base_Type_Decl
   is
   begin
         return Node.Synthetic_Type_Expr_F_Target_Type;
   end;
--# property-start SyntheticTypeExpr.designated_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16693
pragma Warnings (Off, "is not referenced");
 function Synthetic_Type_Expr_P_Designated_Type
  (Node : Bare_Synthetic_Type_Expr
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Synthetic_Type_Expr := Bare_Synthetic_Type_Expr (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Synthetic_Type_Expr :=
     Internal_Entity_Synthetic_Type_Expr'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1189 with Inline_Always;
      Block_Var : Internal_Entity_Generic_Decl;
Block_Var_1 : Env_Rebindings;
Var_Expr : Internal_Entity_Basic_Decl;
Var_Expr_1 : Internal_Entity_Generic_Decl;
Fld : Bare_Base_Type_Decl;
Node_Parents : Bare_Ada_Node_Array_Access;
Is_A : Boolean;
Find_Result : Bare_Ada_Node;
Cast_Result : Bare_Generic_Decl;
As_Entity : Internal_Entity_Generic_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Generic_Decl;
Fld_1 : Internal_Entity_Info;
Fld_2 : Env_Rebindings;
Fld_3 : Internal_Entity_Basic_Decl;
Scope_Result : Internal_Entity_Basic_Decl;
Result_Var : Internal_Entity_Basic_Decl;
Fld_4 : Internal_Entity_Info;
Fld_5 : Env_Rebindings;
Scope_Result_1 : Env_Rebindings;
Result_Var_1 : Env_Rebindings;
Fld_6 : Bare_Base_Type_Decl;
Fld_7 : Internal_Entity_Info;
Fld_8 : Boolean;
New_Struct : Internal_Entity_Info;
New_Struct_1 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
            procedure Finalizer_Scope_1189 is
            begin
                     Dec_Ref (Node_Parents);
            end Finalizer_Scope_1189;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11667 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11656 '<as_bare_entity at ast.py:16707>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16707
--# expr-start 11655 '<Cast to GenericDecl at ast.py:16707>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16707
--# expr-start 11654 '<Find at ast.py:16707>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16707
   --# expr-start 11652 '<FieldAccess for parents at ast.py:16707>' Node_Parents /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16707
--# expr-start 11651 '<FieldAccess for target_type at ast.py:16707>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16707
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Synthetic_Type_Expr_F_Target_Type;
--# expr-done 11651
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Node_Parents := Parents (Fld, True); 
--# expr-done 11652
      Find_Result := No_Bare_Ada_Node;
      declare
         Collection : constant Bare_Ada_Node_Array_Access := Node_Parents;
      begin
         for Item_342 of
               Collection.Items
         loop
   --# scope-start
               --# bind p Item_342
      --# expr-start 11653 '<IsA GenericDecl at ast.py:16708>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16708
Is_A := Item_342 /= null 
and then Item_342.Kind in Ada_Generic_Decl; 
--# expr-done 11653
      if Is_A then
         Find_Result := Item_342;
      end if;
   --# end
               exit when Is_A;
         end loop;
      end;
--# expr-done 11654
   if Find_Result = null
      or else Find_Result.Kind in Ada_Generic_Decl
   then
      Cast_Result := Find_Result;
   else
         Cast_Result := No_Bare_Ada_Node;
   end if;
--# expr-done 11655
Is_Null := Cast_Result = null; 
if Is_Null then
   If_Result := No_Entity_Generic_Decl;
else
As_Entity := (Info => No_Entity_Info, Node => Cast_Result); 
   If_Result := As_Entity;
end if;
--# expr-done 11656
Block_Var := If_Result; 
--# bind gd Block_Var
Var_Expr_1 := Block_Var;
if Var_Expr_1 /= No_Entity_Generic_Decl then
   --# scope-start
--# expr-start 11659 '<FieldAccess for unshed_rebindings at ast.py:16712>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16712
      if Var_Expr_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11658 '<FieldAccess for rebindings at ast.py:16713>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16713
--# expr-start 11657 '<FieldAccess for info at ast.py:16713>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16713
Fld_1 := Ent.Info;
--# expr-done 11657
Fld_2 := Fld_1.Rebindings;
--# expr-done 11658
--# property-call-start BasicDecl.unshed_rebindings
Fld_3 := Libadalang.Implementation.Basic_Decl_P_Unshed_Rebindings (Node => Var_Expr_1.Node, Rebindings => Fld_2, E_Info => Var_Expr_1.Info);
--# end
--# expr-done 11659
Scope_Result := Fld_3;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Basic_Decl;
end if;
Var_Expr := Result_Var;
if Var_Expr /= No_Entity_Basic_Decl then
   --# scope-start
   --# bind fixed Var_Expr
--# expr-start 11661 '<FieldAccess for rebindings at ast.py:16715>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16715
--# expr-start 11660 '<FieldAccess for info at ast.py:16715>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16715
Fld_4 := Var_Expr.Info;
--# expr-done 11660
Fld_5 := Fld_4.Rebindings;
--# expr-done 11661
Scope_Result_1 := Fld_5;
   --# end
   Result_Var_1 := Scope_Result_1;
else
   Result_Var_1 := null;
end if;
Block_Var_1 := Result_Var_1; 
--# bind relevant_rebindings Block_Var_1
--# expr-start 11666 '<New BaseTypeDecl.entity at ???>' New_Struct_1 None
--# expr-start 11664 '<New EntityInfo at ast.py:16722>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16722
--# expr-start 11663 '<FieldAccess for from_rebound at ast.py:16725>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16725
--# expr-start 11662 '<FieldAccess for info at ast.py:16725>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16725
Fld_7 := Ent.Info;
--# expr-done 11662
Fld_8 := Fld_7.From_Rebound;
--# expr-done 11663
New_Struct := (From_Rebound => Fld_8, Md => No_Metadata, Rebindings => Block_Var_1); 
--# expr-done 11664
--# expr-start 11665 '<FieldAccess for target_type at ast.py:16721>' Fld_6 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:16721
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_6 := Self.Synthetic_Type_Expr_F_Target_Type;
--# expr-done 11665
New_Struct_1 := (Info => New_Struct, Node => Fld_6); 
--# expr-done 11666
Let_Result := New_Struct_1; 
--# end
--# expr-done 11667
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_1189;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1189;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Synthetic_Type_Expr_P_Designated_Type;
--# end
   --
   --  Primitives for Bare_Unconstrained_Array_Index
   --
      procedure Initialize_Fields_For_Unconstrained_Array_Index
        (Self : Bare_Unconstrained_Array_Index
         ; Unconstrained_Array_Index_F_Subtype_Indication : Bare_Subtype_Indication
        ) is
      begin
            Self.Unconstrained_Array_Index_F_Subtype_Indication := Unconstrained_Array_Index_F_Subtype_Indication;
      end Initialize_Fields_For_Unconstrained_Array_Index;
   function Unconstrained_Array_Index_F_Subtype_Indication
     (Node : Bare_Unconstrained_Array_Index) return Bare_Subtype_Indication
   is
   begin
         return Node.Unconstrained_Array_Index_F_Subtype_Indication;
   end;
--# property-start UnconstrainedArrayIndex.designated_type /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20302
pragma Warnings (Off, "is not referenced");
 function Unconstrained_Array_Index_P_Designated_Type
  (Node : Bare_Unconstrained_Array_Index
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Type_Decl
is
   Self : Bare_Unconstrained_Array_Index := Bare_Unconstrained_Array_Index (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Unconstrained_Array_Index :=
     Internal_Entity_Unconstrained_Array_Index'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind origin Origin
   Property_Result : Internal_Entity_Base_Type_Decl;
      Fld : Internal_Entity_Subtype_Indication;
Fld_1 : Internal_Entity_Base_Type_Decl;
Let_Result : Internal_Entity_Base_Type_Decl;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11670 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11669 '<FieldAccess for designated_type at ast.py:20303>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20303
--# expr-start 11668 '<FieldAccess for subtype_indication at ast.py:20303>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:20303
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Subtype_Indication (Node => Ent.Node.Unconstrained_Array_Index_F_Subtype_Indication, Info => Ent.Info);
--# expr-done 11668
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]TypeExpr.designated_type'
Fld_1 := Libadalang.Implementation.Dispatcher_Type_Expr_P_Designated_Type (Node => Fld.Node, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11669
Let_Result := Fld_1; 
--# end
--# expr-done 11670
      Property_Result := Let_Result;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Unconstrained_Array_Index_P_Designated_Type;
--# end
   --
   --  Primitives for Bare_Until_Node
   --
--# property-start '[dispatcher]Until.as_bool' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_Until_Node_P_As_Bool
  (Node : Bare_Until_Node
  )
   return Boolean
is
   Self : Bare_Until_Node := Bare_Until_Node (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_Until_Node (Self.Kind) is
               when Ada_Until_Absent =>
                  --# property-call-start Until.Absent.as_bool
                  Property_Result := Until_Absent_P_As_Bool
                    (Self
                    );
                  --# end
               when Ada_Until_Present =>
                  --# property-call-start Until.Present.as_bool
                  Property_Result := Until_Present_P_As_Bool
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_Until_Node_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Until_Absent
   --
--# property-start Until.Absent.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4946
pragma Warnings (Off, "is not referenced");
 function Until_Absent_P_As_Bool
  (Node : Bare_Until_Absent
  )
   return Boolean
is
   Self : Bare_Until_Absent := Bare_Until_Absent (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Until_Absent_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Until_Present
   --
--# property-start Until.Present.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4946
pragma Warnings (Off, "is not referenced");
 function Until_Present_P_As_Bool
  (Node : Bare_Until_Present
  )
   return Boolean
is
   Self : Bare_Until_Present := Bare_Until_Present (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Until_Present_P_As_Bool;
--# end
   --
   --  Primitives for Bare_Use_Clause
   --
--# property-start UseClause.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8273
pragma Warnings (Off, "is not referenced");
 function Use_Clause_P_Xref_Entry_Point
  (Node : Bare_Use_Clause
  )
   return Boolean
is
   Self : Bare_Use_Clause := Bare_Use_Clause (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Use_Clause_P_Xref_Entry_Point;
--# end
--# property-start UseClause.used_envs /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8276
pragma Warnings (Off, "is not referenced");
 function Use_Clause_P_Used_Envs
  (Node : Bare_Use_Clause
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Use_Clause := Bare_Use_Clause (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Use_Clause :=
     Internal_Entity_Use_Clause'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_679 with Inline_Always;
            procedure Finalizer_Scope_3119 with Inline_Always;
            procedure Finalizer_Scope_3120 with Inline_Always;
            procedure Finalizer_Scope_3121 with Inline_Always;
      Match_0 : Internal_Entity_Use_Package_Clause;
Match_1 : Internal_Entity_Use_Type_Clause;
Fld : Lexical_Env_Array_Access;
Group_Env : Lexical_Env := Empty_Env;
Fld_1 : Internal_Entity_Name_List;
Fld_2 : Bare_Name_List;
Fld_3 : Internal_Entity_Info;
As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld_4 : Lexical_Env := Empty_Env;
Item_343 : Internal_Entity_Name;
Bare_Item_343 : Bare_Name;
Map_Result : Lexical_Env_Array_Access;
Group_Env_1 : Lexical_Env := Empty_Env;
Match_Prefix : Internal_Entity_Use_Clause;
Cast_Result : Internal_Entity_Use_Package_Clause;
Let_Result : Lexical_Env := Empty_Env;
Scope_Result : Lexical_Env := Empty_Env;
Cast_Result_1 : Internal_Entity_Use_Type_Clause;
Let_Result_1 : Lexical_Env := Empty_Env;
Scope_Result_1 : Lexical_Env := Empty_Env;
Match_Result : Lexical_Env := Empty_Env;
Let_Result_2 : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_679 is
            begin
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Match_Result);
                     Dec_Ref (Let_Result_2);
            end Finalizer_Scope_679;
            procedure Finalizer_Scope_3119 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Group_Env);
            end Finalizer_Scope_3119;
            procedure Finalizer_Scope_3120 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Group_Env_1);
            end Finalizer_Scope_3120;
            procedure Finalizer_Scope_3121 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_3121;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11678 '<Block at ???>' Let_Result_2 None
--# scope-start
--# expr-start 11677 '<Match at ast.py:8281>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8281
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Ent; 
case Ada_Use_Clause (Match_Prefix.Node.Kind) is
         when Ada_Use_Package_Clause =>
   --# scope-start
--# scope-start
      Cast_Result := Create_Internal_Entity_Use_Package_Clause
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result; 
--# bind upc Match_0
--# expr-start 11672 '<env_group at ast.py:8282>' Group_Env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8282
--# expr-start 11671 '<FieldAccess for designated_envs at ast.py:8282>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8282
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start UsePackageClause.designated_envs
Fld := Libadalang.Implementation.Use_Package_Clause_P_Designated_Envs (Node => Match_0.Node, E_Info => Match_0.Info);
--# end
--# expr-done 11671
Group_Env := Group (Fld, No_Metadata); 
--# expr-done 11672
Let_Result := Group_Env; Inc_Ref (Let_Result);
--# end
Scope_Result := Let_Result;
   Inc_Ref (Scope_Result);
   --# end
      Finalizer_Scope_3119;
            Match_Result := Scope_Result; Inc_Ref (Match_Result);
         when Ada_Use_Type_Clause =>
   --# scope-start
--# scope-start
      Cast_Result_1 := Create_Internal_Entity_Use_Type_Clause
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_1; 
--# bind utc Match_1
--# expr-start 11676 '<env_group at ast.py:8284>' Group_Env_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8284
--# expr-start 11675 '<Map at ast.py:8284>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8284
   --# expr-start 11673 '<FieldAccess for types at ast.py:8284>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8284
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name_List (Node => Match_1.Node.Use_Type_Clause_F_Types, Info => Match_1.Info);
--# expr-done 11673
Fld_2 := Fld_1.Node;
      declare
         Map_Result_Vec : Lexical_Env_Vectors.Vector;
      begin
      if Fld_2 = null then
      Map_Result := Create_Lexical_Env_Array (0);
      else
      declare
         Collection : constant Bare_Name_List := Fld_2;
      begin
         for Untyped_Item_343 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_343 := Untyped_Item_343; 
Is_Null := Bare_Item_343 = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
Fld_3 := Fld_1.Info;
As_Entity := (Info => Fld_3, Node => Bare_Item_343); 
   If_Result := As_Entity;
end if;
                  Item_343 := If_Result; 
   --# scope-start
               --# bind n Item_343
      --# expr-start 11674 '<FieldAccess for name_designated_type_env at ast.py:8284>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8284
      if Item_343.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type_env
Fld_4 := Libadalang.Implementation.Name_P_Name_Designated_Type_Env (Node => Item_343.Node, E_Info => Item_343.Info);
--# end
--# expr-done 11674
         declare
            Item_To_Append : constant Lexical_Env := Fld_4;
         begin
               Inc_Ref (Item_To_Append);
            Lexical_Env_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_3121;
         end loop;
      end;
      end if;
         Map_Result := Create_Lexical_Env_Array
           (Items_Count => Natural (Lexical_Env_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Lexical_Env_Vectors.Get
              (Map_Result_Vec,
               I + Lexical_Env_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Lexical_Env_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 11675
Group_Env_1 := Group (Map_Result, No_Metadata); 
--# expr-done 11676
Let_Result_1 := Group_Env_1; Inc_Ref (Let_Result_1);
--# end
Scope_Result_1 := Let_Result_1;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_3120;
            Match_Result := Scope_Result_1; Inc_Ref (Match_Result);
end case;
--# expr-done 11677
Let_Result_2 := Match_Result; Inc_Ref (Let_Result_2);
--# end
--# expr-done 11678
      Property_Result := Let_Result_2;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_679;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_679;
               Finalizer_Scope_3119;
               Finalizer_Scope_3120;
               Finalizer_Scope_3121;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Use_Clause_P_Used_Envs;
--# end
--# property-start UseClause.initial_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8288
pragma Warnings (Off, "is not referenced");
 function Use_Clause_P_Initial_Env
  (Node : Bare_Use_Clause
  )
   return Internal_Designated_Env
is
   Self : Bare_Use_Clause := Bare_Use_Clause (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_680 with Inline_Always;
      Sym : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
New_Struct_1 : Internal_Designated_Env := No_Designated_Env;
Fld : Bare_Ada_Node;
Fld_1 : Bare_Ada_Node;
Is_A : Boolean;
If_Result : Internal_Designated_Env := No_Designated_Env;
Let_Result : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_680 is
            begin
                     Dec_Ref (New_Struct);
                     Dec_Ref (New_Struct_1);
                     Dec_Ref (If_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_680;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11686 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11685 '<If at ast.py:8293>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8293
--# expr-start 11681 '<IsA CompilationUnit at ast.py:8294>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8294
--# expr-start 11680 '<parent at ast.py:8294>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8294
--# expr-start 11679 '<parent at ast.py:8294>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8294
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 11679
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Parent; 
--# expr-done 11680
Is_A := Fld_1 /= null 
and then Fld_1.Kind in Ada_Compilation_Unit_Range; 
--# expr-done 11681
if Is_A then
   --# expr-start 11683 '<New DesignatedEnv at ast.py:8295>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8295
--# expr-start 11682 '<Symbol Standard at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Standard); 
--# expr-done 11682
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => Sym, Kind => Named_Env); 
--# expr-done 11683
   If_Result := New_Struct;
else
   --# expr-start 11684 '<New DesignatedEnv at ast.py:8296>' New_Struct_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8296
Inc_Ref (Empty_Env);
New_Struct_1 := (Direct_Env => Empty_Env, Env_Name => null, Kind => Current_Env); 
--# expr-done 11684
   If_Result := New_Struct_1;
end if;
      Inc_Ref (If_Result);
--# expr-done 11685
Let_Result := If_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 11686
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_680;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_680;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Use_Clause_P_Initial_Env;
--# end
   --
   --  Primitives for Bare_Use_Package_Clause
   --
      procedure Use_Package_Clause_Pre_Env_Actions
        (Self            : Bare_Use_Package_Clause;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Env : constant Internal_Designated_Env :=
           Libadalang.Implementation.Initial_Env_18 (Node => 
Self);
      begin
         Set_Initial_Env
           (Self,
            State,
            Env,
            DSL_Location => "ast.py:8307");
      end;
      end;
      procedure Use_Package_Clause_Post_Env_Actions
        (Self : Bare_Use_Package_Clause; State : in out PLE_Node_State) is
      begin
      if Libadalang.Implementation.Ref_Cond_20 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_19 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Use_Package_Name_Designated_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      end;
      procedure Initialize_Fields_For_Use_Package_Clause
        (Self : Bare_Use_Package_Clause
         ; Use_Package_Clause_F_Packages : Bare_Name_List
        ) is
      begin
            Self.Use_Package_Clause_F_Packages := Use_Package_Clause_F_Packages;
      end Initialize_Fields_For_Use_Package_Clause;
   function Use_Package_Clause_F_Packages
     (Node : Bare_Use_Package_Clause) return Bare_Name_List
   is
   begin
         return Node.Use_Package_Clause_F_Packages;
   end;
--# property-start UsePackageClause.designated_env /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8320
pragma Warnings (Off, "is not referenced");
 function Use_Package_Clause_P_Designated_Env
  (Node : Bare_Use_Package_Clause
      ; Index : Integer
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env
is
   Self : Bare_Use_Package_Clause := Bare_Use_Package_Clause (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Use_Package_Clause :=
     Internal_Entity_Use_Package_Clause'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind index Index
   Property_Result : Lexical_Env;
            procedure Finalizer_Scope_681 with Inline_Always;
      Block_Var : Bare_Name;
Fld : Bare_Name_List;
Get_Result : Bare_Name;
Fld_1 : Lexical_Env := Empty_Env;
Bound_Env : Lexical_Env := Empty_Env;
Fld_2 : Bare_Ada_Node;
Bound_Origin : Bare_Ada_Node;
As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld_3 : Lexical_Env := Empty_Env;
Dyn_Var_Bind_Result : Lexical_Env := Empty_Env;
Dyn_Var_Bind_Result_1 : Lexical_Env := Empty_Env;
Let_Result : Lexical_Env := Empty_Env;
            procedure Finalizer_Scope_681 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Bound_Env);
                     Dec_Ref (Fld_3);
                     Dec_Ref (Dyn_Var_Bind_Result);
                     Dec_Ref (Dyn_Var_Bind_Result_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_681;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11695 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11688 '<at at ast.py:8325>' Get_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8325
--# expr-start 11687 '<FieldAccess for packages at ast.py:8325>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8325
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Use_Package_Clause_F_Packages;
--# expr-done 11687
Get_Result := Get (Fld, Index, True); 
--# expr-done 11688
Block_Var := Get_Result; 
--# bind pkg Block_Var
--# expr-start 11694 '<bind at ast.py:8326>' Dyn_Var_Bind_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8326
--# expr-start 11689 '<FieldAccess for node_env at ast.py:8327>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8327
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Libadalang.Implementation.Node_Env (Node => Ent.Node, E_Info => Ent.Info);
--# expr-done 11689
Bound_Env := Fld_1; Inc_Ref (Bound_Env);
--# expr-start 11693 '<bind at ast.py:8328>' Dyn_Var_Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8328
--# expr-start 11690 '<FieldAccess for origin_node at ast.py:8329>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8329
      if Block_Var = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.origin_node
Fld_2 := Libadalang.Implementation.Ada_Node_P_Origin_Node (Node => Block_Var);
--# end
--# expr-done 11690
Bound_Origin := Fld_2; 
--# expr-start 11692 '<FieldAccess for designated_env at ast.py:8330>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8330
--# expr-start 11691 '<as_bare_entity at ast.py:8330>' If_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8330
Is_Null := Block_Var = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
As_Entity := (Info => No_Entity_Info, Node => Block_Var); 
   If_Result := As_Entity;
end if;
--# expr-done 11691
      if If_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Expr.designated_env'
Fld_3 := Libadalang.Implementation.Dispatcher_Expr_P_Designated_Env (Node => If_Result.Node, Env => Bound_Env, Origin => Bound_Origin, No_Visibility => False, E_Info => If_Result.Info);
--# end
--# expr-done 11692
Dyn_Var_Bind_Result := Fld_3; Inc_Ref (Dyn_Var_Bind_Result);
--# expr-done 11693
Dyn_Var_Bind_Result_1 := Dyn_Var_Bind_Result; Inc_Ref (Dyn_Var_Bind_Result_1);
--# expr-done 11694
Let_Result := Dyn_Var_Bind_Result_1; Inc_Ref (Let_Result);
--# end
--# expr-done 11695
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_681;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_681;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Use_Package_Clause_P_Designated_Env;
--# end
--# property-start UsePackageClause.designated_envs /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8335
pragma Warnings (Off, "is not referenced");
 function Use_Package_Clause_P_Designated_Envs
  (Node : Bare_Use_Package_Clause
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Lexical_Env_Array_Access
is
   Self : Bare_Use_Package_Clause := Bare_Use_Package_Clause (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Use_Package_Clause :=
     Internal_Entity_Use_Package_Clause'(Node => Self, Info => E_Info);
   --# bind entity Ent
   Property_Result : Lexical_Env_Array_Access;
            procedure Finalizer_Scope_682 with Inline_Always;
            procedure Finalizer_Scope_3122 with Inline_Always;
      Fld : Bare_Name_List;
Fld_1 : Lexical_Env := Empty_Env;
Item_344 : Bare_Name;
Index_345 : Integer;
Map_Result : Lexical_Env_Array_Access;
Let_Result : Lexical_Env_Array_Access;
            procedure Finalizer_Scope_682 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_682;
            procedure Finalizer_Scope_3122 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_3122;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11699 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11698 '<Map at ast.py:8342>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8342
   --# expr-start 11696 '<FieldAccess for packages at ast.py:8342>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8342
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Use_Package_Clause_F_Packages;
--# expr-done 11696
      declare
         Map_Result_Vec : Lexical_Env_Vectors.Vector;
      begin
      if Fld = null then
      Map_Result := Create_Lexical_Env_Array (0);
      else
         Index_345 := 0;
      declare
         Collection : constant Bare_Name_List := Fld;
      begin
         for Untyped_Item_344 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_344 := Untyped_Item_344; 
   --# scope-start
               --# bind _ Item_344
               --# bind i Index_345
      --# expr-start 11697 '<FieldAccess for designated_env at ast.py:8343>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8343
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start UsePackageClause.designated_env
Fld_1 := Libadalang.Implementation.Use_Package_Clause_P_Designated_Env (Node => Ent.Node, Index => Index_345, E_Info => Ent.Info);
--# end
--# expr-done 11697
         declare
            Item_To_Append : constant Lexical_Env := Fld_1;
         begin
               Inc_Ref (Item_To_Append);
            Lexical_Env_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_3122;
               Index_345 := Index_345 + 1;
         end loop;
      end;
      end if;
         Map_Result := Create_Lexical_Env_Array
           (Items_Count => Natural (Lexical_Env_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Lexical_Env_Vectors.Get
              (Map_Result_Vec,
               I + Lexical_Env_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Lexical_Env_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 11698
Let_Result := Map_Result; Inc_Ref (Let_Result);
--# end
--# expr-done 11699
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_682;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_682;
               Finalizer_Scope_3122;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Use_Package_Clause_P_Designated_Envs;
--# end
--# property-start UsePackageClause.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8346
pragma Warnings (Off, "is not referenced");
 function Use_Package_Clause_P_Xref_Equation
  (Node : Bare_Use_Package_Clause
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Use_Package_Clause := Bare_Use_Package_Clause (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Use_Package_Clause :=
     Internal_Entity_Use_Package_Clause'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_683 with Inline_Always;
            procedure Finalizer_Scope_3123 with Inline_Always;
      Fld : Internal_Entity_Name_List;
Fld_1 : Bare_Name_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld_3 : Logic_Equation;
Item_346 : Internal_Entity_Name;
Bare_Item_346 : Bare_Name;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
            procedure Finalizer_Scope_683 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
            end Finalizer_Scope_683;
            procedure Finalizer_Scope_3123 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_3123;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11703 '<LogicAll at ast.py:8347>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8347
--# expr-start 11702 '<Map at ast.py:8347>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8347
   --# expr-start 11700 '<FieldAccess for packages at ast.py:8347>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8347
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name_List (Node => Ent.Node.Use_Package_Clause_F_Packages, Info => Ent.Info);
--# expr-done 11700
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Name_List := Fld_1;
      begin
         for Untyped_Item_346 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_346 := Untyped_Item_346; 
Is_Null := Bare_Item_346 = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_346); 
   If_Result := As_Entity;
end if;
                  Item_346 := If_Result; 
   --# scope-start
               --# bind p Item_346
      --# expr-start 11701 '<FieldAccess for xref_no_overloading at ast.py:8347>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8347
      if Item_346.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_3 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Item_346.Node, Env => Env, Origin => Origin, E_Info => Item_346.Info);
--# end
--# expr-done 11701
         declare
            Item_To_Append : constant Logic_Equation := Fld_3;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_3123;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 11702
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8347") else null)); 
--# expr-done 11703
      Property_Result := Logic_Boolean_Op;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_683;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_683;
               Finalizer_Scope_3123;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Use_Package_Clause_P_Xref_Equation;
--# end
--# property-start UsePackageClause._initial_env_18 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8307
pragma Warnings (Off, "is not referenced");
 function Initial_Env_18
  (Node : Bare_Use_Package_Clause
  )
   return Internal_Designated_Env
is
   Self : Bare_Use_Package_Clause := Bare_Use_Package_Clause (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_1435 with Inline_Always;
      Fld : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_1435 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1435;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11704 '<FieldAccess for initial_env at ast.py:8307>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8307
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start UseClause.initial_env
Fld := Libadalang.Implementation.Use_Clause_P_Initial_Env (Node => Self);
--# end
--# expr-done 11704
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1435;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1435;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Initial_Env_18;
--# end
--# property-start UsePackageClause._ref_env_nodes_19 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8313
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_19
  (Node : Bare_Use_Package_Clause
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Use_Package_Clause := Bare_Use_Package_Clause (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1436 with Inline_Always;
      Fld : Bare_Name_List;
Cast_Expr : Bare_Name;
Cast_Result : Bare_Ada_Node;
Item_347 : Bare_Name;
Map_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1436 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_1436;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11707 '<Map at ast.py:8313>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8313
   --# expr-start 11705 '<FieldAccess for packages at ast.py:8313>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8313
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Use_Package_Clause_F_Packages;
--# expr-done 11705
      declare
         Map_Result_Vec : Bare_Ada_Node_Vectors.Vector;
      begin
      if Fld = null then
      Map_Result := Create_Bare_Ada_Node_Array (0);
      else
      declare
         Collection : constant Bare_Name_List := Fld;
      begin
         for Untyped_Item_347 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_347 := Untyped_Item_347; 
   --# scope-start
               --# bind n Item_347
      --# expr-start 11706 '<Cast to AdaNode at ast.py:8313>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8313
Cast_Expr := Item_347; 
      Cast_Result := Cast_Expr;
--# expr-done 11706
         declare
            Item_To_Append : constant Bare_Ada_Node := Cast_Result;
         begin
            Bare_Ada_Node_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Bare_Ada_Node_Array
           (Items_Count => Natural (Bare_Ada_Node_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Bare_Ada_Node_Vectors.Get
              (Map_Result_Vec,
               I + Bare_Ada_Node_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Bare_Ada_Node_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 11707
      Property_Result := Map_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1436;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1436;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_19;
--# end
--# property-start UsePackageClause._ref_cond_20 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8315
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_20
  (Node : Bare_Use_Package_Clause
  )
   return Boolean
is
   Self : Bare_Use_Package_Clause := Bare_Use_Package_Clause (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Fld_1 : Bare_Ada_Node;
Is_A : Boolean;
Not_Val : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11711 '<Not at ast.py:8315>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8315
--# expr-start 11710 '<IsA CompilationUnit at ast.py:8315>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8315
--# expr-start 11709 '<parent at ast.py:8315>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8315
--# expr-start 11708 '<parent at ast.py:8315>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8315
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 11708
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Parent; 
--# expr-done 11709
Is_A := Fld_1 /= null 
and then Fld_1.Kind in Ada_Compilation_Unit_Range; 
--# expr-done 11710
Not_Val := not (Is_A); 
--# expr-done 11711
      Property_Result := Not_Val;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_20;
--# end
   --
   --  Primitives for Bare_Use_Type_Clause
   --
      procedure Use_Type_Clause_Pre_Env_Actions
        (Self            : Bare_Use_Type_Clause;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Env : constant Internal_Designated_Env :=
           Libadalang.Implementation.Initial_Env_21 (Node => 
Self);
      begin
         Set_Initial_Env
           (Self,
            State,
            Env,
            DSL_Location => "ast.py:8359");
      end;
      end;
      procedure Use_Type_Clause_Post_Env_Actions
        (Self : Bare_Use_Type_Clause; State : in out PLE_Node_State) is
      begin
      if Libadalang.Implementation.Ref_Cond_23 (Node => 
Self) then
         declare
            Ref_Env_Nodes : Bare_Ada_Node_Array_Access :=
               Libadalang.Implementation.Ref_Env_Nodes_22 (Node => 
Self);
            Env : Lexical_Env :=
              Self.Self_Env;
         begin
            Ref_Env
              (Self,
               Env,
               Ref_Env_Nodes,
               Name_P_Name_Designated_Type_Env'Access,
               Normal,
               All_Cats,
               False);
            Dec_Ref (Ref_Env_Nodes);
         end;
      end if;
      end;
      procedure Initialize_Fields_For_Use_Type_Clause
        (Self : Bare_Use_Type_Clause
         ; Use_Type_Clause_F_Has_All : Bare_All_Node
         ; Use_Type_Clause_F_Types : Bare_Name_List
        ) is
      begin
            Self.Use_Type_Clause_F_Has_All := Use_Type_Clause_F_Has_All;
            Self.Use_Type_Clause_F_Types := Use_Type_Clause_F_Types;
      end Initialize_Fields_For_Use_Type_Clause;
   function Use_Type_Clause_F_Has_All
     (Node : Bare_Use_Type_Clause) return Bare_All_Node
   is
   begin
         return Node.Use_Type_Clause_F_Has_All;
   end;
   function Use_Type_Clause_F_Types
     (Node : Bare_Use_Type_Clause) return Bare_Name_List
   is
   begin
         return Node.Use_Type_Clause_F_Types;
   end;
--# property-start UseTypeClause.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8371
pragma Warnings (Off, "is not referenced");
 function Use_Type_Clause_P_Xref_Equation
  (Node : Bare_Use_Type_Clause
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Use_Type_Clause := Bare_Use_Type_Clause (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Use_Type_Clause :=
     Internal_Entity_Use_Type_Clause'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_684 with Inline_Always;
            procedure Finalizer_Scope_3125 with Inline_Always;
      Fld : Internal_Entity_Name_List;
Fld_1 : Bare_Name_List;
Fld_2 : Internal_Entity_Info;
As_Entity : Internal_Entity_Name;
Is_Null : Boolean;
If_Result : Internal_Entity_Name;
Fld_3 : Logic_Equation;
Item_348 : Internal_Entity_Name;
Bare_Item_348 : Bare_Name;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
            procedure Finalizer_Scope_684 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
            end Finalizer_Scope_684;
            procedure Finalizer_Scope_3125 is
            begin
                     Dec_Ref (Fld_3);
            end Finalizer_Scope_3125;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11715 '<LogicAll at ast.py:8372>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8372
--# expr-start 11714 '<Map at ast.py:8372>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8372
   --# expr-start 11712 '<FieldAccess for types at ast.py:8372>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8372
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Name_List (Node => Ent.Node.Use_Type_Clause_F_Types, Info => Ent.Info);
--# expr-done 11712
Fld_1 := Fld.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_1 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Name_List := Fld_1;
      begin
         for Untyped_Item_348 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_348 := Untyped_Item_348; 
Is_Null := Bare_Item_348 = null; 
if Is_Null then
   If_Result := No_Entity_Name;
else
Fld_2 := Fld.Info;
As_Entity := (Info => Fld_2, Node => Bare_Item_348); 
   If_Result := As_Entity;
end if;
                  Item_348 := If_Result; 
   --# scope-start
               --# bind p Item_348
      --# expr-start 11713 '<FieldAccess for xref_no_overloading at ast.py:8372>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8372
      if Item_348.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_3 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Item_348.Node, Env => Env, Origin => Origin, E_Info => Item_348.Info);
--# end
--# expr-done 11713
         declare
            Item_To_Append : constant Logic_Equation := Fld_3;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_3125;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 11714
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8372") else null)); 
--# expr-done 11715
      Property_Result := Logic_Boolean_Op;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_684;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_684;
               Finalizer_Scope_3125;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Use_Type_Clause_P_Xref_Equation;
--# end
--# property-start UseTypeClause._initial_env_21 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8359
pragma Warnings (Off, "is not referenced");
 function Initial_Env_21
  (Node : Bare_Use_Type_Clause
  )
   return Internal_Designated_Env
is
   Self : Bare_Use_Type_Clause := Bare_Use_Type_Clause (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_1438 with Inline_Always;
      Fld : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_1438 is
            begin
                     Dec_Ref (Fld);
            end Finalizer_Scope_1438;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11716 '<FieldAccess for initial_env at ast.py:8359>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8359
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start UseClause.initial_env
Fld := Libadalang.Implementation.Use_Clause_P_Initial_Env (Node => Self);
--# end
--# expr-done 11716
      Property_Result := Fld;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1438;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1438;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Initial_Env_21;
--# end
--# property-start UseTypeClause._ref_env_nodes_22 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8365
pragma Warnings (Off, "is not referenced");
 function Ref_Env_Nodes_22
  (Node : Bare_Use_Type_Clause
  )
   return Bare_Ada_Node_Array_Access
is
   Self : Bare_Use_Type_Clause := Bare_Use_Type_Clause (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1439 with Inline_Always;
      Fld : Bare_Name_List;
Cast_Expr : Bare_Name;
Cast_Result : Bare_Ada_Node;
Item_349 : Bare_Name;
Map_Result : Bare_Ada_Node_Array_Access;
            procedure Finalizer_Scope_1439 is
            begin
                     Dec_Ref (Map_Result);
            end Finalizer_Scope_1439;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11719 '<Map at ast.py:8365>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8365
   --# expr-start 11717 '<FieldAccess for types at ast.py:8365>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8365
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Use_Type_Clause_F_Types;
--# expr-done 11717
      declare
         Map_Result_Vec : Bare_Ada_Node_Vectors.Vector;
      begin
      if Fld = null then
      Map_Result := Create_Bare_Ada_Node_Array (0);
      else
      declare
         Collection : constant Bare_Name_List := Fld;
      begin
         for Untyped_Item_349 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_349 := Untyped_Item_349; 
   --# scope-start
               --# bind n Item_349
      --# expr-start 11718 '<Cast to AdaNode at ast.py:8365>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8365
Cast_Expr := Item_349; 
      Cast_Result := Cast_Expr;
--# expr-done 11718
         declare
            Item_To_Append : constant Bare_Ada_Node := Cast_Result;
         begin
            Bare_Ada_Node_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
         end loop;
      end;
      end if;
         Map_Result := Create_Bare_Ada_Node_Array
           (Items_Count => Natural (Bare_Ada_Node_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Bare_Ada_Node_Vectors.Get
              (Map_Result_Vec,
               I + Bare_Ada_Node_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Bare_Ada_Node_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 11719
      Property_Result := Map_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1439;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1439;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Env_Nodes_22;
--# end
--# property-start UseTypeClause._ref_cond_23 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8367
pragma Warnings (Off, "is not referenced");
 function Ref_Cond_23
  (Node : Bare_Use_Type_Clause
  )
   return Boolean
is
   Self : Bare_Use_Type_Clause := Bare_Use_Type_Clause (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
      Fld : Bare_Ada_Node;
Fld_1 : Bare_Ada_Node;
Is_A : Boolean;
Not_Val : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11723 '<Not at ast.py:8367>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8367
--# expr-start 11722 '<IsA CompilationUnit at ast.py:8367>' Is_A /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8367
--# expr-start 11721 '<parent at ast.py:8367>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8367
--# expr-start 11720 '<parent at ast.py:8367>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8367
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.Parent; 
--# expr-done 11720
      if Fld = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Fld.Parent; 
--# expr-done 11721
Is_A := Fld_1 /= null 
and then Fld_1.Kind in Ada_Compilation_Unit_Range; 
--# expr-done 11722
Not_Val := not (Is_A); 
--# expr-done 11723
      Property_Result := Not_Val;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Ref_Cond_23;
--# end
   --
   --  Primitives for Bare_Value_Sequence
   --
      procedure Initialize_Fields_For_Value_Sequence
        (Self : Bare_Value_Sequence
         ; Value_Sequence_F_Iter_Assoc : Bare_Iterated_Assoc
        ) is
      begin
            Self.Value_Sequence_F_Iter_Assoc := Value_Sequence_F_Iter_Assoc;
      end Initialize_Fields_For_Value_Sequence;
   function Value_Sequence_F_Iter_Assoc
     (Node : Bare_Value_Sequence) return Bare_Iterated_Assoc
   is
   begin
         return Node.Value_Sequence_F_Iter_Assoc;
   end;
--# property-start ValueSequence.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18113
pragma Warnings (Off, "is not referenced");
 function Value_Sequence_P_Xref_Equation
  (Node : Bare_Value_Sequence
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Value_Sequence := Bare_Value_Sequence (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Value_Sequence :=
     Internal_Entity_Value_Sequence'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_1278 with Inline_Always;
      Fld : Internal_Entity_Iterated_Assoc;
Fld_1 : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_1278 is
            begin
                     Dec_Ref (Fld_1);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_1278;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11726 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11725 '<FieldAccess for xref_equation_for_reduce at ast.py:18117>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18117
--# expr-start 11724 '<FieldAccess for iter_assoc at ast.py:18117>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:18117
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Iterated_Assoc (Node => Ent.Node.Value_Sequence_F_Iter_Assoc, Info => Ent.Info);
--# expr-done 11724
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start IteratedAssoc.xref_equation_for_reduce
Fld_1 := Libadalang.Implementation.Iterated_Assoc_P_Xref_Equation_For_Reduce (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11725
Let_Result := Fld_1; Inc_Ref (Let_Result);
--# end
--# expr-done 11726
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1278;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1278;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Value_Sequence_P_Xref_Equation;
--# end
   --
   --  Primitives for Bare_Variant
   --
      procedure Initialize_Fields_For_Variant
        (Self : Bare_Variant
         ; Variant_F_Choices : Bare_Alternatives_List
         ; Variant_F_Components : Bare_Component_List
        ) is
      begin
            Self.Variant_F_Choices := Variant_F_Choices;
            Self.Variant_F_Components := Variant_F_Components;
      end Initialize_Fields_For_Variant;
   function Variant_F_Choices
     (Node : Bare_Variant) return Bare_Alternatives_List
   is
   begin
         return Node.Variant_F_Choices;
   end;
   function Variant_F_Components
     (Node : Bare_Variant) return Bare_Component_List
   is
   begin
         return Node.Variant_F_Components;
   end;
--# property-start Variant.matches /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4544
pragma Warnings (Off, "is not referenced");
 function Variant_P_Matches
  (Node : Bare_Variant
      ; Expr : Internal_Entity_Expr
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Boolean
is
   Self : Bare_Variant := Bare_Variant (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Variant :=
     Internal_Entity_Variant'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind expr Expr
   Property_Result : Boolean;
            procedure Finalizer_Scope_294 with Inline_Always;
      Block_Var : Big_Integer_Type;
Fld : Big_Integer_Type;
Fld_1 : Internal_Entity_Alternatives_List;
Fld_2 : Bare_Alternatives_List;
Fld_3 : Internal_Entity_Info;
As_Entity : Internal_Entity;
Is_Null : Boolean;
If_Result : Internal_Entity;
Fld_4 : Boolean;
Item_350 : Internal_Entity;
Bare_Item_350 : Bare_Ada_Node;
Quantifier_Result : Boolean;
Let_Result : Boolean;
            procedure Finalizer_Scope_294 is
            begin
                     Dec_Ref (Fld);
                     Dec_Ref (Block_Var);
            end Finalizer_Scope_294;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11731 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11727 '<FieldAccess for eval_as_int at ast.py:4549>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4549
      if Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.eval_as_int
Fld := Libadalang.Implementation.Expr_P_Eval_As_Int (Node => Expr.Node, E_Info => Expr.Info);
--# end
--# expr-done 11727
Block_Var := Fld; Inc_Ref (Block_Var);
--# bind expr_val Block_Var
--# expr-start 11730 '<AnyQuantifier at ast.py:4551>' Quantifier_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4551
   --# expr-start 11728 '<FieldAccess for choices at ast.py:4551>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4551
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Alternatives_List (Node => Ent.Node.Variant_F_Choices, Info => Ent.Info);
--# expr-done 11728
Fld_2 := Fld_1.Node;
      Quantifier_Result := False;
      if Fld_2 = null then
      null;
      else
      declare
         Collection : constant Bare_Alternatives_List := Fld_2;
      begin
         for Untyped_Item_350 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_350 := Untyped_Item_350; 
Is_Null := Bare_Item_350 = null; 
if Is_Null then
   If_Result := No_Entity;
else
Fld_3 := Fld_1.Info;
As_Entity := (Info => Fld_3, Node => Bare_Item_350); 
   If_Result := As_Entity;
end if;
                  Item_350 := If_Result; 
   --# scope-start
               --# bind c Item_350
      --# expr-start 11729 '<FieldAccess for choice_match at ast.py:4552>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4552
      if Item_350.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.choice_match
Fld_4 := Libadalang.Implementation.Ada_Node_P_Choice_Match (Node => Item_350.Node, Value => Block_Var, E_Info => Item_350.Info);
--# end
--# expr-done 11729
      Quantifier_Result := Fld_4;
   --# end
               exit when Quantifier_Result;
         end loop;
      end;
      end if;
--# expr-done 11730
Let_Result := Quantifier_Result; 
--# end
--# expr-done 11731
      Property_Result := Let_Result;
   --# end
      Finalizer_Scope_294;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_294;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Variant_P_Matches;
--# end
   --
   --  Primitives for Bare_Variant_Part
   --
      procedure Initialize_Fields_For_Variant_Part
        (Self : Bare_Variant_Part
         ; Variant_Part_F_Discr_Name : Bare_Identifier
         ; Variant_Part_F_Variant : Bare_Variant_List
        ) is
      begin
            Self.Variant_Part_F_Discr_Name := Variant_Part_F_Discr_Name;
            Self.Variant_Part_F_Variant := Variant_Part_F_Variant;
      end Initialize_Fields_For_Variant_Part;
   function Variant_Part_F_Discr_Name
     (Node : Bare_Variant_Part) return Bare_Identifier
   is
   begin
         return Node.Variant_Part_F_Discr_Name;
   end;
   function Variant_Part_F_Variant
     (Node : Bare_Variant_Part) return Bare_Variant_List
   is
   begin
         return Node.Variant_Part_F_Variant;
   end;
--# property-start VariantPart.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4566
pragma Warnings (Off, "is not referenced");
 function Variant_Part_P_Xref_Entry_Point
  (Node : Bare_Variant_Part
  )
   return Boolean
is
   Self : Bare_Variant_Part := Bare_Variant_Part (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Variant_Part_P_Xref_Entry_Point;
--# end
--# property-start VariantPart.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4569
pragma Warnings (Off, "is not referenced");
 function Variant_Part_P_Xref_Equation
  (Node : Bare_Variant_Part
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_Variant_Part := Bare_Variant_Part (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Variant_Part :=
     Internal_Entity_Variant_Part'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_296 with Inline_Always;
            procedure Finalizer_Scope_3128 with Inline_Always;
            procedure Finalizer_Scope_3129 with Inline_Always;
            procedure Finalizer_Scope_3130 with Inline_Always;
            procedure Finalizer_Scope_3132 with Inline_Always;
            procedure Finalizer_Scope_3133 with Inline_Always;
      Block_Var : Boolean;
Match_0 : Internal_Entity_Expr;
Match_1 : Internal_Entity_Subtype_Indication;
Match_2 : Internal_Entity_Others_Designator;
Match_3 : Internal_Entity;
Var_Expr : Internal_Entity_Name;
Fld : Internal_Entity_Identifier;
Fld_1 : Boolean;
Fld_2 : Internal_Entity_Variant_List;
Fld_3 : Bare_Variant_List;
Fld_4 : Internal_Entity_Info;
As_Entity : Internal_Entity_Variant;
Is_Null : Boolean;
If_Result : Internal_Entity_Variant;
Fld_5 : Internal_Entity_Alternatives_List;
Fld_6 : Bare_Alternatives_List;
Fld_7 : Internal_Entity_Info;
As_Entity_1 : Internal_Entity;
Is_Null_1 : Boolean;
If_Result_1 : Internal_Entity;
Cast_Expr : Internal_Entity_Expr;
Cast_Result : Internal_Entity_Name;
Fld_8 : Logic_Equation;
Fld_9 : Logic_Var;
Fld_10 : Bare_Identifier;
Fld_11 : Internal_Entity;
Bind_Result : Logic_Equation;
Fld_12 : Logic_Equation;
And_Pred : Logic_Equation;
Fld_13 : Logic_Equation;
And_Pred_1 : Logic_Equation;
Cast_Expr_1 : Internal_Entity_Expr;
Cast_Result_1 : Internal_Entity_Name;
Fld_14 : Internal_Entity_Base_Type_Decl;
Scope_Result : Internal_Entity_Base_Type_Decl;
Result_Var : Internal_Entity_Base_Type_Decl;
Is_Null_2 : Boolean;
Not_Val : Boolean;
If_Result_2 : Logic_Equation;
Fld_15 : Logic_Equation;
True_Rel : Logic_Equation;
Match_Prefix : Internal_Entity;
Cast_Result_2 : Internal_Entity_Expr;
Let_Result : Logic_Equation;
Scope_Result_1 : Logic_Equation;
Cast_Result_3 : Internal_Entity_Subtype_Indication;
Let_Result_1 : Logic_Equation;
Scope_Result_2 : Logic_Equation;
Cast_Result_4 : Internal_Entity_Others_Designator;
Let_Result_2 : Logic_Equation;
Scope_Result_3 : Logic_Equation;
Cast_Result_5 : Internal_Entity;
Let_Result_3 : Logic_Equation;
Scope_Result_4 : Logic_Equation;
Match_Result : Logic_Equation;
Item_352 : Internal_Entity;
Bare_Item_352 : Bare_Ada_Node;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
Item_351 : Internal_Entity_Variant;
Bare_Item_351 : Bare_Variant;
Map_Result_1 : Logic_Equation_Array_Access;
Logic_Boolean_Op_1 : Logic_Equation;
Let_Result_4 : Logic_Equation;
            procedure Finalizer_Scope_296 is
            begin
                     Dec_Ref (Map_Result_1);
                     Dec_Ref (Logic_Boolean_Op_1);
                     Dec_Ref (Let_Result_4);
            end Finalizer_Scope_296;
            procedure Finalizer_Scope_3128 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
            end Finalizer_Scope_3128;
            procedure Finalizer_Scope_3129 is
            begin
                     Dec_Ref (Let_Result);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (Let_Result_1);
                     Dec_Ref (Scope_Result_2);
                     Dec_Ref (Let_Result_2);
                     Dec_Ref (Scope_Result_3);
                     Dec_Ref (Let_Result_3);
                     Dec_Ref (Scope_Result_4);
                     Dec_Ref (Match_Result);
            end Finalizer_Scope_3129;
            procedure Finalizer_Scope_3130 is
            begin
                     Dec_Ref (Fld_8);
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Fld_12);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Fld_13);
                     Dec_Ref (And_Pred_1);
                     Dec_Ref (If_Result_2);
            end Finalizer_Scope_3130;
            procedure Finalizer_Scope_3132 is
            begin
                     Dec_Ref (Fld_15);
            end Finalizer_Scope_3132;
            procedure Finalizer_Scope_3133 is
            begin
                     Dec_Ref (True_Rel);
            end Finalizer_Scope_3133;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11757 '<Block at ???>' Let_Result_4 None
--# scope-start
--# expr-start 11733 '<FieldAccess for resolve_names_internal at ast.py:4570>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4570
--# expr-start 11732 '<FieldAccess for discr_name at ast.py:4570>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4570
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Create_Internal_Entity_Identifier (Node => Ent.Node.Variant_Part_F_Discr_Name, Info => Ent.Info);
--# expr-done 11732
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.resolve_names_internal
Fld_1 := Libadalang.Implementation.Ada_Node_P_Resolve_Names_Internal (Node => Fld.Node, Env => Env, Origin => Origin, E_Info => Fld.Info);
--# end
--# expr-done 11733
Block_Var := Fld_1; 
--# expr-start 11756 '<LogicAll at ast.py:4572>' Logic_Boolean_Op_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4572
--# expr-start 11755 '<Map at ast.py:4572>' Map_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4572
   --# expr-start 11734 '<FieldAccess for variant at ast.py:4572>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4572
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Create_Internal_Entity_Variant_List (Node => Ent.Node.Variant_Part_F_Variant, Info => Ent.Info);
--# expr-done 11734
Fld_3 := Fld_2.Node;
      declare
         Map_Result_1_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_3 = null then
      Map_Result_1 := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Variant_List := Fld_3;
      begin
         for Untyped_Item_351 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_351 := Untyped_Item_351; 
Is_Null := Bare_Item_351 = null; 
if Is_Null then
   If_Result := No_Entity_Variant;
else
Fld_4 := Fld_2.Info;
As_Entity := (Info => Fld_4, Node => Bare_Item_351); 
   If_Result := As_Entity;
end if;
                  Item_351 := If_Result; 
   --# scope-start
               --# bind var Item_351
      --# expr-start 11754 '<LogicAll at ast.py:4573>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4573
--# expr-start 11753 '<Map at ast.py:4573>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4573
   --# expr-start 11735 '<FieldAccess for choices at ast.py:4573>' Fld_5 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4573
      if Item_351.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_5 := Create_Internal_Entity_Alternatives_List (Node => Item_351.Node.Variant_F_Choices, Info => Item_351.Info);
--# expr-done 11735
Fld_6 := Fld_5.Node;
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld_6 = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Alternatives_List := Fld_6;
      begin
         for Untyped_Item_352 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_352 := Untyped_Item_352; 
Is_Null_1 := Bare_Item_352 = null; 
if Is_Null_1 then
   If_Result_1 := No_Entity;
else
Fld_7 := Fld_5.Info;
As_Entity_1 := (Info => Fld_7, Node => Bare_Item_352); 
   If_Result_1 := As_Entity_1;
end if;
                  Item_352 := If_Result_1; 
   --# scope-start
               --# bind c Item_352
      --# expr-start 11752 '<Match at ast.py:4573>' Match_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4573
      if Item_352.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Match_Prefix := Item_352; 
case Ada_Ada_Node (Match_Prefix.Node.Kind) is
         when Ada_Abstract_State_Decl_Expr .. Ada_Un_Op =>
   --# scope-start
--# scope-start
      Cast_Result_2 := Create_Internal_Entity_Expr
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_0 := Cast_Result_2; 
--# bind e Match_0
--# expr-start 11750 '<If at ast.py:4575>' If_Result_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4575
--# expr-start 11739 '<Not at ast.py:4576>' Not_Val /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4576
--# expr-start 11738 '<IsNull at ast.py:4576>' Is_Null_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4576
--# expr-start 11736 '<Cast to Name at ast.py:4576>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4576
Cast_Expr_1 := Match_0; 
   if Cast_Expr_1.Node = null
      or else Cast_Expr_1.Node.Kind in Ada_Name
   then
      Cast_Result_1 := Create_Internal_Entity_Name
        (Node => Cast_Expr_1.Node,
         Info => Cast_Expr_1.Info);
   else
         Cast_Result_1 := No_Entity_Name;
   end if;
--# expr-done 11736
Var_Expr := Cast_Result_1;
if Var_Expr /= No_Entity_Name then
   --# scope-start
--# expr-start 11737 '<FieldAccess for name_designated_type at ast.py:4576>' Fld_14 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4576
      if Var_Expr.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.name_designated_type
Fld_14 := Libadalang.Implementation.Name_P_Name_Designated_Type (Node => Var_Expr.Node, E_Info => Var_Expr.Info);
--# end
--# expr-done 11737
Scope_Result := Fld_14;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Entity_Base_Type_Decl;
end if;
Is_Null_2 := Result_Var.Node = null; 
--# expr-done 11738
Not_Val := not (Is_Null_2); 
--# expr-done 11739
if Not_Val then
   --# expr-start 11741 '<FieldAccess for xref_no_overloading at ast.py:4578>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4578
--# expr-start 11740 '<Cast to Name at ast.py:4578>' Cast_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4578
Cast_Expr := Match_0; 
   if Cast_Expr.Node = null
      or else Cast_Expr.Node.Kind in Ada_Name
   then
      Cast_Result := Create_Internal_Entity_Name
        (Node => Cast_Expr.Node,
         Info => Cast_Expr.Info);
   else
         Cast_Result := No_Entity_Name;
   end if;
--# expr-done 11740
      if Cast_Result.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Name.xref_no_overloading
Fld_8 := Libadalang.Implementation.Name_P_Xref_No_Overloading (Node => Cast_Result.Node, Env => Env, Origin => Origin, E_Info => Cast_Result.Info);
--# end
--# expr-done 11741
   If_Result_2 := Fld_8;
else
   --# expr-start 11749 '<And at ast.py:4580>' And_Pred_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4580
--# expr-start 11747 '<And at ast.py:4580>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4580
--# expr-start 11745 '<Bind at ast.py:4580>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4580
--# expr-start 11742 '<FieldAccess for expected_type_var at ast.py:4580>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4580
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_9 := Libadalang.Implementation.Extensions.Expr_P_Expected_Type_Var (Node => Match_0.Node);
--# expr-done 11742
Fld_9.Value := No_Entity;
Entity_Vars.Reset (Fld_9);
--# expr-start 11744 '<FieldAccess for type_val at ast.py:4580>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4580
--# expr-start 11743 '<FieldAccess for discr_name at ast.py:4580>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4580
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_10 := Self.Variant_Part_F_Discr_Name;
--# expr-done 11743
      if Fld_10 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.type_val
Fld_11 := Libadalang.Implementation.Expr_P_Type_Val (Node => Fld_10);
--# end
--# expr-done 11744
Bind_Result := Solver.Create_Assign (Fld_9, Fld_11, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:4580") else null)); 
--# expr-done 11745
--# expr-start 11746 '<FieldAccess for sub_equation at ast.py:4581>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4581
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.sub_equation
Fld_12 := Libadalang.Implementation.Ada_Node_P_Sub_Equation (Node => Match_0.Node, Env => Env, Origin => Origin, E_Info => Match_0.Info);
--# end
--# expr-done 11746
And_Pred := Create_And (Bind_Result, Fld_12, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:4580") else null)); 
--# expr-done 11747
--# expr-start 11748 '<FieldAccess for matches_expected_type at ast.py:4582>' Fld_13 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4582
      if Match_0.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start Expr.matches_expected_type
Fld_13 := Libadalang.Implementation.Expr_P_Matches_Expected_Type (Node => Match_0.Node, Origin => Origin);
--# end
--# expr-done 11748
And_Pred_1 := Create_And (And_Pred, Fld_13, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:4580") else null)); 
--# expr-done 11749
   If_Result_2 := And_Pred_1;
end if;
      Inc_Ref (If_Result_2);
--# expr-done 11750
Let_Result := If_Result_2; Inc_Ref (Let_Result);
--# end
Scope_Result_1 := Let_Result;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_3130;
            Match_Result := Scope_Result_1; Inc_Ref (Match_Result);
         when Ada_Subtype_Indication .. Ada_Discrete_Subtype_Indication =>
   --# scope-start
--# scope-start
      Cast_Result_3 := Create_Internal_Entity_Subtype_Indication
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_1 := Cast_Result_3; 
--# bind t Match_1
--# expr-start 11751 '<FieldAccess for xref_equation at ast.py:4587>' Fld_15 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4587
      if Match_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]AdaNode.xref_equation'
Fld_15 := Libadalang.Implementation.Dispatcher_Ada_Node_P_Xref_Equation (Node => Match_1.Node, Env => Env, Origin => Origin, E_Info => Match_1.Info);
--# end
--# expr-done 11751
Let_Result_1 := Fld_15; Inc_Ref (Let_Result_1);
--# end
Scope_Result_2 := Let_Result_1;
   Inc_Ref (Scope_Result_2);
   --# end
      Finalizer_Scope_3132;
            Match_Result := Scope_Result_2; Inc_Ref (Match_Result);
         when Ada_Others_Designator =>
   --# scope-start
--# scope-start
      Cast_Result_4 := Create_Internal_Entity_Others_Designator
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_2 := Cast_Result_4; 
--# bind _ Match_2
True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:4589") else null)); 
Let_Result_2 := True_Rel; Inc_Ref (Let_Result_2);
--# end
Scope_Result_3 := Let_Result_2;
   Inc_Ref (Scope_Result_3);
   --# end
      Finalizer_Scope_3133;
            Match_Result := Scope_Result_3; Inc_Ref (Match_Result);
         when Ada_Abort_Absent .. Ada_Elsif_Stmt_Part | Ada_Handled_Stmts .. Ada_Null_Component_Decl | Ada_Overriding_Not_Overriding .. Ada_Enum_Lit_Synth_Type_Expr | Ada_Synthetic_Type_Expr .. Ada_With_Private_Present =>
   --# scope-start
--# scope-start
      Cast_Result_5 := Create_Internal_Entity
        (Node => Match_Prefix.Node,
         Info => Match_Prefix.Info);
Match_3 := Cast_Result_5; 
--# bind _ Match_3
Let_Result_3 := raise Property_Error with "Should not happen"; Inc_Ref (Let_Result_3);
--# end
Scope_Result_4 := Let_Result_3;
   Inc_Ref (Scope_Result_4);
   --# end
            Match_Result := Scope_Result_4; Inc_Ref (Match_Result);
end case;
--# expr-done 11752
         declare
            Item_To_Append : constant Logic_Equation := Match_Result;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_3129;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 11753
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:4573") else null)); 
--# expr-done 11754
         declare
            Item_To_Append : constant Logic_Equation := Logic_Boolean_Op;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_1_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_3128;
         end loop;
      end;
      end if;
         Map_Result_1 := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_1_Vec)));
         for I in Map_Result_1.Items'Range loop
            Map_Result_1.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_1_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result_1.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_1_Vec);
      end;
--# expr-done 11755
Logic_Boolean_Op_1 := Solver.Create_All (Relation_Array (Map_Result_1.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:4572") else null)); 
--# expr-done 11756
Let_Result_4 := Logic_Boolean_Op_1; Inc_Ref (Let_Result_4);
--# end
--# expr-done 11757
      Property_Result := Let_Result_4;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_296;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_296;
               Finalizer_Scope_3128;
               Finalizer_Scope_3129;
               Finalizer_Scope_3130;
               Finalizer_Scope_3132;
               Finalizer_Scope_3133;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Variant_Part_P_Xref_Equation;
--# end
--# property-start VariantPart.get_components /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4596
pragma Warnings (Off, "is not referenced");
 function Variant_Part_P_Get_Components
  (Node : Bare_Variant_Part
      ; Discriminants : Internal_Param_Match_Array_Access
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Internal_Entity_Base_Formal_Param_Decl_Array_Access
is
   Self : Bare_Variant_Part := Bare_Variant_Part (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_Variant_Part :=
     Internal_Entity_Variant_Part'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind discriminants Discriminants
   Property_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_297 with Inline_Always;
      Block_Var : Internal_Param_Match;
Block_Var_1 : Internal_Entity_Variant;
Fld : Internal_Entity_Defining_Name;
Fld_1 : Internal_Entity_Name;
Fld_2 : Bare_Identifier;
Sym : Symbol_Type;
Fld_3 : Boolean;
Find_Result : Internal_Param_Match;
Fld_4 : Internal_Entity_Variant_List;
Fld_5 : Bare_Variant_List;
Fld_6 : Internal_Entity_Info;
As_Entity : Internal_Entity_Variant;
Is_Null : Boolean;
If_Result : Internal_Entity_Variant;
Fld_7 : Internal_Single_Actual;
Fld_8 : Internal_Entity_Basic_Assoc;
Fld_9 : Internal_Entity_Expr;
Fld_10 : Boolean;
Item_354 : Internal_Entity_Variant;
Bare_Item_354 : Bare_Variant;
Find_Result_1 : Internal_Entity_Variant;
Fld_11 : Internal_Entity_Component_List;
Fld_12 : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
Let_Result : Internal_Entity_Base_Formal_Param_Decl_Array_Access;
            procedure Finalizer_Scope_297 is
            begin
                     Dec_Ref (Fld_12);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_297;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11772 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11763 '<Find at ast.py:4602>' Find_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4602
      Find_Result := No_Param_Match;
      declare
         Collection : constant Internal_Param_Match_Array_Access := Discriminants;
      begin
         for Item_353 of
               Collection.Items
         loop
   --# scope-start
               --# bind d Item_353
      --# expr-start 11762 '<FieldAccess for name_is at ast.py:4603>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4603
--# expr-start 11759 '<FieldAccess for name at ast.py:4603>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4603
--# expr-start 11758 '<FieldAccess for formal at ast.py:4603>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4603
Fld := Item_353.Formal;
--# expr-done 11758
      if Fld.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_1 := Create_Internal_Entity_Name (Node => Fld.Node.Defining_Name_F_Name, Info => Fld.Info);
--# expr-done 11759
      if Fld_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11761 '<GetSymbol at ast.py:4603>' Sym /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4603
--# expr-start 11760 '<FieldAccess for discr_name at ast.py:4603>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4603
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_2 := Self.Variant_Part_F_Discr_Name;
--# expr-done 11760
Sym := Get_Symbol (Fld_2); 
--# expr-done 11761
--# property-call-start Name.name_is
Fld_3 := Libadalang.Implementation.Name_P_Name_Is (Node => Fld_1.Node, Sym => Sym);
--# end
--# expr-done 11762
      if Fld_3 then
         Find_Result := Item_353;
      end if;
   --# end
               exit when Fld_3;
         end loop;
      end;
--# expr-done 11763
Block_Var := Find_Result; 
--# bind discr Block_Var
--# expr-start 11769 '<Find at ast.py:4608>' Find_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4608
   --# expr-start 11764 '<FieldAccess for variant at ast.py:4608>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4608
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_4 := Create_Internal_Entity_Variant_List (Node => Ent.Node.Variant_Part_F_Variant, Info => Ent.Info);
--# expr-done 11764
Fld_5 := Fld_4.Node;
      Find_Result_1 := No_Entity_Variant;
      if Fld_5 = null then
      Find_Result_1 := No_Entity_Variant;
      else
      declare
         Collection : constant Bare_Variant_List := Fld_5;
      begin
         for Untyped_Item_354 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Bare_Item_354 := Untyped_Item_354; 
Is_Null := Bare_Item_354 = null; 
if Is_Null then
   If_Result := No_Entity_Variant;
else
Fld_6 := Fld_4.Info;
As_Entity := (Info => Fld_6, Node => Bare_Item_354); 
   If_Result := As_Entity;
end if;
                  Item_354 := If_Result; 
   --# scope-start
               --# bind v Item_354
      --# expr-start 11768 '<FieldAccess for matches at ast.py:4609>' Fld_10 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4609
      if Item_354.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11767 '<FieldAccess for expr at ast.py:4609>' Fld_9 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4609
--# expr-start 11766 '<FieldAccess for assoc at ast.py:4609>' Fld_8 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4609
--# expr-start 11765 '<FieldAccess for actual at ast.py:4609>' Fld_7 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4609
Fld_7 := Block_Var.Actual;
--# expr-done 11765
Fld_8 := Fld_7.Assoc;
--# expr-done 11766
      if Fld_8.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]BasicAssoc.expr'
Fld_9 := Libadalang.Implementation.Dispatcher_Basic_Assoc_P_Expr (Node => Fld_8.Node, E_Info => Fld_8.Info);
--# end
--# expr-done 11767
--# property-call-start Variant.matches
Fld_10 := Libadalang.Implementation.Variant_P_Matches (Node => Item_354.Node, Expr => Fld_9, E_Info => Item_354.Info);
--# end
--# expr-done 11768
      if Fld_10 then
         Find_Result_1 := Item_354;
      end if;
   --# end
               exit when Fld_10;
         end loop;
      end;
      end if;
--# expr-done 11769
Block_Var_1 := Find_Result_1; 
--# bind variant Block_Var_1
--# expr-start 11771 '<FieldAccess for abstract_formal_params_impl at ast.py:4615>' Fld_12 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4615
--# expr-start 11770 '<FieldAccess for components at ast.py:4615>' Fld_11 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4615
      if Block_Var_1.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_11 := Create_Internal_Entity_Component_List (Node => Block_Var_1.Node.Variant_F_Components, Info => Block_Var_1.Info);
--# expr-done 11770
      if Fld_11.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start ComponentList.abstract_formal_params_impl
Fld_12 := Libadalang.Implementation.Component_List_P_Abstract_Formal_Params_Impl (Node => Fld_11.Node, Discriminants => Discriminants, Include_Discriminants => False, Recurse => False, E_Info => Fld_11.Info);
--# end
--# expr-done 11771
Let_Result := Fld_12; Inc_Ref (Let_Result);
--# end
--# expr-done 11772
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_297;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_297;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Variant_Part_P_Get_Components;
--# end
   --
   --  Primitives for Bare_With_Clause
   --
      procedure With_Clause_Pre_Env_Actions
        (Self            : Bare_With_Clause;
         State           : in out PLE_Node_State;
         Add_To_Env_Only : Boolean := False) is
      begin
      declare
         Env : constant Internal_Designated_Env :=
           Libadalang.Implementation.Initial_Env_17 (Node => 
Self);
      begin
         Set_Initial_Env
           (Self,
            State,
            Env,
            DSL_Location => "ast.py:8264");
      end;
      end;
      procedure Initialize_Fields_For_With_Clause
        (Self : Bare_With_Clause
         ; With_Clause_F_Has_Limited : Bare_Limited_Node
         ; With_Clause_F_Has_Private : Bare_Private_Node
         ; With_Clause_F_Packages : Bare_Name_List
        ) is
      begin
            Self.With_Clause_F_Has_Limited := With_Clause_F_Has_Limited;
            Self.With_Clause_F_Has_Private := With_Clause_F_Has_Private;
            Self.With_Clause_F_Packages := With_Clause_F_Packages;
      end Initialize_Fields_For_With_Clause;
   function With_Clause_F_Has_Limited
     (Node : Bare_With_Clause) return Bare_Limited_Node
   is
   begin
         return Node.With_Clause_F_Has_Limited;
   end;
   function With_Clause_F_Has_Private
     (Node : Bare_With_Clause) return Bare_Private_Node
   is
   begin
         return Node.With_Clause_F_Has_Private;
   end;
   function With_Clause_F_Packages
     (Node : Bare_With_Clause) return Bare_Name_List
   is
   begin
         return Node.With_Clause_F_Packages;
   end;
--# property-start WithClause.child_unit_xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8244
pragma Warnings (Off, "is not referenced");
 function With_Clause_P_Child_Unit_Xref_Equation
  (Node : Bare_With_Clause
      ; Name : Bare_Name
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_With_Clause := Bare_With_Clause (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_With_Clause :=
     Internal_Entity_With_Clause'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind name Name
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_675 with Inline_Always;
            procedure Finalizer_Scope_3138 with Inline_Always;
      Block_Var : Logic_Equation;
Var_Expr : Bare_Compilation_Unit;
Var_Expr_1 : Bare_Dotted_Name;
Fld : Logic_Var;
Fld_1 : Bare_Compilation_Unit;
Fld_2 : Bare_Basic_Decl;
Scope_Result : Bare_Basic_Decl;
Result_Var : Bare_Basic_Decl;
As_Entity : Internal_Entity_Basic_Decl;
Is_Null : Boolean;
If_Result : Internal_Entity_Basic_Decl;
Cast_Result : Internal_Entity;
Bind_Result : Logic_Equation;
Cast_Expr : Bare_Name;
Cast_Result_1 : Bare_Dotted_Name;
Fld_3 : Bare_Name;
Fld_4 : Logic_Equation;
Scope_Result_1 : Logic_Equation;
True_Rel : Logic_Equation;
Result_Var_1 : Logic_Equation;
And_Pred : Logic_Equation;
Let_Result : Logic_Equation;
            procedure Finalizer_Scope_675 is
            begin
                     Dec_Ref (Bind_Result);
                     Dec_Ref (Block_Var);
                     Dec_Ref (Scope_Result_1);
                     Dec_Ref (True_Rel);
                     Dec_Ref (Result_Var_1);
                     Dec_Ref (And_Pred);
                     Dec_Ref (Let_Result);
            end Finalizer_Scope_675;
            procedure Finalizer_Scope_3138 is
            begin
                     Dec_Ref (Fld_4);
            end Finalizer_Scope_3138;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11781 '<Block at ???>' Let_Result None
--# scope-start
--# expr-start 11776 '<Bind at ast.py:8250>' Bind_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8250
--# expr-start 11773 '<FieldAccess for ref_var at ast.py:8250>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8250
      if Name = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start '[dispatcher]Name.ref_var'
Fld := Libadalang.Implementation.Dispatcher_Name_P_Ref_Var (Node => Name);
--# end
--# expr-done 11773
Fld.Value := No_Entity;
Entity_Vars.Reset (Fld);
--# expr-start 11774 '<FieldAccess for withed_unit_helper at ast.py:8250>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8250
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start AdaNode.withed_unit_helper
Fld_1 := Libadalang.Implementation.Ada_Node_P_Withed_Unit_Helper (Node => Self, Unit_Name => Name);
--# end
--# expr-done 11774
Var_Expr := Fld_1;
if Var_Expr /= No_Bare_Ada_Node then
   --# scope-start
--# expr-start 11775 '<FieldAccess for decl at ast.py:8250>' Fld_2 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8250
      if Var_Expr = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start CompilationUnit.decl
Fld_2 := Libadalang.Implementation.Compilation_Unit_P_Decl (Node => Var_Expr);
--# end
--# expr-done 11775
Scope_Result := Fld_2;
   --# end
   Result_Var := Scope_Result;
else
   Result_Var := No_Bare_Ada_Node;
end if;
Is_Null := Result_Var = null; 
if Is_Null then
   If_Result := No_Entity_Basic_Decl;
else
As_Entity := (Info => E_Info, Node => Result_Var); 
   If_Result := As_Entity;
end if;
      Cast_Result := Create_Internal_Entity
        (Node => If_Result.Node,
         Info => If_Result.Info);
Bind_Result := Solver.Create_Assign (Fld, Cast_Result, Solver_Ifc.No_Converter, Debug_String => (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8250") else null)); 
--# expr-done 11776
Block_Var := Bind_Result; Inc_Ref (Block_Var);
--# bind self_eq Block_Var
--# expr-start 11780 '<And at ast.py:8251>' And_Pred /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8251
--# expr-start 11777 '<Cast to DottedName at ast.py:8251>' Cast_Result_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8251
Cast_Expr := Name; 
   if Cast_Expr = null
      or else Cast_Expr.Kind in Ada_Dotted_Name_Range
   then
      Cast_Result_1 := Cast_Expr;
   else
         Cast_Result_1 := No_Bare_Ada_Node;
   end if;
--# expr-done 11777
Var_Expr_1 := Cast_Result_1;
if Var_Expr_1 /= No_Bare_Ada_Node then
   --# scope-start
   --# bind dn Var_Expr_1
--# expr-start 11779 '<FieldAccess for child_unit_xref_equation at ast.py:8252>' Fld_4 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8252
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# expr-start 11778 '<FieldAccess for prefix at ast.py:8252>' Fld_3 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8252
      if Var_Expr_1 = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld_3 := Var_Expr_1.Dotted_Name_F_Prefix;
--# expr-done 11778
--# property-call-start WithClause.child_unit_xref_equation
Fld_4 := Libadalang.Implementation.With_Clause_P_Child_Unit_Xref_Equation (Node => Ent.Node, Name => Fld_3, E_Info => Ent.Info);
--# end
--# expr-done 11779
Scope_Result_1 := Fld_4;
   Inc_Ref (Scope_Result_1);
   --# end
      Finalizer_Scope_3138;
   Result_Var_1 := Scope_Result_1;
else
   True_Rel := Solver.Create_True ((if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8253") else null)); 
   Result_Var_1 := True_Rel;
end if;
      Inc_Ref (Result_Var_1);
And_Pred := Create_And (Block_Var, Result_Var_1, (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8251") else null)); 
--# expr-done 11780
Let_Result := And_Pred; Inc_Ref (Let_Result);
--# end
--# expr-done 11781
      Property_Result := Let_Result;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_675;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_675;
               Finalizer_Scope_3138;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end With_Clause_P_Child_Unit_Xref_Equation;
--# end
--# property-start WithClause.xref_entry_point /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8256
pragma Warnings (Off, "is not referenced");
 function With_Clause_P_Xref_Entry_Point
  (Node : Bare_With_Clause
  )
   return Boolean
is
   Self : Bare_With_Clause := Bare_With_Clause (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end With_Clause_P_Xref_Entry_Point;
--# end
--# property-start WithClause.xref_equation /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8257
pragma Warnings (Off, "is not referenced");
 function With_Clause_P_Xref_Equation
  (Node : Bare_With_Clause
      ; Env : Lexical_Env
      ; Origin : Bare_Ada_Node
   ; E_Info : Internal_Entity_Info :=
      No_Entity_Info
  )
   return Logic_Equation
is
   Self : Bare_With_Clause := Bare_With_Clause (Node);
   Call_Depth : aliased Natural;
   Ent : Internal_Entity_With_Clause :=
     Internal_Entity_With_Clause'(Node => Self, Info => E_Info);
   --# bind entity Ent
   --# bind env Env
   --# bind origin Origin
   Property_Result : Logic_Equation;
            procedure Finalizer_Scope_677 with Inline_Always;
            procedure Finalizer_Scope_3139 with Inline_Always;
      Fld : Bare_Name_List;
Fld_1 : Logic_Equation;
Item_355 : Bare_Name;
Map_Result : Logic_Equation_Array_Access;
Logic_Boolean_Op : Logic_Equation;
            procedure Finalizer_Scope_677 is
            begin
                     Dec_Ref (Map_Result);
                     Dec_Ref (Logic_Boolean_Op);
            end Finalizer_Scope_677;
            procedure Finalizer_Scope_3139 is
            begin
                     Dec_Ref (Fld_1);
            end Finalizer_Scope_3139;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self /= null then
         Reset_Caches (Self.Unit);
      end if;
   --# scope-start
      --# expr-start 11785 '<LogicAll at ast.py:8258>' Logic_Boolean_Op /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8258
--# expr-start 11784 '<Map at ast.py:8258>' Map_Result /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8258
   --# expr-start 11782 '<FieldAccess for packages at ast.py:8258>' Fld /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8258
      if Self = null then
         raise Property_Error with "dereferencing a null access";
      end if;
Fld := Self.With_Clause_F_Packages;
--# expr-done 11782
      declare
         Map_Result_Vec : Logic_Equation_Vectors.Vector;
      begin
      if Fld = null then
      Map_Result := Create_Logic_Equation_Array (0);
      else
      declare
         Collection : constant Bare_Name_List := Fld;
      begin
         for Untyped_Item_355 of
               Collection.Nodes (1 .. Children_Count (Collection))
         loop
                  Item_355 := Untyped_Item_355; 
   --# scope-start
               --# bind p Item_355
      --# expr-start 11783 '<FieldAccess for child_unit_xref_equation at ast.py:8259>' Fld_1 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8259
      if Ent.Node = null then
         raise Property_Error with "dereferencing a null access";
      end if;
--# property-call-start WithClause.child_unit_xref_equation
Fld_1 := Libadalang.Implementation.With_Clause_P_Child_Unit_Xref_Equation (Node => Ent.Node, Name => Item_355, E_Info => Ent.Info);
--# end
--# expr-done 11783
         declare
            Item_To_Append : constant Logic_Equation := Fld_1;
         begin
               Inc_Ref (Item_To_Append);
            Logic_Equation_Vectors.Append (Map_Result_Vec, Item_To_Append);
         end;
   --# end
      Finalizer_Scope_3139;
         end loop;
      end;
      end if;
         Map_Result := Create_Logic_Equation_Array
           (Items_Count => Natural (Logic_Equation_Vectors.Length (Map_Result_Vec)));
         for I in Map_Result.Items'Range loop
            Map_Result.Items (I) := Logic_Equation_Vectors.Get
              (Map_Result_Vec,
               I + Logic_Equation_Vectors.Index_Type'First - Map_Result.Items'First);
         end loop;
         Logic_Equation_Vectors.Destroy (Map_Result_Vec);
      end;
--# expr-done 11784
Logic_Boolean_Op := Solver.Create_All (Relation_Array (Map_Result.Items), (if Langkit_Support.Adalog.Debug.Debug then New_Unit_String (Node.Unit, "ast.py:8258") else null)); 
--# expr-done 11785
      Property_Result := Logic_Boolean_Op;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_677;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_677;
               Finalizer_Scope_3139;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end With_Clause_P_Xref_Equation;
--# end
--# property-start WithClause._initial_env_17 /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8264
pragma Warnings (Off, "is not referenced");
 function Initial_Env_17
  (Node : Bare_With_Clause
  )
   return Internal_Designated_Env
is
   Self : Bare_With_Clause := Bare_With_Clause (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Internal_Designated_Env;
            procedure Finalizer_Scope_1434 with Inline_Always;
      Sym : Symbol_Type;
New_Struct : Internal_Designated_Env := No_Designated_Env;
            procedure Finalizer_Scope_1434 is
            begin
                     Dec_Ref (New_Struct);
            end Finalizer_Scope_1434;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      --# expr-start 11787 '<New DesignatedEnv at ast.py:8264>' New_Struct /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:8264
--# expr-start 11786 '<Symbol Standard at ???>' Sym None
Sym := Precomputed_Symbol (Precomputed_Symbol_Table (Self.Unit.Context.Symbols), Precomputed_Sym_Standard); 
--# expr-done 11786
Inc_Ref (Empty_Env);
New_Struct := (Direct_Env => Empty_Env, Env_Name => Sym, Kind => Named_Env); 
--# expr-done 11787
      Property_Result := New_Struct;
         Inc_Ref (Property_Result);
   --# end
      Finalizer_Scope_1434;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when Exc : Precondition_Failure | Property_Error =>
               Finalizer_Scope_1434;
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Initial_Env_17;
--# end
   --
   --  Primitives for Bare_With_Private
   --
--# property-start '[dispatcher]WithPrivate.as_bool' dispatcher
pragma Warnings (Off, "is not referenced");
 function Dispatcher_With_Private_P_As_Bool
  (Node : Bare_With_Private
  )
   return Boolean
is
   Self : Bare_With_Private := Bare_With_Private (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
      if Self = null then
         raise Property_Error with "dispatching on null node";
      end if;
      case Ada_With_Private (Self.Kind) is
               when Ada_With_Private_Absent =>
                  --# property-call-start WithPrivate.Absent.as_bool
                  Property_Result := With_Private_Absent_P_As_Bool
                    (Self
                    );
                  --# end
               when Ada_With_Private_Present =>
                  --# property-call-start WithPrivate.Present.as_bool
                  Property_Result := With_Private_Present_P_As_Bool
                    (Self
                    );
                  --# end
      end case;
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end Dispatcher_With_Private_P_As_Bool;
--# end
   --
   --  Primitives for Bare_With_Private_Absent
   --
--# property-start WithPrivate.Absent.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4938
pragma Warnings (Off, "is not referenced");
 function With_Private_Absent_P_As_Bool
  (Node : Bare_With_Private_Absent
  )
   return Boolean
is
   Self : Bare_With_Private_Absent := Bare_With_Private_Absent (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := False;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end With_Private_Absent_P_As_Bool;
--# end
   --
   --  Primitives for Bare_With_Private_Present
   --
--# property-start WithPrivate.Present.as_bool /it/sbx/wave/x86_64-linux/libadalang/src/ada/ast.py:4938
pragma Warnings (Off, "is not referenced");
 function With_Private_Present_P_As_Bool
  (Node : Bare_With_Private_Present
  )
   return Boolean
is
   Self : Bare_With_Private_Present := Bare_With_Private_Present (Node);
   Call_Depth : aliased Natural;
   --# bind self Self
   Property_Result : Boolean;
begin
   --# property-body-start
   if Self /= null then
      Enter_Call (Self.Unit.Context, Call_Depth'Access);
   end if;
   --# scope-start
      Property_Result := True;
   --# end
   if Self /= null then
      Exit_Call (Self.Unit.Context, Call_Depth);
   end if;
   return Property_Result;
exception
   when others =>
      if Self /= null then
         Exit_Call (Self.Unit.Context, Call_Depth);
      end if;
      raise;
end With_Private_Present_P_As_Bool;
--# end
   ----------------------------
   -- Destroy_Synthetic_Node --
   ----------------------------
   procedure Destroy_Synthetic_Node (Node : in out Bare_Ada_Node) is
      procedure Free is new Ada.Unchecked_Deallocation
        (Root_Node_Record, Bare_Ada_Node);
   begin
      --  Don't call Node.Destroy, as Node's children may be gone already: they
      --  have their own destructor and there is no specified order for the
      --  call of these destructors.
      Free_User_Fields (Node);
      Free (Node);
   end Destroy_Synthetic_Node;
   -----------
   -- Image --
   -----------
   function Image (Value : Boolean) return String
   is (if Value then "True" else "False");
      -----------------
      -- Trace_Image --
      -----------------
      function Trace_Image
        (Node       : Bare_Ada_Node;
         Decoration : Boolean := True) return String is
      begin
         if Node = null then
            return "None";
         else
            declare
               Result : constant String :=
                 (Kind_Name (Node) & " "
                  & Basename (Node.Unit) & ":"
                  & Image (Sloc_Range (Node)));
            begin
               return (if Decoration then "<" & Result & ">" else Result);
            end;
         end if;
      end Trace_Image;
   Kind_Names : array (Ada_Node_Kind_Type) of Unbounded_String :=
     (Ada_Abort_Absent => To_Unbounded_String ("AbortAbsent"), 
Ada_Abort_Present => To_Unbounded_String ("AbortPresent"), 
Ada_Abstract_Absent => To_Unbounded_String ("AbstractAbsent"), 
Ada_Abstract_Present => To_Unbounded_String ("AbstractPresent"), 
Ada_Ada_Node_List => To_Unbounded_String ("AdaNodeList"), 
Ada_Abstract_State_Decl_List => To_Unbounded_String ("AbstractStateDeclList"), 
Ada_Alternatives_List => To_Unbounded_String ("AlternativesList"), 
Ada_Constraint_List => To_Unbounded_String ("ConstraintList"), 
Ada_Decl_List => To_Unbounded_String ("DeclList"), 
Ada_Stmt_List => To_Unbounded_String ("StmtList"), 
Ada_Aspect_Assoc_List => To_Unbounded_String ("AspectAssocList"), 
Ada_Base_Assoc_List => To_Unbounded_String ("BaseAssocList"), 
Ada_Assoc_List => To_Unbounded_String ("AssocList"), 
Ada_Basic_Decl_List => To_Unbounded_String ("BasicDeclList"), 
Ada_Case_Expr_Alternative_List => To_Unbounded_String ("CaseExprAlternativeList"), 
Ada_Case_Stmt_Alternative_List => To_Unbounded_String ("CaseStmtAlternativeList"), 
Ada_Compilation_Unit_List => To_Unbounded_String ("CompilationUnitList"), 
Ada_Contract_Case_Assoc_List => To_Unbounded_String ("ContractCaseAssocList"), 
Ada_Defining_Name_List => To_Unbounded_String ("DefiningNameList"), 
Ada_Discriminant_Spec_List => To_Unbounded_String ("DiscriminantSpecList"), 
Ada_Elsif_Expr_Part_List => To_Unbounded_String ("ElsifExprPartList"), 
Ada_Elsif_Stmt_Part_List => To_Unbounded_String ("ElsifStmtPartList"), 
Ada_Enum_Literal_Decl_List => To_Unbounded_String ("EnumLiteralDeclList"), 
Ada_Expr_Alternatives_List => To_Unbounded_String ("ExprAlternativesList"), 
Ada_Discriminant_Choice_List => To_Unbounded_String ("DiscriminantChoiceList"), 
Ada_Name_List => To_Unbounded_String ("NameList"), 
Ada_Parent_List => To_Unbounded_String ("ParentList"), 
Ada_Param_Spec_List => To_Unbounded_String ("ParamSpecList"), 
Ada_Pragma_Node_List => To_Unbounded_String ("PragmaNodeList"), 
Ada_Select_When_Part_List => To_Unbounded_String ("SelectWhenPartList"), 
Ada_Unconstrained_Array_Index_List => To_Unbounded_String ("UnconstrainedArrayIndexList"), 
Ada_Variant_List => To_Unbounded_String ("VariantList"), 
Ada_Aliased_Absent => To_Unbounded_String ("AliasedAbsent"), 
Ada_Aliased_Present => To_Unbounded_String ("AliasedPresent"), 
Ada_All_Absent => To_Unbounded_String ("AllAbsent"), 
Ada_All_Present => To_Unbounded_String ("AllPresent"), 
Ada_Constrained_Array_Indices => To_Unbounded_String ("ConstrainedArrayIndices"), 
Ada_Unconstrained_Array_Indices => To_Unbounded_String ("UnconstrainedArrayIndices"), 
Ada_Aspect_Assoc => To_Unbounded_String ("AspectAssoc"), 
Ada_At_Clause => To_Unbounded_String ("AtClause"), 
Ada_Attribute_Def_Clause => To_Unbounded_String ("AttributeDefClause"), 
Ada_Enum_Rep_Clause => To_Unbounded_String ("EnumRepClause"), 
Ada_Record_Rep_Clause => To_Unbounded_String ("RecordRepClause"), 
Ada_Aspect_Spec => To_Unbounded_String ("AspectSpec"), 
Ada_Contract_Case_Assoc => To_Unbounded_String ("ContractCaseAssoc"), 
Ada_Pragma_Argument_Assoc => To_Unbounded_String ("PragmaArgumentAssoc"), 
Ada_Entry_Spec => To_Unbounded_String ("EntrySpec"), 
Ada_Enum_Subp_Spec => To_Unbounded_String ("EnumSubpSpec"), 
Ada_Predefined_Bin_Op_Spec => To_Unbounded_String ("PredefinedBinOpSpec"), 
Ada_Predefined_Un_Op_Spec => To_Unbounded_String ("PredefinedUnOpSpec"), 
Ada_Subp_Spec => To_Unbounded_String ("SubpSpec"), 
Ada_Component_List => To_Unbounded_String ("ComponentList"), 
Ada_Known_Discriminant_Part => To_Unbounded_String ("KnownDiscriminantPart"), 
Ada_Unknown_Discriminant_Part => To_Unbounded_String ("UnknownDiscriminantPart"), 
Ada_Entry_Completion_Formal_Params => To_Unbounded_String ("EntryCompletionFormalParams"), 
Ada_Generic_Formal_Part => To_Unbounded_String ("GenericFormalPart"), 
Ada_Null_Record_Def => To_Unbounded_String ("NullRecordDef"), 
Ada_Record_Def => To_Unbounded_String ("RecordDef"), 
Ada_Aggregate_Assoc => To_Unbounded_String ("AggregateAssoc"), 
Ada_Multi_Dim_Array_Assoc => To_Unbounded_String ("MultiDimArrayAssoc"), 
Ada_Composite_Constraint_Assoc => To_Unbounded_String ("CompositeConstraintAssoc"), 
Ada_Iterated_Assoc => To_Unbounded_String ("IteratedAssoc"), 
Ada_Param_Assoc => To_Unbounded_String ("ParamAssoc"), 
Ada_Abstract_State_Decl => To_Unbounded_String ("AbstractStateDecl"), 
Ada_Anonymous_Expr_Decl => To_Unbounded_String ("AnonymousExprDecl"), 
Ada_Component_Decl => To_Unbounded_String ("ComponentDecl"), 
Ada_Discriminant_Spec => To_Unbounded_String ("DiscriminantSpec"), 
Ada_Generic_Formal_Obj_Decl => To_Unbounded_String ("GenericFormalObjDecl"), 
Ada_Generic_Formal_Package => To_Unbounded_String ("GenericFormalPackage"), 
Ada_Generic_Formal_Subp_Decl => To_Unbounded_String ("GenericFormalSubpDecl"), 
Ada_Generic_Formal_Type_Decl => To_Unbounded_String ("GenericFormalTypeDecl"), 
Ada_Param_Spec => To_Unbounded_String ("ParamSpec"), 
Ada_Synthetic_Formal_Param_Decl => To_Unbounded_String ("SyntheticFormalParamDecl"), 
Ada_Generic_Package_Internal => To_Unbounded_String ("GenericPackageInternal"), 
Ada_Package_Decl => To_Unbounded_String ("PackageDecl"), 
Ada_Discrete_Base_Subtype_Decl => To_Unbounded_String ("DiscreteBaseSubtypeDecl"), 
Ada_Subtype_Decl => To_Unbounded_String ("SubtypeDecl"), 
Ada_Classwide_Type_Decl => To_Unbounded_String ("ClasswideTypeDecl"), 
Ada_Incomplete_Type_Decl => To_Unbounded_String ("IncompleteTypeDecl"), 
Ada_Incomplete_Tagged_Type_Decl => To_Unbounded_String ("IncompleteTaggedTypeDecl"), 
Ada_Protected_Type_Decl => To_Unbounded_String ("ProtectedTypeDecl"), 
Ada_Task_Type_Decl => To_Unbounded_String ("TaskTypeDecl"), 
Ada_Single_Task_Type_Decl => To_Unbounded_String ("SingleTaskTypeDecl"), 
Ada_Type_Decl => To_Unbounded_String ("TypeDecl"), 
Ada_Anonymous_Type_Decl => To_Unbounded_String ("AnonymousTypeDecl"), 
Ada_Synth_Anonymous_Type_Decl => To_Unbounded_String ("SynthAnonymousTypeDecl"), 
Ada_Abstract_Subp_Decl => To_Unbounded_String ("AbstractSubpDecl"), 
Ada_Abstract_Formal_Subp_Decl => To_Unbounded_String ("AbstractFormalSubpDecl"), 
Ada_Concrete_Formal_Subp_Decl => To_Unbounded_String ("ConcreteFormalSubpDecl"), 
Ada_Subp_Decl => To_Unbounded_String ("SubpDecl"), 
Ada_Entry_Decl => To_Unbounded_String ("EntryDecl"), 
Ada_Enum_Literal_Decl => To_Unbounded_String ("EnumLiteralDecl"), 
Ada_Synthetic_Char_Enum_Lit => To_Unbounded_String ("SyntheticCharEnumLit"), 
Ada_Generic_Subp_Internal => To_Unbounded_String ("GenericSubpInternal"), 
Ada_Synthetic_Subp_Decl => To_Unbounded_String ("SyntheticSubpDecl"), 
Ada_Expr_Function => To_Unbounded_String ("ExprFunction"), 
Ada_Null_Subp_Decl => To_Unbounded_String ("NullSubpDecl"), 
Ada_Subp_Body => To_Unbounded_String ("SubpBody"), 
Ada_Subp_Renaming_Decl => To_Unbounded_String ("SubpRenamingDecl"), 
Ada_Package_Body_Stub => To_Unbounded_String ("PackageBodyStub"), 
Ada_Protected_Body_Stub => To_Unbounded_String ("ProtectedBodyStub"), 
Ada_Subp_Body_Stub => To_Unbounded_String ("SubpBodyStub"), 
Ada_Task_Body_Stub => To_Unbounded_String ("TaskBodyStub"), 
Ada_Entry_Body => To_Unbounded_String ("EntryBody"), 
Ada_Package_Body => To_Unbounded_String ("PackageBody"), 
Ada_Protected_Body => To_Unbounded_String ("ProtectedBody"), 
Ada_Task_Body => To_Unbounded_String ("TaskBody"), 
Ada_Entry_Index_Spec => To_Unbounded_String ("EntryIndexSpec"), 
Ada_Error_Decl => To_Unbounded_String ("ErrorDecl"), 
Ada_Exception_Decl => To_Unbounded_String ("ExceptionDecl"), 
Ada_Exception_Handler => To_Unbounded_String ("ExceptionHandler"), 
Ada_For_Loop_Var_Decl => To_Unbounded_String ("ForLoopVarDecl"), 
Ada_Generic_Package_Decl => To_Unbounded_String ("GenericPackageDecl"), 
Ada_Generic_Subp_Decl => To_Unbounded_String ("GenericSubpDecl"), 
Ada_Generic_Package_Instantiation => To_Unbounded_String ("GenericPackageInstantiation"), 
Ada_Generic_Subp_Instantiation => To_Unbounded_String ("GenericSubpInstantiation"), 
Ada_Generic_Package_Renaming_Decl => To_Unbounded_String ("GenericPackageRenamingDecl"), 
Ada_Generic_Subp_Renaming_Decl => To_Unbounded_String ("GenericSubpRenamingDecl"), 
Ada_Label_Decl => To_Unbounded_String ("LabelDecl"), 
Ada_Named_Stmt_Decl => To_Unbounded_String ("NamedStmtDecl"), 
Ada_Number_Decl => To_Unbounded_String ("NumberDecl"), 
Ada_Object_Decl => To_Unbounded_String ("ObjectDecl"), 
Ada_Extended_Return_Stmt_Object_Decl => To_Unbounded_String ("ExtendedReturnStmtObjectDecl"), 
Ada_No_Type_Object_Renaming_Decl => To_Unbounded_String ("NoTypeObjectRenamingDecl"), 
Ada_Package_Renaming_Decl => To_Unbounded_String ("PackageRenamingDecl"), 
Ada_Single_Protected_Decl => To_Unbounded_String ("SingleProtectedDecl"), 
Ada_Single_Task_Decl => To_Unbounded_String ("SingleTaskDecl"), 
Ada_Case_Stmt_Alternative => To_Unbounded_String ("CaseStmtAlternative"), 
Ada_Compilation_Unit => To_Unbounded_String ("CompilationUnit"), 
Ada_Component_Clause => To_Unbounded_String ("ComponentClause"), 
Ada_Component_Def => To_Unbounded_String ("ComponentDef"), 
Ada_Constant_Absent => To_Unbounded_String ("ConstantAbsent"), 
Ada_Constant_Present => To_Unbounded_String ("ConstantPresent"), 
Ada_Composite_Constraint => To_Unbounded_String ("CompositeConstraint"), 
Ada_Delta_Constraint => To_Unbounded_String ("DeltaConstraint"), 
Ada_Digits_Constraint => To_Unbounded_String ("DigitsConstraint"), 
Ada_Range_Constraint => To_Unbounded_String ("RangeConstraint"), 
Ada_Declarative_Part => To_Unbounded_String ("DeclarativePart"), 
Ada_Private_Part => To_Unbounded_String ("PrivatePart"), 
Ada_Public_Part => To_Unbounded_String ("PublicPart"), 
Ada_Elsif_Expr_Part => To_Unbounded_String ("ElsifExprPart"), 
Ada_Elsif_Stmt_Part => To_Unbounded_String ("ElsifStmtPart"), 
Ada_Abstract_State_Decl_Expr => To_Unbounded_String ("AbstractStateDeclExpr"), 
Ada_Allocator => To_Unbounded_String ("Allocator"), 
Ada_Aggregate => To_Unbounded_String ("Aggregate"), 
Ada_Bracket_Aggregate => To_Unbounded_String ("BracketAggregate"), 
Ada_Delta_Aggregate => To_Unbounded_String ("DeltaAggregate"), 
Ada_Bracket_Delta_Aggregate => To_Unbounded_String ("BracketDeltaAggregate"), 
Ada_Null_Record_Aggregate => To_Unbounded_String ("NullRecordAggregate"), 
Ada_Bin_Op => To_Unbounded_String ("BinOp"), 
Ada_Relation_Op => To_Unbounded_String ("RelationOp"), 
Ada_Box_Expr => To_Unbounded_String ("BoxExpr"), 
Ada_Case_Expr_Alternative => To_Unbounded_String ("CaseExprAlternative"), 
Ada_Case_Expr => To_Unbounded_String ("CaseExpr"), 
Ada_If_Expr => To_Unbounded_String ("IfExpr"), 
Ada_Contract_Cases => To_Unbounded_String ("ContractCases"), 
Ada_Decl_Expr => To_Unbounded_String ("DeclExpr"), 
Ada_Membership_Expr => To_Unbounded_String ("MembershipExpr"), 
Ada_Attribute_Ref => To_Unbounded_String ("AttributeRef"), 
Ada_Update_Attribute_Ref => To_Unbounded_String ("UpdateAttributeRef"), 
Ada_Call_Expr => To_Unbounded_String ("CallExpr"), 
Ada_Defining_Name => To_Unbounded_String ("DefiningName"), 
Ada_Synthetic_Defining_Name => To_Unbounded_String ("SyntheticDefiningName"), 
Ada_Discrete_Subtype_Name => To_Unbounded_String ("DiscreteSubtypeName"), 
Ada_Dotted_Name => To_Unbounded_String ("DottedName"), 
Ada_End_Name => To_Unbounded_String ("EndName"), 
Ada_Explicit_Deref => To_Unbounded_String ("ExplicitDeref"), 
Ada_Qual_Expr => To_Unbounded_String ("QualExpr"), 
Ada_Reduce_Attribute_Ref => To_Unbounded_String ("ReduceAttributeRef"), 
Ada_Char_Literal => To_Unbounded_String ("Chr"), 
Ada_Identifier => To_Unbounded_String ("Id"), 
Ada_Op_Abs => To_Unbounded_String ("OpAbs"), 
Ada_Op_And => To_Unbounded_String ("OpAnd"), 
Ada_Op_And_Then => To_Unbounded_String ("OpAndThen"), 
Ada_Op_Concat => To_Unbounded_String ("OpConcat"), 
Ada_Op_Div => To_Unbounded_String ("OpDiv"), 
Ada_Op_Double_Dot => To_Unbounded_String ("OpDoubleDot"), 
Ada_Op_Eq => To_Unbounded_String ("OpEq"), 
Ada_Op_Gt => To_Unbounded_String ("OpGt"), 
Ada_Op_Gte => To_Unbounded_String ("OpGte"), 
Ada_Op_In => To_Unbounded_String ("OpIn"), 
Ada_Op_Lt => To_Unbounded_String ("OpLt"), 
Ada_Op_Lte => To_Unbounded_String ("OpLte"), 
Ada_Op_Minus => To_Unbounded_String ("OpMinus"), 
Ada_Op_Mod => To_Unbounded_String ("OpMod"), 
Ada_Op_Mult => To_Unbounded_String ("OpMult"), 
Ada_Op_Neq => To_Unbounded_String ("OpNeq"), 
Ada_Op_Not => To_Unbounded_String ("OpNot"), 
Ada_Op_Not_In => To_Unbounded_String ("OpNotIn"), 
Ada_Op_Or => To_Unbounded_String ("OpOr"), 
Ada_Op_Or_Else => To_Unbounded_String ("OpOrElse"), 
Ada_Op_Plus => To_Unbounded_String ("OpPlus"), 
Ada_Op_Pow => To_Unbounded_String ("OpPow"), 
Ada_Op_Rem => To_Unbounded_String ("OpRem"), 
Ada_Op_Xor => To_Unbounded_String ("OpXor"), 
Ada_String_Literal => To_Unbounded_String ("Str"), 
Ada_Null_Literal => To_Unbounded_String ("Null"), 
Ada_Int_Literal => To_Unbounded_String ("Int"), 
Ada_Real_Literal => To_Unbounded_String ("Real"), 
Ada_Synthetic_Identifier => To_Unbounded_String ("SyntheticIdentifier"), 
Ada_Target_Name => To_Unbounded_String ("TargetName"), 
Ada_Paren_Expr => To_Unbounded_String ("ParenExpr"), 
Ada_Quantified_Expr => To_Unbounded_String ("QuantifiedExpr"), 
Ada_Raise_Expr => To_Unbounded_String ("RaiseExpr"), 
Ada_Un_Op => To_Unbounded_String ("UnOp"), 
Ada_Handled_Stmts => To_Unbounded_String ("HandledStmts"), 
Ada_Interface_Kind_Limited => To_Unbounded_String ("InterfaceKindLimited"), 
Ada_Interface_Kind_Protected => To_Unbounded_String ("InterfaceKindProtected"), 
Ada_Interface_Kind_Synchronized => To_Unbounded_String ("InterfaceKindSynchronized"), 
Ada_Interface_Kind_Task => To_Unbounded_String ("InterfaceKindTask"), 
Ada_Iter_Type_In => To_Unbounded_String ("IterTypeIn"), 
Ada_Iter_Type_Of => To_Unbounded_String ("IterTypeOf"), 
Ada_Library_Item => To_Unbounded_String ("LibraryItem"), 
Ada_Limited_Absent => To_Unbounded_String ("LimitedAbsent"), 
Ada_Limited_Present => To_Unbounded_String ("LimitedPresent"), 
Ada_For_Loop_Spec => To_Unbounded_String ("ForLoopSpec"), 
Ada_While_Loop_Spec => To_Unbounded_String ("WhileLoopSpec"), 
Ada_Mode_Default => To_Unbounded_String ("ModeDefault"), 
Ada_Mode_In => To_Unbounded_String ("ModeIn"), 
Ada_Mode_In_Out => To_Unbounded_String ("ModeInOut"), 
Ada_Mode_Out => To_Unbounded_String ("ModeOut"), 
Ada_Multi_Abstract_State_Decl => To_Unbounded_String ("MultiAbstractStateDecl"), 
Ada_Not_Null_Absent => To_Unbounded_String ("NotNullAbsent"), 
Ada_Not_Null_Present => To_Unbounded_String ("NotNullPresent"), 
Ada_Null_Component_Decl => To_Unbounded_String ("NullComponentDecl"), 
Ada_Others_Designator => To_Unbounded_String ("OthersDesignator"), 
Ada_Overriding_Not_Overriding => To_Unbounded_String ("OverridingNotOverriding"), 
Ada_Overriding_Overriding => To_Unbounded_String ("OverridingOverriding"), 
Ada_Overriding_Unspecified => To_Unbounded_String ("OverridingUnspecified"), 
Ada_Params => To_Unbounded_String ("Params"), 
Ada_Paren_Abstract_State_Decl => To_Unbounded_String ("ParenAbstractStateDecl"), 
Ada_Pp_Else_Directive => To_Unbounded_String ("PpElseDirective"), 
Ada_Pp_Elsif_Directive => To_Unbounded_String ("PpElsifDirective"), 
Ada_Pp_End_If_Directive => To_Unbounded_String ("PpEndIfDirective"), 
Ada_Pp_If_Directive => To_Unbounded_String ("PpIfDirective"), 
Ada_Pp_Then_Kw => To_Unbounded_String ("PpThenKw"), 
Ada_Pragma_Node => To_Unbounded_String ("PragmaNode"), 
Ada_Private_Absent => To_Unbounded_String ("PrivateAbsent"), 
Ada_Private_Present => To_Unbounded_String ("PrivatePresent"), 
Ada_Protected_Def => To_Unbounded_String ("ProtectedDef"), 
Ada_Protected_Absent => To_Unbounded_String ("ProtectedAbsent"), 
Ada_Protected_Present => To_Unbounded_String ("ProtectedPresent"), 
Ada_Quantifier_All => To_Unbounded_String ("QuantifierAll"), 
Ada_Quantifier_Some => To_Unbounded_String ("QuantifierSome"), 
Ada_Range_Spec => To_Unbounded_String ("RangeSpec"), 
Ada_Renaming_Clause => To_Unbounded_String ("RenamingClause"), 
Ada_Synthetic_Renaming_Clause => To_Unbounded_String ("SyntheticRenamingClause"), 
Ada_Reverse_Absent => To_Unbounded_String ("ReverseAbsent"), 
Ada_Reverse_Present => To_Unbounded_String ("ReversePresent"), 
Ada_Select_When_Part => To_Unbounded_String ("SelectWhenPart"), 
Ada_Accept_Stmt => To_Unbounded_String ("AcceptStmt"), 
Ada_Accept_Stmt_With_Stmts => To_Unbounded_String ("AcceptStmtWithStmts"), 
Ada_For_Loop_Stmt => To_Unbounded_String ("ForLoopStmt"), 
Ada_Loop_Stmt => To_Unbounded_String ("LoopStmt"), 
Ada_While_Loop_Stmt => To_Unbounded_String ("WhileLoopStmt"), 
Ada_Begin_Block => To_Unbounded_String ("BeginBlock"), 
Ada_Decl_Block => To_Unbounded_String ("DeclBlock"), 
Ada_Case_Stmt => To_Unbounded_String ("CaseStmt"), 
Ada_Extended_Return_Stmt => To_Unbounded_String ("ExtendedReturnStmt"), 
Ada_If_Stmt => To_Unbounded_String ("IfStmt"), 
Ada_Named_Stmt => To_Unbounded_String ("NamedStmt"), 
Ada_Select_Stmt => To_Unbounded_String ("SelectStmt"), 
Ada_Error_Stmt => To_Unbounded_String ("ErrorStmt"), 
Ada_Abort_Stmt => To_Unbounded_String ("AbortStmt"), 
Ada_Assign_Stmt => To_Unbounded_String ("AssignStmt"), 
Ada_Call_Stmt => To_Unbounded_String ("CallStmt"), 
Ada_Delay_Stmt => To_Unbounded_String ("DelayStmt"), 
Ada_Exit_Stmt => To_Unbounded_String ("ExitStmt"), 
Ada_Goto_Stmt => To_Unbounded_String ("GotoStmt"), 
Ada_Label => To_Unbounded_String ("Label"), 
Ada_Null_Stmt => To_Unbounded_String ("NullStmt"), 
Ada_Raise_Stmt => To_Unbounded_String ("RaiseStmt"), 
Ada_Requeue_Stmt => To_Unbounded_String ("RequeueStmt"), 
Ada_Return_Stmt => To_Unbounded_String ("ReturnStmt"), 
Ada_Terminate_Alternative => To_Unbounded_String ("TerminateAlternative"), 
Ada_Subp_Kind_Function => To_Unbounded_String ("SubpKindFunction"), 
Ada_Subp_Kind_Procedure => To_Unbounded_String ("SubpKindProcedure"), 
Ada_Subunit => To_Unbounded_String ("Subunit"), 
Ada_Synchronized_Absent => To_Unbounded_String ("SynchronizedAbsent"), 
Ada_Synchronized_Present => To_Unbounded_String ("SynchronizedPresent"), 
Ada_Tagged_Absent => To_Unbounded_String ("TaggedAbsent"), 
Ada_Tagged_Present => To_Unbounded_String ("TaggedPresent"), 
Ada_Task_Def => To_Unbounded_String ("TaskDef"), 
Ada_Access_To_Subp_Def => To_Unbounded_String ("AccessToSubpDef"), 
Ada_Anonymous_Type_Access_Def => To_Unbounded_String ("AnonymousTypeAccessDef"), 
Ada_Type_Access_Def => To_Unbounded_String ("TypeAccessDef"), 
Ada_Array_Type_Def => To_Unbounded_String ("ArrayTypeDef"), 
Ada_Derived_Type_Def => To_Unbounded_String ("DerivedTypeDef"), 
Ada_Enum_Type_Def => To_Unbounded_String ("EnumTypeDef"), 
Ada_Formal_Discrete_Type_Def => To_Unbounded_String ("FormalDiscreteTypeDef"), 
Ada_Interface_Type_Def => To_Unbounded_String ("InterfaceTypeDef"), 
Ada_Mod_Int_Type_Def => To_Unbounded_String ("ModIntTypeDef"), 
Ada_Private_Type_Def => To_Unbounded_String ("PrivateTypeDef"), 
Ada_Decimal_Fixed_Point_Def => To_Unbounded_String ("DecimalFixedPointDef"), 
Ada_Floating_Point_Def => To_Unbounded_String ("FloatingPointDef"), 
Ada_Ordinary_Fixed_Point_Def => To_Unbounded_String ("OrdinaryFixedPointDef"), 
Ada_Record_Type_Def => To_Unbounded_String ("RecordTypeDef"), 
Ada_Signed_Int_Type_Def => To_Unbounded_String ("SignedIntTypeDef"), 
Ada_Anonymous_Type => To_Unbounded_String ("AnonymousType"), 
Ada_Enum_Lit_Synth_Type_Expr => To_Unbounded_String ("EnumLitSynthTypeExpr"), 
Ada_Subtype_Indication => To_Unbounded_String ("SubtypeIndication"), 
Ada_Constrained_Subtype_Indication => To_Unbounded_String ("ConstrainedSubtypeIndication"), 
Ada_Discrete_Subtype_Indication => To_Unbounded_String ("DiscreteSubtypeIndication"), 
Ada_Synthetic_Type_Expr => To_Unbounded_String ("SyntheticTypeExpr"), 
Ada_Unconstrained_Array_Index => To_Unbounded_String ("UnconstrainedArrayIndex"), 
Ada_Until_Absent => To_Unbounded_String ("UntilAbsent"), 
Ada_Until_Present => To_Unbounded_String ("UntilPresent"), 
Ada_Use_Package_Clause => To_Unbounded_String ("UsePackageClause"), 
Ada_Use_Type_Clause => To_Unbounded_String ("UseTypeClause"), 
Ada_Value_Sequence => To_Unbounded_String ("ValueSequence"), 
Ada_Variant => To_Unbounded_String ("Variant"), 
Ada_Variant_Part => To_Unbounded_String ("VariantPart"), 
Ada_With_Clause => To_Unbounded_String ("WithClause"), 
Ada_With_Private_Absent => To_Unbounded_String ("WithPrivateAbsent"), 
Ada_With_Private_Present => To_Unbounded_String ("WithPrivatePresent"));
   ---------------
   -- Kind_Name --
   ---------------
   function Kind_Name (Node : Bare_Ada_Node) return String is
   begin
      return To_String (Kind_Names (Node.Kind));
   end Kind_Name;
   --------------------
   -- Children_Count --
   --------------------
   function Children_Count (Node : Bare_Ada_Node) return Natural is
      C : Integer := Kind_To_Node_Children_Count (Node.Kind);
   begin
      if C = -1 then
         return Node.Count;
      else
         return C;
      end if;
   end Children_Count;
   ----------------------
   -- Free_User_Fields --
   ----------------------
   procedure Free_User_Fields (Node : Bare_Ada_Node) is
      procedure Reset_Logic_Var (LV : in out Logic_Var_Record);
      --  Reset the LV logic variable, clearing the value it stores
      ---------------------
      -- Reset_Logic_Var --
      ---------------------
      procedure Reset_Logic_Var (LV : in out Logic_Var_Record) is
      begin
         LV.Value := No_Entity;
         Entity_Vars.Reset (LV'Unrestricted_Access);
         Entity_Vars.Destroy (LV);
      end Reset_Logic_Var;
      K : constant Ada_Node_Kind_Type := Node.Kind;
   begin
      case Ada_Ada_Node (K) is
when Ada_Base_Type_Decl =>
declare
N_Bare_Base_Type_Decl : constant Bare_Base_Type_Decl := Node;
begin
Dec_Ref (N_Bare_Base_Type_Decl.Lf_Stg_Dottable_Subps_Env);
case Ada_Base_Type_Decl (K) is
when Ada_Type_Decl_Range =>
declare
N_Bare_Type_Decl : constant Bare_Type_Decl := N_Bare_Base_Type_Decl;
begin
Dec_Ref (N_Bare_Type_Decl.Lf_Stg_Direct_Primitives_Env);
Dec_Ref (N_Bare_Type_Decl.Lf_Stg_Direct_Primitive_Subps);
end;
when others => null;
end case;
end;
when Ada_Generic_Instantiation =>
declare
N_Bare_Generic_Instantiation : constant Bare_Generic_Instantiation := Node;
begin
Dec_Ref (N_Bare_Generic_Instantiation.Lf_Stg_Instantiation_Env);
Dec_Ref (N_Bare_Generic_Instantiation.Lf_Stg_Actual_Expr_Decls);
end;
when Ada_Compilation_Unit_Range =>
declare
N_Bare_Compilation_Unit : constant Bare_Compilation_Unit := Node;
begin
Dec_Ref (N_Bare_Compilation_Unit.Compilation_Unit_No_Env);
end;
when Ada_Attribute_Ref_Range =>
declare
N_Bare_Attribute_Ref : constant Bare_Attribute_Ref := Node;
begin
Reset_Logic_Var (N_Bare_Attribute_Ref.Attribute_Ref_R_Ref_Var);
end;
when Ada_Call_Expr_Range =>
declare
N_Bare_Call_Expr : constant Bare_Call_Expr := Node;
begin
Reset_Logic_Var (N_Bare_Call_Expr.Call_Expr_R_Called_Spec);
end;
when Ada_Explicit_Deref_Range =>
declare
N_Bare_Explicit_Deref : constant Bare_Explicit_Deref := Node;
begin
Reset_Logic_Var (N_Bare_Explicit_Deref.Explicit_Deref_R_Called_Spec);
end;
when Ada_Reduce_Attribute_Ref_Range =>
declare
N_Bare_Reduce_Attribute_Ref : constant Bare_Reduce_Attribute_Ref := Node;
begin
Reset_Logic_Var (N_Bare_Reduce_Attribute_Ref.Reduce_Attribute_Ref_R_Ref_Var);
end;
when Ada_Target_Name_Range =>
declare
N_Bare_Target_Name : constant Bare_Target_Name := Node;
begin
Reset_Logic_Var (N_Bare_Target_Name.Target_Name_R_Ref_Var);
end;
when others => null;
end case;
   end Free_User_Fields;
   ----------------
   -- Token_Data --
   ----------------
   function Token_Data (Unit : Internal_Unit) return Token_Data_Handler_Access
   is (Unit.TDH'Access);
   -------------------
   -- Lookup_Symbol --
   -------------------
   function Lookup_Symbol
     (Context : Internal_Context; Symbol : Text_Type) return Symbol_Type
   is
      Canon_Symbol : constant Symbolization_Result :=
            Libadalang.Sources.Canonicalize (Symbol)
      ;
   begin
      if Canon_Symbol.Success then
         return Get_Symbol
           (Context.Symbols, Find (Context.Symbols, Canon_Symbol.Symbol));
      else
         raise Invalid_Symbol_Error with Image (Canon_Symbol.Error_Message);
      end if;
   end Lookup_Symbol;
   -------------------------
   -- Create_Special_Unit --
   -------------------------
   function Create_Special_Unit
     (Context             : Internal_Context;
      Normalized_Filename : Virtual_File;
      Charset             : String;
      Rule                : Grammar_Rule) return Internal_Unit
   is
      Unit : Internal_Unit := new Analysis_Unit_Type'
        (Context                      => Context,
         Is_Internal                  => False,
         Ast_Root                     => null,
         Filename                     => Normalized_Filename,
         Charset                      => To_Unbounded_String (Charset),
         TDH                          => <>,
         Diagnostics                  => <>,
         Is_Env_Populated             => False,
         Rule                         => Rule,
         Ast_Mem_Pool                 => No_Pool,
         Destroyables                 => Destroyable_Vectors.Empty_Vector,
         Referenced_Units             => <>,
         Exiled_Entries               => Exiled_Entry_Vectors.Empty_Vector,
         Foreign_Nodes                =>
            Foreign_Node_Entry_Vectors.Empty_Vector,
         Exiled_Entries_In_NED        =>
            Exiled_Entry_In_NED_Vectors.Empty_Vector,
         Exiled_Envs                  => Exiled_Env_Vectors.Empty_Vector,
         Named_Envs                   => Named_Env_Vectors.Empty_Vector,
         Nodes_With_Foreign_Env       => <>,
         Rebindings                   => Env_Rebindings_Vectors.Empty_Vector,
         Cache_Version                => <>,
         Unit_Version                 => <>,
         Memoization_Map            => <>,
         others => <>
      );
   begin
      Initialize
        (Unit.TDH, Context.Symbols, Unit.all'Address, Context.Tab_Stop);
      return Unit;
   end Create_Special_Unit;
   --------------------
   -- Templates_Unit --
   --------------------
   function Templates_Unit (Context : Internal_Context) return Internal_Unit is
   begin
      if Context.Templates_Unit = No_Analysis_Unit then
         Context.Templates_Unit := Create_Special_Unit
           (Context             => Context,
            Normalized_Filename => No_File,
            Charset             => Default_Charset,
            Rule                => Compilation_Rule);
      end if;
      return Context.Templates_Unit;
   end Templates_Unit;
   --------------
   -- Set_Rule --
   --------------
   procedure Set_Rule (Unit : Internal_Unit; Rule : Grammar_Rule) is
   begin
      Unit.Rule := Rule;
   end Set_Rule;
   ------------------------------
   -- Normalized_Unit_Filename --
   ------------------------------
   function Normalized_Unit_Filename
     (Context : Internal_Context; Filename : String) return Virtual_File
   is
      use Virtual_File_Maps;
      Key : constant Unbounded_String := To_Unbounded_String (Filename);
      Cur : Cursor := Context.Filenames.Find (Key);
   begin
      if Cur = No_Element then
         declare
            F : constant Virtual_File := Create
              (Create_From_Base (+Filename).Full_Name,
               Normalize => True);
         begin
            Context.Filenames.Insert (Key, F);
            return F;
         end;
      else
         return Element (Cur);
      end if;
   end Normalized_Unit_Filename;
   --------------------------
   -- Register_Destroyable --
   --------------------------
   procedure Register_Destroyable_Helper
     (Unit    : Internal_Unit;
      Object  : System.Address;
      Destroy : Destroy_Procedure)
   is
   begin
      Destroyable_Vectors.Append (Unit.Destroyables, (Object, Destroy));
   end Register_Destroyable_Helper;
   --------------------------
   -- Register_Destroyable --
   --------------------------
   procedure Register_Destroyable
     (Unit : Internal_Unit; Node : Bare_Ada_Node)
   is
      procedure Helper is new Register_Destroyable_Gen
        (Root_Node_Record,
         Bare_Ada_Node,
         Destroy_Synthetic_Node);
   begin
      Helper (Unit, Node);
   end Register_Destroyable;
   --------------------------
   -- Register_Destroyable --
   --------------------------
   procedure Register_Destroyable
     (Unit : Internal_Unit; Env : AST_Envs.Lexical_Env_Access)
   is
      procedure Helper is new Register_Destroyable_Gen
        (AST_Envs.Lexical_Env_Record, AST_Envs.Lexical_Env_Access, Destroy);
   begin
      Helper (Unit, Env);
   end Register_Destroyable;
   -----------------------
   -- Invalidate_Caches --
   -----------------------
   procedure Invalidate_Caches
     (Context : Internal_Context; Invalidate_Envs : Boolean) is
   begin
      --  Increase Context's version number. If we are about to overflow, reset
      --  all version numbers from analysis units.
      if Context.Cache_Version = Version_Number'Last then
         Context.Cache_Version := 1;
         for Unit of Context.Units loop
            Unit.Cache_Version := 0;
         end loop;
      else
         Context.Cache_Version := Context.Cache_Version + 1;
      end if;
      if Invalidate_Envs then
         Context.Reparse_Cache_Version := Context.Cache_Version;
      end if;
   end Invalidate_Caches;
   ------------------
   --  Reset_Envs  --
   ------------------
   procedure Reset_Envs (Unit : Internal_Unit) is
      procedure Deactivate_Refd_Envs (Node : Bare_Ada_Node);
      procedure Recompute_Refd_Envs (Node : Bare_Ada_Node);
      --------------------------
      -- Deactivate_Refd_Envs --
      --------------------------
      procedure Deactivate_Refd_Envs (Node : Bare_Ada_Node) is
      begin
         if Node = null then
            return;
         end if;
         Deactivate_Referenced_Envs (Node.Self_Env);
         for I in 1 .. Children_Count (Node) loop
            Deactivate_Refd_Envs (Child (Node, I));
         end loop;
      end Deactivate_Refd_Envs;
      -------------------------
      -- Recompute_Refd_Envs --
      -------------------------
      procedure Recompute_Refd_Envs (Node : Bare_Ada_Node) is
      begin
         if Node = null then
            return;
         end if;
         Recompute_Referenced_Envs (Node.Self_Env);
         for I in 1 .. Children_Count (Node) loop
            Recompute_Refd_Envs (Child (Node, I));
         end loop;
      end Recompute_Refd_Envs;
   begin
      --  First pass will deactivate every referenced envs that Unit possesses
      Deactivate_Refd_Envs (Unit.Ast_Root);
      --  Second pass will recompute the env they are pointing to
      Recompute_Refd_Envs (Unit.Ast_Root);
   end Reset_Envs;
   -------------
   -- Destroy --
   -------------
   procedure Destroy (Reparsed : in out Reparsed_Unit) is
   begin
      Free (Reparsed.TDH);
      Reparsed.Diagnostics := Diagnostics_Vectors.Empty_Vector;
      Free (Reparsed.Ast_Mem_Pool);
      Reparsed.Ast_Root := null;
   end Destroy;
   --------------
   -- Basename --
   --------------
   function Basename (Filename : String) return String is
   begin
      return +Create (+Filename).Base_Name;
   end Basename;
   --------------
   -- Basename --
   --------------
   function Basename (Unit : Internal_Unit) return String is
   begin
      return +Unit.Filename.Base_Name;
   end Basename;
   ------------------
   -- Reset_Caches --
   ------------------
   procedure Reset_Caches (Unit : Internal_Unit) is
      Cache_Version : constant Version_Number := Unit.Cache_Version;
   begin
      if Cache_Version < Unit.Context.Reparse_Cache_Version then
         Unit.Cache_Version := Unit.Context.Reparse_Cache_Version;
         Reset_Envs (Unit);
      end if;
      if Cache_Version < Unit.Context.Cache_Version then
         Unit.Cache_Version := Unit.Context.Cache_Version;
            Destroy (Unit.Memoization_Map);
      end if;
   end Reset_Caches;
   --------------------
   -- Reference_Unit --
   --------------------
   procedure Reference_Unit (From, Referenced : Internal_Unit) is
      Dummy : Boolean;
   begin
      Dummy := Analysis_Unit_Sets.Add (From.Referenced_Units, Referenced);
   end Reference_Unit;
   ------------------------
   -- Is_Referenced_From --
   ------------------------
   function Is_Referenced_From
     (Self, Unit : Internal_Unit) return Boolean is
   begin
      if Unit = null or else Self = null then
         return False;
      elsif Unit = Self then
         return True;
      else
         return Analysis_Unit_Sets.Has (Unit.Referenced_Units, Self);
      end if;
   end Is_Referenced_From;
   ----------------
   -- Do_Parsing --
   ----------------
   procedure Do_Parsing
     (Unit   : Internal_Unit;
      Input  : Internal_Lexer_Input;
      Result : out Reparsed_Unit)
   is
      Context  : constant Internal_Context := Unit.Context;
      Unit_TDH : constant Token_Data_Handler_Access := Token_Data (Unit);
      Saved_TDH : Token_Data_Handler;
      --  Holder to save tokens data in Unit.
      --
      --  By design, parsing is required to bind the nodes it creates to an
      --  analysis unit. However, this procedure is supposed to preserve the
      --  Unit itself and return its parsing result in Result.
      --
      --  In order to implement this, we first move "old" token data in this
      --  variable, then we do parsing. Only then, we can move "new" token data
      --  from the unit to Result, and restore the "old" token data to Unit.
      --  This last step is what Rotate_TDH (see below) is above.
      procedure Rotate_TDH;
      --  Move token data from Unit to Result and restore data in Saved_TDH to
      --  Unit.
      ----------------
      -- Rotate_TDH --
      ----------------
      procedure Rotate_TDH is
      begin
         Move (Result.TDH, Unit_TDH.all);
         Move (Unit_TDH.all, Saved_TDH);
      end Rotate_TDH;
   begin
      GNATCOLL.Traces.Trace (Main_Trace, "Parsing unit " & Basename (Unit));
      Result.Ast_Root := null;
      Move (Saved_TDH, Unit_TDH.all);
      Initialize (Unit_TDH.all,
                  Saved_TDH.Symbols,
                  Unit.all'Address,
                  Unit.Context.Tab_Stop);
      --  This is where lexing occurs, so this is where we get most "setup"
      --  issues: missing input file, bad charset, etc. If we have such an
      --  error, catch it, turn it into diagnostics and abort parsing.
      --
      --  As it is quite common, first check if the file is readable: if not,
      --  don't bother opening it and directly emit a diagnostic. This avoid
      --  pointless exceptions which harm debugging. Note that this
      --  optimization is valid only when there is no file reader, which can
      --  work even when there is no real source file.
      if Context.File_Reader = null
         and then Input.Kind = File
         and then (Input.Filename.Is_Directory
                   or else (not Input.Filename.Is_Readable))
      then
         declare
            Name : constant String := Basename (Unit);
         begin
            GNATCOLL.Traces.Trace
              (Main_Trace, "WARNING: File is not readable: " & Name);
            Append
              (Result.Diagnostics,
               No_Source_Location_Range,
               "Cannot read " & To_Text (Name));
            Rotate_TDH;
            return;
         end;
      end if;
      --  Initialize the parser, which fetches the source buffer and extract
      --  all tokens.
      Init_Parser
        (Input, Context.With_Trivia, Unit, Unit_TDH, Unit.Context.Parser);
      --  If we could run the lexer, run the parser and get the root node
      if Unit_TDH.Source_Buffer /= null then
         Result.Ast_Mem_Pool := Create;
         Unit.Context.Parser.Mem_Pool := Result.Ast_Mem_Pool;
         Result.Ast_Root := Bare_Ada_Node
           (Parse (Unit.Context.Parser, Rule => Unit.Rule));
      end if;
      --  Forward token data and diagnostics to the returned unit
      Rotate_TDH;
      --  TODO we use a for loop rt. than the new ``Append_Vector`` here for
      --  compatibility with old compilers, but someday we'll be able to get
      --  rid of it.
      for Diag of Unit.Context.Parser.Diagnostics loop
         Result.Diagnostics.Append (Diag);
      end loop;
   end Do_Parsing;
   --------------------------
   -- Update_After_Reparse --
   --------------------------
   procedure Update_After_Reparse
     (Unit : Internal_Unit; Reparsed : in out Reparsed_Unit) is
   begin
      --  Remove the `symbol -> AST node` associations for Unit's nodes in
      --  foreign lexical environments.
      Remove_Exiled_Entries (Unit);
      --  Remove the named envs that Unit created
      declare
         Named_Envs_Needing_Update : NED_Maps.Map;
      begin
         Remove_Named_Envs (Unit, Named_Envs_Needing_Update);
         Update_Named_Envs (Unit.Context, Named_Envs_Needing_Update);
      end;
      --  At this point, envs and nodes that don't belong to this unit no
      --  longer reference this unit's envs and nodes. It is thus now safe to
      --  deallocate this unit's obsolete data.
      --  Replace Unit's diagnostics by Reparsed's
      Unit.Diagnostics := Reparsed.Diagnostics;
      Reparsed.Diagnostics.Clear;
      --  As (re-)loading a unit can change how any AST node property in the
      --  whole analysis context behaves, we have to invalidate caches. This
      --  is likely overkill, but kill all caches here as it's easy to do.
      --
      --  As an optimization, invalidate referenced envs cache only if this is
      --  not the first time we parse Unit.
      Invalidate_Caches
        (Unit.Context, Invalidate_Envs => Unit.Ast_Root /= null);
      --  Likewise for token data
      Free (Unit.TDH);
      Move (Unit.TDH, Reparsed.TDH);
      --  Reparsing will invalidate all lexical environments related to this
      --  unit, so destroy all related rebindings as well. This browses AST
      --  nodes, so we have to do this before destroying the old AST nodes
      --  pool.
      Destroy_Rebindings (Unit.Rebindings'Access);
      --  Destroy the old AST node and replace it by the new one
      if Unit.Ast_Root /= null then
         Destroy (Unit.Ast_Root);
      end if;
      Unit.Ast_Root := Reparsed.Ast_Root;
      --  Likewise for memory pools
      Free (Unit.Ast_Mem_Pool);
      Unit.Ast_Mem_Pool := Reparsed.Ast_Mem_Pool;
      Reparsed.Ast_Mem_Pool := No_Pool;
      --  Increment unit version number to invalidate caches and stale node
      --  reference. Also propagate it to the TDH.
      Unit.Unit_Version := Unit.Unit_Version + 1;
      Unit.TDH.Version := Unit.Unit_Version;
      --  If Unit had its lexical environments populated, re-populate them
      if not Unit.Is_Env_Populated then
         return;
      end if;
      declare
         Unit_Name     : constant String := +Unit.Filename.Base_Name;
         Context       : constant Internal_Context := Unit.Context;
         Foreign_Nodes : Bare_Ada_Node_Vectors.Vector :=
            Bare_Ada_Node_Vectors.Empty_Vector;
         Saved_In_Populate_Lexical_Env : constant Boolean :=
            Context.In_Populate_Lexical_Env;
      begin
         GNATCOLL.Traces.Trace
           (Main_Trace, "Updating lexical envs for " & Unit_Name
                        & " after reparse");
         GNATCOLL.Traces.Increase_Indent (Main_Trace);
         Context.In_Populate_Lexical_Env := True;
         --  Collect all nodes that are foreign in this Unit's lexical envs.
         --  Exclude them from the corresponding lists of exiled entries.
         Extract_Foreign_Nodes (Unit, Foreign_Nodes);
         --  Reset the flag so that the call to Populate_Lexical_Env below does
         --  its work.
         Unit.Is_Env_Populated := False;
         --  Now that Unit has been reparsed, we can destroy all its
         --  destroyables, which refer to the old tree (i.e. dangling
         --  pointers).
         Destroy_Unit_Destroyables (Unit);
         for FN of Foreign_Nodes loop
            declare
               Node_Image : constant String := Image (Short_Text_Image (FN));
               Unit_Name  : constant String := +FN.Unit.Filename.Base_Name;
            begin
               GNATCOLL.Traces.Trace
                 (Main_Trace, "Rerooting: " & Node_Image
                              & " (from " & Unit_Name & ")");
            end;
            Reroot_Foreign_Node (FN);
         end loop;
         Foreign_Nodes.Destroy;
         Populate_Lexical_Env (Unit);
         Context.In_Populate_Lexical_Env := Saved_In_Populate_Lexical_Env;
         GNATCOLL.Traces.Decrease_Indent (Main_Trace);
      end;
   end Update_After_Reparse;
   -------------------------------
   -- Destroy_Unit_Destroyables --
   -------------------------------
   procedure Destroy_Unit_Destroyables (Unit : Internal_Unit) is
   begin
      for D of Unit.Destroyables loop
         D.Destroy (D.Object);
      end loop;
      Destroyable_Vectors.Clear (Unit.Destroyables);
   end Destroy_Unit_Destroyables;
   ---------------------------
   -- Remove_Exiled_Entries --
   ---------------------------
   procedure Remove_Exiled_Entries (Unit : Internal_Unit) is
   begin
      for EE of Unit.Exiled_Entries loop
         AST_Envs.Remove (EE.Env, EE.Key, EE.Node);
         --  Also strip foreign nodes information from "outer" units so that it
         --  does not contain stale information (i.e. dangling pointers to
         --  nodes that belong to the units in the queue).
         if EE.Env.Owner /= No_Generic_Unit then
            declare
               Foreign_Nodes : Foreign_Node_Entry_Vectors.Vector renames
                  Convert_Unit (EE.Env.Owner).Foreign_Nodes;
               Current       : Positive := Foreign_Nodes.First_Index;
            begin
               while Current <= Foreign_Nodes.Last_Index loop
                  if Foreign_Nodes.Get (Current).Node = EE.Node then
                     Foreign_Nodes.Pop (Current);
                  else
                     Current := Current + 1;
                  end if;
               end loop;
            end;
         end if;
      end loop;
      Unit.Exiled_Entries.Clear;
   end Remove_Exiled_Entries;
   -----------------------
   -- Remove_Named_Envs --
   -----------------------
   procedure Remove_Named_Envs
     (Unit                      : Internal_Unit;
      Named_Envs_Needing_Update : in out NED_Maps.Map) is
   begin
      --  Remove nodes in this unit from the Named_Env_Descriptor.Foreign_Nodes
      --  components in which they are registered and from the foreign
      --  environments themselves.
      for EE of Unit.Exiled_Entries_In_NED loop
         Remove (EE.Named_Env.Foreign_Nodes, EE.Key, EE.Node);
         Remove (EE.Named_Env.Env_With_Precedence, EE.Key, EE.Node);
      end loop;
      Unit.Exiled_Entries_In_NED.Clear;
      --  Remove nodes in this unit from the
      --  Named_Env_Descriptor.Nodes_With_Foreign_Env components in which they
      --  are registered.
      for Cur in Unit.Nodes_With_Foreign_Env.Iterate loop
         declare
            use Node_To_Named_Env_Maps;
            Node : constant Bare_Ada_Node := Key (Cur);
            NE   : constant Named_Env_Descriptor_Access := Element (Cur);
         begin
            NE.Nodes_With_Foreign_Env.Delete (Node);
         end;
      end loop;
      Unit.Nodes_With_Foreign_Env.Clear;
      --  Remove ends in this unit from the Named_Env_Descriptor.Foreign_Envs
      --  components in which they are registered.
      for EE of Unit.Exiled_Envs loop
         EE.Named_Env.Foreign_Envs.Delete (Env_Node (EE.Env));
      end loop;
      Unit.Exiled_Envs.Clear;
      --  Remove named envs that this unit created
      for NE of Unit.Named_Envs loop
         declare
            NED_Access : constant Named_Env_Descriptor_Access :=
               Unit.Context.Named_Envs.Element (NE.Name);
            NED        : Named_Env_Descriptor renames NED_Access.all;
         begin
            NED.Envs.Delete (Env_Node (NE.Env));
            --  If this named environment had precedence, we must schedule an
            --  update for this name environment entry.
            if NE.Env = NED.Env_With_Precedence then
               Named_Envs_Needing_Update.Include (NE.Name, NED_Access);
               NED.Env_With_Precedence := Empty_Env;
            end if;
         end;
      end loop;
      Unit.Named_Envs.Clear;
   end Remove_Named_Envs;
   ---------------------------
   -- Extract_Foreign_Nodes --
   ---------------------------
   procedure Extract_Foreign_Nodes
     (Unit          : Internal_Unit;
      Foreign_Nodes : in out Bare_Ada_Node_Vectors.Vector) is
   begin
      --  Go through all foreign nodes registered in Unit's lexical
      --  environments.
      for FN of Unit.Foreign_Nodes loop
         --  Collect them
         Foreign_Nodes.Append (FN.Node);
         --  For each foreign node, remove the corresponding exiled entry in
         --  that foreign unit (each foreign node in unit A has a corresponding
         --  exiled entry in unit B).
         declare
            Exiled_Entries : Exiled_Entry_Vectors.Vector renames
               FN.Unit.Exiled_Entries;
            Current        : Positive := Exiled_Entries.First_Index;
         begin
            while Current <= Exiled_Entries.Last_Index loop
               if Exiled_Entries.Get (Current).Node = FN.Node then
                  Exiled_Entries.Pop (Current);
               else
                  Current := Current + 1;
               end if;
            end loop;
         end;
      end loop;
      Unit.Foreign_Nodes.Clear;
   end Extract_Foreign_Nodes;
   --------------------------
   -- Reroot_Foreign_Nodes --
   --------------------------
   procedure Reroot_Foreign_Node (Node : Bare_Ada_Node) is
      Unit : constant Internal_Unit := Node.Unit;
   begin
      --  First, filter the exiled entries in foreign units so that they don't
      --  contain references to this unit's lexical environments.  We need to
      --  do that before running the partial Populate_Lexical_Env pass so that
      --  we don't remove exiled entries that this pass will produce.
      declare
         Exiled_Entries : Exiled_Entry_Vectors.Vector renames
            Unit.Exiled_Entries;
         Current        : Positive := Exiled_Entries.First_Index;
      begin
         while Current <= Exiled_Entries.Last_Index loop
            if Exiled_Entries.Get (Current).Node = Node then
               Exiled_Entries.Pop (Current);
            else
               Current := Current + 1;
            end if;
         end loop;
      end;
      --  Re-do a partial Populate_Lexical_Env pass for each foreign node that
      --  this unit contains so that they are relocated in our new lexical
      --  environments.
      declare
         Unit_State : aliased PLE_Unit_State :=
           (Named_Envs_Needing_Update => <>);
         State      : PLE_Node_State :=
           (Unit_State  => Unit_State'Unchecked_Access,
            Current_Env => Node.Self_Env,
            Current_NED => null);
      begin
         Pre_Env_Actions (Node, State, Add_To_Env_Only => True);
         Post_Env_Actions (Node, State);
      end;
   end Reroot_Foreign_Node;
   ----------
   -- Text --
   ----------
   function Text (Node : Bare_Ada_Node) return String_Type is
   begin
      return Create_String (Text (Node));
   end Text;
   ------------------------
   -- Destroy_Rebindings --
   ------------------------
   procedure Destroy_Rebindings
     (Rebindings : access Env_Rebindings_Vectors.Vector)
   is
      procedure Recurse (R : in out Env_Rebindings);
      --  Destroy R's children and then destroy R. It is up to the caller to
      --  remove R from its parent's Children vector.
      procedure Unregister
        (R          : Env_Rebindings;
         Rebindings : in out Env_Rebindings_Vectors.Vector);
      --  Remove R from Rebindings
      -------------
      -- Recurse --
      -------------
      procedure Recurse (R : in out Env_Rebindings) is
      begin
         for C of R.Children loop
            declare
               C_Var : Env_Rebindings := C;
            begin
               Recurse (C_Var);
            end;
         end loop;
         R.Children.Destroy;
         Unregister (R, Convert_Unit (R.Old_Env.Owner).Rebindings);
         Unregister (R, Convert_Unit (R.New_Env.Owner).Rebindings);
         Release_Rebinding (R);
      end Recurse;
      ----------------
      -- Unregister --
      ----------------
      procedure Unregister
        (R          : Env_Rebindings;
         Rebindings : in out Env_Rebindings_Vectors.Vector) is
      begin
         for I in 1 .. Rebindings.Length loop
            if Rebindings.Get (I) = R then
               Rebindings.Pop (I);
               return;
            end if;
         end loop;
         --  We are always supposed to find R in Rebindings, so this should be
         --  unreachable.
         raise Program_Error;
      end Unregister;
   begin
      while Rebindings.Length > 0 loop
         declare
            R : Env_Rebindings := Rebindings.Get (1);
         begin
            --  Here, we basically undo what has been done in AST_Envs.Append
            --  If this rebinding has no parent, then during its creation we
            --  registered it in its Old_Env. Otherwise, it is registered
            --  in its Parent's Children list.
            if R.Parent = null then
               Unwrap (R.Old_Env).Rebindings_Pool.Delete (R.New_Env);
            else
               Unregister (R, R.Parent.Children);
            end if;
            --  In all cases it's registered in Old_Env's and New_Env's units
            Recurse (R);
         end;
      end loop;
   end Destroy_Rebindings;
   --------------------------
   -- Get_Rewriting_Handle --
   --------------------------
   function Get_Rewriting_Handle
     (Context : Internal_Context) return Rewriting_Handle_Pointer is
   begin
      return Context.Rewriting_Handle;
   end Get_Rewriting_Handle;
   --------------------------
   -- Set_Rewriting_Handle --
   --------------------------
   procedure Set_Rewriting_Handle
     (Context : Internal_Context; Handle : Rewriting_Handle_Pointer) is
   begin
      Context.Rewriting_Handle := Handle;
   end Set_Rewriting_Handle;
   -----------------------
   -- Create_Safety_Net --
   -----------------------
   function Create_Safety_Net
     (Context : Internal_Context) return Iterator_Safety_Net
   is
   begin
      return (Context         => Context,
              Context_Serial  => Context.Serial_Number,
              Context_Version => Context.Cache_Version);
   end Create_Safety_Net;
   ----------------------
   -- Check_Safety_Net --
   ----------------------
   procedure Check_Safety_Net (Self : Iterator_Safety_Net) is
   begin
      if Self.Context = null then
         return;
      end if;
      --  Check that the context is still the same (not released nor reused)
      if Self.Context.Serial_Number /= Self.Context_Serial
         or else Self.Context.Cache_Version /= Self.Context_Version
      then
         raise Stale_Reference_Error;
      end if;
   end Check_Safety_Net;
   ----------------------
   -- String_To_Symbol --
   ----------------------
   function String_To_Symbol
     (Context : Internal_Context; S : String_Type) return Symbol_Type is
   begin
      return (if S.Length > 0
              then Lookup_Symbol (Context, S.Content)
              else null);
   exception
      when Exc : Invalid_Symbol_Error =>
         raise Property_Error with Ada.Exceptions.Exception_Message (Exc);
   end String_To_Symbol;
   -------------
   -- Inc_Ref --
   -------------
   procedure Inc_Ref (Self : String_Type) is
   begin
      if Self.Ref_Count >= 0 then
         Self.Ref_Count := Self.Ref_Count + 1;
      end if;
   end Inc_Ref;
   -------------
   -- Dec_Ref --
   -------------
   procedure Dec_Ref (Self : in out String_Type) is
   begin
      if Self = null or else Self.Ref_Count < 0 then
         return;
      end if;
      if Self.Ref_Count = 1 then
         Free (Self);
      else
         Self.Ref_Count := Self.Ref_Count - 1;
         Self := null;
      end if;
   end Dec_Ref;
   -------------------
   -- Create_String --
   -------------------
   function Create_String (Content : Text_Type) return String_Type is
   begin
      return Result : constant String_Type := new String_Record'
        (Length    => Content'Length,
         Ref_Count => 1,
         Content   => Content);
   end Create_String;
   -------------------
   -- Create_String --
   -------------------
   function Create_String (Content : Unbounded_Text_Type) return String_Type is
      S : Big_Wide_Wide_String_Access;
      L : Natural;
   begin
      Get_Wide_Wide_String (Content, S, L);
      return Create_String (S.all (1 .. L));
   end Create_String;
   -------------------
   -- Concat_String --
   -------------------
   function Concat_String (Left, Right : String_Type) return String_Type is
   begin
      return Result : constant String_Type :=
        new String_Record (Length => Left.Length + Right.Length)
      do
         Result.Ref_Count := 1;
         Result.Content (1 .. Left.Length) := Left.Content;
         Result.Content (Left.Length + 1 .. Result.Length) := Right.Content;
      end return;
   end Concat_String;
   ----------------
   -- Equivalent --
   ----------------
   function Equivalent (Left, Right : String_Type) return Boolean is
   begin
      return Left.Content = Right.Content;
   end Equivalent;
begin
   No_Big_Integer.Value.Set (0);
end Libadalang.Implementation;